#define OLDNSNQC 0
#define NEWNSNQC 1

/*
 *  Copyright (C) 2010 Craig Robinson, Yi Pan, Andrea Taracchini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with with program; see the file COPYING. If not, write to the
 *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA  02110-1301  USA
 */


/**
 * \author Craig Robinson, Yi Pan, Andrea Taracchini
 *
 * \brief More recent versions of the EOB models, such as EOBNRv2 and SEOBNRv1, utilise
 * a non-quasicircular correction (NQC) to bring the peak of the EOB frequency
 * into agreement with that of NR simulations. This file contains the functions
 * used to calculate these NQC corrections, described in DCC document T1100433.
 * The fits to NR peak amplitude, frequency, and their derivatives, are taken
 * from Pan et al. PRD 84 124052 (2011), for EOBNRv2, and
 * from Taracchini et al. PRD 86, 024011 (2012), for SEOBNRv1.
 */


/**
 * Function to interpolate known amplitude NQC coeffcients of spin terms,
 * namely a3s, a4 and a5.
 * The a3s, a4 and a5 values were calculated for
 * 11 mass ratios q=1,1.5,2,3,4,5,6,10,20,50 and 100, and
 * 19 spin (\f$\chi\f$ defined in Taracchini et al. PRD 86, 024011 (2012)) values
 * chi = -1, -0.9, -0.8, ......, 0.3, 0.4, 0.5, 0.55, 0.6, 0.65.
 * The calculation was done by Andrea Taracchini using a C++ code of the UMaryland group.
 * In principle, these numbers can be automatically calculated iteratively by the LAL code.
 * However, since such calcualtion increase the cost of each waveform generation by
 * about an order of magnitude, we prepare these numbers in advance reduce cost.
 * These number can be verified by confirming that
 * the peak amplitude and frequency agree well with the NR-fits predicted values,
 * and to get exact NR-fits predicted values, corrections on these numbers are ~1%.
 */
UNUSED static int XLALSimIMRGetEOBCalibratedSpinNQC( EOBNonQCCoeffs * restrict coeffs,                                   INT4 UNUSED l,                                   INT4 UNUSED m,                                   REAL8 eta,                                   REAL8 a )
{
    const unsigned int nsqdim = 100;
    const unsigned int   qdim = 50;
    const unsigned int   adim = 69;
    UINT4 i;
    /* REAL8 eta2 = eta*eta;*/
    REAL8 a3slist[qdim], a4list[qdim], a5list[qdim];
    
    memset( coeffs, 0, sizeof( *coeffs ) );
    const double nsetalist[100] = {0.0025, 0.005, 0.0075, 0.01, 0.0125, 0.015, 0.0175, 0.02, 0.0225, 0.025, 0.0275, 0.03, 0.0325, 0.035, 0.0375, 0.04, 0.0425, 0.045, 0.0475, 0.05, 0.0525, 0.055, 0.0575, 0.06, 0.0625, 0.065, 0.0675, 0.07, 0.0725, 0.075, 0.0775, 0.08, 0.0825, 0.085, 0.0875, 0.09, 0.0925, 0.095, 0.0975, 0.1, 0.1025, 0.105, 0.1075, 0.11, 0.1125, 0.115, 0.1175, 0.12, 0.1225, 0.125, 0.1275, 0.13, 0.1325, 0.135, 0.1375, 0.14, 0.1425, 0.145, 0.1475, 0.15, 0.1525, 0.155, 0.1575, 0.16, 0.1625, 0.165, 0.1675, 0.17, 0.1725, 0.175, 0.1775, 0.18, 0.1825, 0.185, 0.1875, 0.19, 0.1925, 0.195, 0.1975, 0.2, 0.2025, 0.205, 0.2075, 0.21, 0.2125, 0.215, 0.2175, 0.22, 0.2225, 0.225, 0.2275, 0.23, 0.2325, 0.235, 0.2375, 0.24, 0.2425, 0.245, 0.2475, 0.25};
    const double etalist[50] = {0.005, 0.01, 0.015, 0.02, 0.025, 0.03, 0.035, 0.04, 0.045, 0.05, 0.055, 0.06, 0.065, 0.07, 0.075, 0.08, 0.085, 0.09, 0.095, 0.1, 0.105, 0.11, 0.115, 0.12, 0.125, 0.13, 0.135, 0.14, 0.145, 0.15, 0.155, 0.16, 0.165, 0.17, 0.175, 0.18, 0.185, 0.19, 0.195, 0.2, 0.205, 0.21, 0.215, 0.22, 0.225, 0.23, 0.235, 0.24, 0.245, 0.25};
    const double alist[69]   = {-1., -0.975, -0.95, -0.925, -0.9, -0.875, -0.85, -0.825, -0.8, -0.775, -0.75, -0.725, -0.7, -0.675, -0.65, -0.625, -0.6, -0.575, -0.55, -0.525, -0.5, -0.475, -0.45, -0.425, -0.4, -0.375, -0.35, -0.325, -0.3, -0.275, -0.25, -0.225, -0.2, -0.175, -0.15, -0.125, -0.1, -0.075, -0.05, -0.025, 0., 0.025, 0.05, 0.075, 0.1, 0.125, 0.15, 0.175, 0.2, 0.225, 0.25, 0.275, 0.3, 0.325, 0.35, 0.375, 0.4, 0.425, 0.45, 0.475, 0.5, 0.525, 0.55, 0.575, 0.6, 0.625, 0.65, 0.675, 0.7};
    
    const double a1list[100] = {-21.648106754277453, -20.47585498319721, -19.530321799953853, -18.760004570030972, -18.079213572570588, -17.419521042795658, -16.859312290615286, -16.28756533894047, -15.772560753656803, -15.303367029222859, -14.847798623544204, -14.404050939670224, -13.976897526988164, -13.578815297784823, -13.200763216702892, -12.841976980076495, -12.501692284239752, -12.179144825526786, -11.865421415437213, -11.561721630442458, -11.267369305715642, -10.980678895252732, -10.699964853049693, -10.42829543384451, -10.167524207712544, -9.914674799750443, -9.669372681081281, -9.43124332282813, -9.199912196114067, -8.97500477206216, -8.756146521795488, -8.542962916437121, -8.335079427110134, -8.130867290476582, -7.9313984345769715, -7.736518055359128, -7.545988962963793, -7.35957396753171, -7.177035879203623, -6.998068318971551, -6.8224288840577945, -6.649961333309703, -6.480441962265538, -6.313647066463564, -6.149352941442041, -5.9873358827392344, -5.827372185893404, -5.669238146442815, -5.512513869477759, -5.347444298469859, -5.183197664196798, -5.019955297249104, -4.857898528217308, -4.697208687691938, -4.538067106263523, -4.380655114522592, -4.225154043059675, -4.071745222465302, -3.9206099833299994, -3.771929656244297, -3.6258855717987246, -3.4826590605838113, -3.342431453190086, -3.2053840802080775, -3.0798495456033685, -2.957143501994997, -2.837141996613968, -2.7197210766912865, -2.6047567894579564, -2.4921251821449832, -2.381702301983373, -2.273364196204128, -2.1669869120382543, -2.0624464967167566, -1.9596189974706397, -1.8495581080328225, -1.7409749049276515, -1.6340876818580403, -1.5291147325269023, -1.4262743506371505, -1.3257848298916983, -1.2278644639934593, -1.1327315466453465, -1.0406043715502744, -0.9517012324111543, -0.8662404229309005, -0.7844402368124264, -0.7065189677586451, -0.6326949094724702, -0.5631863556568149, -0.4982116000145925, -0.4379889362487165, -0.3827366580621, -0.3326730591576565, -0.28801643323829923, -0.24898507400694198, -0.21579727516649713, -0.18867133041987855, -0.1678255334699995, -0.15347817801977337};
    
    const double a2list[100] = {178.73204288078207, 167.7345170263427, 158.85457776976878, 151.63702661020528, 145.25886103554546, 139.08071349079492, 133.85300186061994, 128.49833153024582, 123.69581397508206, 119.34304449990263, 115.1225036726121, 111.02138715464824, 107.08437843884623, 103.42908442981195, 99.96941370234742, 96.69762600347904, 93.6059810802332, 90.68673867963636, 87.86042720725196, 85.13659365364732, 82.50827361644501, 79.95961751850369, 77.47477578268207, 75.08044751075573, 72.79299640785891, 70.58542621544011, 68.45392194344085, 66.39466860180258, 64.40385120046682, 62.47765474937501, 60.61226425846867, 58.80386473768926, 57.04864119697828, 55.32723154023535, 53.65375061411054, 52.0271070701423, 50.44518833065014, 48.905881817953585, 47.407074954372156, 45.94482445612376, 44.51687981462722, 43.123270309615144, 41.76224017946817, 40.43203366256695, 39.13089499729211, 37.85706842202429, 36.60879817514412, 35.384328495032264, 34.180608559988464, 32.932965816541305, 31.701640999519693, 30.48755199799916, 29.291616701055233, 28.11475299776345, 26.95787877719934, 25.821911928438432, 24.70777034055626, 23.616371902628366, 22.54863450373026, 21.50547603293748, 20.487814379325563, 19.496567431970035, 18.53265307994643, 17.596989212330282, 16.733055464585682, 15.895477726977264, 15.083579856295431, 14.296685709330593, 13.53411914287315, 12.795204013713512, 12.079264178642092, 11.385623494449275, 10.71360581792548, 10.06253500586111, 9.431734915046569, 8.755551062926262, 8.098379005319565, 7.462063310457031, 6.848448546569211, 6.259379281886656, 5.69670008463992, 5.162255523059555, 4.657890165376111, 4.1854485798201475, 3.7467753346222055, 3.3437149980128416, 2.978112138222609, 2.651811323482059, 2.366657122021744, 2.124494102072216, 1.9271668318640272, 1.77651987962773, 1.674397813593876, 1.6226452019930173, 1.6231066130557064, 1.6776266150124943, 1.7880497760939342, 1.956220664530578, 2.1839838485529777, 2.4731838963916855};
    
    const double a3list[100] = {-198.6684740510964, -185.71560983427335, -175.26102024642407, -166.7798147654891, -159.28986302859136, -152.03810090459336, -145.91734205946045, -139.6355596843493, -134.01653939524905, -128.93945143613453, -124.02034829337536, -119.24645127356649, -114.67037356959001, -110.43118685036107, -106.4264567068743, -102.6466423946065, -99.08220316903453, -95.72359828563523, -92.4798228915885, -89.36096908139594, -86.35837723133453, -83.45329697039512, -80.62697792756852, -77.90968987612793, -75.32016588735331, -72.82709457524146, -70.42579806564999, -68.11159848443651, -65.87981795745866, -63.72577861057402, -61.64480256964021, -59.63221196051487, -57.683328909055604, -55.7731074201788, -53.92035933426217, -52.12397499502319, -50.38150684315194, -48.69050731933857, -47.04852886427319, -45.45070004394427, -43.89439067703587, -42.37998904664536, -40.90552002179141, -39.46900847149271, -38.068479264767944, -36.70195727063581, -35.36746735811498, -34.063034396224154, -32.78536874375959, -31.468633548362497, -30.173781283368523, -28.901550623295833, -27.652680242662605, -26.427908815987003, -25.227975017787198, -24.053617522581362, -22.905575004887663, -21.78458613922428, -20.691389600109368, -19.6267240620611, -18.59132819959765, -17.585940687237187, -16.61130019949788, -15.668145410897901, -14.794309398041404, -13.950185357101319, -13.135122694543167, -12.348470816832473, -11.589579130434757, -10.857797041815543, -10.152473957440362, -9.47295928377472, -8.818602427284151, -8.188752794434174, -7.582759791690314, -6.932160232642741, -6.304213548518645, -5.700899806714059, -5.124199074625014, -4.576091419647542, -4.0585569091776765, -3.573575610611449, -3.123127591344892, -2.709192918774041, -2.3337516602949204, -1.9987838833035663, -1.706269655196011, -1.458189043368287, -1.2565221152164259, -1.1032489381364605, -1.0003495795244224, -0.9498041067763443, -0.953592587288258, -1.0136950884561957, -1.1320916776761898, -1.31076242234427, -1.5516873898564725, -1.8568466476088277, -2.2282202629973678, -2.667788303418125};
    
    const double b1list[100] = {-0.5693500504085347, -0.576434151837257, -0.5827940588889807, -0.5875005969333106, -0.5915255507494274, -0.5970658827548452, -0.6057016775611604, -0.6053160995270499, -0.6070988602490128, -0.6110941958474475, -0.6140262971912503, -0.6172989788502661, -0.6206089494421659, -0.6234488672149939, -0.6258813681301192, -0.6277498501118401, -0.628897711084455, -0.6291683489722621, -0.6308371388180571, -0.6331613723912005, -0.6359581432020665, -0.6393457874435003, -0.6434426413083475, -0.6478966252652467, -0.6524003236766366, -0.6571425380351175, -0.6620695856777987, -0.6671277839417898, -0.6722634501642004, -0.6774229016821397, -0.6825524558327174, -0.6875984299530429, -0.6925071413802255, -0.6961049747777461, -0.6996298196832709, -0.7032241904115, -0.7069570372346382, -0.7108973104248908, -0.7151139602544623, -0.720064149835206, -0.7258460785616149, -0.7320745022572427, -0.7387427060855165, -0.7458439752098636, -0.7533715947937112, -0.7613188500004865, -0.7696790259936167, -0.778445407936529, -0.7876035918776259, -0.7967733093076456, -0.8063159350249182, -0.8162406283658405, -0.8265565486668099, -0.8372728552642233, -0.8483987074944775, -0.8599432646939698, -0.871915686199097, -0.884325131346256, -0.8971807594718443, -0.9104917299122585, -0.9242672020038958, -0.938516335083153, -0.9532482884864274, -0.9684722215501159, -0.9837886712042127, -0.9996512322126136, -1.0160843677317775, -1.0331125409181625, -1.0507602149282274, -1.0690518529184303, -1.08801191804523, -1.1076648734650847, -1.1280351823344532, -1.1491473078097938, -1.171025713047565, -1.1939017528688474, -1.2175927015073384, -1.2421149961497486, -1.2674850739827876, -1.2937193721931661, -1.320834327967594, -1.3488463784927816, -1.377771960955439, -1.4076275125422764, -1.4384294704400042, -1.4701942718353325, -1.5029383539149717, -1.5366781538656316, -1.5714301088740226, -1.6072106561268549, -1.6440362328108387, -1.6819232761126843, -1.7208882232191016, -1.7609475113168012, -1.802117577592493, -1.8444148592328868, -1.8878557934246936, -1.9324568173546235, -1.9782343682093864, -2.0252048831756926};
    
    const double b2list[100] = {1.6514745488753086, 1.6733593678301482, 1.687838328174986, 1.7031979370575185, 1.712831020475929, 1.7266279186283089, 1.7581796869631672, 1.7499867318965905, 1.7518398412177276, 1.7634468469918447, 1.770740685014047, 1.779639998248617, 1.788893228893931, 1.7964389585725973, 1.8024779675983216, 1.8063408969988246, 1.8073583878018264, 1.8048610810350476, 1.8077536017385523, 1.8130592620946404, 1.8200051503317694, 1.8290042591854243, 1.8404695813910905, 1.8531136802761718, 1.8658884601302625, 1.8795206835759333, 1.8938451458175296, 1.908696642059398, 1.923909967505884, 1.9393199173613336, 1.9547612868300928, 1.9700688711165077, 1.9850774654249244, 1.9960816492286722, 2.0069990026439246, 2.0182845185376195, 2.0301606467081585, 2.042849836953944, 2.056574539073377, 2.0726094682881886, 2.0912563552368337, 2.111506538675078, 2.1333773752936596, 2.1568862217833127, 2.1820504348347747, 2.208887371138781, 2.2374143873860683, 2.2676488402673725, 2.299578947658098, 2.3318064402049257, 2.3657435940219895, 2.4014679228653333, 2.439056940491002, 2.478588160655038, 2.520139097113486, 2.56378726362239, 2.609610173937794, 2.657685341815741, 2.708090281012276, 2.760902505283443, 2.8161995283852854, 2.874058864073847, 2.9345580261051722, 2.9977745282353045, 3.0615596376827443, 3.128412230361185, 3.1984931979224718, 3.271963432018449, 3.3489838243009618, 3.429715266421855, 3.5143186500329735, 3.602954866786163, 3.6957848083332676, 3.792969366326133, 3.894669432416603, 4.002044979667201, 4.114256382857298, 4.23142577617022, 4.353675293789295, 4.48112706989785, 4.613903238679211, 4.752125934316707, 4.895917290993665, 5.04539944289341, 5.200694524199272, 5.361924669094577, 5.529212011762653, 5.702678686386827, 5.882446827150425, 6.068638568236777, 6.261376043829206, 6.460781388111044, 6.666976735265615, 6.880084219476247, 7.100225974926268, 7.327524135799001, 7.5621008362777795, 7.804078210545929, 8.053578392786775, 8.310723517183646};
    
    const double a3stab[50][69] = {
        {1298.87913, 1261.29637, 1227.72908, 1197.25947, 1168.96973, 1141.94207, 1115.2587, 1088.00181, 1059.25362, 1022.90618, 985.307903, 947.617046, 910.991876, 881.346609, 853.181173, 825.751447, 798.313312, 767.87765, 736.843339, 705.36426, 673.594293, 642.125047, 610.497581, 578.690684, 546.683144, 514.497622, 482.051486, 449.305976, 416.222332, 382.495195, 348.459044, 314.181759, 279.731222, 245.411071, 210.959123, 176.348956, 141.554145, 106.628243, 71.43286, 35.9095836, 0., -36.1519021, -72.887899, -110.347365, -148.669675, -186.733245, -226.44279, -268.442067, -313.374835, -358.91977, -409.871742, -468.06054, -535.315953, -602.768787, -687.227409, -794.8012, -931.599543, -1073.78457, -1269.39181, -1536.50955, -1893.22609, -2326.65745, -2947.80866, -3741.68954, -4974.29064, -6365.10282, -9538.63496, -15643.1414, -25826.8766},
        {1196.32002, 1167.06104, 1137.27475, 1107.12301, 1076.76769, 1046.37066, 1016.09379, 986.098937, 956.547978, 928.8469, 901.415801, 873.918898, 846.020409, 814.568078, 783.16919, 752.614552, 723.694973, 702.877684, 683.006503, 662.601671, 640.183429, 607.718815, 572.902552, 536.876163, 500.78117, 469.351999, 438.700105, 408.529845, 378.545579, 347.703945, 316.75611, 285.705522, 254.555628, 223.39562, 192.108901, 160.664621, 129.031928, 97.2453174, 65.1824526, 32.7863434, 0., -33.0547348, -66.6852171, -101.01997, -136.187517, -170.971669, -207.383546, -246.089556, -287.756108, -330.313665, -378.258958, -433.352771, -497.35589, -562.191692, -643.393334, -746.656564, -877.67713, -1013.79966, -1200.41146, -1454.54875, -1793.24772, -2204.40576, -2792.4755, -3542.23306, -4707.36116, -6024.93464, -9029.55323, -14806.4354, -24440.7998},
        {1111.34089, 1088.04208, 1061.93086, 1033.65502, 1003.86233, 973.200551, 942.317471, 911.86086, 882.478492, 858.393797, 835.248631, 812.260507, 788.646935, 759.48951, 729.796029, 700.438372, 672.288418, 651.387883, 631.370875, 611.041341, 589.203225, 559.502697, 527.964588, 495.455951, 462.843844, 433.627925, 404.989603, 376.74289, 348.701797, 320.308767, 291.898012, 263.432172, 234.873889, 206.130982, 177.242843, 148.194039, 118.969142, 89.6533759, 60.0903922, 30.2244979, 0., -30.4005164, -61.3663335, -93.0484555, -125.597886, -157.893074, -191.866601, -228.178495, -267.488782, -307.871972, -353.607815, -406.390546, -467.914398, -530.2754, -608.605274, -708.437534, -835.305696, -967.651586, -1148.93709, -1395.53239, -1723.80769, -2122.58454, -2690.87904, -3414.09589, -4533.73512, -5812.62405, -8672.18555, -14119.1936, -23160.4223},
        {1039.53251, 1020.2822, 997.603313, 972.198968, 944.772298, 916.02643, 886.664491, 857.389611, 828.904917, 804.925294, 781.93741, 759.439688, 736.930555, 712.636669, 687.836926, 662.538457, 636.748394, 609.788544, 582.625489, 555.540492, 528.814811, 504.169828, 479.870634, 455.622441, 431.13046, 404.996091, 378.469884, 351.698574, 324.8289, 298.520437, 272.201947, 245.815032, 219.301294, 192.43116, 165.385874, 138.17551, 110.810137, 83.4601202, 55.9111218, 28.1090968, 0., -28.1232207, -56.8004007, -86.2243824, -116.588008, -146.981092, -179.140716, -213.700934, -251.2958, -290.060701, -334.127824, -385.130693, -444.702827, -504.716717, -580.471327, -677.504592, -801.354447, -931.385611, -1109.77852, -1352.54038, -1675.67843, -2068.96254, -2627.11195, -3337.00139, -4427.97475, -5693.23265, -8428.0796, -13552.2941, -21985.6547},
        {976.485655, 959.824144, 940.197989, 918.098323, 894.016279, 868.442993, 841.869596, 814.787222, 787.687004, 761.819814, 736.613152, 712.254257, 688.930369, 670.53532, 652.067116, 632.23036, 609.729651, 574.459963, 537.459377, 500.956344, 467.179318, 448.413342, 432.807639, 418.568025, 403.900314, 381.127188, 356.690849, 331.150363, 305.064801, 280.549729, 255.985115, 231.307428, 206.453135, 181.110154, 155.562924, 129.847331, 103.999264, 78.2732522, 52.3990841, 26.3251899, 0., -26.1568213, -52.856571, -80.3393122, -108.845108, -137.719354, -168.454647, -201.64892, -237.900101, -275.345863, -318.028498, -367.530041, -425.432527, -483.212445, -556.599593, -651.218222, -772.692584, -901.046982, -1077.7456, -1318.65266, -1639.63241, -2031.30849, -2585.26688, -3290.67293, -4364.64226, -5631.82205, -8258.783, -13076.6149, -20916.4077},
        {917.791081, 902.710621, 885.620781, 866.696557, 846.112943, 824.044936, 800.66753, 776.155722, 750.684506, 722.455778, 694.406874, 667.502029, 642.705478, 629.711227, 617.261835, 602.829629, 583.886937, 541.782439, 496.561572, 452.146121, 412.457876, 398.926373, 390.962551, 385.485097, 379.412698, 359.691907, 337.202401, 312.851722, 287.547412, 264.607418, 241.564715, 218.362682, 194.944699, 170.981969, 146.79892, 122.447804, 97.9808711, 73.7004736, 49.308722, 24.7578271, 0., -24.435292, -49.4039968, -75.1848062, -102.056412, -129.591492, -159.057156, -191.014502, -226.024627, -262.193469, -303.519346, -351.545419, -407.814847, -463.459392, -534.598171, -626.938905, -746.189313, -872.680952, -1047.64817, -1286.94915, -1606.44207, -1997.39115, -2549.4365, -3254.83387, -4318.29988, -5593.45384, -8125.84343, -12663.0342, -19952.5918},
        {859.039559, 844.98436, 829.777581, 813.337142, 795.580959, 776.426953, 755.793041, 733.597141, 709.757173, 680.211609, 650.449591, 621.980817, 596.314984, 586.690157, 578.196318, 567.651816, 551.875002, 508.136269, 460.621105, 413.967044, 372.811617, 362.402057, 358.522317, 357.566049, 355.926907, 338.360939, 317.554443, 294.556114, 270.414647, 248.904279, 227.257944, 205.434118, 183.391278, 160.860605, 138.11879, 115.215226, 92.1993096, 69.3494866, 46.3944783, 23.2920582, 0., -22.8926065, -46.3118306, -70.5524255, -95.9091441, -122.081137, -150.197001, -180.789729, -214.392317, -249.069531, -288.809882, -335.133657, -389.561139, -443.154361, -512.075161, -602.027124, -718.713836, -842.332775, -1014.29608, -1250.50977, -1566.87987, -1954.97924, -2503.71346, -3209.20757, -4263.50982, -5543.18962, -7990.80852, -12282.43, -19094.1177},
        {795.821854, 782.688084, 768.57428, 753.36355, 736.939001, 719.18374, 699.980874, 679.213512, 656.764759, 628.465728, 599.872319, 572.488437, 547.817986, 537.997872, 529.6458, 520.012472, 506.348594, 469.901748, 430.32701, 391.276333, 356.401671, 345.533527, 339.673884, 336.003275, 331.702235, 314.804975, 295.296878, 274.017005, 251.804419, 231.651088, 211.382003, 190.97506, 170.40816, 149.560065, 128.547461, 107.387899, 86.098929, 64.8279929, 43.4107958, 21.8129331, 0., -21.4627385, -43.4492248, -66.2337315, -90.0905311, -114.671921, -141.12294, -169.966649, -201.726112, -234.440059, -272.109617, -316.251583, -368.382751, -419.994158, -486.638661, -573.843361, -687.135356, -806.047703, -972.49916, -1202.41444, -1511.71827, -1891.84153, -2432.19041, -3133.5174, -4174.83432, -5446.09099, -7815.22592, -11905.6805, -18340.896},
        {723.728735, 711.864516, 697.916767, 682.119252, 664.705738, 645.909991, 625.965777, 605.106863, 583.567013, 560.596558, 537.806075, 515.822704, 495.273586, 480.160139, 466.385516, 453.227147, 439.962463, 423.459176, 406.36832, 388.931212, 371.389169, 355.013918, 338.6042, 321.989168, 304.997976, 286.694707, 267.979608, 248.987859, 229.854639, 211.058618, 192.254089, 173.438833, 154.610635, 135.894352, 117.109862, 98.2041202, 79.1240795, 59.7436944, 40.1121174, 20.2055015, 0., -20.0796619, -40.6853319, -62.0202854, -84.2877981, -106.847476, -131.083733, -157.53731, -186.748953, -216.771064, -251.628063, -292.856026, -341.991034, -391.675587, -455.896772, -539.748099, -648.323078, -759.870989, -917.067264, -1135.7431, -1431.72971, -1795.74674, -2318.96003, -3007.48672, -4026.8356, -5267.21955, -7560.6433, -11503.6636, -17692.8373},
        {670.702392, 658.608735, 645.231211, 630.627242, 614.85425, 597.969656, 580.030884, 561.095353, 541.220486, 518.602027, 495.903746, 473.927738, 453.476095, 438.689549, 425.696101, 413.962387, 402.955046, 391.824136, 380.479507, 368.51443, 355.522175, 339.17332, 321.752905, 303.623275, 285.146778, 267.626973, 250.10851, 232.577251, 215.019058, 197.328079, 179.618576, 161.913096, 144.234187, 126.856941, 109.450342, 91.9359207, 74.2352057, 56.1190544, 37.7199375, 19.0196532, 0., -18.9413868, -38.4048807, -58.5750179, -79.636335, -100.928376, -123.818666, -148.82974, -176.48413, -204.795257, -237.798413, -277.019773, -323.985518, -371.643004, -433.528758, -514.600485, -619.815891, -726.601121, -878.458066, -1091.35706, -1381.26842, -1738.68566, -2256.0096, -2944.21016, -3956.54775, -5195.43966, -7424.38522, -11191.6133, -17045.3526},
        {623.046755, 610.336599, 597.375254, 583.988785, 570.003256, 555.244732, 539.539276, 522.712953, 504.591828, 481.903784, 458.812338, 436.382826, 415.680585, 401.790876, 390.151139, 380.218743, 371.451052, 363.730969, 355.920111, 347.30563, 337.174678, 321.458633, 304.142732, 285.856435, 267.229206, 250.700686, 234.366085, 218.130789, 201.900189, 185.228258, 168.512366, 151.798468, 135.13252, 118.913825, 102.69365, 86.3766111, 69.8673234, 52.8685906, 35.5675654, 17.9495883, 0., -17.9060924, -36.3290217, -55.439354, -75.4076554, -95.5838756, -117.287444, -141.017173, -167.271875, -194.031478, -225.321234, -262.64751, -307.516673, -353.072705, -412.529312, -490.737815, -592.549535, -694.603535, -841.248297, -1048.62004, -1332.855, -1684.25015, -2196.45943, -2885.12536, -3891.89364, -5131.56784, -7300.31693, -10898.9685, -16428.3503},
        {578.777874, 565.343763, 552.617764, 540.222136, 527.779136, 514.911024, 501.240056, 486.388491, 469.978587, 447.525103, 424.400795, 401.870923, 381.200742, 368.162933, 357.712364, 349.311322, 342.422098, 337.556737, 332.707869, 326.917882, 319.229161, 304.187697, 287.130832, 268.899514, 250.334688, 234.780471, 219.573371, 204.553063, 189.559225, 173.868813, 158.109315, 142.345496, 126.64212, 111.492196, 96.360951, 81.1418536, 65.7283731, 49.7810319, 33.5194242, 16.9301977, 0., -16.9275463, -34.3655585, -52.4701786, -71.3975492, -90.5102836, -111.075464, -133.566646, -158.457381, -183.715399, -213.322409, -248.754293, -291.486936, -334.842038, -391.711338, -466.832391, -564.942754, -661.929092, -802.920208, -1004.19402, -1282.02842, -1626.52282, -2132.4907, -2819.95411, -3819.61549, -5057.44347, -7167.66153, -10607.4294, -15833.9069},
        {537.705661, 523.592259, 510.97538, 499.323946, 488.10688, 476.793102, 464.851536, 451.751104, 436.960726, 415.185168, 392.563172, 370.469324, 350.278208, 338.111834, 328.698393, 321.5135, 316.032768, 313.155404, 310.363995, 306.564723, 300.663765, 286.44217, 269.9813, 252.237386, 234.16666, 219.614044, 205.491601, 191.600088, 177.74026, 163.024266, 148.21691, 133.394391, 118.632905, 104.472791, 90.3404513, 76.1264269, 61.7212598, 46.7790037, 31.5212837, 15.9332368, 0., -15.974157, -32.4508828, -49.5726937, -67.4821057, -85.5612143, -105.017124, -126.29652, -149.846085, -173.651974, -201.605614, -235.1379, -275.67973, -316.701541, -370.778872, -442.526802, -536.56041, -628.112094, -762.932689, -957.39035, -1227.85323, -1564.25632, -2062.26726, -2745.81353, -3735.46447, -4967.31563, -7018.71577, -10306.7003, -15248.3048},
        {499.640027, 485.044119, 472.464742, 461.290869, 450.911477, 440.715539, 430.092032, 418.42993, 405.118208, 384.603164, 363.193521, 342.255325, 323.154622, 311.94369, 303.427849, 297.098651, 292.447647, 290.380935, 288.409705, 285.45969, 280.456626, 267.30371, 251.958228, 235.354929, 218.428561, 204.949119, 191.88201, 179.027883, 166.187391, 152.469136, 138.642637, 124.785364, 110.97479, 97.7363464, 84.5203574, 71.2251094, 57.7488886, 43.7851316, 29.5189138, 14.9304613, 0., -15.0143326, -30.5213867, -46.652101, -63.5374142, -80.5902821, -98.9468193, -119.025158, -141.243429, -163.646157, -189.974524, -221.596107, -259.878479, -298.401753, -349.435951, -417.463636, -506.967366, -592.686844, -720.744632, -907.520433, -1169.39395, -1496.20329, -1983.95294, -2659.82073, -3635.19172, -4855.43337, -6845.77638, -9986.48562, -14657.826},
        {464.390881, 449.661375, 437.102489, 426.119557, 416.117917, 406.502905, 396.679857, 386.054108, 374.030995, 355.498276, 336.185896, 317.306223, 300.071624, 289.964612, 282.219351, 276.34015, 271.831319, 269.087297, 266.366212, 262.816322, 257.585886, 245.853978, 232.325711, 217.737017, 202.823827, 190.533415, 178.505829, 166.592464, 154.644712, 141.977941, 129.193979, 116.358629, 103.537694, 91.1635985, 78.7888757, 66.3326796, 53.7141645, 40.722041, 27.4580845, 13.8936269, 0., -14.0164816, -28.5134619, -43.6136023, -59.439564, -75.4511011, -92.6989449, -111.57092, -132.45485, -153.502899, -178.232816, -207.926688, -243.866605, -279.69321, -327.386613, -391.285482, -475.728482, -555.187646, -675.814927, -853.895645, -1105.71512, -1421.11637, -1895.71161, -2559.09281, -3514.54842, -4716.04577, -6641.14011, -9636.48968, -14048.7528},
        {431.768135, 417.406059, 404.905261, 393.806663, 383.651192, 373.979771, 364.333324, 354.252775, 343.279049, 327.589687, 311.434349, 295.699313, 281.270856, 272.480712, 265.391519, 259.511373, 254.348369, 249.128454, 243.754732, 237.848156, 231.029682, 221.17463, 210.347843, 198.868528, 187.055894, 176.114645, 165.124294, 154.049849, 142.856319, 131.325202, 119.678423, 107.954395, 96.1915321, 84.6352843, 73.0342128, 61.3439161, 49.5199927, 37.5123573, 25.2845654, 12.794489, 0., -12.9490123, -26.3635003, -40.3623994, -55.0646445, -69.9972857, -86.1078972, -103.752168, -123.285788, -143.027154, -166.184163, -193.927421, -227.427532, -260.326451, -304.334894, -363.634929, -442.40862, -515.148803, -627.602467, -795.827371, -1035.88127, -1337.74821, -1795.70711, -2440.7469, -3369.28571, -4543.40188, -6397.10368, -9246.41689, -13407.3673},
        {397.382476, 384.161063, 372.406617, 361.789178, 351.978785, 342.645477, 333.459295, 324.090277, 314.208462, 301.130105, 287.820545, 274.891334, 262.954027, 255.327872, 248.833648, 242.999831, 237.354894, 230.264221, 222.884616, 215.20979, 207.233455, 198.951905, 190.355236, 181.436127, 172.187257, 162.452347, 152.432615, 142.180322, 131.747728, 121.278447, 110.696847, 100.018645, 89.2595619, 78.5197505, 67.6967231, 56.7724257, 45.7288044, 34.6183092, 23.3241805, 11.8001626, 0., -11.943738, -24.3276619, -37.2695573, -50.8872099, -64.7777681, -79.7879094, -96.2436743, -114.471103, -132.961668, -154.609806, -180.475383, -211.618268, -241.71635, -282.164266, -336.974677, -410.160241, -476.397825, -580.770198, -739.024339, -966.907224, -1254.16741, -1694.54713, -2321.20376, -3222.44556, -4362.79135, -6146.7668, -8868.89588, -12823.7025},
        {359.974686, 348.352, 338.125236, 328.975588, 320.58425, 312.632416, 304.801281, 296.772038, 288.225882, 276.655174, 264.805473, 253.233508, 242.496004, 235.582587, 229.643928, 224.263593, 219.02515, 212.45994, 205.624648, 198.523735, 191.16166, 183.591405, 175.749496, 167.620982, 159.190914, 150.290229, 141.119734, 131.726121, 122.156087, 112.539351, 102.806371, 92.9706295, 83.0456113, 73.124859, 63.1097732, 52.9818134, 42.7224396, 32.3549175, 21.8021786, 11.0289603, 0., -11.1220613, -22.6495511, -34.6968934, -47.3785121, -60.3411594, -74.354, -89.7185264, -106.736232, -124.027317, -144.247083, -168.369539, -197.368694, -225.064687, -262.446946, -313.351026, -381.612486, -442.034343, -539.097709, -688.251154, -904.943253, -1178.0799, -1602.7377, -2215.93586, -3095.8159, -4202.28901, -5927.06542, -8553.90562, -12366.5701},
        {324.284553, 314.363487, 305.717204, 298.046437, 291.051921, 284.434388, 277.894571, 271.133205, 263.851022, 253.758508, 243.342743, 233.100562, 223.528797, 217.260142, 211.801225, 206.794534, 201.882558, 195.779693, 189.427757, 182.840476, 176.031575, 169.10102, 161.9418, 154.533147, 146.854291, 138.731414, 130.358013, 121.774536, 113.021431, 104.2128, 95.2859787, 86.2519542, 77.1217145, 67.9818174, 58.7374528, 49.3693806, 39.8583605, 30.1989457, 20.3525856, 10.2945228, 0., -10.342512, -21.0562901, -32.2513829, -44.0378392, -56.1058801, -69.1533129, -83.4581171, -99.2982726, -115.426321, -134.255855, -156.675029, -183.572, -208.924082, -243.294606, -290.336061, -353.700939, -408.371479, -498.138522, -638.12266, -843.444484, -1102.14389, -1510.58355, -2109.2871, -2967.15339, -4038.77567, -5707.59445, -8249.83842, -11941.7363},
        {290.624535, 282.418333, 275.336854, 269.105352, 263.449083, 258.093303, 252.763265, 247.184226, 241.08144, 232.406317, 223.367496, 214.39977, 205.937932, 200.246338, 195.198394, 190.497067, 185.845325, 180.150535, 174.229506, 168.103447, 161.793567, 155.43567, 148.890531, 142.133519, 135.140006, 127.739783, 120.11203, 112.290346, 104.308333, 96.2624774, 88.09834, 79.8243672, 71.4490055, 63.0513927, 54.5410077, 45.8980203, 37.1026006, 28.1249184, 18.959144, 9.58944784, 0., -9.60482061, -19.549136, -29.9368594, -40.8719042, -52.0789138, -64.1927792, -77.4691216, -92.1635624, -107.160693, -124.635576, -145.392245, -170.234732, -193.315202, -224.750304, -268.004818, -326.543524, -375.558951, -458.097032, -588.931449, -782.835883, -1027.0032, -1418.94953, -2002.20341, -2837.66181, -3874.62585, -5491.41154, -7959.08117, -11548.697},
        {259.307093, 252.739349, 247.138517, 242.255957, 237.843023, 233.651073, 229.431464, 224.935552, 219.914693, 212.564812, 204.814872, 197.038402, 189.608933, 184.426978, 179.72829, 175.275605, 170.831659, 165.499518, 159.965458, 154.256084, 148.398001, 142.550276, 136.554067, 130.382995, 124.010678, 117.279218, 110.346361, 103.238334, 95.9813658, 88.6520651, 81.2061239, 73.6496167, 65.9886178, 58.2943521, 50.4816833, 42.5306256, 34.4211932, 26.1073599, 17.6055963, 8.90633292, 0., -8.90871776, -18.1293461, -27.7571564, -37.8874201, -48.2672439, -59.4793298, -71.758215, -85.3384368, -99.2324482, -115.385705, -134.521577, -157.363436, -178.258714, -206.857097, -246.432332, -300.258165, -343.746479, -419.177631, -540.970111, -723.542412, -953.301677, -1328.70045, -1895.63073, -2708.54494, -3712.21403, -5281.5743, -7684.02071, -11186.9482},
        {230.644686, 225.549343, 221.276528, 217.601877, 214.301026, 211.14961, 207.923267, 204.397631, 200.348339, 194.200203, 187.620012, 180.923731, 174.427326, 169.687863, 165.283766, 161.034559, 156.759767, 151.753696, 146.571178, 141.241823, 135.795243, 130.399758, 124.890791, 119.24247, 113.428926, 107.313599, 101.025583, 94.5832827, 88.0051028, 81.3452456, 74.5719997, 67.6894509, 60.7016853, 53.6714628, 46.5207255, 39.2300897, 31.7801716, 24.1207947, 16.2756849, 8.23777562, 0., -8.25393396, -16.7981776, -25.7161074, -35.0911002, -44.6778543, -55.0198957, -66.3320723, -78.829232, -91.6435998, -106.505696, -124.063416, -144.964658, -163.775287, -189.658045, -225.69364, -274.962784, -313.083784, -381.584712, -494.531239, -665.989037, -881.683134, -1240.70113, -1790.51499, -2581.00657, -3553.91473, -5081.14038, -7427.04393, -10855.9858},
        {204.949773, 201.071126, 197.905217, 195.246737, 192.890376, 190.630825, 188.262774, 185.580914, 182.379935, 177.278699, 171.718057, 165.963028, 160.278636, 155.914796, 151.757677, 147.678343, 143.547856, 138.840123, 133.982228, 129.004099, 123.935661, 118.939037, 113.859081, 108.672842, 103.357369, 97.8068086, 92.1142731, 86.2899726, 80.3441169, 74.3057015, 68.1586367, 61.9056181, 55.5493414, 49.1434918, 42.6193799, 35.9593056, 29.1455691, 22.1397473, 14.9531524, 7.57637353, 0., -7.6401998, -15.5568876, -23.817546, -32.4896574, -41.3177284, -50.8214078, -61.1973687, -72.6422837, -84.3961618, -97.9950056, -114.018154, -133.044944, -149.885588, -173.196203, -205.863779, -250.775306, -283.720583, -345.522668, -449.907425, -610.600723, -812.791405, -1155.8164, -1687.80212, -2456.25049, -3402.10245, -4893.16741, -7190.53768, -10555.3056},
        {182.534813, 179.527507, 177.178917, 175.294163, 173.678361, 172.136629, 170.474086, 168.49585, 166.007037, 161.766511, 157.044147, 152.063566, 147.048389, 142.99358, 139.042878, 135.111368, 131.114134, 126.685851, 122.134173, 117.486347, 112.76962, 108.123033, 103.417319, 98.6350069, 93.7586259, 88.7227275, 83.5770077, 78.3231858, 72.9629811, 67.4971152, 61.9287041, 56.2598663, 50.4927201, 44.6712063, 38.738892, 32.6811664, 26.4834191, 20.1387421, 13.621741, 6.91472423, 0., -7.06724582, -14.4067335, -22.0653056, -30.0898049, -38.19385, -46.8907973, -56.3607792, -66.783928, -77.4921481, -89.8530911, -104.386181, -121.610841, -136.610285, -157.514631, -187.017785, -227.813654, -255.806599, -311.195892, -407.391261, -557.802432, -747.270316, -1074.91108, -1588.43805, -2335.48049, -3259.15168, -4720.71301, -6976.88882, -10284.4035},
        {165.290507, 162.286376, 160.078255, 158.450848, 157.188853, 156.076972, 154.899907, 153.442357, 151.489025, 147.77082, 143.54775, 139.026033, 134.411887, 130.567225, 126.78029, 122.995023, 119.155364, 115.021951, 110.795345, 106.492805, 102.131591, 97.7874608, 93.395776, 88.9503958, 84.4451802, 79.8803286, 75.2408251, 70.5179937, 65.7031583, 60.7565616, 55.7130412, 50.5763533, 45.3502544, 40.0687128, 34.693188, 29.2153516, 23.6268752, 17.9944365, 12.2046983, 6.21932983, 0., -6.54855977, -13.3849676, -20.5247794, -27.983551, -35.3990376, -43.315716, -51.9002622, -61.3193524, -70.9677412, -82.0927948, -95.1699579, -110.674675, -124.050902, -142.81816800000001, -169.464514, -206.47798, -229.664784, -278.867519, -367.246955, -507.963863, -685.99481, -999.053177, -1492.11906, -2218.17943, -3129.51904, -4573.22466, -6797.68259, -10051.2791},
        {152.137356, 148.560618, 146.028799, 144.288529, 143.086442, 142.169167, 141.283337, 140.175584, 138.592538, 135.133503, 131.151371, 126.851704, 122.440065, 118.735667, 115.084965, 111.448061, 107.78506, 103.948328, 100.048801, 96.0896769, 92.0741556, 87.9923766, 83.8658206, 79.7029095, 75.5120655, 71.371298, 67.1916066, 62.9535786, 58.6378009, 54.159387, 49.5905871, 44.9381778, 40.2089354, 35.4261284, 30.5734448, 25.6510642, 20.6591665, 15.7676707, 10.7391215, 5.50580328, 0., -6.0717388, -12.4611006, -19.1455073, -26.1023811, -32.8645712, -40.0319015, -47.7596232, -56.2029876, -64.7948193, -74.7017665, -86.3680509, -100.237894, -112.156793, -129.007185, -153.07278, -186.637291, -205.292071, -248.690132, -329.792132, -461.558725, -629.404881, -928.92831, -1400.89323, -2107.06889, -3013.74267, -4448.12437, -6647.19318, -9847.92831},
        {140.781753, 136.682234, 133.820464, 131.915018, 130.684473, 129.847403, 129.122385, 128.227993, 126.882804, 123.587991, 119.766493, 115.623843, 111.365579, 107.785616, 104.265758, 100.776187, 97.2870864, 93.7156472, 90.1062419, 86.4502507, 82.7390537, 78.8895359, 74.9973708, 71.0837365, 67.1698115, 63.3987972, 59.6210392, 55.8089063, 51.9347676, 47.8783341, 43.7416956, 39.5342838, 35.2655307, 30.9463936, 26.5841688, 22.1876782, 17.7657435, 13.5945717, 9.30864537, 4.80983222, 0., -5.61379701, -11.5769522, -17.8296726, -24.3121648, -30.4556994, -36.9129944, -43.8278312, -51.3439913, -58.9189753, -67.6573579, -77.9774331, -90.2974945, -100.815114, -115.857596, -137.531524, -167.94348, -182.552124, -220.771131, -295.366252, -419.10324, -577.762732, -865.065819, -1317.76997, -2006.19433, -2910.75903, -4337.91853, -6512.61869, -9659.80539},
        {130.494647, 126.16847, 123.140568, 121.124242, 119.832792, 118.979516, 118.277715, 117.440689, 116.181737, 112.994264, 109.299423, 105.29847, 101.192663, 97.752079, 94.3816266, 91.0550345, 87.7460315, 84.3994753, 81.0295143, 77.6214257, 74.1604867, 70.521654, 66.8446537, 63.1588911, 59.4937719, 56.0305044, 52.5859699, 49.1288527, 45.6278369, 41.946012, 38.1998949, 34.400408, 30.5584734, 26.6696049, 22.7662965, 18.8656338, 14.9847023, 11.5026577, 7.92968738, 4.13804898, 0., -5.16541025, -10.7114327, -16.5445265, -22.5711508, -28.1368435, -33.9333535, -40.0915086, -46.7421363, -53.3523053, -60.9821059, -70.0278696, -80.8859276, -90.0081075, -103.313046, -122.774876, -150.36773, -161.586559, -195.476351, -264.602911, -381.532045, -532.136331, -809.061255, -1246.06057, -1920.26372, -2824.73561, -4245.42361, -6395.05448, -9486.35496},
        {120.218374, 116.363283, 113.606148, 111.699338, 110.395219, 109.44616, 108.604528, 107.622691, 106.253017, 103.165789, 99.6282923, 95.8257276, 91.9432956, 88.7062903, 85.5437823, 82.4249355, 79.3189135, 76.1411207, 72.935984, 69.6941707, 66.4063484, 62.9681364, 59.5032696, 56.0404345, 52.6083179, 49.374891, 46.1738422, 42.9781445, 39.7607707, 36.4023776, 33.0051808, 29.5790797, 26.1339737, 22.6437028, 19.1686496, 15.7331373, 12.361489, 9.52475412, 6.62183918, 3.4983769, 0., -4.71442103, -9.83778635, -15.2497584, -20.8299995, -25.8728603, -31.0774398, -36.5575249, -42.4269028, -48.1422156, -54.7372534, -62.5886611, -72.0730838, -79.7205999, -91.2930477, -108.705699, -133.873825, -142.639015, -173.419696, -238.560614, -350.406513, -494.295262, -763.592234, -1191.33966, -1857.42283, -2763.52395, -4177.2945, -6299.89362, -9332.48046},
        {110.75343, 107.590506, 105.231651, 103.484465, 102.156549, 101.055506, 99.9889362, 98.7644409, 97.1896215, 94.1790465, 90.790563, 87.1889854, 83.539128, 80.5146295, 77.5679503, 74.660375, 71.7531881, 68.7101427, 65.6290678, 62.5102607, 59.3540188, 56.1009217, 52.8348715, 49.5800526, 46.3606494, 43.309762, 40.2990927, 37.3092595, 34.3208804, 31.2458379, 28.1609798, 25.0744183, 21.9942657, 18.872276, 15.7954629, 12.7944818, 9.89998836, 7.66367431, 5.38674428, 2.89143919, 0., -4.26469321, -8.96129429, -13.9478179, -19.0822789, -23.6261418, -28.2725914, -33.1182624, -38.2597896, -43.1364487, -48.7651769, -55.5055526, -63.7171541, -69.9228172, -79.8535597, -95.4036569, -118.467384, -125.376221, -153.812357, -215.895186, -323.744101, -462.041787, -725.217765, -1146.38373, -1806.37838, -2714.14396, -4118.48155, -6208.22748, -9172.21812},
        {102.109544, 99.7569725, 97.8623984, 96.2978131, 94.9352071, 93.6465714, 92.3038964, 90.7791731, 88.9443922, 85.9985892, 82.7558921, 79.3574736, 75.9445064, 73.1306553, 70.3956043, 67.6915293, 64.9706062, 62.0343038, 59.0457882, 56.0175184, 52.9619535, 49.8802388, 46.8006721, 43.7402377, 40.7159199, 37.8110855, 34.9497834, 32.1224447, 29.3195008, 26.4924822, 23.6862811, 20.9068892, 18.1602981, 15.3777876, 12.6699457, 10.0726484, 7.62177209, 5.93500366, 4.23368424, 2.32096572, 0., -3.81327259, -8.07360397, -12.6229576, -17.3032971, -21.3601287, -25.4704559, -29.714825, -34.1737824, -38.2677164, -43.0013945, -48.7194263, -55.7664212, -60.5911985, -68.9924741, -82.8731736, -104.136223, -109.710419, -136.462466, -196.284943, -301.070425, -434.866515, -693.100714, -1109.41568, -1764.10962, -2672.48073, -4063.17178, -6111.48937, -8992.74014},
        {94.2964493, 92.7695158, 91.3437153, 89.9575739, 88.5496178, 87.0583731, 85.422366, 83.5801226, 81.4701691, 78.5889691, 75.4939364, 72.300422, 69.1237773, 66.5079261, 63.9682178, 61.4485745, 58.8929184, 56.0413661, 53.1231674, 50.1637667, 47.1886082, 44.2663165, 41.3618836, 38.4834818, 35.6392833, 32.8548298, 30.1139763, 27.4179471, 24.7679666, 22.1583995, 19.6000738, 17.0969577, 14.6530196, 12.1827008, 9.81530731, 7.59061801, 5.54841194, 4.35432756, 3.17194063, 1.79068637, 0., -3.35720491, -7.16636284, -11.2594299, -15.4683622, -19.038262, -22.6226808, -26.2883164, -30.1018666, -33.4687308, -37.3814243, -42.1711641, -48.1691674, -51.7021831, -58.7076835, -71.1186729, -90.8681555, -95.5538487, -121.178158, -179.408204, -281.911104, -412.260057, -666.403945, -1078.65842, -1727.59583, -2634.41935, -4005.5522, -6001.11258, -8781.21869},
        {87.7713906, 86.4960091, 85.1811683, 83.7968101, 82.3128763, 80.6993088, 78.9260494, 76.9630399, 74.7802222, 72.0511526, 69.1607128, 66.1973987, 63.2497064, 60.7494344, 58.3044555, 55.8659448, 53.3850774, 50.6152646, 47.784551, 44.9232173, 42.0615442, 39.301138, 36.5724235, 33.8771513, 31.2170719, 28.5794753, 25.9863566, 23.4452507, 20.9636922, 18.5573668, 16.2223978, 13.9630598, 11.783627, 9.60000198, 7.54017972, 5.64378337, 3.95043606, 3.11669301, 2.31847245, 1.34862468, 0., -2.86170759, -6.14562894, -9.68805125, -13.3252617, -16.3257928, -19.3207882, -22.3736369, -25.547728, -28.2713039, -31.4969588, -35.5421401, -40.7242953, -43.2310953, -49.1616752, -60.4853932, -79.1716077, -83.2430703, -108.194389, -165.573564, -266.928597, -396.429249, -647.75824, -1054.36803, -1693.5088, -2596.01584, -3934.58632, -5846.45193, -8468.84437},
        {82.0309902, 80.8879567, 79.6096357, 78.1899457, 76.622805, 74.9021323, 73.0218458, 70.9758641, 68.7581056, 66.1932829, 63.5122026, 60.7764656, 58.0476726, 55.637402, 53.257286, 50.8689344, 48.4339568, 45.7340129, 42.9826428, 40.2134361, 37.4599828, 34.8511951, 32.2872114, 29.7634926, 27.2754996, 24.774658, 22.3180779, 19.918834, 17.5900013, 15.3641527, 13.2270657, 11.1840158, 9.24027873, 7.31904842, 5.54051468, 3.94278571, 2.56396973, 2.03191979, 1.55910133, 0.947724641, 0., -2.3603199, -5.10118416, -8.06899945, -11.1101725, -13.5280581, -15.9293355, -18.3776321, -20.9365754, -23.0678086, -25.6777369, -29.0707811, -33.5513625, -35.1556526, -40.1636214, -50.5869898, -68.4374784, -72.203814, -96.8299089, -153.736681, -254.34505, -383.750348, -632.350248, -1032.43624, -1659.18033, -2552.95931, -3850.45028, -5664.25527, -8106.97631},
        {76.9707195, 75.8621394, 74.5611871, 73.0790334, 71.4268495, 69.6158063, 67.6570748, 65.5618261, 63.3412311, 60.9450827, 58.4704813, 55.9531493, 53.4288089, 51.087234, 48.7484747, 46.3866325, 43.9758091, 41.3365635, 38.6579569, 35.975508, 33.3247354, 30.8558433, 28.4437906, 26.0782215, 23.7487807, 21.3748538, 19.0444472, 16.775309, 14.585187, 12.5188412, 10.5562024, 8.70421353, 6.96981753, 5.28812718, 3.76664768, 2.44105406, 1.34702134, 1.06977331, 0.875616728, 0.580407118, 0., -1.85821402, -4.04652853, -6.42570206, -8.85649313, -10.6846322, -12.4919171, -14.3451174, -16.311003, -17.8933143, -19.9470619, -22.7642269, -26.6367909, -27.4525786, -31.6693904, -41.3408701, -58.5206616, -62.2399847, -86.7838771, -143.414952, -243.395824, -373.020461, -618.457409, -1010.56686, -1621.66818, -2501.0912, -3748.66476, -5451.59296, -7697.07988},
        {72.4860498, 71.3353381, 69.967892, 68.4061262, 66.6724552, 64.7892935, 62.7790558, 60.6641567, 58.4670106, 56.2362749, 53.9576244, 51.6429764, 49.3042485, 47.0143357, 44.6997869, 42.3481283, 39.9468865, 37.3618686, 34.7510078, 32.1505183, 29.5966139, 27.2544382, 24.9797041, 22.7570538, 20.5711298, 18.3145381, 16.1007718, 13.9512876, 11.8875423, 9.96151604, 8.15193297, 6.46804054, 4.91908621, 3.45552525, 2.16891421, 1.09201745, 0.257599308, 0.200019021, 0.249808109, 0.239092969, 0., -1.36056209, -2.99516214, -4.78158666, -6.59762218, -7.83508972, -9.05212751, -10.3209083, -11.7136048, -12.7828905, -14.3282371, -16.629617, -19.9670027, -20.0985972, -23.6348501, -32.6644414, -49.2760512, -53.1554874, -77.7554513, -134.125772, -233.316278, -363.036691, -604.357163, -986.463671, -1578.03011, -2436.25295, -3624.75043, -5205.53533, -7240.62045},
        {68.4724527, 67.2243335, 65.7618201, 64.1132768, 62.3070676, 60.3715567, 58.3351084, 56.2260867, 54.0728559, 51.9965824, 49.8957072, 47.7614734, 45.5851245, 43.3341123, 41.0329879, 38.682511, 36.2834413, 33.7488806, 31.2023098, 28.6795519, 26.21643, 23.9863352, 21.8324951, 19.7357053, 17.6767614, 15.5281868, 13.4223587, 11.3833816, 9.43536035, 7.63226104, 5.95638238, 4.41988452, 3.0349276, 1.76952966, 0.697649771, -0.150895095, -0.746287965, -0.607577646, -0.336535071, -0.0837969507, 0., -0.872536286, -1.96058503, -3.16008079, -4.36695815, -5.01900493, -5.65356114, -6.34982007, -7.18697498, -7.77160669, -8.84456592, -10.6740909, -13.5284199, -13.0704323, -16.0158687, -24.475111, -40.558541, -44.7542269, -69.4437895, -125.386536, -223.341773, -352.596143, -588.326948, -957.830483, -1525.32388, -2354.28602, -3474.22797, -4923.15275, -6739.06338},
        {68.2547617, 64.4030038, 61.0624053, 58.1628557, 55.6342446, 53.4064616, 51.4093962, 49.572938, 47.8269765, 45.957039, 44.0951222, 42.2288605, 40.3458883, 38.4667234, 36.5329635, 34.5190896, 32.3995828, 29.9385344, 27.4049711, 24.8575296, 22.354847, 20.1021213, 17.9528037, 15.9069066, 13.9644425, 12.1438868, 10.4194037, 8.78362046, 7.22916423, 5.71567078, 4.28195543, 2.93384198, 1.67715426, 0.346381608, -0.812783878, -1.72598459, -2.3188629, -1.86839962, -1.20836336, -0.523861146, 0., -0.627400983, -1.46345147, -2.37105288, -3.21310661, -3.25085515, -3.18952241, -3.13267337, -3.18387301, -2.80547233, -2.9987359, -4.1237143, -6.54045811, -6.83371701, -10.6489628, -19.8563666, -36.3260991, -39.4859101, -62.6253602, -116.591589, -212.231736, -342.977808, -571.922343, -920.991834, -1451.388, -2230.85247, -3261.692, -4562.57208, -6152.15819},
        {68.4539243, 61.4483189, 55.8628302, 51.4992593, 48.1594069, 45.6450741, 43.7580616, 42.3001706, 41.0732018, 39.3851077, 37.7290771, 36.1044501, 34.5105672, 33.1173068, 31.6852556, 30.1455386, 28.4292807, 26.0790926, 23.5700193, 20.9885914, 18.4213396, 16.0987041, 13.9057423, 11.871421, 10.0247072, 8.59280239, 7.32714494, 6.17740787, 5.09326418, 3.90357687, 2.727153, 1.56198962, 0.406083746, -1.08429602, -2.43273282, -3.50453824, -4.16502385, -3.33725149, -2.20568241, -1.01252808, 0., -0.497549335, -1.22525202, -1.99042368, -2.60037993, -2.04190726, -1.27106206, -0.423371592, 0.365636893, 1.81885925, 2.59897586, 2.22709019, 0.224305732, -1.26648705, -6.39468643, -16.6879037, -33.6737502, -35.6379855, -56.6468132, -107.524585, -199.095654, -329.609444, -547.615087, -868.664624, -1349.75899, -2057.6127, -2978.63703, -4119.31905, -5486.14585},
        {65.4719479, 57.4208681, 51.1223649, 46.3288876, 42.792886, 40.2668095, 38.5031077, 37.2542302, 36.2726266, 34.6627386, 33.0842266, 31.5487434, 30.0679416, 28.9024083, 27.7152878, 26.4186592, 24.9246012, 22.6910859, 20.2659417, 17.7428902, 15.2156527, 12.9020086, 10.7219986, 8.71972156, 6.93927587, 5.70709501, 4.6720087, 3.76518156, 2.9177782, 1.89110459, 0.854127432, -0.194045212, -1.25430528, -2.72905117, -4.05706578, -5.07863845, -5.63405853, -4.49776766, -3.002242, -1.41410996, 0., -0.232086191, -0.689233061, -1.15585079, -1.41634954, -0.326326355, 1.02947019, 2.49510466, 3.91464163, 6.02483426, 7.41998307, 7.58707717, 6.01310568, 4.25526237, -1.09775015, -11.3870246, -27.9536538, -30.037432, -49.9212697, -97.7867791, -183.815572, -309.038656, -511.089458, -801.715921, -1230.32357, -1857.50504, -2661.44699, -3640.93761, -4794.76507},
        {60.4371914, 52.8944439, 47.0061111, 42.5344545, 39.2417357, 36.8902165, 35.2421584, 34.0598231, 33.1054723, 31.4888045, 29.8856696, 28.3193545, 26.8131459, 25.6686144, 24.5194497, 23.2776252, 21.8551143, 19.7300162, 17.4217281, 15.0157733, 12.5976748, 10.34573, 8.215578, 6.25563199, 4.51430532, 3.33212144, 2.34853948, 1.49512871, 0.703458391, -0.282898639, -1.26917813, -2.24861224, -3.21443316, -4.5363881, -5.68058815, -6.48965946, -6.80622818, -5.41661253, -3.64226975, -1.74834914, 0., 0.144274795, 0.0900209157, 0.0494304766, 0.23469559, 1.80159795, 3.64130425, 5.58857079, 7.47815384, 9.93910759, 11.6941713, 12.260382, 11.1547768, 9.8495151, 5.12446293, -4.2853914, -19.6450596, -22.6973878, -42.0384195, -86.7420329, -165.882106, -281.689148, -463.767142, -721.985092, -1095.87966, -1636.34431, -2319.17767, -3138.9521, -4090.23996},
        {54.4780137, 48.4428385, 43.6791707, 39.9986734, 37.21301, 35.1338437, 33.5728378, 32.3416556, 31.2519604, 29.5621783, 27.8585046, 26.1738973, 24.5413145, 23.2625116, 21.9941304, 20.6616102, 19.1903901, 17.1513851, 14.9663681, 12.7025882, 10.427294, 8.26356395, 6.20048559, 4.282976, 2.55595225, 1.31323844, 0.251281779, -0.684563488, -1.54894311, -2.57958539, -3.57482051, -4.51606121, -5.38472022, -6.45481104, -7.29810532, -7.77897552, -7.76179404, -6.16045051, -4.16999329, -2.03498794, 0., 0.606819966, 1.05792541, 1.54218478, 2.24846652, 4.24757601, 6.49367, 8.82284217, 11.0711862, 13.6883341, 15.6514259, 16.5511399, 15.9781543, 15.634255, 12.1545697, 4.28533388, -9.22721724, -13.6309911, -32.5879525, -73.754209, -144.785868, -247.984626, -407.069825, -631.311504, -949.22515, -1399.94534, -1960.88484, -2624.88686, -3384.79461},
        {48.7227738, 44.6398442, 41.3066454, 38.604258, 36.4137625, 34.6162396, 33.0927698, 31.7244338, 30.3923122, 28.5817329, 26.7278303, 24.8699861, 23.0475818, 21.5306864, 20.0357196, 18.509788, 16.8999986, 14.9106943, 12.8288514, 10.6986821, 8.56439891, 6.48920575, 4.49072666, 2.60557732, 0.870373416, -0.504197236, -1.72521994, -2.82570787, -3.83867419, -4.96010825, -5.99485657, -6.91074186, -7.67558681, -8.43282424, -8.90442271, -8.98796086, -8.58101735, -6.79594598, -4.62964023, -2.2937687, 0., 1.13083567, 2.15989592, 3.23917679, 4.5206743, 6.91731822, 9.51579733, 12.1637342, 14.7087513, 17.3991688, 19.5216327, 20.7634863, 20.8120732, 21.7274658, 19.8751872, 13.9934892, 2.82062372, -2.85138016, -21.1595584, -58.1871699, -120.017474, -208.348795, -342.419193, -531.534524, -793.157935, -1154.12298, -1595.62429, -2110.26622, -2690.65314},
        {44.2998304, 42.0592533, 40.0536373, 38.233922, 36.5510472, 34.9559528, 33.3995785, 31.832864, 30.2067492, 28.2463412, 26.2187454, 24.1652348, 22.1270823, 20.3197254, 18.5406064, 16.7613322, 14.9535099, 12.9634457, 10.9381676, 8.89940251, 6.86887778, 4.85635082, 2.90030642, 1.02725964, -0.73627446, -2.27482884, -3.6864212, -4.98011725, -6.16498272, -7.38561978, -8.46134313, -9.34700391, -9.99745327, -10.418932, -10.4943457, -10.1579897, -9.34415937, -7.38976337, -5.06543819, -2.54443376, 0., 1.69160825, 3.34134796, 5.05717119, 6.94702998, 9.71653493, 12.6369161, 15.5770622, 18.4058619, 21.1982666, 23.5346769, 25.2015565, 25.9853686, 28.2471315, 28.1689322, 24.5074123, 16.0192136, 9.6283068, -7.34292715, -39.4047781, -91.0675361, -163.20536, -271.236932, -424.493521, -630.475917, -904.692043, -1232.45178, -1606.61452, -2020.03966},
        {31.3611692, 33.4138344, 34.6123875, 35.0543707, 34.837326, 34.0587958, 32.8163219, 31.2074467, 29.3297123, 27.404283, 25.3556298, 23.2318459, 21.0810247, 19.0533799, 17.0540359, 15.0902378, 13.1692306, 11.3600238, 9.58339213, 7.82187479, 6.05801097, 4.17746025, 2.29839333, 0.442101287, -1.37012481, -3.08825048, -4.73122545, -6.289256, -7.75254846, -9.29802506, -10.6544898, -11.7374626, -12.4624633, -12.5589911, -12.202995, -11.3844033, -10.0931443, -7.98083331, -5.51103668, -2.80900778, 0., 2.33855583, 4.71445469, 7.18331415, 9.80075179, 13.1193625, 16.4989955, 19.7964776, 22.8686353, 25.1976694, 27.1648828, 28.7769525, 30.0405554, 34.3745852, 37.0086155, 36.5844368, 31.7438393, 24.3660043, 8.56037482, -18.326215, -58.9469312, -111.659095, -192.003407, -312.956989, -466.821193, -649.230038, -863.162643, -1109.76185, -1390.17051},
        {37.107078, 37.9494898, 38.1475584, 37.7748672, 36.9049999, 35.61154, 33.9680713, 32.0481773, 29.9254415, 27.7748251, 25.5279832, 23.2179484, 20.8777534, 18.5825868, 16.3064628, 14.0655515, 11.8760232, 9.83055734, 7.83821112, 5.88455095, 3.9551433, 1.96138975, -0.00731245436, -1.93573091, -3.80863323, -5.6288923, -7.35592834, -8.96726683, -10.4404333, -11.8918334, -13.1045604, -14.0005876, -14.5018883, -14.3327386, -13.6918882, -12.5803894, -10.9992944, -8.70169555, -6.03578926, -3.10181195, 0., 2.87088123, 5.82853593, 8.89213931, 12.0808666, 15.681342, 19.3383122, 22.9639726, 26.4705188, 29.5587418, 32.4368037, 35.1014621, 37.5494744, 42.4526156, 46.0626187, 47.3062344, 45.1102132, 39.9921181, 28.651562, 9.37897069, -19.5352305, -57.2665312, -113.12676, -196.55736, -298.235907, -413.117734, -537.811019, -668.510728, -801.411827},
        {54.2496337, 50.3266662, 46.8385324, 43.7179324, 40.8975662, 38.3101336, 35.8883348, 33.5648695, 31.2724379, 28.7494751, 26.2006517, 23.6363737, 21.0670471, 18.525831, 15.9912767, 13.4646886, 10.9473713, 8.38916945, 5.86343143, 3.39204573, 0.99690084, -1.26416265, -3.41958917, -5.46187108, -7.3835007, -9.22601578, -10.9132451, -12.4180629, -13.7133433, -14.8183369, -15.6409909, -16.1356292, -16.2565754, -15.8913591, -15.0878157, -13.8269866, -12.089913, -9.575889, -6.66040207, -3.43719234, 0., 3.32808289, 6.77440878, 10.3369783, 14.0137919, 17.7915291, 21.6840401, 25.6938539, 29.8234995, 34.4683192, 39.0809036, 43.5066565, 47.5909816, 52.0892308, 55.5728806, 57.5233557, 57.4220807, 57.0603078, 52.6857029, 42.8557596, 26.1279713, 0.306807571, -33.7911656, -77.5102898, -128.97305, -194.828091, -250.841278, -280.499173, -267.288336},
        {45.9632263, 43.5948047, 41.2944117, 39.0497927, 36.8486933, 34.6788586, 32.5280343, 30.3839656, 28.2343979, 26.0514214, 23.8446989, 21.6082379, 19.3360459, 16.9792869, 14.5919492, 12.1851779, 9.7701179, 7.3618403, 4.9659933, 2.59215129, 0.249888704, -2.11170034, -4.39836902, -6.57635081, -8.61187918, -10.3777734, -11.9710468, -13.3952985, -14.6541278, -15.978402, -17.0535448, -17.792248, -18.1072036, -17.711688, -16.7975747, -15.3573219, -13.3833878, -10.6397615, -7.43875759, -3.86422188, 0., 3.91966746, 8.02148764, 12.2817728, 16.6768352, 21.1786397, 25.7695849, 30.427722, 35.1311022, 39.788373, 44.4747508, 49.1960481, 53.9580778, 59.5390344, 64.8633958, 69.6280218, 73.5297724, 77.145323, 78.9397916, 78.2581119, 74.4452177, 65.9571768, 54.8055203, 38.1237308, 26.1576508, 25.2296282, 44.2436292, 91.4582154, 175.131949},
        {51.3870855, 48.6794894, 46.1831565, 43.8403542, 41.59335, 39.3844114, 37.155806, 34.8498012, 32.4086645, 29.4555478, 26.3794804, 23.2503763, 20.1381492, 17.3421612, 14.6110986, 11.9230958, 9.25628726, 6.47079096, 3.70996456, 0.999149184, -1.63631406, -4.15046897, -6.54683554, -8.80831871, -10.9178234, -12.8574803, -14.6112782, -16.1624317, -17.4941555, -18.7345028, -19.6639141, -20.2076685, -20.2910453, -19.6603376, -18.491405, -16.7811208, -14.5263586, -11.5733769, -8.12991022, -4.25307808, 0., 4.55743405, 9.3827832, 14.4248361, 19.6323815, 24.7770149, 30.0555953, 35.4877886, 41.0932604, 46.9053614, 52.9245989, 59.1651649, 65.6412515, 72.5154609, 79.5942109, 86.8323292, 94.184644, 100.942288, 107.989262, 115.545873, 123.832425, 131.321797, 143.476579, 160.747982, 191.051251, 237.852606, 314.950336, 433.559716, 604.896017},
        {52.5791609, 52.189412, 50.9809292, 49.0755733, 46.5952055, 43.6616866, 40.3968776, 36.9226396, 33.3608334, 30.1281937, 26.9337584, 23.7814391, 20.6751474, 17.664573, 14.689538, 11.7356427, 8.78848729, 5.69032192, 2.6274371, -0.357226835, -3.22072955, -5.77288536, -8.1768974, -10.4487235, -12.6043213, -14.8262476, -16.8972218, -18.766562, -20.3835865, -21.7939441, -22.8120902, -23.348811, -23.3148923, -22.3006067, -20.6654593, -18.4484415, -15.6885449, -12.4879459, -8.79717705, -4.62995592, 0., 5.28008011, 10.9150172, 16.810651, 22.8728212, 28.6611796, 34.566229, 40.6322844, 46.9036604, 53.1753873, 59.8407787, 67.0438633, 74.92867, 83.7004277, 93.417485, 104.199391, 116.165694, 127.53876, 141.094195, 157.710418, 178.265852, 202.213911, 234.708038, 276.259409, 335.688216, 421.713219, 535.1499, 678.789558, 855.423494}
    };
    
    const double a4tab[50][69] = {
        {-4795.25483, -4657.11547, -4533.57386, -4420.99082, -4315.72713, -4214.14359, -4112.601, -4007.46016, -3895.08186, -3749.86323, -3598.9142, -3447.38103, -3300.40998, -3183.58631, -3073.44169, -2966.94678, -2861.07226, -2743.59299, -2624.35375, -2504.00351, -2383.19127, -2264.35938, -2145.64608, -2026.98299, -1908.30176, -1789.65691, -1670.80798, -1551.63744, -1432.02772, -1310.85768, -1189.41483, -1067.98309, -946.846341, -827.165104, -707.996037, -589.272406, -470.927472, -353.153181, -235.520643, -117.859652, 0., 117.602012, 235.993705, 355.595891, 476.829381, 595.994952, 719.281464, 848.757742, 986.492612, 1125.29678, 1280.20043, 1456.97565, 1661.39451, 1865.75367, 2122.69078, 2451.36808, 2870.9478, 3306.89371, 3909.54591, 4735.54605, 5841.53576, 7186.8075, 9120.05042, 11591.8611, 15444.2924, 19758.4141, 29767.7712, 49167.7902, 81653.8973},
        {-4386.2486, -4276.4998, -4166.87176, -4057.27842, -3947.63375, -3837.85169, -3727.8462, -3617.53124, -3506.82077, -3396.24985, -3284.86287, -3172.32536, -3058.30283, -2934.12142, -2811.12175, -2692.30507, -2580.67265, -2501.29049, -2426.26919, -2349.78411, -2266.0106, -2143.26267, -2011.92156, -1876.50719, -1741.53944, -1625.69861, -1513.68005, -1404.33951, -1296.53273, -1186.17111, -1076.23248, -966.750311, -857.758075, -749.581249, -641.844515, -534.46455, -427.358033, -320.645434, -213.958122, -107.131257, 0., 107.056887, 214.965285, 324.107473, 434.865731, 543.216635, 655.710448, 774.491733, 901.705051, 1031.01986, 1176.44587, 1343.51767, 1537.76987, 1733.97589, 1980.73599, 2295.88922, 2697.27467, 3113.77504, 3687.76832, 4472.67611, 5521.92004, 6796.91537, 8627.10278, 10961.4315, 14608.3709, 18698.1916, 28214.4395, 46677.342, 77607.1262},
        {-4089.70007, -3999.83272, -3900.40845, -3793.52861, -3681.29454, -3565.8076, -3449.16915, -3333.48054, -3220.84311, -3124.77905, -3031.40055, -2938.24064, -2842.83236, -2728.08787, -2612.00939, -2497.97828, -2389.37592, -2309.8317, -2234.37973, -2158.30216, -2076.88115, -1964.99854, -1846.4969, -1724.8185, -1603.40559, -1496.11382, -1391.80673, -1289.76124, -1189.25426, -1088.05749, -987.55515, -887.626246, -788.149784, -688.774549, -589.701849, -490.902776, -392.348421, -294.336138, -196.38025, -98.321342, 0., 97.9874303, 196.858427, 297.074709, 399.097994, 499.259521, 603.803679, 714.844378, 834.49553, 956.8933, 1095.32044, 1255.08195, 1441.48285, 1629.7946, 1867.36916, 2171.52496, 2559.58043, 2963.77126, 3520.53168, 4281.21321, 5297.16736, 6532.52802, 8300.29955, 10550.9045, 14057.1258, 18028.6165, 27105.1878, 44569.1112, 73702.6582},
        {-3869.29607, -3792.59815, -3700.71666, -3596.83571, -3484.13943, -3365.81194, -3245.03736, -3124.99981, -3008.88343, -2913.32449, -2822.67409, -2734.73548, -2647.31191, -2552.51261, -2456.11249, -2358.19241, -2258.83326, -2155.5432, -2052.00489, -1949.32827, -1848.62329, -1756.61105, -1666.5459, -1577.29332, -1487.71881, -1392.40141, -1296.20764, -1199.71758, -1103.51131, -1010.11081, -917.377477, -825.114634, -733.125587, -640.567837, -548.148831, -455.930204, -363.973591, -272.876208, -181.949879, -91.0420086, 0., 90.1866618, 181.269486, 273.857901, 368.561335, 462.482307, 561.139919, 666.546362, 780.71383, 897.948208, 1031.05052, 1185.11547, 1365.23778, 1545.93007, 1775.10198, 2070.08108, 2448.19491, 2844.61617, 3391.68922, 4139.60356, 5138.54867, 6355.55344, 8090.28923, 10297.6099, 13712.8828, 17642.831, 26328.2418, 42776.3866, 69994.5366},
        {-3688.72344, -3620.28003, -3534.3291, -3434.29407, -3323.59833, -3205.66528, -3083.91832, -2961.78084, -2842.67626, -2739.75984, -2642.83036, -2551.41848, -2465.05485, -2394.42261, -2323.43893, -2247.17347, -2160.69588, -2024.75159, -1882.46418, -1742.633, -1614.05739, -1545.14423, -1489.24234, -1439.30806, -1388.29774, -1306.06025, -1217.90239, -1126.02348, -1032.62284, -945.925062, -859.694111, -773.719216, -687.789602, -600.783951, -513.766253, -426.889953, -340.308496, -254.916558, -169.829861, -84.9053578, 0., 83.447602, 167.794816, 253.817352, 342.290922, 431.243692, 525.297932, 626.328371, 736.209734, 849.215705, 977.862472, 1127.06518, 1301.73898, 1475.10257, 1696.44611, 1983.36334, 2353.44796, 2744.04363, 3285.09418, 4026.29337, 5017.33496, 6227.8996, 7947.72035, 10138.8771, 13497.9681, 17433.977, 25771.8272, 41232.4571, 66536.8049},
        {-3511.669, -3448.36228, -3367.77851, -3272.99804, -3167.10117, -3053.16822, -2934.27953, -2813.5154, -2693.95617, -2581.95877, -2476.01625, -2377.89827, -2289.37451, -2240.84482, -2193.9966, -2139.14746, -2066.61499, -1903.78345, -1729.07712, -1557.98694, -1406.00384, -1357.64213, -1331.76003, -1316.23912, -1298.96098, -1228.58923, -1147.91061, -1060.49389, -969.90785, -889.094259, -808.499694, -727.943732, -647.245948, -565.245724, -483.134905, -401.125143, -319.42809, -239.108105, -159.183049, -79.523491, 0., 77.5632713, 156.030772, 236.313367, 319.321921, 403.902372, 493.856483, 590.921089, 696.833024, 805.726906, 929.982676, 1074.37806, 1243.69077, 1410.03238, 1623.91323, 1903.17752, 2265.66943, 2649.78746, 3184.59978, 3919.72886, 4904.79718, 6111.47446, 7823.24142, 10012.0358, 13334.7074, 17295.1967, 25324.1697, 39870.6114, 63383.5064},
        {-3301.81958, -3242.32883, -3167.59761, -3080.04195, -2982.07786, -2876.12137, -2764.5885, -2649.89526, -2534.45768, -2417.79493, -2306.37861, -2203.78347, -2113.58426, -2078.80618, -2047.79337, -2008.34038, -1948.2418, -1778.96537, -1595.16323, -1415.16067, -1257.28299, -1221.1488, -1211.27277, -1213.4629, -1213.52717, -1151.48722, -1077.25191, -994.943767, -908.685307, -833.212394, -757.788871, -682.291925, -606.598741, -529.775993, -452.835581, -375.978895, -299.407326, -224.101762, -149.172293, -74.5085093, 0., 72.3266902, 145.573709, 220.706247, 298.689499, 378.817046, 464.394334, 557.055201, 658.433482, 762.51293, 881.637499, 1020.50106, 1183.79748, 1343.43977, 1550.015, 1821.32939, 2175.18916, 2549.58149, 3074.05924, 3798.35623, 4772.2063, 5968.186, 7667.50096, 9854.41554, 13145.4268, 17119.632, 24873.4951, 38624.1385, 60588.6845},
        {-3022.862, -2967.66361, -2900.31911, -2822.52015, -2735.95834, -2642.32532, -2543.31273, -2440.61219, -2335.91534, -2225.14199, -2118.06432, -2018.68268, -1930.99743, -1895.33367, -1864.83711, -1828.97824, -1777.2275, -1636.62394, -1484.04202, -1333.92478, -1200.71524, -1162.70827, -1144.95436, -1136.35581, -1125.81492, -1066.2531, -996.945919, -921.188061, -842.274195, -771.873467, -701.556286, -631.267537, -560.952101, -490.197592, -419.449071, -348.794329, -278.321158, -208.548445, -138.960447, -69.472514, 0., 67.5308789, 136.01998, 206.356297, 279.428824, 354.34641, 434.49025, 521.461393, 616.860889, 714.604893, 827.05331, 958.88115, 1114.76342, 1268.04501, 1467.2631, 1729.62474, 2072.337, 2431.15956, 2937.32579, 3640.6217, 4590.83328, 5759.94219, 7431.14749, 9603.34594, 12852.4522, 16800.4251, 24308.029, 37426.3271, 58206.3824},
        {-2638.48309, -2589.85055, -2532.47575, -2467.52697, -2396.17251, -2319.58066, -2238.91972, -2155.35797, -2070.06371, -1981.87362, -1895.22025, -1812.20453, -1734.92739, -1677.45422, -1625.13573, -1575.28704, -1525.22331, -1463.08574, -1399.03301, -1334.04983, -1269.12095, -1209.3646, -1149.97859, -1090.29423, -1029.64285, -964.385738, -898.012251, -831.041724, -763.993491, -698.671476, -633.796583, -569.374308, -505.410146, -442.333357, -379.556167, -316.914565, -254.24454, -191.09907, -127.710361, -64.0276063, 0., 62.9688578, 126.965942, 192.62382, 260.575062, 328.849163, 401.722993, 480.870349, 567.965028, 657.033913, 760.456481, 882.965295, 1029.29292, 1176.56839, 1368.1692, 1619.86934, 1947.44278, 2282.2555, 2758.25267, 3424.97148, 4331.9491, 5448.65099, 7064.82952, 9196.15657, 12378.1095, 16230.7179, 23515.9972, 36210.4661, 56290.6436},
        {-2405.71268, -2357.37371, -2305.09333, -2248.92728, -2188.93131, -2125.16116, -2057.67258, -1986.52131, -1911.7631, -1825.78366, -1739.37678, -1655.66624, -1577.77578, -1521.53728, -1472.28315, -1428.05392, -1386.8901, -1345.84384, -1304.33938, -1260.8126, -1213.69937, -1153.95915, -1090.49479, -1024.73273, -958.099404, -895.75918, -833.905408, -772.469359, -711.382305, -650.167044, -589.326712, -528.955972, -469.149485, -410.875751, -353.00606, -295.28554, -237.459321, -178.716175, -119.580127, -60.018845, 0., 59.1806414, 119.414746, 181.265883, 245.297621, 309.449323, 377.958449, 452.438249, 534.501976, 618.058134, 715.506621, 831.542589, 970.861189, 1111.62937, 1295.68177, 1538.3248, 1854.86492, 2173.99376, 2632.27845, 3279.67134, 4166.12475, 5260.24742, 6856.02255, 8984.53783, 12142.789, 15991.1783, 23070.5741, 35201.7283, 54205.3928},
        {-2204.0302, -2154.47867, -2105.90828, -2057.23597, -2007.37868, -1955.25335, -1899.77692, -1839.86634, -1774.43854, -1688.83889, -1600.98454, -1515.22105, -1435.89401, -1383.22182, -1339.32807, -1302.20921, -1269.86166, -1242.18385, -1214.50944, -1184.07407, -1148.11338, -1090.76646, -1027.60414, -961.10068, -893.730354, -835.115058, -777.722379, -721.167533, -665.065735, -607.620119, -550.422817, -493.653877, -437.493349, -383.360258, -329.700085, -276.19729, -222.536329, -167.664731, -112.298656, -56.4173358, 0., 55.7448171, 112.55891, 170.95477, 231.444888, 291.989405, 356.674102, 427.03241, 504.59776, 583.170483, 675.11035, 785.044031, 917.598198, 1051.62411, 1227.83402, 1461.16475, 1766.55314, 2070.16998, 2511.22455, 3140.16011, 4007.4199, 5080.84751, 6658.68514, 8786.8724, 11925.8417, 15777.5862, 22663.3632, 34249.6139, 52202.7796},
        {-2019.11349, -1967.86498, -1922.03119, -1879.54631, -1838.34453, -1796.36004, -1751.52704, -1701.77973, -1645.05228, -1560.9017, -1472.99026, -1386.60303, -1307.02508, -1257.73185, -1218.54189, -1187.46414, -1162.50754, -1146.0165, -1129.93027, -1110.52355, -1084.07109, -1029.27314, -967.008674, -900.58221, -833.298258, -778.328451, -725.163331, -673.160561, -621.677803, -567.845774, -514.139863, -460.808512, -408.100162, -357.77398, -307.963395, -258.312561, -208.46563, -157.218748, -105.403279, -53.0025778, 0., 52.510405, 106.100252, 161.230464, 218.361962, 275.485215, 336.519778, 402.914754, 476.119245, 549.893005, 636.450223, 740.315742, 866.014403, 993.013679, 1160.91873, 1384.27734, 1677.6373, 1964.76951, 2387.3094, 2996.11552, 3842.0464, 4892.15254, 6448.71667, 8571.55319, 11686.9318, 15534.6254, 22233.9388, 33309.2935, 50285.1113},
        {-1852.906, -1799.92433, -1755.7931, -1717.7438, -1683.00798, -1648.81714, -1612.4028, -1570.99649, -1521.82971, -1440.70191, -1354.84952, -1270.0769, -1192.18842, -1146.35433, -1111.26673, -1084.98361, -1065.56296, -1056.86211, -1048.81999, -1037.17485, -1017.66497, -965.979445, -905.925352, -841.260613, -775.743151, -724.511292, -675.390392, -627.586209, -580.304504, -530.041543, -479.796376, -429.858557, -380.517643, -333.704815, -287.411353, -241.270163, -194.914152, -147.113112, -98.7083065, -49.6778856, 0., 49.3678455, 99.8185964, 151.765844, 205.623179, 259.438588, 316.93751, 379.479777, 448.425223, 517.590152, 598.895338, 696.718027, 815.435464, 935.025994, 1094.02532, 1306.57025, 1586.79759, 1856.40165, 2258.93873, 2845.52263, 3667.26715, 4690.50381, 6220.69328, 8329.94908, 11413.3409, 15245.1024, 21759.8548, 32352.1221, 48416.4288},
        {-1707.35118, -1653.0484, -1609.52501, -1573.71394, -1542.54814, -1512.96053, -1481.88405, -1446.25164, -1402.99624, -1326.96932, -1246.01786, -1165.90739, -1092.40343, -1050.37625, -1018.84473, -995.932496, -979.763174, -974.240986, -969.396721, -961.041762, -944.987495, -897.385648, -841.571127, -781.219183, -720.005065, -672.775514, -627.56569, -583.582244, -540.031828, -493.404962, -446.710883, -400.242696, -354.293506, -310.740658, -267.659322, -224.708907, -181.548824, -137.082706, -92.0280493, -46.3465739, 0., 46.2075789, 93.4937656, 142.233789, 192.802879, 243.351361, 297.36933, 356.121975, 420.874489, 485.626378, 561.814791, 653.61119, 765.18704, 876.888988, 1026.24324, 1226.95119, 1492.71423, 1743.67573, 2124.5183, 2686.36654, 3480.34503, 4472.24266, 5969.19114, 8053.42895, 11092.3504, 14891.8235, 21218.6652, 31349.4551, 46560.7729},
        {-1584.39249, -1529.62886, -1485.55795, -1449.34221, -1418.14411, -1389.1261, -1359.45065, -1326.28022, -1286.77728, -1218.43376, -1145.95085, -1074.35921, -1008.68952, -971.084549, -942.618, -921.475684, -905.843417, -897.67345, -889.878586, -879.138066, -862.131131, -819.992009, -771.162956, -718.541216, -665.024034, -622.233049, -580.851352, -540.286431, -499.94577, -457.133564, -414.20191, -371.399611, -328.975471, -288.469407, -248.322665, -208.267601, -168.036574, -126.862414, -85.1768186, -42.9119574, 0., 42.9200457, 86.9055837, 132.30718, 179.4754, 226.72537, 277.257271, 332.235845, 392.825836, 453.366138, 524.577678, 610.355535, 714.594789, 817.830593, 956.661925, 1144.32784, 1394.06738, 1625.20104, 1982.45386, 2516.63232, 3278.54291, 4233.71041, 5688.78638, 7733.36167, 10711.2418, 14457.5951, 20587.924, 30272.6475, 44682.1844},
        {-1485.97338, -1432.0574, -1386.22294, -1346.5141, -1310.975, -1277.64975, -1244.58246, -1209.81724, -1171.39821, -1113.82504, -1054.10405, -995.697115, -942.066106, -909.766207, -883.928673, -862.778061, -844.53893, -826.679817, -808.483705, -788.47756, -765.188346, -730.298793, -691.917793, -651.310007, -609.740092, -571.99583, -534.409508, -496.836536, -459.132322, -420.424886, -381.587985, -342.767985, -304.111253, -266.47896, -229.016746, -191.585055, -154.044331, -116.18712, -77.9689255, -39.2773507, 0., 39.3956861, 79.8338744, 121.658895, 165.21508, 209.062449, 256.043366, 307.215884, 363.638057, 420.173885, 486.553096, 566.311367, 662.984373, 757.078737, 884.370809, 1057.60788, 1289.53726, 1499.58688, 1831.15114, 2334.30506, 3059.12368, 3971.24838, 5374.05517, 7361.11611, 10257.2967, 13925.2237, 19845.1853, 29093.0545, 42744.7041},
        {-1376.76382, -1326.68783, -1282.07861, -1241.75639, -1204.5414, -1169.25385, -1134.71398, -1099.742, -1063.15815, -1015.4217, -967.058202, -920.232267, -877.1085, -849.472809, -826.019977, -805.066087, -784.927224, -759.752074, -733.691079, -706.727285, -678.843735, -650.016673, -620.238666, -589.495479, -557.772878, -524.522938, -490.478592, -455.839083, -420.803651, -385.914211, -350.890265, -315.793989, -280.687555, -245.903895, -211.126125, -176.308115, -141.403738, -106.570871, -71.4777782, -35.9967311, 0., 36.0990347, 73.185657, 111.604041, 151.698359, 192.269829, 235.822765, 283.318523, 335.718461, 388.430658, 450.191058, 524.182331, 613.587143, 698.997506, 815.223007, 974.482576, 1188.99514, 1378.73642, 1685.06585, 2157.09963, 2843.95399, 3710.28866, 5058.5893, 6988.85511, 9802.94449, 13373.0307, 19088.1684, 27958.192, 40992.9357},
        {-1237.58523, -1193.72216, -1155.05249, -1120.43609, -1088.73286, -1058.80268, -1029.50544, -999.701022, -968.249314, -926.239712, -883.41079, -841.730634, -803.167327, -778.318849, -757.071429, -737.941191, -719.444261, -696.334781, -672.39565, -647.647786, -622.112106, -595.968306, -569.015013, -541.209634, -512.509574, -482.320899, -451.372892, -419.843492, -387.910642, -356.060285, -324.03916, -291.902006, -259.703565, -227.751645, -195.74669, -163.642216, -131.391734, -99.0699117, -66.4606474, -33.4689929, 0., 33.4513572, 67.8059178, 103.394599, 140.548317, 178.214416, 218.660816, 262.771863, 311.431906, 360.448557, 417.813589, 486.442042, 569.248956, 647.275805, 754.06062, 901.267868, 1100.56201, 1272.16848, 1555.76639, 1999.59584, 2651.8969, 3474.58878, 4774.87422, 6664.71547, 9415.52636, 12887.0161, 18426.2608, 27008.4947, 39608.9522},
        {-1105.58134, -1068.28477, -1035.7158, -1006.80631, -980.488224, -955.693435, -931.35385, -906.401371, -879.767902, -843.331603, -805.899619, -769.225353, -735.062207, -712.723629, -693.378956, -675.757571, -658.588856, -637.292406, -615.231308, -592.458861, -569.028363, -545.287037, -520.876688, -495.733044, -469.791833, -442.442738, -414.385952, -385.775624, -356.765902, -327.780257, -298.595783, -269.258899, -239.816024, -210.549023, -181.174687, -151.645255, -121.912965, -91.969331, -61.7116068, -31.0763203, 0., 30.9521313, 62.7237173, 95.6297075, 129.985051, 164.863675, 202.31796, 243.159265, 288.198947, 333.654277, 386.768336, 450.190117, 526.568614, 597.431846, 694.998168, 830.364963, 1014.62961, 1168.42203, 1429.49406, 1845.13006, 2462.6144, 3241.06015, 4492.26561, 6339.07397, 9025.31546, 12395.8862, 17770.4317, 26099.91, 38335.2793},
        {-981.843359, -951.135893, -924.577433, -901.1915, -880.001617, -860.031308, -840.304093, -819.843497, -797.673041, -766.547528, -734.266689, -702.361533, -672.363072, -652.260502, -634.543371, -618.159416, -602.056372, -582.353316, -561.958105, -540.949937, -519.408011, -497.806608, -475.671811, -452.924788, -429.486703, -404.761057, -379.393749, -353.513012, -327.24708, -300.95003, -274.433913, -247.736624, -220.896058, -194.167337, -167.28424, -140.19777, -112.858933, -85.1877245, -57.178561, -28.7948502, 0., 28.5987857, 57.9394136, 88.3159941, 120.022638, 152.2321, 186.808389, 224.494156, 266.032056, 308.046801, 357.046158, 415.419955, 485.558019, 549.519157, 638.156626, 761.992664, 931.549505, 1067.95242, 1306.87941, 1694.61147, 2277.42964, 3011.70205, 4213.44837, 6014.88302, 8636.06863, 11907.0428, 17130.0951, 25239.4523, 37169.3408},
        {-867.462487, -843.035749, -822.146291, -803.916104, -787.467177, -771.921499, -756.401059, -740.027848, -721.923855, -695.737645, -668.254002, -640.784284, -614.639853, -596.502818, -580.165486, -564.790917, -549.542169, -531.245875, -512.336089, -492.910439, -473.066553, -453.360756, -433.248503, -412.643943, -391.461228, -369.148461, -346.272259, -322.93319, -299.231825, -275.445509, -251.427325, -227.207134, -202.814794, -178.477895, -153.949475, -129.180301, -104.121142, -78.6436884, -52.8094149, -26.6007194, 0., 26.3887492, 53.4533645, 81.4600863, 110.675155, 140.334184, 172.146291, 206.789968, 244.943705, 283.625111, 328.637912, 382.124952, 446.229074, 503.591266, 583.656969, 696.36977, 851.673255, 971.215005, 1188.55301, 1548.94926, 2097.66575, 2788.51376, 3941.10738, 5695.09497, 8251.54269, 11427.8877, 16514.6653, 24434.1361, 36108.5609},
        {-763.529923, -744.744569, -728.931269, -715.304576, -703.079038, -691.469208, -679.689636, -666.954874, -652.479471, -630.752109, -607.603558, -584.138715, -561.462481, -545.02393, -529.846115, -515.296264, -500.741608, -483.698449, -466.12531, -448.129792, -429.819492, -411.783222, -393.454883, -374.749589, -355.582454, -335.477555, -314.897458, -293.913693, -272.597788, -251.142599, -229.449798, -207.542384, -185.443357, -163.352002, -141.044517, -118.473386, -95.5910942, -72.2558187, -48.5520731, -24.4700646, 0., 24.3194502, 49.2659279, 75.0686116, 101.95668, 129.18442, 158.345858, 190.060128, 224.946366, 260.38819, 301.534457, 350.298506, 408.59368, 459.701698, 531.62017, 633.715083, 775.352427, 878.665118, 1075.14545, 1409.05263, 1924.64589, 2573.49455, 3677.92749, 5382.66222, 7875.49447, 10965.8228, 15933.5562, 23690.976, 35150.3634},
        {-671.136869, -657.022582, -645.441263, -635.681364, -627.031337, -618.779636, -610.214713, -600.625022, -589.299015, -571.441079, -552.057359, -532.069936, -512.400888, -497.39719, -483.186069, -469.319648, -455.35005, -439.439403, -423.08582, -406.397423, -389.482331, -372.907744, -356.139071, -339.100802, -321.717425, -303.620942, -285.145324, -266.332054, -247.222617, -227.917203, -208.375107, -188.61433, -168.652875, -148.660967, -128.443493, -107.957565, -87.1602927, -65.9427114, -44.3544406, -22.3790227, 0., 22.3883175, 45.3774618, 69.1481977, 93.8812897, 118.797302, 145.421279, 174.318067, 206.052511, 238.33502, 275.726649, 319.934016, 372.66374, 417.903983, 482.167203, 574.247404, 702.938588, 790.758109, 967.287275, 1275.83075, 1759.69318, 2368.6437, 3426.5936, 5080.53715, 7511.68081, 10528.25, 15396.182, 23016.9864, 34292.1725},
        {-591.374525, -580.630018, -572.185168, -565.370918, -559.518208, -553.957981, -548.021179, -541.038742, -532.341613, -517.654709, -501.357407, -484.223055, -467.025006, -453.195949, -439.786162, -426.505259, -413.062857, -398.197102, -382.977668, -367.502757, -351.870573, -336.568062, -321.149189, -305.556662, -289.733187, -273.451226, -256.891831, -240.065808, -222.983963, -205.645227, -188.07703, -170.294927, -152.314472, -134.276096, -116.020528, -97.5133744, -78.7202402, -59.6229627, -40.164422, -20.3037303, 0., 20.5927796, 41.7883242, 63.7054721, 86.463062, 109.187321, 133.386744, 159.577213, 188.274612, 217.464584, 251.205346, 291.024877, 338.451155, 378.251645, 435.419044, 518.185534, 634.783301, 707.949324, 865.609071, 1150.19281, 1604.13079, 2175.96048, 3189.79057, 4791.67214, 7163.85853, 10122.571, 14911.9569, 22419.182, 33531.4121},
        {-531.224895, -520.626258, -512.797608, -506.999762, -502.493538, -498.539755, -494.399231, -489.332783, -482.601229, -469.841541, -455.387921, -439.950727, -424.240315, -411.213282, -398.435247, -385.718072, -372.873616, -359.085644, -345.04535, -330.815832, -316.46019, -302.247325, -287.952211, -273.555623, -259.038339, -244.396642, -229.589599, -214.591784, -199.377772, -183.808107, -168.017004, -152.02465, -135.851231, -119.609428, -103.189935, -86.5759385, -69.7506277, -52.91163, -35.741918, -18.1389039, 0., 18.9692432, 38.6028073, 58.9265352, 79.9662703, 100.621877, 122.495568, 146.063579, 171.802143, 197.878892, 228.002105, 263.571461, 305.986636, 341.038107, 391.978429, 466.450961, 572.099059, 631.266927, 771.016734, 1033.1115, 1459.31424, 1998.35806, 2971.09571, 4515.64123, 6831.24812, 9763.79368, 14511.5205, 21934.5798, 32893.1229},
        {-486.583679, -473.979814, -465.030685, -458.861989, -454.599418, -451.368669, -448.295435, -444.505412, -439.124294, -427.323975, -413.765471, -399.155998, -384.202771, -371.710721, -359.450261, -347.28952, -335.096629, -322.375158, -309.503618, -296.495961, -283.36614, -270.088482, -256.732415, -243.32774, -229.904261, -216.721713, -203.487992, -190.140926, -176.618344, -162.629655, -148.432476, -134.056005, -119.529439, -104.937331, -90.2313794, -75.4186386, -60.5061629, -45.9950208, -31.2006471, -15.9324906, 0., 17.4807597, 35.7296022, 54.6597253, 74.1843266, 92.8924074, 112.55104, 133.603101, 156.491467, 179.486059, 206.071889, 237.561016, 275.265497, 306.105316, 351.541434, 418.642738, 514.478116, 560.661054, 683.898001, 925.440002, 1326.53811, 1837.0297, 2772.40681, 4258.3764, 6521.60732, 9452.61168, 14185.5666, 21543.733, 32350.3718},
        {-448.814905, -434.351947, -424.23195, -417.480755, -413.1242, -410.188124, -407.698365, -404.680764, -400.161159, -388.964091, -375.997217, -361.966893, -347.579478, -335.550049, -323.772757, -312.150471, -300.586061, -288.808322, -276.963829, -265.025081, -252.964578, -240.504183, -227.967286, -215.426643, -202.955006, -191.03281, -179.162057, -167.252429, -155.213609, -142.637141, -129.878101, -116.973426, -103.960054, -90.8894741, -77.7782523, -64.6575056, -51.5583513, -39.2929482, -26.798955, -13.7950721, 0., 16.0603954, 32.9974454, 50.6153159, 68.7181726, 85.5958004, 103.172498, 121.858185, 142.062778, 162.114943, 185.338352, 212.975426, 246.268582, 273.110677, 313.433519, 373.819356, 460.850433, 495.64119, 604.4288, 827.982629, 1207.07205, 1692.46558, 2594.93511, 4028.40848, 6246.18361, 9183.8677, 13908.4607, 21204.1145, 31854.9808},
        {-415.107526, -399.828275, -389.110135, -381.960008, -377.384796, -374.391399, -371.986718, -369.177655, -364.971111, -354.165601, -341.659766, -328.143861, -314.308141, -302.780952, -291.539221, -280.497968, -269.57221, -258.587577, -247.584233, -236.512952, -225.324507, -213.595855, -201.801116, -190.040593, -178.414589, -167.532131, -156.781306, -146.058926, -135.261805, -123.926816, -112.454684, -100.886198, -89.2621454, -77.5906511, -65.9582303, -54.4187358, -43.0260202, -32.8911881, -22.5879395, -11.7472262, 0., 14.6788678, 30.3399124, 46.6897497, 63.4349954, 78.6160763, 94.2722731, 110.776677, 128.50238, 145.787851, 165.854652, 189.889726, 219.080012, 241.990545, 277.478935, 331.780884, 411.132096, 436.62174, 533.690667, 742.633193, 1103.74364, 1567.96956, 2443.64553, 3835.9396, 6019.38928, 8970.19375, 13688.5361, 20918.6033, 31404.5825},
        {-381.404703, -367.810516, -358.064516, -351.312694, -346.701042, -343.375552, -340.482213, -337.167018, -332.575957, -322.140584, -310.207104, -297.407282, -284.372885, -273.56541, -263.055001, -252.741533, -242.524879, -232.132597, -221.705805, -211.213306, -200.6239, -189.585152, -178.515596, -167.512531, -156.673252, -146.561658, -136.621805, -126.76435, -116.899948, -106.625098, -96.2902787, -85.9318111, -75.5860158, -65.1968036, -54.9298686, -44.858495, -35.0559672, -26.8956431, -18.630704, -9.81440464, 0., 13.2974927, 27.6712859, 42.7528299, 58.1735747, 71.8319533, 85.7856399, 100.359291, 115.877565, 130.624358, 147.781391, 168.489624, 193.890017, 212.684812, 243.429178, 292.239562, 365.232414, 384.336086, 473.530361, 672.606928, 1021.35747, 1469.13545, 2327.04725, 3698.40244, 5866.54264, 8835.96394, 13546.5125, 20704.1969, 31015.0258},
        {-350.693994, -339.567532, -331.244199, -325.066325, -320.376238, -316.516268, -312.828743, -308.655992, -303.340345, -293.184706, -281.786597, -269.704117, -257.495364, -247.430419, -237.670605, -228.089227, -218.55959, -208.634708, -198.636295, -188.565772, -178.424563, -168.013895, -157.615462, -147.310764, -137.1813, -127.67297, -118.357114, -109.169472, -100.045784, -90.6878242, -81.3588831, -72.0882854, -62.9053565, -53.6877426, -44.6771193, -35.9634834, -27.6368316, -21.2984096, -14.9224656, -7.99449665, 0., 11.9307478, 25.0149408, 38.8249936, 52.933321, 65.1521334, 77.5181313, 90.3078109, 103.797669, 116.194025, 130.671622, 148.335027, 170.288808, 185.113374, 211.447112, 255.404249, 323.099016, 337.713944, 421.467639, 613.647007, 953.538957, 1388.60832, 2233.60823, 3592.34408, 5751.29326, 8739.15611, 13433.0691, 20498.0598, 30599.1557},
        {-322.965118, -314.722469, -308.053627, -302.529432, -297.72072, -293.198329, -288.533096, -283.29586, -277.057457, -267.132496, -256.250535, -244.884904, -233.50893, -224.173201, -215.142885, -206.260408, -197.368196, -187.807977, -178.123154, -168.356433, -158.550519, -148.713216, -138.936092, -129.275811, -119.789041, -110.754011, -101.917197, -93.246636, -84.7103687, -76.143424, -67.7000552, -59.4015055, -51.2690183, -43.1182337, -35.2582393, -27.7925197, -20.8245593, -16.1400159, -11.4873314, -6.29712101, 0., 10.5704871, 22.3478263, 34.8625739, 47.6452865, 58.4725586, 69.3304936, 80.4512328, 92.0669177, 102.300478, 114.336951, 129.252163, 148.121941, 159.204636, 181.520537, 221.27246, 284.663221, 296.452606, 376.86367, 564.67644, 898.670942, 1324.54377, 2160.32525, 3511.71103, 5663.53469, 8666.0039, 13328.6919, 20271.3306, 30113.6518},
        {-298.207795, -292.898473, -287.897243, -283.010547, -278.044824, -272.806515, -267.102061, -260.737901, -253.520476, -243.818486, -233.451208, -222.800179, -212.246936, -203.590978, -195.228694, -186.974434, -178.64255, -169.366469, -159.913835, -150.371369, -140.825792, -131.514249, -122.312866, -113.248194, -104.346783, -95.6927269, -87.2320163, -78.9681837, -70.9047623, -63.0203273, -55.353353, -47.9173557, -40.7258518, -33.5430427, -26.7314855, -20.4044225, -14.675096, -11.4609901, -8.34940844, -4.73189658, 0., 9.2085648, 19.6468917, 30.8219035, 42.2405231, 51.6891709, 61.0834733, 70.6185542, 80.4895374, 88.7473442, 98.5889822, 111.067257, 127.234972, 134.887001, 153.637255, 189.841712, 249.85635, 260.249361, 339.079625, 524.618236, 855.136285, 1275.09739, 2104.19508, 3450.4498, 5593.16045, 8602.74098, 13213.8667, 19995.1477, 29515.1939},
        {-277.692073, -273.266665, -268.661563, -263.791764, -258.572268, -252.918071, -246.744171, -239.965567, -232.497256, -223.318657, -213.654579, -203.794253, -194.026909, -185.767148, -177.728681, -169.75059, -161.671955, -152.674565, -143.517713, -134.303398, -125.133621, -116.351293, -107.721136, -99.2487811, -90.9398631, -82.7556789, -74.7639327, -66.9879926, -59.4512265, -52.2008259, -45.2268059, -38.543005, -32.163262, -25.856573, -19.9795559, -14.6439865, -9.96164035, -7.81861146, -5.84262963, -3.43574307, 0., 7.74471995, 16.6413325, 26.2149221, 35.990573, 43.8290239, 51.5854432, 59.4506533, 67.6154769, 74.1745522, 82.25336, 92.8811966, 107.087358, 112.157665, 128.36228, 162.227891, 220.281186, 230.228588, 308.945153, 494.485671, 824.904937, 1246.65773, 2072.59887, 3408.55568, 5528.94543, 8536.44835, 13052.1028, 19567.3577, 28573.6617},
        {-259.939683, -255.968472, -251.491441, -246.505072, -241.005847, -234.990248, -228.454756, -221.395854, -213.810024, -205.1907, -196.23863, -187.151517, -178.127059, -170.170813, -162.349484, -154.537632, -146.609818, -137.847271, -128.955213, -120.045537, -111.230135, -102.937193, -94.8357926, -86.9113071, -79.1491114, -71.3967003, -63.8324798, -56.4969763, -49.4307162, -42.7343354, -36.3642068, -30.3368131, -24.6686367, -19.1508851, -14.1154261, -9.66885215, -5.91775591, -4.66091588, -3.63586439, -2.27231968, 0., 6.27642281, 13.5921656, 21.5074915, 29.5826637, 35.7843028, 41.9037718, 48.1387909, 54.6870805, 59.7199925, 66.272163, 75.3518594, 87.9673494, 90.8661607, 105.021597, 137.146221, 193.952597, 204.451318, 284.137707, 470.805114, 802.246893, 1227.94069, 2050.62697, 3373.0398, 5462.06578, 8452.38814, 12843.3511, 19043.1335, 27459.9144},
        {-244.613452, -240.742451, -236.178876, -230.976389, -225.188652, -218.869326, -212.072073, -204.850553, -197.258429, -189.207498, -180.950031, -172.596435, -164.257118, -156.526127, -148.836772, -141.106004, -133.250774, -124.687935, -116.034573, -107.407678, -98.9242389, -91.0767404, -83.4564783, -76.0302437, -68.7648278, -61.4070679, -54.2316906, -47.2934683, -40.6471739, -34.4312784, -28.5833766, -23.1247619, -18.0767276, -13.2663693, -8.98685713, -5.33716324, -2.41625999, -1.89649508, -1.67411532, -1.2187429, 0., 4.81904605, 10.5395963, 16.7694066, 23.1162328, 27.6727934, 32.1678966, 36.8153135, 41.8288147, 45.4844063, 50.7087301, 58.4906628, 69.8190813, 70.9233303, 83.4556316, 114.308675, 170.37515, 182.25048, 263.643527, 451.965888, 784.629157, 1214.99982, 2032.62481, 3336.30428, 5382.70068, 8336.57764, 12572.4318, 18412.3265, 26178.3252},
        {-231.376208, -227.327158, -222.515866, -217.031636, -210.963774, -204.401586, -197.434378, -190.151455, -182.642124, -175.141937, -167.535452, -159.853477, -152.126817, -144.557247, -136.936217, -129.226146, -121.389454, -112.999904, -104.564031, -96.1997155, -88.0248379, -80.5747267, -73.3828348, -66.400063, -59.5773124, -52.5780585, -45.755598, -39.1758021, -32.9045422, -27.1020775, -21.7021359, -16.7328335, -12.2222863, -8.04341573, -4.44161024, -1.50706368, 0.670030087, 0.566059177, 0.0976149603, -0.252129215, 0., 3.38796234, 7.52383021, 12.0704624, 16.6907177, 19.6122814, 22.5072553, 25.6125674, 29.1651461, 31.5685348, 35.6264003, 42.3090248, 52.5866904, 52.2400158, 63.5048119, 93.4272263, 149.053406, 162.959, 246.448854, 436.357316, 769.518734, 1203.88868, 2012.93783, 3290.75124, 5281.02929, 8175.03416, 12224.1654, 17664.7881, 24733.2674},
        {-219.890779, -215.461152, -210.294408, -204.496731, -198.174303, -191.433308, -184.37993, -177.120352, -169.760758, -162.766899, -155.741564, -148.647109, -141.445889, -133.988329, -126.393489, -118.6685, -110.820491, -102.586526, -94.3518247, -86.2315432, -78.3408367, -71.2359444, -64.414504, -57.8152372, -51.3768655, -44.7009486, -38.1982351, -31.9423113, -26.0067637, -20.5571553, -15.5383054, -10.9870099, -6.94006466, -3.32241444, -0.327446578, 1.96330266, 3.46829701, 2.81815514, 1.73432386, 0.650404896, 0., 1.99854432, 4.58507283, 7.48045392, 10.405556, 11.7205528, 13.0512852, 14.6628993, 16.8205414, 18.073119, 21.0885124, 26.8183632, 36.2143128, 34.727059, 45.009565, 74.2138497, 129.491933, 145.909806, 231.539927, 422.368724, 754.382627, 1190.66081, 1985.91147, 3228.78282, 5147.23075, 7953.77503, 11783.3722, 16790.3694, 23129.1142},
        {-226.724899, -212.764992, -200.613432, -190.033165, -180.787138, -172.638298, -165.34959, -158.683961, -152.404358, -145.81283, -139.317579, -132.865911, -126.40513, -119.900384, -113.273999, -106.466146, -99.4169927, -91.5262098, -83.4906648, -75.4667261, -67.6107623, -60.4573783, -53.6334118, -47.1439369, -40.9940276, -35.2433287, -29.8205152, -24.7088327, -19.8915271, -15.2693682, -10.9410679, -6.92286233, -3.23098758, 0.61081133, 3.89681083, 6.41377843, 7.9484816, 6.40498557, 4.20584103, 1.89089637, 0., 1.36847131, 3.32249934, 5.4837441, 7.4738656, 7.10299269, 6.52892902, 6.09794707, 6.15631933, 4.78894311, 5.50801613, 9.56436095, 18.2088001, 19.6826667, 33.4500685, 65.965624, 123.683951, 138.560516, 221.348751, 408.302935, 735.677349, 1179.71831, 1956.70398, 3138.69873, 4942.56472, 7595.34243, 11138.5612, 15660.1283, 21247.9508},
        {-236.84171, -211.295632, -190.740811, -174.495373, -161.877439, -152.205134, -144.796581, -138.969903, -134.043225, -127.670992, -121.500477, -115.515274, -109.698978, -104.443078, -99.1601186, -93.6705368, -87.7947709, -80.3881138, -72.6222059, -64.7035429, -56.8386203, -49.4919717, -42.5088397, -35.9925048, -30.0462478, -25.3821555, -21.2511794, -17.5130774, -14.0276074, -10.327812, -6.73085042, -3.2271669, 0.192794552, 4.54191817, 8.42510028, 11.4505655, 13.2265385, 10.5952775, 7.03736001, 3.26739713, 0., 1.10351971, 2.87733143, 4.77455029, 6.24829142, 4.25773493, 1.74750499, -0.831709239, -3.02921862, -7.34619724, -9.19934742, -6.95723472, 1.01157529, 6.82986723, 25.4411853, 62.2804239, 122.782477, 135.652615, 213.747207, 393.192996, 710.116727, 1156.14026, 1900.90499, 2996.52076, 4644.78844, 7069.78957, 10256.7919, 14250.7667, 19096.6851},
        {-232.280967, -202.805662, -179.493136, -161.488734, -147.937801, -137.985681, -130.77772, -125.459263, -121.175653, -114.86966, -108.770235, -102.903752, -97.2965875, -92.6305548, -88.0144139, -83.2123638, -77.9886036, -70.9785561, -63.5267075, -55.8487674, -48.1604456, -40.8346679, -33.8670415, -27.4103897, -21.6175357, -17.5081853, -14.0215263, -10.963629, -8.14056397, -4.8928611, -1.67834712, 1.51069153, 4.68196843, 9.02829042, 12.8982405, 15.8254948, 17.3437298, 13.841766, 9.25607751, 4.37828269, 0., 0.430831866, 1.53721913, 2.70358662, 3.31435919, -0.0847916974, -4.13518649, -8.31689902, -12.1100031, -18.0667268, -21.3661281, -20.2594189, -12.9978114, -5.65089228, 14.4758508, 52.2585556, 112.57336, 127.359326, 203.604497, 375.359842, 676.676327, 1108.31316, 1804.19659, 2804.49135, 4284.70754, 6449.15256, 9251.59936, 12708.4545, 16836.1246},
        {-217.723575, -189.98152, -168.069873, -151.165799, -138.446465, -129.089038, -122.270683, -117.168567, -112.959856, -106.589909, -100.360423, -94.3412868, -88.602391, -83.9929754, -79.4918381, -74.8571281, -69.8469946, -63.1281994, -55.9868336, -48.6176009, -41.2152054, -34.0498118, -27.2104788, -20.8617257, -15.168072, -11.1893742, -7.83668007, -4.91637443, -2.23484203, 0.921668114, 4.01858121, 7.04145819, 9.97586002, 13.921929, 17.3048121, 19.6642379, 20.5399346, 16.3430655, 10.9933504, 5.28194367, 0., -0.573489477, -0.530341535, -0.4745363, -1.0100539, -5.63864982, -10.9074187, -16.2612305, -21.1449552, -27.7808742, -31.7254815, -31.3126825, -24.8763827, -18.1852808, 0.835428433, 36.825757, 94.4257167, 113.608134, 189.54708, 352.749442, 633.722106, 1037.49814, 1671.00588, 2568.45021, 3871.51835, 5752.88672, 8153.79254, 11073.3328, 14510.6046},
        {-197.850438, -175.509646, -157.670492, -143.67912, -132.881674, -124.624299, -118.253141, -113.114343, -108.55405, -102.012813, -95.5046078, -89.1378169, -83.0208222, -78.0605024, -73.2473444, -68.3703316, -63.2184475, -56.6677061, -49.7852481, -42.7252446, -35.6418668, -28.701748, -22.0416126, -15.8106468, -10.1580371, -5.99367853, -2.40176533, 0.772799518, 3.68511306, 7.00195942, 10.1620738, 13.1158783, 15.8137953, 19.0748347, 21.6333959, 23.0924661, 23.0550322, 18.2977904, 12.3805353, 6.03677068, 0., -1.83334156, -3.15785451, -4.50520788, -6.40707065, -12.1179021, -18.3574648, -24.5683118, -30.1929959, -36.8968677, -41.0105633, -41.0875167, -35.6811617, -31.1989674, -15.1987182, 16.907766, 69.7086655, 94.3265236, 170.201412, 323.307766, 579.620021, 944.956329, 1505.75997, 2294.23711, 3414.41715, 5000.44737, 6994.1804, 9385.54253, 12164.4601},
        {-177.34246, -162.076478, -149.494462, -139.181246, -130.721668, -123.700563, -117.702768, -112.313118, -107.11645, -100.319447, -93.4363589, -86.6032825, -79.9563146, -74.3632984, -68.935886, -63.5174757, -57.9514658, -51.4277386, -44.704615, -37.8868996, -31.0793971, -24.3548212, -17.8629043, -11.7212873, -6.04761157, -1.48905451, 2.57809353, 6.24800591, 9.61485598, 13.2341966, 16.5542697, 19.4846971, 21.9351002, 24.3390085, 25.8725726, 26.2358508, 25.1289017, 19.9045551, 13.5489893, 6.70115434, 0., -3.27262161, -6.17782377, -9.13381751, -12.5588139, -19.2366111, -26.2735981, -33.1417511, -39.3130462, -45.8229357, -49.9545293, -50.5544126, -46.4691713, -45.1176209, -33.3452256, -6.56967931, 39.791324, 69.4419812, 144.193951, 284.980784, 512.73603, 831.948853, 1312.88596, 1987.69175, 2922.60026, 4211.28985, 5803.57192, 7685.22462, 9842.0261},
        {-160.880546, -152.368455, -144.741251, -137.82473, -131.444689, -125.426926, -119.597237, -113.78142, -107.805271, -100.690886, -93.3892444, -86.0476237, -78.8133015, -72.4315259, -66.2124162, -60.0640621, -53.8945533, -47.2389595, -40.5275983, -33.8177671, -27.1667634, -20.5733762, -14.1768151, -8.05778129, -2.29697582, 2.75654163, 7.39777206, 11.6533579, 15.5499414, 19.5045633, 22.9973084, 25.8986598, 28.0791008, 29.5664513, 30.0109229, 29.2200637, 27.0014222, 21.3619739, 14.6300691, 7.33348529, 0., -4.81522689, -9.42275327, -14.1057546, -19.1474064, -26.7088395, -34.4440918, -41.8851566, -48.5640273, -54.9673069, -59.290535, -60.683861, -58.2974344, -60.3669104, -53.3227302, -32.6808411, 6.04281005, 38.8819919, 110.151155, 235.714466, 431.436092, 699.736845, 1096.81095, 1654.65389, 2405.26399, 3404.86948, 4612.77606, 6012.52008, 7587.63786},
        {-117.464876, -123.918365, -127.492752, -128.518553, -127.326281, -124.246452, -119.609581, -113.746181, -106.986767, -100.040866, -92.7083752, -85.1682045, -77.5992633, -70.6372629, -63.8215906, -57.1484355, -50.6139865, -44.2722108, -38.038408, -31.8856556, -25.7870314, -19.5202853, -13.3319539, -7.27324596, -1.3953703, 4.18790751, 9.51295815, 14.5535955, 19.2836336, 24.2243411, 28.5830951, 32.1147276, 34.5740706, 35.1779179, 34.4343549, 32.3134286, 28.7851863, 22.7771943, 15.718973, 7.99756168, 0., -6.61155838, -13.2352339, -19.9940327, -27.010961, -35.9903708, -44.8413841, -53.054469, -60.1200932, -64.6781795, -67.4099591, -68.1461179, -66.7173418, -74.5060328, -75.1704745, -63.9206668, -35.9666093, 0.717528466, 68.7915847, 178.151228, 338.692126, 545.47663, 852.90036, 1304.26458, 1880.44809, 2572.04849, 3393.21069, 4352.19278, 5457.25286},
        {-136.527966, -139.159539, -139.607421, -138.121857, -134.953088, -130.351358, -124.566911, -117.849989, -110.450837, -102.93602, -95.112929, -87.1052792, -79.0367843, -71.261443, -63.5805711, -56.0257688, -48.6286365, -41.5501343, -34.6407587, -27.880366, -21.2488126, -14.5876696, -8.07039264, -1.7321522, 4.3918813, 10.3326244, 15.962385, 21.2195577, 26.0425374, 30.7830152, 34.8007711, 37.8688812, 39.7604217, 39.674461, 38.1876865, 35.3027776, 31.0224139, 24.5660901, 17.0329442, 8.73892973, 0., -8.04750415, -16.2361718, -24.5782045, -33.0858038, -42.6581978, -52.0657511, -60.9658549, -69.0159002, -75.50637, -80.6083267, -84.1259248, -85.8633187, -94.7693226, -97.845567, -91.2383424, -71.0939391, -40.7141918, 13.7723711, 99.0816769, 221.929653, 380.179382, 607.105325, 934.522007, 1336.34321, 1800.19873, 2316.75508, 2875.91963, 3467.59973},
        {-194.664249, -181.230941, -169.302868, -158.648225, -149.035212, -140.232023, -132.006856, -124.127907, -116.363375, -107.812862, -99.1805959, -90.502209, -81.8133354, -73.2086007, -64.6410495, -56.1227188, -47.6656453, -39.1592333, -30.7872051, -22.6106505, -14.6906592, -7.17655082, -0.00589349249, 6.79551487, 13.2018764, 19.3079585, 24.9191718, 29.9614923, 34.3608958, 38.1894679, 41.1686312, 43.1659177, 44.0488595, 43.4935761, 41.6355773, 38.4189602, 33.7878219, 26.7947351, 18.6319311, 9.6001171, 0., -9.24386883, -18.7041645, -28.3297179, -38.0693597, -47.9114983, -57.7495558, -67.516532, -77.1454268, -88.0404396, -98.0748906, -106.593299, -112.940186, -120.019097, -122.191915, -117.379548, -103.502907, -87.2876613, -54.3280535, 0.976913021, 84.2282345, 202.753802, 356.973927, 552.751261, 786.612743, 1083.52895, 1364.03162, 1574.7622, 1662.36213},
        {-175.36164, -166.624748, -158.159801, -149.925768, -141.881618, -133.986319, -126.198841, -118.478152, -110.783222, -103.030633, -95.2386948, -87.3833331, -79.440472, -71.2174429, -62.9261997, -54.6101037, -46.3125163, -38.121428, -30.017719, -22.0268988, -14.1744769, -6.30376836, 1.30464509, 8.55237609, 15.3410373, 21.2894899, 26.6951987, 31.5728771, 35.9372383, 40.4691856, 44.2507664, 47.0302182, 48.5557784, 48.0057096, 45.9262139, 42.2935185, 37.0838506, 29.5326914, 20.6533127, 10.7182403, 0., -10.8380537, -22.0705555, -33.5813111, -45.2541263, -57.0379543, -68.7253946, -80.1741938, -91.2420989, -101.801961, -111.69038, -120.759063, -128.859713, -138.865711, -146.398417, -150.100866, -148.616093, -144.359406, -130.692659, -104.749978, -63.6654913, -2.1174713, 75.3923952, 179.345261, 277.973065, 352.592929, 374.43045, 317.234105, 154.752373},
        {-191.788584, -182.390037, -173.740918, -165.639879, -157.885566, -150.276631, -142.611723, -134.68949, -126.308584, -116.144987, -105.569081, -94.8285798, -84.1712001, -74.6575672, -65.3973217, -56.3130147, -47.3271972, -37.9729249, -28.7180425, -19.6408993, -10.8198446, -2.39999017, 5.63378214, 13.2298281, 20.3365033, 26.8791912, 32.8384087, 38.1717007, 42.8366119, 47.2220462, 50.6816452, 53.00041, 53.9633413, 52.8498506, 50.1527642, 45.859319, 39.9567519, 31.9102235, 22.4378773, 11.7357803, 0., -12.5765892, -25.7934503, -39.453239, -53.3586108, -66.8576235, -80.38937, -93.9383452, -107.489044, -121.262378, -134.911858, -148.327413, -161.398972, -174.938597, -187.545228, -198.739941, -208.043809, -213.510736, -216.715838, -217.767058, -216.77234, -208.338342, -209.076867, -219.416934, -264.909434, -355.793203, -538.433422, -850.163683, -1328.31758},
        {-198.71236, -197.308699, -193.101836, -186.510063, -177.951671, -167.844951, -156.608195, -144.659693, -132.417739, -121.285996, -110.303233, -99.4935919, -88.8812147, -78.707498, -68.6924271, -58.7732421, -48.8871829, -38.459504, -28.1442253, -18.0833812, -8.41900594, 0.229792022, 8.38888204, 16.1070594, 23.4331194, 30.88596, 37.8562326, 44.2046916, 49.7920911, 54.7738453, 58.5981843, 61.0079983, 61.7461775, 59.6423142, 55.7179154, 50.0811906, 42.840349, 34.2000096, 24.1334081, 12.7101898, 0., -14.5670711, -30.0260009, -46.0513217, -62.3175661, -77.5333094, -92.7254243, -107.954826, -123.28243, -138.151262, -153.487283, -169.598565, -186.793178, -205.872273, -226.45361, -248.648032, -272.566377, -292.582224, -316.83858, -347.741193, -387.695808, -434.404278, -504.384028, -598.974796, -746.35905, -974.171705, -1285.39517, -1689.17496, -2194.65663}
    };
    
    const double a5tab[50][69] = {
        {4296.48547, 4145.14906, 4028.39347, 3938.25617, 3866.77463, 3805.98633, 3747.92872, 3684.63929, 3608.15549, 3470.8602, 3320.30733, 3164.39619, 3011.02609, 2895.53131, 2787.40219, 2683.56406, 2580.94224, 2467.08826, 2352.05074, 2236.50453, 2121.12447, 2008.42576, 1896.50673, 1785.30605, 1674.76241, 1564.89358, 1455.52754, 1346.57134, 1237.93205, 1128.57981, 1019.73336, 911.67453, 804.685139, 699.870801, 596.360042, 494.10517, 393.058497, 293.389367, 194.746242, 96.9946212, 0., -95.8821865, -191.472354, -287.100981, -383.098545, -476.415091, -572.113704, -671.877035, -777.387735, -883.098866, -1000.81451, -1135.10914, -1290.55726, -1445.5122, -1641.25806, -1892.85779, -2215.37433, -2550.46591, -3015.96209, -3656.28769, -4515.86756, -5562.5241, -7070.48939, -8999.14401, -12015.737, -15367.8837, -23273.8659, -38709.2987, -64649.7973},
        {3989.50138, 3890.74144, 3791.28858, 3691.26149, 3590.77885, 3489.95935, 3388.92168, 3287.78452, 3186.66657, 3088.65256, 2989.7087, 2888.76729, 2784.76057, 2664.7319, 2544.25806, 2427.02688, 2316.72622, 2240.60956, 2169.37286, 2097.27769, 2018.58566, 1902.04814, 1777.64101, 1649.82996, 1523.08064, 1415.81637, 1312.96214, 1213.40058, 1116.01432, 1016.79356, 918.670313, 821.68419, 725.874783, 631.536102, 538.351566, 446.259007, 355.196258, 265.266634, 176.17629, 87.7968658, 0., -86.9254916, -173.692018, -260.594811, -347.929102, -432.365887, -519.274328, -610.399354, -707.48589, -805.717825, -916.025541, -1042.77838, -1190.34568, -1339.01104, -1526.86385, -1767.90776, -2076.14641, -2395.68728, -2838.38865, -3446.21264, -4261.12134, -5252.34273, -6680.04137, -8501.55957, -11361.8428, -14541.9955, -22090.8829, -36865.43, -61722.5616},
        {3757.56237, 3688.98233, 3602.74083, 3502.53272, 3392.05283, 3274.99599, 3155.05703, 3035.9308, 2921.31211, 2833.32287, 2749.86003, 2667.24761, 2581.80961, 2470.83397, 2357.29521, 2245.13178, 2138.28212, 2062.17793, 1990.6671, 1919.09078, 1842.79012, 1736.93836, 1625.1117, 1510.71846, 1397.16693, 1298.16665, 1202.70421, 1110.06741, 1019.54408, 928.906343, 839.563964, 751.411028, 664.341619, 578.018704, 492.659931, 408.25183, 324.780933, 242.505048, 161.030918, 80.2365616, 0., -79.1939341, -158.316759, -237.733181, -317.807904, -395.541068, -476.007772, -560.918545, -651.983919, -744.762196, -849.577029, -970.599841, -1112.00206, -1254.41911, -1434.97281, -1667.24897, -1964.8334, -2274.39487, -2703.20307, -3291.61063, -4079.9702, -5039.50056, -6417.95592, -8173.42849, -10924.8947, -14014.4248, -21230.8298, -35251.2358, -58752.7686},
        {3573.43954, 3518.13048, 3441.64113, 3347.99242, 3241.20528, 3125.30065, 3004.29946, 2882.22265, 2763.09114, 2665.78538, 2573.523, 2484.38113, 2396.4369, 2303.02573, 2208.86315, 2113.92298, 2018.17905, 1918.81705, 1819.71417, 1721.95946, 1626.642, 1540.32011, 1456.42589, 1373.86068, 1291.52585, 1204.16451, 1116.49954, 1029.09559, 942.517315, 859.174238, 777.048176, 695.965823, 615.753875, 535.637321, 456.285245, 377.765026, 300.144041, 223.94418, 148.596504, 73.9865859, 0., -72.5271837, -145.040088, -218.033339, -292.001563, -364.645482, -440.371185, -520.79086, -607.516694, -696.21608, -796.823916, -913.330305, -1049.72535, -1186.00008, -1359.74331, -1584.54477, -1873.9942, -2177.10138, -2597.86799, -3175.71572, -3950.0663, -4894.3482, -6245.96286, -7966.33323, -10645.6449, -13703.476, -20613.072, -33834.2617, -55826.8739},
        {3409.90395, 3356.44466, 3286.88038, 3203.56312, 3108.8449, 3005.07774, 2894.61363, 2779.80461, 2663.00268, 2546.95431, 2433.45928, 2324.71182, 2222.90615, 2150.49542, 2081.31138, 2009.44471, 1928.98609, 1797.5506, 1660.29474, 1525.89938, 1403.04545, 1339.4171, 1289.09066, 1245.14575, 1200.66196, 1126.02999, 1046.0939, 963.008853, 878.930002, 801.853306, 725.756809, 650.459352, 575.779776, 500.710664, 426.227618, 352.479982, 279.6171, 208.423598, 138.159426, 68.7198163, 0., -66.7649103, -133.555516, -201.012536, -269.776693, -338.38397, -410.421719, -487.372554, -570.71909, -656.063581, -753.121147, -865.726549, -997.714547, -1128.01761, -1295.3337, -1513.4585, -1796.18769, -2094.31954, -2509.84608, -3081.76197, -3849.06182, -4787.23624, -6125.79199, -7831.85622, -10464.8454, -13527.4537, -20156.9751, -32582.0535, -53031.3329},
        {3239.72667, 3182.18361, 3117.34947, 3045.16737, 2965.58039, 2878.53165, 2783.96423, 2681.82123, 2572.04576, 2437.74389, 2302.43055, 2172.78365, 2055.48109, 2002.43127, 1956.9894, 1907.74119, 1843.27233, 1685.40227, 1516.18949, 1350.92621, 1204.90465, 1161.45301, 1140.61312, 1130.46279, 1119.07984, 1055.98318, 983.233088, 904.330944, 822.778121, 751.19961, 680.323725, 610.002396, 540.087553, 469.557447, 399.487158, 330.078085, 261.53163, 194.782874, 129.006064, 64.1091301, 0., -61.7467835, -123.556551, -186.188021, -250.399908, -315.46138, -384.216524, -458.019878, -538.225981, -620.2888, -713.823669, -822.545353, -950.168616, -1074.73535, -1235.90234, -1447.6535, -1723.97275, -2016.56203, -2426.60002, -2992.98342, -3754.60894, -4688.5153, -6019.17312, -7721.57993, -10323.2482, -13404.6625, -19781.9045, -31462.1566, -50452.6013},
        {3035.67878, 2973.60609, 2911.93932, 2848.72771, 2782.02045, 2709.86678, 2630.31591, 2541.41704, 2441.2194, 2299.06835, 2153.19849, 2013.14059, 1888.4254, 1848.02151, 1818.24671, 1784.85662, 1733.60687, 1569.39574, 1391.17913, 1217.05561, 1065.12374, 1033.65155, 1028.50034, 1035.70092, 1041.28404, 986.24413, 919.662878, 845.585609, 768.05765, 701.469213, 635.382783, 569.705732, 504.34543, 438.496382, 373.063973, 308.240723, 244.21915, 181.861577, 120.422796, 59.8274049, 0., -57.3124734, -114.736706, -173.077041, -233.137821, -294.582554, -359.81275, -430.089086, -506.672238, -584.87584, -674.286427, -778.543494, -901.286533, -1020.41696, -1175.60758, -1380.79311, -1649.90829, -1934.34159, -2335.59247, -2892.61414, -3644.35983, -4568.53596, -5887.83607, -7587.08678, -10161.6055, -13253.4069, -19407.2257, -30442.1168, -48177.1347},
        {2770.53136, 2708.97085, 2649.54083, 2590.16668, 2528.77379, 2463.28755, 2391.63335, 2311.73658, 2221.52263, 2091.84191, 1958.52477, 1830.32661, 1716.0028, 1676.45437, 1647.43281, 1616.83524, 1572.5588, 1436.55469, 1289.04433, 1144.30323, 1016.60689, 983.236401, 970.259435, 966.74926, 961.779143, 909.032918, 847.129054, 779.296591, 708.764568, 646.91818, 585.567846, 524.68014, 464.221637, 403.846181, 343.958173, 284.649282, 226.011177, 168.499276, 111.696, 55.5475182, 0., -53.3016495, -106.789491, -161.196848, -217.257045, -274.452337, -335.267548, -400.93643, -472.692737, -545.808803, -629.864368, -728.477751, -845.267274, -959.326113, -1108.60779, -1306.54069, -1566.55318, -1838.17091, -2224.28609, -2763.88818, -3495.96666, -4397.64884, -5693.51063, -7379.95924, -9920.66925, -12991.9915, -18952.3044, -29489.4796, -46291.3886},
        {2417.05548, 2366.53666, 2309.0449, 2245.40682, 2176.44909, 2102.99835, 2025.88123, 1945.9244, 1863.95449, 1776.9788, 1691.17107, 1608.88567, 1532.477, 1476.91808, 1426.8972, 1379.72126, 1332.69721, 1273.9028, 1213.56578, 1152.68474, 1092.25826, 1037.43127, 983.397481, 929.496943, 875.06971, 816.569606, 757.3774, 697.987635, 638.894853, 581.802574, 525.512772, 470.036397, 415.384401, 361.925557, 309.169865, 256.985149, 205.239231, 153.535543, 102.112055, 50.9423473, 0., -49.5539819, -99.4084146, -150.064689, -202.024195, -253.775575, -308.638068, -367.918164, -432.922351, -499.071792, -575.912436, -667.104903, -776.309814, -885.726459, -1029.06132, -1218.55955, -1466.46631, -1718.56271, -2080.14355, -2590.0396, -3287.08162, -4146.20453, -5397.92663, -7051.77975, -9541.19154, -12538.7207, -18336.5059, -28571.7906, -44881.8187},
        {2185.86243, 2129.66864, 2072.54176, 2014.18321, 1954.2944, 1892.57674, 1828.73164, 1762.46052, 1693.46479, 1613.81027, 1533.88821, 1456.45424, 1384.26402, 1331.09789, 1284.27693, 1242.1469, 1203.05357, 1164.97587, 1126.77315, 1086.93792, 1043.96269, 989.074843, 930.938071, 870.950935, 810.511996, 754.725473, 699.802008, 645.657899, 592.209444, 538.934767, 486.363612, 434.587547, 383.69814, 334.548323, 286.163756, 238.331466, 190.838476, 142.966997, 95.210792, 47.5588098, 0., -46.4100171, -93.1756315, -140.734562, -189.524529, -237.936684, -289.27394, -344.79264, -405.74913, -467.474595, -539.520605, -625.513568, -729.079894, -833.298458, -970.562217, -1152.71659, -1391.60701, -1630.84359, -1977.80117, -2471.61931, -3151.43754, -3991.40953, -5225.63246, -6875.89452, -9345.32147, -12339.851, -17972.8354, -27755.7736, -43200.1643},
        {1981.4486, 1917.74747, 1860.59169, 1808.1705, 1758.6731, 1710.28871, 1661.20656, 1609.61586, 1553.70582, 1477.58662, 1399.15813, 1322.2412, 1250.65666, 1201.62972, 1160.21509, 1124.87184, 1094.05904, 1069.0927, 1044.43221, 1017.39389, 985.294068, 932.67365, 874.734515, 813.90314, 752.606002, 700.319407, 649.600069, 600.054533, 551.289343, 501.502803, 452.272994, 403.769755, 356.162928, 310.714537, 266.065364, 221.948375, 178.096535, 133.578698, 89.0575891, 44.5318191, 0., -43.5656426, -87.5295569, -132.282577, -178.215537, -223.723063, -271.99068, -324.207705, -381.563456, -439.300701, -506.933925, -588.031063, -686.16005, -784.990737, -915.94838, -1090.56015, -1320.35322, -1546.9207, -1879.69744, -2358.18424, -3021.88189, -3844.23415, -5062.91291, -6711.52012, -9164.32682, -12162.1323, -17638.9538, -26980.504, -41572.4958},
        {1794.4152, 1724.40808, 1667.3188, 1620.06974, 1579.58327, 1542.78174, 1506.58753, 1467.923, 1423.71051, 1351.59344, 1275.48475, 1200.01839, 1129.82833, 1084.63602, 1047.95293, 1018.37801, 994.510209, 980.512175, 967.19371, 950.928297, 928.089424, 877.880279, 820.71276, 759.828472, 698.469016, 649.593536, 602.839079, 557.560233, 513.111583, 466.653192, 420.611983, 375.220355, 330.710706, 288.654201, 247.408964, 206.671883, 166.13985, 124.746758, 83.2576926, 41.6747436, 0., -40.8980775, -82.2303199, -124.341187, -167.575137, -210.340302, -255.692002, -304.749227, -358.630969, -412.54525, -475.886418, -552.13785, -644.782925, -738.021909, -862.334537, -1028.91743, -1248.96722, -1462.16565, -1779.83017, -2241.76334, -2887.76773, -3690.53634, -4891.20049, -6534.3606, -8967.55178, -11963.4481, -17291.233, -26220.9625, -40022.6926},
        {1628.55148, 1554.65838, 1498.11445, 1454.99224, 1421.36432, 1393.30328, 1366.88168, 1338.17208, 1303.24707, 1235.66334, 1163.01568, 1090.38301, 1022.84422, 981.368108, 948.787774, 923.826191, 905.206329, 898.844712, 893.393335, 884.697746, 868.603494, 821.350866, 766.232776, 706.936878, 647.150825, 601.764906, 558.783084, 517.41196, 476.858131, 433.670357, 390.776215, 348.44544, 306.947768, 268.010996, 229.863574, 192.192013, 154.682824, 116.244139, 77.6521993, 38.9048655, 0., -38.3123138, -77.0873241, -116.628058, -157.237541, -197.364371, -239.907775, -285.912554, -336.423505, -386.690844, -445.871786, -517.328963, -604.425008, -691.796217, -809.022092, -966.955801, -1176.45051, -1375.54392, -1677.03085, -2120.89065, -2747.10266, -3527.644, -4706.50075, -6337.96158, -8745.4887, -11730.9564, -16913.4172, -25457.6291, -38528.3499},
        {1487.64668, 1413.5063, 1358.36997, 1318.04925, 1288.35567, 1265.10079, 1244.09613, 1221.15325, 1192.0837, 1129.6289, 1061.89856, 993.932245, 930.769552, 893.077305, 864.016948, 842.37717, 826.946657, 823.700716, 821.366769, 815.858856, 803.091017, 759.741547, 708.65453, 653.438306, 597.701215, 556.050561, 516.696131, 478.846677, 441.710953, 401.838723, 362.161324, 322.951103, 284.480408, 248.428604, 213.098217, 178.198789, 143.439861, 107.843807, 72.0822061, 36.1394671, 0., -35.7133438, -71.9099732, -108.860855, -146.836957, -184.371235, -224.167871, -267.193037, -314.412902, -361.220084, -416.38373, -483.099433, -564.562788, -645.717902, -755.312447, -903.842611, -1101.80458, -1286.02099, -1570.131, -1994.10022, -2597.89427, -3352.88507, -4504.81926, -6115.86863, -8488.62993, -11451.815, -16489.2504, -24670.9838, -37067.063},
        {1375.49005, 1305.95973, 1253.47674, 1214.35207, 1184.89673, 1161.42172, 1140.23803, 1117.65665, 1089.9886, 1033.32271, 972.280995, 911.263313, 854.669523, 821.014935, 794.937775, 775.19172, 760.530448, 754.690595, 749.449696, 741.568244, 727.806733, 689.708458, 645.337991, 597.542705, 549.16997, 511.667545, 475.842264, 441.101348, 406.852016, 370.442713, 334.162944, 298.243438, 262.914923, 229.550707, 196.781911, 164.382231, 132.125366, 99.3187258, 66.3888102, 33.2958309, 0., -33.0061597, -66.5076708, -100.757245, -136.007596, -170.936864, -208.00216, -248.086024, -292.070995, -335.615571, -386.915951, -448.944294, -524.672755, -599.191206, -700.507005, -838.745224, -1024.03094, -1192.56234, -1457.96212, -1859.92612, -2438.15015, -3163.58747, -4282.16159, -5861.62737, -8187.4678, -11113.1819, -16002.4765, -23841.5066, -35616.4271},
        {1295.87085, 1237.02659, 1188.82609, 1149.01198, 1115.32691, 1085.51353, 1057.31449, 1028.47242, 996.729976, 946.577353, 894.310623, 842.97341, 795.609339, 766.432322, 742.847581, 723.430622, 706.756955, 691.424756, 675.977799, 658.982526, 639.005378, 607.907736, 573.643128, 537.46002, 500.606877, 467.832902, 435.48553, 403.412934, 371.463285, 338.766752, 306.176712, 273.828538, 241.857605, 211.020988, 180.583677, 150.432364, 120.453741, 90.4418588, 60.4131083, 30.291239, 0., -30.0957535, -60.6898206, -92.0348943, -124.383668, -156.637225, -190.940513, -228.086866, -268.869621, -309.359907, -356.962153, -414.358576, -484.231398, -551.62037, -643.907167, -770.830997, -942.131067, -1094.13344, -1339.35572, -1716.90239, -2265.87789, -2957.07913, -4034.53329, -5568.78338, -7832.49467, -10702.2149, -15436.8397, -22949.6777, -34154.0375},
        {1204.65587, 1157.00006, 1114.47906, 1076.03051, 1040.59206, 1007.10134, 974.496, 941.713669, 907.69199, 863.90708, 819.742711, 777.121132, 737.964591, 712.778866, 691.469265, 672.524625, 654.433783, 631.935693, 608.769028, 584.922579, 560.385137, 535.126853, 509.162613, 482.488661, 455.101244, 426.520612, 397.409404, 367.954264, 338.341835, 309.081426, 279.907949, 250.878982, 222.052103, 193.704483, 165.586269, 137.667202, 109.917023, 82.4541578, 55.0401876, 27.5853794, 0., -27.3906496, -55.2573507, -83.8558505, -113.441896, -143.126009, -174.767252, -209.079461, -246.776473, -284.355627, -328.433857, -381.411599, -445.689291, -506.351102, -590.040242, -706.083651, -863.808272, -999.956678, -1225.47393, -1578.72072, -2098.05773, -2753.3297, -3788.44518, -5278.67022, -7480.52598, -10279.4618, -14863.9593, -22096.742, -32840.5336},
        {1074.63988, 1033.0459, 996.321555, 963.439721, 933.373293, 905.095166, 877.57823, 849.795378, 820.719503, 782.391091, 743.488403, 705.757293, 670.943616, 648.479622, 629.350211, 612.226682, 595.780332, 575.305119, 554.200613, 532.489048, 510.192655, 487.464298, 464.143325, 440.199715, 415.603449, 389.833046, 363.546529, 336.910461, 310.091403, 283.54333, 257.03043, 230.6043, 204.316538, 178.421378, 152.686729, 127.083133, 101.581135, 76.2399296, 50.9059493, 25.5142782, 0., -25.256491, -50.9435458, -77.3042049, -104.581509, -131.993253, -161.217822, -192.908352, -227.717982, -262.453961, -303.153673, -352.008612, -411.210273, -466.199332, -542.618428, -649.359384, -795.314019, -917.313637, -1125.13479, -1456.4935, -1949.1058, -2570.59931, -3568.95528, -5028.68341, -7183.5431, -9910.74615, -14364.4451, -21379.5571, -31790.9996},
        {952.045299, 916.813429, 886.000482, 858.645986, 833.789466, 810.47045, 787.728463, 764.603033, 740.133686, 707.0761, 673.26719, 640.260019, 609.607655, 589.582905, 572.331196, 556.717696, 541.607574, 522.902498, 503.616539, 483.800265, 463.504248, 443.03075, 422.07797, 400.595797, 378.534124, 355.357107, 331.694667, 307.69099, 283.49026, 259.485015, 235.471748, 211.495303, 187.600525, 164.018334, 140.533068, 117.115138, 93.7349577, 70.3914253, 47.0150752, 23.564927, 0., -23.2517332, -46.8877878, -71.1367257, -96.2271088, -121.468367, -148.375847, -177.545764, -209.574332, -241.585969, -279.037406, -323.913578, -378.199419, -427.713909, -497.074318, -594.73196, -729.138153, -837.32558, -1027.73164, -1337.3751, -1803.27474, -2390.79442, -3351.91758, -4779.57717, -6886.99293, -9541.19849, -13873.6851, -20699.8296, -30835.0089},
        {837.812697, 808.936927, 783.919108, 761.884766, 741.959432, 723.268633, 704.937899, 686.092758, 665.858739, 637.79203, 608.815237, 580.281624, 553.544456, 535.682276, 520.03296, 505.659659, 491.625529, 474.469091, 456.787981, 438.655203, 420.143763, 401.668489, 382.823831, 363.546065, 343.771465, 322.977172, 301.742249, 280.186624, 258.430227, 236.798139, 215.123074, 193.4429, 171.795483, 150.388204, 129.021612, 107.665771, 86.2907419, 64.8428209, 43.325344, 21.7178806, 0., -21.3729548, -43.0874978, -65.3543693, -88.3843094, -111.556974, -136.246622, -162.996426, -192.349561, -221.745105, -256.071965, -297.114955, -346.658886, -390.927315, -453.490813, -542.358694, -665.540274, -760.3361, -933.747761, -1222.06808, -1661.58988, -2215.46107, -3139.40919, -4533.64928, -6593.803, -9176.58984, -13398.9314, -20062.6987, -29969.7628},
        {732.882625, 710.050697, 690.4807, 673.391519, 658.002041, 643.53115, 629.197732, 614.220674, 597.81886, 574.368803, 549.868711, 525.47442, 502.341765, 486.371293, 472.076243, 458.714566, 445.544212, 429.746158, 413.486118, 396.852833, 379.935045, 363.21979, 346.238194, 328.919679, 311.19367, 292.577619, 273.577705, 254.288137, 234.803126, 215.37436, 195.875577, 176.337996, 156.792834, 137.423841, 118.04869, 98.6375877, 79.1607395, 59.5282925, 39.7945341, 19.9536936, 0., -19.6167344, -39.5400971, -59.9580924, -81.0587245, -102.2647, -124.83544, -149.265068, -176.047706, -202.924822, -234.244259, -271.6012, -316.590832, -355.872032, -411.950817, -492.396897, -604.779981, -686.688787, -843.66641, -1111.27496, -1525.07654, -2046.14531, -2933.5072, -4293.19749, -6306.90082, -8822.69125, -12947.4358, -19473.3033, -29192.4625},
        {638.19564, 620.789034, 606.088527, 593.401702, 582.036144, 571.299435, 560.499161, 548.942905, 535.938251, 516.636341, 496.163778, 475.490721, 455.58733, 441.243514, 428.081786, 415.544409, 403.073644, 388.474958, 373.482125, 358.192127, 342.701942, 327.526931, 312.178341, 296.585799, 280.678933, 264.042825, 247.089466, 229.886301, 212.500777, 195.105336, 177.620429, 160.071501, 142.483999, 125.018098, 107.510627, 89.933143, 72.2572021, 54.3820161, 36.3804242, 18.2529207, 0., -17.9796505, -36.2430068, -54.9488516, -74.2559679, -93.5971696, -114.147596, -136.356417, -160.672804, -185.118576, -213.541195, -247.360771, -287.997414, -322.580542, -372.537235, -445.003884, -547.116876, -616.727231, -757.970853, -1005.69828, -1394.76004, -1884.39319, -2736.28873, -4060.51959, -6029.21391, -8485.27376, -12526.4503, -18936.7826, -28500.3094},
        {554.692301, 541.786237, 531.145858, 522.150773, 514.180592, 506.614925, 498.833382, 490.215574, 480.14111, 464.424568, 447.436604, 429.98284, 412.8689, 399.892497, 387.670329, 375.811181, 363.923841, 350.396751, 336.547181, 322.472055, 308.268298, 294.432189, 280.501557, 266.413583, 252.105452, 237.257168, 222.165962, 206.871889, 191.415003, 175.882728, 160.248799, 144.53432, 128.760398, 113.063829, 97.3037508, 81.4549918, 65.4923811, 49.3381679, 33.0407928, 16.5961166, 0., -16.4582818, -33.193648, -50.3276035, -67.9816533, -85.560007, -104.188383, -124.275204, -146.228893, -168.319821, -193.949684, -224.382126, -260.880791, -291.085326, -335.332971, -400.336967, -492.810558, -550.795024, -677.144355, -906.040578, -1271.66572, -1731.75074, -2549.83088, -3837.91334, -5763.66979, -8170.1084, -12143.2269, -18458.2756, -27890.5049},
        {483.313163, 473.676603, 466.055961, 459.874189, 454.554236, 449.519054, 444.191593, 437.994804, 430.351638, 417.563406, 403.423355, 388.60309, 373.77422, 361.911802, 350.462611, 339.176876, 327.804822, 315.252799, 302.452463, 289.491591, 276.457961, 263.777843, 251.065126, 238.272192, 225.351423, 212.105025, 198.695625, 185.135674, 171.437623, 157.598192, 143.651858, 129.617363, 115.513453, 101.453888, 87.3243877, 73.1056889, 58.7785281, 44.3309239, 29.7334185, 14.9638358, 0., -15.0492067, -30.3894418, -46.0953047, -62.2413945, -78.1588373, -94.9630951, -113.026156, -132.72001, -152.52201, -175.456634, -202.653723, -235.243119, -261.418867, -300.420927, -358.55346, -442.120629, -489.235757, -601.67018, -813.004394, -1156.81889, -1589.764, -2376.21074, -3627.6765, -5513.19596, -7882.96623, -11805.0174, -18042.9214, -27360.2502},
        {430.484418, 421.113456, 414.158086, 408.98146, 404.94673, 401.417046, 397.755561, 393.325426, 387.489794, 376.483753, 364.049742, 350.802139, 337.355319, 326.256844, 315.414632, 304.669786, 293.863408, 282.296992, 270.567093, 258.73066, 246.844638, 235.147337, 223.441797, 211.712417, 199.9436, 188.128389, 176.239085, 164.256634, 152.161977, 139.83341, 127.395587, 114.870511, 102.280187, 89.7188522, 77.1073842, 64.4388941, 51.7064931, 39.0573716, 26.2689294, 13.2726458, 0., -13.7764105, -27.9026719, -42.3837523, -57.2246196, -71.5863824, -86.6554121, -102.77422, -120.285319, -137.798375, -158.081883, -182.171493, -211.102857, -233.791427, -268.24113, -320.335691, -395.958842, -432.859693, -532.310436, -727.448645, -1051.41189, -1460.81344, -2218.36381, -3429.83, -5277.77804, -7636.36413, -11536.5847, -17720.2261, -26929.0746},
        {392.308466, 381.186439, 373.266059, 367.790599, 364.003329, 361.147518, 358.466437, 355.203357, 350.601548, 340.48549, 328.88476, 316.410145, 303.672432, 293.084004, 282.733413, 272.510809, 262.30634, 251.700434, 241.016849, 230.269621, 219.472788, 208.611231, 197.739805, 186.884208, 176.07014, 165.51361, 154.973883, 144.400534, 133.743138, 122.753431, 111.657963, 100.485445, 89.2645885, 78.0710907, 66.8678814, 55.664877, 44.4719937, 33.6607917, 22.7348862, 11.559536, 0., -12.6123, -25.6647331, -39.0785057, -52.774824, -65.6832532, -79.1132972, -93.3828187, -108.809681, -124.072008, -141.783297, -162.917303, -188.447785, -208.078692, -238.561512, -285.377927, -354.009618, -381.598833, -469.302458, -649.937946, -956.32275, -1345.70917, -2077.61012, -3248.67972, -5062.9382, -7430.12351, -11329.7366, -17473.8377, -26574.487},
        {360.66773, 347.89344, 338.935119, 332.947164, 329.083975, 326.499951, 324.349491, 321.786994, 317.966858, 308.406182, 297.351585, 285.41239, 273.197914, 263.040569, 253.137348, 243.408333, 233.77361, 224.009152, 214.236795, 204.434269, 194.579301, 184.439194, 174.28627, 164.182425, 154.189557, 144.710428, 135.329724, 125.972995, 116.565793, 106.761765, 96.8671266, 86.916191, 76.9432708, 67.0016617, 57.0990999, 47.2623045, 37.5179949, 28.466852, 19.3320488, 9.91071997, 0., -11.5081985, -23.54943, -35.9644712, -48.5940986, -60.1461305, -72.0474852, -84.5921225, -98.0740026, -111.204635, -126.493411, -144.867268, -167.253148, -184.019583, -210.875282, -252.97055, -315.455689, -335.026177, -412.669073, -580.916609, -872.30102, -1244.6194, -1954.6094, -3090.28309, -4876.46181, -7258.87633, -11162.8104, -17268.6597, -26256.8197},
        {332.848826, 319.339406, 309.84285, 303.496527, 299.437804, 296.804049, 294.732628, 292.360909, 288.82626, 279.6239, 268.990204, 257.519397, 245.805708, 236.102767, 226.681634, 217.472775, 208.406655, 199.343792, 190.312576, 181.27145, 172.178856, 162.676833, 153.16679, 143.733731, 134.462662, 125.865105, 117.42894, 109.068565, 100.698377, 91.9270899, 83.0970604, 74.2449607, 65.4074631, 56.6069349, 47.9000748, 39.3292765, 30.936934, 23.542257, 16.1000972, 8.34212235, 0., -10.4410908, -21.5043985, -32.9596599, -44.576612, -54.88014, -65.3827734, -76.3521899, -88.0560672, -99.2001149, -112.238766, -128.064485, -147.569738, -161.556655, -185.044169, -222.960877, -280.235378, -293.441953, -363.205243, -521.795574, -801.483269, -1160.10306, -1853.23206, -2962.50825, -4729.38986, -7132.2067, -11041.982, -17106.5658, -25973.8078},
        {304.952152, 292.952218, 284.33093, 278.349426, 274.268842, 271.350315, 268.854983, 266.043982, 262.17845, 253.319746, 243.208695, 232.386348, 221.393753, 212.329699, 203.554397, 194.985801, 186.541861, 178.001582, 169.477443, 160.942978, 152.371717, 143.465933, 134.578921, 125.792715, 117.189349, 109.241916, 101.484971, 93.8441289, 86.2450012, 78.3468177, 70.4481283, 62.5810991, 54.7778966, 47.0124833, 39.3985105, 31.9914262, 24.8466778, 18.9719631, 13.08958, 6.87407634, 0., -9.38029515, -19.4611898, -29.9592584, -40.5910756, -49.7813155, -59.0552134, -68.6461041, -78.7873225, -88.1276716, -99.1188308, -112.627947, -129.522169, -140.631068, -160.874397, -195.134334, -248.293057, -257.393244, -322.292372, -475.008418, -747.559361, -1096.57811, -1780.23784, -2878.9862, -4641.40046, -7069.0396, -10983.4062, -17001.0171, -25738.3888},
        {279.78566, 269.98824, 262.640456, 257.177906, 253.036189, 249.650905, 246.457651, 242.892026, 238.389629, 229.788609, 220.160994, 209.981364, 199.724297, 191.311845, 183.192125, 175.260728, 167.413243, 159.281274, 151.129992, 142.960581, 134.774227, 126.404438, 118.087145, 109.890602, 101.883063, 94.4378202, 87.1960756, 80.1040686, 73.1080386, 65.9558819, 58.8715185, 51.8805254, 45.0084794, 38.1787473, 31.56, 25.2186982, 19.2213027, 14.7377201, 10.2895871, 5.50198627, 0., -8.3384809, -17.4423055, -26.9875147, -36.6501492, -44.7971021, -52.9372211, -61.2702062, -69.9957574, -77.6869199, -86.8207102, -98.2474902, -112.817622, -121.191888, -138.48606, -169.626333, -219.538899, -226.031281, -287.995812, -437.206154, -705.435969, -1047.95037, -1726.04867, -2820.76157, -4583.33267, -7035.48163, -10946.811, -16899.9525, -25477.5381},
        {257.302813, 250.074573, 244.204199, 239.330028, 235.090398, 231.123647, 227.068111, 222.562129, 217.244038, 208.851392, 199.683625, 190.139389, 180.617334, 172.839114, 165.351176, 158.02297, 150.723945, 142.906536, 135.024016, 127.112641, 119.208668, 111.321731, 103.525358, 95.8664567, 88.3919317, 81.3337627, 74.4797527, 67.8027781, 61.2757154, 54.7583607, 48.3819031, 42.1644511, 36.1241133, 30.1364421, 24.4191248, 19.0472927, 14.0960769, 10.8651988, 7.71536327, 4.23186531, 0., -7.30986874, -15.4315422, -24.013753, -32.7052334, -39.8530662, -46.9282936, -54.100308, -61.5385019, -67.7347666, -75.206996, -84.795583, -97.3409206, -103.184816, -117.865682, -146.423345, -193.897633, -199.101319, -259.792105, -407.500641, -673.757574, -1012.59923, -1688.03923, -2782.74657, -4546.80967, -7020.11695, -10915.9374, -16779.1949, -25154.8133},
        {237.457073, 232.838319, 228.454933, 224.153853, 219.782019, 215.186369, 210.213843, 204.71138, 198.525917, 190.328997, 181.613115, 172.69537, 163.892858, 156.701414, 149.787905, 142.997937, 136.177112, 128.601035, 120.91331, 113.187537, 105.497321, 98.0471923, 90.727455, 83.5593415, 76.5640841, 69.8106881, 63.253503, 56.8946518, 50.7362571, 44.7583326, 38.993954, 33.4540878, 28.1497004, 22.9162824, 18.0104665, 13.5134097, 9.5062687, 7.38006997, 5.38215293, 3.0697266, 0., -6.28867938, -13.4126964, -21.0072975, -28.7077289, -34.8747742, -40.9279278, -47.0122213, -53.2726862, -58.1281187, -64.1402798, -72.1446952, -82.9768906, -86.555554, -98.9997838, -125.511841, -171.293985, -176.348612, -237.157795, -385.00374, -651.168655, -988.904083, -1663.58422, -2759.85339, -4523.45463, -7011.52972, -10874.5267, -16614.567, -24733.7721},
        {221.112345, 217.273009, 213.245771, 208.971399, 204.390662, 199.444326, 194.07316, 188.21793, 181.819406, 174.07591, 165.967632, 157.732316, 149.607708, 142.759168, 136.12578, 129.574241, 122.971252, 115.636065, 108.201803, 100.754142, 93.3787588, 86.3649399, 79.5133073, 72.8280932, 66.31353, 59.9400399, 53.7591891, 47.7887339, 42.0464305, 36.5675385, 31.3453094, 26.3904981, 21.7138595, 17.1544906, 12.9634673, 9.22020788, 6.00413027, 4.68026363, 3.52817058, 2.11302482, 0., -5.20156086, -11.1883212, -17.612775, -24.1274164, -29.1515342, -34.0641099, -39.0109194, -44.1377391, -47.8775291, -52.7740078, -59.6580777, -69.3606412, -71.3464121, -82.3589571, -107.775654, -152.973881, -158.692157, -220.802263, -370.537118, -639.129644, -981.466477, -1657.81939, -2751.51569, -4503.96735, -6999.36377, -10792.7001, -16321.2699, -24022.367},
        {207.178203, 203.732178, 199.820678, 195.45373, 190.64136, 185.393593, 179.720456, 173.631975, 167.138176, 159.874527, 152.375435, 144.800749, 137.310317, 130.721435, 124.273526, 117.863461, 111.388109, 104.253972, 97.0444365, 89.8525206, 82.7712422, 76.1562566, 69.7328892, 63.4891028, 57.4128605, 51.3843422, 45.5424068, 39.9181301, 34.5425882, 29.4933277, 24.7363656, 20.2841896, 16.1492874, 12.1875237, 8.63065835, 5.55382834, 3.03217056, 2.36462843, 1.91300975, 1.26292884, 0., -4.1202108, -8.95002157, -14.1717987, -19.4679084, -23.3384285, -27.1209295, -30.9706931, -35.0430011, -37.8103107, -41.7838582, -47.7920555, -56.6633142, -57.3923226, -67.3747053, -92.1723635, -137.347198, -144.531429, -208.788512, -361.25222, -633.056326, -981.531677, -1659.22083, -2747.46148, -4481.02972, -6970.9883, -10669.8149, -15944.6335, -23162.5681},
        {195.375656, 192.004692, 188.015746, 183.466195, 178.413415, 172.914785, 167.027681, 160.809479, 154.317557, 147.537176, 140.626675, 133.672277, 126.760207, 120.360119, 114.021432, 107.676999, 101.25967, 94.2939214, 87.28433, 80.3270978, 73.5184264, 67.2620822, 61.2236764, 55.3763849, 49.6933835, 43.9759018, 38.4378405, 33.1211539, 28.0677965, 23.3846307, 19.0227396, 14.9981141, 11.3267455, 7.89149902, 4.89474193, 2.40571575, 0.493661944, 0.363790214, 0.494986523, 0.502105057, 0., -3.05610252, -6.72786903, -10.7365941, -14.8035724, -17.5234724, -20.1948652, -22.9896958, -26.0799092, -27.9994607, -31.213481, -36.549111, -44.8334916, -44.6143205, -53.9099591, -78.4593255, -124.001338, -133.312004, -200.278317, -355.824359, -630.8742119999999, -985.890452, -1663.18168, -2741.44858, -4446.50068, -6914.71693, -10493.0898, -15475.9586, -22157.6622},
        {185.425713, 181.879415, 177.667063, 172.87414, 167.58613, 161.888515, 155.86678, 149.606408, 143.192881, 136.876184, 130.511499, 124.11851, 117.716901, 111.447124, 105.159785, 98.8262595, 92.4179225, 85.5950808, 78.7646057, 72.0222999, 65.4639662, 59.5233565, 53.8231446, 48.3239537, 42.9864068, 37.5470253, 32.2801747, 27.2361189, 22.4651217, 18.090378, 14.0600481, 10.3952234, 7.11699527, 4.1425339, 1.63842043, -0.332685234, -1.7081232, -1.39162518, -0.767583015, -0.186780138, 0., -2.0207093, -4.55193494, -7.35938705, -10.2087758, -11.7946811, -13.3823954, -15.166081, -17.3399, -18.5179766, -21.1065264, -25.9317269, -33.8197554, -32.9334411, -41.8276492, -66.3938965, -112.5237, -124.479458, -194.433453, -352.92885, -630.508812, -991.333571, -1665.09509, -2727.23475, -4392.23916, -6818.86329, -10249.7435, -14906.5457, -21010.9359},
        {177.049383, 173.145211, 168.610719, 163.542914, 158.038804, 152.195397, 146.109701, 139.878725, 133.599477, 127.703878, 121.820057, 115.911057, 109.93992, 103.754355, 97.4788721, 91.1226459, 84.6948522, 77.9966166, 71.3283849, 64.7825529, 58.4515163, 52.7810191, 47.3687694, 42.1658237, 37.1232385, 31.9300195, 26.9040941, 22.1013388, 17.5776303, 13.4595, 9.70390802, 6.33846907, 3.39079796, 0.816745601, -1.25560381, -2.76992995, -3.66991249, -2.97099193, -1.91638277, -0.821060353, 0., -1.02550445, -2.45229068, -4.09240305, -5.75788596, -6.24006989, -6.77999864, -7.59800213, -8.91441027, -9.4388557, -11.5066445, -15.9423856, -23.5706878, -22.2707196, -30.9907063, -55.7334329, -102.501684, -117.479367, -190.415696, -351.241008, -629.885639, -994.651801, -1660.35421, -2698.57779, -4310.10409, -6671.741, -9926.99437, -14227.6955, -19725.6759},
        {187.82453, 175.380638, 164.522254, 155.048751, 146.759502, 139.453881, 132.931261, 126.991014, 121.432515, 115.682444, 110.061944, 104.519465, 99.0034562, 93.4094581, 87.7599952, 82.0246817, 76.1731321, 69.8497126, 63.4793845, 57.1618613, 50.9968561, 45.3113449, 39.8868738, 34.732251, 29.8562852, 25.3107344, 21.0442775, 17.0485431, 13.3151599, 9.78749463, 6.52474233, 3.53783623, 0.837709548, -1.92123783, -4.22892593, -5.93180815, -6.87633789, -5.53361613, -3.67558966, -1.69885283, 0., -0.623134421, -1.65833802, -2.8432016, -3.91531598, -3.23888229, -2.47423689, -1.90832646, -1.82809767, -0.571915493, -1.15474102, -4.64295363, -12.1029327, -13.5154792, -25.4667823, -53.4574527, -102.988101, -117.986443, -190.155143, -349.623968, -626.522688, -999.557189, -1653.12889, -2646.10523, -4169.23759, -6411.62763, -9437.09199, -13340.7686, -18217.7952},
        {202.594033, 179.751196, 161.257808, 146.529047, 134.980092, 126.026122, 119.082316, 113.563853, 108.885911, 103.059218, 97.4651852, 92.0807714, 86.8829361, 82.0677256, 77.3053773, 72.4852159, 67.4965658, 61.6813602, 55.6952715, 49.6465805, 43.6435684, 37.8332563, 32.2696889, 27.0456511, 22.2539279, 18.4601648, 15.0951418, 12.0624993, 9.26587801, 6.39380025, 3.65107215, 1.02738158, -1.48758357, -4.64361367, -7.41575128, -9.51851721, -10.6664323, -8.53514926, -5.69360421, -2.67186512, 0., -0.508669667, -1.50711222, -2.60515855, -3.41263955, -1.63513838, 0.451567238, 2.47594729, 4.06647178, 7.34644229, 8.45156461, 6.01237612, -1.34058582, -6.53591649, -22.7602928, -54.7595243, -107.279421, -121.97668, -191.921867, -347.096438, -617.481845, -993.994698, -1623.81099, -2550.28408, -3950.79094, -6012.26715, -8749.39028, -12219.9024, -16481.5456},
        {202.282333, 175.871048, 154.829792, 138.423698, 125.9179, 116.577528, 109.667717, 104.453599, 100.200306, 94.3003516, 88.640535, 83.2350361, 78.0980346, 73.6586754, 69.3501876, 65.0207652, 60.5186022, 55.0509575, 49.3633342, 43.5603002, 37.7464236, 31.9534243, 26.3878575, 21.1834305, 16.4738504, 13.064521, 10.1487743, 7.59163891, 5.25814349, 2.701185, 0.222776282, -2.18720142, -4.53886688, -7.71996386, -10.5119361, -12.5738524, -13.5647814, -10.8155658, -7.24479091, -3.44281592, 0., -0.08641868, -0.680828277, -1.34170145, -1.62751084, 1.08043447, 4.17563281, 7.22874608, 9.81043621, 14.0900627, 16.0001108, 14.0717634, 6.83620339, -0.0235436439, -17.9888746, -51.3893436, -104.554505, -121.858854, -191.569027, -341.996602, -601.453156, -968.32838, -1560.69951, -2412.60149, -3678.80089, -5530.33367, -7952.42665, -10976.7589, -14635.0096},
        {191.434042, 166.516391, 146.685199, 131.231755, 119.447344, 110.623254, 104.050772, 99.021184, 94.8257772, 88.8485881, 83.051054, 77.4873619, 72.211699, 67.8168808, 63.6030143, 59.4088351, 55.0730788, 49.8012396, 44.318591, 38.7171651, 33.088994, 27.4028755, 21.9233699, 16.7918031, 12.1495011, 8.82957003, 6.00484342, 3.539935, 1.29945857, -1.20513783, -3.61480841, -5.9236731, -8.12585181, -11.0434715, -13.5114422, -15.1926812, -15.7501056, -12.5223676, -8.42635531, -4.05469173, 0., 0.584774086, 0.691459186, 0.751561269, 1.1965863, 4.6894944, 8.53775573, 12.2802946, 15.4560354, 19.9833366, 22.0699146, 20.30292, 13.2695034, 6.39295062, -11.3101771, -43.9871831, -95.785371, -117.497179, -187.966722, -332.683249, -577.13601, -923.480797, -1467.20723, -2237.65049, -3360.76265, -4981.95802, -7072.18514, -9646.00898, -12717.9945},
        {174.593769, 154.463423, 138.271025, 125.452267, 115.442846, 107.678454, 101.594787, 96.6275385, 92.2124029, 86.1466711, 80.1598024, 74.3428524, 68.7868768, 64.1769151, 59.7724458, 55.426931, 50.993833, 45.7749414, 40.3960599, 34.9313197, 29.4548519, 23.9126366, 18.5582165, 13.5169832, 8.91432833, 5.46107933, 2.46301766, -0.188639368, -2.60267448, -5.23995495, -7.71634509, -9.99979375, -12.0582498, -14.5073195, -16.4082314, -17.4698714, -17.4011254, -13.8030567, -9.33550295, -4.55047905, 0., 1.44606418, 2.48069555, 3.47902116, 4.81616805, 8.97369958, 13.3777319, 17.5608176, 21.055509, 25.3506551, 27.2399934, 25.4735579, 18.8013827, 13.0848776, -2.88184972, -33.1933156, -81.9440363, -108.755869, -179.985053, -317.515168, -543.229795, -860.374511, -1346.74691, -2030.02411, -3004.17146, -4383.27103, -6134.6498, -8262.3235, -10770.3079},
        {156.306122, 142.488343, 131.034262, 121.584286, 113.778824, 107.258283, 101.66307, 96.6335931, 91.81026, 85.6373442, 79.4298407, 73.3066107, 67.3865153, 62.3733517, 57.5670705, 52.8525584, 48.1147021, 42.814798, 37.4307592, 32.0169088, 26.6275694, 21.2137341, 15.9743875, 11.0051847, 6.40178052, 2.66481611, -0.677034553, -3.69011116, -6.44075338, -9.31805307, -11.9364969, -14.2333237, -16.1457723, -18.0046911, -19.1962656, -19.5002906, -18.6965613, -14.805135, -10.0694394, -4.97316436, 0., 2.43860714, 4.5578262, 6.6450698, 8.98775053, 13.7147082, 18.5353573, 23.0005397, 26.6610969, 30.516409, 32.0893643, 30.3513891, 24.2739098, 20.4235486, 7.13845799, -19.6480136, -64.0025181, -95.4991373, -166.494121, -294.851148, -498.433899, -779.932084, -1202.73129, -1794.31537, -2616.52254, -3750.40353, -5165.80466, -6860.37338, -8831.75711},
        {141.115714, 133.367348, 126.421904, 120.126861, 114.329698, 108.877894, 103.618928, 98.4002787, 93.0694261, 86.7633508, 80.3242294, 73.8837403, 67.5735619, 62.0407637, 56.6954766, 51.4632226, 46.2695237, 40.7635444, 35.2577072, 29.788077, 24.3907188, 19.0371946, 13.8538735, 8.90262193, 4.24530605, 0.146547664, -3.61564479, -7.06050731, -10.2072759, -13.3542189, -16.129927, -18.4420232, -20.1981307, -21.4287693, -21.8695065, -21.3788065, -19.8151337, -15.6761045, -10.7253702, -5.36573418, 0., 3.50355851, 6.79379651, 10.0540987, 13.4678499, 18.6941784, 23.8504278, 28.5296855, 32.3250388, 35.8049895, 37.1970445, 35.7041252, 30.5291532, 28.7802748, 18.5930966, -3.99155006, -42.9328337, -77.5911981, -146.364027, -263.049978, -441.44771, -683.076077, -1038.57315, -1535.11731, -2205.31111, -3099.48633, -4191.63378, -5474.82953, -6942.14962},
        {103.940274, 109.246159, 112.100836, 112.786499, 111.585343, 108.779563, 104.651353, 99.482908, 93.556423, 87.4462621, 81.0255828, 74.4597118, 67.9139761, 62.0289608, 56.3046312, 50.7162112, 45.2389241, 39.780798, 34.4111305, 29.1320235, 23.945579, 18.7904806, 13.7576159, 8.87445415, 4.16846479, -0.307552597, -4.56159118, -8.57631379, -12.3343833, -16.2187059, -19.6516038, -22.4556421, -24.4533865, -25.0813324, -24.702543, -23.294012, -20.8327326, -16.4901945, -11.3710965, -5.77463332, 0., 4.75478168, 9.44887446, 14.1426142, 18.8963367, 25.0263267, 30.8345917, 35.8790878, 39.7177713, 41.4854709, 41.3325214, 38.9861302, 34.1735046, 36.3321689, 31.5948865, 15.8047379, -15.1951963, -53.3876082, -119.973336, -223.978989, -374.43118, -567.576215, -850.781612, -1259.91447, -1784.82538, -2422.06506, -3187.40028, -4091.7938, -5146.20839},
        {119.696459, 121.904076, 122.249633, 120.94698, 118.209966, 114.25244, 109.288254, 103.531255, 97.1952948, 90.7420274, 84.0383753, 77.1990662, 70.3388277, 63.8365474, 57.4371289, 51.1496357, 44.9831315, 38.9566103, 33.0652331, 27.3140912, 21.708276, 16.2226188, 10.9045753, 5.77134083, 0.84011102, -3.93325859, -8.44569664, -12.6554717, -16.5208523, -20.3073567, -23.5431039, -26.0634625, -27.7038011, -27.8850365, -27.02277, -25.1181511, -22.1723292, -17.5689495, -12.1736673, -6.23463379, 0., 5.72320321, 11.4692564, 17.2135602, 22.9315156, 29.3276159, 35.3565322, 40.7020283, 45.0478681, 48.0136362, 49.3729471, 48.8352361, 46.1099386, 48.6623724, 45.3437374, 32.7611163, 7.52159136, -26.3360594, -82.6091273, -167.663208, -287.863898, -441.843924, -659.167485, -966.431705, -1346.42536, -1792.66723, -2299.56444, -2861.30205, -3472.06512},
        {169.150239, 157.676836, 147.513519, 138.460033, 130.316123, 122.881534, 115.956011, 109.3393, 102.831146, 95.6550193, 88.4174499, 81.1486922, 73.8790011, 66.6745961, 59.5153815, 52.4172262, 45.3959992, 38.4031129, 31.5446755, 24.8623387, 18.3977542, 12.2411791, 6.36621736, 0.795078651, -4.45002757, -9.41572308, -13.9834346, -18.10342, -21.7259373, -24.9162069, -27.4635394, -29.2722079, -30.2464854, -30.1545278, -29.0911722, -27.0151386, -23.8851471, -18.9590206, -13.1767348, -6.77736834, 0., 6.50069279, 13.0674696, 19.6274911, 26.1079182, 32.5190318, 38.6716244, 44.4596088, 49.7768977, 55.8502074, 60.7075256, 63.7096435, 64.2173524, 64.9421617, 60.5538573, 49.0729433, 28.5199236, 5.13411772, -34.5703119, -95.8603875, -184.003131, -305.004472, -461.914711, -660.672514, -900.983227, -1205.90322, -1511.31808, -1774.90607, -1954.34547},
        {157.588269, 149.87012, 142.397846, 135.138514, 128.05919, 121.126941, 114.308835, 107.571936, 100.883312, 94.184065, 87.4776118, 80.7414055, 73.952899, 66.9294522, 59.8726476, 52.8239751, 45.8249243, 38.9794848, 32.2416468, 25.6279002, 19.1547349, 12.7006072, 6.47525399, 0.550378712, -5.00231531, -9.89727766, -14.3621909, -18.4108905, -22.0572119, -25.8017927, -28.9769454, -31.4017847, -32.8954252, -32.8713233, -31.7165151, -29.4123786, -25.9402917, -20.6885535, -14.4688524, -7.49979804, 0., 7.57072875, 15.3314634, 23.1600759, 30.9344379, 38.6372245, 45.9995829, 52.8574638, 59.0468175, 64.5920658, 69.0652995, 72.2270807, 73.8379713, 76.5085846, 76.0094109, 70.9609919, 59.9838691, 45.5410236, 20.8735817, -16.934891, -70.8008287, -145.218014, -238.370836, -359.648158, -484.977861, -600.901629, -684.966921, -716.969752, -676.706137},
        {170.100056, 162.077358, 154.716959, 147.843675, 141.282325, 134.857726, 128.394696, 121.718051, 114.652611, 106.038997, 97.0798998, 87.993815, 78.9992377, 71.0318317, 63.3060567, 55.7535409, 48.3059124, 40.5721135, 32.9355325, 25.4568718, 18.1968339, 11.2698263, 4.66136436, -1.58933151, -7.44304092, -12.8267762, -17.748591, -22.1827721, -26.1036058, -29.8062716, -32.815806, -34.9781382, -36.1391978, -35.7891867, -34.2720525, -31.5760155, -27.6892961, -22.1624404, -15.5964124, -8.15450233, 0., 8.75252273, 17.8715705, 27.174366, 36.478132, 45.3117409, 53.896106, 62.1637903, 70.0473565, 77.8136669, 84.9272654, 91.1869954, 96.3917, 101.4355, 104.58385, 105.197482, 102.63713, 95.6793504, 84.50272, 68.7016409, 47.8705152, 17.3576345, -10.504268, -36.2030578, -39.5162004, -13.1099957, 81.1941944, 273.863774, 595.366146},
        {177.272638, 176.171898, 172.67048, 167.127495, 159.902055, 151.353269, 141.840249, 131.722107, 121.357952, 111.932907, 102.649669, 93.5369434, 84.6234388, 76.1690857, 67.8788777, 59.6890322, 51.5357667, 42.9013283, 34.3574928, 26.0220658, 18.0128529, 10.8353733, 4.06463329, -2.336647, -8.40574756, -14.5091726, -20.223288, -25.4536843, -30.1059515, -34.3111544, -37.6592191, -39.965546, -41.0455356, -40.0655795, -37.7496906, -34.1728732, -29.4101313, -23.5468726, -16.6435361, -8.77096437, 0., 10.1187145, 20.7860564, 31.7231031, 42.6509319, 52.6140477, 62.2807294, 71.6426832, 80.6916156, 89.0636486, 97.2483064, 105.379529, 113.591258, 122.654121, 131.810694, 140.94024, 149.922024, 154.37936, 160.149841, 168.815111, 181.956813, 197.339586, 227.99609, 274.768336, 359.778501, 504.715443, 711.721124, 987.824231, 1340.05345}
    };
    
    /* Stuff for interpolating the NQC data */
    gsl_spline    *spline = NULL;
    gsl_interp_accel *acc = NULL;
    /* Interpolate the spin NQC data in 2-D parameter space -- spin and mass ratio */
    /* First, interpolate in spin dimension for all mass ratios */
    spline = gsl_spline_alloc( gsl_interp_cspline, adim );
    acc    = gsl_interp_accel_alloc();
    for (i = 0; i < qdim; i++)
    {
        gsl_spline_init( spline, alist, a3stab[i], adim );
        gsl_interp_accel_reset( acc );
        a3slist[i] = gsl_spline_eval( spline, a/(1.0-2.0*eta), acc );
        gsl_spline_init( spline, alist, a4tab[i], adim );
        gsl_interp_accel_reset( acc );
        a4list[i] = gsl_spline_eval( spline, a/(1.0-2.0*eta), acc );
        gsl_spline_init( spline, alist, a5tab[i], adim );
        gsl_interp_accel_reset( acc );
        a5list[i] = gsl_spline_eval( spline, a/(1.0-2.0*eta), acc );
        //printf("%.15f\n",a3slist[i]);
    }
    //printf("%.15f\n",a);
    gsl_spline_free(spline);
    gsl_interp_accel_free(acc);
    /* Second, interpolate in mass ratio dimension */
    spline = gsl_spline_alloc( gsl_interp_cspline, qdim );
    acc    = gsl_interp_accel_alloc();
    gsl_spline_init( spline, etalist, a3slist, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->a3S = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, a4list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->a4 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, a5list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->a5 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_free(spline);
    gsl_interp_accel_free(acc);
    
    /* Interpolate nonspin NQC data in the mass ratio dimension */
    spline = gsl_spline_alloc( gsl_interp_cspline, nsqdim );
    acc    = gsl_interp_accel_alloc();
    gsl_spline_init( spline, nsetalist, a1list, nsqdim );
    gsl_interp_accel_reset( acc );
    coeffs->a1 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, nsetalist, a2list, nsqdim );
    gsl_interp_accel_reset( acc );
    coeffs->a2 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, nsetalist, a3list, nsqdim );
    gsl_interp_accel_reset( acc );
    coeffs->a3 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, nsetalist, b1list, nsqdim );
    gsl_interp_accel_reset( acc );
    coeffs->b1 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, nsetalist, b2list, nsqdim );
    gsl_interp_accel_reset( acc );
    coeffs->b2 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_free(spline);
    gsl_interp_accel_free(acc);
    /* Andrea and I have different sign conventions, so I need to put a minus sign in front */
    coeffs->b1 = - coeffs->b1;
    coeffs->b2 = - coeffs->b2;
#if 0
    coeffs->a1 = -8.02798637014;
    coeffs->a2 = 48.7446843797;
    coeffs->a3 = -45.7900277224;
    coeffs->a3S= 0.;
    coeffs->a4 = 0.;
    coeffs->a5 = 0.;
    coeffs->b1 = 0.834742923041;
    coeffs->b2 = -2.33512320852; // q=1
#endif
#if 0
    coeffs->a1 = -7.79667;
    coeffs->a2 = 47.182;
    coeffs->a3 = 2238.85334023;
    coeffs->a3S= 0.;
    coeffs->a4 = -7143.16738899;
    coeffs->a5 = 5596.0086893;
    coeffs->b1 = 0.85069;
    coeffs->b2 = -2.47071; // q=1, chi1=chi2=0.98
#endif
#if 0
    coeffs->a1 = -6.82562365707;
    coeffs->a2 = 41.5711482044;
    coeffs->a3 = -39.4329799959;
    coeffs->a3S= 0.;
    coeffs->a4 = 0.;
    coeffs->a5 = 0.;
    coeffs->b1 = 0.461925688819;
    coeffs->b2 = -1.38733263299; // q=8
#endif
#if 0
    coeffs->a1 = -7.5758;
    coeffs->a2 = 46.9323;
    coeffs->a3 = -118.368375152;
    //coeffs->a3 = -45.0036; // NS part
    coeffs->a3S= 0.;
    coeffs->a4 = 125.555824111;
    coeffs->a5 = -22.0751068073;
    //coeffs->a4 = 0.;
    //coeffs->a5 = 0.;
    coeffs->b1 = 0.51305;
    coeffs->b2 = -1.55133; // q=8, chi1=0.5
#endif
    
    /* Obsolete polynomial fitting of nonspin NQC coefficients a1, a2, a3, b1 and b2 */
    /*
     coeffs->a1 = -12.67955358602124 + 75.41927959573084 * eta - 106.15933052937714 * eta2;
     coeffs->a2 = 101.45522216901628 - 757.3158549733314 * eta + 1473.314771676588 * eta2;
     coeffs->a3 = -107.6647834845902 + 857.6219519536213 * eta - 1776.2776804623143 * eta2;
     // Andrea and I have different sign conventions, so I need to put a minus sign in front
     coeffs->b1 = - (-1.464129495621165 + 12.81732978488213 * eta - 60.09957767247623 * eta2);
     coeffs->b2 = - ( 7.477426352542122 - 85.26122117590637 * eta + 353.3251639728075 * eta2);
     */
    
    return XLAL_SUCCESS;
    
}

/**
 * Function to interpolate known amplitude NQC coeffcients of spin terms for SEOBNRv2,
 * namely a3s, a4 and a5.
 * The a3s, a4 and a5 values were calculated for
 * 11 mass ratios q=1,1.5,2,4,6,8,10,11,12,14 and 20, and
 * 19 spin (\f$\chi\f$ defined in Taracchini et al. PRD 86, 024011 (2012)) values
 * chi = -1, -0.9, -0.8, ......, 0.3, 0.4, 0.5, 0.55, 0.6, 0.65.
 * The calculation was done by Andrea Taracchini using a C++ code of the UMaryland group.
 * In principle, these numbers can be automatically calculated iteratively by the LAL code.
 * However, since such calcualtion increase the cost of each waveform generation by
 * about an order of magnitude, we prepare these numbers in advance reduce cost.
 * These number can be verified by confirming that
 * the peak amplitude and frequency agree well with the NR-fits predicted values,
 * and to get exact NR-fits predicted values, corrections on these numbers are ~1%.
 */
UNUSED static int XLALSimIMRGetEOBCalibratedSpinNQCv2( EOBNonQCCoeffs * restrict coeffs,                                   INT4 UNUSED l,                                   INT4 UNUSED m,                                   REAL8 eta,                                   REAL8 a )
{
    const unsigned int nsqdim = 101;
    const unsigned int   qdim = 21;
    const unsigned int   adim = 81;
    UINT4 i;
    /* REAL8 eta2 = eta*eta;*/
    REAL8 a3slist[qdim], a4list[qdim], a5list[qdim];
    REAL8 b3list[qdim], b4list[qdim];
    
    memset( coeffs, 0, sizeof( *coeffs ) );
    const double nsetalist[101] = {0.021, 0.0233, 0.0256, 0.0279, 0.0302, 0.0324, 0.0347, 0.037, 0.0393, 0.0416, 0.0439, 0.0462, 0.0485, 0.0508, 0.0531, 0.0553, 0.0576, 0.0599, 0.0622, 0.0645, 0.0668, 0.0691, 0.0714, 0.0737, 0.076, 0.0782, 0.0805, 0.0828, 0.0851, 0.0874, 0.0897, 0.092, 0.0943, 0.0966, 0.0989, 0.1012, 0.1034, 0.1057, 0.108, 0.1103, 0.1126, 0.1149, 0.1172, 0.1195, 0.1218, 0.124, 0.1263, 0.1286, 0.1309, 0.1332, 0.1355, 0.1378, 0.1401, 0.1424, 0.1447, 0.147, 0.1492, 0.1515, 0.1538, 0.1561, 0.1584, 0.1607, 0.163, 0.1653, 0.1676, 0.1698, 0.1721, 0.1744, 0.1767, 0.179, 0.1813, 0.1836, 0.1859, 0.1882, 0.1905, 0.1927, 0.195, 0.1973, 0.1996, 0.2019, 0.2042, 0.2065, 0.2088, 0.2111, 0.2134, 0.2156, 0.2179, 0.2202, 0.2225, 0.2248, 0.2271, 0.2294, 0.2317, 0.234, 0.2363, 0.2385, 0.2408, 0.2431, 0.2454, 0.2477, 0.25};
    const double etalist[21] = {0.05, 0.06, 0.07, 0.08, 0.09, 0.1, 0.11, 0.12, 0.13, 0.14, 0.15, 0.16, 0.17, 0.18, 0.19, 0.2, 0.21, 0.22, 0.23, 0.24, 0.25};
    const double alist[81]   = {-1., -0.975, -0.95, -0.925, -0.9, -0.875, -0.85, -0.825, -0.8, -0.775, -0.75, -0.725, -0.7, -0.675, -0.65, -0.625, -0.6, -0.575, -0.55, -0.525, -0.5, -0.475, -0.45, -0.425, -0.4, -0.375, -0.35, -0.325, -0.3, -0.275, -0.25, -0.225, -0.2, -0.175, -0.15, -0.125, -0.1, -0.075, -0.05, -0.025, 0., 0.025, 0.05, 0.075, 0.1, 0.125, 0.15, 0.175, 0.2, 0.225, 0.25, 0.275, 0.3, 0.325, 0.35, 0.375, 0.4, 0.425, 0.45, 0.475, 0.5, 0.525, 0.55, 0.575, 0.6, 0.625, 0.65, 0.675, 0.7, 0.725, 0.75, 0.775, 0.8, 0.825, 0.85, 0.875, 0.9, 0.925, 0.95, 0.975, 1.};
    
    const double a1list[101] = {-9.70365622284, -9.5487796595, -9.40705218072, -9.28242447497, -9.16287060127, -9.03811669547, -8.92394749379, -8.81661027414, -8.7095851328, -8.60723315388, -8.50836995352, -8.40985391052, -8.314500912, -8.22035943611, -8.12770592382, -8.03757606119, -7.94997595927, -7.86421160183, -7.77953168332, -7.69584399308, -7.61324344682, -7.53192249536, -7.45216063215, -7.37374612092, -7.29648777301, -7.22046897906, -7.14581213839, -7.07265232481, -7.00073270258, -6.92991647363, -6.85992514074, -6.79104613524, -6.72328769905, -6.65664446919, -6.59158891131, -6.52783781218, -6.4667874811, -6.41088676175, -6.35876275395, -6.3073217784, -6.25613493225, -6.20587481079, -6.15851113649, -6.1141742151, -6.07327317284, -6.03621329966, -6.0001893395, -5.96655531812, -5.93844433597, -5.91027864135, -5.88037825648, -5.85358373242, -5.83336708498, -5.82744285781, -5.82818819089, -5.8231589417, -5.81812839946, -5.81887431779, -5.82568678472, -5.83621354379, -5.84931204972, -5.86442954488, -5.88109070102, -5.89755497618, -5.91392008194, -5.93177097875, -5.95199881766, -5.97697408745, -6.00401909417, -6.032531535, -6.06148086596, -6.09023163793, -6.11920032326, -6.14820900838, -6.17641365359, -6.20357599269, -6.23098735047, -6.25903214508, -6.28832401322, -6.32033354968, -6.35344345473, -6.38745961465, -6.42231976572, -6.46800466293, -6.50219865357, -6.52990506622, -6.54965342702, -6.58271099648, -6.626558215, -6.67664802777, -6.72810147477, -6.77989789496, -6.83928403282, -6.91446932169, -6.99994425628, -7.08201636107, -7.21472596367, -7.36827339046, -7.55475356656, -7.81665470799, -8.24936799702};
    
    const double a2list[101] = {71.8355837876, 70.2888163877, 68.858392382, 67.578368761, 66.3483263783, 65.0816770892, 63.9139561883, 62.8042133759, 61.7074812267, 60.6519669001, 59.6324792699, 58.6222094978, 57.6434687679, 56.6791933949, 55.7311388293, 54.8071777364, 53.9062860992, 53.0218701557, 52.1495389357, 51.2895492368, 50.4445863927, 49.6138840037, 48.797074544, 47.9943062521, 47.2056828988, 46.4310888273, 45.6708900349, 44.9258152616, 44.1929201091, 43.4702452949, 42.7537119379, 42.0477065689, 41.3531483593, 40.6709264131, 40.0029858455, 39.3453826346, 38.7105183605, 38.119829408, 37.5632095631, 37.0166865581, 36.4716876499, 35.9345595307, 35.4276417334, 34.9454114602, 34.4897359229, 34.0666726398, 33.6585411373, 33.268152249, 32.9062444867, 32.5494651377, 32.1932422911, 31.8674681969, 31.598287958, 31.4450951286, 31.3504698588, 31.2224679676, 31.1007751184, 31.0231242897, 30.991643584, 30.991534397, 31.0161048282, 31.0616750366, 31.1249606145, 31.1939358163, 31.2696363303, 31.3636652679, 31.4828878696, 31.6456708233, 31.8311325739, 32.034527166, 32.2477681231, 32.4658509298, 32.6919766483, 32.9247097971, 33.1580140019, 33.3900369294, 33.6294910387, 33.8788248397, 34.1421252583, 34.4296046145, 34.7294034713, 35.0398350888, 35.3602204133, 35.7665471711, 36.0881399675, 36.3632831298, 36.5789495976, 36.9023316975, 37.3120873104, 37.7708043813, 38.2382659746, 38.7083475975, 39.2358324951, 39.8798584792, 40.5947669891, 41.2766466634, 42.3250573144, 43.5282809602, 44.9693438982, 46.9606371658, 50.0808316558};
    
    const double a3list[101] = {-75.3870348213, -73.5563054743, -71.8586610238, -70.3326010764, -68.8669937325, -67.3652081804, -65.9797869993, -64.6591310991, -63.3595036445, -62.1068621881, -60.8985027736, -59.7042086628, -58.5479728539, -57.4103808438, -56.2930281651, -55.2041741738, -54.1420295685, -53.0987898958, -52.0706872299, -51.0586173069, -50.0667002046, -49.0926497207, -48.1344431455, -47.1933388413, -46.2704158358, -45.3650449971, -44.4772042283, -43.6074103996, -42.7520134632, -41.9084025308, -41.0711187682, -40.2460776811, -39.4347613879, -38.6386667672, -37.8587722914, -37.0898368809, -36.3458764907, -35.6510020946, -34.9947343733, -34.3515708492, -33.7091654129, -33.0750000728, -32.4772646605, -31.9060849394, -31.3624600752, -30.8542478078, -30.364610911, -29.8930785724, -29.4449207659, -29.0043606667, -28.5703691139, -28.171109369, -27.8341554058, -27.6229720264, -27.4766951091, -27.2985750163, -27.1291617273, -27.0061494789, -26.9317625784, -26.8914625991, -26.8788122164, -26.8902006455, -26.9223881792, -26.9631792328, -27.013848345, -27.0865681299, -27.1887211003, -27.3400465839, -27.5181167553, -27.7178394005, -27.9305187037, -28.1507835284, -28.3821145422, -28.6229907203, -28.8672132002, -29.1128371338, -29.36865915, -29.6371071707, -29.922311561, -30.2347674515, -30.5620356831, -30.9022473284, -31.254609498, -31.7001465622, -32.0565827322, -32.3648712883, -32.6103616483, -32.9736989432, -33.4312219324, -33.9418551023, -34.4614297518, -34.9838283376, -35.5674613231, -36.2737268484, -37.0519888454, -37.7923554764, -38.9140097, -40.1981922876, -41.7294871636, -43.8332602049, -47.0600574352};
    
    const double b1list[101] = {-0.454231789035, -0.453740895324, -0.45114951071, -0.445455559729, -0.44030174389, -0.438246415293, -0.435617386371, -0.430628416274, -0.428155263233, -0.424563996869, -0.421496742404, -0.419542314684, -0.417786543725, -0.416452769889, -0.415181542102, -0.413649756645, -0.411562161412, -0.408719678496, -0.40593475615, -0.403713197659, -0.402947057134, -0.402692696086, -0.401882254635, -0.401361543256, -0.401857592043, -0.403135870926, -0.405023096371, -0.407445456424, -0.410163626113, -0.412943367674, -0.415232607237, -0.417748668538, -0.420844132902, -0.424905043844, -0.429142196414, -0.432959586521, -0.436617400517, -0.440401062979, -0.445048680123, -0.451651705843, -0.457760235328, -0.463733230566, -0.472390233324, -0.480892678638, -0.488468389259, -0.496054924285, -0.505287254412, -0.516712248163, -0.530702320051, -0.543402030632, -0.553376187444, -0.564523083803, -0.576716924164, -0.588673429838, -0.599807622576, -0.6083625871, -0.61629220663, -0.626333930393, -0.638492384934, -0.651370663496, -0.663578126675, -0.675399039036, -0.687337552218, -0.699811934302, -0.712809946825, -0.725971047683, -0.73913751946, -0.752013364118, -0.764774460887, -0.777402692471, -0.789886995534, -0.802209798764, -0.814336203683, -0.826241102491, -0.837954892052, -0.849457493421, -0.860601375162, -0.871314573433, -0.881462999336, -0.890680302818, -0.899437045192, -0.907818061982, -0.915897388877, -0.92305068548, -0.931064125926, -0.939248450894, -0.947470287824, -0.953643840407, -0.957725818058, -0.959847025364, -0.960115447844, -0.959051577361, -0.957073291984, -0.952866371063, -0.945294863841, -0.937774627549, -0.921959850725, -0.901024675728, -0.87472187553, -0.834748962688, -0.733155709571};
    
    const double b2list[101] = {1.24250927283, 1.23820882351, 1.22814744033, 1.20899501569, 1.19187482987, 1.18459068292, 1.17544439367, 1.16014645984, 1.15217431136, 1.14122006693, 1.13199616785, 1.12640121441, 1.121556275, 1.11826795578, 1.11552355075, 1.11238433978, 1.10807035787, 1.10209548232, 1.09668053255, 1.09318287767, 1.09397420715, 1.09653673342, 1.09802015935, 1.10068222122, 1.10643573897, 1.11470254998, 1.12510511039, 1.13755406941, 1.15120517214, 1.16532741744, 1.17824955526, 1.19205557182, 1.20770493667, 1.22624497606, 1.24563734056, 1.26435121321, 1.28310293005, 1.30265970794, 1.32505082997, 1.35327207812, 1.38087484583, 1.40864740792, 1.44367567593, 1.47847338398, 1.51107248182, 1.54412932876, 1.58210883511, 1.62865620041, 1.68826303807, 1.74233039828, 1.78373725125, 1.82868267503, 1.87681370736, 1.92312517747, 1.9656186804, 1.99785596132, 2.02740592217, 2.06495795508, 2.11055684374, 2.15878076242, 2.20440433139, 2.24836632819, 2.29241478185, 2.33795222667, 2.38484419984, 2.43171272272, 2.47786827853, 2.52194361567, 2.56488422632, 2.60668479559, 2.6474159293, 2.6870781889, 2.72548564593, 2.76255774767, 2.79845498187, 2.83311863463, 2.86593650322, 2.89662455284, 2.9246508029, 2.94855630091, 2.97022227107, 2.98999047671, 3.00814380105, 3.02093394891, 3.03817745817, 3.05662290422, 3.07635559396, 3.08574133623, 3.0854061584, 3.07702110074, 3.06226203175, 3.04321477098, 3.01960302068, 2.98621442345, 2.94157099234, 2.89879486315, 2.82337825218, 2.72933462053, 2.61436316871, 2.4455062038, 2.1171325295};
    
    const double a3stab[21][81] = {
        {194.034062, 193.315403, 192.340696, 191.108131, 189.615898, 187.862187, 185.845189, 183.563092, 181.014087, 178.196365, 175.108115, 171.747527, 168.112791, 164.202098, 159.88123, 155.232138, 150.346797, 145.250479, 139.968456, 134.526002, 128.948389, 123.451375, 117.843441, 112.123552, 106.290673, 100.343771, 94.2818106, 88.0767764, 81.7047066, 75.2290824, 68.6598248, 62.0068545, 55.2800926, 48.4894598, 41.7952745, 35.1123156, 28.3439776, 21.4694187, 14.4677972, 7.31827158, 0, -7.06413602, -14.3999825, -22.0896625, -30.2152987, -38.8590141, -48.1029316, -57.9430959, -68.3884858, -79.7270473, -92.0758537, -105.551978, -120.272495, -136.354477, -161.225662, -190.378371, -218.751539, -244.966902, -267.646199, -285.411167, -296.883542, -310.354742, -328.37036, -319.492152, -252.281869, 29.4084482, 249.400555, 588.783155, 1228.64495, 2363.34086, 4134.16095, 7287.44826, 10172.0754, 11510.5972, 12299.3266, 12652.4652, 12433.2947, 11460.0626, 9982.07373, 8101.6811, 5921.23755},
        {160.112548, 160.667985, 160.855096, 160.683878, 160.164332, 159.306454, 158.120244, 156.615701, 154.802822, 152.691607, 150.292055, 147.614163, 144.66793, 141.463355, 137.974809, 134.234829, 130.278648, 126.123553, 121.786828, 117.285761, 112.637636, 107.962912, 103.161454, 98.2362991, 93.1904845, 88.0270475, 82.7490251, 77.3472032, 71.8142083, 66.1823723, 60.4597069, 54.6542236, 48.7739343, 42.8268505, 36.9526253, 31.0759936, 25.1117255, 19.0409065, 12.8446223, 6.50395828, 0, -6.31861818, -12.8891206, -19.7811827, -27.06448, -34.8086881, -43.0834823, -52.0954881, -62.0307527, -72.6314885, -83.9117648, -95.8856513, -108.567217, -121.970532, -144.794469, -171.559059, -196.126931, -216.73575, -231.62318, -239.026886, -237.184533, -230.598764, -228.655499, -197.207571, -102.107815, 154.463951, 449.595197, 893.840096, 1597.75282, 2688.58018, 4226.79843, 6614.56539, 8915.92748, 10410.9574, 11588.5513, 12442.7012, 12813.2517, 12518.0513, 11617.0807, 10087.775, 7907.56954},
        {135.75872, 136.138893, 136.299307, 136.238612, 135.955459, 135.448498, 134.71638, 133.757755, 132.571273, 131.155584, 129.50934, 127.63119, 125.519785, 123.173775, 120.539911, 117.649675, 114.538477, 111.215586, 107.690265, 103.971782, 100.069403, 95.9633653, 91.6959723, 87.2804996, 82.7302224, 78.058416, 73.2783556, 68.4234731, 63.5241716, 58.5455294, 53.4926375, 48.3705869, 43.1844686, 37.9393735, 32.7564707, 27.5674199, 22.2950938, 16.9208406, 11.4260085, 5.79194558, 0, -5.65721456, -11.5526034, -17.7478063, -24.304463, -31.2842132, -38.7486967, -47.177716, -56.9882396, -67.2420324, -77.8309468, -88.646835, -99.5815494, -110.526942, -130.898858, -154.564236, -174.669094, -189.157228, -195.972438, -193.058522, -178.359277, -95.540229, -64.7497577, -36.3204737, 39.4150121, 182.864142, 515.868654, 1052.98428, 1808.76676, 2783.21579, 4034.5552, 5797.96238, 7575.84812, 8998.68735, 10279.6823, 11358.5028, 12072.546, 12243.6693, 11858.1609, 10846.2528, 9138.17686},
        {117.594822, 118.216324, 118.587481, 118.713031, 118.597712, 118.24626, 117.663414, 116.853912, 115.822491, 114.573888, 113.112843, 111.444091, 109.572371, 107.502421, 105.228449, 102.761854, 100.114831, 97.2942708, 94.3070652, 91.1601054, 87.8602826, 84.4097412, 80.820775, 77.1009309, 73.2577557, 69.2987964, 65.2315998, 61.0732898, 56.8395514, 52.5150315, 48.1033884, 43.6082803, 39.0333656, 34.3823025, 29.7493628, 25.0808824, 20.3163379, 15.4408534, 10.439553, 5.29756047, 0, -5.35071899, -10.9029559, -16.6877846, -22.736279, -29.0795131, -35.7485608, -43.2877245, -52.1641879, -61.1818369, -70.1633582, -78.9314385, -87.3087643, -95.1180224, -113.214171, -134.440841, -150.806571, -159.877486, -159.219713, -146.399379, -118.982611, -70.3022292, -22.0562145, 53.8038232, 185.326274, 404.156778, 729.470528, 1193.63258, 1829.00799, 2680.16117, 3742.85913, 5113.35478, 6506.86997, 7696.20069, 8895.67902, 10040.8571, 10932.5801, 11339.6312, 11293.7482, 10737.025, 9611.55528},
        {104.536193, 105.151992, 105.52303, 105.656338, 105.558946, 105.237884, 104.700185, 103.952877, 103.002992, 101.85756, 100.523612, 99.0081792, 97.3182909, 95.4609783, 93.4570201, 91.3047497, 89.00146, 86.5513695, 83.9586968, 81.2276605, 78.362479, 75.3512364, 72.216514, 68.9647586, 65.6024172, 62.1359365, 58.5717633, 54.9331837, 51.2409534, 47.4612548, 43.5936975, 39.637891, 35.593445, 31.4599689, 27.2943493, 23.0599623, 18.715848, 14.2499006, 9.65001414, 4.90408258, 0, -5.02324517, -10.2359098, -15.6571562, -21.306147, -27.2020446, -33.3640114, -40.5534775, -49.4203385, -58.20891, -66.6369699, -74.4222961, -81.2826666, -86.9358592, -101.298089, -117.635564, -128.44252, -131.350726, -123.991956, -103.997981, -69.0005725, -33.6912932, 21.7514718, 116.165345, 268.387948, 496.490784, 821.201233, 1260.43758, 1832.11811, 2569.02123, 3444.48487, 4457.74625, 5529.91911, 6593.29966, 7712.26975, 8816.62971, 9754.9376, 10360.7413, 10608.8717, 10425.9949, 9738.77691},
        {96.9661604, 97.1499366, 97.1688322, 97.0244892, 96.7185494, 96.2526546, 95.6284466, 94.8475673, 93.9116586, 92.8223622, 91.58132, 90.190174, 88.6505658, 86.9641373, 85.1097322, 83.1034146, 80.9629742, 78.6947161, 76.3049452, 73.7999668, 71.1860856, 68.4994692, 65.712436, 62.8271668, 59.8458427, 56.7706443, 53.6037525, 50.3545028, 47.0311555, 43.6187841, 40.1166645, 36.5240723, 32.8402837, 29.0645742, 25.2349668, 21.3262257, 17.3101827, 13.178188, 8.92159236, 4.53174618, 0, -4.6801893, -9.51986849, -14.5279779, -19.713458, -25.0852492, -30.6522918, -37.2783443, -45.6987185, -53.8783872, -61.4792071, -68.1630351, -73.591728, -77.4271424, -90.8801985, -106.306773, -115.188531, -114.825063, -102.515961, -75.5608158, -31.2592188, 33.1149764, 110.137717, 217.631759, 373.41986, 572.166076, 888.81781, 1313.4076, 1835.96799, 2443.46734, 3135.13071, 3899.77877, 4762.59709, 5753.91388, 6767.53656, 7738.27797, 8606.45078, 9341.27386, 9790.84147, 9844.54054, 9391.75801},
        {91.9984129, 92.0724899, 91.9723426, 91.7032712, 91.2705759, 90.6795572, 89.9355152, 89.0437503, 88.0095627, 86.8382527, 85.5351205, 84.1054666, 82.554591, 80.8877943, 79.1289793, 77.2716783, 75.3080151, 73.2394849, 71.067583, 68.7938044, 66.4196446, 63.9380846, 61.3603095, 58.6889905, 55.9267986, 53.0764049, 50.1404805, 47.134602, 44.0724069, 40.9257076, 37.6919349, 34.3685201, 30.9528939, 27.4424876, 23.8547107, 20.1743557, 16.3847026, 12.479096, 8.45088037, 4.29340019, 0, -4.52502743, -9.18698719, -13.9802361, -18.899131, -23.9380288, -29.0912862, -35.37357, -43.6462329, -51.4639546, -58.4068133, -64.0548872, -67.9882546, -69.7869938, -79.8850741, -90.9963626, -95.0131348, -89.2953953, -71.2031485, -38.096399, 12.6648486, 60.4616544, 143.801577, 269.670606, 445.05473, 673.029945, 970.22686, 1338.93951, 1781.46194, 2307.71228, 2897.11226, 3505.21395, 4196.11738, 5020.27394, 5911.38909, 6815.19861, 7663.7143, 8405.30481, 8936.63545, 9171.99775, 9025.68322},
        {88.7697502, 88.7243429, 88.5229502, 88.1697385, 87.668874, 87.0245232, 86.2408524, 85.322028, 84.2722164, 83.0955839, 81.796297, 80.378522, 78.8464252, 77.2041731, 75.4608637, 73.6175437, 71.6748862, 69.6360488, 67.504189, 65.2824646, 62.9740331, 60.5905035, 58.1254149, 55.5807576, 52.9585221, 50.2606987, 47.4892778, 44.6635083, 41.8000455, 38.8596132, 35.8371944, 32.7277721, 29.5263291, 26.2278486, 22.8396849, 19.3489948, 15.7419984, 12.0111483, 8.14889689, 4.14769663, 0, -4.57373064, -9.27148902, -14.063259, -18.9190244, -23.8087691, -28.7024769, -34.6285481, -42.4563432, -49.6250515, -55.6749058, -60.1461388, -62.5789832, -62.5136718, -69.857389, -77.5921962, -77.915422, -68.2468249, -46.0061634, -8.61319596, 46.512319, 98.1593023, 187.472996, 318.664443, 495.944685, 718.350267, 1002.85596, 1347.46346, 1750.17445, 2217.72359, 2721.91372, 3194.0735, 3742.39868, 4446.91421, 5256.03088, 6117.02483, 6949.94367, 7676.88355, 8238.97487, 8570.85908, 8607.17759},
        {86.991682, 86.9009299, 86.6584244, 86.2679651, 85.7333516, 85.0583835, 84.2468603, 83.3025818, 82.2293474, 81.0309567, 79.7112095, 78.2739051, 76.7228433, 75.0618236, 73.2860615, 71.4047868, 69.4278794, 67.3608948, 65.2093882, 62.9789152, 60.6750311, 58.3371277, 55.932251, 53.4612835, 50.9251075, 48.3246055, 45.6606599, 42.952423, 40.2163017, 37.4094932, 34.5254615, 31.5576709, 28.4995856, 25.3446698, 22.0706172, 18.6808685, 15.1800581, 11.5648762, 7.83201255, 3.97815717, 0, -4.44925282, -8.98199005, -13.5540843, -18.121408, -22.639834, -27.0652346, -32.4829047, -39.8084298, -46.2971051, -51.4462215, -54.7530699, -55.7149412, -53.8291262, -58.6640354, -63.3459328, -60.2387428, -46.7797921, -20.4064072, 21.4440851, 81.3343584, 146.856546, 244.47561, 378.327141, 552.546728, 759.30391, 1032.25046, 1361.16283, 1735.81745, 2148.15637, 2581.45923, 2984.80058, 3447.78767, 4035.66729, 4736.79845, 5514.39717, 6296.19014, 7018.13887, 7618.75482, 8035.68015, 8206.55699},
        {87.4052431, 87.2488223, 86.9362164, 86.4718191, 85.8600238, 85.1052242, 84.2118138, 83.184186, 82.0267346, 80.7438529, 79.3399346, 77.8193732, 76.1865622, 74.4458952, 72.5913588, 70.6339301, 68.5853738, 66.4522119, 64.2409669, 61.9581608, 59.6103158, 57.2444895, 54.8210706, 52.3409829, 49.8051505, 47.2144972, 44.5699471, 41.8924106, 39.199795, 36.4452343, 33.6215372, 30.7215125, 27.737969, 24.6637155, 21.4396148, 18.0913369, 14.648484, 11.1144899, 7.49278858, 3.78681398, 0, -4.20404561, -8.43513042, -12.6428882, -16.7769527, -20.7869576, -24.6225368, -29.4472857, -36.2423878, -42.0547553, -46.3411145, -48.5581916, -48.162713, -44.6114048, -47.2420788, -49.2606505, -43.1250808, -26.2710014, 3.86595568, 49.8501587, 114.245976, 188.587294, 293.492819, 432.021615, 607.232747, 806.962399, 1071.81931, 1386.59523, 1736.08191, 2100.7761, 2478.35461, 2845.05278, 3255.50918, 3749.4546, 4345.50524, 5025.01448, 5735.30498, 6443.8395, 7071.55114, 7555.23546, 7831.68799},
        {90.6686478, 90.338945, 89.8498427, 89.2065534, 88.4142895, 87.4782634, 86.4036875, 85.1957743, 83.8597362, 82.4007857, 80.8241351, 79.1349969, 77.3385836, 75.4401075, 73.4416099, 71.3503094, 69.1736645, 66.9175363, 64.5877858, 62.1902741, 59.7308624, 57.2362237, 54.6885327, 52.0907763, 49.445941, 46.7570139, 44.0269815, 41.2851226, 38.5567652, 35.7820291, 32.9532257, 30.0626665, 27.1026627, 24.0655259, 20.8727666, 17.5614852, 14.1741411, 10.7175274, 7.19843724, 3.62366374, 0, -4.06090718, -8.10254597, -12.0635508, -15.8825563, -19.4981968, -22.8491068, -27.132479, -33.3563927, -38.4501259, -41.8412979, -42.9575283, -41.2264364, -36.0756416, -36.4842853, -35.8376593, -26.798047, -6.83938301, 26.5643983, 75.9393624, 143.811575, 212.08357, 319.941909, 466.480743, 650.794218, 862.613943, 1124.12827, 1423.1964, 1747.67751, 2084.00816, 2424.31545, 2731.24107, 3082.92956, 3537.96935, 4077.76369, 4684.63584, 5326.20908, 5987.9228, 6598.65193, 7106.5548, 7459.78976},
        {95.4194996, 94.9744199, 94.3458563, 93.5410786, 92.5673567, 91.4319604, 90.1421596, 88.7052241, 87.1284237, 85.4190283, 83.5843078, 81.6315319, 79.5679705, 77.4008935, 75.1326394, 72.7735975, 70.3345307, 67.8237172, 65.2494357, 62.6199646, 59.9435824, 57.2613442, 54.5442252, 51.7959772, 49.020352, 46.2211015, 43.4019773, 40.5995521, 37.8454667, 35.0649948, 32.2485619, 29.3865936, 26.4695154, 23.4877529, 20.3632018, 17.1296398, 13.8260322, 10.4568216, 7.02645092, 3.53936279, 0, -4.2462037, -8.44478288, -12.5002812, -16.3172421, -19.8002093, -22.8537263, -26.6855879, -32.3077596, -36.5085628, -38.6633793, -38.1475908, -34.3365791, -26.605726, -22.9025218, -17.179601, -2.74072752, 22.792071, 61.7967668, 116.651332, 189.733739, 263.596099, 376.57103, 526.669187, 711.901228, 926.647159, 1177.87322, 1459.23331, 1764.38136, 2097.49779, 2420.65693, 2632.95065, 2903.44544, 3359.22954, 3869.45585, 4404.85502, 4965.83074, 5566.44504, 6136.45351, 6634.35268, 7018.6391},
        {101.846145, 101.147731, 100.262496, 99.1975513, 97.9600035, 96.5569615, 94.9955335, 93.2828279, 91.4259529, 89.432017, 87.3081284, 85.0613955, 82.6989265, 80.2278299, 77.6380801, 74.9476243, 72.1757068, 69.3329406, 66.4299384, 63.4773132, 60.485678, 57.4779833, 54.4508003, 51.4130383, 48.3736061, 45.3414129, 42.3253675, 39.3907795, 36.5944834, 33.8105285, 31.0249232, 28.2236762, 25.3927962, 22.5182917, 19.4346176, 16.2236559, 12.9787608, 9.7169399, 6.45520101, 3.21055181, 0, -3.68378724, -7.22704665, -10.5403555, -13.5342911, -16.1194306, -18.2063515, -21.0997824, -25.8949643, -29.1688988, -30.2660929, -28.5310535, -23.3082876, -13.9423019, -8.20238269, -0.103490751, 16.9771316, 45.4181963, 87.5984154, 145.896501, 222.691165, 302.797812, 421.032832, 575.026689, 762.409844, 975.885659, 1225.23808, 1502.1851, 1798.44471, 2114.12912, 2415.77363, 2618.3873, 2857.47376, 3230.62205, 3675.97438, 4169.32865, 4684.71809, 5189.26697, 5679.50007, 6141.91923, 6563.02631},
        {109.957867, 108.923771, 107.696307, 106.283107, 104.691802, 102.930024, 101.005405, 98.9255759, 96.6981698, 94.330818, 91.8311523, 89.2068045, 86.4654063, 83.6145897, 80.6454548, 77.5760913, 74.4258405, 71.2057155, 67.9267295, 64.5998959, 61.2362277, 57.7997181, 54.3548943, 50.9192633, 47.510332, 44.1456075, 40.8425968, 37.7092791, 34.8400388, 32.036054, 29.278097, 26.5469404, 23.8233566, 21.088118, 18.1070639, 14.9969343, 11.8907385, 8.81321199, 5.78908988, 2.84310746, 0, -3.28293647, -6.33516022, -9.05356875, -11.3350596, -13.0765301, -14.174878, -16.0101729, -19.7396202, -21.7136831, -21.2270435, -17.5743834, -10.0503846, 2.05027093, 12.7196716, 26.9099491, 50.0813972, 84.3124696, 131.68162, 194.267302, 274.14797, 350.227635, 466.768801, 620.509952, 808.189573, 1022.38726, 1270.10006, 1543.07583, 1833.06243, 2139.56356, 2431.05959, 2635.46524, 2856.25309, 3163.22021, 3542.38549, 3979.04626, 4441.16822, 4893.91506, 5330.99006, 5736.36707, 6094.01992},
        {119.074672, 117.759823, 116.217882, 114.459582, 112.495658, 110.336844, 107.993873, 105.477481, 102.7984, 99.9673644, 96.9951093, 93.8923682, 90.6698751, 87.3383642, 83.9063765, 80.3860334, 76.7896223, 73.1283257, 69.4133263, 65.6558067, 61.8669495, 57.9512264, 54.0412684, 50.1629958, 46.3423287, 42.6051873, 38.9774916, 35.5978285, 32.587855, 29.7040929, 26.9267158, 24.2358979, 21.6118129, 19.0346348, 16.2844216, 13.4679411, 10.7071094, 8.02303235, 5.43681598, 2.96956625, 0.642389138, -2.31001716, -4.9715306, -7.21243689, -8.90302176, -9.91357094, -10.1143701, -10.8729696, -13.3319374, -13.7814292, -11.4837931, -5.70137737, 4.30346968, 19.2683999, 34.8047905, 54.8931933, 83.9021962, 123.617988, 175.826758, 242.314694, 324.867985, 399.313533, 514.196004, 666.062089, 851.458481, 1063.45811, 1307.17626, 1574.99112, 1859.28093, 2159.87331, 2446.79825, 2659.00754, 2872.53587, 3134.43182, 3455.95783, 3831.93732, 4236.88674, 4657.31296, 5052.06814, 5389.4322, 5637.68509},
        {128.495326, 127.074107, 125.349104, 123.337657, 121.057102, 118.524777, 115.758021, 112.77417, 109.590562, 106.224536, 102.693428, 99.0145773, 95.2053203, 91.2829951, 87.2836963, 83.2128958, 79.074646, 74.8824479, 70.6498026, 66.3902114, 62.1171752, 57.720481, 53.3544299, 49.049609, 44.836605, 40.746005, 36.8083958, 33.1657559, 29.9433259, 26.9053427, 24.0371649, 21.3241507, 18.7516587, 16.305047, 13.8190975, 11.3744239, 9.06303794, 6.90109728, 4.90475952, 3.09018229, 1.4735232, -0.793048375, -2.71006388, -4.14204299, -4.95350539, -5.00897077, -4.17295882, -3.71878542, -4.70807799, -3.63676357, 0.202654687, 7.51767362, 19.0157901, 35.4045009, 51.4404702, 71.5957135, 100.792698, 140.956113, 194.010651, 261.881001, 346.491854, 436.641211, 558.397569, 710.286632, 890.834104, 1091.36933, 1328.16899, 1591.12322, 1870.12216, 2157.70833, 2435.81478, 2673.27241, 2895.86752, 3111.89571, 3385.10739, 3718.32051, 4076.70075, 4436.6436, 4767.2776, 5040.20866, 5227.04271},
        {138.900398, 137.337422, 135.391602, 133.087098, 130.44807, 127.498678, 124.263082, 120.765443, 117.02992, 113.080674, 108.941865, 104.637653, 100.192198, 95.6296595, 91.0129641, 86.3417484, 81.6127149, 76.8420946, 72.0461186, 67.2410178, 62.4430231, 57.528747, 52.6733214, 47.9122593, 43.2810742, 38.8152792, 34.5503876, 30.751528, 27.6493265, 24.73026, 21.9366103, 19.2106597, 16.49469, 13.7309833, 9.83315317, 5.39422106, 1.08507419, -2.94159973, -6.533113, -9.53677793, -11.7999068, -12.1167774, -11.5331679, -10.0418217, -7.63548268, -4.30689424, -0.0488000308, 3.44288832, 4.72818055, 7.88680239, 13.6175512, 22.6192243, 35.5906192, 53.2305331, 70.1560327, 90.9132406, 120.508639, 160.884993, 213.985069, 281.75163, 366.127442, 463.170325, 585.968126, 734.458352, 908.578505, 1100.01084, 1329.05576, 1585.74455, 1860.10848, 2144.0068, 2421.9869, 2675.82075, 2901.51559, 3083.49661, 3307.1074, 3583.65042, 3885.38171, 4198.36692, 4492.47968, 4743.11362, 4925.66238},
        {150.492141, 148.740377, 146.534433, 143.904424, 140.880463, 137.492664, 133.771141, 129.746008, 125.447379, 120.905368, 116.150089, 111.211655, 106.120181, 100.905781, 95.6494103, 90.3490205, 84.9987135, 79.6182038, 74.2272058, 68.8454341, 63.4926031, 58.0444393, 52.6845308, 47.4524779, 42.3878808, 37.5303397, 32.9194547, 28.9855324, 26.100171, 23.3687472, 20.672222, 17.8915564, 14.9077111, 11.601647, 5.59600934, -1.79962227, -9.10472355, -15.9764138, -22.0718122, -27.0480381, -30.5622108, -28.5501535, -24.904218, -19.7954601, -13.3949353, -5.87369959, 2.59719142, 9.71708841, 13.505339, 18.8829889, 26.5436657, 37.1809971, 51.4886108, 70.1601344, 88.6285408, 110.915367, 141.440359, 181.973167, 234.283443, 300.140838, 381.315003, 474.099497, 590.161596, 730.114744, 894.572384, 1077.6267, 1295.97687, 1542.41088, 1809.71674, 2095.08978, 2378.09595, 2639.604, 2862.82661, 3020.70131, 3195.47472, 3406.67752, 3645.53151, 3922.39098, 4199.38914, 4455.09829, 4668.09072},
        {161.612584, 159.667115, 157.206125, 154.264113, 150.875577, 147.075014, 142.896922, 138.3758, 133.546144, 128.442453, 123.099225, 117.550958, 111.832149, 105.977296, 100.052322, 94.0718457, 88.0481062, 82.0091739, 75.9831191, 69.9980119, 64.0819226, 58.2016273, 52.4549554, 46.878442, 41.5086225, 36.3820321, 31.5352061, 27.2761023, 23.8718937, 20.7101138, 17.7170917, 14.8191564, 11.942637, 9.01386256, 4.4868292, -0.780730804, -5.81963604, -10.4024039, -14.3015518, -17.289597, -19.139057, -17.5768668, -14.7036351, -10.5743881, -5.24415237, 1.23204592, 8.79918041, 15.7981135, 20.8107444, 27.6176602, 36.8151553, 48.9995239, 64.7670605, 84.7140596, 104.239416, 127.227319, 157.955375, 198.082965, 249.269468, 313.174262, 391.456729, 487.744948, 599.674335, 730.018462, 881.550901, 1048.86479, 1254.91206, 1492.64982, 1755.03523, 2037.67667, 2325.57748, 2619.73529, 2861.71536, 2991.38112, 3129.50183, 3302.22954, 3496.20653, 3710.8285, 3933.98728, 4157.49651, 4373.16983},
        {177.48227, 175.503869, 172.943216, 169.838793, 166.229083, 162.152567, 157.647727, 152.753046, 147.507005, 141.948087, 136.114773, 130.045546, 123.778887, 117.353279, 110.83074, 104.234863, 97.5874593, 90.9221961, 84.2727413, 77.6727629, 71.1559286, 64.7702854, 58.533136, 52.4761623, 46.6310463, 41.0294699, 35.7031149, 30.7482842, 26.2515706, 22.09014, 18.2694361, 14.7949029, 11.671984, 8.90612311, 6.4000957, 4.2242939, 2.45688145, 1.1243021, 0.252999572, -0.130582374, 0, 0.107665064, 0.86020865, 2.36190121, 4.71701321, 8.02981511, 12.4045774, 17.1481302, 21.687129, 28.0326146, 36.6126438, 47.8552734, 62.1885602, 80.0405611, 97.444392, 117.608365, 144.073471, 178.166715, 221.215102, 274.545638, 339.48533, 421.429031, 513.437617, 619.067665, 741.875754, 879.019633, 1049.83962, 1250.21376, 1476.0201, 1723.93173, 1987.44155, 2289.09571, 2536.59024, 2645.32769, 2757.98307, 2901.89402, 3057.58613, 3212.0205, 3373.22828, 3542.05763, 3719.35668},
        {203.880451, 202.208214, 199.864409, 196.892449, 193.33575, 189.237727, 184.641796, 179.59137, 174.129866, 168.300697, 162.14728, 155.71303, 149.041361, 142.175688, 135.235082, 128.215097, 121.105564, 113.934421, 106.72961, 99.5190709, 92.3307445, 85.1683129, 78.087325, 71.1190715, 64.294843, 57.6459303, 51.2036241, 45.0011085, 39.0712832, 33.4409113, 28.1405147, 23.2006152, 18.6517348, 14.5243953, 10.9819919, 7.97099029, 5.42779827, 3.35575953, 1.75821776, 0.638516679, 0, -0.204729396, 0.0854206845, 0.880801607, 2.19176474, 4.02866144, 6.40184309, 8.80951187, 10.8268265, 13.6887456, 17.6135694, 22.8195984, 29.525133, 37.9484735, 47.1793729, 58.1500548, 71.8781689, 88.8128507, 109.403236, 134.098459, 163.347656, 184.145996, 219.852907, 270.187002, 334.866894, 420.243061, 509.675555, 610.841165, 731.41668, 883.008518, 1061.50458, 1288.88822, 1489.49727, 1597.87824, 1694.42859, 1798.16764, 1908.18382, 2037.43503, 2163.3321, 2275.40424, 2363.18065}
    };
    
    const double a4tab[21][81] = {
        {-631.969628, -629.931559, -626.93986, -622.993704, -618.092259, -612.234699, -605.420193, -597.647913, -588.91703, -579.226715, -568.576139, -556.964473, -544.390888, -530.854555, -515.846825, -499.688117, -482.737292, -465.097392, -446.871459, -428.162534, -409.073659, -390.431317, -371.515197, -352.328428, -332.874139, -313.155461, -293.175523, -272.835853, -252.053245, -231.073772, -209.941815, -188.701758, -167.397985, -146.074879, -125.285446, -104.752312, -84.159603, -63.4476689, -42.5568579, -21.4275187, 0, 20.492378, 41.5808063, 63.5035042, 86.4986913, 110.804587, 136.659411, 164.137905, 193.339377, 224.892938, 259.103185, 296.274714, 336.71212, 380.720002, 451.856144, 535.71509, 616.131221, 688.652886, 748.828432, 792.206209, 814.334565, 847.454547, 888.473551, 839.16961, 601.32076, -334.087968, -1051.99176, -2143.56025, -4199.96305, -7855.60243, -13571.9501, -23786.9485, -33106.7677, -37360.6537, -39761.0933, -40691.7902, -39750.1263, -36377.5393, -31425.5653, -25265.023, -18266.7309},
        {-505.215416, -508.196276, -509.772603, -509.990387, -508.895619, -506.534288, -502.952385, -498.195899, -492.310822, -485.343142, -477.33885, -468.343937, -458.404391, -447.566204, -435.751584, -423.084826, -409.699594, -395.667196, -381.058941, -365.946138, -350.400093, -334.883419, -319.02208, -302.833344, -286.334476, -269.542745, -252.475417, -235.102086, -217.39951, -199.498947, -181.437021, -163.250356, -144.975574, -126.649301, -108.741298, -91.0141828, -73.197789, -55.240145, -37.0892796, -18.6932216, 0, 17.996917, 36.5393224, 55.8235696, 76.0460123, 97.4030037, 120.090898, 144.841731, 172.307048, 201.402322, 232.1064, 264.398133, 298.256369, 333.659957, 398.01185, 473.942308, 542.005708, 596.571523, 632.009229, 642.688299, 622.97821, 589.700061, 568.112915, 450.1228, 127.635743, -717.548639, -1675.26318, -3105.72753, -5369.16134, -8880.19603, -13835.8159, -21538.445, -28962.0227, -33785.1132, -37519.4657, -40137.7205, -41147.9007, -39965.1597, -36840.7475, -31740.8799, -24631.7726},
        {-416.815775, -419.057598, -420.469505, -421.053329, -420.810899, -419.744046, -417.854601, -415.144394, -411.615257, -407.26902, -402.107514, -396.13257, -389.346018, -381.749689, -373.154365, -363.682735, -353.471949, -342.562916, -330.996543, -318.813739, -306.055411, -292.666119, -278.796426, -264.500547, -249.832696, -234.847087, -219.597935, -204.212585, -188.80739, -173.261705, -157.600052, -141.846951, -126.026926, -110.164496, -94.647616, -79.2708986, -63.8014475, -48.1894474, -32.385083, -16.3385389, 0, 15.8033661, 32.1194751, 49.1192595, 66.9736521, 85.8535854, 105.929992, 128.799096, 155.842954, 183.824465, 212.335847, 240.969319, 269.317099, 296.971406, 353.48309, 419.49243, 473.371969, 508.586135, 518.599357, 496.876061, 436.880677, 151.559476, 37.8684007, -64.67, -316.533178, -780.861526, -1858.22995, -3592.31237, -6026.7827, -9156.64679, -13171.5827, -18819.4885, -24531.9546, -29153.6948, -33275.8654, -36682.7175, -38862.2881, -39230.2155, -37787.9462, -34337.2185, -28679.7708},
        {-352.810903, -355.83479, -357.93123, -359.12342, -359.43456, -358.887848, -357.506483, -355.313665, -352.332593, -348.586464, -344.098478, -338.891834, -332.989731, -326.415368, -319.156081, -311.257756, -302.768993, -293.720326, -284.14229, -274.065418, -263.520245, -252.52262, -241.11979, -229.344318, -217.228767, -204.805698, -192.107676, -179.198016, -166.135417, -152.8789, -139.44854, -125.864412, -112.14659, -98.3151495, -84.6689645, -71.0517401, -57.2860767, -43.3350226, -29.1616263, -14.728936, 0, 14.8074113, 30.0201505, 45.7103479, 61.9501343, 78.81164, 96.3669955, 116.389741, 140.39776, 164.394917, 187.762516, 209.881862, 230.134257, 247.901005, 297.350154, 355.856786, 398.16278, 416.534098, 403.236704, 350.53656, 250.699628, 81.0333436, -86.8683061, -342.335077, -774.696726, -1485.36227, -2533.86603, -4024.03276, -6059.68721, -8785.088, -12182.7583, -16539.1067, -20994.3126, -24867.3396, -28756.8221, -32428.176, -35236.2064, -36409.007, -36092.1278, -34118.3483, -30320.4477},
        {-308.350086, -311.217498, -313.195697, -314.31487, -314.605207, -314.096896, -312.820128, -310.805091, -308.081975, -304.680968, -300.632259, -295.966039, -290.712495, -284.901817, -278.614509, -271.848932, -264.599635, -256.886517, -248.729473, -240.148403, -231.163203, -221.738348, -211.956813, -201.846149, -191.433908, -180.747641, -169.8149, -158.71565, -147.52198, -136.136116, -124.564327, -112.812884, -100.888056, -88.7961136, -76.7068069, -64.522988, -52.135134, -39.5160767, -26.6386475, -13.4756781, 0, 13.7532862, 27.8822177, 42.4225625, 57.4100889, 72.8805651, 88.8697592, 107.833319, 131.863276, 155.20919, 176.924281, 196.061768, 211.674871, 222.816808, 260.477512, 303.506922, 328.337507, 327.490095, 293.485516, 218.844597, 96.0881681, -27.8598655, -216.405701, -529.145665, -1025.67608, -1762.42935, -2806.80613, -4214.60606, -6041.62879, -8393.47272, -11176.5412, -14348.9799, -17747.3316, -21230.7092, -24884.6691, -28450.6124, -31439.9222, -33295.7233, -33964.1906, -33220.0041, -30837.8436},
        {-284.293575, -285.455423, -286.02854, -286.023086, -285.449223, -284.317112, -282.636914, -280.418789, -277.6729, -274.409407, -270.638472, -266.370255, -261.614918, -256.382622, -250.60771, -244.348306, -237.668273, -230.593279, -223.148995, -215.361088, -207.255227, -198.95188, -190.368826, -181.518641, -172.413902, -163.067186, -153.49107, -143.717927, -133.777153, -123.633993, -113.292987, -102.758672, -92.03559, -81.128279, -70.1406469, -59.0143722, -47.6836118, -36.1325801, -24.3454913, -12.3065598, 0, 12.6663563, 25.6213633, 38.8702577, 52.4182761, 66.2706552, 80.4326318, 97.6599695, 120.295132, 141.766533, 160.962606, 176.771785, 188.082505, 193.783199, 229.149613, 270.051435, 289.603997, 279.253008, 230.444179, 134.623223, -16.7641508, -233.206538, -489.797272, -842.946308, -1349.0636, -1986.97268, -3004.46287, -4364.84115, -6031.4145, -7957.99019, -10136.3744, -12487.2395, -15199.5083, -18463.6459, -21789.2919, -24932.21, -27708.8938, -30015.7269, -31368.8032, -31413.8615, -29796.6404},
        {-269.261202, -269.893529, -269.922537, -269.369733, -268.256623, -266.604716, -264.435518, -261.770537, -258.63128, -255.039254, -251.015968, -246.582927, -241.76164, -236.573613, -231.101996, -225.329301, -219.233375, -212.823116, -206.107424, -199.095198, -191.795336, -184.183849, -176.307067, -168.178431, -159.811381, -151.21936, -142.415809, -133.451427, -124.371315, -115.099826, -105.635271, -95.975965, -86.1202194, -76.0663476, -65.8536243, -55.4544499, -44.8381242, -33.9945817, -22.9137572, -11.5855851, 0, 12.2126797, 24.6532377, 37.2820741, 50.0595888, 62.9461818, 75.902253, 92.1509699, 114.42523, 134.883784, 152.162243, 164.896218, 171.721322, 171.273165, 196.398655, 224.091212, 228.754614, 202.026835, 135.545848, 20.9496279, -150.123852, -310.660599, -586.222934, -997.812488, -1566.43089, -2298.62462, -3251.05128, -4427.36643, -5831.22566, -7492.21992, -9336.19869, -11173.0819, -13320.9669, -16036.5623, -18968.5552, -21908.6681, -24641.4471, -26991.1174, -28630.3656, -29283.5396, -28674.987},
        {-260.19861, -260.329225, -259.924836, -259.002871, -257.580756, -255.67592, -253.305787, -250.487786, -247.239344, -243.577888, -239.520843, -235.085639, -230.2897, -225.150455, -219.702272, -213.951861, -207.904649, -201.574597, -194.975667, -188.121822, -181.027023, -173.729491, -166.215579, -158.495896, -150.581055, -142.481666, -134.20834, -125.822067, -117.376265, -108.761087, -99.9666886, -90.9832271, -81.8008583, -72.4097385, -62.81914, -53.0071261, -42.9503135, -32.6349483, -22.0472768, -11.1735453, 0, 12.4164605, 25.0318929, 37.7317017, 50.4012913, 62.9260662, 75.1914308, 90.4528171, 111.45927, 130.038073, 144.706292, 153.980991, 156.379236, 150.418091, 167.309568, 184.88779, 178.512015, 140.011827, 61.2168106, -66.0434513, -249.939375, -421.336883, -714.638337, -1141.92747, -1715.28801, -2427.92313, -3338.48847, -4436.41128, -5711.11882, -7182.28864, -8748.59704, -10135.8843, -11813.1906, -14138.1958, -16810.2724, -19624.9526, -22323.0395, -24636.7153, -26387.3746, -27365.9987, -27363.5686},
        {-255.977882, -255.881463, -255.267847, -254.15314, -252.553448, -250.484879, -247.963539, -245.005536, -241.626976, -237.843965, -233.672612, -229.129022, -224.229303, -218.989561, -213.39858, -207.489751, -201.298535, -194.846627, -188.155723, -181.247519, -174.14371, -166.976683, -159.642156, -152.146537, -144.496235, -136.697658, -128.757215, -120.733546, -112.677443, -104.470422, -96.0976845, -87.54443, -78.7958597, -69.837174, -60.5898415, -51.07938, -41.3361316, -31.3583329, -21.1442203, -10.6920305, 0, 12.0834578, 24.2514848, 36.3470461, 48.2131066, 59.6926314, 70.6285854, 84.449911, 104.046714, 120.68748, 132.759149, 138.648667, 136.742975, 125.429017, 134.94212, 143.521823, 126.996883, 77.2399175, -13.8764533, -154.479611, -352.696935, -567.399046, -886.363643, -1321.38453, -1884.25553, -2545.61083, -3418.77101, -4466.13874, -5650.11671, -6941.4336, -8277.51408, -9446.43994, -10847.9371, -12784.0016, -15099.507, -17645.0049, -20184.1708, -22494.7141, -24385.0854, -25653.2552, -26097.1941},
        {-259.301517, -258.908377, -257.986116, -256.552805, -254.626518, -252.225327, -249.367304, -246.070523, -242.353056, -238.232975, -233.728354, -228.857265, -223.63778, -218.087972, -212.193491, -205.992919, -199.527296, -192.821326, -185.899714, -178.787165, -171.508383, -164.22357, -156.80322, -149.253326, -141.57988, -133.788873, -125.886297, -117.935423, -109.990914, -101.921802, -93.710822, -85.3407098, -76.7941999, -68.0540274, -58.9282556, -49.5101175, -39.9060675, -30.1345682, -20.2140821, -10.1630719, 0, 11.3798946, 22.6798009, 33.7261314, 44.3452987, 54.3637154, 63.6077941, 75.75625, 93.9089438, 108.680992, 118.33465, 121.13217, 115.335806, 99.2078126, 102.39748, 103.311507, 77.9807358, 18.24746, -84.0460249, -237.057424, -448.944442, -690.660789, -1032.33718, -1482.18179, -2048.4028, -2687.14723, -3535.04044, -4537.81771, -5641.21434, -6777.12204, -7932.80681, -8992.50289, -10228.8405, -11850.2394, -13815.456, -16038.211, -18346.2706, -20623.88, -22615.3228, -24112.5969, -24907.7002},
        {-272.576592, -271.509326, -269.905439, -267.78581, -265.171318, -262.08284, -258.541257, -254.567445, -250.182285, -245.406655, -240.261433, -234.767498, -228.945729, -222.817004, -216.396127, -209.707821, -202.777266, -195.626582, -188.277893, -180.753317, -173.074977, -165.336801, -157.479185, -149.514334, -141.454452, -133.311744, -125.098412, -116.905882, -108.813673, -100.644566, -92.3785988, -83.9958093, -75.4762357, -66.7999162, -57.7079606, -48.3315539, -38.8199673, -29.2021095, -19.5068898, -9.76321692, 0, 11.0216661, 21.8299644, 32.2208922, 41.9904468, 50.9346252, 58.8494248, 69.5324439, 86.1799924, 99.0197821, 106.223032, 105.960959, 96.4047834, 75.725723, 72.5747565, 65.8414949, 32.1083658, -36.6877954, -148.610153, -311.721872, -534.086117, -754.965378, -1105.54732, -1581.21004, -2177.33164, -2855.74813, -3694.57455, -4648.93882, -5673.96884, -6720.15437, -7756.5378, -8621.99595, -9668.11695, -11166.9082, -12949.962, -14934.2741, -17017.0813, -19145.3929, -21086.8863, -22670.375, -23724.6727},
        {-290.803667, -289.274472, -287.128883, -284.394981, -281.100843, -277.274549, -272.944177, -268.137807, -262.883517, -257.209386, -251.143493, -244.713917, -237.948737, -230.876032, -223.51659, -215.903102, -208.068811, -200.043286, -191.856098, -183.536816, -175.115011, -166.733726, -158.293387, -149.807891, -141.291137, -132.757024, -124.219449, -115.792997, -107.577122, -99.3449735, -91.0695665, -82.7239183, -74.2810457, -65.7139654, -56.7644909, -47.551899, -38.2129678, -28.7680048, -19.2373174, -9.64121332, 0, 11.7686104, 23.2513321, 34.1374748, 44.1163481, 52.8772616, 60.109525, 69.661843, 84.7579215, 95.1414814, 98.8129842, 93.7728915, 78.021665, 49.5597661, 33.8139514, 11.434708, -39.0035261, -125.110111, -254.494408, -434.765777, -673.533578, -911.692133, -1278.54529, -1765.94828, -2365.75631, -3055.88021, -3862.5713, -4760.95166, -5726.14334, -6770.29687, -7757.44895, -8304.76651, -9080.37076, -10594.8809, -12284.7376, -14027.3107, -15837.8677, -17770.7606, -19586.105, -21144.3248, -22305.8438},
        {-314.860794, -312.399881, -309.315105, -305.634028, -301.38421, -296.593211, -291.288593, -285.497916, -279.248741, -272.568628, -265.485138, -258.025831, -250.218269, -242.090012, -233.621957, -224.871184, -215.898305, -206.740427, -197.434653, -188.018091, -178.527844, -169.058221, -159.581225, -150.12606, -140.721932, -131.398046, -122.183608, -113.282612, -104.872792, -96.565407, -88.3186921, -80.0908819, -71.8402108, -63.5249134, -54.6077698, -45.360441, -36.0920914, -26.8622969, -17.730633, -8.75667548, 0, 10.1361592, 19.7071329, 28.4245933, 36.0002126, 42.145663, 46.5726166, 53.4597832, 66.3146467, 74.1676824, 74.9168018, 66.4599163, 46.6949373, 13.5197764, -8.04792363, -37.2662924, -95.4463037, -190.237163, -329.288076, -520.248248, -770.766884, -1027.86707, -1411.75897, -1912.81564, -2521.41019, -3210.41698, -4013.37242, -4899.65133, -5838.62854, -6829.70559, -7752.17711, -8281.29834, -8959.95466, -10188.4093, -11658.3146, -13267.0828, -14935.01, -16552.2472, -18106.642, -19552.7363, -20845.072},
        {-344.679856, -341.022052, -336.721252, -331.8069, -326.308441, -320.255316, -313.67697, -306.602846, -299.062388, -291.085039, -282.700244, -273.937445, -264.826086, -255.395611, -245.632885, -235.594288, -225.33942, -214.906433, -204.33348, -193.658713, -182.920285, -172.031953, -161.173444, -150.40009, -139.767223, -129.330176, -119.144279, -109.546635, -100.832006, -92.3819813, -84.1374849, -76.0394417, -68.0287763, -60.0464131, -51.3358712, -42.2792565, -33.3114654, -24.5160711, -15.9766467, -7.77676527, 0, 9.05698653, 17.2769261, 24.3294609, 29.8842333, 33.6108856, 35.1790599, 39.0222633, 48.8581746, 52.9654922, 49.0795762, 34.9357869, 8.26948448, -33.1839709, -69.8652883, -117.845006, -194.845637, -307.595725, -462.823816, -667.258454, -927.628182, -1172.06384, -1551.03265, -2052.14941, -2663.0289, -3356.3428, -4156.56543, -5033.37999, -5956.46971, -6923.29205, -7820.20756, -8370.28056, -8997.37166, -9998.44616, -11243.2334, -12663.9936, -14157.567, -15604.3705, -16985.934, -18249.317, -19341.579},
        {-377.73975, -373.104753, -367.713572, -361.606407, -354.823458, -347.404927, -339.391013, -330.821917, -321.73784, -312.178982, -302.185543, -291.797725, -281.055728, -269.999751, -258.680042, -247.130445, -235.384047, -223.478992, -211.453428, -199.3455, -187.193353, -174.725993, -162.333401, -150.096417, -138.095882, -126.412635, -115.127518, -104.688851, -95.4897345, -86.7323237, -78.3482499, -70.2691445, -62.4266393, -54.752366, -46.39279, -37.7762419, -29.3904859, -21.3277529, -13.6802736, -6.54027905, 0, 8.0344215, 14.9827519, 20.4508465, 24.0445606, 25.3697495, 24.0322684, 24.5292753, 30.6229514, 30.2234886, 20.9507164, 0.424463843, -33.7354397, -83.909165, -135.640282, -201.959857, -297.173412, -427.104905, -597.578294, -814.417538, -1083.44659, -1323.86307, -1698.16163, -2193.72605, -2797.94011, -3485.15224, -4274.9, -5138.92486, -6048.9683, -7003.2791, -7894.07684, -8485.75838, -9099.45099, -9946.79918, -10994.753, -12210.4725, -13512.3499, -14852.487, -16096.9162, -17141.0119, -17880.1487},
        {-411.395283, -406.424621, -400.436968, -393.495203, -385.662204, -377.00085, -367.574018, -357.444586, -346.675432, -335.329434, -323.469471, -311.158421, -298.459161, -285.43457, -272.235267, -258.868624, -245.335373, -231.680445, -217.94877, -204.185281, -190.434908, -176.384163, -162.485896, -148.83454, -135.524528, -122.650289, -110.306257, -98.9695798, -89.0598987, -79.7565246, -70.9984936, -62.7248418, -54.8746053, -47.3868203, -39.3994688, -31.3583368, -23.7698461, -16.736882, -10.3623296, -4.74907387, 0, 5.82942382, 10.3061333, 13.044481, 13.6588196, 11.7635017, 6.97287981, 3.62121897, 5.33356082, 0.438408409, -13.3663218, -38.3827132, -76.9128493, -131.258814, -184.276109, -250.568796, -346.212957, -477.488314, -650.674588, -872.051499, -1147.89877, -1439.87244, -1836.72835, -2332.59123, -2921.58585, -3572.41332, -4341.90989, -5193.6922, -6091.37687, -7008.53331, -7878.91978, -8574.4393, -9231.47707, -9918.11911, -10798.793, -11871.3261, -13019.5422, -14158.9815, -15192.3213, -16026.6241, -16568.9523},
        {-447.704261, -442.311139, -435.636192, -427.765098, -418.783533, -408.777175, -397.831702, -386.03279, -373.466118, -360.217362, -346.372199, -332.016307, -317.235363, -302.115045, -286.898914, -271.57277, -256.110459, -240.565366, -224.990874, -209.440365, -193.967224, -178.224329, -162.720881, -147.565575, -132.867107, -118.734172, -105.275468, -93.0029713, -82.3680642, -72.5151463, -63.3891673, -54.9350772, -47.097826, -39.8223634, -32.2517941, -24.8383183, -18.0948322, -12.1302966, -7.05367222, -2.97391977, 0, 3.9144123, 6.20741074, 6.47237482, 4.30268407, -0.708281985, -8.96714383, -16.6217659, -20.4599418, -31.0714639, -50.5921446, -81.1577962, -124.904231, -183.967261, -239.909578, -307.881626, -404.591142, -536.382, -709.598075, -930.583241, -1205.68137, -1520.42345, -1920.72148, -2405.59936, -2974.08102, -3596.46265, -4342.62963, -5177.1329, -6064.52341, -6976.43908, -7856.16996, -8624.20503, -9304.38621, -9874.12416, -10584.9548, -11466.2867, -12424.6604, -13408.4535, -14321.5634, -15083.0508, -15611.977},
        {-486.985302, -481.063369, -473.621951, -464.766211, -454.601306, -443.232398, -430.764646, -417.30321, -402.95325, -387.819926, -372.008398, -355.623825, -338.771369, -321.556188, -304.281376, -286.926879, -269.457662, -251.938399, -234.433764, -217.008433, -199.727079, -182.251363, -165.104632, -148.40722, -132.27946, -116.841686, -102.214231, -88.9322037, -77.46839, -66.9513455, -57.3329911, -48.5652476, -40.6000359, -33.3892771, -26.1172825, -19.2215655, -13.1977443, -8.15474813, -4.20150593, -1.44694686, 0, 2.46866343, 3.07511633, 1.37368948, -3.08128632, -10.7354803, -22.0345617, -33.804155, -43.4138398, -59.9652257, -85.3738362, -121.555195, -170.424826, -233.898252, -294.951034, -367.480185, -466.816392, -598.74919, -769.068117, -983.56271, -1248.02251, -1549.17979, -1927.42098, -2384.28343, -2921.30444, -3517.40583, -4229.90974, -5033.21508, -5901.72077, -6825.70807, -7731.92885, -8541.16776, -9225.3043, -9715.70976, -10261.4576, -10921.8731, -11668.0294, -12532.7126, -13390.1094, -14168.4748, -14796.0635},
        {-523.530987, -517.18179, -509.099572, -499.404296, -488.215926, -475.654428, -461.839764, -446.8919, -430.930799, -414.076427, -396.448746, -378.167723, -359.353319, -340.125501, -320.743696, -301.239643, -281.634523, -262.019777, -242.486841, -223.127154, -204.032154, -185.1605, -166.754747, -148.924671, -131.780048, -115.430653, -99.9862631, -85.8913394, -73.5400536, -62.2419086, -51.970787, -42.7005711, -34.4051436, -27.0583868, -19.9670615, -13.5270739, -8.18482821, -4.05066127, -1.23490977, 0.152089566, 0, 0.63444541, -0.827294392, -4.83197899, -11.826368, -22.2572209, -36.5712973, -52.069956, -66.5271897, -87.9107707, -117.944593, -158.352549, -210.858533, -277.186437, -340.946978, -415.322511, -514.866481, -645.007681, -811.174906, -1018.79695, -1273.30261, -1585.92326, -1950.09153, -2374.72401, -2868.73727, -3413.26677, -4085.75571, -4861.78357, -5716.92985, -6636.49509, -7566.89564, -8502.24447, -9262.53211, -9658.4135, -10079.7924, -10610.8568, -11207.0811, -11865.721, -12545.62, -13218.7166, -13856.9493},
        {-573.544328, -567.3913, -559.264478, -549.297849, -537.625402, -524.381123, -509.699, -493.713021, -476.557174, -458.365445, -439.271823, -419.410296, -398.91485, -377.919473, -356.673279, -335.237426, -313.664358, -292.064516, -270.548338, -249.226266, -228.208738, -207.724455, -187.749263, -168.377269, -149.702582, -131.819308, -114.821554, -98.9907231, -84.5800736, -71.2358692, -58.9761693, -47.8190334, -37.782521, -28.8846914, -20.80112, -13.7665228, -8.04154082, -3.71428576, -0.872869407, 0.394596441, 0, -0.352641544, -2.79107424, -7.65091515, -15.2677814, -25.9772899, -40.115058, -55.4388162, -70.0936536, -90.5792079, -118.277801, -154.571754, -200.84339, -258.47503, -314.439123, -379.233729, -464.447671, -574.410692, -713.452534, -885.902939, -1096.09165, -1361.90342, -1660.1381, -2002.53274, -2400.82438, -2844.97238, -3400.43194, -4052.80709, -4787.70187, -5593.56535, -6447.46639, -7421.88042, -8212.53356, -8539.75889, -8877.49863, -9315.97155, -9789.15075, -10251.9446, -10731.7661, -11231.3108, -11753.2741},
        {-655.005025, -650.22828, -643.153049, -633.930991, -622.713764, -609.653026, -594.900437, -578.607655, -560.926338, -542.008144, -522.004733, -501.067762, -479.348891, -456.999777, -434.44623, -411.666479, -388.618003, -365.396384, -342.097206, -318.816053, -295.648506, -272.707411, -250.068705, -227.82559, -206.071264, -184.898927, -164.401779, -144.67741, -125.822753, -107.920505, -91.0620844, -75.338907, -60.8423896, -47.6639487, -36.2708463, -26.5167378, -18.2268289, -11.4156602, -6.09777208, -2.28770517, 0, 0.899018952, 0.226125436, -2.05369074, -5.97543976, -11.5741318, -18.8847771, -26.3284314, -32.5686664, -41.4996417, -53.8116848, -70.1951234, -91.3402849, -117.937497, -147.032956, -181.622281, -224.976933, -278.532618, -343.72504, -421.989903, -514.762911, -581.305808, -694.756475, -854.432254, -1059.65049, -1330.79573, -1615.21967, -1937.52009, -2322.2948, -2806.55382, -3377.65823, -4109.87456, -4751.12504, -5083.61292, -5376.10302, -5691.28314, -6024.29668, -6414.84462, -6791.94206, -7122.33809, -7372.7818}
    };
    
    const double a5tab[21][81] = {
        {519.523893, 518.131239, 515.8465, 512.673834, 508.617401, 503.681359, 497.869868, 491.187084, 483.637169, 475.224279, 465.952575, 455.826214, 444.849356, 433.026159, 419.876634, 405.711214, 390.876979, 375.477116, 359.614813, 343.393256, 326.915633, 310.97921, 294.897239, 278.677049, 262.325971, 245.851335, 229.260471, 212.468357, 195.403847, 178.295097, 161.186937, 144.124194, 127.151698, 110.314275, 94.0951208, 78.2617512, 62.5485019, 46.9105375, 31.3030226, 15.6811219, 0, -14.8448983, -29.9685477, -45.5456425, -61.7508773, -78.7589465, -96.7445447, -115.834203, -136.161689, -158.016862, -181.593973, -207.087271, -234.691007, -264.599431, -315.508826, -375.908989, -432.889195, -482.859257, -522.228987, -547.408199, -554.806705, -574.962906, -596.142264, -541.649647, -334.789923, 439.372044, 1023.20821, 1899.29847, 3550.22271, 6493.90189, 11106.8927, 19378.4641, 26907.6181, 30293.5484, 32120.0118, 32707.9753, 31765.1304, 28865.1246, 24731.0574, 19694.1679, 14085.6953},
        {401.023924, 404.584455, 406.799776, 407.720454, 407.397055, 405.880146, 403.220293, 399.468062, 394.674018, 388.888729, 382.162761, 374.54668, 366.091051, 356.846442, 346.759978, 335.947662, 324.533328, 312.588699, 300.185499, 287.395453, 274.290286, 261.318544, 248.123086, 234.723595, 221.139753, 207.391243, 193.497745, 179.435614, 165.187713, 150.877329, 136.541738, 122.218216, 107.944036, 93.7564757, 80.056375, 66.6510166, 53.320162, 40.0267222, 26.733608, 13.4037303, 0, -12.7748555, -25.7999156, -39.2144432, -53.157701, -67.768952, -83.1874591, -100.053584, -118.932391, -138.764961, -159.487092, -181.034587, -203.343245, -226.348868, -271.660391, -325.50311, -372.461762, -408.039085, -427.737821, -427.060708, -401.510488, -366.642495, -338.58577, -231.772114, 39.3666742, 733.614349, 1509.27518, 2659.77438, 4478.53714, 7303.46344, 11294.5542, 17504.0752, 23491.4724, 27390.0091, 30357.5262, 32364.0939, 33034.5068, 31900.5993, 29211.3213, 24971.1506, 19184.5649},
        {320.853928, 323.645506, 325.647365, 326.8671, 327.312305, 326.990576, 325.909508, 324.076696, 321.499735, 318.186219, 314.143744, 309.379905, 303.902297, 297.718514, 290.663545, 282.854178, 274.420263, 265.404701, 255.850392, 245.800236, 235.297134, 224.310145, 212.966208, 201.31842, 189.419881, 177.323689, 165.082942, 152.819472, 140.644784, 128.447624, 116.253182, 104.086648, 91.9732114, 79.938062, 68.2956295, 56.8881296, 45.5358834, 34.2049193, 22.8612652, 11.4709494, 0, -10.9706148, -22.1747344, -33.7312581, -45.7590852, -58.377115, -71.7042469, -87.0660335, -105.606713, -124.55994, -143.554672, -162.21987, -180.184491, -197.077497, -236.11891, -282.013897, -317.681883, -337.92647, -337.551259, -311.359852, -254.155851, -10.8720224, 90.9067418, 180.871687, 388.714058, 763.065479, 1633.56374, 3032.62932, 4992.68271, 7505.62609, 10725.4348, 15243.7119, 19832.0917, 23594.4221, 26919.4491, 29614.2054, 31277.3477, 31430.3293, 30111.7931, 27184.0668, 22509.478},
        {264.604952, 268.01221, 270.555054, 272.259659, 273.152196, 273.258839, 272.605761, 271.219136, 269.125135, 266.349933, 262.919703, 258.860616, 254.198848, 248.96057, 243.144059, 236.793135, 229.953732, 222.657727, 214.937001, 206.823432, 198.348899, 189.538801, 180.432392, 171.062448, 161.461742, 151.663049, 141.699143, 131.629487, 121.509533, 111.308241, 101.047548, 90.7493939, 80.4357158, 70.1284518, 60.0684983, 50.1392797, 40.2076459, 30.2507491, 20.2457412, 10.1697742, 0, -10.1716849, -20.49872, -31.0198002, -41.7736206, -52.7988762, -64.1342619, -77.2252061, -93.3057587, -109.050954, -123.92831, -137.40534, -148.949562, -158.028492, -191.573447, -231.678243, -258.357804, -265.462155, -246.841316, -196.345311, -107.824161, 37.7516773, 181.758735, 395.308243, 749.511432, 1325.47786, 2169.65629, 3365.15583, 4995.08563, 7176.1872, 9890.67249, 13346.7352, 16906.2956, 20068.6758, 23231.2194, 26180.9312, 28396.171, 29233.1402, 28845.3183, 27114.3841, 23922.0167},
        {226.936794, 230.072047, 232.394903, 233.936573, 234.728264, 234.801186, 234.186548, 232.915558, 231.019425, 228.529359, 225.476568, 221.89226, 217.807645, 213.253932, 208.310342, 202.977713, 197.253243, 191.158322, 184.714336, 177.942674, 170.864725, 163.458199, 155.794193, 147.900128, 139.803422, 131.531496, 123.11177, 114.613821, 106.100893, 97.4996009, 88.8202487, 80.0731387, 71.2685737, 62.4168564, 53.6474839, 44.895356, 36.0863504, 27.2063895, 18.2413957, 9.17729174, 0, -9.32918894, -18.7963188, -28.4120651, -38.1871037, -48.13211, -58.2577598, -70.546693, -86.685238, -101.968879, -115.607611, -126.811428, -134.790326, -138.754299, -162.950269, -190.74978, -203.62919, -195.677397, -160.983299, -93.6357938, 12.2762194, 119.095052, 277.884525, 535.748638, 939.791391, 1533.80335, 2373.06163, 3500.69416, 4959.82882, 6835.43808, 9045.11621, 11519.1146, 14208.7872, 17071.2823, 20064.301, 22950.6412, 25337.2994, 26758.9674, 27195.3575, 26474.4695, 24424.3036},
        {208.120791, 209.580113, 210.509836, 210.922517, 210.830712, 210.24698, 209.183878, 207.653963, 205.669793, 203.243925, 200.388916, 197.117324, 193.441706, 189.374619, 184.867629, 179.971877, 174.743121, 169.206393, 163.386727, 157.309156, 150.998712, 144.5579, 137.923582, 131.11009, 124.131758, 117.00292, 109.737907, 102.365653, 94.9128929, 87.3590567, 79.7125504, 71.9817802, 64.1751523, 56.3010729, 48.4310367, 40.53354, 32.5723035, 24.5428292, 16.4406192, 8.26117543, 0, -8.46945348, -17.0127789, -25.6176185, -34.2716146, -42.9624094, -51.6776452, -62.6192298, -77.6563512, -91.4820897, -103.185025, -111.853735, -116.576801, -116.442801, -139.244831, -165.912887, -175.204164, -160.335968, -114.525607, -30.9903895, 97.0523769, 277.669417, 490.299248, 779.530743, 1189.95278, 1700.99006, 2517.96931, 3607.28304, 4935.32375, 6461.28195, 8173.15611, 9967.4643, 12094.3003, 14791.2017, 17529.173, 20080.8976, 22307.089, 24120.9478, 25137.6962, 25074.1402, 23647.0861},
        {196.99112, 197.852928, 198.189355, 198.02151, 197.370502, 196.257439, 194.703432, 192.729589, 190.35702, 187.606833, 184.500137, 181.058042, 177.301656, 173.252089, 168.982587, 164.481277, 159.732339, 154.746218, 149.53336, 144.104209, 138.469211, 132.610235, 126.570248, 120.363641, 114.004805, 107.50813, 100.888009, 94.1862905, 87.4406999, 80.5999704, 73.6673442, 66.6460632, 59.5393694, 52.3505049, 45.099863, 37.7798357, 30.3815167, 22.9046398, 15.3489389, 7.71414766, 0, -8.13710569, -16.3066832, -24.4615817, -32.5546505, -40.5387386, -48.3666954, -58.602855, -73.4191447, -86.5240549, -96.8100879, -103.169746, -104.495532, -99.6799483, -114.611839, -131.105722, -128.932611, -101.463129, -42.0679019, 55.8824482, 199.017297, 332.674538, 559.635715, 895.530662, 1355.98921, 1943.25865, 2705.97899, 3643.72114, 4756.05602, 6064.72725, 7502.78762, 8874.38743, 10537.2961, 12783.959, 15205.7867, 17604.9312, 19811.9288, 21677.5237, 22943.2431, 23387.8477, 22790.0979},
        {190.832536, 191.182498, 191.070976, 190.515396, 189.533185, 188.141768, 186.358572, 184.201023, 181.686546, 178.832569, 175.656516, 172.175814, 168.40789, 164.370168, 160.095692, 155.592007, 150.865481, 145.930345, 140.800832, 135.491174, 130.015603, 124.4074, 118.659118, 112.782357, 106.788719, 100.689805, 94.4972168, 88.259642, 82.0201968, 75.7018023, 69.3010013, 62.8143362, 56.2383498, 49.5695846, 42.8059539, 35.9431333, 28.9766944, 21.9028994, 14.7180105, 7.41828999, 0, -8.32900583, -16.675172, -24.9333518, -32.9983983, -40.7651645, -48.1285038, -57.6696175, -71.566051, -83.289292, -91.6434499, -95.4326343, -93.4609546, -84.5325202, -93.2601919, -102.121369, -91.6398416, -55.3407527, 13.2507551, 120.609539, 273.210457, 414.532412, 654.724254, 1002.23974, 1465.53261, 2035.81164, 2764.06811, 3638.06196, 4645.55311, 5800.59664, 7012.06701, 8011.00495, 9284.86209, 11212.7157, 13428.9288, 15737.5879, 17929.6157, 19775.3061, 21140.6326, 21858.9509, 21763.6162},
        {188.572122, 188.667871, 188.320343, 187.545776, 186.360407, 184.780474, 182.822216, 180.50187, 177.835674, 174.839867, 171.530685, 167.924367, 164.03715, 159.885273, 155.464333, 150.803623, 145.934004, 140.875935, 135.649876, 130.276286, 124.775626, 119.260585, 113.646656, 107.941562, 102.153024, 96.2887649, 90.3565074, 84.4012916, 78.4625502, 72.4587743, 66.3825339, 60.226399, 53.9829398, 47.6447264, 41.1445932, 34.5128992, 27.7845253, 20.9642599, 14.0568915, 7.06720873, 0, -8.0991995, -16.1338685, -23.9667389, -31.460543, -38.4780125, -44.8818797, -53.3854886, -66.2738445, -76.5935407, -83.0498692, -84.3481219, -79.1935907, -66.2915677, -69.5515217, -71.7344903, -53.6936662, -8.97903516, 68.8594169, 186.271704, 349.70784, 524.871315, 784.98631, 1138.2079, 1592.69116, 2121.10369, 2819.1361, 3652.35886, 4586.34238, 5594.64432, 6618.87351, 7444.53231, 8495.24079, 10098.3722, 12019.8806, 14111.2512, 16180.113, 18033.9887, 19524.6826, 20488.903, 20763.3579},
        {192.640596, 192.425679, 191.760187, 190.661977, 189.148906, 187.238831, 184.949609, 182.299097, 179.305152, 175.985632, 172.358392, 168.44129, 164.252184, 159.80893, 155.105186, 150.174117, 145.050721, 139.757805, 134.318175, 128.754639, 123.090004, 117.461761, 111.762194, 105.99827, 100.176957, 94.3052222, 88.3900346, 82.4792313, 76.6145088, 70.7051111, 64.7414116, 58.7137839, 52.6126016, 46.4282381, 40.0060334, 33.4281099, 26.7875683, 20.1044475, 13.3987865, 6.69062437, 0, -7.58044865, -14.9752011, -22.0361381, -28.61514, -34.5640876, -39.7348614, -47.0396638, -58.9308472, -67.9427133, -72.6845537, -71.76566, -63.7953237, -47.3828366, -46.0843521, -42.7276859, -18.2522863, 33.8352296, 120.028244, 246.820141, 420.704303, 616.581004, 894.493792, 1259.59357, 1717.03125, 2227.35983, 2905.73808, 3703.79987, 4573.17903, 5454.46437, 6332.42487, 7077.50902, 7998.51352, 9337.6381, 10966.8088, 12792.4299, 14674.4371, 16510.3169, 18093.8205, 19254.0633, 19820.1611},
        {205.201042, 204.348048, 203.041196, 201.300742, 199.146942, 196.600052, 193.680328, 190.408026, 186.803402, 182.886711, 178.67821, 174.198155, 169.466801, 164.504404, 159.33076, 153.966415, 148.431952, 142.747722, 136.934076, 131.011362, 124.999932, 118.983059, 112.90948, 106.790855, 100.638845, 94.4651098, 88.2813103, 82.1587966, 76.15995, 70.1537049, 64.1274852, 58.0687151, 51.9648188, 45.8032204, 39.3720536, 32.7848145, 26.1700843, 19.5560505, 12.9709002, 6.44282098, 0, -7.3490753, -14.4156901, -21.0308295, -27.0254786, -32.2306226, -36.4772465, -42.7807182, -53.6779141, -61.3855962, -64.4417882, -61.3845132, -50.7517947, -31.0816562, -25.2471131, -16.391784, 14.1827643, 72.9160298, 166.247511, 300.616705, 482.463112, 660.681848, 945.390279, 1331.37205, 1813.41082, 2356.24576, 3028.77002, 3789.71416, 4597.80879, 5408.08129, 6190.37179, 6777.22915, 7544.01912, 8786.78629, 10268.4158, 11898.0197, 13595.0589, 15311.8408, 16858.9378, 18095.2411, 18879.6421},
        {221.690131, 220.398247, 218.58656, 216.281616, 213.509957, 210.298129, 206.672674, 202.660137, 198.287063, 193.579994, 188.565475, 183.270051, 177.720264, 171.94266, 165.966042, 159.815525, 153.516051, 147.093703, 140.574562, 133.984708, 127.350225, 120.795949, 114.235566, 107.681522, 101.146258, 94.6422182, 88.1818448, 81.8547985, 75.7391366, 69.6626667, 63.6064793, 57.5516645, 51.4793127, 45.3705141, 39.013682, 32.5117949, 25.9824161, 19.4460464, 12.9231866, 6.43433754, 0, -8.04107527, -15.7543762, -22.8871409, -29.1866072, -34.4000132, -38.2745968, -43.8795544, -53.7849183, -59.7947447, -60.3074499, -53.7214501, -38.4351615, -12.8470002, 2.69274262, 23.672487, 67.242313, 139.493879, 246.518844, 394.408865, 589.255602, 780.905091, 1078.63779, 1474.37621, 1960.04287, 2514.26833, 3161.76182, 3878.49565, 4640.44212, 5456.18109, 6203.86222, 6522.65532, 7063.54625, 8330.475, 9738.76737, 11163.6008, 12630.3005, 14190.7865, 15640.4513, 16862.1399, 17738.6977},
        {243.041167, 240.916442, 238.268739, 235.124196, 231.508948, 227.44913, 222.97088, 218.100331, 212.863622, 207.286886, 201.396261, 195.217881, 188.777884, 182.102404, 175.188008, 168.079538, 160.824073, 153.453798, 146.000897, 138.497553, 130.975952, 123.529313, 116.120354, 108.772831, 101.510497, 94.3571074, 87.336416, 80.6075209, 74.3091827, 68.1415349, 62.0733782, 56.0735135, 50.1107416, 44.1538631, 37.7686538, 31.1788696, 24.6387738, 18.1996027, 11.9125925, 5.82897951, 0, -6.83643511, -13.134385, -18.661234, -23.1843668, -26.4711679, -28.2890216, -31.9863919, -40.3736424, -44.6553882, -43.1449803, -34.15577, -16.0011085, 13.005653, 32.7119582, 58.6200168, 107.88517, 186.661095, 301.10147, 457.359973, 661.59028, 867.621722, 1179.24059, 1586.96251, 2081.30312, 2637.14399, 3283.56464, 3992.32, 4735.16499, 5510.68417, 6210.14351, 6525.55868, 6991.93862, 8010.82803, 9232.1407, 10550.4892, 11905.9158, 13206.7859, 14442.7419, 15575.7181, 16567.6489},
        {269.088009, 265.924239, 262.224134, 258.015465, 253.326007, 248.183532, 242.615814, 236.650625, 230.315739, 223.638929, 216.647968, 209.370629, 201.834685, 194.067909, 186.073201, 177.89407, 169.575906, 161.151569, 152.653921, 144.115822, 135.570133, 126.971984, 118.442702, 110.025882, 101.765121, 93.7040138, 85.886157, 78.5745132, 71.9990828, 65.6789304, 59.5685817, 53.6225628, 47.7953994, 42.0416175, 35.7514821, 29.2364724, 22.8507801, 16.6643459, 10.7471106, 5.16901506, 0, -6.10211882, -11.4602514, -15.8094327, -18.884698, -20.4210821, -20.1536201, -21.6460347, -27.8867423, -29.4559488, -24.5341191, -11.3017181, 12.0607894, 47.3729384, 78.7156359, 119.129938, 182.999637, 275.772557, 402.896523, 569.81936, 781.988894, 978.133378, 1286.266, 1694.78546, 2192.09049, 2753.06761, 3399.44548, 4103.40834, 4837.14042, 5597.6513, 6282.64914, 6628.25706, 7059.49464, 7882.44393, 8911.69889, 10073.7181, 11286.1547, 12446.5037, 13541.6988, 14528.1828, 15362.399},
        {297.538563, 293.556225, 288.943238, 283.736585, 277.973249, 271.690214, 264.924463, 257.712979, 250.092744, 242.100743, 233.773957, 225.149371, 216.263967, 207.154729, 197.88078, 188.465097, 178.928981, 169.304887, 159.625269, 149.922583, 140.229283, 130.357193, 120.589868, 110.990234, 101.621217, 92.545741, 83.8267321, 75.8132842, 68.8114909, 62.2000155, 55.9255896, 49.9349446, 44.1748124, 38.5919243, 32.4983618, 26.2423386, 20.2201084, 14.5082154, 9.18320368, 4.32161718, 0, -5.43618466, -9.94023776, -13.1965411, -14.8894766, -14.7034259, -12.322771, -11.3742724, -14.8923013, -13.1345613, -4.18470192, 13.8736276, 42.956778, 84.9810999, 128.019953, 182.746814, 260.883015, 367.176375, 506.374714, 683.225853, 902.477611, 1096.46704, 1401.4294, 1805.90828, 2298.44724, 2855.37901, 3495.36723, 4192.30257, 4920.07567, 5676.13651, 6363.69775, 6757.39355, 7186.55806, 7877.68362, 8738.18483, 9726.41403, 10777.6599, 11849.6135, 12833.4199, 13642.9101, 14191.9155},
        {325.967927, 321.759463, 316.699275, 310.843686, 304.249017, 296.971589, 289.067725, 280.593744, 271.605969, 262.160722, 252.314323, 242.123094, 231.643357, 220.931433, 210.135985, 199.25492, 188.279152, 177.246116, 166.193245, 155.157973, 144.177733, 133.034222, 122.055929, 111.315608, 100.886012, 90.8398934, 81.2500054, 72.4853239, 64.8703136, 57.7694243, 51.135133, 44.9199163, 39.076251, 33.5566141, 27.6663412, 21.767293, 16.2703157, 11.2602536, 6.82195087, 3.0402516, 0, -3.83572388, -6.53633503, -7.79301253, -7.29693548, -4.73928297, 0.188765894, 3.97288046, 3.67320435, 8.74015049, 21.0348622, 42.4184827, 74.7521554, 119.897023, 163.757313, 218.288572, 296.654215, 403.979242, 545.38865, 726.007439, 950.960608, 1187.36036, 1510.91855, 1915.91188, 2396.61706, 2924.84282, 3550.2867, 4240.26406, 4962.09027, 5692.37005, 6370.55074, 6867.86278, 7344.32832, 7893.98241, 8608.93485, 9476.692, 10400.6984, 11305.6616, 12115.2992, 12753.6661, 13144.8175},
        {355.924553, 351.463197, 345.92638, 339.389563, 331.928211, 323.617786, 314.533751, 304.75157, 294.346705, 283.394619, 271.970776, 260.150639, 248.009671, 235.623334, 223.220949, 210.780647, 198.268917, 185.72975, 173.207141, 160.745081, 148.387565, 135.896053, 123.636089, 111.690687, 100.142859, 89.0756176, 78.5719749, 69.0240383, 60.7774699, 53.1762019, 46.1777443, 39.7396075, 33.8193017, 28.3743371, 22.7181984, 17.2158088, 12.2808414, 8.00253643, 4.47013431, 1.77287539, 0, -2.47767474, -3.61515922, -3.08588707, -0.563291916, 4.27919262, 11.7681329, 18.7689569, 22.6671699, 32.0653158, 48.6952942, 74.2890049, 110.578348, 159.295223, 205.2998, 260.997097, 340.069247, 447.699122, 589.069595, 769.363538, 993.763824, 1249.0487, 1575.51435, 1971.68525, 2436.08585, 2942.24165, 3550.34111, 4228.91015, 4946.47472, 5678.36192, 6372.69412, 6946.10806, 7452.8416, 7901.32133, 8470.01808, 9175.75662, 9940.06053, 10715.5055, 11425.7601, 12004.3475, 12384.791},
        {387.430025, 382.686339, 376.665071, 369.457889, 361.156459, 351.852449, 341.637525, 330.603354, 318.841603, 306.44394, 293.50203, 280.107541, 266.352139, 252.327492, 238.313641, 224.283086, 210.194068, 196.099725, 182.053193, 168.107609, 154.316108, 140.455871, 126.894102, 113.72205, 101.030964, 88.9120906, 77.4566788, 67.070536, 58.1122035, 49.9207821, 42.4597989, 35.6927807, 29.5832546, 24.0947477, 18.578586, 13.385577, 8.91286911, 5.24920986, 2.48334686, 0.704027705, 0, -1.49811338, -1.47300968, 0.434489138, 4.58356111, 11.3333843, 21.0431367, 31.1274123, 39.4432642, 53.390707, 74.5245146, 104.399461, 144.570319, 196.591864, 246.451664, 305.552364, 386.474976, 493.958418, 632.741605, 807.563455, 1023.16289, 1267.6883, 1576.12367, 1949.31661, 2388.11474, 2873.75188, 3455.4562, 4110.53896, 4816.31143, 5564.37504, 6289.17058, 6910.67669, 7431.21712, 7813.68352, 8241.97583, 8760.16318, 9344.65662, 10021.3633, 10685.864, 11278.1692, 11738.2894},
        {415.644597, 410.80384, 404.496223, 396.826118, 387.897897, 377.815933, 366.684597, 354.608262, 341.6913, 328.038083, 313.752984, 298.940374, 283.704625, 268.15011, 252.526896, 236.847188, 221.112163, 205.396391, 189.774444, 174.320894, 159.110312, 144.161103, 129.611761, 115.544616, 102.041996, 89.1862286, 77.0596429, 65.993561, 56.2818907, 47.411588, 39.3638818, 32.1200012, 25.6611749, 19.9686321, 14.4900332, 9.54414462, 5.49012131, 2.41832923, 0.419134351, -0.417097379, 0, -0.19189763, 1.31994565, 4.89557566, 10.8950383, 19.6783793, 31.6056445, 44.4793832, 56.4864388, 74.1021239, 98.7249103, 131.75327, 174.585674, 228.620594, 280.369375, 340.648141, 421.399902, 527.068169, 662.096454, 830.928265, 1038.00711, 1291.98541, 1588.48038, 1934.65271, 2337.66309, 2781.05892, 3330.24684, 3964.0518, 4661.29879, 5409.67817, 6161.41866, 6903.2878, 7499.09516, 7801.54674, 8123.45042, 8532.17509, 8990.76641, 9496.34206, 10014.0548, 10519.9545, 10990.0912},
        {453.17581, 448.837173, 442.811713, 435.216307, 426.167836, 415.783177, 404.179211, 391.472816, 377.78087, 363.220254, 347.907847, 331.960526, 315.495171, 298.628662, 281.605983, 264.462973, 247.225773, 229.98506, 212.831509, 195.855796, 179.148598, 162.951874, 147.184124, 131.915128, 117.214671, 103.152535, 89.7985016, 77.3576024, 66.0145451, 55.5157189, 45.8759913, 37.1102299, 29.2333022, 22.2600757, 15.9192601, 10.4067475, 5.94009081, 2.59268896, 0.43794067, -0.450755246, 0, 0.43664773, 2.55661682, 6.63037865, 12.9284046, 21.7211662, 33.2791347, 45.7879692, 57.7465969, 74.4105134, 96.8966909, 126.322102, 163.803718, 210.458512, 255.584784, 307.776041, 376.520412, 465.35228, 577.806029, 717.416043, 887.716706, 1103.5243, 1345.46792, 1623.23514, 1946.51352, 2306.45731, 2758.46946, 3289.99774, 3888.48991, 4543.93457, 5236.15697, 6023.08053, 6654.59158, 6900.07118, 7152.52932, 7486.30119, 7845.51724, 8191.2698, 8547.00849, 8914.92465, 9297.20964},
        {514.044111, 511.244973, 506.460749, 499.824339, 491.468637, 481.526543, 470.130952, 457.414762, 443.51087, 428.552173, 412.671569, 396.001954, 378.676226, 360.827281, 342.838575, 324.6845, 306.320485, 287.828178, 269.289225, 250.785274, 232.397974, 214.298224, 196.466094, 178.970904, 161.881976, 145.268632, 129.200193, 113.748283, 98.9841805, 84.9717004, 71.779229, 59.4751529, 48.1278582, 37.8057315, 28.8385816, 21.1294181, 14.5614618, 9.14962737, 4.90882923, 1.85398193, 0, -0.744376298, -0.243394776, 1.53252249, 4.61295341, 9.02747592, 14.8056679, 20.7043685, 25.671661, 32.7803872, 42.576898, 55.6075441, 72.4186764, 93.5566457, 116.626628, 144.035572, 178.41306, 220.907037, 272.665445, 334.836229, 408.567332, 461.903817, 552.18437, 679.028025, 842.053812, 1057.64344, 1284.06812, 1541.06193, 1848.35894, 2235.51207, 2692.79891, 3282.76076, 3795.71593, 4050.6964, 4271.96945, 4511.18603, 4762.95519, 5057.70467, 5339.62931, 5582.32165, 5759.37424}
    };
    
    const double b3tab[21][81] = {
        {43.0068882, 42.0198432, 41.0202058, 40.0082996, 38.9844479, 37.9489742, 36.902202, 35.8444546, 34.7760554, 33.697328, 32.6085957, 31.510182, 30.4024104, 29.2856042, 28.1351348, 26.9671105, 25.7995281, 24.6378149, 23.487398, 22.3537045, 21.2421617, 20.2361252, 19.2523312, 18.2854444, 17.3301295, 16.3810513, 15.4328744, 14.4741566, 13.494373, 12.5027911, 11.4965554, 10.4728101, 9.42869978, 8.36136873, 7.43944692, 6.55159635, 5.5734983, 4.46722121, 3.19483351, 1.71840363, 0, -1.82864689, -3.99856215, -6.57110885, -9.60765004, -13.1695488, -17.3181681, -23.1599594, -31.6443369, -40.3337383, -48.8651873, -56.8757074, -64.0023224, -69.8820558, -135.302016, -221.215459, -283.946851, -310.625471, -288.380599, -204.341513, -45.6374928, 1094.39585, 1322.02807, 1275.54747, 1593.24234, 5319.21786, 7157.42945, 10633.1226, 19271.5426, 33460.8266, 66137.5454, 158302.586, 233321.855, 232304.72, 214245.066, 191781.421, 164045.903, 133506.452, 102794.137, 74484.4135, 51152.7378},
        {34.8479771, 34.0828232, 33.2979234, 32.4946199, 31.6742548, 30.8381703, 29.9877087, 29.124212, 28.2490225, 27.3634824, 26.4689337, 25.5667189, 24.6581798, 23.7446589, 22.8204984, 21.8914687, 20.9638698, 20.0404758, 19.1240605, 18.2173979, 17.323262, 16.48684, 15.6626349, 14.8475632, 14.0385411, 13.2324851, 12.4263115, 11.6119553, 10.7820998, 9.94557265, 9.10131288, 8.24825948, 7.38535144, 6.51152776, 5.87655893, 5.32070703, 4.66524731, 3.85781326, 2.84603836, 1.57755611, 0, -1.69213164, -3.83216388, -6.50655697, -9.80177113, -13.8042666, -18.6005036, -25.4550677, -35.4555167, -45.8712791, -56.3104583, -66.381158, -75.6914817, -83.8495329, -163.679532, -268.472799, -345.978518, -380.829042, -357.656722, -261.093911, -75.7729602, 1382.32395, 1669.1229, 1567.52846, 1860.4452, 5398.05478, 7863.99151, 12521.8726, 22635.3153, 38866.4163, 72283.3546, 159204.733, 230487.863, 232792.773, 220036.698, 203067.024, 179727.633, 149951.728, 118318.723, 87824.0906, 61463.3019},
        {27.3016278, 26.3201609, 25.4316053, 24.6262982, 23.8945769, 23.2267788, 22.613241, 22.044301, 21.510296, 21.0015632, 20.5084399, 20.0212635, 19.5303712, 19.0261003, 18.4401786, 17.80002, 17.1374748, 16.4548684, 15.7545261, 15.0387734, 14.3099356, 13.5774014, 12.8354574, 12.0854534, 11.3287394, 10.5666652, 9.80058079, 9.02693396, 8.2429092, 7.46157766, 6.68627963, 5.92035536, 5.16714512, 4.42998917, 4.08113488, 3.89055052, 3.60027976, 3.13820596, 2.43221249, 1.41018271, 0, -1.63570072, -3.83620844, -6.7060606, -10.3497946, -14.8719479, -20.377058, -28.3310151, -39.9951502, -52.2188502, -64.5538995, -76.5520826, -87.7651839, -97.7449876, -189.784157, -310.45969, -400.009772, -440.857681, -415.426698, -306.140102, -95.4211725, 1828.26796, 2199.66409, 1975.70095, 2113.31228, 4971.62859, 8048.84049, 13984.5044, 25418.1767, 42730.9598, 75337.7724, 154867.656, 220221.658, 224088.91, 215027.204, 202272.431, 182777.578, 154477.671, 123662.487, 93773.5817, 68252.5085},
        {19.3140409, 18.8127435, 18.3268192, 17.8544265, 17.3937237, 16.9428693, 16.5000216, 16.0633391, 15.6309802, 15.2011032, 14.7718666, 14.3414288, 13.9079482, 13.4695833, 13.0132296, 12.5441705, 12.0685417, 11.5868053, 11.0994235, 10.6068582, 10.1095716, 9.60921725, 9.10490122, 8.59692112, 8.08557452, 7.57115897, 7.05397204, 6.53172619, 6.00252231, 5.47283926, 4.94402423, 4.41742444, 3.89438708, 3.37625936, 3.26677856, 3.31273845, 3.23047363, 2.93902574, 2.35743642, 1.40474729, 0, -1.64066316, -3.93633285, -7.00899908, -10.9806518, -15.9732812, -22.108877, -31.1013368, -44.4253548, -58.3964885, -72.4903626, -86.1826014, -98.9488294, -110.264671, -213.913405, -349.711289, -450.335908, -495.973035, -466.808444, -343.027907, -104.817199, 1622.64393, 2013.60032, 1968.70522, 2388.61189, 5741.17106, 9061.29224, 15130.2506, 26729.3214, 45277.9892, 77642.9008, 149564.093, 208953.624, 215315.125, 207096.358, 192638.252, 171985.692, 147259.988, 120572.795, 94039.0621, 69773.735},
        {12.7881556, 12.5733124, 12.333449, 12.0708547, 11.7878189, 11.4866307, 11.1695797, 10.838955, 10.497046, 10.146142, 9.78853229, 9.42650626, 9.06235319, 8.69836239, 8.35563103, 8.02455051, 7.69408857, 7.36268754, 7.02878975, 6.69083752, 6.34727319, 5.96573091, 5.57971601, 5.19192562, 4.8050569, 4.42180699, 4.04487306, 3.68477709, 3.35086534, 3.0271248, 2.71307548, 2.40823739, 2.11213051, 1.82427487, 2.01413104, 2.38393289, 2.60034188, 2.56675554, 2.18657142, 1.36318706, 0, -1.7361858, -4.24175745, -7.64969563, -12.092981, -17.7045944, -24.6175163, -34.7775333, -49.8580361, -65.6573797, -81.5724874, -97.0002825, -111.337688, -123.981628, -235.651602, -381.647502, -489.599547, -538.179962, -506.060977, -371.914819, -114.413714, 1427.64716, 1823.98063, 1923.64426, 2575.69561, 6217.63839, 9780.37322, 16019.0778, 27688.9297, 46640.1621, 78342.7864, 145553.488, 200330.975, 204985.626, 196733.976, 183520.692, 164854.785, 142255.175, 117636.798, 92779.5835, 69463.4617},
        {8.67660924, 8.32370542, 8.00786559, 7.72441883, 7.46869424, 7.23602091, 7.02172791, 6.82114435, 6.6295993, 6.44242186, 6.25494111, 6.06248614, 5.86038604, 5.64396989, 5.36030655, 5.0352547, 4.69832613, 4.35472116, 4.0096401, 3.66828328, 3.33585102, 3.07039037, 2.81695643, 2.57345101, 2.33777595, 2.10783307, 1.88152419, 1.63946016, 1.36484999, 1.09694022, 0.840653377, 0.600911977, 0.382638539, 0.190755583, 0.61184877, 1.28287839, 1.79677541, 2.03948795, 1.89696414, 1.25515211, 0, -1.78730835, -4.44302392, -8.10816197, -12.9237377, -19.0307665, -26.5702635, -37.9782779, -55.3460025, -73.3267645, -91.1297216, -107.964032, -123.038852, -135.563341, -253.40089, -407.025591, -518.686891, -565.369623, -524.05862, -371.738712, -85.3947335, 1642.39205, 2114.71184, 2223.45834, 2860.52525, 5983.2463, 9855.43663, 16647.5077, 28529.8711, 46469.7646, 76247.1207, 140366.291, 192371.438, 195823.168, 187334.435, 174699.255, 157267.568, 136599.006, 113858.512, 90345.5841, 67359.7194},
        {6.43126977, 6.27531737, 6.08715296, 5.87043508, 5.62882228, 5.3659731, 5.08554609, 4.7911998, 4.48659277, 4.17538354, 3.86123066, 3.54779269, 3.23872816, 2.93769562, 2.6778981, 2.44430419, 2.21964611, 2.00153933, 1.78759931, 1.57544153, 1.36268145, 1.11185147, 0.860495329, 0.611073717, 0.366047309, 0.127876785, -0.100977178, -0.325939738, -0.551251114, -0.755594685, -0.933307868, -1.07872808, -1.18619273, -1.25003925, -0.612569755, 0.319398433, 1.08969179, 1.57060448, 1.63443064, 1.15346443, 0, -1.83248891, -4.60962345, -8.47584522, -13.5755958, -20.0533167, -28.0534496, -40.4072577, -59.5422772, -79.1784521, -98.3692881, -116.168291, -131.628966, -143.804819, -261.626797, -414.639536, -524.070431, -566.498467, -518.502627, -356.661892, -57.5552467, 1284.42393, 1733.93985, 2047.08474, 2979.95078, 6611.84084, 10434.936, 16793.1685, 28030.4706, 45936.7797, 74518.013, 131486.776, 178779.249, 186637.764, 180981.383, 168083.622, 149637.232, 129364.855, 107450.345, 85283.9669, 64255.9838},
        {5.29448937, 5.04323248, 4.77966155, 4.50506682, 4.22073848, 3.92796675, 3.62804184, 3.32225395, 3.0118933, 2.69825009, 2.38261454, 2.06627686, 1.75052725, 1.43665593, 1.13009179, 0.829506878, 0.53326, 0.241794828, -0.0444449594, -0.325015685, -0.599473671, -0.889773659, -1.16998018, -1.43655617, -1.68596459, -1.91466838, -2.11913049, -2.29959983, -2.45575642, -2.57501055, -2.65228792, -2.68251428, -2.66061534, -2.58151683, -1.71284295, -0.509613024, 0.518100112, 1.22660722, 1.47221906, 1.1112464, 0, -2.14884821, -5.31551114, -9.6238405, -15.197688, -22.1609055, -30.6373446, -43.7251104, -64.075391, -84.7002585, -104.515922, -122.438591, -137.384475, -148.269783, -263.962577, -413.823124, -518.888817, -555.586124, -500.34151, -329.581443, -19.7323883, 1253.13033, 1722.01745, 2095.46885, 3082.02438, 6471.78476, 10305.4467, 16589.4126, 27330.0851, 43794.8627, 70207.1597, 123503.677, 168501.148, 178068.16, 173279.753, 159892.662, 141050.456, 122301.905, 102188.135, 81639.254, 61585.3689},
        {4.84351675, 4.51636659, 4.18454238, 3.84828129, 3.50782051, 3.16339724, 2.81524864, 2.46361192, 2.10872426, 1.75082284, 1.39014485, 1.02692748, 0.661407912, 0.293823333, -0.0856447143, -0.470397806, -0.853076352, -1.23138637, -1.60303387, -1.96572487, -2.31716538, -2.67235893, -3.00932511, -3.32338105, -3.60984386, -3.86403063, -4.0812585, -4.26199265, -4.40592472, -4.49606216, -4.52563179, -4.48786044, -4.37597493, -4.18320209, -3.08069225, -1.58172083, -0.261790799, 0.717722075, 1.19544202, 1.00999325, 0, -2.44350491, -5.97249041, -10.6865168, -16.6851445, -24.0679338, -32.934445, -46.6456857, -68.0725919, -89.5162313, -109.751912, -127.554941, -141.700627, -150.964279, -263.324215, -408.47352, -507.839091, -537.859663, -474.973969, -295.620744, 23.7612774, 1419.05114, 1935.38659, 2283.16973, 3172.80264, 6234.3338, 9909.7101, 16012.9004, 26357.8738, 42320.9255, 67029.0451, 114093.299, 154438.174, 165687.996, 163355.177, 151948.899, 134753.464, 116800.69, 97357.2159, 77655.0931, 58926.3714},
        {4.82260682, 4.44699608, 4.06271369, 3.67030988, 3.27033492, 2.86333905, 2.44987252, 2.0304856, 1.60572854, 1.17615158, 0.742304977, 0.304738993, -0.135996125, -0.579350123, -1.0338924, -1.49330357, -1.95057395, -2.40328793, -2.84902991, -3.2853843, -3.70993548, -4.15038129, -4.57003383, -4.96231861, -5.32066118, -5.63848706, -5.90922177, -6.13521168, -6.31746272, -6.42806964, -6.45683582, -6.39356464, -6.22805947, -5.9501237, -4.63367847, -2.8479176, -1.23136316, 0.0388453389, 0.785568377, 0.831666437, 0, -2.65228347, -6.43086025, -11.4071197, -17.652451, -25.2382436, -34.2358866, -48.1992661, -70.1589809, -91.8593735, -111.957828, -129.111727, -141.978456, -149.215398, -255.278804, -391.897309, -482.791038, -504.977092, -435.47257, -251.294572, 70.5398011, 1522.76762, 2063.48009, 2388.61223, 3194.09905, 6026.91803, 9483.77153, 15281.8373, 25138.2933, 40758.8599, 63895.0867, 104532.188, 139701.079, 151597.138, 151500.514, 142684.992, 127836.263, 110418.605, 91370.037, 72489.8072, 55577.1629},
        {5.12537602, 4.67830998, 4.22600324, 3.7684454, 3.30562605, 2.8375348, 2.36416125, 1.88549499, 1.40152563, 0.91224276, 0.417635984, -0.0823050996, -0.587590892, -1.09823179, -1.62810779, -2.16845532, -2.70946099, -3.24829829, -3.78214069, -4.30816171, -4.82353482, -5.36894761, -5.89204989, -6.38400556, -6.83597852, -7.23913267, -7.58463191, -7.87544036, -8.11274905, -8.25950615, -8.30208429, -8.22685609, -8.02019418, -7.66847118, -6.16973237, -4.13556985, -2.25238776, -0.708712917, 0.306927866, 0.606007777, 0, -2.81622665, -6.7603839, -11.8667879, -18.1697546, -25.7036003, -34.5026409, -48.1725756, -69.7824605, -90.8270269, -109.890495, -125.557084, -136.411015, -141.036506, -236.673829, -359.497024, -438.212308, -451.224619, -376.938899, -193.760087, 119.906876, 1347.48589, 1851.26156, 2237.22252, 3111.35744, 5837.1687, 9152.11593, 14571.197, 23609.4097, 37622.2754, 58603.2218, 96635.7552, 129169.689, 138671.701, 138033.642, 130473.028, 117553.059, 101287.315, 83542.1718, 66234.1913, 51279.9336},
        {6.07046285, 5.585625, 5.08344396, 4.56462091, 4.02985707, 3.47985361, 2.91531175, 2.33693266, 1.74541756, 1.14146763, 0.52578407, -0.100931924, -0.737979157, -1.38465643, -2.06000494, -2.75083438, -3.44246004, -4.13014258, -4.80914267, -5.47472097, -6.12213813, -6.78464611, -7.41426743, -8.00101589, -8.53490532, -9.0059495, -9.40416227, -9.73401699, -9.99781433, -10.1509936, -10.1776975, -10.0620689, -9.78825047, -9.34038497, -7.63854184, -5.34000051, -3.18258604, -1.3680886, -0.0982983435, 0.424994576, 0, -3.08931915, -7.32137822, -12.6888395, -19.1843653, -26.8006179, -35.5302595, -49.0355546, -70.427366, -90.9239181, -109.027897, -123.241988, -132.068878, -134.011252, -220.744836, -331.830711, -399.777169, -404.029121, -324.031477, -139.229147, 170.932957, 1189.77782, 1682.44204, 2145.68871, 3076.28091, 5592.05626, 8657.798, 13515.5527, 21407.3669, 32906.4806, 51261.3599, 89151.949, 120385.793, 124416.287, 122483.423, 118297.419, 109902.396, 95063.6149, 78327.8777, 61930.0868, 48105.1446},
        {7.30722818, 6.78165834, 6.2222456, 5.63092945, 5.00964936, 4.36034482, 3.6849553, 2.98542031, 2.26367931, 1.52167178, 0.761337221, -0.0153848974, -0.80655509, -1.61023387, -2.44139153, -3.2873914, -4.13431683, -4.97676958, -5.80935141, -6.62666409, -7.42330939, -8.25801351, -9.05239776, -9.79220786, -10.4631895, -11.0510886, -11.5416507, -11.9223401, -12.1803631, -12.2973559, -12.2583665, -12.0484429, -11.652633, -11.0559848, -9.17566822, -6.67227475, -4.28722941, -2.22400572, -0.686077233, 0.123082519, 0, -2.92215445, -6.95377737, -12.0686214, -18.2404393, -25.4429835, -33.6500069, -46.3563417, -66.5277358, -85.7186384, -102.473131, -115.335294, -122.849209, -123.558957, -201.015913, -299.988234, -358.809399, -358.63595, -280.62443, -105.93138, 184.286659, 1171.02488, 1639.01505, 2052.88883, 2877.27786, 5016.97965, 7850.88622, 12371.824, 19572.6195, 29795.4396, 45985.0894, 79044.5536, 106928.722, 112598.551, 111337.789, 106287.883, 97341.2319, 84708.7183, 70576.344, 56456.772, 43862.6649},
        {9.20514842, 8.63722588, 8.01374915, 7.3382654, 6.6143218, 5.8454655, 5.03524369, 4.18720353, 3.30489219, 2.39185683, 1.45164463, 0.487802757, -0.496121627, -1.49658135, -2.52715272, -3.5734562, -4.61981607, -5.65918269, -6.68450644, -7.68873768, -8.66482677, -9.67964643, -10.6420155, -11.5346751, -12.3403666, -13.0418309, -13.6218094, -14.0573956, -14.3265315, -14.4254627, -14.3392234, -14.0528478, -13.5513704, -12.8198253, -10.7519076, -8.02303678, -5.39124208, -3.06478227, -1.25191614, -0.160902459, 0, -2.89490659, -6.87163049, -11.873619, -17.8443194, -24.727179, -32.4656451, -44.1354082, -62.3415014, -79.5378213, -94.3960238, -105.587765, -111.784701, -111.658487, -179.78201, -266.811531, -317.657921, -315.467867, -243.388055, -84.5651731, 177.854093, 1055.49709, 1480.67478, 1860.11891, 2600.56127, 4472.5578, 6985.40964, 10982.4341, 17306.9484, 26267.2918, 40311.7159, 68486.5005, 92528.7394, 98412.2179, 97620.7655, 92766.4715, 84466.186, 73852.4758, 61990.28, 49967.5925, 38872.4072},
        {11.9539984, 11.396919, 10.7419251, 9.9961151, 9.16658737, 8.26044025, 7.28477208, 6.24668123, 5.15326603, 4.01162485, 2.82885603, 1.61205793, 0.368328902, -0.895232707, -2.18239559, -3.47918686, -4.77081105, -6.04794703, -7.3012737, -8.52146993, -9.69921461, -10.888297, -12.0075697, -13.0389955, -13.9645373, -14.7661581, -15.4258207, -15.9226181, -16.2360743, -16.3550149, -16.262568, -15.9418619, -15.3760248, -14.548185, -12.3130815, -9.36766573, -6.49430298, -3.90692415, -1.81946011, -0.445841767, 0, -2.83192104, -6.72447656, -11.5965936, -17.367199, -23.95522, -31.2795835, -42.0007638, -58.3672847, -73.7427208, -86.9328428, -96.7434215, -101.980228, -101.449032, -160.189052, -235.106538, -278.094161, -274.410187, -209.312884, -68.0605165, 164.088647, 884.061076, 1259.43462, 1620.23501, 2296.48794, 3967.08945, 6101.52072, 9448.44848, 14756.5395, 22360.0286, 34250.8782, 58000.5016, 78152.0784, 82774.3965, 82064.3976, 78248.386, 71601.9274, 62799.6613, 52870.6706, 42723.2044, 33265.5121},
        {15.6671109, 15.2502665, 14.6526298, 13.8888203, 12.9734576, 11.921161, 10.7465502, 9.46424457, 8.0888636, 6.63502681, 5.11735366, 3.55046366, 1.94897629, 0.327511038, -1.28246908, -2.87699986, -4.45339206, -6.0004714, -7.5070636, -8.96199437, -10.3540894, -11.7138662, -12.9827007, -14.1436609, -15.1798145, -16.0742294, -16.8099733, -17.3781596, -17.7686925, -17.9454023, -17.8880901, -17.576557, -16.9906041, -16.1100324, -13.7676545, -10.6688606, -7.60586307, -4.79307008, -2.4448898, -0.775730382, 0, -2.42321699, -5.8798827, -10.2956087, -15.5960064, -21.7066874, -28.5532632, -38.631874, -54.0524075, -68.5940473, -81.1386877, -90.568223, -95.7645474, -95.6095552, -144.935686, -207.230314, -241.745606, -235.919234, -177.18887, -52.9921887, 149.233139, 762.447176, 1093.2023, 1412.08142, 1989.66745, 3399.62048, 5164.93501, 7919.88365, 12298.739, 18650.8127, 28465.2603, 47623.0852, 63949.4113, 68060.681, 67420.7648, 63840.1625, 57976.2463, 51008.8496, 43181.0877, 35055.4132, 27194.2787},
        {20.7028796, 20.5861279, 20.1539088, 19.4333935, 18.4517532, 17.2361592, 15.8137827, 14.2117952, 12.4573677, 10.5776715, 8.59987803, 6.55115842, 4.45868398, 2.34962596, 0.323808252, -1.63755819, -3.5587619, -5.42749212, -7.23143807, -8.95828898, -10.5957341, -12.1542692, -13.5956273, -14.9043477, -16.06497, -17.0620336, -17.8800781, -18.5192514, -18.9773562, -19.2016102, -19.1702154, -18.8613735, -18.2532866, -17.3241566, -14.8846365, -11.6515239, -8.42999473, -5.43706336, -2.88974419, -1.00505161, 0, -2.21275942, -5.44624247, -9.62451741, -14.6716525, -20.511716, -27.0687761, -36.3999129, -50.2416955, -63.4179179, -74.9865771, -84.00567, -89.5331937, -90.6271451, -131.051248, -181.582493, -209.613846, -205.059116, -157.832113, -57.8466478, 104.98347, 594.229003, 862.120608, 1121.81634, 1586.47426, 2713.72533, 4113.69602, 6292.90951, 9757.88901, 14790.8997, 22571.2387, 37853.8867, 50706.6617, 53429.424, 52491.7453, 49449.0084, 44759.2656, 39356.2846, 33368.3027, 27198.171, 21248.7405},
        {26.2786764, 26.8149455, 26.8307674, 26.3716265, 25.4830075, 24.2103949, 22.5992733, 20.6951271, 18.5434409, 16.1896994, 13.679387, 11.0579882, 8.37098771, 5.66386998, 3.13542106, 0.734146418, -1.60305272, -3.86204835, -6.02871249, -8.08891713, -10.0285343, -11.8312131, -13.4853554, -14.9771402, -16.2927466, -17.4183534, -18.3401396, -19.061775, -19.584301, -19.8520745, -19.8441726, -19.5396726, -18.9176517, -17.9571871, -15.4599763, -12.1499111, -8.84000616, -5.75016203, -3.10027925, -1.11025839, 0, -2.18440409, -5.3852273, -9.51922572, -14.5031554, -20.2537726, -26.6878332, -35.1634238, -46.8220536, -58.1340934, -68.4310726, -77.0445209, -83.305968, -86.5469435, -119.067102, -159.342218, -183.352521, -183.686185, -152.931381, -83.6762811, 31.4909424, 378.75362, 568.315242, 756.471386, 1099.51763, 1936.39386, 2972.8742, 4584.42565, 7146.5152, 10821.9368, 16624.1765, 28601.6773, 38298.484, 39062.8219, 37650.5301, 35460.6162, 32288.2609, 28204.785, 23780.7761, 19416.6161, 15512.6874},
        {30.7209326, 32.3181309, 33.1233752, 33.2056257, 32.6338428, 31.4769868, 29.804018, 27.6838966, 25.1855831, 22.3780376, 19.3302205, 16.1110921, 12.7896127, 9.43474254, 6.36402724, 3.48917116, 0.703061833, -1.97618648, -4.53045955, -6.94164312, -9.19162296, -11.192986, -13.0064874, -14.6235837, -16.0357312, -17.2343863, -18.2110054, -18.9885104, -19.5850951, -19.9169783, -19.9628406, -19.7013622, -19.1112237, -18.1711055, -15.6711163, -12.3418315, -9.00379556, -5.87880157, -3.18864237, -1.15511088, 0, -2.05100706, -5.1331807, -9.17747401, -14.1148401, -19.8762319, -26.3926027, -34.3899995, -44.4749973, -54.6773882, -64.6052914, -73.8668264, -82.0701122, -88.8232684, -115.820738, -148.698752, -171.421702, -179.080132, -166.764588, -129.565612, -62.5737494, 153.460398, 267.996353, 383.268615, 601.511685, 1178.28366, 1837.62182, 2865.74751, 4548.88207, 7038.95973, 11025.0632, 19420.5602, 26011.4796, 25852.1266, 24394.1761, 22732.7513, 20557.3096, 17767.5076, 14827.8707, 12010.3547, 9586.91506},
        {33.2118217, 37.2674517, 40.093284, 41.7904447, 42.4600599, 42.2032557, 41.1211581, 39.3148933, 36.8855873, 33.9343662, 30.5623561, 26.870683, 22.9604731, 18.9328525, 15.2375346, 11.7551279, 8.33985421, 5.02153894, 1.83000748, -1.20491475, -4.05340236, -6.52948658, -8.78104985, -10.7998312, -12.5775698, -14.1060045, -15.3768746, -16.3943759, -17.1608327, -17.6381981, -17.8131533, -17.6723794, -17.2025577, -16.3903692, -14.1693483, -11.1924007, -8.19215066, -5.37069192, -2.93011813, -1.07252293, 0, -1.59714616, -4.1511867, -7.63185005, -12.0088647, -17.251959, -23.3308615, -30.4691984, -38.8524451, -47.8432309, -57.3081937, -67.1139712, -77.1272012, -87.2145214, -108.962979, -134.824838, -156.365179, -171.053327, -176.358606, -169.750341, -148.697858, -43.8984837, 2.44346809, 43.0311626, 130.567765, 428.429069, 716.326467, 1179.76921, 2004.26657, 3330.98416, 5478.46212, 9981.34654, 13433.8445, 13098.0778, 12052.9428, 10942.9317, 9628.42492, 8131.29836, 6611.55752, 5182.11805, 3955.89562},
        {32.9833959, 41.9638812, 49.0019939, 54.2435725, 57.834455, 59.9204799, 60.6474854, 60.1613096, 58.6077911, 56.1327678, 52.8820783, 49.0015607, 44.6370532, 39.9343943, 35.4826326, 31.14723, 26.7600988, 22.3764224, 18.0513838, 13.8401663, 9.79795311, 6.2160221, 2.8808555, -0.184969778, -2.95887684, -5.41828877, -7.54062865, -9.21369604, -10.3387575, -11.1075362, -11.5338454, -11.6314981, -11.4143075, -10.8960868, -9.48107211, -7.56869726, -5.60453816, -3.72709184, -2.07485534, -0.786325716, 0, -0.605736375, -1.88690262, -3.87822766, -6.6144404, -10.1302698, -14.4604447, -19.5041537, -25.1809518, -31.8496605, -39.6000426, -48.5218608, -58.7048776, -70.2388559, -84.6194615, -101.047078, -118.615667, -137.127426, -156.384551, -176.18924, -196.343689, -174.409178, -186.787085, -223.983128, -276.503022, -332.503198, -388.284267, -431.532825, -449.935465, -441.679614, -362.949371, -45.2891724, 163.197081, -19.4728738, -202.899705, -322.640033, -440.529833, -632.727459, -791.822389, -872.984671, -831.384353}
    };
    
    const double b4tab[21][81] = {
        {-76.9329587, -74.9120768, -72.8831959, -70.8465369, -68.8023206, -66.750768, -64.6921, -62.6265373, -60.554301, -58.4756119, -56.3906909, -54.2997588, -52.2030367, -50.1007452, -47.9461158, -45.769095, -43.6031866, -41.4582228, -39.3440357, -37.2704577, -35.247321, -33.4197451, -31.6435908, -29.9100064, -28.2101399, -26.5351394, -24.8761532, -23.2108204, -21.5188101, -19.8235718, -18.1217387, -16.4099441, -14.6848212, -12.943003, -11.4512563, -10.0353269, -8.50051257, -6.78819309, -4.83974807, -2.59655716, 0, 2.76386468, 6.03212818, 9.89720271, 14.4515005, 19.7874337, 25.9974146, 34.7489177, 47.4727453, 60.495156, 73.2690363, 85.2472724, 95.8827507, 104.628358, 205.529346, 338.199126, 435.090883, 476.309414, 441.959521, 312.146002, 66.9736571, -1739.1593, -2089.86491, -1995.19665, -2465.208, -8373.90658, -11200.2837, -16591.1008, -30193.1191, -52541.1941, -104617.805, -252880.445, -373123.584, -369854.392, -339140.269, -301672.616, -256155.659, -206854.116, -157777.913, -113023.84, -76688.683},
        {-61.0665593, -59.5405326, -57.9862446, -56.4062639, -54.803159, -53.1794984, -51.5378507, -49.8807845, -48.2108684, -46.5306708, -44.8427603, -43.1497056, -41.4540751, -39.7584374, -38.0534102, -36.3491224, -34.6566071, -32.9808773, -31.326946, -29.6998262, -28.1045309, -26.6183587, -25.1640537, -23.736646, -22.3311654, -20.9426417, -19.5661048, -18.1863601, -16.7897493, -15.3957502, -14.0035442, -12.6123126, -11.2212368, -9.82949807, -8.82586569, -7.96306628, -6.96433642, -5.74917049, -4.23706286, -2.3475079, 0, 2.5232123, 5.71548762, 9.70743045, 14.6296452, 20.6127364, 27.7873085, 38.0516951, 53.0386074, 68.6518075, 84.304145, 99.4084695, 113.377631, 125.624478, 248.967192, 411.077997, 531.044831, 585.084421, 549.413493, 400.248775, 113.806991, -2210.93178, -2654.26793, -2462.11744, -2880.39627, -8492.70369, -12312.0249, -19589.4638, -35576.1242, -61261.328, -114681.525, -254898.409, -369567.039, -372005.489, -350069.387, -321533.332, -283016.134, -234619.989, -183669.342, -135026.499, -93553.7637},
        {-46.6827849, -44.8401848, -43.166788, -41.64566, -40.2598663, -38.9924727, -37.8265445, -36.7451475, -35.7313471, -34.768209, -33.8387986, -32.9261816, -32.0134235, -31.0835898, -30.014852, -28.8566321, -27.666292, -26.4483527, -25.207335, -23.9477595, -22.6741473, -21.4036985, -20.1265034, -18.8453316, -17.5629529, -16.2821369, -15.0056532, -13.72913, -12.4492681, -11.1859141, -9.94473733, -8.7314073, -7.55159344, -6.41096523, -5.88317934, -5.61459444, -5.20857505, -4.55461355, -3.5422023, -2.06083366, 0, 2.40324205, 5.65496541, 9.91367872, 15.3378906, 22.0861097, 30.3168448, 42.2279784, 59.7109529, 78.0479007, 96.569279, 114.605545, 131.487156, 146.54457, 288.992582, 475.996345, 614.889334, 678.435238, 639.397747, 470.54055, 144.627335, -2932.79777, -3509.49445, -3113.86505, -3274.31192, -7801.70177, -12594.3728, -21919.6626, -40044.909, -67533.3141, -119764.622, -248260.16, -353625.192, -358898.889, -343194.683, -321641.463, -289396.778, -243276.737, -193434.545, -145476.652, -105009.505},
        {-31.8683333, -30.9374697, -30.0361184, -29.1612377, -28.3097862, -27.4787223, -26.6650044, -25.865591, -25.0774406, -24.2975115, -23.5227623, -22.7501514, -21.9766372, -21.1991781, -20.3947106, -19.5728592, -18.7447638, -17.9114784, -17.0740568, -16.2335527, -15.3910199, -14.5491341, -13.7071032, -12.8657571, -12.0259258, -11.188439, -10.3541265, -9.51942752, -8.68144093, -7.85149525, -7.03220309, -6.22617708, -5.43602981, -4.66437391, -4.52962731, -4.64484193, -4.57645592, -4.201124, -3.3955009, -2.03624133, 0, 2.38547115, 5.75691422, 10.2999741, 16.2002957, 23.643524, 32.8153038, 46.2870767, 66.2717898, 87.2503757, 108.439769, 129.056905, 148.318717, 165.442141, 326.167472, 537.020431, 693.43557, 764.659679, 719.939552, 528.521982, 159.653759, -2608.91273, -3214.82827, -3099.76368, -3705.3898, -9026.72777, -14207.2007, -23752.4752, -42168.218, -71709.0263, -123633.775, -239923.275, -335811.755, -345417.838, -331331.384, -307187.809, -273154.191, -232810.415, -189553.475, -146842.712, -108137.47},
        {-20.0302572, -19.6381658, -19.2053708, -18.7358815, -18.2337072, -17.7028571, -17.1473405, -16.5711667, -15.978345, -15.3728846, -14.7587948, -14.1400848, -13.520764, -12.9048416, -12.3275698, -11.7731935, -11.2235926, -10.6763859, -10.1291922, -9.57963022, -9.02531876, -8.41389128, -7.79985521, -7.18773265, -6.58204571, -5.98731649, -5.40806709, -4.86092485, -4.36069817, -3.88296412, -3.42732968, -2.99340183, -2.58078756, -2.18909385, -2.53214569, -3.15773313, -3.5595831, -3.59121517, -3.10614896, -1.95790403, 0, 2.50229889, 6.16408247, 11.1886959, 17.7794844, 26.139793, 36.472967, 51.7037552, 74.3479831, 98.1018087, 122.0636, 145.331724, 167.00455, 186.180444, 359.580345, 586.590006, 754.66385, 830.659851, 781.435983, 573.850219, 174.760534, -2303.86234, -2917.28492, -3028.24439, -3999.47794, -9783.43757, -15354.9138, -25184.2768, -43741.8967, -73989.5527, -124923.387, -233695.183, -322248.095, -329244.738, -315251.626, -293243.744, -262504.614, -225610.988, -185658.156, -145564.631, -108248.927},
        {-12.6902541, -12.0962501, -11.5645809, -11.0876955, -10.6580429, -10.2680723, -9.91023253, -9.5769728, -9.26074213, -8.95398956, -8.64916416, -8.33871498, -8.01509107, -7.67074149, -7.21941031, -6.70333566, -6.17071324, -5.63009056, -5.09001513, -4.55903445, -4.04569602, -3.6434085, -3.26413833, -2.90471312, -2.56196047, -2.23270796, -1.91378321, -1.57505508, -1.19044332, -0.821236911, -0.475209569, -0.160134995, 0.116213103, 0.34606102, -0.359519246, -1.45087998, -2.31167396, -2.76944208, -2.65172521, -1.78606422, 0, 2.55805603, 6.42574797, 11.8198493, 18.9571337, 28.0543746, 39.3283455, 56.4541114, 82.5870872, 109.674877, 136.530078, 161.965285, 184.793094, 203.826103, 387.349059, 626.965857, 801.452096, 874.956242, 811.626764, 575.612131, 131.060809, -2646.10171, -3379.52842, -3502.81122, -4449.542, -9401.98794, -15480.3531, -26216.6228, -45142.7826, -73801.6903, -121692.714, -225579.161, -309758.77, -314876.602, -300584.785, -279611.063, -250946.073, -217184.806, -180241.945, -142264.098, -105397.871},
        {-8.67512051, -8.41984217, -8.11450304, -7.76504885, -7.37742535, -6.95757828, -6.51145338, -6.04499641, -5.56415309, -5.07486918, -4.58309042, -4.09476254, -3.6158313, -3.15224244, -2.75719541, -2.4067431, -2.07336145, -1.75333087, -1.44293172, -1.1384444, -0.836149293, -0.476906124, -0.120069905, 0.230425007, 0.570644256, 0.896653486, 1.20451834, 1.50199322, 1.79507616, 2.05188364, 2.26373529, 2.42195074, 2.51784962, 2.54275157, 1.50268045, 0.0122575862, -1.23931314, -2.059178, -2.25448326, -1.63237517, 0, 2.61669024, 6.65957361, 12.3517224, 19.916209, 29.5761055, 41.5544845, 60.1388366, 89.0059252, 118.660334, 147.67485, 174.622262, 198.075357, 216.606923, 400.555509, 639.834256, 811.354392, 878.555439, 804.876918, 553.758348, 88.6392516, -2073.4964, -2770.75087, -3221.23308, -4643.05199, -10416.5871, -16426.3604, -26485.1848, -44405.8732, -73070.8069, -119084.093, -211386.946, -287977.184, -300412.52, -290833.052, -269475.859, -239172.615, -206071.305, -170460.078, -134624.763, -100851.19},
        {-6.56467003, -6.16903661, -5.75562568, -5.32659038, -4.88408385, -4.43025922, -3.96726963, -3.49726822, -3.02240813, -2.54484249, -2.06672443, -1.5902071, -1.11744363, -0.650587158, -0.199108732, 0.23946785, 0.668172719, 1.08634956, 1.49334205, 1.88849387, 2.2711487, 2.67469139, 3.05972314, 3.42088631, 3.75282326, 4.05017636, 4.30758797, 4.52566205, 4.70410679, 4.82330965, 4.87549239, 4.85287677, 4.74768455, 4.55213748, 3.1606537, 1.25992955, -0.380239742, -1.54308576, -2.01184009, -1.5697343, 0, 3.08044718, 7.69909034, 14.0497285, 22.3261608, 32.7221862, 45.4316038, 65.1716839, 95.9799915, 127.234189, 157.2914, 184.508749, 207.243361, 223.85236, 404.978382, 639.990109, 805.214965, 863.770499, 778.774256, 513.343783, 30.596626, -2021.49672, -2750.28202, -3299.07912, -4811.2079, -10199.2299, -16242.3468, -26206.9516, -43359.4371, -69736.4892, -112293.622, -198712.494, -271662.186, -286963.163, -278845.454, -256704.337, -225749.505, -195124.667, -162423.052, -129175.57, -96913.1306},
        {-5.56124504, -5.05499527, -4.54410271, -4.02898833, -3.51007309, -2.98777796, -2.46252392, -1.93473193, -1.40482298, -0.873218021, -0.340338033, 0.193396014, 0.727563151, 1.26174241, 1.81049577, 2.36392399, 2.91099369, 3.44821924, 3.97211504, 4.47919548, 4.96597494, 5.4534668, 5.91030297, 6.32961438, 6.70453194, 7.02818655, 7.29370913, 7.50304539, 7.65681649, 7.72707617, 7.70337626, 7.57526857, 7.33230493, 6.96403715, 5.21864841, 2.87098488, 0.789858834, -0.781266194, -1.59892669, -1.41965913, 0, 3.51847295, 8.6793345, 15.6411182, 24.5623574, 35.6015858, 48.9173368, 69.6540832, 102.207101, 134.812973, 165.605781, 192.719607, 214.288534, 228.446643, 404.925127, 633.30511, 790.178568, 838.580815, 741.547159, 462.112912, -36.6866138, -2284.05963, -3089.80009, -3601.01998, -4964.8313, -9839.3587, -15644.5593, -25340.7457, -41888.2307, -67517.4983, -107378.347, -183709.23, -249142.889, -267254.155, -263181.338, -244282.62, -216004.911, -186655.952, -155011.43, -123096.606, -92936.7449},
        {-5.27365126, -4.70375745, -4.12317741, -3.53283143, -2.93363981, -2.32652285, -1.71240084, -1.09219407, -0.466822831, 0.162792571, 0.795731847, 1.4310747, 2.06790084, 2.70528998, 3.35530245, 4.00880637, 4.65568726, 5.29236974, 5.91527844, 6.52083796, 7.10547294, 7.70873069, 8.27795759, 8.80362274, 9.27619522, 9.6861441, 10.0239385, 10.2953378, 10.5038039, 10.603245, 10.5779216, 10.4120945, 10.0900245, 9.59597211, 7.53027008, 4.75265703, 2.22762887, 0.222517333, -0.995345823, -1.15862885, 0, 3.83543963, 9.37725417, 16.7425749, 26.048533, 37.4122599, 50.9508867, 72.1330057, 105.623088, 138.74231, 169.43494, 195.645242, 215.317483, 226.395929, 393.567701, 609.294928, 753.418037, 789.775422, 682.205474, 394.546589, -109.362842, -2448.65246, -3295.332, -3774.37078, -5010.73815, -9531.90674, -15003.3467, -24233.018, -40028.8803, -65171.9598, -102551.017, -168472.081, -225525.938, -244736.343, -244336.067, -229668.948, -205209.544, -176713.735, -145681.396, -115063.321, -87810.301},
        {-5.56216828, -4.89659609, -4.22510258, -3.54780906, -2.86483687, -2.17630733, -1.48234175, -0.783061461, -0.0785877887, 0.630957947, 1.34545442, 2.06478032, 2.78881431, 3.51743507, 4.26997829, 5.03401407, 5.7956397, 6.55075409, 7.29525615, 8.02504478, 8.7360189, 9.48666324, 10.2016473, 10.8682266, 11.4736563, 12.005192, 12.4500888, 12.8159221, 13.1072135, 13.2626317, 13.2611816, 13.081868, 12.7036959, 12.1056701, 9.77307445, 6.62984725, 3.71299499, 1.30705091, -0.303451771, -0.833979815, 0, 4.0926515, 9.89657631, 17.4740066, 26.8871747, 38.1983126, 51.4696528, 72.2842673, 105.395656, 137.665093, 166.91317, 190.960483, 207.627624, 214.735188, 366.046031, 560.731102, 686.157209, 708.248243, 592.928098, 306.120667, -186.250159, -2169.11327, -2959.58379, -3540.04735, -4892.88955, -9245.41646, -14509.2179, -23160.5721, -37675.7573, -60261.3122, -94202.7332, -155991.43, -208832.763, -224134.522, -222840.103, -210216.716, -188897.934, -162270.56, -133342.115, -105249.421, -81129.3001},
        {-6.81819222, -6.11133366, -5.37916772, -4.62300956, -3.84417433, -3.04397719, -2.22373331, -1.38475784, -0.528365933, 0.34412724, 1.23140652, 2.13215676, 3.04506279, 3.96880946, 4.92891572, 5.90709105, 6.88301372, 7.84987985, 8.8008856, 9.72922711, 10.6281005, 11.5449591, 12.4112486, 13.2126719, 13.9349318, 14.5637312, 15.0847729, 15.5091816, 15.8442621, 16.0149308, 15.9965684, 15.7645555, 15.2942731, 14.5611019, 11.9296181, 8.39420326, 5.07116982, 2.26537526, 0.281677096, -0.575067161, 0, 4.52295874, 10.7820589, 18.7767731, 28.5065737, 39.9709333, 53.1693243, 73.8052084, 106.724954, 138.289145, 166.181252, 188.084742, 201.683088, 204.659757, 342.470831, 519.318287, 628.218366, 636.62887, 512.007604, 221.812371, -266.499026, -1911.56176, -2686.45663, -3397.68549, -4851.75019, -8873.56267, -13757.553, -21532.3054, -34226.4041, -52759.5347, -82486.9772, -144214.754, -195045.015, -201335.676, -197880.8, -190780.071, -176865.566, -152524.57, -125191.137, -98533.3628, -76219.3402},
        {-8.50563467, -7.75712742, -6.95769135, -6.11052505, -5.21882713, -4.28579619, -3.31463085, -2.30852971, -1.27069137, -0.204314439, 0.887402479, 2.00126078, 3.13406185, 4.2826071, 5.46637267, 6.66781582, 7.86767744, 9.05812097, 10.2313099, 11.3794076, 12.4945777, 13.6589526, 14.7623014, 15.7843622, 16.7048731, 17.5035721, 18.1601975, 18.6635354, 19.0010128, 19.1307322, 19.0287581, 18.6711545, 18.0339858, 17.0933162, 14.2022937, 10.3636453, 6.70294558, 3.52821283, 1.14746542, -0.131278303, 0, 4.29175776, 10.2722567, 17.9121975, 27.182281, 38.0532079, 50.495679, 69.9784769, 101.144231, 130.817102, 156.735391, 176.637403, 188.261439, 189.345803, 312.838843, 470.944493, 565.631885, 566.983223, 445.080713, 170.006561, -288.157028, -1883.52972, -2621.96049, -3258.96568, -4550.06164, -7974.90776, -12502.8055, -19758.2359, -31365.6801, -47873.2169, -74134.5337, -128055.042, -173488.059, -182501.555, -180156.964, -171629.173, -156782.909, -136018.94, -112904.031, -89925.6229, -69571.1584},
        {-11.1687064, -10.3817552, -9.50984147, -8.55865164, -7.53387206, -6.44118915, -5.28628928, -4.07485884, -2.81258425, -1.50515187, -0.158248107, 1.22244065, 2.631228, 4.06242757, 5.53237265, 7.02144711, 8.50836803, 9.98294507, 11.4349879, 12.8543061, 14.2307095, 15.6582985, 17.0081887, 18.2557864, 19.3764979, 20.3457297, 21.138888, 21.7312377, 22.0980647, 22.2151141, 22.0578497, 21.6017354, 20.8222351, 19.6948126, 16.5310094, 12.3588927, 8.33248143, 4.76758106, 1.97999707, 0.28553489, 0, 4.27317723, 10.1955007, 17.6913839, 26.6852403, 37.1014835, 48.8645268, 66.8206199, 95.0764473, 121.787751, 144.88052, 162.280745, 171.914416, 171.707522, 280.53568, 419.811704, 501.793056, 499.666481, 386.61872, 135.836519, -279.493379, -1700.41279, -2372.90976, -2959.46396, -4122.55508, -7123.68991, -11147.7518, -17576.0471, -27789.8824, -42285.6309, -65099.3987, -111099.244, -150308.439, -159708.128, -158151.438, -149949.174, -136152.92, -118673.915, -99244.0018, -79656.1808, -61703.4528},
        {-15.0034096, -14.2724615, -13.3902868, -12.3680307, -11.2168388, -9.94785629, -8.57222864, -7.10110119, -5.54561934, -3.91692846, -2.22617391, -0.484501089, 1.29694464, 3.10701791, 4.94583587, 6.79512844, 8.63577297, 10.4543204, 12.2373218, 13.971328, 15.6428901, 17.328053, 18.911514, 20.3674642, 21.6700948, 22.793597, 23.712162, 24.4049746, 24.8504692, 25.0108968, 24.8584212, 24.3652061, 23.5034152, 22.2452121, 18.8383676, 14.3459026, 9.96136703, 6.00963498, 2.81558045, 0.704077447, 0, 4.20045209, 10.0199711, 17.3453243, 26.0632791, 36.0606028, 47.2240627, 63.77577, 89.2863988, 113.276414, 133.872294, 149.200516, 157.387561, 156.559905, 250.571061, 370.6642, 440.079422, 435.317615, 332.879666, 109.266463, -259.021108, -1425.77493, -2020.73533, -2582.30767, -3648.89731, -6332.28436, -9756.55346, -15148.1542, -23733.536, -36053.0643, -55391.4414, -94200.9791, -127087.449, -134450.543, -133052.038, -126571.412, -115494.953, -100976.185, -84692.8819, -68144.0964, -52828.88},
        {-20.0462326, -19.591834, -18.8574763, -17.8657248, -16.639145, -15.200302, -13.5717613, -11.7760882, -9.8358481, -7.77360624, -5.611928, -3.37337873, -1.08052375, 1.24407159, 3.54735161, 5.82603928, 8.07916554, 10.2904016, 12.4434186, 14.5218878, 16.5094805, 18.4470624, 20.2532111, 21.9036989, 23.3742981, 24.6407807, 25.6789191, 26.486672, 27.0586643, 27.3176179, 27.2302965, 26.7634639, 25.8838841, 24.5583208, 20.9964813, 16.2766499, 11.6101121, 7.32302301, 3.74153784, 1.19181178, 0, 3.60418307, 8.78881258, 15.4502655, 23.4849188, 32.7891494, 43.2593342, 58.8619025, 82.9517093, 105.693172, 125.330096, 140.106285, 148.265546, 148.051682, 227.311307, 327.605162, 383.711452, 375.546468, 283.026502, 86.0678438, -235.413214, -1228.99871, -1754.62657, -2253.85336, -3168.23562, -5438.45292, -8274.89496, -12718.0611, -19808.4507, -30115.6316, -46092.8977, -77411.6011, -104063.578, -110627.87, -109379.491, -103312.972, -93537.7931, -82031.5276, -69184.4152, -55927.8351, -43193.1661},
        {-26.6172289, -26.7224939, -26.3390842, -25.5085305, -24.272363, -22.6721124, -20.7493091, -18.5454836, -16.1021664, -13.460888, -10.6631789, -7.7505696, -4.76459053, -1.74677221, 1.14811707, 3.95018146, 6.69809673, 9.37349424, 11.9580054, 14.4332615, 16.7808939, 19.0100211, 21.0709912, 22.9416395, 24.5998012, 26.0233114, 27.1900054, 28.1121784, 28.7969475, 29.1397503, 29.1044301, 28.6548302, 27.754794, 26.3681648, 22.6649347, 17.7437439, 12.8390194, 8.2815219, 4.40201229, 1.53125138, 0, 3.30355828, 8.17110317, 14.4963509, 22.1730178, 31.0948201, 41.155474, 55.6343922, 77.3022458, 97.9571645, 116.118163, 130.304256, 139.034459, 140.827786, 205.93302, 287.492664, 333.21959, 326.961083, 252.564425, 93.876899, -165.25421, -957.708558, -1384.02209, -1791.95534, -2529.26882, -4346.7656, -6598.03722, -10114.4914, -15727.5357, -23898.7166, -36569.0259, -61556.1781, -82536.9717, -86855.432, -85156.2204, -80011.5735, -72196.0449, -63274.5668, -53448.24, -43383.1097, -33745.2212},
        {-33.3425809, -34.5797467, -35.0142483, -34.7149696, -33.7507949, -32.1906082, -30.1032937, -27.5577354, -24.6228174, -21.3674239, -17.8604389, -14.1707467, -10.3672312, -6.51877659, -2.92507814, 0.49099202, 3.82403315, 7.05237166, 10.1543339, 13.1082464, 15.8924355, 18.472886, 20.8419704, 22.9797195, 24.8661641, 26.4813351, 27.8052633, 28.8560246, 29.6459781, 30.0641847, 30.0752278, 29.6436907, 28.7341567, 27.3112092, 23.5246129, 18.4870122, 13.4484249, 8.74464055, 4.71144862, 1.68463861, 0, 3.27487408, 8.11029377, 14.3888438, 21.9931088, 30.8056735, 40.7091226, 53.8891774, 72.1854853, 89.9735684, 106.200864, 119.81481, 129.762842, 134.992399, 187.316708, 252.240115, 291.265256, 292.52631, 244.157451, 134.292858, -48.9332942, -611.195059, -912.713132, -1208.30045, -1752.76996, -3101.57027, -4767.94823, -7367.47672, -11515.7286, -17477.6065, -26920.694, -46499.002, -62319.7035, -63452.8615, -61014.9319, -57316.892, -52034.8045, -45306.6837, -38057.6528, -30944.5013, -24624.0188},
        {-37.4574626, -40.5138186, -42.3515766, -43.0745254, -42.7864534, -41.5911492, -39.5924013, -36.8939983, -33.5997288, -29.8133813, -25.6387443, -21.1796065, -16.5397563, -11.8229823, -7.50072163, -3.444187, 0.503238223, 4.31296495, 7.9564041, 11.4049666, 14.6300634, 17.4882773, 20.0817059, 22.3976185, 24.4232848, 26.1459741, 27.5529558, 28.6894543, 29.5919855, 30.1092422, 30.2049623, 29.8428838, 28.9867447, 27.6002828, 23.8125351, 18.747878, 13.6695669, 8.91661274, 4.82802626, 1.74281839, 0, 3.08152086, 7.74939156, 13.9085616, 21.4639804, 30.3205975, 40.3833624, 52.8391957, 68.6823865, 84.7525449, 100.434099, 115.111476, 128.169105, 138.991414, 182.3011, 235.127472, 271.82489, 284.54963, 265.45797, 206.706186, 100.450556, -248.727491, -430.444526, -610.947882, -956.48489, -1884.44788, -2941.59479, -4595.54455, -7313.91607, -11342.7906, -17814.4, -31506.4558, -42234.9855, -41893.8713, -39432.2948, -36652.8, -33051.8461, -28475.7232, -23677.4838, -19101.0555, -15190.3661},
        {-37.6999664, -44.6073811, -49.6529908, -52.9865426, -54.757784, -55.1164621, -54.2123245, -52.1951182, -49.2145908, -45.4204894, -40.9625614, -35.9905541, -30.6542148, -25.1032908, -19.9894335, -15.1448838, -10.3638909, -5.69354162, -1.18092301, 3.12687795, 7.18277429, 10.6916503, 13.8887022, 16.7610974, 19.2960035, 21.480588, 23.3020182, 24.7764267, 25.9155933, 26.6374274, 26.9173356, 26.7307249, 26.053002, 24.8595739, 21.493007, 16.9616489, 12.3974458, 8.10978864, 4.40806809, 1.60167494, 0, 2.41668115, 6.31100838, 11.6465188, 18.3867495, 26.4952375, 35.9355201, 47.0895956, 60.2765839, 74.4594328, 89.4317708, 104.987226, 120.919427, 137.022002, 171.755359, 213.104453, 247.641248, 271.341234, 280.179905, 270.132754, 237.175274, 68.9326989, -3.94063173, -66.6118504, -204.248089, -682.503503, -1141.34506, -1882.52258, -3207.78589, -5345.72218, -8817.56921, -16127.4291, -21722.5291, -21138.8662, -19403.4913, -17570.8586, -15415.6481, -12978.8405, -10515.9876, -8209.49146, -6241.75435},
        {-32.9009557, -47.1404618, -58.5117179, -67.2253254, -73.4918858, -77.5220004, -79.5262706, -79.7152978, -78.2996834, -75.4900288, -71.4969354, -66.5310046, -60.8028378, -54.5230364, -48.5130929, -42.5971569, -36.5531358, -30.4666784, -24.4234333, -18.5090492, -12.8091746, -7.78310138, -3.0912324, 1.23238631, 5.15370873, 8.63868887, 11.6532807, 14.0420896, 15.667955, 16.7869894, 17.4144182, 17.5654669, 17.255361, 16.4993258, 14.343992, 11.4173216, 8.42059515, 5.56715624, 3.07034838, 1.14351512, 0, 0.972835271, 3.00103908, 6.14331829, 10.4583798, 16.0049303, 22.8416769, 30.8168166, 39.8101783, 50.3838214, 62.6819266, 76.8486743, 93.0282452, 111.36482, 134.169525, 160.215722, 188.112747, 217.561549, 248.263077, 279.918281, 312.228111, 279.105702, 299.57109, 358.81613, 442.032678, 530.434992, 619.026309, 688.225405, 718.451057, 706.837009, 583.657133, 81.7638928, -247.651353, 41.7735923, 332.221774, 521.477762, 706.957138, 1007.80801, 1256.61738, 1383.66396, 1319.22649}
    };
    
    /* Stuff for interpolating the NQC data */
    gsl_spline    *spline = NULL;
    gsl_interp_accel *acc = NULL;
    /* Interpolate the spin NQC data in 2-D parameter space -- spin and mass ratio */
    /* First, interpolate in spin dimension for all mass ratios */
    spline = gsl_spline_alloc( gsl_interp_cspline, adim );
    acc    = gsl_interp_accel_alloc();
    for (i = 0; i < qdim; i++)
    {
        gsl_spline_init( spline, alist, a3stab[i], adim );
        gsl_interp_accel_reset( acc );
        a3slist[i] = gsl_spline_eval( spline, a/(1.0-2.0*eta), acc );
        gsl_spline_init( spline, alist, a4tab[i], adim );
        gsl_interp_accel_reset( acc );
        a4list[i] = gsl_spline_eval( spline, a/(1.0-2.0*eta), acc );
        gsl_spline_init( spline, alist, a5tab[i], adim );
        gsl_interp_accel_reset( acc );
        a5list[i] = gsl_spline_eval( spline, a/(1.0-2.0*eta), acc );
        gsl_spline_init( spline, alist, b3tab[i], adim );
        gsl_interp_accel_reset( acc );
        b3list[i] = gsl_spline_eval( spline, a/(1.0-2.0*eta), acc );
        gsl_spline_init( spline, alist, b4tab[i], adim );
        gsl_interp_accel_reset( acc );
        b4list[i] = gsl_spline_eval( spline, a/(1.0-2.0*eta), acc );
        //  printf("%.15f\n",a3slist[i]);
    }
    //printf("%.15f\n",a);
    gsl_spline_free(spline);
    gsl_interp_accel_free(acc);
    /* Second, interpolate in mass ratio dimension */
    spline = gsl_spline_alloc( gsl_interp_cspline, qdim );
    acc    = gsl_interp_accel_alloc();
    gsl_spline_init( spline, etalist, a3slist, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->a3S = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, a4list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->a4 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, a5list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->a5 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, b3list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->b3 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, b4list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->b4 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_free(spline);
    gsl_interp_accel_free(acc);
    
    /* Interpolate nonspin NQC data in the mass ratio dimension */
    spline = gsl_spline_alloc( gsl_interp_cspline, nsqdim );
    acc    = gsl_interp_accel_alloc();
    gsl_spline_init( spline, nsetalist, a1list, nsqdim );
    gsl_interp_accel_reset( acc );
    coeffs->a1 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, nsetalist, a2list, nsqdim );
    gsl_interp_accel_reset( acc );
    coeffs->a2 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, nsetalist, a3list, nsqdim );
    gsl_interp_accel_reset( acc );
    coeffs->a3 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, nsetalist, b1list, nsqdim );
    gsl_interp_accel_reset( acc );
    coeffs->b1 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, nsetalist, b2list, nsqdim );
    gsl_interp_accel_reset( acc );
    coeffs->b2 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_free(spline);
    gsl_interp_accel_free(acc);
    /* Andrea and I have different sign conventions, so I need to put a minus sign in front */
    coeffs->b1 = - coeffs->b1;
    coeffs->b2 = - coeffs->b2;
#if 0
    coeffs->a1 = -8.02798637014;
    coeffs->a2 = 48.7446843797;
    coeffs->a3 = -45.7900277224;
    coeffs->a3S= 0.;
    coeffs->a4 = 0.;
    coeffs->a5 = 0.;
    coeffs->b1 = 0.834742923041;
    coeffs->b2 = -2.33512320852; // q=1
#endif
#if 0
    coeffs->a1 = -7.79667;
    coeffs->a2 = 47.182;
    coeffs->a3 = 2238.85334023;
    coeffs->a3S= 0.;
    coeffs->a4 = -7143.16738899;
    coeffs->a5 = 5596.0086893;
    coeffs->b1 = 0.85069;
    coeffs->b2 = -2.47071; // q=1, chi1=chi2=0.98
#endif
#if 0
    coeffs->a1 = -6.82562365707;
    coeffs->a2 = 41.5711482044;
    coeffs->a3 = -39.4329799959;
    coeffs->a3S= 0.;
    coeffs->a4 = 0.;
    coeffs->a5 = 0.;
    coeffs->b1 = 0.461925688819;
    coeffs->b2 = -1.38733263299; // q=8
#endif
#if 0
    coeffs->a1 = -7.5758;
    coeffs->a2 = 46.9323;
    coeffs->a3 = -118.368375152;
    //coeffs->a3 = -45.0036; // NS part
    coeffs->a3S= 0.;
    coeffs->a4 = 125.555824111;
    coeffs->a5 = -22.0751068073;
    //coeffs->a4 = 0.;
    //coeffs->a5 = 0.;
    coeffs->b1 = 0.51305;
    coeffs->b2 = -1.55133; // q=8, chi1=0.5
#endif
#if 0
    coeffs->a1 = -5.79216;
    coeffs->a2 = 30.3902;
    coeffs->a3 = 3329.40898295;
    coeffs->a3S= 0.0;
    coeffs->a4 = -11082.7670039;
    coeffs->a5 = 9152.28860893;
    coeffs->b1 = 0.974779;
    coeffs->b2 = -3.34626;
#endif
    
    /* Obsolete polynomial fitting of nonspin NQC coefficients a1, a2, a3, b1 and b2 */
    /*
     coeffs->a1 = -12.67955358602124 + 75.41927959573084 * eta - 106.15933052937714 * eta2;
     coeffs->a2 = 101.45522216901628 - 757.3158549733314 * eta + 1473.314771676588 * eta2;
     coeffs->a3 = -107.6647834845902 + 857.6219519536213 * eta - 1776.2776804623143 * eta2;
     // Andrea and I have different sign conventions, so I need to put a minus sign in front 
     coeffs->b1 = - (-1.464129495621165 + 12.81732978488213 * eta - 60.09957767247623 * eta2);
     coeffs->b2 = - ( 7.477426352542122 - 85.26122117590637 * eta + 353.3251639728075 * eta2);
     */
    
    return XLAL_SUCCESS;
    
}

/**
 * ==================================================
 *
 *   Below are three interpolation functions for
 *   max, med and min possible chiA values with given
 *   eta and chi values.
 *   They provide 2-D interpolation results for the
 *   3-D interpolation function of SEOBNRv2
 *
 * ==================================================
 */

UNUSED static int XLALSimIMRGetEOBCalibratedSpinNQCv2chiAmax( EOBNonQCCoeffs * restrict coeffs,                                                             INT4 UNUSED l,                                                             INT4 UNUSED m,                                                             REAL8 eta,                                                             REAL8 a )
{
    const unsigned int   qdim = 98;
    const unsigned int   adim = 200;
    UINT4 i;
    /* REAL8 eta2 = eta*eta;*/
    REAL8 a3slist[qdim], a4list[qdim], a5list[qdim];
    REAL8 b3list[qdim], b4list[qdim];
    REAL8 chiA;
    
    chiA = a/(1.0-2.0*eta);
    if (chiA < -1.){
        chiA = -1.;
    }
    if (chiA > 0.99){
        chiA = 0.99;
    }
    
    memset( coeffs, 0, sizeof( *coeffs ) );
    const double alist[200] = {-1., -0.99, -0.98, -0.97, -0.96, -0.95, -0.94, -0.93, -0.92, -0.91, -0.9, -0.89, -0.88, -0.87, -0.86, -0.85, -0.84, -0.83, -0.82, -0.81, -0.8, -0.79, -0.78, -0.77, -0.76, -0.75, -0.74, -0.73, -0.72, -0.71, -0.7, -0.69, -0.68, -0.67, -0.66, -0.65, -0.64, -0.63, -0.62, -0.61, -0.6, -0.59, -0.58, -0.57, -0.56, -0.55, -0.54, -0.53, -0.52, -0.51, -0.5, -0.49, -0.48, -0.47, -0.46, -0.45, -0.44, -0.43, -0.42, -0.41, -0.4, -0.39, -0.38, -0.37, -0.36, -0.35, -0.34, -0.33, -0.32, -0.31, -0.3, -0.29, -0.28, -0.27, -0.26, -0.25, -0.24, -0.23, -0.22, -0.21, -0.2, -0.19, -0.18, -0.17, -0.16, -0.15, -0.14, -0.13, -0.12, -0.11, -0.1, -0.09, -0.08, -0.07, -0.06, -0.05, -0.04, -0.03, -0.02, -0.01, 0, 0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09, 0.1, 0.11, 0.12, 0.13, 0.14, 0.15, 0.16, 0.17, 0.18, 0.19, 0.2, 0.21, 0.22, 0.23, 0.24, 0.25, 0.26, 0.27, 0.28, 0.29, 0.3, 0.31, 0.32, 0.33, 0.34, 0.35, 0.36, 0.37, 0.38, 0.39, 0.4, 0.41, 0.42, 0.43, 0.44, 0.45, 0.46, 0.47, 0.48, 0.49, 0.5, 0.51, 0.52, 0.53, 0.54, 0.55, 0.56, 0.57, 0.58, 0.59, 0.6, 0.61, 0.62, 0.63, 0.64, 0.65, 0.66, 0.67, 0.68, 0.69, 0.7, 0.71, 0.72, 0.73, 0.74, 0.75, 0.76, 0.77, 0.78, 0.79, 0.8, 0.81, 0.82, 0.83, 0.84, 0.85, 0.86, 0.87, 0.88, 0.89, 0.9, 0.91, 0.92, 0.93, 0.94, 0.95, 0.96, 0.97, 0.98, 0.99};
    const double etalist[98]    = {0.0075, 0.01, 0.0125, 0.015, 0.0175, 0.02, 0.0225, 0.025, 0.0275, 0.03, 0.0325, 0.035, 0.0375, 0.04, 0.0425, 0.045, 0.0475, 0.05, 0.0525, 0.055, 0.0575, 0.06, 0.0625, 0.065, 0.0675, 0.07, 0.0725, 0.075, 0.0775, 0.08, 0.0825, 0.085, 0.0875, 0.09, 0.0925, 0.095, 0.0975, 0.1, 0.1025, 0.105, 0.1075, 0.11, 0.1125, 0.115, 0.1175, 0.12, 0.1225, 0.125, 0.1275, 0.13, 0.1325, 0.135, 0.1375, 0.14, 0.1425, 0.145, 0.1475, 0.15, 0.1525, 0.155, 0.1575, 0.16, 0.1625, 0.165, 0.1675, 0.17, 0.1725, 0.175, 0.1775, 0.18, 0.1825, 0.185, 0.1875, 0.19, 0.1925, 0.195, 0.1975, 0.2, 0.2025, 0.205, 0.2075, 0.21, 0.2125, 0.215, 0.2175, 0.22, 0.2225, 0.225, 0.2275, 0.23, 0.2325, 0.235, 0.2375, 0.24, 0.2425, 0.245, 0.2475, 0.25};
    
    /* YPtag: chiAmax tables below */
    const double a3stab[98][200] = {{486.38171, 481.380476, 476.404729, 471.453495, 466.525797, 461.620659, 456.737106, 451.874162, 447.03085, 442.206196, 437.399224, 432.608957, 427.834419, 423.074636, 418.32863, 413.595427, 408.87405, 404.163524, 399.462873, 394.771121, 390.087292, 385.41041, 380.739499, 376.073585, 371.41169, 366.752839, 362.096057, 357.440367, 352.784793, 348.12836, 343.470092, 338.809014, 334.144148, 329.47452, 324.760718, 320.021792, 315.277256, 310.527379, 305.772434, 301.012693, 296.248427, 291.479906, 286.707404, 281.931191, 277.151539, 272.368718, 267.583002, 262.794661, 258.003967, 253.211191, 248.416604, 243.63398, 238.849796, 234.064031, 229.276664, 224.487673, 219.697037, 214.904735, 210.110747, 205.31505, 200.517623, 195.718446, 190.917497, 186.114756, 181.310199, 176.503808, 171.69556, 166.908623, 162.165606, 157.418664, 152.666274, 147.906911, 143.139051, 138.361168, 133.571739, 128.76924, 123.952146, 119.118932, 114.268074, 109.398049, 104.507331, 99.5943955, 94.6577192, 89.6957773, 84.6479933, 79.5436111, 74.412584, 69.255304, 64.0721631, 58.8635533, 53.6298665, 48.3714947, 43.0888299, 37.7822641, 32.4521892, 27.0989972, 21.7230801, 16.3248299, 10.9046385, 5.46289784, 0, -5.02236294, -10.0652441, -15.1402674, -20.2590568, -25.4332361, -30.6744295, -35.9942607, -41.4043538, -46.9163327, -52.5418212, -58.2924434, -64.1798232, -70.2155844, -76.4113512, -82.7787473, -89.3293967, -96.2039592, -103.432492, -110.734753, -118.150017, -125.717557, -133.476649, -141.466567, -149.726584, -158.295977, -167.214019, -176.519984, -186.253147, -196.452783, -207.203609, -218.482794, -230.315979, -242.728807, -255.430166, -268.604073, -282.519016, -297.295662, -313.054678, -329.916733, -348.002494, -367.774213, -388.907013, -411.454731, -435.471203, -461.010266, -487.34786, -515.510191, -545.745569, -578.302302, -613.428701, -651.338492, -692.356838, -735.965331, -782.347584, -833.1205, -889.257464, -951.73186, -1023.26493, -1102.68497, -1186.91349, -1271.49453, -1363.04314, -1467.55843, -1587.86861, -1721.42903, -1868.16125, -2025.72243, -2193.44705, -2378.97259, -2585.23989, -2813.96745, -3069.50649, -3355.22098, -3674.47486, -4030.6321, -4546.27556, -5075.74556, -5592.60132, -6070.40207, -6482.70704, -6760.47171, -6948.26154, -7037.8945, -6938.24518, -6673.96329, -6304.25833, -5852.05204, -5340.26615, -4791.82242, -4229.64258, -3723.43883, -3248.96725, -2806.12957, -2394.82752, -2014.96283, -1666.43723, -1349.15247, -1071.01933, -797.885891},
        {459.284623, 454.253762, 449.282376, 444.368215, 439.509034, 434.702585, 429.946621, 425.238895, 420.57716, 415.959168, 411.382673, 406.845427, 402.345183, 397.879695, 393.446715, 389.043995, 384.669289, 380.32035, 375.99493, 371.690782, 367.40566, 363.137315, 358.883502, 354.641972, 350.410479, 346.186775, 341.968614, 337.753748, 333.53993, 329.324914, 325.106451, 320.882294, 316.650198, 312.407913, 308.0665, 303.668882, 299.259024, 294.837492, 290.404852, 285.961672, 281.508516, 277.045952, 272.574545, 268.094862, 263.607469, 259.112932, 254.611818, 250.104693, 245.592123, 241.074674, 236.552912, 232.055218, 227.553742, 223.048445, 218.539293, 214.026247, 209.509272, 204.988331, 200.463386, 195.934402, 191.401341, 186.864168, 182.322844, 177.777334, 173.227601, 168.673608, 164.115319, 159.573738, 155.069365, 150.558768, 146.040544, 141.513293, 136.975614, 132.426108, 127.863372, 123.286006, 118.69261, 114.081783, 109.452123, 104.802231, 100.130705, 95.4361452, 90.7171501, 85.9723192, 81.1339324, 76.2351437, 71.309904, 66.3589647, 61.3830773, 56.3829932, 51.3594639, 46.3132407, 41.2450752, 36.1557188, 31.045923, 25.9164391, 20.7680186, 15.6014131, 10.4173738, 5.2166523, 0, -4.79856953, -9.6042007, -14.4289622, -19.2849229, -24.1841512, -29.1387161, -34.1606862, -39.2621302, -44.4551168, -49.7517147, -55.1639927, -60.7040194, -66.3838636, -72.2155939, -78.2112791, -84.3829879, -90.8101349, -97.5144097, -104.355551, -111.360059, -118.554433, -125.965173, -133.61878, -141.541752, -149.760592, -158.301797, -167.191869, -176.457307, -186.124611, -196.185579, -206.714138, -217.747197, -229.321668, -241.280348, -253.757203, -266.937965, -280.91778, -295.791793, -311.655148, -328.60299, -346.76356, -366.188834, -386.967483, -409.188176, -432.939583, -458.590821, -485.880001, -514.82568, -545.446415, -577.760764, -611.490337, -647.313576, -684.096495, -721.905562, -763.258166, -809.571282, -862.261886, -925.862718, -997.966861, -1072.76838, -1144.14936, -1218.99033, -1304.23846, -1402.25636, -1507.63834, -1618.54448, -1730.2445, -1840.14912, -1956.2671, -2080.61233, -2218.47154, -2368.08232, -2530.32567, -2706.0826, -2896.23412, -3101.33424, -3322.67273, -3561.21234, -3817.91584, -4093.74598, -4455.02698, -4793.78584, -5082.97327, -5169.40169, -5075.95154, -4878.06091, -4595.98235, -4249.96844, -3860.27176, -3447.14486, -3064.62474, -2701.69608, -2359.20272, -2037.98848, -1738.89719, -1462.77268, -1210.45878, -996.8879, -799.369153},
        {434.112712, 429.650054, 425.21708, 420.812653, 416.435632, 412.084878, 407.759251, 403.457613, 399.178823, 394.921743, 390.685232, 386.468151, 382.269361, 378.087722, 373.922095, 369.77134, 365.634318, 361.509889, 357.396914, 353.294253, 349.200767, 345.115316, 341.036761, 336.963963, 332.895781, 328.831077, 324.768711, 320.707544, 316.646435, 312.584246, 308.519837, 304.452068, 300.379801, 296.301895, 292.179441, 288.03098, 283.875504, 279.713101, 275.543856, 271.367856, 267.185187, 262.995934, 258.800186, 254.598026, 250.389543, 246.174821, 241.953948, 237.72701, 233.494092, 229.255281, 225.010663, 220.753673, 216.491193, 212.223453, 207.950684, 203.673116, 199.390979, 195.104504, 190.813922, 186.519462, 182.221356, 177.919833, 173.615124, 169.30746, 164.99707, 160.684186, 156.369037, 152.085015, 147.864712, 143.63997, 139.40887, 135.169492, 130.919917, 126.658226, 122.3825, 118.09082, 113.781266, 109.45192, 105.100861, 100.726171, 96.3259314, 91.8982218, 87.4411235, 82.9527173, 78.3265578, 73.6151875, 68.8724052, 64.0996836, 59.2984958, 54.4703146, 49.6166129, 44.7388635, 39.8385393, 34.9171132, 29.9760582, 25.016847, 20.0409526, 15.0498478, 10.0450055, 5.02789861, 0, -4.78250344, -9.52437487, -14.2407085, -18.9465985, -23.6571392, -28.3874246, -33.1525491, -37.9676069, -42.8476921, -47.807899, -52.8633217, -58.0290545, -63.3201916, -68.7518272, -74.3390555, -80.0969707, -86.0394014, -92.1812609, -98.5405063, -105.131961, -111.970447, -119.070788, -126.447807, -134.116326, -142.09117, -150.38716, -159.019121, -168.001874, -177.350243, -186.97904, -197.039769, -207.577258, -218.636331, -230.268822, -242.516054, -255.41748, -269.015826, -283.353813, -298.474168, -314.419613, -330.542814, -347.786571, -366.32862, -386.346696, -408.018533, -433.562858, -460.606168, -488.815951, -517.859694, -547.404885, -575.706973, -605.57131, -637.925839, -673.414864, -712.103587, -754.317214, -800.380952, -851.282321, -906.533689, -964.924903, -1024.71467, -1088.38563, -1159.99052, -1240.4031, -1325.11042, -1412.07213, -1498.64422, -1582.62985, -1664.04544, -1741.65545, -1812.44563, -1877.23281, -1935.39716, -1986.31885, -2029.37806, -1966.7829, -1919.37863, -1910.83842, -1964.83547, -2105.04296, -2505.7788, -2939.64165, -3365.74267, -3591.49968, -3636.06913, -3581.81302, -3444.72753, -3240.80883, -2986.05309, -2696.45649, -2408.38342, -2116.37128, -1827.32267, -1548.1402, -1285.72647, -1046.98409, -838.81567, -648.023968, -481.654895},
        {408.975106, 404.985559, 401.008521, 397.04356, 393.090241, 389.148133, 385.2168, 381.295811, 377.384731, 373.483128, 369.590568, 365.706617, 361.830843, 357.962812, 354.10209, 350.248245, 346.400843, 342.559451, 338.723635, 334.892962, 331.066998, 327.245311, 323.427467, 319.613033, 315.801575, 311.992661, 308.185855, 304.380727, 300.576841, 296.773765, 292.971065, 289.168309, 285.365062, 281.560892, 277.767947, 273.979239, 270.187627, 266.392269, 262.592324, 258.786949, 254.975304, 251.156546, 247.329833, 243.494325, 239.64918, 235.793555, 231.926609, 228.0475, 224.155388, 220.249429, 216.328783, 212.332441, 208.321031, 204.295018, 200.254864, 196.201031, 192.133981, 188.054179, 183.962085, 179.858163, 175.742876, 171.616685, 167.480055, 163.333446, 159.177323, 155.012147, 150.838381, 146.680221, 142.561291, 138.433108, 134.294597, 130.14468, 125.982282, 121.806328, 117.615741, 113.409446, 109.186367, 104.945427, 100.685552, 96.4056654, 92.1046909, 87.781553, 83.4351759, 79.0644836, 74.6035898, 70.085187, 65.542747, 60.977297, 56.3898645, 51.7814768, 47.1531613, 42.5059454, 37.8408564, 33.1589218, 28.4611688, 23.748625, 19.0223176, 14.283274, 9.53252162, 4.77108783, 0, -4.44877624, -8.87861571, -13.3025604, -17.7336522, -22.1849332, -26.6694454, -31.2002306, -35.790331, -40.4527883, -45.2006448, -50.0469422, -55.0047226, -60.087028, -65.3069003, -70.6773815, -76.2115136, -81.9965942, -88.056273, -94.2356332, -100.563629, -107.069213, -113.78134, -120.728965, -127.94104, -135.446519, -143.274357, -151.453507, -160.012924, -168.981561, -178.44686, -188.357841, -198.725912, -209.562479, -220.6587, -232.136107, -244.174965, -256.852758, -270.246967, -284.435074, -299.494563, -315.568962, -332.649605, -350.801054, -370.087868, -390.574608, -412.663888, -435.997701, -460.556093, -486.319112, -513.266803, -541.083953, -570.406743, -601.687646, -635.134169, -670.453687, -707.578124, -746.439403, -786.107851, -827.572808, -872.763539, -924.275341, -979.286962, -1037.08167, -1096.52702, -1155.78615, -1213.34554, -1268.72741, -1320.68676, -1364.18089, -1396.31527, -1419.37522, -1424.48952, -1406.97064, -1362.13106, -1285.28326, -966.597996, -657.814899, -405.53188, -256.34685, -256.857717, -691.291436, -1210.19751, -1758.33283, -2094.71786, -2237.63017, -2282.73748, -2244.44785, -2137.16933, -1975.30997, -1773.27782, -1566.24636, -1344.07188, -1114.56605, -885.540522, -664.806968, -460.177048, -279.462428, -103.363394, 51.3919587},
        {385.836919, 382.594083, 379.328095, 376.039662, 372.72949, 369.398284, 366.046749, 362.675593, 359.28552, 355.877237, 352.451449, 349.008862, 345.550182, 342.076115, 338.587366, 335.084641, 331.568647, 328.040088, 324.499671, 320.948102, 317.386086, 313.814329, 310.233537, 306.644416, 303.047671, 299.444009, 295.834135, 292.218755, 288.598575, 284.9743, 281.346637, 277.716291, 274.083968, 270.450374, 266.878819, 263.33739, 259.793421, 256.245586, 252.69256, 249.133016, 245.56563, 241.989075, 238.402024, 234.803154, 231.191137, 227.564648, 223.922361, 220.26295, 216.585089, 212.887453, 209.168716, 205.329318, 201.468297, 197.586457, 193.6846, 189.763531, 185.824053, 181.86697, 177.893086, 173.903204, 169.898128, 165.878662, 161.845609, 157.799773, 153.741958, 149.672967, 145.593604, 141.526206, 137.49259, 133.449152, 129.395299, 125.330441, 121.253984, 117.165336, 113.063905, 108.9491, 104.820327, 100.676995, 96.5185109, 92.3442836, 88.1537205, 83.9462293, 79.721218, 75.4780944, 71.1764429, 66.8364207, 62.4786636, 58.1038718, 53.7127454, 49.3059846, 44.8842897, 40.4483608, 35.998898, 31.5366016, 27.0621718, 22.5763086, 18.0797124, 13.5730832, 9.0571213, 4.53252684, 0, -4.01001769, -8.04263381, -12.1076946, -16.2150461, -20.3745348, -24.5960067, -28.8893081, -33.2642853, -37.7307843, -42.2986514, -46.9777329, -51.777875, -56.7089238, -61.7807256, -67.0031266, -72.385973, -78.0639739, -84.0648132, -90.1159089, -96.2538637, -102.51528, -108.936761, -115.554908, -122.406324, -129.527613, -136.955375, -144.726215, -152.876734, -161.443535, -170.594144, -180.186234, -190.217133, -200.684165, -211.263568, -222.113212, -233.476591, -245.447357, -258.119162, -271.585657, -285.940497, -301.939279, -318.812247, -336.523541, -355.037303, -374.317673, -393.17015, -413.007177, -434.082556, -456.650089, -480.963577, -507.969469, -536.38235, -565.40954, -594.778108, -625.276269, -657.215805, -690.908498, -727.414323, -766.126826, -805.623337, -845.475757, -885.016093, -925.047482, -964.851461, -1001.13131, -1031.77376, -1055.73562, -1071.18106, -1072.35067, -1055.70445, -1037.98114, -992.242162, -907.927943, -774.478902, -581.335462, 16.0848818, 610.813051, 1129.90289, 1500.40825, 1649.38298, 1162.48788, 535.765197, -157.419908, -607.522181, -834.664429, -957.377837, -990.864791, -950.327681, -850.968894, -707.990819, -549.66023, -383.275639, -212.34673, -40.383187, 129.105306, 292.609066, 446.618408, 589.86229, 726.643015},
        {365.912682, 363.744983, 361.490085, 359.150581, 356.729068, 354.22814, 351.650393, 348.998422, 346.274821, 343.482187, 340.623115, 337.700198, 334.716034, 331.673217, 328.574342, 325.422004, 322.218798, 318.967321, 315.670166, 312.32993, 308.949206, 305.530592, 302.076681, 298.590069, 295.07335, 291.529122, 287.959977, 284.368512, 280.757322, 277.129003, 273.486148, 269.831353, 266.167215, 262.496327, 258.931166, 255.417075, 251.898077, 248.373202, 244.841479, 241.301937, 237.753606, 234.195516, 230.626694, 227.046171, 223.452976, 219.846139, 216.224688, 212.587654, 208.934065, 205.26295, 201.57334, 197.758823, 193.926154, 190.076649, 186.211622, 182.33239, 178.440266, 174.536567, 170.622607, 166.699701, 162.769164, 158.832312, 154.89046, 150.944922, 146.997015, 143.048052, 139.099349, 135.200162, 131.398595, 127.59709, 123.793854, 119.987095, 116.175018, 112.355833, 108.527745, 104.688963, 100.837694, 96.972144, 93.0905217, 89.191034, 85.2718882, 81.3312915, 77.3674512, 73.3785747, 69.2579166, 65.0583681, 60.8340818, 56.5866708, 52.3177484, 48.0289278, 43.7218223, 39.398045, 35.0592091, 30.706928, 26.3428146, 21.9684824, 17.5855445, 13.195614, 8.80030435, 4.4012286, 0, -3.83440188, -7.67924906, -11.5453874, -15.4436627, -19.3849207, -23.3800074, -27.4397685, -31.5750499, -35.7966974, -40.1155568, -44.542474, -49.0882947, -53.7638648, -58.5800302, -63.5476366, -68.6775299, -73.9989029, -79.5252225, -85.2258353, -91.1155185, -97.2090496, -103.521206, -110.066765, -116.860503, -123.917199, -131.251629, -138.878571, -146.812803, -155.0691, -163.579717, -172.472214, -181.786126, -191.560988, -201.93223, -212.891439, -224.404631, -236.482572, -249.136028, -262.375766, -276.212551, -290.396536, -305.278419, -320.919957, -337.382905, -354.729019, -373.649022, -393.418463, -413.941855, -435.123714, -456.868554, -478.611292, -501.299992, -524.861263, -549.480354, -575.870577, -604.508156, -635.869314, -673.771219, -714.589845, -751.056497, -775.293263, -794.223421, -815.566468, -841.035697, -862.167215, -875.168627, -877.214619, -864.763519, -830.727682, -770.025264, -710.743517, -609.528568, -449.816964, -215.045253, 111.350017, 1023.44582, 1940.91371, 2760.93875, 3380.70601, 3697.40056, 3148.27271, 2397.06547, 1538.07881, 959.311907, 638.808791, 432.238834, 323.764805, 297.549475, 337.755613, 428.545989, 565.188595, 699.5487, 831.428743, 960.63116, 1086.95839, 1210.21287, 1330.19703, 1426.82895, 1528.14677},
        {346.268512, 344.320946, 342.299719, 340.206661, 338.043602, 335.812371, 333.5148, 331.152718, 328.727956, 326.242343, 323.697709, 321.095886, 318.438702, 315.727989, 312.965575, 310.153292, 307.292969, 304.386437, 301.435525, 298.442065, 295.407885, 292.334816, 289.224689, 286.079333, 282.900578, 279.690255, 276.450193, 273.182223, 269.888176, 266.56988, 263.229167, 259.867866, 256.487807, 253.090821, 249.745437, 246.418732, 243.076983, 239.719855, 236.347013, 232.958124, 229.552853, 226.130865, 222.691826, 219.235403, 215.761259, 212.269062, 208.758476, 205.229168, 201.680803, 198.113046, 194.525563, 190.830217, 187.116379, 183.38562, 179.639507, 175.87961, 172.107498, 168.32474, 164.532904, 160.733561, 156.928278, 153.118625, 149.306171, 145.492484, 141.679135, 137.867692, 134.059723, 130.316305, 126.697084, 123.080902, 119.46547, 115.848499, 112.227701, 108.600788, 104.96547, 101.319459, 97.6604674, 93.9862054, 90.294385, 86.5827175, 82.8489144, 79.0906872, 75.3057473, 71.4918061, 67.5080891, 63.4244637, 59.3124615, 55.1742882, 51.0121494, 46.8282506, 42.6247973, 38.4039953, 34.16805, 29.919167, 25.6595519, 21.3914104, 17.1169478, 12.83837, 8.55788232, 4.27769047, 0, -3.76442042, -7.5164676, -11.2683326, -15.0322065, -18.8202805, -22.6447455, -26.5177926, -30.451613, -34.4583978, -38.5503379, -42.7396244, -47.0384486, -51.4590013, -56.0134738, -60.7140571, -65.5729422, -70.5463537, -75.6384873, -80.9881254, -86.5954661, -92.4607077, -98.5840484, -104.965686, -111.60582, -118.504648, -125.662367, -133.079177, -140.755275, -148.69086, -156.670776, -164.989539, -173.711952, -182.90282, -192.945669, -203.745941, -215.12409, -227.049303, -239.490769, -252.417676, -265.799212, -278.981408, -292.746267, -307.184898, -322.388412, -338.447918, -356.641011, -375.575696, -395.046462, -414.847799, -434.774197, -454.029044, -473.720387, -493.765921, -514.32411, -536.044971, -559.357828, -584.691998, -615.632875, -648.736418, -677.115599, -694.263363, -704.651876, -713.71344, -722.021966, -721.919103, -709.525243, -682.984759, -638.942778, -566.623171, -459.44769, -361.44604, -207.646184, 26.0482623, 363.733683, 829.506461, 2084.01239, 3355.66109, 4509.4116, 5410.22295, 5923.05416, 5328.54282, 4465.51705, 3449.35935, 2745.38025, 2330.44908, 2035.63185, 1845.08805, 1742.97718, 1713.45876, 1740.69228, 1824.24003, 1909.81087, 1996.91077, 2085.04572, 2173.72167, 2262.44459, 2350.72047, 2412.16589, 2463.52166},
        {326.197572, 323.555743, 320.934535, 318.332357, 315.747622, 313.178738, 310.624118, 308.082173, 305.551312, 303.029947, 300.516489, 298.009348, 295.506936, 293.007663, 290.509939, 288.012177, 285.512786, 283.010177, 280.502762, 277.98895, 275.467154, 272.935783, 270.393249, 267.837962, 265.268333, 262.682773, 260.079693, 257.457504, 254.814616, 252.14944, 249.460387, 246.745869, 244.004295, 241.234076, 238.377931, 235.463289, 232.518246, 229.543023, 226.537835, 223.502902, 220.438443, 217.344674, 214.221814, 211.070081, 207.889694, 204.680871, 201.443829, 198.178788, 194.885964, 191.565577, 188.217844, 184.786741, 181.329947, 177.848902, 174.345041, 170.819801, 167.274622, 163.710939, 160.13019, 156.533813, 152.923245, 149.299923, 145.665285, 142.020768, 138.36781, 134.707847, 131.042318, 127.417191, 123.877367, 120.33244, 116.780959, 113.221476, 109.652541, 106.072704, 102.480516, 98.8745278, 95.2532895, 91.6153517, 87.9592651, 84.2835803, 80.5868476, 76.8676177, 73.1244412, 69.3558685, 65.4528234, 61.4699337, 57.4631206, 53.4344274, 49.3858974, 45.3195737, 41.2374997, 37.1417184, 33.0342733, 28.9172075, 24.7925642, 20.6623868, 16.5287184, 12.3936023, 8.25908168, 4.12719984, 0, -3.6889738, -7.35474513, -11.0094906, -14.6653868, -18.3346103, -22.0293377, -25.7617457, -29.5440108, -33.3883095, -37.3068186, -41.3117147, -45.4151742, -49.6293738, -53.9664901, -58.4386998, -63.0581793, -67.799791, -72.6703808, -77.7665268, -83.0924097, -88.6522102, -94.4501089, -100.490286, -106.776924, -113.314201, -120.106299, -127.157399, -134.471681, -142.053326, -149.753532, -157.785555, -166.199472, -175.045358, -184.512954, -194.582503, -205.197003, -216.364629, -228.093559, -240.391966, -253.268027, -266.854795, -280.997562, -295.680072, -310.886066, -326.59929, -342.327363, -358.649182, -375.66752, -393.48515, -412.204846, -432.804075, -453.441848, -472.974776, -490.848834, -507.713278, -523.677115, -538.849354, -553.003007, -566.659438, -580.706546, -600.418643, -615.988203, -619.66566, -606.691417, -577.372217, -529.689272, -465.922891, -385.168865, -270.759621, -114.944237, 19.4044312, 221.526282, 523.61893, 957.879993, 1556.50709, 3138.402, 4752.38213, 6233.96904, 7418.68432, 8142.04954, 7547.37844, 6623.87229, 5503.71317, 4711.04941, 4222.90977, 3853.00369, 3586.52589, 3408.67107, 3304.63392, 3259.60917, 3266.86238, 3286.26722, 3316.23313, 3355.16955, 3401.48592, 3453.5917, 3509.89631, 3544.32202, 3533.28461},
        {308.114675, 305.49012, 302.91027, 300.372586, 297.87453, 295.413563, 292.987146, 290.59274, 288.227806, 285.889804, 283.576197, 281.284446, 279.012011, 276.756353, 274.514934, 272.285215, 270.064657, 267.850721, 265.640867, 263.432558, 261.223255, 259.010417, 256.791507, 254.563986, 252.325314, 250.072954, 247.804365, 245.517009, 243.208347, 240.875841, 238.516951, 236.129138, 233.709864, 231.256589, 228.673016, 226.005457, 223.301351, 220.561204, 217.785519, 214.974798, 212.129548, 209.25027, 206.337468, 203.391648, 200.413311, 197.402963, 194.361107, 191.288246, 188.184885, 185.051526, 181.888675, 178.663712, 175.410982, 172.131706, 168.827107, 165.498405, 162.146823, 158.773583, 155.379906, 151.967015, 148.536131, 145.088475, 141.625271, 138.147738, 134.657101, 131.154579, 127.641395, 124.154, 120.728, 117.291958, 113.844812, 110.3855, 106.912961, 103.426133, 99.9239536, 96.4053607, 92.8692925, 89.314687, 85.7404824, 82.1456167, 78.5290279, 74.889654, 71.2264332, 67.5383034, 63.7235088, 59.8334523, 55.9207472, 51.9875993, 48.0362143, 44.0687979, 40.0875561, 36.0946944, 32.0924187, 28.0829348, 24.0684484, 20.0511652, 16.0332911, 12.0170317, 8.0045929, 3.99818041, 0, -3.51270233, -7.00628527, -10.492272, -13.9821857, -17.4875495, -21.0198866, -24.5907201, -28.2115733, -31.8939693, -35.6494312, -39.4894823, -43.4256457, -47.4694446, -51.6324021, -55.9260415, -60.3618858, -64.9164092, -69.5961278, -74.4918426, -79.607566, -84.9473108, -90.5150897, -96.3149153, -102.3508, -108.626757, -115.146799, -121.914937, -128.935186, -136.211558, -143.62718, -151.351275, -159.42412, -167.885992, -176.814604, -186.231515, -196.148306, -206.594024, -217.597719, -229.188437, -241.395227, -254.705665, -268.550719, -282.869725, -297.602018, -312.686936, -327.546431, -342.766568, -358.416029, -374.563495, -391.277648, -409.411937, -427.291117, -444.587889, -460.861889, -475.441919, -487.76042, -497.249834, -500.648426, -500.695128, -503.067976, -519.267776, -532.103367, -527.28947, -497.329947, -446.232487, -372.724335, -281.404686, -172.523142, -24.7988187, 170.870306, 344.79381, 596.988077, 963.829436, 1481.69422, 2186.95876, 3978.12552, 5813.91317, 7515.16649, 8902.73028, 9797.44933, 9282.47393, 8408.13971, 7311.44628, 6528.36449, 6037.96779, 5653.0916, 5361.12129, 5149.4422, 5005.4397, 4916.49912, 4934.71915, 4929.30355, 4908.7493, 4881.55336, 4856.21271, 4841.22431, 4845.08515, 4814.21138, 4731.18224},
        {291.464229, 289.590747, 287.711549, 285.825864, 283.932921, 282.031951, 280.122181, 278.202843, 276.273165, 274.332376, 272.379707, 270.414386, 268.435643, 266.442707, 264.434808, 262.411175, 260.371037, 258.313625, 256.238167, 254.143893, 252.030032, 249.895814, 247.740468, 245.563223, 243.363309, 241.139956, 238.892393, 236.619848, 234.321553, 231.996735, 229.644625, 227.264452, 224.855445, 222.416834, 219.918973, 217.376136, 214.802946, 212.199567, 209.566167, 206.902913, 204.209969, 201.487504, 198.735684, 195.954674, 193.144642, 190.305754, 187.438176, 184.542075, 181.617618, 178.66497, 175.684298, 172.645848, 169.580356, 166.488637, 163.371506, 160.229778, 157.064268, 153.875791, 150.665162, 147.433197, 144.180711, 140.908518, 137.617434, 134.308274, 130.981853, 127.638985, 124.280487, 120.9292, 117.607439, 114.270588, 110.918035, 107.549167, 104.16337, 100.760033, 97.338541, 93.8982827, 90.4386447, 86.9590143, 83.4587785, 79.9373245, 76.3940394, 72.8283103, 69.2395245, 65.627069, 61.8995717, 58.1037084, 54.2872458, 50.4525161, 46.601852, 42.7375858, 38.8620499, 34.9775769, 31.0864991, 27.191149, 23.2938591, 19.3969618, 15.5027896, 11.6136748, 7.73195, 3.85994758, 0, -3.1252691, -6.27211294, -9.4491745, -12.6650967, -15.9285226, -19.2480951, -22.6324573, -26.0902519, -29.6301221, -33.2607109, -36.9906611, -40.8286157, -44.7832178, -48.8631102, -53.0769361, -57.4333382, -61.9443102, -66.6189738, -71.4583931, -76.4719292, -81.6689429, -87.0587953, -92.6508473, -98.4544598, -104.478994, -110.73381, -117.22827, -123.971734, -130.973564, -138.222167, -145.75554, -153.589329, -161.739183, -170.058303, -178.643559, -187.635138, -197.097419, -207.094782, -217.691606, -228.952273, -241.768953, -255.126298, -268.926729, -283.072665, -297.466529, -312.317498, -327.144545, -341.773402, -356.029799, -369.739468, -382.235048, -394.438028, -408.548014, -424.946701, -440.304227, -452.957143, -461.242, -459.288131, -450.595492, -443.253738, -450.17458, -452.88684, -436.456366, -392.983666, -326.48909, -235.521583, -125.3351, 3.78307966, 176.130694, 402.098803, 619.539404, 922.542477, 1346.60876, 1927.23899, 2699.9339, 4536.57608, 6427.18895, 8198.1778, 9675.9479, 10686.9045, 10360.1579, 9684.27172, 8784.49201, 8143.72269, 7745.63684, 7424.88344, 7172.58677, 6979.87114, 6837.86085, 6737.6802, 6867.28986, 6880.66221, 6810.8189, 6690.78158, 6553.57188, 6432.21147, 6359.72199, 6219.91783, 6057.40346},
        {276.767662, 275.730022, 274.623853, 273.450616, 272.211769, 270.908773, 269.543086, 268.116168, 266.629478, 265.084475, 263.482619, 261.825369, 260.114184, 258.350524, 256.535847, 254.671614, 252.759284, 250.800315, 248.796167, 246.7483, 244.658172, 242.527244, 240.356974, 238.148821, 235.904246, 233.624707, 231.311664, 228.966575, 226.590901, 224.1861, 221.753632, 219.294956, 216.811532, 214.304819, 211.829733, 209.359881, 206.868227, 204.354495, 201.818409, 199.259694, 196.678074, 194.073274, 191.445019, 188.793032, 186.117039, 183.416763, 180.69193, 177.942264, 175.167489, 172.36733, 169.541511, 166.653007, 163.739088, 160.800276, 157.837093, 154.850059, 151.839697, 148.806526, 145.751069, 142.673846, 139.575378, 136.456188, 133.316795, 130.157722, 126.979489, 123.782617, 120.567629, 117.349037, 114.141021, 110.915242, 107.671313, 104.40885, 101.127465, 97.8267743, 94.5063905, 91.1659281, 87.8050011, 84.4232236, 81.0202098, 77.5955736, 74.1489292, 70.6798907, 67.1880721, 63.6730875, 60.0456475, 56.3516878, 52.6382129, 48.9077219, 45.162714, 41.4056884, 37.6391442, 33.8655807, 30.0874969, 26.3073921, 22.5277654, 18.7511161, 14.9799432, 11.216746, 7.46402358, 3.7242752, 0, -2.71017731, -5.48868496, -8.34087076, -11.2720825, -14.2876681, -17.3929754, -20.593352, -23.894146, -27.3007051, -30.8183771, -34.4525099, -38.2084513, -42.0915491, -46.1071511, -50.2606052, -54.5572592, -59.0476355, -63.7435354, -68.5481262, -73.4764359, -78.5434927, -83.7643246, -89.1539597, -94.7274262, -100.499752, -106.485965, -112.701095, -119.160167, -125.878212, -132.939408, -140.264277, -147.8471, -155.682161, -163.446221, -171.292323, -179.458181, -188.033336, -197.107326, -206.769689, -217.109964, -229.336671, -242.079808, -255.209984, -268.597806, -282.113881, -296.716176, -310.9161, -324.312422, -336.50391, -347.089332, -353.966323, -360.513947, -370.92978, -386.331649, -401.549553, -414.2366, -422.045895, -418.646212, -406.580521, -392.738336, -386.828858, -374.33128, -344.559455, -291.054386, -214.519743, -112.371413, 11.6324911, 158.431802, 352.221174, 604.041019, 864.384938, 1214.15505, 1685.35027, 2309.96952, 3120.01173, 4910.4328, 6759.53543, 8508.57928, 9998.82403, 11071.5294, 10945.0051, 10498.7607, 9841.03425, 9383.92154, 9115.38379, 8896.7748, 8722.38612, 8586.50931, 8483.43589, 8407.45742, 8675.81367, 8736.46045, 8645.60269, 8459.44534, 8234.19335, 8026.05166, 7891.22522, 7662.38801, 7425.75215},
        {263.768306, 262.768937, 261.711772, 260.59786, 259.428249, 258.203989, 256.926126, 255.595711, 254.213792, 252.781417, 251.299635, 249.769494, 248.192044, 246.568332, 244.899407, 243.186319, 241.430114, 239.631843, 237.792554, 235.913295, 233.995114, 232.039061, 230.046185, 228.017532, 225.954153, 223.857096, 221.727409, 219.566141, 217.37434, 215.153056, 212.903337, 210.626231, 208.322786, 205.994053, 203.677984, 201.356399, 199.010675, 196.640661, 194.24621, 191.827172, 189.383397, 186.914738, 184.421044, 181.902167, 179.357958, 176.788267, 174.192945, 171.571844, 168.924815, 166.251708, 163.552373, 160.795217, 158.012218, 155.203907, 152.370819, 149.513486, 146.63244, 143.728214, 140.80134, 137.852351, 134.88178, 131.89016, 128.878023, 125.845901, 122.794328, 119.723835, 116.634956, 113.543138, 110.463469, 107.365722, 104.249464, 101.114259, 97.9596741, 94.7852746, 91.5906264, 88.3752952, 85.1388468, 81.8808471, 78.6008619, 75.2984568, 71.9731978, 68.6246506, 65.252381, 61.8559548, 58.3377983, 54.7480902, 51.1381769, 47.5107764, 43.8686069, 40.2143865, 36.5508335, 32.880666, 29.2066021, 25.53136, 21.8576578, 18.1882137, 14.5257458, 10.8729724, 7.23261151, 3.60738132, 0, -2.50115979, -5.08526884, -7.75622804, -10.5179383, -13.3743004, -16.3292154, -19.3865841, -22.5503073, -25.8242861, -29.2124212, -32.7186136, -36.3467641, -40.1007736, -43.9845431, -48.0019733, -52.1569653, -56.5122655, -61.0801814, -65.7315106, -70.4827637, -75.3504514, -80.3510843, -85.5011733, -90.8172289, -96.3157617, -102.013283, -107.926302, -114.071331, -120.46488, -127.211262, -134.206991, -141.442237, -148.90717, -156.255988, -163.646846, -171.327493, -179.388892, -187.922003, -197.01779, -206.767213, -218.584719, -230.834985, -243.35003, -255.961874, -268.502534, -281.562142, -294.025077, -305.53383, -315.730891, -324.258751, -329.336232, -333.769533, -341.128284, -352.334978, -362.907548, -370.790465, -373.928199, -367.56547, -352.960079, -334.315018, -318.211212, -293.342736, -252.398213, -189.917237, -103.575767, 10.3898795, 150.461777, 319.034303, 537.867738, 817.767912, 1114.04489, 1501.53808, 2010.26338, 2670.23669, 3511.47392, 5288.24215, 7125.2433, 8871.43049, 10375.7568, 11487.1754, 11467.8649, 11144.7358, 10618.216, 10269.6219, 10088.1963, 9946.14527, 9838.56345, 9760.54557, 9707.18633, 9673.58046, 10006.9833, 10108.8884, 10039.7507, 9860.02498, 9630.16616, 9410.6291, 9261.86865, 9022.93639, 8729.6314},
        {251.80006, 250.779002, 249.715024, 248.608616, 247.460266, 246.270463, 245.039696, 243.768455, 242.457227, 241.106503, 239.71677, 238.288519, 236.822237, 235.318415, 233.77754, 232.200102, 230.586589, 228.937491, 227.253297, 225.534496, 223.781575, 221.995026, 220.175336, 218.322994, 216.438489, 214.522311, 212.574948, 210.59689, 208.588624, 206.550641, 204.483428, 202.387476, 200.263272, 198.111307, 195.946443, 193.76168, 191.549844, 189.310959, 187.045045, 184.752126, 182.432224, 180.085362, 177.711562, 175.310846, 172.883237, 170.428758, 167.947431, 165.439278, 162.904321, 160.342584, 157.754089, 155.112156, 152.444088, 149.750487, 147.031954, 144.28909, 141.522498, 138.732777, 135.92053, 133.086357, 130.230861, 127.354642, 124.458301, 121.542441, 118.607661, 115.654565, 112.683752, 109.713912, 106.763299, 103.795211, 100.809076, 97.8043229, 94.7803807, 91.736678, 88.6726437, 85.5877064, 82.4812951, 79.3528384, 76.2017651, 73.0275041, 69.829484, 66.6071337, 63.3598819, 60.0871574, 56.679948, 53.1941839, 49.6870986, 46.1616401, 42.6207561, 39.0673946, 35.5045035, 31.9350307, 28.3619241, 24.7881315, 21.2166009, 17.6502801, 14.092117, 10.5450596, 7.01205569, 3.49605319, 0, -2.31217964, -4.71973805, -7.22524719, -9.83127901, -12.5404054, -15.3551985, -18.27823, -21.312072, -24.4592964, -27.7224752, -31.1041803, -34.6069836, -38.2334572, -41.9861729, -45.8677027, -49.8806186, -54.0987389, -58.5348136, -63.030262, -67.6029233, -72.2706365, -77.0512405, -81.9625744, -87.0224773, -92.2487882, -97.6593461, -103.27199, -109.104559, -115.174892, -121.599797, -128.261856, -135.149217, -142.250028, -149.227107, -156.231267, -163.500077, -171.119286, -179.174641, -187.75189, -196.93678, -208.271449, -219.942006, -231.749254, -243.493993, -254.977027, -266.341584, -276.960433, -286.548768, -294.821786, -301.494683, -305.315356, -308.148549, -312.512861, -319.00819, -324.317081, -326.880889, -325.140972, -316.620273, -300.88729, -278.512429, -251.960051, -214.199632, -162.154971, -91.7318057, 2.88917623, 126.735461, 280.910571, 469.42467, 710.675784, 1014.92335, 1343.38289, 1762.90298, 2300.71978, 2984.06947, 3840.18824, 5568.01895, 7355.16441, 9060.93411, 10544.6376, 11665.5843, 11740.2078, 11532.6108, 11134.764, 10896.3766, 10807.3978, 10750.3853, 10720.6578, 10713.5338, 10724.3319, 10748.3707, 11135.911, 11285.7595, 11257.3333, 11110.0496, 10903.3255, 10696.5781, 10549.2247, 10320.5479, 9984.66346},
        {240.493549, 239.453152, 238.383991, 237.286012, 236.159163, 235.003389, 233.818636, 232.604851, 231.361979, 230.089968, 228.788763, 227.458311, 226.098558, 224.709449, 223.290932, 221.842952, 220.365457, 218.858391, 217.321701, 215.755334, 214.159236, 212.533352, 210.87763, 209.192015, 207.476453, 205.730892, 203.955276, 202.149553, 200.313669, 198.447569, 196.5512, 194.624509, 192.667441, 190.679942, 188.653612, 186.592745, 184.501738, 182.380808, 180.230173, 178.050049, 175.840653, 173.602203, 171.334916, 169.039009, 166.714699, 164.362203, 161.981739, 159.573523, 157.137774, 154.674707, 152.18454, 149.645715, 147.080696, 144.490173, 141.874835, 139.235371, 136.57247, 133.886821, 131.179114, 128.450037, 125.70028, 122.930533, 120.141483, 117.333821, 114.508235, 111.665416, 108.806051, 105.953035, 103.12873, 100.288027, 97.4301779, 94.5544312, 91.660037, 88.746245, 85.8123049, 82.8574665, 79.8809794, 76.8820934, 73.8600582, 70.8141236, 67.7435393, 64.647555, 61.5254205, 58.3763855, 55.0785882, 51.6943815, 48.287574, 44.8613457, 41.4188767, 37.9633469, 34.4979365, 31.0258253, 27.5501935, 24.0742211, 20.6010882, 17.1339747, 13.6760606, 10.2305261, 6.80055117, 3.38931578, 0, -2.13372714, -4.37428686, -6.7229933, -9.18116057, -11.7501028, -14.4311342, -17.2255687, -20.1347207, -23.1599041, -26.3024331, -29.5636219, -32.9447846, -36.4472353, -40.0722881, -43.8212572, -47.6954567, -51.7796519, -56.0870783, -60.4302918, -64.8284889, -69.300866, -73.8666197, -78.5449464, -83.3550428, -88.3161052, -93.4473302, -98.7679143, -104.297054, -110.053946, -116.166912, -122.506011, -129.057701, -135.80844, -142.44661, -149.107709, -156.00672, -163.219528, -170.822013, -178.890058, -187.499545, -198.23292, -209.200981, -220.184849, -230.965645, -241.324486, -251.013449, -259.849958, -267.622397, -274.119146, -279.128586, -281.97057, -283.474654, -284.980972, -286.72797, -286.704807, -283.910526, -277.344168, -266.987266, -250.633079, -225.003064, -188.240498, -137.809674, -74.8926527, 3.14857217, 105.424897, 238.076056, 404.672735, 610.68862, 871.017233, 1195.2237, 1551.40651, 1996.97692, 2555.8539, 3251.95643, 4109.20347, 5761.35749, 7470.03309, 9106.68837, 10542.7814, 11649.7704, 11805.0373, 11703.5004, 11428.3644, 11297.7013, 11301.748, 11332.8798, 11386.2268, 11456.9187, 11540.0853, 11630.8566, 12056.7754, 12256.1706, 12282.3533, 12188.6348, 12028.3263, 11854.7391, 11721.1845, 11516.7881, 11155.9431},
        {229.479402, 228.484324, 227.469055, 226.433179, 225.376285, 224.297958, 223.197785, 222.075351, 220.930243, 219.762047, 218.57035, 217.354737, 216.114796, 214.850112, 213.560272, 212.244861, 210.903467, 209.535674, 208.141071, 206.719242, 205.269775, 203.792255, 202.286269, 200.751403, 199.187243, 197.593376, 195.969387, 194.314864, 192.629392, 190.912558, 189.163947, 187.383147, 185.569743, 183.723321, 181.817992, 179.866615, 177.882357, 175.865632, 173.816853, 171.736432, 169.624783, 167.482318, 165.309452, 163.106596, 160.874164, 158.612569, 156.322224, 154.003542, 151.656937, 149.28282, 146.881606, 144.437787, 141.968041, 139.473127, 136.953802, 134.410826, 131.844956, 129.25695, 126.647567, 124.017564, 121.3677, 118.698733, 116.011421, 113.306523, 110.584796, 107.846998, 105.093889, 102.352184, 99.6479787, 96.9284915, 94.1927984, 91.4399745, 88.6690954, 85.8792363, 83.0694727, 80.2388799, 77.3865332, 74.511508, 71.6128798, 68.6897237, 65.7411153, 62.7661298, 59.7638426, 56.7333291, 53.5402103, 50.2530956, 46.9421988, 43.610926, 40.2626836, 36.9008777, 33.5289145, 30.1502002, 26.7681411, 23.3861434, 20.0076133, 16.635957, 13.2745807, 9.9268906, 6.59629299, 3.28619405, 0, -1.95629261, -4.03110957, -6.22453144, -8.53663879, -10.9675122, -13.5172321, -16.1858792, -18.9735341, -21.8802772, -24.9061891, -28.0513504, -31.3158416, -34.6997434, -38.2031362, -41.8261007, -45.5687173, -49.5276008, -53.7166222, -57.9175113, -62.1510344, -66.437958, -70.7990486, -75.2550726, -79.8267964, -84.5349864, -89.400409, -94.4438308, -99.6860181, -105.147737, -110.974503, -117.016593, -123.257348, -129.680111, -136.00153, -142.338293, -148.878209, -155.684625, -162.820894, -170.350365, -178.336388, -188.305189, -198.412019, -208.434014, -218.148307, -227.332035, -235.536682, -242.821447, -249.019876, -253.965517, -257.491918, -259.368746, -259.571506, -258.430074, -255.871002, -250.997383, -243.282027, -232.197742, -219.843093, -202.468372, -173.459415, -127.217674, -65.0805684, 8.32581781, 94.370561, 204.581205, 345.822389, 523.44213, 743.911869, 1019.26401, 1358.38535, 1737.12332, 2202.4871, 2774.80019, 3474.38612, 4321.56841, 5879.85202, 7490.58366, 9038.29148, 10407.5036, 11482.7483, 11705.3571, 11698.519, 11536.7035, 11507.1117, 11600.0068, 11717.0139, 11852.8283, 12002.1451, 12159.6594, 12320.0664, 12763.755, 13009.2184, 13098.8133, 13074.8962, 12979.8237, 12855.9525, 12745.6391, 12573.2225, 12208.5652},
        {218.388245, 217.565456, 216.720598, 215.853247, 214.962977, 214.049364, 213.111983, 212.150407, 211.164213, 210.152975, 209.116267, 208.053666, 206.964745, 205.84908, 204.706246, 203.535817, 202.337368, 201.110474, 199.854711, 198.569653, 197.254874, 195.909951, 194.534457, 193.127968, 191.690058, 190.220302, 188.718276, 187.183554, 185.615712, 184.014323, 182.378963, 180.709207, 179.004629, 177.264805, 175.458084, 173.60031, 171.707703, 169.780851, 167.820343, 165.826767, 163.800712, 161.742764, 159.653513, 157.533547, 155.383454, 153.203822, 150.995239, 148.758295, 146.493576, 144.201671, 141.883169, 139.530264, 137.15212, 134.749506, 132.323194, 129.873954, 127.402555, 124.909768, 122.396363, 119.86311, 117.31078, 114.740143, 112.151968, 109.547026, 106.926088, 104.289923, 101.639302, 99.0030363, 96.4092649, 93.8009239, 91.1769659, 88.536343, 85.8780078, 83.2009125, 80.5040095, 77.7862513, 75.0465901, 72.2839783, 69.4973684, 66.6857126, 63.8479633, 60.983073, 58.089994, 55.1676786, 52.0713058, 48.8747387, 45.6535687, 42.4114137, 39.1518918, 35.8786211, 32.5952198, 29.305306, 26.0124976, 22.7204128, 19.4326697, 16.1528864, 12.884681, 9.63167157, 6.3974762, 3.18571297, 0, -1.77036634, -3.67240045, -5.70492672, -7.86676949, -10.1567531, -12.573702, -15.1164405, -17.7837929, -20.5745836, -23.487637, -26.5217775, -29.6758294, -32.948617, -36.3389648, -39.845697, -43.4676382, -47.3151821, -51.4030916, -55.4778315, -59.5621337, -63.6787305, -67.8503539, -72.0997362, -76.4496093, -80.9227055, -85.5417568, -90.3294955, -95.3086536, -100.501963, -106.084467, -111.87074, -117.83782, -123.962747, -129.978898, -135.985144, -142.145328, -148.509587, -155.128059, -162.05088, -169.328187, -178.324316, -187.375231, -196.273942, -204.813461, -212.786798, -219.870233, -226.00269, -231.006364, -234.703449, -236.916139, -237.576758, -236.262765, -232.757626, -226.813969, -218.121463, -206.398044, -191.36165, -176.364399, -156.664095, -123.553974, -69.0566995, 3.07997934, 86.437516, 181.580825, 300.907908, 451.385186, 638.912618, 870.180135, 1155.78802, 1504.12466, 1899.54089, 2378.16069, 2956.69308, 3651.84712, 4480.33185, 5935.09684, 7437.55048, 8885.34168, 10176.1193, 11207.5324, 11484.1704, 11558.7812, 11497.4675, 11558.1233, 11730.9341, 11926.1726, 12138.02, 12360.6577, 12588.2669, 12815.0287, 13251.0282, 13533.9997, 13690.7159, 13747.9493, 13732.4727, 13671.0586, 13590.4798, 13451.4166, 13107.6244},
        {205.34996, 205.123404, 204.840503, 204.501892, 204.108202, 203.660066, 203.158118, 202.60299, 201.995315, 201.335726, 200.624856, 199.863338, 199.051804, 198.190888, 197.281222, 196.32344, 195.318174, 194.266057, 193.167722, 192.023802, 190.83493, 189.601738, 188.32486, 187.004929, 185.642576, 184.238436, 182.793141, 181.307325, 179.781619, 178.216656, 176.613071, 174.971495, 173.292562, 171.576903, 169.822937, 168.032451, 166.207258, 164.348064, 162.455574, 160.530492, 158.573522, 156.585371, 154.566743, 152.518342, 150.440874, 148.335043, 146.201554, 144.041111, 141.85442, 139.642186, 137.405113, 135.149537, 132.870409, 130.568313, 128.243832, 125.897547, 123.530042, 121.1419, 118.733703, 116.306034, 113.859476, 111.394611, 108.912023, 106.412293, 103.896006, 101.363743, 98.8160866, 96.2776041, 93.7722853, 91.251249, 88.7135232, 86.1581358, 83.584115, 80.9904886, 78.3762847, 75.7405313, 73.0822564, 70.4004879, 67.694254, 64.9625825, 62.2045015, 59.4190389, 56.6052229, 53.7620813, 50.7410957, 47.6181709, 44.4709847, 41.3033535, 38.1190934, 34.9220205, 31.7159511, 28.5047012, 25.292087, 22.0819247, 18.8780306, 15.6842206, 12.504311, 9.34211801, 6.20145773, 3.08614634, 0, -1.49017289, -3.13985657, -4.94589407, -6.90512843, -9.01440268, -11.2705598, -13.670443, -16.2108951, -18.8887592, -21.7008784, -24.6440957, -27.7152542, -30.9111968, -34.2287665, -37.6648066, -41.2161598, -45.0239809, -49.1057288, -53.1318275, -57.1300441, -61.1281455, -65.1538986, -69.2350703, -73.3994276, -77.6747373, -82.0887665, -86.669282, -91.4440508, -96.4408398, -101.927074, -107.602987, -113.42445, -119.347331, -125.008968, -130.524497, -136.093995, -141.768891, -147.600614, -153.640595, -159.940261, -167.634581, -175.361673, -182.96097, -190.271902, -197.133904, -204.254843, -210.388607, -215.15752, -218.183905, -219.090087, -216.901052, -212.566539, -206.815237, -199.746377, -190.174046, -177.489334, -161.08333, -144.028901, -121.198594, -83.4484745, -22.2326226, 58.6732198, 155.038018, 267.40436, 404.403476, 570.953776, 768.712759, 1001.75341, 1286.1054, 1631.03503, 2035.47241, 2520.61018, 3099.29256, 3784.36373, 4588.66789, 5933.39336, 7320.95418, 8662.10855, 9867.61462, 10848.2306, 11172.1823, 11321.1151, 11352.5866, 11494.1797, 11736.1503, 11997.9109, 12273.2599, 12555.9956, 12839.9164, 13118.8205, 13432.269, 13709.4578, 13940.1438, 14114.0837, 14221.0344, 14250.7526, 14192.9952, 14087.9459, 13825.7986},
        {194.679794, 194.802502, 194.850728, 194.825666, 194.728506, 194.560441, 194.32266, 194.016357, 193.642722, 193.202947, 192.698223, 192.129742, 191.498696, 190.806275, 190.053671, 189.242076, 188.372682, 187.446678, 186.465259, 185.429613, 184.340934, 183.200413, 182.00924, 180.768608, 179.479708, 178.143731, 176.76187, 175.335314, 173.865257, 172.352889, 170.799402, 169.205987, 167.573836, 165.90414, 164.210435, 162.487479, 160.729886, 158.938445, 157.113949, 155.257187, 153.368951, 151.450032, 149.50122, 147.523306, 145.517082, 143.483339, 141.422866, 139.336456, 137.224898, 135.088985, 132.929506, 130.761393, 128.570991, 126.358784, 124.125255, 121.870889, 119.596171, 117.301584, 114.987615, 112.654745, 110.303461, 107.934246, 105.547585, 103.143962, 100.723862, 98.2877681, 95.8361657, 93.3919045, 90.9772973, 88.5467013, 86.099151, 83.6336808, 81.1493253, 78.645119, 76.1200964, 73.573292, 71.0037402, 68.4104756, 65.7925327, 63.1489459, 60.4787499, 57.780979, 55.0546677, 52.2988507, 49.3606364, 46.3182179, 43.2516142, 40.1647901, 37.0617104, 33.9463399, 30.8226434, 27.6945857, 24.5661314, 21.4412456, 18.3238928, 15.218038, 12.1276458, 9.05668116, 6.00910878, 2.98889346, 0, -1.34403908, -2.85381904, -4.52547418, -6.35513879, -8.33894718, -10.4730336, -12.7535325, -15.176578, -17.7383045, -20.4348463, -23.2623376, -26.2169128, -29.2947062, -32.4918521, -35.8044847, -39.2287384, -42.9107103, -46.867958, -50.7574141, -54.6073479, -58.4460284, -62.301725, -66.2027067, -70.1772428, -74.2536026, -78.460055, -82.8248695, -87.3763151, -92.142661, -97.4162184, -102.864399, -108.43626, -114.080857, -119.483688, -124.725588, -129.957678, -135.208081, -140.504923, -145.876327, -151.350419, -157.611072, -163.831085, -169.910286, -175.748499, -181.24555, -187.785954, -193.413676, -197.65737, -200.045692, -200.107295, -196.652816, -190.806502, -183.740554, -175.591732, -164.382781, -149.085598, -128.672077, -104.36574, -72.3751234, -26.4524395, 38.9185956, 122.868424, 223.489674, 342.223708, 486.18613, 659.887977, 864.872344, 1104.8808, 1394.53736, 1742.3103, 2150.77783, 2635.20449, 3206.09742, 3873.96379, 4649.31074, 5902.70641, 7194.58174, 8445.42864, 9575.73905, 10506.0049, 10850.0597, 11039.4927, 11126.2206, 11317.7368, 11604.3759, 11911.6487, 12233.1311, 12562.3988, 12893.0274, 13218.5927, 13553.0437, 13868.4903, 14149.414, 14380.2963, 14545.6186, 14629.8624, 14617.5093, 14523.5929, 14305.9194},
        {186.065989, 186.304952, 186.466927, 186.553178, 186.56497, 186.503565, 186.370228, 186.166222, 185.892813, 185.551264, 185.142838, 184.6688, 184.130414, 183.528943, 182.865652, 182.141805, 181.358666, 180.517498, 179.619565, 178.666133, 177.658463, 176.597821, 175.485471, 174.322676, 173.110701, 171.850809, 170.544264, 169.192331, 167.796272, 166.357354, 164.876838, 163.35599, 161.796073, 160.198351, 158.577074, 156.92674, 155.241691, 153.52277, 151.770818, 149.986678, 148.171194, 146.325206, 144.449559, 142.545095, 140.612656, 138.653084, 136.667223, 134.655914, 132.620001, 130.560325, 128.47773, 126.390101, 124.280868, 122.150503, 119.999482, 117.828275, 115.637357, 113.427201, 111.198279, 108.951066, 106.686033, 104.403654, 102.104402, 99.7887507, 97.4571725, 95.1101408, 92.7481285, 90.394614, 88.0725229, 85.7348823, 83.3806738, 81.0088793, 78.6184805, 76.2084593, 73.7777973, 71.3254764, 68.8504783, 66.3517849, 63.8283779, 61.279239, 58.7033501, 56.0996929, 53.4672492, 50.8050009, 47.9549584, 44.997891, 42.0164542, 39.0147238, 35.9967757, 32.9666856, 29.9285292, 26.8863825, 23.8443212, 20.8064211, 17.776758, 14.7594076, 11.7584458, 8.77794832, 5.82199101, 2.89464964, 0, -1.31264276, -2.77856369, -4.39432988, -6.15650848, -8.06166659, -10.1063713, -12.2871899, -14.6006893, -17.0434367, -19.6119992, -22.3029439, -25.1128381, -28.0382487, -31.0757429, -34.2218879, -37.4732507, -40.9527289, -44.6749368, -48.3506941, -52.0036389, -55.6574089, -59.3356422, -63.0619766, -66.8600502, -70.7535007, -74.7659662, -78.9210845, -83.2424936, -87.7538314, -92.7043021, -97.8092697, -103.024702, -108.306568, -113.502558, -118.622779, -123.706211, -128.741304, -133.716511, -138.620283, -143.441069, -148.327123, -153.05846, -157.592264, -161.885722, -165.896018, -171.308418, -175.920008, -179.255954, -180.841423, -180.201579, -176.453366, -170.029113, -162.11136, -152.651335, -139.086345, -119.982397, -93.905498, -59.5035933, -15.2421276, 40.5028438, 108.407041, 191.810265, 290.562455, 407.39595, 549.927152, 723.530598, 932.729352, 1182.6772, 1481.64999, 1836.15754, 2244.6906, 2723.01179, 3280.4068, 3926.16131, 4669.56098, 5846.42472, 7057.87167, 8231.55421, 9295.12472, 10176.2356, 10520.8604, 10726.1162, 10840.3746, 11058.0594, 11369.73, 11705.0913, 12057.7457, 12421.2955, 12789.3432, 13155.491, 13615.6403, 14001.8311, 14315.7799, 14559.203, 14733.8169, 14841.3379, 14883.4825, 14788.1318, 14591.022},
        {178.669995, 178.886664, 179.033134, 179.110457, 179.119684, 179.06187, 178.938065, 178.749322, 178.496693, 178.18123, 177.803987, 177.366014, 176.868364, 176.312089, 175.698242, 175.027875, 174.30204, 173.52179, 172.688175, 171.80225, 170.865066, 169.877674, 168.841129, 167.756481, 166.624783, 165.447087, 164.224445, 162.95791, 161.648534, 160.29737, 158.905468, 157.473882, 156.003664, 154.495866, 152.957742, 151.387113, 149.781725, 148.142429, 146.470077, 144.765519, 143.029606, 141.26319, 139.467121, 137.64225, 135.789428, 133.909507, 132.003337, 130.071769, 128.115655, 126.135844, 124.133189, 122.124183, 120.093696, 118.042239, 115.970323, 113.878461, 111.767164, 109.636945, 107.488315, 105.321786, 103.13787, 100.937079, 98.7199242, 96.4869181, 94.2385724, 91.9753989, 89.6979096, 87.431273, 85.2000659, 82.9539473, 80.6918307, 78.4126294, 76.115257, 73.798627, 71.4616527, 69.1032477, 66.7223254, 64.3177993, 61.8885828, 59.4335894, 56.9517325, 54.4419257, 51.9030823, 49.3341158, 46.5721116, 43.7003009, 40.8038602, 37.8869547, 34.9537494, 32.0084095, 29.0551, 26.097986, 23.1412326, 20.189005, 17.2454682, 14.3147873, 11.4011274, 8.50865364, 5.64153109, 2.80392484, 0, -1.34928305, -2.82793657, -4.43373566, -6.16445541, -8.01787094, -9.99175733, -12.0838897, -14.2920431, -16.6139928, -19.0475137, -21.5903809, -24.2403697, -26.995255, -29.8528121, -32.8108159, -35.8670416, -39.0955091, -42.5048857, -45.9204879, -49.3564289, -52.8268221, -56.3457807, -59.927418, -63.5858473, -67.3351819, -71.189535, -75.16302, -79.2697501, -83.5238386, -88.0780262, -92.7569688, -97.5331913, -102.379219, -107.35735, -112.395223, -117.396537, -122.306884, -127.071858, -131.637052, -135.948059, -139.693876, -143.154786, -146.326586, -149.205074, -151.786045, -155.808815, -159.089782, -161.188865, -161.665981, -160.081046, -155.98728, -148.959486, -140.073409, -129.061363, -112.912817, -89.8487221, -58.0900293, -13.8399127, 42.2043153, 107.144132, 177.032573, 259.000627, 353.839309, 465.410729, 602.780794, 772.626783, 983.418274, 1242.29721, 1549.83379, 1910.31556, 2315.68062, 2784.46587, 3325.23367, 3946.54639, 4656.96638, 5765.67766, 6905.46551, 8009.73685, 9011.89863, 9845.35778, 10181.3955, 10390.2949, 10517.8027, 10747.7119, 11071.1256, 11421.6276, 11793.197, 12179.8129, 12575.4545, 12974.1007, 13594.3279, 14061.4246, 14401.1648, 14639.3225, 14801.6714, 14913.9857, 15002.0391, 14905.7752, 14730.5381},
        {170.921216, 171.128672, 171.272048, 171.352225, 171.370085, 171.326511, 171.222383, 171.058585, 170.835996, 170.555501, 170.217979, 169.824314, 169.375387, 168.872079, 168.315273, 167.705851, 167.044694, 166.332684, 165.570703, 164.759632, 163.900355, 162.993752, 162.040705, 161.042096, 159.998807, 158.911721, 157.781717, 156.60968, 155.396489, 154.143028, 152.850178, 151.518821, 150.149838, 148.744112, 147.305916, 145.834365, 144.328533, 142.789191, 141.217111, 139.613065, 137.977824, 136.312161, 134.616847, 132.892654, 131.140353, 129.360717, 127.554517, 125.722524, 123.865512, 121.984251, 120.079513, 118.164485, 116.227256, 114.268326, 112.2882, 110.287378, 108.266365, 106.225662, 104.165772, 102.087198, 99.9904422, 97.8760075, 95.7443962, 93.5961111, 91.4316548, 89.2515298, 87.0562388, 84.8695455, 82.7146551, 80.5440959, 78.3568625, 76.1519498, 73.9283522, 71.6850646, 69.4210816, 67.1353978, 64.827008, 62.4949068, 60.1380889, 57.755549, 55.3462818, 52.9092819, 50.443544, 47.9480628, 45.2606981, 42.4644013, 39.6440605, 36.8038994, 33.948142, 31.0810119, 28.206733, 25.3295292, 22.4536242, 19.5832418, 16.722606, 13.8759404, 11.0474688, 8.24141524, 5.46200336, 2.713457, 0, -1.42620271, -2.94702471, -4.56256943, -6.27294029, -8.07824071, -9.9785741, -11.9740439, -14.0647535, -16.2508063, -18.5323058, -20.9093553, -23.3820584, -25.9505183, -28.6148386, -31.3751226, -34.2314738, -37.1531204, -40.1357551, -43.2493181, -46.4872965, -49.8431777, -53.3104492, -56.882598, -60.5531116, -64.3154773, -68.1631824, -72.0897142, -76.08856, -80.1532071, -84.2538386, -88.4157909, -92.6395426, -96.9255724, -101.530379, -106.326431, -111.117925, -115.828534, -120.381929, -124.701784, -128.71177, -132.11023, -135.114746, -137.693078, -139.812984, -141.442224, -144.440568, -146.410762, -146.847561, -145.245722, -141.100001, -133.442628, -122.796195, -110.88638, -97.7299068, -79.8543924, -55.3539911, -22.3228577, 22.8391052, 79.9584334, 147.013388, 221.861847, 309.275006, 407.824091, 520.444267, 657.465145, 825.822163, 1034.82548, 1292.02619, 1596.26812, 1951.32041, 2351.8642, 2812.15482, 3339.10724, 3939.63648, 4620.65753, 5655.39825, 6717.88256, 7748.44724, 8687.42908, 9475.16486, 9816.17982, 10043.83, 10199.5142, 10447.7345, 10781.1614, 11141.1725, 11523.0029, 11921.8878, 12333.0624, 12751.7619, 13453.9979, 13982.4998, 14363.8439, 14624.6062, 14791.3631, 14890.6908, 14949.1654, 14872.6005, 14863.9416},
        {163.185019, 163.396772, 163.549537, 163.644065, 163.681109, 163.661417, 163.585743, 163.454837, 163.26945, 163.030333, 162.738238, 162.393916, 161.998117, 161.551594, 161.055096, 160.509376, 159.915184, 159.273272, 158.584391, 157.849291, 157.068724, 156.243442, 155.374194, 154.461734, 153.506811, 152.510176, 151.472582, 150.394778, 149.277517, 148.121549, 146.927626, 145.696498, 144.428917, 143.125634, 141.790367, 140.422322, 139.020668, 137.58606, 136.119152, 134.6206, 133.091058, 131.531181, 129.941624, 128.323042, 126.67609, 125.001422, 123.299693, 121.571559, 119.817674, 118.038692, 116.23527, 114.414813, 112.571078, 110.704574, 108.815809, 106.905291, 104.973529, 103.021032, 101.048308, 99.0558649, 97.0442119, 95.0138574, 92.9653096, 90.8990773, 88.8156687, 86.7155925, 84.599357, 82.4896723, 80.4087155, 78.3112059, 76.1962128, 74.0628053, 71.9100528, 69.7370245, 67.5427895, 65.3264173, 63.0869769, 60.8235377, 58.5351688, 56.2209396, 53.8799192, 51.5111769, 49.113782, 46.6868036, 44.0678591, 41.3401395, 38.5887766, 35.8180791, 33.0323556, 30.2359147, 27.4330652, 24.6281156, 21.8253748, 19.0291512, 16.2437537, 13.4734909, 10.7226714, 7.99560385, 5.29659701, 2.6299595, 0, -1.46554655, -2.99829602, -4.60026387, -6.27346552, -8.01991642, -9.84163201, -11.7406277, -13.718919, -15.7785213, -17.9214501, -20.1497208, -22.4653488, -24.8703495, -27.3667385, -29.9565312, -32.6417429, -35.2871385, -37.8751262, -40.7181694, -43.7888726, -47.0598404, -50.5036774, -54.0929882, -57.8003774, -61.5984495, -65.4598092, -69.357061, -73.2628095, -77.1496593, -80.8116888, -84.4654883, -88.1372203, -91.8530473, -96.0000528, -100.423939, -104.874161, -109.268607, -113.525161, -117.56171, -121.296141, -124.685164, -127.596025, -129.939012, -131.624416, -132.562527, -134.626711, -135.273413, -133.922152, -129.992449, -122.903825, -110.965115, -96.0640288, -80.7315622, -65.3911389, -46.163224, -21.1002231, 11.745458, 54.9982182, 109.775028, 176.454526, 256.280049, 350.387841, 455.163557, 572.212486, 711.984289, 880.868915, 1086.98146, 1337.15914, 1631.91362, 1975.33463, 2368.73859, 2819.41526, 3332.4144, 3912.78577, 4565.57912, 5526.90631, 6512.98948, 7471.11284, 8348.56064, 9092.6171, 9442.24641, 9691.93286, 9878.24072, 10143.8616, 10483.3511, 10847.0448, 11231.7093, 11634.1109, 12051.0162, 12479.1916, 13184.342, 13744.4911, 14171.7406, 14478.1921, 14675.9469, 14777.1067, 14793.7731, 14759.0911, 14892.5826},
        {155.621618, 155.872754, 156.067182, 156.20561, 156.288742, 156.317285, 156.291945, 156.213428, 156.082441, 155.899688, 155.665876, 155.381711, 155.0479, 154.665147, 154.23416, 153.755644, 153.230305, 152.65885, 152.041983, 151.380413, 150.674843, 149.925981, 149.134532, 148.301203, 147.426699, 146.511727, 145.556992, 144.563201, 143.53106, 142.461274, 141.35455, 140.211593, 139.033111, 137.819808, 136.577557, 135.304373, 133.998208, 132.659599, 131.289085, 129.887205, 128.454497, 126.991499, 125.498749, 123.976787, 122.42615, 120.847376, 119.241004, 117.607573, 115.947621, 114.261686, 112.550306, 110.811552, 109.048484, 107.261694, 105.451773, 103.619313, 101.764906, 99.8891444, 97.9926196, 96.0759235, 94.139648, 92.1843849, 90.2107261, 88.2192635, 86.2105889, 84.1852941, 82.1439711, 80.1119619, 78.1140143, 76.0996748, 74.0679308, 72.0177695, 69.9481785, 67.8581452, 65.7466569, 63.612701, 61.455265, 59.2733362, 57.0659021, 54.83195, 52.5704674, 50.2804416, 47.9608601, 45.6107103, 43.0596143, 40.3948045, 37.705549, 34.9963313, 32.2716352, 29.5359442, 26.7937419, 24.0495121, 21.3077383, 18.5729041, 15.8494932, 13.1419892, 10.4548758, 7.79263649, 5.15975502, 2.56071498, 0, -1.37537004, -2.81832477, -4.3307171, -5.91439996, -7.57122626, -9.30304893, -11.1117209, -12.9990951, -14.9670244, -17.0173617, -19.15196, -21.3726722, -23.6813513, -26.07985, -28.5700214, -31.1537184, -33.6474542, -36.0266046, -38.7122433, -41.6665075, -44.8515346, -48.2294618, -51.7624264, -55.4125656, -59.1420167, -62.9129171, -66.6874038, -70.4276143, -74.0956858, -77.4099297, -80.6657119, -83.8983175, -87.1430314, -90.7946026, -94.7085841, -98.644672, -102.530312, -106.29295, -109.860031, -113.159003, -116.489382, -119.293302, -121.425414, -122.740365, -123.092805, -124.126249, -123.459264, -120.49928, -114.653731, -105.330048, -90.5331967, -72.7872016, -55.1252852, -38.1888363, -17.7359171, 8.28289641, 41.917028, 85.7286675, 141.137935, 209.00557, 291.212379, 388.411054, 497.996306, 621.343294, 766.571822, 938.705793, 1143.03091, 1384.63894, 1667.66174, 1996.77413, 2380.07503, 2818.28841, 3314.90386, 3873.41094, 4497.29925, 5390.39785, 6305.77194, 7196.82624, 8016.96545, 8719.59427, 9074.45774, 9341.06001, 9551.51705, 9828.67458, 10168.7347, 10530.0956, 10910.8496, 11309.0889, 11722.9061, 12150.3933, 12758.1224, 13303.8715, 13773.328, 14152.1796, 14426.1137, 14580.8177, 14601.9793, 14638.9387, 14711.9123},
        {148.24968, 148.611409, 148.913131, 149.155669, 149.339849, 149.466496, 149.536436, 149.550492, 149.509489, 149.414254, 149.265611, 149.064384, 148.811399, 148.507482, 148.153455, 147.750146, 147.298378, 146.798978, 146.252768, 145.660576, 145.023225, 144.341541, 143.616349, 142.848473, 142.038739, 141.187972, 140.296996, 139.366637, 138.397719, 137.391068, 136.347508, 135.267865, 134.152963, 133.003628, 131.832171, 130.633433, 129.402114, 128.138668, 126.843547, 125.517202, 124.160086, 122.772652, 121.355351, 119.908636, 118.432958, 116.928771, 115.396526, 113.836676, 112.249673, 110.635968, 108.996015, 107.313556, 105.606116, 103.874507, 102.119545, 100.342045, 98.5428196, 96.7226847, 94.8824544, 93.0229429, 91.1449649, 89.2493346, 87.3368667, 85.4083754, 83.4646752, 81.5065807, 79.5349061, 77.5845622, 75.6896409, 73.7806358, 71.8561511, 69.9147909, 67.9551593, 65.9758603, 63.9754982, 61.952677, 59.9060008, 57.8340739, 55.7355002, 53.608884, 51.4528292, 49.2659402, 47.0468209, 44.7940755, 42.3148424, 39.7074857, 37.0726711, 34.4152159, 31.7399378, 29.0516543, 26.355183, 23.6553414, 20.9569471, 18.2648176, 15.5837704, 12.9186232, 10.2741934, 7.65529868, 5.06675649, 2.51338441, 0, -1.05418016, -2.22612794, -3.51371992, -4.91483263, -6.42734264, -8.0491265, -9.77806077, -11.612022, -13.5488868, -15.5865316, -17.7228331, -19.9556678, -22.2829122, -24.7024429, -27.2121365, -29.8098695, -32.3756736, -34.8905905, -37.6190504, -40.5336775, -43.6070959, -46.8119296, -50.1208027, -53.5063393, -56.9411636, -60.3978996, -63.8491714, -67.2676031, -70.6258187, -73.7613784, -76.8314938, -79.8493081, -82.8279647, -85.9766664, -89.2105268, -92.3923768, -95.476542, -98.4173477, -101.169119, -103.686183, -106.568731, -108.928654, -110.59391, -111.392459, -111.152261, -110.932386, -109.021906, -104.941005, -98.2098635, -88.3486654, -74.0077076, -56.6402512, -39.1188619, -21.8516967, -0.215412429, 28.1542932, 65.6217226, 117.17287, 181.954507, 256.253375, 335.874539, 428.249151, 538.107199, 669.089895, 822.993573, 1002.4614, 1208.25616, 1442.53354, 1714.34385, 2028.83739, 2398.44356, 2819.91374, 3295.87855, 3828.96862, 4421.81458, 5255.28808, 6109.21848, 6941.67616, 7710.73148, 8374.45483, 8726.16771, 8998.52195, 9220.1946, 9499.48851, 9833.56217, 10186.3967, 10556.8093, 10943.6176, 11345.6387, 11761.6901, 12136.5486, 12601.4344, 13103.8228, 13591.1893, 14011.0092, 14310.7578, 14437.9105, 14587.95, 14213.5886},
        {141.68755, 142.08308, 142.421239, 142.702748, 142.928326, 143.098696, 143.214577, 143.276691, 143.285757, 143.242497, 143.147632, 143.001882, 142.805967, 142.560609, 142.266528, 141.924445, 141.535081, 141.099156, 140.61739, 140.090505, 139.519222, 138.904261, 138.246342, 137.546187, 136.804515, 136.022049, 135.199508, 134.337613, 133.437085, 132.498644, 131.523012, 130.510909, 129.463055, 128.380171, 127.268074, 126.124828, 124.948439, 123.739463, 122.498454, 121.225967, 119.922559, 118.588785, 117.2252, 115.832358, 114.410817, 112.96113, 111.483854, 109.979544, 108.448755, 106.892042, 105.309961, 103.690292, 102.046644, 100.379847, 98.6907337, 96.9801369, 95.2488886, 93.497821, 91.7277665, 89.9395573, 88.1340256, 86.3120038, 84.4743242, 82.6218189, 80.7553203, 78.8756606, 76.9836721, 75.117715, 73.3152481, 71.4997057, 69.669487, 67.8229916, 65.9586191, 64.0747688, 62.1698404, 60.2422332, 58.2903469, 56.3125808, 54.3073345, 52.2730075, 50.2079992, 48.1107092, 45.979537, 43.8128821, 41.4050737, 38.8608388, 36.287357, 33.6896505, 31.0727411, 28.4416508, 25.8014015, 23.1570153, 20.513514, 17.8759198, 15.2492544, 12.63854, 10.0487984, 7.48505164, 4.95232168, 2.45563047, 0, -0.902925863, -1.94062149, -3.10938633, -4.40551984, -5.82532148, -7.36509072, -9.021127, -10.7897298, -12.6671985, -14.6498327, -16.7339317, -18.9157951, -21.1917223, -23.5580127, -26.0109658, -28.5468811, -31.0740127, -33.5760823, -36.2485393, -39.0688162, -42.0143457, -45.0625605, -48.190893, -51.376776, -54.5976421, -57.8309239, -61.054054, -64.244465, -67.3795896, -70.3681989, -73.2815628, -76.1177125, -78.8746789, -81.6576509, -84.4110808, -87.0508452, -89.5428278, -91.8529123, -93.9469825, -95.7909222, -98.1544905, -99.9550378, -101.001168, -101.101484, -100.064591, -98.1105283, -94.5336041, -89.0395626, -81.3341482, -71.1231055, -58.2586746, -42.121054, -24.0050533, -3.70616743, 22.0412093, 55.1283111, 97.446372, 153.901281, 222.684468, 298.698649, 374.556406, 461.695838, 569.01417, 701.646014, 858.350444, 1040.81301, 1248.82011, 1483.56492, 1753.20419, 2061.95567, 2418.66634, 2822.21255, 3275.20964, 3780.27293, 4340.01777, 5121.35234, 5921.52591, 6702.08058, 7424.55848, 8050.50172, 8388.58038, 8655.12332, 8877.18923, 9160.20703, 9500.21387, 9860.59269, 10239.0134, 10633.1458, 11040.6599, 11459.2255, 11683.379, 12086.4459, 12595.1407, 13136.178, 13636.2724, 14022.1384, 14220.4905, 14419.2486, 14083.4789},
        {135.976907, 136.32744, 136.628954, 136.881862, 137.086577, 137.243511, 137.353077, 137.41569, 137.43176, 137.401702, 137.325929, 137.204853, 137.038887, 136.828445, 136.573939, 136.275782, 135.934388, 135.550168, 135.123537, 134.654907, 134.144691, 133.593302, 133.001153, 132.368657, 131.696227, 130.984276, 130.233217, 129.443463, 128.615426, 127.74952, 126.846158, 125.905753, 124.928717, 123.915464, 122.854201, 121.7517, 120.614881, 119.444553, 118.241524, 117.006605, 115.740603, 114.444328, 113.11859, 111.764196, 110.381957, 108.972681, 107.537177, 106.076255, 104.590723, 103.081391, 101.549067, 100.003039, 98.4354537, 96.8469365, 95.2381128, 93.609608, 91.9620472, 90.2960558, 88.6122591, 86.9112823, 85.1937508, 83.4602898, 81.7115247, 79.9480806, 78.170583, 76.379657, 74.575928, 72.7936607, 71.0663123, 69.325129, 67.5685556, 65.7950366, 64.0030164, 62.1909398, 60.3572513, 58.5003955, 56.6188168, 54.7109601, 52.7752697, 50.8101902, 48.8141664, 46.7856426, 44.7230635, 42.6248737, 40.2856225, 37.8099907, 35.3051099, 32.7760415, 30.2278468, 27.6655871, 25.094324, 22.5191187, 19.9450326, 17.3771271, 14.8204636, 12.2801033, 9.76110781, 7.26853833, 4.8074563, 2.38292306, 0, -0.959574658, -2.02947038, -3.20742643, -4.49118208, -5.87847659, -7.36704923, -8.95463928, -10.638986, -12.4178286, -14.2889065, -16.2499588, -18.2987248, -20.4329439, -22.6503552, -24.9486981, -27.3257117, -29.6660136, -31.9511826, -34.434829, -37.0904516, -39.8915494, -42.8116212, -45.8241658, -48.9026821, -52.0206689, -55.1516252, -58.2690498, -61.3464415, -64.3572992, -67.2012951, -69.9528246, -72.6075343, -75.1610712, -77.7366695, -80.2661822, -82.6474393, -84.8378114, -86.794669, -88.4753829, -89.8373236, -91.6861858, -92.8728304, -93.1886519, -92.4250443, -90.3734019, -86.2092542, -80.4938263, -73.1724785, -64.1905712, -53.4934648, -42.5083583, -27.8876374, -8.14909921, 17.7986406, 50.2490452, 89.853828, 137.264703, 194.450282, 260.446085, 332.851011, 405.359521, 489.016641, 592.244746, 720.304773, 873.100788, 1052.99575, 1262.58159, 1504.28062, 1779.6756, 2090.82419, 2436.49809, 2822.54643, 3252.16431, 3728.54684, 4254.8891, 4991.01968, 5744.34182, 6478.89226, 7158.70772, 7747.82492, 8063.14311, 8313.92817, 8527.27573, 8816.01558, 9173.30694, 9556.58649, 9960.79957, 10380.8916, 10811.8078, 11248.4936, 11441.5637, 11820.8458, 12312.899, 12844.2821, 13341.5544, 13731.2746, 13940.0019, 14111.8755, 14373.7146},
        {130.320679, 130.732602, 131.093539, 131.403994, 131.664471, 131.875474, 132.037509, 132.151078, 132.216688, 132.234842, 132.206045, 132.1308, 132.009613, 131.842989, 131.63143, 131.375442, 131.075529, 130.732196, 130.345947, 129.917286, 129.446718, 128.934747, 128.381878, 127.788614, 127.155462, 126.482924, 125.771505, 125.02171, 124.234043, 123.409009, 122.547112, 121.648856, 120.714745, 119.745285, 118.731873, 117.679758, 116.594299, 115.476296, 114.326549, 113.145857, 111.935022, 110.694841, 109.426117, 108.129648, 106.806235, 105.456677, 104.081774, 102.682327, 101.259135, 99.8129987, 98.3447172, 96.8673639, 95.3691996, 93.8507583, 92.3125736, 90.7551797, 89.1791102, 87.5848992, 85.9730805, 84.344188, 82.6987556, 81.0373172, 79.3604066, 77.6685577, 75.9623045, 74.2421808, 72.5087205, 70.7939373, 69.1290892, 67.4497853, 65.7545186, 64.0417823, 62.3100697, 60.5578739, 58.7836881, 56.9860054, 55.1633191, 53.3141222, 51.4369081, 49.5301698, 47.5924005, 45.6220934, 43.6177418, 41.5778386, 39.296444, 36.8785676, 34.4316764, 31.9608977, 29.4713591, 26.9681879, 24.4565116, 21.9414576, 19.4281533, 16.9217262, 14.4273036, 11.950013, 9.49498184, 7.06733754, 4.67220751, 2.31471919, 0, -0.898710535, -1.90693952, -3.02203662, -4.24135149, -5.56223376, -6.98203312, -8.49809919, -10.1077816, -11.8084301, -13.5973943, -15.4720238, -17.4296684, -19.4676775, -21.583401, -23.7741884, -26.0373895, -28.2833007, -30.4968357, -32.8722497, -35.3882383, -38.0234968, -40.7567207, -43.5666054, -46.4318464, -49.331139, -52.2431787, -55.1466609, -58.020281, -60.8427345, -63.56985, -66.2115742, -68.7534625, -71.1810704, -73.5831297, -75.8936076, -78.0189577, -79.9137823, -81.5326841, -82.8302654, -83.7611287, -85.2856605, -86.0465711, -85.8016793, -84.3088038, -81.3257631, -74.7190637, -66.6106647, -57.2312126, -46.8113544, -35.5817366, -26.5298565, -13.7600264, 6.22939017, 35.2561338, 71.699991, 115.484319, 166.532477, 225.088882, 290.682864, 362.493499, 435.116366, 518.790712, 621.508612, 748.28357, 899.082062, 1076.18723, 1282.40103, 1520.14111, 1789.92282, 2093.33757, 2428.51952, 2801.0492, 3213.71476, 3669.30432, 4170.60605, 4865.51338, 5575.43283, 6266.87621, 6906.35534, 7460.38203, 7755.22788, 7991.13841, 8194.72839, 8480.02796, 8839.04999, 9227.38425, 9638.86475, 10067.3255, 10506.6005, 10950.5237, 11105.774, 11472.9088, 11969.7715, 12514.2055, 13024.0542, 13417.161, 13611.3693, 13764.7342, 14443.3569},
        {124.827823, 125.383696, 125.878335, 126.312661, 126.687597, 127.004064, 127.262984, 127.465278, 127.611869, 127.703678, 127.741627, 127.726637, 127.65963, 127.541529, 127.373254, 127.155728, 126.889871, 126.576607, 126.216857, 125.811542, 125.361584, 124.867905, 124.331427, 123.753071, 123.13376, 122.474414, 121.775956, 121.039307, 120.265389, 119.455124, 118.609434, 117.72924, 116.815464, 115.869027, 114.902405, 113.9105, 112.888074, 111.835676, 110.753852, 109.643148, 108.504111, 107.337289, 106.143228, 104.922475, 103.675577, 102.40308, 101.105532, 99.7834791, 98.4374683, 97.0680466, 95.6757606, 94.2594216, 92.8213496, 91.3621292, 89.8823448, 88.3825809, 86.8634219, 85.3254522, 83.7692564, 82.1954189, 80.6045242, 78.9971567, 77.3739009, 75.7353412, 74.0820621, 72.4146481, 70.7336835, 69.0720447, 67.4618324, 65.8370319, 64.1961343, 62.5376307, 60.8600122, 59.16177, 57.4413951, 55.6973788, 53.928212, 52.1323859, 50.3083917, 48.4547205, 46.5698633, 44.6523114, 42.7005557, 40.7130875, 38.4800384, 36.1084619, 33.7079155, 31.2836522, 28.8409247, 26.3849858, 23.9210882, 21.4544848, 18.9904283, 16.5341715, 14.0909673, 11.6660683, 9.26472735, 6.89219728, 4.55373084, 2.25458082, 0, -0.707706466, -1.55002823, -2.52200358, -3.6186708, -4.83506818, -6.166234, -7.60720655, -9.15302411, -10.798725, -12.5393474, -14.3699297, -16.2855102, -18.2811271, -20.3518188, -22.4926234, -24.6985794, -26.9463553, -29.2283651, -31.5811976, -33.9959548, -36.4637385, -38.9756507, -41.5227933, -44.0962682, -46.6871774, -49.2866228, -51.8857063, -54.4755298, -57.0471953, -59.6638684, -62.218164, -64.6795648, -67.0175537, -69.2396167, -71.2962352, -73.1277559, -74.6922606, -75.947831, -76.8525491, -77.3644966, -78.6743291, -79.1324233, -78.4557054, -76.3611013, -72.5655371, -63.7014192, -53.3413231, -41.9733048, -30.08542, -18.1657249, -10.3667005, 0.96676457, 20.1451703, 49.385145, 86.6283292, 131.735746, 184.568417, 245.306842, 313.419959, 388.028191, 463.941037, 550.748871, 656.605528, 786.101425, 937.788974, 1112.99213, 1311.97855, 1535.79835, 1789.3747, 2075.43992, 2400.26248, 2762.49459, 3163.64456, 3605.22072, 4088.73137, 4745.44479, 5414.66933, 6065.47331, 6666.92505, 7188.09287, 7466.72775, 7690.76025, 7885.53753, 8158.63197, 8502.84327, 8876.71063, 9274.74679, 9691.46446, 10121.3764, 10558.9953, 10659.4186, 11020.5375, 11542.6093, 12125.891, 12670.6397, 13077.1127, 13245.567, 13409.1244, 14233.8833},
        {120.365703, 120.961423, 121.493393, 121.962686, 122.370374, 122.717532, 123.005231, 123.234545, 123.406547, 123.522311, 123.582908, 123.589413, 123.542898, 123.444437, 123.295102, 123.095966, 122.848103, 122.552586, 122.210487, 121.82288, 121.390837, 120.915433, 120.397739, 119.838829, 119.239777, 118.601654, 117.925534, 117.212491, 116.463597, 115.679925, 114.862549, 114.012541, 113.130974, 112.218922, 111.299221, 110.361605, 109.395545, 108.401408, 107.379562, 106.330372, 105.254207, 104.151432, 103.022414, 101.867521, 100.687119, 99.4815743, 98.2512544, 96.9965258, 95.7177555, 94.4153101, 93.0895566, 91.728487, 90.3451112, 88.940064, 87.5139805, 86.0674959, 84.6012451, 83.1158632, 81.6119852, 80.0902463, 78.5512815, 76.9957257, 75.4242142, 73.8373818, 72.2358638, 70.6202951, 68.9913108, 67.3833918, 65.8302065, 64.2625854, 62.6789692, 61.0777991, 59.457516, 57.8165608, 56.1533745, 54.4663981, 52.7540725, 51.0148386, 49.2471375, 47.4494101, 45.6200973, 43.7576402, 41.8604797, 39.9270566, 37.7410495, 35.4127978, 33.0552222, 30.6737333, 28.2737414, 25.8606569, 23.4398904, 21.0168521, 18.5969525, 16.185602, 13.7882111, 11.4101902, 9.05694971, 6.73390001, 4.44645154, 2.20001473, 0, -0.55620183, -1.26305054, -2.11418287, -3.1032356, -4.22384546, -5.46964921, -6.83428362, -8.31138542, -9.89459139, -11.5775383, -13.3538628, -15.2172018, -17.1611919, -19.17947, -21.2656727, -23.4134369, -25.6165628, -27.8687104, -30.1631468, -32.4935438, -34.8535731, -37.2369065, -39.6372158, -42.0481728, -44.4634493, -46.8767171, -49.2816479, -51.6719135, -54.0411857, -56.4759401, -58.8430166, -61.1082459, -63.2374587, -65.2440464, -67.0700592, -68.6448647, -69.9200253, -70.8471036, -71.377662, -71.463263, -72.3035652, -72.2221796, -70.926476, -68.1238241, -63.5215939, -53.599115, -42.0988073, -29.5350504, -16.422224, -3.27470792, 5.82549011, 18.2920525, 37.9781966, 66.9689809, 103.739474, 148.385556, 201.003108, 262.03531, 331.151811, 407.643386, 487.305416, 578.258393, 687.946771, 820.274362, 973.14556, 1147.21859, 1341.13169, 1555.01938, 1796.4228, 2068.6935, 2382.78235, 2733.27242, 3120.88471, 3546.3402, 4010.3599, 4631.38251, 5262.98188, 5876.44959, 6443.07721, 6934.15631, 7197.90124, 7410.73226, 7596.68832, 7856.23674, 8183.16358, 8539.40974, 8920.3535, 9321.3732, 9737.84713, 10165.1536, 10245.4575, 10607.7982, 11143.713, 11744.7394, 12302.4149, 12708.2769, 12853.8629, 13056.9916, 13991.0943},
        {117.105437, 117.588896, 118.018115, 118.393889, 118.717014, 118.988286, 119.208502, 119.378456, 119.498946, 119.570766, 119.594713, 119.571583, 119.502172, 119.387276, 119.227691, 119.024212, 118.777636, 118.488759, 118.158376, 117.787284, 117.376279, 116.926156, 116.437711, 115.911741, 115.349041, 114.750408, 114.116637, 113.448524, 112.746866, 112.012457, 111.246095, 110.448575, 109.620693, 108.763245, 107.89052, 106.996282, 106.074137, 105.124442, 104.147555, 103.143835, 102.11364, 101.057327, 99.9752556, 98.8677829, 97.7352673, 96.578067, 95.39654, 94.1910445, 92.9619386, 91.7095803, 90.4343279, 89.1234208, 87.79062, 86.4365683, 85.0619079, 83.6672815, 82.2533316, 80.8207006, 79.370031, 77.9019654, 76.4171463, 74.9162162, 73.3998175, 71.8685928, 70.3231846, 68.7642354, 67.1923876, 65.6438111, 64.1538224, 62.649792, 61.1300592, 59.5929634, 58.0368442, 56.4600408, 54.8608927, 53.2377394, 51.5889203, 49.9127747, 48.2076421, 46.471862, 44.7037737, 42.9017166, 41.0640302, 39.1890539, 37.0516718, 34.7666511, 32.4514447, 30.1116435, 27.7528386, 25.3806208, 23.0005812, 20.6183108, 18.2394006, 15.8694414, 13.5140243, 11.1787403, 8.86918025, 6.59093525, 4.34959622, 2.15075415, 0, -0.489813758, -1.12828895, -1.90927231, -2.82661057, -3.87415046, -5.04573871, -6.33522204, -7.7364472, -9.24326091, -10.8495099, -12.5490409, -14.3357006, -16.2033359, -18.1457933, -20.1569196, -22.2305617, -24.3088638, -26.3782866, -28.5496227, -30.8056396, -33.1291051, -35.5027868, -37.9094524, -40.3318695, -42.7528058, -45.1550289, -47.5213066, -49.8344064, -52.077096, -54.2487802, -56.3094892, -58.2369997, -60.0090883, -61.7545082, -63.3755477, -64.7342343, -65.7630517, -66.3944833, -66.5610126, -66.1951232, -66.2205322, -65.2768096, -63.102502, -59.4361558, -54.0163176, -44.4513515, -33.1425318, -20.3609506, -6.37769991, 8.53612833, 21.7129675, 38.2072246, 60.3062799, 89.4434492, 125.308333, 168.373424, 219.111216, 278.151764, 345.774189, 422.085727, 504.794949, 599.756823, 712.662845, 846.555169, 999.986788, 1173.57129, 1366.26819, 1578.26224, 1815.80316, 2081.71, 2384.55205, 2720.51344, 3090.42271, 3495.10844, 3935.39919, 4523.14627, 5120.3998, 5700.23266, 6235.71773, 6699.92788, 6949.69598, 7151.82826, 7329.21473, 7576.67361, 7888.62335, 8229.5123, 8595.25741, 8981.77562, 9384.98387, 9800.79912, 9903.938, 10271.8447, 10800.7093, 11386.7221, 11926.0732, 12314.9529, 12449.5513, 12723.2639, 13785.008},
        {114.315772, 114.655332, 114.952446, 115.207551, 115.421082, 115.593477, 115.725172, 115.816602, 115.868205, 115.880416, 115.853672, 115.788409, 115.685064, 115.544072, 115.36587, 115.150895, 114.899582, 114.612368, 114.28969, 113.931983, 113.539684, 113.113229, 112.653054, 112.159597, 111.633292, 111.074577, 110.483888, 109.861661, 109.208332, 108.524338, 107.810114, 107.066098, 106.292726, 105.490434, 104.661007, 103.804179, 102.919667, 102.007864, 101.069161, 100.103951, 99.1126278, 98.0955826, 97.0532083, 95.9858974, 94.8940424, 93.7780357, 92.63827, 91.4751377, 90.2890314, 89.0803435, 87.8494665, 86.5856593, 85.3006894, 83.9951907, 82.6697971, 81.3251424, 79.9618606, 78.5805853, 77.1819507, 75.7665904, 74.3351384, 72.8882285, 71.4264946, 69.9505706, 68.4610904, 66.9586877, 65.4439966, 63.9548313, 62.5281137, 61.0877955, 59.6321003, 58.1592516, 56.6674732, 55.1549885, 53.6200213, 52.0607951, 50.4755334, 48.86246, 47.2197983, 45.5457721, 43.8386049, 42.0965203, 40.3177419, 38.5004934, 36.4104734, 34.1670589, 31.8924215, 29.5923307, 27.2725559, 24.9388667, 22.5970324, 20.2528226, 17.9120068, 15.5803544, 13.2636349, 10.9676178, 8.69807257, 6.4607687, 4.26147567, 2.10596294, 0, -0.438318446, -1.02160362, -1.74403966, -2.59981073, -3.58310096, -4.6880945, -5.90897551, -7.23992812, -8.67513648, -10.2087847, -11.8350571, -13.5481376, -15.3422104, -17.2114597, -19.1500697, -21.1522245, -23.1068854, -24.9932048, -27.0433707, -29.2290195, -31.5217876, -33.8933115, -36.3152276, -38.7591722, -41.1967817, -43.5996927, -45.9395415, -48.1879644, -50.316598, -52.2396845, -54.0072989, -55.608296, -57.0315303, -58.5144814, -59.9216915, -61.0514029, -61.8178829, -62.1353986, -61.9182172, -61.080606, -60.2242144, -58.3667241, -55.2879145, -50.767565, -44.5854549, -35.7187708, -24.9505331, -12.2611694, 2.36889264, 18.9592256, 36.4397628, 57.2514966, 82.040537, 111.541784, 146.671415, 188.264219, 237.154984, 294.159856, 360.136502, 435.96292, 521.172816, 619.452923, 734.623066, 868.916334, 1021.87717, 1194.28416, 1385.95637, 1597.42363, 1832.73392, 2093.94512, 2385.65702, 2706.9946, 3059.1359, 3443.25895, 3860.54178, 4418.76817, 4985.85898, 5536.34079, 6044.74021, 6485.58382, 6723.11156, 6915.66112, 7085.02482, 7320.54268, 7617.26937, 7942.11468, 8291.54052, 8662.00881, 9049.98147, 9451.92041, 9586.6895, 9958.577, 10472.6155, 11033.8378, 11547.2763, 11917.9638, 12050.933, 12411.4064, 13565.2269},
        {111.077693, 111.372077, 111.627942, 111.845632, 112.025492, 112.167868, 112.273103, 112.341544, 112.373536, 112.369422, 112.329548, 112.25426, 112.143901, 111.998818, 111.819354, 111.605856, 111.358667, 111.078134, 110.7646, 110.418411, 110.039912, 109.629447, 109.187363, 108.714003, 108.209712, 107.674837, 107.109721, 106.51471, 105.890148, 105.236381, 104.553753, 103.84261, 103.103296, 102.336156, 101.540421, 100.717015, 99.8668783, 98.9903928, 98.0879394, 97.1598992, 96.2066534, 95.2285832, 94.2260698, 93.1994943, 92.149238, 91.0756819, 89.9792074, 88.8601955, 87.7190275, 86.5560846, 85.3717478, 84.1562079, 82.9202575, 81.6644987, 80.3895336, 79.0959642, 77.7843928, 76.4554213, 75.1096518, 73.7476866, 72.3701275, 70.9775768, 69.5706366, 68.1499089, 66.7159958, 65.2694994, 63.8110218, 62.3791264, 61.011465, 59.6303475, 58.2339152, 56.8203093, 55.3876709, 53.9341414, 52.4578619, 50.9569736, 49.4296179, 47.8739358, 46.2880687, 44.6701577, 43.0183441, 41.3307691, 39.6055739, 37.8408997, 35.7959593, 33.5933837, 31.3586756, 29.09773, 26.8164416, 24.5207055, 22.2164166, 19.9094697, 17.6057597, 15.3111817, 13.0316304, 10.7730008, 8.5411878, 6.34208632, 4.18159125, 2.06559751, 0, -0.330167432, -0.818760399, -1.45860488, -2.24252686, -3.16335232, -4.21390723, -5.38701758, -6.67550935, -8.07220853, -9.56994108, -11.161533, -12.8398103, -14.5975988, -16.4277247, -18.3230139, -20.2762923, -22.2215515, -24.1432123, -26.1671791, -28.2736703, -30.4429045, -32.6551003, -34.8904763, -37.1292511, -39.3516431, -41.5378711, -43.6681536, -45.7227092, -47.6817564, -49.5552633, -51.2827908, -52.8356328, -54.185083, -55.4793904, -56.6013713, -57.3866535, -57.7534443, -57.6199508, -56.9043802, -55.5249396, -54.0225438, -51.5031727, -47.7631993, -42.5989966, -35.8069378, -26.5920914, -15.4899611, -2.444746, 12.5993547, 29.6981418, 47.8424044, 69.4546361, 95.3266296, 126.227081, 162.877529, 206.020684, 256.399258, 314.660605, 381.664465, 458.374608, 544.644695, 643.449489, 757.708905, 889.319334, 1038.44286, 1206.03764, 1392.41884, 1598.37792, 1827.06397, 2080.29246, 2361.55415, 2670.73149, 3009.06007, 3377.77546, 3778.11322, 4313.78556, 4857.93227, 5386.16977, 5874.1145, 6297.38285, 6526.30427, 6711.9735, 6875.12995, 7099.6061, 7381.16702, 7689.28905, 8021.05172, 8373.53456, 8743.8171, 9128.97887, 9255.48416, 9615.30308, 10116.5474, 10667.3291, 11175.7598, 11549.9514, 11698.0157, 12147.7656, 13315.8052},
        {107.896638, 108.176168, 108.418544, 108.624107, 108.793195, 108.926148, 109.023306, 109.085008, 109.111593, 109.103402, 109.060773, 108.984045, 108.87356, 108.729655, 108.55267, 108.342946, 108.10082, 107.826634, 107.520726, 107.183436, 106.815103, 106.416067, 105.986667, 105.527243, 105.038134, 104.519679, 103.972219, 103.396093, 102.791639, 102.159199, 101.49911, 100.811713, 100.097347, 99.3563514, 98.5883817, 97.7941339, 96.974312, 96.1292778, 95.2593929, 94.3650191, 93.446518, 92.5042513, 91.5385807, 90.5498679, 89.5384745, 88.5047622, 87.4490928, 86.3718279, 85.2733291, 84.1539583, 83.014077, 81.8449279, 80.6561893, 79.4484208, 78.2221816, 76.9780311, 75.7165287, 74.4382337, 73.1437056, 71.8335036, 70.5081872, 69.1683157, 67.8144485, 66.447145, 65.0669644, 63.6744663, 62.2702099, 60.89315, 59.5813195, 58.2560901, 56.9155323, 55.5577164, 54.1807127, 52.7825915, 51.3614233, 49.9152784, 48.4422271, 46.9403397, 45.4076868, 43.8423385, 42.2423652, 40.6058374, 38.9308252, 37.2153992, 35.2132242, 33.049714, 30.85322, 28.629744, 26.3852877, 24.1258529, 21.8574415, 19.5860551, 17.3176956, 15.0583648, 12.8140643, 10.5907961, 8.3945618, 6.23136328, 4.10720229, 2.0280806, 0, -0.215853063, -0.606944492, -1.16436451, -1.87920334, -2.74255121, -3.74549834, -4.87913495, -6.13455128, -7.50283754, -8.97508396, -10.5423808, -12.1958182, -13.9264864, -15.7254757, -17.5838763, -19.4927783, -21.4536302, -23.4590019, -25.4765546, -27.499598, -29.521442, -31.5353965, -33.5347712, -35.5128759, -37.4630205, -39.3785148, -41.2526686, -43.0787917, -44.850194, -46.7125389, -48.4509196, -50.0129398, -51.3462031, -52.4628943, -53.2783265, -53.6905914, -53.6279184, -53.0185369, -51.7906764, -49.8725662, -47.8204521, -44.7434117, -40.4468018, -34.7359789, -27.4162997, -17.574603, -5.91439285, 7.57934473, 22.9216236, 40.1274575, 57.7827604, 79.078324, 105.491117, 138.045467, 176.841558, 222.394498, 275.219393, 335.743199, 404.589207, 482.476864, 568.927822, 666.888242, 778.875655, 906.686517, 1050.89546, 1212.63796, 1392.61139, 1591.83765, 1812.94495, 2057.6529, 2329.09884, 2627.3665, 2953.68471, 3309.28229, 3695.3881, 4212.22567, 4737.28045, 5247.03258, 5717.96221, 6126.54949, 6348.26836, 6527.94266, 6685.34094, 6899.12817, 7165.75969, 7457.1507, 7770.98356, 8104.94063, 8456.70429, 8823.95691, 8935.79229, 9278.9234, 9764.88485, 10305.2113, 10811.4372, 11195.0974, 11367.7264, 11909.3196, 13051.9927},
        {104.915139, 105.192371, 105.432627, 105.636279, 105.8037, 105.935263, 106.031342, 106.09231, 106.118541, 106.110407, 106.068282, 105.992539, 105.883551, 105.741692, 105.567335, 105.360852, 105.122618, 104.853006, 104.552388, 104.221139, 103.85963, 103.468237, 103.047331, 102.597286, 102.118476, 101.611273, 101.076051, 100.513183, 99.923043, 99.3060032, 98.6624372, 97.9927184, 97.2972199, 96.5763151, 95.831421, 95.0623671, 94.26897, 93.4515693, 92.6105043, 91.7461143, 90.8587388, 89.948717, 89.0163885, 88.0620925, 87.0861685, 86.0889558, 85.0707939, 84.032022, 82.9729795, 81.8940059, 80.7954405, 79.6696148, 78.5250497, 77.3622581, 76.1817531, 74.9840475, 73.7696544, 72.5390867, 71.2928574, 70.0314794, 68.7554658, 67.4653296, 66.1615836, 64.8447409, 63.5153144, 62.1738172, 60.8207621, 59.4953221, 58.2357418, 56.962801, 55.6745066, 54.3688654, 53.0438842, 51.6975697, 50.3279288, 48.9329683, 47.510695, 46.0591158, 44.5762373, 43.0600665, 41.5086101, 39.919875, 38.291868, 36.6225958, 34.6605886, 32.5338389, 30.3733379, 28.1851883, 25.9754929, 23.7503545, 21.5158761, 19.2781602, 17.0433098, 14.8174277, 12.6066167, 10.4169795, 8.25461901, 6.12563801, 4.0361393, 1.9922257, 0, -0.114775475, -0.419478007, -0.903602231, -1.55664278, -2.3680943, -3.32745141, -4.42420876, -5.64786098, -6.9879027, -8.43382856, -9.9751332, -11.6013112, -13.3018573, -15.0662661, -16.8840322, -18.7446503, -20.7083848, -22.7748405, -24.7734371, -26.7088343, -28.5856917, -30.4086689, -32.1824256, -33.9116212, -35.6009156, -37.2549682, -38.8784387, -40.4759867, -42.0522718, -43.8692047, -45.5798685, -47.1117476, -48.3923263, -49.3147694, -49.823721, -49.8729769, -49.4003172, -48.3435219, -46.6403713, -44.2286455, -41.6962508, -38.1329765, -33.3494041, -27.1561151, -19.363691, -8.83519289, 3.43439728, 17.397618, 33.0070076, 50.2151046, 67.0971669, 87.7774674, 114.502474, 148.59207, 189.474356, 237.426806, 292.726891, 355.66538, 426.503429, 505.48769, 591.430635, 687.703443, 796.843185, 920.714506, 1060.04267, 1216.07594, 1389.73627, 1582.18734, 1795.78921, 2032.22517, 2294.59946, 2582.95588, 2898.48587, 3242.3809, 3615.83243, 4115.95281, 4624.03237, 5117.28236, 5572.91399, 5968.13849, 6182.87996, 6356.49486, 6507.88893, 6711.12403, 6963.27323, 7238.42791, 7534.81024, 7850.64238, 8184.14651, 8533.54481, 8636.85587, 8963.51301, 9430.83261, 9956.13104, 10456.7247, 10849.93, 11053.0632, 11685.6242, 12781.0914},
        {102.481655, 102.731473, 102.946102, 103.125894, 103.271201, 103.382377, 103.459772, 103.503739, 103.514632, 103.492801, 103.438599, 103.352379, 103.234493, 103.085292, 102.905131, 102.694359, 102.453331, 102.182398, 101.881913, 101.552228, 101.193694, 100.806666, 100.391494, 99.9485306, 99.4781289, 98.9806408, 98.4564187, 97.9058148, 97.3291814, 96.7268708, 96.0992353, 95.4466272, 94.7693987, 94.0679023, 93.3435175, 92.5960615, 91.8253387, 91.0316681, 90.2153688, 89.3767596, 88.5161594, 87.6338874, 86.7302623, 85.8056032, 84.860229, 83.8944586, 82.9086111, 81.9030053, 80.8779602, 79.8337948, 78.770828, 77.6818194, 76.5748112, 75.4502862, 74.3087272, 73.1506171, 71.9764387, 70.7866748, 69.5818082, 68.3623218, 67.1286983, 65.8814207, 64.6209717, 63.3478342, 62.0624909, 60.7654247, 59.4571185, 58.1771534, 56.9641721, 55.7380197, 54.4966445, 53.2379947, 51.9600185, 50.6606641, 49.3378799, 47.989614, 46.6138147, 45.2084301, 43.7714086, 42.3006984, 40.7942476, 39.2500045, 37.6659174, 36.0399344, 34.11496, 32.0218287, 29.8944408, 27.7390211, 25.5617944, 23.3689858, 21.1668201, 18.9615222, 16.759317, 14.5664294, 12.3890842, 10.2335065, 8.10592103, 6.01255275, 3.95962656, 1.95336734, 0, -0.116939789, -0.413941807, -0.880994376, -1.50808582, -2.28520445, -3.20233859, -4.24947656, -5.41660669, -6.6937173, -8.07079669, -9.5378332, -11.0848152, -12.7017309, -14.3785686, -16.1053168, -17.8719637, -19.6793726, -21.5190854, -23.3564922, -25.1839118, -26.9936629, -28.778064, -30.5294339, -32.2400912, -33.9023546, -35.5085428, -37.0509744, -38.5219681, -39.9138426, -41.3729586, -42.6811107, -43.7844047, -44.6289468, -45.2363102, -45.5148674, -45.352866, -44.673772, -43.4010511, -41.4581692, -38.7685921, -35.7671732, -31.7103511, -26.4215376, -19.7241444, -11.4415833, -0.735818369, 11.7429291, 26.0058854, 42.0642769, 59.92933, 78.0144905, 99.8816081, 127.351979, 161.584782, 202.38737, 250.174037, 305.359079, 368.785795, 440.341974, 519.447398, 604.046377, 698.375156, 805.81133, 928.485226, 1066.41374, 1220.58388, 1391.5709, 1580.25504, 1789.02628, 2019.42062, 2274.09109, 2553.28478, 2858.15101, 3189.8391, 3549.49837, 4030.44391, 4518.61783, 4992.128, 5429.0823, 5807.5886, 6012.19908, 6176.94778, 6320.03788, 6511.941, 6750.17074, 7010.46214, 7291.41771, 7591.63997, 7909.73141, 8244.29455, 8374.69743, 8697.53037, 9142.25976, 9638.35198, 10115.2734, 10502.4905, 10729.4695, 11451.4968, 12521.4244},
        {100.344301, 100.565685, 100.753566, 100.908279, 101.030156, 101.119531, 101.176738, 101.202108, 101.195977, 101.158677, 101.090542, 100.991904, 100.863098, 100.704456, 100.516312, 100.299, 100.052852, 99.7782026, 99.4753844, 99.1447308, 98.7865753, 98.4012513, 97.989092, 97.5504308, 97.0856012, 96.5949363, 96.0787696, 95.5374344, 94.9712641, 94.380592, 93.7657514, 93.1270758, 92.4648984, 91.7795526, 91.0724284, 90.3433085, 89.5919631, 88.8186913, 88.0237922, 87.2075647, 86.370308, 85.5123211, 84.633903, 83.735353, 82.8169699, 81.8790528, 80.9219009, 79.9458132, 78.9510887, 77.9380266, 76.9069258, 75.8508379, 74.7774667, 73.6872683, 72.5806988, 71.4582144, 70.3202713, 69.1673256, 67.9998336, 66.8182514, 65.6230351, 64.414641, 63.1935253, 61.960144, 60.7149534, 59.4584096, 58.1909688, 56.9526425, 55.7824923, 54.5993948, 53.4012418, 52.1859253, 50.9513373, 49.6953697, 48.4159145, 47.1108637, 45.7781091, 44.4155429, 43.0210569, 41.5925431, 40.1278934, 38.6249999, 37.0817545, 35.4960492, 33.6050083, 31.5423924, 29.4450965, 27.319475, 25.1718822, 23.0086724, 20.8362002, 18.6608197, 16.4888855, 14.3267518, 12.1807731, 10.0573038, 7.96269813, 5.90331055, 3.88549541, 1.91560711, 0, -0.128875335, -0.425220215, -0.879794788, -1.4833592, -2.2266736, -3.10049814, -4.09559297, -5.20271822, -6.41263406, -7.71610062, -9.10387806, -10.5667265, -12.0954062, -13.6806771, -15.3132995, -16.9840336, -18.6129935, -20.1808473, -21.8381492, -23.5605209, -25.3235843, -27.102961, -28.8742729, -30.6131417, -32.2951891, -33.8960369, -35.3913068, -36.7566206, -37.9676001, -39.01171, -39.8483867, -40.4496989, -40.7877157, -41.0542119, -41.1114034, -40.7629177, -39.9149117, -38.4735425, -36.3449671, -33.4353424, -30.0230822, -25.5287912, -19.7857937, -12.6274139, -3.88697584, 6.92302245, 19.568224, 34.135098, 50.7101137, 69.3797405, 89.0461622, 112.427593, 140.690203, 174.742226, 214.965279, 261.977421, 316.39671, 379.70353, 451.457628, 530.278029, 613.302662, 705.650449, 811.671672, 933.784865, 1071.13533, 1224.37083, 1393.58697, 1579.28838, 1784.00424, 2009.11853, 2256.83214, 2527.58646, 2822.48268, 3142.622, 3489.10562, 3951.65093, 4420.58888, 4874.86663, 5293.43131, 5655.23009, 5848.91436, 6003.92419, 6137.90489, 6318.60552, 6543.83868, 6790.54043, 7057.56721, 7343.77544, 7648.02156, 7969.16202, 8144.54701, 8468.10616, 8885.88205, 9343.91727, 9788.25439, 10164.936, 10420.0047, 11221.5125, 12261.9341},
        {98.4644962, 98.6585766, 98.8206317, 98.950981, 99.0499443, 99.1178411, 99.1549909, 99.1617135, 99.1383284, 99.0851553, 99.0025136, 98.8907232, 98.7501034, 98.580974, 98.3836546, 98.1584648, 97.9057241, 97.6257522, 97.3188687, 96.9853932, 96.6256453, 96.2399446, 95.8286108, 95.3919633, 94.9303219, 94.4440062, 93.9333356, 93.39863, 92.8402088, 92.2583916, 91.6534981, 91.0258479, 90.3757606, 89.7035558, 89.010758, 88.2970591, 87.5621362, 86.80627, 86.029741, 85.2328297, 84.4158165, 83.5789819, 82.7226066, 81.8469709, 80.9523554, 80.0390406, 79.107307, 78.1574351, 77.1897054, 76.2043985, 75.2017947, 74.1751922, 73.1320052, 72.0726658, 70.9976056, 69.9072567, 68.8020509, 67.68242, 66.5487959, 65.4016106, 64.2412958, 63.0682834, 61.8830054, 60.6858935, 59.4773797, 58.2578958, 57.0278738, 55.8277975, 54.6971896, 53.5538771, 52.3956956, 51.2204806, 50.0260674, 48.8102916, 47.5709886, 46.3059938, 45.0131427, 43.6902707, 42.3352133, 40.945806, 39.5198842, 38.0552833, 36.5498389, 35.0013863, 33.1410117, 31.1055361, 29.0349862, 26.935854, 24.8146316, 22.6778108, 20.5318838, 18.3833423, 16.2386785, 14.1043843, 11.9869517, 9.89287266, 7.82863914, 5.80074314, 3.81567663, 1.87993159, 0, -0.122301393, -0.402853393, -0.832991019, -1.40404929, -2.10736322, -2.93426783, -3.87609814, -4.92418917, -6.06987594, -7.30449345, -8.61937674, -10.0058608, -11.4552807, -12.9589714, -14.508268, -16.0945054, -17.5592195, -18.8723455, -20.3640498, -21.9935676, -23.7201342, -25.5029846, -27.3013542, -29.074478, -30.7815914, -32.3819294, -33.8347274, -35.0992204, -36.1346437, -36.7722643, -37.1462072, -37.2540981, -37.0935628, -37.017656, -36.8462889, -36.3045913, -35.28356, -33.6741921, -31.3674842, -28.2544334, -24.4990957, -19.6363038, -13.5036298, -5.93864611, 3.22107518, 14.1413419, 26.9803567, 41.8997024, 59.0609615, 78.6257169, 100.024438, 125.043404, 154.131214, 187.914596, 227.383947, 273.366386, 326.68903, 389.379188, 460.791015, 538.969369, 620.527432, 711.207244, 816.161648, 938.017553, 1075.12181, 1227.7859, 1395.62847, 1578.7814, 1779.91679, 2000.26985, 2241.59493, 2504.5282, 2790.12518, 3099.44134, 3433.5322, 3878.91009, 4329.80945, 4765.92156, 5166.93771, 5512.54916, 5694.83194, 5839.50278, 5963.80235, 6133.6905, 6347.12325, 6581.75697, 6836.56266, 7110.51133, 7402.57399, 7711.72164, 7949.02233, 8277.05504, 8663.03723, 9074.18634, 9477.71985, 9840.8552, 10130.8099, 10996.12, 12002.0021},
        {96.7807852, 96.9550648, 97.0981577, 97.2103887, 97.2920826, 97.3435645, 97.3651591, 97.3571914, 97.3199862, 97.2538685, 97.1591632, 97.036195, 96.885289, 96.70677, 96.5009628, 96.2681924, 96.0087837, 95.7230616, 95.4113509, 95.0739765, 94.7112634, 94.3235364, 93.9111203, 93.4743402, 93.0135208, 92.5289871, 92.021064, 91.4900763, 90.9363489, 90.3602067, 89.7619747, 89.1419776, 88.5005405, 87.8379881, 87.1564535, 86.4553195, 85.7339471, 84.9926026, 84.2315522, 83.451062, 82.6513984, 81.8328274, 80.9956154, 80.1400285, 79.2663329, 78.374795, 77.4656807, 76.5392565, 75.5957884, 74.6355428, 73.6587857, 72.6594951, 71.6443619, 70.6137885, 69.5681775, 68.5079316, 67.4334531, 66.3451446, 65.2434087, 64.1286479, 63.0012647, 61.8616616, 60.7102413, 59.5474061, 58.3735587, 57.1891016, 55.9944374, 54.8305411, 53.7374138, 52.6317801, 51.5114123, 50.3740827, 49.2175639, 48.039628, 46.8380476, 45.610595, 44.3550425, 43.0691625, 41.7507274, 40.3975096, 39.0072815, 37.5778153, 36.1068836, 34.5922586, 32.7585262, 30.7457981, 28.6974695, 26.6201784, 24.5205625, 22.4052597, 20.2809076, 18.1541442, 16.0316073, 13.9199346, 11.8257639, 9.75573307, 7.71647992, 5.71464223, 3.75685781, 1.84976446, 0, 0.0553848848, -0.0735256236, -0.376032411, -0.841436364, -1.45903837, -2.21813931, -3.10804008, -4.11804155, -5.23744462, -6.45555017, -7.76165909, -9.14507226, -10.5950906, -12.1010149, -13.6521462, -15.2377852, -16.7168939, -18.0601533, -19.5516787, -21.1528413, -22.825012, -24.5295621, -26.2278626, -27.8812846, -29.4511992, -30.8989775, -32.1859906, -33.2736096, -34.1232055, -34.5868073, -34.7752203, -34.6826185, -34.3031756, -33.9422851, -33.4385109, -32.547425, -31.1698355, -29.2065501, -26.5583768, -23.1261235, -19.2368923, -14.2354552, -7.93921461, -0.165573139, 9.26806672, 20.5254647, 33.8127654, 49.3172759, 67.2263031, 87.7271539, 110.527315, 136.880361, 166.90443, 201.063482, 240.527526, 286.14957, 338.78262, 400.252036, 470.217485, 547.277886, 628.880496, 719.842585, 824.64639, 945.581453, 1081.22192, 1231.84735, 1397.11264, 1577.13539, 1774.32377, 1989.79029, 2224.90172, 2480.47718, 2757.54116, 3057.11815, 3380.23262, 3811.04644, 4246.66238, 4667.32061, 5053.26126, 5384.7245, 5556.02902, 5690.61738, 5805.55319, 5966.69454, 6171.79057, 6398.25939, 6644.86671, 6910.37821, 7193.55959, 7493.17655, 7842.83432, 8184.94559, 8522.36974, 8857.96616, 9194.59425, 9535.11339, 9882.38296, 10773.1851, 11742.8931},
        {95.2597827, 95.4229555, 95.5551053, 95.6565823, 95.7277363, 95.7689173, 95.7804753, 95.7627602, 95.7161221, 95.6409109, 95.5374765, 95.4061689, 95.2473382, 95.0613342, 94.8485069, 94.6092064, 94.3437825, 94.0525852, 93.7359645, 93.3942704, 93.0278529, 92.6370618, 92.2222472, 91.7837591, 91.3219474, 90.8371621, 90.3297531, 89.8000704, 89.248464, 88.6752839, 88.0808799, 87.4656022, 86.8298006, 86.1738252, 85.5009061, 84.8098926, 84.0995994, 83.370283, 82.6221998, 81.8556063, 81.070759, 80.2679145, 79.4473292, 78.6092596, 77.7539622, 76.8816936, 75.9927102, 75.0872684, 74.1656249, 73.2280361, 72.2747585, 71.3008853, 70.3119481, 69.3083155, 68.2903559, 67.2584378, 66.2129294, 65.1541994, 64.082616, 62.9985478, 61.9023631, 60.7944304, 59.6751182, 58.5447947, 57.4038286, 56.2525882, 55.0914419, 53.9619052, 52.9045057, 51.8347492, 50.7503365, 49.6489688, 48.5283471, 47.3861724, 46.2201458, 45.0279683, 43.8073408, 42.5559645, 41.2715404, 39.9517695, 38.5943529, 37.1969915, 35.7573864, 34.2732387, 32.4620899, 30.4676136, 28.4368506, 26.3765936, 24.2936354, 22.1947687, 20.0867863, 17.9764809, 15.8706452, 13.776072, 11.6995541, 9.64788411, 7.62785486, 5.64625909, 3.70988952, 1.82553891, 0, 0.404232736, 0.562616075, 0.490527516, 0.203344559, -0.283555297, -0.954794553, -1.79499571, -2.78878127, -3.92077373, -5.17559559, -6.53786935, -7.99221752, -9.52326259, -11.1156271, -12.7539335, -14.4228042, -16.1021824, -17.776022, -19.4395301, -21.0763264, -22.6700308, -24.2042629, -25.6626425, -27.0287893, -28.2863231, -29.4188637, -30.4100308, -31.2434441, -31.9027233, -32.4545475, -32.769022, -32.8048491, -32.5207308, -31.9486468, -31.0106603, -29.6092936, -27.6812655, -25.1632945, -21.9920993, -18.1043986, -14.2701004, -9.3391547, -3.08526486, 4.71786573, 14.2965337, 26.1648029, 40.189261, 56.5242626, 75.3241626, 96.7433157, 120.444788, 147.674687, 178.711267, 214.013784, 254.411038, 300.565912, 353.14129, 412.946898, 480.4654, 556.019267, 639.273808, 732.526094, 838.039548, 957.200796, 1089.91078, 1236.75239, 1397.98037, 1574.09253, 1766.78993, 1977.09298, 2206.03579, 2454.62301, 2723.87034, 3014.7935, 3328.40817, 3747.51229, 4170.89376, 4579.12274, 4952.76937, 5272.40378, 5433.34185, 5558.24417, 5664.21843, 5818.61785, 6018.66125, 6240.63182, 6482.79518, 6743.41692, 7020.76266, 7313.098, 7816.77121, 8180.25043, 8454.47475, 8690.38328, 8938.91509, 9251.00929, 9677.60496, 10553.7433, 11483.561},
        {93.9037119, 94.0590911, 94.1833976, 94.2770104, 94.3403087, 94.3736716, 94.377478, 94.3521071, 94.2979378, 94.2153494, 94.1047208, 93.9664312, 93.8008595, 93.6083848, 93.3893863, 93.144243, 92.8733339, 92.5770381, 92.2557347, 91.9098028, 91.5396214, 91.1455695, 90.7280263, 90.2873708, 89.8239821, 89.3382393, 88.8305213, 88.3012074, 87.7506765, 87.1793077, 86.5874801, 85.9755727, 85.3439646, 84.6930349, 84.0271904, 83.3447115, 82.6438304, 81.9247953, 81.1878547, 80.4332569, 79.6612503, 78.8720832, 78.066004, 77.2432611, 76.4041028, 75.5487775, 74.6775335, 73.7906192, 72.8882829, 71.9707731, 71.0383381, 70.087019, 69.1213627, 68.1417087, 67.1483965, 66.1417656, 65.1221557, 64.0899063, 63.045357, 61.9888473, 60.9207168, 59.8413051, 58.7509517, 57.6499961, 56.538778, 55.4176369, 54.2869124, 53.1888093, 52.1645272, 51.1280617, 50.0770385, 49.0090831, 47.9218213, 46.8128784, 45.6798803, 44.5204525, 43.3322205, 42.11281, 40.8598467, 39.5709561, 38.2437637, 36.8758954, 35.4649765, 34.0086328, 32.2168927, 30.2372296, 28.2205736, 26.1738833, 24.1041176, 22.018235, 19.9231942, 17.825954, 15.733473, 13.65271, 11.5906235, 9.55417237, 7.55031518, 5.58601067, 3.6682175, 1.80389439, 0, 0.730763287, 1.15851236, 1.30296743, 1.18384871, 0.820876387, 0.233770687, -0.55774819, -1.53396004, -2.67514464, -3.9615818, -5.37355131, -6.89133296, -8.49520653, -10.1654518, -11.8823487, -13.6261768, -15.4995079, -17.5000919, -19.3315962, -21.0005061, -22.5133066, -23.876483, -25.0965205, -26.1799042, -27.1331192, -27.9626509, -28.6749844, -29.2766049, -29.7739975, -30.4597175, -30.9492876, -31.1633721, -31.0226351, -30.2754975, -28.9287451, -27.0350952, -24.5668847, -21.496451, -17.7961309, -13.4382617, -9.62213638, -4.71971471, 1.5367228, 9.41489566, 19.1825234, 31.7616097, 46.6020189, 63.8078992, 83.4833991, 105.732667, 130.062646, 157.904607, 189.851922, 226.416996, 267.946932, 314.863049, 367.586668, 425.786807, 490.808065, 564.815732, 649.767717, 745.371963, 851.607669, 968.939099, 1098.65243, 1241.65657, 1398.83686, 1571.09611, 1759.42348, 1964.75929, 2187.78293, 2429.73578, 2691.64847, 2974.55166, 3279.47597, 3688.55569, 4101.4419, 4498.88932, 4861.65267, 5170.48668, 5320.95799, 5435.80146, 5532.28099, 5679.67968, 5874.6705, 6092.25175, 6330.18015, 6586.21239, 6858.10518, 7143.61522, 7780.52944, 8159.98193, 8376.30666, 8523.83763, 8696.90884, 8989.85427, 9497.00792, 10342.8514, 11220.0284},
        {92.6925216, 92.8421508, 92.9605937, 93.0482572, 93.1055479, 93.1328726, 93.1306381, 93.0992511, 93.0391185, 92.9506469, 92.8342431, 92.6903139, 92.519266, 92.3215062, 92.0974412, 91.8474779, 91.572023, 91.2714832, 90.9462653, 90.596776, 90.2234221, 89.8266104, 89.4067477, 88.9642406, 88.499496, 88.0129206, 87.5049212, 86.9759045, 86.4262773, 85.8564463, 85.2668183, 84.6578001, 84.0297985, 83.3832201, 82.7235835, 82.048632, 81.3560476, 80.6460712, 79.9189436, 79.1749057, 78.4141984, 77.6370626, 76.8437391, 76.0344688, 75.2094927, 74.3690514, 73.513386, 72.6427374, 71.7573463, 70.8574537, 69.9433004, 69.0114843, 68.0659682, 67.1070721, 66.1351157, 65.150419, 64.1533016, 63.1440836, 62.1230848, 61.0906249, 60.0470238, 58.9926015, 57.9276777, 56.8525723, 55.7676051, 54.673096, 53.5693649, 52.4995114, 51.5056081, 50.4997446, 49.4794675, 48.4423235, 47.3858591, 46.3076209, 45.2051556, 44.0760097, 42.9177299, 41.7278627, 40.5039548, 39.2435527, 37.9442031, 36.6034526, 35.2188478, 33.7879352, 32.0127382, 30.044781, 28.0391397, 26.0029513, 23.9433526, 21.8674805, 19.7824719, 17.6954638, 15.6135931, 13.5439966, 11.4938112, 9.47017395, 7.48022162, 5.53109116, 3.62991946, 1.78384344, 0, 0.962556337, 1.58410981, 1.88729384, 1.89474182, 1.62908717, 1.1129633, 0.369003619, -0.580158469, -1.71188955, -3.00355622, -4.43252506, -5.97616266, -7.61183562, -9.31691053, -11.068754, -12.8447325, -14.8382564, -17.0575554, -18.9913264, -20.6632308, -22.0969304, -23.3160868, -24.3443616, -25.2054165, -25.9229131, -26.520513, -27.0218779, -27.4506694, -27.8305492, -28.6333643, -29.2702563, -29.6304311, -29.6030946, -28.7131852, -27.0320426, -24.7281448, -21.7999779, -18.2460282, -14.0647821, -9.2547256, -5.32280459, -0.299949484, 6.11048684, 14.2051515, 24.2806916, 37.6347845, 53.3127896, 71.3610962, 91.826094, 114.754173, 139.419096, 167.584199, 200.252035, 238.033086, 280.737356, 328.534246, 381.593159, 438.534648, 501.428975, 574.036057, 660.206351, 757.531627, 864.149762, 979.775318, 1106.79446, 1246.36703, 1399.87178, 1568.5253, 1752.74162, 1953.38876, 2170.73799, 2406.34587, 2661.28695, 2936.63582, 3233.46705, 3633.92694, 4037.75039, 4425.74403, 4778.71451, 5077.46845, 5217.22398, 5321.43526, 5407.5897, 5546.84317, 5735.43934, 5947.26003, 6179.66702, 6430.02206, 6695.68693, 6974.02338, 7694.53124, 8078.36407, 8250.88475, 8337.45618, 8463.44126, 8754.20287, 9335.10392, 10143.0506, 10950.1515},
        {91.5947241, 91.739401, 91.8529126, 91.9356844, 91.9881421, 92.0107111, 92.0038169, 91.9678852, 91.9033413, 91.810611, 91.6901196, 91.5422928, 91.367556, 91.1663349, 90.9390549, 90.6861415, 90.4080204, 90.105117, 89.7778569, 89.4266656, 89.0519686, 88.6541915, 88.2337599, 87.7910991, 87.3266349, 86.8407927, 86.333998, 85.8066765, 85.2592535, 84.6921547, 84.1058056, 83.5006317, 82.8770586, 82.2355117, 81.5824321, 80.9151109, 80.2307673, 79.5296315, 78.811934, 78.077905, 77.3277748, 76.5617738, 75.7801323, 74.9830806, 74.1708491, 73.3436681, 72.5017678, 71.6453787, 70.7747311, 69.8900552, 68.9915814, 68.0753565, 67.145885, 66.203488, 65.2484865, 64.2812015, 63.301954, 62.311065, 61.3088556, 60.2956467, 59.2717594, 58.2375146, 57.1932335, 56.139237, 55.0758462, 54.003382, 52.9221654, 51.8766507, 50.9102322, 49.9322098, 48.9400436, 47.9311935, 46.9031197, 45.8532822, 44.779141, 43.6781562, 42.5477878, 41.3854959, 40.1887405, 38.9549817, 37.6816795, 36.366294, 35.0062852, 33.5991132, 31.8378602, 29.8785774, 27.8809743, 25.8523885, 23.8001577, 21.7316197, 19.654112, 17.5749725, 15.5015387, 13.4411484, 11.4011393, 9.38884908, 7.4116154, 5.47677595, 3.59166845, 1.76363056, 0, 0.951050035, 1.57286065, 1.88727715, 1.91614483, 1.68130899, 1.20461492, 0.507907928, -0.386966691, -1.45816364, -2.68383762, -4.04214333, -5.51123547, -7.06926875, -8.69439787, -10.3647775, -12.0585624, -13.9635181, -16.0877437, -17.9352742, -19.5291807, -20.8925347, -22.0484074, -23.0198702, -23.8299944, -24.5018513, -25.0585122, -25.5230485, -25.9185313, -26.2680322, -27.0684938, -27.6953631, -28.02955, -27.9519643, -26.9553083, -25.1145957, -22.6083622, -19.4339799, -15.588821, -11.0702577, -5.87566209, -1.49203787, 4.02624045, 10.9732504, 19.6430694, 30.329775, 44.5832464, 61.1278089, 79.9435897, 101.010716, 124.309314, 148.76113, 176.69825, 209.65063, 248.397787, 292.187106, 340.953866, 394.633348, 451.171566, 512.945172, 584.50093, 670.365831, 767.456116, 873.452198, 987.842599, 1113.17957, 1250.60907, 1401.54112, 1567.19013, 1747.80224, 1944.17129, 2155.97528, 2385.29556, 2633.31223, 2901.2054, 3190.15514, 3583.41812, 3979.57874, 4359.29796, 4703.23674, 4992.05604, 5119.77967, 5211.46383, 5284.8997, 5412.58078, 5590.64522, 5792.43854, 6015.23853, 6256.32297, 6512.96965, 6782.45637, 7485.649, 7850.81447, 8008.85929, 8090.68995, 8227.21296, 8549.33481, 9187.962, 9966.03689, 10671.4956},
        {90.5861288, 90.7253899, 90.8338087, 90.9118144, 90.9598362, 90.9783034, 90.9676452, 90.9282907, 90.8606692, 90.7652098, 90.6423419, 90.4924946, 90.316097, 90.1135785, 89.8853683, 89.6318955, 89.3535894, 89.0508791, 88.7241939, 88.373963, 88.0006157, 87.604581, 87.1862882, 86.7461666, 86.2846454, 85.8021537, 85.2991208, 84.7759759, 84.2331481, 83.6710668, 83.090161, 82.4908601, 81.8735933, 81.2387897, 80.5934863, 79.9346695, 79.2592462, 78.5674311, 77.8594391, 77.1354849, 76.3957833, 75.6405491, 74.869997, 74.0843418, 73.2837983, 72.4685813, 71.6389056, 70.7949859, 69.9370369, 69.0652736, 68.1799106, 67.2749337, 66.3569217, 65.4262244, 64.4831917, 63.5281734, 62.5615193, 61.5835791, 60.5947028, 59.59524, 58.5855407, 57.5659546, 56.5368316, 55.4985214, 54.4513739, 53.3957388, 52.331966, 51.3064516, 50.3644859, 49.411452, 48.4447147, 47.4616387, 46.4595888, 45.43593, 44.3880271, 43.3132448, 42.2089479, 41.0725014, 39.90127, 38.6926185, 37.4439117, 36.1525146, 34.8157918, 33.4311083, 31.6814941, 29.7280931, 27.7358271, 25.7122616, 23.6649622, 21.6014945, 19.5294241, 17.4563166, 15.3897376, 13.3372526, 11.3064272, 9.30482705, 7.3400177, 5.41956474, 3.55103374, 1.7419903, 0, 0.596264668, 0.945307251, 1.06296342, 0.965068862, 0.66745924, 0.185970233, -0.463562484, -1.26530324, -2.20341635, -3.26206614, -4.42541695, -5.67763309, -7.00287889, -8.38531868, -9.80911677, -11.2584375, -12.7741683, -14.3485767, -15.8375249, -17.2373322, -18.5443177, -19.7548009, -20.865101, -21.8715372, -22.7704288, -23.5580951, -24.2308553, -24.7850287, -25.2169346, -25.8185788, -26.1821756, -26.2153383, -25.8256804, -24.7571964, -22.999309, -20.5850725, -17.4711858, -13.6143478, -8.97125731, -3.49861333, 1.80773735, 8.34650615, 16.3643057, 26.1077487, 37.8234477, 53.0996381, 70.5059041, 89.9534528, 111.353491, 134.617226, 158.202524, 185.250237, 217.9058, 257.189906, 301.826899, 351.572137, 406.180978, 463.539444, 525.697077, 596.742698, 680.06567, 774.043353, 877.947461, 991.813755, 1116.97262, 1254.16019, 1404.13093, 1567.62575, 1745.31825, 1937.9201, 2144.26316, 2367.22812, 2608.17736, 2868.47324, 3149.47815, 3536.79466, 3926.4849, 4298.8512, 4634.1959, 4912.82131, 5026.7158, 5103.37166, 5160.94007, 5272.23024, 5433.7579, 5619.26455, 5826.40463, 6052.83254, 6296.2027, 6554.16952, 7102.16155, 7417.51114, 7601.53359, 7755.5442, 7980.85824, 8378.79101, 9050.6578, 9817.66513, 10381.8117},
        {89.7128469, 89.8448223, 89.94645, 90.0181545, 90.0603603, 90.0734917, 90.0579732, 90.0142293, 89.9426845, 89.843763, 89.7178895, 89.5654883, 89.3869839, 89.1828008, 88.9533633, 88.699096, 88.4204232, 88.1177694, 87.7915591, 87.4422167, 87.0701667, 86.6758334, 86.2596415, 85.8220151, 85.363379, 84.8841574, 84.3847748, 83.8656557, 83.3272245, 82.7699056, 82.1941236, 81.6003028, 80.9888677, 80.3602428, 79.7214778, 79.0695453, 78.4013379, 77.717065, 77.0169362, 76.3011608, 75.5699482, 74.823508, 74.0620495, 73.2857822, 72.4949156, 71.6896589, 70.8702218, 70.0368137, 69.1896439, 68.3289219, 67.4548571, 66.5604849, 65.6533444, 64.7338004, 63.802218, 62.8589621, 61.9043977, 60.9388897, 59.9628031, 58.9765029, 57.9803539, 56.9747212, 55.9599698, 54.9364645, 53.9045703, 52.8646522, 51.8170752, 50.8096791, 49.8891641, 48.9579814, 48.0134186, 47.0527629, 46.0733017, 45.0723223, 44.0471122, 42.9949586, 41.9131489, 40.7989705, 39.6497107, 38.4626568, 37.2350962, 35.9643163, 34.6476044, 33.2822479, 31.5425261, 29.5940245, 27.6062106, 25.5868541, 23.5437246, 21.4845916, 19.4172248, 17.3493936, 15.2888677, 13.2434166, 11.22081, 9.22881742, 7.27520842, 5.3677526, 3.51421953, 1.72237881, 0, 0.266282478, 0.361057238, 0.294649958, 0.0773863184, -0.280408003, -0.768407328, -1.37628598, -2.09371827, -2.91037853, -3.81594107, -4.80008023, -5.85247031, -6.96278564, -8.12070055, -9.31588935, -10.5380264, -11.6872713, -12.7405108, -13.8898925, -15.1059091, -16.3590532, -17.6198174, -18.8586943, -20.0461767, -21.152757, -22.148928, -23.0051822, -23.6920124, -24.179911, -24.5562163, -24.6317321, -24.3418973, -23.6221508, -22.4689221, -20.7871549, -18.4655289, -15.421186, -11.5712681, -6.83291696, -1.12327453, 5.06813688, 12.5707586, 21.5794363, 32.2890156, 44.8943421, 60.7258008, 78.558813, 98.3043397, 119.873341, 143.176779, 166.57359, 193.423676, 226.02281, 265.466649, 310.400612, 360.570225, 415.721013, 474.138463, 537.359966, 608.515684, 689.543441, 780.496825, 882.292985, 995.561037, 1120.44204, 1257.30107, 1406.29003, 1567.71875, 1742.67869, 1931.81916, 2133.42266, 2350.86945, 2585.62815, 2839.16735, 3112.95565, 3493.56502, 3876.08487, 4240.70794, 4567.62702, 4837.03485, 4940.46744, 5005.87882, 5051.45753, 5148.73454, 5294.78366, 5464.45265, 5655.92663, 5867.3907, 6097.02995, 6343.02949, 6698.58699, 6957.18462, 7179.19156, 7424.97696, 7754.91002, 8229.3599, 8908.69576, 9647.51447, 10080.6392},
        {89.0085334, 89.1255284, 89.2131763, 89.2718777, 89.3020328, 89.3040422, 89.2783061, 89.225225, 89.1451993, 89.0386294, 88.9059156, 88.7474585, 88.5636584, 88.3549156, 88.1216307, 87.8642039, 87.5830358, 87.2785266, 86.9510769, 86.601087, 86.2289572, 85.8350881, 85.41988, 84.9837333, 84.5270484, 84.0502258, 83.5536657, 83.0377687, 82.502935, 81.9495652, 81.3780596, 80.7888187, 80.1822427, 79.5587322, 78.9237268, 78.2750011, 77.6102696, 76.929769, 76.2337362, 75.5224081, 74.7960215, 74.0548134, 73.2990205, 72.5288798, 71.7446282, 70.9465024, 70.1347394, 69.3095761, 68.4712493, 67.6199958, 66.7560526, 65.8758651, 64.9835438, 64.0794077, 63.163776, 62.2369675, 61.2993015, 60.351097, 59.392673, 58.4243485, 57.4464427, 56.4592746, 55.4631633, 54.4584278, 53.4453872, 52.4243605, 51.3956668, 50.4067273, 49.5038323, 48.5901562, 47.6629646, 46.7195232, 45.7570975, 44.7729533, 43.7643561, 42.7285715, 41.6628652, 40.5645027, 39.4307498, 38.2588719, 37.0461348, 35.789804, 34.4871453, 33.135424, 31.4045199, 29.4622776, 27.480477, 25.4670079, 23.4297603, 21.3766242, 19.3154894, 17.2542458, 15.2007835, 13.1629923, 11.1487622, 9.16598309, 7.2225449, 5.32633757, 3.48525102, 1.70717519, 0, 0.210694432, 0.266424265, 0.176484462, -0.0498300118, -0.403224194, -0.87440312, -1.45407183, -2.13293535, -2.90169873, -3.751067, -4.67174519, -5.65443835, -6.6898515, -7.76868969, -8.88165795, -10.0194613, -11.0427831, -11.9237539, -12.9471754, -14.0758908, -15.2727434, -16.5005763, -17.722233, -18.9005565, -19.9983902, -20.9785773, -21.8039611, -22.4373847, -22.8416915, -23.0406385, -22.9138201, -22.4058053, -21.4611632, -20.1507642, -18.3560267, -15.9246888, -12.7634292, -8.77892673, -3.87786002, 2.0330922, 8.70072113, 16.6703438, 26.103081, 37.1600532, 50.0023812, 64.9829388, 82.0231552, 101.236213, 122.735294, 146.63358, 172.346031, 201.537438, 235.204479, 273.972266, 317.707292, 366.616652, 420.907442, 480.238642, 545.490038, 618.139365, 698.882894, 788.758753, 889.244205, 1001.21879, 1124.54926, 1259.56725, 1406.36999, 1565.22832, 1737.27264, 1923.14712, 2121.61458, 2335.49002, 2566.06884, 2814.64649, 3082.51838, 3452.12969, 3823.33866, 4177.15325, 4494.58147, 4756.63128, 4859.81008, 4925.96019, 4972.30412, 5065.09471, 5202.14666, 5360.84538, 5540.02704, 5738.52779, 5955.18381, 6188.83127, 6333.15413, 6532.76025, 6798.44239, 7140.99333, 7571.20584, 8099.87268, 8737.78664, 9395.21545, 9761.14333},
        {88.4407418, 88.5405767, 88.612147, 88.6558255, 88.6719844, 88.6609966, 88.6232343, 88.5590703, 88.4688769, 88.3530268, 88.2118926, 88.0458466, 87.8552615, 87.6405098, 87.4019641, 87.1399968, 86.8549805, 86.5472878, 86.2172912, 85.8653632, 85.4918763, 85.0972031, 84.6817162, 84.245788, 83.7897911, 83.314098, 82.8190813, 82.3051136, 81.7725672, 81.2218149, 80.653229, 80.0671822, 79.464047, 78.8441959, 78.2109354, 77.5631094, 76.8995265, 76.2204639, 75.526199, 74.8170091, 74.0931715, 73.3549636, 72.6026627, 71.8365461, 71.0568911, 70.2639751, 69.4580753, 68.6394691, 67.8084338, 66.9652468, 66.1101854, 65.2449368, 64.368338, 63.4806355, 62.5820762, 61.6729068, 60.753374, 59.8237247, 58.8842056, 57.9350633, 56.9765448, 56.0088967, 55.0323658, 54.0471989, 53.0536426, 52.0519439, 51.0423493, 50.0710206, 49.1830171, 48.2838893, 47.3709078, 46.4413426, 45.4924641, 44.5215426, 43.5258483, 42.5026516, 41.4492227, 40.3628318, 39.2407494, 38.0802456, 36.8785907, 35.6330551, 34.3409089, 32.9994225, 31.2762809, 29.3404794, 27.3649747, 25.3577326, 23.3267192, 21.2799004, 19.2252422, 17.1707104, 15.1242711, 13.0938903, 11.0875338, 9.11316759, 7.1787577, 5.29227004, 3.46167056, 1.69492523, 0, 0.328635235, 0.481704786, 0.470278843, 0.305427601, -0.00177874853, -0.440270012, -0.998975997, -1.66682651, -2.43275136, -3.28568035, -4.2145433, -5.20827, -6.25579026, -7.3460339, -8.46793072, -9.61041052, -10.6563596, -11.5795585, -12.6087977, -13.7102833, -14.8502215, -15.9948184, -17.1102803, -18.1628133, -19.1186237, -19.9439175, -20.6049011, -21.0677805, -21.298762, -21.3409207, -21.0554086, -20.3853711, -19.2739538, -17.7686864, -15.7605222, -13.1125789, -9.7366868, -5.54467605, -0.448376938, 5.6403803, 12.550454, 20.7082462, 30.2526615, 41.3226045, 54.0569797, 67.5367846, 83.2233077, 101.51993, 122.830034, 147.557001, 176.675621, 209.319476, 244.564581, 282.328218, 324.24526, 371.179417, 423.994398, 484.071942, 551.639997, 626.361394, 707.975737, 797.979445, 897.717846, 1008.09129, 1129.24614, 1261.67485, 1405.74238, 1561.90815, 1731.099, 1913.97736, 2110.23086, 2321.6467, 2549.22475, 2793.96488, 3056.86693, 3414.03867, 3772.59509, 4114.7591, 4422.75358, 4678.80142, 4786.71945, 4859.40734, 4912.69059, 5005.87936, 5137.58646, 5288.30611, 5457.55706, 5644.85805, 5849.72784, 6071.68518, 6020.79174, 6168.24533, 6476.07125, 6906.29482, 7420.94135, 7982.03615, 8551.60453, 9103.5276, 9430.07774},
        {87.976596, 88.0627374, 88.1213411, 88.152764, 88.1573628, 88.1354945, 88.0875159, 88.0137837, 87.914655, 87.7904864, 87.6416349, 87.4684572, 87.2713103, 87.0505509, 86.8065359, 86.5396221, 86.2501664, 85.9385256, 85.6050565, 85.2501161, 84.874061, 84.4772482, 84.0600346, 83.6227768, 83.1658319, 82.6895566, 82.1943077, 81.6804421, 81.1483167, 80.5982882, 80.0307136, 79.4459496, 78.8443531, 78.2262809, 77.5935127, 76.9456638, 76.2823329, 75.6038305, 74.9104674, 74.2025542, 73.4804015, 72.7443201, 71.9946207, 71.2316138, 70.4556101, 69.6669205, 68.8658554, 68.0527255, 67.2278416, 66.3915143, 65.5440543, 64.6916613, 63.8286293, 62.9551412, 62.0713801, 61.1775289, 60.2737707, 59.3602885, 58.4372653, 57.5048841, 56.5633279, 55.6127797, 54.6534225, 53.6854394, 52.7090132, 51.7243271, 50.731564, 49.7758173, 48.9011021, 48.014977, 47.1147139, 46.1975844, 45.26086, 44.3018125, 43.3177136, 42.3058348, 41.2634478, 40.1878242, 39.0762358, 37.9259542, 36.734251, 35.4983978, 34.2156664, 32.8833284, 31.166917, 29.2365508, 27.2663355, 25.2643084, 23.2385065, 21.1969672, 19.1477276, 17.0988249, 15.0582963, 13.034179, 11.0345103, 9.06732727, 7.14066717, 5.2625672, 3.44106454, 1.68419641, 0, 0.521206106, 0.830734193, 0.942591459, 0.870785104, 0.629322327, 0.232210329, -0.306543691, -0.972932535, -1.752949, -2.63258589, -3.59783601, -4.63469214, -5.72914711, -6.8671937, -8.03482471, -9.21803295, -10.3445736, -11.3921197, -12.4783921, -13.5769042, -14.6611693, -15.7047006, -16.6810114, -17.5636151, -18.3260249, -18.9417542, -19.3843163, -19.6272244, -19.6439919, -19.5223593, -19.0797294, -18.2553472, -16.9884579, -15.285244, -13.051482, -10.1750982, -6.57525648, -2.17112081, 3.11814496, 9.37337695, 16.4658881, 24.7498056, 34.3469592, 45.3791787, 57.968294, 70.0599981, 84.4962917, 101.943039, 123.066103, 148.531348, 180.719452, 216.485582, 253.200671, 290.148293, 330.517001, 375.742089, 427.258856, 487.884034, 557.35752, 633.912192, 716.705921, 807.31785, 906.651332, 1015.50651, 1134.50861, 1264.36317, 1405.80244, 1559.53888, 1726.18688, 1906.41632, 2100.69128, 2309.91906, 2534.84073, 2776.19738, 3034.73011, 3380.92398, 3728.34009, 4060.28354, 4360.05942, 4610.97282, 4723.35864, 4802.13895, 4861.89177, 4955.69612, 5082.89006, 5226.76954, 5387.47473, 5565.14576, 5759.9228, 5971.94599, 5778.28546, 5889.95735, 6230.83088, 6724.77522, 7295.65958, 7867.35315, 8363.72512, 8815.59539, 9094.57904},
        {87.5465505, 87.6368163, 87.6986998, 87.7326, 87.7389162, 87.7180472, 87.6703924, 87.5963507, 87.4963212, 87.370703, 87.2198953, 87.0442971, 86.8443074, 86.6203255, 86.3727503, 86.101981, 85.8084167, 85.4924564, 85.1544993, 84.7949443, 84.4141908, 84.0126376, 83.5906839, 83.1487288, 82.6871714, 82.2064108, 81.706846, 81.1888762, 80.6529005, 80.0993179, 79.5285275, 78.9409284, 78.3369198, 77.7169006, 77.0846631, 76.4388384, 75.778017, 75.1024878, 74.4125398, 73.708462, 72.9905434, 72.2590729, 71.5143396, 70.7566324, 69.9862402, 69.2034522, 68.4085572, 67.6018442, 66.7836022, 65.9541202, 65.1136872, 64.2667356, 63.409321, 62.5416428, 61.6639, 60.7762917, 59.8790172, 58.9722756, 58.056266, 57.1311876, 56.1972396, 55.2546211, 54.3035313, 53.3441693, 52.3767343, 51.4014254, 50.4184419, 49.4741395, 48.6137658, 47.7423253, 46.8570251, 45.9550722, 45.0336735, 44.0900361, 43.121367, 42.1248732, 41.0977617, 40.0372395, 38.9405137, 37.8047911, 36.6272789, 35.4051841, 34.1357136, 32.8160745, 31.1053253, 29.1768597, 27.2081358, 25.2073341, 23.1826353, 21.1422198, 19.0942684, 17.0469615, 15.0084798, 12.9870037, 10.990714, 9.02779112, 7.10641569, 5.23476829, 3.42102949, 1.67337987, 0, 0.672794714, 1.10858825, 1.32346605, 1.33351355, 1.15481621, 0.803459467, 0.295528764, -0.352890455, -1.12571275, -2.00685266, -2.98022476, -4.0297436, -5.13932374, -6.29287972, -7.47432612, -8.66757747, -9.81423923, -10.8921818, -11.9810189, -13.0555988, -14.09077, -15.0613806, -15.9422791, -16.7083137, -17.3343327, -17.7951846, -18.0657174, -18.1207797, -17.9352197, -17.5796277, -16.8980046, -15.8364762, -14.3411682, -12.4936796, -10.1723994, -7.21959056, -3.54073665, 0.958678596, 6.37317147, 12.7972583, 20.2856871, 28.9848458, 38.9970315, 50.4245411, 63.3696717, 75.6863188, 90.2872814, 107.836957, 128.999743, 154.440036, 186.448253, 222.075417, 258.941141, 296.407946, 337.396796, 383.230986, 435.233812, 495.896618, 565.109186, 641.487067, 724.578586, 815.505813, 914.76381, 1022.98539, 1141.03677, 1269.67705, 1409.78698, 1562.1572, 1727.13216, 1905.30869, 2096.49511, 2302.19809, 2523.28743, 2760.63293, 3015.10437, 3358.74096, 3703.45072, 4032.31112, 4328.39958, 4574.79356, 4680.63982, 4752.23359, 4804.33688, 4889.92716, 5008.63949, 5144.18584, 5296.99743, 5467.50548, 5656.14122, 5863.33586, 5687.91024, 5809.91656, 6152.58065, 6639.12836, 7192.78552, 7736.77796, 8194.33153, 8619.8157, 8781.65296},
        {87.2064897, 87.3047229, 87.373389, 87.4129405, 87.4238296, 87.4065085, 87.3614297, 87.2890454, 87.1898079, 87.0641695, 86.9125825, 86.7354992, 86.5333719, 86.306653, 86.0557946, 85.7812492, 85.483469, 85.1629064, 84.8200135, 84.4552428, 84.0690465, 83.661877, 83.2341865, 82.7864273, 82.3190518, 81.8325122, 81.3272608, 80.80375, 80.262432, 79.7037592, 79.1281839, 78.5361582, 77.9281347, 77.3045655, 76.6719267, 76.027532, 75.3686228, 74.6954559, 74.0082881, 73.3073762, 72.592977, 71.8653474, 71.1247441, 70.3714241, 69.605644, 68.8276608, 68.0377313, 67.2361122, 66.4230604, 65.5988327, 64.7636859, 63.9188092, 63.0635069, 62.1980156, 61.3225718, 60.4374122, 59.5427734, 58.638892, 57.7260046, 56.8043478, 55.8741582, 54.9356725, 53.9891271, 53.0347588, 52.0728042, 51.1034998, 50.1270823, 49.1917222, 48.3444393, 47.48661, 46.6153634, 45.7278288, 44.8211355, 43.8924127, 42.9387896, 41.9573956, 40.9453597, 39.8998113, 38.8178797, 37.696694, 36.5333835, 35.3250775, 34.0689051, 32.7619957, 31.0561867, 29.1285159, 27.1601715, 25.159488, 23.1348, 21.0944422, 19.0467491, 17.0000555, 14.962696, 12.9430051, 10.9493175, 8.98996781, 7.07329067, 5.20762069, 3.4012925, 1.66264073, 0, 0.831649043, 1.39995517, 1.72314703, 1.81945327, 1.70710252, 1.40432344, 0.929344659, 0.300394835, -0.464297394, -1.34650338, -2.32799449, -3.39054206, -4.51591746, -5.68589205, -6.88223717, -8.08672419, -9.25633613, -10.3693031, -11.4634655, -12.5152826, -13.5012142, -14.3977198, -15.1812589, -15.8282912, -16.3152762, -16.6186735, -16.7149426, -16.5805432, -16.1919347, -15.6033698, -14.6849909, -13.3899198, -11.6712779, -9.69613953, -7.31065004, -4.3039009, -0.564828056, 4.01763256, 9.55454502, 16.1569734, 24.0651172, 33.221603, 43.7122291, 55.622794, 69.0390961, 82.0485937, 97.2350104, 115.18373, 136.480134, 161.709609, 192.64419, 227.232252, 263.889425, 302.256807, 344.477489, 391.571222, 444.557752, 505.106009, 573.439018, 649.071609, 732.23115, 823.409581, 922.668554, 1030.44729, 1147.82478, 1275.58635, 1414.73327, 1566.10686, 1729.7566, 1906.17985, 2094.20098, 2296.24778, 2513.39675, 2746.72438, 2997.30718, 3340.20349, 3684.0125, 4011.31525, 4304.69279, 4546.72614, 4643.73551, 4705.40333, 4746.99387, 4822.77258, 4932.56576, 5059.94915, 5205.56158, 5370.04183, 5554.0287, 5758.161, 5661.672, 5811.40943, 6143.91146, 6595.71622, 7103.36187, 7603.38656, 8032.32845, 8442.61716, 8457.74511},
        {86.969027, 87.0748677, 87.1500037, 87.1949366, 87.2101679, 87.1961991, 87.1535316, 87.082667, 86.9841067, 86.8583523, 86.7059051, 86.5272667, 86.3229387, 86.0934224, 85.8392193, 85.560831, 85.2587589, 84.9335046, 84.5855695, 84.215455, 83.8236628, 83.4106942, 82.9770508, 82.5232341, 82.0497454, 81.5570864, 81.0457586, 80.5162633, 79.9691021, 79.4047765, 78.823788, 78.226638, 77.613828, 76.9858596, 76.3516972, 75.7074329, 75.0490568, 74.3767958, 73.6908767, 72.9915265, 72.2789719, 71.5534398, 70.815157, 70.0643504, 69.3012468, 68.5260731, 67.7390561, 66.9404227, 66.1303997, 65.3092139, 64.4770923, 63.6317365, 62.7759532, 61.910024, 61.0342306, 60.1488544, 59.2541771, 58.3504803, 57.4380456, 56.5171545, 55.5880886, 54.6511295, 53.7065588, 52.7546582, 51.795709, 50.8299931, 49.8577919, 48.9292162, 48.0931806, 47.2471975, 46.3883183, 45.5135943, 44.620077, 43.7048176, 42.7648676, 41.7972783, 40.799101, 39.7673872, 38.6991881, 37.5915551, 36.4415397, 35.2461931, 34.0025668, 32.707712, 31.0063356, 29.0790724, 27.1107919, 25.1099798, 23.0851217, 21.0447032, 18.99721, 16.9511275, 14.9149415, 12.8971376, 10.9062014, 8.95061845, 7.03887445, 5.17945498, 3.38084568, 1.65153214, 0, 0.989488414, 1.68961245, 2.1207169, 2.30314658, 2.25724629, 2.00336085, 1.56183505, 0.953013696, 0.1972416, -0.685136434, -1.6737756, -2.7483311, -3.88845811, -5.07381185, -6.28404749, -7.49882024, -8.69481604, -9.85126593, -10.9545412, -11.9836609, -12.9176438, -13.7355089, -14.4162751, -14.9389613, -15.2825866, -15.4261698, -15.3487298, -15.0292855, -14.446856, -13.6511216, -12.5245306, -11.0249033, -9.11006031, -7.01920832, -4.56947467, -1.50295031, 2.30696018, 6.98685223, 12.6633213, 19.4629627, 27.784721, 37.3999537, 48.3819706, 60.8040815, 74.7395961, 88.6773723, 104.671285, 123.190756, 144.705209, 169.684067, 199.189422, 232.373652, 268.67415, 308.094456, 351.793997, 400.413231, 454.592615, 515.001139, 582.244245, 656.896246, 739.910418, 831.159072, 930.325403, 1037.67412, 1154.4553, 1281.46671, 1419.81731, 1570.38556, 1732.90888, 1907.77012, 2092.76193, 2291.25587, 2504.53143, 2733.86807, 2980.54527, 3322.5935, 3665.3535, 3990.91696, 4281.37562, 4518.82121, 4606.6979, 4658.17664, 4689.05523, 4755.04105, 4856.13436, 4975.70645, 5114.58903, 5273.61381, 5453.61251, 5655.41684, 5670.31618, 5856.63121, 6172.0182, 6574.13341, 7020.63311, 7469.17356, 7877.41102, 8266.30437, 8110.19314},
        {86.8415902, 86.9414277, 87.0107719, 87.0501181, 87.0599618, 87.0407983, 86.9931232, 86.9174319, 86.8142196, 86.683982, 86.5272144, 86.3444121, 86.1360708, 85.9026856, 85.6447522, 85.3627659, 85.0572221, 84.7286162, 84.3774437, 84.0042, 83.6093805, 83.1934807, 82.7569959, 82.3004215, 81.8242531, 81.328986, 80.8151156, 80.2831374, 79.7335467, 79.1668391, 78.5835099, 77.9840545, 77.3689684, 76.7387469, 76.1016864, 75.4542178, 74.7926782, 74.1173099, 73.4283551, 72.7260562, 72.0106554, 71.282395, 70.5415172, 69.7882643, 69.0228786, 68.2456024, 67.4566779, 66.6563474, 65.8448531, 65.0224374, 64.1893425, 63.3435784, 62.487668, 61.621902, 60.746571, 59.8619658, 58.968377, 58.0660951, 57.155411, 56.2366153, 55.3099985, 54.3758515, 53.4344648, 52.4861291, 51.531135, 50.5697733, 49.6023346, 48.6796523, 47.8513465, 47.013467, 46.163028, 45.2970437, 44.4125283, 43.5064958, 42.5759606, 41.6179367, 40.6294384, 39.6074798, 38.5490752, 37.4512385, 36.3109842, 35.1253263, 33.8912789, 32.6058564, 30.9089917, 28.9837502, 27.0174894, 25.0188112, 22.996318, 20.9586118, 18.9142947, 16.871969, 14.8402367, 12.8277, 10.842961, 8.89462194, 6.99128491, 5.14155205, 3.35402552, 1.63730745, 0, 1.15690964, 1.99638663, 2.54105976, 2.8135578, 2.83650953, 2.63254373, 2.22428917, 1.63437464, 0.885428906, 0.000080751784, -0.999041045, -2.08930771, -3.24809046, -4.45276052, -5.68068911, -6.90924746, -8.16776826, -9.44104582, -10.5889194, -11.5998949, -12.4624781, -13.1651748, -13.6964908, -14.044932, -14.1990043, -14.1472133, -13.878065, -13.3800652, -12.6417197, -11.7902222, -10.6245385, -9.09156791, -7.13820993, -4.95797399, -2.37445454, 0.854516619, 4.85602304, 9.75714827, 15.6849759, 22.7665893, 31.4363747, 41.4205862, 52.7861831, 65.6001244, 79.9293695, 94.3940049, 110.86958, 129.784773, 151.568261, 176.648721, 205.836544, 238.712154, 275.070782, 315.044522, 359.453229, 408.807965, 463.619794, 524.318685, 591.515228, 665.908479, 748.415957, 838.955789, 937.204266, 1043.4484, 1159.01003, 1284.73599, 1421.91933, 1571.52256, 1732.87213, 1906.21995, 2089.07958, 2284.86255, 2494.76865, 2719.99766, 2961.74939, 3295.46112, 3629.53576, 3946.61374, 4229.33546, 4460.34134, 4546.06972, 4596.16447, 4625.92404, 4687.94875, 4782.77172, 4895.38361, 5027.09467, 5179.21512, 5353.05522, 5549.92519, 5641.34827, 5852.08707, 6154.67998, 6521.66535, 6925.58155, 7338.96694, 7734.35991, 8042.84947, 7674.33986},
        {86.8441654, 86.9330495, 86.9920873, 87.0217493, 87.0225063, 86.9948288, 86.9391875, 86.8560531, 86.7458961, 86.6091873, 86.4463973, 86.2579966, 86.0444561, 85.8062462, 85.5438376, 85.2577011, 84.9483072, 84.6161265, 84.2616297, 83.8852875, 83.4875705, 83.0689493, 82.6298946, 82.170877, 81.6923672, 81.1948358, 80.6787534, 80.1445907, 79.5928184, 79.023907, 78.4383272, 77.8365497, 77.2190451, 76.586284, 75.9448487, 75.2920254, 74.625027, 73.9441259, 73.2495943, 72.5417045, 71.8207289, 71.0869398, 70.3406094, 69.5820102, 68.8114143, 68.0290941, 67.2353219, 66.4303701, 65.6145109, 64.7880166, 63.9511596, 63.1032933, 62.2456288, 61.3784583, 60.5020741, 59.6167683, 58.7228333, 57.8205612, 56.9102442, 55.9921747, 55.0666448, 54.1339467, 53.1943727, 52.248215, 51.2957659, 50.3373175, 49.3731622, 48.4545141, 47.631365, 46.7989841, 45.9543626, 45.0944917, 44.2163626, 43.3169663, 42.3932941, 41.4423371, 40.4610865, 39.4465335, 38.3956691, 37.3054846, 36.1729711, 34.9951198, 33.7689218, 32.4913683, 30.7987795, 28.8760684, 26.9124826, 24.9167179, 22.8974698, 20.8634341, 18.8233062, 16.7857819, 14.7595567, 12.7533263, 10.7757864, 8.83563242, 6.94156013, 5.10226512, 3.32644301, 1.62278943, 0, 1.29879373, 2.25664732, 2.8981415, 3.24785705, 3.33037472, 3.17027527, 2.79213945, 2.22054803, 1.48008175, 0.595321382, -0.409152323, -1.50875861, -2.67891671, -3.89504588, -5.13256536, -6.36689439, -7.67904511, -9.05952147, -10.2449016, -11.2332319, -12.0225585, -12.6109278, -12.996386, -13.1769795, -13.1507546, -12.9157575, -12.4700345, -11.811632, -10.9385962, -10.0642705, -8.89246177, -7.35662732, -5.3902243, -3.1151159, -0.370556286, 3.05444196, 7.28294346, 12.4380128, 18.6427147, 26.0201137, 34.9845074, 45.2790913, 56.9699497, 70.1231669, 84.8048271, 99.7373157, 116.666341, 135.993911, 158.122035, 183.452723, 212.651917, 245.535106, 282.077475, 322.477202, 367.387744, 417.258146, 472.537454, 533.617073, 601.016789, 675.31927, 757.208701, 846.827881, 943.961515, 1048.97287, 1163.20558, 1287.55324, 1423.47923, 1572.02487, 1732.14247, 1903.96602, 2085.02344, 2278.45569, 2485.29882, 2706.58891, 2943.36203, 3267.08065, 3590.74782, 3897.79299, 4171.64564, 4395.7352, 4480.79655, 4531.41682, 4562.18029, 4621.72495, 4711.15551, 4817.38764, 4942.23551, 5087.51329, 5255.03514, 5446.61522, 5606.98839, 5839.14732, 6129.34663, 6463.84094, 6828.88488, 7210.73306, 7595.64013, 7797.68094, 7211.44455},
        {86.9930403, 87.0682282, 87.1143916, 87.1319685, 87.1213968, 87.0831142, 87.0175587, 86.925168, 86.8063799, 86.6616325, 86.4913634, 86.2960105, 86.0760117, 85.8318048, 85.5638277, 85.2725181, 84.9583141, 84.6216532, 84.2629736, 83.8827129, 83.481309, 83.0591997, 82.616823, 82.1546166, 81.6730184, 81.1724663, 80.653398, 80.1162514, 79.5614644, 78.9894748, 78.4007204, 77.7956392, 77.1746688, 76.5382473, 75.8908141, 75.2307221, 74.5562759, 73.8677836, 73.1655531, 72.4498925, 71.7211096, 70.9795124, 70.225409, 69.4591074, 68.6809154, 67.8911411, 67.0900924, 66.2780774, 65.455404, 64.6223802, 63.779314, 62.9271797, 62.0656045, 61.1948819, 60.3153055, 59.4271687, 58.5307652, 57.6263884, 56.7143319, 55.7948892, 54.8683539, 53.9350194, 52.9951793, 52.0491272, 51.0971566, 50.139561, 49.1766339, 48.2598508, 47.4394583, 46.6101911, 45.7690248, 44.9129349, 44.0388972, 43.1438872, 42.2248805, 41.2788527, 40.3027796, 39.2936366, 38.2483995, 37.1640437, 36.037545, 34.865879, 33.6460213, 32.3749474, 30.6863497, 28.7664867, 26.8059831, 24.8136039, 22.7981139, 20.768278, 18.7328609, 16.7006277, 14.6803431, 12.6807719, 10.7106791, 8.77882957, 6.89398808, 5.06491952, 3.30038874, 1.60916062, 0, 1.39348525, 2.43104808, 3.13859399, 3.54202852, 3.66725717, 3.54018547, 3.18671894, 2.6327631, 1.90422346, 1.02700554, 0.0270148762, -1.06984303, -2.23766265, -3.45053847, -4.68256497, -5.90783662, -7.24258187, -8.68120003, -9.87629214, -10.8324099, -11.5541052, -12.0459297, -12.3124351, -12.3581733, -12.1876959, -11.8055548, -11.2163017, -10.4244883, -9.4346664, -8.52385677, -7.32423684, -5.75861766, -3.74981029, -1.35613231, 1.56735851, 5.20173944, 9.66485138, 15.0745353, 21.548632, 29.2049824, 38.4016314, 48.9431105, 60.900264, 74.3439365, 89.3449724, 104.740324, 122.143201, 141.932921, 164.488801, 190.19016, 219.609498, 252.696836, 289.510315, 330.279087, 375.581451, 425.838945, 481.473108, 542.918213, 610.58017, 684.850998, 766.143822, 854.85455, 950.904101, 1054.73754, 1167.68807, 1290.68122, 1425.29959, 1572.63905, 1731.38627, 1901.59073, 2080.99995, 2272.31946, 2476.39552, 2694.07443, 2926.20244, 3240.21205, 3553.71678, 3850.91634, 4116.01048, 4333.1989, 4417.33612, 4468.19723, 4499.70135, 4556.8896, 4641.36195, 4741.75088, 4860.30474, 4999.27189, 5160.90066, 5347.43942, 5568.61769, 5819.46378, 6098.22434, 6403.14599, 6732.47536, 7084.45906, 7457.34372, 7534.59402, 6745.68346},
        {87.3325968, 87.3942783, 87.4275588, 87.4328506, 87.4105662, 87.3611178, 87.2849179, 87.1823788, 87.053913, 86.8999327, 86.7208504, 86.5170784, 86.2890292, 86.037115, 85.7617483, 85.4633415, 85.1423069, 84.7990568, 84.4340037, 84.04756, 83.640138, 83.2121501, 82.7640087, 82.2961261, 81.8089147, 81.3027869, 80.7781551, 80.2354317, 79.675029, 79.0973593, 78.5028352, 77.8918689, 77.2648728, 76.6222594, 75.9667655, 75.2975924, 74.6139133, 73.9160651, 73.2043849, 72.4792094, 71.7408757, 70.9897206, 70.2260812, 69.4502944, 68.662697, 67.8636261, 67.0534186, 66.2324113, 65.4009414, 64.5593456, 63.7079609, 62.8486498, 61.9801907, 61.1028874, 60.2170437, 59.3229635, 58.4209508, 57.5113093, 56.5943429, 55.6703556, 54.7396511, 53.8025333, 52.8593061, 51.9102734, 50.955739, 49.9960067, 49.0313806, 48.1137831, 47.2938974, 46.4655677, 45.6257517, 44.771407, 43.8994913, 43.0069621, 42.0907771, 41.1478939, 40.1752702, 39.1698636, 38.1286318, 37.0485324, 35.9265229, 34.7595612, 33.5446047, 32.2786112, 30.5938931, 28.6774003, 26.7204658, 24.7318807, 22.7204359, 20.6949226, 18.6641318, 16.6368546, 14.6218821, 12.6280054, 10.6640155, 8.73870357, 6.86086059, 5.03927766, 3.28274587, 1.60005629, 0, 1.35593845, 2.36494585, 3.05247487, 3.44397817, 3.56490841, 3.44071824, 3.09686033, 2.55878734, 1.85195193, 1.00180676, 0.0338044884, -1.02660222, -2.15396072, -3.32281833, -4.50772241, -5.68322028, -6.93503977, -8.25361114, -9.36200208, -10.2585843, -10.9417296, -11.4098096, -11.6611961, -11.6942608, -11.5073755, -11.0989118, -10.4672415, -9.61073634, -8.52776802, -7.44454482, -6.04806185, -4.26833989, -2.0353997, 0.558343127, 3.66406523, 7.47624857, 12.1135908, 17.6947897, 24.3385428, 32.1635479, 41.4562284, 52.1165093, 64.2302724, 77.8833993, 93.1617719, 109.123876, 127.139839, 147.55239, 170.704261, 196.938182, 226.707946, 260.10948, 297.279702, 338.472659, 384.181543, 434.792177, 490.690381, 552.242931, 619.859023, 693.968632, 775.000355, 863.313652, 958.806685, 1061.90485, 1173.92691, 1295.78154, 1429.00466, 1574.66755, 1731.54176, 1899.69971, 2077.31603, 2266.6529, 2468.43964, 2683.40558, 2912.28008, 3221.19024, 3529.11817, 3820.44379, 4079.54699, 4290.80766, 4369.72757, 4415.4835, 4441.97456, 4493.79064, 4572.68239, 4667.61264, 4780.97144, 4915.14882, 5072.53484, 5255.51953, 5529.9522, 5797.14126, 6065.36951, 6342.91977, 6638.07482, 6959.11747, 7314.33052, 7248.6562, 6319.78599},
        {87.8762463, 87.9259711, 87.9475478, 87.9413779, 87.9078626, 87.8474035, 87.7604018, 87.647259, 87.5083764, 87.3441555, 87.1549976, 86.9413041, 86.7034763, 86.4419157, 86.1570237, 85.8492016, 85.5188508, 85.1663727, 84.7921687, 84.3966402, 83.9801885, 83.543215, 83.0861212, 82.6093084, 82.1131779, 81.5981312, 81.0645697, 80.5128947, 79.9435077, 79.3568099, 78.7532029, 78.1330879, 77.4968664, 76.8449397, 76.1791959, 75.4992619, 74.8047464, 74.0960026, 73.3733837, 72.6372428, 71.8879331, 71.1258076, 70.3512195, 69.5645219, 68.766068, 67.956211, 67.1353039, 66.3036999, 65.4617521, 64.6098137, 63.7482378, 62.87855, 61.9999055, 61.1126321, 60.2170576, 59.3135095, 58.4023157, 57.4838039, 56.5583018, 55.626137, 54.6876375, 53.7431308, 52.7929446, 51.8374068, 50.876845, 49.911587, 48.9419604, 48.0206564, 47.199109, 46.3696502, 45.5292133, 44.6747314, 43.8031377, 42.9113656, 41.9963481, 41.0550186, 40.0843101, 39.081156, 38.0424895, 36.9652437, 35.8463519, 34.6827473, 33.471363, 32.2091324, 30.5281461, 28.6154324, 26.6624051, 24.6778372, 22.6705018, 20.6491718, 18.6226204, 16.5996206, 14.5889455, 12.5993682, 10.6396616, 8.71859886, 6.84495304, 5.02749719, 3.27500436, 1.59624761, 0, 1.17161014, 2.03190546, 2.60391166, 2.91065444, 2.9751595, 2.82045254, 2.46955924, 1.94550532, 1.27131646, 0.470018359, -0.435363278, -1.42180276, -2.46627438, -3.54575245, -4.63721127, -5.71762515, -6.76616457, -7.76011759, -8.67149092, -9.47772949, -10.1562782, -10.684582, -11.0400857, -11.2002344, -11.1424729, -10.8442461, -10.2829989, -9.43617628, -8.28122314, -6.85834896, -5.05922041, -2.84245302, -0.166662325, 2.7217744, 6.00812249, 9.95438371, 14.6882711, 20.3374975, 27.029776, 34.8928196, 44.1394129, 54.7863058, 66.9445669, 80.7252645, 96.2394673, 112.886397, 131.666931, 152.870098, 176.78493, 203.700456, 233.907839, 267.69137, 305.293112, 346.995614, 393.164077, 444.126592, 500.211252, 561.547945, 628.708011, 702.479016, 783.688018, 872.273447, 967.893312, 1070.81999, 1182.36684, 1303.36942, 1435.12679, 1578.59468, 1733.02915, 1898.64771, 2074.20869, 2261.61322, 2461.58388, 2674.84324, 2902.11389, 3211.7954, 3520.0141, 3810.57333, 4067.27642, 4273.92671, 4342.25431, 4376.18457, 4390.40935, 4432.99821, 4505.431, 4595.28017, 4704.72093, 4835.92849, 4991.07806, 5172.34485, 5488.41864, 5768.78642, 6028.53976, 6282.77025, 6546.56945, 6835.02896, 7163.24033, 6941.7838, 5951.81363},
        {88.5777614, 88.6154085, 88.6250075, 88.6069546, 88.5616462, 88.4894783, 88.3908472, 88.2661491, 88.1157802, 87.9401367, 87.7396148, 87.5146108, 87.2655208, 86.992741, 86.6966678, 86.3776972, 86.0362254, 85.6726488, 85.2873635, 84.8807656, 84.4532515, 84.0052174, 83.5370593, 83.0491737, 82.5419565, 82.0158042, 81.4711128, 80.9082786, 80.3276978, 79.7297666, 79.1148812, 78.4834379, 77.8358328, 77.1724621, 76.4947564, 75.8025729, 75.0957566, 74.3746701, 73.639676, 72.8911369, 72.1294156, 71.3548746, 70.5678766, 69.7687842, 68.9579602, 68.135767, 67.3025674, 66.4587241, 65.6045996, 64.7405566, 63.8669577, 62.9842718, 62.092753, 61.1927617, 60.2846582, 59.3688029, 58.4455561, 57.5152781, 56.5783293, 55.6350701, 54.6858607, 53.7310615, 52.7710329, 51.8061353, 50.8367288, 49.863174, 48.8858311, 47.9582895, 47.1328608, 46.3001243, 45.4569895, 44.6003663, 43.7271644, 42.8342935, 41.9186633, 40.9771836, 40.006764, 39.0043142, 37.966744, 36.8909631, 35.7738813, 34.6124081, 33.4034534, 32.1439269, 30.466085, 28.5568342, 26.6074634, 24.6267163, 22.6233361, 20.6060662, 18.5836502, 16.5648314, 14.5583532, 12.5729591, 10.6173924, 8.70039657, 6.83071504, 5.0170912, 3.26826847, 1.59299026, 0, 0.961723705, 1.65238122, 2.09223765, 2.3015581, 2.3006077, 2.10965153, 1.74895472, 1.23878237, 0.59939959, -0.148928508, -0.985936817, -1.89136023, -2.84493363, -3.82639192, -4.81546998, -5.7919027, -6.61684314, -7.25308592, -7.94858676, -8.65767017, -9.33466064, -9.93388267, -10.4096608, -10.7163194, -10.8081831, -10.6395763, -10.1648235, -9.33824928, -8.11417808, -6.32828149, -4.09704298, -1.41038292, 1.74177832, 4.94666189, 8.42577666, 12.5133938, 17.3464504, 23.0618838, 29.796631, 37.6876293, 46.8710232, 57.4847838, 69.6660034, 83.5517741, 99.2791883, 116.560843, 136.064449, 158.033223, 182.710382, 210.33914, 241.096421, 275.372761, 313.452394, 355.622379, 402.175543, 453.402123, 509.592356, 570.666512, 637.368878, 710.847343, 792.368184, 881.398581, 977.327701, 1080.2543, 1191.47948, 1311.75245, 1442.09567, 1583.3292, 1735.27115, 1898.3064, 2071.78024, 2257.27705, 2455.54156, 2667.3185, 2893.35258, 3205.03676, 3514.80547, 3805.74141, 4060.92725, 4263.44564, 4320.41314, 4341.52264, 4342.41373, 4375.51034, 4441.8916, 4527.3098, 4633.61039, 4762.63885, 4916.24061, 5096.26114, 5429.33233, 5716.5668, 5973.36318, 6215.12011, 6457.23626, 6715.11024, 7004.14071, 6629.08194, 5612.32502},
        {89.4250614, 89.4505741, 89.4479804, 89.4176776, 89.3600628, 89.2755334, 89.1644865, 89.0273195, 88.8644295, 88.6762139, 88.4630697, 88.2253944, 87.9635851, 87.678039, 87.3691535, 87.0373257, 86.6829529, 86.3064324, 85.9081613, 85.488537, 85.0479566, 84.5868174, 84.1055167, 83.6044517, 83.0840196, 82.5446177, 81.9866432, 81.4104934, 80.8165655, 80.2052567, 79.5769643, 78.9320856, 78.2710177, 77.5941579, 76.9029098, 76.1971462, 75.4767281, 74.74202, 73.9933864, 73.2311921, 72.4558015, 71.6675792, 70.8668899, 70.0540982, 69.2295685, 68.3936655, 67.5467539, 66.6891981, 65.8213627, 64.9436125, 64.0563118, 63.1580872, 62.251079, 61.3356896, 60.4123212, 59.4813761, 58.5432566, 57.5983649, 56.6471033, 55.6898741, 54.7270796, 53.759122, 52.7864035, 51.8093265, 50.8282933, 49.843706, 48.855967, 47.9196893, 47.0881845, 46.2500491, 45.402171, 44.5414384, 43.6647391, 42.7689612, 41.8509927, 40.9077216, 39.9360359, 38.9328236, 37.8949726, 36.8193711, 35.7029069, 34.5424682, 33.3349428, 32.0772189, 30.4017257, 28.4952464, 26.5489448, 24.5715363, 22.5717364, 20.5582603, 18.5398234, 16.5251413, 14.5229291, 12.5419024, 10.5907765, 8.67826685, 6.81308875, 5.00395763, 3.25958887, 1.58869787, 0, 0.769768346, 1.30546897, 1.62486147, 1.74570547, 1.68576057, 1.46278638, 1.0945425, 0.59878854, -0.00671588557, -0.704211173, -1.47593771, -2.3041359, -3.17104612, -4.05890878, -4.94996426, -5.82645295, -6.4417415, -6.74537404, -7.2372816, -7.8506602, -8.51870585, -9.17461457, -9.75158238, -10.1828053, -10.4014793, -10.3408005, -9.93396481, -9.11416829, -7.81460697, -5.68588543, -3.04740797, 0.0828519873, 3.68692099, 7.20637385, 10.893463, 15.1445614, 20.103831, 25.915434, 32.7235325, 40.6722885, 49.8254287, 60.4320311, 72.6519949, 86.6452196, 102.571605, 120.374119, 140.483825, 163.114855, 188.481342, 216.797417, 248.202924, 283.077081, 321.662616, 364.22086, 411.051127, 462.435674, 518.656763, 579.502578, 645.861742, 719.161865, 801.058137, 890.584791, 986.880666, 1089.88024, 1200.86615, 1320.50221, 1449.54486, 1588.68196, 1738.26159, 1898.82405, 2070.21011, 2253.76715, 2450.27758, 2660.52381, 2885.28824, 3198.94542, 3510.28758, 3801.69911, 4055.56438, 4254.26777, 4300.4573, 4309.41126, 4297.68675, 4322.08006, 4383.22028, 4464.97014, 4568.79705, 4696.1684, 4848.55159, 5027.41402, 5344.94306, 5630.91717, 5892.41271, 6136.50605, 6370.27357, 6600.79164, 6835.13664, 6318.1044, 5299.07691},
        {90.3986141, 90.4132684, 90.3994482, 90.357565, 90.2880302, 90.1912553, 90.0676518, 89.9176312, 89.7416049, 89.5399844, 89.3131811, 89.0616066, 88.7856723, 88.4857896, 88.1623701, 87.8158252, 87.4465664, 87.0550051, 86.6415529, 86.2066212, 85.7506214, 85.2739651, 84.7770636, 84.2603286, 83.7241714, 83.1690035, 82.5952364, 82.0032816, 81.3935504, 80.7664545, 80.1224053, 79.4618142, 78.7850927, 78.0926523, 77.3868201, 76.6670094, 75.9326106, 75.183973, 74.4214459, 73.6453786, 72.8561204, 72.0540206, 71.2394284, 70.4126933, 69.5741645, 68.7241913, 67.863123, 66.991309, 66.1090984, 65.2168407, 64.3148851, 63.3984152, 62.4730579, 61.5392747, 60.5975267, 59.6482754, 58.6919819, 57.7291076, 56.7601137, 55.7854616, 54.8056125, 53.8210278, 52.8321686, 51.8394964, 50.8434723, 49.8445577, 48.8432138, 47.8953656, 47.0556062, 46.2100074, 45.3554351, 44.488755, 43.6068331, 42.7065351, 41.7847267, 40.838274, 39.8640425, 38.8588982, 37.8197069, 36.7433343, 35.6266463, 34.4665087, 33.2597874, 32.003348, 30.3282201, 28.4218825, 26.4762112, 24.4999441, 22.5018191, 20.4905741, 18.4749469, 16.4636756, 14.465498, 12.4891521, 10.5433757, 8.63690667, 6.77848302, 4.97684262, 3.24072338, 1.57886321, 0, 0.74587896, 1.26400744, 1.57195121, 1.68727606, 1.62754778, 1.41033214, 1.05319494, 0.57370194, -0.0105810631, -0.682088289, -1.42325396, -2.21651228, -3.04429748, -3.88904377, -4.73318538, -5.55915651, -6.09753897, -6.29478808, -6.7030042, -7.25065318, -7.86620087, -8.47811311, -9.01485574, -9.40489462, -9.57669557, -9.45872445, -8.9794471, -8.06732936, -6.65083708, -4.34648123, -1.50906597, 1.83935637, 5.6767335, 9.43765363, 13.3717442, 17.8735285, 23.083962, 29.1440003, 36.194599, 44.3767135, 53.7106782, 64.4947806, 76.8935761, 91.0716201, 107.193468, 125.092107, 145.346552, 168.204252, 193.912652, 222.719201, 254.983085, 290.703443, 329.832585, 372.518948, 419.311402, 470.57903, 526.690915, 587.623265, 654.227329, 727.782945, 809.951153, 899.679556, 996.027865, 1098.87535, 1209.48995, 1328.50219, 1456.59236, 1594.40388, 1742.39768, 1901.13795, 2070.49327, 2251.83038, 2445.95421, 2653.66969, 2875.78175, 3187.37132, 3496.39834, 3785.09874, 4035.70844, 4230.46338, 4270.8144, 4273.63925, 4255.79601, 4275.78716, 4333.98658, 4413.41634, 4515.3179, 4640.93273, 4791.50231, 4968.26812, 5184.27819, 5448.66871, 5737.07657, 6025.13868, 6288.49192, 6502.7732, 6643.6194, 6028.0876, 5007.09056},
        {91.488041, 91.4928875, 91.4686096, 91.4156451, 91.3344318, 91.2254075, 91.0890098, 90.9256767, 90.7358458, 90.519955, 90.2784421, 90.0117447, 89.7203007, 89.4045478, 89.0649239, 88.7018667, 88.315814, 87.9072035, 87.4764731, 87.0240605, 86.5504034, 86.0559398, 85.5411072, 85.0063436, 84.4520867, 83.8787742, 83.286844, 82.6767338, 82.0488814, 81.4037245, 80.741701, 80.0632486, 79.3688052, 78.6588083, 77.937389, 77.2030606, 76.4542925, 75.6914026, 74.9147088, 74.1245291, 73.3211814, 72.5049836, 71.6762536, 70.8353094, 69.982469, 69.1180501, 68.2423709, 67.3557492, 66.4585029, 65.55095, 64.6334084, 63.6961235, 62.7497042, 61.7946866, 60.8316071, 59.8610021, 58.8834076, 57.8993602, 56.9093959, 55.9140512, 54.9138622, 53.9093654, 52.9010969, 51.889593, 50.8753901, 49.8590244, 48.8410322, 47.8788957, 47.0287298, 46.1736241, 45.3104234, 44.4359724, 43.5471158, 42.6406983, 41.7135647, 40.7625596, 39.7845278, 38.7763139, 37.7347627, 36.656719, 35.5390273, 34.3785325, 33.1720793, 31.9165123, 30.2397814, 28.330938, 26.3834673, 24.4061853, 22.407908, 20.3974516, 18.3836321, 16.3752655, 14.381168, 12.4101556, 10.4710443, 8.57265017, 6.72378935, 4.93327786, 3.20993176, 1.56256712, 0, 0.909182859, 1.56283431, 1.98089138, 2.1832911, 2.1899705, 2.0208666, 1.69591642, 1.23505701, 0.658225387, -0.0146414254, -0.763606396, -1.5687325, -2.4100827, -3.26771998, -4.12170731, -4.95210765, -5.56128542, -5.90391808, -6.36200584, -6.8775334, -7.39248547, -7.84884675, -8.18860197, -8.35373582, -8.28623301, -7.92807825, -7.22125625, -6.10775172, -4.52954937, -2.23669276, 0.566513837, 3.88050338, 7.70570884, 11.622236, 15.8406814, 20.6837287, 26.2779092, 32.7497539, 40.2257939, 48.8325604, 58.5792284, 69.7454021, 82.4805734, 96.9342345, 113.255877, 130.81859, 150.74237, 173.370809, 199.047501, 228.116038, 261.405769, 298.18083, 337.866517, 380.416536, 426.863424, 477.755347, 533.640469, 595.009463, 662.48105, 736.736672, 819.024315, 908.589638, 1004.60754, 1107.02513, 1217.09801, 1335.48117, 1462.98951, 1600.31949, 1747.58112, 1905.21611, 2072.62286, 2251.44694, 2442.49125, 2646.55868, 2864.45212, 3169.30071, 3471.49952, 3753.76986, 3998.83306, 4189.41044, 4229.52534, 4233.06238, 4216.47059, 4236.85316, 4294.56684, 4373.02916, 4473.44385, 4597.01462, 4744.9452, 4918.4393, 4949.46112, 5172.50126, 5509.33586, 5881.74104, 6211.49294, 6420.36768, 6430.14138, 5763.91657, 4734.74761},
        {92.6936721, 92.6877135, 92.6519367, 92.5868072, 92.4927907, 92.3703527, 92.2199587, 92.0420743, 91.8371652, 91.6056967, 91.3481346, 91.0649442, 90.7565913, 90.4235414, 90.06626, 89.6852127, 89.280865, 88.8536825, 88.4041308, 87.9326755, 87.4397819, 86.9259159, 86.3915428, 85.8371283, 85.2631379, 84.6700372, 84.0582917, 83.428367, 82.7807287, 82.1158423, 81.4341734, 80.7361874, 80.0223501, 79.2931269, 78.5545706, 77.8042356, 77.0396092, 76.2609757, 75.4686194, 74.6628244, 73.843875, 73.0120555, 72.1676501, 71.3109431, 70.4422187, 69.5617611, 68.6698546, 67.7667834, 66.8528318, 65.928284, 64.9934244, 64.0332461, 63.0636559, 62.0852694, 61.0987025, 60.1045707, 59.1034899, 58.0960757, 57.0829439, 56.0647102, 55.0419903, 54.0153999, 52.9855548, 51.9530707, 50.9185632, 49.8826482, 48.8459414, 47.867445, 47.0047598, 46.1380832, 45.264246, 44.3800791, 43.4824131, 42.5680787, 41.6339069, 40.6767282, 39.6933734, 38.6806734, 37.6354588, 36.5545604, 35.434809, 34.2730353, 33.0660701, 31.810744, 30.1318636, 28.2200966, 26.2705827, 24.2922254, 22.2939284, 20.2845953, 18.2731299, 16.2684357, 14.2794165, 12.3149759, 10.3840176, 8.49544518, 6.65816234, 4.88107274, 3.17308005, 1.54308791, 0, 1.12554858, 1.95836259, 2.52145932, 2.83785603, 2.93057001, 2.82261853, 2.53701886, 2.09678827, 1.52494405, 0.844503468, 0.0784837987, -0.750097683, -1.6182237, -2.50287698, -3.38104025, -4.22969623, -4.95084573, -5.51075979, -6.05602306, -6.54755072, -6.94625794, -7.21305989, -7.30887175, -7.19460868, -6.83118587, -6.17951848, -5.20052169, -3.85511067, -2.10420059, 0.118709491, 2.81523571, 6.01623806, 9.75257653, 13.7944888, 18.303146, 23.5092186, 29.5217532, 36.4497966, 44.4023955, 53.4885965, 63.7165944, 75.3269819, 88.4485375, 103.21004, 119.740268, 136.882801, 156.372916, 178.660692, 204.196206, 233.429537, 267.689168, 305.473166, 345.663261, 388.014733, 434.066371, 484.556204, 540.222262, 602.1261, 670.608695, 745.658089, 828.001794, 917.291152, 1012.8821, 1114.82638, 1224.35474, 1342.15675, 1469.22864, 1606.33948, 1753.13393, 1909.89268, 2075.42, 2251.70014, 2439.52487, 2639.68596, 2852.97518, 3149.72854, 3443.8075, 3718.61774, 3957.56496, 4144.05488, 4185.877, 4192.464, 4179.62849, 4202.15583, 4260.43194, 4338.60413, 4437.87025, 4559.42817, 4704.47575, 4874.21085, 4704.72951, 4882.1275, 5270.35753, 5733.37225, 6135.12436, 6339.56652, 6210.65142, 5513.68738, 4484.36091},
        {94.0084985, 93.9899385, 93.9409166, 93.8619236, 93.7534501, 93.6159868, 93.4500246, 93.2560541, 93.034566, 92.786051, 92.5109999, 92.2099034, 91.8832522, 91.531537, 91.1552486, 90.7548776, 90.3309148, 89.8838509, 89.4141766, 88.9223826, 88.4089596, 87.8743985, 87.3191898, 86.7438243, 86.1487928, 85.5345858, 84.9016943, 84.2506088, 83.5818201, 82.895819, 82.193096, 81.474142, 80.7394477, 79.9895038, 79.2321152, 78.4639617, 77.6816351, 76.8853888, 76.0754762, 75.2521506, 74.4156654, 73.5662739, 72.7042295, 71.8297855, 70.9431954, 70.0447125, 69.13459, 68.2130815, 67.2804402, 66.3369195, 65.3827727, 64.397871, 63.4032919, 62.3997304, 61.3878819, 60.3684416, 59.3421046, 58.3095661, 57.2715215, 56.2286657, 55.1816942, 54.131302, 53.0781845, 52.0230367, 50.9665539, 49.9094313, 48.8523642, 47.8557198, 46.978433, 46.0981294, 45.211636, 44.3157797, 43.4073874, 42.483286, 41.5403024, 40.5752635, 39.5849962, 38.5663273, 37.5160839, 36.4310928, 35.3081808, 34.144175, 32.9359021, 31.6801892, 29.99907, 28.0846513, 26.1335445, 24.1547392, 22.157225, 20.1499917, 18.1420288, 16.1423261, 14.1598731, 12.2036596, 10.2826751, 8.40590945, 6.58235215, 4.8209929, 3.13082136, 1.52082717, 0, 1.36587957, 2.39782185, 3.12222301, 3.56547922, 3.75398664, 3.71414143, 3.47233975, 3.05497776, 2.48845164, 1.79915754, 1.01349162, 0.157850048, -0.741371011, -1.6577754, -2.56496694, -3.43654949, -4.28757811, -5.09757822, -5.74639495, -6.21651454, -6.49042324, -6.55060727, -6.37955289, -5.95974634, -5.27367386, -4.3038217, -3.03267609, -1.44272328, 0.483550485, 2.60489098, 5.15579488, 8.20140633, 11.8068695, 15.9527081, 20.7463766, 26.3178947, 32.7577927, 40.1566008, 48.6048492, 58.1930683, 68.9339098, 81.0194847, 94.5555602, 109.647904, 126.402283, 143.144804, 162.205811, 184.135986, 209.48601, 238.806566, 273.880598, 312.52043, 353.125344, 395.276749, 440.96956, 491.115079, 546.624606, 609.106371, 678.616352, 754.450239, 836.781093, 925.712681, 1020.82325, 1122.2884, 1231.29588, 1348.56886, 1475.29426, 1612.31548, 1758.77543, 1914.7789, 2078.50904, 2252.28795, 2436.88561, 2633.07201, 2841.61713, 3129.53174, 3414.78486, 3681.58628, 3914.1458, 4096.67321, 4141.44891, 4152.56502, 4145.05823, 4170.86455, 4230.41468, 4308.76456, 4407.11362, 4526.66127, 4668.60693, 4834.15002, 4471.75142, 4604.46617, 5040.74588, 5589.04216, 6057.80661, 6255.49085, 5990.54647, 5277.00231, 4255.5491},
        {95.4255111, 95.3917547, 95.3270368, 95.2318667, 95.1067533, 94.952206, 94.7687338, 94.5568459, 94.3170515, 94.0498596, 93.7557796, 93.4353204, 93.0889913, 92.7173014, 92.3207599, 91.8998759, 91.4551586, 90.9871171, 90.4962606, 89.9830982, 89.4481392, 88.8918925, 88.3148675, 87.7175732, 87.1005188, 86.4642134, 85.8091663, 85.1358865, 84.4448832, 83.7366656, 83.0117427, 82.2706239, 81.5138182, 80.7418347, 79.9637729, 79.1756661, 78.3734444, 77.5573382, 76.7275778, 75.8843938, 75.0280164, 74.158676, 73.2766031, 72.382028, 71.4751811, 70.5562928, 69.6255934, 68.6833134, 67.7296831, 66.7649329, 65.7892933, 64.7780864, 63.7569909, 62.7267768, 61.6882147, 60.6420749, 59.5891277, 58.5301434, 57.4658924, 56.397145, 55.3246716, 54.2492425, 53.171628, 52.0925986, 51.0129244, 49.933376, 48.8547236, 47.8384263, 46.9444862, 46.0485076, 45.1473265, 44.2377791, 43.3167016, 42.38093, 41.4273006, 40.4526494, 39.4538126, 38.4276264, 37.3709268, 36.28055, 35.1533321, 33.9861094, 32.7757178, 31.5189935, 29.8360043, 27.9198952, 25.9683399, 23.9904015, 21.995143, 19.9916273, 17.9889174, 15.9960764, 14.0221672, 12.0762529, 10.1673964, 8.30466076, 6.49710894, 4.75380396, 3.08380881, 1.49618649, 0, 1.60107928, 2.82844165, 3.71175047, 4.28066904, 4.5648607, 4.59398878, 4.39771662, 4.00570753, 3.44762486, 2.75313192, 1.95189205, 1.07356858, 0.147824843, -0.795675839, -1.72727013, -2.61729471, -3.59284075, -4.64663842, -5.39446063, -5.84023429, -5.98788632, -5.84134366, -5.40453322, -4.68138192, -3.67581669, -2.39176445, -0.833152123, 0.996093371, 3.09204511, 5.10701715, 7.50688656, 10.3908541, 13.8581206, 18.0951902, 23.1576119, 29.077654, 35.9283263, 43.7826386, 52.7136006, 62.7942221, 74.0423083, 86.6028751, 100.559733, 115.996693, 132.997565, 149.464667, 168.208676, 189.858777, 215.044154, 244.393991, 280.027379, 319.262598, 360.155293, 402.16579, 447.622312, 497.565455, 553.035816, 616.083472, 686.510111, 763.016166, 845.259715, 933.78281, 1028.40271, 1129.42052, 1237.95721, 1354.75741, 1481.17088, 1618.09914, 1764.22491, 1919.486, 2081.51435, 2252.90837, 2434.40402, 2626.7373, 2830.64419, 3109.58723, 3385.89419, 3644.6194, 3870.81721, 4049.54195, 4097.82057, 4114.08638, 4112.54833, 4142.14872, 4203.34785, 4282.1338, 4379.69048, 4497.2018, 4635.85169, 4796.82405, 4272.19491, 4366.43597, 4841.10521, 5457.76061, 5977.96014, 6163.26179, 5775.22353, 5053.46362, 4047.93082},
        {96.8658079, 96.8078819, 96.7195602, 96.6013221, 96.4536472, 96.2770149, 96.0719047, 95.838796, 95.5781683, 95.2905011, 94.9762737, 94.6359658, 94.2700567, 93.8790258, 93.4633528, 93.0235169, 92.5599978, 92.0732748, 91.5638274, 91.032135, 90.4786772, 89.9039333, 89.308383, 88.6925055, 88.0567804, 87.4016871, 86.7277051, 86.0353138, 85.3249928, 84.5972215, 83.8524792, 83.0912456, 82.314, 81.521222, 80.7201646, 79.907778, 79.0809309, 78.2398831, 77.384894, 76.5162234, 75.6341309, 74.7388762, 73.8307188, 72.9099185, 71.9767349, 71.0314276, 70.0742562, 69.1054804, 68.1253599, 67.1341543, 66.1321232, 65.0940516, 64.046226, 62.9894584, 61.9245606, 60.8523445, 59.773622, 58.689205, 57.5999054, 56.506535, 55.4099057, 54.3108295, 53.2101181, 52.1085835, 51.0070376, 49.9062923, 48.8071594, 47.7710695, 46.8579978, 45.9437347, 45.0251624, 44.099163, 43.1626188, 42.212412, 41.2454248, 40.2585395, 39.2486383, 38.2126034, 37.1473169, 36.0496612, 34.9165185, 33.7447709, 32.5313007, 31.2729901, 29.5910678, 27.6781441, 25.7313445, 23.7597418, 21.772409, 19.778419, 17.7868446, 15.8067588, 13.8472344, 11.9173444, 10.0261617, 8.1827591, 6.39620957, 4.675586, 3.02996128, 1.46840831, 0, 1.73603829, 3.07870517, 4.05929383, 4.70909749, 5.05940937, 5.14152266, 4.98673058, 4.62632635, 4.09160316, 3.41385423, 2.62437277, 1.754452, 0.835385108, -0.101534681, -1.02501416, -1.90376012, -2.91403814, -4.05420627, -4.82349308, -5.23508224, -5.30215744, -5.03790233, -4.4555006, -3.56813592, -2.38899196, -0.931252395, 0.791899096, 2.76727884, 4.98170317, 6.99387283, 9.37569543, 12.242422, 15.7093035, 20.0694896, 25.3492819, 31.5275426, 38.666153, 46.8269945, 56.0719486, 66.4628967, 77.9007251, 90.6573088, 104.826028, 120.500264, 137.773395, 154.998252, 174.443904, 196.63887, 222.111668, 251.390818, 286.048291, 324.293819, 364.842529, 407.403109, 453.712767, 504.597949, 560.8851, 624.1676, 694.330656, 770.422813, 852.131163, 940.094128, 1034.37788, 1135.34267, 1243.84696, 1360.52055, 1486.51998, 1622.6116, 1767.63029, 1921.50351, 2081.91077, 2251.37324, 2430.59649, 2620.28607, 2821.14753, 3095.06844, 3366.27684, 3620.18278, 3842.19632, 4017.72751, 4065.70479, 4082.3409, 4081.53803, 4111.82984, 4173.41021, 4252.04322, 4348.64422, 4464.12858, 4599.41166, 4755.40881, 4259.80408, 4355.71086, 4811.69346, 5396.31619, 5878.14335, 6025.73924, 5607.66818, 4841.58002, 3860.68833},
        {98.4092025, 98.3240728, 98.2092734, 98.065247, 97.892436, 97.6912831, 97.4622309, 97.2057219, 96.9221988, 96.6121042, 96.2758805, 95.9139704, 95.5268166, 95.1148615, 94.6785478, 94.218318, 93.7346148, 93.2278807, 92.6985583, 92.1470903, 91.5739191, 90.9794874, 90.3642378, 89.7286128, 89.0730551, 88.3980072, 87.7039117, 86.9912112, 86.2603483, 85.5117656, 84.7459056, 83.963211, 83.1641244, 82.3490882, 81.5229729, 80.683914, 79.8299939, 78.9615115, 78.0787657, 77.1820554, 76.2716796, 75.347937, 74.4111267, 73.4615474, 72.4994982, 71.5252779, 70.5391855, 69.5415197, 68.5325795, 67.5126639, 66.4820717, 65.4158897, 64.3401755, 63.2557745, 62.1635321, 61.0642939, 59.9589052, 58.8482115, 57.7330583, 56.614291, 55.4927551, 54.3692959, 53.244759, 52.1199899, 50.9958338, 49.8731364, 48.752743, 47.6956976, 46.7615988, 45.8271372, 44.8892558, 43.9448976, 42.9910057, 42.0245229, 41.0423925, 40.0415574, 39.0189606, 37.9715452, 36.8962541, 35.7900305, 34.6498173, 33.4725575, 32.2551942, 30.9946704, 29.3146604, 27.4065929, 25.4663827, 23.5030862, 21.5257597, 19.5434596, 17.5652421, 15.6001635, 13.6572802, 11.7456484, 9.87432451, 8.05236481, 6.28882561, 4.59276321, 2.97323394, 1.4392941, 0, 1.82138133, 3.23924754, 4.28582086, 4.99332352, 5.39397776, 5.5200058, 5.40362987, 5.07707221, 4.57255506, 3.92230063, 3.15853117, 2.31346889, 1.41933605, 0.508354856, -0.387252448, -1.23526363, -2.24107338, -3.40640475, -4.16156865, -4.52526077, -4.51617679, -4.1530124, -3.45446326, -2.43922506, -1.12599348, 0.466535802, 2.31966711, 4.41470476, 6.73295307, 8.77423693, 11.1778826, 14.0696381, 17.5752517, 22.0559472, 27.5197349, 33.9118315, 41.2873427, 49.7013739, 59.2090305, 69.865418, 81.4702282, 94.4117148, 108.794956, 124.725028, 142.30701, 160.456064, 180.764662, 203.63536, 229.470716, 258.673285, 292.145879, 329.179378, 369.279201, 412.451775, 459.726423, 511.673206, 568.862188, 632.312061, 702.042297, 777.582953, 858.746964, 946.188649, 1040.18695, 1141.13825, 1249.63797, 1366.19017, 1491.71617, 1626.82822, 1770.60859, 1923.00498, 2081.91852, 2249.65837, 2426.88065, 2614.24151, 2812.39708, 3081.66357, 3348.12203, 3597.5136, 3815.57941, 3988.06056, 4035.65115, 4052.50404, 4052.23767, 4083.1204, 4145.05507, 4223.50705, 4319.10107, 4432.46184, 4564.21409, 4714.98252, 4296.74371, 4405.07151, 4828.61372, 5356.01817, 5775.93267, 5877.00501, 5447.88302, 4642.27445, 3693.96725},
        {100.067663, 99.9529698, 99.8093687, 99.6372643, 99.437061, 99.209163, 98.9539747, 98.6719002, 98.3633438, 98.0287099, 97.6684027, 97.2828265, 96.8723856, 96.4374842, 95.9785267, 95.4959172, 94.9900601, 94.4613596, 93.9102201, 93.3370457, 92.7422409, 92.1262098, 91.4893567, 90.8320859, 90.1548018, 89.4579084, 88.7418103, 88.0069115, 87.2536164, 86.4823293, 85.6934544, 84.8873961, 84.0645585, 83.225346, 82.3720534, 81.5041, 80.6208823, 79.7227431, 78.8100254, 77.8830721, 76.9422262, 75.9878305, 75.020228, 74.0397616, 73.0467743, 72.0416089, 71.0246084, 69.9961157, 68.9564738, 67.9060255, 66.8451137, 65.74962, 64.644879, 63.5317638, 62.4111478, 61.2839041, 60.1509059, 59.0130264, 57.8711388, 56.7261163, 55.5788321, 54.4301595, 53.2809715, 52.1321415, 50.9845426, 49.839048, 48.6965309, 47.6173649, 46.6604947, 45.7040788, 44.7451332, 43.7806739, 42.807717, 41.8232785, 40.8243744, 39.8080207, 38.7712336, 37.7110291, 36.6244231, 35.5084318, 34.3600711, 33.1763572, 31.954306, 30.6909336, 29.0135817, 27.1118793, 25.1798981, 23.2266509, 21.2611504, 19.2924092, 17.32944, 15.3812556, 13.4568686, 11.5652916, 9.71553741, 7.91661863, 6.17754796, 4.50733808, 2.91500167, 1.40955142, 0, 1.84543728, 3.28864388, 4.36195421, 5.09770269, 5.52822372, 5.68585172, 5.60292111, 5.31176629, 4.84472168, 4.23412169, 3.51230074, 2.71159324, 1.86433359, 1.00285623, 0.159495548, -0.63341403, -1.58842924, -2.70820011, -3.40647179, -3.70338654, -3.61908662, -3.17371429, -2.38741182, -1.28032145, 0.127414539, 1.8156539, 3.76425437, 5.95307368, 8.36196959, 10.4695841, 12.9407698, 15.9057491, 19.4947444, 24.0926504, 29.7023535, 36.2588274, 43.815893, 52.427371, 62.1470825, 73.028848, 84.8133034, 97.9627718, 112.592349, 128.817132, 146.752216, 165.884058, 187.113554, 210.71296, 236.954532, 266.110526, 298.401391, 334.21451, 373.84644, 417.602837, 465.807938, 518.777638, 576.827833, 640.392149, 709.641992, 784.620334, 865.322443, 952.347621, 1046.12214, 1147.05137, 1255.50525, 1371.87, 1496.82923, 1630.84628, 1773.2939, 1924.16172, 2081.72937, 2247.9596, 2423.43394, 2608.73399, 2804.44129, 3069.11539, 3330.8654, 3575.77836, 3789.94137, 3959.44148, 4006.82822, 4024.08457, 4024.528, 4056.07354, 4118.34376, 4196.54557, 4291.02231, 4402.11733, 4530.17398, 4675.53561, 4369.81771, 4499.00202, 4880.88529, 5333.26426, 5673.93567, 5720.69625, 5291.34275, 4455.43783, 3547.50139},
        {101.85316, 101.707216, 101.533038, 101.330997, 101.101464, 100.844807, 100.561398, 100.251607, 99.9158041, 99.5543595, 99.1676435, 98.7560263, 98.3198784, 97.85957, 97.3754712, 96.8679525, 96.3373841, 95.7841362, 95.2085791, 94.6110831, 93.9920185, 93.3517555, 92.6906645, 92.0091157, 91.3074793, 90.5861256, 89.845425, 89.0857477, 88.3074639, 87.5109439, 86.6965581, 85.8646766, 85.0156698, 84.1499079, 83.2672619, 82.3683622, 81.4538453, 80.5240976, 79.5795056, 78.6204557, 77.6473346, 76.6605286, 75.6604242, 74.6474079, 73.6218661, 72.5841854, 71.5347523, 70.4739531, 69.4021745, 68.3198028, 67.2272245, 66.1012618, 64.9663764, 63.8234654, 62.6734263, 61.5171564, 60.3555529, 59.1895131, 58.0199344, 56.8477141, 55.6737494, 54.4989377, 53.3241764, 52.1503626, 50.9783937, 49.809167, 48.6435798, 47.5411256, 46.5598911, 45.5799233, 44.5983209, 43.6121828, 42.6186075, 41.614694, 40.597541, 39.5642472, 38.5119113, 37.4376322, 36.3385085, 35.2116391, 34.0541226, 32.8630579, 31.6355436, 30.3686787, 28.6946311, 26.8006409, 24.878334, 22.9366519, 20.984536, 19.0309278, 17.0847687, 15.1550003, 13.2505638, 11.3804009, 9.55345289, 7.77866129, 6.06496754, 4.42131308, 2.85663937, 1.37988786, 0, 1.79653504, 3.2054693, 4.25831655, 4.98659056, 5.42180511, 5.59547397, 5.53911091, 5.28422971, 4.86234415, 4.30496799, 3.64361501, 2.90979898, 2.13503369, 1.35083289, 0.588710374, -0.119820095, -0.970588458, -1.96455861, -2.55598695, -2.76207621, -2.60002913, -2.08704844, -1.24033688, -0.0770971827, 1.38546791, 3.13015565, 5.13976332, 7.39708817, 9.88492747, 12.101389, 14.6916789, 17.7840015, 21.5065607, 26.2176867, 31.9305206, 38.5968367, 46.2758012, 55.0265804, 64.9083408, 75.9802486, 87.9924367, 101.407159, 116.344044, 132.922721, 151.26282, 171.328189, 193.433184, 217.73638, 244.396351, 273.571673, 304.896074, 339.694453, 378.925381, 423.14734, 472.10197, 525.897657, 584.642789, 648.283159, 717.126698, 791.658707, 872.072922, 958.852292, 1052.47572, 1153.32613, 1261.62383, 1377.66377, 1501.92894, 1634.76304, 1775.82033, 1925.14504, 2081.53509, 2246.4728, 2420.43384, 2603.89388, 2797.32859, 3057.16669, 3313.94254, 3554.14354, 3764.2571, 3930.77064, 3978.4046, 3996.5913, 3998.28972, 4030.74236, 4093.33762, 4171.17903, 4264.36921, 4373.01078, 4497.20633, 4637.0585, 4465.82996, 4621.9865, 4957.52744, 5324.45212, 5574.7599, 5560.45009, 5133.52205, 4280.96113, 3421.02457},
        {103.777659, 103.599453, 103.393474, 103.160069, 102.899585, 102.612368, 102.298764, 101.959119, 101.593781, 101.203094, 100.787406, 100.347062, 99.8824096, 99.3937946, 98.8815633, 98.3460621, 97.7876372, 97.206635, 96.6034019, 95.9782841, 95.3316279, 94.6637798, 93.975086, 93.2658928, 92.5365465, 91.7873935, 91.0187802, 90.2310527, 89.4245576, 88.5996409, 87.7566492, 86.8959287, 86.0178257, 85.1226866, 84.208454, 83.2767266, 82.329132, 81.3660946, 80.3880387, 79.3953886, 78.3885686, 77.3680031, 76.3341165, 75.2873329, 74.2280768, 73.1567725, 72.0738443, 70.9797165, 69.8748135, 68.7595596, 67.6343792, 66.4768345, 65.3107075, 64.1369181, 62.9563862, 61.7700318, 60.5787748, 59.3835351, 58.1852327, 56.9847875, 55.7831194, 54.5811483, 53.3797942, 52.179977, 50.9826166, 49.788633, 48.5989461, 47.4720341, 46.4649935, 45.4600343, 44.4543453, 43.445115, 42.4295321, 41.4047852, 40.3680631, 39.3165544, 38.2474477, 37.1579317, 36.045195, 34.9064263, 33.7388143, 32.5395476, 31.3058148, 30.0348047, 28.364608, 26.4795155, 24.5681339, 22.6393053, 20.7018718, 18.7646754, 16.8365583, 14.9263625, 13.04293, 11.195103, 9.39172348, 7.64163354, 5.95367525, 4.33669068, 2.79952191, 1.35101099, 0, 1.66300351, 2.96829891, 3.94553053, 4.62434271, 5.03437978, 5.20528607, 5.16670591, 4.94828364, 4.57966359, 4.0904901, 3.51040749, 2.8690601, 2.19609226, 1.52114831, 0.873872574, 0.283909388, -0.40203381, -1.18044653, -1.6078986, -1.69394646, -1.44814653, -0.88005525, 0.000770956926, 1.18477566, 2.66240242, 4.42409482, 6.46029642, 8.76145079, 11.3180015, 13.6911262, 16.4579318, 19.7376419, 23.6494796, 28.4691434, 34.2376191, 40.954166, 48.6910648, 57.5205965, 67.5150418, 78.7466816, 91.070114, 104.841554, 120.175872, 137.18794, 155.992629, 176.834411, 199.666155, 224.570332, 251.62941, 280.925861, 311.711175, 345.914443, 384.897157, 429.376331, 478.753177, 533.019676, 592.167809, 655.860387, 724.493372, 798.82182, 879.213725, 965.983909, 1059.53993, 1160.20662, 1268.16871, 1383.67523, 1507.08507, 1638.67577, 1778.32199, 1926.12626, 2081.52743, 2245.39384, 2418.05779, 2599.85156, 2791.10743, 3045.56024, 3296.78908, 3531.77563, 3737.50152, 3900.94841, 3949.54887, 3969.53303, 3973.40356, 4007.18001, 4070.09797, 4147.42769, 4239.10305, 4345.05791, 4465.22614, 4599.5416, 4571.58438, 4758.50904, 5047.55945, 5325.97946, 5481.01292, 5399.90368, 4969.8956, 4118.73526, 3314.2706},
        {105.942093, 105.731493, 105.492907, 105.226688, 104.933194, 104.61278, 104.265802, 103.892616, 103.493578, 103.069044, 102.619369, 102.14491, 101.646023, 101.123063, 100.576386, 100.006349, 99.4133062, 98.7976148, 98.1596304, 97.4997089, 96.8182062, 96.1154781, 95.3918806, 94.6477696, 93.883501, 93.0994307, 92.2959146, 91.4733086, 90.6319687, 89.7722507, 88.8945106, 87.9991042, 87.0863874, 86.1567163, 85.2080004, 84.2418702, 83.2599857, 82.262782, 81.2506945, 80.2241585, 79.1836093, 78.1294821, 77.0622122, 75.982235, 74.8899858, 73.7858997, 72.6704122, 71.5439585, 70.4069738, 69.2598936, 68.103153, 66.9138103, 65.7161845, 64.5112179, 63.2998524, 62.0830301, 60.8616931, 59.6367835, 58.4092433, 57.1800146, 55.9500395, 54.7202599, 53.491618, 52.2650559, 51.0415156, 49.8219391, 48.6072685, 47.4549361, 46.4210178, 45.3899489, 44.3590095, 43.3254796, 42.2866391, 41.2397683, 40.182147, 39.1110554, 38.0237734, 36.9175811, 35.7897586, 34.6375858, 33.4583429, 32.2493098, 31.0077666, 29.7309933, 28.0646135, 26.1891321, 24.2892805, 22.3737509, 20.451235, 18.530425, 16.6200128, 14.7286903, 12.8651495, 11.0380823, 9.2561808, 7.52813688, 5.86264252, 4.26838969, 2.75407035, 1.32837646, 0, 1.24890629, 2.22045636, 2.93900544, 3.42890872, 3.71452142, 3.82019874, 3.77029589, 3.58916808, 3.30117052, 2.93065842, 2.50198697, 2.0395114, 1.56758691, 1.1105687, 0.692811994, 0.338671988, 0.01225811, -0.270680973, -0.349519828, -0.212813056, 0.150884738, 0.753018953, 1.60503498, 2.71837823, 4.10449409, 5.77482796, 7.74082523, 10.0139313, 12.6055916, 15.2580942, 18.3507328, 21.9757, 26.2251883, 31.1385842, 36.8344832, 43.4455628, 51.0798573, 59.8454011, 69.8502286, 81.2023743, 94.1757961, 108.662106, 124.736876, 142.475677, 161.95408, 183.008816, 206.014006, 231.104932, 258.416874, 288.085113, 319.66072, 354.577222, 393.738541, 437.711561, 486.375053, 539.916737, 598.524333, 662.311653, 731.557121, 806.619792, 887.871042, 975.358441, 1069.03149, 1169.05907, 1275.9818, 1390.16964, 1512.15794, 1642.35955, 1780.58098, 1926.97173, 2081.93009, 2245.31847, 2417.20022, 2597.6387, 2786.69727, 3033.17165, 3276.20976, 3503.69185, 3703.49819, 3863.50907, 3916.24583, 3941.8536, 3951.93789, 3988.8124, 4051.92731, 4127.93776, 4216.93369, 4319.00505, 4434.2418, 4562.73388, 4673.63302, 4897.63786, 5150.637, 5348.51905, 5407.17259, 5242.48622, 4770.34853, 3962.03329, 3230.76808},
        {108.201916, 107.959102, 107.687881, 107.388628, 107.061716, 106.707518, 106.326409, 105.918762, 105.48495, 105.025347, 104.540326, 104.030262, 103.495527, 102.936495, 102.35354, 101.747036, 101.117355, 100.464872, 99.7899596, 99.092992, 98.3743426, 97.6343848, 96.8734924, 96.0920388, 95.2903977, 94.4689425, 93.6280469, 92.7680843, 91.8894285, 90.9924529, 90.0775311, 89.1450367, 88.1953432, 87.2288242, 86.2437386, 85.2415618, 84.2237945, 83.190879, 82.1432575, 81.0813721, 80.0056651, 78.9165785, 77.8145547, 76.7000359, 75.5734641, 74.4352816, 73.2859306, 72.1258533, 70.9554919, 69.7752885, 68.5856854, 67.3628723, 66.1320696, 64.8942452, 63.650367, 62.4014031, 61.1483213, 59.8920895, 58.6336757, 57.3740479, 56.1141739, 54.8550217, 53.5975593, 52.3427545, 51.0915753, 49.8449897, 48.6039656, 47.4248796, 46.3627783, 45.304353, 44.2469797, 43.1880345, 42.1248935, 41.0549325, 39.9755276, 38.8840548, 37.7778901, 36.6544096, 35.5109892, 34.345005, 33.1538329, 31.934849, 30.6854292, 29.4029497, 27.7410992, 25.8763264, 23.9892106, 22.0882812, 20.1820676, 18.2790989, 16.3879046, 14.5170139, 12.6749562, 10.8702608, 9.11145705, 7.4070742, 5.76564159, 4.19568855, 2.70574439, 1.30433844, 0, 0.86675525, 1.53033385, 2.01023333, 2.32595122, 2.49698504, 2.54283231, 2.48299058, 2.33695736, 2.12423019, 1.86430659, 1.57668408, 1.2808602, 0.99633247, 0.742598422, 0.539155581, 0.405501475, 0.417012835, 0.601169932, 0.851076856, 1.19820525, 1.67402676, 2.31001302, 3.13763568, 4.18836638, 5.49367676, 7.08503847, 8.99392314, 11.2518024, 13.890148, 16.8275962, 20.2498269, 24.2221622, 28.8099244, 33.8181486, 39.4423006, 45.9472558, 53.4764225, 62.1732091, 72.1810238, 83.6432749, 97.2541379, 112.438629, 129.232398, 147.671094, 167.790366, 188.911894, 211.963789, 237.160195, 264.715252, 294.843104, 327.536758, 363.501763, 403.023414, 446.454799, 494.287421, 546.950639, 604.87381, 668.721745, 738.634837, 814.496624, 896.657682, 984.908922, 1078.73654, 1178.1534, 1284.06484, 1396.96653, 1517.58583, 1646.47847, 1783.35069, 1928.38927, 2082.87864, 2245.73902, 2416.7772, 2595.79998, 2782.61415, 3021.26932, 3256.26877, 3476.35859, 3670.28488, 3826.79374, 3883.26542, 3914.05573, 3929.92467, 3969.7588, 4033.1847, 4108.10959, 4194.74606, 4293.30666, 4404.00397, 4527.05056, 4759.22283, 5015.79212, 5237.06649, 5363.354, 5334.96271, 5092.20067, 4575.37596, 3822.38919, 3163.57458},
        {110.497277, 110.22292, 109.919687, 109.587968, 109.228154, 108.840635, 108.425804, 107.984049, 107.515762, 107.021334, 106.501155, 105.955615, 105.385106, 104.790019, 104.170743, 103.52767, 102.86119, 102.171694, 101.459573, 100.725217, 99.9690167, 99.1913633, 98.3926473, 97.5732592, 96.7335899, 95.87403, 94.9949701, 94.0968009, 93.1799132, 92.2446976, 91.2915448, 90.3208454, 89.3329902, 88.3283698, 87.3054671, 86.2656668, 85.2103765, 84.1400489, 83.0551365, 81.956092, 80.8433679, 79.7174169, 78.5786915, 77.4276443, 76.264728, 75.090395, 73.9050981, 72.7092898, 71.5034228, 70.2879495, 69.0633226, 67.8047055, 66.5383883, 65.2653717, 63.9866565, 62.7032436, 61.4161337, 60.1263277, 58.8348264, 57.5426306, 56.2507411, 54.9601587, 53.6718843, 52.3869186, 51.1062626, 49.8309169, 48.5618824, 47.3545238, 46.2629005, 45.1758919, 44.0909747, 43.0056254, 41.9173208, 40.8235374, 39.7217518, 38.6094407, 37.4840808, 36.3431485, 35.1841206, 34.0044737, 32.8016844, 31.5732294, 30.3165852, 29.0292285, 27.3728942, 25.5199918, 23.6469658, 21.7621883, 19.8740316, 17.9908675, 16.1210683, 14.2730059, 12.4550526, 10.6755803, 8.94296119, 7.26556734, 5.65177084, 4.10994377, 2.64845821, 1.27568626, 0, 0.65939231, 1.15760122, 1.51147613, 1.73786646, 1.85362159, 1.87559094, 1.82062391, 1.7055699, 1.54727832, 1.36259857, 1.16838006, 0.981472181, 0.818724348, 0.696985961, 0.633106425, 0.643935142, 0.872310209, 1.35307941, 1.81811774, 2.31127216, 2.87638967, 3.55731724, 4.39790186, 5.44199049, 6.73343014, 8.31606776, 10.2337504, 12.5303249, 15.2496384, 18.4249918, 22.114645, 26.3656086, 31.2248935, 36.3354352, 41.9462821, 48.4142358, 55.9075297, 64.5943972, 74.6430715, 86.2217859, 100.31395, 116.024523, 133.362249, 152.335869, 172.954126, 194.125457, 217.233986, 242.563532, 270.397913, 301.020949, 334.824768, 371.852501, 411.857186, 454.999483, 502.272283, 554.294822, 611.686336, 675.529882, 745.875463, 822.267098, 905.058352, 993.84752, 1087.81008, 1186.84493, 1292.07674, 1404.06754, 1523.64415, 1651.43724, 1787.10656, 1930.86108, 2084.59704, 2246.57653, 2416.42443, 2593.76565, 2778.22508, 3010.35025, 3238.61275, 3452.40683, 3641.12672, 3794.16665, 3852.42773, 3886.12607, 3905.43868, 3947.28227, 4011.24938, 4085.79099, 4171.05671, 4267.19616, 4374.35893, 4492.69465, 4814.39788, 5092.49869, 5286.1988, 5354.69993, 5257.20377, 4952.91206, 4401.0265, 3705.2248, 3109.46146},
        {112.72494, 112.419836, 112.085681, 111.722864, 111.331775, 110.912802, 110.466333, 109.992759, 109.492467, 108.965846, 108.413287, 107.835176, 107.231904, 106.603859, 105.951429, 105.275005, 104.574974, 103.851726, 103.105649, 102.337133, 101.546565, 100.734336, 99.9008337, 99.0464472, 98.1715654, 97.2765772, 96.3618713, 95.4278368, 94.4748625, 93.5033372, 92.5136498, 91.5061892, 90.4813443, 89.439504, 88.378256, 87.2994487, 86.2049641, 85.0952819, 83.970882, 82.832244, 81.6798478, 80.5141731, 79.3356997, 78.1449074, 76.9422759, 75.7282851, 74.5034146, 73.2681443, 72.022954, 70.7683234, 69.5047322, 68.2063951, 66.9006264, 65.5884753, 64.2709909, 62.9492224, 61.6242188, 60.2970295, 58.9687034, 57.6402898, 56.3128379, 54.9873967, 53.6650154, 52.3467432, 51.0336293, 49.7267227, 48.4270726, 47.1891987, 46.0663366, 44.9492568, 43.8355422, 42.7227756, 41.6085399, 40.4904181, 39.365993, 38.2328474, 37.0885643, 35.9307266, 34.7569171, 33.5647186, 32.3517142, 31.1154866, 29.8536188, 28.5636936, 26.9139539, 25.0735798, 23.2157099, 21.3486147, 19.4805648, 17.6198309, 15.7746835, 13.9533932, 12.1642306, 10.4154664, 8.71537106, 7.07221526, 5.49426958, 3.98980461, 2.56709096, 1.23439923, 0, 1.02843507, 1.83199263, 2.43212471, 2.85028334, 3.10792055, 3.22648836, 3.22743882, 3.13222394, 2.96229577, 2.73910632, 2.48410763, 2.21875173, 1.96449064, 1.74277641, 1.57506105, 1.4827966, 1.52652879, 1.73329444, 2.03611939, 2.46483289, 3.04926417, 3.8192425, 4.80459712, 6.03515728, 7.54075224, 9.35121124, 11.4963635, 14.0060384, 16.910065, 20.1392483, 23.8587509, 28.1281092, 33.0068598, 38.2226157, 44.0008751, 50.6556495, 58.3460523, 67.2311974, 77.4701982, 89.2221685, 103.343023, 119.083004, 136.464895, 155.511479, 176.245538, 197.6079, 220.973792, 246.636487, 274.889257, 306.025374, 340.530473, 378.270353, 418.934866, 462.636246, 510.349086, 562.6608, 620.158798, 683.771895, 753.668509, 829.644619, 912.253705, 1000.98537, 1095.08924, 1194.446, 1300.00568, 1412.22746, 1531.78363, 1659.18865, 1794.17544, 1936.91899, 2088.99826, 2248.96828, 2416.51804, 2591.33654, 2773.11279, 3002.92609, 3228.72755, 3439.85862, 3625.66072, 3775.47528, 3829.48852, 3858.9672, 3874.31929, 3915.03098, 3980.00007, 4056.00841, 4142.5596, 4239.15723, 4345.30489, 4460.50618, 4797.01656, 5068.12454, 5240.421, 5280.4968, 5154.9428, 4830.34989, 4273.30892, 3624.71422, 3066.32564},
        {114.960181, 114.626463, 114.263392, 113.871356, 113.450743, 113.001944, 112.525345, 112.021336, 111.490305, 110.932641, 110.348732, 109.738968, 109.103737, 108.443427, 107.758427, 107.049126, 106.315912, 105.559174, 104.779301, 103.976681, 103.151703, 102.304755, 101.436227, 100.546506, 99.6359815, 98.705042, 97.7540761, 96.7834725, 95.7936197, 94.7849065, 93.7577214, 92.712453, 91.64949, 90.5692211, 89.4682615, 88.348966, 87.2137345, 86.063078, 84.8975076, 83.7175345, 82.5236696, 81.3164242, 80.0963091, 78.8638356, 77.6195146, 76.3638573, 75.0973748, 73.8205781, 72.5339783, 71.2380865, 69.9334137, 68.5930489, 67.2455197, 65.8919318, 64.5333905, 63.1710016, 61.8058704, 60.4391026, 59.0718037, 57.7050792, 56.3400346, 54.9777755, 53.6194074, 52.2660358, 50.9187663, 49.5787045, 48.2469558, 46.9772959, 45.8222357, 44.6742528, 43.5310383, 42.3902834, 41.2496791, 40.1069167, 38.9596872, 37.8056819, 36.6425919, 35.4681082, 34.2799222, 33.0757249, 31.8532074, 30.610061, 29.3439768, 28.0526458, 26.4109285, 24.5847216, 22.7438023, 20.8963383, 19.050497, 17.2144462, 15.3963535, 13.6043865, 11.8467128, 10.1315002, 8.46691612, 6.86112831, 5.32230438, 3.85861195, 2.47821866, 1.18929213, 0, 1.51091858, 2.71242468, 3.63200244, 4.29713603, 4.73530961, 4.97400733, 5.04071336, 4.96291185, 4.76808696, 4.48372286, 4.13730369, 3.75631362, 3.3682368, 3.00055741, 2.68075958, 2.43632749, 2.2170697, 2.03937383, 2.10641888, 2.42904421, 3.01808923, 3.88439331, 5.03879585, 6.49213623, 8.25525384, 10.3389881, 12.7541783, 15.5116639, 18.6222843, 21.8717426, 25.5785644, 29.82113, 34.6778196, 40.0238215, 46.0391119, 52.9578517, 60.9193789, 70.0630313, 80.5281468, 92.4540634, 106.42944, 122.007543, 139.239801, 158.17764, 178.872489, 200.46768, 224.149758, 250.197173, 278.888378, 310.501822, 345.48861, 383.743518, 425.076152, 469.610557, 518.112748, 571.060508, 628.931622, 692.338462, 761.593634, 836.862919, 918.841671, 1007.08, 1101.1659, 1201.04354, 1307.26071, 1420.08811, 1539.89091, 1666.9643, 1801.25718, 1942.91433, 2093.06596, 2250.71983, 2415.67962, 2587.749, 2766.73167, 2995.10553, 3219.33147, 3428.5446, 3611.88005, 3758.47291, 3806.83385, 3830.47211, 3840.21899, 3879.19247, 3945.44374, 4023.57121, 4112.29871, 4210.35005, 4316.44906, 4429.31956, 4747.35868, 5000.40161, 5156.55085, 5183.90892, 5050.57833, 4724.66158, 4174.2612, 3565.56686, 3030.96067},
        {117.215076, 116.855786, 116.46656, 116.047795, 115.599887, 115.123232, 114.618229, 114.085272, 113.524759, 112.937086, 112.322651, 111.681849, 111.015078, 110.322734, 109.605214, 108.862915, 108.096232, 107.305564, 106.491306, 105.653855, 104.793608, 103.910961, 103.006311, 102.080056, 101.132591, 100.164313, 99.1756187, 98.1669051, 97.1385687, 96.0910062, 95.0246142, 93.9397894, 92.8369284, 91.716428, 90.5741745, 89.4129147, 88.2354494, 87.0423216, 85.8340743, 84.6112506, 83.3743935, 82.1240461, 80.8607513, 79.5850523, 78.297492, 76.9986135, 75.6889598, 74.3690739, 73.0394989, 71.7007779, 70.3534537, 68.9692323, 67.578119, 66.1812821, 64.7798896, 63.3751098, 61.9681108, 60.5600608, 59.152128, 57.7454805, 56.3412866, 54.9407143, 53.5449319, 52.1551075, 50.7724092, 49.3980054, 48.033064, 46.7307193, 45.5428574, 44.3634709, 43.1903593, 42.0213219, 40.8541582, 39.6866676, 38.5166496, 37.3419036, 36.1602291, 34.9694254, 33.767292, 32.5516283, 31.3202338, 30.070908, 28.8014501, 27.5096598, 25.8776507, 24.0677052, 22.2458998, 20.4202785, 18.5988855, 16.7897647, 15.0009602, 13.240516, 11.5164762, 9.83688478, 8.20978584, 6.64322339, 5.14524148, 3.72388414, 2.38719542, 1.14321936, 0, 1.94997226, 3.51394175, 4.72487874, 5.61575347, 6.21953619, 6.56919716, 6.69770662, 6.63803483, 6.42315203, 6.08602849, 5.65963444, 5.17694014, 4.67091584, 4.17453179, 3.72075825, 3.34256546, 2.89143164, 2.37439959, 2.23495729, 2.46718383, 3.06515831, 4.02295983, 5.3346675, 6.9943604, 8.99611764, 11.3340183, 14.0021415, 16.9945664, 20.305372, 23.5902654, 27.3057674, 31.5474689, 36.4109605, 41.9249128, 48.2183766, 55.438248, 63.7001938, 73.1198806, 83.8129753, 95.8951446, 109.625775, 124.933073, 141.904586, 160.627861, 181.190446, 203.052051, 227.085021, 253.533862, 282.64308, 314.657183, 349.88638, 388.429171, 430.348523, 475.829837, 525.308482, 579.107597, 637.550319, 700.879728, 769.517216, 843.971515, 924.918409, 1012.19986, 1106.0107, 1206.45024, 1313.4559, 1427.03942, 1547.11443, 1673.66725, 1807.04379, 1947.38657, 2095.52897, 2250.81631, 2413.15195, 2582.43924, 2758.58151, 2985.26417, 3207.663, 3414.73586, 3595.44058, 3738.735, 3781.58318, 3799.59926, 3804.02414, 3841.5149, 3909.29353, 3989.82524, 4081.06986, 4180.98719, 4287.53705, 4398.67926, 4676.52381, 4905.51027, 5050.89341, 5077.92803, 4951.86891, 4637.97084, 4101.4886, 3522.92373, 3000.55072},
        {119.53678, 119.161667, 118.754891, 118.3169, 117.848141, 117.349061, 116.820109, 116.261732, 115.674377, 115.058492, 114.414524, 113.742922, 113.044132, 112.318601, 111.566779, 110.789112, 109.986047, 109.158033, 108.305516, 107.428945, 106.528767, 105.605429, 104.659379, 103.691065, 102.700933, 101.689432, 100.65701, 99.6041126, 98.5311889, 97.4386859, 96.3270511, 95.1967323, 94.0481768, 92.8818323, 91.6946833, 90.4889817, 89.267021, 88.0293611, 86.7765619, 85.5091833, 84.2277852, 82.9329277, 81.6251705, 80.3050737, 78.9731971, 77.6301007, 76.2763443, 74.912488, 73.5390916, 72.1567151, 70.7659183, 69.3356174, 67.8987021, 66.4564183, 65.0100119, 63.5607287, 62.1098148, 60.658516, 59.2080781, 57.7597472, 56.314769, 54.8743895, 53.4398546, 52.0124102, 50.5933022, 49.1837764, 47.7850789, 46.4503274, 45.2313947, 44.0225443, 42.8216594, 41.6266234, 40.4353195, 39.2456309, 38.0554408, 36.8626325, 35.6650893, 34.4606944, 33.247331, 32.0228823, 30.7852317, 29.5322624, 28.2618576, 26.9719005, 25.3531692, 23.563559, 21.7646566, 19.9643113, 18.1703726, 16.3906898, 14.6331124, 12.9054897, 11.2156712, 9.57150619, 7.98084413, 6.45153442, 4.99142645, 3.60836963, 2.31021334, 1.104807, 0, 1.81844633, 3.28004121, 4.41591395, 5.25719383, 5.83501016, 6.18049223, 6.32476934, 6.29897079, 6.13422587, 5.86166388, 5.51241412, 5.11760589, 4.70836847, 4.31583118, 3.97112331, 3.70537414, 3.43382756, 3.17105778, 3.21031593, 3.55789871, 4.22010282, 5.20322497, 6.51356187, 8.1574102, 10.1410667, 12.470828, 15.1529909, 18.1938521, 21.5997082, 25.126656, 29.1229322, 33.6698934, 38.8488964, 44.6521114, 51.2054883, 58.65876, 67.1200393, 76.6974388, 87.499071, 99.6330486, 113.173928, 128.273591, 145.046715, 163.607979, 184.07206, 206.227988, 230.597501, 257.37669, 286.761645, 318.948456, 353.991258, 392.401597, 434.605484, 480.908675, 531.37142, 586.164194, 645.457475, 709.349038, 778.098585, 852.045125, 931.132311, 1016.26593, 1108.96301, 1209.73981, 1317.41683, 1431.59296, 1551.35034, 1676.15391, 1807.36326, 1945.26624, 2091.48519, 2244.76821, 2404.94121, 2571.83015, 2745.26095, 2967.41251, 3185.16955, 3387.76978, 3564.45087, 3704.45054, 3745.90226, 3763.2174, 3767.53545, 3806.10805, 3875.78029, 3958.35057, 4051.41633, 4152.57501, 4259.42405, 4369.5609, 4607.51193, 4818.35988, 4961.82349, 4997.62154, 4885.47279, 4585.096, 4056.20993, 3481.1609, 2962.64006},
        {121.8803, 121.490585, 121.067347, 120.611085, 120.122305, 119.601508, 119.049197, 118.465874, 117.852043, 117.208206, 116.534866, 115.832526, 115.101687, 114.342854, 113.556529, 112.743213, 111.903411, 111.037625, 110.146358, 109.230111, 108.289389, 107.324693, 106.336527, 105.325392, 104.291793, 103.236231, 102.159209, 101.06123, 99.9427973, 98.8044126, 97.6465789, 96.4697991, 95.2745757, 94.0614116, 92.8284832, 91.5775054, 90.3102207, 89.0272074, 87.7290437, 86.4163077, 85.0895776, 83.7494318, 82.3964483, 81.0312053, 79.6542811, 78.2662539, 76.8677018, 75.4592031, 74.0413359, 72.6146785, 71.179809, 69.7031043, 68.2200854, 66.7320718, 65.2403833, 63.7463394, 62.2512597, 60.7564639, 59.2632715, 57.7730023, 56.2869758, 54.8065116, 53.3329293, 51.8675487, 50.4116892, 48.9666705, 47.5338122, 46.1659939, 44.9148571, 43.675383, 42.4455489, 41.2233319, 40.0067091, 38.7936579, 37.5821555, 36.3701789, 35.1557055, 33.9367124, 32.7111768, 31.477076, 30.232387, 28.9750872, 27.7031537, 26.4145637, 24.8102933, 23.0430673, 21.2691801, 19.4962473, 17.7318843, 15.9837065, 14.2593295, 12.5663687, 10.9124395, 9.30515738, 7.75213784, 6.26099631, 4.83934826, 3.49480915, 2.23499443, 1.06751956, 0, 1.56187369, 2.81830062, 3.79708452, 4.52602911, 5.03293812, 5.34561529, 5.49186433, 5.49948899, 5.39629297, 5.21008003, 4.96865388, 4.69981825, 4.43137688, 4.19113348, 4.0068918, 3.90645555, 3.8947151, 3.99620085, 4.29054437, 4.80063937, 5.54937959, 6.55965874, 7.85437055, 9.45640874, 11.388667, 13.6740391, 16.3354188, 19.3956997, 22.8777756, 26.6698834, 30.9789478, 35.8682596, 41.4011096, 47.5121683, 54.3290367, 62.0136142, 70.6677778, 80.3934041, 91.2923701, 103.466553, 116.818466, 131.710025, 148.282112, 166.67561, 187.0314, 209.49722, 234.205385, 261.295063, 290.905423, 323.175634, 357.85984, 395.952823, 438.262626, 485.230901, 536.551253, 592.253145, 652.36604, 716.903255, 785.914069, 859.465377, 936.681121, 1019.58653, 1111.02048, 1211.92717, 1320.04039, 1434.56755, 1553.78448, 1676.65704, 1805.56707, 1940.96411, 2085.57054, 2237.21347, 2395.55575, 2560.26022, 2730.98972, 2947.0276, 3158.51106, 3355.19779, 3526.84551, 3663.21192, 3704.9696, 3723.68481, 3730.15174, 3770.87794, 3842.52107, 3926.85822, 4021.29107, 4123.22129, 4230.05058, 4339.18061, 4532.06194, 4724.66082, 4869.74106, 4920.06646, 4828.40082, 4547.50795, 4030.15163, 3452.93274, 2930.09592},
        {124.234194, 123.828122, 123.386855, 122.910942, 122.400935, 121.857384, 121.280838, 120.671848, 120.030965, 119.358737, 118.655717, 117.922453, 117.159497, 116.367398, 115.546707, 114.697974, 113.821749, 112.918582, 111.989024, 111.033625, 110.052935, 109.047505, 108.017884, 106.964623, 105.888273, 104.789382, 103.668503, 102.526184, 101.362977, 100.179431, 98.9760964, 97.7535239, 96.5122636, 95.2528657, 93.9742671, 92.6778586, 91.3650505, 90.0364454, 88.692646, 87.3342547, 85.9618743, 84.5761073, 83.1775563, 81.766824, 80.3445129, 78.9112256, 77.4675647, 76.014133, 74.5515328, 73.0803669, 71.6012379, 70.0788688, 68.5505196, 67.0175706, 65.4814024, 63.9433951, 62.4049294, 60.8673855, 59.3321439, 57.8005849, 56.274089, 54.7540366, 53.241808, 51.7387838, 50.2463441, 48.7658696, 47.2987406, 45.897027, 44.6115819, 43.3392422, 42.0781022, 40.8262564, 39.581799, 38.3428247, 37.1074276, 35.8737023, 34.6397432, 33.4036445, 32.1635008, 30.9174064, 29.6634558, 28.3997433, 27.1243633, 25.8354102, 24.2459523, 22.5025922, 20.7553583, 19.0115946, 17.2786452, 15.5638542, 13.8745655, 12.2181233, 10.6018715, 9.0331543, 7.51931562, 6.06769956, 4.68565015, 3.38051145, 2.15962749, 1.03034233, 0, 1.2716044, 2.2939432, 3.09125579, 3.6877816, 4.10776001, 4.37543044, 4.51503228, 4.55080495, 4.50698785, 4.40782038, 4.27754194, 4.14039195, 4.0206098, 3.94243491, 3.93010666, 4.00786448, 4.26401725, 4.73195716, 5.29100454, 5.97912796, 6.83429601, 7.89447726, 9.1976403, 10.7817537, 12.6847861, 14.944706, 17.5994819, 20.6870826, 24.2454766, 28.278875, 32.8713815, 38.071092, 43.9261023, 50.3287063, 57.4049008, 65.3222297, 74.1755292, 84.0596359, 95.0693861, 107.299616, 120.48505, 135.190237, 151.580467, 169.821036, 190.077235, 212.858606, 237.900131, 265.281042, 295.080571, 327.377948, 361.625511, 399.295588, 441.530868, 488.900238, 540.801066, 597.156708, 657.890519, 722.989263, 792.298479, 865.594533, 941.290805, 1022.32892, 1112.57691, 1213.32039, 1321.46925, 1435.9419, 1554.40478, 1675.45172, 1802.26718, 1935.4389, 2078.94783, 2229.46649, 2386.40803, 2549.18563, 2717.21247, 2925.4112, 3128.80813, 3317.93907, 3483.33982, 3615.54619, 3659.22539, 3681.36086, 3692.1464, 3735.98957, 3809.54119, 3895.1963, 3990.31849, 4092.27136, 4198.41849, 4306.12348, 4453.29581, 4628.0383, 4776.15646, 4843.45585, 4775.74198, 4518.82039, 4018.49661, 3440.59663, 2906.56469},
        {126.503785, 126.064996, 125.591248, 125.083065, 124.540973, 123.965498, 123.357165, 122.716501, 122.04403, 121.340279, 120.605773, 119.841037, 119.046597, 118.22298, 117.370709, 116.490312, 115.582313, 114.647239, 113.685615, 112.697966, 111.684819, 110.646698, 109.58413, 108.49764, 107.387753, 106.254996, 105.099893, 103.922972, 102.724756, 101.505772, 100.266546, 99.0076022, 97.7294674, 96.4326669, 95.1135668, 93.7748599, 92.4192891, 91.0475151, 89.6601984, 88.2579996, 86.8415793, 85.4115982, 83.9687169, 82.5135959, 81.0468959, 79.5692774, 78.0814011, 76.5839276, 75.0775176, 73.5628315, 72.04053, 70.477382, 68.9086746, 67.3358031, 65.7601626, 64.1831487, 62.6061565, 61.0305813, 59.4578186, 57.8892636, 56.3263115, 54.7703578, 53.2227978, 51.6850266, 50.1584398, 48.6444325, 47.1444, 45.7077881, 44.3828888, 43.071997, 41.7733932, 40.4853578, 39.2061712, 37.9341138, 36.6674661, 35.4045084, 34.143521, 32.8827846, 31.6205793, 30.3551857, 29.0848841, 27.8079549, 26.5226786, 25.2273355, 23.6501621, 21.9301409, 20.2094998, 18.4952827, 16.7945336, 15.1142961, 13.4616143, 11.8435319, 10.2670928, 8.73934083, 7.26731983, 5.85807367, 4.51864619, 3.25608123, 2.07742264, 0.989714286, 0, 1.2194166, 2.19837856, 2.96109699, 3.53178299, 3.93464766, 4.19390209, 4.3337574, 4.37842467, 4.35211503, 4.27903955, 4.18340936, 4.08943554, 4.0213292, 4.00330145, 4.05956337, 4.21432609, 4.52845981, 5.03141266, 5.6644763, 6.45971734, 7.44920243, 8.66499817, 10.1391712, 11.9037882, 13.9909157, 16.4326204, 19.2609689, 22.5080278, 26.2058639, 30.2960894, 34.9343918, 40.1799739, 46.0920386, 52.6712919, 60.0061852, 68.2007688, 77.3317947, 87.4760149, 98.7101812, 111.111046, 124.176799, 138.738838, 154.987111, 173.111569, 193.302158, 216.420396, 241.816771, 269.513342, 299.532166, 331.895298, 365.843859, 403.135324, 444.992971, 492.035104, 543.644873, 599.759987, 660.318152, 725.354234, 794.686704, 868.028052, 943.843331, 1024.89284, 1114.73827, 1214.57615, 1321.59539, 1434.82444, 1552.11441, 1672.18853, 1798.08701, 1930.40817, 2074.43446, 2225.35936, 2382.15969, 2543.81231, 2709.29404, 2906.10747, 3097.57226, 3275.53383, 3431.83759, 3558.32895, 3606.60732, 3635.59478, 3654.385, 3702.83469, 3777.86052, 3863.56123, 3957.49437, 4057.21752, 4160.28824, 4264.26408, 4372.46237, 4529.68358, 4677.6027, 4757.89478, 4712.23482, 4482.29785, 4009.75888, 3453.65015, 2895.23277},
        {128.77986, 128.303291, 127.792525, 127.248045, 126.670332, 126.059867, 125.417131, 124.742607, 124.036776, 123.300119, 122.533118, 121.736255, 120.91001, 120.054866, 119.171303, 118.259804, 117.32085, 116.354922, 115.362502, 114.344072, 113.300113, 112.231106, 111.137533, 110.019875, 108.878615, 107.714233, 106.527211, 105.318031, 104.087173, 102.835121, 101.562354, 100.269355, 98.9566047, 97.6245853, 96.2661508, 94.8857567, 93.4879503, 92.0734605, 90.6430166, 89.1973473, 87.7371819, 86.2632493, 84.7762785, 83.2769987, 81.7661387, 80.2444277, 78.7125947, 77.1713687, 75.6214788, 74.0636539, 72.4986231, 70.8968601, 69.2900052, 67.6794432, 66.066559, 64.4527376, 62.8393639, 61.2278228, 59.6194991, 58.0157778, 56.4180439, 54.8276821, 53.2460774, 51.6746147, 50.114679, 48.5676551, 47.0349279, 45.5622546, 44.1943269, 42.8410154, 41.5008285, 40.1722746, 38.8538622, 37.5440997, 36.2414955, 34.9445581, 33.651796, 32.3617174, 31.072831, 29.783645, 28.4926681, 27.1984085, 25.8993747, 24.5940752, 23.0283421, 21.3323366, 19.6392571, 17.9558119, 16.2887091, 14.6446571, 13.0303641, 11.4525385, 9.91788834, 8.43312206, 7.00494788, 5.64007408, 4.34520891, 3.12706065, 1.99233757, 0.947747928, 0, 1.27507146, 2.29623111, 3.08933779, 3.68025033, 4.09482759, 4.35892841, 4.49841163, 4.53913609, 4.50696064, 4.42774411, 4.32734536, 4.23162322, 4.16643654, 4.15764416, 4.23110493, 4.41267768, 4.6818923, 5.05798922, 5.67161896, 6.53871274, 7.67520176, 9.09701724, 10.8200904, 12.8603524, 15.2337345, 17.9561679, 21.0435837, 24.5119133, 28.3770878, 32.4322411, 36.997794, 42.1565171, 47.9911808, 54.6755617, 62.2469271, 70.7182761, 80.1450773, 90.5827991, 102.08691, 114.712879, 127.724245, 142.209428, 158.378839, 176.44289, 196.611991, 220.047096, 245.770438, 273.754794, 303.972939, 336.397649, 370.129747, 407.079682, 448.403185, 494.721183, 545.562716, 600.947061, 660.893492, 725.532521, 794.746906, 868.298059, 945.101206, 1027.16006, 1117.02376, 1215.57439, 1320.86935, 1432.26264, 1548.23361, 1667.90947, 1793.62445, 1925.8987, 2071.16126, 2223.11436, 2380.23276, 2540.99126, 2703.86462, 2887.66416, 3065.44398, 3230.59474, 3376.50709, 3496.57167, 3550.62895, 3587.98654, 3616.41371, 3670.07169, 3746.28295, 3831.37335, 3923.2152, 4019.68081, 4118.6425, 4217.97258, 4294.98411, 4436.21104, 4581.70294, 4671.50937, 4645.67991, 4444.26412, 4007.31157, 3485.68162, 2900.31146},
        {131.136568, 130.626126, 130.08168, 129.503691, 128.892618, 128.248921, 127.573061, 126.865497, 126.12669, 125.3571, 124.557186, 123.72741, 122.86823, 121.980107, 121.063502, 120.118873, 119.146682, 118.147389, 117.121452, 116.069333, 114.991492, 113.888388, 112.760482, 111.608234, 110.432104, 109.232552, 108.010037, 106.765021, 105.497963, 104.209323, 102.899562, 101.569139, 100.218514, 98.8481479, 97.4483572, 96.024887, 94.5836042, 93.1252981, 91.6507578, 90.1607726, 88.6561318, 87.1376246, 85.6060403, 84.062168, 82.5067971, 80.9407168, 79.3647163, 77.7795848, 76.1861116, 74.585086, 72.9772972, 71.3369774, 69.6920485, 68.0438756, 66.3938236, 64.7432573, 63.0935419, 61.4460422, 59.8021231, 58.1631496, 56.5304868, 54.9054994, 53.2895525, 51.6840109, 50.0902398, 48.5096039, 46.9434683, 45.4334092, 44.0200375, 42.6217852, 41.2374106, 39.8656719, 38.5053272, 37.1551348, 35.8138529, 34.4802396, 33.1530533, 31.8310521, 30.5129943, 29.1976379, 27.8837414, 26.5700628, 25.2553603, 23.9383922, 22.3845294, 20.7141327, 19.0503681, 17.3995677, 15.7680636, 14.1621879, 12.5882726, 11.0526497, 9.56165151, 8.12160994, 6.73885711, 5.41972509, 4.17054597, 2.99765183, 1.90737473, 0.906046764, 0, 1.3443958, 2.41732638, 3.24661991, 3.86010454, 4.28560843, 4.55095973, 4.6839866, 4.7125172, 4.66437968, 4.56740219, 4.4494129, 4.33823997, 4.26171154, 4.24765577, 4.32390082, 4.51827484, 4.71752518, 4.92932557, 5.50061557, 6.42899172, 7.71205058, 9.3473887, 11.3326026, 13.6652889, 16.3430441, 19.3634648, 22.7241474, 26.4226887, 30.456685, 34.4547191, 38.9187065, 43.9569479, 49.6777442, 56.4263688, 64.192636, 72.9031675, 82.5951722, 93.305859, 105.072437, 117.932114, 130.973407, 145.47095, 161.647567, 179.72608, 199.929313, 223.625079, 249.604963, 277.805539, 308.163381, 340.615065, 374.182431, 410.834796, 451.590087, 497.047092, 546.948868, 601.422678, 660.595784, 724.718723, 793.767469, 867.583514, 945.658233, 1029.05886, 1119.09495, 1216.26472, 1319.69159, 1429.13002, 1543.86576, 1663.53174, 1789.47965, 1922.08896, 2068.5898, 2221.45703, 2378.69874, 2538.32304, 2698.33799, 2869.07725, 3033.14196, 3185.45885, 3320.9546, 3434.55593, 3494.30358, 3539.9342, 3577.89705, 3636.60414, 3713.81117, 3798.12264, 3887.75178, 3980.91183, 4075.81602, 4170.67758, 4225.9091, 4353.74858, 4495.08542, 4590.80905, 4581.80888, 4408.97432, 4013.1948, 3531.70662, 2925.92011},
        {133.721218, 133.201864, 132.6455, 132.05269, 131.423999, 130.759992, 130.061234, 129.32829, 128.561724, 127.762101, 126.929987, 126.065945, 125.170542, 124.244341, 123.287908, 122.301807, 121.286604, 120.242862, 119.171148, 118.072026, 116.946061, 115.793817, 114.615859, 113.412753, 112.185063, 110.933354, 109.658191, 108.360139, 107.039763, 105.697626, 104.334295, 102.950335, 101.546309, 100.122783, 98.6733689, 97.2022538, 95.7137092, 94.2085256, 92.6874935, 91.1514035, 89.601046, 88.0372115, 86.4606906, 84.8722738, 83.2727515, 81.6629142, 80.0435525, 78.4154569, 76.7794178, 75.1362258, 73.4866714, 71.8054806, 70.1200735, 68.4318055, 66.7420321, 65.0521089, 63.3633914, 61.6772352, 59.9949958, 58.3180286, 56.6476893, 54.9853334, 53.3323163, 51.6899937, 50.059721, 48.4428538, 46.8407476, 45.2908449, 43.8297216, 42.3843068, 40.9536165, 39.536667, 38.1324745, 36.740055, 35.3584247, 33.9865998, 32.6235965, 31.2684309, 29.9201192, 28.5776775, 27.240122, 25.9064689, 24.5757343, 23.2469343, 21.7076481, 20.0670576, 18.4368861, 16.8230114, 15.2313112, 13.6676632, 12.137945, 10.6480343, 9.20380897, 7.81114659, 6.47592487, 5.20402153, 4.00131426, 2.87368078, 1.82699877, 0.867145943, 0, 1.30317398, 2.3356967, 3.12607611, 3.70282013, 4.09443669, 4.32943374, 4.4363192, 4.44360101, 4.3797871, 4.2733854, 4.15290385, 4.04685038, 3.98373292, 3.9920594, 4.10033777, 4.33707594, 4.53070661, 4.68115554, 5.2694818, 6.28132007, 7.70230502, 9.51807131, 11.7142536, 14.2764866, 17.1904049, 20.4416432, 24.0158361, 27.8986184, 32.0756247, 36.0731119, 40.5045309, 45.4933298, 51.1629565, 57.9328893, 65.7785612, 74.5964635, 84.4212352, 95.2875152, 107.229942, 120.283156, 133.570388, 148.315068, 164.730155, 183.028605, 203.423375, 227.248276, 253.315199, 281.556889, 311.906092, 344.295552, 377.767531, 414.233624, 454.557462, 499.302914, 548.421842, 602.140884, 660.686683, 724.471539, 793.494236, 867.551016, 946.446284, 1030.67491, 1120.69462, 1216.76397, 1318.80417, 1426.89127, 1541.05031, 1661.34413, 1788.02259, 1921.22976, 2067.97302, 2220.47724, 2376.51071, 2533.84175, 2690.23867, 2850.33847, 3003.32358, 3145.24514, 3272.15428, 3380.10213, 3442.99836, 3493.79655, 3538.037, 3600.3633, 3678.75407, 3763.22826, 3852.16247, 3943.93331, 4036.9174, 4129.49134, 4181.26721, 4301.20767, 4435.25009, 4529.33184, 4529.39027, 4381.36275, 4031.18664, 3585.97283, 2982.41138},
        {136.633402, 136.143578, 135.608855, 135.030096, 134.408161, 133.743913, 133.038214, 132.291925, 131.505909, 130.681027, 129.818141, 128.918113, 127.981804, 127.010077, 126.003794, 124.963815, 123.891004, 122.786222, 121.650331, 120.484192, 119.288667, 118.064619, 116.812909, 115.534399, 114.229951, 112.900427, 111.546688, 110.169596, 108.770014, 107.348803, 105.906824, 104.444941, 102.964014, 101.464905, 99.9534323, 98.4278753, 96.8864542, 95.3298701, 93.7588239, 92.1740167, 90.5761495, 88.9659234, 87.3440393, 85.7111982, 84.0681013, 82.4154495, 80.7539438, 79.0842852, 77.4071749, 75.7233137, 74.0334028, 72.3072297, 70.577079, 68.8443219, 67.1103295, 65.376473, 63.6441236, 61.9146524, 60.1894306, 58.4698294, 56.7572199, 55.0529734, 53.3584608, 51.6750535, 50.0041226, 48.3470392, 46.7051746, 45.1123103, 43.6014498, 42.1071201, 40.6285913, 39.1651337, 37.7160173, 36.2805122, 34.8578885, 33.4474164, 32.0483661, 30.6600076, 29.281611, 27.9124466, 26.5517843, 25.1988944, 23.853047, 22.5135122, 20.9930005, 19.3879859, 17.7971847, 16.2259212, 14.6795196, 13.1633045, 11.6826, 10.2427306, 8.84902062, 7.50679436, 6.22137618, 4.99809042, 3.84226143, 2.75921355, 1.75427112, 0.83275849, 0, 1.0678891, 1.8996743, 2.52213285, 2.962042, 3.246179, 3.4013211, 3.45424556, 3.43172963, 3.36055054, 3.26748557, 3.17931195, 3.12280694, 3.12474779, 3.21191174, 3.41107606, 3.74901799, 4.06408446, 4.35613408, 5.07815559, 6.21654829, 7.75771151, 9.68804458, 11.9939468, 14.6618175, 17.6780561, 21.0290618, 24.7012339, 28.6809718, 32.9546749, 37.0661323, 41.6066438, 46.6953918, 52.4515585, 59.2045717, 66.968491, 75.7013107, 85.4591394, 96.2980859, 108.274259, 121.443767, 135.270035, 150.582238, 167.552444, 186.352722, 207.155141, 230.925104, 256.843011, 284.882598, 315.0176, 347.221751, 380.679207, 417.118326, 457.277243, 501.682527, 550.428793, 603.804595, 662.098484, 725.926692, 795.175623, 869.374215, 948.110444, 1032.05063, 1121.62866, 1217.15769, 1318.74629, 1426.59696, 1541.21331, 1662.87591, 1790.76115, 1924.67001, 2069.92837, 2219.96232, 2372.66037, 2525.91104, 2677.60283, 2831.27718, 2977.75645, 3113.51592, 3235.03088, 3338.77661, 3400.57055, 3451.31772, 3496.34676, 3559.93198, 3639.90269, 3726.19421, 3817.03053, 3910.63569, 4005.2337, 4099.04859, 4171.45839, 4290.90963, 4413.6728, 4496.01838, 4494.21688, 4364.53879, 4063.25461, 3644.13305, 3077.02945},
        {139.66004, 139.20204, 138.690888, 138.127756, 137.513815, 136.850235, 136.138188, 135.378845, 134.573377, 133.722954, 132.828749, 131.891931, 130.913671, 129.895142, 128.837513, 127.741957, 126.609643, 125.441743, 124.239427, 123.003868, 121.736236, 120.437701, 119.109435, 117.752609, 116.368395, 114.957962, 113.522482, 112.063125, 110.581064, 109.077469, 107.553511, 106.010361, 104.449189, 102.871168, 101.294657, 99.7118717, 98.1148194, 96.5041137, 94.8803676, 93.2441943, 91.5962071, 89.937019, 88.2672434, 86.5874932, 84.8983818, 83.2005223, 81.4945279, 79.7810117, 78.060587, 76.3338669, 74.6014645, 72.8281677, 71.0511917, 69.2719262, 67.4917611, 65.7120862, 63.9342913, 62.1597661, 60.3899005, 58.6260843, 56.8697073, 55.1221593, 53.3848301, 51.6591094, 49.9463872, 48.2480531, 46.5654971, 44.9289962, 43.3681342, 41.8247228, 40.298279, 38.78832, 37.294363, 35.8159252, 34.3525236, 32.9036755, 31.468898, 30.0477082, 28.6396232, 27.2441603, 25.8608366, 24.4891692, 23.1286753, 21.778872, 20.277684, 18.7088237, 17.1579453, 15.6298102, 14.1291793, 12.660814, 11.2294752, 9.83992416, 8.49692199, 7.20522981, 5.96960877, 4.79482001, 3.68562465, 2.64678383, 1.6830587, 0.799210374, 0, 0.82170609, 1.44444785, 1.89295243, 2.19194702, 2.36615876, 2.44031483, 2.43914238, 2.38736859, 2.30972061, 2.23092562, 2.17571077, 2.16880324, 2.23493018, 2.39881875, 2.68519614, 3.11878949, 3.5702416, 4.0422676, 4.9081188, 6.15950429, 7.78813316, 9.78571448, 12.1439574, 14.8545709, 17.9092641, 21.2997461, 25.0177261, 29.054913, 33.4030159, 37.665436, 42.3645699, 47.6086192, 53.5057854, 60.252647, 67.913217, 76.5279412, 86.1785079, 96.9466056, 108.913923, 122.162147, 136.539749, 152.432615, 169.968063, 189.273412, 210.475979, 234.103668, 259.783066, 287.541344, 317.405674, 349.403228, 382.89861, 419.391364, 459.512669, 503.74944, 552.294061, 605.471153, 663.605339, 727.44834, 796.80061, 870.996681, 949.416607, 1032.94474, 1122.11667, 1217.32629, 1318.72747, 1426.58429, 1541.76169, 1664.67952, 1793.55432, 1927.84894, 2070.90902, 2217.71728, 2366.39255, 2515.05363, 2661.81936, 2810.65877, 2952.3779, 3083.63301, 3201.08037, 3301.37624, 3360.98372, 3410.21434, 3454.37858, 3518.16252, 3599.20712, 3687.0802, 3779.82637, 3875.49026, 3972.11651, 4067.74974, 4169.99219, 4291.99296, 4402.35578, 4469.68439, 4462.58254, 4349.65394, 4099.50235, 3711.04734, 3196.64969},
        {142.743311, 142.307924, 141.812207, 141.257595, 140.645522, 139.977423, 139.254732, 138.478884, 137.651313, 136.773454, 135.846742, 134.87261, 133.852493, 132.787826, 131.680043, 130.530579, 129.340868, 128.112345, 126.846443, 125.544599, 124.208246, 122.838818, 121.43775, 120.006477, 118.546433, 117.059053, 115.545771, 114.008021, 112.447238, 110.864857, 109.262312, 107.641038, 106.002468, 104.348038, 102.706201, 101.064313, 99.409406, 97.7420387, 96.0627685, 94.372153, 92.6707499, 90.9591169, 89.2378117, 87.507392, 85.7684154, 84.0214397, 82.2670226, 80.5057217, 78.7380948, 76.9646994, 75.1860935, 73.363631, 71.5379231, 69.7103773, 67.8824012, 66.0554022, 64.230788, 62.4099661, 60.594344, 58.7853293, 56.9843295, 55.1927522, 53.412005, 51.6434953, 49.8886307, 48.1488187, 46.425467, 44.7455568, 43.1354555, 41.5438265, 39.9704197, 38.4149846, 36.877271, 35.3570286, 33.8540071, 32.3679562, 30.8986255, 29.4457649, 28.0091239, 26.5884524, 25.1834999, 23.7940162, 22.4197511, 21.0604541, 19.5780427, 18.0440818, 16.531793, 15.0453991, 13.5891226, 12.1671862, 10.7838124, 9.44322398, 8.14964346, 6.9072935, 5.72039672, 4.59317575, 3.52985322, 2.53465176, 1.61179398, 0.76550252, 0, 0.615368177, 1.06345593, 1.36711707, 1.5492054, 1.63257475, 1.64007891, 1.59457171, 1.51890694, 1.43593843, 1.36851998, 1.33950541, 1.37174852, 1.48810314, 1.71142305, 2.06456209, 2.57037406, 3.14786174, 3.80504308, 4.79967304, 6.13235162, 7.80367885, 9.81425476, 12.1646793, 14.8555526, 17.8874747, 21.2610454, 24.9768649, 29.0355333, 33.4376504, 37.8777685, 42.7747531, 48.2210183, 54.3089785, 61.087461, 68.6706737, 77.1844476, 86.7342735, 97.4256416, 109.364043, 122.654967, 137.49311, 153.867608, 171.866499, 191.577819, 213.089607, 236.510298, 261.902431, 289.348945, 318.932779, 350.736868, 384.314037, 420.925258, 461.130605, 505.380174, 553.899522, 607.015018, 665.053033, 728.776869, 797.976661, 871.965893, 950.063786, 1033.25684, 1122.17279, 1217.25204, 1318.6175, 1426.5378, 1542.09808, 1665.77867, 1795.06606, 1929.14025, 2069.43592, 2212.53146, 2356.82637, 2500.72017, 2642.61235, 2787.77695, 2926.02031, 3054.02332, 3168.46683, 3266.03173, 3323.11419, 3370.20292, 3412.67223, 3475.78723, 3557.04881, 3645.67001, 3739.56609, 3836.65234, 3934.84402, 4032.05641, 4168.69618, 4295.03871, 4392.90203, 4444.10422, 4430.46333, 4333.79741, 4135.92453, 3787.00386, 3338.26168},
        {145.807159, 145.372063, 144.872239, 144.309268, 143.684733, 143.000218, 142.257304, 141.457575, 140.602613, 139.694001, 138.733322, 137.722158, 136.662092, 135.554706, 134.401584, 133.204309, 131.964462, 130.683626, 129.363385, 128.005321, 126.611016, 125.182054, 123.720016, 122.226487, 120.703048, 119.151282, 117.572771, 115.969099, 114.341849, 112.692602, 111.022942, 109.334452, 107.628713, 105.907309, 104.20292, 102.500926, 100.786333, 99.0597138, 97.3216429, 95.5726933, 93.8134387, 92.0444525, 90.2663084, 88.4795797, 86.68484, 84.8826628, 83.0736217, 81.2582901, 79.4372417, 77.6110499, 75.7802881, 73.9064251, 72.0299868, 70.1523948, 68.2750701, 66.399434, 64.5269078, 62.6589127, 60.7968699, 58.9422007, 57.0963264, 55.260668, 53.4366471, 51.6256846, 49.8292019, 48.0486203, 46.2853609, 44.5634659, 42.9064341, 41.2690331, 39.6512176, 38.0529426, 36.4741629, 34.9148332, 33.3749084, 31.8543434, 30.353093, 28.8711119, 27.408355, 25.9647772, 24.5403333, 23.134978, 21.7486663, 20.3813529, 18.9157812, 17.4129772, 15.9353756, 14.4867349, 13.0708138, 11.6913707, 10.3521643, 9.05695308, 7.80949566, 6.61355057, 5.4728764, 4.3912317, 3.37237503, 2.42006496, 1.53806006, 0.730118882, 0, 0.524622521, 0.896160654, 1.13650768, 1.2675569, 1.31120158, 1.28933501, 1.22385048, 1.13664127, 1.04960066, 0.984621948, 0.96359841, 1.00842333, 1.14099, 1.3831917, 1.75692171, 2.28407333, 2.94089493, 3.74275911, 4.81480356, 6.1691405, 7.81788219, 9.77314084, 12.0470287, 14.651658, 17.599141, 20.9015899, 24.571117, 28.6198344, 33.0598545, 37.6984365, 42.8276584, 48.5210882, 54.852294, 61.7390257, 69.3327606, 77.8265274, 87.3406397, 97.9954112, 109.911155, 123.208186, 138.283653, 154.896779, 173.113714, 193.000607, 214.62361, 237.802191, 262.910851, 290.077409, 319.429686, 351.095501, 384.781202, 421.551213, 461.962807, 506.441283, 555.142485, 608.34324, 666.320371, 729.630028, 798.206227, 871.678268, 949.662218, 1032.89313, 1121.89162, 1217.02384, 1318.39412, 1426.22699, 1541.6539, 1665.13451, 1793.80295, 1926.67445, 2063.78483, 2202.97174, 2342.89712, 2482.22293, 2619.61114, 2761.78214, 2897.32483, 3022.88657, 3135.11469, 3230.65655, 3285.76506, 3331.07829, 3371.98404, 3433.74725, 3513.8877, 3601.64306, 3694.94723, 3791.7341, 3889.93757, 3987.49155, 4158.592, 4289.87861, 4376.18638, 4412.35029, 4393.20536, 4313.58656, 4168.3289, 3874.0066, 3500.67456},
        {148.599376, 148.113621, 147.565791, 146.957321, 146.289643, 145.564189, 144.782394, 143.945689, 143.055507, 142.113282, 141.120447, 140.078433, 138.988674, 137.852603, 136.671653, 135.447256, 134.180845, 132.873854, 131.527715, 130.143861, 128.723725, 127.268739, 125.780337, 124.259952, 122.709015, 121.128961, 119.521222, 117.887231, 116.22842, 114.546223, 112.842073, 111.117401, 109.373642, 107.612228, 105.854736, 104.092102, 102.315022, 100.524276, 98.7206418, 96.9048995, 95.0778279, 93.2402064, 91.3928139, 89.5364298, 87.6718331, 85.799803, 83.9211187, 82.0365594, 80.1469042, 78.2529323, 76.3554228, 74.4249068, 72.4930673, 70.5613392, 68.6311574, 66.7039568, 64.7811722, 62.8642385, 60.9545907, 59.0536636, 57.162892, 55.283711, 53.4175553, 51.5658599, 49.7300596, 47.9115894, 46.111884, 44.3534771, 42.6583955, 40.9845598, 39.3320372, 37.7008948, 36.0911998, 34.5030193, 32.9364205, 31.3914705, 29.8682365, 28.3667856, 26.8871849, 25.4295017, 23.9938031, 22.5801562, 21.1886281, 19.8192861, 18.3643943, 16.8801888, 15.4242015, 13.9999979, 12.6111436, 11.2612041, 9.95374504, 8.69233193, 7.48053035, 6.32190585, 5.220024, 4.17845036, 3.2007505, 2.29048997, 1.45123433, 0.686549155, 0, 0.86676665, 1.52663418, 2.00450238, 2.32527101, 2.51383987, 2.59510872, 2.59397735, 2.53534554, 2.44411306, 2.34517969, 2.26344522, 2.22380941, 2.25117206, 2.37043293, 2.60649181, 2.98424848, 3.53058695, 4.27069047, 5.22497065, 6.41875247, 7.87736089, 9.6261209, 11.6903575, 14.0953956, 16.8665601, 20.0291762, 23.6085687, 27.6300627, 32.118983, 37.0122411, 42.4559939, 48.5020904, 55.2023796, 62.4273366, 70.3194974, 79.0697645, 88.7843991, 99.5696624, 111.531816, 124.77712, 139.450896, 155.608458, 173.348426, 192.769418, 213.970053, 236.968693, 261.96428, 289.075497, 318.421029, 350.11956, 383.84102, 420.701322, 461.501839, 506.703035, 556.069357, 609.677749, 667.605157, 729.506864, 795.977307, 868.070918, 946.964527, 1031.82387, 1122.14457, 1217.76044, 1319.07835, 1426.24213, 1540.13852, 1661.10393, 1786.75088, 1916.23265, 2049.86288, 2185.45605, 2321.7638, 2457.53781, 2591.52971, 2730.44052, 2863.0852, 2986.22809, 3096.63349, 3191.06573, 3247.03219, 3293.39208, 3335.16271, 3395.0011, 3470.94215, 3553.67086, 3641.58963, 3733.10085, 3826.6069, 3920.51017, 4122.91524, 4259.10561, 4336.04929, 4360.71428, 4340.06859, 4281.08024, 4190.71721, 3990.63517, 3700.27556},
        {151.484769, 150.92313, 150.3044, 149.629775, 148.900454, 148.117637, 147.282522, 146.396308, 145.460193, 144.475377, 143.443057, 142.364432, 141.240702, 140.073065, 138.862719, 137.610864, 136.318697, 134.987418, 133.618226, 132.212318, 130.770894, 129.295153, 127.786292, 126.245512, 124.67401, 123.072985, 121.443636, 119.787161, 118.10476, 116.397631, 114.666972, 112.913983, 111.139862, 109.345807, 107.538693, 105.716763, 103.878187, 102.023983, 100.155163, 98.2727419, 96.3777349, 94.4711563, 92.5540206, 90.6273421, 88.6921355, 86.7494151, 84.8001957, 82.8454915, 80.8863172, 78.9236872, 76.958616, 74.9730703, 72.9875254, 71.0034087, 69.0221476, 67.0451696, 65.073902, 63.1097722, 61.1542078, 59.2086361, 57.2744846, 55.3531806, 53.4461516, 51.554825, 49.6806283, 47.8249888, 45.9893341, 44.1939497, 42.4586688, 40.7460248, 39.0562227, 37.3894674, 35.7459637, 34.1259167, 32.5295311, 30.957012, 29.4085643, 27.8843928, 26.3847024, 24.9096982, 23.459585, 22.0345677, 20.6348512, 19.2606405, 17.8159706, 16.3511543, 14.9176599, 13.5188133, 12.1579403, 10.8383667, 9.56341835, 8.33642102, 7.16070053, 6.03958269, 4.97639331, 3.9744582, 3.03710317, 2.16765402, 1.36943657, 0.645776629, 0, 1.11300895, 1.98159341, 2.63256116, 3.09271996, 3.3888776, 3.54784185, 3.5964205, 3.56142131, 3.46965207, 3.34792056, 3.22303454, 3.12180181, 3.07103013, 3.09752728, 3.22810104, 3.4895592, 3.96644858, 4.69382542, 5.56789721, 6.62784438, 7.91284738, 9.46208663, 11.3147426, 13.5099957, 16.0870263, 19.0850149, 22.543142, 26.5005879, 30.9965332, 36.0910557, 41.794776, 48.1406051, 55.1614544, 62.7069251, 70.9015834, 79.9188778, 89.8467125, 100.772991, 112.785619, 125.972499, 140.182357, 155.815069, 173.005335, 191.887855, 212.597328, 235.338465, 260.158453, 287.17449, 316.503772, 348.263498, 382.137062, 419.205667, 460.428366, 506.272962, 556.204302, 610.137539, 667.987826, 728.67248, 793.331273, 864.192531, 943.727339, 1029.79501, 1121.082, 1216.97319, 1318.03698, 1424.29853, 1536.28768, 1654.16041, 1776.2222, 1901.82529, 2031.87395, 2163.92959, 2296.8072, 2429.32177, 2560.28829, 2696.3669, 2826.56614, 2947.73971, 3056.74132, 3150.42466, 3207.48875, 3255.04509, 3297.72796, 3355.44087, 3426.77403, 3503.95545, 3585.89345, 3671.49632, 3759.67239, 3849.32996, 4072.17394, 4210.6683, 4280.97078, 4299.23914, 4281.63114, 4244.30453, 4203.41708, 4097.40899, 3891.88039},
        {154.592035, 153.922288, 153.202676, 152.434097, 151.617448, 150.753625, 149.843525, 148.888046, 147.888084, 146.844536, 145.7583, 144.630272, 143.461349, 142.252428, 141.004406, 139.718181, 138.394648, 137.034705, 135.639249, 134.209177, 132.745386, 131.248772, 129.720233, 128.160666, 126.570967, 124.952034, 123.304764, 121.630052, 119.928798, 118.201896, 116.450245, 114.674741, 112.876281, 111.055763, 109.203422, 107.325711, 105.429208, 103.515157, 101.5848, 99.6393801, 97.6801406, 95.7083241, 93.7251736, 91.731932, 89.729842, 87.7201467, 85.7040889, 83.6829115, 81.6578575, 79.6301696, 77.6010908, 75.5651136, 73.5303777, 71.4982722, 69.4701864, 67.4475094, 65.4316306, 63.4239392, 61.4258244, 59.4386754, 57.4638814, 55.5028317, 53.5569155, 51.627522, 49.7160406, 47.8238603, 45.9523704, 44.1181261, 42.3373185, 40.5800582, 38.8467512, 37.1378036, 35.4536215, 33.7946109, 32.161178, 30.5537288, 28.9726695, 27.418406, 25.8913445, 24.3918911, 22.9204518, 21.4774327, 20.06324, 18.6782796, 17.244988, 15.8043961, 14.398472, 13.0301519, 11.7023723, 10.4180694, 9.1801797, 7.99163942, 6.85538496, 5.77435265, 4.75147885, 3.7896999, 2.89195214, 2.06117192, 1.3002956, 0.612259509, 0, 1.05880507, 1.88523268, 2.50487214, 2.94331276, 3.22614384, 3.37895468, 3.42733458, 3.39687286, 3.31315881, 3.20178174, 3.08833096, 2.99839576, 2.95756545, 2.99142934, 3.12557673, 3.38559692, 3.92448318, 4.78602539, 5.7076372, 6.74220879, 7.94263029, 9.36179186, 11.0525837, 13.0678958, 15.4606186, 18.283642, 21.5898562, 25.4321514, 29.8634178, 35.0412246, 40.8754005, 47.3874319, 54.5988054, 62.3729792, 70.8104538, 80.0538705, 90.1721243, 101.23411, 113.308722, 126.464856, 140.299092, 155.496387, 172.218045, 190.625372, 210.87967, 233.20675, 257.687284, 284.46645, 313.689426, 345.501389, 379.743461, 417.236501, 458.860627, 505.022858, 555.164303, 609.159743, 666.88396, 727.034388, 790.930737, 861.177099, 940.691489, 1026.86411, 1118.18116, 1213.85197, 1314.31107, 1419.43061, 1529.31848, 1643.90794, 1762.26786, 1883.95606, 2010.55483, 2139.28606, 2269.00678, 2398.57404, 2526.84487, 2660.40345, 2788.44789, 2907.90344, 3015.69536, 3108.74889, 3166.81611, 3215.44424, 3258.91843, 3314.66809, 3381.77549, 3453.84612, 3530.23607, 3610.30145, 3693.39835, 3778.88287, 4003.08807, 4138.95052, 4206.78041, 4226.8879, 4219.58316, 4205.17636, 4203.97767, 4179.55883, 4057.26003},
        {158.269136, 157.416606, 156.52594, 155.597615, 154.632109, 153.629901, 152.591467, 151.517285, 150.407833, 149.263589, 148.085031, 146.872636, 145.626882, 144.348247, 143.037209, 141.694245, 140.319832, 138.91445, 137.478575, 136.012685, 134.517258, 132.992771, 131.439703, 129.858532, 128.249734, 126.613787, 124.95117, 123.26236, 121.547835, 119.808073, 118.04355, 116.254746, 114.442137, 112.606202, 110.718811, 108.795346, 106.851536, 104.888785, 102.908501, 100.912089, 98.9009549, 96.8765056, 94.8401469, 92.7932849, 90.7373257, 88.6736753, 86.60374, 84.5289257, 82.4506386, 80.3702848, 78.2892704, 76.2134229, 74.1396311, 72.0692054, 70.0034559, 67.9436929, 65.8912268, 63.8473676, 61.8134257, 59.7907114, 57.7805349, 55.7842064, 53.8030363, 51.8383347, 49.891412, 47.9635783, 46.056144, 44.179403, 42.3434337, 40.5310179, 38.7428837, 36.9797588, 35.2423711, 33.5314485, 31.8477189, 30.1919101, 28.5647501, 26.9669666, 25.3992876, 23.862441, 22.3571546, 20.8841562, 19.4441739, 18.0379354, 16.6209898, 15.2176047, 13.8525906, 12.5281415, 11.2464515, 10.0097146, 8.82012483, 7.67987613, 6.59116255, 5.5561781, 4.57711681, 3.65617268, 2.79553973, 1.99741197, 1.26398342, 0.597448094, 0, 0.201162701, 0.31539633, 0.358712714, 0.347123681, 0.29664106, 0.223276678, 0.143042364, 0.0719499454, 0.0260112507, 0.0212381079, 0.0736423451, 0.19923579, 0.414030272, 0.734037617, 1.17526965, 1.75373821, 2.72159061, 4.12857231, 5.45657916, 6.77222345, 8.14211746, 9.6328735, 11.3111038, 13.2434208, 15.4964366, 18.1367636, 21.2310141, 24.8458003, 29.0477346, 34.0267367, 39.6808988, 46.0398407, 53.1331825, 60.9200085, 69.4652067, 78.8524744, 89.1325924, 100.356342, 112.574502, 125.837856, 139.674889, 154.817634, 171.419062, 189.632141, 209.60984, 231.065619, 254.701833, 280.781327, 309.566948, 341.32154, 376.449342, 414.898995, 456.844423, 502.454804, 551.889623, 605.312717, 662.887922, 724.476541, 790.613703, 862.164577, 940.278951, 1024.04517, 1113.26342, 1207.67286, 1306.90933, 1410.65609, 1518.60131, 1630.42952, 1745.80729, 1864.41135, 1988.24954, 2114.30883, 2241.45902, 2368.56991, 2494.51129, 2625.49497, 2751.21324, 2868.7004, 2974.99076, 3067.11861, 3125.83359, 3175.3111, 3219.5646, 3274.38696, 3339.12893, 3408.23317, 3481.29583, 3557.91307, 3637.68103, 3720.19588, 3906.68642, 4029.62278, 4103.27817, 4141.9258, 4159.83887, 4171.29059, 4190.55416, 4213.1286, 4171.88208},
        {161.951839, 160.946857, 159.912523, 158.849006, 157.756475, 156.635097, 155.485042, 154.306477, 153.099572, 151.864494, 150.601413, 149.310497, 147.991914, 146.645832, 145.272421, 143.871849, 142.444284, 140.989895, 139.50885, 138.001319, 136.467468, 134.907467, 133.321485, 131.709689, 130.072249, 128.409333, 126.721109, 125.007745, 123.269412, 121.506276, 119.718506, 117.906271, 116.06974, 114.209081, 112.28233, 110.311609, 108.319544, 106.30767, 104.277525, 102.230642, 100.168559, 98.092811, 96.0049335, 93.9064623, 91.7989333, 89.6838822, 87.5628448, 85.4373568, 83.3089539, 81.1791719, 79.0495467, 76.9379218, 74.8291717, 72.7244784, 70.6250242, 68.5319913, 66.446562, 64.3699184, 62.3032429, 60.2477176, 58.2045247, 56.1748465, 54.1598652, 52.160763, 50.1787222, 48.214925, 46.2705536, 44.3473279, 42.4469549, 40.569508, 38.7161344, 36.8879816, 35.0861969, 33.3119277, 31.5663213, 29.8505252, 28.1656867, 26.5129531, 24.8934718, 23.3083903, 21.7588558, 20.2460158, 18.7710176, 17.3350085, 15.9340472, 14.5719325, 13.2525243, 11.9771352, 10.7470775, 9.56366398, 8.42820697, 7.34201904, 6.30641269, 5.32270042, 4.39219476, 3.51620822, 2.69605329, 1.93304249, 1.22848834, 0.58370334, 0, -0.72127124, -1.37707989, -1.96125116, -2.46761025, -2.88998237, -3.22219272, -3.45806651, -3.59142896, -3.61610525, -3.52592061, -3.31470023, -2.97626932, -2.50445309, -1.89307675, -1.1359655, -0.226944541, 1.16865659, 3.10394063, 4.84205646, 6.4595708, 8.03305036, 9.63906188, 11.3541721, 13.2549477, 15.4179554, 17.919762, 20.8369341, 24.2460386, 28.2236421, 32.9717612, 38.3955145, 44.5338337, 51.4256506, 59.0934404, 67.5843632, 76.9499673, 87.2341213, 98.4806941, 110.733554, 124.036571, 137.964163, 153.172524, 169.797372, 187.974424, 207.839398, 228.588273, 251.531438, 277.039547, 305.48325, 337.233199, 373.229573, 412.577409, 454.726926, 499.609006, 548.135892, 600.779213, 658.0106, 720.980895, 789.328154, 861.949475, 937.917617, 1018.37308, 1104.71868, 1197.35335, 1294.97495, 1397.06218, 1502.86888, 1611.81545, 1724.14668, 1839.64106, 1960.6156, 2083.91969, 2208.45305, 2333.11543, 2456.80657, 2585.83862, 2709.84582, 2825.87479, 2930.97218, 3022.18463, 3080.62619, 3129.8978, 3173.87435, 3227.55388, 3290.40834, 3357.27508, 3427.8578, 3501.86019, 3578.98597, 3658.93883, 3797.67478, 3904.60586, 3985.34901, 4045.52113, 4090.73915, 4126.61997, 4158.78053, 4212.38249, 4237.43807},
        {165.33743, 164.276512, 163.187006, 162.069043, 160.922759, 159.748286, 158.54576, 157.315314, 156.057083, 154.771199, 153.457797, 152.117012, 150.748977, 149.353825, 147.931692, 146.482711, 145.007016, 143.504741, 141.97602, 140.420987, 138.839776, 137.232521, 135.599355, 133.940414, 132.255831, 130.545739, 128.810274, 127.049568, 125.263757, 123.452973, 121.617351, 119.757024, 117.872128, 115.962796, 113.982821, 111.956484, 109.908619, 107.840864, 105.754856, 103.652234, 101.534635, 99.4036973, 97.2610582, 95.1083557, 92.9472275, 90.7793115, 88.6062453, 86.4296669, 84.251214, 82.0725244, 79.8952358, 77.7518528, 75.6124773, 73.478078, 71.3496235, 69.2280824, 67.1144234, 65.0096152, 62.9146264, 60.8304257, 58.7579817, 56.6982631, 54.6522385, 52.6208765, 50.6051459, 48.6060153, 46.6244534, 44.6501694, 42.6731432, 40.7175648, 38.7851328, 36.8775456, 34.9965018, 33.1437, 31.3208387, 29.5296164, 27.7717318, 26.0488833, 24.3627696, 22.715089, 21.1075403, 19.5418218, 18.0196323, 16.5426702, 15.1536989, 13.833021, 12.560445, 11.3363369, 10.1610627, 9.03498812, 7.95847922, 6.93190186, 5.95562195, 5.0300054, 4.15541812, 3.33222602, 2.56079501, 1.84149099, 1.17467988, 0.560727577, 0, -1.34793107, -2.53370547, -3.55713416, -4.41802812, -5.1161983, -5.65145566, -6.02361116, -6.23247578, -6.27786047, -6.1595762, -5.87743392, -5.43124461, -4.82081922, -4.04596872, -3.10650407, -2.00223623, -0.364452217, 1.85968329, 3.83679324, 5.64603609, 7.36657031, 9.07755434, 10.8581466, 12.7875057, 14.9447899, 17.4091577, 20.2597676, 23.5757781, 27.4363476, 32.0539098, 37.3254804, 43.2902351, 49.9873498, 57.384406, 65.5563769, 74.597327, 84.5679107, 95.5287821, 107.540595, 120.664004, 134.666676, 149.991421, 166.756219, 185.079047, 205.077882, 225.676065, 248.48487, 273.920935, 302.400898, 334.341395, 370.970665, 410.901788, 453.138992, 497.461032, 545.227983, 597.089939, 653.696994, 717.040623, 786.12468, 858.489696, 931.628423, 1008.35767, 1091.51878, 1182.07832, 1277.82611, 1378.01015, 1481.40101, 1587.12294, 1696.05064, 1808.06867, 1925.66013, 2045.71121, 2167.20695, 2289.13239, 2410.47257, 2538.76811, 2662.30956, 2777.94305, 2882.51472, 2972.87071, 3029.20678, 3076.1197, 3117.59146, 3169.19718, 3230.32994, 3295.55236, 3364.49906, 3436.80464, 3512.1037, 3590.03086, 3681.22728, 3771.34063, 3857.48261, 3936.76494, 4006.29931, 4063.19742, 4104.57097, 4176.53959, 4241.61291},
        {168.065254, 167.12278, 166.138483, 165.11293, 164.04669, 162.940329, 161.794416, 160.609519, 159.386205, 158.125042, 156.826598, 155.49144, 154.120137, 152.713256, 151.271364, 149.79503, 148.284821, 146.741305, 145.16505, 143.556624, 141.916594, 140.245528, 138.543993, 136.812559, 135.051791, 133.262259, 131.444529, 129.59917, 127.726749, 125.827834, 123.902993, 121.952794, 119.977803, 117.97859, 115.920095, 113.821449, 111.70221, 109.564102, 107.40885, 105.238176, 103.053805, 100.85746, 98.6508664, 96.4357472, 94.2138264, 91.9868279, 89.7564756, 87.5244935, 85.2926053, 83.0625351, 80.8360068, 78.6659139, 76.5017014, 74.3439838, 72.193376, 70.0504924, 67.9159477, 65.7903566, 63.6743337, 61.5684936, 59.473451, 57.3898205, 55.3182168, 53.2592545, 51.2135483, 49.1817127, 47.1643624, 45.1340625, 43.0640514, 41.012794, 38.9827233, 36.9762722, 34.9958738, 33.043961, 31.1229669, 29.2353243, 27.3834663, 25.5698258, 23.7968359, 22.0669295, 20.3825396, 18.7460992, 17.1600412, 15.6267987, 14.2409923, 12.9570221, 11.7281801, 10.5539078, 9.43364632, 8.36683713, 7.35292149, 6.39134071, 5.48153612, 4.62294903, 3.81502076, 3.05719262, 2.34890593, 1.68960201, 1.07872217, 0.515707727, 0, -1.25416717, -2.37881959, -3.3710563, -4.22797632, -4.9466787, -5.52426246, -5.95782663, -6.24447025, -6.38129234, -6.36539194, -6.19386809, -5.8638198, -5.37234613, -4.71654608, -3.89351871, -2.90036304, -1.42146429, 0.590751892, 2.43185764, 4.17176403, 5.88038213, 7.627623, 9.48339771, 11.5176173, 13.8001929, 16.4010356, 19.3900564, 22.8371663, 26.8122765, 31.5590665, 36.9099638, 42.8827489, 49.4952024, 56.4332668, 63.8806418, 72.1095169, 81.2372241, 91.3810954, 102.658463, 115.186658, 129.122775, 144.532283, 161.524733, 180.209678, 200.69667, 222.173915, 245.902648, 272.222759, 301.474137, 333.996671, 370.802689, 410.737771, 452.97811, 497.335813, 544.921309, 596.262107, 651.885715, 713.515771, 780.211808, 849.728484, 919.381166, 992.479508, 1072.65933, 1161.11998, 1254.93231, 1353.0622, 1453.68057, 1555.5472, 1660.33683, 1768.07534, 1881.25477, 1997.05542, 2114.64951, 2233.20925, 2351.90684, 2481.48184, 2606.64728, 2723.68355, 2828.87102, 2918.49007, 2969.66903, 3010.60836, 3045.79645, 3093.49801, 3152.74831, 3216.84251, 3285.10291, 3356.85179, 3431.41145, 3508.10417, 3562.1873, 3636.83735, 3723.61979, 3814.10014, 3899.84389, 3972.41653, 4023.38356, 4104.51333, 4169.41828},
        {170.892876, 170.096424, 169.242107, 168.331034, 167.364316, 166.343063, 165.268387, 164.141398, 162.963205, 161.734921, 160.457654, 159.132517, 157.760619, 156.343071, 154.880983, 153.375466, 151.827631, 150.238588, 148.609448, 146.94132, 145.235317, 143.492547, 141.714123, 139.901154, 138.05475, 136.176023, 134.266083, 132.326041, 130.357006, 128.36009, 126.336403, 124.287056, 122.213159, 120.115822, 117.978275, 115.810944, 113.624472, 111.420551, 109.200871, 106.967124, 104.720999, 102.464188, 100.198382, 97.925272, 95.646548, 93.3639013, 91.0790228, 88.7936034, 86.5093339, 84.2279051, 81.951008, 79.7387223, 77.5330842, 75.3345189, 73.1434513, 70.9603067, 68.7855101, 66.6194867, 64.4626615, 62.3154597, 60.1783063, 58.0516265, 55.9358454, 53.831388, 51.7386795, 49.658145, 47.5902096, 45.4946884, 43.3323716, 41.1874393, 39.0629492, 36.961959, 34.8875264, 32.8427091, 30.8305647, 28.8541509, 26.9165255, 25.0207461, 23.1698704, 21.3669561, 19.6150609, 17.9172424, 16.2765584, 14.6960666, 13.3230988, 12.0855411, 10.9095453, 9.7934871, 8.73574223, 7.73468637, 6.78869525, 5.89614456, 5.05541002, 4.26486734, 3.52289222, 2.82786038, 2.17814751, 1.57212933, 1.00818155, 0.484679867, 0, -1.25722132, -2.40737465, -3.44556416, -4.366894, -5.16646835, -5.83939138, -6.38076724, -6.78570011, -7.04929416, -7.16665354, -7.13288244, -6.943085, -6.59236541, -6.07582782, -5.38857641, -4.52571534, -3.21889646, -1.4255879, 0.26320214, 1.90882355, 3.57262624, 5.31596011, 7.20017505, 9.28662097, 11.6366478, 14.3116053, 17.3728436, 20.8817124, 24.8995618, 29.6672236, 35.0007556, 40.907663, 47.3954511, 53.9688491, 60.8867503, 68.5421215, 77.0933008, 86.6986262, 97.5164359, 109.705068, 123.753565, 139.388912, 156.704742, 175.79469, 196.752393, 219.038857, 243.538502, 270.503117, 300.184493, 332.834422, 369.169188, 408.408376, 449.992485, 493.807909, 540.651407, 590.911002, 644.974717, 704.167947, 767.728305, 833.870817, 900.123642, 969.789238, 1046.68438, 1132.06047, 1222.82202, 1317.86884, 1415.24013, 1513.61257, 1614.81842, 1718.90494, 1827.94348, 1939.64588, 2053.35877, 2168.42882, 2284.20267, 2413.80326, 2539.35687, 2656.76609, 2761.93348, 2850.76164, 2898.51243, 2935.4896, 2966.44175, 3011.08188, 3068.23518, 3130.49165, 3196.98822, 3266.86185, 3339.24946, 3413.288, 3452.24532, 3517.01193, 3598.77482, 3688.72098, 3778.03739, 3857.91105, 3919.52895, 4001.14824, 4057.53694},
        {174.307038, 173.619954, 172.859481, 172.027313, 171.125145, 170.154672, 169.117587, 168.015586, 166.850362, 165.623612, 164.337029, 162.992307, 161.591142, 160.135229, 158.62626, 157.065933, 155.455939, 153.797976, 152.093736, 150.344915, 148.553208, 146.720308, 144.84791, 142.93771, 140.991401, 139.010678, 136.997236, 134.95277, 132.878973, 130.777541, 128.650168, 126.498549, 124.324378, 122.12935, 119.927087, 117.713068, 115.482631, 113.237084, 110.977733, 108.705886, 106.422851, 104.129935, 101.828446, 99.5196913, 97.2049785, 94.8856152, 92.5629088, 90.2381669, 87.912697, 85.5878067, 83.2648035, 80.968172, 78.6755403, 76.3877134, 74.1054965, 71.8296946, 69.5611127, 67.3005562, 65.0488299, 62.806739, 60.5750887, 58.354684, 56.14633, 53.9508318, 51.7689945, 49.6016233, 47.4495231, 45.2772232, 43.0501235, 40.8438457, 38.6615465, 36.5063827, 34.381511, 32.2900881, 30.2352707, 28.2202156, 26.2480795, 24.3220192, 22.4451912, 20.6207525, 18.8518597, 17.1416694, 15.4933386, 13.9100238, 12.5808412, 11.4120565, 10.3077001, 9.26489402, 8.28076031, 7.3524209, 6.4769978, 5.65161297, 4.87338842, 4.13944612, 3.44690807, 2.79289624, 2.17453263, 1.58893921, 1.03323797, 0.504550909, 0, -2.08829274, -3.95507728, -5.60287051, -7.0341893, -8.25155055, -9.25747111, -10.0544679, -10.6450577, -11.0317576, -11.2170842, -11.2035546, -10.9936857, -10.5899941, -9.99499701, -9.21121112, -8.24115335, -6.78842862, -4.81285208, -2.99305105, -1.26748983, 0.425367248, 2.14705589, 3.95911177, 5.92307057, 8.100468, 10.5528397, 13.3417214, 16.5286488, 20.1751576, 24.386227, 29.1640198, 34.5570386, 40.6137861, 47.1102367, 54.231157, 62.233988, 71.2489906, 81.4064258, 92.8365548, 105.669638, 120.412537, 136.704295, 154.601492, 174.160705, 195.438514, 217.915269, 242.367834, 268.996845, 298.002936, 329.586743, 364.226704, 401.506118, 441.02403, 482.74961, 527.407697, 575.38385, 627.063628, 683.719534, 744.648605, 808.180304, 872.196635, 939.40697, 1012.55478, 1092.67073, 1177.8832, 1267.65275, 1361.27314, 1458.16168, 1558.34739, 1661.51326, 1768.05409, 1876.83159, 1987.28237, 2098.84305, 2210.95023, 2335.51914, 2456.38814, 2569.87419, 2672.29427, 2759.96534, 2811.1795, 2852.29516, 2887.3542, 2932.50767, 2987.00426, 3045.04722, 3106.13753, 3169.77617, 3235.46413, 3302.70238, 3380.75419, 3450.4368, 3515.29151, 3578.85963, 3644.68245, 3716.30128, 3797.2574, 3871.18625, 3933.55122},
        {178.020724, 177.379316, 176.656519, 175.854305, 174.974647, 174.019517, 172.990889, 171.890734, 170.721026, 169.483738, 168.180841, 166.814309, 165.386114, 163.898229, 162.352626, 160.751279, 159.09616, 157.389241, 155.632495, 153.827895, 151.977414, 150.083024, 148.146697, 146.170407, 144.156127, 142.105828, 140.021483, 137.905066, 135.758548, 133.583903, 131.383103, 129.158121, 126.910929, 124.643501, 122.385424, 120.124282, 117.847329, 115.555639, 113.250289, 110.932356, 108.602917, 106.263047, 103.913823, 101.556322, 99.1916207, 96.8207946, 94.4449206, 92.0650752, 89.6823348, 87.2977759, 84.912475, 82.5253121, 80.1396077, 77.7564859, 75.3770707, 73.0024863, 70.6338568, 68.2723062, 65.9189585, 63.574938, 61.2413686, 58.9193745, 56.6100797, 54.3146083, 52.0340844, 49.769632, 47.5223753, 45.2580292, 42.9431588, 40.6519169, 38.3877229, 36.1539964, 33.954157, 31.791624, 29.6698171, 27.5921558, 25.5620595, 23.5829478, 21.6582402, 19.7913561, 17.9857152, 16.2447369, 14.5718408, 12.9704463, 11.6857417, 10.5951769, 9.57329582, 8.61567217, 7.71787958, 6.87549169, 6.08408216, 5.33922465, 4.63649279, 3.97146025, 3.33970068, 2.73678772, 2.15829503, 1.59979627, 1.05686507, 0.525075098, 0, -2.71730287, -5.14550606, -7.29130362, -9.16138958, -10.762458, -12.1012028, -13.1843182, -14.0184982, -14.6104367, -14.9668278, -15.0943656, -14.9997441, -14.6896574, -14.1707994, -13.4498642, -12.5335459, -11.07893, -9.04276627, -7.22253017, -5.54999956, -3.95695237, -2.37516651, -0.736419909, 1.02750952, 2.98484384, 5.20380516, 7.75261554, 10.6994971, 14.1126718, 18.0464713, 22.5881015, 27.8099514, 33.7844104, 40.5141627, 48.10645, 56.6871021, 66.3494197, 77.1867035, 89.2922542, 102.759373, 118.071616, 134.813255, 153.001235, 172.652503, 193.784005, 215.659564, 239.23753, 264.72313, 292.321589, 322.238136, 354.946947, 390.055581, 427.12204, 466.120258, 507.873365, 552.82322, 601.411681, 654.990226, 712.884362, 773.427287, 834.818965, 899.186621, 968.252461, 1042.81735, 1122.12098, 1206.11966, 1295.14954, 1389.26542, 1487.12923, 1588.19078, 1691.85397, 1797.62157, 1904.9593, 2013.33286, 2122.20794, 2242.08348, 2358.63364, 2468.56583, 2568.58746, 2655.40593, 2710.27728, 2755.66117, 2794.88698, 2840.50685, 2892.19951, 2945.80091, 3001.18192, 3058.21342, 3116.76628, 3176.7114, 3291.17495, 3365.2969, 3413.79607, 3451.39129, 3492.8014, 3552.74523, 3645.94162, 3709.6625, 3774.95039},
        {182.127512, 181.042207, 179.925705, 178.777867, 177.598551, 176.387619, 175.144929, 173.870342, 172.563718, 171.224916, 169.853796, 168.450219, 167.014043, 165.54513, 164.043339, 162.508529, 160.940561, 159.339294, 157.704589, 156.036305, 154.334302, 152.59844, 150.828579, 149.024579, 147.186299, 145.3136, 143.406341, 141.464383, 139.487584, 137.475806, 135.428908, 133.346749, 131.22919, 129.076091, 126.807067, 124.463789, 122.08889, 119.684834, 117.254086, 114.799108, 112.322366, 109.826322, 107.313442, 104.786188, 102.247024, 99.6984159, 97.142826, 94.5827185, 92.0205575, 89.4588069, 86.8999307, 84.4366436, 81.9792022, 79.528114, 77.0838865, 74.6470271, 72.2180434, 69.7974426, 67.3857325, 64.9834203, 62.5910136, 60.2090199, 57.8379465, 55.4783011, 53.130591, 50.7953238, 48.4730068, 46.1003164, 43.6154615, 41.1505965, 38.7103668, 36.2994178, 33.9223949, 31.5839434, 29.2887088, 27.0413364, 24.8464717, 22.7087599, 20.6328465, 18.6233768, 16.6849964, 14.8223504, 13.0400844, 11.3428436, 10.0332267, 8.96063539, 7.97089025, 7.05896755, 6.21984352, 5.44849439, 4.7398964, 4.08902577, 3.49085874, 2.94037154, 2.43254041, 1.96234158, 1.52475127, 1.11474573, 0.727301179, 0.357393859, 0, 0.665886977, 1.04867918, 1.17523846, 1.07242666, 0.767105627, 0.286137221, -0.343616715, -1.09529433, -1.94203378, -2.85697321, -3.81325077, -4.78400462, -5.7423729, -6.66149376, -7.51450537, -8.27454586, -8.40996037, -7.82177378, -7.62491795, -7.68436109, -7.86507143, -8.03201719, -8.05016658, -7.78448784, -7.09994917, -5.86151879, -3.93416493, -1.18285581, 2.52744035, 8.10385892, 14.6262235, 21.9979346, 30.1223932, 38.11136, 46.2640556, 55.0908047, 64.7324998, 75.3300337, 87.024299, 99.9561884, 115.331012, 131.901291, 149.599664, 168.358768, 188.111239, 205.868123, 225.214046, 246.812045, 271.325153, 299.416406, 333.695602, 370.499637, 407.441852, 443.692791, 481.602287, 522.09274, 566.086547, 616.462458, 671.741894, 728.312074, 782.143434, 838.062228, 900.926898, 972.486851, 1049.2234, 1130.036, 1212.65905, 1295.68995, 1381.99798, 1472.036, 1570.041, 1672.09953, 1777.35457, 1884.94906, 1994.02599, 2115.24603, 2233.35501, 2344.61644, 2445.29388, 2531.65085, 2582.48341, 2623.16756, 2657.45289, 2700.81776, 2752.36913, 2806.32697, 2862.04258, 2918.86729, 2976.15238, 3033.24917, 3116.97587, 3174.09269, 3215.34169, 3251.4649, 3293.20439, 3351.30219, 3436.50036, 3495.07044, 3531.30989},
        {189.757548, 189.648749, 189.400061, 189.01538, 188.4986, 187.853615, 187.084322, 186.194614, 185.188387, 184.069535, 182.841954, 181.509538, 180.076182, 178.545781, 176.922231, 175.209425, 173.411258, 171.531627, 169.574425, 167.543547, 165.442888, 163.276344, 161.047809, 158.761178, 156.420345, 154.029206, 151.591656, 149.111589, 146.592901, 144.039486, 141.455239, 138.844055, 136.209829, 133.556456, 130.993478, 128.469744, 125.932833, 123.383212, 120.821347, 118.247705, 115.662752, 113.066955, 110.460779, 107.844692, 105.21916, 102.584648, 99.9416246, 97.2905549, 94.6319057, 91.9661433, 89.2937343, 86.5866663, 83.8745018, 81.1583246, 78.4392185, 75.7182673, 72.9965548, 70.2751646, 67.5551807, 64.8376869, 62.1237668, 59.4145043, 56.7109832, 54.0142872, 51.3255002, 48.6457059, 45.9759881, 43.2139365, 40.2596256, 37.3275881, 34.4256168, 31.5615046, 28.7430445, 25.9780292, 23.2742517, 20.6395049, 18.0815816, 15.6082748, 13.2273773, 10.9466821, 8.77398191, 6.71706977, 4.78373851, 2.98178102, 1.98445157, 1.45281526, 1.03993889, 0.732019843, 0.515255511, 0.375843278, 0.299980531, 0.273864659, 0.283693047, 0.315663083, 0.355972155, 0.390817649, 0.406396952, 0.388907451, 0.324546534, 0.199511588, 0, -4.64903933, -8.83734084, -12.5799476, -15.8919027, -18.7882493, -21.2840304, -23.3942892, -25.1340687, -26.518412, -27.5623622, -28.2809625, -28.6892558, -28.8022854, -28.6350942, -28.2027254, -27.5202221, -26.0326385, -23.6735907, -21.7473824, -20.146916, -18.7650944, -17.49482, -16.2289955, -14.8605236, -13.2823068, -11.3872479, -9.06824935, -6.21821386, -2.73004405, 1.93412659, 7.29267761, 13.3234756, 20.0043873, 26.998163, 34.4402273, 42.5500362, 51.3999913, 61.062494, 71.6099459, 83.1147485, 95.6462289, 109.280799, 124.091461, 140.15122, 157.533076, 178.533713, 200.446535, 222.788625, 245.077068, 266.828947, 285.141836, 304.909506, 329.379907, 359.918448, 394.047011, 431.013144, 470.064393, 508.407677, 547.794951, 592.204314, 645.73879, 704.237779, 764.625893, 825.512241, 888.360222, 953.323067, 1020.40409, 1089.71766, 1161.92784, 1237.38776, 1318.06813, 1402.45563, 1490.34348, 1581.52487, 1675.793, 1787.35325, 1897.98359, 2003.87419, 2101.21519, 2186.19675, 2240.04693, 2283.89271, 2320.33657, 2355.85554, 2391.51843, 2426.77965, 2462.78875, 2500.69526, 2541.6487, 2586.79862, 2665.46258, 2738.03981, 2804.1781, 2863.52523, 2915.72898, 2960.43713, 2997.29746, 3107.20523, 3128.99753},
        {192.62829, 192.49083, 192.217802, 191.812901, 191.279817, 190.622246, 189.84388, 188.948411, 187.939534, 186.82094, 185.596325, 184.269379, 182.843797, 181.323272, 179.711497, 178.012165, 176.228968, 174.365601, 172.425756, 170.413127, 168.331406, 166.184286, 163.975461, 161.708625, 159.387469, 157.015687, 154.596973, 152.135018, 149.633518, 147.096164, 144.526649, 141.928667, 139.305911, 136.662075, 134.097805, 131.566278, 129.019506, 126.458035, 123.882412, 121.293184, 118.690897, 116.076098, 113.449334, 110.811152, 108.162098, 105.50272, 102.833563, 100.155175, 97.4681019, 94.7728912, 92.0700894, 89.3380613, 86.6000164, 83.8569823, 81.1099866, 78.360057, 75.608221, 72.8555063, 70.1029405, 67.3515511, 64.6023658, 61.8564121, 59.1147178, 56.3783103, 53.6482174, 50.9254665, 48.2110854, 45.4002746, 42.3907758, 39.4018766, 36.441465, 33.5174289, 30.6376562, 27.8100349, 25.0424529, 22.3427982, 19.7189587, 17.1788223, 14.730277, 12.3812106, 10.1395112, 8.01306666, 6.00976492, 4.13749392, 3.13154623, 2.62080748, 2.22542047, 1.92966748, 1.71783075, 1.57419253, 1.4830351, 1.4286407, 1.39529159, 1.36727002, 1.32885826, 1.26433857, 1.15799319, 0.994104386, 0.756954416, 0.430825535, 0, -6.87835047, -13.0519482, -18.5432954, -23.3748943, -27.5692472, -31.1488564, -34.1362239, -36.5538522, -38.4242434, -39.7698997, -40.6133235, -40.9770169, -40.8834822, -40.3552216, -39.4147374, -38.0845318, -36.2288395, -33.8475529, -31.3211606, -28.6382505, -25.7874102, -22.7572275, -19.5362902, -16.1131859, -12.4765025, -8.6148276, -4.51674907, -0.170854594, 4.43426806, 9.07939199, 14.0911363, 19.550105, 25.5369022, 32.3650855, 39.9987822, 48.3399983, 57.3994517, 67.1878602, 77.7159418, 88.9944142, 100.686655, 113.256435, 126.78243, 141.343314, 157.017765, 175.268931, 194.444896, 214.278217, 234.501451, 254.847153, 273.612385, 293.719698, 317.003128, 344.279395, 374.2882, 406.701783, 441.192383, 476.118868, 512.765344, 553.84869, 602.574171, 655.546252, 709.617433, 763.660818, 819.973313, 879.280243, 942.931593, 1011.81464, 1084.52623, 1160.95883, 1240.86171, 1324.29251, 1411.19325, 1501.50591, 1595.17252, 1708.53075, 1821.028, 1928.50738, 2026.81198, 2111.78488, 2162.58609, 2202.86383, 2235.61108, 2270.37211, 2307.54666, 2344.80483, 2382.68285, 2421.71698, 2462.44348, 2505.39859, 2587.15934, 2649.29956, 2697.25676, 2736.46849, 2772.37227, 2810.40563, 2856.00612, 2913.89874, 2963.61026},
        {202.644418, 202.43914, 202.104309, 201.642974, 201.058186, 200.352995, 199.530452, 198.593605, 197.545507, 196.389205, 195.127752, 193.764197, 192.30159, 190.742981, 189.091421, 187.34996, 185.521647, 183.609534, 181.61667, 179.546105, 177.40089, 175.184075, 172.89871, 170.547845, 168.13453, 165.661816, 163.132752, 160.550389, 157.917778, 155.237967, 152.514008, 149.74895, 146.945844, 144.10774, 141.296455, 138.48442, 135.643358, 132.774413, 129.878729, 126.957447, 124.011712, 121.042666, 118.051453, 115.039215, 112.007096, 108.956239, 105.887786, 102.802882, 99.7026678, 96.5882881, 93.4608855, 90.295565, 87.1200723, 83.9361149, 80.7454003, 77.549636, 74.3505297, 71.1497888, 67.9491209, 64.7502335, 61.5548342, 58.3646304, 55.1813298, 52.0066399, 48.8422681, 45.6899222, 42.5513095, 39.2822013, 35.7418729, 32.2330144, 28.7667936, 25.3543787, 22.0069376, 18.7356383, 15.5516488, 12.4661371, 9.49027115, 6.63521898, 3.91214858, 1.33222792, -1.09337499, -3.35349217, -5.43695561, -7.33259734, -8.05386084, -8.09778835, -7.97314556, -7.7004177, -7.30009003, -6.79264778, -6.19857619, -5.5383605, -4.83248597, -4.10143782, -3.3657013, -2.64576166, -1.96210413, -1.33521396, -0.785576384, -0.333676651, 0, -5.49097175, -10.4293973, -14.8328726, -18.7189938, -22.1053569, -25.0095581, -27.4491933, -29.4418586, -31.0051502, -32.1566639, -32.913996, -33.2947425, -33.3164994, -32.9968628, -32.3534288, -31.4037933, -29.996337, -28.1244822, -26.1885745, -24.1699495, -22.0499427, -19.8098898, -17.4311263, -14.8949877, -12.1828098, -9.27592804, -6.155678, -2.8033953, 0.799584505, 4.76257641, 8.9803557, 13.4443916, 18.1461534, 22.88521, 27.7489809, 32.8760588, 38.3154829, 44.1162927, 50.3275274, 56.9982263, 64.1202615, 71.8172382, 80.1493807, 89.1769132, 98.9600599, 110.693127, 123.018735, 135.71359, 148.554395, 161.317853, 172.569519, 184.777539, 199.806525, 218.382888, 238.909212, 260.831434, 283.595489, 305.147597, 326.774255, 351.39801, 381.687067, 415.324382, 450.873075, 487.666503, 526.343153, 566.942433, 609.018087, 652.483614, 699.033286, 749.354071, 806.135912, 867.754647, 934.2039, 1005.4773, 1081.56847, 1179.16534, 1277.39366, 1372.07348, 1459.02485, 1534.06783, 1576.36727, 1609.50187, 1636.42963, 1667.55378, 1702.88524, 1739.33272, 1777.06224, 1816.23981, 1857.03145, 1899.60317, 2019.13008, 2086.42187, 2118.48724, 2132.33492, 2144.9736, 2173.41199, 2234.65881, 2303.53122, 2315.11357}
    };
    
    const double a4tab[98][200] = {{-1766.7572, -1746.59454, -1726.56578, -1706.66718, -1686.89499, -1667.24545, -1647.71482, -1628.29933, -1608.99525, -1589.79882, -1570.70629, -1551.71391, -1532.81792, -1514.01458, -1495.30014, -1476.67084, -1458.12293, -1439.65266, -1421.25629, -1402.93006, -1384.67021, -1366.473, -1348.33468, -1330.2515, -1312.2197, -1294.23553, -1276.29525, -1258.3951, -1240.53133, -1222.70019, -1204.89792, -1187.12079, -1169.36503, -1151.62689, -1133.74537, -1115.79865, -1097.8668, -1079.95118, -1062.05315, -1044.17405, -1026.31524, -1008.47807, -990.663897, -972.874065, -955.10993, -937.372845, -919.664161, -901.985233, -884.337411, -866.722049, -849.140499, -831.642108, -814.179193, -796.752066, -779.36104, -762.006426, -744.688536, -727.407682, -710.164177, -692.958333, -675.79046, -658.660872, -641.56988, -624.517796, -607.504933, -590.531602, -573.598115, -556.784068, -540.167153, -523.584166, -507.030278, -490.500662, -473.990491, -457.494937, -441.009172, -424.528369, -408.0477, -391.562337, -375.067453, -358.55822, -342.029811, -325.477397, -308.896152, -292.281247, -275.414886, -258.403205, -241.354901, -224.272057, -207.156756, -190.011083, -172.83712, -155.636953, -138.412663, -121.166336, -103.900054, -86.6159006, -69.3159605, -52.0023167, -34.6770529, -17.3422528, 0, 15.5874605, 31.2260846, 46.9499571, 62.7931629, 78.7897866, 94.973913, 111.379627, 128.041013, 144.992156, 162.267141, 179.900052, 197.924975, 216.375994, 235.287193, 254.692659, 274.626474, 295.552045, 317.564787, 339.727705, 362.166879, 385.008392, 408.378325, 432.402762, 457.207783, 482.91947, 509.663906, 537.567173, 566.755352, 597.354525, 629.677491, 663.595153, 699.177577, 736.494832, 774.577317, 814.014934, 855.674826, 899.93896, 947.189304, 997.807825, 1052.17649, 1111.90581, 1175.77531, 1243.92657, 1316.50121, 1393.64081, 1472.96419, 1557.76644, 1648.81985, 1746.89671, 1852.76932, 1967.16894, 2090.95905, 2222.38732, 2361.96984, 2514.8541, 2684.10817, 2872.80013, 3089.68725, 3330.85544, 3586.18425, 3841.24649, 4117.09438, 4433.27118, 4798.65112, 5204.0304, 5648.50333, 6123.68137, 6626.71883, 7182.20718, 7799.21788, 8485.94375, 9252.47006, 10108.1724, 11062.4264, 12124.6076, 13681.8839, 15271.3905, 16808.055, 18206.8049, 19382.5677, 20091.1032, 20512.6184, 20630.2556, 20190.7467, 19270.8129, 18045.6799, 16587.1438, 14967.0007, 13257.0468, 11529.078, 10020.3607, 8629.63832, 7353.5544, 6188.75257, 5131.87645, 4179.56964, 3328.47574, 2598.30101, 1896.87465},
        {-1661.23266, -1640.96932, -1620.97775, -1601.24897, -1581.77402, -1562.54394, -1543.54977, -1524.78253, -1506.23326, -1487.89301, -1469.75279, -1451.80366, -1434.03664, -1416.44278, -1399.01309, -1381.73864, -1364.61043, -1347.61952, -1330.75694, -1314.01372, -1297.3809, -1280.84951, -1264.4106, -1248.05518, -1231.77431, -1215.55902, -1199.40033, -1183.2893, -1167.21694, -1151.1743, -1135.15242, -1119.14232, -1103.13505, -1087.12164, -1070.73828, -1054.1609, -1037.56968, -1020.96717, -1004.35591, -987.738468, -971.117383, -954.495208, -937.874494, -921.257792, -904.647651, -888.046623, -871.457258, -854.882107, -838.323721, -821.784649, -805.267442, -788.880905, -772.519031, -756.182067, -739.870261, -723.58386, -707.32311, -691.08826, -674.879556, -658.697245, -642.541574, -626.412792, -610.311145, -594.236879, -578.190243, -562.171484, -546.180848, -530.288547, -514.563114, -498.860499, -483.176413, -467.506569, -451.846677, -436.192451, -420.5396, -404.883837, -389.220874, -373.546423, -357.856194, -342.1459, -326.411252, -310.647962, -294.851742, -279.018303, -262.912158, -246.649482, -230.349226, -214.014606, -197.648836, -181.255131, -164.836705, -148.396772, -131.938549, -115.465248, -98.9800839, -82.4862721, -65.9870267, -49.4855623, -32.9850933, -16.4888344, 0, 14.8569198, 29.7147625, 44.6096261, 59.5776086, 74.6548082, 89.8773229, 105.281251, 120.90269, 136.777738, 152.942494, 169.433055, 186.285519, 203.535985, 221.220551, 239.375314, 258.036373, 277.452256, 297.68941, 318.303434, 339.375945, 360.988564, 383.222909, 406.1606, 429.883255, 454.472493, 480.009934, 506.577196, 534.255899, 563.127661, 593.173693, 624.612839, 657.556841, 692.117441, 727.790178, 764.994896, 804.315757, 846.049363, 890.492315, 937.941216, 988.692668, 1043.21002, 1101.57238, 1164.04372, 1230.88802, 1302.36926, 1379.75529, 1462.05525, 1549.28215, 1641.44899, 1738.5688, 1839.69945, 1946.97645, 2056.64946, 2168.87932, 2291.81232, 2430.00943, 2588.03164, 2780.69681, 2999.97795, 3226.65873, 3440.43982, 3663.9879, 3920.42426, 4217.46043, 4536.69692, 4871.71963, 5206.52967, 5532.228, 5875.0598, 6241.3907, 6649.49656, 7092.00043, 7571.14516, 8089.17358, 8648.32856, 9255.30047, 9906.77275, 10603.8764, 11347.7423, 12139.5014, 13168.1639, 14121.7288, 14920.8076, 15102.6546, 14739.6266, 14063.8127, 13139.5829, 12031.3076, 10803.357, 9520.10122, 8369.80381, 7294.9199, 6295.895, 5373.17462, 4527.20425, 3758.42941, 3067.2956, 2491.81541, 1969.09585},
        {-1563.20972, -1545.28891, -1527.51745, -1509.89087, -1492.40469, -1475.05445, -1457.83566, -1440.74385, -1423.77455, -1406.92328, -1390.18557, -1373.55695, -1357.03293, -1340.60906, -1324.28084, -1308.04382, -1291.8935, -1275.82543, -1259.83512, -1243.91811, -1228.06991, -1212.28605, -1196.56206, -1180.89347, -1165.27579, -1149.70456, -1134.17531, -1118.68355, -1103.22481, -1087.79463, -1072.38851, -1057.002, -1041.63062, -1026.26988, -1010.7578, -995.171968, -979.591885, -964.018191, -948.451522, -932.892516, -917.341812, -901.800047, -886.267859, -870.745886, -855.234766, -839.735137, -824.247636, -808.772903, -793.311573, -777.864287, -762.43168, -746.988135, -731.561115, -716.151828, -700.76148, -685.391279, -670.042432, -654.716146, -639.413627, -624.136084, -608.884724, -593.660752, -578.465377, -563.299806, -548.165245, -533.062902, -517.993984, -503.07351, -488.413768, -473.781234, -459.169738, -444.57311, -429.985177, -415.39977, -400.810718, -386.211849, -371.596993, -356.959979, -342.294636, -327.594793, -312.85428, -298.066925, -283.226557, -268.327006, -252.994623, -237.414705, -221.776968, -206.087164, -190.351048, -174.574376, -158.762901, -142.922379, -127.058562, -111.177207, -95.2840667, -79.3848965, -63.4854506, -47.5914835, -31.7087498, -15.8430038, 0, 14.8505086, 29.5312302, 44.0893059, 58.5718764, 73.0260825, 87.4990651, 102.037965, 116.689923, 131.50208, 146.521577, 161.795555, 177.371154, 193.295515, 209.61578, 226.379088, 243.632582, 261.39395, 279.706127, 298.682868, 318.365003, 338.793362, 360.008775, 382.052073, 404.964084, 428.785639, 453.557568, 479.3207, 506.115867, 533.983897, 562.63634, 592.564042, 623.90662, 656.803686, 691.447945, 727.952467, 766.416164, 806.962724, 849.715834, 894.799182, 942.336455, 990.198043, 1041.44672, 1096.64703, 1156.36353, 1221.16077, 1298.43436, 1380.21002, 1465.34454, 1552.69471, 1641.1173, 1724.8122, 1812.98489, 1908.67802, 2013.97314, 2128.98948, 2254.72731, 2392.1869, 2544.54794, 2710.13595, 2884.89886, 3063.00941, 3252.62597, 3467.07994, 3709.20491, 3964.04645, 4224.81512, 4482.57319, 4729.97426, 4967.54896, 5191.37224, 5391.45539, 5570.86986, 5727.78953, 5860.38829, 5966.84003, 5745.49967, 5569.3148, 5511.41406, 5644.92607, 6042.97947, 7242.7316, 8543.92989, 9820.83353, 10473.9987, 10565.192, 10353.5562, 9890.98934, 9229.38905, 8420.65319, 7516.67954, 6644.99579, 5777.79864, 4933.24222, 4129.48061, 3384.66795, 2716.95833, 2144.50586, 1627.53437, 1184.91263},
        {-1464.60418, -1448.64276, -1432.76005, -1416.95439, -1401.22412, -1385.56758, -1369.98311, -1354.46905, -1339.02374, -1323.64551, -1308.3327, -1293.08366, -1277.89672, -1262.77022, -1247.70251, -1232.69191, -1217.73676, -1202.83542, -1187.98621, -1173.18748, -1158.43755, -1143.73479, -1129.07751, -1114.46406, -1099.89278, -1085.36201, -1070.87009, -1056.41535, -1041.99614, -1027.61079, -1013.25764, -998.935039, -984.641314, -970.374809, -956.174813, -942.018327, -927.881863, -913.76243, -899.657038, -885.562697, -871.476417, -857.395209, -843.316082, -829.236046, -815.152111, -801.061287, -786.960585, -772.847014, -758.717584, -744.569305, -730.399187, -715.963577, -701.505366, -687.026779, -672.530045, -658.017391, -643.491044, -628.95323, -614.406178, -599.852113, -585.293264, -570.731857, -556.170119, -541.610278, -527.05456, -512.505194, -497.964405, -483.516331, -469.243145, -454.978137, -440.718226, -426.460329, -412.201363, -397.938244, -383.667891, -369.38722, -355.093149, -340.782595, -326.452475, -312.099705, -297.721204, -283.313889, -268.874676, -254.400484, -239.669857, -224.793492, -209.88471, -194.947514, -179.985908, -165.003895, -150.00548, -134.994666, -119.975456, -104.951855, -89.9278661, -74.9074933, -59.8947403, -44.8936107, -29.9081083, -14.9422368, 0, 13.7319382, 27.3661114, 40.942923, 54.5027762, 68.086074, 81.7332199, 95.4846169, 109.380668, 123.461777, 137.768347, 152.340781, 167.219482, 182.444853, 198.057299, 214.097221, 230.605023, 247.847818, 265.898395, 284.284364, 303.094708, 322.41841, 342.344456, 362.961827, 384.359509, 406.626484, 429.851737, 454.124252, 479.533012, 506.167001, 534.32142, 563.803423, 594.640127, 626.858652, 659.771479, 693.763044, 729.408355, 766.948921, 806.626252, 848.681857, 893.357247, 941.143216, 991.956119, 1045.98869, 1103.43367, 1164.48379, 1230.50685, 1300.22676, 1373.5425, 1450.35302, 1530.55731, 1613.08508, 1699.9892, 1792.72752, 1891.94659, 1996.6366, 2106.53141, 2221.36488, 2338.01489, 2459.72037, 2592.83587, 2745.89213, 2909.56452, 3081.33939, 3257.47066, 3432.124, 3600.42359, 3760.69897, 3908.90535, 4029.24508, 4112.56858, 4165.15931, 4160.06038, 4082.78001, 3918.82644, 3653.70791, 2627.74217, 1632.92556, 816.063929, 323.96313, 303.429019, 1648.95758, 3261.20445, 4966.53686, 6002.07698, 6427.43489, 6547.7394, 6410.32463, 6062.5247, 5551.67373, 4925.10585, 4304.14647, 3651.03516, 2987.21362, 2334.12358, 1713.20674, 1145.90481, 653.659527, 180.949463, -226.115744},
        {-1373.5761, -1360.68588, -1347.72756, -1334.70411, -1321.61851, -1308.47374, -1295.27277, -1282.01859, -1268.71417, -1255.36249, -1241.96653, -1228.52926, -1215.05366, -1201.54271, -1187.99939, -1174.42667, -1160.82753, -1147.20496, -1133.56191, -1119.90139, -1106.22635, -1092.53979, -1078.84467, -1065.14398, -1051.44069, -1037.73777, -1024.03822, -1010.345, -996.661091, -982.989473, -969.333122, -955.695016, -942.078133, -928.485449, -915.161209, -901.982682, -888.824239, -875.681027, -862.548194, -849.420888, -836.294258, -823.16345, -810.023614, -796.869898, -783.697448, -770.501414, -757.276942, -744.019182, -730.723281, -717.384388, -703.997649, -690.168774, -676.290793, -662.367295, -648.401871, -634.398112, -620.359607, -606.289946, -592.192721, -578.071521, -563.929936, -549.771557, -535.599975, -521.418778, -507.231558, -493.041905, -478.85341, -464.745523, -450.795875, -436.851598, -422.911365, -408.973847, -395.037718, -381.101649, -367.164314, -353.224385, -339.280533, -325.331433, -311.375755, -297.412173, -283.43936, -269.455987, -255.460727, -241.452252, -227.306741, -213.086691, -198.856068, -184.617522, -170.373699, -156.127247, -141.880814, -127.637048, -113.398595, -99.1681045, -84.9482232, -70.7415988, -56.5508792, -42.3787118, -28.2277444, -14.1006246, 0, 12.2170632, 24.4832114, 36.827677, 49.2796927, 61.8684911, 74.6233046, 87.573366, 100.747908, 114.176162, 127.887363, 141.910741, 156.27553, 171.010962, 186.14627, 201.710687, 217.733444, 234.651903, 252.553601, 270.544622, 288.741656, 307.261391, 326.220516, 345.735719, 365.923688, 386.901114, 408.784683, 431.691085, 455.737009, 481.039143, 508.173191, 536.62852, 566.384116, 597.41896, 628.650595, 660.588725, 694.026108, 729.260162, 766.588303, 806.307948, 848.716516, 896.336869, 946.563669, 999.258919, 1054.28462, 1111.50278, 1166.97415, 1225.3123, 1287.32953, 1353.83818, 1425.65056, 1505.86857, 1590.21655, 1676.04437, 1762.42284, 1851.93634, 1945.59172, 2044.39585, 2151.79585, 2265.8037, 2381.76928, 2498.24025, 2613.12658, 2728.98687, 2843.52984, 2946.24639, 3030.39954, 3092.56874, 3126.87681, 3115.28638, 3046.63856, 2973.49309, 2811.16962, 2526.45282, 2086.12738, 1456.97797, -455.640717, -2358.15651, -4018.42723, -5204.31069, -5683.66473, -4137.80084, -2145.48737, 59.755175, 1488.21783, 2205.70467, 2593.09826, 2700.44875, 2577.80628, 2275.22098, 1842.743, 1383.10257, 907.74269, 426.300494, -51.5868882, -516.282327, -958.148691, -1367.54885, -1741.14511, -2092.20259},
        {-1295.44714, -1286.9622, -1278.14521, -1269.00696, -1259.55822, -1249.80977, -1239.77239, -1229.45687, -1218.87398, -1208.0345, -1196.94923, -1185.62892, -1174.08438, -1162.32637, -1150.36569, -1138.2131, -1125.87939, -1113.37534, -1100.71174, -1087.89936, -1074.94897, -1061.87138, -1048.67734, -1035.37766, -1021.98309, -1008.50444, -994.952467, -981.337965, -967.671712, -953.964487, -940.227072, -926.470247, -912.704793, -898.941492, -885.628885, -872.549665, -859.481262, -846.420251, -833.363204, -820.306698, -807.247306, -794.181603, -781.106164, -768.017562, -754.912372, -741.787169, -728.638527, -715.46302, -702.257224, -689.017711, -675.741058, -662.018918, -648.261563, -634.474346, -620.662619, -606.831735, -592.987045, -579.133902, -565.277659, -551.423668, -537.57728, -523.743849, -509.928727, -496.137265, -482.374817, -468.646734, -454.95837, -441.486345, -428.403171, -415.360968, -402.353986, -389.376474, -376.422682, -363.486859, -350.563256, -337.646121, -324.729704, -311.808254, -298.876022, -285.927257, -272.956208, -259.957125, -246.924258, -233.851855, -220.377137, -206.680378, -192.946144, -179.180269, -165.38859, -151.576944, -137.751165, -123.91709, -110.080554, -96.2473942, -82.4234455, -68.6145439, -54.8265254, -41.0652259, -27.3364813, -13.6461273, 0, 11.5903495, 23.1944077, 34.8444744, 46.5728497, 58.4118335, 70.3937257, 82.5508262, 94.9154349, 107.519852, 120.396377, 133.57731, 147.094951, 160.9816, 175.269556, 189.991121, 205.178593, 220.921496, 237.260306, 254.097887, 271.478802, 289.447611, 308.048879, 327.327166, 347.327036, 368.09305, 389.66977, 412.101759, 435.43358, 459.709793, 484.722598, 510.861454, 538.246632, 566.998403, 597.560455, 629.891349, 663.863408, 699.499877, 736.824004, 775.859036, 816.62822, 858.314577, 902.0373, 947.98403, 996.342404, 1047.30006, 1103.14662, 1161.44224, 1221.84908, 1284.02927, 1347.64496, 1410.79403, 1476.61477, 1544.84042, 1616.06799, 1692.65807, 1776.17944, 1868.20088, 1981.27795, 2103.49565, 2210.95335, 2277.63902, 2326.55417, 2383.03849, 2452.86794, 2508.66889, 2538.28391, 2532.51696, 2479.97838, 2358.42009, 2151.73613, 1949.04892, 1612.83518, 1090.56376, 329.703502, -722.276739, -3646.11166, -6584.07797, -9206.65595, -11184.3258, -12187.5679, -10411.399, -7985.40514, -5212.40798, -3342.65843, -2303.95418, -1629.3308, -1266.32411, -1162.46993, -1265.30405, -1522.3623, -1900.20494, -2265.40003, -2617.91446, -2957.71511, -3284.76887, -3599.04261, -3900.50323, -4131.64011, -4373.70656},
        {-1218.35287, -1210.79363, -1202.95486, -1194.8443, -1186.46968, -1177.83873, -1168.95919, -1159.83879, -1150.48526, -1140.90633, -1131.10974, -1121.10322, -1110.8945, -1100.49132, -1089.9014, -1079.13248, -1068.19229, -1057.08857, -1045.82904, -1034.42145, -1022.87351, -1011.19298, -999.387565, -987.465014, -975.433057, -963.299425, -951.071852, -938.758069, -926.365809, -913.902805, -901.37679, -888.795496, -876.166655, -863.498001, -851.060942, -838.725827, -826.359834, -813.962141, -801.531921, -789.068352, -776.570609, -764.037868, -751.469304, -738.864094, -726.221413, -713.540437, -700.820342, -688.060303, -675.259497, -662.4171, -649.532286, -636.278337, -622.987389, -609.665682, -596.319456, -582.954953, -569.578412, -556.196075, -542.814181, -529.438971, -516.076687, -502.733568, -489.415855, -476.129789, -462.881609, -449.677558, -436.523875, -423.639467, -411.238135, -398.887511, -386.58005, -374.308206, -362.064433, -349.841185, -337.630917, -325.426082, -313.219136, -301.002533, -288.768726, -276.51017, -264.219319, -251.888628, -239.510551, -227.077542, -214.106778, -200.83835, -187.518058, -174.153781, -160.753397, -147.324784, -133.87582, -120.414382, -106.948349, -93.4855987, -80.0340085, -66.6014566, -53.1958211, -39.8249797, -26.4968104, -13.2191912, 0, 11.3459487, 22.6296822, 33.8880574, 45.1579314, 56.4761613, 67.8796039, 79.4051165, 91.0895559, 102.969779, 115.082644, 127.465006, 140.153723, 153.185652, 166.59765, 180.426575, 194.709282, 209.289991, 224.178039, 239.846012, 256.289489, 273.504046, 291.485261, 310.228711, 329.729974, 349.984627, 370.988247, 392.736411, 415.224697, 438.448683, 461.702515, 485.940392, 511.36832, 538.192305, 567.687104, 599.524349, 633.090668, 668.271442, 704.952051, 743.017875, 782.354295, 820.792765, 860.8977, 902.956334, 947.255904, 994.083645, 1047.66584, 1103.36591, 1160.48634, 1218.32959, 1276.19816, 1331.41588, 1387.68217, 1444.70206, 1502.99526, 1564.74483, 1631.38699, 1704.358, 1795.47795, 1893.43926, 1975.60649, 2020.47381, 2043.03721, 2061.13483, 2076.83422, 2065.48062, 2014.68524, 1918.46596, 1766.09497, 1523.35339, 1169.31021, 845.435077, 343.411627, -413.906754, -1503.66668, -3003.01476, -7041.83271, -11131.3483, -14835.5243, -17718.3235, -19343.7089, -17386.0052, -14558.572, -11235.2168, -8927.9098, -7559.93602, -6580.34611, -5936.13681, -5574.30486, -5441.84701, -5485.75998, -5685.17929, -5886.99515, -6090.10267, -6293.39695, -6495.77312, -6696.12627, -6893.35153, -7010.901, -7097.41562},
        {-1139.48068, -1129.12466, -1118.87162, -1108.71535, -1098.64965, -1088.6683, -1078.7651, -1068.93384, -1059.16831, -1049.46232, -1039.80965, -1030.20409, -1020.63943, -1011.10948, -1001.60802, -992.128851, -982.665757, -973.212535, -963.762979, -954.31088, -944.850033, -935.374229, -925.877262, -916.352926, -906.795012, -897.197313, -887.553624, -877.857736, -868.103443, -858.284538, -848.394814, -838.428063, -828.378078, -818.238654, -807.76608, -797.077898, -786.294499, -775.417385, -764.448055, -753.388011, -742.238752, -731.001779, -719.678592, -708.270691, -696.779577, -685.206751, -673.553711, -661.82196, -650.012997, -638.128322, -626.169436, -613.93677, -601.637252, -589.276741, -576.861097, -564.396178, -551.887845, -539.341955, -526.764369, -514.160944, -501.537541, -488.900019, -476.254236, -463.606051, -450.961325, -438.325916, -425.705682, -413.264694, -401.163219, -389.080823, -377.013109, -364.95568, -352.904139, -340.854089, -328.801134, -316.740875, -304.668917, -292.580862, -280.472314, -268.338875, -256.176148, -243.979738, -231.745246, -219.468276, -206.787732, -193.885069, -180.94603, -167.977793, -154.987538, -141.982441, -128.969683, -115.956442, -102.949896, -89.9572242, -76.985605, -64.0422172, -51.1342393, -38.26885, -25.453228, -12.6945517, 0, 11.1156629, 22.1283913, 33.0753746, 43.9938022, 54.9208636, 65.8937482, 76.9496453, 88.1257446, 99.4592353, 110.987307, 122.747149, 134.775951, 147.110902, 159.789191, 172.848009, 186.324544, 200.119179, 214.249558, 229.062318, 244.565331, 260.766471, 277.673612, 295.294627, 313.637389, 332.709773, 352.519651, 373.074898, 394.383386, 416.452989, 438.788645, 462.085573, 486.502527, 512.198263, 539.8121, 569.26251, 600.339813, 633.058593, 667.433436, 703.478924, 741.209643, 781.052073, 822.483543, 865.438048, 909.849584, 955.652147, 1001.20959, 1048.41858, 1097.60567, 1149.09737, 1203.22022, 1263.15759, 1322.88749, 1378.60381, 1428.48096, 1474.73695, 1517.77427, 1557.99545, 1594.68085, 1629.61015, 1665.78716, 1720.5243, 1761.83581, 1764.21893, 1712.2103, 1607.35777, 1443.40075, 1227.93566, 958.294463, 581.000186, 71.316261, -366.711757, -1021.7998, -1997.87824, -3398.87745, -5328.72781, -10446.4597, -15662.1285, -20440.8896, -24247.8982, -26548.3099, -24551.5771, -21482.3598, -17772.5431, -15139.5095, -13506.3014, -12260.3372, -11351.4405, -10729.4349, -10344.144, -10145.3913, -10096.0479, -10080.5049, -10094.4472, -10133.5595, -10193.5268, -10270.0338, -10358.7656, -10381.0559, -10270.3949},
        {-1068.67232, -1058.40742, -1048.34206, -1038.46617, -1028.7697, -1019.24258, -1009.87476, -1000.65617, -991.576764, -982.626473, -973.795239, -965.073002, -956.449703, -947.915282, -939.45968, -931.072837, -922.744692, -914.465188, -906.224262, -898.011857, -889.817912, -881.632368, -873.445165, -865.246243, -857.025542, -848.773003, -840.478567, -832.132173, -823.723762, -815.243273, -806.680649, -798.025828, -789.26875, -780.399358, -771.015568, -761.311577, -751.486295, -741.542384, -731.482505, -721.309321, -711.025494, -700.633687, -690.13656, -679.536777, -668.837, -658.03989, -647.148111, -636.164324, -625.091191, -613.931374, -602.687536, -591.250725, -579.737636, -568.153351, -556.502953, -544.791522, -533.024141, -521.205891, -509.341855, -497.437113, -485.496748, -473.525841, -461.529474, -449.51273, -437.480689, -425.438433, -413.391044, -401.46692, -389.791498, -378.115529, -366.436101, -354.750303, -343.055221, -331.347944, -319.625559, -307.885154, -296.123816, -284.338634, -272.526694, -260.685086, -248.810895, -236.901211, -224.95312, -212.963711, -200.605075, -188.043632, -175.449712, -162.830949, -150.194978, -137.549432, -124.901947, -112.260157, -99.6316956, -87.0241984, -74.4452995, -61.9026333, -49.4038343, -36.956537, -24.5683759, -12.2469854, 0, 10.5526227, 21.010986, 31.4104946, 41.7865532, 52.1745666, 62.6099394, 73.1280765, 83.7643826, 94.5542623, 105.53312, 116.736362, 128.199391, 139.957613, 152.046432, 164.501253, 177.357481, 190.514695, 203.988897, 218.122715, 232.922448, 248.394395, 264.544857, 281.380132, 298.906519, 317.130319, 336.057829, 355.69535, 376.049181, 397.125621, 418.52582, 440.809782, 464.10535, 488.540367, 514.388966, 541.705848, 570.5067, 600.87548, 632.896144, 666.65265, 702.228954, 741.213649, 781.728124, 823.561951, 866.5047, 910.345944, 953.159772, 996.880109, 1041.7254, 1087.91408, 1135.66459, 1187.7515, 1238.71298, 1287.43595, 1332.47615, 1371.7133, 1403.33064, 1425.51141, 1427.52983, 1418.50294, 1417.26674, 1461.68315, 1495.45496, 1471.58326, 1365.60887, 1191.26497, 944.753834, 641.312509, 282.07821, -201.606097, -838.917874, -1402.39128, -2217.33067, -3401.75156, -5073.66948, -7351.09996, -13165.6048, -19118.2667, -24623.7146, -29096.5775, -31951.4842, -30187.0645, -27244.6127, -23574.186, -20941.0446, -19274.6445, -17956.4404, -16943.2885, -16192.0448, -15659.5653, -15302.7063, -15270.6821, -15161.2774, -15001.8821, -14819.886, -14642.6791, -14497.6512, -14412.1924, -14221.0705, -13880.1115},
        {-1003.72827, -996.531557, -989.327998, -982.11475, -974.888964, -967.647793, -960.38839, -953.107908, -945.803499, -938.472316, -931.111511, -923.718238, -916.289649, -908.822896, -901.315132, -893.763511, -886.165184, -878.517304, -870.817024, -863.061497, -855.247875, -847.373311, -839.434957, -831.429967, -823.355492, -815.208686, -806.986702, -798.686691, -790.305807, -781.841202, -773.290029, -764.64944, -755.916589, -747.088627, -738.036852, -728.82399, -719.514284, -710.10897, -700.609284, -691.016465, -681.331748, -671.556371, -661.691571, -651.738584, -641.698648, -631.573, -621.362877, -611.069515, -600.694151, -590.238023, -579.702367, -568.980611, -558.184139, -547.316524, -536.381341, -525.382163, -514.322565, -503.20612, -492.036403, -480.816988, -469.551449, -458.24336, -446.896294, -435.513826, -424.099531, -412.656981, -401.189752, -389.775181, -378.488835, -367.182156, -355.853935, -344.502966, -333.12804, -321.727948, -310.301485, -298.847441, -287.364609, -275.851782, -264.307751, -252.731308, -241.121246, -229.476358, -217.795434, -206.077268, -194.039493, -181.827394, -169.589636, -157.334136, -145.068809, -132.801572, -120.540342, -108.293035, -96.0675669, -83.8718546, -71.7138142, -59.6013622, -47.5424149, -35.5448886, -23.6166998, -11.7657648, 0, 9.28966353, 18.6159067, 28.004897, 37.4828019, 47.0757888, 56.8100253, 66.7116788, 76.8069168, 87.1219067, 97.6828161, 108.515812, 119.647063, 131.102736, 142.908997, 155.092016, 167.677959, 180.680569, 194.12424, 208.063241, 222.521077, 237.521252, 253.087273, 269.242645, 286.010872, 303.41546, 321.479914, 340.227739, 359.682441, 379.867524, 400.728754, 422.39588, 444.915731, 468.335132, 492.183366, 516.766033, 542.526745, 569.667593, 598.390667, 628.89806, 661.391861, 698.792657, 737.756676, 777.954129, 819.055227, 860.730182, 903.646881, 946.228441, 987.895655, 1028.06931, 1066.17021, 1099.95773, 1132.54468, 1171.16499, 1217.09209, 1259.4302, 1292.74724, 1311.61115, 1296.65286, 1259.54483, 1227.16359, 1242.0961, 1243.79652, 1183.12188, 1034.28788, 811.090152, 509.156496, 145.945457, -277.255349, -839.204392, -1573.38155, -2275.7357, -3253.82095, -4623.33943, -6499.9933, -8999.48468, -14978.7934, -21127.0244, -26874.5606, -31651.7847, -34889.0793, -33723.751, -31407.9762, -28354.8849, -26160.4764, -24769.6613, -23634.4877, -22724.5093, -22009.2796, -21458.3522, -21041.2809, -21333.9704, -21262.0719, -20929.5148, -20440.2281, -19898.1412, -19407.1832, -19071.2833, -18531.9506, -17932.6812},
        {-946.797945, -943.00765, -938.953851, -934.642798, -930.08074, -925.273926, -920.228606, -914.951029, -909.447445, -903.724101, -897.787249, -891.643137, -885.298015, -878.758131, -872.029736, -865.119077, -858.032406, -850.775971, -843.356021, -835.778805, -828.050574, -820.177576, -812.16606, -804.022276, -795.752474, -787.362901, -778.859809, -770.249446, -761.538061, -752.731903, -743.837223, -734.860269, -725.807291, -716.684537, -707.709977, -698.779547, -689.786639, -680.730631, -671.610902, -662.42683, -653.177795, -643.863174, -634.482346, -625.03469, -615.519584, -605.936408, -596.284539, -586.563357, -576.772239, -566.910565, -556.977713, -546.830791, -536.614533, -526.3314, -515.983857, -505.574365, -495.105388, -484.579388, -473.998827, -463.36617, -452.683877, -441.954413, -431.18024, -420.363821, -409.507617, -398.614093, -387.685711, -376.76813, -365.905972, -355.01261, -344.087708, -333.130927, -322.14193, -311.120379, -300.065937, -288.978266, -277.85703, -266.70189, -255.512508, -244.288548, -233.029672, -221.735542, -210.405821, -199.040172, -187.370422, -175.535786, -163.678694, -151.807502, -139.930563, -128.056231, -116.192862, -104.348809, -92.5324257, -80.7520673, -69.0160876, -57.332841, -45.7106815, -34.1579635, -22.6830411, -11.2942685, 0, 7.92416365, 16.0380738, 24.3571314, 32.8967375, 41.6722928, 50.6991984, 59.9928551, 69.568664, 79.4420259, 89.6283418, 100.143013, 111.001439, 122.219023, 133.811164, 145.793263, 158.180723, 171.114718, 184.628619, 198.453333, 212.631213, 227.204612, 242.215883, 257.707379, 273.721453, 290.300457, 307.486744, 325.322668, 343.850581, 363.112836, 383.372887, 404.370916, 426.082945, 448.484996, 470.516973, 492.672957, 515.723331, 539.954954, 565.654685, 593.109384, 622.605909, 658.106624, 695.104251, 733.166529, 771.861198, 810.755996, 852.98889, 893.664834, 931.459008, 965.046594, 993.102775, 1008.65384, 1022.92806, 1049.7607, 1092.85887, 1135.25143, 1169.25134, 1187.17159, 1168.0916, 1120.56551, 1067.58405, 1041.23783, 993.974136, 890.647813, 710.222975, 455.571076, 118.58971, -288.121161, -767.296661, -1397.58062, -2214.6786, -3054.91449, -4184.04965, -5706.80641, -7727.90707, -10352.0739, -16194.0548, -22221.0401, -27910.2459, -32738.8881, -36184.1826, -35667.9016, -34093.0007, -31817.6005, -30206.7, -29219.8635, -28397.1226, -27719.0186, -27166.0923, -26718.8848, -26357.9371, -27067.1292, -27125.878, -26710.8402, -25998.6729, -25166.0328, -24389.5769, -23845.9621, -23019.1395, -22173.9482},
        {-896.802199, -893.170384, -889.317809, -885.249068, -880.968756, -876.481471, -871.791805, -866.904356, -861.823717, -856.554485, -851.101255, -845.468622, -839.661181, -833.683527, -827.540257, -821.235965, -814.775247, -808.162697, -801.402912, -794.500486, -787.460015, -780.286094, -772.983318, -765.556283, -758.009584, -750.347816, -742.575575, -734.697456, -726.718055, -718.641965, -710.473784, -702.218105, -693.879525, -685.462639, -677.117447, -668.772782, -660.355733, -651.866176, -643.303987, -634.669044, -625.961222, -617.180399, -608.326451, -599.399254, -590.398686, -581.324623, -572.176942, -562.955519, -553.660232, -544.290956, -534.847568, -525.209572, -515.499826, -505.720818, -495.875034, -485.964958, -475.993079, -465.96188, -455.87385, -445.731473, -435.537236, -425.293626, -415.003127, -404.668226, -394.29141, -383.875165, -373.421976, -362.980254, -352.597308, -342.180908, -331.730562, -321.24578, -310.72607, -300.170941, -289.579902, -278.952463, -268.288132, -257.586418, -246.84683, -236.068877, -225.252069, -214.395914, -203.49992, -192.563598, -181.292515, -169.839334, -158.360249, -146.864307, -135.360558, -123.858049, -112.365828, -100.892943, -89.448442, -78.0413724, -66.6807825, -55.3757202, -44.1352335, -32.9683701, -21.8841781, -10.8917055, 0, 7.23333339, 14.704329, 22.4236738, 30.402055, 38.6501594, 47.1786741, 55.9982862, 65.1196826, 74.5535505, 84.3105768, 94.4014486, 104.836853, 115.627477, 126.784007, 138.317131, 150.237535, 162.727511, 175.822259, 189.144317, 202.741143, 216.660197, 230.948937, 245.654824, 260.825316, 276.507872, 292.749952, 309.599014, 327.102519, 345.307925, 364.545384, 384.476008, 405.06245, 426.26736, 446.954025, 467.634778, 489.115118, 511.687505, 535.644401, 561.278267, 588.881563, 623.091294, 658.533126, 694.649499, 730.882856, 766.675638, 803.951616, 839.05157, 870.797611, 898.011846, 919.516386, 929.395838, 937.000094, 954.090677, 983.734817, 1011.2489, 1029.91907, 1033.03149, 1004.88457, 949.794859, 881.883706, 822.91774, 735.968591, 597.071711, 388.493672, 103.059337, -271.252859, -729.227661, -1278.41783, -1989.5778, -2897.71157, -3854.34688, -5106.1911, -6751.75923, -8889.5663, -11618.1273, -17424.8515, -23422.1361, -29111.2725, -33993.5523, -37570.2668, -37402.0481, -36224.62, -34370.9852, -33106.5277, -32395.1826, -31812.3926, -31341.5174, -30965.9167, -30668.9502, -30433.9775, -31345.9176, -31528.0457, -31171.4668, -30467.2861, -29606.6087, -28780.5396, -28180.184, -27307.728, -26269.5885},
        {-851.022377, -847.29979, -843.41625, -839.374093, -835.17565, -830.823257, -826.319245, -821.66595, -816.865704, -811.920841, -806.833695, -801.606598, -796.241886, -790.74189, -785.108946, -779.345385, -773.453543, -767.435752, -761.294346, -755.031658, -748.650023, -742.151773, -735.539242, -728.814764, -721.980673, -715.039301, -707.992982, -700.844051, -693.59484, -686.247683, -678.804914, -671.268865, -663.641872, -655.926267, -648.179167, -640.374362, -632.484983, -624.511586, -616.454726, -608.314959, -600.09284, -591.788927, -583.403773, -574.937935, -566.391968, -557.766429, -549.061873, -540.278856, -531.417933, -522.47966, -513.464593, -504.273565, -495.009016, -485.673663, -476.270225, -466.801418, -457.26996, -447.678569, -438.029962, -428.326857, -418.571971, -408.768021, -398.917725, -389.023801, -379.088966, -369.115938, -359.107434, -349.123193, -339.221583, -329.287721, -319.320629, -309.319327, -299.282836, -289.210177, -279.100372, -268.952441, -258.765406, -248.538287, -238.270106, -227.959884, -217.606642, -207.2094, -196.76718, -186.279004, -175.411953, -164.337999, -153.233106, -142.107065, -130.96967, -119.830713, -108.699989, -97.5872905, -86.50241, -75.455141, -64.4552765, -53.5126097, -42.6369337, -31.8380416, -21.1257265, -10.5097816, 0, 6.60563213, 13.4903855, 20.6605705, 28.1224975, 35.882477, 43.9468195, 52.3218352, 61.0138347, 70.0291284, 79.3740267, 89.05484, 99.0778788, 109.449453, 120.175874, 131.263452, 142.718497, 154.761731, 167.430097, 180.244923, 193.258466, 206.522981, 220.090725, 234.013954, 248.344923, 263.135888, 278.439106, 294.306831, 310.791321, 327.944831, 346.140229, 364.991602, 384.45502, 404.486556, 423.975395, 443.411053, 463.567548, 484.721019, 507.147602, 531.123437, 556.924662, 589.605636, 623.210035, 657.079125, 690.554177, 722.976457, 754.794912, 783.966213, 809.55471, 830.62475, 846.240685, 852.230628, 854.850559, 862.392667, 876.864383, 887.441764, 889.06189, 876.661845, 842.081279, 784.058666, 704.711498, 612.213675, 483.757432, 309.395799, 75.9058131, -235.486469, -641.007, -1144.0878, -1757.63787, -2541.47732, -3528.89066, -4590.38857, -5947.07158, -7688.56899, -9904.51015, -12684.5244, -18338.8052, -24181.2766, -29746.427, -34568.7447, -38182.7178, -38324.0713, -37525.2325, -36091.5881, -35184.8752, -34771.5087, -34461.091, -34237.8775, -34086.1237, -33990.0853, -33934.0177, -35019.3094, -35349.7461, -35114.3536, -34502.1579, -33702.1851, -32903.461, -32295.0116, -31440.31, -30252.6469},
        {-807.945011, -804.135934, -800.223006, -796.206365, -792.086149, -787.862495, -783.535542, -779.105427, -774.572288, -769.936263, -765.19749, -760.356106, -755.41225, -750.366059, -745.21767, -739.967223, -734.614854, -729.160701, -723.604903, -717.947597, -712.18892, -706.329011, -700.368008, -694.306049, -688.14327, -681.87981, -675.515807, -669.051399, -662.486723, -655.821918, -649.05712, -642.192468, -635.228101, -628.164154, -620.964274, -613.647613, -606.23376, -598.724038, -591.119769, -583.422275, -575.632878, -567.752901, -559.783664, -551.726491, -543.582703, -535.353623, -527.040572, -518.644873, -510.167848, -501.610819, -492.975107, -484.183759, -475.31807, -466.381059, -457.375745, -448.305147, -439.172284, -429.980174, -420.731836, -411.430291, -402.078556, -392.67965, -383.236593, -373.752403, -364.2301, -354.672702, -345.083228, -335.536373, -326.105111, -316.644636, -307.153318, -297.629531, -288.071648, -278.478041, -268.847081, -259.177143, -249.466598, -239.713819, -229.917178, -220.075048, -210.185801, -200.24781, -190.259447, -180.219085, -169.749741, -159.04536, -148.3044, -137.537413, -126.754952, -115.967573, -105.185827, -94.420268, -83.6814503, -72.979927, -62.3262516, -51.7309776, -41.2046584, -30.7578475, -20.4010985, -10.1449648, 0, 6.0114793, 12.3407776, 18.9900326, 25.9613816, 33.2569624, 40.8789125, 48.8293694, 57.1104707, 65.7243539, 74.6731566, 83.9590164, 93.5840708, 103.550457, 113.860314, 124.515777, 135.518986, 147.129378, 159.38585, 171.708546, 184.154741, 196.781708, 209.64672, 222.807051, 236.319975, 250.242765, 264.632696, 279.547041, 295.043072, 311.178066, 328.365094, 346.175171, 364.558829, 383.466602, 401.869228, 420.207137, 439.182041, 459.038411, 480.020719, 502.373437, 526.341039, 557.108198, 588.475644, 619.721289, 650.123042, 678.958814, 705.386122, 728.833369, 748.608566, 764.019722, 774.374846, 777.38652, 775.90815, 774.344308, 773.514728, 766.920098, 751.347211, 723.582856, 683.602876, 624.280226, 535.008893, 409.693026, 240.362615, 31.1567854, -226.344769, -561.843617, -995.272879, -1538.20372, -2208.40136, -3054.29148, -4106.95649, -5259.29196, -6701.8266, -8513.47668, -10773.1585, -13559.7883, -18972.3444, -24564.6655, -29910.6526, -34584.2063, -38159.2275, -38570.2723, -38123.4826, -37095.3357, -36542.8744, -36433.6019, -36409.7856, -36455.1905, -36553.5817, -36688.724, -36844.3825, -38056.6834, -38543.764, -38476.475, -38025.6672, -37362.1915, -36656.8984, -36080.6389, -35288.4824, -34007.1298},
        {-766.056631, -762.418883, -758.711904, -754.934375, -751.084979, -747.162397, -743.16531, -739.092402, -734.942353, -730.713846, -726.405562, -722.016184, -717.544392, -712.98887, -708.348299, -703.621361, -698.806738, -693.903111, -688.909162, -683.823575, -678.645029, -673.372207, -668.003792, -662.538464, -656.974906, -651.3118, -645.547828, -639.68167, -633.712011, -627.63753, -621.45691, -615.168833, -608.771981, -602.265036, -595.541904, -588.655859, -581.661434, -574.560709, -567.355768, -560.048692, -552.641564, -545.136465, -537.535477, -529.840683, -522.054164, -514.178003, -506.214281, -498.165081, -490.032484, -481.818573, -473.52543, -465.101144, -456.602961, -448.034132, -439.397911, -430.697548, -421.936298, -413.117412, -404.244144, -395.319744, -386.347466, -377.330563, -368.272287, -359.175889, -350.044624, -340.881742, -331.690498, -322.559223, -313.57421, -304.563237, -295.524041, -286.454358, -277.351927, -268.214484, -259.039767, -249.825511, -240.569454, -231.269333, -221.922885, -212.527848, -203.081957, -193.582951, -184.028565, -174.416538, -164.326887, -153.974995, -143.581266, -133.156991, -122.713464, -112.261976, -101.813822, -91.3802936, -80.9726837, -70.6022849, -60.2803902, -50.0182922, -39.8272837, -29.7186575, -19.7037063, -9.79372288, 0, 5.42129432, 11.2000398, 17.3342714, 23.822024, 30.6613323, 37.8502313, 45.3867559, 53.2689409, 61.4948212, 70.0624317, 78.9698072, 88.2149826, 97.7959928, 107.710873, 117.957657, 128.534381, 139.74245, 151.623239, 163.488584, 175.40153, 187.425119, 199.622394, 212.056397, 224.790173, 237.886764, 251.409214, 265.420564, 279.983859, 295.16214, 311.427648, 328.284191, 345.672053, 363.531517, 380.923666, 398.228385, 416.060014, 434.622598, 454.120182, 474.756811, 496.736531, 525.05753, 553.670622, 581.841869, 608.837332, 633.923072, 655.59259, 674.077647, 688.837441, 699.331174, 705.018045, 705.09182, 699.601555, 689.62324, 674.953015, 652.773983, 621.442449, 579.314717, 533.3705, 471.382826, 371.717361, 215.923183, 8.80609479, -234.108621, -517.062241, -877.78938, -1338.65066, -1917.07759, -2634.15285, -3529.03262, -4630.6497, -5857.30944, -7365.59166, -9222.7233, -11495.9312, -14252.4424, -19361.8977, -24638.5066, -29698.8924, -34159.6784, -37637.4882, -38276.952, -38148.015, -37498.1547, -37281.6573, -37466.2219, -37725.0442, -38040.5485, -38395.1591, -38771.3004, -39151.3968, -40427.4183, -41062.8842, -41194.8055, -40960.1929, -40496.0573, -39939.4093, -39427.2599, -38723.8419, -37417.0438},
        {-723.843768, -720.888706, -717.856775, -714.746616, -711.556868, -708.286171, -704.933165, -701.496488, -697.97478, -694.366682, -690.670831, -686.885868, -683.010433, -679.043164, -674.982702, -670.827685, -666.576754, -662.228547, -657.781705, -653.234866, -648.586671, -643.835759, -638.980768, -634.02034, -628.953113, -623.777727, -618.492821, -613.097035, -607.589008, -601.96738, -596.230791, -590.377879, -584.407285, -578.317647, -571.981193, -565.462428, -558.827375, -552.078776, -545.219374, -538.251909, -531.179125, -524.003762, -516.728563, -509.35627, -501.889623, -494.331366, -486.68424, -478.950986, -471.134348, -463.237065, -455.261881, -447.186709, -439.039658, -430.824006, -422.543035, -414.200025, -405.798254, -397.341005, -388.831555, -380.273186, -371.669178, -363.022811, -354.337364, -345.616118, -336.862353, -328.079348, -319.270385, -310.531171, -301.955195, -293.355111, -284.728207, -276.071772, -267.383095, -258.659462, -249.898163, -241.096486, -232.25172, -223.361152, -214.42207, -205.431764, -196.387522, -187.286631, -178.126381, -168.904058, -159.164397, -149.140482, -139.070838, -128.967439, -118.842259, -108.707274, -98.5744579, -88.4557852, -78.3632306, -68.3087688, -58.3043743, -48.3620218, -38.4936859, -28.7113411, -19.0269622, -9.45252358, 0, 4.80549664, 10.0127065, 15.6154983, 21.607741, 27.9833033, 34.736054, 41.8598619, 49.3485958, 57.1961246, 65.396317, 73.9430418, 82.8301678, 92.0515638, 101.601099, 111.472641, 121.66006, 132.512945, 144.075981, 155.538433, 166.970393, 178.441957, 190.02322, 201.784276, 213.795219, 226.126146, 238.847149, 252.028323, 265.739765, 280.051567, 295.535562, 311.576137, 328.092865, 345.005319, 361.426854, 377.680151, 394.302886, 411.456497, 429.302422, 448.002102, 467.716975, 492.91218, 518.135634, 542.706747, 565.944929, 587.16959, 605.281662, 620.123654, 631.119597, 637.693521, 639.269457, 635.574837, 625.359463, 607.907103, 582.446407, 548.093497, 504.015022, 449.377625, 395.305294, 326.289757, 213.778371, 31.4715327, -207.890173, -482.863711, -795.05077, -1185.10103, -1675.74052, -2286.21153, -3038.3369, -3966.71306, -5098.71094, -6380.6934, -7933.50228, -9812.54982, -12073.2482, -14771.0098, -19543.8938, -24469.0036, -29206.0896, -33414.9022, -36755.1919, -37580.4115, -37727.474, -37415.9716, -37502.3558, -37954.1288, -38473.4345, -39041.0431, -39637.7244, -40244.2485, -40841.3854, -42100.8927, -42859.8917, -43206.3196, -43228.114, -43013.212, -42649.5511, -42225.0687, -41617.9854, -40366.3952},
        {-673.679382, -673.124023, -672.35271, -671.368386, -670.173992, -668.772469, -667.166758, -665.3598, -663.354537, -661.15391, -658.76086, -656.178329, -653.409257, -650.456586, -647.323257, -644.012212, -640.526392, -636.868737, -633.04219, -629.049691, -624.894181, -620.578603, -616.105897, -611.479004, -606.700866, -601.774424, -596.702619, -591.488393, -586.134686, -580.64444, -575.020596, -569.266096, -563.38388, -557.376891, -551.242548, -544.986671, -538.615141, -532.13108, -525.537608, -518.837846, -512.034913, -505.13193, -498.132018, -491.038297, -483.853887, -476.581908, -469.225482, -461.787728, -454.271767, -446.680719, -439.017704, -431.310356, -423.536751, -415.699478, -407.801125, -399.844283, -391.831539, -383.765484, -375.648705, -367.483792, -359.273334, -351.01992, -342.726138, -334.394579, -326.02783, -317.628481, -309.199121, -300.81974, -292.56847, -284.288266, -275.976699, -267.63134, -259.249761, -250.829533, -242.368228, -233.863417, -225.312672, -216.713563, -208.063662, -199.360541, -190.601771, -181.784923, -172.907568, -163.967279, -154.499314, -144.742125, -134.93972, -125.104673, -115.249559, -105.386952, -95.5294266, -85.6895569, -75.8799172, -66.1130819, -56.4016253, -46.7581218, -37.1951455, -27.7252709, -18.3610722, -9.1151238, 0, 3.88460888, 8.26305334, 13.1226793, 18.4508328, 24.2348598, 30.4621062, 37.119918, 44.1956411, 51.6766216, 59.5502053, 67.8037383, 76.4245665, 85.4000359, 94.7174924, 104.364282, 114.327751, 125.061088, 136.618188, 147.932705, 159.091807, 170.182664, 181.292446, 192.508322, 203.917462, 215.607035, 227.664211, 240.176159, 253.230049, 266.913051, 282.102941, 297.806392, 313.873392, 330.153929, 345.441923, 360.115394, 374.833982, 389.764496, 405.073742, 420.928529, 437.495664, 458.486978, 479.445337, 499.843955, 519.156049, 536.854835, 555.246551, 570.263135, 580.669549, 585.230755, 582.711713, 569.828298, 549.898995, 525.365637, 496.616, 459.844157, 413.12707, 354.5417, 294.278368, 215.547638, 88.3454955, -115.508305, -383.202883, -700.558688, -1069.11759, -1517.06053, -2060.34067, -2704.10137, -3461.49326, -4385.30295, -5505.89669, -6818.13771, -8393.3573, -10274.2112, -12503.3553, -15123.4452, -19535.0278, -24083.395, -28471.7299, -32403.2156, -35581.0351, -36570.0917, -36970.7012, -36973.8808, -37334.706, -38020.9207, -38765.2448, -39547.1734, -40346.2018, -41141.8252, -41913.5389, -42785.2659, -43530.5277, -44118.1515, -44516.9645, -44695.7941, -44623.4675, -44268.812, -43767.5339, -42768.6632},
        {-633.134032, -633.973973, -634.524619, -634.791176, -634.77885, -634.492849, -633.938379, -633.120646, -632.044858, -630.71622, -629.13994, -627.321225, -625.26528, -622.977312, -620.462529, -617.726137, -614.773342, -611.609351, -608.239371, -604.668609, -600.90227, -596.945562, -592.803692, -588.481866, -583.98529, -579.319171, -574.488717, -569.499133, -564.355627, -559.063404, -553.627672, -548.053638, -542.346507, -536.511487, -530.60951, -524.616748, -518.508703, -512.288772, -505.960354, -499.526847, -492.99165, -486.35816, -479.629776, -472.809896, -465.901918, -458.909241, -451.835262, -444.683381, -437.456994, -430.159501, -422.794299, -415.419451, -407.982506, -400.485677, -392.931178, -385.321221, -377.65802, -369.943789, -362.180739, -354.371086, -346.517041, -338.620818, -330.68463, -322.71069, -314.701213, -306.65841, -298.584495, -290.55286, -282.635189, -274.686892, -266.705571, -258.688822, -250.634246, -242.539442, -234.402008, -226.219543, -217.989646, -209.709917, -201.377954, -192.991357, -184.547724, -176.044654, -167.479746, -158.850599, -149.679593, -140.211931, -130.69853, -121.152411, -111.586595, -102.014103, -92.4479548, -82.9011723, -73.3867758, -63.9177862, -54.5072244, -45.168111, -35.9134669, -26.756313, -17.70967, -8.78655873, 0, 3.40832316, 7.33088624, 11.7527569, 16.6590028, 22.0346916, 27.8648909, 34.1346685, 40.829092, 47.9332289, 55.4321471, 63.3109142, 71.5545977, 80.1482655, 89.076985, 98.3258241, 107.87985, 118.209762, 129.370001, 140.247188, 150.930452, 161.508926, 172.071741, 182.708027, 193.506917, 204.557541, 215.949031, 227.770518, 240.111133, 253.060008, 267.577468, 282.562012, 297.841413, 313.243444, 327.718058, 341.531939, 355.185853, 368.77092, 382.378256, 396.098982, 410.024215, 426.384061, 442.479655, 457.983619, 472.568575, 485.907144, 502.517349, 516.015059, 524.861545, 527.518078, 522.445928, 505.667859, 481.064046, 452.544351, 420.620486, 378.866269, 323.97049, 252.621942, 168.885943, 60.3982713, -93.2524138, -310.193905, -587.278542, -918.017747, -1306.95245, -1777.31246, -2343.7491, -3011.08373, -3791.41915, -4732.90131, -5863.28814, -7189.4954, -8763.32963, -10619.7633, -12793.769, -15320.3191, -19432.849, -23668.2527, -27761.8869, -31449.1085, -34465.2743, -35526.2706, -36066.5716, -36258.4501, -36790.5245, -37630.6806, -38531.6605, -39472.1071, -40430.6635, -41385.973, -42316.6787, -43265.8343, -44142.1618, -44897.6094, -45484.1252, -45853.6573, -45958.154, -45749.5634, -45281.9998, -44420.3356},
        {-600.947329, -602.236046, -603.225762, -603.921964, -604.330141, -604.455781, -604.304373, -603.881404, -603.192362, -602.242736, -601.038014, -599.583684, -597.885235, -595.948154, -593.777929, -591.38005, -588.760003, -585.923278, -582.875362, -579.621744, -576.167911, -572.519352, -568.681556, -564.66001, -560.460202, -556.087621, -551.547754, -546.846091, -541.988119, -536.979326, -531.825201, -526.531231, -521.102905, -515.545712, -509.924402, -504.213586, -498.387161, -492.448693, -486.401746, -480.249885, -473.996675, -467.645681, -461.200468, -454.664601, -448.041645, -441.335165, -434.548726, -427.685893, -420.750231, -413.745305, -406.674679, -399.60721, -392.479756, -385.294466, -378.053492, -370.758981, -363.413085, -356.017951, -348.575731, -341.088573, -333.558627, -325.988042, -318.378969, -310.733557, -303.053955, -295.342314, -287.600782, -279.904653, -272.327465, -264.720514, -257.081207, -249.406952, -241.695157, -233.94323, -226.148579, -218.308612, -210.420737, -202.482362, -194.490895, -186.443744, -178.338316, -170.17202, -161.942264, -153.646455, -144.790706, -135.628407, -126.418852, -117.175392, -107.911379, -98.6401645, -89.3750995, -80.1295356, -70.9168243, -61.7503171, -52.6433653, -43.6093204, -34.6615339, -25.8133571, -17.0781416, -8.46923875, 0, 3.31287574, 7.0984024, 11.3431687, 16.0337632, 21.1567748, 26.698792, 32.6464036, 38.9861983, 45.7047647, 52.7886916, 60.2245676, 67.9989814, 76.0985218, 84.5097773, 93.2193368, 102.213789, 111.888037, 122.286999, 132.473697, 142.522215, 152.506638, 162.50105, 172.579537, 182.816181, 193.285069, 204.060285, 215.215913, 226.826039, 238.964746, 252.450374, 266.33986, 280.484011, 294.733636, 308.570878, 322.030878, 335.247086, 348.18091, 360.793756, 373.047032, 384.902145, 396.829137, 408.125978, 418.654561, 428.276779, 436.854523, 449.865389, 460.151639, 466.17124, 466.382157, 459.242355, 441.819928, 415.661449, 384.689474, 348.792774, 299.62448, 232.535359, 142.876176, 26.1929658, -122.403154, -308.014085, -532.716319, -807.341604, -1131.1943, -1513.11507, -1978.10338, -2543.7415, -3224.84234, -4038.27019, -5011.04389, -6164.43831, -7491.66897, -9046.38688, -10859.5798, -12962.2353, -15385.3412, -19247.9894, -23221.0373, -27063.4467, -30534.1791, -33392.1964, -34459.303, -35056.3891, -35344.0551, -35968.3397, -36897.6907, -37898.29, -38948.6986, -40027.4776, -41113.1881, -42184.3913, -43550.5423, -44664.7974, -45534.7812, -46168.1187, -46572.4345, -46755.3535, -46724.5004, -46251.3528, -45453.4},
        {-573.713799, -574.88926, -575.793894, -576.432306, -576.809102, -576.928886, -576.796263, -576.415838, -575.792215, -574.930001, -573.8338, -572.508217, -570.957856, -569.187323, -567.201223, -565.004161, -562.600741, -559.995569, -557.193249, -554.198387, -551.015588, -547.649456, -544.104596, -540.385613, -536.497113, -532.4437, -528.229979, -523.860555, -519.340034, -514.673019, -509.864116, -504.917931, -499.839067, -494.63213, -489.333206, -483.930498, -478.411828, -472.78078, -467.040939, -461.195886, -455.249206, -449.204481, -443.065296, -436.835232, -430.517875, -424.116806, -417.63561, -411.077869, -404.447168, -397.747088, -390.981215, -384.21375, -377.386344, -370.501266, -363.560784, -356.567167, -349.522686, -342.429609, -335.290205, -328.106744, -320.881495, -313.616727, -306.314708, -298.97771, -291.608, -284.207847, -276.779522, -269.404113, -262.160819, -254.889346, -247.586855, -240.250504, -232.877454, -225.464864, -218.009894, -210.509704, -202.961452, -195.362298, -187.709403, -179.999925, -172.231025, -164.399861, -156.503594, -148.539383, -139.997434, -131.139048, -122.231615, -113.288748, -104.324057, -95.3511539, -86.3836495, -77.4351551, -68.5192819, -59.649641, -50.8398437, -42.103501, -33.4542243, -24.9056247, -16.4713133, -8.16490132, 0, 3.44469515, 7.28242127, 11.5038652, 16.099714, 21.0606543, 26.3773732, 32.0405574, 38.0408939, 44.3690697, 51.0157714, 57.9716861, 65.2275006, 72.7739019, 80.6015767, 88.701212, 97.0634947, 105.922846, 115.304773, 124.648657, 133.997414, 143.39396, 152.88121, 162.502081, 172.299488, 182.316346, 192.595573, 203.180082, 214.11279, 225.436613, 237.652117, 250.176762, 262.916169, 275.775958, 288.943614, 302.183826, 315.186117, 327.771551, 339.761186, 350.976085, 361.23731, 369.504555, 376.722403, 382.880441, 387.968259, 391.975446, 400.527331, 406.568828, 408.680591, 405.443273, 395.437529, 377.208937, 349.416096, 315.59792, 274.906765, 217.539103, 137.712413, 29.6441763, -119.193353, -306.304462, -521.83501, -752.551366, -1021.85845, -1332.13144, -1695.96955, -2143.29601, -2696.08206, -3382.34127, -4225.61139, -5227.27495, -6401.24672, -7718.9249, -9243.28185, -11002.9496, -13026.5603, -15342.7458, -18982.8863, -22722.1788, -26338.5684, -29610.0002, -32314.4193, -33356.9965, -33970.3005, -34306.3225, -34976.6485, -35951.3466, -37008.9872, -38129.2131, -39291.6674, -40475.9931, -41661.8333, -43557.0045, -44943.2772, -45905.9066, -46530.1478, -46901.256, -47104.4863, -47225.094, -46749.5328, -46022.0824},
        {-545.055603, -546.174588, -547.04835, -547.680757, -548.075678, -548.236981, -548.168534, -547.874206, -547.357864, -546.623378, -545.674616, -544.515445, -543.149735, -541.581353, -539.814168, -537.852049, -535.698863, -533.358478, -530.834765, -528.131589, -525.252821, -522.202328, -518.983978, -515.60164, -512.059183, -508.360474, -504.509382, -500.509775, -496.365521, -492.08049, -487.658549, -483.103566, -478.41941, -473.609949, -468.697081, -463.67528, -458.538804, -453.290937, -447.934961, -442.47416, -436.911816, -431.251214, -425.495637, -419.648367, -413.712688, -407.691884, -401.589237, -395.40803, -389.151548, -382.823073, -376.425888, -370.011447, -363.533818, -356.995242, -350.397956, -343.7442, -337.036212, -330.276233, -323.4665, -316.609253, -309.70673, -302.761172, -295.774816, -288.749903, -281.68867, -274.593357, -267.466203, -260.383364, -253.419222, -246.423567, -239.393846, -232.327506, -225.221995, -218.07476, -210.883248, -203.644907, -196.357184, -189.017525, -181.623379, -174.172192, -166.661412, -159.088486, -151.450861, -143.745984, -135.469096, -126.87931, -118.241776, -109.570241, -100.878448, -92.1801433, -83.4890707, -74.8189754, -66.1836022, -57.5966961, -49.072002, -40.6232646, -32.2642289, -24.0086397, -15.8702419, -7.86278039, 0, 3.7140203, 7.70565423, 11.9733519, 16.5155634, 21.3307388, 26.4173283, 31.7737819, 37.3985497, 43.2900819, 49.4468285, 55.8672396, 62.5497653, 69.4928557, 76.694961, 84.1545312, 91.8700164, 99.7331501, 107.727137, 116.09181, 124.802752, 133.835545, 143.165771, 152.769012, 162.62085, 172.696869, 182.972649, 193.423774, 204.025825, 214.754385, 225.511372, 236.373043, 247.337079, 258.40116, 270.390316, 282.888553, 295.259256, 307.251899, 318.615962, 329.100921, 338.456254, 345.667833, 351.481141, 355.795058, 358.508462, 359.520233, 364.858561, 366.760684, 363.593152, 353.722513, 335.515319, 305.801975, 266.362685, 222.865237, 175.397738, 112.656444, 28.4523196, -83.4036709, -234.719811, -424.788786, -646.773194, -893.53624, -1180.55632, -1502.7498, -1869.61444, -2315.3554, -2862.83673, -3542.86738, -4380.37122, -5371.14073, -6527.44676, -7830.23882, -9327.92773, -11043.6264, -13000.4475, -15221.5041, -18618.0124, -22102.9558, -25477.4214, -28542.4961, -31099.2669, -32164.1091, -32845.9623, -33282.2183, -34019.8112, -35034.0225, -36129.491, -37290.03, -38499.4531, -39741.5738, -41000.2056, -43177.2961, -44774.7152, -45881.0187, -46584.7629, -46974.5038, -47138.7974, -47166.1999, -46756.2412, -46575.1051},
        {-516.455553, -517.573981, -518.468629, -519.142786, -519.59974, -519.842778, -519.875189, -519.70026, -519.32128, -518.741536, -517.964316, -516.992908, -515.8306, -514.480679, -512.946435, -511.231154, -509.338124, -507.270634, -505.031971, -502.625424, -500.054279, -497.321825, -494.43135, -491.386142, -488.189489, -484.844678, -481.354997, -477.723735, -473.954179, -470.049618, -466.013338, -461.848628, -457.558776, -453.14707, -448.630419, -444.005013, -439.266881, -434.418867, -429.463817, -424.404577, -419.243994, -413.984912, -408.630178, -403.182638, -397.645137, -392.020521, -386.311636, -380.521328, -374.652444, -368.707827, -362.690326, -356.629542, -350.500985, -344.30692, -338.049612, -331.731328, -325.354333, -318.920893, -312.433274, -305.893742, -299.304562, -292.668, -285.986322, -279.261794, -272.496681, -265.69325, -258.853766, -252.050592, -245.354408, -238.622909, -231.853818, -225.044856, -218.193745, -211.298206, -204.355961, -197.364731, -190.322239, -183.226206, -176.074354, -168.864404, -161.594079, -154.261099, -146.863186, -139.398063, -131.363003, -123.016479, -114.622976, -106.196458, -97.7508846, -89.3002195, -80.8584245, -72.4394618, -64.0572934, -55.7258815, -47.4591883, -39.2711759, -31.1758063, -23.1870418, -15.3188445, -7.58517654, 0, 3.86233851, 7.91077311, 12.150161, 16.5853592, 21.2212249, 26.0626153, 31.1143875, 36.3813987, 41.8685059, 47.5805663, 53.5224371, 59.6989754, 66.1150384, 72.7754831, 79.6851668, 86.8489465, 93.8168271, 100.528687, 108.018215, 116.192802, 124.959834, 134.226702, 143.900794, 153.889499, 164.100206, 174.440302, 184.817178, 195.138222, 205.310823, 214.659119, 223.887608, 233.078653, 242.314619, 252.849676, 264.180284, 275.490423, 286.510914, 296.972581, 306.606244, 315.142725, 322.410833, 328.013582, 331.662624, 333.069609, 331.946189, 334.38844, 332.127481, 323.278858, 305.958115, 278.280796, 234.676212, 181.45154, 126.932751, 72.5543152, 5.68056515, -80.0069618, -190.826728, -335.292788, -517.029013, -737.264088, -1000.20304, -1309.16538, -1651.73971, -2032.983, -2487.55135, -3036.51436, -3706.67313, -4520.58322, -5479.7847, -6597.70521, -7877.91584, -9345.21712, -11016.4485, -12908.4496, -15038.0598, -18191.9276, -21424.6312, -24560.558, -27424.0953, -29839.6302, -30939.247, -31701.1714, -32246.4919, -33048.6696, -34089.3122, -35203.2057, -36379.299, -37606.5409, -38873.8804, -40170.2664, -42371.0334, -44082.4486, -45346.5642, -46205.4322, -46701.1048, -46875.6342, -46771.0723, -46500.4007, -46782.3717},
        {-488.567494, -489.825979, -490.870626, -491.704499, -492.33066, -492.752172, -492.9721, -492.993504, -492.819449, -492.452997, -491.897211, -491.155154, -490.22989, -489.124481, -487.841989, -486.385479, -484.758012, -482.962652, -481.002463, -478.880505, -476.599844, -474.163541, -471.57466, -468.836263, -465.951414, -462.923175, -459.75461, -456.448781, -453.008751, -449.437583, -445.738341, -441.914087, -437.967884, -433.902794, -429.742072, -425.478259, -421.103657, -416.620674, -412.031717, -407.339195, -402.545514, -397.653083, -392.66431, -387.581601, -382.407366, -377.144012, -371.793945, -366.359576, -360.84331, -355.247556, -349.574721, -343.820426, -337.994014, -332.098038, -326.135054, -320.107618, -314.018283, -307.869605, -301.664138, -295.404438, -289.09306, -282.732558, -276.325487, -269.874403, -263.381859, -256.850412, -250.282616, -243.759633, -237.36074, -230.926369, -224.453978, -217.941026, -211.384972, -204.783276, -198.133396, -191.432792, -184.678922, -177.869247, -171.001224, -164.072312, -157.079972, -150.021661, -142.89484, -135.696966, -127.900558, -119.776587, -111.602333, -103.392291, -95.1609558, -86.9228211, -78.6923817, -70.4841319, -62.3125662, -54.192179, -46.1374648, -38.162918, -30.2830331, -22.5123043, -14.8652263, -7.35629338, 0, 3.58478179, 7.35529311, 11.3160163, 15.4714338, 19.8260278, 24.3842807, 29.150675, 34.1296928, 39.3258167, 44.7435289, 50.3873117, 56.2616476, 62.3710189, 68.7199079, 75.312797, 82.1541686, 88.635326, 94.6731547, 101.659089, 109.466216, 117.967623, 127.036397, 136.545623, 146.36839, 156.377784, 166.446892, 176.448801, 186.256598, 195.743369, 203.98251, 211.940019, 219.728892, 227.462122, 236.421204, 246.134882, 255.820301, 265.239906, 274.156139, 282.331446, 289.528271, 296.695506, 302.048056, 305.116232, 305.430349, 302.520721, 301.75203, 295.361628, 281.421234, 258.00257, 223.177357, 170.35051, 107.964427, 46.0262687, -13.3005715, -83.8670565, -172.33403, -285.362336, -431.251698, -614.651607, -838.423683, -1108.92133, -1427.84517, -1786.13486, -2188.08432, -2660.61547, -3220.26362, -3884.38443, -4669.7259, -5590.02797, -6660.7321, -7908.36067, -9335.56948, -10953.9643, -12775.151, -14810.7352, -17739.1068, -20738.3916, -23653.4992, -26329.3394, -28610.822, -29733.3348, -30557.6574, -31189.9234, -32037.0203, -33086.0329, -34198.7323, -35368.469, -36588.5938, -37852.4574, -39153.4106, -41042.7934, -42715.1116, -44125.5767, -45229.4003, -45981.7938, -46337.9689, -46253.137, -46220.7832, -46294.2315},
        {-461.473317, -463.156022, -464.612031, -465.844868, -466.858055, -467.655116, -468.239574, -468.614952, -468.784774, -468.752562, -468.52184, -468.096131, -467.478958, -466.673845, -465.684314, -464.51389, -463.166094, -461.64445, -459.952482, -458.093712, -456.071665, -453.889862, -451.551827, -449.061085, -446.421156, -443.635566, -440.707837, -437.641491, -434.440054, -431.107047, -427.645994, -424.060418, -420.353842, -416.52979, -412.636412, -408.653979, -404.562225, -400.363225, -396.059055, -391.651789, -387.143502, -382.536269, -377.832166, -373.033267, -368.141647, -363.159381, -358.088544, -352.931211, -347.689456, -342.365356, -336.960985, -331.42143, -325.80699, -320.120974, -314.366693, -308.547457, -302.666577, -296.727363, -290.733125, -284.687173, -278.592819, -272.453371, -266.272142, -260.05244, -253.797577, -247.510862, -241.195606, -234.965511, -228.93163, -222.869596, -216.775565, -210.645691, -204.476127, -198.263027, -192.002547, -185.690839, -179.324058, -172.898358, -166.409893, -159.854818, -153.229285, -146.52945, -139.751466, -132.891488, -125.346822, -117.425643, -109.44332, -101.415443, -93.3575986, -85.2853749, -77.2143598, -69.1601412, -61.1383066, -53.1644442, -45.2541415, -37.4229865, -29.686567, -22.0604708, -14.5602856, -7.20159944, 0, 2.54507037, 5.43899654, 8.67340884, 12.2399376, 16.1302132, 20.335866, 24.8485262, 29.6598242, 34.7613905, 40.1448552, 45.8018488, 51.7240015, 57.9029438, 64.330306, 70.9977183, 77.8968112, 84.6242705, 91.1153059, 98.2548745, 105.949976, 114.107609, 122.634774, 131.438469, 140.425694, 149.503449, 158.578733, 167.558545, 176.349885, 184.859752, 192.546637, 199.930501, 207.052895, 213.955372, 221.315453, 228.856705, 236.133102, 242.995407, 249.294379, 254.880779, 259.605369, 265.399237, 269.399672, 271.050417, 269.79521, 265.077794, 260.355125, 250.054423, 232.616122, 206.480658, 170.088468, 118.951181, 58.0176903, -3.17753732, -63.2258074, -137.308543, -233.158544, -358.508608, -529.781031, -744.04423, -988.887175, -1250.38817, -1552.88995, -1911.8155, -2339.03657, -2840.40753, -3424.5448, -4093.74755, -4854.99442, -5738.42737, -6761.08586, -7963.87186, -9336.29119, -10887.1232, -12625.1474, -14559.143, -17291.1881, -20088.4387, -22811.3497, -25320.3757, -27475.9715, -28591.9585, -29439.8468, -30114.3628, -30974.5318, -32010.4615, -33102.1347, -34245.1653, -35435.1672, -36667.7542, -37938.54, -39059.6054, -40470.3429, -41997.8876, -43469.3744, -44711.9381, -45552.7136, -45818.836, -46162.6346, -44748.4485},
        {-437.574084, -439.388162, -440.985139, -442.368158, -443.540362, -444.504892, -445.264891, -445.823501, -446.183866, -446.349126, -446.322426, -446.106907, -445.705711, -445.121982, -444.358861, -443.419491, -442.307014, -441.024573, -439.57531, -437.962369, -436.18889, -434.258017, -432.172892, -429.936657, -427.552455, -425.023428, -422.35272, -419.543471, -416.598825, -413.521924, -410.315911, -406.983927, -403.529116, -399.95462, -396.285263, -392.512891, -388.629088, -384.636293, -380.536946, -376.333484, -372.028347, -367.623973, -363.122801, -358.527271, -353.83982, -349.062889, -344.198914, -339.250336, -334.219593, -329.109124, -323.921367, -318.617448, -313.242015, -307.798402, -302.289944, -296.719975, -291.09183, -285.408842, -279.674348, -273.89168, -268.064173, -262.195162, -256.287982, -250.345965, -244.372448, -238.370765, -232.344249, -226.417771, -220.713284, -214.983463, -209.223764, -203.429642, -197.596554, -191.719955, -185.795302, -179.818048, -173.783652, -167.687567, -161.52525, -155.292157, -148.983744, -142.595466, -136.122778, -129.561137, -122.26531, -114.564572, -106.795901, -98.9755432, -91.1197471, -83.2447595, -75.3668278, -67.5021993, -59.6671214, -51.8778415, -44.1506068, -36.5016647, -28.9472625, -21.5036476, -14.1870674, -7.01376903, 0, 2.07218869, 4.54629374, 7.40902556, 10.6470946, 14.2472112, 18.1960859, 22.480429, 27.086951, 32.0023622, 37.2133732, 42.7066943, 48.469036, 54.4871086, 60.7476226, 67.2372884, 73.9428164, 80.5519288, 87.0086234, 93.9758935, 101.376381, 109.132727, 117.167573, 125.40356, 133.763331, 142.169526, 150.544787, 158.811756, 166.893075, 174.711383, 181.95654, 188.869325, 195.442214, 201.667684, 207.881951, 213.905622, 219.467639, 224.457359, 228.764135, 232.277324, 234.886279, 239.081103, 241.358872, 241.100099, 237.685297, 230.494978, 220.222758, 204.607772, 182.702259, 153.558455, 116.228599, 70.1883439, 13.5490032, -49.3468511, -119.11008, -206.4648, -317.594483, -458.682602, -645.95015, -873.261821, -1123.53229, -1372.14871, -1656.87229, -2007.0279, -2439.49317, -2950.05467, -3544.18013, -4220.99785, -4984.332, -5861.25124, -6865.67585, -8026.45228, -9340.27826, -10815.9073, -12462.0928, -14287.5884, -16848.045, -19468.5943, -22021.2653, -24378.0872, -26411.089, -27485.1743, -28317.5814, -28997.6791, -29875.2012, -30936.4551, -32059.2625, -33235.3551, -34456.4644, -35714.3221, -37000.6597, -37631.1091, -38844.9883, -40400.5921, -42056.2153, -43570.1527, -44700.6991, -45206.1494, -45713.5798, -44381.6453},
        {-417.028928, -418.669942, -420.125465, -421.397477, -422.487958, -423.398887, -424.132245, -424.690012, -425.074168, -425.286693, -425.329566, -425.204767, -424.914277, -424.460076, -423.844143, -423.068459, -422.135003, -421.045756, -419.802697, -418.407806, -416.863063, -415.170449, -413.331943, -411.349526, -409.225176, -406.960875, -404.558601, -402.020336, -399.348059, -396.54375, -393.609389, -390.546955, -387.35843, -384.045792, -380.568576, -376.950877, -373.217301, -369.371206, -365.41595, -361.35489, -357.191383, -352.928787, -348.570459, -344.119757, -339.580038, -334.954659, -330.246979, -325.460354, -320.598143, -315.663701, -310.660388, -305.628124, -300.53291, -295.37731, -290.163891, -284.895216, -279.573851, -274.202359, -268.783307, -263.319259, -257.812779, -252.266433, -246.682785, -241.0644, -235.413843, -229.733679, -224.026472, -218.402092, -212.967829, -207.504944, -202.009045, -196.47574, -190.900639, -185.279351, -179.607483, -173.880645, -168.094446, -162.244494, -156.326398, -150.335767, -144.26821, -138.119335, -131.884751, -125.560067, -118.501837, -111.03864, -103.506733, -95.9224636, -88.3021792, -80.6622269, -73.0189542, -65.3887082, -57.7878362, -50.2326856, -42.7396035, -35.3249374, -28.0050344, -20.7962419, -13.7149072, -6.77737742, 0, 2.28780742, 4.8940391, 7.81039329, 11.0285682, 14.540262, 18.337173, 22.4109994, 26.7534394, 31.3561912, 36.2109531, 41.3094232, 46.6432999, 52.2042812, 57.9840655, 63.9743509, 70.1668358, 76.1725182, 81.9287107, 88.2882188, 95.1611619, 102.45766, 110.087832, 117.961799, 125.989679, 134.081592, 142.147659, 150.097998, 157.84273, 165.291974, 172.108196, 178.539976, 184.571729, 190.187869, 195.779056, 201.126583, 205.903412, 209.972088, 213.195151, 215.435146, 216.554614, 219.170198, 219.552136, 217.024127, 210.909866, 200.533051, 183.129653, 160.633026, 132.888801, 99.7426062, 61.040071, 21.4751605, -29.8805768, -97.8858208, -186.090732, -295.373908, -427.813427, -585.487362, -774.896302, -992.690694, -1230.69642, -1467.84414, -1740.65065, -2076.84931, -2493.76683, -2991.09641, -3576.62531, -4258.95455, -5046.08236, -5943.02327, -6956.47956, -8081.64973, -9338.50966, -10737.7437, -12290.0363, -14006.0718, -16417.792, -18884.3095, -21285.9942, -23503.2162, -25416.3455, -26417.8123, -27201.2195, -27856.054, -28756.5453, -29879.4699, -31082.9615, -32349.7285, -33662.4795, -35003.9231, -36356.7678, -36898.4413, -38043.0873, -39549.9005, -41178.0753, -42686.8065, -43835.2887, -44382.7163, -44804.0947, -45367.8811},
        {-396.733209, -398.600028, -400.274243, -401.758166, -403.054109, -404.164383, -405.0913, -405.837172, -406.404309, -406.795024, -407.011628, -407.056432, -406.931748, -406.639888, -406.183164, -405.563886, -404.784366, -403.846916, -402.753847, -401.507471, -400.1101, -398.564045, -396.871617, -395.035128, -393.05689, -390.939214, -388.684411, -386.294794, -383.772674, -381.120362, -378.34017, -375.434409, -372.405392, -369.255428, -365.95619, -362.525953, -358.983362, -355.331724, -351.574343, -347.714527, -343.755582, -339.700812, -335.553524, -331.317024, -326.994617, -322.589611, -318.10531, -313.54502, -308.912048, -304.209699, -299.44128, -294.659839, -289.817861, -284.917573, -279.961204, -274.950979, -269.889128, -264.777878, -259.619455, -254.416089, -249.170005, -243.883433, -238.558599, -233.197732, -227.803058, -222.376805, -216.921201, -211.538006, -206.326589, -201.083901, -195.805716, -190.487811, -185.12596, -179.715939, -174.253523, -168.734488, -163.154608, -157.509659, -151.795416, -146.007655, -140.142151, -134.19468, -128.161015, -122.036934, -115.178667, -107.915213, -100.583258, -93.1993331, -85.779967, -78.3416895, -70.9010306, -63.4745198, -56.078687, -48.7300619, -41.4451742, -34.2405537, -27.1327302, -20.1382333, -13.2735928, -6.55533844, 0, 2.0932669, 4.50337142, 7.22071899, 10.235715, 13.5387649, 17.1202741, 20.9706479, 25.0802918, 29.4396113, 34.0390116, 38.8688983, 43.9196767, 49.1817522, 54.6455303, 60.3014164, 66.1398158, 71.8587776, 77.406942, 83.4359967, 89.8736994, 96.6478076, 103.686079, 110.916271, 118.266142, 125.663448, 133.035948, 140.3114, 147.41756, 154.282187, 160.753538, 166.86802, 172.577242, 177.83281, 182.912821, 187.605637, 191.612559, 194.787247, 196.983361, 198.054562, 197.854511, 199.508398, 198.602672, 194.350912, 185.966697, 172.663605, 147.346773, 117.115334, 82.7599777, 45.0713935, 4.84027163, -28.0665382, -73.0275903, -141.597313, -239.735162, -361.97752, -507.989499, -677.436208, -871.105301, -1087.28422, -1323.03582, -1560.25013, -1832.83299, -2167.15594, -2579.75468, -3070.49857, -3646.90713, -4318.25398, -5092.51339, -5971.22779, -6959.5778, -8050.81386, -9263.88223, -10608.0932, -12092.7572, -13727.1845, -16003.4279, -18327.3697, -20587.6344, -22672.8469, -24471.6318, -25407.5697, -26144.3588, -26769.9284, -27662.6676, -28795.532, -30020.4523, -31316.4352, -32662.4875, -34037.6162, -35420.828, -35841.4969, -36952.8107, -38484.0445, -40164.4733, -41723.3723, -42890.0165, -43393.6812, -43757.4607, -45638.9499},
        {-377.094882, -379.496162, -381.666689, -383.610327, -385.330937, -386.832381, -388.118521, -389.19322, -390.060338, -390.723738, -391.187283, -391.454834, -391.530253, -391.417402, -391.120143, -390.642338, -389.98785, -389.16054, -388.16427, -387.002902, -385.680298, -384.200321, -382.566832, -380.783693, -378.854766, -376.783914, -374.574998, -372.23188, -369.758422, -367.158487, -364.435936, -361.594631, -358.638435, -355.571208, -352.443279, -349.234298, -345.923361, -342.51282, -339.005032, -335.402349, -331.707126, -327.921718, -324.048477, -320.089759, -316.047917, -311.925306, -307.72428, -303.447193, -299.096398, -294.674252, -290.183106, -285.62267, -280.998001, -276.311511, -271.565611, -266.762713, -261.905228, -256.995567, -252.036144, -247.029367, -241.977651, -236.883405, -231.749042, -226.576973, -221.369609, -216.129362, -210.858644, -205.662298, -200.64271, -195.591031, -190.503033, -185.374487, -180.201164, -174.978835, -169.703272, -164.370246, -158.975528, -153.51489, -147.984102, -142.378936, -136.695164, -130.928555, -125.074883, -119.129917, -112.438178, -105.33476, -98.1625874, -90.9385647, -83.6795978, -76.4025924, -69.1244542, -61.8620887, -54.6324015, -47.4522983, -40.3386848, -33.3084664, -26.3785489, -19.5658379, -12.887239, -6.35965784, 0, 1.45281493, 3.30909871, 5.55145621, 8.1624923, 11.1248118, 14.4210197, 18.0337207, 21.9455199, 26.1390219, 30.5968317, 35.3015542, 40.2357942, 45.3821566, 50.7232463, 56.2416681, 61.920027, 67.6777425, 73.4883932, 79.4775027, 85.6141362, 91.8673588, 98.2062358, 104.599832, 111.017214, 117.427445, 123.799591, 130.102718, 136.305891, 142.378174, 148.519633, 154.383633, 159.86994, 164.878318, 169.426174, 173.354452, 176.472726, 178.645468, 179.73715, 179.612244, 178.135224, 179.180114, 177.381536, 171.819484, 161.573952, 145.724936, 113.099501, 75.5938031, 34.8510689, -7.48547455, -49.7726005, -78.2805007, -118.225018, -183.844508, -282.422783, -407.044795, -557.170493, -732.259829, -932.882491, -1157.13648, -1401.90916, -1649.78649, -1932.58217, -2277.33807, -2699.21436, -3193.23846, -3763.67907, -4411.21523, -5139.185, -5964.08839, -6894.98028, -7952.57416, -9132.47263, -10439.6949, -11879.2601, -13456.1876, -15607.0618, -17797.4453, -19924.4658, -21885.2511, -23576.9292, -24461.0628, -25160.7214, -25759.7035, -26615.3826, -27703.267, -28884.8759, -30141.4172, -31454.0988, -32804.1287, -34172.715, -34407.841, -35503.8044, -37128.5602, -38950.0637, -40636.2701, -41855.1345, -42274.6122, -42676.7178, -45006.7268},
        {-361.453756, -363.990002, -366.28682, -368.348611, -370.179778, -371.784724, -373.167849, -374.333558, -375.286252, -376.030333, -376.570204, -376.910266, -377.054923, -377.008577, -376.775629, -376.360482, -375.767539, -375.001202, -374.065872, -372.965953, -371.705847, -370.289955, -368.72268, -367.008425, -365.151592, -363.156583, -361.0278, -358.769646, -356.386523, -353.882833, -351.262979, -348.531362, -345.692386, -342.750452, -339.794074, -336.78383, -333.679286, -330.482115, -327.193989, -323.816582, -320.351565, -316.800611, -313.165394, -309.447586, -305.64886, -301.770888, -297.815343, -293.783898, -289.678226, -285.499998, -281.250889, -276.89038, -272.46325, -267.972084, -263.419471, -258.807998, -254.140253, -249.418821, -244.646292, -239.825252, -234.958288, -230.047988, -225.096939, -220.107729, -215.082944, -210.025173, -204.937001, -199.928724, -195.108048, -190.255425, -185.366459, -180.436757, -175.461923, -170.437562, -165.35928, -160.222681, -155.023372, -149.756956, -144.419039, -139.005227, -133.511124, -127.932336, -122.264468, -116.503125, -109.972652, -103.018404, -95.9938025, -88.9162363, -81.8030943, -74.6717652, -67.5396379, -60.4241011, -53.3425437, -46.3123546, -39.3509224, -32.475636, -25.7038843, -19.053056, -12.54054, -6.18372506, 0, 0.958456258, 2.37442933, 4.22589843, 6.49084279, 9.14724163, 12.1730742, 15.5463197, 19.2449573, 23.2469663, 27.5303259, 32.0730153, 36.8530137, 41.8483005, 47.0368547, 52.3966556, 57.9056825, 63.5410702, 69.2806775, 75.1043932, 80.9900155, 86.9153428, 92.8581734, 98.7963057, 104.707538, 110.569668, 116.360495, 122.057817, 127.639432, 133.083138, 138.664137, 143.953777, 148.840634, 153.213287, 157.112738, 160.351353, 162.700848, 164.004074, 164.103879, 162.843115, 160.06463, 159.665599, 156.200622, 148.71931, 136.271277, 117.906134, 81.9537721, 40.8634554, -3.63527459, -49.8128754, -95.9398049, -128.517586, -171.969865, -239.027784, -336.542728, -459.361571, -607.720969, -781.857575, -983.187474, -1210.49984, -1461.29719, -1721.46733, -2017.9507, -2375.48769, -2806.96905, -3305.2071, -3872.23048, -4503.2219, -5198.43513, -5983.22573, -6868.75026, -7891.52339, -9033.44234, -10296.8859, -11684.2328, -13197.8619, -15231.5991, -17298.5142, -19303.1241, -21149.9459, -22743.4964, -23579.9208, -24244.3558, -24816.335, -25630.0109, -26664.1303, -27791.3488, -28995.7175, -30261.288, -31572.1116, -32912.2396, -33083.9949, -34190.2067, -35868.7411, -37757.4643, -39494.2427, -40716.9422, -41063.4293, -41605.2793, -44264.6257},
        {-350.461653, -352.552208, -354.440298, -356.129245, -357.622372, -358.923002, -360.034457, -360.960061, -361.703137, -362.267006, -362.654992, -362.870418, -362.916606, -362.796879, -362.514561, -362.072973, -361.475439, -360.725281, -359.825822, -358.780386, -357.592294, -356.264869, -354.801435, -353.205315, -351.47983, -349.628304, -347.654059, -345.560419, -343.350706, -341.028243, -338.596352, -336.058357, -333.417581, -330.677345, -327.893311, -325.041532, -322.097431, -319.062634, -315.938763, -312.727445, -309.430303, -306.048961, -302.585043, -299.040175, -295.415981, -291.714084, -287.936109, -284.08368, -280.158423, -276.16196, -272.095917, -267.917208, -263.673136, -259.366296, -254.999279, -250.574681, -246.095094, -241.563113, -236.98133, -232.352339, -227.678734, -222.963109, -218.208056, -213.41617, -208.590044, -203.732272, -198.845447, -194.0452, -189.44445, -184.812657, -180.145088, -175.437007, -170.683681, -165.880376, -161.022357, -156.10489, -151.12324, -146.072675, -140.948459, -135.745857, -130.460137, -125.086564, -119.620402, -114.05692, -107.692356, -100.886193, -94.006313, -87.0706669, -80.0972051, -73.103878, -66.108636, -59.1294296, -52.1842093, -45.2909255, -38.4675287, -31.7319693, -25.1021978, -18.5961647, -12.2318204, -6.02711531, 0, 0.770106128, 1.98803411, 3.63273103, 5.683144, 8.1182201, 10.9169064, 14.0581501, 17.5208981, 21.2840977, 25.3266958, 29.6276397, 34.1658763, 38.9203528, 43.8700162, 48.9938138, 54.2706925, 59.5067797, 64.6563341, 70.0892043, 75.7473048, 81.5725498, 87.5068536, 93.4921305, 99.4702948, 105.383261, 111.172943, 116.781255, 122.150112, 127.221428, 131.982581, 136.313351, 140.142014, 143.396845, 146.497857, 149.127457, 150.83692, 151.407, 150.618452, 148.252028, 144.088484, 141.116683, 134.932887, 124.690175, 109.541628, 88.6403254, 54.0630552, 13.808262, -31.2019012, -80.0452816, -131.799726, -177.651539, -234.21533, -309.045564, -406.838405, -526.454319, -669.373883, -837.077672, -1031.56275, -1253.67581, -1503.70015, -1773.94451, -2083.77257, -2452.04253, -2888.77304, -3389.00478, -3954.65356, -4582.04766, -5271.65429, -6044.42824, -6909.73536, -7896.09126, -8990.76591, -10196.4964, -11516.0198, -12952.0733, -14876.598, -16830.8259, -18725.193, -20470.1353, -21976.0888, -22767.6866, -23398.3696, -23943.9181, -24719.9282, -25707.2528, -26786.8217, -27944.484, -29166.0888, -30437.4853, -31744.5227, -32005.8743, -33139.1305, -34799.163, -36640.8432, -38319.0426, -39488.6327, -39804.485, -40591.0575, -43638.8317},
        {-341.290358, -342.814612, -344.182012, -345.39449, -346.453976, -347.362404, -348.121706, -348.733813, -349.200657, -349.524172, -349.706288, -349.748937, -349.654053, -349.423566, -349.05941, -348.563516, -347.937815, -347.184241, -346.304725, -345.301199, -344.175596, -342.929847, -341.565884, -340.08564, -338.491047, -336.784036, -334.96654, -333.040491, -331.007821, -328.870461, -326.630345, -324.289403, -321.849569, -319.312773, -316.687023, -313.971086, -311.163655, -308.266466, -305.281254, -302.209753, -299.0537, -295.814827, -292.494871, -289.095567, -285.618649, -282.065852, -278.438912, -274.739563, -270.96954, -267.130578, -263.224413, -259.215688, -255.144034, -251.011988, -246.82209, -242.57688, -238.278895, -233.930676, -229.53476, -225.093688, -220.609997, -216.086228, -211.524918, -206.928608, -202.299835, -197.641139, -192.955059, -188.362297, -183.980719, -179.569161, -175.122501, -170.635618, -166.103391, -161.5207, -156.882423, -152.183439, -147.418627, -142.582866, -137.671035, -132.678013, -127.598679, -122.427911, -117.160589, -111.791592, -105.587681, -98.9231076, -92.1808774, -85.3794977, -78.5374765, -71.6733213, -64.80554, -57.9526404, -51.1331302, -44.3655172, -37.6683092, -31.0600139, -24.559139, -18.1841925, -11.9536819, -5.88611519, 0, 0.63719018, 1.70573567, 3.18605001, 5.05854675, 7.30363944, 9.90174161, 12.8332668, 16.0786286, 19.6182405, 23.4325161, 27.5018689, 31.8067125, 36.3274603, 41.044526, 45.9383231, 50.9892652, 55.8264361, 60.3800595, 65.4252231, 70.8670554, 76.6106851, 82.5612408, 88.6238512, 94.7036448, 100.70575, 106.535297, 112.097412, 117.297225, 122.039864, 126.030598, 129.447698, 132.256251, 134.421343, 136.719332, 138.709669, 139.735468, 139.518433, 137.78027, 134.242685, 128.627383, 122.866577, 113.798702, 100.712974, 82.8986089, 59.6448211, 27.5617335, -10.7125729, -55.2191096, -105.998888, -163.092921, -222.973967, -293.612486, -377.136752, -475.934215, -592.921471, -730.777541, -892.181443, -1079.71688, -1296.17985, -1544.47035, -1823.02104, -2143.93712, -2519.7792, -2957.86523, -3456.62984, -4018.58523, -4643.0254, -5331.62822, -6097.8717, -6948.559, -7898.85278, -8945.91068, -10093.6422, -11345.9569, -12706.7643, -14534.5665, -16390.0324, -18188.423, -19844.9999, -21275.0242, -22027.6203, -22628.278, -23149.1745, -23887.7917, -24827.1238, -25856.2879, -26962.9718, -28134.8636, -29359.6511, -30625.0223, -31004.4718, -32158.7667, -33773.8058, -35535.4876, -37129.7109, -38242.3743, -38559.3764, -39635.7087, -42966.5106},
        {-330.401283, -331.738324, -332.934065, -333.990067, -334.907895, -335.689109, -336.335272, -336.847947, -337.228697, -337.479083, -337.600668, -337.595014, -337.463685, -337.208241, -336.830247, -336.331264, -335.712854, -334.97658, -334.124004, -333.15669, -332.076199, -330.884093, -329.581936, -328.171289, -326.653715, -325.030776, -323.304035, -321.475055, -319.545397, -317.516624, -315.390299, -313.167984, -310.851241, -308.441632, -305.93725, -303.341464, -300.657689, -297.887598, -295.032867, -292.095171, -289.076185, -285.977585, -282.801045, -279.54824, -276.220845, -272.820536, -269.348988, -265.807876, -262.198874, -258.523658, -254.783903, -250.947417, -247.050477, -243.095492, -239.08487, -235.021021, -230.906355, -226.743279, -222.534204, -218.281539, -213.987694, -209.655076, -205.286096, -200.883162, -196.448685, -191.985073, -187.494735, -183.100444, -178.922082, -174.713819, -170.470261, -166.186015, -161.855687, -157.473884, -153.035213, -148.534279, -143.96569, -139.324052, -134.603972, -129.800056, -124.90691, -119.919143, -114.831359, -109.638165, -103.58572, -97.0585963, -90.4503638, -83.7799179, -77.0661536, -70.3279661, -63.5842507, -56.8539025, -50.1558167, -43.5088886, -36.9320132, -30.4440858, -24.0640015, -17.8106556, -11.7029433, -5.75975969, 0, 0.307590643, 1.08807832, 2.31767558, 3.97259497, 6.02904905, 8.46325037, 11.2514115, 14.3697449, 17.7944633, 21.5017791, 25.4679048, 29.6690532, 34.0814366, 38.6812677, 43.4447589, 48.348123, 53.1657348, 57.8449731, 62.8185876, 68.0195399, 73.3807917, 78.8353046, 84.3160403, 89.7559605, 95.0880268, 100.245201, 105.160445, 109.766719, 113.996987, 117.866525, 121.195797, 123.893069, 125.866608, 127.601708, 128.718122, 128.681729, 127.227688, 124.091158, 119.007297, 111.711264, 103.9432, 92.8230686, 77.6937495, 57.8981221, 32.7790656, -0.32672054, -39.568632, -85.1022444, -137.083133, -195.666875, -257.500268, -330.536169, -417.395966, -520.588489, -642.392756, -785.190965, -951.365314, -1142.96144, -1362.7746, -1613.96719, -1896.01977, -2218.68625, -2591.53424, -3020.76861, -3506.8963, -4053.0397, -4660.15674, -5330.80866, -6075.49297, -6900.21796, -7816.46146, -8823.92079, -9926.71095, -11128.947, -12434.7438, -14189.2, -15970.2012, -17696.6166, -19287.3151, -20661.166, -21387.5057, -21967.7578, -22470.5911, -23175.036, -24066.4933, -25042.713, -26093.4785, -27208.573, -28377.7799, -29590.8824, -29940.0299, -31056.6368, -32635.3061, -34370.641, -35957.2445, -37089.7198, -37462.6701, -38805.1868, -42196.5757},
        {-319.712611, -320.979408, -322.110652, -323.107862, -323.972559, -324.706264, -325.310497, -325.786779, -326.13663, -326.36157, -326.463121, -326.442802, -326.302134, -326.042638, -325.665833, -325.173241, -324.566382, -323.846777, -323.015945, -322.075408, -321.026685, -319.871298, -318.610767, -317.246612, -315.780353, -314.213512, -312.547609, -310.784163, -308.924696, -306.970729, -304.923781, -302.785373, -300.557026, -298.240259, -295.8346, -293.342607, -290.766865, -288.108958, -285.370472, -282.552992, -279.658103, -276.68739, -273.642439, -270.524835, -267.336162, -264.078007, -260.751954, -257.359588, -253.902495, -250.38226, -246.800468, -243.128394, -239.39859, -235.613299, -231.774763, -227.885224, -223.946926, -219.962109, -215.933016, -211.861891, -207.750974, -203.602508, -199.418735, -195.201898, -190.95424, -186.678001, -182.375425, -178.170129, -174.182815, -170.165399, -166.112256, -162.017759, -157.876283, -153.682201, -149.429889, -145.113719, -140.728066, -136.267303, -131.725806, -127.097948, -122.378103, -117.560645, -112.639948, -107.610387, -101.70061, -95.303959, -88.8229813, -82.2768997, -75.6849377, -69.0663185, -62.4402654, -55.8260017, -49.2427508, -42.7097359, -36.2461804, -29.8713076, -23.6043408, -17.4645034, -11.4710185, -5.64310964, 0, -0.0467632914, 0.432694108, 1.40909645, 2.85316797, 4.73563292, 7.02721556, 9.69864012, 12.7206309, 16.063912, 19.6992078, 23.5972426, 27.7287405, 32.0644258, 36.5750227, 41.2312556, 46.0038486, 50.8866223, 55.8536003, 60.8232654, 65.771291, 70.6733506, 75.5051177, 80.2422655, 84.8604677, 89.3353976, 93.6427288, 97.7581345, 101.657288, 105.315864, 109.190894, 112.600193, 115.375029, 117.346665, 118.555322, 118.727788, 117.535131, 114.74593, 110.128766, 103.452217, 94.4848631, 85.0270551, 72.197236, 55.3664648, 33.9058005, 7.18630228, -27.8811454, -68.8502655, -115.734955, -168.549112, -227.306633, -287.282093, -359.021218, -447.422099, -555.849811, -684.539509, -835.13162, -1009.26657, -1208.30116, -1434.2239, -1689.3327, -1971.95947, -2291.80674, -2657.13177, -3073.82165, -3543.74727, -4070.62304, -4656.66611, -5305.20271, -6025.04889, -6821.91534, -7706.14045, -8678.09527, -9741.8886, -10901.6292, -12161.426, -13854.9364, -15574.3333, -17241.3383, -18777.6731, -20105.0594, -20809.9363, -21372.8294, -21859.1528, -22530.5228, -23374.6901, -24297.6543, -25291.2536, -26347.3262, -27457.7101, -28614.2437, -28906.6251, -29968.3029, -31502.7146, -33213.2983, -34803.4916, -35976.7324, -36436.4584, -38041.0861, -41377.5298},
        {-309.752308, -310.997801, -312.108907, -313.087257, -313.93448, -314.652203, -315.242056, -315.705668, -316.044668, -316.260685, -316.355347, -316.330284, -316.187124, -315.927497, -315.553031, -315.065355, -314.466098, -313.756889, -312.939357, -312.015131, -310.985839, -309.853111, -308.618576, -307.283862, -305.850599, -304.320415, -302.694939, -300.9758, -299.164627, -297.263049, -295.272694, -293.195193, -291.032173, -288.785264, -286.460496, -284.057205, -281.574673, -279.014386, -276.377831, -273.666493, -270.881858, -268.025413, -265.098643, -262.103035, -259.040075, -255.911248, -252.718041, -249.46194, -246.14443, -242.766999, -239.331132, -235.811709, -232.237399, -228.610264, -224.932368, -221.205773, -217.432543, -213.614739, -209.754425, -205.853664, -201.914518, -197.939051, -193.929324, -189.887401, -185.815345, -181.715219, -177.589084, -173.560805, -169.751317, -165.911483, -162.035469, -158.117442, -154.15157, -150.132019, -146.052956, -141.908548, -137.692963, -133.400367, -129.024928, -124.560813, -120.002188, -115.343221, -110.578079, -105.700929, -99.9243092, -93.6496423, -87.287744, -80.8581466, -74.3803825, -67.8739841, -61.3584839, -54.853414, -48.3783071, -41.9526953, -35.5961112, -29.3280871, -23.1681554, -17.1358484, -11.2506986, -5.53223835, 0, -0.361976439, -0.150571994, 0.599855268, 1.85494728, 3.58034597, 5.74169327, 8.30463111, 11.2348014, 14.4978462, 18.0594072, 21.8851265, 25.9406461, 30.1916077, 34.6036535, 39.1424252, 43.7735649, 48.6831996, 53.8684693, 58.7962992, 63.4795781, 67.9311948, 72.1640381, 76.1909966, 80.0249593, 83.6788147, 87.1654518, 90.4977593, 93.688626, 96.7509406, 100.521549, 103.887265, 106.61379, 108.466828, 109.09604, 108.325147, 106.008815, 101.947557, 95.941887, 87.7923178, 77.2993631, 66.3776182, 52.0700985, 33.7636928, 10.8452899, -17.2982215, -54.5262512, -97.3940373, -145.703116, -199.255025, -257.8513, -315.049858, -384.526946, -473.730505, -587.011593, -722.398471, -880.758205, -1062.95786, -1269.95244, -1502.5011, -1761.26705, -2042.1653, -2356.373, -2712.25878, -3115.95751, -3569.81911, -4077.93084, -4643.2375, -5269.53016, -5964.78897, -6734.62449, -7589.30012, -8529.05779, -9557.89745, -10679.8191, -11898.8227, -13538.3811, -15203.1532, -16817.2707, -18304.8654, -19590.069, -20274.1142, -20819.2979, -21288.1372, -21926.6986, -22724.8442, -23595.9573, -24533.7241, -25531.8308, -26583.9633, -27683.8079, -27938.7601, -28945.7523, -30424.6603, -32095.3598, -33677.7266, -34891.6365, -35456.9652, -37316.6838, -40532.5031},
        {-301.815533, -302.944605, -303.946577, -304.82298, -305.575342, -306.205194, -306.714065, -307.103486, -307.374985, -307.530094, -307.570342, -307.497258, -307.312373, -307.017216, -306.613317, -306.102206, -305.485414, -304.764468, -303.940901, -303.01624, -301.992017, -300.869761, -299.651001, -298.337268, -296.930092, -295.431002, -293.841528, -292.1632, -290.397548, -288.546102, -286.610391, -284.591946, -282.492296, -280.31297, -278.059552, -275.73146, -273.328062, -270.850759, -268.300947, -265.680026, -262.989393, -260.230448, -257.404587, -254.513211, -251.557716, -248.539502, -245.459966, -242.320507, -239.122523, -235.867413, -232.556575, -229.166509, -225.724051, -222.23114, -218.689713, -215.101708, -211.469064, -207.793719, -204.077611, -200.322678, -196.530857, -192.704088, -188.844307, -184.953454, -181.033466, -177.086281, -173.113838, -169.240891, -165.589246, -161.907541, -158.189752, -154.429856, -150.621829, -146.759648, -142.837289, -138.848727, -134.787941, -130.648905, -126.425597, -122.111992, -117.702068, -113.189799, -108.569164, -103.834138, -98.1795797, -92.0158329, -85.7628459, -79.440528, -73.0687884, -66.6675365, -60.2566814, -53.8561325, -47.4857991, -41.1655903, -34.9154156, -28.755184, -22.7048051, -16.7841879, -11.0132418, -5.4118761, 0, -0.348100344, -0.154253078, 0.54881337, 1.72837057, 3.3516901, 5.38604353, 7.79870244, 10.5569384, 13.6280229, 16.9792277, 20.5778242, 24.3910841, 28.3862788, 32.5306801, 36.7915594, 41.1361883, 45.5597656, 50.0335526, 54.4616519, 58.8173198, 63.073812, 67.2043846, 71.1822935, 74.9807948, 78.5731444, 81.9325983, 85.0324125, 87.8458429, 90.3461456, 92.9972966, 95.101901, 96.4859989, 96.9756301, 96.6352709, 95.1717432, 92.2282855, 87.5594069, 80.9196165, 72.0634232, 60.7453362, 48.38286, 32.5613789, 12.7100333, -11.7420364, -41.3656901, -79.0178074, -122.411723, -171.546791, -226.422367, -287.037805, -348.085796, -421.358282, -512.872872, -626.442307, -761.377716, -919.01136, -1100.6755, -1309.19359, -1544.06778, -1803.1735, -2079.48132, -2387.12874, -2737.35058, -3137.20104, -3586.65035, -4088.92042, -4645.8174, -5260.19613, -5940.10251, -6690.64605, -7520.53645, -8430.72923, -9425.08772, -10507.4752, -11681.7551, -13258.7924, -14858.6982, -16408.5856, -17835.5675, -19066.7568, -19718.7666, -20236.2095, -20679.27, -21281.9573, -22035.6499, -22859.2995, -23747.9059, -24696.4691, -25699.9891, -26753.4658, -27105.6424, -28104.766, -29511.9744, -31088.4058, -32595.1981, -33793.4893, -34444.4174, -36566.9857, -39718.7795},
        {-294.951689, -295.961205, -296.850493, -297.620992, -298.274142, -298.811381, -299.234149, -299.543885, -299.742028, -299.830017, -299.809292, -299.681292, -299.447456, -299.109223, -298.668033, -298.125324, -297.482536, -296.741108, -295.90248, -294.96809, -293.939377, -292.817782, -291.604742, -290.301698, -288.910088, -287.431353, -285.86693, -284.218259, -282.486779, -280.67393, -278.781151, -276.809881, -274.761559, -272.637624, -270.443368, -268.178224, -265.841576, -263.434739, -260.959027, -258.415754, -255.806234, -253.131782, -250.393712, -247.593337, -244.731973, -241.810933, -238.831531, -235.795082, -232.7029, -229.556299, -226.356593, -223.081488, -219.75642, -216.383212, -212.963692, -209.499684, -205.993016, -202.445513, -198.859001, -195.235305, -191.576253, -187.883669, -184.159381, -180.405213, -176.622991, -172.814542, -168.981692, -165.25016, -161.742693, -158.205594, -154.632656, -151.017675, -147.354445, -143.63676, -139.858414, -136.013202, -132.094919, -128.097358, -124.014314, -119.839582, -115.566956, -111.19023, -106.703198, -102.099655, -96.5563922, -90.4928892, -84.3384473, -78.1133741, -71.8379775, -65.532565, -59.2174446, -52.9129239, -46.6393106, -40.4169124, -34.2660372, -28.2069926, -22.2600864, -16.4456263, -10.7839201, -5.29527539, 0, -0.304647195, -0.107191383, 0.562142938, 1.67313127, 3.19554911, 5.09917196, 7.35377532, 9.9291347, 12.7950256, 15.9212235, 19.2775039, 22.8336423, 26.5594143, 30.4245953, 34.3989608, 38.4522863, 42.3199567, 45.9382632, 49.8071505, 53.8461675, 57.9748631, 62.1127862, 66.1794859, 70.0945111, 73.7774108, 77.1477338, 80.1250291, 82.6288457, 84.5787326, 85.9254779, 86.5459369, 86.3502868, 85.2487047, 83.8563799, 81.7309836, 78.2421836, 73.0886534, 65.9690667, 56.5820969, 44.6264176, 31.0095336, 13.853382, -7.37987412, -33.2280716, -64.2290474, -102.053847, -145.823797, -195.793431, -252.217285, -315.349893, -381.528236, -459.712524, -553.723809, -666.53442, -799.383874, -954.289684, -1133.26936, -1341.28696, -1576.7438, -1834.82677, -2105.77546, -2406.74343, -2752.27135, -3150.40826, -3598.20351, -4097.75548, -4649.2967, -5254.44179, -5921.64695, -6655.49838, -7463.13002, -8346.18842, -9308.3779, -10353.4028, -11484.9674, -13002.0077, -14539.1885, -16026.406, -17393.5569, -18570.5373, -19187.5083, -19673.9253, -20088.1431, -20655.1025, -21367.2177, -22147.4915, -22991.8085, -23896.0531, -24856.1098, -25867.8633, -26375.6845, -27382.0585, -28703.5394, -30156.6816, -31558.0393, -32724.1667, -33471.6181, -35830.0032, -38902.4158},
        {-289.008112, -289.903051, -290.683809, -291.351755, -291.908259, -292.354689, -292.692415, -292.922806, -293.047231, -293.067058, -292.983658, -292.798398, -292.512649, -292.127779, -291.645158, -291.066153, -290.392136, -289.624474, -288.764537, -287.813694, -286.773314, -285.644765, -284.429418, -283.128641, -281.743804, -280.276275, -278.727423, -277.098618, -275.391229, -273.606624, -271.746174, -269.811246, -267.803211, -265.723436, -263.577396, -261.364321, -259.083392, -256.735846, -254.322918, -251.845844, -249.30586, -246.704203, -244.042107, -241.32081, -238.541546, -235.705552, -232.814064, -229.868317, -226.869548, -223.818992, -220.717886, -217.544988, -214.324499, -211.058141, -207.747638, -204.394712, -201.001088, -197.568489, -194.098636, -190.593255, -187.054067, -183.482796, -179.881166, -176.250898, -172.593717, -168.911346, -165.205508, -161.603034, -158.227752, -154.823358, -151.383466, -147.901689, -144.371639, -140.786929, -137.141173, -133.427983, -129.640972, -125.773752, -121.819938, -117.773142, -113.626976, -109.375054, -105.010988, -100.528392, -95.0853071, -89.1107067, -83.0436077, -76.9047389, -70.714829, -64.494607, -58.2648016, -52.0461416, -45.8593559, -39.7251732, -33.6643223, -27.6975322, -21.8455315, -16.1290492, -10.5688139, -5.18555459, 0, -0.321690602, -0.170022857, 0.426633982, 1.43991066, 2.84143794, 4.60284655, 6.69576725, 9.09183078, 11.7626679, 14.6799093, 17.8151859, 21.1401282, 24.6263671, 28.2455334, 31.9692577, 35.7691709, 39.1280948, 41.9478306, 45.3056477, 49.068432, 53.1030698, 57.2764468, 61.4554493, 65.5069632, 69.2978744, 72.6950692, 75.5654334, 77.7758531, 79.1932144, 79.2641626, 78.4299126, 76.6834226, 74.0176508, 71.5695765, 68.7601474, 64.7052387, 59.0546024, 51.4579902, 41.565154, 29.0258456, 14.3740531, -3.89382329, -26.3010344, -53.3708311, -85.6264643, -123.642307, -167.877708, -218.843137, -277.049064, -343.005959, -414.830301, -498.352544, -594.991742, -706.777983, -836.988876, -988.34192, -1163.55461, -1369.42022, -1603.65417, -1859.52555, -2125.49719, -2420.85825, -2762.87593, -3160.32316, -3607.57985, -4105.63253, -4653.15358, -5250.52957, -5906.63236, -6625.53404, -7412.80626, -8270.79085, -9203.04086, -10213.1093, -11304.5492, -12765.7392, -14244.2002, -15672.279, -16982.3222, -18106.6764, -18686.6868, -19139.7025, -19522.7849, -20055.0498, -20729.4123, -21471.2649, -22276.8988, -23142.6052, -24064.6754, -25039.4005, -25753.1118, -26777.9676, -27999.332, -29302.5691, -30573.0432, -31696.1185, -32557.159, -35107.1494, -38081.8825},
        {-283.726184, -284.535746, -285.234737, -285.824529, -286.306493, -286.682, -286.952422, -287.119131, -287.183497, -287.146892, -287.010688, -286.776255, -286.444966, -286.018191, -285.497303, -284.883671, -284.178669, -283.383666, -282.500036, -281.529148, -280.472374, -279.331086, -278.106656, -276.800454, -275.413851, -273.94822, -272.404932, -270.785357, -269.090868, -267.322836, -265.482632, -263.571628, -261.591195, -259.542704, -257.433659, -255.262237, -253.02654, -250.72774, -248.367009, -245.945521, -243.464446, -240.924959, -238.32823, -235.675433, -232.96774, -230.206323, -227.392355, -224.527007, -221.611453, -218.646864, -215.634414, -212.555276, -209.431054, -206.263355, -203.053784, -199.803946, -196.515449, -193.189897, -189.828896, -186.434054, -183.006974, -179.549264, -176.062528, -172.548374, -169.008407, -165.444232, -161.857456, -158.376051, -155.124956, -151.845155, -148.530061, -145.17309, -141.767654, -138.307169, -134.785047, -131.194703, -127.529551, -123.783006, -119.94848, -116.019389, -111.989145, -107.851164, -103.598859, -99.2256443, -93.8697714, -87.9702227, -81.9762592, -75.9090465, -69.7897503, -63.6395363, -57.4795703, -51.3310178, -45.2150448, -39.1528167, -33.1654994, -27.2742585, -21.5002597, -15.8646688, -10.3886514, -5.09337323, 0, -0.890426295, -1.2222853, -1.03052733, -0.350102682, 0.784038319, 2.33694537, 4.27366816, 6.55925637, 9.1587597, 12.0372278, 15.1597105, 18.4912573, 21.996918, 25.6417422, 29.3907797, 33.2090802, 36.6386013, 39.5839511, 42.967173, 46.662654, 50.5447813, 54.4879421, 58.3665234, 62.0549124, 65.4274963, 68.3586621, 70.7227969, 72.394288, 73.2475225, 72.800239, 71.4142441, 69.0709196, 65.7516471, 62.4348846, 58.6034756, 53.4743766, 46.7298461, 38.052143, 27.1235259, 13.6262537, -1.36930305, -19.9914664, -42.8295647, -70.4729264, -103.51088, -142.487097, -188.047976, -240.794261, -301.326695, -370.246019, -446.603306, -534.443012, -634.083921, -747.013772, -877.106915, -1027.16616, -1199.99432, -1401.69556, -1631.02102, -1883.12038, -2149.26461, -2445.6487, -2787.29384, -3181.85044, -3624.47942, -4116.0746, -4655.45392, -5242.97298, -5886.59889, -6589.99326, -7357.41322, -8191.83323, -9096.70876, -10075.4953, -11131.6482, -12546.0864, -13975.4361, -15353.787, -16615.2288, -17693.851, -18237.9184, -18657.8955, -19010.3682, -19514.5814, -20162.7352, -20879.2555, -21659.7312, -22499.7514, -23394.9048, -24340.7807, -25387.0715, -26452.6268, -27528.6196, -28606.2227, -29676.6091, -30730.9517, -31760.4235, -34391.5133, -37261.8464},
        {-278.9773, -279.735142, -280.383449, -280.92367, -281.357254, -281.685649, -281.910306, -282.032673, -282.0542, -281.976334, -281.800527, -281.528226, -281.160881, -280.699941, -280.146855, -279.503072, -278.770042, -277.949212, -277.042033, -276.049954, -274.974424, -273.816891, -272.578805, -271.261615, -269.86677, -268.39572, -266.849913, -265.230798, -263.539825, -261.778443, -259.9481, -258.050246, -256.08633, -254.057802, -251.976116, -249.83751, -247.638096, -245.379, -243.061346, -240.686258, -238.254861, -235.768278, -233.227635, -230.634056, -227.988664, -225.292586, -222.546943, -219.752862, -216.911466, -214.023881, -211.091229, -208.098438, -205.063181, -201.986933, -198.87117, -195.717368, -192.527002, -189.301547, -186.04248, -182.751274, -179.429407, -176.078353, -172.699589, -169.294588, -165.864828, -162.411783, -158.936929, -155.569525, -152.435764, -149.273574, -146.07615, -142.836681, -139.548359, -136.204378, -132.797928, -129.322201, -125.770389, -122.135683, -118.411276, -114.59036, -110.666125, -106.631764, -102.480469, -98.2054312, -92.9237942, -87.0852003, -81.1498216, -75.1392793, -69.0751945, -62.9791887, -56.8728828, -50.7778983, -44.7158563, -38.7083781, -32.7770849, -26.943598, -21.2295385, -15.6565277, -10.2461869, -5.02013724, 0, -2.00983973, -3.26136403, -3.80464093, -3.68973849, -2.96672474, -1.68566774, 0.103364486, 2.35030388, 5.0050824, 8.017632, 11.3378846, 14.9157723, 18.7012268, 22.6441803, 26.6945647, 30.8023118, 34.9045047, 38.9492396, 42.9155119, 46.7505002, 50.4013831, 53.8153392, 56.9395469, 59.721185, 62.107432, 64.0454665, 65.4824671, 66.3656123, 66.6420807, 66.5306497, 65.6073128, 63.7377691, 60.7877176, 56.8424524, 51.6578748, 44.9313276, 36.4626312, 26.0516059, 13.4980723, -1.39814926, -16.1127191, -34.3995375, -56.9918423, -84.622871, -118.025862, -158.892275, -206.757569, -262.115427, -325.459529, -397.28356, -476.491536, -567.109725, -670.004248, -786.652193, -919.778034, -1071.54619, -1244.12109, -1440.1846, -1661.26608, -1908.33047, -2180.12642, -2484.36378, -2828.59259, -3217.42675, -3650.50503, -4129.74987, -4655.99659, -5230.88578, -5860.04416, -6546.84369, -7294.46516, -8106.50107, -8986.38954, -9937.56866, -10963.4766, -12341.1276, -13731.9896, -15071.1066, -16293.5225, -17334.2814, -17844.0833, -18231.879, -18554.5742, -19037.1929, -19670.094, -20373.5873, -21141.5341, -21967.7959, -22846.2341, -23770.7101, -25248.8797, -26370.087, -27262.0192, -28052.3635, -28868.807, -29839.0369, -31090.7404, -33686.3982, -36439.3707},
        {-274.797779, -275.516946, -276.126773, -276.628801, -277.024576, -277.31564, -277.503535, -277.589807, -277.575997, -277.46365, -277.254308, -276.949515, -276.550814, -276.059748, -275.477861, -274.806696, -274.047796, -273.202705, -272.272965, -271.260121, -270.165715, -268.99129, -267.738391, -266.408559, -265.00334, -263.524275, -261.972908, -260.350782, -258.659442, -256.900429, -255.075287, -253.185561, -251.232791, -249.218524, -247.158499, -245.046863, -242.87759, -240.651762, -238.370463, -236.034774, -233.645778, -231.204558, -228.712196, -226.169774, -223.578376, -220.939083, -218.252978, -215.521143, -212.744662, -209.924617, -207.062089, -204.145106, -201.188089, -198.192404, -195.159417, -192.090491, -188.986995, -185.850292, -182.681748, -179.482729, -176.254601, -172.998729, -169.716478, -166.409214, -163.078303, -159.72511, -156.351001, -153.087075, -150.061319, -147.007529, -143.918661, -140.787671, -137.607513, -134.371144, -131.071518, -127.701591, -124.254319, -120.722655, -117.099557, -113.37798, -109.550878, -105.611207, -101.551922, -97.3659794, -92.1479244, -86.3588141, -80.4705036, -74.5051034, -68.4847238, -62.4314755, -56.3674688, -50.3148143, -44.2956224, -38.3320037, -32.4460685, -26.6599275, -20.9956909, -15.4754695, -10.1213735, -4.95551351, 0, -3.05450294, -5.16555382, -6.39722348, -6.81358276, -6.4787025, -5.45665356, -3.81150677, -1.60733297, 1.09179699, 4.22181226, 7.71864201, 11.5182154, 15.5564616, 19.7693097, 24.0926889, 28.4625284, 33.2159121, 38.346077, 42.8815783, 46.8443069, 50.2561537, 53.1390096, 55.5147655, 57.4053123, 58.8325409, 59.8183422, 60.3846071, 60.5532265, 60.3460914, 60.718811, 60.4171944, 59.183017, 56.7580541, 52.302884, 45.8498805, 37.5864036, 27.4285879, 15.292568, 1.09447844, -15.2495463, -29.8073215, -47.9010389, -70.4003127, -98.1747572, -132.093987, -175.191836, -225.632644, -283.744971, -349.857374, -424.298415, -505.449693, -597.966341, -703.782118, -824.55769, -961.396168, -1115.65099, -1288.67561, -1479.30867, -1691.98994, -1933.90263, -2211.52875, -2523.85411, -2870.73219, -3253.67192, -3676.98732, -4143.70482, -4656.77897, -5219.21758, -5834.29203, -6505.12461, -7233.73882, -8024.52478, -8880.98513, -9806.6225, -10804.9395, -12151.451, -13509.6444, -14815.0193, -16003.0753, -17009.312, -17484.2943, -17838.413, -18129.2828, -18589.2269, -19206.7042, -19897.3786, -20653.349, -21466.714, -22329.5725, -23234.0233, -25084.258, -26245.6505, -26968.8853, -27504.6471, -28103.6203, -29016.4896, -30493.9395, -33008.0647, -35601.8727},
        {-271.120774, -271.80938, -272.388562, -272.859954, -273.225189, -273.485901, -273.643725, -273.700292, -273.657239, -273.516197, -273.278801, -272.946685, -272.521482, -272.004826, -271.398351, -270.70369, -269.922477, -269.056347, -268.106932, -267.075867, -265.964784, -264.775319, -263.509104, -262.167774, -260.752962, -259.266302, -257.709427, -256.083972, -254.391569, -252.633854, -250.812458, -248.929018, -246.985165, -244.982534, -242.940936, -240.852534, -238.709272, -236.512194, -234.262342, -231.960763, -229.608498, -227.206592, -224.756088, -222.258032, -219.713465, -217.123433, -214.488978, -211.811146, -209.090979, -206.329522, -203.527817, -200.67558, -197.785429, -194.858655, -191.896545, -188.900391, -185.871481, -182.811104, -179.720551, -176.601111, -173.454072, -170.280725, -167.082359, -163.860264, -160.615728, -157.350042, -154.064494, -150.892805, -147.965515, -145.010785, -142.02132, -138.989825, -135.909003, -132.771561, -129.570201, -126.297629, -122.946549, -119.509666, -115.979684, -112.349307, -108.611241, -104.758189, -100.782857, -96.6779477, -91.5136844, -85.7634158, -79.9115783, -73.9808127, -67.9937601, -61.9730611, -55.9413567, -49.9212877, -43.9354949, -38.0066193, -32.1573015, -26.4101826, -20.7879034, -15.3131047, -10.0084273, -4.89651211, 0, -3.78958061, -6.51318011, -8.24420583, -9.05606505, -9.02216508, -8.21591323, -6.71071681, -4.57998311, -1.89711944, 1.26446689, 4.83136858, 8.73017832, 12.8874888, 17.2298928, 21.6839828, 26.1763518, 31.340165, 37.2029543, 42.0931118, 46.0886389, 49.2675365, 51.7078059, 53.4874482, 54.6844646, 55.3768562, 55.6426243, 55.5597698, 55.206294, 54.6601981, 55.4614506, 55.6900305, 54.9853489, 52.9868165, 48.12804, 40.6513324, 31.1205549, 19.5368597, 5.90139905, -9.78467469, -27.5202093, -42.3673665, -60.7641496, -83.6752797, -112.065478, -146.899464, -192.44998, -245.546723, -306.327407, -374.929751, -451.491469, -533.612767, -627.070276, -735.179908, -859.936225, -1000.63602, -1157.78734, -1331.89821, -1518.27464, -1723.80897, -1961.06855, -2242.92447, -2561.22807, -2909.62506, -3287.09976, -3701.68252, -4157.2514, -4658.42319, -5209.26747, -5811.14524, -6466.94962, -7177.34031, -7947.78747, -8781.95734, -9683.51615, -10656.1301, -11976.0746, -13306.2545, -14582.1837, -15739.3762, -16713.346, -17152.5314, -17470.9059, -17727.0157, -18160.2665, -18757.6144, -19430.658, -20170.1216, -20966.7294, -21811.2057, -22694.2748, -24780.0455, -25950.2208, -26544.4693, -26902.4599, -27363.8613, -28268.3423, -29955.5716, -32364.6679, -34742.8035},
        {-267.841247, -268.502437, -269.054591, -269.4994, -269.838558, -270.073756, -270.206687, -270.239043, -270.172516, -270.008799, -269.749584, -269.396563, -268.951428, -268.415873, -267.791588, -267.080267, -266.283602, -265.403285, -264.441008, -263.398463, -262.277343, -261.079341, -259.806148, -258.459457, -257.04096, -255.552349, -253.995317, -252.371555, -250.682757, -248.930615, -247.11682, -245.243065, -243.311043, -241.322446, -239.300414, -237.235465, -235.117858, -232.948589, -230.728653, -228.459047, -226.140768, -223.774812, -221.362174, -218.903851, -216.40084, -213.854136, -211.264736, -208.633636, -205.961832, -203.250321, -200.500098, -197.698722, -194.860918, -191.987974, -189.081178, -186.141817, -183.171179, -180.17055, -177.141219, -174.084474, -171.001601, -167.893889, -164.762625, -161.609095, -158.434589, -155.240394, -152.027796, -148.934841, -146.096288, -143.231375, -140.332525, -137.392159, -134.4027, -131.35657, -128.24619, -125.063984, -121.802374, -118.453781, -115.010628, -111.465336, -107.810329, -104.038028, -100.140856, -96.1112342, -90.9915862, -85.2693155, -79.4432442, -73.5366237, -67.5727054, -61.5747409, -55.5659816, -49.569679, -43.6090846, -37.7074498, -31.8880262, -26.1740651, -20.5888181, -15.1555366, -9.89747207, -4.83787606, 0, -3.73213075, -6.43779499, -8.18768217, -9.05248171, -9.10288306, -8.40957565, -7.04324892, -5.07459229, -2.57429521, 0.386952893, 3.73846258, 7.40954441, 11.329509, 15.4276668, 19.6333285, 23.8758045, 28.7698241, 34.3426145, 38.9731351, 42.7375719, 45.7121108, 47.9729378, 49.5962388, 50.6581997, 51.2350065, 51.4028452, 51.2379017, 50.8163619, 50.2144118, 51.0532578, 51.2975578, 50.5599918, 48.4532395, 43.310705, 35.3847059, 25.2687001, 12.9591503, -1.54748109, -18.2547314, -37.1661382, -53.4184692, -73.3632223, -97.956129, -128.152921, -164.909329, -213.331699, -269.187495, -332.394796, -402.87168, -480.536226, -561.810439, -654.381449, -763.338954, -891.247406, -1035.51554, -1195.8669, -1372.02503, -1557.02381, -1758.79684, -1992.57554, -2273.52443, -2591.21526, -2937.74633, -3311.34314, -3720.61338, -4169.39914, -4662.44453, -5203.82549, -5794.31059, -6436.53921, -7129.10473, -7879.31505, -8691.21015, -9568.82999, -10516.2145, -11814.0108, -13120.5, -14370.5702, -15499.1095, -16441.0062, -16839.8545, -17116.0325, -17329.2686, -17724.1089, -18287.125, -18927.8808, -19636.7981, -20404.2988, -21220.8046, -22076.7374, -24129.6387, -25248.8783, -25795.2504, -26129.549, -26612.5683, -27605.1024, -29467.9454, -31793.8094, -33854.9813},
        {-264.87853, -265.510504, -266.03493, -266.453505, -266.767924, -266.979883, -267.091078, -267.103205, -267.01796, -266.837038, -266.562136, -266.194949, -265.737174, -265.190506, -264.556641, -263.837275, -263.034104, -262.148823, -261.18313, -260.138719, -259.017286, -257.820528, -256.55014, -255.207818, -253.795258, -252.314157, -250.766209, -249.15311, -247.476558, -245.738246, -243.939873, -242.083132, -240.169721, -238.201335, -236.203164, -234.164664, -232.075006, -229.935123, -227.74595, -225.508418, -223.223462, -220.892016, -218.515011, -216.093383, -213.628065, -211.119989, -208.57009, -205.9793, -203.348553, -200.678784, -197.970924, -195.205168, -192.40364, -189.56772, -186.698793, -183.798242, -180.86745, -177.907798, -174.920672, -171.907453, -168.869524, -165.80827, -162.725072, -159.621313, -156.498377, -153.357647, -150.200505, -147.171504, -144.411756, -141.627372, -138.810452, -135.953099, -133.047416, -130.085503, -127.059464, -123.961399, -120.783412, -117.517603, -114.156076, -110.690931, -107.114271, -103.418199, -99.5948152, -95.6362225, -90.5527863, -84.8481261, -79.0376617, -73.1453542, -67.1951649, -61.2110549, -55.2169856, -49.2369182, -43.2948139, -37.414634, -31.6203396, -25.9358921, -20.3852526, -14.9923824, -9.78124276, -4.77579488, 0, -2.5575157, -4.35673792, -5.44864386, -5.88421067, -5.71441555, -4.99023566, -3.76264818, -2.08263029, -0.00115917075, 2.43078801, 5.16223407, 8.14220183, 11.3197141, 14.6437938, 18.0634636, 21.5277464, 25.1727736, 28.9742977, 32.45812, 35.6133579, 38.4291289, 40.8945505, 42.9987403, 44.7308157, 46.0798943, 47.0350936, 47.585531, 47.7203242, 47.4285906, 47.6626124, 47.0951821, 45.4264677, 42.3566372, 37.0472498, 29.467778, 19.7313235, 7.699637, -6.7655306, -23.8024286, -43.5493061, -62.7704119, -86.0048654, -114.051046, -147.707335, -187.77211, -239.480598, -298.085121, -363.274848, -434.738946, -512.166585, -590.429141, -679.921986, -787.796846, -917.42705, -1064.4717, -1228.05344, -1407.29489, -1595.0289, -1798.0974, -2030.20386, -2302.69389, -2610.06304, -2949.76429, -3321.8731, -3730.90869, -4179.35053, -4669.7884, -5204.73037, -5786.27998, -6416.76959, -7091.76399, -7821.40427, -8610.3654, -9463.32237, -10384.9501, -11664.2758, -12950.5341, -14177.3122, -15278.1969, -16186.7754, -16539.1867, -16764.7641, -16924.5721, -17264.6237, -17772.7747, -18359.8191, -19017.4659, -19737.4239, -20511.4022, -21331.1097, -22986.055, -23974.2239, -24584.3489, -25105.1623, -25825.3966, -27033.7842, -29019.0575, -31314.3024, -32931.628},
        {-262.386706, -262.982958, -263.473745, -263.860735, -264.145599, -264.330007, -264.415631, -264.40414, -264.297205, -264.096496, -263.803684, -263.420439, -262.948431, -262.389331, -261.744809, -261.016536, -260.206181, -259.315416, -258.345911, -257.299336, -256.177362, -254.981658, -253.713896, -252.375746, -250.968878, -249.494962, -247.95567, -246.35267, -244.687635, -242.962234, -241.178137, -239.337016, -237.440539, -235.490379, -233.511791, -231.494156, -229.426572, -227.309945, -225.145179, -222.933179, -220.674851, -218.371098, -216.022827, -213.630943, -211.196349, -208.719952, -206.202655, -203.645365, -201.048987, -198.414424, -195.742582, -193.009976, -190.242429, -187.441375, -184.608249, -181.744484, -178.851514, -175.930773, -172.983694, -170.011711, -167.016259, -163.998769, -160.960677, -157.903417, -154.828421, -151.737124, -148.630959, -145.659341, -142.968133, -140.253568, -137.507486, -134.721729, -131.888136, -128.99855, -126.044811, -123.018758, -119.912234, -116.717079, -113.425133, -110.028238, -106.518234, -102.886961, -99.1262609, -95.2279741, -90.1746519, -84.4840088, -78.6858901, -72.8048899, -66.8656024, -60.8926217, -54.910542, -48.9439572, -43.0174616, -37.1556492, -31.3831143, -25.7244508, -20.2042529, -14.8471148, -9.67763051, -4.72039421, 0, -1.46627225, -2.42148276, -2.89854648, -2.9303784, -2.54989346, -1.79000664, -0.683632909, 0.736312779, 2.43691545, 4.38526015, 6.54843191, 8.89351575, 11.3875967, 13.9977599, 16.6910902, 19.4346727, 21.9060453, 24.0309289, 26.4313371, 29.012309, 31.6788838, 34.3361008, 36.888999, 39.2426177, 41.3019961, 42.9721733, 44.1581885, 44.765081, 44.6978898, 44.240975, 42.7809707, 40.1091198, 36.0166653, 30.4954866, 23.2365084, 13.8771523, 2.14847044, -12.2184853, -29.4926629, -49.9430103, -72.0073257, -98.3430133, -129.57729, -166.337372, -209.250475, -262.699804, -322.617592, -388.692056, -460.611418, -538.063895, -615.584875, -704.313319, -811.908876, -941.999012, -1089.97881, -1254.93954, -1435.97244, -1627.24775, -1833.89616, -2066.4168, -2331.28589, -2628.64191, -2961.4772, -3331.87066, -3740.3535, -4188.16734, -4675.90982, -5204.65558, -5777.84063, -6397.56524, -7057.4216, -7769.32705, -8538.32696, -9369.46674, -10267.7918, -11525.9202, -12789.4316, -13991.478, -15065.2117, -15943.7848, -16261.4043, -16449.4645, -16569.2373, -16862.7197, -17319.6462, -17854.4491, -18460.6281, -19131.6827, -19861.1127, -20642.4177, -21776.7266, -22609.378, -23320.5886, -24090.5755, -25099.5553, -26527.7451, -28555.3616, -30763.9469, -31969.8577},
        {-260.487583, -261.02029, -261.451403, -261.7825, -262.015159, -262.150959, -262.191476, -262.13829, -261.992978, -261.757119, -261.43229, -261.020071, -260.522038, -259.93977, -259.274845, -258.528841, -257.703336, -256.799909, -255.820137, -254.765599, -253.637873, -252.438536, -251.169167, -249.831344, -248.426645, -246.956648, -245.422932, -243.827074, -242.170652, -240.455245, -238.682431, -236.853788, -234.970893, -233.035325, -231.066945, -229.057804, -226.999736, -224.893725, -222.740756, -220.541814, -218.297884, -216.009951, -213.678998, -211.306012, -208.891977, -206.437878, -203.944699, -201.413426, -198.845043, -196.240535, -193.600887, -190.913528, -188.193293, -185.441459, -182.659307, -179.848114, -177.009158, -174.14372, -171.253076, -168.338507, -165.40129, -162.442704, -159.464029, -156.466542, -153.451522, -150.420248, -147.373999, -144.46073, -141.824877, -139.165205, -136.473487, -133.741491, -130.960988, -128.123749, -125.221543, -122.246141, -119.189313, -116.042828, -112.798458, -109.447971, -105.98314, -102.395733, -98.6775202, -94.8202728, -89.7932611, -84.1210119, -78.3403416, -72.4762025, -66.5535468, -60.5973267, -54.6324945, -48.6840024, -42.7768028, -36.9358477, -31.1860895, -25.5524804, -20.0599727, -14.7335185, -9.59807018, -4.67857993, 0, -1.2767387, -2.0956972, -2.4863853, -2.47831283, -2.1009896, -1.38392544, -0.356630142, 0.95138646, 2.51061456, 4.29154433, 6.26466596, 8.40046964, 10.6694455, 13.0420839, 15.4888748, 17.9803085, 20.0654529, 21.654595, 23.671442, 25.9961794, 28.5089924, 31.0900666, 33.6195873, 35.9777401, 38.0447102, 39.7006831, 40.8258443, 41.3003791, 41.0044729, 40.0153648, 37.9439335, 34.6112487, 29.8383795, 23.8600194, 16.2904258, 6.63355626, -5.41360704, -20.154082, -37.8908863, -58.927038, -82.4731351, -110.257091, -142.795657, -180.605586, -224.20363, -274.733633, -331.928481, -396.148156, -467.752635, -547.101897, -632.240703, -728.673965, -839.730274, -967.471437, -1111.37291, -1272.07137, -1450.20348, -1644.52955, -1857.98907, -2095.56852, -2359.60005, -2653.52045, -2982.21565, -3348.60729, -3752.28855, -4194.38046, -4675.27011, -5195.88825, -5759.85654, -6369.2745, -7019.45871, -7720.33515, -8476.35129, -9291.95459, -10171.5925, -11393.9366, -12620.6543, -13787.6369, -14830.7758, -15685.9625, -16004.171, -16196.155, -16319.9137, -16599.7072, -17027.7227, -17526.8717, -18092.8051, -18721.1739, -19407.6288, -20147.8209, -20642.7833, -21294.6081, -22134.4315, -23193.3896, -24502.6187, -26093.255, -27996.4346, -29947.806, -30945.564},
        {-259.058751, -259.520319, -259.884429, -260.152556, -260.326173, -260.406753, -260.395771, -260.2947, -260.105014, -259.828187, -259.465691, -259.019002, -258.489592, -257.878936, -257.188506, -256.419777, -255.574223, -254.653317, -253.658532, -252.591343, -251.453223, -250.245646, -248.970085, -247.628015, -246.220908, -244.750239, -243.217482, -241.624109, -239.971595, -238.261413, -236.495038, -234.673942, -232.7996, -230.873485, -228.90834, -226.899768, -224.843235, -222.739849, -220.590717, -218.396947, -216.159648, -213.879928, -211.558894, -209.197654, -206.797317, -204.358989, -201.88378, -199.372798, -196.827149, -194.247943, -191.636286, -188.996711, -186.326828, -183.62767, -180.900271, -178.145665, -175.364887, -172.558968, -169.728944, -166.875848, -164.000714, -161.104575, -158.188466, -155.253419, -152.300469, -149.33065, -146.344995, -143.487119, -140.897214, -138.28225, -135.634019, -132.944311, -130.204916, -127.407627, -124.544233, -121.606526, -118.586295, -115.475333, -112.26543, -108.948376, -105.515962, -101.95998, -98.272219, -94.4444714, -89.4401644, -83.7869006, -78.0246426, -72.1785535, -66.2737964, -60.3355346, -54.3889311, -48.4591492, -42.5713519, -36.7507024, -31.022364, -25.4114996, -19.9432725, -14.6428459, -9.53538279, -4.64604646, 0, -1.65545421, -2.78493465, -3.42371181, -3.60705619, -3.37023828, -2.74852857, -1.77719755, -0.491515711, 1.07324645, 2.88181845, 4.89892978, 7.08930996, 9.41768851, 11.8487949, 14.3473587, 16.8781093, 19.0626148, 20.8165813, 22.8809375, 25.1468782, 27.5055983, 29.8482925, 32.0661559, 34.0503833, 35.6921695, 36.8827094, 37.5131979, 37.4748299, 36.6588002, 35.2059057, 32.6662186, 28.856053, 23.5917234, 17.0280483, 8.81208997, -1.49935759, -14.1915314, -29.5496686, -47.8590063, -69.4047814, -93.6564348, -121.963286, -154.770184, -192.521978, -235.663518, -281.122545, -333.740293, -394.84089, -465.748462, -547.787135, -644.188852, -752.038149, -868.256975, -992.549767, -1130.30188, -1284.34807, -1457.52305, -1654.34271, -1875.57859, -2120.16828, -2387.26769, -2681.83092, -3008.30918, -3369.68966, -3766.47895, -4200.32225, -4672.4503, -5184.40086, -5739.23151, -6339.14013, -6982.77635, -7676.43219, -8423.53379, -9227.50726, -10091.7787, -11273.5002, -12458.9405, -13588.0942, -14600.956, -15437.5206, -15773.5588, -15989.4382, -16138.3923, -16421.2996, -16832.9505, -17306.6169, -17840.1653, -18431.4621, -19078.3739, -19778.7672, -19650.4321, -20132.2661, -21101.2192, -22434.2414, -24008.2829, -25700.2935, -27387.2235, -29004.5191, -29882.7371},
        {-257.975999, -258.379547, -258.688463, -258.904161, -259.028055, -259.061558, -259.006084, -258.863048, -258.633863, -258.319942, -257.922699, -257.443549, -256.883905, -256.24518, -255.528789, -254.736145, -253.868662, -252.927754, -251.914835, -250.831318, -249.678617, -248.458146, -247.171318, -245.819548, -244.40425, -242.926836, -241.388721, -239.791319, -238.136043, -236.424307, -234.657525, -232.837111, -230.964478, -229.04104, -227.074546, -225.063108, -223.004764, -220.900723, -218.752191, -216.560379, -214.326493, -212.051742, -209.737334, -207.384477, -204.99438, -202.568251, -200.107298, -197.61273, -195.085753, -192.527577, -189.939411, -187.340552, -184.713727, -182.05975, -179.379439, -176.673609, -173.943076, -171.188657, -168.411166, -165.611421, -162.790238, -159.948431, -157.086819, -154.206215, -151.307437, -148.391301, -145.458622, -142.649265, -140.099755, -137.524133, -134.9142, -132.261758, -129.558609, -126.796556, -123.9674, -121.062943, -118.074989, -114.995337, -111.815792, -108.528155, -105.124228, -101.595813, -97.9347116, -94.1327268, -89.1477341, -83.5102458, -77.7632017, -71.9319565, -66.0418653, -60.1182832, -54.186565, -48.2720658, -42.4001406, -36.5961443, -30.8854319, -25.2933585, -19.845279, -14.5665485, -9.48252176, -4.61855394, 0, -2.2752771, -3.90611652, -4.93731821, -5.41368211, -5.38000817, -4.88109632, -3.96174651, -2.66675868, -1.04093277, 0.870931267, 3.02403349, 5.37357397, 7.87475274, 10.4827699, 13.1528254, 15.8401195, 18.3123453, 20.4979163, 22.7761549, 25.0620811, 27.2707151, 29.3170768, 31.1161863, 32.5830637, 33.6327288, 34.1802019, 34.140503, 33.428652, 31.959669, 30.0205312, 27.0179172, 22.75526, 17.0359925, 9.87576239, 0.971895175, -10.0348742, -23.4047771, -39.3980452, -58.2749098, -80.2956026, -105.05754, -133.685495, -166.56938, -204.099107, -246.66459, -287.430043, -335.817501, -394.023302, -464.243783, -548.675281, -655.225081, -773.398526, -894.429824, -1015.91303, -1148.4269, -1296.72032, -1465.54217, -1664.20099, -1891.84974, -2142.66725, -2413.89452, -2710.73572, -3036.15826, -3392.80559, -3782.77208, -4208.40414, -4672.11149, -5176.25693, -5722.97119, -6314.51628, -6952.36876, -7639.69693, -8379.03468, -9172.91592, -10023.8745, -11170.0757, -12319.5142, -13415.8162, -14402.6077, -15223.5148, -15576.3117, -15814.3777, -15986.7039, -16273.3606, -16671.4535, -17123.4722, -17629.2534, -18188.6338, -18801.4499, -19467.5385, -18871.575, -19231.7653, -20302.3185, -21837.4438, -23591.3501, -25318.2467, -26772.3426, -28073.9986, -28806.6574},
        {-256.970622, -257.382003, -257.695879, -257.91381, -258.037359, -258.068087, -258.007554, -257.857321, -257.618951, -257.294003, -256.884039, -256.390621, -255.815309, -255.159665, -254.42525, -253.613624, -252.726349, -251.764987, -250.731097, -249.626243, -248.451984, -247.209881, -245.901497, -244.528391, -243.092126, -241.594262, -240.036361, -238.419983, -236.746691, -235.018044, -233.235604, -231.400933, -229.51559, -227.581139, -225.612518, -223.604318, -221.550968, -219.453595, -217.313327, -215.13129, -212.908612, -210.646419, -208.345838, -206.007996, -203.634021, -201.225039, -198.782177, -196.306562, -193.799322, -191.261582, -188.694471, -186.111288, -183.500724, -180.86364, -178.200901, -175.51337, -172.801909, -170.067382, -167.310652, -164.532581, -161.734034, -158.915872, -156.07896, -153.22416, -150.352335, -147.464349, -144.561065, -141.786318, -139.280514, -136.749643, -134.1853, -131.579079, -128.922574, -126.207382, -123.425096, -120.567311, -117.625622, -114.591623, -111.456909, -108.213076, -104.851716, -101.364426, -97.7427999, -93.9784323, -89.0097185, -83.3770054, -77.633299, -71.8043721, -65.9159979, -59.9939494, -54.0639996, -48.1519216, -42.2834884, -36.484473, -30.7806484, -25.1977879, -19.7616642, -14.4980506, -9.43271996, -4.59144542, 0, -2.7481119, -4.77044878, -6.11842728, -6.84346403, -6.99697568, -6.63037885, -5.79509019, -4.54252633, -2.9241039, -0.991239541, 1.20465011, 3.61214842, 6.17983875, 8.85630446, 11.5901289, 14.3298955, 16.8873402, 19.1914969, 21.5004844, 23.7335617, 25.8099878, 27.6490215, 29.1699219, 30.2919479, 30.9343584, 31.0164123, 30.4573686, 29.1764862, 27.093024, 24.4375119, 20.7038052, 15.7177817, 9.30531914, 1.72876494, -7.4042381, -18.6024391, -32.1709014, -48.4146883, -67.6388629, -90.1484885, -116.143106, -146.065416, -180.241129, -218.995953, -262.655596, -304.082323, -352.931149, -411.393645, -481.661381, -565.925927, -671.787881, -789.418753, -910.888967, -1034.06017, -1168.61776, -1318.93852, -1489.39921, -1688.21182, -1915.04627, -2165.38865, -2437.8238, -2735.99052, -3061.46088, -3416.30729, -3803.44998, -4225.42, -4685.12636, -5185.1981, -5726.8785, -6312.19471, -6940.38596, -7616.69624, -8344.11774, -9125.64265, -9964.26318, -11104.1205, -12246.7706, -13334.9184, -14311.2689, -15118.527, -15449.6788, -15663.6274, -15810.1003, -16069.7329, -16440.4349, -16865.5708, -17345.7775, -17881.6919, -18473.9508, -19123.1911, -18595.3122, -18987.3805, -20055.7095, -21556.6126, -23246.4035, -24881.3955, -26217.9023, -27443.6947, -27810.2081},
        {-256.262747, -256.69641, -257.028446, -257.260603, -257.39463, -257.432274, -257.375285, -257.225411, -256.984399, -256.653999, -256.235958, -255.732025, -255.143949, -254.473477, -253.722358, -252.892341, -251.985173, -251.002604, -249.946381, -248.818253, -247.619967, -246.353274, -245.01992, -243.621655, -242.160226, -240.637382, -239.054871, -237.414441, -235.717842, -233.966821, -232.163127, -230.308507, -228.404711, -226.453487, -224.47926, -222.471962, -220.421255, -218.32815, -216.19366, -214.018798, -211.804576, -209.552007, -207.262102, -204.935875, -202.574338, -200.178503, -197.749382, -195.287989, -192.795336, -190.272434, -187.720297, -185.141425, -182.535311, -179.902933, -177.245274, -174.563312, -171.858028, -169.130402, -166.381414, -163.612044, -160.823274, -158.016082, -155.191449, -152.350356, -149.493782, -146.622707, -143.738113, -140.989638, -138.523355, -136.033643, -133.511844, -130.949304, -128.337363, -125.667366, -122.930656, -120.118576, -117.222469, -114.233678, -111.143546, -107.943417, -104.624633, -101.178539, -97.5964757, -93.869788, -88.9149765, -83.2834133, -77.5393892, -71.7091279, -65.8188533, -59.8947892, -53.9631596, -48.0501882, -42.1820989, -36.3851155, -30.6854618, -25.1093618, -19.6830392, -14.4327179, -9.38462166, -4.56497443, 0, -3.241389, -5.67276198, -7.3523416, -8.3383505, -8.68901133, -8.46254675, -7.71717939, -6.51113191, -4.90262695, -2.94988718, -0.711135225, 1.75540625, 4.39151461, 7.13896719, 9.93954136, 12.7350144, 15.3860765, 17.822921, 20.170737, 22.353926, 24.2968897, 25.9240293, 27.1597466, 27.9284428, 28.1545196, 27.7623785, 26.6764208, 24.8210482, 22.1206621, 18.7520353, 14.2946619, 8.59723193, 1.50843555, -6.42888026, -15.7131039, -27.0277321, -40.7323218, -57.1864301, -76.7496139, -99.7814302, -127.092663, -158.454312, -194.137651, -234.413953, -279.554492, -323.196933, -373.904559, -433.607046, -504.234069, -587.715303, -689.922102, -804.024856, -924.754455, -1050.9049, -1189.56571, -1344.10189, -1517.87845, -1716.35201, -1940.32057, -2188.30034, -2461.18163, -2760.46101, -3086.22648, -3439.89222, -3825.11955, -4244.53831, -4701.47936, -5198.75431, -5736.60434, -6316.72464, -6934.96057, -7599.75705, -8314.83418, -9083.91204, -9910.71073, -11050.4185, -12192.4203, -13277.569, -14246.7179, -15040.7197, -15342.0792, -15522.2302, -15632.7466, -15859.727, -16201.4959, -16600.3257, -17057.3437, -17573.6768, -18150.4522, -18788.7969, -18523.1626, -19002.6579, -20030.4417, -21409.6724, -22943.5089, -24435.1097, -25687.6335, -26873.8728, -26779.6023},
        {-255.903863, -256.358779, -256.708099, -256.953743, -257.097635, -257.141695, -257.087846, -256.938009, -256.694107, -256.35806, -255.931792, -255.417223, -254.816276, -254.130872, -253.362934, -252.514383, -251.587141, -250.58313, -249.504271, -248.352488, -247.1297, -245.837831, -244.478802, -243.054535, -241.566951, -240.017974, -238.409524, -236.743523, -235.021893, -233.246556, -231.419435, -229.54245, -227.617523, -225.646577, -223.662807, -221.65184, -219.598928, -217.504977, -215.370895, -213.197588, -210.985963, -208.736927, -206.451387, -204.130251, -201.774425, -199.384815, -196.96233, -194.507876, -192.022359, -189.506688, -186.961768, -184.3786, -181.768213, -179.131727, -176.470266, -173.78495, -171.076902, -168.347242, -165.597094, -162.827577, -160.039815, -157.234929, -154.414041, -151.578271, -148.728743, -145.866577, -142.992896, -140.263576, -137.830778, -135.376452, -132.89169, -130.367581, -127.795214, -125.165678, -122.470064, -119.699461, -116.844959, -113.897646, -110.848613, -107.68895, -104.409746, -101.00209, -97.4570718, -93.7657817, -88.8230502, -83.1911546, -77.4455483, -71.6128966, -65.7198647, -59.7931181, -53.8593219, -47.9451415, -42.0772421, -36.2822891, -30.5869477, -25.0178832, -19.601761, -14.3652462, -9.3350043, -4.53770046, 0, -3.73089304, -6.5685551, -8.57793107, -9.82396588, -10.3716044, -10.2857916, -9.63147224, -8.47359135, -6.8770938, -4.90692448, -2.6280283, -0.105350148, 2.59616506, 5.41157243, 8.27592705, 11.124284, 13.8806658, 16.4785513, 18.8779506, 21.0115545, 22.8120542, 24.2121404, 25.1445041, 25.5418364, 25.3368282, 24.4621703, 22.8505538, 20.4346697, 17.1472088, 13.1499588, 8.06251192, 1.74882998, -5.92712508, -14.1852465, -23.6196447, -35.0687356, -48.9434006, -65.6545213, -85.6129792, -109.229656, -137.839813, -170.648619, -207.886098, -249.782275, -296.567174, -343.207553, -396.51252, -458.027917, -529.299585, -611.873365, -709.258199, -818.637484, -938.151051, -1067.80298, -1211.40351, -1371.05401, -1548.85582, -1746.92298, -1967.34129, -2212.1827, -2484.78911, -2784.56008, -3110.28721, -3462.79375, -3846.34549, -4263.62803, -4718.36098, -5213.49802, -5748.2015, -6323.77836, -6932.51074, -7586.11881, -8289.03347, -9045.68564, -9860.50622, -10999.5102, -12140.1484, -13221.4557, -14182.4669, -14962.2171, -15232.5767, -15377.8545, -15451.5844, -15645.9194, -15959.5232, -16333.3017, -16768.8065, -17267.5893, -17831.2016, -18461.1952, -18559.3492, -19153.93, -20119.4333, -21330.3548, -22661.1902, -23986.4354, -25180.5859, -26306.667, -25672.3501},
        {-255.940798, -256.367339, -256.68915, -256.908132, -257.026183, -257.045204, -256.967092, -256.793749, -256.527072, -256.168962, -255.721318, -255.186039, -254.565025, -253.860175, -253.073388, -252.206564, -251.261602, -250.240401, -249.144861, -247.976882, -246.738362, -245.431201, -244.057298, -242.618553, -241.116866, -239.554134, -237.932259, -236.253138, -234.518672, -232.730761, -230.891302, -229.002196, -227.065342, -225.08264, -223.085069, -221.059377, -218.991962, -216.883778, -214.735783, -212.548932, -210.324179, -208.062482, -205.764795, -203.432075, -201.065276, -198.665355, -196.233267, -193.769968, -191.276414, -188.75356, -186.202361, -183.615067, -181.001528, -178.36289, -175.700297, -173.014892, -170.307821, -167.580228, -164.833258, -162.068054, -159.285761, -156.487523, -153.674485, -150.847792, -148.008588, -145.158016, -142.297222, -139.584351, -137.173776, -134.742842, -132.282515, -129.783762, -127.23755, -124.634845, -121.966614, -119.223824, -116.397442, -113.478434, -110.457767, -107.326407, -104.075322, -100.695478, -97.1778421, -93.5133806, -88.582728, -82.9553714, -77.2141437, -71.3860438, -65.4980711, -59.5772246, -53.6505035, -47.7449069, -41.8874339, -36.1050838, -30.4248556, -24.8737486, -19.4787618, -14.2668944, -9.2651455, -4.50051433, 0, -4.2590713, -7.53368175, -9.89613625, -11.4187397, -12.1737969, -12.2336129, -11.6704924, -10.5567405, -8.96466193, -6.96656163, -4.6347445, -2.04151541, 0.740820739, 3.63995906, 6.58359466, 9.49942265, 12.4842037, 15.4897852, 18.061452, 20.1631275, 21.7587352, 22.8121986, 23.287441, 23.1483859, 22.3589569, 20.8830773, 18.6846705, 15.7276601, 11.9759694, 7.84675447, 2.68452095, -3.68277742, -11.4271869, -19.9181071, -29.7289077, -41.6469974, -56.0852165, -73.4564054, -94.1734043, -118.649053, -148.335801, -182.290477, -220.72252, -263.84137, -311.856465, -360.155218, -414.9746, -477.729557, -549.835033, -632.705974, -729.031696, -837.395208, -957.110227, -1088.60305, -1234.57147, -1396.69343, -1576.64688, -1775.76872, -1996.15543, -2240.27555, -2511.33655, -2808.90806, -3131.75972, -3480.79659, -3860.54219, -4273.85912, -4725.10992, -5217.54607, -5748.91918, -6320.09193, -6922.46037, -7567.81013, -8260.28076, -9004.01179, -9803.14276, -10914.9154, -12027.0916, -13080.5352, -14016.1102, -14774.6807, -15037.9966, -15178.7785, -15248.9392, -15430.9604, -15725.0714, -16077.0978, -16489.9891, -16966.6946, -17510.1639, -18123.3464, -18463.6604, -19131.8092, -20051.2851, -21145.5806, -22338.1881, -23552.6, -24712.3088, -25584.5459, -24273.8252},
        {-256.431669, -256.811816, -257.089671, -257.267045, -257.345749, -257.327593, -257.214388, -257.007943, -256.71007, -256.322578, -255.847278, -255.285981, -254.640497, -253.912636, -253.104209, -252.217027, -251.252898, -250.213635, -249.101048, -247.916946, -246.66314, -245.341441, -243.95366, -242.501606, -240.98709, -239.411922, -237.777913, -236.086873, -234.340613, -232.540943, -230.689673, -228.788614, -226.839577, -224.844371, -222.828094, -220.780423, -218.690755, -216.560146, -214.389649, -212.18032, -209.933214, -207.649385, -205.329887, -202.975776, -200.588107, -198.167933, -195.71631, -193.234293, -190.722936, -188.183294, -185.616421, -183.01937, -180.397285, -177.751307, -175.082578, -172.392239, -169.681433, -166.951299, -164.202981, -161.437619, -158.656356, -155.860331, -153.050688, -150.228567, -147.39511, -144.551459, -141.698754, -138.996311, -136.599644, -134.183677, -131.739297, -129.257394, -126.728854, -124.144566, -121.495416, -118.772294, -115.966087, -113.067682, -110.067968, -106.957833, -103.728164, -100.369849, -96.8737761, -93.2308332, -88.3110107, -82.6892217, -76.9538517, -71.1321636, -65.2514207, -59.3388859, -53.4218226, -47.5274937, -41.6831624, -35.9160918, -30.2535451, -24.7227853, -19.3510756, -14.1656791, -9.19385895, -4.4628782, 0, -4.7069335, -8.35253154, -11.0154097, -12.7741834, -13.7074682, -13.8938798, -13.4120336, -12.3405451, -10.7580299, -8.74310349, -6.37438147, -3.73047934, -0.890012645, 2.06840309, 5.06615232, 8.02461952, 11.2105347, 14.5946176, 17.3171136, 19.3734098, 20.7588935, 21.4689516, 21.4989712, 20.8443395, 19.5004436, 17.4626705, 14.7264073, 11.2870412, 7.13995914, 2.98617605, -2.14327889, -8.4647069, -16.1944092, -24.9362299, -35.2126987, -47.7096674, -62.8271743, -80.9652578, -102.523956, -127.903308, -158.489146, -193.395689, -232.830104, -276.999556, -326.11121, -375.875254, -432.120077, -496.177088, -569.377699, -653.053318, -749.434528, -857.85458, -978.202538, -1111.11671, -1258.7651, -1422.62893, -1604.18941, -1804.68441, -2025.89377, -2269.86274, -2538.98756, -2833.62288, -3152.91501, -3498.03711, -3873.59663, -4282.62462, -4730.07961, -5219.49239, -5747.32626, -6314.04232, -6911.17142, -7549.47872, -8232.51654, -8963.83717, -9746.99291, -10825.5899, -11903.1132, -12923.1015, -13829.0939, -14564.629, -14825.8877, -14968.6723, -15042.5323, -15217.322, -15494.97, -15827.2785, -16218.6826, -16673.6175, -17196.5186, -17791.8211, -18345.6638, -19076.6677, -19952.5269, -20940.9352, -22009.5866, -23126.1751, -24258.3946, -24791.3387, -22786.2659},
        {-257.431268, -257.75499, -257.979504, -258.106503, -258.137678, -258.074722, -257.919329, -257.673191, -257.338, -256.91545, -256.407232, -255.81504, -255.140566, -254.385503, -253.551544, -252.64038, -251.653705, -250.593212, -249.460593, -248.257541, -246.985748, -245.646907, -244.242711, -242.774852, -241.245024, -239.654918, -238.006228, -236.300645, -234.539863, -232.725575, -230.859473, -228.943249, -226.978597, -224.967208, -222.926768, -220.850638, -218.731984, -216.571981, -214.371802, -212.132622, -209.855613, -207.541949, -205.192805, -202.809354, -200.392769, -197.944224, -195.464893, -192.955949, -190.418567, -187.85392, -185.263181, -182.649235, -180.011508, -177.351136, -174.669257, -171.967006, -169.245521, -166.505937, -163.749392, -160.977023, -158.189965, -155.389355, -152.576331, -149.752028, -146.917583, -144.074134, -141.222815, -138.52374, -136.133203, -133.724466, -131.288361, -128.815718, -126.297367, -123.724141, -121.08687, -118.376384, -115.583515, -112.699093, -109.71395, -106.618917, -103.404823, -100.062501, -96.5827811, -92.9564939, -88.0459093, -82.4298765, -76.7008094, -70.8861628, -65.0133917, -59.1099508, -53.2032952, -47.3208797, -41.4901591, -35.7385884, -30.0936224, -24.5827161, -19.2333243, -14.0729019, -9.12890376, -4.42878484, 0, -5.00303355, -8.8953317, -11.7597936, -13.6793182, -14.7368048, -15.0151523, -14.59726, -13.5660268, -12.004352, -9.99513451, -7.62127355, -4.96566818, -2.11121751, 0.859179363, 3.86262335, 6.81621534, 10.1027332, 13.7058032, 16.4892092, 18.46984, 19.6645844, 20.0903312, 19.7639694, 18.7023876, 16.9224748, 14.4411198, 11.2752114, 7.44163839, 2.95728967, -1.26690126, -6.43590664, -12.8010511, -20.6136591, -29.6851895, -40.4868999, -53.6073451, -69.4298095, -88.3375773, -110.713933, -136.94216, -168.22095, -203.870013, -244.113099, -289.173955, -339.27633, -390.49654, -448.242624, -513.775189, -588.354839, -673.242182, -770.376963, -879.510591, -1000.78483, -1134.9335, -1283.89929, -1449.08211, -1631.88188, -1833.71259, -2055.95691, -2299.98218, -2567.23889, -2858.9704, -3174.80064, -3516.19208, -3887.72252, -4292.53977, -4736.02355, -5221.90031, -5745.71287, -6307.63319, -6900.01751, -7532.05589, -8206.62571, -8926.60434, -9694.86916, -10741.218, -11784.8777, -12771.9955, -13648.7186, -14361.1943, -14619.4101, -14763.1129, -14839.6611, -15007.5285, -15270.0906, -15584.4248, -15956.2217, -16391.1715, -16894.9647, -17473.2917, -18211.9151, -18996.4212, -19832.4961, -20725.8257, -21682.0961, -22706.9932, -23806.2031, -23939.7925, -21289.1416},
        {-259.094316, -259.361922, -259.532746, -259.608387, -259.590445, -259.480519, -259.280209, -258.991114, -258.614834, -258.152968, -257.607115, -256.978875, -256.269847, -255.481631, -254.615826, -253.674031, -252.657847, -251.568872, -250.408706, -249.178949, -247.881199, -246.517056, -245.08812, -243.59599, -242.042266, -240.428546, -238.756431, -237.02752, -235.243411, -233.405706, -231.516003, -229.575901, -227.587, -225.550899, -223.479332, -221.368617, -219.214952, -217.019609, -214.783857, -212.508968, -210.196213, -207.84686, -205.462182, -203.043449, -200.591932, -198.1089, -195.595625, -193.053378, -190.483428, -187.887047, -185.265504, -182.62519, -179.962144, -177.277528, -174.572501, -171.848221, -169.10585, -166.346546, -163.57147, -160.78178, -157.978637, -155.163201, -152.33663, -149.500085, -146.654725, -143.80171, -140.942199, -138.237735, -135.846009, -133.437404, -131.002683, -128.532608, -126.017942, -123.449448, -120.817889, -118.114026, -115.328624, -112.452444, -109.476249, -106.390803, -103.186867, -99.8552052, -96.3865792, -92.7717522, -87.86886, -82.2588019, -76.5360907, -70.7282452, -64.8627838, -58.9672253, -53.0690883, -47.1958914, -41.3751533, -35.6343925, -30.0011278, -24.5028778, -19.1671611, -14.0214963, -9.09340211, -4.41039713, 0, -4.8693068, -8.65662365, -11.4433272, -13.3107943, -14.3404014, -14.6135254, -14.2115429, -13.2158306, -11.7077651, -9.76872325, -7.48008164, -4.92321697, -2.17950594, 0.669674757, 3.54294844, 6.35893842, 9.40035707, 12.6378404, 15.166477, 16.9829092, 18.0837794, 18.4657301, 18.1254035, 17.0594422, 15.2644884, 12.7371847, 9.47417331, 5.47209676, 0.727597392, -4.01523629, -9.78127155, -16.7980998, -25.2933126, -34.9652543, -46.30614, -59.9493169, -76.2811505, -95.6880064, -118.55625, -145.272247, -176.792222, -212.759246, -253.44819, -299.133925, -350.091322, -403.133739, -462.862938, -530.419171, -606.942686, -693.573732, -791.86048, -902.036691, -1024.49795, -1160.06003, -1310.39662, -1476.79621, -1660.54729, -1862.86519, -2085.12822, -2328.7945, -2595.32904, -2885.89302, -3200.05954, -3539.23962, -3907.95856, -4309.31214, -4748.53166, -5229.26654, -5747.33607, -6302.9887, -6890.03802, -7516.15733, -8183.82254, -8895.50955, -9653.69427, -10684.0611, -11710.0753, -12678.4107, -13535.7411, -14228.7403, -14469.0101, -14595.011, -14654.1618, -14804.4715, -15049.5839, -15347.0633, -15702.8575, -16122.914, -16613.1806, -17179.6048, -18079.5328, -18910.2012, -19709.1709, -20514.0028, -21362.2579, -22291.4971, -23339.2814, -23014.4422, -19922.9324},
        {-261.466571, -261.683279, -261.804367, -261.831391, -261.765907, -261.60947, -261.363637, -261.029964, -260.610006, -260.10532, -259.51746, -258.847984, -258.098447, -257.270405, -256.365414, -255.385029, -254.330807, -253.204304, -252.007075, -250.740676, -249.406663, -248.006593, -246.542021, -245.014502, -243.425593, -241.77685, -240.069829, -238.306085, -236.487174, -234.614653, -232.690077, -230.715002, -228.690984, -226.619579, -224.50946, -222.358476, -220.164388, -217.92852, -215.652197, -213.336745, -210.983487, -208.59375, -206.168857, -203.710135, -201.218906, -198.696498, -196.144234, -193.563439, -190.955439, -188.321557, -185.66312, -182.98605, -180.286974, -177.567117, -174.827705, -172.069963, -169.295115, -166.504388, -163.699006, -160.880195, -158.04918, -155.207185, -152.355437, -149.49516, -146.62758, -143.753922, -140.875411, -138.156096, -135.756115, -133.340884, -130.901072, -128.427349, -125.910385, -123.34085, -120.709415, -118.006749, -115.223523, -112.350406, -109.378069, -106.297181, -103.098412, -99.7724333, -96.3099141, -92.7015244, -87.8045116, -82.200097, -76.4831251, -70.6810476, -64.8213164, -58.9313832, -53.0386999, -47.1707181, -41.3548898, -35.6186666, -29.9895004, -24.494843, -19.1621462, -14.0188616, -9.09244124, -4.41033677, 0, -4.25793476, -7.5495215, -9.94815898, -11.527246, -12.3601813, -12.5203636, -12.0811918, -11.1160645, -9.69838064, -7.90153889, -5.79893804, -3.46397687, -0.970054135, 1.60943139, 4.20108093, 6.73149572, 9.13307086, 11.3332353, 13.245485, 14.7976629, 15.9176116, 16.533174, 16.5721928, 15.9625108, 14.6319707, 12.5084154, 9.51968765, 5.59363016, 0.658085729, -5.154625, -12.1961123, -20.5998769, -30.4994192, -41.0863286, -52.9650614, -66.99125, -83.5809685, -103.150291, -126.115291, -152.892043, -184.182596, -220.030013, -260.794416, -306.835927, -358.514669, -413.79565, -476.032884, -546.185272, -625.211713, -714.07111, -813.753194, -925.14835, -1049.01329, -1186.26618, -1338.15481, -1505.77897, -1690.23846, -1891.9741, -2112.89443, -2355.62683, -2622.93927, -2914.61619, -3229.4526, -3568.35848, -3935.82755, -4334.70856, -4769.43104, -5243.25359, -5753.63842, -6301.32649, -6882.02839, -7502.28936, -8164.59166, -8871.41752, -9625.24918, -10660.3723, -11689.5149, -12657.2083, -13507.984, -14186.3733, -14390.0897, -14475.0279, -14491.4984, -14610.6502, -14834.962, -15116.563, -15460.4493, -15871.6171, -16355.0624, -16915.7814, -17941.6251, -18808.8361, -19576.8816, -20305.2286, -21053.3443, -21880.6959, -22846.7506, -22021.821, -18746.8997},
        {-264.383961, -264.54945, -264.619945, -264.596979, -264.482085, -264.276795, -263.982643, -263.601161, -263.133882, -262.58234, -261.948067, -261.232596, -260.43746, -259.564192, -258.614325, -257.589391, -256.490924, -255.320457, -254.079522, -252.769652, -251.39238, -249.94924, -248.441764, -246.871485, -245.239936, -243.548649, -241.799159, -239.992997, -238.131696, -236.21679, -234.249812, -232.232293, -230.165768, -228.051769, -225.897247, -223.700913, -221.461411, -219.180097, -216.858329, -214.497464, -212.09886, -209.663872, -207.193858, -204.690176, -202.154182, -199.587233, -196.990687, -194.365901, -191.71423, -189.037034, -186.335668, -183.613282, -180.869401, -178.105345, -175.322431, -172.521978, -169.705303, -166.873726, -164.028565, -161.171137, -158.302761, -155.424754, -152.538437, -149.645126, -146.746139, -143.842796, -140.936414, -138.193991, -135.778545, -133.349695, -130.898018, -128.414093, -125.888497, -123.311809, -120.674606, -117.967466, -115.180967, -112.305688, -109.332205, -106.251098, -103.052943, -99.7283194, -96.2678046, -92.6619765, -87.7681573, -82.1673894, -76.4543359, -70.6563512, -64.8007895, -58.9150051, -53.0263523, -47.1621853, -41.3498584, -35.6167259, -29.990142, -24.4974611, -19.1660374, -14.0232252, -9.09637867, -4.41285219, 0, -3.56448324, -6.29318339, -8.25044246, -9.50060244, -10.1080053, -10.1369931, -9.65190779, -8.71709136, -7.39688583, -5.75563318, -3.85767541, -1.76735453, 0.450987487, 2.73300863, 5.0143669, 7.23072031, 8.92634855, 9.98099844, 11.2155884, 12.4819098, 13.6317537, 14.5169114, 14.9891741, 14.900333, 14.1021793, 12.4465044, 9.78509934, 5.9697554, 0.85226381, -6.10801221, -14.5229632, -24.4230695, -35.8388115, -47.4140249, -59.8725124, -74.3078487, -91.1665075, -110.894963, -133.939688, -160.747157, -191.75556, -227.422176, -268.195099, -314.522422, -366.852241, -424.212908, -488.827194, -561.49813, -643.028744, -734.222066, -835.666401, -948.641947, -1074.06101, -1212.866, -1366.06082, -1534.62176, -1719.52511, -1920.52452, -2140.09681, -2382.05258, -2650.60353, -2943.98486, -3260.11642, -3599.33658, -3966.08171, -4362.90715, -4793.30277, -5260.06619, -5762.56835, -6302.11849, -6876.31434, -7490.74771, -8148.01262, -8850.7031, -9601.41318, -10645.5789, -11682.2417, -12653.2852, -13500.5929, -14166.0484, -14330.5669, -14370.979, -14341.0119, -14427.808, -14632.2841, -14899.7674, -15233.9541, -15638.5406, -16117.2233, -16673.6984, -17745.0637, -18628.6515, -19384.4201, -20072.3277, -20752.3323, -21484.3924, -22328.466, -21010.8255, -17663.9018},
        {-267.802551, -267.916946, -267.936399, -267.862443, -267.69661, -267.440433, -267.095444, -266.663176, -266.145161, -265.542931, -264.85802, -264.091958, -263.24628, -262.322516, -261.3222, -260.246864, -259.098041, -257.877262, -256.586061, -255.225969, -253.79852, -252.305245, -250.747677, -249.127349, -247.445792, -245.70454, -243.905124, -242.049077, -240.137932, -238.173221, -236.156476, -234.08923, -231.973015, -229.809364, -227.604995, -225.358739, -223.069364, -220.738233, -218.36671, -215.95616, -213.507948, -211.023436, -208.503989, -205.950971, -203.365747, -200.749679, -198.104134, -195.430474, -192.730064, -190.004268, -187.254449, -184.478443, -181.681219, -178.864219, -176.028882, -173.176651, -170.308964, -167.427264, -164.53299, -161.627583, -158.712483, -155.789133, -152.858971, -149.923439, -146.983978, -144.042027, -141.099028, -138.325351, -135.887313, -133.437946, -130.967742, -128.467188, -125.926775, -123.336993, -120.68833, -117.971277, -115.176323, -112.293957, -109.31467, -106.22895, -103.027288, -99.7001725, -96.2380934, -92.6315404, -87.7377029, -82.137558, -76.4257041, -70.6294084, -64.7759383, -58.8925612, -53.0065442, -47.1451549, -41.3356604, -35.6053282, -29.9814255, -24.4912197, -19.161978, -14.0209679, -9.09545667, -4.41271158, 0, -2.93065494, -5.14526076, -6.69995014, -7.65085576, -8.05411031, -7.96584647, -7.44219693, -6.53929435, -5.31327144, -3.82026086, -2.11639531, -0.257807469, 1.69936999, 3.69900437, 5.684963, 7.60111319, 8.63708658, 8.62900253, 9.22673764, 10.2125373, 11.3686469, 12.4773118, 13.3207774, 13.681289, 13.3410922, 12.0824322, 9.68755443, 5.9387043, 0.618127185, -7.42396188, -17.1315427, -28.4427608, -41.2957616, -53.8546172, -66.9445097, -81.8637078, -99.0825791, -119.071491, -142.300812, -169.240908, -200.082266, -235.660317, -276.500187, -323.127004, -376.065895, -435.130299, -501.734952, -576.582904, -660.377202, -753.820896, -857.35844, -972.267536, -1099.33761, -1239.43453, -1393.58032, -1562.72688, -1747.82613, -1948.20015, -2166.80108, -2408.35903, -2678.36753, -2973.63438, -3291.26273, -3631.04841, -3997.34981, -4392.43104, -4818.87669, -5279.03396, -5774.07487, -6305.83615, -6873.49053, -7481.94867, -8133.96947, -8832.31188, -9579.73482, -10632.7576, -11676.9387, -12651.5969, -13496.051, -14149.6198, -14276.9981, -14275.2115, -14201.2748, -14258.1211, -14445.006, -14700.3856, -15026.5552, -15425.8103, -15900.4462, -16452.7585, -17463.1933, -18337.1791, -19106.6519, -19803.5479, -20459.8034, -21107.3543, -21778.137, -20006.0965, -16666.347},
        {-271.647706, -271.716623, -271.689473, -271.567838, -271.353299, -271.047438, -270.651835, -270.168072, -269.59773, -268.942391, -268.203635, -267.383044, -266.482199, -265.502681, -264.446072, -263.313953, -262.107905, -260.829509, -259.480347, -258.061999, -256.576048, -255.024074, -253.407659, -251.728384, -249.987829, -248.187577, -246.329209, -244.414306, -242.444448, -240.421218, -238.346197, -236.220965, -234.047105, -231.826197, -229.568075, -227.27002, -224.929215, -222.546974, -220.124609, -217.663434, -215.164763, -212.629908, -210.060185, -207.456905, -204.821382, -202.154931, -199.458863, -196.734494, -193.983135, -191.206102, -188.404706, -185.56621, -182.706284, -179.826545, -176.928612, -174.014103, -171.084635, -168.141828, -165.187298, -162.222665, -159.249546, -156.269559, -153.284322, -150.295454, -147.304572, -144.313294, -141.323239, -138.509094, -136.041393, -133.564809, -131.069742, -128.546589, -125.985749, -123.377622, -120.712605, -117.981098, -115.173499, -112.280206, -109.291619, -106.198135, -102.990155, -99.6580751, -96.1922955, -92.5832144, -87.6827451, -82.0744753, -76.3556767, -70.5536941, -64.695872, -58.8095552, -52.9220884, -47.0608162, -41.2530834, -35.5262346, -29.9076145, -24.4245679, -19.1044393, -13.9745736, -9.06231533, -4.39500924, 0, -2.84630297, -4.99575848, -6.50386958, -7.42613927, -7.81807058, -7.73516654, -7.23293016, -6.36686447, -5.19247249, -3.76525725, -2.14072176, -0.374369045, 1.47829787, 3.36177596, 5.2205622, 6.99915358, 7.81148788, 7.48341078, 7.83806501, 8.6419706, 9.66164755, 10.6636159, 11.4143956, 11.6805067, 11.2284691, 9.82480304, 7.23602835, 3.2286651, -2.43076672, -11.0052467, -21.3212721, -33.3034729, -46.8764792, -60.1793142, -74.0294114, -89.7203663, -107.712491, -128.466098, -152.441499, -180.099006, -211.487724, -247.604322, -288.989568, -336.184226, -389.729064, -449.070189, -516.11669, -591.682995, -676.583536, -771.632741, -878.002301, -995.712736, -1124.58766, -1265.11972, -1419.16742, -1587.97606, -1772.79089, -1973.56914, -2193.13687, -2435.72525, -2706.84391, -3003.01635, -3321.08807, -3660.67245, -4026.07285, -4419.43986, -4843.09927, -5299.24711, -5789.43691, -6315.58562, -6876.90905, -7478.44087, -8123.03305, -8813.53757, -9552.80638, -10600.3099, -11638.127, -12604.9552, -13439.4917, -14080.434, -14187.8559, -14164.8274, -14069.4557, -14110.6793, -14286.9017, -14533.6802, -14852.5606, -15245.0886, -15712.8102, -16257.2712, -16923.2746, -17721.8723, -18580.2181, -19425.466, -20184.7698, -20785.2832, -21154.1602, -19070.5803, -15737.4698},
        {-275.882505, -275.910852, -275.840923, -275.674394, -275.412939, -275.058236, -274.611959, -274.075784, -273.451387, -272.740444, -271.94463, -271.06562, -270.105092, -269.064719, -267.946178, -266.751144, -265.481294, -264.138303, -262.723846, -261.239599, -259.687238, -258.068438, -256.384875, -254.638225, -252.830164, -250.962366, -249.036508, -247.054266, -245.017315, -242.92733, -240.785988, -238.594964, -236.355933, -234.070572, -231.754941, -229.403222, -227.009421, -224.574746, -222.100408, -219.587614, -217.037574, -214.451496, -211.830589, -209.176063, -206.489125, -203.770984, -201.02285, -198.245931, -195.441437, -192.610575, -189.754554, -186.845145, -183.913633, -180.961865, -177.991689, -175.004951, -172.003499, -168.98918, -165.96384, -162.929327, -159.887487, -156.840168, -153.789216, -150.736479, -147.683804, -144.633038, -141.586027, -138.722581, -136.218235, -133.707805, -131.181599, -128.629929, -126.043105, -123.411438, -120.725237, -117.974814, -115.150479, -112.242542, -109.241314, -106.137105, -102.920226, -99.5809871, -96.1096989, -92.4966718, -87.5830875, -81.9579913, -76.22424, -70.4094354, -64.5411796, -58.6470744, -52.7547217, -46.8917234, -41.0856814, -35.3641976, -29.7548738, -24.2853121, -18.9831143, -13.8758823, -8.99121795, -4.35672323, 0, -3.37359941, -5.95787474, -7.81610745, -9.01157904, -9.60757099, -9.66736479, -9.25424193, -8.43148388, -7.26237214, -5.8101882, -4.13821352, -2.30972961, -0.388017947, 1.56363998, 3.4819627, 5.30366871, 6.37742294, 6.55593627, 7.10604839, 7.83846635, 8.56389718, 9.0930479, 9.23662556, 8.80533718, 7.60988979, 5.46099043, 2.16934614, -2.45433607, -8.59934915, -17.0910787, -27.249491, -39.0730514, -52.5602251, -66.325244, -81.0586899, -97.8202735, -117.02373, -139.082794, -164.411202, -193.422688, -226.142437, -263.49099, -305.958101, -354.033526, -408.207021, -466.371298, -532.262418, -607.019395, -691.781248, -787.686993, -897.490039, -1018.74186, -1149.49671, -1289.59416, -1442.51888, -1610.11895, -1794.24249, -1996.57003, -2219.15531, -2464.2348, -2735.9489, -3031.80811, -3349.03605, -3687.4706, -4051.37935, -4442.99662, -4865.10776, -5320.08996, -5808.2997, -6331.23649, -6886.53467, -7480.15353, -8114.93051, -8793.70306, -9519.30862, -10544.6649, -11560.0091, -12505.6584, -13321.9302, -13949.1418, -14056.1122, -14035.6562, -13944.4476, -13986.0794, -14159.0698, -14400.7211, -14712.6124, -15096.3229, -15553.4318, -16085.5183, -16132.0233, -16791.3993, -17811.5597, -18940.4178, -19925.887, -20515.8805, -20458.3117, -18220.1678, -14871.7741},
        {-280.514156, -280.498876, -280.382934, -280.168106, -279.856168, -279.448896, -278.948067, -278.355457, -277.672841, -276.901997, -276.0447, -275.102726, -274.077852, -272.971854, -271.786508, -270.52359, -269.184876, -267.772143, -266.287166, -264.731723, -263.107588, -261.416538, -259.66035, -257.840799, -255.959663, -254.018716, -252.019735, -249.964496, -247.854776, -245.69235, -243.478995, -241.216487, -238.906603, -236.551117, -234.172754, -231.762375, -229.31059, -226.818498, -224.287193, -221.717772, -219.111331, -216.468968, -213.791778, -211.080858, -208.337304, -205.562213, -202.756681, -199.921805, -197.05868, -194.168404, -191.252073, -188.264956, -185.25497, -182.224205, -179.174751, -176.108698, -173.028135, -169.935152, -166.831839, -163.720287, -160.602585, -157.480822, -154.357089, -151.233475, -148.112071, -144.994966, -141.88425, -138.964749, -136.416903, -133.865921, -131.302047, -128.715524, -126.096596, -123.435508, -120.722502, -117.947822, -115.101713, -112.174418, -109.15618, -106.037245, -102.807854, -99.4582531, -95.9786849, -92.3593934, -87.4304742, -81.7866475, -76.0365771, -70.2081561, -64.3292772, -58.4278332, -52.5317169, -46.6688212, -40.8670389, -35.1542627, -29.5583856, -24.1073003, -18.8288997, -13.7510765, -8.90172365, -4.30873386, 0, -4.07439526, -7.23606303, -9.55839341, -11.1147765, -11.9786024, -12.2232611, -11.9221429, -11.1486377, -9.97613567, -8.47802691, -6.72770149, -4.79854952, -2.76396108, -0.697326271, 1.32796482, 3.23852209, 4.70925468, 5.63081524, 6.49913687, 7.18689359, 7.56675942, 7.5114084, 6.89351454, 5.58575188, 3.46079445, 0.391316264, -3.75000865, -9.09050626, -15.7575025, -23.9745586, -33.7374791, -45.1447194, -58.294735, -72.4236351, -88.0610482, -105.966562, -126.497335, -150.010527, -176.863298, -207.412806, -241.696472, -280.488507, -324.208625, -373.276544, -428.111981, -484.817809, -549.209798, -622.786877, -707.047973, -803.492014, -916.554907, -1041.20896, -1173.68117, -1313.14427, -1464.78517, -1631.09052, -1814.54698, -2018.74447, -2244.81565, -2492.68969, -2764.78406, -3059.95112, -3376.00917, -3713.14207, -4075.54187, -4465.56079, -4886.60221, -5341.29089, -5828.39728, -6348.87205, -6898.3695, -7483.95148, -8108.40202, -8774.50513, -9485.04483, -10483.3424, -11471.5102, -12392.1981, -13188.0558, -13801.733, -13915.0982, -13905.4365, -13827.1606, -13874.9827, -14047.7807, -14285.9748, -14591.3658, -14965.7547, -15410.9424, -15928.7298, -15306.0904, -15813.2156, -17005.757, -18439.3661, -19669.6945, -20252.3937, -19743.1152, -17415.702, -14076.9462},
        {-285.519624, -285.454667, -285.286836, -285.017996, -284.650018, -284.184767, -283.624113, -282.969923, -282.224065, -281.388406, -280.464815, -279.455159, -278.361306, -277.185124, -275.928481, -274.593244, -273.181282, -271.694462, -270.134652, -268.50372, -266.803534, -265.035961, -263.202869, -261.306126, -259.347601, -257.32916, -255.252672, -253.120004, -250.933025, -248.693601, -246.403602, -244.064894, -241.679345, -239.248825, -236.802145, -234.327135, -231.811298, -229.255628, -226.661118, -224.028762, -221.359552, -218.654483, -215.914547, -213.140738, -210.334049, -207.495475, -204.626007, -201.726639, -198.798365, -195.842179, -192.859072, -189.788273, -186.69388, -183.578226, -180.443642, -177.292461, -174.127017, -170.94964, -167.762663, -164.568419, -161.36924, -158.167458, -154.965406, -151.765416, -148.56982, -145.380951, -142.201141, -139.219751, -136.621648, -134.023436, -131.415322, -128.787516, -126.130224, -123.433655, -120.688016, -117.883518, -115.010366, -112.05877, -109.018937, -105.881076, -102.635394, -99.2721005, -95.7814028, -92.1535092, -87.2084661, -81.5461753, -75.7806028, -69.9399309, -64.0523415, -58.1460169, -52.2491391, -46.3898902, -40.5964524, -34.8970077, -29.3197383, -23.8928262, -18.6444536, -13.6028025, -8.79605519, -4.25239365, 0, -4.85343334, -8.6575792, -11.4969217, -13.4559451, -14.6191335, -15.070971, -14.8959417, -14.17853, -13.0032198, -11.4544955, -9.616841, -7.57474061, -5.41267846, -3.21513869, -1.06660547, 0.948437062, 2.8753453, 4.64818375, 5.88878126, 6.54047665, 6.54660874, 5.85051636, 4.39553834, 2.12501348, -1.01771938, -5.08932142, -10.1464538, -16.2457778, -23.4439544, -31.2825423, -40.5221762, -51.3740482, -64.0493501, -78.4726904, -94.9985528, -114.057843, -135.947729, -160.965377, -189.407955, -221.572629, -257.527261, -297.868111, -342.937213, -393.076596, -448.628292, -503.944637, -566.854782, -639.198181, -722.814289, -819.542562, -935.353648, -1062.90657, -1196.80484, -1335.63581, -1486.12049, -1651.32803, -1834.32762, -2040.5354, -2270.14013, -2520.77022, -2793.0093, -3087.20581, -3401.91009, -3737.71569, -4098.67821, -4487.26409, -4907.52626, -5362.34247, -5848.77348, -6367.17048, -6911.13143, -7488.79971, -8102.86415, -8756.01361, -9450.93692, -10419.4076, -11377.7587, -12271.4079, -13045.7727, -13646.2709, -13770.4852, -13776.8915, -13717.1224, -13774.86, -13949.4673, -14185.2868, -14484.4327, -14849.0188, -15281.1594, -15782.9686, -14518.249, -14877.4183, -16231.754, -17952.5338, -19411.0354, -19978.5365, -19026.3147, -16655.7527, -13351.638},
        {-290.875875, -290.752199, -290.523958, -290.193087, -289.761522, -289.231198, -288.604051, -287.882016, -287.06703, -286.161027, -285.165943, -284.083715, -282.916277, -281.665565, -280.333515, -278.922063, -277.433144, -275.868693, -274.230647, -272.52094, -270.741509, -268.894289, -266.981216, -265.004226, -262.965253, -260.866234, -258.709104, -256.4958, -254.228255, -251.908407, -249.53819, -247.119541, -244.654394, -242.144686, -239.623749, -237.077159, -234.490117, -231.863538, -229.198341, -226.495441, -223.755755, -220.9802, -218.169693, -215.32515, -212.447489, -209.537625, -206.596476, -203.624959, -200.62399, -197.594486, -194.537364, -191.377726, -188.193946, -184.988586, -181.764204, -178.523363, -175.26862, -172.002538, -168.727676, -165.446595, -162.161854, -158.876014, -155.591635, -152.311278, -149.037502, -145.772868, -142.519937, -139.471736, -136.81672, -134.164624, -131.505662, -128.830045, -126.127987, -123.3897, -120.605399, -117.765296, -114.859603, -111.878534, -108.812302, -105.65112, -102.385201, -99.0047569, -95.5000021, -91.861149, -86.9006238, -81.2223061, -75.4442316, -69.5948346, -63.7025496, -57.7958111, -51.9030534, -46.052711, -40.2732184, -34.5930099, -29.04052, -23.6441832, -18.4324339, -13.4337065, -8.67643538, -4.18905508, 0, -5.61545651, -10.0496791, -13.3978867, -15.7552983, -17.217133, -17.8786096, -17.8349473, -17.1813649, -16.0130814, -14.425316, -12.5132874, -10.3722148, -8.09731702, -5.78381316, -3.52692218, -1.42186306, 0.94405699, 3.54817782, 5.14643236, 5.75243988, 5.37981961, 4.04219082, 1.75317276, -1.47361533, -5.62455418, -10.6860246, -16.6444072, -23.4860828, -31.1974322, -38.6418852, -47.3405221, -57.6166089, -69.7934118, -84.4706129, -101.83327, -121.992731, -145.189335, -171.663425, -201.655342, -235.405426, -273.012234, -314.901044, -361.339938, -412.596996, -468.940302, -523.286702, -585.09332, -656.466049, -739.510778, -836.333397, -954.043006, -1083.62726, -1218.53151, -1356.93452, -1506.67901, -1671.26879, -1854.20765, -2062.38578, -2295.15096, -2548.15671, -2820.28451, -3113.33263, -3426.64146, -3761.22027, -4120.90618, -4508.23826, -4927.82355, -5382.73729, -5868.47217, -6384.81001, -6923.53837, -7493.66319, -8097.7335, -8738.29835, -9417.90679, -10355.9263, -11283.8829, -12150.1209, -12902.9849, -13490.8192, -13627.9444, -13652.7447, -13613.8606, -13683.1818, -13860.5625, -14094.5028, -14387.4247, -14741.7498, -15159.9002, -15644.2975, -13841.2725, -14074.1046, -15558.4948, -17510.1439, -19144.7529, -19678.0227, -18325.6544, -15938.8896, -12694.501},
        {-296.320558, -296.105001, -295.787093, -295.368665, -294.851545, -294.237564, -293.528552, -292.726338, -291.832753, -290.849627, -289.778788, -288.622068, -287.381295, -286.0583, -284.654913, -283.172963, -281.614281, -279.980695, -278.274037, -276.496136, -274.648821, -272.733924, -270.753272, -268.708697, -266.602029, -264.435096, -262.209729, -259.927758, -257.591013, -255.201323, -252.760519, -250.27043, -247.732886, -245.149717, -242.547791, -239.915894, -237.242507, -234.528647, -231.775332, -228.983579, -226.154406, -223.288829, -220.387866, -217.452535, -214.483853, -211.482836, -208.450502, -205.387869, -202.295954, -199.175774, -196.028346, -192.778108, -189.504317, -186.20965, -182.896785, -179.568399, -176.22717, -172.875775, -169.516892, -166.153198, -162.78737, -159.422085, -156.060022, -152.703858, -149.356269, -146.019934, -142.697529, -139.582128, -136.86223, -134.147838, -131.429288, -128.696914, -125.941051, -123.152033, -120.320197, -117.435876, -114.489405, -111.47112, -108.371356, -105.180446, -101.888727, -98.4865326, -94.9641982, -91.3120586, -86.3448478, -80.6654282, -74.8907923, -69.0494255, -63.1698132, -57.2804407, -51.4097935, -45.5863569, -39.8386165, -34.1950575, -28.6841654, -23.3344256, -18.1743235, -13.2323445, -8.53697393, -4.11669731, 0, -6.04555386, -10.8443651, -14.4969958, -17.1040083, -18.7659647, -19.5834271, -19.6569577, -19.0871188, -17.9744725, -16.4195809, -14.5230063, -12.3853107, -10.1070565, -7.78880577, -5.53112065, -3.43456335, -0.925151875, 1.99291506, 3.6933206, 4.22004775, 3.61707951, 1.92839889, -0.802011114, -4.53016749, -9.21208723, -14.8037873, -21.2612848, -28.5405966, -36.5977398, -43.9873126, -52.5806043, -62.7540574, -74.8841146, -89.9237073, -107.961034, -128.930561, -153.035787, -180.480205, -211.467314, -246.200607, -284.385465, -326.875101, -373.970469, -425.972521, -483.182212, -540.022721, -604.142219, -677.311103, -761.299771, -857.87862, -972.327238, -1098.61782, -1232.40806, -1372.70964, -1525.38198, -1693.21003, -1878.9787, -2088.04103, -2320.03018, -2571.77778, -2842.26766, -3133.62504, -3446.03886, -3780.71996, -4140.61078, -4527.85893, -4946.41624, -5398.89807, -5881.30435, -6393.37706, -6927.08385, -7491.13711, -8087.96994, -8720.01547, -9389.7068, -10310.6522, -11221.3159, -12071.3371, -12810.3551, -13388.0093, -13525.9879, -13553.8487, -13518.9243, -13589.5633, -13765.1962, -13994.8302, -14280.4167, -14623.9074, -15027.2539, -15492.4078, -13792.7207, -14034.5509, -15458.3434, -17304.5431, -18813.5948, -19225.9437, -17782.0345, -15260.3225, -12103.734},
        {-302.119652, -301.801575, -301.383952, -300.868484, -300.256868, -299.550802, -298.751985, -297.862114, -296.882889, -295.816008, -294.663168, -293.426069, -292.106408, -290.705883, -289.226193, -287.669037, -286.036112, -284.329116, -282.549749, -280.699708, -278.780691, -276.794397, -274.742524, -272.62677, -270.448834, -268.210414, -265.913208, -263.558915, -261.149232, -258.685858, -256.170491, -253.60483, -250.990573, -248.329417, -245.639935, -242.915032, -240.147413, -237.33823, -234.488632, -231.599771, -228.672796, -225.70886, -222.709113, -219.674705, -216.606787, -213.50651, -210.375024, -207.213481, -204.023031, -200.804825, -197.560013, -194.21512, -190.847541, -187.460044, -184.055398, -180.636371, -177.205731, -173.766248, -170.320689, -166.871824, -163.422421, -159.975247, -156.533073, -153.098665, -149.674793, -146.264225, -142.86973, -139.683291, -136.892348, -134.109429, -131.325035, -128.52967, -125.713835, -122.868034, -119.982769, -117.048542, -114.055855, -110.995212, -107.857114, -104.632065, -101.310566, -97.8831195, -94.340229, -90.6723964, -85.7006073, -80.024722, -74.2586571, -68.4308682, -62.5698109, -56.7039407, -50.8617131, -45.0715837, -39.3620079, -33.7614414, -28.2983397, -23.0011583, -17.8983528, -13.0183787, -8.38969147, -4.04074673, 0, -6.31373581, -11.3462257, -15.2010923, -17.9819581, -19.7924456, -20.7361774, -20.9167762, -20.4378644, -19.4030645, -17.9159992, -16.0802911, -13.9995625, -11.7774362, -9.51753472, -7.32348052, -5.29889621, -2.77393253, 0.25828393, 1.94461104, 2.34717022, 1.52808284, -0.450529666, -3.52654591, -7.63784449, -12.722304, -18.717803, -25.5622202, -33.193434, -41.5493232, -48.9892024, -57.6083209, -67.8181263, -80.0300666, -95.4242238, -114.027728, -135.662742, -160.510122, -188.750726, -220.565411, -256.135034, -294.820504, -337.872177, -385.631336, -438.439264, -496.637244, -556.533104, -623.509944, -698.91741, -784.10515, -880.422809, -990.909538, -1113.16053, -1245.48542, -1387.89488, -1543.87257, -1715.34284, -1904.23005, -2113.95584, -2344.60699, -2594.63685, -2863.45162, -3153.23811, -3464.91661, -3799.82391, -4160.00249, -4547.17092, -4964.48157, -5414.02475, -5892.63338, -6400.11428, -6929.21034, -7487.879, -8078.37228, -8702.94215, -9363.84062, -10268.9123, -11163.4185, -11998.2129, -12724.1494, -13292.0818, -13430.6264, -13461.033, -13429.4, -13500.9618, -13674.6591, -13899.7922, -14177.7838, -14510.0562, -14898.032, -15343.1337, -13908.2147, -14195.2928, -15513.0487, -17170.1634, -18475.3177, -18737.1924, -17264.4683, -14622.6787, -11578.5359},
        {-308.313295, -307.884642, -307.359391, -306.739104, -306.025342, -305.219666, -304.323638, -303.33882, -302.266772, -301.109056, -299.867234, -298.542867, -297.137516, -295.652743, -294.09011, -292.451177, -290.737506, -288.950659, -287.092198, -285.163682, -283.166675, -281.102737, -278.973429, -276.780314, -274.524953, -272.208907, -269.833737, -267.401005, -264.912273, -262.369101, -259.773052, -257.125686, -254.428566, -251.683252, -248.899397, -246.074347, -243.205349, -240.293704, -237.34071, -234.347666, -231.315871, -228.246623, -225.141222, -222.000967, -218.827156, -215.621088, -212.384062, -209.117378, -205.822333, -202.500227, -199.152358, -195.708895, -192.243808, -188.75994, -185.260132, -181.747223, -178.224055, -174.693469, -171.158306, -167.621406, -164.085611, -160.553761, -157.028698, -153.513262, -150.010293, -146.522634, -143.053124, -139.791773, -136.924095, -134.066912, -131.210931, -128.34686, -125.465406, -122.557277, -119.61318, -116.623824, -113.579915, -110.472161, -107.291269, -104.027948, -100.672905, -97.2168469, -93.6504818, -89.964517, -84.9899133, -79.3216454, -73.5686273, -67.7592026, -61.9217147, -56.0845074, -50.2759241, -44.5243084, -38.858004, -33.3053543, -27.8947029, -22.6543935, -17.6127695, -12.7981746, -8.23895237, -3.9634463, 0, -6.38272729, -11.486852, -15.416403, -18.2754093, -20.1678996, -21.1979029, -21.4694481, -21.0865641, -20.1532797, -18.7736238, -17.0516254, -15.0913133, -12.9967164, -10.8718636, -8.82078376, -6.94750582, -4.55796949, -1.64219092, -0.109400662, 0.107391502, -0.924824213, -3.13905759, -6.46831842, -10.8456165, -16.2039616, -22.4763634, -29.5958319, -37.4953768, -46.1080078, -53.7210824, -62.5166686, -72.9214719, -85.3621977, -101.100212, -120.14489, -142.283032, -167.690943, -196.544932, -229.021305, -265.296371, -304.524536, -348.215021, -396.74407, -450.487927, -509.822836, -572.974734, -643.00775, -720.835706, -807.372423, -903.531723, -1010.06609, -1128.24787, -1259.04206, -1403.46376, -1562.63042, -1737.61354, -1929.48462, -2139.70145, -2368.86143, -2617.14061, -2884.55268, -3173.11143, -3484.25096, -3819.34352, -4179.65653, -4566.50552, -4982.23284, -5428.42024, -5902.88445, -6405.57191, -6930.54276, -7484.5353, -8069.53353, -8687.52143, -9340.48299, -10229.8949, -11108.3752, -11928.0349, -12640.9847, -13199.3355, -13338.8621, -13372.2354, -13344.2816, -13416.9215, -13588.6314, -13809.0986, -14079.239, -14399.9687, -14772.2036, -15196.8597, -14143.1173, -14503.5722, -15685.0639, -17094.432, -18138.5161, -18224.1557, -16758.1903, -14025.4148, -11117.6778},
        {-314.941622, -314.396926, -313.758267, -313.027085, -312.204821, -311.292914, -310.292804, -309.205931, -308.033735, -306.777656, -305.439134, -304.019609, -302.520521, -300.943309, -299.289414, -297.560276, -295.757334, -293.882028, -291.935799, -289.920087, -287.83633, -285.68597, -283.470446, -281.191198, -278.849666, -276.44729, -273.985509, -271.465765, -268.889496, -266.258143, -263.573146, -260.835944, -258.047977, -255.210686, -252.325398, -249.393612, -246.416828, -243.396488, -240.334036, -237.230916, -234.088573, -230.908448, -227.691986, -224.44063, -221.155824, -217.839011, -214.491635, -211.11514, -207.710969, -204.280566, -200.825374, -197.279562, -193.713307, -190.129511, -186.531075, -182.920902, -179.301894, -175.676952, -172.048979, -168.420876, -164.795546, -161.17589, -157.564811, -153.965209, -150.379988, -146.81205, -143.264295, -139.92412, -136.97449, -134.037804, -131.105004, -128.16703, -125.214826, -122.239334, -119.231495, -116.182251, -113.082545, -109.923318, -106.695512, -103.39007, -99.9979333, -96.5100439, -92.9173438, -89.2107752, -84.2347765, -78.5776566, -72.8415045, -67.0544684, -61.2446965, -55.4403371, -49.6695383, -43.9604485, -38.3412158, -32.8399885, -27.4849148, -22.304143, -17.3258213, -12.5780979, -8.08912104, -3.88703901, 0, -6.21525322, -11.197835, -15.0491551, -17.8706237, -19.7636506, -20.8296457, -21.1700191, -20.8861805, -20.07954, -18.8515075, -17.3034929, -15.5369061, -13.6531572, -11.7536559, -9.93981235, -8.31303635, -6.23294723, -3.69498482, -2.47841886, -2.52570415, -3.77929548, -6.18164764, -9.67521543, -14.2024536, -19.7058171, -26.1277605, -33.4107388, -41.4972066, -50.3296189, -58.2564805, -67.3986444, -78.1767502, -91.0114379, -107.079722, -126.424055, -148.885189, -174.656851, -203.932768, -236.906668, -273.772278, -313.704748, -358.226384, -407.7302, -462.609209, -523.256426, -589.504495, -662.446893, -742.616724, -830.547097, -926.771116, -1030.07309, -1144.87234, -1274.35645, -1420.38978, -1582.13519, -1759.96845, -1954.26536, -2164.84904, -2392.77353, -2639.69572, -2906.28711, -3194.18459, -3505.01815, -3840.09022, -4200.14813, -4586.19403, -4999.88332, -5442.38749, -5912.48273, -6410.30014, -6931.70602, -7481.75241, -8062.04672, -8674.19634, -9319.80866, -10192.788, -11054.3706, -11858.0896, -12557.4782, -13106.0695, -13247.6972, -13285.3937, -13262.5634, -13336.9865, -13506.7935, -13722.4587, -13984.4955, -14293.4173, -14649.7377, -15053.9701, -14452.7911, -14906.6311, -15936.842, -17064.7758, -17811.7847, -17699.2205, -16248.4353, -13467.9872, -10719.9313},
        {-322.044773, -321.381149, -320.625435, -319.778986, -318.843156, -317.819299, -316.70877, -315.512923, -314.233113, -312.870693, -311.427019, -309.903444, -308.301323, -306.62201, -304.866859, -303.037226, -301.134464, -299.159927, -297.114971, -295.000948, -292.819215, -290.571125, -288.258032, -285.88129, -283.442255, -280.942281, -278.382721, -275.76493, -273.090263, -270.360074, -267.575717, -264.738546, -261.849917, -258.911183, -255.917157, -252.872603, -249.782361, -246.647997, -243.471079, -240.253173, -236.995845, -233.700663, -230.369193, -227.003001, -223.603655, -220.17272, -216.711763, -213.222352, -209.706052, -206.16443, -202.599054, -198.947255, -195.276226, -191.588927, -187.888317, -184.177355, -180.459, -176.736211, -173.011946, -169.289166, -165.57083, -161.859896, -158.159323, -154.47207, -150.801098, -147.149364, -143.519828, -140.09688, -137.060555, -134.039623, -131.025281, -128.008728, -124.981161, -121.933778, -118.857776, -115.744353, -112.584706, -109.370034, -106.091534, -102.740403, -99.3078387, -95.7850394, -92.1632025, -88.4335256, -83.4572076, -77.8142136, -72.0980902, -66.3367056, -60.5579281, -54.789626, -49.0596676, -43.395921, -37.8262546, -32.3785366, -27.0806352, -21.9604188, -17.0457556, -12.3645139, -7.94456186, -3.81376781, 0, -5.77403853, -10.4107656, -14.0055756, -16.6538631, -18.4510224, -19.4924481, -19.8735345, -19.6896762, -19.0362676, -18.008703, -16.7023771, -15.2126842, -13.6350188, -12.0647753, -10.5973482, -9.32813199, -7.75455026, -5.88657311, -5.17214791, -5.57853251, -7.07298477, -9.62276255, -13.1951237, -17.7573261, -23.2766276, -29.7202861, -37.0555594, -45.2497053, -54.2699818, -62.6689246, -72.347245, -83.6966176, -97.108717, -113.490803, -132.976758, -155.562973, -181.486448, -210.984181, -244.293171, -281.650418, -322.568326, -368.229017, -419.011257, -475.293813, -537.455452, -606.279271, -681.638626, -763.8112, -853.074679, -949.706746, -1051.20671, -1164.02643, -1292.70705, -1439.64644, -1602.86651, -1782.35388, -1978.09518, -2188.96984, -2416.32331, -2662.70887, -2929.37121, -3217.3972, -3528.19443, -3862.87539, -4222.05252, -4606.56777, -5017.6463, -5456.22942, -5921.85338, -6414.8492, -6933.32503, -7480.17676, -8056.5049, -8663.40994, -9301.99236, -10156.7799, -10999.5891, -11785.6634, -12470.2467, -13008.5826, -13154.1341, -13198.4459, -13183.2393, -13260.701, -13428.826, -13639.5822, -13893.2661, -14190.1743, -14530.6033, -14914.8495, -14792.5991, -15351.7116, -16230.8361, -17068.6219, -17503.7181, -17174.7736, -15720.4378, -12949.8526, -10384.0674},
        {-329.973393, -329.191013, -328.316047, -327.349883, -326.293905, -325.149499, -323.918051, -322.600946, -321.19957, -319.71531, -318.149549, -316.503675, -314.779073, -312.977129, -311.099227, -309.146755, -307.121097, -305.023639, -302.855768, -300.618867, -298.314325, -295.943525, -293.507854, -291.008697, -288.44744, -285.825469, -283.144169, -280.404927, -277.609127, -274.758155, -271.853398, -268.89624, -265.888068, -262.830267, -259.717797, -256.555393, -253.347864, -250.096806, -246.803811, -243.470475, -240.09839, -236.689153, -233.244356, -229.765594, -226.254462, -222.712552, -219.14146, -215.54278, -211.918105, -208.269031, -204.597151, -200.838345, -197.061346, -193.269174, -189.464846, -185.651381, -181.831798, -178.009116, -174.186353, -170.366529, -166.552662, -162.74777, -158.954873, -155.176988, -151.417136, -147.678335, -143.963602, -140.454346, -137.327688, -134.218737, -131.118947, -128.019774, -124.91267, -121.789092, -118.640493, -115.458328, -112.234053, -108.95912, -105.624986, -102.223104, -98.7449291, -95.1819161, -91.5255194, -87.7671936, -82.7897476, -77.1622777, -71.4672052, -65.7319625, -59.9839819, -54.2506955, -48.5595358, -42.9379348, -37.413325, -32.0131386, -26.7648079, -21.6957651, -16.8334426, -12.2052725, -7.8386872, -3.76111894, 0, -4.42420863, -7.97005689, -10.7158349, -12.7398329, -14.1203411, -14.9356495, -15.2640484, -15.1838278, -14.7732781, -14.1106893, -13.2743516, -12.3425552, -11.3935903, -10.505747, -9.7573154, -9.22658579, -8.79269911, -8.50549565, -8.89372204, -9.99299361, -11.8389257, -14.4671337, -17.9132329, -22.2128386, -27.4015663, -33.5150312, -40.5888486, -48.6586341, -57.7600028, -67.0361731, -77.7423697, -90.1819269, -104.658179, -121.29951, -140.496729, -162.687301, -188.227044, -217.471778, -250.777322, -288.499498, -331.540522, -379.54352, -432.757409, -491.431104, -555.813521, -625.336789, -701.270807, -784.068686, -874.183538, -972.068476, -1076.21994, -1191.4392, -1320.73063, -1465.99033, -1626.85166, -1803.96392, -1997.97639, -2209.30282, -2438.88156, -2687.90793, -2957.61402, -3248.20778, -3559.57083, -3892.29629, -4248.18283, -4628.47576, -5034.97717, -5469.07672, -5930.12253, -6418.61751, -6936.07908, -7481.83962, -8056.05087, -8658.86461, -9290.43259, -10118.2994, -10933.3758, -11693.9653, -12358.3714, -12884.8977, -13044.4452, -13107.6551, -13113.1464, -13199.5952, -13366.0146, -13569.727, -13811.83, -14093.4209, -14415.5973, -14779.4567, -15111.1513, -15793.5989, -16558.4946, -17137.5335, -17262.4108, -16664.8217, -15076.4614, -12448.2526, -10119.639},
        {-338.221334, -337.322127, -336.32905, -335.243553, -334.067084, -332.801092, -331.447023, -330.006327, -328.480451, -326.870845, -325.178956, -323.406232, -321.554122, -319.624074, -317.617535, -315.535956, -313.380782, -311.153464, -308.855448, -306.488184, -304.053119, -301.551702, -298.98538, -296.355603, -293.663818, -290.911474, -288.100018, -285.2309, -282.305567, -279.325467, -276.292049, -273.20676, -270.07105, -266.886367, -263.649295, -260.363816, -257.033972, -253.661368, -250.24761, -246.794304, -243.303056, -239.775473, -236.21316, -232.617723, -228.990769, -225.333902, -221.648731, -217.93686, -214.199895, -210.439443, -206.657109, -202.786051, -198.897807, -194.995467, -191.082121, -187.16086, -183.234773, -179.30695, -175.38048, -171.458454, -167.543962, -163.640094, -159.749938, -155.876586, -152.023127, -148.192652, -144.388249, -140.788318, -137.567048, -134.365969, -131.176804, -127.99128, -124.801122, -121.598055, -118.373804, -115.120095, -111.828653, -108.491204, -105.099472, -101.645183, -98.1200628, -94.5158361, -90.8242284, -87.036965, -82.0599486, -76.4506548, -70.7796096, -65.0737753, -59.3601145, -53.6655893, -48.0171622, -42.4417954, -36.9664514, -31.6180925, -26.423681, -21.4101794, -16.6045498, -12.0337548, -7.72475654, -3.7045175, 0, -3.17986169, -5.72000203, -7.68302697, -9.13154243, -10.1281544, -10.7354687, -11.0160914, -11.0326283, -10.8476854, -10.5238687, -10.1237841, -9.71003749, -9.34523482, -9.09198205, -9.0128851, -9.17054991, -9.80583067, -11.0067974, -12.4328753, -14.1848666, -16.3635731, -19.0697968, -22.4043398, -26.4680041, -31.3615916, -37.1859043, -44.0417442, -52.0299132, -61.2512135, -71.4243109, -83.1722601, -96.7105037, -112.254485, -129.159567, -148.071233, -169.864319, -195.012292, -223.988617, -257.266761, -295.320191, -340.444873, -390.737099, -446.313759, -507.291745, -573.787946, -643.524913, -719.61246, -802.766064, -893.7012, -993.133343, -1101.03147, -1219.76994, -1350.27866, -1493.73928, -1651.84752, -1826.0683, -2017.86654, -2229.52235, -2461.50017, -2713.37516, -2986.29064, -3279.60565, -3591.65281, -3922.5038, -4275.17707, -4651.33845, -5053.4226, -5483.29468, -5940.0008, -6424.18167, -6940.49741, -7484.95098, -8056.79176, -8655.26912, -9279.63247, -10081.3895, -10869.4667, -11605.0489, -12249.3208, -12763.4673, -12934.5099, -13013.9055, -13037.4803, -13132.126, -13297.6219, -13495.8039, -13728.3484, -13996.9323, -14303.2323, -14648.925, -15375.6905, -16166.2639, -16831.2585, -17181.2876, -17026.9643, -16178.9021, -14447.7141, -12001.7271, -9906.44932},
        {-346.578053, -345.566176, -344.458837, -343.25755, -341.963827, -340.579184, -339.105134, -337.543191, -335.894869, -334.161682, -332.345143, -330.446768, -328.468069, -326.41056, -324.275756, -322.065171, -319.780318, -317.422711, -314.993864, -312.495291, -309.928507, -307.295024, -304.596357, -301.83402, -299.009526, -296.12439, -293.180126, -290.178247, -287.120267, -284.007701, -280.842062, -277.624864, -274.35762, -271.041846, -267.67552, -264.261996, -260.804675, -257.305184, -253.765152, -250.186209, -246.569981, -242.918099, -239.232191, -235.513886, -231.764811, -227.986596, -224.18087, -220.349261, -216.493398, -212.614909, -208.715423, -204.724953, -200.718295, -196.698632, -192.669143, -188.633012, -184.593417, -180.553541, -176.516566, -172.485671, -168.464038, -164.454848, -160.461283, -156.486523, -152.53375, -148.606145, -144.706889, -141.011472, -137.691247, -134.39402, -131.111804, -127.836608, -124.560446, -121.275326, -117.973261, -114.646262, -111.286339, -107.885504, -104.435768, -100.929141, -97.3576358, -93.7132625, -89.9880322, -86.1739563, -81.1997754, -75.6115002, -69.9679211, -64.2955543, -58.6209161, -52.9705228, -47.3708904, -41.8485354, -36.429974, -31.1417223, -26.0102967, -21.0622134, -16.3239887, -11.8221387, -7.58317978, -3.63362813, 0, -2.50498158, -4.50402156, -6.05124187, -7.20076441, -8.0067111, -8.52320384, -8.80436454, -8.90431511, -8.87717747, -8.77707353, -8.65812519, -8.57445436, -8.58018296, -8.72943289, -9.07632608, -9.67498441, -10.9851906, -13.119018, -15.2130226, -17.4082536, -19.8457602, -22.6665917, -26.0117972, -30.022426, -34.8395272, -40.6041502, -47.457344, -55.540158, -64.9936412, -75.910444, -88.4977608, -102.911161, -119.306212, -136.503385, -155.325797, -176.952594, -201.939876, -230.843741, -264.220289, -302.625618, -349.313765, -401.32178, -458.677905, -521.410382, -589.547454, -659.44224, -735.716885, -819.318409, -911.193836, -1012.29019, -1123.90417, -1246.20573, -1378.34562, -1520.86408, -1677.13812, -1849.2711, -2039.36638, -2251.11287, -2484.66802, -2738.45963, -3013.63599, -3308.9109, -3621.57326, -3951.31143, -4301.87959, -4675.16563, -5073.93276, -5500.29582, -5953.16015, -6433.24656, -6947.35985, -7489.20085, -8057.38436, -8650.5252, -9267.23817, -10047.6874, -10813.5511, -11528.0566, -12154.4316, -12655.9034, -12830.3168, -12916.5372, -12948.6705, -13047.8005, -13213.6041, -13409.5177, -13637.0305, -13897.6316, -14192.8102, -14524.0553, -15543.8983, -16406.7614, -16984.6374, -17149.5191, -16773.3992, -15728.2705, -13886.1257, -11628.2999, -9734.89912},
        {-354.681743, -353.564079, -352.350017, -351.041085, -349.638811, -348.144724, -346.560352, -344.887224, -343.126867, -341.28081, -339.350582, -337.33771, -335.243724, -333.070151, -330.818519, -328.490358, -326.087195, -323.610559, -321.061977, -318.442979, -315.755093, -312.999847, -310.178769, -307.293388, -304.345232, -301.33583, -298.266709, -295.139398, -291.955425, -288.716319, -285.423608, -282.078821, -278.683485, -275.23913, -271.741793, -268.195863, -264.605796, -260.973297, -257.300073, -253.58783, -249.838275, -246.053115, -242.234055, -238.382802, -234.501063, -230.590545, -226.652953, -222.689994, -218.703375, -214.694801, -210.665981, -206.544406, -202.407605, -198.258894, -194.101587, -189.939001, -185.77445, -181.61125, -177.452716, -173.302164, -169.162908, -165.038263, -160.931546, -156.846072, -152.785155, -148.752111, -144.750255, -140.952553, -137.527896, -134.129709, -130.75031, -127.382017, -124.017148, -120.648019, -117.266948, -113.866254, -110.438253, -106.975263, -103.469602, -99.9135874, -96.2995365, -92.619767, -88.8665966, -85.0323428, -80.0634785, -74.4972415, -68.88344, -63.2483312, -57.6181724, -52.0192207, -46.4777335, -41.019968, -35.6721814, -30.460631, -25.411574, -20.5512677, -15.9059693, -11.5019361, -7.36542531, -3.5226942, 0, -3.70722268, -6.69797508, -9.04152246, -10.8071301, -12.0640631, -12.8815869, -13.3289667, -13.4754677, -13.3903551, -13.1428943, -12.8023504, -12.4379888, -12.1190746, -11.9148732, -11.8946497, -12.1276694, -12.803949, -14.0100042, -15.5419716, -17.4949918, -19.9642054, -23.0447531, -26.8317753, -31.4204126, -36.9058056, -43.3830949, -50.9474211, -59.6939247, -69.7177462, -80.7751711, -93.424442, -107.862356, -124.285711, -141.793938, -161.132516, -183.339558, -208.941069, -238.463057, -272.431528, -311.37249, -358.110049, -410.17269, -467.636792, -530.578733, -599.074893, -669.600526, -746.733415, -831.450219, -924.727598, -1027.54221, -1141.49038, -1266.17173, -1400.54825, -1545.00554, -1702.83575, -1876.02607, -2066.56368, -2277.58674, -2509.66986, -2762.13204, -3036.83343, -3332.08717, -3645.40611, -3976.40732, -4328.27339, -4702.55027, -5101.50197, -5526.86062, -5977.72544, -6454.68496, -6963.26342, -7498.35431, -8058.83766, -8643.5935, -9251.50184, -10026.6721, -10786.4475, -11493.4009, -12110.1048, -12599.132, -12752.6078, -12817.1838, -12828.4818, -12919.7005, -13087.8372, -13289.232, -13522.9151, -13787.9165, -14083.2665, -14407.9951, -15484.6796, -16328.0494, -16836.1148, -16906.8857, -16438.3725, -15328.5851, -13475.5338, -11375.4012, -9599.94046},
        {-362.793212, -361.578193, -360.265363, -358.856271, -357.352467, -355.7555, -354.066919, -352.288272, -350.42111, -348.466981, -346.427435, -344.30402, -342.098286, -339.811782, -337.446057, -335.00266, -332.483141, -329.889048, -327.221931, -324.483339, -321.67482, -318.797925, -315.854202, -312.845201, -309.77247, -306.637559, -303.442016, -300.187392, -296.875235, -293.507094, -290.084519, -286.609059, -283.082262, -279.505678, -275.873155, -272.190253, -268.462628, -264.69208, -260.880407, -257.029408, -253.140882, -249.216629, -245.258447, -241.268136, -237.247495, -233.198323, -229.122419, -225.021582, -220.897611, -216.752306, -212.587465, -208.327671, -204.053614, -199.768766, -195.476601, -191.180591, -186.88421, -182.59093, -178.304225, -174.027567, -169.76443, -165.518286, -161.292608, -157.090871, -152.916545, -148.773105, -144.664024, -140.760026, -137.227821, -133.725937, -130.247005, -126.783656, -123.328519, -119.874227, -116.41341, -112.938697, -109.442721, -105.918112, -102.357501, -98.7535171, -95.0987927, -91.385958, -87.6076437, -83.7564805, -78.7968622, -73.2571654, -67.6781257, -62.0857421, -56.5060136, -50.9649392, -45.4885178, -40.1027485, -34.8336301, -29.7071619, -24.7493426, -19.9861714, -15.4436471, -11.1477689, -7.12453561, -3.39994632, 0, -5.2773147, -9.56009029, -12.9373682, -15.4981898, -17.3315966, -18.5266299, -19.1723312, -19.3577418, -19.1719033, -18.703857, -18.0426443, -17.2773067, -16.4968855, -15.7904223, -15.2469583, -14.955535, -14.7442039, -14.6647222, -15.3964656, -16.9725491, -19.4260877, -22.7901964, -27.0979902, -32.3825841, -38.6770932, -46.0146324, -54.4283168, -63.9512613, -74.616581, -85.7054254, -98.2785957, -112.594796, -128.912732, -146.818521, -166.907161, -189.953006, -216.416537, -246.758234, -281.438579, -320.918051, -367.126333, -418.607556, -475.534748, -538.080936, -606.41915, -677.712052, -755.895626, -841.895493, -936.637271, -1041.04658, -1156.59214, -1282.99268, -1419.63701, -1566.95816, -1727.52112, -1902.93366, -2094.80357, -2305.16743, -2535.10676, -2785.23527, -3057.93463, -3351.68409, -3665.1109, -3998.03541, -4352.2996, -4728.81747, -5128.83952, -5553.36698, -6002.1674, -6475.70617, -6977.79571, -7505.03946, -8056.74425, -8632.21689, -9230.76418, -10003.2552, -10759.544, -11461.0467, -12069.1796, -12545.3589, -12672.704, -12709.7925, -12694.4538, -12775.9539, -12948.0115, -13157.9134, -13401.7, -13675.4117, -13975.0888, -14296.7718, -15323.0984, -16110.9758, -16565.6814, -16592.493, -16096.6882, -14983.5446, -13158.3397, -11191.4749, -9489.42555},
        {-370.95284, -369.650863, -368.248795, -366.748232, -365.150768, -363.457996, -361.671512, -359.792909, -357.823782, -355.765726, -353.620334, -351.389202, -349.073923, -346.676091, -344.197302, -341.63915, -339.003229, -336.291133, -333.504456, -330.644794, -327.71374, -324.712889, -321.643836, -318.508173, -315.307497, -312.043401, -308.71748, -305.331327, -301.886539, -298.384708, -294.827429, -291.216296, -287.552905, -283.838849, -280.066422, -276.242065, -272.37233, -268.459112, -264.504307, -260.509813, -256.477525, -252.409339, -248.307151, -244.172858, -240.008356, -235.81554, -231.596308, -227.352555, -223.086177, -218.799071, -214.493133, -210.08936, -205.672301, -201.245605, -196.812923, -192.377905, -187.9442, -183.515458, -179.09533, -174.687465, -170.295513, -165.923124, -161.573947, -157.251634, -152.959833, -148.702195, -144.482369, -140.469163, -136.827418, -133.22016, -129.640332, -126.080879, -122.534742, -118.994866, -115.454194, -111.905669, -108.342235, -104.756836, -101.142414, -97.4919132, -93.7982771, -90.0544491, -86.2533725, -82.3879908, -77.4425099, -71.9354594, -66.3974931, -60.85429, -55.331529, -49.854889, -44.4500489, -39.1426876, -33.9584839, -28.9231169, -24.0622654, -19.4016082, -14.9668244, -10.7835927, -6.87759217, -3.27450163, 0, -6.70328335, -12.1604267, -16.4784267, -19.7642796, -22.1249819, -23.6675302, -24.4989209, -24.7261504, -24.4562153, -23.796112, -22.852837, -21.7333867, -20.5447576, -19.3939462, -18.387949, -17.6337624, -16.6389294, -15.4248103, -15.4543083, -16.7059654, -19.1583238, -22.7899256, -27.5793132, -33.5050285, -40.5456137, -48.679611, -57.8855627, -68.1420107, -79.4274973, -90.5983198, -103.166755, -117.448018, -133.757326, -152.187496, -173.164943, -197.175386, -224.600761, -255.823001, -291.224042, -331.185818, -376.538481, -427.079334, -483.102618, -544.902572, -612.773436, -684.939846, -764.283046, -851.614679, -947.746384, -1053.4898, -1169.83426, -1297.19654, -1435.83675, -1586.39995, -1750.31719, -1928.66662, -2122.5264, -2332.65893, -2560.52989, -2807.9494, -3077.2677, -3367.9196, -3680.56345, -4015.5348, -4372.62224, -4751.86545, -5153.00405, -5575.99991, -6021.91504, -6491.18916, -6986.56629, -7505.82733, -8048.61433, -8614.56937, -9203.33451, -9971.04638, -10721.7288, -11415.9003, -12014.0792, -12476.784, -12579.4233, -12591.0318, -12551.4607, -12625.2, -12802.7648, -13022.7371, -13278.2286, -13562.3507, -13868.2152, -14188.9337, -15093.529, -15806.1416, -16224.9872, -16248.2818, -15774.2413, -14701.0816, -12927.0185, -11060.0099, -9393.72023},
        {-379.327632, -377.966969, -376.500657, -374.930459, -373.258133, -371.485442, -369.614147, -367.646008, -365.582786, -363.426243, -361.17814, -358.840238, -356.414297, -353.902078, -351.305344, -348.625854, -345.86537, -343.025652, -340.108462, -337.115561, -334.04871, -330.90967, -327.700201, -324.422065, -321.077023, -317.666836, -314.193265, -310.65807, -307.063014, -303.409856, -299.700359, -295.936282, -292.119388, -288.251436, -284.328307, -280.354825, -276.335887, -272.273446, -268.169452, -264.025859, -259.844617, -255.627679, -251.376997, -247.094522, -242.782206, -238.442002, -234.075861, -229.685735, -225.273576, -220.841336, -216.390966, -211.835627, -207.267986, -202.691922, -198.111311, -193.53003, -188.951955, -184.380963, -179.820932, -175.275737, -170.749256, -166.245365, -161.767941, -157.320861, -152.908002, -148.53324, -144.200453, -140.078584, -136.332617, -132.625987, -128.95187, -125.303442, -121.673879, -118.056356, -114.444051, -110.830139, -107.207796, -103.570198, -99.910521, -96.2219414, -92.4976351, -88.7307781, -84.9145464, -81.0421162, -76.1219616, -70.6603216, -65.1752702, -59.6919389, -54.2354593, -48.8309628, -43.503581, -38.2784453, -33.1806873, -28.2354385, -23.4678303, -18.9029944, -14.5660621, -10.482165, -6.67643462, -3.17400245, 0, -6.26190938, -11.3724423, -15.4325388, -18.5431392, -20.8051837, -22.3196124, -23.1873656, -23.5093833, -23.3866059, -22.9199735, -22.2104263, -21.3589044, -20.4663481, -19.6336975, -18.9618929, -18.5518744, -18.1225271, -17.7202118, -18.3100412, -19.9110863, -22.5424184, -26.2231088, -30.9722288, -36.8088494, -43.752042, -51.8208778, -61.034428, -71.4117639, -82.9719567, -94.9021788, -108.35843, -123.609351, -140.923582, -160.276951, -182.084504, -206.839375, -234.898048, -266.61701, -302.352744, -342.461736, -387.15824, -436.984261, -492.324112, -553.562107, -621.082559, -694.19587, -774.628743, -863.033968, -960.064338, -1066.37264, -1182.09915, -1309.03559, -1448.65997, -1602.02047, -1769.28782, -1951.02668, -2147.80174, -2359.8944, -2588.21701, -2833.99098, -3097.07163, -3380.51496, -3689.48746, -4025.74229, -4385.24889, -4766.6315, -5166.76909, -5583.83343, -6022.39574, -6483.40735, -6972.89748, -7485.95837, -8021.78363, -8579.56686, -9158.50168, -9907.13419, -10637.9674, -11312.8567, -11893.6577, -12342.2257, -12439.4627, -12448.8134, -12409.6851, -12488.4364, -12674.2226, -12903.3025, -13167.4765, -13458.5446, -13768.3071, -14088.5642, -14867.5771, -15523.205, -15936.4975, -15988.5043, -15560.2749, -14532.8588, -12787.3056, -10927.8184, -9270.6994},
        {-387.751961, -386.336967, -384.810387, -383.174157, -381.430218, -379.580507, -377.626964, -375.571525, -373.416131, -371.162718, -368.813227, -366.369596, -363.833762, -361.207665, -358.493242, -355.692434, -352.807177, -349.83941, -346.791073, -343.664103, -340.460439, -337.18202, -333.830784, -330.408669, -326.917614, -323.359558, -319.736439, -316.050195, -312.302766, -308.496089, -304.632103, -300.712747, -296.739958, -292.715677, -288.639527, -284.514654, -280.344228, -276.130263, -271.874772, -267.579769, -263.247267, -258.87928, -254.477822, -250.044905, -245.582545, -241.092753, -236.577545, -232.038932, -227.47893, -222.899551, -218.302809, -213.59484, -208.875614, -204.149222, -199.419757, -194.691311, -189.967974, -185.25384, -180.553001, -175.869548, -171.207573, -166.571168, -161.964426, -157.391438, -152.856296, -148.363093, -143.91592, -139.683192, -135.82938, -132.01967, -128.247504, -124.506319, -120.789556, -117.090654, -113.403053, -109.720192, -106.035511, -102.342449, -98.6344469, -94.9049429, -91.1473771, -87.355189, -83.5218182, -79.6407042, -74.7493472, -69.3392635, -63.9134587, -58.4963946, -53.1125326, -47.7863346, -42.5422621, -37.4047767, -32.3983402, -27.547414, -22.8764599, -18.4099395, -14.1723144, -10.1880463, -6.48159675, -3.07742744, 0, -5.4117809, -9.83982594, -13.3741549, -16.1047878, -18.1217442, -19.5150442, -20.3747074, -20.7907538, -20.8532032, -20.6520754, -20.2773903, -19.8191677, -19.3674274, -19.0121892, -18.843473, -18.9512987, -19.3510272, -20.1220127, -21.5231464, -23.6284498, -26.5119444, -30.2476516, -34.9095931, -40.5717903, -47.3082647, -55.1930378, -64.3001311, -74.7035661, -86.4773643, -99.2442445, -113.695009, -130.039069, -148.485838, -168.824605, -191.474845, -216.968063, -245.639707, -277.825227, -313.860071, -354.079687, -398.126957, -447.240678, -501.891804, -562.551284, -629.690071, -703.797829, -785.322117, -874.72921, -972.485381, -1079.0569, -1193.59255, -1319.48637, -1459.49044, -1615.10574, -1785.27895, -1970.10359, -2169.67316, -2384.00352, -2613.28348, -2857.78657, -3114.5657, -3390.49618, -3695.26002, -4032.06794, -4393.16535, -4775.83089, -5174.1647, -5584.62137, -6015.3101, -6467.74758, -6952.38887, -7460.43716, -7990.31508, -8540.44527, -9109.25038, -9832.0056, -10536.0679, -11185.6467, -11744.9515, -12178.1919, -12277.6283, -12294.0519, -12265.3646, -12354.1974, -12549.0804, -12786.5728, -13057.8032, -13353.8999, -13665.9914, -13985.2061, -14619.0451, -15216.6879, -15636.653, -15737.4586, -15377.6232, -14415.6653, -12710.1031, -10838.058, -9163.58804},
        {-396.18132, -394.707652, -393.117043, -391.411585, -389.59337, -387.664489, -385.627035, -383.4831, -381.234774, -378.88415, -376.43332, -373.884376, -371.239409, -368.500512, -365.669776, -362.749293, -359.741154, -356.647453, -353.47028, -350.211728, -346.873888, -343.458852, -339.968712, -336.40556, -332.771487, -329.068586, -325.298948, -321.464666, -317.567831, -313.610534, -309.594869, -305.522925, -301.396797, -297.218574, -292.990268, -288.714012, -284.391942, -280.026153, -275.618739, -271.171795, -266.687415, -262.167695, -257.614727, -253.030608, -248.417431, -243.777291, -239.112283, -234.424501, -229.716039, -224.988992, -220.245455, -215.38728, -210.518976, -205.644812, -200.769054, -195.895971, -191.02983, -186.174899, -181.335444, -176.515735, -171.720038, -166.95262, -162.217751, -157.519696, -152.862724, -148.251102, -143.689098, -139.342829, -135.374524, -131.454651, -127.576984, -123.7353, -119.923373, -116.134981, -112.363897, -108.603898, -104.848759, -101.092256, -97.3281647, -93.5502602, -89.7523182, -85.9281144, -82.0714242, -78.1760233, -73.3143208, -67.960155, -62.5984388, -57.2528479, -51.9470578, -46.7047443, -41.5495832, -36.50525, -31.5954205, -26.8437704, -22.2739753, -17.909711, -13.7746531, -9.89247743, -6.28685953, -2.98147516, 0, -4.4528948, -8.10515314, -11.0351055, -13.3210823, -15.041414, -16.2744311, -17.098464, -17.5918432, -17.8328992, -17.8999624, -17.8713634, -17.8254324, -17.8405001, -17.9948968, -18.366953, -19.0349993, -20.2905194, -22.2422942, -24.4865217, -27.1472079, -30.348359, -34.2139813, -38.8680809, -44.434664, -51.0377369, -58.8013056, -67.8493764, -78.3059554, -90.295049, -103.82248, -119.173771, -136.508385, -155.985782, -177.257148, -200.736083, -226.971724, -256.276016, -288.960901, -325.338323, -365.720226, -409.194219, -457.669203, -511.696666, -571.828094, -638.614977, -713.736331, -796.33423, -886.678278, -985.038082, -1091.68324, -1204.76661, -1329.2617, -1469.03055, -1625.99632, -1798.12351, -1985.15396, -2186.82954, -2403.10037, -2633.45269, -2877.14549, -3128.79703, -3398.41416, -3699.19964, -4035.52827, -4396.81644, -4779.34667, -5175.12207, -5579.31568, -6002.79187, -6447.53909, -6928.88693, -7433.43026, -7958.53945, -8501.58488, -9059.93694, -9750.34492, -10420.9556, -11039.2945, -11572.8874, -11989.26, -12097.5672, -12129.2857, -12119.8144, -12222.8385, -12426.935, -12671.395, -12947.2599, -13245.5709, -13557.3694, -13873.6969, -14357.3489, -14897.6022, -15329.6313, -15488.6107, -15209.7151, -14328.1191, -12678.9971, -10798.8036, -9084.53262},
        {-404.288181, -402.707269, -401.00991, -399.198129, -397.273948, -395.23939, -393.096478, -390.847237, -388.493687, -386.037853, -383.481758, -380.827425, -378.076877, -375.232137, -372.295228, -369.268173, -366.152995, -362.951718, -359.666364, -356.298956, -352.851519, -349.326073, -345.724644, -342.049254, -338.301925, -334.484682, -330.599547, -326.648543, -322.633693, -318.557021, -314.420549, -310.226301, -305.976299, -301.672567, -297.309288, -292.892569, -288.428613, -283.919698, -279.368101, -274.7761, -270.145971, -265.479994, -260.780445, -256.049601, -251.289741, -246.503141, -241.69208, -236.858835, -232.005683, -227.134902, -222.248769, -217.256249, -212.254956, -207.249189, -202.243249, -197.241437, -192.248052, -187.267396, -182.303768, -177.361469, -172.444799, -167.558059, -162.705549, -157.89157, -153.120421, -148.396403, -143.723817, -139.260776, -135.161699, -131.11366, -127.110988, -123.148013, -119.219064, -115.31847, -111.44056, -107.579664, -103.730111, -99.8862307, -96.042352, -92.1928044, -88.3319172, -84.4540197, -80.5534411, -76.6245109, -71.7841609, -66.483874, -61.1856787, -55.9123776, -50.6867734, -45.5316684, -40.4698654, -35.5241667, -30.7173751, -26.072293, -21.611723, -17.3584677, -13.3353295, -9.56511112, -6.07061503, -2.87464381, 0, -4.27673842, -7.78022516, -10.58873, -12.7805229, -14.4338735, -15.6270516, -16.4383272, -16.9459699, -17.2282497, -17.3634363, -17.4297995, -17.5056092, -17.6691351, -17.9986471, -18.572415, -19.4687086, -20.8916109, -22.9373649, -25.3996632, -28.3837355, -31.9948115, -36.3381211, -41.518894, -47.6423601, -54.813749, -63.1382905, -72.7212145, -83.6677506, -96.0831287, -109.772277, -125.250838, -142.714131, -162.357475, -184.18495, -208.487497, -235.607053, -265.796313, -299.307967, -336.39471, -377.309233, -420.359078, -468.334091, -521.867539, -581.592686, -648.142799, -724.364602, -808.124537, -899.502506, -998.57841, -1105.43215, -1217.51893, -1340.75132, -1479.24664, -1635.08879, -1806.21, -1992.40651, -2193.47457, -2409.54164, -2639.99746, -2883.87039, -3135.94806, -3405.62377, -3705.05389, -4038.3108, -4395.76896, -4774.09041, -5165.96414, -5567.02213, -5987.46416, -6429.24965, -6912.05775, -7417.402, -7941.10807, -8479.00161, -9026.90828, -9674.06868, -10299.5398, -10875.7935, -11375.3017, -11770.5364, -11896.7612, -11955.1285, -11976.9135, -12098.7927, -12310.3203, -12557.1334, -12830.9819, -13123.6159, -13426.7853, -13732.2402, -14082.1429, -14564.8506, -14999.8877, -15206.7784, -15005.0471, -14214.218, -12653.8155, -10841.4647, -9044.1751},
        {-412.375627, -410.673832, -408.85762, -406.928888, -404.889534, -402.741455, -400.486548, -398.126711, -395.663841, -393.099835, -390.436591, -387.676007, -384.819979, -381.870406, -378.829183, -375.69821, -372.479382, -369.174598, -365.785755, -362.314751, -358.763482, -355.133846, -351.42774, -347.647062, -343.79371, -339.86958, -335.87657, -331.816577, -327.691499, -323.503233, -319.253677, -314.944727, -310.578281, -306.156238, -301.662207, -297.107614, -292.504104, -287.854168, -283.160296, -278.424979, -273.650708, -268.839974, -263.995267, -259.119079, -254.213899, -249.28222, -244.326531, -239.349324, -234.353089, -229.340316, -224.313498, -219.193752, -214.066706, -208.936614, -203.807731, -198.684312, -193.570611, -188.470883, -183.389383, -178.330366, -173.298087, -168.2968, -163.33076, -158.404222, -153.52144, -148.68667, -143.904167, -139.320715, -135.079679, -130.891355, -126.750756, -122.652898, -118.592796, -114.565465, -110.565921, -106.589177, -102.63025, -98.6841541, -94.7459045, -90.8105162, -86.8730045, -82.9283842, -78.9716704, -74.9978783, -70.1760016, -64.9314403, -59.6995224, -54.5020689, -49.3609009, -44.2978392, -39.334705, -34.4933194, -29.7955032, -25.2630776, -20.9178636, -16.7816822, -12.8763544, -9.22370126, -5.8455438, -2.76370303, 0, -4.45688243, -8.09447055, -10.9965005, -13.2467085, -14.9288307, -16.1266032, -16.9237621, -17.4040437, -17.6511842, -17.7489196, -17.7809861, -17.8311199, -17.9830572, -18.320534, -18.9272867, -19.8870512, -21.1394058, -22.7474924, -25.1211193, -28.3131314, -32.3763741, -37.3636923, -43.3279312, -50.321936, -58.3985518, -67.6106238, -78.010997, -89.6525167, -102.588028, -116.140316, -131.359975, -148.518867, -167.888856, -190.037243, -215.088173, -243.087004, -274.216969, -308.661298, -346.603223, -388.225974, -431.064751, -478.756741, -532.003267, -591.505655, -657.965228, -735.227914, -820.064285, -912.389515, -1012.11878, -1119.16725, -1230.52643, -1352.60854, -1489.2888, -1642.65494, -1811.14523, -1994.8365, -2193.80556, -2408.51227, -2638.56334, -2883.14757, -3138.58407, -3411.68857, -3711.15811, -4039.99394, -4391.54607, -4763.59458, -5151.05354, -5550.96001, -5970.85969, -6412.35401, -6898.60056, -7406.32824, -7929.67672, -8462.78566, -8999.79474, -9598.97915, -10174.3091, -10703.8905, -11165.8289, -11538.2303, -11686.4286, -11776.4826, -11834.8287, -11977.4071, -12195.2257, -12441.9996, -12710.6015, -12993.904, -13284.7799, -13576.1019, -13814.8306, -14244.316, -14675.8218, -14920.6116, -14789.9492, -14095.0981, -12647.3221, -10945.4592, -9056.50968},
        {-420.690164, -418.881681, -416.958887, -414.92363, -412.777758, -410.523117, -408.161555, -405.694919, -403.125057, -400.453816, -397.683043, -394.814586, -391.850291, -388.792006, -385.641579, -382.400856, -379.071685, -375.655914, -372.155389, -368.571958, -364.907468, -361.163766, -357.3427, -353.446118, -349.475865, -345.43379, -341.32174, -337.141563, -332.895105, -328.584213, -324.210736, -319.776521, -315.283414, -310.733263, -306.102366, -301.405884, -296.659281, -291.865234, -287.02642, -282.145514, -277.225194, -272.268136, -267.277017, -262.254511, -257.203297, -252.12605, -247.025447, -241.904164, -236.764878, -231.610265, -226.443001, -221.196395, -215.943996, -210.689982, -205.438535, -200.193834, -194.96006, -189.741392, -184.542012, -179.366099, -174.217833, -169.101395, -164.020964, -158.980721, -153.984846, -149.03752, -144.142921, -139.435001, -135.044592, -130.708234, -126.421697, -122.180748, -117.981155, -113.818684, -109.689105, -105.588185, -101.511691, -97.4553918, -93.4150546, -89.3864473, -85.3653376, -81.3474933, -77.3286822, -73.304672, -68.5025042, -63.3184788, -58.1581353, -53.0421882, -47.9913518, -43.0263406, -38.167869, -33.4366514, -28.8534022, -24.4388359, -20.2136667, -16.1986091, -12.4143776, -8.88168653, -5.6212503, -2.65378332, 0, -4.6851301, -8.49104464, -11.508031, -13.8263766, -15.5363688, -16.7282951, -17.4924427, -17.9190991, -18.0985517, -18.1210879, -18.076995, -18.0565605, -18.1500718, -18.4478162, -19.0400811, -20.0171539, -21.0159132, -22.0618737, -24.2708903, -27.6360915, -32.1506056, -37.8075609, -44.6000857, -52.5213086, -61.5643577, -71.7223615, -82.9884483, -95.3557464, -108.817384, -122.160657, -137.026665, -153.770286, -172.7464, -195.080083, -220.741115, -249.493888, -281.462219, -316.769929, -355.540834, -397.898755, -440.804322, -488.507249, -541.750241, -601.276, -667.827228, -745.944024, -831.622347, -924.655551, -1024.83699, -1131.96002, -1242.75495, -1363.8219, -1498.55979, -1648.97775, -1814.24728, -1994.81394, -2191.12327, -2404.04634, -2633.50645, -2878.96268, -3138.69774, -3416.33283, -3716.33317, -4040.39019, -4385.49541, -4750.79767, -5134.03166, -5533.97959, -5954.60899, -6396.95426, -6886.37789, -7395.84345, -7917.96412, -8445.35306, -8970.62345, -9522.32423, -10048.1488, -10531.7265, -10956.6865, -11306.658, -11476.5552, -11597.8658, -11692.2396, -11854.8438, -12078.2278, -12324.3088, -12587.1678, -12860.8859, -13139.5439, -13417.223, -13574.8647, -13959.4609, -14381.8909, -14653.0339, -14583.7688, -13984.9751, -12667.5317, -11094.8697, -9135.22986},
        {-429.7245, -427.890143, -425.931247, -423.850019, -421.648662, -419.329381, -416.894382, -414.34587, -411.68605, -408.917126, -406.041304, -403.060789, -399.977785, -396.794498, -393.513132, -390.135894, -386.664986, -383.102616, -379.450987, -375.712305, -371.888774, -367.9826, -363.995988, -359.931142, -355.790268, -351.57557, -347.289254, -342.933525, -338.510587, -334.022646, -329.471906, -324.860573, -320.190851, -315.464947, -310.670685, -305.817762, -300.916049, -295.968217, -290.976937, -285.944882, -280.874722, -275.769129, -270.630776, -265.462333, -260.266472, -255.045864, -249.803182, -244.541098, -239.262281, -233.969405, -228.665141, -223.284398, -217.899079, -212.513325, -207.131275, -201.757072, -196.394854, -191.048764, -185.72294, -180.421525, -175.148658, -169.90848, -164.705132, -159.542754, -154.425486, -149.35747, -144.342845, -139.502964, -134.956363, -130.465444, -126.026751, -121.636824, -117.292206, -112.98944, -108.725067, -104.495631, -100.297672, -96.1277343, -91.9823591, -87.8580888, -83.7514658, -79.6590323, -75.5773305, -71.5029027, -66.7290988, -61.6188478, -56.5435309, -51.5225102, -46.5751479, -41.7208063, -36.9788475, -32.3686337, -27.9095272, -23.6208901, -19.5220847, -15.6324731, -11.9714175, -8.55828025, -5.41242344, -2.55320928, 0, -4.5571159, -8.23440655, -11.1245411, -13.3201889, -14.9140189, -15.9987003, -16.6669024, -17.0112944, -17.1245453, -17.0993245, -17.0283009, -17.0041439, -17.1195227, -17.4671063, -18.139564, -19.2295649, -20.1824699, -20.9984756, -23.234401, -26.8442061, -31.7818514, -38.001297, -45.4565032, -54.1014302, -63.8900384, -74.776288, -86.7141392, -99.6575523, -113.560488, -126.875818, -141.608989, -158.164289, -176.946003, -199.319681, -225.208978, -254.281213, -286.652292, -322.438123, -361.754614, -404.717672, -448.406117, -496.897275, -550.901266, -611.128212, -678.288232, -756.815916, -842.765798, -935.91688, -1036.04817, -1142.93866, -1253.38536, -1373.79395, -1507.03877, -1655.00985, -1817.58764, -1995.55486, -2189.6942, -2401.42241, -2630.74407, -2876.97214, -3139.47037, -3419.88834, -3719.7918, -4040.04454, -4380.32087, -4740.8411, -5121.83848, -5523.53667, -5946.11182, -6389.76696, -6878.79265, -7385.59859, -7902.04983, -8420.01139, -8931.3483, -9444.85778, -9931.23964, -10378.1258, -10773.1483, -11103.9392, -11286.4247, -11427.7463, -11546.2667, -11723.5863, -11953.0365, -12201.5999, -12463.9254, -12734.6619, -13008.4582, -13279.9633, -13418.9009, -13776.9953, -14179.2077, -14450.4993, -14415.8312, -13900.1647, -12728.4608, -11271.5178, -9314.26466},
        {-439.810344, -438.073492, -436.186165, -434.151545, -431.972814, -429.653154, -427.195746, -424.603771, -421.880413, -419.028851, -416.052269, -412.953847, -409.736767, -406.404212, -402.959362, -399.4054, -395.745508, -391.982866, -388.120656, -384.162061, -380.110263, -375.968441, -371.73978, -367.427459, -363.034661, -358.564568, -354.020361, -349.405222, -344.722333, -339.974875, -335.16603, -330.29898, -325.376907, -320.402991, -315.40528, -310.373998, -305.299076, -300.182887, -295.027806, -289.836208, -284.610467, -279.35296, -274.066059, -268.752142, -263.413581, -258.052752, -252.67203, -247.27379, -241.860406, -236.434253, -230.997706, -225.469778, -219.938012, -214.406592, -208.879698, -203.361513, -197.856218, -192.367997, -186.90103, -181.459499, -176.047587, -170.669476, -165.329347, -160.031382, -154.779764, -149.578674, -144.432295, -139.450989, -134.742572, -130.092233, -125.497271, -120.954985, -116.462673, -112.017634, -107.617167, -103.25857, -98.9391416, -94.6561811, -90.4069867, -86.1888572, -81.9990912, -77.8349872, -73.693844, -69.5729601, -64.8410374, -59.8228946, -54.8509058, -49.9427686, -45.116181, -40.3888408, -35.7784458, -31.3026937, -26.9792825, -22.8259099, -18.8602738, -15.1000719, -11.5630021, -8.2667622, -5.22905, -2.46756333, 0, -3.80076397, -6.83052176, -9.17649507, -10.9259056, -12.1659751, -12.9839252, -13.4669777, -13.7023543, -13.7772766, -13.7789664, -13.7946454, -13.9115353, -14.2168579, -14.7978347, -15.7416876, -17.1356382, -18.4546805, -19.6985754, -22.3393449, -26.3330193, -31.6356286, -38.2032028, -45.991772, -54.9573664, -65.0560159, -76.2437506, -88.4766007, -101.710596, -115.901767, -129.557074, -144.612942, -161.460122, -180.489365, -202.771263, -228.356648, -257.115057, -289.23329, -324.898146, -364.296422, -407.614917, -453.058893, -503.399763, -559.212016, -621.070144, -689.548637, -767.86448, -853.289046, -945.736203, -1045.11982, -1151.35376, -1261.70217, -1381.96719, -1514.60596, -1661.38515, -1822.66168, -1999.42544, -2192.66629, -2404.48084, -2634.50071, -2881.15018, -3143.08979, -3422.50638, -3720.90603, -4039.31346, -4377.93774, -4737.36235, -5119.28343, -5524.57292, -5950.02295, -6394.73343, -6877.38361, -7374.48184, -7878.35055, -8381.31219, -8875.6892, -9366.66628, -9830.48804, -10256.2614, -10633.0931, -10950.0901, -11129.8936, -11272.3866, -11395.1614, -11578.5019, -11815.1849, -12071.8656, -12342.6267, -12621.5512, -12902.7217, -13180.2212, -13384.0295, -13740.775, -14108.1492, -14343.844, -14305.551, -13850.962, -12837.7686, -11461.6957, -9617.45018},
        {-450.224459, -448.593411, -446.784789, -444.80279, -442.65161, -440.335445, -437.858492, -435.224947, -432.439006, -429.504866, -426.426723, -423.208773, -419.855213, -416.370239, -412.758047, -409.022834, -405.168796, -401.20013, -397.121031, -392.935696, -388.648321, -384.263104, -379.784239, -375.215923, -370.562354, -365.827726, -361.016237, -356.132082, -351.179459, -346.162563, -341.08559, -335.952738, -330.768202, -325.536179, -320.325916, -315.107715, -309.851097, -304.558148, -299.230953, -293.871598, -288.482167, -283.064746, -277.62142, -272.154275, -266.665396, -261.156869, -255.630777, -250.089208, -244.534246, -238.967977, -233.392485, -227.710808, -222.026226, -216.342973, -210.66528, -204.997381, -199.343507, -193.707892, -188.094767, -182.508366, -176.95292, -171.432662, -165.951825, -160.514641, -155.125343, -149.788163, -144.507334, -139.382743, -134.51198, -129.701998, -124.950828, -120.256501, -115.617049, -111.030504, -106.494897, -102.008261, -97.5686256, -93.1740236, -88.8224864, -84.5120454, -80.2407324, -76.006579, -71.8076166, -67.641877, -62.9534577, -58.0290084, -53.1618371, -48.3679515, -43.6633595, -39.0640688, -34.5860873, -30.2454229, -26.0580832, -22.0400763, -18.2074099, -14.5760918, -11.1621298, -7.98153186, -5.05030567, -2.38445911, 0, -3.01083785, -5.36784663, -7.15175978, -8.44331068, -9.32323276, -9.87225943, -10.1711241, -10.3005601, -10.341301, -10.3740801, -10.4796308, -10.7386866, -11.2319808, -12.0402469, -13.2442182, -14.9246282, -16.6590033, -18.4561902, -21.5351286, -25.8687218, -31.4298732, -38.1914862, -46.126464, -55.2077099, -65.4081275, -76.7006198, -89.0580904, -102.453443, -116.85958, -130.974651, -146.513724, -163.832128, -183.285191, -205.509153, -230.718887, -259.05436, -290.786629, -326.18675, -365.52578, -409.074775, -456.313137, -508.544517, -566.19486, -629.690112, -699.456219, -777.262158, -861.855085, -953.325188, -1051.76265, -1157.25767, -1267.66912, -1388.04565, -1520.49345, -1666.64898, -1827.1896, -2003.22329, -2195.85801, -2407.63825, -2637.909, -2884.44815, -3145.23384, -3423.15568, -3720.09153, -4037.31817, -4375.09385, -4734.14435, -5117.2153, -5525.55614, -5953.01016, -6397.60999, -6871.4745, -7356.38296, -7845.49201, -8331.95826, -8808.93834, -9283.4628, -9731.34586, -10142.2757, -10505.9403, -10812.0281, -10985.2545, -11123.5953, -11244.5838, -11430.0781, -11672.1498, -11936.0688, -12215.2468, -12503.0954, -12793.0266, -13078.452, -13379.137, -13744.9031, -14073.222, -14261.5651, -14207.4041, -13808.2103, -12961.4554, -11680.9892, -10002.4877},
        {-460.767447, -459.2137, -457.458773, -455.507726, -453.365621, -451.037519, -448.528481, -445.843569, -442.987844, -439.966367, -436.7842, -433.446402, -429.958037, -426.324164, -422.549846, -418.640143, -414.600117, -410.434829, -406.14934, -401.748711, -397.238004, -392.622279, -387.906599, -383.096024, -378.195616, -373.210435, -368.145543, -363.006002, -357.796872, -352.523214, -347.190091, -341.802563, -336.365691, -330.884536, -325.461471, -320.050855, -314.605876, -309.128439, -303.620447, -298.083803, -292.52041, -286.932172, -281.320992, -275.688773, -270.037419, -264.368833, -258.684918, -252.987578, -247.278716, -241.560235, -235.834038, -229.992448, -224.149324, -218.308946, -212.475592, -206.653541, -200.847073, -195.060466, -189.297998, -183.563949, -177.862598, -172.198223, -166.575103, -160.997517, -155.469744, -149.996063, -144.580752, -139.313863, -134.283384, -129.316699, -124.412526, -119.569583, -114.786591, -110.062266, -105.395328, -100.784496, -96.2284876, -91.7260221, -87.2758179, -82.8765937, -78.5270682, -74.22596, -69.9719878, -65.7638702, -61.117002, -56.2819298, -51.5150096, -46.8306446, -42.2432384, -37.7671945, -33.4169162, -29.2068069, -25.1512702, -21.2647095, -17.5615281, -14.0561296, -10.7629174, -7.69629485, -4.87066549, -2.30043273, 0, -2.35166923, -4.14904674, -5.46693713, -6.380145, -6.96347495, -7.29173159, -7.43971951, -7.48224333, -7.49410763, -7.55011703, -7.72507613, -8.09378952, -8.73106181, -9.71169761, -11.1105015, -13.0022781, -15.1199539, -17.4894939, -20.9589982, -25.5300119, -31.2040799, -37.9827474, -45.8675593, -54.8600606, -64.9617964, -76.1743116, -88.4991513, -101.93786, -116.491984, -131.154468, -147.305276, -165.248534, -185.288365, -207.578327, -232.49783, -260.466431, -291.833427, -326.948111, -366.159779, -409.817726, -458.54661, -512.336556, -571.482983, -636.281309, -707.026955, -784.099246, -867.688718, -958.069812, -1055.51697, -1160.30464, -1270.9127, -1391.60349, -1524.25516, -1670.38378, -1830.76618, -2006.5111, -2198.72729, -2409.99064, -2639.60928, -2885.29084, -3144.81291, -3421.39216, -3717.24574, -4033.79326, -4371.10295, -4729.86334, -5113.41509, -5523.1343, -5950.67272, -6393.18278, -6856.45447, -7327.67373, -7801.00304, -8270.60485, -8730.64165, -9193.13204, -9629.9183, -10030.6989, -10385.1722, -10683.0367, -10848.4597, -10980.3581, -11096.5152, -11281.046, -11525.4774, -11793.6977, -12078.6156, -12373.1397, -12670.1788, -12962.6415, -13376.7696, -13757.8431, -14046.3996, -14182.9767, -14108.1119, -13762.3429, -13086.2073, -11930.4381, -10459.9179},
        {-471.178999, -469.632186, -467.86959, -465.896762, -463.719253, -461.342612, -458.772393, -456.014144, -453.073417, -449.955763, -446.666732, -443.211875, -439.596744, -435.826888, -431.90786, -427.845208, -423.644485, -419.311242, -414.851028, -410.269395, -405.571893, -400.764074, -395.851489, -390.839687, -385.73422, -380.540639, -375.264494, -369.911336, -364.486717, -358.996186, -353.445295, -347.839595, -342.184636, -336.485969, -330.859916, -325.254313, -319.615663, -313.945923, -308.247048, -302.520994, -296.769718, -290.995176, -285.199324, -279.384117, -273.551512, -267.703465, -261.841933, -255.96887, -250.086233, -244.195979, -238.300062, -232.291881, -226.284303, -220.281638, -214.288196, -208.308286, -202.346217, -196.406299, -190.492842, -184.610154, -178.762546, -172.954326, -167.189804, -161.47329, -155.809093, -150.201522, -144.654886, -139.25045, -134.067628, -128.952019, -123.902943, -118.919722, -114.001675, -109.148124, -104.35839, -99.631794, -94.967656, -90.3652971, -85.8240383, -81.3432003, -76.9221039, -72.5600701, -68.2564195, -64.0104731, -59.3986635, -54.6406505, -49.9611766, -45.3732863, -40.8900239, -36.5244338, -32.2895603, -28.1984478, -24.2641407, -20.4996833, -16.9181199, -13.5324951, -10.355853, -7.40123808, -4.68169471, -2.21026722, 0, -2.06848166, -3.6256955, -4.74345145, -5.49355946, -5.94782946, -6.1780714, -6.25609523, -6.25371089, -6.24272831, -6.29495745, -6.48220823, -6.87629061, -7.54901454, -8.57218994, -10.0176268, -11.957135, -14.3092961, -17.1240305, -20.8197358, -25.4353874, -31.0099606, -37.5824307, -45.191773, -53.8769627, -63.6769753, -74.6307859, -86.77737, -100.155703, -114.804759, -130.084511, -146.961905, -165.679619, -186.48033, -209.089691, -234.008891, -261.87699, -293.091772, -328.051019, -367.152516, -410.794044, -460.269589, -514.807976, -574.631646, -639.963038, -711.024592, -787.237115, -869.825088, -959.21136, -1055.81878, -1160.07019, -1270.95257, -1392.07961, -1525.33214, -1672.14592, -1833.04791, -2008.97311, -2200.8565, -2410.57261, -2637.90337, -2881.60578, -3140.43853, -3416.77891, -3712.5013, -4028.77267, -4365.56128, -4723.38564, -5105.66639, -5513.67442, -5938.03942, -6375.41034, -6826.89815, -7284.00333, -7741.8303, -8195.48347, -8640.06726, -9093.08939, -9521.65014, -9915.25309, -10263.4018, -10555.5999, -10715.1799, -10841.9304, -10953.7142, -11134.9062, -11377.0412, -11643.9406, -11928.5249, -12223.7146, -12522.4302, -12817.5922, -13346.7747, -13745.5755, -13997.2575, -14085.0836, -13992.3167, -13702.2195, -13198.055, -12216.6853, -10986.229},
        {-480.610388, -478.909371, -477.00077, -474.889656, -472.581104, -470.080186, -467.391976, -464.521546, -461.473971, -458.254322, -454.867674, -451.319099, -447.613671, -443.756462, -439.752547, -435.606997, -431.324887, -426.911289, -422.371277, -417.709924, -412.932302, -408.043486, -403.048548, -397.952561, -392.760599, -387.477735, -382.109042, -376.659593, -371.134461, -365.53872, -359.877442, -354.155701, -348.378571, -342.551123, -336.753662, -330.952063, -325.111299, -319.233999, -313.322797, -307.380324, -301.409212, -295.412093, -289.391598, -283.350359, -277.291008, -271.216177, -265.128498, -259.030602, -252.925122, -246.814688, -240.701933, -234.511159, -228.325026, -222.147863, -215.984, -209.837766, -203.713493, -197.615509, -191.548145, -185.515729, -179.522593, -173.573066, -167.671477, -161.822157, -156.029435, -150.297641, -144.631105, -139.106503, -133.798773, -128.563036, -123.398934, -118.306107, -113.284193, -108.332834, -103.451669, -98.6403376, -93.8984804, -89.2257372, -84.621748, -80.0861526, -75.6185911, -71.2187033, -66.8861293, -62.620509, -58.0233913, -53.3018351, -48.6676837, -44.1334961, -39.711831, -35.4152472, -31.2563034, -27.2475585, -23.4015712, -19.7309003, -16.2481045, -12.9657427, -9.8963736, -7.05255601, -4.44684869, -2.09181043, 0, -3.18795368, -5.68789001, -7.58141353, -8.95012879, -9.87564034, -10.4395527, -10.7234705, -10.8089982, -10.7777404, -10.7113016, -10.6912865, -10.7992994, -11.116945, -11.7258278, -12.7075524, -14.1437234, -16.1212822, -18.7225958, -22.0171971, -26.0878341, -31.0172552, -36.8882084, -43.783442, -51.7857041, -60.9777431, -71.4423069, -83.2621439, -96.5200022, -111.29863, -127.386339, -145.268274, -165.115514, -187.099139, -210.791024, -236.66185, -265.342086, -297.18257, -332.534144, -371.747647, -415.173919, -463.272507, -516.25246, -574.443348, -638.174741, -707.776212, -783.325612, -865.467161, -954.593359, -1051.09671, -1155.36971, -1266.27673, -1387.60612, -1522.07521, -1671.24038, -1834.2877, -2011.46744, -2203.02991, -2407.82551, -2627.82255, -2866.51649, -3127.8352, -3408.94799, -3708.26273, -4025.18207, -4360.79426, -4715.41382, -5091.77809, -5490.82974, -5904.62776, -6330.25625, -6769.51168, -7214.04081, -7659.29655, -8100.73181, -8533.79949, -8976.2167, -9395.60609, -9781.85454, -10124.8489, -10414.476, -10578.369, -10710.171, -10826.449, -11002.5906, -11231.8798, -11483.0919, -11750.7362, -12029.3221, -12313.359, -12597.3562, -13232.9237, -13650.0955, -13872.2052, -13922.5867, -13824.5736, -13601.4998, -13276.699, -12600.5031, -11635.3644},
        {-490.273963, -488.338142, -486.210514, -483.8954, -481.397118, -478.719989, -475.868331, -472.846465, -469.65871, -466.309385, -462.802811, -459.143306, -455.33519, -451.382783, -447.290404, -443.062373, -438.70301, -434.216633, -429.607563, -424.880119, -420.03862, -415.087387, -410.030738, -404.872993, -399.618473, -394.271495, -388.836381, -383.317449, -377.719018, -372.04541, -366.300942, -360.489935, -354.616709, -348.685582, -342.72907, -336.736801, -330.698066, -324.616269, -318.494814, -312.337106, -306.14655, -299.92655, -293.680511, -287.411837, -281.123933, -274.820203, -268.504052, -262.178885, -255.848105, -249.515118, -243.183328, -236.815498, -230.456555, -224.110785, -217.782473, -211.475905, -205.195367, -198.945144, -192.729522, -186.552786, -180.419222, -174.333116, -168.298752, -162.320418, -156.402398, -150.548977, -144.764442, -139.11864, -133.679842, -128.31712, -123.030508, -117.820044, -112.685761, -107.627696, -102.645884, -97.7403607, -92.9111616, -88.1583222, -83.481878, -78.8818646, -74.3583174, -69.911272, -65.5407639, -61.2468287, -56.6628556, -51.9799152, -47.3934848, -42.9154984, -38.5578898, -34.332593, -30.2515418, -26.3266702, -22.5699121, -18.9932013, -15.6084718, -12.4276575, -9.46269223, -6.72550995, -4.22804455, -1.98222994, 0, -3.99952687, -7.18694698, -9.65010173, -11.4768325, -12.7549807, -13.5723877, -14.016895, -14.1763439, -14.1385757, -13.991432, -13.8227541, -13.7203834, -13.7721613, -14.0659292, -14.6895284, -15.7308005, -17.4675508, -20.0147584, -23.0305842, -26.6435762, -30.9822822, -36.1752504, -42.3510286, -49.6381648, -58.1652071, -68.0607034, -79.4532017, -92.4712499, -107.243396, -123.965014, -142.673323, -163.476823, -186.484015, -211.199294, -238.033213, -267.557419, -300.061643, -335.835618, -375.169074, -418.351745, -464.861734, -516.047417, -572.357322, -634.239979, -702.143914, -776.760984, -858.235557, -946.955331, -1043.308, -1147.68126, -1258.98693, -1380.89244, -1516.65406, -1667.85282, -1832.64944, -2010.7403, -2201.82177, -2402.26868, -2615.85385, -2849.97362, -3112.84001, -3397.2898, -3698.97663, -4015.81808, -4349.56777, -4700.21853, -5069.47054, -5457.75927, -5859.25979, -6271.6884, -6698.56256, -7130.90492, -7564.42369, -7994.8271, -8417.82335, -8851.27575, -9263.19863, -9643.76144, -9983.1336, -10271.4846, -10440.4974, -10577.8189, -10698.6838, -10869.0007, -11083.9462, -11317.5833, -11566.1444, -11825.8621, -12092.9688, -12363.697, -13070.3247, -13497.4126, -13698.7968, -13728.3131, -13639.7974, -13487.0856, -13324.0135, -12953.0632, -12260.0165},
        {-500.59357, -498.318345, -495.874497, -493.265366, -490.494297, -487.56463, -484.479709, -481.242875, -477.857471, -474.32684, -470.654323, -466.843264, -462.897004, -458.818885, -454.612251, -450.280443, -445.826804, -441.254675, -436.567401, -431.768322, -426.860781, -421.848121, -416.733683, -411.520811, -406.212846, -400.813131, -395.325009, -389.75182, -384.096909, -378.363617, -372.555286, -366.67526, -360.726879, -354.713487, -348.613438, -342.443092, -336.219114, -329.945656, -323.626872, -317.266915, -310.869938, -304.440095, -297.981539, -291.498423, -284.994901, -278.475125, -271.943249, -265.403427, -258.859811, -252.316555, -245.777812, -239.247871, -232.730748, -226.230591, -219.751552, -213.297781, -206.873428, -200.482643, -194.129577, -187.81838, -181.553203, -175.338195, -169.177508, -163.075291, -157.035695, -151.06287, -145.160966, -139.388539, -133.802837, -128.295805, -122.868066, -117.520245, -112.252964, -107.066848, -101.962518, -96.9406002, -92.0017167, -87.1464913, -82.3755474, -77.6895087, -73.0889985, -68.5746405, -64.147058, -59.8068747, -55.2398599, -50.6106868, -46.0878126, -41.6820786, -37.4043258, -33.2653954, -29.2761285, -25.4473663, -21.7899498, -18.3147201, -15.0325185, -11.954186, -9.0905638, -6.45249296, -4.05081463, -1.89636993, 0, -3.84176715, -6.90754351, -9.28130531, -11.0470288, -12.2886901, -13.0902655, -13.5357313, -13.7090635, -13.6942386, -13.5752326, -13.4360217, -13.3605823, -13.4328905, -13.7369226, -14.3566547, -15.3760632, -17.2969116, -20.2628603, -23.4128898, -26.9205022, -30.9591996, -35.7024841, -41.3238579, -47.9968231, -55.8948818, -65.1915361, -76.0602883, -88.6746403, -103.208094, -120.177189, -139.28661, -160.606949, -184.208795, -209.642456, -237.238664, -267.466896, -300.553829, -336.726139, -376.210503, -419.233599, -464.441225, -514.122072, -568.812118, -629.04734, -695.363718, -768.518626, -848.771298, -936.602361, -1032.49245, -1136.92218, -1249.33763, -1372.51845, -1509.44895, -1661.50101, -1826.75453, -2004.76743, -2195.09765, -2393.38586, -2603.99754, -2835.57154, -3097.79475, -3381.95077, -3682.97236, -3998.19928, -4329.04981, -4675.07011, -5036.6827, -5413.66092, -5802.56492, -6201.77241, -6616.85758, -7037.89452, -7460.76986, -7881.37024, -8295.58228, -8721.32969, -9126.95275, -9502.82883, -9839.33528, -10126.8495, -10300.548, -10442.8094, -10567.6687, -10732.5645, -10934.3511, -11151.7974, -11382.662, -11624.7032, -11875.6794, -12133.3489, -12848.1447, -13268.7918, -13463.0702, -13498.7597, -13443.6401, -13365.4914, -13332.0932, -13226.2905, -12801.0824},
        {-512.726318, -509.865085, -506.873401, -503.75324, -500.506572, -497.13537, -493.641603, -490.027244, -486.294264, -482.444635, -478.480328, -474.403314, -470.215565, -465.919053, -461.515748, -457.007622, -452.396647, -447.684794, -442.874034, -437.966339, -432.963681, -427.86803, -422.681359, -417.405638, -412.042839, -406.594934, -401.063894, -395.451689, -389.760293, -383.991676, -378.14781, -372.230665, -366.242214, -360.184428, -353.975666, -347.661464, -341.288364, -334.86105, -328.384207, -321.86252, -315.300674, -308.703354, -302.075244, -295.421029, -288.745394, -282.053025, -275.348605, -268.63682, -261.922354, -255.209892, -248.504119, -241.845934, -235.203023, -228.579286, -221.978621, -215.40493, -208.862111, -202.354064, -195.88469, -189.457887, -183.077557, -176.747598, -170.47191, -164.254393, -158.098947, -152.009472, -145.989867, -140.079819, -134.318157, -128.634971, -123.031842, -117.510348, -112.07207, -106.718588, -101.451482, -96.2723303, -91.1827141, -86.1842129, -81.2784065, -76.4668746, -71.7511972, -67.132954, -62.6137249, -58.1950897, -53.6614769, -49.1276095, -44.7108204, -40.4197364, -36.2629843, -32.2491909, -28.3869829, -24.6849872, -21.1518306, -17.7961399, -14.6265418, -11.6516632, -8.8801309, -6.32057161, -3.98161216, -1.87187936, 0, -1.08831546, -1.86775062, -2.3914323, -2.71248732, -2.8840425, -2.95922467, -2.99116065, -3.03297726, -3.13780133, -3.35875967, -3.74897911, -4.36158647, -5.24970857, -6.46647223, -8.06500428, -10.0984315, -13.3865506, -18.0920125, -22.5338101, -26.9293566, -31.4960652, -36.4513492, -42.0126216, -48.3972958, -55.8227848, -64.506502, -74.6658604, -86.5182734, -100.281154, -116.57284, -135.062814, -155.848213, -179.026172, -204.459969, -232.36367, -263.013701, -296.577356, -333.221929, -373.114711, -416.422998, -461.573209, -511.003342, -565.221295, -624.734968, -690.052258, -760.231258, -837.592126, -923.005211, -1017.34087, -1121.46944, -1236.73605, -1362.95602, -1500.69153, -1650.49738, -1812.91311, -1988.48515, -2177.75986, -2380.2958, -2597.85174, -2833.26407, -3090.35298, -3366.03657, -3659.5949, -3970.12714, -4296.42585, -4637.42435, -4992.16991, -5359.62538, -5738.33571, -6127.08225, -6532.80294, -6944.86766, -7359.23431, -7771.86079, -8178.70501, -8596.8193, -8995.79353, -9366.312, -9699.059, -9984.71885, -10161.5912, -10307.6681, -10436.0843, -10598.8247, -10793.6459, -11002.1172, -11222.8047, -11454.2746, -11695.093, -11943.8259, -12537.5743, -12917.5502, -13131.8715, -13228.6563, -13256.0227, -13262.0888, -13294.9727, -13342.3676, -13179.314},
        {-524.778692, -521.433797, -517.9869, -514.438961, -510.790944, -507.043811, -503.198524, -499.256046, -495.21734, -491.083366, -486.855088, -482.533469, -478.11947, -473.614054, -469.018184, -464.332821, -459.558928, -454.697468, -449.749402, -444.715694, -439.597306, -434.395199, -429.110337, -423.743682, -418.296196, -412.768841, -407.16258, -401.478376, -395.71719, -389.879985, -383.967724, -377.981369, -371.921881, -365.790224, -359.459675, -352.997724, -346.473396, -339.891799, -333.258037, -326.577217, -319.854445, -313.094827, -306.303468, -299.485476, -292.645956, -285.790013, -278.922755, -272.049287, -265.174715, -258.304145, -251.442683, -244.669478, -237.913987, -231.179713, -224.470156, -217.788816, -211.139196, -204.524795, -197.949115, -191.415656, -184.92792, -178.489407, -172.103619, -165.774056, -159.504219, -153.29761, -147.157728, -141.09841, -135.133243, -129.244414, -123.434753, -117.707092, -112.064261, -106.509092, -101.044416, -95.6730625, -90.3978639, -85.2216508, -80.1472542, -75.177505, -70.3152342, -65.5632728, -60.9244518, -56.4016022, -51.9002247, -47.4738625, -43.1772289, -39.0163135, -34.9971057, -31.125595, -27.4077711, -23.8496235, -20.4571416, -17.2363151, -14.1931335, -11.3335862, -8.66366293, -6.18935313, -3.91664635, -1.85153213, 0, 1.87278947, 3.564712, 5.05434748, 6.32027579, 7.34107684, 8.0953305, 8.56161667, 8.71851524, 8.5446061, 8.01846914, 7.11868425, 5.82383133, 4.11249025, 1.96324092, -0.645336777, -3.73466295, -8.38798477, -14.7784119, -20.5256128, -25.878542, -31.0861539, -36.397403, -42.0612437, -48.3266307, -55.4425182, -63.6578609, -73.2216132, -84.3827296, -97.3901645, -112.897664, -130.600967, -150.627591, -173.105054, -198.103739, -225.779731, -256.304347, -289.822245, -326.478081, -366.416512, -409.782196, -455.161457, -504.731559, -558.942049, -618.242474, -683.082382, -750.796933, -825.72866, -909.105706, -1002.15622, -1106.10833, -1224.11417, -1353.12639, -1491.23171, -1638.16909, -1797.05087, -1969.47483, -2157.03871, -2363.6357, -2588.04648, -2826.54759, -3076.0697, -3340.34153, -3623.96379, -3928.22628, -4248.80863, -4583.96562, -4931.29118, -5288.86874, -5657.20495, -6035.43576, -6431.50955, -6834.39406, -7240.17479, -7644.93723, -8044.76685, -8457.49205, -8852.01969, -9218.99952, -9549.08131, -9832.91481, -10009.6907, -10155.8239, -10284.018, -10443.7261, -10633.1053, -10835.0002, -11048.3336, -11272.0281, -11505.0063, -11746.1909, -12184.5927, -12513.2763, -12751.8902, -12920.0828, -13037.5025, -13123.7977, -13198.6168, -13347.2147, -13399.2748},
        {-535.733965, -532.227073, -528.620091, -524.913877, -521.109293, -517.207197, -513.20845, -509.113912, -504.924442, -500.640902, -496.26415, -491.795048, -487.234454, -482.583229, -477.842232, -473.012325, -468.094366, -463.089216, -457.997735, -452.820783, -447.559219, -442.213904, -436.785698, -431.275461, -425.684052, -420.012332, -414.261161, -408.431399, -402.523906, -396.539541, -390.479165, -384.343637, -378.133818, -371.850568, -365.353742, -358.717667, -352.018367, -345.261278, -338.451837, -331.595478, -324.697638, -317.763752, -310.799257, -303.809589, -296.800183, -289.776475, -282.743901, -275.707897, -268.673899, -261.647343, -254.633665, -247.758107, -240.903701, -234.073286, -227.2697, -220.495784, -213.754374, -207.04831, -200.380431, -193.753575, -187.170582, -180.634288, -174.147535, -167.713159, -161.334001, -155.012898, -148.752689, -142.530911, -136.325707, -130.1921, -124.13457, -118.157595, -112.265655, -106.463228, -100.754794, -95.1448311, -89.6378186, -84.2382354, -78.9505606, -73.7792732, -68.7288522, -63.8037764, -59.008525, -54.3475769, -49.8655241, -45.5459455, -41.3719375, -37.3466775, -33.4733426, -29.75511, -26.1951572, -22.7966611, -19.5627992, -16.4967486, -13.6016866, -10.8807904, -8.33723729, -5.97420444, -3.79486911, -1.80240856, 0, 3.87749146, 7.26375453, 10.1566859, 12.5541822, 14.45414, 15.8544561, 16.7530271, 17.1477497, 17.0365204, 16.417236, 15.2877931, 13.6460884, 11.4900185, 8.81748002, 5.6263697, 1.91458417, -3.50902254, -10.816417, -17.3202014, -23.277274, -28.9445329, -34.5788763, -40.4372024, -46.7764094, -53.8533955, -61.9250589, -71.2482978, -82.0800104, -94.6770948, -109.725951, -126.896491, -146.316764, -168.114816, -192.180591, -218.76119, -248.167205, -280.598117, -316.253405, -355.332546, -398.035021, -443.598588, -493.477144, -548.05833, -607.729788, -672.87916, -739.912611, -814.194628, -897.108222, -990.036404, -1094.36218, -1214.20885, -1344.86991, -1482.9222, -1627.59083, -1783.50786, -1952.87774, -2137.90492, -2345.29709, -2571.732, -2808.97475, -3048.7154, -3300.24955, -3572.96356, -3870.03376, -4184.11395, -4512.68793, -4851.72193, -5198.3063, -5555.09593, -5921.59812, -6306.28927, -6698.32765, -7094.11585, -7490.05644, -7882.55199, -8294.08925, -8688.46558, -9055.56253, -9385.26163, -9667.44441, -9837.76926, -9976.48964, -10096.7256, -10250.1035, -10434.5013, -10631.7525, -10840.5375, -11059.5362, -11287.4289, -11522.8957, -11804.4034, -12078.3149, -12335.9477, -12568.6191, -12767.6466, -12924.3474, -13030.0391, -13237.9415, -13420.5099},
        {-544.3831, -541.293812, -538.060051, -534.684121, -531.168328, -527.514975, -523.726367, -519.804809, -515.752604, -511.572057, -507.265473, -502.835155, -498.283409, -493.612539, -488.824849, -483.922644, -478.908228, -473.783905, -468.55198, -463.214757, -457.774541, -452.233636, -446.594346, -440.858977, -435.029831, -429.109215, -423.099431, -417.002785, -410.821582, -404.558124, -398.214717, -391.793666, -385.297274, -378.727846, -371.982706, -365.118856, -358.194562, -351.215535, -344.187485, -337.116124, -330.007163, -322.866314, -315.699288, -308.511795, -301.309548, -294.098258, -286.883635, -279.671391, -272.467237, -265.276884, -258.106044, -251.143579, -244.208079, -237.301285, -230.424937, -223.580776, -216.770544, -209.995981, -203.258828, -196.560826, -189.903716, -183.289239, -176.719135, -170.195146, -163.719012, -157.292474, -150.917274, -144.518968, -138.024945, -131.594067, -125.233012, -118.948462, -112.747094, -106.63559, -100.620627, -94.7088862, -88.9070466, -83.2217877, -77.6597891, -72.2277303, -66.9322908, -61.7801501, -56.7779877, -51.9324831, -47.4413058, -43.2116231, -39.1483069, -35.2518385, -31.5226995, -27.9613715, -24.568336, -21.3440744, -18.2890683, -15.4037992, -12.6887487, -10.1443982, -7.77122934, -5.56972355, -3.54036239, -1.68362736, 0, 3.55560619, 6.72639506, 9.5015781, 11.8703668, 13.8219727, 15.3456072, 16.4304818, 17.065808, 17.2407974, 16.9446614, 16.1666114, 14.8958591, 13.1216158, 10.8330931, 8.0195024, 4.67005527, -0.236153465, -6.85421469, -12.9101293, -18.6311393, -24.2444869, -29.977414, -36.0571629, -42.7109754, -50.1660938, -58.6497601, -68.3892164, -79.6117047, -92.5444671, -107.980806, -125.374348, -144.782517, -166.262736, -188.781886, -212.942663, -239.638573, -269.254204, -302.174143, -338.782976, -379.46529, -424.767335, -474.862833, -530.104743, -590.846022, -657.439627, -727.1115, -804.123369, -889.609944, -984.705938, -1090.54606, -1210.52618, -1340.75621, -1478.44413, -1622.94839, -1778.01806, -1945.43094, -2126.9648, -2328.39787, -2546.59831, -2774.0701, -3001.91403, -3241.07136, -3503.58057, -3793.39639, -4100.77606, -4422.26416, -4751.81978, -5085.31717, -5428.11094, -5780.19307, -6149.99227, -6527.76584, -6910.58536, -7295.52241, -7679.64856, -8096.86991, -8498.21488, -8871.54644, -9204.72753, -9485.62109, -9638.99594, -9757.87193, -9857.15236, -9997.84396, -10176.6008, -10370.867, -10578.2572, -10796.3859, -11022.8675, -11255.3166, -11410.7193, -11633.095, -11894.5961, -12167.3744, -12423.5823, -12635.3718, -12774.895, -13010.6002, -13193.8509},
        {-553.228513, -550.651218, -547.875733, -544.906167, -541.746628, -538.401223, -534.874062, -531.169252, -527.290902, -523.243119, -519.030012, -514.655689, -510.124259, -505.439828, -500.606506, -495.628401, -490.509621, -485.254273, -479.866467, -474.350311, -468.709911, -462.949378, -457.072819, -451.084341, -444.988054, -438.788065, -432.488483, -426.093416, -419.606971, -413.033258, -406.376384, -399.640457, -392.829586, -385.947879, -378.95307, -371.873429, -364.737784, -357.551749, -350.320937, -343.050961, -335.747434, -328.415969, -321.062179, -313.691678, -306.310078, -298.922993, -291.536035, -284.154818, -276.784955, -269.432059, -262.101743, -255.005415, -247.938433, -240.901948, -233.897111, -226.925075, -219.986992, -213.084014, -206.217291, -199.387977, -192.597223, -185.846181, -179.136003, -172.467841, -165.842846, -159.262171, -152.726967, -146.124465, -139.34463, -132.623569, -125.969819, -119.391916, -112.898398, -106.497801, -100.198663, -94.0095202, -87.9389097, -81.9953685, -76.1874335, -70.5236416, -65.0125298, -59.6626351, -54.4824943, -49.4806444, -45.0102341, -40.900246, -36.9755208, -33.2334121, -29.6712736, -26.2864587, -23.0763213, -20.0382148, -17.169493, -14.4675094, -11.9296177, -9.5531715, -7.33552443, -5.27403012, -3.3660422, -1.60891428, 0, 3.53152951, 6.75334498, 9.64820481, 12.1988674, 14.3880911, 16.1986343, 17.6132554, 18.6147129, 19.185765, 19.3091702, 18.9676868, 18.1440733, 16.8210881, 14.9814895, 12.6080359, 9.68348572, 5.33995489, -0.561318562, -6.11350402, -11.5161236, -16.9686995, -22.6707538, -28.8218086, -35.6213862, -43.2690085, -51.9641979, -61.9064763, -73.295366, -86.330389, -101.800374, -119.099457, -138.25037, -159.275841, -180.54842, -202.915926, -227.666231, -255.317121, -286.386379, -321.391793, -360.851146, -406.406705, -457.109542, -513.257433, -575.148158, -643.079494, -715.158255, -794.420925, -881.713023, -977.88007, -1083.76758, -1201.77062, -1329.29129, -1464.47742, -1606.95654, -1759.37722, -1923.03165, -2099.21201, -2292.32833, -2499.84635, -2715.83047, -2932.0875, -3159.54562, -3410.85936, -3690.16278, -3987.15253, -4298.15225, -4616.6699, -4938.29914, -5268.95786, -5608.724, -5964.56986, -6328.6183, -6698.56073, -7072.08854, -7446.89314, -7865.44058, -8269.45395, -8645.43099, -8979.86944, -9259.26704, -9401.66221, -9507.65156, -9593.07254, -9724.03126, -9896.46714, -10085.3744, -10287.7397, -10500.5497, -10720.7912, -10945.4509, -11047.5684, -11236.9991, -11483.7817, -11757.9548, -12029.5568, -12268.6264, -12445.2021, -12681.7708, -12838.5435},
        {-563.878466, -561.692931, -559.257475, -556.57813, -553.66093, -550.511907, -547.137094, -543.542523, -539.734229, -535.718243, -531.500598, -527.087328, -522.484465, -517.698042, -512.734092, -507.598647, -502.297741, -496.837407, -491.223677, -485.462584, -479.560161, -473.52244, -467.355456, -461.06524, -454.657825, -448.139245, -441.515532, -434.792719, -427.976839, -421.073924, -414.090008, -407.031123, -399.903303, -392.712579, -385.514819, -378.29009, -371.017861, -363.702549, -356.348568, -348.960334, -341.542264, -334.098772, -326.634275, -319.153188, -311.659926, -304.158906, -296.654544, -289.151254, -281.653452, -274.165555, -266.691977, -259.333644, -251.99637, -244.682478, -237.394292, -230.134137, -222.904334, -215.707209, -208.545084, -201.420283, -194.335129, -187.291947, -180.293059, -173.34079, -166.437462, -159.5854, -152.786926, -145.944787, -138.964117, -132.052614, -125.219057, -118.472223, -111.820893, -105.273845, -98.8398579, -92.5277102, -86.3461811, -80.3040492, -74.4100936, -68.6730929, -63.1018261, -57.7050719, -52.4916092, -47.4702169, -43.1173813, -39.1981899, -35.4721087, -31.9327384, -28.57368, -25.3885342, -22.3709018, -19.5143838, -16.8125808, -14.2590938, -11.8475235, -9.57147084, -7.42453656, -5.40032151, -3.49242652, -1.6944524, 0, 6.15265964, 11.6329807, 16.4473796, 20.6022725, 24.1040758, 26.9592057, 29.1740786, 30.7551107, 31.7087185, 32.041318, 31.7593258, 30.869158, 29.377231, 27.2899611, 24.6137646, 21.3550577, 16.5565974, 10.0871342, 4.13279039, -1.50651611, -7.03086748, -12.6403459, -18.5350335, -24.9150125, -31.980365, -39.9311732, -48.9675192, -59.2894852, -71.0971534, -84.7480446, -100.227075, -117.687094, -137.280954, -158.260714, -181.229621, -207.031129, -236.088328, -268.824303, -305.662144, -347.024936, -394.542129, -447.063295, -504.775496, -567.865792, -636.521244, -709.003006, -787.905521, -873.897325, -967.646953, -1069.82294, -1182.01435, -1302.84411, -1430.99298, -1566.36313, -1711.3505, -1867.2314, -2035.28212, -2219.71761, -2418.20763, -2625.21613, -2833.72819, -3052.7116, -3291.26718, -3552.80294, -3831.08069, -4124.29002, -4430.10223, -4746.57251, -5073.65647, -5410.23471, -5757.71462, -6112.06133, -6471.28083, -6833.3791, -7196.36213, -7598.79561, -7987.98588, -8351.79899, -8678.10099, -8954.75792, -9109.94839, -9231.01754, -9331.41175, -9465.13593, -9629.58205, -9805.24272, -9990.35489, -10183.1555, -10381.8815, -10584.7699, -10815.672, -11022.332, -11215.058, -11404.1581, -11599.9403, -11812.7127, -12052.7832, -12267.7852, -12443.9962},
        {-575.353228, -573.358249, -571.085667, -568.54246, -565.735606, -562.672082, -559.358866, -555.802936, -552.011268, -547.990841, -543.748632, -539.291618, -534.626777, -529.761087, -524.701525, -519.455068, -514.028695, -508.429382, -502.664107, -496.739848, -490.663582, -484.442287, -478.082941, -471.59252, -464.978002, -458.246366, -451.404588, -444.459645, -437.418517, -430.288179, -423.07561, -415.787787, -408.431687, -401.014289, -393.64325, -386.27309, -378.857118, -371.399045, -363.90258, -356.371435, -348.809319, -341.219942, -333.607015, -325.974248, -318.325351, -310.664034, -302.994008, -295.318983, -287.64267, -279.968777, -272.301017, -264.661662, -257.035456, -249.425709, -241.835726, -234.268817, -226.728289, -219.217449, -211.739607, -204.298068, -196.896143, -189.537137, -182.224359, -174.961117, -167.750718, -160.596471, -153.501683, -146.37395, -139.123166, -131.950038, -124.864077, -117.874797, -110.99171, -104.224327, -97.582163, -91.0747285, -84.7115366, -78.5020995, -72.4559299, -66.58254, -60.8914424, -55.3921494, -50.0941734, -45.007027, -40.7731923, -37.0723832, -33.5767057, -30.2751313, -27.1566312, -24.2101768, -21.4247395, -18.7892905, -16.2928013, -13.9242432, -11.6725875, -9.52680564, -7.47586887, -5.50874858, -3.61441611, -1.7818428, 0, 8.10219764, 15.3219223, 21.6787685, 27.1923309, 31.8822038, 35.767982, 38.8692598, 41.2056319, 42.7966927, 43.6620368, 43.8212588, 43.2939531, 42.0997144, 40.2581371, 37.7888157, 34.7113449, 29.9310917, 23.3084754, 17.3833667, 11.9365971, 6.74899811, 1.6014013, -3.72536182, -9.45045971, -15.7930609, -22.9723337, -31.2074468, -40.7175685, -51.7218674, -64.398497, -79.0226794, -95.8258877, -115.039595, -136.666394, -161.052198, -188.603954, -219.621799, -254.405869, -293.256302, -336.473235, -385.549972, -439.230344, -497.58104, -560.668753, -628.560172, -698.892766, -774.769753, -856.865129, -945.852891, -1042.40703, -1148.08275, -1261.59582, -1381.48819, -1507.67477, -1642.87362, -1788.54425, -1946.14617, -2120.14662, -2308.31379, -2505.13473, -2704.65077, -2913.9053, -3138.61444, -3381.42914, -3639.80653, -3913.5878, -4203.90496, -4510.93383, -4830.11724, -5159.57525, -5497.44767, -5841.83182, -6190.84095, -6542.58831, -6895.18716, -7282.52372, -7657.99502, -8010.77107, -8330.02186, -8604.91741, -8773.64004, -8910.33923, -9026.03681, -9162.4416, -9318.42854, -9480.08636, -9646.92934, -9818.47175, -9994.22783, -10173.7119, -10525.8029, -10748.5169, -10888.8193, -10993.6754, -11110.0507, -11284.9103, -11565.2198, -11750.9308, -11937.0854},
        {-587.760262, -584.400687, -580.921294, -577.322384, -573.604258, -569.767217, -565.811563, -561.737598, -557.545622, -553.235936, -548.808843, -544.264643, -539.603638, -534.826129, -529.932418, -524.922805, -519.797592, -514.557081, -509.201572, -503.731368, -498.146768, -492.448076, -486.635591, -480.709616, -474.670451, -468.518398, -462.253759, -455.876834, -449.387925, -442.787334, -436.07536, -429.252307, -422.318475, -415.274166, -407.878133, -400.256531, -392.538421, -384.731942, -376.845234, -368.886436, -360.86369, -352.785135, -344.658911, -336.493157, -328.296015, -320.075624, -311.840124, -303.597655, -295.356357, -287.12437, -278.909834, -271.032209, -263.181566, -255.359297, -247.566793, -239.805444, -232.076642, -224.381779, -216.722245, -209.099432, -201.514731, -193.969533, -186.465229, -179.00321, -171.584868, -164.211594, -156.884779, -149.423719, -141.65208, -133.946813, -126.321114, -118.788179, -111.361203, -104.053381, -96.8779109, -89.8479868, -82.9768049, -76.2775608, -69.7634505, -63.4476695, -57.3434136, -51.4638786, -45.8222602, -40.4317541, -36.1040812, -32.4363788, -29.0161865, -25.8307862, -22.8674602, -20.1134903, -17.5561587, -15.1827475, -12.9805386, -10.9368142, -9.03885623, -7.27394684, -5.62936804, -4.09240191, -2.65033049, -1.29043583, 0, -2.55751112, -4.18707911, -4.97580059, -5.01077219, -4.37909051, -3.16785217, -1.4641538, 0.644907995, 3.07223659, 5.73073537, 8.53330772, 11.392857, 14.2222866, 16.9345, 19.4424004, 21.6588913, 21.9723425, 20.0778668, 19.3367174, 19.335112, 19.6592681, 19.8954034, 19.6297354, 18.4484818, 15.9378603, 11.6840883, 5.27338347, -3.70803652, -15.6739541, -33.3808686, -54.0408493, -77.3648634, -103.063878, -128.363254, -154.216763, -182.241002, -212.892622, -246.628271, -283.9046, -325.178257, -374.347772, -427.380385, -484.059335, -544.167859, -607.489196, -664.2598, -726.196388, -795.468897, -874.247259, -964.701411, -1075.37571, -1194.2747, -1313.55537, -1430.49768, -1552.84096, -1683.6285, -1825.9036, -1989.17771, -2168.55591, -2352.08708, -2526.43635, -2707.56647, -2911.55087, -3144.17283, -3393.77949, -3656.72143, -3925.50918, -4195.49773, -4476.12253, -4768.85431, -5087.9322, -5420.09422, -5762.39128, -6111.87431, -6465.59424, -6858.11378, -7239.59412, -7597.70826, -7920.12917, -8194.52983, -8350.88275, -8473.02837, -8573.36845, -8703.19399, -8859.51308, -9023.12999, -9191.86301, -9363.53044, -9535.95057, -9706.94171, -9956.45526, -10123.6249, -10241.7539, -10344.1454, -10464.1029, -10634.9294, -10889.9283, -11060.8824, -11155.0355},
        {-612.03065, -611.823335, -611.147064, -610.015078, -608.440622, -606.436935, -604.017261, -601.194843, -597.982921, -594.394739, -590.443539, -586.142563, -581.505053, -576.544252, -571.273402, -565.705745, -559.854524, -553.73298, -547.354356, -540.731894, -533.878837, -526.808426, -519.533905, -512.068515, -504.425499, -496.618098, -488.659555, -480.563113, -472.342014, -464.009499, -455.578812, -447.063194, -438.475888, -429.830135, -421.487489, -413.279709, -405.034371, -396.753414, -388.438777, -380.092399, -371.716219, -363.312175, -354.882207, -346.428254, -337.952255, -329.456148, -320.941873, -312.411368, -303.866573, -295.309426, -286.741867, -278.1112, -269.475183, -260.836938, -252.19959, -243.566262, -234.940077, -226.324159, -217.72163, -209.135614, -200.569235, -192.025615, -183.507879, -175.019149, -166.562549, -158.141202, -149.758231, -141.116453, -131.925895, -122.80904, -113.788479, -104.886803, -96.1266045, -87.530473, -79.121, -70.9207765, -62.9523934, -55.2384419, -47.8015129, -40.6641975, -33.8490867, -27.3787714, -21.2758428, -15.5628918, -12.1720098, -10.1308738, -8.44421027, -7.07264321, -5.9767967, -5.1172948, -4.45476158, -3.94982109, -3.56309741, -3.25521458, -2.98679667, -2.71846774, -2.41085186, -2.02457308, -1.52025547, -0.858523091, 0, 14.0945819, 26.7717266, 38.0776457, 48.0585508, 56.7606537, 64.2301661, 70.5132995, 75.6562657, 79.7052764, 82.7065432, 84.7062777, 85.7506918, 85.8859969, 85.1584049, 83.6141273, 81.2993759, 76.4860194, 68.9667921, 62.8013, 57.6555387, 53.1955033, 49.0871892, 44.9965919, 40.5897066, 35.5325286, 29.4910535, 22.1312764, 13.1191927, 2.1207978, -12.5014089, -29.2999877, -48.2178943, -69.1980846, -91.2065999, -114.674853, -140.294768, -168.302376, -198.933706, -232.424789, -269.011655, -308.945309, -352.44225, -399.735577, -451.05839, -506.64379, -573.862589, -644.025748, -715.581937, -786.97983, -856.668099, -915.291305, -978.640587, -1057.19563, -1155.36655, -1265.17139, -1384.192, -1510.01021, -1633.60804, -1760.66711, -1904.06877, -2077.1232, -2266.33657, -2461.75755, -2658.87412, -2862.39094, -3072.78205, -3290.01885, -3514.44506, -3748.2475, -3992.57044, -4254.07935, -4527.54785, -4812.20901, -5107.2959, -5412.04158, -5772.92272, -6130.11788, -6471.04923, -6783.13893, -7053.80914, -7221.05691, -7354.67961, -7463.2816, -7568.38881, -7673.49082, -7776.69323, -7881.75483, -7992.43441, -8112.49075, -8245.68263, -8486.62777, -8708.05048, -8908.45562, -9086.34806, -9240.23268, -9368.61433, -9469.99788, -9807.88063, -9859.41027},
        {-620.114449, -619.883032, -619.193969, -618.059909, -616.493505, -614.507406, -612.114264, -609.32673, -606.157454, -602.619088, -598.724282, -594.485688, -589.915956, -585.027737, -579.833682, -574.346442, -568.578668, -562.543011, -556.252122, -549.718651, -542.95525, -535.97457, -528.789261, -521.411974, -513.85536, -506.132071, -498.254757, -490.236069, -482.088658, -473.825175, -465.45827, -457.000596, -448.464801, -439.863539, -431.531194, -423.312782, -415.049453, -406.743417, -398.396884, -390.012063, -381.591165, -373.1364, -364.649978, -356.134108, -347.591002, -339.022868, -330.431917, -321.820358, -313.190403, -304.54426, -295.88414, -287.176402, -278.459884, -269.737572, -261.012456, -252.28752, -243.565754, -234.850143, -226.143675, -217.449338, -208.770118, -200.109003, -191.46898, -182.853036, -174.264158, -165.705333, -157.179549, -148.384217, -139.024085, -129.732368, -120.531861, -111.445362, -102.495666, -93.705571, -85.0978723, -76.6953666, -68.5208503, -60.59712, -52.9469719, -45.5932025, -38.5586083, -31.8659857, -25.5381311, -19.5978409, -16.1530782, -14.1401973, -12.4704908, -11.0990872, -9.98111544, -9.07170419, -8.32598222, -7.69907829, -7.14612117, -6.62223961, -6.08256236, -5.48221819, -4.77633585, -3.92004411, -2.86847171, -1.57674742, 0, 20.5891464, 39.0629862, 55.4878825, 69.9301986, 82.4562977, 93.1325428, 102.025297, 109.200924, 114.725786, 118.666247, 121.08867, 122.059418, 121.644854, 119.911342, 116.925244, 112.752924, 106.974675, 99.5874236, 91.7569734, 83.4455303, 74.6153, 65.2284877, 55.2472992, 44.6339398, 33.3506151, 21.3595306, 8.62289173, -4.89709597, -19.238227, -33.6945936, -49.3203833, -66.3765014, -85.123853, -106.574516, -130.613809, -156.926738, -185.548858, -216.515721, -249.86288, -285.62589, -322.729892, -362.658803, -405.665431, -452.002582, -501.923065, -560.159315, -621.364606, -684.67184, -749.213917, -814.123739, -873.873872, -937.953518, -1012.3189, -1099.62276, -1195.77317, -1299.70642, -1410.3588, -1522.37802, -1639.96364, -1771.99363, -1928.9534, -2099.73482, -2274.03265, -2448.16105, -2629.65045, -2820.88282, -3026.30487, -3248.83388, -3483.81649, -3730.88173, -3989.22577, -4258.97708, -4539.91455, -4831.81702, -5134.46336, -5502.00047, -5866.24718, -6213.39033, -6529.61677, -6801.11338, -6958.65303, -7080.77919, -7177.42753, -7280.36296, -7390.88045, -7501.17944, -7613.00966, -7728.12086, -7848.26276, -7975.18509, -8230.05336, -8419.08392, -8560.43623, -8672.26976, -8772.74397, -8880.01834, -9012.25233, -9182.1552, -9326.19517},
        {-650.766465, -650.401968, -649.593433, -648.351663, -646.687464, -644.611642, -642.135001, -639.268348, -636.022487, -632.408223, -628.436362, -624.117708, -619.463069, -614.483247, -609.18905, -603.591281, -597.700747, -591.528252, -585.084603, -578.380603, -571.427059, -564.234775, -556.814557, -549.17721, -541.33354, -533.294351, -525.070449, -516.672639, -508.111727, -499.398517, -490.543816, -481.558427, -472.453157, -463.238811, -454.129609, -445.030369, -435.843467, -426.573106, -417.223491, -407.798826, -398.303315, -388.741161, -379.11657, -369.433744, -359.696888, -349.910207, -340.077903, -330.204182, -320.293247, -310.349302, -300.376552, -290.343268, -280.290365, -270.222826, -260.145635, -250.063775, -239.982228, -229.905978, -219.840007, -209.789299, -199.758836, -189.753602, -179.77858, -169.838752, -159.939102, -150.084613, -140.280267, -130.105707, -119.150788, -108.297726, -97.5790768, -87.0273967, -76.6752418, -66.5551683, -56.6997323, -47.1414898, -37.9129971, -29.0468102, -20.5754853, -12.5315784, -4.94764573, 2.14375661, 8.71007251, 14.7187459, 17.302208, 17.9050397, 18.0058966, 17.664224, 16.9394674, 15.8910725, 14.5784845, 13.0611491, 11.3985118, 9.6500179, 7.87511302, 6.13324263, 4.48385221, 2.98638725, 1.70029324, 0.685015657, 0, 16.3593768, 31.0869257, 44.2328571, 55.8473814, 65.9807092, 74.6830507, 82.0046164, 87.9956168, 92.7062623, 96.1867632, 98.48733, 99.6581732, 99.7495031, 98.8115302, 96.8944649, 94.0485175, 89.8151667, 84.1719467, 78.3197713, 72.1998369, 65.7533399, 58.9214769, 51.6454441, 43.8664381, 35.5256552, 26.5642918, 16.9235445, 6.54460957, -4.6313165, -16.9509838, -30.0796689, -43.9897917, -58.6537717, -73.4315142, -88.6016959, -104.606331, -121.601593, -139.743657, -159.188696, -180.092884, -202.427957, -226.59066, -252.773254, -281.167999, -311.967155, -349.012194, -387.93386, -428.012111, -468.526904, -508.758195, -544.084096, -582.455331, -629.884234, -688.711008, -753.778591, -823.29604, -895.472413, -963.666501, -1032.03996, -1110.04565, -1206.31788, -1313.35536, -1426.50906, -1543.61917, -1666.74382, -1796.00503, -1929.93782, -2068.25275, -2216.47951, -2376.85615, -2558.13982, -2755.04657, -2967.55785, -3195.65513, -3439.31986, -3753.47327, -4069.4221, -4373.41288, -4651.6921, -4890.5063, -5021.34864, -5121.72121, -5201.33621, -5294.40102, -5400.93187, -5510.73868, -5624.34801, -5742.28642, -5865.08044, -5993.25662, -6371.14744, -6578.84146, -6671.59434, -6704.66175, -6733.29936, -6812.76285, -6998.30787, -7207.89873, -7231.63886}
    };
    
    const double a5tab[98][200] = {{1623.72041, 1603.26038, 1582.9671, 1562.83685, 1542.86587, 1523.05045, 1503.38684, 1483.87131, 1464.50013, 1445.26955, 1426.17585, 1407.21528, 1388.38412, 1369.67863, 1351.09507, 1332.6297, 1314.2788, 1296.03863, 1277.90545, 1259.87553, 1241.94513, 1224.11051, 1206.36795, 1188.7137, 1171.14404, 1153.65522, 1136.24351, 1118.90518, 1101.63649, 1084.43371, 1067.29309, 1050.21091, 1033.18343, 1016.20692, 999.118104, 981.99637, 964.923024, 947.899508, 930.927265, 914.007736, 897.142363, 880.332589, 863.579855, 846.885603, 830.251275, 813.678314, 797.168161, 780.722259, 764.342049, 748.028973, 731.784474, 715.655146, 699.596299, 683.608397, 667.691903, 651.847281, 636.074993, 620.375503, 604.749273, 589.196768, 573.71845, 558.314783, 542.986229, 527.733252, 512.556316, 497.455882, 482.432416, 467.555474, 452.892958, 438.302825, 423.78106, 409.323648, 394.926571, 380.585814, 366.297362, 352.057197, 337.861305, 323.705669, 309.586273, 295.499101, 281.440137, 267.405366, 253.390771, 239.392336, 225.216873, 210.958931, 196.715333, 182.488202, 168.279662, 154.091834, 139.926843, 125.786812, 111.673862, 97.5901183, 83.537703, 69.5187392, 55.5353499, 41.5896584, 27.6837875, 13.8198603, 0, -12.1186564, -24.2674317, -36.4713762, -48.7555398, -61.1449727, -73.6647251, -86.3398469, -99.1953884, -112.2564, -125.547931, -139.095031, -152.922752, -167.056143, -181.520255, -196.340136, -211.540838, -227.50695, -244.314885, -261.176449, -278.193736, -295.468841, -313.103857, -331.20088, -349.862004, -369.189322, -389.284928, -410.250918, -432.189385, -455.202424, -479.575975, -505.160876, -532.004066, -560.152486, -588.79689, -618.412311, -649.702096, -682.970043, -718.519944, -756.655596, -797.680794, -842.990751, -891.465673, -943.195819, -998.271446, -1056.78281, -1116.76687, -1180.8805, -1249.7273, -1323.91084, -1404.0347, -1490.72445, -1584.53482, -1683.98873, -1789.44553, -1905.01407, -2033.11973, -2176.18793, -2341.27756, -2525.12745, -2719.42172, -2912.47091, -3121.06027, -3361.10615, -3639.589, -3948.34786, -4286.17191, -4645.67738, -5024.05306, -5441.12197, -5903.90393, -6420.91603, -6997.45046, -7640.00877, -8355.09248, -9149.20315, -10327.923, -11523.9026, -12668.8735, -13694.567, -14532.7144, -14971.0169, -15181.2562, -15156.8911, -14725.3653, -13945.8518, -12946.6962, -11783.6127, -10512.3157, -9188.51922, -7867.93762, -6749.89481, -5735.29053, -4819.2389, -3996.85405, -3263.2501, -2613.54118, -2042.8414, -1562.95934, -1111.20735},
        {1520.58131, 1500.0443, 1479.81369, 1459.88041, 1440.23538, 1420.86953, 1401.77377, 1382.93904, 1364.35626, 1346.01635, 1327.91024, 1310.02885, 1292.36311, 1274.90394, 1257.64227, 1240.56901, 1223.67511, 1206.95147, 1190.38903, 1173.97871, 1157.71143, 1141.57813, 1125.56971, 1109.67711, 1093.89126, 1078.20307, 1062.60348, 1047.0834, 1031.63376, 1016.24549, 1000.90951, 985.616738, 970.358109, 955.124545, 939.545162, 923.799399, 908.071043, 892.36276, 876.677215, 861.017074, 845.385004, 829.78367, 814.215737, 798.683873, 783.190743, 767.739013, 752.331348, 736.970415, 721.65888, 706.399408, 691.194666, 676.151615, 661.166364, 646.239319, 631.370884, 616.561464, 601.811465, 587.121291, 572.491347, 557.922038, 543.41377, 528.966946, 514.581973, 500.259256, 485.999198, 471.802205, 457.668683, 443.658296, 429.829288, 416.059842, 402.346523, 388.685893, 375.074517, 361.508956, 347.985776, 334.501539, 321.052808, 307.636147, 294.24812, 280.885289, 267.544218, 254.221471, 240.91361, 227.6172, 214.129976, 200.552098, 186.986114, 173.435039, 159.90189, 146.389682, 132.901432, 119.440153, 106.008863, 92.6105767, 79.2483101, 65.9250787, 52.6438983, 39.4077845, 26.2197531, 13.0828197, 0, -11.525326, -23.0344724, -34.5546084, -46.1129033, -57.736526, -69.4526459, -81.2884319, -93.2710534, -105.427679, -117.785479, -130.371622, -143.213276, -156.337612, -169.771798, -183.543004, -197.678398, -212.374085, -227.681368, -243.245301, -259.129253, -275.396593, -292.110691, -309.334916, -327.132639, -345.567229, -364.702055, -384.600488, -405.325896, -426.94165, -449.438621, -472.97926, -497.648684, -523.532013, -550.224084, -578.055158, -607.486233, -638.749515, -672.077205, -707.701507, -745.854624, -786.948105, -830.980224, -878.148093, -928.648825, -982.679534, -1041.32383, -1103.6707, -1169.69564, -1239.37415, -1312.6817, -1388.82261, -1469.48611, -1551.57472, -1635.17949, -1726.90139, -1830.41858, -1949.40921, -2095.98975, -2263.48225, -2436.00328, -2596.74012, -2764.34856, -2958.01933, -3184.04821, -3426.88122, -3680.99374, -3932.93077, -4175.11167, -4429.03408, -4699.74733, -5002.86271, -5331.24267, -5686.28049, -6069.36942, -6481.90274, -6933.16819, -7414.69094, -7925.89065, -8466.18698, -9034.99957, -9765.84752, -10434.6514, -10983.3597, -11063.6267, -10732.3089, -10167.2189, -9419.1655, -8538.95753, -7577.40381, -6585.31314, -5724.15795, -4930.84252, -4204.0848, -3542.60272, -2945.11422, -2410.33722, -1936.98966, -1548.9227, -1202.39183},
        {1424.81636, 1406.6982, 1388.76052, 1370.9988, 1353.40854, 1335.98521, 1318.72428, 1301.62125, 1284.6716, 1267.87079, 1251.21433, 1234.69768, 1218.31633, 1202.06575, 1185.94144, 1169.93888, 1154.05353, 1138.28089, 1122.61644, 1107.05565, 1091.59401, 1076.227, 1060.9501, 1045.75879, 1030.64856, 1015.61488, 1000.65324, 985.75911, 970.927981, 956.155332, 941.436643, 926.767396, 912.143072, 897.559152, 882.848309, 868.090852, 853.369049, 838.683664, 824.035462, 809.425209, 794.853667, 780.321604, 765.829782, 751.378967, 736.969923, 722.603415, 708.280209, 694.001067, 679.766756, 665.57804, 651.435684, 637.317148, 623.247006, 609.226529, 595.256985, 581.339646, 567.47578, 553.666658, 539.91355, 526.217725, 512.580453, 499.003005, 485.48665, 472.032658, 458.642299, 445.316843, 432.05756, 418.964619, 406.135816, 393.368446, 380.65737, 367.997444, 355.383529, 342.810482, 330.273163, 317.766429, 305.28514, 292.824154, 280.37833, 267.942526, 255.511601, 243.080414, 230.643824, 218.196688, 205.412935, 192.45464, 179.487734, 166.51749, 153.549182, 140.588083, 127.639466, 114.708606, 101.800775, 88.9212468, 76.075295, 63.268193, 50.5052142, 37.7916321, 25.1327201, 12.5337516, 0, -11.56261, -22.9568568, -34.2199259, -45.3890025, -56.5012718, -67.5939192, -78.70413, -89.8690894, -101.125983, -112.511995, -124.064313, -135.82012, -147.816602, -160.090944, -172.680333, -185.621952, -198.908768, -212.571648, -226.7478, -241.464932, -256.750754, -272.632976, -289.139308, -306.297459, -324.135138, -342.680056, -361.959922, -382.002445, -402.835335, -424.213426, -446.537357, -469.916702, -494.461034, -520.346942, -547.65049, -576.42987, -606.77455, -638.773995, -672.517673, -708.095051, -743.75089, -781.980795, -823.235154, -867.964354, -916.618785, -975.359217, -1037.49806, -1102.0581, -1168.06213, -1234.53295, -1296.65046, -1361.97719, -1433.01117, -1511.4386, -1597.28815, -1691.33271, -1794.34515, -1908.88958, -2033.54054, -2164.9185, -2298.16295, -2439.97692, -2601.31508, -2784.48016, -2977.1146, -3173.59016, -3366.38684, -3549.38594, -3723.40516, -3885.33858, -4026.91504, -4150.98851, -4256.24103, -4341.35465, -4405.01141, -4214.88475, -4058.41748, -3992.04378, -4072.19787, -4355.31393, -5255.02389, -6232.43241, -7190.88893, -7664.48683, -7702.73943, -7509.01676, -7125.1393, -6592.92755, -5954.202, -5250.78314, -4593.25484, -3950.42195, -3333.97552, -2755.6066, -2227.00625, -1759.86552, -1365.87547, -1015.11083, -720.103069},
        {1327.80903, 1311.72413, 1295.74686, 1279.87552, 1264.1084, 1248.44379, 1232.87998, 1217.41525, 1202.0479, 1186.77622, 1171.5985, 1156.51304, 1141.5181, 1126.612, 1111.79302, 1097.05945, 1082.40958, 1067.84171, 1053.35411, 1038.94508, 1024.61291, 1010.3559, 996.172328, 982.060488, 968.018669, 954.045163, 940.138259, 926.296248, 912.517421, 898.800066, 885.142475, 871.542937, 857.999743, 844.511183, 831.108371, 817.772494, 804.484347, 791.241154, 778.040141, 764.878532, 751.753553, 738.662427, 725.602382, 712.570641, 699.564429, 686.580972, 673.617495, 660.671222, 647.739379, 634.819191, 621.907882, 608.765078, 595.630754, 582.507286, 569.39705, 556.302422, 543.225778, 530.169492, 517.135941, 504.127501, 491.146547, 478.195456, 465.276602, 452.392361, 439.54511, 426.737224, 413.971078, 401.321422, 388.861266, 376.443723, 364.066476, 351.72721, 339.423608, 327.153355, 314.914135, 302.703632, 290.51953, 278.359512, 266.221264, 254.10247, 242.000812, 229.913976, 217.839646, 205.775505, 193.537248, 181.217381, 168.910599, 156.620493, 144.350651, 132.104665, 119.886124, 107.698618, 95.5457371, 83.4310715, 71.3582112, 59.3307461, 47.3522663, 35.4263617, 23.5566225, 11.7466386, 0, -10.6266262, -21.1449064, -31.5864849, -41.9830058, -52.3661131, -62.7674512, -73.218664, -83.7513958, -94.3972906, -105.187993, -116.155146, -127.330395, -138.745384, -150.431756, -162.421157, -174.745229, -187.608561, -201.067503, -214.762518, -228.762309, -243.135579, -257.951033, -273.277373, -289.183303, -305.737526, -323.008746, -341.065666, -359.97699, -379.81142, -400.817297, -422.817821, -445.827805, -469.862061, -494.35408, -519.609336, -546.08832, -573.980242, -603.47431, -634.759733, -668.02572, -703.688666, -741.641451, -782.028832, -824.995571, -870.686426, -920.261125, -972.595717, -1027.58122, -1085.10865, -1145.06902, -1206.55577, -1271.23233, -1340.27593, -1414.19316, -1492.12132, -1573.8125, -1659.0188, -1745.12558, -1834.78954, -1933.2492, -2047.51931, -2169.90355, -2298.2179, -2429.37124, -2558.73547, -2682.38798, -2798.88314, -2904.94051, -2988.19749, -3041.42884, -3068.86663, -3050.0657, -2973.83487, -2828.98295, -2604.31873, -1781.5205, -983.310226, -325.279336, 76.980733, 107.878545, -936.301575, -2190.944, -3519.58691, -4318.80246, -4636.96955, -4717.0363, -4597.65317, -4317.47063, -3915.13913, -3429.30913, -2964.09608, -2484.03346, -2003.69125, -1537.63941, -1100.44789, -706.686661, -370.925674, -53.1529409, 214.943335},
        {1238.0202, 1225.09738, 1212.13096, 1199.12394, 1186.07935, 1173.00021, 1159.88953, 1146.75033, 1133.58563, 1120.39844, 1107.19178, 1093.96868, 1080.73215, 1067.4852, 1054.23085, 1040.97213, 1027.71205, 1014.45362, 1001.19986, 987.953803, 974.718451, 961.496826, 948.291946, 935.106828, 921.944491, 908.807951, 895.700226, 882.624335, 869.583294, 856.580121, 843.617834, 830.699451, 817.827989, 805.006465, 792.471419, 780.104195, 767.783333, 755.504271, 743.26245, 731.053309, 718.872286, 706.714823, 694.576357, 682.452329, 670.338177, 658.229342, 646.121263, 634.009379, 621.88913, 609.755955, 597.605293, 585.049753, 572.475904, 559.887486, 547.288236, 534.681892, 522.072194, 509.462881, 496.857689, 484.260358, 471.674627, 459.104234, 446.552916, 434.024414, 421.522465, 409.050808, 396.61318, 384.28221, 372.128868, 360.014818, 347.939332, 335.901682, 323.901143, 311.936986, 300.008484, 288.11491, 276.255537, 264.429637, 252.636484, 240.87535, 229.145507, 217.446229, 205.776789, 194.136459, 182.433315, 170.714146, 159.026839, 147.373625, 135.756736, 124.178405, 112.640864, 101.146345, 89.6970803, 78.2953027, 66.9432442, 55.643137, 44.3972133, 33.2077056, 22.0768459, 11.0068666, 0, -9.32002674, -18.6616475, -28.0467313, -37.497147, -47.0347635, -56.6814499, -66.459075, -76.3895078, -86.4946172, -96.7962722, -107.316342, -118.076695, -129.0992, -140.405727, -152.018144, -163.95832, -176.583243, -189.962655, -203.360419, -216.870217, -230.585728, -244.600631, -259.008607, -273.903337, -289.378498, -305.527773, -322.44484, -340.22338, -358.957072, -379.139881, -400.318432, -422.466319, -445.557137, -468.68514, -492.263593, -516.94025, -542.95251, -570.537767, -599.933419, -631.376862, -666.972164, -704.521933, -743.898346, -784.973582, -827.619817, -868.588343, -911.652446, -957.464527, -1006.67699, -1059.94222, -1119.80369, -1182.71145, -1246.45396, -1310.24478, -1376.20702, -1445.15748, -1517.91293, -1597.28015, -1681.63366, -1767.17712, -1852.68589, -1936.52847, -2020.74576, -2103.51363, -2176.45883, -2234.21427, -2273.98265, -2291.06313, -2271.33221, -2205.99637, -2136.31815, -1995.74704, -1758.16244, -1397.44373, -887.470304, 640.901641, 2160.01374, 3485.23081, 4431.91767, 4815.43914, 3586.69221, 2001.8214, 246.677769, -887.893237, -1455.51754, -1761.90209, -1848.06875, -1755.03941, -1523.83595, -1195.48023, -861.747293, -522.029963, -182.849448, 149.273045, 467.816312, 766.259146, 1038.08034, 1281.16222, 1505.72281},
        {1161.2229, 1152.81576, 1144.09307, 1135.06593, 1125.74538, 1116.14252, 1106.26841, 1096.13412, 1085.75073, 1075.12931, 1064.28093, 1053.21666, 1041.94759, 1030.48477, 1018.83928, 1007.0222, 995.044591, 982.917535, 970.652101, 958.259362, 945.75039, 933.136257, 920.428037, 907.636802, 894.773624, 881.849577, 868.875731, 855.863161, 842.822938, 829.766135, 816.703825, 803.64708, 790.606972, 777.594575, 765.058622, 752.782149, 740.542933, 728.337842, 716.163747, 704.017517, 691.896021, 679.79613, 667.714712, 655.648638, 643.594778, 631.549999, 619.511173, 607.475169, 595.438856, 583.399104, 571.352783, 558.910623, 546.464003, 534.018164, 521.578347, 509.14979, 496.737736, 484.347424, 471.984095, 459.652988, 447.359345, 435.108405, 422.90541, 410.755598, 398.664211, 386.636489, 374.677673, 362.947915, 351.603651, 340.330622, 329.124029, 317.979067, 306.890936, 295.854833, 284.865956, 273.919504, 263.010673, 252.134662, 241.28667, 230.461893, 219.65553, 208.862778, 198.078837, 187.298903, 176.215123, 164.980594, 153.752058, 142.534547, 131.333094, 120.152731, 108.99849, 97.875405, 86.7885077, 75.7428309, 64.743407, 53.7952687, 42.9034485, 32.072979, 21.3088927, 10.6162222, 0, -8.7629998, -17.5228605, -26.3037598, -35.1298753, -44.0253846, -53.0144653, -62.1212951, -71.3700516, -80.7849123, -90.390055, -100.209657, -110.267897, -120.588951, -131.196998, -142.116214, -153.370779, -165.031192, -177.128248, -189.581348, -202.424594, -215.69209, -229.417942, -243.636252, -258.381125, -273.686665, -289.586976, -306.116161, -323.308326, -341.197573, -359.62576, -378.889728, -399.081257, -420.292125, -442.884856, -466.815857, -491.969332, -518.355838, -545.985929, -574.870159, -605.019083, -635.763653, -668.000862, -701.87423, -737.527278, -775.103526, -816.500119, -859.668547, -904.313924, -950.141365, -996.855984, -1042.85975, -1090.75366, -1140.30052, -1191.98409, -1247.76007, -1308.92331, -1376.76862, -1461.62503, -1553.69997, -1633.34536, -1679.09246, -1710.04086, -1746.99066, -1794.8053, -1831.34436, -1846.86262, -1833.87997, -1783.23846, -1677.47474, -1503.82345, -1332.98202, -1055.8819, -630.866789, -16.2803793, 829.533635, 3170.65522, 5520.71111, 7615.75168, 9191.82735, 9984.98848, 8548.10943, 6589.20067, 4351.38806, 2841.58375, 2000.58151, 1450.93082, 1149.33765, 1052.50799, 1117.14781, 1299.96312, 1560.00147, 1806.7442, 2040.60955, 2262.01576, 2471.38109, 2669.12378, 2855.66208, 2989.84669, 3130.55645},
        {1085.38834, 1077.95688, 1070.26061, 1062.30756, 1054.10578, 1045.6633, 1036.98817, 1028.08842, 1018.97211, 1009.64726, 1000.12192, 990.404135, 980.501939, 970.423374, 960.176482, 949.769304, 939.20988, 928.506251, 917.666457, 906.69854, 895.61054, 884.410498, 873.106455, 861.706452, 850.218529, 838.650726, 827.011086, 815.307648, 803.548454, 791.741543, 779.894958, 768.016738, 756.114924, 744.197558, 732.535004, 720.998625, 709.456626, 697.908537, 686.353884, 674.792196, 663.223, 651.645824, 640.060196, 628.465644, 616.861696, 605.247879, 593.623721, 581.988751, 570.342495, 558.684481, 547.014238, 535.032074, 523.043223, 511.053698, 499.069515, 487.096686, 475.141228, 463.209153, 451.306476, 439.439212, 427.613375, 415.834979, 404.110038, 392.444567, 380.84458, 369.316091, 357.865114, 346.689567, 335.982759, 325.35492, 314.799622, 304.310442, 293.880951, 283.504725, 273.175337, 262.886362, 252.631373, 242.403944, 232.19765, 222.006065, 211.822762, 201.641315, 191.455299, 181.258287, 170.636126, 159.794828, 148.94531, 138.094376, 127.248833, 116.415486, 105.60114, 94.8126013, 84.0566755, 73.3401681, 62.6698847, 52.0526307, 41.4952119, 31.0044339, 20.5871021, 10.2500223, 0, -8.55111025, -17.0361809, -25.4832345, -33.9202934, -42.3753804, -50.8765178, -59.4517282, -68.1290341, -76.936458, -85.9020224, -95.05375, -104.419663, -114.027784, -123.906136, -134.082741, -144.585622, -155.279042, -166.16763, -177.649816, -189.718529, -202.366702, -215.587265, -229.373152, -243.717292, -258.612617, -274.052059, -290.028549, -306.535018, -323.564399, -340.537862, -358.229745, -376.804505, -396.426603, -418.154871, -441.706581, -466.560506, -492.612794, -519.75959, -547.897042, -576.921298, -605.033622, -634.340877, -665.070817, -697.451197, -731.70977, -771.342042, -812.491079, -854.567699, -896.982717, -939.146952, -978.815692, -1019.07871, -1059.67646, -1101.03744, -1144.99489, -1192.75137, -1245.50939, -1313.01468, -1385.98497, -1445.81818, -1474.74759, -1485.33329, -1492.22249, -1497.25888, -1480.59125, -1432.32668, -1347.64489, -1217.96812, -1016.11895, -725.440983, -459.576371, -51.5242279, 560.475194, 1438.18164, 2643.35486, 5892.35818, 9178.69687, 12150.4798, 14455.8158, 15742.8137, 14132.0541, 11817.5263, 9101.99427, 7213.43077, 6088.3711, 5277.84678, 4737.66828, 4423.64605, 4291.59056, 4297.31226, 4408.80546, 4519.95418, 4630.21957, 4739.06277, 4845.94492, 4950.32716, 5051.67064, 5094.44178, 5113.97782},
        {1007.70859, 997.464839, 987.344033, 977.340011, 967.446613, 957.657681, 947.967057, 938.368581, 928.856094, 919.423438, 910.064453, 900.772981, 891.542863, 882.36794, 873.242053, 864.159043, 855.112752, 846.09702, 837.105689, 828.1326, 819.171593, 810.216511, 801.261194, 792.299483, 783.325219, 774.332245, 765.314399, 756.265525, 747.179462, 738.050053, 728.871138, 719.636558, 710.340154, 700.975768, 691.2867, 681.397332, 671.435055, 661.401842, 651.299664, 641.130493, 630.8963, 620.599058, 610.240737, 599.823309, 589.348747, 578.819021, 568.236104, 557.601967, 546.918581, 536.187919, 525.411951, 514.417361, 503.385209, 492.321267, 481.231306, 470.121098, 458.996415, 447.863028, 436.726708, 425.593228, 414.468359, 403.357872, 392.267539, 381.203132, 370.170422, 359.175181, 348.22318, 337.462775, 327.038898, 316.663251, 306.332364, 296.042765, 285.790983, 275.573545, 265.38698, 255.227817, 245.092584, 234.977809, 224.880022, 214.79575, 204.721522, 194.653866, 184.589311, 174.524386, 164.161037, 153.64928, 143.142671, 132.647298, 122.169248, 111.71461, 101.289472, 90.8999232, 80.5520503, 70.2519421, 60.0056867, 49.8193722, 39.6990869, 29.6509188, 19.6809562, 9.79528722, 0, -8.37939856, -16.6550084, -24.855446, -33.0093277, -41.1452701, -49.2918897, -57.4778028, -65.7316259, -74.0819756, -82.5574682, -91.1867202, -99.9983481, -109.020968, -118.283197, -127.813652, -137.640948, -147.670918, -157.914637, -168.67845, -179.96466, -191.775574, -204.113498, -216.980736, -230.379594, -244.312378, -258.781394, -273.788946, -289.337341, -305.428884, -321.651525, -338.573855, -356.322487, -375.024033, -395.216843, -416.821662, -439.649434, -463.703248, -488.986197, -515.501369, -543.251856, -572.578397, -603.043669, -634.584702, -667.138523, -700.642161, -733.737164, -767.97991, -803.631298, -840.952224, -880.203589, -923.979343, -967.35582, -1007.16856, -1041.91532, -1073.48753, -1102.25292, -1128.57923, -1151.89921, -1173.72809, -1196.60112, -1234.72882, -1262.02155, -1257.49556, -1208.58552, -1116.99064, -977.862483, -797.524279, -574.023943, -264.446559, 150.951897, 506.624755, 1036.11299, 1823.29971, 2952.06802, 4506.30103, 8644.11745, 12856.6059, 16709.0905, 19766.8956, 21595.3454, 19922.4262, 17375.6919, 14307.8973, 12125.4592, 10764.0578, 9720.34716, 8951.95872, 8416.52386, 8071.67396, 7875.04041, 7784.20116, 7717.38392, 7671.70392, 7644.27635, 7632.21643, 7632.63935, 7642.66032, 7602.94526, 7466.69955},
        {938.236429, 928.1099, 918.202939, 908.505482, 899.007469, 889.698837, 880.569525, 871.609471, 862.808614, 854.156892, 845.644243, 837.260606, 828.995919, 820.840119, 812.783147, 804.814939, 796.925434, 789.10457, 781.342287, 773.628521, 765.953212, 758.306298, 750.677716, 743.057406, 735.435306, 727.801353, 720.145487, 712.457646, 704.727767, 696.94579, 689.101652, 681.185293, 673.186649, 665.09566, 656.494929, 647.586631, 638.577833, 629.471693, 620.271367, 610.980014, 601.600791, 592.136853, 582.591359, 572.967465, 563.268329, 553.497107, 543.656957, 533.751036, 523.782501, 513.754509, 503.670218, 493.444201, 483.17012, 472.85305, 462.49807, 452.110256, 441.694687, 431.256439, 420.80059, 410.332217, 399.856398, 389.378209, 378.902729, 368.435035, 357.980204, 347.543314, 337.129441, 326.853639, 316.828321, 306.831746, 296.861864, 286.916623, 276.99397, 267.091854, 257.208222, 247.341024, 237.488207, 227.647719, 217.817508, 207.995523, 198.179711, 188.368021, 178.5584, 168.748798, 158.676699, 148.475761, 138.282774, 128.104136, 117.94625, 107.815515, 97.7183316, 87.6611013, 77.6502243, 67.6921012, 57.7931326, 47.9597191, 38.1982614, 28.51516, 18.9168156, 9.40962872, 0, -7.93297828, -15.7653053, -23.5244386, -31.237836, -38.9329549, -46.637253, -54.378188, -62.1832175, -70.0797991, -78.0953904, -86.2574491, -94.5934327, -103.130799, -111.897005, -120.91951, -130.22577, -139.715555, -149.398081, -159.589625, -170.290282, -181.500148, -193.219319, -205.447892, -218.185961, -231.433623, -245.190974, -259.458109, -274.235125, -289.522117, -304.978516, -321.069994, -337.898846, -355.567367, -374.316737, -394.17981, -415.152402, -437.295144, -460.668667, -485.3336, -511.350574, -540.01287, -569.773313, -600.45136, -631.866473, -663.838109, -694.763552, -726.239981, -758.4424, -791.545812, -825.72522, -863.237624, -899.631369, -933.963688, -965.051614, -991.221761, -1011.02093, -1022.99593, -1018.33018, -1004.60735, -997.443886, -1028.00528, -1050.18696, -1025.32675, -933.464519, -786.329699, -581.1057, -330.407623, -35.4196825, 359.256785, 877.013967, 1332.26333, 1989.52867, 2944.54163, 4293.03389, 6130.73712, 10848.3559, 15672.9058, 20126.3751, 23730.7524, 26008.026, 24502.4878, 22032.2872, 18966.9971, 16759.7236, 15351.1408, 14230.4504, 13360.9001, 12705.7375, 12228.2102, 11891.566, 11800.7793, 11647.7477, 11454.4786, 11242.9792, 11035.2566, 10853.3184, 10719.1715, 10505.4857, 10185.0648},
        {874.787044, 867.796736, 860.814736, 853.838339, 846.86484, 839.891535, 832.915717, 825.934682, 818.945726, 811.946144, 804.933229, 797.904279, 790.856586, 783.787448, 776.694158, 769.574012, 762.424305, 755.242331, 748.025387, 740.770767, 733.475765, 726.137678, 718.753801, 711.321427, 703.837853, 696.300374, 688.706284, 681.052878, 673.337453, 665.557302, 657.70972, 649.792004, 641.801448, 633.735346, 625.456459, 617.032178, 608.531515, 599.956129, 591.307682, 582.587834, 573.798248, 564.940584, 556.016502, 547.027665, 537.975733, 528.862366, 519.689227, 510.457976, 501.170273, 491.827781, 482.432161, 472.891739, 463.303534, 453.67123, 443.998511, 434.289062, 424.546566, 414.774708, 404.977172, 395.157642, 385.319804, 375.467341, 365.603937, 355.733276, 345.859044, 335.984924, 326.1146, 316.31536, 306.652964, 296.999919, 287.355787, 277.72013, 268.092508, 258.472484, 248.859617, 239.253469, 229.653602, 220.059576, 210.470953, 200.887294, 191.30816, 181.733112, 172.161712, 162.59352, 152.813628, 142.933343, 133.066551, 123.219771, 113.399527, 103.612338, 93.8647258, 84.1632121, 74.5143177, 64.9245638, 55.4004716, 45.9485623, 36.575357, 27.287377, 18.0911434, 8.9931773, 0, -6.9064089, -13.817204, -20.7524485, -27.732206, -34.7765397, -41.9055132, -49.1391896, -56.4976324, -64.0009051, -71.6690709, -79.5221932, -87.5803355, -95.863561, -104.391933, -113.185516, -122.264371, -131.6197, -141.267442, -151.292947, -161.710094, -172.532761, -183.774826, -195.450169, -207.572666, -220.156196, -233.214638, -246.761869, -260.811769, -275.378214, -290.402527, -305.997747, -322.199519, -339.04349, -356.153164, -373.770256, -392.246387, -411.740846, -432.41292, -454.421898, -477.927067, -505.31838, -533.845563, -563.23147, -593.198953, -623.470867, -654.581798, -685.239934, -714.965196, -743.277503, -769.696777, -792.345005, -813.848625, -840.158261, -872.336762, -901.467762, -923.115361, -932.843657, -914.676435, -878.340905, -846.153716, -853.234608, -849.952386, -795.87225, -669.944648, -483.967256, -234.66033, 63.4811688, 409.128726, 866.127732, 1461.45753, 2027.65789, 2816.06151, 3920.95374, 5436.61993, 7457.34543, 12323.0404, 17320.9591, 21983.9808, 25844.9847, 28436.8498, 27407.6574, 25431.6165, 22849.3767, 20979.3564, 19775.1716, 18783.0352, 17976.9334, 17330.8527, 16818.7796, 16414.7003, 16558.8057, 16418.0453, 16074.1359, 15608.7943, 15103.7372, 14640.6812, 14301.3432, 13809.1951, 13279.09},
        {819.559987, 816.042758, 812.271591, 808.25306, 803.993739, 799.500202, 794.779022, 789.836772, 784.680026, 779.315358, 773.749341, 767.988549, 762.039555, 755.908933, 749.603256, 743.129099, 736.493034, 729.701635, 722.761477, 715.679131, 708.461172, 701.114173, 693.644709, 686.059352, 678.364676, 670.567255, 662.673662, 654.69047, 646.624254, 638.481587, 630.269043, 621.993194, 613.660615, 605.277879, 597.062876, 588.912076, 580.71941, 572.484595, 564.207346, 555.88738, 547.524412, 539.118158, 530.668335, 522.174657, 513.636842, 505.054605, 496.427662, 487.75573, 479.038523, 470.275758, 461.467151, 452.47769, 443.44474, 434.370937, 425.258917, 416.111318, 406.930776, 397.719927, 388.481408, 379.217856, 369.931907, 360.626198, 351.303366, 341.966046, 332.616877, 323.258493, 313.893533, 304.559274, 295.292164, 286.023392, 276.753349, 267.482427, 258.211015, 248.939506, 239.668289, 230.397756, 221.128298, 211.860305, 202.594169, 193.33028, 184.069029, 174.810807, 165.556006, 156.305015, 146.860925, 137.326928, 127.808586, 118.312692, 108.84604, 99.4154242, 90.0276378, 80.6894745, 71.407728, 62.1891919, 53.04066, 43.9689258, 34.9807831, 26.0830255, 17.2824467, 8.5858403, 0, -5.78508753, -11.7000624, -17.7561911, -23.9647403, -30.3369766, -36.8841666, -43.6175767, -50.5484737, -57.688124, -65.0477944, -72.6387512, -80.4722613, -88.559591, -96.912007, -105.540776, -114.457164, -123.757658, -133.465697, -143.399791, -153.589468, -164.064256, -174.853682, -185.987275, -197.494562, -209.405071, -221.748329, -234.553865, -247.851206, -261.66988, -276.214752, -291.275722, -306.829717, -322.853662, -338.478906, -354.105165, -370.357641, -387.466934, -405.663645, -425.178374, -446.24172, -472.101785, -499.053298, -526.736476, -554.791539, -582.858706, -613.510226, -642.721279, -669.399078, -692.450834, -710.783757, -718.616738, -725.27548, -742.163395, -772.354425, -801.916511, -824.552941, -833.967004, -812.875863, -768.465778, -719.907872, -693.720374, -650.955554, -562.708709, -411.842829, -201.160606, 75.6879179, 408.15415, 798.206942, 1309.7765, 1971.74419, 2648.8664, 3559.39946, 4789.03721, 6423.47351, 8548.4022, 13313.4151, 18224.3336, 22850.8771, 26762.7648, 29529.7163, 29026.2878, 27647.4703, 25689.4817, 24282.5423, 23392.7438, 22638.677, 22003.8085, 21471.6052, 21025.5338, 20649.0611, 21104.5835, 21051.8662, 20629.6832, 19976.8083, 19232.0154, 18534.0786, 18021.7716, 17297.0979, 16563.3361},
        {771.4088, 768.057881, 764.495505, 760.726583, 756.756024, 752.588738, 748.229636, 743.683627, 738.955622, 734.050531, 728.973264, 723.728731, 718.321842, 712.757507, 707.040637, 701.176141, 695.168929, 689.023912, 682.746, 676.340103, 669.811131, 663.163993, 656.403601, 649.534864, 642.562692, 635.491996, 628.327685, 621.07467, 613.737861, 606.322167, 598.8325, 591.273768, 583.650882, 575.968753, 568.377183, 560.805589, 553.181644, 545.505557, 537.777535, 529.997788, 522.166523, 514.283947, 506.35027, 498.365698, 490.330441, 482.244705, 474.1087, 465.922633, 457.686712, 449.401146, 441.066142, 432.569557, 424.026386, 415.439273, 406.810861, 398.143793, 389.440714, 380.704266, 371.937095, 363.141842, 354.321152, 345.477668, 336.614034, 327.732893, 318.836889, 309.928666, 301.010867, 292.122732, 283.302623, 274.477706, 265.648251, 256.814531, 247.976815, 239.135377, 230.290486, 221.442414, 212.591432, 203.737812, 194.881823, 186.023739, 177.16383, 168.302366, 159.43962, 150.575862, 141.493151, 132.305454, 123.129361, 113.972225, 104.841399, 95.7442352, 86.6880864, 77.6803052, 68.7282442, 59.8392561, 51.0206934, 42.2799089, 33.6242551, 25.0610847, 16.5977503, 8.2416045, 0, -5.21525898, -10.5998448, -16.1611745, -21.9066653, -27.8437342, -33.9797985, -40.3222753, -46.8785816, -53.6561347, -60.6623515, -67.9046494, -75.3904454, -83.1271566, -91.1222001, -99.3829932, -107.916953, -116.855078, -126.222439, -135.746926, -145.462435, -155.402867, -165.60212, -176.094093, -186.912684, -198.091794, -209.66532, -221.667161, -234.131217, -247.091386, -260.807827, -275.005218, -289.649577, -304.706927, -319.243703, -333.675718, -348.658675, -364.428468, -381.220994, -399.272148, -418.817826, -443.659051, -469.381553, -495.523704, -521.623873, -547.220433, -573.88384, -598.612358, -620.436336, -638.386125, -651.492073, -654.845011, -656.229778, -665.337134, -684.716451, -702.336762, -712.713374, -710.361594, -682.319221, -632.278504, -572.088974, -519.750086, -445.091148, -328.465191, -155.465037, 79.4379588, 385.847833, 759.3014, 1205.75564, 1782.74734, 2518.63685, 3290.16714, 4300.64165, 5630.90364, 7361.79639, 9574.16315, 14318.1783, 19213.0211, 23847.2023, 27809.2323, 30687.6215, 30466.6551, 29408.5525, 27789.3496, 26658.7057, 25986.4304, 25419.9903, 24945.312, 24548.3225, 24214.9485, 23931.1168, 24544.0084, 24582.5811, 24197.8495, 23540.8282, 22762.5318, 22013.975, 21446.1723, 20674.0552, 19776.4636},
        {727.568456, 724.124736, 720.529622, 716.78574, 712.895716, 708.862175, 704.687744, 700.375049, 695.926715, 691.345368, 686.633635, 681.794141, 676.829511, 671.742374, 666.535352, 661.211074, 655.772165, 650.22125, 644.560956, 638.793909, 632.922734, 626.950057, 620.878505, 614.710702, 608.449276, 602.096852, 595.656056, 589.129514, 582.519852, 575.829695, 569.06167, 562.218403, 555.302519, 548.316644, 541.317157, 534.278677, 527.175177, 520.007539, 512.776643, 505.483372, 498.128606, 490.713229, 483.23812, 475.704162, 468.112236, 460.463224, 452.758007, 444.997467, 437.182486, 429.313944, 421.392724, 413.329407, 405.217132, 397.058738, 388.857064, 380.61495, 372.335233, 364.020755, 355.674352, 347.298865, 338.897133, 330.471995, 322.026289, 313.562854, 305.084531, 296.594158, 288.094573, 279.634928, 271.263262, 262.886899, 254.505676, 246.11943, 237.727997, 229.331216, 220.928922, 212.520952, 204.107144, 195.687334, 187.261359, 178.829056, 170.390262, 161.944813, 153.492547, 145.033301, 136.316326, 127.472023, 118.633803, 109.809635, 101.007489, 92.2353315, 83.5011328, 74.8128614, 66.1784861, 57.6059757, 49.103299, 40.6784248, 32.3393219, 24.0939592, 15.9503055, 7.91632944, 0, -4.6947176, -9.59355123, -14.7003003, -20.0187641, -25.5527422, -31.3060338, -37.2824383, -43.4857551, -49.9197836, -56.5883232, -63.4951733, -70.6441332, -78.0390023, -85.68358, -93.5816656, -101.737059, -110.314035, -119.33932, -128.453729, -137.695449, -147.102669, -156.713574, -166.566353, -176.699193, -187.15028, -197.957803, -209.159948, -220.794903, -232.900854, -245.77454, -259.100796, -272.840243, -286.953505, -300.526318, -313.956747, -327.876158, -342.507639, -358.074277, -374.799158, -392.90537, -416.535335, -440.799965, -465.15552, -489.058259, -511.964446, -534.227869, -554.182877, -571.061349, -584.095161, -592.516192, -592.851832, -590.343934, -591.819601, -598.957288, -602.923615, -599.604804, -584.887075, -552.04877, -500.176686, -431.202575, -351.904205, -243.770054, -98.7498833, 93.8407544, 349.149821, 680.276453, 1089.8878, 1588.37839, 2224.39342, 3024.94134, 3882.01642, 4978.41271, 6387.87434, 8184.14543, 10440.9701, 15066.6985, 19841.8172, 24381.4189, 28300.5963, 31214.442, 31248.441, 30500.3653, 29223.7109, 28383.1703, 27950.724, 27602.0208, 27323.8579, 27103.0325, 26926.3418, 26780.583, 27531.518, 27683.9477, 27388.2187, 26794.6773, 26053.6701, 25315.5434, 24730.6438, 23960.2895, 22932.1602},
        {686.487617, 682.949948, 679.318433, 675.593476, 671.775484, 667.864861, 663.862013, 659.767346, 655.581265, 651.304174, 646.936481, 642.478589, 637.930905, 633.293833, 628.56778, 623.75315, 618.850349, 613.859782, 608.781855, 603.616974, 598.365542, 593.027967, 587.604652, 582.096005, 576.502429, 570.824331, 565.062116, 559.216189, 553.286955, 547.27482, 541.18019, 535.00347, 528.745064, 522.40538, 515.94686, 509.389689, 502.754508, 496.042954, 489.256665, 482.397278, 475.466429, 468.465757, 461.396899, 454.261492, 447.061173, 439.797579, 432.472348, 425.087118, 417.643525, 410.143207, 402.5878, 394.911552, 387.184952, 379.411097, 371.593087, 363.734018, 355.83699, 347.9051, 339.941448, 331.94913, 323.931245, 315.890892, 307.831169, 299.755173, 291.666004, 283.566758, 275.460536, 267.409805, 259.47561, 251.5386, 243.598026, 235.653135, 227.703178, 219.747404, 211.785063, 203.815403, 195.837674, 187.851126, 179.855008, 171.848569, 163.83106, 155.801728, 147.759824, 139.704597, 131.347037, 122.836585, 114.32615, 105.824335, 97.3397448, 88.8809823, 80.4566521, 72.075358, 63.7457042, 55.4762946, 47.2757331, 39.1526238, 31.1155707, 23.1731778, 15.3340491, 7.60678845, 0, -4.20069216, -8.63840868, -13.3134716, -18.226203, -23.376925, -28.7659595, -34.3936288, -40.2602547, -46.3661595, -52.7116651, -59.2970936, -66.1227671, -73.1890076, -80.4961373, -88.0444781, -95.8343521, -104.064154, -112.762502, -121.481697, -130.264505, -139.153692, -148.192024, -157.422267, -166.887188, -176.629553, -186.692129, -197.11768, -207.948974, -219.228777, -231.288969, -243.777194, -256.649483, -269.861869, -282.565061, -295.117754, -308.093391, -321.689601, -336.104013, -351.534254, -368.177954, -390.283197, -412.764408, -435.026735, -456.475324, -476.515326, -494.434644, -509.78498, -522.000792, -530.51654, -534.766681, -532.830985, -527.154332, -521.356965, -516.141209, -506.251975, -489.108931, -462.131744, -425.330017, -372.944863, -296.392015, -190.667123, -49.5044117, 123.498695, 335.066778, 609.414481, 962.728223, 1404.38729, 1948.8133, 2635.38732, 3489.37194, 4420.95964, 5588.18623, 7056.14625, 8889.93427, 11154.6448, 15587.5691, 20163.5559, 24529.6506, 28332.8987, 31220.3455, 31479.5989, 31023.4335, 30081.5108, 29531.9041, 29347.6118, 29231.4606, 29169.9573, 29149.6089, 29156.9223, 29178.4044, 30033.4589, 30308.8548, 30141.4918, 29668.2694, 29026.0872, 28351.8449, 27782.442, 27049.4914, 25934.9294},
        {646.614948, 643.240145, 639.80643, 636.312736, 632.757996, 629.141143, 625.46111, 621.716829, 617.907233, 614.031256, 610.087829, 606.075887, 601.994362, 597.842187, 593.618294, 589.321617, 584.951088, 580.505641, 575.984208, 571.385722, 566.709117, 561.953324, 557.117277, 552.199908, 547.200152, 542.116939, 536.949205, 531.69588, 526.355898, 520.928193, 515.411696, 509.805341, 504.108061, 498.318788, 492.330353, 486.196971, 479.974134, 473.664218, 467.2696, 460.792655, 454.235759, 447.601289, 440.891621, 434.10913, 427.256194, 420.335188, 413.348488, 406.298471, 399.187512, 392.017988, 384.792275, 377.470307, 370.097822, 362.678117, 355.214487, 347.710229, 340.168639, 332.593013, 324.986647, 317.352838, 309.694882, 302.016075, 294.319713, 286.609092, 278.887509, 271.158259, 263.42464, 255.761305, 248.241197, 240.72044, 233.197704, 225.671658, 218.140975, 210.604323, 203.060373, 195.507796, 187.945261, 180.37144, 172.785002, 165.184617, 157.568956, 149.936689, 142.286487, 134.617019, 126.60187, 118.409091, 110.21064, 102.015737, 93.8336025, 85.6734566, 77.5445191, 69.4560099, 61.417149, 53.4371565, 45.5252523, 37.6906563, 29.9425886, 22.2902691, 14.7429179, 7.30975484, 0, -3.71041147, -7.69164402, -11.9405916, -16.454148, -21.2292072, -26.2626632, -31.5514098, -37.092341, -42.8823506, -48.9183326, -55.1971809, -61.7157894, -68.471052, -75.4598626, -82.6791152, -90.1257036, -98.0350591, -106.438152, -114.792329, -123.145597, -131.545966, -140.041446, -148.680044, -157.509771, -166.578635, -175.934646, -185.625811, -195.700142, -206.205646, -217.525191, -229.249146, -241.325063, -253.700492, -265.598245, -277.328242, -289.393671, -301.959506, -315.19072, -329.252288, -344.309183, -364.455195, -384.730756, -404.532159, -423.255695, -440.297657, -454.396496, -465.770807, -473.981343, -478.588859, -479.154107, -474.976254, -466.197357, -453.695395, -437.332493, -414.896748, -385.107453, -346.683901, -305.428466, -251.36757, -166.800636, -36.5232375, 135.193639, 335.342026, 567.205268, 861.671439, 1236.95759, 1707.27385, 2289.79844, 3016.40161, 3910.64298, 4903.54221, 6125.43578, 7631.91633, 9478.57654, 11721.0091, 15909.3837, 20231.0717, 24368.0215, 28002.1813, 30815.4998, 31268.0824, 31078.2821, 30451.6942, 30180.8749, 30239.0809, 30355.0013, 30514.1212, 30701.9253, 30903.8985, 31105.5257, 32016.1779, 32410.1913, 32398.37, 32091.5184, 31600.4405, 31035.9406, 30508.8229, 29835.351, 28689.2748},
        {606.399111, 603.701953, 600.938107, 598.106466, 595.205922, 592.235369, 589.193699, 586.079806, 582.892581, 579.630917, 576.293708, 572.879846, 569.388224, 565.817734, 562.167271, 558.435725, 554.62199, 550.72496, 546.743525, 542.676581, 538.523018, 534.28173, 529.95161, 525.531551, 521.020445, 516.417185, 511.720664, 506.929774, 502.043409, 497.060461, 491.979824, 486.800388, 481.521049, 476.140698, 470.531699, 464.758871, 458.888553, 452.923744, 446.867444, 440.72265, 434.492362, 428.179578, 421.787298, 415.318521, 408.776244, 402.163468, 395.483191, 388.738412, 381.932129, 375.067343, 368.147051, 361.159983, 354.123717, 347.041562, 339.916824, 332.752812, 325.552835, 318.320201, 311.058218, 303.770194, 296.459437, 289.129256, 281.782958, 274.423853, 267.055248, 259.680451, 252.30277, 245.003372, 237.861553, 230.720046, 223.577111, 216.43101, 209.280004, 202.122353, 194.956319, 187.780163, 180.592146, 173.390528, 166.173571, 158.939536, 151.686684, 144.413276, 137.117572, 129.797835, 122.097413, 114.199488, 106.291508, 98.3832518, 90.4844974, 82.6050235, 74.7546089, 66.9430319, 59.1800712, 51.4755053, 43.8391129, 36.2806724, 28.8099625, 21.4367618, 14.1708487, 7.02200194, 0, -3.20110431, -6.71048416, -10.5215632, -14.6277651, -19.0225135, -23.6992321, -28.6513446, -33.8722745, -39.3554456, -45.0942815, -51.082206, -57.3126425, -63.7790149, -70.4747468, -77.3932617, -84.5279835, -92.1563744, -100.312435, -108.347121, -116.31472, -124.269523, -132.265818, -140.357895, -148.600042, -157.046548, -165.751704, -174.769796, -184.155116, -193.961951, -204.657283, -215.731388, -227.114749, -238.737848, -249.864178, -260.757715, -271.860299, -283.302507, -295.214917, -307.728108, -320.972658, -338.603887, -356.154883, -373.066602, -388.779997, -402.736023, -414.005757, -422.492501, -427.729679, -429.250717, -426.589038, -419.481422, -407.009392, -388.581059, -363.595415, -331.432841, -291.482071, -243.131844, -195.609624, -136.229344, -41.5717776, 110.043058, 307.811958, 533.841589, 789.248458, 1107.36019, 1506.71899, 2003.02142, 2614.0719, 3368.10888, 4287.46875, 5326.30954, 6585.63494, 8111.38153, 9949.48587, 12145.8845, 16060.7356, 20097.1987, 23972.6552, 27404.4861, 30110.0727, 30721.8448, 30765.4359, 30423.2064, 30406.0509, 30687.1184, 31019.335, 31386.8606, 31773.8553, 32164.4789, 32542.8915, 33446.0217, 33940.846, 34099.5547, 33994.3382, 33697.3871, 33280.8915, 32817.0421, 32211.5587, 31099.6997},
        {558.061705, 557.780689, 557.291903, 556.598608, 555.70407, 554.61155, 553.324314, 551.845623, 550.178741, 548.326932, 546.293459, 544.081585, 541.694573, 539.135688, 536.408192, 533.515349, 530.460422, 527.246674, 523.877369, 520.35577, 516.685141, 512.868744, 508.909844, 504.811703, 500.577586, 496.210754, 491.714472, 487.092003, 482.346611, 477.481558, 472.500108, 467.405525, 462.201072, 456.890011, 451.474562, 445.958532, 440.345742, 434.639489, 428.843071, 422.959783, 416.992925, 410.945793, 404.821684, 398.623896, 392.355725, 386.02047, 379.621428, 373.161895, 366.645169, 360.074547, 353.453327, 346.813561, 340.129167, 333.40282, 326.637193, 319.834961, 312.998798, 306.131377, 299.235372, 292.313458, 285.368308, 278.402596, 271.418996, 264.420182, 257.408828, 250.387608, 243.359195, 236.390258, 229.545927, 222.696894, 215.841685, 208.978825, 202.10684, 195.224256, 188.329597, 181.42139, 174.49816, 167.558432, 160.600732, 153.623585, 146.625517, 139.605054, 132.56072, 125.491042, 118.033777, 110.375423, 102.706813, 95.0381814, 87.3797604, 79.7417832, 72.134483, 64.568093, 57.0528461, 49.5989756, 42.2167145, 34.916296, 27.7079531, 20.6019189, 13.6084266, 6.73770926, 0, -2.4457295, -5.2760342, -8.47895805, -12.042545, -15.954839, -20.203884, -24.7777239, -29.6644028, -34.8519644, -40.3284529, -46.0819121, -52.1003859, -58.3719184, -64.8845535, -71.6263351, -78.5853071, -86.1261422, -94.2906882, -102.215091, -109.968101, -117.618469, -125.234943, -132.886275, -140.641214, -148.568511, -156.736915, -165.215177, -174.072046, -183.376273, -193.848327, -204.666275, -215.703351, -226.832791, -237.052416, -246.673167, -256.239427, -265.887053, -275.751903, -285.969834, -296.676704, -310.909705, -325.020346, -338.576831, -351.147366, -362.300156, -373.916822, -382.673799, -387.560939, -387.568093, -381.685113, -367.149308, -346.845066, -322.818544, -295.437727, -261.644326, -219.91838, -168.739928, -116.555308, -49.6131907, 56.7100827, 225.651432, 446.364652, 706.981834, 1008.58016, 1374.17156, 1816.69954, 2340.15196, 2955.15047, 3705.15426, 4615.04774, 5679.31296, 6957.914, 8486.25919, 10299.7569, 12433.8153, 16052.0419, 19779.5962, 23369.8367, 26576.122, 29151.8107, 29905.0186, 30164.5088, 30088.7439, 30303.8174, 30783.088, 31305.0682, 31852.859, 32409.5615, 32958.2767, 33482.1056, 34077.9515, 34564.9212, 34919.4067, 35117.7997, 35136.4922, 34951.8759, 34540.3428, 34016.1238, 33098.6953},
        {519.490016, 520.604075, 521.436213, 521.991988, 522.27696, 522.296687, 522.056728, 521.562642, 520.819986, 519.83432, 518.611203, 517.156194, 515.47485, 513.572731, 511.455395, 509.128402, 506.597309, 503.867675, 500.94506, 497.835022, 494.543119, 491.074911, 487.435955, 483.631811, 479.668038, 475.550193, 471.283837, 466.874526, 462.327821, 457.64928, 452.844462, 447.918924, 442.878227, 437.727928, 432.536584, 427.276929, 421.920942, 416.472137, 410.934027, 405.310128, 399.603954, 393.819019, 387.958837, 382.026923, 376.02679, 369.961954, 363.835928, 357.652228, 351.414366, 345.125858, 338.790218, 332.466153, 326.100788, 319.69644, 313.255429, 306.78007, 300.272683, 293.735585, 287.171094, 280.581528, 273.969205, 267.336442, 260.685557, 254.018868, 247.338693, 240.64735, 233.947157, 227.298487, 220.760322, 214.215242, 207.661802, 201.098554, 194.524055, 187.936858, 181.335517, 174.718586, 168.084621, 161.432174, 154.759801, 148.066056, 141.349493, 134.608666, 127.842129, 121.048437, 113.85579, 106.45571, 99.0441678, 91.6317376, 84.2289919, 76.8465039, 69.4948463, 62.1845923, 54.9263146, 47.7305864, 40.6079805, 33.5690699, 26.6244275, 19.7846264, 13.0602395, 6.46183969, 0, -2.05847428, -4.51836185, -7.36586397, -10.5871819, -14.1685169, -18.0960702, -22.3560431, -26.9346368, -31.8180527, -36.9924918, -42.4441556, -48.1592453, -54.1239621, -60.3245073, -66.7470822, -73.3778879, -80.5969973, -88.446879, -96.0228344, -103.395466, -110.635376, -117.813166, -124.999439, -132.264798, -139.679844, -147.315179, -155.241407, -163.529129, -172.248948, -182.189868, -192.440675, -202.856451, -213.292277, -222.872386, -231.817285, -240.542373, -249.121984, -257.630454, -266.142121, -274.73132, -285.218556, -295.400553, -305.010005, -313.779608, -321.442054, -331.686577, -339.300199, -343.026478, -341.608976, -333.791251, -316.25026, -292.322015, -265.267344, -235.55638, -197.961582, -149.813551, -88.442885, -17.2242582, 73.9294582, 201.698929, 380.984315, 608.968786, 880.145136, 1198.09081, 1581.77671, 2043.10786, 2585.85569, 3219.81651, 3984.61, 4902.98608, 5979.3725, 7257.66011, 8766.94108, 10536.3076, 12594.8519, 15968.2563, 19439.8753, 22789.6534, 25797.5353, 28243.4656, 29060.5302, 29440.1048, 29525.0734, 29886.547, 30497.9031, 31154.0911, 31837.4099, 32530.1582, 33214.6348, 33873.1385, 34538.7407, 35136.9701, 35630.7367, 35982.9507, 36156.5222, 36114.3612, 35819.378, 35319.6369, 34506.9535},
        {489.408433, 490.956453, 492.21316, 493.184383, 493.875954, 494.293704, 494.443464, 494.331065, 493.962337, 493.343113, 492.479222, 491.376495, 490.040765, 488.477861, 486.693615, 484.693857, 482.48442, 480.071132, 477.459827, 474.656333, 471.666484, 468.496109, 465.151039, 461.637106, 457.96014, 454.125972, 450.140434, 446.009356, 441.73857, 437.333906, 432.801195, 428.146268, 423.374957, 418.493092, 413.573866, 408.588013, 403.505455, 398.329838, 393.064807, 387.714006, 382.281081, 376.769676, 371.183437, 365.526008, 359.801034, 354.01216, 348.163032, 342.257293, 336.29859, 330.290567, 324.236868, 318.20591, 312.135163, 306.026868, 299.883265, 293.706595, 287.499099, 281.263018, 275.000592, 268.714063, 262.405672, 256.077658, 249.732263, 243.371729, 236.998294, 230.614202, 224.221691, 217.882538, 211.657087, 205.424795, 199.184044, 192.933214, 186.670688, 180.394847, 174.104073, 167.796746, 161.471249, 155.125962, 148.759268, 142.369548, 135.955183, 129.514555, 123.046046, 116.548036, 109.635405, 102.506352, 95.3636851, 88.2182321, 81.0808194, 73.962274, 66.8734226, 59.8250923, 52.8281099, 45.8933022, 39.0314962, 32.2535187, 25.5701966, 18.9923568, 12.5308262, 6.19643162, 0, -1.98681623, -4.34051091, -7.04860201, -10.0986075, -13.4780453, -17.1744334, -21.1752897, -25.4681323, -30.040479, -34.8798478, -39.9737567, -45.3097236, -50.8752665, -56.6579034, -62.6451522, -68.8245308, -75.5141456, -82.7487412, -89.7680767, -96.6305101, -103.3944, -110.118104, -116.859982, -123.67839, -130.631688, -137.778234, -145.176386, -152.884503, -160.960942, -170.077792, -179.454647, -188.965746, -198.48533, -207.574336, -216.263654, -224.667721, -232.754767, -240.493019, -247.850709, -254.796065, -261.703566, -268.011552, -273.608768, -278.383961, -282.225876, -289.58857, -294.654151, -296.170038, -292.88365, -283.542404, -265.714086, -240.767523, -211.955771, -179.225917, -135.775359, -77.8319702, -1.62361969, 96.5199791, 220.495943, 374.312489, 559.564145, 785.038501, 1050.0007, 1361.60724, 1740.38611, 2200.72754, 2754.77777, 3416.34525, 4207.46656, 5145.52398, 6223.46125, 7486.85898, 8961.55892, 10673.4029, 12648.2326, 15817.6765, 19075.343, 22220.6273, 25052.9246, 27371.6298, 28197.0907, 28627.1149, 28794.9748, 29237.1072, 29927.199, 30670.9517, 31450.4515, 32247.7843, 33045.0361, 33824.293, 34835.1013, 35632.5503, 36224.5323, 36618.9394, 36823.6641, 36846.5983, 36695.6345, 36189.6973, 35425.4022},
        {464.357294, 465.76943, 466.919421, 467.812177, 468.45261, 468.845633, 468.996156, 468.909091, 468.58935, 468.041844, 467.271485, 466.283184, 465.081853, 463.672404, 462.059748, 460.248797, 458.244462, 456.051654, 453.675287, 451.12027, 448.391516, 445.493937, 442.432443, 439.211946, 435.837358, 432.313591, 428.645556, 424.838165, 420.896329, 416.82496, 412.628969, 408.313269, 403.882769, 399.342384, 394.735895, 390.047961, 385.262773, 380.38398, 375.415233, 370.360182, 365.222476, 360.005766, 354.713703, 349.349935, 343.918114, 338.421889, 332.86491, 327.250828, 321.583292, 315.865953, 310.10246, 304.357293, 298.571953, 292.748774, 286.890086, 280.99822, 275.075509, 269.124283, 263.146874, 257.145614, 251.122833, 245.080864, 239.022038, 232.948687, 226.863141, 220.767733, 214.664793, 208.621062, 202.701734, 196.776303, 190.842923, 184.899751, 178.944944, 172.976657, 166.993047, 160.992269, 154.97248, 148.931836, 142.868492, 136.780605, 130.666332, 124.523828, 118.351249, 112.146751, 105.511974, 98.6516711, 91.7753646, 84.8940779, 78.0188348, 71.1606591, 64.3305748, 57.5396056, 50.7987754, 44.1191081, 37.5116274, 30.9873574, 24.5573217, 18.2325443, 12.0240489, 5.94285958, 0, -2.1047361, -4.51022423, -7.20744046, -10.1873608, -13.4409613, -16.9592181, -20.7331071, -24.7536044, -29.011686, -33.498328, -38.2045065, -43.1211975, -48.2393769, -53.5500209, -59.0441056, -64.7126069, -70.7418229, -77.150633, -83.4882131, -89.7873941, -96.0810066, -102.401882, -108.78285, -115.256742, -121.856388, -128.614621, -135.564269, -142.738165, -150.169138, -158.267542, -166.550261, -174.936868, -183.346939, -191.921151, -200.468526, -208.739128, -216.586198, -223.862981, -230.42272, -236.11866, -240.083337, -243.110045, -245.193037, -246.326563, -246.504874, -250.278918, -251.947075, -250.364422, -244.386035, -232.866991, -214.625147, -188.598292, -157.696708, -121.248107, -71.2657855, -3.04706414, 88.1107372, 212.544497, 368.042197, 546.24542, 736.071031, 956.731203, 1209.99618, 1506.14805, 1869.8917, 2319.31173, 2877.59121, 3564.13655, 4379.65977, 5335.44749, 6406.28128, 7645.56899, 9077.23361, 10725.1981, 12613.3856, 15601.2175, 18668.2437, 21629.5125, 24300.0722, 26494.9711, 27303.2112, 27749.9181, 27961.303, 28445.8995, 29178.5827, 29974.9266, 30817.7714, 31689.9573, 32574.3244, 33453.713, 34901.0846, 35926.4035, 36600.2035, 36993.0185, 37175.3823, 37217.8289, 37190.8922, 36682.5593, 35974.2226},
        {437.858147, 439.195971, 440.298293, 441.169237, 441.812925, 442.233481, 442.435028, 442.421689, 442.197587, 441.766845, 441.133586, 440.301934, 439.276011, 438.05994, 436.657845, 435.073849, 433.312075, 431.376645, 429.271684, 427.001314, 424.569658, 421.980839, 419.238981, 416.348206, 413.312638, 410.136399, 406.823614, 403.378404, 399.804894, 396.107205, 392.289462, 388.355787, 384.310304, 380.157135, 375.923669, 371.601906, 367.183569, 362.672026, 358.070644, 353.382792, 348.611838, 343.761151, 338.834097, 333.834046, 328.764365, 323.628423, 318.429587, 313.171226, 307.856708, 302.489401, 297.072672, 291.658753, 286.20109, 280.701991, 275.163766, 269.588724, 263.979174, 258.337424, 252.665784, 246.966562, 241.242067, 235.494609, 229.726495, 223.940035, 218.137538, 212.321313, 206.493668, 200.717037, 195.052411, 189.378095, 183.6925, 177.994039, 172.281122, 166.552159, 160.805564, 155.039746, 149.253118, 143.44409, 137.611073, 131.752479, 125.86672, 119.952205, 114.007347, 108.030557, 101.629459, 95.006078, 88.3671344, 81.7237218, 75.0869331, 68.4678613, 61.8775996, 55.3272409, 48.8278782, 42.3906047, 36.0265133, 29.7466971, 23.5622491, 17.4842624, 11.5238299, 5.69204479, 0, -2.3402798, -4.8855765, -7.63332381, -10.5809554, -13.725905, -17.0656063, -20.5974929, -24.3189987, -28.2275572, -32.3206022, -36.5955674, -41.0498864, -45.680993, -50.4863209, -55.4633037, -60.6093752, -65.830628, -71.1114472, -76.655871, -82.4417602, -88.4469752, -94.6493768, -101.026826, -107.557182, -114.218307, -120.988061, -127.844305, -134.764899, -141.727704, -148.652355, -155.596288, -162.554831, -169.52331, -177.165694, -185.142991, -192.937906, -200.345176, -207.159537, -213.175726, -218.18848, -221.348209, -223.290075, -223.934878, -223.203417, -221.016493, -222.262164, -220.652158, -214.86546, -203.581056, -185.477931, -157.961595, -122.54098, -83.8708373, -42.0487781, 12.1317463, 83.6999065, 177.684873, 303.776742, 461.284458, 644.432319, 847.275094, 1082.37822, 1345.25473, 1643.5951, 2005.72375, 2450.49377, 3003.40945, 3685.04825, 4491.60002, 5433.04963, 6492.47796, 7710.91293, 9107.65235, 10701.994, 12513.2357, 15300.3743, 18158.0837, 20920.7369, 23422.7071, 25498.3671, 26329.7547, 26838.4685, 27138.4537, 27684.0408, 28454.4392, 29287.0312, 30168.1177, 31083.9995, 32020.9773, 32965.352, 34654.0494, 35859.927, 36656.7915, 37118.4495, 37318.7076, 37331.3724, 37230.2505, 36775.6961, 36514.2609},
        {411.415967, 412.737372, 413.845687, 414.744402, 415.437004, 415.926984, 416.21783, 416.313032, 416.216078, 415.930458, 415.459661, 414.807176, 413.976491, 412.971097, 411.794483, 410.450136, 408.941547, 407.272205, 405.445598, 403.465217, 401.334549, 399.057084, 396.636312, 394.07572, 391.3788, 388.549038, 385.589926, 382.504951, 379.297603, 375.971371, 372.529744, 368.976211, 365.314262, 361.547385, 357.695326, 353.753105, 349.715544, 345.585606, 341.36625, 337.06044, 332.671137, 328.201302, 323.653897, 319.031882, 314.338221, 309.575874, 304.747804, 299.85697, 294.906336, 289.898862, 284.83751, 279.753904, 274.621722, 269.443304, 264.220991, 258.957121, 253.654037, 248.314077, 242.939583, 237.532893, 232.09635, 226.632292, 221.14306, 215.630994, 210.098435, 204.547722, 198.981196, 193.457906, 188.035539, 182.599479, 177.148388, 171.680931, 166.195771, 160.691574, 155.167002, 149.62072, 144.051391, 138.457681, 132.838251, 127.191768, 121.516894, 115.812294, 110.076632, 104.308571, 98.1196331, 91.7101958, 85.2852913, 78.8561473, 72.4339912, 66.0300505, 59.6555529, 53.3217258, 47.0397968, 40.8209934, 34.6765432, 28.6176737, 22.6556123, 16.8015868, 11.0668245, 5.46255307, 0, -2.47851376, -5.08575286, -7.82450469, -10.6975566, -13.7076961, -16.8577104, -20.1503871, -23.5885133, -27.1748767, -30.9122644, -34.803464, -38.8512628, -43.0584482, -47.4278076, -51.9621284, -56.6641979, -61.1605981, -65.4003729, -70.2372503, -75.5934023, -81.3910009, -87.552218, -93.9992256, -100.654196, -107.4393, -114.27671, -121.088599, -127.797138, -134.324499, -140.116486, -145.746308, -151.279046, -156.779782, -163.265049, -170.324203, -177.292884, -183.950738, -190.077414, -195.452559, -199.855821, -203.125303, -204.964406, -205.141342, -203.424321, -199.581553, -198.576239, -193.682852, -183.370857, -166.109717, -140.368895, -101.560775, -54.9483307, -7.38135403, 39.9279411, 97.2773971, 169.797076, 262.61704, 382.653226, 532.84394, 714.179262, 930.202385, 1183.3208, 1462.91431, 1772.96895, 2142.22021, 2588.04318, 3132.57881, 3794.43778, 4574.75592, 5484.55381, 6526.18394, 7720.66856, 9082.02879, 10624.2857, 12361.4604, 14948.0004, 17597.3938, 20163.555, 22500.3986, 24461.8391, 25327.1472, 25907.977, 26304.5187, 26909.1892, 27706.5058, 28559.2248, 29457.9517, 30393.2918, 31355.8506, 32336.2334, 34055.4998, 35362.7, 36294.3134, 36886.8194, 37176.6975, 37200.4271, 36994.4875, 36654.9242, 36769.9049},
        {385.697795, 387.143572, 388.386933, 389.431103, 390.279304, 390.934759, 391.400691, 391.680322, 391.776877, 391.693578, 391.433648, 391.000309, 390.396786, 389.626301, 388.692076, 387.597336, 386.345302, 384.939198, 383.382247, 381.677672, 379.828696, 377.838542, 375.710433, 373.447592, 371.053241, 368.530605, 365.882905, 363.113365, 360.225208, 357.221657, 354.105934, 350.881264, 347.550868, 344.11797, 340.60649, 337.008868, 333.317292, 329.534316, 325.662489, 321.704364, 317.662491, 313.539423, 309.337709, 305.059902, 300.708553, 296.286213, 291.795434, 287.238766, 282.618762, 277.937972, 273.198948, 268.402332, 263.552625, 258.652421, 253.704312, 248.710892, 243.674752, 238.598485, 233.484684, 228.335942, 223.154852, 217.944006, 212.705996, 207.443417, 202.158859, 196.854917, 191.534182, 186.263055, 181.106403, 175.935222, 170.747968, 165.543097, 160.319066, 155.074332, 149.80735, 144.516577, 139.200469, 133.857482, 128.486074, 123.0847, 117.651817, 112.185881, 106.685348, 101.148675, 95.1681838, 88.9539403, 82.720892, 76.4806753, 70.2449263, 64.0252812, 57.8333762, 51.6808476, 45.5793315, 39.5404642, 33.5758819, 27.6972207, 21.916117, 16.2442068, 10.6931265, 5.27451212, 0, -2.26642939, -4.6600261, -7.18338767, -9.83911159, -12.6297954, -15.5580366, -18.6264326, -21.8375811, -25.1940795, -28.6985253, -32.3535161, -36.1616494, -40.1255226, -44.2477333, -48.530879, -52.9775572, -57.0837952, -60.7798233, -65.2140527, -70.2805301, -75.8733022, -81.8864159, -88.2139178, -94.7498547, -101.388273, -108.023221, -114.548743, -120.858887, -126.847701, -131.753582, -136.366629, -140.777584, -145.077187, -150.306189, -156.080325, -161.761998, -167.156944, -172.070903, -176.309612, -179.678811, -182.931706, -184.638206, -184.418676, -181.893479, -176.68298, -173.169087, -165.020232, -150.666396, -128.537556, -97.0636905, -50.793947, 3.21762929, 56.7141831, 107.87411, 168.014443, 242.553234, 336.908537, 457.813714, 609.072574, 793.054043, 1015.10772, 1276.28871, 1568.76211, 1895.88186, 2279.96082, 2734.60937, 3274.08299, 3912.15933, 4660.25077, 5531.10759, 6546.49251, 7708.7569, 9027.53141, 10512.4467, 12173.1333, 14573.2567, 17029.904, 19414.1972, 21597.2583, 23450.2091, 24339.0305, 24976.7464, 25450.9676, 26098.2487, 26907.6605, 27764.5475, 28663.1668, 29597.7757, 30562.6312, 31551.9906, 33022.2098, 34303.5459, 35361.1082, 36160.0057, 36665.3477, 36842.2433, 36655.8017, 36513.6104, 36449.3586},
        {360.792502, 362.645892, 364.284319, 365.711451, 366.930956, 367.946504, 368.761763, 369.380401, 369.806087, 370.04249, 370.093278, 369.96212, 369.652685, 369.16864, 368.513655, 367.691399, 366.705539, 365.559744, 364.257684, 362.803026, 361.199439, 359.450592, 357.560154, 355.531792, 353.369176, 351.075974, 348.655855, 346.112487, 343.449539, 340.670679, 337.779577, 334.7799, 331.675318, 328.469498, 325.210449, 321.878737, 318.454396, 314.939654, 311.336743, 307.647891, 303.875329, 300.021285, 296.087991, 292.077674, 287.992566, 283.834896, 279.606893, 275.310787, 270.948809, 266.523187, 262.036151, 257.44363, 252.795159, 248.093971, 243.343299, 238.546377, 233.706437, 228.826715, 223.910442, 218.960853, 213.981181, 208.974658, 203.94452, 198.893998, 193.826327, 188.74474, 183.652471, 178.643534, 173.809768, 168.967173, 164.113096, 159.244888, 154.359895, 149.455466, 144.52895, 139.577694, 134.599048, 129.59036, 124.548977, 119.472249, 114.357523, 109.202148, 104.003472, 98.7588443, 92.9975336, 86.9607732, 80.8954223, 74.8140191, 68.7291022, 62.6532098, 56.5988805, 50.5786526, 44.6050644, 38.6906545, 32.8479612, 27.0895228, 21.4278779, 15.8755647, 10.4451218, 5.14908739, 0, -1.42522145, -3.11027967, -5.04737492, -7.22870745, -9.64647752, -12.2928854, -15.1601313, -18.2404155, -21.5259383, -25.0088998, -28.6815005, -32.5359405, -36.56442, -40.7591394, -45.1122988, -49.6160986, -53.932449, -58.006366, -62.5771343, -67.5661778, -72.8949202, -78.4847854, -84.2571973, -90.1335797, -96.0353563, -101.883951, -107.600788, -113.10729, -118.324883, -122.802721, -126.970995, -130.862808, -134.511264, -138.465766, -142.501269, -146.255048, -149.605315, -152.430286, -154.608175, -156.017195, -158.212727, -158.885377, -157.585219, -153.862325, -147.266768, -140.622952, -129.388037, -112.293513, -88.0708719, -55.4516033, -10.7047396, 41.9660987, 94.6485637, 146.131835, 208.887377, 289.243588, 393.528867, 535.275075, 711.969995, 913.243084, 1127.54014, 1374.81183, 1667.64469, 2015.72326, 2423.81483, 2898.94377, 3442.83401, 4061.13575, 4778.93416, 5610.32089, 6588.99019, 7706.41533, 8969.90247, 10386.7577, 11964.2873, 14202.7403, 16493.2441, 18719.8692, 20766.6858, 22517.7644, 23404.3329, 24065.1991, 24578.5088, 25241.4311, 26045.415, 26890.6496, 27773.147, 28688.919, 29633.9776, 30604.3348, 31440.4484, 32509.8279, 33670.3188, 34779.7667, 35696.0171, 36276.9157, 36380.3078, 36550.0677, 35250.402},
        {339.029877, 341.011679, 342.78719, 344.359732, 345.732627, 346.909198, 347.892767, 348.686656, 349.294188, 349.718685, 349.963469, 350.031863, 349.927189, 349.652769, 349.211926, 348.607982, 347.844259, 346.92408, 345.850767, 344.627642, 343.258028, 341.745248, 340.092622, 338.303474, 336.381127, 334.328901, 332.150121, 329.848107, 327.426183, 324.88767, 322.235892, 319.47417, 316.605827, 313.634185, 310.586385, 307.453339, 304.225673, 300.905936, 297.496678, 294.000447, 290.419794, 286.757268, 283.015418, 279.196793, 275.303944, 271.339418, 267.305766, 263.205538, 259.041281, 254.815547, 250.530883, 246.158877, 241.733712, 237.258609, 232.736788, 228.17147, 223.565875, 218.923225, 214.246739, 209.539639, 204.805145, 200.046477, 195.266857, 190.469504, 185.657639, 180.834484, 176.003258, 171.266925, 166.726053, 162.178151, 157.619975, 153.048277, 148.459814, 143.85134, 139.219609, 134.561377, 129.873398, 125.152427, 120.395217, 115.598525, 110.759105, 105.873711, 100.939099, 95.9520219, 90.4062341, 84.5605706, 78.6799127, 72.7773384, 66.8659258, 60.9587529, 55.0688979, 49.2094387, 43.3934534, 37.6340201, 31.9442168, 26.3371217, 20.8258128, 15.423368, 10.1428656, 4.99738361, 0, -1.05682854, -2.41483982, -4.06239508, -5.98785556, -8.17958248, -10.6259371, -13.3152806, -16.2359743, -19.3763793, -22.7248569, -26.2697684, -29.999475, -33.9023379, -37.9667184, -42.1809776, -46.5334768, -50.7594109, -54.8109746, -59.2491671, -64.0080999, -69.0218842, -74.2246313, -79.5504526, -84.9334592, -90.3077626, -95.6074741, -100.766705, -105.719566, -110.400169, -114.545546, -118.35915, -121.834216, -124.96398, -128.017911, -130.851126, -133.245082, -135.110143, -136.356674, -136.895039, -136.635605, -137.594589, -136.93559, -134.156958, -128.757042, -120.234192, -109.136714, -93.6505132, -73.011449, -46.4553821, -13.2181731, 27.1620079, 76.1210997, 130.020007, 189.312767, 262.810232, 355.571212, 472.654522, 627.474304, 814.8351, 1020.42655, 1223.75052, 1455.98384, 1741.30257, 2093.58319, 2509.25026, 2992.73906, 3543.1994, 4163.69626, 4876.67464, 5693.61709, 6638.07482, 7707.60503, 8909.51268, 10251.1027, 11739.6801, 13836.8937, 15982.1185, 18069.0737, 19991.4781, 21643.0508, 22497.5413, 23148.6177, 23669.986, 24351.2998, 25180.7858, 26056.6924, 26971.7454, 27918.6702, 28890.1925, 29879.0379, 30315.4016, 31227.7763, 32416.9161, 33683.5751, 34828.5074, 35652.4671, 35956.2083, 36262.5984, 34995.6824},
        {320.562039, 322.378136, 324.017819, 325.48331, 326.776836, 327.900622, 328.856891, 329.64787, 330.275783, 330.742855, 331.051311, 331.203377, 331.201276, 331.047234, 330.743476, 330.292227, 329.695712, 328.956156, 328.075783, 327.05682, 325.901489, 324.612018, 323.19063, 321.63955, 319.961004, 318.157217, 316.230413, 314.182817, 312.016654, 309.73415, 307.337529, 304.829017, 302.210837, 299.485216, 296.618757, 293.632245, 290.546893, 287.366082, 284.093192, 280.731604, 277.284699, 273.755858, 270.148463, 266.465893, 262.71153, 258.888754, 255.000947, 251.05149, 247.043763, 242.981148, 238.867024, 234.745083, 230.577522, 226.366848, 222.115568, 217.82619, 213.50122, 209.143166, 204.754535, 200.337833, 195.895568, 191.430247, 186.944378, 182.440466, 177.921019, 173.388545, 168.845549, 164.381254, 160.082797, 155.773846, 151.451286, 147.112004, 142.752884, 138.370813, 133.962676, 129.52536, 125.055749, 120.550731, 116.00719, 111.422013, 106.792085, 102.114292, 97.38552, 92.6026546, 87.2615773, 81.620214, 75.9425123, 70.2416167, 64.5306714, 58.8228208, 53.1312093, 47.4689812, 41.8492808, 36.2852526, 30.7900408, 25.3767899, 20.0586442, 14.8487479, 9.7602456, 4.80628151, 0, -1.25865788, -2.74733074, -4.45867733, -6.38535641, -8.52002674, -10.8553471, -13.3839761, -16.0985727, -18.9917956, -22.0563034, -25.284755, -28.6698092, -32.2041246, -35.8803601, -39.6911743, -43.6292261, -47.3673679, -50.8525718, -54.8008681, -59.1363854, -63.7832526, -68.6655985, -73.7075521, -78.8332421, -83.9667974, -89.0323468, -93.9540192, -98.6559434, -103.062248, -106.889545, -110.34557, -113.416707, -116.08934, -118.673648, -120.994117, -122.78889, -123.947213, -124.358333, -123.911495, -122.495946, -122.238953, -120.110604, -115.56227, -108.045326, -97.0111416, -80.1459977, -59.1076322, -33.7886906, -4.0818185, 30.1203387, 64.9566073, 109.353293, 167.294534, 241.670158, 333.13578, 443.350131, 573.971945, 730.372521, 909.654256, 1104.86947, 1298.41997, 1520.50051, 1793.95468, 2133.05623, 2537.55777, 3013.87736, 3569.15741, 4210.00376, 4940.36632, 5765.69733, 6681.45652, 7704.62599, 8844.11694, 10108.8406, 11507.7081, 13482.4975, 15501.0364, 17464.0196, 19272.1416, 20826.0968, 21622.6815, 22235.7542, 22739.1087, 23442.6505, 24326.7126, 25273.2482, 26267.509, 27294.7471, 28340.2146, 29389.1632, 29763.017, 30625.4363, 31779.0066, 33026.3136, 34169.9429, 35012.4801, 35356.5108, 35593.5176, 35830.0846},
        {302.36832, 304.392341, 306.233512, 307.894358, 309.377403, 310.685174, 311.820196, 312.784993, 313.582091, 314.214016, 314.683292, 314.992445, 315.144, 315.140482, 314.984417, 314.67833, 314.224746, 313.626191, 312.885189, 312.004266, 310.985947, 309.832758, 308.547224, 307.131869, 305.589219, 303.921801, 302.132137, 300.222755, 298.196179, 296.054935, 293.801547, 291.438542, 288.968443, 286.393778, 283.69266, 280.880333, 277.972335, 274.971983, 271.882595, 268.707487, 265.449978, 262.113384, 258.701023, 255.216213, 251.662271, 248.042515, 244.360261, 240.618827, 236.821531, 232.97169, 229.072621, 225.17931, 221.242287, 217.263748, 213.24589, 209.190913, 205.101011, 200.978384, 196.825229, 192.643742, 188.436121, 184.204564, 179.951267, 175.678429, 171.388247, 167.082917, 162.764638, 158.515332, 154.415008, 150.301435, 146.171643, 142.02266, 137.851516, 133.65524, 129.43086, 125.175406, 120.885907, 116.559392, 112.19289, 107.78343, 103.328042, 98.8237532, 94.2675942, 89.6565936, 84.4872112, 79.0172895, 73.5106876, 67.9806789, 62.4405369, 56.9035353, 51.3829474, 45.8920468, 40.444107, 35.0524016, 29.730204, 24.4907878, 19.3474264, 14.3133934, 9.40196237, 4.62640672, 0, -1.103136, -2.43594238, -3.98998886, -5.7568452, -7.72808112, -9.89526638, -12.2499707, -14.7837638, -17.4882155, -20.3548954, -23.3753734, -26.5412191, -29.8440023, -33.2752928, -36.8266602, -40.4896743, -44.0108794, -47.3468415, -51.0433547, -55.0394834, -59.2742917, -63.686844, -68.2162045, -72.8014376, -77.3816074, -81.8957782, -86.2830143, -90.4823801, -94.4329397, -98.0050626, -101.231696, -104.072513, -106.487185, -108.694025, -110.523386, -111.73665, -112.2159, -111.843215, -110.500677, -108.070367, -107.097657, -103.99077, -98.1107094, -88.8184761, -75.4750728, -52.1840902, -24.878295, 5.76695728, 39.0763116, 74.3744126, 103.51234, 142.422661, 200.640423, 283.053538, 385.062022, 506.325764, 646.504651, 806.237369, 983.982553, 1177.13107, 1370.51428, 1592.2116, 1863.99003, 2199.49113, 2598.58827, 3067.47465, 3613.83098, 4244.23609, 4959.81466, 5764.77652, 6653.09953, 7640.7997, 8735.62874, 9945.33835, 11277.6803, 13141.0972, 15042.477, 16889.3988, 18589.4413, 20050.1836, 20794.3224, 21370.907, 21852.466, 22552.496, 23448.1424, 24416.147, 25438.6752, 26497.8923, 27575.964, 28655.0555, 28931.9225, 29772.0578, 30952.4142, 32249.9443, 33441.601, 34304.337, 34615.1051, 34806.9226, 36084.8979},
        {284.82981, 287.349992, 289.651822, 291.739263, 293.616282, 295.286842, 296.754909, 298.024448, 299.099424, 299.983801, 300.681546, 301.196622, 301.532994, 301.694629, 301.685489, 301.509542, 301.170751, 300.673081, 300.020499, 299.216967, 298.266452, 297.172918, 295.940331, 294.572655, 293.073856, 291.447897, 289.698745, 287.830364, 285.84672, 283.751776, 281.549499, 279.243853, 276.838802, 274.338313, 271.793846, 269.184618, 266.489277, 263.710246, 260.849951, 257.910813, 254.895257, 251.805706, 248.644584, 245.414315, 242.117323, 238.75603, 235.332861, 231.850239, 228.310588, 224.716332, 221.069894, 217.376358, 213.63543, 209.849477, 206.020862, 202.151954, 198.245118, 194.302719, 190.327124, 186.320699, 182.28581, 178.224822, 174.140102, 170.034016, 165.908929, 161.767208, 157.611219, 153.525664, 149.59327, 145.646589, 141.682649, 137.698479, 133.691107, 129.657561, 125.59487, 121.500062, 117.370166, 113.20221, 108.993222, 104.740231, 100.440265, 96.0903535, 91.6875237, 87.2288042, 82.2018232, 76.8690238, 71.4989717, 66.1052265, 60.7013475, 55.3008942, 49.917426, 44.5645024, 39.2556827, 34.0045265, 28.824593, 23.7294419, 18.7326324, 13.847724, 9.08827622, 4.4678484, 0, -0.565700319, -1.43582217, -2.59533698, -4.02921618, -5.7224312, -7.65995348, -9.82675444, -12.2078055, -14.7880781, -17.5525438, -20.4861738, -23.5739396, -26.8008128, -30.1517646, -33.6117666, -37.1657901, -40.7449625, -44.326563, -48.0173537, -51.7907678, -55.6202387, -59.4791999, -63.3410848, -67.1793266, -70.9673589, -74.678615, -78.2865283, -81.7645322, -85.0860601, -88.4098803, -91.4561351, -94.1426574, -96.3872801, -98.1992593, -99.4507164, -99.9893929, -99.7056948, -98.4900279, -96.2327982, -92.8244114, -91.4190588, -87.650035, -80.7691792, -70.0283308, -54.6793288, -25.4561947, 7.74195997, 43.5338417, 80.5381568, 117.373612, 142.688939, 177.258342, 232.843359, 315.431144, 419.219481, 543.699341, 688.361695, 853.659247, 1037.90266, 1238.35342, 1440.41069, 1670.46481, 1950.94535, 2294.37438, 2696.48632, 3160.72141, 3687.47836, 4279.40886, 4950.3168, 5707.69775, 6568.69402, 7529.67032, 8594.78296, 9768.18822, 11054.0424, 12814.5294, 14606.271, 16343.9167, 17942.116, 19315.5183, 20018.2436, 20565.8238, 21027.4209, 21699.445, 22561.1059, 23496.9267, 24490.8516, 25526.8247, 26588.7901, 27660.6921, 27780.7101, 28611.6454, 29877.2318, 31301.2034, 32607.294, 33519.2376, 33760.7681, 33987.5563, 35608.8858},
        {271.159829, 273.792753, 276.199922, 278.385782, 280.354777, 282.111353, 283.659955, 285.005028, 286.151017, 287.102367, 287.863524, 288.438933, 288.833038, 289.050285, 289.09512, 288.971986, 288.68533, 288.239596, 287.63923, 286.888677, 285.992382, 284.954789, 283.780345, 282.473494, 281.038682, 279.480353, 277.802953, 276.010927, 274.10872, 272.100777, 269.991544, 267.785465, 265.486985, 263.100551, 260.712741, 258.285207, 255.77861, 253.194731, 250.535348, 247.802243, 244.997195, 242.121982, 239.178386, 236.168185, 233.093159, 229.955089, 226.755753, 223.496932, 220.180405, 216.807951, 213.381351, 209.868231, 206.305264, 202.69497, 199.03987, 195.342483, 191.60533, 187.83093, 184.021805, 180.180473, 176.309455, 172.411271, 168.488442, 164.543486, 160.578925, 156.597278, 152.601066, 148.679629, 144.920224, 141.14631, 137.354778, 133.54252, 129.706428, 125.843394, 121.950309, 118.024065, 114.061555, 110.059669, 106.015301, 101.925341, 97.7866811, 93.5962138, 89.3508307, 85.0474234, 80.1577339, 74.9522748, 69.7078707, 64.4384555, 59.1579632, 53.8803277, 48.619483, 43.389363, 38.2039018, 33.0770332, 28.0226913, 23.0548099, 18.187323, 13.4341646, 8.80926872, 4.32656918, 0, -0.161909691, -0.673867322, -1.51700983, -2.67247413, -4.12139718, -5.84491589, -7.8241672, -10.040288, -12.4744154, -15.1076861, -17.9212371, -20.8962054, -24.0137279, -27.2549415, -30.6009831, -34.0329897, -37.5306372, -41.0748536, -44.6500809, -48.2371429, -51.8168634, -55.3700662, -58.8775752, -62.3202142, -65.678807, -68.9341776, -72.0671496, -75.0585471, -77.8891937, -80.7784084, -83.3810717, -85.6064592, -87.363846, -88.6853536, -89.4188344, -89.3793816, -88.4394168, -86.4713615, -83.3476373, -78.9406655, -76.4182392, -71.3544693, -62.9770309, -50.5135988, -33.191848, -1.34202777, 34.6864048, 73.4414184, 113.470982, 153.323063, 181.836478, 219.135092, 275.738326, 357.277885, 459.406373, 582.256808, 725.962206, 891.656161, 1078.24371, 1283.53835, 1495.70511, 1737.09214, 2028.2605, 2379.8556, 2785.72124, 3247.41346, 3760.69745, 4325.62789, 4963.49251, 5683.56844, 6516.32026, 7446.57802, 8476.28472, 9607.38336, 10841.8169, 12506.1446, 14196.5392, 15833.7897, 17338.6849, 18632.0139, 19296.4722, 19816.3377, 20257.2965, 20896.7989, 21716.7076, 22610.6503, 23564.9014, 24565.7355, 25599.427, 26652.2505, 26722.4366, 27566.725, 28882.9907, 30369.1087, 31722.9538, 32642.4012, 32825.3257, 33174.344, 35042.4127},
        {261.977625, 264.169001, 266.170536, 267.985625, 269.617661, 271.070037, 272.346149, 273.449389, 274.383151, 275.150829, 275.755817, 276.201509, 276.491298, 276.628578, 276.616744, 276.459188, 276.159305, 275.720489, 275.146133, 274.439631, 273.604377, 272.643764, 271.561187, 270.360039, 269.043715, 267.615607, 266.07911, 264.437617, 262.694523, 260.85322, 258.917104, 256.889567, 254.774004, 252.573808, 250.343943, 248.060934, 245.700684, 243.264914, 240.755344, 238.173694, 235.521685, 232.801036, 230.013467, 227.1607, 224.244454, 221.26645, 218.228407, 215.132047, 211.979088, 208.771253, 205.510259, 202.16144, 198.763692, 195.319525, 191.831449, 188.301972, 184.733604, 181.128852, 177.490228, 173.820239, 170.121395, 166.396205, 162.647178, 158.876823, 155.087649, 151.282165, 147.462881, 143.723304, 140.154759, 136.572075, 132.971863, 129.350731, 125.70529, 122.03215, 118.327922, 114.589214, 110.812637, 106.994801, 103.132316, 99.2217917, 95.2598383, 91.2430657, 87.1680838, 83.0315027, 78.2825054, 73.2029296, 68.081193, 62.9316705, 57.7687371, 52.6067679, 47.4601379, 42.3432221, 37.2703956, 32.2560333, 27.3145104, 22.4602018, 17.7074826, 13.0707278, 8.56431237, 4.20261144, 0, -0.0316074413, -0.402696696, -1.09541604, -2.09191375, -3.37433811, -4.92483739, -6.72555988, -8.75865384, -11.0062676, -13.4505493, -16.0736474, -18.85771, -21.7848855, -24.8373222, -27.9971682, -31.246572, -34.4235534, -37.4896709, -40.7530774, -44.1650366, -47.6768121, -51.2396675, -54.8048664, -58.3236725, -61.7473494, -65.0271607, -68.1143701, -70.9602412, -73.5160376, -75.7631197, -77.6116189, -79.0037698, -79.881807, -80.5893009, -80.8678181, -80.3519026, -78.8633883, -76.2241091, -72.2558989, -66.7805915, -62.2181573, -55.0015567, -44.4442923, -29.8598671, -10.561784, 20.0108439, 55.2015266, 94.2281639, 136.308656, 180.660902, 220.003387, 267.997155, 330.882828, 412.510114, 511.847264, 630.054206, 768.290866, 928.139942, 1110.2425, 1314.77841, 1535.29984, 1787.81953, 2087.96457, 2443.98654, 2851.63963, 3312.41455, 3823.09126, 4383.93916, 5012.50037, 5716.54657, 6519.82948, 7411.68232, 8394.34523, 9470.05836, 10641.0619, 12215.7301, 13813.6355, 15360.4846, 16781.9839, 18003.8402, 18632.2926, 19125.4932, 19546.0634, 20156.224, 20939.588, 21796.6017, 22715.0364, 23682.6632, 24687.2534, 25716.5783, 25872.1793, 26745.6933, 28050.3704, 29499.4612, 30806.2159, 31683.8849, 31845.7186, 32405.7134, 34568.1633},
        {254.546663, 256.181573, 257.670736, 259.016201, 260.220018, 261.284235, 262.210902, 263.002067, 263.659781, 264.186092, 264.583049, 264.852701, 264.997098, 265.018289, 264.918323, 264.699249, 264.363116, 263.911973, 263.34787, 262.672856, 261.88898, 260.99829, 260.002837, 258.904668, 257.705835, 256.408385, 255.014367, 253.525832, 251.944827, 250.273403, 248.513608, 246.667492, 244.737103, 242.724491, 240.639151, 238.479252, 236.242875, 233.931827, 231.547916, 229.092948, 226.568732, 223.977076, 221.319786, 218.59867, 215.815536, 212.972192, 210.070444, 207.1121, 204.098969, 201.032856, 197.915571, 194.719704, 191.476912, 188.189637, 184.860318, 181.491398, 178.085316, 174.644514, 171.171432, 167.668511, 164.138193, 160.582918, 157.005126, 153.40726, 149.791759, 146.161065, 142.517618, 138.958756, 135.579538, 132.186688, 128.776495, 125.345247, 121.889234, 118.404746, 114.88807, 111.335497, 107.743315, 104.107813, 100.425281, 96.6920081, 92.9042824, 89.0583935, 85.1506306, 81.1772826, 76.5638013, 71.6039011, 66.5981579, 61.5613855, 56.5083978, 51.4540086, 46.4130319, 41.4002815, 36.4305712, 31.5187149, 26.6795266, 21.9278201, 17.2784093, 12.746108, 8.34573006, 4.09208945, 0, 0.047753574, -0.226868078, -0.807500993, -1.6777812, -2.82134475, -4.22182766, -5.86286597, -7.72809572, -9.80115295, -12.0656737, -14.505294, -17.1036498, -19.8443773, -22.7111124, -25.6874912, -28.7571497, -31.6106484, -34.1897554, -37.1410157, -40.3852635, -43.8433329, -47.4360581, -51.0842731, -54.7088121, -58.2305092, -61.5701985, -64.6487142, -67.3868903, -69.7055611, -71.3523578, -72.4848251, -73.0757578, -73.097951, -73.1870857, -72.9845137, -71.9548174, -69.8719259, -66.5097685, -61.6422741, -55.043372, -48.2659648, -38.7635814, -25.9620909, -9.2873625, 11.8347347, 40.2113274, 73.6253017, 112.09254, 155.628923, 204.250335, 255.049333, 314.538073, 384.47205, 466.792776, 563.821545, 677.709136, 810.606329, 964.556589, 1141.8424, 1344.86331, 1572.30762, 1834.11367, 2140.61305, 2497.80338, 2904.3386, 3362.2435, 3870.84993, 4431.48451, 5055.34811, 5748.05616, 6522.13931, 7375.2339, 8310.56166, 9331.34429, 10440.8035, 11936.2009, 13453.2084, 14921.5376, 16270.9003, 17431.0082, 18028.3506, 18498.0099, 18899.6504, 19480.5214, 20226.0346, 21043.4413, 21922.0611, 22851.2137, 23820.2188, 24818.3962, 25082.4407, 25979.5936, 27250.1902, 28634.5659, 29873.056, 30705.9958, 30873.7208, 31675.3242, 34054.5795},
        {245.454061, 246.896676, 248.208917, 249.392463, 250.448989, 251.380172, 252.18769, 252.873218, 253.438434, 253.885014, 254.214635, 254.428973, 254.529705, 254.518509, 254.397059, 254.167034, 253.83011, 253.387964, 252.842272, 252.194711, 251.446958, 250.600689, 249.657581, 248.619311, 247.487555, 246.26399, 244.950294, 243.548142, 242.059211, 240.485178, 238.827719, 237.088512, 235.269233, 233.371558, 231.395327, 229.343172, 227.217752, 225.020802, 222.75406, 220.41926, 218.01814, 215.552436, 213.023885, 210.434222, 207.785184, 205.078507, 202.315928, 199.499183, 196.630009, 193.710142, 190.741317, 187.698711, 184.611197, 181.481087, 178.310693, 175.102327, 171.858301, 168.580927, 165.272518, 161.935385, 158.571842, 155.184199, 151.774769, 148.345864, 144.899797, 141.438879, 137.965423, 134.578089, 131.373226, 128.154434, 124.917777, 121.659323, 118.375138, 115.061288, 111.71384, 108.328859, 104.902411, 101.430564, 97.909383, 94.3349344, 90.7032844, 87.0104993, 83.2526453, 79.4257888, 74.9389844, 70.0941511, 65.2002633, 60.2724368, 55.3257876, 50.3754315, 45.4364843, 40.524062, 35.6532805, 30.8392555, 26.0971031, 21.441939, 16.8888792, 12.4530395, 8.14953584, 3.99348404, 0, 0.29762241, 0.240935189, -0.150447554, -0.856911713, -1.85884318, -3.13662784, -4.6706516, -6.44130033, -8.42895994, -10.6140163, -12.9768554, -15.4978629, -18.157425, -20.9359273, -23.8137559, -26.7712966, -29.6162352, -32.3042862, -35.2064666, -38.2661551, -41.4267306, -44.6315716, -47.824057, -50.9475655, -53.9454757, -56.7611665, -59.3380165, -61.6194045, -63.5487092, -65.124572, -66.214846, -66.746331, -66.645827, -66.3115854, -65.4346803, -63.5804656, -60.5343055, -56.0815647, -50.0076076, -42.0977985, -33.7532939, -22.6519035, -8.2628588, 9.94460909, 32.5012688, 61.6358441, 95.75666, 134.969996, 179.382133, 229.099351, 281.335707, 342.624641, 415.138833, 500.928072, 601.791245, 719.639887, 856.385538, 1013.64367, 1193.68918, 1399.1199, 1629.48137, 1892.7702, 2196.82972, 2546.73706, 2942.88212, 3387.80633, 3882.23293, 4428.23198, 5034.54014, 5706.12304, 6452.46105, 7273.31052, 8172.07431, 9152.15528, 10216.9563, 11652.6924, 13109.2513, 14519.3326, 15815.6361, 16930.8618, 17508.2897, 17962.9857, 18351.6864, 18906.1696, 19613.885, 20389.3484, 21223.6562, 22107.9046, 23033.1899, 23990.6085, 24226.8433, 25094.4282, 26339.6418, 27708.7632, 28948.0713, 29803.8447, 30022.3626, 31021.0588, 33460.7348},
        {236.525754, 237.889014, 239.127872, 240.243946, 241.238852, 242.114208, 242.871632, 243.512739, 244.039148, 244.452476, 244.75434, 244.946358, 245.030146, 245.007322, 244.879504, 244.648307, 244.315351, 243.882251, 243.350626, 242.722092, 241.998268, 241.180769, 240.271213, 239.271218, 238.182401, 237.006379, 235.744769, 234.399189, 232.971256, 231.462587, 229.874799, 228.209511, 226.468338, 224.652898, 222.764176, 220.804118, 218.774679, 216.677498, 214.514211, 212.286457, 209.995874, 207.644099, 205.23277, 202.763524, 200.238001, 197.657837, 195.02467, 192.340139, 189.60588, 186.823532, 183.994733, 181.097402, 178.157409, 175.176906, 172.158045, 169.102979, 166.013859, 162.892837, 159.742066, 156.563696, 153.359881, 150.132772, 146.884521, 143.61728, 140.333202, 137.034437, 133.723139, 130.498241, 127.456355, 124.400025, 121.32513, 118.227548, 115.103156, 111.947833, 108.757455, 105.527903, 102.255052, 98.934782, 95.5629699, 92.135494, 88.6482322, 85.0970624, 81.4778627, 77.7865108, 73.4182409, 68.6818865, 63.8935007, 59.0684534, 54.2221147, 49.3698546, 44.5270431, 39.7090501, 34.9312457, 30.2089999, 25.5576827, 20.992664, 16.529314, 12.1830026, 7.96909978, 3.90297558, 0, 0.571658978, 0.746766781, 0.549284214, 0.00317208194, -0.867608809, -2.03909765, -3.48733365, -5.18835598, -7.11820385, -9.25291646, -11.568533, -14.0410926, -16.6466346, -19.3611981, -22.1608223, -25.0215463, -27.9295075, -30.8621875, -33.7727848, -36.6395026, -39.4405438, -42.1541115, -44.7584088, -47.2316387, -49.5520042, -51.6977084, -53.6469543, -55.3779451, -56.8688837, -58.4780274, -59.6641717, -60.2915036, -60.2242097, -59.4964365, -57.8873908, -55.1309572, -51.0403344, -45.4287214, -38.109317, -28.8953202, -19.2453723, -6.82644286, 8.87020365, 28.3533028, 52.13159, 82.814653, 118.285162, 158.52664, 203.522609, 253.256591, 303.780603, 363.814848, 437.423416, 527.373006, 633.781484, 757.955988, 901.203656, 1064.60403, 1249.74358, 1458.45702, 1689.29705, 1950.25353, 2248.10375, 2587.67207, 2970.47385, 3399.54331, 3876.63714, 4404.45837, 4990.39439, 5639.18286, 6359.39681, 7151.34845, 8018.44778, 8964.10481, 9991.72955, 11378.1425, 12784.9906, 14147.3311, 15400.2216, 16478.7194, 17040.3419, 17482.7131, 17859.836, 18388.795, 19059.0402, 19792.0708, 20580.7345, 21417.8788, 22296.3512, 23208.9992, 23391.2286, 24213.3887, 25426.9879, 26783.5346, 28034.537, 28931.5035, 29225.9424, 30408.5193, 32825.1774},
        {228.252986, 229.584667, 230.793508, 231.881211, 232.849477, 233.700009, 234.434509, 235.054678, 235.562219, 235.958834, 236.246225, 236.426094, 236.500144, 236.470075, 236.337591, 236.104393, 235.772183, 235.342664, 234.817537, 234.198505, 233.487269, 232.685533, 231.794997, 230.817364, 229.754335, 228.607614, 227.378902, 226.069901, 224.682313, 223.217841, 221.678186, 220.06505, 218.380135, 216.625145, 214.807026, 212.924748, 210.977216, 208.965962, 206.892517, 204.758414, 202.565184, 200.314359, 198.007471, 195.646051, 193.231632, 190.765745, 188.249922, 185.685694, 183.074595, 180.418155, 177.717906, 174.954649, 172.151096, 169.309229, 166.431028, 163.518476, 160.573552, 157.598239, 154.594518, 151.564369, 148.509774, 145.432714, 142.33517, 139.219124, 136.086557, 132.93945, 129.779784, 126.706251, 123.815218, 120.909211, 117.983942, 115.035123, 112.058465, 109.04968, 106.004481, 102.918579, 99.7876864, 96.6075149, 93.3737763, 90.0821827, 86.728446, 83.308278, 79.8173907, 76.251496, 71.9932038, 67.3576728, 62.6674307, 57.9380858, 53.1852463, 48.4245204, 43.6715165, 38.9418428, 34.2511075, 29.614919, 25.0488854, 20.568615, 16.189716, 11.9277968, 7.79846555, 3.81733053, 0, 0.81722445, 1.20005428, 1.17650835, 0.774605507, 0.0223646175, -1.05219546, -2.42105588, -4.05619778, -5.92960229, -8.01325057, -10.2791238, -12.699203, -15.2454694, -17.8899042, -20.6044884, -23.3612033, -26.3033752, -29.4274633, -32.3178819, -34.9833289, -37.4325024, -39.6741004, -41.7168208, -43.5693617, -45.2404211, -46.7386969, -48.0728871, -49.2516899, -50.283803, -51.8378183, -53.020579, -53.642815, -53.5152563, -52.3515126, -50.0108737, -46.3785286, -41.2943432, -34.5981835, -26.1299155, -15.7294051, -4.95747825, 8.59072936, 25.4120613, 46.003361, 70.8614721, 103.257896, 140.221153, 181.554424, 227.060887, 276.543722, 324.612594, 382.611824, 456.717944, 550.51992, 662.323779, 792.807479, 942.648979, 1112.63642, 1303.31254, 1515.09987, 1744.49196, 2000.75555, 2290.79918, 2619.6703, 2989.26278, 3402.91801, 3862.9789, 4372.52798, 4938.30897, 5564.99459, 6261.13165, 7026.92282, 7865.69981, 8780.79435, 9775.53815, 11118.4059, 12481.3005, 13801.2682, 15015.3547, 16060.6062, 16606.8562, 17036.5047, 17401.1173, 17904.6226, 18538.2759, 19229.8476, 19973.767, 20764.4634, 21596.3663, 22463.9048, 22607.371, 23383.9523, 24556.526, 25887.9697, 27141.1605, 28078.9761, 28464.2938, 29821.7223, 32166.208},
        {221.842616, 223.052753, 224.147412, 225.128185, 225.996662, 226.754433, 227.403087, 227.944216, 228.37941, 228.710259, 228.938352, 229.065281, 229.092636, 229.022007, 228.854984, 228.593157, 228.238117, 227.791454, 227.254759, 226.62962, 225.91763, 225.120377, 224.239453, 223.276448, 222.232951, 221.110553, 219.910845, 218.635416, 217.285857, 215.863758, 214.370709, 212.808301, 211.178124, 209.481768, 207.725446, 205.90834, 204.029575, 202.090593, 200.092833, 198.037736, 195.926742, 193.761291, 191.542823, 189.272779, 186.952599, 184.583722, 182.16759, 179.705643, 177.199321, 174.650063, 172.059311, 169.409304, 166.721098, 163.996552, 161.23752, 158.44586, 155.623428, 152.77208, 149.893673, 146.990063, 144.063108, 141.114663, 138.146584, 135.160729, 132.158954, 129.143115, 126.115068, 123.173811, 120.416005, 117.643165, 114.850851, 112.034621, 109.190035, 106.312652, 103.398033, 100.441735, 97.4393185, 94.3863428, 91.2783671, 88.1109507, 84.879653, 81.5800334, 78.2076511, 74.7580655, 70.6000068, 66.0556333, 61.4546376, 56.8129208, 52.146384, 47.4709283, 42.8024549, 38.156865, 33.5500597, 28.9979401, 24.5164074, 20.1213627, 15.8287072, 11.6543421, 7.61416838, 3.72408731, 0, 0.800786563, 1.19221611, 1.20097552, 0.853751658, 0.177231405, -0.801898366, -2.05695078, -3.56123897, -5.28807605, -7.21077515, -9.3026494, -11.5370119, -13.8871758, -16.3264543, -18.8281604, -21.3656073, -23.9286862, -26.4930788, -28.9945996, -31.4101145, -33.7164889, -35.8905884, -37.9092787, -39.7494253, -41.3878939, -42.80155, -43.9672593, -44.8618873, -45.4622996, -46.1362935, -46.3264613, -45.8923892, -44.693663, -42.7788991, -39.9131684, -35.811134, -30.275673, -23.1096625, -14.1159797, -3.09750155, 8.78881802, 23.506289, 41.5169625, 63.2828893, 89.2661206, 121.897576, 159.17822, 201.077887, 247.566411, 298.613626, 349.781326, 410.834988, 486.748087, 580.658588, 691.956963, 821.716231, 971.009414, 1142.20414, 1334.78459, 1546.82264, 1772.31472, 2023.05931, 2308.40736, 2634.20063, 3000.33471, 3409.4346, 3862.91018, 4363.0714, 4916.68355, 5527.99174, 6204.19694, 6946.13362, 7757.02368, 8640.08905, 9598.55163, 10890.4194, 12200.4317, 13468.1139, 14632.9914, 15634.5895, 16155.5342, 16563.5167, 16908.1621, 17383.2616, 17981.3746, 18634.9916, 19339.6757, 20090.9898, 20884.4968, 21715.7599, 21945.7398, 22719.5884, 23835.1831, 25090.4013, 26283.1205, 27211.2179, 27672.5709, 29221.2304, 31528.5291},
        {216.397646, 217.484043, 218.461901, 219.332709, 220.097953, 220.759121, 221.3177, 221.775178, 222.133042, 222.392779, 222.555877, 222.623823, 222.598104, 222.480208, 222.271623, 221.973834, 221.588331, 221.1166, 220.560129, 219.920405, 219.198915, 218.397147, 217.516588, 216.558725, 215.525047, 214.417039, 213.23619, 211.983987, 210.661917, 209.271468, 207.814127, 206.291381, 204.704719, 203.055626, 201.349716, 199.586327, 197.764747, 195.88633, 193.952429, 191.964397, 189.923588, 187.831356, 185.689054, 183.498035, 181.259653, 178.975261, 176.646214, 174.273863, 171.859564, 169.404669, 166.910531, 164.360529, 161.774382, 159.153832, 156.500623, 153.816498, 151.1032, 148.362472, 145.596058, 142.8057, 139.993142, 137.160127, 134.308398, 131.439698, 128.55577, 125.658358, 122.749205, 119.927674, 117.290788, 114.638951, 111.96758, 109.272089, 106.547892, 103.790405, 100.995042, 98.1572175, 95.2723471, 92.3358453, 89.3431267, 86.2896063, 83.1706987, 79.9818187, 76.7183812, 73.3758008, 69.3087059, 64.8463826, 60.3258187, 55.7632237, 51.1748073, 46.5767788, 41.9853479, 37.4167241, 32.8871169, 28.4127358, 24.0097903, 19.69449, 15.4830445, 11.3916632, 7.43655565, 3.63393141, 0, 0.762250868, 1.14664653, 1.17785587, 0.88054779, 0.279391178, -0.600945073, -1.73579207, -3.10048092, -4.67034274, -6.42070862, -8.32690969, -10.364277, -12.5081418, -14.733835, -17.0166879, -19.3320314, -21.4607578, -23.3504211, -25.416155, -27.591625, -29.8104968, -32.0064359, -34.113108, -36.0641787, -37.7933135, -39.2341782, -40.3204382, -40.9857593, -41.163807, -40.8080157, -39.8250336, -38.1425958, -35.6884372, -32.9566786, -29.591862, -25.0874933, -19.2013916, -11.6913761, -2.31526607, 9.16911939, 22.0207812, 37.7643087, 56.8342439, 79.6651289, 106.691506, 139.341813, 176.808222, 219.276802, 266.933618, 319.964741, 375.315024, 440.373231, 518.271427, 611.443731, 720.899294, 848.28705, 995.25593, 1165.97176, 1358.99456, 1570.13865, 1791.08741, 2036.21521, 2317.68083, 2642.18384, 3007.17346, 3414.34499, 3863.82064, 4356.88788, 4900.60202, 5498.75578, 6157.1662, 6877.29018, 7662.21966, 8515.04661, 9438.86298, 10681.7938, 11940.6397, 13157.2343, 14273.4113, 15231.0044, 15723.8401, 16107.0973, 16428.8987, 16875.4748, 17440.2804, 18059.2251, 18728.6409, 19444.8598, 20204.2137, 21003.0347, 21367.3813, 22151.1006, 23198.4194, 24353.5646, 25460.7633, 26364.2423, 26908.2287, 28630.4919, 30886.169},
        {211.768591, 212.736698, 213.602393, 214.367083, 215.032172, 215.599064, 216.069164, 216.443878, 216.724609, 216.912764, 217.009745, 217.016959, 216.93581, 216.767703, 216.514043, 216.176233, 215.75568, 215.253788, 214.671962, 214.011606, 213.274126, 212.460925, 211.573409, 210.612983, 209.581052, 208.479019, 207.308291, 206.070271, 204.766365, 203.397977, 201.966512, 200.473375, 198.919971, 197.307704, 195.642038, 193.922262, 192.147619, 190.319381, 188.438822, 186.507214, 184.525831, 182.495945, 180.41883, 178.295759, 176.128005, 173.91684, 171.663539, 169.369373, 167.035617, 164.663543, 162.254424, 159.792651, 157.296745, 154.768346, 152.209092, 149.620622, 147.004577, 144.362594, 141.696313, 139.007373, 136.297412, 133.568071, 130.820987, 128.057801, 125.280151, 122.489676, 119.688016, 116.975032, 114.448231, 111.906671, 109.345624, 106.760362, 104.146156, 101.498278, 98.8119996, 96.0825925, 93.3053281, 90.4754782, 87.5883142, 84.6391079, 81.6231308, 78.5356544, 75.3719506, 72.1272907, 68.141774, 63.7520389, 59.3025987, 54.8099866, 50.2907358, 45.7613795, 41.238451, 36.7384835, 32.2780101, 27.8735642, 23.5416788, 19.2988873, 15.1617229, 11.1467187, 7.27040802, 3.54932405, 0, 0.773307878, 1.19113423, 1.27666409, 1.05308248, 0.543574419, -0.228675057, -1.24048093, -2.46865816, -3.89002173, -5.48138662, -7.2195678, -9.08138023, -11.0436389, -13.0831588, -15.1767548, -17.3012421, -19.0345732, -20.296583, -21.9662747, -23.9349927, -26.0940815, -28.3348857, -30.5487498, -32.6270182, -34.4610356, -35.9421464, -36.961695, -37.4110261, -37.1814841, -35.819719, -33.6881578, -30.7815535, -27.0946588, -23.5436485, -19.6625644, -14.7397363, -8.49349037, -0.642153064, 9.09594926, 21.0024902, 34.6417331, 51.2311042, 71.1926401, 94.9483772, 122.920352, 155.603195, 193.3282, 236.499257, 285.520251, 340.795073, 400.767049, 470.196869, 550.206232, 642.440596, 749.614767, 873.963439, 1017.7213, 1186.58378, 1378.5383, 1587.79697, 1804.53848, 2044.96644, 2323.55425, 2647.62936, 3012.41081, 3418.6756, 3865.27355, 4352.48202, 4887.64471, 5474.10822, 6116.31182, 6816.34129, 7577.16496, 8401.75117, 9293.06826, 10490.5594, 11701.5994, 12870.0379, 13939.7245, 14854.5088, 15317.3449, 15673.575, 15970.2898, 16388.9786, 16923.5327, 17511.8433, 18150.6017, 18836.4987, 19566.2254, 20336.4728, 20872.2206, 21674.1947, 22642.6181, 23677.714, 24679.7054, 25548.8157, 26185.2678, 28050.6117, 30238.2671},
        {207.691201, 208.569457, 209.3491, 210.031523, 210.618116, 211.110271, 211.509378, 211.816828, 212.034014, 212.162327, 212.203156, 212.157895, 212.027934, 211.814663, 211.519475, 211.143761, 210.688911, 210.156318, 209.547372, 208.863464, 208.105986, 207.276329, 206.375884, 205.406043, 204.368196, 203.263735, 202.094051, 200.860535, 199.564579, 198.207573, 196.790909, 195.315979, 193.784173, 192.196882, 190.561221, 188.8756, 187.138358, 185.350701, 183.513835, 181.628966, 179.697297, 177.720036, 175.698388, 173.633558, 171.526751, 169.379173, 167.19203, 164.966527, 162.703869, 160.405263, 158.071913, 155.690282, 153.276638, 150.832506, 148.359412, 145.85888, 143.332435, 140.781603, 138.207909, 135.612877, 132.998034, 130.364903, 127.71501, 125.049881, 122.371039, 119.680011, 116.978321, 114.366344, 111.942078, 109.503182, 107.044773, 104.561968, 102.049884, 99.5036386, 96.9183487, 94.2891315, 91.6111041, 88.8793838, 86.0890875, 83.2353326, 80.3132362, 77.3179153, 74.2444872, 71.0880691, 67.1737359, 62.8457193, 58.4563044, 54.0223524, 49.5607246, 45.0882825, 40.6218872, 36.1784, 31.7746824, 27.4275956, 23.1540008, 18.9707595, 14.8947329, 10.9427824, 7.13176917, 3.47855461, 0, 1.2293059, 2.03343422, 2.44090663, 2.4802448, 2.17997041, 1.56860511, 0.674670587, -0.473311502, -1.84681949, -3.41733169, -5.15632646, -7.03528211, -9.02567697, -11.0989894, -13.2266977, -15.3802802, -17.1878026, -18.5716843, -20.2801703, -22.2111505, -24.2625148, -26.3321532, -28.3179554, -30.1178115, -31.6296114, -32.7512449, -33.380602, -33.4155726, -32.7540466, -31.0034739, -28.458679, -25.1046839, -20.9265105, -16.7084289, -12.0358368, -6.28099923, 0.810836557, 9.49442314, 20.0245131, 32.6558591, 46.5114835, 63.3223088, 83.564513, 107.714274, 136.247771, 169.614869, 208.324636, 252.859829, 303.703203, 361.337513, 424.994203, 497.936725, 580.370752, 673.489772, 780.504065, 903.718415, 1045.4376, 1210.76929, 1398.57837, 1604.67202, 1821.59251, 2062.95443, 2341.34961, 2663.17126, 3024.30938, 3425.47513, 3865.66089, 4345.13223, 4870.45675, 5444.63724, 6070.9943, 6752.16406, 7491.03931, 8290.51285, 9153.47749, 10313.4922, 11485.1171, 12613.5784, 13644.1023, 14521.9151, 14955.2555, 15284.3289, 15555.9279, 15952.0537, 16465.9786, 17034.6374, 17654.1191, 18320.5127, 19029.907, 19778.3911, 20558.1421, 21389.0702, 22249.3108, 23116.999, 23970.2701, 24787.2593, 25546.1019, 27475.9822, 29589.067},
        {204.04205, 204.863056, 205.58679, 206.214701, 206.748239, 207.188854, 207.537996, 207.797113, 207.967656, 208.051074, 208.048817, 207.962334, 207.793075, 207.54249, 207.212027, 206.803138, 206.317271, 205.755876, 205.120402, 204.4123, 203.633019, 202.784008, 201.866717, 200.882595, 199.833093, 198.71966, 197.543745, 196.306798, 195.010269, 193.655607, 192.244262, 190.777683, 189.25732, 187.684623, 186.070238, 184.410822, 182.702923, 180.947693, 179.146281, 177.299841, 175.409522, 173.476477, 171.501855, 169.486809, 167.43249, 165.340048, 163.210635, 161.045402, 158.8455, 156.612081, 154.346295, 152.037683, 149.699258, 147.332423, 144.938582, 142.519138, 140.075492, 137.609049, 135.121211, 132.613381, 130.086961, 127.543356, 124.983967, 122.410197, 119.82345, 117.225128, 114.616634, 112.098973, 109.770756, 107.427967, 105.065552, 102.678458, 100.261629, 97.8100137, 95.3185567, 92.7822046, 90.1959035, 87.5545995, 84.8532389, 82.0867678, 79.2501324, 76.3382788, 73.3461531, 70.2687017, 66.4152442, 62.1379613, 57.7972742, 53.4103789, 48.9944715, 44.566748, 40.1444045, 35.744637, 31.3846417, 27.0816144, 22.8527514, 18.7152486, 14.6863021, 10.7831079, 7.02286214, 3.42276082, 0, 2.12843815, 3.66963451, 4.66433755, 5.15329573, 5.17725753, 4.77697142, 3.99318585, 2.8666493, 1.43811024, -0.251682862, -2.16198154, -4.25203734, -6.48110177, -8.80842637, -11.1932627, -13.5948622, -15.9639858, -18.2586718, -20.4573773, -22.5175343, -24.3965749, -26.0519312, -27.4410352, -28.521319, -29.2502147, -29.5851544, -29.4835701, -28.902894, -27.800558, -26.3565944, -24.2242151, -21.2940722, -17.4568177, -12.7669616, -7.033227, -0.0206415301, 8.42930016, 18.4751034, 30.2752735, 43.9883159, 57.543235, 74.0065807, 93.9730651, 118.0374, 146.794298, 181.634544, 222.157759, 268.759635, 321.835867, 381.782147, 447.708023, 522.867292, 607.933051, 704.093979, 813.591391, 938.193988, 1079.67047, 1240.24527, 1421.12778, 1623.03028, 1844.798, 2092.90173, 2373.64279, 2690.86046, 3044.22231, 3435.31077, 3864.8119, 4334.07574, 4847.73916, 5408.57983, 6019.0556, 6682.31313, 7401.2408, 8178.72698, 9017.66003, 10148.9075, 11290.3838, 12387.9825, 13387.5971, 14235.1212, 14640.0465, 14942.2697, 15188.9995, 15567.7466, 16070.1835, 16629.5234, 17240.3618, 17897.2941, 18594.9156, 19327.8219, 20402.8256, 21267.7042, 21995.545, 22659.4356, 23332.4632, 24087.7154, 24998.2795, 26909.2325, 28936.5489},
        {200.87636, 201.652992, 202.332864, 202.917503, 203.408432, 203.807176, 204.115261, 204.33421, 204.465549, 204.510803, 204.471496, 204.349154, 204.1453, 203.861459, 203.499157, 203.059919, 202.545268, 201.95673, 201.295829, 200.564091, 199.76304, 198.8942, 197.959097, 196.959256, 195.896201, 194.771457, 193.586548, 192.343, 191.042337, 189.686085, 188.275767, 186.812909, 185.299035, 183.73567, 182.137338, 180.498791, 178.814624, 177.08594, 175.313843, 173.499435, 171.643819, 169.748098, 167.813376, 165.840754, 163.831337, 161.786226, 159.706525, 157.593337, 155.447765, 153.270912, 151.06388, 148.818663, 146.545672, 144.246206, 141.921566, 139.573052, 137.201965, 134.809604, 132.397272, 129.966267, 127.51789, 125.053442, 122.574223, 120.081533, 117.576674, 115.060945, 112.535646, 110.102928, 107.862517, 105.607721, 103.333303, 101.034025, 98.7046518, 96.3399444, 93.9346663, 91.48358, 88.9814486, 86.4230347, 83.8031011, 81.1164107, 78.3577262, 75.5218105, 72.6034264, 69.5973366, 65.7959239, 61.5599147, 57.2584776, 52.909169, 48.5295452, 44.1371626, 39.7495776, 35.3843466, 31.0590259, 26.791172, 22.5983413, 18.49809, 14.5079747, 10.6455516, 6.92837722, 3.37400789, 0, 2.96507057, 5.19300895, 6.73586372, 7.64568342, 7.97451664, 7.77441194, 7.09741787, 5.995583, 4.5209559, 2.72558513, 0.661519264, -1.61919315, -4.06450353, -6.62236333, -9.24072397, -11.8675369, -14.7798134, -17.9725134, -20.6492864, -22.8285966, -24.5289082, -25.7686856, -26.566393, -26.9404947, -26.9094548, -26.4917377, -25.7058077, -24.5701289, -23.1031657, -22.0860214, -20.4948869, -18.1194345, -14.7493368, -9.68489807, -2.96047486, 5.25907792, 15.0372773, 26.4376402, 39.5236836, 54.3589247, 67.7179237, 83.9541411, 103.774585, 127.886265, 156.996188, 193.600257, 236.169364, 284.963293, 340.241829, 402.264759, 469.704355, 546.348206, 633.773555, 733.32732, 845.886169, 972.537904, 1114.37033, 1270.37033, 1444.2041, 1641.82973, 1868.61483, 2123.67933, 2406.84004, 2719.32216, 3064.73731, 3445.60172, 3864.37903, 4323.57183, 4825.87545, 5373.87619, 5969.09714, 6615.35138, 7315.59325, 8072.77708, 8889.85722, 9997.0733, 11113.273, 12184.5894, 13157.1555, 13977.1043, 14352.9662, 14626.8786, 14846.8041, 15207.092, 15697.7604, 16247.8327, 16850.3766, 17498.4599, 18185.1503, 18903.5153, 20231.4611, 21119.4117, 21725.821, 22209.1427, 22727.8306, 23440.3385, 24505.1203, 26363.4399, 28270.7474},
        {198.137604, 198.877953, 199.521759, 200.070621, 200.526135, 200.889901, 201.163515, 201.348576, 201.446682, 201.45943, 201.388419, 201.235246, 201.00151, 200.688807, 200.298737, 199.832897, 199.292885, 198.680298, 197.996736, 197.243794, 196.423073, 195.536168, 194.58468, 193.570204, 192.494339, 191.358684, 190.164835, 188.914391, 187.60895, 186.25011, 184.839468, 183.378622, 181.869171, 180.312712, 178.727469, 177.106377, 175.442171, 173.735911, 171.988654, 170.201459, 168.375384, 166.511488, 164.610829, 162.674465, 160.703455, 158.698857, 156.661729, 154.593131, 152.49412, 150.365754, 148.209092, 146.017316, 143.799532, 141.556968, 139.290854, 137.002419, 134.692892, 132.363502, 130.015478, 127.65005, 125.268447, 122.871897, 120.46163, 118.038874, 115.60486, 113.160816, 110.707972, 108.35035, 106.189508, 104.014668, 101.820395, 99.6012544, 97.3518113, 95.0666313, 92.7402798, 90.367322, 87.9423233, 85.4598491, 82.9144647, 80.3007354, 77.6132267, 74.8465039, 71.9951322, 69.0536772, 65.2960856, 61.0923892, 56.8212813, 52.5007125, 48.1486337, 43.7829957, 39.4217493, 35.0828453, 30.7842345, 26.5438677, 22.3796958, 18.3096694, 14.3517394, 10.5238567, 6.84397205, 3.33003621, 0, 3.54873086, 6.26157356, 8.1980665, 9.41774805, 9.9801566, 9.94483055, 9.37130829, 8.31912819, 6.84782865, 5.01694806, 2.8860248, 0.514597261, -2.03779617, -4.7116171, -7.44732714, -10.1853879, -13.4440924, -17.2464497, -20.2259216, -22.4467908, -23.97334, -24.8698518, -25.2006088, -25.0298937, -24.421989, -23.4411774, -22.1517415, -20.617964, -18.9041275, -18.2675668, -17.1420598, -15.2339737, -12.2496752, -6.89807942, 0.61572065, 9.82064493, 20.7110909, 33.281456, 47.5261376, 63.4395333, 76.974116, 93.3963587, 113.49147, 138.044658, 167.841133, 206.39894, 251.087242, 302.008036, 359.263323, 422.955102, 491.100769, 568.434774, 657.701547, 760.53201, 876.283675, 1005.33477, 1148.06351, 1300.48046, 1468.32417, 1662.09794, 1892.56258, 2152.82285, 2437.4623, 2745.5798, 3083.92518, 3455.73633, 3864.87423, 4314.73832, 4806.44347, 5342.39682, 5922.98993, 6552.95563, 7235.39879, 7973.42424, 8770.13686, 9857.03612, 10951.7336, 12000.2355, 12948.5479, 13742.6772, 14088.5409, 14332.2929, 14522.8344, 14861.1492, 15335.9394, 15872.5391, 16462.8188, 17098.6487, 17771.8993, 18474.4409, 19964.6264, 20855.2587, 21367.5518, 21722.7196, 22141.9761, 22846.5352, 24057.6108, 25845.1459, 27586.6672},
        {195.736755, 196.443977, 197.055282, 197.572314, 197.996714, 198.330125, 198.57419, 198.730551, 198.800851, 198.786732, 198.689838, 198.51181, 198.254292, 197.918925, 197.507352, 197.021217, 196.462161, 195.831828, 195.131859, 194.363897, 193.529586, 192.630566, 191.668482, 190.644976, 189.56169, 188.420267, 187.222349, 185.969579, 184.6636, 183.306054, 181.898583, 180.442831, 178.94044, 177.393053, 175.821875, 174.218357, 172.573712, 170.888947, 169.165071, 167.403092, 165.604016, 163.768852, 161.898607, 159.99429, 158.056907, 156.087467, 154.086977, 152.056444, 149.996878, 147.909284, 145.794672, 143.644086, 141.468713, 139.269775, 137.048497, 134.806103, 132.543815, 130.262859, 127.964456, 125.649832, 123.320209, 120.976811, 118.620863, 116.253587, 113.876207, 111.489947, 109.096031, 106.80192, 104.712527, 102.609965, 100.488572, 98.3426842, 96.1666382, 93.9547708, 91.7014186, 89.4009183, 87.0476065, 84.6358199, 82.1598951, 79.6141689, 76.9929779, 74.2906587, 71.5015479, 68.6199824, 64.8983191, 60.717491, 56.4673492, 52.166309, 47.8327856, 43.4851945, 39.1419509, 34.8214701, 30.5421675, 26.3224584, 22.1807582, 18.1354821, 14.2050454, 10.4078636, 6.76235183, 3.28692553, 0, 3.48671732, 6.17080113, 8.10946072, 9.35990536, 9.97934435, 10.024987, 9.5540425, 8.62372025, 7.29122948, 5.6137795, 3.64857958, 1.45283901, -0.916232924, -3.40142693, -5.94553373, -8.49134402, -11.5420082, -15.1203682, -17.9017348, -19.9489757, -21.3249588, -22.0925516, -22.3146222, -22.0540381, -21.3736672, -20.3363773, -19.0050362, -17.4425116, -15.7116712, -15.1354817, -14.0546758, -12.1540917, -9.11856758, -3.57942454, 4.25174075, 13.8823938, 25.3116414, 38.5385903, 53.5623473, 70.3820191, 85.01899, 102.660701, 124.084507, 150.067767, 181.387837, 222.252042, 269.150278, 322.00241, 380.728304, 445.247825, 512.592223, 589.100505, 679.008923, 784.428713, 903.13253, 1034.84097, 1179.27461, 1330.52956, 1495.2292, 1686.13268, 1915.94531, 2175.84722, 2459.11261, 2764.20279, 3098.36838, 3464.82556, 3867.55948, 4309.98568, 4792.7002, 5317.89394, 5884.16454, 6497.84898, 7162.382, 7881.19831, 8657.73264, 9727.70453, 10804.1927, 11832.5607, 12758.172, 13526.3901, 13838.3512, 14046.4645, 14200.7624, 14507.1068, 14953.8475, 15464.4057, 16030.3781, 16643.361, 17294.9509, 17976.7443, 19459.6589, 20316.8716, 20789.4166, 21118.3282, 21544.6407, 22309.3884, 23653.6054, 25384.5761, 26879.046},
        {193.605523, 194.277934, 194.85606, 195.341536, 195.735999, 196.041088, 196.25844, 196.389691, 196.436479, 196.400441, 196.283215, 196.086438, 195.811746, 195.460778, 195.03517, 194.53656, 193.966585, 193.326883, 192.61909, 191.844843, 191.005781, 190.10354, 189.139757, 188.116071, 187.034117, 185.895534, 184.701958, 183.455028, 182.156379, 180.80765, 179.410477, 177.966498, 176.477351, 174.944671, 173.391404, 171.808084, 170.184991, 168.523073, 166.823273, 165.08654, 163.313817, 161.506051, 159.664189, 157.789175, 155.881955, 153.943476, 151.974684, 149.976523, 147.94994, 145.895881, 143.815292, 141.692603, 139.545633, 137.375686, 135.184066, 132.972078, 130.741024, 128.492209, 126.226936, 123.94651, 121.652234, 119.345412, 117.027348, 114.699346, 112.362709, 110.018743, 107.668749, 105.425631, 103.399611, 101.36183, 99.3063574, 97.2272628, 95.1186156, 92.9744853, 90.7889412, 88.5560528, 86.2698895, 83.9245207, 81.5140158, 79.0324443, 76.4738757, 73.8323792, 71.1020243, 68.2768806, 64.5837601, 60.4164991, 56.1781557, 51.8876984, 47.5640955, 43.2263153, 38.8933264, 34.5840971, 30.3175957, 26.1127907, 21.9886506, 17.9641436, 14.0582381, 10.2899027, 6.67810555, 3.2418152, 0, 2.51536062, 4.44748496, 5.83741959, 6.72621107, 7.15490597, 7.16455086, 6.79619228, 6.09087683, 5.08965104, 3.8335615, 2.36365477, 0.720977406, -1.05342402, -2.91850295, -4.83321281, -6.75650704, -8.80142751, -10.9489403, -12.8094654, -14.3749754, -15.6374424, -16.5888391, -17.2211376, -17.5263105, -17.49633, -17.1231686, -16.3987987, -15.3151926, -13.8643228, -12.8229253, -11.1104621, -8.48347646, -4.6985116, 0.930390292, 8.42543542, 17.6908297, 28.8372795, 41.9754913, 57.2161714, 74.6700263, 91.707361, 112.013318, 136.234771, 165.01859, 199.011647, 242.529523, 291.633204, 346.052386, 405.516763, 469.756031, 534.505271, 608.371097, 697.317699, 804.134672, 925.130149, 1059.5223, 1206.52932, 1360.07728, 1525.8792, 1715.42125, 1938.20458, 2189.55669, 2467.25578, 2771.32632, 3105.599, 3472.15737, 3873.21821, 4310.89961, 4786.83127, 5302.91914, 5855.05646, 6452.08696, 7097.99793, 7796.77667, 8552.41047, 9608.07512, 10668.7723, 11678.6922, 12582.0249, 13322.9607, 13595.857, 13761.2918, 13870.5264, 14131.0034, 14532.1623, 14998.3673, 15522.3116, 16096.6887, 16714.1918, 17367.5144, 18614.2465, 19390.1824, 19897.5696, 20338.6553, 20915.6868, 21830.9115, 23286.5766, 24996.8446, 26142.792},
        {191.875385, 192.507405, 193.047279, 193.496612, 193.857013, 194.130088, 194.317443, 194.420686, 194.441423, 194.381262, 194.241809, 194.024672, 193.731456, 193.36377, 192.923219, 192.411411, 191.829952, 191.180451, 190.464512, 189.683744, 188.839753, 187.934146, 186.96853, 185.944512, 184.863698, 183.727697, 182.538113, 181.296555, 180.00463, 178.663943, 177.276103, 175.842716, 174.365388, 172.845727, 171.306722, 169.73884, 168.132289, 166.487981, 164.80683, 163.089749, 161.337651, 159.551448, 157.732054, 155.880382, 153.997344, 152.083854, 150.140825, 148.169169, 146.1698, 144.14363, 142.091573, 139.994498, 137.873796, 135.730814, 133.5669, 131.383402, 129.181666, 126.963041, 124.728873, 122.480511, 120.219301, 117.946592, 115.66373, 113.372064, 111.07294, 108.767706, 106.45771, 104.259936, 102.288593, 100.306535, 98.3076131, 96.2856785, 94.2345822, 92.1481755, 90.0203095, 87.8448354, 85.6156043, 83.3264674, 80.9712759, 78.5438809, 76.0381337, 73.4478853, 70.7669869, 67.9892898, 64.3195331, 60.1627219, 55.9333458, 51.650864, 47.334736, 43.0044211, 38.6793786, 34.379068, 30.1229486, 25.9304798, 21.8211209, 17.8143312, 13.9295702, 10.1862973, 6.60397165, 3.20205278, 0, 1.61414669, 2.84706317, 3.72499563, 4.27419025, 4.52089323, 4.49135077, 4.21180906, 3.70851429, 3.00771266, 2.13565035, 1.11857356, -0.0172715113, -1.24563868, -2.54028175, -3.87495452, -5.22341081, -6.32487685, -7.11960251, -8.11182063, -9.2250291, -10.3827258, -11.5084087, -12.5255756, -13.3577245, -13.9283532, -14.1609597, -13.9790418, -13.3060974, -12.0656245, -10.4890902, -8.07910106, -4.66676427, -0.0831869452, 5.6744393, 12.8580496, 21.7638681, 32.6106132, 45.617003, 61.0017556, 78.9835894, 98.3164468, 121.129623, 147.928423, 179.218151, 215.504112, 260.397738, 310.521675, 365.604696, 425.375571, 489.563076, 553.617147, 626.775079, 715.430393, 822.587995, 944.334364, 1079.86221, 1228.36425, 1384.88483, 1553.70783, 1743.64264, 1960.10701, 2203.17724, 2475.29354, 2778.17137, 3112.29819, 3478.72511, 3878.02328, 4311.11939, 4780.70009, 5288.45652, 5828.51404, 6411.29349, 7041.10487, 7722.25821, 8459.06354, 9497.99252, 10540.6531, 11530.815, 12412.2477, 13128.7209, 13372.6254, 13506.6954, 13582.4343, 13804.1445, 14162.8464, 14585.894, 15067.4911, 15601.8413, 16183.1484, 16805.616, 17714.8319, 18387.9859, 18959.2195, 19562.674, 20332.4907, 21402.8112, 22907.7767, 24551.4731, 25374.4558},
        {190.658088, 191.224616, 191.702726, 192.093935, 192.399763, 192.621727, 192.761347, 192.82014, 192.799627, 192.701324, 192.52675, 192.277425, 191.954866, 191.560593, 191.096123, 190.562975, 189.962668, 189.296721, 188.566651, 187.773977, 186.920218, 186.006893, 185.03552, 184.007617, 182.924703, 181.788296, 180.599916, 179.36108, 178.073307, 176.738116, 175.357025, 173.931553, 172.463218, 170.953539, 169.420954, 167.858166, 166.257674, 164.620447, 162.947456, 161.239669, 159.498057, 157.723588, 155.917232, 154.079959, 152.212737, 150.316538, 148.392328, 146.44108, 144.463761, 142.461341, 140.43479, 138.373632, 136.29053, 134.186701, 132.063363, 129.921734, 127.76303, 125.58847, 123.399272, 121.196652, 118.981828, 116.756018, 114.52044, 112.276311, 110.024849, 107.76727, 105.504794, 103.353123, 101.425213, 99.4861612, 97.5297646, 95.5498187, 93.5401194, 91.4944626, 89.4066442, 87.2704601, 85.0797062, 82.8281784, 80.5096726, 78.1179847, 75.6469105, 73.090246, 70.4417871, 67.6953296, 64.0456732, 59.9029178, 55.6867665, 51.4169441, 47.1131755, 42.7951857, 38.4826993, 34.1954414, 29.9531367, 25.7755101, 21.6822866, 17.6931908, 13.8279478, 10.1062824, 6.54791933, 3.17258359, 0, 1.45426717, 2.57030847, 3.3715648, 3.88147704, 4.12348609, 4.12103284, 3.89755818, 3.476503, 2.88130819, 2.13541466, 1.26226328, 0.285294955, -0.772049427, -1.88632898, -3.0341028, -4.19193, -4.99447396, -5.36945153, -6.06675742, -6.98968735, -8.04153707, -9.12560231, -10.1451788, -11.0035623, -11.6040485, -11.8499332, -11.6445121, -10.8910809, -9.49293544, -7.51281967, -4.63611668, -0.718287708, 4.385206, 10.4791835, 17.8780434, 26.9867763, 38.0518367, 51.3196793, 67.0367589, 85.4495301, 105.944032, 129.889001, 157.699232, 189.789524, 226.574673, 268.981425, 316.78464, 370.271127, 429.727696, 495.441156, 565.778179, 645.292546, 736.766661, 841.902481, 960.196051, 1092.13383, 1238.20226, 1397.28126, 1571.82895, 1766.05606, 1981.92147, 2222.26463, 2491.12699, 2790.93308, 3121.3676, 3483.37286, 3877.31935, 4304.00111, 4766.30899, 5265.94778, 5798.5863, 6372.89391, 6992.66477, 7661.69303, 8383.77285, 9393.59299, 10406.3293, 11368.0524, 12224.8328, 12922.7409, 13168.3094, 13306.8386, 13387.058, 13598.3495, 13933.7677, 14327.7623, 14776.3182, 15275.4204, 15821.0538, 16409.2034, 16836.6614, 17371.488, 18036.3819, 18854.0422, 19847.1675, 21038.4569, 22450.6093, 23891.038, 24552.9213},
        {189.838058, 190.332549, 190.742561, 191.069515, 191.314831, 191.479929, 191.566231, 191.575157, 191.508128, 191.366565, 191.151888, 190.865518, 190.508875, 190.083381, 189.590456, 189.031521, 188.407996, 187.721303, 186.972861, 186.164092, 185.296416, 184.371254, 183.390027, 182.354155, 181.265059, 180.12416, 178.932879, 177.692636, 176.404851, 175.070946, 173.692341, 172.270457, 170.806715, 169.302535, 167.770392, 166.205945, 164.604726, 162.967795, 161.296215, 159.591047, 157.853355, 156.084198, 154.284641, 152.455744, 150.598569, 148.71418, 146.803637, 144.868003, 142.908339, 140.925708, 138.921172, 136.898133, 134.855262, 132.79357, 130.714069, 128.617769, 126.505683, 124.378821, 122.238194, 120.084815, 117.919694, 115.743842, 113.558271, 111.363992, 109.162016, 106.953355, 104.73902, 102.630964, 100.738478, 98.8337622, 96.9106362, 94.9629193, 92.9844308, 90.9689901, 88.9104165, 86.8025294, 84.6391481, 82.4140919, 80.1211801, 77.7542321, 75.3070672, 72.7735048, 70.1473642, 67.4224647, 63.7899377, 59.6618299, 55.4598359, 51.2038237, 46.9136615, 42.6092173, 38.3103593, 34.0369555, 29.8088741, 25.6459832, 21.5681508, 17.595245, 13.7471341, 10.043686, 6.50476889, 3.15025085, 0, 1.75994538, 3.12531453, 4.12423081, 4.78481759, 5.13519823, 5.20349612, 5.0178346, 4.60633706, 3.99712686, 3.21832737, 2.29806195, 1.26445398, 0.145626822, -1.03029616, -2.23519159, -3.4409361, -4.34143225, -4.86884609, -5.62380514, -6.51862016, -7.4656019, -8.37706112, -9.16530859, -9.74265506, -10.0214113, -9.91388803, -9.33239604, -8.18924609, -6.39674893, -4.07004585, -0.844245884, 3.42918938, 8.89879833, 15.4378399, 23.3324923, 32.9423414, 44.4985095, 58.2321189, 74.3742917, 93.1561502, 114.167487, 138.475941, 166.438113, 198.410601, 234.750007, 272.887921, 316.837204, 367.685706, 426.52128, 494.431777, 574.097936, 663.067859, 758.779336, 860.970722, 974.077363, 1100.42575, 1242.34237, 1403.51586, 1584.60099, 1784.7666, 2003.33384, 2244.39902, 2511.64685, 2807.55356, 3132.54766, 3487.99757, 3874.93336, 4294.63575, 4749.62613, 5241.72411, 5769.59202, 6338.69245, 6951.93791, 7612.24091, 8322.51394, 9299.18506, 10278.2723, 11209.3094, 12041.8298, 12725.3672, 12987.4833, 13147.6649, 13250.5765, 13465.8845, 13788.7735, 14162.3438, 14584.3841, 15052.6831, 15565.0295, 16119.2119, 16048.2179, 16444.1042, 17207.5934, 18239.4079, 19440.2702, 20710.903, 21952.0286, 23127.9659, 23699.3306},
        {189.297894, 189.732807, 190.08597, 190.358748, 190.552503, 190.668599, 190.708401, 190.673271, 190.564573, 190.38367, 190.131927, 189.810707, 189.421373, 188.96529, 188.44382, 187.858327, 187.210175, 186.500728, 185.731349, 184.903401, 184.018249, 183.077255, 182.081784, 181.033199, 179.932864, 178.782142, 177.582397, 176.334993, 175.041292, 173.70266, 172.320458, 170.896052, 169.430803, 167.926077, 166.390243, 164.821013, 163.216017, 161.576392, 159.903272, 158.197796, 156.461098, 154.694316, 152.898585, 151.075042, 149.224823, 147.349064, 145.448901, 143.525472, 141.579911, 139.613355, 137.626941, 135.636093, 133.627349, 131.601536, 129.55948, 127.502008, 125.429946, 123.34412, 121.245358, 119.134486, 117.01233, 114.879716, 112.737472, 110.586423, 108.427397, 106.261219, 104.088717, 102.018545, 100.15677, 98.2818294, 96.3875599, 94.4677978, 92.5163795, 90.5271413, 88.4939197, 86.410551, 84.2708717, 82.0687181, 79.7979266, 77.4523337, 75.0257757, 72.512089, 69.90511, 67.1986752, 63.5806196, 59.4647334, 55.2744944, 51.0298952, 46.7509284, 42.4575869, 38.1698632, 33.9077501, 29.6912402, 25.5403262, 21.4750008, 17.5152568, 13.6810867, 9.99248331, 6.46943926, 3.13194727, 0, 2.26048133, 4.02930452, 5.34233193, 6.23542594, 6.74444891, 6.90526322, 6.75373124, 6.32571533, 5.65707786, 4.78368122, 3.74138775, 2.56605985, 1.29355986, -0.0402498221, -1.39950684, -2.74834883, -3.89973534, -4.79620713, -5.74385224, -6.67441299, -7.51963172, -8.21125076, -8.68101243, -8.86065906, -8.68193298, -8.07657652, -6.976332, -5.31294176, -3.01814812, -0.326861946, 3.24350476, 7.85216024, 13.6583127, 20.6525234, 29.0783241, 39.224219, 51.3000104, 65.5155007, 82.0804921, 101.204787, 122.573754, 147.08124, 175.039653, 206.761402, 242.558895, 276.743717, 317.129308, 365.528282, 423.753256, 493.616844, 581.688216, 679.209866, 778.874455, 878.662569, 987.350984, 1108.86921, 1247.14675, 1409.87541, 1596.36733, 1801.83033, 2024.00878, 2267.18677, 2533.8047, 2826.04428, 3145.64897, 3494.56338, 3874.76019, 4288.19131, 4736.70577, 5222.21078, 5745.92799, 6310.55568, 6918.23836, 7571.12058, 8271.34687, 9219.33264, 10169.3838, 11074.0772, 11885.9896, 12557.6979, 12835.3232, 13015.535, 13139.3913, 13359.159, 13671.8757, 14028.2417, 14427.6116, 14869.3395, 15352.7801, 15877.2876, 15422.3402, 15714.6312, 16555.9907, 17748.2482, 19093.2338, 20392.7771, 21448.708, 22375.7364, 22835.9121},
        {188.77964, 189.215877, 189.56792, 189.837261, 190.025393, 190.133808, 190.163999, 190.117457, 189.995675, 189.800145, 189.532359, 189.193811, 188.785991, 188.310393, 187.768509, 187.16183, 186.49185, 185.76006, 184.967953, 184.117021, 183.208756, 182.244652, 181.226199, 180.15489, 179.032218, 177.859675, 176.638753, 175.370945, 174.057742, 172.700637, 171.301123, 169.860691, 168.380834, 166.863044, 165.3221, 163.75257, 162.148866, 160.512049, 158.843179, 157.143319, 155.413528, 153.654869, 151.868402, 150.055189, 148.21629, 146.352767, 144.465681, 142.556093, 140.625065, 138.673656, 136.70293, 134.723179, 132.726031, 130.712348, 128.682991, 126.638821, 124.580698, 122.509484, 120.42604, 118.331227, 116.225905, 114.110935, 111.98718, 109.855499, 107.716754, 105.571805, 103.421515, 101.377633, 99.54925, 97.7085228, 95.8491238, 93.9647256, 92.0490005, 90.095621, 88.0982595, 86.0505884, 83.94628, 81.7790069, 79.5424414, 77.230256, 74.836123, 72.3537149, 69.7767041, 67.0987631, 63.4932333, 59.3812074, 55.1936372, 50.9508165, 46.6730391, 42.3805987, 38.0937889, 33.8329036, 29.6182365, 25.4700812, 21.4087316, 17.4544813, 13.627624, 9.94845351, 6.43726354, 3.11434779, 0, 2.63009139, 4.70334812, 6.26090174, 7.34388379, 7.99342582, 8.25065938, 8.15671601, 7.75272726, 7.07982467, 6.17913978, 5.09180415, 3.85894932, 2.52170684, 1.12120824, -0.30141492, -1.7050311, -2.93760041, -3.94214727, -4.92840421, -5.83147364, -6.58645791, -7.12845943, -7.39258057, -7.31392371, -6.82759123, -5.86868553, -4.37230897, -2.27356394, 0.492447176, 3.73724158, 7.87200346, 13.0376446, 19.3750766, 26.6731666, 35.2488486, 45.5129354, 57.7119522, 72.0924242, 88.9008765, 108.383834, 130.720856, 156.245814, 185.218338, 217.898053, 254.544588, 289.221532, 329.933559, 378.489308, 436.697416, 506.36652, 593.805155, 690.822184, 790.813288, 891.985725, 1002.37039, 1125.58757, 1265.25755, 1428.14814, 1614.01675, 1819.18711, 2042.55889, 2287.07648, 2553.98764, 2844.98706, 3162.52665, 3508.71075, 3885.93373, 4296.37515, 4741.151, 5221.97891, 5738.15948, 6294.19915, 6892.65218, 7536.07284, 8227.01537, 9171.5971, 10117.9185, 11017.6429, 11822.434, 12483.9552, 12743.5976, 12903.4786, 13005.3787, 13202.7941, 13493.2842, 13827.8598, 14206.4076, 14628.8144, 15094.9667, 15604.7514, 15212.284, 15529.0233, 16361.9405, 17518.0068, 18804.1935, 20027.4716, 20994.8126, 21868.1458, 22042.2963},
        {188.497448, 188.948644, 189.312053, 189.589334, 189.782143, 189.892137, 189.920974, 189.870312, 189.741806, 189.537115, 189.257897, 188.905807, 188.482503, 187.989643, 187.428884, 186.801883, 186.110298, 185.355785, 184.540002, 183.664606, 182.731254, 181.741604, 180.697313, 179.600038, 178.451436, 177.253165, 176.006882, 174.714244, 173.376908, 171.996532, 170.574773, 169.113288, 167.613735, 166.07777, 164.528544, 162.956516, 161.351885, 159.715612, 158.048657, 156.351979, 154.626538, 152.873293, 151.093206, 149.287234, 147.456338, 145.601478, 143.723614, 141.823705, 139.902711, 137.961592, 136.001307, 134.023128, 132.027696, 130.015965, 127.988887, 125.947415, 123.892503, 121.825103, 119.746169, 117.656654, 115.55751, 113.44969, 111.334148, 109.211837, 107.083709, 104.950718, 102.813817, 100.789369, 98.9909678, 97.1815397, 95.3545564, 93.5034891, 91.6218092, 89.7029883, 87.7404976, 85.7278087, 83.6583928, 81.5257215, 79.3232661, 77.044498, 74.6828886, 72.2319094, 69.6850317, 67.0357269, 63.4409043, 59.3299124, 55.1421543, 50.8982496, 46.6188178, 42.3244782, 38.0358504, 33.7735538, 29.5582079, 25.4104322, 21.350846, 17.4000688, 13.5787202, 9.90741957, 6.40678634, 3.09744, 0, 3.01374059, 5.40345901, 7.2156908, 8.49697151, 9.29383667, 9.65282184, 9.62046257, 9.24329439, 8.56785285, 7.6406735, 6.50829189, 5.21724355, 3.81406404, 2.34528889, 0.857453662, -0.602906108, -1.92269166, -3.04585843, -4.0764305, -4.95360878, -5.61659419, -6.00458764, -6.05679003, -5.71240227, -4.91062528, -3.59065996, -1.69170723, 0.847032014, 4.08635686, 7.88200702, 12.5750294, 18.2877409, 25.1424584, 32.6979335, 41.3582657, 51.6778387, 63.9480391, 78.4602532, 95.5058677, 115.376269, 138.754775, 165.421558, 195.591321, 229.478769, 267.298606, 303.758569, 345.957072, 395.485559, 453.935476, 522.898269, 607.23968, 701.27494, 800.62331, 904.272333, 1018.08648, 1144.84262, 1287.31762, 1449.97083, 1633.51427, 1836.8245, 2060.75797, 2306.43824, 2573.84188, 2864.104, 3180.32284, 3524.69554, 3899.98777, 4308.54404, 4750.62433, 5227.66775, 5736.0352, 6283.02528, 6871.83496, 7505.6612, 8187.70096, 9134.16284, 10081.9793, 10981.0943, 11781.452, 12432.9964, 12667.4414, 12798.448, 12869.4875, 13039.5976, 13306.5045, 13619.7817, 13979.563, 14385.9824, 14839.1735, 15339.2702, 15164.7517, 15550.3516, 16344.0859, 17393.9708, 18548.0221, 19654.2558, 20560.688, 21409.7985, 21222.1545},
        {188.502602, 188.968143, 189.342427, 189.627264, 189.824466, 189.935842, 189.963204, 189.908362, 189.773126, 189.559307, 189.268716, 188.903163, 188.464459, 187.954415, 187.37484, 186.727547, 186.014344, 185.237043, 184.397454, 183.497389, 182.538657, 181.523069, 180.452435, 179.328567, 178.153275, 176.928369, 175.655661, 174.33696, 172.974077, 171.568823, 170.123008, 168.638443, 167.116939, 165.560305, 163.999415, 162.420936, 160.811188, 159.171039, 157.501356, 155.803006, 154.076857, 152.323777, 150.544633, 148.740292, 146.911623, 145.059492, 143.184766, 141.288315, 139.371004, 137.433702, 135.477276, 133.493489, 131.492511, 129.475406, 127.44324, 125.397076, 123.337979, 121.267015, 119.185249, 117.093745, 114.993567, 112.885782, 110.771454, 108.651647, 106.527426, 104.399857, 102.270003, 100.259213, 98.485945, 96.7031909, 94.9042183, 93.0822947, 91.2306876, 89.3426643, 87.4114924, 85.4304393, 83.3927724, 81.2917592, 79.1206672, 76.8727637, 74.5413162, 72.1195922, 69.6008591, 66.9783844, 63.392779, 59.281573, 55.092545, 50.8466327, 46.5647739, 42.2679064, 37.9769681, 33.7128967, 29.4966299, 25.3491057, 21.2912617, 17.3440358, 13.5283657, 9.86518934, 6.37544439, 3.08006868, 0, 3.39399473, 6.09761122, 8.16272344, 9.64120539, 10.584931, 11.0457743, 11.0756093, 10.7263099, 10.0497502, 9.09780399, 7.92234539, 6.57524834, 5.10838683, 3.57363483, 2.02286632, 0.507955287, -0.908887079, -2.17431005, -3.25982145, -4.11133218, -4.67475313, -4.89599523, -4.72096935, -4.09558642, -2.96575733, -1.27739298, 1.02359572, 3.99129787, 7.67980256, 11.9571364, 17.1316738, 23.3133228, 30.6119912, 38.3913224, 47.1343563, 57.5231372, 69.8914521, 84.5730882, 101.901833, 122.211473, 146.618662, 174.436058, 205.844278, 241.02394, 280.155662, 319.047995, 363.34664, 414.325232, 473.257405, 541.416792, 621.703052, 711.776435, 810.100386, 916.672726, 1034.62397, 1165.6781, 1311.55913, 1473.93212, 1654.59339, 1855.40354, 2079.28924, 2325.59936, 2593.19932, 2882.72383, 3197.80557, 3540.70111, 3914.52777, 4321.76529, 4761.73724, 5235.5525, 5736.48519, 6274.68429, 6853.97072, 7478.16537, 8151.08916, 9098.79856, 10047.32, 10944.9156, 11739.8472, 12380.377, 12588.4266, 12689.4919, 12728.8379, 12871.6051, 13115.608, 13408.7122, 13751.2293, 14143.4709, 14585.7486, 15078.3738, 15201.1967, 15677.2689, 16414.854, 17322.2156, 18307.6174, 19279.3231, 20145.5961, 20953.3619, 20339.8609},
        {188.855419, 189.290113, 189.634423, 189.890139, 190.059049, 190.142945, 190.143615, 190.06285, 189.90244, 189.664174, 189.349843, 188.961235, 188.500142, 187.968353, 187.367657, 186.699845, 185.966706, 185.170031, 184.311608, 183.393229, 182.416683, 181.38376, 180.296249, 179.15594, 177.964624, 176.724091, 175.436129, 174.102529, 172.725081, 171.305575, 169.8458, 168.347546, 166.812604, 165.242762, 163.66704, 162.07304, 160.448038, 158.79294, 157.108652, 155.396082, 153.656134, 151.889716, 150.097733, 148.281092, 146.440699, 144.57746, 142.692282, 140.786071, 138.859732, 136.914173, 134.9503, 132.96108, 130.95553, 128.934729, 126.899754, 124.851685, 122.7916, 120.720576, 118.639693, 116.550029, 114.452661, 112.348669, 110.23913, 108.125124, 106.007728, 103.888021, 101.767081, 99.7680079, 98.0109713, 96.2453899, 94.4644322, 92.6612662, 90.8290602, 88.9609824, 87.050201, 85.0898843, 83.0732004, 80.9933176, 78.8434041, 76.616628, 74.3061577, 71.9051613, 69.406807, 66.8042631, 63.2277425, 59.1200421, 54.934331, 50.6917819, 46.4135676, 42.1208609, 37.8348345, 33.5766613, 29.3675139, 25.2285652, 21.180988, 17.2459549, 13.4446389, 9.79821261, 6.32784885, 3.05472039, 0, 3.81210897, 6.85984072, 9.20100192, 10.8933993, 11.9948394, 12.5631291, 12.656075, 12.3314837, 11.647162, 10.6609166, 9.43055404, 8.01388113, 6.46870451, 4.85283088, 3.22406691, 1.6402193, 0.0217593359, -1.59312563, -2.83598936, -3.67845419, -4.09214243, -4.04867643, -3.51967849, -2.47677096, -0.891576143, 1.26428362, 4.01918601, 7.4015087, 11.4396294, 15.7911872, 20.9912359, 27.1793745, 34.4952022, 42.4246977, 51.4342706, 62.164629, 74.9514584, 90.1304441, 108.037272, 129.007626, 154.255298, 182.970618, 215.317469, 251.459733, 291.561292, 331.767026, 377.26457, 429.222557, 488.809622, 557.194397, 636.609661, 725.859278, 824.369317, 932.484188, 1052.42324, 1185.56402, 1333.28406, 1496.6168, 1677.36209, 1877.69517, 2100.41818, 2345.03575, 2610.39537, 2897.206, 3209.33079, 3549.18508, 3920.44596, 4325.85587, 4763.53085, 5234.20392, 5730.45634, 6262.42696, 6833.67053, 7447.74181, 8108.19555, 9033.66457, 9958.85605, 10833.5552, 11607.5473, 12230.6176, 12432.8256, 12530.166, 12566.5923, 12699.7574, 12928.6085, 13204.7184, 13529.4032, 13903.9789, 14329.7618, 14808.0679, 15122.2136, 15653.2448, 16349.7475, 17160.3078, 18033.5116, 18917.945, 19762.1941, 20369.6008, 19217.301},
        {189.594339, 189.981866, 190.281297, 190.494342, 190.62271, 190.668111, 190.632256, 190.516854, 190.323616, 190.054251, 189.71047, 189.293983, 188.806499, 188.249728, 187.625381, 186.935167, 186.180797, 185.363981, 184.486428, 183.549849, 182.555953, 181.506451, 180.403052, 179.247467, 178.041405, 176.786577, 175.484693, 174.137462, 172.746595, 171.313802, 169.840792, 168.329275, 166.780963, 165.197564, 163.603023, 161.987465, 160.340748, 158.663861, 156.957791, 155.223528, 153.46206, 151.674376, 149.861463, 148.024311, 146.163907, 144.28124, 142.377299, 140.453072, 138.509548, 136.547715, 134.568561, 132.569309, 130.554794, 128.526088, 126.484261, 124.430381, 122.36552, 120.290748, 118.207134, 116.115749, 114.017663, 111.913946, 109.805668, 107.693899, 105.579709, 103.464169, 101.348348, 99.3562143, 97.608785, 95.8536629, 94.0839512, 92.2927529, 90.4731712, 88.6183092, 86.7212702, 84.7751572, 82.7730734, 80.708122, 78.573406, 76.3620287, 74.0670931, 71.6817025, 69.1989599, 66.6119686, 63.0432665, 58.9395506, 54.7579812, 50.519913, 46.2467006, 41.9596987, 37.6802618, 33.4297448, 29.2295022, 25.1008887, 21.065259, 17.1439676, 13.3583694, 9.7298188, 6.27967062, 3.02927947, 0, 4.16720985, 7.50733226, 10.0832775, 11.9579559, 13.1942776, 13.8551531, 14.0034925, 13.7022061, 13.0142043, 12.0023972, 10.7296952, 9.25900862, 7.65324762, 5.97532254, 4.28814364, 2.65462121, 0.85542769, -1.08684633, -2.47089546, -3.29428897, -3.5545961, -3.24938612, -2.37622829, -0.932691854, 1.08365392, 3.67523978, 6.84449647, 10.5938547, 14.9257453, 19.2639228, 24.4016753, 30.5150365, 37.7800398, 45.8745053, 55.2235734, 66.3852413, 79.6850069, 95.4483677, 114.000821, 135.667866, 161.608352, 191.060795, 224.187645, 261.151351, 302.114365, 343.474917, 390.100732, 443.095313, 503.562165, 572.604794, 652.092837, 741.42728, 840.493199, 949.804713, 1071.16024, 1205.78157, 1354.8905, 1519.4624, 1701.0215, 1901.36083, 2122.57855, 2364.86341, 2627.38954, 2911.10069, 3219.93922, 3556.47108, 3924.89214, 4328.19094, 4763.37956, 5230.85069, 5723.38495, 6250.15821, 6814.19809, 7418.53225, 8066.18833, 8964.09383, 9860.40156, 10707.1642, 11456.4345, 12060.265, 12260.8304, 12359.9802, 12399.7147, 12527.6469, 12743.9722, 13004.881, 13312.7691, 13670.0321, 14079.0659, 14542.2662, 15021.0823, 15597.5574, 16255.8596, 16980.1569, 17754.6173, 18563.4088, 19390.6994, 19727.6243, 18021.4256},
        {190.764245, 191.095524, 191.341588, 191.504041, 191.584488, 191.584533, 191.50578, 191.349833, 191.118296, 190.812775, 190.434873, 189.986194, 189.468342, 188.882923, 188.23154, 187.515797, 186.737299, 185.89765, 184.998455, 184.041316, 183.02784, 181.95963, 180.83829, 179.665424, 178.442637, 177.171534, 175.853718, 174.490793, 173.084365, 171.636036, 170.147412, 168.620097, 167.055694, 165.455809, 163.837948, 162.195429, 160.52138, 158.81689, 157.083045, 155.320935, 153.531648, 151.716272, 149.875894, 148.011605, 146.12449, 144.21564, 142.286141, 140.337082, 138.369552, 136.384638, 134.383429, 132.368346, 130.339114, 128.296794, 126.242446, 124.177127, 122.101898, 120.017817, 117.925945, 115.82734, 113.723062, 111.61417, 109.501724, 107.386782, 105.270404, 103.15365, 101.037578, 99.0467353, 97.3027025, 95.5518567, 93.7872522, 92.0019431, 90.1889837, 88.3414281, 86.4523304, 84.5147449, 82.5217256, 80.4663269, 78.3416027, 76.1406074, 73.8563951, 71.48202, 69.0105361, 66.4349978, 62.872448, 58.7723722, 54.5947848, 50.3611682, 46.0930049, 41.8117773, 37.5389678, 33.2960588, 29.1045327, 24.985872, 20.9615591, 17.0530763, 13.2819062, 9.66953104, 6.23743336, 3.00709554, 0, 4.40028149, 7.93291458, 10.6642742, 12.6607352, 13.9886724, 14.7144609, 14.9044756, 14.6250912, 13.9426827, 12.923625, 11.6342931, 10.1410618, 8.51030598, 6.80840061, 5.10172056, 3.45664074, 1.5582028, -0.581694533, -2.03413687, -2.81446357, -2.938014, -2.42012752, -1.27614351, 0.478598668, 2.82875966, 5.75900009, 9.2539806, 13.2983618, 17.8768044, 22.2400258, 27.3757423, 33.4887975, 40.7840349, 49.1088821, 58.8468907, 70.4769232, 84.3110481, 100.661334, 119.839849, 142.158661, 168.621234, 198.637818, 232.385206, 270.040196, 311.779581, 354.293141, 402.116443, 456.298036, 517.88647, 587.930296, 668.074176, 758.041964, 857.909341, 968.263511, 1090.73602, 1226.48953, 1376.6867, 1542.47732, 1725.03984, 1925.56674, 2145.33256, 2385.30433, 2645.07402, 2925.84041, 3231.52489, 3564.7991, 3930.22819, 4330.97471, 4763.25851, 5227.22664, 5716.44013, 6238.64526, 6796.26886, 7391.73772, 8027.47868, 8898.58164, 9766.64453, 10585.9284, 11310.6943, 11895.2032, 12093.1932, 12193.1304, 12235.2141, 12358.0853, 12562.977, 12810.1052, 13102.754, 13444.2072, 13837.749, 14286.6634, 14904.8852, 15518.8072, 16142.6283, 16790.5476, 17476.7638, 18215.4761, 19020.8832, 19038.1282, 16817.5622},
        {192.497459, 192.772543, 192.964753, 193.075606, 193.106622, 193.05932, 192.935219, 192.735838, 192.462696, 192.117312, 191.701205, 191.215894, 190.662898, 190.043737, 189.359929, 188.612994, 187.80445, 186.935816, 186.008612, 185.024356, 183.984568, 182.890767, 181.744472, 180.547201, 179.300474, 178.00581, 176.664728, 175.278747, 173.849386, 172.378165, 170.866601, 169.316214, 167.728524, 166.105049, 164.45807, 162.783498, 161.077117, 159.340096, 157.573605, 155.778814, 153.956893, 152.109011, 150.236339, 148.340046, 146.421302, 144.481277, 142.52114, 140.542062, 138.545212, 136.531761, 134.502878, 132.464331, 130.412592, 128.348731, 126.273818, 124.188924, 122.095118, 119.993471, 117.885053, 115.770934, 113.652183, 111.529872, 109.405071, 107.278848, 105.152276, 103.026423, 100.90236, 98.9057409, 97.1592275, 95.4069461, 93.6418907, 91.857055, 90.045433, 88.2000186, 86.3138057, 84.3797881, 82.3909599, 80.3403148, 78.2208469, 76.0255498, 73.7474177, 71.3794443, 68.9146236, 66.3459495, 62.7875763, 58.6904028, 54.515891, 50.2855548, 46.0209081, 41.7434648, 37.4747389, 33.2362443, 29.049495, 24.9360048, 20.9172876, 17.0148576, 13.2502284, 9.64491416, 6.22042871, 2.99828602, 0, 4.28380754, 7.72312602, 10.3830516, 12.3286803, 13.6251084, 14.3374319, 14.5307471, 14.2701499, 13.6207367, 12.6476034, 11.4158463, 9.99056148, 8.43684504, 6.81979315, 5.20450193, 3.65606751, 1.94163293, 0.0837295975, -1.17860482, -1.84412127, -1.91157072, -1.37970412, -0.247272447, 1.48697335, 3.82428231, 6.76590348, 10.3130859, 14.4670786, 19.2291306, 23.9869909, 29.5803587, 36.1945329, 44.0148127, 52.7947774, 62.935586, 74.9535224, 89.1634017, 105.880039, 125.418248, 148.092846, 174.702301, 204.930574, 238.997852, 277.124322, 319.530169, 363.518228, 412.955376, 468.79114, 531.975045, 603.456615, 684.557536, 775.400312, 876.264771, 987.805489, 1111.44215, 1248.25092, 1399.30796, 1565.61975, 1748.34798, 1948.73038, 2168.02186, 2407.15448, 2665.69909, 2944.82356, 3248.40186, 3579.06584, 3941.26283, 4338.09534, 4766.00908, 5225.21523, 5710.5194, 6228.36995, 6780.84921, 7370.03949, 7998.02312, 8856.68411, 9710.85269, 10515.1607, 11224.2402, 11792.7229, 11975.201, 12059.706, 12086.6014, 12195.0291, 12386.2525, 12620.3772, 12900.7246, 13230.616, 13613.3728, 14052.3162, 14791.7842, 15437.4308, 16028.7197, 16605.1146, 17206.0789, 17871.0765, 18639.571, 18288.9917, 15721.2087},
        {194.832096, 195.055485, 195.197257, 195.258889, 195.241859, 195.147641, 194.977714, 194.733552, 194.416634, 194.028435, 193.570432, 193.044102, 192.450921, 191.792365, 191.069911, 190.285036, 189.439217, 188.533929, 187.570649, 186.550854, 185.47602, 184.347624, 183.167142, 181.936051, 180.655828, 179.327949, 177.95389, 176.535128, 175.07314, 173.569402, 172.02539, 170.442582, 168.822453, 167.166481, 165.484175, 163.772828, 162.029631, 160.255802, 158.452556, 156.621108, 154.762674, 152.878471, 150.969714, 149.037619, 147.083401, 145.108277, 143.113462, 141.100172, 139.069622, 137.02303, 134.961609, 132.891255, 130.808404, 128.714169, 126.609666, 124.496008, 122.374311, 120.245688, 118.111254, 115.972123, 113.82941, 111.684229, 109.537695, 107.390921, 105.245023, 103.101114, 100.960309, 98.9502612, 97.1955835, 95.4364157, 93.6656694, 91.8762558, 90.0610866, 88.213073, 86.3251267, 84.3901589, 82.4010813, 80.3508051, 78.2322419, 76.0383031, 73.7619002, 71.3959446, 68.9333477, 66.367021, 62.81053, 58.714955, 54.5419453, 50.3129489, 46.0494136, 41.7727874, 37.5045181, 33.2660537, 29.0788421, 24.9643313, 20.943969, 17.0392033, 13.2714821, 9.66225324, 6.23296465, 3.00506427, 0, 3.77841618, 6.80644929, 9.14263622, 10.8455139, 11.9736192, 12.5854891, 12.7396605, 12.4946703, 11.9090553, 11.0413526, 9.95009901, 8.69383144, 7.33108681, 5.92040203, 4.52031402, 3.18935969, 1.98270208, 0.958396132, 0.182610103, -0.28684206, -0.392146418, -0.0754890297, 0.720944044, 2.05496674, 3.98439301, 6.56703678, 9.86071201, 13.9232326, 18.8124126, 24.4192344, 31.029515, 38.7511176, 47.6919056, 57.1883309, 67.7339626, 80.0280794, 94.4099681, 111.218915, 130.794208, 153.475133, 179.841132, 209.918246, 243.998775, 282.375017, 325.339275, 371.167919, 422.673159, 480.651276, 545.898551, 619.211267, 701.431646, 793.253876, 895.267606, 1008.21893, 1133.17337, 1271.05301, 1422.77997, 1588.72848, 1770.49302, 1970.26571, 2190.36501, 2430.5992, 2689.90978, 2969.05515, 3271.87226, 3600.78496, 3959.56399, 4350.98236, 4772.87459, 5225.8687, 5706.29605, 6219.74232, 6768.32283, 7354.15289, 7979.34779, 8843.89247, 9702.56523, 10508.014, 11212.8866, 11769.8311, 11920.6876, 11969.4502, 11959.1129, 12041.0959, 12215.4747, 12437.1469, 12708.4469, 13031.7089, 13409.2672, 13843.456, 14677.4158, 15347.5385, 15910.8914, 16424.5418, 16945.557, 17531.0043, 18237.951, 17485.7726, 14780.9934},
        {197.624867, 197.796465, 197.887235, 197.898629, 197.832099, 197.689098, 197.471078, 197.179491, 196.815789, 196.381424, 195.877849, 195.306515, 194.668875, 193.966381, 193.200486, 192.37264, 191.484298, 190.53691, 189.531929, 188.470807, 187.354996, 186.185949, 184.965117, 183.693953, 182.37391, 181.006438, 179.592991, 178.13502, 176.633978, 175.091317, 173.508489, 171.886947, 170.228141, 168.533526, 166.810984, 165.058616, 163.274447, 161.45972, 159.615677, 157.743564, 155.844622, 153.920097, 151.97123, 149.999266, 148.005447, 145.991018, 143.957222, 141.905303, 139.836503, 137.752066, 135.653237, 133.544177, 131.423147, 129.291328, 127.1499, 125.000042, 122.842936, 120.67976, 118.511696, 116.339923, 114.165622, 111.989973, 109.814156, 107.639351, 105.466738, 103.297498, 101.13281, 99.1026985, 97.3340941, 95.5624463, 93.7805827, 91.9813312, 90.1575195, 88.3019753, 86.4075264, 84.4670006, 82.4732255, 80.4190289, 78.2972386, 76.1006823, 73.8221878, 71.4545828, 68.990695, 66.4233523, 62.8661125, 58.7699325, 54.5963373, 50.3666889, 46.1023489, 41.824679, 37.555041, 33.3147967, 29.1253078, 25.007936, 20.9840431, 17.0749908, 13.3021409, 9.68685512, 6.25049521, 3.01442293, 0, 3.20704487, 5.7698482, 7.73951411, 9.16714671, 10.1038501, 10.6007285, 10.7088859, 10.4794265, 9.96345437, 9.21207367, 8.27638851, 7.20750299, 6.05652126, 4.87454741, 3.71268558, 2.62203988, 1.97697321, 1.87476952, 1.63535236, 1.37909557, 1.22637303, 1.29755859, 1.7130261, 2.59314943, 4.05830243, 6.22885897, 9.22519289, 13.1676781, 18.1766883, 24.6969471, 32.4055505, 41.3255676, 51.4800673, 61.7545882, 72.7409644, 85.3343717, 99.8991382, 116.799592, 136.400061, 159.064874, 185.146566, 215.024847, 249.066352, 287.637718, 331.105579, 378.648794, 432.118719, 492.178935, 559.493022, 634.724558, 718.363535, 811.459286, 914.743603, 1028.99051, 1155.06027, 1293.77442, 1445.9545, 1611.41241, 1792.20879, 1991.50576, 2212.80655, 2454.64789, 2715.25996, 2994.93127, 3297.43802, 3624.9555, 3980.45882, 4366.33065, 4782.02095, 5228.6386, 5704.01147, 6213.03419, 6757.95132, 7341.00748, 7964.44723, 8838.6132, 9705.62743, 10515.71, 11219.081, 11765.9606, 11882.9183, 11892.9249, 11842.0508, 11896.3254, 12054.3351, 12264.6833, 12528.4555, 12846.7372, 13220.6139, 13651.1711, 14514.0391, 15192.5429, 15743.7116, 16224.5739, 16692.1589, 17203.4953, 17815.6121, 16668.0716, 13916.7747},
        {200.836347, 200.956641, 200.996377, 200.957003, 200.839968, 200.646718, 200.378704, 200.037372, 199.624172, 199.140551, 198.587957, 197.967839, 197.281645, 196.530823, 195.716821, 194.841088, 193.905072, 192.910221, 191.857983, 190.749806, 189.587138, 188.371429, 187.104125, 185.786676, 184.420529, 183.007132, 181.547934, 180.044384, 178.497928, 176.910016, 175.282095, 173.615615, 171.912022, 170.172765, 168.405395, 166.60818, 164.779316, 162.920052, 161.03164, 159.115329, 157.17237, 155.204012, 153.211507, 151.196103, 149.159052, 147.101603, 145.025007, 142.930514, 140.819374, 138.692837, 136.552154, 134.3977, 132.231619, 130.05518, 127.869652, 125.676304, 123.476406, 121.271226, 119.062033, 116.850097, 114.636687, 112.423073, 110.210522, 108.000304, 105.793689, 103.591946, 101.396343, 99.3396352, 97.5514187, 95.7617854, 93.963481, 92.1492509, 90.3118407, 88.4439958, 86.5384617, 84.5879841, 82.5853083, 80.52318, 78.3943445, 76.1915476, 73.9075345, 71.535051, 69.0668425, 66.4956544, 62.9347081, 58.8350109, 54.6581411, 50.4253866, 46.1580351, 41.8773744, 37.6046922, 33.3612764, 29.1684146, 25.0473947, 21.0195044, 17.1060316, 13.3282638, 9.70748906, 6.26499498, 3.02206937, 0, 2.68518809, 4.82326813, 6.45861407, 7.63559985, 8.39859944, 8.79198677, 8.8601358, 8.64742047, 8.19821473, 7.55689255, 6.76782785, 5.87539459, 4.92396673, 3.9579182, 3.02162297, 2.15945498, 2.03760443, 2.78927618, 3.05235501, 3.00446121, 2.82321507, 2.68623687, 2.77114692, 3.25556551, 4.31711292, 6.13340944, 8.88207538, 12.740731, 17.8869966, 25.267042, 34.0081364, 44.0573355, 55.3616945, 66.4115418, 77.8822965, 90.8378381, 105.662241, 122.739577, 142.453922, 165.189348, 191.086602, 220.847141, 254.902646, 293.6848, 337.625282, 386.57224, 441.686774, 503.546448, 572.728827, 649.811477, 735.147168, 829.813008, 934.451164, 1049.77959, 1176.67098, 1315.92853, 1468.35548, 1633.41204, 1813.54965, 2012.68477, 2235.37536, 2478.98325, 2741.07337, 3021.48689, 3323.92201, 3650.30648, 4002.8477, 4383.54592, 4793.37605, 5233.89297, 5704.15434, 6208.59649, 6749.63874, 7329.70039, 7951.20078, 8834.76034, 9710.047, 10524.9298, 11227.2778, 11764.9599, 11849.8373, 11823.1256, 11733.8404, 11762.2331, 11905.3714, 12105.6414, 12362.845, 12676.7838, 13047.2599, 13474.075, 14278.7489, 14944.8838, 15505.9041, 15995.2346, 16446.2999, 16892.5246, 17367.3334, 15855.9979, 13122.1393},
        {204.39771, 204.47268, 204.466252, 204.379917, 204.215164, 203.973483, 203.656365, 203.2653, 202.801777, 202.267287, 201.66332, 200.991366, 200.252914, 199.449455, 198.58248, 197.653477, 196.663938, 195.615352, 194.509209, 193.346999, 192.130212, 190.860339, 189.538869, 188.167293, 186.7471, 185.279781, 183.766826, 182.209724, 180.609966, 178.969042, 177.288441, 175.569655, 173.814172, 172.023484, 170.207757, 168.36396, 166.488959, 164.58396, 162.650174, 160.688807, 158.70107, 156.688169, 154.651314, 152.591713, 150.510574, 148.409106, 146.288517, 144.150017, 141.994812, 139.824112, 137.639126, 135.432112, 133.213423, 130.98446, 128.746625, 126.501323, 124.249954, 121.993921, 119.734626, 117.473473, 115.211863, 112.9512, 110.692884, 108.438319, 106.188908, 103.946052, 101.711155, 99.6204738, 97.8070308, 95.9940969, 94.1743324, 92.3403975, 90.4849526, 88.6006578, 86.6801734, 84.7161596, 82.7012767, 80.628185, 78.4895446, 76.2780159, 73.9862591, 71.6069344, 69.1327021, 66.5562224, 62.9861655, 58.8757802, 54.6889402, 50.446993, 46.171286, 41.8831666, 37.6039822, 33.3550803, 29.1578083, 25.0335137, 21.0035438, 17.089246, 13.3119679, 9.69305682, 6.25386019, 3.01572544, 0, 2.61258858, 4.69284531, 6.28462698, 7.43179036, 8.17819223, 8.56768938, 8.64413857, 8.45139659, 8.03332021, 7.43376622, 6.6965914, 5.86565252, 4.98480636, 4.09790971, 3.24881933, 2.48139202, 2.52472152, 3.52055562, 3.96281031, 4.04217885, 3.94935453, 3.87503062, 4.00990039, 4.54465712, 5.66999408, 7.57660456, 10.4551818, 14.4964191, 19.8910098, 27.6791302, 36.8905133, 47.4610075, 59.326461, 70.9549175, 83.0161275, 96.5712565, 111.996494, 129.668031, 149.962056, 173.254759, 199.571537, 229.746136, 264.223379, 303.448089, 347.86509, 397.016162, 452.474932, 514.911981, 584.997894, 663.403254, 751.083563, 848.07625, 954.220128, 1069.92452, 1196.76354, 1335.78833, 1488.05003, 1653.54576, 1834.62025, 2034.76812, 2258.55697, 2503.11655, 2765.80293, 3046.30703, 3348.27607, 3673.54126, 4024.09107, 4401.79751, 4807.95585, 5244.1876, 5709.53549, 6208.59652, 6743.88484, 7317.9146, 7933.19993, 8813.3693, 9684.54397, 10493.9595, 11188.8515, 11716.4555, 11784.2, 11738.999, 11630.8625, 11645.3473, 11778.8094, 11971.1272, 12221.4806, 12529.0497, 12893.0144, 13312.5548, 13825.1807, 14425.6943, 15060.1585, 15674.6365, 16215.1914, 16627.8863, 16858.7843, 15101.0813, 12382.9869},
        {208.276248, 208.311333, 208.263147, 208.133265, 207.923263, 207.634714, 207.269193, 206.828275, 206.313534, 205.726546, 205.068884, 204.342124, 203.54784, 202.687607, 201.762998, 200.77559, 199.726956, 198.618671, 197.45231, 196.229448, 194.951658, 193.620516, 192.237597, 190.804474, 189.322723, 187.793918, 186.219634, 184.601445, 182.940927, 181.239653, 179.499198, 177.721138, 175.907046, 174.058497, 172.191035, 170.298956, 168.376389, 166.424455, 164.444274, 162.43697, 160.403662, 158.345472, 156.263522, 154.158932, 152.032824, 149.886319, 147.720539, 145.536605, 143.335639, 141.11876, 138.887092, 136.62072, 134.342257, 132.053279, 129.755364, 127.45009, 125.139033, 122.823771, 120.505881, 118.186941, 115.868527, 113.552218, 111.239589, 108.93222, 106.631686, 104.339565, 102.057435, 99.9257212, 98.0815155, 96.2400308, 94.3938435, 92.5355299, 90.6576662, 88.7528289, 86.8135941, 84.8325382, 82.8022376, 80.7152685, 78.5642073, 76.3416302, 74.0401136, 71.6522337, 69.170567, 66.5876897, 63.0032758, 58.8751338, 54.6718041, 50.4148395, 46.1257926, 41.8262162, 37.537663, 33.2816856, 29.0798368, 24.9536692, 20.9247357, 17.0145889, 13.2447815, 9.63686615, 6.21239567, 2.99292272, 0, 3.03980786, 5.47060778, 7.34262983, 8.70610409, 9.61126066, 10.1083296, 10.247541, 10.079125, 9.65331152, 9.02033081, 8.23041288, 7.33378783, 6.38068573, 5.42133667, 4.50597073, 3.68481799, 3.49506592, 4.05650995, 4.31792891, 4.43390092, 4.55900406, 4.84781641, 5.45491609, 6.53488118, 8.24228977, 10.73172, 14.1577499, 18.6749575, 24.4379211, 32.128227, 41.1802086, 51.5903414, 63.355101, 75.3306381, 88.0835906, 102.484683, 118.872489, 137.585581, 158.962533, 183.341916, 210.739895, 241.915577, 277.270616, 317.206665, 362.125377, 410.256943, 464.717344, 526.451101, 596.402733, 675.516758, 766.079293, 866.053533, 973.792576, 1089.15799, 1215.09111, 1353.15054, 1504.89486, 1671.76473, 1855.46348, 2057.82307, 2282.27593, 2526.76962, 2788.97131, 3068.75788, 3369.75049, 3693.85161, 4043.44108, 4420.54514, 4825.4397, 5259.39076, 5720.11312, 6212.97032, 6740.45762, 7305.07027, 7909.30355, 8771.2961, 9623.98893, 10415.9665, 11095.8131, 11612.1133, 11679.7048, 11636.7497, 11532.0096, 11546.0386, 11675.4304, 11861.865, 12104.6819, 12403.2202, 12756.8195, 13164.8189, 13158.6324, 13641.9784, 14411.7053, 15264.6618, 15997.6964, 16407.6575, 16291.3936, 14416.2533, 11694.6601},
        {212.481508, 212.474547, 212.382267, 212.206332, 211.94841, 211.610165, 211.193263, 210.699368, 210.130147, 209.487265, 208.772387, 207.987179, 207.133305, 206.212433, 205.226226, 204.176351, 203.064473, 201.892256, 200.661368, 199.373473, 198.030236, 196.633323, 195.184399, 193.68513, 192.137182, 190.542219, 188.901907, 187.217912, 185.491898, 183.725532, 181.920478, 180.078403, 178.200972, 176.289849, 174.366355, 172.421913, 170.447715, 168.444786, 166.414155, 164.356848, 162.273895, 160.166322, 158.035157, 155.881428, 153.706162, 151.510386, 149.29513, 147.061419, 144.810282, 142.542747, 140.25984, 137.928667, 135.584914, 133.230343, 130.866718, 128.495801, 126.119357, 123.739146, 121.356934, 118.974482, 116.593554, 114.215912, 111.84332, 109.477541, 107.120338, 104.773473, 102.438711, 100.260572, 98.3801549, 96.5047887, 94.6269822, 92.739244, 90.8340825, 88.9040062, 86.9415234, 84.9391427, 82.8893724, 80.7847211, 78.6176972, 76.3808091, 74.0665653, 71.6674742, 69.1760443, 66.5847841, 62.9839718, 58.8362083, 54.6149382, 50.3419491, 46.0390287, 41.7279643, 37.4305436, 33.1685542, 28.9637835, 24.8380192, 20.8130487, 16.9106596, 13.1526396, 9.56077602, 6.15685654, 2.96266869, 0, 3.60891903, 6.50677425, 8.75208325, 10.4033636, 11.5191329, 12.1579086, 12.3782085, 12.2385499, 11.7974506, 11.1134281, 10.245, 9.25068377, 8.18899707, 7.11845747, 6.09758253, 5.18488982, 4.65046592, 4.58305255, 4.56261835, 4.69301712, 5.07810263, 5.82172868, 7.02774906, 8.80001756, 11.242388, 14.458714, 18.5528496, 23.6286484, 29.7899643, 37.2246787, 45.9218076, 55.9599965, 67.4178908, 79.6603347, 93.1228744, 108.431403, 125.878706, 145.75757, 168.360781, 193.981123, 222.657495, 255.013841, 291.392622, 332.136298, 377.587328, 424.463462, 477.638048, 538.359724, 607.877128, 687.438897, 780.749339, 883.600046, 992.807935, 1107.67402, 1232.571, 1369.5945, 1520.84015, 1689.34537, 1876.04992, 2080.86615, 2305.80061, 2549.9107, 2811.35261, 3090.28933, 3390.28439, 3713.34041, 4062.36281, 4439.58829, 4843.94329, 5276.22674, 5732.50831, 6219.04372, 6738.26893, 7292.61989, 7884.53257, 8723.90446, 9553.84459, 10324.9235, 10987.7118, 11492.78, 11566.2487, 11532.7717, 11439.1126, 11457.4743, 11584.967, 11766.4346, 12001.6008, 12290.189, 12631.9231, 13026.5266, 12461.8239, 12818.1804, 13732.5268, 14841.7936, 15782.9115, 16192.8111, 15708.4229, 13769.2912, 11063.5486},
        {216.994077, 216.940148, 216.798998, 216.572374, 216.262025, 215.869701, 215.39715, 214.84612, 214.218361, 213.515622, 212.739652, 211.892198, 210.97501, 209.989838, 208.938429, 207.822532, 206.643897, 205.404272, 204.105406, 202.749048, 201.336947, 199.870851, 198.352509, 196.78367, 195.166084, 193.501498, 191.791662, 190.038324, 188.243234, 186.408139, 184.53479, 182.624934, 180.680321, 178.7027, 176.718689, 174.717079, 172.686362, 170.62748, 168.541374, 166.428985, 164.291257, 162.129129, 159.943545, 157.735446, 155.505773, 153.255469, 150.985474, 148.696731, 146.390182, 144.066768, 141.727431, 139.326692, 136.91292, 134.488063, 132.054068, 129.612884, 127.166459, 124.716741, 122.265678, 119.815218, 117.367308, 114.923898, 112.486935, 110.058367, 107.640142, 105.234209, 102.842514, 100.613303, 98.6913052, 96.7767458, 94.8620891, 92.9397994, 91.0023408, 89.0421778, 87.0517745, 85.0235954, 82.9501046, 80.8237665, 78.6370454, 76.3824055, 74.0523112, 71.6392268, 69.1356164, 66.5339445, 62.9158092, 58.7482667, 54.509321, 50.2209959, 45.9053151, 41.584302, 37.2799805, 33.0143741, 28.8095064, 24.6874013, 20.6700822, 16.7795728, 13.0378969, 9.46707799, 6.08913977, 2.92610589, 0, 4.24195007, 7.65995769, 10.3216406, 12.2946165, 13.6465031, 14.4449183, 14.7574797, 14.651805, 14.195512, 13.4562185, 12.5015421, 11.3991006, 10.2165118, 9.02139338, 7.88136305, 6.86403858, 5.93598727, 5.15039107, 4.80343418, 4.94108067, 5.60929462, 6.85404009, 8.72128116, 11.2569819, 14.5071064, 18.5176187, 23.3344829, 29.003663, 35.5711232, 42.6648743, 50.9020831, 60.4540995, 71.4922738, 83.945218, 98.1056512, 114.331022, 132.865503, 153.953264, 177.838477, 204.765313, 234.81067, 268.436902, 305.920909, 347.53959, 393.569844, 439.248964, 491.153357, 550.819821, 619.785157, 699.586164, 795.223177, 900.536253, 1010.97598, 1125.35046, 1249.32266, 1385.47786, 1536.40132, 1706.65583, 1896.39972, 2103.634, 2328.84908, 2572.33916, 2832.86325, 3110.92322, 3409.97411, 3732.11541, 4080.80471, 4458.49359, 4862.65397, 5293.57454, 5745.62922, 6225.93146, 6736.81738, 7280.62312, 7859.68479, 8673.87399, 9478.60752, 10226.8376, 10871.5166, 11365.5966, 11448.9201, 11429.6225, 11351.9887, 11377.7583, 11504.724, 11681.739, 11909.0428, 12186.8752, 12515.4757, 12895.0839, 11795.8801, 12029.7234, 13080.323, 14431.3878, 15566.6268, 15969.7492, 15124.4639, 13158.9183, 10488.4711},
        {221.794543, 221.685964, 221.488728, 221.204645, 220.835526, 220.383184, 219.849428, 219.236072, 218.544924, 217.777798, 216.936503, 216.022852, 215.038656, 213.985725, 212.865871, 211.680905, 210.432639, 209.122884, 207.75345, 206.32615, 204.842793, 203.305193, 201.715159, 200.074503, 198.385037, 196.648571, 194.866917, 193.041885, 191.175288, 189.268936, 187.324641, 185.344214, 183.329466, 181.282208, 179.23301, 177.1687, 175.075758, 172.955062, 170.807491, 168.633922, 166.435234, 164.212304, 161.966011, 159.697231, 157.406844, 155.095728, 152.764759, 150.414816, 148.046778, 145.661521, 143.259925, 140.785534, 138.297801, 135.798848, 133.290796, 130.775765, 128.255876, 125.733249, 123.210007, 120.688269, 118.170157, 115.65779, 113.153291, 110.65878, 108.176377, 105.708203, 103.25638, 100.972187, 99.0033224, 97.0442777, 95.0875048, 93.1254552, 91.1505804, 89.1553317, 87.1321607, 85.0735189, 82.9718577, 80.8196285, 78.609283, 76.3332724, 73.9840484, 71.5540624, 69.0357657, 66.4216101, 62.7863436, 58.6005722, 54.3459314, 50.0446541, 45.7189729, 41.3911205, 37.0833297, 32.8178331, 28.6168634, 24.5026533, 20.4974356, 16.623443, 12.9029081, 9.35806363, 6.01114233, 2.88437689, 0, 4.86092902, 8.78877106, 11.8599552, 14.1509104, 15.7380658, 16.6978504, 17.1066933, 17.0410235, 16.5772701, 15.7918622, 14.7612287, 13.5617988, 12.2700014, 10.9622657, 9.71502064, 8.60469535, 7.29669569, 5.80873313, 5.14693195, 5.29964481, 6.25522438, 8.00202331, 10.5283943, 13.8226899, 17.8732629, 22.6684659, 28.1966516, 34.4461726, 41.4053816, 48.1452027, 55.9078077, 64.9567776, 75.5556936, 88.1864987, 103.003593, 120.103148, 139.683235, 161.941927, 187.077298, 215.287419, 246.685754, 281.580736, 320.186991, 362.719145, 409.391826, 454.226694, 505.179583, 564.013361, 632.490895, 712.375055, 809.630287, 916.682616, 1028.00651, 1142.06518, 1265.46558, 1401.15825, 1552.09379, 1724.0643, 1916.533, 2125.86325, 2351.13944, 2593.85437, 2853.41967, 3130.68139, 3428.91601, 3750.2844, 4098.71523, 4476.82769, 4880.75906, 5310.31315, 5758.384, 6232.74825, 6735.60156, 7269.13959, 7835.55802, 8623.88428, 9402.77432, 10127.7158, 10754.1965, 11237.7042, 11332.8077, 11329.8594, 11270.455, 11304.9948, 11432.0061, 11604.681, 11823.8135, 12090.1977, 12404.6275, 12767.8971, 11221.9259, 11352.0304, 12512.794, 14058.8001, 15344.6322, 15724.8738, 14554.1084, 12583.858, 9968.24625},
        {226.661734, 226.468407, 226.188535, 225.823835, 225.376026, 224.846825, 224.237951, 223.551122, 222.788055, 221.950468, 221.040081, 220.058609, 219.007772, 217.889288, 216.704874, 215.456249, 214.14513, 212.773235, 211.342283, 209.853992, 208.310079, 206.712262, 205.06226, 203.36179, 201.612571, 199.81632, 197.974756, 196.089596, 194.162558, 192.195361, 190.189722, 188.14736, 186.069992, 183.959336, 181.840325, 179.702581, 177.535448, 175.33989, 173.116872, 170.867359, 168.592316, 166.292706, 163.969496, 161.623648, 159.256129, 156.867902, 154.459933, 152.033186, 149.588625, 147.127216, 144.649922, 142.100617, 139.538595, 136.966058, 134.385209, 131.798249, 129.207382, 126.614809, 124.022732, 121.433355, 118.848879, 116.271507, 113.703441, 111.146883, 108.604035, 106.077101, 103.568281, 101.228958, 99.206929, 97.1967281, 95.1908868, 93.1819369, 91.1624098, 89.1248375, 87.0617514, 84.9656834, 82.8291651, 80.6447281, 78.4049042, 76.102225, 73.7292222, 71.2784275, 68.7423726, 66.1135891, 62.4678782, 58.2754191, 54.0174092, 49.71613, 45.3938627, 41.0728888, 36.7754895, 32.5239462, 28.3405403, 24.2475531, 20.267266, 16.4219603, 12.7339173, 9.22541845, 5.91874505, 2.83617845, 0, 5.20537319, 9.42330587, 12.7346016, 15.2200639, 16.9604964, 18.0367025, 18.5294859, 18.5196501, 18.0879985, 17.3153349, 16.2824627, 15.0701855, 13.7593067, 12.4306301, 11.1649591, 10.0430972, 8.59814118, 6.83265072, 6.06629013, 6.2624972, 7.38470972, 9.39636548, 12.2609023, 15.9417579, 20.4023701, 25.6061768, 31.5166156, 38.0971245, 45.3111411, 51.9747625, 59.6194589, 68.5528704, 79.0826368, 91.9833515, 107.329178, 125.069758, 145.372647, 168.405398, 194.335566, 223.330704, 255.173869, 290.534134, 329.65428, 372.77709, 420.145346, 467.039841, 519.905843, 580.226631, 649.485484, 729.165683, 823.701801, 928.018684, 1038.44694, 1154.1484, 1280.06537, 1418.54484, 1571.93379, 1744.72991, 1936.6407, 2145.02746, 2369.06633, 2610.52518, 2869.53519, 3147.11341, 3445.77791, 3767.35778, 4115.23032, 4491.626, 4893.09847, 5319.41254, 5763.62528, 6233.24114, 6730.34679, 7257.02891, 7815.37417, 8589.17121, 9353.37926, 10064.6595, 10679.6732, 11155.0815, 11252.4737, 11253.6307, 11198.8161, 11233.6409, 11356.6684, 11522.5671, 11732.0948, 11986.0097, 12285.0697, 12630.0328, 11175.63, 11316.0848, 12427.7303, 13886.8997, 15069.9262, 15353.1429, 14112.8831, 12040.256, 9502.11885},
        {231.819896, 231.532695, 231.161601, 230.708213, 230.174135, 229.560967, 228.870311, 228.103769, 227.262943, 226.349434, 225.364843, 224.310774, 223.188826, 222.000602, 220.747704, 219.431733, 218.05429, 216.616979, 215.121399, 213.569153, 211.961843, 210.30107, 208.588435, 206.825542, 205.01399, 203.155382, 201.25132, 199.303405, 197.313238, 195.282422, 193.212558, 191.105248, 188.962094, 186.784697, 184.590781, 182.373547, 180.126004, 177.849229, 175.544302, 173.212299, 170.854301, 168.471385, 166.06463, 163.635115, 161.183917, 158.712115, 156.220788, 153.711014, 151.183871, 148.640439, 146.081795, 143.454401, 140.815135, 138.166262, 135.510043, 132.848741, 130.184617, 127.519936, 124.856959, 122.197948, 119.545167, 116.900877, 114.267341, 111.646822, 109.041581, 106.453882, 103.885987, 101.488481, 99.4083897, 97.3420692, 95.282167, 93.2213303, 91.1522063, 89.0674424, 86.9596857, 84.8215836, 82.6457833, 80.424932, 78.1516771, 75.8186658, 73.4185453, 70.943963, 68.387566, 65.7420016, 62.0871549, 57.8912528, 53.6334933, 49.3361445, 45.0214747, 40.7117519, 36.4292443, 32.1962201, 28.0349475, 23.9676945, 20.0167293, 16.2043202, 12.5527352, 9.08424255, 5.82111035, 2.78560678, 0, 5.41772189, 9.81891811, 13.2869075, 15.9050087, 17.7565407, 18.9248221, 19.4931719, 19.5449088, 19.1633515, 18.431819, 17.4336299, 16.2521031, 14.9705573, 13.6723114, 12.4406842, 11.3589944, 9.88141765, 8.00139508, 7.22530758, 7.50165754, 8.77894734, 11.0056794, 14.130356, 18.1014797, 22.8675527, 28.3770776, 34.5785566, 41.4204921, 48.8513866, 55.5253214, 63.1598409, 72.0917739, 82.6579492, 95.8225107, 111.60863, 129.872193, 150.761835, 174.426189, 201.01389, 230.673572, 262.895488, 298.687031, 338.325647, 382.088785, 430.253892, 479.680809, 534.918993, 597.100294, 667.356563, 746.819647, 838.048263, 939.003449, 1048.24142, 1165.76231, 1294.51418, 1436.12156, 1592.20895, 1765.65028, 1956.53668, 2163.5963, 2386.36144, 2626.65618, 2885.23606, 3163.22679, 3462.38133, 3784.16473, 4131.27966, 4505.51201, 4904.10976, 5326.88781, 5767.57426, 6233.00721, 6725.1085, 7245.8, 7797.00354, 8557.25061, 9307.70102, 10006.1242, 10610.2897, 11077.9668, 11177.4121, 11182.25, 11131.4697, 11166.2048, 11285.0125, 11443.8898, 11643.5219, 11884.5941, 12167.7915, 12493.7993, 11265.7949, 11447.1463, 12472.1176, 13774.9729, 14789.9764, 14951.3922, 13693.4843, 11529.9488, 9088.39306},
        {237.302875, 236.915102, 236.446219, 235.897705, 235.271041, 234.567707, 233.789185, 232.936955, 232.012497, 231.017292, 229.952821, 228.820564, 227.622001, 226.358614, 225.031883, 223.643288, 222.19431, 220.686429, 219.121127, 217.499883, 215.824179, 214.095494, 212.31531, 210.485107, 208.606365, 206.680565, 204.709189, 202.693715, 200.635626, 198.5364, 196.39752, 194.220465, 192.006717, 189.757755, 187.483596, 185.181272, 182.847715, 180.484128, 178.091716, 175.671681, 173.225227, 170.753558, 168.257877, 165.739389, 163.199296, 160.638802, 158.059111, 155.461427, 152.846952, 150.216891, 147.572448, 144.863899, 142.144479, 139.416495, 136.682255, 133.944065, 131.204234, 128.465069, 125.728878, 122.997969, 120.274648, 117.561223, 114.860003, 112.173294, 109.503404, 106.85264, 104.223311, 101.764517, 99.6218329, 97.4948179, 95.3762634, 93.258961, 91.1357022, 88.9992787, 86.842482, 84.6581038, 82.4389355, 80.1777687, 77.8673951, 75.5006063, 73.0701937, 70.568949, 67.9896638, 65.3251297, 61.6621525, 57.4655729, 53.2111208, 48.9209886, 44.617369, 40.3224544, 36.0584374, 31.8475107, 27.7118666, 23.6736979, 19.7551971, 15.9785566, 12.3659691, 8.93962719, 5.72172331, 2.73445006, 0, 5.46818055, 9.92094066, 13.4419578, 16.1149093, 18.0234728, 19.2513256, 19.8821452, 19.999609, 19.6873945, 19.0291792, 18.1086404, 17.0094557, 15.8153025, 14.6098581, 13.4768002, 12.499806, 11.1125806, 9.3059483, 8.63376227, 9.04042019, 10.4703197, 12.8678584, 16.1774339, 20.3434439, 25.3102861, 31.0223579, 37.4240571, 44.4597812, 52.073928, 58.8594718, 66.6077554, 75.6686032, 86.39184, 99.8113129, 115.934835, 134.587814, 155.914868, 180.060614, 207.16967, 237.386654, 270.022004, 306.308399, 346.553666, 391.065632, 440.152123, 492.282878, 550.063334, 614.260844, 685.642754, 764.976416, 852.904302, 950.471265, 1058.46246, 1177.72123, 1309.20966, 1453.83652, 1612.51059, 1786.45685, 1976.196, 2181.90378, 2403.62014, 2643.03148, 2901.33703, 3179.69598, 3479.1997, 3800.97071, 4147.02617, 4518.72053, 4914.12786, 5333.17779, 5770.73758, 6232.57809, 6720.38133, 7235.82931, 7780.60405, 8527.48759, 9264.2869, 9949.90897, 10543.2608, 11003.2494, 11104.9314, 11113.6311, 11067.0484, 11101.7532, 11216.3095, 11368.0854, 11557.7047, 11785.7911, 12052.9683, 12359.86, 11454.6772, 11700.3836, 12613.8822, 13712.0762, 14511.8684, 14530.162, 13283.86, 11052.3401, 8725.77511},
        {243.144518, 242.6519, 242.080681, 241.432233, 240.70793, 239.909145, 239.037251, 238.09362, 237.079627, 235.996643, 234.846043, 233.629199, 232.347483, 231.002271, 229.594933, 228.126843, 226.599375, 225.013902, 223.371796, 221.67443, 219.923178, 218.119413, 216.264507, 214.359834, 212.406767, 210.406679, 208.360943, 206.270931, 204.138018, 201.963576, 199.748978, 197.495597, 195.204806, 192.877979, 190.517989, 188.125429, 185.700869, 183.245635, 180.761051, 178.248441, 175.70913, 173.144441, 170.5557, 167.94423, 165.311356, 162.658402, 159.986692, 157.297551, 154.592303, 151.872273, 149.138784, 146.346129, 143.543685, 140.733794, 137.918801, 135.101049, 132.282883, 129.466646, 126.654682, 123.849335, 121.052949, 118.267868, 115.496436, 112.740996, 110.003892, 107.287469, 104.59407, 102.070825, 99.8613872, 97.669491, 95.488094, 93.3101543, 91.1286298, 88.9364787, 86.7266587, 84.4921281, 82.2258447, 79.9207665, 77.5698515, 75.1660578, 72.7023433, 70.171666, 67.5669839, 64.881255, 61.2108502, 57.0158794, 52.767229, 48.4869532, 44.1971058, 39.9197407, 35.6769121, 31.4906738, 27.3830798, 23.3761841, 19.4920406, 15.7527034, 12.1802264, 8.79666361, 5.62406896, 2.68449643, 0, 5.3269546, 9.67470642, 13.1248376, 15.7589304, 17.6585668, 18.9053292, 19.5807995, 19.7665601, 19.544193, 18.9952804, 18.2014045, 17.2441475, 16.2050914, 15.1658186, 14.207911, 13.4129509, 12.2576857, 10.7372925, 10.3014322, 10.9020795, 12.4912092, 15.020796, 18.4428145, 22.7092394, 27.7720455, 33.5832074, 40.0946998, 47.2584975, 55.0265751, 62.0398066, 70.0420043, 79.3784733, 90.3945185, 104.057094, 120.40068, 139.293985, 160.895817, 185.364987, 212.860306, 243.540583, 276.72481, 313.667213, 354.690914, 400.119035, 450.2747, 504.979326, 565.183167, 631.33477, 703.882684, 783.275457, 868.504547, 963.256483, 1070.18255, 1190.83949, 1324.54944, 1471.6378, 1632.43001, 1806.78105, 1995.59369, 2200.28393, 2421.4378, 2660.4352, 2918.65286, 3197.19542, 3496.70645, 3818.04117, 4162.63279, 4531.48637, 4923.48768, 5338.72128, 5773.62187, 6232.48541, 6716.65989, 7227.49328, 7766.33359, 8499.24726, 9221.6842, 9893.81276, 10475.8013, 10927.8183, 11032.3399, 11045.6879, 11004.1846, 11039.3527, 11149.8308, 11294.5903, 11474.253, 11689.4407, 11940.7753, 12228.8785, 11704.5333, 12030.9651, 12820.9505, 13687.2659, 14242.6879, 14099.9931, 12871.958, 10606.8339, 8412.97125},
        {249.378672, 248.779362, 248.10328, 247.351724, 246.525991, 245.627379, 244.657185, 243.616706, 242.507241, 241.330086, 240.086538, 238.777896, 237.405457, 235.970518, 234.474377, 232.91833, 231.303676, 229.631712, 227.903736, 226.121044, 224.284935, 222.396705, 220.457652, 218.469074, 216.432268, 214.348532, 212.219162, 210.045457, 207.828713, 205.570229, 203.271301, 200.933228, 198.557306, 196.144833, 193.69318, 191.205691, 188.685755, 186.134798, 183.554245, 180.945519, 178.310047, 175.649252, 172.96456, 170.257396, 167.529185, 164.78135, 162.015318, 159.232513, 156.43436, 153.622284, 150.797709, 147.918107, 145.02981, 142.135194, 139.236638, 136.336519, 133.437214, 130.541103, 127.650561, 124.767968, 121.895699, 119.036135, 116.191651, 113.364625, 110.557436, 107.772461, 105.012077, 102.421165, 100.141181, 97.8806052, 95.6325767, 93.3902355, 91.1467215, 88.8951744, 86.6287342, 84.3405408, 82.023734, 79.6714536, 77.2768396, 74.8330317, 72.33317, 69.7703941, 67.1378441, 64.4286597, 60.7512267, 56.5596717, 52.3187553, 48.0503292, 43.7762453, 39.5183555, 35.2985115, 31.1385653, 27.0603687, 23.0857734, 19.2366314, 15.5347944, 12.0021144, 8.66044308, 5.53163236, 2.63753405, 0, 4.96424945, 9.02554828, 12.2606321, 14.7462364, 16.5590969, 17.7759492, 18.4735287, 18.7285713, 18.6178123, 18.2179875, 17.6058323, 16.8580825, 16.0514736, 15.2627412, 14.5686209, 14.0458482, 13.2827884, 12.2864097, 12.2380953, 13.10993, 14.8739984, 17.5023856, 20.9671761, 25.2404547, 30.2943063, 36.1008156, 42.6320675, 49.8601466, 57.7571377, 65.1289182, 73.5413897, 83.3164993, 94.7761939, 108.667191, 125.099052, 144.068067, 165.768753, 190.395624, 218.143198, 249.20599, 283.175297, 321.032444, 363.089965, 409.660399, 461.056283, 517.903435, 580.12279, 647.948564, 721.614975, 801.356239, 885.083626, 978.193454, 1084.47416, 1205.93142, 1340.93117, 1489.47351, 1651.55854, 1826.25432, 2014.70482, 2219.07078, 2440.4098, 2679.65146, 2937.99831, 3216.39956, 3515.37501, 3835.64157, 4178.26245, 4544.04433, 4932.52415, 5343.95709, 5776.73378, 6233.26082, 6714.43881, 7221.16836, 7754.35008, 8471.89473, 9178.44023, 9835.63465, 10405.1261, 10848.5626, 10956.9457, 10976.3344, 10941.5109, 10978.0701, 11084.8477, 11222.8407, 11392.7766, 11595.3831, 11831.3878, 12101.5181, 11977.6198, 12394.0595, 13061.2486, 13689.5985, 13989.5205, 13671.426, 12445.7263, 10192.834, 8148.68769},
        {256.310456, 255.605356, 254.823231, 253.965406, 253.033206, 252.027953, 250.950972, 249.803588, 248.587124, 247.302904, 245.952254, 244.536496, 243.056955, 241.514956, 239.911821, 238.248877, 236.527445, 234.748852, 232.91442, 231.025475, 229.083339, 227.089338, 225.044795, 222.951035, 220.809382, 218.621159, 216.387692, 214.110303, 211.790318, 209.429061, 207.027855, 204.588024, 202.110894, 199.597787, 197.046437, 194.460039, 191.841831, 189.193255, 186.515751, 183.810759, 181.079722, 178.324078, 175.545269, 172.744736, 169.92392, 167.08426, 164.227198, 161.354174, 158.46663, 155.566006, 152.653742, 149.686254, 146.710984, 143.730348, 140.746765, 137.762649, 134.780419, 131.80249, 128.831281, 125.869206, 122.918684, 119.98213, 117.061962, 114.160597, 111.28045, 108.42394, 105.593482, 102.931821, 100.578331, 98.2460149, 95.9281926, 93.6181839, 91.309309, 88.9948877, 86.6682399, 84.3226855, 81.9515446, 79.5481369, 77.1057826, 74.6178014, 72.0775133, 69.4782383, 66.8132962, 64.0760071, 60.3904082, 56.202745, 51.969711, 47.7128359, 43.4536498, 39.2136822, 35.0144632, 30.8775225, 26.8243898, 22.8765952, 19.0556683, 15.3831389, 11.880537, 8.56939236, 5.47123475, 2.60759403, 0, 3.86551936, 7.03699394, 9.57731911, 11.5493902, 13.0161026, 14.0403517, 14.6850328, 15.0130414, 15.0872726, 14.970622, 14.7259849, 14.4162566, 14.1043325, 13.853108, 13.7254785, 13.7843392, 13.9284324, 14.197203, 15.0248452, 16.4390787, 18.4676231, 21.1381982, 24.4785236, 28.516319, 33.2793041, 38.7951985, 45.0917221, 52.1965944, 60.1375351, 68.2022111, 77.4297475, 88.0698801, 100.372344, 114.441764, 130.600431, 149.209332, 170.561737, 194.950915, 222.670136, 254.012668, 289.721751, 329.503735, 373.563854, 422.107336, 475.339415, 532.766307, 595.467011, 663.821512, 738.209794, 819.011842, 904.969154, 1000.1028, 1106.90994, 1226.97678, 1360.02955, 1506.62958, 1667.3382, 1842.52961, 2032.99478, 2239.72876, 2463.7565, 2705.28736, 2964.25486, 3241.17654, 3537.55944, 3854.45639, 4193.40054, 4555.56925, 4940.37872, 5348.24135, 5780.57591, 6236.63362, 6716.47853, 7220.1747, 7747.78616, 8442.62472, 9125.69476, 9761.24837, 10313.5377, 10746.8148, 10867.0056, 10902.2393, 10884.6221, 10927.5436, 11030.9353, 11160.8607, 11318.945, 11506.813, 11726.0897, 11978.4002, 12225.7586, 12744.8377, 13321.6309, 13742.1318, 13792.3336, 13258.2297, 11925.8135, 9791.08948, 7939.13674},
        {263.496459, 262.688329, 261.802186, 260.83941, 259.801381, 258.689479, 257.505083, 256.249574, 254.924331, 253.530734, 252.070163, 250.543998, 248.953618, 247.300405, 245.585737, 243.810994, 241.977557, 240.086805, 238.140117, 236.138875, 234.084458, 231.978245, 229.821617, 227.615953, 225.362633, 223.063038, 220.718547, 218.330539, 215.900396, 213.429496, 210.919219, 208.370946, 205.786057, 203.16593, 200.510037, 197.820753, 195.100474, 192.350644, 189.572703, 186.768094, 183.938258, 181.084638, 178.208674, 175.31181, 172.395486, 169.461145, 166.510228, 163.544178, 160.564436, 157.572444, 154.569643, 151.510292, 148.444039, 145.373349, 142.300686, 139.228517, 136.159304, 133.095514, 130.03961, 126.994057, 123.96132, 120.943864, 117.944154, 114.964653, 112.007828, 109.076142, 106.17206, 103.436179, 101.005776, 98.5984326, 96.2076584, 93.8269639, 91.4498593, 89.0698548, 86.6804605, 84.2751869, 81.8475441, 79.3910424, 76.899192, 74.3655032, 71.7834862, 69.1466512, 66.4485085, 63.6825684, 59.9895348, 55.8072587, 51.5838413, 47.3404678, 43.0983239, 38.878595, 34.7024664, 30.5911236, 26.5657522, 22.6475374, 18.8576648, 15.2173199, 11.747688, 8.46995462, 5.40530519, 2.57492517, 0, 2.85380583, 5.20576041, 7.10608714, 8.60500941, 9.7527506, 10.5995341, 11.1955833, 11.5911216, 11.8363724, 11.981559, 12.0769049, 12.1726334, 12.318968, 12.566132, 12.9643488, 13.5638418, 14.5574224, 16.0156837, 17.6663286, 19.5901351, 21.867881, 24.5803444, 27.8083032, 31.6325354, 36.1338189, 41.3929318, 47.490652, 54.5077575, 62.5250262, 71.2997377, 81.3547845, 92.867994, 106.017194, 120.269244, 136.157457, 154.404733, 175.402191, 199.540948, 227.212122, 258.80683, 296.224353, 337.888639, 383.895732, 434.341671, 489.322501, 546.925312, 609.757336, 678.416851, 753.502139, 835.611478, 924.713514, 1022.80572, 1130.64515, 1249.22085, 1379.99548, 1524.21905, 1683.14159, 1858.71916, 2051.33525, 2260.60311, 2487.45507, 2731.40112, 2991.08257, 3266.58206, 3560.42453, 3874.01388, 4209.40199, 4568.16077, 4949.48604, 5353.91766, 5785.662, 6241.02834, 6719.28737, 7219.70977, 7741.56622, 8414.62075, 9075.02737, 9689.43343, 10224.4863, 10646.8333, 10775.7527, 10823.507, 10819.9018, 10868.2409, 10969.2369, 11092.9778, 11241.7072, 11417.6687, 11623.106, 11860.2626, 12429.564, 13038.4639, 13536.7233, 13774.1027, 13600.363, 12865.2648, 11418.5688, 9433.82399, 7769.81901},
        {270.751476, 269.845297, 268.859774, 267.796346, 266.656456, 265.441545, 264.153053, 262.792423, 261.361094, 259.860508, 258.292107, 256.657332, 254.957623, 253.194422, 251.36917, 249.483309, 247.538279, 245.535521, 243.476477, 241.362588, 239.195295, 236.976039, 234.706261, 232.387403, 230.020905, 227.60821, 225.150757, 222.649989, 220.107346, 217.524269, 214.9022, 212.24258, 209.54685, 206.816451, 204.052504, 201.256619, 198.430405, 195.575315, 192.692799, 189.78431, 186.851299, 183.895218, 180.917518, 177.91965, 174.903066, 171.869218, 168.819557, 165.755535, 162.678603, 159.590213, 156.491817, 153.335258, 150.172672, 147.006584, 143.839522, 140.674012, 137.512581, 134.357756, 131.212065, 128.078033, 124.958189, 121.855058, 118.771167, 115.709044, 112.671216, 109.660209, 106.67855, 103.864733, 101.353986, 98.8684167, 96.4017364, 93.9476582, 91.4998948, 89.052159, 86.5981634, 84.1316208, 81.6462439, 79.1357455, 76.5938383, 74.0142349, 71.3906482, 68.7167908, 65.9863755, 63.193115, 59.4940251, 55.3187658, 51.1070535, 46.8797534, 42.6577306, 38.4618502, 34.3129775, 30.2319774, 26.2397153, 22.3570562, 18.6048653, 15.0040077, 11.5753486, 8.33975315, 5.31808646, 2.53121369, 0, 2.30603887, 4.21698317, 5.7762519, 7.02726407, 8.01343867, 8.77819473, 9.36495123, 9.81712718, 10.1781416, 10.4914135, 10.8003618, 11.1484056, 11.5789638, 12.1354556, 12.8612998, 13.7999155, 15.3217927, 17.5170748, 19.6893774, 21.9522061, 24.4190665, 27.2034644, 30.4189053, 34.1788949, 38.5969389, 43.7865428, 49.8612124, 56.9344533, 65.1197712, 74.4794936, 85.1970695, 97.401358, 111.221218, 125.682043, 141.464426, 159.543214, 180.378307, 204.429604, 232.157003, 264.020403, 302.712666, 345.78113, 393.248809, 445.138716, 501.473867, 559.206653, 622.198366, 691.239676, 767.121252, 850.633764, 942.850315, 1043.93394, 1153.15436, 1270.96586, 1400.24131, 1542.76769, 1700.33197, 1876.07751, 2070.12663, 2281.12198, 2509.99422, 2755.71479, 3016.0504, 3290.76206, 3582.99003, 3894.32345, 4227.08381, 4583.05014, 4961.31603, 5362.49412, 5792.67232, 6246.14561, 6721.63133, 7217.84681, 7733.50938, 8389.21919, 9031.34005, 9628.1186, 10147.8015, 10558.6353, 10688.1199, 10739.0799, 10740.082, 10790.2772, 10890.2925, 11011.3448, 11155.5317, 11324.9506, 11521.699, 11747.8746, 12557.1678, 13226.7619, 13656.2793, 13745.3421, 13393.5727, 12500.5935, 10966.0266, 9136.03362, 7633.65619},
        {277.756863, 276.761284, 275.685397, 274.530667, 273.298561, 271.990543, 270.608079, 269.152637, 267.625681, 266.028677, 264.363092, 262.630391, 260.832039, 258.969504, 257.04425, 255.057744, 253.011451, 250.906838, 248.74537, 246.528513, 244.257733, 241.934495, 239.560266, 237.136512, 234.664698, 232.146291, 229.582755, 226.975558, 224.326164, 221.636041, 218.906652, 216.139466, 213.335946, 210.49756, 207.624481, 204.718847, 201.782815, 198.817891, 195.825583, 192.807399, 189.764848, 186.699435, 183.612671, 180.506061, 177.381115, 174.23934, 171.082243, 167.911332, 164.728116, 161.534102, 158.330798, 155.068493, 151.801025, 148.53101, 145.261066, 141.993813, 138.731869, 135.47785, 132.234376, 129.004064, 125.789533, 122.5934, 119.418285, 116.266804, 113.141576, 110.045219, 106.980352, 104.083552, 101.488181, 98.9205755, 96.3746685, 93.8443942, 91.3236866, 88.8064797, 86.2867075, 83.758304, 81.2152032, 78.6513391, 76.0606458, 73.4370571, 70.7745071, 68.0669299, 65.3082594, 62.4924295, 58.7885438, 54.6203006, 50.4213123, 46.2122804, 42.0139061, 37.8468908, 33.7319358, 29.6897425, 25.7410122, 21.9064463, 18.2067461, 14.662613, 11.2947482, 8.12385319, 5.17062925, 2.45577774, 0, 3.28707139, 6.00526952, 8.21046446, 9.95852628, 11.3053251, 12.3067309, 13.0186138, 13.4968439, 13.7972913, 13.975826, 14.0883181, 14.1906377, 14.3386548, 14.5882396, 14.9952621, 15.6155923, 16.5985543, 18.0133687, 19.7045218, 21.7479093, 24.2194273, 27.1949717, 30.7504383, 34.9617231, 39.904722, 45.6553309, 52.2894458, 59.8829624, 68.5117769, 77.9621157, 88.7057561, 100.905495, 114.724129, 129.416925, 145.600444, 164.133257, 185.450419, 209.986985, 238.178011, 270.458552, 309.158851, 352.241981, 399.772201, 451.813768, 508.430942, 566.690455, 630.403472, 700.383633, 777.444576, 862.399941, 956.561633, 1059.63603, 1170.7547, 1290.24912, 1420.90057, 1564.39044, 1722.40011, 1897.58215, 2090.42601, 2300.36166, 2528.95264, 2774.80301, 3035.84558, 3311.77268, 3605.25813, 3917.60717, 4250.73979, 4606.12063, 4982.96035, 5381.74459, 5807.2968, 6255.09746, 6724.13066, 7213.38048, 7721.83101, 8375.25246, 9014.14627, 9605.80001, 10117.5013, 10516.5376, 10622.7689, 10649.1954, 10626.2881, 10666.3599, 10767.4858, 10893.6398, 11044.6021, 11220.1526, 11420.0714, 11644.1385, 12506.8736, 13163.2984, 13535.8419, 13546.9329, 13119.0003, 12174.473, 10635.7799, 8937.43128, 7528.00504},
        {284.741703, 283.664986, 282.506566, 281.267939, 279.950602, 278.556051, 277.085784, 275.541296, 273.924084, 272.235646, 270.477477, 268.651074, 266.757934, 264.799553, 262.777428, 260.693056, 258.547933, 256.343555, 254.081421, 251.763025, 249.389865, 246.963437, 244.485238, 241.956765, 239.379513, 236.754981, 234.084663, 231.370058, 228.612661, 225.81397, 222.97548, 220.098688, 217.185092, 214.236187, 211.250985, 208.232277, 205.182885, 202.104386, 198.998358, 195.866377, 192.710022, 189.530868, 186.330494, 183.110477, 179.872395, 176.617823, 173.34834, 170.065523, 166.770949, 163.466196, 160.152841, 156.77945, 153.40176, 150.022498, 146.644391, 143.270165, 139.902545, 136.54426, 133.198034, 129.866595, 126.552669, 123.258983, 119.988262, 116.743234, 113.526625, 110.341161, 107.189568, 104.206707, 101.524265, 98.8724523, 96.2454309, 93.6373611, 91.0424038, 88.4547196, 85.8684695, 83.2778141, 80.6769143, 78.0599308, 75.4210243, 72.7543558, 70.0540859, 67.3143754, 64.5293851, 61.6932758, 57.9872646, 53.8290833, 49.6462347, 45.459261, 41.2887046, 37.155108, 33.0790134, 29.0809632, 25.1814998, 21.4011655, 17.7605028, 14.2800539, 10.9803612, 7.88196713, 5.00541399, 2.37124416, 0, 4.56631707, 8.33524647, 11.3788636, 13.7692439, 15.5784628, 16.8785956, 17.7417178, 18.2399048, 18.445232, 18.4297748, 18.2656086, 18.0248089, 17.779451, 17.6016103, 17.5633623, 17.7367824, 17.9753017, 18.3197606, 19.3367872, 21.0516046, 23.4894358, 26.6755038, 30.6350319, 35.3932428, 40.9753599, 47.406606, 54.7122043, 62.9173778, 72.0473495, 81.4992055, 92.1566228, 104.233266, 117.942798, 132.94195, 149.722853, 168.926154, 190.932597, 216.122927, 244.877888, 277.578225, 315.805484, 358.374146, 405.430014, 457.118894, 513.586591, 572.483947, 637.07547, 708.130702, 786.419188, 872.710473, 968.198697, 1072.70986, 1185.75556, 1307.71397, 1440.73247, 1586.1641, 1745.3619, 1920.01962, 2111.07215, 2319.08272, 2546.09232, 2790.81169, 3052.08936, 3329.7795, 3625.44024, 3939.84755, 4274.08709, 4629.01512, 5004.35247, 5400.46222, 5820.54778, 6261.69497, 6723.28369, 7204.69388, 7705.30543, 8358.33939, 8995.87427, 9583.82969, 10088.1253, 10474.6807, 10552.9118, 10549.5778, 10497.6714, 10526.3011, 10630.4138, 10764.9079, 10926.7329, 11112.8378, 11320.1719, 11545.6844, 12375.5076, 12989.4748, 13317.8768, 13291.004, 12839.1473, 11892.5972, 10381.6444, 8794.53997, 7441.56955},
        {291.741164, 290.592618, 289.360279, 288.045693, 286.650406, 285.175963, 283.623912, 281.995797, 280.293165, 278.517562, 276.670534, 274.753627, 272.768387, 270.716359, 268.599091, 266.418128, 264.175015, 261.8713, 259.508528, 257.088244, 254.611996, 252.081329, 249.497789, 246.862922, 244.178274, 241.445391, 238.665819, 235.841104, 232.972793, 230.062431, 227.111563, 224.121738, 221.094499, 218.031393, 214.930705, 211.79568, 208.629602, 205.434124, 202.210897, 198.961574, 195.687806, 192.391245, 189.073542, 185.736351, 182.381322, 179.010108, 175.62436, 172.22573, 168.81587, 165.396433, 161.969069, 158.480186, 154.987879, 151.494996, 148.004388, 144.518903, 141.041391, 137.574702, 134.121684, 130.685188, 127.268063, 123.873158, 120.503323, 117.161407, 113.850259, 110.57273, 107.331669, 104.260469, 101.489389, 98.7520422, 96.0428145, 93.3560928, 90.686264, 88.0277149, 85.3748322, 82.7220029, 80.0636137, 77.3940513, 74.7077027, 71.9989546, 69.2621938, 66.4918072, 63.6821815, 60.8277036, 57.1231066, 52.9794238, 48.8172997, 44.6570714, 40.5190759, 36.4236504, 32.391132, 28.4418578, 24.5961648, 20.8743903, 17.2968712, 13.8839447, 10.655948, 7.63321802, 4.83609196, 2.28490692, 0, 5.72586038, 10.447869, 14.2527876, 17.2273776, 19.4584008, 21.0326188, 22.0367931, 22.5576854, 22.6820573, 22.4966705, 22.0882865, 21.543667, 20.9495736, 20.392768, 19.9600116, 19.7380663, 19.3191119, 18.7192557, 19.1439556, 20.5739916, 22.9901434, 26.3731909, 30.7039139, 35.9630925, 42.1315063, 49.1899352, 57.1191592, 65.899958, 75.5131115, 85.0084207, 95.6390032, 107.664933, 121.346283, 136.758199, 154.25322, 174.233196, 197.013679, 222.91022, 252.238372, 285.313686, 322.799404, 364.557569, 410.835707, 461.881343, 517.942003, 577.560541, 643.115321, 715.280038, 794.728383, 882.134051, 978.282665, 1083.59918, 1198.34484, 1323.08116, 1458.98281, 1606.94918, 1767.87964, 1942.36829, 2131.68918, 2337.44941, 2561.69421, 2803.91428, 3064.65615, 3344.20125, 3642.38177, 3959.23613, 4294.58739, 4648.41812, 5021.5005, 5414.16006, 5828.63401, 6263.0421, 6717.0481, 7190.31583, 7682.50906, 8332.31731, 8965.62222, 9547.33117, 10042.3515, 10415.5906, 10467.8516, 10437.549, 10359.9203, 10379.6956, 10488.7719, 10633.4247, 10807.832, 11006.172, 11222.623, 11451.3631, 12189.6119, 12744.7886, 13043.2683, 13011.4261, 12575.6368, 11662.2755, 10197.7173, 8693.335, 7366.11709},
        {298.907735, 297.70907, 296.42206, 295.048393, 293.589752, 292.047825, 290.424295, 288.72085, 286.939174, 285.080953, 283.147872, 281.141617, 279.063874, 276.916328, 274.700664, 272.418569, 270.071727, 267.661824, 265.190546, 262.659579, 260.070607, 257.425316, 254.725393, 251.972522, 249.168388, 246.314678, 243.413078, 240.465271, 237.472945, 234.437784, 231.361475, 228.245702, 225.092152, 221.902509, 218.678115, 215.420835, 212.132538, 208.81492, 205.469679, 202.098511, 198.703113, 195.285182, 191.846413, 188.388505, 184.913153, 181.422055, 177.916907, 174.399406, 170.871248, 167.33413, 163.78975, 160.17905, 156.565798, 152.953007, 149.343691, 145.740864, 142.147538, 138.566729, 135.001449, 131.454712, 127.929532, 124.428923, 120.955899, 117.513473, 114.104658, 110.732469, 107.399919, 104.240625, 101.385067, 98.5669007, 95.7806733, 93.0209324, 90.282225, 87.5590985, 84.8461002, 82.1377774, 79.4286772, 76.713347, 73.9863341, 71.2421857, 68.4754492, 65.6806717, 62.8524005, 59.985183, 56.2907903, 52.1716224, 48.0394876, 43.9143357, 39.8161167, 35.7647804, 31.7802766, 27.8825553, 24.0915662, 20.4272593, 16.9095844, 13.5584914, 10.3939302, 7.43585057, 4.70420242, 2.21893561, 0, 5.35677648, 9.78700298, 13.3724461, 16.1948725, 18.3360488, 19.8777415, 20.9017174, 21.4897431, 21.7235851, 21.68501, 21.4557847, 21.1176755, 20.7524492, 20.4418724, 20.2677116, 20.3117336, 20.3413228, 20.3933341, 21.2606365, 22.9576292, 25.4987111, 28.8982812, 33.1707386, 38.3304823, 44.3919113, 51.3694246, 59.2774212, 68.1303002, 77.9424605, 88.0364222, 99.3721522, 112.171613, 126.656768, 142.809087, 160.970778, 181.548183, 204.835412, 231.126576, 260.715786, 293.897152, 330.820707, 371.968489, 417.662799, 468.225935, 523.980198, 584.363231, 650.803155, 723.843432, 804.027527, 891.898904, 987.539545, 1092.51843, 1208.11809, 1335.2376, 1473.99304, 1624.85203, 1788.28221, 1964.47963, 2154.24521, 2358.67609, 2577.69252, 2813.86618, 3071.58581, 3352.33759, 3652.68944, 3971.46684, 4306.03316, 4654.83482, 5021.67934, 5407.34173, 5817.35641, 6247.0067, 6695.42008, 7161.72403, 7645.04603, 8275.53798, 8888.54684, 9450.444, 9927.60084, 10286.3887, 10333.5843, 10301.5502, 10225.0555, 10250.2653, 10367.8544, 10522.083, 10705.9467, 10912.4412, 11134.5623, 11365.3056, 12004.8492, 12515.2934, 12809.7559, 12801.3542, 12403.2061, 11528.4291, 10090.1411, 8588.56663, 7266.04088},
        {306.085713, 304.841218, 303.503545, 302.074528, 300.555998, 298.949787, 297.257728, 295.481652, 293.623391, 291.684777, 289.667643, 287.573821, 285.405142, 283.163439, 280.850544, 278.468288, 276.018504, 273.503024, 270.923679, 268.282303, 265.580727, 262.820783, 260.004302, 257.133118, 254.209062, 251.233967, 248.209663, 245.137984, 242.020762, 238.859827, 235.657014, 232.414153, 229.133076, 225.815616, 222.466217, 219.085351, 215.673456, 212.23228, 208.763569, 205.269073, 201.750536, 198.209708, 194.648334, 191.068163, 187.470941, 183.858416, 180.232336, 176.594447, 172.946496, 169.290231, 165.6274, 161.89414, 158.159229, 154.425838, 150.697136, 146.976293, 143.266478, 139.570861, 135.892611, 132.234898, 128.600891, 124.993761, 121.416675, 117.872805, 114.36532, 110.897388, 107.47218, 104.223035, 101.280164, 98.3782754, 95.5120988, 92.6763659, 89.865808, 87.0751562, 84.2991418, 81.532496, 78.7699501, 76.0062352, 73.2360827, 70.4542237, 67.6553895, 64.8343113, 61.9857203, 59.1043478, 55.4226905, 51.3326295, 47.2352072, 43.1498909, 39.0961478, 35.0934451, 31.1612499, 27.3190295, 23.5862512, 19.982382, 16.5268892, 13.23924, 10.1389016, 7.24534116, 4.57802596, 2.15642317, 0, 4.65341302, 8.51726987, 11.6643561, 14.1674571, 16.0993585, 17.5328458, 18.5407045, 19.1957201, 19.5706781, 19.738364, 19.7715634, 19.7430617, 19.7256445, 19.7920973, 20.0152055, 20.4677547, 21.1623826, 22.163282, 23.6792873, 25.7702952, 28.4962023, 31.9169053, 36.0923009, 41.0822857, 46.9467565, 53.74561, 61.5387428, 70.3860515, 80.347433, 91.1044381, 103.234036, 116.909627, 132.304612, 149.249175, 168.088325, 189.258594, 213.036346, 239.697948, 269.519763, 302.778158, 339.148682, 379.691372, 424.800143, 474.868911, 530.291591, 591.474576, 658.796184, 732.647214, 813.418462, 901.500723, 996.157224, 1100.28448, 1216.22688, 1345.2605, 1486.48016, 1639.96004, 1805.77429, 1983.91158, 2174.55099, 2377.96487, 2591.67852, 2821.52104, 3075.73741, 3357.0548, 3658.85079, 3978.79456, 4311.85548, 4655.00279, 5015.10509, 5393.43136, 5799.85615, 6225.71889, 6669.3097, 7128.91868, 7602.83594, 8207.03198, 8792.69642, 9328.69926, 9783.91053, 10127.2002, 10177.3309, 10153.3517, 10088.4643, 10124.6221, 10251.9794, 10415.3772, 10607.2355, 10819.9741, 11046.013, 11277.7721, 11800.385, 12264.7445, 12565.727, 12598.2091, 12257.067, 11437.1773, 10033.4162, 8517.00994, 7177.64308},
        {313.231616, 311.939019, 310.548872, 309.063133, 307.483762, 305.812716, 304.051956, 302.203438, 300.269123, 298.250968, 296.150933, 293.970976, 291.713056, 289.379132, 286.971163, 284.491106, 281.940921, 279.322567, 276.638003, 273.889186, 271.078076, 268.206632, 265.276812, 262.290575, 259.24988, 256.156685, 253.012949, 249.820631, 246.581689, 243.298083, 239.971771, 236.604712, 233.198864, 229.756187, 226.283117, 222.77928, 219.244247, 215.679832, 212.087848, 208.470108, 204.828425, 201.164614, 197.480487, 193.777858, 190.05854, 186.324347, 182.577092, 178.818588, 175.050648, 171.275087, 167.493717, 163.640001, 159.785585, 155.933764, 152.087833, 148.251087, 144.426822, 140.618332, 136.828913, 133.061859, 129.320466, 125.608029, 121.927842, 118.283202, 114.677403, 111.11374, 107.595509, 104.2544, 101.219022, 98.2278632, 95.2758947, 92.3580883, 89.4694158, 86.6048488, 83.759359, 80.9279181, 78.1054979, 75.2870699, 72.467606, 69.6420777, 66.8054568, 63.952715, 61.078824, 58.1787555, 54.5102967, 50.4525173, 46.3933471, 42.3516811, 38.3464138, 34.3964401, 30.5206546, 26.7379522, 23.0672276, 19.5273755, 16.1372906, 12.9158678, 9.88200173, 7.05458717, 4.45251887, 2.09469156, 0, 3.86218828, 7.08432293, 9.72958893, 11.8611712, 13.5422549, 14.8360248, 15.8056659, 16.5143633, 17.0253018, 17.4016665, 17.7066424, 18.0034144, 18.3551674, 18.8250866, 19.4763567, 20.3721629, 21.75377, 23.709803, 25.9006469, 28.4276469, 31.3921478, 34.8954946, 39.0390323, 43.9241058, 49.65206, 56.32424, 64.0419906, 72.9066569, 83.0195837, 94.3788274, 107.226894, 121.696115, 137.918823, 155.61266, 175.117303, 196.88301, 221.16652, 248.224569, 278.313896, 311.691239, 347.572697, 387.572362, 432.150576, 481.767679, 536.884013, 598.882833, 667.070836, 741.677636, 822.932844, 911.066072, 1004.51953, 1107.49484, 1223.25617, 1353.42952, 1496.2962, 1651.63925, 1819.24168, 1999.05722, 2190.65937, 2393.43534, 2602.82486, 2827.3326, 3078.2152, 3359.1922, 3661.21199, 3981.08512, 4311.9757, 4649.74891, 5003.64027, 5375.32226, 5779.31897, 6202.47414, 6642.01165, 7095.15536, 7559.12916, 8131.01751, 8683.21853, 9187.99094, 9617.59346, 9944.2848, 10004.0187, 9996.22884, 9951.5904, 10002.8907, 10140.3854, 10311.7836, 10509.4684, 10725.823, 10953.2303, 11184.0736, 11583.2394, 12001.4294, 12313.953, 12396.1191, 12123.2371, 11370.616, 10013.5651, 8485.57233, 7111.04167},
        {320.057931, 318.680458, 317.205692, 315.635542, 313.971916, 312.216724, 310.371873, 308.439273, 306.420832, 304.318459, 302.134063, 299.869552, 297.526835, 295.10782, 292.614417, 290.048534, 287.41208, 284.706962, 281.935091, 279.098375, 276.198722, 273.238041, 270.218241, 267.14123, 264.008917, 260.823211, 257.586021, 254.299255, 250.964822, 247.58463, 244.160588, 240.694606, 237.18859, 233.644452, 230.062541, 226.445578, 222.796302, 219.116671, 215.408646, 211.674184, 207.915246, 204.13379, 200.331776, 196.511163, 192.67391, 188.821977, 184.957321, 181.081904, 177.197683, 173.306619, 169.41067, 165.449503, 161.48872, 157.531631, 153.581545, 149.641772, 145.715621, 141.806403, 137.917425, 134.051999, 130.213434, 126.405039, 122.630123, 118.891997, 115.19397, 111.539352, 107.931452, 104.495832, 101.355119, 98.2604861, 95.2073178, 92.1909983, 89.2069121, 86.2504436, 83.3169772, 80.4018974, 77.5005886, 74.6084352, 71.7208217, 68.8331325, 65.940752, 63.0390647, 60.123455, 57.1893072, 53.527203, 49.4997527, 45.4780381, 41.4803158, 37.5248425, 33.6298748, 29.8136693, 26.0944827, 22.4905717, 19.0201928, 15.7016028, 12.5530582, 9.59281579, 6.8391321, 4.3102638, 2.02446756, 0, 3.71452767, 6.81047496, 9.35099143, 11.3992266, 13.0183301, 14.2714513, 15.2217399, 15.9323455, 16.4664174, 16.8871054, 17.2575589, 17.6409275, 18.1003607, 18.6990081, 19.5000192, 20.5665436, 22.0703496, 24.0901038, 26.4432703, 29.216274, 32.49554, 36.3674933, 40.918559, 46.2351619, 52.4037271, 59.5106797, 67.6424446, 76.8854469, 87.3261115, 98.80139, 111.738654, 126.299172, 142.644209, 160.778626, 180.941893, 203.41519, 228.406707, 256.124632, 286.777156, 320.572467, 356.083352, 395.651135, 439.803976, 489.070035, 543.97747, 606.877902, 676.020165, 751.476555, 833.319367, 921.620894, 1014.24757, 1116.17361, 1230.86432, 1360.07619, 1502.07679, 1656.70014, 1823.78026, 2003.43361, 2195.14759, 2398.10149, 2607.87264, 2832.44825, 3082.19417, 3360.5738, 3659.35471, 3975.67303, 4303.32054, 4638.56631, 4989.94216, 5359.04341, 5764.04267, 6187.40763, 6624.99531, 7072.66276, 7526.26699, 8057.55271, 8567.51736, 9033.04606, 9431.02392, 9738.33605, 9815.3136, 9833.09829, 9818.52694, 9888.58938, 10034.4368, 10209.6568, 10407.2787, 10620.332, 10841.846, 11064.8502, 11349.3961, 11720.2615, 12038.1943, 12163.9425, 11958.2541, 11281.877, 9995.55924, 8520.24782, 7075.12605},
        {326.820878, 325.349509, 323.782267, 322.120966, 320.36742, 318.523446, 316.590858, 314.57147, 312.467097, 310.279554, 308.010656, 305.662218, 303.236054, 300.733979, 298.157809, 295.509357, 292.790439, 290.002869, 287.148462, 284.229034, 281.246398, 278.20237, 275.098764, 271.937395, 268.720078, 265.448628, 262.12486, 258.750588, 255.327627, 251.857792, 248.342898, 244.784759, 241.18519, 237.546007, 233.859375, 230.132136, 226.371249, 222.57884, 218.757039, 214.907973, 211.03377, 207.136558, 203.218465, 199.281618, 195.328146, 191.360176, 187.379837, 183.389256, 179.390561, 175.38588, 171.377341, 167.314692, 163.253576, 159.197257, 155.148998, 151.112062, 147.089714, 143.085215, 139.101831, 135.142824, 131.211457, 127.310994, 123.444698, 119.615833, 115.827663, 112.083449, 108.386457, 104.853524, 101.59876, 98.3911898, 95.2267153, 92.101237, 89.0106557, 85.9508722, 82.9177873, 79.907302, 76.9153171, 73.9377334, 70.9704518, 68.0093731, 65.0503982, 62.089428, 59.1223631, 56.1451047, 52.4868772, 48.4909423, 44.5085745, 40.5573096, 36.6546832, 32.8182313, 29.0654895, 25.4139936, 21.8812795, 18.4848828, 15.2423393, 12.1711849, 9.28895525, 6.61318617, 4.16141342, 1.95117277, 0, 3.86119844, 7.06472862, 9.67826325, 11.769475, 13.4060367, 14.655621, 15.5859006, 16.2645482, 16.7592366, 17.1376385, 17.4674265, 17.8162734, 18.2518519, 18.8418347, 19.6538946, 20.7557042, 22.1040048, 23.7506217, 26.0141442, 28.9384738, 32.5675124, 36.9451615, 42.1153229, 48.1218982, 55.0087891, 62.8198973, 71.5991245, 81.3903723, 92.2375424, 103.586144, 116.297882, 130.596175, 146.704444, 165.085833, 185.843898, 209.01827, 234.760451, 263.221941, 294.554244, 328.90886, 364.222869, 403.536151, 447.433462, 496.499562, 551.319209, 615.077588, 685.108923, 761.347868, 843.729074, 932.187193, 1024.20565, 1125.16627, 1238.31804, 1365.41646, 1505.16778, 1657.6473, 1822.93031, 2001.42223, 2192.79321, 2396.35328, 2608.98081, 2836.46984, 3086.23543, 3360.82955, 3654.564, 3965.53294, 4289.48919, 4623.9199, 4974.89721, 5343.63711, 5750.93478, 6175.41484, 6611.51584, 7053.67638, 7496.335, 7983.6836, 8447.96911, 8871.19179, 9235.35189, 9522.44966, 9620.14807, 9667.67617, 9687.46445, 9777.67526, 9930.76191, 10107.6228, 10302.2891, 10508.7917, 10721.1618, 10933.4305, 11119.4299, 11445.954, 11764.5693, 11926.8424, 11784.3399, 11188.6282, 9991.27395, 8604.50298, 7081.52303},
        {333.727898, 332.175783, 330.527593, 328.785111, 326.950121, 325.024408, 323.009755, 320.907946, 318.720765, 316.449996, 314.097423, 311.664831, 309.154002, 306.56672, 303.904771, 301.169937, 298.364003, 295.488753, 292.54597, 289.537438, 286.464942, 283.330265, 280.135192, 276.881505, 273.57099, 270.20543, 266.786609, 263.316311, 259.79632, 256.22842, 252.614395, 248.956028, 245.255104, 241.513407, 237.717634, 233.87743, 230.00262, 226.095478, 222.158278, 218.193292, 214.202795, 210.189061, 206.154361, 202.100971, 198.031163, 193.947212, 189.85139, 185.745971, 181.633229, 177.515436, 173.394868, 169.231278, 165.07038, 160.91537, 156.769443, 152.635794, 148.517618, 144.418111, 140.340469, 136.287885, 132.263556, 128.270677, 124.312442, 120.392048, 116.512689, 112.677561, 108.88986, 105.256539, 101.882065, 98.5556336, 95.2737145, 92.0327766, 88.8292885, 85.6597193, 82.5205379, 79.408213, 76.3192137, 73.2500087, 70.1970671, 67.1568576, 64.1258492, 61.1005108, 58.0773112, 55.0527194, 51.39921, 47.4383739, 43.4993019, 39.5987127, 35.7533251, 31.9798577, 28.2950292, 24.7155583, 21.2581638, 17.9395643, 14.7764785, 11.7856252, 8.98372308, 6.38749077, 4.01364702, 1.87891053, 0, 4.04991556, 7.39132138, 10.0973282, 12.241047, 13.8955883, 15.134063, 16.029582, 16.6552559, 17.0841956, 17.3895118, 17.6443154, 17.9217172, 18.2948278, 18.8367581, 19.620619, 20.7195211, 21.8432979, 23.0131636, 25.1239275, 28.1708553, 32.1492126, 37.0542651, 42.8812784, 49.6255182, 57.2822502, 65.8467399, 75.3142531, 85.6800554, 96.9394124, 108.10211, 120.510237, 134.454701, 150.226413, 168.742091, 189.979739, 213.750481, 240.157484, 269.303913, 301.292936, 336.227721, 371.573694, 410.874552, 454.74954, 503.817904, 558.698889, 623.15999, 693.885142, 770.70653, 853.456337, 941.966749, 1033.5054, 1133.60347, 1245.0986, 1369.67665, 1506.67192, 1656.47451, 1819.47452, 1996.42989, 2187.27929, 2391.56009, 2607.82225, 2839.1379, 3089.31402, 3359.78624, 3647.9721, 3953.13349, 4273.48223, 4608.008, 4959.55073, 5328.77244, 5737.92467, 6162.75842, 6596.46242, 7032.22537, 7463.23597, 7907.62355, 8327.40105, 8709.52204, 9040.94007, 9308.60871, 9426.80263, 9503.60686, 9557.12306, 9666.78369, 9826.41364, 10004.2833, 10195.4952, 10395.1519, 10598.3561, 10800.2104, 10911.7093, 11200.5469, 11515.3964, 11704.9309, 11617.8237, 11102.7478, 10008.3766, 8725.61959, 7141.61384},
        {341.193481, 339.627413, 337.956607, 336.183151, 334.309131, 332.336633, 330.267745, 328.104551, 325.84914, 323.503597, 321.070009, 318.550462, 315.947043, 313.261839, 310.496936, 307.65442, 304.736378, 301.744896, 298.682061, 295.54996, 292.350678, 289.086303, 285.758921, 282.370619, 278.923482, 275.419597, 271.861052, 268.249932, 264.588323, 260.878313, 257.121988, 253.321435, 249.478739, 245.595988, 241.669795, 237.705098, 233.706901, 229.677468, 225.619063, 221.533951, 217.424394, 213.292658, 209.141006, 204.971703, 200.787012, 196.589198, 192.380525, 188.163256, 183.939657, 179.711991, 175.482521, 171.212371, 166.945837, 162.686076, 158.436245, 154.199499, 149.978994, 145.777886, 141.599332, 137.446487, 133.322506, 129.230547, 125.173765, 121.155316, 117.178356, 113.246041, 109.361528, 105.622145, 102.122962, 98.6729078, 95.269034, 91.9083916, 88.5880315, 85.305005, 82.056363, 78.8391567, 75.6504371, 72.4872553, 69.3466624, 66.2257095, 63.1214477, 60.030928, 56.9512015, 53.8793193, 50.2373976, 46.3220477, 42.4367917, 38.5973383, 34.8193963, 31.1186746, 27.5108821, 24.0117275, 20.6369196, 17.4021674, 14.3231795, 11.415665, 8.69533249, 6.17789093, 3.8790491, 1.81451585, 0, 3.95148948, 7.19151186, 9.79524529, 11.8378679, 13.3945578, 14.5404933, 15.3508523, 15.9008131, 16.2655538, 16.5202525, 16.7400874, 17.0002366, 17.3758783, 17.9421906, 18.7743517, 19.9475396, 21.0143336, 21.9752549, 24.0875506, 27.3144133, 31.6190358, 36.9646107, 43.3143306, 50.6313882, 58.8789761, 68.0202871, 78.0185137, 88.8368485, 100.438484, 111.559906, 123.840806, 137.612391, 153.205865, 171.73275, 193.134092, 217.142854, 243.856149, 273.371085, 305.784774, 341.194327, 377.165757, 417.097606, 461.582197, 511.211856, 566.578907, 631.36921, 702.308173, 779.214736, 861.90784, 950.206428, 1041.43275, 1140.95395, 1251.1616, 1373.64002, 1508.32534, 1655.89369, 1817.02122, 1992.92622, 2183.61945, 2388.52017, 2607.11021, 2840.76608, 3090.82994, 3358.02375, 3642.01867, 3942.9682, 4261.19309, 4596.89017, 4949.64312, 5319.38243, 5727.19419, 6148.59865, 6576.20325, 7002.61542, 7420.44261, 7830.68808, 8215.46448, 8565.28029, 8870.64398, 9122.06403, 9252.58132, 9348.48366, 9424.90852, 9549.09965, 9715.56909, 9897.11792, 10089.3437, 10287.8441, 10488.2166, 10686.0589, 10776.0966, 11042.6388, 11343.8794, 11538.0124, 11483.2318, 11037.7314, 10059.7054, 8869.23083, 7283.2224},
        {349.495856, 348.016876, 346.411673, 344.683142, 342.834178, 340.867673, 338.786524, 336.593623, 334.291866, 331.884146, 329.373357, 326.762395, 324.054154, 321.251526, 318.357408, 315.374693, 312.306276, 309.15505, 305.92391, 302.61575, 299.233465, 295.779949, 292.258096, 288.6708, 285.020956, 281.311458, 277.5452, 273.725077, 269.853982, 265.934811, 261.970457, 257.963814, 253.917778, 249.835241, 245.746118, 241.639225, 237.503053, 233.33962, 229.150944, 224.939042, 220.705932, 216.45363, 212.184155, 207.899524, 203.601754, 199.292864, 194.974869, 190.649788, 186.319639, 181.986438, 177.652203, 173.265231, 168.882425, 164.506967, 160.142038, 155.790821, 151.456498, 147.14225, 142.851261, 138.586711, 134.351784, 130.14966, 125.983522, 121.856553, 117.771933, 113.732845, 109.742472, 105.889875, 102.262054, 98.6850707, 95.1565392, 91.6740745, 88.2352912, 84.8378042, 81.4792282, 78.157178, 74.8692684, 71.6131141, 68.3863298, 65.1865305, 62.0113307, 58.8583454, 55.7251892, 52.6094769, 48.9899623, 45.1346041, 41.3175976, 37.5533955, 33.8564507, 30.2412159, 26.722144, 23.3136877, 20.0303, 16.8864335, 13.8965411, 11.0750756, 8.43648985, 5.99523659, 3.76576866, 1.76253886, 0, 3.34425244, 6.06284509, 8.22666426, 9.90659625, 11.1735274, 12.0983439, 12.7519321, 13.2051784, 13.528969, 13.7941902, 14.0717283, 14.4324697, 14.9473005, 15.6871072, 16.7227761, 18.1251933, 19.4689097, 20.7539063, 23.1737334, 26.6929198, 31.2759941, 36.8874849, 43.491921, 51.0538309, 59.5377433, 68.9081869, 79.1296904, 90.1667823, 101.983991, 113.359355, 125.878939, 139.863221, 155.632678, 174.057194, 195.182541, 218.907982, 245.389169, 274.781757, 307.241398, 342.923746, 380.327679, 421.769861, 467.730098, 518.688193, 575.123953, 639.716646, 710.196747, 786.494194, 868.538926, 956.260881, 1047.36712, 1146.72731, 1256.38605, 1377.82549, 1511.37949, 1657.89748, 1818.22887, 1994.1585, 2185.38777, 2390.59766, 2608.69929, 2841.48843, 3090.2746, 3355.8965, 3638.39498, 3938.17711, 4256.6712, 4594.54907, 4948.73806, 5318.28359, 5719.58752, 6131.66864, 6547.56436, 6960.31214, 7362.94939, 7753.52147, 8118.80592, 8450.58819, 8740.65372, 8980.78797, 9109.9137, 9207.92086, 9289.25032, 9419.98174, 9594.11566, 9784.13651, 9985.12064, 10192.1444, 10400.2842, 10604.6164, 10745.4911, 11011.056, 11285.4055, 11452.6341, 11396.836, 11002.1056, 10152.5374, 9024.52875, 7525.97059},
        {358.020359, 356.636692, 355.104338, 353.42703, 351.608501, 349.652485, 347.562714, 345.342922, 342.996843, 340.528209, 337.940754, 335.238212, 332.424315, 329.502797, 326.477391, 323.35183, 320.129849, 316.815179, 313.411554, 309.922709, 306.352375, 302.704286, 298.982176, 295.189777, 291.330824, 287.409049, 283.428185, 279.391967, 275.304127, 271.168398, 266.988515, 262.768209, 258.511215, 254.221266, 249.962353, 245.706814, 241.42626, 237.122469, 232.79722, 228.452289, 224.089455, 219.710495, 215.317187, 210.911309, 206.494639, 202.068955, 197.636034, 193.197653, 188.755592, 184.311628, 179.867538, 175.358799, 170.854928, 166.35914, 161.874649, 157.40467, 152.952418, 148.52111, 144.113958, 139.73418, 135.384989, 131.069601, 126.79123, 122.553092, 118.358403, 114.210376, 110.112227, 106.145166, 102.388532, 98.6846796, 95.0317675, 91.4279548, 87.8714006, 84.3602636, 80.8927028, 77.4668772, 74.0809456, 70.733067, 67.4214003, 64.1441044, 60.8993382, 57.6852607, 54.5000308, 51.3418074, 47.7458515, 43.9515074, 40.2036832, 36.5155596, 32.9003175, 29.3711378, 25.9412013, 22.6236889, 19.4317814, 16.3786596, 13.4775045, 10.7414968, 8.18381741, 5.81764719, 3.65616699, 1.71255764, 0, 2.71108226, 4.88904952, 6.59964817, 7.90862458, 8.88172513, 9.58469621, 10.0832842, 10.4432355, 10.7302964, 11.0102134, 11.3487328, 11.811601, 12.4645644, 13.3733694, 14.6037623, 16.2214895, 17.8824396, 19.5938078, 22.3483999, 26.1241356, 30.8989347, 36.6507168, 43.3574016, 50.9969089, 59.5471582, 68.9860693, 79.291562, 90.4415558, 102.41397, 114.140084, 127.028227, 141.37031, 157.458247, 175.806766, 196.596371, 219.948148, 246.087164, 275.238487, 307.627183, 343.47832, 382.34528, 425.329241, 472.786686, 525.0741, 582.547965, 646.688871, 716.448171, 791.90132, 873.123777, 960.190998, 1051.30041, 1150.70087, 1260.149, 1381.01997, 1513.91013, 1659.7655, 1819.53211, 1995.36462, 2186.72572, 2391.75956, 2608.81796, 2840.26197, 3087.73844, 3352.28171, 3633.88833, 3933.03121, 4251.88324, 4591.35808, 4946.1361, 5314.42357, 5707.18332, 6107.90251, 6510.37168, 6908.38136, 7295.72209, 7672.36952, 8024.38279, 8344.00615, 8623.48387, 8855.0602, 8979.28892, 9074.56508, 9155.28599, 9289.03169, 9469.14801, 9666.82128, 9876.5053, 10092.6539, 10309.7207, 10522.1597, 10741.6412, 11015.0658, 11258.6157, 11388.4727, 11320.8192, 10971.837, 10257.7083, 9204.02541, 7835.73577},
        {366.595695, 365.285845, 363.80766, 362.165589, 360.364082, 358.40759, 356.300562, 354.04745, 351.652702, 349.12077, 346.456104, 343.663154, 340.74637, 337.710203, 334.559102, 331.297519, 327.929902, 324.460703, 320.894371, 317.235358, 313.488112, 309.657085, 305.746726, 301.761487, 297.705816, 293.584165, 289.400983, 285.160722, 280.86783, 276.526758, 272.141958, 267.717877, 263.258968, 258.76968, 254.341456, 249.933421, 245.503652, 241.053777, 236.585423, 232.100218, 227.599788, 223.085761, 218.559764, 214.023424, 209.478368, 204.926223, 200.368617, 195.807177, 191.243529, 186.679301, 182.116121, 177.481042, 172.851882, 168.231883, 163.62429, 159.032347, 154.459296, 149.908383, 145.38285, 140.885941, 136.420901, 131.990973, 127.599401, 123.249427, 118.944298, 114.687255, 110.481543, 106.401009, 102.517805, 98.6895609, 94.914943, 91.1926181, 87.521253, 83.8995147, 80.3260699, 76.7995853, 73.3187279, 69.8821644, 66.4885617, 63.1365865, 59.8249057, 56.5521861, 53.3170944, 50.1182976, 46.5441461, 42.8070954, 39.1245325, 35.5084398, 31.9707996, 28.5235944, 25.1788065, 21.9484183, 18.8444121, 15.8787703, 13.0634753, 10.4105095, 7.9318551, 5.63949461, 3.54541035, 1.66158469, 0, 2.18501245, 3.91519521, 5.25159726, 6.2552676, 6.98725523, 7.50860915, 7.88037834, 8.16361181, 8.41935854, 8.70866754, 9.0925878, 9.63216832, 10.3884581, 11.4225061, 12.7953614, 14.5680729, 16.521368, 18.6762498, 21.7278281, 25.6770828, 30.5249942, 36.2725423, 42.9207071, 50.4704688, 58.9228074, 68.2787031, 78.5391358, 89.7050857, 101.777533, 113.926094, 127.287945, 142.113477, 158.653077, 177.027147, 197.551069, 220.574892, 246.388002, 275.279785, 307.539626, 343.456913, 383.532528, 427.779991, 476.447309, 529.782488, 588.033534, 651.529615, 720.417286, 794.92426, 875.278254, 961.706984, 1052.92037, 1152.51901, 1262.07678, 1382.87042, 1515.56991, 1661.11765, 1820.45603, 1995.75959, 2186.45855, 2390.63904, 2606.48833, 2836.62819, 3083.01444, 3346.79012, 3627.72111, 3926.20544, 4244.79564, 4584.44829, 4938.22017, 5303.63663, 5686.41277, 6074.62424, 6462.9615, 6846.11501, 7218.77521, 7585.66209, 7928.92919, 8240.75955, 8513.33625, 8738.84234, 8857.05624, 8947.5021, 9024.82213, 9158.77876, 9342.19577, 9544.93017, 9760.9541, 9984.23965, 10208.759, 10428.4841, 10741.4879, 11028.2687, 11240.1215, 11328.3411, 11244.2222, 10939.0598, 10364.1486, 9408.65324, 8205.08232},
        {374.999505, 373.708224, 372.236267, 370.588493, 368.769764, 366.784941, 364.638884, 362.336456, 359.882515, 357.281924, 354.539543, 351.660234, 348.648857, 345.510273, 342.249342, 338.870927, 335.379888, 331.781086, 328.079381, 324.279635, 320.386709, 316.405463, 312.340759, 308.197457, 303.980418, 299.694504, 295.344575, 290.935491, 286.472115, 281.959307, 277.401928, 272.804838, 268.172899, 263.510972, 258.922247, 254.360351, 249.777732, 245.176059, 240.557002, 235.922231, 231.273416, 226.612227, 221.940333, 217.259405, 212.571113, 207.877126, 203.179115, 198.478749, 193.777699, 189.077634, 184.380224, 179.61389, 174.855139, 170.107229, 165.373417, 160.656962, 155.961122, 151.289154, 146.644316, 142.029866, 137.449063, 132.905162, 128.401424, 123.941105, 119.527463, 115.163756, 110.853242, 106.663176, 102.659275, 98.7128091, 94.8228873, 90.988619, 87.2091134, 83.4834796, 79.810827, 76.1902646, 72.6209017, 69.1018476, 65.6322114, 62.2111023, 58.8376296, 55.5109025, 52.2300302, 48.9941218, 45.4363231, 41.7464226, 38.1186074, 34.563863, 31.0931748, 27.7175283, 24.4479088, 21.2953019, 18.270693, 15.3850674, 12.6494106, 10.0747082, 7.67194539, 5.45210775, 3.42618068, 1.60514962, 0, 1.96451446, 3.50682287, 4.68563487, 5.55966008, 6.18760814, 6.62818868, 6.94011134, 7.18208574, 7.41282151, 7.69102829, 8.07541571, 8.6246934, 9.39757099, 10.4527581, 11.8489644, 13.6448995, 15.7718446, 18.2703054, 21.4872223, 25.4539989, 30.2020386, 35.7627451, 42.1675219, 49.4477726, 57.6349007, 66.7603098, 76.8554034, 87.9515851, 100.080258, 112.709861, 126.641184, 142.074519, 159.21016, 177.819191, 198.316511, 221.231472, 246.89313, 275.63054, 307.772759, 343.648842, 384.31317, 429.148723, 478.342644, 532.082078, 590.554167, 653.292913, 721.301889, 794.931526, 874.532254, 960.454502, 1051.82707, 1151.71512, 1261.70473, 1383.00756, 1516.07083, 1661.68498, 1820.64046, 1994.51794, 2183.13803, 2385.45925, 2600.48057, 2830.12193, 3076.06432, 3339.2355, 3619.28696, 3916.45599, 4233.29978, 4570.65708, 4920.86051, 5281.05423, 5653.0316, 6028.57234, 6403.21132, 6772.48338, 7131.92335, 7491.44332, 7828.6066, 8135.35372, 8403.62521, 8625.36162, 8739.2969, 8826.04923, 8900.36196, 9032.45804, 9215.12016, 9418.01451, 9635.07962, 9860.25399, 10087.4762, 10310.6846, 10719.7118, 11022.209, 11204.5611, 11253.1529, 11154.3696, 10894.5959, 10460.2169, 9643.92192, 8631.43485},
        {382.516148, 381.115486, 379.540329, 377.795159, 375.884458, 373.812711, 371.584399, 369.204006, 366.676015, 364.004909, 361.195169, 358.251281, 355.177725, 351.978986, 348.659545, 345.223887, 341.676494, 338.021848, 334.264433, 330.408732, 326.459227, 322.420402, 318.296738, 314.092721, 309.812831, 305.461552, 301.043367, 296.562759, 292.024211, 287.432205, 282.791225, 278.105753, 273.380273, 268.619267, 263.896795, 259.181087, 254.43954, 249.674377, 244.887824, 240.082106, 235.259448, 230.422074, 225.572211, 220.712082, 215.843913, 210.96993, 206.092356, 201.213417, 196.335337, 191.460343, 186.590659, 181.680931, 176.781995, 171.897107, 167.029522, 162.182499, 157.359292, 152.563159, 147.797355, 143.065138, 138.369764, 133.714488, 129.102568, 124.53726, 120.02182, 115.559505, 111.153571, 106.867748, 102.764314, 98.7218031, 94.739552, 90.8168965, 86.9531728, 83.147717, 79.3998654, 75.708954, 72.0743189, 68.4952965, 64.9712227, 61.5014338, 58.0852659, 54.7220552, 51.4111379, 48.15185, 44.5936101, 40.9180734, 37.3111004, 33.7833828, 30.3456123, 27.0084805, 23.7826792, 20.6788998, 17.7078342, 14.8801739, 12.2066105, 9.69783586, 7.36454148, 5.21741905, 3.26716023, 1.52445666, 0, 2.88022254, 5.1932796, 7.00585151, 8.38461859, 9.39626118, 10.1074596, 10.5848942, 10.8952453, 11.1051932, 11.2814183, 11.4906008, 11.7994212, 12.2745596, 12.9826966, 13.9905123, 15.3646872, 17.176431, 19.493071, 22.3710425, 25.8779964, 30.0815836, 35.049455, 40.8492616, 47.5486543, 55.2152839, 63.9168015, 73.7208579, 84.695104, 96.9071908, 110.179357, 124.91465, 141.254345, 159.339716, 178.816764, 200.074401, 223.633611, 249.784248, 278.816171, 311.019235, 346.683295, 386.171493, 429.678097, 477.478626, 529.848596, 587.063528, 649.200255, 716.78265, 790.135902, 869.5852, 955.455734, 1046.7731, 1146.75047, 1257.68959, 1380.90541, 1515.69812, 1662.2725, 1820.83334, 1990.42356, 2172.67385, 2370.48238, 2587.13043, 2820.21381, 3068.47991, 3331.39332, 3609.63397, 3903.32393, 4214.56246, 4543.98426, 4884.97424, 5235.01819, 5596.27018, 5960.82951, 6324.56605, 6683.34964, 7033.05013, 7384.03082, 7714.04474, 8015.33837, 8280.15818, 8500.75067, 8619.50312, 8711.76067, 8791.13802, 8920.12321, 9092.98526, 9283.62928, 9487.34249, 9699.41209, 9915.12528, 10129.7693, 10629.1616, 10948.5701, 11107.506, 11125.4801, 11022.0035, 10816.5871, 10528.7419, 9959.55953, 9159.17045},
        {390.157831, 388.583562, 386.847167, 384.952525, 382.903514, 380.704013, 378.357901, 375.869055, 373.241355, 370.478678, 367.584903, 364.56391, 361.419575, 358.155778, 354.776397, 351.28531, 347.686396, 343.983534, 340.180602, 336.281478, 332.29004, 328.210169, 324.04574, 319.800634, 315.478729, 311.083903, 306.620035, 302.091002, 297.500684, 292.85296, 288.151706, 283.400803, 278.604128, 273.765559, 268.92037, 264.056081, 259.159836, 254.234496, 249.282919, 244.307965, 239.312494, 234.299366, 229.27144, 224.231577, 219.182635, 214.127474, 209.068954, 204.009936, 198.953277, 193.901839, 188.858481, 183.810104, 178.775873, 173.758992, 168.762667, 163.790104, 158.844509, 153.929087, 149.047044, 144.201585, 139.395916, 134.633243, 129.916771, 125.249706, 120.635253, 116.076618, 111.577006, 107.194927, 102.987559, 98.844051, 94.7640244, 90.7470994, 86.7928966, 82.9010366, 79.0711399, 75.3028272, 71.595719, 67.9494358, 64.3635984, 60.8378273, 57.371743, 53.9649661, 50.6171173, 47.3278172, 43.7677785, 40.1079939, 36.5234092, 33.0243186, 29.6210164, 26.3237968, 23.1429542, 20.0887829, 17.171577, 14.4016308, 11.7892387, 9.34469493, 7.0782937, 5.00032932, 3.12109607, 1.4508882, 0, 3.54846632, 6.42739653, 8.70855927, 10.4637231, 11.7646568, 12.6831288, 13.2909078, 13.6597624, 13.8614613, 13.967773, 14.0504663, 14.1813096, 14.4320716, 14.874521, 15.5804263, 16.6215562, 18.2270743, 20.4912341, 23.129792, 26.2482439, 29.952086, 34.3468145, 39.5379254, 45.6309149, 52.731279, 60.9445139, 70.3761156, 81.1315804, 93.3164042, 107.089166, 122.482815, 139.586924, 158.491062, 178.786732, 200.81254, 225.03991, 251.707835, 281.055307, 313.321318, 348.744861, 386.87699, 428.854008, 475.049503, 525.837063, 581.590278, 642.892065, 709.854352, 782.798393, 862.045446, 947.916766, 1039.47941, 1139.84175, 1251.76568, 1376.58689, 1512.7291, 1659.92346, 1817.90108, 1983.62874, 2160.2302, 2353.84491, 2571.3007, 2806.59936, 3056.14321, 3318.16783, 3594.01511, 3883.60108, 4188.28863, 4508.36891, 4838.82795, 5177.65266, 5528.21553, 5882.28934, 6235.99674, 6585.46039, 6926.80295, 7270.79522, 7595.2497, 7892.62702, 8155.38781, 8375.99269, 8500.23587, 8598.35536, 8682.8117, 8807.94415, 8969.63625, 9146.32204, 9334.75751, 9531.69865, 9733.90146, 9938.12194, 10499.1588, 10828.6035, 10971.3124, 10972.1422, 10875.9492, 10727.5898, 10571.9207, 10250.4084, 9668.20376},
        {398.27571, 396.443157, 394.466948, 392.350171, 390.095915, 387.70727, 385.187323, 382.539164, 379.765881, 376.870564, 373.8563, 370.72618, 367.48329, 364.130722, 360.671562, 357.108901, 353.445826, 349.685426, 345.830791, 341.885009, 337.851169, 333.73236, 329.53167, 325.252189, 320.897005, 316.469206, 311.971882, 307.408122, 302.781014, 298.093647, 293.349109, 288.550491, 283.70088, 278.803365, 273.849165, 268.847553, 263.807947, 258.733818, 253.628643, 248.495894, 243.339045, 238.161571, 232.966946, 227.758643, 222.540137, 217.314901, 212.086409, 206.858137, 201.633556, 196.416142, 191.209369, 186.034531, 180.876895, 175.739548, 170.625579, 165.538076, 160.480125, 155.454815, 150.465233, 145.514467, 140.605604, 135.741733, 130.92594, 126.161315, 121.450943, 116.797914, 112.205314, 107.723075, 103.400002, 99.1425737, 94.9508399, 90.8248519, 86.7646608, 82.7703176, 78.8418733, 74.979379, 71.1828857, 67.4524444, 63.7881061, 60.1899218, 56.6579427, 53.1922196, 49.7928036, 46.4597458, 42.9009695, 39.2687341, 35.7188101, 32.2607286, 28.9040207, 25.6582176, 22.5328504, 19.5374502, 16.6815482, 13.9746754, 11.426363, 9.04614211, 6.84354389, 4.82809944, 3.00933989, 1.39679637, 0, 3.43437573, 6.22633736, 8.44457816, 10.1577914, 11.4346702, 12.343908, 12.954198, 13.3342333, 13.5527074, 13.6783133, 13.7797444, 13.925694, 14.1848552, 14.6259213, 15.3175856, 16.3285413, 18.071215, 20.6634048, 23.3963118, 26.4122863, 29.8536788, 33.8628396, 38.5821192, 44.153868, 50.7204364, 58.4241747, 67.4074334, 77.8125629, 89.7819136, 103.738564, 119.441202, 136.947961, 156.316973, 177.177902, 199.803113, 224.579232, 251.692929, 281.330877, 313.679749, 348.926217, 385.934154, 426.615622, 471.416102, 520.781073, 575.156016, 635.174814, 701.047444, 773.172287, 851.947724, 937.772134, 1030.16444, 1131.47737, 1244.23959, 1369.60733, 1505.93481, 1652.83429, 1809.91804, 1973.5397, 2147.31075, 2338.39757, 2554.84415, 2789.36721, 3037.72412, 3297.67539, 3570.37579, 3855.42205, 4153.21962, 4463.57487, 4783.32909, 5111.00077, 5451.49994, 5795.97029, 6140.71788, 6482.04877, 6816.26904, 7154.51287, 7474.55117, 7768.98295, 8030.40723, 8251.42305, 8380.68281, 8484.02991, 8572.89763, 8694.38929, 8845.8188, 9009.63756, 9183.9026, 9366.67096, 9555.99968, 9749.9458, 10320.7513, 10646.6689, 10784.2824, 10790.176, 10720.9335, 10633.139, 10583.3765, 10477.3235, 10110.609},
        {407.835532, 405.541344, 403.134444, 400.616802, 397.990387, 395.257169, 392.419118, 389.478203, 386.436393, 383.295659, 380.05797, 376.725295, 373.299604, 369.782867, 366.177052, 362.48413, 358.706071, 354.844843, 350.902417, 346.880761, 342.781846, 338.607641, 334.360116, 330.04124, 325.652982, 321.197313, 316.676202, 312.091618, 307.445532, 302.739912, 297.976728, 293.157949, 288.285547, 283.361488, 278.328013, 273.218211, 268.065894, 262.874969, 257.649344, 252.392928, 247.109628, 241.803352, 236.478008, 231.137504, 225.785749, 220.426649, 215.064114, 209.702051, 204.344368, 198.994973, 193.657774, 188.383684, 183.128587, 177.895372, 172.686928, 167.506145, 162.35591, 157.239114, 152.158645, 147.117392, 142.118245, 137.164092, 132.257822, 127.402325, 122.600489, 117.855204, 113.169358, 108.578643, 104.117963, 99.7225543, 95.3931795, 91.130601, 86.9355815, 82.8088836, 78.7512697, 74.7635025, 70.8463446, 67.0005584, 63.2269066, 59.5261517, 55.8990564, 52.3463831, 48.8688944, 45.467353, 41.9233007, 38.3517214, 34.8702339, 31.4867149, 28.2090411, 25.0450892, 22.0027359, 19.0898579, 16.3143319, 13.6840346, 11.2068427, 8.89063288, 6.7432819, 4.77266642, 2.98666314, 1.39314877, 0, 1.22297946, 2.17890016, 2.91159544, 3.46489866, 3.88264318, 4.20866234, 4.4867895, 4.76085801, 5.07470123, 5.47215251, 5.99704519, 6.69321265, 7.60448821, 8.77470525, 10.2476971, 12.0672972, 14.9009569, 18.8815981, 22.642489, 26.3610956, 30.2148835, 34.3813187, 39.037867, 44.3619942, 50.5311661, 57.7228485, 66.1145073, 75.8836083, 87.2076173, 100.589442, 115.761778, 132.804458, 151.797315, 172.62655, 195.46881, 220.552384, 248.015193, 277.995162, 310.630213, 346.058269, 382.966048, 423.38435, 467.735032, 516.439949, 569.920955, 627.402759, 690.803649, 760.84477, 838.247262, 923.732269, 1018.41875, 1122.14381, 1235.36394, 1358.53514, 1492.11236, 1636.55104, 1792.30659, 1959.02824, 2138.16085, 2332.02874, 2543.80421, 2770.88492, 3012.63064, 3268.26507, 3536.78143, 3817.27872, 4109.03373, 4411.19156, 4722.24547, 5041.05741, 5373.62691, 5710.58031, 6048.30891, 6383.20401, 6711.6569, 7044.43687, 7359.96272, 7651.03121, 7910.43913, 8130.98329, 8263.69912, 8370.98563, 8463.53733, 8583.9373, 8730.25747, 8887.28286, 9053.74974, 9228.39439, 9409.95307, 9597.16206, 10070.8116, 10364.715, 10519.4504, 10575.5964, 10573.7312, 10554.4334, 10558.2813, 10577.0695, 10422.3577},
        {417.246149, 414.576951, 411.817936, 408.970252, 406.035046, 403.013467, 399.906663, 396.715783, 393.441974, 390.086385, 386.650163, 383.134459, 379.540419, 375.869191, 372.121925, 368.299768, 364.403868, 360.435374, 356.395434, 352.285196, 348.105808, 343.85842, 339.544178, 335.164231, 330.719727, 326.211815, 321.641643, 317.010359, 312.319111, 307.569048, 302.761317, 297.897068, 292.977447, 288.003604, 282.881066, 277.660802, 272.394933, 267.087708, 261.74338, 256.3662, 250.96042, 245.53029, 240.080062, 234.613987, 229.136316, 223.651302, 218.163194, 212.676245, 207.194706, 201.722829, 196.264863, 190.902221, 185.560321, 180.241743, 174.949064, 169.684864, 164.451721, 159.252213, 154.08892, 148.96442, 143.881292, 138.842114, 133.849465, 128.905924, 124.014069, 119.176479, 114.395733, 109.688017, 105.06919, 100.513768, 96.0234448, 91.5999184, 87.2448851, 82.9600413, 78.7470836, 74.6077085, 70.5436125, 66.5564921, 62.6480439, 58.8199644, 55.0739501, 51.4116975, 47.8349031, 44.3452635, 40.8149125, 37.313053, 33.9104454, 30.6129909, 27.4265905, 24.3571452, 21.4105563, 18.5927247, 15.9095515, 13.3669378, 10.9707848, 8.72699345, 6.64146487, 4.72010016, 2.96880039, 1.39346664, 0, -1.15522203, -2.1836572, -3.06704198, -3.78711285, -4.32560627, -4.66425872, -4.78480667, -4.66898659, -4.29853495, -3.65518821, -2.72068286, -1.47675535, 0.0948578267, 2.01242021, 4.29419534, 6.95844673, 10.8829107, 16.2086327, 21.0098305, 25.4889405, 29.8483988, 34.2906418, 39.0181057, 44.2332269, 50.1384416, 56.936186, 64.8288965, 74.0190094, 84.7089608, 97.427157, 111.930542, 128.323761, 146.71146, 167.149337, 189.76651, 214.705152, 242.08459, 272.024155, 304.643176, 340.060982, 377.10244, 417.575309, 461.852182, 510.305652, 563.308314, 618.651177, 679.933815, 748.174217, 824.390373, 909.600271, 1006.44494, 1112.33562, 1225.61582, 1346.04457, 1476.27092, 1617.64635, 1771.52234, 1941.18958, 2125.61962, 2321.66856, 2526.78966, 2744.03906, 2977.20005, 3227.31784, 3490.79823, 3766.17658, 4051.51113, 4345.21355, 4647.44496, 4957.3769, 5281.83401, 5611.15731, 5941.86918, 6270.49201, 6593.54816, 6922.87494, 7235.85108, 7525.17023, 7783.52604, 8003.61216, 8136.88194, 8244.76286, 8337.62299, 8456.06224, 8598.48192, 8750.68671, 8911.70771, 9080.57606, 9256.32287, 9437.97926, 9785.18601, 10037.9282, 10213.3685, 10328.6696, 10400.9941, 10447.5047, 10485.364, 10586.6351, 10606.4136},
        {425.6531, 422.877324, 420.012663, 417.060191, 414.020985, 410.896119, 407.686671, 404.393714, 401.018325, 397.56158, 394.024554, 390.408322, 386.713961, 382.942546, 379.095152, 375.172856, 371.176732, 367.107857, 362.967305, 358.756154, 354.475477, 350.126352, 345.709853, 341.227056, 336.679036, 332.06687, 327.391633, 322.654401, 317.856248, 312.998252, 308.081487, 303.107029, 298.075953, 292.989336, 287.742151, 282.390755, 276.992784, 271.552755, 266.075187, 260.5646, 255.025512, 249.462442, 243.879909, 238.282431, 232.674529, 227.060719, 221.445522, 215.833455, 210.229039, 204.636791, 199.061231, 193.620027, 188.202096, 182.809503, 177.444314, 172.108595, 166.804413, 161.533833, 156.298921, 151.101743, 145.944365, 140.828853, 135.757273, 130.731691, 125.754173, 120.826784, 115.951592, 111.117342, 106.3131, 101.568403, 96.8861805, 92.2693618, 87.7208764, 83.2436537, 78.8406229, 74.5147137, 70.2688553, 66.1059773, 62.0290089, 58.0408797, 54.144519, 50.3428563, 46.638821, 43.0353425, 39.5081888, 36.0744415, 32.7515083, 29.5432, 26.4533275, 23.4857017, 20.6441333, 17.9324333, 15.3544126, 12.9138819, 10.6146521, 8.46053402, 6.45533858, 4.60287661, 2.90695896, 1.37139647, 0, -2.76057923, -5.14494685, -7.15044515, -8.77441641, -10.0142029, -10.8671469, -11.3305908, -11.4018767, -11.0783471, -10.3573441, -9.23621001, -7.7122872, -5.78291791, -3.44544442, -0.697209025, 2.464446, 7.00287879, 13.05799, 18.4594244, 23.4157042, 28.1353514, 32.8268881, 37.6988365, 42.9597185, 48.8180564, 55.4823721, 63.1611879, 72.0630257, 82.3964077, 94.7154596, 108.756379, 124.624006, 142.423182, 162.061257, 183.741818, 207.721115, 234.163237, 263.232272, 295.092308, 329.907433, 367.050432, 407.717528, 452.227631, 500.899648, 554.052489, 608.68838, 669.272083, 736.951676, 812.875239, 898.190852, 996.357295, 1103.38775, 1216.34869, 1334.56536, 1461.97475, 1600.44328, 1751.83734, 1921.80355, 2107.569, 2302.23695, 2498.91061, 2705.27404, 2929.09671, 3172.9938, 3430.84676, 3700.54596, 3978.77693, 4263.1177, 4555.56398, 4855.61258, 5170.5061, 5490.80391, 5813.32159, 6134.87472, 6452.27887, 6781.77632, 7095.89929, 7386.60667, 7645.85738, 7865.61032, 7993.92701, 8095.92867, 8182.38738, 8295.95228, 8434.77744, 8583.7348, 8741.64252, 8907.31875, 9079.58167, 9257.24944, 9474.876, 9682.93264, 9874.98073, 10044.5816, 10185.2967, 10290.6873, 10354.3147, 10503.3922, 10629.5486},
        {432.041947, 429.641794, 427.11537, 424.464958, 421.692838, 418.801292, 415.792601, 412.669048, 409.432912, 406.086477, 402.632022, 399.07183, 395.408182, 391.643359, 387.779643, 383.819315, 379.764657, 375.617949, 371.381474, 367.057513, 362.648347, 358.156258, 353.583526, 348.932434, 344.205263, 339.404295, 334.53181, 329.59009, 324.581416, 319.508071, 314.372335, 309.176489, 303.922816, 298.613597, 293.17538, 287.649907, 282.079828, 276.469881, 270.824807, 265.149344, 259.448231, 253.726207, 247.988011, 242.238383, 236.482061, 230.723784, 224.968292, 219.220324, 213.484618, 207.765914, 202.06895, 196.560983, 191.080712, 185.629353, 180.208122, 174.818234, 169.460907, 164.137355, 158.848795, 153.596443, 148.381515, 143.205226, 138.068794, 132.973433, 127.920361, 122.910792, 117.945944, 112.975158, 107.949023, 102.97574, 98.0598867, 93.206043, 88.418787, 83.7026977, 79.0623538, 74.502334, 70.027217, 65.6415816, 61.3500066, 57.1570706, 53.0673525, 49.0854309, 45.2158847, 41.4632925, 37.9163457, 34.5357976, 31.2813904, 28.1549733, 25.1583954, 22.2935057, 19.5621536, 16.966188, 14.5074581, 12.1878131, 10.009102, 7.97317401, 6.08187825, 4.33706384, 2.74057991, 1.29427558, 0, -2.48767446, -4.68547509, -6.58381154, -8.17309347, -9.44373053, -10.3861324, -10.9907087, -11.2478691, -11.1480233, -10.6815809, -9.83895156, -8.61054498, -6.98677078, -4.95803863, -2.51475819, 0.352660882, 4.47148155, 9.96810453, 15.0026226, 19.7598416, 24.4245675, 29.1816062, 34.2157636, 39.7118455, 45.854658, 52.8290068, 60.819698, 70.0115373, 80.5893308, 93.199084, 107.395297, 123.224414, 140.732884, 159.070892, 178.733015, 200.452829, 224.545657, 251.326824, 281.111653, 314.215466, 351.108563, 391.904126, 436.887155, 486.342654, 540.555624, 597.164438, 659.762384, 729.296123, 806.712314, 892.957617, 990.877789, 1097.19927, 1209.55938, 1327.41169, 1453.91792, 1590.57492, 1738.87953, 1903.67536, 2082.35186, 2268.64744, 2455.17475, 2650.99352, 2866.08869, 3103.72697, 3355.79147, 3619.39055, 3889.52939, 4162.77106, 4443.39043, 4731.30013, 5033.63495, 5341.97429, 5653.73081, 5966.3172, 6277.14612, 6612.66025, 6933.98475, 7231.27479, 7494.68553, 7714.37215, 7828.7139, 7914.15914, 7983.0527, 8086.2345, 8220.81174, 8367.85203, 8525.26455, 8690.95849, 8862.84305, 9038.82741, 9149.32129, 9314.36243, 9510.98071, 9716.20596, 9907.06803, 10060.5968, 10153.8221, 10323.8008, 10451.4318},
        {438.485846, 436.542128, 434.427004, 432.144256, 429.697669, 427.091023, 424.328104, 421.412692, 418.348572, 415.139526, 411.789337, 408.301788, 404.680662, 400.929742, 397.05281, 393.05365, 388.936044, 384.703776, 380.360628, 375.910383, 371.356824, 366.703734, 361.954896, 357.114093, 352.185107, 347.171722, 342.07772, 336.906885, 331.662999, 326.349845, 320.971206, 315.530866, 310.032606, 304.480209, 298.850213, 293.160597, 287.429666, 281.662089, 275.862532, 270.035663, 264.186148, 258.318654, 252.437849, 246.5484, 240.654974, 234.762238, 228.874859, 222.997504, 217.13484, 211.291535, 205.472255, 199.861911, 194.281019, 188.730339, 183.210631, 177.722655, 172.26717, 166.844937, 161.456714, 156.103262, 150.785341, 145.50371, 140.259128, 135.052357, 129.884156, 124.755283, 119.6665, 114.538351, 109.293305, 104.097561, 98.9570787, 93.877818, 88.8657386, 83.9268003, 79.0669631, 74.2921866, 69.6084307, 65.0216552, 60.53782, 56.1628849, 51.9028097, 47.7635541, 43.7510781, 39.8713414, 36.3276322, 33.0230956, 29.858505, 26.8334165, 23.9473863, 21.1999704, 18.5907251, 16.1192064, 13.7849705, 11.5875736, 9.52657169, 7.60152098, 5.8119776, 4.15749767, 2.63763734, 1.25195274, 0, -2.44352063, -4.65954577, -6.6332849, -8.34994751, -9.79474309, -10.9528811, -11.8095711, -12.3500226, -12.5594449, -12.4230477, -11.9260404, -11.0536325, -9.79103354, -8.12345292, -6.03610016, -3.51418477, 0.145965747, 5.0575536, 9.6771839, 14.1672647, 18.6902041, 23.4084102, 28.484291, 34.0802546, 40.3587092, 47.4820627, 55.6127232, 64.9130989, 75.5455977, 88.156865, 102.247518, 117.834694, 134.935529, 152.212278, 170.359519, 190.433133, 212.856721, 238.053884, 266.448225, 298.463345, 335.474806, 376.664522, 422.269843, 472.528116, 527.676689, 586.066662, 650.293196, 721.0652, 799.091587, 885.081267, 981.034495, 1084.79053, 1194.8158, 1310.80482, 1434.95941, 1568.35564, 1712.06961, 1869.77267, 2039.3558, 2215.87881, 2392.54111, 2578.39934, 2783.96101, 3012.64669, 3255.86855, 3510.55081, 3771.31318, 4034.48242, 4304.83515, 4582.36826, 4872.95466, 5169.81124, 5470.90089, 5774.18648, 6077.63092, 6415.37049, 6740.15133, 7040.89298, 7306.515, 7525.9369, 7631.45867, 7706.3658, 7763.84335, 7859.22963, 7989.02405, 8132.16218, 8286.02163, 8447.98002, 8615.41496, 8785.70405, 8849.37821, 8986.89309, 9172.8183, 9381.72346, 9588.1782, 9766.75214, 9892.01489, 10062.0147, 10169.0236},
        {446.318603, 444.732393, 442.931549, 440.921441, 438.707439, 436.294912, 433.689228, 430.895758, 427.919871, 424.766936, 421.442323, 417.9514, 414.299538, 410.492105, 406.534471, 402.432005, 398.190077, 393.814055, 389.30931, 384.68121, 379.935125, 375.076424, 370.110477, 365.042653, 359.87832, 354.62285, 349.28161, 343.859971, 338.363301, 332.79697, 327.166347, 321.476802, 315.733704, 309.942422, 304.158695, 298.361648, 292.529801, 286.666889, 280.776646, 274.862808, 268.929109, 262.979283, 257.017066, 251.046192, 245.070397, 239.093414, 233.118979, 227.150826, 221.19269, 215.248307, 209.32141, 203.511245, 197.723965, 191.961236, 186.224721, 180.516085, 174.836991, 169.189104, 163.574089, 157.99361, 152.44933, 146.942914, 141.476028, 136.050333, 130.667496, 125.32918, 120.03705, 114.724357, 109.325996, 103.984726, 98.7066467, 93.4978576, 88.3644578, 83.3125464, 78.3482226, 73.4775856, 68.7067345, 64.0417687, 59.4887872, 55.0538892, 50.743174, 46.5627407, 42.5186886, 38.6171168, 35.1523651, 31.9803496, 28.953522, 26.0686277, 23.3224117, 20.7116194, 18.232996, 15.8832866, 13.6592366, 11.5575912, 9.57509554, 7.7084949, 5.95453451, 4.30995959, 2.77151538, 1.33594711, 0, -4.51196587, -8.50895049, -11.995051, -14.9743646, -17.4509883, -19.4290195, -20.9125552, -21.9056926, -22.4125288, -22.437161, -21.9836864, -21.0562021, -19.6588053, -17.7955932, -15.4706628, -12.6881114, -8.6735851, -3.31997371, 1.60784347, 6.2725802, 10.8369503, 15.4636675, 20.3154456, 25.5549983, 31.3450396, 37.848283, 45.2274425, 53.6452318, 63.2643647, 74.387823, 86.9866209, 101.181392, 117.092768, 134.096966, 152.686827, 173.553706, 197.041562, 223.494354, 253.256039, 286.670579, 325.050185, 367.475877, 414.102171, 465.083586, 520.574638, 579.121276, 642.88873, 712.433659, 788.312723, 871.082583, 962.062208, 1060.11424, 1164.15269, 1274.09999, 1391.93729, 1518.72144, 1655.50927, 1805.79371, 1967.64183, 2136.46314, 2306.44257, 2485.01012, 2679.72373, 2893.40429, 3120.84609, 3360.52774, 3610.52651, 3869.21697, 4136.44537, 4411.22549, 4694.81162, 4983.63233, 5275.92587, 5569.93044, 5863.88427, 6188.81019, 6501.96568, 6793.3928, 7053.13363, 7271.23024, 7388.43553, 7476.94021, 7547.91029, 7646.41884, 7770.21032, 7902.90127, 8042.94253, 8188.78499, 8338.87949, 8491.6769, 8661.37721, 8813.86679, 8956.53731, 9096.78044, 9241.98784, 9399.5512, 9576.86217, 9732.05302, 9854.81788},
        {454.702129, 453.315212, 451.689614, 449.831508, 447.747065, 445.442459, 442.923862, 440.197446, 437.269383, 434.145847, 430.833009, 427.337043, 423.66412, 419.820413, 415.812095, 411.645337, 407.326313, 402.861195, 398.256155, 393.517366, 388.651, 383.66323, 378.560228, 373.348166, 368.033218, 362.621555, 357.11935, 351.532776, 345.868004, 340.131208, 334.32856, 328.466232, 322.550397, 316.587227, 310.674516, 304.770699, 298.83311, 292.864947, 286.869409, 280.849695, 274.809005, 268.750539, 262.677494, 256.59307, 250.500467, 244.402883, 238.303518, 232.20557, 226.11224, 220.026725, 213.952226, 207.927653, 201.919719, 195.930849, 189.963469, 184.020003, 178.102877, 172.214515, 166.357343, 160.533785, 154.746267, 148.997213, 143.289048, 137.624198, 132.005088, 126.434142, 120.913785, 115.381925, 109.778019, 104.237553, 98.7671365, 93.3733755, 88.0628779, 82.842251, 77.7181023, 72.6970393, 67.7856692, 62.9905997, 58.318438, 53.7757916, 49.3692679, 45.1054744, 40.9910185, 37.0325075, 33.6471107, 30.6275189, 27.7614886, 25.0423035, 22.4632477, 20.0176051, 17.6986594, 15.4996947, 13.4139949, 11.4348438, 9.55552543, 7.76932362, 6.06952229, 4.44940536, 2.90225672, 1.4213603, 0, -6.02342328, -11.3684647, -16.0496207, -20.0813879, -23.4782627, -26.2547417, -28.4253212, -30.004498, -31.0067683, -31.4466288, -31.3385759, -30.6971062, -29.5367161, -27.8719021, -25.7171608, -23.0869886, -19.106212, -13.6622317, -8.78189328, -4.28904955, -0.00755346735, 4.23874206, 8.62598411, 13.3303198, 18.5278962, 24.3948603, 31.1073594, 38.8415404, 47.7735504, 58.0488983, 69.8856035, 83.4690045, 98.9844399, 116.429094, 136.082382, 158.273895, 183.245418, 211.238734, 242.495631, 277.257891, 316.681369, 359.815587, 406.723489, 457.46802, 512.112127, 568.75554, 629.915222, 696.144921, 767.998386, 846.029365, 931.513593, 1023.4004, 1120.48713, 1222.70452, 1332.29738, 1450.47154, 1578.43285, 1719.87438, 1872.94945, 2033.09804, 2195.38719, 2365.6496, 2548.6308, 2746.52551, 2957.20614, 3180.52913, 3417.44739, 3668.10162, 3928.61206, 4197.37316, 4472.94546, 4753.53175, 5037.46895, 5323.094, 5608.74384, 5921.60482, 6223.95211, 6506.91031, 6761.60401, 6979.15779, 7108.73768, 7211.39921, 7296.24922, 7397.87148, 7515.28821, 7636.90622, 7762.2785, 7890.95801, 8022.49775, 8156.45069, 8426.94228, 8593.76403, 8694.37168, 8766.22098, 8846.76769, 8973.46756, 9183.77634, 9318.25749, 9449.41222},
        {463.557869, 461.137458, 458.601943, 455.952233, 453.189234, 450.313856, 447.327005, 444.229591, 441.022521, 437.706703, 434.283044, 430.752454, 427.11584, 423.374109, 419.528171, 415.578932, 411.527301, 407.374186, 403.120494, 398.767135, 394.315014, 389.765042, 385.118125, 380.375172, 375.537091, 370.604789, 365.579174, 360.461155, 355.25164, 349.951535, 344.561751, 339.083193, 333.516771, 327.863393, 321.94446, 315.85441, 309.689835, 303.45747, 297.164048, 290.816302, 284.420964, 277.984769, 271.514449, 265.016739, 258.49837, 251.966077, 245.426592, 238.88665, 232.352982, 225.832323, 219.331406, 213.124263, 206.944534, 200.793157, 194.671071, 188.579215, 182.518528, 176.489948, 170.494414, 164.532865, 158.606239, 152.715476, 146.861513, 141.045291, 135.267746, 129.529819, 123.832448, 118.046474, 112.045862, 106.099867, 100.217861, 94.409217, 88.6833075, 83.0495046, 77.5171809, 72.0957087, 66.7944605, 61.6228088, 56.5901258, 51.7057841, 46.9791561, 42.4196142, 38.0365308, 33.8392783, 30.3690972, 27.3582386, 24.5325614, 21.8841778, 19.4052001, 17.0877406, 14.9239115, 12.9058252, 11.025594, 9.27533005, 7.64714575, 6.13315335, 4.72546515, 3.41619343, 2.19745047, 1.06134856, 0, 2.39302093, 4.02963976, 4.98029592, 5.31542884, 5.10547796, 4.4208827, 3.33208251, 1.90951682, 0.223625053, -1.65515335, -3.65637896, -5.70961234, -7.74441406, -9.69034469, -11.4769648, -13.0338349, -13.1366601, -11.5501643, -10.8138196, -10.6099318, -10.6208067, -10.5287501, -10.0160678, -8.76506552, -6.45804915, -2.77732446, 2.59480277, 9.97602674, 19.6840417, 33.817078, 50.2594305, 68.7946322, 89.2062166, 109.322053, 129.903507, 152.233453, 176.682125, 203.619755, 233.416576, 266.442819, 305.852776, 348.385298, 393.865913, 442.120145, 492.97352, 538.451074, 588.128944, 643.78278, 707.188229, 780.120941, 869.574145, 965.728861, 1062.14299, 1156.58721, 1255.43322, 1361.19105, 1476.37073, 1608.8279, 1754.51203, 1903.54102, 2044.88318, 2191.73858, 2357.40353, 2546.65834, 2749.85335, 2963.96251, 3182.79455, 3402.50282, 3630.84633, 3869.0193, 4128.99503, 4399.53027, 4678.08797, 4962.13109, 5249.12259, 5566.95604, 5875.05613, 6163.27815, 6421.47739, 6639.50917, 6759.65601, 6851.06115, 6923.96822, 7020.8175, 7139.1081, 7262.92386, 7390.43465, 7519.81035, 7649.22083, 7776.83598, 7961.68694, 8083.07874, 8166.75418, 8238.45608, 8323.92725, 8448.91049, 8639.14862, 8763.15919, 8821.43837},
        {482.446566, 482.58624, 482.329686, 481.688211, 480.673122, 479.295726, 477.567331, 475.499244, 473.102773, 470.389224, 467.369905, 464.056124, 460.459187, 456.590401, 452.461076, 448.082516, 443.466031, 438.622926, 433.56451, 428.30209, 422.846972, 417.210465, 411.403876, 405.438512, 399.32568, 393.076687, 386.702842, 380.21545, 373.62582, 366.945259, 360.185074, 353.356573, 346.471062, 339.539849, 332.863836, 326.303442, 319.715604, 313.102233, 306.465237, 299.806526, 293.12801, 286.431598, 279.719199, 272.992722, 266.254077, 259.505174, 252.747922, 245.98423, 239.216008, 232.445165, 225.67361, 218.887996, 212.105819, 205.329321, 198.560741, 191.80232, 185.056298, 178.324914, 171.61041, 164.915024, 158.240998, 151.590572, 144.965985, 138.369478, 131.803291, 125.269664, 118.770838, 112.090845, 105.018958, 98.0074518, 91.0727131, 84.231127, 77.4990794, 70.8929557, 64.4291417, 58.1240229, 51.993985, 46.0554137, 40.3246945, 34.8182132, 29.5523553, 24.5435064, 19.8080523, 15.3623786, 12.5947739, 10.8068164, 9.28359457, 7.99697328, 6.91881723, 6.02099118, 5.27535986, 4.65378802, 4.12814039, 3.67028172, 3.25207674, 2.8453902, 2.42208684, 1.9540314, 1.41308862, 0.771123238, 0, -10.6658072, -20.2372517, -28.7499797, -36.2396372, -42.7418704, -48.2923255, -52.9266487, -56.6804861, -59.5894838, -61.689288, -63.0155448, -63.6039005, -63.4900011, -62.7094929, -61.2980219, -59.2912343, -55.3415266, -49.2869378, -44.2918931, -40.0956282, -36.4373785, -33.0563796, -29.6918668, -26.0830757, -21.9692418, -17.0896006, -11.1833876, -3.98983819, 4.75181203, 16.2893587, 29.5346238, 44.4522625, 61.0069301, 78.4040676, 96.9879371, 117.305258, 139.54845, 163.909933, 190.582125, 219.757446, 251.659733, 286.440426, 324.285799, 365.382124, 409.915674, 463.800525, 520.063195, 577.458009, 634.739287, 690.661351, 737.685301, 788.550401, 851.710044, 930.725546, 1019.17023, 1115.10181, 1216.57803, 1316.32036, 1418.93556, 1534.85175, 1674.86347, 1828.03096, 1986.30435, 2146.02531, 2310.97667, 2481.52563, 2657.6228, 2839.52738, 3029.02616, 3227.04187, 3439.18833, 3660.97541, 3891.70199, 4130.66689, 4377.169, 4669.17849, 4957.65505, 5232.22972, 5482.5335, 5698.19743, 5828.07618, 5929.76136, 6010.35955, 6087.72758, 6164.71675, 6239.6991, 6315.74991, 6395.94445, 6483.35798, 6581.06578, 6765.36613, 6934.08176, 7085.70595, 7218.73198, 7331.65315, 7422.96272, 7491.15398, 7750.82173, 7777.91837},
        {487.525079, 487.711493, 487.508556, 486.927144, 485.978133, 484.672399, 483.020818, 481.034267, 478.723622, 476.099758, 473.173552, 469.95588, 466.457618, 462.689642, 458.662828, 454.388053, 449.876192, 445.138122, 440.184718, 435.026858, 429.675416, 424.14127, 418.435295, 412.568367, 406.551363, 400.395158, 394.110629, 387.708652, 381.200103, 374.595858, 367.906793, 361.143784, 354.317708, 347.439441, 340.788854, 334.236669, 327.650372, 321.032109, 314.384027, 307.708273, 301.006993, 294.282335, 287.536445, 280.771469, 273.989555, 267.192848, 260.383496, 253.563645, 246.735442, 239.901034, 233.062566, 226.221382, 219.38045, 212.541934, 205.707998, 198.880806, 192.062523, 185.255313, 178.46134, 171.682767, 164.92176, 158.180481, 151.461096, 144.765769, 138.096663, 131.455942, 124.845772, 118.0475, 110.847777, 103.704183, 96.633196, 89.6512943, 82.7749569, 76.0206624, 69.4048894, 62.9441168, 56.654823, 50.5534869, 44.656587, 38.9806021, 33.5420108, 28.3572918, 23.4429238, 18.8153854, 15.9869842, 14.1948218, 12.6580207, 11.3445173, 10.2222476, 9.25914773, 8.42315394, 7.68220232, 7.00422903, 6.3571702, 5.708962, 5.02754056, 4.28084204, 3.43680257, 2.46335831, 1.32844541, 0, -15.3950997, -29.1975902, -41.4565791, -52.2211743, -61.5404837, -69.4636149, -76.0396757, -81.3177741, -85.3470177, -88.1765143, -89.8553718, -90.4326979, -89.9576005, -88.4791872, -86.046566, -82.7088446, -78.1406904, -72.3377199, -66.1959891, -59.6843686, -52.771729, -45.4269408, -37.6188748, -29.3164014, -20.4883912, -11.1037147, -1.1312427, 9.46015438, 20.7016059, 32.024183, 44.2790948, 57.6774882, 72.4305102, 89.3547215, 108.358562, 129.189028, 151.875643, 176.447929, 202.935409, 231.367606, 260.885942, 292.678332, 326.948057, 363.898399, 403.732638, 450.277937, 499.207727, 549.819319, 601.410025, 653.277156, 700.935619, 752.088069, 811.570183, 881.536071, 958.664976, 1042.09424, 1130.96122, 1220.90213, 1315.35115, 1421.56187, 1548.11084, 1685.91238, 1826.5295, 1966.94823, 2113.34364, 2267.67286, 2433.60154, 2613.52979, 2803.59313, 3003.4707, 3212.50904, 3430.77101, 3658.05084, 3894.14278, 4138.84107, 4436.98979, 4732.07087, 5012.61609, 5267.15724, 5484.2261, 5606.3588, 5698.74653, 5769.63348, 5845.43684, 5927.20734, 6008.42363, 6090.51503, 6174.91087, 6263.04049, 6356.33321, 6555.09938, 6698.72417, 6802.35953, 6881.15744, 6950.26985, 7024.84874, 7120.04606, 7244.05963, 7347.69821},
        {510.463075, 510.629121, 510.411715, 509.820431, 508.864846, 507.554533, 505.899067, 503.908022, 501.590975, 498.957498, 496.017167, 492.779557, 489.254242, 485.450797, 481.378796, 477.047815, 472.467428, 467.64721, 462.596734, 457.325577, 451.843313, 446.159516, 440.283762, 434.225624, 427.994677, 421.600497, 415.052658, 408.360734, 401.534301, 394.582932, 387.516204, 380.343689, 373.074964, 365.719602, 358.464865, 351.227747, 343.923107, 336.554752, 329.126491, 321.642133, 314.105487, 306.520359, 298.89056, 291.219896, 283.512178, 275.771213, 268.000809, 260.204776, 252.386921, 244.551053, 236.70098, 228.848775, 220.989803, 213.127693, 205.266075, 197.408577, 189.55883, 181.720463, 173.897104, 166.092382, 158.309929, 150.553371, 142.82634, 135.132464, 127.475372, 119.858693, 112.286058, 104.451002, 96.054508, 87.7397475, 79.5304518, 71.4503522, 63.5231799, 55.7726664, 48.2225429, 40.8965407, 33.8183912, 27.0118256, 20.5005752, 14.3083715, 8.45894556, 2.97602885, -2.11664735, -6.79535175, -8.97534702, -9.70675535, -10.0644706, -10.0916451, -9.83143116, -9.32698103, -8.62144704, -7.75798148, -6.77973665, -5.72986484, -4.65151836, -3.58784951, -2.58201057, -1.67715385, -0.916431654, -0.34299627, 0, -12.1398936, -23.074185, -32.8386039, -41.46888, -49.0007433, -55.4699234, -60.9121501, -65.3631532, -68.8586625, -71.4344078, -73.1261188, -73.9695253, -74.0003571, -73.254344, -71.7672157, -69.5747021, -66.3294846, -62.0125719, -57.5261124, -52.823754, -47.8591449, -42.5859329, -36.9577662, -30.9282927, -24.4511605, -17.4800175, -9.96851182, -1.87029142, 6.86099565, 16.5003457, 26.7816301, 37.6826077, 49.1810371, 60.7658635, 72.6592523, 85.2137193, 98.5536673, 112.803499, 128.087617, 144.530425, 162.107319, 181.137057, 201.773195, 224.169291, 248.4789, 277.791771, 308.591219, 340.296753, 372.32788, 404.104109, 431.901856, 462.125277, 499.621733, 546.2764, 597.926657, 653.125222, 710.424811, 764.456046, 818.585125, 880.456894, 957.057468, 1042.31519, 1132.46927, 1225.77042, 1323.87772, 1426.88574, 1533.59256, 1643.75664, 1761.89021, 1889.81652, 2034.66272, 2192.13213, 2362.21199, 2544.88956, 2740.1521, 2993.18315, 3247.4746, 3491.71464, 3714.59144, 3904.79318, 4006.00911, 4081.92564, 4140.51624, 4209.90112, 4290.07897, 4372.65398, 4458.04455, 4546.66909, 4638.94601, 4735.2937, 5034.2266, 5194.56273, 5261.17684, 5278.9437, 5292.73804, 5347.43464, 5487.90823, 5647.35593, 5656.12575}
    };
    
    const double b3tab[98][200] = {{98.4097635, 97.4218531, 96.4411126, 95.4672231, 94.4998661, 93.5387228, 92.5834745, 91.6338025, 90.6893881, 89.7499126, 88.8150574, 87.8845036, 86.9579327, 86.0350259, 85.1154645, 84.1989299, 83.2851033, 82.3736659, 81.4642993, 80.5566845, 79.650503, 78.7454359, 77.8411648, 76.9373707, 76.0337351, 75.1299392, 74.2256643, 73.3205918, 72.4144029, 71.5067789, 70.5974011, 69.6859509, 68.7721096, 67.8555584, 66.9148496, 65.9606735, 65.0039743, 64.0451188, 63.0844742, 62.1224073, 61.1592851, 60.1954747, 59.2313429, 58.2672568, 57.3035834, 56.3406896, 55.3789423, 54.4187087, 53.4603556, 52.50425, 51.550759, 50.6195742, 49.691319, 48.7659414, 47.8433894, 46.9236111, 46.0065545, 45.0921678, 44.1803988, 43.2711958, 42.3645067, 41.4602796, 40.5584626, 39.6590037, 38.7618509, 37.8669523, 36.974256, 36.1089066, 35.2954438, 34.4818496, 33.6664387, 32.8475255, 32.023425, 31.1924515, 30.35292, 29.5031449, 28.641441, 27.7661228, 26.8755052, 25.9679026, 25.0416299, 24.0950016, 23.1263323, 22.1339369, 20.9363852, 19.6256451, 18.2958444, 16.9511309, 15.5956524, 14.2335564, 12.8689908, 11.5061032, 10.1490414, 8.8019531, 7.46898596, 6.15428772, 4.86200609, 3.5962888, 2.36128354, 1.16113803, 0, -0.552595377, -1.10657047, -1.66258326, -2.22129173, -2.78335387, -3.34942765, -3.92017106, -4.49624209, -5.07829871, -5.6669989, -6.26300066, -6.86696196, -7.47954079, -8.10139513, -8.73318296, -9.37556226, -9.91080302, -10.3226506, -10.8795455, -11.5567767, -12.3296334, -13.1734047, -14.0633798, -14.9748477, -15.8830976, -16.7634187, -17.5910999, -18.3414305, -18.9896996, -19.4608292, -19.7989435, -19.9944416, -20.037723, -20.0986524, -20.0778963, -19.8282636, -19.286314, -18.3886069, -17.0717019, -15.2721587, -13.427087, -10.8201547, -7.28998774, -2.6752119, 3.18554698, 9.10403895, 16.9286683, 27.1582154, 40.2914602, 56.8271831, 76.8795517, 101.802041, 131.014707, 165.124392, 207.160952, 259.06636, 322.782588, 397.969448, 489.369747, 604.215915, 744.800208, 920.138539, 1128.23309, 1376.76718, 1689.82832, 2083.97422, 2559.82099, 3129.79331, 3864.76853, 4802.55989, 5705.39299, 6929.06677, 8606.79018, 10871.7721, 13857.2216, 21410.5841, 29022.2728, 35896.9375, 41239.2281, 44253.7943, 40818.9684, 35683.2629, 29476.8921, 24581.4137, 20933.8023, 17741.3059, 14977.3854, 12615.5016, 10629.1152, 8991.68704, 7726.01004, 6753.20196, 6025.06843, 5493.41509, 5110.04759, 4826.77156, 4595.39264, 4501.99075, 4555.25062},
        {88.9798529, 87.8348171, 86.7184098, 85.6295658, 84.5672202, 83.530308, 82.5177643, 81.5285241, 80.5615224, 79.6156942, 78.6899746, 77.7832986, 76.8946011, 76.0228174, 75.1668822, 74.3257308, 73.4982981, 72.6835191, 71.8803289, 71.0876624, 70.3044548, 69.529641, 68.7621561, 68.000935, 67.2449129, 66.4930247, 65.7442054, 64.9973901, 64.2515139, 63.5055117, 62.7583185, 62.0088695, 61.2560995, 60.4989437, 59.6890172, 58.8499104, 58.0057821, 57.157103, 56.3043438, 55.4479751, 54.5884676, 53.7262918, 52.8619185, 51.9958182, 51.1284617, 50.2603195, 49.3918624, 48.5235609, 47.6558857, 46.7893074, 45.9242967, 45.0820481, 44.2418592, 43.4037513, 42.5677458, 41.733864, 40.9021274, 40.0725574, 39.2451753, 38.4200025, 37.5970605, 36.7763705, 35.9579541, 35.1418325, 34.3280272, 33.5165596, 32.707451, 31.9232276, 31.185875, 30.4490005, 29.7111665, 28.9709355, 28.2268701, 27.4775327, 26.7214858, 25.957292, 25.1835137, 24.3987135, 23.6014538, 22.7902971, 21.963806, 21.120543, 20.2590704, 19.377951, 18.3371694, 17.2074917, 16.0613499, 14.9018033, 13.7319118, 12.5547348, 11.373332, 10.1907631, 9.01008752, 7.83436501, 6.66665514, 5.51001755, 4.36751183, 3.2421976, 2.13713447, 1.05538207, 0, -0.56106197, -1.13680422, -1.72665807, -2.33005481, -2.94642576, -3.57520222, -4.2158155, -4.8676969, -5.53027773, -6.2029893, -6.88526292, -7.57652989, -8.27622151, -8.9837691, -9.69860396, -10.4201574, -11.1029521, -11.7400038, -12.4323229, -13.1697175, -13.9419955, -14.738965, -15.550434, -16.3662105, -17.1761027, -17.9699184, -18.7374657, -19.4685527, -20.1529873, -20.8579803, -21.4675561, -21.9483018, -22.2668046, -22.4279774, -22.3792446, -22.0578105, -21.4187644, -20.4171961, -19.0081949, -17.1468505, -15.6015742, -13.2666009, -9.93789194, -5.41140867, 0.516887584, 7.2025123, 15.9101582, 27.0559946, 41.0561912, 58.3269173, 78.4837837, 103.721979, 133.669421, 169.129101, 213.303118, 268.316423, 336.293964, 416.724478, 514.968267, 639.261501, 792.584956, 984.268113, 1210.08951, 1478.40685, 1818.89253, 2251.43514, 2782.06654, 3427.08283, 4253.58821, 5299.94696, 6336.96084, 7711.72008, 9555.20678, 11998.403, 15172.2908, 22875.94, 30655.223, 37724.0999, 43296.5308, 46586.4758, 43580.2304, 38871.1957, 33056.6169, 28335.1272, 24670.0886, 21357.6182, 18394.4405, 15777.2799, 13502.8608, 11567.9076, 9939.42403, 8731.20653, 7867.48849, 7272.50328, 6870.4843, 6585.66493, 6342.27857, 6330.25421, 6662.21957},
        {81.2404201, 80.2072505, 79.1926537, 78.1960443, 77.2168371, 76.2544466, 75.3082876, 74.3777746, 73.4623224, 72.5613455, 71.6742586, 70.8004764, 69.9394135, 69.0904845, 68.2531042, 67.4266872, 66.6106481, 65.8044015, 65.0073622, 64.2189447, 63.4385638, 62.665634, 61.89957, 61.1397865, 60.3856981, 59.6367195, 58.8922654, 58.1517502, 57.4145888, 56.6801958, 55.9479858, 55.2173735, 54.4877735, 53.7586004, 53.0097581, 52.2508271, 51.4916222, 50.7321914, 49.9725823, 49.2128429, 48.453021, 47.6931644, 46.9333208, 46.1735381, 45.4138642, 44.6543468, 43.8950337, 43.1359728, 42.3772119, 41.6187988, 40.8607812, 40.0971879, 39.3342163, 38.5720448, 37.8108515, 37.0508149, 36.2921132, 35.5349247, 34.7794278, 34.0258007, 33.2742217, 32.5248691, 31.7779213, 31.0335565, 30.291953, 29.5532892, 28.8177433, 28.1086091, 27.4486253, 26.7902964, 26.1323022, 25.4733225, 24.8120371, 24.1471258, 23.4772685, 22.8011449, 22.117435, 21.4248184, 20.721975, 20.0075847, 19.2803273, 18.5388825, 17.7819303, 17.0081503, 16.1108422, 15.1435925, 14.1612542, 13.1659269, 12.1597102, 11.1447039, 10.1230074, 9.09672035, 8.06794244, 7.03877323, 6.01131234, 4.98765939, 3.96991398, 2.96017572, 1.96054423, 0.973119121, 0, -0.398298167, -0.859759607, -1.38016123, -1.95527997, -2.58089272, -3.25277641, -3.96670796, -4.71846428, -5.50382228, -6.31855889, -7.15845102, -8.01927558, -8.8968095, -9.78682968, -10.6851131, -11.5874365, -12.5111084, -13.4549814, -14.3661307, -15.244947, -16.0918211, -16.9071438, -17.6913058, -18.444698, -19.1677111, -19.8607358, -20.5241631, -21.1583835, -21.763788, -22.4655822, -23.0935765, -23.6107172, -23.9799505, -24.2285323, -24.2872541, -24.0697583, -23.5196984, -22.580728, -21.1965004, -19.3106693, -17.6298974, -15.1026087, -11.523172, -6.68595658, -0.385331315, 6.35140837, 15.2710514, 26.8874603, 41.7144976, 60.2660258, 82.4843757, 110.153481, 142.346685, 179.756481, 226.381113, 284.734613, 357.331009, 444.387191, 551.236408, 685.717877, 848.846951, 1054.22502, 1301.93616, 1602.4419, 1983.78779, 2465.94805, 3046.99376, 3741.22052, 4643.2354, 5802.21675, 7032.35294, 8653.96463, 10797.5728, 13593.6986, 17172.8631, 25530.0904, 33965.2726, 41642.805, 47727.0826, 51382.5009, 48353.1183, 43503.8912, 37465.0732, 32559.6849, 28747.9267, 25284.679, 22165.4081, 19385.58, 16940.661, 14826.1172, 13107.975, 11711.4202, 10596.3588, 9722.69702, 9050.34097, 8539.19682, 8149.17071, 7933.18319, 7994.09837},
        {74.8104589, 73.8549822, 72.9129376, 71.9840367, 71.0679913, 70.1645129, 69.2733133, 68.394104, 67.5265968, 66.6705033, 65.8255351, 64.9914039, 64.1678214, 63.3544992, 62.5511489, 61.7574823, 60.9732109, 60.1980465, 59.4317007, 58.6738851, 57.9243113, 57.1826912, 56.4487362, 55.7221582, 55.0026686, 54.2899792, 53.5838016, 52.8838475, 52.1898286, 51.5014564, 50.8184427, 50.1404991, 49.4673373, 48.7986688, 48.1390194, 47.4855924, 46.8355335, 46.188398, 45.5437413, 44.901119, 44.2600864, 43.620199, 42.9810122, 42.3420814, 41.7029621, 41.0632096, 40.4223796, 39.7800273, 39.1357082, 38.4889777, 37.8393913, 37.1589304, 36.4753222, 35.7887199, 35.0992766, 34.4071455, 33.7124799, 33.0154328, 32.3161576, 31.6148073, 30.9115351, 30.2064943, 29.499838, 28.7917193, 28.0822916, 27.3717079, 26.6601215, 25.9575591, 25.2738105, 24.5886652, 23.9016365, 23.2122373, 22.5199809, 21.8243803, 21.1249486, 20.4211991, 19.7126447, 18.9987986, 18.279174, 17.5532838, 16.8206414, 16.0807597, 15.3331519, 14.5773311, 13.7713388, 12.9362961, 12.0938228, 11.2447779, 10.3900203, 9.53040901, 8.66680295, 7.80006107, 6.93104234, 6.0606057, 5.18961011, 4.31891452, 3.44937789, 2.58185917, 1.71721731, 0.856311269, 0, -0.427156002, -0.920612215, -1.47572082, -2.08783401, -2.75230395, -3.46448284, -4.21972285, -5.01337617, -5.84079498, -6.69733147, -7.57833782, -8.47916621, -9.39516883, -10.3216979, -11.2541055, -12.1877439, -13.144446, -14.1233471, -15.0599023, -15.9551381, -16.8100812, -17.6257582, -18.4031959, -19.1434208, -19.8474596, -20.516339, -21.1510855, -21.7527259, -22.3222867, -22.9882109, -23.5773896, -24.0526246, -24.3767175, -24.5546693, -24.5281822, -24.2277052, -23.6033804, -22.6053497, -21.1837551, -19.2887388, -17.9715463, -15.7460976, -12.3471011, -7.50926552, -0.967299594, 6.71474737, 16.8388426, 29.8776125, 46.3036837, 66.5896827, 90.2037862, 119.850731, 155.033024, 196.710728, 248.819759, 313.959935, 394.731077, 490.636326, 608.075967, 756.82848, 938.590021, 1168.514, 1446.97003, 1785.78438, 2216.19604, 2760.53322, 3415.77273, 4197.6702, 5214.93653, 6523.70187, 7881.98437, 9689.88956, 12106.7402, 15291.8591, 19404.5689, 29155.3998, 39014.6653, 48003.8866, 55144.5844, 59458.2797, 55948.6998, 50333.6885, 43356.6785, 37816.3259, 33642.5374, 29908.8763, 26588.0663, 23652.8313, 21075.895, 18829.9811, 16911.0343, 15249.2103, 13824.9694, 12618.7722, 11611.0789, 10782.35, 10113.0458, 9670.86568, 9453.89022},
        {70.5057947, 69.7679365, 69.0276573, 68.2852418, 67.5409745, 66.79514, 66.0480227, 65.2999073, 64.5510783, 63.8018202, 63.0524177, 62.3031552, 61.5543173, 60.8061885, 60.0590535, 59.3131967, 58.5689028, 57.8264562, 57.0861416, 56.3482434, 55.6130463, 54.8808347, 54.1518933, 53.4265065, 52.704959, 51.9875353, 51.2745199, 50.5661974, 49.8628524, 49.1647694, 48.4722329, 47.7855275, 47.1049378, 46.4307483, 45.794177, 45.1793912, 44.5701862, 43.9658427, 43.3656413, 42.7688628, 42.1747879, 41.5826973, 40.9918716, 40.4015917, 39.8111381, 39.2197915, 38.6268328, 38.0315426, 37.4332015, 36.8310903, 36.2244897, 35.5689972, 34.9085236, 34.2432967, 33.573544, 32.8994933, 32.2213721, 31.5394083, 30.8538294, 30.164863, 29.472737, 28.7776788, 28.0799163, 27.379677, 26.6771887, 25.9726789, 25.2663754, 24.5645378, 23.8732808, 23.1803917, 22.4857071, 21.7890637, 21.0902981, 20.3892469, 19.6857468, 18.9796344, 18.2707463, 17.5589193, 16.8439898, 16.1257947, 15.4041704, 14.6789537, 13.9499812, 13.2170895, 12.4660893, 11.7041252, 10.9385104, 10.1695368, 9.39749613, 8.62268029, 7.84538106, 7.06589023, 6.28449962, 5.50150105, 4.71718632, 3.93184724, 3.14577563, 2.35926329, 1.57260203, 0.786083661, 0, -0.685887011, -1.38660053, -2.10120513, -2.82876541, -3.56834595, -4.31901134, -5.07982615, -5.84985497, -6.62816239, -7.41381299, -8.20587136, -9.00340207, -9.80546972, -10.6111389, -11.4194742, -12.2295401, -12.9826452, -13.6696031, -14.4201173, -15.2208762, -16.0585678, -16.9198806, -17.7915028, -18.6601227, -19.5124284, -20.3351084, -21.1148508, -21.8383439, -22.4922761, -23.1023504, -23.6019348, -23.966013, -24.1695687, -24.315273, -24.3142661, -24.043638, -23.4400664, -22.4402286, -20.9808024, -18.9984652, -17.6437639, -15.2700683, -11.5765598, -6.26241954, 0.973170935, 9.79304078, 21.2954955, 35.9408512, 54.1894238, 76.5015293, 101.847082, 133.998402, 172.78087, 219.467454, 278.286614, 352.139856, 443.928684, 553.126946, 686.842816, 855.923731, 1062.50674, 1322.88343, 1633.42403, 2007.318, 2486.25411, 3098.83947, 3848.38925, 4756.95339, 5939.31898, 7457.81573, 8855.04795, 10793.0285, 13513.9921, 17260.1733, 22273.8068, 35070.8478, 48051.3803, 59889.2088, 69258.1379, 74831.9722, 69771.4737, 61938.8509, 52370.641, 45113.2811, 40011.8893, 35657.4588, 31957.8141, 28820.7798, 26154.1801, 23865.8397, 21546.9816, 19516.508, 17746.0677, 16207.3095, 14871.8822, 13711.4347, 12697.6157, 11921.8916, 11430.1549},
        {68.2331449, 67.8716052, 67.4764804, 67.0490866, 66.5907398, 66.1027561, 65.5864517, 65.0431426, 64.4741449, 63.8807747, 63.2643481, 62.6261812, 61.9675902, 61.289891, 60.5943998, 59.8824326, 59.1553057, 58.4143349, 57.6608366, 56.8961267, 56.1215213, 55.3383365, 54.5478885, 53.7514933, 52.9504671, 52.1461258, 51.3397856, 50.5327626, 49.7263729, 48.9219326, 48.1207578, 47.3241645, 46.5334689, 45.749987, 45.0387841, 44.3679607, 43.7048514, 43.0487035, 42.3987643, 41.7542811, 41.1145012, 40.478672, 39.8460407, 39.2158547, 38.5873612, 37.9598077, 37.3324413, 36.7045095, 36.0752595, 35.4439386, 34.8097942, 34.1111111, 33.4094206, 32.7052917, 31.999293, 31.2919936, 30.5839623, 29.8757679, 29.1679793, 28.4611654, 27.755895, 27.052737, 26.3522602, 25.6550335, 24.9616258, 24.272606, 23.5885428, 22.9344663, 22.3348194, 21.7397214, 21.1481555, 20.5591049, 19.9715527, 19.384482, 18.796876, 18.2077179, 17.6159907, 17.0206777, 16.4207619, 15.8152265, 15.2030546, 14.5832295, 13.9547341, 13.3165518, 12.6132307, 11.8721165, 11.1212088, 10.3612574, 9.59301188, 8.81722191, 8.03463715, 7.24600724, 6.45208186, 5.65361064, 4.85134325, 4.04602934, 3.23841858, 2.4292606, 1.61930508, 0.809301659, 0, -0.957080673, -1.87269606, -2.75023109, -3.59307069, -4.40459979, -5.18820332, -5.9472662, -6.68517337, -7.40530975, -8.11106027, -8.80580985, -9.49294344, -10.1758459, -10.8579023, -11.5424975, -12.2330163, -12.7644015, -13.1159747, -13.6721214, -14.4001317, -15.2672958, -16.2409038, -17.2882458, -18.3766121, -19.4732926, -20.5455777, -21.5607574, -22.4861219, -23.2889613, -23.8350836, -24.2304713, -24.4728591, -24.5599818, -24.8676177, -25.2044798, -25.2784696, -24.9739089, -24.1751196, -22.7664233, -20.6321419, -18.3586427, -14.9145359, -10.0467865, -3.50235937, 4.97178035, 14.1041722, 26.0534705, 41.4538341, 60.9394219, 85.1443927, 113.354141, 149.20008, 191.704115, 242.120533, 306.261401, 387.892433, 490.779347, 616.533441, 771.564488, 964.632539, 1195.08486, 1483.54185, 1823.88641, 2231.78557, 2759.81833, 3443.62048, 4288.51821, 5322.28927, 6683.84666, 8449.239, 9821.49196, 11883.9883, 15014.977, 19592.7068, 25995.4267, 43645.4879, 61616.0112, 78024.2201, 90987.3378, 98622.5877, 91171.0812, 79876.8947, 66232.7278, 56183.6931, 49460.077, 43936.9938, 39437.4885, 35784.606, 32801.3911, 30310.889, 27484.6564, 24974.5605, 22752.0561, 20788.598, 19055.6409, 17524.6397, 16167.0491, 14883.7163, 13946.4929},
        {66.0053166, 65.6281361, 65.2231028, 64.7912757, 64.3337139, 63.8514762, 63.3456218, 62.8172097, 62.2672989, 61.6969484, 61.1072172, 60.4991643, 59.8738489, 59.2323298, 58.5756661, 57.9049169, 57.221141, 56.5253977, 55.8187458, 55.1022445, 54.3769526, 53.6439293, 52.9042335, 52.1589244, 51.4090608, 50.6557018, 49.8999065, 49.1427338, 48.3852428, 47.6284924, 46.8735418, 46.1214499, 45.3732758, 44.6300784, 43.944173, 43.2899123, 42.6410333, 41.9969314, 41.3570024, 40.7206419, 40.0872456, 39.4562091, 38.8269282, 38.1987984, 37.5712154, 36.9435748, 36.3152724, 35.6857038, 35.0542646, 34.4203505, 33.7833571, 33.0802979, 32.374303, 31.6661206, 30.9564985, 30.2461847, 29.5359272, 28.8264741, 28.1185733, 27.4129727, 26.7104204, 26.0116643, 25.3174524, 24.6285327, 23.9456533, 23.269562, 22.6010068, 21.9730387, 21.4179328, 20.869815, 20.3273391, 19.7891591, 19.2539288, 18.7203022, 18.1869332, 17.6524758, 17.1155838, 16.5749112, 16.0291118, 15.4768397, 14.9167487, 14.3474927, 13.7677257, 13.1761016, 12.504677, 11.7868053, 11.0566403, 10.314997, 9.5626906, 8.80053627, 8.0293491, 7.24994423, 6.46313679, 5.66974191, 4.8705747, 4.06645031, 3.25818385, 2.44659047, 1.63248528, 0.816683412, 0, -1.08662362, -2.10919402, -3.07279772, -3.98252127, -4.84345117, -5.66067396, -6.43927617, -7.18434431, -7.90096492, -8.59422453, -9.26920965, -9.93100681, -10.5847025, -11.2353834, -11.8881358, -12.5480464, -13.0343415, -13.3255562, -13.847175, -14.5644572, -15.4426623, -16.4470495, -17.5428783, -18.6954078, -19.8698976, -21.0316069, -22.145795, -23.1777214, -24.0926454, -24.7506379, -25.2607157, -25.6196947, -25.8243907, -26.299602, -26.8281533, -27.0787405, -26.9197846, -26.2197067, -24.8469279, -22.6698693, -20.1461067, -16.3755985, -11.1114963, -4.10695181, 4.88488329, 14.2868534, 26.6258118, 42.6046077, 62.9260905, 88.2931094, 117.847202, 155.760799, 200.724854, 254.143893, 322.962658, 411.63845, 524.628565, 664.72677, 838.431973, 1054.05784, 1309.2248, 1628.74414, 2007.48735, 2463.54937, 3054.37608, 3819.35061, 4756.40382, 5894.17195, 7413.28308, 9408.39011, 10671.5246, 12800.3641, 16340.4689, 21837.3992, 29836.7154, 53011.6194, 76748.1194, 98559.8653, 115960.507, 126463.694, 117317.093, 103144.326, 85858.7506, 73007.8889, 64274.9079, 56995.4062, 50969.9293, 45999.0231, 41883.2329, 38423.1045, 34967.6531, 31788.3538, 28892.1183, 26285.8581, 23976.4848, 21970.9102, 20276.0458, 18574.6914, 17162.8935},
        {63.1686145, 62.3494887, 61.5480896, 60.7638627, 59.9962534, 59.2447069, 58.5086689, 57.7875846, 57.0808994, 56.3880588, 55.7085081, 55.0416928, 54.3870583, 53.7440499, 53.112113, 52.4906931, 51.8792356, 51.2771858, 50.6839891, 50.099091, 49.5219369, 48.9519721, 48.3886421, 47.8313922, 47.2796678, 46.7329145, 46.1905774, 45.6521021, 45.116934, 44.5845184, 44.0543008, 43.5257265, 42.998241, 42.4712897, 41.9388064, 41.4031083, 40.8665786, 40.3288417, 39.7895217, 39.2482429, 38.7046296, 38.1583061, 37.6088965, 37.0560252, 36.4993163, 35.9383943, 35.3728832, 34.8024075, 34.2265912, 33.6450588, 33.0574344, 32.416809, 31.770349, 31.1186872, 30.4624568, 29.8022907, 29.1388219, 28.4726836, 27.8045085, 27.1349299, 26.4645806, 25.7940937, 25.1241021, 24.455239, 23.7881373, 23.12343, 22.4617501, 21.8246744, 21.2332767, 20.6449951, 20.0591051, 19.4748818, 18.8916007, 18.308537, 17.7249661, 17.1401632, 16.5534037, 15.9639629, 15.371116, 14.7741385, 14.1723056, 13.5648927, 12.951175, 12.3304278, 11.674562, 10.9971106, 10.3119357, 9.61920079, 8.9190691, 8.21170399, 7.49726883, 6.77592695, 6.0478417, 5.31317643, 4.57209448, 3.8247592, 3.07133395, 2.31198205, 1.54686686, 0.77615173, 0, -1.00275613, -1.96597968, -2.89285931, -3.78658362, -4.65034125, -5.48732083, -6.30071099, -7.09370036, -7.86947755, -8.63123121, -9.38214996, -10.1254224, -10.8642372, -11.601783, -12.3412484, -13.0858221, -13.7498989, -14.3239828, -15.0121052, -15.7984273, -16.6671108, -17.6023169, -18.5882072, -19.6089429, -20.6486857, -21.6915968, -22.7218377, -23.7235698, -24.6809545, -25.6348561, -26.4919422, -27.2193632, -27.7842698, -28.362043, -28.8157182, -28.9528025, -28.6779774, -27.8959242, -26.5113245, -24.4288597, -22.6761135, -19.6931122, -15.1648388, -8.77627631, -0.212407996, 9.62135727, 22.5655686, 39.2403494, 60.265823, 86.2621126, 115.596171, 153.895168, 200.326379, 256.683714, 330.123785, 425.395461, 547.247608, 697.846126, 885.109891, 1119.77556, 1399.88178, 1753.41237, 2179.41124, 2698.09892, 3365.57859, 4221.48271, 5248.34783, 6471.00375, 8123.63201, 10321.993, 11453.6356, 13628.4104, 17560.3046, 23963.3057, 33551.4011, 61952.0109, 91237.3315, 118392.992, 140404.622, 154257.85, 144791.159, 129235.423, 109782.189, 94844.4072, 84150.9892, 74838.5944, 66765.746, 59790.9672, 53772.7809, 48569.7104, 44312.5861, 40243.1635, 36428.2723, 32934.7423, 29829.4032, 27179.0849, 25050.6171, 22992.178, 21082.7579},
        {60.9968525, 60.0487976, 59.1332199, 58.2490411, 57.3951829, 56.5705672, 55.7741158, 55.0047504, 54.2613927, 53.5429647, 52.8483881, 52.1765846, 51.526476, 50.8969842, 50.287031, 49.695538, 49.1214271, 48.5636201, 48.0210387, 47.4926048, 46.9772401, 46.4738664, 45.9814055, 45.4987792, 45.0249092, 44.5587174, 44.0991256, 43.6450554, 43.1954288, 42.7491674, 42.3051931, 41.8624277, 41.4197929, 40.9762105, 40.505292, 40.0191467, 39.5301881, 39.0381594, 38.5428038, 38.0438643, 37.5410842, 37.0342065, 36.5229746, 36.0071314, 35.4864202, 34.9605841, 34.4293662, 33.8925098, 33.349758, 32.8008539, 32.2455407, 31.6485575, 31.0454103, 30.4366011, 29.8226318, 29.2040045, 28.5812211, 27.9547836, 27.3251938, 26.6929539, 26.0585658, 25.4225315, 24.7853529, 24.147532, 23.5095707, 22.8719712, 22.2352352, 21.6154724, 21.0284176, 20.4423833, 19.8568598, 19.2713371, 18.6853055, 18.0982552, 17.5096763, 16.9190592, 16.3258939, 15.7296707, 15.1298798, 14.5260113, 13.9175555, 13.3040026, 12.6848427, 12.059566, 11.4111478, 10.7476831, 10.0774655, 9.40052106, 8.71687599, 8.02655643, 7.3295885, 6.62599834, 5.9158121, 5.19905592, 4.47575594, 3.74593829, 3.00962912, 2.26685456, 1.51764077, 0.762013865, 0, -0.768548374, -1.54183514, -2.32015613, -3.10380719, -3.89308415, -4.68828286, -5.48969916, -6.29762888, -7.11236787, -7.93421196, -8.763457, -9.60039882, -10.4453333, -11.2985562, -12.1603634, -13.0310507, -13.9028886, -14.7750265, -15.6659129, -16.5741238, -17.4982353, -18.4368237, -19.388465, -20.3517353, -21.3252107, -22.3074674, -23.2970814, -24.2926289, -25.292686, -26.4668868, -27.5800282, -28.5793687, -29.4121672, -30.1063509, -30.5688442, -30.6850599, -30.3780563, -29.5708913, -28.1866233, -26.1483103, -24.8381146, -22.2759151, -18.0992929, -11.9458294, -3.45310576, 6.44432283, 19.7300925, 37.0908656, 59.2133043, 86.7840707, 118.145146, 159.193595, 208.964103, 269.349218, 348.076299, 450.252995, 580.986953, 742.283366, 943.057442, 1195.60889, 1497.24309, 1880.42382, 2347.26974, 2920.79952, 3659.44582, 4605.38569, 5730.27107, 7057.99473, 8871.04156, 11305.6221, 12476.1984, 14851.7677, 19268.3768, 26562.0722, 37568.9006, 70007.2379, 103610.219, 134993.309, 160771.97, 177561.669, 168610.414, 152812.761, 132513.084, 116523.942, 104629.386, 93918.3878, 84309.7511, 75722.2798, 68074.7776, 61286.0483, 55937.3572, 50795.1473, 45936.8908, 41440.0598, 37382.1264, 33840.5629, 30892.8413, 28070.2204, 25456.1132},
        {58.9297241, 58.168896, 57.4238496, 56.6941269, 55.9792701, 55.2788211, 54.592322, 53.919315, 53.2593421, 52.6119453, 51.9766668, 51.3530485, 50.7406327, 50.1389613, 49.5475764, 48.9660201, 48.3938344, 47.8305615, 47.2757434, 46.7289222, 46.1896399, 45.6574387, 45.1318605, 44.6124475, 44.0987417, 43.5902853, 43.0866202, 42.5872886, 42.0918325, 41.5997939, 41.1107151, 40.624138, 40.1396047, 39.6566572, 39.1716171, 38.6857045, 38.2001781, 37.7146846, 37.2288705, 36.7423826, 36.2548672, 35.7659711, 35.2753407, 34.7826228, 34.2874639, 33.7895105, 33.2884093, 32.7838068, 32.2753497, 31.7626845, 31.2454578, 30.6958167, 30.1415035, 29.5827609, 29.0198314, 28.452958, 27.8823833, 27.30835, 26.7311008, 26.1508785, 25.5679257, 24.9824853, 24.3947999, 23.8051122, 23.213665, 22.620701, 22.0264629, 21.440514, 20.872194, 20.3025224, 19.7311378, 19.1576786, 18.5817833, 18.0030905, 17.4212387, 16.8358663, 16.2466119, 15.6531139, 15.0550109, 14.4519414, 13.8435438, 13.2294567, 12.6093186, 11.982768, 11.3335568, 10.669601, 9.99900743, 9.32193014, 8.63852317, 7.94894059, 7.25333644, 6.55186478, 5.84467965, 5.13193511, 4.41378521, 3.69038399, 2.96188551, 2.22844382, 1.49021297, 0.747347017, 0, -0.318773237, -0.717735852, -1.19248897, -1.7386337, -2.35177118, -3.02750252, -3.76142885, -4.54915128, -5.38627095, -6.26838897, -7.19110646, -8.15002455, -9.14074436, -10.158867, -11.1999936, -12.2597253, -13.4039594, -14.6383394, -15.799463, -16.8979949, -17.9445995, -18.9499416, -19.9246857, -20.8794964, -21.8250383, -22.7719761, -23.7309742, -24.7126974, -25.7278102, -27.0419387, -28.3173001, -29.4880706, -30.4884265, -31.2494559, -31.7068791, -31.7972395, -31.4556397, -30.6171821, -29.2169695, -27.1901041, -26.2642516, -24.0363889, -20.0908997, -14.0121677, -5.38457666, 4.37790815, 17.97788, 36.2883506, 60.1823316, 90.5328348, 126.310275, 172.615657, 227.992077, 294.181131, 379.454688, 489.152656, 628.614947, 799.44964, 1011.57662, 1278.98684, 1597.97097, 2006.05501, 2506.84936, 3126.65018, 3930.19335, 4964.57031, 6197.00333, 7653.87704, 9654.42951, 12352.2444, 13800.294, 16575.9442, 21559.9131, 29632.9189, 41675.68, 75996.7852, 111692.114, 145285.418, 173300.446, 192260.95, 185280.706, 171233.421, 152389.977, 137090.5, 125203.464, 114068.488, 103680.488, 94034.3773, 85125.0724, 76947.4877, 70163.4882, 63737.8602, 57695.2226, 52060.194, 46857.3933, 42111.4391, 37846.9502, 33803.112, 30274.0846},
        {57.0382514, 56.5403257, 56.0334888, 55.5181921, 54.9948874, 54.4640262, 53.9260601, 53.3814406, 52.8306194, 52.2740481, 51.7121782, 51.1454614, 50.5743492, 49.9992933, 49.4207451, 48.8391564, 48.2549786, 47.6686635, 47.0806625, 46.4914273, 45.9014095, 45.3110606, 44.7208322, 44.131176, 43.5425435, 42.9553863, 42.3701561, 41.7873043, 41.2072826, 40.6305426, 40.0575358, 39.488714, 38.9245285, 38.3654311, 37.8416912, 37.3382242, 36.8395873, 36.3452644, 35.8547394, 35.3674964, 34.8830192, 34.4007918, 33.9202982, 33.4410222, 32.9624479, 32.4840591, 32.0053398, 31.5257741, 31.0448457, 30.5620387, 30.076837, 29.562796, 29.0458903, 28.5261659, 28.0036688, 27.478445, 26.9505406, 26.4200016, 25.886874, 25.3512039, 24.8130372, 24.27242, 23.7293984, 23.1840183, 22.6363257, 22.0863667, 21.5341874, 20.9857232, 20.4467685, 19.9052223, 19.360749, 18.8130128, 18.2616779, 17.7064086, 17.146869, 16.5827234, 16.013636, 15.4392711, 14.8592929, 14.2733656, 13.6811534, 13.0823206, 12.4765313, 11.8634499, 11.222813, 10.5646677, 9.89930067, 9.22702291, 8.54814531, 7.86297878, 7.17183424, 6.47502261, 5.7728548, 5.06564173, 4.35369432, 3.63732348, 2.91684014, 2.1925552, 1.46477958, 0.733824213, 0, 0.037427687, -0.0628779014, -0.293098614, -0.645416298, -1.1120128, -1.68506998, -2.35676967, -3.11929373, -3.964824, -4.88554234, -5.87363059, -6.9212706, -8.02064422, -9.16393329, -10.3433197, -11.5509852, -12.89127, -14.3723786, -15.7328014, -16.9887542, -18.1564528, -19.2521129, -20.2919504, -21.2921809, -22.2690202, -23.2386842, -24.2173885, -25.2213489, -26.2667813, -27.6545981, -29.0119295, -30.2695823, -31.3583632, -32.1943209, -32.7156412, -32.8644453, -32.5759675, -31.7854417, -30.4281024, -28.4391834, -27.7922119, -25.7637792, -21.8903231, -15.7082813, -6.75409164, 3.06766245, 17.1807253, 36.6406957, 62.5031722, 95.8237535, 136.168557, 187.90314, 249.118248, 321.422707, 413.608092, 531.241067, 679.8883, 860.548518, 1084.39449, 1367.5822, 1706.78992, 2142.29237, 2675.10564, 3333.94076, 4194.43634, 5310.69048, 6652.31972, 8251.51954, 10450.251, 13415.2543, 15263.8444, 18527.4828, 24082.3257, 32804.5298, 45570.2513, 80433.7284, 116798.515, 151246.249, 180358.564, 200717.098, 195998.435, 184292.63, 167712.054, 153865.621, 142689.989, 131831.591, 121341.597, 111271.178, 101671.504, 92593.7432, 84774.2277, 77262.4885, 70083.5929, 63262.6085, 56824.6024, 50794.6421, 45197.795, 39967.4726, 35426.6465},
        {55.3408179, 54.8137882, 54.2812061, 53.7434166, 53.2007643, 52.653594, 52.1022504, 51.5470782, 50.9884223, 50.4266273, 49.862038, 49.2949991, 48.7258554, 48.1549516, 47.5826325, 47.0092427, 46.4351271, 45.8606303, 45.2860972, 44.7118724, 44.1383007, 43.5657268, 42.9944956, 42.4249516, 41.8574397, 41.2923046, 40.7298911, 40.1705438, 39.6146076, 39.0624271, 38.5143471, 37.9707124, 37.4318677, 36.8981577, 36.3960015, 35.9121581, 35.4330735, 34.9582461, 34.4871746, 34.0193575, 33.5542934, 33.0914809, 32.6304186, 32.170605, 31.7115388, 31.2527185, 30.7936427, 30.33381, 29.872719, 29.4098682, 28.9447562, 28.4519838, 27.9564871, 27.4583045, 26.9574742, 26.4540346, 25.948024, 25.4394807, 24.928443, 24.4149493, 23.8990378, 23.3807469, 22.860115, 22.3371802, 21.811981, 21.2845556, 20.7549424, 20.2289104, 19.7120912, 19.1927038, 18.670415, 18.1448916, 17.6158005, 17.0828083, 16.5455821, 16.0037885, 15.4570944, 14.9051666, 14.3476719, 13.7842771, 13.2146491, 12.6384546, 12.0553606, 11.4650337, 10.851389, 10.2223006, 9.58583183, 8.94216061, 8.29146496, 7.63392288, 6.96971234, 6.29901135, 5.62199788, 4.93884994, 4.2497455, 3.55486257, 2.85437912, 2.14847315, 1.43732264, 0.721105598, 0, 0.0629524099, -0.0176251433, -0.233719857, -0.577318929, -1.04040956, -1.61497894, -2.29301427, -3.06650275, -3.92743158, -4.86778795, -5.87955906, -6.95473211, -8.08529429, -9.26323281, -10.4805349, -11.7291876, -13.122611, -14.6701398, -16.0890918, -17.3974758, -18.6133001, -19.7545732, -20.8393036, -21.8854998, -22.9111703, -23.9343236, -24.9729682, -26.0451125, -27.168765, -28.6556772, -30.1224088, -31.4988456, -32.714873, -33.6979339, -34.3791354, -34.6902358, -34.5618534, -33.9246067, -32.7091141, -30.8459939, -30.4784992, -28.6512033, -24.8618178, -18.6080545, -9.38762498, 0.94016974, 15.8296047, 36.3733654, 63.6641377, 98.7946071, 141.119911, 195.593954, 260.384232, 337.288068, 435.513681, 560.941948, 719.453746, 911.216773, 1149.12354, 1452.2991, 1820.18418, 2292.04805, 2863.23671, 3563.18768, 4479.10462, 5671.67559, 7113.75894, 8843.27206, 11220.1746, 14422.2609, 16590.9779, 20253.7515, 26293.2655, 35592.2038, 49033.2502, 84752.0922, 122065.159, 157541.883, 187751.698, 209264.036, 205854.47, 195415.533, 179999.741, 166933.42, 156181.534, 145511.631, 134998.547, 124717.119, 114742.184, 105148.58, 97190.3112, 89087.4696, 81000.8859, 73091.3908, 65519.8152, 58446.9898, 52033.7455, 46151.4181, 40695.4726},
        {53.6791712, 53.1114875, 52.5427765, 51.9732288, 51.403035, 50.8323857, 50.2614715, 49.6904829, 49.1196105, 48.549045, 47.9789768, 47.4095967, 46.8410951, 46.2736626, 45.7074899, 45.1427675, 44.579686, 44.018436, 43.4592081, 42.9021929, 42.3475809, 41.7955627, 41.2463289, 40.7000701, 40.1569769, 39.6172399, 39.0810496, 38.5485967, 38.0200716, 37.4956651, 36.9755677, 36.4599699, 35.9490624, 35.4430357, 34.9623482, 34.4966302, 34.0352686, 33.5777962, 33.1237459, 32.6726506, 32.2240431, 31.7774563, 31.332423, 30.8884761, 30.4451485, 30.001973, 29.5584824, 29.1142097, 28.6686876, 28.2214491, 27.772027, 27.2956882, 26.8167575, 26.3352938, 25.8513561, 25.3650033, 24.8762942, 24.3852878, 23.8920429, 23.3966184, 22.8990734, 22.3994665, 21.8978568, 21.3943032, 20.8888644, 20.3815996, 19.8725674, 19.3682727, 18.8750651, 18.3797098, 17.8818479, 17.3811204, 16.8771683, 16.3696327, 15.8581547, 15.3423752, 14.8219354, 14.2964763, 13.7656389, 13.2290643, 12.6863936, 12.1372677, 11.5813278, 11.0182149, 10.4360031, 9.84036016, 9.237093, 8.626218, 8.00775159, 7.38171017, 6.74811017, 6.10696801, 5.45830009, 4.80212285, 4.13845269, 3.46730604, 2.78869932, 2.10264893, 1.40917131, 0.708282856, 0, 0.0195150034, -0.0982044682, -0.345774133, -0.715809709, -1.20092692, -1.79374147, -2.48686909, -3.27292549, -4.1445264, -5.09428753, -6.1148246, -7.19875332, -8.33868942, -9.52724861, -10.7570466, -12.0206992, -13.4260653, -14.9822241, -16.4211217, -17.760069, -19.0163766, -20.2073553, -21.3503158, -22.4625687, -23.5614249, -24.664195, -25.7881897, -26.9507199, -28.1690961, -29.7454363, -31.3078148, -32.7881001, -34.1181609, -35.2662965, -36.1461602, -36.6616902, -36.7338259, -36.2835067, -35.231672, -33.4992611, -33.3300473, -31.6151866, -27.8211507, -21.4144116, -11.8614411, -0.94878557, 14.7571764, 36.3699916, 65.003207, 101.770369, 145.693437, 202.533932, 270.664337, 352.133098, 456.493741, 589.930162, 758.626259, 961.839471, 1214.25435, 1538.11158, 1936.02805, 2445.57912, 3055.99377, 3797.10814, 4767.56499, 6034.42373, 7574.86815, 9432.65656, 11981.0912, 15407.0655, 17925.4139, 21998.4829, 28495.8017, 38286.8998, 52241.3065, 88265.3622, 125932.582, 161853.294, 192637.824, 214896.498, 212728.806, 203596.47, 189471.605, 177347.777, 167209.441, 156948.782, 146654.867, 136416.765, 126323.545, 116464.277, 108555.196, 100037.555, 91207.4861, 82361.1247, 73794.6051, 65804.0615, 58685.6283, 52224.4262, 45904.7095},
        {51.9804494, 51.3759305, 50.7748922, 50.1773631, 49.5833715, 48.9929461, 48.4061154, 47.8229079, 47.2433521, 46.6674765, 46.0953098, 45.5268803, 44.9622167, 44.4013475, 43.8443011, 43.2911062, 42.7417912, 42.1963846, 41.6549151, 41.1174111, 40.5839011, 40.0544137, 39.5289774, 39.0076208, 38.4903723, 37.9772605, 37.4683138, 36.963561, 36.4630304, 35.9667505, 35.47475, 34.9870573, 34.503701, 34.0247096, 33.5635987, 33.1133696, 32.6668615, 32.2236652, 31.7833716, 31.3455716, 30.909856, 30.4758156, 30.0430412, 29.6111238, 29.1796542, 28.7482232, 28.3164217, 27.8838404, 27.4500703, 27.0147022, 26.577327, 26.1148474, 25.6500341, 25.1829699, 24.7137374, 24.2424193, 23.7690983, 23.2938571, 22.8167783, 22.3379446, 21.8574387, 21.3753433, 20.8917411, 20.4067147, 19.9203468, 19.4327201, 18.9439172, 18.4614018, 17.9924602, 17.5219525, 17.0494829, 16.5746557, 16.0970751, 15.6163452, 15.1320703, 14.6438546, 14.1513023, 13.6540176, 13.1516047, 12.6436677, 12.1298111, 11.6096388, 11.0827551, 10.5487643, 9.99983601, 9.43944807, 8.87116769, 8.29484035, 7.71031151, 7.11742664, 6.51603121, 5.9059707, 5.28709058, 4.65923631, 4.02225337, 3.37598723, 2.72028336, 2.05498723, 1.37994432, 0.695000081, 0, -0.0607756706, -0.245525714, -0.54799396, -0.961924238, -1.48106038, -2.09914621, -2.80992556, -3.60714225, -4.48454013, -5.43586302, -6.45485474, -7.53525914, -8.67082003, -9.85528124, -11.0823866, -12.34588, -13.7387027, -15.2687698, -16.7054495, -18.0637423, -19.3586487, -20.6051691, -21.818304, -23.0130538, -24.204419, -25.4074, -26.6369974, -27.9082114, -29.2360427, -30.9009701, -32.5551735, -34.1340098, -35.5728357, -36.9007223, -38.0061693, -38.7526863, -39.0475161, -38.7979012, -37.9110844, -36.2943084, -36.2196834, -34.5098423, -30.6093364, -23.9627169, -14.0145351, -2.51290243, 13.9770801, 36.5867514, 66.4474503, 104.690516, 149.940636, 208.899484, 280.24874, 366.332981, 476.975196, 618.640717, 797.794881, 1012.78653, 1280.04215, 1624.84256, 2052.82708, 2600.05904, 3250.59336, 4034.661, 5060.73571, 6401.37652, 8037.64346, 10018.3736, 12727.9003, 16360.7813, 19243.995, 23726.9983, 30650.8184, 40856.4829, 55185.0192, 91078.3096, 128626.313, 164519.843, 195449.713, 218106.737, 217077.332, 209226.307, 196432.071, 185321.274, 175892.818, 166167.171, 156242.159, 146215.604, 136185.33, 126249.161, 118434.321, 109646.403, 100281.358, 90735.1404, 81403.7021, 72682.9959, 64968.9746, 57960.9685, 50876.1859},
        {50.1717908, 49.5496242, 48.9342458, 48.3255537, 47.7234457, 47.1278199, 46.5385742, 45.9556065, 45.378815, 44.8080975, 44.243352, 43.6844765, 43.131369, 42.5839274, 42.0420498, 41.505634, 40.9745781, 40.44878, 39.9281378, 39.4125493, 38.9019126, 38.3961256, 37.8950864, 37.3986928, 36.9068429, 36.4194346, 35.9363659, 35.4575348, 34.9828393, 34.5121773, 34.0454468, 33.5825458, 33.1233723, 32.6678241, 32.2226207, 31.7841058, 31.3485413, 30.9156038, 30.4849699, 30.0563163, 29.6293196, 29.2036563, 28.7790031, 28.3550366, 27.9314334, 27.5078701, 27.0840233, 26.6595697, 26.2341858, 25.8075483, 25.3793337, 24.9303997, 24.4796499, 24.0271688, 23.573041, 23.117351, 22.6601834, 22.2016229, 21.7417539, 21.280661, 20.8184288, 20.3551419, 19.8908849, 19.4257422, 18.9597985, 18.4931384, 18.0258464, 17.5658898, 17.1210466, 16.675144, 16.2277554, 15.7784544, 15.3268146, 14.8724095, 14.4148126, 13.9535977, 13.4883382, 13.0186076, 12.5439796, 12.0640278, 11.5783256, 11.0864466, 10.5879644, 10.0824526, 9.56606871, 9.04016615, 8.50613941, 7.9636729, 7.41245104, 6.85215825, 6.28247895, 5.70309756, 5.11369849, 4.51396617, 3.90358501, 3.28223944, 2.64961387, 2.00539272, 1.34926041, 0.680901367, 0, -0.14581075, -0.400498719, -0.759111856, -1.21669811, -1.76830544, -2.40898178, -3.1337751, -3.93773333, -4.81590443, -5.76333636, -6.77507705, -7.84617447, -8.97167655, -10.1466313, -11.3660865, -12.6250903, -13.9975933, -15.4899151, -16.9186334, -18.295704, -19.6330824, -20.9427244, -22.2365855, -23.5266215, -24.824788, -26.1430407, -27.4933353, -28.8876273, -30.3378725, -32.0993732, -33.8515113, -35.5332384, -37.0835062, -38.6025248, -39.9486167, -40.937102, -41.4585549, -41.4035494, -40.6626596, -39.1264597, -39.0202347, -37.1892836, -33.0673892, -26.0883345, -15.6859023, -3.66587979, 13.5029556, 36.9798222, 67.9239381, 107.494522, 153.913008, 214.867022, 289.427617, 380.262901, 497.384971, 647.508623, 837.348652, 1064.42788, 1346.74216, 1712.31496, 2169.08683, 2752.66126, 3444.25202, 4274.80511, 5359.53501, 6774.97557, 8504.081, 10599.1238, 13455.5018, 17274.5217, 20523.5637, 25404.619, 32719.1999, 43268.8187, 57854.9873, 93295.7059, 130371.877, 165880.89, 196620.136, 219387.006, 219355.938, 212695.907, 201185.568, 191066.496, 182350.772, 173190.928, 163692.026, 153959.124, 144097.282, 134211.559, 126393.123, 117447.672, 107800.467, 97876.768, 88101.8361, 78900.9318, 70699.3157, 63135.5169, 55431.7303},
        {48.1803337, 47.5750754, 46.9775297, 46.3875349, 45.8049295, 45.2295517, 44.6612398, 44.0998323, 43.5451673, 42.9970833, 42.4554187, 41.9200116, 41.3907005, 40.8673237, 40.3497195, 39.8377263, 39.3311824, 38.8299261, 38.3337958, 37.8426298, 37.3562664, 36.874544, 36.3973009, 35.9243754, 35.4556058, 34.9908306, 34.529888, 34.0726164, 33.6188541, 33.1684394, 32.7212107, 32.2770063, 31.8356646, 31.3970238, 30.9622816, 30.530568, 30.100997, 29.6733626, 29.2474591, 28.8230807, 28.4000217, 27.9780762, 27.5570385, 27.1367029, 26.7168635, 26.2973145, 25.8778503, 25.4582649, 25.0383527, 24.6179079, 24.1967246, 23.7632834, 23.3289376, 22.8937265, 22.4576896, 22.0208665, 21.5832966, 21.1450194, 20.7060743, 20.2665009, 19.8263386, 19.3856269, 18.9444053, 18.5027133, 18.0605904, 17.618076, 17.1752096, 16.7393284, 16.3175947, 15.8949966, 15.4711006, 15.0454731, 14.6176804, 14.187289, 13.7538652, 13.3169754, 12.8761861, 12.4310637, 11.9811744, 11.5260848, 11.0653612, 10.59857, 10.1252777, 9.64505047, 9.15788205, 8.66311622, 8.16009167, 7.64836093, 7.12747653, 6.596991, 6.05645688, 5.5054267, 4.94345298, 4.37008827, 3.78488509, 3.18739598, 2.57717346, 1.95377007, 1.31673835, 0.665630811, 0, -0.203481374, -0.504033321, -0.897860341, -1.38116693, -1.9501576, -2.60103683, -3.33000914, -4.13327901, -5.00705096, -5.94752947, -6.95091906, -8.01342421, -9.13124943, -10.3005992, -11.5176781, -12.7786905, -14.1398068, -15.6057982, -17.0372319, -18.4431624, -19.8326439, -21.2147308, -22.5984775, -23.9929381, -25.4071672, -26.8502189, -28.3311477, -29.8590079, -31.4428537, -33.3177403, -35.1838543, -36.9824498, -38.6547809, -40.3730177, -41.9629562, -43.1888151, -43.9225734, -44.0362106, -43.4017059, -41.8910386, -41.6045284, -39.5076235, -35.0363238, -27.6266287, -16.714538, -4.32141661, 13.3484425, 37.5053809, 69.3597405, 110.121863, 157.662052, 220.612956, 298.491144, 394.298042, 518.149994, 676.968887, 877.676613, 1117.13344, 1414.60961, 1800.35168, 2283.31286, 2900.55924, 3634.18632, 4516.49934, 5664.88113, 7157.66249, 8976.17686, 11173.6076, 14158.7951, 18139.4, 21740.9622, 26996.6664, 34661.8305, 45491.7723, 60241.8098, 95022.3222, 131394.801, 166275.798, 196581.866, 219229.558, 220020.512, 214396.137, 204036.522, 194796.022, 186702.411, 178044.182, 168936.07, 159492.811, 149829.143, 140059.8, 131997.04, 122975.022, 113342.777, 103449.338, 93643.7371, 84275.008, 75692.183, 67522.5431, 59393.1713},
        {45.5932658, 45.1046578, 44.6176098, 44.1321195, 43.6481847, 43.165803, 42.6849723, 42.2056903, 41.7279546, 41.2517631, 40.7771135, 40.3040035, 39.8324308, 39.3623932, 38.8938885, 38.4269143, 37.9614684, 37.4975486, 37.0351526, 36.574278, 36.1149227, 35.6570844, 35.2007608, 34.7459497, 34.2926488, 33.8408558, 33.3905685, 32.9417846, 32.4945018, 32.0487179, 31.6044307, 31.1616378, 30.720337, 30.2805261, 29.842279, 29.4055537, 28.9703073, 28.5365352, 28.1042324, 27.6733944, 27.2440164, 26.8160936, 26.3896212, 25.9645946, 25.5410091, 25.1188598, 24.6981421, 24.2788511, 23.8609822, 23.4445307, 23.0294917, 22.6233744, 22.2184972, 21.8146927, 21.4117932, 21.009631, 20.6080386, 20.2068482, 19.8058924, 19.4050035, 19.0040138, 18.6027557, 18.2010617, 17.798764, 17.3956951, 16.9916874, 16.5865731, 16.1833753, 15.7850399, 15.3848195, 14.9823395, 14.5772256, 14.1691032, 13.7575979, 13.3423352, 12.9229408, 12.4990402, 12.0702589, 11.6362224, 11.1965565, 10.7508865, 10.298838, 9.8400367, 9.37410803, 8.90231466, 8.42342916, 7.93620445, 7.44021294, 6.93502707, 6.42021926, 5.89536195, 5.36002755, 4.8137885, 4.25621723, 3.68688616, 3.10536772, 2.51123434, 1.90405846, 1.28341248, 0.648868855, 0, -0.11134339, -0.32879053, -0.648592219, -1.06699926, -1.58026245, -2.1846326, -2.87636049, -3.65169695, -4.50689275, -5.43819872, -6.44186564, -7.51414431, -8.65128555, -9.84954014, -11.1051589, -12.4143926, -13.8301927, -15.3569101, -16.8625442, -18.3554958, -19.844166, -21.3369556, -22.8422656, -24.368497, -25.9240506, -27.5173274, -29.1567284, -30.8506545, -32.6075066, -34.6739371, -36.7327367, -38.7208307, -40.5751448, -42.5046277, -44.3101933, -45.7202159, -46.590014, -46.774906, -46.1302106, -44.5112461, -43.8002224, -41.2086867, -36.195392, -28.2190911, -16.7385368, -4.10379721, 13.8405189, 38.3584875, 70.7141846, 112.171686, 160.040825, 224.372884, 306.101666, 408.797806, 541.415464, 710.49171, 922.56361, 1172.42953, 1481.03313, 1882.12392, 2383.66581, 3029.38532, 3814.50053, 4774.73437, 6014.44297, 7606.47868, 9513.26234, 11779.0159, 14852.877, 18954.9401, 22906.0169, 28540.7588, 36563.6274, 47679.0842, 62591.5907, 96734.0357, 132400.347, 166612.879, 196393.986, 218766.023, 220078.911, 215142.395, 205552.73, 196768.587, 188857.348, 180277.056, 171175.559, 161700.706, 152000.346, 142222.328, 132524.454, 123318.109, 114472.676, 105857.54, 97342.0864, 88795.6988, 80087.762, 70516.6927, 62641.122},
        {43.3985232, 42.9859669, 42.5715309, 42.1552909, 41.7373228, 41.3177025, 40.8965056, 40.473808, 40.0496856, 39.6242141, 39.1974693, 38.769527, 38.3404631, 37.9103534, 37.4792736, 37.0472995, 36.6145071, 36.180972, 35.7467701, 35.3119771, 34.876669, 34.4409215, 34.0048104, 33.5684115, 33.1318006, 32.6950536, 32.2582462, 31.8214542, 31.3847535, 30.9482198, 30.511929, 30.0759569, 29.6403793, 29.2052719, 28.7688791, 28.332231, 27.8963797, 27.4614603, 27.0276082, 26.5949586, 26.1636468, 25.7338079, 25.3055772, 24.8790899, 24.4544814, 24.0318868, 23.6114414, 23.1932804, 22.7775391, 22.3643527, 21.9538565, 21.5653749, 21.179438, 20.795765, 20.4140752, 20.0340878, 19.655522, 19.2780972, 18.9015326, 18.5255474, 18.1498609, 17.7741924, 17.398261, 17.0217862, 16.644487, 16.2660829, 15.886293, 15.5053176, 15.1233457, 14.7391042, 14.3522812, 13.9625648, 13.5696429, 13.1732038, 12.7729354, 12.3685258, 11.9596631, 11.5460354, 11.1273307, 10.7032371, 10.2734426, 9.83763537, 9.39550342, 8.94673484, 8.49862532, 8.04689911, 7.58718899, 7.11893614, 6.64158176, 6.15456701, 5.65733309, 5.14932118, 4.62997247, 4.09872815, 3.55502938, 2.99831737, 2.42803329, 1.84361833, 1.24451367, 0.630160498, 0, -0.119460242, -0.341197284, -0.662074466, -1.07895513, -1.58870261, -2.18818025, -2.87425139, -3.64377938, -4.49362754, -5.42065923, -6.42173777, -7.49372652, -8.63348881, -9.83788798, -11.1037874, -12.4280503, -13.8549115, -15.3880014, -16.9170343, -18.4490243, -19.9909859, -21.5499333, -23.1328809, -24.746843, -26.3988339, -28.095868, -29.8449596, -31.653123, -33.5273726, -35.7084669, -37.8839697, -39.9907722, -41.9657654, -44.0924846, -46.1344987, -47.7629383, -48.810701, -49.1106848, -48.4957873, -46.7989064, -45.6270288, -42.4990233, -36.9006833, -28.3178021, -16.2361733, -3.83612346, 14.0147205, 38.7541983, 71.82015, 114.650415, 164.834286, 232.361931, 317.611395, 424.183385, 562.258202, 739.062044, 961.821111, 1227.05713, 1555.13359, 1978.09163, 2500.82602, 3172.25991, 3992.12054, 4998.52085, 6294.61623, 7953.70651, 9938.62631, 12294.0363, 15469.6364, 19686.0133, 23893.1681, 29777.748, 38000.158, 49220.8029, 64100.0876, 97378.7798, 132116.831, 165454.556, 194532.268, 216490.282, 218262.208, 213999.534, 205216.875, 197053.997, 189590.562, 181395.75, 172624.748, 163432.74, 153974.912, 144406.449, 134677.204, 125634.975, 117050.331, 108693.844, 100336.082, 91747.6168, 82699.0173, 73081.5676, 64820.9729},
        {41.5158516, 41.1409367, 40.7631582, 40.3826021, 39.9993541, 39.6135003, 39.2251264, 38.8343184, 38.4411622, 38.0457437, 37.6481486, 37.248463, 36.8467728, 36.4431637, 36.0377217, 35.6305327, 35.2216826, 34.8112573, 34.3993426, 33.9860244, 33.5713887, 33.1555213, 32.7385081, 32.320435, 31.9013878, 31.4814526, 31.0607151, 30.6392612, 30.2171769, 29.794548, 29.3714605, 28.9480001, 28.5242528, 28.1003045, 27.67264, 27.2432213, 26.8140541, 26.3853411, 25.9572849, 25.5300885, 25.1039545, 24.6790857, 24.2556848, 23.8339546, 23.4140978, 22.9963173, 22.5808156, 22.1677957, 21.7574603, 21.350012, 20.9456537, 20.5679729, 20.1932807, 19.8212728, 19.4516451, 19.0840933, 18.7183134, 18.354001, 17.990852, 17.6285623, 17.2668276, 16.9053438, 16.5438067, 16.1819121, 15.8193558, 15.4558336, 15.0910413, 14.7238769, 14.3532575, 13.9805244, 13.6054253, 13.2277075, 12.8471187, 12.4634065, 12.0763182, 11.6856016, 11.291004, 10.8922732, 10.4891565, 10.0814016, 9.6687559, 9.25096706, 8.82778256, 8.39894996, 7.98117801, 7.56537689, 7.14225289, 6.71100315, 6.27082476, 5.82091485, 5.36047054, 4.88868892, 4.40476713, 3.90790227, 3.39729146, 2.87213181, 2.33162044, 1.77495446, 1.20133098, 0.609947122, 0, -0.209645573, -0.506604521, -0.888904102, -1.35457157, -1.90163419, -2.52811922, -3.2320539, -4.01146551, -4.8643813, -5.78882851, -6.78283443, -7.84442629, -8.97163136, -10.1624769, -11.4149901, -12.7271984, -14.1251204, -15.6107823, -17.118897, -18.6534901, -20.2185869, -21.8182129, -23.4563936, -25.1371544, -26.8645207, -28.6425181, -30.4751719, -32.3665076, -34.3205507, -36.550321, -38.7742183, -40.9335696, -42.9697021, -45.2676217, -47.5468161, -49.4084586, -50.6607726, -51.1119816, -50.570309, -48.8439783, -47.2868128, -43.6910362, -37.5624718, -28.406943, -15.7302735, -3.77600654, 13.8721843, 38.8929056, 72.964764, 117.766366, 171.721872, 243.741991, 332.300808, 440.649432, 581.792065, 764.353967, 996.9604, 1280.2749, 1632.69393, 2081.29951, 2627.18403, 3322.41196, 4164.06665, 5191.51807, 6515.78849, 8215.0106, 10266.861, 12723.4283, 16005.1399, 20324.0699, 24708.9773, 30737.2999, 39030.721, 50210.9236, 64899.5911, 97134.8852, 130767.891, 163066.172, 191297.292, 212728.813, 214903.197, 211296.512, 203344.226, 195934.44, 189135.951, 181578.956, 173404.184, 164752.36, 155764.21, 146580.462, 137957.779, 129295.161, 120551.155, 111684.313, 102653.181, 93416.3082, 83932.2439, 75200.3231, 66144.7354},
        {39.7446927, 39.3881461, 39.0287564, 38.6666018, 38.3017611, 37.9343125, 37.5643345, 37.1919056, 36.8171042, 36.4400087, 36.0606977, 35.6792496, 35.2957428, 34.9102557, 34.5228669, 34.1336547, 33.7426977, 33.3500742, 32.9558627, 32.5601417, 32.1629896, 31.7644849, 31.364706, 30.9637313, 30.5616393, 30.1585086, 29.7544174, 29.3494442, 28.9436676, 28.537166, 28.1300177, 27.7223013, 27.3140952, 26.9054779, 26.492708, 26.0778542, 25.6630311, 25.2484411, 24.8342867, 24.4207702, 24.0080939, 23.5964605, 23.1860722, 22.7771314, 22.3698406, 21.9644021, 21.5610185, 21.159892, 20.7612251, 20.3652202, 19.9720797, 19.6039358, 19.2385857, 18.8757565, 18.5151751, 18.1565686, 17.799664, 17.4441882, 17.0898684, 16.7364314, 16.3836044, 16.0311144, 15.6786882, 15.3260531, 14.9729359, 14.6190637, 14.2641636, 13.9067521, 13.5453752, 13.1822559, 12.8171998, 12.4500122, 12.0804987, 11.7084648, 11.3337158, 10.9560573, 10.5752948, 10.1912337, 9.80367945, 9.41243756, 9.0173135, 8.61811274, 8.21464074, 7.80670298, 7.42151666, 7.04460482, 6.66116102, 6.27010117, 5.87034117, 5.46079693, 5.04038437, 4.60801938, 4.16261787, 3.70309575, 3.22836893, 2.73735331, 2.2289648, 1.70211931, 1.15573274, 0.588721001, 0, -0.332502973, -0.732163573, -1.19849873, -1.73102536, -2.3292604, -2.99272077, -3.7209234, -4.51338522, -5.36962315, -6.28915412, -7.27149506, -8.31616289, -9.42267454, -10.5905469, -11.819297, -13.1084417, -14.4546871, -15.8571486, -17.321701, -18.8472587, -20.4327364, -22.0770488, -23.7791103, -25.5378358, -27.3521397, -29.2209367, -31.1431415, -33.1176685, -35.1434325, -37.3854136, -39.6155701, -41.7829971, -43.8367896, -46.2699846, -48.7597059, -50.8380265, -52.2908588, -52.904115, -52.4637077, -50.7555491, -48.9812101, -45.0800921, -38.5611308, -28.9332622, -15.7054219, -4.09535649, 13.5240118, 39.0709498, 74.4637247, 121.620603, 179.942385, 256.941709, 348.827798, 458.376413, 601.769997, 789.171669, 1030.74455, 1331.69651, 1708.27211, 2182.12157, 2751.18827, 3468.80988, 4326.25168, 5363.26651, 6701.76679, 8425.95234, 10533.3907, 13085.8216, 16462.636, 20863.5436, 25373.0629, 31477.6622, 39763.0784, 50815.0489, 65219.3106, 96315.1386, 128746.348, 159910.29, 187204.319, 208025.787, 210523.558, 207509.132, 200343.499, 193711.708, 187658.976, 180845.496, 173385.737, 165394.169, 156985.259, 148273.476, 141058.68, 132798.01, 123735.178, 114113.897, 104177.881, 94170.8419, 84336.4935, 76732.0557, 66863.7904},
        {37.6880581, 37.3694168, 37.0453899, 36.7161714, 36.381955, 36.0429348, 35.6993046, 35.3512585, 34.9989904, 34.6426942, 34.2825638, 33.9187932, 33.5515764, 33.1811072, 32.8075795, 32.4311875, 32.0521249, 31.6705857, 31.2867638, 30.9008533, 30.5130479, 30.1235417, 29.7325286, 29.3402026, 28.9467575, 28.5523874, 28.157286, 27.7616475, 27.3656657, 26.9695346, 26.573448, 26.1776, 25.7821845, 25.3873953, 24.9963985, 24.6078393, 24.2203276, 23.8339609, 23.4488367, 23.0650525, 22.6827057, 22.3018939, 21.9227145, 21.545265, 21.169643, 20.795946, 20.4242713, 20.0547166, 19.6873793, 19.3223569, 18.9597469, 18.6173823, 18.2772406, 17.9390349, 17.6024781, 17.2672833, 16.9331636, 16.5998318, 16.2670011, 15.9343845, 15.6016951, 15.2686457, 14.9349495, 14.6003195, 14.2644687, 13.9271101, 13.5879567, 13.2431892, 12.889073, 12.5326064, 12.1737314, 11.8123902, 11.4485246, 11.0820767, 10.7129886, 10.3412021, 9.96665946, 9.58930255, 9.20907341, 8.82591407, 8.43976654, 8.05057284, 7.65827498, 7.26281497, 6.90145493, 6.55469185, 6.20257416, 5.84383276, 5.47719856, 5.10140247, 4.7151754, 4.31724825, 3.90635193, 3.48121735, 3.0405754, 2.58315701, 2.10769307, 1.61291449, 1.09755218, 0.56033705, 0, -0.415323906, -0.876065268, -1.38387542, -1.94040568, -2.54730739, -3.20623187, -3.91883046, -4.68675448, -5.51165527, -6.39518415, -7.33899244, -8.34473149, -9.41405263, -10.5486072, -11.7500464, -13.0200218, -14.2833091, -15.5305775, -16.9373632, -18.4888443, -20.1701988, -21.9666046, -23.8632399, -25.8452826, -27.8979108, -30.0063025, -32.1556356, -34.3310883, -36.5178386, -38.7687438, -40.9764868, -43.1059418, -45.121983, -47.6227225, -50.2564156, -52.5024541, -54.1357406, -54.9311781, -54.6636693, -53.1081172, -51.6518275, -47.9665686, -41.5117732, -31.7468738, -18.131303, -6.03943564, 12.4629734, 39.3952268, 76.7766277, 126.626479, 188.706609, 270.052931, 366.388691, 480.371395, 628.815981, 822.180991, 1070.92496, 1379.7968, 1766.26233, 2254.01606, 2844.32903, 3586.92317, 4469.10433, 5529.45098, 6893.42005, 8646.58983, 10793.6406, 13399.177, 16824.43, 21262.8418, 25880.9571, 32057.1001, 40340.1781, 51279.0979, 65422.7665, 95474.2625, 126789.779, 156879.679, 183254.327, 203424.087, 206039.201, 203376.904, 196744.756, 190723.442, 185324.909, 179197.303, 172420.761, 165075.42, 157241.417, 148998.891, 141203.148, 132682.832, 123606.82, 114143.99, 104463.217, 94733.3796, 85123.3534, 77210.8911, 67802.34},
        {35.468434, 35.1940186, 34.9111397, 34.620132, 34.3213305, 34.0150699, 33.7016851, 33.3815109, 33.0548822, 32.7221338, 32.3836006, 32.0396173, 31.6905188, 31.3366399, 30.9783155, 30.6158805, 30.2496696, 29.8800176, 29.5072595, 29.1317301, 28.7537641, 28.3736965, 27.991862, 27.6085955, 27.2242319, 26.8391059, 26.4535525, 26.0679064, 25.6825024, 25.2976755, 24.9137604, 24.5310921, 24.1500052, 23.7708347, 23.4054477, 23.0481702, 22.6931899, 22.3404673, 21.989963, 21.6416376, 21.2954517, 20.9513659, 20.6093407, 20.2693369, 19.9313148, 19.5952353, 19.2610588, 18.9287459, 18.5982572, 18.2695534, 17.942595, 17.6281946, 17.3152256, 17.0034132, 16.6924829, 16.3821599, 16.0721697, 15.7622375, 15.4520887, 15.1414486, 14.8300425, 14.5175959, 14.2038341, 13.8884823, 13.571266, 13.2519105, 12.930141, 12.6009938, 12.2596173, 11.9154084, 11.5683962, 11.2186101, 10.8660794, 10.5108336, 10.1529018, 9.79231342, 9.4290978, 9.06328426, 8.69490213, 8.32398073, 7.95054939, 7.57463743, 7.19627419, 6.815489, 6.47564741, 6.15419911, 5.82807293, 5.49589185, 5.15627887, 4.80785695, 4.44924909, 4.07907827, 3.69596748, 3.29853969, 2.88541789, 2.45522506, 2.0065842, 1.53811828, 1.04845028, 0.536203193, 0, -0.447050583, -0.92598032, -1.4400984, -1.99271402, -2.58713635, -3.22667461, -3.91463797, -4.65433562, -5.44907676, -6.30217058, -7.21692627, -8.19665301, -9.24465999, -10.3642564, -11.5587515, -12.8314543, -14.0402056, -15.1675331, -16.5457824, -18.1470908, -19.9435955, -21.907434, -24.0107436, -26.2256616, -28.5243254, -30.8788722, -33.2614395, -35.6441645, -37.9991847, -40.2792361, -42.4829711, -44.5883473, -46.5733224, -49.10003, -51.8043397, -54.1396747, -55.8787397, -56.7942398, -56.6588799, -55.245365, -54.28336, -50.9930134, -44.7641467, -34.9865813, -21.0501388, -8.01494218, 11.817064, 40.4736336, 79.9825206, 132.371479, 197.299433, 282.058202, 382.755893, 502.166272, 656.67216, 856.546189, 1112.06099, 1426.06235, 1817.93767, 2315.17637, 2922.94369, 3687.92019, 4597.55641, 5689.50719, 7086.49579, 8872.19765, 11050.1417, 13683.2247, 17128.2139, 21575.647, 26291.89, 32533.8321, 40811.9164, 51636.586, 65518.2838, 94527.1614, 124724.026, 153729.395, 179163.782, 198647.705, 201323.465, 198941.935, 192757.152, 187212.482, 182299.518, 176680.967, 170414.867, 163559.254, 156172.168, 148311.645, 139499.994, 130672.398, 121813.344, 112907.319, 103938.81, 94892.3025, 85752.2844, 77030.3661, 68269.7157},
        {33.1454633, 32.9153365, 32.6745952, 32.4236638, 32.1629669, 31.8929288, 31.6139739, 31.3265268, 31.0310117, 30.7278533, 30.4174758, 30.1003037, 29.7767614, 29.4472735, 29.1122642, 28.7721581, 28.4273795, 28.078353, 27.7255029, 27.3692536, 27.0100296, 26.6482554, 26.2843553, 25.9187538, 25.5518753, 25.1841443, 24.8159851, 24.4478222, 24.0800801, 23.7131832, 23.3475558, 22.9836225, 22.6218076, 22.2625357, 21.9236989, 21.5966204, 21.2724133, 20.9509351, 20.6320435, 20.315596, 20.0014502, 19.6894637, 19.3794939, 19.0713985, 18.7650351, 18.4602611, 18.1569342, 17.854912, 17.5540519, 17.2542117, 16.9552487, 16.6580752, 16.3614713, 16.0652717, 15.7693111, 15.4734242, 15.1774457, 14.8812104, 14.584553, 14.287308, 13.9893104, 13.6903947, 13.3903956, 13.0891479, 12.7864863, 12.4822455, 12.1762602, 11.8668809, 11.5524939, 11.2359947, 10.9173144, 10.5963837, 10.2731338, 9.94749533, 9.61939939, 9.28877684, 8.9555586, 8.61967559, 8.28105872, 7.93963892, 7.59534711, 7.24811419, 6.89787109, 6.54454873, 6.22911371, 5.93011597, 5.62561225, 5.31420176, 4.99448375, 4.66505743, 4.32452204, 3.97147681, 3.60452098, 3.22225376, 2.8232744, 2.40618212, 1.96957615, 1.51205572, 1.03222007, 0.528668417, 0, -0.404151741, -0.846194749, -1.3294337, -1.85717328, -2.43271817, -3.05937304, -3.74044257, -4.47923144, -5.27904433, -6.14318592, -7.07496089, -8.07767392, -9.15462969, -10.3091329, -11.5444881, -12.8640002, -14.1110089, -15.2659669, -16.6940039, -18.3641464, -20.245421, -22.306854, -24.5174722, -26.8463019, -29.2623698, -31.7347022, -34.2323257, -36.7242669, -39.1795522, -41.5400538, -43.8119092, -45.9722912, -47.9983726, -50.5282687, -53.208681, -55.51006, -57.2112957, -58.0912783, -57.9288978, -56.5030444, -55.8959524, -52.8821498, -46.7898727, -36.9473568, -22.682838, -8.51671135, 12.7129864, 42.9760592, 84.2423111, 138.481546, 204.828823, 291.553379, 395.343318, 519.261437, 679.385534, 885.949938, 1149.18898, 1471.6842, 1873.06199, 2381.2972, 3000.39546, 3780.1764, 4715.94625, 5846.57803, 7284.76618, 9109.31989, 11315.6286, 13960.8742, 17408.1109, 21847.3742, 26664.9648, 32978.3045, 41255.6447, 51965.2365, 65575.3315, 93464.3537, 122462.838, 150311.494, 174751.028, 193522.15, 196342.286, 194324.279, 188647.385, 183442.432, 178720.345, 173262.231, 167140.508, 160427.593, 153195.904, 145517.859, 136594.634, 127990.425, 119592.735, 111289.063, 102966.911, 94513.7804, 85817.1728, 76516.2393, 67500.4422},
        {30.7355012, 30.548256, 30.3502448, 30.1418679, 29.9235257, 29.6956187, 29.4585473, 29.2127117, 28.9585125, 28.6963501, 28.4266248, 28.1497371, 27.8660873, 27.5760759, 27.2801033, 26.9785698, 26.6718758, 26.3604219, 26.0446083, 25.7248355, 25.4015039, 25.0750138, 24.7457657, 24.4141601, 24.0805972, 23.7454774, 23.4092013, 23.0721692, 22.7347815, 22.3974386, 22.0605408, 21.7244887, 21.3896826, 21.0565229, 20.7423827, 20.4388191, 20.1371852, 19.8373304, 19.5391045, 19.242357, 18.9469376, 18.6526958, 18.3594813, 18.0671436, 17.7755325, 17.4844973, 17.1938879, 16.9035537, 16.6133444, 16.3231096, 16.0326989, 15.7298953, 15.4268766, 15.123754, 14.8206386, 14.5176416, 14.2148743, 13.9124479, 13.6104734, 13.3090621, 13.0083252, 12.7083738, 12.4093192, 12.1112725, 11.814345, 11.5186477, 11.2242919, 10.940746, 10.6772533, 10.414627, 10.1523715, 9.88999155, 9.62699164, 9.36287643, 9.09715051, 8.82931848, 8.55888495, 8.28535453, 8.00823181, 7.7270214, 7.44122791, 7.15035593, 6.85391008, 6.55139495, 6.2675791, 5.98880376, 5.70110651, 5.40317211, 5.09368528, 4.77133078, 4.43479333, 4.08275768, 3.71390857, 3.32693075, 2.92050894, 2.4933279, 2.04407235, 1.57142705, 1.07407673, 0.550706134, 0, -0.254543121, -0.584946765, -0.991620286, -1.47497304, -2.03541437, -2.67335365, -3.38920021, -4.18336342, -5.05625263, -6.00827719, -7.03984646, -8.15136979, -9.34325653, -10.615916, -11.9697577, -13.4051908, -14.8507372, -16.2965366, -17.9056011, -19.6629356, -21.553545, -23.5624341, -25.6746079, -27.8750712, -30.1488289, -32.4808859, -34.8562471, -37.2599173, -39.6769014, -42.2054062, -44.6757274, -47.0389092, -49.2459959, -51.770033, -54.3010642, -56.3899329, -57.8310832, -58.4189589, -57.948004, -56.2126626, -55.5357314, -52.4138028, -46.1466429, -36.0340181, -21.3756943, -6.10227122, 16.2750266, 47.6141414, 89.7730151, 144.60959, 210.282186, 296.870128, 401.322071, 527.14285, 691.222829, 904.442542, 1177.68252, 1517.64933, 1940.3461, 2466.32942, 3087.97992, 3870.13154, 4828.19438, 6006.34639, 7497.51097, 9372.23656, 11609.8345, 14257.1992, 17695.761, 22117.7179, 27059.1381, 33469.2921, 41767.4998, 52373.0806, 65705.3544, 92328.4559, 119980.686, 146545.16, 169904.995, 187943.306, 191125.644, 189697.733, 184721.646, 179691.127, 174705.71, 168851.736, 162279.398, 155138.894, 147580.416, 139754.162, 132813.239, 125515.882, 117899.878, 110003.012, 101863.07, 93517.8383, 85005.102, 75947.3023, 64679.5208},
        {28.6653696, 28.4480612, 28.2254225, 27.9976373, 27.7648895, 27.5273629, 27.2852415, 27.038709, 26.7879494, 26.5331464, 26.274484, 26.0121459, 25.7463161, 25.4771784, 25.2049166, 24.9297146, 24.6517563, 24.3712255, 24.0883061, 23.8031819, 23.5160368, 23.2270546, 22.9364193, 22.6443146, 22.3509243, 22.0564325, 21.7610228, 21.4648793, 21.1681856, 20.8711258, 20.5738835, 20.2766428, 19.9795874, 19.6829012, 19.3942405, 19.1098957, 18.8260675, 18.5426972, 18.2597264, 17.9770962, 17.694748, 17.4126233, 17.1306634, 16.8488097, 16.5670034, 16.285186, 16.0032989, 15.7212833, 15.4390807, 15.1566325, 14.8738799, 14.5787142, 14.2833882, 13.9881044, 13.6930654, 13.3984738, 13.1045321, 12.8114431, 12.5194091, 12.2286329, 11.939317, 11.651664, 11.3658764, 11.082157, 10.8007082, 10.5217326, 10.2454328, 9.98497236, 9.7532034, 9.52359775, 9.2955178, 9.06832593, 8.84138452, 8.61405597, 8.38570266, 8.15568698, 7.9233713, 7.68811802, 7.44928951, 7.20624818, 6.95835639, 6.70497654, 6.44547102, 6.1792022, 5.9296524, 5.68361672, 5.42760034, 5.16018291, 4.87994407, 4.58546347, 4.27532076, 3.94809559, 3.6023676, 3.23671645, 2.84972177, 2.43996322, 2.00602044, 1.54647308, 1.05990079, 0.544883213, 0, -0.11500639, -0.336686124, -0.663371911, -1.09339646, -1.62509247, -2.25679266, -2.98682974, -3.8135364, -4.73524537, -5.75028934, -6.85700103, -8.05371314, -9.33875838, -10.7104695, -12.1671791, -13.70722, -15.3032049, -16.9497923, -18.7034909, -20.5571219, -22.5035068, -24.5354668, -26.6458231, -28.8273971, -31.0730101, -33.3754833, -35.7276381, -38.1222958, -40.5522776, -43.1902116, -45.7831832, -48.2700717, -50.5897564, -53.1740412, -55.7153431, -57.7746323, -59.1429109, -59.6111806, -58.9704433, -57.0117008, -56.1262721, -52.7134411, -46.0554521, -35.4345494, -20.132977, -3.64728289, 20.0081683, 52.6047083, 95.9136687, 151.706381, 217.231981, 304.311125, 410.930265, 541.120009, 711.251126, 932.153907, 1214.65864, 1566.95753, 2003.11852, 2540.08744, 3163.90683, 3947.56841, 4913.52849, 6109.19221, 7624.23911, 9528.94408, 11812.0509, 14522.6967, 18008.9662, 22449.8426, 27408.4641, 33774.4039, 41939.8221, 52296.8787, 65237.7339, 90709.3583, 117160.399, 142594.313, 165014.558, 182424.592, 185894.474, 184983.325, 180666.057, 176064.808, 171304.548, 165615.573, 159167.587, 152130.296, 144673.405, 136966.619, 131579.349, 124949.596, 117379.21, 109170.039, 100623.935, 92042.748, 83728.3271, 75322.7373, 65159.2673},
        {26.9443745, 26.6372785, 26.334256, 26.0351502, 25.7398045, 25.448062, 25.1597659, 24.8747597, 24.5928864, 24.3139894, 24.0379118, 23.764497, 23.4935882, 23.2250287, 22.9586616, 22.6943303, 22.431878, 22.1711479, 21.9119833, 21.6542274, 21.3977236, 21.142315, 20.8878448, 20.6341565, 20.3810931, 20.1284979, 19.8762143, 19.6240854, 19.3719544, 19.1196648, 18.8670596, 18.6139821, 18.3602756, 18.1057834, 17.8426265, 17.5746716, 17.3058784, 17.0363406, 16.7661523, 16.4954071, 16.224199, 15.9526218, 15.6807694, 15.4087356, 15.1366142, 14.8644991, 14.5924842, 14.3206633, 14.0491303, 13.777979, 13.5073032, 13.2381469, 12.9696333, 12.7018356, 12.434827, 12.1686809, 11.9034705, 11.639269, 11.3761497, 11.1141858, 10.8534507, 10.5940174, 10.3359594, 10.0793499, 9.82426203, 9.57076916, 9.3189445, 9.0764334, 8.85069939, 8.62619888, 8.40251424, 8.17922788, 7.95592217, 7.7321795, 7.50758226, 7.28171284, 7.05415363, 6.824487, 6.59229536, 6.35716107, 6.11866655, 5.87639416, 5.62992629, 5.37884534, 5.16380119, 4.96297868, 4.75406905, 4.53532195, 4.30498707, 4.06131407, 3.80255262, 3.52695238, 3.23276302, 2.9182342, 2.58161561, 2.2211569, 1.83510775, 1.42171781, 0.979236762, 0.50591427, 0, -0.0454279127, -0.207998847, -0.485805659, -0.876941208, -1.37949835, -1.99156994, -2.71124884, -3.53662791, -4.4658, -5.49685797, -6.62789468, -7.85700299, -9.18227575, -10.6018058, -12.1136861, -13.7160093, -15.3619288, -17.0431174, -18.8593172, -20.7989912, -22.8506022, -25.0026133, -27.2434873, -29.5616872, -31.9456759, -34.3839163, -36.8648714, -39.377004, -41.9087772, -44.5939971, -47.2224908, -49.7391183, -52.0887394, -54.804086, -57.5360819, -59.7788855, -61.3009952, -61.8709093, -61.2571263, -59.2281446, -58.1112712, -54.3374279, -47.1744769, -35.8902801, -19.7526993, -2.0176959, 23.0319924, 57.1255284, 101.992075, 159.360795, 225.807061, 314.512904, 425.156027, 562.291752, 740.433871, 969.62503, 1259.90787, 1617.59225, 2057.30196, 2597.73291, 3226.31218, 4014.34246, 4975.92439, 6158.44699, 7665.70464, 9575.60076, 11910.3358, 14733.3718, 18312.4145, 22799.6379, 27670.8445, 33858.6369, 41749.3174, 51729.1886, 64184.5529, 88662.7404, 114098.769, 138588.685, 160228.533, 177114.359, 180734.598, 180197.981, 176432.385, 172521.472, 168539.79, 163663.644, 158012.493, 151705.798, 144863.018, 137603.613, 132774.549, 125933.95, 117671.351, 108576.291, 99238.3073, 90246.9362, 82191.7155, 74635.5933, 69214.6451},
        {24.9812298, 24.6845089, 24.3922441, 24.1042741, 23.8204375, 23.540573, 23.2645194, 22.9921152, 22.7231991, 22.4576098, 22.195186, 21.9357663, 21.6791894, 21.425294, 21.1739187, 20.9249022, 20.6780831, 20.4333002, 20.1903921, 19.9491974, 19.7095549, 19.4713032, 19.2342809, 18.9983268, 18.7632795, 18.5289776, 18.2952599, 18.061965, 17.8289315, 17.5959982, 17.3630037, 17.1297866, 16.8961857, 16.6620396, 16.4196518, 16.1727869, 15.9253004, 15.6772755, 15.4287954, 15.1799432, 14.9308021, 14.6814555, 14.4319863, 14.182478, 13.9330136, 13.6836763, 13.4345495, 13.1857161, 12.9372596, 12.689263, 12.4418095, 12.198759, 11.9563363, 11.7145425, 11.4733791, 11.2328473, 10.9929486, 10.7536842, 10.5150555, 10.2770638, 10.0397104, 9.80299668, 9.56692395, 9.33149355, 9.09670681, 8.86256505, 8.6290696, 8.39983463, 8.17838755, 7.9572785, 7.7362746, 7.515143, 7.2936508, 7.07156514, 6.84865314, 6.62468193, 6.39941863, 6.17263037, 5.94408428, 5.71354747, 5.48078708, 5.24557023, 5.00766405, 4.76683566, 4.57942241, 4.41571643, 4.24532993, 4.06619423, 3.87624064, 3.67340047, 3.45560503, 3.22078564, 2.96687361, 2.69180026, 2.3934969, 2.06989483, 1.71892539, 1.33851986, 0.926609584, 0.481125858, 0, -0.187874925, -0.46753279, -0.839047349, -1.30249236, -1.85794157, -2.50546875, -3.24514764, -4.07705201, -5.00125561, -6.01783219, -7.12685551, -8.32839932, -9.62253739, -11.0093435, -12.4888913, -14.0612547, -15.6602167, -17.276381, -19.0597644, -20.9962357, -23.0716633, -25.271916, -27.5828623, -29.9903708, -32.4803102, -35.0385492, -37.6509562, -40.3034, -42.9817492, -45.8254762, -48.6105245, -51.2766813, -53.7637343, -56.5865484, -59.3973722, -61.6951005, -63.2469976, -63.8203275, -63.1823546, -61.1003431, -60.0685862, -56.2973537, -49.0203606, -37.471322, -20.8839528, -2.13489992, 24.0957864, 59.485124, 105.710131, 164.447825, 231.725593, 321.775189, 435.133984, 576.7003, 760.276057, 995.66564, 1292.67344, 1656.86525, 2103.24737, 2651.44996, 3290.62715, 4089.36615, 5059.84804, 6248.21605, 7758.1603, 9666.95595, 12000.9231, 14824.1263, 18387.4654, 22836.1557, 27688.6768, 33813.4004, 41572.4888, 51328.104, 63442.4082, 86982.0882, 111428.65, 134968.126, 155786.546, 172069.941, 175712.87, 175387.152, 171975.162, 168446.411, 164857.119, 160393.862, 155156.342, 149244.259, 142757.312, 135795.202, 130797.547, 123857.567, 115571.564, 106535.843, 97346.7064, 88600.4572, 80893.3988, 73618.841, 71341.4338},
        {22.7887268, 22.5931168, 22.3944325, 22.1927979, 21.9883372, 21.7811746, 21.5714342, 21.3592402, 21.1447167, 20.9279878, 20.7091778, 20.4884108, 20.2658109, 20.0415022, 19.815609, 19.5882554, 19.3595655, 19.1296634, 18.8986734, 18.6667196, 18.4339261, 18.2004172, 17.9663168, 17.7317492, 17.4968386, 17.2617091, 17.0264848, 16.7912899, 16.5562485, 16.3214848, 16.087123, 15.8532872, 15.6201015, 15.3876902, 15.162157, 14.9405104, 14.7196873, 14.4996175, 14.2802313, 14.0614587, 13.8432298, 13.6254747, 13.4081236, 13.1911065, 12.9743535, 12.7577947, 12.5413602, 12.3249802, 12.1085847, 11.8921037, 11.6754675, 11.4543044, 11.2329394, 11.011396, 10.7896973, 10.567867, 10.3459281, 10.1239042, 9.90181851, 9.67969444, 9.45755533, 9.23542453, 9.01332538, 8.79128125, 8.56931547, 8.3474514, 8.1257124, 7.90687976, 7.69366858, 7.48041413, 7.26696096, 7.05315366, 6.83883677, 6.62385486, 6.4080525, 6.19127424, 5.97336465, 5.7541683, 5.53352975, 5.31129355, 5.08730428, 4.86140649, 4.63344475, 4.40326363, 4.23738953, 4.10092353, 3.95816524, 3.80679527, 3.64449423, 3.46894273, 3.27782138, 3.06881079, 2.83959156, 2.58784431, 2.31124964, 2.00748817, 1.67424049, 1.30918723, 0.910008988, 0.474386375, 0, -0.464507651, -0.97438512, -1.53288013, -2.14324039, -2.80871363, -3.53254757, -4.31798993, -5.16828842, -6.08669078, -7.07644471, -8.14079794, -9.28299819, -10.5062932, -11.8139306, -13.2091582, -14.6952238, -16.1993809, -17.7140211, -19.4142832, -21.2871306, -23.3195266, -25.4984344, -27.8108174, -30.2436387, -32.7838618, -35.4184499, -38.1343662, -40.9185742, -43.758037, -46.8525777, -49.8982514, -52.8181632, -55.5354186, -58.4324315, -61.2026529, -63.4170517, -64.8609406, -65.3196325, -64.5784402, -62.4226764, -61.7138759, -58.2248881, -51.1391563, -39.6401237, -22.9112338, -3.36848568, 23.8453722, 60.3550357, 107.7852, 167.760562, 235.972568, 327.23091, 441.925928, 585.061167, 771.058791, 1010.1122, 1312.41478, 1684.32941, 2140.75059, 2700.75157, 3354.2383, 4168.01055, 5160.43512, 6376.02438, 7903.59278, 9811.6194, 12102.7791, 14828.4906, 18281.4547, 22617.8894, 27522.0053, 33694.6442, 41455.4114, 51123.912, 63019.7511, 85614.2667, 109037.398, 131570.817, 151496.195, 167095.206, 170709.059, 170520.195, 167356.197, 163937.986, 160343.7, 155862.587, 150612.229, 144710.21, 138274.112, 131421.518, 125648.509, 118793.07, 111184.633, 103152.633, 95026.5031, 87135.676, 79809.5854, 72270.8995, 71021.7755},
        {20.779917, 20.6459642, 20.5050805, 20.3575358, 20.2035996, 20.0435417, 19.8776317, 19.7061392, 19.5293341, 19.3474859, 19.1608644, 18.9697391, 18.7743799, 18.5750563, 18.3720381, 18.1655949, 17.9559964, 17.7435123, 17.5284122, 17.3109659, 17.091443, 16.8701132, 16.6472462, 16.4231116, 16.1979792, 15.9721186, 15.7457995, 15.5192915, 15.2928644, 15.0667878, 14.8413315, 14.616765, 14.393358, 14.1713803, 13.9643884, 13.7657289, 13.5685891, 13.3728075, 13.1782224, 12.9846725, 12.7919963, 12.6000321, 12.4086185, 12.2175941, 12.0267973, 11.8360666, 11.6452405, 11.4541574, 11.262656, 11.0705747, 10.8777519, 10.6735957, 10.4686011, 10.2628329, 10.0563556, 9.84923377, 9.6415321, 9.43331518, 9.2246476, 9.015594, 8.80621896, 8.59658711, 8.38676306, 8.17681141, 7.96679677, 7.75678376, 7.54683699, 7.3400511, 7.13944798, 6.93884303, 6.73810443, 6.53710037, 6.33569903, 6.1337686, 5.93117727, 5.72779321, 5.52348462, 5.31811968, 5.11156658, 4.90369349, 4.69436861, 4.48346013, 4.27083622, 4.05636507, 3.91291624, 3.80232172, 3.68553614, 3.56005863, 3.42338836, 3.27302447, 3.10646611, 2.92121243, 2.71476259, 2.48461573, 2.22827101, 1.94322758, 1.62698458, 1.27704117, 0.890896507, 0.466049733, 0, -0.568080945, -1.1651027, -1.79557431, -2.46400479, -3.17490319, -3.93277854, -4.74213986, -5.6074962, -6.53335659, -7.52423006, -8.58462564, -9.71905238, -10.9320193, -12.2280354, -13.6116098, -15.0872515, -16.5835341, -18.0941189, -19.7952733, -21.6752345, -23.7222397, -25.924526, -28.2703307, -30.7478909, -33.3454438, -36.0512266, -38.8534764, -41.7404305, -44.700326, -47.9764256, -51.2084316, -54.3080736, -57.187081, -60.1579539, -62.9320367, -65.1138013, -66.494746, -66.8663692, -66.0201692, -63.7476444, -63.173273, -59.7411882, -52.5907835, -40.8614525, -23.6925886, -3.29040317, 25.0392324, 62.9236289, 111.990097, 173.865948, 244.433392, 338.08663, 454.436384, 598.290402, 785.067034, 1025.42069, 1330.00576, 1705.90503, 2168.15627, 2735.69362, 3398.99488, 4224.19567, 5235.04673, 6475.27544, 8022.45862, 9938.63564, 12213.7597, 14891.181, 18278.5662, 22534.0696, 27454.2527, 33614.4923, 41298.1094, 50788.425, 62368.7601, 83881.8476, 106161.744, 127601.917, 146595.835, 161536.965, 165312.223, 165492.667, 162831.715, 159744.362, 156326.322, 151980.985, 146838.682, 141029.744, 134684.499, 127933.277, 120721.266, 113627.679, 106611.678, 99632.4231, 92649.0752, 85620.7953, 78506.7441, 70637.8976, 70388.9216},
        {19.071373, 18.9217661, 18.7684331, 18.6115231, 18.451185, 18.2875676, 18.1208199, 17.9510908, 17.7785293, 17.6032843, 17.4255047, 17.2453394, 17.0629373, 16.8784474, 16.6920186, 16.5037998, 16.31394, 16.122588, 15.9298929, 15.7360034, 15.5410686, 15.3452373, 15.1486585, 14.9514812, 14.7538542, 14.5559264, 14.3578468, 14.1597643, 13.9618279, 13.7641864, 13.5669888, 13.370384, 13.1745209, 12.9795484, 12.7941925, 12.6141331, 12.4349472, 12.2565054, 12.0786783, 11.9013366, 11.7243507, 11.5475912, 11.3709288, 11.1942341, 11.0173776, 10.8402298, 10.6626615, 10.4845432, 10.3057454, 10.1261389, 9.945594, 9.75319708, 9.55983687, 9.36561773, 9.17064405, 8.97502019, 8.77885052, 8.58223942, 8.38529126, 8.18811041, 7.99080124, 7.79346813, 7.59621545, 7.39914756, 7.20236884, 7.00598367, 6.81009641, 6.61878121, 6.43601687, 6.25372044, 6.07173894, 5.88991941, 5.70810886, 5.52615434, 5.34390287, 5.16120148, 4.97789719, 4.79383703, 4.60886804, 4.42283723, 4.23559165, 4.04697832, 3.85684426, 3.66503651, 3.54713719, 3.46353281, 3.37369956, 3.27502674, 3.16490361, 3.04071946, 2.89986357, 2.73972521, 2.55769366, 2.35115821, 2.11750812, 1.85413269, 1.55842118, 1.22776288, 0.859547061, 0.451163009, 0, -0.407489752, -0.873032776, -1.39950915, -1.98979896, -2.64678228, -3.37333919, -4.17234978, -5.04669411, -5.99925228, -7.03290436, -8.15053043, -9.35501057, -10.6492249, -12.0360534, -13.5183762, -15.0990734, -16.7174975, -18.3674532, -20.1955139, -22.1909467, -24.3430186, -26.6409966, -29.0741477, -31.6317389, -34.3030374, -37.07731, -39.9438239, -42.891846, -45.9106433, -49.2541241, -52.549879, -55.707783, -58.6377106, -61.6926914, -64.5610228, -66.8128274, -68.2250333, -68.5745688, -67.6383621, -65.1933415, -64.4163758, -60.649688, -53.0050008, -40.5940366, -22.5285177, -1.17282765, 28.4271372, 67.9728197, 119.165662, 183.707108, 258.229158, 355.698681, 474.5686, 619.183524, 805.916922, 1045.74155, 1349.63017, 1724.80582, 2186.84318, 2755.4075, 3421.48478, 4251.73271, 5271.77138, 6525.72296, 8089.06046, 10022.8663, 12326.4751, 15044.9605, 18449.8055, 22684.4231, 27552.8693, 33590.1232, 41057.0676, 50214.5857, 61323.5602, 81604.627, 102618.978, 122887.558, 140931.311, 155271.182, 159434.967, 160265.526, 158423.723, 155997.597, 153083.09, 149181.491, 144418.475, 138919.719, 132810.898, 126217.69, 117217.229, 109534.299, 102768.279, 96518.5453, 90384.4765, 83965.4506, 76860.8453, 68852.7338, 70031.47},
        {17.4690024, 17.2890253, 17.1098199, 16.9313647, 16.7536383, 16.5766191, 16.4002855, 16.2246161, 16.0495894, 15.8751838, 15.7013778, 15.5281499, 15.3554785, 15.1833422, 15.0117195, 14.8405887, 14.6699285, 14.4997172, 14.3299334, 14.1605556, 13.9915621, 13.8229316, 13.6546425, 13.4866732, 13.3190023, 13.1516082, 12.9844694, 12.8175644, 12.6508717, 12.4843698, 12.3180371, 12.1518521, 11.9857933, 11.8198392, 11.6553914, 11.4916869, 11.3279455, 11.1640996, 11.0000815, 10.8358234, 10.6712576, 10.5063165, 10.3409324, 10.1750374, 10.0085641, 9.84144458, 9.67361121, 9.50499628, 9.3355321, 9.16515095, 8.99378514, 8.8116976, 8.6286996, 8.44493306, 8.26053989, 8.07566201, 7.89044132, 7.70501974, 7.51953918, 7.33414155, 7.14896876, 6.96416272, 6.77986535, 6.59621856, 6.41336426, 6.23144435, 6.05060076, 5.87615856, 5.7133184, 5.55153228, 5.39060612, 5.2303458, 5.07055726, 4.91104638, 4.75161909, 4.59208129, 4.43223888, 4.27189777, 4.11086388, 3.94894311, 3.78594136, 3.62166455, 3.45591858, 3.28850936, 3.19636138, 3.13909916, 3.07542035, 3.00262819, 2.91802595, 2.81891689, 2.70260428, 2.56639137, 2.40758142, 2.2234777, 2.01138347, 1.76860199, 1.49243652, 1.18019033, 0.829166669, 0.436668806, 0, -0.177496122, -0.455001616, -0.832886316, -1.31152006, -1.89127268, -2.57251402, -3.3556139, -4.24094217, -5.22886866, -6.3197632, -7.51399564, -8.8119358, -10.2139535, -11.7204187, -13.331701, -15.0481704, -16.8295373, -18.6703628, -20.6629849, -22.7990605, -25.0702468, -27.4682009, -29.9845798, -32.6110406, -35.3392404, -38.1608363, -41.0674854, -44.0508447, -47.1025714, -50.4906581, -53.8291033, -57.0266633, -59.9920946, -63.1404451, -66.1273254, -68.473335, -69.9353431, -70.2702186, -69.2348304, -66.5860475, -65.4383009, -61.1690308, -52.878192, -39.6657389, -20.6316264, 1.4567469, 32.0837314, 73.0662334, 126.221159, 193.365415, 272.030917, 373.556773, 495.357846, 641.301181, 828.427024, 1067.86112, 1370.72922, 1743.99199, 2203.88687, 2771.19498, 3437.89011, 4270.41817, 5294.4253, 6554.21664, 8125.71398, 10070.3265, 12401.2067, 15167.2527, 18594.3044, 22808.1135, 27594.0927, 33470.6925, 40681.4744, 49469.9996, 60079.8296, 79168.134, 98961.4637, 118099.978, 135223.837, 148973.198, 153467.963, 154882.056, 153792.891, 152010.967, 149625.601, 146212.325, 141886.144, 136762.067, 130955.099, 124580.246, 114453.369, 106365.564, 99704.4684, 93857.7164, 88212.9442, 82157.7875, 75079.882, 67296.1326, 69457.9317},
        {15.7523547, 15.584043, 15.4170666, 15.25138, 15.0869376, 14.9236939, 14.7616033, 14.6006204, 14.4406995, 14.2817951, 14.1238617, 13.9668537, 13.8107256, 13.6554319, 13.5009269, 13.3471652, 13.1941012, 13.0416894, 12.8898843, 12.7386402, 12.5879116, 12.437653, 12.2878189, 12.1383637, 11.9892419, 11.8404079, 11.6918162, 11.5434212, 11.3951774, 11.2470393, 11.0989613, 10.9508978, 10.8028034, 10.6546324, 10.5064589, 10.3581749, 10.2096714, 10.0608988, 9.91180771, 9.76234879, 9.61247258, 9.46212966, 9.3112706, 9.159846, 9.00780643, 8.85510246, 8.70168469, 8.54750369, 8.39251003, 8.23665431, 8.07988709, 7.91289857, 7.74510046, 7.5766441, 7.40768081, 7.23836192, 7.06883875, 6.89926265, 6.72978492, 6.56055691, 6.39172993, 6.22345532, 6.05588441, 5.88916851, 5.72345896, 5.55890708, 5.39566421, 5.23956187, 5.0962952, 4.95430055, 4.81336102, 4.67325972, 4.53377975, 4.39470421, 4.25581621, 4.11689885, 3.97773523, 3.83810846, 3.69780163, 3.55659787, 3.41428025, 3.2706319, 3.12543591, 2.97847539, 2.90959008, 2.87685088, 2.83736647, 2.78832197, 2.72690247, 2.6502931, 2.55567897, 2.44024518, 2.30117684, 2.13565908, 1.940877, 1.7140157, 1.45226031, 1.15279594, 0.812807684, 0.429480668, 0, -0.0435852735, -0.215632195, -0.514626721, -0.939054807, -1.48740241, -2.15815548, -2.94979998, -3.86082185, -4.88970707, -6.03494158, -7.29501134, -8.6684023, -10.1536004, -11.7490917, -13.453362, -15.2648973, -17.1924894, -19.2360965, -21.3708938, -23.5975757, -25.9168365, -28.3293705, -30.8358721, -33.4370356, -36.1335553, -38.9261257, -41.815441, -44.8021956, -47.8870838, -51.4216936, -54.9271648, -58.2989234, -61.4323958, -64.723501, -67.8174189, -70.2258644, -71.6941161, -71.9674522, -70.7911512, -67.9104915, -66.2425322, -61.3954464, -52.4939465, -38.6627452, -19.026555, 2.82243146, 33.3438517, 74.5298632, 128.372623, 196.86429, 278.142029, 382.764645, 508.320725, 658.597581, 850.038251, 1093.40361, 1399.45454, 1773.9164, 2233.73001, 2801.32947, 3471.76627, 4306.7289, 5328.75571, 6579.96409, 8135.64709, 10055.8696, 12357.2533, 15088.6002, 18458.0034, 22583.4521, 27267.9641, 32992.9568, 39985.4476, 48472.454, 58680.9935, 76907.316, 95791.8985, 114044.45, 130374.681, 143492.299, 147808.553, 149197.255, 148210.313, 146599.433, 144441.603, 141313.894, 137318.29, 132556.775, 127131.332, 121143.944, 113039.214, 105731.034, 98995.552, 92608.9136, 86347.2659, 79986.7555, 73303.5289, 66843.1823, 68288.5338},
        {14.0891635, 13.9441452, 13.7997855, 13.6560612, 13.5129491, 13.3704262, 13.2284692, 13.0870551, 12.9461607, 12.8057629, 12.6658384, 12.5263643, 12.3873173, 12.2486742, 12.1104121, 11.9725076, 11.8349377, 11.6976793, 11.5607091, 11.424004, 11.287541, 11.1512968, 11.0152484, 10.8793725, 10.7436461, 10.608046, 10.472549, 10.337132, 10.2017719, 10.0664455, 9.93112977, 9.79580144, 9.66043741, 9.52501455, 9.39014909, 9.25548477, 9.1206576, 8.9856237, 8.85033917, 8.71476013, 8.57884268, 8.44254293, 8.305817, 8.16862099, 8.03091101, 7.89264316, 7.75377357, 7.61425834, 7.47405357, 7.33311538, 7.19139987, 7.03975898, 6.88745035, 6.73462746, 6.58144378, 6.42805276, 6.27460789, 6.12126264, 5.96817047, 5.81548485, 5.66335926, 5.51194716, 5.36140203, 5.21187733, 5.06352653, 4.91650311, 4.77096053, 4.63304783, 4.5087701, 4.3859154, 4.26424853, 4.14353429, 4.02353748, 3.90402291, 3.78475538, 3.66549969, 3.54602064, 3.42608303, 3.30545166, 3.18389134, 3.06116687, 2.93704305, 2.81128468, 2.68365656, 2.63751461, 2.62906985, 2.61352859, 2.58794294, 2.54936502, 2.49484693, 2.42144077, 2.32619867, 2.20617273, 2.05841505, 1.87997775, 1.66791294, 1.41927273, 1.13110921, 0.800474516, 0.424420742, 0, 0.0430252401, -0.0634207632, -0.316357415, -0.712804122, -1.24978029, -1.92430532, -2.73339862, -3.6740796, -4.74336766, -5.9382822, -7.25584264, -8.69306837, -10.2469788, -11.9145934, -13.6929314, -15.5790124, -17.6293051, -19.8493217, -22.1016127, -24.3959367, -26.7420523, -29.149718, -31.6286923, -34.188734, -36.8396014, -39.5910531, -42.4528478, -45.434744, -48.5465002, -52.2258979, -55.897731, -59.4433513, -62.7441105, -66.1695525, -69.3569331, -71.8133229, -73.2736162, -73.4727067, -72.1454886, -69.0268558, -66.7841077, -61.3272618, -51.8174804, -37.4159262, -17.2837617, 4.11873131, 34.2542895, 75.2865301, 129.37907, 198.695528, 281.78605, 388.844187, 518.100612, 673.423266, 870.189498, 1118.60917, 1428.89214, 1805.28539, 2265.31686, 2833.01947, 3506.93688, 4342.72214, 5358.80494, 6594.3992, 8123.93649, 10005.6829, 12256.4116, 14923.6311, 18206.9911, 22220.0813, 26792.5817, 32365.671, 39151.2157, 47361.0825, 57207.138, 74713.4381, 92826.6128, 110305.481, 125908.863, 138395.578, 142347.21, 143484.636, 142356.919, 140746.095, 138713.176, 135805.431, 132109.556, 127712.248, 122700.204, 117160.12, 112060.844, 105899.788, 99029.7065, 91803.3538, 84573.4852, 77692.8555, 71514.2198, 66795.4662, 66950.1304},
        {12.5482653, 12.4299663, 12.3110846, 12.1916413, 12.0716576, 11.9511545, 11.8301533, 11.708675, 11.5867407, 11.4643717, 11.341589, 11.2184138, 11.0948672, 10.9709703, 10.8467443, 10.7222103, 10.5973894, 10.4723028, 10.3469716, 10.2214169, 10.0956598, 9.96972157, 9.84362323, 9.71738595, 9.59103086, 9.46457908, 9.33805175, 9.21147, 9.08485496, 8.95822777, 8.83160955, 8.70502144, 8.57848457, 8.45202007, 8.32774433, 8.2045868, 8.0814517, 7.95829218, 7.83506138, 7.71171243, 7.58819846, 7.46447261, 7.34048803, 7.21619784, 7.09155518, 6.9665132, 6.84102501, 6.71504377, 6.58852261, 6.46141466, 6.33367306, 6.19583396, 6.05747161, 5.91874331, 5.77980634, 5.64081797, 5.50193547, 5.36331614, 5.22511724, 5.08749606, 4.95060987, 4.81461595, 4.67967159, 4.54593405, 4.41356062, 4.28270858, 4.15353521, 4.03247331, 3.92580504, 3.82074485, 3.71704319, 3.61445053, 3.51271734, 3.41159406, 3.31083117, 3.21017912, 3.10938837, 3.00820939, 2.90639263, 2.80368856, 2.69984763, 2.59462032, 2.48775707, 2.37900836, 2.35587263, 2.37238068, 2.38150866, 2.38015944, 2.36523589, 2.33364087, 2.28227726, 2.20804793, 2.10785575, 1.97860359, 1.81719431, 1.62053079, 1.3855159, 1.1090525, 0.788043475, 0.419391684, 0, 0.0765291131, -0.00778436482, -0.249138943, -0.643733133, -1.18776544, -1.87743438, -2.70893846, -3.67847619, -4.78224607, -6.01644663, -7.37727636, -8.86093379, -10.4636174, -12.1815257, -14.0108573, -15.9478105, -18.0775522, -20.4089904, -22.7370863, -25.0771031, -27.4443037, -29.8539513, -32.3213091, -34.86164, -37.4902074, -40.2222741, -43.0731034, -46.0579584, -49.1921021, -52.9690751, -56.7504947, -60.4081408, -63.8137933, -67.3322999, -70.5889067, -73.0773748, -74.521564, -74.6453338, -73.1725438, -69.8270536, -67.0304079, -60.9877466, -50.8951213, -35.9485838, -15.3441856, 5.78338743, 35.6613764, 76.5783881, 130.82303, 200.683908, 284.894875, 393.643992, 525.99659, 685.989653, 887.810635, 1141.08964, 1455.45676, 1833.83613, 2294.0879, 2861.38783, 3537.34799, 4371.77186, 5379.24047, 6596.03176, 8095.18787, 9932.87554, 12124.174, 14714.5976, 17900.3153, 21792.2773, 26251.5111, 31677.3394, 38267.2239, 46218.6265, 55729.0092, 72590.8017, 90007.7561, 106778.592, 121702.03, 133576.789, 137103.543, 137911.09, 136554.455, 134881.679, 132935.97, 130221.779, 126809.251, 122768.533, 118169.769, 113083.104, 110992.468, 105991.957, 99007.0903, 90963.3873, 82786.3667, 75401.5474, 69734.4483, 66855.4478, 65664.0718},
        {11.3204304, 11.2159351, 11.1102607, 11.0034517, 10.8955525, 10.7866076, 10.6766613, 10.5657582, 10.4539427, 10.3412591, 10.227752, 10.1134658, 9.99844487, 9.88273371, 9.76637674, 9.6494184, 9.53190313, 9.41387536, 9.29537952, 9.17646007, 9.05716143, 8.93752805, 8.81760436, 8.69743479, 8.57706379, 8.45653579, 8.33589524, 8.21518656, 8.09445419, 7.97374258, 7.85309616, 7.73255936, 7.61217664, 7.49199241, 7.37525696, 7.26034438, 7.14559023, 7.03093489, 6.91631879, 6.80168231, 6.68696586, 6.57210984, 6.45705465, 6.3417407, 6.22610839, 6.11009811, 5.99365028, 5.87670528, 5.75920353, 5.64108543, 5.52229137, 5.39112329, 5.25941236, 5.12735128, 4.99513275, 4.86294948, 4.73099416, 4.5994595, 4.4685382, 4.33842295, 4.20930647, 4.08138145, 3.95484058, 3.82987659, 3.70668215, 3.58544998, 3.46637278, 3.35655995, 3.26295481, 3.1714849, 3.08189453, 2.99392804, 2.90732973, 2.82184393, 2.73721496, 2.65318714, 2.56950478, 2.48591221, 2.40215375, 2.31797372, 2.23311643, 2.14732621, 2.06034737, 1.97192424, 1.97513851, 2.02110258, 2.05980702, 2.08796717, 2.10229838, 2.09951599, 2.07633534, 2.02947177, 1.95564063, 1.85155726, 1.71393701, 1.5394952, 1.3249472, 1.06700833, 0.762393947, 0.407819389, 0, -0.0320822669, -0.206858175, -0.521947185, -0.974968755, -1.56354234, -2.28528741, -3.13782342, -4.11876982, -5.22574608, -6.45637166, -7.80826601, -9.2790486, -10.8663389, -12.5677563, -14.3809204, -16.3034505, -18.3549912, -20.5363084, -22.795203, -25.134014, -27.5550806, -30.0607419, -32.6533371, -35.3352051, -38.1086852, -40.9761165, -43.9398381, -47.002189, -50.1655085, -53.8129136, -57.4263468, -60.8939139, -64.1037207, -67.5592232, -70.8408521, -73.3649444, -74.8350007, -74.9545217, -73.427008, -69.9559603, -66.8657844, -60.4414089, -49.8735485, -34.352918, -13.0702321, 9.22906088, 40.297663, 82.3335429, 137.534669, 208.09901, 292.360834, 401.104111, 533.779933, 694.570659, 897.318327, 1151.52798, 1466.70464, 1846.13178, 2306.95001, 2873.08715, 3545.52995, 4373.36813, 5372.19705, 6577.4545, 8058.20005, 9868.06032, 12020.1459, 14557.5792, 17672.0416, 21471.1811, 25851.0273, 31163.4035, 37589.2099, 45309.3473, 54504.7161, 70659.7425, 87325.9199, 103358.268, 117611.804, 128941.55, 132274.427, 133012.282, 131693.602, 130151.956, 128413.196, 125963.56, 122855.881, 119142.992, 114877.729, 110112.925, 107830.987, 102907.143, 96177.4721, 88478.0542, 80644.9687, 73514.2956, 67922.1144, 66225.6997, 65171.3487},
        {10.2687401, 10.1744034, 10.0783319, 9.98059319, 9.88125488, 9.78038454, 9.67804976, 9.57431815, 9.4692573, 9.36293479, 9.25541822, 9.14677519, 9.03707329, 8.92638011, 8.81476324, 8.70229028, 8.58902882, 8.47504646, 8.36041078, 8.24518939, 8.12944987, 8.01325981, 7.89668682, 7.77979848, 7.66266239, 7.54534613, 7.42791731, 7.31044352, 7.19299235, 7.07563139, 6.95842824, 6.84145049, 6.72476573, 6.60844155, 6.4969687, 6.38811017, 6.27957576, 6.17128955, 6.06317557, 5.95515789, 5.84716056, 5.73910763, 5.63092315, 5.52253118, 5.41385577, 5.30482099, 5.19535087, 5.08536947, 4.97480085, 4.86356906, 4.75159816, 4.62449346, 4.49680816, 4.36877671, 4.24063356, 4.11261318, 3.98495001, 3.85787851, 3.73163314, 3.60644835, 3.4825586, 3.36019835, 3.23960204, 3.12100414, 3.00463911, 2.89074139, 2.77954544, 2.67888418, 2.5964081, 2.51668039, 2.43944292, 2.36443758, 2.29140625, 2.22009079, 2.1502331, 2.08157505, 2.01385851, 1.94682537, 1.8802175, 1.81377678, 1.74724509, 1.68036431, 1.61287632, 1.54452299, 1.57449647, 1.65072169, 1.71992829, 1.7786308, 1.82334374, 1.85058163, 1.85685901, 1.83869039, 1.79259031, 1.71507329, 1.60265386, 1.45184653, 1.25916585, 1.02112632, 0.734242492, 0.395028875, 0, -0.17746109, -0.472181169, -0.883733191, -1.41169011, -2.05562489, -2.81511049, -3.68971985, -4.67902594, -5.78260171, -7.00002013, -8.33085413, -9.77467669, -11.3310608, -12.9995793, -14.7798052, -16.6713116, -18.6055357, -20.5723168, -22.7253442, -25.0495902, -27.5300274, -30.1516281, -32.899365, -35.7582105, -38.7131371, -41.7491171, -44.8511232, -48.0041278, -51.1931033, -54.6530125, -58.0271746, -61.2255649, -64.1581587, -67.4900605, -70.7536812, -73.2800639, -74.7526454, -74.8548622, -73.2701512, -69.681949, -66.3845472, -59.6559197, -48.6686834, -32.5954558, -10.608854, 13.2565193, 45.9055288, 89.3810534, 145.725972, 216.983165, 300.906942, 409.070334, 541.243503, 701.758179, 904.259288, 1158.21032, 1473.07477, 1852.8126, 2313.64163, 2877.78349, 3544.89888, 4364.32646, 5353.60044, 6547.70721, 8011.20496, 9795.07794, 11910.1881, 14397.1175, 17443.5637, 21154.0072, 25459.246, 30664.1641, 36933.363, 44431.4446, 53323.0104, 68778.3877, 84705.0212, 100016.081, 113624.738, 124444.163, 127645.724, 128378.926, 127160.57, 125787.666, 124270.262, 122084.987, 119268.56, 115857.7, 111889.128, 107399.563, 104349.395, 99284.8874, 92847.7118, 85679.5394, 78422.0416, 71716.89, 66205.7558, 65451.1227, 64868.1914},
        {9.37270529, 9.2850527, 9.19519027, 9.1032068, 9.00919108, 8.91323192, 8.81541811, 8.71583845, 8.61458175, 8.5117368, 8.40739241, 8.30163736, 8.19456046, 8.08625051, 7.97679631, 7.86628666, 7.75481036, 7.6424562, 7.52931298, 7.41546951, 7.30101459, 7.18603701, 7.07062557, 6.95486907, 6.83885631, 6.72267609, 6.60641721, 6.49016847, 6.37401867, 6.25805661, 6.14237108, 6.02705089, 5.91218483, 5.7978617, 5.68975368, 5.58504045, 5.48083456, 5.37704362, 5.27357525, 5.17033704, 5.06723662, 4.96418158, 4.86107954, 4.75783809, 4.65436486, 4.55056745, 4.44635346, 4.34163051, 4.2363062, 4.13028814, 4.02348394, 3.89876668, 3.77344776, 3.64780408, 3.52211252, 3.39664995, 3.27169327, 3.14751935, 3.02440508, 2.90262735, 2.78246303, 2.66418901, 2.54808217, 2.4344194, 2.32347758, 2.21553359, 2.11086431, 2.01800171, 1.94527951, 1.87594916, 1.80975239, 1.74643095, 1.68572657, 1.62738099, 1.57113597, 1.51673322, 1.46391451, 1.41242157, 1.36199613, 1.31237995, 1.26331476, 1.2145423, 1.16580431, 1.11684254, 1.1731872, 1.27946067, 1.37901386, 1.46815544, 1.54319414, 1.60043864, 1.63619765, 1.64677987, 1.628494, 1.57764874, 1.4905528, 1.36351487, 1.19284366, 0.974847867, 0.705836191, 0.382117335, 0, -0.323315444, -0.738439838, -1.24689734, -1.85021212, -2.54990833, -3.34751014, -4.2445417, -5.24252718, -6.34299074, -7.54745654, -8.85744874, -10.2744915, -11.800109, -13.4358254, -15.1831648, -17.0436514, -18.8572152, -20.6022954, -22.6459269, -24.9550065, -27.4964312, -30.2370976, -33.1439027, -36.1837432, -39.323516, -42.5301179, -45.7704458, -49.0113964, -52.2198667, -55.4795141, -58.5976626, -61.5061806, -64.1369368, -67.3054037, -70.5016481, -72.9801086, -74.4075727, -74.4508279, -72.7766614, -69.0518607, -65.6064069, -58.6333564, -47.278437, -30.6873766, -8.00590295, 17.5945504, 52.000388, 97.0723085, 154.671011, 226.657193, 310.158815, 417.553774, 548.883105, 708.609907, 910.225665, 1163.1683, 1476.87574, 1855.96415, 2315.78905, 2876.9661, 3537.89701, 4348.20046, 5327.37028, 6509.98599, 7956.18903, 9714.38752, 11793.4693, 14231.5965, 17211.8389, 20835.298, 25064.4328, 30160.3177, 36272.0462, 43548.7121, 52139.4089, 66898.0286, 82092.6669, 96696.2176, 109681.575, 120021.632, 123149.328, 123937.483, 122876.115, 121699.97, 120406.362, 118473.966, 115926.026, 112785.788, 109076.495, 104821.392, 100979.018, 95725.2992, 89527.8569, 82854.3125, 76172.2876, 69949.4035, 64653.2818, 64599.6809, 64628.6308},
        {8.60440315, 8.51890572, 8.43096512, 8.34068372, 8.2481639, 8.15350804, 8.05681851, 7.95819768, 7.85774794, 7.75557165, 7.65177119, 7.54644894, 7.43970728, 7.33164857, 7.2223752, 7.11198954, 7.00059396, 6.88829084, 6.77518256, 6.66137149, 6.54696001, 6.43205049, 6.31674531, 6.20114684, 6.08535746, 5.96947955, 5.85361548, 5.73786762, 5.62233835, 5.50713005, 5.39234509, 5.27808585, 5.1644547, 5.05155402, 4.94573712, 4.84384943, 4.74266125, 4.64207208, 4.54198146, 4.44228891, 4.34289394, 4.24369608, 4.14459484, 4.04548975, 3.94628033, 3.8468661, 3.74714658, 3.64702129, 3.54638974, 3.44515147, 3.34320599, 3.22167332, 3.09963958, 2.97741141, 2.85529541, 2.73359822, 2.61262646, 2.49268675, 2.37408572, 2.25712999, 2.14212618, 2.02938092, 1.91920083, 1.81189253, 1.70776265, 1.60711781, 1.51026463, 1.42638148, 1.36443373, 1.30643068, 1.25210385, 1.20118474, 1.15340485, 1.10849568, 1.06618875, 1.02621555, 0.988307587, 0.952196372, 0.917613406, 0.884290193, 0.851958237, 0.820349043, 0.789194116, 0.758224959, 0.838302958, 0.971257208, 1.09758111, 1.2133998, 1.31483838, 1.39802201, 1.45907578, 1.49412485, 1.49929432, 1.47070933, 1.404495, 1.29677646, 1.14367884, 0.94132725, 0.685846832, 0.373362707, 0, -0.341680611, -0.77563049, -1.30340701, -1.92656753, -2.64666943, -3.46527008, -4.38392686, -5.40419712, -6.52763824, -7.75580759, -9.09026254, -10.5325605, -12.0842587, -13.7469147, -15.5220858, -17.4113293, -19.2228152, -20.9304742, -22.9732874, -25.3113721, -27.9048455, -30.7138249, -33.6984273, -36.8187701, -40.0349705, -43.3071456, -46.5954127, -49.859889, -53.0606917, -56.2381496, -59.2427574, -62.0105689, -64.4776377, -67.4842463, -70.5143339, -72.8107126, -74.0381673, -73.8614834, -71.9454458, -67.9548399, -64.3427192, -57.1369969, -45.4569276, -28.4217654, -5.15076491, 21.7215614, 57.4700319, 103.854207, 162.633646, 235.56791, 319.524464, 427.13419, 558.747015, 718.864893, 920.466828, 1172.7258, 1484.8148, 1860.52314, 2315.63103, 2871.79173, 3529.65215, 4335.1137, 5306.06097, 6475.23417, 7900.54563, 9628.35329, 11667.5384, 14054.7429, 16964.0225, 20491.7042, 24620.3817, 29577.612, 35499.0909, 42520.5143, 50777.5779, 64801.1295, 79232.9249, 93109.872, 105468.879, 115346.853, 118473.515, 119397.751, 118573.836, 117619.154, 116527.582, 114808.364, 112479.606, 109559.415, 106065.898, 102017.162, 100060.112, 95337.628, 88799.8445, 81396.8946, 74078.9117, 67796.0293, 63498.3807, 63781.4892, 64070.2214},
        {7.94503275, 7.85761018, 7.76776546, 7.67560602, 7.58123925, 7.48477255, 7.38631333, 7.28596899, 7.18384693, 7.08005456, 6.97469928, 6.86788849, 6.7597296, 6.65033001, 6.53979712, 6.42823834, 6.31576106, 6.2024727, 6.08848066, 5.97389233, 5.85881512, 5.74335644, 5.62762369, 5.51172427, 5.39576558, 5.27985503, 5.16410003, 5.04860796, 4.93348624, 4.81884228, 4.70478346, 4.59141721, 4.47885091, 4.36719198, 4.26287155, 4.16270214, 4.06342028, 3.96492815, 3.86712794, 3.76992182, 3.673212, 3.57690064, 3.48088994, 3.38508208, 3.28937926, 3.19368365, 3.09789743, 3.00192281, 2.90566196, 2.80901706, 2.71189031, 2.59492104, 2.47769186, 2.36052255, 2.24373286, 2.12764257, 2.01257144, 1.89883924, 1.78676573, 1.67667069, 1.56887387, 1.46369505, 1.36145399, 1.26247045, 1.16706421, 1.07555503, 0.988262682, 0.914878065, 0.864867097, 0.819222265, 0.777655842, 0.739880102, 0.705607319, 0.674549765, 0.646419714, 0.620929441, 0.597791217, 0.576717318, 0.557420015, 0.539611584, 0.523004296, 0.507310427, 0.492242248, 0.477512034, 0.578218147, 0.733952824, 0.882921452, 1.0210918, 1.14443165, 1.24890877, 1.33049094, 1.38514592, 1.40884148, 1.39754541, 1.34722548, 1.25384945, 1.11338511, 0.921800216, 0.675062553, 0.36913989, 0, -0.217082045, -0.555838014, -1.01571922, -1.59617699, -2.29666262, -3.11662745, -4.05552277, -5.11279991, -6.28791018, -7.5803049, -8.98943539, -10.514753, -12.1557089, -13.9117546, -15.7823413, -17.7669203, -19.7111187, -21.5924129, -23.7581901, -26.1749392, -28.8091493, -31.6273092, -34.5959081, -37.6814347, -40.8503781, -44.0692273, -47.3044713, -50.5225989, -53.6900993, -56.929146, -59.993459, -62.8028217, -65.2770177, -68.1394179, -70.9080122, -72.8865004, -73.7535898, -73.1879876, -70.8684012, -66.4735378, -62.6605087, -55.2231464, -43.2574272, -25.8593271, -2.12482221, 25.4231971, 61.9723976, 109.283532, 169.117352, 243.23461, 328.678922, 437.693568, 570.978497, 732.971938, 935.744428, 1187.93975, 1498.2017, 1867.92767, 2314.65475, 2863.82514, 3521.99669, 4327.12784, 5291.63133, 6444.88855, 7845.03335, 9537.00172, 11531.9321, 13865.998, 16699.2978, 20121.8308, 24123.1862, 28909.2215, 34604.6935, 41334.3592, 49222.9753, 62477.3193, 76119.6224, 89254.1362, 100985.113, 110416.803, 113605.502, 114735.391, 114216.988, 113497.452, 112576.966, 111023.781, 108859.678, 106106.436, 102785.837, 98919.6613, 101590.517, 98152.1991, 90703.2307, 81342.1352, 72167.4354, 65277.6547, 62771.316, 63054.94, 63125.2355},
        {7.38736532, 7.29629265, 7.2028977, 7.10728949, 7.00957703, 6.9098693, 6.80827532, 6.7049041, 6.59986462, 6.49326591, 6.38521695, 6.27582676, 6.16520434, 6.05345869, 5.94069881, 5.82703371, 5.71257239, 5.59742386, 5.48169712, 5.36550116, 5.248945, 5.13213764, 5.01518809, 4.89820533, 4.78129839, 4.66457626, 4.54814794, 4.43212244, 4.31660877, 4.20171592, 4.0875529, 3.97422871, 3.86185236, 3.75053285, 3.64650193, 3.54667845, 3.44790766, 3.35009986, 3.25316538, 3.15701451, 3.06155757, 2.96670487, 2.8723667, 2.77845339, 2.68487524, 2.59154256, 2.49836565, 2.40525483, 2.3121204, 2.21887268, 2.12542196, 2.01257705, 1.89976384, 1.78730674, 1.67553014, 1.56475842, 1.45531598, 1.34752722, 1.24171652, 1.13820827, 1.03732688, 0.939396722, 0.8447422, 0.753687703, 0.666557623, 0.583676352, 0.505368282, 0.441554327, 0.40192499, 0.36701256, 0.336519324, 0.310147573, 0.287599594, 0.268577676, 0.252784108, 0.239921179, 0.229691178, 0.221796392, 0.215939111, 0.211821624, 0.209146218, 0.207615184, 0.20693081, 0.206795383, 0.326818215, 0.504226, 0.674804243, 0.834364021, 0.978716408, 1.10367248, 1.20504331, 1.27863998, 1.32027356, 1.32575512, 1.29089575, 1.21150651, 1.08339848, 0.902382741, 0.664270365, 0.364872426, 0, -0.0767491955, -0.30751503, -0.689431775, -1.2196337, -1.89525509, -2.7134302, -3.67129331, -4.76597869, -5.99462062, -7.35435336, -8.8423112, -10.4556284, -12.1914392, -14.046878, -16.0190789, -18.1051762, -20.198972, -22.2828387, -24.5876384, -27.0883628, -29.7600035, -32.5775522, -35.5160005, -38.55034, -41.6555623, -44.8066592, -47.9786221, -51.1464428, -54.2851128, -57.6218635, -60.786959, -63.6797189, -66.1994631, -68.9192366, -71.4014964, -73.0290396, -73.4990681, -72.5087841, -69.7553897, -64.936087, -60.9511386, -53.3198416, -41.1127122, -23.4002669, 0.746977977, 28.8748612, 66.1424226, 114.325057, 175.19816, 250.537125, 337.63596, 448.2287, 583.304126, 747.205869, 951.127044, 1203.18549, 1511.49904, 1875.17747, 2313.39397, 2855.14869, 3512.58246, 4316.0772, 5272.5987, 6408.326, 7781.99391, 9437.3936, 11388.7083, 13672.0532, 16432.1055, 19752.1338, 23622.6003, 28233.8573, 33701.2983, 40140.3169, 47666.3066, 60179.0351, 73063.4283, 85488.786, 96624.4083, 105639.595, 108911.75, 110263.333, 110060.173, 109568.666, 108798.163, 107382.779, 105350.627, 102729.819, 99548.4699, 95834.6924, 102596.511, 100293.443, 92052.3309, 81000.0171, 70263.3438, 62969.1535, 62244.2886, 62412.8117, 62125.1945},
        {6.91693331, 6.82164821, 6.72416173, 6.62458239, 6.5230187, 6.4195792, 6.31437239, 6.2075068, 6.09909094, 5.98923334, 5.87804251, 5.76562697, 5.65209525, 5.53755586, 5.42211732, 5.30588815, 5.18897687, 5.07149199, 4.95354205, 4.83523555, 4.71668102, 4.59798697, 4.47926193, 4.36061441, 4.24215293, 4.12398601, 4.00622218, 3.88896995, 3.77233783, 3.65643436, 3.54136804, 3.4272474, 3.31418096, 3.20227723, 3.09734902, 2.99653273, 2.89689619, 2.79836281, 2.70085597, 2.6042991, 2.50861559, 2.41372885, 2.31956229, 2.22603929, 2.13308328, 2.04061765, 1.94856581, 1.85685117, 1.76539711, 1.67412706, 1.58296442, 1.47348793, 1.36436333, 1.2559117, 1.14845412, 1.04231166, 0.937805419, 0.835256463, 0.734985876, 0.637314738, 0.542564128, 0.451055128, 0.363108817, 0.279046276, 0.199188583, 0.12385682, 0.053372067, -0.00250379857, -0.0342355677, -0.0609730986, -0.0830073195, -0.100629159, -0.114129546, -0.123799408, -0.129929674, -0.132811272, -0.132735131, -0.12999218, -0.124873346, -0.117669559, -0.108671746, -0.0981708368, -0.0864577588, -0.0738234409, 0.0645840799, 0.263270217, 0.455236264, 0.636130163, 0.801599858, 0.947293295, 1.06885842, 1.16194317, 1.22219549, 1.24526333, 1.22679463, 1.16243734, 1.0478394, 0.878648749, 0.650513337, 0.359081106, 0, 0.0413548791, -0.0987470822, -0.415495785, -0.90408113, -1.55969302, -2.37752135, -3.35275602, -4.48058694, -5.75620401, -7.17479712, -8.73155619, -10.4216711, -12.2403318, -14.1827281, -16.2440499, -18.4194872, -20.6457524, -22.9096813, -25.3387345, -27.915571, -30.6228498, -33.4432299, -36.3593704, -39.3539301, -42.4095681, -45.5089434, -48.634715, -51.7695419, -54.8960832, -58.342417, -61.6191294, -64.6052536, -67.1798229, -69.7625888, -71.9622254, -73.2432152, -73.3223761, -71.9165258, -68.7424821, -63.5170627, -59.4030868, -51.6225878, -39.2181657, -21.2324208, 3.29204689, 32.0081493, 70.0038962, 119.06281, 180.968411, 257.504223, 346.197458, 458.290101, 594.962201, 760.452481, 965.244472, 1217.01791, 1523.45254, 1882.0486, 2312.97886, 2847.52095, 3501.67637, 4300.19367, 5245.88855, 6362.46728, 7709.03406, 9328.44485, 11238.3481, 13475.0647, 16167.3424, 19391.6474, 23134.8114, 27576.7243, 32824.8791, 38986.7691, 46169.8874, 57994.3183, 70176.8158, 81946.7253, 92533.3921, 101166.161, 104521.489, 106083.536, 106175.744, 105890.467, 105246.784, 103943.394, 102015.395, 99497.8844, 96425.9614, 92834.7237, 101963.189, 100329.066, 91678.9625, 79759.4882, 68317.2517, 61098.8615, 61850.9266, 61888.9266, 61151.8756},
        {6.51856216, 6.42020882, 6.31969944, 6.21714252, 6.11264661, 6.00632023, 5.89827191, 5.78861019, 5.6774436, 5.56488065, 5.45102989, 5.33599985, 5.21989905, 5.10283602, 4.9849193, 4.86625741, 4.74695888, 4.62713225, 4.50688605, 4.3863288, 4.26556903, 4.14471528, 4.02387607, 3.90315993, 3.7826754, 3.66253101, 3.54283528, 3.42369674, 3.30522393, 3.18752538, 3.0707096, 2.95488515, 2.84016054, 2.7266443, 2.61965384, 2.5165837, 2.41476579, 2.31413961, 2.21464465, 2.1162204, 2.01880635, 1.92234199, 1.82676683, 1.73202034, 1.63804202, 1.54477136, 1.45214786, 1.360111, 1.26860028, 1.17755519, 1.08691523, 0.979496805, 0.872733685, 0.766936552, 0.662416093, 0.559482993, 0.45844794, 0.35962162, 0.26331472, 0.169837925, 0.0795019235, -0.00738259948, -0.0905049569, -0.169554462, -0.244220429, -0.314192171, -0.379159001, -0.430119735, -0.458281842, -0.481258139, -0.499301306, -0.512664029, -0.521598989, -0.52635887, -0.527196356, -0.524364128, -0.518114872, -0.508701268, -0.496376002, -0.481391755, -0.464001211, -0.444457054, -0.423011965, -0.399918629, -0.243697554, -0.0232383179, 0.190986179, 0.39444829, 0.582620372, 0.750974778, 0.894983862, 1.01011998, 1.09185548, 1.13566273, 1.13701407, 1.09138187, 0.994238465, 0.841056223, 0.627307495, 0.348464636, 0, 0.0738800796, -0.0424924043, -0.343787384, -0.824674792, -1.47982456, -2.30390662, -3.29159091, -4.43754735, -5.73644588, -7.18295644, -8.77174895, -10.4974933, -12.3548596, -14.3385175, -16.4431372, -18.6633884, -20.9459456, -23.2786221, -25.7646495, -28.3884128, -31.1342972, -33.9866878, -36.9299699, -39.9485286, -43.026749, -46.1490163, -49.2997157, -52.4632322, -55.6239511, -59.1626198, -62.5219283, -65.5673531, -68.1643707, -70.6178248, -72.5735083, -73.5600495, -73.3111146, -71.5603701, -68.0414823, -62.4881175, -58.3554254, -50.5229636, -37.996282, -19.7809308, 5.11754018, 34.7386574, 73.7705252, 123.946324, 186.999236, 264.66244, 354.442334, 467.464952, 605.030549, 771.328964, 976.449604, 1227.83309, 1532.92004, 1889.25809, 2316.42925, 2844.80793, 3489.85487, 4275.7436, 5204.89652, 6300.44026, 7620.58365, 9207.43224, 11082.1595, 13281.4442, 15918.7157, 19063.9891, 22698.238, 26994.2172, 32052.7637, 37974.7145, 44860.9064, 56094.2423, 67672.9767, 78877.4302, 88987.9233, 97284.7764, 100697.793, 102424.822, 102753.548, 102633.395, 102090.11, 100873.714, 99023.7401, 96579.7218, 93581.1917, 90067.6827, 97394.2841, 95288.6813, 87128.9199, 76293.0451, 66159.1023, 60105.1369, 61509.1943, 61591.041, 60354.9422},
        {6.1775284, 6.07861052, 5.9774345, 5.87411116, 5.76875133, 5.66146584, 5.55236551, 5.44156117, 5.32916365, 5.21528377, 5.10003235, 4.98352024, 4.86585824, 4.74715719, 4.62752792, 4.50708125, 4.38592801, 4.26417902, 4.14194511, 4.01933711, 3.89646585, 3.77344214, 3.65037682, 3.52738072, 3.40456465, 3.28203945, 3.15991594, 3.03830495, 2.91731731, 2.79706384, 2.67765537, 2.55920272, 2.44181673, 2.32560821, 2.21541131, 2.10887584, 2.00359486, 1.89952592, 1.79662656, 1.69485433, 1.59416678, 1.49452144, 1.39587587, 1.29818762, 1.20141422, 1.10551322, 1.01044218, 0.916158634, 0.822620131, 0.729784215, 0.637608435, 0.530534855, 0.424372848, 0.319416303, 0.215959114, 0.11429517, 0.0147183625, -0.0824774163, -0.176998276, -0.268550324, -0.35683967, -0.441572424, -0.522454692, -0.599192585, -0.671492211, -0.739059679, -0.801601097, -0.851622543, -0.881802965, -0.906697995, -0.926480491, -0.94132331, -0.95139931, -0.956881348, -0.957942282, -0.954754969, -0.947492267, -0.936327032, -0.921432123, -0.902980398, -0.881144712, -0.856097925, -0.828012893, -0.797062474, -0.623226297, -0.379722611, -0.141454676, 0.0868551099, 0.30048435, 0.494710646, 0.664811601, 0.806064818, 0.9137479, 0.983138448, 1.00951407, 0.988152359, 0.914330925, 0.78332737, 0.590419296, 0.330884305, 0, -0.0263255058, -0.223078236, -0.586585636, -1.11317515, -1.79917423, -2.64091031, -3.63471084, -4.77690327, -6.06381504, -7.4917736, -9.0571064, -10.7561409, -12.5852045, -14.5406246, -16.6187288, -18.8158445, -21.0354415, -23.2605822, -25.6976285, -28.32301, -31.113156, -34.044496, -37.0934594, -40.2364756, -43.4499741, -46.7103841, -49.9941353, -53.2776569, -56.5373784, -60.1251274, -63.5042891, -66.5386735, -69.0920906, -71.4272566, -73.2085284, -73.9938878, -73.527473, -71.5534219, -67.8158728, -62.0589638, -58.0511399, -50.2874432, -37.7246473, -19.3195258, 5.97114746, 36.9921901, 77.5348395, 129.191924, 193.556272, 272.220713, 362.272742, 475.317038, 612.689389, 778.62389, 983.272011, 1234.12844, 1538.68786, 1896.92256, 2325.55928, 2849.531, 3477.49691, 4240.24746, 5145.27047, 6217.79286, 7513.09957, 9072.67842, 10920.9227, 13094.8882, 15694.311, 18783.4716, 22337.114, 26522.8302, 31435.894, 37171.5793, 43825.1602, 54596.902, 65700.8393, 76455.998, 86181.4043, 94196.084, 97618.3376, 99435.0673, 99908.4672, 99895.9759, 99424.1699, 98271.6078, 96477.267, 94080.1254, 91119.1606, 87633.3502, 87347.2968, 83183.2828, 76768.5438, 69730.3154, 63695.8333, 60292.333, 61147.0501, 61579.7377, 59840.7429},
        {5.88345485, 5.78449219, 5.68318836, 5.57965296, 5.47399559, 5.36632584, 5.25675332, 5.14538762, 5.03233833, 4.91771505, 4.80162739, 4.68418493, 4.56549728, 4.44567403, 4.32482478, 4.20305913, 4.08048667, 3.957217, 3.83335972, 3.70902443, 3.58432072, 3.45935818, 3.33424643, 3.20909505, 3.08401363, 2.95911179, 2.83449912, 2.7102852, 2.58657965, 2.46349205, 2.34113201, 2.21960912, 2.09903297, 1.97951318, 1.8649762, 1.75354291, 1.64328833, 1.53419817, 1.42625817, 1.31945406, 1.21377158, 1.10919645, 1.00571442, 0.903311201, 0.80197254, 0.701684165, 0.602431807, 0.504201201, 0.406978077, 0.310748167, 0.215497205, 0.108112686, 0.00196252229, -0.102683609, -0.205556033, -0.306385071, -0.404901048, -0.500834289, -0.593915115, -0.683873851, -0.770440821, -0.853346348, -0.932320755, -1.00709437, -1.07739751, -1.1429605, -1.20351367, -1.25330948, -1.28673194, -1.31480901, -1.33762614, -1.35526875, -1.36782228, -1.37537215, -1.3780038, -1.37580265, -1.36885413, -1.35724369, -1.34105674, -1.32037872, -1.29529505, -1.26589118, -1.23225252, -1.19446451, -1.00272546, -0.735217034, -0.471922422, -0.217791183, 0.0222271272, 0.243182952, 0.440126736, 0.608108923, 0.742179956, 0.837390281, 0.888790341, 0.89143058, 0.840361442, 0.730633371, 0.557296811, 0.315402206, 0, -0.150353228, -0.448988078, -0.893942714, -1.4832553, -2.21496401, -3.087107, -4.09772245, -5.24484851, -6.52652335, -7.94078513, -9.48567203, -11.1592222, -12.9594738, -14.8844651, -16.9322341, -19.100819, -21.2490348, -23.3550306, -25.7317542, -28.3474104, -31.1702037, -34.1683387, -37.3100201, -40.5634524, -43.8968403, -47.2783883, -50.6763012, -54.0587833, -57.3940395, -60.991676, -64.353315, -67.3447407, -69.8317371, -72.07502, -73.7429077, -74.3984033, -73.7888114, -71.6614365, -67.7635832, -61.8425562, -57.8074749, -49.9771897, -37.284737, -18.6631533, 6.95452507, 38.7929069, 80.221899, 132.769054, 197.961923, 277.32806, 367.887794, 481.184726, 618.10538, 782.777028, 985.943473, 1235.37783, 1538.85321, 1899.37453, 2330.56679, 2851.25668, 3463.07865, 4203.30823, 5084.3568, 6133.38898, 7402.56799, 8932.58225, 10750.3328, 12892.9336, 15448.0524, 18474.7609, 21939.0137, 26007.3251, 30770.9997, 36321.3418, 42749.6559, 53109.9731, 63790.1896, 74140.9281, 83512.8112, 91256.4618, 94638.0363, 96482.268, 97033.1225, 97082.903, 96658.5247, 95558.4286, 93820.9537, 91484.4388, 88587.2229, 85167.6447, 77541.8611, 71478.8994, 66869.6833, 63605.1361, 61576.1814, 60673.7425, 60788.7431, 61461.109, 59298.3121},
        {5.60830886, 5.51124209, 5.41177704, 5.31001505, 5.20605748, 5.1000057, 4.99196104, 4.88202487, 4.77029853, 4.65688339, 4.5418808, 4.4253921, 4.30751866, 4.18836183, 4.06802296, 3.9466034, 3.82420452, 3.70092765, 3.57687417, 3.45214542, 3.32684275, 3.20106752, 3.07492108, 2.94850479, 2.82192001, 2.69526807, 2.56865035, 2.44216818, 2.31592294, 2.19001596, 2.06454861, 1.93962224, 1.81533819, 1.69179784, 1.57095639, 1.45194928, 1.33388965, 1.21681869, 1.10077759, 0.985807542, 0.871949744, 0.759245385, 0.647735658, 0.537461754, 0.428464865, 0.320786185, 0.214466905, 0.109548218, 0.00607131551, -0.0959226102, -0.196392367, -0.303903123, -0.409620756, -0.513317506, -0.61476561, -0.713737308, -0.810004838, -0.903340439, -0.99351635, -1.08030481, -1.16347806, -1.24280833, -1.31806787, -1.38902891, -1.45546369, -1.51714445, -1.57384344, -1.62170511, -1.65696105, -1.68686548, -1.71142583, -1.7306495, -1.74454391, -1.75311647, -1.7563746, -1.75432569, -1.74697717, -1.73433645, -1.71641093, -1.69320804, -1.66473519, -1.63099978, -1.59200923, -1.54777095, -1.33683077, -1.04332302, -0.753322681, -0.472076896, -0.204832818, 0.0431624009, 0.266661611, 0.460417661, 0.6191834, 0.737711677, 0.810755342, 0.833067242, 0.799400227, 0.704507147, 0.543140849, 0.310054184, 0, -0.276684935, -0.686918139, -1.22953814, -1.90338346, -2.70729262, -3.64010416, -4.7006566, -5.88778847, -7.20033829, -8.63714459, -10.1970459, -11.8788807, -13.6814876, -15.6037051, -17.6443717, -19.8023259, -21.9152954, -23.9591027, -26.2970041, -28.8933143, -31.7123479, -34.7184199, -37.8758449, -41.1489378, -44.5020131, -47.8993858, -51.3053706, -54.6842821, -58.0004352, -61.5305063, -64.8119159, -67.7152702, -70.1111755, -72.3232487, -73.9955908, -74.6514998, -74.0256788, -71.8528308, -67.8676589, -61.8048662, -57.2397057, -48.8974673, -35.7614423, -16.8149217, 8.9588035, 39.7341482, 80.0806887, 131.725707, 196.396487, 275.820309, 368.18838, 483.08593, 619.584768, 781.212578, 980.594448, 1226.27092, 1526.78252, 1888.53135, 2322.68241, 2842.73355, 3446.46825, 4172.80974, 5035.33931, 6061.0041, 7299.39893, 8789.72297, 10559.4873, 12644.8609, 15124.8689, 18054.2944, 21385.1929, 25290.8797, 29861.6978, 35187.9901, 41360.0994, 51336.0482, 61621.58, 71590.118, 80615.0853, 88069.9049, 91313.3483, 93076.5914, 93596.4793, 93633.7613, 93217.904, 92160.0806, 90500.6726, 88280.061, 85538.6273, 82316.7526, 71603.9587, 65017.7299, 61597.634, 60383.2384, 60414.1107, 60729.8184, 60369.9292, 60866.3439, 58537.6604},
        {5.36253757, 5.26678205, 5.16867576, 5.06830599, 4.96576004, 4.86112521, 4.75448878, 4.64593806, 4.53556035, 4.42344292, 4.30967309, 4.19433815, 4.07752538, 3.95932209, 3.83981557, 3.71909312, 3.59724203, 3.4743496, 3.35050312, 3.22578989, 3.1002972, 2.97411235, 2.84732263, 2.72001533, 2.59227776, 2.46419721, 2.33586098, 2.20735635, 2.07877062, 1.9501911, 1.82170506, 1.69339982, 1.56536266, 1.43768088, 1.30987489, 1.18232734, 1.05542771, 0.929281707, 0.803995011, 0.679673318, 0.556422319, 0.434347707, 0.313555172, 0.194150408, 0.0762391056, -0.0400730426, -0.154680345, -0.267477109, -0.378357643, -0.487216256, -0.593947254, -0.702037442, -0.807710754, -0.91078362, -1.01107247, -1.10839373, -1.20256384, -1.29339922, -1.3807163, -1.46433151, -1.54406129, -1.61972206, -1.69113025, -1.75810229, -1.82045461, -1.87800365, -1.93056582, -1.976091, -2.01257388, -2.04368052, -2.06934836, -2.08951482, -2.10411735, -2.11309337, -2.11638031, -2.1139156, -2.10563669, -2.09148098, -2.07138593, -2.04528896, -2.01312749, -1.97483897, -1.93036083, -1.87963049, -1.64865952, -1.32800657, -1.01032078, -0.701183794, -0.406177277, -0.130882889, 0.11911771, 0.338242858, 0.520910897, 0.661540165, 0.754549003, 0.794355749, 0.775378743, 0.692036326, 0.538746837, 0.309928615, 0, -0.379942009, -0.88729715, -1.52089499, -2.2795651, -3.16213705, -4.1674404, -5.29430473, -6.54155959, -7.90803457, -9.39255923, -10.9939631, -12.7110758, -14.5427269, -16.487746, -18.5449626, -20.7132062, -22.8252388, -24.8561659, -27.1804466, -29.7613243, -32.5620428, -35.5458456, -38.6759764, -41.9156788, -45.2281965, -48.5767731, -51.9246522, -55.2350774, -58.4712923, -61.8830724, -65.0420679, -67.8255628, -70.1108413, -72.322317, -74.0637092, -74.7928361, -74.222843, -72.066875, -68.0380774, -61.8495956, -56.4957233, -47.4098467, -33.6631473, -14.3268065, 11.5279944, 40.3588972, 78.6836913, 128.54899, 192.001405, 271.087551, 365.710049, 482.679934, 618.755612, 776.666629, 971.333611, 1212.20364, 1508.72379, 1870.42081, 2306.64401, 2826.65581, 3426.66377, 4144.2122, 4991.80132, 5994.90387, 7200.9739, 8647.37566, 10361.1685, 12377.0449, 14767.4825, 17583.5854, 20762.3704, 24486.581, 28846.8133, 33933.6633, 39837.7271, 49446.3505, 59354.1923, 68952.6609, 77633.1649, 84787.113, 87833.7556, 89451.4313, 89876.7591, 89862.4624, 89440.869, 88429.2631, 86870.7247, 84808.3335, 82285.1696, 79344.3128, 68365.517, 62143.0809, 59463.8547, 59114.6886, 59882.4327, 60553.9372, 59916.0523, 59993.7039, 57642.6077},
        {5.15657962, 5.05892272, 4.95915203, 4.85733604, 4.75354324, 4.64784213, 4.5403012, 4.43098894, 4.31997384, 4.2073244, 4.0931091, 3.97739644, 3.86025491, 3.741753, 3.62195921, 3.50094203, 3.37876994, 3.25551144, 3.13123503, 3.00600919, 2.87990242, 2.75298321, 2.62532005, 2.49698143, 2.36803585, 2.23855179, 2.10859775, 1.97824223, 1.84755371, 1.71660068, 1.58545164, 1.45417508, 1.3228395, 1.19151337, 1.05742458, 0.922121666, 0.787187333, 0.652782256, 0.51906711, 0.386202571, 0.254349315, 0.123668019, -0.00568064311, -0.133535994, -0.259737358, -0.384124059, -0.506535422, -0.626810771, -0.744789429, -0.860310721, -0.973213971, -1.08298157, -1.18981751, -1.29356886, -1.39408268, -1.49120602, -1.58478595, -1.67466954, -1.76070384, -1.84273592, -1.92061283, -1.99418165, -2.06328942, -2.12778322, -2.1875101, -2.24231713, -2.29205137, -2.33613579, -2.37400357, -2.4063764, -2.43312882, -2.4541354, -2.46927069, -2.47840924, -2.48142561, -2.47819434, -2.46858999, -2.45248712, -2.42976027, -2.400284, -2.36393287, -2.32058142, -2.27010421, -2.2123758, -1.96101044, -1.61293058, -1.26729734, -0.930029793, -0.607046982, -0.30426797, -0.0276118136, 0.217002429, 0.4236557, 0.586428942, 0.699403098, 0.756659111, 0.752277921, 0.680340473, 0.534927709, 0.31012057, 0, -0.432979213, -0.997350021, -1.69120201, -2.51262475, -3.45970784, -4.53054086, -5.72321339, -7.03581501, -8.46643532, -10.0131639, -11.6740903, -13.4473041, -15.330895, -17.3229524, -19.421566, -21.6248254, -23.7695066, -25.8306545, -28.1712343, -30.7547686, -33.5447797, -36.5047901, -39.598322, -42.788898, -46.0400404, -49.3152717, -52.5781141, -55.7920902, -58.9207223, -62.1898683, -65.2045254, -67.8495155, -70.0096603, -72.2211225, -74.0430533, -74.8609134, -74.3641222, -72.2420993, -68.1842643, -61.8800366, -55.7237108, -45.8766088, -31.4989352, -11.7508951, 14.2073067, 41.2168658, 77.6158267, 125.743634, 187.939734, 266.54357, 363.008316, 481.64291, 617.271586, 771.923964, 962.34066, 1198.65735, 1491.00973, 1851.10866, 2287.15663, 2805.6377, 3402.60478, 4112.97025, 4947.39066, 5929.49505, 7104.89367, 8509.10702, 10168.4976, 12116.2116, 14418.9746, 17124.5179, 20157.8425, 23708.2914, 27866.0791, 32721.4204, 38364.5299, 47611.8701, 57145.846, 66375.1103, 74708.3159, 81554.1155, 84383.4447, 85834.4848, 86146.3388, 86077.5858, 85661.7028, 84715.7184, 83283.932, 81410.643, 79140.1509, 76516.7551, 66699.1883, 61248.641, 59020.8716, 58871.6384, 59656.7, 60231.8146, 59452.7407, 59040.5368, 56696.9973},
        {5.01139201, 4.90058031, 4.78844207, 4.67501927, 4.56035393, 4.44448804, 4.32746362, 4.20932265, 4.09010716, 3.96985913, 3.84862058, 3.7264335, 3.6033399, 3.47938178, 3.35460115, 3.22904, 3.10274035, 2.97574419, 2.84809353, 2.71983037, 2.59099671, 2.46163456, 2.33178592, 2.20149279, 2.07079717, 1.93974108, 1.80836651, 1.67671546, 1.54482994, 1.41275195, 1.2805235, 1.14818658, 1.01578321, 0.883355378, 0.747181624, 0.609264846, 0.471653173, 0.334510742, 0.198001686, 0.062290141, -0.0724597586, -0.206083878, -0.338418082, -0.469298235, -0.598560204, -0.726039852, -0.851573046, -0.97499565, -1.09614353, -1.21485255, -1.33095857, -1.44485227, -1.55580267, -1.66363363, -1.76816896, -1.86923252, -1.96664814, -2.06023966, -2.14983092, -2.23524576, -2.316308, -2.3928415, -2.46467009, -2.53161761, -2.59350789, -2.65016478, -2.70141212, -2.7469649, -2.78654077, -2.820188, -2.84773748, -2.86902011, -2.88386679, -2.8921084, -2.89357584, -2.8881, -2.87551178, -2.85564206, -2.82832174, -2.79338171, -2.75065287, -2.6999661, -2.64115231, -2.57404238, -2.30269828, -1.9289531, -1.55699292, -1.19300172, -0.843163474, -0.513662146, -0.210681708, 0.059593867, 0.290980609, 0.477294546, 0.612351707, 0.689968122, 0.703959818, 0.648142826, 0.516333172, 0.302346888, 0, -0.333614483, -0.825601453, -1.47200637, -2.26887471, -3.21225193, -4.2981835, -5.52271488, -6.88189154, -8.37175894, -9.98836254, -11.7277478, -13.5859602, -15.5590453, -17.6430484, -19.834015, -22.1279906, -24.3711823, -26.53823, -28.9640995, -31.6127278, -34.4480524, -37.4340103, -40.5345389, -43.7135753, -46.9350569, -50.1629209, -53.3611044, -56.4935448, -59.5241792, -62.65666, -65.5273138, -68.0281634, -70.0512314, -72.2164429, -74.0518696, -74.8914755, -74.408913, -72.2778338, -68.1718902, -61.7647341, -55.2345672, -44.9882377, -30.2093771, -10.0816167, 16.2114118, 43.4077584, 79.92169, 128.089155, 190.246101, 268.728476, 365.066537, 483.386659, 618.797181, 773.308431, 963.0223, 1197.71385, 1487.15816, 1842.2364, 2271.36614, 2781.75827, 3370.53724, 4070.79419, 4892.91728, 5859.57725, 7014.25697, 8390.88621, 10015.4794, 11919.9141, 14165.091, 16795.4936, 19741.1373, 23178.5843, 27191.9418, 31865.3171, 37282.8172, 46126.2403, 55232.5799, 64036.5205, 71972.7465, 78475.9425, 81103.8134, 82419.3429, 82661.635, 82611.5875, 82293.8447, 81507.3038, 80287.8965, 78671.5546, 76694.2099, 74391.794, 65827.6349, 61000.6098, 58942.6109, 58685.5303, 59261.2599, 59701.6918, 59038.7181, 58326.1604, 55863.007},
        {4.90545191, 4.77877057, 4.65165878, 4.52413118, 4.3962024, 4.2678871, 4.13919992, 4.01015549, 3.88076846, 3.75105346, 3.62102515, 3.49069817, 3.36008714, 3.22920673, 3.09807156, 2.96669628, 2.83509554, 2.70328396, 2.57127621, 2.43908691, 2.30673071, 2.17422225, 2.04157617, 1.90880711, 1.77592973, 1.64295865, 1.50990852, 1.37679398, 1.24362967, 1.11043024, 0.977210321, 0.843984564, 0.710767606, 0.57757409, 0.440186441, 0.300824428, 0.161758687, 0.0231412045, -0.114876034, -0.252141044, -0.38850184, -0.523806436, -0.657902848, -0.790639089, -0.921863174, -1.05142312, -1.17916694, -1.30494264, -1.42859825, -1.54998178, -1.66894124, -1.7882772, -1.90482112, -2.018357, -2.12866886, -2.2355407, -2.33875654, -2.43810037, -2.53335622, -2.62430808, -2.71073998, -2.79243591, -2.86917989, -2.94075593, -3.00694803, -3.0675402, -3.12231646, -3.17100105, -3.21331965, -3.24917952, -3.27836855, -3.30067463, -3.31588563, -3.32378945, -3.32417395, -3.31682703, -3.30153657, -3.27809046, -3.24627657, -3.20588279, -3.156697, -3.09850709, -3.03110093, -2.95426642, -2.6627627, -2.26320451, -1.86467091, -1.47360336, -1.09644332, -0.739632243, -0.409611598, -0.112822842, 0.144292565, 0.355293162, 0.513737488, 0.613184085, 0.647191491, 0.609318245, 0.493122889, 0.29216396, 0, -0.192567981, -0.577152135, -1.14735904, -1.89679526, -2.81906738, -3.90778197, -5.1565456, -6.55896486, -8.10864631, -9.79919653, -11.6242221, -13.5773296, -15.6521256, -17.8422166, -20.1412093, -22.5427102, -24.9049322, -27.20214, -29.7341876, -32.4656689, -35.3611773, -38.3853066, -41.5026503, -44.6778022, -47.8753559, -51.059905, -54.1960431, -57.248364, -60.1814612, -63.1808304, -65.9091659, -68.2647906, -70.1460276, -72.2407153, -74.052419, -74.8741667, -74.3674268, -72.1936677, -68.014358, -61.490966, -54.7029088, -44.1578088, -29.0440576, -8.5500467, 18.1358324, 46.1639145, 83.4224001, 132.138216, 194.538289, 272.849546, 368.14443, 485.215387, 620.21966, 775.393903, 965.386914, 1199.27469, 1486.13323, 1835.02124, 2255.03595, 2755.29344, 3334.59337, 4024.09181, 4833.72269, 5785.34022, 6920.99682, 8273.47361, 9869.33528, 11739.7513, 13938.6846, 16507.205, 19381.4428, 22723.5515, 26610.9261, 31120.9614, 36331.0523, 44781.1339, 53470.4258, 61860.6877, 69413.6794, 75591.1606, 78046.0916, 79254.898, 79454.5396, 79438.3794, 79223.1339, 78593.6849, 77578.3407, 76205.4095, 74503.1995, 72500.0191, 65583.1722, 61461.2254, 59443.7351, 58840.2579, 58960.3501, 59113.5683, 58609.4689, 57637.0429, 54990.7007},
        {4.83189122, 4.6894444, 4.54740742, 4.40577015, 4.26452245, 4.12365417, 3.98315516, 3.84301529, 3.7032244, 3.56377237, 3.42464903, 3.28584425, 3.14734788, 3.00914979, 2.87123982, 2.73360784, 2.59624369, 2.45913724, 2.32227834, 2.18565685, 2.04926263, 1.91308552, 1.77711539, 1.6413421, 1.50575549, 1.37034543, 1.23510177, 1.10001437, 0.965073081, 0.830267765, 0.695588275, 0.561024468, 0.4265662, 0.292203327, 0.153506723, 0.0127686823, -0.127665387, -0.267662223, -0.407088566, -0.545811155, -0.68369673, -0.820612028, -0.95642379, -1.09099876, -1.22420366, -1.35590525, -1.48597026, -1.61426543, -1.74065749, -1.8650132, -1.98719928, -2.11301783, -2.23627156, -2.35669855, -2.47403688, -2.58802461, -2.69839983, -2.8049006, -2.90726499, -3.00523108, -3.09853694, -3.18692064, -3.27012026, -3.34787386, -3.41991953, -3.48599533, -3.54583933, -3.5991978, -3.64581679, -3.6854175, -3.71773746, -3.74251423, -3.75948534, -3.76838833, -3.76896075, -3.76094014, -3.74406403, -3.71806998, -3.68269553, -3.63767821, -3.58275557, -3.51766515, -3.44214449, -3.35593114, -3.043965, -2.61790042, -2.19197342, -1.77289216, -1.36736476, -0.982099366, -0.623804119, -0.299187154, -0.0149566098, 0.222179375, 0.405512663, 0.528335115, 0.583938593, 0.565614958, 0.466656074, 0.2803538, 0, -0.0471126859, -0.320291648, -0.810707737, -1.5095318, -2.4079347, -3.49708726, -4.76816036, -6.21232483, -7.82075153, -9.58461131, -11.495075, -13.5433135, -15.7204976, -18.0177982, -20.4263861, -22.9374322, -25.4206393, -27.8498255, -30.4909111, -33.3090379, -36.269348, -39.3369832, -42.4770857, -45.6547974, -48.8352601, -51.9836161, -55.0650071, -58.0445752, -60.8874624, -63.7661487, -66.362414, -68.579199, -70.3194441, -72.317569, -74.0607747, -74.8145344, -74.232345, -71.9677031, -67.6741058, -61.0050496, -54.0228573, -43.239079, -27.8359195, -6.99558358, 20.0997239, 49.3618785, 87.741075, 137.281588, 200.027694, 278.023666, 371.511201, 486.540306, 620.7643, 776.807768, 967.444971, 1200.89317, 1485.36963, 1827.72637, 2237.85619, 2727.14121, 3296.18479, 3974.20055, 4770.13529, 5705.11422, 6820.8982, 8149.77598, 9720.32569, 11563.874, 13725.3535, 16242.0002, 19051.5587, 22306.0635, 26077.1904, 30436.6153, 35456.0138, 43541.3908, 51846.5109, 59859.4675, 67068.3543, 72961.2648, 75292.6151, 76439.9611, 76634.4011, 76655.5642, 76517.4216, 76005.3902, 75144.8691, 73961.2574, 72479.9541, 70726.3584, 65587.797, 62211.1191, 60207.382, 59187.6432, 58762.9604, 58544.3911, 58142.9926, 56943.4755, 54037.8156},
        {4.76909343, 4.61904239, 4.46972981, 4.32113232, 4.17322659, 4.02598926, 3.87939701, 3.73342646, 3.58805429, 3.44325714, 3.29901166, 3.15529451, 3.01208235, 2.86935182, 2.72707959, 2.58524229, 2.4438166, 2.30277915, 2.1621066, 2.02177562, 1.88176284, 1.74204492, 1.60259853, 1.4634003, 1.3244269, 1.18565497, 1.04706117, 0.908622161, 0.770314586, 0.632115101, 0.49400036, 0.355947017, 0.217931724, 0.0799311358, -0.0629524227, -0.208202609, -0.353244496, -0.497943248, -0.642164031, -0.78577201, -0.92863235, -1.07061022, -1.21157078, -1.35137919, -1.48990063, -1.62700026, -1.76254324, -1.89639474, -2.02841992, -2.15848395, -2.286452, -2.41890186, -2.54884055, -2.67598771, -2.800063, -2.92078607, -3.03787655, -3.15105411, -3.26003839, -3.36454904, -3.4643057, -3.55902803, -3.64843567, -3.73224827, -3.81018549, -3.88196696, -3.94731234, -4.00706707, -4.06204954, -4.10965756, -4.1495378, -4.18133692, -4.20470159, -4.21927848, -4.22471426, -4.2206556, -4.20674916, -4.18264162, -4.14797964, -4.10240989, -4.04557903, -3.97713375, -3.8967207, -3.80398655, -3.47062387, -3.01505108, -2.55842644, -2.10779419, -1.67019852, -1.25268367, -0.862293839, -0.506073259, -0.191066141, 0.0756832948, 0.287130832, 0.436232254, 0.515943341, 0.519219877, 0.439017644, 0.268292424, 0, 0.00242263404, -0.238997879, -0.714067997, -1.41259418, -2.32438289, -3.43924057, -4.7469737, -6.23738872, -7.9002921, -9.7254903, -11.7027898, -13.821997, -16.0729184, -18.4453604, -20.9291296, -23.5140323, -26.072317, -28.5769963, -31.2837814, -34.157288, -37.1621314, -40.2629272, -43.4242907, -46.6108375, -49.787183, -52.9179426, -55.9677318, -58.9011661, -61.6828609, -64.4857153, -66.9896904, -69.0969168, -70.7095247, -72.5569245, -74.1276066, -74.6894535, -73.8964116, -71.4024274, -66.8614473, -59.9274177, -52.6634913, -41.5811716, -25.8630385, -4.69167207, 22.7503478, 53.3150379, 92.7765722, 142.943722, 205.625257, 282.629949, 373.098777, 484.768938, 616.621214, 771.731068, 961.533937, 1193.7118, 1475.94664, 1814.41338, 2218.64357, 2699.8128, 3258.56459, 3922.83034, 4698.28245, 5605.40404, 6689.77418, 7985.4525, 9528.86343, 11354.6797, 13488.903, 15962.4399, 18682.1663, 21822.101, 25456.0841, 29657.9559, 34501.5566, 42326.4679, 50374.353, 58152.617, 65168.6646, 70929.9007, 73297.7569, 74522.9268, 74818.2324, 74844.2254, 74636.5016, 74044.1024, 73111.9512, 71884.9715, 70408.0864, 68726.2195, 64806.696, 62148.7861, 60459.4632, 59445.701, 58814.4732, 58272.7533, 57527.515, 56144.4084, 52869.7956},
        {4.73115839, 4.57703635, 4.42371908, 4.27117714, 4.11938112, 3.96830161, 3.81790917, 3.66817439, 3.51906784, 3.3705601, 3.22262176, 3.07522339, 2.92833557, 2.78192887, 2.63597389, 2.49044118, 2.34530135, 2.20052495, 2.05608258, 1.91194481, 1.76808221, 1.62446538, 1.48106488, 1.33785129, 1.1947952, 1.05186719, 0.909037818, 0.766277681, 0.623557352, 0.480847411, 0.338118435, 0.195341003, 0.0524856951, -0.0904769111, -0.238808556, -0.389812446, -0.540728901, -0.691417542, -0.841737991, -0.991549873, -1.14071281, -1.28908642, -1.43653033, -1.58290416, -1.72806754, -1.87188008, -2.01420142, -2.15489116, -2.29380894, -2.43081437, -2.56576708, -2.70559653, -2.84293924, -2.97750157, -3.1089899, -3.23711057, -3.36156996, -3.48207443, -3.59833033, -3.71004404, -3.81692191, -3.9186703, -4.01499558, -4.10560411, -4.19020226, -4.26849638, -4.34019284, -4.40751596, -4.47262961, -4.53004705, -4.57931141, -4.61996582, -4.65155341, -4.67361733, -4.68570069, -4.68734665, -4.67809832, -4.65749884, -4.62509134, -4.58041896, -4.52302484, -4.45245209, -4.36824387, -4.26994329, -3.91485229, -3.42872983, -2.9402464, -2.45680634, -1.98581396, -1.5346736, -1.11078959, -0.721566242, -0.37440789, -0.0767188574, 0.164096528, 0.340633941, 0.445489055, 0.471257544, 0.410535082, 0.255917342, 0, 0.00841199506, -0.237085284, -0.725450306, -1.44564154, -2.38661745, -3.5373365, -4.88675717, -6.42383793, -8.13753723, -10.0168136, -12.0506254, -14.2279311, -16.5376893, -18.9688584, -21.5103968, -24.1512631, -26.7618047, -29.3140356, -32.0652011, -34.978843, -38.0185031, -41.1477233, -44.3300453, -47.529011, -50.7081623, -53.831041, -56.8611889, -59.7621478, -62.4974596, -65.2447304, -67.6749475, -69.6874335, -71.1815107, -72.8597305, -74.219801, -74.5452361, -73.4943897, -70.7256157, -65.897268, -58.6677005, -51.1232201, -39.7552855, -23.7472165, -2.2823326, 25.4560464, 57.2788734, 97.7599882, 148.467503, 210.96953, 286.834181, 374.077587, 482.161153, 611.330799, 765.112604, 953.729646, 1184.39819, 1464.33449, 1799.31893, 2198.32999, 2671.91259, 3220.52631, 3870.78903, 4624.46519, 5501.35232, 6551.80398, 7812.14801, 9327.64345, 11136.9336, 13245.9141, 15679.0044, 18305.9453, 21329.6319, 24828.0263, 28879.0906, 33560.7871, 41165.7257, 49003.559, 56598.5869, 63475.1098, 69157.4278, 71614.9247, 72963.4282, 73393.631, 73414.5966, 73090.9963, 72354.8471, 71277.4529, 69930.1179, 68384.1461, 66710.8415, 63795.659, 61824.6573, 60533.1595, 59656.4885, 58929.9676, 58088.9198, 56868.6681, 55311.5587, 51575.592},
        {4.7216673, 4.56611526, 4.4112137, 4.25693378, 4.10324664, 3.95012341, 3.79753525, 3.64545328, 3.49384866, 3.34269252, 3.19195602, 3.04161028, 2.89162645, 2.74197568, 2.5926291, 2.44355787, 2.29473311, 2.14612597, 1.9977076, 1.84944914, 1.70132172, 1.5532965, 1.4053446, 1.25743719, 1.10954538, 0.961640341, 0.813693199, 0.665675099, 0.517557185, 0.369310597, 0.220906478, 0.0723159705, -0.0764897841, -0.225539643, -0.380259405, -0.53786594, -0.695511252, -0.853049058, -1.01033307, -1.16721702, -1.32355461, -1.47919956, -1.63400558, -1.78782641, -1.94051574, -2.0919273, -2.24191481, -2.39033198, -2.53703254, -2.68187018, -2.82469864, -2.97281633, -3.11847089, -3.26135463, -3.40115989, -3.53757901, -3.6703043, -3.79902811, -3.92344276, -4.04324058, -4.15811391, -4.26775507, -4.3718564, -4.47011021, -4.56220886, -4.64784465, -4.72670994, -4.8024385, -4.8785695, -4.94666629, -5.00616571, -5.05650457, -5.09711968, -5.12744788, -5.14692599, -5.15499082, -5.1510792, -5.13462794, -5.10507388, -5.06185383, -5.00440461, -4.93216304, -4.84456595, -4.74105016, -4.36413812, -3.84712768, -3.3264046, -2.80974309, -2.30491732, -1.81970147, -1.36186975, -0.939196326, -0.559455394, -0.23042114, 0.0401322509, 0.244430591, 0.374699694, 0.423165373, 0.382053442, 0.243589713, 0, -0.021742335, -0.300877879, -0.82595659, -1.58552842, -2.56814334, -3.76235129, -5.15670224, -6.73974615, -8.50003296, -10.4261126, -12.5065351, -14.7298504, -17.0846084, -19.5593592, -22.1426525, -24.8230385, -27.4652079, -30.0400165, -32.8161718, -35.7556827, -38.8205577, -41.9728057, -45.1744353, -48.3874552, -51.5738742, -54.6957009, -57.714944, -60.5936123, -63.2937144, -65.9978418, -68.3665402, -70.2956434, -71.6809852, -73.188479, -74.328919, -74.4077448, -73.0897663, -70.0397937, -64.9226369, -57.4031059, -49.5973099, -37.9727122, -21.714521, -0.00794409195, 27.9618103, 60.8204251, 102.119079, 153.21984, 215.484779, 290.275965, 374.664304, 479.547782, 606.268757, 758.730019, 946.061911, 1175.04771, 1452.47069, 1783.68306, 2177.22441, 2643.19554, 3182.17689, 3819.15013, 4551.98833, 5399.66773, 6416.92276, 7642.07424, 9128.07322, 10917.0335, 12997.425, 15388.0614, 17925.5364, 20838.9182, 24210.0968, 28120.9625, 32653.4056, 40056.1579, 47702.5575, 55132.784, 61887.0171, 67505.4366, 70062.8653, 71541.7444, 72110.2637, 72121.7325, 71671.0641, 70776.0389, 69535.4266, 68047.9975, 66412.5214, 64727.7684, 62642.0351, 61302.6603, 60435.5552, 59766.6309, 59021.799, 57926.9705, 56208.0567, 54465.7628, 50178.3322},
        {4.74779677, 4.59380542, 4.4400053, 4.28637892, 4.13290878, 3.97957738, 3.82636725, 3.67326088, 3.52024079, 3.36728948, 3.21438945, 3.06152322, 2.9086733, 2.75582218, 2.60295238, 2.45004641, 2.29708677, 2.14405597, 1.99093652, 1.83771092, 1.68436169, 1.53087133, 1.37722234, 1.22339724, 1.06937853, 0.915148727, 0.760690328, 0.605985843, 0.451017781, 0.295768649, 0.140220955, -0.0156427946, -0.171840092, -0.328388429, -0.490379176, -0.655186165, -0.820122243, -0.985040245, -1.14979301, -1.31423337, -1.47821416, -1.64158821, -1.80420837, -1.96592746, -2.12659833, -2.28607381, -2.44420673, -2.60084993, -2.75585624, -2.90907851, -3.06036956, -3.2181211, -3.373462, -3.52605997, -3.67558276, -3.82169808, -3.96407367, -4.10237726, -4.23627657, -4.36543934, -4.4895333, -4.60822617, -4.72118569, -4.82807958, -4.92857558, -5.0223414, -5.10904479, -5.19329282, -5.27957398, -5.35736895, -5.42602528, -5.48489049, -5.53331211, -5.57063768, -5.59621473, -5.60939078, -5.60951338, -5.59593005, -5.56798832, -5.52503573, -5.46641981, -5.39148808, -5.29958809, -5.19006736, -4.79158019, -4.24451442, -3.69251135, -3.14370305, -2.60622156, -2.08819893, -1.5977672, -1.14305843, -0.732204646, -0.373337912, -0.0745902699, 0.155906232, 0.310019548, 0.379617629, 0.35656843, 0.232739902, 0, -0.101169223, -0.45393022, -1.04704901, -1.86929163, -2.90942409, -4.15621242, -5.59842265, -7.22482079, -9.02417288, -10.9852449, -13.096803, -15.3476131, -17.7264412, -20.2220533, -22.8232156, -25.518694, -28.1600507, -30.7164511, -33.4870033, -36.4310726, -39.5080241, -42.6772231, -45.8980347, -49.1298241, -52.3319566, -55.4637973, -58.4847113, -61.3540641, -64.0312206, -66.690227, -68.9970513, -70.8466544, -72.1339972, -73.5068483, -74.4837649, -74.3825551, -72.8723377, -69.6222311, -64.3013541, -56.5788252, -48.5608057, -36.7376628, -20.3017025, 1.55476927, 29.6394466, 62.7757769, 104.251763, 155.37116, 217.437724, 291.755211, 375.200366, 478.914746, 604.845692, 756.991885, 943.540148, 1170.79692, 1445.06862, 1770.46426, 2155.98713, 2613.03768, 3144.12057, 3771.22298, 4489.64312, 5317.03193, 6308.9157, 7503.43533, 8954.77826, 10705.2456, 12738.3011, 15070.6901, 17533.6076, 20357.8958, 23624.5293, 27414.4827, 31808.7305, 38981.3672, 46396.9674, 53613.2256, 60187.8363, 65678.4941, 68255.4016, 69788.7399, 70430.8427, 70428.7885, 69897.4081, 68912.0555, 67589.5352, 66046.652, 64400.2106, 62767.0155, 61291.6421, 60460.8884, 59979.9098, 59553.8616, 58887.8989, 57687.1771, 55656.8514, 53677.5066, 48713.4785},
        {4.81153752, 4.66155272, 4.51102221, 4.35995025, 4.2083411, 4.05619904, 3.90352831, 3.7503332, 3.59661797, 3.44238687, 3.28764418, 3.13239417, 2.97664109, 2.82038922, 2.66364281, 2.50640613, 2.34868346, 2.19047905, 2.03179717, 1.87264209, 1.71301806, 1.55292936, 1.39238026, 1.23137501, 1.06991788, 0.908013142, 0.745665056, 0.582877888, 0.419655904, 0.256003369, 0.0919245486, -0.0725762915, -0.237494886, -0.40282697, -0.572785359, -0.745168547, -0.91772434, -1.09031162, -1.26278927, -1.43501618, -1.60685121, -1.77815327, -1.94878122, -2.11859396, -2.28745035, -2.45520929, -2.62172966, -2.78687034, -2.95049022, -3.11244816, -3.27260306, -3.44141714, -3.60791608, -3.7717289, -3.93248463, -4.08981229, -4.2433409, -4.39269948, -4.53751706, -4.67742266, -4.8120453, -4.94101401, -5.0639578, -5.18050571, -5.29028674, -5.39292993, -5.4880643, -5.58067041, -5.67560034, -5.76144596, -5.83748937, -5.90301268, -5.95729798, -5.99962739, -6.029283, -6.04554692, -6.04770126, -6.03502812, -6.0068096, -5.96232781, -5.90086485, -5.82170282, -5.72412383, -5.60740998, -5.18777985, -4.61200116, -4.03024386, -3.450976, -2.88266563, -2.33378083, -1.81278965, -1.32816014, -0.888360372, -0.501858404, -0.177122293, 0.0773798992, 0.253180114, 0.341810291, 0.334802371, 0.223688294, 0, -0.226166108, -0.689391974, -1.37924289, -2.28528414, -3.39708103, -4.70419884, -6.19620287, -7.86265841, -9.69313074, -11.6771852, -13.804387, -16.0643015, -18.4464939, -20.9405296, -23.5359739, -26.222392, -28.8313538, -31.3298537, -34.0648278, -36.9919852, -40.067035, -43.2456864, -46.4836484, -49.7366302, -52.9603409, -56.1104896, -59.1427854, -62.0129374, -64.6766548, -67.2833687, -69.5240348, -71.2962455, -72.4975935, -73.7847172, -74.6756862, -74.4861576, -72.8860102, -69.5451226, -64.1333735, -56.3206416, -48.1479245, -36.1923372, -19.6598437, 2.24359189, 30.3120054, 62.8130019, 103.698656, 154.394611, 216.32651, 290.919998, 375.77561, 480.819229, 607.998165, 761.062757, 947.444517, 1172.92218, 1443.2745, 1760.38955, 2134.82096, 2581.36671, 3106.58116, 3727.92035, 4439.89603, 5258.26604, 6234.78861, 7404.6972, 8815.43417, 10505.3983, 12468.2222, 14722.861, 17130.0051, 19891.2222, 23080.251, 26770.8304, 31036.6991, 37935.5646, 45061.2048, 51991.366, 58303.7949, 63576.238, 66061.265, 67545.2505, 68173.874, 68157.9868, 67617.4713, 66645.0545, 65361.9765, 63889.4772, 62348.7965, 60861.1744, 59787.0459, 59320.1136, 59150.3623, 58967.7765, 58462.3409, 57324.0399, 55242.8583, 52957.1475, 47195.0829},
        {4.9048938, 4.75909395, 4.61197621, 4.46356776, 4.31389582, 4.16298757, 4.01087022, 3.85757096, 3.703117, 3.54753553, 3.39085375, 3.23309887, 3.07429808, 2.91447858, 2.75366757, 2.59189225, 2.42917981, 2.26555747, 2.10105241, 1.93569183, 1.76950294, 1.60251294, 1.43474901, 1.26623837, 1.09700822, 0.927085738, 0.75649814, 0.58527262, 0.413436379, 0.241016615, 0.0680405279, -0.105464684, -0.279471821, -0.453953684, -0.632141652, -0.812310697, -0.992696625, -1.17316649, -1.35358735, -1.53382625, -1.71375025, -1.8932264, -2.07212175, -2.25030337, -2.4276383, -2.60399359, -2.77923631, -2.9532335, -3.12585222, -3.29695951, -3.46642245, -3.64712935, -3.82564373, -4.00155034, -4.17443398, -4.34387942, -4.50947144, -4.67079481, -4.82743432, -4.97897474, -5.12500085, -5.26509742, -5.39884924, -5.52584109, -5.64565773, -5.75788396, -5.86210454, -5.96349388, -6.0670922, -6.16095606, -6.24430791, -6.31637016, -6.37636523, -6.42351556, -6.45704356, -6.47617166, -6.4801223, -6.46811788, -6.43938084, -6.3931336, -6.32859858, -6.24499822, -6.14155493, -6.01749115, -5.57704283, -4.97313029, -4.36221931, -3.75310521, -3.15458328, -2.57544884, -2.02449717, -1.51052359, -1.04232339, -0.62869188, -0.278424362, -0.00031613713, 0.196837492, 0.304241222, 0.313099752, 0.214617779, 0, -0.360618201, -0.941862225, -1.73420233, -2.72810876, -3.91405179, -5.28250168, -6.82392866, -8.52880302, -10.387595, -12.3907748, -14.5288128, -16.7921792, -19.1713442, -21.6567782, -24.2389513, -26.9083338, -29.4785749, -31.9148847, -34.607686, -37.5095586, -40.5730827, -43.7508383, -46.9954054, -50.259364, -53.4952942, -56.6557759, -59.6933893, -62.5607143, -65.2103309, -67.7701487, -69.9531308, -71.6592582, -72.7885123, -74.0039044, -74.8239006, -74.5634923, -72.8937514, -69.4857502, -64.0105606, -56.1392548, -47.8365529, -35.7818122, -19.1973476, 2.694526, 30.6714938, 62.2023314, 102.200861, 152.271994, 214.020644, 289.051724, 375.995366, 483.067103, 611.898482, 765.860207, 951.873757, 1175.26638, 1441.36534, 1749.88872, 2113.04775, 2549.17273, 3069.04974, 3685.53999, 4393.03385, 5205.3514, 6169.23525, 7316.31604, 8685.0762, 10308.9228, 12196.1395, 14369.2212, 16728.3988, 19437.241, 22560.565, 26163.1877, 30309.9263, 36933.3895, 43763.6546, 50398.591, 56436.0677, 61473.9538, 63840.0627, 65249.0233, 65843.0144, 65823.0034, 65307.8987, 64392.1133, 63195.7803, 61839.0326, 60442.0034, 59124.8256, 58357.1205, 58175.8634, 58260.113, 58288.9277, 57941.3662, 56896.4871, 54833.3489, 52199.2538, 45595.9671},
        {5.0255877, 4.88287004, 4.73814632, 4.59146368, 4.44286926, 4.29241018, 4.14013359, 3.98608661, 3.83031639, 3.67287005, 3.51379473, 3.35313758, 3.19094571, 3.02726627, 2.86214639, 2.6956332, 2.52777385, 2.35861546, 2.18820516, 2.01659011, 1.84381742, 1.66993423, 1.49498769, 1.31902491, 1.14209305, 0.964239222, 0.785510574, 0.605954238, 0.425617349, 0.244547041, 0.0627904479, -0.119605294, -0.302593052, -0.486125689, -0.672512738, -0.860479161, -1.0487216, -1.23711643, -1.42554005, -1.61386885, -1.8019792, -1.9897475, -2.17705014, -2.36376349, -2.54976395, -2.73492791, -2.91913175, -3.10225186, -3.28416463, -3.46474644, -3.64387367, -3.83709549, -4.02827575, -4.2169511, -4.40265816, -4.58493356, -4.76331394, -4.93733593, -5.10653617, -5.27045128, -5.42861789, -5.58057264, -5.72585217, -5.8639931, -5.99453206, -6.11700569, -6.23095062, -6.34164137, -6.4542147, -6.55637327, -6.64728176, -6.72610482, -6.79200714, -6.84415337, -6.88170819, -6.90383626, -6.90970226, -6.89847084, -6.86930668, -6.82137445, -6.75383881, -6.66586444, -6.556616, -6.42525815, -5.96437717, -5.33287538, -4.69333997, -4.05488464, -3.42662315, -2.8176692, -2.23713653, -1.69413886, -1.19778992, -0.757203444, -0.381493151, -0.0797727705, 0.138843971, 0.265243345, 0.290311627, 0.204935087, 0, -0.486968585, -1.17946364, -2.0687383, -3.14604569, -4.40263892, -5.82977112, -7.41869542, -9.16066495, -11.0469328, -13.0687522, -15.2173761, -17.4840577, -19.8600502, -22.3366067, -24.9049803, -27.5564241, -30.0929059, -32.4786383, -35.1330194, -38.0067413, -41.0504961, -44.2149756, -47.4508719, -50.7088771, -53.939683, -57.0939817, -60.1224652, -62.9758255, -65.6047546, -68.1254711, -70.2624472, -71.9167172, -72.9893149, -74.1393923, -74.8889244, -74.5577217, -72.8193827, -69.3475063, -63.815691, -55.8975355, -47.4803645, -35.350308, -18.7478443, 3.08654805, 30.9123908, 61.2966688, 100.239575, 149.548764, 211.031893, 286.496617, 375.742372, 485.039525, 615.446759, 769.878565, 955.038386, 1175.9285, 1437.5512, 1737.85198, 2090.49372, 2516.83755, 3031.43695, 3643.02201, 4346.31693, 5153.34642, 6105.44989, 7230.51081, 8557.42806, 10114.3484, 11925.6954, 14017.9987, 16334.5436, 18997.9109, 22063.4455, 25586.4924, 29622.3964, 35977.7846, 42518.8994, 48863.2652, 54628.4063, 59431.8471, 61670.3472, 62996.0392, 63549.6324, 63541.6171, 63085.0812, 62264.0369, 61193.0257, 59986.5892, 58759.2689, 57625.6062, 57124.068, 57166.352, 57433.9554, 57608.3757, 57371.11, 56403.6556, 54387.5098, 51364.69, 43912.5595},
        {5.16508465, 5.01939215, 4.87156923, 4.72166315, 4.56972115, 4.41579046, 4.25991834, 4.10215203, 3.94253876, 3.7811258, 3.61796037, 3.45308972, 3.2865611, 3.11842175, 2.94871891, 2.77749983, 2.60481175, 2.43070191, 2.25521756, 2.07840594, 1.9003143, 1.72098987, 1.54047991, 1.35883165, 1.17609234, 0.992309223, 0.807529544, 0.621800544, 0.435169468, 0.247683559, 0.059390061, -0.129663782, -0.319430728, -0.509863531, -0.702991107, -0.897684465, -1.09278966, -1.28819207, -1.48377707, -1.67943006, -1.8750364, -2.07048148, -2.26565069, -2.4604294, -2.65470299, -2.84835685, -3.04127635, -3.23334688, -3.42445381, -3.61448254, -3.80331843, -4.00930279, -4.213465, -4.41529035, -4.61426412, -4.80987162, -5.00159813, -5.18892895, -5.37134937, -5.54834468, -5.71940018, -5.88400117, -6.04163292, -6.19178074, -6.33392992, -6.46756575, -6.59217352, -6.71317661, -6.83585582, -6.94744959, -7.04705828, -7.13378224, -7.20672182, -7.26497738, -7.30764926, -7.33383782, -7.34264341, -7.33316638, -7.30450708, -7.25576586, -7.18604307, -7.09443908, -6.98005422, -6.84198885, -6.36137809, -5.70299289, -5.03542587, -4.36809597, -3.71042211, -3.07182325, -2.4617183, -1.8895262, -1.3646659, -0.896556317, -0.494616393, -0.168265062, 0.0730787421, 0.219996085, 0.263068032, 0.192875648, 0, -0.542175888, -1.28790587, -2.22824882, -3.35426357, -4.657009, -6.12754396, -7.75692731, -9.5362179, -11.4564746, -13.5087563, -15.6841217, -17.9736299, -20.3683396, -22.8593097, -25.437599, -28.0942665, -30.6438684, -33.052249, -35.709699, -38.5694553, -41.5847547, -44.7088343, -47.8949308, -51.0962812, -54.2661223, -57.357691, -60.3242243, -63.118959, -65.6951321, -68.1972086, -70.3170801, -71.9506151, -72.993682, -74.0553577, -74.6749064, -74.2014031, -72.3167538, -68.7028642, -63.0416401, -55.0149873, -46.4649483, -34.2558595, -17.6469911, 4.10238628, 31.7330023, 61.1236189, 99.0924253, 147.595643, 208.589493, 284.030198, 374.665256, 485.136939, 615.611462, 768.562437, 951.174412, 1168.51682, 1425.65911, 1720.57245, 2067.03778, 2486.57952, 2994.80602, 3598.54016, 4292.56863, 5087.85801, 6022.79084, 7123.16531, 8412.73634, 9916.77238, 11668.0337, 13695.0429, 15968.7166, 18584.0456, 21589.4926, 25033.5207, 28964.5928, 35080.1803, 41367.4855, 47462.7192, 53002.0921, 57621.815, 59790.1963, 61089.9508, 61657.8188, 61708.3037, 61347.3647, 60648.5597, 59719.468, 58667.6689, 57600.7416, 56626.2655, 56581.8693, 56844.368, 57157.0307, 57263.1267, 56905.9252, 55828.6954, 53774.7066, 50310.9546, 42095.0347},
        {5.322536, 5.16736953, 5.01054902, 4.8521007, 4.69205081, 4.5304256, 4.36725131, 4.20255418, 4.03636046, 3.86869638, 3.6995882, 3.52906215, 3.35714447, 3.18386142, 3.00923922, 2.83330413, 2.65608238, 2.47760022, 2.29788389, 2.11695963, 1.9348537, 1.75159232, 1.56720174, 1.3817082, 1.19513795, 1.00751723, 0.818872286, 0.629229352, 0.438614675, 0.247054497, 0.0545750607, -0.138797391, -0.333036615, -0.52811637, -0.726492905, -0.926846515, -1.12782766, -1.32932953, -1.53124533, -1.73346825, -1.93589148, -2.13840823, -2.34091168, -2.54329504, -2.7454515, -2.94727425, -3.1486565, -3.34949143, -3.54967224, -3.74909213, -3.9476443, -4.16646861, -4.38375101, -4.59892409, -4.81142047, -5.02067275, -5.22611356, -5.42717549, -5.62329116, -5.81389317, -5.99841414, -6.17628668, -6.34694339, -6.50981688, -6.66433977, -6.80994466, -6.94606416, -7.07839223, -7.21247247, -7.33482396, -7.4444734, -7.54044751, -7.62177301, -7.6874766, -7.736585, -7.76812491, -7.78112306, -7.77460615, -7.74760089, -7.69913401, -7.6282322, -7.53392219, -7.41523068, -7.27118439, -6.77154886, -6.08690696, -5.39179901, -4.69593688, -4.00903244, -3.34079756, -2.70094412, -2.09918398, -1.54522902, -1.04879109, -0.619582087, -0.267313866, -0.00169830007, 0.16755274, 0.230727384, 0.17811376, 0, -0.519073547, -1.25413781, -2.19499574, -3.33145028, -4.65330438, -6.15036099, -7.81242305, -9.62929351, -11.5907753, -13.6866714, -15.9067848, -18.2409183, -20.678875, -23.2104577, -25.8254695, -28.5137133, -31.1276044, -33.6373191, -36.3434178, -39.2062633, -42.1862183, -45.2436457, -48.3389081, -51.4323684, -54.4843892, -57.4553334, -60.3055636, -62.9954427, -65.4853334, -67.989284, -70.1209536, -71.7645995, -72.8044786, -73.7520953, -74.1780829, -73.4858094, -71.3713575, -67.5308105, -61.6602511, -53.4557625, -44.7509062, -32.4537497, -15.8421736, 5.80594136, 33.2127145, 61.8819578, 99.0586748, 146.769562, 207.041314, 281.900629, 372.700236, 482.964534, 611.733971, 761.102825, 939.40028, 1152.15617, 1404.90035, 1697.53361, 2042.49415, 2458.36092, 2958.80425, 3551.20962, 4230.06286, 5006.10367, 5917.61304, 6990.23007, 8247.80345, 9715.5081, 11425.0837, 13404.5566, 15633.3555, 18195.5807, 21136.0808, 24499.7042, 28331.2993, 34241.5806, 40318.0642, 46214.132, 51583.1661, 56078.5485, 58241.7209, 59579.2991, 60221.2108, 60377.6339, 60146.7306, 59593.2255, 58816.9528, 57917.7465, 56995.4408, 56149.8697, 56767.7982, 57253.4212, 57471.8155, 57288.0581, 56567.2258, 55174.3957, 52974.6445, 49022.5245, 40150.3872},
        {5.5060844, 5.33978131, 5.17242008, 5.00400175, 4.83452734, 4.6639979, 4.49241445, 4.31977803, 4.14608968, 3.97135041, 3.79556128, 3.61872331, 3.44083754, 3.261905, 3.08192672, 2.90090374, 2.71883709, 2.5357278, 2.35157691, 2.16638545, 1.98015445, 1.79288496, 1.60457799, 1.41523459, 1.22485578, 1.03344261, 0.840996103, 0.647517295, 0.453007219, 0.257466908, 0.0608973967, -0.136700282, -0.335325095, -0.534976009, -0.738708897, -0.944929976, -1.15200872, -1.35984488, -1.56833824, -1.77738855, -1.98689558, -2.1967591, -2.40687886, -2.61715462, -2.82748617, -3.03777325, -3.24791563, -3.45781308, -3.66736536, -3.87647223, -4.08503346, -4.31684378, -4.54738998, -4.77605384, -5.00221713, -5.22526161, -5.44456906, -5.65952123, -5.8694999, -6.07388684, -6.27206381, -6.46341258, -6.64731492, -6.8231526, -6.99030739, -7.14816105, -7.29609535, -7.44018827, -7.58635702, -7.72017293, -7.84058368, -7.94653695, -8.03698042, -8.11086178, -8.16712869, -8.20472885, -8.22260993, -8.21971961, -8.19500557, -8.1474155, -8.07589706, -7.97939794, -7.85686583, -7.70724839, -7.18892635, -6.47829424, -5.75583993, -5.03156316, -4.31546367, -3.61754119, -2.94779545, -2.31622618, -1.73283312, -1.20761599, -0.750574542, -0.371708497, -0.0810175893, 0.111498447, 0.195839879, 0.162006975, 0, -0.475859949, -1.18354276, -2.1113507, -3.24758605, -4.58055109, -6.0985481, -7.78987935, -9.64284712, -11.6457537, -13.7869013, -16.0545923, -18.437129, -20.9228135, -23.4999482, -26.1568354, -28.8817773, -31.5713978, -34.2009021, -36.9659606, -39.8352301, -42.7773676, -45.7610299, -48.754874, -51.7275566, -54.6477348, -57.4840653, -60.205205, -62.7798109, -65.1765398, -67.6881744, -69.8403997, -71.5046348, -72.5522989, -73.3891362, -73.6194035, -72.7048704, -70.3566588, -66.2858901, -60.2036859, -51.821168, -42.9715162, -30.59795, -13.9964056, 7.53718059, 34.7068724, 62.8008592, 99.2930476, 146.24147, 205.704158, 279.739146, 370.119963, 479.536866, 605.931605, 751.07344, 924.546133, 1132.42489, 1380.78493, 1671.95112, 2016.60136, 2429.50481, 2921.42252, 3501.24691, 4163.44993, 4918.88357, 5806.15479, 6851.1306, 8078.79961, 9514.8011, 11187.9949, 13125.4192, 15310.0287, 17819.7268, 20696.8114, 23983.58, 27722.3304, 33450.7928, 39341.9743, 45064.3143, 50286.2525, 54676.2283, 56837.7806, 58214.1835, 58930.2624, 59206.542, 59129.967, 58747.6073, 58148.4016, 57421.2882, 56655.2056, 55939.0921, 57184.8506, 57841.2429, 57902.9376, 57364.6031, 56220.908, 54466.5207, 52096.1098, 47632.4219, 38152.0819},
        {5.71771039, 5.53982684, 5.36147451, 5.18262947, 5.00326775, 4.82336542, 4.64289854, 4.46184315, 4.28017532, 4.09787109, 3.91490653, 3.73125769, 3.54690062, 3.36181138, 3.17596602, 2.9893406, 2.80191118, 2.6136538, 2.42454453, 2.23455942, 2.04367452, 1.85186589, 1.65910959, 1.46538166, 1.27065817, 1.07491517, 0.878128718, 0.680274863, 0.481329663, 0.281269173, 0.0800694491, -0.122293455, -0.325843483, -0.530604581, -0.740281717, -0.952980898, -1.16676389, -1.38153518, -1.59719926, -1.81366061, -2.03082373, -2.2485931, -2.46687322, -2.68556856, -2.90458363, -3.1238229, -3.34319087, -3.56259202, -3.78193084, -4.00111183, -4.22003947, -4.46488677, -4.70872025, -4.95087495, -5.1906859, -5.42748814, -5.66061672, -5.88940666, -6.11319302, -6.33131082, -6.54309511, -6.74788092, -6.9450033, -7.13379727, -7.31359789, -7.48374019, -7.6435592, -7.79967781, -7.95854391, -8.10446193, -8.23629446, -8.35290409, -8.45315342, -8.53590506, -8.60002159, -8.64436562, -8.66779975, -8.66918656, -8.64738866, -8.60126865, -8.52968912, -8.43151267, -8.3056019, -8.1508194, -7.61403515, -6.87751175, -6.12774248, -5.37501277, -4.62960807, -3.9018138, -3.2019154, -2.5401983, -1.92694795, -1.37244977, -0.886989206, -0.480851686, -0.164322646, 0.0523124771, 0.158768249, 0.144759235, 0, -0.425936033, -1.10033575, -2.00995512, -3.14155011, -4.48187669, -6.01769083, -7.73574849, -9.62280564, -11.6656182, -13.8509423, -16.1655337, -18.5961485, -21.1295426, -23.752472, -26.4516927, -29.2139606, -31.9811895, -34.7337292, -37.5597647, -40.4364428, -43.3409105, -46.2503147, -49.1418023, -51.9925202, -54.7796154, -57.4802347, -60.071525, -62.5306334, -64.8347066, -67.3552445, -69.5304452, -71.2191495, -72.2801985, -73.01688, -73.0658118, -71.9450922, -69.3802283, -65.096727, -58.8200955, -50.2758407, -41.299042, -28.8635903, -12.28225, 9.13221436, 36.0670381, 63.7740075, 99.7346696, 145.995122, 204.601463, 277.599789, 366.966517, 474.902593, 598.419214, 739.044335, 907.52764, 1110.47845, 1354.50607, 1644.48283, 1989.14982, 2399.14301, 2881.78694, 3448.11374, 4093.01013, 4827.68908, 5691.02746, 6709.20398, 7908.62146, 9315.51669, 10955.3047, 12853.8653, 14994.6073, 17452.4156, 20267.8716, 23481.5569, 27134.0527, 32704.9833, 38436.1265, 44008.3032, 49102.334, 53399.0397, 55548.3508, 56949.2383, 57724.3328, 58124.4799, 58221.5585, 58034.0248, 57636.4291, 57103.3216, 56509.2523, 55928.7716, 57702.186, 58463.8772, 58329.9291, 57416.4257, 55839.451, 53715.0891, 51159.4239, 46171.8644, 36128.0286},
        {5.95939447, 5.77070546, 5.58200439, 5.393247, 5.20438903, 5.01538623, 4.82619434, 4.63676911, 4.44706629, 4.25704163, 4.06665086, 3.87584974, 3.68459401, 3.49283942, 3.30054172, 3.10765664, 2.91413995, 2.71994737, 2.52503467, 2.32935758, 2.13287185, 1.93553323, 1.73729746, 1.53812029, 1.33795747, 1.13676474, 0.93449785, 0.731112544, 0.52656457, 0.320809672, 0.113803598, -0.0944979071, -0.304139096, -0.515164224, -0.731853997, -0.952045334, -1.17352424, -1.39619747, -1.61997182, -1.84475406, -2.07045094, -2.29696926, -2.52421578, -2.75209727, -2.98052051, -3.20939226, -3.43861931, -3.66810843, -3.89776638, -4.12749994, -4.35721589, -4.61505609, -4.87208015, -5.12758273, -5.38085855, -5.63120229, -5.87790864, -6.12027229, -6.35758794, -6.58915027, -6.81425398, -7.03219376, -7.24226429, -7.44376027, -7.6359764, -7.81820735, -7.98974783, -8.1579739, -8.33006761, -8.4886564, -8.63251109, -8.76040248, -8.87110138, -8.96337859, -9.03600493, -9.08775121, -9.11738822, -9.12368679, -9.10541771, -9.0613518, -8.99025987, -8.89091271, -8.76208114, -8.60253598, -8.04739986, -7.28491654, -6.50770051, -5.72632366, -4.95135788, -4.19337506, -3.46294709, -2.77064586, -2.12704327, -1.54271119, -1.02822153, -0.594146166, -0.251056992, -0.00952589932, 0.119875224, 0.126574487, 0, -0.382702737, -1.02873184, -1.92345046, -3.05222171, -4.40040875, -5.95337471, -7.69648273, -9.61509593, -11.6945775, -13.9202905, -16.2775981, -18.7518634, -21.3284497, -23.9927199, -26.7300373, -29.525765, -32.3629204, -35.2265316, -38.1072675, -40.9899885, -43.8595551, -46.7008278, -49.4986671, -52.2379335, -54.9034875, -57.4801896, -59.9529002, -62.3064798, -64.525789, -67.051859, -69.2461169, -70.9565719, -72.0312333, -72.6857261, -72.5842516, -71.2929803, -68.5496368, -64.0919455, -57.6576309, -48.9844175, -39.9057473, -27.4258002, -10.8722692, 10.427153, 37.1447736, 64.6950872, 100.322667, 146.014272, 203.756664, 275.536603, 363.281975, 469.110371, 589.411642, 725.585558, 889.260469, 1087.4723, 1327.25698, 1615.78657, 1959.92992, 2366.40737, 2839.0236, 3391.27184, 4019.02376, 4734.01158, 5574.8424, 6567.7873, 7740.16561, 9118.52016, 10725.5503, 12586.1296, 14682.9623, 17089.5786, 19845.4489, 22990.0435, 26562.833, 32001.3183, 37597.4316, 43041.1358, 48022.3937, 52231.168, 54343.4071, 55739.098, 56542.7812, 57060.8993, 57345.9898, 57374.7977, 57203.6502, 56888.8745, 56486.7976, 56053.7468, 58188.9643, 58977.3683, 58632.3223, 57367.1897, 55395.3342, 52930.1192, 50184.9083, 44672.0695, 34106.137},
        {6.22604782, 6.03675489, 5.84680712, 5.65617936, 5.4648465, 5.27278338, 5.07996488, 4.88636586, 4.69196118, 4.49672571, 4.30063431, 4.10366185, 3.9057832, 3.7069732, 3.50720674, 3.30645868, 3.10470387, 2.90191719, 2.69807349, 2.49314765, 2.28711452, 2.07994898, 1.87162588, 1.66212009, 1.45140648, 1.2394599, 1.02625523, 0.81176733, 0.595971059, 0.378841286, 0.160352874, -0.0595193117, -0.280800408, -0.503515549, -0.731283995, -0.962258223, -1.19454751, -1.42806035, -1.66270525, -1.89839071, -2.13502522, -2.37251729, -2.61077541, -2.84970809, -3.08922382, -3.3292311, -3.56963843, -3.81035431, -4.05128725, -4.29234573, -4.53343826, -4.803815, -5.07340672, -5.34148584, -5.60732481, -5.87019603, -6.12937196, -6.38412501, -6.63372762, -6.87745221, -7.11457122, -7.34435707, -7.5660822, -7.77901904, -7.98244, -8.17561753, -8.35782406, -8.53776238, -8.72390888, -8.89608656, -9.05295651, -9.19317982, -9.31541759, -9.41833092, -9.50058091, -9.56082865, -9.59773524, -9.60996178, -9.59616936, -9.55501908, -9.48517204, -9.38528933, -9.25403205, -9.0900613, -8.51594509, -7.72630168, -6.92026966, -6.10871507, -5.30250395, -4.51250234, -3.74957629, -3.02459183, -2.348415, -1.73191184, -1.1859484, -0.72139071, -0.349104817, -0.0799567589, 0.0751874235, 0.10546169, 0, -0.377339344, -1.02376828, -1.92404995, -3.06294748, -4.42522402, -5.99564271, -7.75896667, -9.69995905, -11.803383, -14.0540016, -16.4365781, -18.9358755, -21.536657, -24.2236858, -26.9817249, -29.7955376, -32.6654493, -35.5784463, -38.4840912, -41.3704819, -44.2257164, -47.0378927, -49.7951085, -52.4854619, -55.0970508, -57.6179731, -60.0363267, -62.3402095, -64.5177196, -67.0289576, -69.2169509, -70.9281966, -72.0091918, -72.6610738, -72.5530196, -71.2596353, -68.521026, -64.0772967, -57.6685524, -49.0348982, -40.0468622, -27.6657365, -11.2148043, 9.98265148, 36.6033477, 65.1423438, 101.503429, 147.408733, 204.58039, 274.740529, 358.587686, 460.118651, 577.224048, 711.806863, 873.961976, 1070.10627, 1306.65664, 1589.90078, 1926.41413, 2322.91305, 2781.60549, 3318.37042, 3934.19673, 4641.57173, 5472.46616, 6449.90758, 7593.71434, 8926.08196, 10480.973, 12285.6938, 14337.7211, 16698.7804, 19406.5036, 22498.5228, 26012.4702, 31396.4395, 36924.9857, 42283.1255, 47155.8756, 51228.2528, 53100.4193, 54265.4824, 54873.3963, 55413.4705, 55901.3312, 56211.2061, 56365.7913, 56387.7826, 56299.8761, 56124.7675, 57781.8103, 58334.8394, 57920.4254, 56675.1393, 54735.5517, 52238.2335, 49319.7554, 43230.3511, 31986.8909},
        {6.52818835, 6.33979973, 6.14990323, 5.95850008, 5.7655915, 5.57117874, 5.37526303, 5.1778456, 4.97892768, 4.77851051, 4.57659533, 4.37318335, 4.16827583, 3.96187399, 3.75397906, 3.54459229, 3.33371489, 3.12134812, 2.90749319, 2.69215135, 2.47532382, 2.25701185, 2.03721666, 1.81593948, 1.59318156, 1.36894412, 1.1432284, 0.916035637, 0.687367056, 0.457223894, 0.225607384, -0.00748123906, -0.242040743, -0.478069893, -0.71842368, -0.961612682, -1.20611315, -1.45183115, -1.69867277, -1.94654408, -2.19535116, -2.44500009, -2.69539694, -2.94644779, -3.19805871, -3.45013578, -3.70258508, -3.95531269, -4.20822468, -4.46122712, -4.7142261, -4.99697103, -5.27887769, -5.55920523, -5.83721278, -6.11215946, -6.38330441, -6.64990676, -6.91122564, -7.16652019, -7.41504952, -7.65607278, -7.8888491, -8.11263761, -8.32669744, -8.53028771, -8.72266757, -8.91410369, -9.11459844, -9.30070873, -9.47098013, -9.62395818, -9.75818846, -9.87221651, -9.96458789, -10.0338482, -10.0785429, -10.0972176, -10.0884179, -10.0506893, -9.98257739, -9.88262771, -9.74938582, -9.58139729, -8.98831313, -8.17130605, -7.33622295, -6.49422976, -5.65649243, -4.83417687, -4.03844902, -3.28047482, -2.5714202, -1.92245109, -1.34473343, -0.849433141, -0.447716162, -0.150748426, 0.030304135, 0.084275588, 0, -0.401776553, -1.07243729, -1.99660564, -3.15890505, -4.54395895, -6.13639078, -7.92082397, -9.88188196, -12.0041882, -14.2723661, -16.6710391, -19.1848306, -21.7983641, -24.4962631, -27.2631509, -30.0836509, -32.9636532, -35.8908193, -38.8016697, -41.685385, -44.5311458, -47.3281325, -50.0655258, -52.7325062, -55.3182542, -57.8119503, -60.2027752, -62.4799094, -64.6325334, -67.1189621, -69.2872258, -70.9857647, -72.063019, -72.7336752, -72.6630504, -71.418796, -68.7394783, -64.3636637, -58.0299186, -49.4768092, -40.5867924, -28.3020556, -11.9417084, 9.1751398, 35.7293795, 65.3904055, 102.603478, 148.802361, 205.42082, 273.892618, 353.554596, 450.50035, 564.365462, 697.626521, 858.560503, 1052.84013, 1286.13812, 1563.40067, 1891.19214, 2276.86948, 2721.52406, 3243.13672, 3847.95664, 4549.16869, 5371.70997, 6335.1231, 7451.16828, 8737.37045, 10239.7902, 11988.3468, 13994.9021, 16310.2106, 18970.4425, 22011.7674, 25470.3554, 30815.6871, 36292.2939, 41578.018, 46350.7016, 50288.1869, 51901.0785, 52812.6146, 53200.8821, 53753.8582, 54447.7983, 55046.5032, 55536.8509, 55905.7192, 56139.9863, 56226.53, 57234.5385, 57482.7584, 57002.7337, 55826.0084, 53984.1266, 51508.6323, 48431.0694, 41798.1049, 29952.0878},
        {6.86966265, 6.68273886, 6.49333968, 6.30149642, 6.10724035, 5.91060278, 5.71161498, 5.51030825, 5.30671388, 5.10086315, 4.89278736, 4.68251778, 4.47008573, 4.25552247, 4.0388593, 3.82012751, 3.59935839, 3.37658323, 3.15183331, 2.92513993, 2.69653437, 2.46604792, 2.23371188, 1.99955753, 1.76361615, 1.52591905, 1.2864975, 1.0453828, 0.802606238, 0.558199099, 0.312192673, 0.0646182507, -0.18449288, -0.435109431, -0.689371269, -0.946115861, -1.20415458, -1.46338566, -1.72370738, -1.98501798, -2.2472157, -2.51019882, -2.77386557, -3.03811422, -3.30284301, -3.56795019, -3.83333403, -4.09889276, -4.36452466, -4.63012796, -4.89560092, -5.19053434, -5.48449026, -5.77672325, -6.06648787, -6.35303872, -6.63563036, -6.91351737, -7.18595433, -7.45219581, -7.7114964, -7.96311066, -8.20629317, -8.44029851, -8.66438126, -8.87779599, -9.07979728, -9.28241632, -9.49737755, -9.69758471, -9.88146415, -10.0474422, -10.1939451, -10.3193993, -10.422231, -10.5008665, -10.5537323, -10.5792546, -10.5758597, -10.5419739, -10.4760236, -10.3764351, -10.2416347, -10.0700488, -9.45805976, -8.61357232, -7.74932343, -6.87678651, -6.00743498, -5.15274224, -4.32418172, -3.53322684, -2.791351, -2.11002763, -1.50073014, -0.974931943, -0.544106463, -0.219727112, -0.0132673074, 0.0637995354, 0, -0.464722346, -1.19081776, -2.16331319, -3.36723558, -4.78761187, -6.40946901, -8.21783394, -10.1977336, -12.334195, -14.6122449, -17.0169105, -19.5332186, -22.1461961, -24.84087, -27.6022673, -30.4154149, -33.2785415, -36.1785602, -39.0706365, -41.9426264, -44.7823857, -47.5777703, -50.3166362, -52.9868393, -55.5762356, -58.0726808, -60.464031, -62.738142, -64.8828698, -67.3331952, -69.4659033, -71.1347128, -72.1933419, -72.8971759, -72.8990999, -71.7448878, -69.1677581, -64.9009291, -58.6776192, -50.231047, -41.4119305, -29.191532, -12.8887765, 8.17741082, 34.6881051, 65.4162234, 103.428038, 149.881664, 205.935214, 272.746804, 348.36224, 440.855877, 551.552192, 683.472116, 843.100168, 1035.36579, 1265.19841, 1536.05167, 1854.66623, 2229.39267, 2660.71753, 3168.15063, 3762.58318, 4457.7054, 5271.95016, 6221.57339, 7310.83475, 8552.88003, 10004.8412, 11698.9692, 13659.1516, 15927.2892, 18538.5281, 21528.0142, 24930.8934, 30240.5276, 35669.2929, 40887.7811, 45566.5842, 49376.2941, 50739.7788, 51407.1702, 51583.799, 52151.8889, 53048.8338, 53929.9889, 54746.8502, 55450.9133, 55993.6743, 56326.629, 56609.676, 56515.5139, 55961.8316, 54866.3181, 53146.6626, 50720.5541, 47505.6817, 40389.101, 28063.274},
        {7.25431735, 7.06847112, 6.8791634, 6.68645565, 6.49040934, 6.29108594, 6.08854692, 5.88285373, 5.67406785, 5.46225074, 5.24746386, 5.0297687, 4.8092267, 4.58589934, 4.35984808, 4.13113439, 3.89981973, 3.66596558, 3.42963339, 3.19088464, 2.94978078, 2.70638329, 2.46075364, 2.21295328, 1.96304369, 1.71108632, 1.45714266, 1.20127416, 0.943542282, 0.684008504, 0.422734287, 0.159781098, -0.104789598, -0.370916334, -0.64022498, -0.911774911, -1.18460522, -1.45859969, -1.73364209, -2.00961619, -2.28640579, -2.56389464, -2.84196653, -3.12050524, -3.39939454, -3.6785182, -3.95776001, -4.23700374, -4.51613317, -4.79503207, -5.07358423, -5.38051514, -5.68624162, -5.99002222, -6.29111548, -6.58877995, -6.88227416, -7.17085666, -7.453786, -7.73032071, -7.99971935, -8.26124046, -8.51414259, -8.75768427, -8.99112405, -9.21372047, -9.42473209, -9.63811878, -9.86748749, -10.0817763, -10.2792908, -10.4583364, -10.6172186, -10.7542431, -10.8677152, -10.9559405, -11.0172246, -11.0498729, -11.052191, -11.0224843, -10.9590584, -10.8602189, -10.7242711, -10.5495207, -9.91874075, -9.04674316, -8.15333416, -7.25030408, -6.34944327, -5.46254207, -4.60139081, -3.77777984, -3.00349949, -2.29034011, -1.65009203, -1.0945456, -0.635491153, -0.284719029, -0.0540195693, 0.044816887, 0, -0.574884706, -1.3949886, -2.44636825, -3.71508023, -5.18718112, -6.8487275, -8.68577593, -10.684383, -12.8306053, -15.1104993, -17.5101217, -20.015529, -22.6127779, -25.2879248, -28.0270263, -30.8161391, -33.6311232, -36.4565784, -39.301625, -42.1501346, -44.9859787, -47.793029, -50.555157, -53.2562344, -55.8801328, -58.4107238, -60.8318789, -63.1274698, -65.2813682, -67.6829791, -69.761945, -71.3804773, -72.4007876, -73.1452217, -73.245924, -72.2123362, -69.7686296, -65.6389752, -59.5475444, -51.2185083, -42.408669, -30.1909399, -13.8918036, 7.16225738, 33.6447606, 65.1967486, 103.782332, 150.333148, 205.780833, 271.057021, 343.190155, 431.785639, 539.500546, 669.771233, 827.625091, 1017.37517, 1243.33451, 1507.61926, 1817.23867, 2181.59862, 2601.12409, 3095.992, 3680.35608, 4368.08477, 5172.56309, 6107.39797, 7171.0211, 8373.10509, 9778.965, 11422.4414, 13335.1161, 15553.4359, 18112.023, 21045.4999, 24388.4888, 29652.4276, 35025.9194, 40174.3827, 44763.236, 48457.8979, 49610.9146, 50075.8251, 50080.7076, 50677.389, 51767.8802, 52910.9633, 54025.8105, 55031.5939, 55847.486, 56392.6588, 55969.7502, 55527.4945, 54880.3037, 53842.5894, 52228.7634, 49853.2376, 46530.4236, 39017.1098, 26381.9961},
        {7.68599905, 7.49989539, 7.30942132, 7.11466505, 6.91571478, 6.7126587, 6.50558503, 6.29458195, 6.07973767, 5.86114039, 5.63887831, 5.41303963, 5.18371256, 4.95098528, 4.71494601, 4.47568294, 4.23328428, 3.98783822, 3.73943297, 3.48815672, 3.23409768, 2.97734405, 2.71798402, 2.45610581, 2.1917976, 1.9251476, 1.65624401, 1.38517503, 1.11202887, 0.836893713, 0.559857771, 0.281009241, 0.000436325892, -0.281772775, -0.567083029, -0.854596981, -1.14339851, -1.43334902, -1.72430989, -2.01614254, -2.30870834, -2.60186871, -2.89548503, -3.1894187, -3.48353113, -3.77768369, -4.0717378, -4.36555484, -4.65899621, -4.95192331, -5.24419754, -5.56292359, -5.88012897, -6.19508451, -6.50706101, -6.81532929, -7.11916016, -7.41782445, -7.71059295, -7.9967365, -8.2755259, -8.54623198, -8.80812553, -9.06047738, -9.30255835, -9.53363924, -9.75299088, -9.97662956, -10.2201695, -10.4483453, -10.6593421, -10.8513453, -11.0225401, -11.1711118, -11.2952457, -11.393127, -11.462941, -11.502873, -11.5111082, -11.485832, -11.4252295, -11.3274862, -11.1907872, -11.0133178, -10.3639119, -9.46446125, -8.54201818, -7.60870124, -6.67662898, -5.75791996, -4.86469271, -4.0090658, -3.20315779, -2.45908721, -1.78897263, -1.2049326, -0.719085667, -0.343550389, -0.0904453163, 0.0281109976, 0, -0.740971614, -1.70102869, -2.86796646, -4.22958016, -5.77366503, -7.48801631, -9.36042924, -11.378699, -13.530621, -15.8039902, -18.1866021, -20.6662518, -23.2307345, -25.8678456, -28.5653801, -31.3111335, -34.0424072, -36.7397833, -39.5052686, -42.3158382, -45.1484674, -47.9801314, -50.7878054, -53.5484644, -56.2390838, -58.8366386, -61.3181041, -63.6604555, -65.8406678, -68.1796364, -70.1843124, -71.728495, -72.6859832, -73.4714583, -73.6882784, -72.7955666, -70.5048572, -66.5276849, -60.5755841, -52.3600896, -43.4634005, -31.157054, -14.7865847, 6.30247233, 32.7645823, 64.708932, 103.471584, 149.843324, 204.614936, 268.577204, 338.217878, 423.890044, 528.926832, 656.951456, 812.17939, 998.560195, 1220.04343, 1477.86887, 1779.31173, 2134.60336, 2544.68195, 3029.24073, 3603.55504, 4281.20973, 5072.92511, 5990.73636, 7030.03464, 8198.54002, 9565.00061, 11163.6436, 13027.4422, 15192.0704, 17692.1897, 20562.4614, 23837.5466, 29032.8536, 34332.1101, 39399.7907, 43900.3698, 47498.3221, 48508.88, 48845.255, 48750.1683, 49400.1851, 50668.3801, 52038.7262, 53403.753, 54655.9901, 55687.9673, 56392.2142, 55377.2881, 54613.0889, 53840.7346, 52801.3433, 51236.0328, 48885.9211, 45492.1263, 37695.9014, 24969.8005},
        {8.1752324, 7.98406735, 7.78800895, 7.58715785, 7.38161469, 7.17148012, 6.95685478, 6.73783931, 6.51453436, 6.28704057, 6.05545859, 5.81988905, 5.5804326, 5.33718989, 5.09026156, 4.83974825, 4.5857506, 4.32836926, 4.06770488, 3.80385809, 3.53692954, 3.26701988, 2.99422974, 2.71865977, 2.44041062, 2.15958292, 1.87627733, 1.59059448, 1.30263501, 1.01249958, 0.720288828, 0.426103391, 0.130043917, -0.167788951, -0.469623025, -0.774144457, -1.08001142, -1.38704769, -1.69507708, -2.00392338, -2.31341037, -2.62336186, -2.93360163, -3.24395349, -3.55424122, -3.86428861, -4.17391947, -4.48295758, -4.79122674, -5.09855075, -5.40475338, -5.73377257, -6.06079524, -6.38512243, -6.70605518, -7.02289456, -7.33494159, -7.64149734, -7.94186283, -8.23533914, -8.52122728, -8.79882833, -9.06744332, -9.32637329, -9.5749193, -9.81238239, -10.0380636, -10.2702913, -10.5269371, -10.7680595, -10.9917262, -11.1960047, -11.3789625, -11.5386674, -11.6731868, -11.7805883, -11.8589396, -11.9063082, -11.9207616, -11.9003675, -11.8431935, -11.7473071, -11.6107759, -11.4316676, -10.7638059, -9.83672451, -8.88529623, -7.92200524, -6.95933567, -6.00977168, -5.0857974, -4.19989697, -3.36455453, -2.59225424, -1.89548023, -1.28671663, -0.778447608, -0.38315729, -0.113329821, 0.0185506565, 0, -1.0933752, -2.34681754, -3.75213535, -5.30113699, -6.9856308, -8.79742514, -10.7283283, -12.7701488, -14.9146947, -17.1537746, -19.4791968, -21.8827696, -24.3563013, -26.8916003, -29.480475, -32.1147337, -34.6424851, -37.0350091, -39.6091488, -42.3259198, -45.1463378, -48.0314182, -50.9421766, -53.8396287, -56.68479, -59.438676, -62.0623024, -64.5166847, -66.7628385, -68.9855203, -70.8399664, -72.2200701, -73.0197247, -73.8194044, -74.169712, -73.4381621, -71.3126735, -67.4811653, -61.6315565, -53.4517659, -44.0984509, -31.3437849, -14.5883249, 6.76737186, 33.3227483, 64.5006819, 102.371322, 147.828254, 201.76506, 265.075326, 335.100572, 420.627856, 523.549626, 646.975376, 796.499412, 976.600412, 1191.75705, 1444.225, 1741.21117, 2092.34759, 2500.31037, 2981.06256, 3544.23373, 4201.7507, 4969.43296, 5861.3132, 6879.04297, 8033.44503, 9380.7397, 10951.4679, 12758.1284, 14852.0798, 17270.1085, 20049.0007, 23225.5424, 28290.5374, 33463.2502, 38416.9625, 42824.9564, 46360.5137, 47375.2113, 47745.1726, 47710.1247, 48475.8434, 49899.6612, 51436.3901, 52963.4886, 54358.4149, 55498.6277, 56261.5853, 55175.1688, 54223.1081, 53200.9564, 51904.2665, 50128.5917, 47669.4848, 44322.4989, 36513.4957, 24107.8279},
        {8.71411517, 8.51657265, 8.313721, 8.10566827, 7.89252249, 7.67439172, 7.451384, 7.22360738, 6.9911699, 6.7541796, 6.51274454, 6.26697277, 6.01697231, 5.76285123, 5.50471756, 5.24267935, 4.97684465, 4.70732151, 4.43421796, 4.15764206, 3.87770185, 3.59450538, 3.30816068, 3.01877582, 2.72645883, 2.43131775, 2.13346064, 1.83299554, 1.5300305, 1.22467355, 0.917032759, 0.607216156, 0.295331791, -0.0185122895, -0.337645864, -0.660169666, -0.984143106, -1.30934658, -1.63556047, -1.96256516, -2.29014106, -2.61806856, -2.94612803, -3.27409987, -3.60176448, -3.92890223, -4.25529353, -4.58071876, -4.90495832, -5.22779259, -5.54900196, -5.88906752, -6.22662021, -6.56099168, -6.89151356, -7.21751749, -7.53833512, -7.85329808, -8.16173802, -8.46298657, -8.75637537, -9.04123606, -9.31690029, -9.58269969, -9.8379659, -10.0820306, -10.3142253, -10.5551205, -10.8247766, -11.0787219, -11.3149112, -11.5312995, -11.7258415, -11.896492, -12.0412059, -12.1579379, -12.244643, -12.2992759, -12.3197914, -12.3041444, -12.2502897, -12.1561821, -12.0197764, -11.8390275, -11.1530115, -10.1987745, -9.21887782, -8.22616786, -7.2334909, -6.25369326, -5.29962124, -4.38412117, -3.52003936, -2.72022213, -1.99751578, -1.36476663, -0.834821009, -0.420525219, -0.134725577, 0.00973160096, 0, -1.42672118, -2.95821414, -4.59004947, -6.31779777, -8.13702967, -10.0433158, -12.0322267, -14.099333, -16.2402053, -18.4504142, -20.7255304, -23.0611244, -25.4527669, -27.8960284, -30.3864795, -32.9196909, -35.2604544, -37.371372, -39.7700143, -42.4024992, -45.2149446, -48.1534687, -51.1641893, -54.1932246, -57.1866924, -60.0907108, -62.8513978, -65.4148713, -67.7272495, -69.8462407, -71.5654559, -72.797536, -73.4551218, -74.2708956, -74.7474775, -74.1688098, -72.201521, -68.5122394, -62.7675935, -54.6342116, -44.8407254, -31.6685416, -14.5765055, 6.97653788, 33.5317433, 63.5444486, 100.16308, 144.450246, 197.468557, 260.280621, 331.42166, 417.5707, 518.803265, 637.569998, 781.170749, 954.728176, 1163.36493, 1410.57163, 1703.47389, 2050.97774, 2457.66558, 2935.41334, 3487.47626, 4123.98233, 4866.70508, 5732.07256, 6728.91455, 7871.68919, 9202.71482, 10748.5505, 12501.293, 14527.2654, 16864.9561, 19552.8531, 22629.4447, 27534.9099, 32555.6236, 37379.6511, 41695.0581, 45189.9102, 46297.2971, 46796.9106, 46914.6629, 47823.3461, 49374.066, 51023.3946, 52642.1642, 54101.2071, 55271.3553, 56023.4412, 54932.6887, 53824.8767, 52559.3031, 50995.2658, 48992.0627, 46408.9918, 43105.3509, 35353.0333, 23471.2867},
        {9.30106999, 9.0968134, 8.88687571, 8.67136912, 8.45040586, 8.22409812, 7.99255814, 7.75589811, 7.51423026, 7.26766679, 7.01631992, 6.76030186, 6.49972482, 6.23470102, 5.96534267, 5.69176198, 5.41407117, 5.13238244, 4.84680801, 4.5574601, 4.26445091, 3.96789267, 3.66789757, 3.36457784, 3.05804569, 2.74841333, 2.43579297, 2.12029683, 1.80203712, 1.48112605, 1.15767584, 0.831798689, 0.503606823, 0.173212451, -0.163989678, -0.505429393, -0.848479824, -1.19287567, -1.53835162, -1.88464238, -2.23148264, -2.5786071, -2.92575046, -3.27264741, -3.61903264, -3.96464086, -4.30920676, -4.65246503, -4.99415038, -5.3339975, -5.67174109, -6.02467058, -6.37458539, -6.72083964, -7.06278749, -7.39978307, -7.73118054, -8.05633403, -8.37459769, -8.68532567, -8.9878721, -9.28159113, -9.56583691, -9.83996358, -10.1033253, -10.3552762, -10.5951703, -10.8455742, -11.1284966, -11.3954185, -11.6441892, -11.872658, -12.0786745, -12.260088, -12.4147479, -12.5405035, -12.6352043, -12.6966997, -12.7228391, -12.7114718, -12.6604472, -12.5676148, -12.4308239, -12.247924, -11.5441395, -10.5635733, -9.55600885, -8.53463314, -7.51263314, -6.50319585, -5.51950823, -4.57475727, -3.68212995, -2.85481324, -2.10599414, -1.44885962, -0.896596652, -0.462392228, -0.159433324, -0.00090692137, 0, -1.64839402, -3.36663592, -5.15265389, -7.00437617, -8.91973097, -10.8966465, -12.933051, -15.0268726, -17.1760396, -19.3784803, -21.6321227, -23.9348952, -26.2847259, -28.679543, -31.1172749, -33.5958496, -35.8287157, -37.7731615, -40.0705811, -42.657943, -45.4722152, -48.4503662, -51.5293643, -54.6461776, -57.7377744, -60.741123, -63.5931918, -66.2309489, -68.5913627, -70.6505124, -72.2919146, -73.4408042, -74.0224162, -74.8660339, -75.444868, -74.9910498, -73.1585996, -69.6015377, -63.9738847, -55.9296608, -45.8719506, -32.4777345, -15.2544767, 6.29035829, 32.6493061, 61.331226, 96.5582491, 139.56883, 191.601423, 253.894484, 326.153895, 413.023825, 513.01444, 627.969242, 766.537295, 934.312103, 1136.88717, 1378.60082, 1666.58701, 2009.349, 2413.09235, 2886.4025, 3427.79102, 4045.93308, 4766.72557, 5608.16033, 6583.87672, 7710.73818, 9021.56715, 10540.1588, 12247.6686, 14217.0912, 16486.9817, 19095.8949, 22082.3859, 26804.649, 31651.6899, 36332.154, 40554.6863, 44027.9321, 45309.4605, 46026.3766, 46380.644, 47451.7252, 49094.2402, 50797.1845, 52433.4593, 53875.9658, 54997.6053, 55671.2792, 54486.9827, 53226.608, 51778.4243, 50030.7008, 47871.707, 45189.7119, 41872.985, 34165.4373, 22946.1409},
        {9.93030726, 9.71802666, 9.4999446, 9.27616459, 9.04679013, 8.81192474, 8.57167191, 8.32613515, 8.07541797, 7.81962387, 7.55885635, 7.29321892, 7.02281509, 6.74774836, 6.46812223, 6.18404021, 5.89560581, 5.60292253, 5.30609387, 5.00522334, 4.70041445, 4.3917707, 4.0793956, 3.76339264, 3.44386534, 3.1209172, 2.79465172, 2.46517242, 2.13258279, 1.79698633, 1.45848657, 1.11718699, 0.773191112, 0.426602431, 0.0711358932, -0.289776341, -0.652625325, -1.01710023, -1.38289024, -1.74968452, -2.11717224, -2.48504258, -2.85298471, -3.22068782, -3.58784106, -3.95413361, -4.31925466, -4.68289337, -5.04473891, -5.40448047, -5.76180721, -6.13215617, -6.49912728, -6.86206833, -7.22032711, -7.57325142, -7.92018903, -8.26048776, -8.59349537, -8.91855967, -9.23502845, -9.5422495, -9.83957061, -10.1263396, -10.4019042, -10.6656122, -10.9168114, -11.1793669, -11.4765548, -11.7571581, -12.0189354, -12.2596451, -12.4770455, -12.6688952, -12.8329525, -12.966976, -13.068724, -13.1359551, -13.1664276, -13.1579, -13.1081307, -13.0148783, -12.875901, -12.6889575, -11.9678018, -10.962375, -9.92843532, -8.87944805, -7.8288785, -6.790192, -5.77685386, -4.80232937, -3.88008386, -3.02358263, -2.246291, -1.56167426, -0.983197743, -0.524326747, -0.198526585, -0.0192625655, 0, -1.51584468, -3.13099143, -4.84127446, -6.64252799, -8.53058625, -10.5012835, -12.5504538, -14.6739316, -16.867551, -19.1271462, -21.4485515, -23.827601, -26.2601291, -28.7419699, -31.2689576, -33.8369264, -36.210561, -38.3526741, -40.7819385, -43.4446564, -46.28713, -49.2556613, -52.2965525, -55.3561058, -58.3806232, -61.3164071, -64.1097594, -66.7069824, -69.0543781, -71.1936026, -72.9406411, -74.2131897, -74.9289442, -75.8453122, -76.4601339, -76.0473266, -74.2726727, -70.8019549, -65.3009557, -57.4354576, -47.5927282, -34.4974828, -17.6743436, 3.35206686, 29.0571261, 56.3388862, 90.1443817, 131.843321, 182.805414, 244.400369, 317.096546, 404.266652, 504.052489, 617.999242, 754.600668, 919.230757, 1117.2635, 1352.22279, 1631.75319, 1965.51754, 2360.78493, 2825.21159, 3358.43599, 3968.70886, 4678.87388, 5505.07633, 6456.2167, 7546.56202, 8816.94338, 10293.1657, 11982.3272, 13931.8176, 16173.9172, 18740.9064, 21665.0656, 26129.4952, 30727.9507, 35205.0071, 39305.24, 42773.2246, 44391.0872, 45507.485, 46279.4715, 47575.8796, 49272.0167, 50946.6158, 52489.8133, 53791.7453, 54742.5483, 55232.3586, 53566.343, 52109.1656, 50653.8892, 48993.5766, 46921.2904, 44230.0934, 40713.0482, 32780.1399, 22361.4086},
        {10.6122373, 10.3939124, 10.1694539, 9.93896349, 9.70254269, 9.46029313, 9.21231641, 8.95871413, 8.69958788, 8.43503928, 8.16516991, 7.89008139, 7.60987531, 7.32465328, 7.03451689, 6.73956775, 6.43990746, 6.13563762, 5.82685982, 5.51367568, 5.19618679, 4.87449475, 4.54870116, 4.21890764, 3.88521576, 3.54772715, 3.20654339, 2.8617661, 2.51349686, 2.16183729, 1.80688898, 1.44875353, 1.08753255, 0.723327634, 0.348458055, -0.0329197299, -0.416555726, -0.802095781, -1.18918574, -1.57747146, -1.96659877, -2.35621353, -2.74596159, -3.13548879, -3.52444098, -3.912464, -4.29920371, -4.68430595, -5.06741658, -5.44818142, -5.82624634, -6.21566078, -6.60135475, -6.98266185, -7.35891568, -7.72944986, -8.09359798, -8.45069366, -8.8000705, -9.14106211, -9.47300208, -9.79522403, -10.1070616, -10.4078483, -10.6969178, -10.9736037, -11.2372396, -11.5127424, -11.8244147, -12.1188266, -12.3936532, -12.6465696, -12.8752509, -13.0773724, -13.2506092, -13.3926364, -13.5011291, -13.5737625, -13.6082118, -13.6021521, -13.5532585, -13.4592062, -13.3176703, -13.126326, -12.3883954, -11.3591804, -10.2999256, -9.22435221, -8.14618125, -7.07913384, -6.03693107, -5.03329405, -4.08194386, -3.19660161, -2.39098838, -1.67882528, -1.0738334, -0.589733839, -0.240247686, -0.0390960416, 0, -1.29608581, -2.73670124, -4.31449084, -6.02209911, -7.8521706, -9.79734983, -11.8502813, -14.0036096, -16.2499792, -18.5820346, -20.9924205, -23.4737812, -26.0187613, -28.6200054, -31.270158, -33.9618636, -36.5347904, -38.9597294, -41.5753426, -44.341494, -47.2180475, -50.164867, -53.1418164, -56.1087596, -59.0255606, -61.8520831, -64.5481911, -67.0737486, -69.3886194, -71.6318365, -73.5183994, -74.9544212, -75.8460152, -76.8332195, -77.4551846, -77.0553482, -75.3196459, -71.9340134, -66.5843865, -58.9567009, -49.5520177, -36.9930656, -20.8062992, -0.518173107, 24.3448579, 50.3432151, 82.841849, 123.292586, 173.147252, 233.857675, 306.339553, 393.236105, 492.862165, 606.638709, 742.328422, 904.846693, 1099.10891, 1327.49654, 1598.03478, 1921.51311, 2305.27302, 2757.71887, 3281.98246, 3886.57854, 4589.14298, 5403.12492, 6330.48354, 7381.6891, 8609.34164, 10042.2102, 11716.9543, 13653.2385, 15877.1008, 18414.5789, 21291.7105, 25523.3211, 29899.4639, 34198.9797, 38200.7094, 41683.4938, 43651.9373, 45175.2748, 46364.163, 47857.2085, 49556.0876, 51142.497, 52529.1118, 53628.607, 54353.6576, 54616.9389, 52379.9785, 50732.5217, 49336.8942, 47855.422, 45950.4308, 43284.2464, 39519.1948, 31425.3934, 21839.8747},
        {11.3515349, 11.1309259, 10.9034136, 10.669114, 10.4281427, 10.1806157, 9.92664861, 9.66635735, 9.39985769, 9.12726543, 8.84869637, 8.56426631, 8.27409104, 7.97828638, 7.67696811, 7.37025205, 7.05825398, 6.74108972, 6.41887505, 6.09172579, 5.75975772, 5.42308666, 5.08182839, 4.73609873, 4.38601347, 4.0316884, 3.67323934, 3.31078208, 2.94443242, 2.57430616, 2.2005191, 1.82318705, 1.44242579, 1.05835114, 0.662533823, 0.25954191, -0.145953828, -0.553560288, -0.962884365, -1.37353296, -1.78511296, -2.19723126, -2.60949477, -3.02151038, -3.43288498, -3.84322547, -4.25213875, -4.65923171, -5.06411125, -5.46638427, -5.86565765, -6.27487348, -6.68001439, -7.0803982, -7.47534272, -7.86416578, -8.24618518, -8.62071874, -8.98708428, -9.34459961, -9.69258255, -10.0303509, -10.3572225, -10.6725152, -10.9755467, -11.2656349, -11.5420976, -11.8307883, -12.1569231, -12.4650923, -12.7528935, -13.0179243, -13.2577823, -13.4700651, -13.6523704, -13.8022958, -13.9174389, -13.9953973, -14.0337687, -14.0301506, -13.9821408, -13.8873367, -13.7433361, -13.5477365, -12.7937598, -11.7418003, -10.6583147, -9.55726858, -8.45262746, -7.35835696, -6.28842266, -5.25679015, -4.27742503, -3.36429286, -2.53135925, -1.79258978, -1.16195003, -0.653405591, -0.280922049, -0.0584649897, 0, -1.07572627, -2.34116625, -3.78576813, -5.39898008, -7.17025031, -9.089027, -11.1447583, -13.3268925, -15.6248777, -18.0281621, -20.5261938, -23.1084212, -25.7642923, -28.4832554, -31.2547586, -34.0682501, -36.8362045, -39.5370735, -42.3344127, -45.2011759, -48.1103168, -51.0347894, -53.9475474, -56.8215446, -59.6297349, -62.3450721, -64.9405099, -67.3890023, -69.6635029, -71.9937111, -74.0016952, -75.5833855, -76.6347122, -77.672333, -78.2818143, -77.8831953, -76.1861879, -72.9005042, -67.7358559, -60.4019552, -51.6002914, -39.7469537, -24.3576107, -4.94793083, 18.9664173, 43.9524144, 75.3910804, 114.746084, 163.481095, 223.059783, 294.453796, 380.220181, 479.299412, 593.166373, 728.470356, 889.537449, 1080.69374, 1303.26068, 1565.25187, 1877.95869, 2247.42717, 2684.59943, 3197.70446, 3796.23154, 4491.80551, 5295.07097, 6201.04662, 7216.3259, 8404.79714, 9797.93944, 11453.8129, 13373.2262, 15578.4574, 18091.7844, 20935.4851, 25005.6169, 29232.2336, 33419.1683, 37370.2543, 40889.3247, 43155.6133, 45013.9522, 46535.86, 48153.0708, 49794.5179, 51239.0446, 52423.0887, 53283.0877, 53755.4794, 53776.7016, 50966.8359, 49151.6225, 47852.7304, 46591.8288, 44890.5871, 42270.6743, 38253.7598, 30188.2529, 21400.3197},
        {12.174701, 11.9637337, 11.743606, 11.5145068, 11.2766254, 11.030151, 10.7752727, 10.5121797, 10.2410612, 9.96210631, 9.67550422, 9.3814441, 9.08011513, 8.77170647, 8.45640729, 8.13440677, 7.80589408, 7.47105837, 7.13008884, 6.78317464, 6.43050494, 6.07226892, 5.70865574, 5.33985458, 4.9660546, 4.58744498, 4.20421489, 3.81655348, 3.42464995, 3.02869345, 2.62887315, 2.22537823, 1.81839785, 1.40812119, 0.987854005, 0.561371947, 0.132533415, -0.298249042, -0.730562873, -1.16399553, -1.59813446, -2.03256711, -2.46688094, -2.90066339, -3.33350192, -3.76498397, -4.19469699, -4.62222844, -5.04716576, -5.4690964, -5.88760782, -6.31490018, -6.7376748, -7.1552457, -7.56692692, -7.97203248, -8.36987642, -8.75977277, -9.14103556, -9.51297882, -9.87491658, -10.2261629, -10.5660317, -10.8938372, -11.2088932, -11.510514, -11.7980134, -12.0982711, -12.437505, -12.758177, -13.0578143, -13.3339438, -13.5840927, -13.805788, -13.9965567, -14.1539261, -14.275423, -14.3585746, -14.4009079, -14.3999501, -14.3532281, -14.258269, -14.1126, -13.913748, -13.1446419, -12.0709317, -10.9644527, -9.8394315, -8.71009494, -7.59066977, -6.49538276, -5.43846067, -4.43413028, -3.49661834, -2.64015163, -1.87895691, -1.22726095, -0.699290506, -0.309272358, -0.0714332669, 0, -1.131893, -2.44760606, -3.93691922, -5.5896125, -7.39546596, -9.34425962, -11.4257735, -13.6297877, -15.9460822, -18.3644371, -20.8746324, -23.4664481, -26.1296643, -28.854061, -31.6294183, -34.4455161, -37.1922722, -39.8452003, -42.61996, -45.4849322, -48.4084979, -51.3590379, -54.3049333, -57.2145649, -60.0563137, -62.7985607, -65.4096867, -67.8580727, -70.1120996, -72.3379011, -74.2335962, -75.7082398, -76.6708871, -77.653813, -78.2544628, -77.9040897, -76.3247831, -73.2386321, -68.367726, -61.434154, -53.3378443, -42.2645745, -27.7059869, -9.15372408, 13.9005716, 38.824074, 70.0516203, 108.876864, 156.593458, 214.495056, 282.670129, 365.090515, 461.970649, 575.198454, 710.078557, 870.381553, 1059.87804, 1280.30073, 1537.92127, 1841.2344, 2190.5078, 2604.00535, 3096.95882, 3681.6611, 4365.42538, 5157.85061, 6053.45257, 7057.91976, 8232.2457, 9606.14027, 11207.193, 13067.5603, 15217.686, 17688.0141, 20508.9882, 24662.824, 28990.2507, 33283.7692, 37335.8807, 40939.0859, 43244.1051, 45113.0738, 46613.5418, 48123.0664, 49586.4861, 50819.4696, 51773.3141, 52399.3165, 52648.7739, 52472.9833, 49207.3172, 47277.5507, 46084.3896, 45028.5397, 43510.7068, 40931.5968, 36691.9154, 29396.2231, 21096.3339},
        {13.0509021, 12.8528947, 12.6430972, 12.4217856, 12.1892361, 11.9457247, 11.6915274, 11.4269204, 11.1521797, 10.8675814, 10.5734015, 10.2699162, 9.95740144, 9.63613337, 9.30638805, 8.96844153, 8.6225699, 8.26904922, 7.90815557, 7.54016502, 7.16535364, 6.78399749, 6.39637266, 6.00275521, 5.60342122, 5.19864675, 4.78870788, 4.37388067, 3.95444121, 3.53066555, 3.10282978, 2.67120996, 2.23608217, 1.79772247, 1.35162112, 0.90054777, 0.447329574, -0.00760208255, -0.463815821, -0.92088026, -1.37836402, -1.83583572, -2.29286398, -2.74901742, -3.20386466, -3.65697432, -4.10791502, -4.55625538, -5.00156401, -5.44340955, -5.8813606, -6.3266173, -6.76686461, -7.201419, -7.62959695, -8.05071491, -8.46408937, -8.8690368, -9.26487366, -9.65091644, -10.0264816, -10.3908856, -10.7434449, -11.083476, -11.4102954, -11.7232196, -12.0215649, -12.3333382, -12.6858574, -13.0192674, -13.3310247, -13.6185861, -13.879408, -14.1109471, -14.31066, -14.4760032, -14.6044334, -14.6934072, -14.7403811, -14.7428118, -14.6981559, -14.6038699, -14.4574104, -14.2562341, -13.4729577, -12.378497, -11.250076, -10.1021834, -8.94930778, -7.80593785, -6.68656226, -5.60566968, -4.57774874, -3.61728813, -2.73877649, -1.95670248, -1.28555476, -0.739821998, -0.333992843, -0.0825559579, 0, -1.24124295, -2.65081252, -4.21945883, -5.93793199, -7.79698212, -9.78735934, -11.8998138, -14.1250955, -16.4539547, -18.8771414, -21.3854058, -23.9694979, -26.620168, -29.3281661, -32.0842423, -34.8791468, -37.5683767, -40.1233605, -42.8407772, -45.6823943, -48.6099791, -51.585299, -54.5701213, -57.5262134, -60.4153426, -63.1992764, -65.839782, -68.2986269, -70.5375783, -72.6438083, -74.407698, -75.7533932, -76.6050401, -77.5173317, -78.0991404, -77.7911927, -76.3284707, -73.4459561, -68.8786308, -62.3614767, -55.0221602, -44.7791182, -31.0948506, -13.4318576, 8.74736061, 33.8591568, 65.0924649, 103.515071, 150.194763, 206.199327, 270.485734, 348.812471, 442.764655, 554.718404, 688.664815, 847.869891, 1035.59964, 1254.28795, 1508.22206, 1802.59694, 2131.49745, 2520.71638, 2991.89999, 3560.21197, 4229.83163, 5009.98031, 5896.42445, 6894.8971, 8060.46576, 9420.29157, 10959.7919, 12754.2937, 14845.408, 17274.7457, 20083.9177, 24405.5107, 28917.4159, 33388.5003, 37587.6309, 41283.6748, 43528.1102, 45285.4519, 46632.0194, 47946.2858, 49184.9678, 50178.8848, 50891.0485, 51284.4706, 51322.1626, 50967.1363, 47398.3037, 45399.3027, 44288.6681, 43384.9347, 42006.6373, 39472.3108, 35100.4898, 28745.1742, 20890.0663},
        {13.9701823, 13.7859525, 13.5872578, 13.3744612, 13.1479261, 12.9080154, 12.6550925, 12.3895205, 12.1116624, 11.8218816, 11.5205411, 11.2080041, 10.8846339, 10.5507934, 10.2068461, 9.85315485, 9.490083, 9.11799366, 8.73725001, 8.34821521, 7.95125242, 7.54672481, 7.13499554, 6.71642778, 6.2913847, 5.86022946, 5.42332523, 4.98103517, 4.53372244, 4.08175022, 3.62548167, 3.16527994, 2.70150822, 2.23452966, 1.76184379, 1.28530582, 0.806805398, 0.326798616, -0.154258429, -0.63590964, -1.11769892, -1.59917018, -2.07986731, -2.55933422, -3.03711482, -3.51275301, -3.98579268, -4.45577776, -4.92225214, -5.38475972, -5.8428444, -6.30608265, -6.76376833, -7.21522785, -7.65978764, -8.09677411, -8.52551368, -8.94533276, -9.35555779, -9.75551517, -10.1445313, -10.5219327, -10.8870456, -11.2391966, -11.577712, -11.9019183, -12.2111418, -12.5346992, -12.901187, -13.2480791, -13.5727579, -13.8726055, -14.1450043, -14.3873366, -14.5969846, -14.7713307, -14.907757, -15.0036459, -15.0563797, -15.0633405, -15.0219108, -14.9294728, -14.7834087, -14.5811009, -13.7847086, -12.6704832, -11.5211094, -10.3513382, -9.17592108, -8.00960926, -6.86715405, -5.76330674, -4.71281863, -3.730441, -2.83092516, -2.0290224, -1.339484, -0.777061273, -0.3565055, -0.0925679776, 0, -1.36266914, -2.87636318, -4.53289004, -6.32405764, -8.24167389, -10.2775467, -12.423484, -14.6712938, -17.0127838, -19.4397621, -21.9440366, -24.5174151, -27.1517057, -29.8387162, -32.5702545, -35.3381285, -37.9707152, -40.4364751, -43.0968816, -45.9087218, -48.8287827, -51.8138511, -54.820714, -57.8061586, -60.7269716, -63.5399401, -66.201851, -68.6694914, -70.8996482, -72.9197, -74.5899586, -75.8460333, -76.6235337, -77.4751144, -78.0278623, -77.7443189, -76.3749801, -73.6703416, -69.3808995, -63.2571496, -56.6714208, -47.2587746, -34.4523919, -17.6854535, 3.60885998, 28.9623959, 60.2356882, 98.2542991, 143.843791, 197.829725, 258.029071, 331.953154, 422.375003, 532.119436, 664.117009, 821.250807, 1006.40391, 1222.75585, 1472.82588, 1758.80989, 2069.66461, 2437.09892, 2887.10939, 3437.08461, 4089.82746, 4854.83588, 5730.24609, 6722.61043, 7880.14085, 9227.48429, 10705.8106, 12436.0864, 14471.8545, 16866.6577, 19674.0388, 24208.2859, 28947.0102, 33628.5682, 37991.3164, 41773.6111, 43884.6115, 45444.6698, 46547.5128, 47612.6573, 48604.8716, 49353.1782, 49829.133, 50004.2914, 49850.2092, 49338.442, 45685.024, 43668.2651, 42586.2487, 41737.0588, 40418.779, 37929.493, 33567.2846, 28160.9733, 20776.9624},
        {14.8745343, 14.6921407, 14.4941465, 14.2809388, 14.052905, 13.8104323, 13.5539079, 13.283719, 13.0002528, 12.7038965, 12.3950375, 12.0740629, 11.7413599, 11.3973157, 11.0423177, 10.6767529, 10.3010087, 9.91547223, 9.52053074, 9.11657146, 8.70398161, 8.28314841, 7.85445908, 7.41830087, 6.97506097, 6.52512663, 6.06888506, 5.6067235, 5.13902916, 4.66618926, 4.18859104, 3.70662172, 3.22066852, 2.73111866, 2.23437306, 1.73289594, 1.22919945, 0.723800169, 0.217214691, -0.290040396, -0.797448504, -1.30449305, -1.81065743, -2.31542508, -2.81827939, -3.31870378, -3.81618167, -4.31019646, -4.80023157, -5.2857704, -5.76629638, -6.24743483, -6.72239411, -7.19052448, -7.65117619, -8.10369953, -8.54744475, -8.98176212, -9.40600191, -9.81951439, -10.2216498, -10.6117585, -10.9891906, -11.3532965, -11.7034264, -12.0389306, -12.3591593, -12.6951801, -13.0772986, -13.4394511, -13.7789317, -14.0930344, -14.3790536, -14.6342834, -14.856018, -15.0415515, -15.1881781, -15.2931919, -15.3538873, -15.3675582, -15.331499, -15.2430038, -15.0993667, -14.8978819, -14.0901067, -12.9576826, -11.7888103, -10.5984943, -9.40173895, -8.2135486, -7.04892764, -5.92288048, -4.85041149, -3.84652506, -2.92622557, -2.10451741, -1.39640496, -0.816892608, -0.380984742, -0.103685744, 0, -1.36139788, -2.8802241, -4.54760043, -6.35464864, -8.29249052, -10.3522478, -12.5250424, -14.8019959, -17.1742302, -19.6328671, -22.1690283, -24.7738356, -27.4384108, -30.1538757, -32.911352, -35.7019615, -38.4016581, -40.9851109, -43.7039406, -46.5245904, -49.4135029, -52.3371213, -55.2618883, -58.154247, -60.9806403, -63.707511, -66.3013022, -68.7284566, -70.9554173, -73.0946222, -74.9129877, -76.3331582, -77.2777783, -78.1437174, -78.6165074, -78.2552203, -76.8402329, -74.1519221, -69.9706648, -64.0768379, -58.1390352, -49.4747421, -37.4949018, -21.6104574, -1.23235208, 23.9485821, 54.871263, 92.1947196, 136.577981, 188.680076, 245.817999, 316.077523, 402.329587, 507.459451, 634.381621, 785.997469, 965.208372, 1175.26968, 1418.64834, 1697.42515, 2001.48777, 2360.41049, 2796.86445, 3327.27599, 3957.48987, 4698.20796, 5549.51228, 6519.35141, 7654.61343, 8980.16004, 10417.596, 12110.9257, 14120.7534, 16507.6838, 19332.3213, 24052.791, 28982.7965, 33833.5622, 38316.3124, 42142.2713, 44083.6452, 45416.6886, 46255.062, 47082.8965, 47867.2083, 48420.6507, 48717.9582, 48733.8651, 48443.1055, 47820.4139, 44606.7634, 42658.9163, 41420.2205, 40334.024, 38843.6747, 36392.5204, 32423.9091, 27484.2953, 20852.7226},
        {15.8130147, 15.6306151, 15.431801, 15.2169684, 14.9865135, 14.7408326, 14.4803218, 14.2053773, 13.9163953, 13.6137719, 13.2979034, 12.9691859, 12.6280155, 12.2747886, 11.9099012, 11.5337495, 11.1467298, 10.7492382, 10.3416708, 9.92442388, 9.49789361, 9.06247617, 8.61856772, 8.16656445, 7.70686255, 7.23985818, 6.76594754, 6.2855268, 5.79899215, 5.30673976, 4.80916581, 4.30666648, 3.79963796, 3.28847643, 2.76758602, 2.24048496, 1.71076353, 1.17901234, 0.645822032, 0.111783245, -0.422513392, -0.956477243, -1.48951768, -2.02104406, -2.55046575, -3.07719212, -3.60063254, -4.12019637, -4.63529298, -5.14533173, -5.64972199, -6.14908868, -6.64159925, -7.12663673, -7.60358414, -8.07182449, -8.53074078, -8.97971605, -9.4181333, -9.84537556, -10.2608258, -10.6638671, -11.0538825, -11.4302549, -11.7923674, -12.1396029, -12.4713446, -12.8207167, -13.2200333, -13.5990887, -13.9550785, -14.2851984, -14.5866442, -14.8566116, -15.0922963, -15.290894, -15.4496003, -15.5656111, -15.636122, -15.6583287, -15.6294269, -15.5466123, -15.4070807, -15.2080277, -14.3899817, -13.2401809, -12.052554, -10.8423617, -9.62486471, -8.41532357, -7.22899893, -6.08115144, -4.98704173, -3.96193043, -3.02107817, -2.17974559, -1.45319332, -0.856681987, -0.405472237, -0.114824696, 0, -1.31135697, -2.79627754, -4.44447632, -6.24566794, -8.18956704, -10.2658882, -12.4643461, -14.7746554, -17.1865306, -19.6896865, -22.2738375, -24.9286985, -27.6439839, -30.4094084, -33.2146866, -36.0495332, -38.8721018, -41.6675984, -44.4971254, -47.3436345, -50.1900771, -53.0194049, -55.8145694, -58.5585222, -61.2342148, -63.8245988, -66.3126257, -68.6812472, -70.9134148, -73.2741362, -75.3608862, -77.0719996, -78.3058109, -79.2387963, -79.6302199, -79.1651745, -77.6585524, -74.9252457, -70.7801467, -65.0381476, -59.6455469, -51.6371418, -40.4108103, -25.3644304, -5.89588015, 18.8329093, 49.1291266, 85.5359068, 128.596385, 178.853697, 233.502753, 300.527189, 382.297329, 481.543465, 601.730813, 745.994622, 917.470143, 1119.14899, 1354.34269, 1626.51947, 1929.66194, 2287.0106, 2713.83864, 3223.95521, 3828.68884, 4540.50668, 5361.49372, 6301.42546, 7408.14541, 8707.96364, 10116.1892, 11787.2102, 13785.7604, 16176.5731, 19024.3819, 23904.9982, 28994.3082, 33979.276, 38546.8658, 42384.0417, 44133.5387, 45222.7026, 45786.0242, 46382.809, 46979.392, 47365.1019, 47515.1241, 47404.6438, 47008.8463, 46302.9169, 43824.7986, 42003.032, 40515.3102, 39039.3267, 37252.7745, 34833.347, 31458.7375, 26745.6343, 20989.6666},
        {16.8250749, 16.6488761, 16.4548041, 16.2432871, 16.0147528, 15.7696294, 15.5083448, 15.2313271, 14.9390042, 14.6318041, 14.3101548, 13.9744843, 13.6252207, 13.2627918, 12.8876257, 12.5001503, 12.1007938, 11.689984, 11.2681489, 10.8357166, 10.3931151, 9.94077227, 9.47911619, 9.00857483, 8.52957619, 8.04254826, 7.54791904, 7.04611651, 6.53756867, 6.02270351, 5.50194903, 4.97573322, 4.44448408, 3.90862959, 3.36142736, 2.8070414, 2.2497219, 1.69012955, 1.12892502, 0.566769013, 0.00432220935, -0.557754705, -1.11880104, -1.67815612, -2.23515925, -2.78914974, -3.33946691, -3.88545007, -4.42643854, -4.96177164, -5.49078866, -6.00907536, -6.519806, -7.02240125, -7.51628179, -8.00086833, -8.47558152, -8.93984207, -9.39307065, -9.83468795, -10.2641147, -10.6807714, -11.084079, -11.473458, -11.8483291, -12.2081131, -12.5522306, -12.916069, -13.3341522, -13.7317222, -14.1058681, -14.453679, -14.772244, -15.0586522, -15.3099927, -15.5233547, -15.6958271, -15.8244992, -15.9064601, -15.9387988, -15.9186044, -15.8429661, -15.7089729, -15.5137141, -14.686005, -13.5190245, -12.3127815, -11.0828067, -9.8446308, -8.61378448, -7.40579841, -6.23620326, -5.1205297, -4.07430842, -3.11307009, -2.25234539, -1.507665, -0.894559586, -0.42855983, -0.125196409, 0, -1.26775123, -2.72450223, -4.35860632, -6.15841679, -8.11228695, -10.2085701, -12.4356196, -14.7817887, -17.2354306, -19.7848988, -22.4185466, -25.1247271, -27.8917938, -30.7080999, -33.5619988, -36.4418437, -39.3921781, -42.4093824, -45.3547133, -48.2285647, -51.0313308, -53.7634056, -56.4251831, -59.0170573, -61.5394224, -63.9926723, -66.3772011, -68.6934029, -70.9416717, -73.5479679, -75.9310782, -77.9637267, -79.5186375, -80.5568284, -80.9068766, -80.3738145, -78.8038784, -76.0433041, -71.9383278, -66.3351857, -61.3785629, -53.9167183, -43.3461914, -29.0635215, -10.4652483, 13.612593, 43.0728327, 78.3788054, 119.993846, 168.381288, 220.811403, 284.843222, 362.113928, 455.109467, 568.048686, 704.372392, 867.521395, 1060.12056, 1286.61207, 1552.32828, 1856.32441, 2214.23646, 2632.64352, 3122.24963, 3700.74832, 4382.62483, 5172.32391, 6081.7277, 7159.5335, 8433.61388, 9818.26277, 11472.405, 13463.2289, 15857.9225, 18723.6743, 23732.2857, 28945.1783, 34028.3871, 38647.9469, 42469.8926, 44021.2417, 44866.3915, 45159.5272, 45532.3608, 45950.0566, 46178.8093, 46193.7318, 45969.9367, 45482.5368, 44706.6448, 43079.1744, 41416.9594, 39660.6902, 37751.0571, 35628.7506, 33234.461, 30508.8786, 25958.4553, 21080.1537},
        {18.0009895, 17.8516944, 17.6805397, 17.4880845, 17.2748878, 17.0415087, 16.7885063, 16.5164397, 16.2258679, 15.9173501, 15.5914452, 15.2487123, 14.8897106, 14.5149991, 14.1251368, 13.7206829, 13.3021964, 12.8702364, 12.425362, 11.9681321, 11.499106, 11.0188427, 10.5279012, 10.0268406, 9.51622007, 8.99659857, 8.4685352, 7.93258905, 7.38931918, 6.83928467, 6.28304459, 5.72115802, 5.15418401, 4.58268166, 4.00328055, 3.41858716, 2.83125515, 2.24197114, 1.65142171, 1.06029344, 0.469272945, -0.120953201, -0.709698401, -1.29627606, -1.8799996, -2.46018241, -3.03613792, -3.60717952, -4.17262062, -4.73177464, -5.28395499, -5.82162104, -6.35108882, -6.8718203, -7.38327749, -7.88492238, -8.37621694, -8.85662318, -9.32560308, -9.78261863, -10.2271318, -10.6586047, -11.0764991, -11.4802772, -11.8694009, -12.2433321, -12.601533, -12.9820479, -13.421995, -13.8412628, -14.2368121, -14.6056038, -14.9445987, -15.2507578, -15.5210417, -15.7524114, -15.9418277, -16.0862515, -16.1826436, -16.2279648, -16.219176, -16.153238, -16.0271118, -15.837758, -14.999768, -13.8135366, -12.5864587, -11.334394, -10.0732025, -8.81874403, -7.58687841, -6.39346553, -5.25436525, -4.18543743, -3.20254193, -2.32153862, -1.55828735, -0.928647995, -0.448480409, -0.133644456, 0, -1.31854842, -2.82542232, -4.50873598, -6.35660367, -8.35713967, -10.4984583, -12.7686737, -15.1559003, -17.6482522, -20.2338439, -22.9007895, -25.6372034, -28.4311997, -31.2708928, -34.144397, -37.0398265, -40.056193, -43.1974533, -46.2008822, -49.0783578, -51.8417582, -54.5029612, -57.073845, -59.5662874, -61.9921665, -64.3633603, -66.6917469, -68.9892041, -71.26761, -74.0494506, -76.6487729, -78.9242727, -80.7346455, -81.9402448, -82.3989374, -81.9681477, -80.5060741, -77.8709147, -73.920868, -68.5141321, -63.7290603, -56.5279965, -46.3347608, -32.5731732, -14.6670539, 8.52167897, 36.8668279, 70.8040973, 110.769192, 157.197815, 207.557595, 268.879963, 342.150265, 429.624136, 536.150726, 665.414749, 821.10092, 1005.7456, 1224.44284, 1483.53946, 1785.44368, 2140.5128, 2548.52815, 3018.04932, 3572.5101, 4228.51675, 4994.26399, 5884.17731, 6943.52466, 8200.12778, 9564.26137, 11199.3783, 13173.9897, 15556.6065, 18415.7399, 23479.5224, 28741.9383, 33856.593, 38477.0924, 42257.042, 43655.5847, 44317.0976, 44407.099, 44588.0721, 44828.8199, 44891.077, 44753.9997, 44396.7446, 43798.4682, 42938.3271, 42129.8231, 40657.0605, 38692.7299, 36409.522, 33980.1274, 31577.2368, 29373.5408, 25129.8318, 20963.6688},
        {19.3992951, 19.3074759, 19.1860283, 19.0357924, 18.8576079, 18.6523147, 18.4207526, 18.1637616, 17.8821815, 17.5768521, 17.2486134, 16.8983051, 16.5267672, 16.1348395, 15.7233618, 15.2931741, 14.8451162, 14.3800279, 13.8987492, 13.4021198, 12.8909797, 12.3661687, 11.8285267, 11.2788935, 10.718109, 10.1470131, 9.56644561, 8.97724642, 8.38025537, 7.77631235, 7.1662572, 6.5509298, 5.93117, 5.30781767, 4.68779119, 4.06876331, 3.44833373, 2.82714504, 2.20583985, 1.58506077, 0.965450391, 0.347651329, -0.267693817, -0.879942443, -1.48845195, -2.09257972, -2.69168316, -3.28511967, -3.87224664, -4.45242147, -5.02500155, -5.58253193, -6.13133003, -6.67090093, -7.20074971, -7.72038143, -8.22930118, -8.72701403, -9.21302505, -9.68683932, -10.1479619, -10.5958979, -11.0301524, -11.4502304, -11.855637, -12.2458773, -12.6204564, -13.0205426, -13.4863036, -13.9313175, -14.3523884, -14.7463207, -15.1099185, -15.4399861, -15.7333276, -15.9867473, -16.1970494, -16.3610381, -16.4755177, -16.5372924, -16.5431664, -16.4899439, -16.3744291, -16.1934263, -15.3437479, -14.134814, -12.8832563, -11.6053497, -10.3173695, -9.03559071, -7.77628848, -6.55573789, -5.39021406, -4.29599211, -3.28934714, -2.38655426, -1.60388857, -0.957625195, -0.464039234, -0.1394058, 0, -1.52220998, -3.20553734, -5.03972009, -7.01449623, -9.11960377, -11.3447807, -13.6797651, -16.1142949, -18.6381082, -21.2409429, -23.912537, -26.6426287, -29.4209558, -32.2372564, -35.0812685, -37.9427301, -40.9168199, -44.0083388, -46.9685282, -49.8097203, -52.5442478, -55.1844429, -57.7426383, -60.2311662, -62.6623591, -65.0485495, -67.4020699, -69.7352526, -72.0604301, -74.8792032, -77.5352375, -79.8940851, -81.8212979, -83.2673964, -84.0554481, -83.9858626, -82.8987012, -80.6340253, -77.0318962, -71.9323753, -66.9649853, -59.6357076, -49.4344917, -35.8512869, -18.3760427, 3.66565062, 30.5783948, 62.8311507, 100.892879, 145.232541, 193.569402, 252.482857, 322.636424, 406.150131, 508.117711, 632.297764, 782.448889, 961.540179, 1174.29917, 1426.3147, 1719.69192, 2064.4394, 2457.82968, 2908.28582, 3443.12482, 4080.99107, 4835.76358, 5724.85128, 6783.13599, 8035.61587, 9379.87409, 10987.9261, 12928.8752, 15271.8248, 18085.8783, 23103.5776, 28314.7278, 33372.5724, 37930.3552, 41641.3197, 42968.0909, 43548.2771, 43545.387, 43584.4073, 43646.1785, 43519.2588, 43194.1029, 42661.1654, 41910.9011, 40933.7647, 40783.8354, 39521.2226, 37471.2895, 34959.3989, 32310.9139, 29851.1975, 27905.6126, 24257.6629, 20519.9689},
        {20.8705639, 20.846342, 20.7838091, 20.6841157, 20.5484126, 20.3778504, 20.1735798, 19.9367515, 19.6685162, 19.3700245, 19.0424271, 18.6868747, 18.3045179, 17.8965075, 17.4639942, 17.0081285, 16.5300612, 16.0309429, 15.5119243, 14.9741562, 14.4187891, 13.8469738, 13.2598609, 12.658601, 12.044345, 11.4182434, 10.781447, 10.1351063, 9.48037218, 8.81839517, 8.150326, 7.47731534, 6.80051386, 6.12107224, 5.45719063, 4.80113635, 4.14497181, 3.4892944, 2.83470152, 2.18179054, 1.53115885, 0.883403852, 0.239122923, -0.401086545, -1.03662717, -1.66690155, -2.29131231, -2.90926206, -3.5201534, -4.12338896, -4.71837134, -5.29796786, -5.86825811, -6.42878635, -6.97909688, -7.51873399, -8.04724195, -8.56416505, -9.06904757, -9.5614338, -10.040868, -10.5068945, -10.9590576, -11.3969015, -11.8199705, -12.227809, -12.6199611, -13.0405067, -13.5324561, -14.0035026, -14.4503036, -14.8695161, -15.2577975, -15.611805, -15.9281958, -16.2036271, -16.4347563, -16.6182404, -16.7507368, -16.8289028, -16.8493954, -16.808872, -16.7039897, -16.5314059, -15.6707004, -14.4400484, -13.1650421, -11.8623641, -10.5486969, -9.24072326, -7.95512563, -6.70858663, -5.51778883, -4.39941483, -3.3701472, -2.44666853, -1.64566141, -0.983808409, -0.477792119, -0.144295122, 0, -1.7438546, -3.61734607, -5.61223784, -7.72029328, -9.93327582, -12.2429489, -14.6410758, -17.11942, -19.6697449, -22.2838139, -24.9533905, -27.6702379, -30.4261197, -33.2127991, -36.0220397, -38.8456048, -41.7539275, -44.7500099, -47.6476718, -50.4555344, -53.182219, -55.8363467, -58.4265388, -60.9614165, -63.4496009, -65.8997133, -68.3203749, -70.7202068, -73.1078304, -75.9234809, -78.5859068, -80.9742451, -82.9676328, -84.6519981, -85.8030822, -86.1414822, -85.4842975, -83.6486276, -80.4515721, -75.7102305, -70.5481135, -63.0783061, -52.8623055, -39.4616091, -22.4377142, -1.69242739, 23.6381406, 54.0775696, 90.1494399, 132.377331, 178.769431, 235.439081, 302.879901, 383.102746, 481.216154, 600.937866, 745.985623, 919.619387, 1126.11872, 1370.26257, 1653.24713, 1985.76301, 2364.01487, 2796.71466, 3313.76715, 3935.34928, 4680.29714, 5568.44002, 6624.5236, 7870.51255, 9194.10412, 10772.187, 12673.203, 14965.5942, 17717.8023, 22639.9529, 27748.3833, 32701.1146, 37156.1675, 40771.5633, 42038.9788, 42560.3416, 42493.5342, 42411.311, 42307.3346, 42002.4041, 41498.8252, 40798.9037, 39904.945, 38819.2551, 39157.3104, 38079.4316, 36010.153, 33374.0086, 30595.5326, 28099.2592, 26309.7225, 23279.7322, 19892.6998},
        {22.3686335, 22.4167353, 22.4177501, 22.3731344, 22.2843443, 22.1528363, 21.9800665, 21.7674914, 21.5165672, 21.2287501, 20.9054966, 20.548263, 20.1585054, 19.7376803, 19.2872438, 18.8086525, 18.3033624, 17.77283, 17.2185115, 16.6418633, 16.0443416, 15.4274027, 14.792503, 14.1410988, 13.4746463, 12.7946018, 12.1024217, 11.3995623, 10.6874798, 9.96763063, 9.241471, 8.51045724, 7.77604566, 7.03969255, 6.33008136, 5.63448213, 4.93983767, 4.2467207, 3.55570395, 2.86736016, 2.18226205, 1.50098235, 0.824093789, 0.152169103, -0.514218981, -1.17449773, -1.82809443, -2.47443633, -3.11295071, -3.74306484, -4.36420599, -4.96894756, -5.56371926, -6.14809696, -6.72165649, -7.2839737, -7.83462445, -8.37318459, -8.89922995, -9.4123364, -9.91207977, -10.3980359, -10.8697807, -11.32689, -11.7689395, -12.1955053, -12.6061631, -13.0473861, -13.5645216, -14.0604231, -14.5316263, -14.9746669, -15.3860805, -15.7624029, -16.1001697, -16.3959167, -16.6461794, -16.8474936, -16.996395, -17.0894192, -17.1231019, -17.0939789, -16.9985858, -16.8334582, -15.963036, -14.713198, -13.4174205, -12.092752, -10.7562408, -9.42493534, -8.11588404, -6.8461353, -5.63273754, -4.49273916, -3.44318859, -2.50113423, -1.68362449, -1.00770779, -0.490432529, -0.14884713, 0, -1.94460589, -3.98912748, -6.12738732, -8.35320797, -10.660412, -13.0428219, -15.4942603, -18.0085496, -20.5795126, -23.2009716, -25.8667493, -28.5706682, -31.3065509, -34.0682198, -36.8494977, -39.6442069, -42.4828468, -45.3644795, -48.1999682, -50.9909949, -53.7392412, -56.4463892, -59.1141207, -61.7441174, -64.3380613, -66.8976342, -69.4245179, -71.9203943, -74.3869453, -77.1749819, -79.8090426, -82.1860707, -84.2030091, -86.0908941, -87.5796222, -88.3099983, -88.0787376, -86.6825551, -83.918166, -79.5822853, -74.3405271, -66.8562601, -56.7561974, -43.667052, -27.2155372, -7.87118476, 15.7928156, 44.3604553, 78.4157261, 118.542619, 163.033347, 217.564745, 282.646304, 360.211832, 455.103105, 570.856281, 711.007517, 878.752605, 1078.04542, 1313.21116, 1584.50514, 1903.66085, 2267.01027, 2683.7029, 3184.60672, 3790.6206, 4523.74062, 5405.15009, 6452.00913, 7683.7536, 8983.96437, 10528.8331, 12384.8223, 14618.3946, 17296.0126, 22085.1048, 27050.9263, 31859.6984, 36177.6421, 39670.9785, 40874.6319, 41340.318, 41219.0997, 41030.4896, 40779.7056, 40318.4445, 39659.6, 38816.0659, 37800.736, 36626.5041, 37212.9081, 36266.5619, 34246.6729, 31612.4488, 28823.0972, 26337.8255, 24615.8413, 22163.6738, 19094.4725},
        {23.831104, 23.9511974, 24.0162215, 24.0278935, 23.9879308, 23.8980508, 23.7599709, 23.5754084, 23.3460806, 23.073705, 22.7599989, 22.4066797, 22.0154648, 21.5880715, 21.1262171, 20.6316192, 20.1059949, 19.5510617, 18.968537, 18.3601381, 17.7275824, 17.0725873, 16.3968701, 15.7021482, 14.9901389, 14.2625597, 13.5211278, 12.7675608, 12.0035758, 11.2308904, 10.4512218, 9.66628747, 8.87780473, 8.08749095, 7.33162773, 6.59392331, 5.8576705, 5.12346499, 4.39190248, 3.66357865, 2.93908921, 2.21902983, 1.50399623, 0.794584074, 0.0913890702, -0.604993093, -1.29396672, -1.97493613, -2.64730561, -3.31047949, -3.96386207, -4.59807282, -5.22149133, -5.83371233, -6.43433058, -7.02294081, -7.59913779, -8.16251624, -8.71267092, -9.24919657, -9.77168794, -10.2797398, -10.7729468, -11.2509038, -11.7132055, -12.1594467, -12.589222, -13.0504767, -13.5899948, -14.1076523, -14.5999095, -15.0632267, -15.4940643, -15.8888828, -16.2441426, -16.5563039, -16.8218273, -17.0371731, -17.1988016, -17.3031733, -17.3467486, -17.3259878, -17.2373514, -17.0772996, -16.199234, -14.9346703, -13.6228532, -12.2811118, -10.9267749, -9.57717141, -8.24963019, -6.96148016, -5.73005021, -4.57266921, -3.50666607, -2.54936967, -1.7181089, -1.03021266, -0.503009839, -0.153829321, 0, -2.07074529, -4.2218058, -6.44859462, -8.74652487, -11.1110097, -13.5374621, -16.0212952, -18.5579223, -21.1427562, -23.7712103, -26.4386975, -29.140631, -31.8724238, -34.6294892, -37.4072401, -40.2010898, -42.9922365, -45.7740628, -48.574134, -51.3848173, -54.1984797, -57.0074885, -59.8042105, -62.5810131, -65.3302631, -68.0443278, -70.7155742, -73.3363695, -75.8990806, -78.6491861, -81.2331342, -83.5673586, -85.5682929, -87.5802233, -89.3056573, -90.3330082, -90.4503541, -89.4457727, -87.1073421, -83.2231401, -78.1820344, -70.9884642, -61.3129616, -48.8260585, -33.1982872, -15.299855, 6.69830015, 33.4255649, 65.5113261, 103.58497, 146.150026, 198.560009, 261.613688, 337.219036, 429.548661, 541.758397, 677.004078, 837.714212, 1027.93726, 1252.51513, 1511.51611, 1817.20588, 2166.95214, 2570.14111, 3056.52373, 3646.5032, 4362.03251, 5223.91665, 6247.27808, 7450.45363, 8722.74369, 10231.3733, 12039.3456, 14209.6641, 16805.3319, 21438.2779, 26235.3485, 30872.3155, 35024.9508, 38369.0264, 39482.9103, 39871.3808, 39680.5259, 39392.8699, 39021.3278, 38438.633, 37664.7099, 36719.4825, 35622.8753, 34394.8124, 34887.8349, 33981.7479, 32087.6696, 29616.718, 26980.0115, 24588.668, 22853.8059, 20863.863, 18137.3333},
        {25.0387119, 25.2286637, 25.3578729, 25.4281936, 25.4414801, 25.3995866, 25.3043673, 25.1576766, 24.9613686, 24.7172975, 24.4273177, 24.0932833, 23.7170487, 23.3004679, 22.8453953, 22.3536851, 21.8271915, 21.2677688, 20.6772713, 20.057553, 19.4104684, 18.7378716, 18.0416168, 17.3235584, 16.5855505, 15.8294473, 15.0571032, 14.2703723, 13.471109, 12.6611673, 11.8424016, 11.0166662, 10.1858151, 9.35170277, 8.55038488, 7.76589554, 6.98185776, 6.19901585, 5.41811415, 4.63989699, 3.86510872, 3.09449364, 2.32879611, 1.56876044, 0.81513098, 0.0686520491, -0.669932018, -1.39987689, -2.12043823, -2.83087172, -3.53043301, -4.20323902, -4.86401235, -5.51233685, -6.14779637, -6.76997476, -7.37845587, -7.97282353, -8.55266161, -9.11755394, -9.66708437, -10.2008368, -10.7183949, -11.2193428, -11.7032641, -12.1697427, -12.6183626, -13.0962688, -13.6494647, -14.1794423, -14.6827024, -15.1557455, -15.5950723, -15.9971835, -16.3585796, -16.6757614, -16.9452294, -17.1634843, -17.3270268, -17.4323574, -17.4759768, -17.4543857, -17.3640847, -17.2015744, -16.3197966, -15.0505677, -13.7334394, -12.385797, -11.0250258, -9.66851119, -8.33363855, -7.03779325, -5.79836064, -4.63272607, -3.55827493, -2.59239255, -1.75246432, -1.05587558, -0.520011711, -0.162258063, 0, -1.92517137, -3.95204517, -6.07535536, -8.28983591, -10.5902208, -12.9712439, -15.4276393, -17.9541408, -20.5454826, -23.1963984, -25.9016224, -28.6558883, -31.4539303, -34.2904822, -37.1602781, -40.0580518, -42.913092, -45.7107833, -48.5938908, -51.5431246, -54.5391945, -57.5628106, -60.5946827, -63.6155209, -66.606035, -69.546935, -72.4189309, -75.2027326, -77.8790501, -80.581657, -83.0820762, -85.3150985, -87.2155147, -89.1093119, -90.7356827, -91.729501, -91.9081985, -91.0892073, -89.0899592, -85.7278862, -81.6972667, -75.6718199, -67.2974208, -56.2199443, -42.0852654, -25.4761585, -4.86737451, 20.3294365, 50.7026242, 86.8405386, 127.069915, 177.004913, 238.614279, 313.979957, 405.414996, 515.128687, 645.330317, 795.326829, 970.88948, 1180.95573, 1428.99132, 1724.46403, 2066.0028, 2461.97738, 2937.26284, 3509.15801, 4191.71543, 5001.3923, 5966.54903, 7108.8129, 8347.75472, 9822.75925, 11590.781, 13708.7745, 16233.6944, 20725.6526, 25362.6568, 29825.872, 33796.4635, 36955.5963, 37885.9421, 38099.4885, 37748.1835, 37345.2409, 36899.6076, 36269.7035, 35472.0021, 34522.9766, 33439.1007, 32236.8478, 31873.4017, 30778.8591, 29145.0046, 27163.6229, 25026.4986, 22925.4164, 21052.1609, 19206.5877, 17027.873},
        {26.3216461, 26.5661142, 26.745786, 26.8625899, 26.918454, 26.9153067, 26.8550763, 26.7396911, 26.5710793, 26.3511693, 26.0818893, 25.7651676, 25.4029326, 24.9971125, 24.5496355, 24.0624301, 23.5374245, 22.9765469, 22.3817257, 21.7548891, 21.0979655, 20.4128831, 19.7015703, 18.9659553, 18.2079663, 17.4295318, 16.63258, 15.8190392, 14.9908376, 14.1499036, 13.2981654, 12.4375514, 11.5699898, 10.6974089, 9.85286625, 9.02207068, 8.19035704, 7.35864342, 6.5278479, 5.69888859, 4.87268356, 4.05015092, 3.23220874, 2.41977512, 1.61376815, 0.815105911, 0.0247065052, -0.756511983, -1.52763146, -2.28773385, -3.03590104, -3.74877214, -4.44835839, -5.13422821, -5.80595004, -6.46309229, -7.1052234, -7.73191179, -8.34272589, -8.93723412, -9.51500491, -10.0756067, -10.6186079, -11.1435769, -11.6500822, -12.1376922, -12.6059753, -13.1009703, -13.6676001, -14.2095917, -14.723501, -15.2058839, -15.6532964, -16.0622945, -16.429434, -16.751271, -17.0243613, -17.245261, -17.4105259, -17.5167121, -17.5603754, -17.5380718, -17.4463572, -17.2817876, -16.3982506, -15.1271538, -13.8076296, -12.4570897, -11.0929458, -9.73260988, -8.39349354, -7.09300862, -5.84856687, -4.67758007, -3.59745999, -2.62561841, -1.7794671, -1.07641783, -0.533882378, -0.169272509, 0, -1.81301624, -3.74214529, -5.78192039, -7.9268748, -10.1715418, -12.5104545, -14.9381463, -17.4491504, -20.038, -22.6992283, -25.4273687, -28.2169543, -31.0625185, -33.9585944, -36.8997153, -39.8804144, -42.7789381, -45.5732072, -48.5207848, -51.5912857, -54.7543243, -57.9795154, -61.2364735, -64.4948133, -67.7241494, -70.8940964, -73.9742689, -76.9342816, -79.7437491, -82.4370621, -84.8953078, -87.0686681, -88.9073248, -90.6913495, -92.2059791, -93.1484796, -93.3700658, -92.7219525, -91.0553544, -88.2214862, -85.2645651, -80.4797157, -73.4847394, -63.8974375, -51.3356113, -36.1551793, -17.0512961, 6.54276633, 35.1937362, 69.4683414, 107.680263, 155.403, 215.44302, 289.951048, 379.739015, 486.219941, 610.806849, 750.244095, 911.674427, 1107.33369, 1343.33744, 1627.02868, 1959.22778, 2347.44465, 2810.52426, 3362.10861, 4007.21588, 4757.25226, 5655.24457, 6726.33323, 7922.32553, 9352.97701, 11068.8126, 13120.3573, 15558.136, 19879.0607, 24325.6726, 28586.8998, 32351.6704, 35308.9127, 36055.8792, 36100.9573, 35600.9361, 35099.1588, 34599.4035, 33944.3841, 33145.9318, 32215.878, 31166.0539, 30008.2907, 28722.228, 27421.1405, 26063.7037, 24608.5935, 23014.4854, 21240.0552, 19243.9783, 17466.7927, 15791.7918},
        {27.8139918, 28.0791442, 28.2781603, 28.4129522, 28.4854318, 28.4975113, 28.4511026, 28.3481178, 28.190469, 27.9800681, 27.7188273, 27.4086584, 27.0514737, 26.6491851, 26.2037046, 25.7169443, 25.1908163, 24.6272325, 24.0281051, 23.3953459, 22.7308672, 22.0365809, 21.314399, 20.5662336, 19.7939968, 18.9996005, 18.1849568, 17.3519778, 16.5025754, 15.6386618, 14.7621489, 13.8749488, 12.9789735, 12.0761351, 11.1934824, 10.3198303, 9.44369161, 8.56616264, 7.68833968, 6.81131903, 5.936197, 5.06406988, 4.19603398, 3.33318558, 2.476621, 1.62743654, 0.786728486, -0.0444068534, -0.86487318, -1.67357419, -2.46941359, -3.2213319, -3.95884553, -4.68150773, -5.38887176, -6.08049084, -6.75591822, -7.41470715, -8.05641088, -8.68058263, -9.28677566, -9.87454321, -10.4434385, -10.9930148, -11.5228254, -12.0324235, -12.5213622, -13.0347207, -13.6164848, -14.1723143, -14.6988112, -15.1925777, -15.6502156, -16.0683269, -16.4435137, -16.772378, -17.0515217, -17.2775469, -17.4470555, -17.5566496, -17.6029311, -17.5825021, -17.4919644, -17.3279203, -16.4442361, -15.1724007, -13.851628, -12.4993781, -11.133111, -9.77028662, -8.428365, -7.12480614, -5.87707002, -4.70261666, -3.61890605, -2.64339818, -1.79355306, -1.08683068, -0.540691047, -0.172594154, 0, -1.84041759, -3.78747401, -5.83694548, -7.98460821, -10.2262384, -12.5576124, -14.9745062, -17.4726962, -20.0479586, -22.6960696, -25.4128054, -28.1939422, -31.0352563, -33.9325238, -36.8815211, -39.8780242, -42.7567626, -45.4905056, -48.4393024, -51.5644192, -54.827122, -58.1886772, -61.6103507, -65.0534089, -68.4791177, -71.8487435, -75.1235524, -78.2648104, -81.2337839, -84.0007116, -86.514597, -88.7340144, -90.6175382, -92.3747334, -93.8386789, -94.7755231, -95.068543, -94.6010158, -93.2562186, -90.9174287, -88.9025394, -85.2235893, -79.4831698, -71.2838722, -60.2282878, -46.7507588, -29.4141881, -7.6132295, 19.2574634, 51.8032367, 88.7416552, 134.873136, 192.933569, 264.883714, 351.101769, 452.676469, 570.69655, 701.091015, 851.280998, 1034.31643, 1256.67817, 1525.59092, 1845.3419, 2223.16467, 2671.28522, 3199.63793, 3804.7686, 4493.1405, 5320.30838, 6314.05807, 7453.69257, 8822.93944, 10466.3671, 12428.5438, 14754.0379, 18864.0562, 23082.369, 27109.3853, 30645.5136, 33391.1629, 33979.8747, 33890.1697, 33279.6857, 32704.0885, 32163.4628, 31492.0484, 30698.2223, 29790.3615, 28776.8431, 27666.0441, 25605.3812, 24138.2476, 23029.4446, 22043.7737, 20946.0363, 19501.0341, 17473.5683, 15726.1345, 14440.6557},
        {29.9100206, 30.0952144, 30.2202671, 30.2868467, 30.296621, 30.2512581, 30.1524258, 30.001792, 29.8010246, 29.5517915, 29.2557606, 28.9145999, 28.5299772, 28.1035604, 27.6370175, 27.1320163, 26.5902248, 26.0133109, 25.4029425, 24.7607874, 24.0885136, 23.387789, 22.6602815, 21.907659, 21.1315894, 20.3337407, 19.5157806, 18.6793772, 17.8261983, 16.9579119, 16.0761858, 15.1826879, 14.2790863, 13.3670486, 12.4611888, 11.5564294, 10.6475375, 9.73576112, 8.82234786, 7.90854556, 6.99560202, 6.08476503, 5.17728238, 4.27440185, 3.37737125, 2.48743835, 1.60585096, 0.733856859, -0.127296156, -0.976360297, -1.81208777, -2.59696195, -3.36679012, -4.12111073, -4.85946221, -5.58138303, -6.28641162, -6.97408643, -7.64394591, -8.2955285, -8.92837265, -9.54201681, -10.1359994, -10.7098589, -11.2631338, -11.7953624, -12.3060833, -12.8400767, -13.4410367, -14.0151938, -14.5591536, -15.0695217, -15.5429038, -15.9759054, -16.3651321, -16.7071895, -16.9986833, -17.236219, -17.4164023, -17.5358387, -17.5911338, -17.5788933, -17.4957227, -17.3382277, -16.4550829, -15.1804878, -13.8564293, -12.5005398, -11.1304514, -9.76379627, -8.41820669, -7.11131483, -5.86075291, -4.68415314, -3.59914771, -2.62336883, -1.77444872, -1.07001957, -0.527713598, -0.165163006, 0, -2.26073499, -4.55391995, -6.88040664, -9.24104681, -11.6366922, -14.0681947, -16.5364059, -19.0421776, -21.5863616, -24.1698097, -26.7933735, -29.457905, -32.1642557, -34.9132775, -37.7058222, -40.5427415, -43.2370996, -45.7629215, -48.5458165, -51.5463961, -54.725272, -58.0430556, -61.4603585, -64.9377925, -68.4359689, -71.9154994, -75.3369956, -78.661069, -81.8483313, -84.901689, -87.7239506, -90.2630389, -92.4668771, -94.4198151, -96.0015624, -97.0554483, -97.4884677, -97.2076157, -96.1198872, -94.1322773, -92.5703611, -89.4908116, -84.5230757, -77.2966, -67.4408314, -56.3866478, -41.5117068, -21.9950974, 2.98409144, 34.2467709, 72.0451513, 118.463923, 173.763698, 239.040139, 317.093808, 409.959801, 519.673214, 646.004396, 793.767908, 970.248941, 1176.17621, 1424.14905, 1724.82245, 2086.85212, 2515.50613, 3017.60753, 3584.43845, 4218.7928, 4980.44796, 5896.94823, 6957.75428, 8234.35309, 9767.47175, 11597.8374, 13766.1771, 17601.7373, 21534.5959, 25283.3504, 28566.5979, 31102.9357, 31608.4041, 31472.5288, 30843.5748, 30235.819, 29652.6462, 28944.9361, 28123.6986, 27199.9438, 26184.6818, 25088.9225, 22761.5366, 21264.6463, 20301.3586, 19574.7804, 18788.0187, 17644.1804, 15846.3725, 14126.5556, 12961.5158},
        {31.9215462, 32.0603583, 32.142305, 32.1689025, 32.1416673, 32.0621155, 31.9317634, 31.7521275, 31.524724, 31.2510691, 30.9326792, 30.5710707, 30.1677597, 29.7242627, 29.2420959, 28.7227756, 28.1678181, 27.5787397, 26.9570568, 26.3042855, 25.6219424, 24.9115435, 24.1746053, 23.4126441, 22.6271761, 21.8197177, 20.9917851, 20.1448947, 19.2805627, 18.4003055, 17.5056394, 16.5980807, 15.6791456, 14.7503506, 13.8172551, 12.8792689, 11.9357528, 10.9880921, 10.0376717, 9.08587683, 8.13409252, 7.18370387, 6.23609597, 5.29265391, 4.35476278, 3.42380767, 2.50117367, 1.58824587, 0.686409363, -0.202950762, -1.07844941, -1.89587206, -2.6975915, -3.48315113, -4.25209429, -5.00396436, -5.73830472, -6.45465871, -7.15256973, -7.83158112, -8.49123627, -9.13107853, -9.75065129, -10.3494979, -10.9271617, -11.4831862, -12.0171145, -12.5743492, -13.1991913, -13.7966037, -14.3631569, -14.8954214, -15.3899678, -15.8433665, -16.252188, -16.613003, -16.9223818, -17.1768951, -17.3731133, -17.507607, -17.5769467, -17.5777029, -17.5064462, -17.359747, -16.4793359, -15.2029772, -13.8763057, -12.5171425, -11.1433092, -9.77262708, -8.42291739, -7.1120015, -5.85770076, -4.67783649, -3.59023003, -2.61270272, -1.76307588, -1.05917086, -0.51880898, -0.159811583, 0, -2.71483777, -5.38191853, -8.00746835, -10.5977133, -13.1588794, -15.6971928, -18.2188795, -20.7301656, -23.2372771, -25.7464402, -28.2638808, -30.7958251, -33.3484991, -35.9281288, -38.5409404, -41.1931599, -43.6804587, -45.9789836, -48.5712153, -51.418261, -54.481228, -57.7212233, -61.0993544, -64.5767282, -68.1144521, -71.6736333, -75.2153789, -78.7007961, -82.0909922, -85.4190308, -88.5476786, -91.4164559, -93.964883, -96.205148, -98.0404372, -99.3545589, -100.065233, -100.09018, -99.3471185, -97.7537697, -96.4744856, -93.8009438, -89.4069578, -82.966341, -74.1529072, -65.3934682, -52.9205896, -35.7198353, -12.7767692, 16.9230446, 55.0947223, 101.195633, 152.939081, 210.557493, 279.4265, 362.612539, 463.182048, 585.1521, 730.422443, 899.855769, 1088.03044, 1312.94231, 1590.81421, 1931.67479, 2335.05696, 2805.31137, 3330.35579, 3910.28048, 4605.42946, 5442.06383, 6414.37236, 7585.62264, 8993.17769, 10674.4005, 12666.6542, 16198.1518, 19817.6937, 23264.9306, 26279.5128, 28601.0911, 29044.5339, 28890.795, 28275.8599, 27660.6216, 27052.4566, 26323.3633, 25487.4869, 24558.9729, 23551.9667, 22480.6135, 20148.0994, 18698.758, 17832.9855, 17251.1779, 16653.7314, 15741.042, 14213.5059, 12559.1897, 11443.4131},
        {33.5062484, 33.7042759, 33.8404364, 33.9163678, 33.933708, 33.894095, 33.7991666, 33.6505608, 33.4499154, 33.1988684, 32.8990577, 32.5521212, 32.1596967, 31.7234223, 31.2449358, 30.7258751, 30.1678782, 29.5725829, 28.9416271, 28.2766488, 27.5792859, 26.8511762, 26.0939577, 25.3092683, 24.4987459, 23.6640284, 22.8067537, 21.9285598, 21.0310845, 20.1159657, 19.1848413, 18.2393494, 17.2811277, 16.3118141, 15.3369799, 14.3562136, 13.3690565, 12.3770187, 11.3816107, 10.3843426, 9.38672481, 8.39026749, 7.39648093, 6.4068754, 5.42296118, 4.44624853, 3.47824771, 2.52046899, 1.57442265, 0.64161894, -0.276431863, -1.12715609, -1.96121384, -2.77820179, -3.57771663, -4.35935505, -5.12271373, -5.86738937, -6.59297864, -7.29907824, -7.98528486, -8.65119517, -9.29640588, -9.92051366, -10.5231152, -11.1038072, -11.6621863, -12.2461342, -12.9023731, -13.5309158, -14.1282287, -14.6907785, -15.2150317, -15.6974549, -16.1345147, -16.5226776, -16.8584103, -17.1381792, -17.3584511, -17.5156923, -17.6063696, -17.6269495, -17.5738985, -17.4436833, -16.5691882, -15.2926284, -13.964195, -12.601889, -11.2237115, -9.84766352, -8.49174624, -7.17396072, -5.91230805, -4.72478932, -3.62940562, -2.64415803, -1.78704765, -1.07607557, -0.529242873, -0.164550653, 0, -3.04218122, -5.97673011, -8.813927, -11.5640522, -14.2373861, -16.844209, -19.3948011, -21.8994429, -24.3684146, -26.8119967, -29.2404693, -31.6641128, -34.0932076, -36.538034, -39.0088723, -41.5160028, -43.833091, -45.9366152, -48.3720558, -51.09899, -54.0769951, -57.265648, -60.624526, -64.1132063, -67.6912658, -71.3182819, -74.9538316, -78.5574921, -82.0888404, -85.5517125, -88.8451985, -91.9155982, -94.709211, -97.3123319, -99.6012627, -101.414973, -102.657765, -103.233938, -103.047793, -102.003633, -100.880059, -98.4409786, -94.419634, -88.5492669, -80.5631193, -73.4541705, -62.8809906, -47.761887, -27.0151672, 0.440861219, 37.0925028, 80.9000901, 128.081544, 178.345775, 238.529131, 312.267888, 403.198319, 518.056774, 656.674889, 815.502405, 985.222944, 1186.59963, 1438.47013, 1751.33179, 2121.54977, 2551.97615, 3029.27182, 3552.09116, 4178.45612, 4932.807, 5812.05773, 6872.40977, 8147.83178, 9672.29241, 11479.7603, 14691.2583, 17981.9373, 21114.0024, 23849.6586, 25951.1111, 26338.0335, 26176.85, 25590.8505, 24989.9512, 24383.986, 23663.361, 22843.8816, 21941.3529, 20971.5801, 19950.3685, 17880.6509, 16564.5478, 15750.8576, 15188.3788, 14625.9099, 13812.2494, 12496.1957, 11006.1739, 9947.94415},
        {34.2768098, 34.721386, 35.0845205, 35.3684503, 35.5754121, 35.7076428, 35.7673793, 35.7568584, 35.678317, 35.5339919, 35.32612, 35.0569381, 34.728683, 34.3435917, 33.903901, 33.4118477, 32.8696688, 32.2796009, 31.6438811, 30.964746, 30.2444327, 29.485178, 28.6892186, 27.8587915, 26.9961336, 26.1034815, 25.1830723, 24.2371428, 23.2679298, 22.2776701, 21.2686007, 20.2429584, 19.20298, 18.1509023, 17.1076762, 16.065788, 15.0174989, 13.9644385, 12.9082364, 11.8505221, 10.7929252, 9.73707526, 8.68460175, 7.63713428, 6.59630237, 5.56373559, 4.54106349, 3.52991562, 2.53192152, 1.54871075, 0.581912859, -0.303747449, -1.17110356, -2.01989369, -2.84985609, -3.66072897, -4.45225057, -5.22415911, -5.97619282, -6.70808994, -7.41958868, -8.11042729, -8.78034399, -9.429077, -10.0563646, -10.6619449, -11.2455562, -11.8607411, -12.5597501, -13.2313543, -13.871804, -14.4773494, -15.0442409, -15.5687288, -16.0470635, -16.4754951, -16.8502742, -17.1676509, -17.4238756, -17.6151986, -17.7378702, -17.7881407, -17.7622606, -17.6564799, -16.7923442, -15.5176478, -14.1882943, -12.8224296, -11.4381994, -10.0537496, -8.68722589, -7.35677413, -6.08054007, -4.87666951, -3.76330825, -2.75860207, -1.88069675, -1.1477381, -0.577871897, -0.189243934, 0, -3.05868787, -6.00032168, -8.83606603, -11.5770856, -14.2345449, -16.8196087, -19.3434415, -21.8172079, -24.2520727, -26.6592003, -29.0497555, -31.4349029, -33.825807, -36.2336325, -38.6695441, -41.1447063, -43.3949508, -45.3921091, -47.7701225, -50.4811558, -53.4773736, -56.7109407, -60.1340218, -63.6987816, -67.3573848, -71.0619962, -74.7647804, -78.4179022, -81.9735264, -85.2773203, -88.426995, -91.4066642, -94.2004419, -97.2336783, -100.269869, -102.954847, -105.140355, -106.678135, -107.419932, -107.217488, -106.154591, -103.780317, -99.9010078, -94.3230065, -86.8526558, -80.1266333, -70.4133632, -56.8116046, -38.4201165, -14.3376579, 17.2819946, 55.2401209, 96.6041666, 141.289406, 195.025429, 260.931317, 342.126151, 444.327998, 567.466274, 708.635136, 859.921077, 1039.17362, 1262.41455, 1538.717, 1865.38006, 2245.1073, 2665.6233, 3125.74816, 3679.22507, 4348.72964, 5135.9072, 6087.69903, 7234.06042, 8604.94665, 10230.313, 13120.689, 16080.3122, 18893.9943, 21346.5471, 23222.7824, 23541.9784, 23364.836, 22804.2525, 22237.5243, 21672.8909, 21008.4259, 20257.8771, 19434.9926, 18553.5202, 17627.208, 16108.7576, 15025.5116, 14219.1928, 13531.5236, 12804.2267, 11879.0247, 10597.6402, 9446.48531, 8544.38444},
        {34.7165028, 35.4902133, 36.1561615, 36.7174171, 37.1770502, 37.5381307, 37.8037285, 37.9769134, 38.0607555, 38.0583245, 37.9726905, 37.8069233, 37.5640929, 37.2472691, 36.8595219, 36.4039212, 35.8835369, 35.301439, 34.6606972, 33.9643816, 33.215562, 32.4173084, 31.5726907, 30.6847787, 29.7566425, 28.7913518, 27.7919767, 26.7615871, 25.7032528, 24.6200437, 23.5150298, 22.3912811, 21.2518673, 20.0998584, 18.9815977, 17.8776081, 16.7678927, 15.6541171, 14.5379468, 13.4210475, 12.3050848, 11.1917243, 10.0826316, 8.97947234, 7.88391212, 6.79761653, 5.72225121, 4.65948175, 3.61097377, 2.57839288, 1.56340468, 0.634880884, -0.274175891, -1.16355693, -2.03305351, -2.88245692, -3.71155845, -4.52014939, -5.30802101, -6.0749646, -6.82077146, -7.54523286, -8.24814008, -8.92928443, -9.58845717, -10.2254496, -10.840053, -11.4884523, -12.2254786, -12.9346153, -13.611998, -14.2537622, -14.8560434, -15.4149771, -15.9266988, -16.3873441, -16.7930484, -17.1399473, -17.4241763, -17.6418708, -17.7891665, -17.8621988, -17.8571031, -17.7700152, -16.9165618, -15.6474463, -14.3215514, -12.9570963, -11.5723004, -10.185383, -8.81456348, -7.47806108, -6.19409516, -4.98088504, -3.85665005, -2.83960951, -1.94798275, -1.1999891, -0.613847869, -0.207778395, 0, -3.03726419, -5.95540377, -8.76593576, -11.4803772, -14.1102451, -16.6670565, -19.1623284, -21.6075779, -24.014322, -26.3940776, -28.7583619, -31.1186918, -33.4865844, -35.8735568, -38.2911258, -40.7508087, -42.9683368, -44.9129723, -47.2652706, -49.9733662, -52.9853935, -56.2494869, -59.7137809, -63.3264099, -67.0355084, -70.7892106, -74.5356512, -78.2229645, -81.7992849, -84.9990784, -88.0624929, -91.0017637, -93.8291256, -97.1984582, -100.801174, -104.157581, -107.087421, -109.410435, -110.946365, -111.514953, -110.674233, -108.585304, -105.119111, -100.146599, -93.5387145, -87.3830485, -78.7797379, -67.0455662, -51.497317, -31.4517737, -5.91553252, 25.1053179, 59.9094217, 98.7312281, 145.757455, 203.400334, 274.072093, 361.869801, 467.137932, 588.382956, 719.94818, 875.759795, 1067.35156, 1302.06979, 1580.16543, 1905.14634, 2268.29298, 2669.94314, 3155.26828, 3744.07941, 4437.42188, 5276.75936, 6288.39867, 7498.6466, 8933.80999, 11494.8078, 14114.6816, 16601.0852, 18761.6724, 20404.0968, 20646.589, 20445.8414, 19904.9751, 19373.257, 18856.3752, 18257.4567, 17587.3173, 16856.7733, 16076.6405, 15257.7349, 14324.3599, 13480.0136, 12672.5987, 11850.0184, 10960.1755, 9950.97318, 8770.31431, 7920.26006, 7153.87273},
        {35.1893876, 36.3581065, 37.3850388, 38.2744066, 39.030432, 39.6573371, 40.159344, 40.5406749, 40.8055519, 40.9581971, 41.0028327, 40.9436806, 40.7849632, 40.5309024, 40.1857204, 39.7536393, 39.2388812, 38.6456684, 37.9782228, 37.2407666, 36.4375219, 35.5727108, 34.6505555, 33.6752781, 32.6511006, 31.5822453, 30.4729342, 29.3273894, 28.1498331, 26.9444874, 25.7155744, 24.4673163, 23.203935, 21.9296529, 20.7348874, 19.57895, 18.420116, 17.2598104, 16.0994579, 14.9404836, 13.7843123, 12.632369, 11.4860785, 10.3468658, 9.21615574, 8.09537328, 6.98594332, 5.88929076, 4.8068405, 3.74001747, 2.69024656, 1.70496186, 0.738581717, -0.20846635, -1.13575482, -2.04285618, -2.9293429, -3.79478746, -4.63876235, -5.46084005, -6.26059303, -7.03759377, -7.79141476, -8.52162847, -9.22780739, -9.90952399, -10.5663508, -11.2479618, -12.0028285, -12.7271922, -13.4173776, -14.0697093, -14.6805118, -15.2461098, -15.7628279, -16.2269906, -16.6349225, -16.9829482, -17.2673924, -17.4845796, -17.6308344, -17.7024814, -17.6958452, -17.6072504, -16.7575946, -15.4951227, -14.1761975, -12.8190624, -11.441961, -10.0631367, -8.70083303, -7.3732935, -6.0987616, -4.89548081, -3.78169463, -2.77564656, -1.89558008, -1.15973869, -0.586365888, -0.193705159, 0, -3.25809013, -6.35918336, -9.31733251, -12.1465904, -14.8610099, -17.4746437, -20.0015448, -22.4557659, -24.8513598, -27.2023795, -29.5228777, -31.8269072, -34.1285208, -36.4417715, -38.7807119, -41.159395, -43.3063679, -45.1948968, -47.4848213, -50.1290144, -53.0803491, -56.2916985, -59.7159356, -63.3059334, -67.014565, -70.7947033, -74.5992216, -78.3809927, -82.0928896, -85.5519502, -88.896689, -92.1207297, -95.217696, -98.5904001, -102.027871, -105.210022, -108.00772, -110.291831, -111.933224, -112.802764, -112.637183, -111.482308, -109.235251, -105.793123, -101.053035, -96.6355778, -90.2835133, -81.4630822, -69.6405256, -54.2820845, -34.8744014, -10.8383813, 16.8079034, 48.4378755, 87.2659274, 135.230914, 194.271691, 267.841375, 356.020409, 457.23731, 566.551308, 695.451549, 850.460129, 1037.8148, 1263.2715, 1532.80732, 1846.22426, 2207.89844, 2644.84812, 3171.2839, 3773.54364, 4497.99878, 5368.50816, 6408.93061, 7643.12498, 9872.41622, 12148.8305, 14301.8602, 16160.9977, 17555.7351, 17696.8473, 17445.0225, 16894.9178, 16371.4666, 15878.1749, 15322.5998, 14713.0445, 14057.8122, 13365.2062, 12643.5296, 12105.3158, 11418.405, 10625.2182, 9768.17646, 8889.70083, 8032.21234, 7238.13201, 6459.94745, 5735.85649},
        {35.0722149, 36.6340657, 38.0242033, 39.2477918, 40.3099958, 41.2159795, 41.9709073, 42.5799436, 43.0482527, 43.3809989, 43.5833466, 43.6604602, 43.617504, 43.4596423, 43.1920395, 42.81986, 42.3482681, 41.7824282, 41.1275046, 40.3886616, 39.5710637, 38.6798751, 37.7202602, 36.6973834, 35.616409, 34.4825014, 33.3008249, 32.0765439, 30.8148227, 29.5208257, 28.1997172, 26.8566616, 25.4968232, 24.1253664, 22.8651527, 21.6600223, 20.4524007, 19.2436328, 18.0350633, 16.8280372, 15.6238993, 14.4239945, 13.2296675, 12.0422633, 10.8631267, 9.69360255, 8.53503569, 7.38877097, 6.25615324, 5.13852737, 4.03723819, 2.9832184, 1.9475836, 0.931037239, -0.065717247, -1.04197641, -1.99703681, -2.930195, -3.84074753, -4.72799097, -5.59122187, -6.42973678, -7.24283227, -8.02980489, -8.78995119, -9.52256773, -10.2269511, -10.9435634, -11.7118787, -12.4462922, -13.1434321, -13.7999262, -14.4124024, -14.9774888, -15.4918134, -15.9520039, -16.3546885, -16.6964951, -16.9740515, -17.1839858, -17.322926, -17.3874999, -17.3743355, -17.2800609, -16.4422758, -15.200984, -13.9041103, -12.5696694, -11.215676, -9.86014468, -8.52109018, -7.21652719, -5.96447037, -4.78293441, -3.68993399, -2.70348379, -1.8415985, -1.12229279, -0.563581353, -0.183478862, 0, -3.18969988, -6.23094607, -9.13723049, -11.9220451, -14.5988817, -17.1812324, -19.6825891, -22.1164436, -24.4962879, -26.8356139, -29.1479136, -31.4466789, -33.7454016, -36.0575739, -38.3966874, -40.7762343, -42.9863415, -45.0085916, -47.3617067, -50.011315, -52.9230443, -56.0625227, -59.3953783, -62.887239, -66.5037328, -70.2104878, -73.973132, -77.7572934, -81.5286, -85.2299182, -88.8579836, -92.3852526, -95.7841818, -99.1755778, -102.457722, -105.489336, -108.198371, -110.512779, -112.360511, -113.669518, -114.39281, -114.425654, -113.6911, -112.112196, -109.611992, -107.522079, -104.004828, -98.6311525, -90.9719666, -80.5981837, -67.2217635, -50.1001838, -29.8217764, -5.98691829, 23.821087, 61.1133112, 107.400826, 165.445691, 235.223675, 315.345829, 401.840522, 503.678372, 623.298777, 765.773753, 940.636138, 1155.37118, 1416.14441, 1730.09892, 2109.21679, 2562.74291, 3064.86031, 3665.26985, 4385.35321, 5246.49206, 6270.06808, 8153.90385, 10073.8299, 11882.1177, 13431.0388, 14572.8645, 14615.1501, 14318.0274, 13767.2176, 13262.8498, 12805.7755, 12307.5092, 11773.3693, 11208.6739, 10618.7412, 10008.8895, 9792.63828, 9252.4493, 8504.19509, 7663.74819, 6846.98115, 6169.76652, 5747.97685, 5025.74918, 4342.07353},
        {33.2162295, 34.8606089, 36.3472735, 37.6804866, 38.864512, 39.903613, 40.8020531, 41.564096, 42.1940052, 42.696044, 43.0744761, 43.333565, 43.4775742, 43.5107672, 43.4374076, 43.2617587, 42.9880843, 42.6206477, 42.1637125, 41.6215423, 40.9984004, 40.2985506, 39.5262562, 38.6857807, 37.7813878, 36.8173409, 35.7979036, 34.7273393, 33.6099116, 32.4498839, 31.2515199, 30.019083, 28.7568367, 27.4690446, 26.2181896, 24.9782008, 23.7223079, 22.4528849, 21.1723062, 19.8829459, 18.5871781, 17.2873772, 15.9859172, 14.6851723, 13.3875167, 12.0953247, 10.8109704, 9.53682797, 8.27527163, 7.02867556, 5.79941394, 4.65915343, 3.53947362, 2.44124658, 1.36534437, 0.312639069, -0.715997276, -1.7196926, -2.69757483, -3.64877192, -4.57241179, -5.46762238, -6.33353164, -7.16926749, -7.97395788, -8.74673074, -9.486714, -10.2227231, -10.9828606, -11.7050264, -12.3864237, -13.0242561, -13.6157269, -14.1580396, -14.6483976, -15.0840043, -15.4620631, -15.7797775, -16.0343508, -16.2229865, -16.342888, -16.3912587, -16.3653021, -16.2622215, -15.4921734, -14.3582325, -13.1705292, -11.9453178, -10.6988525, -9.44738749, -8.20717702, -6.99447529, -5.82553654, -4.71661498, -3.68396483, -2.74384032, -1.91249566, -1.20618508, -0.641162794, -0.233683027, 0, -0.64634697, -1.553406, -2.70828577, -4.09809495, -5.70994224, -7.53093631, -9.54818584, -11.7487995, -14.119886, -16.648554, -19.3219122, -22.1270692, -25.0511338, -28.0812146, -31.2044204, -34.4078597, -37.7874873, -41.3459613, -44.824191, -48.2326094, -51.5816491, -54.8817431, -58.1433241, -61.376825, -64.5926787, -67.8013179, -71.0131754, -74.2386842, -77.4882769, -80.8060692, -84.1564608, -87.5397799, -90.9563544, -94.7793628, -98.822708, -102.800866, -106.60231, -110.115513, -113.228947, -115.831087, -117.054283, -117.773254, -118.024409, -117.844158, -117.268911, -118.679967, -119.182624, -118.227068, -115.263488, -109.742069, -100.578237, -88.4105405, -73.9181249, -57.0046249, -35.990343, -9.90646573, 22.2158205, 61.1351888, 108.078352, 164.501268, 229.934886, 309.149638, 404.861448, 520.208859, 659.046512, 824.90035, 1016.98019, 1237.69297, 1505.27144, 1828.99635, 2189.56519, 2629.23946, 3167.19413, 3822.60421, 4614.64469, 6146.13254, 7706.69027, 9169.58236, 10408.0733, 11295.4276, 11222.0588, 10865.9829, 10307.4007, 9844.86444, 9472.14394, 9092.02885, 8702.83781, 8302.88946, 7890.50247, 7463.99546, 7860.94077, 7605.09083, 6910.44997, 5991.02253, 5060.81284, 4333.82525, 4024.06408, 3570.82463, 2979.04932},
        {36.6432225, 39.9550389, 42.9619825, 45.6731549, 48.0976578, 50.2445928, 52.1230617, 53.742166, 55.1110074, 56.2386877, 57.1343083, 57.8069711, 58.2657776, 58.5198296, 58.5782286, 58.4500764, 58.1444746, 57.6705248, 57.0373287, 56.2539881, 55.3296044, 54.2732795, 53.0941149, 51.8012123, 50.4036734, 48.9105998, 47.3310932, 45.6742553, 43.9491877, 42.164992, 40.33077, 38.4556232, 36.5486534, 34.6189622, 32.8791199, 31.2322142, 29.5788871, 27.9216374, 26.2629638, 24.605365, 22.9513397, 21.3033866, 19.6640044, 18.0356919, 16.4209478, 14.8222707, 13.2421594, 11.6831126, 10.1476289, 8.63820722, 7.15734612, 5.78528935, 4.44510526, 3.13760719, 1.86360852, 0.623922574, -0.580637281, -1.7492577, -2.88112533, -3.97542681, -5.0313488, -6.04807795, -7.0248009, -7.96070431, -8.85497481, -9.70679907, -10.5153637, -11.3013794, -12.0850398, -12.8211402, -13.5074719, -14.1418261, -14.7219944, -15.2457679, -15.710938, -16.115296, -16.4566334, -16.7327414, -16.9414114, -17.0804347, -17.1476027, -17.1407066, -17.0575379, -16.8958878, -16.040013, -14.8073769, -13.5228175, -12.2040365, -10.8687356, -9.53461649, -8.21938095, -6.94073068, -5.71636742, -4.56399287, -3.50130877, -2.54601684, -1.7158188, -1.02841637, -0.501511277, -0.152805247, 0, -3.28034312, -6.39804729, -9.36732413, -12.2023853, -14.9174424, -17.526707, -20.0443909, -22.4847056, -24.8618628, -27.1900741, -29.4835511, -31.7565054, -34.0231488, -36.2976928, -38.594349, -40.9273292, -43.1148889, -45.1432462, -47.4698463, -50.0669103, -52.9066592, -55.9613141, -59.2030962, -62.6042264, -66.1369259, -69.7734157, -73.485917, -77.2466507, -81.027838, -84.6858371, -88.3512149, -92.0309515, -95.7320267, -99.6909624, -103.799968, -107.89004, -111.899297, -115.765856, -119.427835, -122.82335, -125.477171, -127.866566, -130.010523, -131.928033, -133.638086, -135.216915, -136.611955, -137.827884, -138.86938, -139.741122, -141.324473, -141.675923, -140.60321, -137.569615, -131.335147, -120.975573, -105.566655, -83.7547288, -55.1425866, -19.8014928, 21.7440472, 72.050378, 131.905986, 202.958544, 288.31157, 390.400325, 507.904976, 642.28724, 808.777515, 1014.81785, 1254.1662, 1550.05396, 1914.66, 2360.16318, 2898.74236, 3934.96013, 4990.5157, 5979.49198, 6815.9719, 7414.03839, 7361.70805, 7116.50769, 6732.26294, 6406.71415, 6136.72071, 5859.17758, 5573.94238, 5280.87275, 4979.82632, 4670.66071, 4797.34813, 4584.75702, 4141.56232, 3576.43901, 2998.06202, 2515.10633, 2236.24689, 1973.98466, 1531.57087},
        {32.139939, 35.5369083, 38.6420056, 41.4636489, 44.0102563, 46.2902458, 48.3120353, 50.0840429, 51.6146867, 52.9123845, 53.9855544, 54.8426145, 55.4919827, 55.942077, 56.2013155, 56.2781161, 56.1808969, 55.918076, 55.4980712, 54.9293006, 54.2201822, 53.3791341, 52.4145741, 51.3349205, 50.1485911, 48.8640039, 47.489577, 46.0337285, 44.5048762, 42.9114382, 41.2618325, 39.5644772, 37.8277902, 36.0601896, 34.4469342, 32.904302, 31.3464455, 29.7760438, 28.1957762, 26.6083218, 25.0163598, 23.4225693, 21.8296296, 20.2402199, 18.6570193, 17.082707, 15.5199622, 13.971464, 12.4398917, 10.9279245, 9.43824148, 8.03981, 6.6675841, 5.32280599, 4.00671783, 2.72056182, 1.46558015, 0.243014987, -0.94589147, -2.09989704, -3.21775953, -4.29823676, -5.34008655, -6.34206671, -7.30293505, -8.22144938, -9.09636753, -9.93989075, -10.7638975, -11.5394196, -12.2643433, -12.9365549, -13.5539407, -14.1143872, -14.6157806, -15.0560072, -15.4329535, -15.7445057, -15.9885501, -16.1629731, -16.265661, -16.2945002, -16.2473769, -16.1221775, -15.3584587, -14.2450154, -13.0773531, -11.871477, -10.6433923, -9.40910422, -8.18461798, -6.98593875, -5.82907175, -4.73002217, -3.70479522, -2.7693961, -1.93983001, -1.23210216, -0.662217736, -0.246181949, 0, -2.96057998, -5.73424836, -8.33765029, -10.7874309, -13.1002354, -15.2927089, -17.3814966, -19.3832436, -21.3145952, -23.1921963, -25.0326923, -26.8527282, -28.6689493, -30.4980006, -32.3565273, -34.2611747, -36.1304381, -37.9669415, -40.009335, -42.2532318, -44.694245, -47.3279876, -50.1500727, -53.1561135, -56.341723, -59.7025144, -63.2341006, -66.9320949, -70.7921103, -74.8544592, -79.0536657, -83.3719329, -87.7914642, -92.2183283, -96.6727957, -101.195439, -105.791303, -110.465431, -115.222865, -120.068651, -125.137099, -130.264643, -135.431035, -140.616027, -145.79937, -150.896982, -155.968409, -161.009361, -166.015549, -170.982683, -176.216516, -181.023776, -185.632495, -189.88047, -192.808757, -193.816136, -192.301382, -187.504008, -179.018252, -166.612098, -150.28166, -128.691758, -101.818068, -68.6683414, -26.6102426, 26.235738, 89.5030478, 164.488884, 260.726, 383.088654, 531.468458, 716.488874, 944.748922, 1222.84762, 1557.38399, 2189.77563, 2833.09833, 3435.24647, 3944.11441, 4307.59654, 4277.29598, 4128.25918, 3892.50532, 3685.68985, 3506.72593, 3319.67832, 3125.20263, 2923.95447, 2716.58944, 2503.76315, 2283.33515, 2059.46958, 1833.50758, 1606.79031, 1380.65894, 1156.45463, 935.518528, 727.026283, 483.59234},
        {31.6116123, 36.1689115, 40.3598168, 44.1945866, 47.683479, 50.8367526, 53.6646656, 56.1774764, 58.3854434, 60.2988248, 61.9278791, 63.2828646, 64.3740397, 65.2116626, 65.8059918, 66.1672856, 66.3058023, 66.2318004, 65.9555381, 65.4872738, 64.8372658, 64.0157725, 63.0330523, 61.8993635, 60.6249644, 59.2201134, 57.6950689, 56.0600891, 54.3254325, 52.5013574, 50.5981222, 48.6259851, 46.5952046, 44.516039, 42.5796926, 40.7021449, 38.7972004, 36.8692454, 34.9226662, 32.9618491, 30.9911804, 29.0150465, 27.0378335, 25.0639279, 23.097716, 21.1435839, 19.2059181, 17.2891049, 15.3975304, 13.5355812, 11.7076433, 10.0486591, 8.42962882, 6.85210846, 5.31765418, 3.82782209, 2.38416829, 0.988248894, -0.358379993, -1.65416226, -2.89754181, -4.08696252, -5.22086828, -6.297703, -7.31591056, -8.27393486, -9.17021978, -9.97588139, -10.6626919, -11.2854567, -11.8443914, -12.3397113, -12.771632, -13.1403687, -13.4461371, -13.6891525, -13.8696304, -13.9877863, -14.0438355, -14.0379935, -13.9704758, -13.8414977, -13.6512749, -13.4000226, -12.7642597, -11.9128585, -11.0185819, -10.0921499, -9.14428244, -8.18569962, -7.22712144, -6.27926793, -5.35285911, -4.458615, -3.60725563, -2.80950101, -2.07607116, -1.41768612, -0.845065893, -0.368930513, 0, -0.0321096849, -0.216258288, -0.550987934, -1.03484075, -1.66635885, -2.44408438, -3.36655944, -4.43232617, -5.63992669, -6.98790313, -8.47479761, -10.0991523, -11.8595092, -13.7544105, -15.7823984, -17.942015, -20.3401688, -22.9908831, -25.647585, -28.3320383, -31.0660063, -33.8712526, -36.7695408, -39.7826343, -42.9322966, -46.2402914, -49.7283821, -53.4183322, -57.3319053, -61.5918638, -66.0819395, -70.7935959, -75.7182971, -80.7727976, -85.9859159, -91.4063928, -97.0481051, -102.924929, -109.050742, -115.43942, -122.063831, -128.991342, -136.243852, -143.84326, -151.811467, -160.128912, -168.869319, -178.064956, -187.748086, -197.950976, -208.744382, -220.07503, -232.040874, -244.658666, -257.77938, -271.328422, -285.231198, -299.454164, -313.872345, -328.315986, -342.763131, -356.885644, -370.299193, -382.787028, -394.416354, -405.124038, -415.668256, -426.198808, -433.854012, -437.475653, -437.79868, -431.48045, -416.707421, -391.666046, -354.54278, -254.605197, -151.188354, -54.7084283, 24.4184052, 75.7759704, 49.668391, 1.14565862, -63.3742454, -120.987319, -171.801297, -222.793089, -273.726973, -324.367224, -374.478119, -423.823935, -439.081398, -477.313406, -530.616936, -591.08897, -650.826486, -701.926463, -736.485883, -755.714371, -806.256731}
    };
    
    const double b4tab[98][200] = {{-186.925396, -184.835303, -182.761961, -180.70479, -178.663211, -176.636644, -174.62451, -172.62623, -170.641225, -168.668915, -166.708722, -164.760064, -162.822364, -160.895043, -158.97752, -157.069216, -155.169553, -153.27795, -151.393829, -149.51661, -147.645713, -145.780561, -143.920573, -142.065169, -140.213772, -138.3658, -136.520676, -134.677819, -132.836651, -130.996592, -129.157062, -127.317483, -125.477276, -123.63586, -121.752887, -119.8485, -117.943317, -116.038051, -114.133412, -112.230112, -110.328862, -108.430373, -106.535357, -104.644524, -102.758587, -100.878256, -99.0042424, -97.1372579, -95.2780137, -93.4272209, -91.5855908, -89.789941, -88.0040937, -86.2279776, -84.4615212, -82.704653, -80.9573018, -79.2193959, -77.490864, -75.7716347, -74.0616364, -72.3607978, -70.6690475, -68.9863139, -67.3125258, -65.6476115, -63.9914997, -62.3854729, -60.8698206, -59.359182, -57.850805, -56.3419373, -54.8298267, -53.3117209, -51.7848676, -50.2465147, -48.6939099, -47.124301, -45.5349356, -43.9230617, -42.2859269, -40.6207789, -38.9248657, -37.1954348, -35.1323921, -32.8879105, -30.6177368, -28.3287681, -26.0279014, -23.7220339, -21.4180626, -19.1228847, -16.8433971, -14.5864969, -12.3590814, -10.1680474, -8.02029218, -5.92271275, -3.88220618, -1.90566957, 0, 0.864827967, 1.72997818, 2.59631864, 3.46471734, 4.33604229, 5.21116149, 6.09094293, 6.97625461, 7.86796453, 8.76694069, 9.6740511, 10.5901637, 11.5161466, 12.4528677, 13.4011951, 14.3619967, 15.1548217, 15.7546355, 16.5724318, 17.5702248, 18.7100283, 19.9538564, 21.2637229, 22.601642, 23.9296274, 25.2096932, 26.4038534, 27.474122, 28.3825128, 29.0127609, 29.4338616, 29.6313124, 29.5906109, 29.5704979, 29.4198494, 28.9146767, 27.9585046, 26.4548578, 24.3072609, 21.4192388, 18.461251, 14.3364724, 8.79837528, 1.60043184, -7.50388565, -16.7022099, -28.8146875, -44.6025698, -64.8271085, -90.249555, -121.051693, -159.28248, -204.066638, -256.326969, -320.657363, -400.003467, -497.310928, -612.03855, -751.411286, -926.45792, -1140.78358, -1407.87563, -1724.65392, -2102.68176, -2578.33488, -3176.59972, -3898.87704, -4764.03847, -5877.43634, -7295.50461, -8665.3891, -10515.7793, -13044.7858, -16450.5192, -20931.0901, -32314.7424, -43761.9202, -54063.2006, -62009.161, -66390.3786, -60900.1688, -52824.5455, -43138.6266, -35556.7613, -29969.7886, -25127.2016, -20976.9068, -17466.8106, -14544.8196, -12158.8402, -10320.8061, -8900.66198, -7828.23538, -7033.35386, -6445.84503, -5995.53647, -5612.25575, -5405.08697, -5374.82903},
        {-168.555271, -166.178371, -163.859789, -161.597488, -159.389434, -157.233589, -155.127919, -153.070388, -151.058961, -149.091601, -147.166273, -145.280941, -143.43357, -141.622125, -139.844568, -138.098865, -136.38298, -134.694878, -133.032522, -131.393878, -129.776908, -128.179579, -126.599854, -125.035697, -123.485073, -121.945946, -120.41628, -118.894041, -117.377191, -115.863696, -114.35152, -112.838627, -111.322982, -109.802549, -108.18438, -106.513809, -104.837259, -103.155646, -101.469884, -99.7808882, -98.0895726, -96.3968519, -94.7036408, -93.0108538, -91.3194056, -89.6302106, -87.9441836, -86.2622392, -84.5852918, -82.9142561, -81.2500467, -79.6330898, -78.0239317, -76.4226307, -74.8292448, -73.2438319, -71.6664503, -70.0971579, -68.5360128, -66.9830731, -65.4383968, -63.902042, -62.3740667, -60.854529, -59.343487, -57.8409988, -56.3471223, -54.8988805, -53.5324087, -52.1715279, -50.8138997, -49.4571861, -48.0990487, -46.7371494, -45.3691499, -43.9927121, -42.6054976, -41.2051683, -39.789386, -38.3558124, -36.9021093, -35.4259385, -33.9249618, -32.396841, -30.6100545, -28.6816762, -26.731535, -24.7647298, -22.7863601, -20.801525, -18.8153237, -16.8328554, -14.8592195, -12.899515, -10.9588412, -9.04229738, -7.1549827, -5.30199641, -3.48843771, -1.71940583, 0, 0.886339582, 1.79195411, 2.71588912, 3.65719014, 4.61490271, 5.58807236, 6.57574462, 7.57696503, 8.59077912, 9.61623241, 10.6523705, 11.6982388, 12.7528829, 13.8153484, 14.8846807, 15.9599255, 16.9713049, 17.9080326, 18.9248259, 20.0059823, 21.1357997, 22.2985757, 23.478608, 24.6601943, 25.8276322, 26.9652195, 28.0572538, 29.0880328, 30.0418542, 31.0209162, 31.8483858, 32.4731905, 32.8442578, 32.9676672, 32.7637702, 32.137678, 31.0211725, 29.3460359, 27.0440501, 24.046997, 21.5325568, 17.807427, 12.559627, 5.47717607, -3.75190645, -14.1403928, -27.6242742, -44.8403334, -66.4253531, -93.016116, -124.030893, -162.814271, -208.810629, -263.24634, -330.986022, -415.260797, -519.301783, -642.282932, -792.414618, -982.333224, -1216.74165, -1509.6275, -1854.41182, -2263.72911, -2782.77004, -3441.78148, -4250.46451, -5233.7393, -6491.86057, -8082.46957, -9663.09232, -11753.0415, -14548.4597, -18245.4896, -23040.2739, -34728.1977, -46506.3507, -57171.0647, -65518.6719, -70345.5043, -65467.068, -57981.0715, -48818.4867, -41434.5191, -35760.9646, -30677.8549, -26169.5739, -22220.5057, -18815.0343, -15937.5437, -13558.5224, -11775.2924, -10480.143, -9565.36358, -8923.24356, -8446.07234, -8026.1393, -7903.15278, -8205.31554},
        {-153.482925, -151.337397, -149.230949, -147.162453, -145.130779, -143.134799, -141.173385, -139.245406, -137.349736, -135.485244, -133.650803, -131.845284, -130.067557, -128.316495, -126.590968, -124.889848, -123.212007, -121.556314, -119.921643, -118.306863, -116.710847, -115.132465, -113.57059, -112.024091, -110.491842, -108.972711, -107.465573, -105.969296, -104.482753, -103.004815, -101.534354, -100.07024, -98.6113451, -97.1565404, -95.6663644, -94.1596612, -92.6557353, -91.1547021, -89.6566768, -88.1617747, -86.670111, -85.1818011, -83.6969601, -82.2157034, -80.7381463, -79.2644039, -77.7945917, -76.3288247, -74.8672184, -73.4098879, -71.9569485, -70.4983299, -69.0445537, -67.5959561, -66.1528731, -64.7156408, -63.2845953, -61.8600727, -60.442409, -59.0319403, -57.6290027, -56.2339322, -54.847065, -53.4687371, -52.0992845, -50.7390434, -49.3883499, -48.0860083, -46.8698996, -45.6610216, -44.4572166, -43.2563271, -42.0561954, -40.8546637, -39.6495744, -38.4387699, -37.2200923, -35.9913842, -34.7504878, -33.4952454, -32.2234994, -30.9330921, -29.6218658, -28.2876628, -26.7528343, -25.1066594, -23.4405265, -21.7579731, -20.0625363, -18.3577537, -16.6471626, -14.9343003, -13.2227041, -11.5159115, -9.81745978, -8.1308863, -6.45972842, -4.80752349, -3.17780888, -1.57412193, 0, 0.643421548, 1.37937144, 2.20132983, 3.10277686, 4.07719267, 5.11805742, 6.21885125, 7.3730543, 8.57414672, 9.81560866, 11.0909203, 12.3935617, 13.717013, 15.0547545, 16.4002662, 17.7470284, 19.1216924, 20.5224771, 21.8678324, 23.1583466, 24.3946078, 25.5772045, 26.7067248, 27.7837571, 28.8088897, 29.7827109, 30.7058089, 31.578772, 32.4021885, 33.3675054, 34.2144708, 34.8864154, 35.3266699, 35.5770558, 35.5316584, 35.0582989, 34.0707606, 32.4828271, 30.2082818, 27.1609083, 24.4179703, 20.3756682, 14.7201481, 7.13755586, -2.68596243, -13.1551319, -26.9702176, -44.9223559, -67.8026829, -96.4023348, -130.655168, -173.257386, -222.777403, -280.267405, -351.850604, -441.373423, -552.682287, -686.103454, -849.803551, -1055.78939, -1305.66937, -1620.15482, -1999.37619, -2459.28503, -3042.64092, -3779.8981, -4668.29109, -5729.66164, -7107.65714, -8877.0248, -10760.0316, -13238.1328, -16508.0879, -20766.6565, -26210.5984, -38976.1271, -51835.6846, -63501.1672, -72684.4709, -78097.4918, -73142.0326, -65380.145, -55799.4796, -48064.9728, -42104.5161, -36730.4647, -31926.4868, -27676.2505, -23963.4238, -20771.6749, -18199.6898, -16103.0164, -14420.9161, -13092.6501, -12057.4797, -11254.6664, -10623.4714, -10224.7671, -10182.7088},
        {-140.878578, -138.896878, -136.944493, -135.020846, -133.12536, -131.257457, -129.41656, -127.602091, -125.813472, -124.050126, -122.311475, -120.596942, -118.905949, -117.237918, -115.592272, -113.968434, -112.365825, -110.783868, -109.221986, -107.679601, -106.156135, -104.651011, -103.163651, -101.693478, -100.239914, -98.8023809, -97.380302, -95.9730995, -94.5801957, -93.201013, -91.8349738, -90.4815006, -89.1400157, -87.8099417, -86.497666, -85.1989819, -83.9095992, -82.6287142, -81.3555233, -80.0892228, -78.8290092, -77.5740789, -76.3236281, -75.0768534, -73.8329511, -72.5911175, -71.3505491, -70.1104423, -68.8699933, -67.6283987, -66.3848548, -65.0862724, -63.7852668, -62.4821681, -61.1773059, -59.8710101, -58.5636106, -57.2554372, -55.9468196, -54.6380879, -53.3295717, -52.0216009, -50.7145054, -49.408615, -48.1042595, -46.8017688, -45.5014726, -44.2202016, -42.9743901, -41.7303364, -40.4873008, -39.2445433, -38.0013241, -36.7569035, -35.5105415, -34.2614983, -33.0090341, -31.7524091, -30.4908833, -29.223717, -27.9501703, -26.6695033, -25.3809763, -24.0838494, -22.7096066, -21.2928217, -19.8688831, -18.4392506, -17.0053836, -15.5687418, -14.1307849, -12.6929725, -11.2567641, -9.8236195, -8.39499824, -6.97235994, -5.55716425, -4.15087079, -2.75493917, -1.37082903, 0, 0.689067875, 1.47492338, 2.35041306, 3.30838349, 4.34168122, 5.44315281, 6.60564482, 7.8220038, 9.08507631, 10.3877089, 11.7227482, 13.0830406, 14.4614328, 15.8507714, 17.2439028, 18.6336737, 20.0545105, 21.5052, 22.8845385, 24.1942827, 25.4361891, 26.6120143, 27.7235148, 28.7724472, 29.7605681, 30.689634, 31.5614015, 32.3776272, 33.1400675, 34.0472344, 34.8319854, 35.4370507, 35.8051602, 35.9411799, 35.7567719, 35.1470284, 34.0360384, 32.3478911, 30.0066756, 26.9364811, 24.7521944, 21.1725162, 15.7907274, 8.20010885, -2.00605866, -13.958477, -29.6588869, -49.8330118, -75.2065752, -106.5053, -142.917787, -188.58559, -242.742103, -306.858125, -386.972481, -487.073045, -611.147691, -758.428343, -938.739724, -1167.09486, -1446.10779, -1799.01644, -2226.42886, -2746.48152, -3407.01188, -4242.22416, -5247.57586, -6447.22578, -8007.40392, -10013.9768, -12098.9858, -14871.9418, -18575.8044, -23453.5336, -29748.089, -44747.6137, -59888.5885, -73652.6772, -84521.5436, -90976.8519, -85227.9824, -76210.4048, -65093.9046, -56302.8207, -49718.6777, -43866.9364, -38696.9244, -34157.9696, -30199.3998, -26770.5429, -23844.8155, -21317.8945, -19156.6252, -17327.8531, -15798.4236, -14535.182, -13504.9737, -12795.2437, -12393.8753},
        {-132.327653, -130.776012, -129.223098, -127.66943, -126.115527, -124.561907, -123.009091, -121.457595, -119.907939, -118.360642, -116.816223, -115.2752, -113.738091, -112.205417, -110.677695, -109.155444, -107.639184, -106.129432, -104.626708, -103.13153, -101.644418, -100.165889, -98.6964627, -97.236658, -95.7869934, -94.3479877, -92.9201597, -91.5040282, -90.1001118, -88.7089295, -87.3309998, -85.9668417, -84.6169738, -83.2819149, -82.0186135, -80.7981865, -79.5910728, -78.3959597, -77.211535, -76.0364859, -74.8695002, -73.7092652, -72.5544684, -71.4037975, -70.2559399, -69.1095831, -67.9634146, -66.8161219, -65.6663926, -64.5129141, -63.354374, -62.1069054, -60.8535397, -59.5947542, -58.331026, -57.0628321, -55.7906498, -54.5149561, -53.2362282, -51.9549432, -50.6715782, -49.3866103, -48.1005168, -46.8137746, -45.526861, -44.2402531, -42.9544279, -41.6802213, -40.4282203, -39.1775382, -37.9279805, -36.6793529, -35.431461, -34.1841104, -32.9371068, -31.6902557, -30.4433627, -29.1962336, -27.9486739, -26.7004893, -25.4514853, -24.2014676, -22.9502418, -21.6976135, -20.4219983, -19.1343466, -17.8458663, -16.557057, -15.2684186, -13.9804508, -12.6936533, -11.4085261, -10.1255687, -8.8452811, -7.56816293, -6.29471401, -5.02543409, -3.76082297, -2.50138039, -1.24760615, 0, 1.08353659, 2.18456832, 3.30166086, 4.43337989, 5.57829107, 6.73496009, 7.90195259, 9.07783427, 10.2611708, 11.4505278, 12.644471, 13.8415661, 15.0403787, 16.2394745, 17.4374191, 18.6327783, 19.736342, 20.7341366, 21.8232678, 22.9834924, 24.1945669, 25.4362478, 26.6882919, 27.9304557, 29.1424958, 30.3041688, 31.3952312, 32.3954398, 33.2845511, 34.1037688, 34.7488718, 35.1811825, 35.3620233, 35.4451516, 35.2956724, 34.7273747, 33.6438503, 31.9486913, 29.5454894, 26.3378368, 24.0999861, 20.295537, 14.4620823, 6.13721531, -5.14147108, -18.8654674, -36.710826, -59.3826829, -87.5861743, -122.026436, -161.110427, -210.650344, -270.384312, -342.271904, -432.808151, -546.451755, -687.66142, -855.600089, -1061.2258, -1321.2738, -1639.07048, -2039.6578, -2517.35172, -3092.44984, -3829.27422, -4771.93897, -5925.70042, -7324.59843, -9145.15144, -11483.2839, -13630.5029, -16610.9068, -20798.8725, -26568.7766, -34294.9959, -54140.3646, -74243.6876, -92532.227, -106933.245, -115374.005, -107124.117, -94540.2622, -79260.1249, -67695.8532, -59595.2393, -52716.7152, -46907.541, -42014.9769, -37886.2829, -34368.7191, -30803.4402, -27689.4739, -24981.4709, -22634.0819, -20601.9575, -18839.7482, -17302.1048, -16109.9855, -15320.8816},
        {-127.695887, -126.877418, -125.99736, -125.058231, -124.062548, -123.012828, -121.911588, -120.761346, -119.564618, -118.323923, -117.041778, -115.720699, -114.363204, -112.97181, -111.549035, -110.097396, -108.61941, -107.117595, -105.594467, -104.052544, -102.494344, -100.922383, -99.3391784, -97.7472483, -96.1491096, -94.5472796, -92.9442755, -91.3426147, -89.7448143, -88.1533917, -86.5708642, -84.9997491, -83.4425635, -81.9018249, -80.499995, -79.177096, -77.8717083, -76.5824568, -75.3079664, -74.0468619, -72.7977681, -71.5593099, -70.3301122, -69.1087998, -67.8939975, -66.6843302, -65.4784228, -64.2749001, -63.0723869, -61.869508, -60.6648885, -59.3457083, -58.024453, -56.7021633, -55.37988, -54.0586437, -52.7394952, -51.4234754, -50.1116248, -48.8049844, -47.5045948, -46.2114968, -44.9267311, -43.6513385, -42.3863597, -41.1328356, -39.8918068, -38.7061723, -37.6178262, -36.5414797, -35.4754601, -34.4180945, -33.3677103, -32.3226346, -31.2811948, -30.241718, -29.2025316, -28.1619627, -27.1183387, -26.0699867, -25.015234, -23.9524079, -22.8798357, -21.7958444, -20.6074583, -19.3605765, -18.1021957, -16.8336063, -15.5560984, -14.2709622, -12.979488, -11.6829659, -10.3826862, -9.07993908, -7.77601478, -6.47220351, -5.16979547, -3.87008088, -2.57434997, -1.28389293, 0, 1.49727428, 2.92525102, 4.28916709, 5.59425935, 6.84576465, 8.04891988, 9.20896188, 10.3311275, 11.4206536, 12.4827771, 13.5227349, 14.5457637, 15.5571004, 16.561982, 17.5656453, 18.573327, 19.3310642, 19.807065, 20.5928522, 21.6381196, 22.8925613, 24.3058712, 25.8277435, 27.4078721, 28.9959509, 30.5416741, 31.9947356, 33.3048293, 34.4216493, 35.1380006, 35.6179922, 35.8583847, 35.8559391, 36.1883196, 36.5618353, 36.5278872, 35.9089649, 34.527558, 32.2061561, 28.7672488, 25.1109522, 19.654156, 12.0085099, 1.78566357, -11.4027334, -25.5921236, -44.1119921, -67.9389162, -98.0494733, -135.420241, -178.954151, -234.235881, -299.749274, -377.422766, -476.217488, -601.937037, -760.385008, -954.004746, -1192.72379, -1490.1354, -1845.31024, -2290.00047, -2814.49266, -3442.96457, -4257.02074, -5311.90586, -6615.90131, -8212.07611, -10315.6982, -13044.6301, -15150.2881, -18327.2033, -23166.2389, -30258.2579, -40194.1234, -67778.7437, -95835.4252, -121401.52, -141514.38, -153211.356, -141088.913, -122919.216, -101071.424, -85001.7762, -74274.1308, -65499.3991, -58389.0877, -52654.7036, -48007.7537, -44159.7448, -39777.2058, -35896.9559, -32472.6566, -29457.9695, -26806.5559, -24472.0775, -22408.1957, -20472.7869, -19041.2338},
        {-123.130142, -122.289872, -121.398897, -120.459242, -119.472931, -118.441988, -117.368437, -116.254302, -115.101608, -113.912378, -112.688637, -111.432409, -110.145718, -108.830588, -107.489044, -106.123109, -104.734808, -103.326165, -101.899204, -100.455949, -98.9984245, -97.5286545, -96.0486631, -94.5604745, -93.0661129, -91.5676024, -90.0669672, -88.5662315, -87.0674194, -85.5725551, -84.0836627, -82.6027664, -81.1318903, -79.6730586, -78.323925, -77.0366875, -75.7623945, -74.499967, -73.2483256, -72.0063913, -70.7730847, -69.5473267, -68.3280382, -67.1141398, -65.9045524, -64.6981968, -63.4939937, -62.290864, -61.0877285, -59.883508, -58.6771232, -57.3558088, -56.0325929, -54.7088175, -53.3858243, -52.0649555, -50.7475529, -49.4349585, -48.1285142, -46.8295619, -45.5394436, -44.2595012, -42.9910767, -41.7355119, -40.4941488, -39.2683293, -38.0593955, -36.9239469, -35.9172567, -34.9267017, -33.9500416, -32.9850364, -32.0294457, -31.0810294, -30.1375473, -29.1967592, -28.2564248, -27.314304, -26.3681565, -25.4157422, -24.4548209, -23.4831523, -22.4984963, -21.4986125, -20.3686152, -19.1649562, -17.9454416, -16.7114866, -15.4645068, -14.2059174, -12.9371339, -11.6595715, -10.3746458, -9.08377191, -7.78836535, -6.48984147, -5.18961562, -3.88910318, -2.58971952, -1.29288, 0, 1.69526659, 3.28624337, 4.78081051, 6.18684819, 7.5122366, 8.76485592, 9.95258632, 11.083308, 12.1649011, 13.2052458, 14.2122224, 15.1937109, 16.1575917, 17.1117447, 18.0640503, 19.0223886, 19.708006, 20.0878349, 20.818094, 21.8452419, 23.1157372, 24.5760387, 26.1726049, 27.8518945, 29.5603661, 31.2444784, 32.85069, 34.3254596, 35.6152458, 36.5025254, 37.1578656, 37.5769195, 37.7553404, 38.3485253, 39.0222556, 39.2663806, 38.8786305, 37.6567351, 35.3984245, 31.9014288, 27.8654231, 21.9116871, 13.6614833, 2.73607405, -11.2432781, -25.8349958, -44.9432103, -69.6517378, -101.044395, -140.204997, -185.811572, -244.294135, -313.616008, -395.939938, -502.0025, -638.690798, -812.891931, -1028.88385, -1296.75581, -1629.44337, -2023.29974, -2516.73031, -3101.53701, -3805.76528, -4718.87318, -5902.1294, -7352.01991, -9113.01787, -11467.1329, -14562.0107, -16488.449, -19767.3793, -25258.4328, -33821.2409, -46315.435, -82795.0697, -120126.747, -154371.493, -181590.332, -197844.29, -182872.556, -159939.216, -132100.371, -111438.591, -97436.997, -85817.6772, -76251.8027, -68410.5443, -61965.0731, -56586.56, -51180.323, -46233.8799, -41750.8344, -37734.7898, -34189.3497, -31118.1176, -28524.697, -25952.4506, -23814.7537},
        {-117.374593, -115.692322, -114.047185, -112.438084, -110.863922, -109.323601, -107.816025, -106.340095, -104.894715, -103.478787, -102.091213, -100.730897, -99.3967412, -98.0876477, -96.8025195, -95.5402591, -94.2997693, -93.0799527, -91.8797119, -90.6979495, -89.5335682, -88.3854706, -87.2525593, -86.1337371, -85.0279065, -83.9339701, -82.8508307, -81.7773908, -80.7125531, -79.6552202, -78.6042948, -77.5586794, -76.5172768, -75.4789896, -74.4279241, -73.3706924, -72.3140843, -71.2574825, -70.2002699, -69.1418293, -68.0815434, -67.0187951, -65.9529672, -64.8834424, -63.8096037, -62.7308337, -61.6465153, -60.5560313, -59.4587645, -58.3540977, -57.2414136, -56.0378095, -54.8267376, -53.6093645, -52.3868568, -51.160381, -49.9311039, -48.7001919, -47.4688116, -46.2381298, -45.0093129, -43.7835276, -42.5619404, -41.345718, -40.1360269, -38.9340337, -37.7409051, -36.5937214, -35.5287005, -34.4729398, -33.4252776, -32.3845525, -31.3496029, -30.3192673, -29.2923841, -28.2677917, -27.2443287, -26.2208335, -25.1961445, -24.1691002, -23.138539, -22.1032995, -21.06222, -20.0141391, -18.9123453, -17.7794101, -16.6384524, -15.4897888, -14.3337361, -13.1706108, -12.0007295, -10.8244089, -9.6419657, -8.45371643, -7.25997776, -6.06106633, -4.85729878, -3.64899173, -2.43646183, -1.22002571, 0, 1.5668462, 3.06689776, 4.5051345, 5.88653626, 7.21608286, 8.49875413, 9.7395299, 10.94339, 12.1153143, 13.2602825, 14.3832746, 15.4892703, 16.5832494, 17.6701919, 18.7550775, 19.8428861, 20.8016027, 21.6166366, 22.6028365, 23.7358263, 24.9912298, 26.3446709, 27.7717733, 29.2481608, 30.7494573, 32.2512866, 33.7292725, 35.1590388, 36.5162093, 37.8618662, 39.0548402, 40.0451176, 40.7826846, 41.5382842, 42.1015244, 42.1764779, 41.6169039, 40.2765618, 38.0092107, 34.66861, 31.8372038, 27.1139449, 20.0143717, 10.0540223, -3.25156482, -18.5074663, -38.5473754, -64.3256, -96.7964482, -136.914228, -182.1415, -241.192011, -312.763243, -399.627931, -512.879109, -659.874165, -847.97049, -1080.51501, -1369.78704, -1732.44096, -2165.30815, -2712.0519, -3371.10983, -4174.00096, -5208.29892, -6535.84737, -8128.69494, -10025.4051, -12593.7897, -16015.4191, -17723.4334, -21070.6686, -27190.0747, -37214.6017, -52277.1997, -97269.6788, -143626.414, -186540.64, -221205.592, -242814.505, -227115.611, -201710.484, -170125.93, -145942.184, -128706.571, -113777.492, -100911.719, -89866.0258, -80397.1853, -72261.9708, -65544.9676, -59179.2808, -53253.633, -47856.747, -43077.3457, -39004.1519, -35725.8882, -32576.4732, -29673.4673},
        {-112.925689, -111.003804, -109.147347, -107.354209, -105.622282, -103.949457, -102.333626, -100.77268, -99.2645112, -97.8070107, -96.3980701, -95.0355808, -93.7174345, -92.4415225, -91.2057364, -90.0079677, -88.846108, -87.7180487, -86.6216813, -85.5548973, -84.5155883, -83.5016458, -82.5109612, -81.5414261, -80.590932, -79.6573703, -78.7386327, -77.8326105, -76.9371954, -76.0502788, -75.1697522, -74.2935072, -73.4194352, -72.5454278, -71.6163322, -70.6576778, -69.6956315, -68.7298061, -67.7598145, -66.7852697, -65.8057845, -64.8209719, -63.8304447, -62.8338158, -61.8306981, -60.8207045, -59.8034479, -58.7785411, -57.7455972, -56.704229, -55.6540493, -54.5333687, -53.4044314, -52.2681792, -51.1255539, -49.9774973, -48.8249512, -47.6688575, -46.5101579, -45.3497942, -44.1887083, -43.0278419, -41.868137, -40.7105352, -39.5559784, -38.4054085, -37.2597671, -36.1465356, -35.092558, -34.0440405, -33.0002045, -31.9602713, -30.9234624, -29.888999, -28.8561026, -27.8239946, -26.7918962, -25.759029, -24.7246141, -23.687873, -22.6480272, -21.6042978, -20.5559063, -19.5020741, -18.4156528, -17.3096033, -16.1972033, -15.0785301, -13.9536608, -12.8226724, -11.6856421, -10.542647, -9.39376426, -8.23907095, -7.07864419, -5.91256112, -4.74089884, -3.56373447, -2.38114513, -1.19320793, 0, 1.20466733, 2.41060951, 3.6183716, 4.82849866, 6.04153576, 7.25802796, 8.47852032, 9.7035579, 10.9336858, 12.169449, 13.4113926, 14.6600616, 15.9160013, 17.1797564, 18.4518723, 19.7328939, 21.0103207, 22.2828341, 23.5804871, 24.9010291, 26.2422097, 27.6017786, 28.9774853, 30.3670794, 31.7683106, 33.1789283, 34.5966823, 36.019322, 37.4445971, 39.1310295, 40.71998, 42.1309664, 43.2835066, 44.2218875, 44.8032425, 44.8514332, 44.2485469, 42.8766708, 40.6178919, 37.3542977, 35.2153436, 31.1517667, 24.6059518, 15.0202835, 1.83714657, -13.497186, -34.04519, -60.8654527, -95.0165616, -137.557104, -185.908763, -249.212136, -325.943809, -419.030685, -540.485721, -698.24242, -900.234285, -1149.58967, -1460.13931, -1850.95592, -2317.60931, -2911.00143, -3634.402, -4523.83698, -5670.836, -7141.4516, -8890.30937, -10954.5734, -13779.9732, -17581.1305, -19341.799, -23004.5253, -29902.3086, -41368.1481, -58735.043, -110355.134, -163787.494, -213610.336, -254401.874, -280740.321, -265704.746, -239705.272, -206534.033, -180493.595, -161215.713, -143952.792, -128555.556, -114874.724, -102761.018, -92065.16, -83575.0147, -75483.3992, -67893.2232, -60907.396, -54628.8273, -49160.4265, -44605.1033, -40268.1829, -36283.13},
        {-108.709406, -107.157153, -105.637939, -104.150855, -102.69499, -101.269436, -99.8732818, -98.5056183, -97.1655355, -95.8521237, -94.564473, -93.3016739, -92.0628164, -90.8469908, -89.6532874, -88.4807964, -87.3286079, -86.1958123, -85.0814998, -83.9847606, -82.904685, -81.8403631, -80.7908852, -79.7553416, -78.7328225, -77.7224181, -76.7232186, -75.7343143, -74.7547954, -73.7837522, -72.8202748, -71.8634536, -70.9123787, -69.9661404, -69.0140195, -68.0602173, -67.109053, -66.1599351, -65.2122721, -64.2654726, -63.3189453, -62.3720986, -61.4243411, -60.4750815, -59.5237282, -58.5696898, -57.6123749, -56.6511921, -55.6855499, -54.7148569, -53.7385217, -52.7063178, -51.6683647, -50.6251471, -49.5771494, -48.5248564, -47.4687524, -46.4093221, -45.34705, -44.2824206, -43.2159187, -42.1480285, -41.0792349, -40.0100222, -38.9408751, -37.8722781, -36.8047157, -35.753943, -34.7353483, -33.7179219, -32.7011587, -31.6845532, -30.6676, -29.6497939, -28.6306294, -27.6096011, -26.5862038, -25.559932, -24.5302804, -23.4967436, -22.4588163, -21.4159931, -20.3677686, -19.3136375, -18.2293186, -17.1266945, -16.0179336, -14.9033022, -13.7830665, -12.6574926, -11.5268468, -10.3913953, -9.25140436, -8.1071401, -6.95886877, -5.80685658, -4.65136973, -3.49267444, -2.33103691, -1.16672336, 0, 0.507913685, 1.13340008, 1.86978696, 2.7104021, 3.64857327, 4.67762824, 5.79089479, 6.98170069, 8.24337372, 9.56924164, 10.9526322, 12.3868733, 13.8652925, 15.3812177, 16.9279767, 18.4988973, 20.1941765, 22.0224093, 23.7351949, 25.3487619, 26.8793385, 28.3431531, 29.756434, 31.1354096, 32.4963083, 33.8553584, 35.2287883, 36.6328263, 38.0837007, 39.9866744, 41.8262954, 43.5020816, 44.9135512, 45.9562186, 46.5316038, 46.5422939, 45.8890081, 44.4724655, 42.1933851, 38.9524859, 37.4114001, 33.8696558, 27.6877935, 18.2263536, 4.84587668, -10.2492472, -31.2532897, -59.5166729, -96.3898188, -143.223149, -198.420348, -269.880145, -355.291799, -457.345153, -588.93957, -758.390545, -974.013574, -1238.34351, -1566.79028, -1981.06995, -2475.08894, -3107.85869, -3884.93809, -4847.59948, -6097.62959, -7709.03806, -9629.06703, -11898.8605, -15024.3765, -19248.6482, -21441.7669, -25738.6212, -33549.8007, -46285.8952, -65357.4943, -120183.722, -177164.5, -230708.285, -275223.532, -305118.697, -293147.098, -269809.088, -238793.899, -213691.745, -194272.717, -176181.493, -159394.433, -143887.898, -129638.246, -116621.84, -105810.959, -95638.4102, -86130.2257, -77312.4373, -69211.0772, -61852.1776, -55261.7703, -49050.0293, -43658.0217},
        {-104.863058, -103.825531, -102.773263, -101.707102, -100.627893, -99.5364851, -98.4337238, -97.3204566, -96.1975304, -95.0657922, -93.926089, -92.7792678, -91.6261756, -90.4676592, -89.3045659, -88.1377424, -86.9680358, -85.796293, -84.6233611, -83.4500871, -82.2773178, -81.1059003, -79.9366816, -78.7705086, -77.6082284, -76.4506878, -75.298734, -74.1532138, -73.0149743, -71.8848624, -70.7637251, -69.6524094, -68.5517622, -67.4626307, -66.440406, -65.4575157, -64.4857317, -63.5241311, -62.5717907, -61.6277874, -60.6911981, -59.7610998, -58.8365693, -57.9166836, -57.0005196, -56.0871542, -55.1756643, -54.2651268, -53.3546186, -52.4432167, -51.5299979, -50.5653829, -49.5981597, -48.6284599, -47.6564153, -46.6821574, -45.7058181, -44.7275289, -43.7474216, -42.7656279, -41.7822794, -40.7975078, -39.8114448, -38.8242222, -37.8359715, -36.8468245, -35.8569128, -34.8753153, -33.9108957, -32.9453333, -31.9781795, -31.0089863, -30.0373052, -29.0626879, -28.0846861, -27.1028514, -26.1167357, -25.1258905, -24.1298676, -23.1282185, -22.1204951, -21.106249, -20.0850319, -19.0563954, -17.9903002, -16.9017159, -15.8059679, -14.7035679, -13.5950282, -12.4808605, -11.3615769, -10.2376894, -9.10970996, -7.97815049, -6.84352299, -5.70633945, -4.56711183, -3.4263521, -2.28457224, -1.14228421, 0, -0.0450627005, 0.116155894, 0.471730443, 1.00973561, 1.71824604, 2.5853364, 3.59908135, 4.74755555, 6.01883366, 7.40099033, 8.88210022, 10.450238, 12.0934783, 13.7998958, 15.5575652, 17.3545611, 19.3496398, 21.555259, 23.5734278, 25.4287956, 27.1460116, 28.7497251, 30.2645855, 31.7152419, 33.1263436, 34.5225401, 35.9284805, 37.3688142, 38.8681904, 40.8852947, 42.8515936, 44.6615253, 46.2095282, 47.3680965, 48.0426408, 48.1444232, 47.5744656, 46.2337895, 44.0234166, 40.8443687, 39.7370792, 36.5076848, 30.4429744, 20.8297369, 6.95476117, -8.201797, -29.9698572, -59.9859722, -99.8866948, -151.308578, -213.585807, -293.471306, -387.934352, -499.457276, -641.778231, -823.581734, -1053.5523, -1333.26701, -1680.13315, -2119.3116, -2645.31351, -3321.56333, -4149.4054, -5173.98749, -6514.67985, -8256.7834, -10350.9198, -12847.473, -16290.6512, -20945.3277, -23770.9521, -28846.3481, -37580.5742, -51382.6888, -71661.7504, -127533.587, -185773.795, -240864.741, -287288.79, -319528.307, -311156.604, -291504.473, -264015.304, -241120.492, -222705.674, -204911.489, -187810.824, -171476.568, -155981.61, -141398.839, -128883.842, -116923.655, -105552.969, -94806.4788, -84718.8765, -75324.8555, -66659.1085, -58613.9209, -51649.5912},
        {-101.405685, -100.319022, -99.2240199, -98.1213189, -97.0115585, -95.8953786, -94.7734186, -93.6463184, -92.5147174, -91.3792554, -90.2405721, -89.0993071, -87.9561, -86.8115905, -85.6664183, -84.521223, -83.3766443, -82.2333218, -81.0918952, -79.9530041, -78.8172883, -77.6853873, -76.5579409, -75.4355886, -74.3189701, -73.2087251, -72.1054933, -71.0099143, -69.9226277, -68.8442733, -67.7754906, -66.7169193, -65.6691991, -64.6329697, -63.6561139, -62.714656, -61.7840527, -60.8634103, -59.9518355, -59.0484348, -58.1523146, -57.2625815, -56.378342, -55.4987027, -54.6227701, -53.7496506, -52.8784508, -52.0082773, -51.1382365, -50.267435, -49.3949793, -48.4735, -47.549587, -46.6233545, -45.6949163, -44.7643866, -43.8318794, -42.8975086, -41.9613884, -41.0236327, -40.0843555, -39.143671, -38.2016932, -37.2585359, -36.3143134, -35.3691396, -34.4231286, -33.4849734, -32.5631611, -31.6401122, -30.7153845, -29.7885358, -28.8591241, -27.9267072, -26.9908431, -26.0510895, -25.1070044, -24.1581457, -23.2040712, -22.2443389, -21.2785065, -20.3061321, -19.3267734, -18.3399884, -17.3224973, -16.2857571, -15.2414995, -14.1900234, -13.1316278, -12.0666117, -10.9952742, -9.91791421, -8.83483078, -7.7463229, -6.65268959, -5.55422986, -4.4512427, -3.34402713, -2.23288215, -1.11810677, 0, -0.086065554, 0.0429181388, 0.374746579, 0.897215268, 1.59811971, 2.46525539, 3.48641783, 4.64940253, 5.94200497, 7.35202067, 8.86724512, 10.4754738, 12.1645023, 13.922126, 15.7361405, 17.5943412, 19.66926, 21.975083, 24.0817503, 26.0166435, 27.8071445, 29.4806351, 31.0644972, 32.5861126, 34.0728632, 35.5521307, 37.051297, 38.597744, 40.2188534, 42.3896786, 44.5257838, 46.5202492, 48.2661553, 49.6534634, 50.5768133, 50.9316768, 50.61207, 49.5120091, 47.5255102, 44.5465895, 43.8787795, 40.968901, 35.0438908, 25.3306856, 11.0562222, -4.86390133, -27.8135747, -59.4880264, -101.582485, -155.792179, -221.118187, -305.242737, -405.254552, -523.933561, -675.67923, -869.674091, -1115.10067, -1412.20011, -1781.12847, -2251.79607, -2823.07478, -3556.89763, -4445.52221, -5535.32666, -6964.39344, -8828.71119, -11083.7659, -13789.1196, -17519.5108, -22558.2653, -25894.2653, -31612.1171, -41135.5304, -55888.2151, -77293.8809, -134675.885, -194583.379, -251465.159, -299770.023, -333946.769, -327665.839, -310006.624, -284322.932, -262609.011, -244798.494, -227224.832, -210001.666, -193242.64, -177061.396, -161571.577, -148751.488, -135782.875, -122914.284, -110394.261, -98471.3517, -87394.1012, -77411.0552, -68312.0996, -59896.2774},
        {-98.0372641, -96.8787191, -95.7204908, -94.5629208, -93.406351, -92.2511232, -91.097579, -89.9460603, -88.7969089, -87.6504665, -86.5070749, -85.3670758, -84.2308111, -83.0986226, -81.9708519, -80.8478409, -79.7299313, -78.617465, -77.5107836, -76.410229, -75.3161429, -74.2288672, -73.1487435, -72.0761137, -71.0113195, -69.9547027, -68.9066051, -67.8673685, -66.8373345, -65.8168451, -64.806242, -63.8058669, -62.8160616, -61.8371679, -60.9055256, -60.0027203, -59.109905, -58.2262531, -57.3509383, -56.483134, -55.622014, -54.7667516, -53.9165206, -53.0704945, -52.2278468, -51.3877511, -50.549381, -49.71191, -48.8745118, -48.0363598, -47.1966277, -46.3096483, -45.420408, -44.5290524, -43.635727, -42.7405775, -41.8437496, -40.9453889, -40.0456409, -39.1446514, -38.2425659, -37.3395301, -36.4356895, -35.5311899, -34.6261768, -33.720796, -32.8151928, -31.9192834, -31.0427488, -30.1655845, -29.2873025, -28.4074153, -27.5254351, -26.6408741, -25.7532446, -24.8620588, -23.966829, -23.0670676, -22.1622866, -21.2519985, -20.3357155, -19.4129497, -18.4832136, -17.5460192, -16.5845222, -15.6067572, -14.6209555, -13.6271602, -12.6254143, -11.615761, -10.5982432, -9.57290418, -8.53978688, -7.49893439, -6.4503898, -5.39419617, -4.33039658, -3.25903409, -2.18015179, -1.09379273, 0, -0.0205575168, 0.164062313, 0.542642333, 1.10396539, 1.83681432, 2.72997198, 3.77222121, 4.95234484, 6.25912574, 7.68134673, 9.20779067, 10.8272404, 12.5284787, 14.3002886, 16.1314527, 18.0107541, 20.1020981, 22.4192852, 24.5549873, 26.5355137, 28.3871734, 30.1362756, 31.8091293, 33.4320439, 35.0313283, 36.6332917, 38.2642432, 39.9504921, 41.7183473, 44.0279155, 46.3126494, 48.4687188, 50.3922937, 52.0359926, 53.2677612, 53.9404921, 53.9334208, 53.1257825, 51.3968128, 48.6257469, 48.2731218, 45.5469099, 39.6256572, 29.6879097, 14.9122133, -1.89479453, -26.0896823, -59.4009269, -103.557005, -160.286395, -228.065407, -315.849552, -421.048874, -546.839233, -708.168364, -914.698672, -1176.09256, -1491.12162, -1882.81425, -2386.07975, -3004.81506, -3798.3771, -4749.22859, -5904.46627, -7420.70559, -9404.29109, -11817.2975, -14728.6661, -18736.3275, -24138.9861, -28035.5143, -34416.7302, -44688.9722, -60258.5788, -82531.8884, -140528.955, -201138.973, -258864.851, -308209.499, -343675.827, -339407.08, -323838.941, -300199.888, -279955.794, -263074.504, -246091.617, -229146.591, -212378.885, -195927.956, -179933.262, -167104.315, -153403.647, -139292.562, -125232.364, -111684.355, -99109.8388, -87970.1198, -77917.5759, -68118.6112},
        {-94.6155069, -93.3920942, -92.1776058, -90.972072, -89.7755229, -88.5879888, -87.4094998, -86.2400861, -85.079778, -83.9286056, -82.786599, -81.6537886, -80.5302044, -79.4158767, -78.3108356, -77.2151114, -76.1287342, -75.0517343, -73.9841417, -72.9259868, -71.8772996, -70.8381105, -69.8084495, -68.7883469, -67.7778328, -66.7769375, -65.7856911, -64.8041238, -63.8322659, -62.8701474, -61.9177987, -60.9752498, -60.042531, -59.1196725, -58.2297235, -57.3607205, -56.5004233, -55.6481151, -54.8030791, -53.9645984, -53.1319562, -52.3044357, -51.4813201, -50.6618926, -49.8454362, -49.0312343, -48.2185699, -47.4067263, -46.5949866, -45.782634, -44.9689517, -44.1117088, -43.2526024, -42.3918157, -41.5295319, -40.6659339, -39.8012051, -38.9355284, -38.0690871, -37.2020641, -36.3346427, -35.467006, -34.5993371, -33.731819, -32.864635, -31.9979682, -31.1320016, -30.2782915, -29.4481209, -28.6182169, -27.7880254, -26.9569922, -26.1245632, -25.2901842, -24.4533011, -23.6133596, -22.7698058, -21.9220854, -21.0696442, -20.2119282, -19.3483832, -18.4784549, -17.6015894, -16.7172324, -15.8147437, -14.8988242, -13.9742964, -13.0409333, -12.0985082, -11.1467942, -10.1855645, -9.21459228, -8.23365066, -7.24251281, -6.24095193, -5.22874116, -4.20565368, -3.17146266, -2.12594126, -1.06886265, 0, 0.101876771, 0.388343789, 0.849937705, 1.47719517, 2.26065283, 3.19084733, 4.25831533, 5.45359348, 6.76721842, 8.1897268, 9.71165528, 11.3235405, 13.0159191, 14.7793278, 16.6043031, 18.4813818, 20.5515758, 22.8269182, 24.9575849, 26.9663572, 28.8760164, 30.709344, 32.4891213, 34.2381296, 35.9791503, 37.7349647, 39.5283543, 41.3821004, 43.3189844, 45.7658583, 48.1932734, 50.5027901, 52.5959683, 54.5185993, 56.1001277, 57.1315365, 57.4711165, 56.9771586, 55.5079536, 52.9217924, 52.7255014, 50.0184119, 43.9449709, 33.6496251, 18.2768215, 0.575863257, -24.8148731, -59.6521547, -105.692749, -164.693422, -234.505069, -325.564175, -435.763792, -568.745977, -739.890344, -959.306582, -1237.10438, -1570.58709, -1985.58127, -2521.88407, -3188.17649, -4041.53718, -5056.13717, -6279.79299, -7885.10927, -9987.41086, -12554.6573, -15663.9676, -19932.8002, -25672.9671, -30157.0588, -37203.6811, -48176.6842, -64439.9182, -87357.2331, -145256.694, -205796.883, -263600.595, -313290.627, -349489.776, -347088.083, -333596.268, -312094.021, -293456.399, -277677.426, -261503.058, -245089.722, -228593.845, -212171.856, -195980.182, -183220.077, -169018.959, -153995.025, -138766.474, -123951.504, -110168.312, -98035.0981, -87069.9803, -76033.1982},
        {-90.998127, -89.7466187, -88.5102946, -87.2889365, -86.0823262, -84.8902454, -83.712476, -82.5487995, -81.3989979, -80.2628529, -79.1401463, -78.0306597, -76.934175, -75.8504739, -74.7793383, -73.7205498, -72.6738902, -71.6391413, -70.6160849, -69.6045027, -68.6041764, -67.6148879, -66.6364188, -65.6685511, -64.7110663, -63.7637464, -62.826373, -61.8987279, -60.9805929, -60.0717497, -59.1719802, -58.281066, -57.3987889, -56.5249307, -55.6697898, -54.8276684, -53.9927423, -53.1644521, -52.3422383, -51.5255412, -50.7138015, -49.9064595, -49.1029557, -48.3027306, -47.5052248, -46.7098786, -45.9161326, -45.1234272, -44.3312029, -43.5389003, -42.7459596, -41.9175619, -41.0881499, -40.2579067, -39.4270156, -38.5956597, -37.7640221, -36.932286, -36.1006346, -35.269251, -34.4383184, -33.60802, -32.7785388, -31.9500582, -31.1227612, -30.296831, -29.4724507, -28.6620437, -27.8777394, -27.0944765, -26.3116448, -25.5286339, -24.7448336, -23.9596334, -23.1724231, -22.3825924, -21.5895309, -20.7926283, -19.9912743, -19.1848586, -18.3727709, -17.5544008, -16.7291381, -15.8963724, -15.0515303, -14.1960663, -13.3314828, -12.4572979, -11.5730301, -10.6781975, -9.77231854, -8.85491144, -7.9254945, -6.98358602, -6.02870428, -5.06036759, -4.07809424, -3.08140251, -2.0698107, -1.0428371, 0, 0.231652668, 0.624517939, 1.17115269, 1.8641138, 2.69595815, 3.65924262, 4.74652407, 5.9503594, 7.26330548, 8.67791918, 10.1867574, 11.782377, 13.4573348, 15.2041878, 17.0154929, 18.8838068, 20.9211148, 23.1370351, 25.2539889, 27.2901248, 29.2635916, 31.1925378, 33.0951123, 34.9894636, 36.8937405, 38.8260915, 40.8046654, 42.8476109, 44.9730766, 47.5693596, 50.1487391, 52.6183191, 54.8852039, 57.1041982, 59.0585559, 60.4654771, 61.157756, 60.9681863, 59.7295622, 57.2746775, 57.0413141, 54.1601077, 47.7585288, 36.9640474, 20.9041339, 2.41841189, -24.0058401, -60.1691903, -107.872207, -168.915458, -240.514776, -334.65903, -449.845782, -590.225481, -771.489881, -1004.14893, -1298.71256, -1651.15208, -2089.82031, -2658.93043, -3370.80123, -4281.91333, -5361.86058, -6659.69349, -8359.09746, -10581.9583, -13298.9879, -16592.879, -21100.6279, -27145.6852, -32221.2583, -39916.4633, -51534.4509, -68378.3714, -91751.3753, -149023.002, -208913.415, -266209.169, -315696.816, -352162.91, -351416.605, -339873.453, -320453.177, -303406.381, -288750.982, -273450.367, -257675.179, -241596.063, -225383.663, -209208.625, -196376.525, -181861.794, -166328.892, -150442.28, -134866.418, -120265.766, -107304.786, -95408.9431, -83356.6439},
        {-87.0428371, -85.8297641, -84.6334868, -83.4536786, -82.2900129, -81.1421631, -80.0098024, -78.8926042, -77.7902419, -76.7023888, -75.6287183, -74.5689036, -73.5226183, -72.4895356, -71.4693288, -70.4616714, -69.4662366, -68.4826978, -67.5107285, -66.5500018, -65.6001912, -64.66097, -63.7320116, -62.8129893, -61.9035764, -61.0034464, -60.1122726, -59.2297282, -58.3554868, -57.4892215, -56.6306058, -55.779313, -54.9350165, -54.0973897, -53.2668067, -52.4425758, -51.6239967, -50.8107199, -50.0023961, -49.1986758, -48.3992097, -47.6036484, -46.8116424, -46.0228424, -45.2368989, -44.4534627, -43.6721842, -42.892714, -42.1147029, -41.3378013, -40.5616599, -39.7650885, -38.9690303, -38.1735876, -37.3788629, -36.5849585, -35.7919768, -35.0000203, -34.2091912, -33.4195921, -32.6313253, -31.8444931, -31.059198, -30.2755424, -29.4936286, -28.713559, -27.9354361, -27.1705862, -26.4300664, -25.6908301, -24.9522524, -24.2137078, -23.4745712, -22.7342174, -21.9920212, -21.2473573, -20.4996005, -19.7481256, -18.9923073, -18.2315206, -17.46514, -16.6925405, -15.9130967, -15.1261836, -14.3332506, -13.5325916, -12.7224752, -11.902209, -11.0711004, -10.2284568, -9.37358573, -8.50579457, -7.6243908, -6.72868185, -5.81797518, -4.89157823, -3.94879845, -2.98894329, -2.0113202, -1.01523662, 0, 0.319185534, 0.781340134, 1.38080729, 2.11193049, 2.96905323, 3.946519, 5.03867129, 6.23985359, 7.54440938, 8.94668217, 10.4410154, 12.0217527, 13.6832374, 15.419813, 17.2258231, 19.0956112, 21.1141368, 23.2886886, 25.408645, 27.4877674, 29.5398168, 31.5785545, 33.6177419, 35.67114, 37.7525102, 39.8756136, 42.0542116, 44.3020653, 46.632936, 49.4042722, 52.1601293, 54.811162, 57.2680256, 59.7957044, 62.1276889, 63.9029813, 64.9259378, 65.0009146, 63.9322679, 61.5243538, 61.0259553, 57.7486979, 50.8230281, 39.3793925, 22.5482376, 3.50319124, -23.6792766, -60.8795147, -109.977872, -172.854696, -246.17213, -343.406538, -463.741318, -611.849429, -803.611686, -1049.8768, -1361.4935, -1733.37214, -2195.92215, -2796.94024, -3550.33142, -4515.04101, -5662.01143, -7042.55445, -8844.16312, -11191.821, -14053.4321, -17513.2556, -22231.5096, -28542.6174, -34190.4723, -42498.5704, -54698.0569, -72020.0768, -95695.7752, -151991.775, -210844.874, -267227.348, -316111.473, -352469.524, -353100.401, -343265.34, -325725.205, -310101.298, -296438.894, -281924.755, -266747.083, -251094.079, -235153.945, -219114.883, -205851.413, -191165.139, -175601.384, -159705.469, -144022.715, -129098.445, -115477.98, -102574.095, -89805.5539},
        {-81.9500307, -80.9683203, -79.9912275, -79.0187472, -78.0508745, -77.0876042, -76.1289313, -75.1748507, -74.2253573, -73.2804462, -72.3401121, -71.4043501, -70.4731551, -69.546522, -68.6244457, -67.7069213, -66.7939435, -65.8855075, -64.981608, -64.0822401, -63.1873987, -62.2970786, -61.4112749, -60.5299825, -59.6531963, -58.7809112, -57.9131222, -57.0498243, -56.1910123, -55.3366812, -54.4868259, -53.6414413, -52.8005225, -51.9640643, -51.1319823, -50.3043128, -49.481093, -48.6623205, -47.8479927, -47.0381072, -46.2326615, -45.431653, -44.6350793, -43.842938, -43.0552265, -42.2719423, -41.4930829, -40.7186459, -39.9486287, -39.1830289, -38.4218441, -37.6768602, -36.9360308, -36.1990976, -35.4658027, -34.735888, -34.0090955, -33.2851672, -32.5638449, -31.8448706, -31.1279863, -30.412934, -29.6994556, -28.9872931, -28.2761883, -27.5658834, -26.8561201, -26.1506525, -25.453138, -24.7550444, -24.0558535, -23.3550471, -22.6521072, -21.9465156, -21.2377543, -20.5253051, -19.8086498, -19.0872704, -18.3606487, -17.6282666, -16.889606, -16.1441488, -15.3913768, -14.630772, -13.8665661, -13.0957661, -12.315322, -11.5245615, -10.7228123, -9.90940215, -9.08365879, -8.24490996, -7.39248338, -6.52570679, -5.64390792, -4.74641451, -3.83255429, -2.901655, -1.95304436, -0.986050118, 0, 0.176074227, 0.509117449, 0.993532835, 1.62372355, 2.39409277, 3.29904365, 4.33297936, 5.49030308, 6.76541797, 8.1527272, 9.64663393, 11.2415413, 12.9318526, 14.7119708, 16.5762993, 18.5192411, 20.6215164, 22.8898595, 25.1234325, 27.3351352, 29.5378672, 31.7445282, 33.9680179, 36.2212359, 38.5170819, 40.8684557, 43.2882568, 45.789385, 48.38474, 51.4501889, 54.5025757, 57.4459101, 60.1842013, 63.0383186, 65.7038416, 67.7651871, 69.0013067, 69.1911518, 68.113674, 65.5478247, 64.4104666, 60.3876236, 52.644308, 40.3455322, 22.6563085, 3.26318511, -24.3207454, -62.0608546, -111.922514, -175.871096, -249.700341, -349.090355, -475.404758, -634.199324, -839.588801, -1101.84513, -1431.24026, -1819.45932, -2299.58533, -2924.97787, -3707.91223, -4718.09366, -5947.33111, -7452.47694, -9400.59592, -11908.1635, -14912.666, -18484.1744, -23349.0529, -29862.4292, -36080.3329, -45010.697, -57804.491, -75612.6841, -99586.2456, -154929.448, -212726.632, -268115.441, -316233.517, -352218.505, -353734.449, -345041.657, -328766.458, -313893.034, -300531.063, -286141.117, -270965.328, -255245.829, -239224.753, -223144.234, -207297.322, -192299.824, -177943.822, -164021.396, -150324.626, -136645.593, -122776.376, -107623.41, -95200.3918},
        {-77.6286638, -76.7989479, -75.9669642, -75.13287, -74.2968225, -73.458979, -72.6194967, -71.7785331, -70.9362452, -70.0927904, -69.2483259, -68.403009, -67.556997, -66.7104471, -65.8635166, -65.0163627, -64.1691428, -63.3220141, -62.4751338, -61.6286593, -60.7827477, -59.9375564, -59.0932427, -58.2499636, -57.4078767, -56.567139, -55.727908, -54.8903408, -54.0545947, -53.2208269, -52.3891948, -51.5598556, -50.7329666, -49.908685, -49.0846232, -48.2622644, -47.4431226, -46.6274374, -45.8154488, -45.0073967, -44.2035208, -43.4040611, -42.6092574, -41.8193495, -41.0345774, -40.2551808, -39.4813996, -38.7134737, -37.9516429, -37.196147, -36.447226, -35.7371988, -35.0335307, -34.3357662, -33.6434496, -32.9561255, -32.2733383, -31.5946324, -30.9195523, -30.2476423, -29.578447, -28.9115108, -28.2463781, -27.5825934, -26.9197011, -26.2572456, -25.5947714, -24.9311792, -24.2653853, -23.5982617, -22.9293944, -22.2583697, -21.5847738, -20.9081928, -20.2282129, -19.5444202, -18.8564009, -18.1637412, -17.4660273, -16.7628453, -16.0537814, -15.3384217, -14.6163525, -13.8871598, -13.16494, -12.4417188, -11.709348, -10.9669427, -10.2136184, -9.44849018, -8.67067343, -7.87928344, -7.07343548, -6.25224485, -5.41482684, -4.56029674, -3.68776984, -2.79636142, -1.88518678, -0.953361212, 0, 0.186352681, 0.523875309, 1.00791213, 1.63380738, 2.39690531, 3.29255015, 4.31608615, 5.46285756, 6.72820861, 8.10748356, 9.59602663, 11.1891821, 12.8822941, 14.670707, 16.5497651, 18.5148124, 20.6335162, 22.9115525, 25.1806781, 27.4517352, 29.7355655, 32.0430113, 34.3849143, 36.7721167, 39.2154605, 41.7257875, 44.3139399, 46.9907596, 49.7670886, 53.0100263, 56.2435295, 59.3715632, 62.298092, 65.4581485, 68.4901636, 70.8909502, 72.4051527, 72.7774157, 71.7523835, 69.0747008, 67.2413017, 62.4068874, 53.7776111, 40.5596256, 21.9590839, 2.94309977, -24.4833747, -62.5544269, -113.504144, -179.566613, -256.954133, -361.282543, -493.032739, -657.812478, -871.69685, -1146.06133, -1492.28139, -1904.76786, -2415.71646, -3075.82974, -3892.49504, -4943.79829, -6228.62903, -7807.73284, -9846.66702, -12462.8786, -15594.5321, -19312.594, -24345.3994, -31049.4383, -37687.7934, -47032.4252, -60164.4648, -78165.0437, -102115.293, -156121.798, -212483.873, -266526.285, -313573.801, -348951.191, -351254.832, -343690.14, -328751.192, -314904.374, -302279.725, -288518.818, -273876.115, -258606.076, -242963.161, -227201.831, -211304.183, -196547.6, -182572.167, -169017.965, -155525.078, -141733.586, -127283.573, -112017.532, -98944.4579},
        {-73.9283896, -73.1753362, -72.4179872, -71.6565255, -70.8911337, -70.1219948, -69.3492915, -68.5732066, -67.793923, -67.0116233, -66.2264905, -65.4387072, -64.6484564, -63.8559208, -63.0612832, -62.2647264, -61.4664332, -60.6665864, -59.8653688, -59.0629633, -58.2595525, -57.4553193, -56.6504465, -55.8451169, -55.0395133, -54.2338185, -53.4282153, -52.6228864, -51.8180148, -51.0137831, -50.2103743, -49.407971, -48.6067561, -47.8069124, -47.0032313, -46.1987047, -45.3963893, -44.5966428, -43.799823, -43.0062877, -42.2163946, -41.4305014, -40.648966, -39.8721461, -39.1003994, -38.3340837, -37.5735568, -36.8191764, -36.0713003, -35.3302862, -34.596492, -33.9097114, -33.2300112, -32.5568943, -31.8898635, -31.2284218, -30.5720719, -29.9203167, -29.2726591, -28.6286019, -27.9876481, -27.3493004, -26.7130617, -26.0784349, -25.4449229, -24.8120285, -24.1792545, -23.5433951, -22.9013094, -22.2580868, -21.6134059, -20.9669451, -20.3183829, -19.6673976, -19.0136679, -18.356872, -17.6966886, -17.0327961, -16.3648728, -15.6925973, -15.0156481, -14.3337035, -13.6464421, -12.9535424, -12.2839236, -11.6220287, -10.9519493, -10.2724148, -9.58215492, -8.87989908, -8.16437682, -7.4343177, -6.68845125, -5.925507, -5.1442145, -4.34330326, -3.52150284, -2.67754277, -1.81015258, -0.918061813, 0, 0.322376422, 0.772983825, 1.34896159, 2.04744908, 2.86558569, 3.80051078, 4.84936375, 6.00928395, 7.27741077, 8.65088359, 10.1268418, 11.7024247, 13.3747718, 15.1410224, 16.9983158, 18.9437915, 21.0173307, 23.2221785, 25.4587967, 27.7334836, 30.0525373, 32.4222564, 34.8489391, 37.3388837, 39.8983886, 42.5337521, 45.2512726, 48.0572483, 50.9579777, 54.2782005, 57.5890117, 60.8011771, 63.8254626, 67.252426, 70.6529373, 73.4165883, 75.2502074, 75.8606227, 74.9546626, 72.2391554, 69.824804, 64.2829479, 54.8500919, 40.762741, 21.2574001, 2.93853595, -24.1672989, -62.66559, -115.161823, -184.261483, -267.447625, -378.708915, -515.552479, -683.052833, -901.734579, -1185.18555, -1546.99356, -1988.09096, -2537.59731, -3238.4382, -4091.98376, -5181.3929, -6501.20737, -8114.25269, -10199.0574, -12881.0405, -16122.2356, -20005.7027, -25213.5806, -32089.1352, -39022.3782, -48610.3302, -61872.3836, -79827.9304, -103496.363, -155860.252, -210485.016, -262899.255, -308631.564, -343210.543, -346213.611, -339754.662, -326198.514, -313598.362, -302067.374, -289348.559, -275673.591, -261274.145, -246381.897, -231228.522, -217072.389, -202896.865, -188641.89, -174247.406, -159653.353, -144799.671, -129626.3, -115716.177, -101355.783},
        {-70.4665418, -69.7517256, -69.0326839, -68.3095833, -67.5825903, -66.8518715, -66.1175936, -65.3799232, -64.6390268, -63.895071, -63.1482225, -62.3986478, -61.6465136, -60.8919864, -60.1352328, -59.3764195, -58.615713, -57.8532799, -57.0892869, -56.3239005, -55.5572874, -54.789614, -54.0210471, -53.2517533, -52.481899, -51.711651, -50.9411758, -50.1706401, -49.4002103, -48.6300532, -47.8603353, -47.0912232, -46.3228835, -45.5554829, -44.7832905, -44.0095456, -43.2375585, -42.4676872, -41.7002896, -40.9357238, -40.1743476, -39.4165192, -38.6625964, -37.9129373, -37.1678998, -36.4278419, -35.6931215, -34.9640968, -34.2411256, -33.5245659, -32.8147758, -32.1490574, -31.4900237, -30.8372316, -30.1902383, -29.5486009, -28.9118765, -28.2796222, -27.6513951, -27.0267523, -26.4052509, -25.786448, -25.1699007, -24.5551661, -23.9418013, -23.3293634, -22.7174095, -22.1022869, -21.4804198, -20.8579855, -20.2347491, -19.6104758, -18.9849308, -18.3578793, -17.7290864, -17.0983174, -16.4653374, -15.8299116, -15.1918053, -14.5507834, -13.9066114, -13.2590543, -12.6078773, -11.9528456, -11.3392703, -10.7431856, -10.1400783, -9.52823567, -8.90594468, -8.27149251, -7.62316629, -6.95925315, -6.27804022, -5.57781464, -4.85686352, -4.11347401, -3.34593322, -2.5525283, -1.73154637, -0.881274559, 0, 0.508567147, 1.11446337, 1.81712199, 2.61597633, 3.51045971, 4.50000546, 5.5840469, 6.76201734, 8.03335012, 9.39747855, 10.853836, 12.4018557, 14.040971, 15.7706152, 17.5902218, 19.4992239, 21.4928579, 23.5699576, 25.7394495, 27.9998675, 30.3497457, 32.787618, 35.3120183, 37.9214806, 40.6145389, 43.3897272, 46.2455794, 49.1806294, 52.1934114, 55.5354571, 58.8595367, 62.0882847, 65.1443357, 68.78372, 72.5123743, 75.6139965, 77.7612025, 78.6266081, 77.8828293, 75.2024819, 72.4649969, 66.4645357, 56.4419461, 41.6380756, 21.2937722, 3.51584988, -23.5373017, -62.8413264, -117.371868, -190.104571, -280.013718, -398.966866, -540.939325, -710.277879, -932.5044, -1223.63358, -1599.68011, -2068.72213, -2656.51464, -3397.47067, -4287.99649, -5413.33582, -6758.63469, -8387.40049, -10495.9195, -13219.5304, -16552.2373, -20592.7366, -25957.9508, -32971.6643, -40114.6602, -49837.2631, -63101.2198, -80868.2768, -104100.181, -154654.928, -207373.954, -257994.94, -302255.573, -335893.535, -339468.467, -334014.793, -321775.071, -310465.91, -300163.058, -288661.41, -276150.758, -262820.891, -248861.602, -234462.681, -222519.503, -208961.58, -194180.606, -178568.276, -162516.284, -146416.325, -130660.095, -118489.107, -102817.757},
        {-66.4848201, -65.8435595, -65.1936498, -64.5354561, -63.8693433, -63.1956764, -62.5148203, -61.82714, -61.1330004, -60.4327665, -59.7268033, -59.0154757, -58.2991487, -57.5781872, -56.8529563, -56.1238207, -55.3911456, -54.6552959, -53.9166365, -53.1755323, -52.4323485, -51.6874498, -50.9412013, -50.193968, -49.4461147, -48.6980064, -47.9500082, -47.2024849, -46.4558016, -45.7103231, -44.9664145, -44.2244406, -43.4847665, -42.7477572, -42.0197221, -41.297929, -40.5795739, -39.864829, -39.1538662, -38.4468576, -37.7439752, -37.0453912, -36.3512774, -35.661806, -34.9771491, -34.2974786, -33.6229666, -32.9537851, -32.2901063, -31.6321021, -30.9799445, -30.3628985, -29.7514125, -29.145028, -28.5432863, -27.9457289, -27.351897, -26.761332, -26.1735753, -25.5881684, -25.0046525, -24.422569, -23.8414593, -23.2608647, -22.6803267, -22.0993866, -21.5175858, -20.9276337, -20.3224037, -19.7155277, -19.1069899, -18.4967744, -17.8848656, -17.2712477, -16.655905, -16.0388217, -15.4199821, -14.7993704, -14.1769709, -13.5527679, -12.9267456, -12.2988882, -11.6691801, -11.0376054, -10.4656304, -9.92120518, -9.37154086, -8.81462647, -8.24845107, -7.6710037, -7.0802734, -6.47424923, -5.85092023, -5.20827544, -4.54430392, -3.85699472, -3.14433687, -2.40431942, -1.63493143, -0.834161943, 0, 0.63266266, 1.32977798, 2.09401241, 2.92803237, 3.83450431, 4.81609467, 5.87546988, 7.01529637, 8.23824059, 9.54696898, 10.944148, 12.432444, 14.0145235, 15.6930529, 17.4706987, 19.3501272, 21.218395, 23.0616527, 25.1440656, 27.4435266, 29.9379285, 32.6051643, 35.4231267, 38.3697086, 41.422803, 44.5603026, 47.7601003, 51.0000889, 54.2581614, 57.6168901, 60.9111059, 64.0872977, 67.0919546, 70.8402386, 74.7943022, 78.1579851, 80.5871743, 81.7377571, 81.2656203, 78.8266512, 76.5891616, 70.9319627, 61.0193804, 46.0157404, 25.0853683, 6.59445717, -21.7950013, -63.2191483, -120.814125, -197.716072, -293.462568, -419.186561, -568.101339, -744.372987, -974.529544, -1275.05899, -1662.44929, -2144.02333, -2747.51717, -3510.66518, -4435.18139, -5600.63545, -6985.83507, -8652.4858, -10802.759, -13574.3499, -16972.6183, -21100.9504, -26547.9619, -33627.7955, -40954.8448, -50804.5009, -64079.2808, -81681.7015, -104514.28, -153410.595, -204359.337, -253280.257, -296093.106, -328717.637, -332525.227, -327682.917, -316345.477, -306055.08, -296833.982, -286459.684, -275066.679, -262789.46, -249762.517, -236120.344, -223229.08, -209240.329, -194424.001, -179050.004, -163388.245, -147708.634, -132281.079, -119590.81, -104622.294},
        {-62.2183179, -61.6622567, -61.0921709, -60.5086683, -59.9123567, -59.3038438, -58.6837374, -58.0526454, -57.4111756, -56.7599356, -56.0995333, -55.4305765, -54.753673, -54.0694305, -53.3784569, -52.68136, -51.9787474, -51.2712271, -50.5594067, -49.8438942, -49.1252972, -48.4042236, -47.6812811, -46.9570776, -46.2322208, -45.5073185, -44.7829785, -44.0598087, -43.3384166, -42.6194103, -41.9033974, -41.1909857, -40.4827831, -39.7793973, -39.102309, -38.4412505, -37.7857033, -37.1355977, -36.4908642, -35.8514332, -35.2172351, -34.5882004, -33.9642594, -33.3453426, -32.7313805, -32.1223033, -31.5180416, -30.9185258, -30.3236862, -29.7334534, -29.1477577, -28.5831859, -28.022651, -27.4657224, -26.9119693, -26.3609612, -25.8122673, -25.2654571, -24.7200998, -24.1757649, -23.6320216, -23.0884393, -22.5445874, -22.0000352, -21.454352, -20.9071071, -20.35787, -19.7976252, -19.2175633, -18.6349592, -18.0499387, -17.4626277, -16.8731522, -16.2816378, -15.6882105, -15.0929963, -14.4961208, -13.89771, -13.2978898, -12.6967859, -12.0945244, -11.491231, -10.8870316, -10.282052, -9.74768005, -9.24691167, -8.74188643, -8.23041762, -7.71031851, -7.17940239, -6.63548253, -6.07637221, -5.49988473, -4.90383334, -4.28603133, -3.64429199, -2.97642859, -2.28025441, -1.55358274, -0.794226838, 0, 0.679414183, 1.40285091, 2.17548522, 3.00249214, 3.88904672, 4.84032398, 5.86149895, 6.95774668, 8.1342422, 9.39616054, 10.7486767, 12.1969658, 13.7462028, 15.4015628, 17.1682208, 19.0513518, 20.8374371, 22.5004099, 24.5401095, 26.914848, 29.5829375, 32.50269, 35.6324178, 38.9304329, 42.3550473, 45.8645733, 49.4173228, 52.971608, 56.485741, 59.8915755, 63.1835834, 66.3280143, 69.2911176, 73.0845533, 77.1518658, 80.6578235, 83.2550531, 84.596181, 84.3338337, 82.1206379, 80.6540356, 75.6151548, 66.0608972, 51.0481645, 29.6338585, 9.72212861, -20.6891831, -64.7549866, -125.630192, -206.46971, -306.664478, -437.73379, -593.47995, -778.249192, -1017.91494, -1328.68535, -1726.76856, -2216.46473, -2828.59638, -3606.97664, -4559.51066, -5761.02623, -7190.59356, -8908.51633, -11112.7438, -13938.0436, -17387.8368, -21562.8383, -27045.3672, -34145.3201, -41639.8928, -51607.4828, -64888.537, -82323.5022, -104752.825, -151988.342, -201156.263, -248354.188, -289679.717, -321230.448, -325175.65, -320826.808, -310250.807, -300762.797, -292352.223, -282821.754, -272270.175, -260796.27, -248498.825, -235476.623, -220957.143, -206470.563, -191991.234, -177493.508, -162951.736, -148340.271, -133633.463, -119638.456, -105658.642},
        {-57.7832348, -57.3133615, -56.8256364, -56.3208226, -55.7996829, -55.2629803, -54.7114778, -54.1459384, -53.567125, -52.9758006, -52.3727282, -51.7586707, -51.134391, -50.5006522, -49.8582172, -49.2078489, -48.5503104, -47.8863646, -47.2167745, -46.5423029, -45.863713, -45.1817676, -44.4972297, -43.8108623, -43.1234284, -42.4356908, -41.7484126, -41.0623568, -40.3782862, -39.6969639, -39.0191528, -38.3456159, -37.6771162, -37.0144165, -36.389444, -35.786724, -35.1904074, -34.600246, -34.0159912, -33.4373948, -32.8642084, -32.2961835, -31.733072, -31.1746253, -30.6205951, -30.070733, -29.5247907, -28.9825198, -28.4436719, -27.9079988, -27.3752519, -26.8451113, -26.3174019, -25.7918767, -25.268289, -24.746392, -24.2259389, -23.7066828, -23.1883769, -22.6707743, -22.1536284, -21.6366923, -21.119719, -20.602462, -20.0846742, -19.566109, -19.0465194, -18.5223558, -17.9901478, -17.4564605, -16.9212613, -16.3845173, -15.846196, -15.3062645, -14.7646903, -14.2214406, -13.6764826, -13.1297838, -12.5813113, -12.0310325, -11.4789147, -10.9249252, -10.3690312, -9.81120008, -9.31884279, -8.85678618, -8.38904654, -7.91340248, -7.42763262, -6.92951558, -6.41682999, -5.88735447, -5.33886763, -4.76914809, -4.17597448, -3.55712542, -2.91037952, -2.23351541, -1.5243117, -0.780547028, 0, 0.61444239, 1.28176193, 2.00715246, 2.79580782, 3.65292183, 4.58368834, 5.59330117, 6.68695417, 7.86984117, 9.14715601, 10.5240925, 12.0058445, 13.5976059, 15.3045704, 17.1319319, 19.0848843, 20.9278763, 22.6317095, 24.7461203, 27.2247142, 30.021097, 33.0888742, 36.3816513, 39.853034, 43.4566279, 47.1460386, 50.8748716, 54.5967326, 58.2652272, 61.795562, 65.1938211, 68.42513, 71.4546138, 75.2578975, 79.2948567, 82.7559001, 85.3030032, 86.5981415, 86.3032903, 84.0804251, 83.169997, 78.5664043, 69.2314859, 54.1270807, 32.2150276, 10.5832377, -21.9640406, -68.4964864, -132.083779, -215.795598, -318.219167, -452.385177, -612.976462, -804.7871, -1053.25496, -1374.54926, -1784.83924, -2288.01981, -2915.32413, -3711.37547, -4682.22145, -5907.80379, -7379.78656, -9160.48753, -11431.9432, -14321.2373, -17818.6582, -22015.824, -27505.9357, -34598.4708, -42266.3895, -52361.3043, -65655.333, -82920.5936, -104929.204, -150372.607, -197623.764, -242974.962, -282718.487, -313146.626, -317363.378, -313641.509, -303925.429, -295019.316, -286943.92, -277695.092, -267394.827, -256165.123, -244127.977, -231405.386, -216719.745, -202594.534, -188849.669, -175305.067, -161780.645, -148096.32, -134072.009, -119141.156, -104696.178},
        {-53.2138641, -52.8294697, -52.426727, -52.0063606, -51.5690953, -51.1156556, -50.6467662, -50.1631516, -49.6655365, -49.1546455, -48.6312033, -48.0959344, -47.5495635, -46.9928153, -46.4264143, -45.8510851, -45.2675524, -44.6765409, -44.078775, -43.4749796, -42.8658791, -42.2521982, -41.6346615, -41.0139937, -40.3909194, -39.7661632, -39.1404497, -38.5145036, -37.8890494, -37.2648118, -36.6425155, -36.022885, -35.406645, -34.79452, -34.2173951, -33.6602803, -33.1078232, -32.5597696, -32.0158655, -31.4758567, -30.939489, -30.4065083, -29.8766606, -29.3496916, -28.8253472, -28.3033732, -27.7835157, -27.2655203, -26.749133, -26.2340997, -25.7201661, -25.1859562, -24.6527957, -24.1208883, -23.5904378, -23.0616481, -22.5347227, -22.0098655, -21.4872802, -20.9671705, -20.4497403, -19.9351932, -19.4237331, -18.9155636, -18.4108885, -17.9099115, -17.4128365, -16.934157, -16.4880236, -16.0451683, -15.6048683, -15.1664012, -14.7290442, -14.2920748, -13.8547704, -13.4164083, -12.976266, -12.5336209, -12.0877503, -11.6379316, -11.1834423, -10.7235597, -10.2575612, -9.78472425, -9.34578853, -8.91842818, -8.47981897, -8.02789274, -7.56058137, -7.0758167, -6.57153061, -6.04565494, -5.49612156, -4.92086233, -4.3178091, -3.68489374, -3.0200481, -2.32120405, -1.58629344, -0.81324814, 0, 0.390247936, 0.889989393, 1.50011472, 2.22151428, 3.0550784, 4.00169746, 5.06226179, 6.23766175, 7.52878769, 8.93652996, 10.4617789, 12.1054249, 13.8683583, 15.7514694, 17.7556485, 19.8817862, 22.02153, 24.1601643, 26.5456758, 29.1555456, 31.967255, 34.9582852, 38.1061175, 41.388233, 44.782113, 48.2652387, 51.8150915, 55.4091524, 59.0249027, 62.8118873, 66.5124338, 70.0524044, 73.357661, 77.1561349, 80.9726535, 84.1181597, 86.277895, 87.1371006, 86.3810183, 83.6948894, 82.6867466, 77.9251472, 68.32783, 52.812534, 30.296998, 6.95929871, -27.3582474, -75.552986, -140.522263, -225.163423, -326.542014, -460.514935, -622.12665, -816.857845, -1071.49649, -1403.26401, -1829.38184, -2360.31902, -3021.58402, -3846.05151, -4821.25911, -6051.18747, -7559.65016, -9417.49749, -11775.281, -14746.968, -18297.0323, -22500.6836, -27981.2749, -35052.0982, -42930.5641, -53194.4433, -66536.3717, -83648.985, -105224.919, -148632.814, -193720.299, -237011.011, -275028.582, -304296.648, -309137.101, -306409.813, -297867.736, -289277.734, -280803.919, -270937.944, -259927.545, -248020.461, -235464.427, -222507.181, -211023.18, -199005.185, -186515.83, -173617.747, -160373.569, -146845.93, -133097.462, -118533.145, -100424.896},
        {-49.2914599, -48.860891, -48.4215334, -47.9737279, -47.5178154, -47.0541368, -46.5830329, -46.1048447, -45.6199129, -45.1285784, -44.6311822, -44.128065, -43.6195677, -43.1060312, -42.5877964, -42.0652041, -41.5385952, -41.0083105, -40.4746909, -39.9380774, -39.3988107, -38.8572317, -38.3136812, -37.7685003, -37.2220296, -36.6746101, -36.1265826, -35.5782881, -35.0300673, -34.4822611, -33.9352104, -33.3892561, -32.8447391, -32.3020001, -31.7746411, -31.2560936, -30.7396296, -30.2251595, -29.7125939, -29.2018432, -28.6928181, -28.185429, -27.6795865, -27.175201, -26.6721831, -26.1704433, -25.6698921, -25.17044, -24.6719976, -24.1744753, -23.6777837, -23.1616147, -22.6465357, -22.1328955, -21.6210429, -21.1113269, -20.6040961, -20.0996994, -19.5984857, -19.1008037, -18.6070022, -18.1174302, -17.6324363, -17.1523695, -16.6775784, -16.2084121, -15.7452192, -15.3084787, -14.9181861, -14.5331734, -14.1524847, -13.7751637, -13.4002542, -13.0268003, -12.6538456, -12.2804342, -11.9056098, -11.5284164, -11.1478977, -10.7630977, -10.3730603, -9.97682924, -9.57344847, -9.16196184, -8.78040417, -8.40750373, -8.02152014, -7.62023195, -7.20141768, -6.76285588, -6.30232508, -5.81760381, -5.30647062, -4.76670403, -4.19608259, -3.59238483, -2.95338929, -2.2768745, -1.56061899, -0.802401316, 0, 0.183243109, 0.521527844, 1.01268827, 1.65455843, 2.4449724, 3.38176424, 4.46276801, 5.68581775, 7.04874755, 8.54939145, 10.1855835, 11.9551578, 13.8559484, 15.8857893, 18.0425146, 20.3239584, 22.6874947, 25.1251775, 27.7263577, 30.4801993, 33.3758663, 36.4025228, 39.5493329, 42.8054606, 46.16007, 49.6023251, 53.1213899, 56.7064286, 60.3466052, 64.3033117, 68.1936672, 71.9251675, 75.4053081, 79.3005184, 83.1388272, 86.245881, 88.3014953, 88.9854855, 87.9776672, 84.9578559, 83.643312, 78.4476185, 68.2606562, 51.9723057, 28.4724477, 3.26205377, -33.0328591, -83.1751829, -149.92781, -236.053632, -337.187955, -471.932752, -636.948718, -838.544479, -1102.72651, -1446.65736, -1887.49956, -2438.10853, -3120.93284, -3963.11976, -4942.09028, -6174.93475, -7696.64853, -9583.30489, -11980.5813, -15002.2056, -18628.105, -22937.3183, -28499.4878, -35606.7429, -43521.6627, -53724.319, -66863.2239, -83586.8895, -104543.828, -146099.259, -189255.809, -230732.815, -267249.611, -295525.533, -300868.617, -299017.029, -291580.659, -283716.952, -275631.559, -266051.986, -255257.792, -243528.536, -231143.777, -218383.075, -209407.743, -198465.268, -186044.65, -172634.891, -158724.991, -144803.952, -131360.775, -117819.37, -101337.482},
        {-46.0338544, -45.4485787, -44.8712927, -44.3017278, -43.7396156, -43.1846879, -42.6366762, -42.0953121, -41.5603272, -41.0314532, -40.5084216, -39.9909642, -39.4788124, -38.9716979, -38.4693523, -37.9715073, -37.4778944, -36.9882452, -36.5022914, -36.0197646, -35.5403965, -35.0639185, -34.5900623, -34.1185596, -33.6491419, -33.1815409, -32.7154882, -32.2507154, -31.7869541, -31.3239359, -30.8613924, -30.3990553, -29.9366561, -29.4739265, -28.9969028, -28.5124523, -28.027607, -27.5425428, -27.0574359, -26.5724624, -26.0877981, -25.6036193, -25.1201019, -24.6374221, -24.1557558, -23.6752792, -23.1961682, -22.7185989, -22.2427474, -21.7687898, -21.296902, -20.8292811, -20.3640385, -19.9013062, -19.4412166, -18.9839019, -18.5294944, -18.0781262, -17.6299297, -17.185037, -16.7435804, -16.3056922, -15.8715046, -15.4411498, -15.0147601, -14.5924677, -14.1744049, -13.7721029, -13.3968191, -13.0251764, -12.6565681, -12.2903873, -11.9260274, -11.5628817, -11.2003434, -10.837806, -10.4746626, -10.1103066, -9.74413118, -9.37552974, -9.00389553, -8.62862184, -8.24910197, -7.86472921, -7.54057854, -7.24182088, -6.93283771, -6.61089083, -6.27324203, -5.9171531, -5.53988584, -5.13870205, -4.71086352, -4.25363204, -3.76426941, -3.24003743, -2.67819788, -2.07601257, -1.43074329, -0.739651837, 0, 0.0821226793, 0.334196613, 0.753759272, 1.33834812, 2.08550064, 2.99275429, 4.05764655, 5.27771488, 6.65049676, 8.17352965, 9.84435102, 11.6604984, 13.6195091, 15.7189208, 17.9562708, 20.3290966, 22.7655315, 25.253198, 27.9466194, 30.8284609, 33.8813876, 37.0880645, 40.4311569, 43.8933298, 47.4572485, 51.1055779, 54.8209833, 58.5861297, 62.3836824, 66.4174085, 70.3678003, 74.1511922, 77.6839189, 81.784707, 85.9186949, 89.3103831, 91.6053885, 92.4493279, 91.4878183, 88.3664763, 86.7140091, 80.9806986, 70.0328613, 52.7368139, 27.958873, 0.81697182, -37.6370938, -90.0999114, -159.268068, -247.838152, -350.369763, -487.641683, -658.995901, -871.578189, -1148.46705, -1505.5875, -1958.86459, -2518.18741, -3206.88872, -4054.88591, -5041.75802, -6282.00893, -7797.13482, -9663.26138, -12049.081, -15080.7965, -18792.5914, -23286.8119, -29003.2071, -36190.6719, -43971.0473, -53893.6751, -66597.5842, -82721.8036, -102905.362, -142863.444, -184389.884, -224354.673, -259627.802, -287079.26, -292701.658, -291493.953, -284987.87, -278269.401, -271463.372, -263213.669, -253719.168, -243178.744, -231791.272, -219755.629, -211682.683, -200398.587, -186856.95, -172011.379, -156815.483, -142222.872, -129187.154, -116988.179, -107881.081},
        {-42.358798, -41.7974739, -41.2448403, -40.7006165, -40.1645217, -39.6362752, -39.1155962, -38.6022039, -38.0958176, -37.5961565, -37.1029398, -36.6158868, -36.1347166, -35.6591486, -35.188902, -34.7236959, -34.2632497, -33.8072826, -33.3555138, -32.9076624, -32.4634479, -32.0225894, -31.5848061, -31.1498172, -30.7173421, -30.2870999, -29.8588099, -29.4321912, -29.0069633, -28.5828451, -28.1595561, -27.7368155, -27.3143424, -26.8918561, -26.4555625, -26.0122217, -25.5687564, -25.1253242, -24.6820829, -24.2391904, -23.7968043, -23.3550824, -22.9141826, -22.4742625, -22.03548, -21.5979927, -21.1619586, -20.7275353, -20.2948806, -19.8641522, -19.4355081, -19.0156465, -18.5980429, -18.1827131, -17.7696732, -17.3589391, -16.9505268, -16.5444522, -16.1407314, -15.7393804, -15.3404151, -14.9438514, -14.5497054, -14.1579931, -13.7687303, -13.3819332, -12.9976176, -12.6208693, -12.256652, -11.894526, -11.5341787, -11.1752974, -10.8175694, -10.460682, -10.1043226, -9.74817849, -9.39193704, -9.03528556, -8.67791139, -8.31950186, -7.95974431, -7.59832605, -7.23493444, -6.8692568, -6.5909421, -6.35280463, -6.10657742, -5.84902838, -5.57692539, -5.28703638, -4.97612924, -4.64097188, -4.27833221, -3.88497812, -3.45767753, -2.99319834, -2.48830845, -1.93977576, -1.34436819, -0.698853634, 0, 0.292223165, 0.716788876, 1.27418686, 1.96490683, 2.78943852, 3.74827165, 4.84189594, 6.07080111, 7.4354769, 8.93641302, 10.5740992, 12.3490252, 14.2616806, 16.3125553, 18.5021389, 20.8309212, 23.1987975, 25.5918867, 28.2382137, 31.1167124, 34.2063165, 37.4859598, 40.9345759, 44.5310988, 48.254462, 52.0835995, 55.9974449, 59.974932, 63.9949947, 68.2708885, 72.4613073, 76.4748883, 80.2202685, 84.48999, 88.7507377, 92.2334881, 94.581707, 95.43886, 94.4484127, 91.2538309, 89.7434655, 84.0239753, 72.9083049, 55.2093991, 29.7402024, 1.02289636, -39.2661204, -93.7412126, -165.016745, -255.707081, -359.49379, -498.841895, -674.459441, -894.042488, -1179.5543, -1546.55245, -2010.59452, -2580.38964, -3279.9003, -4140.56005, -5144.67869, -6402.53944, -7932.64973, -9809.21551, -12200.8209, -15232.7288, -18945.7713, -23443.322, -29138.3342, -36268.6453, -44026.3419, -53855.897, -66357.1389, -82129.8956, -101773.996, -140223.886, -180161.121, -218601.874, -252562.318, -279058.627, -284745.732, -283908.543, -278003.762, -271926.656, -265772.831, -258205.164, -249390.898, -239497.278, -228691.55, -217140.957, -208801.909, -197360.963, -183782.305, -169030.122, -154068.601, -139861.928, -127374.289, -115611.472, -111375.383},
        {-38.288519, -37.9132339, -37.5332024, -37.1486479, -36.7597939, -36.3668639, -35.9700813, -35.5696696, -35.1658523, -34.7588529, -34.3488948, -33.9362016, -33.5209966, -33.1035033, -32.6839453, -32.262546, -31.8395288, -31.4151173, -30.9895349, -30.5630051, -30.1357513, -29.7079971, -29.2799659, -28.8518812, -28.4239664, -27.996445, -27.5695405, -27.1434764, -26.7184762, -26.2947632, -25.8725611, -25.4520931, -25.033583, -24.617254, -24.2135291, -23.8173176, -23.4234061, -23.031687, -22.6420528, -22.254396, -21.8686091, -21.4845847, -21.102215, -20.7213927, -20.3420102, -19.96396, -19.5871346, -19.2114265, -18.8367281, -18.4629319, -18.0899304, -17.7099612, -17.3307376, -16.952318, -16.5747609, -16.1981246, -15.8224677, -15.4478485, -15.0743255, -14.701957, -14.3308016, -13.9609176, -13.5923634, -13.2251976, -12.8594784, -12.4952644, -12.1326139, -11.7754171, -11.427472, -11.0809346, -10.7356149, -10.3913229, -10.0478686, -9.70506212, -9.36271341, -9.02063253, -8.67862952, -8.3365144, -7.99409721, -7.65118799, -7.30759675, -6.96313354, -6.61760839, -6.27083132, -6.02813755, -5.83435494, -5.63304302, -5.42058733, -5.19337341, -4.9477868, -4.68021303, -4.38703764, -4.06464617, -3.70942416, -3.31775714, -2.88603066, -2.41063024, -1.88794143, -1.31434976, -0.68624077, 0, 0.698520493, 1.46110372, 2.29292131, 3.1991449, 4.18494613, 5.25549665, 6.41596807, 7.67153205, 9.02736021, 10.4886242, 12.0604957, 13.7481462, 15.5567475, 17.4914711, 19.5574888, 21.7599721, 23.9903126, 26.2374276, 28.7638486, 31.5503656, 34.5777689, 37.8268486, 41.278395, 44.9131982, 48.7120484, 52.6557359, 56.7250508, 60.9007832, 65.1637235, 69.8186749, 74.4036097, 78.802114, 82.8977741, 87.2849644, 91.491877, 94.8571606, 97.0511649, 97.7442398, 96.606735, 93.3090004, 92.3002982, 87.0176141, 76.1962932, 58.5716807, 32.8791217, 2.92538012, -38.8934723, -95.109945, -168.256547, -260.865789, -366.079581, -507.298429, -685.000037, -907.053399, -1196.41808, -1569.27827, -2041.81819, -2623.99265, -3339.63101, -4219.35834, -5246.69368, -6529.17597, -8095.46158, -10017.2124, -12438.9605, -15471.7897, -19118.183, -23461.0683, -28981.5748, -35935.5862, -43785.3035, -53702.3986, -66217.4921, -81861.204, -101164.155, -138093.207, -176383.678, -213207.128, -245735.117, -271139.205, -276801.525, -276209.348, -270729.703, -264851.717, -258706.589, -251124.64, -242302.111, -232435.243, -221720.278, -210353.455, -200771.805, -189473.995, -176992.658, -163860.424, -150609.925, -137773.791, -125884.655, -113687.006, -111004.347},
        {-34.5767865, -34.316882, -34.0452934, -33.7625019, -33.4689888, -33.1652355, -32.8517232, -32.5289332, -32.1973469, -31.8574455, -31.5097104, -31.1546228, -30.7926641, -30.4243156, -30.0500585, -29.6703742, -29.2857439, -28.8966491, -28.5035709, -28.1069906, -27.7073897, -27.3052493, -26.9010509, -26.4952756, -26.0884048, -25.6809198, -25.2733019, -24.8660324, -24.4595926, -24.0544638, -23.6511274, -23.2500645, -22.8517566, -22.4566849, -22.0883415, -21.7352184, -21.3855308, -21.0390134, -20.6954006, -20.354427, -20.0158273, -19.6793358, -19.3446873, -19.0116162, -18.6798571, -18.3491446, -18.0192132, -17.6897974, -17.360632, -17.0314513, -16.7019899, -16.353787, -16.0051669, -15.6562587, -15.3071915, -14.9580941, -14.6090957, -14.2603252, -13.9119116, -13.5639839, -13.2166711, -12.8701023, -12.5244063, -12.1797123, -11.8361492, -11.493846, -11.1529317, -10.817898, -10.4931318, -10.1697644, -9.84764201, -9.52661085, -9.2065171, -8.88720694, -8.56852655, -8.25032213, -7.93243986, -7.61472593, -7.29702654, -6.97918785, -6.66105608, -6.34247739, -6.02329799, -5.70336406, -5.49764991, -5.34601612, -5.18693947, -5.01653003, -4.83089786, -4.62615301, -4.39840555, -4.14376553, -3.85834302, -3.53824806, -3.17959073, -2.77848107, -2.33102915, -1.83334502, -1.28153875, -0.671720389, 0, 0.849891146, 1.73970852, 2.67649401, 3.66728948, 4.7191368, 5.83907785, 7.03415451, 8.31140865, 9.67788214, 11.1406169, 12.7066547, 14.3830375, 16.1768072, 18.0950056, 20.1446745, 22.332856, 24.5532634, 26.7967488, 29.3266922, 32.1258509, 35.176982, 38.4628427, 41.9661903, 45.669782, 49.5563749, 53.6087263, 57.8095934, 62.1417335, 66.5879036, 71.518147, 76.3859302, 81.0578244, 85.4004011, 89.904665, 94.1249708, 97.4491578, 99.5564677, 100.126142, 98.8374217, 95.3695488, 94.5747823, 89.3849493, 78.4671791, 60.4886012, 34.1163447, 2.84430938, -40.6938383, -99.0376617, -174.726724, -270.300588, -379.200272, -524.184329, -704.504748, -927.727535, -1218.39315, -1593.42088, -2069.73, -2658.3818, -3383.4849, -4275.53847, -5318.99889, -6620.4333, -8217.24659, -10179.9078, -12634.9116, -15683.0358, -19306.2723, -23574.6029, -28993.5576, -35822.2958, -43707.0719, -53614.243, -66015.1277, -81381.0445, -100183.312, -135361.543, -171801.688, -206857.992, -237884.702, -262236.061, -268187.994, -268225.281, -263592.876, -258276.64, -252436.585, -245090.677, -236456.03, -226749.757, -216188.973, -204990.791, -193089.545, -181409.245, -169884.278, -158449.032, -147037.895, -135585.256, -124025.503, -111294.597, -110130.866},
        {-31.4352978, -31.1524338, -30.8634822, -30.5687115, -30.2683902, -29.9627868, -29.6521698, -29.3368076, -29.0169688, -28.6929218, -28.3649352, -28.0332775, -27.6982171, -27.3600225, -27.0189622, -26.6753047, -26.3293186, -25.9812722, -25.6314341, -25.2800728, -24.9274568, -24.5738546, -24.2195346, -23.8647654, -23.5098154, -23.1549532, -22.8004471, -22.4465659, -22.0935778, -21.7417514, -21.3913553, -21.0426578, -20.6959276, -20.351433, -20.0241239, -19.7066193, -19.3913618, -19.0781435, -18.7667564, -18.4569925, -18.148644, -17.8415029, -17.5353611, -17.2300109, -16.9252442, -16.6208531, -16.3166297, -16.012366, -15.707854, -15.4028859, -15.0972536, -14.7723388, -14.446743, -14.1206576, -13.7942735, -13.467782, -13.1413743, -12.8152414, -12.4895745, -12.1645649, -11.8404036, -11.5172819, -11.1953908, -10.8749216, -10.5560654, -10.2390133, -9.92395654, -9.6169643, -9.32396462, -9.03302567, -8.74395756, -8.4565704, -8.17067429, -7.88607935, -7.6025957, -7.32003343, -7.03820266, -6.7569135, -6.47597605, -6.19520044, -5.91439677, -5.63337514, -5.35194568, -5.06991848, -4.90626025, -4.7986965, -4.68352059, -4.55667576, -4.41410526, -4.25175234, -4.06556025, -3.85147225, -3.60543157, -3.32338147, -3.0012652, -2.63502602, -2.22060716, -1.75395188, -1.23100342, -0.647705047, 0, 0.612548283, 1.30785469, 2.09058759, 2.96541537, 3.93700639, 5.01002903, 6.18915166, 7.47904266, 8.88437039, 10.4098032, 12.0600096, 13.8396578, 15.7534162, 17.8059532, 20.0019372, 22.3460366, 24.7471275, 27.1961936, 29.9145762, 32.8864169, 36.0958571, 39.5270382, 43.1641017, 46.9911892, 50.9924419, 55.1520015, 59.4540093, 63.8826069, 68.4219357, 73.4569272, 78.4239765, 83.186988, 87.609866, 92.2471154, 96.6173405, 100.054985, 102.216774, 102.759431, 101.33968, 97.614246, 96.5198795, 90.8263136, 79.1572239, 60.1362861, 32.3871757, -0.325376692, -45.8040142, -106.710325, -185.705897, -285.452319, -400.578851, -551.596697, -735.940266, -960.450243, -1251.17181, -1625.52772, -2100.94073, -2688.64992, -3413.6671, -4307.74989, -5356.19565, -6666.46846, -8278.90177, -10264.6575, -12747.1345, -15825.7542, -19498.0171, -23836.9207, -29289.1715, -36090.0852, -43900.4261, -53618.4535, -65678.3661, -80514.3625, -98560.6414, -131733.693, -166114.851, -199267.743, -228755.993, -252143.227, -258760.587, -259892.506, -256630.818, -252418.703, -247417.252, -240807.832, -232800.52, -223605.397, -213432.54, -202492.028, -187699.829, -175063.676, -163937.161, -153673.877, -143627.417, -133151.374, -121599.34, -108644.141, -109682.421},
        {-28.5074737, -28.1749381, -27.8442133, -27.5152669, -27.1880669, -26.8625811, -26.5387774, -26.2166236, -25.8960876, -25.5771372, -25.2597404, -24.9438649, -24.6294786, -24.3165494, -24.0050451, -23.6949335, -23.3861826, -23.0787602, -22.7726342, -22.4677723, -22.1641425, -21.8617126, -21.5604505, -21.260324, -20.961301, -20.6633493, -20.3664368, -20.0705314, -19.7756008, -19.4816131, -19.1885359, -18.8963372, -18.6049848, -18.3144466, -18.0267601, -17.7408146, -17.455475, -17.1706419, -16.886216, -16.6020979, -16.3181885, -16.0343884, -15.7505982, -15.4667188, -15.1826506, -14.8982946, -14.6135513, -14.3283215, -14.0425058, -13.756005, -13.4687197, -13.1645364, -12.8597172, -12.55451, -12.2491625, -11.9439227, -11.6390383, -11.3347573, -11.0313275, -10.7289967, -10.4280128, -10.1286237, -9.83107713, -9.53562101, -9.24250319, -8.95197151, -8.66427383, -8.38744966, -8.12935144, -7.8741573, -7.62160999, -7.37145226, -7.12342685, -6.87727651, -6.632744, -6.38957206, -6.14750344, -5.90628089, -5.66564716, -5.42534499, -5.18511714, -4.94470636, -4.70385539, -4.46230699, -4.34089679, -4.27627406, -4.20363266, -4.11878562, -4.01754599, -3.89572684, -3.74914121, -3.57360215, -3.36492272, -3.11891597, -2.83139494, -2.4981727, -2.11506228, -1.67787676, -1.18242917, -0.624532563, 0, 0.273093771, 0.690703633, 1.25382553, 1.96345539, 2.82058916, 3.82622279, 4.9813522, 6.28697335, 7.74408216, 9.35367459, 11.1167466, 13.034294, 15.1073129, 17.3367992, 19.7237487, 22.2691576, 24.9110964, 27.6415719, 30.6039107, 33.7856248, 37.1742262, 40.7572269, 44.5221388, 48.4564739, 52.5477441, 56.7834615, 61.1511381, 65.6382857, 70.2324165, 75.3382985, 80.3731935, 85.1994368, 89.6793635, 94.463437, 99.0199284, 102.606941, 104.850371, 105.376116, 103.810071, 99.778134, 98.1266303, 91.6722011, 79.0193546, 58.7725991, 29.5364428, -4.27590288, -51.3208085, -114.445942, -196.49897, -300.32756, -421.98126, -579.414672, -768.435106, -995.136485, -1286.61456, -1660.53569, -2134.56623, -2719.49676, -3441.43209, -4333.97791, -5384.02841, -6698.7644, -8318.47579, -10314.672, -12811.7622, -15909.8737, -19629.1917, -24049.1177, -29542.1745, -36315.0788, -44000.7814, -53468.0368, -65122.173, -79368.518, -96612.4001, -127841.715, -160233.581, -191547.684, -219543.711, -241981.347, -249167.766, -251283.509, -249282.194, -246145.417, -242024.058, -236224.006, -228938.392, -220360.343, -210682.99, -200099.461, -183486.18, -170192.249, -159229.884, -149611.299, -140348.709, -130454.33, -118940.376, -106327.096, -108882.613},
        {-25.3925761, -25.0847166, -24.7796127, -24.4771899, -24.1773735, -23.8800888, -23.5852612, -23.2928161, -23.0026786, -22.7147743, -22.4290284, -22.1453662, -21.8637132, -21.5839946, -21.3061358, -21.0300621, -20.7556989, -20.4829715, -20.2118053, -19.9421255, -19.6738576, -19.4069268, -19.1412585, -18.8767781, -18.6134108, -18.3510821, -18.0897172, -17.8292415, -17.5695804, -17.3106591, -17.0524031, -16.7947376, -16.537588, -16.2808797, -16.0243384, -15.7679936, -15.5118768, -15.2559198, -15.0000546, -14.7442127, -14.4883262, -14.2323268, -13.9761463, -13.7197165, -13.4629693, -13.2058364, -12.9482496, -12.6901409, -12.4314419, -12.1720845, -11.9120005, -11.6359669, -11.3593988, -11.0825566, -10.8057007, -10.5290913, -10.2529889, -9.97765373, -9.70334615, -9.4303265, -9.15885513, -8.88919237, -8.62159857, -8.35633406, -8.09365918, -7.83383428, -7.57711969, -7.33229996, -7.10795497, -6.88676452, -6.66843636, -6.45267826, -6.23919794, -6.02770318, -5.81790171, -5.60950129, -5.40220966, -5.19573458, -4.9897838, -4.78406507, -4.57828613, -4.37215474, -4.16537865, -3.9576656, -3.87428047, -3.84951153, -3.81613793, -3.76979286, -3.7061095, -3.62072103, -3.50926063, -3.36736147, -3.19065675, -2.97477963, -2.7153633, -2.40804095, -2.04844574, -1.63221086, -1.15496949, -0.612354809, 0, 0.0752843512, 0.33707251, 0.783593421, 1.41307603, 2.22374928, 3.21384211, 4.38158348, 5.72520232, 7.24292758, 8.9329882, 10.7936131, 12.8230313, 15.0194717, 17.3811632, 19.9063349, 22.5932155, 25.452763, 28.4850246, 31.6594379, 34.9769594, 38.4385458, 42.0451535, 45.7977391, 49.6972592, 53.7446703, 57.9409289, 62.2869917, 66.7838151, 71.4323557, 76.7623021, 82.0520104, 87.1438175, 91.8800604, 96.8845636, 101.60892, 105.296327, 107.554674, 107.991852, 106.215753, 101.834266, 99.3949165, 92.0625941, 78.4787395, 57.284793, 27.1221953, -6.32321337, -53.2154921, -116.6743, -199.819298, -305.770143, -431.498556, -593.786285, -788.729034, -1022.30925, -1320.6948, -1700.99072, -2180.30204, -2767.48768, -3489.77289, -4383.37868, -5440.08323, -6759.53573, -8377.13896, -10360.9399, -12834.2991, -15895.1873, -19569.3748, -23936.1137, -29339.3912, -35973.4456, -43501.6739, -52730.8231, -64038.8839, -77803.8472, -94403.7036, -124233.939, -155150.676, -185027.53, -211738.12, -233156.06, -240066.564, -242157.258, -240339.644, -237498.645, -233764.62, -228445.676, -221713.287, -213738.927, -204694.07, -194750.19, -181415.586, -169386.537, -158304.477, -147810.841, -137547.061, -127154.572, -116274.807, -105703.968, -107126.244},
        {-22.3891136, -22.1261242, -21.8646584, -21.6046805, -21.3461548, -21.0890458, -20.8333177, -20.5789347, -20.3258614, -20.0740619, -19.8235007, -19.574142, -19.3259502, -19.0788895, -18.8329244, -18.5880191, -18.344138, -18.1012454, -17.8593057, -17.618283, -17.3781419, -17.1388466, -16.9003614, -16.6626507, -16.4256787, -16.1894099, -15.9538085, -15.7188389, -15.4844654, -15.2506523, -15.0173639, -14.7845646, -14.5522188, -14.3202906, -14.0894868, -13.8593848, -13.6295534, -13.3999328, -13.1704633, -12.941085, -12.7117383, -12.4823633, -12.2529003, -12.0232896, -11.7934713, -11.5633857, -11.3329731, -11.1021737, -10.8709278, -10.6391755, -10.4068571, -10.1591258, -9.91102949, -9.6628289, -9.41478484, -9.16715811, -8.9202095, -8.67419981, -8.42938981, -8.18604032, -7.94441212, -7.70476601, -7.46736277, -7.23246321, -7.00032811, -6.77121828, -6.54539449, -6.33206281, -6.14021452, -5.95166148, -5.7660846, -5.5831648, -5.40258297, -5.22402003, -5.04715688, -4.87167444, -4.69725362, -4.52357531, -4.35032043, -4.1771699, -4.00380461, -3.82990548, -3.65515341, -3.47922931, -3.43277819, -3.44724425, -3.45249761, -3.44396915, -3.41708974, -3.36729025, -3.29000156, -3.18065452, -3.03468003, -2.84750894, -2.61457213, -2.33130048, -1.99312484, -1.59547611, -1.13378514, -0.603482815, 0, -0.0527767014, 0.112013662, 0.490442372, 1.07858071, 1.87249996, 2.86827139, 4.0619663, 5.44965597, 7.02741167, 8.79130468, 10.7374063, 12.8617878, 15.1605204, 17.6296755, 20.2653244, 23.0635382, 26.106343, 29.4020892, 32.7524269, 36.1718463, 39.6748374, 43.2758905, 46.9894956, 50.8301428, 54.8123224, 58.9505245, 63.2592391, 67.7529564, 72.4461666, 77.9978155, 83.5416371, 88.8987849, 93.8904122, 99.1024025, 103.973544, 107.738698, 109.989597, 110.317978, 108.315573, 103.574119, 100.262059, 92.0015057, 77.4887494, 55.4200811, 24.4917915, -8.27923539, -54.59745, -117.84671, -201.410872, -308.673794, -437.219911, -603.320685, -804.096216, -1045.67108, -1352.554, -1740.99961, -2227.2625, -2817.90824, -3541.08257, -4435.50115, -5498.46929, -6820.06696, -8429.26118, -10389.4849, -12822.5714, -15823.5476, -19418.2949, -23683.9137, -28951.0356, -35406.8807, -42759.7115, -51749.3911, -62728.9326, -76051.3488, -92069.6527, -120733.678, -150400.412, -179023.662, -204557.235, -224954.94, -231279.887, -232970.378, -230933.375, -228105.584, -224590.527, -219646.003, -213417.823, -206051.797, -197693.735, -188489.448, -180023.601, -169855.428, -158559.033, -146708.519, -134877.991, -123641.552, -113573.307, -105691.521, -105094.044},
        {-19.6174481, -19.4049132, -19.1916939, -18.9778317, -18.7633678, -18.5483435, -18.3328004, -18.1167795, -17.9003224, -17.6834703, -17.4662645, -17.2487464, -17.0309574, -16.8129387, -16.5947317, -16.3763778, -16.1579182, -15.9393943, -15.7208474, -15.5023189, -15.2838502, -15.0654824, -14.847257, -14.6292153, -14.4113987, -14.1938484, -13.9766059, -13.7597124, -13.5432092, -13.3271378, -13.1115394, -12.8964554, -12.6819271, -12.4679959, -12.2580086, -12.0502844, -11.8431024, -11.6363968, -11.4301015, -11.2241507, -11.0184784, -10.8130186, -10.6077055, -10.402473, -10.1972553, -9.99198634, -9.78660027, -9.58103112, -9.37521294, -9.1690798, -8.96256577, -8.74041642, -8.51808356, -8.29583048, -8.07392051, -7.85261696, -7.63218315, -7.41288239, -7.19497801, -6.97873331, -6.76441161, -6.55227624, -6.34259049, -6.1356177, -5.93162118, -5.73086424, -5.5336102, -5.3494229, -5.18764289, -5.02935183, -4.87421014, -4.7218782, -4.57201643, -4.42428523, -4.27834498, -4.13385611, -3.990479, -3.84787405, -3.70570167, -3.56362226, -3.42129622, -3.27838395, -3.13454585, -2.98944232, -2.98004328, -3.03446627, -3.07917842, -3.1093842, -3.12028804, -3.10709438, -3.06500767, -2.98923235, -2.87497286, -2.71743365, -2.51181915, -2.25333381, -1.93718207, -1.55856838, -1.11269718, -0.594772902, 0, -0.102567403, 0.0293045919, 0.390479873, 0.975822329, 1.78019585, 2.79846432, 4.02549163, 5.45614166, 7.08527831, 8.90776547, 10.918467, 13.1122468, 15.4839688, 18.0284969, 20.7406949, 23.6154267, 26.7776078, 30.240681, 33.7053457, 37.194334, 40.7303782, 44.3362101, 48.034562, 51.8481659, 55.7997539, 59.9120581, 64.2078105, 68.7097433, 73.4405884, 79.1433177, 84.8563357, 90.3863026, 95.5398788, 100.897142, 105.878044, 109.696291, 111.926518, 112.14336, 109.921452, 104.835429, 100.681318, 91.5275896, 76.1229547, 53.2161232, 21.5558053, -10.8104592, -56.757497, -119.861305, -203.697881, -311.843222, -442.139601, -610.904624, -816.579288, -1065.5668, -1380.52653, -1776.81918, -2269.80542, -2864.01311, -3588.09778, -4482.53263, -5549.4677, -6869.7134, -8466.23101, -10397.7864, -12783.8141, -15715.9007, -19217.0453, -23361.0525, -28473.215, -34736.5698, -41911.6181, -50668.9378, -61337.8282, -74247.5885, -89727.518, -117348.956, -145888.652, -173365.394, -197797.972, -217205.175, -222843.148, -224001.753, -221598.053, -218679.999, -215321.178, -210707.313, -204956.387, -198186.383, -190515.284, -182061.073, -178465.235, -170179.153, -158702.157, -145533.578, -132172.747, -120118.995, -110871.652, -105829.108, -103139.13},
        {-17.4163321, -17.2304457, -17.0428621, -16.8536612, -16.6629228, -16.4707268, -16.277153, -16.0822815, -15.8861919, -15.6889642, -15.4906783, -15.291414, -15.0912512, -14.8902698, -14.6885496, -14.4861706, -14.2832126, -14.0797554, -13.875879, -13.6716632, -13.4671878, -13.2625329, -13.0577781, -12.8530035, -12.6482889, -12.443714, -12.239359, -12.0353035, -11.8316274, -11.6284108, -11.4257333, -11.2236749, -11.0223155, -10.8217349, -10.6271358, -10.435929, -10.2454635, -10.055653, -9.86641119, -9.6776516, -9.48928789, -9.30123367, -9.11340259, -8.92570827, -8.73806433, -8.5503844, -8.3625821, -8.17457107, -7.98626493, -7.7975773, -7.60842182, -7.40019243, -7.1917239, -6.98333134, -6.77532984, -6.56803449, -6.36176041, -6.15682267, -5.95353638, -5.75221665, -5.55317855, -5.3567372, -5.16320768, -4.9729051, -4.78614455, -4.60324113, -4.42450994, -4.26052835, -4.12162736, -3.98695687, -3.85616672, -3.72890674, -3.60482676, -3.48357663, -3.36480617, -3.24816522, -3.13330362, -3.01987121, -2.9075178, -2.79589325, -2.68464739, -2.57343004, -2.46189105, -2.34968025, -2.38213009, -2.48298485, -2.57423784, -2.65081122, -2.70762712, -2.73960772, -2.74167517, -2.70875161, -2.63575921, -2.51762011, -2.34925648, -2.12559046, -1.84154421, -1.49203989, -1.07199964, -0.576345626, 0, 0.0573234949, 0.322305386, 0.791922773, 1.46315275, 2.33297242, 3.39835888, 4.65628923, 6.10374056, 7.73768998, 9.55511458, 11.5529915, 13.7282977, 16.0780105, 18.5991068, 21.2885638, 24.1433585, 27.1905456, 30.4313988, 33.794863, 37.2843605, 40.9033135, 44.6551442, 48.543275, 52.5711281, 56.7421258, 61.0596902, 65.5272438, 70.1482087, 74.9260072, 80.4389682, 85.9048082, 91.1544776, 96.0189267, 101.287118, 106.310995, 110.187034, 112.458779, 112.669779, 110.363579, 105.083727, 100.484097, 90.7469376, 74.611602, 50.8174416, 18.1038087, -16.0752363, -63.873727, -128.730988, -214.086346, -323.379124, -453.84104, -622.706104, -828.997286, -1079.34179, -1395.89212, -1793.83023, -2288.3381, -2884.55509, -3609.98817, -4503.09986, -5565.00138, -6875.45805, -8459.135, -10373.5052, -12731.6328, -15621.0645, -19061.7401, -23123.6788, -28124.7921, -34242.7695, -41295.9894, -49876.7311, -60287.0611, -72829.046, -87804.7523, -114281.85, -141605.401, -167886.071, -191234.526, -209761.431, -215092.116, -216146.14, -213813.598, -211120.052, -208110.309, -203939.399, -198696.616, -192471.259, -185352.622, -177430.001, -173523.751, -165361.97, -154298.563, -141687.439, -128882.503, -117237.663, -108106.825, -104874.992, -102443.052},
        {-15.5373657, -15.3712542, -15.2025251, -15.031296, -14.8576847, -14.6818087, -14.5037856, -14.3237331, -14.1417689, -13.9580105, -13.7725756, -13.5855818, -13.3971468, -13.2073882, -13.0164236, -12.8243707, -12.6313471, -12.4374704, -12.2428583, -12.0476284, -11.8518983, -11.6557856, -11.4594081, -11.2628833, -11.0663289, -10.8698625, -10.6736017, -10.4776642, -10.2821675, -10.0872295, -9.89296756, -9.69949946, -9.50694279, -9.3154152, -9.13212999, -8.95350771, -8.77588363, -8.5991451, -8.4231795, -8.24787419, -8.07311653, -7.8987939, -7.72479366, -7.55100317, -7.3773098, -7.20360091, -7.02976388, -6.85568606, -6.68125482, -6.50635753, -6.33088156, -6.13215013, -5.9331039, -5.73411937, -5.53557305, -5.33784146, -5.1413011, -4.9463285, -4.75330015, -4.56259258, -4.3745823, -4.18964581, -4.00815962, -3.83050026, -3.65704423, -3.48816804, -3.3242482, -3.17695105, -3.05767183, -2.94350266, -2.8340882, -2.72907308, -2.62810193, -2.5308194, -2.43687013, -2.34589875, -2.25754991, -2.17146824, -2.08729838, -2.00468498, -1.92327267, -1.84270608, -1.76262987, -1.68268867, -1.75735222, -1.90559608, -2.04453492, -2.16878899, -2.27297857, -2.35172392, -2.39964531, -2.411363, -2.38149727, -2.30466837, -2.17549659, -1.98860217, -1.73860539, -1.42012652, -1.02778582, -0.556203557, 0, 0.271194116, 0.712558045, 1.3239661, 2.1052926, 3.05641186, 4.1771982, 5.46752593, 6.92726936, 8.55630282, 10.3545006, 12.3217371, 14.4578865, 16.7628232, 19.2364215, 21.8785558, 24.6891003, 27.5634518, 30.4865595, 33.6946155, 37.1651063, 40.8755181, 44.8033372, 48.92605, 53.2211425, 57.6661013, 62.2384125, 66.9155625, 71.6750375, 76.4943238, 81.7308768, 86.8428917, 91.6938641, 96.1472896, 101.235167, 106.23674, 110.119653, 112.396849, 112.581272, 110.185869, 104.723583, 99.8094192, 89.6058087, 72.8231191, 48.1717179, 14.3619727, -22.2381127, -72.490216, -139.598379, -226.766643, -337.199049, -467.245775, -635.341672, -840.980301, -1091.01498, -1407.32518, -1805.04396, -2299.30434, -2896.35113, -3622.22566, -4512.66553, -5566.63678, -6864.39, -8433.75229, -10331.5255, -12663.6078, -15513.3611, -18897.2729, -22881.3673, -27776.8758, -33756.3269, -40695.4529, -49109.2391, -59272.951, -71461.8539, -85951.2133, -111294.889, -137419.904, -162531.874, -184836.416, -202539.148, -207664.65, -208720.265, -206560.624, -204152.229, -201513.763, -197783.963, -193025.511, -187301.088, -180673.374, -173205.05, -168043.873, -159652.012, -149065.045, -137318.552, -125448.111, -114489.301, -105477.701, -103683.944, -102043.484},
        {-13.9407675, -13.7880307, -13.631909, -13.4725542, -13.3101181, -13.1447523, -12.9766086, -12.8058389, -12.6325947, -12.4570278, -12.2792901, -12.0995332, -11.9179088, -11.7345688, -11.5496648, -11.3633486, -11.1757719, -10.9870865, -10.7974441, -10.6069965, -10.4158954, -10.2242925, -10.0323396, -9.84018848, -9.64799079, -9.4558983, -9.26406276, -9.07263591, -8.88176949, -8.69161524, -8.50232491, -8.31405024, -8.12694297, -7.94115483, -7.76580181, -7.59636499, -7.42821769, -7.26122074, -7.09523497, -6.93012123, -6.76574035, -6.60195315, -6.43862049, -6.27560318, -6.11276206, -5.94995798, -5.78705176, -5.62390424, -5.46037625, -5.29632863, -5.13162221, -4.93945738, -4.74693336, -4.55448895, -4.36256291, -4.17159404, -3.9820211, -3.79428289, -3.60881817, -3.42606574, -3.24646437, -3.07045285, -2.89846995, -2.73095446, -2.56834515, -2.41108081, -2.25960022, -2.12661553, -2.02454421, -1.92851214, -1.83816246, -1.75313833, -1.67308289, -1.59763928, -1.52645066, -1.45916018, -1.39541098, -1.33484621, -1.27710902, -1.22184256, -1.16868998, -1.11729441, -1.06729902, -1.01834694, -1.1343942, -1.3294714, -1.51563425, -1.68719363, -1.83846042, -1.96374549, -2.05735974, -2.11361403, -2.12681925, -2.09128628, -2.001326, -1.85124928, -1.63536701, -1.34799007, -0.983429335, -0.535995685, 0, 0.485355444, 1.10345632, 1.85706504, 2.74894401, 3.78185562, 4.95856229, 6.28182642, 7.75441041, 9.37907667, 11.1585876, 13.0957056, 15.1931931, 17.4538125, 19.8803262, 22.4754966, 25.242086, 27.9389921, 30.5341394, 33.5818894, 37.0327478, 40.8372203, 44.9458124, 49.3090298, 53.8773782, 58.6013631, 63.4314903, 68.3182654, 73.2121941, 78.0637818, 83.0051613, 87.7386147, 92.1601596, 96.1658136, 101.013235, 105.91762, 109.731063, 111.94109, 112.035225, 109.500993, 103.825921, 98.6870103, 88.1072515, 70.7544903, 45.296572, 10.4013421, -28.8826552, -81.8634482, -151.468366, -240.624739, -352.259896, -481.774182, -648.821559, -853.291475, -1102.23415, -1417.31973, -1813.6543, -2306.34395, -2902.69374, -3627.38377, -4513.60433, -5558.30172, -6842.24231, -8396.42946, -10277.0589, -12583.0134, -15393.5816, -18722.2666, -22631.3016, -27424.1339, -33267.7667, -40090.1748, -48334.1629, -58249.4745, -70085.8532, -84093.0425, -108306.778, -133244.313, -157208.893, -178503.764, -195432.171, -200446.582, -201601.538, -199707.763, -197629.927, -195365.731, -192057.063, -187744.413, -182468.273, -176269.133, -169187.484, -162716.655, -154013.014, -143821.917, -132888.722, -121958.784, -111777.458, -103090.102, -102365.835, -101739.831},
        {-12.5705434, -12.423385, -12.2724856, -12.118018, -11.960155, -11.7990694, -11.6349342, -11.467922, -11.2982057, -11.1259582, -10.9513522, -10.7745606, -10.5957561, -10.4151117, -10.2328001, -10.0489941, -9.86386661, -9.67759039, -9.49033827, -9.30228308, -9.11359762, -8.92445473, -8.73502721, -8.5454879, -8.3560096, -8.16676513, -7.97792733, -7.789669, -7.60216297, -7.41558205, -7.23009906, -7.04588683, -6.86311817, -6.6819659, -6.5126279, -6.35005359, -6.18907192, -6.02953036, -5.8712764, -5.71415753, -5.55802124, -5.40271501, -5.24808634, -5.0939827, -4.94025158, -4.78674047, -4.63329686, -4.47976824, -4.32600209, -4.17184589, -4.01714715, -3.83254215, -3.64772282, -3.46316987, -3.27936404, -3.09678606, -2.91591664, -2.73723653, -2.56122644, -2.38836711, -2.21913926, -2.05402363, -1.89350093, -1.7380519, -1.58815727, -1.44429776, -1.30695411, -1.1897571, -1.10602181, -1.02910793, -0.958643731, -0.894257471, -0.835577414, -0.782231823, -0.733848959, -0.690057085, -0.650484464, -0.614759359, -0.582510031, -0.553364743, -0.526951759, -0.50289934, -0.480835749, -0.460389248, -0.613574402, -0.850200409, -1.07800452, -1.29102072, -1.48328301, -1.64882539, -1.78168185, -1.87588639, -1.925473, -1.92447568, -1.86692843, -1.74686523, -1.5583201, -1.29532701, -0.951919964, -0.522132963, 0, 0.509527961, 1.15302778, 1.93328801, 2.85309717, 3.91524381, 5.12251647, 6.47770368, 7.98359399, 9.64297593, 11.458638, 13.4333689, 15.5699569, 17.8711908, 20.339859, 22.97875, 25.7906524, 28.4875244, 31.0306072, 34.0805201, 37.5777308, 41.4627072, 45.6759169, 50.1578276, 54.8489071, 59.6896231, 64.6204433, 69.5818355, 74.5142675, 79.3582068, 84.182856, 88.7527454, 92.9697223, 96.735634, 101.344197, 106.001323, 109.541762, 111.449799, 111.209721, 108.305815, 102.222367, 96.8313317, 85.8939512, 68.0360562, 41.8834769, 6.06204347, -35.1993599, -90.2787203, -161.950971, -252.991045, -366.173877, -496.507804, -664.026859, -869.102042, -1118.84361, -1434.12113, -1829.62651, -2320.05168, -2911.37316, -3628.97904, -4507.76503, -5548.3054, -6825.42549, -8367.57945, -10228.0351, -12502.1055, -15265.8043, -18532.7353, -22362.7316, -27044.4631, -32737.1566, -39402.5325, -47427.7097, -57040.283, -68467.8468, -81937.9957, -104955.761, -128651.941, -151434.772, -171712.488, -187893.325, -192917.134, -194306.123, -192812.123, -191106.303, -189180.754, -186227.837, -182279.627, -177368.2, -171525.629, -164783.99, -161263.437, -153399.825, -142689.331, -130628.129, -118712.396, -108438.307, -101302.038, -101098.805, -100921.984},
        {-11.3905937, -11.2421155, -11.089967, -10.9343276, -10.7753768, -10.613294, -10.4482587, -10.2804504, -10.1100486, -9.93723262, -9.76218204, -9.5850763, -9.40609487, -9.22541721, -9.04322278, -8.85969106, -8.67500151, -8.4893336, -8.30286679, -8.11578055, -7.92825435, -7.74046765, -7.55259991, -7.36483061, -7.17733922, -6.99030519, -6.80390799, -6.61832709, -6.43374196, -6.25033206, -6.06827686, -5.88775583, -5.70894842, -5.53203412, -5.36731131, -5.20968717, -5.05394722, -4.89994256, -4.74752427, -4.59654344, -4.44685116, -4.29829851, -4.15073658, -4.00401647, -3.85798925, -3.71250602, -3.56741785, -3.42257585, -3.27783109, -3.13303467, -2.98803767, -2.81292363, -2.63795649, -2.46363264, -2.29044847, -2.11890036, -1.94948471, -1.78269789, -1.61903632, -1.45899636, -1.30307442, -1.15176688, -1.00557013, -0.86498056, -0.730494559, -0.602608516, -0.481818821, -0.382421394, -0.318380833, -0.261733031, -0.21207981, -0.169022992, -0.132164399, -0.101105853, -0.075449175, -0.054796188, -0.0387487135, -0.0269085733, -0.0188775894, -0.0142575836, -0.0126503779, -0.0136577941, -0.0168816543, -0.0219237801, -0.207310492, -0.479417927, -0.742471686, -0.990267148, -1.21659969, -1.41526469, -1.58005753, -1.70477358, -1.78320822, -1.80915683, -1.77641479, -1.67877747, -1.51004026, -1.26399853, -0.934447658, -0.515183022, 0, 0.321040077, 0.820384759, 1.49766135, 2.35249714, 3.38451945, 4.59335557, 5.9786328, 7.53997845, 9.27701981, 11.1893842, 13.2766989, 15.5385912, 17.9746884, 20.5846179, 23.3680069, 26.3244827, 29.2227172, 32.0293442, 35.26639, 38.8839914, 42.8322854, 47.0614087, 51.5214984, 56.1626911, 60.9351238, 65.7889333, 70.6742565, 75.5412303, 80.3399915, 85.2648581, 89.9322527, 94.219058, 98.0021564, 102.397599, 106.662685, 109.724266, 111.087475, 110.257444, 106.739303, 100.038186, 94.3436989, 83.0507889, 64.7480596, 38.0241151, 1.46755953, -40.8610535, -97.213057, -170.367835, -263.10477, -378.203245, -510.943962, -680.770044, -888.628304, -1141.53817, -1458.91733, -1854.6169, -2342.488, -2924.66058, -3629.37084, -4497.64173, -5539.58073, -6817.24794, -8350.3546, -10186.7776, -12422.1188, -15130.0737, -18327.8944, -22074.6428, -26636.3395, -32161.9376, -38625.8148, -46378.3573, -55628.948, -66586.9699, -79461.806, -101224.23, -123631.886, -145203.81, -164459.038, -179916.604, -185054.47, -186793.463, -185813.077, -184502.937, -182865.613, -180191.121, -176517.66, -171883.427, -166326.619, -159885.435, -163687.147, -157869.866, -145738.263, -130597.01, -115750.778, -104504.239, -100162.063, -99977.2221, -99482.3417},
        {-10.3900562, -10.2371688, -10.0808168, -9.92118058, -9.75844059, -9.5927772, -9.42437083, -9.25340185, -9.08005068, -8.9044977, -8.72692332, -8.54750794, -8.36643195, -8.18387574, -8.00001973, -7.8150443, -7.62912985, -7.44245678, -7.25520549, -7.06755637, -6.87968983, -6.69178626, -6.50402605, -6.31658962, -6.12965734, -5.94340963, -5.75802688, -5.57368948, -5.39057784, -5.20887235, -5.02875341, -4.85040141, -4.67399676, -4.49971986, -4.33752827, -4.18250817, -4.02962819, -3.87875144, -3.72974102, -3.58246006, -3.43677166, -3.29253892, -3.14962496, -3.00789289, -2.86720581, -2.72742684, -2.58841908, -2.45004564, -2.31216964, -2.17465418, -2.03736237, -1.87081935, -1.7048622, -1.53999, -1.37670186, -1.21549687, -1.05687414, -0.901332762, -0.749371836, -0.601490461, -0.458187735, -0.319962759, -0.18731463, -0.0607424468, 0.0592546908, 0.172177684, 0.277527435, 0.36077637, 0.407733735, 0.446833105, 0.478484792, 0.503099104, 0.521086352, 0.532856846, 0.538820896, 0.539388811, 0.534970903, 0.525977481, 0.512818855, 0.495905335, 0.475647232, 0.452454854, 0.426738513, 0.398908518, 0.183489925, -0.122254482, -0.418827723, -0.699787154, -0.95869013, -1.18909401, -1.38455614, -1.53863389, -1.64488461, -1.69686566, -1.68813438, -1.61224815, -1.46276431, -1.23324022, -0.91723323, -0.508300706, 0, 0.109846076, 0.446592952, 1.00640015, 1.78542721, 2.77983364, 3.98577899, 5.39942277, 7.01692451, 8.83444374, 10.84814, 13.0541728, 15.4487016, 18.0278861, 20.7878857, 23.7248599, 26.8349683, 29.9594647, 33.0723792, 36.5202534, 40.2660527, 44.2727424, 48.5032882, 52.9206554, 57.4878094, 62.1677158, 66.9233399, 71.7176471, 76.5136031, 81.274173, 86.3533783, 91.1797413, 95.598911, 99.4565359, 103.64104, 107.476684, 110.009657, 110.772773, 109.29885, 105.120704, 97.7711499, 91.8161868, 80.2236107, 61.5414847, 34.317872, -2.89916418, -46.1524427, -103.655298, -178.211946, -272.626604, -389.703487, -525.107429, -697.507942, -908.337968, -1164.4801, -1483.94333, -1879.74113, -2364.88694, -2937.82578, -3629.45058, -4486.54719, -5528.24021, -6804.39686, -8326.01059, -10135.8116, -12330.3062, -14981.3168, -18111.0211, -21778.3708, -26224.3464, -31586.9624, -37843.2683, -45317.0078, -54201.9142, -64691.7208, -76980.1608, -97531.1552, -118700.703, -139114.99, -157400.203, -172182.529, -177469.708, -179585.336, -179134.933, -178208.781, -176824.683, -174381.788, -170928.829, -166514.542, -161187.658, -154996.913, -165273.563, -161267.715, -147902.946, -130102.83, -112790.945, -100890.865, -99326.1661, -98989.1801, -97950.7805},
        {-9.54228205, -9.38388377, -9.22226239, -9.05759593, -8.89006237, -8.71983972, -8.54710598, -8.37203916, -8.19481724, -8.01561824, -7.83462015, -7.65200097, -7.4679387, -7.28261134, -7.0961969, -6.90887338, -6.72081877, -6.53221107, -6.34322829, -6.15404842, -5.96484948, -5.77580944, -5.58710633, -5.39891813, -5.21142285, -5.02479849, -4.83922305, -4.65487453, -4.47193093, -4.29057025, -4.11097048, -3.93330964, -3.75776573, -3.58451673, -3.42283984, -3.26817206, -3.11584107, -2.96572958, -2.81772029, -2.67169593, -2.5275392, -2.38513283, -2.24435953, -2.10510201, -1.96724299, -1.83066518, -1.69525129, -1.56088405, -1.42744616, -1.29482034, -1.1628893, -1.00351643, -0.845211166, -0.688463645, -0.533763979, -0.381602289, -0.232468693, -0.0868533102, 0.0547537401, 0.191862339, 0.323982367, 0.450623706, 0.571296237, 0.68550984, 0.792774396, 0.892599787, 0.984495894, 1.05428337, 1.08811155, 1.11372331, 1.13151595, 1.14188676, 1.14523305, 1.14195211, 1.13244124, 1.11709774, 1.0963189, 1.07050202, 1.0400444, 1.00534334, 0.966796131, 0.924800078, 0.879752475, 0.83205062, 0.588165764, 0.249537201, -0.0800649314, -0.393950063, -0.685427624, -0.947807043, -1.17439775, -1.35850917, -1.49345074, -1.57253188, -1.58906202, -1.5363506, -1.40770704, -1.19644077, -0.895861221, -0.499277821, 0, -0.0671047531, 0.133792737, 0.595952898, 1.31263616, 2.27710294, 3.48261368, 4.92242879, 6.58980871, 8.47801386, 10.5803047, 12.8899416, 15.400185, 18.1042954, 20.9955331, 24.0671586, 27.3124323, 30.6370015, 34.0216102, 37.6576919, 41.5197327, 45.5822187, 49.819636, 54.206471, 58.7172096, 63.3263381, 68.0083425, 72.7377092, 77.4889242, 82.2364736, 87.487252, 92.4881208, 97.0538437, 100.999184, 104.98121, 108.393531, 110.405148, 110.578132, 108.474556, 103.656493, 95.6860157, 89.5338828, 77.7073593, 58.7102968, 31.0465469, -6.78003884, -50.9754819, -109.648589, -185.618041, -281.702518, -400.720699, -538.693036, -713.545384, -927.04362, -1185.91802, -1507.03454, -1902.7084, -2385.25485, -2950.51816, -3631.02792, -4477.28208, -5514.66277, -6783.9566, -8289.50021, -10070.0469, -12222.6667, -14817.6616, -17882.8519, -21477.5257, -25816.7359, -31027.4258, -37081.8677, -44285.472, -52818.7221, -62862.1016, -74596.094, -94022.4129, -114045.004, -133389.044, -150779.709, -164942.174, -170378.03, -172851.762, -172898.655, -172320.432, -171151.078, -168896.897, -165618.258, -161375.533, -156229.09, -150239.3, -164259.207, -161324.303, -147330.826, -128175.011, -109753.094, -97961.3116, -98695.9, -98189.2517, -96458.4121},
        {-8.81868184, -8.65657673, -8.49136239, -8.32321551, -8.15231276, -7.9788308, -7.80294633, -7.624836, -7.44467649, -7.26264449, -7.07891666, -6.89366968, -6.70708022, -6.51932496, -6.33058057, -6.14102373, -5.95083112, -5.76017939, -5.56924525, -5.37820534, -5.18723636, -4.99651498, -4.80621786, -4.61652169, -4.42760314, -4.23963888, -4.05280559, -3.86727994, -3.68323862, -3.50085828, -3.32031562, -3.14178729, -2.96544999, -2.79148037, -2.62837206, -2.47196831, -2.31801245, -2.16641125, -2.01707147, -1.8698999, -1.72480331, -1.58168847, -1.44046215, -1.30103114, -1.1633022, -1.02718211, -0.892577639, -0.759395567, -0.627542666, -0.496925711, -0.367451475, -0.212979319, -0.0600280874, 0.0909307884, 0.239425878, 0.384985749, 0.527138972, 0.665414115, 0.799339747, 0.928444438, 1.05225676, 1.17030527, 1.28211855, 1.38722516, 1.48515368, 1.57543267, 1.6575907, 1.71867125, 1.74601757, 1.76490779, 1.77567982, 1.77867159, 1.77422104, 1.76266608, 1.74434464, 1.71959465, 1.68875404, 1.65216072, 1.61015264, 1.5630677, 1.51124385, 1.45501901, 1.39473109, 1.33071804, 1.05934237, 0.68722006, 0.323418713, -0.0251043529, -0.351391818, -0.648486363, -0.909430666, -1.12726741, -1.29503927, -1.40578893, -1.45255906, -1.42839236, -1.32633149, -1.13941914, -0.860697991, -0.483210717, 0, -0.114265702, 0.0523839173, 0.492461409, 1.19847932, 2.16295021, 3.37838662, 4.8373011, 6.53220621, 8.45561448, 10.6000385, 12.9579908, 15.5219839, 18.2845303, 21.2381427, 24.3753336, 27.6886155, 31.0981026, 34.5859647, 38.3084728, 42.2426254, 46.3654212, 50.6538588, 55.0849368, 59.6356537, 64.2830082, 69.0039988, 73.7756243, 78.5748831, 83.3787739, 88.7737167, 93.9038342, 98.5632985, 102.546282, 106.337847, 109.38672, 110.957392, 110.636767, 108.01175, 102.669247, 94.1961632, 88.0117135, 76.0954418, 56.8931488, 28.8506353, -9.58629769, -55.2197805, -115.544098, -193.301468, -291.234105, -412.084227, -551.831871, -728.246317, -943.307256, -1203.6794, -1525.59269, -1920.98247, -2401.78409, -2963.89841, -3638.93947, -4476.0135, -5499.69203, -6749.80827, -8230.03126, -9978.18439, -12089.9638, -14634.501, -17645.4685, -21182.8046, -25436.465, -30522.8603, -36405.6216, -43377.4434, -51607.6217, -61265.4522, -72520.2308, -90981.3347, -110017.958, -128439.376, -145054.864, -158673.698, -164216.341, -166973.422, -167420.851, -167122.627, -166123.957, -164016.679, -160868.668, -156747.8, -151721.952, -145858.999, -157006.886, -153327.645, -140124.891, -122702.244, -106363.32, -96411.7365, -98151.1114, -97750.8457, -95245.0383},
        {-8.19190423, -8.03021644, -7.86528313, -7.69728382, -7.52639806, -7.35280538, -7.17668532, -6.99821741, -6.81758119, -6.6349562, -6.45052196, -6.26445803, -6.07694393, -5.8881592, -5.69828337, -5.50749599, -5.31597658, -5.12390469, -4.93145984, -4.73882158, -4.54616945, -4.35368297, -4.16154168, -3.96992512, -3.77901283, -3.58898434, -3.40001918, -3.2122969, -3.02599703, -2.84129911, -2.65838266, -2.47742723, -2.29861236, -2.12211757, -1.95570025, -1.79559161, -1.63793182, -1.48265448, -1.32969322, -1.17898166, -1.03045343, -0.884042131, -0.739681395, -0.597304842, -0.456846092, -0.318238765, -0.181416482, -0.0463128628, 0.0871384709, 0.219003899, 0.349349801, 0.501811884, 0.652376263, 0.800598381, 0.946033679, 1.0882376, 1.22676559, 1.36117308, 1.49101553, 1.61584837, 1.73522705, 1.84870701, 1.95584368, 2.05619252, 2.14930897, 2.23474847, 2.31206646, 2.37060393, 2.39994714, 2.42071805, 2.43313426, 2.43741338, 2.43377301, 2.42243076, 2.40360422, 2.37751101, 2.34436871, 2.30439495, 2.25780732, 2.20482342, 2.14566086, 2.08053724, 2.00967017, 1.93327725, 1.63478719, 1.2273719, 0.826808936, 0.44035111, 0.0752512419, -0.261237847, -0.561863338, -0.819372411, -1.02651225, -1.17603002, -1.26067292, -1.27318813, -1.20632281, -1.05282416, -0.805439359, -0.456915577, 0, 0.039282874, 0.329213403, 0.864838716, 1.64120594, 2.65336221, 3.89635465, 5.3652304, 7.05503657, 8.96082031, 11.0776287, 13.400509, 15.9245082, 18.6446734, 21.5560519, 24.6536907, 27.932637, 31.2470891, 34.5719729, 38.220922, 42.1588017, 46.3504773, 50.7608139, 55.3546769, 60.0969315, 64.9524429, 69.8860763, 74.8626971, 79.8471705, 84.8043617, 90.276366, 95.4391676, 100.084463, 104.003948, 107.622365, 110.414889, 111.687878, 111.043116, 108.082384, 102.407466, 93.6201428, 87.6179569, 75.7908301, 56.5087657, 28.141767, -10.9401628, -58.7828246, -121.49646, -201.607115, -301.640835, -424.123666, -564.37667, -740.936967, -955.850763, -1215.86041, -1537.29621, -1932.18369, -2412.54836, -2978.16297, -3656.0905, -4486.7606, -5483.87512, -6697.87684, -8140.47709, -9852.88665, -11926.3014, -14428.9733, -17400.0944, -20900.3827, -25097.1085, -30097.2698, -35854.909, -42653.5119, -50653.023, -60013.3869, -70894.548, -88603.5484, -106866.461, -124556.455, -140546.702, -153710.371, -159280.078, -162196.583, -162893.296, -162780.058, -161903.773, -159904.493, -156849.464, -152805.929, -147841.133, -142022.32, -141074.931, -134124.454, -123692.596, -112301.064, -102471.565, -96725.8068, -97585.4967, -97771.5473, -94481.1112},
        {-7.64652465, -7.48605769, -7.32222926, -7.15521667, -6.98519726, -6.81234835, -6.63684726, -6.45887132, -6.27859785, -6.09620419, -5.91186766, -5.72576558, -5.53807528, -5.34897408, -5.15863932, -4.96724831, -4.77497839, -4.58200687, -4.38851109, -4.19466837, -4.00065603, -3.80665141, -3.61283182, -3.4193746, -3.22645706, -3.03425654, -2.84295036, -2.65271585, -2.46373033, -2.27617113, -2.09021557, -1.90604098, -1.72382468, -1.54374401, -1.37219023, -1.206103, -1.04234728, -0.88089873, -0.721733033, -0.564825858, -0.410152875, -0.257689755, -0.10741217, 0.0407042093, 0.186683712, 0.330550666, 0.472329402, 0.612044247, 0.74971953, 0.88537958, 1.01904873, 1.17063, 1.3198381, 1.46626642, 1.60950835, 1.7491573, 1.88480666, 2.01604983, 2.14248021, 2.26369119, 2.37927618, 2.48882856, 2.59194174, 2.68820912, 2.77722408, 2.85858004, 2.93187038, 2.98909675, 3.02244305, 3.04716012, 3.06333349, 3.07104872, 3.07039134, 3.0614469, 3.04430092, 3.01903896, 2.98574655, 2.94450924, 2.89541255, 2.83854205, 2.77398326, 2.70182172, 2.62214298, 2.53503258, 2.20912011, 1.76508746, 1.32639842, 0.900649711, 0.495438076, 0.118360243, -0.222987058, -0.521007095, -0.768103139, -0.956678458, -1.07913632, -1.12788, -1.09531276, -0.973837867, -0.755858598, -0.43377822, 0, 0.227192514, 0.671520154, 1.33063013, 2.20216967, 3.28378597, 4.57312625, 6.06783772, 7.76556759, 9.66396309, 11.7606714, 14.0533398, 16.5396154, 19.2171455, 22.0835773, 25.136558, 28.3737348, 31.5812426, 34.7265125, 38.2856126, 42.2108659, 46.4545956, 50.9691249, 55.706777, 60.6198748, 65.6607416, 70.7817005, 75.9350747, 81.0731872, 86.1483613, 91.63831, 96.7773236, 101.360108, 105.18137, 108.665973, 111.293544, 112.375669, 111.518006, 108.326214, 102.405952, 93.3628773, 87.3080124, 75.3617602, 55.8569909, 27.1265749, -12.4966174, -61.6379855, -125.766822, -207.310688, -308.697148, -432.353761, -573.498402, -750.580165, -964.936459, -1223.14613, -1542.48927, -1935.44135, -2414.47783, -2984.21798, -3666.77893, -4492.84815, -5464.90229, -6643.77822, -8049.02736, -9724.98254, -11757.9672, -14215.0305, -17139.6825, -20592.8698, -24721.8478, -29625.0505, -35242.4538, -41855.3285, -49616.9596, -58680.6323, -69199.6318, -86239.6984, -103813.548, -120846.352, -136263.282, -148989.509, -154501.681, -157472.508, -158305.101, -158300.852, -157507.302, -155597.705, -152638.313, -148695.379, -143835.158, -138123.903, -125539.862, -115578.671, -108009.207, -102600.349, -99120.9748, -97339.9611, -97026.1862, -97618.4239, -93668.9369},
        {-7.13641705, -6.98024269, -6.82061303, -6.65769316, -6.49164821, -6.32264328, -6.15084348, -5.97641393, -5.79951973, -5.620326, -5.43899784, -5.25570037, -5.0705987, -4.88385793, -4.69564319, -4.50611957, -4.31545219, -4.12380617, -3.9313466, -3.73823861, -3.5446473, -3.35073778, -3.15667517, -2.96262457, -2.76875109, -2.57521986, -2.38219596, -2.18984453, -1.99833066, -1.80781947, -1.61847607, -1.43046557, -1.24395308, -1.05910372, -0.879356941, -0.703171817, -0.528968047, -0.356804483, -0.186739976, -0.0188333764, 0.146856464, 0.310270693, 0.471350461, 0.630036916, 0.786271205, 0.93999448, 1.09114789, 1.23967257, 1.38550969, 1.52860039, 1.66888582, 1.81933333, 1.96657548, 2.11027104, 2.25007877, 2.38565745, 2.51666583, 2.64276268, 2.76360677, 2.87885687, 2.98817174, 3.09121014, 3.18763085, 3.27709262, 3.35925423, 3.43377444, 3.50031202, 3.55378267, 3.58921597, 3.6160529, 3.6342597, 3.64380261, 3.64464786, 3.63676171, 3.62011037, 3.59466011, 3.56037715, 3.51722773, 3.4651781, 3.40419449, 3.33424314, 3.2552903, 3.16730219, 3.07024507, 2.71514251, 2.23166826, 1.75248815, 1.28564709, 0.839189997, 0.421161769, 0.039607316, -0.297428455, -0.581900637, -0.805764322, -0.960974604, -1.03948657, -1.03325533, -0.934235951, -0.734383544, -0.425653196, 0, 0.41323853, 1.02208643, 1.8253721, 2.82192392, 4.01057028, 5.39013957, 6.95946019, 8.71736052, 10.662669, 12.7942139, 15.1108237, 17.6113268, 20.2945515, 23.1593263, 26.2044796, 29.4288396, 32.5864347, 35.6411216, 39.1454436, 43.0457721, 47.288478, 51.8199327, 56.5865073, 61.534573, 66.610501, 71.7606625, 76.9314287, 82.0691707, 87.1202599, 92.5127908, 97.5347792, 101.988079, 105.674545, 109.115333, 111.752648, 112.836876, 111.954081, 108.690326, 102.631676, 93.3641916, 86.4860921, 73.7414996, 53.5401862, 24.291924, -15.5935149, -63.0909504, -125.559871, -205.744355, -306.388484, -430.236338, -574.354592, -754.103779, -967.934795, -1221.47909, -1535.02839, -1922.29227, -2396.9803, -2969.13922, -3656.97387, -4482.68492, -5442.63109, -6600.26059, -7977.04946, -9616.97431, -11602.1377, -13997.1416, -16846.4987, -20209.6027, -24219.5088, -28967.6053, -34370.8605, -40721.9161, -48172.5364, -56874.4854, -66979.5274, -83398.1507, -100333.714, -116748.302, -131603.996, -143862.88, -149150.962, -151990.451, -152773.175, -152751.568, -151976.804, -150141.471, -147314.673, -143565.512, -138963.092, -133576.516, -116173.701, -105411.47, -99722.9857, -97541.4088, -97299.9007, -97431.6224, -96369.7352, -96699.2802, -92504.6966},
        {-6.67630935, -6.52350181, -6.36730138, -6.20785225, -6.04529866, -5.87978481, -5.71145491, -5.5404532, -5.36692387, -5.19101116, -5.01285926, -4.83261241, -4.65041481, -4.46641068, -4.28074424, -4.0935597, -3.90500128, -3.71521319, -3.52433966, -3.33252489, -3.1399131, -2.94664851, -2.75287533, -2.55873779, -2.36438008, -2.16994644, -1.97558108, -1.78142821, -1.58763205, -1.39433681, -1.20168671, -1.00982597, -0.818898805, -0.629049425, -0.440074341, -0.252298937, -0.0660527754, 0.118508645, 0.301229825, 0.481955263, 0.660529461, 0.83679692, 1.01060214, 1.18178962, 1.35020386, 1.51568936, 1.67809062, 1.83725215, 1.99301844, 2.14523399, 2.2937433, 2.44377052, 2.58966388, 2.73115126, 2.86796055, 2.99981962, 3.12645636, 3.24759865, 3.36297436, 3.47231138, 3.57533758, 3.67178086, 3.76136908, 3.84383013, 3.91889188, 3.98628223, 4.04572905, 4.09495297, 4.13172283, 4.1599063, 4.17936138, 4.18994606, 4.19151836, 4.18393626, 4.16705779, 4.14074093, 4.10484369, 4.05922406, 4.00374006, 3.93824969, 3.86261094, 3.77668181, 3.68032032, 3.57338445, 3.18793719, 2.66336675, 2.14228005, 1.63322556, 1.14475174, 0.685407065, 0.263739994, -0.111701007, -0.432367472, -0.689710935, -0.875182929, -0.980234991, -0.996318653, -0.914885451, -0.727386918, -0.42527459, 0, 0.56223716, 1.31187262, 2.24767562, 3.36841539, 4.67286118, 6.15978222, 7.82794775, 9.67612702, 11.7030893, 13.9076037, 16.2884396, 18.8443662, 21.5741527, 24.4765683, 27.5503824, 30.7943641, 33.9549158, 36.9947544, 40.4834789, 44.3657799, 48.586348, 53.0898739, 57.8210481, 62.7245614, 67.7451043, 72.8273675, 77.9160416, 82.9558172, 87.891385, 93.1106828, 97.9526301, 102.228943, 105.751339, 109.193566, 111.936325, 113.13044, 112.329019, 109.085166, 102.95199, 93.4825967, 85.384535, 71.4877258, 50.336798, 20.4763801, -19.5488991, -64.0253266, -123.369629, -200.829449, -299.652427, -423.086208, -570.89784, -754.069616, -967.36989, -1215.16673, -1521.42765, -1901.32039, -2370.0127, -2942.54328, -3634.23829, -4460.56471, -5415.41775, -6559.99174, -7914.12841, -9519.42138, -11454.3217, -13780.3623, -16541.2203, -19793.7994, -23660.4741, -28226.5849, -33383.9816, -39441.2632, -46550.7476, -54864.7528, -64535.5968, -80361.5886, -96687.5574, -112504.323, -126802.707, -138573.528, -143537.61, -146135.769, -146760.251, -146655.516, -145876.477, -144122.961, -141467.633, -137983.162, -133742.217, -128817.463, -111104.057, -100959.947, -96438.79, -95594.241, -96479.9563, -97149.5921, -95656.8042, -95332.3911, -91122.4917},
        {-6.28084603, -6.12632, -5.96876361, -5.80829261, -5.64502277, -5.47906983, -5.31054958, -5.13957774, -4.9662701, -4.7907424, -4.61311039, -4.43348985, -4.25199653, -4.06874618, -3.88385456, -3.69743743, -3.50961055, -3.32048968, -3.13019057, -2.93882898, -2.74652067, -2.5533814, -2.35952692, -2.16507299, -1.97013537, -1.77482982, -1.5792721, -1.38357796, -1.18786316, -0.992243461, -0.796834617, -0.601752387, -0.40711253, -0.213030802, -0.0158726868, 0.182292039, 0.379348523, 0.575059302, 0.769186916, 0.961493903, 1.1517428, 1.33969615, 1.52511649, 1.70776636, 1.88740829, 2.06380483, 2.23671851, 2.40591187, 2.57114745, 2.73218779, 2.88879543, 3.04009109, 3.18649303, 3.32777771, 3.46372158, 3.5941011, 3.71869272, 3.83727288, 3.94961804, 4.05550466, 4.15470918, 4.24700806, 4.33217775, 4.4099947, 4.48023536, 4.54267618, 4.59709363, 4.64352151, 4.68198747, 4.71173716, 4.73253034, 4.74412679, 4.74628627, 4.73876855, 4.72133339, 4.69374057, 4.65574986, 4.60712101, 4.5476138, 4.476988, 4.39500337, 4.30141969, 4.19599672, 4.07849422, 3.66215479, 3.09602328, 2.53258828, 1.98090758, 1.45003899, 0.949040307, 0.486969335, 0.0728838755, -0.284158271, -0.575099303, -0.79088142, -0.922446819, -0.960737699, -0.896696258, -0.721264696, -0.42538521, 0, 0.636380987, 1.46708196, 2.48971737, 3.70190164, 5.10124921, 6.68537453, 8.45189202, 10.3984161, 12.5225613, 14.821942, 17.2941726, 19.9368675, 22.7476413, 25.7241083, 28.863883, 32.1645799, 35.3784347, 38.4680078, 41.9859368, 45.8772547, 50.086995, 54.5601911, 59.2418761, 64.0770834, 69.0108464, 73.9881983, 78.9541726, 83.8538023, 88.632121, 93.6413894, 98.2700965, 102.341108, 105.677288, 109.121508, 111.986629, 113.313514, 112.619025, 109.420025, 103.233378, 93.5759476, 84.2361105, 69.1591903, 47.0263192, 16.5186296, -23.6827461, -65.3353708, -121.709743, -196.515057, -293.460507, -416.255287, -567.111704, -753.065365, -965.805642, -1208.59323, -1508.31364, -1881.24531, -2343.6667, -2914.10695, -3606.08228, -4430.65358, -5381.52586, -6515.63248, -7849.9557, -9423.11377, -11310.3859, -13570.2233, -16245.0776, -19389.2519, -23115.7148, -27504.2501, -32426.6289, -38202.6503, -44984.1027, -52922.7748, -62170.4552, -77412.2678, -93133.3319, -108352.102, -122087.033, -133356.579, -137962.378, -140284.247, -140719.277, -140527.102, -139763.389, -138124.414, -135683.813, -132515.224, -128692.283, -124288.628, -108522.963, -99640.7886, -95826.2784, -95263.6057, -96136.9436, -96630.4654, -94928.3444, -93834.5552, -89656.4543},
        {-5.97837511, -5.80433372, -5.62851554, -5.45099445, -5.27184433, -5.09113906, -4.90895249, -4.72535853, -4.54043103, -4.35424387, -4.16687093, -3.97838609, -3.78886322, -3.59837619, -3.40699888, -3.21480517, -3.02186894, -2.82826405, -2.63406438, -2.43934381, -2.24417622, -2.04863548, -1.85279546, -1.65673004, -1.4605131, -1.26421851, -1.06792015, -0.871691888, -0.675607606, -0.479741178, -0.284166479, -0.0889573847, 0.105812229, 0.300068488, 0.498880911, 0.699495741, 0.8990974, 1.0974451, 1.29429805, 1.48941547, 1.68255656, 1.87348054, 2.06194661, 2.247714, 2.4305419, 2.61018954, 2.78641612, 2.95898085, 3.12764295, 3.29216163, 3.4522961, 3.60855751, 3.75993683, 3.90617697, 4.04702084, 4.18221136, 4.31149144, 4.43460397, 4.55129189, 4.66129809, 4.76436549, 4.86023699, 4.94865552, 5.02936397, 5.10210527, 5.16662231, 5.22265802, 5.27077451, 5.31151421, 5.3429305, 5.36471318, 5.37655207, 5.37813695, 5.36915764, 5.34930394, 5.31826567, 5.27573261, 5.22139459, 5.1549414, 5.07606285, 4.98444874, 4.87978888, 4.76177308, 4.63009114, 4.18325712, 3.57788354, 2.97420278, 2.38167839, 1.8097739, 1.26795285, 0.765678783, 0.312415227, -0.0823742769, -0.409226192, -0.658676982, -0.821263109, -0.887521036, -0.847987225, -0.69319814, -0.413690244, 0, 0.488337255, 1.21145595, 2.16391569, 3.34027609, 4.73509676, 6.3429373, 8.15835732, 10.1759164, 12.3901743, 14.7956904, 17.3870244, 20.158736, 23.1053847, 26.2215301, 29.5017319, 32.9405497, 36.3034815, 39.5523638, 43.1998717, 47.1914801, 51.4726642, 55.9888987, 60.6856589, 65.5084195, 70.4026556, 75.3138423, 80.1874544, 84.9689671, 89.6038551, 94.4102857, 98.824388, 102.679829, 105.810277, 109.187391, 112.075842, 113.436837, 112.762646, 109.545538, 103.277784, 93.4516534, 83.5255949, 67.8186012, 45.0469506, 13.9269211, -26.825209, -68.8069653, -125.409815, -200.339029, -297.299877, -419.997632, -570.760715, -756.35406, -968.927226, -1211.71742, -1510.59916, -1881.28326, -2339.48054, -2902.39417, -3583.90346, -4396.25972, -5334.91708, -6453.87701, -7769.69985, -9319.4686, -11175.0407, -13392.3116, -16013.2987, -19090.2994, -22726.497, -26996.2908, -31780.1846, -37376.3032, -43926.1637, -51571.283, -60453.1781, -75033.3391, -90053.3166, -104574.634, -117658.816, -128367.385, -132645.696, -134748.888, -135073.987, -134919.644, -134325.989, -132957.479, -130873.027, -128131.549, -124791.96, -120913.177, -107158.947, -99267.9381, -95712.0645, -94963.2399, -95493.378, -95774.3925, -94278.1973, -92718.029, -88365.7415},
        {-5.73832998, -5.54046978, -5.34226362, -5.1437418, -4.94493462, -4.74587238, -4.54658538, -4.34710392, -4.14745831, -3.94767884, -3.74779581, -3.54783952, -3.34784028, -3.14782839, -2.94783414, -2.74788783, -2.54801977, -2.34826026, -2.1486396, -1.94918809, -1.74993602, -1.5509137, -1.35215144, -1.15367952, -0.955528257, -0.757727943, -0.560308882, -0.363301373, -0.166735717, 0.0293577848, 0.224948832, 0.420007125, 0.614502363, 0.808404244, 1.00753802, 1.20882244, 1.4091059, 1.60816807, 1.80578863, 2.00174726, 2.19582362, 2.3877974, 2.57744826, 2.7645559, 2.94889997, 3.13026016, 3.30841615, 3.4831476, 3.6542342, 3.82145562, 3.98459154, 4.1478282, 4.30644319, 4.46012064, 4.60854472, 4.75139957, 4.88836933, 5.01913817, 5.14339023, 5.26080966, 5.37108061, 5.47388723, 5.56891367, 5.65584407, 5.7343626, 5.8041534, 5.86490061, 5.91726868, 5.96189851, 5.99645248, 6.02055119, 6.03381524, 6.03586523, 6.02632177, 6.00480545, 5.97093687, 5.92433665, 5.86462537, 5.79142365, 5.70435208, 5.60303126, 5.4870818, 5.3561243, 5.20977936, 4.73215013, 4.0872524, 3.4428943, 2.80893555, 2.19523591, 1.61165514, 1.06805298, 0.574289177, 0.140223489, -0.224284337, -0.509374552, -0.705187406, -0.801863148, -0.789542028, -0.658364297, -0.398470204, 0, 0.279306314, 0.843649634, 1.68395295, 2.79113924, 4.1561315, 5.76985273, 7.6232259, 9.707174, 12.01262, 14.530487, 17.2516978, 20.1671755, 23.2678432, 26.5446236, 29.98844, 33.5902151, 37.1325448, 40.576591, 44.3844931, 48.5025327, 52.876991, 57.4541497, 62.18029, 67.0016934, 71.8646414, 76.7154152, 81.5002965, 86.1655665, 90.6575067, 95.2663605, 99.4683279, 103.106502, 106.023975, 109.297384, 112.15305, 113.488349, 112.776311, 109.489965, 103.102342, 93.0864713, 82.751241, 66.5663874, 43.2544461, 11.5379529, -29.8605565, -73.1859063, -131.006361, -206.857544, -304.275081, -426.794595, -576.017079, -759.777339, -971.900775, -1215.99055, -1515.61274, -1885.37049, -2339.86695, -2893.39793, -3560.94374, -4357.81992, -5282.1853, -6384.93445, -7681.87065, -9208.7736, -11035.2719, -13215.457, -15792.4386, -18817.4609, -22382.0345, -26554.8932, -31227.2207, -36672.616, -43021.2724, -50403.3831, -58949.1415, -72883.496, -87218.1959, -101059.746, -113514.65, -123689.413, -127688.163, -129621.366, -129882.258, -129791.526, -129375.292, -128273.241, -126530.817, -124193.464, -121306.625, -117915.745, -106773.625, -99993.5947, -96492.1759, -95185.8902, -94991.2592, -94824.8047, -93603.0483, -91640.597, -87011.0621},
        {-5.55046846, -5.32893921, -5.10840838, -4.88886666, -4.67030474, -4.45271332, -4.23608311, -4.02040478, -3.80566904, -3.59186659, -3.37898811, -3.16702431, -2.95596587, -2.7458035, -2.53652789, -2.32812974, -2.12059973, -1.91392857, -1.70810695, -1.50312557, -1.29897512, -1.0956463, -0.893129794, -0.691416311, -0.49049654, -0.290361179, -0.0910009208, 0.107593538, 0.305431502, 0.502522276, 0.698875165, 0.894499473, 1.08940451, 1.28359957, 1.48323337, 1.68511638, 1.88598527, 2.08565012, 2.28392098, 2.48060794, 2.67552106, 2.8684704, 3.05926605, 3.24771806, 3.4336365, 3.61683145, 3.79711297, 3.97429114, 4.14817602, 4.31857768, 4.48530619, 4.65711497, 4.82467686, 4.98760806, 5.14552477, 5.29804317, 5.44477947, 5.58534985, 5.71937052, 5.84645766, 5.96622747, 6.07829615, 6.18227989, 6.27779488, 6.36445732, 6.44188341, 6.50968934, 6.56865992, 6.61955191, 6.65957131, 6.68825856, 6.7051541, 6.70979837, 6.70173181, 6.68049485, 6.64562793, 6.5966715, 6.53316598, 6.45465182, 6.36066945, 6.25075932, 6.12446186, 5.9813175, 5.8208667, 5.31201026, 4.62656299, 3.94031374, 3.26353283, 2.6064906, 1.97945736, 1.39270345, 0.856499206, 0.381114947, -0.023178993, -0.346112286, -0.577414601, -0.706815611, -0.724044985, -0.618832394, -0.380907509, 0, 0.0640505102, 0.463983915, 1.18709077, 2.22066162, 3.55198702, 5.16835753, 7.05706369, 9.20539607, 11.6006452, 14.2301017, 17.081056, 20.1407987, 23.3966204, 26.8358116, 30.445663, 34.2134649, 37.9383122, 41.5806102, 45.5533306, 49.8034363, 54.2778902, 58.9236551, 63.6876938, 68.5169694, 73.3584445, 78.1590822, 82.8658452, 87.4256964, 91.7855988, 96.216047, 100.221844, 103.652892, 106.359096, 109.489462, 112.244341, 113.477657, 112.649583, 109.22029, 102.649951, 92.3987368, 81.7516334, 65.1799262, 41.3954544, 9.11005698, -32.9644268, -78.2667614, -137.896853, -215.105212, -313.142349, -435.258772, -581.742093, -762.427043, -973.521292, -1219.2486, -1520.19894, -1889.61417, -2340.73618, -2884.34759, -3536.70843, -4316.76153, -5225.56655, -6310.90174, -7586.92558, -9088.25631, -10884.18, -13028.0695, -15566.5423, -18551.2198, -22058.4942, -26150.7458, -30722.3894, -36029.612, -42192.9361, -49332.8842, -57569.9785, -70905.8539, -84609.6424, -97831.5885, -109721.936, -119430.931, -123231.721, -125069.71, -125328.181, -125306.375, -125025.011, -124123.16, -122640.625, -120617.206, -118092.706, -115106.926, -106772.217, -101163.304, -97676.2417, -95707.0847, -94651.8883, -93906.7077, -92867.5979, -90554.8203, -85525.0798},
        {-5.38083911, -5.14857648, -4.91783422, -4.68858242, -4.46079121, -4.23443067, -4.00947091, -3.78588205, -3.56363417, -3.3426974, -3.12304182, -2.90463756, -2.6874547, -2.47146336, -2.25663363, -2.04293563, -1.83033946, -1.61881522, -1.40833302, -1.19886296, -0.990375143, -0.782839678, -0.576226668, -0.370506217, -0.165648429, 0.0383765902, 0.241598737, 0.444047908, 0.645753997, 0.846746901, 1.04705651, 1.24671273, 1.44574545, 1.64418457, 1.84886055, 2.05625997, 2.26278765, 2.4682528, 2.67246462, 2.87523232, 3.07636509, 3.27567215, 3.4729627, 3.66804594, 3.86073107, 4.0508273, 4.23814383, 4.42248987, 4.60367462, 4.78150729, 4.95579707, 5.13646596, 5.31299115, 5.48496262, 5.65197035, 5.81360432, 5.9694545, 6.11911087, 6.26216341, 6.3982021, 6.52681693, 6.64759785, 6.76013487, 6.86401795, 6.95883707, 7.04418221, 7.11964335, 7.18773193, 7.25088923, 7.30267996, 7.3425047, 7.36976405, 7.38385859, 7.38418892, 7.37015563, 7.34115931, 7.29660056, 7.23587997, 7.15839812, 7.06355562, 6.95075306, 6.81939102, 6.66887009, 6.49859089, 5.95721593, 5.22689214, 4.49398205, 3.76926837, 3.06353381, 2.38756111, 1.75213296, 1.16803209, 0.646041217, 0.196943054, -0.168479683, -0.439444275, -0.605168007, -0.654868162, -0.577762024, -0.363066875, 0, -0.0106222181, 0.341660262, 1.04207323, 2.07584249, 3.42819382, 5.08435303, 7.02954589, 9.24899822, 11.7279358, 14.4515844, 17.4051699, 20.5739179, 23.9430544, 27.4978051, 31.2233959, 35.1050524, 38.9461584, 42.7059621, 46.7809991, 51.1175289, 55.6618111, 60.3601054, 65.1586713, 70.0037685, 74.8416567, 79.6185953, 84.2808442, 88.7746628, 93.0463108, 97.3691869, 101.242461, 104.51425, 107.032673, 109.938481, 112.433475, 113.374754, 112.222646, 108.437477, 101.479576, 90.8092706, 79.7155026, 62.6647549, 38.3682782, 5.53732361, -37.116858, -84.5044024, -145.887325, -224.149029, -322.172916, -442.842389, -584.71866, -760.289999, -967.851945, -1212.33944, -1512.24285, -1879.96644, -2327.91446, -2865.82793, -3509.37963, -4277.14772, -5170.25814, -6234.41414, -7478.39434, -8935.78565, -10682.2113, -12773.9993, -15270.752, -18229.6463, -21695.8523, -25721.3319, -30149.5815, -35273.1, -41216.0104, -48102.4355, -56056.4983, -68972.4065, -82261.6772, -95105.912, -106686.713, -116185.681, -120050.329, -122019.075, -122445.274, -122434.41, -122042.12, -121010.652, -119411.314, -117315.411, -114794.248, -111919.132, -105526.938, -101062.099, -98074.4399, -96113.7843, -94729.9573, -93472.7835, -91892.0875, -89299.3782, -83693.1561},
        {-5.2529074, -5.01545359, -4.77962741, -4.54538974, -4.31270141, -4.08152328, -3.85181621, -3.62354104, -3.39665862, -3.17112982, -2.94691547, -2.72397643, -2.50227355, -2.28176769, -2.06241969, -1.84419041, -1.6270407, -1.41093141, -1.19582339, -0.981677496, -0.768454575, -0.556115482, -0.344621068, -0.133932183, 0.0759903191, 0.285185588, 0.493692772, 0.701551019, 0.908799477, 1.1154773, 1.32162362, 1.52727761, 1.73247839, 1.93726514, 2.14898262, 2.36386352, 2.57805268, 2.79135216, 3.00356404, 3.21449038, 3.42393324, 3.6316947, 3.83757681, 4.04138165, 4.24291128, 4.44196776, 4.63835317, 4.83186957, 5.02231902, 5.20950359, 5.39322535, 5.5839039, 5.77049361, 5.95256636, 6.12969407, 6.30144862, 6.46740192, 6.62712587, 6.78019236, 6.92617329, 7.06464056, 7.19516607, 7.31732171, 7.4306794, 7.53481102, 7.62928847, 7.71368366, 7.79265819, 7.87075149, 7.93703829, 7.99076054, 8.0311602, 8.05747922, 8.06895955, 8.06484315, 8.04437196, 8.00678794, 7.95133305, 7.87724923, 7.78377844, 7.67016263, 7.53564376, 7.37946377, 7.20086463, 6.62658352, 5.84994992, 5.06877052, 4.29437575, 3.53809607, 2.81126192, 2.12520377, 1.49125204, 0.920737196, 0.424989683, 0.0153399478, -0.29688156, -0.500344392, -0.583718101, -0.535672238, -0.344876353, 0, -0.0208115878, 0.336900525, 1.05706128, 2.12359561, 3.52042847, 5.2314848, 7.24068953, 9.53196761, 12.089244, 14.8964436, 17.9374913, 21.1963122, 24.6568311, 28.3029731, 32.1186629, 36.0878257, 40.0116939, 43.8480938, 47.9941827, 52.3947373, 56.9945341, 61.7383498, 66.5709609, 71.437144, 76.2816758, 81.0493328, 85.6848916, 90.1331288, 94.3388211, 98.5851969, 102.354482, 105.489917, 107.834744, 110.487805, 112.66454, 113.245563, 111.697432, 107.48671, 100.079956, 88.9437293, 77.4053365, 59.8955834, 35.1256672, 1.80678513, -41.3498656, -90.7546363, -153.794394, -232.977555, -330.812534, -449.807747, -586.76295, -756.87247, -960.415025, -1203.0342, -1501.32683, -1866.97196, -2311.64864, -2844.51492, -3480.34381, -4236.65851, -5114.28598, -6156.80325, -7366.58376, -8776.11146, -10468.8909, -12505.0073, -14958.6386, -17893.5236, -21321.7283, -25284.5479, -29564.0509, -34500.6399, -40225.3421, -46869.185, -54563.1958, -67127.3838, -80081.5489, -92634.4728, -103994.937, -113371.724, -117387.377, -119561.075, -120209.987, -120192.512, -119611.583, -118344.622, -116505.837, -114209.436, -111569.63, -108700.626, -103911.643, -100543.268, -98190.8155, -96449.6049, -94914.9532, -93182.1779, -90846.5963, -87988.7383, -81657.3761},
        {-5.17322923, -4.93467588, -4.69751369, -4.46170463, -4.22721065, -3.99399369, -3.76201572, -3.53123868, -3.30162453, -3.07313521, -2.84573269, -2.61937891, -2.39403582, -2.16966538, -1.94622955, -1.72369026, -1.50200948, -1.28114915, -1.06107123, -0.841737678, -0.623110435, -0.405151458, -0.187822699, 0.0289138898, 0.245096356, 0.460762748, 0.675951113, 0.890699498, 1.10504595, 1.31902852, 1.53268526, 1.7460542, 1.95917341, 2.17208092, 2.39233035, 2.61604388, 2.83925337, 3.06175298, 3.28333685, 3.50379914, 3.722934, 3.94053559, 4.15639806, 4.37031556, 4.58208225, 4.79149228, 4.9983398, 5.20241896, 5.40352393, 5.60144885, 5.79598787, 5.998048, 6.19606963, 6.38960603, 6.57821042, 6.76143608, 6.93883623, 7.10996413, 7.27437304, 7.43161619, 7.58124684, 7.72281823, 7.85588362, 7.97999625, 8.09470938, 8.19957624, 8.29415009, 8.38529963, 8.47971792, 8.56187064, 8.6308368, 8.68569542, 8.72552552, 8.74940612, 8.75641623, 8.74563487, 8.71614106, 8.66701382, 8.59733216, 8.50617511, 8.39262167, 8.25575087, 8.09464172, 7.90837325, 7.30118284, 6.4778753, 5.64800109, 4.82345261, 4.01612225, 3.23790239, 2.50068543, 1.81636374, 1.19682972, 0.653975747, 0.199694218, -0.154122483, -0.395581967, -0.512791849, -0.49385974, -0.326893253, 0, 0.0230301251, 0.430373134, 1.20531043, 2.3311234, 3.79109345, 5.56850198, 7.64663039, 10.0087601, 12.6381724, 15.5181488, 18.6319707, 21.9629195, 25.4942765, 29.2093232, 33.091341, 37.1236112, 41.0994269, 44.974928, 49.163132, 53.6066084, 58.2479268, 63.0296567, 67.8943678, 72.7846296, 77.6430117, 82.4120837, 87.0344151, 91.4525756, 95.6091347, 99.7972454, 103.480347, 106.495833, 108.681099, 111.078423, 112.922758, 113.127754, 111.169341, 106.523448, 98.666004, 87.0729393, 75.1201435, 57.1966411, 32.015326, -1.71090822, -45.2691678, -96.3442029, -160.726065, -240.602449, -338.161051, -455.589565, -588.20881, -753.466208, -953.358051, -1194.12003, -1490.64326, -1853.93277, -2294.99357, -2822.34561, -3450.04374, -4194.85377, -5057.78799, -6079.78458, -7256.85974, -8620.1769, -10260.4545, -12241.0742, -14648.8951, -17553.3857, -20937.9305, -24834.6081, -28970.584, -33729.7593, -39249.7591, -45668.2087, -53122.7333, -65364.2913, -78013.8414, -90305.6755, -101474.085, -110753.363, -114937.015, -117327.309, -118204.588, -118172.832, -117384.277, -115856.313, -113747.918, -111218.073, -108425.757, -105529.949, -102063.443, -99703.7317, -98030.561, -96623.6753, -95062.8191, -92927.7373, -89798.1743, -86656.1472, -79454.7682},
        {-5.15647109, -4.92124252, -4.68671236, -4.45285984, -4.21966419, -3.98710464, -3.75516043, -3.52381077, -3.29303491, -3.06281207, -2.83312149, -2.60394238, -2.375254, -2.14703555, -1.91926629, -1.69192542, -1.4649922, -1.23844584, -1.01226558, -0.786430642, -0.560920266, -0.335713678, -0.110790109, 0.11387121, 0.33829105, 0.562490179, 0.786489368, 1.01030938, 1.233971, 1.45749498, 1.6809021, 1.90421313, 2.12744883, 2.35062998, 2.58077677, 2.81426303, 3.04737847, 3.27991671, 3.51167139, 3.74243612, 3.97200452, 4.20017024, 4.42672688, 4.65146807, 4.87418744, 5.09467862, 5.31273523, 5.52815088, 5.74071922, 5.95023386, 6.15648842, 6.3719097, 6.5833843, 6.79043197, 6.99257249, 7.1893256, 7.38021108, 7.56474869, 7.74245819, 7.91285935, 8.07547192, 8.22981567, 8.37541036, 8.51177575, 8.63843162, 8.75489771, 8.86069379, 8.96427009, 9.07386242, 9.17057213, 9.25334005, 9.32110703, 9.37281389, 9.40740149, 9.42381065, 9.4209822, 9.397857, 9.35337588, 9.28647966, 9.1961092, 9.08120532, 8.94070887, 8.77356067, 8.57870158, 7.93922338, 7.07073827, 6.1938566, 5.32101431, 4.46464733, 3.63719162, 2.8510831, 2.11875772, 1.4526514, 0.865200099, 0.368839741, -0.0239937327, -0.300864383, -0.449336273, -0.456973465, -0.311340019, 0, 0.141884936, 0.659731403, 1.53713261, 2.75768175, 4.30497203, 6.16259667, 8.31414885, 10.7432218, 13.4334087, 16.3683028, 19.5314972, 22.9065852, 26.47716, 30.2268148, 34.1391428, 38.1977371, 42.1755597, 46.0261135, 50.2094198, 54.6639367, 59.3281218, 64.1404333, 69.0393289, 73.9632666, 78.8507044, 83.6401, 88.2699114, 92.6785966, 96.8046134, 100.928243, 104.520785, 107.41815, 109.456249, 111.650799, 113.247807, 113.178668, 110.92535, 105.969823, 97.7940556, 85.8800181, 73.5913146, 55.3469119, 29.8693293, -4.11891447, -47.8953003, -99.4552809, -164.178873, -244.164064, -341.508842, -458.311196, -589.587053, -753.172313, -952.022414, -1192.52739, -1488.08959, -1848.96297, -2285.40149, -2803.96346, -3419.4384, -4150.64751, -5001.67646, -6008.66779, -7163.49789, -8495.20631, -10095.8036, -12028.3956, -14381.9254, -17226.1976, -20536.2794, -24340.8266, -28358.0189, -32974.7541, -38327.1609, -44551.3677, -51783.503, -63652.5163, -75928.5101, -87874.4076, -98753.1323, -107827.607, -112047.767, -114531.517, -115533.061, -115482.631, -114564.755, -112890.614, -110648.403, -108026.316, -105212.547, -102395.291, -99890.3091, -98338.1921, -97283.3866, -96270.3393, -94843.4968, -92547.3056, -88926.2123, -85414.4393, -77141.7815},
        {-5.20643077, -4.97806998, -4.74930433, -4.52014573, -4.29060611, -4.06069741, -3.83043154, -3.59982044, -3.36887604, -3.13761026, -2.90603503, -2.67416228, -2.44200394, -2.20957193, -1.97687818, -1.74393462, -1.51075318, -1.27734578, -1.04372436, -0.809900835, -0.57588714, -0.341695201, -0.107336947, 0.127175696, 0.3618308, 0.596616436, 0.831520677, 1.0665316, 1.30163726, 1.53682575, 1.77208514, 2.00740349, 2.24276887, 2.47816937, 2.71928441, 2.96313667, 3.20668043, 3.44971907, 3.69205597, 3.9334945, 4.17383803, 4.41288995, 4.65045364, 4.88633246, 5.1203298, 5.35224903, 5.58189353, 5.80906668, 6.03357185, 6.25521241, 6.47379176, 6.70466507, 6.93174678, 7.15450312, 7.37240036, 7.58490472, 7.79148246, 7.99159981, 8.18472302, 8.37031834, 8.54785201, 8.71679027, 8.87659936, 9.02674554, 9.16669504, 9.29591411, 9.413869, 9.52972581, 9.65241777, 9.76140587, 9.85552755, 9.93362026, 9.99452143, 10.0370685, 10.0600989, 10.0624501, 10.0429595, 10.0004646, 9.93380284, 9.8418116, 9.72332834, 9.57719051, 9.40223556, 9.19730092, 8.52645699, 7.61506947, 6.69372964, 5.77538395, 4.87297888, 3.99946088, 3.16777644, 2.39087201, 1.68169406, 1.05318907, 0.518303495, 0.0899838054, -0.218823531, -0.395172048, -0.426115276, -0.298706749, 0, 0.33057479, 1.01538015, 2.03921972, 3.38689718, 5.04321618, 6.99298038, 9.22099342, 11.712059, 14.4509807, 17.4225623, 20.6116073, 24.0029195, 27.5813025, 31.3315599, 35.2384955, 39.2869128, 43.2177059, 46.9808342, 51.1128017, 55.5461459, 60.2134038, 65.047113, 69.9798106, 74.9440341, 79.8723206, 84.6972074, 89.351232, 93.7669316, 97.8768434, 101.921219, 105.412052, 108.189568, 110.093988, 112.157847, 113.625213, 113.422203, 111.031348, 105.935177, 97.6162217, 85.5570107, 73.0241201, 54.5643679, 28.9199272, -5.16702877, -48.9543269, -99.5707533, -163.435436, -242.838461, -340.069917, -457.41989, -591.034134, -756.85681, -957.86916, -1200.07915, -1495.67432, -1854.06771, -2284.67235, -2790.49512, -3388.81116, -4103.88414, -4946.27944, -5944.90477, -7090.49344, -8409.05682, -9986.38287, -11880.8211, -14170.3036, -16918.2568, -20116.3045, -23796.6862, -27726.3779, -32243.6921, -37472.6902, -43537.4335, -50561.9833, -61981.4745, -73781.1265, -85257.2322, -95706.0848, -104423.977, -108498.483, -110907.094, -111892.697, -111825.937, -110899.13, -109251.359, -107077.6, -104572.832, -101932.03, -99350.1726, -97459.4845, -96477.6275, -95921.0219, -95306.0882, -94149.2467, -91966.9178, -88275.5219, -84280.1048, -74740.9828},
        {-5.30597629, -5.08480714, -4.86206994, -4.6378113, -4.41207784, -4.18491617, -3.95637288, -3.7264946, -3.49532793, -3.26291948, -3.02931587, -2.79456369, -2.55870956, -2.3218001, -2.0838819, -1.84500158, -1.60520575, -1.36454102, -1.123054, -0.880791291, -0.637799511, -0.394125267, -0.149815169, 0.0950841739, 0.340526152, 0.586464155, 0.832851575, 1.0796418, 1.32678822, 1.57424423, 1.82196322, 2.06989858, 2.31800369, 2.56623196, 2.81876519, 3.07335363, 3.32769669, 3.58161054, 3.83491135, 4.08741529, 4.33893852, 4.58929723, 4.83830757, 5.08578571, 5.33154784, 5.57541011, 5.8171887, 6.05669978, 6.29375951, 6.52818407, 6.75978962, 7.00736128, 7.25133507, 7.49111595, 7.72610887, 7.95571879, 8.17935069, 8.39640951, 8.60630021, 8.80842777, 9.00219713, 9.18701326, 9.36228112, 9.52740566, 9.68179186, 9.82484466, 9.95596903, 10.0847932, 10.2206998, 10.3420092, 10.4474637, 10.5358055, 10.6057767, 10.6561197, 10.6855766, 10.6928897, 10.6768013, 10.6360534, 10.5693885, 10.4755486, 10.3532761, 10.2013132, 10.018402, 9.80328492, 9.10167371, 8.14835468, 7.18355921, 6.22073124, 5.27331469, 4.3547535, 3.47849162, 2.65797296, 1.90664148, 1.2379411, 0.665315766, 0.202209406, -0.137934042, -0.341670642, -0.395556459, -0.286147557, 0, 0.53354527, 1.3967379, 2.57575082, 4.05675697, 5.82592929, 7.8694407, 10.1734642, 12.7241726, 15.5077389, 18.5103361, 21.718137, 25.1173146, 28.6940419, 32.4344918, 36.3248372, 40.3512511, 44.2254526, 47.894407, 51.9650754, 56.3648194, 61.0210002, 65.8609794, 70.8121182, 75.8017783, 80.7573209, 85.6061076, 90.2754996, 94.6928585, 98.7855457, 102.757942, 106.158816, 108.832423, 110.623017, 112.576466, 113.936219, 113.625296, 111.127471, 105.926516, 97.5062029, 85.350304, 72.6106811, 53.9866422, 28.2442899, -5.85027312, -49.5309442, -98.6782647, -161.217827, -239.721867, -336.762621, -454.912326, -591.927089, -761.081655, -964.886972, -1208.76247, -1504.06572, -1859.4882, -2283.72137, -2776.29492, -3357.14432, -4056.19976, -4890.80001, -5882.51848, -7022.04479, -8332.2901, -9890.7632, -11749.9116, -13973.0115, -16615.3177, -19692.3808, -23242.2105, -27097.2211, -31532.6476, -36657.6673, -42581.4576, -49413.1957, -60379.149, -71692.1321, -82682.0499, -92678.807, -101012.308, -104896.686, -107188.132, -108123.311, -108055.955, -107176.389, -105626.161, -103598.036, -101284.777, -98879.1491, -96573.9171, -95144.9109, -94608.0178, -94458.2808, -94190.7431, -93300.4477, -91282.4378, -87631.7565, -83084.1039, -72209.692},
        {-5.44995917, -5.23442488, -5.01627556, -4.79558831, -4.57244022, -4.3469084, -4.11906994, -3.88900194, -3.65678151, -3.42248573, -3.18619172, -2.94797656, -2.70791736, -2.46609121, -2.22257522, -1.97744649, -1.73078211, -1.48265918, -1.2331548, -0.982346077, -0.7303101, -0.477123971, -0.222864791, 0.0323903436, 0.288564332, 0.545580076, 0.803360477, 1.06182843, 1.32090685, 1.58051863, 1.84058666, 2.10103386, 2.36178312, 2.62275734, 2.88672467, 3.15212553, 3.41736618, 3.68227717, 3.94668907, 4.21043244, 4.47333787, 4.73523591, 4.99595713, 5.25533211, 5.5131914, 5.76936557, 6.02368519, 6.27598084, 6.52608307, 6.77382246, 7.01902956, 7.28424863, 7.54610416, 7.80393434, 8.05707735, 8.30487137, 8.54665458, 8.78176515, 9.00954127, 9.22932112, 9.44044287, 9.64224471, 9.83406482, 10.0152414, 10.1851125, 10.3430165, 10.4882915, 10.630898, 10.7805414, 10.9146523, 11.0318801, 11.1308745, 11.2102849, 11.2687611, 11.3049525, 11.3175088, 11.3050795, 11.2663142, 11.1998624, 11.1043738, 10.978498, 10.8208844, 10.6301827, 10.4050425, 9.67329615, 8.6789049, 7.67148871, 6.66497606, 5.67329545, 4.71037537, 3.79014433, 2.92653083, 2.13346334, 1.42487038, 0.814680443, 0.316822018, -0.0547763944, -0.286186299, -0.363479199, -0.272726598, 0, 0.724085337, 1.7553181, 3.08105063, 4.68863526, 6.56542434, 8.69877021, 11.0760252, 13.6845417, 16.511672, 19.5447684, 22.7711833, 26.1782691, 29.7533781, 33.4838625, 37.3570748, 41.3603674, 45.1858057, 48.7785582, 52.793888, 57.1558715, 61.7885855, 66.6161064, 71.5625109, 76.5518756, 81.5082772, 86.3557921, 91.0184971, 95.4204686, 99.4857835, 103.398002, 106.726574, 109.317728, 111.017698, 112.869821, 114.121776, 113.702011, 111.097625, 105.795716, 97.2833849, 85.0477289, 72.1232052, 53.3699244, 27.5809374, -6.45070526, -49.9319528, -97.331339, -158.278833, -235.665987, -332.384354, -451.325489, -592.077766, -764.873133, -971.342893, -1216.19601, -1510.425, -1862.20119, -2279.69592, -2759.59472, -3324.16511, -4008.20462, -4835.08534, -5819.78538, -6953.68515, -8256.82162, -9797.78872, -11622.892, -13779.7156, -16314.8966, -19270.3832, -22690.776, -26479.777, -30844.5617, -35878.5033, -41674.9748, -48327.3491, -58851.4407, -69687.5709, -80198.5024, -89746.9977, -97695.8196, -101375.434, -103535.765, -104410.559, -104367.921, -103588.414, -102196.62, -100375.901, -98309.6189, -96181.1374, -94173.8185, -93144.9266, -92954.8284, -93099.03, -93073.0372, -92372.3561, -90492.4926, -86928.9527, -81763.6235, -69542.5868},
        {-5.62176446, -5.40307577, -5.18154157, -4.95724068, -4.73025195, -4.50065422, -4.26852633, -4.03394711, -3.7969954, -3.55775005, -3.31628988, -3.07269374, -2.82704046, -2.57940889, -2.32987785, -2.0785262, -1.82543277, -1.57067639, -1.31433591, -1.05649016, -0.797217979, -0.53659821, -0.274709688, -0.0116312514, 0.252558261, 0.517780011, 0.78395516, 1.05100487, 1.3188503, 1.58741262, 1.85661299, 2.12637256, 2.3966125, 2.66725398, 2.94058409, 3.21529126, 3.49003549, 3.76466115, 4.03901262, 4.3129343, 4.58627055, 4.85886577, 5.13056433, 5.40121062, 5.67064902, 5.93872392, 6.20527968, 6.47016071, 6.73321137, 6.99427606, 7.25319915, 7.53647229, 7.81671495, 8.09319383, 8.36517567, 8.63192717, 8.89271508, 9.1468061, 9.39346696, 9.63196438, 9.86156508, 10.0815358, 10.2911432, 10.4896541, 10.6763351, 10.8504531, 11.0112746, 11.1691696, 11.3342411, 11.4828568, 11.6135634, 11.7249081, 11.8154377, 11.8836993, 11.9282397, 11.9476059, 11.940345, 11.9050037, 11.8401291, 11.7442682, 11.6159679, 11.4537752, 11.2562369, 11.0219002, 10.260988, 9.22648042, 8.17722696, 7.1276216, 6.09205832, 5.08493115, 4.12063407, 3.21356111, 2.37810626, 1.62866352, 0.979626902, 0.445390411, 0.0403480501, -0.221106176, -0.324578263, -0.255674205, 0, 0.806067954, 1.91688058, 3.31947004, 5.0008685, 6.9481081, 9.14822102, 11.5882394, 14.2551954, 17.1361213, 20.218049, 23.4880109, 26.9330391, 30.5401657, 34.2964229, 38.1888429, 42.2044577, 46.0535274, 49.6835453, 53.7075759, 58.0533429, 62.6485703, 67.4209819, 72.2983014, 77.2082528, 82.0785599, 86.8369464, 91.4111363, 95.7288532, 99.7178212, 103.603694, 106.907024, 109.466157, 111.119437, 112.834986, 113.886261, 113.245445, 110.411949, 104.885186, 96.164567, 83.7495047, 70.6034215, 51.7074547, 25.8832751, -8.04744656, -51.2630395, -97.1434253, -156.614943, -232.805523, -328.843097, -447.855595, -590.9297, -765.729453, -972.475328, -1215.18183, -1505.60964, -1852.04153, -2262.76032, -2734.49553, -3289.70898, -3963.46601, -4780.84052, -5753.74556, -6873.72445, -8159.11027, -9673.62984, -11460.1735, -13557.8972, -16008.7231, -18868.2663, -22184.3967, -25906.6897, -30196.4314, -35135.5161, -40805.8385, -47289.2929, -57420.479, -67840.4099, -77942.804, -87121.3795, -94769.8547, -98338.0527, -100459.517, -101361.064, -101415.254, -100791.213, -99597.3396, -98005.7956, -96188.743, -94318.3437, -92566.7596, -92260.3266, -92418.4192, -92634.7259, -92502.9347, -91616.7341, -89569.8122, -85955.8574, -80089.5305, -66660.9763},
        {-5.81886307, -5.58757267, -5.35407909, -5.11843218, -4.8806818, -4.64087779, -4.39907, -4.15530828, -3.90964249, -3.66212246, -3.41279806, -3.16171912, -2.9089355, -2.65449705, -2.39845362, -2.14085505, -1.8817512, -1.62119191, -1.35922704, -1.09590643, -0.831279937, -0.565397406, -0.298308686, -0.0300636283, 0.239287919, 0.509696106, 0.781111083, 1.053483, 1.32676201, 1.60089826, 1.87584191, 2.1515431, 2.42795198, 2.7050187, 2.98558388, 3.26809164, 3.55095136, 3.83401939, 4.11715208, 4.40020579, 4.68303685, 4.96550162, 5.24745646, 5.5287577, 5.8092617, 6.08882481, 6.36730339, 6.64455377, 6.9204323, 7.19479535, 7.46749925, 7.76899187, 8.06787489, 8.36334148, 8.65458484, 8.94079815, 9.2211746, 9.49490736, 9.76118964, 10.0192146, 10.2681755, 10.5072654, 10.7356775, 10.9526051, 11.1572413, 11.3487794, 11.5264124, 11.7011016, 11.8835255, 12.0486068, 12.1947758, 12.3204628, 12.4240982, 12.5041122, 12.5589352, 12.5869975, 12.5867294, 12.5565613, 12.4949233, 12.400246, 12.2709595, 12.1054941, 11.9022803, 11.6597484, 10.8706039, 9.7967768, 8.70627539, 7.61394168, 6.53461775, 5.48314565, 4.47436742, 3.52312512, 2.64426081, 1.85261653, 1.16303433, 0.590356283, 0.149424426, -0.144919183, -0.277832493, -0.234473449, 0, 0.768639298, 1.86166562, 3.26416219, 4.96121225, 6.93789901, 9.17930572, 11.6705156, 14.3966119, 17.3426778, 20.4937965, 23.8350514, 27.3515255, 31.0283022, 34.8504647, 38.8030962, 42.8712799, 46.8229895, 50.6122927, 54.7153536, 59.0707322, 63.6169881, 68.2926811, 73.0363709, 77.7866173, 82.4819802, 87.0610191, 91.4622939, 95.6243644, 99.4857903, 103.379674, 106.705367, 109.283066, 110.932969, 112.47307, 113.224668, 112.242981, 109.048867, 103.163184, 94.1067881, 81.4005382, 67.9904061, 48.9297088, 23.0691727, -10.7404756, -53.6485095, -98.425921, -156.694835, -231.699097, -326.682549, -444.889036, -588.378874, -763.025969, -967.243173, -1204.43891, -1488.22085, -1827.61744, -2231.6571, -2700.17288, -3253.47843, -3921.91869, -4727.48833, -5682.95791, -6779.346, -8034.60237, -9512.31023, -11255.1063, -13302.2908, -15695.6355, -18489.1458, -21729.9069, -25381.8389, -29587.987, -34424.2245, -39966.4246, -46290.4606, -56088.4576, -66165.974, -75944.8203, -84846.8066, -92293.7435, -95855.862, -98040.9378, -99064.315, -99288.5222, -98870.6431, -97906.4349, -96556.0476, -94979.6313, -93337.3362, -91789.3124, -92551.8254, -93070.0845, -93135.0077, -92537.5127, -91068.5173, -88518.9394, -84679.6966, -78037.0053, -63576.2982},
        {-6.05520415, -5.8088978, -5.56120639, -5.31214449, -5.06172667, -4.80996751, -4.55688157, -4.30248342, -4.04678763, -3.78980877, -3.53156142, -3.27206013, -3.01131949, -2.74935405, -2.4861784, -2.2218071, -1.95625471, -1.68953582, -1.42166499, -1.15265678, -0.882525775, -0.611286538, -0.338953639, -0.0655416479, 0.208934865, 0.484461329, 0.761023177, 1.03860584, 1.31719474, 1.59677531, 1.87733299, 2.15885321, 2.44132138, 2.72472295, 3.01271479, 3.30336938, 3.59471507, 3.88661816, 4.17894493, 4.47156167, 4.76433465, 5.05713016, 5.3498145, 5.64225393, 5.93431476, 6.22586325, 6.51676571, 6.80688841, 7.09609763, 7.38425967, 7.6712408, 7.99122923, 8.3090253, 8.62375126, 8.93452936, 9.24048185, 9.54073099, 9.83439902, 10.1206082, 10.3984808, 10.6671391, 10.9257052, 11.1733015, 11.4090503, 11.6320737, 11.841494, 12.0364335, 12.2286133, 12.4294521, 12.6120877, 12.7748259, 12.915972, 13.0338317, 13.1267104, 13.1929137, 13.230747, 13.2385159, 13.2145259, 13.1570826, 13.0644914, 12.9350579, 12.7670876, 12.558886, 12.3087587, 11.4918444, 10.3790361, 9.24751484, 8.11256397, 6.98946673, 5.89350643, 4.83996633, 3.84412974, 2.92127993, 2.08670019, 1.35567381, 0.743484066, 0.265414253, -0.0632523458, -0.227232445, -0.211242758, 0, 0.700702108, 1.75063187, 3.13254675, 4.82920418, 6.82336161, 9.0977765, 11.6352063, 14.4184084, 17.4301403, 20.6531595, 24.0702234, 27.6640894, 31.417515, 35.3132576, 39.3340747, 43.4627238, 47.5349003, 51.5123527, 55.7104788, 60.0805227, 64.5737286, 69.1413405, 73.7346028, 78.3047596, 82.803055, 87.1807332, 91.3890383, 95.3792147, 99.1025063, 103.014128, 106.376564, 108.989867, 110.654092, 112.023871, 112.471969, 111.143257, 107.581415, 101.330121, 91.9330539, 78.9338913, 65.2729933, 46.0635844, 20.1847753, -13.4843239, -56.0646027, -99.9670224, -157.199883, -231.061556, -324.850414, -441.864827, -584.850567, -758.334007, -958.957731, -1189.60804, -1465.91886, -1797.80068, -2195.16399, -2661.73865, -3215.0205, -3879.2479, -4671.7816, -5607.75806, -6678.11356, -7900.95664, -9340.43766, -11039.573, -13039.5125, -15382.7886, -18118.7781, -21292.9861, -24875.4535, -28998.7599, -33734.5327, -39154.3995, -45329.9878, -54834.2489, -64612.1556, -74110.0046, -82774.0925, -90050.7158, -93611.7904, -95860.9142, -97004.8328, -97419.8414, -97245.5721, -96551.0114, -95479.3238, -94173.6738, -92777.2257, -91433.1441, -93214.3568, -94011.2392, -93825.0998, -92657.2467, -90508.9885, -87381.6334, -83276.4898, -75820.2442, -60405.4266},
        {-6.3334436, -6.07146908, -5.80892267, -5.54578369, -5.2820315, -5.01764541, -4.75260476, -4.48688888, -4.22047712, -3.95334879, -3.68548324, -3.41685979, -3.14745778, -2.87725655, -2.60623543, -2.33437375, -2.06165084, -1.78804603, -1.51353867, -1.23810808, -0.961733602, -0.684394561, -0.406070294, -0.126740133, 0.153616587, 0.435020534, 0.717492374, 1.00105277, 1.2857224, 1.57152192, 1.858472, 2.14659331, 2.43590651, 2.72643227, 3.02273347, 3.32246414, 3.6232237, 3.92488541, 4.22732253, 4.53040833, 4.83401606, 5.138019, 5.4422904, 5.74670353, 6.05113166, 6.35544803, 6.65952592, 6.9632386, 7.26645931, 7.56906133, 7.87091792, 8.20954805, 8.54636319, 8.88042051, 9.2107772, 9.53649043, 9.85661739, 10.1702152, 10.4763412, 10.7740523, 11.0624059, 11.3404592, 11.6072692, 11.8618931, 12.1033882, 12.3308117, 12.5432206, 12.7533325, 12.973537, 13.1747178, 13.3550466, 13.5126949, 13.6458344, 13.7526367, 13.8312735, 13.8799165, 13.8967371, 13.8799071, 13.8275982, 13.7379818, 13.6092298, 13.4395136, 13.227005, 12.9698755, 12.1254811, 10.9737766, 9.80121697, 8.62352444, 7.45642111, 6.31562914, 5.21687066, 4.17586782, 3.20834278, 2.33001766, 1.55661461, 0.903855788, 0.387463326, 0.0231593723, -0.17333393, -0.186294435, 0, 0.622791602, 1.62088382, 2.97463687, 4.66441097, 6.67056632, 8.97346315, 11.5534617, 14.3909221, 17.4662046, 20.7596695, 24.2516769, 27.9225871, 31.7527602, 35.7225566, 39.8123363, 44.0024597, 48.1982114, 52.3692268, 56.6656794, 61.0518418, 65.4919867, 69.9503866, 74.3913144, 78.7790424, 83.0778434, 87.2519901, 91.265755, 95.0834107, 98.6692299, 102.600927, 106.004405, 108.659902, 110.347659, 111.563604, 111.730133, 110.078792, 106.175115, 99.5846344, 89.872885, 76.6054001, 62.720572, 43.3845536, 17.5029714, -16.018548, -58.2743781, -101.598842, -158.036376, -230.871367, -323.388199, -438.871261, -580.407306, -751.7199, -947.947409, -1171.58622, -1440.15534, -1764.42894, -2155.18117, -2620.23628, -3173.98877, -3834.05123, -4612.303, -5527.26395, -6570.47672, -7760.60868, -9162.28862, -10819.0483, -12774.3285, -15071.6375, -17754.8167, -20867.5191, -24380.8248, -28422.1246, -33060.1928, -38363.8036, -44401.7314, -53652.7085, -63173.0616, -72429.0755, -80887.035, -88013.2248, -91554.9483, -93843.4594, -95081.7491, -95692.5931, -95791.9845, -95404.3395, -94650.2301, -93650.2281, -92524.9057, -91394.8348, -94036.2119, -95007.9988, -94508.9406, -92737.7826, -89893.2699, -86174.1477, -81779.1613, -73489.6356, -57192.9751},
        {-6.65623729, -6.37970441, -6.10322712, -5.82675588, -5.55024119, -5.2736335, -4.99688329, -4.71994105, -4.44275723, -4.16528232, -3.88746679, -3.60926111, -3.33061576, -3.0514812, -2.77180792, -2.49154639, -2.21064708, -1.92906047, -1.64673702, -1.36362722, -1.07968153, -0.794850433, -0.509084401, -0.222333907, 0.0654505744, 0.354318571, 0.644319607, 0.93550321, 1.22791891, 1.52161622, 1.81664468, 2.11305381, 2.41089314, 2.7102122, 3.01639661, 3.32671562, 3.63837432, 3.95124907, 4.26521622, 4.58015215, 4.8959332, 5.21243574, 5.52953612, 5.84711071, 6.16503587, 6.48318794, 6.8014433, 7.1196783, 7.4377693, 7.75559266, 8.07302473, 8.43031199, 8.78608554, 9.13934659, 9.48909637, 9.83433608, 10.1740669, 10.5072902, 10.833007, 11.1502186, 11.4579262, 11.755131, 12.0408343, 12.3140372, 12.573741, 12.8189468, 13.048656, 13.2768866, 13.5172964, 13.7379152, 13.9367706, 14.1118904, 14.2613024, 14.3830342, 14.4751135, 14.5355682, 14.5624259, 14.5537143, 14.5074612, 14.4216943, 14.2944413, 14.12373, 13.907588, 13.6440431, 12.7722857, 11.5815167, 10.3676534, 9.1468591, 7.93529675, 6.74912944, 5.60452025, 4.51763225, 3.5046285, 2.58167208, 1.76492605, 1.07055348, 0.514717442, 0.113581007, -0.116692758, -0.159940783, 0, 0.555442999, 1.50952593, 2.84044572, 4.52639929, 6.54558357, 8.87619548, 11.4964319, 14.3844899, 17.5185662, 20.8768579, 24.4375619, 28.178875, 32.0789942, 36.1161164, 40.2684386, 44.5141577, 48.8218748, 53.1684162, 57.5536832, 61.9538167, 66.3449573, 70.7032457, 75.0048226, 79.2258287, 83.3424047, 87.3306914, 91.1668293, 94.8269593, 98.2872219, 102.233943, 105.672678, 108.36651, 110.078525, 111.168483, 111.101132, 109.182106, 104.995488, 98.1253588, 88.1558023, 74.6709008, 60.6025316, 41.1680883, 15.29665, -18.0827041, -60.040895, -103.153492, -159.110606, -231.106994, -322.337416, -435.996629, -575.111622, -743.249978, -934.540617, -1151.2704, -1412.382, -1729.33995, -2113.60881, -2576.7092, -3130.03684, -3784.92631, -4547.63517, -5440.59354, -6456.88506, -7615.99409, -8982.13959, -10599.0066, -12511.505, -14763.6373, -17394.9151, -20447.3905, -23891.2439, -27851.4558, -32394.9569, -37588.6775, -43499.5484, -52538.692, -61842.7988, -70892.7514, -79169.4322, -86153.7238, -89634.4466, -91912.5866, -93194.1958, -93990.1589, -94385.865, -94339.6894, -93943.3718, -93288.652, -92467.2695, -91570.9642, -94805.6817, -95826.4787, -94990.4689, -92654.766, -89176.4836, -84912.7355, -80220.6352, -71095.5679, -53983.5578},
        {-7.01295006, -6.73629337, -6.45878381, -6.18039754, -5.9011107, -5.62089946, -5.33973995, -5.05760833, -4.77448075, -4.49033337, -4.20514232, -3.91888377, -3.63153386, -3.34306874, -3.05346457, -2.7626975, -2.47074367, -2.17757923, -1.88318035, -1.58752316, -1.29058383, -0.992338493, -0.692763308, -0.391834424, -0.0895279931, 0.214179835, 0.519312909, 0.825895078, 1.13395019, 1.4435021, 1.75457465, 2.06719169, 2.38137708, 2.69715465, 3.01908538, 3.34482911, 3.67199118, 4.00044821, 4.3300768, 4.66075357, 4.99235512, 5.32475808, 5.65783905, 5.99147463, 6.32554146, 6.65991612, 6.99447525, 7.32909544, 7.66365331, 7.99802548, 8.33208854, 8.70744944, 9.08134983, 9.45276169, 9.82065701, 10.1840078, 10.5417859, 10.8929635, 11.2365124, 11.5714047, 11.8966123, 12.2111072, 12.5138615, 12.803847, 13.0800357, 13.3413997, 13.5869109, 13.8327405, 14.0946466, 14.3361293, 14.5550456, 14.7492525, 14.916607, 15.0549662, 15.1621871, 15.2361266, 15.2746419, 15.27559, 15.2368279, 15.1562126, 15.0316011, 14.8608505, 14.6418177, 14.37236, 13.4722456, 12.2405352, 10.9832377, 9.71697011, 8.45834954, 7.22399311, 6.03051794, 4.89454115, 3.83267983, 2.86155111, 1.99777209, 1.25795988, 0.658731607, 0.216704369, -0.0515047191, -0.129278546, 0, 0.547565369, 1.50300004, 2.84356095, 4.54650505, 6.5890893, 8.94857063, 11.602206, 14.5272524, 17.7009666, 21.1006058, 24.7034268, 28.4866865, 32.427642, 36.5035501, 40.6916679, 44.9692522, 49.3314846, 53.7581828, 58.1860599, 62.5962138, 66.9697425, 71.2877439, 75.5313159, 79.6815565, 83.7195636, 87.6264353, 91.3832693, 94.9711636, 98.3712162, 102.300133, 105.733681, 108.432276, 110.156333, 111.241942, 111.161682, 109.234618, 105.048464, 98.1909322, 88.2497374, 74.8125925, 60.8754241, 41.5804503, 15.8485601, -17.3993573, -59.2424127, -103.977532, -161.161557, -233.569145, -323.974954, -435.153641, -568.171132, -729.599264, -915.975295, -1130.39342, -1389.33491, -1703.27051, -2082.67097, -2537.63733, -3079.09406, -3718.36893, -4459.12426, -5327.68782, -6325.54078, -7473.40234, -8824.58947, -10417.4562, -12284.5471, -14462.7101, -17010.0946, -19972.8006, -23345.6338, -27233.88, -31701.3951, -36812.0349, -42629.6551, -51583.9084, -60781.4045, -69694.55, -77795.7514, -84557.4154, -87642.6667, -89539.3815, -90495.3729, -91319.5661, -92037.6672, -92441.5856, -92568.7174, -92456.4585, -92142.205, -91663.3528, -94120.7642, -94762.5011, -93820.0943, -91525.0743, -88108.9719, -83803.3176, -78839.6422, -68795.1633, -50619.2152},
        {-7.42228891, -7.1477411, -6.8711277, -6.59246088, -6.31175285, -6.02901577, -5.74426185, -5.45750326, -5.1687522, -4.87802086, -4.5853214, -4.29066604, -3.99406694, -3.69553631, -3.39508632, -3.09272916, -2.78847702, -2.48234209, -2.17433654, -1.86447258, -1.55276239, -1.23921814, -0.923852042, -0.606676265, -0.287703003, 0.0330555597, 0.355587235, 0.679879837, 1.00592118, 1.33369907, 1.66320133, 1.99441577, 2.3273302, 2.66193243, 3.00181835, 3.34509584, 3.68982942, 4.03588983, 4.38314779, 4.73147401, 5.08073923, 5.43081417, 5.78156956, 6.13287612, 6.48460457, 6.83662564, 7.18881005, 7.54102853, 7.8931518, 8.24505059, 8.59659563, 8.98988196, 9.38164065, 9.77082708, 10.1563966, 10.5373047, 10.9125066, 11.2809578, 11.6416136, 11.9934295, 12.3353608, 12.6663629, 12.9853912, 13.2914011, 13.5833479, 13.8601871, 14.1208739, 14.3840729, 14.6679751, 14.9308877, 15.1704883, 15.3844547, 15.5704647, 15.7261959, 15.8493261, 15.937533, 15.9884944, 15.9998879, 15.9693913, 15.8946824, 15.7734389, 15.6033384, 15.3820588, 15.1072777, 14.1787731, 12.9057487, 11.6045943, 10.2923907, 8.98621872, 7.70315896, 6.46029221, 5.27469918, 4.16346061, 3.14365722, 2.23236975, 1.44667893, 0.80366548, 0.320410133, 0.0139936175, -0.0985033365, 0, 0.585427895, 1.57882773, 2.95721919, 4.69762194, 6.77705565, 9.17254, 11.8610947, 14.8197393, 18.0254936, 21.4553772, 25.0864099, 28.8956112, 32.8600008, 36.9565985, 41.1624239, 45.4544966, 49.8366604, 54.2897669, 58.7301625, 63.1406147, 67.5038914, 71.8027602, 76.0199887, 80.1383446, 84.1405956, 88.0095094, 91.7278536, 95.278396, 98.6439041, 102.539419, 105.946935, 108.629537, 110.350312, 111.465024, 111.44042, 109.585196, 105.484636, 98.7240193, 88.8886291, 75.5637465, 61.7711155, 42.6136757, 17.0043576, -16.1439085, -57.9181921, -104.489458, -163.089907, -236.035634, -325.642737, -434.227311, -560.68081, -714.929644, -896.312095, -1108.84281, -1366.084, -1677.31123, -2051.80005, -2497.49866, -3025.31163, -3647.5916, -4366.16883, -5210.82403, -6191.68417, -7330.61485, -8669.37211, -10240.5819, -12063.5141, -14167.3319, -16630.1713, -19502.4667, -22802.944, -26618.624, -31010.9169, -36041.233, -41770.9825, -50666.247, -59782.5976, -68580.2766, -76519.5264, -83060.5895, -85717.995, -87194.8407, -87785.246, -88622.5361, -89668.645, -90535.8372, -91203.53, -91651.1406, -91858.0863, -91803.7843, -93207.2226, -93357.7996, -92315.5953, -90140.6895, -86893.1621, -82633.093, -77420.562, -66509.9117, -47389.2013},
        {-7.89041817, -7.61892175, -7.3439707, -7.06561891, -6.78392025, -6.49892859, -6.21069782, -5.9192818, -5.62473441, -5.32710953, -5.02646102, -4.72284278, -4.41630866, -4.10691255, -3.79470832, -3.47974985, -3.162091, -2.84178567, -2.51888772, -2.19345102, -1.86552945, -1.53517689, -1.20244722, -0.867394297, -0.530072009, -0.190534229, 0.151165169, 0.49497231, 0.840833319, 1.18869432, 1.53850144, 1.8902008, 2.24373853, 2.59906075, 2.95885207, 3.32163178, 3.68588623, 4.0514727, 4.41824843, 4.78607068, 5.1547967, 5.52428375, 5.89438909, 6.26496997, 6.63588365, 7.00698738, 7.37813842, 7.74919403, 8.12001145, 8.49044795, 8.86036079, 9.27141758, 9.68075884, 10.0873355, 10.4900984, 10.8879984, 11.2799865, 11.6650134, 12.0420302, 12.4099877, 12.7678368, 13.1145284, 13.4490133, 13.7702424, 14.0771668, 14.3687371, 14.6439044, 14.9241075, 15.2302454, 15.5148891, 15.7755318, 16.0096665, 16.2147862, 16.3883842, 16.5279534, 16.6309869, 16.6949779, 16.7174193, 16.6958044, 16.6276261, 16.5103776, 16.3415518, 16.118642, 15.8391412, 14.8822883, 13.5677077, 12.2224545, 10.8640853, 9.51015693, 8.17822585, 6.8858487, 5.65058211, 4.4899827, 3.42160708, 2.46301188, 1.63175371, 0.945389193, 0.421474945, 0.0775675841, -0.0687762703, 0, 0.682075939, 1.7610899, 3.21465183, 5.02037168, 7.1558594, 9.59872494, 12.3265783, 15.3170293, 18.547688, 21.9961644, 25.6400683, 29.4570097, 33.4245987, 37.520445, 41.7221588, 46.0073499, 50.3683578, 54.7848969, 59.20126, 63.5982134, 67.9565235, 72.2569565, 76.4802788, 80.6072566, 84.6186562, 88.4952439, 92.2177861, 95.767049, 99.1237989, 102.968308, 106.325351, 108.965843, 110.660698, 111.827567, 111.913997, 110.195071, 106.247683, 99.6487261, 89.9750943, 76.8036813, 63.1136339, 44.0441242, 18.5063878, -14.58834, -56.3288236, -104.649555, -164.582139, -238.003909, -326.792198, -432.824339, -552.928856, -700.202656, -876.695217, -1087.29489, -1342.68603, -1650.95073, -2020.17105, -2455.90155, -2969.32626, -3574.38648, -4271.90149, -5094.18226, -6058.99775, -7189.04991, -8515.39947, -10065.2703, -11845.5857, -13878.3064, -16259.8289, -19044.4448, -22270.8173, -26011.2705, -30325.5053, -35273.2231, -40914.125, -49754.997, -58796.6842, -67487.1166, -75274.2242, -81605.9368, -83850.9142, -84922.5368, -85159.8505, -86013.782, -87383.2576, -88703.9201, -89897.8599, -90887.1678, -91593.9341, -91940.2496, -92167.5502, -91766.6038, -90611.8053, -88577.5498, -85538.2325, -81368.2483, -75941.9923, -64262.027, -44391.6032},
        {-8.42350218, -8.15470944, -7.88102475, -7.60254462, -7.31936556, -7.03158407, -6.73929668, -6.44259988, -6.14159019, -5.83636411, -5.52701816, -5.21364883, -4.89635265, -4.57522611, -4.25036573, -3.92186802, -3.58982949, -3.25434664, -2.91551598, -2.57343402, -2.22819728, -1.87990225, -1.52864545, -1.17452339, -0.817632568, -0.458069505, -0.0959307046, 0.268687324, 0.635688072, 1.00497503, 1.37645169, 1.75002155, 2.12558808, 2.5030548, 2.88444306, 3.26855288, 3.65415879, 4.04109557, 4.42919798, 4.81830077, 5.2082387, 5.59884655, 5.98995907, 6.38141102, 6.77303717, 7.16467227, 7.5561511, 7.9473084, 8.33797895, 8.7279975, 9.11719882, 9.54586428, 9.97250522, 10.3960796, 10.8155454, 11.2298606, 11.637983, 12.0388708, 12.4314818, 12.814774, 13.1877053, 13.5492338, 13.8983173, 14.2339139, 14.5549815, 14.860478, 15.1493614, 15.4460676, 15.7744207, 16.0808324, 16.3626091, 16.6170571, 16.8414826, 17.0331921, 17.1894918, 17.3076881, 17.3850873, 17.4189958, 17.4067198, 17.3455658, 17.23284, 17.0658487, 16.8418984, 16.5582952, 15.5732111, 14.2169629, 12.8275494, 11.4230183, 10.0214168, 8.64079261, 7.29919307, 6.0146657, 4.80525796, 3.68901733, 2.68399129, 1.80822732, 1.07977288, 0.516675452, 0.136982513, -0.0412584626, 0, 0.850554865, 2.0738674, 3.64909023, 5.555376, 7.77187732, 10.2777468, 13.0521372, 16.0742009, 19.3230908, 22.7779594, 26.4179593, 30.2222431, 34.1699636, 38.2402733, 42.4123248, 46.6652708, 50.9575326, 55.2653012, 59.6146215, 63.9802038, 68.3367582, 72.6589949, 76.9216241, 81.099356, 85.1669009, 89.0989689, 92.8702703, 96.4555151, 99.8294137, 103.603304, 106.881839, 109.448739, 111.087726, 112.319409, 112.559066, 111.025472, 107.281285, 100.889159, 91.4117503, 78.4117154, 64.7270075, 45.6481553, 20.0969958, -13.0046341, -54.7348975, -104.418106, -165.324738, -238.971418, -326.874772, -430.551427, -545.203472, -686.379835, -858.268862, -1066.42598, -1319.19776, -1623.67761, -1986.95892, -2412.45436, -2911.77461, -3500.54569, -4179.45483, -4981.94263, -5931.16405, -7050.12582, -8361.58351, -9888.40797, -11627.9418, -13596.4375, -15903.7509, -18606.7905, -21756.8965, -25417.4022, -29647.1436, -34504.9566, -40049.677, -48819.4475, -57773.9706, -66352.2556, -73993.3117, -80136.1482, -82031.9065, -82766.0422, -82715.2222, -83608.0167, -85285.9642, -87027.3101, -88701.7574, -90179.0093, -91328.7689, -92020.7392, -91104.2402, -90143.143, -88843.5575, -86911.5936, -84053.3611, -79974.9702, -74382.5305, -62073.7229, -41724.5081},
        {-9.02770528, -8.75997831, -8.48600176, -8.205911, -7.9198414, -7.62792836, -7.33030726, -7.02711346, -6.71848237, -6.40454935, -6.08544978, -5.76131906, -5.43229255, -5.09850564, -4.76009372, -4.41719216, -4.06993634, -3.71846164, -3.36290345, -3.00339714, -2.6400781, -2.27308171, -1.90254335, -1.5285984, -1.15138223, -0.771030242, -0.387677803, -0.00146029783, 0.387486893, 0.779028389, 1.17302881, 1.56935277, 1.9678649, 2.3684298, 2.77284786, 3.1799751, 3.58864429, 3.99865721, 4.40981567, 4.82192147, 5.23477641, 5.64818229, 6.06194092, 6.47585408, 6.88972358, 7.30335122, 7.7165388, 8.12908812, 8.54080098, 8.95147919, 9.36092453, 9.80703007, 10.2506806, 10.6908522, 11.1265209, 11.5566628, 11.9802539, 12.3962702, 12.8036879, 13.201483, 13.5886315, 13.9641096, 14.3268932, 14.6759584, 15.0102812, 15.3288378, 15.6306042, 15.9431767, 16.2934645, 16.6214165, 16.9241534, 17.1987958, 17.4424646, 17.6522805, 17.8253641, 17.9588362, 18.0498175, 18.0954287, 18.0927905, 18.0390237, 17.931249, 17.7665871, 17.5421588, 17.2550846, 16.2419614, 14.844065, 13.4106105, 11.9601539, 10.5112511, 9.08245811, 7.69233099, 6.35942571, 5.10229826, 3.93950463, 2.88960083, 1.97114284, 1.20268667, 0.6027883, 0.190003736, -0.017111029, 0, 1.10391003, 2.54124111, 4.29376579, 6.34325664, 8.6714862, 11.260227, 14.0912517, 17.1463328, 20.4072428, 23.8557543, 27.4736399, 31.2426722, 35.1446236, 39.1612667, 43.2743741, 47.4657184, 51.6351405, 55.7527084, 59.9855162, 64.2977798, 68.6537148, 73.0175371, 77.3534624, 81.6257065, 85.7984853, 89.8360144, 93.7025097, 97.362187, 100.779262, 104.460914, 107.629309, 110.085773, 111.631633, 112.930387, 113.352277, 112.03763, 108.52912, 102.369423, 93.101214, 80.2671675, 66.4352644, 47.2021285, 21.5185269, -11.6647731, -53.3970043, -103.755397, -165.004187, -238.435608, -325.341893, -427.015275, -537.792857, -674.422718, -842.17723, -1046.91242, -1295.67596, -1594.98049, -1951.33865, -2366.76546, -2853.29338, -3427.86138, -4091.96144, -4878.28522, -5811.86561, -6915.26087, -8206.8362, -9706.88148, -11407.7624, -13322.5293, -15566.621, -18197.5598, -21268.8243, -24842.6021, -28977.815, -33733.3847, -39168.2329, -47828.8876, -56664.7629, -65112.8789, -72610.256, -78593.9143, -80251.4546, -80768.9297, -80547.3967, -81519.9533, -83481.224, -85587.4831, -87665.2727, -89541.1346, -91041.6111, -91993.244, -90119.7857, -88641.6467, -87145.685, -85218.7589, -82447.7263, -78419.4453, -72720.7742, -59967.2132, -39486.0033},
        {-9.72480593, -9.45169299, -9.17146031, -8.88426442, -8.59026184, -8.2896091, -7.98246272, -7.66897923, -7.34931516, -7.02362703, -6.69207137, -6.35480469, -6.01198354, -5.66376443, -5.31030388, -4.95175843, -4.5882846, -4.22003892, -3.84717791, -3.46985809, -3.088236, -2.70246815, -2.31271108, -1.91912131, -1.52185536, -1.12106976, -0.716921044, -0.309565725, 0.100839667, 0.514138606, 0.930174566, 1.34879102, 1.76983145, 2.19313932, 2.62144631, 3.05309104, 3.48637743, 3.92105522, 4.35687415, 4.79358399, 5.23093446, 5.66867533, 6.10655633, 6.54432721, 6.98173772, 7.41853761, 7.85447662, 8.2893045, 8.722771, 9.15462586, 9.58461882, 10.0462902, 10.5048667, 10.9593655, 11.4088039, 11.8521991, 12.2885682, 12.7169287, 13.1362976, 13.5456923, 13.9441299, 14.3306277, 14.7042029, 15.0638728, 15.4086545, 15.7375654, 16.0496227, 16.3758797, 16.7466207, 17.0947907, 17.4173302, 17.7111797, 17.9732795, 18.2005702, 18.3899922, 18.538486, 18.6429921, 18.7004508, 18.7078027, 18.6619882, 18.5599477, 18.3986218, 18.1749509, 17.8858754, 16.8448307, 15.4048865, 13.9273274, 12.4312752, 10.9358512, 9.46017701, 8.02337413, 6.64456406, 5.34286832, 4.13740842, 3.04730589, 2.09168223, 1.28965895, 0.660357577, 0.222899618, -0.00359341236, 0, 1.64044021, 3.52491848, 5.64129076, 7.97741295, 10.521141, 13.2603308, 16.1828384, 19.2765195, 22.5292302, 25.9288264, 29.463164, 33.120099, 36.8874872, 40.7531845, 44.705047, 48.7309306, 52.5947579, 56.2523945, 60.1982119, 64.3720801, 68.7138695, 73.1634504, 77.660693, 82.1454675, 86.5576442, 90.8370933, 94.9236851, 98.7572898, 102.277778, 105.784255, 108.725636, 110.93402, 112.241508, 113.570527, 114.208013, 113.150146, 109.901926, 103.968353, 94.8544291, 82.0651539, 67.4968765, 47.5354474, 21.2179945, -12.4183541, -54.3364705, -103.524075, -163.412979, -235.459842, -321.121325, -421.854088, -533.362127, -669.885798, -834.373825, -1031.85764, -1271.62085, -1561.03795, -1907.4834, -2314.63417, -2794.40259, -3362.73457, -4023.93154, -4804.6302, -5720.64506, -6791.84391, -8044.96263, -9503.58455, -11170.459, -13063.7678, -15278.0843, -17864.7705, -20842.8834, -24301.06, -28301.8446, -32907.7816, -38181.4154, -46631.3244, -55262.5106, -63526.0099, -70872.8583, -76754.0918, -78414.7514, -78982.5311, -78853.8933, -80009.6696, -82219.1762, -84591.5752, -86929.1905, -89034.3458, -90709.3649, -91756.5716, -89776.0592, -87999.6393, -86103.4936, -83763.8041, -80656.7525, -76458.5206, -70845.2903, -58084.4595, -38124.4189},
        {-10.4934981, -10.2134565, -9.92554817, -9.62994468, -9.32681755, -9.01633828, -8.69867842, -8.37400948, -8.042503, -7.70433049, -7.35966348, -7.0086735, -6.65153208, -6.28841074, -5.91948101, -5.54491441, -5.16488246, -4.7795567, -4.38910865, -3.99370984, -3.59353178, -3.18874602, -2.77952406, -2.36603745, -1.9484577, -1.52695634, -1.1017049, -0.672874899, -0.240637867, 0.194834671, 0.63337119, 1.07480016, 1.51895006, 1.96564937, 2.41900001, 2.87660385, 3.33601138, 3.7969124, 4.25899668, 4.72195403, 5.18547423, 5.64924708, 6.11296238, 6.5763099, 7.03897946, 7.50066082, 7.9610438, 8.41981818, 8.87667375, 9.33130031, 9.78338765, 10.2617327, 10.7362872, 11.2061098, 11.6702594, 12.1277948, 12.5777748, 13.0192583, 13.4513039, 13.8729705, 14.283317, 14.681402, 15.0662846, 15.4370233, 15.7926771, 16.1323047, 16.454965, 16.7950683, 17.1861763, 17.5544377, 17.8966197, 18.2094894, 18.4898139, 18.7343603, 18.9398957, 19.1031872, 19.2210019, 19.2901069, 19.3072694, 19.2692565, 19.1728351, 19.0147726, 18.7918359, 18.5007922, 17.4322821, 15.9509884, 14.430081, 12.889242, 11.3481538, 9.82649863, 8.34395875, 6.92021643, 5.57495395, 4.32785356, 3.19859754, 2.20686814, 1.37234765, 0.714718318, 0.253662423, 0.00886222737, 0, 2.14726935, 4.45498467, 6.9166951, 9.5259498, 12.2762979, 15.1612885, 18.1744709, 21.309394, 24.5596072, 27.9186595, 31.3801, 34.9374779, 38.5843424, 42.3142426, 46.1207276, 49.9973466, 53.5812374, 56.8150331, 60.4984516, 64.548668, 68.8828577, 73.4181959, 78.0718579, 82.761019, 87.4028544, 91.9145396, 96.2132497, 100.21616, 103.840446, 107.191148, 109.928692, 111.913894, 113.007569, 114.370083, 115.212477, 114.399245, 111.401337, 105.689704, 96.7352965, 84.0090653, 68.7308709, 48.0904772, 21.2166565, -12.7618183, -54.7161748, -102.09687, -160.051095, -230.305268, -314.58581, -414.619141, -528.017712, -665.650095, -827.54171, -1017.66788, -1248.06834, -1527.16204, -1863.36797, -2262.37019, -2735.94415, -3298.84884, -3958.48077, -4734.82675, -5633.28179, -6670.78905, -7883.85942, -9300.02186, -10933.9158, -12809.7285, -14998.9088, -17546.172, -20436.3295, -23783.2158, -27652.2818, -32108.9783, -37218.7562, -45409.422, -53793.482, -61845.7981, -69041.2323, -74854.6464, -76662.3315, -77436.7677, -77552.2046, -78936.1343, -81347.9811, -83899.9849, -86383.9293, -88591.5979, -90314.7742, -91345.2418, -89367.3552, -87345.1432, -85059.2879, -82290.471, -78819.3744, -74426.68, -68893.0698, -56237., -37121.6499},
        {-11.3273078, -11.0401941, -10.74451, -10.4404379, -10.1281604, -9.80785985, -9.47971867, -9.1439193, -8.80064417, -8.45007573, -8.0923964, -7.72778861, -7.35643479, -6.97851739, -6.59421883, -6.20372154, -5.80720796, -5.40486052, -4.99686165, -4.58339378, -4.16463935, -3.74078079, -3.31200053, -2.878481, -2.44040465, -1.99795389, -1.55131116, -1.10065889, -0.646179527, -0.188055489, 0.273530787, 0.738396869, 1.20636032, 1.67723872, 2.15674958, 2.64163639, 3.12857174, 3.61718173, 4.10709246, 4.59793003, 5.08932054, 5.58089011, 6.07226482, 6.56307079, 7.05293411, 7.5414809, 8.02833724, 8.51312925, 8.99548303, 9.47502468, 9.9513803, 10.4489184, 10.9419864, 11.4296739, 11.9110706, 12.3852664, 12.851351, 13.308414, 13.7555452, 14.1918344, 14.6163713, 15.0282456, 15.4265471, 15.8103654, 16.1787904, 16.5309118, 16.8658192, 17.2209575, 17.6328506, 18.0214837, 18.3834602, 18.7153836, 19.013857, 19.2754839, 19.4968676, 19.6746114, 19.8053187, 19.8855927, 19.9120368, 19.8812545, 19.7898489, 19.6344234, 19.4115814, 19.1179262, 18.0226016, 16.5012418, 14.9382246, 13.3537607, 11.7680611, 10.2013363, 8.6737974, 7.205655, 5.81711996, 4.52840307, 3.35971513, 2.33126693, 1.46326928, 0.77593296, 0.289468776, 0.0240875242, 0, 2.48309807, 5.0742659, 7.77059306, 10.5691691, 13.4670836, 16.4614261, 19.5492863, 22.7277535, 25.9939175, 29.3448678, 32.777694, 36.2894855, 39.877332, 43.5383231, 47.2695483, 51.0680972, 54.4934233, 57.4800966, 61.0150979, 65.0017376, 69.3433262, 73.943174, 78.7045914, 83.5308888, 88.3253767, 92.9913654, 97.4321653, 101.551087, 105.25144, 108.514075, 111.136332, 112.998259, 113.979906, 115.39259, 116.400724, 115.787054, 113.002246, 107.49697, 98.7218928, 86.1276842, 70.4201397, 49.4128046, 22.3107789, -11.6808374, -53.3569444, -98.6657206, -154.460468, -222.747767, -305.534199, -404.826342, -520.104429, -658.989151, -818.997278, -1003.13362, -1225.60773, -1495.59846, -1822.28466, -2212.71604, -2678.68449, -3234.30464, -3889.62359, -4659.25958, -5540.84983, -6548.96963, -7726.89488, -9104.72743, -10705.0682, -12556.0744, -14713.4065, -17217.1677, -20033.8275, -23288.4507, -27046.6288, -31373.9532, -36336.0156, -44227.9711, -52328.4567, -60145.673, -67187.8206, -72963.1, -75049.0314, -76171.6163, -76666.4899, -78310.3039, -80868.2911, -83505.2607, -86016.4665, -88197.1626, -89842.6028, -90748.0412, -88626.8554, -86364.0419, -83787.9265, -80726.8345, -77009.0916, -72463.0234, -66916.9553, -54345.7006, -36295.7187},
        {-12.2111996, -11.9154198, -11.6107239, -11.2972866, -10.9752826, -10.6448866, -10.3062733, -9.95961754, -9.60509394, -9.24287724, -8.87314218, -8.49606348, -8.11181584, -7.720574, -7.32251267, -6.91780659, -6.50663046, -6.08915901, -5.66556696, -5.23602904, -4.80071996, -4.35981445, -3.91348722, -3.46191301, -3.00526652, -2.54372249, -2.07745563, -1.60664067, -1.13145232, -0.65206531, -0.168654359, 0.318605811, 0.809540479, 1.30397492, 1.80988319, 2.32284476, 2.83834122, 3.35593338, 3.87518208, 4.39564817, 4.91689247, 5.43847583, 5.95995907, 6.48090303, 7.00086855, 7.51941646, 8.03610761, 8.55050281, 9.06216292, 9.57064876, 10.0755212, 10.5983869, 11.1162829, 11.6282921, 12.1334976, 12.6309821, 13.1198287, 13.5991203, 14.0679397, 14.5253699, 14.9704938, 15.4023944, 15.8201546, 16.2228572, 16.6095853, 16.9794218, 17.3314495, 17.7052561, 18.1394568, 18.5495965, 18.9321325, 19.2835218, 19.6002216, 19.8786892, 20.1153816, 20.3067561, 20.4492698, 20.5393799, 20.5735435, 20.5482179, 20.4598601, 20.3049274, 20.0798769, 19.7811658, 18.6599114, 17.1009179, 15.4979481, 13.871647, 12.2426594, 10.6316302, 9.05920431, 7.54602657, 6.11274188, 4.77999511, 3.56843117, 2.49869492, 1.59143124, 0.867285033, 0.346901168, 0.050924529, 0, 2.27576178, 4.70602547, 7.28458627, 10.0052394, 12.8617801, 15.8480036, 18.957705, 22.1846796, 25.5227226, 28.9656292, 32.5071946, 36.1412141, 39.8614828, 43.661796, 47.5359489, 51.4777366, 55.1267553, 58.4262003, 62.1710315, 66.2791443, 70.6684342, 75.2567965, 79.9621267, 84.70232, 89.395272, 93.958878, 98.3110333, 102.369633, 106.052574, 109.443565, 112.233888, 114.291694, 115.485134, 117.012594, 118.077107, 117.526977, 114.831285, 109.459111, 100.879535, 88.5616365, 73.1899219, 52.6481343, 26.1675535, -7.02054072, -47.6848685, -90.8119062, -144.398179, -210.657968, -291.805554, -390.055219, -506.091115, -645.527809, -805.326988, -988.026055, -1207.54617, -1472.69741, -1792.28986, -2172.00485, -2624.49231, -3165.81531, -3807.81148, -4563.53042, -5432.45243, -6428.44, -7589.72534, -8943.35457, -10504.0639, -12295.7079, -14387.2777, -16822.2326, -19610.3299, -22833.9159, -26548.116, -30808.0557, -35668.8603, -43139.9265, -50833.5407, -58316.2607, -65154.6441, -70915.2487, -73538.902, -75301.7119, -76466.9774, -78472.0651, -81115.5391, -83706.7687, -86068.7777, -88024.5899, -89397.2291, -90009.7191, -87109.7623, -84533.4408, -81953.7023, -79043.4944, -75475.7648, -70923.461, -65059.5306, -52135.7619, -35373.2741},
        {-13.164245, -12.8630655, -12.5523028, -12.232134, -11.9027363, -11.5642866, -11.2169621, -10.8609398, -10.4963969, -10.1235103, -9.74245712, -9.35341448, -8.95655942, -8.55206901, -8.14012032, -7.72089043, -7.2945564, -6.86129532, -6.42128423, -5.97470023, -5.52172037, -5.06252174, -4.59728139, -4.12617641, -3.64938386, -3.16708081, -2.67944433, -2.1866515, -1.68887938, -1.18630505, -0.679105574, -0.167458027, 0.348460522, 0.868473003, 1.40241856, 1.94490133, 2.49040508, 3.03842769, 3.58846704, 4.14002102, 4.69258751, 5.2456644, 5.79874957, 6.3513409, 6.90293628, 7.45303359, 8.00113071, 8.54672553, 9.08931593, 9.6283998, 10.163475, 10.7139807, 11.2590411, 11.797722, 12.3290889, 12.8522075, 13.3661432, 13.8699618, 14.3627288, 14.8435099, 15.3113705, 15.7653764, 16.2045931, 16.6280862, 17.0349213, 17.424164, 17.7948799, 18.1884649, 18.6452983, 19.077143, 19.4803204, 19.8511521, 20.1859594, 20.4810639, 20.7327872, 20.9374506, 21.0913758, 21.1908842, 21.2322974, 21.2119367, 21.1261238, 20.9711801, 20.7434271, 20.4391864, 19.2929197, 17.697903, 16.0565694, 14.3899643, 12.7191328, 11.0651204, 9.44897233, 7.89173393, 6.4144505, 5.03816736, 3.78392983, 2.67278322, 1.72577285, 0.963944024, 0.408342065, 0.0800122856, 0, 1.93648875, 4.09777871, 6.47271321, 9.05013555, 11.818889, 14.767817, 17.8857627, 21.1615695, 24.5840808, 28.1421397, 31.8245896, 35.6202739, 39.5180358, 43.5067186, 47.5751657, 51.7122204, 55.6735996, 59.414843, 63.4521033, 67.7242683, 72.1702255, 76.7288626, 81.3390674, 85.9397276, 90.4697306, 94.8679644, 99.0733165, 103.024675, 106.660926, 110.21699, 113.227178, 115.541569, 117.010242, 118.647414, 119.721095, 119.188855, 116.551532, 111.309961, 102.96498, 91.0174255, 76.3356322, 56.6367427, 31.1540397, -0.879194431, -40.2296773, -81.3566237, -132.91113, -197.236788, -276.677192, -373.575936, -489.316221, -628.375842, -788.035869, -970.646512, -1188.90748, -1450.87176, -1764.59236, -2133.84942, -2571.93988, -3096.82197, -3720.51088, -4457.21893, -5312.13491, -6299.52883, -7449.05833, -8783.31787, -10305.5428, -12033.3087, -14055.1741, -16419.5456, -19185.7984, -22389.683, -26075.8091, -30288.7867, -35073.2255, -42162.8008, -49490.7905, -56679.5379, -63351.3864, -69128.6793, -72312.6351, -74727.472, -76561.7297, -78881.7742, -81528.6724, -83977.7301, -86088.3022, -87719.7437, -88731.4093, -88982.654, -85158.2218, -82279.247, -79805.2214, -77195.6367, -73909.9847, -69407.757, -63148.4455, -49974.792, -34551.113},
        {-14.1934407, -13.8927534, -13.5811675, -13.2588859, -12.9261114, -12.583047, -12.2298954, -11.8668596, -11.4941424, -11.1119466, -10.7204751, -10.3199309, -9.91051667, -9.49243538, -9.06588987, -8.631083, -8.18821764, -7.73749664, -7.27912288, -6.81329922, -6.34022851, -5.86011363, -5.37315744, -4.87956279, -4.37953256, -3.8732696, -3.36097679, -2.84285698, -2.31911304, -1.78994783, -1.25556421, -0.716165047, -0.171953207, 0.376868449, 0.941085945, 1.51477084, 2.0918626, 2.67180177, 3.25402887, 3.83798441, 4.42310894, 5.00884297, 5.59462703, 6.17990164, 6.76410734, 7.34668465, 7.92707409, 8.50471619, 9.07905148, 9.64952048, 10.2155637, 10.7948326, 11.3681621, 11.9345978, 12.4931855, 13.042971, 13.5829999, 14.1123182, 14.6299714, 15.1350054, 15.6264659, 16.1033987, 16.5648495, 17.009864, 17.437488, 17.8467673, 18.2367475, 18.6504564, 19.1298653, 19.5833109, 20.0069878, 20.3970906, 20.7498138, 21.0613521, 21.3279, 21.5456522, 21.7108032, 21.8195475, 21.8680799, 21.8525948, 21.7692869, 21.6143508, 21.383981, 21.0743722, 19.9041149, 18.2745615, 16.5964108, 14.891092, 13.1800344, 11.4846675, 9.82642051, 8.22672278, 6.70700364, 5.28869244, 3.9932185, 2.84201115, 1.85649973, 1.05811358, 0.468282015, 0.108434378, 0, 1.59915707, 3.49280284, 5.6648579, 8.09924278, 10.7798781, 13.6906843, 16.8155821, 20.1384919, 23.6433344, 27.31403, 31.1344995, 35.0886632, 39.1604418, 43.3337559, 47.592526, 51.9206726, 56.1857956, 60.3551983, 64.6759078, 69.1069185, 73.6072254, 78.135823, 82.651706, 87.1138692, 91.4813072, 95.7130147, 99.7679864, 103.605217, 107.183701, 110.878121, 114.079365, 116.621722, 118.33948, 120.053265, 121.104197, 120.570373, 117.990179, 112.902002, 104.84423, 93.3552502, 79.6234925, 61.0351138, 36.8056667, 6.15070389, -31.7142222, -71.2659447, -121.173429, -183.798025, -261.501086, -356.64396, -470.686101, -607.993079, -766.88759, -949.816703, -1167.64809, -1427.46874, -1736.36566, -2096.366, -2520.76655, -3028.38389, -3629.16428, -4341.41569, -5178.65253, -6156.71614, -7295.3907, -8612.65057, -10100.2565, -11769.3866, -13727.3359, -16027.0301, -18764.2597, -21942.4915, -25599.8136, -29774.3142, -34504.0815, -41326.6549, -48405.8085, -55404.7675, -61986.7573, -67815.0031, -71474.2234, -74425.155, -76793.2407, -79311.174, -81864.0274, -84083.8519, -85868.3711, -87115.3083, -87722.387, -87587.3307, -82835.7726, -79691.4567, -77384.2212, -75143.9044, -72200.3444, -67783.3795, -61122.848, -48002.9631, -33859.5826},
        {-15.3459475, -15.0636219, -14.7668957, -14.4560836, -14.1315004, -13.7934609, -13.4422799, -13.0782722, -12.7017526, -12.3130358, -11.9124367, -11.5002701, -11.0768507, -10.6424933, -10.1975128, -9.74222385, -9.27694133, -8.80197999, -8.31765464, -7.82428007, -7.32217106, -6.81164241, -6.2930089, -5.76658533, -5.23268648, -4.69162715, -4.14372213, -3.5892862, -3.02863417, -2.46208081, -1.88994092, -1.31252929, -0.730160708, -0.143149968, 0.456643477, 1.06449932, 1.67559564, 2.28934324, 2.90515294, 3.52243557, 4.14060194, 4.75906286, 5.37722916, 5.99451165, 6.61032116, 7.22406849, 7.83516447, 8.44301991, 9.04704564, 9.64665246, 10.2412512, 10.8473611, 11.4469137, 12.0389489, 12.6225067, 13.1966269, 13.7603496, 14.3127147, 14.8527621, 15.3795317, 15.8920635, 16.3893975, 16.8705736, 17.3346317, 17.7806117, 18.2075537, 18.6144975, 19.046121, 19.5460063, 20.0190686, 20.4613893, 20.8690497, 21.2381315, 21.5647158, 21.8448842, 22.0747182, 22.250299, 22.3677081, 22.423027, 22.412337, 22.3317196, 22.1772563, 21.9450283, 21.6311172, 20.4378392, 18.7748542, 17.0613622, 15.319204, 13.5702202, 11.8362513, 10.139138, 8.50072089, 6.94284065, 5.48733789, 4.15605322, 2.97082726, 1.95350064, 1.12591398, 0.509907904, 0.127323036, 0, 1.69446226, 3.67306093, 5.92033022, 8.42080436, 11.1590176, 14.119504, 17.286798, 20.6454336, 24.1799452, 27.8748669, 31.7147329, 35.6840775, 39.7674348, 43.9493392, 48.2143247, 52.5469257, 56.7771915, 60.8675872, 65.1520768, 69.5820905, 74.1090587, 78.6844118, 83.25958, 87.7859938, 92.2150835, 96.4982795, 100.587012, 104.432711, 107.986808, 111.522694, 114.551786, 116.928925, 118.508952, 120.135915, 121.170053, 120.707817, 118.307287, 113.526542, 105.923663, 95.0567297, 82.4278266, 65.0593756, 42.1291081, 12.8147555, -23.7059511, -63.1767563, -112.768084, -174.571833, -250.679904, -343.184196, -452.134743, -584.160924, -739.634636, -921.66666, -1138.95967, -1397.7057, -1704.09676, -2060.86847, -2478.45482, -2971.0605, -3539.15838, -4213.04645, -5017.66647, -5973.45086, -7093.19106, -8393.18585, -9864.46801, -11516.7859, -13452.5297, -15721.6864, -18371.3542, -21454.8527, -25023.1686, -29127.2883, -33818.1987, -40768.5443, -48002.2392, -55164.8554, -61901.9649, -67859.1398, -71577.7724, -74543.7344, -76875.8177, -79217.3862, -81479.5887, -83355.0025, -84765.7048, -85633.7732, -85881.285, -85430.3176, -79946.9138, -76626.2832, -74501.4229, -72605.3303, -69971.0025, -65631.437, -58619.6309, -46747.0206, -33384.7818},
        {-16.5634187, -16.3047364, -16.0276075, -15.7324779, -15.4197936, -15.0900006, -14.7435451, -14.3808729, -14.0024301, -13.6086627, -13.2000168, -12.7769383, -12.3398733, -11.8892677, -11.4255677, -10.9492191, -10.4606681, -9.96036064, -9.44874273, -8.92626039, -8.39335963, -7.85048648, -7.29808696, -6.73660706, -6.16649282, -5.58819025, -5.00214536, -4.40880417, -3.80861269, -3.20201695, -2.58946295, -1.97139671, -1.34826424, -0.720511576, -0.0832693069, 0.560247039, 1.20675805, 1.85564522, 2.50629004, 3.15807399, 3.81037857, 4.46258526, 5.11407555, 5.76423094, 6.4124329, 7.05806293, 7.70050252, 8.33913316, 8.97333634, 9.60249354, 10.2259863, 10.8588463, 11.4844656, 12.1018863, 12.7101507, 13.3083009, 13.8953793, 14.470428, 15.0324893, 15.5806053, 16.1138183, 16.6311705, 17.1317042, 17.6144615, 18.0784847, 18.5228159, 18.9464975, 19.3959746, 19.916554, 20.4095132, 20.8708214, 21.296448, 21.682362, 22.024533, 22.31893, 22.5615225, 22.7482797, 22.8751709, 22.9381654, 22.9332324, 22.8563413, 22.7034612, 22.4705616, 22.1536117, 20.9386405, 19.2436235, 17.4962709, 15.7188361, 13.9335724, 12.1627334, 10.4285724, 8.75334275, 7.15929792, 5.66869129, 4.30377625, 3.08680621, 2.04003457, 1.18571473, 0.54610009, 0.143444046, 0, 1.87339308, 4.00544746, 6.38231851, 8.99016159, 11.815132, 14.8433852, 18.0610765, 21.4543612, 25.0093947, 28.7123324, 32.5493296, 36.5065417, 40.5701239, 44.7262318, 48.9610206, 53.2606458, 57.3987452, 61.3311147, 65.5246035, 69.9198277, 74.4574031, 79.0779457, 83.7220714, 88.3303961, 92.8435358, 97.2021063, 101.346724, 105.218004, 108.756562, 112.109935, 114.935948, 117.113141, 118.520054, 120.035157, 121.037026, 120.637586, 118.4154, 113.949028, 106.817031, 96.5979703, 85.1524315, 69.082422, 47.5200198, 19.5973031, -15.5536502, -55.3368972, -104.950191, -166.127419, -240.60247, -330.109233, -432.896651, -558.442932, -709.31266, -889.397174, -1105.2966, -1362.39488, -1666.07595, -2020.27091, -2432.14672, -2910.45526, -3445.51892, -4079.97752, -4849.23466, -5778.42822, -6875.30133, -8155.5816, -9612.49956, -11255.9502, -13178.3179, -15425.3263, -17975.9457, -20952.923, -24425.0887, -28461.2736, -33130.3084, -40344.4605, -47868.2645, -55309.6918, -62276.7139, -68377.3021, -71994.9912, -74778.4804, -76860.4999, -78884.0072, -80779.7674, -82265.9506, -83283.9515, -83775.1647, -83680.985, -82942.8071, -76976.6128, -73554.0612, -71574.4019, -69936.8843, -67540.7581, -63285.2729, -56069.6784, -45718.2239, -33066.3557},
        {-17.8259259, -17.5925257, -17.3365732, -17.0586454, -16.759319, -16.4391711, -16.0987783, -15.7387177, -15.3595661, -14.9619002, -14.5462971, -14.1133336, -13.6635865, -13.1976327, -12.7160491, -12.2194126, -11.7082999, -11.1832881, -10.6449539, -10.0938742, -9.53062593, -8.95578589, -8.36993098, -7.77363806, -7.167484, -6.55204567, -5.92789993, -5.29562366, -4.65579372, -4.00898697, -3.3557803, -2.69675056, -2.03247463, -1.36352936, -0.688029503, -0.00783477205, 0.675165515, 1.36031459, 2.04695568, 2.73443202, 3.42208684, 4.10926338, 4.79530485, 5.4795545, 6.16135555, 6.84005124, 7.51498479, 8.18549945, 8.85093843, 9.51064497, 10.1639623, 10.8235372, 11.4751209, 12.1177684, 12.7505344, 13.3724737, 13.9826413, 14.5800918, 15.1638803, 15.7330615, 16.2866902, 16.8238213, 17.3435096, 17.84481, 18.3267773, 18.7884663, 19.2289318, 19.6966369, 20.238858, 20.7527496, 21.2341636, 21.678952, 22.0829666, 22.4420594, 22.7520823, 23.0088871, 23.2083259, 23.3462504, 23.4185127, 23.4209647, 23.3494581, 23.1998451, 22.9679774, 22.649707, 21.4145786, 19.6889779, 17.9092223, 16.097979, 14.2779152, 12.4716981, 10.7019948, 8.99147243, 7.36279828, 5.83863947, 4.44166318, 3.19453658, 2.11992684, 1.24050114, 0.578926653, 0.157870548, 0, 2.07085284, 4.37212575, 6.89172826, 9.61756993, 12.5375603, 15.6396089, 18.9116252, 22.3415189, 25.9171994, 29.6265763, 33.4575591, 37.3980574, 41.4359807, 45.5592386, 49.7557406, 54.0133962, 58.0607278, 61.8485896, 65.9516003, 70.3020864, 74.8323745, 79.4747911, 84.1616629, 88.8253164, 93.3980782, 97.8122749, 102.000233, 105.894279, 109.42674, 112.648776, 115.330308, 117.368011, 118.658562, 120.07731, 121.031595, 120.667776, 118.588929, 114.398127, 107.698445, 98.0929558, 87.8262423, 73.0559338, 52.8682878, 26.3495618, -7.41398677, -47.5923896, -97.2790276, -157.823571, -230.57569, -316.885054, -413.189291, -531.753722, -677.033555, -853.634361, -1066.4687, -1320.31129, -1619.93688, -1970.54603, -2376.39084, -2841.25882, -3347.04436, -3946.09143, -4680.89011, -5580.22662, -6649.66667, -7905.43456, -9344.81483, -10979.1215, -12889.1142, -15116.2705, -17567.5721, -20439.6834, -23820.638, -27798.4692, -32461.2107, -40012.7388, -47896.2832, -55670.9168, -62895.7122, -69129.7421, -72528.1344, -74989.87, -76675.7122, -78293.6895, -79787.9394, -80874.5727, -81509.1284, -81647.1457, -81244.1635, -80255.721, -74162.0581, -70721.5058, -68799.7598, -67262.5162, -64975.4708, -60804.3197, -53614.7587, -44797.6013, -32897.2775},
        {-19.0363347, -18.8107262, -18.560778, -18.287103, -17.9903142, -17.6710246, -17.3298471, -16.9673947, -16.5842804, -16.1811171, -15.7585178, -15.3170955, -14.8574632, -14.3802338, -13.8860203, -13.3754356, -12.8490928, -12.3076049, -11.7515846, -11.1816452, -10.5983994, -10.0024604, -9.39444096, -8.77495419, -8.14461302, -7.50403041, -6.85381935, -6.1945928, -5.52696373, -4.85154512, -4.16894992, -3.47979112, -2.78468168, -2.08423457, -1.37492634, -0.659525184, 0.0591510052, 0.780355021, 1.50333966, 2.22735771, 2.95166197, 3.67550524, 4.39814031, 5.11881997, 5.83679703, 6.55132427, 7.26165449, 7.96704048, 8.66673505, 9.35999097, 10.0460611, 10.731787, 11.4086681, 12.0757927, 12.7322492, 13.3771257, 14.0095107, 14.6284922, 15.2331587, 15.8225985, 16.3958997, 16.9521507, 17.4904398, 18.0098553, 18.5094853, 18.9884183, 19.4457426, 19.9326183, 20.4989558, 21.0364485, 21.5408091, 22.0077503, 22.4329847, 22.8122249, 23.1411837, 23.4155737, 23.6311075, 23.7834979, 23.8684574, 23.8816988, 23.8189347, 23.6758778, 23.4482407, 23.1317362, 21.8791935, 20.1254959, 18.3156574, 16.4727439, 14.6198212, 12.7799552, 10.9762117, 9.23165652, 7.56935554, 6.01237459, 4.58377953, 3.30663619, 2.20401043, 1.29896808, 0.614574997, 0.173897022, 0, 2.07478023, 4.38919545, 6.93015077, 9.6845513, 12.6393021, 15.7813084, 19.0974753, 22.5747078, 26.199911, 29.9599902, 33.8418503, 37.8323966, 41.918534, 46.0871678, 50.3252031, 54.6195449, 58.7703234, 62.7377616, 66.9312317, 71.2976158, 75.7837956, 80.3366531, 84.9030702, 89.4299288, 93.8641108, 98.1524981, 102.241973, 106.079416, 109.611711, 113.018711, 115.928903, 118.219277, 119.766822, 121.205829, 122.034637, 121.549641, 119.40064, 115.237434, 108.709824, 99.4676097, 90.2195519, 76.625502, 57.7367676, 32.6046565, 0.280476626, -39.6420589, -88.7896491, -148.246588, -219.097168, -302.425684, -393.882145, -506.627032, -645.24995, -814.444988, -1019.11956, -1264.08529, -1554.15383, -1894.6695, -2289.79017, -2743.09257, -3237.98628, -3823.31803, -4536.11884, -5403.59586, -6436.31037, -7652.46439, -9052.67409, -10650.5822, -12524.3128, -14715.5157, -17098.2934, -19907.6194, -23243.1407, -27204.504, -31891.3564, -39742.3272, -47933.3352, -55979.2816, -63395.0677, -69695.5947, -72812.161, -74899.0061, -76149.718, -77379.1091, -78534.8663, -79305.8333, -79652.6392, -79535.9134, -78916.2849, -77754.3829, -72382.7605, -69064.0061, -66898.9038, -64988.2378, -62432.7921, -58333.3509, -51790.6985, -43729.8548, -33030.9981},
        {-20.2695359, -20.0496049, -19.8040241, -19.5334199, -19.2384188, -18.9196472, -18.5777313, -18.2132977, -17.8269728, -17.4193828, -16.9911544, -16.5429137, -16.0752873, -15.5889015, -15.0843827, -14.5623574, -14.0234519, -13.4682926, -12.897506, -12.3117184, -11.7115562, -11.0976458, -10.4706137, -9.83108622, -9.17968975, -8.51705071, -7.8437955, -7.16055051, -6.46794214, -5.7665968, -5.05714089, -4.3402008, -3.61640293, -2.88637368, -2.1436198, -1.39247726, -0.637367624, 0.120851669, 0.881323172, 1.64318944, 2.40559302, 3.16767647, 3.92858235, 4.68745321, 5.4434316, 6.19566007, 6.94328119, 7.6854375, 8.42127155, 9.14992591, 9.87054313, 10.5826291, 11.2849551, 11.976656, 12.6568662, 13.3247206, 13.9793538, 14.6199004, 15.2454952, 15.8552729, 16.448368, 17.0239153, 17.5810495, 18.1189052, 18.6366171, 19.13332, 19.6081483, 20.1154103, 20.7080892, 21.2715288, 21.8012871, 22.2929222, 22.741992, 23.1440546, 23.4946681, 23.7893903, 24.0237793, 24.1933932, 24.2937899, 24.3205275, 24.2691639, 24.1352573, 23.9143655, 23.6020467, 22.3338148, 20.5533398, 18.7146286, 16.8411492, 14.9563698, 13.0837584, 11.2467832, 9.46891223, 7.77361358, 6.18435537, 4.72460569, 3.41783263, 2.28750429, 1.35708876, 0.650054137, 0.18986852, 0, 2.00056999, 4.26558015, 6.77977549, 9.52790104, 12.4947018, 15.6649228, 19.023309, 22.5546055, 26.2435573, 30.0749093, 34.0334067, 38.1037944, 42.2708174, 46.5192207, 50.8337495, 55.1991486, 59.5415944, 63.8374651, 68.2039822, 72.6133405, 77.0377345, 81.4493588, 85.8204079, 90.1230766, 94.3295593, 98.4120506, 102.342745, 106.093838, 109.637523, 113.387641, 116.712804, 119.452712, 121.447066, 122.989107, 123.691764, 123.047025, 120.758528, 116.529913, 110.064816, 101.066878, 92.6789411, 80.1187252, 62.4169805, 38.6044578, 7.7119076, -31.5127044, -79.6851463, -137.703971, -206.467732, -286.874982, -374.39192, -481.988997, -613.427142, -773.142123, -966.947506, -1200.03826, -1477.60934, -1804.57604, -2186.47657, -2629.1542, -3122.72729, -3705.65671, -4402.86158, -5237.25381, -6228.3441, -7397.20437, -8747.84406, -10296.9115, -12123.6174, -14271.7836, -16603.6751, -19372.6025, -22685.2346, -26648.2398, -31368.2867, -39478.5273, -47927.5259, -56190.3301, -63741.9878, -70057.5466, -72854.7869, -74537.5353, -75333.9539, -76184.2955, -77034.8532, -77535.2156, -77646.511, -77329.8675, -76546.4135, -75257.2772, -71087.0542, -67982.4963, -65423.907, -62891.5902, -59865.8495, -55826.9887, -50255.3114, -42563.4388, -33263.4036},
        {-21.5849875, -21.3809001, -21.1488683, -20.8895664, -20.6036686, -20.2918494, -19.954783, -19.5931436, -19.2076057, -18.7988435, -18.3675312, -17.9143433, -17.4399539, -16.9450374, -16.4302682, -15.8963204, -15.3438684, -14.7735865, -14.186149, -13.5822301, -12.9625043, -12.3276457, -11.6783288, -11.0152277, -10.3390168, -9.65037039, -8.94996276, -8.23846822, -7.51656105, -6.78491557, -6.04420607, -5.29510684, -4.5382922, -3.77443644, -2.99535141, -2.20632904, -1.41276779, -0.615629553, 0.18412376, 0.985530249, 1.78762801, 2.58945515, 3.39004975, 4.18844992, 4.98369375, 5.77481934, 6.56086479, 7.34086819, 8.11386764, 8.87890125, 9.6350071, 10.3743184, 11.1029278, 11.8200231, 12.5247921, 13.2164226, 13.8941023, 14.5570191, 15.2043607, 15.8353149, 16.4490694, 17.0448122, 17.6217309, 18.1790133, 18.7158473, 19.2314205, 19.7249208, 20.2540836, 20.8752386, 21.4668234, 22.0242306, 22.5428523, 23.0180811, 23.4453093, 23.8199293, 24.1373334, 24.3929141, 24.5820638, 24.7001747, 24.7426393, 24.7048499, 24.582199, 24.370079, 24.0638821, 22.7808678, 20.9739653, 19.1066584, 17.2028364, 15.286389, 13.3812054, 11.5111753, 9.70018819, 7.97213343, 6.35090057, 4.86037908, 3.52445844, 2.36702815, 1.41197767, 0.683196499, 0.204574114, 0, 1.93424018, 4.15702013, 6.65096883, 9.39871525, 12.3828884, 15.5861171, 18.9910305, 22.5802576, 26.3364271, 30.2421683, 34.2801099, 38.4328811, 42.6831107, 47.0134277, 51.4064612, 55.84484, 60.3879861, 65.0294989, 69.578309, 74.0335011, 78.3941597, 82.6593695, 86.828215, 90.8997808, 94.8731516, 98.7474119, 102.521646, 106.19494, 109.766376, 113.897593, 117.682186, 120.920475, 123.412779, 125.118896, 125.759405, 125.012358, 122.630231, 118.365497, 111.970633, 103.198114, 95.4992909, 83.802018, 67.1350779, 44.5272529, 15.0073252, -23.2017023, -70.0638232, -126.34881, -192.826437, -270.266475, -354.265652, -457.089399, -581.298432, -730.927025, -913.017977, -1133.26332, -1397.35508, -1709.6024, -2077.3945, -2509.62923, -3004.74116, -3588.7154, -4272.24794, -5073.15083, -6021.29381, -7141.01196, -8440.2518, -9939.09952, -11717.7326, -13822.3376, -16111.7911, -18848.4063, -22143.2076, -26107.2195, -30851.4663, -39168.7618, -47817.8939, -56239.4398, -63873.9764, -70162.0811, -72630.2873, -73909.2448, -74259.8352, -74743.2886, -75303.6818, -75551.6134, -75447.8342, -74953.0951, -74028.147, -72633.7407, -69851.6668, -67015.5686, -64031.7453, -60806.4956, -57246.1188, -53256.9136, -48745.1791, -41319.5574, -33422.4579},
        {-23.1223766, -22.9662305, -22.7760117, -22.5525923, -22.2968443, -22.0096397, -21.6918504, -21.3443485, -20.9680059, -20.5636946, -20.1322866, -19.674654, -19.1916686, -18.6842025, -18.1531276, -17.5993161, -17.0236397, -16.4269707, -15.8101808, -15.1741422, -14.5197267, -13.8478065, -13.1592534, -12.4549395, -11.7357368, -11.0025173, -10.2561528, -9.49751554, -8.72747737, -7.9469103, -7.15668634, -6.35767746, -5.55075566, -4.73679293, -3.91269261, -3.08139447, -2.24588603, -1.40716804, -0.566241309, 0.2758934, 1.1182353, 1.95978362, 2.79953757, 3.63649638, 4.46965926, 5.29802544, 6.12059413, 6.93636455, 7.74433593, 8.54350749, 9.33287844, 10.1001286, 10.855822, 11.5992033, 12.3295169, 13.0460076, 13.7479199, 14.4344984, 15.1049877, 15.7586325, 16.3946774, 17.0123668, 17.6109455, 18.1896581, 18.7477491, 19.2844631, 19.7990448, 20.3531555, 21.0069581, 21.6309671, 22.2203811, 22.7703982, 23.276217, 23.7330358, 24.1360529, 24.4804668, 24.7614757, 24.9742782, 25.1140725, 25.1760571, 25.1554302, 25.0473904, 24.8471359, 24.5498652, 23.2510276, 21.4146211, 19.5154578, 17.5779346, 15.6264485, 13.6853964, 11.7791753, 9.93218215, 8.16881393, 6.51346759, 4.99054007, 3.62442834, 2.43952935, 1.46024007, 0.710957441, 0.216078433, 0, 2.01112652, 4.31044123, 6.88016143, 9.70250442, 12.7596875, 16.033928, 19.5074431, 23.1624502, 26.9811666, 30.9458096, 35.0385965, 39.2417445, 43.537471, 47.9079932, 52.3355285, 56.8022942, 61.4528034, 66.2924587, 70.9363794, 75.4015604, 79.7049968, 83.8636835, 87.8946156, 91.814788, 95.6411958, 99.3908339, 103.080697, 106.727781, 110.349081, 114.756193, 118.880156, 122.499585, 125.393093, 127.364598, 128.180063, 127.598702, 125.391538, 121.329595, 115.183894, 106.72546, 99.2965721, 88.010092, 71.9383184, 50.1535501, 21.7280859, -15.0811275, -60.1826012, -114.300198, -178.157782, -252.479216, -333.18175, -431.670609, -549.447511, -690.171109, -861.903809, -1070.73082, -1322.73736, -1622.07304, -1977.19857, -2398.68623, -2890.48536, -3469.89438, -4136.50378, -4904.58823, -5813.30391, -6890.38596, -8149.89703, -9616.02754, -11363.1005, -13437.0811, -15688.9582, -18389.6697, -21652.6726, -25591.4238, -30319.3803, -38722.6168, -47448.8182, -55918.2871, -63551.3261, -69768.2377, -71984.3374, -72961.5732, -72979.5277, -73150.7388, -73425.078, -73405.3521, -73058.7859, -72352.6041, -71254.0314, -69730.2924, -68284.8886, -65765.9039, -62455.2887, -58634.9936, -54586.9691, -50593.1656, -46935.5336, -40009.2087, -33252.2891},
        {-24.9710939, -24.9099733, -24.8029194, -24.6512278, -24.4561942, -24.2191141, -23.9412832, -23.6239971, -23.2685514, -22.8762417, -22.4483636, -21.9862128, -21.4910848, -20.9642754, -20.40708, -19.8207943, -19.206714, -18.5661346, -17.9003518, -17.2106612, -16.4983584, -15.764739, -15.0110986, -14.2387328, -13.4489374, -12.6430078, -11.8222397, -10.9879288, -10.1413705, -9.28386064, -8.41669472, -7.54116838, -6.65857725, -5.77021692, -4.88846843, -4.00885151, -3.12675269, -2.24310783, -1.35885282, -0.474923527, 0.407744177, 1.28821442, 2.16555132, 3.038819, 3.9070816, 4.76940324, 5.62484804, 6.47248013, 7.31136364, 8.14056269, 8.95914141, 9.75520241, 10.539009, 11.3098628, 12.0670657, 12.8099194, 13.5377256, 14.2497861, 14.9454026, 15.623877, 16.2845109, 16.926606, 17.5494643, 18.1523873, 18.7346768, 19.2956347, 19.8345625, 20.417618, 21.1093536, 21.7711858, 22.3980823, 22.985011, 23.5269397, 24.0188359, 24.4556677, 24.8324025, 25.1440084, 25.3854529, 25.5517039, 25.6377291, 25.6384962, 25.5489731, 25.3641275, 25.0789271, 23.7619731, 21.8909014, 19.9544822, 17.9777457, 15.9857218, 14.0034409, 12.055933, 10.1682282, 8.36535677, 6.67234874, 5.11423428, 3.71604352, 2.5028066, 1.49955365, 0.731314807, 0.223120213, 0, 2.32128781, 4.8898724, 7.6904082, 10.7075496, 13.9259511, 17.3302671, 20.905152, 24.6352602, 28.5052462, 32.4997643, 36.6034691, 40.8010149, 45.0770561, 49.4162472, 53.8032426, 58.2226968, 62.8155286, 67.5887161, 72.1734581, 76.5878943, 80.8501645, 84.9784082, 88.9907652, 92.9053751, 96.7403777, 100.513913, 104.24412, 107.949138, 111.647108, 116.120374, 120.342662, 124.102851, 127.189818, 129.546774, 130.88543, 130.876346, 129.262099, 125.78527, 120.188438, 112.214181, 104.492726, 93.0001565, 76.9140757, 55.412089, 27.6718009, -7.31894419, -50.1453413, -101.582346, -162.404913, -233.387998, -310.848597, -405.468241, -518.236518, -652.593983, -816.985261, -1017.60825, -1260.66085, -1550.97875, -1896.43165, -2306.37531, -2784.30506, -3346.88046, -3989.62201, -4726.54162, -5602.98381, -6649.91094, -7890.53547, -9353.81053, -11097.1054, -13161.7156, -15377.4207, -18029.573, -21232.6851, -25101.2695, -29749.839, -38069.6757, -46704.33, -55074.1222, -62599.3723, -68700.4005, -70801.2335, -71649.3472, -71520.616, -71464.5042, -71450.6736, -71126.457, -70477.6329, -69489.9796, -68149.2755, -66441.2988, -66072.2966, -63904.2842, -60466.1001, -56286.5825, -51894.5699, -47818.9005, -44588.4129, -38628.6993, -32561.0987},
        {-26.8909579, -26.9414272, -26.9326961, -26.8665292, -26.7446913, -26.5689469, -26.3410607, -26.0627973, -25.7359214, -25.3621977, -24.9433908, -24.4812652, -23.9775858, -23.4341171, -22.8526237, -22.2348704, -21.5826217, -20.8976423, -20.1816968, -19.43655, -18.6639663, -17.8657106, -17.0435474, -16.1992414, -15.3345572, -14.4512594, -13.5511128, -12.6358819, -11.7073314, -10.767226, -9.81733027, -8.85940888, -7.89522648, -6.9265477, -5.98269127, -5.05106522, -4.11874644, -3.18660538, -2.25551252, -1.32633831, -0.399953221, 0.522772289, 1.44096776, 2.35376271, 3.2602867, 4.15966925, 5.05103991, 5.9335282, 6.80626367, 7.66837585, 8.51899429, 9.34692732, 10.1618494, 10.9631139, 11.7500739, 12.5220829, 13.2784941, 14.0186607, 14.7419361, 15.4476734, 16.1352261, 16.8039474, 17.4531905, 18.0823088, 18.6906555, 19.2775839, 19.8424473, 20.4555756, 21.1855947, 21.8854739, 22.5499651, 23.1738207, 23.7517926, 24.2786332, 24.7490944, 25.1579285, 25.4998877, 25.769724, 25.9621896, 26.0720367, 26.0940174, 26.0228839, 25.8533883, 25.5802827, 24.2462222, 22.3420335, 20.3699869, 18.355732, 16.3249184, 14.3031955, 12.3162129, 10.3896201, 8.54906665, 6.8202021, 5.22867597, 3.8001378, 2.56023712, 1.53462347, 0.748946373, 0.228855373, 0, 2.65810214, 5.51615403, 8.56186009, 11.7829248, 15.1670525, 18.7019477, 22.3753148, 26.1748582, 30.0882824, 34.1032917, 38.2075907, 42.3888838, 46.6348753, 50.9332698, 55.2717716, 59.6380851, 64.1374412, 68.7743335, 73.2706335, 77.6392297, 81.8930107, 86.0448653, 90.107682, 94.0943494, 98.0177561, 101.890791, 105.726342, 109.537299, 113.336549, 117.814892, 122.058738, 125.877605, 129.081008, 131.821216, 133.736369, 134.373203, 133.438409, 130.638676, 125.680693, 118.27115, 110.238303, 98.5215806, 82.4125827, 61.2029095, 34.1841611, 1.25433441, -39.0529661, -87.5977394, -145.239984, -212.8397, -287.170985, -378.173617, -486.575834, -615.64427, -773.828111, -967.249682, -1202.03131, -1483.46614, -1818.69413, -2215.75978, -2676.70415, -3219.23801, -3837.17441, -4544.95755, -5391.99005, -6411.62805, -7635.00846, -9094.91207, -10832.2567, -12883.3575, -15061.6871, -17660.6863, -20793.8081, -24574.5053, -29116.2307, -37270.5574, -45728.2879, -53918.345, -61269.6514, -67211.1301, -69218.3361, -69975.8049, -69749.6167, -69501.357, -69221.802, -68611.934, -67677.019, -66422.3228, -64853.1114, -62974.6507, -63405.893, -61547.733, -58089.2781, -53719.6356, -49127.9133, -45003.2184, -42034.6584, -37078.0679, -31576.9333},
        {-28.8054375, -28.9761373, -29.0741983, -29.1018465, -29.0613078, -28.9548084, -28.7845741, -28.552831, -28.2618051, -27.9137223, -27.5108087, -27.0552903, -26.5493931, -25.995343, -25.3953661, -24.7516883, -24.0665357, -23.3421342, -22.5807099, -21.7844888, -20.9556967, -20.0965599, -19.2093041, -18.2961555, -17.3593401, -16.4010838, -15.4236126, -14.4291525, -13.4199295, -12.3981697, -11.366099, -10.3259434, -9.27992896, -8.23028161, -7.22201143, -6.23505249, -5.24882461, -4.26416537, -3.28191234, -2.30290308, -1.32797518, -0.357966203, 0.606286277, 1.56394469, 2.51417146, 3.45612901, 4.38897978, 5.3118862, 6.22401068, 7.12451566, 8.01256356, 8.87661064, 9.72675759, 10.5623989, 11.3829292, 12.1877428, 12.9762344, 13.7477985, 14.5018295, 15.237722, 15.9548705, 16.6526695, 17.3305136, 17.9877972, 18.6239149, 19.2382612, 19.8302306, 20.4735871, 21.2403092, 21.9764099, 22.6764626, 23.3350407, 23.9467177, 24.506067, 25.0076621, 25.4460765, 25.8158834, 26.1116565, 26.3279691, 26.4593947, 26.5005068, 26.4458787, 26.2900839, 26.0276959, 24.6785798, 22.7451648, 20.7415956, 18.6940766, 16.6288125, 14.5720079, 12.5498674, 10.5885956, 8.71439713, 6.95347652, 5.33203842, 3.87628742, 2.61242812, 1.56666513, 0.765203041, 0.234246465, 0, 2.96226533, 6.07984847, 9.34355934, 12.7442078, 16.2726039, 19.9195575, 23.6758784, 27.5323767, 31.4798622, 35.5091449, 39.6110346, 43.7763413, 47.995875, 52.2604455, 56.5608627, 60.8879366, 65.2874804, 69.7581615, 74.1663778, 78.5147256, 82.8058012, 87.0422009, 91.2265211, 95.3613582, 99.4493083, 103.492968, 107.494934, 111.457801, 115.384167, 119.829446, 124.040716, 127.854728, 131.108234, 134.177257, 136.628911, 137.886507, 137.625012, 135.519397, 131.244632, 124.475684, 116.315165, 104.575903, 88.6535484, 67.9437491, 41.8421538, 11.1440797, -26.5044105, -72.0575852, -126.469713, -190.695062, -261.956181, -349.500051, -454.096715, -578.891361, -731.878439, -918.875178, -1145.69881, -1417.52863, -1740.96479, -2123.30334, -2565.05947, -3085.59911, -3678.98321, -4360.33493, -5180.4661, -6173.80689, -7376.47348, -8823.33313, -10543.071, -12567.8692, -14704.5903, -17245.4005, -20300.4772, -23979.9978, -28394.1396, -36320.8148, -44535.5323, -52481.5356, -59602.0686, -65340.3747, -67247.3459, -67920.1456, -67613.0253, -67197.7815, -66684.2209, -65824.84, -64642.3109, -63159.3059, -61398.4974, -59382.5575, -60226.2132, -58591.6911, -55224.0272, -50868.2573, -46269.4173, -42172.5432, -39322.6707, -35305.1553, -30319.3642},
        {-30.6128386, -30.9049869, -31.1122232, -31.2371676, -31.2824404, -31.2506616, -31.1444514, -30.96643, -30.7192176, -30.4054343, -30.0277003, -29.5886358, -29.0908608, -28.5369956, -27.9296604, -27.2714752, -26.5650603, -25.8130358, -25.0180218, -24.1826386, -23.3095063, -22.4012451, -21.4604751, -20.4898164, -19.4918893, -18.4693139, -17.4247104, -16.3606989, -15.2798996, -14.1849327, -13.0784183, -11.9629765, -10.8412276, -9.71579167, -8.64296272, -7.59739503, -6.5530978, -5.51094937, -4.47182808, -3.43661227, -2.40618029, -1.38141049, -0.363181196, 0.647629236, 1.65014246, 2.64348015, 3.62676394, 4.5991155, 5.55965648, 6.50750854, 7.44179334, 8.34790492, 9.23899013, 10.1144682, 10.9737584, 11.81628, 12.6414522, 13.4486943, 14.2374254, 15.007065, 15.7570321, 16.486746, 17.195626, 17.8830913, 18.5485613, 19.191455, 19.8111918, 20.4836844, 21.2830093, 22.0508233, 22.7815868, 23.4697604, 24.1098045, 24.6961796, 25.2233462, 25.6857647, 26.0778956, 26.3941995, 26.6291367, 26.7771677, 26.8327531, 26.7903533, 26.6444288, 26.38944, 25.0285426, 23.0727004, 21.0447916, 18.9714464, 16.8792951, 14.7949679, 12.7450949, 10.7563065, 8.85523279, 7.06850404, 5.42275047, 3.94460229, 2.66068972, 1.59764299, 0.782092312, 0.240667909, 0, 3.15175611, 6.4296391, 9.82685826, 13.3366228, 16.9521421, 20.6666254, 24.4732819, 28.3653209, 32.3359516, 36.3783834, 40.4858256, 44.6514873, 48.8685779, 53.1303066, 57.4298827, 61.7605154, 66.0932328, 70.4180753, 74.7784242, 79.1627356, 83.5594658, 87.9570708, 92.3440068, 96.7087298, 101.039696, 105.325362, 109.554183, 113.714615, 117.795115, 122.189678, 126.332522, 130.090445, 133.330239, 136.60197, 139.430797, 141.159675, 141.450417, 139.964839, 136.364753, 130.311974, 122.469839, 111.193827, 95.9498009, 76.2036242, 51.4211606, 23.030515, -11.955251, -54.5608342, -105.810931, -166.73024, -234.874159, -318.976934, -420.292555, -541.927075, -690.767843, -872.004486, -1090.82663, -1351.16854, -1659.7605, -2024.7023, -2446.18069, -2944.40983, -3515.18034, -4173.97504, -4969.64415, -5935.7217, -7108.09276, -8520.90088, -10199.6692, -12174.797, -14262.8375, -16740.9356, -19713.5143, -23284.9966, -27559.8055, -35220.7585, -43149.2856, -50805.2111, -57648.3593, -63138.5546, -64902.4586, -65455.167, -65042.1454, -64472.2662, -63767.9799, -62716.9838, -61353.4658, -59711.6139, -57825.6159, -55729.6599, -56432.772, -54873.9071, -51720.2288, -47638.9008, -43297.0865, -39361.9496, -36500.6538, -33236.4988, -28807.0228},
        {-31.968384, -32.380614, -32.6991413, -32.9267915, -33.06639, -33.1207622, -33.0927339, -32.9851303, -32.800777, -32.5424995, -32.2131233, -31.8154739, -31.3523767, -30.8266574, -30.2411413, -29.598654, -28.9020209, -28.1540676, -27.3576195, -26.5155022, -25.6305412, -24.7055618, -23.7433897, -22.7468503, -21.7187691, -20.6619716, -19.5792834, -18.4735298, -17.3475364, -16.2041288, -15.0461323, -13.8763725, -12.6976748, -11.5128649, -10.3775585, -9.26707568, -8.15610175, -7.04574905, -5.93712995, -4.83135677, -3.72954186, -2.63279757, -1.54223623, -0.458970192, 0.615888203, 1.68122661, 2.7359327, 3.77889411, 4.80899851, 5.82513355, 6.82618689, 7.78711898, 8.73126338, 9.65802644, 10.5668145, 11.457034, 12.3280911, 13.1793924, 14.010344, 14.8203525, 15.608824, 16.3751651, 17.118782, 17.839081, 18.5354686, 19.2073511, 19.8541349, 20.5507706, 21.3703951, 22.1566104, 22.9039254, 23.6068492, 24.2598912, 24.8575602, 25.3943655, 25.8648161, 26.2634213, 26.58469, 26.8231315, 26.9732548, 27.0295691, 26.9865834, 26.838807, 26.5807489, 25.2143294, 23.2512299, 21.2150617, 19.1325454, 17.0304016, 14.9353508, 12.8741134, 10.8734101, 8.95996133, 7.16048769, 5.50170969, 4.01034786, 2.71312274, 1.63675486, 0.807964752, 0.253472954, 0, 2.92509464, 6.00963733, 9.24579604, 12.6257387, 16.1416334, 19.7856479, 23.5499504, 27.4267087, 31.4080908, 35.4862648, 39.6533984, 43.9016599, 48.223217, 52.6102378, 57.0548902, 61.5493422, 65.9827963, 70.332711, 74.8241526, 79.4272249, 84.1120319, 88.8486775, 93.6072656, 98.3579002, 103.070685, 107.715725, 112.263122, 116.682982, 120.945408, 125.268446, 129.283347, 132.88583, 135.971617, 139.060602, 141.73642, 143.416259, 143.808588, 142.621876, 139.564592, 134.345206, 128.107502, 118.687798, 105.513741, 88.0129768, 65.6131527, 39.2830217, 6.52384667, -33.6220028, -82.112157, -139.904246, -204.26861, -284.356858, -383.32495, -504.559835, -651.920069, -829.052466, -1039.60384, -1282.47436, -1567.13627, -1908.23995, -2311.39856, -2792.31953, -3348.88903, -3994.93081, -4771.27394, -5706.3516, -6823.0741, -8148.44277, -9729.70949, -11602.6713, -13633.9591, -16054.5615, -18958.8332, -22441.1286, -26595.8022, -34016.6606, -41673.7434, -49036.542, -55574.5481, -60757.2532, -62221.9716, -62491.8236, -61821.5194, -61069.4538, -60251.3787, -59131.511, -57738.1796, -56099.7131, -54244.4405, -52200.6904, -51528.8086, -49674.7923, -46951.2765, -43670.896, -40146.2858, -36690.0808, -33614.9159, -30592.8296, -27049.4602},
        {-33.3787552, -33.890169, -34.301454, -34.6155483, -34.8353897, -34.9639163, -35.0040661, -34.958777, -34.8309871, -34.6236343, -34.3396567, -33.9819921, -33.5535787, -33.0573543, -32.496257, -31.8732248, -31.1911956, -30.4531075, -29.6618984, -28.8205063, -27.9318692, -26.9989251, -26.024612, -25.0118679, -23.9636307, -22.8828385, -21.7724293, -20.6353409, -19.4745115, -18.2928789, -17.0933813, -15.8789565, -14.6525426, -13.4170776, -12.2233717, -11.0494198, -9.87264156, -8.6944213, -7.5161435, -6.33919259, -5.16495303, -3.99480923, -2.83014566, -1.67234675, -0.522796931, 0.617119346, 1.74601764, 2.86251352, 3.96522254, 5.05276025, 6.12374223, 7.14120245, 8.14010938, 9.11984993, 10.079811, 11.0193795, 11.9379423, 12.8348863, 13.7095984, 14.5614655, 15.3898745, 16.1942123, 16.9738658, 17.7282219, 18.4566675, 19.1585895, 19.8333748, 20.5545546, 21.3938799, 22.1978205, 22.9609569, 23.6778696, 24.3431388, 24.9513452, 25.497069, 25.9748907, 26.3793908, 26.7051496, 26.9467476, 27.0987652, 27.1557829, 27.112381, 26.9631401, 26.7026404, 25.3337732, 23.3678223, 21.3279911, 19.2410422, 17.1337382, 15.0328418, 12.9651156, 10.9573223, 9.03622456, 7.228585, 5.56116628, 4.06073105, 2.75404196, 1.66786166, 0.828952796, 0.264078027, 0, 2.74950821, 5.6810666, 8.7865372, 12.0577821, 15.4866633, 19.0650428, 22.7847827, 26.6377451, 30.615792, 34.7107855, 38.9145875, 43.2190602, 47.6160656, 52.0974657, 56.6551226, 61.2808983, 65.7835755, 70.128862, 74.7227281, 79.5178044, 84.4667217, 89.5221108, 94.6366024, 99.7628273, 104.853416, 109.861, 114.738209, 119.437674, 123.912027, 128.222396, 132.17313, 135.684312, 138.676021, 141.594851, 144.097627, 145.700771, 146.166412, 145.256678, 142.733697, 138.359596, 133.829, 126.38339, 115.406796, 100.283248, 80.3967787, 56.3542931, 26.0112279, -11.5541053, -57.2633944, -112.038328, -173.128861, -249.616088, -346.02045, -465.83934, -610.481417, -782.293138, -983.620961, -1209.23277, -1470.77599, -1788.16561, -2171.23511, -2632.27103, -3172.75461, -3805.10323, -4560.26493, -5460.68189, -6514.46563, -7740.34296, -9209.58637, -10963.7456, -12922.7664, -15268.5106, -18084.6215, -21454.7425, -25462.5168, -32593.8017, -39929.4731, -46952.6208, -53146.3346, -57993.7044, -59159.7594, -59157.6904, -58249.7564, -57342.925, -56444.2576, -55292.1869, -53907.2059, -52309.8076, -50520.4847, -48559.7301, -46405.6149, -44226.9676, -41959.5784, -39539.2374, -36901.7349, -33982.8611, -30718.4061, -27815.7635, -25088.0328},
        {-35.0428378, -35.6053824, -36.0653165, -36.4255591, -36.6890289, -36.8586449, -36.9373259, -36.9279906, -36.8335581, -36.656947, -36.4010763, -36.0688648, -35.6632313, -35.1870947, -34.6433739, -34.0349876, -33.3648548, -32.6358942, -31.8510248, -31.0131653, -30.1252346, -29.1901516, -28.2108351, -27.1902039, -26.1311769, -25.0366729, -23.9096109, -22.7529095, -21.5694877, -20.3622643, -19.1341582, -17.8880882, -16.6269732, -15.3537319, -14.1100151, -12.8785611, -11.6416426, -10.4009214, -9.15805954, -7.9147189, -6.67256141, -5.433249, -4.19844357, -2.96980707, -1.7490014, -0.537688497, 0.662469726, 1.84981134, 3.02267443, 4.17939707, 5.31831734, 6.39034917, 7.44228285, 8.47348452, 9.48332032, 10.4711564, 11.4363588, 12.3782939, 13.2963276, 14.1898261, 15.0581556, 15.9006822, 16.716772, 17.5057913, 18.2671061, 19.0000825, 19.7040867, 20.4513758, 21.3124567, 22.1363636, 22.9177373, 23.6512187, 24.3314489, 24.9530686, 25.510719, 25.9990407, 26.4126749, 26.7462623, 26.994444, 27.1518608, 27.2131537, 27.1729635, 27.0259312, 26.7666978, 25.3978911, 23.4310243, 21.3895336, 19.3002526, 17.1900151, 15.0856547, 13.014005, 11.0018996, 9.07617229, 7.26365662, 5.59118626, 4.08559487, 2.77371608, 1.68238355, 0.838430931, 0.268691864, 0, 2.78720587, 5.7425945, 8.85991927, 12.1329336, 15.5553908, 19.1210443, 22.8236475, 26.6569537, 30.6147164, 34.690689, 38.8786248, 43.1722771, 47.5653995, 52.0517453, 56.6250678, 61.2791205, 65.755077, 70.0106078, 74.6107778, 79.4952159, 84.6035513, 89.8754133, 95.2504309, 100.668233, 106.06845, 111.390709, 116.574641, 121.559875, 126.286039, 130.713464, 134.753488, 138.339529, 141.405006, 144.284397, 146.710591, 148.309528, 148.894308, 148.278034, 146.273805, 142.694722, 139.665924, 133.98481, 125.012127, 112.108621, 94.6350365, 73.3232803, 45.8201482, 11.1435963, -31.6884196, -83.6579436, -142.665955, -216.541308, -309.700811, -425.333963, -564.124515, -727.881246, -918.412937, -1129.17115, -1372.22124, -1668.74557, -2029.1254, -2465.45322, -2984.85437, -3599.27894, -4328.85221, -5189.94694, -6176.8205, -7300.11202, -8651.55027, -10276.9951, -12141.8422, -14384.7948, -17079.5658, -20299.868, -24119.4143, -30895.3705, -37847.1334, -44477.5527, -50289.4779, -54785.7588, -55695.1437, -55477.3179, -54395.4601, -53375.5835, -52418.4949, -51248.8204, -49881.1761, -48330.1779, -46610.4418, -44736.584, -41339.4055, -38901.5931, -37045.4143, -35393.1366, -33567.0275, -31189.3546, -27882.3853, -25036.2328, -22940.7585},
        {-37.544799, -38.0124978, -38.3857569, -38.6671463, -38.8592364, -38.9645973, -38.9857992, -38.9254125, -38.7860073, -38.5701539, -38.2804225, -37.9193833, -37.4896065, -36.9936625, -36.4341213, -35.8135532, -35.1345286, -34.3996175, -33.6113902, -32.772417, -31.885268, -30.9525136, -29.9767239, -28.9604691, -27.9063195, -26.8168454, -25.6946169, -24.5422042, -23.3621777, -22.1571075, -20.9295638, -19.682117, -18.4173371, -17.1377945, -15.8669018, -14.5963696, -13.317658, -12.0326609, -10.7432722, -9.45138562, -8.1588951, -6.86769447, -5.57967758, -4.29673829, -3.02077044, -1.75366789, -0.497324482, 0.74636592, 1.97550947, 3.18821232, 4.38258061, 5.49975327, 6.59603861, 7.67077772, 8.72331168, 9.75298158, 10.7591285, 11.7410935, 12.6982177, 13.6298422, 14.5353081, 15.4139564, 16.2651283, 17.0881648, 17.882407, 18.647196, 19.3818729, 20.1582149, 21.04626, 21.8959552, 22.701946, 23.4588776, 24.1613954, 24.8041448, 25.3817711, 25.8889196, 26.3202356, 26.6703647, 26.933952, 27.1056429, 27.1800828, 27.151917, 27.0157909, 26.7663498, 25.3985366, 23.4278752, 21.3819053, 19.2877145, 17.1723905, 15.0630208, 12.9866931, 10.970495, 9.04151407, 7.22683797, 5.55355431, 4.0487507, 2.73951476, 1.65293411, 0.816096349, 0.256089109, 0, 3.42370294, 6.90338575, 10.4404879, 14.0364487, 17.6927077, 21.4107043, 25.1918778, 29.0376679, 32.9495138, 36.928855, 40.9771309, 45.095781, 49.2862446, 53.5499613, 57.8883704, 62.3029114, 66.5051945, 70.455255, 74.8100983, 79.5090576, 84.491466, 89.6966568, 95.0639631, 100.532718, 106.042255, 111.531907, 116.941007, 122.208888, 127.274885, 132.154494, 136.682957, 140.776757, 144.352377, 147.544638, 150.160854, 151.950118, 152.763414, 152.451723, 150.866028, 147.857313, 145.562048, 140.850145, 133.125424, 121.791709, 106.252821, 88.8284526, 65.2775876, 34.2750799, -5.50421636, -55.385447, -115.73911, -190.011791, -278.631794, -383.390333, -508.861232, -658.368977, -835.238059, -1039.10765, -1277.82512, -1563.2582, -1896.50264, -2298.29414, -2786.07003, -3374.04871, -4070.9942, -4888.17419, -5811.26146, -6844.88797, -8087.53361, -9584.59258, -11318.1815, -13406.7072, -15917.4675, -18917.7605, -22474.884, -28790.3819, -35263.7449, -41428.7094, -46819.0118, -50968.3888, -51745.975, -51459.8432, -50357.1305, -49293.8961, -48276.2457, -47056.5674, -45653.6528, -44086.2935, -42373.281, -40533.4069, -36714.5114, -34238.4424, -32627.362, -31403.4323, -30088.8156, -28205.6738, -25276.1692, -22481.5058, -20585.7067},
        {-39.8659112, -40.2892826, -40.6224116, -40.8676541, -41.0273661, -41.1039036, -41.0996224, -41.0168786, -40.8580281, -40.6254269, -40.3214309, -39.9483961, -39.5086784, -39.0046338, -38.4386182, -37.8129877, -37.1300981, -36.3923054, -35.6019656, -34.7614346, -33.8730684, -32.9392229, -31.9622541, -30.944518, -29.8883705, -28.7961675, -27.6702651, -26.5130191, -25.3267856, -24.1139204, -22.8767796, -21.6177191, -20.3390949, -19.0432628, -17.7403089, -16.4285616, -15.1062562, -13.7754979, -12.4383919, -11.0970431, -9.75355671, -8.41003782, -7.06859154, -5.73132296, -4.40033718, -3.07773932, -1.76563446, -0.466127718, 0.818675809, 2.08667102, 3.33575281, 4.49701264, 5.63659702, 6.75384901, 7.84811167, 8.91872807, 9.96504127, 10.9863943, 11.9821303, 12.9515924, 13.8941234, 14.8090666, 15.695765, 16.5535617, 17.3817996, 18.179822, 18.9469718, 19.7557975, 20.6770898, 21.5592112, 22.3967593, 23.1843315, 23.9165253, 24.5879381, 25.1931674, 25.7268106, 26.1834652, 26.5577288, 26.8441986, 27.0374722, 27.1321471, 27.1228208, 27.0040906, 26.770554, 25.4072907, 23.434891, 21.3860786, 19.2882117, 17.1686488, 15.054748, 12.9738675, 10.9533657, 9.02060078, 7.20293099, 5.5277146, 4.02230987, 2.71407505, 1.63036839, 0.798548165, 0.245972615, 0, 4.11548897, 8.16480528, 12.1575819, 16.1034519, 20.0120481, 23.8930037, 27.7559515, 31.6105245, 35.4663558, 39.3330783, 43.2203251, 47.137729, 51.0949231, 55.1015403, 59.1672137, 63.3015762, 67.1924192, 70.8033981, 74.8721213, 79.3392558, 84.1454682, 89.2314257, 94.5377951, 100.005243, 105.574437, 111.186044, 116.78073, 122.299162, 127.682008, 132.994649, 138.007309, 142.623239, 146.745692, 150.394275, 153.414064, 155.619105, 156.877742, 157.058324, 156.029193, 153.658698, 151.826478, 147.777451, 140.986448, 130.9283, 117.077838, 103.340825, 83.6534291, 56.382747, 19.8958773, -27.4400818, -88.3506057, -162.040654, -244.83114, -337.117292, -447.612738, -581.296314, -743.146858, -939.643205, -1173.92328, -1447.48866, -1751.52523, -2115.42528, -2565.61598, -3118.49747, -3773.47958, -4537.77097, -5391.57972, -6335.11432, -7467.58542, -8832.20248, -10418.8204, -12332.2115, -14634.0542, -17386.0266, -20649.8072, -26456.4386, -32405.8936, -38067.5094, -43010.6233, -46804.5725, -47485.845, -47178.5266, -46108.8899, -45043.0932, -43993.7042, -42748.3386, -41330.8117, -39764.9391, -38074.5362, -36283.4184, -32465.4634, -30075.8025, -28631.8421, -27650.9886, -26650.6483, -25148.2278, -22661.1334, -19978.0625, -18167.1449},
        {-41.5002906, -42.0353056, -42.4722038, -42.8135235, -43.0618034, -43.2195817, -43.2893971, -43.273788, -43.1752928, -42.99645, -42.7397982, -42.4078757, -42.003221, -41.5283726, -40.985869, -40.3782486, -39.7080499, -38.9778114, -38.1900716, -37.3473689, -36.4522417, -35.5072286, -34.514868, -33.4776984, -32.3982583, -31.2790861, -30.1227203, -28.9316994, -27.7085618, -26.4558461, -25.1760906, -23.8718338, -22.5456143, -21.1999704, -19.8447536, -18.478692, -17.100426, -15.7122567, -14.3164852, -12.9154128, -11.5113406, -10.1065697, -8.70340127, -7.30413648, -5.91107649, -4.52652244, -3.15277548, -1.79213677, -0.446907454, 0.880611304, 2.18811836, 3.39416227, 4.577308, 5.73697021, 6.87256357, 7.98350275, 9.06920242, 10.1290773, 11.1625419, 12.1690111, 13.1478994, 14.0986216, 15.0205923, 15.9132262, 16.7759379, 17.6081422, 18.4092536, 19.2552149, 20.2201312, 21.1455839, 22.026027, 22.8559139, 23.6296984, 24.3418341, 24.9867747, 25.558974, 26.0528855, 26.462963, 26.7836601, 27.0094304, 27.1347278, 27.1540058, 27.0617182, 26.8523185, 25.4988189, 23.5278785, 21.4784334, 19.378085, 17.2544347, 15.1350839, 13.0476341, 11.0196866, 9.07884291, 7.25270443, 5.56887257, 4.05494875, 2.73853439, 1.64723092, 0.808639753, 0.250362305, 0, 4.62229525, 9.08547483, 13.405486, 17.5982761, 21.6797923, 25.6659819, 29.5727922, 33.4161704, 37.2120639, 40.9764199, 44.7251856, 48.4743084, 52.2397355, 56.0374142, 59.8832917, 63.7933153, 67.4203636, 70.728517, 74.5549494, 78.8378077, 83.5152389, 88.5253897, 93.8064069, 99.2964374, 104.933628, 110.656125, 116.402077, 122.109629, 127.716928, 133.241395, 138.512836, 143.445617, 147.9541, 152.175358, 155.912401, 158.908853, 161.012269, 162.070202, 161.930203, 160.439828, 158.864145, 155.201773, 149.022925, 139.897815, 127.396656, 116.325211, 99.7092559, 75.8101152, 42.8891145, -0.792420539, -59.2038337, -129.126757, -204.469562, -284.779013, -381.113587, -499.363236, -645.417913, -830.132336, -1053.30334, -1309.31164, -1583.09018, -1908.41477, -2315.88371, -2822.60795, -3422.84487, -4121.34201, -4896.30064, -5745.58593, -6764.39937, -7992.88351, -9425.32477, -11154.7766, -13237.2549, -15728.7755, -18685.3541, -23957.5949, -29358.2783, -34494.7733, -38974.4486, -42404.673, -42999.0167, -42687.8461, -41675.8723, -40643.1947, -39606.3499, -38383.7956, -37001.9724, -35487.3205, -33866.2803, -32165.2922, -28781.1256, -26615.7034, -25264.4899, -24322.9494, -23386.5462, -22050.7446, -19911.0088, -17497.8076, -15784.2861},
        {-41.8758817, -42.7982295, -43.5933351, -44.2646127, -44.8154768, -45.2493419, -45.5696224, -45.7797326, -45.8830871, -45.8831002, -45.7831863, -45.5867599, -45.2972355, -44.9180273, -44.45255, -43.9042178, -43.2764452, -42.5726467, -41.7962366, -40.9506293, -40.0392394, -39.0654813, -38.0327692, -36.9445178, -35.8041413, -34.6150543, -33.3806711, -32.1044062, -30.789674, -29.4398889, -28.0584653, -26.6488177, -25.2143606, -23.7585082, -22.313042, -20.8665964, -19.4074645, -17.9381403, -16.4611177, -14.9788905, -13.4939526, -12.0087977, -10.5259199, -9.04781291, -7.57697064, -6.11588695, -4.66705569, -3.23297073, -1.81612593, -0.419015148, 0.955867748, 2.20875777, 3.43654084, 4.63883173, 5.81524524, 6.96539614, 8.08889923, 9.18536929, 10.2544211, 11.2956695, 12.3087291, 13.2932149, 14.2487416, 15.174924, 16.0713769, 16.937715, 17.7735531, 18.6627862, 19.6872855, 20.6728446, 21.6136145, 22.5037467, 23.3373923, 24.1087026, 24.8118288, 25.4409222, 25.990134, 26.4536155, 26.825518, 27.0999926, 27.2711907, 27.3332635, 27.2803622, 27.1066381, 25.7701986, 23.8050086, 21.7574855, 19.6554023, 17.5265318, 15.3986468, 13.2995203, 11.2569251, 9.2986341, 7.45242011, 5.74605602, 4.2073147, 2.86396901, 1.74379184, 0.87455603, 0.284034462, 0, 4.66951365, 9.16076846, 13.4914318, 17.679171, 21.7416535, 25.6965466, 29.5615177, 33.3542342, 37.0923633, 40.7935725, 44.4755292, 48.1559007, 51.8523543, 55.5825574, 59.3641775, 63.2148818, 66.720359, 69.8365653, 73.5582626, 77.8105515, 82.5185325, 87.6073061, 93.0019728, 98.627633, 104.409387, 110.272336, 116.14158, 121.94222, 127.599355, 132.871332, 137.911149, 142.693935, 147.194815, 152.089669, 157.003249, 161.373438, 164.96514, 167.543255, 168.872686, 168.718334, 167.237308, 163.682935, 157.743383, 149.106818, 137.461404, 127.04129, 111.852165, 90.4456984, 61.3735602, 23.187421, -27.058319, -87.484723, -153.368784, -224.585857, -310.399618, -415.858578, -546.011248, -710.061579, -907.958224, -1135.11663, -1378.78265, -1667.90613, -2028.41135, -2475.04744, -3003.62792, -3618.65806, -4300.13931, -5046.22409, -5944.91222, -7033.38077, -8313.67303, -9863.50502, -11732.2155, -13969.1431, -16623.6267, -21360.7368, -26210.1473, -30817.2642, -34827.4931, -37886.24, -38375.2925, -38046.0866, -37085.5786, -36117.9762, -35157.157, -34034.811, -32773.9254, -31397.4872, -29928.4837, -28389.9021, -25905.6338, -24124.3714, -22792.0404, -21654.5663, -20457.8744, -18947.8901, -16870.5389, -15007.5978, -13548.6344},
        {-41.7005097, -43.1318052, -44.3969678, -45.5006253, -46.4474057, -47.2419368, -47.8888463, -48.392762, -48.7583119, -48.9901236, -49.0928251, -49.0710442, -48.9294086, -48.6725462, -48.3050848, -47.8316523, -47.2568763, -46.5853849, -45.8218058, -44.9707667, -44.0368956, -43.0248203, -41.9391685, -40.7845681, -39.5656469, -38.2870327, -36.9533534, -35.5692367, -34.1393105, -32.6682027, -31.1605409, -29.6209532, -28.0540671, -26.4645107, -24.9204074, -23.3933017, -21.8539747, -20.3049938, -18.7489262, -17.1883391, -15.6257999, -14.0638758, -12.505134, -10.9521418, -9.40746644, -7.87367524, -6.35333542, -4.84901426, -3.363279, -1.89869691, -0.457835246, 0.852805355, 2.13684408, 3.39396673, 4.62385912, 5.82620707, 7.00069639, 8.14701288, 9.26484235, 10.3538706, 11.4137835, 12.4442668, 13.4450063, 14.4156879, 15.3559974, 16.2656205, 17.144243, 18.0792784, 19.1560336, 20.1932631, 21.1849656, 22.1251398, 23.0077847, 23.8268989, 24.5764814, 25.2505308, 25.8430459, 26.3480257, 26.7594687, 27.071374, 27.2777401, 27.372566, 27.3498504, 27.2035921, 25.8852178, 23.9319416, 21.8935081, 19.7977496, 17.6724985, 15.545587, 13.4448474, 11.3981118, 9.43321265, 7.57798211, 5.86025246, 4.30785597, 2.9486249, 1.8103915, 0.920988047, 0.308246792, 0, 4.64943176, 9.11441333, 13.4133963, 17.5648321, 21.5871724, 25.4988687, 29.3183725, 33.0641354, 36.7546089, 40.4082447, 44.0434941, 47.6788089, 51.3326404, 55.0234403, 58.7696602, 62.5897515, 66.0323802, 69.0488856, 72.7203296, 76.9644954, 81.6991661, 86.8421249, 92.3111552, 98.02404, 103.898563, 109.852506, 115.803654, 121.669789, 127.368695, 132.476084, 137.377236, 142.092555, 146.642446, 152.06679, 157.876254, 163.309642, 168.081514, 171.906432, 174.498956, 175.573649, 174.446987, 171.35277, 166.083445, 158.431457, 148.189255, 138.709022, 125.333532, 106.965299, 82.5068333, 50.8606481, 10.4482074, -38.7589808, -94.0232579, -155.730919, -230.64633, -322.670233, -435.703371, -576.326529, -745.151306, -939.865616, -1151.39808, -1402.25813, -1711.00808, -2089.55426, -2538.52473, -3063.72446, -3650.98735, -4300.94067, -5087.43895, -6042.91522, -7168.51197, -8532.76674, -10178.9439, -12150.3079, -14490.1231, -18679.7933, -22964.9086, -27029.1986, -30556.3928, -33230.2209, -33599.1332, -33238.9912, -32320.0725, -31418.4775, -30543.8458, -29536.7243, -28415.2074, -27197.3894, -25901.3645, -24545.2271, -23013.5714, -21625.473, -20298.295, -18949.4005, -17496.1525, -15855.9139, -13946.0479, -12571.9813, -11334.585},
        {-41.4908839, -43.5277993, -45.3486246, -46.9596585, -48.3671998, -49.5775473, -50.5969999, -51.4318563, -52.0884155, -52.5729761, -52.8918369, -53.0512969, -53.0576547, -52.9172093, -52.6362594, -52.2211038, -51.6780413, -51.0133707, -50.2333909, -49.3444006, -48.3526987, -47.2645839, -46.0863551, -44.8243111, -43.4847506, -42.0739726, -40.5982757, -39.0639589, -37.4773208, -35.8446604, -34.1722764, -32.4664676, -30.7335329, -28.979771, -27.3365347, -25.7449654, -24.1447017, -22.5379841, -20.9270531, -19.3141494, -17.7015135, -16.0913859, -14.4860074, -12.8876183, -11.2984593, -9.720771, -8.1567939, -6.60876859, -5.07893564, -3.56953561, -2.08280908, -0.694862551, 0.667530617, 2.00373112, 3.31309967, 4.59499695, 5.84878367, 7.07382053, 8.26946822, 9.43508746, 10.5700389, 11.6736834, 12.7453814, 13.7844938, 14.7903813, 15.7624044, 16.6999241, 17.6805408, 18.7799767, 19.8362285, 20.843585, 21.7963348, 22.6887667, 23.5151694, 24.2698316, 24.9470421, 25.5410895, 26.0462627, 26.4568502, 26.7671409, 26.9714235, 27.0639866, 27.0391191, 26.8911096, 25.5807559, 23.6408303, 21.6165269, 19.5356732, 17.4260965, 15.3156245, 13.2320846, 11.2033043, 9.25711114, 7.42133254, 5.72379601, 4.19232905, 2.85475916, 1.73891381, 0.872620512, 0.283706743, 0, 4.96728855, 9.69560562, 14.2070532, 18.5237333, 22.6677479, 26.6611991, 30.5261887, 34.2848189, 37.9591916, 41.5714088, 45.1435725, 48.6977848, 52.2561475, 55.8407627, 59.4737325, 63.1771588, 66.5256016, 69.4772286, 73.0664388, 77.2194322, 81.8624092, 86.9215699, 92.3231144, 97.9932429, 103.858156, 109.844053, 115.877134, 121.883601, 127.789652, 133.304879, 138.651514, 143.820741, 148.803743, 154.245563, 159.810453, 164.988305, 169.574141, 173.362988, 176.14987, 177.729811, 177.69455, 176.104268, 172.793763, 167.597834, 160.351278, 153.651085, 143.879314, 130.180216, 111.69804, 87.577038, 57.007902, 19.0316768, -24.7177091, -74.846896, -136.537019, -212.915358, -307.109194, -424.65057, -565.715467, -727.856487, -903.16339, -1110.17038, -1359.31613, -1660.69172, -2023.79935, -2458.41153, -2964.17849, -3548.24844, -4254.88463, -5107.35591, -6083.0266, -7258.13239, -8671.83219, -10363.2849, -12371.6496, -16011.7236, -19728.1177, -23241.0811, -26270.863, -28537.7128, -28745.3536, -28308.2448, -27382.2898, -26502.219, -25673.9833, -24745.689, -23731.1791, -22644.2968, -21498.8851, -20308.7872, -19418.6632, -18291.4348, -16995.6408, -15599.8204, -14172.5127, -12782.2567, -11497.5914, -10243.0814, -9077.9173},
        {-40.3788505, -43.0186901, -45.3985053, -47.5259592, -49.408715, -51.0544356, -52.4707842, -53.6654237, -54.6460174, -55.4202282, -55.9957192, -56.3801535, -56.5811942, -56.6065043, -56.4637469, -56.1605851, -55.7046819, -55.1037004, -54.3653037, -53.4971548, -52.5069169, -51.402253, -50.1908261, -48.8802993, -47.4783357, -45.9925984, -44.4307505, -42.8004549, -41.1093748, -39.3651733, -37.5755134, -35.7480582, -33.8904707, -32.010414, -30.2856276, -28.6351589, -26.9760114, -25.3103289, -23.6402552, -21.9679339, -20.2955089, -18.6251239, -16.9589227, -15.2990491, -13.6476468, -12.0068596, -10.3788312, -8.76570554, -7.16962623, -5.59273708, -4.03718187, -2.55569911, -1.09874102, 0.332645433, 1.73741327, 3.11451553, 4.46290524, 5.78153541, 7.0693591, 8.32532931, 9.54839908, 10.7375214, 11.8916494, 13.0097361, 14.0907344, 15.1335974, 16.1372781, 17.1656156, 18.2808911, 19.348235, 20.3623941, 21.318115, 22.2101446, 23.0332295, 23.7821165, 24.4515523, 25.0362836, 25.5310572, 25.9306198, 26.2297182, 26.423099, 26.505509, 26.4716949, 26.3164036, 25.0272775, 23.1237987, 21.1375535, 19.0959711, 17.0264807, 14.9565115, 12.9134927, 10.9248536, 9.01802319, 7.2204308, 5.55950559, 4.06267675, 2.75737349, 1.67102499, 0.831060441, 0.264909047, 0, 4.83403883, 9.44701426, 13.8598494, 18.0934675, 22.1687916, 26.106745, 29.9282507, 33.6542319, 37.3056118, 40.9033135, 44.4682602, 48.0213751, 51.5835812, 55.1758018, 58.8189599, 62.5339788, 65.998768, 69.1852487, 72.8902082, 77.0610667, 81.6452445, 86.5901617, 91.8432386, 97.3518956, 103.063553, 108.92563, 114.885548, 120.890728, 126.888588, 132.793547, 138.598129, 144.259654, 149.735444, 155.223395, 160.560541, 165.519761, 169.986204, 173.845016, 176.981347, 179.280345, 180.672404, 180.983655, 180.090394, 177.868917, 174.195519, 171.190005, 165.924285, 157.713776, 145.873898, 129.720069, 108.809493, 81.9202858, 49.9860222, 12.3586347, -34.839128, -94.0346797, -167.655434, -260.109348, -371.393167, -499.343046, -637.606882, -800.634902, -992.272352, -1220.71291, -1501.5184, -1846.86858, -2266.72537, -2772.69351, -3384.50968, -4117.31067, -4928.79807, -5900.38005, -7067.07482, -8463.90058, -10125.8755, -13195.2454, -16322.994, -19267.3327, -21786.4727, -23638.6255, -23689.3017, -23184.5471, -22265.1588, -21423.8238, -20662.0497, -19834.7236, -18950.6841, -18018.77, -17047.8199, -16046.6727, -15678.3088, -14793.6669, -13578.2728, -12217.6526, -10897.3324, -9802.83826, -9119.69614, -7958.23246, -6861.85018},
        {-36.7036308, -39.5083493, -42.0713855, -44.3991544, -46.4980708, -48.3745495, -50.0350055, -51.4858535, -52.7335084, -53.7843851, -54.6448984, -55.3214632, -55.8204943, -56.1484066, -56.3116149, -56.3165341, -56.169579, -55.8771646, -55.4457056, -54.8816168, -54.1913133, -53.3812097, -52.457721, -51.427262, -50.2962476, -49.0710926, -47.7582118, -46.3640202, -44.8949326, -43.3573638, -41.7577286, -40.1024421, -38.3979189, -36.6505739, -34.9531206, -33.2670094, -31.5526535, -29.813667, -28.0536643, -26.2762598, -24.4850676, -22.6837021, -20.8757777, -19.0649085, -17.2547089, -15.4487933, -13.6507758, -11.8642708, -10.0928925, -8.34025541, -6.60997366, -5.01237094, -3.44203894, -1.90027872, -0.388391328, 1.09232219, 2.5405608, 3.95502343, 5.33440905, 6.6774166, 7.98274504, 9.24909332, 10.4751604, 11.6596452, 12.8012467, 13.8986638, 14.9505956, 16.0035983, 17.1030794, 18.1490354, 19.1370628, 20.0627583, 20.9217183, 21.7095395, 22.4218183, 23.0541513, 23.602135, 24.0613661, 24.427441, 24.6959564, 24.8625087, 24.9226944, 24.8721103, 24.7063527, 23.5235684, 21.7869554, 19.9704987, 18.0989187, 16.1969359, 14.2892709, 12.4006442, 10.5557764, 8.77938793, 7.09619933, 5.53093114, 4.10830389, 2.85303809, 1.78985428, 0.943472978, 0.33861471, 0, 0.94919764, 2.30268582, 4.04103124, 6.14480059, 8.59456058, 11.3708779, 14.4543193, 17.8254513, 21.4648408, 25.3530544, 29.4706589, 33.7982208, 38.316307, 43.0054841, 47.8463188, 52.8193778, 58.0817659, 63.6392697, 69.0731433, 74.4017829, 79.643585, 84.8169457, 89.9402614, 95.0319285, 100.110343, 105.193902, 110.301001, 115.450036, 120.659404, 126.017482, 131.447025, 136.945437, 142.510119, 148.712383, 155.262676, 161.718405, 167.904797, 173.647084, 178.770494, 183.100258, 185.289178, 186.691736, 187.362549, 187.356235, 186.727411, 189.193908, 190.231325, 188.978479, 184.574181, 176.157247, 162.076982, 143.226664, 120.622804, 94.0807075, 60.9632124, 19.7342649, -31.1421889, -92.8067728, -167.280842, -257.017131, -361.360133, -487.897865, -640.859092, -825.27583, -1047.54115, -1313.42339, -1621.70273, -1976.33576, -2406.90223, -2928.48796, -3509.24176, -4218.40758, -5087.31868, -6147.30834, -7429.70982, -9918.11755, -12453.5384, -14829.2402, -16838.4911, -18274.559, -18142.3437, -17549.0602, -16625.8485, -15860.5609, -15243.1387, -14614.8936, -13973.1984, -13315.4261, -12638.9499, -11941.1426, -12558.2743, -12135.1433, -11013.3517, -9534.50189, -8040.19597, -6872.03617, -6371.62472, -5643.63544, -4697.85634},
        {-40.407689, -45.6133109, -50.3699414, -54.690749, -58.5889019, -62.0775685, -65.1699171, -67.879116, -70.2183337, -72.2007384, -73.8394985, -75.1477823, -76.1387582, -76.8255945, -77.2214595, -77.3395216, -77.1929491, -76.7949103, -76.1585737, -75.2971074, -74.22368, -72.9514597, -71.4936148, -69.8633137, -68.0737247, -66.1380162, -64.0693565, -61.8809139, -59.5858568, -57.1973536, -54.7285725, -52.1926819, -49.6028502, -46.9722456, -44.5992744, -42.3484863, -40.0810029, -37.8007361, -35.5115975, -33.2174991, -30.9223526, -28.6300699, -26.3445628, -24.0697432, -21.8095228, -19.5678134, -17.348527, -15.1555753, -12.9928701, -10.8643233, -8.77384667, -6.84941854, -4.96819472, -3.13139751, -1.34024924, 0.404027813, 2.10021133, 3.747079, 5.34340853, 6.88797762, 8.37956395, 9.81694522, 11.1988991, 12.5242034, 13.7916356, 14.9999736, 16.147995, 17.2696805, 18.3981654, 19.459624, 20.450552, 21.3674451, 22.2067988, 22.9651088, 23.6388707, 24.2245802, 24.7187328, 25.1178243, 25.4183502, 25.6168062, 25.709688, 25.6934911, 25.5647112, 25.3198439, 24.0233633, 22.1573797, 20.2152054, 18.2235819, 16.2092509, 14.1989538, 12.2194321, 10.2974273, 8.45968112, 6.73293491, 5.14393024, 3.71940863, 2.48611162, 1.47078073, 0.700157478, 0.200983393, 0, 4.89219611, 9.553754, 14.0060151, 18.2703209, 22.3680127, 26.3204321, 30.1489204, 33.874819, 37.5194695, 41.1042131, 44.6503915, 48.1793459, 51.7124178, 55.2709487, 58.8762799, 62.549753, 66.0207475, 69.2688962, 72.97593, 77.1006272, 81.601766, 86.4381246, 91.5684811, 96.9516138, 102.546301, 108.311321, 114.205451, 120.18747, 126.216157, 132.080175, 137.970791, 143.897819, 149.871071, 156.257763, 162.889006, 169.500602, 175.995143, 182.275221, 188.243426, 193.80235, 198.201581, 202.195454, 205.814321, 209.088534, 212.048447, 214.878398, 217.416257, 219.653879, 221.583121, 223.195837, 225.817968, 226.47674, 224.945429, 220.401036, 210.803174, 194.658037, 170.47182, 136.043961, 90.7480396, 34.7286443, -31.110271, -110.977396, -206.214372, -319.515893, -455.869325, -619.20966, -807.382453, -1022.74395, -1289.97835, -1621.13988, -2005.72024, -2481.80739, -3069.26567, -3787.95943, -4657.75303, -6336.96182, -8047.38641, -9649.27839, -11002.8894, -11968.4709, -11875.3457, -11468.6469, -10836.1673, -10300.4103, -9856.26226, -9400.81302, -8933.83858, -8455.11494, -7964.41809, -7461.52402, -7653.1292, -7305.15498, -6590.85125, -5683.46794, -4756.25494, -3982.46216, -3535.33951, -3113.88376, -2409.81894},
        {-33.1017163, -38.4527941, -43.3735768, -47.8762346, -51.9729377, -55.6758561, -58.9971601, -61.9490197, -64.543605, -66.7930862, -68.7096334, -70.3054168, -71.5926064, -72.5833725, -73.289885, -73.7243142, -73.8988303, -73.8256032, -73.5168031, -72.9846003, -72.2411647, -71.2986665, -70.1692759, -68.865163, -67.3984979, -65.7814508, -64.0261917, -62.1448908, -60.1497183, -58.0528442, -55.8664386, -53.6026718, -51.2737139, -48.8917349, -46.714893, -44.6271903, -42.5096723, -40.3665264, -38.2019399, -36.0201002, -33.8251945, -31.6214103, -29.4129348, -27.2039555, -24.9986597, -22.8012346, -20.6158677, -18.4467463, -16.2980578, -14.1739894, -12.0787286, -10.1247734, -8.20565253, -6.32320526, -4.479271, -2.67568913, -0.914299029, 0.803059923, 2.47454835, 4.09832687, 5.6725561, 7.19539667, 8.6650092, 10.0795543, 11.4371926, 12.7360848, 13.9743913, 15.1732479, 16.3532385, 17.4651396, 18.5056224, 19.4713582, 20.3590182, 21.1652736, 21.8867958, 22.520256, 23.0623253, 23.5096752, 23.8589767, 24.1069012, 24.2501199, 24.2853041, 24.2091249, 24.0182538, 22.8652951, 21.1867074, 19.429094, 17.6165178, 15.7730416, 13.9227283, 12.0896406, 10.2978414, 8.57139364, 6.93436003, 5.41080343, 4.02478669, 2.80037263, 1.7616241, 0.932603929, 0.337374951, 0, 4.39147165, 8.51614545, 12.3991242, 16.0655108, 19.540408, 22.8489186, 26.0161455, 29.0671915, 32.0271594, 34.921152, 37.7742722, 40.6116228, 43.4583066, 46.3394265, 49.2800852, 52.3053856, 55.2926119, 58.2457501, 61.5242545, 65.1215524, 69.0310713, 73.2462386, 77.7604818, 82.5672282, 87.6599053, 93.0319405, 98.6767612, 104.587795, 110.758469, 117.248196, 123.960224, 130.868184, 137.945708, 145.054278, 152.2236, 159.513289, 166.930622, 174.482876, 182.177328, 190.021256, 198.218155, 206.519365, 214.893774, 223.310267, 231.737732, 240.080614, 248.388352, 256.645943, 264.838383, 272.950671, 281.424559, 289.23003, 296.77588, 303.823523, 308.812644, 310.776356, 308.747772, 301.482743, 288.354661, 269.039394, 243.600819, 209.813896, 167.564889, 115.231321, 48.634245, -35.2318241, -135.72906, -254.918385, -408.219532, -603.495735, -840.385434, -1136.24201, -1501.77574, -1947.6969, -2484.71578, -3503.11458, -4539.13866, -5508.60532, -6327.33186, -6911.13556, -6858.4512, -6614.06695, -6229.8212, -5892.91173, -5601.59717, -5297.73636, -4982.40793, -4656.69045, -4321.66255, -3978.40283, -3622.87972, -3262.87695, -2900.43325, -2537.58733, -2176.37791, -1818.84371, -1467.02346, -1135.30892, -749.490878},
        {-31.0761186, -38.1179052, -44.6242007, -50.6097214, -56.0891837, -61.077304, -65.5887987, -69.6383842, -73.2407768, -76.410693, -79.1628492, -81.5119617, -83.4727469, -85.0599212, -86.288201, -87.1723027, -87.7269426, -87.9668373, -87.9067029, -87.561256, -86.945213, -86.0732901, -84.9602038, -83.6206706, -82.0694067, -80.3211285, -78.3905525, -76.2923951, -74.0413725, -71.6522013, -69.1395978, -66.5182784, -63.8029595, -61.0083574, -58.3968477, -55.8541082, -53.2628396, -50.6297214, -47.9614329, -45.2646536, -42.5460627, -39.8123397, -37.0701639, -34.3262147, -31.5871715, -28.8597136, -26.1505204, -23.4662713, -20.8136456, -18.1993228, -15.6299822, -13.3137356, -11.0514632, -8.8454779, -6.69809224, -4.61161898, -2.58837082, -0.63066048, 1.25919934, 3.07889593, 4.82611657, 6.49854857, 8.09387921, 9.60979578, 11.0439856, 12.3941359, 13.657934, 14.7955315, 15.7679811, 16.6503848, 17.4428631, 18.1455367, 18.758526, 19.2819517, 19.7159345, 20.0605947, 20.3160531, 20.4824302, 20.5598465, 20.5484228, 20.4482794, 20.2595371, 19.9823165, 19.616738, 18.6719332, 17.4038446, 16.0743159, 14.699401, 13.2951543, 11.8776297, 10.4628814, 9.06696347, 7.70593009, 6.39583536, 5.15273338, 3.99267828, 2.93172417, 1.98592517, 1.1713354, 0.504008968, 0, -0.0150103308, 0.21624665, 0.691340598, 1.40784117, 2.36331802, 3.55534081, 4.9814792, 6.63930283, 8.52638137, 10.6402845, 12.9785818, 15.538843, 18.3186377, 21.3155356, 24.5271064, 27.9509197, 31.7566329, 35.9663999, 40.1885438, 44.4575102, 48.807745, 53.2736935, 57.8898015, 62.6905146, 67.7102784, 72.9835386, 78.5447407, 84.4283305, 90.6687534, 97.461289, 104.620577, 112.132812, 119.98419, 128.039962, 136.346794, 144.983609, 153.972884, 163.337099, 173.098734, 183.280267, 193.840133, 204.884348, 216.447921, 228.56586, 241.273177, 254.560268, 268.517907, 283.192257, 298.629478, 314.875734, 332.015348, 350.009678, 369.047555, 389.168378, 410.106463, 431.733102, 453.919588, 476.572403, 499.51965, 522.551043, 545.68622, 568.357932, 589.93074, 610.047309, 628.821532, 646.151003, 663.224439, 680.274169, 692.798394, 698.973207, 700.073659, 690.700119, 667.955964, 628.944568, 570.769307, 412.974295, 249.611984, 97.1755669, -27.8417664, -108.946824, -67.4453983, 9.48734037, 111.687004, 202.872772, 283.218011, 363.777936, 444.18182, 524.058938, 603.038564, 680.749974, 705.417047, 765.687549, 849.280353, 943.914335, 1037.30837, 1117.18134, 1171.25211, 1201.32955, 1280.02151}
    };
    
    /* YPtag: chiAmax tables above */
    
    /* Stuff for interpolating the NQC data */
    gsl_spline    *spline = NULL;
    gsl_interp_accel *acc = NULL;
    /* Interpolate the spin NQC data in 2-D parameter space -- spin and mass ratio */
    /* First, interpolate in spin dimension for all mass ratios */
    spline = gsl_spline_alloc( gsl_interp_cspline, adim );
    acc    = gsl_interp_accel_alloc();
    for (i = 0; i < qdim; i++)
    {
        gsl_spline_init( spline, alist, a3stab[i], adim );
        gsl_interp_accel_reset( acc );
        a3slist[i] = gsl_spline_eval( spline, chiA, acc );
        gsl_spline_init( spline, alist, a4tab[i], adim );
        gsl_interp_accel_reset( acc );
        a4list[i] = gsl_spline_eval( spline, chiA, acc );
        gsl_spline_init( spline, alist, a5tab[i], adim );
        gsl_interp_accel_reset( acc );
        a5list[i] = gsl_spline_eval( spline, chiA, acc );
        gsl_spline_init( spline, alist, b3tab[i], adim );
        gsl_interp_accel_reset( acc );
        b3list[i] = gsl_spline_eval( spline, chiA, acc );
        gsl_spline_init( spline, alist, b4tab[i], adim );
        gsl_interp_accel_reset( acc );
        b4list[i] = gsl_spline_eval( spline, chiA, acc );
        //  printf("%.15f\n",a3slist[i]);
    }
    //printf("%.15f\n",a);
    gsl_spline_free(spline);
    gsl_interp_accel_free(acc);
    /* Second, interpolate in mass ratio dimension */
    spline = gsl_spline_alloc( gsl_interp_cspline, qdim );
    acc    = gsl_interp_accel_alloc();
    gsl_spline_init( spline, etalist, a3slist, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->a3S = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, a4list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->a4 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, a5list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->a5 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, b3list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->b3 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, b4list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->b4 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_free(spline);
    gsl_interp_accel_free(acc);
    
    return XLAL_SUCCESS;
    
}



UNUSED static int XLALSimIMRGetEOBCalibratedSpinNQCv2chiAmed( EOBNonQCCoeffs * restrict coeffs,                                                             INT4 UNUSED l,                                                             INT4 UNUSED m,                                                             REAL8 eta,                                                             REAL8 a )
{
    const unsigned int   qdim = 98;
    const unsigned int   adim = 200;
    UINT4 i;
    /* REAL8 eta2 = eta*eta;*/
    REAL8 a3slist[qdim], a4list[qdim], a5list[qdim];
    REAL8 b3list[qdim], b4list[qdim];
    REAL8 chiA;
    
    chiA = a/(1.0-2.0*eta);
    if (chiA < -1.){
        chiA = -1.;
    }
    if (chiA > 0.99){
        chiA = 0.99;
    }
    
    memset( coeffs, 0, sizeof( *coeffs ) );
    const double alist[200] = {-1., -0.99, -0.98, -0.97, -0.96, -0.95, -0.94, -0.93, -0.92, -0.91, -0.9, -0.89, -0.88, -0.87, -0.86, -0.85, -0.84, -0.83, -0.82, -0.81, -0.8, -0.79, -0.78, -0.77, -0.76, -0.75, -0.74, -0.73, -0.72, -0.71, -0.7, -0.69, -0.68, -0.67, -0.66, -0.65, -0.64, -0.63, -0.62, -0.61, -0.6, -0.59, -0.58, -0.57, -0.56, -0.55, -0.54, -0.53, -0.52, -0.51, -0.5, -0.49, -0.48, -0.47, -0.46, -0.45, -0.44, -0.43, -0.42, -0.41, -0.4, -0.39, -0.38, -0.37, -0.36, -0.35, -0.34, -0.33, -0.32, -0.31, -0.3, -0.29, -0.28, -0.27, -0.26, -0.25, -0.24, -0.23, -0.22, -0.21, -0.2, -0.19, -0.18, -0.17, -0.16, -0.15, -0.14, -0.13, -0.12, -0.11, -0.1, -0.09, -0.08, -0.07, -0.06, -0.05, -0.04, -0.03, -0.02, -0.01, 0, 0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09, 0.1, 0.11, 0.12, 0.13, 0.14, 0.15, 0.16, 0.17, 0.18, 0.19, 0.2, 0.21, 0.22, 0.23, 0.24, 0.25, 0.26, 0.27, 0.28, 0.29, 0.3, 0.31, 0.32, 0.33, 0.34, 0.35, 0.36, 0.37, 0.38, 0.39, 0.4, 0.41, 0.42, 0.43, 0.44, 0.45, 0.46, 0.47, 0.48, 0.49, 0.5, 0.51, 0.52, 0.53, 0.54, 0.55, 0.56, 0.57, 0.58, 0.59, 0.6, 0.61, 0.62, 0.63, 0.64, 0.65, 0.66, 0.67, 0.68, 0.69, 0.7, 0.71, 0.72, 0.73, 0.74, 0.75, 0.76, 0.77, 0.78, 0.79, 0.8, 0.81, 0.82, 0.83, 0.84, 0.85, 0.86, 0.87, 0.88, 0.89, 0.9, 0.91, 0.92, 0.93, 0.94, 0.95, 0.96, 0.97, 0.98, 0.99};
    const double etalist[98]    = {0.0075, 0.01, 0.0125, 0.015, 0.0175, 0.02, 0.0225, 0.025, 0.0275, 0.03, 0.0325, 0.035, 0.0375, 0.04, 0.0425, 0.045, 0.0475, 0.05, 0.0525, 0.055, 0.0575, 0.06, 0.0625, 0.065, 0.0675, 0.07, 0.0725, 0.075, 0.0775, 0.08, 0.0825, 0.085, 0.0875, 0.09, 0.0925, 0.095, 0.0975, 0.1, 0.1025, 0.105, 0.1075, 0.11, 0.1125, 0.115, 0.1175, 0.12, 0.1225, 0.125, 0.1275, 0.13, 0.1325, 0.135, 0.1375, 0.14, 0.1425, 0.145, 0.1475, 0.15, 0.1525, 0.155, 0.1575, 0.16, 0.1625, 0.165, 0.1675, 0.17, 0.1725, 0.175, 0.1775, 0.18, 0.1825, 0.185, 0.1875, 0.19, 0.1925, 0.195, 0.1975, 0.2, 0.2025, 0.205, 0.2075, 0.21, 0.2125, 0.215, 0.2175, 0.22, 0.2225, 0.225, 0.2275, 0.23, 0.2325, 0.235, 0.2375, 0.24, 0.2425, 0.245, 0.2475, 0.25};
    
    /* YPtag: chiAmed tables below */
    const double a3stab[98][200] = {{486.38171, 480.717682, 475.141347, 469.649366, 464.238399, 458.905105, 453.646145, 448.458178, 443.337865, 438.281866, 433.28684, 428.349447, 423.466348, 418.634203, 413.849671, 409.109412, 404.410087, 399.748355, 395.120876, 390.524311, 385.955318, 381.41056, 376.886694, 372.380381, 367.888282, 363.407056, 358.933363, 354.463863, 349.995216, 345.524082, 341.047121, 336.560993, 332.062358, 327.547876, 322.85644, 318.066911, 313.260049, 308.437632, 303.601441, 298.753255, 293.894855, 289.028019, 284.154529, 279.276163, 274.394701, 269.511924, 264.62961, 259.749541, 254.873496, 250.003254, 245.140595, 240.419085, 235.705861, 230.999847, 226.299964, 221.605137, 216.914287, 212.226339, 207.540214, 202.854836, 198.169129, 193.482014, 188.792414, 184.099254, 179.401455, 174.697941, 169.987635, 165.263148, 160.517242, 155.761859, 150.99633, 146.219987, 141.432162, 136.632189, 131.819397, 126.993121, 122.152692, 117.297442, 112.426703, 107.539807, 102.636087, 97.7148744, 92.7755015, 87.8173004, 82.8521522, 77.8728507, 72.8720386, 67.8486408, 62.8015824, 57.729788, 52.6321827, 47.5076912, 42.3552386, 37.1737496, 31.9621492, 26.7193622, 21.4443135, 16.1359281, 10.7931308, 5.41484646, 0, -5.64616978, -11.2279518, -16.7624524, -22.2667776, -27.7580337, -33.2533268, -38.7697632, -44.3244491, -49.9344907, -55.6169941, -61.3890656, -67.2678114, -73.2703376, -79.4137505, -85.7151562, -92.191661, -98.9128639, -105.90337, -113.061552, -120.415764, -127.994361, -135.825698, -143.938129, -152.360009, -161.119693, -170.245535, -179.76589, -189.709113, -200.103558, -210.89047, -222.217254, -234.138398, -246.708389, -259.960116, -273.958866, -288.775686, -304.471543, -321.107404, -338.744237, -357.443009, -376.766052, -397.424726, -419.57756, -443.383078, -468.999807, -496.859488, -526.77913, -558.848955, -593.159185, -629.800044, -668.247114, -709.956484, -754.975967, -803.755303, -857.564837, -917.306481, -983.882145, -1060.76288, -1145.69757, -1233.63236, -1317.2123, -1406.22411, -1508.76945, -1628.83824, -1763.45321, -1912.83514, -2074.3334, -2247.42431, -2442.11276, -2662.44809, -2921.64657, -3213.18031, -3537.92547, -3896.75821, -4290.5547, -4834.23223, -5386.11555, -5918.57057, -6403.96316, -6814.65921, -7071.96019, -7233.33935, -7293.04733, -7170.75575, -6890.2361, -6506.33449, -6041.26869, -5517.25647, -4956.51559, -4381.26384, -3834.66505, -3332.94793, -2872.90023, -2451.30966, -2064.96394, -1710.6508, -1385.15797, -1098.49334, -816.747593},
        {459.284623, 453.711419, 448.235842, 442.854574, 437.564295, 432.361686, 427.243427, 422.206201, 417.246687, 412.361566, 407.547519, 402.801227, 398.119371, 393.498632, 388.93569, 384.427226, 379.969922, 375.560457, 371.195513, 366.871771, 362.585912, 358.334616, 354.114563, 349.922436, 345.754915, 341.60868, 337.480413, 333.366794, 329.264504, 325.170224, 321.080635, 316.992418, 312.902253, 308.806821, 304.603643, 300.341064, 296.068624, 291.786222, 287.493757, 283.191128, 278.878233, 274.55497, 270.22124, 265.876941, 261.521971, 257.15623, 252.779616, 248.392028, 243.993364, 239.583524, 235.162407, 230.710069, 226.246681, 221.772572, 217.288071, 212.793507, 208.289209, 203.775506, 199.252727, 194.721201, 190.181257, 185.633223, 181.077429, 176.514204, 171.943876, 167.366775, 162.78323, 158.211751, 153.670414, 149.122047, 144.565802, 140.000828, 135.426275, 130.841295, 126.245036, 121.63665, 117.015286, 112.380095, 107.730227, 103.064833, 98.3830616, 93.6840646, 88.9669917, 84.2309932, 79.4749897, 74.6982512, 69.9000502, 65.0795446, 60.2358919, 55.36825, 50.4757764, 45.557629, 40.6129654, 35.6409434, 30.6407206, 25.6114548, 20.5523037, 15.4624249, 10.3409762, 5.18711536, 0, -5.47359405, -10.8681705, -16.2012908, -21.4905163, -26.7534083, -32.0075281, -37.2704372, -42.5596968, -47.8928684, -53.2875133, -58.7611929, -64.3314685, -70.0159015, -75.8320532, -81.797485, -87.9297582, -94.2706696, -100.841243, -107.604221, -114.582359, -121.798411, -129.275132, -137.035276, -145.101599, -153.496855, -162.243798, -171.365184, -180.883766, -190.8223, -201.086164, -211.858526, -223.197355, -235.160618, -247.905318, -261.439906, -275.74642, -290.85312, -306.788261, -323.5801, -341.256894, -359.047543, -378.022943, -398.36775, -420.266614, -443.904191, -470.222546, -498.459567, -528.610552, -560.670801, -594.635616, -629.785979, -667.70456, -708.882005, -753.791953, -802.873312, -856.580584, -915.368269, -980.994537, -1052.31393, -1126.75882, -1200.65169, -1278.83162, -1365.42075, -1462.36714, -1567.93506, -1682.07971, -1802.91262, -1929.911, -2069.31212, -2223.52937, -2413.17972, -2619.67245, -2839.1197, -3067.63365, -3301.32644, -3503.07364, -3710.53316, -3928.12628, -4160.27432, -4411.39858, -4759.79629, -5086.7622, -5365.05649, -5442.6309, -5341.64897, -5136.27772, -4846.08051, -4490.62069, -4089.4616, -3662.16659, -3209.37647, -2806.28425, -2445.20744, -2118.46353, -1818.37003, -1537.24444, -1267.40428, -1044.83788, -847.172309},
        {434.112712, 429.565752, 425.02971, 420.504828, 415.991345, 411.489504, 406.999544, 402.521707, 398.056233, 393.603364, 389.163339, 384.7364, 380.322788, 375.922743, 371.536507, 367.164319, 362.806421, 358.463054, 354.134458, 349.820874, 345.522544, 341.239707, 336.972605, 332.721478, 328.486568, 324.268115, 320.06636, 315.881543, 311.713906, 307.563689, 303.431133, 299.316479, 295.219968, 291.141841, 287.169183, 283.256987, 279.3592, 275.473244, 271.596542, 267.726516, 263.860589, 259.996183, 256.13072, 252.261624, 248.386317, 244.502221, 240.606758, 236.697352, 232.771424, 228.826398, 224.859695, 220.721497, 216.55966, 212.374798, 208.167525, 203.938456, 199.688204, 195.417386, 191.126614, 186.816504, 182.487669, 178.140725, 173.776285, 169.394964, 164.997376, 160.584136, 156.155858, 151.716601, 147.27034, 142.810587, 138.337733, 133.852168, 129.354284, 124.844473, 120.323124, 115.790631, 111.247382, 106.69377, 102.130186, 97.5570209, 92.9746658, 88.3835117, 83.7839499, 79.1763715, 74.6303081, 70.1101264, 65.5793619, 61.036162, 56.478674, 51.9050454, 47.3134235, 42.7019558, 38.0687895, 33.4120722, 28.7299511, 24.0205737, 19.2820873, 14.5126394, 9.71037728, 4.87344835, 0, -5.23209552, -10.3891728, -15.4882938, -20.5465204, -25.5809146, -30.6085383, -35.6464535, -40.7117221, -45.8214061, -50.9925673, -56.2422678, -61.5875694, -67.0455342, -72.633224, -78.3677008, -84.2660265, -90.3626907, -96.6772449, -103.187332, -109.913747, -116.877288, -124.098751, -131.598932, -139.398627, -147.518633, -155.979747, -164.802765, -174.008483, -183.617697, -193.559212, -203.979548, -214.927097, -226.450255, -238.703174, -251.681369, -265.352153, -279.732193, -294.838156, -310.686708, -327.294517, -343.868308, -361.481192, -380.308304, -400.524777, -422.305744, -445.581921, -470.833964, -498.298111, -528.2106, -560.80767, -596.936692, -635.47583, -675.819585, -717.748516, -761.831382, -808.283052, -857.318401, -908.25966, -962.417214, -1022.07524, -1090.89822, -1164.76947, -1239.90306, -1314.85313, -1392.13893, -1472.45969, -1557.24048, -1647.36871, -1741.07022, -1838.0764, -1946.45413, -2056.3169, -2164.51068, -2267.88149, -2363.27532, -2361.29241, -2366.58597, -2397.56344, -2472.63226, -2610.19988, -2949.45463, -3307.50246, -3650.98759, -3820.24927, -3831.9274, -3755.28898, -3604.34254, -3393.09667, -3135.55991, -2845.74082, -2487.30045, -2173.32562, -1894.26791, -1640.57888, -1402.7101, -1171.11314, -936.239567, -732.773861, -559.416359},
        {408.975106, 404.924904, 400.899172, 396.896763, 392.916528, 388.957319, 385.017987, 381.097384, 377.194361, 373.307771, 369.436464, 365.579292, 361.735107, 357.90276, 354.081104, 350.268989, 346.465268, 342.668791, 338.878411, 335.092979, 331.311346, 327.532365, 323.754886, 319.977762, 316.199844, 312.419984, 308.637033, 304.849843, 301.057265, 297.258151, 293.451353, 289.635722, 285.81011, 281.973368, 278.069998, 274.127169, 270.172704, 266.207221, 262.231332, 258.245655, 254.250804, 250.247395, 246.236043, 242.217364, 238.191973, 234.160485, 230.123516, 226.081681, 222.035596, 217.985875, 213.933135, 209.94595, 205.955503, 201.960936, 197.961391, 193.95601, 189.943936, 185.924311, 181.896276, 177.858975, 173.811548, 169.753139, 165.68289, 161.599942, 157.503438, 153.39252, 149.26633, 145.103232, 140.88209, 136.644899, 132.392147, 128.124324, 123.841919, 119.545421, 115.235319, 110.912102, 106.576259, 102.22828, 97.8686534, 93.4978683, 89.116414, 84.7247794, 80.3234538, 75.9129262, 71.5633899, 67.2390157, 62.9031368, 58.5539803, 54.1897732, 49.8087427, 45.4091157, 40.9891194, 36.5469809, 32.0809271, 27.5891853, 23.0699824, 18.5215455, 13.9421018, 9.32987825, 4.68310196, 0, -5.01479186, -9.96138683, -14.8559964, -19.7148322, -24.5541056, -29.3900282, -34.2388114, -39.1166669, -44.0398062, -49.0244406, -54.0867819, -59.2430414, -64.5094307, -69.9021613, -75.4374448, -81.1314926, -87.0184916, -93.1172213, -99.4032346, -105.896595, -112.617366, -119.58561, -126.821392, -134.344774, -142.17582, -150.334593, -158.841157, -167.715575, -176.97791, -186.599507, -196.667011, -207.215103, -218.278462, -229.776332, -241.801111, -254.475978, -267.870241, -282.05321, -297.094196, -313.062507, -330.140205, -348.249533, -367.437738, -387.75207, -409.239779, -430.805523, -453.92479, -478.930476, -506.155478, -535.932693, -569.378063, -605.084383, -643.026761, -683.034311, -724.638082, -767.50295, -811.29379, -853.596749, -896.627867, -944.870885, -1003.96933, -1067.90375, -1132.86913, -1196.69301, -1259.96916, -1322.02158, -1383.2635, -1443.30134, -1497.74769, -1544.47429, -1586.36361, -1615.50574, -1628.03791, -1620.09735, -1587.82128, -1367.03216, -1154.26069, -985.722809, -897.634445, -926.211525, -1298.13965, -1732.18529, -2182.93453, -2450.38986, -2550.32669, -2562.93023, -2500.75538, -2376.35704, -2202.29013, -1991.10955, -1771.87556, -1536.55129, -1293.52538, -1051.18646, -817.923188, -602.124207, -412.17816, -228.490399, -65.6287844},
        {385.836919, 382.142754, 378.485238, 374.862283, 371.271796, 367.711687, 364.179867, 360.674245, 357.192731, 353.733233, 350.293662, 346.871928, 343.465939, 340.073607, 336.692839, 333.321546, 329.957638, 326.599023, 323.243612, 319.889315, 316.53404, 313.175698, 309.812198, 306.44145, 303.061363, 299.669847, 296.264811, 292.844166, 289.405821, 285.947684, 282.467667, 278.963679, 275.433629, 271.875427, 268.16126, 264.354543, 260.520202, 256.660228, 252.77661, 248.871337, 244.946398, 241.003784, 237.045484, 233.073488, 229.089784, 225.096363, 221.095214, 217.088326, 213.07769, 209.065295, 205.05313, 201.17162, 197.291535, 193.41208, 189.532462, 185.651884, 181.769553, 177.884674, 173.996452, 170.104093, 166.206801, 162.303782, 158.394242, 154.477386, 150.552418, 146.618545, 142.674972, 138.711401, 134.717761, 130.712859, 126.696515, 122.668551, 118.628788, 114.577048, 110.513152, 106.436921, 102.348178, 98.2467425, 94.1324367, 90.005082, 85.8644997, 81.7105113, 77.542938, 73.3616012, 69.2005095, 65.0416716, 60.8666853, 56.6742627, 52.4631157, 48.2319563, 43.9794963, 39.7044478, 35.4055228, 31.0814331, 26.7308908, 22.3526077, 17.9452959, 13.5076672, 9.03843372, 4.53630733, 0, -4.75784138, -9.46360207, -14.1318829, -18.7772847, -23.4144084, -28.0578546, -32.7222244, -37.4221184, -42.1721376, -46.9868827, -51.8809547, -56.8689543, -61.9654823, -67.1851397, -72.5425271, -78.0522456, -83.7771402, -89.7387038, -95.8424132, -102.113207, -108.576025, -115.255805, -122.177486, -129.366007, -136.846308, -144.643327, -152.782003, -161.287274, -170.184081, -179.514085, -189.27937, -199.499857, -210.195468, -221.125611, -232.440464, -244.359676, -256.981323, -270.403481, -284.724225, -300.041632, -317.057973, -335.083241, -354.097301, -374.080016, -395.011249, -416.201502, -438.467341, -461.955968, -486.814588, -513.190403, -541.151381, -571.020807, -603.587189, -639.115943, -676.804499, -716.32979, -757.368752, -799.315356, -842.193807, -886.336991, -931.790563, -978.41149, -1028.58778, -1082.0118, -1133.80831, -1181.19865, -1221.82097, -1253.00469, -1270.55102, -1271.12562, -1265.77441, -1234.57042, -1169.20152, -1061.3556, -902.720506, -395.793862, 106.248638, 539.421559, 839.739465, 943.21692, 479.557862, -104.705434, -742.282421, -1146.25086, -1336.25774, -1423.46327, -1422.93881, -1349.75567, -1218.98521, -1045.69876, -912.513651, -729.716774, -515.844935, -289.434941, -69.0235994, 126.852283, 279.655898, 437.343758, 577.080712},
        {365.912682, 363.112049, 360.249951, 357.328864, 354.351262, 351.319619, 348.23641, 345.10411, 341.925194, 338.702136, 335.437412, 332.133496, 328.792862, 325.417985, 322.011341, 318.575404, 315.112648, 311.625548, 308.11658, 304.588217, 301.042935, 297.483207, 293.91151, 290.330318, 286.742105, 283.149346, 279.554516, 275.96009, 272.368542, 268.782348, 265.203981, 261.635916, 258.080629, 254.540594, 251.197258, 247.959845, 244.735428, 241.520674, 238.31225, 235.106822, 231.901058, 228.691623, 225.475185, 222.248411, 219.007966, 215.750519, 212.472735, 209.171281, 205.842825, 202.484032, 199.09157, 195.421355, 191.716023, 187.977461, 184.207552, 180.408184, 176.581242, 172.728611, 168.852177, 164.953827, 161.035445, 157.098918, 153.146131, 149.17897, 145.199321, 141.209069, 137.2101, 133.236029, 129.31971, 125.397588, 121.469493, 117.535253, 113.594698, 109.647656, 105.693955, 101.733425, 97.7658951, 93.7911932, 89.8091486, 85.81959, 81.8223463, 77.8172463, 73.8041188, 69.7827928, 65.7994462, 61.8297584, 57.8488516, 53.8550505, 49.8466799, 45.8220644, 41.7795287, 37.7173976, 33.6339958, 29.527648, 25.3966788, 21.239413, 17.0541754, 12.8392905, 8.59308321, 4.31387813, 0, -4.44417937, -8.86134847, -13.2640317, -17.6647533, -22.0760378, -26.5104095, -30.9803927, -35.4985118, -40.0772912, -44.7292552, -49.4669283, -54.3028347, -59.2494988, -64.3194451, -69.5251977, -74.8792813, -80.4912529, -86.387499, -92.3610645, -98.4452665, -104.673422, -111.078849, -117.694863, -124.554783, -131.691924, -139.139606, -146.931143, -155.099855, -163.679057, -172.781253, -182.331539, -192.339478, -202.814629, -213.638206, -224.883945, -236.659808, -249.01386, -261.994169, -275.6488, -290.025819, -305.052785, -320.934947, -337.74395, -355.551436, -374.429049, -395.129798, -416.873621, -439.561819, -463.095696, -487.376553, -511.74843, -537.350993, -564.280049, -592.788742, -623.451449, -656.698321, -692.959506, -735.349402, -781.003855, -824.130445, -859.084942, -890.11089, -921.985227, -955.470888, -984.529294, -1006.24387, -1019.26499, -1021.08231, -1003.44007, -961.332451, -917.752607, -835.388208, -699.541484, -495.514661, -208.609968, 615.153168, 1441.36902, 2174.91466, 2720.66715, 2983.50358, 2435.23364, 1702.51335, 875.820095, 333.273359, 51.4921791, -112.755298, -176.857497, -158.20284, -74.1797503, 57.8233477, 209.591802, 364.561406, 520.758863, 676.210875, 828.944143, 976.985372, 1118.36126, 1249.26349, 1358.94984},
        {346.268512, 344.166001, 341.947232, 339.616997, 337.180087, 334.641292, 332.005403, 329.277212, 326.461509, 323.563086, 320.586733, 317.537242, 314.419403, 311.238007, 307.997846, 304.70371, 301.360391, 297.972679, 294.545366, 291.083242, 287.591098, 284.073726, 280.535916, 276.98246, 273.418147, 269.847771, 266.27612, 262.707987, 259.148162, 255.601437, 252.072602, 248.566448, 245.087767, 241.641348, 238.548983, 235.650294, 232.781096, 229.935892, 227.109187, 224.295485, 221.489289, 218.685102, 215.87743, 213.060775, 210.229641, 207.378533, 204.501953, 201.594406, 198.650396, 195.664426, 192.631, 189.164687, 185.648162, 182.084165, 178.475436, 174.824715, 171.134743, 167.408258, 163.648002, 159.856714, 156.037134, 152.192002, 148.324059, 144.436045, 140.530698, 136.61076, 132.678971, 128.776423, 124.943288, 121.103207, 117.256433, 113.40322, 109.543822, 105.678493, 101.807487, 97.9310574, 94.0494578, 90.1629423, 86.2717645, 82.3761783, 78.4764376, 74.572796, 70.6655074, 66.7548256, 62.9111702, 59.0982362, 55.2788754, 51.4510645, 47.6127803, 43.7619996, 39.8966992, 36.0148557, 32.1144461, 28.1934471, 24.2498354, 20.281588, 16.2866814, 12.2630925, 8.20879816, 4.12177506, 0, -4.24015241, -8.46078817, -12.6737278, -16.8907917, -21.1238004, -25.3845744, -29.6849342, -34.0367003, -38.4516932, -42.9417333, -47.5186411, -52.1942372, -56.980342, -61.888776, -66.9313597, -72.1199136, -77.5564362, -83.2656306, -89.0553431, -94.9567181, -101.0009, -107.219033, -113.642261, -120.30173, -127.228583, -134.453964, -142.009018, -149.92489, -158.232723, -167.012884, -176.229247, -185.898191, -196.036093, -206.672491, -217.817182, -229.476456, -241.662744, -254.388475, -267.666079, -281.507987, -295.409617, -310.057763, -325.566009, -342.04794, -359.617141, -379.513321, -400.442409, -422.23646, -444.727528, -467.747666, -490.629618, -514.315017, -538.897165, -564.533032, -591.509579, -620.055406, -650.399112, -684.36274, -720.219298, -754.503489, -785.830416, -812.422927, -832.31099, -843.871142, -846.067159, -837.593267, -821.622564, -798.010463, -750.189841, -670.883089, -594.53777, -465.73185, -262.74439, 36.1455448, 452.658892, 1598.66564, 2758.20042, 3805.44688, 4614.58871, 5059.80958, 4462.53706, 3618.21498, 2637.92249, 1977.33327, 1611.12445, 1370.39212, 1236.99247, 1192.78171, 1219.61603, 1299.35161, 1455.79748, 1577.65517, 1677.0063, 1765.93251, 1856.51541, 1960.83666, 2090.97787, 2173.42335, 2263.06813},
        {326.197572, 324.372358, 322.462156, 320.469717, 318.397792, 316.249132, 314.026487, 311.732609, 309.370247, 306.942154, 304.45108, 301.899775, 299.290991, 296.627478, 293.911988, 291.14727, 288.336076, 285.481157, 282.585263, 279.651146, 276.681556, 273.679243, 270.64696, 267.587456, 264.503482, 261.39779, 258.27313, 255.132252, 251.977909, 248.81285, 245.639826, 242.461589, 239.280888, 236.100476, 233.069435, 230.114272, 227.159734, 224.203825, 221.244545, 218.279897, 215.307883, 212.326505, 209.333765, 206.327664, 203.306206, 200.267392, 197.209225, 194.129705, 191.026836, 187.898619, 184.743056, 181.428804, 178.086015, 174.715494, 171.318047, 167.894481, 164.445601, 160.972213, 157.475124, 153.95514, 150.413065, 146.849707, 143.265872, 139.662365, 136.039992, 132.39956, 128.741874, 125.067307, 121.376238, 117.670372, 113.950541, 110.217582, 106.472326, 102.71561, 98.9482659, 95.1711293, 91.385034, 87.5908142, 83.789304, 79.9813376, 76.1677492, 72.3493729, 68.5270431, 64.7015938, 60.9614584, 57.2618289, 53.5568446, 49.8444969, 46.1227772, 42.3896768, 38.6431873, 34.8813001, 31.1020065, 27.303298, 23.483166, 19.6396019, 15.7705972, 11.8741432, 7.94823138, 3.99085318, 0, -4.18328316, -8.33099326, -12.4559858, -16.5711164, -20.6892404, -24.8232135, -28.9858911, -33.1901288, -37.448782, -41.7747064, -46.1807575, -50.6797907, -55.2846616, -60.0082257, -64.8633385, -69.8628556, -75.0399906, -80.4105074, -85.9412132, -91.6493261, -97.5520641, -103.666645, -110.010287, -116.600208, -123.453626, -130.587759, -138.019825, -145.767042, -153.846628, -162.209723, -170.963852, -180.146055, -189.793374, -199.994854, -210.761535, -222.090587, -234.00345, -246.521563, -259.666365, -273.459298, -287.807216, -302.881017, -318.724557, -335.381696, -352.89629, -371.499458, -391.000983, -411.397909, -432.687278, -454.866134, -478.025116, -501.953276, -527.745056, -555.463966, -583.068704, -609.462984, -633.550522, -650.591743, -663.961673, -681.009836, -713.780008, -745.509914, -760.555496, -752.051617, -728.008536, -689.608461, -646.327398, -601.497798, -528.041523, -416.082381, -308.155748, -133.910553, 134.671855, 525.610131, 1066.92293, 2504.29076, 3968.65495, 5308.61869, 6372.78517, 7009.75756, 6422.6511, 5535.88225, 4474.69188, 3742.8086, 3315.65523, 3010.95141, 2812.00511, 2702.12432, 2664.61705, 2682.79127, 2804.17019, 2883.09019, 2935.50806, 2977.3806, 3024.66461, 3093.31689, 3199.29424, 3209.76559, 3284.0173},
        {308.114675, 306.589172, 304.988845, 303.315623, 301.571435, 299.758212, 297.877882, 295.932376, 293.923623, 291.853551, 289.724092, 287.537175, 285.294728, 282.998683, 280.650967, 278.253512, 275.808245, 273.317098, 270.781999, 268.204878, 265.587665, 262.932289, 260.24068, 257.514768, 254.756481, 251.967749, 249.150503, 246.306671, 243.438183, 240.546969, 237.634959, 234.704081, 231.756266, 228.793442, 225.892809, 223.017079, 220.128058, 217.225234, 214.308093, 211.376123, 208.428811, 205.465642, 202.486105, 199.489686, 196.475872, 193.44415, 190.394008, 187.324931, 184.236407, 181.127923, 177.998966, 174.821412, 171.622957, 168.403686, 165.163686, 161.903041, 158.621838, 155.320161, 151.998097, 148.65573, 145.293147, 141.910432, 138.507671, 135.084951, 131.642355, 128.17997, 124.697881, 121.180159, 117.611258, 114.024294, 110.420391, 106.800672, 103.166261, 99.5182823, 95.8578586, 92.1861139, 88.504172, 84.8131567, 81.1141914, 77.4084, 73.6969062, 69.9808336, 66.261306, 62.539447, 58.9133261, 55.3335551, 51.7495803, 48.1593793, 44.5609298, 40.9522094, 37.3311956, 33.6958662, 30.0441986, 26.3741706, 22.6837596, 18.9709435, 15.2336996, 11.4700057, 7.67783944, 3.85517832, 0, -4.12352975, -8.19624771, -12.2319428, -16.2444041, -20.2474205, -24.2547809, -28.2802743, -32.3376896, -36.4408159, -40.603442, -44.839357, -49.1623497, -53.5862092, -58.1247244, -62.7916842, -67.6008777, -72.5473927, -77.6423468, -82.9418286, -88.4556197, -94.1935019, -100.165257, -106.380666, -112.84951, -119.581572, -126.586633, -133.874475, -141.454879, -149.337627, -157.422819, -165.870135, -174.72226, -184.02188, -193.882203, -204.310655, -215.295853, -226.859326, -239.022604, -251.807216, -265.23469, -279.428073, -294.276481, -309.781579, -325.945035, -342.768516, -359.569324, -377.204582, -395.847047, -415.669477, -436.84463, -460.399555, -484.608585, -510.267028, -537.110567, -562.711487, -585.613391, -604.359885, -612.29102, -614.336582, -621.102965, -649.280366, -676.367597, -681.052647, -654.329843, -608.04592, -544.476016, -476.424785, -408.897239, -308.290166, -162.839351, -19.3683953, 201.488246, 531.775188, 1003.53705, 1648.81844, 3290.6255, 4972.30094, 6528.149, 7792.4739, 8599.5799, 8095.38295, 7261.50106, 6227.26202, 5503.62073, 5068.84423, 4742.18823, 4509.87035, 4358.10816, 4273.11927, 4241.12127, 4391.59633, 4450.95044, 4450.31595, 4420.8252, 4393.61053, 4399.80428, 4470.53879, 4412.10702, 4436.04124},
        {291.464229, 290.067835, 288.621151, 287.124769, 285.57928, 283.985277, 282.343352, 280.654096, 278.9181, 277.135958, 275.308261, 273.435601, 271.518569, 269.557758, 267.553759, 265.507165, 263.418566, 261.288556, 259.117726, 256.906667, 254.655972, 252.366233, 250.038041, 247.671988, 245.268666, 242.828668, 240.352584, 237.841007, 235.294529, 232.713742, 230.099236, 227.451605, 224.771441, 222.059334, 219.304324, 216.513024, 213.692181, 210.842764, 207.965739, 205.062073, 202.132732, 199.178684, 196.200895, 193.200332, 190.177961, 187.13475, 184.071665, 180.989674, 177.889742, 174.772836, 171.639924, 168.543177, 165.431247, 162.303991, 159.161265, 156.002926, 152.828831, 149.638836, 146.432799, 143.210576, 139.972024, 136.717, 133.44536, 130.156961, 126.85166, 123.529313, 120.189778, 116.81757, 113.39757, 109.961278, 106.509546, 103.043223, 99.5631627, 96.070215, 92.5652315, 89.0490636, 85.5225626, 81.9865798, 78.4419664, 74.8895739, 71.3302534, 67.7648564, 64.1942341, 60.6192378, 57.1216786, 53.6602251, 50.1925727, 46.7169454, 43.2315672, 39.7346622, 36.2244544, 32.6991676, 29.1570261, 25.5962537, 22.0150744, 18.4117124, 14.7843915, 11.1313358, 7.45076934, 3.74091606, 0, -4.08932722, -8.10891844, -12.0736538, -15.9984136, -19.8980779, -23.787527, -27.681641, -31.5953001, -35.5433845, -39.5407744, -43.60235, -47.7429914, -51.9775788, -56.3209924, -60.7881125, -65.3938191, -70.1235676, -74.9880346, -80.0686565, -85.3740084, -90.912665, -96.6932012, -102.724192, -109.014212, -115.571836, -122.405639, -129.524195, -136.93608, -144.649869, -152.606903, -160.907642, -169.580832, -178.655216, -188.125662, -198.037854, -208.446509, -219.390535, -230.90884, -243.040332, -255.823919, -269.82999, -284.404217, -299.481522, -314.996828, -330.885058, -345.208169, -360.24229, -376.390585, -394.056217, -413.642349, -437.316156, -461.56078, -485.758761, -509.154523, -530.710498, -549.51573, -564.659258, -572.43811, -575.367891, -579.010042, -594.307075, -604.520358, -593.501343, -554.066084, -494.220654, -414.998659, -327.68228, -235.962161, -105.950833, 74.9828668, 262.282209, 534.564386, 924.256367, 1463.78512, 2185.57763, 3907.65551, 5680.45242, 7339.99666, 8722.31652, 9663.44032, 9337.97288, 8684.31496, 7821.96207, 7212.63001, 6840.35207, 6546.56981, 6322.44611, 6159.14389, 6047.82605, 5979.65548, 6203.83759, 6264.28497, 6212.3471, 6099.37345, 5976.7135, 5895.71673, 5907.73261, 5783.00016, 5714.67433},
        {276.767662, 275.443116, 274.086555, 272.697689, 271.276228, 269.821881, 268.334357, 266.813365, 265.258615, 263.669816, 262.046677, 260.388909, 258.696219, 256.968317, 255.204914, 253.405717, 251.570437, 249.698782, 247.790463, 245.845188, 243.862666, 241.842607, 239.784721, 237.688716, 235.554303, 233.381189, 231.169085, 228.9177, 226.626744, 224.295924, 221.924952, 219.513536, 217.061386, 214.56821, 211.975579, 209.313495, 206.612659, 203.874666, 201.101112, 198.293594, 195.453708, 192.583049, 189.683214, 186.755798, 183.802399, 180.824611, 177.824032, 174.802256, 171.76088, 168.701501, 165.625713, 162.603827, 159.567235, 156.516044, 153.450362, 150.370293, 147.275945, 144.167424, 141.044837, 137.90829, 134.75789, 131.593743, 128.415955, 125.224634, 122.019885, 118.801816, 115.570532, 112.321491, 109.050262, 105.76654, 102.470733, 99.1632491, 95.8444955, 92.5148804, 89.1748116, 85.824697, 82.4649446, 79.0959621, 75.7181575, 72.3319387, 68.9377135, 65.5358897, 62.1268754, 58.7110783, 55.3477985, 52.0067592, 48.6569754, 45.296944, 41.9251616, 38.5401249, 35.1403307, 31.7242757, 28.2904565, 24.8373699, 21.3635126, 17.8673814, 14.3474728, 10.8022837, 7.23031069, 3.63005056, 0, -3.96983052, -7.86895704, -11.7119983, -15.5135729, -19.2882996, -23.0507972, -26.8156842, -30.5975795, -34.4111016, -38.2708694, -42.1915014, -46.1876164, -50.2738332, -54.4647703, -58.7750465, -63.2192806, -67.8054061, -72.5470868, -77.4740622, -82.5995186, -87.9366421, -93.4986189, -99.2986352, -105.349877, -111.665531, -118.258783, -125.14282, -132.330827, -139.83599, -147.694514, -155.888128, -164.423112, -173.305747, -182.35911, -191.681085, -201.418408, -211.632323, -222.38407, -233.734894, -245.746035, -259.414116, -273.580319, -288.122876, -302.920021, -317.849987, -330.195397, -343.078996, -357.027922, -372.569309, -390.230294, -412.817052, -435.792187, -457.29052, -476.222831, -493.08416, -507.658248, -519.728834, -529.64797, -536.501923, -538.756991, -538.633394, -528.745352, -502.012387, -454.35528, -386.78027, -297.959289, -195.088694, -79.0504527, 80.5296995, 296.345734, 530.144432, 853.404985, 1296.45693, 1889.62981, 2663.25316, 4385.49461, 6164.38608, 7845.79758, 9275.59914, 10299.6608, 10157.361, 9705.38903, 9049.53962, 8591.87168, 8320.90898, 8101.73193, 7929.03647, 7797.5185, 7701.87395, 7636.79874, 7900.31017, 7969.63786, 7901.96195, 7754.46258, 7584.3199, 7448.71405, 7404.82517, 7220.67393, 7089.22202},
        {263.768306, 262.574345, 261.349658, 260.093904, 258.806744, 257.487836, 256.136839, 254.753414, 253.33722, 251.887916, 250.405162, 248.888617, 247.337941, 245.752793, 244.132832, 242.477719, 240.787112, 239.060671, 237.298056, 235.498925, 233.662939, 231.789757, 229.879038, 227.930442, 225.943628, 223.918256, 221.853985, 219.750474, 217.607384, 215.424373, 213.201101, 210.937228, 208.632413, 206.286315, 203.841177, 201.326575, 198.772774, 196.181298, 193.553668, 190.891407, 188.196039, 185.469086, 182.71207, 179.926515, 177.113943, 174.275878, 171.41384, 168.529355, 165.623943, 162.699129, 159.756434, 156.85815, 153.943715, 151.013332, 148.067209, 145.105549, 142.128558, 139.136443, 136.129407, 133.107657, 130.071398, 127.020836, 123.956174, 120.877621, 117.785379, 114.679656, 111.560655, 108.425523, 105.271477, 102.105035, 98.9266013, 95.7365787, 92.5353716, 89.3233835, 86.1010184, 82.8686801, 79.6267722, 76.3756987, 73.1158632, 69.8476697, 66.5715218, 63.2878235, 59.9969783, 56.6993903, 53.4532514, 50.2288558, 46.9958031, 43.7526217, 40.4978401, 37.2299868, 33.9475901, 30.6491785, 27.3332805, 23.9984246, 20.6431391, 17.2659526, 13.8653935, 10.4399902, 6.98827126, 3.50876503, 0, -3.73246667, -7.41796539, -11.0691226, -14.6985649, -18.3189187, -21.9428104, -25.5828666, -29.2517136, -32.9619781, -36.7262865, -40.5572652, -44.4675408, -48.4697397, -52.5764884, -56.8004133, -61.154141, -65.6710589, -70.3667593, -75.2069094, -80.2085844, -85.3888597, -90.7648103, -96.3535117, -102.172039, -108.237467, -114.566872, -121.177329, -128.085913, -135.309699, -142.934437, -150.883346, -159.152902, -167.739575, -176.387598, -185.219564, -194.425332, -204.077047, -214.246855, -225.0069, -236.42933, -249.683099, -263.40973, -277.466776, -291.711789, -306.00232, -317.746289, -329.863289, -342.823279, -357.09622, -373.152072, -393.629107, -414.178813, -433.190203, -449.617092, -463.566448, -474.627496, -482.389461, -486.861985, -487.118329, -481.773121, -472.731339, -453.178461, -417.648047, -362.229627, -285.649225, -185.422671, -66.7160414, 70.9714386, 256.191441, 501.628463, 774.58142, 1140.10279, 1626.58746, 2262.43032, 3076.02627, 4802.0364, 6586.02285, 8279.81395, 9735.23802, 10804.1234, 10759.5736, 10417.9584, 9879.13101, 9524.06373, 9341.33356, 9202.38452, 9101.65025, 9033.56436, 8992.56048, 8973.07222, 9180.78844, 9251.29953, 9220.9995, 9126.28234, 9003.54208, 8889.17272, 8819.56826, 8585.96512, 8474.88863},
        {251.80006, 250.723921, 249.614963, 248.472997, 247.297838, 246.0893, 244.847195, 243.571336, 242.261539, 240.917615, 239.539379, 238.126643, 236.679221, 235.196928, 233.679575, 232.126977, 230.538947, 228.915298, 227.255844, 225.560399, 223.828775, 222.060786, 220.256247, 218.414969, 216.536767, 214.621454, 212.668843, 210.678748, 208.650983, 206.58536, 204.481694, 202.339797, 200.159483, 197.940566, 195.640971, 193.282337, 190.886806, 188.45555, 185.989741, 183.490554, 180.95916, 178.396732, 175.804443, 173.183467, 170.534975, 167.86014, 165.160136, 162.436135, 159.689309, 156.920832, 154.131876, 151.360255, 148.569706, 145.760609, 142.933341, 140.088281, 137.225807, 134.346299, 131.450134, 128.537692, 125.609349, 122.665486, 119.70648, 116.73271, 113.744555, 110.742393, 107.726602, 104.696838, 101.652775, 98.5962815, 95.5277825, 92.4477037, 89.3564701, 86.2545073, 83.1422405, 80.0200951, 76.8884964, 73.7478697, 70.5986404, 67.4412337, 64.2760751, 61.1035899, 57.9242033, 54.7383408, 51.6046636, 48.4932541, 45.3734949, 42.2439214, 39.1030691, 35.9494733, 32.7816697, 29.5981936, 26.3975806, 23.178366, 19.9390853, 16.678274, 13.3944677, 10.0862016, 6.75201133, 3.39043231, 0, -3.4801297, -6.9411473, -10.393377, -13.8471432, -17.3127699, -20.8005815, -24.3209021, -27.8840561, -31.5003677, -35.180161, -38.9337604, -42.77149, -46.7036742, -50.7406371, -54.892703, -59.170196, -63.6336684, -68.3006196, -73.0677635, -77.9561873, -82.9869784, -88.1812241, -93.5600118, -99.1444288, -104.955562, -111.0145, -117.342329, -123.960137, -130.88901, -138.259, -145.942277, -153.927241, -162.202289, -170.45482, -178.823731, -187.52819, -196.646894, -206.258543, -216.441835, -227.275468, -240.008252, -253.192654, -266.67844, -280.315373, -293.953217, -305.430101, -317.110334, -329.346589, -342.49154, -356.897859, -374.763005, -392.340129, -408.503926, -422.349644, -433.422821, -441.066827, -444.625028, -443.574788, -437.095025, -424.218479, -406.718008, -378.271513, -335.115298, -273.504503, -189.72618, -80.0527345, 52.8139345, 210.934516, 419.944519, 692.1442, 1001.45509, 1404.46059, 1926.7459, 2593.89623, 3431.49678, 5124.70284, 6874.637, 8541.99194, 9987.46034, 11071.7349, 11113.0055, 10876.1361, 10454.3206, 10206.8783, 10121.6617, 10075.6379, 10062.2038, 10074.7561, 10106.6918, 10151.4077, 10285.778, 10363.6163, 10392.6868, 10380.7535, 10335.5804, 10264.9315, 10176.5709, 9910.60163, 9829.40073},
        {240.493549, 239.532799, 238.535194, 237.500807, 236.429712, 235.321983, 234.177692, 232.996912, 231.779717, 230.52618, 229.236375, 227.910374, 226.54825, 225.150078, 223.71593, 222.245879, 220.739999, 219.198362, 217.621043, 216.008114, 214.359649, 212.67572, 210.956401, 209.201766, 207.411887, 205.586838, 203.726691, 201.831521, 199.9014, 197.936402, 195.9366, 193.902066, 191.832875, 189.7291, 187.571481, 185.370165, 183.13553, 180.868276, 178.569104, 176.238715, 173.877809, 171.487086, 169.067248, 166.618995, 164.143027, 161.640045, 159.11075, 156.555842, 153.976022, 151.371991, 148.744448, 146.100168, 143.433647, 140.745454, 138.036156, 135.306324, 132.556526, 129.787332, 126.999309, 124.193027, 121.369056, 118.527963, 115.670318, 112.796689, 109.907646, 107.003758, 104.085593, 101.155172, 98.2144798, 95.2610925, 92.2954849, 89.3181318, 86.329508, 83.3300884, 80.3203477, 77.3007609, 74.2718028, 71.2339482, 68.187672, 65.1334489, 62.071754, 59.0030619, 55.9278475, 52.8465857, 49.8197749, 46.8166155, 43.8055868, 40.7852157, 37.7540291, 34.710554, 31.6533175, 28.5808463, 25.4916676, 22.3843083, 19.2572953, 16.1091555, 12.9384161, 9.74360381, 6.52324574, 3.27586882, 0, -3.24451756, -6.4955438, -9.76125201, -13.0498155, -16.3694076, -19.7282017, -23.1343709, -26.5960887, -30.1215284, -33.7188632, -37.3962665, -41.1619115, -45.0239717, -48.9906203, -53.0700306, -57.2703759, -61.6725286, -66.2950474, -70.9816676, -75.7561408, -80.6422187, -85.663653, -90.8441952, -96.2075972, -101.77761, -107.577987, -113.632478, -119.964835, -126.59881, -133.691694, -141.084736, -148.761623, -156.706048, -164.586326, -172.543833, -180.804046, -189.445266, -198.545794, -208.183931, -218.437979, -230.529721, -243.045961, -255.841277, -268.770245, -281.687443, -293.075945, -304.504707, -316.171181, -328.272822, -341.007081, -355.963309, -370.245854, -383.31084, -394.448522, -402.610508, -406.900452, -406.422006, -400.141094, -387.330402, -367.412873, -341.989874, -305.285038, -255.246977, -188.321554, -98.4083133, 19.4243982, 165.095843, 342.238164, 572.862277, 868.583022, 1210.27985, 1645.03153, 2195.09205, 2882.71544, 3730.15571, 5362.16851, 7047.8808, 8658.92121, 10066.9183, 11143.5008, 11260.2855, 11122.2539, 10815.3251, 10675.6669, 10690.1418, 10741.3425, 10821.3453, 10922.2263, 11036.0619, 11154.9283, 11212.5679, 11305.2511, 11408.3606, 11497.2792, 11547.3897, 11534.0748, 11432.7174, 11156.9974, 11099.9835},
        {229.479402, 228.641932, 227.763076, 226.843173, 225.882566, 224.881594, 223.840599, 222.759921, 221.639901, 220.48088, 219.283199, 218.047198, 216.773218, 215.4616, 214.112685, 212.726814, 211.304326, 209.845564, 208.350868, 206.820578, 205.255036, 203.654581, 202.019556, 200.350301, 198.647156, 196.910462, 195.14056, 193.337791, 191.502496, 189.635015, 187.735689, 185.80486, 183.842867, 181.850051, 179.829224, 177.77944, 175.699722, 173.590331, 171.451527, 169.283572, 167.086724, 164.861246, 162.607398, 160.32544, 158.015633, 155.678238, 153.313515, 150.921724, 148.503127, 146.057984, 143.586555, 141.067917, 138.523974, 135.955446, 133.363053, 130.747514, 128.109549, 125.449879, 122.769222, 120.068299, 117.34783, 114.608534, 111.851132, 109.076342, 106.284886, 103.477482, 100.654851, 97.8202594, 94.976915, 92.1202823, 89.2509162, 86.3693712, 83.4762021, 80.5719634, 77.6572099, 74.7324963, 71.7983772, 68.8554072, 65.9041411, 62.9451334, 59.9789389, 57.0061123, 54.0272081, 51.0427812, 48.1163252, 45.2156015, 42.3076147, 39.3908772, 36.4639009, 33.5251982, 30.5732812, 27.606662, 24.6238528, 21.6233658, 18.6037131, 15.563407, 12.5009595, 9.41488294, 6.30368937, 3.165891, 0, -3.05732821, -6.13819589, -9.24923808, -12.3970898, -15.5883862, -18.8297621, -22.1278528, -25.4892931, -28.9207183, -32.4287632, -36.0200629, -39.7012526, -43.4789672, -47.3598417, -51.3505113, -55.4576109, -59.7669334, -64.2964223, -68.8736649, -73.5222586, -78.265801, -83.1278895, -88.1321215, -93.3020944, -98.6614057, -104.233653, -110.042433, -116.111345, -122.463984, -129.25601, -136.330536, -143.671543, -151.263009, -158.807666, -166.430117, -174.329965, -182.575563, -191.235265, -200.377423, -210.070391, -221.387651, -233.08652, -245.038696, -257.115877, -269.189761, -280.512932, -291.730981, -302.870385, -313.957621, -325.019165, -336.974579, -347.865708, -357.690094, -365.941766, -371.086738, -372.052582, -367.766869, -356.915616, -338.722849, -312.676108, -279.943411, -235.479567, -178.87592, -106.822426, -11.0928011, 114.282603, 271.724292, 466.281765, 716.018054, 1031.635, 1400.57015, 1860.36876, 2429.78571, 3127.57586, 3972.49408, 5523.10801, 7123.40656, 8657.19141, 10008.2642, 11060.4267, 11244.0428, 11198.6438, 11002.3616, 10965.7812, 11075.0221, 11219.3485, 11389.7229, 11577.1076, 11772.4649, 11966.7572, 11958.4474, 12074.8666, 12259.3574, 12455.2627, 12605.925, 12654.6871, 12544.8918, 12287.5662, 12233.8622},
        {218.388245, 217.692276, 216.951334, 216.165935, 215.336599, 214.463843, 213.548185, 212.590143, 211.590237, 210.548982, 209.466898, 208.344503, 207.182315, 205.980851, 204.740629, 203.462169, 202.145988, 200.792603, 199.402533, 197.976297, 196.514411, 195.017394, 193.485765, 191.920041, 190.32074, 188.68838, 187.02348, 185.326557, 183.598129, 181.838715, 180.048833, 178.229, 176.379735, 174.501555, 172.61072, 170.699546, 168.76016, 166.79257, 164.796783, 162.772806, 160.720646, 158.640311, 156.531809, 154.395145, 152.230328, 150.037364, 147.816261, 145.567027, 143.289668, 140.984191, 138.650604, 136.253529, 133.829125, 131.378167, 128.901429, 126.399686, 123.873711, 121.324279, 118.752165, 116.158142, 113.542986, 110.90747, 108.252368, 105.578456, 102.886507, 100.177296, 97.4515962, 94.7118366, 91.9604046, 89.1946646, 86.415284, 83.6229297, 80.8182691, 78.0019691, 75.1746971, 72.3371201, 69.4899052, 66.6337197, 63.7692307, 60.8971053, 58.0180107, 55.132614, 52.2415825, 49.3455832, 46.5120545, 43.7068735, 40.8951147, 38.0752786, 35.2458653, 32.4053752, 29.5523086, 26.6851658, 23.802447, 20.9026527, 17.984283, 15.0458383, 12.0858189, 9.10272505, 6.09505709, 3.06131531, 0, -2.95025959, -5.9261446, -8.93382579, -11.9794739, -15.0692598, -18.2093541, -21.4059276, -24.6651512, -27.9931955, -31.3962313, -34.8804294, -38.4519606, -42.1169956, -45.8817052, -49.7522601, -53.7348311, -57.896177, -62.2511239, -66.6687984, -71.1683544, -75.7689457, -80.4897263, -85.34985, -90.3684708, -95.5647426, -100.957819, -106.566855, -112.411003, -118.509418, -124.975435, -131.699494, -138.672492, -145.88533, -153.14439, -160.53283, -168.183011, -176.141188, -184.453614, -193.166544, -202.326232, -212.722188, -223.431202, -234.354108, -245.391741, -256.444936, -267.570174, -278.473732, -289.01753, -299.063491, -308.473537, -317.541376, -325.169409, -331.720838, -336.857412, -338.808739, -336.447428, -328.646089, -314.25307, -292.170756, -261.327994, -221.975094, -170.115629, -106.834964, -29.1487644, 72.8231793, 205.795758, 374.293892, 584.464703, 850.485195, 1181.99021, 1571.84039, 2049.02543, 2628.98666, 3327.16538, 4159.0029, 5616.19594, 7118.86655, 8563.39218, 9846.15029, 10863.5183, 11106.9065, 11147.6377, 11055.6468, 11112.5727, 11304.5509, 11529.5063, 11777.985, 12040.5326, 12307.6948, 12570.0176, 12520.7057, 12671.1257, 12937.014, 13234.1071, 13478.1413, 13584.8529, 13469.9785, 13264.722, 13178.262},
        {205.34996, 204.894268, 204.39336, 203.847584, 203.257289, 202.622823, 201.944536, 201.222775, 200.457889, 199.650227, 198.800138, 197.90797, 196.974071, 195.998792, 194.982479, 193.925482, 192.82815, 191.69083, 190.513873, 189.297625, 188.042437, 186.748656, 185.416631, 184.046711, 182.639245, 181.194581, 179.713067, 178.195053, 176.640887, 175.050918, 173.425494, 171.764964, 170.069676, 168.33998, 166.572143, 164.768639, 162.931994, 161.062688, 159.161201, 157.228016, 155.263613, 153.268473, 151.243077, 149.187906, 147.103442, 144.990165, 142.848557, 140.679099, 138.482271, 136.258555, 134.008431, 131.735524, 129.437104, 127.113584, 124.765376, 122.392894, 119.996551, 117.576759, 115.133931, 112.668481, 110.180821, 107.671364, 105.140523, 102.588712, 100.016342, 97.4238276, 94.8115808, 92.1706951, 89.4924872, 86.7965916, 84.0840254, 81.3558056, 78.6129494, 75.8564737, 73.0873959, 70.3067329, 67.5155018, 64.7147197, 61.9054039, 59.0885712, 56.2652389, 53.4364241, 50.6031438, 47.7664151, 45.0069659, 42.2842814, 39.5568885, 36.8232176, 34.0816991, 31.3307632, 28.5688403, 25.7943608, 23.0057549, 20.201453, 17.3798855, 14.5394826, 11.6786748, 8.79589222, 5.88956537, 2.95812452, 0, -3.11090118, -6.19650602, -9.26640702, -12.3301967, -15.3974675, -18.477812, -21.5808226, -24.7160919, -27.8932124, -31.1217765, -34.4113768, -37.7716057, -41.2120558, -44.7423196, -48.3719895, -52.1106581, -55.8987041, -59.7358325, -63.7881904, -68.0505389, -72.517639, -77.1842516, -82.0451379, -87.0950587, -92.328775, -97.7410478, -103.326638, -109.080307, -114.996816, -120.954107, -127.106592, -133.484079, -140.116373, -147.217784, -154.725865, -162.528973, -170.601563, -178.91809, -187.453011, -196.180782, -204.984383, -213.957584, -223.09274, -232.382202, -241.818324, -252.821634, -263.599265, -273.786527, -283.018728, -290.931178, -297.067136, -301.266465, -303.411861, -303.220667, -300.072635, -293.499087, -283.031349, -270.754708, -253.066076, -223.576221, -176.821328, -114.280138, -41.457989, 42.9422452, 151.744083, 292.480238, 472.271626, 698.544197, 980.233841, 1325.42234, 1728.24367, 2212.87324, 2792.22093, 3479.19661, 4286.71014, 5642.77861, 7041.42789, 8391.79104, 9603.00115, 10584.1913, 10892.5213, 11023.7463, 11037.845, 11173.5657, 11420.1242, 11692.2491, 11982.8807, 12284.9592, 12591.4247, 12895.2177, 12871.6778, 13086.8913, 13437.0531, 13818.358, 14127.0008, 14259.1764, 14111.0798, 14044.8547, 13808.0621},
        {194.679794, 194.354322, 193.985167, 193.572561, 193.116732, 192.617912, 192.07633, 191.492217, 190.865802, 190.197315, 189.486987, 188.735047, 187.941727, 187.107255, 186.231862, 185.315778, 184.359233, 183.362457, 182.32568, 181.249133, 180.133044, 178.977646, 177.783167, 176.549837, 175.277887, 173.967547, 172.619046, 171.232616, 169.808485, 168.346884, 166.848044, 165.312194, 163.739564, 162.130384, 160.472568, 158.772763, 157.037765, 155.268204, 153.464709, 151.62791, 149.758436, 147.856919, 145.923986, 143.960269, 141.966397, 139.942999, 137.890706, 135.810148, 133.701953, 131.566752, 129.405175, 127.228555, 125.026586, 122.799666, 120.548193, 118.272564, 115.973176, 113.650429, 111.304719, 108.936444, 106.546003, 104.133791, 101.700209, 99.245652, 96.7705192, 94.2752079, 91.7601159, 89.2167021, 86.6366399, 84.0387628, 81.424048, 78.7934728, 76.1480142, 73.4886496, 70.8163561, 68.1321109, 65.4368913, 62.7316743, 60.0174373, 57.2951575, 54.5658119, 51.8303779, 49.0898326, 46.3451533, 43.6727468, 41.0342889, 38.3905488, 35.7400559, 33.0813396, 30.4129291, 27.7333538, 25.0411432, 22.3348266, 19.6129333, 16.8739927, 14.1165341, 11.339087, 8.54018067, 5.71834448, 2.87210781, 0, -3.08460503, -6.1325423, -9.15384131, -12.1585316, -15.1566425, -18.1582037, -21.1732446, -24.2117947, -27.2838835, -30.3995405, -33.5687952, -36.8016771, -40.1082156, -43.4984403, -46.9823807, -50.5700662, -54.1926943, -57.8490328, -61.7274217, -65.820998, -70.1228986, -74.6262603, -79.3242201, -84.209915, -89.2764817, -94.5170573, -99.9247786, -105.492783, -111.214206, -116.964397, -122.89747, -129.0419, -135.426161, -142.309608, -149.605274, -157.164623, -164.946863, -172.911204, -181.016853, -189.223021, -197.340159, -205.521506, -213.755375, -222.030079, -230.333931, -240.007431, -249.348659, -258.007882, -265.635367, -271.881379, -276.370281, -278.809908, -279.285261, -277.510264, -272.441216, -263.374576, -249.606802, -232.21226, -208.305427, -173.061249, -122.998106, -57.2697896, 21.1419162, 114.34599, 233.077421, 384.111721, 573.774755, 808.72546, 1097.27146, 1446.78778, 1854.90307, 2340.23818, 2913.54192, 3585.56307, 4367.0504, 5634.35188, 6941.21734, 8206.99575, 9351.03612, 10292.6874, 10639.7176, 10830.7775, 10918.7507, 11114.6917, 11408.8749, 11725.6706, 12058.6485, 12401.3783, 12747.4299, 13090.373, 13086.9112, 13339.388, 13737.8987, 14172.5387, 14533.4035, 14710.5883, 14594.1885, 14610.4091, 14297.8028},
        {186.065989, 185.771469, 185.436806, 185.062139, 184.647604, 184.193339, 183.699483, 183.166174, 182.593549, 181.981746, 181.330903, 180.641158, 179.91265, 179.145514, 178.339891, 177.495917, 176.613731, 175.69347, 174.735272, 173.739276, 172.705618, 171.634437, 170.525872, 169.380059, 168.197136, 166.977242, 165.720515, 164.427091, 163.09711, 161.730709, 160.328027, 158.889199, 157.414366, 155.903664, 154.345679, 152.746568, 151.112627, 149.444369, 147.742306, 146.006953, 144.23882, 142.438422, 140.606272, 138.742881, 136.848762, 134.92443, 132.970395, 130.987172, 128.975273, 126.93521, 124.867497, 122.766972, 120.639945, 118.487052, 116.308928, 114.106211, 111.879535, 109.629537, 107.356851, 105.062115, 102.745964, 100.409033, 98.0519594, 95.6753779, 93.2799246, 90.8662355, 88.4349463, 85.9861561, 83.519977, 81.0381518, 78.541351, 76.0302453, 73.5055053, 70.9678016, 68.4178048, 65.8561856, 63.2836147, 60.7007625, 58.1082998, 55.5068972, 52.8972253, 50.2799547, 47.6557561, 45.0253001, 42.4489749, 39.8963362, 37.3362221, 34.7673652, 32.1884982, 29.5983538, 26.9956646, 24.3791634, 21.7475827, 19.0996553, 16.4341139, 13.7496911, 11.0451195, 8.31913201, 5.57046112, 2.79783956, 0, -2.91231846, -5.80742921, -8.69349831, -11.5786918, -14.4711758, -17.3791163, -20.3106793, -23.274031, -26.2773374, -29.3287645, -32.4364785, -35.6086452, -38.8534309, -42.1790016, -45.5935233, -49.1051621, -52.7195225, -56.4444045, -60.2937683, -64.2752311, -68.3964099, -72.6649219, -77.0883843, -81.6744142, -86.4306288, -91.3646452, -96.4840806, -101.796552, -107.309677, -113.061224, -119.017603, -125.177386, -131.539143, -138.143289, -144.967475, -151.978191, -159.161457, -166.50329, -173.98971, -181.606734, -189.710897, -197.804936, -205.802378, -213.616747, -221.16157, -228.617035, -235.56334, -241.847349, -247.315921, -251.815919, -255.710055, -257.698852, -258.402287, -257.447433, -252.434183, -241.872595, -224.272724, -197.020329, -160.005285, -114.343976, -63.3619209, 0.894935999, 82.0905601, 184.653791, 314.309502, 476.187669, 673.875812, 912.104009, 1201.258, 1548.52435, 1956.06913, 2436.87182, 2999.85686, 3653.94869, 4408.07175, 5596.85693, 6822.09562, 8011.28567, 9091.92491, 9991.51118, 10356.171, 10582.3544, 10718.1469, 10958.898, 11294.4011, 11653.0902, 12027.8267, 12411.4725, 12796.889, 13176.938, 13184.87, 13452.9312, 13874.0153, 14341.0161, 14746.8273, 14984.3427, 14946.4561, 14992.222, 14680.0418},
        {178.669995, 178.342757, 177.981427, 177.585999, 177.15647, 176.692833, 176.195086, 175.663222, 175.097238, 174.497128, 173.862887, 173.194511, 172.491996, 171.755335, 170.984525, 170.179561, 169.340438, 168.467152, 167.559697, 166.618068, 165.642262, 164.632273, 163.588097, 162.509728, 161.397162, 160.250395, 159.069421, 157.854235, 156.604834, 155.321212, 154.003364, 152.651286, 151.264973, 149.844421, 148.379575, 146.875667, 145.338048, 143.76704, 142.162965, 140.526142, 138.856895, 137.155543, 135.422409, 133.657814, 131.862078, 130.035524, 128.178473, 126.291246, 124.374165, 122.42755, 120.451724, 118.418468, 116.357262, 114.269045, 112.154758, 110.01534, 107.851732, 105.664873, 103.455703, 101.225163, 98.9741921, 96.7037305, 94.4147183, 92.1080955, 89.784802, 87.4457778, 85.0919629, 82.7342045, 80.3831118, 78.0191921, 75.6427431, 73.2540626, 70.8534483, 68.441198, 66.0176094, 63.5829802, 61.1376083, 58.6817914, 56.2158271, 53.7400134, 51.2546478, 48.7600283, 46.2564524, 43.7442181, 41.2652384, 38.798128, 36.3210015, 33.8328061, 31.3324892, 28.8189979, 26.2912795, 23.7482812, 21.1889503, 18.6122341, 16.0170798, 13.4024347, 10.7672459, 8.11046084, 5.43102665, 2.72789062, 0, -2.69043655, -5.39393703, -8.11602802, -10.8622361, -13.6380879, -16.4491099, -19.3008288, -22.1987711, -25.1484634, -28.1554324, -31.2252045, -34.3633065, -37.5752648, -40.8666061, -44.2428569, -47.7095438, -51.3690547, -55.2407534, -59.1110757, -63.0063041, -66.9527212, -70.9766095, -75.1042516, -79.3619301, -83.7759276, -88.3725265, -93.1780095, -98.2186592, -103.520758, -109.329463, -115.371929, -121.608775, -128.000622, -134.285187, -140.534542, -146.8802, -153.34965, -159.970385, -166.769896, -173.775673, -182.117238, -190.384651, -198.389788, -205.944526, -212.860742, -217.680818, -221.803498, -225.358029, -228.473661, -231.279644, -235.249903, -237.525517, -239.045066, -239.230881, -234.409611, -222.297803, -200.612002, -162.861234, -111.925822, -55.2761128, -3.42885521, 58.9026221, 142.077615, 252.87418, 392.002158, 562.9568, 766.585231, 1005.69615, 1292.80806, 1634.94704, 2037.8553, 2510.04034, 3058.97264, 3692.12273, 4416.96111, 5533.51863, 6683.56536, 7801.43172, 8821.44813, 9677.945, 10048.2463, 10295.0266, 10462.7617, 10736.3633, 11104.7181, 11498.1986, 11908.6175, 12327.7875, 12747.5215, 13159.632, 13176.2247, 13451.0151, 13883.4742, 14373.0727, 14819.2812, 15121.5707, 15179.4117, 15222.271, 14965.8835},
        {170.921216, 170.509457, 170.079226, 169.629938, 169.161011, 168.671859, 168.161901, 167.630552, 167.077228, 166.501346, 165.902322, 165.279573, 164.632514, 163.960563, 163.263136, 162.539648, 161.789517, 161.012158, 160.206988, 159.373424, 158.510881, 157.618777, 156.696527, 155.743548, 154.759255, 153.743067, 152.694398, 151.612666, 150.497286, 149.347674, 148.163249, 146.943424, 145.687618, 144.395246, 143.03052, 141.611199, 140.155465, 138.663878, 137.136996, 135.575378, 133.979582, 132.350169, 130.687695, 128.992722, 127.265806, 125.507507, 123.718384, 121.898996, 120.049901, 118.171658, 116.264827, 114.315758, 112.339527, 110.336999, 108.309041, 106.256521, 104.180304, 102.081258, 99.9602496, 97.8181456, 95.6558126, 93.4741176, 91.2739273, 89.0561086, 86.8215282, 84.5710529, 82.3055495, 80.0343005, 77.7663861, 75.4853166, 73.1914133, 70.8849975, 68.5663903, 66.2359131, 63.8938871, 61.5406335, 59.1764736, 56.8017287, 54.41672, 52.0217687, 49.6171961, 47.2033235, 44.7804721, 42.3489631, 39.9511104, 37.5653557, 35.169636, 32.7629098, 30.3441356, 27.9122719, 25.4662773, 23.0051102, 20.5277291, 18.0330926, 15.5201592, 12.9878874, 10.4352356, 7.86116253, 5.26462655, 2.6445862, 0, -2.65997828, -5.32351227, -7.99662917, -10.6853562, -13.3957206, -16.1337495, -18.9054701, -21.7169098, -24.5740956, -27.4830548, -30.4498146, -33.4804022, -36.5808448, -39.7571697, -43.015404, -46.3615749, -49.8601831, -53.525609, -57.2316187, -60.9967695, -64.8396184, -68.7787229, -72.83264, -77.019927, -81.359141, -85.8688394, -90.5675794, -95.473918, -100.606413, -106.161886, -111.915266, -117.83163, -123.876055, -129.815855, -135.71499, -141.690155, -147.765758, -153.966205, -160.315902, -166.839255, -174.945937, -182.853487, -190.31528, -197.084695, -202.915107, -204.837488, -206.008221, -206.861283, -207.830652, -209.350307, -214.462321, -217.804902, -218.38975, -214.947272, -205.633571, -188.862602, -163.048319, -124.387327, -74.0148726, -15.4851292, 45.3643289, 118.424963, 208.844017, 321.484579, 460.728254, 631.177653, 836.821996, 1082.10487, 1373.71897, 1717.08476, 2113.87529, 2573.83498, 3103.68151, 3710.13252, 4399.90568, 5440.09632, 6511.95002, 7557.09002, 8517.13956, 9333.72189, 9720.63875, 9999.21589, 10208.7144, 10512.3761, 10900.3864, 11311.2717, 11737.8003, 12172.7404, 12608.8601, 13038.9277, 13036.2744, 13324.1266, 13793.9497, 14337.2092, 14845.3702, 15209.8983, 15322.2587, 15340.5943, 15171.1257},
        {163.185019, 162.711551, 162.234533, 161.752799, 161.26518, 160.77051, 160.26762, 159.755344, 159.232513, 158.69796, 158.150518, 157.589018, 157.012294, 156.419178, 155.808502, 155.179099, 154.529801, 153.85944, 153.166849, 152.450861, 151.710308, 150.944022, 150.150836, 149.329582, 148.479092, 147.5982, 146.685737, 145.740536, 144.76143, 143.74725, 142.69683, 141.609002, 140.482597, 139.31645, 138.045593, 136.702102, 135.318815, 133.896635, 132.436466, 130.939209, 129.405768, 127.837046, 126.233944, 124.597367, 122.928217, 121.227396, 119.495807, 117.734354, 115.943939, 114.125465, 112.279834, 110.4196, 108.533762, 106.622971, 104.687877, 102.729131, 100.747382, 98.7432806, 96.7174775, 94.6706227, 92.6033664, 90.5163589, 88.4102505, 86.2856912, 84.1433315, 81.9838215, 79.8078114, 77.6181129, 75.4174857, 73.2021223, 70.972533, 68.7292278, 66.4727168, 64.2035102, 61.9221181, 59.6290507, 57.3248181, 55.0099303, 52.6848976, 50.35023, 48.0064377, 45.6540309, 43.2935195, 40.9254139, 38.6006613, 36.2934925, 33.9775317, 31.6516524, 29.3147278, 26.9656313, 24.6032362, 22.2264159, 19.8340438, 17.4249931, 14.9981372, 12.5523495, 10.0865032, 7.59947184, 5.0901286, 2.55734687, 0, -2.67949339, -5.34351187, -7.99931437, -10.6541598, -13.3153072, -15.9900154, -18.6855433, -21.4091499, -24.1680942, -26.969635, -29.8210312, -32.7295419, -35.702426, -38.7469423, -41.8703499, -45.0799075, -48.3639539, -51.7270448, -55.2192356, -58.8437308, -62.6037349, -66.5024524, -70.543088, -74.7288461, -79.0629314, -83.5485484, -88.1889016, -92.9871956, -97.9466349, -103.137005, -108.470516, -113.9304, -119.499885, -125.100597, -130.746569, -136.468262, -142.267388, -148.145661, -154.104791, -160.146491, -167.615527, -174.761802, -181.324256, -187.041832, -191.653469, -191.406127, -190.403725, -189.258197, -188.581479, -188.985509, -194.647081, -198.256221, -196.537331, -187.55789, -172.127491, -149.824577, -120.227588, -84.2639448, -39.8565217, 16.5434236, 87.2385702, 173.244583, 271.483086, 383.788662, 520.320632, 687.417127, 893.548147, 1145.60453, 1442.66026, 1788.21096, 2177.9077, 2624.29756, 3133.59159, 3712.00087, 4365.73648, 5332.09259, 6328.42638, 7303.17814, 8204.78816, 8981.69672, 9382.45964, 9688.65801, 9934.39687, 10259.3743, 10655.898, 11072.2789, 11503.0247, 11942.6431, 12385.6417, 12826.5283, 12833.8597, 13150.0423, 13660.6114, 14251.1024, 14807.0505, 15213.991, 15357.4594, 15365.9137, 15299.1542},
        {155.621618, 155.191911, 154.762868, 154.333139, 153.901375, 153.466225, 153.026339, 152.580367, 152.126959, 151.664765, 151.192435, 150.708619, 150.211968, 149.70113, 149.174756, 148.631495, 148.069999, 147.488916, 146.886897, 146.262592, 145.61465, 144.941723, 144.242458, 143.515508, 142.75952, 141.973147, 141.155036, 140.30384, 139.418206, 138.496786, 137.53823, 136.541186, 135.504306, 134.426239, 133.230639, 131.955232, 130.638643, 129.281958, 127.88626, 126.452633, 124.98216, 123.475925, 121.935012, 120.360504, 118.753486, 117.115041, 115.446253, 113.748205, 112.021982, 110.268667, 108.489343, 106.713304, 104.912813, 103.088341, 101.240362, 99.3693473, 97.4757692, 95.5601, 93.622812, 91.6643774, 89.6852684, 87.6859572, 85.6669161, 83.6286173, 81.5715331, 79.4961356, 77.4028971, 75.2889217, 73.1513941, 70.9977333, 68.8286299, 66.6447746, 64.4468577, 62.2355701, 60.0116022, 57.7756445, 55.5283878, 53.2705225, 51.0027393, 48.7257287, 46.4401813, 44.1467877, 41.8462385, 39.5392242, 37.2856538, 35.0553632, 32.8174763, 30.570762, 28.3139892, 26.0459266, 23.7653431, 21.4710075, 19.1616886, 16.8361553, 14.4931765, 12.1315208, 9.74995721, 7.3472545, 4.92218151, 2.47350707, 0, -2.62970738, -5.24120216, -7.8416257, -10.4381194, -13.0378247, -15.6478829, -18.2754355, -20.9276238, -23.6115892, -26.3344731, -29.103417, -31.9255621, -34.8080499, -37.7580218, -40.7826192, -43.8889834, -47.0340444, -50.2177705, -53.5608767, -57.0597447, -60.7107563, -64.5102934, -68.4547376, -72.5404709, -76.763875, -81.1213317, -85.6092227, -90.22393, -94.9618352, -99.8370752, -104.821767, -109.906965, -115.083725, -120.388307, -125.789165, -131.242696, -136.726393, -142.217751, -147.694263, -153.133422, -159.507717, -165.496823, -170.88356, -175.450748, -178.981209, -178.474253, -177.192089, -175.613414, -174.216927, -173.481325, -177.281904, -178.54937, -173.096889, -158.871969, -138.179738, -111.368838, -78.7879164, -44.3497228, -4.02877207, 50.0885376, 124.782919, 216.648326, 319.081809, 432.503848, 569.235739, 736.136105, 943.548481, 1199.23499, 1498.17975, 1842.59484, 2225.10049, 2658.97631, 3149.75464, 3702.96783, 4324.1482, 5224.95155, 6153.25591, 7063.06276, 7908.37359, 8643.18986, 9043.25747, 9359.67055, 9622.82584, 9954.46211, 10348.8199, 10762.0805, 11190.3873, 11629.884, 12076.7139, 12527.0205, 12637.4312, 13011.727, 13547.6881, 14143.0943, 14695.7258, 15103.3626, 15263.7847, 15318.3487, 15355.6689},
        {148.24968, 148.05729, 147.849241, 147.62479, 147.383194, 147.123712, 146.845602, 146.54812, 146.230526, 145.892076, 145.532028, 145.149641, 144.744172, 144.314878, 143.861019, 143.38185, 142.876631, 142.344618, 141.78507, 141.197245, 140.5804, 139.933792, 139.256681, 138.548323, 137.807976, 137.034899, 136.228348, 135.387582, 134.511858, 133.600435, 132.65257, 131.66752, 130.644544, 129.582899, 128.432191, 127.217549, 125.964696, 124.674503, 123.347839, 121.985571, 120.58857, 119.157705, 117.693844, 116.197856, 114.670611, 113.112979, 111.525826, 109.910024, 108.266441, 106.595946, 104.899408, 103.195709, 101.467314, 99.7147025, 97.9383533, 96.1387446, 94.3163552, 92.4716636, 90.6051484, 88.7172883, 86.8085618, 84.8794476, 82.9304243, 80.9619704, 78.9745645, 76.9686854, 74.9448115, 72.8995679, 70.8296722, 68.7435507, 66.6419317, 64.5255438, 62.3951152, 60.2513745, 58.09505, 55.9268702, 53.7475634, 51.5578582, 49.3584828, 47.1501657, 44.9336354, 42.7096202, 40.4788486, 38.2420489, 36.0624094, 33.9081148, 31.7465992, 29.5765641, 27.3967109, 25.2057411, 23.0023562, 20.7852576, 18.5531469, 16.3047254, 14.0386948, 11.7537564, 9.44861178, 7.12196234, 4.77250958, 2.39895497, 0, -2.40680797, -4.83167252, -7.27841765, -9.75086733, -12.2528456, -14.7881763, -17.3606836, -19.9741913, -22.6325235, -25.3395042, -28.0989574, -30.914707, -33.7905769, -36.7303913, -39.7379741, -42.8171493, -46.011878, -49.3317182, -52.6897072, -56.0982699, -59.5698312, -63.1168157, -66.7516484, -70.486754, -74.3345574, -78.3074833, -82.4179567, -86.6784023, -91.101245, -95.8476685, -100.726794, -105.706418, -110.754338, -115.847392, -120.948857, -126.019597, -131.024699, -135.929246, -140.698324, -145.297017, -149.980051, -154.334717, -158.269432, -161.692611, -164.512669, -166.968216, -168.554926, -169.098666, -168.425303, -166.360705, -163.963991, -158.320468, -147.339734, -129.903923, -106.88075, -78.2464466, -43.9772413, -5.94636379, 38.198092, 92.877486, 159.460317, 241.592792, 340.740375, 459.956912, 604.987684, 780.342557, 992.102394, 1245.18436, 1538.7453, 1875.20043, 2251.55282, 2676.46245, 3154.21797, 3689.10804, 4285.4213, 5134.05894, 6006.04392, 6859.01175, 7650.59795, 8338.43801, 8712.87416, 9010.36406, 9260.24033, 9578.00943, 9958.89352, 10362.209, 10784.9733, 11224.204, 11676.9183, 12140.1339, 12515.0861, 12995.5971, 13525.5527, 14048.8385, 14509.3405, 14850.9443, 15017.5357, 15218.9908, 15347.9575},
        {141.68755, 141.660844, 141.606473, 141.524174, 141.413685, 141.274741, 141.10708, 140.910437, 140.68455, 140.429156, 140.143991, 139.828791, 139.483294, 139.107236, 138.700354, 138.262384, 137.793063, 137.292128, 136.759315, 136.194362, 135.597004, 134.966979, 134.304022, 133.607872, 132.878264, 132.114936, 131.317623, 130.486063, 129.619991, 128.719146, 127.783263, 126.81208, 125.805332, 124.762757, 123.656111, 122.499709, 121.308203, 120.082238, 118.822458, 117.529508, 116.204032, 114.846676, 113.458084, 112.038901, 110.589771, 109.11134, 107.604251, 106.06915, 104.50668, 102.917488, 101.302217, 99.6672492, 98.0073679, 96.3230935, 94.6149463, 92.8834466, 91.1291148, 89.3524711, 87.554036, 85.7343296, 83.8938725, 82.0331847, 80.1527868, 78.2531991, 76.3349417, 74.3985352, 72.4444997, 70.4708008, 68.4754651, 66.4642823, 64.4379383, 62.3971192, 60.3425109, 58.2747992, 56.1946702, 54.1028098, 51.999904, 49.8866386, 47.7636998, 45.6317733, 43.4915452, 41.3437014, 39.1889278, 37.0279104, 34.921589, 32.8392551, 30.7494764, 28.6509835, 26.5425068, 24.4227771, 22.290525, 20.144481, 17.9833757, 15.8059397, 13.6109037, 11.3969983, 9.16295394, 6.90750135, 4.62937109, 2.32729378, 0, -2.28722344, -4.60298506, -6.95013205, -9.33151158, -11.7499708, -14.208357, -16.7095173, -19.2562989, -21.851549, -24.4981147, -27.1988433, -29.9565819, -32.7741778, -35.654478, -38.6003299, -41.6145805, -44.7287992, -47.9499364, -51.2158724, -54.5356094, -57.9181491, -61.3724934, -64.9076445, -68.532604, -72.2563741, -76.0879565, -80.0363533, -84.1105663, -88.3195975, -92.8067236, -97.3974377, -102.060459, -106.764507, -111.547722, -116.344113, -121.069177, -125.670807, -130.096898, -134.295343, -138.214034, -141.745152, -144.909261, -147.665155, -149.971627, -151.787472, -154.510539, -156.300803, -156.757294, -155.479042, -152.065077, -146.571406, -137.581556, -124.070707, -105.297772, -81.1009468, -51.0583422, -14.7480674, 28.1980415, 78.2678117, 136.007681, 197.41001, 272.572743, 370.289259, 495.879947, 648.780403, 831.127439, 1044.35143, 1290.40604, 1573.8352, 1897.71764, 2265.41576, 2679.68097, 3143.4938, 3659.83479, 4231.68447, 5029.04541, 5846.12059, 6644.13503, 7384.31376, 8027.88179, 8378.92903, 8660.57236, 8901.38042, 9216.34989, 9599.83831, 10010.1916, 10443.5678, 10896.125, 11364.0215, 11843.4154, 12313.5197, 12827.0534, 13349.0314, 13844.4683, 14278.3793, 14615.7791, 14821.6827, 15127.7919, 15244.116},
        {135.976907, 135.993856, 135.980666, 135.937225, 135.86342, 135.759141, 135.624274, 135.458709, 135.262333, 135.035035, 134.776703, 134.487225, 134.166489, 133.814383, 133.430797, 133.015616, 132.568731, 132.090029, 131.579398, 131.036727, 130.461903, 129.854815, 129.215351, 128.5434, 127.838848, 127.101586, 126.3315, 125.528478, 124.69241, 123.823184, 122.920686, 121.984806, 121.015432, 120.012452, 118.957001, 117.858737, 116.727545, 115.563924, 114.368369, 113.141377, 111.883445, 110.59507, 109.276748, 107.928976, 106.552252, 105.147072, 103.713932, 102.25333, 100.765762, 99.2517244, 97.7117151, 96.142762, 94.5489054, 92.9307173, 91.2887696, 89.6236345, 87.9358837, 86.2260893, 84.4948232, 82.7426574, 80.9701639, 79.1779147, 77.3664817, 75.5364369, 73.6883522, 71.8227996, 69.9403512, 68.0414718, 66.1266291, 64.1966156, 62.2520102, 60.2933919, 58.3213396, 56.3364321, 54.3392484, 52.3303673, 50.3103679, 48.2798289, 46.2393292, 44.1894479, 42.1307637, 40.0638556, 37.9893025, 35.9076833, 33.872651, 31.857131, 29.8334111, 27.8003478, 25.7567975, 23.7016169, 21.6336625, 19.5517909, 17.4548585, 15.341722, 13.2112379, 11.0622627, 8.89365299, 6.70426533, 4.49295624, 2.25858228, 0, -2.2806992, -4.57253196, -6.87986836, -9.20707847, -11.5585324, -13.9386002, -16.351652, -18.8020578, -21.2941877, -23.8324119, -26.4211004, -29.0646232, -31.7673505, -34.5336524, -37.3678989, -40.2744601, -43.1688896, -46.0428694, -49.1016641, -52.3306116, -55.7150499, -59.2403169, -62.8917507, -66.6546893, -70.5144705, -74.4564324, -78.465913, -82.5282503, -86.6287822, -90.7175453, -94.8281228, -98.9564432, -103.098435, -107.479279, -111.980279, -116.4216, -120.730396, -124.833819, -128.659021, -132.133154, -135.269486, -137.882846, -139.883537, -141.181863, -141.688129, -141.2844, -139.916277, -137.501125, -133.956307, -129.199187, -124.648916, -116.885551, -104.297069, -85.7409525, -61.0332643, -29.5596845, 9.29410691, 57.157063, 113.398272, 176.279949, 239.148464, 314.713573, 414.990206, 546.030108, 703.780748, 888.82762, 1099.30965, 1335.17803, 1605.35474, 1913.68737, 2270.13295, 2671.48973, 3119.4905, 3615.86804, 4162.35516, 4906.76848, 5668.236, 6411.96953, 7103.18092, 7707.08198, 8043.81789, 8320.37824, 8564.1463, 8891.79887, 9295.37385, 9729.70348, 10188.6859, 10666.2194, 11156.2021, 11652.5322, 12026.6565, 12493.5446, 13009.7002, 13531.6271, 14015.8292, 14418.8102, 14697.074, 15034.1479, 15058.621},
        {130.320679, 130.357614, 130.365784, 130.345064, 130.295329, 130.216455, 130.108316, 129.970788, 129.803745, 129.607063, 129.380617, 129.124281, 128.837932, 128.521444, 128.174692, 127.797551, 127.389897, 126.951604, 126.482548, 125.982604, 125.451646, 124.889551, 124.296192, 123.671446, 123.015186, 122.327289, 121.60763, 120.856083, 120.072523, 119.256826, 118.408867, 117.528521, 116.615663, 115.670168, 114.67496, 113.638745, 112.570435, 111.470456, 110.339233, 109.177191, 107.984754, 106.76235, 105.510401, 104.229335, 102.919576, 101.581549, 100.215679, 98.8223922, 97.4021134, 95.9552677, 94.4822803, 92.9742213, 91.4410739, 89.8834662, 88.3020262, 86.6973819, 85.0701614, 83.4209926, 81.7505037, 80.0593226, 78.3480773, 76.617396, 74.8679065, 73.1002371, 71.3150155, 69.51287, 67.6944286, 65.8625031, 64.019853, 62.1626023, 60.2912374, 58.4062448, 56.5081109, 54.5973221, 52.674365, 50.7397259, 48.7938914, 46.8373478, 44.8705817, 42.8940794, 40.9083275, 38.9138124, 36.9110206, 34.9004384, 32.9296772, 30.9746699, 29.0107827, 27.036973, 25.0521979, 23.0554145, 21.04558, 19.0216515, 16.9825863, 14.9273414, 12.8548741, 10.7641415, 8.65410084, 6.52370916, 4.37192369, 2.19770158, 0, -2.09379645, -4.22340475, -6.39112535, -8.59925869, -10.8501052, -13.1459653, -15.4891395, -17.8819282, -20.3266319, -22.8255509, -25.3809858, -27.9952369, -30.6706047, -33.4093896, -36.2138922, -39.0864128, -41.963787, -44.8389633, -47.8623171, -51.0221208, -54.3066465, -57.7041667, -61.2029535, -64.7912793, -68.4574162, -72.1896366, -75.9762127, -79.8054169, -83.6655213, -87.5177636, -91.3873635, -95.2707036, -99.1641664, -103.315787, -107.596122, -111.808621, -115.87417, -119.713656, -123.247967, -126.39799, -129.407293, -131.775873, -133.361486, -134.021884, -133.614822, -130.801037, -126.934552, -122.172375, -116.671514, -110.588976, -106.337557, -99.0613999, -86.5558869, -67.3788582, -41.6448426, -8.7694477, 31.8317188, 81.0852414, 139.15555, 205.833772, 277.495287, 362.570574, 469.050465, 601.220991, 757.090605, 937.549272, 1141.09916, 1368.01117, 1627.31148, 1923.07376, 2269.4896, 2658.95822, 3092.05095, 3569.33914, 4091.39414, 4783.33432, 5490.04724, 6180.96751, 6825.52972, 7393.16846, 7723.59219, 8002.44573, 8254.76059, 8589.52013, 8998.52493, 9438.5956, 9903.28173, 10386.1329, 10880.6986, 11380.5284, 11806.5311, 12281.5547, 12778.8117, 13271.5142, 13732.8746, 14136.1051, 14454.418, 14781.5916, 14889.5161},
        {124.827823, 124.869976, 124.887285, 124.8795, 124.84637, 124.787644, 124.703073, 124.592406, 124.455394, 124.291785, 124.10133, 123.883779, 123.638881, 123.366385, 123.066043, 122.737603, 122.380815, 121.995429, 121.581195, 121.137862, 120.665181, 120.162901, 119.630771, 119.068543, 118.475964, 117.852786, 117.198757, 116.513628, 115.797149, 115.049069, 114.269137, 113.457104, 112.61272, 111.735734, 110.805449, 109.832267, 108.826839, 107.789577, 106.720896, 105.621208, 104.490927, 103.330466, 102.140238, 100.920657, 99.6721363, 98.3950888, 97.089928, 95.7570673, 94.3969199, 93.0098992, 91.5964185, 90.1448547, 88.6679185, 87.1662841, 85.6406259, 84.091618, 82.5199347, 80.9262503, 79.311239, 77.6755752, 76.0199329, 74.3449866, 72.6514104, 70.9398786, 69.2110654, 67.4656452, 65.7042922, 63.9315205, 62.151752, 60.3577416, 58.5499146, 56.7286962, 54.8945119, 53.0477869, 51.1889466, 49.3184163, 47.4366214, 45.5439871, 43.6409388, 41.7279019, 39.8053015, 37.8735632, 35.9331122, 33.9843738, 32.0711926, 30.1713711, 28.2621898, 26.342665, 24.4118131, 22.4686502, 20.5121928, 18.541457, 16.5554593, 14.5532158, 12.5337429, 10.4960569, 8.43917402, 6.36211061, 4.26388294, 2.1435073, 0, -1.77512299, -3.64298691, -5.60098867, -7.64652515, -9.77699327, -11.9897899, -14.282312, -16.6519564, -19.0961201, -21.6121998, -24.1975927, -26.8496954, -29.565905, -32.3436183, -35.1802323, -38.0731438, -41.0976167, -44.2621718, -47.3880783, -50.4894188, -53.5802761, -56.6747329, -59.7868717, -62.9307753, -66.1205264, -69.3702077, -72.6939017, -76.1056913, -79.6196591, -83.3918576, -87.242344, -91.1426101, -95.0641478, -99.1267835, -103.227842, -107.225091, -111.045522, -114.616126, -117.863895, -120.715819, -123.610572, -125.807735, -127.134186, -127.416806, -126.482475, -122.679042, -117.682175, -111.68851, -104.894685, -97.4973367, -92.009502, -83.4802638, -69.5153049, -48.6506976, -21.3220589, 12.8878511, 54.3962724, 102.090571, 158.265559, 226.885002, 311.1381, 410.747462, 525.020584, 654.921093, 804.218841, 975.395579, 1169.99038, 1390.2406, 1641.84006, 1928.52741, 2265.86364, 2643.94624, 3062.4215, 3520.9357, 4019.13514, 4661.05854, 5315.86167, 5957.09271, 6558.29986, 7093.0313, 7422.04595, 7706.87412, 7969.40228, 8303.21564, 8701.8199, 9128.84626, 9579.29413, 10048.1629, 10530.4521, 11021.161, 11624.4528, 12156.6573, 12627.6979, 13047.4982, 13425.9818, 13773.0721, 14098.6927, 14389.7615, 14732.1703},
        {120.365703, 120.359781, 120.335275, 120.291749, 120.228769, 120.145901, 120.042711, 119.918763, 119.773624, 119.606858, 119.418032, 119.206711, 118.972461, 118.714846, 118.433434, 118.127788, 117.797474, 117.442059, 117.061108, 116.654185, 116.220857, 115.76069, 115.273248, 114.758098, 114.214804, 113.642932, 113.042049, 112.411718, 111.751507, 111.06098, 110.339703, 109.587241, 108.80316, 107.987026, 107.112712, 106.193169, 105.24166, 104.258582, 103.244335, 102.199319, 101.123933, 100.018576, 98.8836467, 97.7195454, 96.526671, 95.3054227, 94.0561999, 92.7794019, 91.4754279, 90.1446771, 88.7875489, 87.391353, 85.9698619, 84.5237587, 83.0537265, 81.5604483, 80.0446071, 78.5068861, 76.9479681, 75.3685363, 73.7692738, 72.1508634, 70.5139883, 68.8593316, 67.1875762, 65.4994052, 63.7955016, 62.0811426, 60.3614951, 58.6277671, 56.8803437, 55.1196103, 53.3459519, 51.5597539, 49.7614014, 47.9512797, 46.1297739, 44.2972694, 42.4541513, 40.6008048, 38.7376152, 36.8649676, 34.9832474, 33.0928397, 31.2345575, 29.3877219, 27.5311679, 25.6639687, 23.7851976, 21.8939278, 19.9892326, 18.0701852, 16.135859, 14.1853272, 12.2176629, 10.2319396, 8.22723035, 6.20260853, 4.15714736, 2.0899201, 0, -1.66308209, -3.43275997, -5.30493996, -7.2755284, -9.34043161, -11.4955559, -13.7368077, -16.0600932, -18.4613188, -20.9363908, -23.4812156, -26.0916995, -28.7637488, -31.4932698, -34.2761689, -37.1083524, -40.105539, -43.2807509, -46.3588905, -49.3615383, -52.3102747, -55.2266799, -58.1323345, -61.0488188, -63.9977132, -67.000598, -70.0790537, -73.2546607, -76.5489993, -80.1898347, -83.9169615, -87.6901045, -91.4689887, -95.3231398, -99.1573824, -102.847261, -106.31956, -109.501064, -112.318559, -114.698828, -117.045526, -118.663434, -119.386036, -119.046817, -117.479261, -113.553064, -108.306444, -101.813835, -94.1496678, -85.3883733, -77.4210015, -66.2850548, -50.034937, -27.4441176, 1.24584134, 36.4525213, 78.5935038, 125.756651, 181.218745, 250.798082, 340.802904, 447.015555, 565.041818, 694.333705, 840.860451, 1007.59978, 1197.30583, 1412.89838, 1658.11694, 1936.23733, 2261.8133, 2625.10766, 3025.49234, 3462.33926, 3935.02032, 4533.9416, 5144.68232, 5743.85587, 6308.07563, 6813.95498, 7137.94154, 7423.59161, 7690.44678, 8021.58175, 8411.38954, 8827.79104, 9266.5029, 9723.24175, 10193.7242, 10673.667, 11296.1278, 11829.2808, 12290.3734, 12696.653, 13065.367, 13413.7629, 13759.0881, 14001.4264, 14496.4056},
        {117.105437, 116.997983, 116.87965, 116.749804, 116.607811, 116.453035, 116.284842, 116.102598, 115.905668, 115.693417, 115.465212, 115.220416, 114.958397, 114.678518, 114.380147, 114.062647, 113.725385, 113.367726, 112.989035, 112.588678, 112.166021, 111.720428, 111.251266, 110.757899, 110.239693, 109.696014, 109.126226, 108.529696, 107.905789, 107.253871, 106.573306, 105.86346, 105.123698, 104.353387, 103.521176, 102.642435, 101.732901, 100.792937, 99.8229055, 98.8231679, 97.7940868, 96.7360244, 95.6493428, 94.5344043, 93.391571, 92.2212052, 91.0236691, 89.7993249, 88.5485347, 87.2716609, 85.9690655, 84.6279947, 83.2622112, 81.8723615, 80.459092, 79.0230494, 77.5648801, 76.0852307, 74.5847478, 73.0640778, 71.5238673, 69.9647629, 68.387411, 66.7924582, 65.180551, 63.5523361, 61.9084598, 60.2539484, 58.5937232, 56.9193977, 55.2313345, 53.5298963, 51.8154457, 50.0883453, 48.3489577, 46.5976455, 44.8347714, 43.0606979, 41.2757877, 39.4804034, 37.6749075, 35.8596628, 34.0350318, 32.2013771, 30.3968557, 28.6021381, 26.7974405, 24.981899, 23.1546498, 21.3148289, 19.4615725, 17.5940167, 15.7112976, 13.8125513, 11.8969139, 9.96352165, 8.01151052, 6.04001668, 4.04817623, 2.0351253, 0, -1.83060817, -3.72195339, -5.67328402, -7.68384845, -9.75289505, -11.8796722, -14.0634283, -16.3034116, -18.5988707, -20.9490537, -23.3532092, -25.8105855, -28.320431, -30.881994, -33.4945229, -36.1572661, -38.875018, -41.6478191, -44.4623734, -47.3191178, -50.218489, -53.1609239, -56.1468592, -59.1767316, -62.2509781, -65.3700353, -68.5343402, -71.7443293, -75.0004397, -78.4014654, -81.8134215, -85.2072375, -88.5538429, -92.0067817, -95.4456761, -98.7014516, -101.690253, -104.328227, -106.531517, -108.216269, -109.489205, -110.017893, -109.68119, -108.357954, -105.927045, -102.519848, -97.6995616, -91.2819133, -83.0826294, -72.9174365, -61.4550628, -46.6156755, -27.5416099, -3.46828871, 26.1788116, 62.0595448, 104.833765, 153.759574, 211.217157, 281.115882, 367.005398, 468.159143, 584.223301, 715.921729, 865.804345, 1035.58288, 1226.12908, 1438.9369, 1678.5802, 1947.89069, 2258.45612, 2603.00503, 2981.33816, 3393.25627, 3838.56009, 4402.7013, 4978.41699, 5544.09517, 6078.12387, 6558.89109, 6871.20113, 7149.41488, 7412.02767, 7740.19051, 8127.59753, 8541.83631, 8977.9063, 9430.80696, 9895.53773, 10367.0981, 10781.2157, 11234.2047, 11708.6574, 12187.1664, 12652.3241, 13086.7229, 13472.9553, 13669.1079, 14167.5035},
        {114.315772, 114.097521, 113.876096, 113.650666, 113.4204, 113.184468, 112.942039, 112.692283, 112.43437, 112.167469, 111.89075, 111.603382, 111.304534, 110.993377, 110.669079, 110.330811, 109.977741, 109.60904, 109.223877, 108.821421, 108.400842, 107.961309, 107.501993, 107.022062, 106.520686, 105.997035, 105.450277, 104.879584, 104.284124, 103.663066, 103.015581, 102.340837, 101.638005, 100.906253, 100.108877, 99.2637374, 98.3891276, 97.4853814, 96.5528326, 95.5918149, 94.6026621, 93.585708, 92.5412862, 91.4697306, 90.371375, 89.246553, 88.0955985, 86.9188453, 85.716627, 84.4892774, 83.2371303, 81.9486861, 80.6363684, 79.3007675, 77.9424738, 76.5620775, 75.1601689, 73.7373382, 72.2941758, 70.831272, 69.3492171, 67.8486012, 66.3300148, 64.7940481, 63.2412914, 61.6723349, 60.0877691, 58.4919811, 56.8892672, 55.2723865, 53.6416832, 51.9975014, 50.3401853, 48.6700789, 46.9875265, 45.2928722, 43.5864601, 41.8686344, 40.1397391, 38.4001185, 36.6501167, 34.8900777, 33.1203459, 31.3412652, 29.5886583, 27.8443839, 26.0898741, 24.3243219, 22.54692, 20.7568612, 18.9533383, 17.1355442, 15.3026715, 13.4539131, 11.5884618, 9.70551046, 7.80425179, 5.8838786, 3.94358371, 1.98255991, 0, -2.00717277, -4.02743762, -6.06359991, -8.11846497, -10.1948381, -12.2955248, -14.4233303, -16.5810599, -18.771519, -20.997513, -23.2618472, -25.5673269, -27.9167575, -30.3129444, -32.7586928, -35.2568082, -37.6880553, -40.0378051, -42.5849071, -45.306015, -48.1777827, -51.1768639, -54.2799125, -57.4635822, -60.7045269, -63.9794004, -67.2648564, -70.5375487, -73.7741312, -76.9285822, -80.0085453, -82.9974769, -85.8788336, -88.8838021, -91.8719737, -94.6368782, -97.087653, -99.1334353, -100.683363, -101.646572, -101.871635, -101.346688, -99.9927178, -97.7307128, -94.4816597, -91.5796501, -87.1792911, -80.8482939, -72.1543697, -60.6652299, -45.8954355, -27.5308093, -5.7296091, 19.843341, 50.530645, 87.222592, 130.809472, 181.968028, 241.850628, 311.842598, 391.71077, 485.484996, 598.018219, 731.785739, 885.234068, 1058.659, 1250.84303, 1461.68961, 1696.65095, 1958.04055, 2253.72052, 2579.60215, 2936.07833, 3323.54191, 3742.38574, 4275.81156, 4820.70113, 5356.74511, 5863.63415, 6321.05887, 6619.66348, 6887.54938, 7142.56998, 7465.88944, 7850.30067, 8262.2171, 8695.75062, 9145.01313, 9604.11649, 10067.1726, 10227.0901, 10586.7304, 11083.2608, 11653.8484, 12235.6604, 12765.8641, 13181.6268, 13351.993, 13825.3551},
        {111.077693, 110.805614, 110.534985, 110.264857, 109.994277, 109.722298, 109.447968, 109.170337, 108.888456, 108.601373, 108.308139, 108.007804, 107.699418, 107.38203, 107.05469, 106.716448, 106.366354, 106.003458, 105.626809, 105.235458, 104.828455, 104.404848, 103.963689, 103.504026, 103.02491, 102.52539, 102.004517, 101.46134, 100.89491, 100.304275, 99.6884858, 99.0465923, 98.3776444, 97.6806918, 96.9146691, 96.0995276, 95.2557007, 94.38354, 93.4833971, 92.5556238, 91.6005717, 90.6185925, 89.6100378, 88.5752593, 87.5146086, 86.4284374, 85.3170974, 84.1809403, 83.0203176, 81.8355811, 80.6270825, 79.386788, 78.1236164, 76.8381011, 75.5307757, 74.2021736, 72.8528281, 71.4832727, 70.0940409, 68.6856661, 67.2586817, 65.8136213, 64.3510181, 62.8714057, 61.3753176, 59.863287, 58.3358476, 56.7970091, 55.2506981, 53.690278, 52.1160569, 50.5283429, 48.927444, 47.3136683, 45.6873239, 44.0487188, 42.3981612, 40.7359591, 39.0624205, 37.3778536, 35.6825665, 33.9768671, 32.2610636, 30.5354641, 28.8333543, 27.1378599, 25.4317099, 23.7141424, 21.9843952, 20.2417061, 18.4853131, 16.714454, 14.9283667, 13.1262891, 11.3074592, 9.47111464, 7.61649347, 5.74283355, 3.84937274, 1.93534893, 0, -1.80607623, -3.6562605, -5.55091889, -7.49041749, -9.47512239, -11.5053997, -13.5816154, -15.7041358, -17.8733267, -20.0895544, -22.353185, -24.6645844, -27.0241189, -29.4321544, -31.8890571, -34.3951931, -36.8601139, -39.2712119, -41.8342674, -44.5301861, -47.3398737, -50.2442362, -53.2241794, -56.260609, -59.3344309, -62.4265509, -65.5178749, -68.5893086, -71.6217578, -74.6247092, -77.5400082, -80.3399864, -82.9969753, -85.6304634, -88.1392038, -90.3827079, -92.2891603, -93.7867456, -94.8036482, -95.2680529, -95.3052138, -94.586268, -93.0008429, -90.438566, -86.7890647, -83.1167384, -77.8427495, -70.5630324, -60.8735216, -48.3701514, -32.7225782, -13.36291, 9.86260153, 37.373896, 70.1343869, 108.863479, 154.280577, 206.632992, 267.219516, 337.853956, 418.895844, 513.454347, 625.185634, 756.194724, 905.958307, 1075.15952, 1263.79066, 1472.35578, 1703.89193, 1960.00335, 2245.78045, 2558.80497, 2899.4744, 3268.18623, 3665.33797, 4166.70525, 4678.46287, 5182.1638, 5659.36099, 6091.60739, 6378.35661, 6637.99386, 6887.25754, 7202.69225, 7577.36311, 7979.41594, 8403.16179, 8842.91172, 9292.97679, 9747.66805, 9788.37694, 10114.8546, 10635.3514, 11258.1178, 11891.4042, 12443.4611, 12822.539, 13039.2696, 13583.9808},
        {107.896638, 107.589707, 107.287612, 106.989315, 106.693779, 106.399968, 106.106845, 105.813372, 105.518513, 105.221231, 104.920488, 104.615249, 104.304476, 103.987132, 103.66218, 103.328583, 102.985305, 102.631309, 102.265557, 101.887012, 101.494638, 101.087399, 100.664256, 100.224173, 99.7661126, 99.2890387, 98.791914, 98.2737017, 97.7333647, 97.1698661, 96.5821691, 95.9692366, 95.3300319, 94.6635179, 93.9248809, 93.1358932, 92.3188532, 91.4741447, 90.6021513, 89.7032569, 88.7778452, 87.8262998, 86.8490046, 85.8463431, 84.8186992, 83.7664566, 82.689999, 81.5897101, 80.4659736, 79.3191733, 78.1496929, 76.9535638, 75.7356164, 74.4963288, 73.236179, 71.9556451, 70.6552053, 69.3353375, 67.9965198, 66.6392304, 65.2639473, 63.8711486, 62.4613123, 61.0349166, 59.5924394, 58.134359, 56.6611532, 55.176586, 53.6843422, 52.1781233, 50.6581946, 49.1248209, 47.5782675, 46.0187993, 44.4466816, 42.8621793, 41.2655575, 39.6570814, 38.0370159, 36.4056263, 34.7631774, 33.1099345, 31.4461626, 29.7721268, 28.1187926, 26.4704292, 24.8109366, 23.1395836, 21.455639, 19.7583716, 18.0470503, 16.3209437, 14.5793207, 12.82145, 11.0466006, 9.254041, 7.44304018, 5.61286688, 3.76278989, 1.892078, 0, -1.50216164, -3.10214527, -4.79610109, -6.58017934, -8.45053022, -10.403304, -12.4346508, -14.5407209, -16.7176645, -18.9616319, -21.2687733, -23.6352388, -26.0571787, -28.5307432, -31.0520826, -33.617347, -36.2137057, -38.836026, -41.5007721, -44.2021698, -46.9344448, -49.6918228, -52.4685295, -55.2587907, -58.0568322, -60.8568795, -63.6531586, -66.439895, -69.2113146, -72.0967984, -74.9058598, -77.5921779, -80.1094318, -82.4076098, -84.4422364, -86.1698204, -87.5451481, -88.523006, -89.0581805, -89.1054579, -89.2359107, -88.6004739, -87.0333561, -84.3687662, -80.4409126, -75.5197897, -68.8948743, -60.2914286, -49.4347148, -36.0499954, -20.4243002, -1.03451899, 22.9861885, 52.3112485, 87.2192005, 128.16588, 175.607123, 228.925608, 289.894226, 361.456588, 445.683442, 543.249492, 654.845993, 781.630723, 925.551956, 1088.19512, 1271.66238, 1477.67315, 1706.05205, 1957.6955, 2235.41668, 2537.90035, 2865.37942, 3218.08682, 3596.25545, 4067.64009, 4548.07134, 5020.90165, 5469.48349, 5877.1693, 6151.9447, 6402.77355, 6645.30839, 6950.64439, 7312.48803, 7701.44471, 8112.37597, 8540.14336, 8979.60842, 9425.63271, 9408.30891, 9725.70647, 10266.6323, 10919.8933, 11574.2965, 12118.6486, 12441.7568, 12735.5012, 13369.4895},
        {104.915139, 104.580894, 104.254217, 103.934003, 103.619148, 103.308548, 103.001096, 102.695689, 102.391222, 102.086591, 101.780689, 101.472414, 101.16066, 100.844323, 100.522297, 100.193479, 99.8567631, 99.5110451, 99.1552202, 98.7881838, 98.4088311, 98.0160576, 97.6087584, 97.185829, 96.7461647, 96.2886607, 95.8122125, 95.3157152, 94.7980643, 94.258155, 93.6948827, 93.1071427, 92.4938303, 91.8538408, 91.1392232, 90.3732814, 89.5798822, 88.7594412, 87.912374, 87.0390962, 86.1400233, 85.2155709, 84.2661546, 83.2921899, 82.2940925, 81.2722779, 80.2271616, 79.1591594, 78.0686866, 76.9561589, 75.8219919, 74.665907, 73.4890288, 72.2917882, 71.0746157, 69.8379419, 68.5821975, 67.3078131, 66.0152192, 64.7048466, 63.3771259, 62.0324876, 60.6713625, 59.294181, 57.901374, 56.4933718, 55.0706053, 53.6367916, 52.195569, 50.7405898, 49.2720715, 47.7902318, 46.2952882, 44.7874582, 43.2669594, 41.7340095, 40.1888258, 38.6316262, 37.062628, 35.4820488, 33.8901063, 32.287018, 30.6730014, 29.0482742, 27.4418766, 25.8390087, 24.2245239, 22.5977044, 20.9578324, 19.3041902, 17.6360599, 15.9527237, 14.2534639, 12.5375627, 10.8043023, 9.05296484, 7.28283261, 5.4931878, 3.68331263, 1.85248929, 0, -1.20173153, -2.55648711, -4.05598667, -5.69195011, -7.45609738, -9.34014838, -11.3358231, -13.4348413, -15.6289231, -17.9097883, -20.2691569, -22.6987488, -25.1902838, -27.7354821, -30.3260633, -32.9537476, -35.6969316, -38.5597177, -41.3377711, -44.0413852, -46.6808536, -49.2664698, -51.8085274, -54.3173199, -56.8031408, -59.2762838, -61.7470422, -64.2257097, -66.7225798, -69.5050825, -72.2320914, -74.8367593, -77.2522387, -79.2526293, -80.8506102, -82.1012745, -82.9854907, -83.4841273, -83.5780527, -83.2481355, -83.5342489, -83.0359507, -81.5269129, -78.7808072, -74.5713057, -68.195434, -60.0226716, -49.945852, -37.8578085, -23.6513745, -8.37822938, 10.6440068, 35.0256087, 65.7274786, 102.384716, 145.22768, 194.486729, 248.763512, 310.287256, 383.063962, 470.942572, 571.983477, 683.670746, 806.126411, 943.942391, 1099.65885, 1277.4933, 1480.42075, 1705.26603, 1952.33286, 2222.74584, 2515.86152, 2831.69948, 3170.27927, 3531.62047, 3976.26859, 4428.58577, 4873.46009, 5295.77965, 5680.43252, 5942.63692, 6183.39739, 6417.46051, 6710.80228, 7057.891, 7431.84937, 7828.21297, 8242.51736, 8670.29813, 9107.09083, 9083.97518, 9408.16251, 9960.19644, 10620.6206, 11269.9785, 11788.8138, 12057.6701, 12441.3342, 13155.8644},
        {102.481655, 102.100806, 101.730976, 101.370979, 101.01963, 100.675743, 100.338134, 100.005617, 99.6770055, 99.3511154, 99.0267608, 98.7027564, 98.3779167, 98.0510564, 97.7209901, 97.3865322, 97.0464975, 96.6997006, 96.3449559, 95.9810781, 95.6068819, 95.2211818, 94.8227923, 94.4105282, 93.9832039, 93.5396341, 93.0786334, 92.5990164, 92.0995976, 91.5791917, 91.0366132, 90.4706768, 89.8801971, 89.2639886, 88.5718736, 87.8281935, 87.0578783, 86.2613324, 85.4389602, 84.5911664, 83.7183554, 82.8209317, 81.8992997, 80.9538639, 79.9850289, 78.9931992, 77.9787791, 76.9421732, 75.883786, 74.8040219, 73.7032856, 72.5811592, 71.4388872, 70.2768921, 69.095596, 67.8954212, 66.6767902, 65.4401251, 64.1858483, 62.9143822, 61.6261489, 60.3215709, 59.0010704, 57.6650698, 56.3139913, 54.9482572, 53.5682899, 52.1784965, 50.7831883, 49.3745694, 47.9528038, 46.5180554, 45.0704882, 43.6102663, 42.1375537, 40.6525142, 39.1553119, 37.6461109, 36.125075, 34.5923684, 33.0481549, 31.4925986, 29.9258634, 28.3481134, 26.7878596, 25.2304963, 23.6610769, 22.0788455, 20.4830462, 18.8729231, 17.2477202, 15.6066815, 13.9490513, 12.2740735, 10.5809923, 8.86905162, 7.13749564, 5.38556842, 3.61251402, 1.81757652, 0, -1.19261072, -2.54004096, -4.03324907, -5.66319344, -7.42083242, -9.29712438, -11.2830277, -13.3695007, -15.5475018, -17.8079894, -20.1419217, -22.5402572, -24.9939543, -27.4939713, -30.0312665, -32.5967984, -35.2992701, -38.1464608, -40.8630011, -43.4650804, -45.9688882, -48.3906139, -50.7464468, -53.0525765, -55.3251921, -57.5804833, -59.8346393, -62.1038496, -64.4043036, -67.0647273, -69.6741766, -72.1550802, -74.4298664, -76.2197177, -77.5476854, -78.4904867, -79.0309239, -79.1517991, -78.8359147, -78.066073, -77.9768745, -77.0487759, -75.0392274, -71.7056795, -66.8055826, -59.2744067, -49.8970776, -38.6365408, -25.4557419, -10.3176262, 5.54183672, 24.9886436, 49.3910869, 79.6885208, 115.944552, 158.615979, 208.159603, 264.025569, 327.906115, 402.591642, 490.928628, 591.70451, 703.002475, 825.082905, 961.895317, 1115.69585, 1290.23738, 1488.16407, 1706.63214, 1945.9021, 2206.85205, 2489.02991, 2792.48244, 3117.25636, 3463.39842, 3889.47422, 4322.38193, 4747.53858, 5150.36119, 5516.26678, 5762.5654, 5988.18573, 6207.73348, 6486.00757, 6817.7456, 7176.77136, 7558.86812, 7959.81918, 8375.40784, 8801.41737, 8836.68266, 9172.77794, 9705.11819, 10329.1184, 10940.1935, 11433.7585, 11705.2283, 12159.5924, 12875.6988},
        {100.344301, 99.9154578, 99.5009194, 99.0994268, 98.709721, 98.3305431, 97.960634, 97.5987347, 97.2435862, 96.8939295, 96.5485056, 96.2060554, 95.86532, 95.5250404, 95.1839575, 94.8408122, 94.4943458, 94.143299, 93.7864129, 93.4224285, 93.0500867, 92.6681286, 92.2752952, 91.8703273, 91.4519661, 91.0189526, 90.5700276, 90.1039322, 89.6194074, 89.1151942, 88.5900335, 88.0426663, 87.4718338, 86.8762767, 86.2041478, 85.4805463, 84.7311794, 83.9564299, 83.1566803, 82.3323135, 81.4837119, 80.6112584, 79.7153357, 78.7963263, 77.8546129, 76.8905783, 75.9046051, 74.8970761, 73.8683737, 72.8188809, 71.7489801, 70.6571675, 69.5457532, 68.4151609, 67.2658142, 66.0981365, 64.9125516, 63.7094829, 62.4893541, 61.2525887, 59.9996104, 58.7308426, 57.446709, 56.1476331, 54.8340386, 53.506349, 52.1649878, 50.8152051, 49.462135, 48.096247, 46.7176517, 45.3264598, 43.9227821, 42.5067293, 41.078412, 39.637941, 38.1854269, 36.7209805, 35.2447124, 33.7567334, 32.2571542, 30.7460854, 29.2236378, 27.6899221, 26.1732093, 24.6589377, 23.132207, 21.592214, 20.0381556, 18.4692285, 16.8846297, 15.283556, 13.6652042, 12.0287711, 10.3734537, 8.69844872, 7.00295302, 5.28616346, 3.54727688, 1.78549011, 0, -1.26968957, -2.679999, -4.22225387, -5.88777977, -7.66790226, -9.55394693, -11.5372394, -13.6091051, -15.7608698, -17.9838589, -20.2693981, -22.6088129, -24.993429, -27.4145718, -29.863567, -32.3317402, -34.937126, -37.6891513, -40.2925378, -42.7657629, -45.1273042, -47.3956393, -49.5892456, -51.7266007, -53.8261822, -55.9064675, -57.9859342, -60.0830598, -62.2163219, -64.7469261, -67.2249548, -69.5660671, -71.6859221, -73.2915905, -74.4030256, -75.0958041, -75.3481617, -75.1383338, -74.4445562, -73.2450644, -72.6761894, -71.2056075, -68.5849703, -64.5659296, -58.9001369, -50.3122748, -39.8377064, -27.4848256, -13.2620264, 2.82229712, 19.4996915, 39.5618953, 63.8753721, 93.2756364, 128.535015, 170.454203, 219.833899, 277.373267, 343.997611, 420.742995, 508.831404, 608.508881, 719.278249, 841.791871, 978.649953, 1131.55861, 1303.21646, 1495.58695, 1706.99428, 1937.82123, 2189.2345, 2460.71231, 2752.36606, 3064.30713, 3396.64695, 3807.35011, 4224.21151, 4632.87922, 5019.00135, 5368.22599, 5598.53499, 5808.35351, 6012.32944, 6274.89333, 6590.98747, 6934.81134, 7302.34438, 7689.56599, 8092.45559, 8506.99259, 8640.26237, 8995.64349, 9495.05726, 10060.425, 10613.6681, 11076.7078, 11371.4655, 11886.7323, 12576.2446},
        {98.4644962, 97.9882771, 97.529402, 97.0865471, 96.6583885, 96.2436022, 95.8408643, 95.448851, 95.0662383, 94.6917023, 94.3239191, 93.9615648, 93.6033154, 93.2478471, 92.893836, 92.5399582, 92.1848896, 91.8273065, 91.4658849, 91.0993009, 90.7262307, 90.3453502, 89.9553355, 89.5548629, 89.1426083, 88.7172478, 88.2774576, 87.8219138, 87.3492923, 86.8582693, 86.347521, 85.8157233, 85.2615524, 84.6836844, 84.0290373, 83.323255, 82.5926032, 81.8374377, 81.0581142, 80.2549884, 79.4284162, 78.5787532, 77.7063551, 76.8115778, 75.8947769, 74.9563082, 73.9965275, 73.0157903, 72.0144526, 70.99287, 69.9513982, 68.88713, 67.8037548, 66.7016992, 65.5813895, 64.4432524, 63.2877143, 62.1152016, 60.9261409, 59.7209585, 58.500081, 57.2639349, 56.0129466, 54.7475426, 53.4681493, 52.1751933, 50.869101, 49.555933, 48.2416147, 46.9149524, 45.5760076, 44.2248414, 42.8615151, 41.48609, 40.0986271, 38.6991879, 37.2878335, 35.8646252, 34.4296242, 32.9828918, 31.5244891, 30.0544775, 28.5729181, 27.0798723, 25.6033027, 24.1287329, 22.6413513, 21.1403136, 19.6247755, 18.093893, 16.5468217, 14.9827174, 13.400736, 11.8000332, 10.1797647, 8.53908647, 6.87715415, 5.19312356, 3.48615048, 1.7553907, 0, -1.40300074, -2.92104115, -4.54668501, -6.27249607, -8.0910381, -9.99487486, -11.9765701, -14.0286876, -16.1437912, -18.3144445, -20.5332114, -22.7926556, -25.0853408, -27.4038309, -29.7406897, -32.0884808, -34.5528673, -37.1425702, -39.5943332, -41.9249556, -44.1512368, -46.2899761, -48.3579729, -50.3720264, -52.3489361, -54.3055013, -56.2585213, -58.2247954, -60.221123, -62.609557, -64.9350501, -67.1108259, -69.0501075, -70.4763608, -71.397688, -71.8727929, -71.8718262, -71.3649383, -70.3222798, -68.7140012, -67.6252993, -65.5719166, -62.3058422, -57.5790654, -51.1435757, -41.6477439, -30.2230822, -16.8974847, -1.698845, 15.3449428, 33.066994, 53.9705056, 78.2671332, 106.630557, 140.677759, 181.602196, 230.597328, 289.7688, 359.190568, 437.941477, 525.376862, 623.501522, 733.571983, 856.741421, 993.987887, 1146.36982, 1315.30657, 1501.95015, 1706.12919, 1928.42091, 2170.53431, 2431.74067, 2712.22545, 3012.17406, 3331.77195, 3728.84898, 4131.53505, 4525.60448, 4896.8316, 5230.99072, 5446.34945, 5641.19333, 5830.22813, 6077.58798, 6378.41328, 6707.24928, 7060.26612, 7433.63395, 7823.52293, 8226.10321, 8485.41379, 8865.0953, 9322.04414, 9813.15667, 10295.3293, 10725.4584, 11060.4403, 11621.4729, 12264.0148},
        {96.7807852, 96.2631122, 95.765305, 95.2859896, 94.8237922, 94.3773388, 93.9452554, 93.5261682, 93.1187031, 92.7214862, 92.3331436, 91.9523014, 91.5775855, 91.2076221, 90.8410373, 90.476457, 90.1125073, 89.7478144, 89.3810041, 89.0107027, 88.6355362, 88.2541306, 87.865112, 87.4671064, 87.0587399, 86.6386386, 86.2054285, 85.7577357, 85.2941862, 84.8134061, 84.3140214, 83.7946583, 83.2539427, 82.6905008, 82.0501402, 81.359007, 80.6438824, 79.9051065, 79.1430194, 78.3579613, 77.5502721, 76.7202921, 75.8683614, 74.9948201, 74.1000083, 73.1842661, 72.2479336, 71.2913511, 70.3148584, 69.3187959, 68.3035036, 67.2662782, 66.2105692, 65.1367827, 64.0453248, 62.9366016, 61.8110193, 60.6689838, 59.5109012, 58.3371778, 57.1482195, 55.9444324, 54.7262227, 53.4939965, 52.2481598, 50.9891187, 49.7172793, 48.4389096, 47.1601368, 45.8692774, 44.5663574, 43.2514029, 41.9244399, 40.5854947, 39.2345933, 37.8717617, 36.4970261, 35.1104126, 33.7119472, 32.3016561, 30.8795653, 29.4457009, 28.0000891, 26.5427558, 25.1000994, 23.6584051, 22.2036413, 20.7349783, 19.2515864, 17.7526357, 16.2372966, 14.7047393, 13.1541342, 11.5846514, 9.99546117, 8.38573387, 6.75463972, 5.101349, 3.42503196, 1.72485887, 0, -1.57484139, -3.22661114, -4.95037598, -6.74120266, -8.59415792, -10.5043085, -12.4667212, -14.4764626, -16.5285996, -18.6181989, -20.7403272, -22.8900513, -25.062438, -27.2525539, -29.4554657, -31.6662404, -33.9267083, -36.2386169, -38.4912575, -40.6897177, -42.839085, -44.944447, -47.0108913, -49.0435055, -51.047377, -53.0275935, -54.9892426, -56.9374117, -58.8771885, -61.1015372, -63.222114, -65.1576432, -66.8268495, -67.9722255, -68.6006117, -68.7658438, -68.4395157, -67.5932216, -66.1985555, -64.2271116, -62.8470863, -60.469288, -56.8311929, -51.6702773, -44.7240174, -34.5955053, -22.4401975, -8.27916607, 7.86651663, 25.9757784, 45.0986364, 67.278264, 92.052308, 119.90113, 153.229803, 193.579243, 242.490367, 303.299389, 375.343905, 456.003004, 543.165934, 640.043305, 748.798564, 870.890433, 1006.58351, 1156.69052, 1322.20394, 1503.9831, 1702.2282, 1917.51229, 2151.62285, 2403.73165, 2673.99136, 2962.55463, 3269.57413, 3650.07142, 4035.61304, 4412.63442, 4767.571, 5086.85821, 5292.09829, 5477.78202, 5658.13051, 5894.50467, 6182.46609, 6497.93477, 6837.45438, 7197.56857, 7574.82102, 7965.7554, 8393.96566, 8808.66586, 9213.15384, 9610.72744, 10004.6845, 10398.3229, 10794.9404, 11360.8758, 11951.5071},
        {95.2597827, 94.7078375, 94.177723, 93.6680307, 93.177352, 92.7042784, 92.2474013, 91.8053122, 91.3766024, 90.9598635, 90.5536869, 90.156664, 89.7673863, 89.3844452, 89.0064322, 88.6319387, 88.2595561, 87.8878759, 87.5154896, 87.1409885, 86.7629642, 86.380008, 85.9907115, 85.593666, 85.187463, 84.7706939, 84.3419503, 83.8998234, 83.4429049, 82.9697861, 82.4790584, 81.9693133, 81.4391423, 80.8871368, 80.2580394, 79.5784987, 78.8758102, 78.1503131, 77.4023461, 76.6322483, 75.8403585, 75.0270158, 74.1925591, 73.3373273, 72.4616593, 71.5658941, 70.6503707, 69.7154279, 68.7614048, 67.7886402, 66.7974731, 65.7873484, 64.7595184, 63.7143416, 62.6521761, 61.5733804, 60.4783127, 59.3673314, 58.2407948, 57.0990613, 55.9424891, 54.7714365, 53.586262, 52.3873237, 51.1749801, 49.9495895, 48.7115102, 47.4665165, 46.2202527, 44.961907, 43.6914867, 42.4089989, 41.114451, 39.80785, 38.4892033, 37.1585181, 35.8158016, 34.4610611, 33.0943036, 31.7155366, 30.3247671, 28.9220025, 27.5072499, 26.0805166, 24.6652108, 23.2491471, 21.8198586, 20.3765933, 18.9185988, 17.445123, 15.9554138, 14.4487189, 12.9242862, 11.3813634, 9.81919848, 8.23703913, 6.63413319, 5.0097285, 3.36307286, 1.69341409, 0, -1.75245914, -3.53712793, -5.35243511, -7.19680945, -9.06867969, -10.9664746, -12.8886228, -14.8335533, -16.7996946, -18.7854756, -20.7893249, -22.8096714, -24.8449438, -26.8935709, -28.9539813, -31.0246039, -33.0378397, -34.982685, -37.0069168, -39.094815, -41.2306596, -43.3987306, -45.5833079, -47.7686716, -49.9391015, -52.0788777, -54.1722801, -56.2035887, -58.1570834, -60.1919464, -62.0534247, -63.6733276, -64.9834643, -65.7441326, -65.9728975, -65.7330606, -65.0078847, -63.7806325, -62.0345669, -59.7529505, -58.3278947, -55.9050337, -52.1919862, -46.8963705, -39.7258052, -29.2726348, -16.6385702, -1.8100487, 15.2264928, 34.4846172, 55.3290436, 79.2152102, 105.08831, 133.227699, 166.636647, 207.09544, 256.384367, 318.671753, 392.807113, 475.034826, 562.49881, 658.726874, 765.613854, 884.620116, 1016.46984, 1162.22513, 1323.50667, 1501.52138, 1695.42804, 1905.544, 2133.08895, 2377.33902, 2638.29919, 2915.97445, 3210.36975, 3570.61782, 3935.31396, 4292.18119, 4628.94256, 4933.3211, 5133.79123, 5316.82369, 5495.53378, 5725.70715, 6003.56874, 6307.56425, 6634.80761, 6982.41277, 7347.49365, 7727.16419, 8349.5485, 8805.79048, 9152.45886, 9446.12238, 9743.3498, 10100.7099, 10574.7713, 11104.0363, 11643.8757},
        {93.9037119, 93.3208061, 92.7613931, 92.2240388, 91.7073086, 91.2097683, 90.7299836, 90.2665202, 89.8179438, 89.3828199, 88.9597144, 88.5471928, 88.1438209, 87.7481643, 87.3587887, 86.9742599, 86.5931434, 86.2140049, 85.8354102, 85.4559248, 85.0741146, 84.6885451, 84.297782, 83.9003911, 83.494938, 83.0799883, 82.6541078, 82.2158621, 81.763817, 81.296538, 80.8125909, 80.3105413, 79.788955, 79.2463976, 78.6268081, 77.9572145, 77.2653015, 76.5514074, 75.8158706, 75.0590296, 74.2812227, 73.4827885, 72.6640652, 71.8253914, 70.9671054, 70.0895456, 69.1930506, 68.2779586, 67.3446081, 66.3933376, 65.4244854, 64.4399114, 63.4383996, 62.4202554, 61.3857843, 60.3352917, 59.269083, 58.1874637, 57.0907393, 55.979215, 54.8531965, 53.7129891, 52.5588982, 51.3912294, 50.2102879, 49.0163793, 47.8098091, 46.595659, 45.3788965, 44.1499757, 42.9088924, 41.6556425, 40.3902218, 39.1126259, 37.8228508, 36.5208922, 35.206746, 33.8804079, 32.5418736, 31.1911391, 29.8282001, 28.4530524, 27.0656918, 25.666114, 24.2744346, 22.8801664, 21.4725801, 20.0510189, 18.6148254, 17.1633427, 15.6959135, 14.2118808, 12.7105873, 11.1913761, 9.6535899, 8.09657165, 6.5196642, 4.92221045, 3.30355324, 1.66303547, 0, -1.88400984, -3.76380148, -5.64060458, -7.51564883, -9.39016391, -11.2653795, -13.1425253, -15.0228309, -16.9075261, -18.7978406, -20.6950039, -22.6002459, -24.5147961, -26.4398843, -28.3767401, -30.3265932, -32.1178721, -33.7271204, -35.546428, -37.5399956, -39.6720241, -41.9067144, -44.2082672, -46.5408834, -48.8687638, -51.1561093, -53.3671207, -55.4659988, -57.4169445, -59.2411463, -60.8249219, -62.1153757, -63.0596123, -63.4611327, -63.3388433, -62.7499444, -61.6846217, -60.1330605, -58.0854466, -55.5319653, -54.0077174, -51.4877817, -47.6600777, -42.212525, -34.8330431, -24.2769233, -11.3978704, 3.88303899, 21.6447283, 41.9661208, 64.6084069, 90.3565836, 117.703806, 146.741087, 180.821735, 221.834353, 271.667544, 334.841324, 410.014838, 492.976604, 580.756977, 676.633121, 781.807938, 897.578845, 1025.40342, 1166.66572, 1323.76271, 1498.34105, 1688.33344, 1893.77341, 2115.16294, 2351.99503, 2604.14107, 2871.47241, 3153.86043, 3494.52836, 3839.15776, 4176.78202, 4496.43456, 4787.1488, 4984.55983, 5166.69824, 5345.48241, 5570.81588, 5839.58127, 6132.83402, 6448.24811, 6783.4975, 7136.25615, 7504.19803, 8287.33493, 8774.88892, 9069.47911, 9273.72461, 9490.24455, 9821.65803, 10370.5842, 10852.6295, 11336.9585},
        {92.6925216, 92.0813056, 91.4950123, 90.9321884, 90.3913805, 89.8711354, 89.3699995, 88.8865196, 88.4192423, 87.9667143, 87.5274822, 87.1000927, 86.6830923, 86.2750278, 85.8744458, 85.4798929, 85.0899159, 84.7030612, 84.3178756, 83.9329058, 83.5466983, 83.1577999, 82.7647571, 82.3661166, 81.9604251, 81.5462292, 81.1220755, 80.6865108, 80.2380815, 79.7753345, 79.2968163, 78.8010736, 78.286653, 77.7521012, 77.1408467, 76.4801546, 75.7979524, 75.0945753, 74.3703588, 73.6256382, 72.8607487, 72.0760258, 71.2718047, 70.4484208, 69.6062095, 68.7455061, 67.8666458, 66.9699641, 66.0557962, 65.1244776, 64.1763435, 63.215518, 62.2384657, 61.2454396, 60.236693, 59.212479, 58.173051, 57.118662, 56.0495653, 54.966014, 53.8682615, 52.7565608, 51.6311652, 50.4923279, 49.340302, 48.1753408, 46.9976976, 45.8116898, 44.6215376, 43.4191115, 42.2044014, 40.9773968, 39.7380876, 38.4864635, 37.2225141, 35.9462292, 34.6575986, 33.3566118, 32.0432588, 30.7175291, 29.3794125, 28.0288987, 26.6659774, 25.2906385, 23.9197159, 22.544423, 21.1557704, 19.7532012, 18.3361586, 16.9040856, 15.4564253, 13.9926208, 12.5121152, 11.0143516, 9.49877294, 7.96482246, 6.41194319, 4.83957822, 3.24717065, 1.63416354, 0, -1.92959194, -3.83611302, -5.72224718, -7.59067835, -9.44409049, -11.2851675, -13.1165934, -14.941052, -16.7612273, -18.5798033, -20.3994639, -22.2228929, -24.0527745, -25.8917924, -27.7426306, -29.6079732, -31.2405184, -32.6072382, -34.2762596, -36.1966985, -38.3176702, -40.5882906, -42.9576752, -45.3749397, -47.7891996, -50.1495706, -52.4051683, -54.5051083, -56.3985063, -57.9902731, -59.2899375, -60.2598766, -60.8624675, -60.9773709, -60.608322, -59.7811137, -58.4829379, -56.7009866, -54.4224516, -51.634525, -49.8555418, -47.0755507, -42.982172, -37.2630264, -29.6057341, -19.1565282, -6.27976342, 9.20159297, 27.4645737, 48.6862115, 73.0889662, 100.748921, 129.932676, 160.578072, 196.035076, 238.121677, 288.655869, 351.859024, 426.599529, 509.123902, 597.147677, 693.039665, 796.963545, 909.873408, 1034.06266, 1171.20993, 1324.42369, 1495.75326, 1682.0053, 1882.95198, 2098.2994, 2327.89605, 2571.53702, 2829.01741, 3100.13229, 3423.06599, 3749.62706, 4070.0133, 4374.42248, 4653.05241, 4848.07643, 5029.73304, 5208.80191, 5429.72865, 5689.96683, 5972.9426, 6276.81092, 6599.72677, 6939.84511, 7295.3209, 8172.80969, 8672.54834, 8927.85907, 9072.06411, 9238.48569, 9560.44604, 10171.2674, 10606.7169, 11032.4211},
        {91.5947241, 90.9572207, 90.3459165, 89.7593439, 89.1960355, 88.6545238, 88.1333412, 87.6310202, 87.1460933, 86.677093, 86.2225518, 85.7810021, 85.3509765, 84.9310074, 84.5196273, 84.1153687, 83.7167641, 83.322346, 82.9306468, 82.5401991, 82.1495353, 81.7571879, 81.3616894, 80.9615723, 80.5553691, 80.1416122, 79.7188341, 79.2855674, 78.8403445, 78.3816978, 77.90816, 77.4182634, 76.9105406, 76.383524, 75.7805925, 75.1290155, 74.4567252, 73.764044, 73.0512943, 72.3187983, 71.5668783, 70.7958568, 70.0060559, 69.1977981, 68.3714057, 67.527201, 66.6655063, 65.786644, 64.8909364, 63.9787057, 63.0502744, 62.1111018, 61.1562617, 60.1859652, 59.2004233, 58.199847, 57.1844471, 56.1544347, 55.1100208, 54.0514162, 52.9788321, 51.8924792, 50.7925687, 49.6793115, 48.5529185, 47.4136007, 46.2615692, 45.1004803, 43.933908, 42.754957, 41.5636148, 40.3598691, 39.1437075, 37.9151176, 36.674087, 35.4206033, 34.1546541, 32.8762271, 31.5853097, 30.2818897, 28.9659546, 27.637492, 26.2964896, 24.9429349, 23.5908586, 22.2329314, 20.8616552, 19.4765619, 18.0771834, 16.6630516, 15.2336983, 13.7886555, 12.327455, 10.8496286, 9.35470838, 7.8422261, 6.31171367, 4.76270295, 3.19472585, 1.60731424, 0, -1.80502135, -3.60452111, -5.39972867, -7.19187342, -8.98218475, -10.771892, -12.5622247, -14.3544121, -16.1496836, -17.9492686, -19.7543966, -21.5662968, -23.3861987, -25.2153317, -27.0549251, -28.9062084, -30.5160377, -31.8493034, -33.4826028, -35.3626568, -37.4361864, -39.6499122, -41.9505551, -44.284836, -46.5994757, -48.841195, -50.9567149, -52.892756, -54.5960393, -55.919062, -56.9373172, -57.6257927, -57.9594765, -57.9561419, -57.569384, -56.7413886, -55.434308, -53.6102945, -51.2315004, -48.2600779, -45.8526951, -42.4134403, -37.670756, -31.353085, -23.1888698, -13.1208747, -0.609640445, 14.6699708, 33.0430967, 54.8348751, 80.794577, 110.304822, 141.878038, 175.386612, 213.479515, 257.559087, 309.027668, 370.551841, 441.982377, 521.790875, 609.8909, 706.540484, 810.365281, 921.652154, 1043.50279, 1177.72683, 1327.79291, 1495.94077, 1678.32717, 1874.54977, 2083.39194, 2305.39085, 2540.42602, 2788.37695, 3049.12313, 3358.41446, 3671.29246, 3978.66905, 4271.45615, 4540.56566, 4732.55637, 4912.26208, 5089.60318, 5305.07852, 5556.50453, 5828.95661, 6120.88824, 6430.75291, 6757.00408, 7098.09525, 7942.05606, 8417.06911, 8657.4945, 8797.69233, 8972.02271, 9314.84573, 9960.5215, 10369.4781, 10735.012},
        {90.5861288, 89.9237117, 89.2886904, 88.6795861, 88.0949198, 87.5332129, 86.9929864, 86.4727615, 85.9710594, 85.4864012, 85.0173082, 84.5623015, 84.1199023, 83.6886317, 83.267011, 82.8535612, 82.4468036, 82.0452594, 81.6474496, 81.2518955, 80.8571183, 80.4616391, 80.063979, 79.6626594, 79.2562012, 78.8431258, 78.4219542, 77.9912077, 77.5494074, 77.0950745, 76.6267302, 76.1428955, 75.6420918, 75.1228402, 74.5290114, 73.8876014, 73.2262629, 72.5452904, 71.8449787, 71.1256225, 70.3875165, 69.6309553, 68.8562337, 68.0636462, 67.2534876, 66.4260526, 65.5816358, 64.7205319, 63.8430356, 62.9494416, 62.0400446, 61.1201616, 60.1849561, 59.2346138, 58.2693206, 57.2892623, 56.2946246, 55.2855933, 54.2623544, 53.2250935, 52.1739965, 51.1092492, 50.0310374, 48.9395468, 47.8349634, 46.7174728, 45.587261, 44.4475709, 43.3015722, 42.1431453, 40.9722779, 39.7889576, 38.593172, 37.3849087, 36.1641555, 34.9308998, 33.6851292, 32.4268315, 31.1559943, 29.872605, 28.5766515, 27.2681212, 25.9470018, 24.6132809, 23.2790328, 21.9379475, 20.5835698, 19.2154952, 17.833319, 16.4366365, 15.0250433, 13.5981346, 12.1555059, 10.6967525, 9.22146977, 7.72925317, 6.21969804, 4.69239978, 3.14695376, 1.58295538, 0, -1.45436844, -2.96990469, -4.54240813, -6.16767813, -7.84151405, -9.55971526, -11.3180811, -13.112411, -14.9385043, -16.7921603, -18.6691785, -20.5653582, -22.4764987, -24.3983995, -26.3268598, -28.2576791, -30.0313708, -31.6215504, -33.3752596, -35.255022, -37.2233615, -39.2428019, -41.2758669, -43.2850805, -45.2329663, -47.0820482, -48.7948499, -50.3338954, -51.6617082, -52.674035, -53.4246619, -53.8961459, -54.071044, -54.1825123, -54.0878084, -53.5773631, -52.5585537, -50.9387574, -48.6253515, -45.5257133, -41.9726312, -37.3185983, -31.387759, -24.0042579, -14.9922397, -5.59551589, 6.13635268, 20.7341382, 38.7286128, 60.6505486, 87.8310106, 119.022343, 153.598833, 191.488507, 233.749942, 280.934117, 333.592015, 391.251594, 455.719365, 529.917864, 617.836552, 716.165754, 821.487971, 933.036915, 1054.53863, 1187.6571, 1335.63026, 1500.53018, 1678.65774, 1869.57833, 2071.05402, 2284.73096, 2510.79875, 2749.44703, 3000.86539, 3302.17, 3607.39231, 3907.49029, 4193.42193, 4456.14521, 4643.31634, 4818.06293, 4989.90087, 5197.75331, 5439.49577, 5700.75073, 5980.00634, 6275.75077, 6586.47219, 6910.65874, 7549.9169, 7951.18045, 8209.81449, 8421.1841, 8680.65436, 9083.59036, 9725.35718, 10142.103, 10447.5125},
        {89.7128469, 89.0260346, 88.3677571, 87.7365233, 87.1308425, 86.5492237, 85.9901761, 85.4522089, 84.9338311, 84.4335518, 83.9498803, 83.4813256, 83.0263968, 82.583603, 82.1514535, 81.7284573, 81.3131235, 80.9039613, 80.4994797, 80.098188, 79.6985952, 79.2992105, 78.898543, 78.4951018, 78.087396, 77.6739348, 77.2532272, 76.8237825, 76.3841097, 75.932718, 75.4681165, 74.9888143, 74.4933205, 73.9801443, 73.3934347, 72.7600243, 72.107399, 71.435832, 70.7455965, 70.0369658, 69.3102129, 68.5656112, 67.8034339, 67.0239541, 66.2274451, 65.41418, 64.5844322, 63.7384748, 62.8765811, 61.9990241, 61.1060772, 60.2031092, 59.2851871, 58.3524738, 57.405132, 56.4433245, 55.467214, 54.4769632, 53.472735, 52.4546921, 51.4229971, 50.377813, 49.3193024, 48.247628, 47.1629527, 46.0654391, 44.9552501, 43.8353104, 42.7084786, 41.5692209, 40.4175208, 39.2533623, 38.0767288, 36.8876042, 35.6859721, 34.4718162, 33.2451202, 32.0058678, 30.7540427, 29.4896286, 28.2126093, 26.9229683, 25.6206894, 24.3057563, 22.9885702, 21.663687, 20.3255372, 18.9737665, 17.6080205, 16.2279447, 14.8331849, 13.4233866, 11.9981955, 10.5572572, 9.10021731, 7.62672148, 6.13641533, 4.62894448, 3.10395456, 1.56109119, 0, -1.09391923, -2.31891498, -3.66508802, -5.12253909, -6.68136897, -8.33167841, -10.0635682, -11.867139, -13.7324917, -15.649727, -17.6089456, -19.6002484, -21.613736, -23.6395093, -25.667669, -27.6883158, -29.6488018, -31.5331206, -33.4280669, -35.3145808, -37.1736027, -38.986073, -40.7329318, -42.3951196, -43.9535765, -45.389243, -46.6830592, -47.8159655, -48.7689022, -49.5086124, -50.0354393, -50.3345822, -50.3912406, -50.6147527, -50.7782485, -50.5417545, -49.7632286, -48.3006286, -46.0119123, -42.7550377, -38.1129477, -32.3023152, -25.2349052, -16.8224827, -6.97681285, 1.91172422, 13.0296323, 27.0848, 44.7851162, 66.8384694, 95.0932489, 127.722898, 164.728387, 206.063207, 251.583908, 301.190562, 354.783245, 409.519449, 468.665143, 538.478198, 625.903395, 725.148137, 831.639082, 943.894031, 1065.66855, 1198.314, 1344.46031, 1505.7903, 1679.29867, 1864.63201, 2059.20039, 2265.23095, 2483.14441, 2713.36151, 2956.303, 3250.26826, 3548.32971, 3841.43841, 4120.54544, 4376.60187, 4557.74576, 4726.2832, 4891.51369, 5091.82609, 5325.00536, 5576.71593, 5845.3488, 6129.29495, 6426.94541, 6736.69116, 7131.97336, 7452.97017, 7743.71019, 8048.22206, 8410.53442, 8874.67587, 9484.67506, 9906.61375, 10153.753},
        {89.0085334, 88.2955434, 87.6122777, 86.9572271, 86.3288825, 85.7257347, 85.1462745, 84.5889928, 84.0523805, 83.5349283, 83.0351272, 82.5514679, 82.0824413, 81.6265382, 81.1822495, 80.7480661, 80.3224787, 79.9039781, 79.4910554, 79.0822012, 78.6759064, 78.2706619, 77.8649585, 77.457287, 77.0461383, 76.6300033, 76.2073727, 75.7767374, 75.3365883, 74.8854161, 74.4217118, 73.9439662, 73.4506701, 72.9403143, 72.3564175, 71.7261128, 71.0771943, 70.4099367, 69.7246149, 69.0215036, 68.3008775, 67.5630116, 66.8081806, 66.0366593, 65.2487225, 64.444645, 63.6247016, 62.7891671, 61.9383163, 61.0724239, 60.1917648, 59.3032599, 58.4003938, 57.4832972, 56.5521009, 55.6069355, 54.6479317, 53.6752204, 52.6889322, 51.6891978, 50.6761479, 49.6499133, 48.6106247, 47.5584129, 46.4934084, 45.4157421, 44.3255447, 43.2253997, 42.1178319, 40.9979131, 39.865615, 38.7209094, 37.5637679, 36.3941623, 35.2120643, 34.0174455, 32.8102778, 31.5905328, 30.3581821, 29.1131977, 27.855551, 26.5852139, 25.3021581, 24.0063552, 22.7073124, 21.400033, 20.0794059, 18.7450935, 17.3967581, 16.0340618, 14.656667, 13.264236, 11.856431, 10.4329142, 8.99334805, 7.5373947, 6.06471646, 4.57497559, 3.06783438, 1.54295509, 0, -0.867727801, -1.91132946, -3.11711646, -4.47140029, -5.96049246, -7.57070445, -9.28834776, -11.0997339, -12.9911743, -14.9489805, -16.959464, -19.0089362, -21.0837088, -23.1700931, -25.2544006, -27.3229429, -29.4032928, -31.4876563, -33.4690155, -35.3425238, -37.1033343, -38.7466001, -40.2674747, -41.6611111, -42.9226625, -44.0472823, -45.0301235, -45.8663395, -46.5510834, -47.1345313, -47.5366385, -47.7360514, -47.7114163, -47.904964, -48.0635485, -47.8104016, -46.9850941, -45.4271972, -42.9762819, -39.471919, -34.21867, -27.7539446, -20.0219896, -10.9670519, -0.533378296, 8.40862282, 19.5724066, 33.7452666, 51.7144964, 74.2673896, 103.660958, 137.416454, 174.977242, 216.110344, 261.243594, 310.508135, 364.035112, 419.876072, 480.714388, 551.502085, 637.741056, 735.183735, 841.005651, 954.213978, 1076.91951, 1209.80842, 1354.08992, 1510.58579, 1678.20003, 1856.92147, 2045.88646, 2246.0674, 2457.74819, 2681.21272, 2916.74489, 3201.06163, 3488.90553, 3771.45221, 4039.8773, 4285.35642, 4456.44616, 4615.3539, 4771.52056, 4964.91207, 5192.88655, 5440.0834, 5704.49668, 5984.12048, 6276.94886, 6580.97588, 6733.55304, 6986.40502, 7324.75915, 7733.84276, 8198.88319, 8705.10777, 9237.74383, 9649.0957, 9849.44914},
        {88.4407418, 87.7014209, 86.993027, 86.314028, 85.6628916, 85.0380858, 84.4380782, 83.8613368, 83.3063293, 82.7715237, 82.2553876, 81.756389, 81.2729957, 80.8036755, 80.3468962, 79.9011256, 79.4648316, 79.0364821, 78.6145447, 78.1974874, 77.783778, 77.3718843, 76.9602741, 76.5474153, 76.1317756, 75.711823, 75.2860252, 74.8528501, 74.4107654, 73.9582391, 73.4937389, 73.0157327, 72.5226883, 72.0130735, 71.4291804, 70.7987463, 70.1501837, 69.4837834, 68.7998362, 68.0986329, 67.3804644, 66.6456216, 65.8943952, 65.1270761, 64.3439551, 63.5453231, 62.7314708, 61.9026892, 61.0592691, 60.2015013, 59.3296766, 58.4529162, 57.5624891, 56.6584948, 55.7410328, 54.8102023, 53.8661029, 52.908834, 51.9384949, 50.9551851, 49.959004, 48.9500509, 47.9284254, 46.8942268, 45.8475545, 44.788508, 43.7171866, 42.6360174, 41.5473712, 40.4465473, 39.3334941, 38.2081602, 37.0704941, 35.9204444, 34.7579595, 33.582988, 32.3954784, 31.1953793, 29.9826391, 28.7572065, 27.5190299, 26.2680578, 25.0042389, 23.7275215, 22.4464171, 21.1564126, 19.8528922, 18.5355265, 17.2039862, 15.8579419, 14.4970643, 13.1210241, 11.7294919, 10.3221383, 8.89863397, 7.45864963, 6.00185589, 4.52792341, 3.03652283, 1.52732481, 0, -0.724398496, -1.65303804, -2.76976197, -4.05841365, -5.50283642, -7.08687361, -8.79436859, -10.6091647, -12.5151053, -14.4960337, -16.5357933, -18.6182274, -20.7271793, -22.8464925, -24.9600102, -27.0515758, -29.2060811, -31.4218049, -33.4540287, -35.308249, -36.9899624, -38.5046654, -39.8578548, -41.055027, -42.1016786, -43.0033063, -43.7654065, -44.3934759, -44.8930111, -45.400431, -45.7423049, -45.8848528, -45.7942943, -45.8499388, -45.8114613, -45.32838, -44.2429879, -42.397578, -39.6344435, -35.7958773, -30.2204686, -23.4075157, -15.2978622, -5.83235198, 5.04817106, 14.4911054, 26.1953044, 40.9478637, 59.5358788, 82.7464454, 113.053667, 147.495732, 184.451578, 223.441018, 266.313161, 313.871309, 366.918767, 426.081605, 492.380638, 567.030019, 651.781549, 746.013683, 850.057496, 964.028645, 1087.67778, 1220.92309, 1363.31251, 1514.66823, 1676.1701, 1848.22999, 2032.55593, 2228.06424, 2434.71816, 2652.4809, 2881.31567, 3154.66082, 3430.63567, 3700.83465, 3956.85221, 4190.28277, 4350.65433, 4499.67208, 4646.95896, 4834.22663, 5058.39623, 5303.01855, 5565.68323, 5843.97991, 6135.49822, 6437.82779, 6376.83566, 6576.71787, 6966.04155, 7473.37384, 8027.28187, 8556.33278, 8989.0937, 9371.97542, 9526.63823},
        {87.976596, 87.2124948, 86.4804962, 85.7790417, 85.1065729, 84.4615315, 83.8423592, 83.2474975, 82.675388, 82.1244724, 81.5931923, 81.0799893, 80.583305, 80.1015811, 79.6332591, 79.1767807, 78.7305876, 78.2931212, 77.8628233, 77.4381355, 77.0174994, 76.5993565, 76.1821486, 75.7643173, 75.3443041, 74.9205507, 74.4914987, 74.0555897, 73.6112653, 73.1569673, 72.6911371, 72.2122164, 71.7186468, 71.20887, 70.6236679, 69.9915247, 69.3416178, 68.6742598, 67.9897636, 67.288442, 66.5706078, 65.8365737, 65.0866525, 64.3211571, 63.5404002, 62.7446946, 61.934353, 61.1096884, 60.2710133, 59.4186408, 58.5528834, 57.6848468, 56.803817, 55.9098729, 55.0030932, 54.0835569, 53.1513427, 52.2065294, 51.2491959, 50.2794211, 49.2972836, 48.3028624, 47.2962363, 46.2774841, 45.2466845, 44.2039166, 43.1492589, 42.0853873, 41.014915, 39.9325651, 38.838248, 37.7318743, 36.6133543, 35.4825987, 34.3395179, 33.1840223, 32.0160225, 30.8354289, 29.6421519, 28.4361022, 27.2171902, 25.9853262, 24.740421, 23.4823848, 22.2177087, 20.9428744, 19.6542849, 18.3516368, 17.0346273, 15.7029532, 14.3563115, 12.9943991, 11.616913, 10.22355, 8.81400707, 7.3879812, 5.9451693, 4.48526829, 3.00797512, 1.51298671, 0, -0.612289368, -1.4494172, -2.49349938, -3.72665183, -5.13099044, -6.68863111, -8.38168974, -10.1922822, -12.1025245, -14.0945324, -16.1504219, -18.2523089, -20.3823093, -22.5225389, -24.6551137, -26.7621496, -28.9664307, -31.2701684, -33.3373004, -35.1800858, -36.8107837, -38.2416532, -39.4849534, -40.5529434, -41.4578822, -42.2120291, -42.8276431, -43.3169833, -43.6923088, -44.1698392, -44.4830876, -44.5831249, -44.421022, -44.2628118, -43.9020843, -43.0484396, -41.55846, -39.2887278, -36.0958254, -31.8363351, -26.0407598, -18.9910029, -10.6074448, -0.810465944, 10.4795531, 20.7661737, 33.3490881, 48.9519304, 68.2983346, 92.1119347, 122.800808, 157.345382, 193.238649, 229.847095, 270.362251, 316.258281, 369.009348, 431.911127, 504.20229, 583.134408, 666.490461, 757.409111, 859.2788, 973.356485, 1097.28745, 1230.37479, 1370.85083, 1517.74107, 1673.99716, 1840.34991, 2020.67059, 2212.06638, 2414.18097, 2626.65807, 2849.14138, 3111.13951, 3374.96505, 3632.79547, 3876.80827, 4099.18092, 4251.63429, 4393.77356, 4535.11479, 4717.26929, 4937.04871, 5177.89168, 5437.27853, 5712.68956, 6001.60509, 6301.50542, 6085.89303, 6251.59118, 6682.70401, 7263.33562, 7877.59016, 8409.57173, 8743.38446, 9077.30144, 9176.83762},
        {87.5465505, 86.7651477, 86.0167968, 85.2999152, 84.6129199, 83.9542279, 83.3222565, 82.7154227, 82.1321437, 81.5708365, 81.0299183, 80.5078062, 80.0029173, 79.5136688, 79.0384777, 78.5757612, 78.1239363, 77.6814202, 77.2466301, 76.8179829, 76.3938959, 75.9727862, 75.5530708, 75.1331669, 74.7114916, 74.2864621, 73.8564953, 73.4200085, 72.9754188, 72.5211432, 72.055599, 71.5772031, 71.0843728, 70.5755251, 69.9902218, 69.3575455, 68.7072866, 68.039772, 67.3553289, 66.6542843, 65.9369653, 65.203699, 64.4548125, 63.6906329, 62.9114872, 62.1177026, 61.3096061, 60.4875249, 59.6517859, 58.8027164, 57.9406433, 57.0764988, 56.199775, 55.3105692, 54.4089786, 53.4951002, 52.5690314, 51.6308693, 50.6807112, 49.7186541, 48.7447953, 47.7592319, 46.7620613, 45.7533805, 44.7332868, 43.7018773, 42.6592492, 41.6095942, 40.5570053, 39.4931356, 38.4178167, 37.3308805, 36.2321587, 35.1214832, 33.9986856, 32.8635977, 31.7160514, 30.5558784, 29.3829104, 28.1969792, 26.9979167, 25.7855545, 24.5597245, 23.3202584, 22.0674903, 20.8009903, 19.5203224, 18.2253021, 16.9157447, 15.5914657, 14.2522807, 12.8980051, 11.5284544, 10.143444, 8.74278943, 7.32630611, 5.89380953, 4.44511517, 2.98003848, 1.49839493, 0, -0.414221841, -1.08435464, -1.99049178, -3.11272663, -4.43115259, -5.92586304, -7.57695135, -9.36451091, -11.2686351, -13.2694173, -15.3469509, -17.4813293, -19.6526458, -21.8409939, -24.0264668, -26.1891581, -28.4760389, -30.8910421, -33.0367996, -34.9291644, -36.5839892, -38.017127, -39.2444307, -40.2817532, -41.1449474, -41.8498663, -42.4123626, -42.8482893, -43.1734993, -43.685591, -44.0153651, -44.0941509, -43.8532778, -43.3990777, -42.5753786, -41.1793409, -39.089793, -36.1855634, -32.3454803, -27.4483723, -21.2523951, -13.7937762, -4.97495397, 5.30163324, 17.1335472, 28.7532071, 42.5896027, 59.2065812, 79.1679901, 103.037677, 132.405504, 165.555284, 200.745158, 237.674774, 278.986855, 326.00274, 380.043773, 444.375556, 517.933292, 597.531169, 680.26423, 769.84413, 869.834973, 981.665151, 1103.14421, 1233.74281, 1372.1832, 1517.74199, 1672.43999, 1836.74575, 2013.40652, 2200.26155, 2396.99062, 2603.27348, 2818.78989, 3069.51836, 3322.2652, 3570.1355, 3806.23431, 4023.6667, 4180.44679, 4328.16458, 4474.96413, 4654.62187, 4865.1357, 5093.82339, 5339.21687, 5599.84807, 5874.24891, 6160.95132, 5957.45285, 6130.29601, 6561.5537, 7133.2988, 7727.60423, 8226.54286, 8512.18761, 8755.91429, 8768.76092},
        {87.2064897, 86.4079694, 85.6435214, 84.911533, 84.2103917, 83.5384849, 82.8942, 82.2759244, 81.6820455, 81.1109507, 80.5610275, 80.0306632, 79.5182453, 79.0221611, 78.5407982, 78.0725439, 77.6157855, 77.1689106, 76.7303066, 76.2983608, 75.8714606, 75.4479936, 75.026347, 74.6049084, 74.182065, 73.7562044, 73.3257139, 72.888981, 72.444393, 71.9903374, 71.5252016, 71.047373, 70.5552391, 70.0471871, 69.4613493, 68.8275082, 68.1761703, 67.5076785, 66.8223756, 66.1206043, 65.4027076, 64.6690282, 63.9199089, 63.1556926, 62.376722, 61.58334, 60.7758893, 59.9547128, 59.1201534, 58.2725537, 57.4122566, 56.5496811, 55.6748753, 54.7879637, 53.8890708, 52.9783207, 52.055838, 51.1217471, 50.1761722, 49.2192378, 48.2510683, 47.271788, 46.2815214, 45.2803928, 44.2685266, 43.2460471, 42.2130788, 41.1755417, 40.1392166, 39.0922749, 38.0344653, 36.9655363, 35.8852367, 34.7933152, 33.6895204, 32.5736011, 31.4453058, 30.3043833, 29.1505822, 27.9836512, 26.8033391, 25.6093943, 24.4015658, 23.179602, 21.9370628, 20.6769215, 19.4022258, 18.1129251, 16.8089691, 15.4903072, 14.1568889, 12.8086638, 11.4455814, 10.0675911, 8.67464257, 7.26668524, 5.84366865, 4.40554231, 2.95225575, 1.48375847, 0, -0.262939405, -0.802224863, -1.59680186, -2.6256159, -3.86761246, -5.30173703, -6.90693511, -8.66215218, -10.5463337, -12.5384253, -14.6173723, -16.7621202, -18.9516146, -21.164801, -23.3806248, -25.5780315, -27.8993687, -30.3469251, -32.5300461, -34.462692, -36.158823, -37.6323994, -38.8973813, -39.967729, -40.8574027, -41.5803627, -42.1505691, -42.5819823, -42.8885624, -43.4082802, -43.7122816, -43.7173235, -43.3401631, -42.5515768, -41.241312, -39.2847107, -36.5823243, -33.0347039, -28.5424006, -23.0059658, -16.4188378, -8.56041063, 0.686938127, 11.4408308, 23.8188899, 36.8695678, 52.0469492, 69.7359491, 90.3214821, 114.188463, 141.964708, 173.495352, 208.165295, 246.104397, 288.96348, 337.710486, 393.313356, 458.282834, 531.693422, 610.936565, 693.432111, 782.461454, 880.876069, 989.961847, 1108.35189, 1235.89697, 1372.2753, 1517.29292, 1671.3886, 1834.64322, 2007.9767, 2190.50176, 2382.00883, 2582.28833, 2791.13069, 3031.20207, 3273.69823, 3512.69063, 3742.25077, 3956.45013, 4120.32057, 4276.33296, 4431.00433, 4608.58067, 4808.56132, 5023.05854, 5251.94101, 5495.07741, 5752.33643, 6023.58675, 5891.05076, 6086.70918, 6504.78808, 7039.51354, 7585.11164, 8035.80846, 8285.83007, 8424.82015, 8336.65722},
        {86.969027, 86.1508408, 85.3679434, 84.6186844, 83.9014136, 83.2144807, 82.5562353, 81.9250271, 81.3192058, 80.7371211, 80.1771226, 79.63756, 79.116783, 78.6131413, 78.1249846, 77.6506624, 77.1885246, 76.7369208, 76.2942006, 75.8587138, 75.42881, 75.0028389, 74.5791502, 74.1560935, 73.7320186, 73.3052751, 72.8742127, 72.4371811, 71.9925299, 71.5386088, 71.0737675, 70.5963558, 70.1047231, 69.5972194, 69.0102086, 68.3743371, 67.720996, 67.0505465, 66.3633501, 65.6597677, 64.9401607, 64.2048903, 63.4543177, 62.6888042, 61.908711, 61.1143992, 60.3062302, 59.4845652, 58.6497653, 57.8021919, 56.9422061, 56.0798946, 55.2056824, 54.3197199, 53.4221575, 52.5131455, 51.5928344, 50.6613745, 49.7189163, 48.7656101, 47.8016064, 46.8270556, 45.842108, 44.846914, 43.8416241, 42.8263886, 41.801358, 40.7739998, 39.751606, 38.7192357, 37.6765651, 36.6232702, 35.559027, 34.4835116, 33.3964, 32.2973684, 31.1860926, 30.0622489, 28.9255133, 27.7755617, 26.6120704, 25.4347152, 24.2431723, 23.0371178, 21.8039507, 20.549744, 19.2807177, 17.9969464, 16.6985045, 15.3854664, 14.0579068, 12.7158999, 11.3595203, 9.98884235, 8.60394064, 7.20488957, 5.79176361, 4.3646372, 2.92358483, 1.46868094, 0, -0.204967233, -0.688543254, -1.42980915, -2.40784599, -3.60173487, -4.99055687, -6.55339306, -8.26932454, -10.1174324, -12.0767977, -14.1265015, -16.2456249, -18.413249, -20.6084549, -22.8103237, -24.9979363, -27.2722861, -29.6298699, -31.7740156, -33.7099282, -35.4428128, -36.9778745, -38.3203184, -39.4753495, -40.448173, -41.2439939, -41.8680173, -42.3254483, -42.6214921, -43.068359, -43.2516801, -43.0845589, -42.4800989, -41.3452465, -39.5961838, -37.1507345, -33.9238492, -29.8304784, -24.7855727, -18.7040827, -11.7618093, -3.53346344, 6.11704037, 17.3257875, 30.2288634, 44.6027545, 61.0330449, 79.7457197, 100.966764, 124.922163, 151.370512, 181.57644, 216.076599, 255.408463, 300.111188, 350.723177, 407.782831, 472.650694, 544.856174, 623.031797, 705.656572, 794.863106, 892.247403, 998.722661, 1114.0451, 1238.50202, 1372.81987, 1517.3998, 1671.0327, 1833.42028, 2003.65127, 2182.13473, 2368.7846, 2563.5148, 2766.23927, 2996.77293, 3230.15349, 3461.31962, 3685.21001, 3896.76335, 4067.5357, 4231.4368, 4393.42645, 4568.44901, 4757.51209, 4957.46305, 5169.51753, 5394.89113, 5634.79947, 5890.45817, 5854.24987, 6079.3638, 6478.29142, 6963.52418, 7447.55351, 7842.87089, 8061.96774, 8091.99857, 7892.16714},
        {86.8415902, 85.9915347, 85.1787927, 84.4016498, 83.6583916, 82.9473039, 82.2666721, 81.614782, 80.9899191, 80.3903692, 79.8144178, 79.2603506, 78.7264533, 78.2110114, 77.7123107, 77.2286366, 76.758275, 76.2995113, 75.8506314, 75.4099207, 74.9756649, 74.5461497, 74.1196606, 73.6944834, 73.2689037, 72.8412071, 72.4096792, 71.9726057, 71.5282722, 71.0749644, 70.6109679, 70.1345682, 69.6440512, 69.1377023, 68.5489151, 67.9097862, 67.2531922, 66.5795246, 65.8891749, 65.1825347, 64.4599954, 63.7219486, 62.9687858, 62.2008984, 61.418678, 60.6225162, 59.8128044, 58.9899341, 58.1542969, 57.3062842, 56.4462876, 55.5862609, 54.7147826, 53.8319936, 52.9380348, 52.0330471, 51.1171713, 50.1905482, 49.2533188, 48.3056239, 47.3476044, 46.3794012, 45.401155, 44.4130068, 43.4150975, 42.4075678, 41.3905587, 40.3714674, 39.3575168, 38.3338822, 37.300234, 36.2562427, 35.2015787, 34.1359126, 33.0589148, 31.9702558, 30.8696061, 29.756636, 28.6310161, 27.4924169, 26.3405088, 25.1749622, 23.9954477, 22.8016357, 21.5800568, 20.3372277, 19.0798232, 17.8079402, 16.5216756, 15.2211262, 13.906389, 12.5775609, 11.2347386, 9.8780192, 8.50749946, 7.12327629, 5.72544658, 4.31410721, 2.88935507, 1.45128703, 0, -0.498471856, -1.21423903, -2.13064484, -3.23103264, -4.49874576, -5.91712752, -7.46952126, -9.13927032, -10.909718, -12.7642077, -14.6860827, -16.6586863, -18.665362, -20.6894529, -22.7143025, -24.7232541, -26.6674162, -28.5255274, -30.3538427, -32.128798, -33.8268293, -35.4243725, -36.8978634, -38.2237379, -39.378432, -40.3383814, -41.0800222, -41.5797902, -41.8141213, -41.840527, -41.5246398, -40.8185731, -39.67444, -38.2460061, -36.3845586, -33.8876105, -30.6467793, -26.5536825, -21.4999376, -15.3771623, -8.24059038, 0.231573322, 10.1797232, 21.7442538, 35.0655594, 49.937671, 66.9339373, 86.2813435, 108.206875, 132.937517, 160.289182, 191.40235, 226.933235, 267.400114, 313.039641, 364.214916, 421.289035, 484.761534, 554.828065, 631.535441, 714.974425, 805.623904, 903.888851, 1010.12015, 1124.57705, 1247.56083, 1379.75161, 1521.54888, 1671.96328, 1830.79106, 1997.64137, 2172.52637, 2355.3071, 2545.8446, 2743.9999, 2968.05875, 3194.8513, 3419.63239, 3637.65689, 3844.17965, 4013.05435, 4175.20447, 4335.0568, 4503.91329, 4683.47931, 4872.59558, 5073.11071, 5286.87331, 5515.73198, 5761.53532, 5762.999, 6002.34564, 6395.74925, 6859.38381, 7309.42332, 7662.04178, 7833.41318, 7784.35146, 7459.05174},
        {86.8441654, 85.9571453, 85.1097063, 84.3000627, 83.526429, 82.7870196, 82.0800488, 81.4037312, 80.756281, 80.1359128, 79.5408409, 78.9692798, 78.4194438, 77.8895474, 77.377805, 76.882431, 76.4016398, 75.9336458, 75.4766635, 75.0289072, 74.5885914, 74.1539304, 73.7231388, 73.2944308, 72.866021, 72.4361237, 72.0029533, 71.5647243, 71.119651, 70.665948, 70.2018295, 69.72551, 69.2352039, 68.7291257, 68.1374114, 67.4937464, 66.8326347, 66.1544998, 65.4597655, 64.7488555, 64.0221934, 63.2802029, 62.5233077, 61.7519315, 60.9664979, 60.1674308, 59.3551537, 58.5300903, 57.6926643, 56.8432995, 55.9824194, 55.1244363, 54.2554822, 53.3756773, 52.4851423, 51.5839976, 50.6723635, 49.7503606, 48.8181093, 47.87573, 46.9233432, 45.9610693, 44.9890287, 44.007342, 43.0161295, 42.0155116, 41.0056089, 39.9932204, 38.9849848, 37.9672484, 36.9396987, 35.9020232, 34.8539094, 33.7950447, 32.7251168, 31.643813, 30.5508208, 29.4458279, 28.3285216, 27.1985894, 26.0557189, 24.8995976, 23.7299129, 22.5463523, 21.3362008, 20.1056081, 18.8608728, 17.6020849, 16.3293344, 15.0427113, 13.7423055, 12.4282072, 11.1005061, 9.75929243, 8.40465602, 7.0366869, 5.65547505, 4.26111045, 2.85368309, 1.43328294, 0, -0.88908598, -1.91688119, -3.07206583, -4.3433201, -5.71932421, -7.18875835, -8.74030273, -10.3626376, -12.044443, -13.7743993, -15.5411867, -17.3334853, -19.1399753, -20.949337, -22.7502506, -24.5313962, -26.0683643, -27.3193936, -28.7551533, -30.3186613, -31.9529358, -33.6009949, -35.2058567, -36.7105393, -38.0580608, -39.1914393, -40.0536929, -40.5878398, -40.736898, -40.2453544, -39.3275532, -37.9860717, -36.2234875, -34.5265803, -32.6558264, -30.2425816, -27.1441627, -23.2178863, -18.3210694, -12.3110287, -5.02830723, 3.64789941, 13.8569927, 25.7383739, 39.4314447, 54.4980689, 71.7995697, 91.6197329, 114.242344, 139.951189, 168.93387, 201.687914, 238.644732, 280.161509, 326.443895, 377.765576, 434.400238, 496.100232, 563.779066, 638.918979, 723.359353, 815.806252, 915.511286, 1022.21852, 1136.50647, 1258.57063, 1388.73238, 1527.21987, 1673.79956, 1828.49905, 1992.14445, 2163.84202, 2343.34299, 2530.39855, 2724.75993, 2943.87098, 3165.3671, 3384.57631, 3596.82666, 3797.4462, 3960.71705, 4117.04378, 4270.7741, 4432.16633, 4603.22857, 4783.5062, 4975.13045, 5180.23258, 5400.94383, 5639.39546, 5658.24178, 5905.81323, 6297.25727, 6747.72133, 7172.35286, 7486.29929, 7604.70807, 7482.95624, 7031.71986},
        {86.9930403, 86.0661316, 85.1810957, 84.336077, 83.5292194, 82.7586671, 82.0225643, 81.3190549, 80.6462833, 80.0023934, 79.3855294, 78.7938354, 78.2254555, 77.6785338, 77.1512145, 76.6416417, 76.1479595, 75.668312, 75.2008433, 74.7436976, 74.2950189, 73.8529514, 73.4156392, 72.9812264, 72.5478571, 72.1136755, 71.6768257, 71.2354517, 70.7876978, 70.3317079, 69.8656263, 69.387597, 68.8957642, 68.388272, 67.7921041, 67.1424816, 66.4754808, 65.7915551, 65.0911581, 64.374743, 63.6427634, 62.8956725, 62.1339239, 61.3579709, 60.568267, 59.7652654, 58.9494198, 58.1211834, 57.2810097, 56.4293521, 55.566664, 54.7098672, 53.8425898, 52.9649281, 52.0769785, 51.1788374, 50.2706013, 49.3523665, 48.4242295, 47.4862866, 46.5386343, 45.5813689, 44.614587, 43.6383848, 42.6528588, 41.6581054, 40.654221, 39.6472247, 38.6429937, 37.6294089, 36.6061828, 35.573028, 34.5296568, 33.4757817, 32.4111152, 31.3353697, 30.2482577, 29.1494916, 28.0387839, 26.915847, 25.7803935, 24.6321357, 23.4707862, 22.2960573, 21.096401, 19.8773969, 18.6447598, 17.3985677, 16.1388985, 14.8658301, 13.5794402, 12.2798068, 10.9670079, 9.64112117, 8.30222462, 6.95039611, 5.58571352, 4.20825473, 2.81809762, 1.41532009, 0, -1.27338614, -2.6089181, -4.00037169, -5.44152274, -6.92614705, -8.44802046, -10.0009188, -11.5786178, -13.1748934, -14.7835213, -16.3982775, -18.0129376, -19.6212775, -21.2170731, -22.7941001, -24.3461344, -25.4849522, -26.1497579, -27.1983741, -28.5427195, -30.0947129, -31.766273, -33.4693186, -35.1157685, -36.6175414, -37.8865561, -38.8347314, -39.373986, -39.4162386, -38.4070459, -36.8956882, -34.9339931, -32.5737881, -30.6330751, -28.7805944, -26.4807729, -23.5555857, -19.8270079, -15.1170147, -9.24758111, -1.82411776, 7.0489248, 17.5072001, 29.6863618, 43.7220633, 58.8360407, 76.3063445, 96.4971078, 119.772464, 146.496545, 177.356669, 211.998781, 250.459033, 292.873599, 339.582865, 390.835529, 446.880289, 506.914014, 572.476285, 646.254137, 731.62561, 825.73995, 926.893024, 1034.2919, 1148.68764, 1270.12257, 1398.44945, 1533.66145, 1676.44694, 1827.10105, 1987.88946, 2156.83358, 2333.54642, 2517.64099, 2708.73031, 2923.9094, 3140.93875, 3355.06087, 3561.51828, 3755.55349, 3910.57163, 4058.21082, 4203.07245, 4356.4101, 4520.19198, 4693.61456, 4878.78464, 5077.80902, 5292.79451, 5525.84791, 5544.11022, 5793.96088, 6185.68276, 6629.55873, 7035.87165, 7314.9044, 7376.93983, 7182.38588, 6613.22474},
        {87.3325968, 86.3670406, 85.4451757, 84.5650964, 83.7248971, 82.9226719, 82.1565153, 81.4245214, 80.7247846, 80.0553993, 79.4144596, 78.8000599, 78.2102945, 77.6432577, 77.0970437, 76.5697469, 76.0594616, 75.564282, 75.0823025, 74.6116174, 74.1503209, 73.6965073, 73.248271, 72.8037062, 72.3609073, 71.9179685, 71.4729841, 71.0240484, 70.5692558, 70.1067005, 69.6344767, 69.1506789, 68.6534013, 68.1407383, 67.537509, 66.8799866, 66.2053248, 65.5139958, 64.8064718, 64.083225, 63.3447276, 62.5914518, 61.8238698, 61.0424539, 60.2476763, 59.4400091, 58.6199246, 57.787895, 56.9443924, 56.0898892, 55.2248574, 54.368059, 53.50128, 52.6245961, 51.7380832, 50.8418169, 49.9358729, 49.020327, 48.095255, 47.1607324, 46.2168352, 45.2636389, 44.3012193, 43.3296522, 42.3490133, 41.3593783, 40.3608229, 39.3588973, 38.3590205, 37.3499773, 36.3314885, 35.3032751, 34.2650578, 33.2165574, 32.1574949, 31.087591, 30.0065666, 28.9141425, 27.8100395, 26.6939786, 25.5656805, 24.424866, 23.271256, 22.1045714, 20.9127013, 19.7015312, 18.4770889, 17.2394791, 15.9888068, 14.7251766, 13.4486934, 12.1594619, 10.857587, 9.54317352, 8.21632616, 6.87714975, 5.5257491, 4.16222901, 2.78669426, 1.39924966, 0, -1.42446384, -2.87976439, -4.36150826, -5.86530207, -7.38675244, -8.92146597, -10.4650493, -12.013109, -13.5612517, -15.1050841, -16.6402127, -18.1622441, -19.6667851, -21.1494421, -22.6058218, -24.0315308, -24.9761612, -25.3716033, -26.222306, -27.4283015, -28.8896218, -30.5062989, -32.178365, -33.8058521, -35.2887923, -36.5272176, -37.4211602, -37.8706521, -37.7757253, -36.4651581, -34.6196962, -32.3107478, -29.6097213, -27.4733004, -25.5302558, -23.1732843, -20.2082115, -16.440863, -11.6770644, -5.72264131, 1.93685754, 11.0778542, 21.8318602, 34.3303871, 48.7049465, 63.9355814, 81.5931391, 102.096998, 125.866537, 153.321135, 185.490353, 221.437608, 260.837838, 303.666117, 350.510292, 401.683089, 457.497233, 517.175764, 582.368747, 655.914994, 741.520792, 835.963009, 937.334787, 1044.69859, 1158.75932, 1279.468, 1406.42585, 1539.49318, 1679.81283, 1827.80217, 1986.61791, 2153.51665, 2327.96753, 2509.43967, 2697.40223, 2908.76519, 3121.19663, 3329.80547, 3529.70064, 3715.99106, 3860.37165, 3996.97469, 4130.94368, 4276.70805, 4435.73563, 4605.62509, 4788.03773, 4984.63482, 5197.07764, 5427.02748, 5409.05865, 5648.87943, 6044.73455, 6494.86873, 6897.5267, 7150.95318, 7153.3929, 6869.09162, 6207.89389},
        {87.8762463, 86.8746822, 85.9180021, 85.0042765, 84.1315762, 83.297972, 82.5015346, 81.7403348, 81.0124431, 80.3159305, 79.6488676, 79.0093251, 78.3953739, 77.8050845, 77.2365278, 76.6877745, 76.1568952, 75.6419609, 75.1410421, 74.6522096, 74.1735341, 73.7030864, 73.2389373, 72.7791573, 72.3218173, 71.8649881, 71.4067402, 70.9451446, 70.4782718, 70.0041926, 69.5209778, 69.0266981, 68.5194243, 67.9972269, 67.3840347, 66.7165488, 66.0323618, 65.3319504, 64.6157915, 63.8843618, 63.138138, 62.3775971, 61.6032158, 60.8154708, 60.014839, 59.2017972, 58.3768221, 57.5403905, 56.6929793, 55.8350651, 54.9671249, 54.108696, 53.2407814, 52.3634447, 51.4767495, 50.5807593, 49.6755378, 48.7611485, 47.837655, 46.905121, 45.96361, 45.0131856, 44.0539114, 43.085851, 42.109068, 41.123626, 40.1295886, 39.1325906, 38.1381335, 37.1347901, 36.1222629, 35.1002544, 34.0684669, 33.0266028, 31.9743647, 30.9114549, 29.8375759, 28.75243, 27.6557197, 26.5471474, 25.4264156, 24.2932266, 23.1472829, 21.9882869, 20.8010043, 19.59292, 18.3716986, 17.1375272, 15.8905929, 14.6310829, 13.3591843, 12.0750844, 10.7789701, 9.47102874, 8.15144738, 6.8204132, 5.47811333, 4.12473493, 2.76046514, 1.38549112, 0, -1.26792717, -2.59457794, -3.97316776, -5.3969121, -6.85902641, -8.35272615, -9.87122679, -11.4077438, -12.9554926, -14.5076887, -16.0575475, -17.5982845, -19.1231152, -20.625255, -22.0979193, -23.5343237, -24.5507587, -25.0865933, -25.9876113, -27.1662584, -28.5349806, -30.0062235, -31.4924332, -32.9060554, -34.1595359, -35.1653207, -35.8358554, -36.0835861, -35.8209584, -34.4943361, -32.653144, -30.3496526, -27.6361323, -25.352902, -23.1582079, -20.5001502, -17.1945849, -13.057368, -7.90435556, -1.55140349, 6.46864776, 15.9707914, 27.0837989, 39.9364416, 54.6574908, 70.1810776, 88.1292738, 108.929511, 133.009221, 160.795834, 193.377718, 229.71356, 269.259859, 311.9445, 358.657778, 409.857887, 466.003019, 527.005789, 593.993962, 668.69048, 753.637466, 846.698096, 946.664337, 1052.93903, 1165.96008, 1285.69031, 1411.81755, 1544.23332, 1683.83729, 1830.95885, 1988.79264, 2154.36201, 2327.00455, 2506.05786, 2690.85952, 2898.16928, 3105.54704, 3307.97481, 3500.43466, 3677.90862, 3809.90151, 3933.85742, 4055.67722, 4194.83061, 4351.84745, 4521.6039, 4704.9125, 4902.58578, 5115.43627, 5344.27653, 5254.09329, 5471.0785, 5874.40395, 6343.24148, 6756.76289, 6994.13999, 6934.54461, 6539.99004, 5819.72244},
        {88.5777614, 87.5388987, 86.5459999, 85.5971176, 84.6903046, 83.8236135, 82.9950971, 82.2028082, 81.4447992, 80.7191231, 80.0238325, 79.35698, 78.7166184, 78.1008004, 77.5075788, 76.9350061, 76.3811351, 75.8440185, 75.3217091, 74.8122594, 74.3137223, 73.8241504, 73.3415964, 72.8641131, 72.389753, 71.916569, 71.4426138, 70.96594, 70.4846003, 69.9966474, 69.5001342, 68.9931131, 68.473637, 67.9397586, 67.3149239, 66.6360582, 65.9409835, 65.2301736, 64.5041023, 63.7632435, 63.008071, 62.2390587, 61.4566803, 60.6614098, 59.8537209, 59.0340874, 58.2029833, 57.3608823, 56.5082583, 55.6455851, 54.7733365, 53.9109962, 53.0396162, 52.159258, 51.2699836, 50.3718546, 49.4649327, 48.5492796, 47.6249573, 46.6920272, 45.7505513, 44.8005911, 43.8422086, 42.8754653, 41.9004231, 40.9171436, 39.9256886, 38.9320928, 37.9422474, 36.9438955, 35.9367115, 34.9203701, 33.8945456, 32.8589128, 31.8131461, 30.7569201, 29.6899093, 28.6117881, 27.5222313, 26.4209132, 25.3075085, 24.1816917, 23.0431373, 21.8915198, 20.707615, 19.5009443, 18.2812564, 17.048839, 15.80398, 14.5469672, 13.2780885, 11.9976317, 10.7058847, 9.40313526, 8.08967125, 6.76578049, 5.43175084, 4.08787013, 2.73442619, 1.37170687, 0, -1.0230849, -2.14951443, -3.36891314, -4.67090556, -6.04511626, -7.48116977, -8.96869066, -10.4973035, -12.0566327, -13.636303, -15.2259388, -16.8151648, -18.3936054, -19.9508852, -21.4766288, -22.9604607, -24.1208151, -24.908575, -25.9267718, -27.1069178, -28.3805254, -29.6791068, -30.9341744, -32.0772405, -33.0398175, -33.7534177, -34.1495533, -34.1597368, -33.7154805, -32.4442247, -30.6930467, -28.4846806, -25.8418602, -23.4179404, -20.9203448, -17.8883311, -14.155447, -9.55524011, -3.92125814, 2.9129513, 11.3096506, 21.1798878, 32.6518044, 45.8535423, 60.913243, 76.8022513, 95.094705, 116.207945, 140.559313, 168.566149, 201.273431, 237.703752, 277.16227, 319.570817, 366.110274, 417.396235, 474.044295, 536.808282, 606.134139, 682.317007, 766.343964, 857.56062, 955.724364, 1060.71314, 1172.60914, 1291.40083, 1416.92498, 1549.13075, 1688.52327, 1835.29319, 1992.41406, 2156.86544, 2327.87466, 2504.66902, 2686.47584, 2890.19592, 3092.96471, 3289.59115, 3474.88418, 3643.65272, 3762.85623, 3873.71947, 3982.98688, 4115.83524, 4271.71284, 4442.38787, 4627.69265, 4827.45949, 5041.52071, 5269.70863, 5104.39147, 5296.53656, 5706.03736, 6192.78732, 6616.67988, 6837.6085, 6715.46663, 6212.1507, 5452.0654},
        {89.4250614, 88.3467298, 87.3154617, 86.3292918, 85.3862547, 84.4843852, 83.6217178, 82.7962872, 82.0061281, 81.2492752, 80.523763, 79.8276264, 79.1588999, 78.5156182, 77.895816, 77.297528, 76.7187887, 76.1576329, 75.6120952, 75.0802103, 74.5600128, 74.0495375, 73.5468189, 73.0498918, 72.5567908, 72.0655506, 71.5742057, 71.080791, 70.5833411, 70.0798905, 69.5684741, 69.0471264, 68.5138821, 67.9667759, 67.3289734, 66.6375186, 65.9303553, 65.2079478, 64.4707603, 63.7192572, 62.9539028, 62.1751615, 61.3834974, 60.579375, 59.7632584, 58.9356122, 58.0969004, 57.2475875, 56.3881378, 55.5190155, 54.640685, 53.7716212, 52.8938874, 52.0075575, 51.1127052, 50.2094044, 49.2977291, 48.3777531, 47.4495502, 46.5131944, 45.5687594, 44.6163193, 43.6559477, 42.6877187, 41.711706, 40.7279836, 39.7366254, 38.7444174, 37.757985, 36.7635581, 35.7607755, 34.7492759, 33.728698, 32.6986806, 31.6588624, 30.6088822, 29.5483786, 28.4769905, 27.3943564, 26.3001153, 25.1939057, 24.0753665, 22.9441364, 21.799854, 20.6186871, 19.4125036, 18.1934536, 16.9619375, 15.7183556, 14.4631085, 13.1965965, 11.9192201, 10.6313796, 9.33347549, 8.02590817, 6.70907805, 5.38338554, 4.04923107, 2.70701505, 1.35713789, 0, -0.741085213, -1.63670266, -2.67243003, -3.83384499, -5.10652522, -6.47604842, -7.92799225, -9.4479344, -11.0214525, -12.6341244, -14.2715276, -15.9192398, -17.5628387, -19.1879021, -20.7800075, -22.3247327, -23.6537652, -24.7306984, -25.8891948, -27.0818556, -28.261282, -29.3800753, -30.3908367, -31.2461673, -31.8986686, -32.3009415, -32.4055875, -32.1652076, -31.5324032, -30.3360382, -28.6978214, -26.6074751, -24.0547218, -21.4784058, -18.6436886, -15.1959655, -10.9902224, -5.8814448, 0.275381393, 7.62527049, 16.3851018, 26.6061526, 38.4193765, 51.9557273, 67.3461585, 83.6899367, 102.407625, 123.888098, 148.520232, 176.692902, 209.310848, 245.616581, 284.860914, 327.000574, 373.433534, 424.910676, 482.182884, 546.824894, 618.576497, 696.278731, 779.282121, 868.513679, 964.776452, 1068.46151, 1179.26187, 1297.19094, 1422.2482, 1554.44347, 1693.83773, 1840.46303, 1996.89246, 2160.22609, 2329.61636, 2504.2157, 2683.17658, 2883.86788, 3082.6715, 3274.18575, 3453.00898, 3613.73953, 3720.69176, 3818.93728, 3916.05327, 4042.89737, 4197.83158, 4369.51769, 4556.80273, 4758.53368, 4973.55757, 5200.72139, 4971.29116, 5140.92414, 5553.02857, 6051.0127, 6478.28476, 6678.25301, 6494.32568, 5891.10442, 5108.64429},
        {90.3986141, 89.2763605, 88.2026638, 87.1755251, 86.1929453, 85.2529255, 84.3534668, 83.4925701, 82.6682364, 81.8784667, 81.1212622, 80.3946237, 79.6965524, 79.0250492, 78.3781151, 77.7537513, 77.1499585, 76.564738, 75.9960908, 75.4420177, 74.9005199, 74.3695984, 73.8472542, 73.3314883, 72.8203017, 72.3116954, 71.8036705, 71.294228, 70.7813689, 70.2630942, 69.7374049, 69.2023021, 68.6557868, 68.0958599, 67.4451858, 66.7410185, 66.0215174, 65.2871283, 64.5382972, 63.7754699, 62.9990922, 62.2096099, 61.4074688, 60.593115, 59.766994, 58.9295519, 58.0812344, 57.2224874, 56.3537567, 55.4754882, 54.5881277, 53.7072466, 52.8178373, 51.9200177, 51.0139058, 50.0996195, 49.1772768, 48.2469957, 47.308894, 46.3630897, 45.4097008, 44.4488451, 43.4806408, 42.5052056, 41.5226576, 40.5331147, 39.5366949, 38.541542, 37.5556075, 36.5624561, 35.5616855, 34.5528934, 33.5356774, 32.5096352, 31.4743644, 30.4294626, 29.3745276, 28.309157, 27.2329484, 26.1454994, 25.0464079, 23.9352712, 22.8116873, 21.6752536, 20.4975775, 19.2928406, 18.0755608, 16.8462442, 15.6053967, 14.3535243, 13.091133, 11.8187289, 10.5368178, 9.24590585, 7.94649897, 6.63910316, 5.32422442, 4.00236875, 2.67404212, 1.33975054, 0, -0.566931572, -1.31731584, -2.23434133, -3.30119658, -4.5010701, -5.81715043, -7.23262608, -8.7306856, -10.2945175, -11.9073103, -13.5522526, -15.2125328, -16.8713395, -18.5118612, -20.1172865, -21.6708038, -23.0558104, -24.2412389, -25.4359963, -26.6018873, -27.7007165, -28.6942887, -29.5444085, -30.2128805, -30.6615094, -30.8521, -30.7464567, -30.3063845, -29.4936878, -28.2304105, -26.5326972, -24.3742806, -21.7288937, -18.9149328, -15.7337991, -11.8949834, -7.26881965, -1.72564168, 4.86421666, 12.6304215, 21.6253504, 32.0794808, 44.1376004, 57.9444972, 73.6449588, 90.6555629, 110.03136, 132.099191, 157.185896, 185.618315, 218.008363, 254.049383, 293.199294, 335.51946, 382.303266, 434.290944, 492.222728, 557.914425, 630.786243, 709.085037, 791.359858, 879.358644, 974.503197, 1077.46075, 1187.61994, 1304.95629, 1429.46572, 1561.12901, 1699.85235, 1845.58411, 2000.48379, 2161.94851, 2329.16129, 2501.30515, 2677.56309, 2876.59184, 3073.23229, 3261.79902, 3436.60662, 3591.9697, 3689.95421, 3778.62245, 3866.10984, 3986.89342, 4138.68172, 4308.20762, 4493.72846, 4693.50155, 4905.78423, 5128.83382, 4896.6932, 5061.74537, 5464.44368, 5945.24143, 6344.59197, 6502.94861, 6260.76468, 5589.08981, 4800.96179},
        {91.488041, 90.3168534, 89.1961869, 88.12399, 87.0982112, 86.1167991, 85.177702, 84.2788687, 83.4182475, 82.593787, 81.8034357, 81.0451421, 80.3168548, 79.6165222, 78.9420929, 78.2915154, 77.6627381, 77.0537097, 76.4623786, 75.8866933, 75.3246024, 74.7740544, 74.2329977, 73.6993809, 73.1711526, 72.6462612, 72.1226552, 71.5982832, 71.0710937, 70.5390352, 70.0000562, 69.4521052, 68.8931307, 68.3210813, 67.657742, 66.9407451, 66.2086382, 65.4618415, 64.7007751, 63.9258593, 63.1375141, 62.3361597, 61.5222164, 60.6961042, 59.8582434, 59.009054, 58.1489563, 57.2783705, 56.3977166, 55.507415, 54.6078856, 53.709962, 52.8034254, 51.8884701, 50.9652906, 50.0340812, 49.0950365, 48.1483508, 47.1942186, 46.2328342, 45.2643922, 44.2890869, 43.3071127, 42.3186642, 41.3239356, 40.3231215, 39.3164162, 38.3139149, 37.325475, 36.3308715, 35.3296569, 34.3213839, 33.3056051, 32.281873, 31.2497404, 30.2087597, 29.1584836, 28.0984647, 27.0282556, 25.9474089, 24.8554771, 23.752013, 22.6365691, 21.508698, 20.3356474, 19.133802, 17.9199344, 16.6946456, 15.4585366, 14.2122084, 12.9562618, 11.6912979, 10.4179176, 9.13672189, 7.84831169, 6.55328796, 5.25225167, 3.94580377, 2.63454523, 1.31907698, 0, -0.53033209, -1.24486667, -2.12649171, -3.15809516, -4.32256498, -5.60278913, -6.98165556, -8.44205224, -9.96686712, -11.5389882, -13.1413033, -14.7567006, -16.3680678, -17.9582931, -19.5102643, -21.0068694, -22.3083738, -23.3801477, -24.4824401, -25.5718625, -26.6050267, -27.5385442, -28.3290268, -28.9330862, -29.307334, -29.4083819, -29.1928416, -28.6173247, -27.6384429, -26.1395514, -24.1773791, -21.7305146, -18.7775465, -15.6355923, -12.1139762, -7.93174792, -2.96593744, 2.90642531, 9.80831041, 17.8626879, 26.9544641, 37.517127, 49.7202243, 63.733304, 79.7259136, 97.6081032, 117.873793, 140.757405, 166.493361, 195.316084, 227.411378, 263.121705, 302.372716, 345.393336, 393.03167, 445.857824, 504.441904, 570.177544, 642.624157, 720.442754, 802.386665, 890.095558, 985.07001, 1087.96046, 1197.97685, 1314.99033, 1438.80836, 1569.28557, 1706.51014, 1850.43813, 2002.86202, 2161.61893, 2326.02926, 2495.41337, 2669.09166, 2867.75976, 3064.01898, 3251.8459, 3425.21708, 3578.10908, 3670.93675, 3753.61287, 3834.49761, 3949.23962, 4095.39738, 4259.15916, 4438.65197, 4632.00278, 4837.3386, 5052.78641, 4884.95385, 5065.11439, 5445.52758, 5878.45295, 6216.15003, 6310.87837, 6014.8975, 5310.10538, 4530.9619},
        {92.6936721, 91.4699967, 90.2990066, 89.17859, 88.1066354, 87.0810311, 86.0996654, 85.1604266, 84.2612032, 83.3998834, 82.5743557, 81.7825083, 81.0222296, 80.291408, 79.5879317, 78.9096892, 78.2545688, 77.6204588, 77.0052476, 76.4068235, 75.8230748, 75.25189, 74.6911573, 74.1387652, 73.5926019, 73.0505558, 72.5105152, 71.9703685, 71.428004, 70.8813102, 70.3281752, 69.7664875, 69.1941354, 68.6090073, 67.9317858, 67.2005861, 66.4544511, 65.6937747, 64.9189507, 64.130373, 63.3284354, 62.5135316, 61.6860555, 60.846401, 59.9949617, 59.1321317, 58.2583046, 57.3738743, 56.4792346, 55.5747793, 54.6609023, 53.743096, 52.816545, 51.8815327, 50.9383423, 49.9872571, 49.0285605, 48.0625357, 47.089466, 46.1096346, 45.123325, 44.1308203, 43.1324039, 42.128359, 41.118969, 40.1045171, 39.0852866, 38.0735924, 37.0814602, 36.0843595, 35.0817934, 34.0732653, 33.0582786, 32.0363367, 31.0069427, 29.9696002, 28.9238124, 27.8690826, 26.8049143, 25.7308106, 24.6462751, 23.550811, 22.4439216, 21.3251103, 20.1570191, 18.9583573, 17.7482774, 16.5274788, 15.2966611, 14.0565239, 12.8077667, 11.5510891, 10.2871907, 9.01677103, 7.7405296, 6.459166, 5.17337979, 3.88387052, 2.59133776, 1.29648107, 0, -0.526116988, -1.22993692, -2.09459803, -3.10323853, -4.23899666, -5.48501064, -6.8244187, -8.24035906, -9.71596995, -11.2343896, -12.7787562, -14.3322081, -15.8778833, -17.3989203, -18.8784571, -20.2996321, -21.4808762, -22.3817981, -23.358088, -24.3575897, -25.328147, -26.2176037, -26.9738038, -27.5445909, -27.877809, -27.921302, -27.6229135, -26.9304875, -25.7918678, -24.019022, -21.7454915, -18.959883, -15.6508032, -12.165475, -8.31319698, -3.8076369, 1.4701834, 7.63924205, 14.8185172, 23.1269869, 32.2891619, 42.9385431, 55.271287, 69.4835501, 85.7714889, 104.570642, 125.777939, 149.52969, 175.962206, 205.211798, 237.011967, 272.394158, 311.81819, 355.695024, 404.335855, 458.096671, 517.333459, 582.828033, 654.413776, 731.459538, 813.22672, 900.956883, 996.037275, 1099.00889, 1208.97876, 1325.71215, 1448.81341, 1578.00607, 1713.60361, 1855.5858, 2005.36005, 2161.25885, 2322.76779, 2489.37245, 2660.55843, 2859.08609, 3055.34754, 3243.00966, 3415.73935, 3567.20348, 3656.61758, 3735.06749, 3810.98497, 3920.28531, 4060.52138, 4217.79458, 4390.23035, 4575.95415, 4773.09142, 4979.76761, 4896.508, 5097.0269, 5450.26112, 5825.14748, 6090.62278, 6115.62384, 5769.08747, 5048.50992, 4289.40644},
        {94.0084985, 92.7288573, 91.504233, 90.3324476, 89.2113227, 88.1386802, 87.1123419, 86.1301295, 85.189865, 84.28937, 83.4264665, 82.5989761, 81.8047208, 81.0415222, 80.3072022, 79.5995827, 78.9164853, 78.255732, 77.6151445, 76.9925446, 76.3857541, 75.7925948, 75.2108885, 74.6384571, 74.0731223, 73.5127059, 72.9550298, 72.3979157, 71.8391854, 71.2766607, 70.7081635, 70.1315156, 69.5445386, 68.9450546, 68.2523615, 67.5051947, 66.7432332, 65.9668471, 65.1764064, 64.3722811, 63.5548414, 62.7244572, 61.8814987, 61.0263357, 60.1593385, 59.280877, 58.3913213, 57.4910414, 56.5804073, 55.6597892, 54.7295571, 53.7896882, 52.8409538, 51.8837325, 50.918403, 49.9453437, 48.9649332, 47.9775501, 46.9835729, 45.9833803, 44.9773507, 43.9658627, 42.9492949, 41.9280258, 40.9024341, 39.8728982, 38.8397968, 37.8178218, 36.8213216, 35.8211544, 34.8167708, 33.8076215, 32.7931571, 31.7728284, 30.746086, 29.7123806, 28.6711629, 27.6218835, 26.5639932, 25.4969426, 24.4201823, 23.3331631, 22.2353356, 21.1261506, 19.9633622, 18.7681464, 17.5621803, 16.3462677, 15.1212123, 13.887818, 12.6468886, 11.3992279, 10.1456396, 8.88692766, 7.62389582, 6.35734788, 5.08808765, 3.81691895, 2.54464557, 1.27207132, 0, -0.541554183, -1.2495995, -2.10788672, -3.10016661, -4.21018993, -5.42170743, -6.71846989, -8.08422808, -9.50273274, -10.9577347, -12.4329846, -13.9122333, -15.3792315, -16.81773, -18.2114796, -19.5442311, -20.5824298, -21.2787831, -22.1085654, -23.0089621, -23.9171585, -24.7703401, -25.5056921, -26.0603999, -26.3716489, -26.3766243, -26.0125117, -25.2164962, -23.9257633, -21.8632073, -19.2588779, -16.1142479, -12.4307898, -8.60680094, -4.44534165, 0.356347489, 5.91584116, 12.350714, 19.7785408, 28.3168961, 37.5511995, 48.293141, 60.7644128, 75.1867067, 91.7817149, 111.48806, 133.63127, 158.253805, 185.398125, 215.106689, 246.681424, 281.810416, 321.492421, 366.314988, 416.026115, 470.75074, 530.613802, 595.687495, 666.161491, 742.283, 824.021376, 912.015792, 1007.38003, 1110.47761, 1220.40082, 1336.82543, 1459.17882, 1587.07238, 1721.0284, 1861.05394, 2008.17385, 2161.22076, 2319.84938, 2483.71443, 2652.47061, 2850.6936, 3046.88691, 3234.47503, 3406.88242, 3557.53356, 3644.98817, 3720.77865, 3793.27152, 3897.96444, 4032.48451, 4183.15421, 4348.16588, 4525.71185, 4713.98448, 4911.1761, 4921.89178, 5144.68039, 5467.75865, 5779.3433, 5967.65106, 5920.89867, 5527.30286, 4805.2687, 4074.72084},
        {95.4255111, 94.0865019, 92.8049765, 91.5786851, 90.4053779, 89.282805, 88.2087167, 87.1808631, 86.1969944, 85.2548608, 84.3522125, 83.4867995, 82.6563722, 81.8586807, 81.0914751, 80.3525057, 79.6395226, 78.950276, 78.2825161, 77.633993, 77.002457, 76.3856581, 75.7813467, 75.1872728, 74.6011867, 74.0208385, 73.4439784, 72.8683566, 72.2917232, 71.7118285, 71.1264226, 70.5332557, 69.9300779, 69.3146395, 68.6045132, 67.8392243, 67.0592618, 66.2649778, 65.4567243, 64.6348535, 63.7997175, 62.9516684, 62.0910582, 61.218239, 60.333563, 59.4373821, 58.5300486, 57.6119145, 56.6833318, 55.7446528, 54.7962294, 53.8327784, 52.8604095, 51.8795968, 50.8908147, 49.8945373, 48.8912389, 47.8813938, 46.8654763, 45.8439605, 44.8173208, 43.7860313, 42.7505664, 41.7114003, 40.6690072, 39.6238615, 38.5764372, 37.5438505, 36.5428177, 35.5394907, 34.5332651, 33.5235363, 32.5096997, 31.4911507, 30.4672849, 29.4374975, 28.4011842, 27.3577402, 26.3065611, 25.2470423, 24.1785793, 23.1005674, 22.0124021, 20.9134789, 19.7563465, 18.5648091, 17.3632338, 16.1525364, 14.9336325, 13.7074379, 12.4748683, 11.2368393, 9.99426669, 8.74806618, 7.49915345, 6.24844421, 4.99685416, 3.74529901, 2.49469447, 1.24595623, 0, -0.563911588, -1.28092734, -2.13558425, -3.11241933, -4.19596958, -5.37077199, -6.62136358, -7.93228134, -9.28806227, -10.6732434, -12.0723617, -13.4699541, -14.8505578, -16.1987096, -17.4989466, -18.7358058, -19.6221468, -20.1036955, -20.7794973, -21.5758727, -22.4191421, -23.2356259, -23.9516445, -24.4935183, -24.7875678, -24.7601132, -24.3374752, -23.445974, -22.0119301, -19.6664922, -16.739382, -13.2454714, -9.19963251, -5.06179041, -0.624290417, 4.43920307, 10.2473341, 16.9187468, 24.5720851, 33.3259931, 42.6623321, 53.5303323, 66.173226, 80.8342451, 97.7566215, 118.305234, 141.321256, 166.767509, 194.606812, 224.801986, 256.291041, 291.314154, 331.352113, 377.143694, 427.912745, 483.563286, 543.999337, 608.577537, 677.873691, 753.060748, 834.911985, 923.34546, 1019.07331, 1122.23824, 1232.01817, 1348.03384, 1469.60255, 1596.26637, 1728.68017, 1866.86936, 2011.4994, 2161.85717, 2317.74652, 2478.97132, 2645.3354, 2842.70503, 3038.30605, 3225.42672, 3397.35528, 3547.37999, 3634.03998, 3708.5387, 3779.05688, 3880.21099, 4009.7176, 4154.27843, 4312.1608, 4481.63206, 4660.95954, 4848.41057, 4951.64131, 5195.27234, 5487.1345, 5735.05868, 5846.87573, 5730.41651, 5293.51191, 4581.34704, 3885.33056},
        {96.8658079, 95.4608311, 94.116282, 92.8298196, 91.599103, 90.4217912, 89.2955433, 88.2180184, 87.1868755, 86.1997737, 85.254372, 84.3483295, 83.4793053, 82.6449584, 81.8429478, 81.0709326, 80.326572, 79.6075249, 78.9114503, 78.2360075, 77.5788553, 76.9376529, 76.3100594, 75.6937337, 75.086335, 74.4855223, 73.8889547, 73.2942912, 72.6991908, 72.1013127, 71.4983159, 70.8878595, 70.2676025, 69.6352039, 68.9046727, 68.1172524, 67.3151922, 66.4988656, 65.6686463, 64.8249081, 63.9680245, 63.0983692, 62.216316, 61.3222384, 60.4165101, 59.4995049, 58.5715963, 57.6331581, 56.6845639, 55.7261873, 54.7584021, 53.7748333, 52.7827494, 51.7826705, 50.7751165, 49.7606074, 48.7396631, 47.7128036, 46.6805489, 45.6434189, 44.6019337, 43.556613, 42.5079771, 41.4565457, 40.4028389, 39.3473766, 38.2906787, 37.2510595, 36.2464058, 35.2405384, 34.2328238, 33.2226285, 32.2093189, 31.1922614, 30.1708224, 29.1443685, 28.1122661, 27.0738815, 26.0285813, 24.9757319, 23.9146998, 22.8448513, 21.765553, 20.6761712, 19.5239626, 18.3354447, 17.1377611, 15.9319693, 14.7191269, 13.5002914, 12.2765202, 11.0488709, 9.81840094, 8.5861679, 7.35322925, 6.1206425, 4.88946516, 3.66075473, 2.43556872, 1.21496464, 0, -0.434498795, -1.03558382, -1.78674177, -2.67145934, -3.67322323, -4.77552014, -5.96183677, -7.21565981, -8.52047596, -9.85977193, -11.2170344, -12.5757501, -13.9194057, -15.2314878, -16.4954833, -17.6948788, -18.5512379, -19.0106296, -19.6489859, -20.3936671, -21.1720333, -21.9114451, -22.5392626, -22.9828463, -23.1695564, -23.0267532, -22.4817971, -21.4620483, -19.8948673, -17.4104613, -14.3422998, -10.7068888, -6.52073475, -2.2821823, 2.2331816, 7.37826033, 13.2810993, 20.0697438, 27.8722392, 36.8166309, 46.4685439, 57.6896154, 70.7179295, 85.7915703, 103.148622, 124.030933, 147.421882, 173.308612, 201.678267, 232.517989, 264.899509, 300.844221, 341.78216, 388.403255, 439.886402, 496.088922, 556.868133, 621.320854, 690.237748, 765.239117, 847.725712, 936.854939, 1032.49899, 1134.52898, 1242.81417, 1357.22466, 1477.60797, 1603.82837, 1735.83291, 1873.52183, 2017.24608, 2166.38581, 2320.68524, 2479.88857, 2643.74001, 2836.14345, 3026.64349, 3208.94442, 3376.75051, 3523.76607, 3610.86161, 3686.46435, 3758.35017, 3860.64169, 3990.68589, 4134.85195, 4291.03574, 4457.13313, 4631.03998, 4810.65217, 4946.85894, 5188.71409, 5453.46549, 5658.36103, 5720.64858, 5557.57604, 5086.39129, 4389.26732, 3721.0376},
        {98.4092025, 96.9356873, 95.5255361, 94.1763177, 92.8856007, 91.6509537, 90.4699457, 89.3401451, 88.2591208, 87.2244415, 86.2336758, 85.2843924, 84.3741601, 83.5005476, 82.6611236, 81.8534567, 81.0751157, 80.3236693, 79.5966863, 78.8917352, 78.2063848, 77.5382038, 76.8847609, 76.2436249, 75.6123644, 74.9885481, 74.3697448, 73.7535231, 73.1374518, 72.5190995, 71.896035, 71.2658269, 70.626044, 69.974255, 69.220745, 68.4085605, 67.5817968, 66.7408551, 65.8861368, 65.018043, 64.136975, 63.2433339, 62.3375211, 61.4199377, 60.4909849, 59.5510641, 58.6005763, 57.6399229, 56.669505, 55.6897239, 54.7009808, 53.6961979, 52.6834177, 51.6632033, 50.6361183, 49.6027259, 48.5635895, 47.5192724, 46.4703381, 45.4173499, 44.360871, 43.301465, 42.239695, 41.1761246, 40.111317, 39.0458356, 37.9802437, 36.9341196, 35.9265851, 34.9189866, 33.9106549, 32.9009207, 31.8891146, 30.8745674, 29.8566097, 28.8345722, 27.8077857, 26.7755809, 25.7372883, 24.6922388, 23.639763, 22.5791916, 21.5098554, 20.4310849, 19.2844, 18.0992528, 16.9057903, 15.7052193, 14.4987465, 13.2875787, 12.0729227, 10.8559852, 9.63797307, 8.42009301, 7.20355182, 5.98955626, 4.77931311, 3.57402913, 2.37491111, 1.18316581, 0, -0.316852105, -0.811166982, -1.46555588, -2.26263006, -3.18500077, -4.21527926, -5.33607679, -6.5300046, -7.77967395, -9.0676961, -10.3766823, -11.6892438, -12.9879918, -14.2555376, -15.4744925, -16.6274677, -17.4487406, -17.885446, -18.481903, -19.1675084, -19.871659, -20.5237517, -21.0531832, -21.3893506, -21.4616505, -21.1994798, -20.5322355, -19.3893142, -17.7001129, -15.104958, -11.9283093, -8.18628459, -3.89500194, 0.414202638, 4.98282022, 10.1897334, 16.1733895, 23.0722362, 31.0247208, 40.1692908, 50.1907589, 61.81927, 75.2820264, 90.80623, 108.619083, 129.740542, 153.406867, 179.647069, 208.490163, 239.965161, 273.359471, 310.35012, 352.222154, 399.583614, 451.660331, 508.29872, 569.345195, 633.737734, 702.437654, 777.397294, 860.568019, 950.400286, 1045.93313, 1146.77869, 1253.5202, 1366.29514, 1485.51919, 1611.40193, 1743.13303, 1880.47907, 2023.54063, 2171.69896, 2324.60503, 2481.90984, 2643.26437, 2830.04435, 3014.74484, 3191.58567, 3354.78664, 3498.56755, 3586.61152, 3664.03285, 3738.13851, 3842.40427, 3973.77859, 4118.3063, 4273.48107, 4436.79656, 4605.74644, 4777.82438, 4933.62687, 5169.30883, 5409.04842, 5577.02381, 5597.41317, 5394.39466, 4892.14646, 4216.03777, 3578.605},
        {100.067663, 98.5239789, 97.046456, 95.6325792, 94.2798331, 92.9857024, 91.7476714, 90.5632248, 89.4298472, 88.3450232, 87.3062371, 86.3109738, 85.3567176, 84.4409531, 83.561165, 82.7148377, 81.8994559, 81.112504, 80.3514667, 79.6138285, 78.897074, 78.1986877, 77.5161541, 76.8469579, 76.1885837, 75.5385158, 74.894239, 74.2532378, 73.6129967, 72.9710003, 72.3247331, 71.6716798, 71.0093248, 70.3351528, 69.5558862, 68.7162997, 67.8622588, 66.9941935, 66.1125337, 65.2177094, 64.3101506, 63.3902872, 62.4585492, 61.5153666, 60.5611693, 59.5963872, 58.6214504, 57.6367888, 56.6428324, 55.6400111, 54.6287548, 53.6011811, 52.5662126, 51.5244595, 50.4765319, 49.42304, 48.364594, 47.301804, 46.2352802, 45.1656328, 44.0934718, 43.0194075, 41.9440501, 40.8680097, 39.7918964, 38.7163204, 37.6418919, 36.589715, 35.5804016, 34.5722949, 33.5646763, 32.5568276, 31.5480303, 30.5375662, 29.5247167, 28.5087636, 27.4889884, 26.4646729, 25.4350985, 24.399547, 23.3573, 22.307639, 21.2498458, 20.1832019, 19.0429392, 17.8617115, 16.6729422, 15.4779915, 14.2782196, 13.0749866, 11.8696527, 10.6635781, 9.45812286, 8.25464725, 7.05451138, 5.85907542, 4.66969954, 3.48774389, 2.31456864, 1.15153396, 0, -0.236768103, -0.654810001, -1.23632866, -1.96352704, -2.81860811, -3.78377484, -4.84123019, -5.97317713, -7.16181862, -8.38935763, -9.63799712, -10.8899401, -12.1273894, -13.3325482, -14.4876192, -15.5748056, -16.3390512, -16.7286648, -17.2625068, -17.871939, -18.488323, -19.0430207, -19.4673938, -19.6928039, -19.6506129, -19.2721824, -18.4888742, -17.23205, -15.4330716, -12.7477462, -9.48402659, -5.65594091, -1.27751719, 3.09659793, 7.72268569, 13.0011923, 19.0802752, 26.108092, 34.2328001, 43.6025572, 54.0347041, 66.1088982, 80.0380221, 96.0349582, 114.312589, 135.626048, 159.510404, 186.042976, 215.301085, 247.362051, 281.756726, 319.7768, 362.515768, 410.48769, 463.023612, 519.98573, 581.236239, 645.649493, 714.300039, 789.340062, 873.135254, 963.570057, 1058.97937, 1158.76217, 1264.12041, 1375.42844, 1493.60816, 1619.17587, 1750.64028, 1887.64568, 2030.12674, 2177.39276, 2328.98756, 2484.45493, 2643.33869, 2824.29281, 3002.97341, 3174.14674, 3332.57909, 3473.03671, 3562.15921, 3641.59065, 3718.15161, 3824.5723, 3957.4174, 4102.45584, 4256.79208, 4417.53053, 4581.77563, 4746.63183, 4910.25087, 5136.73619, 5354.62225, 5492.4435, 5478.73442, 5242.02945, 4710.86307, 4059.57294, 3456.09485},
        {101.85316, 100.238615, 98.6927587, 97.213004, 95.7967627, 94.4414468, 93.1444683, 91.9032393, 90.7151718, 89.5776779, 88.4881697, 87.4440592, 86.4427584, 85.4816795, 84.5582345, 83.6698355, 82.8138944, 81.9878235, 81.1890346, 80.41494, 79.6629516, 78.9304815, 78.2149417, 77.5137444, 76.8243016, 76.1440253, 75.4703276, 74.8006205, 74.1323162, 73.4628266, 72.7895639, 72.10994, 71.4213671, 70.7212572, 69.9132522, 69.0436215, 68.1597617, 67.2621278, 66.351175, 65.4273584, 64.4911329, 63.5429538, 62.583276, 61.6125547, 60.631245, 59.6398019, 58.6386805, 57.6283359, 56.6092232, 55.5817974, 54.5465137, 53.4940916, 52.4349328, 51.3697034, 50.2990695, 49.2236973, 48.1442528, 47.0614022, 45.9758116, 44.8881472, 43.7990751, 42.7092613, 41.6193721, 40.5300736, 39.4420319, 38.3559131, 37.2723833, 36.2145299, 35.2049015, 34.1979226, 33.1928059, 32.1887639, 31.1850095, 30.1807551, 29.1752136, 28.1675974, 27.1571194, 26.1429922, 25.1244284, 24.1006407, 23.0708418, 22.0342442, 20.9900608, 19.9375041, 18.8048609, 17.6282986, 16.4448372, 15.255991, 14.0632744, 12.8682016, 11.6722871, 10.4770451, 9.28399009, 8.09463632, 6.91049814, 5.73308992, 4.56392598, 3.40452068, 2.25638835, 1.12104335, 0, -0.22004338, -0.613646048, -1.16336214, -1.8517458, -2.66135115, -3.57473235, -4.57444352, -5.6430388, -6.76307234, -7.91709826, -9.08767071, -10.2573438, -11.4086717, -12.5242086, -13.5865085, -14.5781257, -15.2465663, -15.5408062, -15.9750556, -16.4815011, -16.9923295, -17.4397274, -17.7558817, -17.8729791, -17.7232063, -17.23875, -16.351797, -14.994534, -13.0991478, -10.3365895, -6.99606779, -3.08814003, 1.37663642, 5.83423713, 10.5508383, 15.9402071, 22.1578264, 29.359179, 37.6997474, 47.3350145, 58.2061066, 70.7481018, 85.1584219, 101.634489, 120.373725, 141.879434, 165.966685, 192.756431, 222.369621, 254.927208, 290.17707, 329.06921, 372.506677, 420.918402, 473.765328, 530.943005, 592.346981, 656.877449, 725.651534, 800.872206, 885.123766, 975.952811, 1071.2414, 1170.25425, 1274.59896, 1384.80769, 1502.14686, 1627.339, 1758.41441, 1894.92626, 2036.74808, 2183.06339, 2333.31449, 2486.94365, 2643.39316, 2818.77395, 2991.69249, 3157.4239, 3311.24331, 3448.42584, 3538.37419, 3619.48418, 3698.11912, 3806.21939, 3940.02403, 4085.11495, 4238.26404, 4396.24318, 4555.82427, 4713.77918, 4875.03671, 5090.67579, 5290.92591, 5406.01656, 5366.17727, 5101.63751, 4542.6268, 3917.78741, 3351.56923},
        {103.777659, 102.092503, 100.478161, 98.931992, 97.4513517, 96.0335968, 94.676084, 93.37617, 92.1312114, 90.9385648, 89.7955868, 88.6996342, 87.6480636, 86.6382315, 85.6674947, 84.7332098, 83.8327335, 82.9634223, 82.1226329, 81.307722, 80.5160462, 79.7449622, 78.9918265, 78.2539959, 77.528827, 76.8136764, 76.1059007, 75.4028567, 74.7019009, 74.00039, 73.2956806, 72.5851295, 71.8660931, 71.1359282, 70.2959989, 69.3936771, 68.4774886, 67.5479052, 66.6053987, 65.6504408, 64.6835034, 63.7050582, 62.7155771, 61.7155317, 60.7053939, 59.6856355, 58.6567283, 57.6191441, 56.5733545, 55.5198315, 54.4590468, 53.3792383, 52.2933768, 51.2021994, 50.1064431, 49.0068449, 47.9041418, 46.7990707, 45.6923686, 44.5847727, 43.4770197, 42.3698468, 41.263991, 40.1601892, 39.0591784, 37.9616956, 36.8684778, 35.8052484, 34.7971307, 33.7933293, 32.7929615, 31.7951444, 30.7989954, 29.8036316, 28.8081703, 27.8117287, 26.8134241, 25.8123736, 24.8076945, 23.7985041, 22.7839195, 21.763058, 20.7350369, 19.6989734, 18.5754456, 17.4044923, 16.2270961, 15.0449231, 13.8596392, 12.6729105, 11.4864028, 10.3017821, 9.12071454, 7.94486595, 6.77590234, 5.61548969, 4.46529397, 3.32698117, 2.20221726, 1.09266821, 0, -0.292474522, -0.734808294, -1.31095838, -2.00488186, -2.8005358, -3.68187726, -4.63286331, -5.63745103, -6.67959748, -7.74325973, -8.81239485, -9.8709599, -10.902912, -11.8922081, -12.8228054, -13.6786609, -14.1956821, -14.3223906, -14.6038077, -14.970737, -15.3539822, -15.6843468, -15.8926345, -15.9096488, -15.6661934, -15.0930718, -14.1210876, -12.6810445, -10.7037461, -7.86925185, -4.45104894, -0.455164135, 4.11237577, 8.69635378, 13.5653382, 19.134348, 25.5621136, 33.0073651, 41.6288328, 51.5852471, 62.9106937, 75.9264826, 90.815731, 107.761556, 126.947075, 148.692686, 173.009905, 200.047529, 229.954356, 262.879183, 298.7063, 338.172298, 382.038552, 430.678668, 483.67456, 540.963594, 602.483139, 667.242918, 736.318769, 811.79851, 896.229903, 987.137107, 1082.32287, 1181.02976, 1284.94002, 1394.61603, 1511.40725, 1636.08013, 1766.51519, 1902.22542, 2043.14832, 2188.307, 2337.0675, 2488.79583, 2642.858, 2813.37285, 2981.26539, 3142.21341, 3291.89474, 3425.98719, 3516.12596, 3598.05987, 3677.77073, 3786.41911, 3920.0202, 4064.098, 4215.19226, 4369.84269, 4524.58905, 4675.97106, 4826.29017, 5030.80726, 5218.69835, 5319.13947, 5261.30665, 4974.37592, 4387.52331, 3788.59573, 3263.09024},
        {105.942093, 104.169684, 102.471295, 100.844191, 99.2856361, 97.7928966, 96.3632376, 94.9939242, 93.6822219, 92.4253958, 91.2207113, 90.0654337, 88.9568282, 87.8921601, 86.8686948, 85.8836975, 84.9344335, 84.018168, 83.1321664, 82.273694, 81.440016, 80.6283978, 79.8361045, 79.0604016, 78.2985543, 77.5478278, 76.8054875, 76.0687987, 75.3350266, 74.6014365, 73.8652937, 73.1238636, 72.3744113, 71.6142022, 70.7412708, 69.8045851, 68.8543054, 67.8909174, 66.9149064, 65.9267581, 64.9269578, 63.9159913, 62.8943439, 61.8625011, 60.8209486, 59.7701717, 58.710656, 57.6428871, 56.5673504, 55.4845315, 54.3949158, 53.2818354, 52.1633011, 51.0401703, 49.9133004, 48.7835488, 47.6517727, 46.5188297, 45.385577, 44.252872, 43.1215721, 41.9925347, 40.8666171, 39.7446767, 38.6275708, 37.5161569, 36.4112923, 35.3433703, 34.3420751, 33.3473483, 32.3581327, 31.373371, 30.3920057, 29.4129796, 28.4352353, 27.4577155, 26.4793629, 25.4991201, 24.5159299, 23.5287348, 22.5364777, 21.538101, 20.5325476, 19.51876, 18.4072331, 17.2429938, 16.0721322, 14.8964613, 13.7177942, 12.5379437, 11.3587229, 10.1819446, 9.00942203, 7.84296799, 6.68439551, 5.53551756, 4.39814712, 3.27409716, 2.16518068, 1.07321063, 0, -0.704518665, -1.47344615, -2.29630628, -3.16262288, -4.06191978, -4.98372082, -5.91754982, -6.8529306, -7.779387, -8.68644284, -9.56362195, -10.4004482, -11.1864453, -11.9111372, -12.5640477, -13.1347006, -13.2853847, -12.9588761, -12.884873, -12.9827772, -13.171991, -13.3719164, -13.5019554, -13.48151, -13.2299824, -12.6667745, -11.7112884, -10.2829261, -8.30108972, -5.40999986, -1.90513983, 2.2115339, 6.93806485, 11.7643119, 16.942411, 22.8600138, 29.6676192, 37.5157261, 46.5548335, 56.9354402, 68.6310121, 82.0229607, 97.2964221, 114.636532, 134.228426, 156.667295, 181.625707, 209.186283, 239.431647, 272.444419, 307.811998, 346.717505, 390.071511, 438.3421, 491.093943, 548.297325, 609.922534, 675.267711, 745.128044, 821.031967, 904.714246, 994.668624, 1089.83099, 1189.77986, 1295.18271, 1406.20713, 1523.43201, 1647.13158, 1776.07216, 1909.87301, 2048.69747, 2191.53708, 2337.82277, 2486.9855, 2638.45624, 2806.59577, 2972.17274, 3130.88564, 3278.43297, 3410.51321, 3499.23951, 3579.61927, 3657.459, 3763.01987, 3892.42842, 4031.6217, 4177.23571, 4325.90643, 4474.26988, 4618.96202, 4729.03731, 4921.53579, 5114.06172, 5224.21939, 5169.61306, 4867.84701, 4236.52549, 3661.52883, 3189.38703},
        {108.201916, 106.357922, 104.589265, 102.893191, 101.266947, 99.7077777, 98.2129307, 96.779652, 95.4051879, 94.0867846, 92.8216884, 91.6071457, 90.4404027, 89.3187056, 88.2393009, 87.1994347, 86.1963534, 85.2273033, 84.2895306, 83.3802816, 82.4968027, 81.63634, 80.79614, 79.9734488, 79.1655129, 78.3695784, 77.5828916, 76.8026989, 76.0262465, 75.2507808, 74.4735479, 73.6917943, 72.9027661, 72.1037097, 71.192427, 70.217978, 69.2306183, 68.2308213, 67.2190605, 66.1958092, 65.1615411, 64.1167294, 63.0618477, 61.9973694, 60.923768, 59.8415169, 58.7510896, 57.6529595, 56.5476001, 55.4354848, 54.3170871, 53.1694119, 52.01691, 50.8605635, 49.7013547, 48.5402657, 47.3782788, 46.2163761, 45.05554, 43.8967524, 42.7409958, 41.5892522, 40.4425039, 39.3017331, 38.167922, 37.0420528, 35.9251076, 34.8522943, 33.8579986, 32.8726154, 31.8949081, 30.9236403, 29.9575754, 28.995477, 28.0361086, 27.0782337, 26.1206158, 25.1620184, 24.2012049, 23.236939, 22.2679841, 21.2931038, 20.3110614, 19.3206206, 18.2223596, 17.0661997, 15.9032423, 14.7354373, 13.5647345, 12.3930837, 11.2224346, 10.0547372, 8.8919411, 7.73599625, 6.58885242, 5.45245941, 4.32876704, 3.2197251, 2.12728341, 1.05339178, 0, -1.08672092, -2.1588586, -3.21099887, -4.23772753, -5.23363041, -6.19329334, -7.11130212, -7.98224258, -8.80070053, -9.5612618, -10.2585122, -10.8870376, -11.4414237, -11.9162564, -12.3061215, -12.6056049, -12.408843, -11.6532147, -11.2390839, -11.0752258, -11.0704158, -11.133429, -11.1730408, -11.0980264, -10.817161, -10.2392199, -9.27297846, -7.8272118, -5.81069523, -2.8395971, 0.777411221, 5.04377241, 9.96292914, 15.0860036, 20.6425986, 26.9829361, 34.2461546, 42.5713932, 52.0977905, 62.9644855, 75.0857723, 88.8940653, 104.572495, 122.304191, 142.272283, 165.274791, 190.726234, 218.656019, 249.093556, 282.068251, 316.908276, 355.201344, 398.003692, 445.874369, 498.35733, 555.452288, 617.158953, 683.059401, 753.665898, 829.946312, 912.79913, 1001.73001, 1096.88186, 1198.18587, 1305.21448, 1417.70481, 1535.37787, 1657.9666, 1785.26295, 1917.02549, 2053.79663, 2194.43051, 2338.41435, 2485.23539, 2634.38086, 2800.75365, 2964.57143, 3121.46752, 3267.07523, 3397.02788, 3483.38468, 3561.06913, 3635.81775, 3737.25076, 3861.65069, 3995.34854, 4135.12444, 4277.75848, 4420.03081, 4558.72152, 4640.13381, 4824.60028, 5021.06397, 5138.4679, 5085.75509, 4771.86857, 4105.75136, 3549.25084, 3127.34808},
        {110.497277, 108.613293, 106.802582, 105.0625, 103.3904, 101.783638, 100.239568, 98.7555452, 97.3289243, 95.95706, 94.6373069, 93.3670199, 92.1435537, 90.964263, 89.8265025, 88.727627, 87.6649911, 86.6359498, 85.6378576, 84.6680693, 83.7239396, 82.8028234, 81.9020752, 81.0190499, 80.1511022, 79.2955869, 78.4498586, 77.611272, 76.777182, 75.9449433, 75.1119106, 74.2754386, 73.4328821, 72.5815958, 71.6244244, 70.6079656, 69.5799529, 68.5408081, 67.490953, 66.4308093, 65.3607988, 64.2813433, 63.1928645, 62.0957842, 60.9905242, 59.8775062, 58.757152, 57.6298834, 56.4961221, 55.3562899, 54.2108086, 53.0291125, 51.8432826, 50.6544123, 49.4635952, 48.2719248, 47.0804946, 45.8903981, 44.7027288, 43.5185803, 42.339046, 41.1652195, 39.9981943, 38.8390638, 37.6889217, 36.5488614, 35.4199764, 34.3416612, 33.3523908, 32.3742658, 31.4058967, 30.4458942, 29.4928688, 28.5454313, 27.6021922, 26.6617621, 25.7227517, 24.7837716, 23.8434323, 22.9003446, 21.953119, 21.0003661, 20.0406965, 19.0727209, 17.9883835, 16.8418716, 15.6886548, 14.5308051, 13.3703947, 12.2094959, 11.0501806, 9.89452117, 8.74458962, 7.60245814, 6.47019887, 5.34988396, 4.24358556, 3.15337582, 2.08132688, 1.02951089, 0, -1.27818439, -2.49916843, -3.66004, -4.75788696, -5.78979717, -6.75285852, -7.64415885, -8.46078605, -9.19982798, -9.85837251, -10.4335075, -10.9223208, -11.3219004, -11.629334, -11.8417095, -11.9561149, -11.5323322, -10.5049771, -9.86028134, -9.50162441, -9.3323858, -9.25594503, -9.17568159, -8.99497498, -8.61720471, -7.94575028, -6.88399118, -5.33530693, -3.20307701, -0.104863152, 3.66533751, 8.11840013, 13.2651999, 18.7280781, 24.7121768, 31.5262472, 39.2977245, 48.1540441, 58.2226414, 69.6309516, 82.2963559, 96.6202733, 112.771237, 130.917779, 151.228433, 174.408212, 199.955049, 227.903356, 258.287546, 291.142031, 325.768572, 363.829699, 406.411044, 454.097986, 506.454557, 563.503353, 625.266971, 691.515126, 762.580769, 839.072717, 921.354374, 1009.58505, 1104.72128, 1206.92074, 1314.9875, 1428.34695, 1546.07578, 1667.50895, 1793.26015, 1923.21975, 2058.47858, 2197.54187, 2339.88448, 2484.98124, 2632.307, 2796.91958, 2958.8151, 3113.57263, 3256.77128, 3383.99013, 3467.06303, 3541.14449, 3611.99012, 3709.06325, 3828.73521, 3957.49911, 4092.22018, 4229.76366, 4366.99479, 4500.77881, 4586.29214, 4770.09607, 4961.89299, 5071.38529, 5008.27537, 4682.26561, 4003.05842, 3456.39364, 3074.42823},
        {112.72494, 110.874709, 109.088152, 107.363018, 105.697051, 104.087998, 102.533605, 101.031618, 99.5797836, 98.1758471, 96.817555, 95.5026534, 94.2288883, 92.9940058, 91.7957521, 90.6318733, 89.5001154, 88.3982246, 87.323947, 86.2750286, 85.2492155, 84.2442539, 83.2578899, 82.2878695, 81.3319389, 80.3878441, 79.4533313, 78.5261466, 77.604036, 76.6847456, 75.7660216, 74.8456101, 73.9212571, 72.9907087, 71.9733552, 70.9077686, 69.833463, 68.7507272, 67.6598502, 66.5611209, 65.4548281, 64.3412608, 63.2207078, 62.0934581, 60.9598007, 59.8200242, 58.6744178, 57.5232702, 56.3668704, 55.2055073, 54.0394698, 52.8294745, 51.6162405, 50.4009145, 49.1846432, 47.9685734, 46.7538518, 45.5416252, 44.3330404, 43.1292441, 41.931383, 40.7406038, 39.5580535, 38.3848786, 37.222226, 36.0712423, 34.9330745, 33.8490148, 32.8593919, 31.882555, 30.9170484, 29.9614164, 29.0142032, 28.073953, 27.1392103, 26.2085191, 25.2804239, 24.3534688, 23.4261981, 22.4971561, 21.5648871, 20.6279353, 19.684845, 18.7341604, 17.6625177, 16.5267689, 15.3851114, 14.2397211, 13.0927739, 11.9464457, 10.8029124, 9.66435001, 8.53293432, 7.41084127, 6.30024677, 5.20332672, 4.12225703, 3.05921361, 2.01637236, 0.995909186, 0, -0.85224824, -1.72136463, -2.59886374, -3.47626012, -4.34506836, -5.19680301, -6.02297864, -6.81510981, -7.5647111, -8.26329707, -8.90238227, -9.47348129, -9.96810868, -10.377779, -10.6940069, -10.9083068, -10.6474271, -9.85077285, -9.33492521, -9.01323452, -8.79905114, -8.60572544, -8.3466078, -7.93504857, -7.28439814, -6.30800686, -4.91922512, -3.03140327, -0.55789169, 2.79908936, 6.83764504, 11.5810859, 17.0527226, 22.9631615, 29.4920652, 36.9024844, 45.3115408, 54.8363564, 65.5940531, 77.7017526, 91.2996255, 106.47473, 123.339679, 142.007085, 162.58956, 184.691689, 209.061119, 235.937469, 265.560357, 298.169403, 333.858844, 373.191369, 416.857114, 465.231906, 518.049868, 575.333236, 637.104244, 702.949632, 773.426109, 849.565469, 932.370611, 1021.25548, 1116.45813, 1217.86913, 1324.79024, 1436.79348, 1553.15654, 1673.37511, 1798.02424, 1927.06845, 2062.58851, 2202.1071, 2344.85616, 2490.06766, 2636.97354, 2799.03617, 2957.6995, 3108.6379, 3247.52574, 3370.03737, 3447.86915, 3516.65878, 3582.3569, 3675.73717, 3792.92444, 3919.70068, 4052.70789, 4188.58809, 4323.98329, 4455.53551, 4655.67133, 4862.252, 5017.02344, 5061.73158, 4938.12235, 4587.94168, 3952.93549, 3390.1729, 3025.73814},
        {114.960181, 113.172483, 111.435894, 109.748662, 108.109034, 106.51526, 104.965587, 103.458262, 101.991535, 100.563652, 99.1728619, 97.8174127, 96.4955524, 95.2055289, 93.9455901, 92.7139841, 91.5089588, 90.3287623, 89.1716424, 88.0358472, 86.9196247, 85.8212229, 84.7388897, 83.670873, 82.615421, 81.5707816, 80.5352027, 79.5069324, 78.4842186, 77.4653094, 76.4484526, 75.4318963, 74.4138885, 73.3926771, 72.3088114, 71.1906025, 70.0670546, 68.9382883, 67.804424, 66.665582, 65.5218827, 64.3734467, 63.2203943, 62.0628459, 60.900922, 59.7347429, 58.5644291, 57.390101, 56.211879, 55.0298835, 53.8442349, 52.6056288, 51.3646818, 50.122586, 48.880533, 47.6397148, 46.4013232, 45.1665501, 43.9365873, 42.7126267, 41.4958601, 40.2874794, 39.0886765, 37.9006431, 36.7245711, 35.5616525, 34.4130789, 33.3213817, 32.3280995, 31.3491648, 30.3830897, 29.4283861, 28.483566, 27.5471415, 26.6176244, 25.6935269, 24.7733608, 23.8556382, 22.9388711, 22.0215715, 21.1022513, 20.1794226, 19.2515973, 18.3172874, 17.2596016, 16.137181, 15.0100536, 13.8804764, 12.7507064, 11.6230008, 10.4996165, 9.38281062, 8.27484023, 7.17796236, 6.09443406, 5.02651238, 3.97645437, 2.94651707, 1.93895755, 0.956032839, 0, -0.290658405, -0.697746645, -1.20538177, -1.79768082, -2.45876084, -3.17273889, -3.92373201, -4.69585725, -5.47323165, -6.23997227, -6.98019615, -7.67802033, -8.31756187, -8.88293781, -9.35826521, -9.7276611, -9.71471075, -9.2663124, -8.95588115, -8.71170581, -8.4620752, -8.13527812, -7.65960339, -6.9633398, -5.97477618, -4.62220133, -2.83390405, -0.538173163, 2.33670253, 5.9801715, 10.3030373, 15.34169, 21.1325195, 27.467505, 34.505242, 42.469502, 51.4699984, 61.6164444, 73.0185533, 85.7860384, 100.317727, 116.346227, 133.924686, 153.106251, 173.94407, 194.844439, 217.91907, 243.632822, 272.450556, 304.837132, 341.766494, 382.572205, 427.501591, 476.693665, 530.066286, 587.636608, 649.421782, 714.773671, 784.525921, 860.237949, 943.725142, 1033.43229, 1128.58933, 1228.73964, 1333.95767, 1444.07411, 1558.79032, 1677.90347, 1801.68485, 1930.13756, 2066.30509, 2206.68238, 2350.2201, 2495.86889, 2642.57942, 2801.84287, 2956.93422, 3103.66899, 3237.86269, 3355.33084, 3427.76319, 3491.18487, 3551.75101, 3641.67047, 3756.78545, 3882.08242, 4013.94636, 4148.76224, 4282.91506, 4412.78978, 4737.1897, 4967.73783, 5081.95365, 5057.35663, 4871.46627, 4501.80204, 3925.88342, 3345.38867, 2983.64112},
        {117.215076, 115.504571, 113.831118, 112.193522, 110.590586, 109.021116, 107.483913, 105.977784, 104.501532, 103.05396, 101.633873, 100.240076, 98.8713709, 97.5265635, 96.2044573, 94.9038564, 93.6235648, 92.3623866, 91.119126, 89.8925868, 88.6815733, 87.4848894, 86.3013393, 85.1297269, 83.9688565, 82.8175319, 81.6745574, 80.5387369, 79.4088745, 78.2837743, 77.1622404, 76.0430767, 74.9250875, 73.8070767, 72.6531213, 71.4801193, 70.3053863, 69.1288531, 67.950451, 66.7701108, 65.5877636, 64.4033405, 63.2167724, 62.0279904, 60.8369255, 59.6435088, 58.4476712, 57.2493439, 56.0484577, 54.8449438, 53.6387331, 52.3691295, 51.0980056, 49.8266065, 48.5561777, 47.2879643, 46.0232118, 44.7631654, 43.5090704, 42.2621722, 41.023716, 39.7949472, 38.577111, 37.3714527, 36.1792178, 35.0016513, 33.8399988, 32.7378353, 31.7377197, 30.7535951, 29.7839628, 28.827324, 27.8821798, 26.9470317, 26.0203807, 25.1007281, 24.1865751, 23.2764231, 22.3687731, 21.4621265, 20.5549844, 19.6458482, 18.7332189, 17.815598, 16.7740731, 15.6683218, 14.559433, 13.4497182, 12.3414888, 11.2370565, 10.1387327, 9.04882897, 7.96965684, 6.90352783, 5.85275345, 4.81964525, 3.80651473, 2.81567342, 1.84943284, 0.91010453, 0, 0.246448932, 0.281887114, 0.129333893, -0.188191386, -0.647669376, -1.22608073, -1.9004061, -2.64762615, -3.44472152, -4.26867286, -5.09646085, -5.90506611, -6.67146932, -7.37265112, -7.98559217, -8.48727312, -8.70573862, -8.59669276, -8.47399524, -8.28271186, -7.9679084, -7.47465067, -6.74800447, -5.73303558, -4.37480981, -2.61839296, -0.408850814, 2.30875082, 5.58934616, 9.50982008, 14.0951075, 19.3935575, 25.4535189, 32.1284986, 39.5642667, 47.9585512, 57.4181538, 68.0498763, 79.9605202, 93.2568873, 108.56932, 125.321741, 143.51852, 163.164026, 184.262629, 204.260089, 226.359038, 251.203498, 279.437491, 311.70504, 349.681895, 391.719349, 437.70926, 487.60068, 541.45944, 599.298938, 661.132575, 726.134192, 795.347555, 870.732306, 954.751542, 1045.04779, 1139.95267, 1238.71932, 1342.16232, 1450.37953, 1563.5402, 1681.7607, 1804.89558, 1932.94748, 2069.68888, 2210.77264, 2354.89645, 2500.75803, 2647.05509, 2803.33741, 2954.73761, 3097.24037, 3226.83037, 3339.49229, 3407.15973, 3465.90253, 3522.013, 3608.74106, 3721.70827, 3845.30388, 3975.71108, 4109.11306, 4241.69301, 4369.63413, 4790.30418, 5037.32807, 5117.67044, 5038.29588, 4806.16902, 4428.25447, 3911.51685, 3321.23659, 2947.93275},
        {119.53678, 117.873043, 116.235459, 114.623267, 113.035708, 111.47202, 109.931443, 108.413216, 106.916577, 105.440768, 103.985025, 102.54859, 101.130701, 99.7305979, 98.3475193, 96.9807048, 95.6293936, 94.292825, 92.9702384, 91.6608729, 90.3639679, 89.0787626, 87.8044963, 86.5404084, 85.2857379, 84.0397244, 82.8016069, 81.5706248, 80.3460174, 79.127024, 77.9128837, 76.702836, 75.49612, 74.2919751, 73.073994, 71.8495684, 70.6262779, 69.4038696, 68.1820906, 66.9606877, 65.7394082, 64.5179989, 63.2962071, 62.0737796, 60.8504635, 59.6260059, 58.4001538, 57.1726542, 55.9432542, 54.7117008, 53.477741, 52.1656237, 50.8522307, 49.5389457, 48.2271525, 46.9182347, 45.6135761, 44.3145602, 43.0225708, 41.7389916, 40.4652063, 39.2025986, 37.9525521, 36.7164506, 35.4956776, 34.291617, 33.1056524, 31.9839147, 30.9714603, 29.9773854, 29.0001727, 28.0383051, 27.0902658, 26.1545374, 25.2296029, 24.3139454, 23.4060475, 22.5043924, 21.6074628, 20.7137417, 19.8217121, 18.9298567, 18.0366586, 17.1406007, 16.1227942, 15.0428766, 13.9618962, 12.8821447, 11.8059139, 10.7354959, 9.67318233, 8.6212651, 7.58203607, 6.55778706, 5.55080993, 4.56339653, 3.59783869, 2.65642828, 1.74145713, 0.855217085, 0, 0.232047457, 0.262586582, 0.114889084, -0.187773333, -0.622128961, -1.16490609, -1.79283303, -2.48263805, -3.21104946, -3.95479555, -4.69060461, -5.39520494, -6.04532483, -6.61769257, -7.08903646, -7.43608479, -7.54615044, -7.38318807, -7.12617513, -6.73267947, -6.16026893, -5.36651136, -4.30897462, -2.94522654, -1.23283497, 0.870632237, 3.40760723, 6.42052218, 9.95180922, 13.9998519, 18.6672821, 24.0097466, 30.0828922, 36.7362445, 44.1285109, 52.4733641, 61.8882874, 72.4907637, 84.3982762, 97.728308, 113.119067, 130.008829, 148.413198, 168.347775, 189.828163, 210.802727, 233.871115, 259.565738, 288.419006, 320.96333, 358.408469, 399.781617, 445.213913, 494.791791, 548.510418, 606.405939, 668.514498, 734.324284, 804.543934, 880.479852, 963.770665, 1052.89733, 1146.83808, 1245.12471, 1348.22705, 1456.18434, 1569.10751, 1687.05438, 1809.82006, 1937.34826, 2073.20816, 2213.16025, 2355.89326, 2500.09595, 2644.45705, 2797.89424, 2946.31011, 3085.83614, 3212.60386, 3322.74476, 3389.20229, 3446.75474, 3501.47136, 3585.27047, 3693.98941, 3812.69506, 3937.76328, 4065.56991, 4192.4908, 4314.90182, 4663.88881, 4886.50125, 4976.37636, 4927.15138, 4732.46352, 4385.95002, 3881.24811, 3326.1873, 2912.22593},
        {121.8803, 120.269156, 118.672748, 117.090767, 115.522902, 113.968844, 112.428284, 110.900912, 109.386419, 107.884495, 106.394831, 104.917116, 103.451043, 101.9963, 100.552579, 99.1195699, 97.6969635, 96.28445, 94.8817201, 93.4884643, 92.1043729, 90.7291365, 89.3624456, 88.0039906, 86.6534621, 85.3105506, 83.9749465, 82.6463403, 81.3244225, 80.0088836, 78.699414, 77.3957044, 76.0974451, 74.8043267, 73.5199277, 72.2419128, 70.9678999, 69.6974533, 68.4301374, 67.1655164, 65.9031547, 64.6426166, 63.3834663, 62.1252682, 60.8675867, 59.6099859, 58.3520303, 57.0932841, 55.8333116, 54.5716772, 53.3079452, 51.9512006, 50.5934483, 49.2362142, 47.881024, 46.5294033, 45.1828779, 43.8429736, 42.511216, 41.1891308, 39.8782439, 38.5800809, 37.2961676, 36.0280296, 34.7771928, 33.5451827, 32.3335252, 31.1908625, 30.1647056, 29.159406, 28.1734349, 27.2052637, 26.2533638, 25.3162066, 24.3922633, 23.4800054, 22.5779042, 21.6844311, 20.7980573, 19.9172543, 19.0404934, 18.1662459, 17.2929833, 16.4191768, 15.4264092, 14.3740546, 13.3228924, 12.2751872, 11.2332035, 10.199206, 9.17545921, 8.16422786, 7.16777651, 6.1883698, 5.22827233, 4.28974871, 3.37506357, 2.48648151, 1.62626715, 0.796685111, 0, 0.102890894, 0.0366819614, -0.176689982, -0.515288121, -0.95717564, -1.48041572, -2.06307155, -2.68320631, -3.31888319, -3.94816537, -4.54911603, -5.09979836, -5.57827554, -5.96261076, -6.2308672, -6.36110805, -6.28220114, -5.96518174, -5.49938824, -4.85234196, -3.99156422, -2.88457635, -1.49889968, 0.197944487, 2.23843481, 4.65504998, 7.48026866, 10.7465695, 14.4864313, 18.642565, 23.3701317, 28.7285404, 34.7771999, 41.3364095, 48.5851329, 56.7660966, 66.0104425, 76.4493125, 88.2138484, 101.435192, 116.689946, 133.528217, 151.993991, 172.131254, 193.983995, 216.354372, 240.838655, 267.791287, 297.566712, 330.519373, 367.139216, 407.479566, 452.033771, 501.091409, 554.526031, 612.397975, 674.76758, 741.50692, 812.907385, 889.465743, 971.763668, 1059.38983, 1152.22846, 1250.18605, 1353.20675, 1461.21743, 1574.20348, 1692.10696, 1814.65522, 1941.69708, 2076.26029, 2214.52562, 2355.24147, 2497.15622, 2639.01828, 2789.40576, 2934.77989, 3071.43163, 3195.65193, 3303.73176, 3369.86575, 3427.17206, 3481.41148, 3562.65345, 3667.0366, 3780.57098, 3899.90167, 4021.67375, 4142.53229, 4259.12235, 4497.30672, 4686.14355, 4796.03829, 4797.39635, 4660.62319, 4356.12422, 3854.3049, 3344.34702, 2886.37483},
        {124.234194, 122.688106, 121.144171, 119.602571, 118.063487, 116.5271, 114.99359, 113.463139, 111.935928, 110.412137, 108.891948, 107.375541, 105.863098, 104.3548, 102.850827, 101.35136, 99.8565816, 98.3666713, 96.8818105, 95.4021801, 93.9279612, 92.4593347, 90.9964816, 89.5395829, 88.0888196, 86.6443726, 85.206423, 83.7751517, 82.3507397, 80.933368, 79.5232176, 78.1204695, 76.7253046, 75.3379039, 73.9828285, 72.6475564, 71.319273, 69.9973678, 68.6812309, 67.3702518, 66.0638206, 64.7613269, 63.4621605, 62.1657113, 60.8713691, 59.5785237, 58.2865648, 56.9948823, 55.702866, 54.4099057, 53.1153912, 51.7145603, 50.3132126, 48.9129958, 47.5155574, 46.122545, 44.7356064, 43.3563889, 41.9865404, 40.6277084, 39.2815405, 37.9496843, 36.6337874, 35.3354975, 34.0564622, 32.798329, 31.5627457, 30.4004534, 29.3610148, 28.3448255, 27.3503504, 26.3760548, 25.4204036, 24.4818619, 23.5588948, 22.6499673, 21.7535446, 20.8680917, 19.9920737, 19.1239556, 18.2622025, 17.4052795, 16.5516516, 15.699784, 14.7313591, 13.7056897, 12.6834921, 11.6670211, 10.6585317, 9.66027879, 8.67451719, 7.70350182, 6.74948756, 5.81472929, 4.90148192, 4.01200032, 3.14853939, 2.313354, 1.50869905, 0.736829418, 0, -0.0552081612, -0.239733539, -0.533604991, -0.916851376, -1.36950155, -1.87158438, -2.40312872, -2.94416343, -3.47471736, -3.97481938, -4.42449834, -4.80378311, -5.09270253, -5.27128548, -5.31956081, -5.21755737, -4.91622606, -4.39144174, -3.68900487, -2.78271331, -1.64636492, -0.253757571, 1.42131087, 3.40504254, 5.72363957, 8.4033041, 11.4702383, 14.9506442, 18.870724, 23.1471468, 27.9558099, 33.3557753, 39.4061052, 45.8816014, 52.9834561, 60.988664, 70.0415653, 80.2865, 91.8678083, 104.92983, 119.944575, 136.628989, 155.063302, 175.327744, 197.502547, 221.333991, 247.319745, 275.623527, 306.409054, 339.840046, 375.653543, 414.961435, 458.626965, 507.144421, 560.254815, 618.037364, 680.571286, 748.071802, 820.451217, 897.473466, 978.743014, 1064.89011, 1156.6696, 1254.34435, 1357.34359, 1465.4792, 1578.62814, 1696.61914, 1819.04239, 1945.62301, 2078.6326, 2214.85805, 2353.14303, 2492.33119, 2631.26619, 2778.29859, 2920.38842, 3054.00263, 3175.60816, 3281.67194, 3347.73531, 3405.14121, 3459.32469, 3538.44828, 3638.95073, 3747.80101, 3861.91133, 3978.19395, 4093.56109, 4204.92499, 4324.14526, 4477.72661, 4610.56774, 4667.56734, 4593.6241, 4333.63673, 3832.50391, 3368.69378, 2874.31968},
        {126.503785, 125.064768, 123.611345, 122.14433, 120.664534, 119.172769, 117.669848, 116.156583, 114.633787, 113.10227, 111.562846, 110.016327, 108.463524, 106.905251, 105.342319, 103.775541, 102.205728, 100.633693, 99.0602483, 97.4862057, 95.9123776, 94.3395762, 92.7686136, 91.2003021, 89.6354539, 88.074881, 86.5193959, 84.9698105, 83.4269372, 81.8915882, 80.3645755, 78.8467115, 77.3388083, 75.8416781, 74.4044312, 73.0027148, 71.6115957, 70.2303188, 68.8581289, 67.4942708, 66.1379894, 64.7885294, 63.4451358, 62.1070533, 60.7735268, 59.4438011, 58.117121, 56.7927314, 55.4698771, 54.1478029, 52.8257537, 51.3916244, 49.9584235, 48.5278097, 47.1014418, 45.6809784, 44.2680782, 42.8643999, 41.4716022, 40.0913436, 38.725283, 37.375079, 36.0423903, 34.7288755, 33.4361934, 32.1660026, 30.9199618, 29.7489326, 28.7025955, 27.6811315, 26.6830098, 25.7066994, 24.7506693, 23.8133887, 22.8933265, 21.9889518, 21.0987337, 20.2211412, 19.3546434, 18.4977093, 17.6488079, 16.8064083, 15.9689797, 15.1349909, 14.1833544, 13.1748323, 12.1716236, 11.1760774, 10.1905424, 9.21736757, 8.25890181, 7.31749402, 6.39549308, 5.4952479, 4.61910736, 3.76942038, 2.94853583, 2.15880263, 1.40256965, 0.682185813, 0, -0.0106689959, -0.144368407, -0.381285884, -0.701609076, -1.08552563, -1.5132232, -1.96488944, -2.42071198, -2.86087849, -3.26557662, -3.614994, -3.8893183, -4.06873715, -4.13343822, -4.06360915, -3.83943758, -3.39977507, -2.71890409, -1.87051065, -0.825924657, 0.443523978, 1.96650534, 3.77168953, 5.88774662, 8.3433467, 11.1671599, 14.3878562, 18.0341058, 22.1345788, 26.6418669, 31.688614, 37.3263135, 43.6064589, 50.267393, 57.5171847, 65.64741, 74.8035078, 85.1309169, 96.7750763, 109.881425, 124.810753, 141.427605, 159.835288, 180.137106, 202.436364, 226.813399, 253.400226, 282.305892, 313.639441, 347.509922, 383.488629, 422.879613, 466.621953, 515.235591, 568.384726, 626.117767, 688.483125, 755.56682, 827.371102, 903.857196, 984.82738, 1070.72854, 1162.14423, 1259.28518, 1361.73028, 1469.34846, 1582.23949, 1700.33216, 1822.70868, 1948.93013, 2080.4764, 2214.69455, 2350.48144, 2486.73394, 2622.34893, 2765.05789, 2902.71442, 3032.00672, 3149.62301, 3252.2515, 3316.48066, 3372.49827, 3425.60675, 3503.51855, 3602.66541, 3710.14153, 3822.84552, 3937.67598, 4051.53153, 4161.31077, 4241.11891, 4380.57987, 4518.05706, 4591.91385, 4540.51364, 4302.21982, 3815.39578, 3371.38489, 2884.97364},
        {128.77986, 127.468415, 126.124046, 124.748264, 123.342581, 121.908507, 120.447554, 118.961234, 117.451058, 115.918538, 114.365185, 112.79251, 111.202025, 109.595242, 107.973671, 106.338824, 104.692213, 103.035349, 101.369743, 99.6969073, 98.0183528, 96.335591, 94.6501334, 92.9634913, 91.2771763, 89.5926998, 87.9115732, 86.2353078, 84.5654153, 82.9034069, 81.2507941, 79.6090884, 77.9798011, 76.3644438, 74.8382594, 73.3643427, 71.9049024, 70.459057, 69.0259254, 67.6046262, 66.1942781, 64.7939998, 63.40291, 62.0201273, 60.6447706, 59.2759585, 57.9128096, 56.5544428, 55.1999766, 53.8485298, 52.4992211, 51.0359115, 49.5754758, 48.1195315, 46.6696957, 45.2275858, 43.7948189, 42.3730124, 40.9637834, 39.5687494, 38.1895274, 36.8277349, 35.4849889, 34.1629069, 32.8631061, 31.5872036, 30.3368169, 29.1620528, 28.1118541, 27.0878316, 26.0884594, 25.1122111, 24.1575607, 23.2229821, 22.3069491, 21.4079356, 20.5244154, 19.6548625, 18.7977507, 17.9515538, 17.1147458, 16.2858005, 15.4631918, 14.6453935, 13.7069207, 12.7108341, 11.7216841, 10.7419671, 9.77417986, 8.82081887, 7.88438071, 6.96736195, 6.07225916, 5.20156891, 4.35778776, 3.54341229, 2.76093906, 2.01286465, 1.30168562, 0.629898551, 0, 0.143299008, 0.151944856, 0.0465353503, -0.152331705, -0.424058505, -0.748047244, -1.10370012, -1.47041932, -1.82760705, -2.15466549, -2.43099685, -2.63600332, -2.74908709, -2.74965036, -2.61709532, -2.33082417, -1.79755461, -0.986305336, -0.0408839448, 1.07488262, 2.39716741, 3.96214348, 5.80598388, 7.96486167, 10.4749499, 13.3724216, 16.6934499, 20.4742078, 24.7508683, 29.5493542, 34.9198474, 40.9015958, 47.5338478, 54.5164391, 62.0583239, 70.4589666, 79.8594391, 90.4008132, 102.224161, 115.470554, 130.39729, 146.993841, 165.378541, 185.669721, 207.985712, 232.409789, 259.104107, 288.195761, 319.811847, 354.079463, 390.72136, 430.763176, 475.039556, 523.995823, 577.282441, 634.906748, 696.876085, 762.959274, 833.456379, 908.927664, 989.898783, 1076.08417, 1167.43339, 1263.92107, 1365.56436, 1472.36092, 1584.76634, 1702.897, 1825.19005, 1951.03251, 2081.13197, 2213.35171, 2346.62163, 2479.87162, 2612.03159, 2750.03841, 2882.81326, 3007.28428, 3120.37961, 3219.02742, 3280.72386, 3334.78373, 3386.46317, 3463.64451, 3562.62768, 3670.28509, 3783.38805, 3898.70789, 4013.01596, 4123.08358, 4204.28681, 4340.55557, 4473.5955, 4545.11228, 4496.81155, 4270.39897, 3807.58022, 3370.96891, 2916.16954},
        {131.136568, 129.956266, 128.724142, 127.442419, 126.113318, 124.73906, 123.321867, 121.863961, 120.367563, 118.834896, 117.268179, 115.669636, 114.041488, 112.385956, 110.705262, 109.001627, 107.277274, 105.534424, 103.775298, 102.002118, 100.217106, 98.4224828, 96.6204708, 94.8132913, 93.003166, 91.1923164, 89.3829641, 87.5773308, 85.7776381, 83.9861076, 82.204961, 80.4364197, 78.6827055, 76.9460399, 75.3278078, 73.7785634, 72.247669, 70.7341284, 69.236945, 67.7551224, 66.2876643, 64.8335742, 63.3918557, 61.9615123, 60.5415477, 59.1309654, 57.728769, 56.3339621, 54.9455483, 53.5625311, 52.1839141, 50.6903904, 49.2018388, 47.7198277, 46.2459254, 44.7817001, 43.3287202, 41.888554, 40.4627699, 39.0529361, 37.660621, 36.2873929, 34.9348201, 33.604471, 32.2979138, 31.0167168, 29.7624485, 28.5844501, 27.5309159, 26.504886, 25.5048318, 24.5292247, 23.5765361, 22.6452373, 21.7337998, 20.840695, 19.9643942, 19.1033688, 18.2560902, 17.4210299, 16.5966591, 15.7814494, 14.973872, 14.1723983, 13.2464267, 12.26168, 11.2854323, 10.3203437, 9.36907427, 8.43428411, 7.51863325, 6.62478176, 5.7553897, 4.91311712, 4.1006241, 3.32057069, 2.57561695, 1.86842296, 1.20164876, 0.577954417, 0, 0.34262395, 0.531583936, 0.588640902, 0.53555579, 0.394089546, 0.186003112, -0.0669425683, -0.342986551, -0.620367893, -0.877325651, -1.09209888, -1.24292664, -1.30804798, -1.26570197, -1.09412765, -0.771564086, -0.164119651, 0.765985268, 1.78688269, 2.94436156, 4.28421083, 5.85221944, 7.69417635, 9.85587049, 12.3830908, 15.3216263, 18.7172658, 22.6157984, 27.0630129, 32.1761183, 37.9032962, 44.2689097, 51.2973216, 58.6399936, 66.5077392, 75.2108126, 84.8854473, 95.6678766, 107.694334, 121.101053, 136.065895, 152.670796, 171.043844, 191.313129, 213.606739, 237.87085, 264.460941, 293.55058, 325.313334, 359.922768, 397.380205, 438.241978, 483.134529, 532.383815, 585.702306, 643.077915, 704.498554, 769.447975, 838.532836, 912.90979, 993.860469, 1080.32909, 1171.60532, 1267.41332, 1368.21339, 1474.12796, 1585.94906, 1703.97271, 1826.03957, 1951.37921, 2080.01233, 2210.25561, 2341.06475, 2471.39542, 2600.20333, 2733.60659, 2861.60786, 2981.37223, 3090.06477, 3184.85059, 3244.10445, 3296.30862, 3346.68116, 3423.2682, 3522.25171, 3630.30194, 3744.07643, 3860.23274, 3975.4284, 4086.32096, 4182.32656, 4319.06887, 4445.10139, 4508.97765, 4459.25115, 4244.47543, 3813.20399, 3381.83256, 2966.73423},
        {133.721218, 132.657199, 131.523196, 130.322127, 129.056909, 127.730458, 126.345691, 124.905526, 123.412878, 121.870667, 120.281807, 118.649217, 116.975813, 115.264511, 113.51823, 111.739886, 109.932396, 108.098677, 106.241646, 104.36422, 102.469316, 100.55985, 98.6387403, 96.7089033, 94.7732559, 92.8347152, 90.8961981, 88.9606216, 87.0309026, 85.1099582, 83.2007052, 81.3060606, 79.4289414, 77.5722646, 75.8649143, 74.2441745, 72.6458147, 71.0686641, 69.5115518, 67.9733068, 66.4527583, 64.9487352, 63.4600668, 61.985582, 60.52411, 59.0744798, 57.6355205, 56.2060613, 54.7849311, 53.370959, 51.9629742, 50.4320102, 48.9074619, 47.3909288, 45.8840103, 44.3883059, 42.9054151, 41.4369372, 39.9844717, 38.549618, 37.1339756, 35.7391439, 34.3667224, 33.0183105, 31.6955076, 30.3999131, 29.1331266, 27.9454716, 26.8861017, 25.8561266, 24.8539872, 23.8781242, 22.9269786, 21.9989912, 21.0926028, 20.2062543, 19.3383865, 18.4874404, 17.6518567, 16.8300763, 16.02054, 15.2216888, 14.4319634, 13.6498047, 12.7394726, 11.7693207, 10.8093155, 9.86225215, 8.93092607, 8.0181325, 7.12666672, 6.25932402, 5.41889968, 4.60818898, 3.82998721, 3.08708965, 2.38229159, 1.71838831, 1.0981751, 0.524447231, 0, 0.591322616, 0.999833493, 1.24916951, 1.36296753, 1.36486444, 1.27849712, 1.12750244, 0.935517283, 0.726178516, 0.523123018, 0.349987668, 0.23040934, 0.188024911, 0.246471258, 0.429385256, 0.760403783, 1.40711472, 2.41371938, 3.47825117, 4.65519359, 5.99903018, 7.56424446, 9.40531994, 11.5767402, 14.1329886, 17.1285488, 20.6179044, 24.6555387, 29.2959354, 34.7566023, 40.869223, 47.6393737, 55.0726305, 62.7245849, 70.8258054, 79.7268567, 89.5683104, 100.490738, 112.634711, 126.140802, 141.210019, 157.904102, 176.351799, 196.681856, 219.023021, 243.140761, 269.617923, 298.674072, 330.528774, 365.401596, 403.588166, 445.139024, 490.319006, 539.307635, 592.110255, 648.810415, 709.491661, 773.822008, 842.394977, 916.257393, 996.69216, 1082.63746, 1173.36891, 1268.60788, 1368.83108, 1474.16853, 1585.49594, 1703.13665, 1824.67969, 1949.26076, 2076.46343, 2204.90664, 2333.57107, 2461.4374, 2587.4863, 2717.13808, 2841.32388, 2957.41447, 3062.78063, 3154.79311, 3212.82969, 3264.24949, 3314.13456, 3389.81162, 3487.55756, 3594.38065, 3707.0212, 3822.21954, 3936.71598, 4047.25082, 4145.69971, 4279.34162, 4401.67522, 4466.19919, 4426.41221, 4235.81295, 3847.9001, 3427.03608, 3035.7362},
        {136.633402, 135.656415, 134.593047, 133.446865, 132.221437, 130.920331, 129.547114, 128.105354, 126.598619, 125.030477, 123.404495, 121.724241, 119.993282, 118.215187, 116.393522, 114.531857, 112.633757, 110.702792, 108.742529, 106.756535, 104.748378, 102.721626, 100.679847, 98.6266076, 96.5654764, 94.500021, 92.4338089, 90.3704079, 88.3133856, 86.2663098, 84.2327481, 82.2162683, 80.220438, 78.2488249, 76.4593876, 74.7752509, 73.1176841, 71.4852446, 69.8764898, 68.2899771, 66.7242639, 65.1779078, 63.649466, 62.137496, 60.6405552, 59.1572011, 57.685991, 56.2254824, 54.7742327, 53.3307993, 51.8937397, 50.3135197, 48.7399987, 47.1749443, 45.6201243, 44.0773064, 42.5482585, 41.0347483, 39.5385435, 38.0614118, 36.6051211, 35.1714391, 33.7621335, 32.3789721, 31.0237226, 29.6981529, 28.4040305, 27.1955586, 26.1256811, 25.088022, 24.0809501, 23.1028338, 22.1520418, 21.2269426, 20.3259049, 19.4472972, 18.5894882, 17.7508463, 16.9297403, 16.1245387, 15.3336101, 14.5553231, 13.7880463, 13.0301483, 12.1412868, 11.1921036, 10.2549073, 9.33256795, 8.42795558, 7.54394022, 6.68339189, 5.8491806, 5.04417639, 4.27124927, 3.53326927, 2.83310641, 2.17363071, 1.5577122, 0.988220896, 0.468026822, 0, 0.881216657, 1.54040805, 2.00390857, 2.29805264, 2.44917465, 2.48360901, 2.42769012, 2.30775238, 2.15013021, 1.98115799, 1.82717014, 1.71450106, 1.66948516, 1.71845682, 1.88775047, 2.20370051, 2.84902746, 3.87240661, 4.94644183, 6.13097883, 7.48586333, 9.07094105, 10.9460577, 13.171059, 15.8057907, 18.9100985, 22.5438281, 26.7668252, 31.6389355, 37.4765916, 43.9889706, 51.1589421, 58.9693757, 66.8117736, 74.9646892, 83.8643741, 93.671108, 104.545171, 116.646843, 130.136405, 145.390191, 162.28667, 180.94385, 201.479739, 224.012345, 248.162017, 274.668837, 303.775227, 335.72361, 370.756408, 409.387153, 451.255804, 496.097198, 543.936703, 595.392849, 650.817849, 710.563914, 775.237553, 844.878888, 919.250627, 998.351579, 1082.41938, 1171.81577, 1266.6842, 1366.79825, 1472.10127, 1583.1393, 1700.01192, 1820.60899, 1944.07033, 2069.91649, 2196.84797, 2323.87278, 2449.99894, 2574.23448, 2701.52443, 2823.45553, 2937.55154, 3041.33621, 3132.33331, 3191.14434, 3243.49682, 3294.16689, 3368.27634, 3462.54785, 3565.22675, 3673.45958, 3784.3929, 3895.17324, 4002.94716, 4073.10673, 4194.94922, 4321.10216, 4404.19305, 4396.84944, 4251.69883, 3921.36875, 3522.40209, 3122.84363},
        {139.66004, 138.752904, 137.745237, 136.641158, 135.444786, 134.160237, 132.791629, 131.343082, 129.818711, 128.222636, 126.558974, 124.831843, 123.045361, 121.203646, 119.310815, 117.370987, 115.388279, 113.36681, 111.310696, 109.224057, 107.111009, 104.975672, 102.822161, 100.654597, 98.4770957, 96.2937757, 94.1087549, 91.9261511, 89.7500821, 87.584666, 85.4340204, 83.3022634, 81.1935127, 79.1118863, 77.2387915, 75.4859722, 73.7630419, 72.068365, 70.400306, 68.7572294, 67.1374996, 65.5394813, 63.9615388, 62.4020366, 60.8593394, 59.3318114, 57.8178173, 56.3157215, 54.8238886, 53.3406829, 51.8644691, 50.2290956, 48.6010094, 46.9821411, 45.3744218, 43.7797823, 42.2001535, 40.6374662, 39.0936514, 37.5706399, 36.0703626, 34.5947503, 33.145734, 31.7252446, 30.3352128, 28.9775696, 27.6542458, 26.4233758, 25.3417443, 24.295367, 23.2825313, 22.3015247, 21.3506345, 20.4281482, 19.5323533, 18.6615372, 17.8139873, 16.9879911, 16.181836, 15.3938094, 14.6221989, 13.8652917, 13.1213754, 12.3887374, 11.5226092, 10.5958152, 9.68289917, 8.78679106, 7.91042072, 7.05671805, 6.22861292, 5.4290352, 4.6609148, 3.92718157, 3.2307654, 2.57459618, 1.96160378, 1.39471808, 0.876868964, 0.41098631, 0, 1.12421804, 1.99568192, 2.64287104, 3.09426482, 3.37834266, 3.52358397, 3.55846815, 3.51147462, 3.41108279, 3.28577205, 3.16402183, 3.07431152, 3.04512054, 3.10492829, 3.28221419, 3.60545763, 4.26679136, 5.31807384, 6.41761621, 7.62896648, 9.01567261, 10.6412826, 12.5693444, 14.863406, 17.5870153, 20.8037205, 24.5770693, 28.9706099, 34.0478902, 40.2030366, 47.0478065, 54.5465745, 62.6637148, 70.6484664, 78.8227714, 87.699275, 97.4568921, 108.274538, 120.331128, 133.805576, 149.261755, 166.376463, 185.253296, 205.995853, 228.707731, 252.866052, 279.357508, 308.442314, 340.380687, 375.432842, 414.287725, 456.25282, 500.576655, 547.152974, 597.192836, 651.315852, 710.141634, 775.256223, 846.093157, 921.031688, 998.652331, 1080.63071, 1168.54559, 1263.00705, 1362.98386, 1468.19817, 1578.76893, 1694.52118, 1813.82501, 1935.87354, 2060.36143, 2185.90304, 2311.51783, 2436.22527, 2559.04482, 2684.48517, 2804.70427, 2917.3493, 3020.06744, 3110.50586, 3170.4337, 3223.96151, 3275.54134, 3347.80389, 3438.0055, 3535.82802, 3638.9065, 3744.87598, 3851.3715, 3956.0281, 3997.2343, 4107.30961, 4238.07439, 4341.34897, 4368.95371, 4272.70895, 4004.43504, 3636.32838, 3231.65426},
        {142.743311, 141.887773, 140.92056, 139.846199, 138.669214, 137.394129, 136.025468, 134.567756, 133.025519, 131.403279, 129.705563, 127.936894, 126.101796, 124.204796, 122.250416, 120.243182, 118.187618, 116.088248, 113.949598, 111.776192, 109.572553, 107.343208, 105.09268, 102.825493, 100.546174, 98.2592449, 95.9692316, 93.6806584, 91.3980498, 89.1259304, 86.8688246, 84.631257, 82.4177522, 80.2328348, 78.272218, 76.4408106, 74.6412234, 72.8717766, 71.1307901, 69.4165841, 67.7274787, 66.0617939, 64.41785, 62.7939669, 61.1884648, 59.5996637, 58.0258839, 56.4654453, 54.916668, 53.3778723, 51.847378, 50.1516541, 48.4645973, 46.7882532, 45.1246672, 43.4758848, 41.8439516, 40.230913, 38.6388146, 37.0697018, 35.5256201, 34.0086151, 32.5207322, 31.0640169, 29.6405148, 28.2522713, 26.9013318, 25.6489524, 24.554967, 23.4993044, 22.4801717, 21.4957761, 20.5443248, 19.6240247, 18.7330831, 17.8697072, 17.0321039, 16.2184806, 15.4270443, 14.6560021, 13.9035612, 13.1679287, 12.4473117, 11.7399175, 10.896822, 9.99289267, 9.10476525, 8.23542155, 7.38784339, 6.56501258, 5.76991096, 5.00552034, 4.27482253, 3.58079937, 2.92643266, 2.31470424, 1.74859591, 1.23108951, 0.765166838, 0.353809729, 0, 1.27191814, 2.27821235, 3.04831675, 3.61166546, 3.99769259, 4.23583227, 4.35551862, 4.38618575, 4.35726779, 4.29819886, 4.23841307, 4.20734454, 4.2344274, 4.34909576, 4.58078375, 4.95892548, 5.68002618, 6.79738728, 7.96254395, 9.24073127, 10.6971843, 12.3971381, 14.4058278, 16.7884884, 19.6103551, 22.9366628, 26.8326466, 31.3635417, 36.594583, 42.961206, 50.0227052, 57.7332559, 66.0470329, 74.127881, 82.3251403, 91.1989057, 100.940452, 111.741052, 123.791981, 137.284512, 152.936421, 170.252242, 189.320238, 210.228672, 233.065806, 257.189793, 283.601535, 312.571824, 344.371451, 379.271207, 418.064286, 459.860585, 503.565339, 548.988039, 597.78437, 650.781113, 708.805049, 774.245205, 846.011062, 921.307828, 997.479582, 1077.45977, 1163.91267, 1257.85609, 1357.51796, 1462.40678, 1572.19644, 1686.43833, 1804.07754, 1924.40209, 2047.51075, 2171.75613, 2296.14564, 2419.68672, 2541.38677, 2665.31489, 2784.1514, 2895.63831, 2997.5176, 3087.53127, 3148.46849, 3202.99264, 3255.25409, 3325.41115, 3411.27768, 3503.9976, 3601.7183, 3702.58714, 3804.75151, 3906.3588, 3923.96207, 4024.1533, 4159.02674, 4280.67664, 4341.19726, 4292.68284, 4087.22765, 3760.7617, 3363.81459},
        {145.807159, 144.983001, 144.040069, 142.983092, 141.8168, 140.545922, 139.175188, 137.709328, 136.153071, 134.511147, 132.788284, 130.989214, 129.118664, 127.181365, 125.182047, 123.125438, 121.016269, 118.859269, 116.659167, 114.420693, 112.148576, 109.847547, 107.522334, 105.177668, 102.818277, 100.448891, 98.0742402, 95.6990535, 93.3280607, 90.9659913, 88.6175747, 86.2875406, 83.9806185, 81.701538, 79.6464943, 77.7204568, 75.8260937, 73.9619212, 72.1264554, 70.3182122, 68.535708, 66.7774587, 65.0419805, 63.3277894, 61.6334017, 59.9573333, 58.2981004, 56.6542192, 55.0242057, 53.4065759, 51.7998462, 50.0391388, 48.2899057, 46.5542049, 44.8340946, 43.131633, 41.4488782, 39.7878885, 38.1507219, 36.5394366, 34.9560908, 33.4027426, 31.8814503, 30.3942719, 28.9432656, 27.5304896, 26.158002, 24.8883712, 23.7827127, 22.7182869, 21.6932294, 20.7056757, 19.7537614, 18.8356219, 17.9493929, 17.0932098, 16.2652082, 15.4635236, 14.6862916, 13.9316476, 13.1977273, 12.4826661, 11.7845996, 11.1016633, 10.2809811, 9.39937069, 8.53546479, 7.69229955, 6.87291112, 6.08033565, 5.31760929, 4.58776819, 3.89384851, 3.23888639, 2.62591798, 2.05797944, 1.53810691, 1.06933656, 0.654704519, 0.29724695, 0, 1.25361013, 2.26018833, 3.04801314, 3.64536306, 4.08051663, 4.38175236, 4.57734877, 4.69558439, 4.76473774, 4.81308734, 4.86891172, 4.96048938, 5.11609886, 5.36401868, 5.73252735, 6.24990341, 7.11250477, 8.37262129, 9.67835191, 11.0939922, 12.6838376, 14.5121837, 16.643326, 19.1415601, 22.0711815, 25.4964858, 29.4817685, 34.091325, 39.389451, 45.7938115, 52.8857636, 60.6235923, 68.965582, 77.0973128, 85.3634217, 94.3146005, 104.140945, 115.032553, 127.179518, 140.771939, 156.586306, 174.047853, 193.231944, 214.213946, 237.069227, 261.08849, 287.32793, 316.059077, 347.553464, 382.082621, 420.435131, 461.733526, 504.811991, 549.475953, 597.513795, 649.823594, 707.303427, 772.691327, 844.627244, 919.7439, 994.7307, 1073.19589, 1158.42882, 1251.66516, 1350.66065, 1454.76835, 1563.29204, 1675.57199, 1791.12935, 1909.38296, 2031.08063, 2154.10777, 2277.42319, 2399.98571, 2520.75414, 2643.27265, 2760.76835, 2871.05371, 2971.9412, 3061.24328, 3122.52021, 3177.33816, 3229.6177, 3297.42866, 3379.09084, 3467.02855, 3559.85637, 3656.1889, 3754.64073, 3853.82646, 3860.47417, 3954.95688, 4091.96053, 4226.17102, 4312.27427, 4304.95621, 4158.90275, 3887.9513, 3524.48976},
        {148.599376, 147.775764, 146.834089, 145.778879, 144.61466, 143.345962, 141.977312, 140.513238, 138.958267, 137.316928, 135.593747, 133.793254, 131.919975, 129.978438, 127.973172, 125.908704, 123.789561, 121.620272, 119.405365, 117.149366, 114.856805, 112.532208, 110.180104, 107.80502, 105.411484, 103.004023, 100.587167, 98.1654417, 95.7433758, 93.3254968, 90.9163326, 88.5204109, 86.1422595, 83.7864062, 81.619775, 79.5622797, 77.5318824, 75.5278409, 73.5494128, 71.5958558, 69.6664277, 67.7603862, 65.876989, 64.0154938, 62.1751583, 60.3552403, 58.5549974, 56.7736874, 55.010568, 53.2648969, 51.5359318, 49.7048525, 47.8915544, 46.0978549, 44.3255713, 42.5765211, 40.8525218, 39.1553906, 37.486945, 35.8490025, 34.2433805, 32.6718963, 31.1363674, 29.6386111, 28.180445, 26.7636864, 25.3901527, 24.1195888, 23.0103485, 21.9447785, 20.920941, 19.9368981, 18.9907121, 18.0804452, 17.2041596, 16.3599176, 15.5457813, 14.759813, 14.0000749, 13.2646292, 12.5515381, 11.8588638, 11.1846686, 10.5270147, 9.72763172, 8.86603647, 8.02362462, 7.20353168, 6.40889316, 5.64284456, 4.9085214, 4.20905917, 3.54759339, 2.92725956, 2.35119319, 1.82252979, 1.34440486, 0.919953904, 0.55231244, 0.24461597, 0, 0.783175506, 1.42382157, 1.94330416, 2.36298921, 2.7042427, 2.98843056, 3.23691877, 3.47107327, 3.71226002, 3.98184498, 4.3011941, 4.69167334, 5.17464865, 5.77148599, 6.50355132, 7.39221058, 8.62812086, 10.2568326, 11.9167913, 13.6656395, 15.5610201, 17.6605754, 20.0219483, 22.7027812, 25.7607169, 29.2533979, 33.2384669, 37.7735665, 42.9163394, 48.9120716, 55.5619597, 62.8673532, 70.8296016, 78.9604211, 87.505978, 96.8430074, 107.119749, 118.484443, 131.085328, 145.070644, 160.967084, 178.429253, 197.531346, 218.347557, 240.95208, 264.694419, 290.554632, 318.788086, 349.650148, 383.396186, 420.576551, 460.79109, 503.493274, 548.674566, 597.424825, 650.340751, 708.019046, 772.237188, 842.142742, 915.595146, 990.420921, 1069.10448, 1154.12746, 1246.35771, 1343.92875, 1446.22918, 1552.49211, 1662.06585, 1774.8689, 1890.49723, 2010.82292, 2132.81564, 2255.28347, 2377.03448, 2496.87673, 2617.27402, 2732.46478, 2840.34317, 2938.80332, 3025.73939, 3085.03792, 3137.93907, 3188.34024, 3253.55667, 3332.17454, 3417.18926, 3507.46655, 3601.87211, 3699.27168, 3798.53096, 3826.55626, 3926.06484, 4060.00998, 4191.34498, 4283.02311, 4297.99768, 4199.22196, 4013.06893, 3752.83681},
        {151.484769, 150.650555, 149.701005, 148.640346, 147.472809, 146.202622, 144.834014, 143.371213, 141.818448, 140.179949, 138.459944, 136.662661, 134.79233, 132.85318, 130.849439, 128.785336, 126.6651, 124.492959, 122.273144, 120.009882, 117.707402, 115.369933, 113.001704, 110.606943, 108.18988, 105.754744, 103.305763, 100.847166, 98.3831813, 95.9180386, 93.4559663, 91.0011934, 88.5579486, 86.1304608, 83.8500976, 81.6548443, 79.481158, 77.3291414, 75.1988977, 73.0905296, 71.0041401, 68.9398322, 66.8977087, 64.8778726, 62.8804268, 60.9054742, 58.9531177, 57.0234603, 55.1166049, 53.2326543, 51.3717116, 49.4687385, 47.5903912, 45.7381848, 43.9136343, 42.1182547, 40.3535611, 38.6210686, 36.9222923, 35.2587471, 33.6319481, 32.0434104, 30.494649, 28.987179, 27.5225155, 26.1021734, 24.7276679, 23.4541082, 22.3353169, 21.2622749, 20.233023, 19.2456019, 18.2980524, 17.3884152, 16.5147311, 15.6750408, 14.8673852, 14.0898049, 13.3403408, 12.6170335, 11.9179239, 11.2410526, 10.5844605, 9.94618835, 9.16500925, 8.32194775, 7.49994266, 6.70220327, 5.93193888, 5.19235879, 4.48667229, 3.81808868, 3.18981725, 2.6050673, 2.06704813, 1.57896904, 1.14403931, 0.765468244, 0.44646514, 0.190239294, 0, 0.329027733, 0.617314425, 0.879460568, 1.13006665, 1.38373317, 1.65506062, 1.95864949, 2.30910026, 2.72101344, 3.20898952, 3.78762898, 4.47153232, 5.27530003, 6.2135326, 7.30083053, 8.5517943, 10.1502349, 12.1349258, 14.1377295, 16.2095056, 18.4011143, 20.7634153, 23.3472686, 26.203534, 29.3830715, 32.9367408, 36.915402, 41.3699148, 46.3511393, 51.9369687, 58.1413171, 65.0069345, 72.5765707, 80.6748084, 89.4534811, 99.1226, 109.790365, 121.564976, 134.554634, 148.867537, 164.790377, 182.19854, 201.165304, 221.763947, 244.067747, 267.519782, 292.98108, 320.682469, 350.854777, 383.72883, 419.56335, 458.527182, 500.7897, 546.558374, 596.118458, 649.720281, 707.614173, 770.477013, 838.035639, 909.551567, 984.152622, 1063.03205, 1147.7122, 1238.63947, 1334.43717, 1434.56542, 1538.24986, 1644.88982, 1754.74449, 1867.58665, 1986.36455, 2107.18701, 2228.72765, 2349.66008, 2468.6579, 2587.4377, 2700.86937, 2806.86576, 2903.33976, 2988.2042, 3045.36254, 3096.07669, 3144.26353, 3206.71968, 3282.26383, 3364.34812, 3452.05958, 3544.48522, 3640.71208, 3739.82716, 3797.85393, 3904.73771, 4033.30254, 4156.3725, 4246.77163, 4277.32401, 4220.85368, 4121.04962, 3968.9479},
        {154.592035, 153.743867, 152.783213, 151.714003, 150.540165, 149.265629, 147.894322, 146.430175, 144.877115, 143.239072, 141.519974, 139.72375, 137.854329, 135.915639, 133.91161, 131.84617, 129.723249, 127.546773, 125.320674, 123.048879, 120.735317, 118.383916, 115.998607, 113.583317, 111.141976, 108.678511, 106.196852, 103.700928, 101.194668, 98.6819997, 96.1668525, 93.6531551, 91.1448364, 88.645825, 86.2522553, 83.920014, 81.6038793, 79.3047879, 77.0236762, 74.7614809, 72.5191385, 70.2975855, 68.0977584, 65.9205938, 63.7670283, 61.6379984, 59.5344406, 57.4572914, 55.4074875, 53.3859654, 51.3936616, 49.420417, 47.4785043, 45.5691006, 43.693383, 41.8525285, 40.0477143, 38.2801173, 36.5509148, 34.8612837, 33.2124011, 31.6054441, 30.0415898, 28.5220153, 27.0478976, 25.6204138, 24.2407409, 22.9573163, 21.8174423, 20.7248257, 19.6775797, 18.6738178, 17.7116534, 16.7891999, 15.9045706, 15.055879, 14.2412384, 13.4587622, 12.7065639, 11.9827568, 11.2854544, 10.6127699, 9.96281684, 9.33370858, 8.56583178, 7.73948076, 6.93684561, 6.16115827, 5.41565065, 4.70355468, 4.0281023, 3.39252543, 2.80005601, 2.25392596, 1.7573672, 1.31361168, 0.925891313, 0.597438033, 0.331483768, 0.131260447, 0, 0.0419024707, 0.114046083, 0.226268775, 0.388408483, 0.610303146, 0.9017907, 1.27270908, 1.73289623, 2.29219009, 2.96042858, 3.74744966, 4.66309125, 5.71719129, 6.91958772, 8.28011849, 9.80862152, 11.6850289, 13.9434778, 16.2090692, 18.52809, 20.9468268, 23.5115662, 26.268595, 29.2641998, 32.5446674, 36.1562844, 40.1453375, 44.5581133, 49.4408987, 54.767367, 60.6830431, 67.2559974, 74.5543006, 82.5894877, 91.4578972, 101.270944, 112.113658, 124.071072, 137.228217, 151.670124, 167.564632, 184.888762, 203.711344, 224.101209, 246.127185, 269.343873, 294.462891, 321.681624, 351.197459, 383.207785, 417.658709, 455.306015, 496.999318, 543.15738, 593.319308, 647.419158, 705.390985, 766.947785, 832.294912, 901.878871, 975.894437, 1054.47489, 1138.33415, 1227.57406, 1321.25931, 1418.93066, 1519.83694, 1623.44323, 1730.28498, 1840.29218, 1957.26068, 2076.65986, 2197.08153, 2317.11749, 2435.35955, 2553.39645, 2666.07383, 2771.23425, 2866.72028, 2950.37449, 3005.81815, 3054.59599, 3100.64542, 3160.21802, 3232.37684, 3311.05404, 3395.56921, 3485.24198, 3579.39193, 3677.33866, 3771.17826, 3886.56005, 4007.30919, 4117.25085, 4200.21021, 4240.01243, 4220.48268, 4199.91579, 4142.61444},
        {158.269136, 157.420863, 156.458918, 155.387131, 154.209334, 152.929358, 151.551034, 150.078192, 148.514663, 146.86428, 145.130872, 143.318271, 141.430307, 139.470813, 137.443618, 135.352553, 133.201451, 130.994141, 128.734455, 126.426224, 124.073278, 121.679449, 119.248568, 116.784466, 114.290973, 111.771922, 109.231142, 106.672464, 104.099721, 101.516743, 98.9273597, 96.3354038, 93.7447056, 91.1590963, 88.6522161, 86.191523, 83.7436359, 81.3101206, 78.8925424, 76.492467, 74.11146, 71.7510868, 69.4129131, 67.0985045, 64.8094264, 62.5472444, 60.3135242, 58.1098312, 55.937731, 53.7987892, 51.6945714, 49.6622661, 47.6670436, 45.7096973, 43.7910205, 41.9118064, 40.0728485, 38.27494, 36.5188743, 34.8054446, 33.1354444, 31.5096669, 29.9289054, 28.3939533, 26.9056039, 25.4646506, 24.0718866, 22.7649401, 21.580555, 20.4435555, 19.3523468, 18.3053347, 17.3009244, 16.3375215, 15.4135314, 14.5273597, 13.6774116, 12.8620929, 12.0798088, 11.3289648, 10.6079666, 9.91521935, 9.2491287, 8.60810008, 7.84448281, 7.03278252, 6.24926033, 5.4970615, 4.77933131, 4.09921501, 3.45985788, 2.86440519, 2.31600221, 1.8177942, 1.37292644, 0.984544182, 0.65579271, 0.389817287, 0.189763182, 0.0587756632, 0, 0.207816799, 0.433659692, 0.688050527, 0.981511151, 1.32456341, 1.72772916, 2.20153024, 2.7564885, 3.40312579, 4.15196396, 5.01352486, 5.99833033, 7.11690222, 8.37976238, 9.79743265, 11.3804349, 13.3220342, 15.6588585, 17.988082, 20.3593858, 22.8224508, 25.4269584, 28.2225895, 31.2590253, 34.585947, 38.2530355, 42.3099721, 46.8064378, 51.7921137, 57.331119, 63.4534029, 70.204315, 77.6292051, 85.5416238, 94.1028203, 103.535857, 113.955623, 125.477009, 138.214903, 152.284194, 168.139144, 185.451984, 204.271204, 224.645297, 246.622751, 269.862447, 294.899889, 321.880973, 350.951593, 382.257643, 415.409758, 451.743297, 492.601444, 538.580316, 588.750768, 642.868465, 700.689073, 761.554164, 825.72761, 893.925025, 966.424067, 1043.2816, 1125.12258, 1212.02887, 1303.16207, 1398.10621, 1496.09379, 1596.61769, 1700.4597, 1807.67254, 1922.47347, 2040.10995, 2159.19313, 2278.33416, 2396.14416, 2515.0934, 2628.96913, 2735.41769, 2832.08546, 2916.61878, 2972.19373, 3020.57383, 3065.60697, 3122.2219, 3189.88037, 3263.42714, 3342.55482, 3426.95602, 3516.32335, 3610.34942, 3747.27235, 3872.34674, 3981.88497, 4072.19938, 4139.60235, 4180.40623, 4190.92339, 4230.24435, 4220.30632},
        {161.951839, 161.095369, 160.125295, 159.045307, 157.8591, 156.570365, 155.182797, 153.700087, 152.125928, 150.464014, 148.718037, 146.89169, 144.988665, 143.012656, 140.967355, 138.856456, 136.68365, 134.452631, 132.167092, 129.830726, 127.447225, 125.020282, 122.553589, 120.050841, 117.515729, 114.951947, 112.363187, 109.753142, 107.125505, 104.483969, 101.832226, 99.1739694, 96.5128922, 93.8526869, 91.2454886, 88.66975, 86.1033419, 83.548385, 81.0070001, 78.4813079, 75.9734294, 73.4854852, 71.0195961, 68.577883, 66.1624666, 63.7754678, 61.4190072, 59.0952057, 56.806184, 54.554063, 52.3409635, 50.2435865, 48.1878557, 46.1742753, 44.2033492, 42.2755814, 40.391476, 38.551537, 36.7562685, 35.0061744, 33.3017589, 31.6435259, 30.0319794, 28.4676236, 26.9509624, 25.4824998, 24.06274, 22.7200683, 21.4822009, 20.2921384, 19.1485774, 18.0502143, 16.9957459, 15.9838688, 15.0132794, 14.0826744, 13.1907504, 12.336204, 11.5177318, 10.7340303, 9.98379609, 9.26572585, 8.57851612, 7.9208635, 7.15803653, 6.3582523, 5.59133159, 4.86030101, 4.16818712, 3.51801652, 2.91281579, 2.35561151, 1.84943027, 1.39729866, 1.00224326, 0.667290648, 0.395467419, 0.189800155, 0.0533154403, -0.0109601397, 0, 0.549973633, 1.07126124, 1.57735307, 2.08173933, 2.59791028, 3.13935613, 3.71956712, 4.35203349, 5.05024547, 5.82769328, 6.69786716, 7.67425735, 8.77035408, 9.99964757, 11.3756281, 12.9117858, 14.8040216, 17.0918844, 19.376269, 21.7097535, 24.1449163, 26.7343354, 29.5305893, 32.586256, 35.9539138, 39.6861409, 43.8355155, 48.4546159, 53.5960204, 59.4354775, 65.8572326, 72.8769126, 80.5101447, 88.3375129, 96.5921659, 105.623264, 115.576946, 126.599353, 138.836624, 152.4349, 168.105487, 185.257351, 203.926056, 224.147165, 245.956242, 269.118799, 294.007965, 320.726816, 349.378429, 380.065881, 412.182306, 447.40843, 487.321267, 532.608371, 582.141314, 635.607006, 692.692356, 752.671167, 815.737346, 882.535459, 953.153239, 1027.83506, 1107.29751, 1191.68767, 1280.18758, 1372.42225, 1467.655, 1565.41706, 1666.56588, 1771.20874, 1883.43224, 1998.7521, 2115.89814, 2233.60014, 2350.58793, 2471.18053, 2587.12121, 2695.74246, 2794.37678, 2880.35665, 2935.86314, 2983.48114, 3027.03664, 3080.64119, 3144.04939, 3212.73016, 3286.65239, 3365.78494, 3450.0967, 3539.55654, 3710.52084, 3841.15284, 3938.6458, 4010.19294, 4062.98753, 4104.22279, 4141.09198, 4223.58802, 4245.51848},
        {165.33743, 164.4517, 163.456297, 162.354625, 161.150087, 159.846087, 158.446028, 156.953314, 155.371349, 153.703536, 151.953278, 150.12398, 148.219045, 146.241877, 144.195879, 142.084455, 139.911008, 137.678942, 135.391661, 133.052568, 130.665068, 128.232562, 125.758456, 123.246153, 120.699056, 118.120569, 115.514095, 112.883039, 110.230804, 107.560793, 104.87641, 102.181059, 99.4781427, 96.7710656, 94.0872892, 91.4176819, 88.7528232, 86.0953358, 83.4478426, 80.8129666, 78.1933306, 75.5915574, 73.01027, 70.4520913, 67.919644, 65.4155511, 62.9424355, 60.50292, 58.0996275, 55.735181, 53.4122032, 51.2356004, 49.1034948, 47.0162921, 44.9743977, 42.9782173, 41.0281563, 39.1246205, 37.2680153, 35.4587464, 33.6972192, 31.9838395, 30.3190127, 28.7031444, 27.1366403, 25.6199058, 24.1533466, 22.7608268, 21.4656475, 20.2198325, 19.0222666, 17.8718349, 16.7674219, 15.7079127, 14.6921921, 13.7191449, 12.787656, 11.8966101, 11.0448922, 10.2313871, 9.4549797, 8.71455475, 8.00899714, 7.33719172, 6.57407194, 5.7831055, 5.02924988, 4.31541239, 3.64450035, 3.01942105, 2.44308181, 1.91838993, 1.44825272, 1.03557748, 0.683271531, 0.394242172, 0.171396712, 0.0176424574, -0.0641132843, -0.0709632059, 0, 1.04188664, 1.97722179, 2.82456792, 3.60248748, 4.32954293, 5.02429673, 5.70531133, 6.39114919, 7.10037277, 7.85154452, 8.66322691, 9.55398239, 10.5423734, 11.6469624, 12.8863119, 14.2789843, 15.9925107, 18.0667349, 20.1832663, 22.3925893, 24.7451882, 27.2915473, 30.082151, 33.1674835, 36.5980291, 40.4242722, 44.696697, 49.4657879, 54.7820292, 60.8779855, 67.5552981, 74.8098269, 82.6374323, 90.5092943, 98.6836133, 107.558504, 117.28723, 128.023057, 139.919247, 153.129066, 168.355311, 185.034463, 203.21227, 222.934477, 244.246833, 267.03034, 291.536675, 317.852769, 346.065554, 376.261963, 407.865234, 442.437174, 481.354668, 525.241274, 573.18252, 624.954476, 680.333214, 738.869532, 800.615975, 865.870838, 934.422999, 1006.79346, 1083.82928, 1165.76769, 1251.81262, 1341.64228, 1434.67083, 1530.50805, 1629.73193, 1732.37277, 1841.42394, 1953.49684, 2067.59614, 2182.72647, 2297.89247, 2420.7327, 2539.45943, 2650.91882, 2751.95706, 2839.42032, 2893.45771, 2938.74383, 2979.28072, 3029.78517, 3090.01798, 3155.27548, 3225.5417, 3300.80069, 3381.03645, 3466.23304, 3649.89174, 3777.23768, 3862.73788, 3920.85939, 3966.06922, 4012.83442, 4075.62202, 4178.78858, 4225.7811},
        {168.065254, 167.112236, 166.060574, 164.913097, 163.672631, 162.342003, 160.924041, 159.421572, 157.837424, 156.174422, 154.435395, 152.62317, 150.740574, 148.790433, 146.775576, 144.69883, 142.563021, 140.370977, 138.125525, 135.829492, 133.485705, 131.096992, 128.66618, 126.196096, 123.689567, 121.14942, 118.578483, 115.979582, 113.355546, 110.7092, 108.043373, 105.360891, 102.664582, 99.957273, 97.2333387, 94.5000102, 91.7646204, 89.0302706, 86.3000623, 83.577097, 80.8644762, 78.1653013, 75.4826739, 72.8196954, 70.1794673, 67.565091, 64.9796681, 62.4263001, 59.9080883, 57.4281343, 54.9895396, 52.7106421, 50.4768087, 48.2886428, 46.1467478, 44.051727, 42.0041838, 40.0047217, 38.0539439, 36.1524539, 34.300855, 32.4997507, 30.7497442, 29.051439, 27.4054384, 25.8123458, 24.2727647, 22.8142022, 21.4635197, 20.1658333, 18.9200023, 17.724886, 16.5793438, 15.4822349, 14.4324186, 13.4287542, 12.4701012, 11.5553186, 10.683266, 9.85280245, 9.06278743, 8.31208019, 7.59954005, 6.9240263, 6.16226484, 5.3767506, 4.63144645, 3.92917516, 3.27275951, 2.66502227, 2.1087862, 1.60687409, 1.16210869, 0.777312785, 0.455309143, 0.198920535, 0.010969733, -0.105720493, -0.148327371, -0.114028131, 0, 1.68080576, 3.14485648, 4.41801778, 5.52615523, 6.49513444, 7.350821, 8.11908051, 8.82577856, 9.49678075, 10.1579527, 10.8351599, 11.5542681, 12.3411428, 13.2216496, 14.2216541, 15.367022, 16.739924, 18.3742696, 20.1685668, 22.1607469, 24.3887407, 26.8904793, 29.7038938, 32.8669151, 36.4174743, 40.3935024, 44.8329305, 49.7736895, 55.2537106, 61.4189111, 68.1596407, 75.4814343, 83.389827, 91.5561371, 100.153008, 109.442208, 119.529536, 130.520794, 142.521782, 155.638298, 170.012151, 185.702176, 202.807126, 221.425759, 241.656828, 263.518999, 287.211141, 312.85203, 340.560446, 370.455164, 402.395697, 437.076972, 475.019808, 516.541295, 561.542574, 610.11154, 662.336085, 718.46045, 778.380648, 841.822134, 908.336722, 978.625003, 1053.54442, 1133.38255, 1217.4614, 1305.54618, 1397.42562, 1492.87105, 1591.56753, 1693.24891, 1798.27003, 1905.64824, 2014.90245, 2125.55157, 2237.1145, 2362.77801, 2484.97619, 2599.81099, 2703.38435, 2791.79822, 2841.44528, 2881.27625, 2915.83991, 2963.07136, 3022.22063, 3086.96009, 3156.81864, 3231.32517, 3310.0086, 3392.39781, 3553.24278, 3662.78686, 3737.01288, 3791.90367, 3843.44207, 3907.61091, 4000.39303, 4094.09392, 4169.33848},
        {170.892876, 169.894275, 168.806456, 167.631736, 166.372435, 165.03087, 163.609362, 162.110228, 160.535788, 158.888359, 157.170262, 155.383813, 153.531332, 151.615139, 149.63755, 147.600886, 145.507464, 143.359604, 141.159624, 138.909843, 136.612579, 134.270152, 131.884879, 129.45908, 126.995074, 124.495178, 121.961712, 119.396994, 116.803344, 114.183079, 111.538518, 108.871981, 106.185786, 103.48225, 100.729967, 97.9488264, 95.1591268, 92.3642812, 89.5677028, 86.7728049, 83.9830007, 81.2017033, 78.4323261, 75.6782823, 72.942985, 70.2298475, 67.542283, 64.8837048, 62.2575261, 59.66716, 57.1160199, 54.7173974, 52.3624453, 50.052195, 47.7876775, 45.5699243, 43.3999666, 41.2788357, 39.2075629, 37.1871794, 35.2187164, 33.3032054, 31.4416775, 29.635164, 27.8846962, 26.1913054, 24.5560228, 23.0161789, 21.6082326, 20.2583509, 18.9652118, 17.7274936, 16.5438744, 15.4130323, 14.3336455, 13.3043922, 12.3239504, 11.3909984, 10.5042142, 9.662276, 8.86386199, 8.10765028, 7.39231903, 6.71654638, 5.95399745, 5.16848639, 4.42532989, 3.727263, 3.07702079, 2.4773383, 1.93095061, 1.44059277, 1.00899983, 0.638906857, 0.333048909, 0.0941610438, -0.0750216808, -0.171764206, -0.193331473, -0.136988424, 0, 2.00042668, 3.72876037, 5.21452486, 6.48724391, 7.5764413, 8.5116408, 9.32236618, 10.0381412, 10.6884897, 11.3029354, 11.911002, 12.5422134, 13.2260934, 13.9921656, 14.8699539, 15.888982, 17.0728067, 18.4500992, 20.0638803, 21.9423952, 24.1138889, 26.6066066, 29.4487933, 32.6686943, 36.2945545, 40.354619, 44.8771332, 49.8903419, 55.4224903, 61.5446682, 68.2265664, 75.4835766, 83.3310905, 91.6055156, 100.411736, 109.902364, 120.146487, 131.213193, 143.171568, 156.090699, 169.69614, 184.505065, 200.653774, 218.278566, 237.515742, 258.423383, 281.235562, 306.108135, 333.196957, 362.657882, 394.79236, 429.432702, 466.288796, 505.380664, 547.361532, 592.60033, 641.465989, 694.793249, 752.379366, 813.513437, 877.608755, 945.484984, 1018.06883, 1095.68207, 1177.62143, 1263.65417, 1353.66463, 1447.45044, 1544.37976, 1644.06368, 1745.44789, 1848.91125, 1954.29518, 2061.44109, 2170.19042, 2297.30474, 2421.47526, 2538.31337, 2643.43045, 2732.43787, 2779.46235, 2815.92304, 2846.63904, 2891.2416, 2948.68669, 3011.75848, 3079.72153, 3151.84035, 3227.37947, 3305.60342, 3429.46287, 3519.59225, 3588.35699, 3648.1225, 3711.25421, 3790.11753, 3897.07788, 3979.23207, 4070.60884},
        {174.307038, 173.31096, 172.229365, 171.064295, 169.817792, 168.491897, 167.088654, 165.610104, 164.058288, 162.435249, 160.743029, 158.98367, 157.159213, 155.2717, 153.323174, 151.315677, 149.25125, 147.131935, 144.959774, 142.73681, 140.465084, 138.146638, 135.783514, 133.377754, 130.9314, 128.446493, 125.925077, 123.369193, 120.780882, 118.162187, 115.515149, 112.841812, 110.144215, 107.424403, 104.640199, 101.816686, 98.979473, 96.1320378, 93.2778572, 90.4204078, 87.5631666, 84.7096104, 81.863216, 79.0274603, 76.2058201, 73.4017723, 70.6187936, 67.860361, 65.1299512, 62.431041, 59.7671075, 57.2192093, 54.7115595, 52.2459531, 49.8241851, 47.4480505, 45.1193443, 42.8398615, 40.6113971, 38.4357461, 36.3147036, 34.2500645, 32.2436239, 30.2971767, 28.412518, 26.5914428, 24.8357461, 23.2033958, 21.7510111, 20.3645347, 19.04212, 17.7819207, 16.5820902, 15.4407822, 14.3561501, 13.3263475, 12.349528, 11.4238451, 10.5474524, 9.71850334, 8.93515154, 8.19555054, 7.49785388, 6.84021511, 6.0802912, 5.28943913, 4.54070456, 3.83680043, 3.18043966, 2.57433519, 2.02119995, 1.52374687, 1.08468888, 0.706738914, 0.392609901, 0.145014772, -0.03333354, -0.139722105, -0.17143799, -0.125768266, 0, 1.41877747, 2.67221333, 3.78273311, 4.77276233, 5.66472652, 6.48105121, 7.24416191, 7.97648417, 8.7004435, 9.43846542, 10.2129755, 11.0463992, 11.9611621, 12.9796896, 14.1244075, 15.417741, 16.9156967, 18.6454973, 20.5536117, 22.6696613, 25.0232675, 27.6440517, 30.5616354, 33.80564, 37.4056868, 41.3913974, 45.7923931, 50.6382954, 55.9587256, 61.8483392, 68.2478778, 75.1674526, 82.6171751, 90.3009328, 98.3819488, 107.105106, 116.572382, 126.885756, 138.147205, 150.458709, 163.937609, 178.665843, 194.742386, 212.266209, 231.336285, 251.766043, 274.011385, 298.24267, 324.630256, 353.344502, 384.84126, 418.656459, 454.275643, 491.6239, 531.523711, 574.394646, 620.656273, 671.215759, 725.89426, 783.98101, 844.844556, 909.4849, 979.353592, 1054.84329, 1134.55244, 1217.90305, 1303.75319, 1391.37867, 1482.12309, 1576.16041, 1675.74371, 1778.64792, 1884.32729, 1992.23607, 2101.8285, 2226.47157, 2348.2286, 2463.07563, 2566.98874, 2655.94396, 2706.9286, 2747.56665, 2781.97222, 2825.58522, 2878.03662, 2934.2383, 2994.05718, 3057.36019, 3124.01425, 3193.88627, 3265.85342, 3342.61769, 3422.69557, 3504.60359, 3586.85826, 3667.97608, 3746.47357, 3841.95824, 3925.52477},
        {178.020724, 177.110604, 176.109962, 175.020923, 173.845613, 172.586154, 171.244673, 169.823293, 168.32414, 166.749337, 165.101009, 163.381281, 161.592277, 159.736122, 157.81494, 155.830857, 153.785996, 151.682482, 149.52244, 147.307993, 145.041268, 142.724388, 140.359478, 137.948662, 135.494065, 132.997811, 130.462026, 127.888833, 125.280358, 122.638724, 119.966056, 117.26448, 114.536118, 111.783097, 108.972984, 106.125909, 103.262417, 100.385753, 97.4991637, 94.605895, 91.7091928, 88.812303, 85.9184716, 83.0309446, 80.1529679, 77.2877875, 74.4386494, 71.6087995, 68.8014838, 66.0199482, 63.2674388, 60.5783132, 57.9240312, 55.3071643, 52.7302841, 50.195962, 47.7067696, 45.2652785, 42.87406, 40.5356858, 38.2527274, 36.0277562, 33.8633439, 31.7620618, 29.7264817, 27.7591749, 25.862713, 24.1126262, 22.5826932, 21.1250135, 19.7374289, 18.4177815, 17.1639131, 15.9736656, 14.8448809, 13.7754009, 12.7630675, 11.8057227, 10.9012083, 10.0473662, 9.24203834, 8.48306661, 7.76829293, 7.09555918, 6.31909914, 5.51157939, 4.74739255, 4.02904089, 3.35902667, 2.73985215, 2.1740196, 1.66403127, 1.21238944, 0.821596359, 0.494154297, 0.232565515, 0.039332275, -0.0830431591, -0.132058525, -0.105211559, 0, 0.632057286, 1.23966898, 1.83574486, 2.43319473, 3.04492837, 3.68385557, 4.36288613, 5.09492983, 5.89289646, 6.76969581, 7.73823768, 8.81143185, 10.0021881, 11.3234163, 12.7880261, 14.4089273, 16.3239823, 18.5639512, 20.8591131, 23.2491534, 25.7737574, 28.4726105, 31.3853979, 34.5518049, 38.011517, 41.8042193, 45.9695973, 50.5473361, 55.5771213, 61.2338039, 67.3723426, 73.9918729, 81.0915303, 88.2448263, 95.6637086, 103.673624, 112.401396, 121.973847, 132.517799, 144.160076, 157.34147, 171.779278, 187.538893, 204.685709, 223.285118, 242.939315, 264.292692, 287.526445, 312.821767, 340.359852, 370.675109, 403.163808, 437.307913, 473.001612, 510.98072, 551.603179, 595.226931, 642.470735, 693.372441, 747.68537, 805.22571, 866.618079, 933.199135, 1005.24815, 1081.25272, 1160.52286, 1241.39123, 1322.91446, 1407.73272, 1496.45911, 1594.12667, 1696.2486, 1801.90805, 1910.18816, 2020.17208, 2141.67948, 2260.37284, 2372.65117, 2474.91348, 2563.55878, 2618.75841, 2663.95749, 2702.50975, 2745.48055, 2793.1711, 2842.66939, 2894.43727, 2948.93659, 3006.6292, 3067.97695, 3104.44393, 3171.81803, 3258.73169, 3353.81737, 3445.70751, 3523.03455, 3574.43094, 3678.69192, 3747.2901},
        {182.127512, 181.480733, 180.720494, 179.849725, 178.87136, 177.78833, 176.603568, 175.320005, 173.940573, 172.468204, 170.905831, 169.256385, 167.522799, 165.708004, 163.814932, 161.846516, 159.805687, 157.695377, 155.51852, 153.278045, 150.976886, 148.617975, 146.204243, 143.738622, 141.224045, 138.663444, 136.05975, 133.415896, 130.734814, 128.019435, 125.272691, 122.497516, 119.69684, 116.873595, 114.041618, 111.198159, 108.340338, 105.470733, 102.591923, 99.7064849, 96.8169976, 93.9260391, 91.0361876, 88.1500213, 85.2701183, 82.3990569, 79.5394152, 76.6937714, 73.8647036, 71.05479, 68.2666089, 65.4835293, 62.7277549, 60.0022802, 57.3100999, 54.6542084, 52.0376005, 49.4632706, 46.9342134, 44.4534235, 42.0238954, 39.6486237, 37.330603, 35.0728279, 32.878293, 30.7499929, 28.6909221, 26.774898, 25.0740374, 23.4452684, 21.8869943, 20.3976188, 18.9755451, 17.6191768, 16.3269172, 15.0971698, 13.928338, 12.8188253, 11.7670351, 10.7713709, 9.830236, 8.94203394, 8.10516813, 7.318042, 6.49161996, 5.66986412, 4.89778756, 4.17663127, 3.50763625, 2.89204351, 2.33109405, 1.82602886, 1.37808896, 0.988515334, 0.658548995, 0.389430941, 0.182402174, 0.0387036947, -0.040423495, -0.0537383938, 0, -0.416973565, -0.699241324, -0.844817974, -0.851718215, -0.717956747, -0.441548267, -0.0205074759, 0.547150928, 1.26341225, 2.13026178, 3.14968483, 4.32366669, 5.65419268, 7.14324808, 8.7928182, 10.6048883, 13.0257566, 16.1208815, 18.8872552, 21.4220728, 23.8225296, 26.1858206, 28.6091411, 31.1896863, 34.0246514, 37.2112315, 40.8466219, 45.0280176, 49.852614, 55.8952716, 62.6003762, 69.9218234, 77.8135086, 86.0507321, 94.6766872, 103.782192, 113.374722, 123.46175, 134.050751, 145.149198, 155.621378, 166.965879, 179.413844, 193.196415, 208.544733, 226.258484, 245.858128, 267.432673, 291.071124, 316.862486, 345.112805, 375.428776, 407.45681, 441.108119, 476.834548, 514.845211, 555.349218, 598.081893, 643.833816, 693.912433, 750.421635, 811.134717, 873.904853, 938.001258, 1005.09255, 1075.74598, 1151.15482, 1232.04862, 1316.86153, 1405.3261, 1497.1213, 1592.04151, 1689.83783, 1790.26136, 1893.0632, 2012.93231, 2130.94747, 2243.12529, 2345.48242, 2434.03548, 2487.39442, 2530.58702, 2567.08992, 2609.96196, 2658.88684, 2709.55566, 2761.85176, 2815.65845, 2870.85905, 2927.33688, 2931.84786, 2974.65034, 3044.94373, 3131.92744, 3224.80089, 3312.76347, 3385.0146, 3475.29526, 3540.73349},
        {189.757548, 189.459674, 189.006076, 188.401366, 187.650151, 186.757042, 185.726647, 184.563576, 183.272438, 181.857843, 180.324399, 178.676717, 176.919406, 175.057074, 173.094331, 171.035787, 168.886051, 166.649732, 164.33144, 161.935783, 159.467372, 156.930815, 154.330722, 151.671702, 148.958364, 146.195319, 143.387174, 140.53854, 137.654026, 134.738241, 131.795795, 128.831296, 125.849354, 122.854579, 119.959706, 117.113007, 114.261454, 111.406147, 108.548188, 105.688677, 102.828713, 99.9693979, 97.1118315, 94.2571144, 91.4063469, 88.5606296, 85.7210628, 82.888747, 80.0647825, 77.2502699, 74.4463094, 71.528939, 68.6270328, 65.7444024, 62.8848592, 60.0522149, 57.2502809, 54.4828688, 51.7537902, 49.0668565, 46.4258794, 43.8346703, 41.2970408, 38.8168025, 36.3977668, 34.0437454, 31.7585497, 29.6200537, 27.7043532, 25.8625115, 24.0935391, 22.3964461, 20.7702432, 19.2139405, 17.7265486, 16.3070777, 14.9545383, 13.6679407, 12.4462953, 11.2886124, 10.1939025, 9.16117596, 8.18944308, 7.27771427, 6.37872466, 5.51559559, 4.71200458, 3.96846372, 3.28548512, 2.66358086, 2.10326304, 1.60504376, 1.16943511, 0.796949179, 0.488098073, 0.243393883, 0.0633487041, -0.0515253688, -0.10071624, -0.0837118159, 0, 0.0844557096, 0.225222091, 0.426883157, 0.694022921, 1.0312254, 1.4430746, 1.93415454, 2.50904923, 3.17234268, 3.92861892, 4.78246194, 5.73845577, 6.80118442, 7.9752319, 9.26518223, 10.6756194, 12.5790489, 15.0326149, 17.2086982, 19.1907232, 21.062114, 22.9062951, 24.8066908, 26.8467254, 29.1098232, 31.6794085, 34.6389057, 38.0717391, 42.061333, 47.2104446, 52.8927433, 59.0358535, 65.5673997, 71.6041569, 77.4791743, 83.7417276, 90.562696, 98.1129587, 106.563395, 116.084885, 128.207744, 141.329673, 155.355574, 170.190349, 185.7389, 200.008377, 215.275873, 231.920728, 250.322282, 270.859876, 294.772201, 320.528928, 348.287695, 377.893962, 408.555834, 439.767575, 471.023449, 498.175203, 525.187465, 559.998522, 612.679844, 672.54889, 729.876959, 781.098668, 833.092034, 887.94138, 949.56502, 1020.52276, 1096.64977, 1177.58508, 1264.81489, 1355.84682, 1449.68046, 1545.3154, 1641.75123, 1743.82521, 1843.23985, 1937.5353, 2024.25173, 2100.92932, 2153.98722, 2199.50063, 2239.77586, 2281.2515, 2324.59818, 2368.76477, 2414.50798, 2462.58455, 2513.75119, 2568.76463, 2872.92415, 3008.58208, 3028.0857, 2983.78231, 2928.01921, 2913.14369, 2991.50305, 3060.40921, 3117.90423},
        {192.62829, 192.393492, 192.011292, 191.485798, 190.82112, 190.021367, 189.090647, 188.03307, 186.852743, 185.553777, 184.14028, 182.61636, 180.986128, 179.253691, 177.423159, 175.49864, 173.484244, 171.384079, 169.202255, 166.94288, 164.610063, 162.207912, 159.740538, 157.212049, 154.626553, 151.98816, 149.300978, 146.569117, 143.796686, 140.987792, 138.146546, 135.277055, 132.38343, 129.469779, 126.620736, 123.798453, 120.964116, 118.11922, 115.265261, 112.403734, 109.536135, 106.663959, 103.788704, 100.911863, 98.0349329, 95.1594095, 92.2867884, 89.4185652, 86.5562356, 83.7012953, 80.8552401, 77.9421671, 75.0426483, 72.1598574, 69.2969678, 66.4571532, 63.6435869, 60.8594426, 58.1078939, 55.3921143, 52.7152772, 50.0805563, 47.4911252, 44.9501573, 42.4608261, 40.0263054, 37.6497685, 35.3810604, 33.2689057, 31.2202215, 29.2351557, 27.3138566, 25.4564719, 23.6631499, 21.9340385, 20.2692858, 18.6690397, 17.1334483, 15.6626596, 14.2568217, 12.9160826, 11.6405902, 10.4304927, 9.28593797, 8.18005321, 7.12706521, 6.14152561, 5.2244593, 4.3768912, 3.59984621, 2.89434923, 2.26142519, 1.70209897, 1.21739549, 0.808339658, 0.475956378, 0.221270557, 0.0453071018, -0.0509090807, -0.0663530836, 0, 0.326370212, 0.692641171, 1.10419982, 1.56643309, 2.08472794, 2.66447129, 3.31105009, 4.02985129, 4.82626181, 5.7056686, 6.6734586, 7.73501875, 8.895736, 10.1609973, 11.5361895, 13.0266997, 14.816274, 16.9357795, 18.987171, 21.0140555, 23.0600397, 25.1687304, 27.3837344, 29.7486586, 32.3071097, 35.1026946, 38.17902, 41.5796927, 45.3483196, 49.7155382, 54.4693466, 59.5972424, 65.0867231, 70.4489823, 75.9096697, 81.8214491, 88.3147092, 95.5198387, 103.567226, 112.587261, 123.82027, 135.948895, 148.886365, 162.545905, 176.840742, 189.962054, 203.97563, 219.225208, 236.054529, 254.807331, 276.548366, 300.019124, 325.131253, 351.753215, 379.665301, 408.687392, 438.639363, 467.838247, 497.948326, 532.273353, 574.505435, 621.319441, 670.051543, 719.752227, 772.376783, 828.547147, 889.297177, 955.355602, 1025.94078, 1101.12542, 1182.97539, 1269.2636, 1359.37282, 1452.68582, 1548.58537, 1657.83139, 1765.58522, 1868.38534, 1962.77022, 2045.27836, 2098.24991, 2141.88722, 2178.81377, 2216.72657, 2256.24081, 2295.85772, 2336.29818, 2378.28309, 2422.53333, 2469.76981, 2690.8768, 2795.04125, 2819.27292, 2800.58154, 2775.97687, 2782.46864, 2857.06661, 2895.45241, 2932.92104},
        {202.644418, 202.376182, 201.976711, 201.449209, 200.796884, 200.022942, 199.130588, 198.12303, 197.003474, 195.775126, 194.441192, 193.004879, 191.469393, 189.83794, 188.113726, 186.299959, 184.399844, 182.416587, 180.353396, 178.213475, 176.000032, 173.716273, 171.365405, 168.950632, 166.475163, 163.942202, 161.354957, 158.716634, 156.030439, 153.299578, 150.527258, 147.716685, 144.871066, 141.993606, 139.10426, 136.197508, 133.267629, 130.317285, 127.34914, 124.365857, 121.370097, 118.364524, 115.351801, 112.33459, 109.315555, 106.297358, 103.282662, 100.274129, 97.2744233, 94.2862069, 91.3121426, 88.4078896, 85.5219655, 82.6558844, 79.8111602, 76.9893069, 74.1918385, 71.4202688, 68.6761119, 65.9608817, 63.2760922, 60.6232574, 58.0038912, 55.4195075, 52.8716204, 50.3617438, 47.8913916, 45.4561674, 43.0518167, 40.6920432, 38.3787439, 36.1138162, 33.8991569, 31.7366634, 29.6282326, 27.5757616, 25.5811477, 23.6462879, 21.7730793, 19.963419, 18.2192041, 16.5423318, 14.9346992, 13.3982033, 11.9983501, 10.7038945, 9.48282654, 8.33497921, 7.26018537, 6.25827789, 5.32908966, 4.47245356, 3.68820247, 2.97616928, 2.33618686, 1.7680881, 1.27170587, 0.846873071, 0.49342257, 0.211187252, 0, -0.579862167, -1.01088636, -1.29774187, -1.44509799, -1.45762402, -1.33998925, -1.09686296, -0.732914458, -0.25281303, 0.338772029, 1.03717143, 1.83771587, 2.73573607, 3.72656273, 4.80552656, 5.96795826, 7.62667379, 9.83664461, 11.6488896, 13.1482005, 14.4193691, 15.5471875, 16.6164472, 17.7119403, 18.9184584, 20.3207935, 22.0037374, 24.0520819, 26.5506189, 30.1930372, 34.2319693, 38.569538, 43.1078659, 47.0181711, 50.5680287, 54.2199215, 58.0952436, 62.3153891, 67.0017519, 72.2757263, 78.9159639, 86.1865662, 94.0803334, 102.590066, 111.708563, 121.427794, 131.741599, 142.642985, 154.124961, 166.180535, 178.559986, 191.795721, 206.397605, 222.559914, 239.832596, 258.054885, 277.06602, 295.36308, 314.432494, 337.224864, 367.115643, 400.968142, 436.44116, 472.586106, 510.814086, 551.453068, 594.964731, 641.71171, 691.566352, 744.678336, 801.200386, 861.278673, 925.061824, 992.698467, 1064.33723, 1153.25799, 1243.19531, 1331.01501, 1413.5829, 1487.76481, 1538.73232, 1582.84162, 1621.97035, 1661.93351, 1703.05018, 1743.99886, 1785.18066, 1826.99666, 1869.84799, 1914.13575, 2028.06635, 2098.77793, 2138.22659, 2158.36842, 2171.15951, 2188.55595, 2222.51385, 2278.80295, 2329.96594}
    };
    
    const double a4tab[98][200] = {{-1766.7572, -1743.88214, -1721.39625, -1699.28611, -1677.5383, -1656.13939, -1635.07595, -1614.33455, -1593.90178, -1573.76419, -1553.90837, -1534.32088, -1514.98831, -1495.89722, -1477.03418, -1458.38578, -1439.93858, -1421.67915, -1403.59408, -1385.66992, -1367.89327, -1350.25068, -1332.72874, -1315.31401, -1297.99306, -1280.75248, -1263.57884, -1246.4587, -1229.37865, -1212.32525, -1195.28507, -1178.2447, -1161.1907, -1144.10965, -1126.34727, -1108.22403, -1090.06812, -1071.88691, -1053.68777, -1035.47807, -1017.26517, -999.056457, -980.859287, -962.681036, -944.529073, -926.410769, -908.333493, -890.304614, -872.331505, -854.421533, -836.58207, -819.33603, -802.164063, -785.062362, -768.027122, -751.054536, -734.140798, -717.282103, -700.474645, -683.714617, -666.998214, -650.32163, -633.681058, -617.072693, -600.492729, -583.937359, -567.402778, -550.853265, -534.253866, -517.666596, -501.089718, -484.521496, -467.960192, -451.404069, -434.85139, -418.300418, -401.749416, -385.196648, -368.640374, -352.07886, -335.510368, -318.93316, -302.345499, -285.74565, -269.164874, -252.584242, -235.984424, -219.362613, -202.716, -186.041777, -169.337137, -152.599271, -135.825372, -119.01263, -102.158239, -85.2593904, -68.3132759, -51.3170876, -34.2680175, -17.1632577, 0, 17.7523869, 35.2525046, 52.5540932, 69.7108928, 86.7766435, 103.805085, 120.849959, 137.965003, 155.20396, 172.620568, 190.268567, 208.201699, 226.473703, 245.138319, 264.249287, 283.860347, 304.170146, 325.253125, 346.835259, 369.001341, 391.836162, 415.424512, 439.851185, 465.20097, 491.55866, 519.009045, 547.636918, 577.52707, 608.764291, 641.151332, 675.158442, 710.955023, 748.71048, 788.557488, 830.683816, 875.287025, 922.547539, 972.64578, 1025.76217, 1082.07713, 1140.18709, 1202.33855, 1269.02185, 1340.72733, 1417.94532, 1502.17427, 1592.64437, 1689.59395, 1793.2613, 1903.88475, 2019.71468, 2145.40699, 2281.1348, 2428.30718, 2590.85671, 2771.58293, 2973.28543, 3207.08409, 3465.56718, 3732.24629, 3983.27192, 4250.02202, 4558.81443, 4922.31725, 5330.06992, 5782.22832, 6269.95128, 6791.0621, 7376.37331, 8038.03687, 8817.96206, 9693.96547, 10667.8984, 11741.6123, 12916.9584, 14558.7763, 16215.182, 17797.28, 19216.1743, 20382.9695, 21022.3307, 21356.0938, 21375.2656, 20861.2923, 19888.1316, 18617.0581, 17117.6637, 15459.5402, 13712.2795, 11945.4736, 10324.3132, 8855.63002, 7528.37765, 6331.50979, 5253.98009, 4284.74222, 3412.74983, 2660.64522, 1938.41116},
        {-1661.23266, -1638.7405, -1616.67986, -1595.03721, -1573.79903, -1552.95179, -1532.48196, -1512.37602, -1492.62043, -1473.20168, -1454.10623, -1435.32056, -1416.83114, -1398.62444, -1380.68694, -1363.00512, -1345.56543, -1328.35437, -1311.35839, -1294.56397, -1277.9576, -1261.52573, -1245.25485, -1229.13142, -1213.14192, -1197.27282, -1181.5106, -1165.84173, -1150.25268, -1134.72992, -1119.25994, -1103.82919, -1088.42416, -1073.03132, -1057.21956, -1041.19293, -1025.16049, -1009.12227, -993.078286, -977.028571, -960.973148, -944.912039, -928.84527, -912.772865, -896.694848, -880.611244, -864.522078, -848.427373, -832.327155, -816.221447, -800.110274, -783.916358, -767.718702, -751.519006, -735.31897, -719.120294, -702.924679, -686.733825, -670.549431, -654.373199, -638.206828, -622.052018, -605.91047, -589.783884, -573.673961, -557.582399, -541.510901, -525.524995, -509.688679, -493.872009, -478.072548, -462.287857, -446.5155, -430.753038, -414.998034, -399.248051, -383.50065, -367.753394, -352.003846, -336.249567, -320.488121, -304.717069, -288.933974, -273.136398, -257.308023, -241.453643, -225.578221, -209.679768, -193.756298, -177.805824, -161.826359, -145.815915, -129.772505, -113.694142, -97.5788394, -81.4246094, -65.2294649, -48.9914189, -32.7084842, -16.3786736, 0, 17.1411341, 33.9841607, 50.5842693, 66.9966496, 83.2764912, 99.4789839, 115.659317, 131.872681, 148.174264, 164.619258, 181.26285, 198.160231, 215.366591, 232.937119, 250.927005, 269.391438, 288.43946, 308.133952, 328.408299, 349.32923, 370.963473, 393.377758, 416.638814, 440.813371, 465.968157, 492.169901, 519.485334, 547.981183, 577.724179, 608.398175, 640.593163, 674.490735, 710.272484, 748.478772, 789.111807, 832.078122, 877.451675, 925.306428, 975.71634, 1028.75537, 1081.91096, 1138.63079, 1199.49489, 1265.08326, 1335.97593, 1415.34393, 1500.52851, 1591.46194, 1688.07649, 1790.30442, 1895.74097, 2009.5118, 2133.23504, 2268.41518, 2416.32464, 2578.34003, 2755.83795, 2954.49285, 3170.40672, 3394.99299, 3615.90306, 3849.1892, 4108.90342, 4401.4217, 4720.11329, 5064.31778, 5427.36975, 5807.05194, 6223.16521, 6683.05571, 7251.46783, 7869.51899, 8524.68674, 9204.44859, 9896.2821, 10490.5233, 11096.0766, 11724.7049, 12388.1711, 13098.2381, 14079.9282, 14989.572, 15747.5356, 15895.5743, 15503.921, 14800.5627, 13847.8443, 12708.1107, 11443.7069, 10116.9779, 8766.27563, 7575.43928, 6521.30907, 5580.72519, 4730.52784, 3947.55719, 3208.65344, 2608.83865, 2085.37415},
        {-1563.20972, -1544.92192, -1526.71237, -1508.58197, -1490.53161, -1472.56218, -1454.67457, -1436.86967, -1419.14838, -1401.51157, -1383.96015, -1366.49501, -1349.11702, -1331.8271, -1314.62612, -1297.51497, -1280.49455, -1263.56575, -1246.72946, -1229.98657, -1213.33797, -1196.78455, -1180.32719, -1163.96681, -1147.70427, -1131.54047, -1115.47631, -1099.51268, -1083.65046, -1067.89054, -1052.23382, -1036.68119, -1021.23354, -1005.89175, -990.989909, -976.3553, -961.81121, -947.347718, -932.954902, -918.62284, -904.34161, -890.10129, -875.891958, -861.703693, -847.526573, -833.350676, -819.16608, -804.962863, -790.731103, -776.460879, -762.142269, -747.170304, -732.143009, -717.063361, -701.934337, -686.758916, -671.540075, -656.280792, -640.984043, -625.652808, -610.290063, -594.898786, -579.481955, -564.042547, -548.58354, -533.107912, -517.618639, -502.132447, -486.665728, -471.19311, -455.716679, -440.238522, -424.760725, -409.285375, -393.814559, -378.350363, -362.894873, -347.450176, -332.018359, -316.601507, -301.201709, -285.821049, -270.461615, -255.125492, -240.053629, -225.123656, -210.210336, -195.308004, -180.410994, -165.513642, -150.610283, -135.695251, -120.762882, -105.807511, -90.8234732, -75.8051032, -60.7467363, -45.6427075, -30.4873518, -15.2750043, 0, 16.2719273, 32.267669, 48.0402684, 63.6427688, 79.1282134, 94.5496455, 109.960108, 125.412645, 140.9603, 156.656115, 172.553134, 188.7044, 205.162956, 221.981846, 239.214114, 256.912801, 275.172386, 294.051829, 313.510457, 333.610191, 354.412953, 375.980664, 398.375247, 421.658624, 445.892715, 471.139444, 497.460732, 524.918501, 553.574673, 583.200456, 614.25508, 646.887681, 681.247395, 717.85821, 756.681835, 797.58002, 840.589444, 885.746788, 933.088731, 982.651954, 1031.82836, 1084.10433, 1140.03402, 1200.17155, 1265.07106, 1334.58505, 1410.14471, 1492.47958, 1582.31921, 1680.39316, 1789.44067, 1905.72527, 2027.23671, 2153.20916, 2285.41754, 2424.49604, 2571.07882, 2722.92051, 2884.18929, 3062.19463, 3268.69027, 3490.35549, 3715.00781, 3937.92657, 4167.03472, 4404.4516, 4654.5967, 4920.1857, 5195.50028, 5479.59286, 5797.27527, 6117.87755, 6431.53541, 6728.38456, 6998.5607, 6967.94615, 6956.99337, 7021.90142, 7218.86936, 7604.09625, 8601.32294, 9654.17884, 10661.3451, 11130.3337, 11114.6637, 10830.841, 10324.2353, 9640.2161, 8824.15297, 7921.41549, 6850.2941, 5920.46824, 5103.61631, 4371.4167, 3695.54782, 3047.68805, 2399.51579, 1847.03125, 1384.21255},
        {-1464.60418, -1448.43476, -1432.38877, -1416.46185, -1400.64961, -1384.94768, -1369.35167, -1353.85721, -1338.45992, -1323.15542, -1307.93933, -1292.80727, -1277.75487, -1262.77775, -1247.87152, -1233.03181, -1218.25425, -1203.53444, -1188.86802, -1174.25061, -1159.67782, -1145.14528, -1130.6486, -1116.18342, -1101.74536, -1087.33002, -1072.93304, -1058.55004, -1044.17664, -1029.80845, -1015.44111, -1001.07023, -986.691431, -972.30034, -957.674638, -942.9235, -928.158719, -913.382989, -898.599005, -883.80946, -869.017048, -854.224463, -839.434399, -824.64955, -809.87261, -795.106273, -780.353232, -765.616183, -750.897818, -736.200832, -721.527918, -707.137984, -692.771956, -678.426974, -664.100178, -649.788707, -635.489702, -621.200302, -606.917647, -592.638876, -578.361129, -564.081547, -549.797268, -535.505433, -521.203181, -506.887652, -492.555986, -478.122874, -463.504986, -448.869507, -434.218921, -419.555712, -404.882366, -390.201366, -375.515198, -360.826345, -346.137293, -331.450526, -316.768528, -302.093784, -287.428779, -272.775997, -258.137923, -243.51704, -229.158448, -214.93822, -200.729515, -186.526946, -172.325124, -158.118659, -143.902163, -129.670248, -115.417525, -101.138605, -86.8280987, -72.4806186, -58.0907755, -43.6531808, -29.1624458, -14.6131817, 0, 15.5318349, 30.810716, 45.8868728, 60.8105346, 75.6319311, 90.4012916, 105.168846, 119.984823, 134.899452, 149.962964, 165.225586, 180.73755, 196.549084, 212.710418, 229.271781, 246.283403, 263.837935, 281.991667, 300.698873, 320.018875, 340.010992, 360.734543, 382.248849, 404.61323, 427.887004, 452.129493, 477.400016, 503.757892, 531.262442, 559.821323, 589.701127, 621.006673, 653.842779, 687.954817, 723.627328, 761.24071, 801.007615, 843.140697, 887.852609, 935.356005, 986.279954, 1040.29396, 1097.5292, 1158.11685, 1222.18811, 1286.13041, 1354.75461, 1429.12781, 1510.31711, 1599.38964, 1700.02391, 1807.48389, 1921.62221, 2041.82354, 2166.51706, 2294.56095, 2424.81338, 2549.32758, 2675.31324, 2817.40367, 2994.0075, 3185.42035, 3379.06573, 3567.8206, 3753.80238, 3934.88694, 4112.40587, 4285.13097, 4439.16321, 4567.77089, 4679.14815, 4749.3412, 4766.45174, 4718.58146, 4593.83204, 3879.31657, 3189.37249, 2637.34864, 2336.59386, 2400.45697, 3537.10926, 4868.52885, 6253.14065, 7062.09047, 7346.68457, 7361.2622, 7146.97798, 6744.98653, 6196.44248, 5542.50044, 4883.2411, 4189.44177, 3484.24529, 2790.7945, 2132.23222, 1531.70128, 1012.34452, 517.616728, 86.9457832},
        {-1373.5761, -1358.92541, -1344.44171, -1330.11702, -1315.94331, -1301.91258, -1288.0168, -1274.24798, -1260.59809, -1247.05913, -1233.62308, -1220.28194, -1207.02768, -1193.8523, -1180.74779, -1167.70613, -1154.71931, -1141.77932, -1128.87815, -1116.00779, -1103.16022, -1090.32744, -1077.50142, -1064.67416, -1051.83765, -1038.98387, -1026.10482, -1013.19247, -1000.23883, -987.235871, -974.175587, -961.049966, -947.850996, -934.570665, -920.701979, -906.496913, -892.213438, -877.859735, -863.443985, -848.97437, -834.45907, -819.906267, -805.32414, -790.720872, -776.104644, -761.483635, -746.866028, -732.260002, -717.673741, -703.115423, -688.59323, -674.6181, -660.684558, -646.789886, -632.931368, -619.106284, -605.311918, -591.545551, -577.804465, -564.085944, -550.387269, -536.705722, -523.038586, -509.383143, -495.736674, -482.096463, -468.459791, -454.782338, -441.020781, -427.258205, -413.494591, -399.729918, -385.964164, -372.197309, -358.429332, -344.660213, -330.88993, -317.118463, -303.34579, -289.571892, -275.796747, -262.020334, -248.242633, -234.463623, -220.794086, -207.17627, -193.551089, -179.914927, -166.264167, -152.595193, -138.904389, -125.188136, -111.44282, -97.6648238, -83.8505301, -69.9963227, -56.0985851, -42.1537005, -28.1580526, -14.1080246, 0, 14.6591022, 29.1192352, 43.4252753, 57.6220988, 71.7545819, 85.867601, 100.006032, 114.214752, 128.538637, 143.022562, 157.711406, 172.650042, 187.883349, 203.456202, 219.413477, 235.800052, 252.805173, 270.494343, 288.58588, 307.1556, 326.279314, 346.032836, 366.491979, 387.732556, 409.83038, 432.861266, 456.901024, 482.02547, 508.310416, 535.896513, 564.770962, 594.990126, 626.610366, 658.843736, 692.168752, 727.289078, 764.51437, 804.154281, 846.518467, 891.916581, 942.708436, 996.529567, 1053.28851, 1112.8938, 1175.25398, 1238.05564, 1303.98475, 1373.50531, 1447.08136, 1525.17692, 1608.02323, 1696.60159, 1793.45577, 1899.41881, 2011.83103, 2129.60088, 2251.63686, 2375.88849, 2502.44112, 2632.42626, 2765.97021, 2902.6596, 3050.21477, 3207.67025, 3359.34299, 3496.30555, 3610.79771, 3694.19467, 3733.59176, 3718.50524, 3683.17643, 3565.51575, 3339.55773, 2979.33686, 2458.88767, 837.126826, -768.013858, -2153.72041, -3117.17885, -3455.57522, -1995.4771, -151.76791, 1862.38701, 3133.35858, 3725.24833, 3994.01771, 3989.17525, 3760.22951, 3356.68905, 2828.0624, 2432.47659, 1903.42702, 1293.50769, 655.312561, 41.4356171, -495.529164, -902.987803, -1319.43779, -1679.97424},
        {-1295.44714, -1284.41597, -1273.15844, -1261.68473, -1250.00509, -1238.1297, -1226.06879, -1213.83257, -1201.43126, -1188.87506, -1176.17419, -1163.33887, -1150.3793, -1137.3057, -1124.12828, -1110.85726, -1097.50285, -1084.07526, -1070.5847, -1057.04139, -1043.45554, -1029.83737, -1016.19708, -1002.54489, -988.891019, -975.245672, -961.619064, -948.02141, -934.462921, -920.953812, -907.504294, -894.124582, -880.824888, -867.615426, -855.219415, -843.275568, -831.414028, -819.621869, -807.886167, -796.193995, -784.532429, -772.888543, -761.249411, -749.60211, -737.933712, -726.231293, -714.481928, -702.67269, -690.790656, -678.822899, -666.756494, -653.618754, -640.377321, -627.040076, -613.6149, -600.109671, -586.532272, -572.890582, -559.192482, -545.445852, -531.658574, -517.838526, -503.99359, -490.131645, -476.260574, -462.388255, -448.52257, -434.79569, -421.336803, -407.897447, -394.477446, -381.076623, -367.694801, -354.331803, -340.987452, -327.661571, -314.353984, -301.064513, -287.792982, -274.539213, -261.30303, -248.084257, -234.882715, -221.698228, -208.678914, -195.747328, -182.824246, -169.904678, -156.983636, -144.05613, -131.117171, -118.161769, -105.184935, -92.1816795, -79.1470139, -66.0759486, -52.9634941, -39.8046613, -26.5944609, -13.3279036, 0, 13.5557825, 27.0032363, 40.3798799, 53.7232315, 67.0708096, 80.4601323, 93.9287181, 107.514085, 121.253752, 135.185237, 149.346058, 163.773734, 178.505783, 193.579723, 209.033073, 224.903351, 241.548007, 259.050265, 276.723987, 294.675247, 313.010121, 331.834684, 351.255012, 371.377179, 392.307262, 414.151334, 437.015472, 461.005751, 486.228246, 513.076555, 541.263805, 570.809816, 601.734407, 633.639827, 666.754678, 701.418918, 737.777634, 775.975919, 816.15886, 858.471548, 902.686044, 949.433997, 998.933479, 1051.40257, 1107.05933, 1168.44157, 1232.86771, 1299.97588, 1369.40422, 1440.79087, 1511.8805, 1586.51897, 1665.06255, 1748.35123, 1838.21261, 1936.03087, 2043.19022, 2169.82051, 2306.5726, 2434.55661, 2535.25995, 2622.71755, 2712.54276, 2807.32953, 2887.61111, 2944.04695, 2972.13095, 2963.77593, 2893.16851, 2744.52226, 2590.75465, 2313.72858, 1866.95244, 1203.93459, 278.183412, -2358.88389, -5001.14499, -7344.56873, -9085.12393, -9918.77943, -8156.43127, -5802.50296, -3146.56595, -1403.12397, -495.629479, 37.8535413, 254.603223, 211.897701, -32.9848863, -422.766404, -850.271746, -1281.36303, -1710.46732, -2132.01168, -2540.42318, -2930.12889, -3295.55586, -3626.95729, -3894.40339},
        {-1218.35287, -1210.16745, -1201.53457, -1192.4737, -1183.00435, -1173.146, -1162.91813, -1152.34024, -1141.43181, -1130.21232, -1118.70128, -1106.91816, -1094.88245, -1082.61365, -1070.13123, -1057.45469, -1044.60351, -1031.59719, -1018.45521, -1005.19705, -991.842209, -978.410172, -964.920426, -951.392459, -937.84576, -924.299814, -910.774112, -897.28814, -883.861386, -870.513338, -857.263484, -844.131312, -831.136309, -818.297963, -806.898207, -796.298752, -785.846136, -775.518879, -765.2955, -755.154517, -745.074451, -735.03382, -725.011143, -714.984939, -704.933727, -694.836027, -684.670357, -674.415238, -664.049186, -653.550723, -642.898366, -630.559601, -618.056737, -605.401049, -592.603814, -579.676306, -566.629801, -553.475575, -540.224902, -526.889058, -513.479318, -500.006958, -486.483253, -472.919479, -459.32691, -445.716823, -432.100492, -418.642254, -405.502771, -392.37764, -379.268215, -366.175849, -353.101895, -340.047707, -327.014637, -314.004039, -301.017264, -288.055668, -275.120602, -262.21342, -249.335474, -236.488118, -223.672705, -210.890588, -198.378045, -186.014026, -173.674655, -161.353662, -149.044777, -136.741728, -124.438246, -112.128059, -99.8048962, -87.4624877, -75.0945625, -62.6948498, -50.2570789, -37.7749792, -25.24228, -12.6527105, 0, 12.8555949, 25.6285243, 38.3539629, 51.0670856, 63.8030671, 76.5970822, 89.4843057, 102.499912, 115.679077, 129.056974, 142.668778, 156.549665, 170.734809, 185.259384, 200.158566, 215.46753, 231.522454, 248.401517, 265.459045, 282.794715, 300.508203, 318.699185, 337.467336, 356.912334, 377.133854, 398.231571, 420.305162, 443.454302, 467.778669, 493.57251, 520.669586, 549.111199, 578.938654, 610.23792, 643.027969, 677.31586, 713.129497, 750.496784, 789.445625, 830.003923, 870.494902, 913.169962, 958.390534, 1006.51804, 1057.91392, 1116.75715, 1178.63722, 1242.96117, 1309.13603, 1376.56884, 1442.9313, 1511.48678, 1582.60717, 1656.83507, 1735.06165, 1818.02162, 1906.44964, 2006.22674, 2111.77164, 2211.88887, 2302.07754, 2376.64685, 2429.06676, 2454.13579, 2448.90372, 2409.38699, 2345.99261, 2258.46789, 2093.79465, 1828.80183, 1573.53507, 1151.1118, 492.247518, -472.342266, -1811.94205, -5492.51086, -9212.49003, -12566.9954, -15151.1429, -16560.0483, -14607.8102, -11857.9066, -8668.74652, -6514.83888, -5312.80306, -4514.21671, -4058.95128, -3886.8782, -3937.86889, -4151.7948, -4571.11048, -4886.63578, -5133.41904, -5346.50857, -5560.95271, -5811.79979, -6134.09814, -6313.91537, -6514.56154},
        {-1139.48068, -1132.48133, -1125.15509, -1117.5134, -1109.56773, -1101.32952, -1092.81021, -1084.02127, -1074.97413, -1065.68025, -1056.15108, -1046.39807, -1036.43267, -1026.26632, -1015.91048, -1005.37659, -994.676117, -983.820496, -972.821182, -961.689623, -950.437269, -939.075571, -927.615978, -916.06994, -904.448906, -892.764326, -881.02765, -869.250328, -857.443809, -845.619543, -833.78898, -821.96357, -810.154762, -798.374006, -787.218591, -776.394724, -765.601572, -754.831573, -744.077167, -733.33079, -722.584881, -711.831877, -701.064219, -690.274342, -679.454686, -668.597689, -657.695788, -646.741423, -635.72703, -624.645049, -613.487917, -601.731899, -589.896796, -577.986236, -566.003848, -553.953258, -541.838094, -529.661985, -517.428558, -505.14144, -492.804261, -480.420646, -467.994225, -455.528625, -443.027474, -430.494399, -417.933029, -405.346701, -392.738759, -380.113431, -367.474361, -354.825198, -342.169588, -329.511176, -316.85361, -304.200537, -291.555603, -278.922454, -266.304737, -253.706099, -241.130186, -228.580646, -216.061123, -203.575266, -191.432956, -179.478281, -167.551648, -155.646765, -143.757341, -131.877084, -119.999704, -108.118909, -96.2284072, -84.3219077, -72.393119, -60.4357499, -48.4435089, -36.4101048, -24.3292462, -12.1946417, 0, 12.6973806, 25.2507453, 37.6993017, 50.0822573, 62.4388199, 74.8081968, 87.2295958, 99.7422243, 112.38529, 125.198001, 138.219563, 151.489186, 165.046077, 178.929442, 193.17849, 207.832429, 222.989285, 238.696671, 254.860747, 271.533325, 288.766216, 306.611234, 325.12019, 344.344896, 364.337164, 385.148806, 406.831634, 429.437459, 453.018095, 477.428513, 502.989539, 529.812037, 558.00687, 587.859187, 619.392709, 652.584681, 687.493681, 724.178287, 762.697078, 803.108631, 845.09672, 889.208798, 935.576775, 984.33256, 1035.60806, 1090.18593, 1147.38464, 1207.17344, 1269.52154, 1334.39817, 1402.04079, 1471.82255, 1547.3366, 1628.79587, 1709.51198, 1785.8951, 1854.3554, 1899.25784, 1931.79535, 1976.30106, 2072.38635, 2165.18524, 2203.21713, 2164.10206, 2074.76945, 1939.51491, 1789.65912, 1636.50418, 1392.26025, 1025.18947, 671.650118, 105.947305, -761.843927, -2021.64853, -3763.39146, -8397.02758, -13112.4445, -17419.5596, -20828.2904, -22848.5544, -20897.1669, -17972.549, -14481.6621, -12065.6907, -10643.0495, -9617.89324, -8934.34953, -8536.54592, -8368.60994, -8374.66916, -8675.3714, -8846.31124, -8935.36878, -8990.42414, -9059.35742, -9190.04872, -9430.37815, -9386.52823, -9530.42985},
        {-1068.67232, -1062.9308, -1056.89941, -1050.58642, -1044.0001, -1037.14871, -1030.0405, -1022.68375, -1015.08672, -1007.25767, -999.20486, -990.936557, -982.461024, -973.786525, -964.921322, -955.873679, -946.651858, -937.264124, -927.718739, -918.023967, -908.188071, -898.219314, -888.125959, -877.91627, -867.598509, -857.180941, -846.671829, -836.079434, -825.412022, -814.677855, -803.885196, -793.042308, -782.157456, -771.238901, -760.600817, -750.092077, -739.557879, -728.99656, -718.406454, -707.785899, -697.133229, -686.446781, -675.724891, -664.965895, -654.168128, -643.329927, -632.449627, -621.525564, -610.556074, -599.539494, -588.474158, -577.245347, -565.966903, -554.639614, -543.264265, -531.841645, -520.37254, -508.857736, -497.298021, -485.694181, -474.047002, -462.357273, -450.625778, -438.853306, -427.040643, -415.188575, -403.29789, -391.308248, -379.160777, -366.982332, -354.777662, -342.551517, -330.308647, -318.0538, -305.791726, -293.527175, -281.264895, -269.009636, -256.766148, -244.53918, -232.33348, -220.153799, -208.004886, -195.89149, -184.161048, -172.639757, -161.149698, -149.684496, -138.237783, -126.803185, -115.374331, -103.944851, -92.5083712, -81.0585216, -69.5889302, -58.0932255, -46.5650361, -34.9979903, -23.3857167, -11.7218438, 0, 12.5023082, 24.8070046, 36.9566458, 48.9937889, 60.9609906, 72.9008077, 84.8557971, 96.8685155, 108.98152, 121.237367, 133.678614, 146.347817, 159.287533, 172.54032, 186.148733, 200.15533, 214.52733, 229.296527, 244.675885, 260.691816, 277.370734, 294.739052, 312.823183, 331.64954, 351.244536, 371.634583, 392.846096, 414.905486, 437.839168, 461.327452, 485.869757, 511.596326, 538.637404, 567.358326, 597.77179, 629.827801, 663.586075, 699.106328, 736.448274, 775.67163, 817.17101, 860.569305, 905.860707, 953.039407, 1002.0996, 1050.78372, 1101.90066, 1156.00753, 1213.66147, 1275.41961, 1344.62734, 1415.64563, 1491.05365, 1569.98089, 1644.51373, 1709.90074, 1761.39052, 1777.07171, 1773.25282, 1784.96227, 1867.06021, 1945.89021, 1951.38304, 1854.16523, 1693.81995, 1478.27816, 1249.8838, 1025.48971, 695.767686, 222.76468, -242.5492, -956.1153, -2021.28317, -3541.40236, -5619.82242, -10928.4394, -16359.9197, -21375.4763, -25436.322, -28003.67, -26295.9864, -23512.3952, -20075.2864, -17657.4832, -16186.4127, -15068.4876, -14257.1856, -13705.9842, -13368.361, -13197.7934, -13571.0509, -13665.4232, -13575.638, -13396.4228, -13222.5052, -13148.6128, -13269.4733, -13005.2246, -12985.4324},
        {-1003.72827, -998.557267, -993.190441, -987.63084, -981.881508, -975.945492, -969.825835, -963.525583, -957.04778, -950.395472, -943.571703, -936.579519, -929.421964, -922.102083, -914.622922, -906.987525, -899.198937, -891.260203, -883.174368, -874.944477, -866.573575, -858.064706, -849.420917, -840.645251, -831.740754, -822.710471, -813.557446, -804.284724, -794.895351, -785.392372, -775.778831, -766.057773, -756.232243, -746.305287, -736.245036, -726.072726, -715.810013, -705.461074, -695.030087, -684.52123, -673.938681, -663.286617, -652.569218, -641.790661, -630.955123, -620.066782, -609.129817, -598.148406, -587.126726, -576.068955, -564.979271, -554.060715, -543.114291, -532.139865, -521.137306, -510.10648, -499.047253, -487.959494, -476.843069, -465.697844, -454.523687, -443.320466, -432.088046, -420.826294, -409.535079, -398.214266, -386.863723, -375.424361, -363.838504, -352.227614, -340.595379, -328.945489, -317.281631, -305.607496, -293.926772, -282.243148, -270.560312, -258.881955, -247.211763, -235.553427, -223.910636, -212.287077, -200.686441, -189.112416, -177.850685, -166.75801, -155.68776, -144.634473, -133.592687, -122.556939, -111.521767, -100.481709, -89.4313017, -78.3650834, -67.2775914, -56.1633634, -45.0169371, -33.83285, -22.6056397, -11.3298438, 0, 12.3744472, 24.4890641, 36.3900273, 48.1235131, 59.735698, 71.2727583, 82.7808706, 94.3062112, 105.894957, 117.593283, 129.447367, 141.503386, 153.807514, 166.40593, 179.344809, 192.670328, 206.315868, 220.311492, 234.958508, 250.278922, 266.29474, 283.027968, 300.500613, 318.73468, 337.752176, 357.575106, 378.225476, 399.725294, 422.096564, 445.150607, 469.197366, 494.322054, 520.609881, 548.035653, 576.739786, 606.892136, 638.611038, 672.014825, 707.221833, 744.350395, 785.278206, 827.828782, 871.776234, 916.894674, 962.958212, 1003.54762, 1046.17869, 1092.17386, 1142.85558, 1199.54629, 1269.35418, 1340.74451, 1411.64068, 1479.55889, 1541.18394, 1593.57393, 1633.78693, 1649.65261, 1649.55485, 1651.94489, 1692.78491, 1717.40736, 1672.77766, 1535.53804, 1332.6165, 1067.85903, 778.588171, 477.329035, 53.8594268, -532.61003, -1137.05958, -2014.67374, -3270.56046, -5009.82766, -7337.58325, -12921.1211, -18662.8168, -24027.2316, -28478.9271, -31482.4646, -30338.446, -28118.0322, -25213.1956, -23141.4221, -21849.1688, -20813.929, -20005.6081, -19394.1112, -18949.3439, -18641.2112, -19225.8972, -19305.5626, -19038.4394, -18582.7596, -18096.755, -17738.6575, -17666.6991, -17183.6925, -16872.1398},
        {-946.797945, -941.948495, -936.973611, -931.872893, -926.645943, -921.292361, -915.811747, -910.203701, -904.467825, -898.603719, -892.610983, -886.489218, -880.238024, -873.857002, -867.345752, -860.703875, -853.930972, -847.026642, -839.990487, -832.822107, -825.521102, -818.087072, -810.51962, -802.818344, -794.982845, -787.012725, -778.907582, -770.667019, -762.290635, -753.778031, -745.128808, -736.342566, -727.418905, -718.357425, -708.940501, -699.280915, -689.494061, -679.58659, -669.565152, -659.436395, -649.206972, -638.883531, -628.472722, -617.981195, -607.415601, -596.782589, -586.08881, -575.340912, -564.545547, -553.709364, -542.839013, -532.210322, -521.554928, -510.873646, -500.16729, -489.436674, -478.682615, -467.905925, -457.107421, -446.287917, -435.448227, -424.589166, -413.71155, -402.816192, -391.903907, -380.975511, -370.031817, -359.055853, -348.031073, -336.994977, -325.949533, -314.89671, -303.838474, -292.776794, -281.713638, -270.650973, -259.590768, -248.53499, -237.485607, -226.444586, -215.413896, -204.395505, -193.39138, -182.403489, -171.631299, -160.973874, -150.327905, -139.68895, -129.052567, -118.414317, -107.769756, -97.1144439, -86.4439392, -75.7538005, -65.0395864, -54.2968554, -43.5211663, -32.7080775, -21.8531478, -10.9519358, 0, 11.9596202, 23.6578612, 35.1399484, 46.4511069, 57.6365617, 68.7415381, 79.8112613, 90.8909563, 102.025848, 113.261163, 124.642124, 136.213958, 148.02189, 160.111145, 172.526948, 185.314523, 198.481531, 212.067828, 226.203613, 240.926061, 256.272347, 272.279645, 288.985132, 306.425983, 324.639371, 343.662474, 363.532465, 384.28652, 405.961813, 428.679702, 452.362313, 477.021569, 502.669389, 528.713228, 555.467241, 583.406773, 612.725086, 643.61544, 676.271097, 710.885319, 750.749772, 792.016319, 834.27201, 877.103896, 920.099029, 954.248148, 989.883694, 1028.7418, 1072.55858, 1123.07018, 1189.49109, 1256.93885, 1319.23925, 1372.81786, 1419.40816, 1458.36046, 1489.0251, 1512.55778, 1526.09315, 1524.79634, 1515.98455, 1475.74638, 1380.91369, 1218.34717, 991.900878, 697.628629, 359.367412, -19.6259086, -537.966073, -1236.64348, -1989.78208, -3030.91048, -4458.78054, -6372.14412, -8869.75308, -14465.8384, -20239.8031, -25686.5291, -30300.8986, -33577.7936, -33020.9782, -31443.8645, -29194.6708, -27598.5546, -26617.0293, -25804.5501, -25142.9834, -24614.1953, -24200.052, -23882.4199, -24574.3493, -24665.9304, -24334.0291, -23755.5113, -23107.2428, -22566.0897, -22308.9177, -21623.9637, -21110.7606},
        {-896.802199, -892.484763, -888.046738, -883.487514, -878.806479, -874.003024, -869.076536, -864.026405, -858.85202, -853.552771, -848.128046, -842.577235, -836.899726, -831.094909, -825.162173, -819.100906, -812.910499, -806.59034, -800.139819, -793.558323, -786.845244, -779.999969, -773.021887, -765.910389, -758.664863, -751.284697, -743.769283, -736.118007, -728.33026, -720.40543, -712.342907, -704.14208, -695.802338, -687.32307, -678.488285, -669.409592, -660.201189, -650.869456, -641.420771, -631.861512, -622.198058, -612.436788, -602.58408, -592.646313, -582.629865, -572.541116, -562.386444, -552.172227, -541.904844, -531.590673, -521.236094, -511.088039, -500.907118, -490.694495, -480.451333, -470.178797, -459.87805, -449.550256, -439.196579, -428.818183, -418.416232, -407.991889, -397.54632, -387.080686, -376.596153, -366.093885, -355.575044, -345.028748, -334.444401, -323.848016, -313.241537, -302.626908, -292.006076, -281.380983, -270.753577, -260.1258, -249.499599, -238.876918, -228.259701, -217.649895, -207.049443, -196.460291, -185.884383, -175.323665, -164.973357, -154.734702, -144.506618, -134.284776, -124.064849, -113.842511, -103.613434, -93.3732908, -83.1177535, -72.8424952, -62.5431884, -52.2155058, -41.8551202, -31.4577041, -21.0189303, -10.5344714, 0, 11.1725937, 22.1610631, 33.0041535, 43.7406104, 54.409179, 65.0486049, 75.6976333, 86.3950096, 97.1794793, 108.089788, 119.16468, 130.442902, 141.963199, 153.764316, 165.884999, 178.363993, 191.291404, 204.713315, 218.552297, 232.858103, 247.680484, 263.06919, 279.073974, 295.744586, 313.130777, 331.282299, 350.248903, 370.080341, 390.826364, 412.768622, 435.639937, 459.420491, 484.090465, 508.797588, 533.938268, 560.1309, 587.605401, 616.59169, 647.319682, 680.019297, 718.549888, 758.407325, 799.111416, 840.181971, 881.138801, 913.356867, 946.53704, 982.235341, 1022.00779, 1067.41042, 1127.12402, 1186.87177, 1241.31079, 1287.00847, 1324.43232, 1352.29876, 1369.32423, 1375.52775, 1368.02711, 1342.51908, 1305.31994, 1234.39464, 1111.91523, 925.305156, 670.886096, 340.895298, -47.4900655, -495.56684, -1096.07827, -1890.04033, -2769.17154, -3946.7542, -5515.48421, -7568.05743, -10197.1698, -15813.1175, -21611.596, -27105.9011, -31809.3285, -35235.1739, -34992.7074, -33768.6011, -31892.162, -30621.0853, -29917.1909, -29355.47, -28917.0839, -28583.1939, -28334.9612, -28153.5471, -28661.7773, -28747.8246, -28523.3419, -28099.9821, -27589.3984, -27103.2435, -26753.1705, -25897.9715, -25438.2772},
        {-851.022377, -847.176145, -843.201326, -839.09789, -834.865806, -830.505043, -826.01557, -821.397355, -816.650369, -811.774579, -806.769955, -801.636465, -796.374079, -790.982766, -785.462494, -779.813232, -774.03495, -768.127617, -762.091201, -755.925671, -749.630996, -743.207146, -736.654089, -729.971794, -723.16023, -716.219367, -709.149172, -701.949616, -694.620666, -687.162293, -679.574464, -671.85715, -664.010318, -656.033938, -647.772558, -639.307127, -630.72046, -622.017572, -613.203473, -604.283177, -595.261697, -586.144044, -576.935233, -567.640275, -558.264183, -548.81197, -539.288648, -529.699231, -520.04873, -510.342158, -500.584529, -490.928719, -481.228671, -471.486192, -461.703089, -451.88117, -442.022242, -432.128113, -422.200588, -412.241476, -402.252584, -392.23572, -382.192689, -372.1253, -362.03536, -351.924676, -341.795055, -331.645054, -321.473306, -311.288325, -301.092127, -290.886732, -280.674156, -270.456419, -260.235537, -250.01353, -239.792414, -229.574208, -219.360931, -209.154599, -198.957231, -188.770845, -178.597459, -168.439091, -158.492515, -148.658276, -138.834574, -129.017109, -119.201577, -109.383676, -99.5591051, -89.7235607, -79.8727409, -70.0023435, -60.1080662, -50.1856069, -40.2306632, -30.2389329, -20.2061138, -10.1279036, 0, 10.3355035, 20.5781194, 30.7590634, 40.9095511, 51.0607979, 61.2440195, 71.4904315, 81.8312494, 92.2976889, 102.920965, 113.732295, 124.762892, 136.043974, 147.606755, 159.482451, 171.702277, 184.445147, 197.763374, 211.354099, 225.280187, 239.604502, 254.38991, 269.699275, 285.595462, 302.141336, 319.399761, 337.433604, 356.305728, 376.078999, 397.178938, 419.172779, 442.014589, 465.658437, 489.064798, 512.684536, 537.233472, 562.963751, 590.127518, 618.97692, 649.764101, 686.609437, 724.719556, 763.589773, 802.715408, 841.591777, 872.98787, 904.806915, 938.22581, 974.421455, 1014.57075, 1065.91865, 1116.15748, 1161.54124, 1199.10394, 1227.47218, 1244.55751, 1248.27149, 1236.86917, 1207.84054, 1158.30088, 1094.17108, 995.022134, 848.651815, 643.029473, 366.415171, 6.93553023, -426.318759, -939.785948, -1616.77255, -2497.10403, -3493.78025, -4792.96393, -6478.38174, -8633.76032, -11342.8263, -16858.0178, -22552.1722, -27966.8383, -32643.5648, -36123.9007, -36161.7466, -35278.0648, -33780.5388, -32856.2092, -32464.6808, -32199.16, -32037.5715, -31957.8403, -31937.8911, -31955.6489, -32228.994, -32332.005, -32286.5199, -32114.3763, -31837.4122, -31477.4654, -31056.3736, -30084.3951, -29706.5315},
        {-807.945011, -804.555175, -801.02129, -797.34432, -793.525227, -789.564972, -785.464518, -781.224828, -776.846864, -772.331589, -767.679964, -762.892952, -757.971516, -752.916618, -747.72922, -742.410284, -736.960774, -731.381651, -725.673877, -719.838416, -713.876229, -707.78828, -701.575529, -695.23894, -688.779476, -682.198097, -675.495768, -668.673449, -661.732105, -654.672696, -647.496185, -640.203535, -632.795708, -625.273667, -617.570681, -609.722983, -601.767619, -593.707748, -585.546529, -577.287121, -568.932683, -560.486375, -551.951355, -543.330783, -534.627818, -525.845619, -516.987344, -508.056154, -499.055207, -489.987663, -480.85668, -471.69454, -462.474649, -453.199534, -443.871723, -434.493743, -425.068124, -415.597392, -406.084075, -396.530701, -386.939798, -377.313894, -367.655516, -357.967193, -348.251451, -338.51082, -328.747826, -318.970158, -309.185383, -299.385382, -289.572349, -279.748478, -269.915961, -260.076993, -250.233765, -240.388472, -230.543307, -220.700462, -210.862131, -201.030507, -191.207784, -181.396155, -171.597813, -161.814951, -152.250847, -142.802923, -133.366183, -123.936295, -114.508926, -105.079743, -95.6444146, -86.1986084, -76.7379917, -67.2582323, -57.7549978, -48.2239557, -38.6607738, -29.0611197, -19.4206609, -9.73506513, 0, 9.55133313, 19.0942601, 28.6529264, 38.2514774, 47.9140584, 57.6648149, 67.5278923, 77.527436, 87.6875914, 98.0325039, 108.586319, 119.373182, 130.417238, 141.742633, 153.373512, 165.33402, 177.870115, 191.037628, 204.358989, 217.905876, 231.749964, 245.962929, 260.616448, 275.782198, 291.531854, 307.937094, 325.069592, 343.001026, 361.803073, 381.989806, 403.028292, 424.857485, 447.416344, 469.602298, 491.875068, 514.972113, 539.144846, 564.644683, 591.723036, 620.63132, 655.396736, 691.34576, 727.991063, 764.845319, 801.421201, 832.598055, 863.680213, 895.338682, 928.244466, 963.068572, 1005.0624, 1044.7183, 1080.20624, 1109.20504, 1128.39069, 1134.88943, 1125.82749, 1097.76723, 1048.52692, 976.540008, 887.198407, 761.849494, 593.94293, 372.071926, 76.6007034, -308.329273, -782.229256, -1356.7196, -2103.36069, -3059.80746, -4161.61259, -5564.24126, -7340.69147, -9563.9612, -12307.0485, -17627.432, -23117.5088, -28354.1569, -32914.2538, -36374.6775, -36662.895, -36104.8021, -34984.1669, -34411.4794, -34343.2636, -34391.5763, -34530.1651, -34732.7775, -34973.161, -35225.0631, -35256.5668, -35403.4698, -35585.9663, -35724.2505, -35738.5165, -35548.9586, -35075.771, -34059.27, -33744.5849},
        {-766.056631, -763.154388, -760.090549, -756.867103, -753.486044, -749.949363, -746.25905, -742.417099, -738.425501, -734.286247, -730.001329, -725.57274, -721.002469, -716.29251, -711.444854, -706.461493, -701.344417, -696.09562, -690.717093, -685.210827, -679.578814, -673.823045, -667.945514, -661.94821, -655.833126, -649.602254, -643.257585, -636.801111, -630.234824, -623.560715, -616.780776, -609.896999, -602.911375, -595.825896, -588.660015, -581.406624, -574.05841, -566.616796, -559.083209, -551.459072, -543.745812, -535.944852, -528.057619, -520.085538, -512.030032, -503.892529, -495.674451, -487.377226, -479.002277, -470.55103, -462.02491, -453.347682, -444.600115, -435.785318, -426.906398, -417.966464, -408.968625, -399.915989, -390.811665, -381.658761, -372.460387, -363.21965, -353.93966, -344.623524, -335.274352, -325.895251, -316.489331, -307.069448, -297.648225, -288.208666, -278.753247, -269.284445, -259.804736, -250.316597, -240.822505, -231.324937, -221.826369, -212.329277, -202.836139, -193.34943, -183.871628, -174.40521, -164.952651, -155.516429, -146.310428, -137.226975, -128.155853, -119.09268, -110.033071, -100.972642, -91.9070092, -82.8317886, -73.7425963, -64.6350484, -55.5047607, -46.3473495, -37.1584307, -27.9336202, -18.6685343, -9.35878889, 0, 8.92306631, 17.8947151, 26.9339907, 36.0599378, 45.2916008, 54.6480242, 64.1482524, 73.81133, 83.6563014, 93.7022111, 103.968104, 114.473023, 125.236015, 136.276123, 147.612392, 159.263865, 171.493663, 184.355697, 197.316938, 210.448734, 223.822431, 237.509373, 251.580908, 266.108382, 281.16314, 296.816529, 313.139895, 330.204583, 348.081941, 367.280383, 387.273928, 408.0028, 429.407223, 450.49753, 471.678891, 493.60445, 516.493402, 540.564939, 566.038254, 593.132542, 625.380102, 658.678684, 692.599264, 726.712817, 760.590321, 791.644317, 822.143826, 852.199431, 881.921719, 911.421275, 943.742668, 972.476517, 997.58146, 1017.41254, 1027.0508, 1023.04723, 1001.95284, 959.407134, 893.079758, 801.634726, 689.062333, 539.097359, 350.608071, 112.984316, -200.444711, -608.977713, -1120.33214, -1750.80416, -2559.15449, -3580.12352, -4770.67298, -6255.2878, -8095.63173, -10353.3685, -13090.1619, -18148.2526, -23363.5831, -28352.6729, -32732.0415, -36118.2082, -36630.9518, -36381.3598, -35627.4121, -35394.449, -35636.7035, -35988.6752, -36420.5837, -36902.6488, -37405.0901, -37898.1274, -37725.0633, -37947.217, -38384.0844, -38855.1612, -39179.9433, -39177.9265, -38668.6065, -37698.6316, -37381.4985},
        {-723.843768, -721.506321, -718.993018, -716.306538, -713.449562, -710.424768, -707.234836, -703.882444, -700.370272, -696.701, -692.877306, -688.90187, -684.777371, -680.506488, -676.091901, -671.536288, -666.84233, -662.012704, -657.050091, -651.957169, -646.736619, -641.391118, -635.923347, -630.335985, -624.63171, -618.813202, -612.883141, -606.844205, -600.699074, -594.450427, -588.100943, -581.653302, -575.110183, -568.474264, -561.817919, -555.107514, -548.308578, -541.421528, -534.446783, -527.38476, -520.235877, -513.000551, -505.679199, -498.272239, -490.78009, -483.203167, -475.54189, -467.796675, -459.96794, -452.056103, -444.061581, -435.850325, -427.560134, -419.194341, -410.756279, -402.249279, -393.676675, -385.041799, -376.347983, -367.598561, -358.796865, -349.946226, -341.049979, -332.111454, -323.133986, -314.120906, -305.075547, -296.008311, -286.92943, -277.827656, -268.705865, -259.56693, -250.413725, -241.249124, -232.076002, -222.897234, -213.715693, -204.534253, -195.35579, -186.183176, -177.019287, -167.866997, -158.72918, -149.60871, -140.733334, -131.98876, -123.257993, -114.53661, -105.820189, -97.1043048, -88.384535, -79.6564562, -70.9156451, -62.1576785, -53.378133, -44.5725853, -35.7366122, -26.8657904, -17.9556967, -9.00190762, 0, 8.55368671, 17.1647141, 25.8505047, 34.6284809, 43.5160652, 52.5306802, 61.6897482, 71.0106917, 80.5109332, 90.2078953, 100.119, 110.261671, 120.653329, 131.311398, 142.253299, 153.496456, 165.243147, 177.537202, 189.977918, 202.622327, 215.527466, 228.750368, 242.348068, 256.377601, 270.896, 285.960301, 301.627538, 317.954745, 334.998957, 353.129825, 371.977141, 391.504154, 411.674115, 431.837933, 452.265027, 473.388109, 495.35413, 518.310044, 542.402801, 567.779355, 597.027852, 627.111077, 657.698356, 688.459015, 719.062381, 749.583555, 779.184645, 807.433536, 833.898109, 858.146248, 881.146861, 899.354443, 913.942544, 923.827218, 923.315465, 908.783657, 876.608165, 822.974085, 744.492535, 637.983297, 504.423283, 330.986366, 121.466729, -133.681554, -466.608475, -899.08839, -1445.73801, -2126.47599, -2987.46573, -4060.02508, -5318.96579, -6860.80515, -8736.42086, -10996.6906, -13692.4922, -18447.3722, -23346.3722, -28047.2026, -32207.5739, -35485.1965, -36200.7161, -36240.2845, -35834.64, -35912.6712, -36428.7651, -37046.4128, -37734.5462, -38462.0973, -39197.9978, -39911.1797, -39615.0508, -39948.2445, -40643.2772, -41432.6652, -42048.9248, -42224.5724, -41692.1242, -40878.5154, -40446.3337},
        {-673.679382, -672.322987, -670.788002, -669.076477, -667.190462, -665.132004, -662.903154, -660.50596, -657.942471, -655.214736, -652.324804, -649.274725, -646.066547, -642.702319, -639.184091, -635.513911, -631.693829, -627.725893, -623.612152, -619.354657, -614.955454, -610.416595, -605.740126, -600.928099, -595.982561, -590.905562, -585.699151, -580.365376, -574.906287, -569.323934, -563.620364, -557.797626, -551.857771, -545.802847, -539.628886, -533.341073, -526.944663, -520.441901, -513.835031, -507.126297, -500.317944, -493.412215, -486.411356, -479.31761, -472.133222, -464.860435, -457.501495, -450.058646, -442.534131, -434.930195, -427.249082, -419.502999, -411.684012, -403.794149, -395.835438, -387.809908, -379.719588, -371.566505, -363.352688, -355.080165, -346.750965, -338.367115, -329.930645, -321.443582, -312.907956, -304.325793, -295.699123, -286.997962, -278.193095, -269.352572, -260.480498, -251.580975, -242.658107, -233.715997, -224.758749, -215.790467, -206.815254, -197.837214, -188.860449, -179.889064, -170.927163, -161.978847, -153.048222, -144.139391, -135.526389, -127.072676, -118.638467, -110.219107, -101.809939, -93.406306, -85.0035527, -76.5970224, -68.1820589, -59.7540059, -51.3082071, -42.8400062, -34.3447469, -25.817773, -17.254428, -8.65005579, 0, 9.04319276, 17.9816826, 26.8438404, 35.6580369, 44.452643, 53.2560295, 62.0965672, 71.002627, 80.0025796, 89.124796, 98.3976469, 107.849503, 117.508736, 127.403715, 137.562813, 148.014399, 158.549035, 169.161118, 180.416923, 192.293862, 204.769345, 217.820785, 231.425592, 245.561179, 260.204958, 275.334339, 290.926734, 306.959556, 323.410215, 339.876316, 356.85434, 374.435642, 392.711575, 412.373294, 433.212252, 454.859955, 477.227817, 500.227251, 523.769669, 547.766487, 571.809534, 596.22707, 620.993035, 646.081371, 671.466017, 701.767898, 731.152224, 758.431188, 782.416983, 801.921805, 815.393367, 822.453817, 822.714289, 815.291738, 798.29416, 770.282554, 729.817918, 683.794765, 620.546598, 519.315813, 362.225531, 154.488724, -85.5486753, -361.88226, -716.364095, -1173.46951, -1756.30447, -2488.98931, -3400.66524, -4517.63338, -5819.22882, -7385.94424, -9260.33822, -11484.9694, -14102.3962, -18526.3705, -23085.4595, -27479.4233, -31408.0224, -34571.017, -35506.334, -35850.1227, -35800.0722, -36142.941, -36843.2447, -37620.0529, -38450.1536, -39310.3344, -40177.3832, -41028.0878, -40825.9426, -41373.7918, -42342.3066, -43402.1584, -44224.0186, -44478.5585, -43836.4494, -43461.008, -42544.4746},
        {-633.134032, -632.306366, -631.305285, -630.132385, -628.78926, -627.277508, -625.598723, -623.754502, -621.746441, -619.576134, -617.245179, -614.75517, -612.107704, -609.304377, -606.346783, -603.23652, -599.975183, -596.564367, -593.005669, -589.300683, -585.451007, -581.458236, -577.323966, -573.049792, -568.63731, -564.088117, -559.403807, -554.585977, -549.636222, -544.556139, -539.347323, -534.01137, -528.549876, -522.964436, -517.218857, -511.334424, -505.332878, -499.217039, -492.989731, -486.653774, -480.211992, -473.667207, -467.02224, -460.279915, -453.443052, -446.514474, -439.497004, -432.393463, -425.206673, -417.939457, -410.594637, -403.215013, -395.762562, -388.23924, -380.647002, -372.987805, -365.263603, -357.476352, -349.628007, -341.720524, -333.755859, -325.735967, -317.662804, -309.538324, -301.364484, -293.143239, -284.876545, -276.535464, -268.091804, -259.611231, -251.097703, -242.555179, -233.987616, -225.398972, -216.793206, -208.174276, -199.546139, -190.912754, -182.278079, -173.646072, -165.020691, -156.405895, -147.80564, -139.223886, -130.91924, -122.762979, -114.62332, -106.495956, -98.3765819, -90.2608922, -82.1445807, -74.0233418, -65.8928696, -57.7488584, -49.5870023, -41.4029956, -33.1925324, -24.9513071, -16.6750138, -8.35934667, 0, 8.94392636, 17.7472497, 26.4397198, 35.0510865, 43.6110994, 52.1495084, 60.6960631, 69.2805134, 77.9326089, 86.6820996, 95.558735, 104.592265, 113.812439, 123.249008, 132.93172, 142.890326, 152.88704, 162.913394, 173.604275, 184.932103, 196.8693, 209.388287, 222.461484, 236.061312, 250.160193, 264.730546, 279.744794, 295.175357, 310.994656, 326.79235, 343.063968, 359.896759, 377.377974, 396.347523, 416.516324, 437.394588, 458.843766, 480.725309, 502.900668, 525.231295, 547.055265, 568.916694, 590.779432, 612.607332, 634.364245, 660.443949, 685.272885, 707.707423, 726.603931, 740.818779, 749.035578, 750.494601, 745.491839, 733.145992, 710.172132, 674.364513, 623.517387, 561.220842, 478.15606, 358.681491, 191.428868, -26.0450226, -283.692164, -588.189288, -974.55229, -1464.78892, -2079.39059, -2839.97189, -3773.70747, -4904.62689, -6223.76545, -7793.22303, -9648.4503, -11824.8979, -14358.0166, -18491.0046, -22749.628, -26867.4006, -30577.8362, -33614.4486, -34682.6969, -35229.5192, -35429.0241, -35976.1613, -36838.8732, -37768.0849, -38742.5896, -39741.1804, -40742.6506, -41725.7932, -41592.9912, -42265.9578, -43395.1815, -44631.1508, -45624.3543, -46025.2806, -45484.418, -45369.5792, -44207.8829},
        {-600.947329, -600.241446, -599.375937, -598.35203, -597.170955, -595.833942, -594.342218, -592.697014, -590.89956, -588.951083, -586.852814, -584.605981, -582.211815, -579.671543, -576.986396, -574.157603, -571.186393, -568.073995, -564.821639, -561.430554, -557.901968, -554.237112, -550.437215, -546.503505, -542.437213, -538.239567, -533.911796, -529.455131, -524.870799, -520.160031, -515.324056, -510.364103, -505.281401, -500.07718, -494.717422, -489.221721, -483.610095, -477.884916, -472.048559, -466.103395, -460.051798, -453.896141, -447.638798, -441.28214, -434.828542, -428.280376, -421.640015, -414.909832, -408.092201, -401.189495, -394.204086, -387.119222, -379.956815, -372.719655, -365.410529, -358.032224, -350.587528, -343.079229, -335.510115, -327.882973, -320.200592, -312.465757, -304.681259, -296.849883, -288.974418, -281.057652, -273.102372, -265.11007, -257.082271, -249.024434, -240.939429, -232.830129, -224.699405, -216.550129, -208.385172, -200.207407, -192.019704, -183.824936, -175.625974, -167.42569, -159.226956, -151.032642, -142.845621, -134.668765, -126.704262, -118.851133, -111.006004, -103.165276, -95.3253549, -87.4826427, -79.6335432, -71.7744599, -63.9017961, -56.0119552, -48.1013408, -40.1663561, -32.2034047, -24.20889, -16.1792153, -8.11078421, 0, 8.38449879, 16.6909594, 24.9431277, 33.1647498, 41.3795715, 49.6113388, 57.8837976, 66.220694, 74.6457738, 83.182783, 91.8554675, 100.687573, 109.702847, 118.925033, 128.377878, 138.085129, 148.054064, 158.306078, 168.902161, 179.862531, 191.207406, 202.957002, 215.131538, 227.751231, 240.836298, 254.406957, 268.483425, 283.085919, 298.234657, 314.053019, 330.420234, 347.32557, 364.758297, 382.839851, 401.493416, 420.606934, 440.130025, 460.012307, 480.203399, 500.65292, 522.481814, 544.111886, 565.263581, 585.657347, 605.01363, 623.922917, 641.018105, 655.802132, 667.777934, 676.448448, 682.915661, 683.129068, 679.163009, 669.858548, 647.455479, 607.157431, 544.168035, 449.93271, 324.267434, 171.088048, 1.43886698, -210.53552, -476.642849, -811.207422, -1232.81773, -1758.10493, -2398.60112, -3169.61548, -4105.15375, -5228.64586, -6546.11882, -8101.0828, -9922.99178, -12041.2998, -14485.4607, -18360.76, -22351.8624, -26219.2641, -29723.4613, -32624.95, -33756.4264, -34424.7201, -34787.956, -35488.9081, -36493.8608, -37567.1614, -38685.1912, -39824.3316, -40960.9638, -42071.4691, -41975.4883, -42703.0922, -43912.0834, -45260.2646, -46405.4383, -47005.4072, -46717.9739, -46697.1754, -45534.6875},
        {-573.713799, -572.8715, -571.893471, -570.780373, -569.532869, -568.15162, -566.637288, -564.990534, -563.21202, -561.302408, -559.26236, -557.092537, -554.793601, -552.366214, -549.811037, -547.128732, -544.319961, -541.385386, -538.325668, -535.141469, -531.833451, -528.402275, -524.848603, -521.173098, -517.376419, -513.459231, -509.422193, -505.265967, -500.991217, -496.598602, -492.088785, -487.462428, -482.720192, -477.862739, -472.860891, -467.730857, -462.489205, -457.137565, -451.677567, -446.110842, -440.439019, -434.663728, -428.7866, -422.809265, -416.733352, -410.560492, -404.292315, -397.930451, -391.476529, -384.932181, -378.299036, -371.476017, -364.569688, -357.583906, -350.522526, -343.389405, -336.188401, -328.923369, -321.598166, -314.216649, -306.782674, -299.300097, -291.772776, -284.204567, -276.599326, -268.96091, -261.293176, -253.63557, -246.026686, -238.396977, -230.747992, -223.081281, -215.398392, -207.700876, -199.990282, -192.268158, -184.536055, -176.795522, -169.048107, -161.295361, -153.538833, -145.780072, -138.020627, -130.262048, -122.641462, -115.089777, -107.536274, -99.9781004, -92.4124073, -84.836344, -77.2470601, -69.6417052, -62.0174289, -54.3713807, -46.7007104, -39.0025674, -31.2741015, -23.5124622, -15.7147991, -7.87826184, 0, 7.670116, 15.3592814, 23.0827498, 30.855775, 38.6936106, 46.6115102, 54.6247275, 62.7485161, 70.9981299, 79.3888223, 87.935847, 96.6544578, 105.559908, 114.667452, 123.992343, 133.549834, 143.665451, 154.398769, 165.063243, 175.74061, 186.512613, 197.46099, 208.667483, 220.21383, 232.181773, 244.653052, 257.709406, 271.432576, 285.904302, 301.929392, 318.601393, 335.785125, 353.345409, 370.409386, 387.210714, 404.179768, 421.402669, 438.965542, 456.95451, 475.455695, 498.123353, 520.389523, 541.642215, 561.26944, 578.659208, 589.012954, 596.95191, 602.910728, 607.324066, 610.626576, 617.577583, 619.001365, 618.036561, 612.8555, 591.490968, 546.528197, 470.552421, 342.23305, 171.233829, -17.600598, -189.543832, -394.601742, -666.585998, -1027.56347, -1479.70011, -2034.29673, -2693.95498, -3467.72032, -4396.53512, -5503.29902, -6805.77258, -8332.78522, -10109.0156, -12159.1422, -14507.8437, -18145.8374, -21890.2535, -25524.261, -28831.0288, -31593.7258, -32748.7492, -33490.5543, -33965.2123, -34780.7102, -35900.2775, -37094.6788, -38336.7667, -39599.3942, -40855.4139, -42077.6787, -42007.8485, -42761.7349, -44016.0209, -45447.3896, -46732.5237, -47548.1063, -47570.8204, -47542.1766, -46556.938},
        {-545.055603, -543.886118, -542.64125, -541.319422, -539.919055, -538.438572, -536.876396, -535.230948, -533.500652, -531.683928, -529.7792, -527.78489, -525.699421, -523.521213, -521.24869, -518.880275, -516.414388, -513.849454, -511.183893, -508.416128, -505.544582, -502.567677, -499.483835, -496.291479, -492.98903, -489.574911, -486.047544, -482.405352, -478.646757, -474.770181, -470.774047, -466.656776, -462.416792, -458.052515, -453.43624, -448.632106, -443.705769, -438.659745, -433.496548, -428.218696, -422.828702, -417.329083, -411.722354, -406.01103, -400.197627, -394.28466, -388.274646, -382.170098, -375.973533, -369.687466, -363.314413, -356.807725, -350.220147, -343.555261, -336.816648, -330.007888, -323.132564, -316.194256, -309.196546, -302.143014, -295.037242, -287.882812, -280.683304, -273.442299, -266.16338, -258.850127, -251.50612, -244.165204, -236.860493, -229.533158, -222.184817, -214.81709, -207.431598, -200.029959, -192.613793, -185.184719, -177.744358, -170.294328, -162.836248, -155.37174, -147.902421, -140.429912, -132.955832, -125.4818, -118.14609, -110.879121, -103.609667, -96.3349144, -89.0520472, -81.7582505, -74.4507091, -67.1266078, -59.7831315, -52.4174649, -45.0267931, -37.6083007, -30.1591726, -22.6765936, -15.1577487, -7.59982251, 0, 7.56811296, 15.1247583, 22.6868138, 30.2711571, 37.8946658, 45.5742178, 53.3266906, 61.1689621, 69.1179098, 77.1904116, 85.403345, 93.7735878, 102.318018, 111.053512, 119.996949, 129.165207, 138.761919, 148.830644, 158.965832, 169.22438, 179.663185, 190.339144, 201.309155, 212.630115, 224.35892, 236.552468, 249.267656, 262.561381, 276.490539, 291.703799, 307.449305, 323.606423, 340.054519, 356.013589, 371.692684, 387.476688, 403.442778, 419.66813, 436.229922, 453.205331, 475.172272, 496.337395, 515.897296, 533.048574, 546.987827, 547.947785, 546.32988, 543.571675, 541.110735, 540.384624, 551.302247, 556.475967, 552.724362, 535.951724, 500.195669, 440.331915, 351.236178, 220.427898, 51.810947, -142.685773, -343.788069, -583.681158, -878.902396, -1245.19759, -1696.9713, -2249.24375, -2915.10176, -3709.06432, -4652.73944, -5763.72518, -7046.26622, -8533.36836, -10247.252, -12210.1374, -14444.2451, -17831.159, -21318.895, -24714.8325, -27826.351, -30460.8301, -31675.971, -32528.6168, -33147.4375, -34068.2096, -35258.4568, -36516.0754, -37817.0707, -39137.4479, -40453.2123, -41740.3691, -41606.5106, -42408.1404, -43793.3849, -45410.3704, -46907.2235, -47932.0703, -48133.0375, -48026.3754, -47323.4756},
        {-516.455553, -515.050596, -513.627677, -512.182967, -510.712635, -509.21285, -507.679781, -506.109597, -504.498467, -502.842561, -501.138047, -499.381095, -497.567874, -495.694552, -493.7573, -491.752286, -489.675679, -487.523649, -485.292364, -482.977994, -480.576708, -478.084675, -475.498064, -472.813044, -470.025784, -467.132454, -464.129223, -461.012259, -457.777733, -454.421812, -450.940666, -447.330465, -443.587377, -439.707572, -435.45188, -430.939088, -426.290816, -421.510868, -416.603052, -411.571173, -406.419038, -401.150454, -395.769226, -390.279161, -384.684065, -378.987745, -373.194007, -367.306657, -361.329501, -355.266347, -349.120999, -342.943965, -336.691337, -330.365912, -323.970482, -317.507841, -310.980784, -304.392105, -297.744597, -291.041054, -284.284271, -277.477042, -270.62216, -263.722419, -256.780614, -249.799539, -242.781987, -235.738851, -228.680831, -221.594017, -214.480677, -207.34308, -200.183495, -193.004191, -185.807438, -178.595503, -171.370657, -164.135167, -156.891303, -149.641334, -142.387529, -135.132157, -127.877486, -120.625786, -113.54513, -106.551397, -99.5584738, -92.5632485, -85.5626094, -78.5534449, -71.5326431, -64.4970924, -57.443681, -50.3692973, -43.2708295, -36.1451658, -28.9891946, -21.7998042, -14.5738827, -7.30831858, 0, 7.62368821, 15.1753186, 22.6757228, 30.1457327, 37.6061798, 45.0778958, 52.5817126, 60.1384617, 67.7689749, 75.4940839, 83.3346203, 91.3114159, 99.4453024, 107.757111, 116.267675, 124.997824, 133.903832, 142.997306, 152.445107, 162.254232, 172.431678, 182.984442, 193.919523, 205.243918, 216.964624, 229.088639, 241.62296, 254.574586, 267.950513, 281.984211, 296.373167, 311.056434, 325.973069, 340.845602, 355.721351, 370.705375, 385.801689, 401.014305, 416.347237, 431.804499, 451.757961, 470.514433, 487.223347, 501.034137, 511.096235, 505.058564, 496.446195, 487.283689, 479.595607, 475.406509, 488.355401, 494.656967, 483.615006, 448.958029, 393.448324, 315.7922, 214.695964, 93.2601998, -55.2017618, -242.170344, -475.212666, -757.245455, -1077.63047, -1442.13089, -1884.30122, -2424.93927, -3092.00503, -3908.15318, -4869.77715, -5988.12521, -7247.52167, -8690.28093, -10336.9711, -12208.1602, -14324.4165, -17468.4972, -20708.2339, -23873.6475, -26794.7588, -29301.5886, -30567.4116, -31516.8256, -32261.314, -33258.3341, -34482.4213, -35763.9557, -37084.6804, -38426.3389, -39770.6745, -41099.4306, -40999.9101, -41900.2219, -43427.4044, -45208.4963, -46870.536, -48040.5621, -48345.6131, -48209.6478, -47838.7355},
        {-488.567494, -487.335064, -486.100637, -484.859681, -483.607664, -482.340054, -481.052318, -479.739923, -478.398338, -477.02303, -475.609466, -474.153114, -472.649441, -471.093916, -469.482006, -467.809177, -466.070899, -464.262639, -462.379863, -460.41804, -458.372637, -456.239122, -454.012963, -451.689626, -449.26458, -446.733293, -444.091231, -441.333862, -438.456654, -435.455075, -432.324592, -429.060672, -425.658784, -422.114395, -418.145178, -413.89134, -409.49643, -404.964928, -400.301318, -395.510082, -390.595703, -385.562663, -380.415444, -375.15853, -369.796403, -364.333544, -358.774438, -353.123566, -347.385411, -341.564456, -335.665182, -329.800067, -323.863258, -317.856896, -311.783122, -305.64408, -299.441909, -293.178751, -286.856748, -280.47804, -274.044771, -267.55908, -261.02311, -254.439002, -247.808897, -241.134936, -234.419262, -227.652542, -220.82572, -213.9646, -207.072067, -200.151007, -193.204304, -186.234845, -179.245515, -172.239198, -165.21878, -158.187146, -151.147181, -144.101771, -137.0538, -130.006155, -122.96172, -115.92338, -109.089658, -102.361506, -95.6375234, -88.9142472, -82.1882136, -75.4559591, -68.7140202, -61.9589332, -55.1872347, -48.3954609, -41.5801484, -34.7378337, -27.865053, -20.9583428, -14.0142397, -7.02927992, 0, 7.45196552, 14.8246127, 22.1383221, 29.4134745, 36.6704506, 43.929631, 51.2113965, 58.5361276, 65.9242052, 73.3960098, 80.9719223, 88.6723231, 96.5175931, 104.528113, 112.724263, 121.126425, 129.588871, 138.108252, 147.080669, 156.490907, 166.323753, 176.563994, 187.196414, 198.205801, 209.576941, 221.29462, 233.343624, 245.70874, 258.374753, 271.390788, 284.653702, 298.128981, 311.78211, 325.712268, 339.818092, 353.962569, 368.071077, 382.068992, 395.881691, 409.434552, 425.882623, 440.938666, 453.896165, 464.048604, 470.689467, 463.917015, 454.518763, 444.087, 434.214019, 426.492108, 433.589102, 432.484974, 409.472968, 357.882196, 285.406665, 193.290834, 82.7791663, -33.1951895, -167.73298, -346.686246, -592.345624, -893.109804, -1226.75606, -1594.41281, -2036.77569, -2576.37833, -3247.38262, -4075.33691, -5043.15271, -6157.85934, -7393.83205, -8795.77796, -10382.0297, -12170.92, -14180.7815, -17109.2917, -20125.6022, -23078.2094, -25815.6099, -28186.3, -29453.8907, -30441.6876, -31248.8524, -32273.4589, -33496.3494, -34773.8194, -36092.9383, -37440.7759, -38804.4017, -40170.8851, -40415.5203, -41513.3467, -43131.1274, -44935.6254, -46593.6039, -47771.8259, -48137.0546, -48155.6498, -48114.3905},
        {-461.473317, -461.160519, -460.782335, -460.33658, -459.821071, -459.233625, -458.572059, -457.83419, -457.017836, -456.120812, -455.140936, -454.076024, -452.923895, -451.682364, -450.349248, -448.922365, -447.399531, -445.778564, -444.057279, -442.233495, -440.305028, -438.269694, -436.125312, -433.869697, -431.500667, -429.016039, -426.41363, -423.691255, -420.846734, -417.877881, -414.782515, -411.558452, -408.203509, -404.715503, -400.912047, -396.88485, -392.727787, -388.444524, -384.038725, -379.514055, -374.874179, -370.122762, -365.263468, -360.299964, -355.235913, -350.07498, -344.820831, -339.47713, -334.047543, -328.535733, -322.945367, -317.350386, -311.682655, -305.944314, -300.137506, -294.264371, -288.32705, -282.327687, -276.268421, -270.151394, -263.978748, -257.752623, -251.475162, -245.148506, -238.774797, -232.356174, -225.894781, -219.379592, -212.799896, -206.184992, -199.537874, -192.861538, -186.158978, -179.433188, -172.687166, -165.923904, -159.146398, -152.357643, -145.560635, -138.758367, -131.953836, -125.150035, -118.34996, -111.556606, -104.9782, -98.510803, -92.0480119, -85.5861611, -79.1215853, -72.6506195, -66.1695983, -59.6748566, -53.1627291, -46.6295508, -40.0716563, -33.4853805, -26.8670583, -20.2130243, -13.5196133, -6.78316033, 0, 6.71683955, 13.4739666, 20.2809061, 27.1471828, 34.0823216, 41.0958473, 48.1972848, 55.3961589, 62.7019945, 70.1243165, 77.6726496, 85.3565188, 93.1854489, 101.168965, 109.316591, 117.637853, 126.270602, 135.242696, 144.273396, 153.399725, 162.658708, 172.087366, 181.722725, 191.601807, 201.761636, 212.239236, 223.071629, 234.29584, 245.948892, 258.554503, 271.484547, 284.63004, 297.881996, 311.153359, 324.32418, 337.268663, 349.871245, 362.016363, 373.588453, 384.471953, 395.464718, 405.25977, 413.562833, 420.079631, 424.515889, 427.566288, 427.700357, 424.376581, 417.053446, 405.189438, 392.24909, 368.788559, 327.947429, 266.108132, 186.273909, 88.475393, -27.2567825, -154.705363, -301.432388, -481.748938, -699.995, -967.985598, -1290.28682, -1676.77804, -2146.34171, -2713.72772, -3398.94225, -4218.09792, -5168.03433, -6256.49299, -7472.52727, -8845.44193, -10389.4475, -12118.7543, -14047.5731, -16804.6928, -19636.1006, -22402.3625, -24964.0442, -27181.7115, -28367.237, -29295.6753, -30062.7466, -31046.7704, -32231.6045, -33483.357, -34791.7732, -36146.5985, -37537.5782, -38954.4577, -40079.8664, -41534.4957, -43137.8187, -44709.3085, -46068.4381, -47034.6807, -47427.5092, -47930.6693, -48167.0846},
        {-437.574084, -437.898602, -438.111519, -438.212486, -438.201154, -438.077176, -437.840202, -437.489885, -437.025875, -436.447824, -435.755383, -434.948205, -434.025941, -432.988241, -431.834758, -430.565144, -429.179049, -427.676125, -426.056024, -424.318397, -422.462896, -420.489172, -418.396877, -416.185662, -413.855179, -411.405079, -408.835013, -406.144634, -403.333593, -400.401541, -397.34813, -394.173011, -390.875836, -387.456256, -383.816631, -380.007306, -376.079793, -372.0369, -367.881437, -363.616213, -359.244035, -354.767714, -350.190057, -345.513874, -340.741973, -335.877163, -330.922253, -325.880052, -320.753368, -315.545011, -310.257788, -304.919514, -299.50745, -294.023863, -288.471021, -282.85119, -277.166636, -271.419627, -265.612429, -259.747309, -253.826533, -247.85237, -241.827084, -235.752944, -229.632215, -223.467165, -217.26006, -211.00444, -204.694053, -198.349165, -191.97261, -185.567221, -179.135829, -172.681267, -166.206367, -159.713963, -153.206887, -146.687971, -140.160048, -133.62595, -127.08851, -120.550561, -114.014935, -107.484464, -101.158633, -94.9378116, -88.7200073, -82.5016708, -76.279253, -70.0492047, -63.8079768, -57.55202, -51.2777853, -44.9817234, -38.6602852, -32.3099216, -25.9270832, -19.5082211, -13.0497859, -6.54822862, 0, 6.32235644, 12.7202364, 19.1999489, 25.7678027, 32.4301067, 39.1931698, 46.063301, 53.046809, 60.1500028, 67.3791912, 74.7406833, 82.2407877, 89.8858135, 97.6820695, 105.635865, 113.753508, 122.134284, 130.797783, 139.540013, 148.387206, 157.365593, 166.501408, 175.820881, 185.350247, 195.115736, 205.143582, 215.460016, 226.091271, 237.063579, 248.842206, 260.853372, 272.991598, 285.151409, 297.44786, 309.665206, 321.528894, 332.867286, 343.508744, 353.281631, 362.014309, 369.309073, 375.289156, 379.827148, 382.795644, 384.067234, 388.14059, 389.105707, 385.678658, 376.575516, 360.512353, 337.670659, 303.510028, 254.670657, 188.769573, 105.418161, 3.3323795, -118.771811, -262.097012, -428.027213, -618.035251, -818.614615, -1063.02261, -1380.18734, -1787.5276, -2282.96033, -2873.35414, -3563.22042, -4358.81664, -5275.04346, -6321.91244, -7509.97984, -8848.62877, -10347.6823, -12016.9634, -13866.2951, -16454.6648, -19105.4402, -21691.1532, -24084.3357, -26157.5197, -27268.2796, -28147.41, -28888.1389, -29867.547, -31066.5039, -32347.6077, -33697.6415, -35103.3889, -36551.6331, -38029.1576, -39469.8617, -41038.9386, -42624.0878, -44113.0087, -45393.4007, -46352.9635, -46879.3963, -47728.544, -47901.6559},
        {-417.028928, -417.514161, -417.878814, -418.123093, -418.247208, -418.251366, -418.135775, -417.900643, -417.546178, -417.072588, -416.480081, -415.768866, -414.93915, -413.991141, -412.925047, -411.741076, -410.439437, -409.020336, -407.483984, -405.830586, -404.060352, -402.173488, -400.170205, -398.050708, -395.815207, -393.463909, -390.997023, -388.414756, -385.717316, -382.904912, -379.977751, -376.936042, -373.779991, -370.509809, -367.0631, -363.47269, -359.772157, -355.963741, -352.049682, -348.032219, -343.913592, -339.69604, -335.381803, -330.97312, -326.472231, -321.881376, -317.202794, -312.438724, -307.591407, -302.663082, -297.655988, -292.563597, -287.397107, -282.158947, -276.851547, -271.477336, -266.038745, -260.538203, -254.978139, -249.360983, -243.689165, -237.965114, -232.19126, -226.370033, -220.503862, -214.595178, -208.646408, -202.659603, -196.636821, -190.581275, -184.495422, -178.381714, -172.242607, -166.080555, -159.898012, -153.697432, -147.48127, -141.25198, -135.012017, -128.763834, -122.509887, -116.252629, -109.994515, -103.737999, -97.6572831, -91.6653348, -85.6730585, -79.6773349, -73.675045, -67.6630694, -61.638289, -55.5975846, -49.5378371, -43.4559272, -37.3487358, -31.2131436, -25.0460315, -18.8442802, -12.6047707, -6.3243837, 0, 6.30173087, 12.6227581, 18.9743511, 25.3677792, 31.8143118, 38.3252182, 44.9117678, 51.58523, 58.3568742, 65.2379697, 72.2397859, 79.3735922, 86.6506579, 94.0822523, 101.679645, 109.454105, 117.130301, 124.67856, 132.758414, 141.319719, 150.312329, 159.6861, 169.390886, 179.376542, 189.592924, 199.989885, 210.517281, 221.124966, 231.762797, 242.270763, 252.748867, 263.179922, 273.546742, 284.570291, 295.864307, 306.845688, 317.275803, 326.916019, 335.527705, 342.872228, 348.941721, 353.196555, 355.352951, 355.127127, 352.235302, 346.232717, 337.036814, 324.404056, 308.090906, 287.853828, 268.327248, 238.427712, 192.9064, 128.072989, 43.4190883, -62.9923164, -193.098238, -352.315697, -538.31079, -744.953241, -949.922619, -1195.20473, -1520.49388, -1945.66302, -2456.9989, -3056.20488, -3736.77321, -4498.27847, -5370.40255, -6365.79694, -7517.49979, -8814.6395, -10262.8106, -11867.6078, -13634.6255, -16048.3554, -18515.7706, -20922.7417, -23155.1389, -25098.8325, -26165.134, -27030.8452, -27785.2185, -28810.6468, -30080.348, -31445.3985, -32885.1234, -34378.8479, -35905.8969, -37445.5958, -38565.7276, -39985.9299, -41562.5282, -43151.8481, -44610.2152, -45793.955, -46559.3932, -47513.4502, -47366.9727},
        {-396.733209, -397.29269, -397.736785, -398.065648, -398.279434, -398.378298, -398.362395, -398.231879, -397.986904, -397.627626, -397.154199, -396.566778, -395.865517, -395.050572, -394.122095, -393.080244, -391.925171, -390.657032, -389.275981, -387.782173, -386.175762, -384.456904, -382.625752, -380.682462, -378.627189, -376.460086, -374.181308, -371.791011, -369.289348, -366.676475, -363.952545, -361.117715, -358.172137, -355.115968, -351.893738, -348.534583, -345.068308, -341.496874, -337.82224, -334.046365, -330.171209, -326.198733, -322.130896, -317.969657, -313.716976, -309.374814, -304.945129, -300.429882, -295.831032, -291.150539, -286.390364, -281.522086, -276.578702, -271.562832, -266.477094, -261.324105, -256.106484, -250.826849, -245.487819, -240.092011, -234.642045, -229.140538, -223.590109, -217.993376, -212.352957, -206.671471, -200.951535, -195.203402, -189.437141, -183.639626, -177.812981, -171.959328, -166.080791, -160.179493, -154.257559, -148.317111, -142.360273, -136.389168, -130.40592, -124.412652, -118.411487, -112.40455, -106.393962, -100.381849, -94.5212273, -88.7356, -82.9466554, -77.15162, -71.3477204, -65.5321833, -59.7022352, -53.8551027, -47.9880124, -42.0981909, -36.1828647, -30.2392606, -24.264605, -18.2561245, -12.2110458, -6.12659544, 0, 5.69744439, 11.4941356, 17.3946808, 23.4036869, 29.5257611, 35.7655104, 42.1275419, 48.6164626, 55.2368797, 61.9934002, 68.8906311, 75.9331796, 83.1256527, 90.4726575, 97.978801, 105.64869, 113.277657, 120.84041, 128.814921, 137.16095, 145.838261, 154.806615, 164.025775, 173.455502, 183.05556, 192.78571, 202.605715, 212.475336, 222.354337, 232.120226, 241.845177, 251.51363, 261.110022, 271.42826, 282.048048, 292.333231, 302.025406, 310.86617, 318.597121, 324.959856, 330.70414, 334.256569, 335.161491, 332.963252, 327.2062, 313.422449, 296.171637, 276.00117, 253.458455, 229.090895, 210.811814, 182.799914, 137.84943, 71.281318, -16.4247469, -127.105249, -262.596671, -425.964868, -617.537552, -836.301306, -1070.0194, -1346.47023, -1691.99079, -2120.60097, -2625.44993, -3209.26681, -3866.75777, -4598.57191, -5434.77609, -6388.79683, -7507.91797, -8766.49988, -10166.2492, -11708.8727, -13396.077, -15636.2796, -17922.7986, -20155.6631, -22234.9019, -24060.544, -25109.6006, -25987.1303, -26774.4556, -27827.7617, -29119.5711, -30507.8798, -31970.9367, -33486.9904, -35034.2897, -36591.0832, -37894.6073, -39350.6828, -40871.5048, -42369.2684, -43756.1689, -44944.4013, -45846.1609, -46775.074, -46882.7865},
        {-377.094882, -377.675563, -378.155306, -378.533801, -378.810737, -378.985804, -379.058689, -379.029084, -378.896677, -378.661157, -378.322214, -377.879538, -377.332816, -376.68174, -375.925997, -375.065279, -374.099272, -373.027668, -371.850155, -370.566423, -369.176161, -367.679058, -366.074804, -364.363088, -362.543599, -360.616026, -358.58006, -356.435389, -354.181702, -351.818689, -349.346039, -346.763442, -344.070587, -341.267163, -338.284559, -335.158052, -331.923738, -328.583523, -325.139313, -321.593013, -317.94653, -314.201768, -310.360634, -306.425034, -302.396874, -298.278058, -294.070493, -289.776085, -285.39674, -280.934363, -276.390859, -271.727813, -266.988326, -262.175178, -257.29115, -252.339022, -247.321572, -242.241582, -237.101831, -231.905099, -226.654166, -221.351812, -216.000816, -210.60396, -205.164021, -199.683781, -194.16602, -188.627129, -183.083173, -177.508859, -171.906084, -166.276745, -160.62274, -154.945967, -149.248322, -143.531704, -137.79801, -132.049138, -126.286984, -120.513447, -114.730423, -108.939811, -103.143508, -97.3434111, -91.6789423, -86.0803443, -80.4762068, -74.8639646, -69.241052, -63.6049038, -57.9529543, -52.2826383, -46.5913903, -40.8766447, -35.1358363, -29.3663995, -23.5657689, -17.7313791, -11.8606647, -5.9510601, 0, 4.66864002, 9.62048776, 14.8443451, 20.329014, 26.0632963, 32.0359939, 38.2359087, 44.6518427, 51.2725977, 58.0869756, 65.0837783, 72.2518078, 79.579866, 87.0567546, 94.6712757, 102.412231, 110.524707, 119.034118, 127.349574, 135.514796, 143.573504, 151.569417, 159.546255, 167.547739, 175.617589, 183.799524, 192.137265, 200.674531, 209.455042, 218.987676, 228.680438, 238.4375, 248.163038, 258.235276, 268.321362, 277.962027, 286.919227, 294.954921, 301.831064, 307.309613, 312.788972, 315.896602, 316.074286, 312.763807, 305.406947, 288.455016, 267.587887, 243.494963, 216.865643, 188.389329, 166.359733, 134.567786, 85.1276898, 13.2697857, -79.413463, -194.186056, -332.311992, -490.019675, -674.753155, -899.449854, -1174.57321, -1498.93904, -1869.99854, -2290.73456, -2773.50305, -3326.35756, -3954.17997, -4664.20153, -5475.28294, -6399.70662, -7489.11605, -8710.37704, -10062.147, -11543.0835, -13151.8438, -15226.3624, -17341.2004, -19410.1962, -21347.1879, -23066.0137, -24114.4581, -25016.4487, -25842.704, -26897.5098, -28159.1306, -29508.5113, -30928.7822, -32403.0739, -33914.5169, -35446.2417, -37359.2253, -39015.5457, -40452.6162, -41707.8502, -42818.6611, -43822.4623, -44756.6672, -45577.0698, -46434.0225},
        {-361.453756, -361.84382, -362.156946, -362.392113, -362.548301, -362.624487, -362.619651, -362.532773, -362.36283, -362.108802, -361.769668, -361.344407, -360.831998, -360.23142, -359.541651, -358.761672, -357.890461, -356.926996, -355.870257, -354.719223, -353.472873, -352.130185, -350.69014, -349.151715, -347.51389, -345.775644, -343.935955, -341.993804, -339.948168, -337.798026, -335.542359, -333.180144, -330.710361, -328.131988, -325.355797, -322.426726, -319.390773, -316.24978, -313.005588, -309.66004, -306.214976, -302.672239, -299.033669, -295.301109, -291.476399, -287.561381, -283.557897, -279.467789, -275.292897, -271.035063, -266.69613, -262.233448, -257.694313, -253.081531, -248.397909, -243.646252, -238.829366, -233.950058, -229.011132, -224.015396, -218.965656, -213.864716, -208.715384, -203.520465, -198.282764, -193.005089, -187.690246, -182.357381, -177.025253, -171.662961, -166.272253, -160.854874, -155.412571, -149.94709, -144.460179, -138.953584, -133.429051, -127.888327, -122.333158, -116.765291, -111.186473, -105.59845, -100.002969, -94.4017759, -88.9234385, -83.5036254, -78.0764809, -72.6396358, -67.1907213, -61.7273683, -56.2472077, -50.7478705, -45.2269878, -39.6821904, -34.1111093, -28.5113755, -22.88062, -17.2164737, -11.5165676, -5.77853273, 0, 4.32268638, 8.97067251, 13.9281763, 19.1794157, 24.7086086, 30.4999729, 36.5377266, 42.8060876, 49.2892738, 55.9715031, 62.8369935, 69.8699628, 77.0546291, 84.3752102, 91.815924, 99.3609885, 107.383907, 115.924508, 124.086486, 131.937476, 139.545114, 146.977037, 154.30088, 161.58428, 168.894873, 176.300295, 183.868181, 191.666169, 199.761894, 208.891421, 218.208866, 227.581339, 236.875947, 246.312959, 255.582902, 264.282128, 272.171796, 279.013067, 284.5671, 288.595055, 292.378872, 293.696084, 292.010307, 286.785158, 277.484252, 260.247758, 238.693601, 213.116259, 183.810211, 151.069935, 121.176114, 81.1205466, 24.4529802, -52.8289033, -149.729553, -267.497359, -407.380711, -562.891655, -744.773185, -972.207945, -1266.07954, -1612.08551, -1995.36015, -2413.86555, -2887.29975, -3425.38363, -4037.08088, -4731.91607, -5522.19018, -6418.63369, -7469.44994, -8642.13152, -9934.43755, -11344.1272, -12868.9595, -14801.2158, -16770.5025, -18700.9482, -20516.6814, -22141.8306, -23176.8469, -24085.3215, -24930.1021, -25980.2853, -27217.0785, -28535.752, -29921.8099, -31360.7565, -32838.096, -34339.3326, -36328.6387, -37995.4414, -39403.3197, -40615.8528, -41696.6198, -42709.1997, -43717.1717, -44385.4806, -45726.0864},
        {-350.461653, -350.446535, -350.384842, -350.274754, -350.114452, -349.902118, -349.635931, -349.314073, -348.934723, -348.496064, -347.996274, -347.433536, -346.806029, -346.111935, -345.349434, -344.516706, -343.611933, -342.633295, -341.578972, -340.447146, -339.235997, -337.943706, -336.568454, -335.10842, -333.561786, -331.926733, -330.201441, -328.38409, -326.472862, -324.465937, -322.361496, -320.15772, -317.852788, -315.444883, -312.824142, -310.045041, -307.163351, -304.180759, -301.098953, -297.919618, -294.644442, -291.275112, -287.813314, -284.260735, -280.619062, -276.889982, -273.075181, -269.176348, -265.195167, -261.133326, -256.992513, -252.729301, -248.391468, -243.981679, -239.502597, -234.956888, -230.347217, -225.676247, -220.946645, -216.161075, -211.322201, -206.432688, -201.495201, -196.512405, -191.486964, -186.421544, -181.318809, -176.196916, -171.073651, -165.919726, -160.736801, -155.526538, -150.290596, -145.030636, -139.748318, -134.445302, -129.12325, -123.783821, -118.428676, -113.059476, -107.67788, -102.285549, -96.8841437, -91.4753245, -86.178412, -80.9337616, -75.6803146, -70.4159131, -65.1383993, -59.8456152, -54.5354031, -49.2056049, -43.8540628, -38.4786189, -33.0771154, -27.6473943, -22.1872979, -16.6946681, -11.1673471, -5.60317702, 0, 4.89644429, 9.9644418, 15.1994507, 20.5969291, 26.1523352, 31.8611272, 37.7187632, 43.7207014, 49.8623999, 56.1393168, 62.5469104, 69.0806389, 75.7359602, 82.5083327, 89.3932145, 96.3860637, 103.49411, 110.714494, 118.016046, 125.396748, 132.85458, 140.387523, 147.993557, 155.670664, 163.416823, 171.230016, 179.108223, 187.049424, 195.051601, 203.422236, 211.736324, 219.898994, 227.815377, 235.988382, 244.024249, 251.369811, 257.750865, 262.893206, 266.52263, 268.364934, 268.712567, 266.552212, 261.4988, 253.16726, 241.172522, 225.821439, 205.864036, 180.742262, 149.898065, 112.773394, 71.6257021, 19.640261, -46.0583214, -127.954026, -227.731702, -347.434994, -489.107543, -650.110576, -838.234339, -1066.37717, -1346.29386, -1675.27246, -2051.88768, -2478.33211, -2962.92859, -3511.18607, -4125.78622, -4811.50501, -5583.48503, -6451.00491, -7452.55542, -8563.7089, -9783.63208, -11111.4917, -12546.4546, -14364.0407, -16217.9752, -18038.3364, -19755.2029, -21298.653, -22297.0103, -23183.2778, -24016.8576, -25060.8283, -26293.9699, -27610.1966, -28992.5314, -30423.9974, -31887.6175, -33366.4147, -34662.7636, -36066.2576, -37524.3219, -38984.3821, -40393.8635, -41700.1913, -42850.7911, -43370.6164, -44710.9896},
        {-341.290358, -340.833786, -340.361012, -339.869435, -339.356453, -338.819466, -338.255871, -337.663068, -337.038454, -336.37943, -335.683394, -334.947744, -334.169878, -333.347197, -332.477098, -331.55698, -330.584242, -329.556283, -328.470501, -327.324295, -326.115064, -324.840206, -323.49712, -322.083205, -320.595859, -319.032482, -317.390472, -315.667227, -313.860147, -311.96663, -309.984074, -307.90988, -305.741444, -303.476166, -300.983182, -298.32672, -295.572549, -292.72223, -289.777325, -286.739394, -283.61, -280.390701, -277.083061, -273.688639, -270.208998, -266.645697, -263.000299, -259.274363, -255.469452, -251.587127, -247.628948, -243.55545, -239.410111, -235.19538, -230.913708, -226.567545, -222.159342, -217.691549, -213.166616, -208.586995, -203.955135, -199.273486, -194.5445, -189.770627, -184.954317, -180.098021, -175.204188, -170.288518, -165.366392, -160.412936, -155.429742, -150.418401, -145.380505, -140.317645, -135.231414, -130.123403, -124.995203, -119.848407, -114.684604, -109.505389, -104.31235, -99.1070818, -93.8911742, -88.6662193, -83.5434867, -78.4674221, -73.3811529, -68.2827113, -63.1701297, -58.0414404, -52.8946758, -47.7278683, -42.53905, -37.3262534, -32.0875108, -26.8208546, -21.524317, -16.1959304, -10.8337271, -5.43573957, 0, 5.49856036, 11.009398, 16.5398931, 22.0974258, 27.6893761, 33.3231243, 39.0060504, 44.7455347, 50.5489571, 56.4236979, 62.3771372, 68.4166552, 74.5496319, 80.7834474, 87.125482, 93.5831158, 99.7551737, 105.590671, 112.0211, 118.966294, 126.346086, 134.080308, 142.088792, 150.291371, 158.607878, 166.958146, 175.262007, 183.439293, 191.409837, 199.002193, 206.260941, 213.133298, 219.566482, 226.323221, 232.942976, 238.747737, 243.440067, 246.722529, 248.297685, 247.868098, 244.863898, 239.342996, 231.061256, 219.774543, 205.238723, 191.746462, 173.382623, 148.768869, 116.526866, 75.2782763, 23.468093, -39.8894143, -114.314548, -200.874502, -303.794702, -425.882591, -569.945615, -738.040611, -933.896222, -1162.05993, -1421.74329, -1726.03274, -2090.81613, -2524.10917, -3020.58865, -3581.05415, -4201.21677, -4880.55684, -5637.21246, -6478.76783, -7431.50858, -8481.44035, -9629.6738, -10877.3196, -12225.4883, -13941.9749, -15694.5347, -17417.6072, -19045.6321, -20513.0489, -21467.5455, -22321.4809, -23132.1884, -24163.3273, -25390.5601, -26703.5798, -28082.3471, -29506.8227, -30956.9676, -32412.7424, -32858.6559, -33953.4587, -35491.5, -37267.1284, -39074.6932, -40708.543, -41963.027, -42395.8887, -43649.9829},
        {-330.401283, -329.732106, -329.064298, -328.394804, -327.72057, -327.03854, -326.34566, -325.638875, -324.91513, -324.171371, -323.404542, -322.611589, -321.789457, -320.935091, -320.045437, -319.11744, -318.148044, -317.134195, -316.072839, -314.96092, -313.795383, -312.573175, -311.29124, -309.946523, -308.535969, -307.056524, -305.505133, -303.878741, -302.174293, -300.388735, -298.519012, -296.562068, -294.51485, -292.374302, -289.993876, -287.445284, -284.801958, -282.065501, -279.237515, -276.319601, -273.313361, -270.220398, -267.042312, -263.780705, -260.437181, -257.013339, -253.510783, -249.931113, -246.275932, -242.546842, -238.745444, -234.844307, -230.874695, -226.83884, -222.738972, -218.577323, -214.356125, -210.077608, -205.744003, -201.357543, -196.920457, -192.434977, -187.903335, -183.327761, -178.710488, -174.053745, -169.359764, -164.642567, -159.915894, -155.157658, -150.369325, -145.552362, -140.708237, -135.838416, -130.944365, -126.027552, -121.089444, -116.131507, -111.155208, -106.162014, -101.153393, -96.1308097, -91.0957323, -86.0496274, -81.0949446, -76.1805351, -71.2540369, -66.3136396, -61.357533, -56.3839068, -51.3909509, -46.3768549, -41.3398086, -36.2780019, -31.1896243, -26.0728658, -20.9259161, -15.7469649, -10.534202, -5.28581707, 0, 4.85163108, 9.81576679, 14.8919644, 20.0797814, 25.3787749, 30.7885023, 36.308521, 41.9383882, 47.6776612, 53.5258975, 59.4826542, 65.5474887, 71.7199584, 77.9996206, 84.3860325, 90.8787514, 97.1709062, 103.218278, 109.713538, 116.590579, 123.783295, 131.225581, 138.85133, 146.594437, 154.388795, 162.168298, 169.866842, 177.418318, 184.756623, 191.892448, 198.654729, 204.954321, 210.702077, 216.291001, 221.390873, 225.542899, 228.51329, 230.068255, 229.974003, 227.996746, 224.47772, 218.433099, 209.516586, 197.381885, 181.682702, 165.845068, 144.807276, 117.27995, 81.9737089, 37.5991764, -16.9085823, -83.3377095, -162.203561, -254.842516, -364.26721, -492.737675, -642.513942, -814.261961, -1012.19814, -1242.27786, -1505.69765, -1812.46193, -2174.42768, -2598.37362, -3082.47367, -3628.85148, -4237.30644, -4909.35961, -5655.05442, -6479.61355, -7400.14034, -8408.14905, -9504.74984, -10691.0529, -11968.1683, -13578.9255, -15222.2856, -16838.929, -18369.5361, -19754.7872, -20671.7857, -21500.507, -22294.5931, -23302.491, -24500.7324, -25784.68, -27134.9237, -28532.0535, -29956.6594, -31389.3315, -31433.1952, -32419.921, -34045.2149, -36004.7834, -37994.3326, -39709.5688, -40846.1982, -41416.989, -42916.6905},
        {-319.712611, -318.907765, -318.116729, -317.336133, -316.562609, -315.792789, -315.023304, -314.250786, -313.471865, -312.683173, -311.881342, -311.063002, -310.224786, -309.363325, -308.47525, -307.557192, -306.605783, -305.617655, -304.589437, -303.517763, -302.399264, -301.23057, -300.008313, -298.729125, -297.389636, -295.98648, -294.516285, -292.975685, -291.36131, -289.669792, -287.897763, -286.041853, -284.098694, -282.064917, -279.78115, -277.325306, -274.777177, -272.138456, -269.410838, -266.596016, -263.695684, -260.711537, -257.645269, -254.498573, -251.273144, -247.970676, -244.592862, -241.141397, -237.617975, -234.02429, -230.362035, -226.618058, -222.809222, -218.937542, -215.005035, -211.013717, -206.965603, -202.86271, -198.707053, -194.500648, -190.245511, -185.943658, -181.597105, -177.207868, -172.777963, -168.309405, -163.804211, -159.275133, -154.734668, -150.162686, -145.560507, -140.929451, -136.270838, -131.585988, -126.876219, -122.142853, -117.38721, -112.610608, -107.814368, -102.999809, -98.1682524, -93.3210169, -88.4594227, -83.5847894, -78.7914204, -74.032188, -69.258847, -64.4696999, -59.663049, -54.8371966, -49.9904451, -45.1210969, -40.2274543, -35.3078197, -30.3604955, -25.383784, -20.3759876, -15.3354086, -10.2603494, -5.14911246, 0, 3.86318046, 8.01421801, 12.438984, 17.1233498, 22.0531868, 27.2143663, 32.5927597, 38.1742383, 43.9446735, 49.8899366, 55.995899, 62.2484321, 68.6334072, 75.1366956, 81.7441688, 88.441698, 95.1775675, 101.932279, 108.776723, 115.688715, 122.646073, 129.626614, 136.608155, 143.568513, 150.485505, 157.336947, 164.100657, 170.754451, 177.276147, 184.071331, 190.533202, 196.511246, 201.854949, 206.395105, 209.990545, 212.505089, 213.793829, 213.71186, 212.114277, 208.856172, 205.750972, 200.099426, 191.373477, 179.045065, 162.586135, 142.827844, 117.543114, 85.8640859, 46.9228976, -0.148311931, -54.4052413, -120.74278, -201.984811, -300.290407, -416.461312, -551.908683, -708.043677, -882.714131, -1081.70437, -1314.68596, -1588.54194, -1905.17513, -2266.53503, -2676.24888, -3140.78674, -3665.3138, -4256.72538, -4920.63522, -5656.31306, -6466.6172, -7361.06858, -8334.83776, -9388.47667, -10522.5373, -11737.5715, -13249.5373, -14789.7292, -16304.8475, -17741.5927, -19046.6653, -19925.6216, -20727.0688, -21501.0547, -22477.9844, -23636.5142, -24880.152, -26191.3119, -27552.4078, -28945.8538, -30354.0638, -30198.5442, -31157.2234, -32858.3625, -34930.2225, -37001.0645, -38699.1496, -39652.7388, -40457.1687, -42268.3441},
        {-309.752308, -308.843064, -307.957392, -307.091688, -306.242349, -305.40577, -304.578347, -303.756477, -302.936556, -302.11498, -301.288145, -300.452446, -299.604281, -298.740045, -297.856134, -296.948944, -296.014872, -295.050313, -294.051664, -293.01532, -291.937678, -290.815134, -289.644084, -288.420924, -287.14205, -285.803859, -284.402745, -282.935106, -281.397337, -279.785835, -278.096995, -276.327214, -274.472888, -272.530413, -270.330061, -267.954785, -265.489533, -262.936092, -260.296249, -257.571792, -254.764508, -251.876183, -248.908605, -245.863561, -242.742838, -239.548223, -236.281504, -232.944467, -229.538899, -226.066588, -222.52932, -218.926973, -215.263285, -211.540085, -207.759202, -203.922463, -200.031698, -196.088735, -192.095402, -188.053527, -183.964941, -179.83147, -175.654944, -171.437191, -167.180039, -162.885317, -158.554854, -154.200836, -149.8352, -145.438413, -141.011632, -136.556014, -132.072716, -127.562895, -123.027708, -118.468312, -113.885865, -109.281523, -104.656443, -100.011783, -95.3487001, -90.6683504, -85.9718914, -81.2604802, -76.6215219, -72.0112354, -67.384803, -62.7405829, -58.0769333, -53.3922122, -48.6847779, -43.9529886, -39.1952022, -34.4097771, -29.5950713, -24.7494431, -19.8712505, -14.9588518, -10.010605, -5.02486835, 0, 2.88408078, 6.23661404, 10.029063, 14.2328909, 18.819561, 23.7605365, 29.0272806, 34.5912566, 40.4239276, 46.4967569, 52.7812078, 59.2487434, 65.870827, 72.6189218, 79.4644911, 86.378998, 93.6107805, 101.170856, 108.404422, 115.342273, 122.015202, 128.454004, 134.689471, 140.752397, 146.673575, 152.483799, 158.213863, 163.89456, 169.556684, 176.060613, 182.303375, 188.066886, 193.133067, 196.749153, 198.964404, 199.970659, 199.710239, 198.125467, 195.158664, 190.752152, 188.268556, 183.188933, 174.786416, 162.334138, 145.105232, 120.745819, 90.5627948, 54.2360444, 11.4454528, -38.1290951, -91.0467017, -155.985288, -238.180012, -340.725855, -462.348261, -603.734506, -765.571868, -943.168783, -1143.81383, -1380.66342, -1666.49152, -1994.52708, -2356.03787, -2751.21535, -3195.37168, -3696.87838, -4269.72729, -4923.74704, -5648.15841, -6443.83916, -7314.67253, -8258.06839, -9273.86411, -10361.8971, -11522.0047, -12946.1011, -14393.9278, -15817.3027, -17168.0442, -18397.9703, -19236.6489, -20006.3153, -20754.0378, -21693.3049, -22805.3102, -24001.9056, -25267.752, -26587.5101, -27945.8406, -29327.4043, -29146.4055, -30129.8786, -31875.7954, -33982.1277, -36046.8474, -37667.9264, -38443.3366, -39522.4672, -41619.2134},
        {-301.815533, -300.726726, -299.674265, -298.65425, -297.662778, -296.695949, -295.749863, -294.820619, -293.904316, -292.997053, -292.09493, -291.194046, -290.290499, -289.38039, -288.459817, -287.524879, -286.571677, -285.596309, -284.594874, -283.563472, -282.498201, -281.395162, -280.250453, -279.060173, -277.820423, -276.5273, -275.176905, -273.765336, -272.288693, -270.743074, -269.12458, -267.42931, -265.653362, -263.792836, -261.670279, -259.372216, -256.98726, -254.517142, -251.963593, -249.328344, -246.613127, -243.819674, -240.949714, -238.004981, -234.987205, -231.898117, -228.739449, -225.512932, -222.220297, -218.863277, -215.443601, -211.960663, -208.418584, -204.819145, -201.164129, -197.455318, -193.694493, -189.883437, -186.023932, -182.117759, -178.166702, -174.172541, -170.137059, -166.062039, -161.949261, -157.800508, -153.617562, -149.414789, -145.20625, -140.967777, -136.700336, -132.404893, -128.082415, -123.733868, -119.360218, -114.962431, -110.541475, -106.098315, -101.633917, -97.1492491, -92.6452761, -88.1229647, -83.5832814, -79.0271925, -74.5394805, -70.0774407, -65.597361, -61.0974877, -56.5760671, -52.0313455, -47.4615694, -42.864985, -38.2398387, -33.5843768, -28.8968457, -24.1754916, -19.4185611, -14.6243003, -9.7909556, -4.9167734, 0, 2.86454501, 6.20231324, 9.98243805, 14.1740528, 18.7462908, 23.6682854, 28.9091701, 34.438078, 40.2241427, 46.2364974, 52.4442755, 58.8166104, 65.3226354, 71.9314838, 78.6122891, 85.3341845, 92.4442054, 99.9654712, 107.012337, 113.634915, 119.883317, 125.807657, 131.458046, 136.884596, 142.13742, 147.266631, 152.322341, 157.354661, 162.413706, 168.55902, 174.461157, 179.867397, 184.525024, 187.507989, 188.900238, 188.965273, 187.652376, 184.910828, 180.689908, 174.938899, 171.332563, 164.960858, 155.044163, 140.802863, 121.457339, 93.470197, 59.5090394, 19.4836928, -26.6960164, -79.1202619, -133.739006, -199.842892, -281.872579, -382.855086, -502.931251, -643.537749, -806.111254, -988.781496, -1197.04367, -1440.00055, -1727.02923, -2053.88929, -2413.98023, -2808.00294, -3249.03005, -3744.45539, -4306.65268, -4944.30685, -5647.84297, -6418.01493, -7257.92694, -8165.62093, -9141.03721, -10184.1161, -11294.7978, -12658.8223, -14043.8803, -15403.4624, -16691.0587, -17860.1599, -18646.7083, -19366.7742, -20067.0827, -20957.7309, -22020.776, -23170.2034, -24391.4741, -25670.0487, -26991.3881, -28340.9529, -28353.3525, -29379.6793, -31066.0395, -33058.5396, -35003.2858, -36546.3844, -37333.9418, -38630.748, -40746.3002},
        {-294.951689, -293.678681, -292.454174, -291.273996, -290.133976, -289.029941, -287.95772, -286.913141, -285.892031, -284.89022, -283.903534, -282.927801, -281.958851, -280.99251, -280.024608, -279.050972, -278.067429, -277.069809, -276.053939, -275.015648, -273.950762, -272.855111, -271.724522, -270.554824, -269.341844, -268.081411, -266.769352, -265.401496, -263.973671, -262.481704, -260.921424, -259.288659, -257.579237, -255.788986, -253.73466, -251.50539, -249.192461, -246.797512, -244.322182, -241.768111, -239.136938, -236.430302, -233.649843, -230.797199, -227.874011, -224.881918, -221.822558, -218.697572, -215.508599, -212.257277, -208.945247, -205.56807, -202.133594, -198.643591, -195.099832, -191.504089, -187.858131, -184.163731, -180.42266, -176.636688, -172.807587, -168.937127, -165.027081, -161.079218, -157.095311, -153.07713, -149.026446, -144.960391, -140.895726, -136.802544, -132.681621, -128.533732, -124.359652, -120.160156, -115.936021, -111.68802, -107.416931, -103.123527, -98.8085855, -94.4728804, -90.1171877, -85.7422826, -81.3489405, -76.9379368, -72.5924122, -68.270227, -63.9282515, -59.5645882, -55.1773395, -50.764608, -46.3244961, -41.8551063, -37.3545411, -32.8209029, -28.2522941, -23.6468174, -19.0025751, -14.3176698, -9.59020382, -4.81827974, 0, 3.1271997, 6.68012953, 10.6292817, 14.9451486, 19.5982224, 24.5589953, 29.7979596, 35.2856075, 40.9924314, 46.8889235, 52.945576, 59.1328811, 65.4213312, 71.7814185, 78.1836353, 84.5984737, 91.4030305, 98.6258842, 105.317827, 111.53648, 117.339465, 122.784405, 127.92892, 132.830633, 137.547166, 142.136139, 146.655176, 151.161897, 155.713924, 161.475168, 166.991322, 171.988124, 176.191307, 178.630316, 179.379031, 178.697013, 176.518885, 172.77927, 167.412791, 160.354072, 155.282666, 147.248504, 135.453508, 119.099596, 97.3886882, 66.0932707, 28.7020554, -14.7256797, -64.1306564, -119.453596, -176.529673, -244.42305, -325.933084, -423.751042, -540.347501, -678.292121, -840.154566, -1028.20412, -1245.37908, -1494.94546, -1780.80587, -2103.70349, -2461.9325, -2857.55536, -3299.01974, -3791.8424, -4344.78848, -4964.21689, -5644.57587, -6387.05099, -7195.75805, -8068.50164, -9005.45302, -10006.7835, -11072.6643, -12387.3325, -13720.8773, -15027.4534, -16261.2157, -17376.3191, -18110.2769, -18778.3132, -19427.4294, -20267.047, -21279.8758, -22380.9508, -23556.3655, -24792.2136, -26074.5888, -27389.5848, -27727.3763, -28821.1335, -30403.7517, -32208.1269, -33967.1544, -35413.73, -36280.7493, -37766.2718, -39806.4525},
        {-289.008112, -287.553911, -286.159431, -284.820262, -283.531992, -282.290211, -281.090507, -279.92847, -278.799688, -277.699751, -276.624246, -275.568765, -274.528894, -273.500225, -272.478344, -271.458842, -270.437307, -269.409329, -268.370495, -267.316397, -266.242621, -265.144758, -264.018396, -262.859124, -261.662532, -260.424208, -259.139741, -257.80472, -256.414735, -254.965374, -253.452226, -251.870881, -250.216926, -248.485952, -246.490459, -244.321431, -242.072053, -239.743855, -237.338368, -234.857123, -232.301649, -229.673478, -226.97414, -224.205166, -221.368086, -218.46443, -215.49573, -212.463515, -209.369317, -206.214666, -203.001092, -199.719175, -196.381634, -192.990236, -189.546751, -186.052945, -182.510587, -178.921444, -175.287286, -171.609878, -167.890991, -164.132391, -160.335846, -156.503125, -152.635995, -148.736224, -144.805581, -140.863884, -136.93052, -132.970026, -128.983002, -124.970043, -120.931749, -116.868717, -112.781544, -108.670828, -104.537168, -100.38116, -96.2034034, -92.0044948, -87.7850323, -83.5456136, -79.2868365, -75.0092988, -70.7944309, -66.6007138, -62.3856574, -58.1472362, -53.8834249, -49.5921979, -45.2715299, -40.9193954, -36.5337691, -32.1126254, -27.6539391, -23.1556846, -18.6158365, -14.0323694, -9.40325791, -4.72647658, 0, 3.57296172, 7.48722135, 11.7174582, 16.2383517, 21.0245811, 26.0508258, 31.2917651, 36.7220783, 42.3164448, 48.049544, 53.8960552, 59.8306577, 65.8280308, 71.8628539, 77.9098064, 83.9435676, 90.3007276, 97.0076672, 103.220458, 108.991332, 114.372521, 119.416256, 124.174768, 128.700291, 133.045054, 137.26129, 141.401231, 145.517108, 149.661153, 154.998403, 160.060255, 164.5651, 168.231327, 170.144528, 170.339492, 169.019756, 166.09355, 161.469105, 155.05465, 146.758417, 140.089876, 130.259987, 116.472389, 97.9307214, 73.8386236, 39.7204188, -0.621108035, -47.0624883, -99.4802534, -157.750935, -218.034925, -288.466842, -369.703223, -463.909423, -576.331301, -710.831631, -871.273188, -1064.58259, -1290.86244, -1546.87297, -1830.24055, -2147.68713, -2503.51275, -2901.50241, -3344.56832, -3836.02331, -4380.27181, -4980.90965, -5637.53019, -6351.9908, -7130.26691, -7969.46432, -8870.01492, -9832.35062, -10856.9033, -12127.9144, -13416.0539, -14675.8014, -15861.6365, -16928.0385, -17612.8725, -18231.6424, -18831.6627, -19621.8321, -20585.5047, -21638.7073, -22768.1603, -23960.5839, -25202.6983, -26481.2237, -27236.3664, -28413.9742, -29861.4105, -31426.0392, -32955.2239, -34296.3282, -35296.7159, -36925.2384, -38821.557},
        {-283.726184, -282.114162, -280.5711, -279.092405, -277.673484, -276.309743, -274.996588, -273.729426, -272.503664, -271.314708, -270.157964, -269.028839, -267.92274, -266.835073, -265.761245, -264.696662, -263.63673, -262.576856, -261.512447, -260.438909, -259.351648, -258.246071, -257.117585, -255.961597, -254.773511, -253.548736, -252.282677, -250.970741, -249.608335, -248.190865, -246.713738, -245.172359, -243.562136, -241.878475, -239.930178, -237.809878, -235.612453, -233.339364, -230.992074, -228.572044, -226.080739, -223.519618, -220.890145, -218.193782, -215.43199, -212.606233, -209.717973, -206.76867, -203.759789, -200.69279, -197.569137, -194.379964, -191.137283, -187.842783, -184.498148, -181.105064, -177.665219, -174.180297, -170.651985, -167.081969, -163.471934, -159.823568, -156.138556, -152.418585, -148.665339, -144.880506, -141.065772, -137.241521, -133.427689, -129.587398, -125.721122, -121.829333, -117.912507, -113.971117, -110.005636, -106.01654, -102.004301, -97.9693933, -93.9122914, -89.8334689, -85.7333997, -81.6125577, -77.4714167, -73.3104507, -69.2058572, -65.1186556, -61.0089549, -56.8747719, -52.7141228, -48.5250243, -44.3054928, -40.053545, -35.7671972, -31.444466, -27.083368, -22.6819195, -18.2381373, -13.7500377, -9.21563725, -4.63295252, 0, 4.13943923, 8.49606775, 13.0527885, 17.7925046, 22.6981189, 27.7525344, 32.9386541, 38.239381, 43.6376181, 49.1162683, 54.6582347, 60.2464202, 65.8637278, 71.4930605, 77.1173213, 82.7194131, 88.4295465, 94.2516684, 99.8354862, 105.195469, 110.346085, 115.301804, 120.077094, 124.686424, 129.144264, 133.465081, 137.663346, 141.753526, 145.750091, 150.593674, 155.032986, 158.804646, 161.645274, 162.703463, 162.009847, 159.751864, 155.842544, 150.194914, 142.722002, 133.336836, 125.818005, 115.0365, 100.149043, 80.3123576, 54.6831652, 18.6481247, -23.9234619, -72.9323565, -128.279321, -189.865116, -254.568198, -329.005566, -411.604506, -503.893057, -613.729962, -746.131605, -906.114373, -1104.69209, -1339.52065, -1601.71328, -1884.00001, -2197.16247, -2548.58283, -2943.08591, -3381.16325, -3865.29545, -4398.43103, -4983.1719, -5620.40423, -6311.98474, -7064.24404, -7873.8766, -8741.19045, -9666.49361, -10650.0941, -11866.7227, -13098.659, -14302.6053, -15435.264, -16453.3375, -17105.8618, -17695.6496, -18268.4032, -19023.694, -19946.3176, -20956.9571, -22043.6138, -23194.2887, -24396.983, -25639.6977, -26940.7128, -28234.8443, -29517.8602, -30785.5286, -32033.6175, -33257.8949, -34454.1289, -36099.1556, -37832.6628},
        {-278.9773, -277.23559, -275.569946, -273.975651, -272.447987, -270.982237, -269.573685, -268.217614, -266.909306, -265.644044, -264.417112, -263.223792, -262.059367, -260.91912, -259.798335, -258.692293, -257.596279, -256.505575, -255.415464, -254.321228, -253.218152, -252.101518, -250.966608, -249.808706, -248.623095, -247.405057, -246.149877, -244.852835, -243.509217, -242.114304, -240.663379, -239.151726, -237.574627, -235.927365, -234.015338, -231.932766, -229.776149, -227.546935, -225.246567, -222.876491, -220.438152, -217.932995, -215.362465, -212.728008, -210.031068, -207.27309, -204.455521, -201.579804, -198.647385, -195.659709, -192.618221, -189.521326, -186.373576, -183.17648, -179.931551, -176.6403, -173.304237, -169.924873, -166.50372, -163.042289, -159.54209, -156.004635, -152.431434, -148.824, -145.183842, -141.512472, -137.8114, -134.099104, -130.393652, -126.661565, -122.903255, -119.119133, -115.30961, -111.475098, -107.616008, -103.732751, -99.8257382, -95.895381, -91.9420907, -87.9662786, -83.9683558, -79.9487336, -75.9078233, -71.8460362, -67.830312, -63.8263985, -59.7992435, -55.7470994, -51.6682185, -47.560853, -43.4232551, -39.2536773, -35.0503716, -30.8115905, -26.5355862, -22.2206109, -17.864917, -13.4667566, -9.02438218, -4.53604588, 0, 4.71921713, 9.51126479, 14.3699758, 19.2891828, 24.2627186, 29.284416, 34.3481078, 39.4476267, 44.5768055, 49.7294769, 54.8994737, 60.0806287, 65.2667747, 70.4517443, 75.6293704, 80.7934857, 85.72132, 90.3757628, 95.2405821, 100.263196, 105.391022, 110.571478, 115.751982, 120.879952, 125.902805, 130.76796, 135.422834, 139.814845, 143.891412, 148.16066, 151.803705, 154.599753, 156.328008, 156.193715, 154.263059, 150.755282, 145.621778, 138.813941, 130.283163, 119.980838, 112.416992, 101.596977, 86.5802799, 66.4263877, 40.1947879, 3.25712034, -40.7173184, -91.7470793, -149.850713, -215.046771, -285.26864, -365.168567, -451.187587, -544.182401, -654.025029, -786.539795, -947.551025, -1150.85442, -1392.49126, -1659.80675, -1943.08461, -2254.12002, -2599.34297, -2983.56959, -3408.88144, -3878.61529, -4397.8578, -4970.3994, -5593.61433, -6268.50616, -6999.6339, -7783.89877, -8621.07366, -9510.93144, -10453.245, -11602.3398, -12764.7979, -13901.7543, -14974.3436, -15943.7006, -16582.5177, -17166.0002, -17736.0442, -18472.978, -19363.9013, -20338.2332, -21385.9574, -22497.0574, -23661.5168, -24869.319, -26787.2995, -28216.9286, -29321.4254, -30264.009, -31207.8986, -32316.3133, -33752.4723, -35285.2887, -36857.3052},
        {-274.797779, -272.93949, -271.163217, -269.464153, -267.83749, -266.278422, -264.782141, -263.343839, -261.958711, -260.621948, -259.328742, -258.074288, -256.853778, -255.662403, -254.495358, -253.347835, -252.215026, -251.092125, -249.974323, -248.856815, -247.734792, -246.603448, -245.457974, -244.293565, -243.105412, -241.888708, -240.638647, -239.35042, -238.019221, -236.640242, -235.208676, -233.719716, -232.168554, -230.550384, -228.668159, -226.617235, -224.495278, -222.303719, -220.04399, -217.717522, -215.325746, -212.870093, -210.351995, -207.772883, -205.134189, -202.437342, -199.683776, -196.874921, -194.012208, -191.097069, -188.130934, -185.120378, -182.061578, -178.955853, -175.804525, -172.608911, -169.370333, -166.090109, -162.769561, -159.410007, -156.012767, -152.579161, -149.11051, -145.608132, -142.073348, -138.507477, -134.911839, -131.302323, -127.694466, -124.059543, -120.397929, -116.709998, -112.996127, -109.256692, -105.492066, -101.702626, -97.8887474, -94.0508051, -90.1891747, -86.3042316, -82.3963511, -78.4659086, -74.5132796, -70.5388394, -66.599487, -62.666147, -58.7090649, -54.7267817, -50.7178385, -46.6807764, -42.6141365, -38.51646, -34.3862878, -30.2221611, -26.022621, -21.7862086, -17.511465, -13.1969312, -8.84114838, -4.44265762, 0, 5.14776424, 10.2512928, 15.3135034, 20.3373134, 25.3256405, 30.2814022, 35.2075162, 40.1069, 44.9824712, 49.8371473, 54.6738459, 59.4954846, 64.304981, 69.1052527, 73.8992171, 78.689792, 82.9192491, 86.5104139, 90.734331, 95.4737797, 100.611539, 106.030388, 111.613107, 117.242474, 122.801268, 128.17227, 133.238257, 137.88201, 141.986307, 145.613349, 148.400707, 150.177332, 150.772178, 149.529823, 146.521407, 141.947237, 135.781579, 127.998697, 118.572856, 107.478323, 99.692327, 88.6635271, 73.387347, 52.8592109, 26.0745428, -11.0393491, -55.6518956, -108.000643, -168.32314, -236.856931, -312.860064, -398.746752, -489.435539, -585.153416, -696.965448, -831.070629, -993.667947, -1199.74186, -1444.70919, -1714.40265, -1998.76924, -2308.71207, -2648.27402, -3021.70784, -3433.6219, -3888.46133, -4393.96943, -4955.44642, -5566.0992, -6225.97534, -6937.30765, -7697.62226, -8506.21031, -9362.36294, -10265.3713, -11349.1133, -12444.6467, -13517.6157, -14533.6648, -15458.4384, -16089.2752, -16672.3292, -17245.3759, -17968.4281, -18830.8756, -19771.1762, -20781.2666, -21853.0831, -22978.5626, -24149.6413, -26579.0178, -28111.4221, -29054.8544, -29717.3153, -30406.8049, -31431.3236, -33098.8718, -34488.2078, -35883.1893},
        {-271.120774, -269.156366, -267.279041, -265.483926, -263.76615, -262.120841, -260.543125, -259.028132, -257.570988, -256.166821, -254.810759, -253.497931, -252.223463, -250.982483, -249.77012, -248.5815, -247.411752, -246.256004, -245.109383, -243.967017, -242.824034, -241.675561, -240.516727, -239.342659, -238.148485, -236.929332, -235.680329, -234.396603, -233.073281, -231.705493, -230.288365, -228.817025, -227.286601, -225.692221, -223.835444, -221.812254, -219.72096, -217.562973, -215.3397, -213.052553, -210.702939, -208.29227, -205.821953, -203.293399, -200.708017, -198.067216, -195.372406, -192.624997, -189.826397, -186.978016, -184.081264, -181.150341, -178.173587, -175.152136, -172.087119, -168.979669, -165.830918, -162.641998, -159.414041, -156.148179, -152.845544, -149.507269, -146.134486, -142.728327, -139.289923, -135.820408, -132.320913, -128.804512, -125.283991, -121.735855, -118.160461, -114.558168, -110.929334, -107.274317, -103.593475, -99.8871664, -96.1557485, -92.3995798, -88.6190182, -84.814422, -80.986149, -77.1345575, -73.2600053, -69.3628507, -65.490068, -61.6177264, -57.7213353, -53.7997398, -49.8517852, -45.8763168, -41.8721798, -37.8382196, -33.7732813, -29.6762104, -25.545852, -21.3810514, -17.1806538, -12.9435047, -8.66844912, -4.35433247, 0, 5.2962775, 10.4885006, 15.5842904, 20.5912677, 25.5170535, 30.3692688, 35.1555346, 39.8834717, 44.5607012, 49.1948441, 53.7935213, 58.3643537, 62.9149623, 67.4529682, 71.9859921, 76.5216552, 80.2589955, 83.0901221, 86.8515935, 91.3777627, 96.5029831, 102.061608, 107.88799, 113.816483, 119.68144, 125.317214, 130.558159, 135.238627, 139.192972, 142.107851, 144.019469, 144.806486, 144.347566, 142.234382, 138.48909, 133.210376, 126.362998, 117.911715, 107.821284, 96.0564642, 87.5398569, 75.7694683, 59.7400436, 38.4463285, 10.8830682, -25.7207, -70.1630954, -123.007945, -184.819078, -256.16032, -337.82395, -429.866127, -526.435406, -627.247428, -743.38397, -880.823646, -1045.54507, -1251.56063, -1494.98931, -1763.18598, -2048.41241, -2358.49612, -2693.93665, -3057.82473, -3457.6417, -3898.85354, -4391.65819, -4942.74843, -5541.46651, -6186.96915, -6878.80783, -7615.68371, -8396.61682, -9220.6272, -10086.7349, -11111.5294, -12147.0688, -13162.981, -14128.8934, -15014.4339, -15639.3146, -16223.0222, -16799.4339, -17509.7899, -18345.4694, -19253.1169, -20226.3484, -21258.7797, -22344.0268, -23475.7055, -26208.8314, -27783.3607, -28604.4696, -29077.3344, -29607.1313, -30599.0367, -32458.2268, -33707.9137, -34916.68},
        {-267.841247, -265.778504, -263.807343, -261.922844, -260.120089, -258.394157, -256.740128, -255.153083, -253.628101, -252.160263, -250.744649, -249.376339, -248.050414, -246.761952, -245.506035, -244.277743, -243.072156, -241.884353, -240.709416, -239.542424, -238.378458, -237.212597, -236.039922, -234.855512, -233.654449, -232.431812, -231.182682, -229.902138, -228.585261, -227.227131, -225.822827, -224.367431, -222.856022, -221.283681, -219.452157, -217.457264, -215.397134, -213.273123, -211.086585, -208.838873, -206.531342, -204.165346, -201.742239, -199.263375, -196.730108, -194.143792, -191.505782, -188.817431, -186.080093, -183.295122, -180.463873, -177.604922, -174.702027, -171.75617, -168.76833, -165.739488, -162.670626, -159.562724, -156.416762, -153.233722, -150.014584, -146.760329, -143.471937, -140.15039, -136.796667, -133.41175, -129.99662, -126.561944, -123.11816, -119.646267, -116.146617, -112.619564, -109.065461, -105.48466, -101.877513, -98.2443743, -94.5855955, -90.9015296, -87.1925293, -83.4589474, -79.7011364, -75.9194492, -72.1142385, -68.2858568, -64.472816, -60.6555865, -56.8141804, -52.9477121, -49.0552959, -45.1360462, -41.1890774, -37.2135038, -33.2084399, -29.1729999, -25.1062984, -21.0074496, -16.8755679, -12.7097677, -8.50916343, -4.2728694, 0, 4.89210627, 9.73850287, 14.5420773, 19.3057172, 24.0323099, 28.724743, 33.385904, 38.0186804, 42.6259598, 47.2106296, 51.7755775, 56.3236907, 60.857857, 65.3809639, 69.8958987, 74.4055491, 78.0922455, 80.841653, 84.5126809, 88.9323833, 93.9278139, 99.3260266, 104.954075, 110.639014, 116.207896, 121.487775, 126.305706, 130.488743, 133.863938, 135.948932, 136.993644, 136.917954, 135.641739, 133.206869, 129.472225, 124.264161, 117.465957, 108.960895, 98.6322542, 86.3633168, 75.8953573, 62.0794909, 44.0441738, 20.9178622, -8.17098794, -43.3675916, -86.4514035, -138.47555, -200.493157, -273.55735, -360.198738, -458.187154, -562.484302, -672.590939, -797.275053, -941.143835, -1108.80448, -1309.12053, -1541.47546, -1800.60946, -2086.09868, -2398.73785, -2733.86378, -3092.34439, -3484.41978, -3916.02804, -4398.65638, -4939.68156, -5526.13368, -6156.55206, -6827.18012, -7539.2063, -8291.96466, -9084.78926, -9917.01416, -10897.3476, -11888.406, -12862.1797, -13790.6593, -14645.8352, -15261.998, -15840.6381, -16412.7605, -17106.2913, -17913.8707, -18787.7547, -19722.6025, -20713.0734, -21753.8266, -22839.5214, -25480.0272, -26980.5468, -27752.3806, -28206.8287, -28755.1915, -29808.7691, -31778.8619, -32954.5514, -33975.0512},
        {-264.87853, -262.722575, -260.662424, -258.693117, -256.809696, -255.007204, -253.280682, -251.625172, -250.035717, -248.507357, -247.035135, -245.614092, -244.239272, -242.905714, -241.608462, -240.342558, -239.103043, -237.884958, -236.683347, -235.493251, -234.309711, -233.12777, -231.94247, -230.748852, -229.541959, -228.316832, -227.068512, -225.792043, -224.482466, -223.134823, -221.744156, -220.305506, -218.813916, -217.264427, -215.460817, -213.497783, -211.47232, -209.385679, -207.239107, -205.033854, -202.771167, -200.452297, -198.078491, -195.650999, -193.171069, -190.63995, -188.058892, -185.429142, -182.751949, -180.028563, -177.260232, -174.464759, -171.62648, -168.746285, -165.825064, -162.863706, -159.863102, -156.824141, -153.747714, -150.63471, -147.48602, -144.302533, -141.085139, -137.834728, -134.552191, -131.238416, -127.894295, -124.529019, -121.151582, -117.745751, -114.311876, -110.85031, -107.361404, -103.845511, -100.30298, -96.7341655, -93.1394177, -89.5190886, -85.8735299, -82.2030934, -78.5081307, -74.7889935, -71.0460336, -67.2796027, -63.5222724, -59.7576068, -55.9687826, -52.1551059, -48.3158827, -44.4504191, -40.5580213, -36.6379953, -32.6896473, -28.7122833, -24.7052094, -20.6677317, -16.5991564, -12.4987894, -8.36593698, -4.19990514, 0, 3.75438176, 7.68072325, 11.7642935, 15.9903615, 20.3441964, 24.8110672, 29.3762429, 34.0249925, 38.7425852, 43.5142899, 48.3253756, 53.1611115, 58.0067666, 62.8476099, 67.6689104, 72.4559371, 76.6966559, 80.3052923, 84.3919683, 88.8353878, 93.5142549, 98.3072737, 103.093148, 107.750582, 112.15828, 116.194946, 119.739284, 122.669997, 124.86579, 125.984556, 126.206774, 125.477392, 123.741356, 121.744339, 119.030926, 114.932175, 109.152816, 101.39758, 91.3711953, 78.7783928, 64.669128, 46.9934886, 25.1930083, -1.29077936, -33.0163408, -65.8580356, -106.229464, -155.860118, -216.479492, -289.817078, -380.30555, -483.667333, -597.745304, -722.235224, -860.615829, -1014.66738, -1186.17013, -1373.57015, -1582.73965, -1823.18816, -2108.00157, -2426.16523, -2766.24407, -3125.62711, -3516.65546, -3944.80694, -4420.88423, -4951.74569, -5524.72537, -6138.20136, -6784.51115, -7469.00279, -8192.14539, -8954.40805, -9756.25986, -10712.2391, -11680.2284, -12632.1796, -13540.0443, -14375.7745, -14976.3625, -15538.6925, -16092.556, -16761.1101, -17537.1892, -18374.7255, -19268.4984, -20213.2876, -21203.8726, -22235.0328, -24253.1101, -25525.5835, -26347.1897, -27012.6655, -27816.7475, -29054.1725, -31019.6772, -32231.7935, -33068.6163},
        {-262.386706, -260.140192, -257.993429, -255.941422, -253.979172, -252.101685, -250.303963, -248.581009, -246.927828, -245.339422, -243.810796, -242.336953, -240.912896, -239.533628, -238.194154, -236.889476, -235.614598, -234.364524, -233.134257, -231.918801, -230.713159, -229.512334, -228.311331, -227.105152, -225.888801, -224.657281, -223.405597, -222.128751, -220.821747, -219.479588, -218.097278, -216.669821, -215.192219, -213.659477, -211.876726, -209.937899, -207.939212, -205.881829, -203.766916, -201.595637, -199.369159, -197.088646, -194.755264, -192.370177, -189.934551, -187.449551, -184.916343, -182.33609, -179.70996, -177.039116, -174.324724, -171.58455, -168.802799, -165.980275, -163.117784, -160.216132, -157.276124, -154.298565, -151.284261, -148.234016, -145.148636, -142.028927, -138.875694, -135.689742, -132.471876, -129.222902, -125.943626, -122.642075, -119.326107, -115.981628, -112.608974, -109.208483, -105.780492, -102.325338, -98.8433576, -95.3348881, -91.8002665, -88.2398299, -84.6539152, -81.0428596, -77.4069999, -73.7466734, -70.0622169, -66.3539676, -62.6493999, -58.9347113, -55.1957732, -51.4320419, -47.6429738, -43.8280252, -39.9866525, -36.1183121, -32.2224602, -28.2985534, -24.3460479, -20.3644002, -16.3530665, -12.3115033, -8.23916691, -4.13551368, 0, 2.5865028, 5.57279504, 8.92567875, 12.611956, 16.5984287, 20.851899, 25.3391689, 30.0270405, 34.8823157, 39.8717966, 44.9622852, 50.1205836, 55.3134938, 60.5078177, 65.6703576, 70.7679153, 75.6366487, 80.2246963, 84.794365, 89.2844617, 93.6337931, 97.7811662, 101.665388, 105.225264, 108.399603, 111.127211, 113.346895, 114.997461, 116.017717, 116.296097, 115.84025, 114.604094, 112.541548, 110.974307, 109.172401, 106.04112, 101.124049, 93.9647744, 84.1068809, 71.0939544, 53.5260294, 32.1774101, 6.77628965, -22.9491386, -57.2706813, -88.3004258, -126.509828, -174.210625, -233.714552, -307.333347, -401.188963, -509.125987, -631.090142, -766.872369, -915.94759, -1077.93261, -1252.44424, -1430.05349, -1621.47883, -1847.30688, -2130.42036, -2451.59472, -2795.58414, -3157.4181, -3549.53698, -3976.38926, -4446.80353, -4966.36383, -5524.59395, -6120.10228, -6743.64728, -7402.89492, -8099.17088, -8833.80081, -9608.11037, -10541.7492, -11488.138, -12419.0214, -13306.1441, -14121.2508, -14701.9724, -15243.5765, -15775.285, -16419.1347, -17167.6082, -17974.5158, -18834.274, -19741.2991, -20690.0078, -21674.8165, -22939.8531, -23961.7396, -24878.1199, -25826.638, -26944.9379, -28370.6636, -30241.459, -31480.9531, -32140.6276},
        {-260.487583, -258.145377, -255.906917, -253.76715, -251.721025, -249.763491, -247.889497, -246.093991, -244.371923, -242.71824, -241.127891, -239.595825, -238.116992, -236.686338, -235.298814, -233.949367, -232.632947, -231.344502, -230.07898, -228.831332, -227.596504, -226.369446, -225.145107, -223.918435, -222.684378, -221.437886, -220.173908, -218.887391, -217.573285, -216.226538, -214.8421, -213.414917, -211.93994, -210.412117, -208.634943, -206.70312, -204.713651, -202.6677, -200.566427, -198.410995, -196.202564, -193.942297, -191.631356, -189.270902, -186.862097, -184.406103, -181.904081, -179.357194, -176.766602, -174.133468, -171.458954, -168.766256, -166.034023, -163.26294, -160.45369, -157.606957, -154.723426, -151.80378, -148.848703, -145.85888, -142.834995, -139.777731, -136.687773, -133.565805, -130.412511, -127.228574, -124.014679, -120.777512, -117.523614, -114.241292, -110.93084, -107.592553, -104.226727, -100.833655, -97.4136336, -93.9669573, -90.4939211, -86.9948198, -83.4699486, -79.9196024, -76.3440761, -72.7436648, -69.1186633, -65.4693667, -61.8204948, -58.1596163, -54.4740065, -50.7631679, -47.0266026, -43.2638132, -39.4743018, -35.6575709, -31.8131228, -27.9404599, -24.0390844, -20.1084988, -16.1482053, -12.1577064, -8.13650426, -4.08410136, 0, 1.85772362, 4.26098214, 7.16435018, 10.5224024, 14.2897134, 18.4208578, 22.8704104, 27.5929456, 32.5430382, 37.6752628, 42.9441941, 48.3044066, 53.710475, 59.116974, 64.4784781, 69.7495621, 75.0282378, 80.2895708, 85.1636945, 89.6359201, 93.6915586, 97.3159214, 100.49432, 103.212064, 105.454467, 107.206838, 108.45449, 109.182733, 109.376878, 109.197937, 108.391098, 106.888961, 104.624129, 103.025958, 101.278671, 98.1673586, 93.1755944, 85.7869535, 75.4850107, 61.7533409, 42.2801894, 18.8908651, -8.57979722, -40.2969627, -76.4257964, -107.497255, -145.719264, -193.665541, -253.909803, -329.025767, -426.479303, -537.972678, -661.616194, -796.588579, -944.249934, -1104.98097, -1279.1624, -1460.31036, -1657.23027, -1886.21611, -2165.39008, -2480.65108, -2822.59543, -3187.87993, -3583.43057, -4011.4596, -4476.01838, -4979.79599, -5518.73812, -6092.6049, -6697.94636, -7338.22658, -8014.31689, -8727.08859, -9477.413, -10379.9366, -11293.3117, -12189.966, -13042.3268, -13822.8218, -14369.9482, -14879.3507, -15380.8554, -16002.6113, -16735.5252, -17529.536, -18377.599, -19272.6694, -20207.7022, -21175.6527, -21667.6278, -22471.8843, -23538.8483, -24818.9456, -26262.6021, -27820.2438, -29442.2964, -30655.7015, -31173.6156},
        {-259.058751, -256.621067, -254.291093, -252.063711, -249.933799, -247.89624, -245.945912, -244.077696, -242.286473, -240.567123, -238.914526, -237.323562, -235.789112, -234.306056, -232.869274, -231.473646, -230.114053, -228.785376, -227.482493, -226.200287, -224.933636, -223.677421, -222.426523, -221.175822, -219.920198, -218.654531, -217.373701, -216.07259, -214.746077, -213.389042, -211.996367, -210.56293, -209.083612, -207.553295, -205.771872, -203.835818, -201.843951, -199.797479, -197.697608, -195.545549, -193.342507, -191.089691, -188.78831, -186.439571, -184.044681, -181.604849, -179.121283, -176.595191, -174.02778, -171.420259, -168.773835, -166.119448, -163.42793, -160.699845, -157.935754, -155.136223, -152.301814, -149.43309, -146.530616, -143.594954, -140.626668, -137.626321, -134.594476, -131.531698, -128.438549, -125.315592, -122.163392, -118.987879, -115.794857, -112.573817, -109.324975, -106.048545, -102.744745, -99.4137879, -96.0558903, -92.6712674, -89.2601346, -85.8227073, -82.3592008, -78.8698306, -75.3548121, -71.8143606, -68.2486916, -64.6580204, -61.0639395, -57.455526, -53.8216004, -50.1616843, -46.4752994, -42.7619673, -39.0212097, -35.2525483, -31.4555048, -27.6296008, -23.774358, -19.8892981, -15.9739427, -12.0278135, -8.05043223, -4.0413205, 0, 1.40115188, 3.43973118, 6.06241129, 9.2158656, 12.8467675, 16.9017904, 21.3276076, 26.0708926, 31.0783188, 36.2965594, 41.6722881, 47.152178, 52.6829027, 58.2111354, 63.6835497, 69.0468189, 74.5861225, 80.2964921, 85.3589322, 89.7926531, 93.6168655, 96.8507795, 99.5136059, 101.624555, 103.202837, 104.267663, 104.838242, 104.933786, 104.573505, 104.200586, 103.254804, 101.628177, 99.2127242, 97.2326532, 94.9138864, 91.1270946, 85.3646379, 77.1188765, 65.8821707, 51.1468806, 30.7098504, 6.27498288, -22.3336303, -55.2918975, -92.7757272, -125.369229, -165.238059, -214.956077, -277.09714, -354.235107, -454.553084, -568.159933, -689.642988, -817.364718, -957.395151, -1112.34378, -1284.82009, -1476.8592, -1691.77551, -1933.51006, -2207.78286, -2512.50731, -2848.79423, -3216.97609, -3616.06649, -4045.68439, -4504.25752, -4991.09602, -5509.87796, -6061.81065, -6652.29228, -7277.69456, -7937.77414, -8632.28765, -9360.99174, -10227.3111, -11101.4174, -11957.1501, -12768.3488, -13508.8532, -14020.2385, -14496.785, -14969.017, -15571.7069, -16294.114, -17082.0616, -17926.9808, -18820.3032, -19753.4601, -20717.883, -20516.1931, -21147.9183, -22379.6088, -23977.8146, -25709.0858, -27339.9725, -28637.0249, -29765.0435, -30144.102},
        {-257.975999, -255.448651, -253.032869, -250.723456, -248.515212, -246.402939, -244.381437, -242.445509, -240.589955, -238.809576, -237.099174, -235.45355, -233.867505, -232.33584, -230.853357, -229.414857, -228.01514, -226.649009, -225.311264, -223.996707, -222.700139, -221.416361, -220.140174, -218.866379, -217.589778, -216.305172, -215.007363, -213.69115, -212.351336, -210.982722, -209.580109, -208.138298, -206.652091, -205.116288, -203.326253, -201.380702, -199.380746, -197.327658, -195.222712, -193.067181, -190.862339, -188.609459, -186.309814, -183.964678, -181.575323, -179.143024, -176.669054, -174.154685, -171.601192, -169.009847, -166.381925, -163.755215, -161.093682, -158.397807, -155.668074, -152.904962, -150.108955, -147.280533, -144.420179, -141.528374, -138.6056, -135.652339, -132.669072, -129.656281, -126.614448, -123.544054, -120.445582, -117.325611, -114.190576, -111.028379, -107.839109, -104.62285, -101.37969, -98.1097155, -94.813012, -91.4896662, -88.1397646, -84.7633935, -81.3606393, -77.9315885, -74.4763275, -70.9949425, -67.4875201, -63.9541466, -60.410062, -56.8474572, -53.2583409, -49.6423078, -45.9989524, -42.3278694, -38.6286534, -34.9008991, -31.1442011, -27.3581541, -23.5423526, -19.6963913, -15.8198648, -11.9123678, -7.97349496, -4.0028408, 0, 1.04913371, 2.80189112, 5.19948282, 8.18311941, 11.6940115, 15.6733696, 20.0624045, 24.8023266, 29.8343466, 35.099675, 40.5395225, 46.0950997, 51.7076172, 57.3182855, 62.8683152, 68.298917, 74.0186127, 80.0353716, 85.2333908, 89.654019, 93.3386051, 96.3284976, 98.6650454, 100.389597, 101.543501, 102.168107, 102.304763, 101.994817, 101.279619, 100.8629, 99.8807519, 98.1758091, 95.5907051, 92.9806916, 89.6820936, 84.7612046, 77.7568732, 68.2079478, 55.653277, 39.6317094, 18.5661467, -6.54897983, -35.9564843, -69.8991807, -108.619883, -143.873739, -186.513146, -238.902835, -303.407536, -382.391979, -483.819557, -597.613531, -715.427228, -835.117496, -967.206713, -1116.53069, -1287.92525, -1492.26782, -1726.9795, -1982.89052, -2252.58403, -2546.43741, -2875.74506, -3244.62447, -3645.03227, -4074.50713, -4527.01076, -4999.15446, -5500.66505, -6033.85047, -6611.62981, -7224.06736, -7869.79795, -8547.45644, -9255.67765, -10084.2528, -10917.8713, -11732.3789, -12503.6212, -13207.444, -13692.8812, -14147.1316, -14600.3888, -15187.5853, -15897.4556, -16675.0933, -17511.4151, -18397.3378, -19323.778, -20281.6524, -19571.2605, -20089.459, -21457.1505, -23295.2378, -25224.6233, -26866.2099, -27840.9, -28816.7625, -29026.9376},
        {-256.970622, -254.379514, -251.903138, -249.536219, -247.273481, -245.109647, -243.039442, -241.05759, -239.158814, -237.337838, -235.589387, -233.908183, -232.288952, -230.726417, -229.215302, -227.750331, -226.326228, -224.937717, -223.579521, -222.246365, -220.932973, -219.634068, -218.344374, -217.058616, -215.771517, -214.477801, -213.172193, -211.849415, -210.504193, -209.131249, -207.725309, -206.281095, -204.793332, -203.256744, -201.463174, -199.513055, -197.509255, -195.453083, -193.345845, -191.18885, -188.983406, -186.73082, -184.4324, -182.089454, -179.70329, -177.275215, -174.806536, -172.298563, -169.752602, -167.169962, -164.55195, -161.935015, -159.284562, -156.601137, -153.885285, -151.137553, -148.358487, -145.548633, -142.708536, -139.838742, -136.939798, -134.01225, -131.056642, -128.073523, -125.063436, -122.026929, -118.964546, -115.887985, -112.808672, -109.704158, -106.574266, -103.418821, -100.237644, -97.0305593, -93.79739, -90.5379591, -87.2520899, -83.9396055, -80.6003291, -77.234084, -73.8406932, -70.41998, -66.9717675, -63.4958789, -59.9881777, -56.4505502, -52.8849303, -49.2912697, -45.6695202, -42.0196334, -38.3415609, -34.6352546, -30.9006661, -27.137747, -23.346449, -19.5267239, -15.6785234, -11.801799, -7.89650252, -3.96258562, 0, 0.422387494, 1.64797194, 3.61158737, 6.24806781, 9.49224728, 13.2789598, 17.5430394, 22.2193201, 27.242636, 32.547821, 38.0697092, 43.7431346, 49.5029313, 55.2839332, 61.0209744, 66.648889, 72.6531799, 79.0473484, 84.5206673, 89.1259712, 92.9160946, 95.9438721, 98.2621383, 99.9237276, 100.981475, 101.488214, 101.49678, 101.060008, 100.230731, 99.975223, 99.0979526, 97.3777232, 94.5933381, 91.0821446, 86.3436737, 79.7285117, 70.8478983, 59.3130737, 44.7352777, 26.7257504, 4.45540514, -21.8901791, -52.6136114, -88.0175011, -128.404457, -167.950075, -214.61573, -270.235787, -336.644609, -415.676556, -512.573846, -621.597836, -736.941098, -857.569441, -992.194022, -1145.15054, -1320.77468, -1529.86865, -1768.83198, -2027.0098, -2294.66848, -2584.12137, -2907.25812, -3268.77792, -3661.19615, -4082.62076, -4528.67316, -4996.81666, -5493.63179, -6020.54186, -6586.70894, -7184.627, -7813.04054, -8470.69404, -9156.33201, -9947.53272, -10744.4985, -11526.2653, -12271.8693, -12960.3464, -13462.4077, -13937.6311, -14411.4778, -14992.1208, -15672.2811, -16410.2166, -17200.3714, -18037.1898, -18915.1159, -19828.5939, -19154.235, -19693.4221, -21060.4103, -22869.4546, -24734.8103, -26270.7324, -27091.476, -27783.137, -27725.3232},
        {-256.262747, -253.608644, -251.072704, -248.649556, -246.333827, -244.120145, -242.003136, -239.97743, -238.037653, -236.178433, -234.394398, -232.680175, -231.030391, -229.439676, -227.902655, -226.413956, -224.968208, -223.560038, -222.184073, -220.834941, -219.50727, -218.195687, -216.89482, -215.599296, -214.303743, -213.002789, -211.69106, -210.363186, -209.013793, -207.637509, -206.228962, -204.782778, -203.293587, -201.756015, -199.957645, -198.000983, -195.991022, -193.929107, -191.816587, -189.654806, -187.445112, -185.188851, -182.887371, -180.542017, -178.154136, -175.725074, -173.256179, -170.748797, -168.204274, -165.623957, -163.009193, -160.39378, -157.745908, -155.066223, -152.355365, -149.613979, -146.842708, -144.042194, -141.213081, -138.356012, -135.471629, -132.560577, -129.623498, -126.661034, -123.67383, -120.662529, -117.627772, -114.587138, -111.557797, -108.505466, -105.429692, -102.330019, -99.2059933, -96.0571599, -92.8830641, -89.6832513, -86.4572668, -83.2046559, -79.924964, -76.6177365, -73.2825185, -69.9188556, -66.5262929, -63.1043759, -59.6277626, -56.1089654, -52.5607959, -48.9836071, -45.3777519, -41.7435834, -38.0814546, -34.3917183, -30.6747276, -26.9308355, -23.1603949, -19.3637588, -15.5412801, -11.693312, -7.82020725, -3.92231893, 0, -0.0496634385, 0.768577157, 2.38604162, 4.73404977, 7.74392146, 11.3469765, 15.4745347, 20.057916, 25.0284401, 30.3174269, 35.8561962, 41.5760679, 47.4083617, 53.2843976, 59.1354953, 64.8929747, 71.0249882, 77.5395462, 83.153704, 87.9138171, 91.8662407, 95.0573305, 97.5334417, 99.3409297, 100.52615, 101.135458, 101.215209, 100.811759, 99.9714622, 99.7887802, 98.8776576, 96.9700185, 93.7977865, 89.2593599, 83.0034255, 74.6342766, 63.8338951, 50.2842623, 33.6673598, 13.6651693, -9.77554579, -37.3181716, -69.3325317, -106.18845, -148.255749, -192.423375, -243.412249, -302.462415, -370.813915, -449.706793, -541.182807, -644.700412, -758.192442, -882.056384, -1021.72172, -1180.35866, -1361.13742, -1572.38483, -1810.94255, -2068.02683, -2334.93904, -2622.60284, -2940.54232, -3292.99731, -3675.25386, -4086.70792, -4526.21961, -4993.04595, -5488.40892, -6012.41901, -6568.06903, -7152.14372, -7763.75586, -8402.01823, -9066.04359, -9821.95112, -10585.0956, -11337.8382, -12062.54, -12741.5623, -13269.8248, -13775.4243, -14278.1971, -14853.6346, -15499.5427, -16190.5372, -16925.5835, -17703.6467, -18523.692, -19384.6848, -18937.684, -19548.5906, -20871.2934, -22559.6811, -24267.6427, -25649.0669, -26357.8427, -26718.1447, -26346.5252},
        {-255.903863, -253.177994, -250.57441, -248.087615, -245.712111, -243.442401, -241.272988, -239.198375, -237.213064, -235.311559, -233.488363, -231.737977, -230.054906, -228.433652, -226.868718, -225.354606, -223.885821, -222.456863, -221.062238, -219.696446, -218.353992, -217.029378, -215.717106, -214.411681, -213.107604, -211.799379, -210.481508, -209.148494, -207.794841, -206.415051, -205.003627, -203.555071, -202.063887, -200.524578, -198.719246, -196.753062, -194.733748, -192.662702, -190.541318, -188.370993, -186.153121, -183.889099, -181.580322, -179.228186, -176.834087, -174.39942, -171.92558, -169.413964, -166.865968, -164.282986, -161.666415, -159.048082, -156.398291, -153.717778, -151.007279, -148.267531, -145.499268, -142.703228, -139.880146, -137.030758, -134.1558, -131.256008, -128.332118, -125.384866, -122.414988, -119.42322, -116.410297, -113.399078, -110.41189, -107.403842, -104.374239, -101.322381, -98.247571, -95.1491102, -92.0263008, -88.8784447, -85.7048439, -82.5048002, -79.2776157, -76.0225921, -72.7390315, -69.4262357, -66.0835067, -62.7101465, -59.2617315, -55.7603462, -52.2286004, -48.6672148, -45.0769105, -41.4584083, -37.8124291, -34.1396939, -30.4409234, -26.7168387, -22.9681605, -19.1956098, -15.3999075, -11.5817745, -7.74193162, -3.88109983, 0, -0.216583442, 0.440256594, 1.90250512, 4.10214715, 6.9711677, 10.4415518, 14.4452844, 18.9143505, 23.7807353, 28.9764236, 34.4334005, 40.083651, 45.8591601, 51.6919128, 57.5138942, 63.2570892, 69.2525709, 75.4893367, 80.9966728, 85.792083, 89.8930712, 93.3171413, 96.0817971, 98.2045424, 99.7028812, 100.594317, 100.896354, 100.626497, 99.8022476, 99.4307435, 98.1769909, 95.7616038, 91.9051966, 86.3005809, 78.6822723, 68.7922, 56.3593188, 41.1125837, 22.7809496, 1.09337125, -23.3977885, -52.037585, -85.2581651, -123.491675, -167.170262, -215.607968, -270.63457, -332.961741, -403.301155, -482.364483, -569.298586, -668.292498, -781.106916, -909.497367, -1055.20882, -1219.99099, -1405.59357, -1616.54895, -1851.19174, -2104.82094, -2372.21903, -2660.52494, -2975.08388, -3318.90885, -3691.061, -4092.41545, -4525.38557, -4991.24528, -5485.62821, -6007.35846, -6553.22168, -7124.39402, -7720.40518, -8340.78487, -8985.06281, -9709.53846, -10442.7793, -11170.1227, -11876.9058, -12548.466, -13102.2026, -13636.6827, -14166.3598, -14734.8185, -15344.9813, -15987.6325, -16666.3012, -17384.5166, -18145.8081, -18953.7049, -18818.0116, -19522.5378, -20781.1193, -22307.5915, -23815.7901, -25019.5507, -25632.7087, -25647.5785, -24927.4113},
        {-255.940798, -253.100458, -250.38933, -247.8017, -245.331856, -242.974085, -240.722673, -238.571907, -236.516074, -234.54946, -232.666354, -230.86104, -229.127808, -227.460942, -225.85473, -224.303459, -222.801416, -221.342888, -219.922161, -218.533522, -217.171258, -215.829656, -214.503003, -213.185586, -211.87169, -210.555605, -209.231615, -207.894008, -206.537072, -205.155091, -203.742355, -202.293148, -200.801759, -199.262474, -197.447547, -195.46695, -193.433327, -191.348169, -189.21297, -187.029221, -184.798415, -182.522044, -180.2016, -177.838575, -175.434463, -172.990754, -170.508941, -167.990516, -165.436973, -162.849802, -160.230497, -157.617199, -154.973956, -152.301466, -149.600427, -146.871537, -144.115492, -141.332992, -138.524734, -135.691416, -132.833735, -129.952389, -127.048077, -124.121495, -121.173341, -118.204314, -115.215111, -112.228286, -109.265869, -106.283481, -103.280401, -100.255912, -97.2092929, -94.1398261, -91.0467919, -87.9294714, -84.7871453, -81.6190946, -78.4246002, -75.202943, -71.9534039, -68.6752637, -65.3678034, -62.0303039, -58.6166529, -55.1497828, -51.6531714, -48.1275725, -44.5737402, -40.9924285, -37.3843914, -33.7503829, -30.0911569, -26.4074675, -22.7000686, -18.9697142, -15.2171584, -11.443155, -7.6484582, -3.83382186, 0, 0.761532105, 2.19530693, 4.24738764, 6.8638374, 9.99071938, 13.5740967, 17.5600327, 21.8945903, 26.5238328, 31.3938233, 36.4506251, 41.6403012, 46.9089149, 52.2025293, 57.4672076, 62.6490129, 67.5895362, 72.219916, 76.7225218, 81.0210297, 85.0391161, 88.7004572, 91.9287293, 94.6476086, 96.7807714, 98.251894, 98.9846525, 98.9027234, 97.9297828, 96.2421784, 93.4182689, 89.3059292, 83.7530341, 77.2555378, 69.3372756, 59.3492615, 46.9449465, 31.7777818, 13.5012183, -8.2312928, -33.2446999, -62.5659005, -96.6434958, -135.926087, -180.862276, -230.845001, -287.642442, -351.967114, -424.531535, -506.04822, -595.834258, -697.703117, -813.842207, -945.954841, -1094.75597, -1261.4043, -1447.05853, -1653.35999, -1880.87507, -2129.6437, -2399.84095, -2693.0786, -3010.71933, -3353.89886, -3723.36841, -4120.05568, -4546.22974, -5003.16606, -5487.2218, -5997.53617, -6532.62047, -7092.33848, -7676.04678, -8283.10199, -8912.86068, -9616.58008, -10328.741, -11035.7249, -11723.9133, -12379.6874, -12928.1468, -13457.8097, -13981.3213, -14530.0177, -15109.1456, -15716.1639, -16356.7085, -17036.4157, -17760.9214, -18535.8618, -18529.7057, -19281.1694, -20518.5595, -21970.1829, -23364.3462, -24429.3564, -24893.5201, -24657.6262, -23541.4915},
        {-256.431669, -253.459304, -250.623907, -247.919524, -245.340202, -242.879986, -240.532923, -238.29306, -236.154442, -234.111116, -232.157129, -230.286527, -228.493355, -226.771661, -225.115491, -223.51889, -221.975906, -220.480585, -219.026972, -217.609115, -216.221059, -214.856852, -213.510538, -212.176165, -210.84778, -209.519427, -208.185154, -206.839006, -205.475031, -204.087275, -202.669783, -201.216602, -199.721779, -198.17936, -196.350698, -194.351082, -192.298594, -190.194831, -188.041393, -185.839875, -183.591876, -181.298994, -178.962825, -176.584969, -174.167022, -171.710581, -169.217245, -166.688612, -164.126278, -161.531842, -158.906901, -156.298201, -153.661213, -150.996555, -148.304847, -145.586707, -142.842754, -140.073608, -137.279886, -134.462208, -131.621194, -128.757461, -125.871628, -122.964315, -120.036141, -117.087724, -114.119683, -111.152459, -108.206019, -105.240098, -102.25403, -99.2471504, -96.2187916, -93.168288, -90.0949734, -86.9981819, -83.8772471, -80.7315032, -77.5602839, -74.3629231, -71.1387547, -67.8871128, -64.607331, -61.2987434, -57.9188063, -54.4886733, -51.0300011, -47.5434827, -44.0298108, -40.4896785, -36.9237786, -33.332804, -29.7174476, -26.0784023, -22.4163611, -18.7320167, -15.0260622, -11.2991903, -7.55209407, -3.78546633, 0, 2.05489446, 4.52497117, 7.37384633, 10.5651361, 14.0624567, 17.8294243, 21.8296551, 26.0267653, 30.384371, 34.8660885, 39.4355339, 44.0563235, 48.6920734, 53.3063997, 57.8629188, 62.3252468, 65.96238, 68.6387034, 71.8889967, 75.5280293, 79.3705701, 83.2313882, 86.9252527, 90.2669328, 93.0711974, 95.1528158, 96.326557, 96.40719, 95.209484, 91.8895812, 87.1623741, 81.0402198, 73.5354755, 66.2275522, 58.34528, 48.6996082, 36.8327777, 22.2870295, 4.60460462, -16.672256, -42.0530706, -71.9280862, -106.744935, -146.951249, -192.994661, -243.519217, -301.227032, -367.016634, -441.786551, -526.435312, -621.492718, -728.676692, -849.463837, -985.035418, -1135.96971, -1303.11573, -1487.32248, -1687.70199, -1907.23501, -2150.79718, -2424.43444, -2723.8149, -3046.43689, -3391.40435, -3760.5419, -4154.42531, -4574.12412, -5020.34213, -5491.97268, -5988.93321, -6513.81154, -7063.44389, -7636.82333, -8232.94294, -8850.79578, -9539.05708, -10234.6172, -10924.0487, -11593.9241, -12230.816, -12760.9789, -13271.5152, -13774.8479, -14299.093, -14850.4617, -15428.6266, -16040.126, -16691.4982, -17389.2813, -18140.0136, -18199.3677, -18978.7714, -20206.1719, -21609.5164, -22916.7519, -23855.8256, -24154.6847, -23688.2172, -22173.5389},
        {-257.431268, -254.317075, -251.347665, -248.516846, -245.81843, -243.246225, -240.794042, -238.45569, -236.22498, -234.095722, -232.061726, -230.1168, -228.254756, -226.469403, -224.754552, -223.104012, -221.511592, -219.971104, -218.476356, -217.02116, -215.599324, -214.204659, -212.830974, -211.47208, -210.121787, -208.773904, -207.422241, -206.060609, -204.682817, -203.282674, -201.853992, -200.39058, -198.886248, -197.334806, -195.487073, -193.463466, -191.387322, -189.260334, -187.0842, -184.860614, -182.591272, -180.277869, -177.922101, -175.525663, -173.09025, -170.617559, -168.109283, -165.56712, -162.992764, -160.38791, -157.754255, -155.147447, -152.514058, -149.854615, -147.169642, -144.459666, -141.725213, -138.966808, -136.184977, -133.380246, -130.553141, -127.704188, -124.833912, -121.942839, -119.031495, -116.100406, -113.150098, -110.198283, -107.262264, -104.307117, -101.332254, -98.337087, -95.3210271, -92.2834861, -89.2238755, -86.1416069, -83.0360921, -79.9067425, -76.7529698, -73.5741855, -70.3698013, -67.1392288, -63.8818796, -60.5971652, -57.2473037, -53.8510857, -50.427749, -46.9779124, -43.5021943, -40.0012134, -36.4755883, -32.9259376, -29.3528799, -25.7570339, -22.139018, -18.499451, -14.8389515, -11.158138, -7.4576292, -3.73804366, 0, 3.32544809, 6.81609814, 10.4523023, 14.2144127, 18.0827814, 22.0377607, 26.0597027, 30.1289594, 34.225883, 38.3308258, 42.4241397, 46.486177, 50.4972897, 54.4378301, 58.2881503, 62.0286024, 64.3939347, 65.1865562, 67.2042558, 70.1604707, 73.7686379, 77.7421945, 81.7945775, 85.639224, 88.989571, 91.5590556, 93.0611149, 93.2091858, 91.7167054, 86.765587, 80.1623501, 72.0798891, 62.6910982, 54.6536166, 46.897966, 37.6920695, 26.4633979, 12.639422, -4.3523876, -25.08456, -50.8211061, -81.2326226, -116.756349, -157.829525, -204.889389, -255.468458, -313.634876, -380.552062, -457.383436, -545.292418, -646.449751, -759.780361, -885.482051, -1024.01402, -1176.36914, -1343.3007, -1525.56194, -1720.36977, -1932.81756, -2171.85653, -2448.71307, -2753.8228, -3081.47461, -3428.97046, -3798.7178, -4190.78057, -4604.65611, -5040.26149, -5499.59129, -5983.46494, -6499.30717, -7040.31589, -7605.02408, -8191.96471, -8799.67076, -9475.9852, -10157.8035, -10831.3311, -11482.7736, -12098.3363, -12600.8511, -13082.1462, -13555.5873, -14053.1601, -14580.8599, -15136.9231, -15727.7329, -16359.6721, -17039.1239, -17772.4711, -17838.1192, -18625.7994, -19850.5129, -21227.2609, -22471.0447, -23296.8656, -23419.7246, -22722.3549, -20834.46},
        {-259.094316, -255.843419, -252.743411, -249.787939, -246.97065, -244.28519, -241.725208, -239.28435, -236.956262, -234.734592, -232.612988, -230.585095, -228.644561, -226.785033, -225.000157, -223.283582, -221.628953, -220.029918, -218.480124, -216.973217, -215.502846, -214.062656, -212.646295, -211.247409, -209.859647, -208.476654, -207.092077, -205.699565, -204.292763, -202.865318, -201.410879, -199.92309, -198.395601, -196.822057, -194.946311, -192.892161, -190.786408, -188.630805, -186.427106, -184.177065, -181.882433, -179.544966, -177.166415, -174.748535, -172.293078, -169.801799, -167.276449, -164.718784, -162.130555, -159.513516, -156.869422, -154.260349, -151.62642, -148.968078, -146.28577, -143.579941, -140.851037, -138.099504, -135.325786, -132.530329, -129.71358, -126.875983, -124.017984, -121.140029, -118.242563, -115.326032, -112.390882, -109.453213, -106.52875, -103.585651, -100.623346, -97.6412675, -94.6388446, -91.6155085, -88.5706898, -85.5038193, -82.4143277, -79.3016456, -76.1652037, -73.0044328, -69.8187636, -66.6076267, -63.3704528, -60.1066727, -56.777139, -53.401383, -49.9993993, -46.5718706, -43.1194795, -39.6429088, -36.1428409, -32.6199586, -29.0749444, -25.508481, -21.921251, -18.3139371, -14.6872218, -11.0417878, -7.3783178, -3.69749431, 0, 3.83041265, 7.72436051, 11.6682425, 15.6484575, 19.6514045, 23.6634823, 27.67109, 31.6606263, 35.6184902, 39.5310807, 43.3847966, 47.1660368, 50.8612003, 54.456686, 57.9388928, 61.2942196, 63.0541208, 62.997146, 64.4006402, 66.9392285, 70.287536, 74.1201876, 78.1118086, 81.937024, 85.2704589, 87.7867383, 89.1604874, 89.0663312, 87.1788948, 81.2980151, 73.6605278, 64.5034942, 54.063976, 45.4542365, 37.4737362, 28.1552155, 16.8731751, 3.00211617, -14.0834603, -35.0090534, -61.4298246, -92.6282348, -129.028327, -171.054146, -219.129733, -269.983999, -328.659904, -396.505275, -474.867938, -565.09572, -670.491985, -788.058921, -916.808593, -1056.67046, -1209.44701, -1376.09975, -1557.59023, -1751.19637, -1962.40048, -2200.70339, -2478.46981, -2784.85542, -3113.44794, -3461.05618, -3829.94689, -4219.88157, -4629.51689, -5058.3279, -5509.84057, -5985.28946, -6495.12391, -7029.946, -7587.80052, -8166.73227, -8764.78605, -9429.51757, -10097.083, -10753.1494, -11383.3838, -11973.4534, -12439.8199, -12883.4925, -13319.7976, -13792.2467, -14305.0097, -14850.5128, -15433.5131, -16058.7678, -16731.0339, -17455.0685, -17403.6662, -18157.0008, -19392.3567, -20787.0181, -22018.2692, -22763.3942, -22699.6774, -21718.0917, -19540.579},
        {-261.466571, -258.089107, -254.866392, -251.792005, -248.859524, -246.062527, -243.394592, -240.849296, -238.420218, -236.100936, -233.885027, -231.76607, -229.737642, -227.793321, -225.926686, -224.131315, -222.400784, -220.728673, -219.108559, -217.53402, -215.998634, -214.49598, -213.019634, -211.563175, -210.120181, -208.684229, -207.248899, -205.807767, -204.354412, -202.882411, -201.385343, -199.856786, -198.290316, -196.679514, -194.765856, -192.674274, -190.532727, -188.342974, -186.106775, -183.825888, -181.502074, -179.13709, -176.732695, -174.29065, -171.812712, -169.300641, -166.756197, -164.181137, -161.577221, -158.946209, -156.289859, -153.672771, -151.032502, -148.369447, -145.684004, -142.976569, -140.24754, -137.497312, -134.726284, -131.93485, -129.123409, -126.292357, -123.442091, -120.573007, -117.685503, -114.779975, -111.85682, -108.932477, -106.022998, -103.095696, -100.149927, -97.1850489, -94.2004181, -91.1953914, -88.1693258, -85.121578, -82.0515049, -78.9584634, -75.8418104, -72.7009026, -69.535097, -66.3437505, -63.1262198, -59.8818618, -56.561236, -53.1891243, -49.7908952, -46.3674891, -42.9198465, -39.4489077, -35.9556133, -32.4409036, -28.905719, -25.3510001, -21.7776871, -18.1867206, -14.579041, -10.9555887, -7.31730418, -3.66512779, 0, 3.32652052, 6.80873207, 10.4252682, 14.1547624, 17.9758481, 21.867159, 25.8073285, 29.7749901, 33.7487774, 37.7073238, 41.6292629, 45.4932283, 49.2778533, 52.9617715, 56.5236165, 59.9420218, 61.9647921, 62.3947284, 63.9944814, 66.4789356, 69.5629755, 72.9614857, 76.3893507, 79.5614549, 82.192683, 83.9979193, 84.6920484, 83.9899549, 81.6065231, 75.7214461, 68.1477036, 59.0607374, 48.6359898, 39.6138344, 30.8872475, 20.6652231, 8.35372366, -6.64128838, -24.9138506, -47.0580007, -74.58963, -106.900358, -144.40386, -187.51381, -236.643883, -288.346553, -347.861996, -416.569188, -495.847105, -587.074721, -693.764666, -812.554464, -941.718972, -1081.02788, -1233.30281, -1399.99514, -1582.55623, -1780.56202, -1997.76556, -2239.9658, -2515.67577, -2817.64478, -3141.76546, -3485.97307, -3851.66392, -4238.64563, -4645.87342, -5072.93392, -5522.53929, -5995.63372, -6502.85889, -7033.96957, -7586.55312, -8158.19691, -8746.48829, -9398.7683, -10050.4322, -10686.629, -11292.5076, -11853.2169, -12277.1341, -12677.3612, -13071.9492, -13522.4941, -14029.8105, -14576.5694, -15164.4926, -15795.3019, -16470.719, -17192.4659, -16897.6903, -17571.8661, -18829.824, -20286.3947, -21556.4091, -22254.6978, -21996.0917, -20665.8271, -18305.5149},
        {-264.383961, -260.87599, -257.526209, -254.328147, -251.275332, -248.361294, -245.579563, -242.923667, -240.387135, -237.963496, -235.64628, -233.429016, -231.305233, -229.26846, -227.312226, -225.43006, -223.615492, -221.86205, -220.163264, -218.512663, -216.903775, -215.330131, -213.785259, -212.262688, -210.755948, -209.258567, -207.764075, -206.266001, -204.757874, -203.233223, -201.685577, -200.108466, -198.495418, -196.839962, -194.882642, -192.7488, -190.566821, -188.338444, -186.065409, -183.749454, -181.392319, -178.995742, -176.561463, -174.09122, -171.586753, -169.049801, -166.482103, -163.885397, -161.261423, -158.61192, -155.938628, -153.305897, -150.651498, -147.97581, -145.279218, -142.5621, -139.824841, -137.067821, -134.291422, -131.496026, -128.682014, -125.849769, -122.999672, -120.132104, -117.247447, -114.346084, -111.428396, -108.51222, -105.614976, -102.701044, -99.7696726, -96.8201126, -93.8516141, -90.8634271, -87.8548018, -84.8249882, -81.7732365, -78.6987967, -75.600919, -72.4788534, -69.33185, -66.1591589, -62.9600303, -59.7337142, -56.4174475, -53.0427914, -49.6420521, -46.2164923, -42.7673745, -39.2959611, -35.8035149, -32.2912982, -28.7605737, -25.2126039, -21.6486513, -18.0699785, -14.477848, -10.8735223, -7.25826409, -3.6333358, 0, 2.532414, 5.36767047, 8.4726991, 11.8144295, 15.3597915, 19.0757146, 22.9291286, 26.8869631, 30.9161477, 34.9836123, 39.0562863, 43.1010996, 47.0849818, 50.9748625, 54.7376715, 58.3403384, 60.8626715, 62.1448689, 64.1604443, 66.6862299, 69.4990579, 72.3757605, 75.0931701, 77.4281189, 79.1574391, 80.057963, 79.9065227, 78.4799506, 75.5550789, 69.8983404, 62.6674466, 53.9423495, 43.8030012, 34.3885245, 24.749286, 13.3865404, -0.237511412, -16.6606685, -36.4207299, -60.0554949, -88.7566778, -122.209145, -160.822755, -205.007367, -255.17284, -307.965099, -368.498921, -438.125149, -518.194627, -610.058198, -717.10362, -836.155257, -964.966273, -1103.28071, -1254.90731, -1421.82985, -1606.0321, -1809.90083, -2034.92524, -2282.15489, -2554.94228, -2850.95649, -3169.28874, -3509.45159, -3871.67142, -4255.84694, -4661.43741, -5088.22761, -5537.61368, -6010.08029, -6515.59872, -7043.70754, -7591.60758, -8156.49966, -8735.58462, -9376.9993, -10014.2745, -10631.8771, -11214.2739, -11745.9317, -12125.8938, -12480.9996, -12832.3261, -13261.8167, -13766.4554, -14317.8649, -14914.2973, -15554.0044, -16235.2382, -16956.2507, -16406.5576, -16994.5608, -18271.4665, -19788.4813, -21096.8115, -21747.6636, -21292.244, -19618.2511, -17137.6765},
        {-267.802551, -264.156991, -260.67311, -257.344388, -254.164308, -251.126349, -248.223992, -245.450718, -242.800008, -240.265342, -237.840201, -235.518067, -233.292419, -231.156739, -229.104507, -227.129204, -225.22431, -223.383308, -221.599676, -219.866896, -218.178449, -216.527816, -214.908477, -213.313913, -211.737604, -210.173032, -208.613677, -207.053021, -205.484543, -203.901724, -202.298046, -200.666989, -199.002033, -197.29666, -195.29115, -193.11091, -190.884389, -188.613285, -186.299296, -183.944119, -181.549452, -179.116993, -176.648438, -174.145486, -171.609833, -169.043179, -166.44722, -163.823654, -161.174178, -158.500491, -155.804288, -153.146572, -150.468451, -147.770338, -145.052643, -142.31578, -139.56016, -136.786196, -133.994299, -131.184881, -128.358354, -125.515131, -122.655624, -119.780243, -116.889402, -113.983513, -111.062987, -108.148222, -105.259137, -102.354924, -99.4346992, -96.4975793, -93.5426806, -90.5691197, -87.5760129, -84.5624766, -81.5276272, -78.4705812, -75.390455, -72.2863649, -69.1574274, -66.002759, -62.8214759, -59.6126947, -56.2982181, -52.9174455, -49.5107056, -46.0796238, -42.6258255, -39.150936, -35.6565809, -32.1443854, -28.6159749, -25.0729749, -21.5170107, -17.9497078, -14.3726914, -10.7875871, -7.19602014, -3.59961598, 0, 1.61598645, 3.70367845, 6.21679574, 9.10905807, 12.3341852, 15.8458968, 19.5979127, 23.5439526, 27.6377362, 31.8329833, 36.0834137, 40.342747, 44.564703, 48.7030015, 52.7113621, 56.5435047, 59.6477358, 61.9055732, 64.4139332, 67.0182327, 69.5638886, 71.8963179, 73.8609374, 75.303164, 76.0684148, 76.0021065, 74.9496561, 72.7564806, 69.2679968, 63.9042186, 57.0919471, 48.80422, 39.0140752, 29.1690925, 18.5050388, 5.8644694, -9.22194037, -27.2235154, -48.6095805, -73.8494605, -103.677811, -138.217873, -177.887059, -223.102782, -274.282454, -328.465407, -390.291653, -461.023126, -541.921758, -634.24948, -740.949274, -859.54741, -987.594796, -1124.94485, -1276.15196, -1443.65986, -1629.9123, -1840.0435, -2073.19526, -2325.5743, -2595.09456, -2884.68204, -3196.90233, -3532.97447, -3891.83578, -4273.47095, -4677.88309, -5105.06166, -5554.92843, -6027.44344, -6531.30278, -7056.38646, -7599.63055, -8157.97108, -8728.34411, -9360.8232, -9985.92251, -10587.2937, -11148.5885, -11653.4584, -11991.2684, -12302.8148, -12612.1582, -13021.3884, -13523.679, -14079.6915, -14684.2252, -15332.0794, -16018.0533, -16736.9462, -15968.1475, -16477.8667, -17762.6079, -19318.8751, -20643.1722, -21232.0034, -20581.8726, -18592.3889, -16048.4555},
        {-271.647706, -267.848957, -264.216787, -260.744573, -257.42569, -254.253515, -251.221424, -248.322792, -245.550997, -242.899414, -240.361418, -237.930388, -235.599697, -233.362723, -231.212842, -229.143429, -227.147862, -225.219515, -223.351765, -221.537988, -219.77156, -218.045858, -216.354257, -214.690133, -213.046864, -211.417823, -209.796389, -208.175937, -206.549842, -204.911482, -203.254232, -201.571468, -199.856567, -198.102904, -196.049493, -193.822241, -191.550113, -189.234741, -186.877754, -184.480784, -182.045461, -179.573416, -177.066279, -174.525682, -171.953254, -169.350626, -166.71943, -164.061295, -161.377853, -158.670735, -155.941569, -153.242067, -150.522694, -147.783996, -145.026518, -142.250804, -139.4574, -136.646852, -133.819703, -130.976499, -128.117785, -125.244106, -122.356007, -119.454034, -116.538731, -113.610644, -110.670317, -107.742648, -104.851948, -101.948539, -99.0313878, -96.09946, -93.1517224, -90.1871414, -87.2046834, -84.2033149, -81.1820023, -78.139712, -75.0754106, -71.9880643, -68.8766396, -65.7401031, -62.577421, -59.3875599, -56.0772078, -52.6934114, -49.2843262, -45.8519133, -42.3981338, -38.9249487, -35.4343189, -31.9282057, -28.40857, -24.8773728, -21.3365753, -17.7881384, -14.2340231, -10.6761906, -7.11660193, -3.55721804, 0, 1.05201337, 2.67209974, 4.80616466, 7.40011368, 10.3998524, 13.7512863, 17.4003209, 21.2928619, 25.3748147, 29.592085, 33.8905782, 38.2162, 42.5148558, 46.7324513, 50.8148919, 54.7080833, 58.0291688, 60.6770881, 63.3413752, 65.8974867, 68.2208791, 70.1870089, 71.6713327, 72.549307, 72.6963883, 71.9880332, 70.2996981, 67.5068396, 63.4849143, 57.9572424, 51.0071999, 42.5558844, 32.5243932, 21.9706859, 10.2474289, -3.59587484, -19.9791866, -39.3224675, -62.0456789, -88.5687821, -119.093393, -154.324271, -194.724097, -240.755552, -292.881318, -349.176718, -413.08863, -485.676574, -568.00007, -661.118638, -767.014558, -884.696775, -1012.3938, -1150.30377, -1302.64617, -1471.83497, -1660.28417, -1873.9383, -2110.87839, -2365.33393, -2632.55304, -2918.20036, -3226.91826, -3560.84866, -3917.89651, -4297.91159, -4700.86343, -5126.63286, -5574.66161, -6044.63976, -6543.90487, -7063.32307, -7599.93726, -8150.79036, -8712.92527, -9341.25845, -9960.49087, -10553.1971, -11101.9516, -11589.3289, -11896.4116, -12174.9275, -12450.9476, -12839.2586, -13330.815, -13879.6063, -14478.4354, -15120.105, -15797.4178, -16503.1765, -15722.0788, -16215.0394, -17468.9067, -18970.5288, -20206.7543, -20664.4312, -19830.408, -17624.7141, -15073.729},
        {-275.882505, -271.913049, -268.116752, -264.486817, -261.016448, -257.698849, -254.527224, -251.494776, -248.59471, -245.82023, -243.164538, -240.620839, -238.182337, -235.842236, -233.593739, -231.43005, -229.344374, -227.329913, -225.379872, -223.487455, -221.645865, -219.848307, -218.087983, -216.358099, -214.651857, -212.962461, -211.283116, -209.607026, -207.927393, -206.237422, -204.530317, -202.799282, -201.03752, -199.238235, -197.137616, -194.862779, -192.543933, -190.182621, -187.780388, -185.338777, -182.859335, -180.343604, -177.793131, -175.209458, -172.594131, -169.948694, -167.274691, -164.573667, -161.847167, -159.096735, -156.323915, -153.565423, -150.786869, -147.989037, -145.172707, -142.338663, -139.487685, -136.620557, -133.73806, -130.840976, -127.930088, -125.006176, -122.070024, -119.122414, -116.164127, -113.195945, -110.218651, -107.263523, -104.361106, -101.449287, -98.526871, -95.5926628, -92.6454677, -89.6840907, -86.707337, -83.7140116, -80.7029196, -77.6728661, -74.6226562, -71.5510951, -68.4569878, -65.3391395, -62.1963551, -59.0274399, -55.7248585, -52.342823, -48.9368235, -45.509116, -42.0619567, -38.5976015, -35.1183066, -31.6263279, -28.1239215, -24.6133435, -21.0968499, -17.5766968, -14.0551402, -10.5344362, -7.01684075, -3.50461001, 0, 0.938284713, 2.44914745, 4.47748666, 6.96820083, 9.8661884, 13.1163478, 16.6635776, 20.4527762, 24.428842, 28.5366736, 32.7211693, 36.9272277, 41.0997472, 45.1836262, 49.1237633, 52.8650569, 55.9505318, 58.267676, 60.6703878, 63.0174499, 65.1676449, 66.9797556, 68.3125647, 69.0248548, 68.9754086, 68.0230088, 66.0264381, 62.8444792, 58.3359147, 52.1003478, 44.3507732, 35.0237276, 24.0557476, 12.498338, -0.269448809, -15.1658851, -32.5889247, -52.9365215, -76.6066294, -103.997202, -134.751747, -170.252278, -211.04436, -257.673555, -310.685426, -369.792542, -436.581709, -511.806738, -596.221442, -690.579631, -795.449213, -911.99712, -1040.00896, -1180.24113, -1335.43207, -1507.4303, -1698.08437, -1911.93222, -2147.52176, -2400.46699, -2666.69229, -2951.52002, -3259.89498, -3593.91173, -3950.83522, -4330.14725, -4731.14378, -5153.25844, -5596.60633, -6060.91706, -6552.28397, -7063.09392, -7590.8737, -8133.1501, -8687.4499, -9316.19559, -9935.79434, -10527.549, -11072.7625, -11552.7376, -11842.3848, -12100.3276, -12353.4657, -12720.4617, -13191.8923, -13720.0998, -14297.5725, -14916.799, -15570.2679, -16250.4675, -15682.9897, -16226.7712, -17408.1973, -18753.6534, -19789.5248, -20042.1969, -19038.0551, -16727.8089, -14219.2956},
        {-280.514156, -276.362273, -272.390796, -268.592733, -264.961089, -261.488869, -258.169079, -254.994725, -251.958812, -249.054346, -246.274332, -243.611777, -241.059685, -238.611064, -236.258917, -233.996251, -231.816071, -229.711384, -227.675194, -225.700508, -223.78033, -221.907668, -220.075525, -218.276909, -216.504824, -214.752276, -213.012271, -211.277815, -209.541912, -207.79757, -206.037793, -204.255587, -202.443958, -200.595911, -198.444025, -196.116996, -193.746652, -191.334451, -188.881848, -186.390301, -183.861267, -181.296203, -178.696566, -176.063813, -173.3994, -170.704786, -167.981427, -165.23078, -162.454302, -159.653449, -156.82968, -154.002842, -151.155602, -148.289019, -145.404151, -142.502057, -139.583794, -136.650421, -133.702996, -130.742578, -127.770225, -124.786995, -121.793947, -118.792139, -115.782629, -112.766476, -109.744738, -106.755921, -103.837633, -100.913699, -97.9827483, -95.0434121, -92.0943211, -89.1341061, -86.1613977, -83.1748266, -80.1730236, -77.1546193, -74.1182445, -71.0625299, -67.9861061, -64.8876039, -61.7656539, -58.618887, -55.3248316, -51.9452169, -48.5432272, -45.1214233, -41.6823659, -38.2286158, -34.7627335, -31.2872799, -27.8048155, -24.3179012, -20.8290975, -17.3409653, -13.8560652, -10.3769579, -6.90620417, -3.44636462, 0, 0.93157521, 2.4165561, 4.40062947, 6.82948214, 9.64880093, 12.8042726, 16.2415841, 19.9064221, 23.7444734, 27.7014249, 31.7229634, 35.7547757, 39.7425486, 43.631969, 47.3687235, 50.8984991, 53.629571, 55.430853, 57.4636055, 59.5583558, 61.5456311, 63.2559586, 64.5198654, 65.1678788, 65.0305259, 63.938334, 61.7218302, 58.2115418, 53.2379959, 46.1714837, 37.4715216, 27.1067078, 15.0456402, 2.4339891, -11.3481836, -27.2347018, -45.610089, -66.8588688, -91.3655647, -119.5147, -150.410506, -186.107453, -227.240557, -274.444834, -328.3553, -390.432522, -460.279577, -538.325092, -624.997694, -720.726011, -824.57429, -940.003115, -1068.57832, -1211.66341, -1370.2088, -1545.35038, -1738.22404, -1951.32911, -2184.12816, -2434.59638, -2700.34278, -2985.38622, -3294.33947, -3628.94252, -3986.06748, -4364.82092, -4763.77169, -5181.88695, -5620.10536, -6078.25035, -6561.10406, -7062.76842, -7581.35042, -8114.95706, -8661.69535, -9291.56451, -9912.80626, -10505.5545, -11049.9433, -11526.1065, -11804.4038, -12047.9263, -12283.8539, -12631.4089, -13081.3379, -13586.0368, -14138.1699, -14730.4014, -15355.3955, -16005.8167, -15719.3105, -16331.8822, -17425.383, -18581.6641, -19382.5765, -19409.9714, -18245.6998, -15886.3068, -13455.7665},
        {-285.519624, -281.174101, -277.01677, -273.040418, -269.237833, -265.601802, -262.12511, -258.800546, -255.620896, -252.578947, -249.667486, -246.879299, -244.207174, -241.643897, -239.182255, -236.815036, -234.535026, -232.335011, -230.20778, -228.146118, -226.142813, -224.190651, -222.282419, -220.410905, -218.568895, -216.749176, -214.944534, -213.147758, -211.351633, -209.548946, -207.732485, -205.895037, -204.029387, -202.128323, -199.919894, -197.534823, -195.107017, -192.637855, -190.128715, -187.580975, -184.996013, -182.375208, -179.719936, -177.031578, -174.31151, -171.561112, -168.78176, -165.974834, -163.141711, -160.28377, -157.402389, -154.500043, -151.576989, -148.634581, -145.674175, -142.697125, -139.704784, -136.698508, -133.679651, -130.649566, -127.609609, -124.561134, -121.505496, -118.444048, -115.378145, -112.309141, -109.238392, -106.212113, -103.275447, -100.337219, -97.395877, -94.4498664, -91.4976334, -88.5376242, -85.5682848, -82.5880613, -79.5953998, -76.5887464, -73.5665473, -70.5272485, -67.4692962, -64.3911364, -61.2912152, -58.1679788, -54.8831941, -51.5065217, -48.1092649, -44.6943052, -41.2645243, -37.8228037, -34.3720252, -30.9150701, -27.4548203, -23.9941572, -20.5359625, -17.0831178, -13.6385047, -10.2050047, -6.78549956, -3.38287078, 0, 0.990906539, 2.500551, 4.4765943, 6.86669734, 9.61852105, 12.6797263, 15.9979741, 19.5209253, 23.1962408, 26.9715815, 30.7946084, 34.6129823, 38.3743642, 42.026415, 45.5167956, 48.7931669, 51.0961218, 52.2723115, 53.8687175, 55.681486, 57.5067635, 59.1406961, 60.3794304, 61.0191125, 60.855889, 59.685906, 57.30531, 53.5102473, 48.0968643, 40.1490856, 30.4364273, 18.9687019, 5.75572246, -7.89528084, -22.6232039, -39.4129206, -58.6438434, -80.6953848, -105.946957, -134.777973, -165.820923, -201.733812, -243.237844, -291.054217, -345.904135, -410.924255, -483.816458, -564.698081, -653.68646, -750.898934, -853.964122, -968.529845, -1097.95851, -1244.2028, -1406.33728, -1584.72871, -1779.74388, -1991.51841, -2220.70276, -2468.19952, -2733.96649, -3020.03931, -3330.16469, -3665.50147, -4022.82353, -4400.91751, -4797.71337, -5211.77565, -5644.81009, -6096.74561, -6571.06441, -7063.59544, -7573.03759, -8098.08978, -8637.45092, -9267.78538, -9890.33524, -10484.308, -11028.9113, -11503.3526, -11775.3976, -12009.9903, -12234.0991, -12564.9814, -12993.8662, -13474.362, -13999.5786, -14562.6259, -15156.614, -15774.6528, -15799.6739, -16487.5443, -17483.8843, -18434.3143, -18984.4546, -18779.9256, -17466.3476, -15103.9956, -12777.8683},
        {-290.875875, -286.326007, -281.972521, -277.807968, -273.824901, -270.015875, -266.37344, -262.890151, -259.558559, -256.371219, -253.320683, -250.399504, -247.600234, -244.915427, -242.337635, -239.859411, -237.473309, -235.171881, -232.94768, -230.793258, -228.70117, -226.663967, -224.674203, -222.72443, -220.807201, -218.91507, -217.040588, -215.17631, -213.314787, -211.448574, -209.570222, -207.672284, -205.747314, -203.787864, -201.516397, -199.066191, -196.573773, -194.04046, -191.46757, -188.85642, -186.208329, -183.524615, -180.806595, -178.055586, -175.272908, -172.459877, -169.617812, -166.748031, -163.851851, -160.930589, -157.985565, -155.002741, -151.999122, -148.976359, -145.936103, -142.880005, -139.809715, -136.726885, -133.633163, -130.530203, -127.419653, -124.303165, -121.182389, -118.058977, -114.934579, -111.810845, -108.689427, -105.624369, -102.668461, -99.7152913, -96.7631143, -93.8101843, -90.8547554, -87.8950817, -84.9294173, -81.9560165, -78.9731333, -75.9790219, -72.9719365, -69.9501311, -66.9118599, -63.8553771, -60.7789368, -57.6807932, -54.4060131, -51.0326665, -47.6406643, -44.2332319, -40.8135946, -37.3849778, -33.9506069, -30.5137073, -27.0775042, -23.6452231, -20.2200893, -16.8053281, -13.404165, -10.0198253, -6.65553428, -3.31451741, 0, 1.07530038, 2.62735748, 4.60638237, 6.96258613, 9.64617986, 12.6073746, 15.7963815, 19.1634116, 22.6586759, 26.2323856, 29.8347518, 33.4159855, 36.9262977, 40.3158997, 43.5350024, 46.533817, 48.3800194, 48.8977442, 50.0334129, 51.5481219, 53.2029677, 54.7590467, 55.9774552, 56.6192896, 56.4456465, 55.2176222, 52.696313, 48.6428155, 42.818226, 34.0115892, 23.3124722, 10.773587, -3.55235436, -18.162392, -33.728938, -51.3111372, -71.291352, -94.0519449, -119.975278, -149.443715, -180.734249, -216.975375, -258.961376, -307.486532, -363.345128, -431.095336, -506.826573, -590.392151, -681.645376, -780.439559, -883.193041, -997.392393, -1128.00616, -1277.49155, -1443.17844, -1624.6988, -1821.68456, -2031.88955, -2257.2507, -2501.75378, -2768.02542, -3055.71966, -3367.28355, -3703.14904, -4060.33358, -4437.42193, -4831.93501, -5242.18176, -5670.37194, -6116.50876, -6582.86428, -7066.82387, -7567.60541, -8084.4268, -8616.50589, -9245.27837, -9867.18988, -10460.9039, -11005.0837, -11478.3929, -11748.295, -11978.7865, -12196.1881, -12514.0606, -12924.1919, -13382.0198, -13881.1496, -14415.1867, -14977.7366, -15562.4045, -15892.7129, -16650.9294, -17547.1216, -18291.3574, -18593.7046, -18164.2308, -16713.0039, -14384.6632, -12180.3278},
        {-296.320558, -291.547181, -286.979686, -282.610339, -278.431403, -274.435145, -270.613828, -266.959717, -263.465077, -260.122172, -256.923268, -253.860628, -250.926519, -248.113203, -245.412947, -242.818014, -240.32067, -237.913178, -235.587805, -233.336814, -231.15247, -229.027038, -226.952783, -224.921969, -222.92686, -220.959723, -219.012821, -217.078419, -215.148781, -213.216173, -211.272859, -209.311104, -207.323173, -205.301329, -202.956908, -200.428548, -197.858265, -195.24744, -192.597455, -189.909691, -187.18553, -184.426352, -181.63354, -178.808475, -175.952539, -173.067112, -170.153576, -167.213313, -164.247704, -161.258131, -158.245974, -155.194037, -152.122683, -149.033695, -145.928858, -142.809957, -139.678774, -136.537096, -133.386705, -130.229386, -127.066923, -123.9011, -120.733702, -117.566513, -114.401316, -111.239896, -108.084038, -104.991613, -102.019145, -99.0527425, -96.0905546, -93.1307292, -90.1714147, -87.2107591, -84.2469109, -81.2780184, -78.3022297, -75.3176932, -72.3225572, -69.3149699, -66.2930796, -63.2550347, -60.1989833, -57.1230738, -53.8555399, -50.4830612, -47.0943593, -43.6930965, -40.2829352, -36.8675376, -33.4505661, -30.0356831, -26.6265508, -23.2268316, -19.8401879, -16.4702819, -13.120776, -9.79533258, -6.49761386, -3.23128222, 0, 0.672253236, 1.86597416, 3.52798015, 5.60508862, 8.04411695, 10.7918825, 13.7952028, 17.000895, 20.3557768, 23.8066653, 27.3003781, 30.7837324, 34.2035458, 37.5066356, 40.6398192, 43.5499139, 45.334475, 45.8189966, 46.8712465, 48.2560573, 49.7382616, 51.0826921, 52.0541813, 52.4175619, 51.9376666, 50.3793279, 47.5073784, 43.0866508, 36.8819777, 27.6830145, 16.587336, 3.65232798, -11.0646237, -25.9382951, -41.6952196, -59.4800205, -79.7056635, -102.785114, -129.131338, -159.157301, -191.407472, -228.731986, -271.909385, -321.718208, -378.936998, -447.70942, -524.607609, -609.568826, -702.530329, -803.42938, -909.140628, -1026.40713, -1160.01554, -1312.25682, -1480.32656, -1663.70809, -1861.88473, -2071.75594, -2295.97613, -2540.01846, -2808.61265, -3098.85867, -3410.37974, -3742.6948, -4095.14607, -4467.15689, -4858.02287, -5267.13422, -5694.34956, -6139.26256, -6603.22647, -7083.80467, -7579.98174, -8090.74225, -8615.07079, -9227.38789, -9832.38318, -10410.1822, -10940.9107, -11404.6941, -11677.3706, -11912.8782, -12136.037, -12457.6592, -12868.5265, -13323.5905, -13815.4245, -14336.6016, -14879.6949, -15437.2776, -15860.4923, -16609.0393, -17420.5105, -18032.498, -18182.5937, -17608.3897, -16047.4779, -13774.8523, -11659.7415},
        {-302.119652, -297.114942, -292.325535, -287.743412, -283.360555, -279.168945, -275.160562, -271.327389, -267.661406, -264.154595, -260.798936, -257.58641, -254.509, -251.558686, -248.727449, -246.00727, -243.390131, -240.868013, -238.432896, -236.076763, -233.791593, -231.569369, -229.402071, -227.281681, -225.20018, -223.149548, -221.121768, -219.10882, -217.102685, -215.095345, -213.078781, -211.044973, -208.985904, -206.893553, -204.467707, -201.85259, -199.195939, -196.499219, -193.763892, -190.991422, -188.183272, -185.340906, -182.465788, -179.559381, -176.623148, -173.658554, -170.66706, -167.650132, -164.609232, -161.545824, -158.461371, -155.336801, -152.194558, -149.036551, -145.86469, -142.680882, -139.487036, -136.285061, -133.076866, -129.864359, -126.649449, -123.434045, -120.220054, -117.009387, -113.803952, -110.605656, -107.41641, -104.298078, -101.311085, -98.3336851, -95.3638997, -92.399751, -89.4392608, -86.4804511, -83.5213438, -80.5599608, -77.5943242, -74.6224559, -71.6423778, -68.6521118, -65.6496799, -62.6331041, -59.6004063, -56.5496084, -53.2907562, -49.9198137, -46.5350039, -43.140449, -39.7402711, -36.3385924, -32.9395349, -29.5472209, -26.1657724, -22.7993116, -19.4519606, -16.1278415, -12.8310765, -9.56578776, -6.33609732, -3.14612736, 0, 0.310230037, 1.17786683, 2.54691812, 4.36139162, 6.56529507, 9.1026362, 11.9174227, 14.9536624, 18.1553629, 21.4665321, 24.8311776, 28.1933071, 31.4969284, 34.6860492, 37.7046773, 40.4968203, 42.2026261, 42.6512653, 43.6050001, 44.8355828, 46.1147652, 47.2142995, 47.9059376, 47.9614316, 47.1525334, 45.2509952, 42.0285689, 37.2570065, 30.7080601, 21.2075124, 9.81994009, -3.39911394, -18.3941069, -33.4343799, -49.3019482, -67.2255246, -87.6521011, -111.02867, -137.802222, -168.419751, -201.811738, -240.403231, -284.937932, -336.159539, -394.811754, -464.294307, -542.030861, -628.101108, -722.58474, -825.561451, -934.635565, -1055.38759, -1192.10712, -1346.80849, -1516.86075, -1701.71861, -1900.83676, -2110.59885, -2334.22858, -2578.29987, -2849.38971, -3142.22001, -3453.60476, -3782.19068, -4129.72991, -4496.54916, -4883.85564, -5292.20433, -5718.92169, -6163.16042, -6625.57582, -7103.58682, -7595.82597, -8100.92583, -8617.51895, -9211.00265, -9796.55355, -10356.113, -10871.6225, -11325.0232, -11602.2344, -11845.2345, -12077.1173, -12405.4068, -12819.7766, -13274.7207, -13761.6714, -14272.0612, -14797.3224, -15328.8874, -15800.3434, -16524.5541, -17258.2883, -17758.3148, -17781.4025, -17084.3201, -15423.8366, -13226.7177, -11208.5777},
        {-308.313295, -303.072305, -298.055548, -293.254745, -288.661614, -284.267874, -280.065244, -276.045445, -272.200195, -268.521214, -265.000221, -261.628935, -258.399076, -255.302363, -252.330515, -249.475251, -246.728291, -244.081355, -241.526161, -239.054428, -236.657877, -234.328226, -232.057195, -229.836503, -227.657869, -225.513013, -223.393654, -221.291511, -219.198304, -217.105752, -215.005574, -212.889489, -210.749217, -208.576478, -206.0601, -203.349618, -200.598209, -197.807421, -194.978803, -192.113903, -189.21427, -186.281453, -183.317, -180.32246, -177.299382, -174.249314, -171.173806, -168.074405, -164.95266, -161.810121, -158.648335, -155.446071, -152.228152, -148.996621, -145.753519, -142.50089, -139.240776, -135.975219, -132.706262, -129.435948, -126.166319, -122.899418, -119.637286, -116.381967, -113.135504, -109.899937, -106.677311, -103.534279, -100.535941, -97.5510862, -94.5775686, -91.6132422, -88.6559612, -85.7035794, -82.7539511, -79.8049303, -76.854371, -73.9001274, -70.9400535, -67.9720033, -64.993831, -62.0033906, -58.9985362, -55.9771218, -52.7292951, -49.3611322, -45.9812057, -42.5941102, -39.2044401, -35.81679, -32.4357545, -29.0659281, -25.7119054, -22.3782808, -19.069649, -15.7906044, -12.5457417, -9.33965538, -6.17693995, -3.06218997, 0, 0.0727211457, 0.715543277, 1.87119892, 3.48242062, 5.49194089, 7.84249227, 10.4768073, 13.3376185, 16.3676584, 19.5096595, 22.7063544, 25.9004756, 29.0347556, 32.0519269, 34.8947222, 37.5058738, 39.0609315, 39.3930302, 40.1801971, 41.2007684, 42.2330804, 43.0554691, 43.4462709, 43.1838218, 42.0464581, 39.8125159, 36.2603315, 31.1682411, 24.3145807, 14.5778825, 2.96621505, -10.4721442, -25.6889179, -40.8784243, -56.8710882, -74.9659751, -95.6420288, -119.378193, -146.653412, -177.946629, -212.62068, -252.610303, -298.612808, -351.325508, -411.445717, -481.48638, -559.875268, -646.855785, -742.671335, -847.565322, -959.963521, -1084.14851, -1223.75803, -1380.48974, -1552.08034, -1738.04494, -1937.89861, -2147.82778, -2371.43274, -2615.9451, -2889.33214, -3184.40978, -3495.61221, -3820.8674, -4164.02323, -4526.21028, -4910.34761, -5318.02946, -5744.29215, -6187.8815, -6649.02155, -7124.75254, -7613.30865, -8112.92405, -8621.8329, -9195.77092, -9761.09535, -10301.665, -10801.3386, -11243.975, -11526.0188, -11777.0191, -12018.2978, -12353.8457, -12772.2616, -13227.6993, -13710.4799, -14210.9247, -14719.3547, -15226.0912, -15709.0299, -16399.5656, -17065.5842, -17474.9714, -17395.613, -16595.3949, -14842.2028, -12732.7616, -10821.0696},
        {-314.941622, -309.462284, -304.21521, -299.191896, -294.383838, -289.782532, -285.379475, -281.166163, -277.134092, -273.274759, -269.579659, -266.040288, -262.648143, -259.394721, -256.271516, -253.270026, -250.381747, -247.598174, -244.910805, -242.311135, -239.79066, -237.340877, -234.953282, -232.61937, -230.330639, -228.078585, -225.854703, -223.65049, -221.457442, -219.267055, -217.070826, -214.86025, -212.626824, -210.362044, -207.745393, -204.930935, -202.076488, -199.183672, -196.254111, -193.289425, -190.291236, -187.261165, -184.200835, -181.111867, -177.995883, -174.854504, -171.689352, -168.502049, -165.294216, -162.067475, -158.823447, -155.536889, -152.23687, -148.925594, -145.605266, -142.278089, -138.946268, -135.612007, -132.27751, -128.94498, -125.616622, -122.29464, -118.981238, -115.67862, -112.388991, -109.114553, -105.857512, -102.690729, -99.6853703, -96.6979128, -93.7259801, -90.7671957, -87.8191832, -84.8795662, -81.9459683, -79.0160132, -76.0873242, -73.1575252, -70.2242396, -67.285091, -64.3377031, -61.3796993, -58.4087034, -55.4223389, -52.1887897, -48.8252252, -45.4515725, -42.0729009, -38.6942797, -35.3207783, -31.957466, -28.6094122, -25.2816861, -21.9793572, -18.7074948, -15.4711683, -12.2754469, -9.12540003, -6.02609703, -2.98260724, 0, 0.0432169234, 0.63151126, 1.70882525, 3.21910111, 5.1062811, 7.31430744, 9.78712237, 12.4686681, 15.3028869, 18.233721, 21.2051127, 24.1610041, 27.0453375, 29.8020551, 32.3750992, 34.708412, 35.9858499, 36.0427715, 36.5423607, 37.2656842, 37.9938088, 38.5078011, 38.5887279, 38.0176558, 36.5756516, 34.0437819, 30.2031135, 24.834713, 17.7196471, 7.78692451, -4.01790874, -17.6581685, -33.0971705, -48.4982062, -64.7246036, -83.1196974, -104.186811, -128.429267, -156.35039, -188.453502, -224.507935, -265.974393, -313.499805, -367.731102, -429.315213, -499.922024, -578.919771, -666.699646, -763.652841, -870.170547, -985.410162, -1112.50464, -1254.44543, -1412.64376, -1585.28467, -1772.00167, -1972.4283, -2182.85226, -2407.01332, -2652.30126, -2927.4155, -3224.0341, -3535.05571, -3857.95567, -4197.96417, -4556.75179, -4938.41306, -5345.24699, -5770.66476, -6213.10499, -6672.67285, -7145.88405, -7630.60033, -8124.68346, -8625.99518, -9181.34098, -9727.40294, -10249.8069, -10734.1787, -11166.144, -11451.8562, -11709.3954, -11958.4472, -12299.5177, -12720.301, -13174.8157, -13652.4399, -14142.5512, -14634.5275, -15117.7465, -15583.3156, -16236.1657, -16847.5275, -17188.6312, -17030.7075, -16144.9866, -14302.699, -12285.486, -10491.4503},
        {-322.044773, -316.327895, -310.850001, -305.602422, -300.576485, -295.76352, -291.154856, -286.741821, -282.515745, -278.467955, -274.589782, -270.872554, -267.307599, -263.886247, -260.599826, -257.439666, -254.397095, -251.463442, -248.630036, -245.888205, -243.22928, -240.644587, -238.125457, -235.663219, -233.2492, -230.87473, -228.531138, -226.209752, -223.901902, -221.598917, -219.292124, -216.972854, -214.632434, -212.262195, -209.534892, -206.607841, -203.642187, -200.639597, -197.601739, -194.530278, -191.426882, -188.29322, -185.130956, -181.941759, -178.727296, -175.489234, -172.22924, -168.948981, -165.650125, -162.334337, -159.003287, -155.624293, -152.234115, -148.835164, -145.429851, -142.020589, -138.60979, -135.199864, -131.793223, -128.39228, -124.999445, -121.617131, -118.24775, -114.893712, -111.55743, -108.241316, -104.94778, -101.757942, -98.7510317, -95.7671321, -92.8035531, -89.857604, -86.9265943, -84.0078334, -81.0986308, -78.1962958, -75.298138, -72.4014667, -69.5035914, -66.6018215, -63.6934664, -60.7758355, -57.8462384, -54.9019844, -51.6868731, -48.3303011, -44.964712, -41.5956421, -38.2286277, -34.8692051, -31.5229106, -28.1952805, -24.8918511, -21.6181588, -18.3797399, -15.1821306, -12.0308673, -8.93148627, -5.88952384, -2.91051631, 0, 0.305207731, 1.07827855, 2.26779975, 3.82235859, 5.69054238, 7.82093838, 10.1621339, 12.6627162, 15.2712725, 17.9363902, 20.6066565, 23.2306587, 25.7569842, 28.13422, 30.3109537, 32.2357724, 33.05384, 32.5989691, 32.6370141, 32.9444001, 33.2975522, 33.4728957, 33.2468555, 32.3958569, 30.696325, 27.9246849, 23.8573617, 18.2707806, 10.9413667, 0.827437875, -11.1765009, -25.0485921, -40.7669786, -56.5215033, -73.1844587, -92.1050173, -113.797812, -138.777474, -167.558638, -200.655935, -238.147136, -281.116694, -330.164714, -385.891303, -448.896565, -520.237621, -599.943309, -688.499479, -786.391985, -894.106677, -1011.26116, -1140.27072, -1283.64645, -1442.61373, -1615.77305, -1802.9034, -2003.78381, -2215.08179, -2440.39499, -2686.71543, -2962.61534, -3259.69907, -3570.58886, -3892.68623, -4231.49086, -4588.78524, -4968.96628, -5374.4943, -5798.24336, -6238.51004, -6695.63895, -7165.56358, -7645.87158, -8134.1506, -8627.9883, -9167.36109, -9696.87069, -10203.5075, -10674.2621, -11096.125, -11382.879, -11643.5271, -11896.4344, -12238.9652, -12658.2139, -13108.359, -13578.1409, -14056.2998, -14531.5764, -14992.7109, -15419.9642, -16036.4463, -16609.2473, -16905.4577, -16692.1678, -15736.4678, -13805.4481, -11877.393, -10213.9527},
        {-329.973393, -323.961561, -318.19921, -312.677377, -307.387095, -302.3194, -297.465328, -292.815912, -288.362189, -284.095193, -280.005958, -276.085522, -272.324917, -268.71518, -265.247345, -261.912448, -258.701523, -255.605605, -252.61573, -249.722933, -246.918248, -244.192711, -241.537356, -238.943219, -236.401335, -233.902739, -231.438465, -228.99955, -226.577027, -224.161932, -221.7453, -219.318166, -216.871566, -214.396533, -211.555257, -208.510111, -205.427426, -202.308906, -199.15626, -195.971193, -192.755412, -189.510625, -186.238538, -182.940858, -179.619291, -176.275544, -172.911325, -169.528339, -166.128294, -162.712896, -159.283852, -155.793626, -152.294234, -148.788453, -145.279054, -141.768814, -138.260506, -134.756904, -131.260783, -127.774917, -124.30208, -120.845047, -117.406591, -113.989488, -110.596511, -107.230434, -103.894032, -100.683113, -97.6912493, -94.7293423, -91.7941349, -88.8823704, -85.9907921, -83.1161433, -80.2551674, -77.4046076, -74.5612072, -71.7217096, -68.8828579, -66.0413956, -63.1940659, -60.3376121, -57.4687775, -54.5843054, -51.3960397, -48.0491659, -44.6922938, -41.3314121, -37.9725097, -34.6215754, -31.2845981, -27.9675665, -24.6764696, -21.4172961, -18.1960349, -15.0186748, -11.8912046, -8.81961324, -5.80988945, -2.86802209, 0, 1.67471073, 3.5386676, 5.55869298, 7.70160923, 9.93423873, 12.2234038, 14.5359269, 16.8386303, 19.0983364, 21.2818676, 23.3560462, 25.2876946, 27.0436352, 28.5906903, 29.8956823, 30.9254335, 30.5821688, 28.680625, 27.595643, 27.0659171, 26.8301417, 26.6270109, 26.1952192, 25.2734608, 23.6004299, 20.914821, 16.9553283, 11.4606461, 4.16946877, -6.08114372, -18.3202636, -32.5387062, -48.7272871, -65.2185108, -82.8323482, -102.830986, -125.702734, -151.935899, -182.018792, -216.439719, -255.069338, -299.201591, -349.445632, -406.410614, -470.705693, -544.326702, -626.149445, -716.436404, -815.450064, -923.452908, -1039.06069, -1166.19529, -1307.87024, -1465.62953, -1638.01688, -1824.92308, -2026.2389, -2239.60462, -2467.673, -2715.55199, -2989.07242, -3283.11244, -3594.10837, -3920.59846, -4264.68263, -4626.82593, -5008.86959, -5411.63544, -5830.89942, -6265.29173, -6715.73615, -7178.21641, -7650.56875, -8130.62943, -8616.23469, -9148.79061, -9671.67114, -10171.8201, -10636.1813, -11051.6984, -11333.1513, -11587.7572, -11833.3873, -12164.452, -12568.5259, -13001.292, -13451.9195, -13909.5775, -14363.4349, -14802.6608, -15117.403, -15701.3703, -16287.7239, -16609.6252, -16400.2354, -15392.716, -13320.2283, -11471.0851, -9987.84551},
        {-338.221334, -331.967503, -325.967352, -320.211862, -314.692016, -309.398797, -304.323187, -299.456169, -294.788724, -290.311835, -286.016485, -281.893656, -277.93433, -274.12949, -270.470118, -266.947197, -263.551709, -260.274636, -257.106961, -254.039666, -251.063733, -248.170145, -245.349885, -242.593934, -239.893276, -237.238892, -234.621765, -232.032877, -229.46321, -226.903748, -224.345472, -221.779366, -219.19641, -216.587588, -213.61485, -210.440615, -207.231257, -203.988436, -200.713813, -197.409047, -194.075798, -190.715727, -187.330494, -183.921759, -180.491181, -177.040422, -173.571141, -170.084998, -166.583654, -163.068768, -159.542001, -155.936243, -152.323416, -148.706668, -145.089153, -141.474019, -137.864419, -134.263503, -130.674423, -127.100328, -123.54437, -120.0097, -116.499469, -113.016827, -109.564926, -106.146917, -102.765949, -99.5329676, -96.5563271, -93.6168645, -90.7107432, -87.8341263, -84.9831768, -82.154058, -79.3429329, -76.5459649, -73.7593169, -70.9791523, -68.201634, -65.4229253, -62.6391894, -59.8465893, -57.0412883, -54.2194494, -51.0614709, -47.7280401, -44.3836396, -41.0346796, -37.6875707, -34.348723, -31.024547, -27.7214529, -24.4458511, -21.204152, -18.0027658, -14.848103, -11.7465737, -8.7045885, -5.72855757, -2.82489129, 0, 2.94646174, 5.824727, 8.61820038, 11.3102865, 13.8843901, 16.3239156, 18.6122678, 20.7328513, 22.6690706, 24.4043305, 25.9220355, 27.2055904, 28.2383996, 29.0038678, 29.4853997, 29.6663999, 28.2236345, 24.9538455, 22.7959294, 21.4532968, 20.6293584, 20.0275249, 19.3512069, 18.3038151, 16.5887601, 13.9094527, 9.96930338, 4.47172291, -2.87987806, -13.3447442, -25.903836, -40.564946, -57.3358671, -74.7423808, -93.5332853, -114.852582, -139.152667, -166.885937, -198.504787, -234.461613, -274.439695, -319.894624, -371.429277, -429.64653, -495.149258, -570.61392, -654.051413, -745.546216, -845.182809, -953.045671, -1066.88028, -1191.96889, -1331.81563, -1488.27098, -1659.80941, -1846.42114, -2048.09639, -2263.43278, -2494.12966, -2743.40563, -3014.26823, -3305.01679, -3616.15544, -3947.41732, -4297.22918, -4664.61742, -5048.57276, -5448.11233, -5862.38416, -6290.46225, -6734.34832, -7189.7382, -7654.69245, -8127.27161, -8605.53624, -9133.52717, -9651.82958, -10147.0089, -10605.6307, -11014.2604, -11286.5729, -11530.618, -11764.1999, -12080.0469, -12466.3058, -12879.9054, -13310.5534, -13747.9578, -14181.8264, -14601.8671, -14845.402, -15410.9483, -16007.983, -16345.9832, -16134.426, -15082.7887, -12900.5481, -11111.6674, -9798.50257},
        {-346.578053, -340.189392, -334.04725, -328.142982, -322.467948, -317.013503, -311.771004, -306.73181, -301.887276, -297.22876, -292.74762, -288.435211, -284.282892, -280.28202, -276.423951, -272.700042, -269.101652, -265.620136, -262.246852, -258.973157, -255.790409, -252.689963, -249.663179, -246.701411, -243.796018, -240.938357, -238.119785, -235.331659, -232.565336, -229.812172, -227.063527, -224.310755, -221.545215, -218.758264, -215.629345, -212.312334, -208.964852, -205.588381, -202.184397, -198.754382, -195.299814, -191.822173, -188.322937, -184.803587, -181.265601, -177.710458, -174.139639, -170.554621, -166.956885, -163.34791, -159.729174, -156.009606, -152.285242, -148.559569, -144.836072, -141.118237, -137.409549, -133.713494, -130.033558, -126.373226, -122.735984, -119.125317, -115.544712, -111.997653, -108.487627, -105.018118, -101.592614, -98.3351996, -95.3670682, -92.4429731, -89.5585818, -86.7095619, -83.8915809, -81.1003063, -78.3314055, -75.5805462, -72.8433959, -70.1156219, -67.392892, -64.6708735, -61.9452339, -59.2116409, -56.4657619, -53.7032644, -50.5771451, -47.2616145, -43.9350019, -40.6040951, -37.2756819, -33.9565501, -30.6534876, -27.3732823, -24.122722, -20.9085945, -17.7376877, -14.6167894, -11.5526875, -8.55216971, -5.62202398, -2.76903813, 0, 3.59521001, 6.98361531, 10.1568386, 13.1065025, 15.8242299, 18.3016433, 20.5303654, 22.5020191, 24.2082269, 25.6406116, 26.7907959, 27.6504024, 28.2110539, 28.4643731, 28.4019827, 28.0155053, 25.8667761, 21.7431625, 18.8703304, 16.9335194, 15.6179693, 14.6089197, 13.5916102, 12.2512805, 10.2731702, 7.34251902, 3.14456655, -2.63544753, -10.3122836, -21.1444489, -34.1569164, -49.3813222, -66.8493026, -85.3098801, -105.435998, -128.242629, -154.146194, -183.563114, -216.909809, -254.602699, -296.370743, -343.52705, -396.622544, -456.20815, -522.83479, -598.833113, -682.529388, -774.029607, -873.439762, -980.865846, -1093.97635, -1218.29393, -1357.42, -1513.29337, -1684.45844, -1870.98363, -2072.93737, -2289.55423, -2521.92599, -2772.0541, -3041.13053, -3329.67502, -3640.9407, -3975.43215, -4328.99672, -4699.61951, -5084.15022, -5480.27956, -5889.86146, -6312.39488, -6751.58623, -7202.07977, -7661.91808, -8129.14371, -8601.79924, -9125.27541, -9638.42957, -10127.4672, -10578.5939, -10978.0152, -11237.5195, -11467.3407, -11685.5775, -11985.2989, -12354.9275, -12751.6484, -13165.4465, -13586.3062, -14004.2122, -14409.1493, -14685.1585, -15255.2612, -15835.3205, -16141.1999, -15888.7626, -14793.8719, -12572.3911, -10815.6349, -9635.84514},
        {-354.681743, -348.408498, -342.349198, -336.496526, -330.843168, -325.381809, -320.105135, -315.00583, -310.076579, -305.310067, -300.69898, -296.236002, -291.913819, -287.725115, -283.662576, -279.718887, -275.886733, -272.158798, -268.527768, -264.986329, -261.527164, -258.14296, -254.826401, -251.570172, -248.366959, -245.209446, -242.090318, -239.002261, -235.937959, -232.890098, -229.851363, -226.814439, -223.772011, -220.716763, -217.384585, -213.901962, -210.398464, -206.87511, -203.332919, -199.772908, -196.196098, -192.603505, -188.99615, -185.37505, -181.741224, -178.095691, -174.439469, -170.773577, -167.099034, -163.416857, -159.728066, -155.912702, -152.095384, -148.279752, -144.469447, -140.668112, -136.879386, -133.106912, -129.35433, -125.625281, -121.923407, -118.252348, -114.615746, -111.017242, -107.460478, -103.949093, -100.48673, -97.2034182, -94.2261536, -91.2979094, -88.4141334, -85.5702737, -82.7617779, -79.9840941, -77.2326699, -74.5029534, -71.7903923, -69.0904345, -66.3985279, -63.7101204, -61.0206597, -58.3255938, -55.6203705, -52.9004377, -49.8018413, -46.5072897, -43.203811, -39.8985162, -36.5985166, -33.3109233, -30.0428475, -26.8014002, -23.5936927, -20.426836, -17.3079414, -14.2441199, -11.2424827, -8.31014102, -5.45420589, -2.68178851, 0, 2.22887735, 4.49306643, 6.76606068, 9.02135359, 11.2324386, 13.3728092, 15.4159589, 17.3353811, 19.1045693, 20.697017, 22.0862176, 23.2456647, 24.1488516, 24.7692718, 25.0804188, 25.0557862, 23.4736091, 20.1366298, 17.7219021, 15.9467926, 14.5286681, 13.1848952, 11.6328407, 9.58987129, 6.77335368, 2.90065459, -2.31085926, -9.14382116, -17.8808644, -29.5047788, -43.341317, -59.4630654, -77.9426103, -97.7999722, -119.63402, -144.324308, -172.259192, -203.827027, -239.416171, -279.414978, -324.411101, -374.532945, -430.129873, -491.551251, -559.146441, -631.408776, -711.007659, -798.75646, -895.468552, -1001.95731, -1118.60387, -1247.18211, -1389.95058, -1548.13936, -1720.87875, -1908.24179, -2110.30153, -2325.70022, -2556.26691, -2805.39156, -3076.38078, -3367.28422, -3678.96357, -4011.03755, -4361.01881, -4727.38671, -5107.66641, -5500.08991, -5906.3879, -6326.312, -6766.94857, -7219.58608, -7681.43723, -8149.7147, -8621.63116, -9137.9847, -9641.50626, -10118.5122, -10555.3188, -10938.2424, -11177.071, -11385.6682, -11583.1481, -11868.7136, -12229.777, -12620.3807, -13029.6058, -13446.5331, -13860.2438, -14259.8189, -14914.0167, -15560.1451, -16019.5422, -16113.5462, -15663.4953, -14490.7277, -12416.5814, -10609.7548, -9480.09966},
        {-362.793212, -356.73307, -350.844355, -345.12145, -339.558736, -334.150595, -328.89141, -323.775561, -318.79743, -313.9514, -309.231853, -304.633169, -300.149731, -295.775921, -291.50612, -287.334711, -283.256075, -279.264594, -275.354649, -271.520623, -267.756898, -264.057855, -260.417876, -256.831343, -253.292638, -249.796142, -246.336237, -242.907306, -239.50373, -236.11989, -232.750169, -229.388949, -226.030611, -222.669537, -219.113424, -215.453924, -211.784931, -208.106884, -204.420225, -200.725392, -197.022826, -193.312967, -189.596255, -185.873131, -182.144034, -178.409404, -174.669682, -170.925307, -167.17672, -163.424361, -159.668669, -155.756446, -151.845102, -147.938407, -144.040131, -140.154047, -136.283923, -132.433532, -128.606643, -124.807027, -121.038456, -117.304699, -113.609528, -109.956712, -106.350024, -102.793233, -99.2901097, -95.9745038, -92.977141, -90.033515, -87.1389641, -84.2888266, -81.4784407, -78.7031446, -75.9582765, -73.2391748, -70.5411775, -67.859623, -65.1898494, -62.5271951, -59.8669982, -57.204597, -54.5353296, -51.8545344, -48.7869781, -45.521079, -42.2497097, -38.9802342, -35.7200164, -32.4764203, -29.25681, -26.0685494, -22.9190025, -19.8155333, -16.7655058, -13.776284, -10.8552319, -8.00971345, -5.24709265, -2.5747335, 0, 0.418148869, 1.19735145, 2.28698207, 3.63641503, 5.19502467, 6.9121853, 8.73727123, 10.6196568, 12.5087163, 14.3538241, 16.1043544, 17.7096817, 19.1191802, 20.2822242, 21.1481881, 21.6664461, 20.9301197, 18.7662611, 17.0610077, 15.5802507, 14.0898819, 12.3557927, 10.1438746, 7.22001919, 3.350118, -1.69993744, -8.1642556, -16.276945, -26.272114, -38.7778746, -53.4898638, -70.5239892, -89.996158, -111.176305, -134.603323, -161.0417, -190.861135, -224.431328, -262.121977, -304.302783, -352.539417, -405.641613, -463.745079, -526.985517, -595.498633, -663.610167, -738.71828, -822.411168, -916.277025, -1021.90405, -1142.679, -1276.19326, -1423.20438, -1584.14245, -1758.76898, -1947.14568, -2149.33422, -2363.20883, -2591.51584, -2839.3879, -3112.83206, -3406.6699, -3718.39484, -4046.46813, -4390.99094, -4751.31169, -5126.37126, -5515.4124, -5919.17222, -6337.54252, -6780.97961, -7237.18551, -7702.39508, -8172.8432, -8644.76474, -9152.88279, -9645.32196, -10108.6951, -10529.6149, -10894.6943, -11112.282, -11299.431, -11476.2958, -11748.6746, -12102.8214, -12489.3054, -12896.1651, -13311.4388, -13723.1648, -14119.3813, -15177.8797, -15902.3523, -16230.5541, -16100.24, -15449.1649, -14215.0835, -12335.7508, -10473.4168, -9342.80375},
        {-370.95284, -365.155592, -359.482333, -353.929317, -348.492801, -343.169041, -337.954292, -332.844812, -327.836855, -322.926678, -318.110537, -313.384688, -308.745386, -304.188889, -299.711452, -295.30933, -290.978781, -286.716059, -282.517422, -278.379125, -274.297424, -270.268574, -266.288833, -262.354457, -258.4617, -254.606819, -250.786071, -246.995711, -243.231995, -239.491179, -235.76952, -232.063273, -228.368694, -224.68204, -220.890457, -217.04705, -213.206238, -209.367816, -205.531581, -201.69733, -197.864861, -194.033969, -190.204452, -186.376107, -182.54873, -178.722118, -174.896068, -171.070376, -167.244841, -163.419258, -159.593424, -155.576722, -151.563492, -147.557656, -143.563141, -139.58387, -135.623769, -131.686761, -127.776771, -123.897725, -120.053546, -116.248159, -112.485489, -108.769461, -105.103999, -101.493027, -97.9404709, -94.5833094, -91.5553274, -88.586034, -85.6707284, -82.8047098, -79.9832776, -77.2017308, -74.4553686, -71.7394904, -69.0493953, -66.3803825, -63.7277512, -61.0868007, -58.4528301, -55.8211387, -53.1870257, -50.5457904, -47.5157724, -44.2886795, -41.0607504, -37.8395172, -34.6325122, -31.4472677, -28.2913159, -25.1721892, -22.0974198, -19.07454, -16.111082, -13.2145781, -10.3925606, -7.6525618, -5.00211391, -2.44874922, 0, -1.31521865, -1.95921213, -2.00593175, -1.52932884, -0.60335475, 0.698039207, 2.3009017, 4.13128139, 6.11522696, 8.17878707, 10.2480104, 12.2489456, 14.1076414, 15.7501464, 17.1025093, 18.0907787, 18.1492113, 17.1335995, 16.0823792, 14.816215, 13.1557714, 10.9217132, 7.93470492, 4.01541127, -1.01550313, -7.33737367, -15.1295357, -24.5713246, -35.8420757, -49.1998875, -64.7164525, -82.547477, -102.848667, -125.080618, -149.746592, -177.535213, -208.810724, -243.937362, -283.279369, -327.200985, -378.122463, -433.726286, -493.97443, -558.828873, -628.251591, -693.239764, -764.961365, -845.619567, -937.417546, -1042.55848, -1166.82171, -1304.46337, -1455.03932, -1618.33938, -1794.63103, -1983.96728, -2186.40113, -2399.22192, -2625.87446, -2872.81849, -3148.22151, -3444.22388, -3755.31047, -4078.97163, -4417.79038, -4772.01213, -5142.13844, -5528.48113, -5930.41277, -6347.83726, -6793.85346, -7253.14023, -7721.03396, -8192.87101, -8663.98776, -9162.78212, -9643.26353, -10092.503, -10497.5715, -10845.54, -11044.5515, -11213.2237, -11372.4538, -11632.9875, -11980.1571, -12361.858, -12765.2942, -13177.6699, -13586.1892, -13978.0561, -15346.2041, -16123.8386, -16343.5684, -16038.0023, -15239.7493, -13981.4182, -12295.6178, -10402.7845, -9224.40968},
        {-379.327632, -373.690349, -368.140442, -362.675624, -357.29361, -351.99211, -346.76884, -341.621511, -336.547838, -331.545532, -326.612308, -321.745877, -316.943954, -312.204252, -307.524482, -302.90236, -298.335596, -293.821906, -289.359001, -284.944596, -280.576402, -276.252133, -271.969502, -267.726223, -263.520007, -259.348569, -255.209622, -251.100878, -247.020051, -242.964853, -238.932999, -234.9222, -230.93017, -226.954622, -222.947647, -218.930727, -214.925897, -210.932349, -206.949278, -202.975876, -199.011338, -195.054856, -191.105626, -187.16284, -183.225691, -179.293375, -175.365083, -171.440011, -167.517351, -163.596297, -159.676043, -155.518517, -151.365321, -147.220792, -143.089269, -138.975087, -134.882584, -130.816097, -126.779964, -122.77852, -118.816104, -114.897052, -111.025701, -107.20639, -103.443453, -99.7412302, -96.1040567, -92.6765864, -89.6001035, -86.5895537, -83.6401778, -80.7472169, -77.9059119, -75.1115037, -72.3592334, -69.6443418, -66.9620701, -64.3076591, -61.6763498, -59.0633832, -56.4640002, -53.8734419, -51.2869491, -48.6997629, -45.7305487, -42.5707558, -39.4163601, -36.2748231, -33.1536064, -30.0601716, -27.0019801, -23.9864937, -21.0211738, -18.1134821, -15.2708801, -12.5008294, -9.81079148, -7.20822804, -4.70060059, -2.29537073, 0, -1.25832888, -1.87252738, -1.91760329, -1.46856441, -0.600418558, 0.611826482, 2.0931629, 3.7685829, 5.56307868, 7.40164244, 9.20926638, 10.9109427, 12.4316636, 13.6964212, 14.6302079, 15.1580157, 14.9118179, 13.7747469, 12.3345759, 10.4535072, 7.99374322, 4.81748646, 0.786939328, -4.23569576, -10.3882164, -17.8084201, -26.6341045, -37.0030671, -49.0531056, -62.7807613, -78.516882, -96.4416419, -116.735215, -138.857654, -163.349194, -190.942104, -222.032594, -257.016876, -296.291162, -340.251662, -391.266841, -447.160405, -507.942692, -573.624036, -644.214773, -712.50293, -787.526728, -871.102081, -965.0449, -1071.1711, -1193.65306, -1329.0701, -1477.80286, -1640.11135, -1816.00931, -2005.62105, -2209.07086, -2424.68009, -2654.78577, -2903.69184, -3176.83657, -3469.13879, -3777.23059, -4099.52642, -4437.46086, -4791.08219, -5160.68485, -5546.38094, -5947.38005, -6363.40228, -6807.14653, -7263.35737, -7727.26225, -8194.0886, -8659.06386, -9147.62112, -9617.23078, -10055.5689, -10450.3114, -10789.1344, -10984.0112, -11148.7891, -11303.5881, -11554.4661, -11887.5261, -12253.0639, -12638.973, -13033.1468, -13423.4788, -13797.8626, -14928.3751, -15627.2898, -15881.8449, -15679.2783, -15006.8282, -13851.7324, -12201.2292, -10421.0694, -9106.93731},
        {-387.751961, -382.294364, -376.885916, -371.525835, -366.213337, -360.94764, -355.727963, -350.553522, -345.423536, -340.337222, -335.293797, -330.29248, -325.332489, -320.41304, -315.533352, -310.692642, -305.890127, -301.125027, -296.396557, -291.703937, -287.046383, -282.423113, -277.833345, -273.276297, -268.751186, -264.25723, -259.793647, -255.359654, -250.954469, -246.57731, -242.227394, -237.903939, -233.606163, -229.333283, -225.103568, -220.906309, -216.730569, -212.574947, -208.438044, -204.318458, -200.214788, -196.125635, -192.049598, -187.985276, -183.931268, -179.886174, -175.848594, -171.817127, -167.790373, -163.76693, -159.745399, -155.440217, -151.140305, -146.850424, -142.575331, -138.319788, -134.088552, -129.886385, -125.718044, -121.588291, -117.501883, -113.463582, -109.478146, -105.550334, -101.684907, -97.886624, -94.1602441, -90.6578968, -87.5281733, -84.4718938, -81.4842643, -78.5604911, -75.6957801, -72.8853376, -70.1243697, -67.4080824, -64.7316818, -62.0903742, -59.4793656, -56.8938621, -54.3290698, -51.7801948, -49.2424434, -46.7110215, -43.8064296, -40.719108, -37.644001, -34.5884744, -31.5598943, -28.5656269, -25.613038, -22.7094937, -19.8623601, -17.0790033, -14.3667891, -11.7330838, -9.18525326, -6.73066358, -4.37668079, -2.13067091, 0, -0.827278069, -1.11200046, -0.92508153, -0.33743563, 0.580022885, 1.75637966, 3.12072035, 4.60213059, 6.12969603, 7.63250231, 9.03963509, 10.28018, 11.2832227, 11.9778489, 12.2931441, 12.158194, 11.3455428, 9.76191314, 7.69047247, 5.02676188, 1.66632241, -2.4953049, -7.562579, -13.6399588, -20.8319034, -29.2428716, -38.9773225, -50.1397149, -62.8345078, -76.8708818, -92.7568431, -110.685434, -130.849697, -152.632338, -176.631568, -203.661687, -234.158839, -268.559167, -307.298815, -350.813926, -401.150824, -456.645417, -517.418814, -583.592123, -655.286452, -728.308034, -808.171571, -896.076892, -993.223824, -1100.81219, -1220.53744, -1352.49804, -1498.31776, -1658.94984, -1833.9785, -2023.6026, -2228.02103, -2446.81397, -2680.93959, -2932.03095, -3202.02674, -3489.51881, -3794.15053, -4115.57989, -4453.48957, -4807.55094, -5177.63174, -5563.45425, -5964.02059, -6378.74023, -6818.7515, -7269.93058, -7727.62697, -8187.19016, -8643.96964, -9121.15678, -9579.29873, -10006.7845, -10392.0031, -10723.3436, -10917.8529, -11082.8235, -11237.2197, -11480.2305, -11799.0646, -12147.5031, -12514.426, -12888.713, -13259.2439, -13614.8985, -14379.6462, -14969.3056, -15293.1832, -15260.5857, -14780.8198, -13763.1922, -12117.0095, -10480.4299, -9020.82647},
        {-396.18132, -390.945514, -385.716835, -380.49613, -375.284249, -370.08204, -364.890352, -359.710034, -354.541934, -349.386902, -344.245785, -339.119433, -334.008695, -328.914419, -323.837454, -318.778648, -313.738852, -308.718912, -303.719679, -298.742, -293.786725, -288.854702, -283.94678, -279.063808, -274.206634, -269.376108, -264.573078, -259.798392, -255.0529, -250.337451, -245.652892, -241.000074, -236.379844, -231.793051, -227.327321, -222.93829, -218.580548, -214.25213, -209.951067, -205.675392, -201.423137, -197.192337, -192.981022, -188.787226, -184.608982, -180.444322, -176.291279, -172.147885, -168.012174, -163.882177, -159.755928, -155.304569, -150.86011, -146.427668, -142.012364, -137.619316, -133.253643, -128.920465, -124.6249, -120.372067, -116.167085, -112.015074, -107.921152, -103.890438, -99.9280516, -96.0391115, -92.2287368, -88.6552254, -85.4731975, -82.371852, -79.346378, -76.3919644, -73.5038004, -70.6770749, -67.9069769, -65.1886956, -62.51742, -59.888339, -57.2966418, -54.7375174, -52.2061547, -49.697743, -47.2074711, -44.7305281, -41.888117, -38.8702862, -35.8715791, -32.8993212, -29.9608382, -27.0634556, -24.2144991, -21.4212942, -18.6911667, -16.0314419, -13.4494457, -10.9525035, -8.54794105, -6.24308385, -4.04525754, -1.96178772, 0, -0.298802706, -0.181145413, 0.288235202, 1.04460246, 2.02321969, 3.15935021, 4.38825734, 5.64520442, 6.86545475, 7.98427167, 8.93691849, 9.65865855, 10.0847552, 10.1504716, 9.79107132, 8.94181753, 7.4517566, 5.24383517, 2.44833112, -1.01796728, -5.23827177, -10.2957941, -16.273746, -23.2553392, -31.3237855, -40.5622966, -51.0540843, -62.8823602, -76.1303362, -90.5173192, -106.623857, -124.642334, -144.765132, -166.245506, -189.745403, -216.177205, -246.015036, -279.733015, -317.805267, -360.705912, -410.008119, -464.752471, -525.198061, -591.603978, -664.229315, -742.209642, -827.208451, -919.765714, -1020.4214, -1129.71549, -1246.75337, -1375.26296, -1518.1359, -1677.02358, -1851.04519, -2040.45682, -2245.51455, -2466.92251, -2704.38694, -2957.1253, -3223.84175, -3506.58185, -3807.88173, -4128.58254, -4466.65332, -4821.38146, -5192.27648, -5578.68337, -5979.13259, -6392.61533, -6827.90237, -7272.70078, -7722.61651, -8173.25551, -8620.22373, -9084.83173, -9530.55471, -9946.57248, -10322.0648, -10646.2116, -10841.6217, -11008.4263, -11164.3357, -11401.3219, -11707.6998, -12040.8416, -12390.6094, -12746.8651, -13099.4705, -13438.2877, -13810.8523, -14286.3108, -14688.7301, -14842.1768, -14570.7177, -13698.4196, -12049.3492, -10558.8969, -8978.88581},
        {-404.288181, -399.416273, -394.496159, -389.530789, -384.523115, -379.476087, -374.392656, -369.275772, -364.128386, -358.953449, -353.75391, -348.532722, -343.292833, -338.037196, -332.76876, -327.490476, -322.205295, -316.916168, -311.626044, -306.337875, -301.054611, -295.779204, -290.514602, -285.263758, -280.029621, -274.815142, -269.623272, -264.456962, -259.319162, -254.212823, -249.140895, -244.106328, -239.112075, -234.161084, -229.422809, -224.813447, -220.247194, -215.721598, -211.234206, -206.782566, -202.364224, -197.976729, -193.617626, -189.284465, -184.974791, -180.686153, -176.416097, -172.162171, -167.921922, -163.692898, -159.472646, -154.90858, -150.35597, -145.819954, -141.30567, -136.818256, -132.362848, -127.944584, -123.568602, -119.24004, -114.964034, -110.745723, -106.590244, -102.502734, -98.4883312, -94.5521728, -90.6993964, -87.0884088, -83.8739368, -80.7450117, -77.696835, -74.7246086, -71.823534, -68.988813, -66.2156473, -63.4992385, -60.8347884, -58.2174987, -55.6425711, -53.1052072, -50.6006087, -48.1239775, -45.6705151, -43.2354232, -40.4313529, -37.4528276, -34.4989688, -31.5773927, -28.6957155, -25.8615534, -23.0825226, -20.3662392, -17.7203196, -15.1523799, -12.6700362, -10.2809049, -7.99260211, -5.81274401, -3.74894682, -1.80882675, 0, -0.418199869, -0.439435975, -0.127967523, 0.451946283, 1.23604624, 2.16007313, 3.15976777, 4.17087093, 5.12912343, 5.97026604, 6.63003957, 7.04418482, 7.14844257, 6.87855362, 6.17025877, 4.95929881, 3.05552273, 0.376686786, -2.85671296, -6.73591253, -11.3521479, -16.7966552, -23.1606704, -30.5354294, -39.0121684, -48.6821233, -59.6365302, -71.9666251, -85.7636439, -100.863404, -117.706213, -136.459934, -157.292428, -179.345648, -203.30041, -230.111106, -260.253371, -294.202841, -332.435149, -375.42593, -424.331288, -478.739289, -538.992434, -605.433222, -678.404153, -758.216479, -845.251758, -939.860303, -1042.39243, -1153.19844, -1270.80829, -1399.61754, -1542.77618, -1702.02432, -1876.2236, -2065.52804, -2270.09167, -2490.18608, -2725.821, -2976.87785, -3242.72792, -3524.79938, -3825.04364, -4144.12734, -4480.54013, -4833.77094, -5204.12632, -5591.30716, -5992.01646, -6404.65301, -6835.17337, -7273.25582, -7714.68296, -8155.23743, -8590.70186, -9039.81938, -9469.67199, -9870.30216, -10231.7524, -10544.0652, -10733.927, -10896.9738, -11050.4005, -11284.4785, -11587.477, -11917.2164, -12263.4882, -12616.084, -12964.7952, -13299.4132, -13542.1073, -13971.8377, -14390.6604, -14600.6317, -14403.8076, -13602.2442, -11997.9976, -10568.0417, -9009.78682},
        {-412.375627, -407.936773, -403.387906, -398.734301, -393.981236, -389.133986, -384.197828, -379.178038, -374.079893, -368.908668, -363.669641, -358.368087, -353.009283, -347.598505, -342.14103, -336.642133, -331.107092, -325.541183, -319.949681, -314.337863, -308.711007, -303.074387, -297.43328, -291.792963, -286.158712, -280.535803, -274.929513, -269.345118, -263.787894, -258.263118, -252.776066, -247.332014, -241.936239, -236.594017, -231.56202, -226.714539, -221.922842, -217.184049, -212.495277, -207.853644, -203.256268, -198.700267, -194.18276, -189.700864, -185.251698, -180.832379, -176.440025, -172.071755, -167.724687, -163.395939, -159.082628, -154.418546, -149.772013, -145.148026, -140.551576, -135.987661, -131.461272, -126.977405, -122.541054, -118.157213, -113.830877, -109.56704, -105.370696, -101.24684, -97.200465, -93.2365663, -89.360138, -85.7269999, -82.4893504, -79.3411177, -76.2775186, -73.29377, -70.3850889, -67.5466921, -64.7737966, -62.0616191, -59.4053767, -56.8002862, -54.2415646, -51.7244286, -49.2440953, -46.7957815, -44.374704, -41.9760799, -39.198842, -36.2446844, -33.3200645, -30.4330593, -27.5917458, -24.8042012, -22.0785026, -19.422727, -16.8449515, -14.3532531, -11.955709, -9.66039627, -7.4753919, -5.40877301, -3.46861668, -1.66299998, 0, -0.886753461, -1.33915771, -1.42396515, -1.20792818, -0.757799202, -0.140330619, 0.577725166, 1.32961575, 2.04858874, 2.66789173, 3.12077231, 3.34047809, 3.26025666, 2.81335563, 1.93302258, 0.552505124, -1.62526045, -4.69992816, -8.21730817, -12.2935053, -17.0446244, -22.5867702, -29.0360477, -36.5085616, -45.1204167, -54.9877178, -66.2265699, -78.9530777, -93.283346, -109.294243, -127.155496, -146.994983, -168.940579, -192.009828, -216.885772, -244.547542, -275.456113, -310.072461, -348.857561, -392.272388, -441.123618, -495.421314, -555.558815, -621.929459, -694.926586, -774.869086, -862.243359, -957.461355, -1060.93503, -1173.07632, -1292.92767, -1423.94441, -1568.91846, -1729.3378, -1904.02815, -2093.01053, -2296.30597, -2513.13674, -2744.50414, -2992.28087, -3258.24792, -3541.39354, -3841.54283, -4158.57734, -4492.47425, -4843.16685, -5212.22087, -5599.99285, -6000.85359, -6412.55979, -6838.15261, -7269.29122, -7701.90304, -8131.91549, -8555.25598, -8987.89302, -9400.70268, -9784.60208, -10130.5084, -10429.3387, -10611.1127, -10768.2434, -10917.7469, -11150.5887, -11454.5132, -11786.4526, -12135.6921, -12491.5166, -12843.2113, -13180.0613, -13427.2239, -13846.5507, -14250.9519, -14453.3375, -14266.6175, -13503.702, -11977.501, -10567.5162, -9106.85657},
        {-420.690164, -416.696091, -412.529087, -408.196784, -403.706814, -399.06681, -394.284403, -389.367227, -384.322911, -379.15909, -373.883394, -368.503457, -363.026909, -357.461383, -351.814512, -346.093927, -340.307261, -334.462144, -328.566211, -322.627092, -316.652419, -310.649826, -304.626943, -298.591403, -292.550838, -286.51288, -280.485162, -274.475315, -268.490971, -262.539762, -256.629321, -250.76728, -244.96127, -239.218924, -233.88392, -228.788799, -223.762063, -218.800437, -213.90065, -209.059427, -204.273495, -199.539582, -194.854413, -190.214716, -185.617217, -181.058643, -176.535721, -172.045178, -167.58374, -163.148133, -158.735086, -153.967635, -149.224298, -144.509901, -139.829272, -135.187239, -130.58863, -126.038271, -121.54099, -117.101616, -112.724975, -108.415895, -104.179204, -100.019729, -95.9422985, -91.951739, -88.0528785, -84.3988897, -81.1393835, -77.9732369, -74.8956608, -71.9018661, -68.9870638, -66.1464649, -63.3752803, -60.668721, -58.0219979, -55.430322, -52.8889042, -50.3929555, -47.9376869, -45.5183093, -43.1300337, -40.768071, -38.0154628, -35.0821741, -32.183124, -29.3268992, -26.5220863, -23.7772719, -21.1010428, -18.5019855, -15.9886867, -13.569733, -11.2537112, -9.04920777, -6.96480945, -5.00910286, -3.19067466, -1.51811149, 0, -1.49940761, -2.50369333, -3.08328712, -3.30861892, -3.2501187, -2.97821642, -2.56334202, -2.07592547, -1.58639672, -1.16518572, -0.882722442, -0.80943683, -1.01575884, -1.57211844, -2.54894557, -4.0166702, -6.4079365, -9.84491933, -13.5787545, -17.7574894, -22.5291713, -28.0418475, -34.4435652, -41.8823717, -50.5063144, -60.4634405, -71.9017973, -84.9694321, -99.8143922, -116.814406, -135.803623, -156.861187, -180.06624, -204.276663, -230.182229, -258.798384, -290.570649, -325.944546, -365.365596, -409.279322, -458.239696, -512.550783, -572.636884, -638.922304, -711.831345, -791.198566, -878.185451, -973.363739, -1077.30517, -1190.58148, -1313.18281, -1446.97334, -1594.03558, -1755.45378, -1930.27408, -2118.45768, -2319.96582, -2533.08238, -2759.8271, -3004.04949, -3270.0437, -3554.23681, -3854.24695, -4169.12365, -4500.34188, -4848.26332, -5215.64173, -5603.45897, -6003.92577, -6414.21432, -6834.69608, -7258.85173, -7682.74601, -8102.44363, -8514.00932, -8930.77211, -9327.21634, -9695.09068, -10026.1438, -10312.1243, -10486.1372, -10637.6706, -10783.6061, -11015.6827, -11321.151, -11656.0709, -12009.2525, -12369.5058, -12725.6409, -13066.4678, -13361.8473, -13782.5075, -14163.7643, -14340.9338, -14149.332, -13424.2748, -12001.0783, -10603.6875, -9266.59295},
        {-429.7245, -426.12283, -422.288209, -418.230554, -413.959784, -409.48582, -404.818579, -399.967981, -394.943946, -389.75639, -384.415235, -378.930399, -373.3118, -367.569358, -361.712992, -355.752621, -349.698163, -343.559539, -337.346666, -331.069464, -324.737852, -318.361749, -311.951074, -305.515745, -299.065682, -292.610805, -286.161031, -279.726279, -273.31647, -266.941521, -260.611353, -254.335883, -248.125031, -241.988716, -236.360912, -231.030592, -225.781629, -220.610183, -215.512409, -210.484467, -205.522513, -200.622705, -195.781201, -190.994158, -186.257734, -181.568086, -176.921372, -172.31375, -167.741377, -163.20041, -158.687008, -153.793619, -148.928861, -144.097643, -139.304875, -134.555465, -129.854323, -125.206359, -120.61648, -116.089598, -111.63062, -107.244455, -102.936015, -98.710206, -94.571939, -90.5261228, -86.5776665, -82.8824559, -79.5927516, -76.4020854, -73.3055791, -70.2983549, -67.3755348, -64.5322407, -61.7635946, -59.0647187, -56.4307349, -53.8567652, -51.3379317, -48.8693563, -46.4461612, -44.0634682, -41.7163994, -39.4000768, -36.6819572, -33.7799907, -30.9172673, -28.1027877, -25.3455525, -22.6545622, -20.0388175, -17.5073188, -15.0690668, -12.733062, -10.5083051, -8.40379656, -6.428537, -4.59152698, -2.90176709, -1.36825791, 0, -2.27226002, -3.95557114, -5.12637245, -5.86110299, -6.23620183, -6.32810804, -6.21326068, -5.9680988, -5.66906148, -5.39258778, -5.21511677, -5.21308749, -5.46293903, -6.04111043, -7.02404077, -8.48816911, -10.9790367, -14.6400971, -18.4867251, -22.6958816, -27.4445275, -32.9096238, -39.2681314, -46.6970113, -55.3732245, -65.4737319, -77.1754945, -90.6554733, -106.090629, -124.189268, -144.40218, -166.746921, -191.24105, -216.426151, -243.057769, -272.285038, -304.568309, -340.367929, -380.144248, -424.357614, -473.648745, -528.242728, -588.56462, -655.039482, -728.092374, -806.931329, -893.502688, -988.535769, -1092.75989, -1206.90436, -1331.95605, -1468.07194, -1616.11748, -1776.68448, -1949.8061, -2135.7663, -2334.84907, -2545.96405, -2771.0819, -3013.67255, -3278.02179, -3560.55031, -3858.84517, -4171.95108, -4501.38257, -4847.52047, -5213.40632, -5600.11075, -5998.94857, -6406.75302, -6822.18421, -7239.9675, -7656.30379, -8067.39399, -8469.43899, -8873.71976, -9257.5871, -9613.47189, -9933.80499, -10211.0173, -10381.5531, -10531.1543, -10676.2458, -10906.3273, -11208.8985, -11541.0117, -11891.6889, -12249.9525, -12604.8246, -12945.3275, -13246.2342, -13656.1446, -14025.8771, -14206.2496, -14048.0804, -13402.1876, -12119.3894, -10750.9716, -9485.96723},
        {-439.810344, -436.499055, -432.9013, -429.029108, -424.89451, -420.509535, -415.886215, -411.036578, -405.972655, -400.706477, -395.250074, -389.615475, -383.814711, -377.859812, -371.762808, -365.535729, -359.190607, -352.73947, -346.194349, -339.567274, -332.870275, -326.115383, -319.314627, -312.480038, -305.623646, -298.757482, -291.893574, -285.043955, -278.220652, -271.435698, -264.701122, -258.028954, -251.431224, -244.919963, -239.022848, -233.48324, -228.038301, -222.683329, -217.413618, -212.224467, -207.11117, -202.069025, -197.093328, -192.179376, -187.322464, -182.517889, -177.760948, -173.046937, -168.371153, -163.728891, -159.115449, -154.059299, -149.03268, -144.04101, -139.089704, -134.184179, -129.329851, -124.532135, -119.79645, -115.128209, -110.532831, -106.015731, -101.582325, -97.2380296, -92.9882611, -88.8384356, -84.7939694, -81.0223566, -77.6872001, -74.4596424, -71.334593, -68.3069612, -65.3716564, -62.523588, -59.7576654, -57.068798, -54.4518952, -51.9018663, -49.4136208, -46.9820679, -44.6021172, -42.268678, -39.9766596, -37.7209716, -35.0557317, -32.2055167, -29.4002009, -26.6489984, -23.961123, -21.3457887, -18.8122093, -16.3695988, -14.0271711, -11.7941401, -9.67971976, -7.69312395, -5.84356658, -4.14026159, -2.59242289, -1.20926439, 0, -3.18109156, -5.64603852, -7.48001705, -8.76820333, -9.59577355, -10.0479039, -10.2097705, -10.1665496, -10.0034174, -9.80554997, -9.65812357, -9.64631437, -9.85529855, -10.3702523, -11.2763517, -12.6587731, -15.1143764, -18.8014353, -22.6477462, -26.8481318, -31.5974147, -37.0904177, -43.5219634, -51.0868746, -59.979974, -70.3960842, -82.530028, -96.5766281, -112.730707, -132.026221, -153.511177, -177.128658, -202.821747, -228.592729, -255.355086, -284.539844, -316.672325, -352.27785, -391.881741, -436.009319, -485.878615, -541.111417, -602.097516, -669.226704, -742.888772, -821.808591, -908.457103, -1003.64033, -1108.16429, -1222.83502, -1349.36665, -1486.54715, -1633.47417, -1790.22576, -1958.88165, -2140.62286, -2336.63041, -2548.94997, -2777.70139, -3022.06123, -3282.01685, -3558.34254, -3852.29563, -4164.33334, -4493.55712, -4839.69074, -5204.59366, -5588.54335, -5983.98636, -6387.79885, -6798.42291, -7210.92786, -7621.64781, -8026.91686, -8423.06913, -8820.23079, -9197.49586, -9547.75043, -9863.8806, -10138.7724, -10312.7869, -10466.352, -10614.8646, -10840.4401, -11132.066, -11450.9304, -11787.4109, -12131.8848, -12474.7298, -12806.3234, -13008.5906, -13378.8635, -13763.2435, -14007.832, -13958.7304, -13462.0403, -12363.8631, -11060.5052, -9764.00876},
        {-450.224459, -447.14661, -443.736312, -440.007372, -435.973595, -431.648786, -427.046752, -422.181298, -417.066229, -411.715352, -406.142472, -400.361394, -394.385925, -388.229869, -381.907033, -375.431223, -368.816243, -362.0759, -355.223999, -348.274346, -341.240746, -334.137005, -326.97693, -319.774324, -312.542995, -305.296748, -298.049388, -290.814721, -283.606553, -276.438689, -269.324936, -262.279098, -255.314981, -248.446392, -242.276043, -236.510899, -230.850848, -225.290583, -219.824799, -214.44819, -209.155451, -203.941276, -198.80036, -193.727398, -188.717083, -183.76411, -178.863173, -174.008968, -169.196188, -164.419528, -159.673683, -154.436245, -149.23022, -144.061512, -138.936026, -133.859666, -128.838335, -123.877939, -118.984381, -114.163565, -109.421396, -104.763778, -100.196616, -95.7258123, -91.3572724, -87.0969003, -82.9506002, -79.0975577, -75.7127982, -72.4448457, -69.2883714, -66.2380463, -63.2885415, -60.4345281, -57.6706771, -54.9916596, -52.3921467, -49.8668095, -47.4103189, -45.0173462, -42.6825624, -40.4006384, -38.1662456, -35.9740548, -33.3654901, -30.5713795, -27.8279989, -25.1447279, -22.5309464, -19.9960342, -17.5493711, -15.200337, -12.9583116, -10.8326747, -8.8328062, -6.96808586, -5.24789349, -3.68160891, -2.27861192, -1.04828235, 0, -3.93750379, -7.05926838, -9.45741227, -11.2240539, -12.4513119, -13.2313046, -13.6561505, -13.8179681, -13.808876, -13.7209926, -13.6464363, -13.6773257, -13.9057792, -14.4239154, -15.3238527, -16.6977097, -19.1748126, -22.9240239, -26.8223491, -31.0770226, -35.8952789, -41.4843525, -48.0514777, -55.8038892, -64.9488212, -75.6935083, -88.245185, -102.811086, -119.598445, -139.897472, -162.435335, -187.094377, -213.75694, -239.957029, -266.751153, -295.822048, -327.756557, -363.141523, -402.563789, -446.610199, -497.117611, -553.042414, -614.726884, -682.513295, -756.743924, -835.668598, -922.245151, -1017.33897, -1121.81545, -1236.53996, -1363.81027, -1501.30873, -1646.51312, -1799.06362, -1963.01529, -2140.44099, -2333.41357, -2547.24977, -2780.04135, -3026.34131, -3281.39231, -3550.78153, -3839.87148, -4150.6791, -4479.55248, -4825.44183, -5188.68636, -5568.59638, -5959.38912, -6358.16273, -6764.00621, -7171.72061, -7577.71494, -7978.39823, -8370.17951, -8762.21301, -9134.97623, -9481.69191, -9795.58276, -10069.8715, -10248.5698, -10406.9188, -10559.2317, -10779.2636, -11057.7986, -11360.8533, -11680.4501, -12008.6115, -12337.3599, -12658.7178, -12760.2651, -13091.1744, -13492.9818, -13807.2233, -13875.435, -13539.1531, -12639.9137, -11430.8839, -10113.2324},
        {-460.767447, -457.864138, -454.592242, -450.966871, -447.003137, -442.716152, -438.121028, -433.232876, -428.066808, -422.637936, -416.961373, -411.052229, -404.925616, -398.596647, -392.080433, -385.392086, -378.546718, -371.559441, -364.445366, -357.219606, -349.897272, -342.493476, -335.023329, -327.501945, -319.944433, -312.365907, -304.781478, -297.206259, -289.655359, -282.143893, -274.686971, -267.299705, -259.997207, -252.794589, -246.339878, -240.318415, -234.407935, -228.603011, -222.898215, -217.28812, -211.767298, -206.330323, -200.971765, -195.686198, -190.468195, -185.312328, -180.213169, -175.165291, -170.163267, -165.201669, -160.275069, -154.839951, -149.440642, -144.083378, -138.774397, -133.519935, -128.326232, -123.199523, -118.146045, -113.172037, -108.283735, -103.487376, -98.7891988, -94.1954392, -89.7123349, -85.3461232, -81.1030412, -77.1713682, -73.7350124, -70.4247635, -67.2350576, -64.1603309, -61.1950197, -58.3335601, -55.5703885, -52.8999409, -50.3166536, -47.8149628, -45.3893047, -43.0341156, -40.7438317, -38.512889, -36.335724, -34.2067728, -31.6555671, -28.9187742, -26.2386488, -23.6247141, -21.0864933, -18.6335099, -16.2752869, -14.0213477, -11.8812155, -9.86441349, -7.98046503, -6.23889333, -4.64922163, -3.2209732, -1.96367128, -0.886839131, 0, -4.38427731, -7.91136094, -10.6764367, -12.7746905, -14.3013082, -15.3514755, -16.0203784, -16.4032026, -16.5951341, -16.6913586, -16.787062, -16.9774301, -17.3576489, -18.022904, -19.0683814, -20.589267, -23.2306981, -27.166425, -31.2476956, -35.6875424, -40.6989981, -46.4950952, -53.2888664, -61.2933443, -70.7215615, -81.7865505, -94.7013441, -109.678975, -126.932475, -147.88925, -171.102691, -196.421517, -223.69445, -250.173295, -277.005232, -306.029948, -337.875239, -373.168901, -412.538728, -456.612517, -507.736429, -564.296911, -626.585555, -694.893956, -769.513704, -848.301415, -934.592405, -1029.28701, -1133.28557, -1247.48842, -1374.53846, -1511.46367, -1654.60031, -1803.30896, -1963.12492, -2136.81114, -2327.13061, -2542.0801, -2777.99928, -3025.51821, -3275.74556, -3538.47334, -3822.71704, -4131.85412, -4459.70881, -4804.49042, -5164.96663, -5539.49153, -5924.37189, -6317.07257, -6718.14367, -7121.49014, -7523.50885, -7920.5967, -8309.15055, -8697.32778, -9066.82466, -9411.09794, -9723.60439, -9997.80074, -10180.8198, -10343.325, -10498.6261, -10712.1825, -10976.6616, -11263.0019, -11564.9571, -11876.2812, -12190.7279, -12502.0511, -12521.6662, -12819.8228, -13237.2865, -13614.8232, -13793.1986, -13613.1783, -12915.5281, -11836.4481, -10539.4568},
        {-471.178999, -468.386737, -465.202798, -461.64295, -457.722958, -453.458589, -448.86561, -443.959788, -438.756888, -433.272678, -427.522924, -421.523393, -415.28985, -408.838063, -402.183798, -395.342822, -388.330902, -381.163803, -373.857292, -366.427136, -358.889102, -351.258956, -343.552464, -335.785393, -327.97351, -320.132582, -312.278374, -304.426653, -296.593186, -288.793739, -281.04408, -273.359973, -265.757187, -258.251488, -251.491711, -245.163773, -238.94598, -232.833555, -226.821721, -220.905699, -215.080711, -209.34198, -203.684728, -198.104177, -192.59555, -187.154068, -181.774955, -176.453431, -171.18472, -165.964043, -160.786623, -155.139104, -149.536312, -143.984494, -138.489895, -133.058765, -127.697349, -122.411895, -117.208651, -112.093863, -107.073778, -102.154645, -97.3427089, -92.6442182, -88.0654197, -83.6125607, -79.2918883, -75.2950585, -71.8092759, -68.4583956, -65.2366456, -62.1382538, -59.1574481, -56.2884562, -53.5255062, -50.8628259, -48.2946431, -45.8151857, -43.4186816, -41.0993588, -38.8514449, -36.6691681, -34.546756, -32.4784366, -29.9822746, -27.3005949, -24.6814455, -22.1345067, -19.6694588, -17.2959822, -15.023757, -12.8624636, -10.8217822, -8.91139322, -7.14097685, -5.52021339, -4.05878315, -2.7663664, -1.65264343, -0.727294538, 0, -4.29149641, -7.78683174, -10.5773814, -12.754521, -14.4096258, -15.6340714, -16.5192331, -17.1564866, -17.6372072, -18.0527703, -18.4945516, -19.0539263, -19.82227, -20.8909581, -22.3513661, -24.2948695, -27.3674894, -31.7398365, -36.2487303, -41.1043989, -46.5170704, -52.6969731, -59.8543349, -68.199384, -77.9423486, -89.2934568, -102.462937, -117.661016, -135.097924, -156.144966, -179.425564, -204.801621, -232.135043, -258.746384, -285.768222, -315.010832, -347.098524, -382.655606, -422.30639, -466.675185, -518.305362, -575.318106, -637.96226, -706.486665, -781.14016, -859.557908, -945.25585, -1039.13625, -1142.10136, -1255.05345, -1380.61722, -1515.86728, -1656.90685, -1803.08424, -1960.37322, -2131.77285, -2320.28219, -2535.0613, -2771.54805, -3018.45924, -3264.71758, -3522.36242, -3802.49711, -4109.22213, -4434.77196, -4776.82779, -5132.87221, -5500.5349, -5878.17368, -6263.73452, -6660.0652, -7059.45569, -7458.15517, -7852.41279, -8238.4777, -8623.12351, -8989.44383, -9331.05671, -9641.58019, -9914.63234, -10099.651, -10263.8879, -10419.8954, -10626.1481, -10877.0208, -11147.7543, -11433.6802, -11730.13, -12032.4354, -12335.928, -12317.7794, -12597.6659, -13023.8144, -13444.4519, -13707.8054, -13662.1019, -13155.5685, -12252.6578, -11060.0153},
        {-480.610388, -477.835605, -474.670965, -471.131587, -467.23259, -462.989093, -458.416214, -453.529073, -448.342788, -442.872479, -437.133264, -431.140263, -424.908594, -418.453377, -411.789729, -404.932771, -397.897621, -390.699398, -383.353222, -375.87421, -368.277482, -360.578157, -352.791354, -344.932191, -337.015789, -329.057265, -321.071738, -313.074329, -305.080154, -297.104335, -289.161988, -281.268234, -273.438192, -265.686979, -258.568971, -251.818316, -245.162682, -238.599688, -232.126952, -225.742094, -219.442733, -213.226489, -207.090979, -201.033824, -195.052642, -189.145052, -183.308674, -177.541127, -171.840029, -166.202999, -160.627658, -154.750223, -148.937548, -143.195087, -137.528294, -131.942621, -126.443523, -121.036454, -115.726866, -110.520213, -105.421949, -100.437527, -95.5724016, -90.8320251, -86.2218516, -81.7473345, -77.4139275, -73.4036028, -69.8940947, -66.5267998, -63.2957289, -60.1948923, -57.2183007, -54.3599646, -51.6138945, -48.974101, -46.4345947, -43.989386, -41.6324855, -39.3579038, -37.1596514, -35.0317389, -32.9681768, -30.9629756, -28.5176249, -25.8827554, -23.3149172, -20.8241041, -18.4203099, -16.1135285, -13.9137537, -11.8309791, -9.87519872, -8.05640624, -6.3845955, -4.86976028, -3.52189438, -2.3509916, -1.36704575, -0.580050617, 0, -2.7269627, -4.9990233, -6.88495307, -8.45352331, -9.77350527, -10.9136702, -11.9427895, -12.9296343, -13.9429759, -15.0515856, -16.3242347, -17.8296944, -19.6367361, -21.8141309, -24.4306503, -27.5550653, -31.8165537, -37.3639445, -42.9984237, -48.9088496, -55.2840804, -62.3129745, -70.1843902, -79.0871858, -89.2102197, -100.74235, -113.872436, -128.789334, -145.681904, -165.360388, -187.164419, -211.09644, -237.158897, -263.737955, -291.644195, -322.119208, -355.650317, -392.724849, -433.830131, -479.453488, -531.322926, -588.307494, -650.651775, -718.600355, -792.397817, -869.876285, -954.29592, -1046.50442, -1147.34949, -1257.67882, -1379.32042, -1510.94353, -1650.72144, -1798.62379, -1958.28768, -2131.7036, -2320.86203, -2531.71366, -2761.38873, -3002.69731, -3248.3702, -3506.66228, -3785.90481, -4088.95719, -4409.40712, -4745.09813, -5093.34318, -5451.84823, -5820.2649, -6197.14424, -6589.19706, -6985.55929, -7381.95745, -7774.11804, -8157.76757, -8536.05718, -8895.43259, -9229.76418, -9532.92229, -9798.77728, -9977.75132, -10136.1284, -10286.3996, -10484.6042, -10725.9953, -10987.6811, -11265.8217, -11556.5778, -11856.1097, -12160.5778, -12217.8069, -12516.6042, -12932.9867, -13342.9714, -13622.5751, -13647.815, -13294.708, -12665.7894, -11803.4805},
        {-490.273963, -487.48227, -484.308992, -480.768292, -476.874328, -472.64126, -468.083249, -463.214454, -458.049036, -452.601155, -446.88497, -440.914642, -434.70433, -428.268195, -421.620396, -414.775094, -407.746449, -400.548621, -393.195769, -385.702053, -378.081635, -370.348673, -362.517328, -354.601759, -346.616127, -338.574592, -330.491314, -322.380452, -314.256167, -306.132619, -298.023967, -289.944373, -281.907995, -273.928994, -266.447297, -259.255226, -252.139984, -245.101913, -238.141357, -231.258658, -224.45416, -217.728205, -211.081137, -204.513299, -198.025033, -191.616683, -185.288591, -179.041101, -172.874556, -166.789299, -160.785672, -154.673657, -148.648086, -142.713427, -136.874152, -131.134729, -125.499628, -119.973319, -114.560272, -109.264956, -104.091841, -99.0453965, -94.1300925, -89.3503985, -84.7107843, -80.2157195, -75.8696738, -71.8393323, -68.2874852, -64.8840454, -61.6229667, -58.498203, -55.5037082, -52.6334364, -49.8813413, -47.241377, -44.7074974, -42.2736563, -39.9338078, -37.6819057, -35.511904, -33.4177567, -31.3934175, -29.4328405, -27.0283599, -24.4352903, -21.9149756, -19.4776481, -17.1335404, -14.8928849, -12.7659139, -10.7628601, -8.89395588, -7.16943362, -5.59952582, -4.19446493, -2.9644834, -1.91981369, -1.07068825, -0.427339538, 0, -1.2183943, -2.313872, -3.33311124, -4.32279017, -5.32958691, -6.40017962, -7.58124643, -8.91946549, -10.4615149, -12.2540729, -14.3438175, -16.777427, -19.6015793, -22.8629528, -26.6082255, -30.8840756, -36.298171, -42.9773312, -49.6993289, -56.6310544, -63.9393978, -71.7912494, -80.3534995, -89.7930383, -100.276756, -111.971543, -125.044289, -139.661885, -155.991221, -174.296507, -194.611629, -217.074282, -241.82216, -268.26481, -296.903999, -328.435669, -363.215959, -401.601008, -443.946954, -490.609936, -542.533568, -599.307718, -661.173582, -728.372359, -801.145247, -877.637438, -960.710137, -1051.12854, -1149.65786, -1257.06328, -1374.19937, -1501.63275, -1639.92641, -1789.77013, -1952.11253, -2127.78604, -2317.62307, -2523.89697, -2745.67173, -2980.43935, -3225.29111, -3484.13573, -3762.02039, -4060.38394, -4374.55114, -4702.6532, -5042.05722, -5390.69631, -5749.30542, -6117.03462, -6504.2777, -6897.21122, -7291.09321, -7681.18168, -8062.73467, -8436.4087, -8790.71365, -9119.55795, -9416.84997, -9676.49812, -9848.84801, -10000.4127, -10143.7703, -10333.3355, -10565.1262, -10817.7284, -11088.0316, -11372.9253, -11669.2991, -11974.0423, -12136.0716, -12461.7252, -12860.4033, -13241.506, -13514.4336, -13588.5862, -13373.364, -13024.2323, -12508.7828},
        {-500.59357, -497.775061, -494.583588, -491.032353, -487.134557, -482.903402, -478.352088, -473.493816, -468.341788, -462.909205, -457.209268, -451.255179, -445.060138, -438.637346, -432.000006, -425.161317, -418.134482, -410.932701, -403.569176, -396.057107, -388.409697, -380.640145, -372.761654, -364.787425, -356.730658, -348.604555, -340.422317, -332.197145, -323.942241, -315.670805, -307.39604, -299.131145, -290.889322, -282.683773, -274.841434, -267.21204, -259.641554, -252.132995, -244.689385, -237.313742, -230.009086, -222.778438, -215.624817, -208.551243, -201.560736, -194.656316, -187.841002, -181.117814, -174.489773, -167.959898, -161.531209, -155.190014, -148.956408, -142.833771, -136.825487, -130.934937, -125.165503, -119.520568, -114.003514, -108.617722, -103.366576, -98.2534568, -93.2817468, -88.4548282, -83.776083, -79.2488935, -74.8766417, -70.8045079, -67.1742676, -63.6968551, -60.3664605, -57.1772736, -54.1234845, -51.199283, -48.398859, -45.7164025, -43.1461035, -40.6821519, -38.3187376, -36.0500506, -33.8702807, -31.773618, -29.7542524, -27.8063738, -25.4272833, -22.8700668, -20.393782, -18.0087441, -15.7252678, -13.5536681, -11.50426, -9.58735834, -7.81327814, -6.1923343, -4.73484175, -3.45111544, -2.35147028, -1.44622123, -0.745683209, -0.260171154, 0, -0.258693305, -0.625197321, -1.1306918, -1.80635649, -2.68337114, -3.79291551, -5.16616933, -6.83431237, -8.82852437, -11.1799851, -13.9198743, -17.0793717, -20.689657, -24.7819101, -29.3873106, -34.5370384, -40.8251912, -48.3633658, -55.9094746, -63.6153228, -71.6327153, -80.1134573, -89.2093539, -99.07221, -109.853831, -121.706021, -134.780587, -149.229332, -165.204062, -182.627078, -201.963841, -223.435007, -247.26123, -273.465258, -302.366702, -334.337948, -369.659026, -408.609965, -451.470794, -498.521543, -550.31662, -606.778167, -668.132533, -734.606062, -806.425101, -882.108821, -964.017538, -1052.80439, -1149.12252, -1253.62507, -1366.1261, -1489.14337, -1625.50475, -1776.60224, -1940.8963, -2118.1636, -2308.18079, -2510.00094, -2724.28876, -2952.49838, -3195.30219, -3453.0384, -3727.98717, -4020.42617, -4327.24001, -4646.87076, -4976.83124, -5315.32263, -5663.95336, -6022.40454, -6403.95081, -6792.58841, -7183.29296, -7571.04011, -7950.80549, -8322.93234, -8675.68679, -9002.70257, -9297.61343, -9554.05309, -9721.23082, -9866.82113, -10003.4968, -10183.9598, -10405.0379, -10646.8805, -10907.1493, -11183.5063, -11473.6133, -11775.1322, -12060.8609, -12416.857, -12789.6827, -13125.9003, -13372.0722, -13474.7607, -13380.528, -13288.4159, -13077.6369},
        {-512.726318, -509.924824, -506.745878, -503.202382, -499.30724, -495.073353, -490.513622, -485.640951, -480.468242, -475.008395, -469.274315, -463.278903, -457.035061, -450.555691, -443.853695, -436.941976, -429.833435, -422.540976, -415.077499, -407.455907, -399.689103, -391.789988, -383.771465, -375.646436, -367.427802, -359.128467, -350.761331, -342.339299, -333.87527, -325.382149, -316.872836, -308.360234, -299.857245, -291.376772, -283.174694, -275.137314, -267.148014, -259.211811, -251.333722, -243.518764, -235.771954, -228.098309, -220.502847, -212.990583, -205.566537, -198.235723, -191.00316, -183.873864, -176.852852, -169.945143, -163.155751, -156.621071, -150.211895, -143.930393, -137.778734, -131.759087, -125.873621, -120.124505, -114.513908, -109.044, -103.716949, -98.5349252, -93.5000967, -88.6146331, -83.8807033, -79.3004765, -74.8761217, -70.7186955, -66.9366398, -63.3075518, -59.8265419, -56.4887205, -53.2891979, -50.2230845, -47.2854907, -44.4715268, -41.7763032, -39.1949304, -36.7225186, -34.3541782, -32.0850196, -29.9101532, -27.8246893, -25.8237384, -23.4417957, -20.9139675, -18.4808963, -16.1526406, -13.9392584, -11.8508079, -9.89734744, -8.08893518, -6.43562934, -4.94748814, -3.6345698, -2.50693253, -1.57463456, -0.847734109, -0.336289384, -0.0503586083, 0, -0.785813299, -1.63679881, -2.58656369, -3.66871508, -4.91686013, -6.364606, -8.04555983, -9.99332877, -12.24152, -14.8237406, -17.7735978, -21.1246987, -24.9106504, -29.1650602, -33.9215351, -39.2136823, -45.6739711, -53.4215604, -61.1354883, -68.9776893, -77.1100984, -85.6946503, -94.8932797, -104.867921, -115.78051, -127.79298, -141.067267, -155.765305, -172.04903, -190.12948, -210.101481, -232.112234, -256.308945, -282.07165, -309.931135, -340.622764, -374.523888, -412.011862, -453.464036, -499.257765, -550.882479, -607.264994, -668.565082, -734.942514, -806.557062, -882.282554, -963.886193, -1051.84924, -1146.65294, -1248.77857, -1356.93136, -1475.53927, -1609.05609, -1759.45692, -1923.6562, -2100.8405, -2290.1964, -2489.53842, -2699.73701, -2923.15944, -3160.77669, -3412.65577, -3680.76239, -3965.30472, -4263.51291, -4573.98409, -4894.18816, -5222.43, -5561.14755, -5910.44086, -6284.97296, -6668.04044, -7054.70228, -7440.0175, -7819.04509, -8195.45283, -8553.53874, -8886.20963, -9186.37232, -9446.93359, -9615.33387, -9760.25728, -9894.09625, -10065.1557, -10271.6126, -10496.6805, -10739.2847, -10998.3503, -11272.8026, -11561.5667, -11992.8751, -12382.2872, -12718.1332, -12988.7435, -13182.4484, -13287.5781, -13292.4629, -13394.0271, -13336.0912},
        {-524.778692, -521.970094, -518.783559, -515.231561, -511.326573, -507.08107, -502.507526, -497.618414, -492.426208, -486.943382, -481.18241, -475.155765, -468.875922, -462.355355, -455.606537, -448.641942, -441.474045, -434.115318, -426.578236, -418.875272, -411.018902, -403.021597, -394.895833, -386.654083, -378.308821, -369.872521, -361.357656, -352.776702, -344.14213, -335.466416, -326.762033, -318.041455, -309.317156, -300.60161, -292.08284, -283.681852, -275.317544, -266.996692, -258.726074, -250.512466, -242.362646, -234.283389, -226.281475, -218.363678, -210.536776, -202.807547, -195.182767, -187.669213, -180.273661, -173.00289, -165.863676, -159.117326, -152.51057, -146.044665, -139.720873, -133.540451, -127.504659, -121.614757, -115.872003, -110.277656, -104.832977, -99.5392239, -94.3976562, -89.4095332, -84.5761141, -79.8986581, -75.3784245, -71.0974024, -67.135643, -63.3279057, -59.6702165, -56.1586011, -52.7890854, -49.5576953, -46.4604565, -43.4933948, -40.6525362, -37.9339065, -35.3335314, -32.8474368, -30.4716486, -28.2021926, -26.0350946, -23.9663804, -21.570635, -19.0634692, -16.6655589, -14.3866067, -12.2363149, -10.224386, -8.36052221, -6.65442609, -5.11579999, -3.75434632, -2.57976747, -1.60176584, -0.830043826, -0.274303838, 0.0557517288, 0.150420475, 0, -1.88615062, -3.68269867, -5.43310485, -7.18082988, -8.96933446, -10.8420793, -12.8425251, -15.0141325, -17.4003624, -20.0446753, -22.990532, -26.2813932, -29.9607195, -34.0719718, -38.6586107, -43.7640969, -50.0245944, -57.5682357, -65.0978434, -72.7838858, -80.7968316, -89.3071493, -98.4853074, -108.501775, -119.527019, -131.73151, -145.285715, -160.360104, -177.125144, -196.149987, -217.060235, -239.906753, -264.740403, -290.184617, -317.003976, -346.343708, -378.682907, -414.500667, -454.276082, -498.488246, -549.465007, -605.274039, -666.032725, -731.858446, -802.868581, -878.296247, -959.364156, -1046.41076, -1139.7745, -1239.79383, -1344.45722, -1459.32528, -1589.61464, -1737.59834, -1899.53961, -2074.39999, -2261.14098, -2457.34948, -2663.67405, -2882.26287, -3113.47345, -3357.98889, -3618.07741, -3894.16498, -4183.55605, -4484.98791, -4796.03437, -5115.13108, -5444.97972, -5785.86887, -6151.95173, -6527.51923, -6908.06062, -7289.06518, -7666.02217, -8049.34143, -8415.86153, -8757.3416, -9065.54076, -9332.21816, -9501.06668, -9643.95585, -9773.24459, -9934.62299, -10127.2564, -10336.4222, -10561.9803, -10803.7902, -11061.7119, -11335.605, -11881.1719, -12289.826, -12586.1096, -12794.565, -12939.7346, -13046.1607, -13138.3855, -13379.6892, -13424.9903},
        {-535.733965, -532.853423, -529.606934, -526.006052, -522.062333, -517.787331, -513.192602, -508.289702, -503.090184, -497.605606, -491.847521, -485.827486, -479.557055, -473.047783, -466.311227, -459.35894, -452.202478, -444.853398, -437.323252, -429.623598, -421.76599, -413.761983, -405.623133, -397.360995, -388.987124, -380.513075, -371.950404, -363.310666, -354.605416, -345.846208, -337.0446, -328.212145, -319.360399, -310.500917, -301.743203, -293.047777, -284.373984, -275.730199, -267.1248, -258.566163, -250.062664, -241.622681, -233.254589, -224.966766, -216.767588, -208.665432, -200.668675, -192.785692, -185.024861, -177.394558, -169.90316, -162.901525, -156.048124, -149.343909, -142.789833, -136.386847, -130.135903, -124.037955, -118.093955, -112.304854, -106.671605, -101.19516, -95.8764717, -90.7164923, -85.716174, -80.876469, -76.1983297, -71.7495897, -67.5964767, -63.6020054, -59.7627923, -56.0754541, -52.5366075, -49.142869, -45.8908553, -42.7771831, -39.7984689, -36.9513294, -34.2323813, -31.6382411, -29.1655256, -26.8108513, -24.5708348, -22.4420929, -20.0290162, -17.5325844, -15.158491, -12.9160814, -10.8147005, -8.86369359, -7.07240569, -5.45018194, -4.00636748, -2.75030741, -1.69134685, -0.838830925, -0.202104755, 0.209486541, 0.386597846, 0.319884039, 0, -3.46625272, -6.58837675, -9.42641126, -12.0403955, -14.4903685, -16.8363697, -19.1384381, -21.4566129, -23.8509333, -26.3814386, -29.1081679, -32.0911604, -35.3904553, -39.0660918, -43.1781091, -47.7865463, -53.4328446, -60.2458148, -67.1966109, -74.4484296, -82.1644677, -90.5079219, -99.641989, -109.729866, -120.934749, -133.419835, -147.34832, -162.883403, -180.188279, -200.014276, -221.720812, -245.294643, -270.722528, -296.274753, -322.79631, -351.589917, -383.157273, -418.000075, -456.620022, -499.518812, -548.994208, -603.205215, -662.302128, -726.435242, -795.754852, -869.873328, -949.613369, -1035.25975, -1127.09725, -1225.41064, -1328.29098, -1440.89797, -1567.77786, -1710.9889, -1867.50963, -2036.59929, -2217.51709, -2408.76596, -2610.53332, -2823.83165, -3048.03796, -3284.75215, -3536.67642, -3804.53888, -4085.72261, -4379.14614, -4682.89301, -4995.66526, -5319.22647, -5653.60843, -6009.16582, -6374.01941, -6744.62771, -7117.4492, -7488.94237, -7881.42251, -8259.02712, -8611.7505, -8929.58696, -9202.53079, -9366.81783, -9502.03982, -9621.23038, -9772.4493, -9954.8657, -10153.0545, -10366.9145, -10596.3445, -10841.2434, -11101.51, -11688.6784, -12086.6988, -12343.9658, -12508.8741, -12629.8181, -12755.1925, -12933.3918, -13241.9826, -13368.47},
        {-544.3831, -541.312202, -537.910317, -534.187145, -530.15239, -525.815752, -521.186932, -516.275633, -511.091557, -505.644403, -499.943875, -493.999674, -487.821501, -481.419058, -474.802047, -467.980168, -460.963124, -453.760617, -446.382347, -438.838016, -431.137326, -423.289979, -415.305676, -407.194118, -398.965008, -390.628046, -382.192935, -373.669375, -365.067069, -356.395718, -347.665023, -338.884687, -330.06441, -321.213895, -312.336016, -303.44403, -294.55128, -285.667689, -276.803179, -267.967674, -259.171096, -250.42337, -241.734417, -233.114162, -224.572526, -216.119435, -207.76481, -199.518574, -191.390651, -183.390964, -175.529437, -168.201006, -161.022235, -153.994699, -147.119976, -140.399641, -133.835273, -127.428447, -121.180741, -115.093731, -109.168994, -103.408106, -97.8126453, -92.3841874, -87.1243093, -82.0345879, -77.1165999, -72.4497055, -68.1113974, -63.9428297, -59.9405368, -56.1010532, -52.4209131, -48.896651, -45.5248011, -42.3018979, -39.2244758, -36.289069, -33.492212, -30.8304391, -28.3002846, -25.8982831, -23.6209687, -21.4648759, -19.0394315, -16.5428659, -14.1780623, -11.9541174, -9.88012794, -7.96519076, -6.21840257, -4.64886011, -3.26566014, -2.07789941, -1.09467467, -0.325082663, 0.221779852, 0.536816126, 0.610929409, 0.435022951, 0, -5.50853084, -10.3165854, -14.5077854, -18.1657526, -21.3741087, -24.2164755, -26.7764748, -29.1377282, -31.3838575, -33.5984845, -35.865231, -38.2677186, -40.889569, -43.8144042, -47.1258457, -50.9075154, -55.4259323, -60.7908462, -66.6721827, -73.1927559, -80.4753798, -88.6428683, -97.8180356, -108.123696, -119.682662, -132.61775, -147.051772, -163.107543, -180.907877, -200.924211, -222.802908, -246.562193, -272.220294, -298.704817, -326.579299, -356.698109, -389.40666, -425.050367, -463.974641, -506.524896, -553.162594, -604.081779, -659.605161, -720.055448, -785.755346, -856.752632, -933.713679, -1017.02993, -1107.09283, -1204.29381, -1308.16897, -1421.01053, -1544.53378, -1679.78415, -1826.43931, -1984.79104, -2155.1311, -2338.25588, -2533.83783, -2740.99892, -2958.33688, -3188.0434, -3432.84107, -3693.62582, -3968.19848, -4255.78056, -4555.73624, -4867.32405, -5189.27941, -5520.63363, -5862.67784, -6211.83587, -6566.35679, -6924.4897, -7284.48364, -7688.35542, -8079.64445, -8445.65791, -8773.70292, -9051.08664, -9200.43737, -9316.86033, -9415.38219, -9555.57677, -9734.83747, -9931.99638, -10145.3749, -10373.2946, -10614.0769, -10866.0433, -11374.4151, -11712.9614, -11934.2062, -12090.6735, -12234.8874, -12419.3716, -12696.6504, -12975.6461, -13192.9684},
        {-553.228513, -550.042171, -546.554326, -542.773052, -538.706427, -534.362525, -529.749423, -524.875196, -519.747921, -514.375672, -508.766527, -502.92856, -496.869848, -490.598466, -484.122491, -477.449998, -470.589063, -463.547762, -456.33417, -448.956364, -441.42242, -433.740413, -425.918419, -417.964513, -409.886773, -401.693273, -393.39209, -384.991299, -376.498976, -367.923197, -359.272039, -350.553575, -341.775884, -332.947039, -323.986704, -314.949021, -305.889194, -296.81817, -287.746893, -278.686307, -269.647359, -260.640993, -251.678155, -242.769788, -233.92684, -225.160253, -216.480974, -207.899948, -199.428119, -191.076433, -182.855835, -175.147659, -167.584433, -160.16907, -152.904488, -145.793601, -138.839325, -132.044576, -125.412271, -118.945324, -112.646651, -106.519167, -100.56579, -94.7894341, -89.1930152, -83.7794491, -78.5516515, -73.6196854, -69.0910412, -64.7476515, -60.585486, -56.6005145, -52.7887069, -49.146033, -45.6684625, -42.3519652, -39.192511, -36.1860697, -33.328611, -30.6161047, -28.0445207, -25.6098287, -23.3079985, -21.135, -18.6905673, -16.1771224, -13.8018495, -11.573577, -9.50113339, -7.59334703, -5.85904633, -4.30705971, -2.94621555, -1.78534228, -0.833268284, -0.098821982, 0.409168225, 0.681873932, 0.710466734, 0.486118225, 0, -6.51356123, -12.148598, -17.0004685, -21.164531, -24.7361436, -27.8106647, -30.4834523, -32.8498648, -35.0052603, -37.0449971, -39.0644334, -41.1589273, -43.4238371, -45.954521, -48.8463373, -52.1946441, -56.0777762, -60.5886598, -65.8611587, -71.9869832, -79.0578437, -87.1654503, -96.4015136, -106.857744, -118.625851, -131.797546, -146.464539, -162.71854, -180.651259, -200.493261, -222.146489, -245.660998, -271.086842, -297.89289, -326.419789, -357.163169, -390.34744, -426.19701, -464.936289, -506.789688, -550.854509, -598.8253, -651.146994, -708.264521, -770.622812, -838.387771, -912.353112, -993.033522, -1080.94369, -1176.5983, -1280.99408, -1393.57452, -1513.37735, -1640.4675, -1777.00708, -1924.2166, -2083.31656, -2257.05513, -2444.77798, -2644.16429, -2853.34718, -3074.92536, -3311.85842, -3565.16222, -3832.55899, -4113.28279, -4407.02857, -4713.14985, -5029.31032, -5354.12953, -5684.02389, -6020.15505, -6361.90519, -6708.65651, -7059.79119, -7470.17611, -7869.83759, -8244.28664, -8579.03427, -8859.59149, -9000.20382, -9105.15809, -9190.50765, -9322.04168, -9496.12644, -9688.40497, -9896.2797, -10117.1531, -10348.4275, -10587.5055, -10971.5155, -11245.7077, -11450.0476, -11624.5011, -11809.0338, -12043.6114, -12368.1996, -12611.0188, -12880.3443},
        {-563.878466, -560.738737, -557.307823, -553.592949, -549.60134, -545.34022, -540.816814, -536.038346, -531.012041, -525.745124, -520.244818, -514.518349, -508.572941, -502.415819, -496.054207, -489.49533, -482.746412, -475.814678, -468.707353, -461.431661, -453.994827, -446.404075, -438.666629, -430.789715, -422.780557, -414.64638, -406.394408, -398.031865, -389.565977, -381.003967, -372.353061, -363.620482, -354.813457, -345.939208, -336.883207, -327.716118, -318.510064, -309.276222, -300.025767, -290.769875, -281.519723, -272.286484, -263.081336, -253.915453, -244.800013, -235.746189, -226.765158, -217.868096, -209.066178, -200.37058, -191.792478, -183.614657, -175.570796, -167.666181, -159.9061, -152.295842, -144.840694, -137.545944, -130.416879, -123.458787, -116.676957, -110.076674, -103.663229, -97.4419072, -91.4179974, -85.5967872, -79.9835642, -74.7523513, -70.0731201, -65.6031547, -61.3368046, -57.268419, -53.3923472, -49.7029387, -46.1945427, -42.8615085, -39.6981854, -36.6989228, -33.8580699, -31.1699762, -28.6289908, -26.2294632, -23.9657426, -21.8321784, -19.3797147, -16.8337297, -14.4249314, -12.1620582, -10.053849, -8.10904241, -6.33637704, -4.7445916, -3.34242476, -2.13861518, -1.14190154, -0.361022515, 0.195283232, 0.518277028, 0.599220201, 0.429374082, 0, -4.60130548, -8.67046618, -12.2797955, -15.5016069, -18.4082138, -21.0719297, -23.5650678, -25.9599417, -28.3288648, -30.7441505, -33.2781122, -36.0030634, -38.9913175, -42.3151879, -46.046988, -50.2590312, -55.1337782, -60.7592777, -66.9587007, -73.8279635, -81.4629826, -89.9596744, -99.4139552, -109.921741, -121.578949, -134.481496, -148.725297, -164.406269, -181.620328, -200.676127, -221.378844, -243.760572, -267.853408, -292.691495, -318.805903, -346.993827, -377.586746, -410.916143, -447.313498, -487.110293, -530.676453, -578.293315, -630.284839, -686.974985, -748.687712, -814.802518, -886.823941, -965.312056, -1050.82694, -1143.92866, -1246.11497, -1355.86223, -1471.47754, -1592.71498, -1722.28296, -1861.56346, -2011.93845, -2176.38678, -2354.33063, -2543.45019, -2741.70324, -2952.33436, -3180.10948, -3426.29978, -3686.25476, -3958.04208, -4237.88462, -4523.37175, -4818.85703, -5124.86779, -5448.99212, -5783.61037, -6126.8058, -6476.66167, -6831.26124, -7234.1183, -7626.52794, -7995.21576, -8326.90738, -8608.32843, -8763.36335, -8883.47304, -8982.31502, -9111.40424, -9269.41302, -9439.23976, -9620.34547, -9812.19118, -10014.2379, -10225.9466, -10441.9672, -10673.6046, -10915.6927, -11163.0653, -11410.5562, -11652.9992, -11885.2281, -12172.3081, -12417.2051},
        {-575.353228, -572.533312, -569.404243, -565.973557, -562.24879, -558.237477, -553.947155, -549.385359, -544.559624, -539.477488, -534.146484, -528.574149, -522.768019, -516.735629, -510.484515, -504.022213, -497.356258, -490.494187, -483.443534, -476.211837, -468.806629, -461.235448, -453.505828, -445.625306, -437.601417, -429.441697, -421.153682, -412.744907, -404.222908, -395.595221, -386.869381, -378.052925, -369.153387, -360.178305, -351.044568, -341.806943, -332.521281, -323.198062, -313.847761, -304.480857, -295.107827, -285.739147, -276.385296, -267.05675, -257.763986, -248.517483, -239.327717, -230.205166, -221.160307, -212.203616, -203.345573, -194.725025, -186.221295, -177.842077, -169.595067, -161.487958, -153.528445, -145.724223, -138.082986, -130.612428, -123.320243, -116.214127, -109.301773, -102.590877, -96.0891316, -89.8042325, -83.7438738, -78.1364333, -73.2049903, -68.5020964, -64.02114, -59.7555095, -55.6985933, -51.8437798, -48.1844576, -44.714015, -41.4258405, -38.3133225, -35.3698494, -32.5888096, -29.9635917, -27.487584, -25.154175, -22.9567531, -20.4366183, -17.8227123, -15.3494106, -13.0247728, -10.8568585, -8.85372715, -7.02343842, -5.37405186, -3.91362704, -2.65022353, -1.59190092, -0.746718765, -0.122736648, 0.27198586, 0.429389186, 0.341413757, 0, -2.02325532, -3.97171433, -5.88678433, -7.80987263, -9.78238651, -11.8457333, -14.0413202, -16.4105546, -18.9948438, -21.835595, -24.9742156, -28.4521128, -32.310694, -36.5913665, -41.3355374, -46.5846143, -52.7847769, -60.0352575, -67.4619058, -75.1928661, -83.3562828, -92.0803002, -101.493063, -111.722714, -122.8974, -135.145263, -148.594448, -163.3731, -179.609363, -197.872201, -217.687304, -239.050572, -261.957902, -285.016035, -308.915449, -334.717063, -362.833521, -393.67747, -427.661555, -465.198422, -507.712659, -554.296986, -605.166058, -660.534534, -720.617068, -784.113277, -853.131619, -928.265509, -1010.10836, -1099.2536, -1197.45017, -1302.72364, -1413.37848, -1529.06903, -1652.20558, -1783.96106, -1925.50839, -2078.86748, -2244.17177, -2420.63074, -2607.65142, -2807.28682, -3023.98794, -3258.67134, -3506.2647, -3764.44466, -4027.6269, -4292.64264, -4568.28009, -4856.56389, -5174.52455, -5506.87229, -5850.43744, -6202.05033, -6558.54129, -6951.4111, -7334.15202, -7694.92677, -8021.89804, -8303.22857, -8473.6737, -8610.40841, -8724.48428, -8852.39924, -8995.13236, -9143.06004, -9297.69149, -9460.53588, -9633.10241, -9816.90027, -9919.92053, -10122.514, -10387.6255, -10678.1996, -10957.1813, -11187.5152, -11332.1461, -11648.6279, -11846.4087},
        {-587.760262, -585.841603, -583.537615, -580.858486, -577.814404, -574.415557, -570.672135, -566.594325, -562.192315, -557.476293, -552.456448, -547.142969, -541.546043, -535.675858, -529.542603, -523.156466, -516.527636, -509.6663, -502.582647, -495.286865, -487.789142, -480.099667, -472.228628, -464.186212, -455.982609, -447.628007, -439.132594, -430.506557, -421.760086, -412.903368, -403.946593, -394.899947, -385.773619, -376.577799, -367.378002, -358.15559, -348.891257, -339.593395, -330.270397, -320.930656, -311.582565, -302.234515, -292.894901, -283.572114, -274.274547, -265.010593, -255.788645, -246.617095, -237.504336, -228.45876, -219.488761, -210.574429, -201.753072, -193.033697, -184.425309, -175.936915, -167.577522, -159.356135, -151.28176, -143.363405, -135.610075, -128.030776, -120.634514, -113.430297, -106.427129, -99.6340184, -93.0599699, -86.9278679, -81.4554615, -76.2106499, -71.1885745, -66.3843767, -61.7931979, -57.4101796, -53.230463, -49.2491896, -45.4615008, -41.8625381, -38.4474427, -35.2113562, -32.1494199, -29.2567752, -26.5285636, -23.9599264, -21.2725297, -18.6040049, -16.09527, -13.7503412, -11.5732349, -9.56796706, -7.73855403, -6.08901196, -4.62335706, -3.34560554, -2.25977358, -1.36987739, -0.679933166, -0.193957117, 0.0840345604, 0.150025666, 0, 1.43531361, 2.42390398, 2.96015965, 3.03846918, 2.65322112, 1.79880404, 0.469606475, -1.33998301, -3.63557588, -6.42278355, -9.7072175, -13.4944891, -17.79021, -22.5999914, -27.9294448, -33.7841818, -41.6093556, -51.6162267, -60.5542996, -68.737659, -76.4803897, -84.0965765, -91.900304, -100.205657, -109.32672, -119.577578, -131.272316, -144.725018, -160.249768, -179.70623, -201.296198, -224.870085, -250.278303, -276.791044, -304.548828, -333.846905, -364.709751, -397.161845, -431.227666, -466.93169, -500.587872, -537.060504, -577.100036, -621.456918, -670.881602, -727.933733, -791.102266, -860.685352, -936.981143, -1020.28779, -1111.63014, -1209.69146, -1313.31313, -1422.20079, -1537.82465, -1660.86266, -1791.99278, -1930.30134, -2078.41964, -2240.71533, -2424.18284, -2621.42182, -2825.31598, -3033.43567, -3251.2925, -3480.75291, -3725.76682, -3988.74084, -4264.44216, -4551.95927, -4850.41732, -5158.86248, -5476.37055, -5802.0173, -6134.87853, -6522.85465, -6903.99067, -7265.15621, -7593.22088, -7875.05433, -8040.16122, -8171.01943, -8279.0836, -8407.35699, -8554.80622, -8707.25258, -8864.31994, -9025.63217, -9190.81311, -9359.48663, -9364.64839, -9489.17593, -9699.38172, -9961.57824, -10242.078, -10507.1934, -10723.2369, -10995.2702, -11184.0619},
        {-612.03065, -611.234516, -609.918161, -608.097111, -605.786893, -603.003034, -599.76106, -596.0765, -591.964879, -587.441725, -582.522565, -577.222925, -571.558331, -565.544313, -559.196395, -552.530105, -545.56097, -538.304516, -530.776271, -522.991762, -514.966515, -506.716057, -498.255916, -489.601617, -480.768689, -471.772657, -462.629049, -453.353392, -443.961212, -434.468037, -424.889392, -415.240807, -405.537806, -395.795917, -386.392789, -377.155271, -367.90586, -358.648331, -349.386459, -340.12402, -330.86479, -321.612542, -312.371053, -303.144099, -293.935453, -284.748893, -275.588192, -266.457127, -257.359473, -248.299004, -239.279497, -229.9475, -220.67176, -211.463794, -202.335124, -193.297267, -184.361743, -175.540072, -166.843773, -158.284365, -149.873367, -141.622299, -133.542681, -125.64603, -117.943867, -110.447711, -103.169082, -96.3439799, -90.2030176, -84.2947365, -78.6161035, -73.1640858, -67.9356503, -62.9277641, -58.1373942, -53.5615076, -49.1970714, -45.0410526, -41.0904182, -37.3421353, -33.7931709, -30.440492, -27.2810658, -24.3118591, -21.3779128, -18.5553527, -15.9221305, -13.4801435, -11.231289, -9.17746432, -7.32056685, -5.6624939, -4.20514279, -2.95041086, -1.90019544, -1.05639385, -0.420903425, 0.00437850088, 0.217554601, 0.216727544, 0, -0.210344857, -0.608496259, -1.20872805, -2.02531406, -3.07252815, -4.36464414, -5.91593589, -7.74067724, -9.85314202, -12.2676041, -14.9983373, -18.0596154, -21.4657124, -25.230902, -29.369458, -33.8956545, -40.0076279, -47.8887755, -54.8755861, -61.2360183, -67.2380309, -73.1495827, -79.2386324, -85.7731387, -93.0210604, -101.250356, -110.728985, -121.724905, -134.506076, -151.008554, -169.220563, -188.909634, -209.843296, -229.177378, -247.98526, -268.036774, -289.88296, -314.074857, -341.163505, -371.699943, -410.617995, -452.752025, -497.795571, -545.442171, -595.385364, -641.132916, -690.11058, -743.558336, -802.716166, -868.824051, -945.945708, -1029.04615, -1118.53902, -1213.89918, -1312.68471, -1413.31429, -1514.20658, -1602.18178, -1689.89307, -1802.64648, -1972.62519, -2165.66157, -2350.5412, -2515.85885, -2683.77486, -2861.04241, -3060.35457, -3290.00447, -3536.50577, -3798.69181, -4081.78229, -4377.24174, -4681.69288, -4991.75841, -5304.06106, -5634.72335, -5955.99324, -6259.61848, -6537.34685, -6780.92609, -6944.78526, -7082.86998, -7202.92176, -7326.86807, -7456.83513, -7589.03835, -7725.8995, -7869.84034, -8023.28265, -8188.64818, -9161.19735, -9586.72028, -9635.01241, -9475.8692, -9279.08609, -9214.45852, -9451.78195, -9655.14341, -9822.23619},
        {-620.114449, -619.594743, -618.578088, -617.078484, -615.109934, -612.686438, -609.821998, -606.530616, -602.826292, -598.723029, -594.234828, -589.375691, -584.159618, -578.600611, -572.712672, -566.509802, -560.006003, -553.215276, -546.151622, -538.829044, -531.261541, -523.463117, -515.447772, -507.229508, -498.822326, -490.240227, -481.497214, -472.607288, -463.584449, -454.4427, -445.196042, -435.858477, -426.444005, -416.966629, -407.714846, -398.559638, -389.368682, -380.147073, -370.899904, -361.632269, -352.34926, -343.055972, -333.757498, -324.458931, -315.165366, -305.881895, -296.613611, -287.365609, -278.142982, -268.950824, -259.794227, -250.473303, -241.202572, -231.99157, -222.849835, -213.786904, -204.812313, -195.9356, -187.166302, -178.513957, -169.9881, -161.598269, -153.354002, -145.264836, -137.340306, -129.589952, -122.023309, -114.790395, -108.037852, -101.485491, -95.13374, -88.9830308, -83.0337934, -77.2864582, -71.7414555, -66.3992157, -61.2601691, -56.3247463, -51.5933774, -47.0664929, -42.7445232, -38.6278986, -34.7170495, -31.0124063, -27.4162276, -23.9800946, -20.7567684, -17.7498652, -14.9630011, -12.3997926, -10.0638557, -7.95880668, -6.08826184, -4.45583737, -3.06514951, -1.91981448, -1.02344852, -0.379667864, 0.00791126347, 0.135672629, 0, -0.983845609, -2.09848818, -3.36097924, -4.78837029, -6.39771285, -8.20605843, -10.2304586, -12.4879647, -14.9956285, -17.7705013, -20.8296347, -24.1900803, -27.8688895, -31.8831138, -36.2498048, -40.9860139, -46.6799665, -53.4303104, -59.9621563, -66.4149502, -72.9281379, -79.6411654, -86.6934785, -94.2245232, -102.373745, -111.280591, -121.084505, -131.924935, -143.941326, -157.871535, -173.03718, -189.398182, -206.914464, -224.009988, -241.412656, -260.259961, -280.972613, -303.971324, -329.676805, -358.509766, -394.481012, -433.328523, -474.770599, -518.52554, -564.311648, -606.246998, -651.050172, -699.839528, -753.733423, -813.850214, -883.661521, -959.056227, -1039.71273, -1125.19487, -1214.83257, -1308.06078, -1404.31446, -1498.23319, -1595.13715, -1705.57781, -1841.32948, -1991.82694, -2148.60434, -2308.65982, -2478.25003, -2659.38178, -2855.37707, -3068.5837, -3296.52724, -3539.46099, -3804.37174, -4083.74336, -4375.49832, -4677.5591, -4987.84815, -5342.19374, -5691.13609, -6023.12125, -6326.59524, -6590.00408, -6754.43105, -6887.26009, -6997.23555, -7110.14272, -7227.98058, -7345.64772, -7465.49813, -7589.8858, -7721.16474, -7861.68894, -8565.12357, -8889.43749, -8954.40308, -8879.79271, -8785.37874, -8790.93355, -9016.2295, -9122.559, -9226.2932},
        {-650.766465, -650.196436, -649.177873, -647.721985, -645.839982, -643.543073, -640.842468, -637.749376, -634.275007, -630.430571, -626.227276, -621.676332, -616.788949, -611.576337, -606.049704, -600.22026, -594.099215, -587.697779, -581.02716, -574.098568, -566.923213, -559.512304, -551.877051, -544.028663, -535.97835, -527.737321, -519.316786, -510.727953, -501.982034, -493.090237, -484.063771, -474.913847, -465.651673, -456.288459, -446.909059, -437.486312, -427.99217, -418.435453, -408.82498, -399.169572, -389.478048, -379.759228, -370.02193, -360.274975, -350.527183, -340.787372, -331.064363, -321.366975, -311.704028, -302.084341, -292.516734, -283.212512, -273.97362, -264.804487, -255.709545, -246.693222, -237.75995, -228.914157, -220.160275, -211.502733, -202.945962, -194.494391, -186.152451, -177.924572, -169.815184, -161.828717, -153.969601, -146.222822, -138.573834, -131.067168, -123.708515, -116.503565, -109.458009, -102.577537, -95.8678396, -89.334608, -82.9835325, -76.8203035, -70.8506117, -65.0801476, -59.5146018, -54.1596649, -49.0210274, -44.1043799, -39.5996272, -35.416478, -31.4624271, -27.7371867, -24.2404695, -20.9719878, -17.9314542, -15.118581, -12.5330808, -10.1746659, -8.04304897, -6.13794232, -4.45905845, -3.00610983, -1.7788089, -0.776868142, 0, 2.00442108, 3.52490223, 4.57687503, 5.17577105, 5.33702189, 5.07605912, 4.40831432, 3.34921909, 1.914205, 0.118703625, -2.02185344, -4.49203462, -7.27640833, -10.359543, -13.726007, -17.3603688, -22.5845874, -29.574287, -35.2889858, -39.9998361, -43.9779899, -47.4945994, -50.8208165, -54.2277936, -57.9866826, -62.3686357, -67.644805, -74.0863426, -81.9644007, -93.5031339, -106.304591, -120.054023, -134.436681, -146.792147, -157.978506, -169.479358, -181.683655, -194.980348, -209.758392, -226.406737, -247.427003, -270.45249, -295.458802, -322.421545, -351.316322, -382.122158, -414.810383, -449.355746, -485.732997, -523.916886, -563.098023, -604.993689, -651.189511, -702.3039, -756.980535, -814.749708, -875.141713, -933.580246, -994.635514, -1067.35765, -1162.0807, -1269.26381, -1381.81918, -1496.88168, -1618.74124, -1748.40352, -1887.26622, -2036.43666, -2195.58452, -2365.19272, -2545.67181, -2737.58814, -2941.44962, -3157.76417, -3387.03972, -3673.53519, -3963.06975, -4245.21357, -4509.53684, -4745.60972, -4904.00411, -5039.28732, -5157.74307, -5279.12609, -5404.38784, -5528.86683, -5653.79516, -5780.40492, -5909.9282, -6043.59709, -6398.91916, -6615.76517, -6732.31535, -6786.7499, -6817.24902, -6861.99293, -6959.16182, -7127.65459, -7279.75167}
    };
    
    const double a5tab[98][200] = {{1623.72041, 1600.48232, 1577.67306, 1555.27899, 1533.28645, 1511.68181, 1490.45141, 1469.58163, 1449.05881, 1428.86932, 1408.9995, 1389.43571, 1370.16432, 1351.17167, 1332.44412, 1313.96804, 1295.72977, 1277.71568, 1259.91211, 1242.30543, 1224.88199, 1207.62815, 1190.53026, 1173.57469, 1156.74778, 1140.0359, 1123.4254, 1106.90264, 1090.45397, 1074.06575, 1057.72433, 1041.41608, 1025.12735, 1008.84449, 991.903965, 974.630743, 957.357616, 940.092033, 922.841441, 905.613286, 888.415016, 871.254077, 854.137918, 837.073984, 820.069724, 803.132584, 786.270012, 769.489454, 752.798358, 736.20417, 719.714338, 703.843765, 688.081441, 672.423814, 656.867328, 641.408431, 626.04357, 610.769191, 595.58174, 580.477663, 565.453408, 550.505421, 535.630148, 520.824035, 506.08353, 491.405078, 476.785126, 462.183943, 447.562664, 432.992352, 418.471798, 403.999794, 389.575132, 375.196604, 360.863001, 346.573115, 332.325739, 318.119663, 303.95368, 289.82658, 275.737157, 261.684202, 247.666507, 233.682862, 219.755572, 205.871177, 192.015938, 178.187882, 164.385039, 150.605438, 136.847107, 123.108075, 109.386371, 95.6800232, 81.9870613, 68.3055139, 54.6334096, 40.9687772, 27.3096456, 13.6540436, 0, -14.0013435, -27.7618601, -41.3242107, -54.7310563, -68.0250576, -81.2488757, -94.4451714, -107.656606, -120.925839, -134.295533, -147.808348, -161.506945, -175.433985, -189.632129, -204.144037, -219.012371, -234.376733, -250.293631, -266.586456, -283.318642, -300.553623, -318.354833, -336.785705, -355.909674, -375.790174, -396.490637, -418.074499, -440.605193, -464.146153, -488.530438, -514.136327, -541.096367, -569.543103, -599.605116, -631.416936, -665.114148, -700.830488, -738.699692, -778.855496, -821.431636, -865.298883, -912.238317, -962.630777, -1016.8571, -1075.29812, -1139.24635, -1207.94303, -1281.5411, -1360.19347, -1444.05307, -1531.66097, -1626.75197, -1729.48464, -1840.96769, -1964.24974, -2101.50842, -2254.92137, -2433.40628, -2630.86891, -2833.8622, -3023.03961, -3223.59323, -3456.90788, -3733.06209, -4042.97641, -4386.36523, -4755.87335, -5149.38231, -5590.69535, -6088.95292, -6677.39197, -7337.34931, -8069.62429, -8875.01628, -9754.32463, -10996.7528, -12242.595, -13420.5496, -14459.315, -15287.5894, -15668.2024, -15806.3003, -15701.6683, -15209.5823, -14387.0289, -13351.2064, -12156.0908, -10855.658, -9503.88366, -8154.74375, -6958.37272, -5888.07957, -4934.26905, -4087.34594, -3337.715, -2675.78099, -2091.94866, -1598.05655, -1133.82715},
        {1520.58131, 1497.75139, 1475.39511, 1453.49854, 1432.04776, 1411.02885, 1390.42787, 1370.23092, 1350.42405, 1330.99335, 1311.9249, 1293.20476, 1274.81902, 1256.75375, 1238.99503, 1221.52892, 1204.34152, 1187.41888, 1170.7471, 1154.31224, 1138.10038, 1122.0976, 1106.28996, 1090.66356, 1075.20446, 1059.89873, 1044.73246, 1029.69172, 1014.76258, 999.931131, 985.183433, 970.505569, 955.883612, 941.303639, 926.313912, 911.128623, 895.967223, 880.829999, 865.717232, 850.629206, 835.566205, 820.528513, 805.516413, 790.530188, 775.570123, 760.6365, 745.729603, 730.849717, 715.997124, 701.172108, 686.374953, 671.533924, 656.722884, 641.943678, 627.19815, 612.488145, 597.815509, 583.182085, 568.589718, 554.040254, 539.535537, 525.077412, 510.667724, 496.308317, 482.001036, 467.747725, 453.550231, 439.467765, 425.558164, 411.704954, 397.906261, 384.160211, 370.46493, 356.818543, 343.219177, 329.664957, 316.154009, 302.684459, 289.254433, 275.862057, 262.505455, 249.182755, 235.892082, 222.631561, 209.379802, 196.1451, 182.935983, 169.751212, 156.589544, 143.44974, 130.330558, 117.230758, 104.149099, 91.0843403, 78.0352406, 65.0005593, 51.9790555, 38.9694886, 25.9706175, 12.9812016, 0, -13.463472, -26.6507305, -39.6055667, -52.3717718, -64.9931372, -77.513454, -89.9765133, -102.426107, -114.906025, -127.460059, -140.132002, -152.965642, -166.004773, -179.293185, -192.874669, -206.793017, -221.115787, -235.890165, -251.106183, -266.812727, -283.058681, -299.892927, -317.364352, -335.521838, -354.414269, -374.090531, -394.599506, -415.99008, -438.311135, -461.297415, -485.42713, -510.843408, -537.689377, -566.428544, -597.043847, -629.432785, -663.642371, -699.719616, -737.711534, -777.665135, -817.52988, -860.088719, -905.799056, -955.118294, -1008.50384, -1068.61954, -1133.16474, -1202.04523, -1275.16679, -1352.43521, -1431.84103, -1517.54615, -1610.88824, -1713.06715, -1825.00135, -1947.73564, -2082.31484, -2233.32218, -2397.45983, -2567.56982, -2733.31813, -2908.00089, -3103.52727, -3325.11727, -3566.65656, -3827.23344, -4101.04224, -4385.90246, -4697.57821, -5041.68311, -5469.1114, -5933.22983, -6423.97799, -6931.29543, -7445.12173, -7884.64926, -8328.25158, -8783.55505, -9258.18602, -9759.77087, -10449.0348, -11078.4395, -11589.9974, -11639.621, -11282.5709, -10693.3159, -9921.15598, -9015.39125, -8025.32176, -7000.24756, -5996.25218, -5119.40897, -4352.06951, -3676.58538, -3075.30815, -2530.58941, -2024.78073, -1620.33653, -1273.14176},
        {1424.81636, 1406.29628, 1387.8899, 1369.5979, 1351.42096, 1333.35975, 1315.41495, 1297.58724, 1279.8773, 1262.28581, 1244.81344, 1227.46087, 1210.22878, 1193.11785, 1176.12876, 1159.26218, 1142.51879, 1125.89927, 1109.4043, 1093.03455, 1076.79071, 1060.67345, 1044.68344, 1028.82138, 1013.08793, 997.483774, 982.009589, 966.666053, 951.453844, 936.37364, 921.426119, 906.611959, 891.931838, 877.386434, 863.296241, 849.495302, 835.813817, 822.242083, 808.770401, 795.389069, 782.088388, 768.858656, 755.690172, 742.573236, 729.498147, 716.455205, 703.434708, 690.426956, 677.422248, 664.410884, 651.383163, 637.73191, 624.057852, 610.364237, 596.654319, 582.931348, 569.198576, 555.459254, 541.716632, 527.973964, 514.234499, 500.501489, 486.778186, 473.067841, 459.373705, 445.699029, 432.047065, 418.436419, 404.885327, 391.365374, 377.878816, 364.427908, 351.014906, 337.642066, 324.311645, 311.025898, 297.787082, 284.597451, 271.459263, 258.374772, 245.346236, 232.375909, 219.466049, 206.61891, 194.047405, 181.64403, 169.298752, 157.00699, 144.764163, 132.565691, 120.406994, 108.283493, 96.1906052, 84.1237521, 72.0783531, 60.0498279, 48.0335962, 36.0250778, 24.0196925, 12.01286, 0, -12.6860857, -25.1219374, -37.3491237, -49.4092134, -61.3437751, -73.1943775, -85.0025894, -96.8099795, -108.658116, -120.588569, -132.642905, -144.862695, -157.289506, -169.964908, -182.930468, -196.227757, -209.92068, -224.053998, -238.618752, -253.661298, -269.22799, -285.365185, -302.119238, -319.536505, -337.66334, -356.5461, -376.231141, -396.764817, -418.193484, -440.332907, -463.544582, -487.944043, -513.646823, -541.097046, -570.245951, -600.957149, -633.247595, -667.134245, -702.634055, -739.763979, -776.37659, -815.311951, -857.010486, -901.912615, -950.458761, -1002.59164, -1059.3738, -1121.3701, -1189.14538, -1263.2645, -1345.9418, -1434.07657, -1526.00407, -1621.06393, -1720.64637, -1825.22093, -1935.25715, -2048.89892, -2169.47, -2302.83117, -2458.42294, -2625.46502, -2794.1435, -2960.60903, -3131.11885, -3307.29112, -3492.56589, -3689.03365, -3892.1046, -4100.96771, -4334.72676, -4569.59205, -4797.85893, -5011.82276, -5203.77889, -5163.79815, -5136.95656, -5166.1056, -5294.09676, -5563.78152, -6297.39736, -7072.15245, -7811.16144, -8135.29145, -8087.29678, -7835.86815, -7417.46302, -6868.53886, -6225.55311, -5524.96325, -4726.33674, -4039.0608, -3441.99808, -2914.01126, -2433.96301, -1980.716, -1533.1329, -1157.56778, -848.084168},
        {1327.80903, 1311.54568, 1295.43188, 1279.46334, 1263.63576, 1247.94485, 1232.3863, 1216.95583, 1201.64915, 1186.46194, 1171.38992, 1156.42879, 1141.57425, 1126.82202, 1112.16778, 1097.60726, 1083.13614, 1068.75014, 1054.44496, 1040.2163, 1026.05987, 1011.97137, 997.94651, 983.980985, 970.070503, 956.210769, 942.397487, 928.62636, 914.893093, 901.19339, 887.522954, 873.87749, 860.252702, 846.644293, 832.830391, 818.920065, 805.024998, 791.147957, 777.291705, 763.459008, 749.652631, 735.875338, 722.129894, 708.419065, 694.745615, 681.112308, 667.52191, 653.977186, 640.480901, 627.035819, 613.644705, 600.55455, 587.518599, 574.534323, 561.59919, 548.710674, 535.866243, 523.063368, 510.29952, 497.57217, 484.878788, 472.216844, 459.583809, 446.977154, 434.394349, 421.832864, 409.290171, 396.683634, 383.932544, 371.19958, 358.487406, 345.798686, 333.136082, 320.502259, 307.899879, 295.331606, 282.800104, 270.308035, 257.858063, 245.452853, 233.095066, 220.787366, 208.532417, 196.332883, 184.406985, 172.645073, 160.934907, 149.272154, 137.652484, 126.071563, 114.52506, 103.008643, 91.5179805, 80.04874, 68.5965899, 57.1571982, 45.7262331, 34.2993626, 22.8722549, 11.440578, 0, -12.0586586, -23.8870339, -35.52436, -47.0098712, -58.3828018, -69.6823859, -80.9478579, -92.2184519, -103.533402, -114.931943, -126.453308, -138.136733, -150.021451, -162.146696, -174.551703, -187.275705, -200.380347, -213.908063, -227.847401, -242.242398, -257.137088, -272.57551, -288.601698, -305.259688, -322.593518, -340.647222, -359.464837, -379.0904, -399.567945, -420.82252, -443.060779, -466.362457, -490.807288, -516.194361, -542.743732, -570.750294, -600.377972, -631.790692, -665.152382, -700.626967, -738.75399, -779.207443, -822.077761, -867.455381, -915.430737, -963.023034, -1014.16175, -1069.70512, -1130.5114, -1197.43884, -1273.52083, -1354.78192, -1441.05046, -1531.78129, -1625.6666, -1721.741, -1819.03909, -1911.02471, -2003.56932, -2108.6216, -2241.20152, -2385.18917, -2530.20147, -2670.40586, -2807.68038, -2940.36377, -3069.53736, -3194.25093, -3303.49816, -3391.96104, -3465.39558, -3505.70604, -3503.74109, -3450.3494, -3336.37963, -2761.48707, -2205.51211, -1757.10179, -1504.90311, -1537.5631, -2408.05243, -3431.14535, -4496.49303, -5110.14413, -5313.59246, -5308.16624, -5127.36756, -4804.69853, -4373.66126, -3867.75785, -3372.74731, -2861.43056, -2349.62552, -1853.15014, -1387.82236, -969.460121, -613.881354, -280.195793, 5.0307902},
        {1238.0202, 1223.37624, 1208.92003, 1194.64378, 1180.53968, 1166.59996, 1152.81683, 1139.18249, 1125.68915, 1112.32903, 1099.09434, 1085.97728, 1072.97007, 1060.06492, 1047.25404, 1034.52963, 1021.88391, 1009.30909, 996.797376, 984.340983, 971.932121, 959.562998, 947.225826, 934.912814, 922.616174, 910.328115, 898.040849, 885.746584, 873.437533, 861.105904, 848.743909, 836.343758, 823.897661, 811.397829, 798.342347, 784.980882, 771.569039, 758.115062, 744.627197, 731.113691, 717.582788, 704.042735, 690.501777, 676.968159, 663.450128, 649.955928, 636.493806, 623.072008, 609.698778, 596.382362, 583.131007, 570.447961, 557.835736, 545.291846, 532.813807, 520.399134, 508.045341, 495.749943, 483.510456, 471.324394, 459.189272, 447.102606, 435.061909, 423.064698, 411.108486, 399.19079, 387.309123, 375.417774, 363.47207, 351.558677, 339.677912, 327.830092, 316.015535, 304.234557, 292.487475, 280.774608, 269.096271, 257.452782, 245.844458, 234.271617, 222.734575, 211.233649, 199.769158, 188.341417, 177.043805, 165.828152, 154.645168, 143.49215, 132.366395, 121.2652, 110.185862, 99.1256788, 88.081947, 77.0519639, 66.0330264, 55.0224318, 44.0174771, 33.0154595, 22.013676, 11.0094238, 0, -11.3204132, -22.4558659, -33.4411424, -44.3110269, -55.1003038, -65.8437574, -76.5761719, -87.3323317, -98.147021, -109.055024, -120.091125, -131.290109, -142.686759, -154.31586, -166.212197, -178.410553, -191.053864, -204.192366, -217.616213, -231.383365, -245.551782, -260.179424, -275.324249, -291.044219, -307.397292, -324.441428, -342.234586, -360.834727, -380.29981, -400.747887, -422.154792, -444.560456, -468.00481, -491.842192, -516.455332, -542.409783, -569.951155, -599.325059, -630.777105, -664.552903, -702.632486, -742.999173, -785.559232, -830.218928, -876.884528, -923.614366, -972.624624, -1024.28355, -1078.9594, -1137.02042, -1198.66518, -1264.639, -1336.99604, -1416.39242, -1500.62994, -1588.7919, -1679.96164, -1772.41115, -1866.21947, -1962.35073, -2060.89473, -2161.55972, -2270.59072, -2387.23418, -2498.87777, -2598.35255, -2679.29227, -2734.73609, -2754.77993, -2731.18459, -2690.47833, -2584.30563, -2392.39237, -2094.46445, -1670.24776, -375.729835, 904.690477, 2010.35269, 2780.59631, 3054.76085, 1903.34098, 447.084297, -1145.16476, -2146.38113, -2608.66686, -2816.69924, -2810.95154, -2631.89704, -2320.00901, -1915.76072, -1621.94059, -1238.81136, -803.584122, -353.470009, 74.3198664, 442.574389, 714.082443, 988.908967, 1220.97527},
        {1161.2229, 1150.24808, 1139.06657, 1127.68875, 1116.12505, 1104.38585, 1092.48157, 1080.42261, 1068.21936, 1055.88224, 1043.42164, 1030.84796, 1018.17162, 1005.40301, 992.552542, 979.630609, 966.647619, 953.613973, 940.540076, 927.436329, 914.313136, 901.180899, 888.050021, 874.930906, 861.833956, 848.769575, 835.748164, 822.780126, 809.875866, 797.045785, 784.300287, 771.649773, 759.104649, 746.675315, 735.083385, 723.968698, 712.962544, 702.052245, 691.225125, 680.468506, 669.76971, 659.116062, 648.494883, 637.893495, 627.299223, 616.699389, 606.081314, 595.432323, 584.739738, 573.990882, 563.173077, 551.320152, 539.393593, 527.401395, 515.351548, 503.252046, 491.110881, 478.936046, 466.735534, 454.517337, 442.289448, 430.059859, 417.836564, 405.627554, 393.440823, 381.284363, 369.166167, 357.21745, 345.566471, 333.967083, 322.41929, 310.923097, 299.478508, 288.085529, 276.744163, 265.454416, 254.216291, 243.029794, 231.894928, 220.811699, 209.780111, 198.800169, 187.871877, 176.99524, 166.290933, 155.696088, 145.146384, 134.637911, 124.166756, 113.729009, 103.320758, 92.9380908, 82.5770969, 72.2338647, 61.9044825, 51.5850389, 41.2716226, 30.960322, 20.6472256, 10.3284221, 0, -10.3539294, -20.6046784, -30.7805184, -40.9097208, -51.0205568, -61.141298, -71.3002156, -81.525581, -91.8456657, -102.288741, -112.883078, -123.656948, -134.638624, -145.856375, -157.338473, -169.113191, -181.474271, -194.487911, -207.581907, -220.841417, -234.351601, -248.197615, -262.464619, -277.237771, -292.60223, -308.643154, -325.445702, -343.095032, -361.676302, -381.532798, -402.396904, -424.276342, -447.178831, -470.770608, -495.230134, -520.826935, -547.671177, -575.873024, -605.542643, -636.7902, -669.435454, -703.967361, -740.552905, -779.359069, -820.552838, -866.271001, -914.218286, -964.069225, -1015.49835, -1068.1802, -1120.1813, -1174.74952, -1232.21553, -1293.27966, -1359.39693, -1431.68353, -1511.25564, -1606.35396, -1709.35096, -1804.8469, -1877.66577, -1939.40873, -2002.838, -2070.15464, -2125.66434, -2161.88707, -2175.06764, -2158.69172, -2092.58713, -1964.30735, -1831.6772, -1601.16784, -1236.00803, -699.426553, 45.3478072, 2153.63521, 4264.02072, 6133.63831, 7519.62198, 8179.10571, 6761.86896, 4870.63661, 2737.12315, 1336.31807, 605.722842, 173.14107, -8.44085798, 13.9634479, 193.340378, 482.676321, 783.247374, 1082.46196, 1376.44987, 1661.34092, 1933.26489, 2188.3516, 2422.73084, 2630.31017, 2790.29669},
        {1085.38834, 1077.31645, 1068.81473, 1059.90288, 1050.60058, 1040.92752, 1030.90339, 1020.54789, 1009.8807, 998.921504, 987.690003, 976.205883, 964.488833, 952.558543, 940.434702, 928.137001, 915.685129, 903.098775, 890.397629, 877.601382, 864.729722, 851.802339, 838.838923, 825.859164, 812.882751, 799.929374, 787.018723, 774.170487, 761.404356, 748.74002, 736.197168, 723.79549, 711.554676, 699.494414, 688.892434, 679.112945, 669.505035, 660.047568, 650.719409, 641.49942, 632.366467, 623.299413, 614.277122, 605.278457, 596.282284, 587.267465, 578.212864, 569.097347, 559.899775, 550.599014, 541.173928, 530.104017, 518.900011, 507.573278, 496.135185, 484.597099, 472.970386, 461.266415, 449.496552, 437.672164, 425.80462, 413.905285, 401.985527, 390.056713, 378.13021, 366.217386, 354.329607, 342.632209, 331.286829, 319.987288, 308.734973, 297.531268, 286.37756, 275.275236, 264.225681, 253.230281, 242.290423, 231.407492, 220.582875, 209.817957, 199.114125, 188.472764, 177.895261, 167.383002, 157.136883, 147.054338, 137.030405, 127.059994, 117.138018, 107.259387, 97.4190146, 87.6118109, 77.8326879, 68.0765573, 58.3383305, 48.6129191, 38.8952349, 29.1801892, 19.4626938, 9.73766022, 0, -9.75640131, -19.4316838, -29.0521658, -38.644166, -48.2340026, -57.8479942, -67.5124593, -77.2537162, -87.0980834, -97.0718795, -107.201423, -117.513032, -128.033025, -138.787721, -149.803437, -161.106494, -172.976188, -185.474979, -198.056969, -210.802684, -223.792655, -237.107409, -250.827475, -265.033382, -279.805658, -295.224831, -311.37143, -328.325983, -346.16902, -365.167703, -385.147493, -406.132928, -428.148546, -451.254831, -475.458346, -500.756072, -527.161764, -554.689176, -583.352061, -613.164174, -642.731422, -673.903882, -706.970756, -742.221248, -779.94456, -823.66135, -869.621501, -917.306352, -966.197243, -1015.77551, -1064.01817, -1113.74951, -1165.33274, -1219.23864, -1276.15761, -1336.68147, -1401.402, -1475.06656, -1553.16695, -1626.66164, -1691.89792, -1744.31532, -1778.50424, -1790.2965, -1777.62735, -1737.46655, -1678.34877, -1600.24209, -1460.70974, -1241.30154, -1029.90872, -685.379434, -152.453001, 624.131267, 1699.63406, 4652.95067, 7634.79853, 10319.5297, 12381.4961, 13495.0498, 11900.0785, 9661.70819, 7069.06148, 5314.7322, 4330.21881, 3670.57341, 3286.11678, 3127.1697, 3144.05297, 3287.08737, 3564.83542, 3765.11061, 3913.37842, 4035.10433, 4155.75383, 4300.79239, 4495.68551, 4584.7724, 4689.48834},
        {1007.70859, 1000.90376, 993.785046, 986.36422, 978.653058, 970.663338, 962.406836, 953.895329, 945.140593, 936.154406, 926.948543, 917.534781, 907.924898, 898.130669, 888.163872, 878.036283, 867.759678, 857.345835, 846.806529, 836.153538, 825.398638, 814.553606, 803.630219, 792.640252, 781.595484, 770.50769, 759.388646, 748.250131, 737.10392, 725.96179, 714.835517, 703.736879, 692.677651, 681.669612, 671.31196, 661.310405, 651.363596, 641.464246, 631.605069, 621.778779, 611.978089, 602.195713, 592.424365, 582.656757, 572.885604, 563.103619, 553.303517, 543.478009, 533.619811, 523.721636, 513.776197, 503.264245, 492.701554, 482.091938, 471.439207, 460.747175, 450.019652, 439.260451, 428.473384, 417.662262, 406.830898, 395.983104, 385.122692, 374.253473, 363.379259, 352.503864, 341.631098, 330.767236, 319.918494, 309.083604, 298.26622, 287.469993, 276.698575, 265.955619, 255.244778, 244.569703, 233.934046, 223.341461, 212.7956, 202.300114, 191.858657, 181.47488, 171.152436, 160.894977, 150.978155, 141.263902, 131.61088, 122.013913, 112.467828, 102.96745, 93.5076037, 84.0831151, 74.6888095, 65.3195123, 55.9700488, 46.6352446, 37.3099249, 27.9889152, 18.6670409, 9.33912734, 0, -9.65535587, -19.1717386, -28.5793301, -37.9083126, -47.1888681, -56.4511787, -65.7254265, -75.0417934, -84.4304616, -93.9216131, -103.54543, -113.332095, -123.311789, -133.514694, -143.970993, -154.710868, -165.80745, -177.297055, -189.11272, -201.293832, -213.879774, -226.909934, -240.423695, -254.460443, -269.059563, -284.260441, -300.102463, -316.625012, -333.867475, -351.723639, -370.431873, -390.075242, -410.736811, -432.643998, -455.807691, -480.200286, -505.86154, -532.831215, -561.149068, -590.85486, -621.679603, -654.065769, -688.113524, -723.923034, -761.594465, -801.790813, -843.908706, -887.907603, -933.746962, -981.386243, -1030.97307, -1082.04875, -1137.56583, -1197.70482, -1256.98669, -1312.47343, -1361.22704, -1390.35549, -1409.13706, -1437.70894, -1508.64582, -1577.03909, -1600.53432, -1560.88572, -1480.69828, -1363.82582, -1236.14628, -1107.22343, -905.865883, -606.561914, -318.570836, 139.124079, 838.686049, 1852.27829, 3252.06401, 6984.84961, 10779.4943, 14239.5006, 16968.3708, 18569.6074, 16950.1791, 14541.1444, 11673.0918, 9682.50479, 8501.80475, 7644.29024, 7063.3389, 6712.32837, 6544.63629, 6513.64031, 6691.52276, 6770.89832, 6787.70053, 6777.86297, 6777.31919, 6822.00275, 6947.84721, 6862.93289, 6916.82987},
        {938.236429, 932.751583, 926.986165, 920.948866, 914.648375, 908.093381, 901.292575, 894.254647, 886.988286, 879.502182, 871.805024, 863.905503, 855.812308, 847.53413, 839.079657, 830.457579, 821.676588, 812.745371, 803.672619, 794.467022, 785.13727, 775.692051, 766.140057, 756.489977, 746.7505, 736.930317, 727.038116, 717.082589, 707.072424, 697.016312, 686.922943, 676.801005, 666.659189, 656.506185, 646.662422, 636.974164, 627.283927, 617.590282, 607.891804, 598.187065, 588.474639, 578.753098, 569.021017, 559.276969, 549.519525, 539.747261, 529.958748, 520.15256, 510.327271, 500.481453, 490.61368, 480.609924, 470.583801, 460.536324, 450.468507, 440.381364, 430.27591, 420.153158, 410.014124, 399.85982, 389.691261, 379.509462, 369.315435, 359.110196, 348.894759, 338.670138, 328.437346, 318.140428, 307.724796, 297.308961, 286.897629, 276.495508, 266.107305, 255.737728, 245.391484, 235.07328, 224.787823, 214.539821, 204.333981, 194.175011, 184.067616, 174.016506, 164.026386, 154.101965, 144.55543, 135.231282, 125.970323, 116.767282, 107.616889, 98.5138722, 89.4529607, 80.4288834, 71.4363692, 62.4701471, 53.524946, 44.5954948, 35.6765225, 26.7627579, 17.84893, 8.92976772, 0, -9.49933364, -18.8127142, -27.9733161, -37.014314, -45.9688825, -54.870196, -63.7514291, -72.6457563, -81.5863522, -90.6063914, -99.7390483, -109.017498, -118.474914, -128.144471, -138.059345, -148.252708, -158.68544, -169.380385, -180.534249, -192.164714, -204.289461, -216.926173, -230.092533, -243.806222, -258.084923, -272.946318, -288.408089, -304.487919, -321.20349, -338.299216, -356.166245, -374.904241, -394.612865, -415.586265, -437.826862, -461.285212, -486.002632, -512.02044, -539.37995, -568.122482, -598.563705, -630.387083, -663.580255, -698.13086, -734.026536, -769.402137, -806.561284, -845.95481, -888.033551, -933.248341, -984.325978, -1036.6596, -1092.3427, -1150.6599, -1205.1611, -1251.97099, -1287.21424, -1292.87004, -1282.42344, -1284.79095, -1345.058, -1402.93223, -1400.81574, -1314.74865, -1177.76879, -996.751105, -806.69623, -621.77096, -353.014521, 30.1990757, 406.21051, 981.35959, 1838.98645, 3062.43123, 4735.03406, 9024.92018, 13409.4483, 17450.7624, 20711.0062, 22752.3237, 21310.5195, 18991.636, 16140.5339, 14126.0651, 12887.7603, 11938.4872, 11239.1099, 10750.4922, 10433.498, 10248.9913, 10469.1952, 10476.7284, 10343.6612, 10142.0639, 9944.00693, 9821.56058, 9846.79521, 9585.65271, 9506.64373},
        {874.787044, 869.929395, 864.881064, 859.645657, 854.226775, 848.628025, 842.853009, 836.905332, 830.788597, 824.506409, 818.062372, 811.460089, 804.703165, 797.795204, 790.739809, 783.540585, 776.201136, 768.725066, 761.115978, 753.377477, 745.513167, 737.526652, 729.421535, 721.201421, 712.869914, 704.430618, 695.887137, 687.243074, 678.502034, 669.667621, 660.743439, 651.733091, 642.640183, 633.468317, 624.198352, 614.845744, 605.42622, 595.944124, 586.403797, 576.809582, 567.165821, 557.476855, 547.747028, 537.980681, 528.182157, 518.355797, 508.505944, 498.636939, 488.753126, 478.858847, 468.958442, 459.252288, 449.544444, 439.835003, 430.124057, 420.411699, 410.698021, 400.983116, 391.267077, 381.549996, 371.831965, 362.113078, 352.393426, 342.673103, 332.952201, 323.230812, 313.509029, 303.731617, 293.844667, 283.962384, 274.088446, 264.226532, 254.380322, 244.553495, 234.749731, 224.972708, 215.226106, 205.513604, 195.838881, 186.205617, 176.617491, 167.078182, 157.59137, 148.160733, 139.039631, 130.101652, 121.217381, 112.382397, 103.592275, 94.8425929, 86.1289262, 77.4468521, 68.7919472, 60.1597881, 51.5459516, 42.9460144, 34.355553, 25.7701442, 17.1853647, 8.59679104, 0, -9.38296431, -18.5290501, -27.4744248, -36.2552561, -44.9077113, -53.467958, -61.9721638, -70.456496, -78.9571223, -87.5102101, -96.151927, -104.91844, -113.845918, -122.970527, -132.328435, -141.95581, -151.783843, -161.833706, -172.379011, -183.43343, -195.010636, -207.124302, -219.788101, -233.015706, -246.820788, -261.217022, -276.218079, -291.837633, -308.089357, -324.821454, -342.273738, -360.509522, -379.592119, -399.493707, -420.323166, -442.213683, -465.255911, -489.540505, -515.15812, -542.19941, -572.209819, -603.38245, -635.523326, -668.438469, -701.9339, -730.695327, -760.929166, -793.721517, -830.158479, -871.326153, -923.056439, -975.889217, -1028.08154, -1077.59284, -1121.77493, -1158.25243, -1184.64996, -1190.96914, -1184.19009, -1179.60892, -1206.77504, -1221.0388, -1178.85061, -1061.21585, -890.747039, -670.958338, -432.708069, -186.59962, 157.025535, 631.112112, 1117.68827, 1823.83322, 2834.73201, 4235.56971, 6111.53137, 10636.4231, 15284.6537, 19619.2529, 23203.2505, 25599.6763, 24601.9785, 22722.4896, 20282.6314, 18529.017, 17416.8469, 16515.8249, 15800.4264, 15245.1266, 14824.4005, 14512.7234, 14882.8889, 14864.4784, 14579.3832, 14149.4946, 13696.704, 13342.9029, 13209.9826, 12768.9095, 12458.9559},
        {819.559987, 815.060002, 810.436976, 805.691155, 800.822782, 795.832104, 790.719363, 785.484806, 780.128677, 774.651221, 769.052682, 763.333305, 757.493335, 751.533017, 745.452595, 739.252314, 732.932419, 726.493154, 719.934765, 713.257496, 706.461592, 699.547298, 692.514858, 685.364517, 678.096519, 670.71111, 663.208535, 655.589037, 647.852863, 640.000256, 632.031461, 623.946723, 615.746286, 607.430396, 598.798162, 589.954626, 581.00725, 571.962805, 562.828063, 553.609797, 544.314778, 534.949779, 525.521571, 516.036926, 506.502617, 496.925415, 487.312092, 477.669421, 468.004173, 458.32312, 448.633034, 439.207205, 429.780109, 420.352741, 410.926096, 401.501167, 392.078951, 382.66044, 373.246629, 363.838513, 354.437086, 345.043343, 335.658277, 326.282884, 316.918158, 307.565094, 298.224685, 288.882203, 279.523297, 270.181388, 260.858492, 251.556621, 242.277789, 233.02401, 223.797298, 214.599665, 205.433127, 196.299696, 187.201387, 178.140213, 169.118187, 160.137324, 151.199637, 142.307139, 133.63045, 125.083735, 116.57913, 108.11318, 99.6824258, 91.2834096, 82.9126736, 74.5667601, 66.242211, 57.9355687, 49.6433752, 41.3621726, 33.0885033, 24.8189091, 16.5499325, 8.27811537, 0, -9.02478904, -17.8130648, -26.4001232, -34.8212601, -43.1117714, -51.306953, -59.4421009, -67.5525109, -75.6734789, -83.8403008, -92.0882726, -100.45269, -108.968849, -117.672046, -126.597576, -135.780735, -145.214079, -154.926796, -165.05086, -175.612407, -186.637576, -198.152501, -210.183322, -222.756174, -235.897196, -249.632523, -263.988294, -278.990644, -294.665711, -311.114986, -328.261622, -346.109151, -364.661105, -383.421399, -402.643371, -422.713593, -443.785482, -466.012453, -489.547924, -514.54531, -543.72332, -573.889337, -604.694873, -635.79144, -666.830549, -690.346132, -714.886675, -741.883084, -772.766266, -808.967127, -858.053603, -907.818758, -953.155022, -991.130048, -1023.25256, -1049.03734, -1067.99915, -1081.08802, -1086.05728, -1079.09458, -1066.22844, -1028.09476, -945.735479, -808.563552, -620.176065, -377.659351, -100.742442, 207.618001, 627.488543, 1191.97311, 1797.55638, 2634.97989, 3784.63885, 5326.92848, 7342.24398, 11887.3802, 16572.2328, 20983.0971, 24706.2684, 27328.0421, 26800.8324, 25434.0697, 23514.2841, 22133.245, 21258.7418, 20523.0479, 19910.7285, 19406.349, 18994.4749, 18659.6714, 19100.6969, 19078.6583, 18730.2901, 18192.3268, 17601.503, 17094.553, 16808.2113, 16196.4704, 15725.0044},
        {771.4088, 767.451892, 763.376474, 759.182567, 754.870193, 750.439371, 745.890122, 741.222467, 736.436426, 731.53202, 726.509271, 721.368197, 716.10882, 710.731161, 705.23524, 699.621078, 693.888695, 688.038113, 682.069351, 675.98243, 669.777372, 663.454196, 657.012923, 650.453574, 643.776169, 636.980729, 630.067275, 623.035828, 615.886407, 608.619034, 601.23373, 593.730514, 586.109407, 578.370431, 570.313233, 562.042235, 553.664267, 545.185851, 536.61351, 527.953769, 519.213149, 510.398173, 501.515366, 492.57125, 483.572347, 474.525182, 465.436277, 456.312155, 447.159339, 437.984353, 428.79372, 419.834704, 410.867869, 401.894517, 392.915954, 383.933484, 374.94841, 365.962037, 356.97567, 347.990612, 339.008169, 330.029643, 321.05634, 312.089563, 303.130618, 294.180807, 285.241436, 276.303235, 267.357188, 258.426407, 249.512883, 240.618603, 231.745559, 222.89574, 214.071135, 205.273735, 196.505528, 187.768506, 179.064656, 170.39597, 161.764437, 153.172046, 144.620787, 136.11265, 127.814167, 119.641595, 111.509214, 103.413675, 95.351627, 87.3197209, 79.3146064, 71.3329334, 63.3713521, 55.4265123, 47.4950642, 39.5736577, 31.6589428, 23.7475696, 15.8361881, 7.9214482, 0, -8.37473907, -16.575884, -24.6334616, -32.5774985, -40.4380214, -48.2450571, -56.0286322, -63.8187735, -71.6455076, -79.5388613, -87.5288613, -95.6455342, -103.918907, -112.379006, -121.055858, -129.97949, -139.209112, -148.77892, -158.65293, -168.867423, -179.458678, -190.462977, -201.916599, -213.855825, -226.316936, -239.33621, -252.949929, -267.194374, -282.105824, -297.915364, -314.393043, -331.516697, -349.264166, -366.925583, -384.82264, -403.460416, -423.023062, -443.694727, -465.659561, -489.101717, -517.207805, -546.245721, -575.812176, -605.503884, -634.917555, -656.880489, -679.451164, -703.918645, -731.571996, -763.700282, -807.447881, -851.092053, -890.204868, -921.970355, -946.863623, -963.882153, -972.023427, -971.293017, -959.451201, -933.158521, -897.649541, -835.100252, -730.968859, -575.136773, -364.980271, -94.4353617, 222.233361, 585.819515, 1071.65485, 1712.90814, 2419.99215, 3367.71717, 4631.74446, 6287.73525, 8411.35076, 12978.9832, 17690.3801, 22146.02, 25946.3812, 28691.9424, 28417.4796, 27332.9752, 25709.9229, 24586.172, 23929.8671, 23389.8379, 22950.195, 22595.0486, 22308.5091, 22074.6868, 22364.1604, 22329.8694, 22057.3927, 21632.309, 21140.1972, 20666.6361, 20297.2046, 19539.5521, 19101.3984},
        {727.568456, 724.083669, 720.472765, 716.736307, 712.874857, 708.888981, 704.77924, 700.546198, 696.190418, 691.712464, 687.112899, 682.392286, 677.551188, 672.590169, 667.509792, 662.31062, 656.993217, 651.558146, 646.005969, 640.337251, 634.552554, 628.652443, 622.637479, 616.508227, 610.26525, 603.90911, 597.440372, 590.859599, 584.167353, 577.364199, 570.450698, 563.427416, 556.294914, 549.053757, 541.561897, 533.894203, 526.127257, 518.266248, 510.31637, 502.282813, 494.170768, 485.985426, 477.731979, 469.415618, 461.041534, 452.614919, 444.140963, 435.624858, 427.071796, 418.486966, 409.875561, 401.394335, 392.893631, 384.375353, 375.841409, 367.293702, 358.73414, 350.164628, 341.58707, 333.003374, 324.415445, 315.825187, 307.234508, 298.645312, 290.059506, 281.478994, 272.905683, 264.339158, 255.779061, 247.232081, 238.700276, 230.185701, 221.690411, 213.216464, 204.765915, 196.340821, 187.943236, 179.575218, 171.238822, 162.936105, 154.669122, 146.43993, 138.250584, 130.103141, 122.165405, 114.353073, 106.579846, 98.8424029, 91.1374196, 83.4615736, 75.8115422, 68.1840024, 60.5756314, 52.9831065, 45.4031046, 37.832303, 30.2673789, 22.7050094, 15.1418716, 7.57464278, 0, -7.68280688, -15.2667915, -22.775822, -30.2337668, -37.6644942, -45.0918724, -52.5397698, -60.0320546, -67.592595, -75.2452595, -83.0139162, -90.9224335, -98.9946797, -107.254523, -115.725832, -124.432474, -133.506165, -142.986179, -152.652447, -162.551946, -172.731654, -183.238551, -194.119614, -205.421821, -217.192151, -229.477582, -242.325092, -255.781659, -269.894263, -285.01079, -300.766975, -317.119525, -334.025144, -350.619823, -367.270626, -384.563471, -402.701278, -421.886966, -442.323455, -464.213665, -490.957853, -518.588497, -546.682952, -574.81857, -602.572705, -623.90216, -645.409977, -668.078646, -692.89066, -720.828508, -757.867343, -793.894854, -825.795708, -851.138468, -868.887982, -877.382197, -874.959056, -860.161551, -831.075979, -785.564955, -728.569661, -643.575236, -520.686062, -350.23308, -122.931121, 170.681196, 522.976116, 938.984506, 1486.39739, 2197.46166, 2999.68102, 4046.15989, 5405.40242, 7145.91276, 9336.19505, 13826.8712, 18458.7981, 22854.9506, 26638.3033, 29431.831, 29386.2495, 28578.2982, 27261.9201, 26417.3264, 26010.9787, 25706.7266, 25486.162, 25330.8772, 25222.4641, 25142.5149, 25243.2475, 25216.7702, 25078.22, 24842.734, 24525.4494, 24141.5032, 23706.0325, 22844.1392, 22460.2315},
        {686.487617, 683.454805, 680.281109, 676.96804, 673.517106, 669.929818, 666.207684, 662.352216, 658.364921, 654.247311, 650.000893, 645.627179, 641.127678, 636.503899, 631.757351, 626.889546, 621.901991, 616.796197, 611.573673, 606.235929, 600.784475, 595.22082, 589.546474, 583.762946, 577.871746, 571.874383, 565.772368, 559.567209, 553.260417, 546.853501, 540.34797, 533.745335, 527.047105, 520.254789, 513.312464, 506.251565, 499.104216, 491.87379, 484.56366, 477.1772, 469.717783, 462.188782, 454.593572, 446.935525, 439.218015, 431.444414, 423.618098, 415.742438, 407.820809, 399.856583, 391.853134, 383.84992, 375.813447, 367.746307, 359.651092, 351.530391, 343.386797, 335.2229, 327.041293, 318.844565, 310.635308, 302.416114, 294.189573, 285.958277, 277.724816, 269.491783, 261.261768, 253.043153, 244.844183, 236.655504, 228.479332, 220.317882, 212.173371, 204.048014, 195.944027, 187.863625, 179.809025, 171.782443, 163.786093, 155.822191, 147.892955, 140.000598, 132.147337, 124.335388, 116.738337, 109.269109, 101.838572, 94.4433794, 87.0801856, 79.7456453, 72.4364128, 65.1491427, 57.8804893, 50.6271072, 43.3856507, 36.1527743, 28.9251325, 21.6993798, 14.4721704, 7.24015904, 0, -7.03254819, -14.0359936, -21.0283911, -28.0277956, -35.052262, -42.119845, -49.2485997, -56.4565808, -63.7618433, -71.182442, -78.7364318, -86.4418675, -94.3168041, -102.379296, -110.647399, -119.139168, -128.041565, -137.396775, -146.840799, -156.427886, -166.212285, -176.248247, -186.590021, -197.291856, -208.408002, -219.992707, -232.100223, -244.784798, -258.100681, -272.467753, -287.440568, -302.963687, -318.981669, -334.578523, -350.129085, -366.237673, -383.107014, -400.939834, -419.938859, -440.306816, -465.364113, -491.246935, -517.548027, -543.860136, -569.776005, -590.97928, -611.949081, -633.255431, -655.46835, -679.157858, -708.665057, -736.179789, -760.165151, -778.722349, -789.213717, -789.333327, -776.775251, -748.680948, -703.422697, -639.975628, -562.874254, -457.049237, -317.268363, -134.372653, 107.450552, 420.959763, 805.586542, 1270.63426, 1874.27179, 2647.0269, 3534.67541, 4665.53584, 6099.47315, 7896.35236, 10116.0384, 14451.8594, 18921.8514, 23177.5138, 26870.3459, 29651.8468, 29813.6764, 29273.7331, 28266.3044, 27708.3187, 27563.8515, 27512.5451, 27532.6969, 27602.6042, 27700.5644, 27804.875, 27714.2812, 27719.3307, 27755.3989, 27757.8609, 27662.0919, 27403.4673, 26917.3622, 26008.8094, 25663.7781},
        {646.614948, 644.064771, 641.356767, 638.493434, 635.477268, 632.310765, 628.996422, 625.536735, 621.934202, 618.191317, 614.310579, 610.294484, 606.145528, 601.866207, 597.459018, 592.926458, 588.271024, 583.495211, 578.601516, 573.592437, 568.470469, 563.238108, 557.897852, 552.452197, 546.90364, 541.254676, 535.507803, 529.665518, 523.730316, 517.704694, 511.591149, 505.392177, 499.110274, 492.747939, 486.333242, 479.855356, 473.303142, 466.678267, 459.982398, 453.217201, 446.384342, 439.485489, 432.522307, 425.496464, 418.409625, 411.263457, 404.059627, 396.799801, 389.485646, 382.118829, 374.701015, 367.165281, 359.583371, 351.958438, 344.293635, 336.592117, 328.857035, 321.091545, 313.298799, 305.48195, 297.644153, 289.78856, 281.918325, 274.036601, 266.146542, 258.251301, 250.354032, 242.468401, 234.607823, 226.753769, 218.908709, 211.075117, 203.255462, 195.452219, 187.667858, 179.904851, 172.16567, 164.452787, 156.768674, 149.115802, 141.496644, 133.913672, 126.369357, 118.866171, 111.587133, 104.440645, 97.3329357, 90.260619, 83.2203073, 76.2086134, 69.22215, 62.2575299, 55.3113658, 48.3802705, 41.4608568, 34.5497373, 27.6435248, 20.7388321, 13.8322719, 6.92045692, 0, -6.50751874, -13.0336971, -19.5923558, -26.1973155, -32.8623969, -39.6014208, -46.4282079, -53.3565789, -60.4003545, -67.5733554, -74.8894023, -82.362316, -90.0059171, -97.8340264, -105.860465, -114.099052, -122.751639, -131.858908, -141.009375, -150.257152, -159.656354, -169.261091, -179.125478, -189.303626, -199.849648, -210.817657, -222.261765, -234.236086, -246.794731, -260.352746, -274.470969, -289.095233, -304.171373, -318.876086, -333.539774, -348.697939, -364.527153, -381.203991, -398.905025, -417.806829, -440.817233, -464.550301, -488.64823, -512.753217, -536.50746, -557.679982, -578.254446, -598.341342, -618.051159, -637.494388, -659.194091, -677.899485, -693.550809, -704.809964, -707.728907, -699.531404, -677.441223, -637.838535, -578.982431, -500.052292, -404.448768, -279.051781, -123.097038, 71.9243254, 327.643916, 659.669785, 1074.16536, 1584.29007, 2237.83369, 3062.99711, 4023.02773, 5221.07279, 6707.81708, 8533.9454, 10750.1426, 14874.7631, 19123.9046, 23181.3344, 26730.8199, 29456.1285, 29806.295, 29522.9743, 28819.1048, 28540.7592, 28650.2603, 28846.1251, 29103.8669, 29398.9989, 29707.0343, 30003.4863, 29753.5841, 29817.521, 30051.556, 30311.9484, 30454.9572, 30336.8416, 29813.8608, 28932.1403, 28574.3128},
        {606.399111, 604.413037, 602.254999, 599.928157, 597.435672, 594.780706, 591.966419, 588.995972, 585.872526, 582.599243, 579.179283, 575.615807, 571.911976, 568.070951, 564.095894, 559.989964, 555.756323, 551.398132, 546.918552, 542.320744, 537.607869, 532.783088, 527.849561, 522.81045, 517.668916, 512.42812, 507.091222, 501.661384, 496.141766, 490.53553, 484.845836, 479.075846, 473.22872, 467.30762, 461.39254, 455.446609, 449.432031, 443.349473, 437.199602, 430.983087, 424.700594, 418.352792, 411.940347, 405.463928, 398.924202, 392.321836, 385.657498, 378.931856, 372.145577, 365.299328, 358.393777, 351.30424, 344.159453, 336.962802, 329.717671, 322.427446, 315.095511, 307.725251, 300.320052, 292.883298, 285.418373, 277.928664, 270.417555, 262.888431, 255.344676, 247.789676, 240.226816, 232.668081, 225.12525, 217.58397, 210.047068, 202.517373, 194.997711, 187.490909, 179.999796, 172.527198, 165.075943, 157.648859, 150.248773, 142.878512, 135.540903, 128.238775, 120.974954, 113.752267, 106.765966, 99.9186191, 93.1104815, 86.3381364, 79.5981667, 72.8871555, 66.2016855, 59.5383398, 52.8937012, 46.2643527, 39.6468771, 33.0378575, 26.4338767, 19.8315176, 13.2273632, 6.61799633, 0, -6.19127428, -12.4101085, -18.6689027, -24.980057, -31.3559713, -37.8090458, -44.3516805, -50.9962754, -57.7552306, -64.6409461, -71.665822, -78.8422584, -86.1826552, -93.6994125, -101.40493, -109.311609, -117.572717, -126.22078, -134.949563, -143.801655, -152.819641, -162.046107, -171.523639, -181.294825, -191.402249, -201.8885, -212.796162, -224.167822, -236.046067, -248.732261, -261.915326, -275.560216, -289.631883, -303.586916, -317.64445, -332.159183, -347.248579, -363.030099, -379.621207, -397.139364, -417.707862, -438.827864, -460.224389, -481.622452, -502.747074, -523.572402, -543.512039, -562.22872, -579.385179, -594.644151, -608.807513, -619.00657, -626.190291, -629.489277, -624.321629, -607.772285, -576.926182, -528.621638, -460.24626, -369.456695, -257.178654, -113.112396, 59.4466376, 268.137675, 539.128138, 890.080704, 1332.81328, 1883.47322, 2579.63874, 3446.76555, 4462.79039, 5707.99855, 7224.29464, 9053.58324, 11237.769, 15116.3976, 19109.3223, 22934.0372, 26308.0366, 28948.8149, 29470.6397, 29429.7162, 29016.3501, 28996.2584, 29331.9797, 29746.2979, 30213.7395, 30708.8308, 31206.0982, 31680.068, 31337.4793, 31491.3111, 31929.3182, 32439.2555, 32808.8776, 32825.9393, 32278.1954, 31512.7098, 31054.1103},
        {558.061705, 557.077395, 555.916638, 554.582013, 553.076094, 551.40146, 549.560687, 547.556351, 545.391031, 543.067303, 540.587743, 537.954928, 535.171436, 532.239843, 529.162726, 525.942661, 522.582227, 519.083999, 515.450555, 511.684471, 507.788324, 503.764692, 499.61615, 495.345276, 490.954647, 486.446839, 481.82443, 477.089996, 472.246114, 467.295361, 462.240314, 457.083549, 451.827645, 446.475176, 441.033131, 435.501788, 429.881373, 424.174321, 418.383065, 412.510038, 406.557675, 400.52841, 394.424675, 388.248906, 382.003535, 375.690997, 369.313724, 362.874152, 356.374713, 349.817842, 343.205972, 336.551859, 329.84739, 323.094777, 316.296229, 309.453956, 302.570167, 295.647074, 288.686886, 281.691813, 274.664065, 267.605852, 260.519384, 253.406871, 246.270523, 239.11255, 231.935163, 224.714353, 217.426743, 210.128614, 202.823875, 195.516437, 188.21021, 180.909101, 173.617022, 166.337882, 159.07559, 151.834056, 144.617189, 137.4289, 130.273097, 123.15369, 116.074589, 109.039704, 102.28466, 95.6926344, 89.1440204, 82.6352077, 76.1625865, 69.7225465, 63.3114779, 56.9257704, 50.5618141, 44.2159989, 37.8847148, 31.5643517, 25.2512996, 18.9419484, 12.6326881, 6.31990867, 0, -6.56266698, -13.0243977, -19.4063248, -25.7295809, -32.0152988, -38.284611, -44.5586503, -50.8585493, -57.2054407, -63.6204572, -70.1247313, -76.7393959, -83.4855835, -90.3844268, -97.4570585, -104.724611, -112.005124, -119.290715, -127.061898, -135.296284, -143.971485, -153.065115, -162.554786, -172.418111, -182.632702, -193.176171, -204.026131, -215.160195, -226.555976, -237.881852, -249.538055, -261.594967, -274.12297, -287.679742, -302.092017, -317.056585, -332.49764, -348.339374, -364.505983, -380.921659, -397.233827, -413.727684, -430.381575, -447.173844, -464.082835, -484.870045, -504.784878, -522.85989, -538.127637, -549.620677, -556.024083, -557.142594, -552.653891, -541.868128, -523.345082, -495.981434, -458.673865, -417.121095, -361.871848, -276.054438, -145.038549, 26.6139919, 223.63123, 449.086722, 737.139475, 1107.67976, 1579.46167, 2172.08089, 2909.29919, 3812.52176, 4863.6904, 6129.745, 7645.75117, 9446.77451, 11567.8806, 15175.7916, 18891.0024, 22465.6646, 25651.9296, 28201.949, 28909.5268, 29124.0601, 29008.4209, 29210.5503, 29701.3134, 30249.4632, 30835.9705, 31441.806, 32047.9405, 32635.3447, 32376.7063, 32707.5397, 33366.6969, 34093.0297, 34625.3901, 34702.6298, 34063.6009, 33642.4121, 32792.9151},
        {519.490016, 519.042824, 518.423423, 517.633955, 516.676564, 515.553393, 514.266585, 512.818283, 511.21063, 509.44577, 507.525844, 505.452997, 503.229372, 500.857111, 498.338358, 495.675255, 492.869947, 489.924575, 486.841283, 483.622214, 480.269511, 476.785317, 473.171775, 469.431029, 465.565221, 461.576495, 457.466993, 453.238859, 448.894236, 444.435266, 439.864093, 435.182861, 430.393711, 425.498787, 420.4739, 415.334914, 410.098006, 404.766176, 399.342421, 393.829738, 388.231125, 382.54958, 376.7881, 370.949683, 365.037326, 359.054027, 353.002784, 346.886594, 340.708456, 334.471365, 328.178321, 321.872362, 315.515577, 309.110094, 302.658044, 296.161556, 289.62276, 283.043785, 276.426761, 269.773817, 263.087084, 256.36869, 249.620765, 242.845439, 236.044841, 229.221102, 222.37635, 215.48725, 208.531081, 201.562488, 194.585253, 187.603157, 180.619978, 173.639499, 166.665498, 159.701756, 152.752055, 145.820173, 138.909892, 132.024991, 125.169252, 118.346454, 111.560377, 104.814803, 98.3315939, 92.0009021, 85.710258, 79.4563648, 73.2359255, 67.0456433, 60.8822213, 54.7423626, 48.6227704, 42.5201478, 36.4311979, 30.3526238, 24.2811286, 18.2134155, 12.1461877, 6.07614811, 0, -6.47190359, -12.814736, -19.0507049, -25.2020177, -31.290882, -37.3395053, -43.3700952, -49.4048593, -55.466005, -61.57574, -67.7562717, -74.0298078, -80.4185557, -86.944723, -93.6305173, -100.498146, -107.34387, -114.157618, -121.472669, -129.262812, -137.50184, -146.16354, -155.221705, -164.650124, -174.422588, -184.512887, -194.89481, -205.54215, -216.428695, -227.216874, -238.306005, -249.763288, -261.655921, -274.664475, -288.549463, -302.907832, -317.622768, -332.577458, -347.65509, -362.73885, -377.255124, -391.682927, -405.99482, -420.163365, -434.161123, -451.592957, -467.89105, -482.11989, -493.343962, -500.627753, -502.824567, -499.468182, -490.815342, -476.188151, -453.000487, -419.522982, -374.026271, -319.508331, -248.438053, -148.127218, -9.28820842, 169.820138, 380.748512, 628.783913, 942.488753, 1339.74165, 1837.14576, 2452.24904, 3207.27609, 4121.80611, 5187.57437, 6456.28773, 7957.16371, 9719.41981, 11772.2736, 15141.6664, 18610.9109, 21960.0436, 24969.1011, 27418.1198, 28239.5927, 28626.1291, 28720.9984, 29096.0142, 29724.8141, 30402.4755, 31111.5641, 31834.6455, 32554.2851, 33253.0486, 33053.5415, 33488.65, 34280.559, 35151.4533, 35823.5181, 36018.9383, 35459.8988, 35243.7613, 34195.2425},
        {489.408433, 489.081264, 488.595352, 487.952476, 487.154412, 486.202935, 485.099824, 483.846855, 482.445803, 480.898447, 479.206563, 477.371926, 475.396315, 473.281506, 471.029275, 468.641399, 466.119655, 463.465819, 460.681669, 457.76898, 454.72953, 451.565095, 448.277451, 444.868376, 441.339647, 437.693039, 433.930329, 430.053295, 426.063713, 421.963359, 417.754011, 413.437444, 409.015436, 404.489763, 399.837859, 395.074183, 390.21349, 385.258345, 380.211315, 375.074968, 369.851869, 364.544585, 359.155683, 353.68773, 348.143292, 342.524935, 336.835228, 331.076736, 325.252026, 319.363664, 313.414218, 307.392839, 301.3158, 295.185958, 289.00617, 282.779294, 276.508188, 270.195708, 263.844712, 257.458058, 251.038602, 244.589204, 238.112719, 231.612006, 225.089922, 218.549324, 211.99307, 205.424575, 198.84724, 192.262773, 185.673995, 179.083728, 172.494794, 165.910012, 159.332205, 152.764194, 146.208801, 139.668846, 133.147151, 126.646538, 120.169827, 113.71984, 107.299398, 100.911323, 94.7277517, 88.6632867, 82.6304946, 76.626702, 70.6492357, 64.6954222, 58.7625884, 52.8480608, 46.9491663, 41.0632314, 35.1875829, 29.3195474, 23.4564517, 17.5956224, 11.7343862, 5.87006987, 0, -6.01962909, -11.9604982, -17.839976, -23.6754307, -29.4842311, -35.2837455, -41.0913425, -46.9243906, -52.8002582, -58.736314, -64.7499263, -70.8584638, -77.0792948, -83.429788, -89.9273118, -96.5892347, -103.413559, -110.414887, -117.654391, -125.14529, -132.900803, -140.934147, -149.258542, -157.887207, -166.833359, -176.110218, -185.731003, -195.708931, -206.057222, -216.876873, -228.061139, -239.596903, -251.471051, -263.774713, -276.442652, -289.38183, -302.547857, -315.896346, -329.382906, -342.963149, -357.518825, -371.797536, -385.573695, -398.621709, -410.715989, -422.345034, -432.390643, -440.448703, -446.115101, -448.985725, -449.893731, -445.685521, -438.096306, -426.226482, -403.797023, -366.944156, -311.804105, -231.375615, -125.645456, 1.97689421, 142.327997, 316.430006, 533.804386, 806.041164, 1148.23543, 1573.87339, 2092.2309, 2715.70259, 3472.12108, 4380.5864, 5445.21431, 6702.39437, 8176.41346, 9891.55844, 11872.1162, 15030.0084, 18279.9785, 21424.4046, 24265.665, 26606.1377, 27483.8734, 27975.1297, 28209.8535, 28716.4834, 29467.2304, 30268.2673, 31100.1071, 31943.2626, 32778.2467, 33585.5724, 33415.4069, 33897.574, 34758.794, 35725.7871, 36525.2735, 36883.9734, 36528.607, 36386.4734, 35334.7462},
        {464.357294, 463.888202, 463.284024, 462.545972, 461.675253, 460.673078, 459.540656, 458.279196, 456.889907, 455.374001, 453.732684, 451.967169, 450.078662, 448.068375, 445.937517, 443.687296, 441.318923, 438.833607, 436.232558, 433.516984, 430.688096, 427.747102, 424.695213, 421.533637, 418.263585, 414.886265, 411.402887, 407.814661, 404.122796, 400.328502, 396.432987, 392.437462, 388.343136, 384.151218, 379.843327, 375.43088, 370.92553, 366.329121, 361.643499, 356.870509, 352.011995, 347.069805, 342.045781, 336.941771, 331.759618, 326.501169, 321.168268, 315.76276, 310.286492, 304.741307, 299.129051, 293.361522, 287.532565, 281.645976, 275.705554, 269.715095, 263.678397, 257.599257, 251.481472, 245.32884, 239.145158, 232.934223, 226.699832, 220.445783, 214.175873, 207.893899, 201.603659, 195.342034, 189.145114, 182.948458, 176.753721, 170.562554, 164.376611, 158.197544, 152.027004, 145.866645, 139.71812, 133.58308, 127.463178, 121.360068, 115.2754, 109.210828, 103.168005, 97.148582, 91.2673359, 85.4669779, 79.6888597, 73.9309629, 68.1912688, 62.4677589, 56.7584147, 51.0612176, 45.3741489, 39.6951903, 34.0223231, 28.3535287, 22.6867886, 17.0200842, 11.351397, 5.67870849, 0, -5.44678427, -10.8926084, -16.3480105, -21.8235285, -27.3297005, -32.8770644, -38.4761583, -44.1375202, -49.8716882, -55.6892002, -61.6005943, -67.6164084, -73.7471807, -80.0034491, -86.3957516, -92.9346263, -99.879619, -107.27684, -114.573596, -121.833784, -129.121299, -136.50004, -144.033903, -151.786784, -159.822582, -168.205191, -176.99851, -186.266435, -196.072864, -207.078653, -218.531854, -230.317273, -242.31972, -253.813779, -264.98937, -276.19914, -287.510964, -298.992717, -310.712275, -322.737513, -338.026089, -352.876597, -366.791519, -379.27334, -389.824542, -394.500465, -397.112522, -398.02498, -397.602109, -396.208178, -397.685529, -394.669378, -389.784308, -381.584652, -360.314654, -319.949617, -254.464846, -146.329933, -3.64082797, 152.954835, 294.793022, 462.78327, 684.526996, 977.994061, 1344.8342, 1794.18836, 2328.05887, 2953.73593, 3704.68461, 4599.56398, 5652.10889, 6886.66023, 8323.58161, 9983.23666, 11885.989, 14848.8862, 17896.4369, 20849.8338, 23530.2696, 25758.9368, 26660.1065, 27216.5074, 27548.0129, 28154.0396, 29004.2249, 29910.0634, 30849.0998, 31798.8787, 32736.9448, 33640.8427, 33489.9775, 33996.5711, 34900.7003, 35942.4421, 36861.8733, 37399.071, 37294.1119, 37146.0766, 36234.2937},
        {437.858147, 437.072284, 436.209679, 435.269375, 434.250411, 433.151828, 431.972668, 430.711972, 429.368779, 427.942132, 426.431071, 424.834636, 423.151869, 421.381811, 419.523502, 417.575984, 415.538296, 413.409481, 411.188579, 408.87463, 406.466676, 403.963758, 401.364916, 398.669192, 395.875625, 392.983258, 389.99113, 386.898284, 383.703758, 380.406596, 377.005837, 373.500522, 369.889692, 366.172389, 362.235765, 358.137159, 353.935253, 349.632719, 345.232229, 340.736455, 336.148068, 331.469741, 326.704145, 321.853951, 316.921833, 311.910461, 306.822508, 301.660644, 296.427543, 291.125876, 285.758315, 280.28747, 274.756944, 269.170275, 263.531005, 257.842673, 252.10882, 246.332987, 240.518712, 234.669537, 228.789002, 222.880647, 216.948013, 210.994639, 205.024065, 199.039833, 193.045482, 187.072856, 181.153119, 175.231438, 169.309518, 163.389064, 157.471783, 151.559379, 145.653558, 139.756026, 133.868488, 127.992649, 122.130215, 116.282892, 110.452385, 104.640399, 98.8486393, 93.0788125, 87.4463791, 81.8937742, 76.3620654, 70.8492665, 65.3533915, 59.8724541, 54.4044683, 48.9474479, 43.4994069, 38.0583591, 32.6223184, 27.1892987, 21.7573137, 16.3243775, 10.8885039, 5.44770679, 0, -5.36290154, -10.7004392, -16.0244597, -21.3468098, -26.6793364, -32.0338861, -37.4223058, -42.8564423, -48.3481423, -53.9092526, -59.55162, -65.2870913, -71.1275131, -77.0847324, -83.1705959, -89.3969504, -95.9252838, -102.78882, -109.660932, -116.585533, -123.606535, -130.76785, -138.113392, -145.687074, -153.532808, -161.694506, -170.216083, -179.14145, -188.51452, -198.870092, -209.581202, -220.544495, -231.656619, -242.264927, -252.540712, -262.801747, -273.109466, -283.525304, -294.110697, -304.927078, -319.693908, -333.701285, -346.294943, -356.820616, -364.624038, -361.675675, -356.540346, -350.407604, -344.467, -339.908086, -344.801959, -345.045847, -338.080463, -320.600869, -287.779752, -235.473315, -159.53776, -49.7292007, 90.6096887, 251.481591, 416.981359, 613.300226, 853.725675, 1151.01004, 1516.99888, 1963.95399, 2502.62251, 3144.87355, 3908.12985, 4806.67268, 5843.01291, 7045.01309, 8431.02888, 10019.416, 11828.53, 14584.995, 17421.8312, 20180.3271, 22701.7715, 24827.4528, 25782.5172, 26435.1574, 26890.723, 27588.4479, 28501.5192, 29464.0056, 30456.0503, 31457.7963, 32449.3866, 33410.9644, 33206.5493, 33755.7761, 34773.4812, 35974.5009, 37073.6717, 37785.8299, 37825.812, 37614.9526, 36931.5993},
        {411.415967, 410.405551, 409.373672, 408.317194, 407.232982, 406.117901, 404.968816, 403.782593, 402.556095, 401.286188, 399.969737, 398.603607, 397.184663, 395.709769, 394.175791, 392.579595, 390.918044, 389.188003, 387.386338, 385.509914, 383.555596, 381.520248, 379.400736, 377.193924, 374.896678, 372.505862, 370.018342, 367.430982, 364.740648, 361.944204, 359.038515, 356.020447, 352.886864, 349.634631, 346.044283, 342.225399, 338.290162, 334.242455, 330.086165, 325.825177, 321.463376, 317.004647, 312.452875, 307.811947, 303.085747, 298.27816, 293.393073, 288.43437, 283.405936, 278.311657, 273.155418, 267.989964, 262.769262, 257.496138, 252.173417, 246.803927, 241.390493, 235.935941, 230.443098, 224.914789, 219.353842, 213.763081, 208.145333, 202.503425, 196.840183, 191.158432, 185.460998, 179.759359, 174.064782, 168.362253, 162.654038, 156.942403, 151.229613, 145.517932, 139.809627, 134.106964, 128.412206, 122.727621, 117.055472, 111.398026, 105.757549, 100.136304, 94.5365585, 88.9605772, 83.5494478, 78.2331046, 72.9398135, 67.6673348, 62.4134288, 57.1758559, 51.9523764, 46.7407508, 41.5387393, 36.3441023, 31.1546002, 25.9679934, 20.7820421, 15.5945068, 10.4031478, 5.20572537, 0, -5.40274393, -10.7320447, -16.0029096, -21.2303461, -26.4293616, -31.6149634, -36.8021589, -42.0059554, -47.2413605, -52.5233813, -57.8670254, -63.2873001, -68.7992127, -74.4177707, -80.1579815, -86.0348523, -92.0089844, -98.0876127, -104.408806, -110.975915, -117.792286, -124.861269, -132.186214, -139.770468, -147.617381, -155.730302, -164.112579, -172.767562, -181.698599, -191.101225, -200.716134, -210.489022, -220.365581, -230.102433, -239.739808, -249.368354, -258.990485, -268.608617, -278.225164, -287.842542, -301.017286, -313.116001, -323.44573, -331.313518, -336.026407, -327.424783, -316.649011, -305.372801, -295.26986, -288.013897, -294.742668, -296.098661, -283.348132, -251.40008, -202.60077, -135.957286, -50.4767132, 51.2533557, 174.620098, 328.916701, 520.367584, 751.031331, 1011.74805, 1307.04218, 1664.62549, 2101.56637, 2640.95193, 3301.41097, 4079.50374, 4984.27371, 6001.75677, 7167.54361, 8498.64185, 10012.0591, 11724.8031, 14281.5528, 16914.7267, 19484.4143, 21850.7056, 23873.6902, 24874.6195, 25611.6473, 26175.7943, 26941.1424, 27886.6671, 28873.4013, 29886.2177, 30909.9887, 31929.5866, 32929.8842, 32754.5901, 33388.9543, 34529.2172, 35871.6195, 37112.4016, 37947.8041, 38074.0674, 37840.9126, 37425.1145},
        {385.697795, 384.857514, 384.01091, 383.154173, 382.283492, 381.395055, 380.485053, 379.549674, 378.585107, 377.587542, 376.553167, 375.478173, 374.358748, 373.191081, 371.971361, 370.695778, 369.360521, 367.961779, 366.495741, 364.958596, 363.346534, 361.655743, 359.882413, 358.022734, 356.072893, 354.029081, 351.887486, 349.644298, 347.295706, 344.837898, 342.267065, 339.579396, 336.771078, 333.838303, 330.520586, 326.947852, 323.25311, 319.440878, 315.515675, 311.48202, 307.344431, 303.107428, 298.775528, 294.353251, 289.845115, 285.255638, 280.589341, 275.85074, 271.044355, 266.174705, 261.246307, 256.369302, 251.440298, 246.461523, 241.435207, 236.36358, 231.248869, 226.093305, 220.899116, 215.668531, 210.403779, 205.10709, 199.780692, 194.426815, 189.047687, 183.645537, 178.222595, 172.772381, 167.288625, 161.791516, 156.283849, 150.768415, 145.24801, 139.725427, 134.203459, 128.6849, 123.172543, 117.669182, 112.177611, 106.700622, 101.241011, 95.8015699, 90.3850926, 84.9943728, 79.7962936, 74.7081525, 69.6452745, 64.6051281, 59.585182, 54.5829047, 49.5957647, 44.6212307, 39.656771, 34.6998544, 29.7479493, 24.7985243, 19.8490479, 14.8969887, 9.93981534, 4.97499624, 0, -5.25601836, -10.4341606, -15.5489985, -20.6151036, -25.6470477, -30.6594024, -35.6667394, -40.6836304, -45.724647, -50.804361, -55.9373439, -61.1381675, -66.4214035, -71.8016235, -77.2933992, -82.9113023, -88.5330794, -94.1537556, -100.097388, -106.349229, -112.89453, -119.718543, -126.80652, -134.143714, -141.715376, -149.506758, -157.503113, -165.689691, -174.051747, -182.632147, -191.337401, -200.135477, -208.99434, -217.980144, -227.011763, -235.981889, -244.829033, -253.491706, -261.908421, -270.017689, -280.381828, -289.516992, -296.848341, -301.801032, -303.800224, -294.679523, -283.353527, -271.14528, -259.377828, -249.374217, -251.490618, -246.976462, -224.610282, -178.968323, -116.463834, -38.1954897, 54.7380329, 151.651611, 263.211753, 410.542914, 611.973528, 857.643215, 1128.8723, 1426.38669, 1783.79404, 2219.60872, 2762.04974, 3432.1475, 4215.34882, 5117.22862, 6115.65302, 7248.1686, 8529.94178, 9976.13896, 11601.9266, 13981.388, 16431.0434, 18826.33, 21042.6848, 22955.5451, 23961.22, 24733.6932, 25353.7427, 26146.3433, 27095.6114, 28083.9145, 29100.4564, 30134.4408, 31175.0716, 32211.5524, 32323.019, 33124.517, 34344.5105, 35711.4637, 36953.8409, 37800.106, 37978.7233, 37874.244, 37718.9634},
        {360.792502, 360.843447, 360.82653, 360.740211, 360.582948, 360.3532, 360.049427, 359.670088, 359.213641, 358.678546, 358.063262, 357.366247, 356.585962, 355.720864, 354.769414, 353.73007, 352.601291, 351.381537, 350.069265, 348.662936, 347.161009, 345.561942, 343.864195, 342.066227, 340.166496, 338.163463, 336.055585, 333.841322, 331.519133, 329.087478, 326.544814, 323.889602, 321.1203, 318.235368, 315.070535, 311.709048, 308.23611, 304.655459, 300.970836, 297.185981, 293.304634, 289.330535, 285.267423, 281.119039, 276.889122, 272.581412, 268.19965, 263.747575, 259.228927, 254.647447, 250.006873, 245.381605, 240.703192, 235.973843, 231.195765, 226.371167, 221.502256, 216.591242, 211.640332, 206.651734, 201.627657, 196.570309, 191.481898, 186.364632, 181.220719, 176.052368, 170.861786, 165.640925, 160.38198, 155.108317, 149.822807, 144.528324, 139.227742, 133.923934, 128.619772, 123.31813, 118.021881, 112.733898, 107.457054, 102.194223, 96.9482776, 91.7220908, 86.5185359, 81.340486, 76.3617856, 71.4962259, 66.6555436, 61.8370634, 57.03811, 52.2560084, 47.4880833, 42.7316596, 37.984062, 33.2426154, 28.5046445, 23.7674743, 19.0284294, 14.2848347, 9.53401504, 4.7732952, 0, -4.65076717, -9.32236587, -14.0204746, -18.750772, -23.5189364, -28.3306464, -33.1915807, -38.1074175, -43.0838356, -48.1265134, -53.2411295, -58.4333623, -63.7088905, -69.0733924, -74.5325467, -80.0920319, -85.8605414, -91.8584701, -97.8584867, -103.888344, -109.975796, -116.148595, -122.434494, -128.861247, -135.456607, -142.248327, -149.264159, -156.531858, -164.079177, -172.331869, -180.773754, -189.313182, -197.858508, -206.330396, -214.631042, -222.65936, -230.320007, -237.517642, -244.156924, -250.142511, -256.100911, -260.995241, -264.588927, -266.645398, -266.928081, -265.938419, -262.51732, -256.243707, -246.696504, -233.454635, -219.353331, -196.735274, -159.969341, -106.125897, -37.7908421, 44.9262876, 141.915955, 248.023081, 369.330385, 517.424811, 695.748447, 913.88013, 1175.37669, 1488.23432, 1867.97143, 2326.6536, 2880.75053, 3543.46961, 4311.86984, 5192.14466, 6174.43355, 7283.45375, 8530.95582, 9928.69034, 11488.4079, 13727.0101, 16024.8086, 18267.2665, 20339.8463, 22128.011, 23067.9966, 23793.9772, 24383.9412, 25147.2212, 26070.232, 27042.9952, 28056.7452, 29102.7166, 30172.1439, 31256.2616, 32100.1369, 33200.6215, 34412.5838, 35590.892, 36590.4144, 37266.0193, 37472.5749, 37766.9618, 37821.1707},
        {339.029877, 339.693978, 340.245741, 340.685382, 341.013115, 341.229157, 341.333723, 341.327029, 341.209291, 340.980724, 340.641543, 340.191965, 339.632205, 338.962479, 338.183002, 337.29399, 336.295658, 335.188222, 333.971899, 332.646902, 331.213449, 329.671755, 328.022034, 326.264504, 324.399379, 322.426876, 320.347209, 318.160595, 315.867249, 313.467386, 310.961223, 308.348975, 305.630858, 302.807086, 299.794433, 296.636591, 293.378256, 290.02235, 286.571797, 283.029521, 279.398446, 275.681495, 271.881591, 268.001658, 264.04462, 260.013399, 255.910921, 251.740107, 247.503882, 243.20517, 238.846893, 234.460765, 230.020296, 225.527784, 220.985531, 216.395834, 211.760994, 207.083309, 202.36508, 197.608605, 192.816185, 187.990118, 183.132704, 178.246242, 173.333032, 168.395373, 163.435565, 158.44938, 153.432747, 148.401427, 143.358143, 138.305617, 133.246571, 128.183729, 123.119813, 118.057545, 112.999648, 107.948844, 102.907856, 97.8794072, 92.866219, 87.8710145, 82.8965161, 77.9454465, 73.1832504, 68.5278667, 63.8952783, 59.2829269, 54.6882547, 50.1087033, 45.5417147, 40.9847309, 36.4351936, 31.8905449, 27.3482265, 22.8056804, 18.2603484, 13.7096725, 9.15109454, 4.58205641, 0, -4.32648878, -8.70346019, -13.1339358, -17.6209373, -22.1674863, -26.7766043, -31.4513129, -36.1946339, -41.0095887, -45.899199, -50.8664864, -55.9144725, -61.046179, -66.2646273, -71.5728392, -76.9738362, -82.5463271, -88.3041459, -94.0791172, -99.8904812, -105.757478, -111.699349, -117.735333, -123.88467, -130.166602, -136.600368, -143.205209, -150.000365, -157.005077, -164.597445, -172.306267, -180.043124, -187.719596, -195.423027, -202.977116, -210.158697, -216.826624, -222.83975, -228.056929, -232.337015, -235.320653, -237.151316, -237.730551, -236.959905, -234.740923, -234.696456, -232.076421, -225.852039, -214.994533, -198.475121, -176.442064, -145.250943, -102.146083, -45.2093235, 25.76757, 111.760554, 213.745585, 332.70964, 469.615151, 625.412523, 788.730626, 986.965673, 1243.92425, 1573.87246, 1974.91062, 2452.5594, 3010.38112, 3653.38875, 4393.77586, 5239.67429, 6199.39345, 7280.86023, 8492.14497, 9841.318, 11336.4497, 13436.1334, 15585.2857, 17679.346, 19613.754, 21283.9493, 22164.181, 22851.8726, 23422.9739, 24186.1462, 25125.2241, 26127.0793, 27180.3946, 28273.8531, 29396.1376, 30535.931, 31640.72, 32839.7124, 34042.8606, 35160.1168, 36101.4332, 36776.7622, 37096.0561, 37675.7184, 37660.7597},
        {320.562039, 321.373916, 322.065352, 322.637071, 323.089797, 323.424255, 323.64117, 323.741266, 323.725268, 323.593902, 323.34789, 322.987959, 322.514833, 321.929236, 321.231893, 320.423529, 319.504868, 318.476636, 317.339556, 316.094354, 314.741754, 313.282481, 311.717259, 310.046813, 308.271869, 306.393149, 304.41138, 302.327286, 300.141591, 297.855021, 295.468299, 292.982151, 290.397301, 287.714474, 284.883483, 281.93158, 278.886627, 275.751002, 272.52708, 269.21724, 265.823857, 262.349309, 258.795973, 255.166225, 251.462441, 247.687, 243.842278, 239.930651, 235.954497, 231.916192, 227.818113, 223.660312, 219.447542, 215.18223, 210.866802, 206.503685, 202.095308, 197.644097, 193.152479, 188.622882, 184.057733, 179.459459, 174.830487, 170.173244, 165.490158, 160.783657, 156.056166, 151.310658, 146.550091, 141.775769, 136.990086, 132.195432, 127.394199, 122.58878, 117.781566, 112.974949, 108.171321, 103.373075, 98.5826013, 93.8022927, 89.0345408, 84.2817378, 79.5462754, 74.8305456, 70.2780535, 65.8176663, 61.3765551, 56.9525325, 52.543411, 48.1470031, 43.7611214, 39.3835784, 35.0121866, 30.6447586, 26.2791069, 21.9130439, 17.5443823, 13.1709346, 8.7905133, 4.40093092, 0, -4.31142498, -8.62794181, -12.9566032, -17.3044621, -21.6785711, -26.085983, -30.5337506, -35.0289268, -39.5785642, -44.1897156, -48.8694338, -53.6247716, -58.4627817, -63.390517, -68.4150302, -73.543374, -78.55184, -83.4145149, -88.6604118, -94.2471344, -100.132287, -106.273473, -112.628297, -119.154362, -125.809272, -132.550632, -139.336046, -146.123116, -152.869447, -159.447475, -165.9312, -172.303777, -178.54836, -185.24295, -192.073279, -198.566452, -204.527171, -209.760137, -214.070051, -217.261615, -219.289287, -219.762431, -218.456449, -215.14674, -209.608705, -201.433775, -190.627313, -177.01071, -160.40536, -140.632655, -121.478119, -93.953965, -53.7854169, 2.01037675, 73.7785884, 163.048654, 271.350011, 403.191777, 556.446506, 725.736193, 892.34846, 1090.97201, 1354.40971, 1698.98578, 2113.10232, 2597.97814, 3147.94887, 3762.44878, 4466.15049, 5269.45039, 6199.8184, 7247.95069, 8418.33356, 9715.45333, 11143.7963, 13099.3981, 15097.3084, 17044.1262, 18846.4507, 20410.8808, 21256.6546, 21935.9727, 22521.4458, 23325.9075, 24326.8805, 25401.7945, 26533.1769, 27703.5547, 28895.4555, 30091.4063, 30928.6387, 32008.6847, 33213.0104, 34423.082, 35520.3657, 36386.3275, 36902.4336, 37570.2687, 37278.7334},
        {302.36832, 303.249171, 304.014501, 304.664979, 305.201277, 305.624062, 305.934005, 306.131775, 306.218043, 306.193476, 306.058747, 305.814522, 305.461474, 305.00027, 304.431581, 303.756077, 302.974426, 302.087299, 301.095365, 299.999294, 298.799756, 297.497419, 296.092955, 294.587031, 292.980318, 291.273486, 289.467205, 287.562143, 285.55897, 283.458356, 281.260971, 278.967485, 276.578566, 274.094885, 271.472901, 268.73632, 265.909376, 262.994173, 259.992817, 256.907412, 253.740061, 250.492869, 247.167941, 243.76738, 240.293291, 236.747779, 233.132947, 229.450901, 225.703743, 221.893579, 218.022513, 214.070387, 210.062049, 206.000088, 201.887089, 197.725641, 193.518329, 189.267741, 184.976463, 180.647083, 176.282187, 171.884362, 167.456196, 163.000274, 158.519185, 154.015515, 149.491851, 144.958314, 140.424845, 135.878492, 131.321352, 126.755525, 122.183109, 117.606201, 113.026902, 108.447308, 103.86952, 99.295634, 94.7277499, 90.1679659, 85.6183803, 81.0810918, 76.5581987, 72.0517995, 67.686789, 63.4012847, 59.131928, 54.8768326, 50.6341118, 46.4018792, 42.1782483, 37.9613326, 33.7492455, 29.5401006, 25.3320113, 21.1230912, 16.9114537, 12.6952123, 8.47248061, 4.241372, 0, -3.82384404, -7.71757871, -11.6828856, -15.7214464, -19.8349426, -24.0250559, -28.2934678, -32.64186, -37.071914, -41.5853115, -46.1837341, -50.8688634, -55.6423809, -60.5059684, -65.4613073, -70.5100793, -75.4871822, -80.3704714, -85.5388775, -90.9583427, -96.5948092, -102.414219, -108.382515, -114.465639, -120.629533, -126.840139, -133.063401, -139.265259, -145.411656, -151.40652, -157.300546, -163.078281, -168.724272, -174.874622, -181.188098, -187.149722, -192.548572, -197.173728, -200.81427, -203.259277, -205.086588, -205.056467, -202.80367, -197.962956, -190.16908, -175.699899, -158.386296, -138.702255, -117.121758, -94.1187871, -76.1838378, -50.4208642, -10.9392031, 46.0596867, 120.102931, 212.635435, 325.102103, 460.044156, 617.561019, 796.556133, 986.711723, 1210.91963, 1490.89451, 1838.10791, 2246.67503, 2718.67788, 3249.46601, 3839.37597, 4513.43025, 5282.68758, 6186.5929, 7203.45163, 8334.49652, 9580.9603, 10944.0757, 12756.1243, 14604.5279, 16407.7568, 18084.2818, 19552.5732, 20387.0567, 21079.6112, 21696.1548, 22527.1901, 23549.7376, 24647.2411, 25801.4002, 26993.9141, 28206.4822, 29420.804, 30418.9364, 31535.6003, 32698.9373, 33837.089, 34878.1967, 35750.4021, 36381.8467, 37034.457, 36935.7498},
        {284.82981, 285.73221, 286.532391, 287.230592, 287.82705, 288.322003, 288.715688, 289.008342, 289.200204, 289.291512, 289.282502, 289.173412, 288.96448, 288.655944, 288.248041, 287.741009, 287.135085, 286.430508, 285.627514, 284.726341, 283.727228, 282.630411, 281.436128, 280.144617, 278.756116, 277.270862, 275.689092, 274.011045, 272.236957, 270.367068, 268.401613, 266.340832, 264.184961, 261.934237, 259.533192, 257.011086, 254.397855, 251.695544, 248.906197, 246.031861, 243.074581, 240.036403, 236.919371, 233.725532, 230.45693, 227.115612, 223.703622, 220.223007, 216.675811, 213.064081, 209.389862, 205.623645, 201.799715, 197.920799, 193.989628, 190.008931, 185.981438, 181.909878, 177.79698, 173.645474, 169.45809, 165.237556, 160.986603, 156.70796, 152.404356, 148.078521, 143.733184, 139.383992, 135.046282, 130.696141, 126.335463, 121.96614, 117.590063, 113.209124, 108.825217, 104.440232, 100.056063, 95.6746005, 91.2977374, 86.9273657, 82.5653775, 78.2136648, 73.8741199, 69.5486348, 65.3497832, 61.2217883, 57.1075432, 53.0053482, 48.9135036, 44.8303096, 40.7540665, 36.6830747, 32.6156343, 28.5500458, 24.4846093, 20.4176252, 16.3473937, 12.2722152, 8.19038984, 4.10021802, 0, -2.99406384, -6.20668706, -9.62680512, -13.2433535, -17.0452675, -21.0214827, -25.1609346, -29.4525585, -33.8852899, -38.4480642, -43.1298169, -47.9194834, -52.8059993, -57.7782998, -62.8253205, -67.9359968, -73.3104939, -78.9679228, -84.4194376, -89.6992372, -94.8415205, -99.8804864, -104.850334, -109.785262, -114.71947, -119.687156, -124.72252, -129.85976, -135.133076, -140.957979, -146.847541, -152.721566, -158.499859, -164.481527, -170.396723, -175.874456, -180.720739, -184.741587, -187.743015, -189.531036, -191.222764, -190.914086, -188.154495, -182.493489, -173.480561, -156.462854, -136.242806, -113.420497, -88.5960135, -62.369437, -41.5847502, -12.9667059, 29.9804601, 91.1451798, 169.092583, 264.777854, 379.156177, 509.037041, 660.463528, 844.001295, 1068.24358, 1331.99212, 1632.9311, 1973.36255, 2363.41322, 2809.61817, 3315.8541, 3887.96686, 4541.54973, 5286.6823, 6166.25421, 7152.4873, 8244.10382, 9439.82602, 10738.3762, 12413.1115, 14119.4605, 15787.4868, 17347.2539, 18728.8252, 19566.1448, 20282.8088, 20935.789, 21771.8307, 22772.7873, 23841.4397, 24963.5946, 26125.0591, 27311.6398, 28509.1437, 30029.2946, 31320.1601, 32416.4109, 33352.7175, 34163.7505, 34884.1804, 35548.6779, 36119.821, 36619.5013},
        {271.159829, 271.873559, 272.508088, 273.062969, 273.537754, 273.931996, 274.245246, 274.477057, 274.626982, 274.694573, 274.679382, 274.580961, 274.398864, 274.132642, 273.781848, 273.346035, 272.824754, 272.217557, 271.523998, 270.743629, 269.876002, 268.92067, 267.877184, 266.745098, 265.523963, 264.213332, 262.812758, 261.321793, 259.739988, 258.066897, 256.302072, 254.445066, 252.49543, 250.452717, 248.241821, 245.901194, 243.470185, 240.95077, 238.344923, 235.65462, 232.881836, 230.028547, 227.096726, 224.088351, 221.005395, 217.849834, 214.623643, 211.328797, 207.967272, 204.541043, 201.052085, 197.46666, 193.823231, 190.124546, 186.373356, 182.572409, 178.724455, 174.832244, 170.898525, 166.926046, 162.917559, 158.875811, 154.803552, 150.703532, 146.5785, 142.431205, 138.264397, 134.096196, 129.944354, 125.779917, 121.60464, 117.420275, 113.228575, 109.031293, 104.830183, 100.626996, 96.4234865, 92.2214067, 88.0225098, 83.8285486, 79.6412762, 75.4624455, 71.2938093, 67.1371208, 63.0951885, 59.1171119, 55.1507761, 51.1946543, 47.2472202, 43.3069473, 39.372309, 35.4417788, 31.5138303, 27.5869369, 23.6595721, 19.7302094, 15.7973223, 11.8593844, 7.91486899, 3.9622497, 0, -2.72859858, -5.70755629, -8.92228881, -12.3582118, -16.000741, -19.835292, -23.8472806, -28.0221223, -32.345233, -36.8020282, -41.3779236, -46.058335, -50.828678, -55.6743682, -60.5808215, -65.5334533, -70.8341451, -76.513522, -81.8411844, -86.870362, -91.6542847, -96.2461822, -100.699284, -105.06682, -109.402021, -113.758115, -118.188332, -122.745903, -127.484057, -132.998053, -138.600347, -144.181561, -149.632315, -155.127932, -160.416681, -165.170914, -169.195841, -172.296675, -174.278625, -174.946902, -175.32181, -173.623657, -169.419919, -162.27807, -151.765587, -134.596188, -113.904545, -89.9715736, -63.078188, -33.5053035, -6.46766301, 28.7178806, 77.3995872, 142.845756, 224.078097, 322.024955, 437.614672, 565.350615, 714.046323, 899.524399, 1139.08259, 1420.60042, 1731.5132, 2069.95214, 2452.17203, 2886.10865, 3379.06684, 3938.81878, 4575.45015, 5297.73797, 6145.62955, 7091.93623, 8134.64532, 9271.74417, 10501.2201, 12058.1911, 13644.2312, 15198.0449, 16658.337, 17963.8121, 18792.6462, 19517.7583, 20189.5083, 21025.1645, 22009.0186, 23055.8174, 24153.33, 25289.3255, 26451.573, 27627.8416, 29221.1356, 30526.3144, 31601.2042, 32503.631, 33291.4208, 34022.3996, 34754.3937, 35206.6146, 36089.3839},
        {261.977625, 262.28781, 262.548696, 262.75904, 262.917598, 263.023127, 263.074383, 263.070123, 263.009104, 262.890082, 262.711813, 262.473055, 262.172563, 261.809094, 261.381406, 260.888254, 260.328394, 259.700584, 259.003581, 258.23614, 257.397018, 256.484971, 255.498757, 254.437132, 253.298852, 252.082674, 250.787355, 249.41165, 247.954317, 246.414113, 244.789793, 243.080114, 241.283833, 239.399706, 237.332195, 235.129187, 232.839703, 230.46556, 228.008574, 225.470562, 222.853341, 220.158727, 217.388537, 214.544588, 211.628695, 208.642677, 205.588349, 202.467528, 199.282031, 196.033674, 192.724274, 189.318822, 185.856759, 182.340699, 178.773258, 175.15705, 171.49469, 167.788793, 164.041974, 160.256848, 156.43603, 152.582136, 148.697778, 144.785574, 140.848137, 136.888082, 132.908026, 128.925093, 124.956061, 120.973618, 116.979437, 112.975192, 108.962558, 104.943209, 100.918818, 96.8910613, 92.8616116, 88.8321437, 84.8043315, 80.7798495, 76.7603717, 72.7475725, 68.7431259, 64.7487062, 60.8590452, 57.0273309, 53.2056331, 49.3926059, 45.5869037, 41.7871809, 37.9920917, 34.2002905, 30.4104315, 26.6211692, 22.8311578, 19.0390516, 15.2435049, 11.4431721, 7.63670747, 3.82276533, 0, -3.21987607, -6.5612656, -10.0190228, -13.5880018, -17.2630569, -21.0390422, -24.910812, -28.8732205, -32.9211217, -37.0493701, -41.2528197, -45.5263247, -49.8647394, -54.2629179, -58.7157145, -63.2179834, -67.7690978, -72.3645577, -76.9889952, -81.638233, -86.3080937, -90.9943998, -95.692974, -100.399639, -105.110217, -109.820531, -114.526403, -119.223656, -123.908113, -128.818979, -133.619454, -138.232344, -142.580459, -147.076773, -151.399011, -155.094185, -157.938053, -159.70637, -160.174896, -159.119387, -156.735711, -152.251655, -145.360782, -135.756652, -123.132828, -107.643269, -88.4060399, -64.9996028, -37.0024199, -3.99295344, 32.1251141, 76.9404765, 132.864532, 201.917823, 285.32289, 384.660562, 501.511666, 633.545457, 787.143329, 972.953848, 1200.7198, 1467.94321, 1773.24022, 2118.26043, 2509.79342, 2952.26944, 3447.74782, 4000.04413, 4621.66727, 5320.20872, 6127.67706, 7023.49014, 8006.78737, 9076.70813, 10232.3918, 11694.7475, 13185.7025, 14648.9537, 16028.1981, 17267.1326, 18067.205, 18775.8609, 19440.6184, 20274.0463, 21258.3301, 22306.8841, 23405.3321, 24539.2981, 25694.4063, 26856.2805, 27873.3607, 28962.1199, 30083.1091, 31196.8794, 32263.9818, 33244.9674, 34100.3873, 34432.8854, 35306.2598},
        {254.546663, 254.417866, 254.2697, 254.100142, 253.907168, 253.688756, 253.442882, 253.167522, 252.860653, 252.520252, 252.144296, 251.73076, 251.277623, 250.78286, 250.244448, 249.660364, 249.028584, 248.347085, 247.613844, 246.826838, 245.984043, 245.083435, 244.122992, 243.100689, 242.014505, 240.862415, 239.642396, 238.352424, 236.990477, 235.554531, 234.042562, 232.452548, 230.782465, 229.030289, 227.079813, 224.988507, 222.815214, 220.561617, 218.229398, 215.820239, 213.335822, 210.777829, 208.147943, 205.447846, 202.679219, 199.843746, 196.943108, 193.978988, 190.953067, 187.867028, 184.722553, 181.487614, 178.198335, 174.857127, 171.466405, 168.028581, 164.546067, 161.021278, 157.456625, 153.854522, 150.217382, 146.547617, 142.847641, 139.119867, 135.366707, 131.590574, 127.793882, 123.991378, 120.197513, 116.389261, 112.568236, 108.73605, 104.894318, 101.044652, 97.1886656, 93.3279727, 89.4641864, 85.59892, 81.7337869, 77.8704004, 74.0103739, 70.1553207, 66.3068542, 62.4665877, 58.722309, 55.0307317, 51.3475336, 47.6715315, 44.0015423, 40.3363827, 36.6748697, 33.0158201, 29.3580506, 25.7003781, 22.0416194, 18.3805914, 14.7161109, 11.0469946, 7.37205951, 3.69012236, 0, -3.73327502, -7.45494862, -11.1698531, -14.8828207, -18.5986837, -22.3222745, -26.0584252, -29.8119683, -33.5877359, -37.3905604, -41.225274, -45.096709, -49.0096977, -52.9690724, -56.9796654, -61.046309, -64.832023, -68.2928094, -72.2066475, -76.5051242, -81.1198262, -85.9823402, -91.0242531, -96.1771516, -101.372623, -106.542253, -111.617629, -116.530337, -121.211965, -125.505708, -129.463954, -133.044807, -136.206373, -139.578997, -142.784663, -145.26609, -146.77971, -147.081955, -145.929256, -143.078046, -138.004023, -130.829793, -121.366713, -109.426142, -94.8194372, -81.0026274, -63.2312316, -40.4054399, -11.4254423, 24.8085712, 69.541449, 123.550693, 186.477627, 259.174058, 344.963198, 446.058649, 564.674014, 702.36496, 862.152554, 1047.77561, 1258.57565, 1505.17653, 1800.59073, 2151.30879, 2552.77008, 3005.48663, 3505.69874, 4052.81095, 4661.89085, 5339.14611, 6106.60511, 6952.22088, 7876.72473, 8880.84793, 9965.32178, 11344.3933, 12751.8991, 14135.1916, 15441.6231, 16618.5459, 17383.4866, 18066.8406, 18714.4572, 19539.827, 20522.441, 21571.9398, 22671.307, 23803.5263, 24951.5811, 26098.4551, 26404.5058, 27237.9286, 28430.4348, 29813.7352, 31219.5408, 32479.5628, 33425.5122, 33684.5299, 34483.6831},
        {245.454061, 245.115745, 244.774766, 244.428673, 244.075011, 243.711325, 243.335163, 242.94407, 242.535593, 242.107277, 241.656669, 241.181315, 240.67876, 240.146552, 239.582236, 238.983358, 238.347465, 237.672102, 236.954816, 236.193153, 235.384659, 234.52688, 233.617362, 232.653651, 231.633294, 230.553837, 229.412825, 228.207805, 226.936323, 225.595926, 224.184158, 222.698567, 221.136698, 219.496098, 217.646404, 215.651806, 213.578034, 211.426783, 209.199752, 206.898638, 204.525136, 202.080945, 199.567761, 196.987281, 194.341203, 191.631223, 188.859039, 186.026347, 183.134844, 180.186228, 177.182196, 174.101049, 170.968387, 167.786413, 164.557333, 161.28335, 157.966669, 154.609493, 151.214028, 147.782476, 144.317042, 140.819931, 137.293347, 133.739494, 130.160575, 126.558796, 122.93636, 119.306203, 115.681003, 112.040832, 108.387198, 104.721609, 101.045574, 97.3606017, 93.6682, 89.9698777, 86.2671434, 82.5615054, 78.8544724, 75.1475527, 71.442255, 67.7400877, 64.0425594, 60.3511785, 56.7461161, 53.1878951, 49.6360923, 46.0896637, 42.5475649, 39.0087517, 35.4721799, 31.9368051, 28.4015832, 24.8654699, 21.3274209, 17.786392, 14.2413389, 10.6912174, 7.1349833, 3.57159223, 0, -3.21340602, -6.49620618, -9.84694962, -13.2641855, -16.7464629, -20.2923311, -23.9003391, -27.5690361, -31.2969713, -35.0826938, -38.9247527, -42.8216973, -46.7720765, -50.7744397, -54.8273358, -58.9293142, -62.8205934, -66.4628181, -70.4388566, -74.6919015, -79.1651453, -83.8017807, -88.5450003, -93.3379967, -98.1239625, -102.84609, -107.447573, -111.871602, -116.061372, -120.009569, -123.591743, -126.736238, -129.371398, -131.821558, -133.817066, -134.982674, -135.127926, -134.062369, -131.59555, -127.537015, -122.112696, -114.589027, -104.694089, -92.1559593, -76.7027189, -61.0939312, -41.2683203, -16.1960945, 15.1525376, 53.8073676, 100.626596, 157.021328, 223.425837, 300.90543, 391.813772, 497.926083, 621.017581, 761.517504, 922.852957, 1109.91939, 1323.72076, 1572.29978, 1865.27348, 2208.09825, 2599.1805, 3040.16268, 3530.73812, 4072.04394, 4672.36412, 5335.93999, 6077.19464, 6888.62117, 7770.93643, 8724.85727, 9751.10054, 11042.9128, 12360.3487, 13655.9928, 14882.4295, 15992.2431, 16727.5389, 17391.7001, 18027.5162, 18835.945, 19797.1682, 20825.4641, 21904.3097, 23017.1817, 24147.557, 25278.9126, 25246.1418, 25991.345, 27262.2506, 28806.5872, 30372.0831, 31706.4668, 32557.4668, 32918.2881, 33927.6811},
        {236.525754, 236.055766, 235.594544, 235.139348, 234.687442, 234.236088, 233.782548, 233.324084, 232.85796, 232.381437, 231.891779, 231.386247, 230.862103, 230.316611, 229.747032, 229.15063, 228.524666, 227.866403, 227.173104, 226.44203, 225.670444, 224.855609, 223.994787, 223.085241, 222.124232, 221.109024, 220.036879, 218.905058, 217.710826, 216.451443, 215.124173, 213.726277, 212.255019, 210.707661, 208.942914, 207.030231, 205.040728, 202.976165, 200.8383, 198.628892, 196.349698, 194.002479, 191.588993, 189.110997, 186.570251, 183.968514, 181.307545, 178.589101, 175.814941, 172.986825, 170.10651, 167.164725, 164.174499, 161.137828, 158.056712, 154.933146, 151.76913, 148.566662, 145.327737, 142.054356, 138.748515, 135.412212, 132.047444, 128.656211, 125.240509, 121.802336, 118.34369, 114.876006, 111.410492, 107.929683, 104.434965, 100.927723, 97.4093439, 93.8812131, 90.3447167, 86.8012408, 83.2521714, 79.6988944, 76.1427958, 72.5852615, 69.0276777, 65.4714303, 61.9179052, 58.3684884, 54.8962752, 51.4652885, 48.0386896, 44.6155375, 41.1948908, 37.7758087, 34.3573498, 30.9385732, 27.5185377, 24.0963022, 20.6709256, 17.2414668, 13.8069846, 10.3665379, 6.91918568, 3.46398674, 0, -2.40971064, -5.03199433, -7.85427733, -10.8639859, -14.0485463, -17.3953849, -20.8919278, -24.5256014, -28.283832, -32.1540457, -36.1236689, -40.1801278, -44.3108488, -48.503258, -52.7447817, -57.0228463, -61.2873736, -65.5204321, -69.7942795, -74.0883052, -78.3818989, -82.6544504, -86.885349, -91.0539846, -95.1397467, -99.122025, -102.980209, -106.693688, -110.241853, -113.942671, -117.312807, -120.230078, -122.5723, -124.197526, -124.993454, -124.85305, -123.660058, -121.298223, -117.651289, -112.603, -107.594934, -100.478878, -90.8337846, -78.2386034, -62.2722855, -43.5746133, -20.3984983, 7.94231661, 42.1340884, 82.863074, 129.352576, 185.539861, 253.727562, 335.648468, 431.871942, 543.489703, 671.593468, 814.31568, 976.379895, 1165.73797, 1388.11966, 1644.82411, 1937.18836, 2268.04601, 2642.76653, 3065.55538, 3542.0729, 4076.90173, 4669.29004, 5321.50348, 6041.64874, 6825.25182, 7672.55647, 8583.80644, 9559.24549, 10769.6036, 12002.0167, 13214.107, 14363.4969, 15407.8083, 16112.8883, 16755.9846, 17376.9089, 18161.676, 19092.2255, 20089.5791, 21138.7222, 22224.6404, 23332.3191, 24446.7439, 24242.9954, 24966.4697, 26306.5988, 27952.815, 29594.5507, 30921.238, 31622.3093, 32159.6905, 33438.871},
        {228.252986, 227.683678, 227.131822, 226.594477, 226.068701, 225.551555, 225.040099, 224.531391, 224.022492, 223.510461, 222.992357, 222.46524, 221.92617, 221.372206, 220.800408, 220.207834, 219.591546, 218.948602, 218.276062, 217.570986, 216.830432, 216.051461, 215.231132, 214.366505, 213.454639, 212.492594, 211.47743, 210.406205, 209.275979, 208.083813, 206.826765, 205.501896, 204.106264, 202.636929, 200.94418, 199.101549, 197.184333, 195.194353, 193.133433, 191.003394, 188.80606, 186.543253, 184.216795, 181.82851, 179.380219, 176.873745, 174.310911, 171.693539, 169.023452, 166.302472, 163.532422, 160.715429, 157.853004, 154.946963, 151.999123, 149.011299, 145.985308, 142.922964, 139.826085, 136.696485, 133.535982, 130.346391, 127.129528, 123.887209, 120.621249, 117.333466, 114.025674, 110.708426, 107.392063, 104.060385, 100.714642, 97.3560818, 93.9859555, 90.6055125, 87.2160022, 83.8186744, 80.4147788, 77.0055649, 73.5922824, 70.1761809, 66.7585102, 63.3405198, 59.9234595, 56.5085787, 53.1628813, 49.8533567, 46.5462043, 43.2405399, 39.9354793, 36.6301383, 33.3236327, 30.0150783, 26.7035908, 23.3882862, 20.0682802, 16.7426886, 13.4106273, 10.0712119, 6.72355838, 3.36678247, 0, -1.61181554, -3.58398196, -5.89219422, -8.51214733, -11.4195362, -14.590056, -17.9994015, -21.6232678, -25.4373498, -29.4173426, -33.538941, -37.7778402, -42.1097351, -46.5103206, -50.9552918, -55.4203436, -60.1021433, -65.0079533, -69.6136504, -73.9422808, -78.0168907, -81.8605262, -85.4962335, -88.9470588, -92.2360483, -95.3862482, -98.4207046, -101.362464, -104.234572, -107.729698, -110.955737, -113.734848, -115.889191, -116.792693, -116.49163, -115.151806, -112.729847, -109.182384, -104.466045, -98.5374594, -94.118361, -87.5587187, -78.2741621, -65.6803209, -49.1928247, -26.8439902, 0.221411777, 32.2419233, 69.4560864, 112.102443, 157.366193, 212.271084, 281.074922, 366.272222, 466.757449, 583.041417, 715.63494, 860.606148, 1023.91824, 1216.38098, 1448.60452, 1714.69028, 2007.07763, 2325.74937, 2683.47025, 3087.13787, 3548.36556, 4075.27354, 4658.69105, 5299.2281, 6000.27959, 6759.24217, 7575.76182, 8449.48455, 9380.05635, 10518.1543, 11674.6355, 12811.3882, 13890.3007, 14873.2612, 15546.0603, 16164.0822, 16764.6855, 17519.9036, 18413.7864, 19374.2692, 20388.2074, 21442.4565, 22523.8722, 23619.31, 23389.3032, 24135.0405, 25518.5274, 27201.7697, 28846.773, 30115.543, 30670.0853, 31416.7418, 32946.6325},
        {221.842616, 221.100933, 220.388508, 219.702128, 219.038581, 218.394653, 217.767131, 217.152802, 216.548452, 215.950869, 215.356839, 214.76315, 214.166588, 213.56394, 212.951992, 212.327533, 211.687348, 211.028225, 210.34695, 209.64031, 208.905093, 208.138084, 207.336071, 206.495841, 205.614181, 204.687877, 203.713716, 202.688485, 201.608972, 200.471962, 199.274243, 198.012602, 196.683826, 195.2847, 193.65883, 191.882816, 190.035098, 188.117435, 186.131586, 184.079308, 181.962358, 179.782496, 177.541478, 175.241064, 172.88301, 170.469074, 168.001016, 165.480592, 162.909561, 160.289681, 157.622709, 154.910265, 152.154249, 149.356422, 146.518543, 143.642376, 140.729679, 137.782215, 134.801743, 131.790026, 128.748824, 125.679899, 122.58501, 119.465919, 116.324387, 113.162175, 109.981044, 106.793213, 103.610652, 100.41355, 97.2029915, 93.9800581, 90.7458333, 87.5014001, 84.2478412, 80.9862398, 77.7176788, 74.4432411, 71.1640096, 67.8810674, 64.5954973, 61.3083824, 58.0208056, 54.7338497, 51.5116376, 48.3224063, 45.1336352, 41.9443617, 38.7536228, 35.5604558, 32.3638979, 29.1629862, 25.956758, 22.7442505, 19.5245007, 16.2965461, 13.0594236, 9.81217058, 6.55382418, 3.28342159, 0, -1.6042331, -3.57182793, -5.87669252, -8.49273494, -11.3938632, -14.5539854, -17.9470095, -21.5468437, -25.3273958, -29.2625741, -33.3262865, -37.492441, -41.7349458, -46.0277089, -50.3446382, -54.6596419, -59.2497156, -64.1320655, -68.5950445, -72.677508, -76.4183112, -79.8563097, -83.0303587, -85.9793137, -88.7420301, -91.3573632, -93.8641684, -96.3013012, -98.7076168, -101.937649, -104.915492, -107.4353, -109.291223, -109.715226, -108.782555, -106.718372, -103.485485, -99.0467031, -93.3648331, -86.4026836, -81.1392031, -73.6031035, -63.1670786, -49.203822, -31.0860273, -5.87688592, 24.1640305, 59.0866527, 98.9409112, 143.776737, 190.275937, 245.973161, 314.473766, 398.218656, 497.271303, 612.762599, 745.823433, 894.868252, 1064.36177, 1261.73208, 1494.71083, 1759.59798, 2050.72193, 2368.5299, 2723.83706, 3122.69953, 3575.32296, 4088.83926, 4655.16537, 5274.82469, 5950.59688, 6680.40194, 7463.98223, 8301.08013, 9191.43801, 10280.9428, 11386.6562, 12471.7844, 13499.5336, 14433.1101, 15063.1215, 15640.4384, 16202.2376, 16918.1731, 17772.9831, 18696.2073, 19675.344, 20697.8914, 21751.3479, 22823.2116, 22753.2795, 23537.1447, 24875.7627, 26470.0888, 28021.0785, 29229.687, 29796.8699, 30710.7855, 32266.5873},
        {216.397646, 215.479903, 214.602641, 213.762397, 212.955709, 212.179114, 211.42915, 210.702354, 209.995264, 209.304417, 208.626351, 207.957604, 207.294712, 206.634213, 205.972645, 205.306545, 204.632451, 203.946901, 203.246431, 202.527579, 201.786883, 201.020881, 200.226109, 199.399106, 198.536408, 197.634554, 196.69008, 195.699525, 194.659425, 193.566319, 192.416743, 191.207236, 189.934334, 188.594576, 187.026555, 185.309103, 183.522949, 181.669757, 179.751189, 177.768908, 175.724579, 173.619863, 171.456424, 169.235925, 166.960028, 164.630398, 162.248697, 159.816588, 157.335734, 154.807799, 152.234444, 149.613839, 146.951218, 144.248318, 141.506879, 138.72864, 135.91534, 133.068717, 130.190511, 127.282459, 124.346302, 121.383777, 118.396624, 115.386582, 112.355389, 109.304784, 106.236506, 103.165005, 100.104427, 97.0302933, 93.9435194, 90.8450199, 87.7357097, 84.6165036, 81.4883164, 78.3520628, 75.2086578, 72.0590161, 68.9040525, 65.7446819, 62.581819, 59.4163787, 56.2492757, 53.081425, 49.9748777, 46.8986945, 43.8212013, 40.7413289, 37.658008, 34.5701696, 31.4767444, 28.3766633, 25.268857, 22.1522565, 19.0257924, 15.8883956, 12.7389969, 9.57652725, 6.39991731, 3.20809796, 0, -1.8275074, -3.97867852, -6.42864893, -9.15255421, -12.1255299, -15.3227117, -18.719235, -22.2902355, -26.0108487, -29.8562103, -33.8014557, -37.8217206, -41.8921406, -45.9878511, -50.0839879, -54.1556864, -58.5041036, -63.1509494, -67.3339304, -71.0980438, -74.4882868, -77.5496568, -80.327151, -82.8657667, -85.2105013, -87.4063519, -89.498316, -91.5313907, -93.5505735, -96.4942558, -99.1864628, -101.404173, -102.924367, -102.943889, -101.529787, -98.903806, -95.0169647, -89.8202818, -83.2647756, -75.3014648, -68.9124994, -60.0952481, -48.2076817, -32.6077714, -12.6534881, 15.1573234, 47.8915342, 85.4761419, 127.838144, 174.904538, 223.256272, 280.256008, 348.166593, 429.157408, 525.207001, 638.379601, 770.739436, 924.132425, 1100.89072, 1303.58463, 1535.33176, 1796.6876, 2086.19616, 2405.48512, 2761.40729, 3158.41705, 3603.63203, 4102.19708, 4649.49187, 5246.4198, 5896.6442, 6597.88396, 7350.08714, 8153.20177, 9007.17592, 10057.0625, 11121.4285, 12163.9458, 13148.286, 14038.1212, 14624.8121, 15159.2155, 15678.8506, 16353.2726, 17167.7365, 18052.4829, 18995.5177, 19984.8469, 21008.4765, 22054.4126, 22254.5193, 23096.9992, 24353.8744, 25797.1669, 27198.899, 28331.0927, 28965.7701, 30025.7115, 31528.8179},
        {211.768591, 210.678377, 209.638974, 208.6467, 207.697872, 206.78881, 205.915829, 205.07525, 204.263389, 203.476565, 202.711095, 201.963298, 201.229492, 200.505993, 199.789122, 199.075194, 198.360529, 197.641445, 196.914258, 196.175288, 195.420852, 194.647269, 193.850855, 193.02793, 192.174811, 191.287816, 190.363263, 189.39747, 188.386755, 187.327436, 186.21583, 185.048257, 183.821034, 182.530478, 181.011539, 179.344613, 177.612036, 175.815364, 173.956151, 172.035951, 170.056321, 168.018813, 165.924984, 163.776387, 161.574577, 159.32111, 157.01754, 154.665421, 152.266308, 149.821756, 147.333321, 144.794683, 142.215441, 139.597321, 136.942047, 134.251344, 131.52694, 128.770558, 125.983924, 123.168763, 120.326801, 117.459764, 114.569376, 111.657362, 108.72545, 105.775363, 102.808826, 99.8424896, 96.892642, 93.9302317, 90.9560166, 87.9707549, 84.9752044, 81.9701233, 78.9562693, 75.9344007, 72.9052752, 69.869651, 66.828286, 63.7819381, 60.7313655, 57.677326, 54.6205776, 51.5618784, 48.561261, 45.5885992, 42.6130546, 39.6334615, 36.6486543, 33.6574676, 30.6587356, 27.6512929, 24.6339737, 21.6056126, 18.5650439, 15.5111021, 12.4426215, 9.35843659, 6.25738176, 3.13829143, 0, -2.19970179, -4.65338622, -7.33971266, -10.2373405, -13.324929, -16.5811377, -19.9846258, -23.5140528, -27.148078, -30.8653608, -34.6445605, -38.4643365, -42.3033483, -46.1402551, -49.9537163, -53.7223913, -57.7142186, -61.9491831, -65.7516224, -69.1621843, -72.2215164, -74.9702665, -77.4490824, -79.6986119, -81.7595027, -83.6724025, -85.4779591, -87.2168202, -88.9296337, -91.5543861, -93.9053621, -95.7540077, -96.8717692, -96.503427, -94.6837601, -91.5879915, -87.145567, -81.2859323, -73.9385332, -65.0328156, -57.4094665, -47.2006604, -33.7662525, -16.4660977, 5.33994868, 35.3561361, 70.3924789, 110.323096, 155.022106, 204.363628, 255.188928, 314.111797, 381.641402, 459.519668, 552.005405, 662.227395, 793.314415, 950.969626, 1135.16234, 1343.05347, 1572.48259, 1829.08604, 2116.54782, 2437.97545, 2795.49958, 3191.69928, 3629.97554, 4113.12053, 4640.91255, 5214.83473, 5840.14363, 6513.96838, 7236.48591, 8007.87314, 8828.30699, 9843.22082, 10871.221, 11876.1703, 12821.9317, 13672.3678, 14218.6361, 14712.4418, 15191.6643, 15825.8121, 16600.6616, 17447.1434, 18353.7759, 19309.0776, 20301.5669, 21319.7623, 21865.4231, 22779.9902, 23929.1983, 25178.7823, 26394.477, 27442.017, 28187.1373, 29358.7174, 30751.677},
        {207.691201, 206.451616, 205.271273, 204.146323, 203.072918, 202.047211, 201.065355, 200.123501, 199.217802, 198.344412, 197.499481, 196.679163, 195.879609, 195.096973, 194.327407, 193.567063, 192.812093, 192.05865, 191.302887, 190.540955, 189.769007, 188.983196, 188.179675, 187.354594, 186.504108, 185.624367, 184.711526, 183.761735, 182.771148, 181.735917, 180.652194, 179.516132, 178.323883, 177.071599, 175.591183, 173.964366, 172.274856, 170.524136, 168.713687, 166.844993, 164.919536, 162.938799, 160.904264, 158.817415, 156.679734, 154.492703, 152.257806, 149.976524, 147.650341, 145.28074, 142.869202, 140.409701, 137.911393, 135.375922, 132.804934, 130.200075, 127.562989, 124.895323, 122.198722, 119.474831, 116.725296, 113.951762, 111.155874, 108.339279, 105.503622, 102.650548, 99.781702, 96.914076, 94.0642923, 91.2023473, 88.3288914, 85.4445755, 82.5500501, 79.6459659, 76.7329735, 73.8117237, 70.8828672, 67.9470545, 65.0049363, 62.0571633, 59.1043862, 56.1472557, 53.1864223, 50.2225368, 47.3111829, 44.4243892, 41.5335243, 38.6374561, 35.7350525, 32.8251816, 29.9067114, 26.9785097, 24.0394447, 21.0883843, 18.1241964, 15.1457491, 12.1519103, 9.14154798, 6.11353019, 3.06672487, 0, -2.66601926, -5.4849634, -8.4422349, -11.5232362, -14.7133699, -17.9980384, -21.3626441, -24.7925897, -28.2732775, -31.79011, -35.3284897, -38.8738192, -42.4115008, -45.9269371, -49.4055306, -52.8326837, -56.3094272, -59.8376819, -63.1413101, -66.2304919, -69.115407, -71.8062356, -74.3131576, -76.646353, -78.8160018, -80.8322839, -82.7053793, -84.445468, -86.06273, -88.3128656, -90.187177, -91.4721879, -91.9544223, -90.9308066, -88.4326634, -84.6218744, -79.4318427, -72.7959713, -64.6476632, -54.9203216, -46.6727118, -35.7616772, -21.5091369, -3.23700986, 19.7327848, 51.2086284, 87.9557264, 129.869585, 176.845709, 228.779605, 283.107209, 345.189738, 413.692286, 489.831171, 580.031456, 688.379582, 818.961992, 980.876804, 1172.05977, 1384.97935, 1613.38524, 1866.33118, 2150.1194, 2468.74163, 2822.27461, 3212.59212, 3641.85327, 4112.00582, 4623.95135, 5179.18326, 5783.33341, 6433.11996, 7128.60371, 7869.84548, 8656.90605, 9627.14675, 10609.0027, 11568.2097, 12470.5033, 13281.6192, 13802.0979, 14273.0004, 14730.8603, 15337.3277, 16079.4109, 16891.9865, 17764.6692, 18687.0741, 19648.8161, 20639.5101, 21626.7504, 22638.2824, 23658.6214, 24672.2829, 25663.7822, 26617.6345, 27518.3554, 28703.739, 29968.8535},
        {204.04205, 202.680689, 201.384951, 200.150878, 198.974515, 197.851904, 196.77909, 195.752115, 194.767022, 193.819856, 192.906659, 192.023475, 191.166347, 190.331318, 189.514432, 188.711733, 187.919263, 187.133065, 186.349184, 185.563663, 184.772544, 183.971872, 183.157689, 182.32604, 181.472966, 180.594513, 179.686722, 178.745638, 177.767304, 176.747762, 175.683057, 174.569232, 173.40233, 172.178395, 170.726671, 169.130109, 167.473665, 165.758796, 163.986957, 162.159605, 160.278197, 158.344188, 156.359035, 154.324194, 152.241122, 150.111274, 147.936107, 145.717077, 143.455641, 141.153255, 138.811375, 136.430067, 134.012208, 131.559284, 129.072782, 126.554188, 124.004988, 121.42667, 118.820718, 116.188621, 113.531863, 110.851932, 108.150314, 105.428496, 102.687963, 99.9302023, 97.1567004, 94.3826199, 91.6227953, 88.8505065, 86.0663533, 83.2709356, 80.4648531, 77.6487058, 74.8230935, 71.9886159, 69.145873, 66.2954645, 63.4379903, 60.5740503, 57.7042442, 54.8291719, 51.9494333, 49.0656281, 46.226096, 43.4065629, 40.5821806, 37.7519998, 34.915071, 32.0704449, 29.2171721, 26.354303, 23.4808884, 20.5959787, 17.6986246, 14.7878767, 11.8627855, 8.92240169, 5.96577576, 2.99195834, 0, -3.1383602, -6.3131429, -9.51862213, -12.7490719, -15.9987662, -19.261979, -22.5329844, -25.8060564, -29.075469, -32.3354962, -35.5804119, -38.8044903, -42.0020054, -45.1672311, -48.2944414, -51.3779104, -54.2339125, -56.8312931, -59.5669445, -62.3969979, -65.2775844, -68.1648354, -71.0148821, -73.7838557, -76.4278875, -78.9031087, -81.1656506, -83.1716443, -84.8772212, -86.6884039, -87.9464721, -88.4725895, -88.0879199, -86.134453, -82.6729396, -77.8919104, -71.7562815, -64.2309689, -55.2808886, -44.8709567, -36.65724, -25.7901101, -11.5123012, 6.9334523, 30.304416, 62.4046279, 100.182888, 143.63477, 192.755845, 247.541686, 306.312592, 372.786967, 443.964819, 520.501637, 610.516171, 718.776319, 850.049979, 1015.76048, 1212.50769, 1429.63101, 1658.87264, 1910.0941, 2188.75601, 2498.83063, 2841.76815, 3220.17644, 3638.03329, 4098.30174, 4598.92169, 5140.67442, 5727.81838, 6357.12272, 7028.16506, 7740.52304, 8493.7743, 9407.58648, 10331.4247, 11234.8441, 12087.3998, 12858.6469, 13369.5139, 13837.267, 14295.1585, 14888.2191, 15605.4554, 16389.2997, 17231.0872, 18122.1533, 19053.8332, 20017.4623, 21495.3782, 22617.5561, 23500.1879, 24259.4657, 25011.5815, 25872.7273, 26959.0952, 28058.7145, 29195.2177},
        {200.87636, 199.406257, 198.007053, 196.674715, 195.405207, 194.194497, 193.038549, 191.93333, 190.874805, 189.858941, 188.881703, 187.939058, 187.02697, 186.141406, 185.278332, 184.433713, 183.603516, 182.783705, 181.970248, 181.15911, 180.346257, 179.527655, 178.699269, 177.857065, 176.99701, 176.115069, 175.207208, 174.269392, 173.297589, 172.287763, 171.23588, 170.137906, 168.989808, 167.787551, 166.35861, 164.786741, 163.157723, 161.472991, 159.733978, 157.942117, 156.098843, 154.205588, 152.263788, 150.274875, 148.240283, 146.161446, 144.039797, 141.876771, 139.6738, 137.432319, 135.153762, 132.845049, 130.502009, 128.125954, 125.718201, 123.280064, 120.812857, 118.317895, 115.796493, 113.249966, 110.679629, 108.086795, 105.47278, 102.838898, 100.186465, 97.5167947, 94.831202, 92.1424445, 89.4630053, 86.7705989, 84.065798, 81.3491756, 78.6213046, 75.8827579, 73.1341084, 70.3759291, 67.6087929, 64.8332726, 62.0499411, 59.2593715, 56.4621365, 53.6588091, 50.8499622, 48.0361687, 45.2580508, 42.4953145, 39.7271842, 36.9529331, 34.1718345, 31.3831616, 28.5861876, 25.7801858, 22.9644294, 20.1381916, 17.3007457, 14.4513648, 11.5893223, 8.71389133, 5.82434513, 2.91995694, 0, -3.48637764, -6.91514205, -10.2879214, -13.6063438, -16.8720375, -20.0866306, -23.2517512, -26.3690275, -29.4400877, -32.4665599, -35.4500722, -38.3922529, -41.29473, -44.1591317, -46.9870863, -49.7802217, -52.0849492, -53.8378659, -56.0702575, -58.6862057, -61.5897921, -64.6850984, -67.8762062, -71.0671972, -74.162153, -77.0651552, -79.6802855, -81.9116256, -83.663257, -84.9808411, -85.5749674, -85.3072435, -84.0392772, -81.2254267, -76.9309246, -71.3296031, -64.405245, -56.1416328, -46.522549, -35.5317764, -27.2069237, -16.2441741, -1.83417032, 16.8324449, 40.5650288, 72.6944059, 110.8781, 155.2951, 206.124398, 263.544985, 326.983935, 398.291161, 473.184746, 551.793604, 643.252086, 752.651288, 885.082302, 1052.97238, 1252.40915, 1471.47716, 1701.67275, 1952.06608, 2226.06027, 2527.14067, 2858.93192, 3224.99471, 3631.5785, 4082.94105, 4573.48143, 5103.17532, 5674.3999, 6284.35994, 6932.1999, 7617.06418, 8338.09724, 9197.26226, 10065.1802, 10915.2908, 11721.0339, 12455.8491, 12961.9002, 13431.4202, 13894.1101, 14477.4455, 15172.4247, 15929.3065, 16741.1263, 17600.9193, 18501.7208, 19436.5662, 21321.7714, 22528.4064, 23286.5508, 23826.2841, 24377.686, 25170.8361, 26435.8139, 27426.6587, 28421.7668},
        {198.137604, 196.569151, 195.076031, 193.654154, 192.299433, 191.007781, 189.775109, 188.597329, 187.470354, 186.390096, 185.352467, 184.35338, 183.388746, 182.454477, 181.546486, 180.660685, 179.792987, 178.939302, 178.095544, 177.257624, 176.421456, 175.58295, 174.738019, 173.882576, 173.012532, 172.1238, 171.212292, 170.273919, 169.304595, 168.300231, 167.256739, 166.170033, 165.036023, 163.850622, 162.440479, 160.8897, 159.28442, 157.626046, 155.915982, 154.155634, 152.346406, 150.489704, 148.586933, 146.639498, 144.648803, 142.616254, 140.543256, 138.431214, 136.281533, 134.095619, 131.874875, 129.632599, 127.358045, 125.052363, 122.716699, 120.3522, 117.960014, 115.541288, 113.097169, 110.628805, 108.137343, 105.62393, 103.089714, 100.535841, 97.9634601, 95.3737172, 92.76776, 90.1557638, 87.5476868, 84.9260571, 82.2914368, 79.644388, 76.9854726, 74.3152528, 71.6342905, 68.943148, 66.2423872, 63.5325701, 60.8142589, 58.0880157, 55.3544023, 52.613981, 49.8673138, 47.1149628, 44.3900005, 41.6760501, 38.9563435, 36.230388, 33.4976904, 30.7577578, 28.0100973, 25.2542158, 22.4896205, 19.7158182, 16.9323161, 14.1386212, 11.3342404, 8.51868092, 5.69144966, 2.85205368, 0, -3.60615183, -7.10728384, -10.5088133, -13.8161576, -17.0347339, -20.1699595, -23.2272517, -26.2120279, -29.1297052, -31.985701, -34.7854326, -37.5343171, -40.2377721, -42.9012146, -45.530062, -48.1297316, -50.0512802, -51.206645, -53.0813444, -55.5405755, -58.4495355, -61.6734217, -65.0774311, -68.526761, -71.8866085, -75.0221709, -77.7986454, -80.0812291, -81.7351192, -82.5026773, -82.4169759, -81.3800629, -79.2939862, -75.8135681, -70.9644691, -64.838277, -57.4112074, -48.6594758, -38.5592979, -27.0868893, -18.2351525, -6.74114625, 8.20478731, 27.4123061, 51.6910681, 83.2882828, 121.216669, 165.926495, 217.868033, 277.491553, 345.500192, 421.78229, 501.402126, 584.061677, 678.93156, 790.931166, 924.979886, 1092.71343, 1290.80843, 1508.61463, 1739.58346, 1990.18112, 2260.79077, 2553.91409, 2875.64253, 3230.41976, 3626.60986, 4069.67289, 4550.69422, 5068.89036, 5624.38192, 6215.34222, 6840.6748, 7499.28323, 8190.07105, 9000.15378, 9818.17001, 10620.9703, 11385.4053, 12088.3254, 12591.0606, 13062.9959, 13530.4908, 14104.8671, 14778.8533, 15509.7769, 16292.1365, 17120.4307, 17989.1581, 18892.8173, 21022.9594, 22265.929, 22928.7997, 23318.6451, 23742.5389, 24507.5548, 25920.7664, 26807.411, 27654.4313},
        {195.736755, 194.077538, 192.497555, 190.992682, 189.558792, 188.19176, 186.887459, 185.641764, 184.450549, 183.309688, 182.215055, 181.162525, 180.147971, 179.167267, 178.216289, 177.290909, 176.387003, 175.500443, 174.627106, 173.762863, 172.903591, 172.045162, 171.183451, 170.314333, 169.43368, 168.537368, 167.621271, 166.681263, 165.713217, 164.713009, 163.676511, 162.599599, 161.478147, 160.308028, 158.916434, 157.387127, 155.805901, 154.174102, 152.493079, 150.764181, 148.988755, 147.168151, 145.303715, 143.396797, 141.448744, 139.460906, 137.43463, 135.371264, 133.272157, 131.138657, 128.972112, 126.789355, 124.575914, 122.332801, 120.06103, 117.761613, 115.435562, 113.08389, 110.70761, 108.307734, 105.885275, 103.441245, 100.976657, 98.4925242, 95.9898583, 93.4696721, 90.9329784, 88.3877644, 85.8418499, 83.2818628, 80.7083634, 78.1219122, 75.5230696, 72.9123961, 70.2904521, 67.6577982, 65.0149946, 62.3626019, 59.7011805, 57.0312909, 54.3534936, 51.6683488, 48.9764172, 46.2782592, 43.6005401, 40.9302188, 38.2539406, 35.5714186, 32.8823662, 30.1864966, 27.4835231, 24.7731591, 22.0551178, 19.3291125, 16.5948565, 13.8520631, 11.1004455, 8.33971715, 5.56959125, 2.7897811, 0, -3.27694016, -6.49736573, -9.66282711, -12.7748746, -15.8350587, -18.8449297, -21.8060381, -24.719934, -27.5881681, -30.4122905, -33.1938518, -35.9344022, -38.6354922, -41.2986722, -43.9254924, -46.5175033, -48.4140425, -49.5220585, -51.340964, -53.7304065, -56.5500338, -59.6594935, -62.9184331, -66.1865004, -69.323343, -72.1886084, -74.6419444, -76.5429986, -77.7514185, -77.8733114, -77.1148302, -75.4116848, -72.6995849, -69.0002618, -64.2064146, -58.1878032, -50.8543311, -42.1159015, -31.8824178, -20.0637834, -9.6878509, 3.40236676, 19.9069998, 40.5261785, 65.9600329, 96.2916107, 132.992395, 176.916788, 228.919189, 289.854, 361.858482, 442.936235, 528.824627, 619.049285, 720.870208, 838.074817, 974.450537, 1137.36866, 1326.21822, 1536.47859, 1767.68747, 2020.4513, 2290.81999, 2579.44323, 2894.76493, 3241.65132, 3629.61385, 4064.72309, 4536.01707, 5042.1684, 5580.34689, 6150.9591, 6753.19784, 7386.25588, 8049.32602, 8823.14789, 9604.98071, 10373.6306, 11107.9035, 11786.6057, 12283.1918, 12752.0534, 13217.1471, 13778.5596, 14430.1127, 15133.9074, 15885.3515, 16679.853, 17512.8197, 18379.6595, 20447.731, 21635.6207, 22257.6317, 22628.0671, 23061.2301, 23871.4239, 25372.9516, 26209.0685, 26908.7226},
        {193.605523, 191.860423, 190.198213, 188.614736, 187.105838, 185.66736, 184.295148, 182.985044, 181.732894, 180.534541, 179.385829, 178.282601, 177.220703, 176.195976, 175.204266, 174.241417, 173.303272, 172.385675, 171.48447, 170.595501, 169.714612, 168.837646, 167.960449, 167.078862, 166.188732, 165.285901, 164.366212, 163.425512, 162.459642, 161.464447, 160.435771, 159.369458, 158.261351, 157.107295, 155.736577, 154.231821, 152.677656, 151.075331, 149.426095, 147.731197, 145.991885, 144.209409, 142.385018, 140.51996, 138.615485, 136.672842, 134.69328, 132.678048, 130.628394, 128.545568, 126.430819, 124.300032, 122.139504, 119.950164, 117.732945, 115.488778, 113.218596, 110.92333, 108.603911, 106.261271, 103.896343, 101.510057, 99.1033454, 96.6771399, 94.2323721, 91.7699737, 89.2908764, 86.8017363, 84.3090721, 81.8020092, 79.2811084, 76.7469301, 74.2000351, 71.6409838, 69.070337, 66.4886552, 63.896499, 61.294429, 58.6830058, 56.06279, 53.4343423, 50.7982232, 48.1549933, 45.5052133, 42.8710452, 40.2417947, 37.6065074, 34.965043, 32.317261, 29.6630211, 27.0021828, 24.3346058, 21.6601496, 18.9786738, 16.2900379, 13.5941017, 10.8907247, 8.17976637, 5.46108644, 2.73454445, 0, -2.35253879, -4.8262348, -7.40832823, -10.0860593, -12.8466681, -15.6773949, -18.5654799, -21.4981633, -24.4626852, -27.4462859, -30.4362055, -33.4196843, -36.3839624, -39.31628, -42.2038773, -45.0339945, -47.3954211, -49.2184759, -51.3916563, -53.8168204, -56.3958261, -59.0305313, -61.6227938, -64.0744717, -66.2874226, -68.1635047, -69.6045757, -70.5124937, -70.7891164, -70.1543754, -68.7587614, -66.55871, -63.5106572, -60.2113124, -56.2969745, -51.2332029, -44.7843513, -36.7147734, -26.788823, -14.7708537, -1.4901846, 14.6779158, 34.1774565, 57.4524467, 84.9468953, 113.238667, 147.604452, 189.454794, 240.200239, 301.251333, 376.299554, 461.686705, 555.559354, 657.613718, 770.709875, 896.287394, 1035.78584, 1187.92952, 1357.49036, 1552.20247, 1782.7999, 2040.11498, 2314.58579, 2603.98853, 2918.53347, 3262.72343, 3645.56763, 4072.7372, 4533.37989, 5025.99003, 5544.06151, 6091.85836, 6669.60582, 7277.52911, 7915.85348, 8671.27652, 9435.933, 10188.4301, 10907.3748, 11571.3744, 12055.3071, 12510.6615, 12960.5, 13501.357, 14127.2124, 14801.4323, 15519.5271, 16277.0072, 17069.383, 17892.1648, 19488.4158, 20500.1467, 21155.0382, 21680.7711, 22305.0263, 23255.4844, 24759.8263, 25634.4595, 26194.039},
        {191.875385, 190.047613, 188.30608, 186.646602, 185.064994, 183.557073, 182.118655, 180.745555, 179.433589, 178.178573, 176.976323, 175.822654, 174.713383, 173.644326, 172.611297, 171.610114, 170.636591, 169.686546, 168.755793, 167.840148, 166.935428, 166.037448, 165.142024, 164.244972, 163.342107, 162.429247, 161.502205, 160.556799, 159.588844, 158.594156, 157.568551, 156.507845, 155.407853, 154.264391, 152.908328, 151.421422, 149.887415, 148.307477, 146.682775, 145.014479, 143.303757, 141.551779, 139.759712, 137.928726, 136.059989, 134.15467, 132.213938, 130.238962, 128.23091, 126.19095, 124.120253, 122.034445, 119.919924, 117.777544, 115.60816, 113.412627, 111.191801, 108.946537, 106.677691, 104.386117, 102.072671, 99.7382079, 97.3835834, 95.0096525, 92.6172704, 90.2072926, 87.7805742, 85.3426958, 82.8991243, 80.4409696, 77.9687808, 75.4831066, 72.9844961, 70.4734983, 67.9506621, 65.4165364, 62.8716701, 60.3166124, 57.751912, 55.1781179, 52.5957792, 50.0054446, 47.4076633, 44.8029841, 42.2096834, 39.6190734, 37.0222536, 34.4191977, 31.8098793, 29.1942721, 26.5723497, 23.9440859, 21.3094543, 18.6684285, 16.0209822, 13.3670892, 10.706723, 8.03985727, 5.36646577, 2.68652212, 0, -1.40488349, -3.11659723, -5.10740204, -7.34955875, -9.81532818, -12.4769711, -15.3067485, -18.276921, -21.3597495, -24.5274949, -27.7524179, -31.0067795, -34.2628403, -37.4928612, -40.6691031, -43.7638268, -46.6496305, -49.2845375, -51.8662359, -54.3456303, -56.6736252, -58.8011253, -60.679035, -62.2582589, -63.4897015, -64.3242675, -64.7128613, -64.6063875, -63.9557507, -62.6681685, -60.7542508, -58.1780085, -54.9034519, -51.9981271, -48.874277, -44.6498685, -38.9578518, -31.431177, -21.702794, -9.40565292, 6.6326807, 25.7287552, 48.0920453, 73.9320256, 103.458171, 130.161743, 162.649982, 202.811917, 252.536575, 313.712985, 391.411384, 480.451445, 580.744548, 692.074822, 813.966586, 946.060812, 1087.99847, 1231.96114, 1386.7445, 1569.2817, 1798.427, 2058.22505, 2335.97319, 2627.50303, 2943.10703, 3286.43985, 3664.9071, 4083.13592, 4532.00012, 5010.1532, 5509.43339, 6036.35792, 6591.93938, 7177.1904, 7793.12358, 8531.70067, 9280.24785, 10017.0404, 10720.3537, 11368.463, 11835.9431, 12273.9036, 12705.0628, 13225.5917, 13829.1336, 14478.727, 15169.5454, 15896.7628, 16655.5527, 17441.089, 18456.8365, 19272.9594, 19997.2044, 20737.318, 21601.0468, 22696.1374, 24130.3363, 25035.647, 25460.8845},
        {190.658088, 188.744273, 186.919962, 185.180933, 183.522964, 181.941835, 180.433323, 178.993207, 177.617266, 176.301278, 175.041022, 173.832276, 172.670818, 171.552427, 170.472882, 169.427961, 168.413443, 167.425106, 166.458728, 165.510088, 164.574965, 163.649137, 162.728383, 161.80848, 160.885209, 159.954346, 159.011671, 158.052961, 157.073997, 156.070555, 155.038415, 153.973355, 152.871153, 151.727589, 150.372674, 148.888556, 147.359376, 145.786292, 144.170463, 142.513045, 140.815198, 139.078079, 137.302845, 135.490655, 133.642667, 131.760039, 129.843928, 127.895493, 125.915892, 123.906281, 121.867821, 119.820829, 117.746887, 115.646738, 113.521123, 111.370787, 109.19647, 106.998915, 104.778866, 102.537065, 100.274254, 97.9911754, 95.6885724, 93.3671872, 91.0277625, 88.6710407, 86.2977645, 83.9122024, 81.5185384, 79.1102428, 76.6878296, 74.2518127, 71.8027061, 69.3410237, 66.8672795, 64.3819876, 61.8856618, 59.3788161, 56.8619645, 54.3356209, 51.8002994, 49.2565138, 46.7047782, 44.1456065, 41.5951065, 39.0456673, 36.4894905, 33.926584, 31.3569557, 28.7806136, 26.1975655, 23.6078195, 21.0113833, 18.4082649, 15.7984722, 13.1820132, 10.5588958, 7.92912785, 5.29271728, 2.64967203, 0, -0.816164366, -2.05772283, -3.68705701, -5.66654852, -7.95857898, -10.52553, -13.3297832, -16.3337202, -19.4997227, -22.7901722, -26.1674503, -29.5939387, -33.032019, -36.4440728, -39.7924817, -43.0396274, -46.2718663, -49.4692907, -52.3138629, -54.7945306, -56.9002415, -58.6199432, -59.9425833, -60.8571096, -61.3524696, -61.4176111, -61.0414816, -60.213029, -58.9212007, -57.2959376, -55.1334968, -52.3805279, -48.9836806, -46.0985948, -43.0674251, -38.9099288, -33.2100585, -25.5517669, -15.5190067, -2.69573062, 14.8375923, 35.5359438, 59.5212116, 86.915283, 117.840046, 144.484394, 176.886456, 217.151369, 267.384268, 329.69029, 410.245172, 502.108267, 603.679968, 714.241563, 834.872836, 965.846082, 1107.4336, 1254.24251, 1413.49782, 1598.6047, 1824.49034, 2079.44711, 2355.65216, 2650.28983, 2969.02894, 3313.64379, 3687.52214, 4092.8566, 4525.9238, 4986.34679, 5470.82489, 5982.35467, 6521.57124, 7089.10971, 7685.60518, 8399.07322, 9120.92336, 9829.94559, 10504.9299, 11124.6663, 11564.0097, 11974.9751, 12380.8962, 12883.7569, 13475.7582, 14116.3304, 14799.3133, 15518.5466, 16267.8699, 17041.1231, 17440.6006, 18082.7144, 18926.0499, 19929.1922, 21050.7266, 22249.2384, 23483.3128, 24374.0537, 24692.206},
        {189.838058, 187.839194, 185.933003, 184.115217, 182.381568, 180.727787, 179.149608, 177.642761, 176.202979, 174.825995, 173.50754, 172.243346, 171.029145, 169.86067, 168.733653, 167.643824, 166.586918, 165.558665, 164.554798, 163.571049, 162.603149, 161.646832, 160.697829, 159.751872, 158.804693, 157.852024, 156.889597, 155.913145, 154.9184, 153.901092, 152.856956, 151.781722, 150.671123, 149.520891, 148.15858, 146.66755, 145.133179, 143.556657, 141.939173, 140.281917, 138.58608, 136.85285, 135.083418, 133.278973, 131.440705, 129.569803, 127.667458, 125.734859, 123.773196, 121.783658, 119.767436, 117.751591, 115.71088, 113.645932, 111.557375, 109.445839, 107.311952, 105.156343, 102.979642, 100.782476, 98.5654756, 96.3292686, 94.0744842, 91.8017511, 89.5116984, 87.2049549, 84.8821493, 82.5467096, 80.2019962, 77.8428655, 75.4697649, 73.0831419, 70.683444, 68.2711184, 65.8466127, 63.4103742, 60.9628504, 58.5044888, 56.0357368, 53.5570417, 51.068851, 48.5716121, 46.0657725, 43.5517796, 41.0432143, 38.533681, 36.0166264, 33.4920716, 30.9600378, 28.4205463, 25.8736184, 23.3192751, 20.7575378, 18.1884277, 15.611966, 13.0281738, 10.4370725, 7.83868328, 5.23302729, 2.62012579, 0, -0.450904671, -1.40161493, -2.80817679, -4.6266363, -6.81303946, -9.32343232, -12.1138609, -15.1403712, -18.3590093, -21.7258212, -25.1968529, -28.7281505, -32.2757599, -35.7957272, -39.2440985, -42.5769197, -46.0333792, -49.6099719, -52.622303, -55.0870917, -57.0210575, -58.4409198, -59.3633979, -59.8052111, -59.783079, -59.3137208, -58.413856, -57.1002038, -55.3894838, -53.6425067, -51.4057335, -48.5926562, -45.1167667, -41.9664657, -38.5177907, -33.860137, -27.5845239, -19.2819705, -8.54349632, 5.03987948, 23.3014486, 44.7923947, 69.643029, 97.9836627, 129.944607, 157.753709, 191.419359, 233.047486, 284.744015, 348.614876, 431.428976, 524.93006, 624.601932, 729.051478, 843.261882, 969.3535, 1109.44669, 1265.19737, 1439.29589, 1634.93924, 1856.79898, 2103.12099, 2374.84277, 2672.19592, 2994.21589, 3340.48721, 3709.63757, 4101.00341, 4517.42947, 4959.77602, 5432.35187, 5932.03815, 6458.50137, 7011.40805, 7590.42469, 8273.94619, 8963.22859, 9638.2563, 10279.0137, 10865.4853, 11274.9762, 11658.603, 12040.3554, 12528.3994, 13113.4053, 13750.9696, 14433.516, 15153.4683, 15903.2503, 16675.2856, 16510.7396, 17012.1183, 17988.651, 19249.5666, 20604.0944, 21861.4634, 22830.9029, 23657.9143, 23870.8463},
        {189.297894, 187.219564, 185.236961, 183.345763, 181.541646, 179.820287, 178.177364, 176.608552, 175.109529, 173.675972, 172.303558, 170.987963, 169.724864, 168.509938, 167.338863, 166.207314, 165.110969, 164.045505, 163.006598, 161.989926, 160.991165, 160.005992, 159.030085, 158.059119, 157.088772, 156.11472, 155.132641, 154.138211, 153.127108, 152.095008, 151.037588, 149.950524, 148.829495, 147.670176, 146.296901, 144.794623, 143.250353, 141.665329, 140.040789, 138.377969, 136.678106, 134.942439, 133.172204, 131.368639, 129.532981, 127.666467, 125.770334, 123.845821, 121.894163, 119.916599, 117.914366, 115.920378, 113.90351, 111.86431, 109.80333, 107.721121, 105.618232, 103.495215, 101.35262, 99.1909971, 97.0108977, 94.812872, 92.5974705, 90.365244, 88.1167429, 85.8525179, 83.5731194, 81.2823517, 78.983941, 76.6717274, 74.3460505, 72.00725, 69.6556655, 67.2916368, 64.9155034, 62.5276049, 60.1282812, 57.7178717, 55.2967162, 52.8651542, 50.4235256, 47.9721698, 45.5114266, 43.0416357, 40.5713064, 38.0965218, 35.6132664, 33.1216149, 30.6216417, 28.1134213, 25.5970283, 23.0725373, 20.5400226, 17.999559, 15.4512208, 12.8950826, 10.3312189, 7.7597043, 5.18061326, 2.59402032, 0, -0.173041434, -0.899037908, -2.12970826, -3.81677131, -5.91194592, -8.3669509, -11.1335051, -14.1633273, -17.4081365, -20.8196513, -24.3495907, -27.9496735, -31.5716185, -35.1671445, -38.6879705, -42.0858152, -45.7002443, -49.5383834, -52.6746787, -55.1439588, -56.9810528, -58.2207896, -58.897998, -59.0475068, -58.7041451, -57.9027416, -56.6781253, -55.0651251, -53.0985697, -51.3519642, -49.1239469, -46.2877438, -42.716581, -39.0983667, -34.8989859, -29.3670752, -22.131456, -12.82095, -1.06437863, 13.5094367, 32.2234175, 54.2073383, 79.6461668, 108.724871, 141.628419, 171.547167, 207.409347, 251.14858, 304.698487, 369.992688, 453.617911, 547.167541, 643.694597, 741.388571, 848.925914, 970.274645, 1109.40278, 1275.29031, 1465.75421, 1673.14384, 1891.25748, 2128.67159, 2394.80557, 2693.02962, 3016.46397, 3362.93377, 3727.27634, 4106.5422, 4508.73719, 4935.6702, 5398.18161, 5887.65936, 6402.78604, 6942.24426, 7504.71662, 8156.75795, 8812.21056, 9452.789, 10060.2078, 10616.1815, 11004.3856, 11369.933, 11736.555, 12212.9408, 12789.1924, 13420.0128, 14097.2742, 14812.8486, 15558.6082, 16326.4249, 15742.9674, 16149.957, 17237.2218, 18694.5897, 20211.889, 21478.9476, 22185.5936, 22894.4763, 22978.3044},
        {188.77964, 186.644426, 184.607476, 182.66441, 180.810847, 179.042407, 177.354709, 175.743373, 174.204019, 172.732266, 171.323733, 169.97404, 168.678807, 167.433654, 166.234199, 165.076063, 163.954864, 162.866223, 161.80576, 160.769092, 159.751841, 158.749626, 157.758066, 156.772781, 155.78939, 154.803513, 153.81077, 152.80678, 151.787162, 150.747537, 149.683524, 148.590741, 147.46481, 146.301349, 144.922331, 143.413852, 141.864093, 140.274311, 138.645759, 136.979692, 135.277365, 133.540033, 131.768951, 129.965373, 128.130553, 126.265748, 124.372211, 122.451197, 120.503961, 118.531758, 116.535842, 114.547316, 112.536938, 110.505319, 108.453065, 106.380784, 104.289083, 102.178572, 100.049856, 97.9035442, 95.7402442, 93.5605635, 91.36511, 89.1544912, 86.9293151, 84.6901891, 82.4377212, 80.1800309, 77.9250573, 75.6579156, 73.3787265, 71.0876108, 68.7846892, 66.4700826, 64.1439117, 61.8062972, 59.4573599, 57.0972205, 54.7259999, 52.3438187, 49.9507978, 47.5470579, 45.1327197, 42.7079041, 40.2658136, 37.8101736, 35.3447926, 32.8700158, 30.3861886, 27.8936561, 25.3927636, 22.8838565, 20.3672799, 17.8433792, 15.3124996, 12.7749863, 10.2311847, 7.68144001, 5.12609748, 2.56550239, 0, 0.324300592, 0.0159409468, -0.871766609, -2.28550975, -4.17197615, -6.47785348, -9.14982943, -12.1345916, -15.3788278, -18.8292256, -22.4324727, -26.1352568, -29.8842656, -33.6261866, -37.3077077, -40.8755164, -44.7303577, -48.8837844, -52.2554502, -54.8893408, -56.8294418, -58.1197388, -58.8042175, -58.9268635, -58.5316624, -57.6625999, -56.3636616, -54.6788331, -52.6521001, -51.0686893, -48.9595569, -46.1463161, -42.4505801, -38.1402249, -32.8137325, -25.9505814, -17.2385062, -6.36524128, 6.98147883, 23.1139196, 42.7404726, 65.6567172, 92.0974161, 122.297332, 156.491227, 189.879476, 228.989827, 275.315639, 330.350273, 395.587088, 475.349955, 564.842209, 659.223436, 757.605904, 867.171934, 991.480559, 1134.09081, 1303.94214, 1497.99033, 1707.32163, 1923.78052, 2157.58148, 2418.65029, 2710.8513, 3027.77823, 3367.73919, 3726.97628, 4103.26199, 4501.94417, 4924.08561, 5377.38853, 5855.25485, 6356.44913, 6879.73588, 7423.87965, 8044.8567, 8669.91691, 9283.52188, 9870.13323, 10414.2126, 10818.1254, 11203.1601, 11588.9624, 12061.3298, 12613.7229, 13212.0388, 13851.1163, 14525.7939, 15230.9103, 15961.3043, 15404.31, 15826.5005, 16912.2368, 18345.8798, 19811.7905, 20994.3299, 21577.8591, 22062.882, 21939.823},
        {188.497448, 186.306281, 184.216157, 182.222621, 180.321218, 178.507493, 176.776991, 175.125257, 173.547835, 172.040272, 170.598111, 169.216898, 167.892178, 166.619496, 165.394396, 164.212424, 163.069124, 161.960042, 160.880723, 159.826711, 158.793551, 157.776789, 156.771969, 155.774637, 154.780336, 153.784614, 152.783013, 151.771079, 150.744358, 149.698394, 148.628731, 147.530916, 146.400493, 145.233007, 143.84741, 142.331257, 140.774228, 139.177597, 137.542641, 135.870636, 134.162857, 132.420579, 130.64508, 128.837633, 126.999516, 125.132004, 123.236372, 121.313897, 119.365854, 117.393519, 115.398168, 113.40785, 111.396487, 109.364774, 107.313406, 105.243079, 103.154488, 101.048329, 98.9252964, 96.7860861, 94.6313932, 92.4619132, 90.2783413, 88.081373, 85.8717035, 83.6500282, 81.4170425, 79.1858718, 76.9693434, 74.7425161, 72.5052795, 70.2575232, 67.9991367, 65.7300095, 63.4500312, 61.1590914, 58.8570795, 56.5438852, 54.2193979, 51.8835072, 49.5361027, 47.1770739, 44.8063103, 42.4237014, 40.0057995, 37.5646536, 35.1125933, 32.6502656, 30.1783173, 27.6973953, 25.2081466, 22.7112179, 20.2072561, 17.6969083, 15.1808212, 12.6596417, 10.1340168, 7.60459324, 5.07201801, 2.53693797, 0, 0.693859821, 0.703943179, 0.0862501222, -1.1032193, -2.80846505, -4.97348708, -7.54228534, -10.4588598, -13.6672104, -17.111337, -20.7352398, -24.4829185, -28.2983731, -32.1256037, -35.9086102, -39.5913924, -43.5586266, -47.8172664, -51.3145443, -54.0888908, -56.1787365, -57.6225119, -58.4586477, -58.7255743, -58.4617223, -57.7055223, -56.4954049, -54.8698006, -52.8671399, -51.3742754, -49.2701276, -46.3386006, -42.3635982, -37.2575194, -30.7400206, -22.4964931, -12.2722926, 0.18722564, 15.1367059, 32.8307928, 53.3381398, 77.1559883, 104.575372, 135.887326, 171.382882, 208.518539, 251.128502, 300.212439, 356.770018, 421.800908, 496.966644, 581.796079, 674.552998, 775.540866, 889.227486, 1018.21096, 1165.08938, 1336.77425, 1530.56995, 1739.07536, 1954.95469, 2187.30421, 2444.08841, 2728.80275, 3037.37019, 3369.18958, 3723.24558, 4098.82967, 4496.75204, 4916.96383, 5361.9595, 5828.75554, 6316.42269, 6824.03169, 7350.65326, 7942.34063, 8539.43643, 9129.26575, 9699.15372, 10236.4254, 10664.3445, 11077.0052, 11489.1876, 11958.63, 12483.1082, 13042.4987, 13635.4722, 14260.6994, 14916.8511, 15602.5979, 15227.7821, 15706.2973, 16754.834, 18090.0831, 19428.7353, 20487.4812, 20983.0115, 21205.9431, 20839.3326},
        {188.502602, 186.248095, 184.098011, 182.047795, 180.092894, 178.228753, 176.450817, 174.754531, 173.135342, 171.588694, 170.110033, 168.694804, 167.338453, 166.036426, 164.784167, 163.577122, 162.410737, 161.280458, 160.181728, 159.109995, 158.060703, 157.029298, 156.011226, 155.001931, 153.99686, 152.991457, 151.981168, 150.961439, 149.927715, 148.875442, 147.800064, 146.697028, 145.561779, 144.389762, 142.995813, 141.469483, 139.902489, 138.29614, 136.65174, 134.970598, 133.254018, 131.503309, 129.719776, 127.904727, 126.059467, 124.185303, 122.283542, 120.35549, 118.402454, 116.425741, 114.426656, 112.430834, 110.414727, 108.379114, 106.324774, 104.252487, 102.163031, 100.057185, 97.9357297, 95.7994432, 93.649105, 91.4854943, 89.3093902, 87.1215719, 84.9228187, 82.7139098, 80.4956242, 78.2855954, 76.1010522, 73.9080131, 71.7061649, 69.495194, 67.2747872, 65.0446311, 62.8044121, 60.553817, 58.2925323, 56.0202447, 53.7366407, 51.4414069, 49.13423, 46.8147965, 44.4827931, 42.1379064, 39.7418654, 37.3141338, 34.8746316, 32.4242774, 29.9639894, 27.494686, 25.0172858, 22.532707, 20.0418681, 17.5456874, 15.0450834, 12.5409745, 10.0342791, 7.52591549, 5.01680219, 2.50785756, 0, 0.814039323, 0.941399563, 0.43736465, -0.642781488, -2.24375492, -4.31027172, -6.78704795, -9.61879969, -12.750243, -16.126094, -19.6910686, -23.3898831, -27.1672534, -30.9678957, -34.736526, -38.4178603, -42.2829083, -46.3229993, -49.7448047, -52.5629604, -54.7921027, -56.4468677, -57.5418913, -58.0918098, -58.1112593, -57.6148758, -56.6172955, -55.1331545, -53.1770889, -51.5619372, -49.2114589, -45.9008293, -41.4052239, -35.4714359, -27.8888318, -18.4543465, -6.95167013, 6.83550738, 23.123496, 42.1286057, 63.4166438, 88.0524021, 116.379463, 148.741408, 185.48182, 226.077249, 271.955068, 323.675616, 381.799233, 446.88626, 518.187322, 599.173236, 691.288746, 795.968588, 914.627062, 1048.68764, 1199.57381, 1371.06819, 1561.69894, 1767.42056, 1983.75489, 2216.68325, 2470.6937, 2748.27277, 3048.51548, 3372.0736, 3720.94318, 4096.12452, 4493.68888, 4912.49564, 5349.77626, 5806.26854, 6281.39539, 6774.57974, 7285.2445, 7850.97935, 8423.49875, 8992.68394, 9548.41615, 10080.5766, 10531.8093, 10970.7242, 11407.4471, 11872.3158, 12367.4554, 12886.5693, 13432.2038, 14006.9048, 14613.2188, 15253.6917, 15130.6221, 15683.5568, 16678.4131, 17881.1078, 19057.558, 19973.6807, 20395.393, 20344.5064, 19706.0192},
        {188.855419, 186.500362, 184.255536, 182.116206, 180.07764, 178.135105, 176.283867, 174.519194, 172.836353, 171.230611, 169.697234, 168.23149, 166.828646, 165.483969, 164.192725, 162.950182, 161.751607, 160.592267, 159.467428, 158.372359, 157.302324, 156.252593, 155.218431, 154.195107, 153.177885, 152.162035, 151.142822, 150.115514, 149.075378, 148.017681, 146.937689, 145.83067, 144.691891, 143.516618, 142.111634, 140.570412, 138.988693, 137.367859, 135.709295, 134.014385, 132.284511, 130.521059, 128.725411, 126.898951, 125.043064, 123.159132, 121.24854, 119.312671, 117.352909, 115.370637, 113.36724, 111.373662, 109.361084, 107.33025, 105.281903, 103.216786, 101.135642, 99.0392128, 96.9282423, 94.8034731, 92.6656482, 90.5155105, 88.3538028, 86.1812681, 83.9986494, 81.8066894, 79.6061311, 77.4142959, 75.2481072, 73.074116, 70.8919904, 68.7013988, 66.5020091, 64.2934898, 62.0755088, 59.8477345, 57.6098351, 55.3614787, 53.1023335, 50.8320677, 48.5503495, 46.2568471, 43.9512287, 41.6331625, 39.2638208, 36.8629295, 34.450677, 32.0279809, 29.5957586, 27.1549275, 24.7064051, 22.2511088, 19.7899562, 17.3238646, 14.8537516, 12.3805345, 9.90513091, 7.42845819, 4.95143383, 2.47497528, 0, 0.0017221766, -0.518100044, -1.51580025, -2.94771201, -4.77016893, -6.93950457, -9.41205253, -12.1441464, -15.0921197, -18.2123061, -21.4610392, -24.7946524, -28.1694795, -31.541854, -34.8681095, -38.1045795, -41.1224819, -43.8659909, -46.4839638, -48.9144951, -51.0956792, -52.9656108, -54.4623842, -55.524094, -56.0888346, -56.0947006, -55.4797864, -54.1821865, -52.1399955, -49.4882771, -45.8959345, -41.2419711, -35.4053908, -28.7865333, -21.0037342, -11.5363061, -0.106851982, 13.5620254, 29.7477232, 48.7276386, 70.3622765, 95.4728065, 124.418192, 157.557395, 195.24938, 237.053178, 284.327666, 337.63179, 397.524495, 464.564727, 538.127892, 621.403024, 716.213927, 823.961444, 945.182869, 1080.80321, 1231.74749, 1399.37037, 1584.06954, 1785.77398, 2004.51769, 2241.66182, 2498.3629, 2775.54003, 3073.70996, 3393.57412, 3737.01977, 4105.05575, 4494.34302, 4904.00195, 5332.68329, 5780.04928, 6245.3829, 6727.96706, 7227.08473, 7776.17654, 8331.82831, 8884.78356, 9425.78582, 9945.57861, 10393.2634, 10828.9872, 11261.3399, 11710.3377, 12180.048, 12669.7771, 13183.8303, 13726.513, 14302.1304, 14914.9879, 14903.1739, 15494.8751, 16469.8408, 17607.8201, 18688.5624, 19491.8167, 19797.3324, 19547.6031, 18596.626},
        {189.594339, 187.123779, 184.769957, 182.52794, 180.392795, 178.359591, 176.423396, 174.579277, 172.822303, 171.147541, 169.55006, 168.024926, 166.567209, 165.171976, 163.834294, 162.549233, 161.311859, 160.117241, 158.960446, 157.836543, 156.740599, 155.667683, 154.612862, 153.571204, 152.537777, 151.507649, 150.475888, 149.437561, 148.387738, 147.321485, 146.23387, 145.119962, 143.974829, 142.793538, 141.373866, 139.813678, 138.213206, 136.573921, 134.897294, 133.184795, 131.437894, 129.658061, 127.846767, 126.005483, 124.135678, 122.238823, 120.316389, 118.369845, 116.400662, 114.410311, 112.400261, 110.40895, 108.400079, 106.374319, 104.332339, 102.274806, 100.20239, 98.1157607, 96.0155863, 93.9025361, 91.7772791, 89.6404842, 87.4928205, 85.334957, 83.1675626, 80.9913063, 78.8068573, 76.6296703, 74.4748452, 72.3125563, 70.1425141, 67.964429, 65.7780116, 63.5829724, 61.3790217, 59.1658703, 56.9432284, 54.7108067, 52.4683156, 50.2154656, 47.9519672, 45.677531, 43.3918673, 41.0946867, 38.7508661, 36.3782655, 33.9951635, 31.602401, 29.200819, 26.7912582, 24.3745598, 21.9515645, 19.5231133, 17.0900471, 14.6532068, 12.2134333, 9.77156751, 7.32845039, 4.88492282, 2.44182572, 0, -1.06637041, -2.44394254, -4.1034775, -6.0157364, -8.15148037, -10.4814705, -12.9764679, -15.6072338, -18.3445292, -21.1591152, -24.0217529, -26.9032036, -29.7742282, -32.6055879, -35.3680438, -38.0323571, -40.0025778, -41.1685084, -42.7827562, -44.6946444, -46.7534959, -48.8086338, -50.7093812, -52.3050611, -53.4449965, -53.9785104, -53.7549261, -52.6235663, -50.4337544, -46.4889552, -41.3844977, -35.1334624, -27.7489297, -20.5127641, -12.8036541, -3.66194554, 7.27991649, 20.3894868, 36.0343201, 54.5819714, 76.4384889, 101.921219, 131.390184, 165.205409, 203.726916, 245.910007, 293.870608, 348.319924, 409.969159, 479.529517, 557.365613, 644.95885, 743.572708, 854.18514, 977.191158, 1113.2475, 1263.01092, 1425.69433, 1603.72644, 1801.11098, 2022.81227, 2265.20673, 2526.21064, 2805.04135, 3103.12076, 3420.85883, 3759.14335, 4118.50816, 4497.73514, 4896.59715, 5317.14559, 5756.5243, 6213.71746, 6687.70924, 7177.48379, 7714.53512, 8257.21013, 8796.36553, 9322.85804, 9827.54439, 10260.2533, 10680.2215, 11095.8891, 11524.7956, 11971.7794, 12437.8872, 12928.1468, 13447.5859, 14001.2325, 14594.1143, 14643.0827, 15258.5426, 16222.4094, 17316.5983, 18323.0246, 19023.6037, 19200.2509, 18767.5762, 17501.1613},
        {190.764245, 188.170147, 185.699314, 183.34662, 181.106938, 178.975141, 176.946103, 175.014698, 173.175798, 171.424277, 169.755009, 168.162866, 166.642723, 165.189453, 163.797929, 162.463025, 161.179613, 159.942568, 158.746763, 157.58707, 156.458365, 155.35552, 154.273408, 153.206903, 152.150878, 151.100207, 150.049763, 148.99442, 147.92905, 146.848528, 145.747727, 144.62152, 143.464781, 142.272382, 140.833423, 139.249992, 137.626652, 135.964955, 134.26645, 132.532691, 130.765227, 128.96561, 127.135391, 125.276122, 123.389353, 121.476636, 119.539523, 117.579564, 115.59831, 113.597313, 111.578124, 109.587031, 107.579878, 105.557247, 103.51972, 101.467878, 99.4023016, 97.3235732, 95.2322739, 93.1289852, 91.0142884, 88.8887651, 86.7529965, 84.6075641, 82.4530493, 80.2900336, 78.1190983, 75.9533083, 73.8054285, 71.6502944, 69.4876782, 67.3173521, 65.1390883, 62.9526589, 60.757836, 58.554392, 56.3420988, 54.1207288, 51.8900541, 49.6498469, 47.3998793, 45.1399235, 42.8697517, 40.5891361, 38.267524, 35.9204878, 33.5639645, 31.1987104, 28.8254816, 26.4450344, 24.0581253, 21.6655103, 19.2679458, 16.8661881, 14.4609934, 12.053118, 9.64331814, 7.23235015, 4.82097029, 2.40993481, 0, -2.1139965, -4.33501079, -6.64754119, -9.03608601, -11.4851436, -13.9792122, -16.5027901, -19.0403757, -21.5764673, -24.0955632, -26.5821617, -29.0207611, -31.3958598, -33.6919559, -35.8935479, -37.9851341, -38.9350252, -38.5825498, -39.2107258, -40.586297, -42.4760072, -44.6466004, -46.8648202, -48.8974106, -50.5111154, -51.4726783, -51.5488433, -50.5063541, -48.1119546, -42.8750311, -36.280716, -28.4729604, -19.5957155, -11.8086054, -4.24774459, 4.48826505, 14.8601744, 27.3287342, 42.3546953, 60.3988085, 82.4723838, 108.318051, 138.288844, 172.737794, 212.017935, 254.176346, 302.448502, 357.763921, 421.052127, 493.242638, 576.047364, 668.657185, 771.302078, 884.370198, 1008.57265, 1144.47556, 1292.64502, 1450.67652, 1622.78194, 1816.41384, 2040.89974, 2288.21627, 2553.58186, 2834.70106, 3133.49096, 3449.93503, 3783.60584, 4134.38035, 4503.64208, 4891.92236, 5305.32237, 5737.94667, 6188.39856, 6655.28135, 7137.19832, 7665.2499, 8197.41799, 8724.18159, 9236.01972, 9723.41141, 10132.8979, 10528.1878, 10918.5907, 11325.3364, 11753.0161, 12201.2538, 12674.8888, 13178.7604, 13717.708, 14296.5708, 14357.4918, 14980.4387, 15939.1671, 17007.4322, 17958.9895, 18567.5942, 18607.0019, 17991.134, 16429.2872},
        {192.497459, 189.785152, 187.201094, 184.740032, 182.396708, 180.165869, 178.04226, 176.020625, 174.095709, 172.262257, 170.515014, 168.848725, 167.258136, 165.73799, 164.283033, 162.88801, 161.547666, 160.256745, 159.009993, 157.802155, 156.627975, 155.482199, 154.359571, 153.254836, 152.16274, 151.078027, 149.995442, 148.90973, 147.815636, 146.707906, 145.581283, 144.430513, 143.250341, 142.035512, 140.569661, 138.957528, 137.306391, 135.617846, 133.893489, 132.134914, 130.343719, 128.521498, 126.669849, 124.790366, 122.884646, 120.954283, 119.000876, 117.026018, 115.031306, 113.018335, 110.988702, 108.994277, 106.98529, 104.962249, 102.925659, 100.876027, 98.8138584, 96.739659, 94.6539351, 92.5571926, 90.4499376, 88.3326761, 86.205914, 84.0701574, 81.9259124, 79.7736848, 77.6139807, 75.4584632, 73.3185273, 71.1716685, 69.0176698, 66.8563137, 64.687383, 62.5106605, 60.325929, 58.1329711, 55.9315696, 53.7215073, 51.5025669, 49.2745311, 47.0371828, 44.7903046, 42.5336793, 40.2670896, 37.9592171, 35.6260493, 33.2839475, 30.9337038, 28.5761102, 26.2119588, 23.8420416, 21.4671506, 19.0880779, 16.7056156, 14.3205556, 11.9336901, 9.54581099, 7.15771039, 4.77018033, 2.38401285, 0, -2.53331596, -5.09098941, -7.66250578, -10.2373505, -12.805009, -15.3549668, -17.8767092, -20.3597217, -22.7934897, -25.1674987, -27.4712342, -29.6941814, -31.825826, -33.8556532, -35.7731486, -37.5677976, -38.0417009, -37.0147175, -37.1620841, -38.2188464, -39.9200504, -42.0007419, -44.1959668, -46.2407708, -47.8701999, -48.8192999, -48.8231167, -47.616696, -44.9350838, -38.9750229, -31.5739066, -22.9282717, -13.2346548, -5.02534472, 2.67099784, 11.4485792, 21.8115881, 34.2642133, 49.3106437, 67.455068, 90.0281212, 116.456019, 147.081254, 182.246319, 222.293707, 264.602513, 313.219477, 369.227942, 433.711248, 507.75274, 594.001457, 690.061412, 795.18973, 909.33902, 1033.88183, 1169.5532, 1317.08814, 1474.11002, 1645.17272, 1838.2247, 2063.58055, 2312.1162, 2578.44266, 2859.8436, 3158.13141, 3473.03974, 3803.44042, 4148.84368, 4511.91962, 4893.55095, 5302.39775, 5730.36882, 6175.65468, 6636.44587, 7110.93292, 7630.27393, 8151.44997, 8664.40968, 9159.10169, 9625.47465, 10004.1783, 10367.326, 10725.9948, 11111.8186, 11527.737, 11968.1237, 12436.3465, 12935.7731, 13469.7712, 14041.7085, 14008.0818, 14601.6824, 15566.7608, 16647.5673, 17588.3523, 18133.3661, 18026.859, 17185.8825, 15396.1002},
        {194.832096, 192.011405, 189.321928, 186.758366, 184.315418, 181.987787, 179.770171, 177.657273, 175.643792, 173.724428, 171.893884, 170.146858, 168.478052, 166.882166, 165.353902, 163.887958, 162.479036, 161.121837, 159.811061, 158.541408, 157.30758, 156.104276, 154.926198, 153.768045, 152.624519, 151.49032, 150.360149, 149.228705, 148.09069, 146.940805, 145.773749, 144.584224, 143.36693, 142.116567, 140.61545, 138.968939, 137.284948, 135.565071, 133.810901, 132.024029, 130.20605, 128.358556, 126.48314, 124.581394, 122.654913, 120.705289, 118.734114, 116.742982, 114.733485, 112.707217, 110.665771, 108.663027, 106.647156, 104.618619, 102.577874, 100.525382, 98.4616005, 96.3869903, 94.3020105, 92.2071204, 90.1027795, 87.9894472, 85.8675829, 83.7376461, 81.6000962, 79.4553926, 77.3039947, 75.1578697, 73.0287082, 70.8932361, 68.7511668, 66.6022136, 64.4460902, 62.2825098, 60.1111859, 57.931832, 55.7441614, 53.5478876, 51.3427241, 49.1283842, 46.9045814, 44.6710292, 42.4274409, 40.17353, 37.8695272, 35.5357179, 33.1928617, 30.8419534, 28.4839876, 26.1199593, 23.7508629, 21.3776933, 19.0014452, 16.6231133, 14.2436923, 11.864177, 9.48556198, 7.10884207, 4.73501195, 2.36506636, 0, -2.12545543, -4.35127575, -6.66063548, -9.03670914, -11.4626713, -13.9216964, -16.396959, -18.8716337, -21.3288949, -23.7519173, -26.1238752, -28.4279433, -30.6472961, -32.7651081, -34.7645538, -36.6288077, -37.3354695, -36.7240601, -37.0522688, -38.0877897, -39.598317, -41.351545, -43.1151677, -44.6568795, -45.7443744, -46.1453466, -45.6274904, -43.9584999, -40.9060693, -34.9737993, -27.6569784, -19.1025291, -9.4573734, -0.956644096, 7.29684183, 16.7571235, 27.9037419, 41.2162378, 57.1741522, 76.257026, 99.6943681, 126.9875, 158.469231, 194.472367, 235.329718, 278.254452, 327.478927, 384.115859, 449.277967, 524.077969, 611.349909, 708.381334, 813.80763, 927.446209, 1051.52779, 1187.19957, 1335.60874, 1496.29311, 1672.37503, 1868.73252, 2092.49398, 2337.50758, 2600.28509, 2879.04219, 3174.8818, 3487.58154, 3816.26984, 4160.55601, 4522.42983, 4902.53456, 5309.74575, 5735.20787, 6176.71453, 6632.05936, 7099.03599, 7608.87924, 8117.58122, 8614.57526, 9089.29467, 9531.17277, 9873.43905, 10199.1999, 10521.9759, 10889.5683, 11301.9284, 11744.724, 12218.6206, 12724.2838, 13262.3792, 13833.5722, 13594.9087, 14120.0661, 15102.1276, 16234.1763, 17209.2956, 17720.5687, 17461.0787, 16344.3615, 14413.1848},
        {197.624867, 194.693029, 191.894969, 189.225364, 186.678887, 184.250213, 181.934018, 179.724976, 177.617761, 175.60705, 173.687516, 171.853834, 170.10068, 168.422728, 166.814653, 165.271129, 163.786833, 162.356437, 160.974618, 159.636051, 158.335409, 157.067368, 155.826602, 154.607787, 153.405598, 152.214708, 151.029794, 149.84553, 148.65659, 147.457649, 146.243383, 145.008466, 143.747574, 142.45538, 140.914147, 139.229224, 137.508513, 135.753582, 133.966, 132.147333, 130.299151, 128.423021, 126.520511, 124.59319, 122.642624, 120.670383, 118.678035, 116.667147, 114.639287, 112.596024, 110.538925, 108.521568, 106.492384, 104.451814, 102.400298, 100.338276, 98.2661908, 96.1844812, 94.0935882, 91.9939525, 89.8860147, 87.7702152, 85.6469948, 83.516794, 81.3800533, 79.2372134, 77.0887149, 74.947722, 72.8270937, 70.7010287, 68.5691428, 66.4310517, 64.2863713, 62.1347172, 59.9757053, 57.8089512, 55.6340708, 53.4506797, 51.2583938, 49.0568288, 46.8456005, 44.6243246, 42.3926169, 40.1500931, 37.8459211, 35.5060016, 33.1568419, 30.7996948, 28.4358133, 26.0664502, 23.6928585, 21.3162909, 18.9380005, 16.5592401, 14.1812626, 11.8053208, 9.43266774, 7.06455623, 4.70223918, 2.34696947, 0, -1.47908737, -3.17972943, -5.07555782, -7.14020416, -9.34730008, -11.6704772, -14.0833671, -16.5596016, -19.072812, -21.5966302, -24.1046878, -26.5706163, -28.9680473, -31.2706126, -33.4519438, -35.4856724, -36.6192708, -36.7226336, -37.4118956, -38.5050828, -39.8202213, -41.1753372, -42.3884565, -43.2776052, -43.6608096, -43.3560955, -42.1814892, -39.9550166, -36.4947039, -30.7796032, -23.7743383, -15.5486272, -6.17218828, 2.60334948, 11.5492233, 21.8851799, 34.0460743, 48.4667616, 65.5820968, 85.826935, 110.181288, 138.368937, 170.718076, 207.556897, 249.213596, 292.954523, 342.935176, 400.249209, 465.990276, 541.252032, 628.780329, 725.997271, 831.089412, 943.848315, 1067.34244, 1203.16898, 1352.92515, 1518.49694, 1701.12711, 1901.74338, 2123.19186, 2363.40336, 2621.53978, 2897.12615, 3190.30286, 3500.92782, 3828.55496, 4172.96338, 4535.04688, 4915.06872, 5321.40321, 5744.9547, 6183.17879, 6633.53107, 7093.46712, 7595.17495, 8092.69464, 8574.79866, 9030.25949, 9447.84964, 9752.27866, 10039.0906, 10324.5134, 10674.4414, 11085.443, 11533.3925, 12015.8928, 12530.5473, 13074.9591, 13646.7313, 13191.9278, 13642.518, 14639.0477, 15822.0628, 16832.1089, 17309.732, 16895.4778, 15507.1248, 13485.3624},
        {200.836347, 197.787792, 194.8756, 192.094423, 189.438911, 186.903714, 184.483481, 182.172863, 179.966511, 177.859074, 175.845202, 173.919546, 172.076755, 170.31148, 168.618372, 166.992079, 165.427252, 163.918541, 162.460597, 161.04807, 159.675609, 158.337865, 157.029488, 155.745128, 154.479435, 153.227059, 151.982651, 150.74086, 149.496337, 148.243732, 146.977695, 145.692876, 144.383925, 143.045492, 141.460361, 139.73357, 137.972708, 136.179299, 134.354869, 132.500944, 130.61905, 128.710711, 126.777454, 124.820804, 122.842287, 120.843429, 118.825754, 116.79079, 114.74006, 112.675092, 110.597411, 108.557765, 106.507389, 104.446743, 102.376285, 100.296474, 98.2077677, 96.1106249, 94.0055044, 91.8928646, 89.7731641, 87.6468615, 85.5144154, 83.3762842, 81.2329266, 79.0848011, 76.9323664, 74.7909405, 72.675485, 70.5558114, 68.4314151, 66.3017914, 64.1664354, 62.0248426, 59.876508, 57.7209271, 55.5575951, 53.3860073, 51.2056589, 49.0160451, 46.8166614, 44.6070029, 42.3865649, 40.1548428, 37.8481525, 35.4989079, 33.1402821, 30.7738206, 28.4010689, 26.0235726, 23.6428773, 21.2605286, 18.8780719, 16.4970528, 14.119017, 11.7455099, 9.37807706, 7.01826411, 4.66761658, 2.32768003, 0, -0.731968779, -1.82462321, -3.24080894, -4.94337163, -6.89515692, -9.05901046, -11.3977779, -13.8743049, -16.4514371, -19.0920201, -21.7588997, -24.4149213, -27.0229308, -29.5457737, -31.9462958, -34.1873425, -35.8162385, -36.7364693, -37.8507496, -39.0328851, -40.156681, -41.0959428, -41.7244758, -41.9160855, -41.5445773, -40.4837565, -38.6074286, -35.789399, -31.903473, -26.4590619, -19.8279765, -11.9933404, -2.93827719, 6.14309359, 15.8731462, 27.1971872, 40.4953916, 56.1479346, 74.5349912, 96.0367365, 121.274597, 150.314073, 183.488136, 221.129762, 263.571923, 308.382225, 359.350353, 417.500621, 483.857345, 559.444841, 646.65649, 743.474239, 847.898848, 959.807381, 1082.9077, 1219.18772, 1370.63536, 1541.43562, 1730.87999, 1935.86306, 2154.71143, 2389.72576, 2642.95485, 2915.34283, 3205.96203, 3514.74397, 3841.69632, 4186.77035, 4549.63804, 4930.12936, 5335.60792, 5757.211, 6192.16076, 6637.67938, 7090.989, 7586.22628, 8074.47026, 8543.71449, 8981.95249, 9377.17781, 9645.27949, 9894.42522, 10143.5104, 10476.2647, 10885.9199, 11338.6939, 11829.1694, 12351.929, 12901.5556, 13472.6317, 12830.7826, 13213.4797, 14215.8532, 15433.0332, 16460.1498, 16892.333, 16324.7131, 14687.2353, 12621.3267},
        {204.39771, 201.218955, 198.180405, 195.276635, 192.50222, 189.851735, 187.319755, 184.900855, 182.58961, 180.380596, 178.268386, 176.247556, 174.312682, 172.458337, 170.679098, 168.969538, 167.324234, 165.73776, 164.204691, 162.719602, 161.277068, 159.871664, 158.497966, 157.150547, 155.823984, 154.51285, 153.211722, 151.915174, 150.617782, 149.314119, 147.998762, 146.666284, 145.311262, 143.92827, 142.299547, 140.530302, 138.728297, 136.894997, 135.031869, 133.140377, 131.221988, 129.278167, 127.310381, 125.320094, 123.308773, 121.277884, 119.228891, 117.163261, 115.08246, 112.987953, 110.881206, 108.805505, 106.719588, 104.624015, 102.519346, 100.406139, 98.2849534, 96.1563487, 94.0208838, 91.8791179, 89.7316102, 87.5789199, 85.4216061, 83.260228, 81.0953447, 78.9275154, 76.7572993, 74.6037222, 72.4853665, 70.3647053, 68.2411007, 66.1139146, 63.9825092, 61.8462464, 59.7044884, 57.5565972, 55.4019348, 53.2398632, 51.0697446, 48.8909409, 46.7028142, 44.5047266, 42.29604, 40.0761166, 37.7687949, 35.4127873, 33.0477137, 30.6753872, 28.2976206, 25.9162268, 23.5330188, 21.1498095, 18.7684118, 16.3906386, 14.018303, 11.6532178, 9.29719584, 6.95205018, 4.6195937, 2.30163934, 0, -0.272688736, -0.986175797, -2.09691312, -3.56135264, -5.3359463, -7.37714604, -9.64140378, -12.0851715, -14.664901, -17.3370444, -20.0580536, -22.7843805, -25.472477, -28.078795, -30.5597866, -32.8719036, -34.7003314, -35.9627694, -37.229249, -38.3980936, -39.3676264, -40.0361708, -40.3020503, -40.0635882, -39.2191079, -37.6669327, -35.3053861, -32.0327913, -27.7474718, -22.2057199, -15.499968, -7.57114891, 1.63980482, 11.2539606, 21.7993856, 34.0542806, 48.3591127, 65.0543492, 84.4804575, 106.977905, 132.736262, 162.292817, 196.017562, 234.280487, 277.451582, 323.944565, 376.574768, 436.201251, 503.683071, 579.879288, 666.39476, 762.431215, 866.543995, 978.894236, 1102.92157, 1240.59366, 1393.87817, 1567.6407, 1760.29233, 1967.0827, 2184.12143, 2415.99321, 2666.4889, 2937.3236, 3226.69514, 3534.41582, 3860.44792, 4204.64261, 4566.30128, 4945.03631, 5347.09316, 5764.43067, 6194.36514, 6634.21289, 7081.29023, 7574.25904, 8058.75369, 8521.7541, 8950.2402, 9331.19192, 9572.89967, 9793.49257, 10013.6746, 10328.3216, 10728.7779, 11175.4829, 11661.3261, 12179.1973, 12721.9858, 13282.5813, 12627.8042, 12995.3185, 13972.3433, 15146.0973, 16103.7993, 16432.6682, 15719.9226, 13911.6395, 11849.0237},
        {208.276248, 204.952171, 201.773629, 198.735061, 195.830907, 193.055606, 190.403598, 187.869323, 185.44722, 183.131729, 180.91729, 178.798341, 176.769322, 174.824674, 172.958835, 171.166245, 169.441344, 167.778571, 166.172366, 164.617169, 163.107418, 161.637554, 160.202016, 158.795243, 157.411676, 156.045754, 154.691915, 153.344601, 151.99825, 150.647303, 149.286198, 147.909375, 146.511273, 145.086333, 143.4147, 141.602468, 139.758306, 137.883608, 135.979768, 134.04818, 132.090238, 130.107335, 128.100865, 126.072223, 124.022802, 121.953995, 119.867197, 117.763801, 115.645202, 113.512792, 111.367966, 109.242137, 107.106029, 104.960384, 102.805947, 100.643459, 98.4736641, 96.2973054, 94.1151257, 91.9278683, 89.7362761, 87.5410922, 85.3430597, 83.1429216, 80.9414211, 78.7393011, 76.5373048, 74.3596608, 72.2300339, 70.1007298, 67.9709692, 65.8399726, 63.7069607, 61.5711541, 59.4317733, 57.2880389, 55.1391715, 52.9843918, 50.8229202, 48.6539775, 46.4767841, 44.2905608, 42.094528, 39.8879063, 37.5829487, 35.2241888, 32.8572057, 30.4840421, 28.106741, 25.7273451, 23.3478975, 20.9704409, 18.5970182, 16.2296722, 13.8704458, 11.521382, 9.18452344, 6.86191312, 4.55559386, 2.26760854, 0, -0.1817048, -0.809418534, -1.83874762, -3.22529849, -4.92467756, -6.89249124, -9.08434597, -11.4558482, -13.9626042, -16.5602206, -19.2043037, -21.85046, -24.4542958, -26.9714176, -29.3574318, -31.5679449, -33.2287603, -34.2483697, -35.3283624, -36.3536727, -37.2092351, -37.7799838, -37.9508533, -37.6067779, -36.6326919, -34.9135298, -32.3342259, -28.7797145, -24.1349301, -18.0572116, -10.7425403, -2.14412931, 7.78480844, 18.1724609, 29.5249147, 42.5932161, 57.6997318, 75.1668285, 95.316873, 118.472232, 144.357422, 174.074614, 208.063146, 246.762356, 290.611581, 339.38329, 394.350407, 456.118985, 525.295081, 602.484747, 688.118461, 783.192451, 887.558696, 1001.84351, 1128.25415, 1268.28806, 1423.44269, 1597.41113, 1788.99618, 1994.60535, 2210.90879, 2442.22019, 2692.61354, 2963.7712, 3253.32298, 3560.75888, 3885.44423, 4226.83669, 4584.85114, 4959.14381, 5354.8973, 5765.39132, 6188.36917, 6621.57412, 7062.74948, 7557.44393, 8043.64401, 8507.14166, 8933.72883, 9309.19745, 9536.09497, 9738.95415, 9939.24317, 10235.0809, 10617.5945, 11045.9735, 11512.9421, 12011.2243, 12533.5443, 13072.6263, 12595.2904, 13005.5375, 13923.671, 14969.9943, 15764.8107, 15928.4234, 15081.1359, 13190.2288, 11172.7491},
        {212.481508, 209.00268, 205.67527, 202.493559, 199.45183, 196.544366, 193.765449, 191.109361, 188.570385, 186.142804, 183.820899, 181.598954, 179.47125, 177.432071, 175.475698, 173.596415, 171.788503, 170.046246, 168.363925, 166.735823, 165.156222, 163.619405, 162.119655, 160.651254, 159.208484, 157.785628, 156.376969, 154.976788, 153.579368, 152.178992, 150.769942, 149.346501, 147.902951, 146.433574, 144.715741, 142.856752, 140.966516, 139.046355, 137.09759, 135.121544, 133.119537, 131.092892, 129.04293, 126.970973, 124.878342, 122.76636, 120.636348, 118.489628, 116.327521, 114.15135, 111.962435, 109.778869, 107.58484, 105.381305, 103.169224, 100.949554, 98.7232531, 96.4912798, 94.2545921, 92.0141483, 89.7709063, 87.5258245, 85.2798609, 83.0339737, 80.7891211, 78.5462612, 76.3063522, 74.0994954, 71.9550922, 69.8139954, 67.6752739, 65.5379966, 63.4012324, 61.2640503, 59.1255192, 56.984708, 54.8406856, 52.6925209, 50.539283, 48.3800406, 46.2138628, 44.0398184, 41.8569764, 39.6644057, 37.3622583, 35.0011422, 32.6328543, 30.2596734, 27.8838787, 25.507749, 23.1335633, 20.7636006, 18.4001398, 16.04546, 13.70184, 11.3715589, 9.0568956, 6.7601291, 4.48353835, 2.22940233, 0, -0.179007945, -0.78989873, -1.78889717, -3.13222807, -4.77611624, -6.6767865, -8.79046365, -11.0733725, -13.4817379, -15.9717846, -18.4997374, -21.0218212, -23.4942608, -25.8732809, -28.1151065, -30.1759622, -31.5731134, -32.2000764, -33.0059607, -33.8529282, -34.6031411, -35.1187613, -35.261951, -34.8948723, -33.8796871, -32.0785575, -29.3536456, -25.5671135, -20.5811233, -13.8687806, -5.82395788, 3.57446598, 14.347612, 25.5501247, 37.6863634, 51.5184145, 67.3573421, 85.5142101, 106.300083, 130.026024, 155.963347, 185.779313, 219.988414, 259.105145, 303.643998, 354.831741, 412.292525, 476.362775, 547.378915, 625.67737, 710.438255, 804.567573, 909.401226, 1026.07191, 1155.29746, 1297.98197, 1455.02952, 1628.43412, 1817.76225, 2021.39147, 2237.38935, 2469.00147, 2720.06083, 2991.96228, 3281.9694, 3589.24206, 3912.50053, 4250.79074, 4604.77056, 4974.18608, 5363.09495, 5766.26831, 6181.95978, 6608.42294, 7043.91139, 7540.90616, 8029.87656, 8495.51929, 8922.5311, 9295.60873, 9513.00635, 9703.49161, 9888.80802, 10167.3023, 10530.396, 10937.5924, 11381.9089, 11856.3628, 12353.9715, 12867.7522, 12623.8362, 13092.1709, 13939.1726, 14831.2579, 15434.8434, 15416.3454, 14442.1806, 12512.9427, 10569.1312},
        {216.994077, 213.351738, 209.867117, 206.534322, 203.347464, 200.30065, 197.387991, 194.603594, 191.94157, 189.396026, 186.961074, 184.63082, 182.399376, 180.260848, 178.209348, 176.238983, 174.343863, 172.518097, 170.755794, 169.051063, 167.398013, 165.790754, 164.223394, 162.690042, 161.184808, 159.7018, 158.235128, 156.778901, 155.327228, 153.874217, 152.413979, 150.940622, 149.448255, 147.930987, 146.16267, 144.252175, 142.31102, 140.34046, 138.341754, 136.316158, 134.264929, 132.189323, 130.090599, 127.970011, 125.828818, 123.668276, 121.489642, 119.294173, 117.083126, 114.857757, 112.619323, 110.372236, 108.114529, 105.847391, 103.572011, 101.289577, 99.0012784, 96.7083027, 94.4118387, 92.1130749, 89.8131999, 87.513402, 85.2148698, 82.9187919, 80.6263566, 78.3387525, 76.0571681, 73.8179472, 71.6565898, 69.5017789, 67.352424, 65.2074346, 63.0657203, 60.9261905, 58.7877548, 56.6493226, 54.5098036, 52.3681071, 50.2231428, 48.07382, 45.9190484, 43.7577375, 41.5887967, 39.4111355, 37.1122023, 34.7489815, 32.3797975, 30.0071768, 27.6336463, 25.2617323, 22.8939617, 20.5328611, 18.1809571, 15.8407763, 13.5148454, 11.2056911, 8.91583995, 6.64781862, 4.40415375, 2.187372, 0, -0.231632605, -0.868283403, -1.86776656, -3.18789626, -4.78648665, -6.62135191, -8.6503062, -10.8311637, -13.1217386, -15.479845, -17.8632971, -20.2299091, -22.5374952, -24.7438695, -26.8068461, -28.6842394, -29.758057, -29.9038551, -30.3818669, -31.0265197, -31.6722404, -32.1534562, -32.3045941, -31.9600813, -30.9543448, -29.1218118, -26.2969093, -22.3140644, -17.0077043, -9.62067115, -0.795891361, 9.45473249, 21.1192978, 33.1247381, 45.9897328, 60.5159384, 77.0098015, 95.7777687, 117.126287, 141.361802, 167.355918, 197.286617, 231.741074, 271.306467, 316.569973, 370.154884, 430.103232, 496.493166, 569.402834, 648.910383, 733.001495, 826.404245, 931.952339, 1051.27008, 1183.51257, 1328.94352, 1487.82666, 1660.18681, 1846.58088, 2047.82641, 2263.94017, 2496.53412, 2748.75501, 3021.52156, 3311.97727, 3618.99849, 3940.73489, 4275.87443, 4625.76976, 4990.26625, 5372.30914, 5768.167, 6176.61211, 6596.41681, 7026.3534, 7525.00473, 8016.37993, 8484.2987, 8912.58071, 9285.04563, 9497.41077, 9680.33309, 9855.39044, 10118.8481, 10462.7201, 10847.9008, 11267.9601, 11716.4683, 12186.9952, 12673.111, 12687.4449, 13219.3971, 13988.1188, 14712.761, 15112.4747, 14906.411, 13813.7209, 11883.3947, 10033.7602},
        {221.794543, 217.980599, 214.330957, 210.839542, 207.500279, 204.307092, 201.253907, 198.334647, 195.543238, 192.873605, 190.319673, 187.875365, 185.534608, 183.291326, 181.139443, 179.072884, 177.085575, 175.171439, 173.324403, 171.53839, 169.807325, 168.125133, 166.48574, 164.883069, 163.311046, 161.763595, 160.234641, 158.718109, 157.207924, 155.69801, 154.182292, 152.654696, 151.109145, 149.539566, 147.71549, 145.747758, 143.749907, 141.723145, 139.668679, 137.587714, 135.481457, 133.351116, 131.197897, 129.023006, 126.827651, 124.613039, 122.380375, 120.130867, 117.865722, 115.586145, 113.293345, 110.978774, 108.653605, 106.319254, 103.977141, 101.628683, 99.275298, 96.9184047, 94.5594211, 92.1997653, 89.8408555, 87.4841099, 85.1309465, 82.7827836, 80.4410392, 78.1071316, 75.7824789, 73.5097378, 71.3305748, 69.1613568, 67.0008285, 64.8477347, 62.7008202, 60.5588299, 58.4205086, 56.2846011, 54.1498523, 52.0150069, 49.8788097, 47.7400057, 45.5973395, 43.4495561, 41.2954002, 39.1336167, 36.8382592, 34.4730408, 32.1031733, 29.731448, 27.3606563, 24.9935894, 22.6330386, 20.2817951, 17.9426504, 15.6183956, 13.3118222, 11.0257212, 8.76288413, 6.52610217, 4.31816663, 2.14186881, 0, -0.306613213, -0.985239572, -1.99576063, -3.29805795, -4.85201308, -6.61750759, -8.55442302, -10.6226409, -12.7820429, -14.9925104, -17.2139251, -19.4061685, -21.5291222, -23.5426677, -25.4066865, -27.0810603, -27.8082571, -27.4456707, -27.5759039, -28.0051066, -28.5394289, -28.985021, -29.1480329, -28.8346148, -27.8509167, -26.0030886, -23.0972808, -18.9396433, -13.3363261, -5.29312719, 4.28998869, 15.3667657, 27.8909481, 40.6340874, 54.1410241, 69.2718503, 86.3349679, 105.638779, 127.491686, 152.20209, 178.337015, 208.476226, 243.268831, 283.363935, 329.410645, 385.217683, 447.484636, 516.070708, 590.8351, 671.637015, 755.455537, 848.550133, 955.092786, 1077.12863, 1212.36051, 1360.44084, 1521.02207, 1692.14636, 1875.4424, 2074.29551, 2290.93831, 2525.0152, 2778.62031, 3052.07373, 3342.68946, 3649.16135, 3969.2654, 4301.45748, 4647.55894, 5007.48743, 5383.16294, 5772.19272, 6173.8013, 6587.2132, 7011.65296, 7510.09858, 8002.08275, 8470.8916, 8899.81131, 9272.12803, 9483.08515, 9662.70678, 9832.01168, 10083.5801, 10410.1044, 10774.4596, 11170.8291, 11593.3959, 12036.3434, 12493.8547, 12760.1196, 13351.3949, 14039.7806, 14597.3762, 14796.2816, 14408.5966, 13206.421, 11305.1982, 9562.22603},
        {226.661734, 222.663478, 218.836908, 215.17573, 211.673651, 208.324377, 205.121615, 202.059072, 199.130453, 196.329466, 193.649818, 191.085214, 188.629361, 186.275967, 184.018738, 181.851379, 179.767599, 177.761103, 175.825597, 173.95479, 172.142387, 170.382094, 168.667619, 166.992668, 165.350948, 163.736165, 162.142025, 160.562236, 158.990504, 157.420535, 155.846037, 154.260715, 152.658277, 151.032429, 149.144337, 147.108868, 145.043687, 142.950045, 140.829197, 138.682395, 136.510892, 134.31594, 132.098794, 129.860706, 127.602929, 125.326715, 123.033319, 120.723993, 118.399989, 116.062562, 113.712963, 111.340351, 108.958336, 106.568434, 104.172159, 101.771026, 99.3665519, 96.9602504, 94.5536371, 92.1482273, 89.7455361, 87.3470787, 84.9543703, 82.5689261, 80.1922612, 77.8258909, 75.4713303, 73.1742821, 70.979388, 68.7970298, 66.6258583, 64.4645243, 62.3116783, 60.1659711, 58.0260534, 55.890576, 53.7581894, 51.6275444, 49.4972917, 47.366082, 45.232566, 43.0953944, 40.9532179, 38.8046872, 36.5104765, 34.1411269, 31.7688379, 29.3967377, 27.0279547, 24.6656171, 22.3128534, 19.9727917, 17.6485604, 15.3432878, 13.0601022, 10.8021318, 8.57250506, 6.3743502, 4.21079553, 2.08496937, 0, 0.00991290227, -0.389125361, -1.15425396, -2.24261207, -3.61133885, -5.21757348, -7.01845512, -8.97112295, -11.0327161, -13.1603739, -15.3112353, -17.4424396, -19.5111259, -21.4744334, -23.2895013, -24.9134688, -25.6143739, -25.2509129, -25.3389059, -25.6878274, -26.1071522, -26.4063549, -26.3949102, -25.8822928, -24.6779772, -22.5914383, -19.4321505, -15.0095887, -9.13322739, -0.812310248, 9.05003896, 20.4042762, 33.2008576, 46.1140657, 59.7324431, 74.9848457, 92.2045813, 111.724958, 133.879284, 159.000868, 185.870435, 216.846401, 252.565839, 293.665822, 340.783425, 397.377443, 460.557797, 530.256132, 606.40409, 688.933316, 775.212643, 870.868846, 979.910707, 1104.24159, 1241.46627, 1391.11953, 1552.73612, 1723.65701, 1906.10934, 2104.7135, 2323.46325, 2559.85995, 2813.62503, 3084.31158, 3371.18741, 3673.65128, 3990.96334, 4322.48641, 4668.09154, 5027.3623, 5401.56977, 5788.35038, 6186.70357, 6595.62875, 7014.12535, 7498.90789, 7976.83193, 8432.46812, 8850.38713, 9215.15961, 9429.52606, 9614.44072, 9789.54431, 10044.2089, 10370.4731, 10731.3211, 11120.2444, 11530.7343, 11956.2825, 12390.3803, 12720.2461, 13300.3701, 13922.6547, 14379.0026, 14461.3163, 13961.4983, 12671.4511, 10820.9711, 9149.53118},
        {231.819896, 227.631666, 223.622385, 219.785547, 216.114647, 212.603179, 209.244636, 206.032514, 202.960305, 200.021504, 197.209606, 194.518104, 191.940492, 189.470265, 187.100917, 184.825942, 182.638834, 180.533087, 178.502195, 176.539652, 174.638954, 172.793592, 170.997063, 169.24286, 167.524476, 165.835407, 164.169146, 162.519188, 160.879026, 159.242155, 157.602069, 155.952262, 154.286228, 152.597461, 150.638503, 148.528392, 146.389063, 144.22183, 142.028006, 139.808904, 137.565838, 135.300121, 133.013067, 130.70599, 128.380203, 126.037019, 123.677751, 121.303715, 118.916222, 116.516587, 114.106122, 111.672752, 109.231471, 106.783881, 104.331587, 101.876192, 99.4193007, 96.9625163, 94.5074428, 92.0556837, 89.608843, 87.1685244, 84.7363316, 82.3138685, 79.9027387, 77.5045461, 75.1208944, 72.8006333, 70.5914784, 68.397592, 66.2175152, 64.0497888, 61.892954, 59.7455517, 57.606123, 55.4732088, 53.3453501, 51.2210879, 49.0989632, 46.9775171, 44.8552905, 42.7308245, 40.6026599, 38.4693379, 36.1771217, 33.8038942, 31.4293683, 29.0570269, 26.6903525, 24.332828, 21.987936, 19.6591592, 17.3499804, 15.0638822, 12.8043474, 10.5748586, 8.37889851, 6.21994993, 4.10149551, 2.02701797, 0, 0.290916181, 0.143316766, -0.397682898, -1.28696746, -2.47942158, -3.9299299, -5.59337708, -7.42464777, -9.37862662, -11.4101983, -13.4742474, -15.5256586, -17.5193167, -19.4101061, -21.1529116, -22.7026178, -23.3629077, -22.9956371, -23.0286469, -23.2772785, -23.5568734, -23.682773, -23.4703188, -22.734852, -21.2917143, -18.9562469, -15.5437914, -10.8696891, -4.74928151, 3.77622218, 13.8334999, 25.3749705, 38.353053, 51.3577715, 65.0227423, 80.344887, 97.685343, 117.405248, 139.865738, 165.427952, 193.185503, 225.152821, 261.939035, 304.153278, 352.40468, 409.527535, 473.349522, 543.923479, 621.302247, 705.538666, 794.619403, 893.188795, 1004.82874, 1131.20858, 1270.09292, 1420.99952, 1583.44611, 1754.35643, 1936.4318, 2135.20341, 2356.21313, 2594.96365, 2848.80958, 3116.56991, 3399.54504, 3697.89638, 4012.4908, 4343.67261, 4689.19933, 5048.29172, 5421.76367, 5806.9977, 6202.66308, 6607.4291, 7019.96502, 7488.94011, 7950.52367, 8390.88497, 8796.19332, 9152.61798, 9371.90957, 9564.26848, 9747.75771, 10008.0637, 10336.4818, 10696.1309, 11079.7154, 11479.9393, 11889.507, 12301.1225, 12657.3013, 13214.0329, 13776.0196, 14147.9633, 14134.5661, 13540.5303, 12170.5579, 10387.2226, 8791.97514},
        {237.302875, 232.921068, 228.72505, 224.708118, 220.863575, 217.18472, 213.664854, 210.297276, 207.075287, 203.992187, 201.041277, 198.215857, 195.509226, 192.914686, 190.425537, 188.035078, 185.736611, 183.523435, 181.38885, 179.326157, 177.328656, 175.389648, 173.502432, 171.66031, 169.85658, 168.084544, 166.337501, 164.608753, 162.891598, 161.179338, 159.465273, 157.742703, 156.004928, 154.245248, 152.208107, 150.016468, 147.796285, 145.548934, 143.275792, 140.978234, 138.657637, 136.315376, 133.952829, 131.571371, 129.172378, 126.757227, 124.327293, 121.883953, 119.428583, 116.962559, 114.487257, 111.989181, 109.484902, 106.976118, 104.464529, 101.951832, 99.4397259, 96.9299099, 94.4240822, 91.9239415, 89.4311862, 86.9475149, 84.4746263, 82.0142188, 79.5679911, 77.1376417, 74.7248692, 72.3822737, 70.1612334, 67.958466, 65.7723704, 63.6013453, 61.4437895, 59.298102, 57.1626814, 55.0359267, 52.9162366, 50.80201, 48.6916457, 46.5835425, 44.4760992, 42.3677147, 40.2567877, 38.1417171, 35.8530262, 33.4765827, 31.1002718, 28.7279402, 26.3634347, 24.010602, 21.6732887, 19.3553415, 17.0606073, 14.7929325, 12.5561641, 10.3541486, 8.19073271, 6.06976323, 3.99508682, 1.97055018, 0, 0.468816641, 0.488670755, 0.105671579, -0.634071654, -1.68444971, -2.99935335, -4.53267335, -6.23830047, -8.07012547, -9.98203912, -11.9279322, -13.8616954, -15.7372196, -17.5083955, -19.1291139, -20.5532656, -21.1138981, -20.6762105, -20.5983814, -20.7012638, -20.8057106, -20.7325748, -20.3027094, -19.3369672, -17.6562014, -15.0812649, -11.4330105, -6.53229132, -0.199960282, 8.47794985, 18.6763318, 30.3538867, 43.4693153, 56.5522404, 70.2759022, 85.6946283, 103.195843, 123.166971, 145.995436, 172.068664, 200.833267, 233.903816, 271.851804, 315.24873, 364.666089, 422.194873, 486.507209, 557.794717, 636.24902, 722.061739, 813.913169, 915.353436, 1029.40891, 1157.48178, 1297.65822, 1449.51324, 1612.62182, 1783.75687, 1965.93232, 2165.21893, 2388.32371, 2629.14594, 2883.03069, 3148.19147, 3427.7022, 3722.40577, 4034.61484, 4365.55337, 4711.05624, 5070.00672, 5442.97118, 5826.89082, 6220.06624, 6620.79809, 7027.387, 7479.93094, 7924.48386, 8348.89705, 8741.0218, 9088.70941, 9313.06164, 9513.179, 9705.51972, 9971.95018, 10303.0379, 10662.097, 11041.0576, 11431.8499, 11826.4041, 12216.6503, 12570.5286, 13096.6785, 13606.3061, 13910.6174, 13820.8183, 13148.115, 11703.7134, 9997.27005, 8485.303},
        {243.144518, 238.567594, 234.182565, 229.982568, 225.960741, 222.110222, 218.424149, 214.895659, 211.517891, 208.283983, 205.187071, 202.220294, 199.376791, 196.649698, 194.032153, 191.517295, 189.098262, 186.76819, 184.520218, 182.347484, 180.243126, 178.200282, 176.212089, 174.271684, 172.372208, 170.506796, 168.668586, 166.850718, 165.046327, 163.248554, 161.450534, 159.645406, 157.826308, 155.986378, 153.863269, 151.583234, 149.275599, 146.941792, 144.583242, 142.201379, 139.79763, 137.373425, 134.930192, 132.469361, 129.99236, 127.500619, 124.995565, 122.478627, 119.951236, 117.414818, 114.870804, 112.302842, 109.730525, 107.15567, 104.580089, 102.005597, 99.4340091, 96.8671385, 94.3067998, 91.7548074, 89.2129754, 86.6831182, 84.16705, 81.6665851, 79.1835378, 76.7197223, 74.276953, 71.9126857, 69.6830402, 67.4750744, 65.2869954, 63.1170105, 60.9633269, 58.824152, 56.6976929, 54.582157, 52.4757515, 50.3766836, 48.2831605, 46.1933897, 44.1055782, 42.0179335, 39.9286626, 37.8359729, 35.5530213, 33.1744326, 30.7970555, 28.4251025, 26.0627863, 23.7143192, 21.3839138, 19.0757825, 16.7941379, 14.5431924, 12.3271586, 10.1502489, 8.01667588, 5.93065197, 3.89638969, 1.91810154, 0, 0.476034296, 0.523520552, 0.187528494, -0.486872154, -1.45461167, -2.67062032, -4.08982839, -5.66716615, -7.35756388, -9.11595185, -10.8972603, -12.6564196, -14.34836, -15.9280117, -17.350305, -18.5701702, -18.9273849, -18.2890006, -18.001364, -17.8875871, -17.7707817, -17.4740596, -16.8205328, -15.633313, -13.7355122, -10.9502422, -7.10061483, -2.00974197, 4.49926451, 13.2983525, 23.6144952, 35.4160628, 48.6714255, 61.884508, 75.7559034, 91.3767233, 109.154671, 129.497452, 152.812768, 179.508324, 209.364777, 243.607717, 282.767532, 327.374608, 377.959332, 435.90637, 500.678259, 572.591817, 651.963861, 739.111207, 833.331293, 937.206225, 1053.21326, 1182.51335, 1323.57994, 1476.09314, 1639.73303, 1811.37056, 1994.1334, 2194.21375, 2418.93076, 2661.22644, 2915.14512, 3178.51903, 3455.5987, 3747.68855, 4058.10251, 4388.66597, 4733.83621, 5092.23835, 5464.41884, 5846.78586, 6237.29943, 6633.91958, 7034.60634, 7471.61608, 7900.0384, 8309.25923, 8688.66452, 9027.64021, 9255.80823, 9462.16119, 9661.69817, 9932.67407, 10265.0492, 10622.4273, 10996.0866, 11377.3051, 11757.3611, 12127.5327, 12459.1714, 12952.6017, 13419.9454, 13673.3242, 13524.8599, 12786.6743, 11270.8893, 9644.43083, 8225.25983},
        {249.378672, 244.607151, 240.032595, 235.648022, 231.446453, 227.420908, 223.564406, 219.869967, 216.330611, 212.939358, 209.689227, 206.573238, 203.584411, 200.715766, 197.960322, 195.311099, 192.761117, 190.303396, 187.930955, 185.636815, 183.413994, 181.255514, 179.154392, 177.10365, 175.096308, 173.125383, 171.183898, 169.264871, 167.361322, 165.46627, 163.572737, 161.67374, 159.762301, 157.831439, 155.614107, 153.238829, 150.837252, 148.410838, 145.961044, 143.489331, 140.997158, 138.485984, 135.95727, 133.412473, 130.853054, 128.280472, 125.696186, 123.101657, 120.498343, 117.887703, 115.271198, 112.626937, 109.980235, 107.333058, 104.687372, 102.045141, 99.4083318, 96.7789096, 94.1588399, 91.5500883, 88.9546205, 86.374402, 83.8113984, 81.2675752, 78.7448981, 76.2453325, 73.7708442, 71.3853516, 69.1512861, 66.9428396, 64.7579616, 62.5946013, 60.4507083, 58.3242319, 56.2131216, 54.1153268, 52.0287968, 49.9514811, 47.8813291, 45.8162902, 43.7543138, 41.6933493, 39.6313461, 37.5662537, 35.2919385, 32.9126839, 30.5352265, 28.1641385, 25.8039921, 23.4593597, 21.1348135, 18.8349259, 16.5642691, 14.3274155, 12.1289373, 9.97340676, 7.86539624, 5.80947802, 3.81022437, 1.87220761, 0, 0.244989163, 0.124450103, -0.320393127, -1.04831647, -2.01809588, -3.18850729, -4.51832666, -5.96632993, -7.49129303, -9.05199193, -10.6072026, -12.1157009, -13.5362628, -14.8276643, -15.9486814, -16.8580899, -16.8634079, -15.8303746, -15.1908492, -14.7640523, -14.3692047, -13.825527, -12.9522399, -11.5685641, -9.49372027, -6.54692917, -2.54741145, 2.6856122, 9.33292109, 18.24291, 28.6839505, 40.6365371, 54.0811645, 67.5416098, 81.7267264, 97.7338259, 115.980418, 136.884013, 160.862121, 188.332251, 219.33108, 254.772857, 295.149604, 340.953342, 392.676091, 451.188937, 516.51007, 589.036746, 669.166219, 757.295744, 853.111128, 958.590621, 1075.80383, 1205.75544, 1347.27585, 1500.17169, 1664.24956, 1836.70976, 2020.55758, 2221.64155, 2447.17004, 2690.02477, 2944.00961, 3206.89533, 3483.17437, 3774.25384, 4083.72085, 4413.54769, 4757.71317, 5114.71765, 5485.33319, 5865.43898, 6252.74905, 6644.97741, 7039.83807, 7463.73123, 7878.51318, 8274.72642, 8642.91342, 8973.61665, 9202.97527, 9412.20397, 9615.16092, 9887.04101, 10217.4233, 10570.33, 10936.6179, 11307.144, 11672.7652, 12024.3385, 12322.4731, 12786.0974, 13223.3684, 13442.4429, 13251.4779, 12458.6305, 10872.0575, 9322.02234, 8007.59071},
        {256.310456, 251.294801, 246.484498, 241.872338, 237.451109, 233.213603, 229.152609, 225.260917, 221.531317, 217.9566, 214.529556, 211.242974, 208.089644, 205.062358, 202.153904, 199.357072, 196.664654, 194.069438, 191.564216, 189.141776, 186.794909, 184.516406, 182.299056, 180.135649, 178.018975, 175.941824, 173.896987, 171.877254, 169.875414, 167.884257, 165.896574, 163.905155, 161.90279, 159.882268, 157.568338, 155.093448, 152.59327, 150.069292, 147.523003, 144.95589, 142.369444, 139.765151, 137.144502, 134.508983, 131.860085, 129.199295, 126.528102, 123.847994, 121.160461, 118.46699, 115.769071, 113.033614, 110.297436, 107.562773, 104.831864, 102.106948, 99.390262, 96.6840443, 93.9905331, 91.3119666, 88.6505829, 86.00862, 83.3883161, 80.7919092, 78.2216376, 75.6797392, 73.1684523, 70.7632476, 68.5378376, 66.3434235, 64.1774961, 62.037546, 59.9210641, 57.825541, 55.7484675, 53.6873342, 51.6396321, 49.6028517, 47.5744838, 45.5520191, 43.5329485, 41.5147625, 39.494952, 37.4710077, 35.2114626, 32.8331022, 30.4554637, 28.083468, 25.7220358, 23.3760881, 21.0505457, 18.7503295, 16.4803602, 14.2455588, 12.0508461, 9.90114295, 7.80137017, 5.75644864, 3.77129921, 1.85084271, 0, -0.889236016, -1.9167098, -3.05613573, -4.28122818, -5.56570152, -6.88327014, -8.20764839, -9.51255067, -10.7716913, -11.9587848, -13.0475454, -14.0116875, -14.8249255, -15.4609738, -15.8935467, -16.0963587, -15.1764657, -12.983774, -11.4522962, -10.3700344, -9.52499042, -8.70516611, -7.69856339, -6.29318414, -4.27703024, -1.43810357, 2.43559397, 7.55606051, 14.1352942, 23.1242142, 33.7249598, 45.9278527, 59.7232148, 73.7474546, 88.6678505, 105.512724, 124.678572, 146.561889, 171.559171, 200.066914, 231.945827, 268.291259, 309.604531, 356.386968, 409.139893, 469.534484, 536.609747, 610.574541, 691.637726, 780.008162, 874.526576, 978.442123, 1094.24797, 1223.21755, 1364.13392, 1516.8983, 1681.41186, 1855.69394, 2041.95529, 2244.45958, 2468.10639, 2708.53818, 2962.76494, 3229.51805, 3510.44562, 3805.85695, 4117.22342, 4445.15559, 4786.00304, 5138.52564, 5503.82233, 5877.95398, 6258.87073, 6644.52271, 7032.86009, 7452.03286, 7862.24134, 8253.88571, 8617.36612, 8943.08276, 9167.46895, 9370.86962, 9566.81362, 9828.11002, 10144.8226, 10482.3149, 10831.8857, 11184.8337, 11532.4577, 11866.0564, 12087.8626, 12530.4486, 12977.4384, 13212.4562, 13019.1261, 12181.0724, 10481.9191, 8996.88369, 7834.48732},
        {263.496459, 258.280886, 253.273866, 248.468157, 243.856517, 239.431704, 235.186478, 231.113597, 227.20582, 223.455905, 219.856611, 216.400696, 213.080919, 209.890038, 206.820813, 203.866001, 201.018362, 198.270653, 195.615634, 193.046063, 190.554698, 188.134299, 185.777623, 183.47743, 181.226477, 179.017524, 176.84333, 174.696651, 172.570249, 170.45688, 168.349303, 166.240277, 164.122561, 161.988913, 159.564376, 156.981196, 154.374841, 151.746755, 149.098387, 146.431181, 143.746586, 141.046047, 138.33101, 135.602923, 132.863231, 130.113382, 127.354821, 124.588996, 121.817352, 119.041336, 116.262396, 113.432376, 110.603399, 107.777987, 104.958661, 102.147943, 99.3483554, 96.5624193, 93.7926565, 91.0415887, 88.3117378, 85.6056254, 82.9257732, 80.2747031, 77.6549367, 75.0689958, 72.5194021, 70.0935267, 67.8767033, 65.6964584, 63.5498133, 61.4337893, 59.3454077, 57.2816898, 55.2396569, 53.2163302, 51.2087311, 49.2138809, 47.2288008, 45.2505123, 43.2760365, 41.3023947, 39.3266083, 37.3456986, 35.1032481, 32.7283282, 30.3530564, 27.9826788, 25.622441, 23.2775891, 20.9533688, 18.655026, 16.3878067, 14.1569567, 11.9677218, 9.825348, 7.73508109, 5.70216696, 3.73185148, 1.82938054, 0, -1.94330934, -3.81493918, -5.60218854, -7.29235647, -8.872742, -10.3306442, -11.653362, -12.8281945, -13.8424407, -14.6833997, -15.3383705, -15.7946522, -16.0395437, -16.0603441, -15.8443524, -15.3788677, -13.5844813, -10.2961053, -7.91383011, -6.19637459, -4.90245749, -3.79079762, -2.6201138, -1.14912481, 0.863450537, 3.65889344, 7.47848509, 12.5635067, 19.1552395, 28.2870977, 39.1177807, 51.6509296, 65.8901858, 80.6242059, 96.4641234, 114.345068, 134.635176, 157.702584, 183.915428, 213.641847, 246.594335, 283.996213, 326.343894, 374.133791, 427.86232, 489.77218, 558.176927, 633.136404, 714.710454, 802.95892, 895.991357, 998.201579, 1112.49669, 1240.40862, 1380.66165, 1533.24066, 1698.13049, 1874.15619, 2062.72604, 2266.5136, 2488.03768, 2725.82978, 2980.32409, 3251.26315, 3537.21756, 3837.29583, 4150.60436, 4476.25112, 4813.35197, 5161.0183, 5521.08049, 5889.51266, 6264.48501, 6644.16776, 7026.73113, 7443.25741, 7850.77671, 8239.23122, 8598.56314, 8918.71466, 9134.38081, 9327.58238, 9511.93893, 9759.27838, 10060.1795, 10380.9674, 10713.4336, 11049.3694, 11380.5663, 11698.8156, 11880.9287, 12314.8835, 12768.755, 13010.618, 12808.5475, 11930.6184, 10144.9056, 8708.91625, 7690.35304},
        {270.751476, 265.426019, 260.302879, 255.375137, 250.635873, 246.078171, 241.695111, 237.479774, 233.425242, 229.524596, 225.770917, 222.157288, 218.676789, 215.322502, 212.087507, 208.964888, 205.947724, 203.029097, 200.202089, 197.459781, 194.795254, 192.20159, 189.67187, 187.199175, 184.776588, 182.397189, 180.054059, 177.74028, 175.448934, 173.173101, 170.905864, 168.640303, 166.3695, 164.086536, 161.531474, 158.829225, 156.107745, 153.368325, 150.612259, 147.840838, 145.055354, 142.257101, 139.447371, 136.627455, 133.798646, 130.962237, 128.11952, 125.271787, 122.420331, 119.566444, 116.711418, 113.788248, 110.868005, 107.953462, 105.047391, 102.152565, 99.2717572, 96.4077407, 93.5632881, 90.7411725, 87.9441667, 85.1750436, 82.4365761, 79.7315372, 77.0626997, 74.4328365, 71.8447206, 69.3960433, 67.1822171, 65.0102526, 62.8767695, 60.7783876, 58.7117266, 56.6734063, 54.6600463, 52.6682664, 50.6946863, 48.7359258, 46.7886045, 44.8493422, 42.9147587, 40.9814736, 39.0461067, 37.1052777, 34.8808013, 32.5124799, 30.1434189, 27.7791558, 25.4252285, 23.0871745, 20.7705316, 18.4808374, 16.2236297, 14.004446, 11.8288241, 9.7023017, 7.63041638, 5.61870588, 3.67270786, 1.79796001, 0, -2.4884047, -4.79231719, -6.9057861, -8.82286006, -10.5375877, -12.0440176, -13.3361985, -14.4081789, -15.2540074, -15.8677328, -16.2434036, -16.3750684, -16.2567759, -15.8825747, -15.2465135, -14.3426408, -11.9955141, -8.03211191, -5.09176044, -2.91790305, -1.25398311, 0.156555984, 1.57027085, 3.24371812, 5.43345439, 8.39603629, 12.3880204, 17.6659635, 24.4864219, 33.8852612, 45.0539826, 58.0153503, 72.7921282, 88.3478556, 105.234909, 124.288124, 145.848033, 170.255168, 197.850059, 228.973238, 263.402396, 302.212204, 345.853316, 394.776383, 449.43206, 511.748627, 580.329701, 655.256526, 736.610348, 824.472411, 916.895427, 1018.46849, 1132.18022, 1259.63697, 1399.6231, 1552.18998, 1717.38899, 1894.58478, 2084.67153, 2289.29258, 2509.42805, 2745.49362, 3000.2442, 3274.07874, 3563.40408, 3866.45952, 4180.574, 4503.75081, 4837.33107, 5180.76783, 5537.20169, 5901.83084, 6272.83184, 6648.38129, 7026.65576, 7440.59689, 7844.92496, 8229.12528, 8582.68317, 8895.08396, 9098.46323, 9277.88919, 9447.42564, 9679.94041, 9966.22067, 10272.5519, 10590.9464, 10913.4163, 11231.9741, 11538.6318, 11763.027, 12206.2245, 12644.7462, 12855.1138, 12613.8492, 11697.4744, 9882.51102, 8472.71184, 7565.50915},
        {277.756863, 272.532119, 267.482227, 262.601386, 257.883796, 253.323656, 248.915165, 244.652523, 240.52993, 236.541584, 232.681686, 228.944434, 225.324029, 221.814669, 218.410554, 215.105884, 211.894858, 208.771675, 205.730536, 202.765638, 199.871183, 197.041369, 194.270395, 191.552462, 188.881768, 186.252513, 183.658897, 181.095119, 178.555378, 176.033875, 173.524807, 171.022375, 168.520779, 166.014217, 163.290398, 160.45111, 157.600624, 154.73984, 151.869658, 148.990979, 146.104703, 143.211732, 140.312964, 137.409302, 134.501645, 131.590894, 128.67795, 125.763712, 122.849082, 119.934959, 117.022245, 114.020252, 111.023454, 108.034737, 105.056986, 102.093089, 99.1459291, 96.2183941, 93.3133696, 90.4337413, 87.5823954, 84.7622176, 81.976094, 79.2269104, 76.5175528, 73.8509072, 71.2298594, 68.7567574, 66.531561, 64.3520233, 62.2145825, 60.1156769, 58.0517447, 56.019224, 54.0145532, 52.0341703, 50.0745138, 48.1320217, 46.2031322, 44.2842837, 42.3719143, 40.4624623, 38.5523658, 36.6380631, 34.4278027, 32.0681794, 29.7092517, 27.3568102, 25.0166453, 22.6945476, 20.3963077, 18.127716, 15.8945631, 13.7026396, 11.5577359, 9.46564254, 7.43215011, 5.46304911, 3.56413007, 1.74118353, 0, -1.38856263, -2.79051962, -4.18516908, -5.5518091, -6.86973778, -8.11825321, -9.2766535, -10.3242367, -11.240301, -12.0041444, -12.595065, -12.992361, -13.1753304, -13.1232713, -12.8154819, -12.2312601, -10.3702033, -7.07165223, -4.5308951, -2.51729411, -0.800211465, 0.850990659, 2.66695007, 4.87830457, 7.71569197, 11.4097501, 16.1911167, 22.2904297, 29.9383268, 39.9458446, 51.7504092, 65.408539, 80.9767521, 97.6267874, 115.857553, 136.403897, 159.587773, 185.731131, 215.155925, 248.184105, 285.396681, 326.777704, 372.598441, 423.13016, 478.644127, 537.743021, 602.783844, 674.45501, 753.444934, 840.442029, 935.813543, 1040.96159, 1157.73621, 1287.14518, 1428.47678, 1581.79131, 1747.14907, 1923.43666, 2112.15483, 2316.08473, 2537.95206, 2776.12969, 3031.38758, 3303.3916, 3589.93676, 3889.67684, 4200.50356, 4520.87314, 4852.0359, 5193.66167, 5551.78336, 5918.72874, 6291.96501, 6668.95935, 7047.17897, 7456.34749, 7853.61157, 8228.37429, 8570.03874, 8868.00801, 9051.45269, 9210.16337, 9359.35704, 9578.49713, 9857.36979, 10158.9918, 10474.4878, 10794.9824, 11111.6003, 11415.4659, 11952.0132, 12458.6446, 12798.7432, 12835.6916, 12432.8728, 11453.6697, 9761.46509, 8312.75618, 7440.69129},
        {284.741703, 279.699738, 274.796935, 270.028915, 265.391302, 260.879715, 256.489776, 252.217107, 248.057329, 244.006064, 240.058933, 236.211557, 232.459559, 228.798558, 225.224178, 221.732039, 218.317763, 214.97697, 211.705284, 208.498324, 205.351714, 202.261073, 199.222023, 196.230187, 193.281185, 190.370638, 187.494169, 184.647399, 181.825948, 179.02544, 176.241494, 173.469733, 170.705777, 167.94525, 165.036137, 162.050983, 159.064106, 156.075919, 153.086832, 150.09726, 147.107614, 144.118308, 141.129752, 138.142361, 135.156545, 132.172719, 129.191293, 126.212681, 123.237295, 120.265547, 117.297851, 114.216259, 111.142108, 108.078377, 105.028045, 101.994088, 98.9794858, 95.9872163, 93.0202578, 90.0815883, 87.1741861, 84.3010295, 81.4650967, 78.6693659, 75.9168154, 73.2104233, 70.5531678, 68.0503441, 65.8047905, 63.6084646, 61.4577117, 59.3488774, 57.2783071, 55.2423461, 53.23734, 51.2596342, 49.3055741, 47.3715052, 45.4537729, 43.5487227, 41.6526999, 39.7620501, 37.8731187, 35.9822511, 33.7886406, 31.443459, 29.1014494, 26.7686005, 24.4509009, 22.1543392, 19.884904, 17.6485838, 15.4513673, 13.2992431, 11.1981998, 9.15422594, 7.17331018, 5.26144111, 3.42460732, 1.66879742, 0, 0.0753748936, -0.129139595, -0.573172564, -1.21635311, -2.01831034, -2.93867334, -3.93707121, -4.97313306, -6.00648797, -6.99676505, -7.90359341, -8.68660212, -9.3054203, -9.71967704, -9.88900144, -9.77302259, -8.62758031, -6.31176235, -4.37710218, -2.63241691, -0.886523648, 1.05176051, 3.37361845, 6.27023308, 9.93278727, 14.5524639, 20.320446, 27.4279163, 36.0660577, 46.7552238, 59.2367309, 73.6030108, 89.9464952, 107.631114, 127.113551, 149.044754, 173.735706, 201.497389, 232.640784, 267.476875, 307.489091, 351.459134, 399.468594, 451.599062, 507.932127, 563.466404, 624.637202, 692.796856, 769.2977, 855.492066, 954.307754, 1063.59607, 1183.93607, 1315.66059, 1458.59962, 1612.80895, 1778.34433, 1953.4659, 2140.43315, 2343.4644, 2567.52614, 2808.30754, 3063.77119, 3332.61982, 3614.8103, 3909.72382, 4216.4286, 4534.2247, 4863.55979, 5204.23239, 5565.23114, 5935.75058, 6312.40802, 6691.82081, 7070.60625, 7473.83526, 7862.55819, 8226.27896, 8554.50151, 8836.72975, 8999.74789, 9137.59274, 9266.64812, 9473.38736, 9746.4505, 10045.2736, 10359.957, 10680.6008, 10997.3054, 11300.1709, 12165.7759, 12736.4395, 12970.3257, 12825.5983, 12260.4214, 11232.9588, 9701.37461, 8209.2445, 7328.44202},
        {291.741164, 286.923699, 282.205708, 277.584377, 273.056893, 268.620442, 264.272209, 260.009381, 255.829144, 251.728684, 247.705187, 243.755839, 239.877826, 236.068335, 232.324551, 228.643661, 225.02285, 221.459305, 217.950213, 214.492758, 211.084127, 207.721507, 204.402083, 201.123041, 197.881568, 194.67485, 191.500072, 188.354421, 185.235084, 182.139245, 179.064091, 176.006809, 172.964584, 169.934603, 166.831189, 163.694703, 160.566501, 157.446457, 154.334447, 151.230346, 148.134029, 145.045372, 141.964248, 138.890534, 135.824105, 132.764835, 129.7126, 126.667275, 123.628735, 120.596855, 117.571511, 114.404359, 111.246706, 108.10164, 104.972249, 101.861621, 98.7728452, 95.7090086, 92.6731999, 89.6685072, 86.6980188, 83.7648229, 80.8720076, 78.0226612, 75.2198718, 72.4667277, 69.7663171, 67.2262813, 64.9517513, 62.7301825, 60.5578851, 58.4311696, 56.3463465, 54.2997262, 52.2876192, 50.3063359, 48.3521867, 46.4214821, 44.5105326, 42.6156485, 40.7331403, 38.8593185, 36.9904934, 35.1229757, 32.9508069, 30.6277896, 28.3113597, 26.0076366, 23.72274, 21.4627896, 19.233905, 17.0422058, 14.8938118, 12.7948424, 10.7514175, 8.76965661, 6.85567938, 5.01560546, 3.25555449, 1.58164612, 0, 1.47820341, 2.4221752, 2.89135459, 2.9451808, 2.64309302, 2.04453047, 1.20893237, 0.195737925, -0.935613657, -2.12568316, -3.31503138, -4.4442191, -5.45380711, -6.2843562, -6.87642715, -7.17058077, -6.70140711, -5.35147116, -3.9796008, -2.43936308, -0.584325071, 1.73194616, 4.65588355, 8.33392004, 12.9124885, 18.538022, 25.3569534, 33.5157156, 43.1607416, 54.5079956, 67.6088844, 82.5889816, 99.5738607, 118.075555, 138.526407, 161.522373, 187.373089, 216.388188, 248.877309, 285.150085, 327.464458, 373.588796, 423.451545, 476.981151, 534.10606, 586.955187, 645.206392, 710.738002, 785.428347, 871.155755, 972.882493, 1085.63369, 1208.97344, 1342.69541, 1487.06201, 1642.1252, 1807.93692, 1982.27791, 2167.98753, 2370.38288, 2596.22954, 2838.98018, 3094.08395, 3359.43303, 3637.05916, 3927.09394, 4229.90103, 4545.67217, 4873.74838, 5213.95187, 5577.66453, 5951.37054, 6330.89075, 6712.04599, 7090.65708, 7486.64278, 7865.70151, 8217.62962, 8532.22344, 8799.27933, 8944.55654, 9064.57989, 9176.59022, 9372.41539, 9639.77219, 9935.29942, 10248.2677, 10567.9478, 10883.6104, 11184.526, 12299.2108, 12912.7434, 13059.7184, 12774.7303, 12092.3738, 11047.2433, 9673.93336, 8158.05248, 7230.0431},
        {298.907735, 294.227514, 289.616369, 285.072693, 280.594876, 276.181309, 271.830383, 267.54049, 263.31002, 259.137365, 255.020915, 250.959062, 246.950197, 242.99271, 239.084994, 235.225439, 231.412436, 227.644376, 223.91965, 220.23665, 216.593766, 212.98939, 209.421913, 205.889725, 202.391218, 198.924783, 195.488811, 192.081693, 188.70182, 185.347584, 182.017375, 178.709584, 175.422603, 172.154823, 168.873905, 165.594253, 162.330637, 159.082444, 155.849065, 152.629885, 149.424295, 146.231682, 143.051435, 139.882941, 136.72559, 133.578769, 130.441866, 127.314271, 124.19537, 121.084554, 117.981208, 114.699912, 111.42887, 108.171477, 104.931128, 101.711218, 98.5151408, 95.3462917, 92.2080651, 89.1038556, 86.0370579, 83.0110667, 80.0292765, 77.095082, 74.2118778, 71.3830586, 68.612019, 66.0121538, 63.6942167, 61.4347353, 59.2299734, 57.0761947, 54.969663, 52.9066422, 50.8833959, 48.896188, 46.9412822, 45.0149424, 43.1134322, 41.2330154, 39.369956, 37.5205175, 35.6809638, 33.8475586, 31.7147635, 29.4360914, 27.1686826, 24.918595, 22.6918863, 20.4946146, 18.3328375, 16.212613, 14.1399988, 12.1210529, 10.161833, 8.26839699, 6.44680275, 4.70310808, 3.04337085, 1.47364887, 0, 1.42743863, 2.33964727, 2.79711436, 2.86032836, 2.58977771, 2.04595084, 1.28933621, 0.380422258, -0.620302573, -1.65234984, -2.65523109, -3.56845789, -4.3315418, -4.88399436, -5.16532713, -5.11505167, -4.43257958, -3.0231224, -1.36927496, 0.640901192, 3.11934448, 6.17799335, 9.92878623, 14.4836616, 19.9545578, 26.4534133, 34.0921666, 42.9827561, 53.2371203, 64.8534578, 78.0991514, 93.1202614, 110.062848, 128.448174, 148.734699, 171.547494, 197.220059, 226.085894, 258.478498, 294.731371, 336.999909, 383.241225, 433.43236, 487.550358, 545.572258, 601.207213, 662.267128, 730.296015, 806.83789, 893.436766, 993.67431, 1104.56642, 1226.38665, 1359.33034, 1503.43321, 1658.80265, 1825.54604, 2002.28993, 2190.9591, 2395.0938, 2619.20281, 2859.03912, 3111.83413, 3376.24563, 3653.34839, 3943.10775, 4245.70227, 4561.15256, 4888.69746, 5228.01801, 5590.39887, 5962.13233, 6338.88283, 6716.31481, 7090.09269, 7476.97186, 7846.25307, 8188.328, 8493.58834, 8752.4258, 8894.15641, 9010.96462, 9119.41741, 9307.23334, 9562.81523, 9844.91936, 10143.4309, 10448.2349, 10749.2166, 11036.261, 11953.9579, 12503.9887, 12682.3854, 12485.1799, 11908.4041, 10948.0899, 9600.2694, 8174.51284, 7133.15197},
        {306.085713, 301.559337, 297.07041, 292.618561, 288.203422, 283.824621, 279.481789, 275.174556, 270.902551, 266.665406, 262.462749, 258.294211, 254.159422, 250.058012, 245.989611, 241.953849, 237.950355, 233.978761, 230.038696, 226.129789, 222.251671, 218.403973, 214.586323, 210.798353, 207.039691, 203.309969, 199.608815, 195.935861, 192.290735, 188.673069, 185.082492, 181.518634, 177.981125, 174.469595, 171.005769, 167.577766, 164.173437, 160.791695, 157.431453, 154.091624, 150.771121, 147.468857, 144.183745, 140.914697, 137.660628, 134.420449, 131.193074, 127.977416, 124.772388, 121.576902, 118.389872, 114.989035, 111.599275, 108.224299, 104.867814, 101.533528, 98.2251485, 94.9463835, 91.7009405, 88.4925271, 85.3248507, 82.2016192, 79.1265399, 76.1033207, 73.135669, 70.2272925, 67.3818989, 64.718435, 62.3530809, 60.0518718, 57.8110449, 55.6268371, 53.4954857, 51.4132277, 49.3763002, 47.3809405, 45.4233855, 43.4998724, 41.6066383, 39.7399204, 37.8959557, 36.0709813, 34.2612345, 32.4629522, 30.372279, 28.1418102, 25.9277484, 23.7360692, 21.5727483, 19.4437614, 17.3550842, 15.3126924, 13.3225617, 11.3906679, 9.5229865, 7.72549333, 6.00416406, 4.3649744, 2.81390003, 1.35691666, 0, 1.07153402, 1.70740523, 1.96497181, 1.90159194, 1.57462382, 1.04142561, 0.359355518, -0.414228287, -1.22196762, -2.00650429, -2.71048011, -3.27653691, -3.6473165, -3.76546069, -3.57361131, -3.01441016, -1.90594282, -0.173057382, 1.95987002, 4.57688819, 7.76204593, 11.599392, 16.1729753, 21.5668446, 27.8650487, 35.1516364, 43.5106565, 53.0261578, 63.7821891, 75.6195686, 88.9547603, 103.944782, 120.746652, 138.832429, 158.70161, 181.036348, 206.199151, 234.552524, 266.458972, 302.281001, 343.816839, 389.556311, 439.581022, 493.972575, 552.812575, 612.456095, 677.642902, 749.386233, 828.699326, 916.595416, 1014.53237, 1122.53536, 1241.94295, 1373.54103, 1516.98715, 1672.44546, 1840.08009, 2019.54791, 2211.63566, 2417.68351, 2639.30617, 2875.29874, 3125.39276, 3389.27691, 3666.56725, 3956.91319, 4260.13017, 4575.91065, 4903.33821, 5241.84087, 5601.60816, 5969.65092, 6341.67198, 6713.37419, 7080.4604, 7456.92028, 7815.59814, 8147.62512, 8444.13236, 8696.251, 8838.25304, 8956.03484, 9065.09574, 9246.84779, 9490.69988, 9758.59727, 10041.322, 10329.6562, 10614.382, 10886.2813, 11502.0072, 11963.6241, 12201.8253, 12147.3043, 11730.7544, 10882.869, 9534.3414, 8222.84688, 7061.2156},
        {313.231616, 308.894762, 304.560647, 300.23024, 295.904513, 291.584434, 287.270975, 282.965106, 278.667797, 274.380018, 270.10274, 265.836932, 261.583565, 257.34361, 253.118036, 248.907814, 244.713914, 240.537307, 236.378962, 232.239849, 228.12094, 224.023204, 219.947612, 215.895134, 211.86674, 207.8634, 203.886085, 199.935765, 196.013409, 192.11999, 188.256476, 184.423838, 180.623046, 176.85507, 173.198211, 169.613147, 166.059628, 162.536115, 159.041067, 155.572947, 152.130214, 148.71133, 145.314756, 141.938951, 138.582377, 135.243495, 131.920765, 128.612648, 125.317606, 122.034097, 118.760584, 115.241322, 111.734486, 108.24405, 104.773984, 101.328261, 97.9108504, 94.525725, 91.176856, 87.8682147, 84.6037727, 81.3875014, 78.2233722, 75.1153566, 72.0674261, 69.0835521, 66.1677061, 63.4433509, 61.0310804, 58.6884241, 56.4116072, 54.1968551, 52.0403931, 49.9384464, 47.8872405, 45.8830006, 43.921952, 42.0003201, 40.1143301, 38.2602074, 36.4341773, 34.6324651, 32.8512961, 31.0868956, 29.0362221, 26.8514291, 24.6882668, 22.5526724, 20.450583, 18.3879356, 16.3706674, 14.4047154, 12.4960168, 10.6505086, 8.87412797, 7.17281198, 5.55249772, 4.0191223, 2.57862281, 1.23693634, 0, 0.63371165, 0.931701322, 0.946475069, 0.730538942, 0.336398995, -0.183438719, -0.776468149, -1.39018324, -1.97207794, -2.4696462, -2.83038196, -3.00177918, -2.93133179, -2.56653375, -1.854879, -0.743861496, 0.882487237, 3.08573931, 5.77485483, 9.01593895, 12.8750968, 17.4184335, 22.7120543, 28.8220642, 35.8145684, 43.7556721, 52.7114803, 62.7480983, 73.9316311, 86.02575, 99.5098864, 114.540876, 131.275554, 149.094975, 168.543865, 190.373825, 214.974423, 242.735228, 274.045811, 309.29574, 349.794445, 394.731678, 444.317035, 498.760113, 558.270509, 622.115905, 691.683291, 767.417739, 849.764326, 939.168126, 1034.86701, 1139.98873, 1256.9567, 1387.15119, 1529.82271, 1685.17797, 1853.42371, 2035.13519, 2230.06683, 2437.57098, 2656.58887, 2888.77542, 3136.26239, 3399.71939, 3677.33858, 3968.4493, 4272.57169, 4589.08457, 4916.6672, 5254.39453, 5610.61026, 5973.69467, 6339.51433, 6703.93581, 7062.82567, 7427.69981, 7774.86313, 8096.26986, 8383.87422, 8629.63046, 8773.31464, 8893.8446, 9005.53664, 9183.07641, 9416.88402, 9672.22565, 9940.7791, 10214.2222, 10484.2326, 10742.4882, 11034.8411, 11403.9024, 11709.1549, 11810.082, 11566.1665, 10836.8916, 9481.74043, 8285.8871, 7024.64893},
        {320.057931, 316.031379, 311.961697, 307.85159, 303.703764, 299.520922, 295.305769, 291.06101, 286.789349, 282.493492, 278.176143, 273.840006, 269.487786, 265.122188, 260.745916, 256.361675, 251.97217, 247.580105, 243.188186, 238.799115, 234.415599, 230.040342, 225.676049, 221.325424, 216.991171, 212.675996, 208.382603, 204.113697, 199.871982, 195.660163, 191.480945, 187.337033, 183.23113, 179.165942, 175.287184, 171.523047, 167.800004, 164.116118, 160.469453, 156.858072, 153.28004, 149.73342, 146.216275, 142.72667, 139.262668, 135.822332, 132.403727, 129.004916, 125.623962, 122.25893, 118.907883, 115.296292, 111.700724, 108.125149, 104.573543, 101.049877, 97.5581236, 94.1022571, 90.6862497, 87.3140745, 83.9897044, 80.7171123, 77.5002711, 74.3431538, 71.2497334, 68.2239827, 65.2698746, 62.5107489, 60.067079, 57.6966533, 55.3957064, 53.1604734, 50.9871891, 48.8720883, 46.811406, 44.8013771, 42.8382363, 40.9182186, 39.0375589, 37.192492, 35.3792529, 33.5940763, 31.8331972, 30.0928504, 28.0633331, 25.900284, 23.7630189, 21.6577011, 19.5904937, 17.5675598, 15.5950626, 13.6791652, 11.8260308, 10.0418225, 8.33270348, 6.70483686, 5.16438579, 3.71751342, 2.3703829, 1.12915738, 0, 0.713927611, 1.10723615, 1.23207454, 1.14059173, 0.884936664, 0.517258265, 0.0897054753, -0.345572768, -0.736427528, -1.03070987, -1.17627085, -1.12096154, -0.812632992, -0.199136279, 0.771677539, 2.1519574, 4.08944974, 6.64996029, 9.66940554, 13.2204196, 17.3756366, 22.2076906, 27.7892158, 34.1928463, 41.4912161, 49.7569595, 59.0627105, 69.4811032, 81.0847718, 93.731954, 107.788292, 123.39074, 140.676249, 158.940958, 178.742226, 200.861634, 225.688378, 253.611657, 285.020669, 320.304612, 360.390288, 404.965672, 454.314778, 508.72162, 568.470212, 633.854467, 705.146027, 782.626431, 866.577219, 957.27993, 1053.47584, 1158.8693, 1276.10341, 1406.63487, 1549.4981, 1704.81506, 1872.70772, 2053.38846, 2246.8683, 2453.05959, 2671.46739, 2903.22562, 3149.96575, 3412.21108, 3688.60718, 3978.66154, 4282.60498, 4600.13252, 4928.28692, 5265.6112, 5617.95523, 5975.43106, 6334.05241, 6689.83303, 7038.78663, 7389.88745, 7722.94859, 8030.74365, 8306.04624, 8541.62996, 8681.14158, 8799.23276, 8909.73145, 9085.74492, 9317.6364, 9571.06943, 9837.64223, 10108.953, 10376.6001, 10632.1815, 10817.4758, 11149.4389, 11469.0411, 11617.2526, 11435.0436, 10763.3844, 9443.24514, 8293.417, 7046.52456},
        {326.820878, 323.163392, 319.411544, 315.569958, 311.643257, 307.636065, 303.553006, 299.398704, 295.177782, 290.894863, 286.554572, 282.161532, 277.720367, 273.235701, 268.712157, 264.154359, 259.56693, 254.954495, 250.321676, 245.673098, 241.013384, 236.347159, 231.679045, 227.013666, 222.355646, 217.709609, 213.080178, 208.471977, 203.88963, 199.33776, 194.820991, 190.343946, 185.911251, 181.527527, 177.4105, 173.45376, 169.548337, 165.69194, 161.882276, 158.117054, 154.393982, 150.710767, 147.065119, 143.454745, 139.877353, 136.330651, 132.812348, 129.320152, 125.851771, 122.404913, 118.977286, 115.283282, 111.610067, 107.961492, 104.341405, 100.753657, 97.2020984, 93.690578, 90.2229462, 86.8030526, 83.4347474, 80.1218803, 76.8683012, 73.67786, 70.5544066, 67.5017908, 64.5238626, 61.7417634, 59.2738184, 56.8819726, 54.5624723, 52.3115638, 50.1254935, 48.0005077, 45.9328529, 43.9187752, 41.9545212, 40.036337, 38.1604691, 36.3231639, 34.5206676, 32.7492266, 31.0050873, 29.284496, 27.2669759, 25.1137969, 22.9900354, 20.902216, 18.8568633, 16.860502, 14.9196567, 13.040852, 11.2306126, 9.495463, 7.84192789, 6.27653188, 4.80579958, 3.43625562, 2.17442463, 1.02683121, 0, 1.07272616, 1.79475888, 2.22022576, 2.4032544, 2.3979724, 2.25850736, 2.03898686, 1.79353852, 1.57628992, 1.44136866, 1.44290235, 1.63501858, 2.07184494, 2.80750904, 3.89613847, 5.39186083, 7.53132599, 10.3947361, 13.6233709, 17.3104698, 21.5492723, 26.433018, 32.0549463, 38.5082968, 45.886309, 54.2822224, 63.7892765, 74.5007108, 86.5097649, 99.8728326, 114.733509, 131.196088, 149.364861, 168.435391, 188.966336, 211.758685, 237.18935, 265.635244, 297.473281, 333.080373, 373.088876, 417.542517, 466.76823, 521.09295, 580.84361, 646.31796, 717.879416, 795.862212, 880.600579, 972.428747, 1070.52168, 1177.78976, 1296.56949, 1428.10453, 1571.40741, 1726.49239, 1893.37372, 2071.39549, 2261.39445, 2464.93845, 2683.53999, 2916.30057, 3163.06806, 3423.713, 3698.14422, 3986.25297, 4289.38222, 4607.79958, 4936.44974, 5273.28831, 5621.44761, 5972.91017, 6323.83971, 6670.39992, 7008.75451, 7345.47453, 7663.7145, 7957.03627, 8219.00172, 8443.17271, 8576.21613, 8689.85213, 8797.74043, 8973.67998, 9207.52537, 9464.07319, 9734.43092, 10009.7061, 10281.0061, 10539.4386, 10728.2722, 11051.6784, 11359.459, 11501.4155, 11327.3496, 10687.0628, 9430.35688, 8293.42005, 7121.7122},
        {333.727898, 330.449427, 327.024546, 323.459817, 319.761803, 315.937068, 311.992173, 307.933683, 303.768159, 299.502166, 295.142265, 290.69502, 286.166993, 281.564748, 276.894847, 272.163854, 267.378331, 262.544841, 257.669947, 252.760213, 247.8222, 242.862472, 237.887592, 232.904123, 227.918627, 222.937668, 217.967809, 213.015612, 208.08764, 203.190456, 198.330624, 193.514706, 188.749264, 184.040863, 179.67889, 175.522546, 171.427634, 167.391538, 163.411643, 159.485334, 155.609996, 151.783012, 148.001769, 144.26365, 140.56604, 136.906324, 133.281887, 129.690114, 126.128388, 122.594095, 119.084619, 115.305625, 111.552541, 107.829077, 104.13894, 100.485841, 96.8734865, 93.3055863, 89.7858488, 86.3179826, 82.9056963, 79.5526986, 76.2626982, 73.0394037, 69.8865236, 66.8077668, 63.8068418, 61.0026279, 58.5112395, 56.098854, 53.7617142, 51.4960625, 49.2981417, 47.1641942, 45.0904628, 43.07319, 41.1086183, 39.1929904, 37.3225489, 35.4935363, 33.7021953, 31.9447685, 30.2174984, 28.5166277, 26.5092619, 24.3631352, 22.2499374, 20.1765924, 18.1500239, 16.1771555, 14.264911, 12.4202142, 10.6499887, 8.96115826, 7.36064663, 5.85537752, 4.45227465, 3.15826173, 1.98026251, 0.925200689, 0, 1.54595663, 2.69263213, 3.49706074, 4.01627668, 4.30731419, 4.42720749, 4.43299082, 4.38169841, 4.33036449, 4.33602328, 4.45570904, 4.74645597, 5.26529832, 6.06927031, 7.21540618, 8.76074016, 11.0550313, 14.1971318, 17.58196, 21.3292769, 25.5588435, 30.3904209, 35.9437699, 42.3386516, 49.6948269, 58.1320569, 67.7701026, 78.7287249, 91.1276848, 105.271316, 121.02713, 138.459517, 157.632864, 177.611005, 198.958606, 222.507151, 248.621195, 277.665297, 310.00401, 346.001893, 386.089995, 430.54614, 479.721877, 533.968751, 593.638308, 658.603061, 729.813349, 807.740478, 892.855755, 985.630483, 1086.0449, 1195.66158, 1316.21319, 1448.60219, 1592.03601, 1746.48313, 1911.91199, 2086.89435, 2273.11282, 2473.7737, 2692.47838, 2926.19922, 3172.93814, 3431.8749, 3704.18498, 3990.12781, 4292.09379, 4610.89481, 4939.52913, 5275.41483, 5619.13824, 5964.46178, 6307.70684, 6645.19481, 6973.24705, 7296.32553, 7600.57593, 7880.28449, 8129.73746, 8343.22108, 8470.08864, 8579.51464, 8685.04233, 8861.31228, 9097.68335, 9358.05717, 9633.06547, 9913.33995, 10189.5124, 10452.2144, 10680.2789, 11004.5148, 11293.0749, 11414.1121, 11235.7793, 10626.2293, 9453.61507, 8323.27576, 7247.61169},
        {341.193481, 338.248103, 335.107054, 331.778764, 328.271661, 324.594173, 320.75473, 316.76176, 312.623691, 308.348952, 303.945971, 299.423178, 294.789001, 290.051868, 285.220208, 280.302449, 275.307021, 270.242352, 265.116869, 259.939003, 254.717181, 249.459832, 244.175385, 238.872268, 233.55891, 228.243739, 222.935185, 217.641675, 212.371638, 207.133503, 201.935698, 196.786652, 191.694794, 186.668552, 182.070713, 177.725074, 173.45117, 169.245931, 165.106286, 161.029163, 157.011493, 153.050203, 149.142223, 145.284481, 141.473908, 137.707431, 133.981979, 130.294483, 126.64187, 123.021069, 119.42901, 115.547505, 111.695431, 107.876547, 104.094614, 100.353392, 96.6566406, 93.0081206, 89.4115917, 85.8708141, 82.3895479, 78.9715531, 75.6205898, 72.3404181, 69.1347982, 66.00749, 62.9622537, 60.1198161, 57.5980954, 55.159617, 52.8005584, 50.5170972, 48.305411, 46.1616776, 44.0820744, 42.0627791, 40.0999693, 38.1898227, 36.3285168, 34.5122294, 32.7371379, 30.99942, 29.2952534, 27.6208156, 25.6313255, 23.5006117, 21.4066009, 19.3565355, 17.3576578, 15.4172103, 13.5424353, 11.7405753, 10.0188725, 8.38456948, 6.8449085, 5.40713199, 4.07848234, 2.86620193, 1.77753314, 0.819718371, 0, 2.1491446, 3.82345936, 5.08479237, 5.99499169, 6.61590538, 7.00938151, 7.23726815, 7.36141337, 7.44366523, 7.5458718, 7.72988115, 8.05754134, 8.59070045, 9.39120653, 10.5209077, 12.0416519, 14.3977033, 17.7055407, 21.1620238, 24.9109469, 29.0961043, 33.8612904, 39.3502996, 45.7069261, 53.0749643, 61.5982086, 71.4204532, 82.6854926, 95.5371211, 110.552122, 127.282538, 145.742267, 165.945206, 186.694302, 208.608927, 232.631376, 259.138831, 288.508477, 321.117495, 357.343069, 397.695727, 442.378724, 491.743154, 546.14011, 605.920687, 670.416525, 741.253034, 819.036172, 904.371895, 997.866161, 1100.34228, 1211.9232, 1333.31784, 1465.01481, 1607.05288, 1759.6728, 1923.11531, 2096.48305, 2281.41352, 2480.78581, 2698.18419, 2930.58673, 3175.9818, 3433.54869, 3704.4846, 3989.0605, 4289.91535, 4607.93406, 4935.31871, 5269.18287, 5608.47096, 5948.18405, 6284.80193, 6614.80444, 6934.67138, 7247.41487, 7541.34936, 7811.32157, 8052.17824, 8258.76612, 8383.19462, 8491.53935, 8596.96303, 8772.37921, 9007.29926, 9266.33835, 9540.25265, 9819.79833, 10095.7315, 10358.8085, 10589.2985, 10904.1524, 11183.662, 11308.1188, 11157.8146, 10613.041, 9554.0898, 8443.05255, 7421.75671},
        {349.495856, 346.794745, 343.854662, 340.68573, 337.298076, 333.701823, 329.907098, 325.924024, 321.762727, 317.433332, 312.945963, 308.310746, 303.537806, 298.637267, 293.619255, 288.493894, 283.27131, 277.961627, 272.57497, 267.121464, 261.611235, 256.054406, 250.461104, 244.841453, 239.205577, 233.563603, 227.925654, 222.301856, 216.702334, 211.137212, 205.616617, 200.150671, 194.749501, 189.423232, 184.609383, 180.095392, 175.66356, 171.310143, 167.031394, 162.82357, 158.682926, 154.605716, 150.588197, 146.626622, 142.717248, 138.856329, 135.04012, 131.264877, 127.526855, 123.822309, 120.147494, 116.134879, 112.152391, 108.204173, 104.294365, 100.427108, 96.6065445, 92.8368152, 89.1220616, 85.4664249, 81.8740466, 78.3490679, 74.8956302, 71.5178749, 68.2199432, 65.0059765, 61.8801161, 58.9717642, 56.4073154, 53.9325701, 51.5435494, 49.2362746, 47.0067668, 44.8510473, 42.7651372, 40.7450577, 38.78683, 36.8864754, 35.0400149, 33.2434697, 31.4928612, 29.7842104, 28.1135385, 26.4768668, 24.5197305, 22.4207659, 20.3628215, 18.3532925, 16.399574, 14.5090611, 12.6891491, 10.947233, 9.29070786, 7.72696893, 6.2634113, 4.9074301, 3.66642045, 2.54777749, 1.55889634, 0.707172134, 0, 2.8644905, 5.1509471, 6.92829563, 8.26546194, 9.23137188, 9.89495127, 10.325126, 10.5908218, 10.7609647, 10.9044803, 11.0902947, 11.3873335, 11.8645227, 12.5907881, 13.6350555, 15.0662508, 17.372099, 20.6813543, 24.1156589, 27.8336811, 31.9940897, 36.755553, 42.2767395, 48.7163179, 56.2329565, 64.985324, 75.1320889, 86.8319196, 100.243485, 116.211664, 133.957305, 153.43321, 174.592186, 195.793908, 217.787747, 241.747904, 268.105125, 297.290153, 329.73373, 365.866601, 406.674609, 451.864454, 501.758274, 556.678207, 616.946389, 681.492591, 752.37941, 830.277073, 915.855811, 1009.78585, 1113.497, 1225.97154, 1346.44437, 1474.9761, 1613.31323, 1762.44534, 1923.36199, 2097.7855, 2285.80617, 2486.7149, 2700.49982, 2927.78418, 3169.65498, 3426.47286, 3697.36763, 3982.03405, 4282.05736, 4597.62245, 4921.98251, 5252.31168, 5587.36366, 5922.49005, 6254.3182, 6579.47552, 6894.58937, 7202.10757, 7491.38196, 7757.5848, 7995.88836, 8201.4649, 8329.5166, 8441.83256, 8550.7151, 8722.90254, 8949.14962, 9197.52939, 9459.93066, 9728.24225, 9994.35296, 10250.1516, 10394.8312, 10676.3131, 10969.4972, 11149.2837, 11090.5724, 10668.2633, 9757.25638, 8694.09042, 7643.44218},
        {358.020359, 355.516792, 352.737205, 349.693141, 346.396143, 342.857756, 339.089523, 335.102987, 330.909692, 326.521182, 321.948999, 317.204687, 312.299791, 307.245853, 302.054417, 296.737027, 291.305226, 285.770557, 280.144564, 274.438791, 268.664782, 262.834079, 256.958226, 251.048767, 245.117245, 239.175204, 233.234187, 227.305738, 221.401401, 215.532718, 209.711234, 203.948492, 198.256035, 192.645408, 187.61353, 182.919047, 178.314772, 173.796496, 169.360012, 165.001111, 160.715585, 156.499228, 152.34783, 148.257184, 144.223083, 140.241317, 136.30768, 132.417963, 128.567958, 124.753458, 120.970255, 116.812161, 112.685661, 108.595263, 104.545472, 100.540795, 96.5857385, 92.6848078, 88.8425095, 85.06335, 81.3518353, 77.7124719, 74.1497658, 70.6682234, 67.2723509, 63.9666546, 60.7556406, 57.7774008, 55.1668199, 52.6528937, 50.2314685, 47.8983909, 45.6495073, 43.480664, 41.3877075, 39.3664843, 37.4128407, 35.5226232, 33.6916782, 31.9158522, 30.1909915, 28.5129426, 26.8775519, 25.2806658, 23.3587213, 21.2948293, 19.2764203, 17.3110045, 15.4060917, 13.5691919, 11.8078151, 10.1294712, 8.54167014, 7.05192193, 5.66773651, 4.39662382, 3.24609384, 2.22365651, 1.33682179, 0.593099632, 0, 3.45582915, 6.25287476, 8.46568217, 10.1687967, 11.4367638, 12.3441287, 12.9654368, 13.3752335, 13.648064, 13.8584737, 14.0810081, 14.3902123, 14.8606319, 15.566812, 16.5832981, 17.9846355, 20.2864804, 23.6263938, 27.0811707, 30.81988, 35.0115908, 39.8253724, 45.4302936, 51.9954236, 59.6898315, 68.6825864, 79.1427574, 91.2394135, 105.141624, 121.905642, 140.488051, 160.791766, 182.7197, 204.246052, 226.238094, 250.07742, 276.245558, 305.224034, 337.494378, 373.538117, 414.851624, 460.592715, 511.044363, 566.489537, 627.211211, 691.745593, 762.559106, 840.371413, 925.902173, 1019.87105, 1124.19294, 1236.93142, 1355.97255, 1481.01285, 1615.44522, 1761.00295, 1919.41936, 2095.1477, 2286.58313, 2489.15363, 2698.8795, 2920.4106, 3158.29461, 3414.19527, 3684.88973, 3969.39814, 4267.95811, 4579.90543, 4900.112, 5225.97478, 5556.84059, 5887.85648, 6215.74443, 6537.22641, 6849.0244, 7153.22297, 7439.84083, 7704.25931, 7941.85972, 8148.02338, 8280.75321, 8397.72786, 8510.34797, 8678.34674, 8893.99315, 9129.41764, 9377.89612, 9632.70451, 9887.11875, 10134.4147, 10191.6587, 10440.1328, 10749.4212, 10989.1083, 11028.7783, 10738.0155, 9986.40422, 8995.00995, 7923.50376},
        {366.595695, 364.24284, 361.584896, 358.634446, 355.40407, 351.906351, 348.153873, 344.159215, 339.934962, 335.493695, 330.847996, 326.010447, 320.993631, 315.81013, 310.472525, 304.9934, 299.385336, 293.660916, 287.832722, 281.913335, 275.915339, 269.851315, 263.733845, 257.575512, 251.388898, 245.186585, 238.981155, 232.785191, 226.611274, 220.471987, 214.379912, 208.347631, 202.387727, 196.512781, 191.255193, 186.356703, 181.552803, 176.839206, 172.211626, 167.665775, 163.197366, 158.802112, 154.475728, 150.213925, 146.012418, 141.866918, 137.773139, 133.726795, 129.723599, 125.759262, 121.8295, 117.51322, 113.231975, 108.990514, 104.793585, 100.645938, 96.5523214, 92.5174828, 88.5461715, 84.6431359, 80.8131248, 77.0608867, 73.3911702, 69.8087241, 66.3182968, 62.9246372, 59.6324937, 56.586582, 53.9282575, 51.3735969, 48.9182756, 46.5579686, 44.288351, 42.105098, 40.0038847, 37.9803862, 36.0302777, 34.1492343, 32.3329311, 30.5770433, 28.8772459, 27.2292142, 25.6286232, 24.0711481, 22.1847281, 20.1566411, 18.178585, 16.2581686, 14.4030007, 12.6206903, 10.9188461, 9.30507703, 7.786992, 6.37219985, 5.06830944, 3.88292964, 2.82366931, 1.89813731, 1.11394252, 0.478693793, 0, 3.79542549, 6.89865509, 9.38669857, 11.3365657, 12.8252663, 13.9298102, 14.7272071, 15.2944668, 15.7085991, 16.0466139, 16.3855208, 16.8023296, 17.3740502, 18.1776924, 19.2902658, 20.7887804, 23.203065, 26.6748177, 30.2568148, 34.1230988, 38.4477119, 43.4046968, 49.1680958, 55.9119514, 63.810306, 73.0372021, 83.7666821, 96.1727885, 110.429564, 127.707213, 146.818356, 167.638187, 190.041899, 211.770497, 233.766269, 257.540619, 283.609, 312.48686, 344.68965, 380.732821, 422.534067, 468.779825, 519.711192, 575.569266, 636.595146, 701.000436, 771.563101, 849.031613, 934.154441, 1027.68006, 1131.8088, 1244.06277, 1361.37902, 1483.22217, 1614.2173, 1756.67255, 1912.89606, 2089.57775, 2283.64811, 2487.25752, 2692.97021, 2908.95089, 3142.82083, 3397.37912, 3667.25388, 3950.82385, 4246.94278, 4554.11254, 4869.09331, 5189.63045, 5516.3835, 5843.73535, 6168.42263, 6487.18198, 6796.75001, 7098.81689, 7383.92734, 7647.57961, 7885.27194, 8092.50258, 8229.59929, 8350.67779, 8466.30941, 8629.28065, 8833.45265, 9055.09686, 9288.9539, 9529.76438, 9772.26892, 10011.2081, 9997.4574, 10218.7073, 10542.5462, 10836.5628, 10968.3455, 10805.4829, 10215.5634, 9325.36841, 8267.0465},
        {374.999505, 372.748576, 370.17389, 367.288547, 364.105648, 360.638292, 356.899582, 352.902616, 348.660497, 344.186323, 339.493197, 334.594218, 329.502487, 324.231105, 318.793172, 313.201789, 307.470056, 301.611074, 295.637944, 289.563766, 283.40164, 277.164668, 270.865949, 264.518584, 258.135675, 251.730321, 245.315623, 238.904681, 232.510597, 226.146471, 219.825402, 213.560493, 207.364843, 201.251553, 195.753038, 190.611818, 185.564172, 180.606347, 175.734592, 170.945154, 166.234281, 161.598221, 157.033222, 152.535532, 148.101398, 143.727068, 139.40879, 135.142813, 130.925383, 126.752748, 122.621157, 118.135269, 113.691409, 109.294313, 104.948719, 100.659363, 96.4309816, 92.2683115, 88.1760895, 84.1590523, 80.2219365, 76.369479, 72.6064163, 68.9374852, 65.3674223, 61.9009644, 58.5428481, 55.4399205, 52.7356167, 50.1415813, 47.6533387, 45.266413, 42.9763287, 40.7786098, 38.6687807, 36.6423656, 34.6948888, 32.8218745, 31.018847, 29.2813306, 27.6048495, 25.984928, 24.4170903, 22.8968606, 21.0438634, 19.0494792, 17.1096071, 15.2319699, 13.4242905, 11.6942919, 10.0496971, 8.49822894, 7.04761039, 5.70556437, 4.47981383, 3.3780817, 2.40809092, 1.57756443, 0.894225149, 0.365796029, 0, 3.69625901, 6.75040841, 9.23631622, 11.2278504, 12.7988791, 14.0232702, 14.9748917, 15.7276117, 16.3552982, 16.9318192, 17.5310427, 18.2268367, 19.0930693, 20.2036084, 21.6323221, 23.4530783, 26.1975561, 30.0050249, 33.9138281, 38.0959361, 42.7233192, 47.9679478, 54.0017922, 60.9968227, 69.1250098, 78.5583238, 89.4687349, 102.028214, 116.40873, 133.736287, 152.879011, 173.722662, 196.153002, 217.965355, 240.0907, 264.017466, 290.258543, 319.326823, 351.735197, 387.996556, 430.194006, 476.792333, 527.997214, 584.014324, 645.049339, 709.132235, 779.188313, 855.967174, 940.218418, 1032.69165, 1135.56967, 1246.41718, 1361.97999, 1481.71353, 1610.60534, 1751.15918, 1905.87879, 2082.42428, 2276.97111, 2480.06965, 2682.45718, 2894.17232, 3124.58259, 3377.08799, 3644.97681, 3926.17881, 4218.43219, 4519.61703, 4828.31568, 5142.71326, 5465.49751, 5789.65819, 6111.82182, 6428.6149, 6736.66395, 7036.85515, 7320.49042, 7583.13133, 7820.33949, 8027.67648, 8167.12477, 8290.21115, 8406.88824, 8564.1186, 8757.20511, 8966.02984, 9186.66673, 9415.18968, 9647.67262, 9880.18946, 9833.90777, 10040.4672, 10372.7354, 10703.5803, 10905.8695, 10852.4708, 10416.252, 9665.74374, 8688.59914},
        {382.516148, 380.298553, 377.758106, 374.907398, 371.759018, 368.325554, 364.619595, 360.653732, 356.440552, 351.992645, 347.3226, 342.443007, 337.366453, 332.10553, 326.672824, 321.080926, 315.342425, 309.46991, 303.47597, 297.373194, 291.174171, 284.89149, 278.537741, 272.125512, 265.667392, 259.175972, 252.663839, 246.143583, 239.627793, 233.129059, 226.659968, 220.233111, 213.861076, 207.556453, 201.776153, 196.301257, 190.907508, 185.593077, 180.356132, 175.194845, 170.107385, 165.091922, 160.146626, 155.269667, 150.459216, 145.713441, 141.030514, 136.408604, 131.845881, 127.340516, 122.890678, 118.221398, 113.609907, 109.060295, 104.576654, 100.163075, 95.8236489, 91.5624665, 87.3836193, 83.2911983, 79.2892946, 75.3819994, 71.5734039, 67.867599, 64.2686759, 60.7807258, 57.4078398, 54.2887333, 51.5588893, 48.9447466, 46.4416691, 44.0450208, 41.7501656, 39.5524677, 37.4472908, 35.4299991, 33.4959565, 31.6405269, 29.8590744, 28.1469628, 26.4995563, 24.9122188, 23.3803142, 21.8992065, 20.0757769, 18.107802, 16.1977276, 14.3535248, 12.5831645, 10.8946177, 9.29585551, 7.79484878, 6.39956855, 5.11798582, 3.95807159, 2.92779683, 2.03513255, 1.28804973, 0.694519372, 0.262512465, 0, 2.39858222, 4.43375728, 6.16090664, 7.63541178, 8.91265416, 10.0480152, 11.0968765, 12.1146194, 13.1566253, 14.2782759, 15.5349524, 16.9820365, 18.6749095, 20.6689529, 23.0195482, 25.7820768, 29.4759369, 34.2227979, 39.02848, 44.0477968, 49.4355618, 55.3465887, 61.9356911, 69.3576824, 77.7673764, 87.3195867, 98.1691267, 110.47081, 124.379451, 140.564408, 158.477282, 178.118524, 199.488582, 221.253515, 244.080969, 268.994426, 296.394655, 326.682421, 360.258493, 397.523637, 439.895282, 486.448115, 537.383991, 592.904766, 653.212293, 716.501141, 785.482273, 860.859369, 943.336105, 1033.61616, 1133.2168, 1241.03372, 1355.54201, 1476.71559, 1607.5885, 1749.82083, 1905.0727, 2078.32535, 2267.16293, 2465.54655, 2667.39542, 2879.5705, 3109.07113, 3358.25353, 3621.6064, 3897.22978, 4182.77139, 4476.21402, 4777.19879, 5084.42877, 5403.91311, 5725.92477, 6046.63778, 6362.22618, 6668.86399, 6964.43395, 7242.97421, 7500.23163, 7731.95306, 7933.88535, 8068.7023, 8187.27253, 8299.51706, 8450.4611, 8636.13025, 8837.9234, 9052.59749, 9276.90949, 9507.61633, 9741.47498, 9761.37372, 9985.38258, 10309.7484, 10630.7182, 10844.5387, 10847.4567, 10535.7193, 10006.5739, 9293.7277},
        {390.157831, 387.947698, 385.420719, 382.588724, 379.463547, 376.057018, 372.380968, 368.44723, 364.267634, 359.854012, 355.218196, 350.372016, 345.327305, 340.095894, 334.689614, 329.120296, 323.399773, 317.539876, 311.552436, 305.449284, 299.242252, 292.943172, 286.563875, 280.116192, 273.611955, 267.062995, 260.481145, 253.878234, 247.266095, 240.656559, 234.061458, 227.492623, 220.961885, 214.481076, 208.415907, 202.593825, 196.838026, 191.148858, 185.526669, 179.971804, 174.484613, 169.065442, 163.714639, 158.432551, 153.219525, 148.07591, 143.002052, 137.998299, 133.064998, 128.202497, 123.411144, 118.555101, 113.773853, 109.070699, 104.448938, 99.91187, 95.4627945, 91.105011, 86.8418193, 82.6765188, 78.6124093, 74.6527903, 70.8009613, 67.060222, 63.433872, 59.9252108, 56.5375381, 53.3971173, 50.6272595, 47.9776606, 45.4436489, 43.0205529, 40.7037009, 38.4884211, 36.3700421, 34.3438921, 32.4052994, 30.5495925, 28.7720996, 27.0681492, 25.4330696, 23.8621891, 22.350836, 20.8943388, 19.0924818, 17.1466851, 15.2631229, 13.4499596, 11.7153599, 10.0674882, 8.51450904, 7.06458693, 5.72588639, 4.50657193, 3.41480806, 2.45875929, 1.64659014, 0.986465127, 0.486548756, 0.155005542, 0, 1.14876368, 2.2049108, 3.2057789, 4.1887055, 5.19102813, 6.25008432, 7.4032116, 8.68774749, 10.1410295, 11.8003953, 13.7031822, 15.8867278, 18.3883697, 21.2454454, 24.4952925, 28.1752483, 32.7878536, 38.4369034, 44.1074903, 49.9366382, 56.0613709, 62.6187124, 69.7456866, 77.5793173, 86.2566285, 95.914644, 106.690388, 118.720884, 132.143155, 147.181159, 163.853111, 182.269957, 202.54264, 224.174371, 247.579962, 273.336286, 301.736608, 333.074193, 367.642306, 405.73421, 448.126154, 494.481424, 544.998789, 599.877017, 659.314876, 721.768348, 789.614683, 863.488347, 944.023802, 1031.85552, 1127.68904, 1232.00086, 1345.25968, 1468.03942, 1601.12882, 1745.22015, 1901.00571, 2070.39695, 2252.5899, 2445.4506, 2646.5448, 2859.08447, 3087.26826, 3332.25859, 3590.07323, 3859.09142, 4137.06816, 4422.22084, 4715.05584, 5014.78477, 5330.57131, 5650.1439, 5969.2692, 6283.71387, 6589.24457, 6881.69552, 7156.74893, 7410.15455, 7637.66217, 7835.02154, 7964.68814, 8077.90222, 8184.68255, 8328.81412, 8506.94988, 8701.6735, 8910.34141, 9130.31001, 9358.93571, 9593.57493, 9704.72122, 9952.8942, 10262.5814, 10558.2702, 10764.4481, 10805.6028, 10606.2215, 10303.7597, 9869.11787},
        {398.27571, 396.065984, 393.545635, 390.725739, 387.617374, 384.231618, 380.579548, 376.672242, 372.520777, 368.136231, 363.529681, 358.712206, 353.694882, 348.488787, 343.104999, 337.554595, 331.848653, 325.99825, 320.014464, 313.908372, 307.691053, 301.373583, 294.96704, 288.482502, 281.931046, 275.32375, 268.671691, 261.985947, 255.277596, 248.557714, 241.83738, 235.127672, 228.439666, 221.78444, 215.438062, 209.273542, 203.160701, 197.102019, 191.099974, 185.157044, 179.275707, 173.458441, 167.707724, 162.026034, 156.41585, 150.87965, 145.41991, 140.039111, 134.73973, 129.524244, 124.395133, 119.356934, 114.410021, 109.556827, 104.799785, 100.14133, 95.5838936, 91.1299103, 86.7818135, 82.5420363, 78.4130124, 74.3971752, 70.4969581, 66.7147945, 63.053118, 59.5143619, 56.1009598, 52.9219879, 50.0839621, 47.369373, 44.7737407, 42.2925854, 39.9214275, 37.655787, 35.4911843, 33.4231395, 31.4471728, 29.5588046, 27.753555, 26.0269442, 24.3744924, 22.79172, 21.274147, 19.8172937, 18.0247493, 16.0962432, 14.2362157, 12.4529059, 10.754553, 9.14939612, 7.64567434, 6.25162684, 4.97549277, 3.82551125, 2.80992142, 1.93696244, 1.21487344, 0.651893552, 0.256261924, 0.036217693, 0, 0.349974576, 0.794877598, 1.3594334, 2.0683663, 2.94640063, 4.01826071, 5.30867089, 6.84235548, 8.64403881, 10.7384452, 13.150299, 15.9043245, 19.0252461, 22.537788, 26.4666747, 30.8366304, 36.1383702, 42.4631885, 48.7817831, 55.2187333, 61.8986186, 68.9460186, 76.4855128, 84.6416806, 93.5391015, 103.302355, 114.056021, 125.924678, 139.032906, 153.324144, 169.170531, 186.750987, 206.244435, 227.657099, 251.25425, 277.34721, 306.166707, 337.943473, 372.908237, 411.291731, 453.551447, 499.622338, 549.690767, 603.943096, 662.56569, 724.327586, 791.186805, 863.684041, 942.359989, 1027.75534, 1119.71037, 1220.32228, 1331.94452, 1455.73427, 1590.40632, 1735.77202, 1891.64275, 2057.24153, 2233.10178, 2420.39871, 2619.69824, 2831.21195, 3056.7785, 3296.58455, 3548.01749, 3809.74959, 4079.71157, 4356.38345, 4640.96445, 4933.11566, 5244.45109, 5560.84001, 5877.80595, 6190.87245, 6495.56304, 6787.58164, 7062.22633, 7314.97555, 7541.30775, 7736.70139, 7862.45712, 7971.01637, 8072.50523, 8209.39401, 8379.00587, 8565.18718, 8765.93492, 8979.24609, 9203.11766, 9435.54663, 9653.28402, 9928.31295, 10216.5132, 10473.7646, 10655.9468, 10718.9397, 10618.6229, 10524.6129, 10334.7894},
        {407.835532, 405.660291, 403.170088, 400.375783, 397.288234, 393.918299, 390.276837, 386.374707, 382.222767, 377.831877, 373.212894, 368.376678, 363.334086, 358.095978, 352.673213, 347.076648, 341.317143, 335.405556, 329.352746, 323.169571, 316.86689, 310.455562, 303.946446, 297.350399, 290.678281, 283.940951, 277.149266, 270.314086, 263.446269, 256.556674, 249.656159, 242.755583, 235.865805, 228.997684, 222.372002, 215.890242, 209.451359, 203.059401, 196.718413, 190.432442, 184.205533, 178.041734, 171.94509, 165.919647, 159.969452, 154.09855, 148.310989, 142.610813, 137.002069, 131.488804, 126.075064, 120.883089, 115.796168, 110.815786, 105.943427, 101.180574, 96.5287111, 91.9893227, 87.5638926, 83.2539046, 79.0608427, 74.9861908, 71.0314328, 67.1980527, 63.4875345, 59.901362, 56.4410191, 53.1888019, 50.2250661, 47.3846267, 44.6637291, 42.0586185, 39.5655401, 37.1807391, 34.9004607, 32.7209501, 30.6384526, 28.6492134, 26.7494777, 24.9354907, 23.2034976, 21.5497437, 19.9704741, 18.4619342, 16.6565647, 14.7401652, 12.9029078, 11.152844, 9.49802514, 7.94650252, 6.50632755, 5.1855516, 3.99222604, 2.93440224, 2.02013157, 1.2574654, 0.654455101, 0.219152041, -0.04039241, -0.11612688, 0, 0.771629732, 1.60107475, 2.51520224, 3.54087935, 4.70497327, 6.03435116, 7.55588019, 9.29642754, 11.2828604, 13.5420459, 16.1008512, 18.9861435, 22.22479, 25.8436578, 29.8696141, 34.3295261, 39.7411645, 46.2015255, 52.6270993, 59.1499467, 65.9021283, 73.0157051, 80.6227378, 88.8552871, 97.845414, 107.725179, 118.626643, 130.681867, 144.022912, 158.823518, 175.158784, 193.148265, 212.91152, 233.933106, 256.650079, 281.668828, 309.299409, 339.851878, 373.636291, 410.962705, 453.052061, 499.026304, 549.017272, 603.156805, 661.576741, 723.347049, 789.926907, 861.713623, 939.104502, 1022.49685, 1110.81498, 1207.72953, 1316.93267, 1440.0607, 1574.55247, 1719.7314, 1874.9209, 2038.31149, 2210.61696, 2393.78701, 2588.65488, 2795.19289, 3014.96316, 3248.09232, 3492.27461, 3746.32078, 4008.13554, 4276.29474, 4552.69639, 4837.35928, 5143.03879, 5455.05769, 5769.02599, 6080.55371, 6385.25083, 6681.95657, 6962.24443, 7220.91711, 7452.77731, 7652.62774, 7780.01381, 7888.50037, 7988.0004, 8117.34925, 8275.01146, 8447.40055, 8633.58359, 8832.62765, 9043.59982, 9265.56716, 9606.15099, 9909.84235, 10167.453, 10369.7946, 10507.679, 10571.9179, 10553.323, 10615.8881, 10549.3541},
        {417.246149, 415.08776, 412.61327, 409.833213, 406.758124, 403.398541, 399.764997, 395.86803, 391.718175, 387.325967, 382.701942, 377.856635, 372.800584, 367.544322, 362.098386, 356.473312, 350.679635, 344.727892, 338.628616, 332.392345, 326.029614, 319.550959, 312.966915, 306.288018, 299.524804, 292.687809, 285.787567, 278.834616, 271.83949, 264.812725, 257.764857, 250.706421, 243.647954, 236.599991, 229.730523, 222.968046, 216.239164, 209.549302, 202.903888, 196.308346, 189.768102, 183.288584, 176.875216, 170.533424, 164.268636, 158.086275, 151.99177, 145.990544, 140.088025, 134.289639, 128.600811, 123.241494, 117.997936, 112.870914, 107.861203, 102.969577, 98.1968134, 93.5436866, 89.0109722, 84.5994458, 80.3098828, 76.1430586, 72.0997486, 68.1807284, 64.3867733, 60.7186589, 57.1771605, 53.8218284, 50.7108016, 47.7236371, 44.8573001, 42.108756, 39.4749702, 36.9529078, 34.5395343, 32.231815, 30.0267151, 27.9212001, 25.9122352, 23.9967858, 22.1718171, 20.4342945, 18.7811833, 17.2094489, 15.3825897, 13.4713184, 11.6503559, 9.92748908, 8.31050498, 6.80719054, 5.42533271, 4.17271846, 3.05713476, 2.08636855, 1.26820681, 0.610436494, 0.120844569, -0.192782006, -0.322656267, -0.260991252, 0, 1.65945275, 3.248446, 4.80202703, 6.35524311, 7.94314153, 9.60076954, 11.3631744, 13.2654035, 15.342504, 17.6295232, 20.1615083, 22.9735068, 26.1005657, 29.5777325, 33.4400544, 37.7225786, 42.9420137, 49.2022158, 55.4487601, 61.8199072, 68.4539174, 75.489051, 83.0635686, 91.3157305, 100.383797, 110.406029, 121.520686, 133.866029, 147.580319, 163.124398, 180.195665, 198.835603, 219.085698, 239.816102, 261.653967, 285.538801, 311.86349, 341.020921, 373.403978, 409.405547, 450.930603, 496.399742, 545.910006, 599.558439, 657.442083, 718.933238, 785.034874, 856.025218, 932.182499, 1013.78494, 1099.16642, 1192.92596, 1299.37758, 1420.40014, 1552.90072, 1696.01861, 1848.89311, 2009.52323, 2178.44773, 2357.44933, 2546.86742, 2747.17928, 2960.19546, 3186.22417, 3423.02805, 3669.5382, 3923.74783, 4184.34487, 4453.45621, 4731.26347, 5030.03067, 5335.99819, 5645.16531, 5953.53127, 6257.09533, 6560.84678, 6849.54732, 7116.94871, 7356.80269, 7562.86102, 7691.13973, 7798.28342, 7894.21628, 8015.70583, 8162.05078, 8321.36513, 8493.50685, 8678.33389, 8875.70419, 9085.47573, 9521.8265, 9842.04865, 10067.0859, 10217.8818, 10315.3802, 10380.5248, 10434.2591, 10609.8637, 10626.9007},
        {425.6531, 423.462308, 420.964323, 418.168958, 415.086028, 411.725348, 408.096732, 404.209994, 400.074949, 395.70141, 391.099193, 386.278111, 381.247979, 376.018612, 370.599823, 365.001427, 359.233239, 353.305072, 347.226742, 341.008062, 334.658846, 328.18891, 321.608068, 314.926133, 308.15292, 301.298244, 294.371919, 287.383759, 280.343579, 273.261192, 266.146414, 259.009059, 251.858941, 244.705874, 237.655951, 230.668822, 223.70298, 216.765116, 209.861918, 203.000077, 196.186283, 189.427226, 182.729595, 176.100081, 169.545374, 163.072163, 156.687139, 150.396992, 144.20841, 138.128085, 132.162707, 126.602705, 121.164879, 115.849768, 110.657909, 105.589842, 100.646107, 95.8272418, 91.1337856, 86.5662775, 82.1252564, 77.8112614, 73.6248314, 69.5665054, 65.6368224, 61.8363214, 58.1655414, 54.6729732, 51.4059565, 48.266133, 45.2509332, 42.3577875, 39.5841263, 36.9273802, 34.3849796, 31.9543549, 29.6329366, 27.4181552, 25.307441, 23.2982247, 21.3879365, 19.574007, 17.8538667, 16.2249459, 14.3737229, 12.4596446, 10.6458957, 8.9399975, 7.34947149, 5.88183912, 4.54462186, 3.34534115, 2.29151846, 1.39067524, 0.650332936, 0.0780130122, -0.318763078, -0.532473879, -0.555597937, -0.380613796, 0, 2.9313388, 5.58410728, 8.00690558, 10.2483339, 12.3569923, 14.381481, 16.3704002, 18.37235, 20.4359305, 22.609742, 24.9423845, 27.4824583, 30.2785634, 33.3793001, 36.8332685, 40.6890688, 45.3842983, 51.0231282, 56.774284, 62.7697224, 69.1414002, 76.021274, 83.5413005, 91.8334364, 101.029639, 111.261863, 122.662068, 135.362209, 149.494242, 165.66496, 183.357378, 202.561003, 223.265341, 244.055575, 265.623374, 289.034891, 314.700929, 343.032291, 374.439781, 409.334201, 449.594228, 493.716047, 541.82327, 594.039507, 650.488369, 710.853724, 775.808862, 845.587327, 920.422667, 1000.54843, 1084.38562, 1176.19564, 1279.73126, 1396.69114, 1524.58016, 1662.78613, 1810.69684, 1967.06826, 2132.06363, 2306.53547, 2490.02007, 2683.75334, 2889.90554, 3109.03768, 3338.98402, 3578.83043, 3827.00156, 4082.41188, 4346.40028, 4618.93427, 4908.97726, 5206.11684, 5507.20652, 5809.09981, 6108.65024, 6421.52606, 6721.06236, 6999.40897, 7248.71572, 7461.13244, 7585.8355, 7686.59716, 7773.98444, 7887.06004, 8025.11145, 8175.55296, 8338.26569, 8513.1308, 8700.02943, 8898.84272, 9369.19044, 9680.72781, 9873.95394, 9989.36797, 10067.469, 10148.7562, 10273.7287, 10504.044, 10586.7837},
        {432.041947, 429.727602, 427.133522, 424.268039, 421.139484, 417.75619, 414.126488, 410.258712, 406.161193, 401.842262, 397.310253, 392.573496, 387.640325, 382.519071, 377.218066, 371.745642, 366.110132, 360.319867, 354.38318, 348.308402, 342.103866, 335.777903, 329.338846, 322.795026, 316.154777, 309.426429, 302.618315, 295.738767, 288.796117, 281.798696, 274.754838, 267.672874, 260.561137, 253.427957, 246.293877, 239.160866, 232.03075, 224.911463, 217.810942, 210.737123, 203.697941, 196.701333, 189.755233, 182.867579, 176.046305, 169.299348, 162.634644, 156.060128, 149.583736, 143.213404, 136.957068, 131.141317, 125.448525, 119.879721, 114.435933, 109.11819, 103.927518, 98.8649471, 93.9315044, 89.1282182, 84.4561166, 79.9162277, 75.5095796, 71.2372004, 67.1001181, 63.0993608, 59.2359567, 55.5673976, 52.1498196, 48.8677987, 45.7187025, 42.6998988, 39.8087555, 37.0426403, 34.3989209, 31.8749652, 29.4681409, 27.1758158, 24.9953577, 22.9241343, 20.9595135, 19.0988629, 17.3395504, 15.6789438, 13.8070963, 11.8814968, 10.0633281, 8.35993089, 6.77864587, 5.32681376, 4.0117753, 2.84087121, 1.8214422, 0.960829006, 0.266372342, -0.254587068, -0.594708499, -0.746651231, -0.70307454, -0.456637704, 0, 4.56577379, 8.56482506, 12.0647956, 15.1333274, 17.8380622, 20.2466419, 22.4267084, 24.4459034, 26.371869, 28.2722468, 30.2146789, 32.266807, 34.4962729, 36.9707187, 39.757786, 42.9251168, 46.6888902, 51.1379676, 56.0040138, 61.3865, 67.3848976, 74.0986777, 81.6273117, 90.0702709, 99.5270264, 110.09705, 121.879812, 134.974784, 149.481438, 165.780456, 183.586987, 202.916139, 223.78302, 245.312701, 267.965308, 292.43831, 319.013824, 347.973969, 379.600864, 414.176626, 452.076872, 493.461767, 538.595135, 587.740802, 641.162591, 698.888924, 761.477882, 829.252142, 902.534379, 981.647272, 1066.20776, 1158.10682, 1258.75705, 1369.02043, 1488.63482, 1617.84278, 1756.88687, 1906.41947, 2066.18019, 2235.46157, 2413.15902, 2600.99682, 2801.15115, 3014.32328, 3238.71536, 3473.67669, 3718.72578, 3973.2558, 4236.0394, 4506.20023, 4784.91669, 5068.94155, 5356.68719, 5646.56598, 5936.9903, 6261.12279, 6573.93801, 6865.16078, 7124.51592, 7341.72825, 7453.66034, 7538.14074, 7607.54952, 7711.23919, 7846.95488, 7997.03654, 8159.99958, 8334.35939, 8518.63137, 8711.33092, 9113.70227, 9375.49503, 9539.89714, 9650.09651, 9749.28108, 9880.63877, 10087.3575, 10294.5082, 10450.2146},
        {438.485846, 436.112152, 433.481566, 430.601126, 427.477872, 424.11884, 420.53107, 416.7216, 412.697468, 408.465712, 404.033372, 399.407485, 394.595089, 389.603224, 384.438926, 379.109236, 373.62119, 367.981828, 362.198187, 356.277307, 350.226225, 344.051979, 337.761609, 331.362152, 324.860647, 318.264132, 311.579645, 304.814226, 297.974911, 291.06874, 284.102751, 277.083982, 270.019472, 262.916258, 255.727894, 248.489286, 241.235983, 233.976759, 226.720387, 219.475643, 212.251299, 205.056131, 197.898912, 190.788416, 183.733417, 176.742689, 169.825007, 162.989144, 156.243875, 149.597973, 143.060213, 136.948704, 130.956179, 125.084706, 119.336353, 113.713189, 108.217283, 102.850701, 97.6155128, 92.5137864, 87.54759, 82.7189919, 78.0300603, 73.4828634, 69.0794697, 64.8219472, 60.7123644, 56.8320177, 53.2603014, 49.8358815, 46.5556901, 43.4166595, 40.4157218, 37.5498093, 34.8158541, 32.2107886, 29.7315449, 27.3750552, 25.1382517, 23.0180668, 21.0114325, 19.115281, 17.3265447, 15.6421558, 13.7446157, 11.7946857, 9.9569062, 8.23841317, 6.64634277, 5.18783108, 3.87001419, 2.7000282, 1.68500919, 0.832093276, 0.148416534, -0.358884939, -0.682675049, -0.815817703, -0.751176808, -0.481616272, 0, 5.35750875, 10.0050452, 14.0196647, 17.4784227, 20.4583747, 23.0365761, 25.2900822, 27.2959485, 29.1312305, 30.8729834, 32.5982629, 34.3841242, 36.3076228, 38.4458141, 40.8757535, 43.6744965, 46.9071752, 50.6491416, 55.0044206, 60.0475125, 65.8529179, 72.4951372, 80.0486707, 88.5880189, 98.1876822, 108.922161, 120.865956, 134.093567, 148.679495, 164.810558, 182.407756, 201.511893, 222.163774, 243.932277, 267.094172, 292.052072, 318.988363, 348.085428, 379.525649, 413.491409, 449.240419, 488.161157, 530.616922, 576.971012, 627.586724, 682.580535, 742.624269, 808.142929, 879.561519, 957.305041, 1042.197, 1133.77684, 1231.24698, 1334.66044, 1445.80685, 1565.69613, 1695.33821, 1836.99813, 1990.14545, 2152.88044, 2323.7098, 2504.70421, 2698.24085, 2905.12544, 3123.50091, 3352.73244, 3592.62517, 3842.65835, 4100.69805, 4365.52285, 4634.09424, 4907.28743, 5184.50998, 5465.16946, 5748.67342, 6079.73267, 6401.12574, 6700.93438, 6967.24035, 7188.1254, 7292.71434, 7367.35053, 7425.38299, 7521.84562, 7653.58718, 7800.08776, 7959.06454, 8128.23472, 8305.31552, 8488.02412, 8785.80083, 8994.18358, 9145.48419, 9272.01448, 9406.08629, 9580.01144, 9826.10176, 10004.886, 10202.4499},
        {446.318603, 444.020316, 441.472115, 438.680382, 435.651503, 432.391862, 428.907843, 425.205831, 421.29221, 417.173364, 412.855678, 408.345536, 403.649322, 398.773421, 393.724216, 388.508093, 383.131435, 377.600627, 371.922054, 366.102099, 360.147147, 354.063582, 347.857788, 341.536151, 335.105053, 328.57088, 321.940016, 315.218845, 308.413751, 301.53112, 294.577334, 287.558779, 280.481839, 273.352897, 266.098137, 258.76573, 251.404814, 244.024374, 236.633398, 229.240873, 221.855787, 214.487125, 207.143876, 199.835027, 192.569563, 185.356473, 178.204743, 171.123361, 164.121314, 157.207587, 150.39117, 143.915787, 137.550598, 131.299501, 125.166396, 119.155179, 113.26975, 107.514007, 101.891848, 96.4071717, 91.0638766, 85.8658609, 80.817023, 75.9212612, 71.182474, 66.6045596, 62.1914166, 58.0736188, 54.3786985, 50.8492955, 47.4810963, 44.2697874, 41.2110553, 38.3005866, 35.5340676, 32.9071851, 30.4156254, 28.0550751, 25.8212208, 23.7097488, 21.7163459, 19.8366983, 18.0664928, 16.4014157, 14.4870014, 12.5013775, 10.6268802, 8.87055907, 7.23946376, 5.74064391, 4.38114915, 3.1680291, 2.10833341, 1.20911169, 0.477413581, -0.0797112892, -0.45521329, -0.642042792, -0.633150165, -0.421485777, 0, 3.78796061, 7.14698376, 10.1354152, 12.8116007, 15.2338859, 17.4606167, 19.5501388, 21.5607979, 23.5509398, 25.5789103, 27.703055, 29.9817198, 32.4732504, 35.2359925, 38.3282918, 41.8084943, 45.8252885, 50.4499266, 55.5364067, 61.1624339, 67.4057131, 74.3439494, 82.0548477, 90.616113, 100.10545, 110.600564, 122.179161, 134.918944, 148.897618, 164.366698, 181.166351, 199.322137, 218.859621, 238.991163, 260.148925, 282.981926, 307.759688, 334.751735, 364.22759, 396.456775, 431.730879, 470.290644, 512.401277, 558.327983, 608.335968, 661.908723, 720.288599, 783.936229, 853.312249, 928.877293, 1011.86214, 1101.016, 1194.94564, 1293.44933, 1398.75798, 1512.01026, 1634.34489, 1768.20971, 1913.13672, 2067.22976, 2228.83277, 2400.58466, 2586.40661, 2787.33014, 2999.49036, 3221.26994, 3449.54228, 3682.29874, 3923.06456, 4172.23474, 4436.20211, 4708.44103, 4987.27025, 5271.00858, 5557.97477, 5883.41821, 6199.49443, 6495.28958, 6759.88978, 6982.38117, 7099.97546, 7188.21616, 7258.4213, 7353.48934, 7472.23772, 7600.32538, 7737.22769, 7882.42001, 8035.37773, 8195.5762, 8357.30942, 8531.76626, 8714.48064, 8900.98653, 9086.81786, 9267.50859, 9438.59266, 9654.05041, 9832.46838},
        {454.702129, 452.703609, 450.439119, 447.91533, 445.138914, 442.116543, 438.854889, 435.360625, 431.640422, 427.700953, 423.54889, 419.190905, 414.633669, 409.883856, 404.948136, 399.833183, 394.545668, 389.092264, 383.479642, 377.714475, 371.803434, 365.753192, 359.570421, 353.261794, 346.833981, 340.293655, 333.647489, 326.902154, 320.064322, 313.140667, 306.137858, 299.06257, 291.921473, 284.721241, 277.413222, 270.032915, 262.61648, 255.172384, 247.709096, 240.235083, 232.758811, 225.288749, 217.833364, 210.401123, 203.000494, 195.639944, 188.32794, 181.07295, 173.883442, 166.767882, 159.734738, 152.91771, 146.197317, 139.579314, 133.069452, 126.673484, 120.397162, 114.246239, 108.226468, 102.343601, 96.6033915, 91.0115913, 85.5739532, 80.2962298, 75.1841738, 70.2435377, 65.4800742, 61.0661812, 57.1702557, 53.4543566, 49.9134337, 46.5424365, 43.3363148, 40.2900183, 37.3984966, 34.6566993, 32.0595762, 29.6020769, 27.2791511, 25.0857484, 23.0168186, 21.0673112, 19.232176, 17.5063625, 15.5288147, 13.4799731, 11.5445575, 9.72907058, 8.04001509, 6.48389377, 5.06720938, 3.79646466, 2.67816234, 1.71880519, 0.924895934, 0.302937324, -0.140567897, -0.399116984, -0.466207193, -0.33533578, 0, 1.67794929, 3.29816796, 4.89389998, 6.49838936, 8.14488008, 9.86661614, 11.6968415, 13.6688002, 15.8157362, 18.1708936, 20.7675162, 23.6388481, 26.8181333, 30.3386158, 34.2335395, 38.5361485, 43.6076646, 49.5281859, 55.5891015, 61.8939364, 68.5462156, 75.6494641, 83.3072067, 91.6229686, 100.700275, 110.64265, 121.553619, 133.536708, 146.69544, 161.492787, 177.541031, 194.835864, 213.372977, 232.014319, 251.322451, 272.162269, 294.869587, 319.780218, 347.229977, 377.554678, 411.905937, 449.555478, 490.679503, 535.454213, 584.055807, 635.419948, 691.273511, 752.10283, 818.394243, 890.634083, 970.254552, 1055.64006, 1145.40222, 1239.25855, 1339.18448, 1446.14169, 1561.09186, 1685.685, 1820.03802, 1963.51691, 2115.64285, 2278.10431, 2454.60834, 2645.90763, 2847.74883, 3058.17644, 3272.51519, 3488.10446, 3712.25613, 3946.64107, 4205.66806, 4476.31037, 4755.8296, 5041.48735, 5330.54519, 5648.1073, 5956.63205, 6246.4204, 6507.7733, 6730.9917, 6862.53671, 6965.77569, 7049.79809, 7144.6334, 7251.0823, 7361.22049, 7476.28494, 7597.51258, 7726.14038, 7863.40529, 7935.19299, 8087.19371, 8289.21803, 8511.07653, 8722.57977, 8893.53832, 8993.76276, 9233.7648, 9375.06886},
        {463.557869, 462.344297, 460.800628, 458.935743, 456.758526, 454.277858, 451.50262, 448.441695, 445.103965, 441.49831, 437.633614, 433.518757, 429.162623, 424.574092, 419.762046, 414.735368, 409.502939, 404.073641, 398.456356, 392.659966, 386.693352, 380.565396, 374.284981, 367.860988, 361.302299, 354.617796, 347.81636, 340.906874, 333.89822, 326.799278, 319.618932, 312.366062, 305.049552, 297.678282, 290.324174, 282.963263, 275.57083, 268.153709, 260.718737, 253.27275, 245.822584, 238.375075, 230.937059, 223.515372, 216.11685, 208.74833, 201.416648, 194.128639, 186.89114, 179.710987, 172.595016, 165.55342, 158.589606, 151.710337, 144.922375, 138.232485, 131.647429, 125.173971, 118.818875, 112.588903, 106.490818, 100.531385, 94.7173662, 89.0555252, 83.5526252, 78.2154295, 73.0507015, 68.2263341, 63.9063515, 59.7651993, 55.7991955, 52.004658, 48.3779046, 44.9152534, 41.6130223, 38.4675292, 35.475092, 32.6320286, 29.9346571, 27.3792953, 24.9622611, 22.6798725, 20.5284475, 18.5043039, 16.3896429, 14.2923444, 12.3228621, 10.4844622, 8.7804113, 7.21397578, 5.78842207, 4.50701661, 3.37302583, 2.38971616, 1.56035403, 0.888205885, 0.376538151, 0.0286172631, -0.152290347, -0.162918246, 0, -1.16687128, -1.96459474, -2.38923645, -2.43686246, -2.10353882, -1.3853316, -0.278306854, 1.22146936, 3.11793099, 5.41501197, 8.11664626, 11.2267678, 14.7493105, 18.6882083, 23.0473952, 27.8308051, 34.2093038, 42.3535296, 49.6279281, 56.2864899, 62.5832058, 68.7720666, 75.107063, 81.8421858, 89.2314257, 97.5287735, 106.98822, 117.863756, 130.409372, 146.130681, 163.571124, 182.609204, 203.123425, 224.520202, 246.914082, 270.545507, 295.434607, 321.601513, 349.066358, 377.849271, 404.9587, 434.343061, 466.611725, 502.374066, 542.239459, 588.256012, 639.234679, 695.425149, 757.077113, 824.440258, 898.372233, 977.77171, 1061.68908, 1149.88049, 1243.54309, 1343.22703, 1449.48246, 1561.52412, 1681.54111, 1813.17942, 1962.2516, 2122.60994, 2288.35809, 2457.47699, 2634.51746, 2821.01464, 3020.23661, 3234.16777, 3458.4385, 3692.27336, 3935.11654, 4185.93922, 4443.88998, 4708.11737, 4977.76996, 5291.78189, 5599.56974, 5890.3357, 6153.28193, 6377.61059, 6505.34202, 6604.31479, 6683.95187, 6779.60059, 6890.42217, 7004.77604, 7122.36351, 7242.88591, 7366.04456, 7491.54079, 7488.61093, 7578.57199, 7735.20709, 7932.29934, 8143.63184, 8342.98772, 8504.15008, 8708.7145, 8843.70243},
        {482.446566, 482.127966, 481.373089, 480.195008, 478.606794, 476.621518, 474.252251, 471.512067, 468.414034, 464.971227, 461.196715, 457.103571, 452.704865, 448.01367, 443.043057, 437.806098, 432.315864, 426.585426, 420.627856, 414.456226, 408.083607, 401.523071, 394.787689, 387.890533, 380.844674, 373.663183, 366.359134, 358.945595, 351.435641, 343.842341, 336.178767, 328.457992, 320.693086, 312.897121, 305.385617, 298.01406, 290.6343, 283.249593, 275.863197, 268.478368, 261.098361, 253.726434, 246.365843, 239.019845, 231.691695, 224.384651, 217.10197, 209.846906, 202.622718, 195.432661, 188.279991, 180.917158, 173.603662, 166.348197, 159.159458, 152.046137, 145.016928, 138.080525, 131.245621, 124.520909, 117.915084, 111.436838, 105.094866, 98.8978604, 92.8545152, 86.9735238, 81.2635798, 75.9032103, 71.0668651, 66.4129684, 61.939204, 57.6432559, 53.522808, 49.5755442, 45.7991484, 42.1913045, 38.7496964, 35.4720081, 32.3559233, 29.3991261, 26.5993003, 23.9541299, 21.4612987, 19.1184906, 16.798588, 14.5643005, 12.4798379, 10.5469343, 8.76732348, 7.14273955, 5.67491642, 4.36558804, 3.21648837, 2.22935138, 1.40591103, 0.747901269, 0.257056066, -0.0648906227, -0.216204836, -0.195152615, 0, 0.181345035, 0.518531952, 1.02269072, 1.7049513, 2.57644367, 3.6482978, 4.93164365, 6.43761119, 8.17733039, 10.1619312, 12.4025436, 14.9102976, 17.6963232, 20.7717502, 24.1477087, 27.8353286, 32.7985986, 39.1847713, 44.8497011, 50.008702, 54.877088, 59.6701731, 64.6032711, 69.8916961, 75.750762, 82.3957829, 90.0420726, 98.9049451, 109.199714, 122.481941, 137.135269, 152.972938, 169.808187, 185.350369, 200.464669, 216.577305, 234.132684, 253.575213, 275.349297, 299.899342, 331.202692, 365.095573, 401.331166, 439.662649, 479.843202, 516.585266, 555.942944, 598.929598, 646.558594, 699.843295, 762.11556, 829.236542, 901.468506, 978.367435, 1058.04733, 1139.26961, 1220.7957, 1292.1386, 1363.41007, 1454.81103, 1592.09079, 1747.88849, 1897.13165, 2030.68631, 2166.42548, 2309.82554, 2471.1798, 2657.21347, 2856.98965, 3069.56211, 3299.48312, 3539.46207, 3786.64942, 4038.19563, 4291.25115, 4559.16286, 4818.83572, 5063.37108, 5285.87031, 5479.43477, 5605.94944, 5710.54298, 5799.69522, 5891.9805, 5989.09123, 6087.68039, 6189.69236, 6297.07149, 6411.76214, 6535.70869, 7313.83362, 7647.79866, 7675.36359, 7534.28824, 7362.33239, 7297.25586, 7476.81845, 7626.18174, 7746.78912},
        {487.525079, 487.500363, 487.054791, 486.200291, 484.94879, 483.312218, 481.302501, 478.931567, 476.211345, 473.153762, 469.770747, 466.074226, 462.076128, 457.788381, 453.222913, 448.391652, 443.306524, 437.97946, 432.422386, 426.647229, 420.665919, 414.490383, 408.132549, 401.604345, 394.917698, 388.084537, 381.116789, 374.026383, 366.825245, 359.525305, 352.13849, 344.676728, 337.151947, 329.576074, 322.194635, 314.897846, 307.573114, 300.224788, 292.857214, 285.47474, 278.081712, 270.682479, 263.281386, 255.882783, 248.491015, 241.110429, 233.745374, 226.400197, 219.079244, 211.786863, 204.527401, 197.175389, 189.867805, 182.611809, 175.414565, 168.283232, 161.224972, 154.246946, 147.356317, 140.560245, 133.865892, 127.280418, 120.810986, 114.464757, 108.248892, 102.170553, 96.2369004, 90.5607664, 85.2524451, 80.1011612, 75.1072262, 70.2709511, 65.5926473, 61.0726258, 56.711198, 52.5086751, 48.4653682, 44.5815886, 40.8576476, 37.2938563, 33.890526, 30.6479678, 27.566493, 24.6464129, 21.799997, 19.0734293, 16.5139517, 14.1247325, 11.9089402, 9.86974307, 8.01030942, 6.33380761, 4.84340598, 3.54227287, 2.43357662, 1.52048555, 0.806168023, 0.293792364, -0.0134730857, -0.112459986, 0, 0.800282488, 1.70739789, 2.73485252, 3.89615269, 5.20480472, 6.67431491, 8.31818959, 10.1499351, 12.1830576, 14.4310636, 16.9074594, 19.6257512, 22.5994453, 25.8420482, 29.367066, 33.1880051, 37.7757389, 43.2091119, 48.4671096, 53.6612456, 58.9030333, 64.3039864, 69.9756183, 76.0294427, 82.576973, 89.7297229, 97.5992059, 106.296935, 115.934425, 127.101808, 139.256485, 152.366383, 166.399428, 180.084983, 194.010257, 209.092747, 225.671951, 244.087366, 264.678489, 287.78482, 316.649326, 347.824369, 381.081377, 416.191776, 452.926994, 486.504859, 522.388797, 561.488635, 604.7142, 652.97532, 709.101684, 769.736757, 834.59374, 903.311805, 975.378977, 1050.35115, 1127.7842, 1203.40294, 1281.46504, 1370.40752, 1479.63037, 1600.73009, 1726.97149, 1855.97912, 1992.76494, 2138.94997, 2297.20691, 2469.42944, 2653.65506, 2850.10257, 3064.67034, 3291.02377, 3527.4156, 3772.09859, 4023.3255, 4310.88368, 4593.60765, 4861.8665, 5106.02935, 5316.46531, 5444.08148, 5545.01697, 5626.55316, 5710.25344, 5797.74591, 5884.70784, 5973.06487, 6064.74266, 6161.66684, 6265.76305, 6825.65394, 7078.21265, 7120.37182, 7049.06411, 6961.22216, 6953.77862, 7123.66614, 7195.89622, 7266.38062},
        {510.463075, 510.459455, 510.070294, 509.305389, 508.174535, 506.687527, 504.854162, 502.684235, 500.187541, 497.373877, 494.253038, 490.834819, 487.129016, 483.145425, 478.893841, 474.384061, 469.625879, 464.629092, 459.403494, 453.958882, 448.305052, 442.451798, 436.408917, 430.186204, 423.793455, 417.240465, 410.537031, 403.692947, 396.718009, 389.622013, 382.414756, 375.106031, 367.705635, 360.223364, 352.745376, 345.24168, 337.681365, 330.071749, 322.420149, 314.733883, 307.020267, 299.286621, 291.540261, 283.788504, 276.038669, 268.298072, 260.574032, 252.873865, 245.204889, 237.574422, 229.989781, 222.646818, 215.360229, 208.133245, 200.969095, 193.871011, 186.842223, 179.885961, 173.005456, 166.203938, 159.484638, 152.850786, 146.305612, 139.852348, 133.494222, 127.234467, 121.076312, 115.00705, 109.014355, 103.134329, 97.3712358, 91.7293395, 86.2129034, 80.8261912, 75.5734665, 70.4589931, 65.4870345, 60.6618545, 55.9877167, 51.4688847, 47.1096222, 42.9141928, 38.8868603, 35.0318883, 31.4854877, 28.1831733, 25.0588738, 22.1125709, 19.3442462, 16.7538816, 14.3414586, 12.1069591, 10.0503648, 8.17165723, 6.47081826, 4.94782956, 3.60267284, 2.43532982, 1.44578219, 0.634011685, 0, -1.65129453, -2.91031605, -3.78976358, -4.30233609, -4.4607326, -4.2776521, -3.76579359, -2.93785606, -1.80653852, -0.384539964, 1.31544061, 3.28070421, 5.49855183, 7.95628447, 10.6412031, 13.5406088, 17.713023, 23.2987783, 27.8621768, 31.6200669, 34.7892967, 37.5867147, 40.2291689, 42.9335076, 45.9165792, 49.3952317, 53.5863136, 58.706673, 64.9731582, 74.1689238, 84.3701993, 95.3239411, 106.777106, 116.594316, 125.463696, 134.575323, 144.240853, 154.771943, 166.480248, 179.677425, 196.37299, 214.664, 234.529919, 255.950214, 278.90435, 303.379492, 329.345483, 356.779866, 385.660182, 415.963973, 447.035296, 480.259438, 516.871042, 557.362727, 600.712142, 646.577127, 694.615521, 741.339724, 790.267887, 848.349549, 923.504863, 1008.47108, 1097.87914, 1189.56426, 1286.79101, 1390.33183, 1501.24605, 1620.3805, 1747.52984, 1883.08389, 2027.31052, 2180.74025, 2343.80511, 2516.93715, 2700.56839, 2931.52669, 3164.7493, 3391.56928, 3603.31972, 3791.33369, 3914.47329, 4018.18987, 4107.73265, 4199.82174, 4295.16576, 4389.69382, 4484.35353, 4580.0925, 4677.85835, 4778.59869, 5055.9632, 5222.26432, 5308.03174, 5343.79515, 5360.08426, 5387.42878, 5456.35838, 5582.26266, 5695.08446}
    };
    
    const double b3tab[98][200] = {{98.4097635, 97.1760915, 95.9730506, 94.799433, 93.6540305, 92.5356354, 91.4430395, 90.3750351, 89.330414, 88.3079684, 87.3064903, 86.3247717, 85.3616047, 84.4157813, 83.4860937, 82.5713337, 81.6702935, 80.781765, 79.9045405, 79.0374118, 78.1791711, 77.3286103, 76.4845216, 75.6456969, 74.8109283, 73.9790079, 73.1487277, 72.3188798, 71.4882561, 70.6556487, 69.8198498, 68.9796512, 68.1338451, 67.2812235, 66.355427, 65.3891938, 64.416045, 63.4368869, 62.452626, 61.4641686, 60.4724209, 59.4782894, 58.4826803, 57.4865001, 56.4906549, 55.4960512, 54.5035954, 53.5141936, 52.5287523, 51.5481778, 50.5733764, 49.6682217, 48.7692878, 47.876116, 46.9882476, 46.1052241, 45.2265866, 44.3518766, 43.4806354, 42.6124042, 41.7467245, 40.8831375, 40.0211846, 39.1604072, 38.3003464, 37.4405438, 36.5805405, 35.7330266, 34.9103763, 34.0850128, 33.2556252, 32.4209025, 31.5795334, 30.7302072, 29.8716125, 29.0024386, 28.1213742, 27.2271084, 26.3183301, 25.3937283, 24.4519919, 23.4918099, 22.5118712, 21.5108649, 20.3371478, 19.067737, 17.7814563, 16.4818732, 15.1725553, 13.8570701, 12.538985, 11.2218677, 9.90928565, 8.60480633, 7.31199729, 6.03442604, 4.77566012, 3.53926703, 2.3288143, 1.14786945, 0, -0.0159923779, -0.153916723, -0.405638979, -0.763025093, -1.21794101, -1.76225267, -2.38782603, -3.08652703, -3.85022161, -4.67077572, -5.5400553, -6.4499263, -7.39225467, -8.35890635, -9.34174728, -10.3326434, -11.4052827, -12.56322, -13.6132475, -14.5647646, -15.4271705, -16.2098644, -16.9222455, -17.5737132, -18.1736665, -18.7315047, -19.2566272, -19.758433, -20.2463215, -20.8648893, -21.4387658, -21.936791, -22.3278051, -22.6794273, -22.911108, -22.9159564, -22.6331786, -22.001981, -20.9615697, -19.4511509, -17.8944717, -15.5987285, -12.408326, -8.16766893, -2.72116204, 3.25058181, 10.9484181, 20.736994, 32.980957, 48.0449543, 65.2627953, 87.2898783, 114.961552, 149.305631, 191.74288, 243.517638, 305.874244, 376.394452, 460.81759, 568.87853, 704.740707, 875.950684, 1081.02909, 1327.16872, 1636.25687, 2023.4358, 2483.92395, 3027.69813, 3737.78933, 4655.90497, 5559.49816, 6795.18515, 8496.14601, 10795.5608, 13826.6096, 21500.1379, 29227.244, 36196.6917, 41597.2445, 44617.6664, 41038.9421, 35729.4667, 29338.4807, 24336.1642, 20650.969, 17452.6218, 14707.5105, 12382.023, 10442.5474, 8855.47136, 7591.24265, 6623.20723, 5904.70549, 5389.07778, 5029.66447, 4779.80594, 4592.84254, 4453.81101, 4428.88647},
        {88.9798529, 87.6384075, 86.3382259, 85.0779503, 83.8562226, 82.6716849, 81.5229792, 80.4087474, 79.3276317, 78.2782741, 77.2593165, 76.269401, 75.3071696, 74.3712643, 73.4603273, 72.5730004, 71.7079257, 70.8637452, 70.039101, 69.232635, 68.4429893, 67.668806, 66.908727, 66.1613943, 65.4254501, 64.6995362, 63.9822947, 63.2723677, 62.5683972, 61.8690252, 61.1728937, 60.4786447, 59.7849202, 59.0903624, 58.3528459, 57.592254, 56.8289596, 56.0629279, 55.2941237, 54.5225122, 53.7480582, 52.9707268, 52.190483, 51.4072918, 50.6211181, 49.831927, 49.0396835, 48.2443525, 47.4458991, 46.6442882, 45.8394849, 45.0163497, 44.1902796, 43.3615668, 42.5305038, 41.6973832, 40.8624972, 40.0261383, 39.188599, 38.3501717, 37.5111487, 36.6718227, 35.8324859, 34.9934308, 34.1549498, 33.3173354, 32.4808799, 31.6685967, 30.9029531, 30.137382, 29.3707029, 28.6017353, 27.8292987, 27.0522127, 26.2692968, 25.4793705, 24.6812534, 23.8737651, 23.055725, 22.2259528, 21.3832679, 20.5264898, 19.6544382, 18.7659325, 17.7508519, 16.6647775, 15.5645999, 14.4526739, 13.3313545, 12.2029964, 11.0699546, 9.93458386, 8.79923911, 7.6662752, 6.538047, 5.41690936, 4.30521716, 3.20532526, 2.11958852, 1.05036181, 0, -0.388021037, -0.846035596, -1.36929094, -1.95303435, -2.59251307, -3.28297439, -4.01966556, -4.79783385, -5.61272654, -6.45959087, -7.33367413, -8.23022358, -9.14448649, -10.0717101, -11.0071417, -11.9460286, -12.9244536, -13.9434976, -14.9060417, -15.8160836, -16.6776212, -17.494652, -18.271174, -19.0111848, -19.7186822, -20.3976639, -21.0521277, -21.6860714, -22.3034926, -23.0078702, -23.6672446, -24.255769, -24.7475969, -25.2925024, -25.7768285, -26.0400862, -26.0037428, -25.5892655, -24.7181215, -23.3117781, -21.5778167, -19.0645121, -15.6373528, -11.1618271, -5.50342338, 0.475233378, 8.15507551, 17.9198988, 30.153499, 45.2396718, 62.4118971, 84.6102284, 112.114335, 145.978112, 188.836164, 242.613391, 309.234691, 388.808222, 485.488518, 605.412014, 747.648881, 926.470114, 1144.9921, 1413.22357, 1752.85196, 2180.20398, 2689.77775, 3292.24071, 4078.29852, 5093.3827, 6156.17734, 7573.74559, 9470.00845, 11968.8869, 15194.3021, 22920.019, 30707.6535, 37768.6655, 43314.5152, 46556.6623, 43449.1498, 38632.4661, 32714.1072, 27950.0211, 24295.1915, 21017.7466, 18107.0129, 15552.3166, 13342.984, 11468.3414, 9989.6802, 8813.27937, 7903.56495, 7224.96295, 6741.89938, 6418.80026, 6220.09159, 6217.92725, 6532.91105},
        {81.2404201, 80.2115989, 79.1918898, 78.1813176, 77.179907, 76.187683, 75.2046702, 74.2308935, 73.2663776, 72.3111474, 71.3652276, 70.428643, 69.5014183, 68.5835785, 67.6751483, 66.7761524, 65.8866157, 65.0065629, 64.1360189, 63.2750084, 62.4235562, 61.5816871, 60.7494259, 59.9267973, 59.1138263, 58.3105375, 57.5169557, 56.7331057, 55.9590124, 55.1947005, 54.4401948, 53.6955201, 52.9607012, 52.2357628, 51.554373, 50.8990274, 50.251817, 49.6116749, 48.9775341, 48.3483276, 47.7229884, 47.1004494, 46.4796438, 45.8595045, 45.2389644, 44.6169567, 43.9924142, 43.36427, 42.7314572, 42.0929086, 41.4475573, 40.7316872, 40.0082385, 39.2775024, 38.5397698, 37.7953321, 37.0444801, 36.2875052, 35.5246983, 34.7563506, 33.9827533, 33.2041973, 32.4209739, 31.6333741, 30.841689, 30.0462099, 29.2472277, 28.451951, 27.6674224, 26.8796662, 26.0885251, 25.2938417, 24.4954588, 23.693219, 22.8869651, 22.0765395, 21.2617852, 20.4425447, 19.6186608, 18.789976, 17.9563331, 17.1175748, 16.2735438, 15.4240826, 14.5470907, 13.6538438, 12.7558818, 11.8537594, 10.9480313, 10.0392525, 9.12797758, 8.21476144, 7.30015884, 6.38472456, 5.46901341, 4.55358015, 3.63897959, 2.72576651, 1.81449569, 0.905721924, 0, -0.886193644, -1.76498891, -2.63690887, -3.50247659, -4.36221513, -5.21664756, -6.06629696, -6.91168637, -7.75333889, -8.59177756, -9.42752546, -10.2611056, -11.0930412, -11.9238552, -12.7540707, -13.5842107, -14.3590863, -15.0712616, -15.8472752, -16.6757117, -17.545156, -18.4441929, -19.3614071, -20.2853833, -21.2047063, -22.1079609, -22.9837318, -23.8206039, -24.6071618, -25.3508027, -26.0144011, -26.5808981, -27.0332348, -27.5250254, -27.9538743, -28.1718732, -28.1107612, -27.7022773, -26.8781606, -25.5701503, -24.3042021, -22.2369898, -19.1839928, -14.9606905, -9.38256218, -2.45240388, 6.24845099, 16.9513523, 29.8876499, 45.2886937, 61.4397746, 82.8968178, 110.322545, 145.149456, 190.381683, 248.317725, 321.256085, 410.32486, 519.258954, 653.070074, 808.881286, 1004.09412, 1241.92292, 1534.02015, 1906.33623, 2378.25857, 2945.92843, 3622.70647, 4507.18921, 5649.75893, 6888.66355, 8519.05579, 10664.7491, 13449.5568, 16997.2924, 25185.7631, 33446.29, 40964.1882, 46924.7727, 50513.3584, 47583.6229, 42873.6104, 36996.4803, 32207.4119, 28469.5986, 25062.0594, 21982.1964, 19227.4122, 16795.109, 14682.6891, 12836.2218, 11396.4304, 10294.3959, 9461.19934, 8827.92158, 8325.64363, 7885.44646, 7768.35572, 7902.53999},
        {74.8104589, 73.8844318, 72.9734349, 72.0769787, 71.1945738, 70.3257308, 69.4699602, 68.6267724, 67.7956782, 66.9761879, 66.1678122, 65.3700615, 64.5824466, 63.8044777, 63.0356656, 62.2755208, 61.5235538, 60.7792751, 60.0421953, 59.3118249, 58.5876745, 57.8692546, 57.1560758, 56.4476486, 55.7434835, 55.0430911, 54.3459819, 53.6516665, 52.9596555, 52.2694592, 51.5805884, 50.8925535, 50.2048651, 49.5170338, 48.8079731, 48.0879253, 47.3673801, 46.6465166, 45.9255137, 45.2045502, 44.4838053, 43.7634578, 43.0436867, 42.3246709, 41.6065893, 40.8896211, 40.173945, 39.45974, 38.7471852, 38.0364594, 37.3277416, 36.6462264, 35.9665349, 35.2883037, 34.6111694, 33.9347687, 33.2587382, 32.5827146, 31.9063346, 31.2292348, 30.5510519, 29.8714225, 29.1899833, 28.506371, 27.8202222, 27.1311735, 26.4388617, 25.7374737, 25.0213272, 24.3012985, 23.5773777, 22.8495546, 22.1178191, 21.3821613, 20.6425711, 19.8990384, 19.1515532, 18.4001054, 17.6446849, 16.8852817, 16.1218858, 15.354487, 14.5830754, 13.8076409, 13.0298425, 12.2488006, 11.4636154, 10.6742225, 9.88055777, 9.082557, 8.28015594, 7.47329037, 6.66189606, 5.84590879, 5.02526434, 4.19989847, 3.36974695, 2.53474558, 1.69483011, 0.849936327, 0, -0.972202188, -1.92966602, -2.87346881, -3.80468788, -4.72440052, -5.63368406, -6.53361581, -7.42527307, -8.30973317, -9.18807341, -10.0613711, -10.9307036, -11.7971481, -12.6617821, -13.5256827, -14.3899274, -15.2227862, -16.0206497, -16.858802, -17.7312902, -18.6321616, -19.5554634, -20.4952427, -21.4455468, -22.4004229, -23.3539181, -24.3000797, -25.2329548, -26.1465907, -27.1202977, -28.0315966, -28.8489559, -29.5408438, -30.1590414, -30.6303606, -30.8593965, -30.7896234, -30.3645157, -29.5275478, -28.2221942, -27.1343611, -25.2391335, -22.3347274, -18.2193589, -12.6912438, -6.7076883, 1.38195421, 12.0692402, 25.8457262, 43.2029687, 63.5497607, 89.7837998, 121.300037, 159.109821, 207.524647, 269.374313, 347.488615, 443.441046, 561.603232, 707.71731, 879.524346, 1095.06589, 1353.53122, 1667.6293, 2072.97735, 2594.67734, 3237.70119, 4021.93195, 5041.06283, 6347.03582, 7698.32436, 9485.48803, 11862.0541, 14981.5499, 18997.5026, 28447.3725, 38004.7707, 46727.2413, 53672.3282, 57897.5756, 54604.6053, 49277.4983, 42626.3368, 37310.2043, 33267.7034, 29625.3522, 26362.3574, 23457.9255, 20891.2631, 18641.5766, 16334.7945, 14577.3557, 13251.1515, 12238.0733, 11420.0124, 10678.8603, 9896.50829, 9500.74792, 9265.54573},
        {70.5057947, 69.5911794, 68.7000529, 67.8314576, 66.9844361, 66.1580308, 65.3512843, 64.5632391, 63.7929377, 63.0394226, 62.3017363, 61.5789214, 60.8700203, 60.1740755, 59.4901296, 58.8172251, 58.1544045, 57.5007103, 56.855185, 56.2168711, 55.5848112, 54.9580478, 54.3356233, 53.7165802, 53.0999612, 52.4848086, 51.8701651, 51.2550731, 50.6385751, 50.0197137, 49.3975314, 48.7710706, 48.1393739, 47.5014838, 46.801655, 46.0674764, 45.3271952, 44.581632, 43.831607, 43.0779409, 42.3214542, 41.5629673, 40.8033006, 40.0432747, 39.2837101, 38.5254271, 37.7692464, 37.0159884, 36.2664735, 35.5215223, 34.7819552, 34.1083144, 33.4404041, 32.77775, 32.119878, 31.4663139, 30.8165836, 30.1702128, 29.5267275, 28.8856533, 28.2465163, 27.6088421, 26.9721567, 26.3359857, 25.6998552, 25.0632909, 24.4258186, 23.7830522, 23.1306996, 22.4763545, 21.8197966, 21.1608051, 20.4991596, 19.8346394, 19.167024, 18.4960929, 17.8216253, 17.1434008, 16.4611988, 15.7747987, 15.0839799, 14.3885219, 13.6882041, 12.9828059, 12.2712328, 11.5537196, 10.8305116, 10.1014165, 9.3662421, 8.62479624, 7.87688667, 7.12232119, 6.36090757, 5.5924536, 4.81676706, 4.03365575, 3.24292745, 2.44438993, 1.637851, 0.823118425, 0, -0.946967845, -1.8841558, -2.81254045, -3.73309838, -4.64680616, -5.55464037, -6.4575776, -7.35659442, -8.25266742, -9.14677317, -10.0398883, -10.9329893, -11.8270528, -12.7230554, -13.6219736, -14.5247841, -15.4168438, -16.2968979, -17.2012531, -18.1275388, -19.0733847, -20.0364202, -21.0142748, -22.0045781, -23.0049596, -24.0130488, -25.0264752, -26.0428685, -27.059858, -28.2075372, -29.3025016, -30.3026416, -31.1658477, -31.9227845, -32.4949553, -32.7844566, -32.7273465, -32.259683, -31.3175242, -29.8369281, -28.4824336, -26.2399062, -22.9028752, -18.2648696, -12.1194185, -5.97208208, 2.52364921, 14.0022539, 29.0982104, 48.4459975, 71.8432111, 101.784069, 137.388549, 179.618784, 233.197965, 301.160644, 386.541373, 489.507438, 616.61231, 777.537387, 973.42358, 1221.19388, 1517.56222, 1875.81071, 2337.29525, 2930.48541, 3659.4693, 4546.39536, 5702.81071, 7189.69352, 8597.00718, 10529.2085, 13210.3191, 16864.3609, 21715.3555, 33869.6407, 46198.3433, 57454.906, 66392.7717, 71765.3832, 67169.4479, 59952.9674, 51079.4138, 44280.5986, 39424.1929, 35224.3932, 31606.5445, 28495.9916, 25818.0794, 23498.153, 20908.8917, 18833.7294, 17168.9203, 15810.7185, 14655.3783, 13599.1539, 12538.2996, 11844.3095, 11255.8106},
        {68.2331449, 67.529362, 66.816162, 66.0941873, 65.36408, 64.6264826, 63.8820374, 63.1313865, 62.3751724, 61.6140373, 60.8486235, 60.0795734, 59.3075291, 58.5331331, 57.7570276, 56.979855, 56.2022574, 55.4248773, 54.6483569, 53.8733385, 53.1004644, 52.330377, 51.5637184, 50.8011311, 50.0432573, 49.2907394, 48.5442195, 47.8043401, 47.0717433, 46.3470716, 45.6309672, 44.9240724, 44.2270296, 43.5404809, 42.9188189, 42.3346802, 41.7600555, 41.1938427, 40.6349399, 40.0822451, 39.5346562, 38.9910713, 38.4503884, 37.9115055, 37.3733205, 36.8347316, 36.2946366, 35.7519337, 35.2055207, 34.6542958, 34.0971569, 33.4667542, 32.8296697, 32.1862374, 31.5367916, 30.8816663, 30.2211956, 29.5557137, 28.8855546, 28.2110527, 27.5325418, 26.8503563, 26.1648301, 25.4762975, 24.7850925, 24.0915494, 23.3960021, 22.701919, 22.0126931, 21.3221947, 20.6305545, 19.9379037, 19.2443731, 18.5500937, 17.8551964, 17.1598122, 16.4640721, 15.768107, 15.0720479, 14.3760256, 13.6801712, 12.9846157, 12.2894899, 11.5949248, 10.9318965, 10.2844646, 9.63631797, 8.98658658, 8.33440039, 7.67888937, 7.01918348, 6.3544127, 5.68370699, 5.00619632, 4.32101065, 3.62727996, 2.92413421, 2.21070338, 1.48611742, 0.749506301, 0, -1.12453843, -2.20249071, -3.23781596, -4.23447331, -5.19642188, -6.12762079, -7.03202916, -7.91360612, -8.77631079, -9.62410229, -10.4609397, -11.2907823, -12.117589, -12.9453191, -13.7779316, -14.6193856, -15.3579602, -15.9810886, -16.7543876, -17.6570277, -18.6681794, -19.7670133, -20.9326997, -22.1444093, -23.3813126, -24.62258, -25.847382, -27.0348892, -28.164272, -29.225444, -30.1828935, -31.0125682, -31.6904158, -32.3956745, -33.0026465, -33.3314234, -33.2969656, -32.8142334, -31.7981874, -30.1637879, -28.657169, -26.1091517, -22.2720752, -16.8982791, -9.74010291, -0.702917923, 10.6522262, 24.6112481, 41.460066, 61.4845987, 82.2350505, 110.076704, 146.38206, 193.178932, 253.833066, 331.109502, 427.773283, 541.18718, 680.746294, 861.739108, 1089.40221, 1379.72653, 1730.63037, 2156.01471, 2696.86236, 3381.72515, 4199.58622, 5168.73885, 6452.56138, 8132.36382, 9555.34022, 11649.8575, 14751.3427, 19195.2229, 25316.9252, 41692.8598, 58357.2249, 73585.2017, 85651.9715, 92832.7156, 86219.5158, 76059.3858, 63705.9781, 54508.0694, 48237.9088, 43003.1103, 38659.1547, 35061.5229, 32065.6957, 29527.1539, 26898.7233, 24491.283, 22308.898, 20355.6338, 18635.5556, 17152.7285, 15911.2179, 14882.0893, 13990.1027},
        {66.0053166, 65.4823761, 64.9292281, 64.3474318, 63.7385465, 63.1041312, 62.4457451, 61.7649474, 61.0632973, 60.3423538, 59.6036763, 58.8488238, 58.0793555, 57.2968306, 56.5028083, 55.6988476, 54.8865079, 54.0673481, 53.2429275, 52.4148054, 51.5845407, 50.7536927, 49.9238206, 49.0964835, 48.2732406, 47.455651, 46.645274, 45.8436686, 45.0523941, 44.2730096, 43.5070742, 42.7561472, 42.0217877, 41.3055549, 40.7183113, 40.2046654, 39.7079118, 39.2260626, 38.7571298, 38.2991255, 37.8500618, 37.4079506, 36.9708041, 36.5366344, 36.1034534, 35.6692732, 35.2321059, 34.7899635, 34.3408582, 33.8828019, 33.4138068, 32.8073614, 32.1887007, 31.5585362, 30.9175792, 30.2665414, 29.6061341, 28.9370689, 28.2600572, 27.5758105, 26.8850404, 26.1884582, 25.4867754, 24.7807036, 24.0709542, 23.3582387, 22.6432685, 21.9336616, 21.23687, 20.5393612, 19.8413991, 19.1432474, 18.4451698, 17.7474301, 17.0502922, 16.3540197, 15.6588764, 14.9651261, 14.2730326, 13.5828597, 12.894871, 12.2093303, 11.5265016, 10.8466483, 10.2162891, 9.61158745, 9.00815091, 8.40474223, 7.80012413, 7.19305931, 6.5823105, 5.9666404, 5.34481175, 4.71558724, 4.0777296, 3.43000154, 2.77116578, 2.09998503, 1.415222, 0.715639423, 0, -1.0743492, -2.110917, -3.11354409, -4.0860712, -5.03233902, -5.95618826, -6.86145963, -7.75199383, -8.63163157, -9.50421355, -10.3735805, -11.2435731, -12.118032, -13.0007981, -13.8957119, -14.8066142, -15.5756068, -16.183425, -16.9997472, -17.9937558, -19.1346333, -20.3915621, -21.7337244, -23.1303027, -24.5504794, -25.9634368, -27.3383573, -28.6444234, -29.8508174, -30.8351809, -31.691802, -32.4173253, -33.0083954, -33.7714818, -34.5483166, -35.0980121, -35.3242654, -35.1307736, -34.4212338, -33.0993431, -32.0516034, -29.8997927, -26.3553202, -21.1295951, -13.9340267, -4.18778614, 8.03241966, 22.9421219, 40.756852, 61.6921409, 82.4852538, 111.081202, 149.456205, 200.08262, 266.445756, 351.576128, 458.504251, 584.385957, 739.461598, 940.38027, 1192.30008, 1513.83418, 1901.963, 2372.59594, 2973.71657, 3738.5859, 4653.05085, 5738.07991, 7188.49311, 9100.77021, 10443.4668, 12614.5137, 16100.8258, 21389.3181, 28966.9057, 50413.6792, 72350.0841, 92489.7417, 108546.273, 118233.299, 109809.88, 96747.2395, 80810.953, 68972.0856, 60935.7073, 54237.9346, 48692.2851, 44112.2767, 40311.427, 37103.2537, 34371.925, 31493.4647, 28612.9088, 25875.2935, 23425.6548, 21409.0288, 19970.4517, 18356.4855, 17206.0995},
        {63.1686145, 62.7109816, 62.2308201, 61.7291199, 61.2068711, 60.6650635, 60.1046872, 59.5267321, 58.9321882, 58.3220453, 57.6972936, 57.0589228, 56.4079231, 55.7452843, 55.0719964, 54.3890494, 53.6974332, 52.9981378, 52.2921531, 51.5804691, 50.8640758, 50.1439631, 49.4211209, 48.6965393, 47.9712082, 47.2461176, 46.5222573, 45.8006174, 45.0821878, 44.3679586, 43.6589195, 42.9560607, 42.260372, 41.5728434, 40.9518152, 40.3683958, 39.7930045, 39.2247702, 38.6628217, 38.1062878, 37.5542973, 37.0059791, 36.460462, 35.9168748, 35.3743464, 34.8320056, 34.2889813, 33.7444022, 33.1973972, 32.6470951, 32.0926248, 31.4795413, 30.8617086, 30.2394169, 29.6129564, 28.9826173, 28.3486898, 27.7114642, 27.0712307, 26.4282793, 25.7829005, 25.1353842, 24.4860209, 23.8351006, 23.1829136, 22.5297501, 21.8759003, 21.2260151, 20.5846407, 19.9430737, 19.3013216, 18.6593919, 18.0172922, 17.37503, 16.7326128, 16.0900481, 15.4473435, 14.8045064, 14.1615444, 13.518465, 12.8752758, 12.2319842, 11.5885978, 10.945124, 10.3360527, 9.74342493, 9.14886738, 8.55126856, 7.94951699, 7.34250117, 6.72910962, 6.10823083, 5.47875332, 4.83956559, 4.18955615, 3.5276135, 2.85262616, 2.16348264, 1.45907143, 0.738281047, 0, -0.74865051, -1.52443599, -2.32716492, -3.15664575, -4.01268697, -4.89509703, -5.80368441, -6.73825758, -7.698625, -8.68459514, -9.69597647, -10.7325775, -11.7942066, -12.8806723, -13.9917831, -15.1273474, -16.1775168, -17.1264344, -18.2219423, -19.4403511, -20.7579712, -22.1511133, -23.5960879, -25.0692055, -26.5467767, -28.0051121, -29.4205222, -30.7693175, -32.0278086, -33.0693579, -34.0109718, -34.8598452, -35.6231731, -36.6403643, -37.7525069, -38.7120987, -39.4266703, -39.8037526, -39.7508764, -39.1755726, -39.1223812, -38.0157774, -35.5408338, -31.3826229, -25.2262171, -17.8533875, -7.57833336, 6.18804744, 24.0348571, 46.5511977, 71.9902152, 106.132026, 149.02794, 202.591575, 272.53876, 362.87821, 477.618638, 616.848129, 789.387078, 1008.33293, 1274.12395, 1611.86196, 2014.94013, 2502.78199, 3138.918, 3966.63283, 4983.78132, 6221.87002, 7878.98152, 10058.9736, 11270.2005, 13457.4856, 17276.9763, 23384.82, 32437.1641, 59044.1677, 86419.4637, 111730.697, 132145.511, 144831.551, 135432.195, 120322.198, 101608.177, 87494.9822, 77675.3454, 69301.0675, 62191.9268, 56167.7019, 51048.1709, 46653.1124, 43600.8265, 40046.7738, 36277.2677, 32578.6218, 29237.1495, 26539.1643, 24770.9796, 22301.8352, 20843.9884},
        {60.9968525, 60.5810023, 60.1459149, 59.6923331, 59.2209999, 58.7326581, 58.2280506, 57.7079205, 57.1730106, 56.6240639, 56.0618232, 55.4870315, 54.9004318, 54.3027668, 53.6947796, 53.0772131, 52.4508102, 51.8163138, 51.1744669, 50.5260123, 49.871693, 49.2122519, 48.5484319, 47.8809759, 47.210627, 46.5381279, 45.8642216, 45.1896511, 44.5151592, 43.8414889, 43.1693831, 42.4995847, 41.8328367, 41.1698819, 40.546521, 39.9452306, 39.3480662, 38.7546328, 38.1645358, 37.5773803, 36.9927717, 36.4103151, 35.8296157, 35.2502789, 34.6719097, 34.0941136, 33.5164956, 32.9386611, 32.3602152, 31.7807632, 31.1999103, 30.5955503, 29.9894705, 29.3817468, 28.7724552, 28.1616715, 27.5494717, 26.9359315, 26.321127, 25.705134, 25.0880283, 24.469886, 23.8507829, 23.2307948, 22.6099977, 21.9884675, 21.3662801, 20.74472, 20.1250426, 19.5048313, 18.8840835, 18.2627968, 17.6409687, 17.0185967, 16.3956784, 15.7722114, 15.1481931, 14.5236212, 13.8984931, 13.2728064, 12.6465587, 12.0197475, 11.3923703, 10.7644247, 10.1728133, 9.59830482, 9.02122448, 8.44037219, 7.85454784, 7.26255134, 6.6631826, 6.0552415, 5.43752797, 4.80884189, 4.16798317, 3.51375172, 2.84494743, 2.16037021, 1.45881997, 0.739096595, 0, -0.637212523, -1.3243473, -2.06004434, -2.84294365, -3.67168525, -4.54490914, -5.46125535, -6.41936388, -7.41787474, -8.45542794, -9.5306635, -10.6422214, -11.7887417, -12.9688644, -14.1812295, -15.4244771, -16.6085554, -17.7194344, -18.956366, -20.2989847, -21.7269254, -23.2198226, -24.7573111, -26.3190255, -27.8846005, -29.4336708, -30.945871, -32.4008358, -33.7781998, -34.9560006, -36.0527223, -37.0784788, -38.043384, -39.3085328, -40.7085486, -41.9844595, -43.0410852, -43.783245, -44.1157584, -43.9434449, -44.4198242, -43.8209764, -41.80741, -38.0396338, -32.1781562, -25.4323829, -15.5267012, -1.73439527, 16.6712505, 40.4169517, 68.0440116, 105.135618, 151.084439, 207.76402, 282.113041, 378.796038, 502.47755, 654.672938, 843.911637, 1082.15855, 1366.84756, 1728.93463, 2162.70333, 2690.80773, 3383.97402, 4290.86248, 5406.0256, 6764.09571, 8598.10048, 11028.844, 12258.7524, 14622.1426, 18903.8726, 25888.8004, 36361.7837, 67194.567, 99063.4001, 128731.419, 152961.761, 168517.562, 159228.199, 143413.074, 123378.363, 107953.952, 96863.7157, 87113.3157, 78562.5365, 71071.163, 64498.9798, 58705.7718, 55157.8487, 50737.7883, 45872.7947, 40990.0723, 36516.8253, 32880.2578, 30507.5742, 27178.6977, 25112.1428},
        {58.9297241, 58.4551898, 57.9742915, 57.4871706, 56.9939684, 56.4948264, 55.9898861, 55.4792888, 54.963176, 54.4416891, 53.9149696, 53.3831589, 52.8463984, 52.3048295, 51.7585937, 51.2078323, 50.6526869, 50.0932989, 49.5298096, 48.9623606, 48.3910931, 47.8161488, 47.2376689, 46.655795, 46.0706684, 45.4824306, 44.8912231, 44.2971871, 43.7004642, 43.1011959, 42.4995234, 41.8955883, 41.289532, 40.6814959, 40.0688128, 39.4530874, 38.8359584, 38.2176584, 37.59842, 36.9784757, 36.3580581, 35.7373999, 35.1167335, 34.4962916, 33.8763067, 33.2570115, 32.6386384, 32.0214201, 31.4055891, 30.7913781, 30.1790196, 29.5902048, 29.0032425, 28.4179001, 27.833945, 27.2511445, 26.669266, 26.088077, 25.5073449, 24.9268369, 24.3463207, 23.7655634, 23.1843326, 22.6023955, 22.0195197, 21.4354725, 20.8500213, 20.2577639, 19.6534222, 19.0474255, 18.4398764, 17.8308773, 17.2205308, 16.6089394, 15.9962056, 15.3824319, 14.7677209, 14.152175, 13.5358969, 12.9189889, 12.3015537, 11.6836937, 11.0655114, 10.4471095, 9.87377273, 9.32206213, 8.76799561, 8.21020943, 7.64733986, 7.07802315, 6.50089557, 5.91459339, 5.31775287, 4.70901027, 4.08700186, 3.45036389, 2.79773264, 2.12774436, 1.43903531, 0.730241774, 0, -0.718328755, -1.47246998, -2.26231391, -3.08775078, -3.94867083, -4.84496429, -5.77652139, -6.74323238, -7.74498749, -8.78167694, -9.85319099, -10.9594199, -12.1002538, -13.275583, -14.4852977, -15.7292883, -16.9443394, -18.1213265, -19.4028779, -20.7753613, -22.2251443, -23.7385947, -25.30208, -26.9019681, -28.5246264, -30.1564228, -31.7837248, -33.3929001, -34.9703164, -36.4790462, -37.9372939, -39.3384157, -40.6757679, -42.2554865, -43.914538, -45.4064805, -46.6308363, -47.4871276, -47.8748767, -47.6936059, -48.114412, -47.3782417, -45.1358307, -41.0379149, -34.7352302, -27.4941862, -16.9459267, -2.33726901, 17.0849693, 42.0739707, 70.9101909, 109.841761, 158.215638, 218.094077, 297.083067, 400.299572, 532.860557, 696.494051, 900.476168, 1157.78005, 1464.41408, 1856.63282, 2333.95064, 2921.99237, 3690.62555, 4689.40962, 5895.83044, 7340.76169, 9319.34675, 11977.2431, 13428.3235, 16162.6124, 21029.0775, 28876.6863, 40554.4065, 73905.9277, 108536.815, 141047.355, 168037.836, 186108.543, 178624.143, 164244.292, 145241.687, 130077.086, 118573.745, 107976.56, 98237.7044, 89309.3524, 81143.6772, 73692.8524, 69329.861, 63791.4225, 57613.95, 51333.8569, 45487.5566, 40611.4622, 37241.9871, 33042.0611, 29963.7439},
        {57.0382514, 56.483186, 55.9335163, 55.3888953, 54.8489758, 54.3134105, 53.7818524, 53.2539541, 52.7293687, 52.2077488, 51.6887473, 51.172017, 50.6572108, 50.1439815, 49.6319819, 49.1208648, 48.610283, 48.0998895, 47.5893369, 47.0782782, 46.5663661, 46.0532535, 45.5385932, 45.022038, 44.5032408, 43.9818543, 43.4575315, 42.929925, 42.3986879, 41.8634728, 41.3239326, 40.7797201, 40.2304882, 39.6758896, 39.0868396, 38.4779643, 37.8642351, 37.2462375, 36.624557, 35.9997788, 35.3724885, 34.7432714, 34.112713, 33.4813986, 32.8499138, 32.2188439, 31.5887743, 30.9602904, 30.3339777, 29.7104216, 29.0902075, 28.5188679, 27.9510666, 27.386415, 26.8245238, 26.2650043, 25.7074675, 25.1515244, 24.5967861, 24.0428636, 23.489368, 22.9359104, 22.3821017, 21.8275531, 21.2718756, 20.7146802, 20.155578, 19.5849654, 18.9934598, 18.3996772, 17.8038257, 17.2061139, 16.6067501, 16.0059428, 15.4039003, 14.8008312, 14.1969436, 13.5924462, 12.9875473, 12.3824552, 11.7773784, 11.1725254, 10.5681044, 9.96432397, 9.41653641, 8.89621828, 8.37418333, 7.84885045, 7.3186385, 6.78196637, 6.23725294, 5.68291708, 5.11737767, 4.53905359, 3.94636372, 3.33772693, 2.7115621, 2.06628812, 1.40032386, 0.712088191, 0, -0.805960148, -1.63159821, -2.47820895, -3.34708714, -4.23952756, -5.15682496, -6.10027411, -7.07116979, -8.07080677, -9.10047981, -10.1614837, -11.2551132, -12.382663, -13.545428, -14.7447029, -15.9817824, -17.2247969, -18.4703032, -19.7946112, -21.191909, -22.6563846, -24.1822262, -25.7636218, -27.3947594, -29.0698273, -30.7830133, -32.5285057, -34.3004926, -36.0931619, -37.9734872, -39.8361832, -41.6536023, -43.398097, -45.3053851, -47.2161363, -48.9007897, -50.252688, -51.1651739, -51.5315904, -51.2452802, -51.4380106, -50.387788, -47.7456551, -43.1626541, -36.2898275, -28.5661388, -17.4077289, -2.01865996, 18.397006, 44.6352068, 74.7185142, 115.605902, 166.940026, 231.032157, 315.641992, 426.082997, 567.668636, 741.124223, 957.394137, 1232.42801, 1562.75479, 1988.24272, 2513.73744, 3167.24769, 4016.03073, 5109.32812, 6404.44931, 7926.8016, 10038.8766, 12912.4921, 14711.9247, 17925.3855, 23419.6103, 32061.3347, 44717.2945, 79459.1939, 115647.558, 149847.88, 178625.654, 198546.373, 193097.977, 180641.883, 163348.251, 149114.945, 137844.91, 127054.545, 116766.122, 107001.913, 97784.1891, 89135.2206, 84171.469, 77623.26, 70163.5257, 62465.1982, 55201.2096, 49044.492, 44667.9776, 39522.9074, 35365.6076},
        {55.3408179, 54.7766647, 54.2196794, 53.6694624, 53.1256139, 52.5877341, 52.0554231, 51.5282813, 51.0059088, 50.4879059, 49.9738728, 49.4634096, 48.9561166, 48.4515941, 47.9494422, 47.4492611, 46.9506511, 46.4532124, 45.9565452, 45.4602497, 44.9639261, 44.4671747, 43.9695956, 43.4707891, 42.9703554, 42.4678947, 41.9630073, 41.4552933, 40.9443529, 40.4297864, 39.911194, 39.388176, 38.8603324, 38.3272636, 37.7593808, 37.1714926, 36.5787585, 35.981726, 35.3809426, 34.7769556, 34.1703125, 33.5615609, 32.9512481, 32.3399216, 31.7281288, 31.1164173, 30.5053344, 29.8954276, 29.2872444, 28.6813322, 28.0782385, 27.5178984, 26.9606179, 26.4060904, 25.8540098, 25.3040695, 24.7559631, 24.2093844, 23.6640268, 23.119584, 22.5757496, 22.0322172, 21.4886805, 20.9448329, 20.4003682, 19.8549799, 19.3083616, 18.7528895, 18.1811153, 17.6078244, 17.0331848, 16.4573646, 15.8805316, 15.3028538, 14.7244993, 14.1456359, 13.5664316, 12.9870545, 12.4076725, 11.8284535, 11.2495656, 10.6711767, 10.0934547, 9.51656769, 8.99965327, 8.51215422, 8.02280456, 7.5298586, 7.03157062, 6.52619491, 6.01198578, 5.4871975, 4.95008437, 4.3989007, 3.83190075, 3.24733885, 2.64346926, 2.01854629, 1.37082423, 0.698557367, 0, -0.675885449, -1.39642853, -2.16165628, -2.97159576, -3.82627402, -4.7257181, -5.66995506, -6.65901194, -7.69291579, -8.77169367, -9.89537262, -11.0639797, -12.2775419, -13.5360864, -14.8396401, -16.1882302, -17.5859583, -19.0334336, -20.5214666, -22.0509574, -23.6228064, -25.2379137, -26.8971794, -28.6015039, -30.3517872, -32.1489295, -33.9938312, -35.8873923, -37.8305131, -39.9367628, -42.0530606, -44.146506, -46.1841982, -48.3901429, -50.6029864, -52.592867, -54.2498124, -55.4638504, -56.1250086, -56.1233146, -56.5843787, -55.7865993, -53.3782596, -49.0076431, -42.3230329, -35.611739, -25.2232615, -10.1461272, 10.6311373, 38.1200053, 70.8466486, 115.345433, 171.122945, 240.49622, 331.52029, 449.673944, 600.43597, 783.615221, 1011.64904, 1303.16017, 1655.53752, 2111.0351, 2674.83085, 3376.97634, 4289.67006, 5465.7644, 6855.34295, 8484.61435, 10756.606, 13861.1951, 15971.6036, 19604.6192, 25648.1922, 34990.2732, 48518.8126, 84814.4946, 122649.353, 158488.154, 188795.665, 210036.653, 205288.819, 193328.707, 176358.399, 162416.578, 151395.003, 140753.516, 130505.472, 120664.224, 111243.128, 102255.538, 97189.2742, 90012.9019, 81572.8691, 72715.6233, 64287.6121, 57135.283, 52105.0834, 45912.805, 41007.2404},
        {53.6791712, 53.1133961, 52.5547182, 52.0027789, 51.4572194, 50.9176812, 50.3838056, 49.8552338, 49.3316073, 48.8125674, 48.2977554, 47.7868128, 47.2793807, 46.7751005, 46.2736137, 45.7745615, 45.2775853, 44.7823264, 44.2884261, 43.7955259, 43.303267, 42.8112908, 42.3192386, 41.8267518, 41.3334717, 40.8390396, 40.3430969, 39.845285, 39.3452451, 38.8426186, 38.3370469, 37.8281713, 37.3156331, 36.7990736, 36.2548502, 35.694736, 35.1307839, 34.5633909, 33.992954, 33.4198701, 32.8445362, 32.2673492, 31.6887061, 31.1090038, 30.5286394, 29.9480096, 29.3675116, 28.7875422, 28.2084985, 27.6307773, 27.0547756, 26.5072467, 25.96166, 25.4178409, 24.8756151, 24.3348082, 23.7952457, 23.2567533, 22.7191567, 22.1822812, 21.6459527, 21.1099967, 20.5742387, 20.0385044, 19.5026194, 18.9664093, 18.4296996, 17.8877817, 17.3360556, 16.7836987, 16.2308306, 15.6775708, 15.1240388, 14.5703542, 14.0166365, 13.4630053, 12.9095802, 12.3564806, 11.8038261, 11.2517363, 10.7003307, 10.1497288, 9.60005021, 9.05141447, 8.56675081, 8.11345281, 7.65815913, 7.198951, 6.73390969, 6.26111644, 5.7786525, 5.28459912, 4.77703754, 4.25404903, 3.71371482, 3.15411617, 2.57333433, 1.96945054, 1.34054606, 0.684702126, 0, -0.550161501, -1.16922809, -1.85602887, -2.60939291, -3.4281493, -4.31112713, -5.25715549, -6.26506346, -7.33368013, -8.46183458, -9.64835591, -10.8920732, -12.1918155, -13.546412, -14.9546917, -16.4154837, -17.9646115, -19.606189, -21.2553677, -22.9189039, -24.603554, -26.3160747, -28.0632222, -29.8517531, -31.6884238, -33.5799908, -35.5332105, -37.5548394, -39.6516339, -41.9738511, -44.33213, -46.6901769, -49.0116981, -51.5184294, -54.0450624, -56.3574808, -58.3419821, -59.8848639, -60.8724236, -61.1909587, -61.9645325, -61.4649654, -59.3363834, -55.2229123, -48.7686782, -43.0977796, -33.5043768, -18.7626023, 2.35341092, 31.0695301, 66.4051969, 114.489447, 174.714076, 249.408678, 346.900848, 472.825198, 632.816343, 825.647964, 1065.37491, 1373.53669, 1748.90494, 2235.34385, 2837.16591, 3586.17897, 4560.7801, 5817.98116, 7303.6326, 9045.92666, 11480.6473, 14813.0217, 17274.0414, 21346.8974, 27920.2058, 37882.5829, 52122.6447, 89396.6499, 128258.662, 165130.386, 196433.527, 218589.793, 214515.114, 203140.818, 186676.8, 173245.83, 162717.29, 152496.86, 142577.105, 132950.591, 123609.883, 114547.546, 109231.288, 101489.33, 92265.8666, 82505.0922, 73151.2022, 65148.3917, 59440.8558, 52266.5469, 46644.9003},
        {51.9804494, 51.4233022, 50.8717451, 50.3255254, 49.7843906, 49.2480882, 48.7163656, 48.1889702, 47.6656496, 47.1461512, 46.6302225, 46.1176108, 45.6080637, 45.1013286, 44.597153, 44.0952843, 43.5954699, 43.0974575, 42.6009943, 42.1058279, 41.6117057, 41.1183752, 40.6255837, 40.1330789, 39.6406081, 39.1479188, 38.6547585, 38.1608745, 37.6660145, 37.1699257, 36.6723557, 36.173052, 35.671762, 35.1682331, 34.648496, 34.1194451, 33.5881397, 33.0547722, 32.5195354, 31.9826217, 31.4442238, 30.9045343, 30.3637458, 29.8220508, 29.2796419, 28.7367118, 28.193453, 27.6500581, 27.1067198, 26.5636305, 26.0209829, 25.4895728, 24.9587597, 24.4285063, 23.8987755, 23.3695299, 22.8407322, 22.3123453, 21.7843318, 21.2566545, 20.7292761, 20.2021594, 19.675267, 19.1485618, 18.6220065, 18.0955637, 17.5691963, 17.0410393, 16.5092716, 15.9776255, 15.446182, 14.9150226, 14.3842283, 13.8538804, 13.3240602, 12.7948488, 12.2663274, 11.7385773, 11.2116797, 10.6857159, 10.1607669, 9.63691411, 9.11423868, 8.59282183, 8.13998198, 7.7207675, 7.29941864, 6.87383464, 6.44191473, 6.00155817, 5.5506642, 5.08713205, 4.60886097, 4.11375021, 3.599699, 3.06460659, 2.50637222, 1.92289513, 1.31207457, 0.671809781, 0, -0.460878457, -1.00835648, -1.6405195, -2.35545296, -3.15124229, -4.02597293, -4.97773033, -6.00459991, -7.10466711, -8.27601738, -9.51673615, -10.8249089, -12.1986209, -13.6359578, -15.135005, -16.6938478, -18.3704511, -20.1714545, -21.959502, -23.7455104, -25.5403965, -27.3550769, -29.2004683, -31.0874876, -33.0270513, -35.0300763, -37.1074792, -39.2701769, -41.5290859, -44.0566589, -46.643047, -49.2507061, -51.8420922, -54.6490267, -57.4992827, -60.148803, -62.479234, -64.3722221, -65.7094139, -66.3724555, -67.5035719, -67.3401768, -65.5172818, -61.6698986, -55.4330385, -50.6046641, -41.6160984, -27.0616151, -5.53548817, 24.3680086, 62.0194698, 113.347136, 177.640116, 257.274958, 360.931743, 494.460392, 663.71083, 866.562491, 1118.64726, 1444.2921, 1843.86972, 2362.29431, 3001.8425, 3795.84397, 4830.24595, 6166.84337, 7750.88464, 9613.94891, 12214.9525, 15771.0858, 18606.8626, 23123.3306, 30192.8372, 40687.7298, 55480.3558, 93201.5177, 132525.494, 169885.019, 201712.826, 224441.649, 221075.566, 210428.403, 194688.032, 181956.597, 172080.902, 162443.473, 153014.329, 143763.489, 134660.971, 125676.795, 119762.661, 111471.55, 101717.997, 91416.5384, 81481.7113, 72828.0513, 66370.0946, 58355.3232, 52047.3305},
        {50.1717908, 49.636305, 49.1038724, 48.5743826, 48.0477256, 47.5237911, 47.0024691, 46.4836492, 45.9672213, 45.4530753, 44.9411009, 44.431188, 43.9232264, 43.417106, 42.9127164, 42.4099476, 41.9086895, 41.4088317, 40.9102641, 40.4128766, 39.9165589, 39.4212009, 38.9266925, 38.4329233, 37.9397833, 37.4471623, 36.95495, 36.4630364, 35.9713111, 35.4796642, 34.9879853, 34.4961642, 34.0040909, 33.5116552, 33.015566, 32.5173706, 32.0186543, 31.51941, 31.0196308, 30.5193097, 30.0184398, 29.5170142, 29.0150258, 28.5124677, 28.009333, 27.5056146, 27.0013056, 26.4963992, 25.9908882, 25.4847657, 24.9780249, 24.4675366, 23.9564836, 23.4449267, 22.9329266, 22.420544, 21.9078396, 21.3948741, 20.8817083, 20.3684029, 19.8550186, 19.3416161, 18.8282561, 18.3149994, 17.8019066, 17.2890386, 16.776456, 16.2640596, 15.7517541, 15.23993, 14.7286584, 14.2180103, 13.7080569, 13.1988693, 12.6905184, 12.1830755, 11.6766116, 11.1711978, 10.6669051, 10.1638047, 9.66196762, 9.16146498, 8.66236785, 8.16474732, 7.74149971, 7.35475172, 6.96575472, 6.57221648, 6.17184476, 5.76234733, 5.34143196, 4.90680641, 4.45617845, 3.98725585, 3.49774637, 2.98535778, 2.44779785, 1.88277433, 1.28799501, 0.661167645, 0, -0.440126473, -0.972173241, -1.59432096, -2.30475029, -3.10164189, -3.98317641, -4.94753452, -5.99289686, -7.11744411, -8.31935691, -9.59681593, -10.9480018, -12.3710952, -13.8642768, -15.4257273, -17.0536272, -18.8131716, -20.7121146, -22.5970569, -24.4805393, -26.3751027, -28.2932881, -30.2476362, -32.2506878, -34.314984, -36.4530655, -38.6774732, -41.0007479, -43.4354306, -46.1570932, -48.9554674, -51.7941846, -54.6368763, -57.7407166, -60.9225656, -63.9210055, -66.611605, -68.8699328, -70.5715576, -71.592048, -73.1265966, -73.3295236, -71.8182105, -68.2100385, -62.1223892, -57.712796, -48.9234503, -34.2366954, -12.1348749, 18.899668, 58.314778, 112.227693, 179.827759, 263.600484, 372.761053, 513.503161, 692.020507, 905.698844, 1171.54162, 1516.16094, 1941.44449, 2493.01182, 3169.9605, 4006.9597, 5098.95271, 6513.21601, 8198.66546, 10191.8915, 12963.4733, 16738.5016, 19957.692, 24905.0293, 32423.2723, 43355.1799, 58543.5109, 96224.956, 135499.859, 172862.497, 204807.144, 227828.078, 225268.883, 215541.653, 200776.674, 188902.77, 179754.973, 170752.255, 161850.449, 153005.389, 144172.908, 135308.841, 128248.542, 119378.564, 109404.737, 99032.8955, 88968.8705, 79918.4946, 72587.5999, 63950.3237, 56983.2738},
        {48.1803337, 47.6823264, 47.1842123, 46.6860312, 46.1878226, 45.6896263, 45.1914819, 44.6934292, 44.1955079, 43.6977575, 43.2002179, 42.7029288, 42.2059297, 41.7092604, 41.2129606, 40.71707, 40.2216282, 39.726675, 39.23225, 38.7383929, 38.2451435, 37.7525414, 37.2606263, 36.7694378, 36.2790158, 35.7893998, 35.3006296, 34.8127448, 34.3257852, 33.8397904, 33.3548001, 32.870854, 32.3879919, 31.9062533, 31.4313084, 30.9602632, 30.4901561, 30.0208442, 29.5521843, 29.0840334, 28.6162486, 28.1486868, 27.6812049, 27.21366, 26.745909, 26.2778089, 25.8092166, 25.3399892, 24.8699836, 24.3990567, 23.9270656, 23.443798, 22.9593984, 22.473942, 21.9875041, 21.5001599, 21.0119848, 20.5230538, 20.0334424, 19.5432257, 19.052479, 18.5612775, 18.0696966, 17.5778114, 17.0856972, 16.5934293, 16.1010829, 15.6082398, 15.1144938, 14.6209377, 14.1276787, 13.6348242, 13.1424813, 12.6507573, 12.1597594, 11.6695949, 11.180371, 10.6921949, 10.2051739, 9.71941519, 9.23502606, 8.75211374, 8.27078545, 7.79114844, 7.39345694, 7.0360589, 6.67633899, 6.31180351, 5.93995879, 5.55831115, 5.16436688, 4.75563232, 4.32961378, 3.88381758, 3.41575002, 2.92291743, 2.40282612, 1.8529824, 1.2708926, 0.654063029, 0, -0.519995704, -1.11903795, -1.79662605, -2.55225929, -3.385437, -4.29565846, -5.28242299, -6.34522989, -7.48357845, -8.696968, -9.98489783, -11.3468672, -12.7823755, -14.290922, -15.872006, -17.5251268, -19.3024677, -21.2110542, -23.1312196, -25.0737525, -27.0494418, -29.069076, -31.143444, -33.2833345, -35.4995361, -37.8028377, -40.204028, -42.7138956, -45.3432293, -48.247061, -51.2390471, -54.2867035, -57.3575459, -60.7522811, -64.2718295, -67.6282603, -70.6891322, -73.3220037, -75.3944331, -76.7739792, -78.7587067, -79.3502963, -78.1364247, -74.7047689, -68.6430055, -64.0025787, -54.7914563, -39.4813731, -16.5440642, 15.5487356, 55.9164322, 111.440311, 181.203701, 267.890681, 381.536854, 528.877138, 716.646451, 942.39706, 1224.1335, 1589.87775, 2042.64189, 2628.62171, 3342.61976, 4220.51455, 5367.78549, 6857.96402, 8648.54146, 10782.9648, 13730.1615, 17718.3832, 21314.1541, 26663.1037, 34568.6971, 45834.3991, 61263.6749, 98462.8225, 137231.766, 174173.261, 205890.065, 228984.935, 227393.771, 218830.757, 205327.304, 194438.245, 186008.635, 177582.102, 169118.769, 160578.764, 151922.211, 143109.235, 134154.081, 124629.377, 114801.567, 104937.097, 95302.4114, 86163.9546, 77788.1715, 68822.7383, 61221.4733},
        {45.5932658, 45.1811964, 44.7642364, 44.3425806, 43.9164237, 43.4859605, 43.0513859, 42.6128945, 42.1706812, 41.7249408, 41.2758679, 40.8236575, 40.3685042, 39.9106028, 39.4501482, 38.9873351, 38.5223582, 38.0554124, 37.5866924, 37.116393, 36.644709, 36.1718352, 35.6979663, 35.2232971, 34.7480223, 34.2723369, 33.7964355, 33.3205128, 32.8447638, 32.3693832, 31.8945657, 31.4205061, 30.9473992, 30.4754398, 30.0116777, 29.5527361, 29.0951556, 28.6389084, 28.1839668, 27.7303032, 27.2778899, 26.8266992, 26.3767034, 25.9278748, 25.4801857, 25.0336085, 24.5881155, 24.143679, 23.7002712, 23.2578645, 22.8164313, 22.3872971, 21.9588348, 21.5307707, 21.1028308, 20.6747415, 20.2462289, 19.8170192, 19.3868386, 18.9554133, 18.5224694, 18.0877333, 17.650931, 17.2117888, 16.7700329, 16.3253895, 15.8775847, 15.4157366, 14.9292178, 14.4396332, 13.9473966, 13.452922, 12.9566231, 12.458914, 11.9602085, 11.4609204, 10.9614636, 10.4622521, 9.96369972, 9.46622031, 8.97022777, 8.47613599, 7.98435885, 7.49531024, 7.10889205, 6.77368157, 6.4370603, 6.09621355, 5.7483266, 5.39058479, 5.02017339, 4.63427773, 4.23008311, 3.80477483, 3.35553819, 2.8795585, 2.37402107, 1.8361112, 1.2630142, 0.651915361, 0, -0.949064175, -1.90128996, -2.86172576, -3.83541999, -4.82742103, -5.84277731, -6.88653722, -7.96374917, -9.07946156, -10.2387228, -11.4465813, -12.7080854, -14.0282836, -15.4122243, -16.8649558, -18.3915267, -19.9606465, -21.5721726, -23.3133479, -25.1814338, -27.173692, -29.2873839, -31.5197713, -33.8681155, -36.3296782, -38.901721, -41.5815054, -44.3662929, -47.2533451, -50.2871621, -53.4004461, -56.5762871, -59.7977752, -63.5099648, -67.4649638, -71.2916892, -74.8346418, -77.9383223, -80.4472313, -82.2058695, -84.6171252, -85.4928195, -84.3725511, -80.7959188, -74.3025212, -67.7331445, -56.5009032, -39.3200991, -14.9050343, 18.0299895, 57.2786076, 111.886657, 180.931089, 267.521555, 383.00114, 535.016288, 731.21344, 969.750147, 1269.91831, 1661.36135, 2150.99836, 2780.56772, 3530.19206, 4435.12516, 5624.0375, 7182.71977, 9096.80384, 11426.1133, 14597.7207, 18830.9619, 22810.6776, 28526.7747, 36729.7604, 48170.1418, 63598.4261, 99543.3774, 137032.682, 172872.282, 203868.12, 226826.14, 226836.567, 220231.539, 208838.022, 199305.236, 191551.338, 183485.206, 175125.459, 166490.715, 157599.592, 148470.709, 134857.163, 124136.861, 115367.412, 107606.425, 99911.5094, 91340.275, 80950.3309, 72392.8859, 64221.9877},
        {43.3985232, 43.0305056, 42.6559738, 42.2751723, 41.8883457, 41.4957386, 41.0975956, 40.6941612, 40.2856801, 39.8723969, 39.4545561, 39.0324023, 38.6061801, 38.1761341, 37.7425089, 37.305549, 36.8654991, 36.4226037, 35.9771075, 35.5292549, 35.0792906, 34.6274592, 34.1740053, 33.7191734, 33.2632081, 32.806354, 32.3488558, 31.8909579, 31.432905, 30.9749417, 30.5173125, 30.0602621, 29.604035, 29.1488757, 28.7027594, 28.2619028, 27.8224297, 27.3843339, 26.9476091, 26.512249, 26.0782473, 25.6455977, 25.214294, 24.7843299, 24.3556991, 23.9283953, 23.5024122, 23.0777436, 22.6543832, 22.2323246, 21.8115617, 21.4054307, 21.0002936, 20.5958547, 20.1918186, 19.7878897, 19.3837725, 18.9791715, 18.5737912, 18.167336, 17.7595104, 17.3500189, 16.938566, 16.5248561, 16.1085938, 15.6894834, 15.2672296, 14.8313079, 14.3714399, 13.9084259, 13.4426334, 12.9744301, 12.5041836, 12.0322612, 11.5590307, 11.0848596, 10.6101154, 10.1351658, 9.66037817, 9.18612021, 8.71275945, 8.24066344, 7.77019976, 7.30173595, 6.93684344, 6.62315917, 6.3071037, 5.98576034, 5.65621237, 5.31554311, 4.96083585, 4.5891739, 4.19764056, 3.78331914, 3.34329292, 2.87464522, 2.37445934, 1.83981857, 1.26780622, 0.6555056, 0, -1.08442396, -2.14887757, -3.20039728, -4.24601956, -5.29278088, -6.34771768, -7.41786644, -8.5102636, -9.63194565, -10.789949, -11.9913102, -13.2430656, -14.5522518, -15.9259051, -17.3710621, -18.8947592, -20.4463277, -22.0245604, -23.7670174, -25.6683699, -27.7232888, -29.9264453, -32.2725104, -34.7561551, -37.3720505, -40.1148678, -42.9792779, -45.9599519, -49.0515609, -52.2857886, -55.6067221, -58.9979287, -62.4429755, -66.3869844, -70.5827454, -74.6599672, -78.4637506, -81.8391963, -84.6314053, -86.6854784, -89.7207255, -91.1376269, -90.4145903, -87.0300236, -80.4623344, -72.6120317, -59.9298968, -41.2888122, -15.5616602, 18.3786767, 57.3092905, 112.024022, 181.328869, 268.67084, 386.972339, 544.901511, 751.1265, 1006.58119, 1329.42578, 1746.25782, 2259.59868, 2917.80564, 3704.86048, 4659.06282, 5910.75082, 7548.01543, 9561.20243, 12013.0614, 15324.7411, 19714.6592, 23978.187, 29975.7193, 38418.2671, 50016.8415, 65482.4534, 100608.742, 137253.433, 172356.88, 202859.438, 225701.46, 226731.734, 221376.558, 211329.815, 202639.049, 195274.818, 187393.657, 179056.824, 170325.582, 161261.189, 151924.905, 136348.695, 125151.901, 116881.646, 110085.053, 103309.246, 95101.3456, 84008.4764, 75531.1193, 66722.7294},
        {41.5158516, 41.1558253, 40.7903626, 40.4196688, 40.0439492, 39.6634092, 39.278254, 38.888689, 38.4949194, 38.0971506, 37.6955879, 37.2904365, 36.8819018, 36.470189, 36.0555036, 35.6380507, 35.2180357, 34.795664, 34.3711407, 33.9446713, 33.5164609, 33.086715, 32.6556388, 32.2234376, 31.7903168, 31.3564816, 30.9221373, 30.4874893, 30.0527429, 29.6181033, 29.1837758, 28.7499659, 28.3168787, 27.8847196, 27.4617604, 27.0440035, 26.6273543, 26.2117564, 25.7971532, 25.3834883, 24.9707052, 24.5587475, 24.1475587, 23.7370823, 23.3272618, 22.9180408, 22.5093628, 22.1011714, 21.69341, 21.2860223, 20.8789516, 20.4731871, 20.0676042, 19.6621236, 19.2566663, 18.8511531, 18.4455049, 18.0396426, 17.633487, 17.226959, 16.8199794, 16.4124692, 16.0043492, 15.5955402, 15.1859632, 14.775539, 14.3641884, 13.9493736, 13.5286158, 13.1069068, 12.6843269, 12.2609562, 11.8368751, 11.4121637, 10.9869025, 10.5611715, 10.1350512, 9.70862164, 9.28196318, 8.85515607, 8.42828056, 8.0014169, 7.57464535, 7.14804617, 6.81276649, 6.52143765, 6.22559646, 5.92236787, 5.60887684, 5.28224833, 4.93960731, 4.57807873, 4.19478756, 3.78685875, 3.35141727, 2.88558807, 2.38649612, 1.85126638, 1.2770238, 0.660893357, 0, -0.992943759, -1.98331398, -2.97741768, -3.98156188, -5.00205359, -6.04519983, -7.11730762, -8.22468396, -9.37363589, -10.5704704, -11.8214945, -13.1330153, -14.5113397, -15.9627747, -17.4936275, -19.1102049, -20.7898351, -22.5346853, -24.41661, -26.4357065, -28.592072, -30.8858038, -33.3169991, -35.8857552, -38.5921692, -41.4363386, -44.4183605, -47.5383321, -50.7963508, -54.2814265, -57.8721424, -61.5419219, -65.2641883, -69.3869688, -73.6963849, -77.8786639, -81.7948482, -85.30598, -88.2731016, -90.5572555, -94.2971839, -96.3830161, -96.2301573, -93.2540131, -86.8699886, -78.4231277, -64.9167877, -45.2839644, -18.4576534, 16.6291493, 56.0926615, 112.001635, 182.842755, 272.330122, 394.850482, 559.998781, 777.369965, 1051.511, 1398.21208, 1838.7703, 2366.58858, 3043.26639, 3868.79646, 4886.70735, 6214.29666, 7935.00081, 10027.7649, 12548.9351, 15937.9921, 20417.694, 24884.9924, 31092.4221, 39721.6744, 51454.4402, 66972.4108, 101601.928, 137718.87, 172343.764, 202497.14, 225199.524, 226733.21, 222015.786, 212669.883, 204421.131, 197268.055, 189499.599, 181199.768, 172452.566, 163341.999, 153952.072, 138746.628, 127685.355, 119406.818, 112549.583, 105752.217, 97653.2865, 86891.3567, 78239.3699, 68815.7516},
        {39.7446927, 39.3731033, 38.9988603, 38.6220684, 38.2428321, 37.8612561, 37.4774449, 37.0915031, 36.7035354, 36.3136464, 35.9219406, 35.5285226, 35.1334971, 34.7369687, 34.3390419, 33.9398214, 33.5394117, 33.1379175, 32.7354434, 32.3320939, 31.9279737, 31.5231874, 31.1178395, 30.7120347, 30.3058775, 29.8994726, 29.4929246, 29.0863381, 28.6798176, 28.2734679, 27.8673933, 27.4616987, 27.0564886, 26.6518675, 26.2546158, 25.8613598, 25.4686457, 25.0763613, 24.6843947, 24.2926338, 23.9009665, 23.5092809, 23.1174649, 22.7254063, 22.3329933, 21.9401138, 21.5466557, 21.152507, 20.7575556, 20.3616896, 19.9647968, 19.5523353, 19.1389357, 18.724799, 18.3101258, 17.895117, 17.4799733, 17.0648954, 16.6500843, 16.2357405, 15.822065, 15.4092585, 14.9975217, 14.5870555, 14.1780606, 13.7707378, 13.3652879, 12.9690017, 12.5889999, 12.2108605, 11.8343246, 11.4591334, 11.085028, 10.7117495, 10.3390392, 9.96663822, 9.59428763, 9.22172863, 8.84870236, 8.47494999, 8.10021266, 7.72423152, 7.34674774, 6.96750245, 6.66476157, 6.39735223, 6.1231577, 5.83937088, 5.54318462, 5.23179181, 4.90238532, 4.55215803, 4.17830279, 3.77801249, 3.34848001, 2.8868982, 2.39045996, 1.85635814, 1.28178563, 0.663935289, 0, -0.819540126, -1.66814275, -2.55045873, -3.47113893, -4.43483421, -5.44619541, -6.50987341, -7.63051907, -8.81278323, -10.0613168, -11.3807705, -12.7757954, -14.2510422, -15.8111618, -17.460805, -19.2046228, -21.062213, -23.0403615, -25.1099109, -27.2787147, -29.5546269, -31.9455012, -34.4591913, -37.1035511, -39.8864342, -42.8156946, -45.8991859, -49.144762, -52.5602765, -56.3117504, -60.1908757, -64.1580562, -68.1736956, -72.4689223, -76.8687776, -81.1261095, -85.1201043, -88.7299483, -91.8348276, -94.3139287, -98.6916972, -101.396981, -101.791415, -99.236636, -93.0942782, -84.4938246, -70.5871019, -50.2937839, -22.5335442, 13.7739434, 54.3323801, 112.170148, 185.673821, 278.81057, 406.941274, 580.311269, 809.165891, 1100.92179, 1469.29587, 1931.09089, 2470.86741, 3163.60793, 4027.49417, 5111.28619, 6515.55748, 8317.09833, 10479.9159, 13051.5605, 16493.1202, 21031.7629, 25655.0552, 32020.4738, 40784.2791, 52602.7314, 68132.0913, 102333.72, 137982.502, 172158.422, 201941.466, 224411.618, 226142.224, 221724.335, 212740.782, 204752.169, 197766.793, 190135.541, 181949.578, 173300.071, 164278.185, 154975.084, 141463.589, 130892.048, 122321.961, 114814.828, 107432.148, 99235.4226, 89286.151, 80406.7804, 70499.6233},
        {37.6880581, 37.2950834, 36.9044237, 36.5159781, 36.1296461, 35.7453268, 35.3629195, 34.9823235, 34.603438, 34.2261622, 33.8503955, 33.4760371, 33.1029862, 32.7311421, 32.3604041, 31.9906715, 31.6218434, 31.2538192, 30.8864981, 30.5197793, 30.1535622, 29.787746, 29.4222299, 29.0569133, 28.6916953, 28.3264752, 27.9611523, 27.5956259, 27.2297952, 26.8635594, 26.4968179, 26.1294698, 25.7614145, 25.3925512, 25.018399, 24.6411393, 24.2630063, 23.8840414, 23.504286, 23.1237815, 22.7425693, 22.3606909, 21.9781875, 21.5951008, 21.2114719, 20.8273424, 20.4427537, 20.0577471, 19.6723641, 19.286646, 18.9006343, 18.5109231, 18.1210757, 17.7312084, 17.3414373, 16.9518785, 16.5626482, 16.1738625, 15.7856376, 15.3980896, 15.0113346, 14.6254887, 14.2406682, 13.8569891, 13.4745676, 13.0935198, 12.7139619, 12.3410903, 11.9799793, 11.6202676, 11.2617418, 10.9041887, 10.5473953, 10.1911482, 9.83523438, 9.47944053, 9.12355347, 8.76736003, 8.410647, 8.0532012, 7.69480942, 7.33525848, 6.97433517, 6.61182631, 6.33206688, 6.09079125, 5.84271755, 5.58488886, 5.31434821, 5.02813866, 4.72330328, 4.3968851, 4.04592719, 3.6674726, 3.25856439, 2.8162456, 2.3375593, 1.81954854, 1.25925636, 0.653725831, 0, -0.888986312, -1.79226306, -2.71612446, -3.66686474, -4.65077812, -5.67415883, -6.74330109, -7.86449913, -9.04404717, -10.2882394, -11.6033701, -12.9957335, -14.4716238, -16.0373352, -17.6991619, -19.4633982, -21.319123, -23.2701711, -25.3617762, -27.5965436, -29.9770784, -32.5059859, -35.1858713, -38.0193398, -41.0089967, -44.1574472, -47.4672965, -50.9411498, -54.5816124, -58.5126105, -62.5709439, -66.7228216, -70.9344525, -75.5040087, -80.2317175, -84.8292827, -89.1633245, -93.1004628, -96.5073175, -99.2505087, -103.882485, -106.766611, -107.24402, -104.655842, -98.3432114, -89.3097349, -74.818449, -53.7948663, -25.1644997, 12.1471383, 53.1723928, 112.412734, 189.587643, 289.201581, 425.528336, 608.455468, 847.870536, 1148.34123, 1526.28403, 2003.91896, 2571.0595, 3298.84863, 4200.44024, 5324.01897, 6777.12712, 8640.06086, 10890.1068, 13580.8549, 17143.5969, 21795.975, 26591.3103, 33090.6926, 41914.7985, 53684.3046, 69019.8875, 102246.698, 136854.82, 170038.811, 198993.229, 220912.634, 222907.214, 218978.808, 210643.845, 203261.549, 196827.246, 189735.094, 182060.474, 173878.77, 165265.362, 156295.634, 144400.602, 134249.855, 125291.115, 116972.104, 108740.544, 100044.155, 90330.6608, 81805.0053, 71672.3795},
        {35.468434, 35.0718986, 34.6814062, 34.2966902, 33.9174843, 33.5435219, 33.1745365, 32.8102616, 32.4504308, 32.0947776, 31.7430355, 31.394938, 31.0502186, 30.7086109, 30.3698483, 30.0336644, 29.6997927, 29.3679668, 29.03792, 28.7093861, 28.3820983, 28.0557904, 27.7301958, 27.4050479, 27.0800805, 26.7550268, 26.4296206, 26.1035952, 25.7766842, 25.4486212, 25.1191395, 24.7879729, 24.4548547, 24.1195185, 23.7665103, 23.4034769, 23.0382475, 22.6710484, 22.302106, 21.9316468, 21.5598971, 21.1870832, 20.8134316, 20.4391686, 20.0645206, 19.689714, 19.3149752, 18.9405305, 18.5666063, 18.1934291, 17.8212251, 17.4648512, 17.1095862, 16.7553392, 16.4020197, 16.0495366, 15.6977993, 15.346717, 14.9961989, 14.6461542, 14.2964922, 13.9471221, 13.597953, 13.2488943, 12.899855, 12.5507446, 12.2014721, 11.8505528, 11.4965354, 11.1422041, 10.7875586, 10.4325983, 10.077323, 9.72173206, 9.36582523, 9.00960203, 8.65306207, 8.29620494, 7.9390302, 7.58153747, 7.22372632, 6.86559634, 6.50714711, 6.14837823, 5.8889236, 5.67686996, 5.45910656, 5.23239965, 4.9935155, 4.73922036, 4.4662805, 4.17146219, 3.85153168, 3.50325524, 3.12339912, 2.70872959, 2.25601292, 1.76201536, 1.22350318, 0.63724264, 0, -1.05041048, -2.08414567, -3.11022819, -4.13768065, -5.17552569, -6.23278591, -7.31848395, -8.44164242, -9.61128395, -10.8364312, -12.1261067, -13.4893331, -14.935133, -16.4725292, -18.1105441, -19.8582004, -21.6569067, -23.5060266, -25.567519, -27.8359178, -30.305757, -32.9715705, -35.8278922, -38.8692559, -42.0901957, -45.4852453, -49.0489388, -52.77581, -56.6603928, -60.7602239, -64.9837335, -69.3065546, -73.7043202, -78.5956929, -83.7347909, -88.7575915, -93.5068192, -97.825198, -101.555452, -104.540307, -109.173688, -111.970665, -112.274812, -109.429707, -102.778924, -93.5826238, -78.7886517, -57.2614378, -27.8654117, 10.5349966, 52.5922727, 113.848842, 195.501507, 302.53447, 447.664539, 640.136736, 889.196089, 1194.84603, 1577.67354, 2068.34732, 2666.51651, 3433.6742, 4370.11712, 5521.43517, 7009.96507, 8922.81551, 11259.8083, 14085.4221, 17784.1877, 22559.5961, 27523.3835, 34138.7585, 42987.1275, 54649.8968, 69708.4731, 101757.183, 135111.282, 167098.947, 195048.355, 216287.681, 218554.178, 215160.897, 207546.409, 200825.051, 194980.405, 188464.486, 181337.456, 173659.477, 165490.708, 156891.312, 146167.085, 136382.324, 127224.603, 118381.494, 109540.569, 100389.4, 90615.5602, 82516.4056, 72497.9256},
        {33.1454633, 32.7952264, 32.4498892, 32.1092077, 31.7729377, 31.4408355, 31.1126569, 30.7881579, 30.4670946, 30.1492229, 29.8342989, 29.5220786, 29.2123179, 28.9047728, 28.5991995, 28.2953537, 27.9929916, 27.6918692, 27.3917424, 27.0923673, 26.7934999, 26.4948961, 26.1963119, 25.8975034, 25.5982266, 25.2982374, 24.9972919, 24.695146, 24.3915558, 24.0862773, 23.7790664, 23.4696792, 23.1578716, 22.8433997, 22.5097232, 22.1650889, 21.8179398, 21.4685605, 21.117236, 20.7642511, 20.4098907, 20.0544396, 19.6981826, 19.3414046, 18.9843904, 18.627425, 18.270793, 17.9147795, 17.5596691, 17.2057468, 16.8532975, 16.5261524, 16.2005396, 15.8762333, 15.553008, 15.230638, 14.9088978, 14.5875618, 14.2664044, 13.9451999, 13.6237228, 13.3017474, 12.9790483, 12.6553997, 12.3305761, 12.0043518, 11.6765014, 11.3399726, 10.9878771, 10.6340687, 10.2787643, 9.9221808, 9.56453514, 9.20604429, 8.84692516, 8.48739471, 8.12766987, 7.76796758, 7.40850479, 7.04949843, 6.69116543, 6.33372275, 5.97738732, 5.62237607, 5.38599974, 5.20746547, 5.024573, 4.83373937, 4.63138158, 4.41391666, 4.17776164, 3.91933353, 3.63504935, 3.32132613, 2.97458088, 2.59123062, 2.16769238, 1.70038318, 1.18572004, 0.620119969, 0, -1.19036312, -2.34046612, -3.46137147, -4.56414163, -5.65983906, -6.75952624, -7.87426562, -9.01511966, -10.1931508, -11.4194216, -12.7049944, -14.0609317, -15.4982961, -17.0281498, -18.6615555, -20.4095755, -22.1978623, -24.0252767, -26.0960709, -28.4030052, -30.9388398, -33.6963352, -36.6682514, -39.8473489, -43.2263879, -46.7981287, -50.5553316, -54.4907569, -58.5971648, -62.9270136, -67.3914763, -71.9654041, -76.6236476, -81.8463519, -87.355721, -92.7392136, -97.8200924, -102.42162, -106.36706, -109.479674, -113.926783, -116.474183, -116.486517, -113.328427, -106.364557, -97.8448303, -83.5272882, -62.0552524, -32.072045, 7.77901163, 52.6384102, 117.676573, 204.478588, 318.110175, 470.743354, 671.359563, 928.94024, 1240.73447, 1630.12248, 2133.28344, 2756.22449, 3554.52578, 4520.3553, 5697.85079, 7219.20981, 9176.20901, 11585.0974, 14518.0703, 18324.258, 23196.1408, 28312.6203, 35035.6905, 43908.686, 55474.9412, 70277.7905, 101221.683, 133398.56, 164261.477, 191263.49, 211857.655, 214387.377, 211488.463, 204518.105, 198263.594, 192724.375, 186470.683, 179573.424, 172103.501, 164131.82, 155729.284, 145146.844, 135652.988, 126850.971, 118344.044, 109735.46, 100628.471, 90626.3293, 82554.2434, 73119.6933},
        {30.7355012, 30.5164282, 30.2924738, 30.0637561, 29.8303933, 29.5925034, 29.3502047, 29.1036154, 28.8528534, 28.5980371, 28.3392845, 28.0767139, 27.8104433, 27.5405909, 27.2672749, 26.9906133, 26.7107244, 26.4277264, 26.1417372, 25.8528752, 25.5612585, 25.2670051, 24.9702333, 24.6710613, 24.369607, 24.0659888, 23.7603247, 23.452733, 23.1433317, 22.832239, 22.519573, 22.205452, 21.889994, 21.5733173, 21.2569501, 20.9402758, 20.6226609, 20.3041778, 19.9848989, 19.6648965, 19.344243, 19.0230109, 18.7012724, 18.3791, 18.0565661, 17.7337431, 17.4107033, 17.0875191, 16.764263, 16.4410072, 16.1178243, 15.8039924, 15.4901785, 15.1762554, 14.862096, 14.547573, 14.2325593, 13.9169277, 13.600551, 13.283302, 12.9650536, 12.6456786, 12.3250497, 12.0030399, 11.6795219, 11.3543686, 11.0274527, 10.691835, 10.3407396, 9.98808905, 9.63419763, 9.27937978, 8.92394994, 8.56822253, 8.21251196, 7.85713267, 7.50239907, 7.14862558, 6.79612664, 6.44521666, 6.09621006, 5.74942128, 5.40516472, 5.06375482, 4.85768803, 4.7184075, 4.57576762, 4.42579326, 4.26450927, 4.08794051, 3.89211185, 3.67304815, 3.42677426, 3.14931506, 2.8366954, 2.48494014, 2.09007414, 1.64812226, 1.15510938, 0.607060335, 0, -1.22132808, -2.40515923, -3.56243994, -4.70411671, -5.84113601, -6.98444434, -8.14498819, -9.33371405, -10.5615684, -11.8394978, -13.1784486, -14.5893674, -16.0832007, -17.6708949, -19.3633966, -21.1716522, -23.0819271, -25.1016419, -27.2975695, -29.6753675, -32.2406937, -34.9992056, -37.9565611, -41.1184177, -44.4904331, -48.078265, -51.8875711, -55.924009, -60.1932365, -64.8830621, -69.7502039, -74.7456743, -79.8204856, -85.357404, -91.0915649, -96.6429696, -101.833105, -106.483456, -110.41551, -113.450753, -117.543865, -119.733905, -119.424997, -116.021264, -108.926826, -102.504403, -89.9598732, -69.4577092, -39.1623852, 2.76162536, 53.4037958, 125.146437, 217.680334, 335.415186, 492.395135, 698.338653, 962.964216, 1285.76754, 1689.0132, 2206.1179, 2838.91987, 3649.07325, 4635.93818, 5846.06935, 7405.98993, 9405.51622, 11858.3503, 14834.5752, 18683.3595, 23596.005, 28842.4131, 35677.0404, 44610.135, 56151.9453, 70812.7197, 100955.057, 132273.768, 162316.012, 188628.95, 208759.742, 211556.694, 209065.059, 202560.078, 196359.569, 190534.421, 183887.781, 176553.924, 168667.125, 160361.658, 151771.799, 139571.247, 130249.102, 122775.678, 116121.287, 109256.242, 101150.856, 90775.4411, 81884.4288, 73666.9211},
        {28.6653696, 28.5437409, 28.4091753, 28.2621035, 28.1029561, 27.9321638, 27.7501571, 27.5573667, 27.3542231, 27.141157, 26.918599, 26.6869797, 26.4467297, 26.1982797, 25.9420601, 25.6785018, 25.4080351, 25.1310909, 24.8480996, 24.5594919, 24.2656985, 23.9671498, 23.6642766, 23.3575094, 23.0472788, 22.7340156, 22.4181502, 22.1001132, 21.7803354, 21.4592473, 21.1372795, 20.8148626, 20.4924273, 20.1704041, 19.8662818, 19.571603, 19.2777057, 18.9844668, 18.6917633, 18.3994723, 18.1074709, 17.8156361, 17.523845, 17.2319745, 16.9399018, 16.6475038, 16.3546577, 16.0612405, 15.7671291, 15.4722007, 15.1763323, 14.8744649, 14.5715185, 14.2674773, 13.9623253, 13.6560466, 13.3486251, 13.040045, 12.7302903, 12.419345, 12.1071932, 11.793819, 11.4792064, 11.1633394, 10.8462021, 10.5277786, 10.2080529, 9.88184004, 9.54407923, 9.20541515, 8.86616694, 8.52665371, 8.18719462, 7.84810879, 7.50971536, 7.17233345, 6.83628222, 6.50188078, 6.16944827, 5.83930383, 5.51176659, 5.18715568, 4.86579025, 4.54798941, 4.37466614, 4.27288546, 4.16802254, 4.05583399, 3.93207639, 3.79250635, 3.63288045, 3.44895528, 3.23648745, 2.99123355, 2.70895016, 2.38539389, 2.01632132, 1.59748906, 1.12465369, 0.593571803, 0, -1.25505813, -2.46843675, -3.65189228, -4.81718116, -5.9760598, -7.14028465, -8.32161212, -9.53179864, -10.7826007, -12.0857746, -13.4530768, -14.8962639, -16.4270921, -18.057318, -19.7986979, -21.6629883, -23.6269441, -25.6973216, -27.9650471, -30.4343769, -33.109567, -35.9948736, -39.0945527, -42.4128604, -45.9540529, -49.7223863, -53.7221167, -57.9575001, -62.4327927, -67.3363119, -72.4207633, -77.6351848, -82.9286139, -88.6901696, -94.6488489, -100.416294, -105.809519, -110.645536, -114.74136, -117.914003, -122.230154, -124.572468, -124.317804, -120.843021, -113.52498, -107.247941, -94.5045115, -73.2947016, -41.6185205, 2.5240226, 56.7528353, 132.801426, 229.843633, 352.161036, 514.463514, 726.778855, 999.134845, 1331.93074, 1747.01115, 2276.72457, 2920.00975, 3743.66877, 4754.94728, 6002.50571, 7605.17956, 9649.59282, 12145.0284, 15158.059, 19038.841, 23977.1198, 29332.3366, 36252.2801, 45214.109, 56694.9817, 71172.0566, 100430.645, 130812.715, 159968.386, 185547.778, 205201.012, 208297.104, 206288.014, 200372.906, 194549.571, 188909.385, 182377.629, 175103.982, 167238.125, 158929.737, 150328.5, 136863.909, 127215.356, 120083.085, 114167.339, 108168.362, 100786.397, 90721.6875, 81660.9202, 74214.7715},
        {26.9443745, 26.8616956, 26.7623784, 26.6470131, 26.5161898, 26.3704987, 26.21053, 26.0368738, 25.8501204, 25.6508599, 25.4396825, 25.2171783, 24.9839376, 24.7405505, 24.4876071, 24.2256977, 23.9554124, 23.6773415, 23.392075, 23.1002031, 22.802316, 22.499004, 22.1908571, 21.8784655, 21.5624195, 21.2433091, 20.9217246, 20.5982562, 20.2734939, 19.948028, 19.6224487, 19.2973461, 18.9733104, 18.6509318, 18.3561069, 18.0762403, 17.7984326, 17.5224526, 17.2480694, 16.9750518, 16.7031688, 16.4321894, 16.1618824, 15.8920167, 15.6223614, 15.3526853, 15.0827574, 14.8123466, 14.5412219, 14.2691521, 13.9959062, 13.7077585, 13.418265, 13.1274874, 12.835487, 12.5423253, 12.2480638, 11.9527639, 11.6564872, 11.359295, 11.0612489, 10.7624104, 10.4628408, 10.1626017, 9.86175456, 9.56036081, 9.25848195, 8.95333007, 8.64218567, 8.33098685, 8.01997979, 7.70941066, 7.39952565, 7.09057095, 6.78279273, 6.47643717, 6.17175047, 5.86897881, 5.56836836, 5.2701653, 4.97461583, 4.68196613, 4.39246237, 4.10635074, 3.96615532, 3.89799071, 3.82626147, 3.74659657, 3.65462502, 3.5459758, 3.4162779, 3.26116033, 3.07625206, 2.85718209, 2.59957941, 2.29907302, 1.9512919, 1.55186505, 1.09642145, 0.580590105, 0, -1.29221095, -2.53108262, -3.73015372, -4.90296297, -6.0630491, -7.22395082, -8.39920686, -9.60235593, -10.8469368, -12.1464881, -13.5145486, -14.964657, -16.5103521, -18.1651725, -19.942657, -21.8563443, -23.7992577, -25.7677194, -28.0478626, -30.6274012, -33.4940493, -36.6355208, -40.0395297, -43.6937899, -47.5860156, -51.7039206, -56.035219, -60.5676248, -65.2888519, -70.2418153, -75.3387877, -80.5509227, -85.849374, -91.7517583, -97.9559979, -104.014291, -109.733853, -114.921898, -119.385641, -122.932297, -128.040974, -131.033809, -131.195253, -127.809759, -120.161781, -111.93924, -96.9222525, -73.2944044, -39.2392813, 7.0595312, 62.1312861, 139.541645, 239.583952, 367.038522, 535.846578, 755.836282, 1036.8358, 1378.1462, 1802.51525, 2344.17482, 3002.4307, 3845.75185, 4886.92859, 6175.48453, 7823.18587, 9913.22845, 12450.9232, 15499.2734, 19406.8609, 24360.728, 29809.2544, 36791.7858, 45751.0556, 57129.7971, 71370.7439, 99636.4267, 128969.832, 157133.744, 181890.947, 201004.225, 204364.198, 202853.923, 197609.967, 192519.84, 187631.195, 181845.463, 175267.608, 168002.597, 160155.395, 151830.967, 138002.965, 127658.815, 119717.995, 113099.983, 106724.254, 99510.2874, 90377.5588, 81903.3591, 74724.0332},
        {24.9812298, 24.9102365, 24.8230683, 24.7203038, 24.6025211, 24.4702987, 24.3242149, 24.1648481, 23.9927765, 23.8085785, 23.6128326, 23.4061169, 23.1890099, 22.9620899, 22.7259352, 22.4811242, 22.2282352, 21.9678465, 21.7005366, 21.4268837, 21.1474662, 20.8628625, 20.5736508, 20.2804095, 19.983717, 19.6841515, 19.3822915, 19.0787153, 18.7740012, 18.4687276, 18.1634728, 17.8588151, 17.5553329, 17.2536046, 16.9796423, 16.7207725, 16.4640159, 16.2091255, 15.9558543, 15.7039552, 15.4531812, 15.2032852, 14.9540202, 14.7051392, 14.4563951, 14.207541, 13.9583297, 13.7085142, 13.4578475, 13.2060825, 12.9529723, 12.6825517, 12.4106324, 12.1373083, 11.8626729, 11.58682, 11.3098433, 11.0318364, 10.7528932, 10.4731072, 10.1925722, 9.91138194, 9.62963004, 9.34741023, 9.06481622, 8.78194169, 8.49888036, 8.21383558, 7.92505609, 7.63653428, 7.34848641, 7.06112871, 6.77467743, 6.48934882, 6.20535912, 5.92292456, 5.64226141, 5.3635859, 5.08711427, 4.81306277, 4.54164765, 4.27308514, 4.0075915, 3.74538297, 3.63344226, 3.59551535, 3.5535842, 3.50310218, 3.43952269, 3.35829912, 3.25488486, 3.12473329, 2.96329782, 2.76603182, 2.5283887, 2.24582183, 1.91378461, 1.52773042, 1.08311267, 0.57538473, 0, -1.25026008, -2.45751028, -3.63505646, -4.7962045, -5.95426027, -7.12252964, -8.31431848, -9.54293266, -10.8216781, -12.1638605, -13.582786, -15.0917602, -16.7040892, -18.4330787, -20.2920346, -22.2942629, -24.3278143, -26.3881011, -28.7717442, -31.4652091, -34.4549612, -37.7274662, -41.2691894, -45.0665964, -49.1061528, -53.3743239, -57.8575754, -62.5423727, -67.4151813, -72.5015708, -77.7345645, -83.0888966, -88.5393015, -94.6507657, -101.102897, -107.417904, -113.393444, -118.827177, -123.51676, -127.259853, -132.624999, -135.79566, -136.027362, -132.575636, -124.69601, -116.147639, -100.556519, -76.0522727, -40.7645217, 7.17711159, 64.3169613, 144.361058, 247.216075, 377.690376, 550.599936, 776.267517, 1065.01588, 1416.58309, 1854.28221, 2412.97328, 3093.59799, 3962.87178, 5035.89949, 6361.25786, 8044.24001, 10164.1053, 12716.4489, 15758.8394, 19655.6143, 24597.585, 30163.8632, 37251.0683, 46261.7543, 57598.475, 71663.7844, 99069.9233, 127457.337, 154676.158, 178576.519, 197008.552, 200206.516, 198713.665, 193644.077, 188916.502, 184534.835, 179334.357, 173377.393, 166726.267, 159443.307, 151590.836, 138192.385, 127875.985, 119696.449, 112708.589, 105967.216, 98527.1439, 89443.1834, 80988.8684, 74832.5641},
        {22.7887268, 22.7191435, 22.6363215, 22.5407189, 22.4327934, 22.3130032, 22.1818062, 22.0396603, 21.8870235, 21.7243539, 21.5521093, 21.3707478, 21.1807273, 20.9825058, 20.7765412, 20.5632916, 20.3432148, 20.116769, 19.8844119, 19.6466017, 19.4037963, 19.1564536, 18.9050316, 18.6499884, 18.3917818, 18.1308698, 17.8677105, 17.6027617, 17.3364815, 17.0693278, 16.8017585, 16.5342318, 16.2672055, 16.0011376, 15.7566527, 15.5237014, 15.2919916, 15.0613269, 14.8315108, 14.6023469, 14.3736386, 14.1451895, 13.9168032, 13.6882831, 13.4594327, 13.2300556, 12.9999554, 12.7689354, 12.5367993, 12.3033506, 12.0683928, 11.8187259, 11.5674389, 11.3146171, 11.060346, 10.8047109, 10.5477974, 10.2896907, 10.0304764, 9.77023987, 9.50906647, 9.24704164, 8.98425079, 8.72077935, 8.45671272, 8.19213632, 7.92713556, 7.65937843, 7.38659093, 7.11384428, 6.84138059, 6.56944199, 6.29827061, 6.02810858, 5.75919802, 5.49178107, 5.22609985, 4.96239648, 4.70091309, 4.44189182, 4.18557478, 3.93220411, 3.68202194, 3.43527038, 3.34707519, 3.33697892, 3.32266253, 3.29934188, 3.26223283, 3.20655122, 3.12751292, 3.02033378, 2.88022966, 2.70241641, 2.48210989, 2.21452596, 1.89488046, 1.51838926, 1.08026822, 0.575733176, 0, -1.13340224, -2.2548397, -3.37548638, -4.50651627, -5.65910337, -6.84442167, -8.07364517, -9.35794787, -10.7085038, -12.1364868, -13.6530711, -15.2694305, -16.9967391, -18.8461709, -20.8288999, -22.9561, -25.190134, -27.535203, -30.1128871, -32.923901, -35.9689589, -39.2487755, -42.764065, -46.5155421, -50.503921, -54.7299163, -59.1942424, -63.8976138, -68.8407449, -74.1679053, -79.6836174, -85.3455292, -91.1112885, -97.4939654, -104.173497, -110.681707, -116.809617, -122.348248, -127.088623, -130.821761, -135.910936, -138.792058, -138.752884, -135.081166, -127.064661, -119.858416, -105.416069, -81.5585528, -46.1067992, 3.11826021, 63.8740795, 148.165536, 253.997075, 385.704134, 560.547925, 789.98625, 1085.47691, 1448.94704, 1903.55118, 2482.84096, 3189.25569, 4086.25514, 5189.09179, 6544.42723, 8252.1441, 10387.6957, 12934.7385, 15944.9264, 19806.5025, 24717.3159, 30402.0305, 37604.4787, 46686.3046, 58009.1524, 71934.6663, 98614.2253, 126172.305, 152523.116, 175580.868, 193259.773, 196000.415, 194173.047, 188894.862, 184159.578, 179945.763, 175033.431, 169461.079, 163267.201, 156490.292, 149168.847, 136776.004, 127185.719, 119449.063, 112617.107, 105740.921, 97871.5763, 88060.1426, 79069.4807, 74524.9983},
        {20.779917, 20.7317989, 20.6709027, 20.5976588, 20.5124976, 20.4158496, 20.3081452, 20.1898148, 20.0612889, 19.9229979, 19.7753723, 19.6188424, 19.4538387, 19.2807917, 19.1001317, 18.9122892, 18.7176947, 18.5167786, 18.3099712, 18.0977031, 17.8804047, 17.6585063, 17.4324386, 17.2026318, 16.9695164, 16.7335228, 16.4950815, 16.254623, 16.0125775, 15.7693757, 15.5254479, 15.2812245, 15.037136, 14.7936128, 14.5693909, 14.355363, 14.1422017, 13.9297432, 13.717824, 13.5062805, 13.2949491, 13.0836661, 12.8722679, 12.6605909, 12.4484715, 12.235746, 12.0222509, 11.8078225, 11.5922972, 11.3755114, 11.1573015, 10.9271311, 10.6954342, 10.462272, 10.2277058, 9.99179685, 9.75460633, 9.5161955, 9.27662561, 9.03595789, 8.7942536, 8.55157396, 8.30798022, 8.06353363, 7.81829542, 7.57232683, 7.32568911, 7.07504196, 6.81712678, 6.55902019, 6.30100395, 6.04335981, 5.78636951, 5.5303148, 5.27547743, 5.02213915, 4.77058171, 4.52108685, 4.27393633, 4.02941189, 3.78779528, 3.54936825, 3.31441254, 3.08320991, 3.01884711, 3.03677899, 3.05050363, 3.05501944, 3.04532485, 3.01641825, 2.96329807, 2.88096272, 2.76441061, 2.60864016, 2.40864977, 2.15943786, 1.85600285, 1.49334314, 1.06645716, 0.570343306, 0, -0.995039784, -2.0094201, -3.05239506, -4.13321877, -5.26114536, -6.44542892, -7.69532358, -9.02008344, -10.4289626, -11.9312152, -13.5360954, -15.2528572, -17.0907548, -19.0590422, -21.1669737, -23.4238032, -25.85211, -28.4630516, -31.2357424, -34.1822918, -37.3148093, -40.6454043, -44.1861864, -47.9492649, -51.9467494, -56.1907493, -60.6933741, -65.4667332, -70.5229361, -76.1149132, -81.9256521, -87.8950158, -93.9628675, -100.556741, -107.372665, -113.97662, -120.162169, -125.722874, -130.452297, -134.143999, -139.16174, -141.94665, -141.789428, -137.98077, -129.811373, -123.461782, -109.610386, -85.825421, -49.6751228, 1.27227247, 65.8094468, 154.455014, 263.54993, 397.005698, 574.190344, 807.53223, 1109.45972, 1484.6355, 1955.01854, 2551.03942, 3273.09196, 4186.01801, 5307.551, 6684.44498, 8412.62779, 10565.4557, 13125.4813, 16142.5191, 19999.3307, 24890.3441, 30595.89, 37788.2009, 46805.0459, 57984.1943, 71663.4154, 97608.2595, 124371.769, 149934.77, 172278.084, 189382.536, 191974.282, 190125.259, 184925.436, 180262.611, 176119.276, 171312.162, 165882.048, 159869.716, 153315.947, 146261.521, 134913.966, 125984.686, 118593.717, 111861.098, 104906.865, 96851.055, 86813.7064, 77523.3727, 73973.3218},
        {19.071373, 19.0737815, 19.0599563, 19.0304516, 18.9858221, 18.926622, 18.853406, 18.7667283, 18.6671435, 18.5552059, 18.4314702, 18.2964905, 18.1508215, 17.9950176, 17.8296331, 17.6552226, 17.4723405, 17.2815411, 17.0833791, 16.8784087, 16.6671845, 16.4502609, 16.2281923, 16.0015332, 15.770838, 15.5366611, 15.299557, 15.0600802, 14.8187851, 14.576226, 14.3329576, 14.0895341, 13.84651, 13.6044399, 13.3863876, 13.1811794, 12.9773709, 12.7747861, 12.573249, 12.3725835, 12.1726136, 11.9731634, 11.7740566, 11.5751173, 11.3761695, 11.1770371, 10.9775441, 10.7775145, 10.5767722, 10.3751411, 10.1724453, 9.95879408, 9.7439366, 9.52790745, 9.31074118, 9.09247236, 8.87313555, 8.6527653, 8.43139618, 8.20906275, 7.98579956, 7.76164117, 7.53662214, 7.31077704, 7.08414041, 6.85674683, 6.62863085, 6.39546082, 6.15300991, 5.91031767, 5.66770171, 5.42547962, 5.183969, 4.94348747, 4.70435262, 4.46688206, 4.23139339, 3.99820422, 3.76763213, 3.53999475, 3.31560967, 3.09479449, 2.87786683, 2.66514427, 2.62625655, 2.67330397, 2.7163518, 2.75022314, 2.76974109, 2.76972875, 2.74500921, 2.69040558, 2.60074096, 2.47083844, 2.29552113, 2.06961211, 1.78793449, 1.44531138, 1.03656585, 0.55652103, 0, -0.860667199, -1.76473572, -2.72045583, -3.73607778, -4.81985185, -5.98002828, -7.22485733, -8.56258928, -10.0014744, -11.5497629, -13.2157051, -15.0075512, -16.9335515, -19.0019563, -21.2210158, -23.5989802, -26.1753305, -28.9627783, -31.8989335, -34.9987384, -38.2771356, -41.7490677, -45.4294771, -49.3333064, -53.4754982, -57.8709949, -62.534739, -67.4816732, -72.7267399, -78.573543, -84.6425211, -90.8620185, -97.1603792, -103.877194, -110.735183, -117.347402, -123.518822, -129.054412, -133.759143, -137.437986, -142.785225, -145.837159, -145.833458, -142.01379, -133.617826, -126.954453, -112.426818, -87.507285, -49.6682188, 3.61801592, 71.5255697, 164.034266, 276.734456, 413.370215, 594.332809, 832.539258, 1140.90658, 1526.70852, 2009.78793, 2616.14395, 3340.09313, 4253.75899, 5380.60285, 6770.1647, 8516.17339, 10692.0743, 13293.4873, 16373.6806, 20271.2837, 25163.5094, 30772.7638, 37801.0373, 46582.976, 57453.2263, 70746.4344, 95899.1446, 121868.63, 146714.064, 168494.616, 185269.459, 188200.127, 186841.855, 182209.942, 177837.979, 173744.443, 168898.531, 163371.187, 157233.355, 150555.98, 143410.007, 132897.293, 124310.012, 116951.51, 110125.132, 103134.225, 95282.1339, 85872.2051, 76802.6002, 73256.1218},
        {17.4690024, 17.5229586, 17.5569516, 17.5716693, 17.5677994, 17.5460296, 17.5070477, 17.4515414, 17.3801984, 17.2937066, 17.1927536, 17.0780272, 16.9502151, 16.8100052, 16.658085, 16.4951424, 16.3218651, 16.1389409, 15.9470575, 15.7469026, 15.539164, 15.3245294, 15.1036866, 14.8773233, 14.6461273, 14.4107863, 14.171988, 13.9304203, 13.6867707, 13.4417272, 13.1959773, 12.9502089, 12.7051098, 12.4613676, 12.2465609, 12.0473665, 11.8501379, 11.6546901, 11.4608382, 11.2683974, 11.0771826, 10.8870089, 10.6976916, 10.5090455, 10.3208859, 10.1330277, 9.94528617, 9.75747627, 9.56941312, 9.38091179, 9.19178738, 8.99279446, 8.79300504, 8.59243061, 8.39108266, 8.18897269, 7.9861122, 7.78251269, 7.57818565, 7.37314258, 7.16739497, 6.96095432, 6.75383214, 6.54603991, 6.33758913, 6.12849131, 5.91875792, 5.70337502, 5.47744926, 5.25135683, 5.02543497, 4.80002099, 4.57545215, 4.35206573, 4.13019901, 3.91018926, 3.69237377, 3.47708981, 3.26467467, 3.0554656, 2.84979991, 2.64801485, 2.45044772, 2.25743578, 2.24415761, 2.3198523, 2.39165745, 2.45423642, 2.50225254, 2.53036915, 2.5332496, 2.50555722, 2.44195535, 2.33710735, 2.18567654, 1.98232628, 1.7217199, 1.39852073, 1.00739214, 0.542997446, 0, -0.737508558, -1.53968443, -2.41401063, -3.36797014, -4.40904597, -5.5447211, -6.78247855, -8.12980131, -9.59417238, -11.1830748, -12.9039914, -14.7644054, -16.7717997, -18.9336573, -21.2574612, -23.7506943, -26.4575035, -29.3906094, -32.4745651, -35.7247102, -39.1563842, -42.7849265, -46.6256767, -50.6939743, -55.0051588, -59.5745697, -64.4175465, -69.5494288, -74.985556, -81.0565569, -87.3468757, -93.7772651, -100.268478, -107.10037, -114.014144, -120.655237, -126.836673, -132.371473, -137.072658, -140.753251, -146.464246, -149.79522, -149.925681, -146.035131, -137.303077, -129.800599, -114.092733, -87.6360889, -47.8872787, 7.69708685, 78.245228, 173.889798, 290.216399, 431.019139, 616.851146, 860.741089, 1175.71764, 1570.76323, 2063.87237, 2677.45362, 3400.12717, 4311.09836, 5438.62134, 6835.0906, 8593.80509, 10789.2879, 13434.1315, 16586.8029, 20532.8484, 25429.3657, 30918.5085, 37751.5401, 46263.8078, 56790.6587, 69667.4402, 94001.583, 119163.33, 143295.007, 164538.942, 181037.459, 184434.696, 183703.296, 179770.524, 175746.108, 171693.421, 166766.479, 161074.125, 154725.2, 147828.548, 140493.01, 130718.252, 122385.931, 114995.842, 108047.775, 101041.522, 93476.8758, 84853.6284, 76365.7882, 72412.4233},
        {15.7523547, 15.8258447, 15.879586, 15.9142441, 15.9304843, 15.9289721, 15.910373, 15.8753523, 15.8245756, 15.7587082, 15.6784157, 15.5843634, 15.4772167, 15.3576412, 15.2263022, 15.0838652, 14.9309957, 14.768359, 14.5966206, 14.4164459, 14.2285005, 14.0334496, 13.8319589, 13.6246936, 13.4123192, 13.1955012, 12.9749051, 12.7511962, 12.5250399, 12.2971018, 12.0680473, 11.8385417, 11.6092506, 11.3808394, 11.1785686, 10.9902887, 10.8035547, 10.618234, 10.4341937, 10.2513012, 10.0694238, 9.88842873, 9.70818327, 9.52855471, 9.34941034, 9.17061743, 8.99204327, 8.81355514, 8.63502032, 8.45630609, 8.27727974, 8.09101837, 7.90432664, 7.71721902, 7.52971, 7.34181404, 7.15354563, 6.96491926, 6.77594939, 6.58665051, 6.39703709, 6.20712362, 6.01692457, 5.82645443, 5.63572767, 5.44475877, 5.25356221, 5.05789018, 4.85359721, 4.64948842, 4.4458546, 4.24298654, 4.04117503, 3.84071085, 3.64188478, 3.44498761, 3.25031013, 3.05814312, 2.86877738, 2.68250367, 2.4996128, 2.32039554, 2.14514269, 1.97414503, 1.98561153, 2.08713163, 2.18415629, 2.27120252, 2.34278731, 2.39342766, 2.41764057, 2.40994303, 2.36485205, 2.27688463, 2.14055776, 1.95038843, 1.70089366, 1.38659043, 1.00199574, 0.541626602, 0, -0.657345109, -1.39785737, -2.22811553, -3.15469834, -4.18418456, -5.32315292, -6.57818217, -7.95585107, -9.46273837, -11.1054228, -12.8904831, -14.8244981, -16.9140465, -19.165707, -21.5860584, -24.1816794, -27.0054415, -30.0705367, -33.278834, -36.6468321, -40.1910296, -43.927925, -47.8740171, -52.0458044, -56.4597855, -61.132459, -66.0803236, -71.3198778, -76.8676203, -83.05241, -89.4638531, -96.0247401, -102.657862, -109.710701, -116.893702, -123.804058, -130.237151, -135.988364, -140.85308, -144.626682, -150.380605, -153.637119, -153.55064, -149.275585, -139.966369, -131.221383, -114.140074, -86.2658641, -45.1421745, 11.6875724, 82.44285, 178.996025, 298.712489, 445.99401, 639.608289, 891.873018, 1215.10589, 1616.57447, 2114.79432, 2733.79023, 3463.25398, 4381.30203, 5514.29758, 6914.36907, 8677.30283, 10878.8455, 13541.4281, 16726.9746, 20692.8999, 25586.038, 31060.7052, 37832.0256, 46217.3203, 56533.9105, 69099.1173, 92571.9325, 116842.589, 140142.99, 160705.04, 176760.641, 180419.339, 180116.968, 176709.301, 172989.234, 169045.761, 164160.739, 158463.519, 152083.45, 145149.882, 137792.165, 128552.129, 120482.883, 113171.3, 106204.249, 99168.6007, 91651.2259, 83238.995, 75976.4223, 71482.4073},
        {14.0891635, 14.1683726, 14.2294736, 14.2730553, 14.2997064, 14.3100155, 14.3045713, 14.2839625, 14.2487778, 14.1996058, 14.1370354, 14.0616551, 13.9740536, 13.8748196, 13.7645419, 13.643809, 13.5132097, 13.3733327, 13.2247666, 13.0681002, 12.9039221, 12.732821, 12.5553856, 12.3722046, 12.1838666, 11.9909604, 11.7940746, 11.5937979, 11.3907191, 11.1854267, 10.9785096, 10.7705563, 10.5621555, 10.353896, 10.1664589, 9.98996395, 9.81428962, 9.63937249, 9.46514922, 9.29155645, 9.11853083, 8.94600901, 8.77392763, 8.60222333, 8.43083277, 8.2596926, 8.08873945, 7.91790998, 7.74714083, 7.57636864, 7.40553008, 7.2302631, 7.05489622, 6.87945927, 6.70398208, 6.52849449, 6.35302632, 6.17760741, 6.00226759, 5.8270367, 5.65194458, 5.47702104, 5.30229593, 5.12779908, 4.95356033, 4.7796095, 4.60597642, 4.42970723, 4.24791968, 4.06679728, 3.88656329, 3.70744096, 3.52965356, 3.35342433, 3.17897653, 3.00653342, 2.83631825, 2.66855427, 2.50346475, 2.34127293, 2.18220208, 2.02647544, 1.87431628, 1.72594784, 1.76207049, 1.88887284, 2.01037429, 2.12094129, 2.21494024, 2.28673759, 2.33069975, 2.34119316, 2.31258423, 2.2392394, 2.11552509, 1.93580772, 1.69445373, 1.38582953, 1.00430156, 0.544236243, 0, -0.605079326, -1.30866343, -2.11651485, -3.03439612, -4.06806979, -5.22329839, -6.50584447, -7.92147055, -9.47593918, -11.1750129, -13.0244542, -15.0300258, -17.19749, -19.5326094, -22.0411467, -24.7288642, -27.6555023, -30.8345344, -34.1496309, -37.6181211, -41.257334, -45.0845988, -49.1172447, -53.3726009, -57.8679965, -62.6207608, -67.6482228, -72.9677118, -78.5965568, -84.8491554, -91.3368434, -97.9878295, -104.730323, -112.023563, -119.530243, -126.771449, -133.516078, -139.53303, -144.591204, -148.4595, -154.125505, -157.159829, -156.701628, -151.890059, -141.864277, -131.6316, -112.995982, -83.6295386, -41.2043844, 16.6073649, 86.8668195, 183.605807, 306.433937, 460.498115, 662.250558, 923.067632, 1254.3257, 1660.39683, 2161.25357, 2784.5095, 3523.79727, 4452.53018, 5592.14395, 6993.08466, 8754.95482, 10954.7933, 13622.6768, 16823.4688, 20794.2278, 25674.6088, 31157.9468, 37898.8821, 46191.5655, 56330.1479, 68608.7801, 91205.3921, 114559.411, 136994.042, 156832.492, 172397.967, 176259.021, 176329.946, 173400.228, 169983.178, 166185.934, 161397.372, 155760.54, 149418.49, 142514.271, 135190.933, 126391.158, 118556.526, 111332.109, 104362.979, 97294.2088, 89770.8704, 81438.0359, 75600.259, 70543.6395},
        {12.5482653, 12.6245248, 12.6849416, 12.7300065, 12.7602106, 12.7760446, 12.7779994, 12.7665658, 12.7422347, 12.705497, 12.6568436, 12.5967653, 12.5257529, 12.4442974, 12.3528895, 12.2520202, 12.1421804, 12.0238608, 11.8975523, 11.7637459, 11.6229323, 11.4756025, 11.3222472, 11.1633574, 10.999424, 10.8309377, 10.6583895, 10.4822701, 10.3030706, 10.1212817, 9.93739426, 9.7518992, 9.56528736, 9.37804962, 9.20564511, 9.04076573, 8.87592343, 8.71112335, 8.54637058, 8.38167024, 8.21702744, 8.0524473, 7.88793493, 7.72349543, 7.55913394, 7.39485555, 7.23066538, 7.06656855, 6.90257016, 6.73867533, 6.57488917, 6.40900395, 6.24328559, 6.07778718, 5.9125618, 5.74766255, 5.58314249, 5.41905472, 5.25545232, 5.09238837, 4.92991595, 4.76808815, 4.60695806, 4.44657874, 4.2870033, 4.1282848, 3.97047634, 3.81204987, 3.65151533, 3.49218673, 3.33421965, 3.17776969, 3.02299242, 2.87004342, 2.71907827, 2.57025256, 2.42372186, 2.27964176, 2.13816783, 1.99945566, 1.86366083, 1.73093892, 1.60144551, 1.47533618, 1.5363504, 1.68899037, 1.83555175, 1.97023246, 2.08723045, 2.18074363, 2.24496994, 2.27410732, 2.26235368, 2.20390697, 2.09296512, 1.92372604, 1.69038768, 1.38714797, 1.00820483, 0.547756195, 0, -0.577556941, -1.26550623, -2.06895474, -2.99300938, -4.04277704, -5.22336459, -6.53987893, -7.99742696, -9.60111555, -11.3560516, -13.267342, -15.3400936, -17.5794134, -19.9904082, -22.5781849, -25.3478504, -28.3639009, -31.6399276, -35.0467042, -38.602064, -42.32384, -46.2298655, -50.3379736, -54.6659976, -59.2317707, -64.0531259, -69.1478966, -74.5339158, -80.2290168, -86.5311331, -93.0752941, -99.7953029, -106.624962, -114.142635, -121.959844, -129.516231, -136.552231, -142.808278, -148.02481, -151.942259, -157.440615, -160.165244, -159.242321, -153.798023, -142.958523, -131.141945, -110.859529, -79.9144635, -36.109937, 22.7508624, 92.6671114, 189.608144, 315.173056, 475.164918, 683.969921, 952.120608, 1290.14952, 1699.38179, 2201.64997, 2828.8311, 3580.08985, 4521.8868, 5668.94725, 7069.07609, 8826.12961, 11018.2356, 13681.793, 16884.111, 20845.5045, 25699.7396, 31186.4021, 37894.0818, 46092.992, 56053.346, 68045.357, 89767.5111, 112204.681, 133770.011, 152876.646, 167937.733, 171966.07, 172375.378, 169893.235, 166783.622, 163165.958, 158519.496, 152995.663, 146745.887, 139921.596, 132674.218, 124274.21, 116642.713, 109482.898, 102497.936, 95390.9977, 87865.2537, 79623.875, 75151.5226, 69664.161},
        {11.3204304, 11.3945867, 11.4540625, 11.4993041, 11.5307577, 11.5488697, 11.5540861, 11.5468533, 11.5276175, 11.4968249, 11.4549218, 11.4023544, 11.3395689, 11.2670116, 11.1851288, 11.0943666, 10.9951713, 10.8879891, 10.7732663, 10.6514491, 10.5229838, 10.3883165, 10.2478936, 10.1021613, 9.95156571, 9.7965532, 9.63756996, 9.47506225, 9.3094763, 9.14125833, 8.9708546, 8.79871135, 8.6252748, 8.4509912, 8.28933713, 8.13396959, 7.97838895, 7.82261857, 7.66668185, 7.51060217, 7.35440289, 7.1981074, 7.04173908, 6.8853213, 6.72887744, 6.57243089, 6.41600501, 6.2596232, 6.10330882, 5.94708525, 5.79097588, 5.63238321, 5.47400831, 5.31593137, 5.15823258, 5.00099213, 4.84429023, 4.68820706, 4.53282282, 4.37821769, 4.22447189, 4.07166559, 3.919879, 3.7691923, 3.61968569, 3.47143936, 3.32453351, 3.17804841, 3.03108831, 2.88579571, 2.74231561, 2.60079302, 2.46137296, 2.32420045, 2.18942051, 2.05717813, 1.92761835, 1.80088618, 1.67712663, 1.55648471, 1.43910545, 1.32513385, 1.21471493, 1.10799371, 1.1959362, 1.37926325, 1.55640219, 1.72130553, 1.86792578, 1.99021544, 2.08212703, 2.13761304, 2.15062599, 2.11511838, 2.02504271, 1.8743515, 1.65699725, 1.36693247, 0.998109663, 0.544481338, 0, -0.567631919, -1.25442247, -2.06502911, -3.00410929, -4.07632047, -5.28632009, -6.63876561, -8.13831448, -9.78962416, -11.5973521, -13.5661558, -15.7006926, -18.00562, -20.4855955, -23.1452766, -25.9893206, -29.0914045, -32.4660457, -35.9617862, -39.5980733, -43.3943543, -47.3700766, -51.5446876, -55.9376345, -60.5683647, -65.4563255, -70.6209643, -76.0817285, -81.8580653, -88.269764, -94.9258046, -101.755532, -108.688291, -116.327724, -124.266026, -131.915582, -139.003447, -145.256677, -150.402326, -154.167452, -159.503139, -161.931186, -160.547861, -154.449432, -142.732164, -129.637712, -107.830611, -75.1207797, -29.3181404, 31.7673865, 104.350885, 203.900202, 331.251083, 491.983863, 701.365749, 970.31474, 1309.74884, 1720.69576, 2226.21158, 2860.14174, 3626.89815, 4585.24995, 5743.08388, 7144.92697, 8897.39114, 11078.5903, 13731.8089, 16926.131, 20858.349, 25653.0162, 31053.895, 37624.9129, 45622.4355, 55302.8283, 66922.4568, 87771.6943, 109314.397, 130048.428, 148471.652, 163081.931, 167319.575, 168111.706, 166123.712, 163378.654, 160004.509, 155563.554, 150212.049, 144106.254, 137402.43, 130256.836, 122291.264, 114793.784, 107590.55, 100507.716, 93371.4364, 86007.8665, 78243.1603, 74291.6036, 69146.4993},
        {10.2687401, 10.3371354, 10.3919018, 10.4334501, 10.4621912, 10.4785361, 10.4828955, 10.4756805, 10.4573018, 10.4281704, 10.3886972, 10.339293, 10.2803688, 10.2123354, 10.1356037, 10.0505847, 9.95768922, 9.85732811, 9.74991231, 9.6358527, 9.51556016, 9.3894456, 9.25791991, 9.12139396, 8.98027867, 8.83498491, 8.68592358, 8.53350557, 8.37814177, 8.22024307, 8.06022037, 7.89848455, 7.73544651, 7.57151714, 7.41896118, 7.27201327, 7.12476557, 6.97724429, 6.82947564, 6.68148582, 6.53330106, 6.38494757, 6.23645154, 6.0878392, 5.93913676, 5.79037042, 5.6415664, 5.49275091, 5.34395016, 5.19519036, 5.04649772, 4.89411477, 4.74193342, 4.59006191, 4.43860848, 4.28768136, 4.13738877, 3.98783896, 3.83914016, 3.6914006, 3.54472852, 3.39923215, 3.25501972, 3.11219946, 2.97087962, 2.83116842, 2.69317411, 2.55632417, 2.42006246, 2.28590117, 2.15399266, 2.0244893, 1.89754344, 1.77330746, 1.6519337, 1.53357454, 1.41838234, 1.30650945, 1.19810826, 1.0933311, 0.99233036, 0.895258388, 0.80226755, 0.71351021, 0.82842744, 1.04333548, 1.25215526, 1.44857186, 1.62627033, 1.77893573, 1.90025312, 1.98390756, 2.02358412, 2.01296786, 1.94574382, 1.81559709, 1.61621271, 1.34127576, 0.984471277, 0.539484338, 0, -0.580971382, -1.28556377, -2.1182675, -3.0835729, -4.18597032, -5.42995008, -6.82000252, -8.36061798, -10.0562868, -11.9114993, -13.9307458, -16.1185167, -18.4793022, -21.0175928, -23.7378788, -26.6446504, -29.816145, -33.2673882, -36.8360617, -40.5424589, -44.406873, -48.4495972, -52.6909247, -57.1511489, -61.850563, -66.8094602, -72.0481337, -77.5868768, -83.4459828, -89.9705636, -96.7369936, -103.66812, -110.686791, -118.398672, -126.385201, -134.045732, -141.098268, -147.260811, -152.251364, -155.787927, -160.933611, -163.043235, -161.180324, -154.408402, -141.790996, -127.493295, -104.201743, -69.7044475, -21.7895174, 41.7549389, 117.67249, 220.326856, 349.118588, 509.187695, 717.39224, 985.32912, 1324.59523, 1736.61397, 2245.4679, 2886.3379, 3668.84047, 4643.69679, 5811.90775, 7214.4468, 8960.01878, 11126.2387, 13762.2157, 16938.1743, 20828.8609, 25549.8982, 30842.853, 37254.9514, 45031.2973, 54416.9946, 65657.1473, 85674.0538, 106366.325, 126309.765, 144080.18, 158253.375, 162657.924, 163781.684, 162234.989, 159838.608, 156723.744, 152517.248, 147375.54, 141455.039, 134912.162, 127903.33, 120416.051, 113040.193, 105752.09, 98528.0769, 91344.489, 84177.6607, 77003.9268, 73299.8936, 68706.2959},
        {9.37270529, 9.43215899, 9.47895035, 9.51346225, 9.5360776, 9.54717928, 9.54715019, 9.53637322, 9.51523128, 9.48410725, 9.44338402, 9.39344451, 9.33467159, 9.26744816, 9.19215713, 9.10918137, 9.01890379, 8.92170729, 8.81797475, 8.70808907, 8.59243315, 8.47138988, 8.34534215, 8.21467287, 8.07976491, 7.94100119, 7.7987646, 7.65343802, 7.50540435, 7.3550465, 7.20274734, 7.04888978, 6.89385672, 6.73803104, 6.59302416, 6.45336853, 6.31346166, 6.17332206, 6.03296823, 5.89241868, 5.75169193, 5.61080648, 5.46978083, 5.3286335, 5.18738299, 5.04604781, 4.90464648, 4.76319748, 4.62171935, 4.48023057, 4.33874967, 4.19197306, 4.0453567, 3.89903449, 3.75314029, 3.607808, 3.46317148, 3.31936463, 3.17652131, 3.03477542, 2.89426082, 2.75511141, 2.61746106, 2.48144365, 2.34719306, 2.21484317, 2.08452787, 1.95578224, 1.82815573, 1.7030172, 1.58053935, 1.46089488, 1.34425648, 1.23079685, 1.12068869, 1.0141047, 0.911217573, 0.812200006, 0.7172247, 0.626464351, 0.540091658, 0.458279318, 0.381200029, 0.309026489, 0.450321861, 0.696680397, 0.937191657, 1.16526567, 1.37431245, 1.55774204, 1.70896445, 1.82138972, 1.88842788, 1.90348894, 1.85998293, 1.75131989, 1.57090983, 1.31216279, 0.968488782, 0.533297845, 0, -0.618689869, -1.36051357, -2.23014193, -3.23224576, -4.37149588, -5.65256312, -7.08011829, -8.65883222, -10.3933757, -12.2884196, -14.3486347, -16.5786918, -18.9832618, -21.5670155, -24.3346236, -27.2907571, -30.5100845, -34.0072766, -37.6286754, -41.3939514, -45.3227749, -49.4348163, -53.7497459, -58.2872342, -63.0669515, -68.1085681, -73.4317545, -79.0561811, -85.0015181, -91.6335254, -98.4988845, -105.513439, -112.593032, -120.331644, -128.306049, -135.912187, -142.862462, -148.869275, -153.645031, -156.902131, -161.839742, -163.622315, -161.280059, -153.843184, -140.341899, -124.909308, -100.197003, -63.9594723, -13.9512008, 52.0733247, 131.496196, 237.370531, 367.443855, 526.343936, 732.745688, 999.017143, 1337.52633, 1750.49035, 2262.7352, 2910.16064, 3707.03469, 4696.45722, 5873.13606, 7274.56259, 9010.7226, 11158.3257, 13770.8686, 16919.0425, 20758.6538, 25397.3629, 30573.666, 36821.1823, 44375.36, 53471.6472, 64345.4921, 83595.7524, 103503.614, 122713.673, 139870.525, 153618.766, 158120.785, 159491.521, 158296.515, 156204.595, 153344.583, 149385.6, 144479.453, 138777.949, 132432.897, 125596.105, 118663.605, 111420.17, 104018.305, 96610.5144, 89349.3041, 82387.1785, 75876.6428, 72236.2051, 68267.716},
        {8.60440315, 8.65177243, 8.68743736, 8.71175632, 8.72508769, 8.72778986, 8.72022121, 8.70274013, 8.675705, 8.63947421, 8.59440614, 8.54085918, 8.47919171, 8.40976211, 8.33292877, 8.24905008, 8.15848442, 8.06159016, 7.95872571, 7.85024944, 7.73651974, 7.61789499, 7.49473357, 7.36739388, 7.23623429, 7.10161319, 6.96388897, 6.82342, 6.68056468, 6.53568139, 6.38912851, 6.24126443, 6.09244753, 5.9430362, 5.80408938, 5.6703901, 5.53659278, 5.40270856, 5.26874857, 5.13472393, 5.00064578, 4.86652525, 4.73237347, 4.59820157, 4.46402068, 4.32984193, 4.19567646, 4.06153539, 3.92742985, 3.79337098, 3.65936991, 3.51928142, 3.37940644, 3.23988957, 3.1008754, 2.9625085, 2.82493347, 2.6882949, 2.55273737, 2.41840548, 2.28544381, 2.15399695, 2.02420949, 1.89622601, 1.77019112, 1.64624938, 1.5245454, 1.4043922, 1.28512278, 1.16859676, 1.05501262, 0.94456886, 0.83746398, 0.733896471, 0.634064828, 0.538167546, 0.44640312, 0.358970044, 0.276066814, 0.197891923, 0.124643868, 0.0565211417, -0.00627776006, -0.0635543428, 0.101522335, 0.376277973, 0.645434515, 0.902160317, 1.13962373, 1.35099312, 1.52943683, 1.66812322, 1.76022065, 1.79889747, 1.77732204, 1.6886627, 1.52608783, 1.28276576, 0.951864869, 0.526553495, 0, -0.708172813, -1.52685721, -2.46186312, -3.51900052, -4.70407933, -6.02290952, -7.48130103, -9.08506381, -10.8400078, -12.751943, -14.8266792, -17.0700265, -19.4877949, -22.0857942, -24.8698344, -27.8457254, -31.0569163, -34.5145938, -38.1394323, -41.9453071, -45.9460936, -50.1556674, -54.5879038, -59.2566783, -64.1758664, -69.3593435, -74.820985, -80.5746664, -86.6342631, -93.3505763, -100.277016, -107.326381, -114.411468, -122.115069, -130.014986, -137.510353, -144.31297, -150.134636, -154.687149, -157.682311, -162.489982, -164.050576, -161.360185, -153.414899, -139.21081, -122.613887, -96.5328753, -58.7463954, -7.03306933, 60.828481, 142.294118, 250.175255, 381.550322, 540.893268, 747.777227, 1014.99611, 1355.34382, 1772.15645, 2289.83522, 2943.099, 3747.52948, 4742.37353, 5919.2752, 7313.59452, 9034.76588, 11158.2223, 13738.1025, 16844.3186, 20624.8624, 25183.5596, 30270.3049, 36397.3055, 43790.9016, 52677.4331, 63283.24, 82002.3865, 101351.558, 120015.162, 136677.61, 150023.309, 154373.275, 155684.241, 154510.641, 152521.437, 149829.897, 146084.906, 141422.952, 135980.525, 129894.113, 123300.205, 117351.669, 110379.327, 102800.765, 95033.5719, 87495.3348, 80603.6412, 74776.0786, 71280.2578, 67618.6621},
        {7.94503275, 7.9777515, 7.99972972, 8.01130327, 8.01280805, 8.00457991, 7.98695473, 7.96026839, 7.92485676, 7.8810557, 7.82920111, 7.76962884, 7.70267478, 7.62867479, 7.54796475, 7.46088053, 7.36775801, 7.26893306, 7.16474156, 7.05551937, 6.94160237, 6.82332643, 6.70102743, 6.57504124, 6.44570373, 6.31335079, 6.17831827, 6.04094205, 5.90155802, 5.76050203, 5.61810997, 5.47471771, 5.33066112, 5.18627607, 5.05201342, 4.92293881, 4.79399627, 4.66519344, 4.53653793, 4.40803737, 4.27969939, 4.1515316, 4.02354162, 3.89573709, 3.76812562, 3.64071484, 3.51351236, 3.38652581, 3.25976282, 3.13323101, 3.006938, 2.87482508, 2.74309772, 2.61189504, 2.48135617, 2.35162023, 2.22282636, 2.09511369, 1.96862133, 1.84348844, 1.71985412, 1.59785751, 1.47763774, 1.35933394, 1.24308523, 1.12903075, 1.01730962, 0.906729898, 0.796131579, 0.688399051, 0.583757731, 0.482433037, 0.384650387, 0.290635196, 0.200612884, 0.114808867, 0.0334485617, -0.0432426134, -0.115039242, -0.181715905, -0.243047187, -0.29880767, -0.348771937, -0.39271457, -0.206788317, 0.0927177387, 0.386826749, 0.668507172, 0.930727467, 1.16645609, 1.36866151, 1.53031217, 1.64437654, 1.70382308, 1.70162024, 1.63073648, 1.48414027, 1.25480005, 0.935684298, 0.519761461, 0, -0.844569072, -1.77573186, -2.80133745, -3.9292349, -5.16727328, -6.52330169, -8.00516917, -9.62072482, -11.3778177, -13.2842969, -15.3480114, -17.5768105, -19.978543, -22.5610582, -25.332205, -28.2998325, -31.4486467, -34.7831904, -38.3636604, -42.1929465, -46.2739386, -50.6095264, -55.2025999, -60.0560487, -65.1727627, -70.5556317, -76.2075456, -82.1313941, -88.3300671, -95.100833, -102.044264, -109.074937, -116.107427, -123.722328, -131.501208, -138.84803, -145.477564, -151.104582, -155.443855, -158.210155, -162.95554, -164.389624, -161.476402, -153.179872, -138.464031, -120.687856, -93.3216184, -54.2305711, -1.27996583, 67.6649454, 149.473744, 257.981548, 390.865829, 552.886182, 763.26153, 1034.71886, 1379.98514, 1803.49523, 2328.15281, 2985.90754, 3790.32302, 4780.76363, 5949.22567, 7330.42405, 9031.45806, 11126.2014, 13665.7513, 16718.0767, 20434.6592, 24919.7409, 29952.3465, 36011.9689, 43315.4253, 52079.5332, 62521.1098, 80936.9038, 99943.8184, 118238.688, 134518.347, 147479.63, 151435.74, 152389.563, 150916.636, 148829.403, 146213.787, 142640.333, 138220.732, 133066.678, 127289.863, 121001.979, 116426.539, 109863.384, 102068.328, 93797.1842, 85805.7645, 78849.8824, 73685.3506, 70464.6352, 66714.2425},
        {7.38736532, 7.40412748, 7.41105483, 7.40846426, 7.39667266, 7.37599691, 7.34675391, 7.30926053, 7.26383368, 7.21079023, 7.15044708, 7.08312112, 7.00912922, 6.92878829, 6.8424152, 6.75032686, 6.65284013, 6.55027192, 6.44293911, 6.33115859, 6.21524724, 6.09552196, 5.97229963, 5.84589715, 5.71663139, 5.58481925, 5.45077762, 5.31482338, 5.17727342, 5.03844464, 4.89865391, 4.75821813, 4.61745418, 4.47667895, 4.34590509, 4.22039766, 4.09530521, 3.97062998, 3.84637421, 3.72254015, 3.59913003, 3.4761461, 3.3535906, 3.23146578, 3.10977387, 2.98851712, 2.86769778, 2.74731807, 2.62738025, 2.50788656, 2.38883924, 2.26431817, 2.14037635, 2.01714439, 1.89475292, 1.77333258, 1.65301399, 1.53392777, 1.41620456, 1.29997499, 1.18536967, 1.07251924, 0.961554331, 0.852605561, 0.745803561, 0.641278958, 0.539162381, 0.43769406, 0.335159615, 0.235588473, 0.139233809, 0.0463487964, -0.0428133901, -0.127999577, -0.208956589, -0.285431252, -0.357170393, -0.423920836, -0.485429409, -0.541442936, -0.591708243, -0.635972157, -0.673981502, -0.705483106, -0.499861265, -0.176889436, 0.140890146, 0.446254948, 0.731982438, 0.990850083, 1.21563535, 1.39911571, 1.53406862, 1.61327155, 1.62950198, 1.57553736, 1.44415517, 1.22813287, 0.920247932, 0.513277819, 0, -0.982132469, -2.02666515, -3.14347802, -4.34245108, -5.63346431, -7.0263977, -8.53113123, -10.1575449, -11.9155187, -13.8149326, -15.8656666, -18.0776006, -20.4606148, -23.024589, -25.7794032, -28.7349375, -31.8146877, -35.0161931, -38.5447265, -42.391657, -46.5483536, -51.0061854, -55.7565216, -60.7907313, -66.1001835, -71.6762474, -77.5102921, -83.5936866, -89.9178001, -96.7203018, -103.655951, -110.642226, -117.596606, -125.107705, -132.757434, -139.948736, -146.397751, -151.820618, -155.933475, -158.452462, -163.057413, -164.294429, -161.104143, -152.427189, -137.2042, -118.390855, -89.9089798, -49.695446, 4.31287445, 74.1791099, 156.166354, 265.226701, 400.025031, 565.683947, 780.510568, 1056.89243, 1407.21708, 1836.30618, 2365.83266, 3025.72313, 3827.7379, 4811.54876, 5969.99944, 7337.33055, 9017.92739, 11083.9777, 13582.9229, 16580.5349, 20233.3224, 24646.4685, 29633.7931, 35636.667, 42861.1294, 51513.2197, 61798.977, 79906.0848, 98563.5272, 116481.932, 132371.928, 144944.144, 148510.571, 149112.897, 147346.881, 145162.982, 142620.723, 139215.164, 135034.048, 130165.119, 124696.121, 118714.798, 115363.29, 109183.906, 101228.23, 92547.8447, 84194.3353, 77219.2852, 72674.2783, 69635.2652, 65721.6808},
        {6.91693331, 6.9172654, 6.90859541, 6.89122439, 6.86545344, 6.8315836, 6.78991596, 6.74075158, 6.68439153, 6.62113688, 6.55128871, 6.47514808, 6.39301606, 6.30519373, 6.21198214, 6.11368238, 6.01059551, 5.9030226, 5.79126472, 5.67562294, 5.55639833, 5.43389196, 5.3084049, 5.18023823, 5.049693, 4.91707029, 4.78267117, 4.64679671, 4.50974798, 4.37182605, 4.23333198, 4.09456686, 3.95583174, 3.8174277, 3.68908557, 3.5661932, 3.44402509, 3.3225763, 3.20184189, 3.08181691, 2.96249641, 2.84387547, 2.72594912, 2.60871243, 2.49216045, 2.37628823, 2.26109084, 2.14656333, 2.03270076, 1.91949818, 1.80695064, 1.68914055, 1.57210378, 1.45596358, 1.34084318, 1.22686581, 1.11415469, 1.00283306, 0.893024154, 0.784851198, 0.678437424, 0.573906063, 0.471380347, 0.370983505, 0.27283877, 0.177069372, 0.0837985412, -0.00925250414, -0.104304876, -0.19628137, -0.284903042, -0.369890949, -0.450966146, -0.52784969, -0.600262637, -0.667926044, -0.730560967, -0.787888461, -0.839629583, -0.88550539, -0.925236937, -0.958545281, -0.985151478, -1.00477658, -0.780156932, -0.434490595, -0.0938240188, 0.23443113, 0.542863184, 0.824060479, 1.07061135, 1.27510412, 1.43012714, 1.52826873, 1.56211723, 1.52426097, 1.40728829, 1.20378752, 0.906346991, 0.50755504, 0, -1.10069812, -2.24397899, -3.44136853, -4.70439266, -6.04457727, -7.4734483, -9.00253164, -10.6433532, -12.4074389, -14.3063147, -16.3515064, -18.55454, -20.9269415, -23.4802366, -26.2259513, -29.1756116, -32.1984991, -35.2858191, -38.7708405, -42.6346082, -46.8581672, -51.4225625, -56.3088392, -61.4980423, -66.9712167, -72.7094075, -78.6936598, -84.9050184, -91.3245284, -98.1345095, -105.040931, -111.964457, -118.825748, -126.237427, -133.774177, -140.826159, -147.106447, -152.328116, -156.204241, -158.447897, -162.773112, -163.674326, -160.081819, -150.925871, -135.136762, -115.496205, -86.1201874, -44.9761312, 9.96854262, 80.7464128, 163.173865, 273.09724, 410.405536, 580.576826, 800.500212, 1081.94148, 1436.66641, 1868.69636, 2399.30164, 3058.20101, 3856.89808, 4834.27276, 5983.53449, 7338.21252, 9000.15583, 11039.8535, 13501.664, 16449.3019, 20042.294, 24385.9701, 29332.1339, 35278.6458, 42418.5948, 50945.07, 61051.1605, 78752.1016, 96963.2996, 114422.307, 129866.676, 142033.96, 145311.82, 145687.627, 143763.171, 141560.28, 139118.646, 135886.284, 131932.474, 127326.492, 122137.618, 116435.131, 113931.765, 108049.144, 100039.019, 91153.1371, 82643.2484, 75761.1016, 71758.4456, 68754.8352, 64675.5149},
        {6.51856216, 6.50378743, 6.48067504, 6.44951605, 6.41060153, 6.36422252, 6.3106701, 6.25023532, 6.18320924, 6.10988292, 6.03054742, 5.9454938, 5.85501312, 5.75939643, 5.65893481, 5.5539193, 5.44464096, 5.33139087, 5.21446006, 5.09413962, 4.97072059, 4.84449403, 4.71575101, 4.58478258, 4.4518798, 4.31733374, 4.18143545, 4.04447598, 3.90674642, 3.7685378, 3.63014119, 3.49184765, 3.35394824, 3.21673402, 3.08994174, 2.96894092, 2.84898759, 2.73006627, 2.61216149, 2.49525779, 2.37933968, 2.2643917, 2.15039837, 2.03734423, 1.92521381, 1.81399162, 1.70366221, 1.5942101, 1.48561982, 1.3778759, 1.27096287, 1.15851555, 1.04700583, 0.936555885, 0.827287886, 0.719324016, 0.612786452, 0.507797372, 0.404478955, 0.302953377, 0.203342817, 0.105769452, 0.0103554617, -0.0827769772, -0.173505687, -0.261708488, -0.347263205, -0.432740455, -0.520646204, -0.605304584, -0.686418853, -0.763692269, -0.836828093, -0.905529583, -0.969499998, -1.0284426, -1.08206064, -1.13005738, -1.17213609, -1.20800001, -1.23735242, -1.25989656, -1.2753357, -1.2833731, -1.04027954, -0.672594807, -0.309784522, 0.0405482003, 0.370800248, 0.673368512, 0.940649881, 1.16504124, 1.33893949, 1.45474151, 1.50484419, 1.48164443, 1.3775391, 1.18492511, 0.896199332, 0.503758667, 0, -1.16651195, -2.36784708, -3.61616487, -4.92362483, -6.30238644, -7.76460922, -9.32245266, -10.9880762, -12.7736395, -14.6913019, -16.7532229, -18.9715621, -21.3584789, -23.9261329, -26.6866835, -29.6522902, -32.6610107, -35.7001329, -39.175618, -43.0623178, -47.3350844, -51.9687697, -56.9382257, -62.2183044, -67.7838578, -73.6097378, -79.6707963, -85.9418854, -92.397857, -99.1865642, -106.046424, -112.900388, -119.671408, -127.029073, -134.523016, -141.503766, -147.670284, -152.721531, -156.356467, -158.274053, -162.028719, -162.290552, -158.004182, -148.114239, -131.565352, -111.478009, -81.5770176, -39.7630427, 16.0632501, 88.0011954, 171.904927, 283.752004, 424.853391, 600.921687, 826.656158, 1112.72124, 1469.78137, 1898.27413, 2421.41509, 3073.57638, 3871.08664, 4847.53618, 5993.27381, 7339.82924, 8988.51082, 11008.5974, 13444.7591, 16359.8948, 19907.1348, 24188.5693, 29095.3888, 34973.0295, 41996.5854, 50341.1505, 60181.8189, 77185.052, 94661.7346, 111414.119, 126244.457, 137955.001, 141229.097, 141733.84, 140036.731, 138047.347, 135801.946, 132777.586, 129038.486, 124648.864, 119672.94, 114174.93, 111707.288, 105907.273, 98033.7107, 89345.4261, 81101.2452, 74559.9934, 70980.4965, 67723.663, 63614.395},
        {6.1775284, 6.15015164, 6.11488937, 6.07202916, 6.0218586, 5.96466527, 5.90073676, 5.83036065, 5.75382453, 5.67141597, 5.58342258, 5.49013192, 5.39183158, 5.28880915, 5.18135222, 5.06974836, 4.95428516, 4.8352502, 4.71293107, 4.58761536, 4.45959064, 4.32914451, 4.19656454, 4.06213833, 3.92615344, 3.78889748, 3.65065802, 3.51172265, 3.37237894, 3.2329145, 3.09361689, 2.95477371, 2.81667254, 2.67960096, 2.5536283, 2.43394552, 2.31562603, 2.19863997, 2.08295749, 1.96854874, 1.85538387, 1.74343303, 1.63266636, 1.52305402, 1.41456614, 1.30717289, 1.2008444, 1.09555082, 0.991262315, 0.887949016, 0.785581078, 0.676662013, 0.568790466, 0.462098447, 0.356717967, 0.252781034, 0.15041966, 0.0497658541, -0.0490483731, -0.145891012, -0.240630051, -0.333133482, -0.423269294, -0.510905477, -0.59591002, -0.678150914, -0.757496149, -0.836445209, -0.917434399, -0.994904447, -1.06855275, -1.13807672, -1.20317375, -1.26354123, -1.31887658, -1.3688772, -1.41324047, -1.45166381, -1.48384461, -1.50948028, -1.52826822, -1.53990582, -1.54409049, -1.54051963, -1.27912207, -0.889733035, -0.505189791, -0.133295206, 0.218147848, 0.541336501, 0.82846788, 1.07173912, 1.26334734, 1.39548967, 1.46036324, 1.45016519, 1.35709264, 1.17334271, 0.891112547, 0.502599266, 0, -1.15449048, -2.35385086, -3.60939442, -4.93243443, -6.33428415, -7.82625686, -9.41966582, -11.1258243, -12.9560456, -14.9216429, -17.0339296, -19.3042188, -21.7438239, -24.3640582, -27.1762348, -30.1916672, -33.2520058, -36.3443266, -39.8637034, -43.7850932, -48.0834529, -52.7337397, -57.7109104, -62.9899222, -68.5457319, -74.3532967, -80.3875734, -86.6235191, -93.0360908, -99.772161, -106.571736, -113.358197, -120.054927, -127.431262, -134.987608, -141.998782, -148.134381, -153.064, -156.457235, -157.983684, -160.783131, -159.998838, -154.621452, -143.641618, -126.049983, -106.001288, -76.031059, -33.8389192, 22.8755099, 96.4126061, 183.381064, 298.729447, 445.277707, 628.756802, 860.840991, 1150.53611, 1506.848, 1922.96604, 2427.30718, 3065.54268, 3866.03735, 4850.54143, 6001.70014, 7347.11468, 8990.56167, 11000.852, 13428.1414, 16336.4042, 19858.0168, 24086.6633, 28952.0977, 34737.1547, 41592.2635, 49667.8537, 59114.3546, 74999.3037, 91326.7453, 107017.832, 120993.715, 132175.548, 135859.02, 137007.721, 136096.574, 134657.926, 132748.136, 129982.967, 126440.808, 122200.046, 117339.07, 111936.266, 108389.296, 102372.41, 94889.2232, 86943.353, 79538.416, 73678.0291, 70365.8092, 66481.8612, 62574.3475},
        {5.88345485, 5.84123549, 5.79182066, 5.7354833, 5.67249633, 5.60313268, 5.52766528, 5.44636707, 5.35951098, 5.26736993, 5.17021685, 5.06832469, 4.96196636, 4.8514148, 4.73694293, 4.6188237, 4.49733003, 4.37273484, 4.24531108, 4.11533167, 3.98306954, 3.84879763, 3.71278886, 3.57531616, 3.43665247, 3.29707071, 3.15684382, 3.01624473, 2.87554636, 2.73502165, 2.59494353, 2.45558493, 2.31721878, 2.18011801, 2.05416018, 1.93461388, 1.81663227, 1.7001766, 1.5852081, 1.47168802, 1.3595776, 1.24883808, 1.1394307, 1.03131671, 0.924457338, 0.818813839, 0.714347451, 0.611019416, 0.508790975, 0.407623371, 0.307477845, 0.199582755, 0.0928215383, -0.0126552498, -0.116697056, -0.219153327, -0.319873508, -0.418707046, -0.515503387, -0.610111977, -0.702382262, -0.792163689, -0.879305704, -0.963657753, -1.04506928, -1.12338974, -1.19846856, -1.27264563, -1.34820102, -1.41984787, -1.48727416, -1.55016792, -1.60821713, -1.6611098, -1.70853394, -1.75017754, -1.7857286, -1.81487514, -1.83730515, -1.85270664, -1.86076759, -1.86117603, -1.85361995, -1.83778735, -1.55616599, -1.14245385, -0.733440278, -0.337159907, 0.0383526144, 0.385062643, 0.694935532, 0.959936639, 1.17203132, 1.32318493, 1.40536282, 1.41053034, 1.33065287, 1.15769574, 0.883624314, 0.50040395, 0, -1.12308629, -2.30601721, -3.55885283, -4.89165326, -6.31447859, -7.83738892, -9.47044434, -11.2237049, -13.1072308, -15.1310821, -17.3053188, -19.6400011, -22.145189, -24.8309427, -27.7073223, -30.7843878, -33.9138759, -37.0832292, -40.6606201, -44.6221823, -48.9440494, -53.6023551, -58.573233, -63.8328169, -69.3572403, -75.1226369, -81.1051404, -87.2808845, -93.6260027, -100.293645, -107.018023, -113.722165, -120.329101, -127.691685, -135.268033, -142.268306, -148.336585, -153.116949, -156.253481, -157.390259, -159.311419, -157.565318, -151.181679, -139.190222, -120.62067, -100.570401, -70.4845674, -27.8759758, 29.7425657, 104.858249, 194.894057, 313.590981, 464.560735, 653.41647, 889.857614, 1181.74894, 1536.95522, 1943.76431, 2434.06721, 3061.11154, 3861.0946, 4848.28948, 6002.04419, 7347.41686, 8986.91891, 10987.5094, 13401.7401, 16292.8349, 19776.8445, 23939.9378, 28743.1217, 34416.9828, 41091.0154, 48894.7138, 57957.5724, 72801.1185, 88064.805, 102780.118, 115978.542, 126691.565, 130792.63, 132576.626, 132428.486, 131499.846, 129872.151, 127307.387, 123902.973, 119756.329, 114964.873, 109626.024, 104872.951, 98657.4265, 91668.8903, 84596.7815, 78130.5398, 72959.6044, 69773.415, 65352.9463, 61566.1192},
        {5.60830886, 5.54339307, 5.47286024, 5.39693094, 5.31582573, 5.22976516, 5.13896979, 5.04366016, 4.94405684, 4.84038039, 4.73285135, 4.62169029, 4.50711775, 4.38935429, 4.26862048, 4.14513686, 4.01912399, 3.89080243, 3.76039272, 3.62811543, 3.49419112, 3.35884033, 3.22228362, 3.08474156, 2.94643468, 2.80758355, 2.66840873, 2.52913077, 2.38997022, 2.25114765, 2.1128836, 1.97539863, 1.83891329, 1.70364815, 1.57772944, 1.45725858, 1.33824199, 1.22064367, 1.10442762, 0.989557839, 0.87599833, 0.763713094, 0.652666131, 0.542821443, 0.43414303, 0.326594893, 0.220141034, 0.114745453, 0.0103721522, -0.0930148688, -0.195451608, -0.3074991, -0.418440147, -0.528082584, -0.636234251, -0.742702983, -0.847296617, -0.94982299, -1.05008994, -1.1479053, -1.24307691, -1.33541261, -1.42472024, -1.51080762, -1.5934826, -1.67255302, -1.7478267, -1.82136404, -1.89517134, -1.96441071, -2.02874399, -2.08783298, -2.14133949, -2.18892534, -2.23025234, -2.26498231, -2.29277706, -2.3132984, -2.32620814, -2.33116811, -2.32784011, -2.31588595, -2.29496746, -2.26474644, -1.95716352, -1.51137227, -1.06974404, -0.640628719, -0.232376186, 0.146663676, 0.488140987, 0.783705867, 1.02500843, 1.20369881, 1.31142711, 1.33984345, 1.28059796, 1.12534075, 0.865721936, 0.493391648, 0, -1.10343036, -2.2807079, -3.54070558, -4.89229635, -6.34435315, -7.90574896, -9.58535671, -11.3920494, -13.3346999, -15.4221812, -17.6633662, -20.067128, -22.6423395, -25.3978735, -28.3426032, -31.4854013, -34.6794163, -37.9112747, -41.5420837, -45.5473469, -49.9025675, -54.5832489, -59.5648946, -64.8230078, -70.3330921, -76.0706507, -82.0111871, -88.1302046, -94.4032066, -100.98131, -107.600013, -114.182135, -120.650496, -127.875766, -135.30684, -142.139849, -148.013258, -152.565531, -155.43513, -156.260521, -157.814513, -155.647292, -148.784079, -136.250095, -117.070563, -96.5694632, -65.8985687, -22.5084117, 36.1504757, 112.627561, 204.889167, 325.669529, 477.190183, 664.680098, 899.508423, 1190.28749, 1545.62964, 1957.99946, 2455.82697, 3085.15789, 3873.606, 4844.42995, 5989.48449, 7336.39485, 8973.39737, 10960.9069, 13341.3317, 16170.4184, 19569.9372, 23624.3117, 28308.9005, 33833.9713, 40321.7008, 47894.2657, 56673.8428, 70912.1639, 85569.462, 99735.525, 112500.141, 122953.098, 127286.248, 129419.272, 129683.932, 128975.93, 127403.8, 124841.099, 121408.11, 117225.116, 112412.398, 107090.239, 101437.606, 95229.0033, 88842.464, 82656.0219, 77047.7105, 72395.5636, 69077.6146, 64681.8506, 60677.8423},
        {5.36253757, 5.27186621, 5.17750485, 5.07960524, 4.97831913, 4.87379828, 4.76619442, 4.65565931, 4.54234471, 4.42640235, 4.30798399, 4.18724138, 4.06432626, 3.9393904, 3.81258553, 3.68406341, 3.55397578, 3.4224744, 3.28971101, 3.15583737, 3.02100523, 2.88536633, 2.74907242, 2.61227525, 2.47512658, 2.33777815, 2.20038171, 2.06308902, 1.92605181, 1.78942185, 1.65335088, 1.51799064, 1.3834929, 1.2500094, 1.12301893, 0.999897688, 0.877957809, 0.757178172, 0.637537655, 0.519015133, 0.401589483, 0.285239581, 0.169944305, 0.0556825292, -0.0575668684, -0.169825012, -0.281113024, -0.391452029, -0.500863151, -0.609367513, -0.716986237, -0.835732508, -0.953375425, -1.06967615, -1.18439583, -1.29729564, -1.40813674, -1.51668027, -1.62268741, -1.72591931, -1.82613712, -1.92310202, -2.01657515, -2.10631768, -2.19209077, -2.27365557, -2.35077325, -2.4252755, -2.49894432, -2.56727052, -2.62988104, -2.68640282, -2.73646279, -2.77968788, -2.81570502, -2.84414116, -2.86462323, -2.87677816, -2.88023288, -2.87461433, -2.85954944, -2.83466515, -2.7995884, -2.75394611, -2.41760997, -1.93596881, -1.45777391, -0.991730807, -0.546545028, -0.130922103, 0.246432438, 0.576813063, 0.851514243, 1.06183045, 1.19905614, 1.25448581, 1.2194139, 1.0851349, 0.842943266, 0.484133477, 0, -1.07967154, -2.24856287, -3.51427477, -4.884408, -6.36656332, -7.9683415, -9.69734332, -11.5611695, -13.5674209, -15.7236982, -18.0376022, -20.5167336, -23.1686933, -26.0010819, -29.0215003, -32.2375492, -35.5024662, -38.8018, -42.4923065, -46.5485086, -50.9449292, -55.6560912, -60.6565176, -65.9207312, -71.4232551, -77.1386121, -83.0413251, -89.1059171, -95.306911, -101.796495, -108.306384, -114.757799, -121.071964, -128.093619, -135.28223, -141.850989, -147.444064, -151.705625, -154.279839, -154.810873, -156.241741, -153.91377, -146.825703, -133.976282, -114.364248, -93.1347134, -61.6034587, -17.2326343, 42.5156097, 120.179123, 214.07667, 336.121847, 486.126028, 668.82432, 899.004457, 1186.94075, 1542.9075, 1967.98099, 2483.72402, 3121.73363, 3893.20696, 4837.97696, 5967.12104, 7313.14659, 8944.87285, 10914.4513, 13249.0766, 15994.4297, 19287.7009, 23214.3176, 27751.2303, 33109.0322, 39410.7005, 46779.2124, 55337.5453, 69245.5151, 83580.0504, 97454.9189, 109983.888, 120280.725, 124673.902, 126919.346, 127324.522, 126642.594, 125004.605, 122355.89, 118836.779, 114587.601, 109748.684, 104460.357, 98256.6371, 92217.562, 86453.2192, 81073.6964, 76189.0813, 71909.4615, 68344.9247, 64227.7714, 59837.6286},
        {5.15657962, 5.04196534, 4.92538935, 4.80693967, 4.68670436, 4.56477146, 4.44122899, 4.31616501, 4.18966756, 4.06182468, 3.9327244, 3.80245477, 3.67110383, 3.53875961, 3.40551017, 3.27144354, 3.13664776, 3.00121087, 2.86522092, 2.72876594, 2.59193397, 2.45481306, 2.31749125, 2.18005657, 2.04259707, 1.90520079, 1.76795577, 1.63095004, 1.49427166, 1.35800866, 1.22224908, 1.08708096, 0.952592348, 0.818871278, 0.688512054, 0.560296869, 0.432977795, 0.306561541, 0.181054813, 0.0564643198, -0.0672032309, -0.189941131, -0.311742674, -0.432601151, -0.552509854, -0.671462076, -0.78945111, -0.906470246, -1.02251278, -1.137572, -1.2516412, -1.37695296, -1.50099596, -1.62349817, -1.74418756, -1.86279209, -1.97903974, -2.09265847, -2.20337625, -2.31092106, -2.41502084, -2.51540359, -2.61179726, -2.70392982, -2.79152925, -2.8743235, -2.95204056, -3.02651147, -3.0995168, -3.16644706, -3.22689388, -3.28044888, -3.32670371, -3.36525, -3.39567938, -3.41758349, -3.43055395, -3.43418241, -3.42806049, -3.41177983, -3.38493207, -3.34710883, -3.29790175, -3.23690247, -2.87199505, -2.35476059, -1.84028628, -1.33763021, -0.855850507, -0.404005261, 0.00884741803, 0.373649426, 0.681342659, 0.92286901, 1.08917038, 1.17118865, 1.15986573, 1.04614351, 0.820963886, 0.47526875, 0, -1.03542166, -2.17924487, -3.43755714, -4.81644598, -6.32199887, -7.96030333, -9.73744686, -11.659517, -13.7326011, -15.9627869, -18.3561617, -20.9188131, -23.6568285, -26.5762956, -29.6833017, -32.9839344, -36.3377786, -39.7299929, -43.4980396, -47.6166129, -52.0604068, -56.8041155, -61.8224331, -67.0900535, -72.581671, -78.2719795, -84.1356732, -90.1474462, -96.2819925, -102.707731, -109.134598, -115.479172, -121.658028, -128.467249, -135.383658, -141.649546, -146.917637, -150.840656, -153.071329, -153.262379, -154.54682, -152.03787, -144.707332, -131.527003, -111.468684, -89.3992707, -56.9216916, -11.5339722, 49.2658623, 127.979787, 223.16126, 346.086736, 494.312857, 672.131978, 897.58853, 1182.55254, 1538.89405, 1976.07072, 2508.9132, 3154.8838, 3909.55837, 4827.99396, 5938.00385, 7276.47541, 8895.67166, 10840.7979, 13126.3714, 15789.641, 18980.3797, 22784.7103, 27172.2483, 32363.5485, 38485.0636, 45663.2461, 54024.5486, 67718.4812, 81842.6747, 95517.817, 107864.597, 118003.701, 122293.007, 124457.889, 124803.099, 124044.549, 122324.237, 119612.848, 116060.947, 111819.101, 107037.876, 101867.837, 95518.4697, 89772.6305, 84560.2166, 79811.125, 75455.2529, 71422.4972, 67642.7549, 63747.8641, 58973.4667},
        {5.01139201, 4.88568533, 4.75834771, 4.62945353, 4.49907718, 4.36729304, 4.23417549, 4.09979892, 3.96423771, 3.82756625, 3.68985892, 3.55119011, 3.41163419, 3.27126556, 3.13015859, 2.98838768, 2.8460272, 2.70315153, 2.55983508, 2.41615221, 2.27217731, 2.12798477, 1.98364897, 1.83924429, 1.69484512, 1.55052584, 1.40636084, 1.2624245, 1.11879121, 0.975535339, 0.832731285, 0.690453428, 0.548776152, 0.407773841, 0.267873139, 0.128964892, -0.00906428998, -0.146151455, -0.282233649, -0.417247921, -0.551131318, -0.683820887, -0.815253675, -0.94536673, -1.0740971, -1.20138183, -1.32715797, -1.45136257, -1.57393267, -1.69480532, -1.81391758, -1.93940467, -2.06282774, -2.18394611, -2.3025191, -2.41830604, -2.53106627, -2.6405591, -2.74654386, -2.84877988, -2.94702649, -3.04104301, -3.13058877, -3.21542309, -3.29530531, -3.36999474, -3.43925072, -3.50564024, -3.57166291, -3.63128741, -3.68409108, -3.72965123, -3.76754517, -3.79735022, -3.81864369, -3.83100291, -3.83400519, -3.82722784, -3.81024818, -3.78264353, -3.74399121, -3.69386852, -3.6318528, -3.55752134, -3.17026146, -2.62605957, -2.08451014, -1.55493224, -1.04664491, -0.568967224, -0.131218238, 0.257282989, 0.587217396, 0.849265925, 1.03410951, 1.13242911, 1.13490564, 1.03222006, 0.815053294, 0.474086295, 0, -0.91853828, -1.97729305, -3.18005779, -4.53062595, -6.032791, -7.69034642, -9.50708567, -11.4868022, -13.6332895, -15.950341, -18.4417503, -21.1113107, -23.9628157, -27.0000588, -30.2268335, -33.6469332, -37.1560526, -40.7425424, -44.6547054, -48.873171, -53.3785687, -58.1515278, -63.1726778, -68.4226481, -73.8820682, -79.5315675, -85.3517754, -91.3233213, -97.4268348, -103.908466, -110.385965, -116.760307, -122.932465, -129.629082, -136.338296, -142.328067, -147.251666, -150.762365, -152.513436, -152.15815, -152.796042, -149.585259, -141.504804, -127.533681, -106.650892, -83.7304215, -50.3825469, -4.11252637, 57.5743819, 137.17292, 233.254108, 357.032069, 506.445234, 685.807258, 912.447868, 1197.85284, 1553.50795, 1988.22671, 2516.7745, 3156.83187, 3905.55764, 4814.30026, 5905.32878, 7213.82562, 8796.85414, 10701.434, 12950.099, 15583.5206, 18732.1518, 22475.66, 26762.2328, 31823.2453, 37783.8774, 44769.3089, 52904.7197, 66310.1404, 80117.1874, 93452.3276, 105442.028, 115212.756, 119079.414, 120854.409, 120869.167, 119997.374, 118355.808, 115835.304, 112563.935, 108669.778, 104280.906, 99525.3951, 94011.5954, 88748.4548, 83762.7613, 79081.3029, 74730.8676, 70738.2435, 67130.2185, 62748.3103, 57962.9708},
        {4.90545191, 4.77168523, 4.63625569, 4.49923701, 4.36070294, 4.22072721, 4.07938354, 3.93674568, 3.79288735, 3.64788229, 3.50180424, 3.35472692, 3.20672406, 3.05786942, 2.9082367, 2.75789966, 2.60693202, 2.45540751, 2.30339988, 2.15098284, 1.99823014, 1.84521551, 1.69201268, 1.53869539, 1.38533736, 1.23201234, 1.07879405, 0.925756229, 0.77297261, 0.620516926, 0.468462911, 0.316884296, 0.165854817, 0.0154482052, -0.135803179, -0.287022494, -0.437314374, -0.586555065, -0.734620813, -0.881387864, -1.02673246, -1.17053086, -1.3126593, -1.45299402, -1.59141128, -1.72778732, -1.86199839, -1.99392073, -2.12343058, -2.2504042, -2.37471784, -2.4995875, -2.62147727, -2.74019098, -2.8555325, -2.96730565, -3.0753143, -3.17936228, -3.27925345, -3.37479164, -3.46578071, -3.5520245, -3.63332686, -3.70949164, -3.78032267, -3.84562381, -3.90519891, -3.96259437, -4.02126675, -4.07330114, -4.11825877, -4.15570086, -4.18518866, -4.20628338, -4.21854627, -4.22153856, -4.21482146, -4.19795623, -4.17050408, -4.13202625, -4.08208397, -4.02023847, -3.94605099, -3.85908275, -3.45083432, -2.88137692, -2.31448439, -1.75971512, -1.22662751, -0.72477995, -0.263730826, 0.146961466, 0.497738534, 0.779041987, 0.981313433, 1.09499448, 1.11052673, 1.0183518, 0.808911301, 0.47264683, 0, -0.793360176, -1.75945715, -2.89975974, -4.21573676, -5.70885701, -7.38058932, -9.23240249, -11.2657653, -13.4821467, -15.8830153, -18.4698401, -21.2440898, -24.2072332, -27.3607392, -30.7060766, -34.2447141, -37.9093148, -41.691518, -45.7484249, -50.0667602, -54.6332485, -59.4346145, -64.4575831, -69.6888787, -75.1152263, -80.7233504, -86.4999757, -92.431827, -98.505629, -105.047416, -111.580189, -117.988881, -124.158423, -130.757186, -137.278383, -143.006309, -147.590863, -150.681945, -151.929457, -150.983297, -150.822015, -146.753794, -137.777279, -122.891111, -101.093933, -77.4039189, -43.2952982, 3.73816818, 66.2027197, 146.604596, 243.565657, 368.224097, 519.460673, 701.680651, 930.568483, 1216.7445, 1570.82902, 1998.60669, 2516.63254, 3146.73694, 3892.40635, 4796.1275, 5869.24196, 7143.53056, 8685.40733, 10545.3891, 12757.4877, 15367.9408, 18483.5046, 22176.7026, 26375.841, 31319.8479, 37131.1683, 43932.2469, 51845.5287, 64961.4607, 78442.4852, 91419.0465, 103021.589, 112380.557, 115741.694, 117043.279, 116652.058, 115654.83, 114133.319, 111869.225, 108959.275, 105500.193, 101588.703, 97321.5312, 92867.519, 88123.8541, 83278.6237, 78519.9148, 74035.8147, 70014.4104, 66643.789, 61638.1685, 56879.1163},
        {4.83189122, 4.69047374, 4.54726304, 4.40233646, 4.25577132, 4.10764497, 3.95803473, 3.80701792, 3.65467189, 3.50107396, 3.34630147, 3.19043174, 3.0335421, 2.87570989, 2.71701244, 2.55752708, 2.39733113, 2.23650193, 2.07511681, 1.9132531, 1.75098814, 1.58839925, 1.42556376, 1.262559, 1.09946231, 0.936351013, 0.773302445, 0.610393933, 0.447702809, 0.285306405, 0.12328205, -0.0382929242, -0.199341187, -0.359785408, -0.52267554, -0.686304834, -0.848928948, -1.01036907, -1.17044637, -1.32898205, -1.48579728, -1.64071324, -1.79355112, -1.9441321, -2.09227736, -2.23780809, -2.38054547, -2.52031069, -2.65692491, -2.79020934, -2.91998514, -3.0449525, -3.16607791, -3.28320686, -3.39618481, -3.50485727, -3.6090697, -3.70866761, -3.80349646, -3.89340175, -3.97822895, -4.05782356, -4.13203105, -4.20069691, -4.26366662, -4.32078567, -4.37189954, -4.42162996, -4.47448396, -4.5204564, -4.55908314, -4.58990004, -4.61244296, -4.62624776, -4.6308503, -4.62578644, -4.61059205, -4.58480297, -4.54795507, -4.4995842, -4.43922624, -4.36641704, -4.28069245, -4.18158835, -3.75261786, -3.15797532, -2.56565342, -1.98544355, -1.42713712, -0.900525548, -0.415400226, 0.0184474391, 0.391226041, 0.693144175, 0.914410436, 1.04523342, 1.07582172, 0.996383934, 0.797128655, 0.468264479, 0, -0.684042333, -1.56927832, -2.65512668, -3.94100614, -5.42633541, -7.11053322, -8.99301829, -11.0732093, -13.3505251, -15.8243842, -18.4942055, -21.3594077, -24.4194094, -27.6736295, -31.1214866, -34.7623994, -38.5560337, -42.4961292, -46.6720221, -51.0746126, -55.6948009, -60.5234872, -65.5515719, -70.769955, -76.1695369, -81.7412177, -87.4758977, -93.3644771, -99.3978562, -105.961219, -112.506612, -118.906649, -125.033944, -131.531356, -137.886379, -143.381107, -147.657083, -150.35585, -151.118949, -149.587923, -148.537962, -143.523241, -133.572197, -117.713268, -94.9748914, -70.6921506, -36.0101745, 11.6192615, 74.7443819, 155.913411, 253.962066, 379.688589, 532.972723, 718.266199, 949.355225, 1235.83502, 1587.30079, 2005.43924, 2509.55149, 3127.56621, 3871.54167, 4772.91006, 5830.9416, 7073.26745, 8575.57481, 10391.4923, 12564.9475, 15147.0539, 18224.4962, 21863.838, 25983.5131, 30821.0043, 36494.1508, 43120.7919, 50818.7671, 63626.5574, 76761.1999, 89360.3736, 100561.757, 109503.03, 112376.864, 113229.283, 112460.11, 111364.049, 109986.701, 108000.803, 105470.597, 102460.331, 99034.2482, 95256.5941, 91789.3738, 87524.6338, 82802.1771, 77961.8069, 73343.326, 69286.5376, 66131.2446, 60515.0217, 55727.4828},
        {4.76909343, 4.61293916, 4.45551368, 4.29687769, 4.13709185, 3.97621683, 3.81431331, 3.65144197, 3.48766347, 3.32303849, 3.15762771, 2.9914918, 2.82469143, 2.65728727, 2.48934001, 2.32091031, 2.15205884, 1.98284629, 1.81333332, 1.64358061, 1.47364883, 1.30359866, 1.13349077, 0.963385833, 0.793344523, 0.623427515, 0.453695481, 0.284209097, 0.115029035, -0.0537840307, -0.222169426, -0.390066477, -0.55741451, -0.724152852, -0.894269115, -1.06559333, -1.23590688, -1.4050177, -1.57273376, -1.73886301, -1.90321339, -2.06559285, -2.22580936, -2.38367085, -2.53898529, -2.69156061, -2.84120479, -2.98772576, -3.13093147, -3.27062989, -3.40662896, -3.53765029, -3.66461173, -3.78734477, -3.90568092, -4.01945167, -4.12848853, -4.23262301, -4.33168659, -4.42551078, -4.51392709, -4.59676701, -4.67386205, -4.74504371, -4.81014348, -4.86899287, -4.92142338, -4.97287009, -5.02863354, -5.07698826, -5.11740252, -5.14934454, -5.17228258, -5.18568487, -5.18901966, -5.1817552, -5.16335973, -5.1333015, -5.09104874, -5.0360697, -4.96783263, -4.88580577, -4.78945737, -4.67825566, -4.22485228, -3.59899792, -2.97437138, -2.36104671, -1.769098, -1.20859932, -0.689624732, -0.222248314, 0.183455863, 0.517413728, 0.769551208, 0.929794232, 0.988068729, 0.934300625, 0.758415851, 0.450340333, 0, -0.679378557, -1.56629017, -2.65957702, -3.95808126, -5.46064509, -7.16611067, -9.07332017, -11.1811158, -13.4883396, -15.9938339, -18.6964409, -21.5950026, -24.6883613, -27.9753591, -31.4548382, -35.1256408, -38.9165154, -42.8162907, -46.9823541, -51.3985279, -56.0486341, -60.9164948, -65.9859321, -71.240768, -76.6648247, -82.2419242, -87.9558887, -93.7905401, -99.7297007, -106.09419, -112.407267, -118.551654, -124.410075, -130.753266, -137.019943, -142.412018, -146.545809, -149.037634, -149.503812, -147.560661, -145.585658, -139.593609, -128.660608, -111.862747, -88.2761166, -64.1150819, -29.532894, 18.1789232, 81.728846, 163.825351, 264.348779, 392.29235, 546.95669, 732.347555, 962.077017, 1245.44411, 1591.74785, 2000.60345, 2493.19464, 3101.26937, 3842.12847, 4741.12603, 5797.29932, 7035.59952, 8524.88755, 10313.869, 12435.5462, 14934.5535, 17913.1043, 21440.842, 25477.637, 30219.7939, 35774.9547, 42250.7613, 49754.8554, 62158.1373, 74864.176, 87039.7986, 97851.8325, 106467.105, 109183.385, 109949.263, 109161.164, 108134.474, 106901.861, 105113.09, 102819.974, 100074.325, 96927.9564, 93432.6797, 89754.9429, 85608.5141, 81211.2968, 76781.1949, 72536.1119, 68693.9515, 65472.6174, 59794.6605, 54620.4552},
        {4.73115839, 4.55748467, 4.38328672, 4.20860191, 4.03346765, 3.85792132, 3.68200032, 3.50574203, 3.32918384, 3.15236314, 2.97531733, 2.79808379, 2.62069992, 2.4432031, 2.26563073, 2.08802019, 1.91040888, 1.73283419, 1.5553335, 1.3779442, 1.2007037, 1.02364937, 0.846818608, 0.670248804, 0.493977348, 0.31804163, 0.14247904, -0.0326730307, -0.207377193, -0.381596056, -0.555292229, -0.728428323, -0.900966947, -1.07287071, -1.24879973, -1.42626907, -1.6027373, -1.77801459, -1.95191111, -2.12423703, -2.29480251, -2.46341772, -2.62989283, -2.794038, -2.95566339, -3.11457918, -3.27059554, -3.42352262, -3.57317059, -3.71934963, -3.86186989, -4.00110179, -4.1362831, -4.26721184, -4.39368605, -4.51550373, -4.63246292, -4.74436163, -4.85099788, -4.95216971, -5.04767512, -5.13731214, -5.22087879, -5.29817309, -5.36899307, -5.43313674, -5.49040213, -5.54692713, -5.60869741, -5.66243549, -5.7075284, -5.74336318, -5.76932686, -5.78480647, -5.78918904, -5.78186162, -5.76221124, -5.72962492, -5.6834897, -5.62319262, -5.54812071, -5.457661, -5.35120053, -5.22812633, -4.7495761, -4.09117617, -3.43261834, -2.78426649, -2.1564845, -1.55963625, -1.0040856, -0.500196431, -0.0583326214, 0.311141956, 0.597863427, 0.791467915, 0.881591546, 0.857870445, 0.709940737, 0.427438547, 0, -0.711626232, -1.62912631, -2.75127386, -4.07684251, -5.60460586, -7.33333756, -9.26181121, -11.3888004, -13.7130789, -16.2334201, -18.9485979, -21.8573856, -24.9585571, -28.2508859, -31.7331456, -35.4041099, -39.1456546, -42.9398536, -47.0498922, -51.4494946, -56.1123847, -61.0122867, -66.1229247, -71.4180228, -76.8713051, -82.4564957, -88.1473187, -93.9174982, -99.7407582, -105.840738, -111.849611, -117.666127, -123.189036, -129.35114, -135.534162, -140.844078, -144.869421, -147.198724, -147.420522, -145.123346, -142.264249, -135.342393, -123.482906, -105.810914, -81.4515455, -57.5931175, -23.2817708, 24.3731642, 88.2623573, 171.276478, 274.536099, 404.865439, 560.511804, 744.898878, 972.018799, 1251.11868, 1591.44565, 1991.3551, 2473.46125, 3072.26023, 3808.8345, 4704.41276, 5761.44741, 7002.82332, 8486.04707, 10252.7333, 12319.3607, 14721.027, 17584.9957, 20986.3962, 24939.6201, 29589.8788, 35033.326, 41366.1158, 48684.402, 60654.4891, 72909.8428, 84654.0794, 95090.8153, 103423.667, 106114.051, 106935.916, 106268.106, 105371.624, 104276.128, 102643.707, 100522.397, 97960.233, 95005.2493, 91705.4812, 87498.5969, 83377.636, 79361.1791, 75467.8066, 71716.099, 68124.6368, 64712.0004, 59122.2699, 53411.7667},
        {4.7216673, 4.52852011, 4.33574013, 4.14333727, 3.95132138, 3.75970235, 3.56849007, 3.3776944, 3.18732522, 2.99739243, 2.80790588, 2.61887547, 2.43031107, 2.24222256, 2.05461981, 1.86751272, 1.68091114, 1.49482498, 1.30926409, 1.12423837, 0.939757681, 0.755831915, 0.572470945, 0.38968465, 0.20748291, 0.0258756023, -0.155127394, -0.335516201, -0.515280939, -0.69441173, -0.872898695, -1.05073196, -1.22790163, -1.40439785, -1.58545785, -1.76833782, -1.95023089, -2.13095691, -2.31033572, -2.48818716, -2.66433107, -2.8385873, -3.01077569, -3.18071608, -3.34822831, -3.51313222, -3.67524766, -3.83439447, -3.9903925, -4.14306157, -4.29222154, -4.44068348, -4.58521116, -4.72555957, -4.86148372, -4.99273859, -5.1190792, -5.24026054, -5.3560376, -5.46616539, -5.5703989, -5.66849313, -5.76020308, -5.84528375, -5.92349014, -5.99457724, -6.05830006, -6.12139017, -6.19041166, -6.25073084, -6.30165044, -6.3424732, -6.37250186, -6.39103914, -6.39738779, -6.39085054, -6.37073012, -6.33632926, -6.28695071, -6.2218972, -6.14047146, -6.04197622, -5.92571422, -5.7909882, -5.28752012, -4.59675605, -3.90434378, -3.22093061, -2.55716383, -1.92369076, -1.33115869, -0.790214934, -0.311506787, 0.0943184448, 0.416613459, 0.644730952, 0.768023622, 0.775844164, 0.657545277, 0.402479656, 0, -0.759143517, -1.71914684, -2.87889494, -4.23727281, -5.79316543, -7.5454578, -9.49303488, -11.6347817, -13.9695831, -16.4963243, -19.2138901, -22.1211655, -25.2170356, -28.5003852, -31.9700995, -35.6250633, -39.2998078, -42.9697388, -47.0054941, -51.3707401, -56.0291431, -60.9443693, -66.080085, -71.3999565, -76.8676501, -82.4468322, -88.1011689, -93.7943266, -99.4899716, -105.315465, -111.010758, -116.490406, -121.668967, -127.626924, -133.695871, -138.896492, -142.793567, -144.951878, -144.936206, -142.31133, -138.679285, -130.947565, -118.282357, -99.8498495, -74.81623, -51.2232359, -17.1426177, 30.4783243, 94.6922897, 178.551978, 284.36235, 416.837748, 572.995303, 755.6602, 979.515742, 1253.92118, 1588.23576, 1980.42145, 2453.82508, 3044.22689, 3774.70691, 4664.29917, 5720.56446, 6964.98424, 8442.62769, 10187.7369, 12199.7441, 14503.8667, 17252.9574, 20527.6712, 24395.2029, 28951.7384, 34282.5918, 40473.0775, 47608.5096, 59141.5761, 70948.3741, 82272.3743, 92357.0475, 100445.864, 103194.007, 104158.761, 103692.863, 102953.384, 101974.678, 100455.963, 98448.3516, 96002.9553, 93170.8868, 90003.2585, 85237.3822, 81143.6968, 77524.6892, 74182.8459, 70920.6538, 67540.5995, 63845.1698, 58369.9818, 52031.3248},
        {4.74779677, 4.53336761, 4.32033605, 4.10867961, 3.8983758, 3.68940212, 3.48173608, 3.27535519, 3.07023697, 2.86635891, 2.66369853, 2.46223333, 2.26194083, 2.06279854, 1.86478395, 1.66787459, 1.47204796, 1.27728156, 1.08355291, 0.890839518, 0.699118889, 0.508368533, 0.318565958, 0.129688674, -0.0582858108, -0.245379988, -0.431616349, -0.617017385, -0.801605587, -0.985403446, -1.16843345, -1.3507181, -1.53227988, -1.71314128, -1.8994894, -2.08813476, -2.2758138, -2.46234894, -2.64756264, -2.83127733, -3.01331544, -3.19349942, -3.37165171, -3.54759473, -3.72115094, -3.89214276, -4.06039264, -4.22572301, -4.38795632, -4.54691499, -4.70242148, -4.85904361, -5.01175556, -5.1602769, -5.30432718, -5.44362597, -5.57789285, -5.70684737, -5.83020911, -5.94769763, -6.05903251, -6.1639333, -6.26211957, -6.35331089, -6.43722683, -6.51358696, -6.58211084, -6.65000525, -6.72429721, -6.78926447, -6.84414123, -6.8881617, -6.92056006, -6.94057053, -6.9474273, -6.94036458, -6.91861657, -6.88141746, -6.82800147, -6.75760279, -6.66945562, -6.56279417, -6.43685263, -6.2908652, -5.76461781, -5.04462929, -4.32173881, -3.60684535, -2.91084791, -2.24464547, -1.61913703, -1.04522157, -0.533798079, -0.0957655555, 0.257977015, 0.516530643, 0.668996338, 0.704475112, 0.612067973, 0.380875932, 0, -0.78046735, -1.76246618, -2.94450792, -4.32510404, -5.90276596, -7.67600514, -9.64333301, -11.803261, -14.1543006, -16.6949633, -19.4237604, -22.3392034, -25.4398038, -28.724073, -32.1905224, -35.8376636, -39.4874869, -43.1132867, -47.1128478, -51.4468557, -56.0759961, -60.9609545, -66.0624162, -71.3410671, -76.7575924, -82.2726778, -87.8470088, -93.441271, -99.0161498, -104.680645, -110.192746, -115.468645, -120.424531, -126.157485, -131.993255, -136.942682, -140.567693, -142.430211, -142.092162, -139.115471, -135.051696, -126.867587, -113.735792, -94.8289596, -69.3197369, -45.466685, -11.0850619, 36.9239622, 101.659217, 186.219532, 293.515873, 427.064546, 582.85216, 763.298625, 983.958513, 1254.4892, 1584.54807, 1973.5959, 2443.80406, 3028.46752, 3748.52919, 4624.42289, 5667.79652, 6900.05931, 8359.16064, 10075.4577, 12041.2517, 14277.0335, 16942.8323, 20119.7463, 23891.7207, 28336.5691, 33534.4045, 39565.3399, 46509.4881, 57654.6561, 69071.3394, 80037.7275, 89832.0098, 97732.3757, 100576.46, 101710.043, 101457.267, 100837.342, 99899.2789, 98402.4754, 96410.4229, 93986.6124, 91194.5347, 88097.6809, 83311.9176, 79463.5338, 76224.6589, 73267.4227, 70263.9547, 66886.3846, 62806.8419, 57244.2918, 50278.578},
        {4.81153752, 4.57448265, 4.33995003, 4.10788122, 3.87821779, 3.65090133, 3.42587339, 3.20307557, 2.98244943, 2.76393654, 2.54747848, 2.33301682, 2.12049314, 1.90984901, 1.70102601, 1.4939657, 1.28860967, 1.08489948, 0.882776708, 0.682182931, 0.483059722, 0.285348653, 0.0889912987, -0.106070767, -0.299895971, -0.492542738, -0.684069496, -0.87453467, -1.06399669, -1.25251397, -1.44014495, -1.62694805, -1.8129817, -1.99830432, -2.19054407, -2.38581529, -2.58014134, -2.77333499, -2.96520903, -3.15557622, -3.34424933, -3.53104115, -3.71576445, -3.89823199, -4.07825656, -4.25565093, -4.43022787, -4.60180015, -4.77018056, -4.93518186, -5.09661683, -5.25959844, -5.41852437, -5.5730925, -5.72300069, -5.86794683, -6.0076288, -6.14174446, -6.2699917, -6.39206839, -6.5076724, -6.61650162, -6.71825392, -6.81262717, -6.89931925, -6.97802804, -7.04845141, -7.11813696, -7.19444381, -7.26088038, -7.31663569, -7.36089875, -7.39285856, -7.41170414, -7.4166245, -7.40680864, -7.38144559, -7.33972434, -7.28083392, -7.20396332, -7.10830157, -6.99303767, -6.85736064, -6.70045948, -6.15418084, -5.40911146, -4.66025326, -3.91871623, -3.19561034, -2.50204556, -1.84913187, -1.24797925, -0.709697676, -0.245397126, 0.133812423, 0.416820995, 0.592518613, 0.649795299, 0.577541077, 0.36464597, 0, -0.7612775, -1.73355496, -2.91426394, -4.30083602, -5.89070277, -7.68129576, -9.67004656, -11.8543868, -14.2317479, -16.7995616, -19.5552593, -22.4962728, -25.6200334, -28.9239729, -32.4055228, -36.0621146, -39.7485539, -43.441897, -47.4621822, -51.7762782, -56.351054, -61.1533783, -66.15012, -71.3081479, -76.594331, -81.9755379, -87.4186376, -92.890499, -98.3579908, -104.012713, -109.514402, -114.762507, -119.656478, -125.144818, -130.602449, -135.124547, -138.295845, -139.701076, -138.924974, -135.552271, -131.453509, -123.231932, -110.025053, -90.9703806, -65.2054257, -40.4104752, -5.0445686, 43.8904778, 109.392848, 194.460725, 301.851352, 435.098338, 589.549787, 767.496653, 985.404884, 1253.37614, 1581.5121, 1972.78985, 2446.05485, 3027.92476, 3732.65591, 4585.87419, 5601.16279, 6801.29593, 8224.62156, 9902.33901, 11832.9094, 14039.0443, 16663.4957, 19781.114, 23446.2749, 27757.4533, 32796.5872, 38645.6143, 45386.4724, 56210.0887, 67312.1643, 77997.3894, 87570.4546, 95336.0503, 98284.2989, 99577.5041, 99512.3141, 98951.9827, 97968.8766, 96399.2768, 94327.3544, 91837.2803, 89013.2255, 85939.3611, 81865.6259, 78543.6555, 75641.5045, 72827.2274, 69768.8789, 66134.5136, 61592.186, 55659.2521, 48103.9687},
        {4.9048938, 4.64494896, 4.38862529, 4.13582987, 3.88646976, 3.64045204, 3.39768378, 3.15807205, 2.92152392, 2.68794646, 2.45724674, 2.22933184, 2.00410882, 1.78148476, 1.56136672, 1.34366178, 1.12827701, 0.915119477, 0.704096256, 0.495114416, 0.28808103, 0.082903167, -0.120512101, -0.322257704, -0.522426569, -0.721111627, -0.918405806, -1.11440204, -1.30919324, -1.50287236, -1.69553231, -1.88726603, -2.07816645, -2.26832649, -2.46675173, -2.66889131, -2.87008725, -3.07014324, -3.26886298, -3.46605016, -3.66150849, -3.85504166, -4.04645336, -4.23554729, -4.42212715, -4.60599664, -4.78695944, -4.96481926, -5.1393798, -5.31044474, -5.47781779, -5.64724325, -5.81245543, -5.97312924, -6.1289396, -6.27956143, -6.42466963, -6.56393914, -6.69704485, -6.82366169, -6.94346458, -7.05612842, -7.16132814, -7.25873865, -7.34803487, -7.4288917, -7.50098408, -7.57203757, -7.64958448, -7.71669581, -7.7725246, -7.81622386, -7.84694662, -7.86384591, -7.86607474, -7.85278613, -7.82313313, -7.77626873, -7.71134598, -7.62751789, -7.52393749, -7.3997578, -7.25413184, -7.08621264, -6.52075691, -5.75199076, -4.97864023, -4.21200823, -3.46339767, -2.74411146, -2.06545248, -1.43872365, -0.87522788, -0.386268061, 0.0168528993, 0.322832096, 0.520366626, 0.598153585, 0.54489007, 0.349273176, 0, -0.739522595, -1.70000313, -2.87771936, -4.26894903, -5.8699699, -7.6770597, -9.6864962, -11.8945571, -14.2975203, -16.8916633, -19.6732641, -22.6386003, -25.7839497, -29.1055901, -32.5997991, -36.2628546, -39.990671, -43.7651885, -47.8096969, -52.0989676, -56.607772, -61.3108813, -66.183067, -71.1991003, -76.3337526, -81.5617952, -86.8579994, -92.1971367, -97.5539783, -103.222807, -108.741729, -113.989662, -118.845522, -124.069019, -129.098675, -133.138135, -135.802079, -136.705187, -135.462141, -131.68762, -127.669997, -119.536527, -106.378779, -87.2883194, -61.3567148, -35.4209481, 1.10918341, 51.0784667, 117.331689, 202.713637, 309.543796, 441.83429, 594.524505, 769.892139, 985.19728, 1250.97323, 1577.75329, 1972.43926, 2450.02129, 3029.45112, 3717.06151, 4544.83778, 5528.06673, 6691.34753, 8075.05867, 9712.33566, 11610.8662, 13797.1883, 16391.1485, 19459.8125, 23017.7456, 27192.5913, 32069.589, 37733.9777, 44270.9966, 54787.8419, 65592.3065, 76014.1406, 85383.0942, 93028.9171, 96089.4698, 97547.6517, 97672.5939, 97169.7337, 96135.4981, 94484.5193, 92322.3139, 89754.3983, 86886.289, 83823.5024, 80662.1897, 77915.9441, 75296.253, 72514.6042, 69282.4849, 65311.3828, 60312.7854, 53928.6076, 45748.7083},
        {5.0255877, 4.7432161, 4.46546207, 4.19220249, 3.92331423, 3.65867417, 3.39815918, 3.14164615, 2.88901194, 2.64013345, 2.39488753, 2.15315107, 1.91480095, 1.67971405, 1.44776723, 1.21883738, 0.992801376, 0.76953609, 0.548918402, 0.330825188, 0.115133324, -0.0982803114, -0.309538843, -0.518765393, -0.726083085, -0.931615042, -1.13548439, -1.33781425, -1.53872774, -1.73834799, -1.93679812, -2.13420125, -2.33068052, -2.52635903, -2.73138684, -2.9406627, -3.14896485, -3.35609104, -3.56183899, -3.76600643, -3.96839109, -4.1687907, -4.36700299, -4.5628257, -4.75605654, -4.94649326, -5.13393359, -5.31817525, -5.49901597, -5.67625349, -5.84968553, -6.02602651, -6.19800754, -6.3652764, -6.52748089, -6.6842688, -6.83528792, -6.98018603, -7.11861093, -7.25021041, -7.37463226, -7.49152427, -7.60053423, -7.70130993, -7.79349916, -7.87674971, -7.95070937, -8.02309764, -8.10144023, -8.16873762, -8.22411435, -8.26669496, -8.29560398, -8.30996595, -8.30890539, -8.29154684, -8.25701484, -8.20443393, -8.13292863, -8.04162347, -7.92964301, -7.79611176, -7.64015426, -7.46089505, -6.87665584, -6.08497226, -5.28795687, -4.497092, -3.72385995, -2.97974306, -2.27622363, -1.62478399, -1.03690645, -0.524073334, -0.0977669551, 0.230530367, 0.449336315, 0.54716857, 0.512544814, 0.33398273, 0, -0.723224324, -1.67608447, -2.85377393, -4.25148623, -5.86441487, -7.68775335, -9.71669518, -11.9464339, -14.3721629, -16.9890758, -19.7923661, -22.7772273, -25.9388528, -29.2724362, -32.7731711, -36.4362508, -40.1971925, -44.0426645, -48.1028426, -52.3601327, -56.7969403, -61.3956713, -66.1387313, -71.0085259, -75.9874609, -81.057942, -86.2023749, -91.4031652, -96.6427187, -102.316709, -107.842743, -113.079244, -117.884639, -122.823513, -127.40121, -130.934768, -133.070761, -133.455768, -131.736366, -127.559131, -123.650617, -115.64004, -102.571374, -83.4885923, -57.4356673, -30.2770699, 7.46884888, 58.4632404, 125.367256, 210.842047, 316.584811, 447.398819, 598.085322, 771.003382, 983.942587, 1247.76453, 1573.33082, 1971.26022, 2452.71233, 3029.11165, 3698.80878, 4500.30708, 5450.53877, 6576.03089, 7919.56843, 9516.47374, 11384.2907, 13553.7324, 16120.6995, 19144.2191, 22598.348, 26640.1554, 31358.0809, 36840.5641, 43176.0446, 53388.7199, 63897.3321, 74056.3813, 83220.3676, 90743.7909, 93907.5984, 95522.2118, 95830.7968, 95395.1183, 94332.2297, 92626.4347, 90401.4244, 87780.8895, 84888.5211, 81848.0101, 79682.7146, 77498.5441, 75087.47, 72241.4641, 68752.4976, 64412.5423, 59013.5695, 52137.2562, 43287.5264},
        {5.16508465, 4.86447885, 4.56901548, 4.27855944, 3.9929756, 3.71212885, 3.43588409, 3.16410618, 2.89666003, 2.63341051, 2.3742225, 2.11896091, 1.8674906, 1.61967646, 1.37538339, 1.13447626, 0.896819962, 0.662279379, 0.430719396, 0.202004899, -0.0239992267, -0.247428096, -0.468416824, -0.687100525, -0.903614314, -1.11809331, -1.33067262, -1.54148736, -1.75067265, -1.9583636, -2.16469533, -2.36980295, -2.57382157, -2.77688632, -2.98925236, -3.20578191, -3.42121558, -3.63536004, -3.84802202, -4.0590082, -4.26812528, -4.47517997, -4.67997896, -4.88232896, -5.08203666, -5.27890876, -5.47275196, -5.66337296, -5.85057847, -6.03417517, -6.21396978, -6.39910181, -6.57984283, -6.75579722, -6.92656934, -7.0917636, -7.25098436, -7.40383602, -7.54992295, -7.68884953, -7.82022015, -7.94363918, -8.05871102, -8.16504004, -8.26223063, -8.34988716, -8.42761402, -8.50280401, -8.58266292, -8.65073212, -8.7061111, -8.74789934, -8.77519634, -8.78710159, -8.78271458, -8.7611348, -8.72146174, -8.66279488, -8.58423373, -8.48487778, -8.3638265, -8.2201794, -8.05303597, -7.86149568, -7.25743393, -6.44170698, -5.61982527, -4.80345447, -4.00426021, -3.23390817, -2.504064, -1.82639336, -1.21256191, -0.674235302, -0.223079198, 0.129240744, 0.371058867, 0.490709512, 0.476527021, 0.316845736, 0, -0.746682769, -1.72350792, -2.92517065, -4.34636617, -5.9817897, -7.82613644, -9.8741016, -12.1203804, -14.559668, -17.1866597, -19.9960506, -22.982536, -26.1408111, -29.465571, -32.951511, -36.5933264, -40.3321787, -44.1551157, -48.1779201, -52.3838157, -56.7560262, -61.2777754, -65.932287, -70.7027847, -75.5724924, -80.5246338, -85.5424327, -90.6091127, -95.7078976, -101.267073, -106.661612, -111.741219, -116.355599, -120.968263, -125.122015, -128.195973, -129.855703, -129.766768, -127.594731, -123.005156, -118.952035, -110.811677, -97.6062566, -78.3579499, -52.0889319, -23.9242052, 14.742589, 66.4149824, 133.596507, 218.790693, 323.116938, 452.154629, 601.062077, 772.346907, 983.52159, 1245.36173, 1568.64292, 1965.59083, 2445.20144, 3014.88882, 3668.25641, 4447.90562, 5373.6668, 6473.20389, 7787.45447, 9351.26328, 11184.3917, 13317.7741, 15837.6502, 18798.9759, 22167.6365, 26101.3617, 30685.9394, 36007.1578, 42150.8048, 52007.3886, 62156.797, 71983.6382, 80872.52, 88208.0504, 91419.7746, 93150.7387, 93623.4348, 93318.721, 92356.8196, 90750.3004, 88624.7682, 86105.8281, 83319.085, 80390.1437, 78996.4827, 77162.4605, 74787.1181, 71769.4968, 68008.6376, 63403.5817, 57853.3701, 50543.7532, 40854.3897},
        {5.322536, 5.00793031, 4.69851028, 4.39414739, 4.09471307, 3.80007879, 3.51011601, 3.22469617, 2.94369074, 2.66697117, 2.39440892, 2.12587543, 1.86124218, 1.60038061, 1.34316218, 1.08945835, 0.839140567, 0.592080292, 0.348148982, 0.10721809, -0.130840926, -0.366156611, -0.598857509, -0.829072164, -1.05692912, -1.28255692, -1.50608411, -1.72763924, -1.94735084, -2.16534747, -2.38175766, -2.59670996, -2.81033292, -3.02275507, -3.24327359, -3.46723985, -3.68989505, -3.91107018, -4.13059625, -4.34830427, -4.56402524, -4.77759017, -4.98883006, -5.19757591, -5.40365874, -5.60690954, -5.80715932, -6.00423908, -6.19797984, -6.38821259, -6.57476833, -6.77072205, -6.96237366, -7.14926709, -7.33094622, -7.50695497, -7.67683723, -7.8401369, -7.99639789, -8.1451641, -8.28597942, -8.41838777, -8.54193305, -8.65615914, -8.76060997, -8.85482942, -8.93836141, -9.01795648, -9.10019218, -9.16974921, -9.22570429, -9.26713415, -9.29311552, -9.3027251, -9.29503964, -9.26913584, -9.22409045, -9.15898017, -9.07288173, -8.96487185, -8.83402726, -8.67942469, -8.50014085, -8.29525247, -7.6700906, -6.82888672, -5.98060481, -5.1370994, -4.31022501, -3.51183616, -2.7537874, -2.04793324, -1.40612821, -0.84022684, -0.362083657, 0.016446813, 0.283510042, 0.427251502, 0.435816666, 0.297351008, 0, -0.81191665, -1.84571126, -3.09622505, -4.55829923, -6.226775, -8.09649358, -10.1622962, -12.419024, -14.8615183, -17.4846203, -20.2831711, -23.252012, -26.3859841, -29.6799288, -33.1286872, -36.7271006, -40.3833913, -44.0800266, -48.0077709, -52.1429043, -56.4617069, -60.9404587, -65.5554397, -70.2829301, -75.0992098, -79.980559, -84.9032576, -89.8435859, -94.7778237, -100.087247, -105.194641, -109.954788, -114.22247, -118.47287, -122.25057, -124.934706, -126.19394, -125.696931, -123.112339, -118.108826, -113.613435, -105.04476, -91.4339509, -71.8121577, -45.2105305, -16.331746, 22.8824539, 74.8188006, 141.864026, 226.40486, 329.097302, 456.174159, 603.644801, 774.212985, 984.25017, 1243.9913, 1563.67131, 1954.69041, 2425.85344, 2984.69403, 3623.99863, 4387.39661, 5298.78281, 6385.93268, 7683.19717, 9221.90893, 11015.3028, 13090.0195, 15539.0583, 18417.8824, 21720.7185, 25573.687, 30053.4483, 35236.6626, 41199.9903, 50646.0545, 60369.0619, 69789.1822, 78326.585, 85401.4399, 88595.2011, 90392.5639, 91001.719, 90893.9362, 90173.2705, 88834.1605, 86986.142, 84738.7508, 82201.5228, 79483.9937, 78564.457, 76836.6065, 74322.6835, 71044.9297, 67025.5863, 62286.895, 56851.0971, 49193.3266, 38510.4955},
        {5.5060844, 5.17744106, 4.85395469, 4.53550643, 4.22197737, 3.91324864, 3.60920135, 3.30971661, 3.01467555, 2.72395928, 2.4374489, 2.15502554, 1.87657032, 1.60196434, 1.33108872, 1.06382458, 0.800053031, 0.53965519, 0.28251217, 0.0285050875, -0.222484943, -0.470576806, -0.715889387, -0.958541571, -1.19865224, -1.43634029, -1.67172459, -1.90492403, -2.1360575, -2.36524388, -2.59260206, -2.81825093, -3.04230936, -3.26489624, -3.49417927, -3.72608362, -3.95643776, -4.18509935, -4.41192609, -4.63677566, -4.85950574, -5.07997403, -5.29803821, -5.51355596, -5.72638498, -5.93638293, -6.14340752, -6.34731642, -6.54796733, -6.74521792, -6.93892589, -7.14642609, -7.34972016, -7.54828692, -7.74160518, -7.92915376, -8.11041148, -8.28485714, -8.45196956, -8.61122756, -8.76210995, -8.90409555, -9.03666317, -9.15929163, -9.27145973, -9.37264631, -9.46233016, -9.54647485, -9.63088823, -9.70167483, -9.75789309, -9.79860144, -9.82285831, -9.82972214, -9.81825137, -9.78750443, -9.73653976, -9.66441579, -9.57019096, -9.4529237, -9.31167245, -9.14549565, -8.95345172, -8.73459912, -8.08821817, -7.22166515, -6.34710137, -5.47655883, -4.62206953, -3.79566547, -3.00937866, -2.27524111, -1.60528481, -1.01154177, -0.506043989, -0.100823475, 0.192087771, 0.360657745, 0.392854443, 0.276645863, 0, -0.883180442, -1.97796397, -3.27950464, -4.78295652, -6.48347366, -8.37621013, -10.45632, -12.7189573, -15.1592761, -17.7724305, -20.5535745, -23.4978622, -26.6004477, -29.856485, -33.2611281, -36.8095312, -40.3663189, -43.9079981, -47.7249681, -51.7844124, -56.0535142, -60.4994573, -65.089425, -69.7906007, -74.5701681, -79.3953105, -84.2332115, -89.0510544, -93.8160228, -98.8381833, -103.616112, -108.014129, -111.896551, -115.786143, -119.218002, -121.551637, -122.453831, -121.591371, -118.631041, -113.239625, -108.238472, -99.1797178, -85.1129497, -65.0877548, -38.1537204, -8.66703873, 30.9559593, 82.9923378, 149.719161, 233.413492, 334.266119, 459.190277, 605.275044, 775.457354, 984.613513, 1242.25929, 1557.91046, 1941.2356, 2401.56294, 2948.05403, 3574.26407, 4323.38613, 5222.82497, 6299.97929, 7582.23752, 9096.99282, 10849.4832, 12861.3578, 15235.5015, 18028.8477, 21270.8961, 25050.2344, 29433.2922, 34486.4991, 40276.2849, 49315.7457, 58612.978, 67622.7447, 75799.8086, 82598.9329, 85740.1286, 87569.4117, 88285.0104, 88381.6047, 87940.0772, 86917.7886, 85401.7975, 83479.1625, 81236.9421, 78762.1949, 78220.8836, 76517.6389, 73803.5912, 70229.8706, 65947.6073, 61107.9316, 55861.9736, 47881.6442, 36231.9124},
        {5.71771039, 5.37367601, 5.03482, 4.70103158, 4.37220001, 4.04821453, 3.72896436, 3.41433876, 3.10422696, 2.79851821, 2.49710174, 2.1998668, 1.90670262, 1.61749845, 1.33214352, 1.05052708, 0.772538362, 0.498066614, 0.227001074, -0.0407690196, -0.305354425, -0.566865902, -0.82541421, -1.08111011, -1.33406436, -1.58438772, -1.83219095, -2.07758481, -2.32068006, -2.56158746, -2.80041776, -3.03728174, -3.27229014, -3.50555373, -3.74420353, -3.9846925, -4.22338924, -4.46017668, -4.69493776, -4.92755543, -5.15791261, -5.38589226, -5.6113773, -5.83425067, -6.05439532, -6.27169419, -6.48603021, -6.69728632, -6.90534545, -7.11009056, -7.31140458, -7.53087148, -7.74620273, -7.95681083, -8.16210828, -8.36150759, -8.55442127, -8.7402618, -8.9184417, -9.08837346, -9.2494696, -9.40114261, -9.542805, -9.67386926, -9.7937479, -9.90185343, -9.99759835, -10.0860716, -10.172226, -10.2437667, -10.2997381, -10.3391848, -10.3611512, -10.3646821, -10.3488217, -10.3126148, -10.2551057, -10.175339, -10.0723592, -9.94521083, -9.79293844, -9.61458652, -9.4091996, -9.17582219, -8.50830428, -7.61681038, -6.71637512, -5.81919236, -4.93745595, -4.08335975, -3.26909764, -2.50686346, -1.80885108, -1.18725436, -0.654267161, -0.222083337, 0.0971032488, 0.291098736, 0.347709263, 0.254740971, 0, -0.950610104, -2.10225422, -3.45045586, -4.99073853, -6.71862576, -8.62964104, -10.7193079, -12.9831499, -15.4166904, -18.0154531, -20.7749614, -23.6907389, -26.758309, -29.9731953, -33.3309213, -36.8270104, -40.2774475, -43.6521076, -47.3526139, -51.3360173, -55.5593688, -59.9797192, -64.5541195, -69.2396206, -73.9932733, -78.7721287, -83.5332375, -88.2336508, -92.8304194, -97.5482096, -101.978331, -105.997551, -109.482634, -113.028469, -116.152761, -118.181717, -118.776667, -117.598942, -114.309871, -108.570785, -103.043943, -93.4764195, -78.9424069, -58.5160968, -31.2716811, -1.27641942, 38.636831, 90.6421454, 156.913599, 239.625266, 338.51226, 461.168572, 605.907322, 775.918901, 984.351459, 1239.88231, 1551.18878, 1925.56101, 2373.37875, 2906.53508, 3520.67849, 4257.16125, 5145.74274, 6212.965, 7479.9185, 8970.1961, 10681.0997, 12629.4059, 14928.2915, 17636.4041, 20819.002, 24527.1046, 28817.0708, 33745.2596, 39368.03, 48021.4448, 56912.2331, 65526.828, 73351.6624, 79873.1696, 82929.8141, 84754.6428, 85540.3607, 85831.8608, 85682.4405, 84998.6525, 83841.1314, 82270.512, 80347.429, 78132.5173, 77886.8531, 76155.787, 73214.5755, 69338.475, 64802.7421, 59882.6331, 54853.4046, 46575.4082, 34053.0515},
        {5.95939447, 5.59730007, 5.24057748, 4.8891179, 4.54281254, 4.20155261, 3.86522932, 3.53373387, 3.20695749, 2.88479137, 2.56712672, 2.25385476, 1.9448667, 1.64005374, 1.33930709, 1.04251796, 0.749577564, 0.460377109, 0.174807805, -0.107239141, -0.385872518, -0.661201117, -0.93333373, -1.20237915, -1.46844616, -1.73164355, -1.99208012, -2.24986466, -2.50510596, -2.7579128, -3.00839398, -3.2566583, -3.50281453, -3.74697147, -3.99558052, -4.24544576, -4.49329503, -4.7390313, -4.98255752, -5.22377666, -5.46259167, -5.69890553, -5.93262118, -6.16364159, -6.39186973, -6.61720855, -6.83956101, -7.05883008, -7.27491871, -7.48772987, -7.69716652, -7.92871574, -8.15614176, -8.37879293, -8.59601758, -8.80716407, -9.01158075, -9.20861596, -9.39761804, -9.57793535, -9.74891624, -9.90990904, -10.0602621, -10.1993238, -10.3264425, -10.4409664, -10.542244, -10.6344594, -10.7216805, -10.7932824, -10.8482998, -10.8857677, -10.9047209, -10.9041942, -10.8832227, -10.840841, -10.7760841, -10.6879869, -10.5755841, -10.4379107, -10.2740016, -10.0828916, -9.86361548, -9.61520823, -8.92683661, -8.01109054, -7.08548624, -6.16235958, -5.25404644, -4.37288267, -3.53120417, -2.74134679, -2.01564641, -1.3664389, -0.80606014, -0.346845992, -0.00113233181, 0.218744969, 0.30045004, 0.231647007, 0, -1.00434159, -2.20057019, -3.58452513, -5.15204574, -6.89897134, -8.82114126, -10.9143948, -13.1745714, -15.5975102, -18.1790507, -20.9150322, -23.8012939, -26.8336753, -30.0080156, -33.3201542, -36.7659304, -40.1132633, -43.3254322, -46.9138102, -50.8253968, -55.0071911, -59.4061924, -63.9694001, -68.6438133, -73.3764314, -78.1142536, -82.8042792, -87.3935075, -91.8289377, -96.2456528, -100.333604, -103.983367, -107.085513, -110.320234, -113.183294, -114.959898, -115.303734, -113.868494, -110.307866, -104.275541, -98.2466461, -88.1947349, -73.2214761, -52.4285389, -24.9175924, 5.49377592, 45.5987947, 97.4747745, 163.199026, 244.848859, 341.7246, 462.074635, 605.496152, 775.436513, 983.203849, 1236.57697, 1543.33468, 1908.00122, 2342.34964, 2861.70341, 3464.86765, 4190.00902, 5067.48555, 6122.51109, 7371.58306, 8835.19988, 10504.3193, 12391.781, 14618.7405, 17245.0839, 20365.8687, 24000.3982, 28196.3837, 33001.5364, 38463.5675, 46768.1339, 55290.5145, 63543.9343, 71041.6178, 77296.7899, 80239.5147, 82021.6179, 82834.8219, 83294.8387, 83425.5614, 83074.2199, 82273.5402, 81056.2481, 79455.0697, 77502.7306, 77483.4563, 75701.28, 72540.3709, 68384.8986, 63619.0322, 58626.9413, 53792.795, 45241.3208, 32008.3238},
        {6.22604782, 5.82836985, 5.43790477, 5.05448353, 4.67793711, 4.30809646, 3.94479254, 3.58785632, 3.23711877, 2.89241083, 2.55356348, 2.22040767, 1.89277438, 1.57049455, 1.25339915, 0.941319154, 0.634085509, 0.331529183, 0.0334811378, -0.260227665, -0.549766263, -0.835303694, -1.117009, -1.39505121, -1.66959937, -1.94082251, -2.20888968, -2.4739699, -2.73623223, -2.99584569, -3.25297933, -3.50780218, -3.76048328, -4.01119166, -4.26893348, -4.52927332, -4.7876696, -5.04400457, -5.2981605, -5.55001962, -5.79946421, -6.04637652, -6.2906388, -6.53213332, -6.77074233, -7.00634809, -7.23883285, -7.46807888, -7.69396843, -7.91638375, -8.13520711, -8.37650562, -8.61340904, -8.84523199, -9.0712891, -9.29089498, -9.50336426, -9.70801155, -9.90415147, -10.0910987, -10.2681677, -10.4346733, -10.5899299, -10.7332524, -10.8639551, -10.9813529, -11.0847602, -11.178099, -11.2651804, -11.3358171, -11.389025, -11.42382, -11.4392181, -11.4342352, -11.4078874, -11.3591906, -11.2871606, -11.1908136, -11.0691653, -10.9212319, -10.7460292, -10.5425732, -10.3098798, -10.0469651, -9.33993528, -8.40294621, -7.45511644, -6.50886166, -5.57659752, -4.67073969, -3.80370383, -2.98790562, -2.2357607, -1.55968475, -0.972093432, -0.48540241, -0.112027349, 0.135616087, 0.245112234, 0.204045426, 0, -0.920331079, -2.04551221, -3.37034974, -4.88964998, -6.59821928, -8.49086396, -10.5623904, -12.8076048, -15.2213136, -17.7983232, -20.5334397, -23.4214697, -26.4572193, -29.635495, -32.9511031, -36.3988499, -39.7351843, -42.9208617, -46.4898301, -50.3858191, -54.5525586, -58.9337782, -63.4732079, -68.1145772, -72.801616, -77.478054, -82.087621, -86.5740467, -90.8810609, -95.0916193, -98.9591762, -102.385693, -105.273133, -108.384931, -111.192311, -112.936773, -113.261838, -111.811024, -108.227852, -102.155842, -95.937992, -85.6967634, -70.5475173, -49.6056151, -21.9864176, 8.75085254, 49.0456608, 100.893611, 166.290308, 247.231355, 342.631021, 461.332323, 604.172774, 774.845279, 982.872498, 1235.15965, 1538.61196, 1894.74606, 2317.08043, 2825.01201, 3423.65183, 4142.22494, 4999.8072, 6018.17614, 7223.68718, 8640.59442, 10271.1514, 12133.9084, 14328.0842, 16907.2671, 19928.9272, 23442.1734, 27499.25, 32152.4014, 37453.8721, 45613.7354, 53986.9494, 62086.3011, 69424.5776, 75514.5661, 78248.3277, 79839.8594, 80496.5449, 80943.4452, 81183.0303, 81002.0714, 80413.8235, 79431.5415, 78068.4803, 76337.895, 76746.6115, 75131.1343, 71941.1122, 67626.1939, 62636.0283, 57420.2642, 52428.5502, 43551.4294, 30397.9594},
        {6.52818835, 6.09243813, 5.66587683, 5.24827044, 4.83938497, 4.4389864, 4.04684074, 3.66271398, 3.2863721, 2.91758112, 2.55610702, 2.2017158, 1.85417345, 1.51324597, 1.17869935, 0.850299593, 0.52781269, 0.211004637, -0.100358572, -0.40651094, -0.707686473, -1.00411918, -1.29604305, -1.58369211, -1.86730035, -2.14710178, -2.4233304, -2.69622022, -2.96600524, -3.23291947, -3.49719691, -3.75907157, -4.01877746, -4.27654857, -4.54428871, -4.81615153, -5.08615041, -5.35414065, -5.61997755, -5.8835164, -6.14461251, -6.40312116, -6.65889766, -6.9117973, -7.16167539, -7.40838721, -7.65178806, -7.89173325, -8.12807807, -8.36067781, -8.58938778, -8.84038725, -9.08663688, -9.32742132, -9.56202522, -9.78973323, -10.00983, -10.2216001, -10.4243283, -10.6172992, -10.7997974, -10.9711075, -11.1305143, -11.2773024, -11.4107563, -11.5301608, -11.6348005, -11.7282912, -11.8141447, -11.8827129, -11.9329998, -11.9640092, -11.974745, -11.9642111, -11.9314113, -11.8753496, -11.7950297, -11.6894557, -11.5576313, -11.3985604, -11.2112469, -10.9946947, -10.7479077, -10.4698897, -9.74453183, -8.78712574, -7.81795231, -6.84948664, -5.89420379, -4.96457885, -4.07308691, -3.23220304, -2.45440232, -1.75215984, -1.13795068, -0.624249924, -0.223532648, 0.0517260646, 0.189051131, 0.17596747, 0, -0.826605159, -1.87193832, -3.12974477, -4.59376979, -6.25775864, -8.11545662, -10.160609, -12.3869611, -14.7882581, -17.3582454, -20.0906682, -22.9792717, -26.0178014, -29.2000024, -32.5196201, -35.9703997, -39.3028134, -42.4758531, -46.0335248, -49.9174438, -54.0692253, -58.4304848, -62.9428376, -67.5478989, -72.1872842, -76.8026087, -81.3354878, -85.7275367, -89.9203709, -93.9540837, -97.6357037, -100.877303, -103.590953, -106.614596, -109.397368, -111.141508, -111.481329, -110.05114, -106.485254, -100.417982, -94.0708794, -83.7035907, -68.4442271, -47.4208997, -19.7617198, 11.2236162, 51.6339788, 103.386653, 168.398924, 248.588077, 342.522479, 459.561456, 601.678281, 772.781006, 980.727846, 1231.60347, 1531.49256, 1879.05899, 2289.49477, 2786.08729, 3380.30695, 4092.1403, 4928.40147, 5907.31847, 7066.45954, 8434.51542, 10026.675, 11868.1286, 14033.5733, 16569.7021, 19488.9646, 22877.566, 26794.0531, 31296.9728, 36444.8718, 44499.2981, 52765.047, 60749.9153, 67961.6994, 73908.1958, 76423.0413, 77804.2988, 78277.2618, 78679.5467, 78994.0694, 78948.6113, 78538.845, 77760.4434, 76609.0788, 75080.4241, 75821.0349, 74363.229, 71189.7092, 66783.1788, 61626.3405, 56201.8974, 50992.5524, 41828.2799, 28935.3497},
        {6.86966265, 6.39560979, 5.93262151, 5.48040307, 5.03865975, 4.60709681, 4.18541951, 3.77333312, 3.37054289, 2.9767541, 2.59167201, 2.21500188, 1.84644897, 1.48571856, 1.1325159, 0.78654626, 0.447514902, 0.115127092, -0.210911907, -0.530896831, -0.845122415, -1.15388339, -1.45747451, -1.75619049, -2.05032607, -2.340176, -2.626035, -2.90819781, -3.18695917, -3.46261381, -3.73545647, -4.00578188, -4.27388479, -4.54005992, -4.81902618, -5.10356282, -5.38627575, -5.6669916, -5.94553702, -6.22173865, -6.49542314, -6.76641713, -7.03454727, -7.2996402, -7.56152256, -7.820021, -8.07496216, -8.32617268, -8.57347922, -8.81670841, -9.0556869, -9.31660036, -9.572345, -9.82217605, -10.0653487, -10.3011183, -10.5287399, -10.7474689, -10.9565605, -11.1552698, -11.3428522, -11.5185628, -11.681657, -11.8313898, -11.9670166, -12.0877926, -12.1929731, -12.2857186, -12.369096, -12.434306, -12.4803506, -12.5062318, -12.5109518, -12.4935126, -12.4529163, -12.3881648, -12.2982604, -12.1822049, -12.0390006, -11.8676494, -11.6671534, -11.4365147, -11.1747353, -10.8808173, -10.1372109, -9.15995579, -8.17009806, -7.18015771, -6.20265472, -5.25010903, -4.33504064, -3.46996949, -2.66741556, -1.93989881, -1.29993922, -0.760056745, -0.332771356, -0.0306030197, 0.133928296, 0.148302625, 0, -0.743439871, -1.7171089, -2.9138658, -4.32656929, -5.94807808, -7.7712509, -9.78894645, -11.9940235, -14.3793406, -16.9377567, -19.6621304, -22.5453203, -25.5801854, -28.7595841, -32.0763753, -35.5234178, -38.8482905, -42.0088124, -45.5526411, -49.4200754, -53.5514142, -57.8869563, -62.3670005, -66.9318455, -71.5217903, -76.0771336, -80.5381741, -84.8452108, -88.9385425, -92.8263234, -96.3561165, -99.4478639, -102.021508, -104.970413, -107.727811, -109.467352, -109.816952, -108.404527, -104.857991, -98.805261, -92.3672644, -81.9201604, -66.6041008, -45.5592377, -17.9257228, 13.2378943, 53.6894589, 105.268418, 169.814221, 249.166313, 341.594315, 456.870625, 597.920541, 768.819612, 975.991829, 1224.80678, 1520.63406, 1859.729, 2258.64686, 2743.88572, 3332.44788, 4036.0245, 4850.10181, 5789.68768, 6902.86839, 8222.54669, 9776.18548, 11595.2772, 13731.2603, 16224.2909, 19042.432, 22309.0234, 26089.1762, 30448.0015, 35450.6102, 43415.8296, 51591.6256, 59479.6802, 66581.6754, 72399.2933, 74705.9134, 75883.722, 76175.1022, 76520.8695, 76886.9444, 76948.2334, 76685.2076, 76078.3378, 75108.0951, 73754.9502, 74676.6974, 73340.7221, 70224.5961, 65805.8909, 60562.1781, 54971.0293, 49510.016, 40116.3546, 27584.7374},
        {7.25431735, 6.74398972, 6.24626665, 5.76080586, 5.28726508, 4.82530203, 4.37457444, 3.93474003, 3.50545652, 3.08638164, 2.67717312, 2.27748867, 1.88698602, 1.5053229, 1.13215703, 0.767146138, 0.409947943, 0.0602201716, -0.282379452, -0.618193203, -0.947563357, -1.27083219, -1.58834198, -1.900435, -2.20745352, -2.50973983, -2.80763619, -3.10148489, -3.39162819, -3.67840838, -3.96216773, -4.24324851, -4.52199301, -4.79874349, -5.09052586, -5.38898966, -5.68558391, -5.98010948, -6.27236722, -6.562158, -6.84928267, -7.1335421, -7.41473714, -7.69266866, -7.96713751, -8.23794456, -8.50489066, -8.76777667, -9.02640346, -9.28057188, -9.5300828, -9.80138469, -10.0670531, -10.3263113, -10.5783824, -10.8224896, -11.0578562, -11.2837052, -11.4992599, -11.7037436, -11.8963793, -12.0763903, -12.2429997, -12.3954309, -12.5329068, -12.6546508, -12.7598861, -12.8510637, -12.9305572, -12.9909325, -13.0312036, -13.0503844, -13.0474887, -13.0215307, -12.9715242, -12.8964831, -12.7954214, -12.667353, -12.5112919, -12.326252, -12.1112472, -11.8652915, -11.5873989, -11.2765832, -10.5145573, -9.51776301, -8.50765788, -7.49679808, -6.49773977, -5.52303911, -4.58525227, -3.69693541, -2.8706447, -2.1189363, -1.45436637, -0.889491084, -0.436866601, -0.109049086, 0.0814052949, 0.121940379, 0, -0.691111252, -1.61828432, -2.77386838, -4.15021263, -5.73966625, -7.53457842, -9.52729835, -11.7101752, -14.0755582, -16.6157965, -19.3232392, -22.1902357, -25.209135, -28.3722864, -31.672039, -35.100742, -38.4037559, -41.5381456, -45.0549255, -48.8935187, -52.9933481, -57.2938368, -61.7344078, -66.254484, -70.7934885, -75.2908444, -79.6859746, -83.9183022, -87.9272501, -91.7016157, -95.1133438, -98.0870452, -100.54733, -103.413567, -106.112986, -107.807554, -108.123453, -106.686868, -103.123982, -97.0609794, -90.5491027, -80.0514156, -64.719634, -43.7054737, -16.1606507, 15.119514, 55.5378111, 106.853426, 170.825545, 249.213352, 340.041867, 453.368424, 592.807419, 762.537014, 967.886386, 1213.66793, 1504.69404, 1835.54508, 2223.59093, 2697.36373, 3277.68946, 3970.14696, 4761.74168, 5665.03337, 6735.882, 8010.27203, 9524.97804, 11316.1893, 13417.1981, 15862.9352, 18585.7802, 21738.9927, 25393.0023, 29618.2385, 34485.131, 42354.3378, 50433.5029, 58220.4988, 65213.1979, 70909.4729, 73039.2024, 74046.9151, 74188.1952, 74485.1395, 74889.9208, 75035.3316, 74889.5137, 74420.6088, 73596.7588, 72386.1053, 73283.5695, 72006.7719, 68984.2067, 64644.3683, 59415.7507, 53726.8483, 48006.1551, 38460.136, 26310.3654},
        {7.68599905, 7.14368282, 6.61494011, 6.09940325, 5.59670458, 5.10647643, 4.62835113, 4.161961, 3.70693839, 3.26291561, 2.82952501, 2.40639892, 1.99316966, 1.58946957, 1.19493098, 0.809186211, 0.431867609, 0.062607498, -0.298961789, -0.653207921, -1.00049857, -1.3412014, -1.67568408, -2.00431428, -2.32745967, -2.64548791, -2.95876669, -3.26766366, -3.57254649, -3.87378285, -4.17174042, -4.46678686, -4.75928983, -5.04961701, -5.3561677, -5.66991447, -5.98161319, -6.29104637, -6.5979965, -6.9022461, -7.20357768, -7.50177374, -7.7966168, -8.08788935, -8.37537391, -8.65885299, -8.93810909, -9.21292471, -9.48308238, -9.74836459, -10.0085539, -10.29098, -10.567281, -10.8366422, -11.098249, -11.3512867, -11.5949407, -11.8283962, -12.0508386, -12.2614534, -12.4594257, -12.6439409, -12.8141844, -12.9693415, -13.1085975, -13.2311379, -13.3361478, -13.4250092, -13.4990511, -13.5529288, -13.5856852, -13.5963632, -13.5840059, -13.5476561, -13.4863569, -13.3991511, -13.2850817, -13.1431916, -12.9725239, -12.7721213, -12.541027, -12.2782839, -11.9829348, -11.6540227, -10.8731556, -9.85687404, -8.82673599, -7.79533095, -6.77524843, -5.77907797, -4.81940907, -3.90883125, -3.05993403, -2.28530693, -1.59753947, -1.00922115, -0.532941511, -0.181290057, 0.0331436918, 0.0977702165, 0, -0.68989534, -1.61272495, -2.76090807, -4.12686394, -5.70301178, -7.48177084, -9.45556036, -11.6167996, -13.9579077, -16.471304, -19.1494077, -21.984638, -24.9694142, -28.0961555, -31.3572811, -34.7452103, -38.0013495, -41.0822589, -44.5481249, -48.3375784, -52.3892499, -56.6417704, -61.0337706, -65.5038813, -69.9907332, -74.4329571, -78.7691837, -82.9380439, -86.8781684, -90.5732381, -93.9003152, -96.7845156, -99.1509547, -101.90524, -104.48224, -106.055363, -106.255575, -104.713846, -101.061144, -94.9284376, -88.3383502, -77.8022998, -62.483322, -41.5444526, -14.1487275, 17.1943026, 57.5047457, 108.456195, 171.722243, 248.976484, 338.060475, 449.163442, 586.246784, 753.509129, 955.633455, 1197.08527, 1482.3301, 1805.29621, 2183.38119, 2645.47779, 3213.64657, 3890.77705, 4660.15453, 5533.10514, 6568.46864, 7803.27524, 9278.34799, 11031.7004, 13087.4391, 15477.5372, 18115.4604, 21169.9211, 24713.9144, 28820.435, 33562.4779, 41305.8304, 49257.4969, 56917.2742, 63784.9593, 69360.3487, 71365.1663, 72262.664, 72314.6703, 72590.0827, 73031.2644, 73244.2998, 73188.366, 72822.6403, 72106.2997, 70998.5216, 71611.6219, 70304.5362, 67406.975, 63248.6487, 58159.2678, 52468.5428, 46506.1841, 36904.1066, 25076.4764},
        {8.1752324, 7.60905975, 7.05643624, 6.51700763, 5.99041969, 5.47631819, 4.9743489, 4.48415757, 4.00538999, 3.5376919, 3.08070909, 2.63408731, 2.19747233, 1.77050992, 1.35284585, 0.944125872, 0.543995765, 0.152101291, -0.231911783, -0.608397691, -0.977710667, -1.34020494, -1.69623476, -2.04615434, -2.39031792, -2.72907974, -3.06279403, -3.39181503, -3.71649696, -4.03719406, -4.35426057, -4.66805072, -4.97891874, -5.28721886, -5.61077782, -5.94084609, -6.26846414, -6.59341919, -6.91549846, -7.23448915, -7.5501785, -7.86235372, -8.17080203, -8.47531063, -8.77566676, -9.07165763, -9.36307046, -9.64969246, -9.93131085, -10.2077129, -10.4786857, -10.7749384, -11.064666, -11.3469855, -11.6210136, -11.8858674, -12.1406638, -12.3845195, -12.6165515, -12.8358767, -13.041612, -13.2328743, -13.4087804, -13.5684473, -13.7109919, -13.8355311, -13.9411816, -14.0270247, -14.092142, -14.1357246, -14.1568916, -14.1547623, -14.1284559, -14.0770916, -13.9997886, -13.8956662, -13.7638435, -13.6034399, -13.4135744, -13.1933663, -12.9419349, -12.6583994, -12.3418789, -11.9914927, -11.1901526, -10.1538358, -9.10352226, -8.0518378, -7.01140844, -5.99486012, -5.0148188, -4.08391042, -3.21476095, -2.41999633, -1.71224252, -1.10412547, -0.608271134, -0.237305467, -0.00385442012, 0.0794560533, 0, -0.886655444, -1.96779641, -3.23822817, -4.69275602, -6.32618523, -8.13332108, -10.1089688, -12.2479338, -14.5450213, -16.9950365, -19.5927848, -22.3330713, -25.2107015, -28.2204806, -31.3572138, -34.6157065, -37.7167606, -40.616038, -43.9281135, -47.5890773, -51.5350198, -55.7020311, -60.0262017, -64.4436217, -68.8903815, -73.3025713, -77.6162813, -81.767602, -85.6926235, -89.3976767, -92.7228564, -95.5831808, -97.8936681, -100.397451, -102.59549, -103.767916, -103.581313, -101.702264, -97.7973526, -91.5331635, -85.3716527, -75.3332657, -60.5376659, -40.1045161, -13.1534796, 18.9144311, 59.8312398, 111.047621, 174.014248, 250.181797, 336.943381, 444.766474, 577.195709, 738.318902, 933.33287, 1166.93651, 1443.82872, 1760.35513, 2131.46638, 2581.22576, 3124.43332, 3772.7824, 4522.30726, 5387.20316, 6412.43809, 7628.85484, 9062.95592, 10744.4588, 12718.9956, 15023.1961, 17610.5562, 20613.6295, 24097.8232, 28128.5441, 32771.1994, 40271.8401, 47970.0681, 55386.1295, 62040.2702, 67452.7363, 69436.7472, 70357.5932, 70467.103, 70812.4608, 71330.6671, 71627.3247, 71656.0043, 71370.2763, 70723.7114, 69669.8802, 69390.159, 67828.7112, 65106.806, 61345.7128, 56666.7009, 51191.0396, 45039.9982, 35625.5853, 23615.4393},
        {8.71411517, 8.13168319, 7.56187632, 7.00438797, 6.45891151, 5.92514035, 5.40276788, 4.89148749, 4.39099257, 3.90097652, 3.42113272, 2.95115457, 2.49073547, 2.0395688, 1.59734795, 1.16376633, 0.738517319, 0.321294313, -0.0882092952, -0.490300114, -0.88528475, -1.27346981, -1.65516191, -2.03066764, -2.40029362, -2.76434646, -3.12313276, -3.47695913, -3.82613217, -4.1709585, -4.51174473, -4.84879745, -5.18242329, -5.51292883, -5.85622924, -6.20449855, -6.54972318, -6.89170321, -7.23023874, -7.56512983, -7.89617656, -8.22317903, -8.5459373, -8.86425146, -9.17792159, -9.48674777, -9.79053007, -10.0890686, -10.3821634, -10.6696146, -10.9512222, -11.2612652, -11.5643174, -11.8594314, -12.1456599, -12.4220555, -12.6876708, -12.9415586, -13.1827714, -13.410362, -13.6233828, -13.8208867, -14.0019262, -14.165554, -14.3108228, -14.4367851, -14.5424937, -14.6247403, -14.6803711, -14.7131015, -14.7221307, -14.7066579, -14.6658822, -14.599003, -14.5052194, -14.3837306, -14.2337358, -14.0544342, -13.8450251, -13.6047076, -13.3326809, -13.0281443, -12.6902969, -12.318338, -11.4963307, -10.439774, -9.36915669, -8.29714988, -7.23642476, -6.19965248, -5.19950423, -4.24865116, -3.35976446, -2.54551528, -1.81857479, -1.19161418, -0.677304595, -0.288317217, -0.0373232129, 0.0630062492, 0, -1.07895914, -2.31592862, -3.70791972, -5.25194374, -6.94501194, -8.78413562, -10.7663261, -12.8885945, -15.1479523, -17.5414107, -20.065981, -22.7186745, -25.4965024, -28.3964761, -31.4156068, -34.5509058, -37.5055122, -40.2344556, -43.3994581, -46.9345584, -50.7737952, -54.851207, -59.1008327, -63.4567108, -67.8528801, -72.2233793, -76.502247, -80.6235218, -84.5212425, -88.2329365, -91.5512079, -94.3790487, -96.6194509, -98.8675085, -100.680163, -101.429794, -100.811764, -98.521434, -94.254166, -87.7053214, -81.5210544, -71.5478665, -56.9037901, -36.7068579, -10.0751024, 22.1427743, 62.9920766, 113.78744, 175.843498, 250.474887, 334.954355, 439.578506, 567.234816, 721.80787, 909.218027, 1134.47162, 1402.575, 1712.55847, 2076.76258, 2514.04717, 3031.4168, 3650.08516, 4379.98508, 5238.42369, 6255.20347, 7454.39131, 8846.50855, 10452.1103, 12341.4187, 14556.5615, 17098.0733, 20056.0738, 23487.0883, 27447.6421, 31994.2605, 39224.8578, 46644.223, 53798.5343, 60233.9696, 65496.7071, 67529.2722, 68550.5972, 68794.14, 69242.0694, 69831.4784, 70187.4966, 70262.8152, 70010.125, 69382.1171, 68331.4824, 67012.5624, 65135.5138, 62593.0862, 59278.0291, 55083.0921, 49901.0246, 43624.5762, 34435.3514, 22299.4388},
        {9.30106999, 8.71110479, 8.13184285, 7.56306254, 7.00454226, 6.4560604, 5.91739534, 5.38832548, 4.86862921, 4.35808492, 3.856471, 3.36356583, 2.87914781, 2.40299533, 1.93488677, 1.47460053, 1.021915, 0.576608558, 0.138459607, -0.292753468, -0.717252277, -1.13525843, -1.54699354, -1.95267921, -2.35253706, -2.74678869, -3.13565571, -3.51935974, -3.89812239, -4.27216526, -4.64170997, -5.00697812, -5.36819132, -5.7255712, -6.09182345, -6.46066505, -6.82566293, -7.18663356, -7.54339342, -7.89575899, -8.24354673, -8.58657313, -8.92465467, -9.25760782, -9.58524907, -9.90739487, -10.2238617, -10.5344661, -10.8390245, -11.1373533, -11.4292691, -11.7519135, -12.0669686, -12.3734418, -12.6703405, -12.956672, -13.2314436, -13.4936627, -13.7423366, -13.9764727, -14.1950782, -14.3971606, -14.5817272, -14.7477853, -14.8943423, -15.0204054, -15.1249821, -15.2031099, -15.2499214, -15.2726116, -15.2704453, -15.242687, -15.1886014, -15.1074534, -14.9985075, -14.8610285, -14.694281, -14.4975297, -14.2700393, -14.0110746, -13.7199002, -13.3957807, -13.037981, -12.6457656, -11.8034223, -10.7264327, -9.63526624, -8.54265439, -7.46132866, -6.40402054, -5.3834615, -4.41238306, -3.5035167, -2.66959391, -1.9233462, -1.27750505, -0.744801953, -0.337968406, -0.0697359035, 0.0471640631, 0, -1.18244192, -2.50424184, -3.96336502, -5.55777674, -7.28544227, -9.14432688, -11.1323958, -13.2476144, -15.4879478, -17.8513614, -20.3358205, -22.9392902, -25.6597359, -28.4951228, -31.4434162, -34.5025813, -37.3688256, -39.997006, -43.0676355, -46.5140169, -50.269453, -54.2672468, -58.4407011, -62.7231187, -67.0478026, -71.3480555, -75.5571803, -79.6084799, -83.4352571, -87.1068569, -90.370658, -93.1191506, -95.2448249, -97.2816653, -98.8014149, -99.2047519, -98.1917181, -95.4623555, -90.716706, -83.6548117, -76.4983437, -65.6582624, -50.3412475, -29.7539788, -3.10313595, 28.0752403, 67.4861715, 116.505318, 176.508341, 248.8709, 331.378381, 433.384833, 557.073361, 706.036912, 886.746862, 1104.38223, 1364.12205, 1666.78908, 2022.90865, 2447.75834, 2943.26717, 3536.51415, 4246.37902, 5091.95644, 6092.70529, 7267.91727, 8616.95783, 10153.9526, 11966.0902, 14099.2304, 16588.4821, 19492.1106, 22859.6976, 26740.8246, 31185.0733, 38145.4249, 45292.2115, 52199.1649, 58440.0168, 63588.4989, 65737.7886, 66929.2083, 67371.0129, 67939.746, 68579.1318, 68954.4409, 69023.32, 68743.416, 68072.3757, 66967.8459, 64642.5781, 62447.5279, 60064.3764, 57174.8046, 53460.4935, 48603.1243, 42284.3779, 33268.7155, 21283.5686},
        {9.93030726, 9.34637536, 8.76981201, 8.20053096, 7.63844596, 7.08347076, 6.53551909, 5.99450472, 5.46034139, 4.93294285, 4.41222284, 3.89809511, 3.39047341, 2.88927149, 2.39440309, 1.90578197, 1.42332186, 0.946936524, 0.476539702, 0.012045143, -0.446633405, -0.899582195, -1.34688748, -1.78863551, -2.22491254, -2.65580482, -3.0813986, -3.50178014, -3.91703569, -4.32725149, -4.73251381, -5.1329089, -5.528523, -5.91944237, -6.31398353, -6.70794447, -7.0970243, -7.48104217, -7.85981726, -8.23316871, -8.60091571, -8.96287741, -9.31887297, -9.66872157, -10.0122424, -10.3492545, -10.6795772, -11.0030296, -11.3194308, -11.6286001, -11.9303565, -12.2613304, -12.5837318, -12.896582, -13.1989019, -13.489713, -13.7680362, -14.0328928, -14.2833039, -14.5182908, -14.7368746, -14.9380764, -15.1209175, -15.284419, -15.427602, -15.5494878, -15.6490976, -15.7214298, -15.7615795, -15.7768644, -15.7665664, -15.7299674, -15.6663494, -15.5749942, -15.4551839, -15.3062004, -15.1273257, -14.9178415, -14.67703, -14.4041731, -14.0985527, -13.7594507, -13.386149, -12.9779298, -12.1166125, -11.0189912, -9.90687707, -8.79309959, -7.69048814, -6.6118721, -5.57008088, -4.57794387, -3.64829045, -2.79395003, -2.027752, -1.36252575, -0.811100675, -0.386306168, -0.100971625, 0.0320735624, 0, -0.979268786, -2.13933242, -3.47497549, -4.98098257, -6.65213826, -8.48322714, -10.4690338, -12.6043428, -14.8839387, -17.3026062, -19.8551298, -22.536294, -25.3408836, -28.263683, -31.2994769, -34.4430498, -37.4113098, -40.1593449, -43.3104698, -46.7999241, -50.5629475, -54.5347796, -58.65066, -62.8458283, -67.0555241, -71.2149872, -75.259457, -79.1241732, -82.7443754, -86.2080222, -89.2416373, -91.7346447, -93.5764684, -95.5421671, -97.0783473, -97.3954465, -96.1171982, -92.8673358, -87.269593, -78.9477033, -67.7337625, -52.9797272, -34.2685646, -11.1832414, 16.6932751, 43.1756364, 76.9338526, 120.035552, 174.548363, 242.539914, 324.26057, 425.816323, 548.876901, 696.54262, 874.834573, 1088.46248, 1342.13607, 1634.6731, 1978.01434, 2390.52816, 2879.80918, 3464.58623, 4153.72604, 4962.98922, 5919.8177, 7046.29148, 8349.0698, 9846.23471, 11612.411, 13690.2397, 16100.6599, 18911.3533, 22172.473, 25934.1723, 30246.6045, 36951.0784, 43858.8025, 50572.1412, 56693.4587, 61825.1193, 64248.0738, 65767.0422, 66550.7089, 67281.3637, 67924.3889, 68231.1641, 68177.7717, 67740.2938, 66894.8128, 65617.4108, 62545.8695, 60101.3145, 57822.2355, 55247.1225, 51914.4653, 47362.7537, 41130.4777, 31939.5805, 20943.2408},
        {10.6122373, 10.0427914, 9.47637143, 8.91306551, 8.35296184, 7.79614861, 7.242714, 6.69274619, 6.14633339, 5.60356376, 5.06452549, 4.52930678, 3.99799581, 3.47068076, 2.94744983, 2.42839119, 1.91359303, 1.40314355, 0.897130916, 0.395643327, -0.101231035, -0.593403983, -1.08078733, -1.56329289, -2.04083248, -2.51331791, -2.980661, -3.44277355, -3.89956739, -4.35095433, -4.79684618, -5.23715475, -5.67179186, -6.10066932, -6.52608816, -6.94671534, -7.36118908, -7.76934368, -8.1710134, -8.56603253, -8.95423535, -9.33545613, -9.70952916, -10.0762887, -10.4355691, -10.7872045, -11.1310293, -11.4668778, -11.7945841, -12.1139827, -12.4249078, -12.7634965, -13.0924934, -13.4109456, -13.7179004, -14.0124052, -14.2935073, -14.560254, -14.8116926, -15.0468703, -15.2648345, -15.4646325, -15.6453116, -15.805919, -15.9455022, -16.0631083, -16.1577847, -16.225018, -16.2603802, -16.2702623, -16.2539425, -16.210699, -16.1398099, -16.0405532, -15.9122072, -15.7540499, -15.5653595, -15.3454141, -15.0934919, -14.808871, -14.4908295, -14.1386456, -13.7515973, -13.3289629, -12.4490292, -11.330425, -10.196839, -9.06121118, -7.93648178, -6.83559086, -5.77147851, -4.75708485, -3.80534999, -2.92921401, -2.14161703, -1.45549916, -0.883800494, -0.439461141, -0.135421204, 0.0153792108, 0, -0.715774265, -1.665413, -2.83969371, -4.22939393, -5.82529118, -7.61816296, -9.59878681, -11.7579402, -14.0864008, -16.5749459, -19.2143532, -21.9954002, -24.9088643, -27.9455231, -31.0961542, -34.3515349, -37.4598858, -40.3773332, -43.643296, -47.196575, -50.9759713, -54.9202859, -58.9683198, -63.0588739, -67.1307492, -71.1227466, -74.9736673, -78.6223121, -82.0074821, -85.232861, -88.01191, -90.2339652, -91.7883627, -93.7334642, -95.3740932, -95.7033326, -94.259811, -90.5821567, -84.2089983, -74.6789642, -59.2917917, -40.5064019, -18.2994673, 7.35233982, 36.4723471, 57.970928, 85.7626029, 122.648938, 171.4315, 234.911854, 315.912821, 417.188737, 539.995253, 686.949573, 863.448734, 1073.64168, 1321.67737, 1603.79553, 1933.85188, 2334.33034, 2820.38638, 3399.70088, 4066.9849, 4833.69289, 5739.65002, 6811.65885, 8068.58881, 9535.18574, 11265.2835, 13296.262, 15618.8163, 18324.1938, 21465.9347, 25097.5792, 29272.6676, 35750.0049, 42451.5503, 49004.5278, 55036.1615, 60173.6753, 62870.6759, 64710.4162, 65823.0994, 66696.2488, 67318.628, 67530.1171, 67326.9236, 66705.2555, 65661.3203, 64191.3259, 60550.4122, 57898.1004, 55678.8864, 53337.2662, 50317.7356, 46064.7906, 40022.9273, 30694.295, 20874.4194},
        {11.3515349, 10.8034383, 10.2531314, 9.70091336, 9.14708319, 8.59193989, 8.03578247, 7.47890997, 6.92162139, 6.36421575, 5.80699208, 5.25024938, 4.69428668, 4.13940299, 3.58589734, 3.03406873, 2.48421619, 1.93663874, 1.39163538, 0.849505153, 0.31054706, -0.224939878, -0.756656643, -1.28430422, -1.80758358, -2.32619572, -2.83984161, -3.34822225, -3.8510386, -4.34799166, -4.8387824, -5.3231118, -5.80068086, -6.27119055, -6.7290535, -7.17672611, -7.61672832, -8.04893275, -8.47321197, -8.88943861, -9.29748524, -9.69722449, -10.0885289, -10.4712712, -10.8453238, -11.2105595, -11.5668507, -11.9140702, -12.2520905, -12.5807841, -12.9000238, -13.2467191, -13.5829026, -13.9076442, -14.2200136, -14.5190803, -14.8039142, -15.073585, -15.3271623, -15.5637159, -15.7823155, -15.9820308, -16.1619315, -16.3210873, -16.4585679, -16.573443, -16.6647824, -16.7287378, -16.7615311, -16.7682499, -16.7481531, -16.7004997, -16.6245485, -16.5195583, -16.384788, -16.2194965, -16.0229426, -15.7943853, -15.5330833, -15.2382956, -14.909281, -14.5452984, -14.1456067, -13.7094647, -12.810988, -11.6709495, -10.5152318, -9.35689057, -8.20898134, -7.08455971, -5.99668127, -4.95840159, -3.98277625, -3.08286084, -2.27171095, -1.56238215, -0.967930032, -0.501410174, -0.175878161, -0.00438957522, 0, -0.473959617, -1.2301489, -2.25562838, -3.5374586, -5.0627001, -6.81841343, -8.79165911, -10.9694977, -13.3389897, -15.8871957, -18.6011762, -21.4679917, -24.4747029, -27.6083701, -30.8560541, -34.2048152, -37.4397865, -40.5191817, -43.8868024, -47.4864387, -51.2618812, -55.1569201, -59.1153457, -63.0809485, -66.9975188, -70.8088468, -74.458723, -77.8909377, -81.0492812, -84.0520437, -86.6045325, -88.5981878, -89.9244502, -91.8755, -93.6424075, -94.0427125, -92.5476334, -88.6283886, -81.7561965, -71.4022755, -52.8881451, -31.0976745, -6.31397944, 21.179824, 51.1006198, 68.6303144, 91.6545139, 123.523847, 167.588941, 227.200426, 307.136432, 407.57536, 529.483185, 675.099647, 849.265086, 1055.65223, 1297.9338, 1570.25283, 1887.97757, 2276.8723, 2758.67416, 3331.01964, 3974.93337, 4697.92752, 5552.39602, 6570.13131, 7783.21101, 9223.50148, 10921.8236, 12909.5897, 15139.4348, 17733.668, 20750.9706, 24250.024, 28289.5094, 34582.8942, 41120.3772, 47546.9432, 53507.5771, 58647.2637, 61548.9881, 63627.7347, 64983.6309, 65948.8403, 66531.6428, 66643.6366, 66297.9359, 65507.6545, 64285.9065, 62645.8059, 58639.1395, 55828.6409, 53619.7952, 51418.0878, 48629.0042, 44658.0297, 38910.6496, 29626.1056, 20949.0219},
        {12.174701, 11.6476663, 11.1127425, 10.570467, 10.0213773, 9.4660109, 8.90490527, 8.33859793, 7.76762636, 7.19252806, 6.61384051, 6.03210122, 5.44784767, 4.86161737, 4.27394779, 3.68537644, 3.09644081, 2.50767839, 1.91962668, 1.33282316, 0.747805337, 0.165110696, -0.414723268, -0.991159062, -1.56365919, -2.13168616, -2.69470249, -3.25217067, -3.80355321, -4.34831262, -4.88591142, -5.41581209, -5.93747716, -6.45036912, -6.93779294, -7.40762973, -7.86795485, -8.31875515, -8.76001748, -9.19172868, -9.6138756, -10.0264451, -10.429424, -10.8227993, -11.2065576, -11.5806859, -11.9451711, -12.2999999, -12.6451593, -12.9806361, -13.306417, -13.6671043, -14.0171024, -14.3554309, -14.6811095, -14.993158, -15.2905959, -15.572443, -15.8377189, -16.0854433, -16.314636, -16.5243165, -16.7135047, -16.88122, -17.0264823, -17.1483112, -17.2457264, -17.3145228, -17.3505729, -17.359547, -17.3406739, -17.2931823, -17.2163009, -17.1092585, -16.9712838, -16.8016054, -16.5994522, -16.3640529, -16.0946361, -15.7904307, -15.4506652, -15.0745686, -14.6613694, -14.2102964, -13.290777, -12.1259808, -10.9442435, -9.75874163, -8.58265184, -7.42915072, -6.31141488, -5.24262091, -4.23594544, -3.30456506, -2.46165638, -1.72039601, -1.09396055, -0.595526619, -0.238270811, -0.0353697357, 0, -0.537299042, -1.34476614, -2.4097395, -3.71955734, -5.26155785, -7.02307926, -8.99145976, -11.1540376, -13.4981509, -16.0111379, -18.6803369, -21.493086, -24.4367235, -27.4985875, -30.6660163, -33.9263481, -37.1019307, -40.1565322, -43.4506835, -46.9363675, -50.5655675, -54.2902665, -58.0624477, -61.8340943, -65.5571893, -69.1837159, -72.6656572, -75.9549964, -79.0037165, -81.9606054, -84.5086799, -86.5408816, -87.9501523, -89.8843821, -91.6090386, -92.0549364, -90.7385325, -87.1762843, -80.8846489, -71.3800835, -55.1372707, -35.6401998, -13.0004577, 12.6703686, 41.2606921, 60.7930347, 85.9881734, 119.700994, 164.786382, 224.099224, 300.972524, 397.198682, 513.696616, 652.822776, 819.868564, 1018.80765, 1253.61369, 1520.85718, 1833.5975, 2212.97029, 2676.04813, 3224.11793, 3839.37773, 4530.32913, 5350.04361, 6331.13434, 7510.06363, 8920.44256, 10581.7688, 12521.5233, 14661.8268, 17151.7299, 20055.1844, 23436.1422, 27358.5552, 33575.1799, 40038.9623, 46391.6531, 52275.0031, 57330.763, 60118.4554, 62083.5444, 63331.5928, 64227.1567, 64773.2805, 64865.0949, 64511.0401, 63719.5563, 62499.0837, 60858.0625, 56891.7766, 54005.3161, 51680.2059, 49397.9706, 46640.135, 42888.2238, 37623.7617, 29161.6233, 20828.6044},
        {13.0509021, 12.5520994, 12.0388537, 11.5119759, 10.9722767, 10.420567, 9.85765755, 9.2843591, 8.70148247, 8.10983842, 7.51023776, 6.90349126, 6.29040972, 5.67180391, 5.04848463, 4.42126267, 3.7909488, 3.15835382, 2.52428851, 1.88956366, 1.25499005, 0.621378474, -0.0104602835, -0.639715437, -1.2655762, -1.88723179, -2.50387141, -3.11468428, -3.71885961, -4.31558663, -4.90405453, -5.48345253, -6.05296986, -6.61179571, -7.13045063, -7.62306116, -8.10409846, -8.57368209, -9.03193161, -9.4789666, -9.91490662, -10.3398712, -10.75398, -11.1573525, -11.5501084, -11.932367, -12.3042482, -12.6658713, -13.017356, -13.3588218, -13.6903884, -14.0653469, -14.4294925, -14.7817923, -15.1212131, -15.4467218, -15.7572853, -16.0518705, -16.3294444, -16.5889738, -16.8294257, -17.049767, -17.2489645, -17.4259852, -17.5797961, -17.7093639, -17.8136557, -17.8881198, -17.9282893, -17.9403875, -17.9236095, -17.8771503, -17.8002048, -17.6919681, -17.5516351, -17.3784008, -17.1714603, -16.9300085, -16.6532405, -16.3403512, -15.9905356, -15.6029887, -15.1769056, -14.7114811, -13.7718595, -12.5833909, -11.3766422, -10.1648944, -8.96142808, -7.77952416, -6.63246337, -5.5335265, -4.49599433, -3.53314765, -2.65826723, -1.88463385, -1.2255283, -0.694231356, -0.304023805, -0.0681864256, 0, -0.660019488, -1.56545681, -2.70482842, -4.06665074, -5.63944023, -7.41171331, -9.37198644, -11.5087761, -13.8105986, -16.2659705, -18.8634082, -21.5914282, -24.4385468, -27.3932806, -30.444146, -33.5796593, -36.6551413, -39.6400803, -42.8242692, -46.1676823, -49.6302943, -53.1720796, -56.7530125, -60.3330677, -63.8722196, -67.3304426, -70.6677112, -73.8439999, -76.8192831, -79.7657111, -82.3532848, -84.478326, -86.0371564, -87.9489825, -89.5986836, -90.0983701, -89.0374981, -86.0055242, -80.5919046, -72.3860957, -59.7851806, -43.9338852, -24.6549561, -1.77114027, 24.8948157, 47.7593765, 76.700781, 113.836479, 161.283922, 221.160559, 294.578447, 385.889245, 496.081044, 627.741599, 786.724642, 977.417549, 1204.20769, 1466.8095, 1775.34533, 2145.03532, 2586.19408, 3106.52931, 3692.97536, 4356.03328, 5146.13658, 6095.3891, 7241.49943, 8618.02448, 10237.9706, 12125.9686, 14177.8746, 16566.9059, 19362.2693, 22633.1717, 26448.8199, 32629.6823, 39055.8889, 45358.8311, 51169.9005, 56120.4885, 58698.0912, 60440.5924, 61469.6246, 62233.4098, 62724.3054, 62798.5897, 62455.424, 61693.9695, 60513.3875, 58912.8394, 55133.787, 52221.0103, 49764.601, 47354.6508, 44581.2512, 41034.4941, 36304.4711, 28851.748, 20727.3195},
        {13.9701823, 13.5089191, 13.0256741, 12.5215675, 11.9977195, 11.4552505, 10.8952806, 10.3189303, 9.72731964, 9.12156903, 8.50279872, 7.87212897, 7.23068007, 6.57957228, 5.91992588, 5.25286114, 4.57949834, 3.90095774, 3.21835963, 2.53282428, 1.84547195, 1.15742293, 0.469797486, -0.216284108, -0.899701578, -1.57933465, -2.25406305, -2.92276651, -3.58432475, -4.23761749, -4.88152447, -5.51492541, -6.13670004, -6.74572809, -7.29855952, -7.81612929, -8.31988112, -8.81006843, -9.28694461, -9.75076308, -10.2017773, -10.6402405, -11.0664063, -11.480528, -11.8828591, -12.2736528, -12.6531627, -13.0216422, -13.3793447, -13.7265235, -14.063432, -14.4513401, -14.828162, -15.1928285, -15.5442703, -15.881418, -16.2032024, -16.5085542, -16.796404, -17.0656825, -17.3153205, -17.5442487, -17.7513976, -17.9356982, -18.0960809, -18.2314765, -18.3408158, -18.4194307, -18.4627397, -18.4770954, -18.4616618, -18.415603, -18.3380828, -18.2282654, -18.0853146, -17.9083944, -17.6966689, -17.449302, -17.1654577, -16.8443, -16.4849928, -16.0867001, -15.648586, -15.1698144, -14.2122818, -13.0029833, -11.7741696, -10.5391947, -9.31141284, -8.1041782, -6.93084491, -5.80476713, -4.73929901, -3.74779469, -2.84360834, -2.0400941, -1.35060613, -0.788498582, -0.367125609, -0.0998413644, 0, -0.789062652, -1.79616942, -3.01129704, -4.42442225, -6.02552181, -7.80457245, -9.75155092, -11.8564339, -14.1091983, -16.4998207, -19.0182779, -21.6545466, -24.3986036, -27.2404256, -30.1699894, -33.1772717, -36.1415048, -39.0368448, -42.1037618, -45.3085017, -48.6173103, -51.9964332, -55.4121164, -58.8306055, -62.2181464, -65.5409849, -68.7653668, -71.8575377, -74.7837436, -77.7239138, -80.3522599, -82.5709225, -84.2820423, -86.1588512, -87.7179444, -88.2702926, -87.4867092, -85.0380077, -80.5950014, -73.8285037, -65.2636692, -53.4569532, -37.9120151, -18.1325148, 6.37788831, 32.7574124, 65.7000505, 106.541673, 156.618152, 217.265357, 287.254009, 373.620313, 477.500438, 601.728156, 752.603176, 934.869072, 1153.26942, 1410.43004, 1713.69357, 2072.71292, 2491.03271, 2983.22654, 3542.58461, 4181.78713, 4946.36956, 5866.78627, 6978.89231, 8314.54222, 9885.78799, 11715.8831, 13682.1658, 15976.2532, 18671.9076, 21842.8917, 25562.9679, 31730.2821, 38138.1175, 44404.1407, 50146.0184, 54981.4171, 57303.6531, 58770.977, 59525.7769, 60118.4485, 60529.8769, 60570.9443, 60231.0362, 59499.5377, 58365.8344, 56819.3115, 53283.2186, 50384.9456, 47825.0772, 45304.1981, 42522.8929, 39181.7465, 34981.3435, 28566.2521, 20658.1033},
        {14.8745343, 14.4685821, 14.0315826, 13.565014, 13.0703547, 12.5490829, 12.0026769, 11.432615, 10.8403753, 10.2274363, 9.59527607, 8.945373, 8.27920533, 7.59825132, 6.90398924, 6.19789737, 5.48145396, 4.75613731, 4.02342566, 3.28479729, 2.54173047, 1.79570346, 1.04819455, 0.30068199, -0.445355943, -1.18844098, -1.92709486, -2.6598393, -3.38519604, -4.10168681, -4.80783333, -5.50215735, -6.18318059, -6.84942478, -7.44358603, -7.99327309, -8.52676571, -9.04439726, -9.54650114, -10.0334107, -10.5054594, -10.9629805, -11.4063074, -11.8357736, -12.2517124, -12.6544571, -13.0443412, -13.4216981, -13.7868611, -14.1401635, -14.4819389, -14.8748588, -15.2555671, -15.6230456, -15.9762764, -16.3142415, -16.6359228, -16.9403025, -17.2263624, -17.4930847, -17.7394512, -17.9644441, -18.1670453, -18.3462368, -18.5010007, -18.6303189, -18.7331734, -18.8059336, -18.8450315, -18.8548375, -18.8345031, -18.7831796, -18.7000185, -18.5841711, -18.4347887, -18.2510228, -18.0320247, -17.7769459, -17.4849376, -17.1551512, -16.7867383, -16.37885, -15.9306378, -15.4412531, -14.4726658, -13.2518137, -12.0108887, -10.7632298, -9.52217553, -8.30106478, -7.11323627, -5.97202876, -4.89078098, -3.88283171, -2.96151969, -2.14018367, -1.43216241, -0.850794658, -0.409419171, -0.121374701, 0, -0.776799891, -1.76890363, -2.96640598, -4.35940167, -5.93798547, -7.69225211, -9.61229636, -11.688213, -13.9100966, -16.2680422, -18.7521443, -21.3524978, -24.0591974, -26.8623378, -29.7520138, -32.7183202, -35.6359499, -38.4785119, -41.4971284, -44.6571653, -47.9239885, -51.2629637, -54.6394568, -58.0188335, -61.3664597, -64.6477012, -67.8279237, -70.8724932, -73.7467754, -76.5912353, -79.1319365, -81.281715, -82.953407, -84.7355599, -86.2031541, -86.7509803, -86.0891609, -83.9278186, -79.977076, -73.9470558, -67.194926, -57.2824892, -43.5976197, -25.5281921, -2.46208111, 24.4212641, 57.9734373, 99.2544576, 149.324344, 209.243116, 277.138164, 360.586459, 461.028977, 581.305917, 727.114235, 902.868259, 1112.98232, 1359.40864, 1649.58323, 1993.62803, 2395.17136, 2868.75336, 3410.3743, 4032.47221, 4775.33797, 5666.47755, 6734.85727, 8008.36172, 9509.5208, 11263.8933, 13151.8017, 15366.3852, 17983.4768, 21078.9096, 24728.5165, 30877.6945, 37265.3216, 43499.8398, 49189.6912, 53943.3177, 56099.0649, 57382.2023, 57945.4989, 58370.8332, 58641.1184, 58560.4721, 58120.7473, 57313.797, 56131.474, 54565.6315, 51001.1295, 48127.9875, 45650.3813, 43272.4865, 40698.479, 37632.5343, 33778.8284, 27966.2004, 20737.1904},
        {15.8130147, 15.4745142, 15.0947929, 14.6757235, 14.2191787, 13.7270314, 13.2011541, 12.6434195, 12.0557005, 11.4398697, 10.7977999, 10.1313637, 9.44243379, 8.73288301, 8.00458402, 7.25940954, 6.49923228, 5.72592495, 4.94136027, 4.14741096, 3.34594973, 2.5388493, 1.72798237, 0.915221669, 0.102439907, -0.7084902, -1.51569594, -2.31730459, -3.11144345, -3.89623979, -4.6698209, -5.43031407, -6.17584658, -6.90454572, -7.5448635, -8.13123037, -8.69891421, -9.24830342, -9.77978636, -10.2937514, -10.790587, -11.2706814, -11.7344232, -12.1822005, -12.6144019, -13.0314157, -13.4336304, -13.8214341, -14.1952155, -14.5553627, -14.9022643, -15.296817, -15.6775891, -16.0436573, -16.3940982, -16.7279885, -17.0444049, -17.342424, -17.6211226, -17.8795773, -18.1168649, -18.3320619, -18.524245, -18.692491, -18.8358765, -18.9534782, -19.0443728, -19.1064758, -19.1377306, -19.1396087, -19.111262, -19.0518424, -18.960502, -18.8363926, -18.6786662, -18.4864748, -18.2589703, -17.9953046, -17.6946297, -17.3560976, -16.9788602, -16.5620694, -16.1048773, -15.6064356, -14.6307072, -13.4035923, -12.1562216, -10.9018698, -9.65381173, -8.42532193, -7.2296751, -6.08014594, -4.99000912, -3.97253931, -3.04101119, -2.20869945, -1.48887875, -0.894823784, -0.439809216, -0.137109729, 0, -0.686317427, -1.59878319, -2.7268846, -4.06010895, -5.58794356, -7.29987571, -9.18539272, -11.2339819, -13.4351305, -15.7783259, -18.2530554, -20.8488062, -23.5550658, -26.3613213, -29.25706, -32.2317694, -35.1402285, -37.9526805, -40.96331, -44.132738, -47.421586, -50.7904751, -54.2000267, -57.6108619, -60.9836021, -64.2788685, -67.4572824, -70.479465, -73.3060377, -76.0108173, -78.3997244, -80.3994215, -81.9365711, -83.6065197, -85.001588, -85.5357807, -84.935155, -82.9257685, -79.2336783, -73.5849419, -67.4613012, -58.2987902, -45.4799624, -28.3873718, -6.4035722, 19.3396815, 51.646337, 91.5711408, 140.168839, 198.494179, 264.956396, 346.489149, 444.816211, 562.634374, 704.627014, 874.585569, 1076.30148, 1309.47101, 1582.91151, 1909.90775, 2297.23115, 2756.32811, 3284.39961, 3892.56436, 4615.5239, 5477.15483, 6496.96267, 7697.69077, 9118.10999, 10787.9251, 12602.8451, 14747.1853, 17297.0338, 20328.4784, 23917.6073, 30035.1248, 36388.8483, 42581.2118, 48214.6491, 52891.5941, 54918.1175, 56057.2585, 56467.0353, 56726.6592, 56827.4261, 56585.1344, 55999.8547, 55071.6574, 53800.6131, 52186.7925, 48521.1839, 45701.216, 43393.2245, 41263.5449, 38978.513, 36204.4643, 32607.7345, 27177.6853, 20850.6522},
        {16.8250749, 16.5618009, 16.2463567, 15.8810362, 15.468133, 15.009941, 14.5087537, 13.9668651, 13.3865688, 12.7701585, 12.119928, 11.4381711, 10.7271814, 9.98925279, 9.22667889, 8.44175348, 7.63677029, 6.81402308, 5.97580556, 5.1244115, 4.26213461, 3.39126865, 2.51410734, 1.63294443, 0.750073665, -0.132211229, -1.01161651, -1.88584843, -2.75261327, -3.60961727, -4.4545667, -5.28516783, -6.09912691, -6.8941502, -7.58371089, -8.2094035, -8.81383498, -9.39744512, -9.96067372, -10.5039605, -11.0277454, -11.5324681, -12.0185683, -12.486486, -12.9366608, -13.3695325, -13.7855411, -14.1851261, -14.5687275, -14.9367849, -15.2897383, -15.6861004, -16.0669791, -16.4315552, -16.7790097, -17.1085233, -17.419277, -17.7104516, -17.981228, -18.2307871, -18.4583098, -18.6629769, -18.8439693, -19.0004679, -19.1316536, -19.2367072, -19.3148096, -19.3654276, -19.3880211, -19.3811813, -19.3440705, -19.2758512, -19.1756856, -19.0427362, -18.8761652, -18.6751351, -18.4388081, -18.1663465, -17.8569129, -17.5096694, -17.1237784, -16.6984023, -16.2327034, -15.7258441, -14.7446277, -13.5134901, -12.262064, -11.0035425, -9.75111883, -8.51798623, -7.31733791, -6.16236711, -5.06626704, -4.04223094, -3.10345203, -2.26312352, -1.53443865, -0.930590639, -0.464772711, -0.15017809, 0, -0.561797887, -1.3665611, -2.40285983, -3.65926425, -5.12434454, -6.78667089, -8.63481349, -10.6573425, -12.8428281, -15.1798406, -17.65695, -20.2627265, -22.9857404, -25.8145618, -28.737761, -31.743908, -34.6626571, -37.4598762, -40.4835883, -43.6883102, -47.0285585, -50.4588501, -53.9337017, -57.4076301, -60.835152, -64.1707843, -67.3690436, -70.3844467, -73.1715105, -75.7256403, -77.9454718, -79.7732549, -81.1512397, -82.7255315, -84.0864528, -84.6366316, -84.1071614, -82.2291357, -78.7336483, -73.3517927, -67.3835352, -58.5136136, -46.1661683, -29.7653396, -8.73526776, 15.182026, 45.4597527, 83.2532423, 129.717825, 186.00883, 251.202142, 331.074083, 427.453488, 542.802787, 680.877991, 844.85432, 1037.90699, 1257.41619, 1513.66922, 1823.27526, 2196.37027, 2641.08226, 3155.77608, 3750.16428, 4453.18748, 5284.9603, 6255.02382, 7380.75138, 8718.28593, 10301.8402, 12047.8636, 14126.4495, 16612.5746, 19581.2157, 23107.3497, 29169.1434, 35462.5862, 41586.8572, 47141.1358, 51724.6013, 53624.2186, 54626.191, 54892.0754, 54989.527, 54919.3554, 54513.1629, 53780.2195, 52729.7954, 51371.1607, 49713.5856, 46025.6766, 43300.0706, 41173.6661, 39283.3619, 37266.0565, 34758.6486, 31398.0369, 26285.0808, 20905.2082},
        {18.0009895, 17.8147897, 17.5647845, 17.253723, 16.8843541, 16.4594271, 15.9816908, 15.4538945, 14.8787871, 14.2591178, 13.5976355, 12.8970894, 12.1602285, 11.389802, 10.5885587, 9.75924787, 8.90461852, 8.0274197, 7.13040049, 6.21630995, 5.28789714, 4.34791114, 3.39910099, 2.44421577, 1.48600454, 0.527216369, -0.429399684, -1.38109455, -2.32511918, -3.25872449, -4.17916142, -5.08368091, -5.9695339, -6.83397132, -7.57269642, -8.2358699, -8.87487298, -9.49025201, -10.0825533, -10.6523234, -11.2001084, -11.7264548, -12.2319089, -12.717017, -13.1823257, -13.628381, -14.0557296, -14.4649176, -14.8564914, -15.2309975, -15.5889821, -15.9914193, -16.3771178, -16.745314, -17.0952444, -17.4261452, -17.7372529, -18.0278039, -18.2970346, -18.5441814, -18.7684807, -18.9691689, -19.1454824, -19.2966576, -19.4219309, -19.5205386, -19.5917173, -19.6355008, -19.651904, -19.6385183, -19.5945284, -19.5191191, -19.4114751, -19.270781, -19.0962215, -18.8869814, -18.6422453, -18.3611979, -18.0430239, -17.6869081, -17.2920351, -16.8575896, -16.3827563, -15.8667199, -14.8788248, -13.6422168, -12.3852085, -11.1209336, -9.8625259, -8.62311934, -7.41584775, -6.25384495, -5.15024479, -4.11818113, -3.17078779, -2.32119863, -1.58254749, -0.96796821, -0.49059464, -0.163560622, 0, -0.391604402, -1.05234921, -1.96924253, -3.12929247, -4.51950711, -6.12689458, -7.93846296, -9.94122037, -12.1221749, -14.4683346, -16.9667077, -19.6043022, -22.3681263, -25.2451879, -28.2224953, -31.2870566, -34.2516466, -37.0783832, -40.148373, -43.4112886, -46.8168023, -50.3145867, -53.854314, -57.3856569, -60.8582876, -64.2218787, -67.4261025, -70.4206316, -73.1551383, -75.541142, -77.5804577, -79.2342037, -80.4634983, -82.0105672, -83.445975, -84.1319911, -83.7954018, -82.162993, -78.9615511, -73.917862, -68.0363335, -59.3772889, -47.4175449, -31.6339183, -11.5032258, 10.5824101, 38.7903048, 74.3724681, 118.58091, 172.667639, 236.473627, 314.386526, 407.562881, 518.080265, 649.896686, 806.125875, 989.881562, 1198.74288, 1442.61602, 1737.44493, 2092.49402, 2516.67274, 3011.4832, 3586.63146, 4265.72472, 5065.98936, 5989.81815, 7050.59171, 8316.08044, 9823.2888, 11507.9347, 13523.8916, 15943.2249, 18837.9998, 22280.2814, 28225.6677, 34391.808, 40379.8843, 45791.0786, 50226.5729, 51971.1907, 52820.0448, 52938.4712, 52896.1791, 52698.4397, 52182.035, 51360.6607, 50248.0124, 48857.7855, 47203.6757, 43915.9605, 41381.6734, 39316.2992, 37435.3225, 35454.2282, 33088.5008, 30053.625, 25356.5722, 20776.4992},
        {19.3992951, 19.287765, 19.1002542, 18.8400022, 18.5102482, 18.1142318, 17.6551923, 17.1363691, 16.5610016, 15.9323293, 15.2535916, 14.5280278, 13.7588774, 12.9493797, 12.1027742, 11.2223003, 10.3111974, 9.3727049, 8.41006219, 7.42650868, 6.42528378, 5.4096269, 4.38277745, 3.34797483, 2.30845845, 1.26746772, 0.228242036, -0.80597919, -1.83195655, -2.84645064, -3.84622206, -4.82803139, -5.78863923, -6.72480618, -7.51057077, -8.20663593, -8.87517957, -9.51694488, -10.132675, -10.7231131, -11.2890023, -11.8310859, -12.3501069, -12.8468085, -13.3219339, -13.7762262, -14.2104286, -14.6252843, -15.0215365, -15.3999281, -15.7612026, -16.1770856, -16.5757989, -16.956547, -17.3185341, -17.6609647, -17.9830431, -18.2839738, -18.5629612, -18.8192097, -19.0519236, -19.2603074, -19.4435653, -19.600902, -19.7315216, -19.8346287, -19.9094276, -19.9548849, -19.9699726, -19.9543859, -19.9073446, -19.8280685, -19.7157775, -19.5696912, -19.3890296, -19.1730125, -18.9208597, -18.6317909, -18.3050261, -17.939785, -17.5352874, -17.0907533, -16.6054023, -16.0784543, -15.0806232, -13.8343948, -12.5674053, -11.2927803, -10.0236454, -8.77312638, -7.55434884, -6.38043845, -5.26452086, -4.21972173, -3.25916672, -2.39598148, -1.64329168, -1.01422298, -0.521901022, -0.179451477, 0, -0.177695443, -0.660563515, -1.43329139, -2.48056624, -3.78707524, -5.33750557, -7.11654439, -9.10887889, -11.2991962, -13.6721836, -16.2125282, -18.9049171, -21.7340376, -24.6845768, -27.7412219, -30.8886601, -33.9407796, -36.8578679, -40.0109428, -43.3480918, -46.8174022, -50.3669616, -53.9448573, -57.4991768, -60.9780074, -64.3294366, -67.5015517, -70.4424403, -73.1001897, -75.3043541, -77.1650163, -78.6658238, -79.790424, -81.4161361, -83.0797716, -84.0798296, -84.131856, -82.9513966, -80.2539975, -75.7552044, -70.209037, -61.9765103, -50.5699905, -35.5018438, -16.2844366, 4.48422672, 31.1488276, 64.9684095, 107.202016, 159.108689, 221.150418, 296.35748, 384.070409, 485.80741, 607.532764, 753.214555, 926.820866, 1130.35927, 1370.20393, 1654.86776, 1985.53077, 2378.85726, 2842.72841, 3389.23225, 4037.58518, 4803.73155, 5688.15612, 6702.79523, 7916.27034, 9365.13861, 10997.4591, 12951.9021, 15295.6588, 18095.9202, 21419.8774, 27163.1239, 33107.3479, 38862.6396, 44039.0897, 48246.7886, 49787.7198, 50452.152, 50410.7931, 50264.7213, 50014.2894, 49480.5291, 48672.759, 47600.2981, 46272.465, 44698.5786, 42449.3934, 40235.2657, 38022.3231, 35776.6932, 33464.5036, 31051.8818, 28504.9555, 24438.1279, 20370.2404},
        {20.8705639, 20.8380195, 20.7172932, 20.5121044, 20.2261729, 19.863218, 19.4269594, 18.9211167, 18.3494093, 17.7155568, 17.0232789, 16.276295, 15.4783246, 14.6330874, 13.744303, 12.8156907, 11.8509703, 10.8538612, 9.82808309, 8.77735541, 7.70539776, 6.61592968, 5.51267074, 4.39934048, 3.27965846, 2.15734423, 1.03611734, -0.0803026392, -1.18819617, -2.28384369, -3.36352565, -4.42352249, -5.46011467, -6.46958262, -7.30554392, -8.03861116, -8.74111167, -9.41395556, -10.058053, -10.674314, -11.2636489, -11.8269676, -12.3651804, -12.8791974, -13.3699287, -13.8382844, -14.2851747, -14.7115096, -15.1181994, -15.5061542, -15.8762841, -16.309642, -16.7252579, -17.1223046, -17.4999549, -17.8573816, -18.1937575, -18.5082553, -18.800048, -19.0683082, -19.3122087, -19.5309223, -19.7236219, -19.8894803, -20.0276701, -20.1373642, -20.2177355, -20.2666137, -20.281861, -20.2654199, -20.2165501, -20.1345116, -20.0185642, -19.8679677, -19.681982, -19.4598669, -19.2008824, -18.9042881, -18.569344, -18.19531, -17.7814458, -17.3270113, -16.8312664, -16.2934709, -15.2854755, -14.0292345, -12.7518395, -11.4664207, -10.1861081, -8.92403189, -7.69332209, -6.50710883, -5.3785222, -4.32069231, -3.34674925, -2.46982313, -1.70304404, -1.05954209, -0.552447361, -0.194889965, 0, -0.0111141548, -0.354874997, -1.01422678, -1.97211378, -3.21148023, -4.71527041, -6.46642857, -8.44789897, -10.6426259, -13.0335535, -15.6036262, -18.3357881, -21.2129836, -24.2181569, -27.3342522, -30.5442138, -33.666474, -36.6604753, -39.8812725, -43.2765572, -46.7940211, -50.381356, -53.9862534, -57.5564051, -61.0395029, -64.3832382, -67.5353029, -70.4433886, -73.0551871, -75.1348559, -76.8809166, -78.2961211, -79.3832212, -81.1361621, -83.062099, -84.403869, -84.866866, -84.1564839, -81.9781167, -78.0371582, -72.821108, -65.0152222, -54.1718742, -39.843437, -21.5822841, -2.14254526, 22.925602, 54.8702237, 94.9393857, 144.381154, 204.179321, 276.169227, 357.893756, 450.345893, 561.56258, 696.418166, 859.786999, 1058.40924, 1294.86937, 1569.71647, 1875.56001, 2237.15916, 2668.42327, 3183.93845, 3799.43753, 4530.12698, 5376.38021, 6349.55809, 7515.4093, 8908.91773, 10487.0893, 12374.8825, 14634.2734, 17327.2382, 20515.7532, 26021.9043, 31707.5306, 37194.5807, 42105.0036, 46060.7479, 47393.0409, 47875.0339, 47681.842, 47446.3969, 47162.5947, 46633.5745, 45862.77, 44853.6153, 43609.5441, 42133.9903, 40939.7954, 39050.0355, 36684.3908, 34062.5416, 31404.1681, 28928.9508, 26856.5698, 23430.139, 19789.2119},
        {22.3686335, 22.4206627, 22.372614, 22.2286546, 21.9929514, 21.6696717, 21.2629825, 20.7770509, 20.216044, 19.5841289, 18.8854728, 18.1242427, 17.3046057, 16.4307289, 15.5067795, 14.5369244, 13.5253309, 12.476166, 11.3935969, 10.2817905, 9.14491408, 7.98713468, 6.81261941, 5.62553535, 4.43004963, 3.23032932, 2.03054154, 0.834853391, -0.352568036, -1.52755564, -2.68594231, -3.82356096, -4.93624448, -6.01982577, -6.91078409, -7.68798094, -8.43219316, -9.14442496, -9.82568058, -10.4769643, -11.0992802, -11.6936327, -12.261026, -12.8024642, -13.3189516, -13.8114925, -14.281091, -14.7287515, -15.1554781, -15.5622751, -15.9501467, -16.404868, -16.8408347, -17.2572133, -17.6531705, -18.0278728, -18.3804868, -18.710179, -19.016116, -19.2974644, -19.5533907, -19.7830615, -19.9856433, -20.1603028, -20.3062065, -20.4225209, -20.5084127, -20.560971, -20.5773347, -20.5609551, -20.5111333, -20.4271705, -20.308368, -20.1540271, -19.9634489, -19.7359347, -19.4707858, -19.1673034, -18.8247887, -18.4425429, -18.0198673, -17.5560632, -17.0504317, -16.5022741, -15.4843846, -14.2182889, -12.9306392, -11.6345775, -10.3432462, -9.06978743, -7.82734338, -6.62905628, -5.48806834, -4.41752178, -3.4305588, -2.54032162, -1.75995244, -1.10259348, -0.581386948, -0.20947505, 0, 0.0613315924, -0.219876078, -0.826018855, -1.73949258, -2.9426931, -4.41801624, -6.14785787, -8.11461381, -10.3006799, -12.688452, -15.260326, -17.9986976, -20.8859628, -23.9045173, -27.0367571, -30.2650779, -33.4133953, -36.4414649, -39.6901503, -43.1078871, -46.643111, -50.2442579, -53.8597635, -57.4380637, -60.927594, -64.2767904, -67.4340885, -70.3479242, -72.9667332, -75.0168925, -76.7503182, -78.1820637, -79.3271823, -81.2456868, -83.430151, -85.0944926, -85.9402771, -85.6690701, -83.982437, -80.5819434, -75.7205027, -68.3805311, -58.1557215, -44.6397672, -27.4263612, -9.33423459, 14.0742169, 44.0115599, 81.6903605, 128.323185, 185.218584, 253.375827, 328.961637, 412.571052, 513.840256, 638.346142, 791.665604, 984.450656, 1216.04964, 1480.27438, 1762.58618, 2094.26639, 2493.72277, 2977.50253, 3559.00022, 4253.0575, 5061.20315, 5994.82735, 7114.13423, 8451.716, 9969.20119, 11781.0632, 13944.4585, 16516.5437, 19554.4753, 24801.7007, 30207.5129, 35407.4959, 40037.2334, 43732.3093, 44862.8053, 45173.4756, 44842.2622, 44527.7181, 44217.5434, 43699.1837, 42970.3519, 42028.7607, 40872.1231, 39498.1519, 39169.5033, 37561.4057, 35096.6414, 32197.993, 29288.2428, 26790.1733, 25126.5668, 22305.2832, 19038.3405},
        {23.831104, 23.9749415, 24.0075245, 23.933407, 23.7571429, 23.4832862, 23.1163908, 22.6610107, 22.1216999, 21.5030124, 20.8095021, 20.045723, 19.2162291, 18.3255743, 17.3783126, 16.3789979, 15.3321843, 14.2424257, 13.1142761, 11.9522894, 10.7610197, 9.54502081, 8.30884678, 7.05705158, 5.79418916, 4.5248135, 3.25347856, 1.98473833, 0.72314675, -0.526742193, -1.76037453, -2.9731963, -4.16065354, -5.31819226, -6.27097753, -7.10317749, -7.90100695, -8.66543788, -9.39744225, -10.097992, -10.7680592, -11.4086158, -12.0206338, -12.605085, -13.1629416, -13.6951754, -14.2027585, -14.6866629, -15.1478604, -15.5873232, -16.0060231, -16.4859785, -16.9453581, -17.383377, -17.7992501, -18.1921926, -18.5614195, -18.9061458, -19.2255866, -19.5189569, -19.7854719, -20.0243464, -20.2347957, -20.4160347, -20.5672785, -20.6877421, -20.7766406, -20.8312344, -20.8488311, -20.8326768, -20.7821132, -20.6964821, -20.5751252, -20.4173843, -20.2226012, -19.9901176, -19.7192752, -19.4094159, -19.0598813, -18.6700132, -18.2391535, -17.7666437, -17.2518258, -16.6940414, -15.6669079, -14.3915837, -13.0943444, -11.7883471, -10.4867492, -9.20270818, -7.94938122, -6.73992568, -5.58749891, -4.50525823, -3.50636098, -2.60396448, -1.81122608, -1.1413031, -0.607352871, -0.222532727, 0, -0.0300934069, -0.381735319, -1.03887411, -1.98545814, -3.20543579, -4.68275542, -6.40136541, -8.34521412, -10.4982499, -12.8444212, -15.3676763, -18.0519636, -20.8812315, -23.8394284, -26.9105025, -30.0784024, -33.169957, -36.1466691, -39.347876, -42.7238576, -46.2248939, -49.8012648, -53.4032503, -56.9811303, -60.4851847, -63.8656936, -67.0729368, -70.0571944, -72.7687463, -74.9400468, -76.8190708, -78.421446, -79.7628, -81.8730677, -84.2607229, -86.1637579, -87.2935081, -87.3613089, -86.0784959, -83.1564045, -78.7097783, -71.9237681, -62.430782, -49.8632275, -33.8535126, -17.1370908, 4.5324374, 32.2984255, 67.3042273, 110.693196, 163.769314, 227.318983, 297.080039, 373.500035, 466.642202, 582.252529, 726.077003, 910.40858, 1132.95833, 1384.2952, 1646.61725, 1953.68354, 2325.29873, 2778.59326, 3326.16626, 3982.69681, 4751.60935, 5644.66508, 6714.89224, 7991.97609, 9436.26972, 11157.4571, 13209.223, 15645.2524, 18519.2302, 23501.2809, 28624.5398, 33538.5817, 37892.9818, 41337.3148, 42287.2289, 42448.8439, 42000.5604, 41611.8911, 41266.8381, 40744.774, 40039.9326, 39146.5478, 38058.8533, 36771.0831, 36850.8165, 35420.114, 32985.3643, 30052.9565, 27129.2793, 24720.7218, 23333.6728, 21027.5037, 18120.4467},
        {25.0387119, 25.2868614, 25.4145122, 25.4264613, 25.3275063, 25.1224443, 24.8160727, 24.4131887, 23.9185897, 23.337073, 22.6734359, 21.9324756, 21.1189895, 20.2377749, 19.293629, 18.2913493, 17.2357329, 16.1315772, 14.9836794, 13.796837, 12.5758471, 11.3255071, 10.0506143, 8.75596593, 7.44635939, 6.12659193, 4.80146085, 3.47576346, 2.15429706, 0.841858947, -0.456753575, -1.73674321, -2.99331265, -4.22166459, -5.24887012, -6.15773417, -7.03308248, -7.87557394, -8.68586746, -9.46462191, -10.2124962, -10.9301492, -11.6182399, -12.277427, -12.9083696, -13.5117265, -14.0881566, -14.6383187, -15.1628719, -15.6624749, -16.1377867, -16.6480153, -17.1340012, -17.5951341, -18.0308036, -18.4403995, -18.8233114, -19.1789289, -19.5066417, -19.8058394, -20.0759119, -20.3162486, -20.5262393, -20.7052736, -20.8527412, -20.9680318, -21.050535, -21.0996049, -21.1145967, -21.0949732, -21.0401263, -20.9494481, -20.8223304, -20.6581652, -20.4563445, -20.2162603, -19.9373045, -19.6188691, -19.2603461, -18.8611274, -18.4206049, -17.9381707, -17.4132167, -16.8451349, -15.8086369, -14.5243873, -13.2181553, -11.9031145, -10.5924386, -9.29930127, -8.0368761, -6.81833678, -5.65685697, -4.56561033, -3.55777053, -2.64651121, -1.84500603, -1.16642867, -0.623952768, -0.230751992, 0, -0.576613768, -1.36826738, -2.36627026, -3.56193182, -4.94656149, -6.5114687, -8.24796285, -10.1473534, -12.2009497, -14.4000613, -16.7359975, -19.2000677, -21.7835814, -24.4778481, -27.2741771, -30.1638778, -32.9607838, -35.6308508, -38.5668218, -41.7219757, -45.0495913, -48.5029475, -52.0353232, -55.5999972, -59.1502485, -62.6393559, -66.0205983, -69.2472545, -72.2726035, -74.9454669, -77.3618818, -79.5064641, -81.3638299, -83.661715, -86.0131754, -87.8331021, -88.8831752, -88.9250748, -87.7204809, -85.0310738, -81.1529126, -75.1506611, -66.6814471, -55.4023982, -40.970642, -25.6932029, -5.91483764, 19.3698002, 51.1660571, 90.4792795, 137.81456, 195.288921, 260.869321, 335.519956, 426.324374, 537.618662, 673.738903, 841.75524, 1042.64832, 1274.41619, 1527.69523, 1826.80411, 2184.47605, 2614.39159, 3131.83637, 3751.35928, 4478.5287, 5325.56523, 6337.61794, 7541.20967, 8881.72859, 10479.3145, 12388.5754, 14664.1192, 17360.5539, 22110.4877, 26996.0283, 31677.2833, 35814.3608, 39067.3684, 39897.2294, 39962.6928, 39437.8027, 38963.3941, 38528.7398, 37928.6163, 37162.021, 36227.9514, 35125.4051, 33853.3795, 33019.4397, 31458.7971, 29421.3842, 27157.1331, 24915.9761, 22947.8456, 21502.6737, 19476.3589, 17018.0013},
        {26.3216461, 26.6791931, 26.9073435, 27.0111034, 26.9954789, 26.8654761, 26.6261012, 26.2823602, 25.8392594, 25.3018047, 24.6750025, 23.9638586, 23.1733794, 22.3085709, 21.3744392, 20.3759905, 19.3182309, 18.2061665, 17.0448034, 15.8391478, 14.5942057, 13.3149833, 12.0064868, 10.6737222, 9.32169566, 7.95541332, 6.57988129, 5.2001057, 3.82109267, 2.44784832, 1.08537877, -0.261309842, -1.58721141, -2.8873198, -3.99422383, -4.98753588, -5.94882032, -6.87834138, -7.77636328, -8.64315028, -9.47896659, -10.2840765, -11.0587441, -11.8032338, -12.5178097, -13.2027361, -13.8582773, -14.4846974, -15.0822606, -15.6512313, -16.1918737, -16.7344284, -17.2485334, -17.7338031, -18.1898519, -18.6162943, -19.0127445, -19.378817, -19.7141262, -20.0182865, -20.2909124, -20.5316181, -20.7400182, -20.915727, -21.0583589, -21.1675284, -21.2428497, -21.2869729, -21.3024749, -21.2827096, -21.2270948, -21.1350479, -21.0059867, -20.8393288, -20.6344917, -20.3908932, -20.1079508, -19.7850822, -19.4217049, -19.0172367, -18.5710951, -18.0826978, -17.5514623, -16.9768064, -15.9340817, -14.6436584, -13.3310037, -12.0092971, -10.6917179, -9.39144546, -8.12165921, -6.89553851, -5.72626275, -4.6270113, -3.61096355, -2.69129888, -1.88119667, -1.19383629, -0.642397133, -0.240058576, 0, -1.09363337, -2.30042339, -3.61874496, -5.04697302, -6.58348246, -8.22664822, -9.9748452, -11.8264483, -13.7798325, -15.8333727, -17.9854437, -20.2344206, -22.5786782, -25.0165915, -27.5465353, -30.1668846, -32.6836628, -35.0677659, -37.7526498, -40.6954712, -43.8533868, -47.1835534, -50.6431277, -54.1892664, -57.7791262, -61.369864, -64.9186363, -68.3825999, -71.7189116, -74.8936803, -77.851828, -80.5478259, -82.9361448, -85.4509395, -87.8068392, -89.5905576, -90.6126607, -90.6837146, -89.6142851, -87.2149384, -83.9501138, -78.777499, -71.3797281, -61.4394355, -48.6392553, -34.8089584, -16.9472583, 5.799995, 34.2869515, 69.3677611, 110.654741, 161.76167, 223.102796, 296.219291, 384.953132, 492.113279, 620.508695, 771.538004, 949.741035, 1161.19617, 1405.74288, 1697.38875, 2041.25356, 2447.23501, 2933.32673, 3513.80621, 4196.02165, 4992.45407, 5940.9917, 7065.15101, 8295.65373, 9763.08718, 11523.0891, 13631.2972, 16143.3492, 20650.4265, 25288.7372, 29730.0333, 33646.0663, 36708.5883, 37430.8826, 37415.4816, 36830.6237, 36281.8104, 35764.1965, 35090.7439, 34265.7173, 33293.3811, 32177.9999, 30923.8384, 29081.016, 27375.6902, 25771.4502, 24231.885, 22720.5839, 21201.1359, 19637.1304, 17826.9842, 15784.7545},
        {27.8139918, 28.2847413, 28.6169902, 28.8159518, 28.8868391, 28.8348655, 28.665244, 28.383188, 27.9939105, 27.5026247, 26.9145439, 26.2348813, 25.4688499, 24.621663, 23.6985339, 22.7046756, 21.6453014, 20.5256245, 19.350858, 18.1262151, 16.856909, 15.548153, 14.2051602, 12.8331437, 11.4373168, 10.0228927, 8.59508453, 7.15910549, 5.72016878, 4.28348758, 2.85427508, 1.43774446, 0.0391089253, -1.33641835, -2.52825506, -3.61214055, -4.6656816, -5.68870704, -6.6810457, -7.64252642, -8.57297806, -9.47222943, -10.3401094, -11.1764468, -11.9810704, -12.7538091, -13.4944918, -14.2029472, -14.8790043, -15.5224918, -16.1332386, -16.7081785, -17.2500981, -17.758889, -18.2344429, -18.6766512, -19.0854057, -19.4605979, -19.8021194, -20.1098618, -20.3837167, -20.6235758, -20.8293305, -21.0008725, -21.1380934, -21.2408848, -21.3091382, -21.350428, -21.3721438, -21.3583324, -21.3083874, -21.2217024, -21.0976709, -20.9356865, -20.7351428, -20.4954333, -20.2159516, -19.8960912, -19.5352458, -19.1328089, -18.688174, -18.2007348, -17.6698847, -17.0950174, -16.0511339, -14.7587504, -13.4434844, -12.1185016, -10.7969681, -9.49204965, -8.21691231, -6.98472195, -5.8086445, -4.70184587, -3.67749199, -2.74874878, -1.92878216, -1.23075805, -0.667842375, -0.25320105, 0, -1.41740813, -2.88074941, -4.39313838, -5.95768956, -7.57751748, -9.25573666, -10.9954616, -12.799807, -14.6718871, -16.6148167, -18.6317101, -20.725682, -22.8998469, -25.1573193, -27.5012136, -29.9346446, -32.2674675, -34.4751884, -36.9980556, -39.797771, -42.8360364, -46.0745537, -49.4750248, -52.9991514, -56.6086355, -60.2651789, -63.9304834, -67.566251, -71.1341835, -74.6794464, -78.0496745, -81.1815306, -84.0116775, -86.7889956, -89.2940386, -91.2241027, -92.4221852, -92.7312836, -91.9943954, -90.0545181, -87.4144776, -83.0566255, -76.6948623, -68.0430888, -56.8152055, -44.3982084, -28.4146288, -8.16009339, 17.069771, 47.9793378, 83.3760041, 128.179646, 184.838438, 255.374644, 340.940955, 443.080482, 563.336336, 698.682457, 856.269576, 1048.23297, 1281.44312, 1561.69308, 1888.26129, 2267.90703, 2720.06546, 3258.35331, 3890.74411, 4629.39152, 5507.14094, 6545.13336, 7664.65769, 9002.06885, 10613.072, 12553.3723, 14878.675, 19127.0355, 23501.2212, 27686.3494, 31367.538, 34229.9043, 34845.5654, 34754.6399, 34119.2455, 33511.9248, 32932.3358, 32209.1263, 31350.5049, 30364.6802, 29259.8608, 28044.2554, 25426.8212, 23644.9942, 22426.6325, 21499.5942, 20591.7374, 19430.9201, 17745.0006, 16120.7662, 14438.4116},
        {29.9100206, 30.4978754, 30.9361988, 31.2304972, 31.386277, 31.4090445, 31.3043061, 31.0775682, 30.7343372, 30.2801194, 29.7204212, 29.0607489, 28.3066091, 27.4635079, 26.5369518, 25.5324472, 24.4555004, 23.3116178, 22.1063058, 20.8450707, 19.533419, 18.1768569, 16.7808909, 15.3510274, 13.8927727, 12.4116331, 10.9131151, 9.40272505, 7.88596928, 6.36835419, 4.85538613, 3.3525715, 1.86541665, 0.399427958, -0.888284702, -2.07120548, -3.2246407, -4.3480038, -5.44070823, -6.50216746, -7.53179493, -8.5290041, -9.49320842, -10.4238213, -11.3202563, -12.1819268, -13.0082462, -13.7986281, -14.5524858, -15.2692328, -15.9482827, -16.5494539, -17.1126132, -17.6380322, -18.1259829, -18.5767369, -18.9905662, -19.3677423, -19.7085373, -20.0132228, -20.2820706, -20.5153526, -20.7133405, -20.8763061, -21.0045212, -21.0982576, -21.157787, -21.1992726, -21.2384955, -21.2429534, -21.2118878, -21.1445403, -21.0401527, -20.8979665, -20.7172234, -20.4971649, -20.2370328, -19.9360686, -19.5935141, -19.2086107, -18.7806002, -18.3087241, -17.7922242, -17.230342, -16.1947086, -14.907365, -13.595492, -12.2722081, -10.9506316, -9.64388104, -8.36507473, -7.12733114, -5.94376867, -4.82750576, -3.79166081, -2.84935226, -2.01369851, -1.29781799, -0.714829124, -0.277850319, 0, -1.21853001, -2.51030444, -3.87679952, -5.31949147, -6.83985651, -8.43937087, -10.1195108, -11.8817524, -13.727572, -15.6584459, -17.6758501, -19.781261, -21.9761547, -24.2620076, -26.6402957, -29.1124954, -31.5159241, -33.8286067, -36.423331, -39.2663963, -42.3241019, -45.5627472, -48.9486316, -52.4480543, -56.0273147, -59.6527123, -63.2905462, -66.907116, -70.4687209, -73.9732538, -77.3438362, -80.5372894, -83.5104347, -86.6683344, -89.7436893, -92.349669, -94.3086225, -95.4428986, -95.5748464, -94.5268147, -92.4084104, -88.667298, -83.069624, -75.3815345, -65.3691758, -54.159157, -39.8170458, -21.7688728, 0.559331439, 27.7415364, 58.1301983, 97.2359825, 148.031012, 212.672171, 291.651893, 386.209917, 497.585981, 622.855725, 768.369372, 945.019798, 1158.71542, 1414.96857, 1713.61506, 2060.82075, 2473.47743, 2963.5535, 3536.63178, 4203.02518, 4996.46039, 5937.41994, 6952.36977, 8170.27309, 9644.15663, 11427.0471, 13571.9713, 17542.6594, 21628.7587, 25530.6202, 28948.5948, 31583.0333, 32064.4517, 31876.259, 31177.3067, 30531.0444, 29936.1702, 29222.7996, 28398.0597, 27469.0778, 26442.981, 25326.8965, 22797.0422, 21174.1884, 20148.1831, 19408.8743, 18646.1098, 17549.7376, 15809.6056, 14409.3941, 12981.7663},
        {31.9215462, 32.6247723, 33.1686485, 33.5589188, 33.8013276, 33.901619, 33.8655373, 33.6988267, 33.4072314, 32.9964956, 32.4723636, 31.8405796, 31.1068878, 30.2770324, 29.3567576, 28.3518077, 27.2679269, 26.1108594, 24.8863494, 23.6001411, 22.2579788, 20.8656067, 19.428769, 17.9532099, 16.4446737, 14.9089046, 13.3516467, 11.7786443, 10.1956417, 8.60838303, 7.02261252, 5.44407441, 3.87851291, 2.33167224, 0.955366234, -0.320768077, -1.56884929, -2.78787343, -3.9768365, -5.13473454, -6.26056355, -7.35331956, -8.41199859, -9.43559665, -10.4231098, -11.3735339, -12.2858652, -13.1590996, -13.9922331, -14.7842618, -15.5341816, -16.1667693, -16.7568492, -17.3050261, -17.8119051, -18.278091, -18.7041889, -19.0908036, -19.4385402, -19.7480035, -20.0197985, -20.2545302, -20.4528035, -20.6152233, -20.7423946, -20.8349224, -20.8934116, -20.9416193, -21.0027464, -21.0296487, -21.0214302, -20.9771949, -20.8960471, -20.7770907, -20.6194299, -20.4221687, -20.1844113, -19.9052617, -19.5838241, -19.2192025, -18.810501, -18.3568237, -17.8572748, -17.3109582, -16.2872943, -15.0092669, -13.7050243, -12.3876147, -11.0700861, -9.76548682, -8.48686496, -7.24726869, -6.05974617, -4.93734555, -3.893115, -2.94010268, -2.09135674, -1.35992534, -0.758856642, -0.30119881, 0, -0.856806139, -1.84443051, -2.96069201, -4.20340958, -5.57040211, -7.05948852, -8.66848773, -10.3952186, -12.2375002, -14.1931513, -16.2599908, -18.4358377, -20.7185108, -23.1058292, -25.5956116, -28.1856771, -30.7534122, -33.2794313, -36.0339596, -38.9890091, -42.1165918, -45.3887196, -48.7774047, -52.254659, -55.7924945, -59.3629231, -62.937957, -66.489608, -69.9898882, -73.3786296, -76.6718234, -79.8511359, -82.8982328, -86.3937266, -90.0198102, -93.2989572, -96.0331499, -98.0243702, -99.0746003, -98.9858223, -97.5090085, -94.5126754, -89.8087853, -83.2093005, -74.5261832, -64.6871675, -52.1095008, -36.3262026, -16.8702921, 6.72521132, 32.6355053, 66.4204217, 110.904351, 168.174303, 238.811797, 324.074286, 425.219226, 540.63989, 675.108862, 836.523282, 1028.01408, 1256.64417, 1524.21132, 1836.89226, 2208.28362, 2648.57615, 3160.54045, 3752.44354, 4459.75973, 5302.57333, 6218.03687, 7321.92521, 8663.02983, 10290.1422, 12252.0539, 15914.5942, 19680.2572, 23268.575, 26399.0796, 28791.3028, 29153.0116, 28890.0133, 28155.4536, 27497.2059, 26912.1695, 26232.7739, 25464.1419, 24611.3963, 23679.6599, 22674.0555, 20480.2832, 19068.1754, 18157.2795, 17467.1429, 16717.313, 15627.3372, 13916.7629, 12701.6958, 11479.65},
        {33.5062484, 34.309998, 34.9482375, 35.4267956, 35.7515013, 35.9281834, 35.9626708, 35.8607923, 35.6283768, 35.2712532, 34.7952503, 34.2061969, 33.509922, 32.7122544, 31.819023, 30.8360565, 29.769184, 28.6242342, 27.407036, 26.1234182, 24.7792098, 23.3802395, 21.9323363, 20.441329, 18.9130464, 17.3533175, 15.767971, 14.1628359, 12.543741, 10.9165151, 9.28698718, 7.66098601, 6.04434049, 4.44287947, 2.99826976, 1.64556408, 0.318182473, -0.982454398, -2.25492588, -3.49781131, -4.70969003, -5.8891414, -7.03474474, -8.1450794, -9.21872472, -10.2542601, -11.2502647, -12.2053181, -13.1179995, -13.9868883, -14.8105638, -15.4850029, -16.1136117, -16.6971936, -17.2365523, -17.7324914, -18.1858142, -18.5973245, -18.9678259, -19.2981217, -19.5890157, -19.8413114, -20.0558123, -20.2333221, -20.3746442, -20.4805822, -20.5519398, -20.6166433, -20.701968, -20.7527786, -20.7681203, -20.7470384, -20.6885782, -20.591785, -20.4557041, -20.2793808, -20.0618604, -19.8021882, -19.4994094, -19.1525694, -18.7607135, -18.3228869, -17.8381349, -17.3055029, -16.2952778, -15.0299025, -13.736973, -12.4294487, -11.120289, -9.82245309, -8.54890038, -7.31259018, -6.12648181, -5.0035346, -3.95670788, -2.99896097, -2.14325319, -1.40254389, -0.789792365, -0.317957962, 0, -0.41332445, -1.03144328, -1.84746339, -2.85449168, -4.04563505, -5.4140004, -6.95269463, -8.65482465, -10.5134973, -12.5218196, -14.6728984, -16.9598405, -19.3757529, -21.9137425, -24.5669162, -27.3283809, -30.1298, -32.9555028, -35.9375754, -39.0559583, -42.290592, -45.6214167, -49.0283731, -52.4914016, -55.9904425, -59.5054364, -63.0163237, -66.5030449, -69.9455404, -73.2677146, -76.5260905, -79.7174195, -82.8384528, -86.5124689, -90.4229555, -94.0863264, -97.3113747, -99.9068935, -101.681676, -102.444516, -101.876428, -99.9528721, -96.5076411, -91.3745281, -84.3873259, -76.2980102, -65.7926459, -52.4754803, -35.9507608, -15.8227349, 6.28772276, 35.2577588, 73.1258853, 121.665422, 182.108255, 255.929362, 344.603722, 448.169883, 569.865716, 714.496105, 882.510846, 1082.50664, 1318.02608, 1595.29747, 1925.09995, 2316.12371, 2769.0663, 3290.54574, 3916.48665, 4666.23618, 5491.43613, 6489.55522, 7703.37707, 9175.68532, 10949.2636, 14257.6355, 17655.1597, 20886.9348, 23698.0593, 25833.6319, 26113.1301, 25824.3543, 25109.0979, 24477.6201, 23925.8632, 23296.2422, 22592.3757, 21817.8826, 20976.3814, 20071.4909, 18311.4585, 17101.3357, 16236.3654, 15511.7905, 14722.8539, 13664.7986, 12132.8676, 11027.6383, 9979.45002},
        {34.2768098, 35.1466572, 35.8517162, 36.3976198, 36.7900011, 37.034493, 37.1367287, 37.1023411, 36.9369632, 36.6462281, 36.2357688, 35.7112183, 35.0782097, 34.342376, 33.5093502, 32.5847653, 31.5742544, 30.4834505, 29.3179866, 28.0834958, 26.785611, 25.4299653, 24.0221917, 22.5679233, 21.0727931, 19.542434, 17.9824792, 16.3985617, 14.7963144, 13.1813705, 11.5593628, 9.93592458, 8.31668871, 6.70728825, 5.23029075, 3.83040242, 2.450924, 1.09369377, -0.239449997, -1.54666902, -2.82612503, -4.07597975, -5.29439489, -6.47953219, -7.62955336, -8.74262013, -9.81689422, -10.8505374, -11.8417113, -12.7885777, -13.6892983, -14.4216816, -15.1068515, -15.7455789, -16.3386345, -16.886789, -17.3908131, -17.8514777, -18.2695535, -18.6458112, -18.9810215, -19.2759553, -19.5313832, -19.748076, -19.9268045, -20.0683394, -20.1734514, -20.2681639, -20.3778939, -20.4513305, -20.4876076, -20.4858588, -20.4452179, -20.3648186, -20.2437946, -20.0812797, -19.8764075, -19.6283119, -19.3361265, -18.9989851, -18.6160213, -18.186369, -17.7091618, -17.1835335, -16.1857059, -14.9351641, -13.6564762, -12.3625002, -11.0660942, -9.78011613, -8.51742406, -7.29087603, -6.11333006, -4.99764418, -3.95667642, -3.00328481, -2.15032737, -1.41066215, -0.797147156, -0.322640431, 0, 0.0412982337, -0.200740664, -0.714297917, -1.48755475, -2.50869237, -3.76589202, -5.24733491, -6.94120227, -8.83567532, -10.9189353, -13.1791634, -15.6045408, -18.1832488, -20.9034686, -23.7533815, -26.7211686, -29.8111469, -33.0138027, -36.2808197, -39.603837, -42.9744932, -46.3844273, -49.8252782, -53.2886847, -56.7662858, -60.2497203, -63.7306271, -67.2006451, -70.6514131, -74.0960649, -77.4968631, -80.8389981, -84.1076602, -87.6555132, -91.2840108, -94.6966139, -97.7682709, -100.37393, -102.38854, -103.687049, -104.477428, -104.200249, -102.665302, -99.6823798, -95.0612742, -89.3494758, -81.4346527, -70.9421719, -57.4974003, -40.725705, -21.5271387, 3.30556642, 34.3761307, 72.8400105, 120.979124, 180.56963, 253.387688, 340.494185, 444.543116, 568.968776, 713.757819, 886.973429, 1091.86365, 1334.12598, 1623.60839, 1968.25371, 2368.76473, 2831.20724, 3388.19385, 4057.32075, 4807.04301, 5711.6582, 6805.77293, 8123.99385, 9700.92756, 12588.7229, 15550.0587, 18365.1562, 20814.2362, 22677.52, 22940.4909, 22707.9328, 22100.6688, 21550.4136, 21054.929, 20482.4382, 19837.1797, 19123.392, 18345.3135, 17507.1828, 16120.3113, 15035.6564, 14150.1282, 13360.6371, 12564.0931, 11657.4067, 10537.488, 9420.38473, 8532.7226},
        {34.7165028, 35.7229672, 36.5604878, 37.2346608, 37.7510824, 38.1153485, 38.3330554, 38.4097991, 38.3511758, 38.1627815, 37.8502125, 37.4190647, 36.8749344, 36.2234176, 35.4701104, 34.620609, 33.6805095, 32.655408, 31.5509006, 30.3725834, 29.1260525, 27.8169041, 26.4507343, 25.0331391, 23.5697147, 22.0660573, 20.5277628, 18.9604275, 17.3696475, 15.7610188, 14.1401376, 12.5126, 10.8840021, 9.25994007, 7.75154167, 6.30941745, 4.88290956, 3.4741894, 2.08542838, 0.71879791, -0.623530607, -1.93938576, -3.22659614, -4.48299035, -5.70639697, -6.89464459, -8.04556182, -9.15697724, -10.2267194, -11.252617, -12.2324986, -13.0368562, -13.7936154, -14.5033652, -15.1666945, -15.7841924, -16.3564478, -16.8840498, -17.3675873, -17.8076493, -18.2048249, -18.559703, -18.8728727, -19.1449228, -19.3764425, -19.5680207, -19.7202465, -19.8526566, -19.984333, -20.0767862, -20.1293769, -20.1414656, -20.1124132, -20.0415803, -19.9283275, -19.7720156, -19.5720053, -19.3276572, -19.038332, -18.7033904, -18.3221931, -17.8941007, -17.418474, -16.8946737, -15.9106327, -14.6800795, -13.4216874, -12.1481686, -10.8722355, -9.60660044, -8.36397565, -7.15707345, -5.99860614, -4.90128602, -3.87782541, -2.94093659, -2.10333186, -1.37772354, -0.776823925, -0.31334531, 0, 0.230678701, 0.146432905, -0.239036537, -0.912028772, -1.85884295, -3.06577821, -4.51913372, -6.20520861, -8.11030203, -10.2207131, -12.5227411, -15.002685, -17.646844, -20.4415173, -23.3730041, -26.4276034, -29.6553488, -33.0516444, -36.4586282, -39.8762567, -43.3044864, -46.7432738, -50.1925756, -53.6523481, -57.1225479, -60.6031315, -64.0940555, -67.5952763, -71.1067505, -74.7043936, -78.2843515, -81.823793, -85.2998869, -88.8449425, -92.3585583, -95.6989623, -98.7967811, -101.582642, -103.98717, -105.940994, -108.035224, -109.338985, -109.653679, -108.780708, -106.521473, -103.402458, -98.3187122, -90.8903652, -80.7375481, -67.4803916, -51.2673348, -30.5444901, -5.94178552, 23.1215201, 59.6979505, 105.73025, 163.161162, 233.69478, 319.566737, 423.273012, 544.627252, 690.95278, 865.224218, 1072.55356, 1321.67447, 1619.6582, 1967.16627, 2369.60816, 2855.8957, 3441.64648, 4108.51937, 4911.31506, 5877.40605, 7034.16485, 8408.96398, 10883.1514, 13416.6303, 15823.2793, 17916.977, 19511.6022, 19750.9557, 19565.7124, 19056.927, 18582.473, 18141.7479, 17627.1412, 17043.401, 16395.275, 15687.5111, 14924.8572, 13831.6897, 12867.1606, 12002.1707, 11207.6209, 10454.4119, 9713.44452, 8955.61959, 7851.13964, 7110.68594},
        {35.1893876, 36.4685747, 37.5622973, 38.4765531, 39.2173402, 39.7906563, 40.2024994, 40.4588674, 40.5657581, 40.5291695, 40.3550994, 40.0495458, 39.6185065, 39.0679794, 38.4039624, 37.6324534, 36.7594502, 35.7909509, 34.7329532, 33.591455, 32.3724543, 31.0819489, 29.7259368, 28.3104157, 26.8413837, 25.3248385, 23.7667782, 22.1732005, 20.5501033, 18.9034847, 17.2393423, 15.5636742, 13.8824782, 12.2017522, 10.6399178, 9.14439636, 7.66125797, 6.19285218, 4.74152852, 3.30963651, 1.89952568, 0.513545559, -0.845954335, -2.17662447, -3.47611532, -4.74207737, -5.97216108, -7.16401693, -8.31529539, -9.42364694, -10.4867221, -11.383259, -12.2323983, -13.0343682, -13.7893968, -14.4977125, -15.1595435, -15.7751181, -16.3446645, -16.8684111, -17.3465859, -17.7794174, -18.1671337, -18.5099631, -18.8081339, -19.0618744, -19.2714127, -19.4427255, -19.5816515, -19.6765634, -19.7273166, -19.7337668, -19.6957697, -19.6131808, -19.4858557, -19.3136502, -19.0964196, -18.8340198, -18.5263063, -18.1731347, -17.7743606, -17.3298396, -16.8394274, -16.3029796, -15.3319133, -14.1284738, -12.8995747, -11.6576773, -10.4152427, -9.18473219, -7.97860685, -6.8093279, -5.68935652, -4.6311539, -3.64718121, -2.74989964, -1.95177038, -1.2652546, -0.702813487, -0.276908227, 0, -0.225248685, -0.682439882, -1.36327002, -2.25943554, -3.36263286, -4.66455842, -6.15690864, -7.83137996, -9.67966882, -11.6934716, -13.8644848, -16.1844049, -18.6449281, -21.2377511, -23.9545702, -26.7870818, -29.7742756, -32.9146041, -36.0927911, -39.3106674, -42.5700635, -45.8728102, -49.2207382, -52.6156782, -56.0594607, -59.5539166, -63.1008764, -66.7021709, -70.3596307, -74.1146318, -77.9149597, -81.7505814, -85.6114641, -89.7239784, -93.9598896, -98.1279221, -102.147122, -105.936535, -109.415207, -112.502185, -115.369038, -117.605433, -119.08101, -119.665407, -119.228264, -118.795071, -116.790653, -112.795687, -106.39085, -97.1568185, -84.8079934, -68.6278866, -49.3915752, -26.7332031, 2.04249144, 38.5849145, 84.5434722, 141.527798, 211.23611, 295.410015, 393.048259, 511.616595, 655.572185, 829.715331, 1039.42776, 1289.82431, 1579.40424, 1911.56723, 2313.97016, 2800.54866, 3355.58088, 4027.22264, 4839.67563, 5817.14156, 6983.82212, 9115.1343, 11296.2607, 13363.5991, 15153.5475, 16502.5036, 16652.1941, 16430.1358, 15927.5855, 15467.7891, 15050.1748, 14577.5086, 14054.052, 13484.0661, 12871.8121, 12221.5514, 11152.1747, 10321.2141, 9657.71583, 9090.72586, 8549.29025, 7962.45505, 7259.26633, 6294.69517, 5709.04092},
        {35.0722149, 36.778297, 38.2702103, 39.5547751, 40.6388113, 41.529139, 42.2325782, 42.7559491, 43.1060717, 43.289766, 43.3138522, 43.1851502, 42.9104802, 42.4966622, 41.9505163, 41.2788625, 40.488521, 39.5863117, 38.5790548, 37.4735703, 36.2766782, 34.9951987, 33.6359518, 32.2057575, 30.711436, 29.1598073, 27.5576914, 25.9119084, 24.2292785, 22.5166216, 20.7807578, 19.0285072, 17.2666898, 15.5021258, 13.8781953, 12.330566, 10.7932641, 9.26867811, 7.7591964, 6.26720739, 4.79509951, 3.34526119, 1.92008086, 0.521946929, -0.846752172, -2.18362802, -3.48629219, -4.75235626, -5.97943181, -7.1651304, -8.30706362, -9.30075637, -10.2481199, -11.1489789, -12.003158, -12.8104817, -13.5707747, -14.2838615, -14.9495669, -15.5677155, -16.1381317, -16.6606404, -17.135066, -17.5612331, -17.9389665, -18.2680907, -18.5484304, -18.7724827, -18.932921, -19.0446818, -19.1080649, -19.1233698, -19.0908963, -19.0109438, -18.883812, -18.7098005, -18.489209, -18.222337, -17.9094842, -17.5509502, -17.1470346, -16.698037, -16.2042571, -15.6659944, -14.7158984, -13.5458267, -12.352056, -11.146817, -9.9423402, -8.75085616, -7.58459548, -6.45578872, -5.37666648, -4.35945932, -3.41639782, -2.55971256, -1.80163411, -1.15439305, -0.630219965, -0.241345422, 0, -0.919341557, -1.94088577, -3.06500155, -4.29205783, -5.62242352, -7.05646753, -8.59455879, -10.2370662, -11.9843587, -13.8368052, -15.7947746, -17.8586359, -20.0287579, -22.3055096, -24.6892598, -27.1803776, -29.7397385, -32.3620694, -35.1370696, -38.056645, -41.1127017, -44.2971458, -47.6018833, -51.0188203, -54.5398628, -58.1569169, -61.8618887, -65.6466843, -69.5032096, -73.347679, -77.2754439, -81.301118, -85.4393148, -90.0767725, -95.0420423, -100.064442, -105.046949, -109.892538, -114.504186, -118.784869, -122.474144, -125.688143, -128.361815, -130.430109, -131.827974, -133.962907, -134.929173, -134.293586, -131.622959, -126.484104, -118.339045, -106.987462, -93.0017475, -76.1225449, -54.3923394, -26.6160546, 8.40138617, 51.5675427, 104.430353, 168.85141, 244.179349, 336.551677, 450.130312, 588.875892, 756.407995, 956.502753, 1187.01075, 1450.17186, 1769.95296, 2158.03092, 2595.73334, 3129.75544, 3782.2796, 4575.48818, 5531.56357, 7334.60907, 9179.40587, 10922.6561, 12421.0619, 13531.3254, 13580.8725, 13308.5323, 12797.8399, 12347.5643, 11956.0436, 11531.9345, 11078.0591, 10597.2391, 10092.2967, 9566.05361, 8542.59229, 7860.01976, 7403.9822, 7060.12582, 6714.09679, 6251.54132, 5558.1056, 4816.67584, 4357.96906},
        {33.2162295, 35.7823797, 38.0707372, 40.0901586, 41.8495007, 43.3576201, 44.6233734, 45.6556173, 46.4632084, 47.0550034, 47.4398589, 47.6266317, 47.6241782, 47.4413553, 47.0870196, 46.5700276, 45.899236, 45.0835016, 44.131681, 43.0526307, 41.8552075, 40.548268, 39.1406689, 37.6412667, 36.0589183, 34.4024802, 32.680809, 30.9027614, 29.0771942, 27.2129638, 25.3189271, 23.4039405, 21.4768609, 19.5465447, 17.8321898, 16.233058, 14.645883, 13.0726956, 11.5155264, 9.97640617, 8.4573655, 6.96043512, 5.48764569, 4.0410279, 2.62261243, 1.23442997, -0.121488809, -1.44311322, -2.72841259, -3.97535623, -5.18191345, -6.28002897, -7.335128, -8.34661114, -9.31387901, -10.2363322, -11.1133713, -11.9443969, -12.7288097, -13.4660102, -14.155399, -14.7963768, -15.3883441, -15.9307016, -16.4228498, -16.8641893, -17.2541208, -17.5775317, -17.8196579, -18.0098323, -18.1483964, -18.2356915, -18.2720591, -18.2578406, -18.1933775, -18.0790112, -17.9150831, -17.7019347, -17.4399073, -17.1293425, -16.7705816, -16.3639662, -15.9098375, -15.4085371, -14.4863412, -13.3384918, -12.1647268, -10.9775269, -9.78937247, -8.61274421, -7.46012265, -6.34398835, -5.27682188, -4.27110381, -3.33931471, -2.49393513, -1.74744565, -1.11232683, -0.601059241, -0.226123441, 0, -2.48746551, -4.78009527, -6.89912118, -8.86577512, -10.701289, -12.4268946, -14.0638239, -15.6333087, -17.156581, -18.6548725, -20.1494153, -21.6614411, -23.2121818, -24.8228694, -26.5147357, -28.3090126, -29.661294, -30.5120063, -32.1618006, -34.5107008, -37.4587306, -40.9059136, -44.7522737, -48.8978345, -53.2426199, -57.6866535, -62.1299591, -66.4725605, -70.6144813, -73.7296374, -76.7104, -79.6746252, -82.7401692, -86.6507918, -91.2113973, -96.0599823, -101.126632, -106.34143, -111.634463, -116.935816, -122.454547, -127.756862, -132.718264, -137.214257, -141.120343, -144.746185, -147.424586, -148.922511, -149.006921, -147.444782, -144.166574, -138.575886, -130.476089, -119.541291, -105.18169, -86.9259739, -64.3028318, -39.1331431, -8.13245233, 34.4842685, 94.4502561, 168.82021, 251.25341, 342.125614, 453.193284, 590.988949, 762.151561, 973.241241, 1230.42791, 1540.10221, 1863.69521, 2259.47397, 2749.39208, 3355.4031, 4099.46059, 5595.03691, 7124.68716, 8562.48522, 9782.50496, 10658.8203, 10586.0944, 10237.419, 9692.3294, 9249.94726, 8902.88864, 8552.27547, 8195.29855, 7829.14867, 7451.01664, 7058.09324, 6181.04884, 5624.40663, 5277.80568, 5030.88507, 4773.28387, 4394.64117, 3784.59602, 3553.18484, 3171.4667},
        {36.6432225, 39.4459311, 41.9597539, 44.1936276, 46.1564889, 47.8572745, 49.304921, 50.5083651, 51.4765436, 52.218393, 52.74285, 53.0588514, 53.1753338, 53.1012338, 52.8454883, 52.4170337, 51.8248069, 51.0777445, 50.1847831, 49.1548595, 47.9969104, 46.7198723, 45.332682, 43.8442762, 42.2635915, 40.5995646, 38.8611322, 37.057231, 35.1967976, 33.2887688, 31.3420811, 29.3656714, 27.3684762, 25.3594322, 23.5378941, 21.8135845, 20.0939371, 18.3817091, 16.6796579, 14.9905406, 13.3171145, 11.6621368, 10.0283649, 8.41855588, 6.83546705, 5.28185566, 3.76047896, 2.27409417, 0.825458565, -0.582670623, -1.94753615, -3.15843217, -4.32289016, -5.44049296, -6.51082345, -7.5334645, -8.50799897, -9.43400973, -10.3110797, -11.1387916, -11.9167284, -12.644473, -13.3216082, -13.9477169, -14.522382, -15.0451863, -15.5157127, -15.9249806, -16.2642149, -16.5506602, -16.7844543, -16.9657352, -17.0946411, -17.1713099, -17.1958795, -17.1684881, -17.0892735, -16.9583738, -16.7759271, -16.5420712, -16.2569443, -15.9206842, -15.5334291, -15.0953169, -14.2277983, -13.1283202, -11.9988809, -10.8519075, -9.6998272, -8.55506711, -7.43005436, -6.33721608, -5.28897942, -4.29777149, -3.37601944, -2.5361504, -1.7905915, -1.15176988, -0.632112667, -0.244046996, 0, -1.54253487, -3.0549122, -4.5477035, -6.03148025, -7.51681397, -9.01427614, -10.5344383, -12.0878718, -13.6851484, -15.3368393, -17.0535162, -18.8457506, -20.7241139, -22.6991776, -24.7815133, -26.9816923, -28.9983713, -30.7975623, -33.095359, -35.8354941, -38.9616999, -42.4177092, -46.1472544, -50.0940682, -54.2018829, -58.4144314, -62.6754459, -66.9286592, -71.1178038, -74.7630249, -78.3869577, -82.060411, -85.8541936, -90.1532381, -94.8712912, -99.8383336, -105.030937, -110.425673, -115.999113, -121.72783, -127.654301, -133.669134, -139.736003, -145.818585, -151.880557, -160.068072, -167.616712, -173.944533, -178.469593, -180.609951, -178.13999, -174.130376, -172.053281, -172.751248, -171.697996, -166.777734, -155.874674, -132.105448, -99.2053848, -66.1108104, -43.0652258, -17.8146307, 15.8603547, 62.9532825, 126.380608, 210.012207, 317.300362, 452.006685, 619.423967, 823.978881, 1041.96352, 1310.26753, 1643.05647, 2054.4959, 2558.75137, 3574.63071, 4610.49666, 5579.35422, 6394.20839, 6968.06416, 6877.33145, 6596.00706, 6181.35103, 5853.00614, 5603.95592, 5359.52443, 5116.07818, 4869.98367, 4617.60741, 4355.31592, 4104.92347, 3806.32885, 3474.19461, 3123.18334, 2767.95761, 2423.17999, 2103.51307, 1984.9796, 1579.21194},
        {32.139939, 35.5609321, 38.664182, 41.4592776, 43.9558078, 46.1633616, 48.0915278, 49.7498953, 51.1480531, 52.2955899, 53.2020949, 53.8771567, 54.3303644, 54.5713069, 54.6095729, 54.4547516, 54.1164316, 53.604202, 52.9276517, 52.0963695, 51.1199444, 50.0079652, 48.7700208, 47.4157003, 45.9545923, 44.396286, 42.75037, 41.0264335, 39.2340652, 37.3828541, 35.482389, 33.5422589, 31.5720527, 29.5813593, 27.78148, 26.0769047, 24.369699, 22.6629058, 20.959568, 19.2627286, 17.5754306, 15.9007168, 14.2416302, 12.6012137, 10.9825104, 9.38856304, 7.82241472, 6.28710832, 4.7856868, 3.3211931, 1.89667015, 0.624687976, -0.603611883, -1.78756081, -2.9264902, -4.01973144, -5.06661592, -6.06647502, -7.01864013, -7.92244265, -8.77721395, -9.58228543, -10.3369885, -11.0406545, -11.6926148, -12.2922009, -12.8387441, -13.3152051, -13.7049375, -14.0410362, -14.3238938, -14.553903, -14.7314563, -14.8569465, -14.9307662, -14.953308, -14.9249645, -14.8461284, -14.7171923, -14.5385488, -14.3105907, -14.0337105, -13.7083008, -13.3347544, -12.5911802, -11.6458916, -10.6710752, -9.67758255, -8.67626492, -7.67797367, -6.69356017, -5.73387578, -4.80977185, -3.93209974, -3.1117108, -2.35945641, -1.68618791, -1.10275666, -0.620014016, -0.248811345, 0, -0.750259283, -1.56652565, -2.45301413, -3.41393975, -4.45351754, -5.57596254, -6.78548978, -8.08631428, -9.48265108, -10.9787152, -12.5787217, -14.2868856, -16.1074218, -18.0445456, -20.1024718, -22.2854155, -24.6385603, -27.1719739, -29.7972043, -32.5272455, -35.3750917, -38.3537367, -41.4761747, -44.7553995, -48.2044051, -51.8361857, -55.6637351, -59.7000473, -63.9581165, -68.5237872, -73.3104908, -78.3093662, -83.5115521, -88.8799084, -94.4197134, -100.143786, -106.051749, -112.143225, -118.417836, -124.875205, -131.431653, -138.195455, -145.182533, -152.408807, -159.890198, -169.147151, -178.314931, -187.033326, -194.942126, -201.681119, -205.828864, -209.383442, -214.257968, -220.973701, -227.210194, -231.922869, -234.067148, -230.773368, -223.236824, -214.643813, -208.981651, -200.680549, -186.734888, -164.659114, -132.848903, -89.2954297, -33.0230541, 37.7091856, 128.430583, 242.527538, 369.979737, 529.642861, 729.544094, 977.710621, 1282.16962, 1888.85354, 2508.408, 3089.38385, 3580.33197, 3929.80324, 3890.68153, 3737.62936, 3503.05488, 3307.59958, 3148.74593, 2987.21246, 2822.31981, 2653.38862, 2479.73954, 2300.6932, 2427.53736, 2307.72336, 2024.49048, 1661.07797, 1300.7251, 1026.67112, 922.155289, 741.435195, 452.763737},
        {31.6116123, 35.7658006, 39.5787693, 43.0601093, 46.2194113, 49.0662663, 51.6102651, 53.8609986, 55.8280577, 57.5210333, 58.9495161, 60.1230971, 61.0513671, 61.7439171, 62.2103378, 62.4602202, 62.5031551, 62.3487334, 62.0065459, 61.4861835, 60.7972372, 59.9492976, 58.9519558, 57.8148026, 56.5474289, 55.1594254, 53.6603832, 52.059893, 50.3675458, 48.5929323, 46.7456435, 44.8352703, 42.8714034, 40.8636338, 38.9743304, 37.1334814, 35.2692391, 33.3862363, 31.4891061, 29.5824813, 27.6709948, 25.7592797, 23.8519687, 21.9536949, 20.0690912, 18.2027904, 16.3594256, 14.5436297, 12.7600356, 11.0132762, 9.30798441, 7.83212198, 6.40301886, 5.02133376, 3.6877254, 2.4028525, 1.16737379, -0.0180520166, -1.1527662, -2.23611003, -3.2674248, -4.24605178, -5.17133226, -6.0426075, -6.85921879, -7.62050741, -8.32581464, -8.95546958, -9.49025782, -9.96873182, -10.3914653, -10.7590321, -11.0720059, -11.3309606, -11.5364697, -11.6891072, -11.7894468, -11.8380623, -11.8355273, -11.7824157, -11.6793012, -11.5267576, -11.3253587, -11.0756781, -10.5085101, -9.76499261, -8.98950626, -8.19137968, -7.3799415, -6.56452036, -5.75444487, -4.95904368, -4.18764541, -3.44957868, -2.75417214, -2.11075441, -1.52865413, -1.01719991, -0.5857204, -0.243544219, 0, -0.631595761, -1.2960111, -1.9998198, -2.74959563, -3.55191239, -4.41334386, -5.34046381, -6.33984602, -7.41806428, -8.58169237, -9.83730407, -11.1914732, -12.6507734, -14.2217787, -15.9110626, -17.7251991, -19.5674166, -21.4295254, -23.5518563, -25.9188377, -28.5148979, -31.3244653, -34.3319683, -37.5218353, -40.8784946, -44.3863747, -48.0299038, -51.7935104, -55.6616229, -59.4590448, -63.3883584, -67.4818795, -71.7719239, -76.307994, -81.1138123, -86.2085182, -91.6192715, -97.3732321, -103.49756, -110.019414, -117.055912, -124.516878, -132.411872, -140.750454, -149.542183, -159.215084, -169.255636, -179.568781, -190.059465, -200.632629, -210.842043, -221.373039, -234.329092, -250.067529, -265.380635, -278.661323, -288.30251, -285.514935, -277.506002, -279.318215, -309.240499, -347.656062, -378.518536, -395.043901, -406.142671, -413.521313, -420.79133, -430.153088, -436.821811, -439.963915, -444.581707, -443.107881, -432.688736, -410.470571, -373.599686, -254.965588, -132.035565, -18.0201138, 73.8702698, 130.425089, 86.1474082, 14.9707959, -73.9067704, -141.837052, -191.279466, -236.964374, -280.737645, -324.445151, -369.932762, -419.046349, -897.821028, -1085.49787, -1070.24367, -940.225247, -783.6094, -688.562938, -743.252665, -754.413519, -809.763205}
    };
    
    const double b4tab[98][200] = {{-186.925396, -184.352525, -181.842459, -179.392881, -177.001471, -174.665911, -172.383881, -170.153063, -167.971138, -165.835786, -163.74469, -161.695529, -159.685985, -157.713739, -155.776473, -153.871866, -151.997601, -150.151359, -148.330819, -146.533665, -144.757576, -143.000233, -141.259319, -139.532513, -137.817497, -136.111952, -134.413559, -132.72, -131.028954, -129.338104, -127.64513, -125.947714, -124.243536, -122.530278, -120.680754, -118.757705, -116.825372, -114.885489, -112.939788, -110.990004, -109.037869, -107.085117, -105.133481, -103.184693, -101.240488, -99.3025987, -97.3727579, -95.4526991, -93.5441555, -91.6488604, -89.768547, -88.0226562, -86.2926618, -84.5777456, -82.8770889, -81.1898735, -79.5152807, -77.8524923, -76.2006897, -74.5590546, -72.9267684, -71.3030128, -69.6869693, -68.0778195, -66.4747449, -64.8769271, -63.2835477, -61.7134664, -60.1850706, -58.6569569, -57.1270313, -55.5931997, -54.053368, -52.505442, -50.9473277, -49.3769311, -47.7921579, -46.1909141, -44.5711056, -42.9306384, -41.2674182, -39.5793511, -37.8643428, -36.1202994, -34.0976577, -31.9232632, -29.726936, -27.5146128, -25.2922304, -23.0657256, -20.841035, -18.6240954, -16.4208437, -14.2372165, -12.0791505, -9.95258263, -7.86344952, -5.81768795, -3.82123468, -1.88002644, 0, 0.0434581599, 0.272749192, 0.675186147, 1.23808208, 1.94875003, 2.79450306, 3.76265422, 4.84051656, 6.01540312, 7.27462697, 8.60550115, 9.99533872, 11.4314527, 12.9011562, 14.3917622, 15.8905838, 17.5127663, 19.2638843, 20.8421068, 22.2621394, 23.5386878, 24.6864575, 25.7201544, 26.654484, 27.504152, 28.2838641, 29.008326, 29.6922432, 30.3503215, 31.2052016, 31.9874112, 32.6492755, 33.1431197, 33.5715277, 33.8116952, 33.7007491, 33.1459369, 32.0545062, 30.3337047, 27.8907797, 25.3697236, 21.7168125, 16.6951485, 10.0678332, 1.59796868, -7.67457687, -19.5826596, -34.6823693, -53.5297957, -76.6810286, -103.124037, -136.899623, -179.306522, -231.912595, -296.835178, -375.944902, -471.112401, -578.590857, -707.145046, -871.667868, -1078.66356, -1339.33343, -1651.40524, -2025.67476, -2495.08082, -3082.39822, -3780.68292, -4605.00483, -5679.40299, -7066.27733, -8436.54447, -10304.3159, -12867.5049, -16324.025, -20871.7896, -32431.6495, -44047.8462, -54487.5588, -62517.9663, -66906.2476, -61201.5804, -52867.8127, -42908.4099, -35170.5652, -29532.1018, -24685.9559, -20569.1951, -17118.8867, -14272.098, -11965.8966, -10141.5357, -8733.36197, -7674.88453, -6899.61269, -6341.05569, -5932.72281, -5608.12327, -5341.81396, -5213.41287},
        {-168.555271, -165.78756, -163.103997, -160.50192, -157.978668, -155.531577, -153.157987, -150.855235, -148.620659, -146.451596, -144.345386, -142.299365, -140.310872, -138.377245, -136.495821, -134.663939, -132.878937, -131.138151, -129.438922, -127.778585, -126.15448, -124.563944, -123.004316, -121.472932, -119.967132, -118.484252, -117.021632, -115.576608, -114.146519, -112.728703, -111.320498, -109.919241, -108.522271, -107.126926, -105.648744, -104.127684, -102.604687, -101.079744, -99.5528491, -98.0239942, -96.4931718, -94.9603743, -93.4255943, -91.8888242, -90.3500565, -88.8092836, -87.2664981, -85.7216923, -84.1748589, -82.6259902, -81.0750787, -79.4954129, -77.9142682, -76.3322158, -74.7498273, -73.1676739, -71.586327, -70.0063579, -68.4283381, -66.8528389, -65.2804317, -63.7116878, -62.1471785, -60.5874753, -59.0331495, -57.4847725, -55.9429156, -54.4463975, -53.0331186, -51.6247681, -50.2194378, -48.8152197, -47.4102058, -46.0024881, -44.5901583, -43.1713086, -41.7440308, -40.3064169, -38.8565588, -37.3925484, -35.9124778, -34.4144389, -32.8965235, -31.3568237, -29.6129418, -27.7570108, -25.8833332, -23.9958582, -22.0985351, -20.1953129, -18.290141, -16.3869683, -14.4897442, -12.6024178, -10.7289383, -8.87325483, -7.03931664, -5.23107288, -3.45247273, -1.70746538, 0, 0.630126151, 1.36372938, 2.19344837, 3.11192183, 4.11178846, 5.18568694, 6.32625598, 7.52613428, 8.77796052, 10.0743734, 11.4080116, 12.7715139, 14.1575189, 15.5586653, 16.9675919, 18.3769373, 19.8429884, 21.3674765, 22.7990742, 24.144059, 25.4087086, 26.5993005, 27.7221123, 28.7834217, 29.7895061, 30.7466433, 31.6611109, 32.5391863, 33.3871473, 34.3637756, 35.2669264, 36.057126, 36.6949007, 37.4099502, 38.0282141, 38.3038525, 38.1166398, 37.3463503, 35.8727582, 33.575638, 30.7657188, 26.7606599, 21.3559184, 14.3469512, 5.52921545, -3.77444403, -15.6778049, -30.767257, -49.6291902, -72.8499945, -99.2618331, -133.349377, -175.54819, -227.467444, -293.112433, -375.412641, -477.297552, -598.925376, -746.626706, -929.755344, -1146.94357, -1419.83068, -1753.22308, -2162.30443, -2679.84198, -3330.53185, -4106.24175, -5023.15671, -6217.83249, -7758.73637, -9377.18534, -11530.6662, -14404.3522, -18183.4163, -23053.0318, -34764.1587, -46544.7365, -57188.4916, -65489.1504, -70240.4395, -65215.3108, -57575.448, -48266.3377, -40825.8243, -35173.2531, -30147.8204, -25722.8792, -21871.7821, -18567.882, -15784.5317, -13603.5652, -11864.2351, -10511.2725, -9489.40856, -8743.37446, -8217.90137, -7857.72043, -7748.75498, -8029.84348},
        {-153.482925, -151.33362, -149.206595, -147.101835, -145.019324, -142.959046, -140.920988, -138.905133, -136.911467, -134.939974, -132.990638, -131.063445, -129.15838, -127.275426, -125.41457, -123.575795, -121.759086, -119.964429, -118.191808, -116.441207, -114.712611, -113.006006, -111.321376, -109.658706, -108.01798, -106.399183, -104.8023, -103.227316, -101.674216, -100.142984, -98.6336045, -97.1460631, -95.6803444, -94.2364329, -92.8746705, -91.5635938, -90.2710142, -88.9949578, -87.7334507, -86.4845188, -85.2461883, -84.0164852, -82.7934355, -81.5750654, -80.3594009, -79.1444681, -77.928293, -76.7089016, -75.4843201, -74.2525745, -73.0116909, -71.6402925, -70.2583966, -68.8666177, -67.4655703, -66.0558688, -64.6381277, -63.2129616, -61.7809849, -60.342812, -58.8990575, -57.4503359, -55.9972615, -54.540449, -53.0805127, -51.6180672, -50.153727, -48.7001889, -47.2698597, -45.8384349, -44.4057456, -42.9716232, -41.5358988, -40.0984036, -38.6589689, -37.217426, -35.773606, -34.3273401, -32.8784597, -31.4267958, -29.9721799, -28.514443, -27.0534164, -25.5889314, -24.0862507, -22.563217, -21.0380882, -19.5118174, -17.9853575, -16.4596616, -14.9356827, -13.414374, -11.8966883, -10.3835789, -8.87599858, -7.37490052, -5.88123774, -4.39596325, -2.92003011, -1.45439135, 0, 1.40819313, 2.79867414, 4.17231018, 5.52996845, 6.87251608, 8.20082027, 9.51574817, 10.8181669, 12.1089438, 13.3889458, 14.6590403, 15.9200942, 17.1729749, 18.4185495, 19.6576852, 20.891249, 22.0341193, 23.0748791, 24.2088943, 25.418606, 26.686455, 27.9948824, 29.3263291, 30.663236, 31.9880443, 33.2831947, 34.5311282, 35.7142859, 36.8151087, 37.8440576, 38.7452795, 39.4928093, 40.0606821, 40.6852645, 41.2094257, 41.4060795, 41.1705612, 40.3982064, 38.9843502, 36.8243282, 34.7167793, 31.3788169, 26.5290429, 19.8860596, 11.1684689, 0.372006086, -13.1311432, -29.6916601, -49.6602258, -73.3875217, -98.2334846, -131.194894, -173.29869, -226.742986, -296.117043, -384.936547, -496.717182, -633.216505, -800.107904, -1004.98272, -1243.38475, -1541.94018, -1905.69005, -2352.43392, -2921.58992, -3642.65289, -4509.88213, -5543.63735, -6893.47525, -8635.8714, -10530.3323, -13019.067, -16288.2713, -20524.1407, -25912.871, -38401.1863, -50974.6215, -62379.2402, -71361.1058, -76666.2818, -71876.0932, -64344.5011, -55031.0282, -47488.3277, -41649.4752, -36367.0749, -31628.2259, -27420.0272, -23729.5776, -20543.9762, -17797.9503, -15644.7736, -13983.9277, -12714.8943, -11737.1551, -10950.1917, -10253.4859, -9989.70525, -10046.4791},
        {-140.878578, -138.952874, -137.059274, -135.196834, -133.364612, -131.561665, -129.78705, -128.039824, -126.319044, -124.623768, -122.953053, -121.305955, -119.681532, -118.078841, -116.496939, -114.934883, -113.39173, -111.866538, -110.358363, -108.866263, -107.389295, -105.926515, -104.476982, -103.039752, -101.613882, -100.198429, -98.7924512, -97.3950048, -96.0051471, -94.6219353, -93.2444265, -91.8716779, -90.5027465, -89.1366894, -87.7326791, -86.310554, -84.8906318, -83.4732641, -82.0588022, -80.6475977, -79.240002, -77.8363665, -76.4370428, -75.0423822, -73.6527362, -72.2684564, -70.8898941, -69.5174008, -68.1513279, -66.792027, -65.4398495, -64.1397755, -62.8465604, -61.5595881, -60.2782426, -59.001908, -57.7299682, -56.4618072, -55.196809, -53.9343577, -52.6738371, -51.4146314, -50.1561244, -48.8977002, -47.6387429, -46.3786363, -45.1167645, -43.8417309, -42.5423979, -41.2403835, -39.9357705, -38.6286418, -37.3190803, -36.0071687, -34.69299, -33.376627, -32.0581625, -30.7376795, -29.4152607, -28.090989, -26.7649473, -25.4372184, -24.1078851, -22.7770304, -21.4487477, -20.1210301, -18.7918227, -17.4610781, -16.1287492, -14.7947885, -13.4591488, -12.1217827, -10.7826431, -9.44168251, -8.09885373, -6.75410945, -5.40740236, -4.05868517, -2.70791058, -1.35503129, 0, 1.54240269, 3.05507522, 4.53977032, 5.99824076, 7.43223928, 8.84351863, 10.2338316, 11.6049308, 12.9585691, 14.2964993, 15.620474, 16.9322461, 18.2335682, 19.5261932, 20.8118737, 22.0923626, 23.3181876, 24.4837836, 25.7067694, 26.9779208, 28.2880139, 29.6278246, 30.9881289, 32.3597029, 33.7333224, 35.0997634, 36.4498021, 37.7742142, 39.0637758, 40.4392753, 41.7138051, 42.8391373, 43.7670443, 44.5759574, 45.1543194, 45.3567971, 45.0971649, 44.2891974, 42.8466691, 40.6833544, 38.8517722, 35.7803783, 31.1601493, 24.682062, 16.0370929, 6.68244398, -5.89868958, -22.4568873, -43.7427288, -70.5067937, -101.83272, -142.175401, -190.60369, -248.669125, -322.992062, -417.917064, -537.788692, -685.06462, -866.405135, -1090.52895, -1353.89933, -1684.21865, -2080.26861, -2561.53741, -3182.59919, -3981.92305, -4967.41742, -6169.62826, -7731.15739, -9731.22149, -11803.2407, -14540.8123, -18177.4673, -22946.7364, -29082.1505, -43590.0773, -58238.0018, -71566.2455, -82115.1301, -88424.9772, -83027.1229, -74476.8651, -63889.8049, -55463.5258, -49093.6779, -43391.7519, -38317.6059, -33831.0977, -29892.0854, -26460.4269, -23005.79, -20352.0952, -18330.4233, -16771.8552, -15507.4717, -14368.3537, -13185.5821, -12539.9599, -12114.3284},
        {-132.327653, -130.441789, -128.60339, -126.810647, -125.061752, -123.354895, -121.688267, -120.060059, -118.468462, -116.911667, -115.387865, -113.895247, -112.432004, -110.996327, -109.586406, -108.200433, -106.836599, -105.493095, -104.168111, -102.859839, -101.566469, -100.286193, -99.0172012, -97.7576848, -96.5058348, -95.259842, -94.0178975, -92.7781922, -91.538917, -90.298263, -89.0544211, -87.8055821, -86.5499372, -85.2856771, -83.9074221, -82.4673268, -81.0187912, -79.5633671, -78.1026068, -76.6380621, -75.1712851, -73.7038279, -72.2372425, -70.7730809, -69.3128952, -67.8582373, -66.4106594, -64.9717134, -63.5429514, -62.1259254, -60.7221874, -59.443624, -58.1790609, -56.9276582, -55.6885762, -54.4609751, -53.244015, -52.0368562, -50.8386588, -49.648583, -48.4657891, -47.2894373, -46.1186876, -44.9527004, -43.7906358, -42.631654, -41.4749153, -40.3111971, -39.1314782, -37.9522074, -36.7730884, -35.5938248, -34.4141202, -33.2336782, -32.0522024, -30.8693964, -29.6849639, -28.4986083, -27.3100335, -26.1189429, -24.9250402, -23.7280289, -22.5276128, -21.3234954, -20.1133199, -18.8978636, -17.6779282, -16.4532843, -15.2237023, -13.9889527, -12.748806, -11.5030326, -10.251403, -8.99368766, -7.7296571, -6.45908176, -5.18173213, -3.89737866, -2.60579183, -1.30674212, 0, 1.49729896, 2.97288603, 4.42833015, 5.86520026, 7.28506531, 8.68949426, 10.080056, 11.4583196, 12.8258539, 14.1842279, 15.5350105, 16.8797707, 18.2200774, 19.5574995, 20.8936061, 22.2299661, 23.5437843, 24.8331491, 26.1547386, 27.504901, 28.8799843, 30.2763365, 31.6903058, 33.1182403, 34.5564879, 36.0013969, 37.4493152, 38.8965909, 40.3395722, 41.9775083, 43.529449, 44.9308719, 46.1172548, 47.1334377, 47.8602171, 48.1492262, 47.9031339, 47.0246092, 45.4163209, 42.980938, 40.7438624, 37.1433288, 31.862341, 24.5839027, 14.9910178, 5.38827188, -7.8183085, -25.6011151, -48.9325395, -78.7849734, -114.846357, -160.943419, -215.715178, -280.635921, -362.968595, -467.377158, -598.525567, -756.666556, -951.877856, -1199.04944, -1499.95705, -1880.57786, -2335.76653, -2885.92905, -3594.76664, -4506.10729, -5626.41034, -6989.8156, -8767.48133, -11053.0604, -13213.9735, -16181.8337, -20301.682, -25918.5592, -33377.5064, -52178.1702, -71222.3343, -88566.3882, -102266.721, -110379.723, -102881.584, -91295.6919, -77141.3782, -66321.7119, -58620.7956, -51996.1746, -46323.2795, -41477.5407, -37334.3886, -33769.2535, -29836.1352, -26668.7884, -24116.4655, -22028.4187, -20253.9003, -18642.1625, -17042.4577, -15968.3614, -15048.7111},
        {-127.695887, -126.217184, -124.723581, -123.216295, -121.696543, -120.165545, -118.624517, -117.074679, -115.517247, -113.953439, -112.384475, -110.81157, -109.235945, -107.658816, -106.081401, -104.504918, -102.930586, -101.359622, -99.7932443, -98.2326706, -96.679119, -95.1338073, -93.5979536, -92.0727757, -90.5594917, -89.0593195, -87.573477, -86.1031821, -84.6496528, -83.2141071, -81.7977628, -80.401838, -79.0275506, -77.6761184, -76.450175, -75.2980973, -74.1670436, -73.0549406, -71.9597152, -70.8792941, -69.8116043, -68.7545725, -67.7061256, -66.6641903, -65.6266935, -64.591562, -63.5567226, -62.5201021, -61.4796274, -60.4332252, -59.3788225, -58.1864322, -56.9846678, -55.7742291, -54.5558158, -53.3301277, -52.0978645, -50.8597259, -49.6164116, -48.3686215, -47.1170551, -45.8624124, -44.605393, -43.3466966, -42.0870229, -40.8270718, -39.567543, -38.3157699, -37.0789268, -35.8440318, -34.6113546, -33.3811649, -32.1537323, -30.9293266, -29.7082174, -28.4906745, -27.2769676, -26.0673663, -24.8621404, -23.6615595, -22.4658933, -21.2754116, -20.0903841, -18.9110804, -17.7880527, -16.695372, -15.6065046, -14.5200884, -13.4347614, -12.3491616, -11.2619267, -10.1716948, -9.07710385, -7.9767917, -6.8693963, -5.7535556, -4.62790752, -3.49108998, -2.34174093, -1.1784983, 0, 1.76285546, 3.44792918, 5.0613459, 6.60923035, 8.09770727, 9.53290141, 10.9209375, 12.2679403, 13.5800345, 14.8633449, 16.1239962, 17.3681131, 18.6018205, 19.8312429, 21.0625053, 22.3017322, 23.3777128, 24.2712379, 25.3895491, 26.7007705, 28.1730263, 29.7744407, 31.4731381, 33.2372425, 35.0348782, 36.8341694, 38.6032404, 40.3102153, 41.9232184, 43.4269254, 44.7668402, 45.9061214, 46.8079279, 47.7463689, 48.5291278, 48.8809683, 48.671764, 47.7713885, 46.0497155, 43.3766186, 40.897239, 36.818058, 30.7594404, 22.3417511, 11.1853548, -2.85087352, -20.4447067, -42.0354072, -68.0622379, -98.9644614, -130.978127, -173.882972, -229.786183, -301.803142, -395.107623, -513.949222, -662.577537, -836.862929, -1051.3386, -1329.67875, -1680.055, -2127.04139, -2667.3323, -3322.34286, -4155.37624, -5210.50827, -6470.52131, -7963.6003, -9942.67314, -12533.5404, -14717.1854, -17940.3517, -22725.1228, -29593.5822, -39067.8133, -64585.3163, -90523.9039, -114176.805, -132837.25, -143798.467, -133035.457, -116705.163, -96951.486, -82265.6223, -72278.9985, -63978.3329, -57127.2872, -51489.523, -46828.7019, -42908.4856, -38824.2732, -35106.1683, -31753.7844, -28766.7347, -26144.6325, -23887.0911, -21993.7239, -20417.8735, -19050.1709},
        {-123.130142, -122.002919, -120.821326, -119.588315, -118.306838, -116.979846, -115.610292, -114.201127, -112.755304, -111.275774, -109.765489, -108.227402, -106.664464, -105.079627, -103.475842, -101.856063, -100.223241, -98.5803271, -96.9302742, -95.276034, -93.6205583, -91.9667991, -90.3177083, -88.6762378, -87.0453397, -85.4279657, -83.8270678, -82.2455981, -80.6865083, -79.1527504, -77.6472764, -76.1730381, -74.7329875, -73.3300766, -72.1732589, -71.1581525, -70.1778996, -69.2287669, -68.3070213, -67.4089296, -66.5307584, -65.6687747, -64.8192451, -63.9784364, -63.1426154, -62.308049, -61.4710038, -60.6277467, -59.7745444, -58.9076637, -58.0233714, -56.8809717, -55.7188308, -54.5383524, -53.3409402, -52.1279978, -50.9009288, -49.661137, -48.410026, -47.1489995, -45.8794611, -44.6028145, -43.3204633, -42.0338113, -40.744262, -39.4532192, -38.1620865, -36.8863693, -35.6412347, -34.3990022, -33.1601615, -31.9252019, -30.6946131, -29.4688845, -28.2485056, -27.0339659, -25.8257548, -24.624362, -23.4302769, -22.243989, -21.0659878, -19.8967628, -18.7368035, -17.5865994, -16.522546, -15.5055831, -14.4957385, -13.4910382, -12.4895086, -11.4891758, -10.4880661, -9.48420553, -8.47562044, -7.46033698, -6.43638137, -5.40177982, -4.35455852, -3.29274368, -2.21436152, -1.11743822, 0, 1.68035993, 3.2968786, 4.85548006, 6.36208834, 7.8226275, 9.24302158, 10.6291946, 11.9870706, 13.3225737, 14.6416279, 15.9501572, 17.2540856, 18.5593373, 19.8718362, 21.1975065, 22.5422721, 23.6647946, 24.5356748, 25.7201209, 27.1710723, 28.8414681, 30.6842476, 32.6523499, 34.6987143, 36.77628, 38.8379863, 40.8367723, 42.7255774, 44.4573406, 45.8453531, 47.0334066, 48.0163347, 48.7889712, 49.8209566, 50.8697212, 51.5660799, 51.7624244, 51.3111462, 50.064637, 47.8752883, 46.0997687, 42.6283694, 37.0191669, 28.8302377, 17.6196584, 2.5086056, -16.3987193, -39.4350148, -66.9329795, -99.2253118, -131.295591, -175.363447, -234.456441, -312.37655, -414.506848, -545.520531, -710.090794, -903.782404, -1142.44708, -1451.873, -1840.04259, -2335.76673, -2934.30627, -3660.29239, -4588.12264, -5769.35099, -7181.63961, -8857.38535, -11099.9216, -14059.3247, -16110.8021, -19452.3558, -24848.4399, -33063.5086, -44862.0159, -78509.248, -112893.619, -144404.375, -169430.761, -184362.025, -170601.761, -149515.3, -123914.309, -104925.142, -92068.2357, -81399.5761, -72612.823, -65401.6365, -59459.6766, -54480.6036, -50143.7678, -45656.52, -41220.2944, -37036.5252, -33306.6467, -30232.093, -28014.2984, -25565.2493, -23800.1293},
        {-117.374593, -116.384705, -115.354609, -114.286188, -113.181321, -112.04189, -110.869777, -109.666862, -108.435028, -107.176155, -105.892124, -104.584818, -103.256116, -101.907901, -100.542054, -99.1604554, -97.7649872, -96.3575305, -94.9399665, -93.5141767, -92.0820421, -90.6454441, -89.206264, -87.7663831, -86.3276825, -84.8920437, -83.4613478, -82.0374762, -80.6223101, -79.2177308, -77.8256195, -76.4478576, -75.0863263, -73.7429068, -72.5271338, -71.3847975, -70.2603944, -69.152312, -68.0589382, -66.9786607, -65.9098672, -64.8509454, -63.800283, -62.7562677, -61.7172874, -60.6817295, -59.647982, -58.6144325, -57.5794688, -56.5414785, -55.4988494, -54.3491435, -53.19376, -52.0332723, -50.8682536, -49.6992776, -48.5269175, -47.3517469, -46.174339, -44.9952675, -43.8151055, -42.6344267, -41.4538044, -40.273812, -39.0950229, -37.9180106, -36.7433485, -35.579265, -34.4338046, -33.2917531, -32.1531874, -31.018185, -29.8868228, -28.7591782, -27.6353283, -26.5153502, -25.3993211, -24.2873183, -23.1794188, -22.0757, -20.9762388, -19.8811127, -18.7903986, -17.7041738, -16.6787636, -15.6849381, -14.6927911, -13.7005745, -12.7065401, -11.7089396, -10.7060251, -9.69604814, -8.67726068, -7.6479145, -6.60626143, -5.55055328, -4.47904187, -3.38997902, -2.28161655, -1.15220627, 0, 1.17733761, 2.39078622, 3.64007033, 4.92491442, 6.24504298, 7.6001805, 8.99005147, 10.4143804, 11.8728917, 13.3653099, 14.8913596, 16.4507651, 18.043251, 19.6685417, 21.3263618, 23.0164358, 24.5718829, 25.968627, 27.5832374, 29.3797377, 31.3221511, 33.3745011, 35.500811, 37.6651042, 39.8314041, 41.9637342, 44.0261177, 45.982578, 47.7971386, 49.2775854, 50.6014662, 51.7796758, 52.8231087, 54.250409, 55.8185967, 57.1492918, 58.1010641, 58.5324835, 58.3021196, 57.2685425, 57.0332846, 55.181486, 51.2307021, 44.6984882, 35.1023995, 23.6580387, 7.76040214, -13.4974667, -41.0225241, -75.7217267, -114.899873, -167.468714, -233.473234, -315.859226, -423.494974, -562.589686, -739.352568, -953.957556, -1220.02078, -1557.74231, -1967.61078, -2488.78613, -3111.01296, -3864.52042, -4848.13642, -6129.20093, -7703.74709, -9620.62793, -12190.1547, -15574.7431, -17410.188, -20774.2369, -26703.9047, -36236.2067, -50408.1581, -92401.3127, -135572.012, -175421.137, -207449.568, -227158.185, -211675.254, -187122.68, -156875.321, -134110.619, -118323.244, -104923.453, -93610.134, -84082.1767, -76038.4695, -69177.9011, -64264.0818, -58687.9392, -52859.241, -47187.7548, -42083.2485, -37955.4896, -35214.2458, -31489.0808, -29244.2053},
        {-112.925689, -112.025613, -111.091242, -110.123999, -109.125304, -108.096579, -107.039245, -105.954723, -104.844434, -103.709799, -102.552239, -101.373176, -100.174031, -98.9562252, -97.7211791, -96.4703143, -95.2050519, -93.926813, -92.6370189, -91.3370907, -90.0284495, -88.7125165, -87.3907129, -86.0644598, -84.7351784, -83.4042899, -82.0732154, -80.743376, -79.416193, -78.0930875, -76.7754806, -75.4647936, -74.1624475, -72.8698635, -71.6542274, -70.4826947, -69.3216306, -68.1703222, -67.0280563, -65.8941201, -64.7678003, -63.6483841, -62.5351585, -61.4274103, -60.3244265, -59.2254942, -58.1299003, -57.0369318, -55.9458756, -54.8560188, -53.7666484, -52.6364539, -51.5061998, -50.3760531, -49.2461808, -48.11675, -46.9879278, -45.859881, -44.7327769, -43.6067825, -42.4820647, -41.3587907, -40.2371275, -39.1172421, -37.9993016, -36.883473, -35.7699234, -34.6603456, -33.5563959, -32.4550944, -31.3565092, -30.2607085, -29.1677605, -28.0777332, -26.9906948, -25.9067134, -24.8258572, -23.7481942, -22.6737927, -21.6027208, -20.5350465, -19.470838, -18.4101635, -17.3530911, -16.3598432, -15.3991454, -14.4390007, -13.4775249, -12.512834, -11.5430442, -10.5662713, -9.58063147, -8.58424064, -7.57521484, -6.55167008, -5.51172237, -4.45348772, -3.37508215, -2.27462166, -1.15022228, 0, 1.0035918, 2.07830525, 3.22208173, 4.43286263, 5.70858933, 7.04720322, 8.44664569, 9.90485812, 11.4197819, 12.9893584, 14.611529, 16.2842352, 18.0054182, 19.7730195, 21.5849805, 23.4392425, 25.1994677, 26.8444147, 28.6757518, 30.6626463, 32.7742654, 34.9797763, 37.2483464, 39.5491429, 41.851333, 44.1240841, 46.3365633, 48.4579379, 50.4573752, 52.1500381, 53.7155665, 55.1693288, 56.5266936, 58.3389362, 60.3534718, 62.1748069, 63.6575379, 64.6562613, 65.0255735, 64.6200711, 65.211102, 64.1531519, 60.9258006, 55.0086278, 45.8812133, 35.4129651, 20.0961778, -1.18702583, -29.5545229, -66.1241908, -108.631178, -165.710536, -236.386223, -323.548038, -437.978541, -586.916605, -777.601101, -1012.43377, -1304.58996, -1672.52191, -2111.97336, -2671.40005, -3341.94369, -4158.97364, -5232.8006, -6639.33598, -8369.03248, -10475.6456, -13326.2792, -17110.7896, -18967.0369, -22606.4115, -29275.2301, -40219.8094, -56686.466, -105563.164, -156044.161, -202965.362, -241162.669, -265471.99, -250019.82, -224157.743, -191597.125, -166584.686, -148661.548, -132979.096, -119297.936, -107378.673, -96981.9146, -87868.2653, -82094.1428, -75116.51, -67555.898, -60032.8376, -53167.8597, -47581.4953, -43894.2753, -38856.7086, -35676.871},
        {-108.709406, -107.708905, -106.698022, -105.677057, -104.646308, -103.606074, -102.556652, -101.498343, -100.431444, -99.356254, -98.2730721, -97.1821968, -96.0839266, -94.9785603, -93.8663965, -92.7477338, -91.622871, -90.4921067, -89.3557396, -88.2140682, -87.0673913, -85.9160076, -84.7602157, -83.6003142, -82.4366018, -81.2693772, -80.098939, -78.9255859, -77.7496166, -76.5713297, -75.3910239, -74.2089978, -73.0255501, -71.8409795, -70.6509211, -69.4581035, -68.2653111, -67.072994, -65.8816022, -64.6915856, -63.5033943, -62.3174783, -61.1342875, -59.954272, -58.7778818, -57.6055668, -56.4377771, -55.2749626, -54.1175734, -52.9660595, -51.8208708, -50.7209894, -49.627498, -48.5400112, -47.4581438, -46.3815104, -45.3097258, -44.2424046, -43.1791616, -42.1196115, -41.0633688, -40.0100485, -38.9592651, -37.9106333, -36.8637678, -35.8182835, -34.7737948, -33.7202625, -32.6478788, -31.5761695, -30.5053749, -29.4357358, -28.3674925, -27.3008857, -26.2361558, -25.1735435, -24.1132892, -23.0556336, -22.000817, -20.9490801, -19.9006634, -18.8558075, -17.8147528, -16.77774, -15.818307, -14.898504, -13.9795001, -13.059157, -12.1353367, -11.2059011, -10.2687121, -9.32163158, -8.36252141, -7.38924349, -6.39965972, -5.39163198, -4.36302217, -3.31169217, -2.23550389, -1.1323192, 0, 1.12648763, 2.3020783, 3.5266692, 4.80015752, 6.12244045, 7.49341517, 8.91297889, 10.3810288, 11.897462, 13.4621758, 15.0750674, 16.7360339, 18.4449725, 20.2017804, 22.0063548, 23.8585929, 25.6627363, 27.4050169, 29.3016961, 31.3321732, 33.475848, 35.7121199, 38.0203885, 40.3800533, 42.7705139, 45.1711698, 47.5614207, 49.9206659, 52.2283051, 54.429094, 56.5497788, 58.580152, 60.5100065, 62.8058573, 65.2191359, 67.3741478, 69.117669, 70.2964754, 70.757343, 70.3470477, 70.8708338, 69.6209533, 66.0610039, 59.6545827, 49.8652872, 38.6383406, 22.3353081, -0.200619328, -30.1262507, -68.598395, -112.934607, -172.823372, -247.223818, -339.323682, -460.944107, -620.030012, -824.526317, -1077.12508, -1392.21791, -1789.92683, -2263.8021, -2870.64417, -3609.66764, -4520.95342, -5713.82807, -7265.83328, -9140.28649, -11385.1157, -14467.071, -18615.9931, -20813.0084, -25036.474, -32640.7855, -44980.3385, -63409.5287, -116487.349, -171562.448, -223188.072, -265917.469, -294303.883, -281617.283, -257883.047, -226775.828, -202010.028, -183286.233, -166121.795, -150426.457, -136109.965, -123082.064, -111252.497, -104124.706, -95336.8787, -85679.7725, -75944.1439, -66920.7493, -59400.3453, -54173.6883, -47775.2501, -43060.6477},
        {-104.863058, -103.718964, -102.586412, -101.464788, -100.353478, -99.2518689, -98.1593461, -97.0752959, -95.9991046, -94.9301583, -93.8678433, -92.8115456, -91.7606515, -90.7145472, -89.6726188, -88.6342524, -87.5988344, -86.5657507, -85.5343876, -84.5041314, -83.474368, -82.4444838, -81.4138649, -80.3818974, -79.3479675, -78.3114615, -77.2717654, -76.2282655, -75.1803479, -74.1273989, -73.0688044, -72.0039508, -70.9322243, -69.8530109, -68.7129857, -67.5389993, -66.3585354, -65.1726911, -63.9825628, -62.7892475, -61.5938418, -60.3974425, -59.2011464, -58.0060502, -56.8132507, -55.6238445, -54.4389285, -53.2595994, -52.0869539, -50.9220888, -49.7661009, -48.700898, -47.645014, -46.5977936, -45.558582, -44.5267239, -43.5015643, -42.4824481, -41.4687203, -40.4597257, -39.4548094, -38.4533161, -37.4545909, -36.4579786, -35.4628242, -34.4684725, -33.4742686, -32.4631705, -31.4185296, -30.3734875, -29.3284515, -28.2838288, -27.2400267, -26.1974523, -25.1565129, -24.1176156, -23.0811677, -22.0475764, -21.017249, -19.9905926, -18.9680144, -17.9499218, -16.9367218, -15.9288217, -15.0155685, -14.1510288, -13.2882001, -12.4246036, -11.55776, -10.6851905, -9.80441585, -8.9129571, -8.00833518, -7.08807104, -6.14968564, -5.19069993, -4.20863485, -3.20101137, -2.16535043, -1.09917299, 0, 1.26111916, 2.54583947, 3.85627327, 5.19453288, 6.56273066, 7.96297893, 9.39739003, 10.8680763, 12.3771501, 13.9267237, 15.5189096, 17.1558199, 18.8395671, 20.5722635, 22.3560214, 24.1929533, 26.0343284, 27.8749958, 29.8320698, 31.8967677, 34.0603071, 36.3139052, 38.6487797, 41.0561477, 43.5272269, 46.0532346, 48.6253882, 51.2349052, 53.873003, 56.6423698, 59.3818797, 62.0493087, 64.602433, 67.3995514, 70.1981787, 72.6490238, 74.5897061, 75.857845, 76.2910599, 75.7269701, 75.917731, 74.2037412, 70.0480371, 62.9136548, 52.2636307, 40.2971948, 23.0571411, -0.677542297, -32.1278677, -72.5148472, -118.74323, -181.625724, -260.587009, -359.206289, -489.544926, -659.855963, -878.392447, -1146.2392, -1480.44663, -1905.88503, -2416.95828, -3076.17296, -3890.93647, -4905.12547, -6224.41169, -7926.10192, -9941.48923, -12310.025, -15606.5442, -20103.0385, -22845.5081, -27828.2426, -36439.923, -50069.2302, -70104.8454, -125582.589, -183332.218, -237830.629, -283554.716, -314981.376, -305478.826, -284705.095, -256183.655, -232766.924, -214285.668, -196683.593, -179985.816, -164217.456, -149403.633, -135569.464, -127433.031, -116994.693, -105264.884, -93254.0391, -81972.5926, -72430.9788, -65639.6322, -57739.6826, -51375.8778},
        {-101.405685, -100.250489, -99.1101872, -97.9840634, -96.8714015, -95.771485, -94.6835977, -93.6070232, -92.5410453, -91.4849476, -90.4380138, -89.3995276, -88.3687727, -87.3450328, -86.3275916, -85.3157326, -84.3087398, -83.3058966, -82.3064868, -81.3097941, -80.3151022, -79.3216948, -78.3288555, -77.335868, -76.342016, -75.3465833, -74.3488534, -73.3481101, -72.3436371, -71.334718, -70.3206366, -69.3006765, -68.2741214, -67.2402549, -66.1444856, -65.0141678, -63.8773038, -62.7349254, -61.5880648, -60.4377541, -59.2850251, -58.1309101, -56.976441, -55.8226499, -54.6705688, -53.5212298, -52.3756649, -51.2349061, -50.0999856, -48.9719353, -47.8517872, -46.8119754, -45.7805821, -44.7570915, -43.7409878, -42.7317552, -41.7288779, -40.73184, -39.7401258, -38.7532194, -37.7706051, -36.791767, -35.8161893, -34.8433562, -33.8727519, -32.9038606, -31.9361665, -30.956039, -29.9501624, -28.9450691, -27.9410935, -26.9385699, -25.9378328, -24.9392164, -23.9430552, -22.9496835, -21.9594356, -20.972646, -19.989649, -19.010779, -18.0363703, -17.0667573, -16.1022743, -15.1432558, -14.2845987, -13.4773669, -12.6714882, -11.864228, -11.0528521, -10.234626, -9.40681542, -8.56668596, -7.71150324, -6.8385329, -5.94504057, -5.02829188, -4.08555245, -3.11408792, -2.11116391, -1.07404606, 0, 1.06283118, 2.18681633, 3.37218167, 4.61915344, 5.92795787, 7.29882119, 8.73196964, 10.2276294, 11.7860268, 13.407388, 15.0919392, 16.8399067, 18.6515167, 20.5269955, 22.4665692, 24.4704641, 26.5443153, 28.6891216, 30.8928751, 33.156961, 35.4827646, 37.8716712, 40.325066, 42.8443343, 45.4308614, 48.0860324, 50.8112327, 53.6078476, 56.4772622, 59.592573, 62.7204934, 65.8108953, 68.8136504, 72.0689163, 75.3314221, 78.2518208, 80.6628987, 82.3974421, 83.2882372, 83.1680704, 83.7827648, 82.4698412, 78.6877959, 71.8951253, 61.5503259, 51.1552794, 35.1142505, 11.8748893, -20.1151542, -62.4082298, -112.671069, -181.09073, -266.883414, -373.644489, -513.910115, -696.202219, -929.042726, -1212.08898, -1564.74216, -2016.07394, -2561.78508, -3268.26428, -4143.36463, -5234.25995, -6654.75084, -8488.05511, -10653.6832, -13192.4102, -16745.2719, -21613.7624, -24848.107, -30501.2644, -40000.1686, -54771.7531, -76242.9518, -134346.265, -194876.667, -252134.703, -300420.912, -334035.839, -325620.439, -305574.564, -277481.434, -254446.083, -236282.683, -218837.982, -202124.52, -186154.84, -170941.483, -156496.991, -148139.358, -136647.237, -123308.88, -109412.535, -96246.4546, -85098.8889, -77258.0886, -67693.9537, -60146.0586},
        {-98.0372641, -96.8840391, -95.7456685, -94.6215069, -93.5109092, -92.41323, -91.3278243, -90.2540467, -89.1912521, -88.1387951, -87.0960306, -86.0623133, -85.036998, -84.0194395, -83.0089925, -82.0050118, -81.0068522, -80.0138684, -79.0254153, -78.0408475, -77.0595199, -76.0807872, -75.1040041, -74.1285256, -73.1537062, -72.1789008, -71.2034642, -70.2267512, -69.2481164, -68.2669147, -67.2825009, -66.2942296, -65.3014557, -64.303534, -63.2565888, -62.1824994, -61.1036646, -60.0208421, -58.9347896, -57.8462648, -56.7560253, -55.6648289, -54.5734331, -53.4825958, -52.3930745, -51.3056269, -50.2210108, -49.1399837, -48.0633035, -46.9917276, -45.9260139, -44.9150738, -43.9104673, -42.9119083, -41.9191104, -40.9317875, -39.9496534, -38.9724219, -37.9998068, -37.031522, -36.0672811, -35.106798, -34.1497865, -33.1959604, -32.2450335, -31.2967195, -30.3507324, -29.3985496, -28.4318463, -27.4673232, -26.5052278, -25.545808, -24.5893115, -23.6359859, -22.686079, -21.7398385, -20.7975121, -19.8593476, -18.9255927, -17.9964952, -17.0723026, -16.1532628, -15.2396234, -14.3316323, -13.529683, -12.7818438, -12.0349795, -11.2860878, -10.5321665, -9.77021334, -8.99722623, -8.21020291, -7.40614118, -6.58203884, -5.73489368, -4.86170349, -3.95946607, -3.02517922, -2.05584072, -1.04844839, 0, 0.870870367, 1.83939386, 2.9040116, 4.06316469, 5.31529425, 6.65884138, 8.0922472, 9.61395282, 11.2223993, 12.9160279, 14.6932796, 16.5525955, 18.4924168, 20.5111845, 22.6073398, 24.7793239, 27.0807172, 29.5178382, 31.9644073, 34.4306814, 36.9269173, 39.4633715, 42.0503009, 44.6979621, 47.416612, 50.2165073, 53.1079046, 56.1010607, 59.2062324, 62.6518051, 66.1499263, 69.6454141, 73.0830865, 76.7994824, 80.5435595, 83.9598168, 86.875556, 89.1180792, 90.5146881, 90.8926847, 91.9976832, 91.1548389, 87.8161319, 81.433542, 71.4590491, 62.6831476, 47.8846744, 25.1809808, -7.31058189, -51.4726623, -105.720734, -179.642279, -272.274902, -387.239757, -537.516083, -731.881532, -979.113754, -1277.25917, -1648.26558, -2125.78756, -2707.63804, -3462.88331, -4397.93561, -5562.82448, -7081.48655, -9043.93389, -11362.6258, -14081.5421, -17895.6896, -23131.8173, -26923.8367, -33282.5403, -43639.8904, -59427.849, -82078.3783, -141876.696, -204188.195, -263231.523, -313225.329, -348388.26, -341045.223, -321904.436, -294567.58, -272290.484, -254853.395, -238013.996, -221753.662, -206053.766, -190895.683, -176260.787, -167458.441, -155009.913, -140364.699, -124972.296, -110282.202, -97743.9141, -88806.9285, -77669.7101, -68975.3308},
        {-94.6155069, -93.4830111, -92.3626737, -91.2540407, -90.1566586, -89.0700736, -87.9938318, -86.9274796, -85.8705632, -84.8226288, -83.7832227, -82.7518911, -81.7281803, -80.7116365, -79.701806, -78.698235, -77.7004697, -76.7080564, -75.7205414, -74.7374709, -73.7583911, -72.7828484, -71.8103888, -70.8405587, -69.8729044, -68.906972, -67.9423078, -66.9784581, -66.014969, -65.051387, -64.0872581, -63.1221287, -62.1555449, -61.1870531, -60.1904546, -59.1787095, -58.1650875, -57.1499704, -56.1337399, -55.1167776, -54.0994653, -53.0821848, -52.0653176, -51.0492456, -50.0343505, -49.021014, -48.0096177, -47.0005435, -45.994173, -44.9908879, -43.9910701, -43.0147501, -42.0422349, -41.07348, -40.1084414, -39.1470746, -38.1893356, -37.2351801, -36.2845638, -35.3374425, -34.3937719, -33.4535079, -32.5166062, -31.5830226, -30.6527128, -29.7256326, -28.8017377, -27.8775479, -26.9496652, -26.0251322, -25.1041273, -24.1868291, -23.2734161, -22.3640668, -21.4589597, -20.5582733, -19.6621862, -18.7708768, -17.8845237, -17.0033053, -16.1274003, -15.256987, -14.3922441, -13.53335, -12.7871567, -12.0982623, -11.4099827, -10.7190346, -10.0221347, -9.31599992, -8.59734689, -7.86289239, -7.10935317, -6.33344599, -5.5318876, -4.70139475, -3.8386842, -2.94047269, -2.00347699, -1.02441384, 0, 0.733553898, 1.59143694, 2.57098577, 3.66953702, 4.88442733, 6.21299333, 7.65257167, 9.20049897, 10.8541119, 12.610747, 14.4677411, 16.4224307, 18.4721524, 20.6142429, 22.8460388, 25.1648769, 27.6578548, 30.3351323, 32.9910147, 35.6420733, 38.3048791, 40.9960033, 43.7320173, 46.529492, 49.4049988, 52.3751087, 55.456393, 58.6654229, 62.0187695, 65.7781386, 69.6250842, 73.5026369, 77.3538276, 81.5305193, 85.7713268, 89.7058426, 93.1544479, 95.9375238, 97.8754514, 98.7886118, 100.451601, 100.135824, 97.2793162, 91.3201114, 81.6962447, 74.2406741, 60.3977801, 38.0068667, 4.90723762, -41.0618032, -98.8495459, -177.753144, -276.645737, -399.225792, -559.043488, -765.225788, -1026.89965, -1340.71745, -1731.11588, -2236.14025, -2856.06456, -3661.75644, -4656.3385, -5892.33343, -7505.9696, -9595.05311, -12070.762, -14982.5173, -19064.0429, -24661.9824, -29052.7705, -36125.5994, -47289.877, -63955.0113, -87530.4104, -148160.382, -211335.71, -271282.077, -322225.166, -358390.659, -352201.042, -334220.662, -308019.43, -286826.631, -270388.674, -254427.228, -238889.938, -223724.449, -208878.406, -194299.454, -184514.935, -171138.459, -155582.162, -139258.181, -123578.654, -109955.717, -99801.5056, -87304.5627, -77501.2637},
        {-90.998127, -89.9108021, -88.8310205, -87.7585872, -86.693307, -85.6349849, -84.5834259, -83.5384348, -82.4998167, -81.4673765, -80.440919, -79.4202494, -78.4051724, -77.395493, -76.3910162, -75.391547, -74.3968901, -73.4068507, -72.4212336, -71.4398438, -70.4624862, -69.4889658, -68.5190875, -67.5526562, -66.5894769, -65.6293545, -64.6720939, -63.7175002, -62.7653782, -61.8155329, -60.8677692, -59.9218921, -58.9777064, -58.0350173, -57.087242, -56.1375138, -55.1890423, -54.2418397, -53.2959184, -52.3512904, -51.4079681, -50.4659637, -49.5252893, -48.5859573, -47.6479797, -46.711369, -45.7761373, -44.8422967, -43.9098596, -42.9788382, -42.0492446, -41.1155612, -40.1834499, -39.253043, -38.3244725, -37.3978706, -36.4733693, -35.5511007, -34.6311969, -33.7137901, -32.7990124, -31.8869958, -30.9778724, -30.0717745, -29.168834, -28.269183, -27.3729538, -26.4798794, -25.5897027, -24.7033784, -23.8210646, -22.9429192, -22.0691001, -21.1997654, -20.335073, -19.4751808, -18.6202468, -17.7704289, -16.9258853, -16.0867737, -15.2532522, -14.4254787, -13.6036112, -12.7878077, -12.0933551, -11.4604253, -10.8278063, -10.1919201, -9.54918849, -8.89603318, -8.22887601, -7.5441388, -6.83824335, -6.10761147, -5.34866497, -4.55782564, -3.73151531, -2.86615577, -1.95816883, -1.00397631, 0, 0.699198951, 1.53081042, 2.49232689, 3.5812408, 4.79504465, 6.13123088, 7.58729197, 9.16072039, 10.8490086, 12.6496491, 14.5601343, 16.5779567, 18.7006087, 20.9255829, 23.2503717, 25.6724675, 28.2900485, 31.1149907, 33.9170454, 36.7152806, 39.5287642, 42.3765641, 45.2777483, 48.2513848, 51.3165415, 54.4922863, 57.7976872, 61.2518122, 64.8737292, 68.929646, 73.1008728, 77.3323355, 81.5689602, 86.2012968, 90.9514595, 95.422729, 99.426344, 102.773543, 105.275566, 106.74365, 109.033634, 109.289887, 106.92362, 101.346041, 91.9683572, 85.1877336, 71.682935, 49.1186814, 15.1596928, -32.5293107, -93.0148265, -175.896096, -279.880185, -408.836295, -577.172987, -794.566868, -1070.69454, -1401.4315, -1813.39207, -2348.24635, -3008.61203, -3866.61006, -4920.26241, -6224.30119, -7929.5508, -10142.7277, -12780.5368, -15900.4323, -20256.5774, -26209.037, -31214.9819, -38983.9707, -50880.9167, -68270.7333, -92518.3335, -153183.821, -216388.121, -276447.349, -327677.622, -364395.054, -359535.759, -343049.19, -318414.32, -298581.03, -283279.392, -268293.269, -253550.044, -238977.101, -224501.823, -210051.596, -198433.495, -184088.615, -168111.089, -151595.049, -135634.628, -121323.959, -109757.174, -96236.1222, -85361.4259},
        {-87.0428371, -86.0308089, -85.0205268, -84.0120684, -83.0055114, -82.0009335, -80.9984123, -79.9980256, -78.9998509, -78.0039661, -77.0104488, -76.0193766, -75.0308272, -74.0448783, -73.0616076, -72.0810928, -71.1034115, -70.1286415, -69.1568603, -68.1881457, -67.2225753, -66.2602269, -65.3011781, -64.3455066, -63.3932901, -62.4446062, -61.4995326, -60.558147, -59.6205271, -58.6867505, -57.756895, -56.8310382, -55.9092578, -54.9916315, -54.0881104, -53.1936279, -52.3029984, -51.4159793, -50.5323278, -49.6518011, -48.7741566, -47.8991515, -47.026543, -46.1560885, -45.2875451, -44.4206702, -43.555221, -42.6909548, -41.8276288, -40.9650003, -40.1028266, -39.2220637, -38.3416777, -37.4618334, -36.5826957, -35.7044293, -34.8271991, -33.9511699, -33.0765066, -32.2033741, -31.331937, -30.4623604, -29.5948089, -28.7294475, -27.866441, -27.0059542, -26.148152, -25.2923898, -24.4380426, -23.5869444, -22.7393126, -21.8953643, -21.055317, -20.2193879, -19.3877943, -18.5607535, -17.7384828, -16.9211995, -16.109121, -15.3024644, -14.5014472, -13.7062865, -12.9171998, -12.1344043, -11.4846133, -10.9021355, -10.319759, -9.73359621, -9.13975929, -8.53436054, -7.91351222, -7.27332658, -6.6099159, -5.91939242, -5.19786842, -4.44145616, -3.64626789, -2.80841588, -1.92401238, -0.989169671, 0, 0.816122703, 1.74537917, 2.78725766, 3.94124643, 5.20683375, 6.58350787, 8.07075705, 9.66806955, 11.3749336, 13.1908376, 15.1152696, 17.147718, 19.287671, 21.5346169, 23.8880439, 26.3474403, 28.991619, 31.8314002, 34.6868476, 37.5744477, 40.5106867, 43.512051, 46.5950271, 49.7761014, 53.0717601, 56.4984897, 60.0727765, 63.811107, 67.7299675, 72.0643999, 76.532198, 81.0842817, 85.6715706, 90.7510846, 96.0206935, 101.043307, 105.618014, 109.543906, 112.620071, 114.645599, 117.632896, 118.494118, 116.595315, 111.302536, 101.981831, 94.8842007, 80.7695062, 57.2825594, 22.0681722, -27.2288433, -89.1738969, -174.543906, -281.86251, -415.304963, -590.585237, -818.236655, -1108.79254, -1458.36897, -1895.19316, -2463.22023, -3166.82785, -4079.17058, -5191.39644, -6560.24216, -8353.58091, -10688.2724, -13494.3951, -16840.3839, -21479.5386, -27777.7607, -33390.5439, -41811.1832, -54343.7982, -72292.5084, -96961.4332, -156933.51, -219414.336, -278888.327, -329839.897, -366753.461, -363497.236, -348915.968, -326329.587, -308080.185, -293916.418, -279827.709, -265750.675, -251621.936, -237378.11, -222955.814, -208338.776, -192916.122, -177101.3, -161307.757, -145948.942, -131438.302, -118189.287, -104102., -92193.3866},
        {-81.9500307, -81.1058026, -80.2544688, -79.3963882, -78.5319193, -77.6614209, -76.7852517, -75.9037703, -75.0173354, -74.1263058, -73.2310401, -72.331897, -71.4292352, -70.5234134, -69.6147902, -68.7037245, -67.7905747, -66.8756998, -65.9594582, -65.0422088, -64.1243102, -63.2061211, -62.2880002, -61.3703062, -60.4533977, -59.5376335, -58.6233723, -57.7109727, -56.8007934, -55.8931931, -54.9885306, -54.0871645, -53.1894534, -52.2957561, -51.4188687, -50.5526694, -49.6908873, -48.8334774, -47.9803949, -47.1315947, -46.2870319, -45.4466617, -44.6104391, -43.7783192, -42.950257, -42.1262076, -41.3061262, -40.4899677, -39.6776873, -38.86924, -38.0645808, -37.2811868, -36.5011113, -35.7239295, -34.9492167, -34.1765481, -33.405499, -32.6356445, -31.86656, -31.0978206, -30.3290016, -29.5596783, -28.7894258, -28.0178194, -27.2444344, -26.4688459, -25.6906293, -24.8914824, -24.0535326, -23.2132254, -22.3712951, -21.5284759, -20.6855018, -19.8431071, -19.0020258, -18.1629921, -17.3267402, -16.4940041, -15.6655182, -14.8420164, -14.0242329, -13.212902, -12.4087576, -11.6125341, -10.9827951, -10.4380403, -9.89487221, -9.34890317, -8.79574543, -8.23101126, -7.65031292, -7.04926269, -6.42347284, -5.76855563, -5.08012334, -4.35378824, -3.5851626, -2.76985868, -1.90348876, -0.981665112, 0, 1.45716319, 2.91357847, 4.37708271, 5.85551276, 7.35670549, 8.88849776, 10.4587264, 12.0752283, 13.7458404, 15.4783994, 17.2807422, 19.1607058, 21.1261269, 23.1848424, 25.3446892, 27.6135041, 29.9447229, 32.3384109, 34.9254551, 37.7020348, 40.6643297, 43.8085192, 47.1307827, 50.6272997, 54.2942498, 58.1278124, 62.124167, 66.2794931, 70.5899701, 75.123719, 79.7785993, 84.529208, 89.3501422, 94.9180973, 100.856621, 106.602035, 111.918308, 116.569406, 120.319296, 122.931948, 126.578456, 127.883056, 126.138754, 120.63856, 110.675481, 100.670986, 83.5075063, 57.1959329, 19.747158, -30.8279262, -91.0763074, -175.080694, -281.327858, -414.642719, -592.790582, -827.726786, -1131.40667, -1500.7969, -1966.248, -2574.46174, -3336.07673, -4317.25954, -5485.80478, -6897.54776, -8757.50385, -11202.2123, -14206.288, -17864.2898, -22864.4996, -29559.356, -35795.2001, -44816.1214, -57842.054, -76092.9317, -100788.688, -158792.299, -219269.897, -277030.654, -326883.746, -363638.345, -363018.281, -351641.635, -332497.013, -316455.753, -303386.215, -289872.287, -275946.552, -261641.592, -246989.991, -232024.332, -210095.261, -192764.382, -178558.161, -166003.064, -153625.56, -139952.114, -123509.194, -110005.96, -97170.7006},
        {-77.6286638, -76.8735856, -76.1083464, -75.3333946, -74.5491785, -73.7561465, -72.9547469, -72.1454281, -71.3286384, -70.5048261, -69.6744397, -68.8379274, -67.9957375, -67.1483185, -66.2961187, -65.4395864, -64.57917, -63.7153178, -62.8484781, -61.9790993, -61.1076298, -60.2345179, -59.3602119, -58.4851601, -57.609811, -56.7346129, -55.8600141, -54.9864629, -54.1144077, -53.2442969, -52.3765787, -51.5117016, -50.6501138, -49.7922638, -48.9525979, -48.1242711, -47.3002699, -46.4805884, -45.6652207, -44.8541609, -44.0474031, -43.2449413, -42.4467696, -41.6528822, -40.863273, -40.0779362, -39.2968658, -38.520056, -37.7475008, -36.9791942, -36.2151305, -35.4764948, -34.7416306, -34.0100727, -33.2813557, -32.5550143, -31.8305832, -31.1075972, -30.3855909, -29.664099, -28.9426562, -28.2207973, -27.4980568, -26.7739695, -26.0480701, -25.3198933, -24.5889738, -23.837465, -23.0479376, -22.2557739, -21.4616354, -20.6661835, -19.8700796, -19.0739854, -18.2785621, -17.4844712, -16.6923743, -15.9029327, -15.1168079, -14.3346614, -13.5571546, -12.784949, -12.018706, -11.2590871, -10.6668638, -10.1592747, -9.65163437, -9.13940823, -8.61806189, -8.08306093, -7.52987092, -6.95395744, -6.35078607, -5.71582238, -5.04453194, -4.33238034, -3.57483315, -2.76735595, -1.9054143, -0.984473794, 0, 1.65878361, 3.28187482, 4.88012892, 6.46440117, 8.04554683, 9.63442119, 11.2418795, 12.8787771, 14.5559691, 16.284311, 18.0746578, 19.937865, 21.8847878, 23.9262815, 26.0732012, 28.3364024, 30.6404523, 32.9838795, 35.5727141, 38.3993212, 41.4560658, 44.7353129, 48.2294275, 51.9307746, 55.8317192, 59.9246263, 64.201861, 68.6557882, 73.278773, 78.1197629, 83.0937934, 88.1762548, 93.3425373, 99.2702097, 105.583573, 111.71968, 117.443269, 122.519075, 126.711835, 129.786286, 134.394484, 136.535097, 135.407951, 130.212874, 120.14969, 108.214398, 88.8616098, 60.3421104, 20.9066833, -31.1938874, -90.9503931, -175.130772, -281.761567, -416.183358, -598.669542, -842.788985, -1162.11055, -1557.93747, -2058.89188, -2706.97667, -3505.91267, -4532.42531, -5760.3447, -7250.40997, -9210.62265, -11781.3885, -14944.7238, -18799.9704, -24027.2308, -30978.0964, -37675.8512, -47158.595, -60583.8715, -79109.2245, -103892.198, -160629.118, -219804.051, -276439.844, -325559.345, -362185.403, -363256.828, -353936.529, -336997.659, -322335.604, -309904.296, -296690.066, -282795.222, -268322.074, -253372.934, -238050.111, -213021.304, -194916.331, -181461.91, -170384.761, -159411.6, -146269.148, -128684.121, -115257.992, -101387.925},
        {-73.9283896, -73.1947253, -72.4527948, -71.7029746, -70.9456411, -70.1811708, -69.4099403, -68.6323259, -67.8487041, -67.0594515, -66.2649446, -65.4655597, -64.6616734, -63.8536621, -63.0419024, -62.2267707, -61.4086435, -60.5878972, -59.7649084, -58.9400535, -58.113709, -57.2862515, -56.4580572, -55.6295028, -54.8009648, -53.9728195, -53.1454435, -52.3192133, -51.4945053, -50.671696, -49.8511619, -49.0332795, -48.2184252, -47.4069756, -46.6136673, -45.8313946, -45.0528792, -44.2780318, -43.5067626, -42.7389821, -41.9746009, -41.2135294, -40.4556781, -39.7009573, -38.9492777, -38.2005496, -37.4546835, -36.7115898, -35.9711791, -35.2333618, -34.4980483, -33.7658673, -33.0359955, -32.3083279, -31.5827592, -30.8591844, -30.1374985, -29.4175962, -28.6993725, -27.9827223, -27.2675405, -26.553722, -25.8411616, -25.1297543, -24.419395, -23.7099785, -23.0013998, -22.2890813, -21.568553, -20.8489337, -20.1304082, -19.4131614, -18.697378, -17.983243, -17.2709411, -16.5606571, -15.8525759, -15.1468823, -14.4437611, -13.7433972, -13.0459754, -12.3516804, -11.6606971, -10.9732104, -10.4322141, -9.96348479, -9.49108239, -9.01055736, -8.51746006, -8.00734091, -7.47575028, -6.91823859, -6.33035621, -5.70765355, -5.045681, -4.33998894, -3.58612779, -2.77964792, -1.91609974, -0.991033635, 0, 1.52062468, 3.03133766, 4.54193957, 6.06223101, 7.60201261, 9.17108497, 10.7792487, 12.4363045, 14.1520528, 15.9362944, 17.7988298, 19.7494597, 21.7979847, 23.9542053, 26.2279223, 28.6289362, 31.1243085, 33.7177343, 36.515686, 39.5188057, 42.7277358, 46.1431184, 49.7655959, 53.5958104, 57.6344041, 61.8820194, 66.3392985, 71.0068835, 75.8854168, 81.1107384, 86.4987202, 92.009445, 97.6029956, 103.810318, 110.306063, 116.612652, 122.51891, 127.81366, 132.285727, 135.723935, 141.419066, 144.592172, 144.346912, 139.786944, 130.015926, 117.184089, 96.5868752, 66.5663, 25.4643786, -28.376874, -88.9257283, -174.928423, -283.873789, -421.51344, -610.468836, -865.787222, -1202.51584, -1627.66576, -2166.14681, -2851.63553, -3673.42564, -4729.30308, -6018.39931, -7609.8104, -9691.18353, -12395.749, -15687.4092, -19655.4944, -25010.2044, -32110.5399, -39141.316, -48970.2096, -62708.8242, -81468.7634, -106361.631, -162349.427, -220730.76, -276654.634, -325270.053, -361726.021, -363650.655, -355394.44, -339615.332, -325687.268, -313609.795, -300584.139, -286749.764, -272246.135, -257212.717, -241788.975, -217310.211, -199396.075, -185912.091, -174723.785, -163696.681, -150696.303, -133588.176, -119847.641, -104975.981},
        {-70.4665418, -69.7177237, -68.9656419, -68.21049, -67.4524616, -66.6917503, -65.9285498, -65.1630536, -64.3954555, -63.6259489, -62.8547276, -62.0819852, -61.3079152, -60.5327113, -59.7565671, -58.9796762, -58.2022323, -57.424429, -56.6464598, -55.8685184, -55.0907985, -54.3134936, -53.5367974, -52.7609035, -51.9860054, -51.2122969, -50.4399715, -49.6692229, -48.9002447, -48.1332305, -47.3683738, -46.6058685, -45.845908, -45.0886859, -44.3459647, -43.6119101, -42.8805492, -42.1517, -41.425181, -40.7008101, -39.9784056, -39.2577857, -38.5387686, -37.8211725, -37.1048156, -36.389516, -35.675092, -34.9613618, -34.2481435, -33.5352554, -32.8225157, -32.0846219, -31.3470575, -30.6101851, -29.8743675, -29.1399675, -28.4073478, -27.6768712, -26.9489004, -26.2237983, -25.5019275, -24.7836508, -24.069331, -23.3593308, -22.654013, -21.9537403, -21.2588755, -20.5812477, -19.9324106, -19.2890786, -18.6508711, -18.0174076, -17.3883076, -16.7631904, -16.1416755, -15.5233823, -14.9079304, -14.2949391, -13.6840279, -13.0748162, -12.4669235, -11.8599693, -11.2535729, -10.6473538, -10.1624062, -9.73505709, -9.30017349, -8.85343274, -8.3905122, -7.90708923, -7.39884119, -6.86144543, -6.2905793, -5.68192017, -5.03114539, -4.33393231, -3.5859583, -2.7829007, -1.92043689, -0.994244199, 0, 1.25928418, 2.55586823, 3.89710232, 5.29033663, 6.74292135, 8.26220666, 9.85554275, 11.5302798, 13.293768, 15.1533575, 17.1163985, 19.1902412, 21.3822357, 23.6997323, 26.1500812, 28.7406324, 31.5025022, 34.4464359, 37.5260273, 40.7537193, 44.1419548, 47.7031768, 51.449828, 55.3943513, 59.5491897, 63.9267861, 68.5395832, 73.4000241, 78.5205516, 84.1537211, 89.9838216, 95.9512625, 101.996453, 108.473412, 115.115745, 121.546759, 127.582779, 133.040131, 137.73514, 141.484134, 148.166959, 152.299696, 152.903633, 149.000058, 139.610262, 126.552112, 105.351173, 74.329588, 31.8095001, -23.8869469, -86.0776055, -175.018118, -288.008641, -431.194696, -628.781092, -896.864385, -1251.54113, -1704.38187, -2277.03782, -2996.09768, -3836.83972, -4918.34608, -6268.50482, -7964.77305, -10168.5999, -13002.8806, -16407.8729, -20459.1215, -25901.8433, -33102.4509, -40389.7155, -50481.1741, -64448.5766, -83363.6727, -108298.212, -163641.821, -221318.903, -276571.741, -324642.615, -360773.806, -363054.735, -355315.783, -340150.455, -326665.116, -314875.267, -302082.716, -288438.647, -274094.246, -259200.697, -243909.184, -222055.959, -204893.486, -190944.071, -178730.014, -166773.622, -153597.196, -137723.041, -123592.885, -107925.619},
        {-66.4848201, -65.7044941, -64.9298022, -64.1605676, -63.3966134, -62.6377629, -61.8838393, -61.1346656, -60.3900652, -59.6498611, -58.9138766, -58.1819348, -57.4538589, -56.7294722, -56.0085977, -55.2910588, -54.5766784, -53.86528, -53.1566865, -52.4507212, -51.7472074, -51.0459681, -50.3468265, -49.6496059, -48.9541294, -48.2602202, -47.5677015, -46.8763965, -46.1861283, -45.4967202, -44.8079952, -44.1197767, -43.4318877, -42.7441515, -42.0482377, -41.3482177, -40.6482611, -39.9484555, -39.2488888, -38.5496488, -37.8508232, -37.1524998, -36.4547664, -35.7577108, -35.0614207, -34.365984, -33.6714884, -32.9780217, -32.2856716, -31.5945261, -30.9046727, -30.2101317, -29.51719, -28.826067, -28.1369819, -27.4501542, -26.765803, -26.0841478, -25.4054078, -24.7298024, -24.0575509, -23.3888726, -22.7239868, -22.0631128, -21.40647, -20.7542777, -20.1067551, -19.4722365, -18.8588613, -18.2501125, -17.6456832, -17.0452668, -16.4485565, -15.8552456, -15.2650274, -14.677595, -14.0926419, -13.5098611, -12.9289461, -12.34959, -11.7714862, -11.1943278, -10.6178082, -10.0416206, -9.59621585, -9.21315816, -8.82244005, -8.41951143, -7.99982224, -7.55882239, -7.09196182, -6.59469045, -6.06245821, -5.49071502, -4.87491081, -4.21049551, -3.49291903, -2.71763132, -1.8800823, -0.975721879, 0, 1.35981315, 2.73499761, 4.13537012, 5.57074745, 7.05094634, 8.58578355, 10.1850758, 11.8586399, 13.6162926, 15.4678506, 17.4231307, 19.4919497, 21.6841242, 24.0094711, 26.4778071, 29.0989489, 31.8589093, 34.7641046, 37.8781937, 41.2058927, 44.7519173, 48.5209836, 52.5178074, 56.7471046, 61.2135911, 65.9219828, 70.8769956, 76.0833454, 81.545748, 87.4542417, 93.5602686, 99.8129478, 106.161399, 113.059982, 120.205196, 127.158808, 133.718365, 139.681414, 144.845502, 149.008175, 156.100204, 160.527974, 161.280358, 157.346228, 147.714457, 133.99797, 111.905573, 79.7701252, 35.9244862, -21.2984852, -84.1586827, -175.228241, -293.882446, -447.108452, -657.434316, -940.410544, -1311.58765, -1778.04438, -2365.7284, -3109.82927, -3993.92911, -5131.12809, -6541.22999, -8300.85978, -10582.9228, -13516.13, -17062.1054, -21306.2411, -26946.5821, -34334.6224, -41905.8557, -52222.5453, -66299.4162, -85151.1933, -109792.601, -163591.035, -219620.382, -273307.199, -320078.044, -355359.474, -358108.671, -351201.151, -337121.776, -324645.833, -313768.223, -301862.836, -289055.622, -275472.531, -261239.515, -246482.525, -227163.298, -210650.02, -196072.727, -182561.455, -169246.238, -155257.114, -139724.117, -126126.263, -110075.917},
        {-62.2183179, -61.4393137, -60.6725692, -59.9176092, -59.1739584, -58.4411416, -57.7186836, -57.0061089, -56.3029425, -55.608709, -54.9229332, -54.2451398, -53.5748535, -52.9115991, -52.2549013, -51.6042849, -50.9592745, -50.319395, -49.684171, -49.0531274, -48.4257888, -47.8016799, -47.1803256, -46.5612505, -45.9439795, -45.3280371, -44.7129482, -44.0982376, -43.4834298, -42.8680497, -42.2516221, -41.6336716, -41.013723, -40.391301, -39.7386112, -39.0694126, -38.3977924, -37.7241619, -37.0489325, -36.3725154, -35.6953219, -35.0177634, -34.340251, -33.6631962, -32.9870102, -32.3121043, -31.6388898, -30.967778, -30.2991802, -29.6335077, -28.9711718, -28.3374659, -27.7073797, -27.0807852, -26.4575543, -25.8375589, -25.2206709, -24.6067623, -23.9957048, -23.3873705, -22.7816311, -22.1783587, -21.5774252, -20.9787023, -20.3820621, -19.7873764, -19.1945171, -18.600722, -18.0032918, -17.4075316, -16.8134838, -16.2211912, -15.6306964, -15.0420422, -14.4552711, -13.8704259, -13.2875491, -12.7066836, -12.1278719, -11.5511567, -10.9765807, -10.4041866, -9.83401697, -9.26611455, -8.85531176, -8.5210006, -8.18071259, -7.82946719, -7.46228382, -7.07418194, -6.66018099, -6.21530042, -5.73455965, -5.21297814, -4.64557534, -4.02737067, -3.35338359, -2.61863354, -1.81813997, -0.946922304, 0, 1.59731699, 3.16397076, 4.71386006, 6.26088365, 7.81894031, 9.40192879, 11.0237478, 12.6982962, 14.4394728, 16.2611761, 18.1773051, 20.2017585, 22.3484351, 24.6312335, 27.0640527, 29.6607913, 32.336472, 35.0908684, 38.1615277, 41.5411607, 45.2224785, 49.1981921, 53.4610127, 58.0036511, 62.8188186, 67.899226, 73.2375844, 78.8266049, 84.6589984, 90.8252298, 97.1844133, 103.699934, 110.335176, 117.726219, 125.500099, 133.104471, 140.300913, 146.851001, 152.516313, 157.058425, 164.175676, 168.494737, 169.00695, 164.703656, 154.576198, 140.619415, 118.070276, 85.1692462, 40.1567932, -18.7266166, -83.1302719, -177.280916, -302.871772, -467.646263, -691.700295, -989.583764, -1375.84656, -1850.35771, -2445.68454, -3210.40987, -4143.65777, -5343.36788, -6808.87676, -8612.78833, -10951.8541, -13965.8397, -17652.4784, -22114.7127, -27976.2709, -35566.2954, -43416.3816, -53929.7646, -68057.8145, -86751.9017, -110963.396, -162878.756, -216905.491, -268686.201, -313863.484, -348079.938, -351310.783, -345310.915, -332437.525, -321062.759, -311162.886, -300207.942, -288299.506, -275539.159, -262028.479, -247869.047, -230392.823, -214448.428, -199544.829, -185190.995, -170895.894, -156168.494, -140517.762, -127570.725, -111667.057},
        {-57.7832348, -57.0958962, -56.4186165, -55.7509619, -55.0924985, -54.4427926, -53.8014101, -53.1679174, -52.5418806, -51.9228657, -51.310439, -50.7041667, -50.1036147, -49.5083494, -48.9179369, -48.3319432, -47.7499346, -47.1714773, -46.5961373, -46.0234808, -45.4530739, -44.884483, -44.3172739, -43.751013, -43.1852664, -42.6196002, -42.0535806, -41.4867737, -40.9187456, -40.3490627, -39.7772908, -39.2029964, -38.6257454, -38.045104, -37.4314183, -36.799479, -36.1644282, -35.5267803, -34.8870498, -34.2457509, -33.6033982, -32.960506, -32.3175887, -31.6751608, -31.0337365, -30.3938304, -29.7559568, -29.1206302, -28.4883649, -27.8596753, -27.2350759, -26.6550624, -26.079301, -25.5074396, -24.9391258, -24.3740073, -23.8117318, -23.2519469, -22.6943003, -22.1384398, -21.5840129, -21.0306675, -20.4780512, -19.9258116, -19.3735964, -18.8210534, -18.2678302, -17.7020391, -17.1120695, -16.5213598, -15.9303055, -15.3393021, -14.748745, -14.1590297, -13.5705518, -12.9837067, -12.3988898, -11.8164966, -11.2369227, -10.6605634, -10.0878143, -9.51907089, -8.95472856, -8.39518282, -8.02314712, -7.74402315, -7.46102125, -7.16862033, -6.8612993, -6.53353705, -6.17981251, -5.79460457, -5.37239214, -4.90765413, -4.39486945, -3.828517, -3.20307569, -2.51302443, -1.75284213, -0.917007679, 0, 1.80280596, 3.53992097, 5.2282988, 6.88489322, 8.526658, 10.1705469, 11.8335138, 13.5325123, 15.2844963, 17.1064195, 19.0152357, 21.0278986, 23.1613622, 25.43258, 27.858506, 30.4560938, 33.1168216, 35.839718, 38.9255508, 42.3643862, 46.1462904, 50.2613293, 54.6995693, 59.4510763, 64.5059165, 69.854156, 75.485861, 81.3910976, 87.5599319, 94.074335, 100.79877, 107.69573, 114.727712, 122.625455, 130.967331, 139.126846, 146.836021, 153.826877, 159.831435, 164.581716, 171.434942, 175.394611, 175.483464, 170.724242, 160.139689, 147.242123, 125.442316, 92.6406142, 46.7373663, -14.36708, -83.053993, -183.019433, -316.586532, -491.638814, -727.412276, -1038.04589, -1437.67862, -1921.82297, -2527.47766, -3311.96123, -4284.39807, -5533.53488, -7045.82767, -8891.68482, -11283.7673, -14369.4398, -18172.7476, -22809.0837, -28846.0122, -36595.3236, -44698.6792, -55394.9418, -69574.0236, -88125.8368, -111940.294, -162088.118, -214233.207, -264220.271, -307894.02, -341099.162, -344795.65, -339636.122, -327844.468, -317257.729, -307877.925, -297368.26, -285847.694, -273435.187, -260249.698, -246410.186, -229152.363, -213665.828, -199323.772, -185499.387, -171565.863, -156896.392, -140864.163, -127939.597, -112906.357},
        {-53.2138641, -52.7716956, -52.3217474, -51.8642402, -51.3993947, -50.9274315, -50.4485711, -49.9630343, -49.4710416, -48.9728137, -48.4685713, -47.9585348, -47.442925, -46.9219625, -46.3958679, -45.8648619, -45.3291651, -44.788998, -44.2445814, -43.6961358, -43.1438819, -42.5880404, -42.0288317, -41.4664767, -40.9011958, -40.3332098, -39.7627392, -39.1900047, -38.6152269, -38.0386264, -37.460424, -36.88084, -36.3000954, -35.7184105, -35.1382974, -34.5587829, -33.9788662, -33.3986935, -32.8184112, -32.2381655, -31.6581027, -31.0783691, -30.4991109, -29.9204744, -29.3426058, -28.7656516, -28.1897578, -27.6150708, -27.0417369, -26.4699024, -25.8997134, -25.3464507, -24.7947981, -24.2445737, -23.6955958, -23.1476825, -22.600652, -22.0543225, -21.5085122, -20.9630393, -20.4177219, -19.8723783, -19.3268266, -18.7808851, -18.2343718, -17.6871051, -17.138903, -16.5784063, -15.9945238, -15.4101266, -14.8257577, -14.2419597, -13.6592754, -13.0782476, -12.4994192, -11.9233328, -11.3505313, -10.7815575, -10.2169541, -9.65726398, -9.10302985, -8.55479453, -8.01310079, -7.47849142, -7.15678868, -6.94157784, -6.72397687, -6.49786684, -6.25712883, -5.9956439, -5.70729314, -5.3859576, -5.02551837, -4.61985651, -4.1628531, -3.6483892, -3.0703459, -2.42260425, -1.69904534, -0.893550232, 0, 1.84603137, 3.6305793, 5.37044107, 7.08241398, 8.7832953, 10.4898823, 12.2189723, 13.9873626, 15.8118505, 17.7092332, 19.696308, 21.7898722, 24.0067232, 26.3636581, 28.8774744, 31.5649691, 34.4078834, 37.4180063, 40.6914294, 44.2374379, 48.0653171, 52.184352, 56.603828, 61.3330301, 66.3812437, 71.7577538, 77.4718458, 83.5328048, 89.949916, 97.0059804, 104.336478, 111.86864, 119.529696, 127.908741, 136.602072, 145.029492, 152.919672, 160.001285, 166.003, 170.653491, 176.979605, 180.408043, 180.022183, 174.905399, 164.141067, 154.503254, 135.461974, 104.177928, 57.8118208, -6.47564444, -84.0614645, -194.367825, -336.793993, -518.219142, -760.793927, -1079.81242, -1490.56869, -1992.11069, -2619.67092, -3426.20491, -4414.11835, -5682.02431, -7227.93705, -9126.22068, -11580.6099, -14735.4563, -18610.7568, -23318.6229, -29427.1324, -37246.4429, -45565.4041, -56449.5285, -70735.5802, -89260.3232, -112860.521, -161733.638, -212515.564, -261202.887, -303792.196, -336280.082, -340444.621, -335979.922, -324976.221, -314508.326, -304693.415, -293571.761, -281364.389, -268292.321, -254576.582, -240438.195, -220611.51, -205404.452, -193181.153, -182305.745, -171142.358, -158055.126, -141408.178, -127171.088, -113978.379},
        {-49.2914599, -49.0311281, -48.7484903, -48.444326, -48.1194147, -47.7745359, -47.410469, -47.0279936, -46.6278892, -46.2109352, -45.7779112, -45.3295966, -44.866771, -44.3902137, -43.9007044, -43.3990225, -42.8859475, -42.3622589, -41.8287362, -41.2861589, -40.7353064, -40.1769583, -39.6118941, -39.0408933, -38.4647353, -37.8841996, -37.3000657, -36.7131132, -36.1241215, -35.5338702, -34.9431386, -34.3527063, -33.7633529, -33.1758577, -32.6209861, -32.0838941, -31.5493769, -31.0172411, -30.487293, -29.9593391, -29.4331856, -28.9086391, -28.3855058, -27.8635923, -27.3427048, -26.8226498, -26.3032337, -25.7842629, -25.2655437, -24.7468826, -24.228086, -23.6997901, -23.1711702, -22.6422315, -22.1129792, -21.5834185, -21.0535546, -20.5233926, -19.9929378, -19.4621953, -18.9311704, -18.3998681, -17.8682937, -17.3364524, -16.8043494, -16.2719898, -15.7393788, -15.1982518, -14.6405427, -14.0833125, -13.5271027, -12.9724544, -12.4199089, -11.8700075, -11.3232915, -10.7803023, -10.241581, -9.7076689, -9.17910738, -8.65643769, -8.14020111, -7.63093892, -7.12919241, -6.63550285, -6.36808225, -6.21405812, -6.05794175, -5.89321058, -5.71334208, -5.5118137, -5.28210287, -5.01768707, -4.71204373, -4.35865031, -3.95098425, -3.48252303, -2.94674407, -2.33712484, -1.64714278, -0.870275351, 0, 1.89267789, 3.71791044, 5.49370256, 7.23805915, 8.96898513, 10.7044854, 12.4625648, 14.2612284, 16.1184809, 18.0523273, 20.0807726, 22.2218215, 24.4934791, 26.9137502, 29.5006397, 32.2721525, 35.196584, 38.2848376, 41.6673567, 45.3514941, 49.3446026, 53.6540352, 58.2871446, 63.2512837, 68.5538054, 74.2020624, 80.2034077, 86.565194, 93.2947742, 100.676286, 108.33881, 116.206663, 124.204162, 132.931238, 141.9744, 150.739998, 158.949664, 166.325032, 172.587736, 177.459409, 184.13567, 187.806867, 187.514941, 182.301832, 171.209481, 161.831696, 142.520584, 110.180118, 61.7142719, -5.97297959, -89.0736466, -206.026014, -355.42138, -543.946513, -794.81644, -1123.82533, -1546.76737, -2064.15473, -2710.53676, -3537.13409, -4541.60981, -5831.00584, -7416.0564, -9374.59246, -11898.1366, -15126.0252, -19071.1133, -23839.9458, -30002.8318, -37867.749, -46369.5122, -57400.518, -71744.0698, -90183.471, -113502.025, -160962.929, -210249.62, -257525.426, -298953.678, -330697.705, -335382.322, -331735.052, -321721.445, -311891.621, -302396.406, -291476.39, -279377.912, -266347.308, -252630.918, -238475.079, -216624.445, -200893.21, -189212.383, -179512.976, -169725.999, -157782.462, -141613.376, -127092.201, -115038.38},
        {-46.0338544, -45.8473407, -45.632037, -45.3890004, -45.1192884, -44.8239585, -44.504068, -44.1606742, -43.7948347, -43.4076069, -43.000048, -42.5732155, -42.1281668, -41.6659594, -41.1876505, -40.6942977, -40.1869582, -39.6666896, -39.1345491, -38.5915943, -38.0388824, -37.4774709, -36.9084172, -36.3327787, -35.7516128, -35.1659769, -34.5769284, -33.9855247, -33.3928231, -32.7998811, -32.2077561, -31.6175055, -31.0301866, -30.446857, -29.9129589, -29.4064236, -28.9046491, -28.4072524, -27.9138504, -27.4240602, -26.9374989, -26.4537834, -25.9725308, -25.4933581, -25.0158823, -24.5397204, -24.0644894, -23.5898065, -23.1152885, -22.6405525, -22.1652156, -21.6652317, -21.1643938, -20.662832, -20.1606762, -19.6580564, -19.1551026, -18.6519448, -18.1487129, -17.6455371, -17.1425472, -16.6398733, -16.1376453, -15.6359932, -15.135047, -14.6349368, -14.1357925, -13.6333048, -13.1232714, -12.6149774, -12.1088409, -11.6052793, -11.1047107, -10.6075527, -10.114223, -9.62513945, -9.14071982, -8.66138185, -8.1875433, -7.71962194, -7.25803552, -6.80320181, -6.35553858, -5.91546358, -5.70293866, -5.60399173, -5.50201194, -5.39029243, -5.26212632, -5.11080677, -4.92962689, -4.71187982, -4.4508587, -4.13985666, -3.77216684, -3.34108237, -2.83989639, -2.26190203, -1.60039242, -0.848660697, 0, 1.94409055, 3.80373663, 5.59959003, 7.35230252, 9.08252591, 10.810912, 12.5581125, 14.3447793, 16.1915642, 18.1191188, 20.1480952, 22.2991449, 24.5929198, 27.0500717, 29.6912524, 32.5371137, 35.432581, 38.3732025, 41.777106, 45.6272877, 49.9067436, 54.59847, 59.6854631, 65.1507189, 70.9772338, 77.1480037, 83.6460249, 90.4542936, 97.5558058, 105.019947, 112.711647, 120.587988, 128.606047, 137.556397, 146.979371, 156.193039, 164.904432, 172.820582, 179.648522, 185.095284, 192.989924, 197.662921, 198.014823, 192.946179, 181.357538, 169.019208, 146.244538, 110.216637, 58.1186133, -12.8664234, -97.2483063, -216.304057, -370.33845, -566.811956, -827.794582, -1168.79695, -1605.32969, -2136.24115, -2797.4911, -3643.21632, -4671.52731, -5992.34841, -7625.41752, -9650.04633, -12246.491, -15548.6724, -19562.9423, -24390.1482, -30598.8915, -38493.1855, -47154.0176, -58296.5849, -72648.2559, -90936.3989, -113888.382, -159756.042, -207361.163, -253049.995, -293168.791, -324063.803, -329210.628, -326406.607, -317513.985, -308893.846, -300627.646, -290922.325, -279953.066, -267895.049, -254923.458, -241213.473, -218753.418, -201898.624, -188924.413, -178106.104, -167719.018, -156038.477, -141339.799, -127732.131, -116031.802},
        {-42.358798, -42.1988371, -42.0108906, -41.7959927, -41.5551772, -41.2894783, -40.9999299, -40.6875661, -40.3534209, -39.9985283, -39.6239223, -39.230637, -38.8197063, -38.3921643, -37.9490451, -37.4913825, -37.0202107, -36.5365637, -36.0414754, -35.53598, -35.0211113, -34.4979035, -33.9673906, -33.4306066, -32.8885854, -32.3423612, -31.7929679, -31.2414396, -30.6888102, -30.1361139, -29.5843846, -29.0346563, -28.487963, -27.9453389, -27.4523151, -26.986741, -26.5259314, -26.0694763, -25.6169655, -25.1679892, -24.7221374, -24.2789999, -23.8381669, -23.3992283, -22.9617741, -22.5253944, -22.0896791, -21.6542181, -21.2186016, -20.7824196, -20.3452619, -19.8793921, -19.4123192, -18.9442254, -18.4752932, -18.0057049, -17.5356429, -17.0652897, -16.5948276, -16.124439, -15.6543062, -15.1846117, -14.7155378, -14.2472669, -13.7799815, -13.3138638, -12.8490963, -12.3830377, -11.9131146, -11.4453329, -10.980058, -10.5176554, -10.0584904, -9.60292849, -9.15133511, -8.70407569, -8.26151563, -7.82402039, -7.39195538, -6.96568602, -6.54557776, -6.13199601, -5.72530621, -5.32587378, -5.15989286, -5.11006453, -5.05637556, -4.9918628, -4.90956311, -4.80251333, -4.66375033, -4.48631095, -4.26323204, -3.98755045, -3.65230304, -3.25052666, -2.77525816, -2.21953439, -1.57639221, -0.838868461, 0, 1.87939333, 3.68971577, 5.45129455, 7.18445687, 8.90952997, 10.6468411, 12.4167174, 14.2394862, 16.1354746, 18.1250099, 20.2284193, 22.4660301, 24.8581694, 27.4251645, 30.1873425, 33.1650308, 36.1963202, 39.2755042, 42.8351112, 46.8564178, 51.3207006, 56.2092361, 61.5033011, 67.1841719, 73.2331252, 79.6314376, 86.3603857, 93.401246, 100.735295, 108.406898, 116.311111, 124.410283, 132.666764, 141.941853, 151.748426, 161.359638, 170.468155, 178.766641, 185.947762, 191.704183, 200.007997, 204.970008, 205.4456, 200.290159, 188.359071, 175.526297, 151.874003, 114.502932, 60.5138265, -12.9925721, -100.532445, -223.634109, -382.001028, -583.145122, -850.520567, -1200.42379, -1649.15123, -2196.18915, -2878.46463, -3751.24306, -4815.33455, -6177.85527, -7862.20309, -9946.20532, -12599.917, -15951.059, -19990.3972, -24810.1456, -31004.6873, -38884.3504, -47742.6228, -59063.1092, -73508.1204, -91739.9673, -114420.961, -158912.24, -205002.581, -249179.586, -287930.861, -317744.008, -322652.529, -319900.866, -311314.093, -303300.952, -295872.152, -287128.542, -277176.324, -266121.701, -254070.875, -241130.048, -219347.631, -202533.636, -189177.144, -177767.235, -166792.992, -154743.493, -140107.82, -126519.851, -116380.672},
        {-38.288519, -38.1380585, -37.9646915, -37.7692393, -37.5525233, -37.3153648, -37.058585, -36.7830055, -36.4894474, -36.1787322, -35.8516811, -35.5091155, -35.1518567, -34.780726, -34.3965449, -34.0001345, -33.5923163, -33.1739115, -32.7457416, -32.3086278, -31.8633915, -31.4108539, -30.9518366, -30.4871606, -30.0176475, -29.5441185, -29.067395, -28.5882983, -28.1076497, -27.6262705, -27.1449822, -26.664606, -26.1859632, -25.7098752, -25.2723235, -24.8558097, -24.442413, -24.0318137, -23.6236922, -23.2177287, -22.8136036, -22.4109972, -22.0095898, -21.6090617, -21.2090932, -20.8093647, -20.4095565, -20.0093489, -19.6084222, -19.2064567, -18.8031328, -18.3755159, -17.9463915, -17.5159301, -17.0843022, -16.6516784, -16.2182294, -15.7841256, -15.3495376, -14.914636, -14.4795913, -14.0445741, -13.609755, -13.1753046, -12.7413933, -12.3081918, -11.8758706, -11.440997, -11.0002246, -10.5611557, -10.1241945, -9.68974501, -9.25821145, -8.82999797, -8.40550873, -7.98514787, -7.56931956, -7.15842793, -6.75287715, -6.35307137, -5.95941474, -5.57231142, -5.19216555, -4.8193813, -4.69220464, -4.68711612, -4.67773763, -4.6567549, -4.61685366, -4.55071964, -4.45103856, -4.31049615, -4.12177814, -3.87757027, -3.57055825, -3.19342783, -2.73886471, -2.19955465, -1.56818335, -0.837436561, 0, 1.70546057, 3.38759842, 5.06362439, 6.75074929, 8.46618396, 10.2271392, 12.0508259, 13.9544548, 15.9552368, 18.0703827, 20.3171033, 22.7126095, 25.274112, 28.0188218, 30.9639496, 34.1267062, 37.4554494, 40.9575539, 44.8059696, 49.0031532, 53.5515612, 58.4536501, 63.7118767, 69.3286975, 75.3065691, 81.647948, 88.3552909, 95.4310544, 102.877695, 110.915828, 119.24976, 127.816499, 136.553056, 146.245007, 156.405078, 166.319708, 175.671335, 184.142397, 191.415332, 197.172579, 205.079163, 209.622844, 209.706857, 204.23444, 192.108828, 181.330387, 159.430943, 123.03945, 68.7848619, -6.70386614, -99.7853946, -229.413848, -392.360008, -595.417045, -865.840508, -1221.69875, -1681.06014, -2246.71395, -2955.48013, -3860.84673, -4966.2901, -6373.53259, -8106.01696, -10238.48, -12932.5732, -16310.0531, -20342.6722, -25113.2973, -31254.8073, -39088.9633, -48144.5257, -59658.2558, -74226.4148, -92445.2636, -114911.064, -158241.925, -203006.797, -245796.479, -283201.768, -311813.465, -315996.706, -312718.392, -303808.891, -295801.987, -288666.279, -280410.607, -271101.556, -260805.713, -249589.663, -237519.991, -217357.859, -201710.106, -189061.201, -177895.614, -166697.813, -153952.268, -138143.448, -123697.712, -116060.406},
        {-34.5767865, -34.4692711, -34.3396695, -34.1887512, -34.0172856, -33.8260424, -33.6157909, -33.3873006, -33.1413409, -32.8786815, -32.6000917, -32.306341, -31.9981988, -31.6764348, -31.3418182, -30.9951187, -30.6371057, -30.2685486, -29.890217, -29.5028803, -29.107308, -28.7042695, -28.2945344, -27.8788721, -27.458052, -27.0328438, -26.6040168, -26.1723405, -25.7385844, -25.3035179, -24.8679106, -24.4325319, -23.9981513, -23.5655383, -23.1672423, -22.7874745, -22.4100637, -22.0347482, -21.6612662, -21.2893557, -20.918755, -20.5492021, -20.1804352, -19.8121926, -19.4442122, -19.0762324, -18.7079912, -18.3392267, -17.9696773, -17.5990809, -17.2271757, -16.8355559, -16.442497, -16.0481305, -15.6525877, -15.2560003, -14.8584997, -14.4602173, -14.0612847, -13.6618333, -13.2619946, -12.8619001, -12.4616812, -12.0614695, -11.6613964, -11.2615934, -10.8621919, -10.458193, -10.0447209, -9.63248815, -9.22195897, -8.81359739, -8.40786747, -8.00523328, -7.60615889, -7.21110837, -6.82054578, -6.4349352, -6.05474068, -5.6804263, -5.31245613, -4.95129423, -4.59740467, -4.25125152, -4.16293244, -4.202844, -4.23838235, -4.26191047, -4.26579134, -4.24238794, -4.18406324, -4.08318022, -3.93210186, -3.72319114, -3.44881103, -3.10132451, -2.67309456, -2.15648416, -1.54385628, -0.827573901, 0, 1.50273567, 3.02773696, 4.58945051, 6.20232294, 7.88080091, 9.63933104, 11.49236, 13.4543344, 15.5397008, 17.762906, 20.1383966, 22.6806191, 25.4040203, 28.3230468, 31.4521451, 34.805762, 38.4207118, 42.3146366, 46.4513892, 50.8502095, 55.5303372, 60.511012, 65.8114737, 71.4509618, 77.4487162, 83.8239765, 90.5959825, 97.7839739, 105.40719, 113.846414, 122.626776, 131.659055, 140.854027, 150.872615, 161.250523, 171.323417, 180.777032, 189.297101, 196.569357, 202.279535, 210.09218, 214.503271, 214.420079, 208.749874, 196.399926, 186.95003, 165.966801, 129.689379, 74.3569033, -3.79148715, -102.699419, -239.054938, -407.054677, -612.844053, -886.963926, -1249.0021, -1718.54638, -2302.60162, -3036.19833, -3968.09388, -5098.70666, -6531.84331, -8294.78093, -10462.3966, -13190.2228, -16596.7699, -20652.1855, -25436.5156, -31573.8029, -39380.3748, -48475.4066, -59979.632, -74450.7004, -92446.2611, -114523.963, -156677.853, -200170.084, -241699.206, -277963.77, -305662.326, -309622.745, -306328.042, -297564.2, -289693.988, -282691.744, -274624.814, -265562.889, -255575.657, -244732.807, -233104.03, -214626.448, -200042.379, -187947.732, -176938.416, -165610.341, -152559.417, -136381.553, -121455.633, -115343.787},
        {-31.4352978, -31.4194436, -31.3755558, -31.3046159, -31.2076052, -31.0855051, -30.9392971, -30.7699626, -30.578483, -30.3658398, -30.1330143, -29.8809879, -29.6107421, -29.3232583, -29.019518, -28.7005024, -28.3671931, -28.0205714, -27.6616189, -27.2913168, -26.9106466, -26.5205898, -26.1221277, -25.7162418, -25.3039134, -24.8861241, -24.4638551, -24.038088, -23.6098041, -23.1799849, -22.7496118, -22.3196661, -21.8911294, -21.464983, -21.0812472, -20.7205628, -20.3631016, -20.008578, -19.6567065, -19.3072018, -18.9597782, -18.6141505, -18.270033, -17.9271404, -17.5851871, -17.2438877, -16.9029567, -16.5621086, -16.221058, -15.8795195, -15.5372075, -15.1766981, -14.8152159, -14.4528468, -14.089677, -13.7257926, -13.3612794, -12.9962237, -12.6307115, -12.2648289, -11.8986618, -11.5322964, -11.1658187, -10.7993148, -10.4328707, -10.0665725, -9.70050623, -9.3281164, -8.9430069, -8.5589616, -8.17649709, -7.79612998, -7.41837689, -7.04375441, -6.67277916, -6.30596773, -5.94383674, -5.58690279, -5.23568249, -4.89069244, -4.55244925, -4.22146953, -3.89826988, -3.58336692, -3.53658171, -3.62332452, -3.70588894, -3.77637666, -3.82688935, -3.84952869, -3.83639635, -3.77959402, -3.67122337, -3.50338608, -3.26818383, -2.95771829, -2.56409114, -2.07940406, -1.49575872, -0.805256812, 0, 1.30879798, 2.67421706, 4.10933189, 5.62721711, 7.24094735, 8.96359727, 10.8082415, 12.7879547, 14.9158114, 17.2048864, 19.6682543, 22.3189897, 25.1701673, 28.2348617, 31.5261475, 35.0570994, 38.8878884, 43.0383174, 47.4149332, 52.0409025, 56.9393922, 62.1335689, 67.6465995, 73.5016505, 79.7218888, 86.3304811, 93.3505942, 100.805395, 108.718049, 117.542345, 126.712934, 136.123798, 145.668917, 155.881736, 166.336504, 176.436947, 185.885206, 194.383424, 201.633741, 207.338302, 215.656695, 220.476999, 220.629247, 214.943471, 202.249703, 192.359231, 170.375519, 132.383284, 74.4672448, -7.28788008, -111.426503, -253.799171, -427.423971, -638.188125, -918.253326, -1287.98457, -1767.74684, -2368.66398, -3122.4424, -4070.86977, -5204.71471, -6639.44694, -8411.49275, -10600.2229, -13357.7753, -16802.8865, -20926.8216, -25815.2689, -32021.3125, -39833.6511, -48779.1883, -60025.6168, -74125.3337, -91630.7357, -113094.22, -153973.618, -196189.533, -236568.004, -271935.067, -299116.762, -303651.137, -301177.545, -293359.735, -285981.226, -279075.022, -270957.558, -261747.669, -251564.188, -240525.95, -228751.788, -211616.579, -197585.209, -185546.396, -174388.86, -163001.318, -150272.489, -135091.091, -120509.558, -114355.322},
        {-28.5074737, -28.5844005, -28.6268698, -28.6360915, -28.613276, -28.5596335, -28.4763741, -28.364708, -28.2258455, -28.0609967, -27.871372, -27.6581813, -27.4226351, -27.1659434, -26.8893165, -26.5939646, -26.2810979, -25.9519265, -25.6076608, -25.2495108, -24.8786868, -24.4963991, -24.1038577, -23.702273, -23.292855, -22.8768141, -22.4553604, -22.0297041, -21.6010554, -21.1706246, -20.7396218, -20.3092573, -19.8807411, -19.4552837, -19.0807545, -18.7340527, -18.3915165, -18.052842, -17.717725, -17.3858617, -17.0569481, -16.7306802, -16.4067539, -16.0848654, -15.7647105, -15.4459854, -15.1283861, -14.8116084, -14.4953486, -14.1793026, -13.8631663, -13.5305268, -13.1975382, -12.864246, -12.5306952, -12.1969312, -11.8629992, -11.5289443, -11.1948118, -10.860647, -10.5264951, -10.1924012, -9.85841071, -9.52456875, -9.19092058, -8.85751143, -8.52438652, -8.18389482, -7.82857009, -7.4743305, -7.12172021, -6.77128337, -6.42356412, -6.07910661, -5.73845498, -5.40215338, -5.07074595, -4.74477685, -4.42479022, -4.11133019, -3.80494093, -3.50616657, -3.21555126, -2.93363916, -2.9284611, -3.06119166, -3.18979043, -3.30612109, -3.40204733, -3.46943284, -3.50014131, -3.48603643, -3.41898187, -3.29084134, -3.09347852, -2.81875709, -2.45854075, -2.00469319, -1.44907808, -0.783559121, 0, 1.1312553, 2.3494206, 3.66654388, 5.09467311, 6.64585625, 8.33214129, 10.1655762, 12.1582089, 14.3220875, 16.6692598, 19.2117739, 21.9616777, 24.9310193, 28.1318464, 31.5762073, 35.2761497, 39.2971265, 43.6588149, 48.2524664, 53.1015727, 58.2296258, 63.6601176, 69.4165398, 75.5223843, 82.0011429, 88.8763076, 96.1713701, 103.909822, 112.115156, 121.2793, 130.78555, 140.516865, 150.356206, 160.750997, 171.301966, 181.459322, 190.936685, 199.447679, 206.705926, 212.425048, 221.301521, 226.549594, 226.907981, 221.115399, 207.910564, 196.77672, 173.021923, 132.698757, 71.8598047, -13.4423489, -121.697749, -268.981303, -448.285519, -665.574599, -953.296621, -1332.00894, -1822.26892, -2437.93378, -3207.78377, -4167.90838, -5299.9778, -6730.85093, -8504.63103, -10705.3042, -13484.4356, -16962.5707, -21158.3727, -26165.9732, -32453.0709, -40276.0536, -49033.7455, -59972.0312, -73644.1478, -90603.3321, -111402.821, -150962.332, -191877.001, -231108.446, -265618.284, -292368.131, -297684.477, -296254.153, -289596.688, -282801.649, -275975.426, -267736.895, -258266.821, -247745.971, -236355.112, -224275.01, -208326.262, -194708.329, -182623.535, -171274.206, -159862.666, -147591.242, -133662.258, -119990.6, -113156.647},
        {-25.3925761, -25.5073511, -25.588022, -25.6357599, -25.651736, -25.6371212, -25.5930867, -25.5208035, -25.4214426, -25.2961752, -25.1461724, -24.9726051, -24.7766445, -24.5594616, -24.3222275, -24.0661133, -23.79229, -23.5019288, -23.1962006, -22.8762765, -22.5433277, -22.1985251, -21.84304, -21.4780432, -21.1047059, -20.7241992, -20.3376941, -19.9463618, -19.5513731, -19.1538994, -18.7551115, -18.3561807, -17.9582778, -17.5625741, -17.2129958, -16.888437, -16.567278, -16.2493025, -15.9342939, -15.622036, -15.3123122, -15.0049063, -14.6996016, -14.3961819, -14.0944306, -13.7941315, -13.495068, -13.1970238, -12.8997824, -12.6031275, -12.3068425, -11.9986301, -11.6906169, -11.3828481, -11.0753693, -10.7682261, -10.4614637, -10.1551278, -9.84926377, -9.54391711, -9.23913328, -8.93495778, -8.63143605, -8.32861359, -8.02653586, -7.72524834, -7.4247965, -7.11866624, -6.80050097, -6.48387478, -6.16925836, -5.85712241, -5.54793766, -5.24217478, -4.9403045, -4.64279751, -4.35012452, -4.06275622, -3.78116333, -3.50581654, -3.23718657, -2.9757441, -2.72195985, -2.47630452, -2.51110207, -2.6851785, -2.85409236, -3.00949219, -3.14302653, -3.24634392, -3.31109291, -3.32892204, -3.29147986, -3.19041489, -3.0173757, -2.76401081, -2.42196878, -1.98289814, -1.43844743, -0.780265205, 0, 1.01184321, 2.13787074, 3.38882513, 4.77544888, 6.3084845, 7.99867453, 9.85676146, 11.8934878, 14.1195961, 16.5458289, 19.1829286, 22.0416378, 25.1326991, 28.4668548, 32.0548476, 35.9074199, 40.1029978, 44.6619927, 49.4420537, 54.468427, 59.7663586, 65.3610946, 71.2778812, 77.5419645, 84.1785905, 91.2130054, 98.6704552, 106.576186, 114.955444, 124.298149, 133.994493, 143.93032, 153.991475, 164.725517, 175.687433, 186.255753, 196.117805, 204.960919, 212.472425, 218.339652, 227.301499, 232.456292, 232.503013, 226.140639, 212.06815, 199.04811, 173.200018, 130.706957, 67.7520122, -19.4817339, -128.078689, -276.8169, -461.444521, -688.871576, -988.796899, -1380.68761, -1884.01084, -2509.96518, -3288.16549, -4257.24685, -5400.58999, -6843.33924, -8626.47849, -10833.5958, -13620.4697, -17109.7419, -21336.0829, -26399.3273, -32721.8712, -40543.6517, -49284.7117, -60131.02, -73606.3232, -90234.3679, -110538.901, -148708.307, -188185.535, -226078.171, -259493.802, -285540.015, -291303.47, -290593.297, -284811.765, -278528.836, -271892.901, -263732.478, -254261.953, -243695.708, -232248.128, -220133.598, -205047.964, -191855.891, -179899.289, -168520.069, -157060.141, -144861.416, -131265.804, -119501.764, -111809.924},
        {-22.3891136, -22.5176432, -22.6148801, -22.6818625, -22.7196284, -22.7292161, -22.7116638, -22.6680096, -22.5992918, -22.5065484, -22.3908176, -22.2531378, -22.0945469, -21.9160832, -21.7187849, -21.5036902, -21.2718371, -21.024264, -20.7620089, -20.48611, -20.1976056, -19.8975338, -19.5869328, -19.2668407, -18.9382957, -18.602336, -18.2599998, -17.9123253, -17.5603506, -17.2051139, -16.8476533, -16.4890072, -16.1302135, -15.7723106, -15.4513765, -15.1501924, -14.8511184, -14.5540555, -14.2589048, -13.9655672, -13.673944, -13.3839361, -13.0954447, -12.8083707, -12.5226153, -12.2380794, -11.9546643, -11.6722708, -11.3908001, -11.1101533, -10.8302314, -10.5433608, -10.2571815, -9.97175859, -9.6871575, -9.40344344, -9.12068169, -8.8389375, -8.55827614, -8.27876289, -8.000463, -7.72344175, -7.44776439, -7.17349619, -6.90070243, -6.62944836, -6.35979925, -6.08719782, -5.80719778, -5.52939804, -5.25416354, -4.98185919, -4.71284994, -4.4475007, -4.1861764, -3.92924198, -3.67706234, -3.43000243, -3.18842718, -2.95270149, -2.72319032, -2.50025857, -2.28427118, -2.07559307, -2.14997771, -2.36430332, -2.57213819, -2.76491035, -2.93404785, -3.07097871, -3.16713098, -3.21393271, -3.20281192, -3.12519667, -2.97251498, -2.7361949, -2.40766447, -1.97835172, -1.4396847, -0.783091448, 0, 0.931834535, 2.00083203, 3.2165327, 4.58847675, 6.1262044, 7.83925586, 9.73717134, 11.8294911, 14.1257552, 16.635504, 19.3682777, 22.3336165, 25.5410606, 29.0001501, 32.7204254, 36.7114266, 41.062381, 45.7942127, 50.7362174, 55.915011, 61.3572096, 67.0894292, 73.1382857, 79.5303953, 86.2923739, 93.4508375, 101.032402, 109.063684, 117.571299, 127.025267, 136.846218, 146.927748, 157.163451, 168.265622, 179.718507, 190.788028, 201.122171, 210.36892, 218.17626, 224.192174, 233.045212, 237.88697, 237.390976, 230.230754, 215.079831, 199.787897, 171.558274, 126.770449, 61.8039072, -26.9618645, -134.815482, -283.892789, -473.420767, -711.46918, -1024.16181, -1429.51657, -1945.55134, -2578.94959, -3361.63966, -4337.91474, -5497.42175, -6957.83449, -8752.17452, -10961.3722, -13747.5368, -17235.5149, -21472.4625, -26563.0842, -32896.9513, -40702.4982, -49464.3667, -60269.2139, -73602.9409, -89951.4492, -109800.64, -146554.201, -184550.8, -221046.893, -253298.933, -278563.377, -284677.071, -284595.816, -279613.039, -273839.43, -267453.253, -259458.912, -250093.363, -239593.563, -228196.466, -216139.03, -201763.804, -188951.128, -177136.183, -165754.154, -154240.221, -142029.567, -128557.376, -119007.337, -110440.23},
        {-19.6174481, -19.7443413, -19.8438393, -19.9168111, -19.9641254, -19.986651, -19.9852565, -19.9608108, -19.9141827, -19.8462408, -19.7578541, -19.6498911, -19.5232207, -19.3787117, -19.2172329, -19.0396529, -18.8468405, -18.6396646, -18.4189938, -18.185697, -17.9406429, -17.6847002, -17.4187378, -17.1436244, -16.8602287, -16.5694195, -16.2720657, -15.9690358, -15.6611988, -15.3494234, -15.0345783, -14.7175323, -14.3991542, -14.0803127, -13.7881108, -13.5097484, -13.23211, -12.955213, -12.6790749, -12.4037129, -12.1291447, -11.8553875, -11.5824588, -11.310376, -11.0391565, -10.7688177, -10.499377, -10.2308519, -9.9632598, -9.69661804, -9.43094407, -9.16254434, -8.89522768, -8.62909194, -8.36423497, -8.10075464, -7.83874881, -7.57831533, -7.31955207, -7.06255688, -6.80742763, -6.55426217, -6.30315836, -6.05421407, -5.80752714, -5.56319544, -5.32131683, -5.07950932, -4.83545045, -4.5943526, -4.35647437, -4.12207439, -3.89141127, -3.66474364, -3.44233011, -3.22442929, -3.01129982, -2.8032003, -2.60038935, -2.40312559, -2.21166764, -2.02627412, -1.84720364, -1.67471482, -1.78879005, -2.04394245, -2.29132308, -2.52211297, -2.72749318, -2.89864474, -3.02674871, -3.10298611, -3.11853801, -3.06458544, -2.93230944, -2.71289107, -2.39751135, -1.97735134, -1.44359209, -0.787414625, 0, 0.887906155, 1.93097148, 3.13776268, 4.51684643, 6.07678942, 7.82615835, 9.7735199, 11.9274408, 14.2964876, 16.8892272, 19.7142262, 22.7800512, 26.095269, 29.6684462, 33.5081496, 37.6229459, 42.1099057, 46.9902392, 52.0723254, 57.383696, 62.9518831, 68.8044183, 74.9688335, 81.4726607, 88.3434315, 95.608678, 103.295932, 111.432725, 120.046589, 129.585, 139.501566, 149.697837, 160.07536, 171.522041, 183.446251, 194.993331, 205.768923, 215.378669, 223.428209, 229.523186, 238.147036, 242.54306, 241.362548, 233.256789, 216.877074, 199.159922, 168.400087, 121.177553, 54.0723039, -36.3356778, -143.687962, -293.134447, -486.985224, -734.331619, -1058.10998, -1475.04052, -2001.84346, -2640.44466, -3425.72103, -4408.68879, -5587.76973, -7069.61844, -8876.51224, -11085.208, -13864.775, -17341.9097, -21574.0941, -26670.08, -32992.4057, -40760.0532, -49533.8692, -60292.2234, -73481.9143, -89549.7405, -108942.5, -144281.293, -180795.041, -215886.969, -246960.299, -271418.255, -277824.37, -278314.685, -274080.864, -268823.326, -262740.686, -254986.24, -245810.867, -235465.448, -224200.859, -212267.982, -198534.97, -186049.988, -174340.715, -162934.832, -151360.017, -139143.952, -125814.316, -118379.561, -109156.614},
        {-17.4163321, -17.5419591, -17.6422327, -17.7179426, -17.7698788, -17.7988312, -17.8055895, -17.7909437, -17.7556836, -17.7005991, -17.6264801, -17.5341163, -17.4242977, -17.2978141, -17.1554554, -16.9980114, -16.8262721, -16.6410273, -16.4430667, -16.2331804, -16.0121582, -15.7807898, -15.5398653, -15.2901744, -15.032507, -14.767653, -14.4964022, -14.2195446, -13.9378698, -13.6521679, -13.3632287, -13.0718421, -12.7787978, -12.4848859, -12.2137432, -11.9542556, -11.6950341, -11.4361272, -11.1775833, -10.9194508, -10.6617781, -10.4046137, -10.148006, -9.89200341, -9.63665436, -9.38200728, -9.1281106, -8.87501274, -8.62276214, -8.37140721, -8.12099639, -7.86743937, -7.61501304, -7.36385553, -7.11410498, -6.86589955, -6.61937739, -6.37467663, -6.13193542, -5.89129191, -5.65288424, -5.41685057, -5.18332903, -4.95245778, -4.72437495, -4.4992187, -4.27712716, -4.05669207, -3.83654226, -3.62000527, -3.4073195, -3.19872334, -2.99445519, -2.79475344, -2.59985648, -2.41000272, -2.22543055, -2.04637835, -1.87308453, -1.70578747, -1.54472558, -1.39013725, -1.24226088, -1.10133485, -1.2578859, -1.56090244, -1.85588381, -2.13364801, -2.38501301, -2.60079678, -2.77181732, -2.88889259, -2.94284059, -2.92447928, -2.82462665, -2.63410068, -2.34371935, -1.94430063, -1.42666252, -0.781622981, 0, 0.873332792, 1.91458001, 3.13168285, 4.53258251, 6.12522021, 7.91753714, 9.91747451, 12.1329735, 14.5719754, 17.2424213, 20.1522524, 23.30941, 26.7218353, 30.3974694, 34.3442536, 38.570129, 43.1856955, 48.2135598, 53.4294591, 58.8633329, 64.5451206, 70.5047618, 76.7721959, 83.3773624, 90.3502007, 97.7206504, 105.518651, 113.774142, 122.517062, 132.227962, 142.320946, 152.690772, 163.232197, 174.871863, 186.988584, 198.686002, 209.549309, 219.163695, 227.11435, 232.986465, 241.380113, 245.339333, 243.468143, 234.37056, 216.650603, 196.979098, 163.876551, 113.930278, 43.7275949, -50.1441839, -161.664496, -315.208796, -511.890377, -760.456508, -1085.24049, -1503.57782, -2032.80399, -2674.38259, -3465.12825, -4459.17045, -5663.16574, -7171.75101, -8996.56964, -11209.2648, -13983.1458, -17444.8556, -21662.9473, -26749.0034, -33027.6026, -40702.8408, -49343.4491, -59886.6368, -72756.4342, -88376.8712, -107171.978, -141094.265, -176157.162, -209902.579, -239872.425, -263608.61, -270376.145, -271511.772, -268104.644, -263457.517, -257783.325, -250372.669, -241484.603, -231378.177, -220312.445, -208546.458, -195507.612, -183237.635, -171460.388, -159899.729, -148279.519, -136323.618, -123755.885, -117098.201, -108441.665},
        {-15.5373657, -15.6549551, -15.7490578, -15.8204003, -15.8697091, -15.8977108, -15.9051317, -15.8926984, -15.8611374, -15.8111752, -15.7435383, -15.6589531, -15.5581463, -15.4418443, -15.3107736, -15.1656607, -15.007232, -14.8362142, -14.6533337, -14.4593169, -14.2548905, -14.0407808, -13.8177144, -13.5864179, -13.3476176, -13.1020401, -12.8504119, -12.5934595, -12.3319094, -12.0664881, -11.7979221, -11.5269379, -11.2542619, -10.9806208, -10.7274866, -10.4847766, -10.2421589, -9.99968672, -9.75741326, -9.51539182, -9.27367567, -9.03231807, -8.79137227, -8.55089155, -8.31092917, -8.07153839, -7.83277247, -7.59468468, -7.35732829, -7.12075655, -6.88502272, -6.64431206, -6.40467305, -6.16628616, -5.92933187, -5.69399064, -5.46044294, -5.22886925, -4.99945004, -4.77236577, -4.54779692, -4.32592396, -4.10692736, -3.89098758, -3.67828511, -3.46900041, -3.26331394, -3.06040488, -2.85947644, -2.66277419, -2.47054352, -2.28302982, -2.10047846, -1.92313482, -1.7512443, -1.58505226, -1.42480409, -1.27074517, -1.12312088, -0.982176601, -0.84815772, -0.721309614, -0.601877664, -0.490107253, -0.688961925, -1.04096071, -1.3849842, -1.71145399, -2.01079167, -2.27341884, -2.48975709, -2.65022802, -2.74525322, -2.76525429, -2.70065282, -2.54187041, -2.27932865, -1.90344913, -1.40465345, -0.773363212, 0, 0.893561393, 1.96136787, 3.21115693, 4.65066606, 6.28763277, 8.12979454, 10.1848889, 12.4606533, 14.9648252, 17.7051422, 20.6893418, 23.9251614, 27.4203385, 31.1826107, 35.2197154, 39.5393901, 44.2587586, 49.4011849, 54.7269854, 60.2673373, 66.0534182, 72.1164054, 78.4874763, 85.1978083, 92.2785789, 99.7609654, 107.676145, 116.055296, 124.929595, 134.817609, 145.084417, 155.614978, 166.294253, 178.053663, 190.254938, 201.980751, 212.802124, 222.290078, 230.015635, 235.549818, 243.668978, 247.154575, 244.559198, 234.435435, 215.335873, 193.822298, 158.435799, 105.726667, 32.245188, -65.4583482, -182.18694, -340.605218, -539.584819, -787.190759, -1110.25787, -1527.19371, -2056.40585, -2699.97917, -3496.35362, -4501.77572, -5731.05355, -7266.27135, -9108.43655, -11323.652, -14088.3209, -17528.217, -21721.2104, -26780.6552, -32995.5516, -40555.5447, -49026.8143, -59318.3879, -71835.7252, -86984.2859, -105169.53, -137740.414, -171423.026, -203886.951, -232801.774, -255837.081, -262895.374, -264591.376, -261923.422, -257859.909, -252619.354, -245599.417, -237059.658, -227259.637, -216458.916, -204917.054, -192634.996, -180558.186, -168648.227, -156866.722, -145175.273, -133535.483, -121908.955, -115599.468, -107842.847},
        {-13.9407675, -14.0445437, -14.1265635, -14.1875025, -14.2280361, -14.2488397, -14.2505889, -14.2339591, -14.1996259, -14.1482646, -14.0805508, -13.9971599, -13.8987674, -13.7860488, -13.6596796, -13.5203353, -13.3686912, -13.205423, -13.031206, -12.8467158, -12.6526278, -12.4496175, -12.2383603, -12.0195319, -11.7938075, -11.5618628, -11.3243732, -11.0820141, -10.8354611, -10.5853895, -10.332475, -10.0773929, -9.82081881, -9.56342811, -9.32547394, -9.09743118, -8.86953938, -8.64183869, -8.41436925, -8.18717121, -7.96028471, -7.7337499, -7.50760691, -7.2818959, -7.05665701, -6.83193039, -6.60775617, -6.3841745, -6.16122553, -5.93894941, -5.71738626, -5.48830853, -5.26020329, -5.03328994, -4.80778783, -4.58391634, -4.36189484, -4.14194271, -3.92427932, -3.70912404, -3.49669624, -3.28721529, -3.08090058, -2.87797147, -2.67864732, -2.48314753, -2.29169145, -2.10368052, -1.91853579, -1.73816359, -1.56283633, -1.3928264, -1.22840619, -1.0698481, -0.917424532, -0.77140788, -0.632070541, -0.499684912, -0.374523389, -0.25685837, -0.14696225, -0.0451074279, 0.0484337008, 0.133388739, -0.10664354, -0.507221773, -0.900104974, -1.27530655, -1.62283991, -1.93271846, -2.19495561, -2.39956477, -2.53655935, -2.59595275, -2.56775839, -2.44198967, -2.20866, -1.85778279, -1.37937145, -0.763439383, 0, 0.94971405, 2.07272432, 3.3770763, 4.87081547, 6.56198731, 8.4586373, 10.5688109, 12.9005537, 15.4619111, 18.2609285, 21.3056515, 24.6041256, 28.1643962, 31.9945089, 36.102509, 40.4964421, 45.2878402, 50.4995326, 55.905488, 61.5359278, 67.4210729, 73.5911447, 80.0763643, 86.9069528, 94.1131314, 101.725121, 109.773143, 118.287419, 127.29817, 137.35496, 147.778241, 158.442431, 169.221948, 181.032886, 193.228824, 204.885565, 215.565023, 224.829115, 232.239756, 237.358863, 245.173762, 248.169747, 244.847237, 233.706657, 213.248428, 189.995665, 152.420424, 97.017455, 20.2815062, -81.292673, -203.503195, -366.982286, -568.020616, -813.889009, -1134.2761, -1548.80539, -2077.10038, -2722.49252, -3524.60009, -4540.81447, -5793.22333, -7352.02824, -9208.55388, -11423.4717, -14174.9577, -17587.2513, -21745.2062, -26762.851, -32898.6277, -40329.2901, -48616.9516, -58647.5659, -70810.6465, -85495.707, -103092.261, -134418.934, -166828.85, -198104.244, -226027.351, -248380.408, -255613.821, -257729.539, -255652.011, -252098.342, -247283.386, -240675.056, -232525.552, -223087.074, -212611.825, -201352.004, -189938.92, -178070.339, -165984.235, -153918.582, -142111.353, -130800.521, -120224.061, -113979.744, -107239.877},
        {-12.5705434, -12.6551499, -12.7197145, -12.7648675, -12.7912393, -12.7994601, -12.79016, -12.7639695, -12.7215187, -12.663438, -12.5903575, -12.5029076, -12.4017185, -12.2874204, -12.1606436, -12.0220185, -11.8721752, -11.7117439, -11.5413551, -11.3616389, -11.1732256, -10.9767454, -10.7728286, -10.5621055, -10.3452063, -10.1227614, -9.89540084, -9.66375504, -9.42845422, -9.19012863, -8.94940854, -8.70692421, -8.4633059, -8.21918388, -7.99377499, -7.77802526, -7.5626573, -7.34769822, -7.1331751, -6.91911503, -6.70554511, -6.49249241, -6.27998404, -6.06804708, -5.85670862, -5.64599575, -5.43593557, -5.22655517, -5.01788162, -4.80994203, -4.60276349, -4.38679947, -4.17185821, -3.95817434, -3.74598249, -3.53551729, -3.32701336, -3.12070532, -2.91682782, -2.71561546, -2.51730289, -2.32212473, -2.13031561, -1.94211015, -1.75774298, -1.57744873, -1.40146203, -1.22887178, -1.05879438, -0.893827436, -0.734279855, -0.580460534, -0.432678374, -0.291242273, -0.156461134, -0.0286438548, 0.0919006627, 0.204863519, 0.309935813, 0.406808645, 0.495173115, 0.574720322, 0.645141366, 0.706127347, 0.429053924, -0.0153244111, -0.452307064, -0.871548468, -1.26270306, -1.61542526, -1.91936951, -2.16419025, -2.3395419, -2.4350789, -2.44045568, -2.34532668, -2.13934632, -1.81216904, -1.35344927, -0.752841445, 0, 1.08108374, 2.31683008, 3.7169942, 5.29133129, 7.04959653, 9.0015451, 11.1569322, 13.525513, 16.1170426, 18.9412763, 22.0079693, 25.3268767, 28.9077536, 32.7603554, 36.8944372, 41.3197541, 46.1013685, 51.2569367, 56.6711141, 62.3655077, 68.361724, 74.6813698, 81.3460518, 88.3773767, 95.7969513, 103.626382, 111.887276, 120.60124, 129.789881, 139.981145, 150.50464, 161.230073, 172.027147, 183.796296, 195.891861, 207.39332, 217.861159, 226.855864, 233.937922, 238.667816, 246.296552, 248.9622, 245.111886, 233.192732, 211.651863, 186.613275, 146.928001, 89.1239484, 9.72902225, -94.7288711, -220.17074, -386.848869, -589.989975, -836.624718, -1157.88347, -1574.07633, -2105.51338, -2757.31558, -3568.43425, -4594.39066, -5859.3321, -7427.4232, -9285.25962, -11490.1674, -14219.3212, -17594.8833, -21702.6899, -26655.4852, -32699.2091, -40004.4583, -48152.4981, -57993.9235, -69903.3307, -84255.316, -101424.476, -131896.489, -163407.099, -193803.131, -220931.41, -242638.763, -249632.955, -251659.243, -249624.052, -246245.528, -241712.952, -235460.72, -227715.856, -218705.384, -208656.329, -197795.717, -187889.748, -176440.833, -164091.846, -151485.663, -139265.16, -128073.212, -118552.694, -112527.916, -106295.814},
        {-11.3905937, -11.4518006, -11.4946842, -11.5198329, -11.5278352, -11.5192793, -11.4947538, -11.4548469, -11.4001473, -11.3312432, -11.248723, -11.1531753, -11.0451883, -10.9253505, -10.7942503, -10.6524761, -10.5006164, -10.3392595, -10.1689938, -9.99040781, -9.80408987, -9.61062841, -9.41061184, -9.20462856, -8.99326699, -8.77711554, -8.55676261, -8.33279663, -8.10580599, -7.87637911, -7.6451044, -7.41257027, -7.17936512, -6.94607738, -6.73080793, -6.52502059, -6.31996869, -6.11567232, -5.91215159, -5.70942659, -5.50751741, -5.30644416, -5.10622692, -4.9068858, -4.7084409, -4.51091229, -4.3143201, -4.1186844, -3.9240253, -3.73036289, -3.53771728, -3.33669668, -3.1369371, -2.93866266, -2.74209749, -2.54746571, -2.35499145, -2.16489885, -1.97741202, -1.79275509, -1.6111522, -1.43282746, -1.25800501, -1.08690897, -0.919763465, -0.756792632, -0.598220592, -0.442353315, -0.28754282, -0.137969296, 0.00601878319, 0.144072944, 0.275844713, 0.400985617, 0.519147181, 0.629980932, 0.733138397, 0.828271101, 0.915030571, 0.993068334, 1.06203592, 1.12158484, 1.17136664, 1.21103283, 0.901541507, 0.419014867, -0.0563540401, -0.513922435, -0.943047536, -1.33308657, -1.67339674, -1.95333528, -2.16225942, -2.28952635, -2.32449332, -2.25651753, -2.07495621, -1.76916658, -1.32850585, -0.742331252, 0, 1.28012458, 2.67990716, 4.21212255, 5.88954557, 7.72495103, 9.73111377, 11.9208086, 14.3068103, 16.9018937, 19.7188337, 22.770405, 26.0693825, 29.628541, 33.4606553, 37.5785002, 41.9948505, 46.6876816, 51.6647967, 57.0176842, 62.7516616, 68.8720469, 75.3841578, 82.293312, 89.6048273, 97.3240215, 105.456212, 114.006718, 122.980855, 132.383942, 142.665363, 153.223545, 163.930586, 174.658585, 186.304495, 198.227986, 209.515434, 219.731482, 228.440771, 235.207941, 239.597637, 247.142535, 249.622883, 245.43566, 232.97785, 210.646435, 183.797804, 142.12968, 82.2991926, 0.963473041, -105.220349, -231.275506, -399.032951, -604.61668, -855.497624, -1182.31336, -1605.30011, -2144.6941, -2807.40502, -3630.02439, -4663.68176, -5929.38361, -7491.40657, -9336.83473, -11521.9428, -14220.2453, -17551.4417, -21596.4714, -26464.9848, -32408.7532, -39599.1562, -47665.1574, -57404.0012, -69174.8267, -83336.773, -100248.979, -130243.414, -161213.179, -191024.207, -217542.428, -238633.775, -244987.043, -246430.057, -243904.665, -240368.199, -235964.709, -229998.463, -222655.502, -214121.868, -204583.602, -194226.744, -186407.418, -175590.291, -162927.657, -149571.806, -136675.032, -125389.626, -116867.88, -111296.65, -104939.577},
        {-10.3900562, -10.4256586, -10.444552, -10.447289, -10.4344223, -10.4065045, -10.3640884, -10.3077266, -10.2379718, -10.1553766, -10.0604937, -9.95387578, -9.8360755, -9.70764553, -9.56913854, -9.42110718, -9.26410414, -9.09868207, -8.92539365, -8.74479154, -8.5574284, -8.36385691, -8.16462974, -7.96029955, -7.751419, -7.53854077, -7.32221752, -7.10300192, -6.88144664, -6.65810435, -6.4335277, -6.20826938, -5.98288204, -5.75791836, -5.55065618, -5.35293378, -5.15638843, -4.96103004, -4.7668685, -4.57391372, -4.3821756, -4.19166404, -4.00238896, -3.81436024, -3.6275878, -3.44208154, -3.25785136, -3.07490716, -2.89325885, -2.71291633, -2.5338895, -2.34701575, -2.16167633, -1.97808, -1.79643551, -1.61695159, -1.439837, -1.26530048, -1.09355078, -0.924796633, -0.759246797, -0.597110013, -0.438595028, -0.283910587, -0.133265435, 0.0131316823, 0.155072018, 0.295141761, 0.435859992, 0.571253627, 0.700932736, 0.824507389, 0.941587657, 1.05178361, 1.15470532, 1.24996285, 1.33716628, 1.41592567, 1.4858511, 1.54655264, 1.59764036, 1.63872432, 1.6694146, 1.68932126, 1.34932055, 0.830724584, 0.319037027, -0.174814971, -0.639904256, -1.06530367, -1.44008607, -1.75332429, -1.99409119, -2.1514596, -2.21450237, -2.17229235, -2.01390239, -1.72840533, -1.30487401, -0.732381285, 0, 1.47987135, 3.04420551, 4.70876959, 6.48933074, 8.40165605, 10.4615127, 12.6846677, 15.0868882, 17.6839414, 20.4915944, 23.5256143, 26.8017682, 30.3358232, 34.1435464, 38.2407051, 42.6430662, 47.2379783, 52.0228631, 57.3039582, 63.0695126, 69.3077751, 76.0069945, 83.1554197, 90.7412995, 98.7528828, 107.178418, 116.006155, 125.224342, 134.821228, 145.158698, 155.714364, 166.364385, 176.984918, 188.484893, 200.224082, 211.28685, 221.239523, 229.648427, 236.079888, 240.100232, 247.440267, 249.62569, 245.016732, 231.973624, 208.856597, 180.414249, 137.021354, 75.4410507, -7.56352103, -115.229222, -241.648184, -410.378885, -619.056688, -875.715163, -1209.61755, -1640.4954, -2188.08026, -2859.94097, -3690.73605, -4728.39242, -5991.13552, -7543.62496, -9374.17375, -11538.2644, -14205.1946, -17492.0188, -21473.8311, -26256.8074, -32100.9792, -39179.3488, -47177.8373, -56831.4634, -68481.834, -82470.5563, -99139.2372, -128646.686, -159063.007, -188275.507, -214171.492, -234638.269, -240356.505, -241224.573, -238218.337, -234525.508, -230246.43, -224559.846, -217612.181, -209549.859, -200519.305, -190666.944, -184713.6, -174492.83, -161602.327, -147639.782, -134202.888, -122889.337, -115296.821, -110038.944, -103437.353},
        {-9.54228205, -9.55153785, -9.54556654, -9.52489066, -9.49003274, -9.44151532, -9.37986092, -9.30559207, -9.21923131, -9.12130118, -9.01232419, -8.89282288, -8.76331979, -8.62433744, -8.47639837, -8.32002511, -8.15574019, -7.98406613, -7.80552548, -7.62064077, -7.42993452, -7.23392926, -7.03314754, -6.82811187, -6.61934479, -6.40736884, -6.19270655, -5.97588043, -5.75741304, -5.53782689, -5.31764453, -5.09738848, -4.87758126, -4.65874543, -4.45760402, -4.26623855, -4.07653582, -3.88849262, -3.70210574, -3.51737199, -3.33428815, -3.15285102, -2.97305739, -2.79490405, -2.61838781, -2.44350544, -2.27025375, -2.09862952, -1.92862956, -1.76025065, -1.5934896, -1.41920089, -1.24672181, -1.07624733, -0.90797244, -0.742092111, -0.578801327, -0.418295069, -0.260768316, -0.10641605, 0.0445667498, 0.191985102, 0.335644027, 0.475348543, 0.610903671, 0.742114429, 0.868785839, 0.994329942, 1.12207218, 1.24437835, 1.36081966, 1.47096728, 1.57439241, 1.67066625, 1.75935997, 1.84004478, 1.91229186, 1.97567241, 2.02975761, 2.07411865, 2.10832674, 2.13195305, 2.14456878, 2.14574511, 1.77643145, 1.22305301, 0.676345356, 0.147517657, -0.352220933, -0.811661256, -1.21959415, -1.56481047, -1.83610104, -2.02225672, -2.11206833, -2.09432674, -1.95782277, -1.69134726, -1.28369108, -0.72364504, 0, 1.65052771, 3.35701327, 5.13763061, 7.01055364, 8.99395629, 11.1060125, 13.3648961, 15.7887812, 18.3958415, 21.2042511, 24.2321838, 27.4978136, 31.0193143, 34.81486, 38.9026245, 43.3007818, 47.8165612, 52.4380017, 57.6604137, 63.4567688, 69.8000384, 76.6631941, 84.0192074, 91.8410497, 100.101693, 108.774108, 117.831267, 127.246141, 136.991701, 147.348672, 157.869431, 168.434622, 178.924893, 190.285662, 201.865189, 212.727793, 222.434688, 230.547088, 236.626206, 240.233256, 247.1539, 248.830506, 243.606462, 229.825154, 205.82997, 176.114332, 131.334083, 68.2951006, -16.1967366, -125.335551, -252.525915, -422.716503, -635.441547, -899.289727, -1241.33589, -1680.35498, -2235.12193, -2911.97235, -3744.94763, -4781.6593, -6040.00588, -7583.28944, -9400.28096, -11545.2982, -14183.7241, -17429.9611, -21354.2318, -26059.4847, -31810.6594, -38781.0641, -46718.9477, -56288.0716, -67808.8798, -81601.8166, -97987.3259, -126845.855, -156547.844, -185023.736, -210203.976, -230019.007, -235268.642, -235767.707, -232502.632, -228781.35, -224671.361, -219272.866, -212701.728, -205073.81, -196504.975, -187111.085, -182466.757, -172714.641, -159755.68, -145490.819, -131821.003, -120647.174, -113870.278, -108695.246, -101844.358},
        {-8.81868184, -8.80389596, -8.77507468, -8.73272026, -8.67733499, -8.60942115, -8.52948103, -8.43801692, -8.33553109, -8.22252582, -8.09950341, -7.96696613, -7.82541628, -7.67535612, -7.51728795, -7.35171404, -7.17913669, -7.00005817, -6.81498077, -6.62440678, -6.42883846, -6.22877812, -6.02472803, -5.81719047, -5.60666774, -5.3936621, -5.17867585, -4.96221127, -4.74477064, -4.52685625, -4.30897037, -4.0916153, -3.87529332, -3.6605067, -3.46389755, -3.27755877, -3.09338925, -2.91136753, -2.73147211, -2.5536815, -2.37797424, -2.20432882, -2.03272376, -1.86313759, -1.69554881, -1.52993594, -1.3662775, -1.204552, -1.04473796, -0.886813894, -0.730758313, -0.566737459, -0.404754834, -0.245001666, -0.0876691795, 0.0670513983, 0.218968841, 0.367891923, 0.513629418, 0.655990098, 0.794782739, 0.929816113, 1.060899, 1.18784016, 1.31044838, 1.42853242, 1.54190107, 1.65444619, 1.7699616, 1.87983502, 1.98361052, 2.0808322, 2.17104412, 2.25379039, 2.32861507, 2.39506226, 2.45267604, 2.50100049, 2.53957969, 2.56795773, 2.58567869, 2.59228665, 2.58732571, 2.57033994, 2.1726039, 1.58554251, 1.00500696, 0.442492398, -0.0905060475, -0.582493235, -1.02197403, -1.39745329, -1.69743588, -1.91042666, -2.0249305, -2.02945225, -1.91249679, -1.66256897, -1.26817366, -0.717815714, 0, 1.74191549, 3.52939636, 5.38150421, 7.31730063, 9.35584722, 11.5162056, 13.8174373, 16.278604, 18.9187672, 21.7569885, 24.8123296, 28.1038521, 31.6506174, 35.4716873, 39.5861233, 44.012987, 48.5136762, 53.0704432, 58.2851609, 64.1216772, 70.5438399, 77.5154969, 85.0004963, 92.9626858, 101.365913, 110.174027, 119.350874, 128.860304, 138.666163, 148.99892, 159.458042, 169.92739, 180.290827, 191.581119, 203.107677, 213.873533, 223.417878, 231.279903, 236.998796, 240.113749, 246.169089, 246.871218, 240.584407, 225.672927, 200.501051, 170.0947, 124.491082, 60.3890576, -25.5125154, -136.514778, -266.076199, -439.370524, -658.182662, -931.477979, -1282.88193, -1729.4379, -2288.18926, -2959.79718, -3781.41047, -4808.0494, -6065.31792, -7608.08282, -9420.49469, -11553.6917, -14172.319, -17388.9199, -21274.403, -25930.5175, -31611.6747, -38485.9316, -46366.6777, -55831.2017, -67170.3303, -80674.8898, -96635.7067, -124362.247, -152872.039, -180201.247, -204386.039, -223462.582, -228714.165, -229431.749, -226545.593, -223179.94, -219396.594, -214343.196, -208127.259, -200856.292, -192637.808, -183579.317, -179041.404, -169439.827, -156694.241, -142724.303, -129449.667, -118789.989, -112664.924, -107105.927, -100222.082},
        {-8.19190423, -8.15741797, -8.10972261, -8.04931134, -7.97667728, -7.8923136, -7.79671345, -7.69036998, -7.57377635, -7.4474257, -7.31181119, -7.16742598, -7.01476321, -6.85431604, -6.68657762, -6.51204111, -6.33119965, -6.1445464, -5.95257452, -5.75577715, -5.55464745, -5.34967857, -5.14136366, -4.93019588, -4.71666838, -4.50127431, -4.28450682, -4.06685907, -3.84882422, -3.6308954, -3.41356578, -3.19732851, -2.98267675, -2.77010363, -2.57670284, -2.39431811, -2.21459362, -2.03748381, -1.86294312, -1.69092597, -1.52138682, -1.35428008, -1.18956019, -1.02718159, -0.867098703, -0.709265975, -0.553637835, -0.400168717, -0.248813054, -0.0995252822, 0.0477401659, 0.204572031, 0.359222471, 0.511486818, 0.661160405, 0.808038564, 0.951916628, 1.09258993, 1.2298538, 1.36350357, 1.49333458, 1.61914216, 1.74072163, 1.85786834, 1.97037762, 2.07804479, 2.18066519, 2.28205124, 2.38591888, 2.48377854, 2.57516513, 2.65961359, 2.73665882, 2.80583575, 2.8666793, 2.9187244, 2.96150597, 2.99455892, 3.01741818, 3.02961868, 3.03069533, 3.02018305, 2.99761677, 2.96253141, 2.53669854, 1.91648051, 1.30278007, 0.707391298, 0.142108259, -0.381274979, -0.850964346, -1.25516577, -1.58208519, -1.81992854, -1.95690173, -1.98121071, -1.88106141, -1.64465976, -1.26021169, -0.715923122, 0, 1.71686113, 3.49553225, 5.35374944, 7.30924877, 9.37976631, 11.5830381, 13.9368003, 16.458789, 19.1667401, 22.0783898, 25.2114742, 28.5837293, 32.2128913, 36.1166961, 40.3128799, 44.8191787, 49.403016, 54.046369, 59.3331569, 65.2274774, 71.6934283, 78.695107, 86.1966114, 94.1620391, 102.555488, 111.341055, 120.482838, 129.944936, 139.691445, 149.952121, 160.327996, 170.703471, 180.962945, 192.292705, 203.926207, 214.74974, 224.257138, 231.942233, 237.298862, 239.820858, 244.421576, 243.52591, 235.567353, 218.979398, 192.195538, 161.842358, 116.111987, 51.389644, -35.9394491, -149.490072, -283.872922, -462.711852, -690.236777, -975.466579, -1337.19761, -1789.83634, -2347.78923, -3000.19259, -3792.46626, -4797.59758, -6060.28366, -7616.66302, -9438.66404, -11571.2333, -14183.0425, -17385.9717, -21260.0571, -25908.923, -31552.9529, -38346.4848, -46167.4547, -55489.1256, -66561.5128, -79634.6316, -94958.4974, -120856.427, -147486.809, -173081.334, -195871.69, -214089.569, -220026.156, -221813.981, -220230.665, -217778.047, -214551.245, -209926.95, -204036.199, -197010.031, -188979.484, -180075.594, -173993.227, -164096.282, -151937.197, -139068.404, -127042.341, -117411.44, -111728.137, -105175.21, -98628.0138},
        {-7.64652465, -7.58897863, -7.51940704, -7.43827684, -7.34605501, -7.24320853, -7.13020437, -7.00750951, -6.87559091, -6.73491557, -6.58595044, -6.42916251, -6.26501876, -6.09398614, -5.91653165, -5.73312225, -5.54422493, -5.35030665, -5.15183439, -4.94927512, -4.74309583, -4.53376348, -4.32174505, -4.10750752, -3.89151785, -3.67424303, -3.45615003, -3.23770583, -3.01937739, -2.8016317, -2.58493573, -2.36975645, -2.15656084, -1.94581587, -1.7542426, -1.57383907, -1.39640798, -1.22188883, -1.05022112, -0.881344357, -0.715198029, -0.55172164, -0.390854688, -0.232536673, -0.0767070935, 0.0766945515, 0.227728763, 0.376456041, 0.522936886, 0.667231801, 0.809401285, 0.962944754, 1.11419246, 1.26291359, 1.4088773, 1.55185276, 1.69160916, 1.82791565, 1.96054141, 2.08925562, 2.21382744, 2.33402605, 2.44962062, 2.56038031, 2.66607431, 2.76647178, 2.86134189, 2.95428573, 3.04881237, 3.13678797, 3.21773328, 3.29116906, 3.35661609, 3.41359512, 3.46162692, 3.50023226, 3.5289319, 3.5472466, 3.55469712, 3.55080424, 3.53508872, 3.50707131, 3.46627279, 3.41221392, 2.95554766, 2.29840798, 1.64760276, 1.01527227, 0.413556772, -0.145403449, -0.649468125, -1.08649698, -1.44434975, -1.71088616, -1.87396593, -1.9214488, -1.84119449, -1.62106272, -1.24891324, -0.712605752, 0, 1.66348349, 3.41226574, 5.26215808, 7.22897184, 9.32851835, 11.5766089, 13.9890549, 16.5816677, 19.3702585, 22.3706386, 25.5986196, 29.0700125, 32.8006289, 36.8062799, 41.102777, 45.7059315, 50.3987909, 55.1639147, 60.5436035, 66.5037908, 73.0104099, 80.0293942, 87.5266771, 95.468192, 103.819872, 112.547651, 121.617463, 130.995239, 140.646915, 150.81253, 161.083404, 171.34324, 181.475739, 192.795869, 204.471698, 215.289622, 224.703963, 232.16904, 237.139174, 239.068687, 242.329644, 239.961916, 230.457658, 212.309022, 184.008164, 153.641762, 107.708813, 42.3028382, -46.482641, -162.554102, -301.770815, -485.921944, -720.545772, -1014.4988, -1383.41226, -1839.87564, -2396.47846, -3034.55165, -3805.04753, -4793.01864, -6055.58124, -7617.08261, -9444.29029, -11578.1604, -14185.3577, -17374.8444, -21230.7622, -25855.6386, -31443.0101, -38134.9332, -45861.7008, -55009.7125, -65794.3276, -78430.9056, -93134.8062, -117330.421, -142221.82, -166222.103, -187744.372, -205201.728, -211836.227, -214679.379, -214360.45, -212753.431, -209994.882, -205702.725, -200039.452, -193167.554, -185249.522, -176447.849, -168623.404, -158460.683, -147052.604, -135492.084, -124872.04, -116285.389, -110825.049, -103422.55, -97081.052},
        {-7.13641705, -7.04355943, -6.94125091, -6.82987306, -6.70980749, -6.58143577, -6.44513949, -6.30130024, -6.1502996, -5.99251916, -5.82834051, -5.65814523, -5.48231491, -5.30123113, -5.11527549, -4.92482957, -4.73027495, -4.53199321, -4.33036596, -4.12577477, -3.91860123, -3.70922693, -3.49803345, -3.28540238, -3.0717153, -2.8573538, -2.64269948, -2.4281339, -2.21403867, -2.00079536, -1.78878557, -1.57839088, -1.36999287, -1.16397314, -0.974166716, -0.79394549, -0.616519644, -0.441834175, -0.269834079, -0.100464351, 0.0663300117, 0.230604014, 0.39241266, 0.551810953, 0.708853898, 0.863596499, 1.01609376, 1.16640068, 1.31457227, 1.46066354, 1.60472948, 1.76273995, 1.9184901, 2.07168993, 2.22204945, 2.36927865, 2.51308754, 2.65318611, 2.78928438, 2.92109233, 3.04831998, 3.17067732, 3.28787435, 3.39962107, 3.5056275, 3.60560362, 3.69925944, 3.78981152, 3.88039221, 3.96347952, 4.03855613, 4.10510473, 4.162608, 4.21054863, 4.2484093, 4.2756727, 4.29182151, 4.29633842, 4.28870612, 4.26840729, 4.23492462, 4.18774079, 4.12633849, 4.0502004, 3.55433858, 2.84897852, 2.14920715, 1.46763301, 0.816864615, 0.209510519, -0.341820748, -0.824520651, -1.22598065, -1.53359222, -1.73474681, -1.8168359, -1.76725095, -1.57338342, -1.22262477, -0.702366478, 0, 1.62927128, 3.36555098, 5.22284798, 7.21517111, 9.35652924, 11.6609312, 14.1423859, 16.8149022, 19.6924888, 22.7891548, 26.1189088, 29.6957598, 33.5337167, 37.6467883, 42.0489834, 46.7543109, 51.5484732, 56.4128641, 61.8778996, 67.9086659, 74.470249, 81.527735, 89.0462099, 96.9907598, 105.326471, 114.018429, 123.03172, 132.33143, 141.882646, 151.924152, 162.04698, 172.134104, 182.068501, 193.192502, 204.659405, 215.233349, 224.359501, 231.483031, 236.04911, 237.502906, 240.199938, 237.180572, 226.929258, 207.930444, 178.668581, 147.593612, 100.733117, 34.0801727, -56.372146, -174.630763, -317.404783, -504.918509, -740.524347, -1032.48113, -1399.06499, -1854.0545, -2411.22821, -3058.52508, -3841.43595, -4833.6403, -6078.64517, -7615.11828, -9429.91301, -11568.0872, -14173.323, -17344.0324, -21148.0535, -25675.8718, -31130.0765, -37649.8448, -45189.0847, -54100.7452, -64588.1347, -76854.5618, -91103.3352, -114317.055, -138224.916, -161335.403, -182157., -199198.195, -206223.016, -209647.376, -210013.098, -208773.499, -206109.605, -201822.65, -196112.649, -189179.617, -181223.567, -172444.513, -163300.287, -153182.238, -142712.319, -132512.483, -123204.684, -115410.874, -109753.008, -102402.023, -95723.6716},
        {-6.67630935, -6.54350469, -6.40435753, -6.25913874, -6.10811921, -5.95156981, -5.78976144, -5.62296495, -5.45145125, -5.2754912, -5.09535569, -4.9113156, -4.72364181, -4.5326052, -4.33847665, -4.14152704, -3.94202726, -3.74024818, -3.53646068, -3.33093564, -3.12394395, -2.91575648, -2.70664412, -2.49687774, -2.28672823, -2.07646646, -1.86636333, -1.6566897, -1.44771646, -1.23971449, -1.03295467, -0.827707882, -0.624245003, -0.422836916, -0.233031213, -0.0502652969, 0.130134932, 0.308199639, 0.483958992, 0.657443156, 0.8286823, 0.997706588, 1.16454619, 1.32923127, 1.49179199, 1.65225853, 1.81066104, 1.9670297, 2.12139467, 2.27378612, 2.42423421, 2.59059248, 2.75468135, 2.91614462, 3.07462607, 3.2297695, 3.38121869, 3.52861744, 3.67160953, 3.80983877, 3.94294893, 4.07058382, 4.19238722, 4.30800292, 4.41707471, 4.51924639, 4.61416174, 4.70473138, 4.79378748, 4.87423626, 4.94550974, 5.00703993, 5.05825885, 5.09859853, 5.12749096, 5.14436818, 5.14866219, 5.13980502, 5.11722867, 5.08036517, 5.02864653, 4.96150478, 4.87837191, 4.77867996, 4.23956035, 3.48047793, 2.72597324, 1.98918277, 1.283243, 0.621290419, 0.0164614999, -0.518107272, -0.969279413, -1.32391844, -1.56888788, -1.69105124, -1.67727204, -1.51441379, -1.18934003, -0.688914258, 0, 1.59036814, 3.3112303, 5.1746854, 7.19283239, 9.37777022, 11.7415978, 14.2964141, 17.0543181, 20.0274087, 23.2277849, 26.6675455, 30.3587895, 34.313616, 38.5441237, 43.0624117, 47.880579, 52.7847831, 57.7548459, 63.3139236, 69.4256993, 76.0538559, 83.1620764, 90.7140438, 98.6734411, 107.003951, 115.669257, 124.633042, 133.858988, 143.310779, 153.230818, 163.20187, 173.104002, 182.817282, 193.644441, 204.754214, 214.935958, 223.642939, 230.328427, 234.445687, 235.447987, 237.951289, 234.674911, 224.062028, 204.555816, 174.599449, 142.382018, 94.1643032, 25.9530011, -66.2451927, -186.423582, -331.854055, -521.466395, -754.790857, -1039.30791, -1398.72972, -1849.48074, -2407.98545, -3075.83764, -3887.66304, -4894.26628, -6113.12456, -7609.25444, -9400.31174, -11539.1173, -14138.0769, -17282.2461, -21014.8842, -25410.2105, -30695.6964, -37012.1253, -44315.577, -52959.9543, -63149.9763, -75090.3622, -88985.8312, -111671.472, -135064.042, -157710.668, -178158.476, -194954.593, -202084.378, -205697.904, -206297.342, -205103.505, -202334.1, -197907.401, -192056.086, -185012.835, -177010.325, -168281.237, -158322.948, -148494.834, -139005.83, -130064.87, -121880.888, -114662.818, -108619.594, -101727.156, -94440.3441},
        {-6.28084603, -6.11119927, -5.93798254, -5.76136501, -5.58151579, -5.39860404, -5.2127989, -5.02426949, -4.83318497, -4.63971446, -4.44402712, -4.24629207, -4.04667846, -3.84535542, -3.6424921, -3.43825764, -3.23282117, -3.02635183, -2.81901877, -2.61099111, -2.40243801, -2.1935286, -1.98443201, -1.77531739, -1.56635388, -1.35771061, -1.14955673, -0.942061378, -0.735393686, -0.529722795, -0.325217847, -0.122047979, 0.0796176691, 0.279609958, 0.472962984, 0.662006884, 0.849129448, 1.0343172, 1.21755667, 1.39883437, 1.57813684, 1.75545061, 1.93076218, 2.1040581, 2.27532488, 2.44454905, 2.61171713, 2.77681566, 2.93983115, 3.10075013, 3.25955913, 3.43419825, 3.60631123, 3.77549541, 3.94134812, 4.10346667, 4.26144839, 4.41489061, 4.56339066, 4.70654587, 4.84395355, 4.97521104, 5.09991567, 5.21766475, 5.32805563, 5.43068561, 5.52515204, 5.61439957, 5.70129251, 5.77852455, 5.84547601, 5.90152723, 5.94605854, 5.97845027, 5.99808275, 6.00433632, 5.99659131, 5.97422805, 5.93662688, 5.88316812, 5.8132321, 5.72619916, 5.62144964, 5.49836385, 4.91626277, 4.10381246, 3.29496818, 2.50339103, 1.74274212, 1.02668255, 0.368873431, -0.217024136, -0.717349044, -1.11844019, -1.40663647, -1.56827677, -1.58969999, -1.45724503, -1.15725078, -0.676056142, 0, 1.52212021, 3.20369418, 5.05456656, 7.08458196, 9.30358502, 11.7214204, 14.3479326, 17.1929664, 20.2663664, 23.5779771, 27.1376433, 30.9552095, 35.0405204, 39.4034206, 44.0537548, 49.0013675, 54.0417795, 59.1542179, 64.8333073, 71.0429658, 77.7471115, 84.9096625, 92.4945368, 100.465653, 108.786928, 117.422281, 126.33563, 135.490893, 144.851987, 154.687166, 164.544424, 174.296379, 183.815648, 194.331352, 205.037858, 214.767799, 222.986841, 229.160651, 232.754898, 233.235247, 235.509009, 231.944056, 220.941387, 200.901999, 170.226892, 136.685887, 86.9689526, 17.1348803, -76.7575376, -198.649509, -346.189472, -537.302568, -767.939696, -1044.8833, -1397.03023, -1843.34843, -2402.80582, -3090.29731, -3929.78955, -4949.69248, -6142.71264, -7598.05706, -9360.18595, -11488.8782, -14069.8694, -17176.9805, -20832.9733, -25098.4311, -30221.155, -36343.0453, -43407.7061, -51785.8406, -61687.99, -73324.6954, -86906.4979, -109255.817, -132318.345, -154651.655, -174813.318, -191360.906, -198323.755, -201806.495, -202307.37, -200989.299, -198087.425, -193559.941, -187655.838, -180624.103, -172713.728, -164173.701, -154013.61, -144661.724, -136046.856, -128097.82, -120743.43, -113912.5, -107533.841, -101008.201, -93115.3441},
        {-5.97837511, -5.79258398, -5.603748, -5.41201554, -5.21753499, -5.02045472, -4.82092313, -4.61908858, -4.41509946, -4.20910415, -4.00125103, -3.79168848, -3.58056488, -3.36802862, -3.15422806, -2.9393116, -2.72342761, -2.50672448, -2.28935058, -2.07145429, -1.853184, -1.63468809, -1.41611493, -1.19761291, -0.979330413, -0.76141581, -0.544017486, -0.327283822, -0.111363198, 0.103596005, 0.317445407, 0.530036627, 0.741221284, 0.950850999, 1.15713549, 1.36078187, 1.56251696, 1.76224568, 1.95987292, 2.15530358, 2.34844258, 2.53919481, 2.72746517, 2.91315856, 3.09617989, 3.27643406, 3.45382598, 3.62826053, 3.79964264, 3.96787719, 4.13286909, 4.30652823, 4.47649429, 4.64241191, 4.80392575, 4.96068048, 5.11232075, 5.25849122, 5.39883655, 5.53300138, 5.66063039, 5.78136823, 5.89485955, 6.00074901, 6.09868128, 6.188301, 6.26925285, 6.34560329, 6.42131266, 6.48690592, 6.54174107, 6.58517614, 6.61656911, 6.63527802, 6.64066085, 6.63207562, 6.60888035, 6.57043302, 6.51609167, 6.44521429, 6.35715889, 6.25128348, 6.12694607, 5.98350466, 5.36745659, 4.51411551, 3.66423813, 2.83187865, 2.0310913, 1.27593026, 0.580449766, -0.0412959907, -0.575252797, -1.00736644, -1.32358272, -1.50984743, -1.55210634, -1.43630527, -1.14838999, -0.674306306, 0, 1.34633107, 2.89979956, 4.66684194, 6.65389467, 8.86739421, 11.313777, 13.9994796, 16.9309385, 20.11459, 23.5568706, 27.2642169, 31.2430653, 35.4998523, 40.0410142, 44.8729877, 50.0022092, 55.2771285, 60.6816126, 66.5794486, 72.9432187, 79.745505, 86.9588899, 94.5559554, 102.509284, 110.791457, 119.375058, 128.232668, 137.33687, 146.660246, 156.587317, 166.527683, 176.330342, 185.844296, 196.196193, 206.59621, 215.913817, 223.614718, 229.16462, 232.029228, 231.674248, 232.959995, 228.315727, 216.151685, 194.878106, 162.90523, 128.00088, 76.8783131, 5.60837317, -89.7380975, -213.090256, -362.103953, -554.649028, -787.293302, -1066.86552, -1421.14382, -1868.50049, -2427.30782, -3111.47724, -3944.85572, -4956.15659, -6140.57786, -7581.44534, -9314.34188, -11396.3438, -13920.9466, -16964.9289, -20564.0916, -24784.5894, -29842.8284, -35870.224, -42775.8049, -50947.5867, -60593.9649, -71923.3351, -85144.0928, -107033.837, -129589.459, -151377.054, -170962.714, -186912.537, -193179.749, -196018.556, -195971.994, -194468.392, -191700.36, -187491.633, -182053.263, -175596.302, -168331.802, -160470.814, -151644.121, -143058.065, -134798.813, -126952.529, -119605.381, -112843.536, -106753.158, -99452.4701, -91551.8112},
        {-5.73832998, -5.54140087, -5.34136164, -5.13836115, -4.93254827, -4.72407187, -4.51308081, -4.29972396, -4.08415019, -3.86650835, -3.64694734, -3.42561599, -3.2026632, -2.97823781, -2.7524887, -2.52556474, -2.29761479, -2.06878772, -1.8392324, -1.60909768, -1.37853245, -1.14768557, -0.916705897, -0.685742307, -0.454943664, -0.224458836, 0.00556331013, 0.234973908, 0.46362409, 0.691364991, 0.918047741, 1.14352348, 1.36764333, 1.59025843, 1.81225339, 2.03294086, 2.25162108, 2.46811165, 2.68223017, 2.89379422, 3.1026214, 3.30852931, 3.51133554, 3.71085768, 3.90691334, 4.09932011, 4.28789557, 4.47245733, 4.65282298, 4.82881012, 5.00023634, 5.17183015, 5.33839178, 5.49963235, 5.65526301, 5.80499488, 5.94853911, 6.08560683, 6.21590917, 6.33915728, 6.45506228, 6.56333532, 6.66368753, 6.75583004, 6.83947399, 6.91433052, 6.98011076, 7.04236937, 7.10652069, 7.16022406, 7.20281179, 7.2336162, 7.25196962, 7.25720438, 7.2486528, 7.2256472, 7.1875199, 7.13360324, 7.06322953, 6.97573111, 6.87044028, 6.74668939, 6.60381075, 6.44113668, 5.79316423, 4.90142055, 4.01301975, 3.14237895, 2.30391528, 1.51204586, 0.781187817, 0.125758268, -0.439825661, -0.901146847, -1.24378817, -1.4533325, -1.51536272, -1.4154617, -1.13921232, -0.672197463, 0, 1.15956585, 2.57480035, 4.24869984, 6.18426061, 8.38447901, 10.8523513, 13.5908739, 16.6030431, 19.8918552, 23.4603064, 27.3113932, 31.4481119, 35.8734587, 40.5904301, 45.6020222, 50.9112315, 56.42014, 62.1173277, 68.2340494, 74.7516768, 81.6515818, 88.9151363, 96.5237121, 104.458681, 112.701415, 121.233286, 130.035665, 139.089925, 148.377437, 158.401289, 168.429842, 178.287951, 187.800475, 198.00935, 208.130894, 217.056868, 224.247006, 229.161041, 231.258706, 229.999734, 230.064609, 224.10339, 210.554356, 187.855791, 154.445976, 118.297563, 65.9308696, -6.59941378, -103.238599, -227.931996, -378.388858, -572.412404, -808.082879, -1092.38936, -1450.5098, -1899.45571, -2456.23861, -3130.00267, -3947.41479, -4943.72443, -6124.15377, -7557.95093, -9263.20793, -11291.5995, -13751.7047, -16725.9825, -20268.8762, -24455.5814, -29464.0181, -35413.9995, -42182.6663, -50169.8385, -59579.3313, -70614.9598, -83480.5392, -104908.457, -126941.813, -148152.279, -167111.529, -182391.233, -187825.339, -189881.729, -189162.527, -187453.029, -184887.024, -181105.311, -176266.329, -170528.514, -164050.305, -156990.138, -149831.311, -142061.593, -134022.799, -126056.745, -118505.246, -111710.118, -106013.175, -97716.8969, -89867.6319},
        {-5.55046846, -5.3431988, -5.13258933, -4.91879636, -4.70197621, -4.48228519, -4.25987962, -4.0349158, -3.80755006, -3.57793869, -3.34623803, -3.11260438, -2.87719406, -2.64016338, -2.40166865, -2.16186619, -1.92091231, -1.67896333, -1.43617555, -1.1927053, -0.94870888, -0.704342613, -0.459762809, -0.215125782, 0.0294121531, 0.273694684, 0.517565496, 0.760868276, 1.00344671, 1.24514449, 1.48580529, 1.7252728, 1.96339071, 2.20000271, 2.43817398, 2.67606969, 2.91181631, 3.14515297, 3.37581882, 3.60355301, 3.82809467, 4.04918295, 4.26655699, 4.47995593, 4.68911892, 4.8937851, 5.09369361, 5.28858359, 5.47819418, 5.66226454, 5.8405338, 6.01116031, 6.17549828, 6.33332112, 6.48440223, 6.62851503, 6.76543293, 6.89492934, 7.01677767, 7.13075133, 7.23662373, 7.33416829, 7.42315841, 7.5033675, 7.57456899, 7.63653626, 7.68904275, 7.73928304, 7.79427354, 7.83848204, 7.87120086, 7.89172232, 7.89933876, 7.89334249, 7.87302584, 7.83768114, 7.78660072, 7.7190769, 7.634402, 7.53186836, 7.41076829, 7.27039413, 7.1100382, 6.92899282, 6.24954711, 5.31952811, 4.39249611, 3.48322309, 2.60648104, 1.77704198, 1.00967789, 0.319160768, -0.279737382, -0.772244566, -1.14358878, -1.37899804, -1.46370034, -1.38292368, -1.12189607, -0.66584551, 0, 0.998100267, 2.29410961, 3.8880078, 5.77977462, 7.96938986, 10.4568333, 13.2420847, 16.3251238, 19.7059305, 23.3844845, 27.3607656, 31.6347535, 36.2064281, 41.0757691, 46.2427563, 51.7073695, 57.4099064, 63.3418208, 69.6380876, 76.2858975, 83.2724414, 90.5849099, 98.2104939, 106.136384, 114.349771, 122.837846, 131.5878, 140.586823, 149.822105, 159.883399, 169.934396, 179.781519, 189.231189, 199.290705, 209.166052, 217.742982, 224.467654, 228.78623, 230.144868, 227.98973, 226.695863, 219.280444, 204.225284, 180.012197, 145.122995, 108.002745, 54.6791928, -18.856664, -116.613826, -242.601294, -394.836531, -590.643067, -829.710876, -1119.11961, -1481.03617, -1930.85359, -2483.96489, -3143.03553, -3939.07876, -4916.99251, -6095.6033, -7526.58419, -9208.72275, -11187.0985, -13585.2303, -16490.6318, -19973.8569, -24118.0224, -29068.4322, -34935.2138, -41579.974, -49399.5949, -58590.2301, -69348.0334, -81869.1588, -102803.582, -124280.179, -144881.649, -163190.689, -177789.998, -182421.2, -183735.451, -182390.194, -180518.744, -178194.347, -174875.304, -170665.608, -165669.252, -159990.231, -153732.538, -148107.936, -141084.092, -133239.672, -125153.339, -117403.761, -110569.601, -105229.525, -95958.323, -88072.1111},
        {-5.38083911, -5.15189639, -4.92044114, -4.68660367, -4.45051427, -4.21230324, -3.97210088, -3.73003749, -3.48624336, -3.24084879, -2.99398409, -2.74577954, -2.49636544, -2.24587209, -1.9944298, -1.74216885, -1.48921955, -1.23571219, -0.981777068, -0.727544485, -0.473144738, -0.218708124, 0.0356350599, 0.289754515, 0.543519945, 0.796801051, 1.04946754, 1.3013891, 1.55243545, 1.80247629, 2.05138131, 2.29902023, 2.54526274, 2.78997854, 3.03762472, 3.28568081, 3.53157118, 3.77501667, 4.01573811, 4.25345634, 4.48789219, 4.7187665, 4.94580008, 5.16871379, 5.38722845, 5.60106489, 5.80994395, 6.01358647, 6.21171326, 6.40404518, 6.59030304, 6.76864279, 6.94038407, 7.10528166, 7.2630903, 7.41356475, 7.55645976, 7.69153011, 7.81853053, 7.9372158, 8.04734066, 8.14865988, 8.2409282, 8.32390039, 8.39733121, 8.46097541, 8.51458775, 8.56664421, 8.62541138, 8.67265715, 8.70757091, 8.72934205, 8.73715997, 8.73021407, 8.70769374, 8.66878838, 8.61268738, 8.53858013, 8.44565604, 8.3331045, 8.2001149, 8.04587664, 7.86957911, 7.67041172, 6.95415109, 5.97707128, 5.00137678, 4.04226916, 3.11495003, 2.23462097, 1.41648356, 0.675739396, 0.0275900665, -0.512762842, -0.930117741, -1.20927304, -1.33502716, -1.2921785, -1.06552548, -0.639866509, 0, 0.997779603, 2.30196038, 3.91174944, 5.82635389, 8.04498085, 10.5668374, 13.3911307, 16.5170679, 19.943856, 23.6707021, 27.6968135, 32.0213971, 36.6436601, 41.5628097, 46.7780529, 52.2885967, 57.9853292, 63.8519832, 70.1327736, 76.8036964, 83.8407474, 91.2199224, 98.9172172, 106.908628, 115.17015, 123.677779, 132.407512, 141.335343, 150.43727, 160.203073, 169.906575, 179.369635, 188.414113, 198.244674, 207.991776, 216.417128, 222.92775, 226.930659, 227.832875, 225.041417, 222.327891, 213.406375, 196.829946, 171.151681, 134.924659, 97.9538818, 44.7275204, -29.0143298, -127.531573, -255.084112, -411.313178, -610.742396, -852.191177, -1142.05192, -1502.17853, -1947.48302, -2492.87739, -3137.67741, -3915.93592, -4878.71989, -6053.2707, -7481.55312, -9162.09618, -11135.7028, -13514.8063, -16378.9607, -19781.0667, -23793.768, -28587.4365, -34277.1635, -40792.3482, -48462.342, -57466.5142, -67984.234, -80194.8708, -100474.716, -121256.987, -141171.82, -158849.357, -172919.734, -177295.003, -178468.783, -177093.248, -175338.499, -173255.867, -170264.214, -166446.496, -161885.666, -156664.68, -150866.492, -144852.968, -137995.918, -130671.073, -123254.163, -116120.922, -109647.078, -104208.364, -94845.2674, -86345.7148},
        {-5.2529074, -4.99801147, -4.74178235, -4.48431374, -4.22569934, -3.96603284, -3.70540794, -3.44391834, -3.18165773, -2.9187198, -2.65519826, -2.3911868, -2.12677911, -1.8620689, -1.59714985, -1.33211566, -1.06706003, -0.802076662, -0.537259241, -0.272701465, -0.00849703253, 0.255260362, 0.518477022, 0.781059251, 1.04291335, 1.30394564, 1.5640624, 1.82316995, 2.08117459, 2.33798262, 2.59350035, 2.84763408, 3.10029012, 3.35137477, 3.60638242, 3.86230778, 4.11607843, 4.36741932, 4.61605541, 4.86171167, 5.10411305, 5.34298451, 5.57805102, 5.80903753, 6.035669, 6.2576704, 6.47476668, 6.68668281, 6.89314374, 7.09387443, 7.28859985, 7.47784077, 7.66050908, 7.83631249, 8.00495872, 8.16615546, 8.31961043, 8.46503133, 8.60212588, 8.73060178, 8.85016674, 8.96052848, 9.06139469, 9.15247308, 9.23347137, 9.30409727, 9.36405848, 9.42295971, 9.49016805, 9.54497938, 9.58645983, 9.61367553, 9.62569261, 9.62157718, 9.60039538, 9.56121332, 9.50309715, 9.42511298, 9.32632693, 9.20580515, 9.06261374, 8.89581884, 8.70448658, 8.48768308, 7.73369579, 6.70772371, 5.68110502, 4.6694816, 3.68849532, 2.75378806, 1.88100169, 1.08577809, 0.383759143, -0.209413284, -0.678097311, -1.00665106, -1.17943266, -1.18080023, -0.995111893, -0.606725776, 0, 1.05332068, 2.4095357, 4.06783881, 6.02742377, 8.28748432, 10.8472142, 13.7058072, 16.862457, 20.3163574, 24.0667022, 28.112685, 32.4534997, 37.08834, 42.0163996, 47.2368723, 52.7489518, 58.3701138, 64.0735922, 70.26961, 76.9184212, 83.9802801, 91.4154408, 99.1841574, 107.246684, 115.563275, 124.094184, 132.799666, 141.639974, 150.575364, 159.946174, 169.193209, 178.162696, 186.700864, 196.261348, 205.886673, 214.190722, 220.5375, 224.291014, 224.81527, 221.474274, 217.40079, 207.043053, 189.027703, 161.98138, 124.530724, 87.9939659, 35.1332583, -38.597656, -137.745034, -266.855133, -427.506129, -630.840016, -874.056947, -1162.64581, -1519.01832, -1957.98818, -2494.36912, -3125.39304, -3887.45132, -4836.11546, -6004.74798, -7428.69556, -9112.08597, -11092.4152, -13463.7201, -16294.0843, -19609.6925, -23467.7655, -28079.154, -33567.8066, -39952.1199, -47477.8341, -56305.4355, -66595.4104, -78508.2451, -98085.1861, -118135.77, -137350.292, -154419.049, -168032.337, -172370.003, -173639.76, -172464.952, -170927.608, -169073.679, -166341.332, -162806.908, -158546.746, -153637.187, -148154.571, -141239.518, -134400.328, -127684.963, -121141.383, -114817.551, -108761.428, -103020.974, -93807.1156, -84454.4065},
        {-5.17322923, -4.88950303, -4.6058374, -4.32228311, -4.03889089, -3.75571149, -3.47279564, -3.19019411, -2.90795762, -2.62613692, -2.34478275, -2.06394586, -1.783677, -1.5040269, -1.22504631, -0.946785969, -0.669296627, -0.392629025, -0.116833906, 0.158037986, 0.431935907, 0.704809114, 0.976606864, 1.24727841, 1.51677302, 1.78503993, 2.05202841, 2.31768772, 2.58196711, 2.84481583, 3.10618315, 3.36601832, 3.6242706, 3.88088923, 4.14227086, 4.40500544, 4.66560544, 4.92381087, 5.17936178, 5.43199818, 5.68146011, 5.92748759, 6.16982064, 6.4081993, 6.64236359, 6.87205353, 7.09700916, 7.31697051, 7.53167759, 7.74087043, 7.94428907, 8.14597719, 8.34127786, 8.52983781, 8.71130377, 8.88532249, 9.05154067, 9.20960507, 9.35916241, 9.49985943, 9.63134284, 9.7532594, 9.86525582, 9.96697885, 10.0580752, 10.1381916, 10.2069748, 10.2750006, 10.3525822, 10.4168262, 10.4666707, 10.5010541, 10.5189146, 10.5191905, 10.5008199, 10.4627412, 10.4038925, 10.3232122, 10.2196385, 10.0921097, 9.93956389, 9.76093944, 9.55517458, 9.32120756, 8.53000112, 7.45558652, 6.37834087, 5.31433623, 4.27964469, 3.29033829, 2.36248912, 1.51216925, 0.755450736, 0.108405656, -0.412893923, -0.792375932, -1.0139683, -1.06159896, -0.919195844, -0.57068688, 0, 1.13045387, 2.55555404, 4.27472893, 6.28740696, 8.59301653, 11.1909861, 14.080744, 17.2617187, 20.7333386, 24.4950322, 28.5462278, 32.8863539, 37.5148388, 42.4311111, 47.634599, 53.1247311, 58.6451883, 64.1588638, 70.2436621, 76.8442087, 83.905129, 91.3710483, 99.1865922, 107.296386, 115.645055, 124.177225, 132.837521, 141.570568, 150.320992, 159.281417, 168.057537, 176.519578, 184.537766, 193.792619, 203.249218, 211.389896, 217.541791, 221.032042, 221.187785, 217.336161, 212.070178, 200.457142, 181.185213, 152.942555, 114.417331, 78.2588651, 25.7038673, -48.0497898, -147.804234, -278.361593, -443.155395, -650.0291, -894.277431, -1180.45716, -1532.03507, -1963.98775, -2491.29176, -3110.4721, -3859.15048, -4795.07761, -5954.90361, -7370.42433, -9054.0621, -11041.0589, -13405.3236, -16203.0162, -19432.7703, -23135.8319, -27564.3768, -32851.2671, -39100.9977, -46479.1617, -55128.4403, -65191.5149, -76811.0668, -95678.2124, -115000.089, -133532.269, -150030.325, -163249.83, -167688.558, -169198.411, -168359.732, -167084.079, -165424.855, -162880.792, -159533.026, -155462.696, -150750.938, -145478.889, -137617.808, -130801.221, -124722.695, -119075.799, -113554.098, -107851.161, -101660.554, -92638.1841, -82286.2229},
        {-5.15647109, -4.84147892, -4.52812125, -4.21639957, -3.90631536, -3.59787012, -3.29106533, -2.98590247, -2.68238305, -2.38050853, -2.08028042, -1.78170019, -1.48476934, -1.18948936, -0.895861723, -0.603887929, -0.313569462, -0.024907809, 0.262095542, 0.547439103, 0.831121387, 1.11314091, 1.39349617, 1.6721857, 1.949208, 2.22456158, 2.49824496, 2.77025665, 3.04059516, 3.309259, 3.57624669, 3.84155674, 4.10518766, 4.36713796, 4.63521747, 4.90535472, 5.17338442, 5.43905157, 5.70210119, 5.96227831, 6.21932795, 6.47299513, 6.72302486, 6.96916217, 7.21115208, 7.44873961, 7.68166978, 7.90968761, 8.13253813, 8.34996635, 8.56171728, 8.77437572, 8.98069866, 9.18028283, 9.37272499, 9.55762188, 9.73457026, 9.90316687, 10.0630085, 10.2136918, 10.3548136, 10.4859706, 10.6067596, 10.7167773, 10.8156205, 10.9028859, 10.9781702, 11.052832, 11.137947, 11.2088547, 11.2643892, 11.303385, 11.3246761, 11.3270971, 11.3094821, 11.2706654, 11.2094813, 11.1247641, 11.0153482, 10.8800677, 10.717757, 10.5272503, 10.307382, 10.0569864, 9.23200028, 8.11412586, 6.99159039, 5.88084622, 4.79834569, 3.76054116, 2.78388499, 1.88482952, 1.0798271, 0.385330097, -0.182209152, -0.606338289, -0.870604964, -0.958556824, -0.853741518, -0.539706694, 0, 1.16246522, 2.62090698, 4.37418843, 6.42117275, 8.7607231, 11.3917026, 14.3129746, 17.523402, 21.0218481, 24.8071761, 28.8782492, 33.2339304, 37.873083, 42.7945701, 47.997255, 53.4800006, 58.9662796, 64.4156134, 70.449773, 77.0086092, 84.0319729, 91.4597149, 99.2316861, 107.287737, 115.567719, 124.011483, 132.55888, 141.14976, 149.723974, 158.445243, 166.947462, 175.103322, 182.785513, 191.702295, 200.808572, 208.569765, 214.307895, 217.344979, 217.003039, 212.604092, 206.664706, 194.340099, 174.327271, 145.323223, 106.024952, 69.431748, 16.3627491, -58.060617, -158.716923, -290.48474, -457.755755, -666.490512, -910.366031, -1193.3048, -1540.1471, -1966.34157, -2487.33684, -3102.0091, -3846.14586, -4773.6136, -5917.73045, -7312.47295, -8976.82202, -10945.8889, -13282.1169, -16035.4033, -19192.934, -22788.3478, -27085.226, -32217.0814, -38315.5629, -45516.3339, -53953.5524, -63761.3763, -75073.9637, -93312.7976, -112002.88, -129956.537, -145986.095, -158903.882, -163506.725, -165299.449, -164815.307, -163739.52, -162149.354, -159642.094, -156319.006, -152281.36, -147630.423, -142467.462, -134542.176, -128105.003, -122635.395, -117612.804, -112516.68, -106826.475, -100021.641, -90865.7734, -79520.2494},
        {-5.20643077, -4.85853562, -4.51395152, -4.17262674, -3.83450952, -3.49954812, -3.16769081, -2.83888583, -2.51308144, -2.1902259, -1.87026747, -1.55315439, -1.23883493, -0.927257339, -0.618369878, -0.312120801, -0.00845836309, 0.292669178, 0.591313568, 0.887526549, 1.18135987, 1.47286526, 1.76209448, 2.04909927, 2.33393137, 2.61664253, 2.89728448, 3.17590898, 3.45256777, 3.72731259, 4.00019518, 4.27126729, 4.54058067, 4.80818705, 5.083977, 5.3628759, 5.63969089, 5.91415443, 6.18599897, 6.45495697, 6.72076089, 6.98314317, 7.24183628, 7.49657267, 7.7470848, 7.99310512, 8.23436609, 8.47060016, 8.70153979, 8.92691744, 9.14646555, 9.36756336, 9.58213079, 9.78973452, 9.98994124, 10.1823176, 10.3664304, 10.5418462, 10.7081318, 10.8648538, 11.0115789, 11.1478739, 11.2733053, 11.38744, 11.4898445, 11.5800855, 11.6577299, 11.7347016, 11.8226283, 11.8955901, 11.9523528, 11.9916819, 12.012343, 12.0131017, 11.9927237, 11.9499744, 11.8836196, 11.7924248, 11.6751557, 11.5305778, 11.3574567, 11.1545581, 10.9206475, 10.6544906, 9.80008664, 8.64525322, 7.48447564, 6.33452273, 5.21216333, 4.13416628, 3.11730041, 2.17833455, 1.33403753, 0.601178198, -0.00347462036, -0.463152089, -0.761085375, -0.880505643, -0.804644059, -0.51673179, 0, 1.1265284, 2.56482626, 4.31204507, 6.36533634, 8.72185158, 11.3787423, 14.33316, 17.5822562, 21.1231824, 24.9530901, 29.0691308, 33.4684561, 38.1482173, 43.1055661, 48.337654, 53.8416323, 59.3904822, 64.9493308, 71.0223819, 77.5587506, 84.5075518, 91.8179005, 99.4389116, 107.3197, 115.409381, 123.657069, 132.01188, 140.422927, 148.839327, 157.549754, 166.039258, 174.155086, 181.744484, 190.292528, 198.82755, 205.94113, 210.989165, 213.327555, 212.312198, 207.298992, 201.290429, 188.885982, 168.728085, 139.459172, 99.7216773, 61.6368613, 6.99862682, -68.9202977, -170.847184, -303.509304, -471.074807, -679.513459, -921.470776, -1200.66294, -1543.40876, -1965.87893, -2484.24412, -3103.00565, -3852.65573, -4776.41674, -5896.97765, -7256.55068, -8877.10794, -10795.925, -13076.2083, -15769.2616, -18872.3976, -22422.9225, -26656.1094, -31695.2501, -37623.4849, -44610.4344, -52793.2889, -62309.2389, -73295.4747, -91016.184, -109199.811, -126701.796, -142377.58, -155082.606, -159862.903, -161922.93, -161750.732, -160775.645, -159113.12, -156486.406, -153030.564, -148880.656, -144171.745, -139038.893, -132245.083, -126647.036, -121716.273, -116924.314, -111742.678, -105642.885, -98096.4552, -88351.666, -76077.421},
        {-5.30597629, -4.92502784, -4.54900892, -4.1778173, -3.81135075, -3.44950705, -3.09218395, -2.73927923, -2.39069066, -2.04631601, -1.70605304, -1.36979953, -1.03745325, -0.708911958, -0.38407343, -0.0628354354, 0.254904258, 0.56924788, 0.880297662, 1.18815583, 1.49292463, 1.79470628, 2.093603, 2.38971705, 2.68315063, 2.974006, 3.26238537, 3.54839097, 3.83212505, 4.11368982, 4.39318752, 4.67072038, 4.94639063, 5.22030051, 5.50430306, 5.79237794, 6.07836357, 6.36198086, 6.64295068, 6.92099393, 7.19583152, 7.46718433, 7.73477325, 7.99831919, 8.25754302, 8.51216566, 8.76190799, 9.0064909, 9.2456353, 9.47906206, 9.7064921, 9.93623567, 10.1592381, 10.3750341, 10.5831583, 10.7831454, 10.9745302, 11.1568473, 11.3296314, 11.4924172, 11.6447394, 11.7861327, 11.9161318, 12.0342714, 12.1400862, 12.2331108, 12.31288, 12.3916425, 12.481342, 12.5552911, 12.6122004, 12.6507804, 12.6697416, 12.6677944, 12.6436494, 12.596017, 12.5236078, 12.4251322, 12.2993007, 12.1448238, 11.960412, 11.7447758, 11.4966256, 11.214672, 10.3322643, 9.14268503, 7.94601226, 6.75930351, 5.59961632, 4.48400824, 3.42953683, 2.45325963, 1.57223419, 0.803518056, 0.164168787, -0.328756072, -0.658198973, -0.807102366, -0.758408701, -0.495060428, 0, 1.0843313, 2.49740301, 4.23451286, 6.29095855, 8.66203781, 11.3430484, 14.3292879, 17.6160542, 21.198645, 25.0723579, 29.2324907, 33.6743411, 38.3932068, 43.3843856, 48.6431751, 54.1648731, 59.7849551, 65.475887, 71.5944693, 78.101752, 84.9587852, 92.1266191, 99.5663037, 107.238889, 115.105425, 123.126963, 131.264551, 139.479241, 147.732082, 156.467624, 164.986134, 173.103613, 180.636061, 188.783903, 196.670926, 203.051739, 207.327013, 208.897421, 207.163635, 201.526328, 195.624754, 183.331, 163.216453, 133.852496, 93.810514, 53.9221371, -2.56655576, -80.1492408, -183.319594, -316.571293, -483.405509, -690.521482, -929.887478, -1205.19361, -1544.0621, -1963.37014, -2479.99492, -3104.73034, -3861.91945, -4782.54162, -5876.66813, -7196.61725, -8766.9612, -10627.8764, -12845.9796, -15475.7512, -18529.5956, -22051.0845, -26238.0675, -31201.1255, -36958.2438, -43726.8453, -51649.5637, -60869.0328, -71527.8864, -88754.4521, -106459.246, -123538.479, -138888.361, -151405.103, -156376.34, -158712.573, -158856.544, -157978.089, -156232.142, -153470.682, -149863.792, -145581.553, -140794.049, -135671.361, -130339.145, -125663.017, -121184.165, -116443.781, -110983.051, -104343.168, -96065.3182, -85601.7555, -72344.402},
        {-5.44995917, -5.03689986, -4.63021741, -4.22976605, -3.83540002, -3.44697355, -3.06434086, -2.6873562, -2.3158738, -1.94974788, -1.58883268, -1.23298244, -0.882051381, -0.535893742, -0.194363754, 0.14268435, 0.475396339, 0.803917978, 1.12839504, 1.44897328, 1.76579848, 2.0790164, 2.3887728, 2.69521347, 2.99848415, 3.29873063, 3.59609866, 3.89073402, 4.18278247, 4.47238978, 4.75970172, 5.04486406, 5.32802255, 5.60932298, 5.90221643, 6.19991616, 6.49547548, 6.78860837, 7.07902883, 7.36645082, 7.65058835, 7.93115539, 8.20786592, 8.48043394, 8.74857341, 9.01199833, 9.27042269, 9.52356046, 9.77112562, 10.0128322, 10.2483941, 10.4875563, 10.7197844, 10.944575, 11.1614245, 11.3698294, 11.5692864, 11.7592919, 11.9393425, 12.1089347, 12.267565, 12.41473, 12.5499262, 12.6726501, 12.7823983, 12.8786673, 12.9609536, 13.0415628, 13.1324929, 13.2068227, 13.2632185, 13.3003464, 13.3168727, 13.3114634, 13.2827848, 13.2295031, 13.1502844, 13.0437948, 12.9087007, 12.7436681, 12.5473632, 12.3184522, 12.0556013, 11.7574767, 10.8477471, 9.62467177, 8.39342083, 7.17132024, 5.97569594, 4.82387388, 3.73318002, 2.72094029, 1.80448064, 1.00112702, 0.328205374, -0.196958349, -0.557038203, -0.734708241, -0.712642518, -0.473515087, 0, 1.04917253, 2.44235267, 4.17307311, 6.23486654, 8.62126565, 11.3258031, 14.3420117, 17.6634239, 21.2835726, 25.1959904, 29.3942101, 33.8717642, 38.6221855, 43.6390066, 48.9157604, 54.4459793, 60.1256063, 65.9342324, 72.0861297, 78.5539187, 85.3102199, 92.3276539, 99.5788412, 107.036402, 114.672958, 122.461129, 130.373535, 138.382797, 146.461536, 155.208613, 163.740786, 171.842803, 179.299414, 187.016264, 194.217654, 199.828975, 203.298708, 204.075331, 201.607323, 195.343163, 189.590513, 177.459787, 157.448755, 128.055186, 87.776849, 45.9517598, -12.4726198, -91.7085821, -195.968419, -329.464423, -494.740017, -699.716092, -936.109664, -1207.73362, -1543.09659, -1959.6193, -2474.72251, -3105.16959, -3869.18804, -4785.72248, -5852.1192, -7131.06942, -8649.66349, -10451.1474, -12606.1417, -15172.7017, -18179.3217, -21676.3929, -25822.6967, -30715.7029, -36307.2484, -42862.8409, -50530.3615, -59457.6914, -69792.7117, -86528.7013, -103756.794, -120413.522, -135435.417, -147759.01, -152906.271, -155504.668, -155954.119, -155187.496, -153393.753, -150539.776, -146825.622, -142451.347, -137617.007, -132522.658, -128791.877, -125018.263, -120873.049, -116027.468, -110152.754, -102920.137, -94000.8528, -82753.145, -68441.1999},
        {-5.62176446, -5.18305314, -4.75145394, -4.32680483, -3.90894374, -3.49770865, -3.09293749, -2.69446822, -2.30213879, -1.91578716, -1.53525128, -1.16036909, -0.790978557, -0.426917626, -0.068024252, 0.285863614, 0.63490802, 0.979271013, 1.31911464, 1.65460095, 1.98589199, 2.31314981, 2.63653646, 2.95621398, 3.27234442, 3.58508983, 3.89461225, 4.20107375, 4.50463635, 4.80546211, 5.10371309, 5.39955131, 5.69313884, 5.98463773, 6.28755594, 6.59510186, 6.90032344, 7.20294963, 7.50270936, 7.79933159, 8.09254527, 8.38207934, 8.66766275, 8.94902444, 9.22589335, 9.49799845, 9.76506867, 10.026833, 10.2830203, 10.5333595, 10.7775797, 11.0289951, 11.2734549, 11.5103934, 11.739245, 11.9594443, 12.1704257, 12.3716235, 12.5624723, 12.7424064, 12.9108604, 13.0672687, 13.2110657, 13.3416858, 13.4585635, 13.5611332, 13.6488294, 13.7335571, 13.8269218, 13.9026384, 13.959333, 13.9956315, 14.01016, 14.0015445, 13.968411, 13.9093857, 13.8230944, 13.7081632, 13.5632182, 13.3868853, 13.1777907, 12.9345603, 12.6558201, 12.3401962, 11.4015855, 10.1432775, 8.87558992, 7.61612425, 6.38248213, 5.19226518, 4.06307499, 3.01251319, 2.05818139, 1.21768118, 0.508614198, -0.0514179624, -0.444813686, -0.653971362, -0.66128938, -0.449166129, 0, 1.07552131, 2.49772877, 4.25929143, 6.35287837, 8.77115863, 11.5068013, 14.5524754, 17.9008499, 21.5445941, 25.4763768, 29.6888673, 34.1747344, 38.9266474, 43.9372752, 49.1992869, 54.7053516, 60.3603041, 66.1442657, 72.2485781, 78.6470887, 85.3136451, 92.2220946, 99.3462846, 106.660063, 114.137276, 121.751773, 129.477399, 137.288004, 145.157434, 153.734965, 162.07136, 169.937835, 177.105611, 184.327166, 190.88309, 195.792299, 198.531636, 198.577939, 195.40805, 188.49881, 182.50777, 170.154321, 149.901684, 120.213083, 79.5517392, 36.1078449, -23.8140883, -104.18258, -208.966149, -342.133316, -505.326008, -707.682948, -941.473701, -1210.74562, -1543.5924, -1957.31141, -2469.20002, -3098.61281, -3860.32241, -4766.85724, -5807.97623, -7052.93443, -8533.48175, -10294.6891, -12404.1992, -14919.2938, -17872.0811, -21313.3165, -25386.059, -30181.1556, -35637.6015, -42021.4569, -49476.1329, -58145.0403, -68171.5901, -84329.2183, -100973.804, -117091.253, -131667.468, -143688.356, -148920.554, -151715.41, -152438.513, -151887.344, -150256.536, -147560.766, -144003.666, -139788.867, -135119.999, -130200.693, -127708.335, -124495.984, -120407.209, -115285.581, -108974.668, -101318.039, -92159.2648, -80223.245, -64583.8322},
        {-5.81886307, -5.36104081, -4.91033855, -4.46660574, -4.02969185, -3.59944632, -3.17571861, -2.75835819, -2.3472145, -1.942137, -1.54297515, -1.14957839, -0.7617962, -0.379478019, -0.00247330618, 0.369368482, 0.736197889, 1.09816546, 1.45542174, 1.80811727, 2.1564026, 2.50042827, 2.84034483, 3.17630281, 3.50845277, 3.83694525, 4.16193079, 4.48355994, 4.80198324, 5.11735123, 5.42981446, 5.73952348, 6.04662883, 6.35128105, 6.6654702, 6.98317804, 7.29824403, 7.61043449, 7.91951576, 8.22525417, 8.52741604, 8.8257677, 9.12007548, 9.41010572, 9.69562474, 9.97639887, 10.2521944, 10.5227778, 10.7879152, 11.0473731, 11.3009177, 11.5676459, 11.8275745, 12.0800507, 12.3244218, 12.5600352, 12.7862381, 13.0023777, 13.2078014, 13.4018565, 13.5838901, 13.7532497, 13.9092825, 14.0513357, 14.1787567, 14.2908927, 14.387091, 14.4784093, 14.575624, 14.6539306, 14.7119174, 14.7481725, 14.761284, 14.74984, 14.7124288, 14.6476385, 14.5540572, 14.4302731, 14.2748743, 14.086449, 13.8635853, 13.6048713, 13.3088953, 12.9742453, 12.004807, 10.7090285, 9.40250596, 8.1031267, 6.82877814, 5.59734767, 4.42672267, 3.33479055, 2.33943868, 1.45855447, 0.710025291, 0.111738549, -0.318418371, -0.562558076, -0.602793176, -0.421236281, 0, 1.16727135, 2.67026512, 4.50178613, 6.65463921, 9.12162916, 11.8955608, 14.969239, 18.3354685, 21.9870541, 25.9168008, 30.1175132, 34.5819962, 39.3030547, 44.2734933, 49.4861171, 54.9337307, 60.4713618, 66.0721328, 72.0404874, 78.3397066, 84.9330718, 91.7838639, 98.8553641, 106.110854, 113.513613, 121.026925, 128.614069, 136.238326, 143.862979, 152.067296, 159.972708, 167.357699, 174.000754, 180.671196, 186.652091, 190.962279, 193.082994, 192.495469, 188.680937, 181.12063, 174.437054, 161.405434, 140.501103, 110.199394, 68.9756383, 24.3474493, -36.5126925, -117.390025, -222.069787, -354.337216, -515.097268, -714.53581, -946.280578, -1214.69503, -1546.06141, -1956.82086, -2463.41453, -3083.89097, -3832.71861, -4722.61242, -5741.9944, -6961.83568, -8420.55886, -10163.4406, -12247.3715, -14723.911, -17614.5099, -20962.9219, -24923.2994, -29587.347, -34941.3505, -41198.6517, -48487.4464, -56935.9303, -66672.2992, -82159.5723, -98107.416, -113560.32, -127562.775, -139159.271, -144367.831, -147277.055, -148228.491, -147999.87, -146760.082, -144496.275, -141386.428, -137608.52, -133340.529, -128760.433, -127024.612, -123980.254, -119668.525, -114130.592, -107407.623, -99540.7824, -90571.2389, -78085.1257, -60870.3741},
        {-6.05520415, -5.57853714, -5.1088701, -4.64606916, -4.19000046, -3.74053012, -3.29752429, -2.86084908, -2.43037064, -2.0059551, -1.58746859, -1.17477724, -0.767747192, -0.366244567, 0.0298644973, 0.420713869, 0.806437415, 1.187169, 1.5630425, 1.93419177, 2.30075069, 2.66285312, 3.02063293, 3.37422399, 3.72376016, 4.06937531, 4.41120332, 4.74937804, 5.08403334, 5.41530309, 5.74332116, 6.06822142, 6.39013774, 6.70920397, 7.03562098, 7.3642774, 7.68994085, 8.01241852, 8.33151756, 8.64704516, 8.9588085, 9.26661474, 9.57027107, 9.86958465, 10.1643627, 10.4544123, 10.7395407, 11.019555, 11.2942625, 11.5634703, 11.8269855, 12.1101624, 12.3867074, 12.6558737, 12.9169148, 13.169084, 13.4116346, 13.6438201, 13.8648938, 14.0741091, 14.2707193, 14.4539778, 14.6231379, 14.7774531, 14.9161767, 15.038562, 15.1438625, 15.2420858, 15.3429814, 15.4236227, 15.4825659, 15.5183672, 15.5295827, 15.5147687, 15.4724814, 15.401277, 15.2997117, 15.1663416, 14.9997231, 14.7984122, 14.5609653, 14.2859385, 13.971888, 13.61737, 12.6169396, 11.283858, 9.9386458, 8.59946027, 7.28445859, 6.01179797, 4.7996356, 3.66612869, 2.62943443, 1.70771003, 0.919112681, 0.281799592, -0.186072041, -0.466345018, -0.540862136, -0.391466197, 0, 1.26915075, 2.85986547, 4.76534605, 6.97879443, 9.49341251, 12.3024022, 15.3989655, 18.7763042, 22.4276202, 26.3461156, 30.5249922, 34.9574519, 39.6366966, 44.5559283, 49.7083489, 55.0871603, 60.4813883, 65.8550669, 71.6632248, 77.8553118, 84.3807779, 91.189073, 98.2296469, 105.45195, 112.805431, 120.239541, 127.70373, 135.147447, 142.520142, 150.294506, 157.704894, 164.543782, 170.603649, 176.724778, 182.175742, 185.945704, 187.512799, 186.355164, 181.950934, 173.778246, 166.305119, 152.501146, 130.868183, 99.9080828, 58.1227015, 12.4700696, -49.1181781, -130.254231, -234.550278, -365.618508, -523.620388, -719.83638, -949.620043, -1217.71081, -1548.01096, -1955.81902, -2456.43355, -3065.14792, -3797.26702, -4668.10141, -5667.26545, -6865.11299, -8305.87352, -10034.2505, -12095.7508, -14535.5122, -17361.9242, -20610.5974, -24451.8874, -28979.9305, -34239.7247, -40382.1042, -47518.2144, -55759.2012, -65216.2099, -80038.8704, -95290.2231, -110071.793, -123485.104, -134631.681, -139759.183, -142725.578, -143855.474, -143960.588, -143172.718, -141420.697, -138846.468, -135591.977, -131799.167, -127609.981, -126479.139, -123473.146, -118840.222, -112828.588, -105686.465, -97662.0739, -89003.6355, -76009.0315, -57260.926},
        {-6.3334436, -5.83627996, -5.34604647, -4.86262438, -4.3858949, -3.91573928, -3.45203873, -2.99467449, -2.54352779, -2.09847986, -1.65941193, -1.22620522, -0.798740971, -0.37690041, 0.0394352349, 0.450384733, 0.856066854, 1.25660037, 1.65210405, 2.04269666, 2.42849698, 2.80962377, 3.18619581, 3.55833187, 3.9261507, 4.2897711, 4.64931182, 5.00489164, 5.35662932, 5.70464364, 6.04905336, 6.38997727, 6.72753412, 7.06184269, 7.40143045, 7.74203496, 8.0792927, 8.41304956, 8.74315143, 9.06944421, 9.39177379, 9.70998605, 10.0239269, 10.3334422, 10.6383779, 10.9385798, 11.2338939, 11.524166, 11.8092421, 12.0889679, 12.3631895, 12.6635001, 12.95731, 13.2437768, 13.5220582, 13.7913119, 14.0506956, 14.2993669, 14.5364834, 14.761203, 14.9726832, 15.1700817, 15.3525561, 15.5192642, 15.6693637, 15.8020121, 15.9163671, 16.0212682, 16.1253221, 16.2077189, 16.2669886, 16.3016612, 16.3102668, 16.2913353, 16.243397, 16.1649818, 16.0546197, 15.9108409, 15.7321753, 15.517153, 15.264304, 14.9721584, 14.6392463, 14.2640977, 13.2328063, 11.863002, 10.479676, 9.1012329, 7.74607758, 6.4326148, 5.17924934, 4.00438599, 2.92642954, 1.96378475, 1.13485643, 0.458049348, -0.0482317081, -0.365581954, -0.475596607, -0.359870884, 0, 1.36643641, 3.03964886, 5.01333264, 7.28118302, 9.83689527, 12.6741647, 15.7866865, 19.1681561, 22.8122686, 26.7127194, 30.8632038, 35.2574169, 39.8890542, 44.7518108, 49.8393821, 55.1454633, 60.3853759, 65.5130476, 71.1518285, 77.2357625, 83.6988932, 90.4752645, 97.49892, 104.703904, 112.024259, 119.39403, 126.74726, 134.017993, 141.140273, 148.458697, 155.346219, 161.613717, 167.072069, 172.669628, 177.648535, 180.947602, 182.036465, 180.384759, 175.462119, 166.73818, 158.445059, 143.840459, 121.462224, 89.8481998, 47.536231, 1.01249841, -61.1205721, -142.312219, -246.011682, -375.668198, -530.713616, -723.52471, -951.41196, -1219.5232, -1549.00751, -1953.82976, -2447.9548, -3042.88635, -3755.60975, -4605.79517, -5586.37034, -6764.81779, -8189.31433, -9903.21902, -11941.7275, -14343.7826, -17104.8118, -20252.4764, -23973.9929, -28366.3243, -33533.9538, -39565.1674, -46554.3249, -54595.7862, -63783.9114, -77975.5709, -92561.9863, -106696.89, -119534.014, -130227.09, -135220.229, -138183.199, -139431., -139853.243, -139537.475, -138331.824, -136335.808, -133648.947, -130370.76, -126600.766, -125945.813, -122895.862, -117898.998, -111403.307, -103856.874, -95707.7854, -87404.1253, -73941.2809, -53810.1482},
        {-6.65623729, -6.13500706, -5.62086552, -5.11370067, -4.6134005, -4.119853, -3.63294617, -3.15256801, -2.67860651, -2.21094966, -1.74948546, -1.2941019, -0.844686977, -0.401128684, 0.0366849832, 0.468866031, 0.895526466, 1.31677829, 1.73273352, 2.14350414, 2.54920218, 2.94993963, 3.3458285, 3.73698079, 4.12350852, 4.50552368, 4.88313828, 5.25646434, 5.62561385, 5.99069881, 6.35183124, 6.70912314, 7.06268652, 7.41263338, 7.76632074, 8.12008572, 8.47017833, 8.81647547, 9.15885404, 9.49719092, 9.83136301, 10.1612472, 10.4867204, 10.8076594, 11.1239412, 11.4354427, 11.7420407, 12.0436122, 12.340034, 12.631183, 12.9169362, 13.2346143, 13.5458386, 13.8496742, 14.1451864, 14.4314403, 14.7075012, 14.9724341, 15.2253042, 15.4651769, 15.6911171, 15.9021901, 16.0974611, 16.2759953, 16.4368578, 16.5791138, 16.7018285, 16.8126384, 16.9189743, 17.0022235, 17.0608957, 17.0935002, 17.0985468, 17.0745449, 17.020004, 16.9334337, 16.8133435, 16.6582429, 16.4666415, 16.2370489, 15.9679744, 15.6579278, 15.3054185, 14.908956, 13.8472299, 12.4416968, 11.021263, 9.60455256, 8.21018921, 6.85679689, 5.56299947, 4.34742085, 3.2286849, 2.22541552, 1.35623659, 0.639771997, 0.0946456196, -0.260518654, -0.407096939, -0.326465349, 0, 1.44440518, 3.18273435, 5.20910734, 7.51764397, 10.1024641, 12.9576875, 16.0774339, 19.4558234, 23.0869755, 26.9650103, 31.0840474, 35.4382068, 40.0216082, 44.8283715, 49.8526165, 55.088463, 60.1783173, 65.0660546, 70.5413369, 76.522917, 82.9295475, 89.679981, 96.6929703, 103.887268, 111.181626, 118.494799, 125.745537, 132.852595, 139.734724, 146.60197, 152.974986, 158.685136, 163.563785, 168.687465, 173.264957, 176.172999, 176.869404, 174.811989, 169.458567, 160.266953, 151.189966, 135.822373, 112.74253, 80.5287964, 37.759529, -9.48847195, -72.009902, -153.101014, -256.058061, -384.177297, -536.195197, -725.540851, -951.576192, -1219.86242, -1548.61755, -1950.37695, -2437.676, -3017.60897, -3709.38893, -4538.16464, -5501.89, -6663.00148, -8070.76993, -9766.44629, -11777.6923, -14138.4074, -16833.6607, -19884.6921, -23491.7853, -27753.9464, -32825.2675, -38741.1945, -45581.6656, -53426.6189, -62355.9927, -75978.132, -89962.4664, -103506.832, -115809.066, -126067.005, -130876.584, -133772.141, -135066.612, -135761.577, -135897.383, -135227.448, -133806.467, -131689.137, -128930.154, -125584.214, -125298.53, -122169.605, -116821.554, -109878.491, -101964.531, -93703.7884, -85720.379, -71828.1923, -50572.7013},
        {-7.01295006, -6.44248812, -5.88164886, -5.33023715, -4.78805782, -4.25491573, -3.73061572, -3.21496264, -2.70776133, -2.20881665, -1.71793344, -1.23491655, -0.759570827, -0.291701118, 0.16888773, 0.622390868, 1.06900345, 1.50892063, 1.94233755, 2.36944938, 2.79045126, 3.20553835, 3.61490579, 4.01874875, 4.41726237, 4.8106418, 5.19908221, 5.58277873, 5.96192653, 6.33672075, 6.70735656, 7.07402909, 7.43693351, 7.79626496, 8.16290546, 8.53148194, 8.89649291, 9.25778669, 9.61521164, 9.9686161, 10.3178484, 10.6627569, 11.00319, 11.3389959, 11.6700231, 11.9961198, 12.3171344, 12.6329153, 12.9433108, 13.2481692, 13.547339, 13.879, 14.203838, 14.5208705, 14.8291147, 15.1275881, 15.4153081, 15.691292, 15.9545573, 16.2041214, 16.4390015, 16.6582152, 16.8607798, 17.0457126, 17.2120312, 17.3587528, 17.4848948, 17.5978482, 17.7048027, 17.7875061, 17.8444329, 17.8740577, 17.8748551, 17.8452996, 17.7838659, 17.6890284, 17.5592618, 17.3930405, 17.1888393, 16.9451327, 16.6603951, 16.3331013, 15.9617258, 15.5447431, 14.4557898, 13.0189022, 11.5656265, 10.1147144, 8.68491769, 7.29498815, 5.96367754, 4.70973761, 3.55192013, 2.50897687, 1.59965957, 0.84272002, 0.256909965, -0.139018827, -0.326314594, -0.286225572, 0, 1.3160802, 2.94641888, 4.88353002, 7.11992758, 9.64812551, 12.4606378, 15.5499784, 18.9086612, 22.5292003, 26.4041096, 30.525903, 34.8870946, 39.4801982, 44.2977279, 49.3321976, 54.5761213, 59.6547243, 64.5080508, 69.9593866, 75.922541, 82.3113231, 89.0395424, 96.021008, 103.169529, 110.398915, 117.622975, 124.755519, 131.710355, 138.401294, 144.96025, 151.002954, 156.377783, 160.933115, 165.872239, 170.366075, 173.224236, 173.888625, 171.801146, 166.403703, 157.138201, 147.757551, 132.080602, 108.7058, 76.2315895, 33.2564159, -14.5390763, -77.4091905, -158.526031, -261.061702, -388.188307, -538.044974, -724.988202, -950.342003, -1220.05564, -1549.52162, -1949.8926, -2432.32126, -2999.10103, -3672.2573, -4483.48012, -5440.91702, -6592.53489, -7969.28791, -9606.86799, -11548.9949, -13835.7035, -16470.5371, -19484.0699, -23040.6781, -27229.0519, -32141.6201, -37862.7592, -44479.8651, -52080.334, -60751.5617, -74144.0481, -87891.3096, -101189.966, -113236.638, -123227.946, -127695.788, -130277.987, -131313.185, -131979.991, -132284.562, -131883.068, -130799.165, -129056.51, -126678.757, -123689.563, -124124.548, -121278.176, -115885.347, -108680.963, -100399.925, -91777.1338, -83547.4915, -69141.4716, -48027.0107},
        {-7.42228891, -6.79927535, -6.18859417, -5.58996079, -5.0030906, -4.427699, -3.86350139, -3.31021318, -2.76754976, -2.23522654, -1.71295892, -1.20046229, -0.697452053, -0.203643616, 0.281247622, 0.757506263, 1.2254169, 1.68526415, 2.13733259, 2.58190683, 3.01927147, 3.44971111, 3.87351035, 4.29095378, 4.70232602, 5.10791164, 5.50799527, 5.90286149, 6.2927949, 6.67808011, 7.05900171, 7.43584431, 7.80889249, 8.17843087, 8.55935333, 8.94433252, 9.32586554, 9.70376289, 10.0778351, 10.4478926, 10.8137459, 11.1752055, 11.532082, 11.8841858, 12.2313274, 12.5733174, 12.9099662, 13.2410843, 13.5664822, 13.8859705, 14.1993596, 14.544993, 14.8833129, 15.2132945, 15.533913, 15.8441435, 16.1429613, 16.4293415, 16.7022592, 16.9606898, 17.2036082, 17.4299898, 17.6388098, 17.8290432, 17.9996652, 18.1496512, 18.2779761, 18.3917305, 18.4978098, 18.5784523, 18.632107, 18.6572231, 18.6522496, 18.6156356, 18.5458303, 18.4412828, 18.3004422, 18.1217575, 17.9036779, 17.6446525, 17.3431304, 16.9975607, 16.6063925, 16.1680749, 15.0523955, 13.5853573, 12.1005191, 10.6167313, 9.15284434, 7.72770853, 6.36017428, 5.06909195, 3.87331192, 2.79168458, 1.8430603, 1.04628945, 0.42022242, -0.0162904182, -0.244398686, -0.245252006, 0, 1.17476271, 2.68522798, 4.52227804, 6.67679507, 9.13966128, 11.9017589, 14.9539701, 18.287177, 21.892262, 25.7601071, 29.8815947, 34.2476068, 38.8490258, 43.6767338, 48.7216129, 53.9745455, 59.0517023, 63.8904354, 69.3291883, 75.2785481, 81.6491016, 88.3514356, 95.2961372, 102.393793, 109.554991, 116.690316, 123.710357, 130.5257, 137.046932, 143.341565, 149.105722, 154.202911, 158.496642, 163.303999, 167.761703, 170.618856, 171.302895, 169.241257, 163.861379, 154.590698, 144.997259, 129.107706, 105.539354, 72.9095238, 29.8355315, -18.3812096, -81.4785008, -162.510048, -264.529558, -390.590736, -538.301367, -722.817206, -947.255606, -1217.90456, -1547.52544, -1945.97321, -2423.10282, -2976.65502, -3631.35221, -4425.13229, -5376.43055, -6518.24999, -7861.61932, -9436.48265, -11304.8352, -13513.9603, -16088.5593, -19070.0608, -22582.6937, -26703.8504, -31451.9671, -36972.4625, -43363.1367, -50721.7901, -59146.2227, -72374.2518, -85953.6565, -99072.2327, -110917.776, -120678.084, -124786.7, -127021.839, -127751.689, -128337.824, -128755.216, -128564.931, -127762.638, -126344.005, -124304.703, -121640.401, -122643.929, -120067.182, -114703.863, -107347.674, -98792.3197, -89831.5027, -81258.9262, -66400.6089, -45715.9718},
        {-7.89041817, -7.21480918, -6.55407633, -5.9078523, -5.27576972, -4.65746126, -4.05255956, -3.46069728, -2.88150706, -2.31462157, -1.75967346, -1.21629537, -0.684119955, -0.162779875, 0.348092222, 0.848863682, 1.33990185, 1.82157409, 2.29424772, 2.75829012, 3.21406861, 3.66195056, 4.1023033, 4.53549419, 4.96189057, 5.38185979, 5.7957692, 6.20398614, 6.60687797, 7.00481203, 7.39815566, 7.78727622, 8.17254106, 8.55431752, 8.95138081, 9.35450718, 9.75425158, 10.150384, 10.5426744, 10.9308929, 11.3148092, 11.6941935, 12.0688157, 12.4384458, 12.8028537, 13.1618095, 13.5150831, 13.8624445, 14.2036637, 14.5385106, 14.8667552, 15.2267356, 15.5788176, 15.921935, 16.2550219, 16.577012, 16.8868394, 17.1834378, 17.4657412, 17.7326834, 17.9831985, 18.2162202, 18.4306825, 18.6255192, 18.7996643, 18.9520517, 19.0816153, 19.1949345, 19.2984054, 19.3751934, 19.4237366, 19.4424733, 19.4298418, 19.3842804, 19.3042272, 19.1881206, 19.0343989, 18.8415002, 18.6078629, 18.3319251, 18.0121252, 17.6469015, 17.2346921, 16.7739354, 15.6316702, 14.1353151, 12.6198776, 11.1042862, 9.6074691, 8.14835483, 6.74587173, 5.4189482, 4.1865126, 3.06749334, 2.08081878, 1.24541733, 0.580217357, 0.104147253, -0.163864598, -0.20488981, 0, 1.05119946, 2.45564837, 4.20288043, 6.28242933, 8.68382874, 11.3966124, 14.4103139, 17.7144671, 21.2986055, 25.1522629, 29.264973, 33.6262694, 38.2256859, 43.0527562, 48.0970139, 53.3479927, 58.4171649, 63.2400556, 68.6613371, 74.58953, 80.9331549, 87.6007322, 94.5007826, 101.541826, 108.632385, 115.680977, 122.596125, 129.286349, 135.660169, 141.736854, 147.273569, 152.14561, 156.228272, 160.923661, 165.343669, 168.193036, 168.889017, 166.848863, 161.489828, 152.229165, 142.480729, 126.448814, 102.768727, 70.0757788, 27.0052772, -21.5226028, -84.7286348, -165.548726, -266.918785, -391.774719, -537.267907, -719.188764, -942.155191, -1212.71785, -1541.37347, -1936.84706, -2407.86364, -2948.32333, -3585.14535, -4361.42155, -5304.54119, -6434.06278, -7742.59537, -9254.86553, -11050.0212, -13182.23, -15696.2763, -18643.8949, -22111.2264, -26164.9493, -30750.4836, -36074.422, -42245.4175, -49372.1229, -57563.1913, -70653.6986, -84094.2689, -97061.9494, -108733.787, -118286.829, -122052.694, -123950.81, -124377.695, -124863.21, -125355.258, -125329.51, -124757.477, -123610.672, -121860.606, -119478.79, -120806.234, -118441.303, -113174.409, -105795.962, -97096.375, -87866.0583, -78895.4238, -63677.2062, -43582.3837},
        {-8.42350218, -7.69853004, -6.99047024, -6.2988924, -5.6233661, -4.96346096, -4.31874658, -3.68879255, -3.07316848, -2.47144397, -1.88318863, -1.30797205, -0.745363833, -0.194933587, 0.34374909, 0.871114595, 1.38759333, 1.89361568, 2.38961206, 2.87601286, 3.35324848, 3.82174932, 4.28194578, 4.73426824, 5.17914712, 5.61701281, 6.04829571, 6.47342622, 6.89283473, 7.30695164, 7.71620736, 8.12103227, 8.52185678, 8.91911129, 9.33470439, 9.75787562, 10.1776064, 10.59363, 11.0056799, 11.4134895, 11.8167922, 12.2153213, 12.6088103, 12.9969925, 13.3796013, 13.7563702, 14.1270324, 14.4913214, 14.8489706, 15.1997134, 15.5432832, 15.9183701, 16.2849063, 16.6417805, 16.9878818, 17.3220989, 17.6433208, 17.9504363, 18.2423344, 18.5179038, 18.7760335, 19.0156124, 19.2355293, 19.4346732, 19.6119328, 19.7661972, 19.8963551, 20.0081094, 20.1069994, 20.1778607, 20.2191404, 20.2292856, 20.2067436, 20.1499614, 20.0573863, 19.9274654, 19.7586459, 19.5493749, 19.2980996, 19.0032671, 18.6633246, 18.2767193, 17.8418983, 17.3573087, 16.1882372, 14.6630288, 13.1176388, 11.5710617, 10.0422919, 8.55032383, 7.11415195, 5.75277069, 4.4851745, 3.33035781, 2.30731506, 1.43504069, 0.732529147, 0.218774866, -0.0872277119, -0.166484147, 0, 0.976137213, 2.31416677, 4.00286625, 6.03101323, 8.38738529, 11.06076, 14.0399149, 17.3136277, 20.8706758, 24.6998369, 28.7898885, 33.1296082, 37.7077736, 42.5131623, 47.5345518, 52.7607197, 57.7990256, 62.5837587, 67.966428, 73.8540784, 80.1537552, 86.7725035, 93.6173684, 100.595395, 107.613629, 114.579115, 121.398897, 127.980023, 134.229536, 140.137055, 145.496775, 150.190969, 154.10191, 158.672141, 163.003802, 165.782955, 166.423792, 164.340504, 158.947282, 149.658318, 139.779598, 123.649058, 99.9194528, 67.2435339, 24.2740547, -24.4709872, -87.6703943, -168.137725, -268.686537, -392.130389, -535.248126, -714.263779, -934.87895, -1203.80422, -1529.81015, -1920.74241, -2384.44669, -2912.15832, -3532.10839, -4290.6483, -5221.35956, -6333.88928, -7607.04729, -9061.59188, -10789.3607, -12849.5649, -15302.2374, -18206.8025, -21619.6704, -25598.9562, -30031.3448, -35172.7555, -41140.6445, -48052.4682, -56025.6829, -68967.3437, -82257.9086, -95067.434, -106565.977, -115923.593, -119397.144, -121012.012, -121186.77, -121584.286, -122130.596, -122233.274, -121844.276, -120915.559, -119399.079, -117246.794, -118561.025, -116305.219, -111194.294, -103943.165, -95266.7498, -85879.964, -76497.7244, -61042.8656, -41569.0455},
        {-9.02770528, -8.25987837, -7.5101508, -6.77806181, -6.06315065, -5.36495657, -4.68301882, -4.01687663, -3.36606927, -2.73013598, -2.108616, -1.50104859, -0.906972989, -0.325928447, 0.242545789, 0.79891047, 1.34362635, 1.87715418, 2.39995471, 2.9124887, 3.41521689, 3.90860004, 4.39309891, 4.86917424, 5.33728679, 5.7978973, 6.25146654, 6.69845525, 7.13932418, 7.5745341, 8.00454574, 8.42981987, 8.85081723, 9.26799858, 9.70504053, 10.1503076, 10.5918852, 11.0294808, 11.4628016, 11.8915551, 12.3154486, 12.7341894, 13.147485, 13.5550425, 13.9565695, 14.3517733, 14.7403611, 15.1220404, 15.4965185, 15.8635028, 16.2227007, 16.6140388, 16.9961333, 17.3678196, 17.7279332, 18.0753095, 18.4087842, 18.7271925, 19.02937, 19.3141523, 19.5803746, 19.8268726, 20.0524817, 20.2560374, 20.4363751, 20.5923303, 20.7227385, 20.8319045, 20.9240017, 20.9865856, 21.018137, 21.017137, 20.9820665, 20.9114064, 20.8036375, 20.6572409, 20.4706974, 20.242488, 19.9710936, 19.6549951, 19.2926735, 18.8826096, 18.4232844, 17.9131788, 16.7167197, 15.1627515, 13.5877395, 12.0107408, 10.4508127, 8.92701235, 7.45839698, 6.06402378, 4.76294994, 3.57423266, 2.51692914, 1.61009658, 0.872792163, 0.324073101, -0.0170034117, -0.131380178, 0, 0.980322719, 2.3172699, 3.99976455, 6.01672967, 8.35708828, 11.0097634, 13.963678, 17.2077551, 20.7309178, 24.5220889, 28.5701916, 32.8641489, 37.3928837, 42.1453191, 47.110378, 52.2769835, 57.2451977, 61.9483921, 67.255056, 73.070785, 79.3011747, 85.8518209, 92.6283193, 99.5362656, 106.481255, 113.368884, 120.104748, 126.594442, 132.743563, 138.533106, 143.76562, 148.32408, 152.091461, 156.490356, 160.63393, 163.224789, 163.684022, 161.43272, 155.891975, 146.482876, 136.465504, 120.253572, 96.5170652, 63.9259683, 21.1502655, -27.7340938, -90.8145813, -170.772704, -270.289968, -392.047881, -532.545556, -708.203152, -925.265074, -1190.47235, -1511.57993, -1895.88755, -2350.69494, -2866.21239, -3470.71302, -4211.11291, -5122.99625, -6211.64549, -7449.80629, -8856.23696, -10527.6617, -12525.0172, -14914.9919, -17760.0136, -21101.4199, -24992.4784, -29288.726, -34271.5806, -40062.755, -46783.9617, -54556.9133, -67300.1427, -80389.3371, -92997.0042, -104295.651, -113457.787, -116723.423, -118152.558, -118174.483, -118529.187, -119127.142, -119332.696, -119083.628, -118317.716, -116972.738, -114986.474, -115857.863, -113563.61, -108660.827, -101706.621, -93258.1035, -83872.3828, -74106.5687, -58569.1892, -39618.7562},
        {-9.72480593, -8.92582175, -8.14468824, -7.38097029, -6.63423275, -5.90404053, -5.1899585, -4.49155153, -3.80838451, -3.14002232, -2.48602983, -1.84597194, -1.21941351, -0.605919437, -0.00505458795, 0.583616152, 1.1605279, 1.72611579, 2.28081492, 2.82506042, 3.35928742, 3.88393102, 4.39942636, 4.90620854, 5.4047127, 5.89537395, 6.37862741, 6.85490819, 7.32465143, 7.78829224, 8.24626574, 8.69900704, 9.14695128, 9.59053356, 10.0524109, 10.52144, 10.9862107, 11.4464368, 11.9018324, 12.3521114, 12.7969879, 13.2361757, 13.6693889, 14.0963414, 14.5167473, 14.9303205, 15.336775, 15.7358247, 16.1271837, 16.5105658, 16.8856852, 17.2972699, 17.699044, 18.0897456, 18.4681128, 18.8328838, 19.1827967, 19.5165898, 19.8330011, 20.1307689, 20.4086313, 20.6653264, 20.8995924, 21.1101674, 21.2957897, 21.4551974, 21.5871286, 21.692703, 21.772983, 21.8217952, 21.8377234, 21.8193512, 21.7652626, 21.6740412, 21.5442708, 21.3745353, 21.1634183, 20.9095037, 20.6113752, 20.2676167, 19.8768118, 19.4375443, 18.9483981, 18.4079569, 17.1817889, 15.5983002, 13.9935055, 12.3865311, 10.7965034, 9.2425486, 7.74379319, 6.31936347, 4.9883858, 3.76998653, 2.683292, 1.74742857, 0.981522574, 0.404700372, 0.0360883094, -0.105187264, 0, 1.29076397, 2.87817408, 4.75489346, 6.9135852, 9.34691243, 12.0475383, 15.0081258, 18.2213381, 21.6798384, 25.3762897, 29.3033552, 33.4536979, 37.819981, 42.3948676, 47.1710208, 52.1411037, 56.8707352, 61.2915721, 66.3622109, 71.9838051, 78.0575085, 84.4844746, 91.1658572, 98.0028097, 104.896486, 111.74804, 118.458624, 124.929393, 131.0615, 136.867408, 142.096148, 146.615481, 150.29317, 154.304677, 157.863913, 159.836067, 159.696589, 156.92093, 150.98454, 141.36287, 131.996132, 116.536174, 93.5849063, 61.7442399, 19.6160846, -30.470913, -94.5732631, -175.020739, -274.143116, -394.270166, -531.1929, -701.771678, -911.587375, -1167.20582, -1477.2038, -1849.25522, -2291.03399, -2796.6482, -3390.42072, -4111.47348, -4983.6425, -6026.48001, -7233.43743, -8627.92673, -10284.1263, -12252.912, -14577.634, -17307.217, -20518.1795, -24271.4316, -28488.4759, -33391.0768, -39088.4008, -45689.614, -53303.8828, -65654.3468, -78332.7056, -90544.6322, -101495.8, -110391.881, -113620.913, -115084.628, -115196.259, -115660.472, -116379.305, -116716.697, -116601.591, -115962.929, -114729.653, -112830.706, -112245.937, -109540.906, -104931.579, -98633.9202, -90863.8929, -81837.4618, -71770.5908, -56541.8282, -37303.7702},
        {-10.4934981, -9.67496461, -8.87268656, -8.08630532, -7.31546227, -6.55979882, -5.81895635, -5.09257626, -4.38029993, -3.68176876, -2.99662415, -2.32450748, -1.66506015, -1.01792355, -0.382739065, 0.240851904, 0.853207968, 1.45468773, 2.04564981, 2.62645281, 3.19745533, 3.75901598, 4.31149338, 4.85524613, 5.39063283, 5.9180121, 6.43774255, 6.95018277, 7.45569139, 7.954627, 8.44734821, 8.93421364, 9.41558189, 9.89181157, 10.382551, 10.8781081, 11.3685592, 11.8536369, 12.3330737, 12.8066024, 13.2739554, 13.7348655, 14.1890653, 14.6362873, 15.0762643, 15.5087287, 15.9334133, 16.3500507, 16.7583735, 17.1581142, 17.5490056, 17.9808794, 18.402283, 18.811863, 19.208266, 19.5901384, 19.956127, 20.3048782, 20.6350386, 20.9452548, 21.2341734, 21.5004409, 21.7427039, 21.9596089, 22.1498026, 22.3119315, 22.4446421, 22.5458795, 22.6136056, 22.6479138, 22.6474961, 22.6110447, 22.5372515, 22.4248086, 22.272408, 22.0787418, 21.8425021, 21.5623808, 21.2370701, 20.8652619, 20.4456484, 19.9769215, 19.4577734, 18.886896, 17.6307637, 16.0174752, 14.3827281, 12.7457312, 11.125693, 9.54182239, 8.01332806, 6.55941873, 5.19930313, 3.95219, 2.83728805, 1.87380602, 1.08095262, 0.477936596, 0.0839666656, -0.0817484421, 0, 1.59291993, 3.42576184, 5.49463857, 7.79566298, 10.3249479, 13.0786061, 16.0527506, 19.2434941, 22.6469494, 26.2592295, 30.0764471, 34.0947152, 38.3101464, 42.7188538, 47.3169501, 52.1005481, 56.6054297, 60.761946, 65.607435, 71.0393672, 76.955213, 83.252443, 89.8285275, 96.580937, 103.407142, 110.204613, 116.87082, 123.303235, 129.399326, 135.216291, 140.433308, 144.89993, 148.46571, 152.081606, 155.046468, 156.365439, 155.557649, 152.142231, 145.638315, 135.565034, 126.1436, 110.759994, 88.0133759, 56.5029036, 14.8277361, -35.5620526, -99.6309477, -179.492519, -277.260337, -395.04797, -528.451509, -694.067814, -896.436298, -1141.80228, -1439.89398, -1798.87585, -2226.91236, -2722.43601, -3305.55505, -4006.9944, -4837.97224, -5833.47609, -7009.53867, -8394.62015, -10038.2383, -11980.2452, -14237.9368, -16845.3173, -19919.0961, -23529.2, -27674.6051, -32507.104, -38121.1833, -44611.3294, -52072.0289, -63984.482, -76209.7829, -87995.7394, -98590.1597, -107240.852, -110545.405, -112168.657, -112494.226, -113121.644, -113953.227, -114384.657, -114343.581, -113757.646, -112554.499, -110661.787, -108379.195, -105162.635, -100853.432, -95292.907, -88322.3852, -79783.1894, -69516.6433, -54655.8019, -35218.7963},
        {-11.3273078, -10.5028002, -9.6914403, -8.8930024, -8.10726069, -7.3339894, -6.57296273, -5.8239549, -5.08674012, -4.36109262, -3.64678659, -2.94359627, -2.25129585, -1.56965956, -0.898461613, -0.237476218, 0.41352241, 1.05476005, 1.6864625, 2.30885554, 2.92216495, 3.52661652, 4.12243604, 4.70984929, 5.28908205, 5.86036012, 6.42390928, 6.97995531, 7.528724, 8.07044114, 8.60533251, 9.1336239, 9.65554109, 10.1713099, 10.6955988, 11.2211085, 11.740366, 12.2531284, 12.7591528, 13.2581961, 13.7500157, 14.2343685, 14.7110115, 15.179702, 15.640197, 16.0922536, 16.5356289, 16.9700799, 17.3953638, 17.8112377, 18.2174586, 18.6680055, 19.1072383, 19.5337384, 19.9460878, 20.3428678, 20.7226604, 21.084047, 21.4256094, 21.7459293, 22.0435882, 22.317168, 22.5652501, 22.7864164, 22.9792484, 23.1423278, 23.2742363, 23.3705084, 23.426752, 23.4478331, 23.4325308, 23.3796243, 23.2878928, 23.1561156, 22.9830718, 22.7675406, 22.5083012, 22.2041328, 21.8538146, 21.4561257, 21.0098455, 20.513753, 19.9666275, 19.3672481, 18.0816504, 16.4382941, 14.7732446, 13.1058146, 11.4553172, 9.84106535, 8.2823719, 6.79854984, 5.40891212, 4.13277168, 2.98944147, 1.99823443, 1.17846352, 0.549441673, 0.13048184, -0.059103033, 0, 1.75524785, 3.72164347, 5.89678726, 8.27827963, 10.863721, 13.6507117, 16.6368522, 19.8197428, 23.196984, 26.7661762, 30.5249198, 34.4708152, 38.6014627, 42.9144628, 47.4074159, 52.0779223, 56.4492289, 60.4511712, 65.1542924, 70.4545457, 76.2478842, 82.4302612, 88.8976299, 95.5459434, 102.271155, 108.969218, 115.536085, 121.86771, 127.860045, 133.617319, 138.750843, 143.094088, 146.480524, 149.770283, 152.283506, 153.066556, 151.644908, 147.544034, 140.289406, 129.406498, 118.448069, 101.68661, 77.8596444, 45.7046982, 3.95929629, -44.8607824, -106.74083, -183.887884, -278.508983, -392.811165, -523.191218, -684.767849, -880.967684, -1117.57638, -1405.19597, -1752.26601, -2167.22607, -2651.39897, -3221.95849, -3903.84217, -4700.04591, -5655.07464, -6799.50727, -8164.70663, -9783.36383, -11687.5602, -13876.4088, -16373.4436, -19323.083, -22801.3662, -26864.1482, -31610.7581, -37124.2219, -43487.5658, -50783.8161, -62258.3492, -74041.2535, -85424.9675, -95701.9297, -104164.579, -107653.776, -109547.923, -110190.132, -111010.055, -111920.718, -112382.426, -112330.843, -111701.631, -110430.454, -108452.976, -104529.734, -100799.002, -96756.5221, -91898.0361, -85719.2859, -77716.0135, -67383.9609, -52806.801, -33611.5972},
        {-12.2111996, -11.4022486, -10.6011375, -9.80785193, -9.02237731, -8.24469918, -7.47480305, -6.71267444, -5.95829886, -5.21166184, -4.47274889, -3.74154552, -3.01803724, -2.30220959, -1.59404806, -0.893538178, -0.200665458, 0.484584584, 1.16222643, 1.83227457, 2.49474349, 3.14964766, 3.79700158, 4.43681973, 5.06911659, 5.69390664, 6.31120438, 6.92102428, 7.52338083, 8.11828852, 8.70576182, 9.28581523, 9.85846322, 10.4237203, 10.9892934, 11.5511891, 12.1053212, 12.6514546, 13.1893542, 13.7187848, 14.2395112, 14.7512983, 15.253911, 15.7471141, 16.2306723, 16.7043507, 17.167914, 17.6211271, 18.0637548, 18.495562, 18.9163135, 19.379508, 19.8300118, 20.2664247, 20.6873467, 21.091378, 21.4771184, 21.8431679, 22.1881267, 22.5105946, 22.8091717, 23.0824579, 23.3290533, 23.5475579, 23.7365717, 23.8946946, 24.0205267, 24.1096352, 24.15766, 24.169456, 24.1438201, 24.0795486, 23.9754383, 23.8302857, 23.6428875, 23.4120403, 23.1365406, 22.8151851, 22.4467704, 22.0300931, 21.5639498, 21.0471372, 20.4784517, 19.85669, 18.5435881, 16.8698726, 15.1739271, 13.4752221, 11.7932281, 10.1474155, 8.55725477, 7.04221637, 5.62177074, 4.31538834, 3.14253962, 2.12269504, 1.27532504, 0.61990009, 0.175890629, -0.0372328854, 0, 1.43766898, 3.15148305, 5.13408021, 7.37809845, 9.87617579, 12.6209502, 15.6050597, 18.8211424, 22.2618361, 25.9197789, 29.7876088, 33.8579638, 38.1234819, 42.5768011, 47.2105594, 52.0173948, 56.5527305, 60.7467454, 65.5810154, 70.9544899, 76.7661178, 82.9148484, 89.2996307, 95.8194138, 102.373147, 108.859779, 115.17826, 121.227537, 126.906562, 132.350382, 137.135277, 141.089366, 144.040766, 147.223604, 149.762994, 150.409118, 148.568295, 143.646842, 135.051076, 122.187315, 104.876158, 81.9835558, 52.8347695, 16.7550618, -26.9303049, -68.4209997, -121.485596, -189.417599, -275.510512, -383.057842, -512.331858, -673.339919, -868.706641, -1103.44157, -1387.42348, -1728.34496, -2133.8986, -2602.20359, -3152.7017, -3815.2784, -4602.05977, -5544.06112, -6655.64391, -7962.78083, -9511.11913, -11337.2751, -13452.2798, -15886.1162, -18762.5803, -22152.4021, -26087.6709, -30683.7024, -36024.6021, -42194.4759, -49277.4294, -60341.5448, -71740.9572, -82813.7782, -92898.1195, -101332.093, -105256.093, -107671.093, -108857.081, -109933.308, -110847.556, -111197.917, -110949.658, -110068.047, -108518.351, -106265.838, -101144.167, -97016.837, -93148.2957, -88802.9921, -83245.3749, -75739.8929, -65550.9948, -50695.0338, -33081.2788},
        {-13.164245, -12.3839002, -11.6045871, -10.82656, -10.0500734, -9.27538162, -8.50273907, -7.73240013, -6.96461917, -6.19965057, -5.43774872, -4.67916799, -3.92416277, -3.17298743, -2.42589635, -1.68314393, -0.94498453, -0.21167254, 0.516537662, 1.23939169, 1.95663517, 2.66801372, 3.37327296, 4.07215851, 4.76441598, 5.449791, 6.12802919, 6.79887615, 7.46207753, 8.11737893, 8.76452596, 9.40326427, 10.0333394, 10.6544971, 11.2651093, 11.8656374, 12.456559, 13.0376645, 13.6087441, 14.1695879, 14.7199862, 15.259729, 15.7886067, 16.3064094, 16.8129274, 17.3079507, 17.7912697, 18.2626744, 18.7219552, 19.1689021, 19.6033055, 20.0781975, 20.5389722, 20.9842654, 21.4127133, 21.8229517, 22.2136169, 22.5833447, 22.9307711, 23.2545322, 23.553264, 23.8256025, 24.0701837, 24.2856437, 24.4706183, 24.6237437, 24.7436558, 24.8267247, 24.8693745, 24.874915, 24.8421291, 24.7697996, 24.6567095, 24.5016418, 24.3033792, 24.0607048, 23.7724013, 23.4372519, 23.0540393, 22.6215464, 22.1385563, 21.6038517, 21.0162156, 20.374431, 19.034376, 17.3298086, 15.6022138, 13.871238, 12.1565277, 10.4777292, 8.85448905, 7.3064535, 5.85326902, 4.51458201, 3.31003887, 2.25928601, 1.38196983, 0.697736732, 0.226233126, -0.012894586, 0, 1.02445001, 2.40846959, 4.13842395, 6.20067828, 8.5815978, 11.2675477, 14.2448932, 17.4999995, 21.0192318, 24.7889553, 28.7955352, 33.0253368, 37.4647251, 42.1000655, 46.9177231, 51.9040631, 56.6640331, 61.1295099, 66.1495894, 71.6289044, 77.4720877, 83.5837721, 89.8685905, 96.2311757, 102.576161, 108.808178, 114.831861, 120.551842, 125.872753, 130.95744, 135.357647, 138.900541, 141.413294, 144.567995, 147.269354, 147.9301, 145.823927, 140.224528, 130.405597, 115.640828, 91.7969375, 62.5915006, 27.9647922, -12.1429124, -57.7913377, -91.4939788, -135.243347, -193.485726, -270.667395, -371.234639, -499.505251, -660.211041, -855.333367, -1089.12077, -1370.45063, -1706.12206, -2102.93419, -2554.87827, -3084.47218, -3728.20598, -4510.41134, -5444.23034, -6521.06749, -7759.92501, -9226.56784, -10965.2961, -13007.0844, -15392.7781, -18212.0825, -21527.0649, -25319.7258, -29744.5174, -34891.2224, -40849.6236, -47709.5038, -58409.0623, -69477.5612, -80292.6793, -90232.0953, -98673.4879, -103032.151, -105956.405, -107664.949, -108966.359, -109845.844, -110042.83, -109554.817, -108379.301, -106513.78, -103955.753, -97926.4164, -93471.5253, -89704.7707, -85739.8438, -80690.4356, -73670.2373, -63792.9399, -48717.0982, -32983.508},
        {-14.1934407, -13.4520895, -12.7036698, -11.9487569, -11.1879258, -10.4217517, -9.65080977, -8.87567511, -8.0969229, -7.31512828, -6.53086638, -5.74471237, -4.95724139, -4.16902858, -3.38064909, -2.59267807, -1.80569067, -1.02026203, -0.236967303, 0.543618369, 1.32091984, 2.09436195, 2.86336957, 3.62736755, 4.38578073, 5.13803397, 5.88355212, 6.62176004, 7.35208258, 8.07394459, 8.78677093, 9.48998644, 10.183016, 10.8652844, 11.5231985, 12.1629659, 12.7909509, 13.4070008, 14.0109628, 14.6026844, 15.1820127, 15.7487951, 16.302879, 16.8441116, 17.3723403, 17.8874124, 18.3891751, 18.8774759, 19.352162, 19.8130807, 20.2600794, 20.7474361, 21.2193874, 21.6746007, 22.1117434, 22.5294827, 22.9264862, 23.3014211, 23.6529548, 23.9797547, 24.2804881, 24.5538224, 24.798425, 25.0129632, 25.1961044, 25.3465159, 25.4628652, 25.5426089, 25.5832327, 25.5859009, 25.5493595, 25.4723541, 25.3536308, 25.1919353, 24.9860135, 24.7346112, 24.4364743, 24.0903486, 23.69498, 23.2491142, 22.7514973, 22.2008749, 21.595993, 20.9355973, 19.5687473, 17.8326972, 16.0725102, 14.3080158, 12.5590434, 10.8454224, 9.18698219, 7.60355227, 6.11496205, 4.74104093, 3.50161835, 2.41652373, 1.50558651, 0.788636095, 0.285501923, 0.0160134159, 0, 0.643412419, 1.72287968, 3.21891492, 5.11203128, 7.38274187, 10.0115598, 12.9789983, 16.2655705, 19.8517894, 23.7181682, 27.84522, 32.213458, 36.8033954, 41.5955452, 46.5704205, 51.7085346, 56.6737348, 61.4012964, 66.5880001, 72.146502, 77.9894584, 84.0295255, 90.1793594, 96.3516165, 102.458953, 108.414025, 114.129488, 119.518, 124.492215, 129.241709, 133.300682, 136.498717, 138.665393, 141.833045, 144.729877, 145.496694, 143.302249, 137.315299, 126.704598, 110.638902, 81.896622, 47.9817446, 9.31330883, -33.6896455, -80.6080787, -108.112044, -144.421136, -194.844041, -264.689446, -359.266039, -485.966925, -645.468749, -839.305449, -1071.13108, -1348.92829, -1678.73627, -2066.59421, -2503.13074, -3013.29802, -3638.85978, -4414.79333, -5337.96386, -6377.55134, -7546.16681, -8930.0127, -10581.5118, -12553.2184, -14897.5893, -17666.5992, -20912.4954, -24554.6963, -28798.7135, -33742.774, -39485.1043, -46123.9313, -56526.3572, -67329.5144, -77939.4107, -87762.0538, -96203.4517, -100884.886, -104186.908, -106280.408, -107726.78, -108542.339, -108580.693, -107866.294, -106423.594, -104277.044, -101451.097, -94843.3431, -90141.1378, -86396.3217, -82660.7352, -77986.219, -71424.6135, -62027.7591, -47023.0444, -33114.1754},
        {-15.3459475, -14.641643, -13.9217241, -13.1871184, -12.4387533, -11.6775566, -10.9044557, -10.1203783, -9.32625188, -8.52300406, -7.7115624, -6.89285449, -6.06780787, -5.23735014, -4.40240887, -3.56391162, -2.72278597, -1.8799595, -1.03635977, -0.192914367, 0.649449147, 1.4898032, 2.32722021, 3.16077261, 3.98953282, 4.81257328, 5.6289664, 6.43778462, 7.23810037, 8.02898606, 8.80951414, 9.57875701, 10.3357871, 11.0796769, 11.780693, 12.4529167, 13.1107753, 13.7542761, 14.3834262, 14.998233, 15.5987037, 16.1848454, 16.7566655, 17.3141711, 17.8573694, 18.3862678, 18.9008733, 19.4011933, 19.887235, 20.3590057, 20.8165124, 21.3247416, 21.8173129, 22.2928247, 22.7498757, 23.1870645, 23.6029896, 23.9962498, 24.3654437, 24.7091697, 25.0260266, 25.3146129, 25.5735273, 25.8013683, 25.9967346, 26.1582247, 26.2844373, 26.3725703, 26.419855, 26.4277791, 26.395032, 26.3203031, 26.2022819, 26.0396576, 25.8311198, 25.5753579, 25.2710612, 24.9169192, 24.5116214, 24.053857, 23.5423156, 22.9756865, 22.3526592, 21.671923, 20.2747912, 18.5024682, 16.7040599, 14.8995887, 13.1090772, 11.3525479, 9.65002342, 8.02152626, 6.48707895, 5.06670404, 3.78042407, 2.64826156, 1.69023907, 0.92637913, 0.376704278, 0.0612370552, 0, 0.733717137, 1.88706293, 3.44082281, 5.37578217, 7.67272643, 10.312441, 13.2757113, 16.5433228, 20.0960608, 23.9147108, 27.9800581, 32.2728883, 36.7739867, 41.4641387, 46.3241297, 51.3347452, 56.2201721, 60.9245389, 66.0090323, 71.3994523, 77.0215989, 82.8012722, 88.6642723, 94.5363991, 100.343453, 106.011233, 111.46554, 116.632174, 121.436935, 126.120108, 130.177697, 133.441156, 135.741939, 138.885977, 141.717485, 142.554155, 140.635097, 135.199425, 125.486249, 110.734683, 85.5219182, 55.1678311, 19.8236481, -20.3594046, -65.2301004, -95.9187528, -135.67221, -189.018862, -260.487096, -354.605301, -476.489808, -629.362467, -814.683177, -1036.29097, -1302.88229, -1620.97268, -1997.07772, -2425.70508, -2928.10726, -3538.6363, -4284.8207, -5169.31815, -6163.08337, -7280.28897, -8608.33797, -10201.0195, -12118.1092, -14414.9484, -17124.927, -20293.8516, -23792.5863, -27868.795, -32628.8413, -38179.0893, -44625.9028, -54894.7692, -65568.148, -76047.622, -85734.7742, -94031.1875, -98522.6631, -101637.087, -103554.234, -104889.441, -105650.071, -105662.01, -104941.696, -103505.569, -101370.071, -98551.6411, -92006.1845, -87184.5431, -83259.558, -79404.0702, -74790.921, -68592.9512, -59983.0021, -46298.2728, -32933.3449},
        {-16.5634187, -15.9072879, -15.2257222, -14.5200513, -13.7916049, -13.0417127, -12.2717045, -11.4829099, -10.6766586, -9.85428051, -9.01710517, -8.16646236, -7.30368177, -6.43009314, -5.54702618, -4.6558106, -3.75777611, -2.85425245, -1.94656931, -1.03605642, -0.124043493, 0.788139754, 1.69916361, 2.60769835, 3.51241426, 4.41198163, 5.30507074, 6.19035187, 7.06649531, 7.93217134, 8.78605025, 9.62680232, 10.4530978, 11.2636071, 12.0101202, 12.7157336, 13.4041071, 14.0754324, 14.7299012, 15.3677051, 15.9890356, 16.5940845, 17.1830434, 17.7561038, 18.3134576, 18.8552962, 19.3818113, 19.8931945, 20.3896376, 20.871332, 21.3384695, 21.8680599, 22.3818113, 22.8782501, 23.3559027, 23.8132955, 24.2489548, 24.661407, 25.0491784, 25.4107955, 25.7447846, 26.0496721, 26.3239843, 26.5662476, 26.7749884, 26.948733, 27.0860078, 27.1837802, 27.2390548, 27.2535735, 27.2259637, 27.1548529, 27.0388684, 26.8766377, 26.6667883, 26.4079476, 26.098743, 25.7378019, 25.3237518, 24.8552202, 24.3308344, 23.7492219, 23.1090102, 22.4088266, 20.9828229, 19.1758408, 17.3407084, 15.4976165, 13.6667562, 11.8683184, 10.1224941, 8.44947423, 6.86944968, 5.40261143, 4.06915044, 2.88925765, 1.88312402, 1.07094049, 0.47289801, 0.10918753, 0, 0.916670601, 2.21686495, 3.88303337, 5.89762622, 8.24309383, 10.9018865, 13.8564547, 17.0892486, 20.5827186, 24.3193151, 28.2814884, 32.4516889, 36.8123667, 41.3459725, 46.0349563, 50.8617687, 55.6043452, 60.2159198, 65.1415356, 70.3198184, 75.6893938, 81.1888876, 86.7569257, 92.3321338, 97.8531377, 103.258563, 108.487036, 113.477182, 118.167626, 122.837756, 126.960491, 130.372228, 132.909365, 136.020525, 138.735993, 139.656126, 138.133652, 133.521303, 125.171806, 112.437892, 92.896266, 68.2162102, 38.0979376, 2.24166043, -39.6524087, -75.6091519, -121.271988, -180.009432, -255.189997, -350.182198, -466.607005, -611.716359, -787.078055, -996.896107, -1250.75681, -1555.82994, -1919.28524, -2340.6532, -2836.47907, -3431.64185, -4142.85047, -4983.02545, -5930.67092, -7003.15116, -8283.67683, -9825.23401, -11689.6975, -13932.3579, -16575.9138, -19661.4997, -23018.635, -26932.6253, -31517.8562, -36888.7136, -43159.5832, -53358.6854, -63958.1125, -74343.7916, -83901.6498, -92017.614, -96164.96, -98917.367, -100480.47, -101605.58, -102282.335, -102273.007, -101578.603, -100200.128, -98138.5886, -95394.9909, -89146.9206, -84285.6737, -80157.0965, -76107.0354, -71481.3368, -65625.847, -57886.4123, -45821.8774, -32783.2601},
        {-17.8259259, -17.2329178, -16.603161, -15.9384409, -15.2405435, -14.5112543, -13.752359, -12.9656433, -12.152893, -11.3158936, -10.4564309, -9.5762906, -8.67725834, -7.76111984, -6.82966076, -5.88466682, -4.92792368, -3.96121704, -2.98633258, -2.005056, -1.01917299, -0.0304692227, 0.959269605, 1.94825781, 2.93470969, 3.91683958, 4.89286177, 5.86099059, 6.81944034, 7.76642533, 8.70015989, 9.61885831, 10.5207349, 11.404004, 12.2002492, 12.9424016, 13.6641933, 14.3660008, 15.0482008, 15.7111699, 16.3552846, 16.9809217, 17.5884576, 18.1782691, 18.7507327, 19.306225, 19.8451226, 20.3678022, 20.8746403, 21.3660136, 21.8422987, 22.3914851, 22.9244372, 23.4396324, 23.9355479, 24.4106611, 24.8634493, 25.2923899, 25.6959601, 26.0726373, 26.4208989, 26.7392221, 27.0260843, 27.2799628, 27.4993349, 27.682678, 27.8284694, 27.9337545, 27.9956131, 28.0154756, 27.9919124, 27.9234936, 27.8087893, 27.6463696, 27.4348048, 27.172665, 26.8585204, 26.490941, 26.0684971, 25.5897588, 25.0532963, 24.4576798, 23.8014793, 23.0832651, 21.6315861, 19.7941161, 17.9265819, 16.0492972, 14.1825756, 12.3467308, 10.5620763, 8.84892577, 7.22759289, 5.71839126, 4.34163451, 3.11763629, 2.06671021, 1.20916991, 0.565329022, 0.155501175, 0, 1.11138453, 2.56577058, 4.34773417, 6.44185132, 8.83269805, 11.5048504, 14.4428843, 17.6313759, 21.0549012, 24.6980361, 28.5453567, 32.581439, 36.7908591, 41.1581929, 45.6680165, 50.3049059, 54.8859587, 59.3718255, 64.1259024, 69.0968772, 74.2334378, 79.4842719, 84.7980673, 90.1235118, 95.4092932, 100.604099, 105.656618, 110.515537, 115.129544, 119.792006, 123.980549, 127.540457, 130.317015, 133.375121, 135.950251, 136.952652, 135.861725, 132.15687, 125.317486, 114.822973, 101.570588, 83.1903538, 58.8842622, 27.854307, -10.6975186, -52.2019541, -104.16609, -168.72975, -248.032758, -344.214938, -455.216113, -592.509439, -757.895447, -955.96515, -1197.00679, -1488.75068, -1838.92711, -2251.71387, -2739.20118, -3317.35474, -3992.03471, -4787.15892, -5691.43935, -6725.72641, -7965.32725, -9460.55247, -11270.3022, -13447.114, -16012.0304, -19003.9095, -22223.4577, -25984.3231, -30407.8005, -35615.185, -41727.7716, -51892.4698, -62448.5562, -72760.9222, -82194.4589, -90114.0577, -93845.1429, -96151.7173, -97272.7292, -98124.7162, -98678.4144, -98622.2313, -97941.0193, -96619.6312, -94642.9195, -91995.7367, -86135.6304, -81301.4005, -77014.91, -72798.0216, -68172.5982, -62660.5026, -55783.5977, -45385.0304, -32684.9364},
        {-19.0363347, -18.5363648, -17.9858198, -17.3870215, -16.742292, -16.0539529, -15.3243263, -14.555734, -13.7504979, -12.9109399, -12.0393818, -11.1381456, -10.2095531, -9.25592626, -8.27958692, -7.28285696, -6.26805828, -5.23751276, -4.19354229, -3.13846874, -2.07461401, -1.00429998, 0.0701514625, 1.14641844, 2.22217906, 3.29511145, 4.36289371, 5.42320397, 6.47372033, 7.51212092, 8.53608384, 9.54328721, 10.5314092, 11.4981278, 12.354569, 13.1435075, 13.9087584, 14.6508085, 15.3701447, 16.067254, 16.7426233, 17.3967393, 18.0300892, 18.6431596, 19.2364376, 19.8104101, 20.3655639, 20.902386, 21.4213632, 21.9229824, 22.4077307, 22.9654405, 23.5053162, 24.0259078, 24.5257656, 25.0034394, 25.4574793, 25.8864355, 26.288858, 26.6632968, 27.008302, 27.3224237, 27.6042119, 27.8522167, 28.0649881, 28.2410762, 28.3790311, 28.4774949, 28.5351078, 28.5502296, 28.5214044, 28.4471765, 28.3260897, 28.1566883, 27.9375164, 27.6671179, 27.344037, 26.9668178, 26.5340044, 26.0441408, 25.4957712, 24.8874395, 24.21769, 23.4850667, 22.0176092, 20.1632169, 18.2778483, 16.3818067, 14.4953954, 12.6389175, 10.8326764, 9.09697519, 7.45211726, 5.91840581, 4.51614408, 3.26563534, 2.18718282, 1.30108977, 0.627659458, 0.187195117, 0, 1.09364663, 2.52648837, 4.2832566, 6.3486827, 8.70749802, 11.3444339, 14.2442218, 17.3915931, 20.771279, 24.3680111, 28.1665206, 32.1515389, 36.3077974, 40.6200275, 45.0729606, 49.6513279, 54.1649884, 58.5736914, 63.2577137, 68.1643138, 73.2407505, 78.4342824, 83.6921682, 88.9616665, 94.1900361, 99.3245355, 104.312423, 109.100959, 113.6374, 118.153345, 122.207456, 125.661688, 128.378, 131.289269, 133.72199, 134.717083, 133.815226, 130.557101, 124.483388, 115.134768, 104.706457, 89.2766984, 67.866806, 39.4980951, 3.19188057, -39.113417, -92.0428648, -157.304425, -236.606058, -331.655726, -439.393907, -572.12297, -732.133638, -924.023799, -1157.10181, -1438.56111, -1775.59515, -2171.27828, -2637.85921, -3192.08027, -3839.95038, -4605.32393, -5481.36581, -6488.52607, -7693.48792, -9141.71279, -10880.8523, -12956.9357, -15407.9616, -18276.4712, -21368.1233, -24998.7849, -29292.9951, -34375.2927, -40370.2167, -50498.3808, -61014.2305, -71268.286, -80611.0677, -88393.0957, -91853.0393, -93861.1704, -94673.0884, -95255.7823, -95583.2596, -95333.1153, -94494.1777, -93055.2748, -91005.2346, -88332.8855, -82430.9362, -77644.962, -73501.2226, -69525.9779, -65245.4877, -60186.012, -53873.8105, -44444.2769, -32824.8333},
        {-20.2695359, -19.8826548, -19.4296236, -18.9133576, -18.3367717, -17.7027809, -17.0143004, -16.2742451, -15.4855302, -14.6510706, -13.7737814, -12.8565777, -11.9023745, -10.9140869, -9.8946299, -8.84691856, -7.77386795, -6.67839309, -5.56340906, -4.43183089, -3.28657363, -2.13055234, -0.966682068, 0.20212214, 1.37294523, 2.54287216, 3.70898787, 4.86837732, 6.01812545, 7.15531722, 8.27703758, 9.38037147, 10.4624038, 11.5202197, 12.4437711, 13.285963, 14.1009876, 14.8894065, 15.6517814, 16.3886738, 17.1006454, 17.7882578, 18.4520726, 19.0926514, 19.7105558, 20.3063475, 20.880588, 21.433839, 21.966662, 22.4796188, 22.9732708, 23.5348413, 24.0763516, 24.5964848, 25.0939237, 25.5673514, 26.0154509, 26.436905, 26.8303968, 27.1946091, 27.528225, 27.8299275, 28.0983994, 28.3323238, 28.5303836, 28.6912618, 28.8136413, 28.8984181, 28.9464352, 28.9518112, 28.9130856, 28.8287979, 28.6974877, 28.5176944, 28.2879575, 28.0068165, 27.672811, 27.2844804, 26.8403642, 26.3390019, 25.778933, 25.158697, 24.4768335, 23.7318818, 22.2542943, 20.3907063, 18.4957832, 16.5897453, 14.6928127, 12.8252058, 11.0071446, 9.25884948, 7.60054058, 6.05243813, 4.63476236, 3.36773348, 2.2715717, 1.36649725, 0.672730336, 0.210491181, 0, 0.958588899, 2.27277, 3.92636902, 5.90321166, 8.18712362, 10.7619306, 13.6114583, 16.7195325, 20.0699788, 23.646623, 27.4332908, 31.4138078, 35.5719998, 39.8916924, 44.3567115, 48.9508827, 53.4516805, 57.8134521, 62.4867696, 67.4112406, 72.5264727, 77.7720736, 83.0876508, 88.4128121, 93.6871649, 98.8503169, 103.841876, 108.601449, 113.068645, 117.369614, 121.189021, 124.410511, 126.917726, 129.654453, 131.974264, 132.948031, 132.141356, 129.119841, 123.44909, 114.694703, 105.245483, 90.9845985, 70.9252882, 44.0807897, 9.46434068, -31.0489887, -82.0232513, -145.160668, -222.163458, -314.733842, -420.274195, -550.041951, -706.785418, -894.871106, -1121.96935, -1394.26715, -1717.95148, -2092.45888, -2532.26099, -3059.19356, -3684.31521, -4426.51813, -5281.05442, -6266.24627, -7439.5068, -8840.30576, -10500.8112, -12458.5997, -14778.0211, -17508.2806, -20479.5341, -23993.7364, -28175.9179, -33151.1087, -39044.3391, -49116.5249, -59572.8394, -69754.3415, -79002.0898, -86657.1434, -89907.0484, -91682.0513, -92246.1448, -92561.2657, -92615.1372, -92104.6726, -91032.1357, -89399.7901, -87209.8996, -84464.728, -78408.2385, -73718.0668, -69858.5555, -66294.047, -62488.8839, -57907.4086, -52013.9637, -43201.7543, -33020.78},
        {-21.5849875, -21.3236086, -20.9793602, -20.5557886, -20.0564402, -19.4848614, -18.8445985, -18.139198, -17.3722062, -16.5471696, -15.6676344, -14.7371471, -13.7592541, -12.7375017, -11.6754363, -10.5766043, -9.44455214, -8.2828261, -7.0949726, -5.88453802, -4.65506874, -3.41011114, -2.15321161, -0.887916531, 0.382227725, 1.65367477, 2.92287823, 4.18629172, 5.44036886, 6.68156327, 7.90632856, 9.11111836, 10.2923863, 11.446586, 12.4414895, 13.3405978, 14.2089587, 15.0472018, 15.8559565, 16.6358525, 17.3875194, 18.1115867, 18.8086841, 19.4794411, 20.1244872, 20.7444522, 21.3399655, 21.9116567, 22.4601555, 22.9860914, 23.490094, 24.0558624, 24.5991558, 25.1188031, 25.6136331, 26.0824746, 26.5241563, 26.937507, 27.3213556, 27.6745308, 27.9958615, 28.2841764, 28.5383043, 28.7570741, 28.9393144, 29.0838543, 29.1895223, 29.2594494, 29.2966631, 29.2911196, 29.2413688, 29.1459605, 29.0034448, 28.8123714, 28.5712904, 28.2787516, 27.933305, 27.5335004, 27.0778878, 26.5650171, 25.9934381, 25.3617009, 24.6683553, 23.9119512, 22.4266162, 20.5570868, 18.6560848, 16.7437222, 14.8401104, 12.9653614, 11.1395866, 9.38289792, 7.71540693, 6.15722535, 4.72846487, 3.4492372, 2.33965401, 1.419827, 0.709867869, 0.229888305, 0, 0.772694412, 1.92624076, 3.44310772, 5.30576394, 7.4966781, 9.99831885, 12.7931549, 15.8636548, 19.1922873, 22.7615211, 26.5538248, 30.551667, 34.7375165, 39.093842, 43.6031119, 48.2477952, 52.7643315, 57.1000426, 61.7938908, 66.7756245, 71.9749918, 77.3217408, 82.7456198, 88.1763771, 93.5437608, 98.777519, 103.8074, 108.563152, 112.974524, 117.043194, 120.600605, 123.554925, 125.814324, 128.402349, 130.669478, 131.668758, 130.973744, 128.15799, 122.795053, 114.458487, 105.251891, 91.4487633, 72.127652, 46.3671037, 13.2456655, -24.3926964, -72.1762091, -131.96768, -205.629918, -295.02573, -398.626824, -525.831786, -679.561764, -863.8067, -1084.73866, -1347.54997, -1657.43297, -2010.09768, -2422.37381, -2921.43547, -3523.73143, -4242.84386, -5076.07689, -6039.48753, -7180.90857, -8533.5872, -10113.3621, -11948.9672, -14132.9675, -16721.8215, -19578.6865, -22983.0608, -27058.3151, -31927.8197, -37714.9451, -47692.5845, -58047.2051, -68114.7967, -77231.3493, -84732.8526, -87776.8334, -89330.0534, -89662.1293, -89715.4747, -89493.3911, -88718.848, -87409.1635, -85581.6556, -83253.6425, -80442.4422, -74360.6917, -69833.7464, -66277.9089, -63109.4814, -59744.7662, -55600.0656, -50091.682, -41792.7118, -33123.6763},
        {-23.1223766, -22.9882058, -22.7536243, -22.4228522, -22.0001092, -21.4896152, -20.8955901, -20.2222538, -19.4738261, -18.6545269, -17.7685761, -16.8201935, -15.813599, -14.7530126, -13.6426539, -12.486743, -11.2894997, -10.0551439, -8.78789532, -7.49197399, -6.17159974, -4.83099241, -3.4743719, -2.10595807, -0.729970783, 0.649370085, 2.02784467, 3.40123309, 4.76531549, 6.11587199, 7.44868273, 8.75952783, 10.0441874, 11.2984416, 12.3642413, 13.3175168, 14.2360474, 15.1206054, 15.9719628, 16.7908918, 17.5781648, 18.3345537, 19.0608309, 19.7577686, 20.4261388, 21.0667138, 21.6802658, 22.267567, 22.8293895, 23.3665055, 23.8796873, 24.455805, 25.0076664, 25.5341772, 26.0342431, 26.50677, 26.9506635, 27.3648294, 27.7481735, 28.0996016, 28.4180192, 28.7023323, 28.9514466, 29.1642678, 29.3397017, 29.476654, 29.5740304, 29.6358156, 29.6658723, 29.6526314, 29.5946695, 29.490563, 29.3388885, 29.1382224, 28.8871413, 28.5842217, 28.2280401, 27.817173, 27.3501969, 26.8256883, 26.2422237, 25.5983796, 24.8927326, 24.1238592, 22.6290081, 20.7513299, 18.8419237, 16.9208274, 15.0080787, 13.1237157, 11.2877761, 9.52029788, 7.84131879, 6.27087674, 4.8290096, 3.53575522, 2.41115146, 1.4752362, 0.748047296, 0.249622608, 0, 0.515878903, 1.45247094, 2.78987875, 4.50820494, 6.58755215, 9.00802301, 11.7497201, 14.7927461, 18.1172037, 21.7031954, 25.5308238, 29.5801917, 33.8314016, 38.2645561, 42.8597579, 47.5971097, 52.1850244, 56.5647919, 61.3310507, 66.4056843, 71.7105761, 77.1676096, 82.6986681, 88.2256352, 93.6703941, 98.9548285, 104.000822, 108.730257, 113.065018, 116.873776, 120.151138, 122.834951, 124.863061, 127.409999, 129.79527, 131.008598, 130.616826, 128.186795, 123.28535, 115.479331, 106.40775, 92.9346211, 74.2213618, 49.4293903, 17.7201247, -17.0352918, -61.5625975, -117.877806, -187.99693, -273.935982, -375.39005, -499.532762, -648.209435, -824.792481, -1035.77212, -1286.23875, -1581.28276, -1916.94642, -2309.42462, -2784.78294, -3358.01933, -4044.01585, -4845.29124, -5777.89174, -6880.78511, -8182.41777, -9686.86329, -11416.2492, -13482.07, -15945.2693, -18699.8723, -21999.4204, -25962.6365, -30708.2434, -36354.964, -46138.0633, -56280.5864, -66122.1206, -75002.2532, -82260.5716, -85052.5531, -86357.9681, -86452.4512, -86281.3583, -85855.2256, -84905.5387, -83456.7865, -81533.4578, -79160.0414, -76361.026, -70936.0709, -66728.4952, -63281.2017, -60137.0933, -56839.0728, -52930.043, -47952.9068, -40327.0201, -32935.573},
        {-24.9710939, -24.9582359, -24.8269495, -24.5821665, -24.2288183, -23.7718366, -23.2161529, -22.5666989, -21.8284062, -21.0062063, -20.1050309, -19.1298115, -18.0854799, -16.9769674, -15.8092058, -14.5871267, -13.3156616, -11.9997422, -10.6443001, -9.25426674, -7.83457387, -6.39015305, -4.92593587, -3.44685394, -1.95783886, -0.463822216, 1.03026438, 2.51948935, 3.99892107, 5.46362795, 6.90867839, 8.3291408, 9.72008358, 11.0765751, 12.2097392, 13.2104855, 14.1719348, 15.0951247, 15.981093, 16.8308775, 17.6455159, 18.426046, 19.1735056, 19.8889324, 20.5733641, 21.2278386, 21.8533935, 22.4510667, 23.0218959, 23.5669189, 24.0871734, 24.6841906, 25.2563363, 25.8024698, 26.3214504, 26.8121372, 27.2733897, 27.7040669, 28.1030283, 28.469133, 28.8012403, 29.0982094, 29.3588997, 29.5821703, 29.7668806, 29.9118897, 30.016057, 30.0819467, 30.1120341, 30.0975372, 30.0370751, 29.9292668, 29.7727314, 29.566088, 29.3079556, 28.9969535, 28.6317005, 28.2108158, 27.7329186, 27.1966278, 26.6005625, 25.9433418, 25.2235849, 24.4399107, 22.9307645, 21.0387382, 19.1143584, 17.1776682, 15.2487104, 13.3475282, 11.4941644, 9.70866202, 8.01106415, 6.42141371, 4.9597537, 3.64612711, 2.50057692, 1.54314613, 0.79387772, 0.27281468, 0, 0.189624323, 0.855561066, 1.97433353, 3.52246504, 5.47647887, 7.81289835, 10.5082468, 13.5390475, 16.8818237, 20.5130988, 24.4093961, 28.5472389, 32.9031504, 37.4536541, 42.1752731, 47.0445308, 51.7705229, 56.2915686, 61.1890872, 66.3822962, 71.7904129, 77.3326546, 82.9282387, 88.4963827, 93.9563038, 99.2272195, 104.228347, 108.878904, 113.098107, 116.62513, 119.62525, 122.071699, 123.937708, 126.610126, 129.355373, 131.062909, 131.281878, 129.561425, 125.450697, 118.498839, 109.948233, 97.1394552, 79.2903653, 55.6188233, 25.3426894, -7.34398892, -49.4435861, -102.982289, -169.986285, -252.48176, -351.157184, -471.01145, -610.976206, -773.526524, -968.371963, -1201.97394, -1480.79389, -1808.02558, -2194.14072, -2653.17572, -3187.04331, -3823.15544, -4574.41831, -5460.71162, -6513.73128, -7759.74938, -9199.59358, -10852.9392, -12832.758, -15199.2478, -17866.6308, -21063.5387, -24900.6996, -29488.8415, -34938.6927, -44385.1709, -54158.7671, -63614.162, -72106.0361, -78989.0701, -81448.9807, -82455.3886, -82292.5304, -81956.9397, -81450.9023, -80479.9137, -79061.0995, -77211.5848, -74948.4953, -72288.9561, -68549.7057, -64868.1206, -61191.6557, -57467.7657, -53643.9055, -49667.53, -45486.094, -38878.6471, -32306.2357},
        {-26.8909579, -27.006274, -26.9851528, -26.8332202, -26.5561022, -26.1594244, -25.6488127, -25.0298929, -24.308291, -23.4896326, -22.5795437, -21.5836501, -20.5075775, -19.3569519, -18.137399, -16.8545447, -15.5140148, -14.1214352, -12.6824316, -11.20263, -9.68765609, -8.14313574, -6.57469479, -4.98795907, -3.3885544, -1.78210661, -0.174241545, 1.42941498, 3.02323713, 4.60159907, 6.15887498, 7.68943902, 9.18766537, 10.6479282, 11.852475, 12.906205, 13.9165672, 14.8848197, 15.8122205, 16.7000279, 17.5494997, 18.3618942, 19.1384694, 19.8804834, 20.5891943, 21.2658602, 21.9117392, 22.5280893, 23.1161687, 23.6772354, 24.2125476, 24.8361394, 25.4340481, 26.005087, 26.5480693, 27.0618085, 27.5451178, 27.9968106, 28.4157002, 28.8005999, 29.150323, 29.4636829, 29.7394929, 29.9765662, 30.1737164, 30.3297565, 30.4435001, 30.5159975, 30.5482456, 30.5344437, 30.4732601, 30.3633631, 30.2034209, 29.9921019, 29.7280743, 29.4100065, 29.0365668, 28.6064234, 28.1182447, 27.5706989, 26.9624543, 26.2921793, 25.5585421, 24.7602111, 23.2364586, 21.3295204, 19.3895576, 17.4366373, 15.4908264, 13.5721917, 11.7008004, 9.89671938, 8.18001555, 6.5707559, 5.08900739, 3.75483697, 2.58831162, 1.60949828, 0.838463928, 0.295275514, 0, -0.0632728937, 0.392084302, 1.33991323, 2.75405554, 4.60835288, 6.87664688, 9.5327792, 12.5505915, 15.9039254, 19.5666225, 23.5125245, 27.7154731, 32.1493099, 36.7878764, 41.6050145, 46.5745657, 51.4116827, 56.0532516, 61.0582438, 66.3450677, 71.8321314, 77.4378433, 83.0806116, 88.6788448, 94.1509511, 99.4153388, 104.390416, 108.994592, 113.146273, 116.480884, 119.303579, 121.617663, 123.426438, 126.3003, 129.459008, 131.704429, 132.569741, 131.58812, 128.292744, 122.21679, 114.180181, 102.037729, 85.0708567, 62.5609863, 33.7895401, 3.1919378, -36.392895, -86.9720352, -150.55256, -229.141546, -324.261243, -438.996124, -569.392783, -717.083328, -895.103616, -1111.36945, -1373.79664, -1693.22157, -2073.9757, -2517.20441, -3010.95635, -3595.7064, -4294.14822, -5130.17564, -6129.71601, -7317.76329, -8694.83955, -10279.5942, -12180.4067, -14454.9002, -17032.0447, -20117.9092, -23814.6503, -28224.425, -33449.39, -42501.7941, -51846.1791, -60857.1789, -68909.4272, -75377.5578, -77497.4371, -78208.3113, -77801.4288, -77325.3296, -76771.5008, -75815.8914, -74465.7656, -72728.3875, -70611.021, -68120.9304, -66093.8069, -62944.7324, -59030.5122, -54707.9519, -50333.8568, -46265.0324, -42858.2841, -37286.3054, -31398.085},
        {-28.8054375, -29.0584366, -29.1574142, -29.1086427, -28.9183948, -28.592943, -28.1385599, -27.561518, -26.8680899, -26.0645482, -25.1571653, -24.1522139, -23.0559665, -21.8746957, -20.614674, -19.2821739, -17.8834681, -16.4248291, -14.9125295, -13.3528418, -11.7520385, -10.1163922, -8.45217555, -6.765661, -5.06312115, -3.35082855, -1.63505575, 0.0779246944, 1.78184022, 3.47041827, 5.13738628, 6.77647171, 8.381402, 9.94590458, 11.2281261, 12.3445394, 13.4144492, 14.4392278, 15.4202478, 16.3588817, 17.2565019, 18.114481, 18.9341914, 19.7170057, 20.4642963, 21.1774358, 21.8577965, 22.5067511, 23.125672, 23.7159318, 24.2789028, 24.9344744, 25.5629331, 26.1630822, 26.7337249, 27.2736645, 27.7817044, 28.2566477, 28.6972979, 29.1024581, 29.4709316, 29.8015218, 30.0930319, 30.3442652, 30.554025, 30.7211146, 30.8443372, 30.9237906, 30.9595416, 30.9477237, 30.887056, 30.7762582, 30.6140494, 30.3991491, 30.1302766, 29.8061513, 29.4254926, 28.9870198, 28.4894523, 27.9315095, 27.3119106, 26.6293752, 25.8826225, 25.0703719, 23.5324564, 21.6108974, 19.6556355, 17.6867721, 15.7244088, 13.7886471, 11.8995885, 10.0773344, 8.34198644, 6.713646, 5.2124146, 3.85839376, 2.67168495, 1.67238967, 0.880609425, 0.316445703, 0, -0.17018646, 0.193247743, 1.06332156, 2.41305394, 4.21546384, 6.44357021, 9.07039201, 12.0689482, 15.4122577, 19.0733394, 23.0252124, 27.2408956, 31.6934079, 36.3557683, 41.2009957, 46.2021091, 51.0817416, 55.7771428, 60.8276806, 66.1527196, 71.6716248, 77.303761, 82.9684931, 88.5851859, 94.0732043, 99.3519132, 104.340677, 108.958862, 113.125831, 116.416185, 119.220468, 121.561474, 123.461996, 126.596927, 130.163013, 132.916438, 134.384369, 134.093969, 131.572404, 126.346839, 118.864507, 107.452484, 91.4579207, 70.2279692, 43.1097809, 14.6334828, -22.3324931, -69.7367393, -129.527848, -203.654412, -294.157602, -402.780281, -523.35972, -656.888393, -818.967309, -1018.63854, -1264.94417, -1575.03908, -1948.00871, -2374.08815, -2829.76815, -3366.02823, -4012.84148, -4797.24148, -5741.28102, -6869.29919, -8183.60804, -9702.83596, -11526.3769, -13707.9392, -16184.1821, -19143.8707, -22681.2304, -26890.4867, -31865.8648, -40488.4487, -49368.8907, -57904.7016, -65493.3921, -71532.4731, -73323.7839, -73757.621, -73129.6399, -72530.1456, -71940.2451, -71009.9414, -69736.8966, -68118.7728, -66153.2323, -63837.9372, -63216.8947, -60531.1754, -56465.5152, -51704.6502, -46933.3163, -42836.2494, -40098.1855, -35505.5243, -30218.292},
        {-30.6128386, -31.0163886, -31.249281, -31.3183498, -31.2304288, -30.992352, -30.6109534, -30.0930668, -29.4455263, -28.6751657, -27.788819, -26.7933201, -25.695503, -24.5022016, -23.2202498, -21.8564816, -20.4177309, -18.9108316, -17.3426177, -15.7199231, -14.0495818, -12.3384276, -10.5932945, -8.82101654, -7.02842753, -5.22236144, -3.40965221, -1.59713378, 0.20835992, 1.99999496, 3.77093739, 5.51435329, 7.22340871, 8.89126972, 10.2602975, 11.4543358, 12.600243, 13.6993244, 14.7528851, 15.7622303, 16.728665, 17.6534946, 18.5380241, 19.3835588, 20.1914037, 20.9628641, 21.6992451, 22.4018519, 23.0719897, 23.7109636, 24.3200788, 25.0129365, 25.6761115, 26.3084746, 26.9088966, 27.4762483, 28.0094004, 28.5072239, 28.9685895, 29.392368, 29.7774303, 30.122647, 30.4268892, 30.6890275, 30.9079327, 31.0824757, 31.2115273, 31.2955225, 31.3348588, 31.3251808, 31.2652581, 31.1538599, 30.9897558, 30.7717151, 30.4985072, 30.1689016, 29.7816676, 29.3355747, 28.8293922, 28.2618895, 27.6318362, 26.9380015, 26.1791548, 25.3540657, 23.8030125, 21.8678606, 19.8983899, 17.9147391, 15.9370469, 13.985452, 12.0800931, 10.2411089, 8.48863818, 6.84281957, 5.32379179, 3.95169354, 2.74666354, 1.72884048, 0.918363066, 0.335370005, 0, -0.0227282673, 0.455090383, 1.40892687, 2.81425213, 4.64653708, 6.88125264, 9.49386974, 12.4598593, 15.7546923, 19.3538395, 23.232772, 27.3669607, 31.7318764, 36.3029902, 41.0557729, 45.9656954, 50.7586597, 55.3744839, 60.3531403, 65.616621, 71.0869176, 76.686022, 82.335926, 87.9586216, 93.4761005, 98.8103546, 103.883376, 108.617156, 112.933686, 116.414821, 119.447408, 122.055479, 124.263069, 127.699206, 131.585424, 134.715263, 136.631255, 136.875938, 134.991844, 130.52151, 123.690629, 113.150659, 98.3104731, 78.5789455, 53.3649495, 27.0585388, -7.15788874, -51.1207546, -106.66648, -175.631488, -260.050929, -361.319602, -472.585216, -594.601012, -743.648204, -929.036262, -1160.07466, -1456.57261, -1814.95354, -2220.18658, -2643.4923, -3139.80014, -3741.3078, -4475.95528, -5364.46816, -6430.88358, -7680.56687, -9132.71433, -10874.9854, -12956.3131, -15311.3002, -18120.7988, -21473.3172, -25457.3641, -30161.4479, -38344.2507, -46756.5642, -54819.3534, -61953.5835, -67580.2196, -69078.0981, -69271.7321, -68457.2848, -67742.701, -67102.8137, -66174.2134, -64948.6452, -63417.8543, -61573.5856, -59407.5841, -59454.2816, -57063.4519, -53054.7814, -48247.9566, -43462.664, -39518.5902, -37235.4217, -33477.709, -28770.6047},
        {-31.968384, -32.5455792, -32.9379976, -33.1528356, -33.1972896, -33.0785561, -32.8038314, -32.3803119, -31.8151941, -31.1156744, -30.2889492, -29.3422149, -28.2826679, -27.1175047, -25.8539217, -24.4991152, -23.0602816, -21.5446175, -19.9593192, -18.3115831, -16.6086057, -14.8575833, -13.0657124, -11.2401893, -9.38821057, -7.51697253, -5.63367162, -3.74550424, -1.85966682, 0.0166442416, 1.87623252, 3.71190161, 5.51645508, 7.28269654, 8.75596749, 10.0580076, 11.3135308, 12.5233713, 13.6883633, 14.8093411, 15.8871388, 16.9225906, 17.9165307, 18.8697934, 19.7832128, 20.6576232, 21.4938588, 22.2927537, 23.0551422, 23.7818585, 24.4737367, 25.2107432, 25.9128646, 26.5792198, 27.2089275, 27.8011065, 28.3548756, 28.8693536, 29.3436592, 29.7769112, 30.1682284, 30.5167296, 30.8215335, 31.081759, 31.2965248, 31.4649496, 31.5861523, 31.6636431, 31.7008268, 31.687765, 31.623292, 31.5062418, 31.3354485, 31.1097463, 30.8279692, 30.4889514, 30.0915268, 29.6345297, 29.1167942, 28.5371543, 27.8944441, 27.1874977, 26.4151493, 25.576233, 24.0119815, 22.0638614, 20.081243, 18.0843008, 16.0932094, 14.1281433, 12.2092771, 10.3567851, 8.59084207, 6.93162241, 5.39930066, 4.01405136, 2.79604904, 1.76546822, 0.942483418, 0.347269171, 0, 0.832966725, 1.99995957, 3.48798145, 5.28403525, 7.37512389, 9.74825027, 12.3904173, 15.2886279, 18.4298849, 21.8011913, 25.3895499, 29.1819637, 33.1654356, 37.3269685, 41.6535652, 46.1322287, 50.4660163, 54.6013671, 59.1675421, 64.0906987, 69.2969944, 74.7125865, 80.2636324, 85.8762896, 91.4767154, 96.9910673, 102.345503, 107.466179, 112.279253, 116.544056, 120.414742, 123.867516, 126.878585, 130.606166, 134.435457, 137.436457, 139.230767, 139.43999, 137.685729, 133.589584, 127.65728, 118.357218, 105.138019, 87.4483059, 64.7367009, 40.708244, 9.49103466, -30.5304094, -80.9715707, -143.447932, -218.723961, -310.306282, -414.919204, -534.13905, -679.448339, -857.881913, -1076.47461, -1346.61464, -1669.99407, -2043.55585, -2452.18014, -2935.45025, -3514.01784, -4210.19939, -5049.13245, -6054.65963, -7235.75089, -8612.39299, -10259.1058, -12219.5623, -14403.483, -17009.101, -20126.4576, -23845.5935, -28256.55, -36054.7929, -44073.582, -51751.6021, -58527.5379, -63840.0743, -65144.3885, -65185.0111, -64250.7013, -63401.8598, -62621.8047, -61572.3369, -60252.8501, -58662.7383, -56801.3955, -54668.2154, -53247.6411, -50655.598, -47297.3751, -43578.2608, -39903.5439, -36678.5131, -34308.4571, -31007.8186, -27025.7054},
        {-33.3787552, -34.1388605, -34.7004958, -35.0711726, -35.2584026, -35.2696974, -35.1125687, -34.794528, -34.3230871, -33.7057574, -32.9500508, -32.0634787, -31.0535528, -29.9277848, -28.6936863, -27.3587689, -25.9305442, -24.4165239, -22.8242196, -21.161143, -19.4348055, -17.652719, -15.822395, -13.9513452, -12.0470811, -10.1171144, -8.16895677, -6.21011981, -4.24811514, -2.29045439, -0.344649202, 1.5817888, 3.48134799, 5.34651673, 6.93062754, 8.35077703, 9.7269363, 11.0593548, 12.3482819, 13.5939672, 14.7966599, 15.9566096, 17.0740658, 18.1492778, 19.182495, 20.173967, 21.1239432, 22.0326729, 22.9004057, 23.7273909, 24.5138781, 25.2976551, 26.0406192, 26.742206, 27.4018512, 28.0189904, 28.5930594, 29.1234937, 29.6097291, 30.0512013, 30.4473459, 30.7975986, 31.101395, 31.3581709, 31.567362, 31.7284038, 31.8407321, 31.9125993, 31.9520459, 31.940324, 31.8762977, 31.758831, 31.5867882, 31.3590333, 31.0744304, 30.7318437, 30.3301374, 29.8681754, 29.344822, 28.7589413, 28.1093974, 27.3950544, 26.6147765, 25.7674277, 24.1946732, 22.2378, 20.2459938, 18.2394464, 16.2383491, 14.2628935, 12.3332711, 10.4696733, 8.69229166, 7.02131769, 5.47694287, 4.0793587, 2.84875668, 1.8053283, 0.96926506, 0.360758462, 0, 1.64064626, 3.45640474, 5.44531989, 7.60543613, 9.93479792, 12.4314497, 15.0934359, 17.9188009, 20.9055893, 24.0518453, 27.3556136, 30.8149384, 34.4278644, 38.1924358, 42.1066971, 46.1686928, 50.067598, 53.7573329, 57.934575, 62.5311824, 67.4790133, 72.7099258, 78.1557782, 83.7484285, 89.4197348, 95.1015555, 100.725748, 106.224172, 111.528684, 116.578754, 121.295839, 125.609512, 129.44935, 133.515231, 137.351577, 140.297397, 142.051176, 142.311397, 140.776543, 137.145099, 132.182483, 124.195524, 112.673959, 97.1075225, 76.9859506, 55.2519291, 27.0790056, -8.90632194, -54.0775555, -109.808197, -175.442841, -256.863673, -354.720644, -471.509179, -613.492539, -785.242307, -991.330061, -1233.9991, -1520.67844, -1861.33705, -2255.75076, -2726.75046, -3282.56412, -3939.34757, -4726.69002, -5668.01452, -6775.31331, -8069.04589, -9611.58486, -11441.5944, -13444.4158, -15835.6243, -18706.9574, -22150.1526, -26256.9474, -33653.0365, -41262.7109, -48544.2188, -54955.8085, -59955.7282, -61087.2603, -61000.2627, -59973.6832, -59009.3235, -58100.0019, -56937.3474, -55529.2376, -53883.55, -52008.1625, -49910.9524, -46869.3858, -44052.1597, -41401.95, -38861.4324, -36373.2826, -33880.1763, -31324.7893, -28378.7479, -25070.0972},
        {-35.0428378, -35.9928596, -36.7303496, -37.2631351, -37.599044, -37.7459037, -37.7115418, -37.503786, -37.1304638, -36.5994027, -35.9184304, -35.0953745, -34.1380625, -33.054322, -31.8519807, -30.538866, -29.1228056, -27.6116271, -26.013158, -24.335226, -22.5856585, -20.7722833, -18.9029278, -16.9854197, -15.0275866, -13.0372559, -11.0222555, -8.99041267, -6.94955517, -4.90751054, -2.87210637, -0.851170237, 1.14747027, 3.11598757, 4.81727211, 6.36295994, 7.86744287, 9.33033396, 10.7512463, 12.1297929, 13.4655868, 14.7582411, 16.0073688, 17.2125831, 18.3734968, 19.4897233, 20.5608753, 21.5865661, 22.5664087, 23.5000161, 24.3870014, 25.2173012, 26.0004147, 26.7361648, 27.4243743, 28.0648661, 28.6574629, 29.2019876, 29.6982631, 30.1461121, 30.5453574, 30.8958219, 31.1973285, 31.4496999, 31.6527589, 31.8063285, 31.9102313, 31.9797048, 32.0296162, 32.0279971, 31.9736709, 31.8654613, 31.7021919, 31.4826861, 31.2057675, 30.8702598, 30.4749864, 30.018771, 29.5004371, 28.9188082, 28.272708, 27.5609599, 26.7823876, 25.9358146, 24.3620974, 22.4028825, 20.4077249, 18.3968002, 16.3902845, 14.4083534, 12.471183, 10.598949, 8.8118273, 7.12999378, 5.57362429, 4.16289468, 2.91798082, 1.85905857, 1.00630377, 0.379892296, 0, 2.1438152, 4.35861512, 6.64979862, 9.02276456, 11.4829118, 14.0356393, 16.6863458, 19.4404302, 22.3032914, 25.2803282, 28.3769396, 31.5985244, 34.9504814, 38.4382095, 42.0671077, 45.8425747, 49.4600043, 52.880509, 56.8106896, 61.1895153, 65.9559554, 71.0489791, 76.4075555, 81.9706538, 87.6772433, 93.4662932, 99.2767726, 105.047651, 110.717897, 116.350206, 121.714456, 126.712497, 131.246181, 135.724547, 139.795853, 142.973106, 145.006001, 145.644232, 144.637495, 141.735484, 137.75921, 131.061, 121.180942, 107.659127, 90.035643, 70.5455818, 45.3602809, 13.3460796, -26.6306828, -75.7036672, -131.938451, -203.286658, -293.658165, -406.313223, -543.18577, -706.805232, -899.701035, -1116.98442, -1370.28895, -1679.34075, -2055.29335, -2507.62973, -3035.04934, -3648.59536, -4380.22663, -5252.30134, -6277.92336, -7477.05159, -8903.57763, -10592.1988, -12412.5603, -14589.8043, -17215.7616, -20382.263, -24181.1392, -31148.8123, -38321.374, -45179.5075, -51203.8956, -55875.2215, -56835.742, -56630.1837, -55527.1263, -54473.5114, -53469.4006, -52232.2053, -50776.1195, -49115.3377, -47264.0542, -45236.4632, -40952.4057, -38019.8735, -36000.7935, -34457.0925, -32950.6971, -31043.5343, -28297.5307, -25657.0557, -22931.6341},
        {-37.544799, -38.6917925, -39.6092518, -40.3054535, -40.7886742, -41.0671904, -41.1492789, -41.0432161, -40.7572788, -40.2997436, -39.678887, -38.9029856, -37.9803162, -36.9191552, -35.7277793, -34.4144652, -32.9874894, -31.4551285, -29.8256592, -28.107358, -26.3085016, -24.4373666, -22.5022296, -20.5113673, -18.4730561, -16.3955728, -14.287194, -12.1561962, -10.010856, -7.85945022, -5.71025529, -3.57154788, -1.45160459, 0.641297967, 2.47394414, 4.15525813, 5.79686772, 7.397796, 8.95706606, 10.473701, 11.9467239, 13.3751579, 14.758026, 16.0943513, 17.383157, 18.6234661, 19.8143017, 20.9546869, 22.0436449, 23.0801986, 24.0633712, 24.9311272, 25.7448718, 26.5049519, 27.2117142, 27.8655053, 28.466672, 29.0155609, 29.512519, 29.9578927, 30.352029, 30.6952744, 30.9879758, 31.2304798, 31.4231332, 31.5662827, 31.6602749, 31.7323895, 31.8092594, 31.8356843, 31.8102651, 31.7316024, 31.5982972, 31.40895, 31.1621619, 30.8565334, 30.4906654, 30.0631587, 29.572614, 29.0176321, 28.3968139, 27.70876, 26.9520712, 26.1253484, 24.5646192, 22.6137073, 20.6244, 18.6167997, 16.6110091, 14.6271306, 12.6852669, 10.8055204, 9.00799379, 7.31278953, 5.74001018, 4.3097583, 3.04213643, 1.95724712, 1.07519291, 0.416076357, 0, 1.82434412, 3.76481572, 5.82406949, 8.00476014, 10.3095423, 12.7410708, 15.3020002, 17.9949852, 20.8226806, 23.787741, 26.8928211, 30.1405756, 33.5336592, 37.0747266, 40.7664325, 44.6114316, 48.3530271, 51.9568236, 56.0121039, 60.4659474, 65.2654335, 70.3576416, 75.6896511, 81.2085414, 86.8613918, 92.5952818, 98.3572908, 104.094498, 109.753983, 115.331265, 120.707222, 125.814402, 130.585353, 135.668212, 140.637733, 144.877843, 148.106412, 150.041311, 150.400409, 148.901578, 145.743435, 140.016788, 131.345448, 119.353225, 103.663931, 86.0944516, 63.5272533, 35.0378772, -0.298135078, -43.4052418, -91.6300767, -153.84782, -234.797507, -337.94763, -464.172657, -615.511722, -794.003955, -994.937979, -1228.63764, -1512.79099, -1856.98735, -2270.23148, -2752.13418, -3312.7571, -3979.87029, -4773.11537, -5701.68503, -6782.51191, -8070.84129, -9600.24667, -11248.8947, -13229.6744, -15629.954, -18537.1017, -22038.4856, -28545.8173, -35241.0357, -41630.4231, -47220.2617, -51516.8339, -52261.1283, -51901.5833, -50701.0355, -49590.5633, -48568.4911, -47354.7094, -45961.4644, -44401.0022, -42685.5689, -40827.4109, -36693.9715, -34026.0809, -32324.6053, -31090.4108, -29824.3637, -28027.3301, -25200.1763, -22925.6394, -20613.57},
        {-39.8659112, -41.2073808, -42.304196, -43.164997, -43.7984241, -44.2131173, -44.4177169, -44.4208632, -44.2311962, -43.8573562, -43.3079834, -42.5917179, -41.7172001, -40.69307, -39.5279679, -38.2305339, -36.8094084, -35.2732314, -33.6306431, -31.8902838, -30.0607937, -28.1508129, -26.1689816, -24.1239401, -22.0243286, -19.8787871, -17.695956, -15.4844755, -13.2529856, -11.0101267, -8.76453893, -6.52486246, -4.29973751, -2.09780426, -0.145522776, 1.66195822, 3.43210804, 5.16335516, 6.85412805, 8.50285518, 10.107965, 11.6678861, 13.1810468, 14.6458757, 16.0608013, 17.4242519, 18.7346561, 19.9904424, 21.1900392, 22.331875, 23.4143783, 24.3264539, 25.1784282, 25.9711039, 26.7052837, 27.3817705, 28.001367, 28.5648758, 29.0730997, 29.5268414, 29.9269038, 30.2740895, 30.5692012, 30.8130417, 31.0064137, 31.1501201, 31.2449634, 31.3288238, 31.4386914, 31.4988995, 31.5078471, 31.4639335, 31.3655577, 31.211119, 30.9990166, 30.7276495, 30.3954169, 30.0007181, 29.541952, 29.017518, 28.4258151, 27.7652425, 27.0341994, 26.2310849, 24.6888718, 22.7526688, 20.7755825, 18.7776064, 16.7787341, 14.798959, 12.8582747, 10.9766747, 9.17415261, 7.47070181, 5.8863159, 4.44098838, 3.15471278, 2.04748262, 1.13929142, 0.450132707, 0, 1.25258635, 2.71339561, 4.37920295, 6.24678354, 8.31291256, 10.5743652, 13.0279166, 15.6703419, 18.4984164, 21.5089151, 24.6986134, 28.0642863, 31.602709, 35.3106566, 39.1849045, 43.2222277, 47.2336372, 51.1893705, 55.5063837, 60.1416452, 65.0521236, 70.1947876, 75.5266057, 81.0045464, 86.5855783, 92.2266699, 97.88479, 103.516907, 109.079989, 114.485994, 119.753405, 124.852696, 129.754338, 135.375045, 141.212167, 146.510727, 150.957322, 154.238554, 156.041021, 156.051322, 153.900779, 149.348094, 142.090683, 131.825963, 118.251347, 102.862971, 83.1098508, 58.2397223, 27.5003207, -9.86061829, -50.9127105, -104.591884, -175.391973, -266.663282, -379.42141, -515.730194, -677.653468, -862.613568, -1078.37071, -1337.74858, -1645.83498, -2014.12749, -2445.44057, -2949.8363, -3549.65737, -4261.60949, -5090.26043, -6049.16902, -7196.39978, -8564.8423, -10050.3392, -11843.9113, -14025.8873, -16676.5961, -19876.3668, -25873.2663, -32037.9509, -37908.8149, -43024.2525, -46922.6582, -47474.9055, -46998.5901, -45746.7578, -44623.7373, -43624.7753, -42473.7508, -41179.4791, -39750.7754, -38196.4548, -36525.3325, -32943.3483, -30623.3821, -29114.4974, -27965.7574, -26726.2258, -24944.966, -22171.0413, -20199.4234, -18221.5026},
        {-41.5002906, -43.0134448, -44.2724432, -45.2860539, -46.0630445, -46.6121829, -46.9422368, -47.0619742, -46.9801627, -46.7055703, -46.2469646, -45.6131136, -44.812785, -43.8547466, -42.7477662, -41.5006117, -40.1220508, -38.6208514, -37.0057812, -35.2856081, -33.4690998, -31.5650242, -29.5821491, -27.5292423, -25.4150716, -23.2484047, -21.0380096, -18.792654, -16.5211057, -14.2321325, -11.9345022, -9.6369827, -7.34834171, -5.07734707, -3.03587394, -1.12686435, 0.749180843, 2.59008507, 4.39367174, 6.15776427, 7.88018609, 9.55876061, 11.1913113, 12.7756614, 14.3096346, 15.7910541, 17.2177434, 18.5875259, 19.898225, 21.1476642, 22.3336669, 23.3042852, 24.2103606, 25.0529632, 25.8331631, 26.5520307, 27.2106359, 27.810049, 28.3513401, 28.8355794, 29.2638371, 29.6371834, 29.9566883, 30.2234221, 30.438455, 30.602857, 30.7176984, 30.8266146, 30.9722191, 31.0677944, 31.1116513, 31.1021005, 31.037453, 30.9160195, 30.7361109, 30.496038, 30.1941117, 29.8286428, 29.397942, 28.9003203, 28.3340885, 27.6975573, 26.9890377, 26.2068404, 24.6853313, 22.7687589, 20.8093519, 18.8269673, 16.8414622, 14.8726939, 12.9405194, 11.0647957, 9.26538018, 7.56212979, 5.97490169, 4.52355301, 3.22794085, 2.10792236, 1.18335463, 0.474094808, 0, 0.560797503, 1.4460207, 2.64502809, 4.14717819, 5.94182951, 8.01834055, 10.3660698, 12.9743759, 15.8326171, 18.9301522, 22.2563395, 25.8005375, 29.5521049, 33.5004001, 37.6347815, 41.9446078, 46.3285114, 50.7628899, 55.4423155, 60.3367054, 65.4159768, 70.6500469, 76.0088331, 81.4622523, 86.980222, 92.5326593, 98.0894813, 103.620605, 109.095949, 114.407117, 119.631053, 124.761168, 129.790872, 135.698093, 141.983982, 147.887155, 153.105668, 157.337576, 160.280936, 161.633803, 160.915343, 158.056947, 152.791671, 144.85257, 133.972701, 121.363478, 104.910009, 83.9757603, 57.9241973, 26.1187867, -8.83813554, -54.7805915, -114.943735, -192.172166, -288.513101, -406.371768, -548.153391, -713.928581, -908.967775, -1141.08826, -1411.03782, -1732.77775, -2111.97135, -2558.74777, -3090.80337, -3722.36104, -4454.5627, -5298.23987, -6312.18765, -7528.07025, -8865.75262, -10485.8981, -12458.8705, -14855.0339, -17744.7522, -23155.8211, -28711.8147, -33993.7391, -38582.6002, -42059.404, -42481.6541, -41969.5271, -40756.9585, -39684.9809, -38747.2205, -37684.3467, -36502.6336, -35208.3556, -33807.7866, -32307.201, -29433.0784, -27446.514, -26018.6529, -24820.6401, -23523.6205, -21798.7392, -19317.1412, -17526.9132, -15831.9721},
        {-41.8758817, -43.5079972, -44.8869726, -46.0212727, -46.9193623, -47.5897062, -48.0407692, -48.2810163, -48.3189121, -48.1629215, -47.8215094, -47.3031405, -46.6162797, -45.7693918, -44.7709416, -43.6293939, -42.3532136, -40.9508655, -39.4308143, -37.801525, -36.0714623, -34.249091, -32.342876, -30.3612821, -28.3127742, -26.2058169, -24.0488753, -21.850414, -19.6188978, -17.3627918, -15.0905605, -12.8106689, -10.5315818, -8.26176395, -6.18406645, -4.216093, -2.27335137, -0.3586472, 1.52521385, 3.37542613, 5.18918399, 6.96368177, 8.69611382, 10.3836745, 12.0235581, 13.6129591, 15.1490717, 16.6290903, 18.0502093, 19.4096229, 20.7045257, 21.7558555, 22.740884, 23.6606264, 24.5160979, 25.3083137, 26.0382891, 26.7070393, 27.3155796, 27.8649252, 28.3560913, 28.7900932, 29.1679461, 29.4906653, 29.7592659, 29.9747633, 30.1381726, 30.2903281, 30.4711082, 30.5994042, 30.6736501, 30.6922797, 30.6537271, 30.5564261, 30.3988107, 30.1793148, 29.8963723, 29.5484172, 29.1338834, 28.6512048, 28.0988153, 27.475149, 26.7786396, 26.0077212, 24.5053767, 22.611552, 20.6740711, 18.7126386, 16.7469591, 14.796737, 12.8816771, 11.0214839, 9.23586189, 7.54451573, 5.96714997, 4.52346918, 3.23317795, 2.11598084, 1.19158243, 0.479687289, 0, -0.133909215, 0.176985607, 0.914513379, 2.06050302, 3.59678344, 5.50518356, 7.7675323, 10.3656586, 13.2813913, 16.4965593, 19.9929917, 23.7525172, 27.7569649, 31.9881635, 36.4279422, 41.0581296, 45.8886283, 50.9052776, 56.026407, 61.2398611, 66.5334846, 71.8951222, 77.3126184, 82.773818, 88.2665656, 93.7787059, 99.2980835, 104.812543, 110.309929, 115.819651, 121.272748, 126.644597, 131.910572, 137.621657, 143.465424, 148.975948, 153.955921, 158.208036, 161.534985, 163.73946, 165.169039, 164.915696, 162.675516, 158.144583, 151.018982, 142.178429, 129.839471, 113.402285, 92.2670472, 65.833936, 35.5580855, -3.72689832, -52.9752685, -114.043981, -190.633012, -285.614858, -401.862018, -541.063519, -707.544298, -906.920348, -1139.20039, -1417.45119, -1746.85571, -2136.66486, -2603.02253, -3158.9086, -3805.44283, -4552.53026, -5453.56313, -6537.3349, -7751.65607, -9218.76189, -10995.4783, -13138.6314, -15705.047, -20421.6117, -25257.5757, -29852.25, -33844.9453, -36874.9726, -37275.7703, -36863.1032, -35835.9418, -34904.2841, -34064.8189, -33101.3067, -32021.0826, -30831.4818, -29539.8392, -28153.4901, -25887.068, -24109.9059, -22656.3934, -21360.92, -20057.8751, -18581.6484, -16766.6294, -14961.7212, -13528.2089},
        {-41.7005097, -43.5546254, -45.1496404, -46.4939415, -47.5959154, -48.4639491, -49.1064292, -49.5317425, -49.7482759, -49.7644161, -49.58855, -49.2290642, -48.6943457, -47.9927811, -47.1327573, -46.1226611, -44.9708792, -43.6857984, -42.2758056, -40.7492875, -39.1146309, -37.3802226, -35.5544494, -33.6456981, -31.6623554, -29.6128082, -27.5054432, -25.3486473, -23.1508072, -20.9203096, -18.6655415, -16.3948896, -14.1167407, -11.8394815, -9.72756622, -7.70806534, -5.70617345, -3.72521264, -1.76850503, 0.160627279, 2.0588622, 3.92287761, 5.74935143, 7.53496154, 9.27638585, 10.9703022, 12.6133886, 14.2023229, 15.733783, 17.2044467, 18.610992, 19.7620561, 20.846434, 21.8648798, 22.8181479, 23.7069924, 24.5321677, 25.294428, 25.9945275, 26.6332205, 27.2112612, 27.7294039, 28.1884028, 28.5890123, 28.9319865, 29.2180796, 29.4480461, 29.6534838, 29.8652502, 30.0204867, 30.117948, 30.1563889, 30.1345641, 30.0512285, 29.9051367, 29.6950436, 29.419704, 29.0778726, 28.6683042, 28.1897536, 27.6409755, 27.0207247, 26.327756, 25.5608242, 24.0797292, 22.2161427, 20.3094163, 18.3790401, 16.444504, 14.5252983, 12.6409129, 10.810838, 9.05456374, 7.39158008, 5.84137716, 4.42344506, 3.15727388, 2.06235369, 1.15817459, 0.464226667, 0, -0.416513462, -0.340776918, 0.206235431, 1.20354938, 2.63019074, 4.4651853, 6.68755887, 9.27633723, 12.2105462, 15.4692116, 19.0313591, 22.8760147, 26.9822041, 31.328953, 35.8952874, 40.6602329, 45.7026403, 51.0157958, 56.3529313, 61.7144635, 67.1008093, 72.5123855, 77.9496089, 83.4128964, 88.9026647, 94.4193307, 99.9633111, 105.535023, 111.134883, 116.887527, 122.623605, 128.30627, 133.898671, 139.605421, 145.268941, 150.667261, 155.691095, 160.231156, 164.178157, 167.422811, 170.913167, 173.160804, 173.849565, 172.663294, 169.285833, 164.559193, 156.719508, 145.161078, 129.278204, 108.465187, 82.9803932, 50.2979793, 11.4223901, -34.5852069, -92.6311678, -165.840512, -257.338258, -369.82608, -506.948558, -672.812096, -867.150283, -1101.80696, -1381.51683, -1714.56646, -2115.26083, -2595.14237, -3155.2559, -3804.42224, -4589.95343, -5537.38869, -6616.46072, -7917.10089, -9484.16607, -11362.5131, -13596.9989, -17631.3901, -21761.9064, -25683.6771, -29091.8316, -31681.4995, -32049.02, -31720.8394, -30862.997, -30062.4817, -29318.7071, -28455.6912, -27481.6174, -26404.669, -25233.0293, -23974.8818, -22188.3316, -20611.6185, -19197.9868, -17900.6808, -16672.9447, -15468.023, -14239.1598, -12458.4682, -11264.5219},
        {-41.4908839, -43.7699152, -45.7656025, -47.4869014, -48.9427673, -50.1421558, -51.0940222, -51.8073222, -52.2910113, -52.5540448, -52.6053784, -52.4539674, -52.1087675, -51.5787341, -50.8728226, -49.9999887, -48.9691877, -47.7893753, -46.4695068, -45.0185378, -43.4454237, -41.7591201, -39.9685825, -38.0827663, -36.1106271, -34.0611204, -31.9432016, -29.7658262, -27.5379497, -25.2685277, -22.9665156, -20.640869, -18.3005432, -15.9544939, -13.7760258, -11.6884629, -9.61315382, -7.5537206, -5.51378528, -3.4969699, -1.50689652, 0.452812831, 2.3785361, 4.26665125, 6.11353623, 7.915569, 9.66912752, 11.3705897, 13.0163336, 14.6027371, 16.1261782, 17.4044296, 18.6163463, 19.762178, 20.8421746, 21.8565858, 22.8056613, 23.689651, 24.5088045, 25.2633717, 25.9536023, 26.5797461, 27.1420529, 27.6407724, 28.0761543, 28.4484486, 28.7579048, 29.0170151, 29.2379775, 29.3957931, 29.489918, 29.5198084, 29.4849204, 29.3847101, 29.2186337, 28.9861475, 28.6867074, 28.3197697, 27.8847905, 27.381226, 26.8085323, 26.1661655, 25.4535819, 24.6702375, 23.2092885, 21.3860938, 19.5232995, 17.6400373, 15.7554389, 13.8886363, 12.058761, 10.2849448, 8.5863196, 6.982017, 5.49116879, 4.13290674, 2.92636258, 1.89066808, 1.04495498, 0.408355037, 0, 0.292257986, 0.948485711, 1.95604466, 3.30229633, 4.97460219, 6.96032375, 9.24682248, 11.8214599, 14.6715974, 17.7845966, 21.147819, 24.7486259, 28.574379, 32.6124396, 36.8501694, 41.2749297, 45.948803, 50.8698252, 55.8563461, 60.911739, 66.0393768, 71.2426327, 76.5248798, 81.8894913, 87.3398402, 92.8792997, 98.5112429, 104.239043, 110.066073, 116.061264, 122.138394, 128.281167, 134.473289, 141.068183, 147.864696, 154.563083, 161.036134, 167.15664, 172.79739, 177.831175, 182.538844, 186.260936, 188.790404, 189.9202, 189.443278, 188.99758, 186.069821, 179.991705, 170.094938, 155.711224, 136.403181, 110.989375, 80.7002864, 44.941759, -0.605664054, -58.5888555, -131.654688, -222.354838, -333.45301, -467.816763, -623.855868, -813.623093, -1044.23621, -1323.46103, -1660.16146, -2062.69735, -2528.65955, -3063.61592, -3712.62084, -4498.39288, -5394.80124, -6480.92917, -7796.40463, -9380.85562, -11273.9101, -14743.813, -18294.4212, -21658.2082, -24567.648, -26755.2141, -26979.9077, -26596.657, -25755.1382, -24984.9158, -24285.2936, -23497.2889, -22628.12, -21685.005, -20675.1624, -19605.8102, -17866.9667, -16512.4312, -15428.6412, -14502.0338, -13619.0465, -12666.1164, -11529.6809, -9977.72893, -9034.65033},
        {-40.3788505, -43.3029335, -45.9021266, -48.1865514, -50.1663293, -51.8515818, -53.2524304, -54.3789965, -55.2414017, -55.8497675, -56.2142152, -56.3448665, -56.2518428, -55.9452655, -55.4352563, -54.7319365, -53.8454276, -52.7858511, -51.5633286, -50.1879814, -48.6699312, -47.0192993, -45.2462072, -43.3607765, -41.3731287, -39.2933851, -37.1316673, -34.8980968, -32.6027951, -30.2558836, -27.8674838, -25.4477173, -23.0067055, -20.5545699, -18.2986687, -16.1463564, -14.0026169, -11.871171, -9.75573947, -7.66004303, -5.58780248, -3.54273858, -1.52857212, 0.450976125, 2.39218539, 4.29133491, 6.14470389, 7.94857157, 9.69921718, 11.3929199, 13.0259591, 14.4376144, 15.7845679, 17.0665024, 18.2831005, 19.4340449, 20.5190182, 21.5377031, 22.4897823, 23.3749384, 24.1928542, 24.9432122, 25.6256952, 26.2399858, 26.7857667, 27.2627205, 27.67053, 28.0026448, 28.2526646, 28.4331266, 28.5441178, 28.5857247, 28.5580342, 28.4611329, 28.2951075, 28.0600448, 27.7560315, 27.3831544, 26.9415, 26.4311552, 25.8522067, 25.2047411, 24.4888453, 23.7046059, 22.2755616, 20.5022001, 18.6919387, 16.8635741, 15.0359033, 13.2277229, 11.4578299, 9.74502093, 8.10809294, 6.5658427, 5.13706702, 3.8405627, 2.69512656, 1.71955541, 0.932646061, 0.35319532, 0, 1.37256172, 2.90791077, 4.60685122, 6.47018717, 8.49872271, 10.6932619, 13.0546089, 15.5835677, 18.2809424, 21.1475372, 24.1841561, 27.3916032, 30.7706825, 34.3221983, 38.0469544, 41.9457552, 45.9591206, 50.079243, 54.4432826, 59.0391257, 63.8546584, 68.8777668, 74.096337, 79.4982551, 85.0714071, 90.8036794, 96.6829578, 102.697129, 108.834078, 114.96511, 121.237439, 127.673927, 134.297433, 141.715278, 149.658093, 157.700653, 165.69048, 173.475098, 180.902029, 187.818796, 193.817715, 199.079186, 203.500666, 206.979609, 209.41347, 213.050237, 214.849197, 214.120173, 210.172985, 202.317455, 189.710602, 172.001809, 150.094723, 123.566807, 89.2878966, 45.3434972, -10.1808861, -78.7231539, -162.782713, -265.378889, -385.453896, -532.925444, -714.456541, -936.451103, -1204.87404, -1525.89176, -1896.06528, -2319.07071, -2833.80406, -3459.23897, -4164.31928, -5025.74534, -6079.80882, -7362.80139, -8911.0147, -11841.5041, -14839.6067, -17671.4232, -20103.0542, -21900.6007, -21965.2894, -21504.6775, -20655.5259, -19906.7703, -19255.7692, -18553.1413, -17803.582, -17011.7868, -16182.4508, -15320.2695, -13663.4979, -12554.1477, -11809.2584, -11245.8696, -10681.0206, -9931.75108, -8815.10039, -7624.26995, -6887.06924},
        {-36.7036308, -40.8928298, -44.665527, -48.0347566, -51.013553, -53.6149505, -55.8519833, -57.7376857, -59.285092, -60.5072365, -61.4171535, -62.0278773, -62.3524421, -62.4038822, -62.195232, -61.7395257, -61.0497976, -60.1390819, -59.0204131, -57.7068252, -56.2113528, -54.5470299, -52.726891, -50.7639702, -48.6713019, -46.4619204, -44.14886, -41.7451549, -39.2638394, -36.7179478, -34.1205144, -31.4845735, -28.8231594, -26.1493063, -23.7781024, -21.5652017, -19.3626279, -17.173613, -15.0013891, -12.8491882, -10.7202424, -8.61778381, -6.54504441, -4.5052563, -2.50165151, -0.537462126, 1.38407981, 3.25974224, 5.0862931, 6.86050032, 8.57913187, 10.1318017, 11.6247547, 13.0570816, 14.4278732, 15.7362203, 16.9812139, 18.1619447, 19.2775036, 20.3269813, 21.3094688, 22.2240568, 23.0698361, 23.8458977, 24.5513323, 25.1852307, 25.7466839, 26.218921, 26.5855523, 26.8783817, 27.0975284, 27.2431113, 27.3152496, 27.3140623, 27.2396685, 27.0921873, 26.8717376, 26.5784386, 26.2124094, 25.773769, 25.2626364, 24.6791308, 24.0233712, 23.2954767, 21.9080645, 20.1673621, 18.3866578, 16.5851364, 14.7819824, 12.9963805, 11.2475153, 9.55457146, 7.93673356, 6.41318625, 5.00311414, 3.72570187, 2.60013407, 1.64559535, 0.881270352, 0.326343692, 0, 3.84055686, 7.37729274, 10.6438042, 13.6736878, 16.5005402, 19.1579578, 21.6795372, 24.098875, 26.4495678, 28.7652121, 31.0794045, 33.4257415, 35.8378196, 38.3492356, 40.9935858, 43.804467, 45.9279075, 47.2707085, 49.874061, 53.5813684, 58.236034, 63.6814613, 69.7610537, 76.3182146, 83.1963474, 90.2388555, 97.2891424, 104.190611, 110.786666, 115.7803, 120.573477, 125.351723, 130.300562, 136.596932, 143.930654, 151.727172, 159.874591, 168.261013, 176.774543, 185.303283, 194.148216, 202.658908, 210.64268, 217.906855, 224.258754, 230.226109, 234.715732, 237.354844, 237.770664, 235.590414, 230.706903, 222.157152, 209.623715, 192.582572, 170.087957, 141.383459, 105.712666, 65.9855343, 16.9460217, -50.6615864, -145.971261, -264.357897, -395.688312, -540.614012, -718.060558, -938.565421, -1212.75817, -1551.20014, -1964.11501, -2461.91744, -2981.76874, -3618.60671, -4408.20294, -5386.32905, -6588.75664, -9014.2552, -11494.849, -13825.5601, -15801.4108, -17217.4231, -17087.5456, -16508.5898, -15610.3236, -14880.5579, -14307.3354, -13729.5143, -13142.5963, -12542.0833, -11923.4773, -11282.2801, -9871.80228, -8971.63435, -8406.4765, -8001.02897, -7579.99198, -6968.06578, -5989.95059, -5615.29684, -5002.94136},
        {-40.407689, -44.9437335, -49.050512, -52.7410199, -56.0282524, -58.9252046, -61.4448718, -63.6002492, -65.4043319, -66.8701154, -68.0105946, -68.8387649, -69.3676215, -69.6101596, -69.5793744, -69.2882612, -68.749815, -67.9770313, -66.9829051, -65.7804316, -64.3826062, -62.8024241, -61.0528803, -59.1469703, -57.0976891, -54.9180319, -52.6209941, -50.2195709, -47.7267573, -45.1555487, -42.5189403, -39.8299273, -37.1015049, -34.3466683, -31.8445609, -29.4695057, -27.092627, -24.7182829, -22.3508317, -19.9946315, -17.6540405, -15.333417, -13.037119, -10.7695049, -8.53493282, -6.33776093, -4.18234745, -2.07305058, -0.0142285163, 1.98976055, 3.93455842, 5.64651325, 7.29423044, 8.87702175, 10.3941989, 11.8450738, 13.228958, 14.5451633, 15.7930016, 16.9717845, 18.0808239, 19.1194314, 20.0869189, 20.9825981, 21.8057807, 22.5557785, 23.2319033, 23.8246425, 24.3246953, 24.7495731, 25.0991597, 25.3733388, 25.5719943, 25.69501, 25.7422697, 25.7136572, 25.6090563, 25.4283507, 25.1714244, 24.838161, 24.4284445, 23.9421586, 23.3791871, 22.7394138, 21.4332859, 19.7678036, 18.0570515, 16.3200418, 14.5757866, 12.8432981, 11.1415884, 9.48966963, 7.90655401, 6.41125367, 5.02278077, 3.76014745, 2.64236587, 1.68844818, 0.917406538, 0.348253093, 0, 2.35229248, 4.66060476, 6.94165429, 9.21215853, 11.4888349, 13.788401, 16.1275741, 18.5230718, 20.9916114, 23.5499106, 26.2146866, 29.0026571, 31.9305393, 35.0150509, 38.2729092, 41.7208317, 44.8911639, 47.7314272, 51.3539424, 55.671633, 60.5974225, 66.0442342, 71.9249918, 78.1526185, 84.640038, 91.3001735, 98.0459486, 104.790287, 111.446111, 117.269937, 123.07178, 128.961485, 135.048898, 141.92606, 149.46172, 157.396041, 165.694212, 174.321422, 183.242857, 192.423707, 201.957351, 211.641771, 221.417072, 231.223361, 241.000746, 254.105442, 266.207419, 276.392758, 283.747537, 287.357835, 283.736244, 277.687705, 274.634077, 275.890443, 274.387228, 266.819397, 249.881915, 212.961792, 161.722849, 109.801225, 72.8456261, 31.9917151, -22.102497, -97.1872816, -198.315833, -331.779349, -503.188784, -718.658975, -986.798983, -1314.80699, -1664.1007, -2094.70194, -2629.65599, -3292.00813, -4104.80364, -5747.84897, -7423.23793, -8989.82553, -10306.4667, -11232.0166, -11078.9507, -10616.7563, -9938.45085, -9400.50795, -8991.66045, -8590.86773, -8192.35128, -7790.33256, -7379.03303, -6952.67417, -6555.02574, -6074.7786, -5537.36238, -4968.20672, -4392.74124, -3836.39558, -3324.59936, -3132.0603, -2486.28916},
        {-33.1017163, -38.5287178, -43.4866407, -47.989349, -52.0507067, -55.6845776, -58.9048259, -61.7253153, -64.1599099, -66.2224735, -67.9268702, -69.2869638, -70.3166183, -71.0296977, -71.4400658, -71.5615867, -71.4081243, -70.9935424, -70.3317052, -69.4364764, -68.3217201, -67.0013001, -65.4890805, -63.7989251, -61.944698, -59.940263, -57.7994841, -55.5362252, -53.1643504, -50.6977234, -48.1502083, -45.535669, -42.8679695, -40.1609736, -37.7087341, -35.379127, -33.036661, -30.6861074, -28.3322374, -25.9798224, -23.6336335, -21.2984422, -18.9790197, -16.6801372, -14.406566, -12.1630775, -9.95444288, -7.78543345, -5.6608205, -3.5853753, -1.56386913, 0.228615903, 1.96131134, 3.63313791, 5.24301632, 6.78986731, 8.27261158, 9.69016988, 11.0414629, 12.3254114, 13.5409361, 14.6869577, 15.7623969, 16.7661745, 17.6972111, 18.5544275, 19.3367445, 20.0219721, 20.5884274, 21.07857, 21.4926892, 21.8310743, 22.0940145, 22.2817989, 22.394717, 22.4330578, 22.3971106, 22.2871647, 22.1035093, 21.8464336, 21.5162269, 21.1131784, 20.6375774, 20.089713, 18.9672098, 17.5322623, 16.0531065, 14.5463442, 13.028577, 11.5164066, 10.0264347, 8.57526299, 7.17949306, 5.85572663, 4.62056536, 3.49061093, 2.48246502, 1.6127293, 0.898005442, 0.354895117, 0, 1.12328851, 2.35404117, 3.69891812, 5.1645795, 6.75768546, 8.48489612, 10.3528716, 12.3682721, 14.5377578, 16.8679887, 19.365625, 22.0373269, 24.8897544, 27.9295678, 31.1634271, 34.5979926, 38.3092115, 42.3136421, 46.4612236, 50.7734634, 55.2718688, 59.9779473, 64.9132062, 70.099153, 75.557295, 81.3091397, 87.3761943, 93.7799664, 100.541963, 107.804505, 115.423989, 123.385677, 131.674834, 140.222492, 149.041031, 158.157289, 167.5728, 177.289096, 187.30771, 197.630174, 208.149823, 219.009318, 230.231359, 241.83865, 253.85389, 268.643082, 283.299802, 297.260929, 309.963338, 320.843907, 327.678957, 333.59549, 341.593492, 352.489776, 362.643835, 370.416408, 374.168233, 369.449223, 358.069767, 344.906613, 336.084814, 322.964418, 300.9645, 266.201149, 215.9715, 147.030566, 57.7965736, -54.5442618, -198.904156, -380.745705, -583.763121, -838.536369, -1158.06326, -1555.34159, -2043.36919, -3019.09627, -4015.58012, -4949.83044, -5738.85692, -6299.66928, -6233.4918, -5983.6432, -5602.46975, -5284.7288, -5026.39561, -4764.1075, -4496.80665, -4223.4352, -3942.93533, -3654.24919, -3853.54193, -3660.26545, -3207.01769, -2626.39659, -2051.00009, -1613.42614, -1446.27266, -1158.55761, -701.310287},
        {-31.0761186, -37.5574255, -43.5386235, -49.0334671, -54.055711, -58.6191099, -62.7374186, -66.4243916, -69.6937837, -72.5593494, -75.0348436, -77.1340208, -78.8706358, -80.2584432, -81.3111976, -82.0426539, -82.4665665, -82.5966903, -82.4467798, -82.0305898, -81.361875, -80.4543899, -79.3218893, -77.9781279, -76.4368603, -74.7118412, -72.8168253, -70.7655672, -68.5718216, -66.2493432, -63.8118867, -61.2732068, -58.6470581, -55.9471952, -53.3928302, -50.8906676, -48.3449425, -45.762742, -43.1511533, -40.5172636, -37.8681601, -35.2109299, -32.5526604, -29.9004386, -27.2613518, -24.6424871, -22.0509317, -19.4937729, -16.9780977, -14.5109935, -12.0995474, -10.0293323, -8.02291273, -6.08133874, -4.20566051, -2.3969282, -0.656191973, 1.015498, 2.61709157, 4.14753857, 5.60578883, 6.9907922, 8.30149851, 9.5368576, 10.6958193, 11.7773335, 12.78035, 13.6794237, 14.4496955, 15.1404276, 15.7521514, 16.285398, 16.7406988, 17.118585, 17.4195878, 17.6442386, 17.7930684, 17.8666088, 17.8653907, 17.7899456, 17.6408047, 17.4184993, 17.1235605, 16.7565197, 15.8922654, 14.753103, 13.5664538, 12.3466621, 11.1080721, 9.86502781, 8.63187354, 7.42295343, 6.25261168, 5.13519246, 4.08503995, 3.11649834, 2.2439118, 1.48162451, 0.843980662, 0.345324432, 0, 0.973145912, 2.00120636, 3.09447464, 4.26324406, 5.51780793, 6.86845954, 8.3254922, 9.89919921, 11.5998739, 13.4378095, 15.4232994, 17.5666369, 19.8781152, 22.3680277, 25.0466677, 27.9243285, 30.856879, 33.832552, 37.2109339, 40.9692269, 45.0846336, 49.5343563, 54.2955973, 59.3455591, 64.661444, 70.2204545, 75.9997928, 81.9766613, 88.1282626, 94.1939945, 100.476059, 107.022999, 113.883359, 121.119076, 128.771998, 136.880399, 145.488804, 154.641738, 164.383728, 174.759299, 185.965657, 197.848179, 210.421516, 223.70032, 237.699246, 253.090845, 269.067395, 285.479075, 302.176062, 319.008533, 335.27129, 352.048683, 372.611726, 397.519802, 421.796053, 442.949282, 458.488289, 454.783909, 443.014267, 446.507996, 493.637637, 553.962724, 602.586701, 628.979543, 646.95451, 659.150895, 671.178427, 686.446511, 697.472954, 702.936466, 710.783578, 708.999363, 693.054441, 658.419431, 600.564953, 413.231207, 219.051836, 38.9300675, -106.230876, -195.527768, -125.323877, -12.605158, 128.074396, 235.636027, 313.941795, 386.239078, 455.42067, 524.379366, 596.007958, 673.199241, 1419.05307, 1711.85231, 1688.78909, 1487.05554, 1243.8438, 1096.34601, 1181.75429, 1199.20829, 1285.61877}
    };
    
    /* YPtag: chiAmed tables below */
    
    /* Stuff for interpolating the NQC data */
    gsl_spline    *spline = NULL;
    gsl_interp_accel *acc = NULL;
    /* Interpolate the spin NQC data in 2-D parameter space -- spin and mass ratio */
    /* First, interpolate in spin dimension for all mass ratios */
    spline = gsl_spline_alloc( gsl_interp_cspline, adim );
    acc    = gsl_interp_accel_alloc();
    for (i = 0; i < qdim; i++)
    {
        gsl_spline_init( spline, alist, a3stab[i], adim );
        gsl_interp_accel_reset( acc );
        a3slist[i] = gsl_spline_eval( spline, chiA, acc );
        gsl_spline_init( spline, alist, a4tab[i], adim );
        gsl_interp_accel_reset( acc );
        a4list[i] = gsl_spline_eval( spline, chiA, acc );
        gsl_spline_init( spline, alist, a5tab[i], adim );
        gsl_interp_accel_reset( acc );
        a5list[i] = gsl_spline_eval( spline, chiA, acc );
        gsl_spline_init( spline, alist, b3tab[i], adim );
        gsl_interp_accel_reset( acc );
        b3list[i] = gsl_spline_eval( spline, chiA, acc );
        gsl_spline_init( spline, alist, b4tab[i], adim );
        gsl_interp_accel_reset( acc );
        b4list[i] = gsl_spline_eval( spline, chiA, acc );
        //  printf("%.15f\n",a3slist[i]);
    }
    //printf("%.15f\n",a);
    gsl_spline_free(spline);
    gsl_interp_accel_free(acc);
    /* Second, interpolate in mass ratio dimension */
    spline = gsl_spline_alloc( gsl_interp_cspline, qdim );
    acc    = gsl_interp_accel_alloc();
    gsl_spline_init( spline, etalist, a3slist, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->a3S = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, a4list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->a4 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, a5list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->a5 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, b3list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->b3 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, b4list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->b4 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_free(spline);
    gsl_interp_accel_free(acc);
    
    return XLAL_SUCCESS;
    
}



UNUSED static int XLALSimIMRGetEOBCalibratedSpinNQCv2chiAmin( EOBNonQCCoeffs * restrict coeffs,                                                             INT4 UNUSED l,                                                             INT4 UNUSED m,                                                             REAL8 eta,                                                             REAL8 a )
{
    const unsigned int   qdim = 98;
    const unsigned int   adim = 200;
    UINT4 i;
    /* REAL8 eta2 = eta*eta;*/
    REAL8 a3slist[qdim], a4list[qdim], a5list[qdim];
    REAL8 b3list[qdim], b4list[qdim];
    REAL8 chiA;
    
    chiA = a/(1.0-2.0*eta);
    if (chiA < -1.){
        chiA = -1.;
    }
    if (chiA > 0.99){
        chiA = 0.99;
    }
    
    memset( coeffs, 0, sizeof( *coeffs ) );
    const double alist[200] = {-1., -0.99, -0.98, -0.97, -0.96, -0.95, -0.94, -0.93, -0.92, -0.91, -0.9, -0.89, -0.88, -0.87, -0.86, -0.85, -0.84, -0.83, -0.82, -0.81, -0.8, -0.79, -0.78, -0.77, -0.76, -0.75, -0.74, -0.73, -0.72, -0.71, -0.7, -0.69, -0.68, -0.67, -0.66, -0.65, -0.64, -0.63, -0.62, -0.61, -0.6, -0.59, -0.58, -0.57, -0.56, -0.55, -0.54, -0.53, -0.52, -0.51, -0.5, -0.49, -0.48, -0.47, -0.46, -0.45, -0.44, -0.43, -0.42, -0.41, -0.4, -0.39, -0.38, -0.37, -0.36, -0.35, -0.34, -0.33, -0.32, -0.31, -0.3, -0.29, -0.28, -0.27, -0.26, -0.25, -0.24, -0.23, -0.22, -0.21, -0.2, -0.19, -0.18, -0.17, -0.16, -0.15, -0.14, -0.13, -0.12, -0.11, -0.1, -0.09, -0.08, -0.07, -0.06, -0.05, -0.04, -0.03, -0.02, -0.01, 0, 0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09, 0.1, 0.11, 0.12, 0.13, 0.14, 0.15, 0.16, 0.17, 0.18, 0.19, 0.2, 0.21, 0.22, 0.23, 0.24, 0.25, 0.26, 0.27, 0.28, 0.29, 0.3, 0.31, 0.32, 0.33, 0.34, 0.35, 0.36, 0.37, 0.38, 0.39, 0.4, 0.41, 0.42, 0.43, 0.44, 0.45, 0.46, 0.47, 0.48, 0.49, 0.5, 0.51, 0.52, 0.53, 0.54, 0.55, 0.56, 0.57, 0.58, 0.59, 0.6, 0.61, 0.62, 0.63, 0.64, 0.65, 0.66, 0.67, 0.68, 0.69, 0.7, 0.71, 0.72, 0.73, 0.74, 0.75, 0.76, 0.77, 0.78, 0.79, 0.8, 0.81, 0.82, 0.83, 0.84, 0.85, 0.86, 0.87, 0.88, 0.89, 0.9, 0.91, 0.92, 0.93, 0.94, 0.95, 0.96, 0.97, 0.98, 0.99};
    const double etalist[98]    = {0.0075, 0.01, 0.0125, 0.015, 0.0175, 0.02, 0.0225, 0.025, 0.0275, 0.03, 0.0325, 0.035, 0.0375, 0.04, 0.0425, 0.045, 0.0475, 0.05, 0.0525, 0.055, 0.0575, 0.06, 0.0625, 0.065, 0.0675, 0.07, 0.0725, 0.075, 0.0775, 0.08, 0.0825, 0.085, 0.0875, 0.09, 0.0925, 0.095, 0.0975, 0.1, 0.1025, 0.105, 0.1075, 0.11, 0.1125, 0.115, 0.1175, 0.12, 0.1225, 0.125, 0.1275, 0.13, 0.1325, 0.135, 0.1375, 0.14, 0.1425, 0.145, 0.1475, 0.15, 0.1525, 0.155, 0.1575, 0.16, 0.1625, 0.165, 0.1675, 0.17, 0.1725, 0.175, 0.1775, 0.18, 0.1825, 0.185, 0.1875, 0.19, 0.1925, 0.195, 0.1975, 0.2, 0.2025, 0.205, 0.2075, 0.21, 0.2125, 0.215, 0.2175, 0.22, 0.2225, 0.225, 0.2275, 0.23, 0.2325, 0.235, 0.2375, 0.24, 0.2425, 0.245, 0.2475, 0.25};
    
    /* YPtag: chiAmin tables below */
    const double a3stab[98][200] = {{486.38171, 482.014119, 477.586317, 473.100898, 468.560452, 463.967572, 459.324849, 454.634874, 449.90024, 445.123537, 440.307358, 435.454294, 430.566937, 425.647878, 420.699709, 415.725022, 410.726408, 405.706459, 400.667766, 395.612921, 390.544517, 385.465143, 380.377392, 375.283856, 370.187127, 365.089795, 359.994452, 354.903691, 349.820102, 344.746278, 339.684809, 334.638289, 329.609307, 324.600456, 319.739539, 314.963908, 310.20941, 305.474574, 300.757928, 296.058, 291.373319, 286.702413, 282.04381, 277.396038, 272.757626, 268.127103, 263.502995, 258.883832, 254.268143, 249.654454, 245.041295, 240.396283, 235.749527, 231.100226, 226.447578, 221.790781, 217.129033, 212.461533, 207.78748, 203.106071, 198.416506, 193.717981, 189.009697, 184.29085, 179.56064, 174.818265, 170.062923, 165.27696, 160.443128, 155.59538, 150.734009, 145.859303, 140.971555, 136.071055, 131.158094, 126.232963, 121.295953, 116.347354, 111.387457, 106.416553, 101.434933, 96.4428872, 91.4407074, 86.4286838, 81.4974507, 76.600227, 71.6891456, 66.7615657, 61.8148462, 56.846346, 51.8534243, 46.83344, 41.7837521, 36.7017197, 31.5847016, 26.4300569, 21.2351447, 15.9973239, 10.7139535, 5.38239254, 0, -6.08763789, -12.0608823, -17.9402314, -23.7461834, -29.4992367, -35.2198894, -40.9286397, -46.645986, -52.3924264, -58.1884591, -64.0545825, -70.0112947, -76.079094, -82.2784786, -88.6299467, -95.1539966, -101.894803, -108.876246, -116.06527, -123.487448, -131.168351, -139.133551, -147.40862, -156.019128, -164.990649, -174.348753, -184.119013, -194.327, -204.998285, -216.100994, -227.739209, -239.955736, -252.79338, -266.150682, -280.142581, -294.922484, -310.576478, -327.190647, -344.851076, -363.643851, -383.401623, -404.541043, -427.197779, -451.507502, -477.605881, -504.513317, -533.759566, -565.759114, -600.92645, -639.67606, -683.197252, -730.182688, -779.288705, -830.275285, -885.15569, -944.931507, -1010.60432, -1083.7882, -1164.73305, -1253.02061, -1348.39703, -1452.45417, -1565.58855, -1688.86004, -1824.45251, -1974.0339, -2138.78785, -2320.25673, -2521.75865, -2745.60728, -2989.38409, -3260.86294, -3563.99556, -3902.73366, -4281.02897, -4852.58648, -5434.16631, -5992.28187, -6493.44657, -6904.17379, -7109.65732, -7211.94326, -7212.39631, -7069.49428, -6801.57789, -6444.85737, -6016.77997, -5534.79294, -5016.3435, -4478.8789, -3925.25323, -3414.7935, -2944.95691, -2513.20067, -2116.98199, -1753.75807, -1420.98611, -1112.62652, -824.627661},
        {459.284623, 455.294794, 451.230698, 447.095748, 442.893358, 438.626943, 434.299914, 429.915688, 425.477677, 420.989294, 416.453955, 411.875072, 407.256059, 402.600331, 397.911301, 393.192382, 388.446989, 383.678535, 378.890434, 374.0861, 369.268946, 364.442387, 359.609836, 354.774707, 349.940413, 345.110369, 340.287989, 335.476685, 330.679872, 325.900964, 321.143374, 316.410517, 311.705805, 307.032653, 302.612369, 298.334837, 294.087321, 289.866163, 285.667706, 281.488293, 277.324267, 273.17197, 269.027745, 264.887935, 260.748881, 256.606928, 252.458417, 248.299692, 244.127095, 239.936968, 235.725654, 231.272658, 226.795861, 222.296307, 217.775038, 213.233099, 208.671532, 204.091382, 199.49369, 194.879501, 190.249858, 185.605804, 180.948382, 176.278636, 171.597609, 166.906345, 162.205886, 157.511834, 152.839439, 148.159722, 143.472781, 138.778718, 134.07763, 129.369619, 124.654783, 119.933222, 115.205036, 110.470324, 105.729186, 100.981722, 96.228031, 91.4682129, 86.7023672, 81.9305935, 77.247984, 72.6051439, 67.951537, 63.2841801, 58.6000902, 53.8962842, 49.1697791, 44.4175917, 39.6367391, 34.8242381, 29.9771057, 25.0923589, 20.1670145, 15.1980894, 10.1826007, 5.11756527, 0, -5.9259591, -11.7291727, -17.4300365, -23.0489465, -28.6062984, -34.122488, -39.6179112, -45.1129638, -50.6280416, -56.1835405, -61.7998562, -67.4973846, -73.2965215, -79.2176627, -85.2812041, -91.5075414, -97.9184851, -104.534633, -111.373181, -118.454829, -125.800275, -133.430219, -141.365358, -149.626393, -158.234022, -167.208944, -176.571858, -186.343463, -196.544458, -207.089667, -218.144485, -229.761372, -241.992791, -254.8952, -268.519062, -282.913772, -298.130594, -314.220788, -331.235617, -349.226342, -367.703928, -387.424373, -408.544648, -431.221726, -455.61258, -481.537286, -509.573935, -539.963725, -572.94785, -608.767509, -647.972797, -690.118464, -734.362578, -780.642936, -830.489277, -884.716592, -944.139869, -1010.02413, -1082.63205, -1161.73537, -1249.03305, -1341.90293, -1434.56161, -1526.17421, -1624.29082, -1732.61269, -1858.43532, -2006.39182, -2167.93629, -2341.97762, -2524.25325, -2717.33143, -2921.21886, -3135.92225, -3361.44829, -3585.1549, -3822.85975, -4077.73175, -4352.93978, -4651.65275, -5058.34714, -5440.6792, -5766.9717, -5857.0752, -5738.08274, -5498.95078, -5163.67917, -4756.26778, -4300.71646, -3821.02508, -3372.25, -2957.29785, -2574.67695, -2222.89566, -1900.46232, -1605.88528, -1337.67288, -1095.85441, -888.527963},
        {434.112712, 430.090631, 426.034601, 421.946307, 417.827436, 413.679675, 409.504708, 405.304224, 401.079907, 396.833445, 392.566522, 388.280826, 383.978043, 379.659858, 375.327958, 370.98403, 366.629758, 362.266831, 357.896932, 353.52175, 349.14297, 344.762278, 340.381361, 336.001904, 331.625595, 327.254118, 322.88916, 318.532408, 314.185548, 309.850265, 305.528246, 301.221178, 296.930745, 292.658636, 288.546796, 284.523833, 280.516664, 276.522424, 272.538248, 268.56127, 264.588625, 260.617447, 256.64487, 252.668029, 248.684059, 244.690093, 240.683267, 236.660714, 232.61957, 228.556968, 224.470042, 220.148706, 215.801808, 211.430974, 207.037832, 202.624007, 198.191127, 193.740818, 189.274706, 184.794419, 180.301583, 175.797825, 171.284771, 166.764047, 162.237281, 157.706099, 153.172127, 148.666886, 144.221177, 139.774974, 135.327967, 130.879844, 126.430294, 121.979006, 117.525669, 113.069971, 108.611601, 104.150248, 99.6856008, 95.2173481, 90.7451787, 86.2687813, 81.7878448, 77.3020579, 72.905106, 68.5477025, 64.179431, 59.79693, 55.3968377, 50.9757927, 46.5304333, 42.057398, 37.5533252, 33.0148532, 28.4386205, 23.8212654, 19.1594265, 14.4497421, 9.68885062, 4.87339044, 0, -5.70837099, -11.3065105, -16.8133494, -22.2478185, -27.6288485, -32.9753702, -38.3063145, -43.6406121, -48.9971939, -54.3949906, -59.8529331, -65.3899521, -71.0249784, -76.7769429, -82.6647763, -88.7074095, -94.9273491, -101.344037, -107.968315, -114.819881, -121.918431, -129.283663, -136.935274, -144.892961, -153.176421, -161.805351, -170.799447, -180.178408, -189.96193, -200.035467, -210.602181, -221.722044, -233.455025, -245.915197, -259.135478, -273.134361, -287.955585, -303.642891, -320.240018, -337.790706, -355.964843, -375.293802, -395.894465, -417.883718, -441.378444, -466.855804, -493.982338, -522.78486, -553.290186, -585.525131, -618.697719, -654.654303, -694.353388, -738.410522, -786.741039, -839.574659, -897.141095, -960.293546, -1028.49655, -1100.53448, -1175.57001, -1253.98173, -1335.14896, -1419.11907, -1507.07141, -1599.66572, -1699.44658, -1807.56238, -1918.25054, -2029.65773, -2134.90362, -2237.93526, -2338.63939, -2436.90281, -2532.61227, -2550.59925, -2584.56965, -2653.17407, -2775.06311, -2968.88738, -3391.92504, -3831.78075, -4247.69327, -4434.55066, -4416.6472, -4286.75662, -4065.74925, -3774.49536, -3433.86528, -3064.72929, -2742.22085, -2420.50355, -2105.75969, -1804.17155, -1521.92145, -1265.19167, -1040.16452, -826.039404, -636.713941},
        {408.975106, 404.811377, 400.68602, 396.597255, 392.543305, 388.522388, 384.532726, 380.57254, 376.640051, 372.733478, 368.851044, 364.990967, 361.15147, 357.330773, 353.527096, 349.738661, 345.963688, 342.200397, 338.44701, 334.701747, 330.962829, 327.228477, 323.496911, 319.766351, 316.03502, 312.301136, 308.562922, 304.818598, 301.066384, 297.304501, 293.53117, 289.744612, 285.943047, 282.124697, 278.210793, 274.23967, 270.250587, 266.244264, 262.221418, 258.182771, 254.12904, 250.060944, 245.979204, 241.884537, 237.777663, 233.659302, 229.530172, 225.390992, 221.242482, 217.08536, 212.920346, 208.779426, 204.631374, 200.476232, 196.314041, 192.144841, 187.968675, 183.785583, 179.595607, 175.398788, 171.195168, 166.984787, 162.767687, 158.543909, 154.313495, 150.076485, 145.832922, 141.576005, 137.2991, 133.016356, 128.728258, 124.435291, 120.137939, 115.836687, 111.532021, 107.224423, 102.914381, 98.6023771, 94.2888974, 89.9744263, 85.6594486, 81.3444489, 77.0299122, 72.716323, 68.5487667, 64.4523862, 60.3505867, 56.2391602, 52.1138988, 47.9705949, 43.8050405, 39.6130278, 35.390349, 31.1327963, 26.8361619, 22.4962378, 18.1088164, 13.6696897, 9.17464997, 4.61948934, 0, -5.55175292, -10.9970051, -16.3540744, -21.6412786, -26.8769354, -32.0793626, -37.2668781, -42.4577997, -47.6704451, -52.9231321, -58.2341785, -63.6219022, -69.1046209, -74.7006523, -80.4283144, -86.3059248, -92.3792483, -98.6705235, -105.135986, -111.799836, -118.686271, -125.819492, -133.223697, -140.923086, -148.941857, -157.304212, -166.034347, -175.156464, -184.69476, -194.602547, -205.000905, -215.9353, -227.451198, -239.535279, -252.262401, -265.7231, -279.980476, -295.097631, -311.137666, -328.163683, -346.236532, -365.42225, -385.784379, -407.386459, -430.292032, -454.393507, -479.968341, -507.122859, -535.963386, -566.596246, -598.687374, -633.32174, -671.816865, -714.674624, -760.731612, -809.572099, -860.780358, -913.234573, -967.385576, -1024.45448, -1085.60795, -1149.46027, -1213.95115, -1278.64213, -1345.8429, -1416.60172, -1496.78719, -1588.69727, -1676.95531, -1756.18242, -1807.81815, -1841.69309, -1856.99122, -1852.8965, -1828.59288, -1649.08674, -1481.28402, -1357.9131, -1311.70232, -1375.38004, -1764.03073, -2206.45592, -2657.23136, -2910.31484, -2983.86007, -2962.94488, -2862.37676, -2696.96318, -2481.51165, -2230.82964, -1984.51884, -1726.92513, -1466.12629, -1210.20013, -967.224428, -745.276976, -552.435566, -353.697537, -171.729325},
        {385.836919, 382.118356, 378.432453, 374.777342, 371.151156, 367.552029, 363.978093, 360.42748, 356.898325, 353.388759, 349.896916, 346.420928, 342.958928, 339.509049, 336.069425, 332.638187, 329.213468, 325.793402, 322.376121, 318.959759, 315.542447, 312.122319, 308.697508, 305.266147, 301.826368, 298.376304, 294.914089, 291.437854, 287.945733, 284.435859, 280.906364, 277.355382, 273.781045, 270.181486, 266.446684, 262.631123, 258.790589, 254.926724, 251.041171, 247.135571, 243.211569, 239.270806, 235.314926, 231.34557, 227.364382, 223.373004, 219.373079, 215.366249, 211.354158, 207.338447, 203.32076, 199.416907, 195.511888, 191.60487, 187.695022, 183.78151, 179.863504, 175.940169, 172.010675, 168.074188, 164.129877, 160.17691, 156.214453, 152.241675, 148.257743, 144.261825, 140.253089, 136.198671, 132.066476, 127.921735, 123.765691, 119.59959, 115.424674, 111.242189, 107.053378, 102.859486, 98.6617569, 94.4614347, 90.2597638, 86.0579882, 81.8573523, 77.6591002, 73.4644761, 69.2747241, 65.2734682, 61.3669265, 57.4591253, 53.5453901, 49.6210467, 45.6814205, 41.7218372, 37.7376225, 33.7241019, 29.676601, 25.5904456, 21.4609611, 17.2834732, 13.0533075, 8.7657897, 4.41624532, 0, -5.43212607, -10.7520904, -15.978402, -21.1295701, -26.2241036, -31.2805117, -36.3173035, -41.3529879, -46.4060742, -51.4950714, -56.6384886, -61.8548348, -67.1626192, -72.5803508, -78.1265387, -83.8196919, -89.7283826, -95.8782715, -102.17463, -108.646696, -115.323705, -122.234894, -129.409501, -136.876762, -144.665913, -152.806193, -161.326837, -170.257083, -179.626167, -189.479755, -199.824632, -210.685104, -222.08548, -233.879279, -246.176204, -259.131499, -272.82071, -287.31938, -302.703054, -319.047276, -336.6654, -355.322784, -375.048446, -395.871401, -417.820667, -440.221826, -463.983188, -489.309627, -516.406018, -545.477237, -576.822313, -610.436887, -647.453037, -688.122865, -730.901853, -775.052127, -819.835815, -862.91455, -905.514703, -950.608638, -1001.24757, -1054.06156, -1105.80229, -1155.31759, -1205.00714, -1255.64028, -1313.96355, -1382.29594, -1441.0399, -1482.99511, -1493.12839, -1473.27733, -1418.95297, -1325.66636, -1188.92855, -771.999169, -360.703538, -8.61556821, 230.690837, 303.641771, -103.710004, -610.982838, -1160.16206, -1503.84212, -1658.97557, -1721.42785, -1704.20588, -1620.31657, -1482.76684, -1304.56362, -1105.59042, -895.096213, -679.530962, -465.344621, -258.987145, -66.9084913, 104.441384, 270.078387, 433.118704},
        {365.912682, 363.109694, 360.25406, 357.347546, 354.391918, 351.38894, 348.34038, 345.248001, 342.113571, 338.938854, 335.725616, 332.475624, 329.190641, 325.872435, 322.522771, 319.143413, 315.736129, 312.302684, 308.844842, 305.364371, 301.863035, 298.3426, 294.804832, 291.251497, 287.684359, 284.105185, 280.51574, 276.91779, 273.313101, 269.703438, 266.090566, 262.476252, 258.862261, 255.250359, 251.739604, 248.28107, 244.824659, 241.36898, 237.91264, 234.454249, 230.992413, 227.525743, 224.052845, 220.572328, 217.082801, 213.582872, 210.071148, 206.546239, 203.006752, 199.451295, 195.878478, 192.204425, 188.512015, 184.801646, 181.073712, 177.328612, 173.56674, 169.788493, 165.994268, 162.184461, 158.359469, 154.519687, 150.665512, 146.79734, 142.915568, 139.020592, 135.112808, 131.184141, 127.226719, 123.25841, 119.280161, 115.292916, 111.297621, 107.295222, 103.286665, 99.2728938, 95.2548552, 91.2334945, 87.2097571, 83.1845886, 79.1589346, 75.1337405, 71.1099519, 67.0885143, 63.2343829, 59.4630487, 55.6880315, 51.9049542, 48.10944, 44.2971121, 40.4635934, 36.6045071, 32.7154763, 28.7921241, 24.8300736, 20.8249479, 16.7723701, 12.6679633, 8.50735066, 4.28615518, 0, -5.25949092, -10.4105459, -15.4711942, -20.4594653, -25.3933884, -30.290993, -35.1703084, -40.0493639, -44.9461889, -49.8788127, -54.8652647, -59.9235743, -65.0717707, -70.3278834, -75.7099417, -81.2359748, -86.9762977, -92.9564091, -99.0760735, -105.364524, -111.850995, -118.564719, -125.53493, -132.79086, -140.361744, -148.276814, -156.565305, -165.256449, -174.37948, -183.989543, -194.080458, -204.673686, -215.790686, -227.334498, -239.385792, -252.056816, -265.404555, -279.485996, -294.358124, -310.077924, -326.852519, -344.543064, -363.17717, -382.782448, -403.386509, -424.771211, -447.271356, -470.975994, -495.974173, -522.354943, -549.803269, -579.306165, -612.295927, -649.25031, -688.698223, -730.043565, -772.690239, -815.624193, -858.762268, -902.477256, -946.355343, -990.842795, -1040.08833, -1093.9031, -1144.74844, -1188.45938, -1221.03277, -1238.34559, -1235.68147, -1208.65967, -1170.75975, -1096.99293, -976.796048, -799.605949, -554.859476, 126.241303, 806.466075, 1406.8193, 1848.30545, 2051.92897, 1573.16136, 942.228702, 236.792452, -215.777688, -436.663116, -551.090336, -575.03849, -524.486721, -415.414172, -263.799987, -130.310601, 28.8387208, 204.935896, 389.268845, 573.125485, 747.793736, 904.561517, 1018.79555, 1157.43195},
        {346.268512, 343.649868, 340.988407, 338.285432, 335.542248, 332.760157, 329.940462, 327.084469, 324.193479, 321.268797, 318.311727, 315.323571, 312.305633, 309.259218, 306.185627, 303.086166, 299.962137, 296.814844, 293.645591, 290.455681, 287.246417, 284.019104, 280.775045, 277.515542, 274.241901, 270.955424, 267.657415, 264.349178, 261.032015, 257.707231, 254.37613, 251.040014, 247.700187, 244.357953, 241.091871, 237.862991, 234.631437, 231.396004, 228.155489, 224.908688, 221.654398, 218.391414, 215.118535, 211.834556, 208.538273, 205.228483, 201.903983, 198.563568, 195.206036, 191.830182, 188.434804, 184.946201, 181.437237, 177.908281, 174.3597, 170.791864, 167.20514, 163.599895, 159.976498, 156.335318, 152.676722, 149.001077, 145.308753, 141.600117, 137.875538, 134.135383, 130.38002, 126.597412, 122.77582, 118.941198, 115.094716, 111.237547, 107.370864, 103.495838, 99.6136423, 95.7254485, 91.8324289, 87.9357558, 84.0366014, 80.1361378, 76.2355374, 72.3359724, 68.4386149, 64.5446371, 60.8299577, 57.2047001, 53.5768875, 49.9420213, 46.2956028, 42.6331334, 38.9501145, 35.2420473, 31.5044334, 27.732774, 23.9225705, 20.0693243, 16.1685367, 12.2157091, 8.20634291, 4.13593941, 0, -5.01159832, -9.93786259, -14.794955, -19.5990376, -24.3662726, -29.1128221, -33.8548482, -38.6085131, -43.389979, -48.215408, -53.1009621, -58.0628036, -63.1170946, -68.2799972, -73.5676735, -78.9962858, -84.6186541, -90.4561773, -96.4421009, -102.600442, -108.955219, -115.530448, -122.350147, -129.438333, -136.819024, -144.516238, -152.553991, -160.956301, -169.747185, -178.901561, -188.510549, -198.612898, -209.247356, -220.508007, -232.405931, -244.937451, -258.124713, -271.989864, -286.555052, -301.842422, -317.618465, -334.238793, -351.775573, -370.300969, -389.88715, -410.446094, -432.250201, -455.411685, -480.04276, -506.255638, -534.353424, -564.02413, -595.357391, -628.349791, -662.807936, -698.623728, -735.689067, -774.70916, -814.577762, -853.301393, -888.247811, -921.929329, -961.786227, -1008.09711, -1049.00552, -1078.22521, -1087.94601, -1071.48659, -1027.75321, -952.491469, -880.117524, -761.757106, -579.769854, -316.515405, 45.6465991, 1026.76416, 2016.46808, 2906.79683, 3589.78885, 3957.48259, 3421.06354, 2673.99176, 1814.38554, 1249.57883, 955.292071, 778.405809, 701.082461, 705.484443, 773.774171, 888.114057, 951.377502, 1060.51622, 1201.83684, 1361.64603, 1526.25041, 1681.95663, 1815.07134, 1888.54052, 2001.90689},
        {326.197572, 322.931095, 319.746404, 316.639285, 313.605523, 310.640903, 307.741209, 304.902226, 302.11974, 299.389536, 296.707398, 294.069112, 291.470462, 288.907233, 286.375211, 283.870179, 281.387925, 278.924231, 276.474884, 274.035667, 271.602367, 269.170768, 266.736655, 264.295813, 261.844027, 259.377082, 256.890763, 254.380854, 251.843142, 249.27341, 246.667445, 244.021029, 241.32995, 238.589991, 235.564375, 232.370059, 229.126797, 225.837922, 222.506764, 219.136658, 215.730933, 212.292923, 208.82596, 205.333376, 201.818503, 198.284672, 194.735217, 191.173469, 187.60276, 184.026423, 180.447789, 177.092656, 173.737069, 170.379536, 167.018568, 163.652672, 160.28036, 156.90014, 153.510521, 150.110014, 146.697128, 143.270372, 139.828255, 136.369288, 132.891979, 129.394838, 125.876375, 122.278186, 118.543234, 114.787408, 111.012905, 107.221925, 103.416667, 99.5993302, 95.7721126, 91.9372135, 88.0968318, 84.2531664, 80.4084161, 76.5647798, 72.7244565, 68.8896451, 65.0625443, 61.2453532, 57.6707508, 54.2208479, 50.7749855, 47.3278829, 43.8742594, 40.4088344, 36.9263273, 33.4214575, 29.8889442, 26.3235069, 22.719865, 19.0727377, 15.3768446, 11.6269049, 7.81763795, 3.94376321, 0, -4.72283176, -9.39701609, -14.0358595, -18.6526685, -23.2607497, -27.8734095, -32.5039545, -37.1656911, -41.8719259, -46.6359654, -51.4711162, -56.3906847, -61.4079775, -66.536301, -71.7889619, -77.1792665, -82.7362737, -88.4755401, -94.3747425, -100.450563, -106.719683, -113.198785, -119.904551, -126.853663, -134.062803, -141.548652, -149.327893, -157.417208, -165.833279, -174.427587, -183.442588, -192.944525, -202.999639, -213.935553, -225.687819, -238.122287, -251.226786, -264.989143, -279.397186, -294.438742, -309.339494, -325.081374, -341.801324, -359.636289, -378.723212, -398.342005, -419.7009, -443.1511, -469.043807, -497.730221, -531.710591, -566.56046, -599.477243, -629.18821, -657.544098, -684.993271, -711.984093, -740.024084, -768.261736, -794.6901, -818.254597, -838.992792, -859.911882, -880.400373, -893.715109, -895.927468, -886.358894, -861.923375, -807.617992, -715.1807, -622.910422, -468.820925, -228.53291, 122.332923, 608.155874, 1895.37867, 3205.80133, 4403.28729, 5351.69999, 5914.90288, 5372.19006, 4561.70721, 3597.84748, 2944.05906, 2576.5675, 2326.15874, 2176.33031, 2110.57973, 2112.40451, 2165.30218, 2161.68064, 2226.04576, 2337.52114, 2475.23041, 2618.29718, 2745.84504, 2836.99762, 2885.42327, 2966.37507},
        {308.114675, 304.916178, 301.828887, 298.84715, 295.965318, 293.17774, 290.478767, 287.862747, 285.324031, 282.856969, 280.45591, 278.115203, 275.8292, 273.592249, 271.3987, 269.242903, 267.119209, 265.021965, 262.945524, 260.884233, 258.832443, 256.784504, 254.734765, 252.677576, 250.607288, 248.518249, 246.404809, 244.261319, 242.082128, 239.861586, 237.594042, 235.273846, 232.895349, 230.4529, 227.625837, 224.572643, 221.455584, 218.279232, 215.04816, 211.76694, 208.440146, 205.072348, 201.66812, 198.232035, 194.768664, 191.282581, 187.778357, 184.260565, 180.733777, 177.202567, 173.671506, 170.439666, 167.210736, 163.982904, 160.75436, 157.52329, 154.287884, 151.04633, 147.796816, 144.53753, 141.26666, 137.982396, 134.682925, 131.366436, 128.031116, 124.675154, 121.296739, 117.82787, 114.202134, 110.55423, 106.886638, 103.201836, 99.5023045, 95.7905214, 92.0689662, 88.3401179, 84.6064557, 80.8704585, 77.1346056, 73.401376, 69.6732488, 65.952703, 62.2422178, 58.5442722, 55.1110604, 51.8149514, 48.5253131, 45.2365209, 41.9429501, 38.6389761, 35.3189741, 31.9773197, 28.6083881, 25.2065546, 21.7661948, 18.2816838, 14.7473971, 11.15771, 7.50699787, 3.78963609, 0, -4.52711422, -9.02362609, -13.5012812, -17.971825, -22.447003, -26.9385609, -31.458244, -36.017798, -40.6289684, -45.3035006, -50.0531404, -54.889633, -59.8247242, -64.8701594, -70.0376842, -75.339044, -80.8075008, -86.4578736, -92.2532405, -98.2099574, -104.344381, -110.672866, -117.211771, -123.97745, -130.98626, -138.254557, -145.798697, -153.635036, -161.779931, -170.064031, -178.75749, -187.932379, -197.660766, -208.323399, -219.838006, -232.040002, -244.90885, -258.424014, -272.564959, -287.311148, -301.863185, -317.21644, -333.502764, -350.854005, -369.402014, -387.7962, -408.021465, -430.580269, -455.975074, -484.708341, -520.387387, -556.614989, -589.507952, -617.241523, -642.193615, -664.855231, -685.717371, -705.722821, -724.808123, -742.416965, -760.218215, -774.62991, -785.235286, -789.987238, -784.076134, -763.960386, -730.551412, -681.46211, -597.977686, -470.619213, -347.245897, -152.219424, 143.786951, 570.099971, 1156.04638, 2662.38828, 4204.15822, 5627.82411, 6779.85383, 7506.71531, 7010.96455, 6212.25594, 5233.01962, 4562.51671, 4177.97343, 3903.76972, 3724.7741, 3625.85507, 3591.88116, 3607.72087, 3546.57721, 3575.6043, 3664.88331, 3784.49538, 3904.52168, 3995.04337, 4026.14159, 4035.74983, 4053.71386},
        {291.464229, 289.028473, 286.646671, 284.31535, 282.031035, 279.790252, 277.589526, 275.425384, 273.29435, 271.192952, 269.117714, 267.065162, 265.031822, 263.01422, 261.008881, 259.012331, 257.021097, 255.031703, 253.040675, 251.04454, 249.039822, 247.023048, 244.990744, 242.939434, 240.865645, 238.765903, 236.636733, 234.474661, 232.276213, 230.037914, 227.756291, 225.427868, 223.049172, 220.616729, 217.920854, 215.065515, 212.15716, 209.199004, 206.194267, 203.146165, 200.057917, 196.93274, 193.773852, 190.584469, 187.367811, 184.127094, 180.865536, 177.586355, 174.292769, 170.987995, 167.67525, 164.55129, 161.421599, 158.285199, 155.141114, 151.988364, 148.825972, 145.652961, 142.468352, 139.271167, 136.060429, 132.83516, 129.594382, 126.337116, 123.062386, 119.769213, 116.45662, 113.074619, 109.574399, 106.056062, 102.521806, 98.97383, 95.4143339, 91.8455166, 88.2695773, 84.6887154, 81.10513, 77.5210203, 73.9385855, 70.3600249, 66.7875377, 63.2233231, 59.6695803, 56.1285086, 52.84593, 49.697108, 46.5543779, 43.4120331, 40.2643666, 37.1056716, 33.9302414, 30.732369, 27.5063478, 24.2464708, 20.9470314, 17.6023226, 14.2066376, 10.7542697, 7.23951198, 3.65665769, 0, -4.46365193, -8.88903286, -13.2881123, -17.6728598, -22.055245, -26.4472372, -30.8608061, -35.3079211, -39.8005518, -44.3506678, -48.9702385, -53.6712335, -58.4656224, -63.3653745, -68.3824595, -73.528847, -78.8839671, -84.4694266, -90.1446053, -95.9359287, -101.869822, -107.972711, -114.27102, -120.791176, -127.559603, -134.602728, -141.946974, -149.618769, -157.644537, -165.988009, -174.761293, -184.009623, -193.778232, -204.215552, -215.315217, -227.043344, -239.414205, -252.442075, -266.141229, -280.525939, -295.540179, -311.289922, -327.803194, -345.108027, -363.232447, -379.81902, -397.877606, -418.032599, -440.908394, -467.129388, -500.520574, -534.593903, -566.448822, -594.421414, -619.372553, -641.029537, -659.119661, -671.383797, -679.987123, -689.261825, -707.04356, -722.495157, -729.764376, -724.042348, -702.288139, -660.649303, -599.089425, -514.745399, -390.762017, -218.198795, -52.2356447, 190.052987, 539.803105, 1028.15071, 1686.23182, 3291.88199, 4942.86278, 6483.63528, 7758.66063, 8612.39991, 8252.73088, 7585.08693, 6726.75062, 6139.48762, 5804.22257, 5558.34554, 5389.95791, 5287.16107, 5238.05643, 5230.74539, 5129.42571, 5136.46742, 5211.55873, 5314.38783, 5404.64292, 5442.01221, 5386.18388, 5346.6087, 5263.90237},
        {276.767662, 275.248375, 273.698831, 272.118947, 270.50864, 268.867825, 267.19642, 265.494341, 263.761504, 261.997826, 260.203222, 258.37761, 256.520905, 254.633024, 252.713884, 250.7634, 248.78149, 246.768069, 244.723054, 242.646362, 240.537908, 238.39761, 236.225382, 234.021143, 231.784808, 229.516294, 227.215516, 224.882392, 222.516838, 220.11877, 217.688105, 215.224758, 212.728647, 210.199687, 207.610015, 204.974021, 202.306431, 199.608062, 196.879733, 194.12226, 191.336463, 188.523158, 185.683165, 182.8173, 179.926382, 177.011229, 174.072658, 171.111488, 168.128536, 165.124621, 162.10056, 159.082009, 156.04441, 152.988042, 149.913185, 146.820118, 143.70912, 140.580471, 137.434451, 134.271339, 131.091414, 127.894955, 124.682244, 121.453558, 118.209177, 114.949381, 111.674449, 108.362392, 104.991752, 101.60874, 98.2150786, 94.8124909, 91.4027001, 87.9874294, 84.5684017, 81.1473404, 77.7259685, 74.3060091, 70.8891853, 67.4772204, 64.0718374, 60.6747594, 57.2877096, 53.9124112, 50.7789756, 47.7701281, 44.7658485, 41.7604483, 38.7482391, 35.7235322, 32.6806393, 29.6138717, 26.517541, 23.3859586, 20.2134361, 16.9942848, 13.7228163, 10.3933421, 7.0001737, 3.53762249, 0, -4.49069793, -8.91898961, -13.298233, -17.6417862, -21.9630069, -26.2752534, -30.5918834, -34.9262551, -39.2917263, -43.701655, -48.1693992, -52.7083169, -57.331766, -62.0531046, -66.8856905, -71.8428818, -77.0395781, -82.5036432, -88.0187577, -93.6200384, -99.3426023, -105.221566, -111.292047, -117.589162, -124.148027, -131.00376, -138.191477, -145.746296, -153.703333, -162.16635, -171.076649, -180.448754, -190.297189, -200.523914, -211.199734, -222.42547, -234.249414, -246.719858, -259.885095, -273.793415, -289.044567, -304.967553, -321.502772, -338.590622, -356.171501, -371.593297, -388.037049, -406.091282, -426.344523, -449.385299, -478.371104, -508.181648, -537.692621, -565.779886, -591.319643, -613.187944, -630.260837, -637.620455, -638.799018, -641.467564, -656.990348, -670.490692, -674.293964, -661.977177, -629.238185, -570.80134, -482.983903, -360.923202, -193.915906, 25.4375116, 237.14557, 526.366644, 923.827912, 1460.25655, 2166.37975, 3790.99136, 5467.23522, 7046.32182, 8379.46168, 9317.8653, 9116.8588, 8620.79334, 7936.25847, 7478.54457, 7232.0471, 7052.53613, 6930.83838, 6857.78055, 6824.18935, 6820.89151, 6617.9907, 6595.30914, 6685.76257, 6822.26671, 6937.73731, 6965.09012, 6837.24086, 6751.95526, 6556.40516},
        {263.768306, 262.511119, 261.209677, 259.864479, 258.476022, 257.044803, 255.57132, 254.056069, 252.499548, 250.902254, 249.264685, 247.587337, 245.870708, 244.115295, 242.321595, 240.490106, 238.621324, 236.715748, 234.773874, 232.796199, 230.783221, 228.735438, 226.653345, 224.537441, 222.388222, 220.206187, 217.991832, 215.745654, 213.468151, 211.15982, 208.821158, 206.452662, 204.05483, 201.628159, 199.181651, 196.711371, 194.213284, 191.687611, 189.134573, 186.554392, 183.947288, 181.313483, 178.653198, 175.966654, 173.254073, 170.515676, 167.751684, 164.962319, 162.147801, 159.308352, 156.444193, 153.53283, 150.597693, 147.639493, 144.658947, 141.656766, 138.633665, 135.590358, 132.527558, 129.445978, 126.346333, 123.229336, 120.0957, 116.94614, 113.78137, 110.602102, 107.40905, 104.188733, 100.928011, 97.6568726, 94.3769527, 91.0898851, 87.7973036, 84.5008421, 81.2021345, 77.9028148, 74.6045167, 71.3088743, 68.0175213, 64.7320917, 61.4542194, 58.1855382, 54.9276822, 51.6822851, 48.6819473, 45.8079614, 42.938844, 40.0687336, 37.1917689, 34.3020885, 31.3938308, 28.4611346, 25.4981384, 22.4989808, 19.4578004, 16.3687358, 13.2259255, 10.0235082, 6.75562249, 3.41640688, 0, -4.49149474, -8.90440585, -13.2529514, -17.5513496, -21.8138184, -26.054576, -30.2878405, -34.5278299, -38.7887624, -43.0848562, -47.4303292, -51.8393996, -56.3262855, -60.9052049, -65.5903761, -70.3960171, -75.4092151, -80.654598, -85.981561, -91.4199371, -96.9995592, -102.75026, -108.701873, -114.88423, -121.327165, -128.06051, -135.114099, -142.517764, -150.301338, -158.5379, -167.198181, -176.299038, -185.857333, -195.833945, -206.273722, -217.236441, -228.755754, -240.865313, -253.59877, -266.989779, -281.330262, -296.316997, -311.933105, -328.161706, -344.985921, -360.75567, -377.495576, -395.597061, -415.451547, -437.450455, -463.999344, -491.013776, -517.789856, -543.40374, -566.482525, -585.854929, -600.349667, -606.024504, -605.108869, -602.855042, -607.124182, -606.890166, -598.861435, -577.86464, -535.537841, -464.982715, -359.213938, -211.31063, -14.6709082, 237.487553, 484.039529, 810.122211, 1246.02439, 1822.03484, 2568.44236, 4217.36628, 5921.80721, 7536.59628, 8916.56464, 9916.54345, 9816.24205, 9429.02792, 8856.21299, 8493.14263, 8324.74053, 8213.41642, 8150.21968, 8126.19969, 8132.40582, 8159.88746, 7774.25988, 7707.50874, 7847.89828, 8083.69274, 8303.15639, 8394.55346, 8246.14821, 8135.46555, 7856.02698},
        {251.80006, 250.725247, 249.598023, 248.419262, 247.189837, 245.910623, 244.582493, 243.20632, 241.782978, 240.313341, 238.798283, 237.238677, 235.635396, 233.989316, 232.301308, 230.572247, 228.803007, 226.99446, 225.147482, 223.262945, 221.341722, 219.384689, 217.392718, 215.366683, 213.307458, 211.215916, 209.092932, 206.939377, 204.756128, 202.544056, 200.304036, 198.036941, 195.743645, 193.425022, 191.117765, 188.804085, 186.465761, 184.102505, 181.714027, 179.300039, 176.860253, 174.394378, 171.902127, 169.383211, 166.837341, 164.264228, 161.663583, 159.035119, 156.378545, 153.693573, 150.979914, 148.173871, 145.339938, 142.479201, 139.592747, 136.68166, 133.747026, 130.789932, 127.811464, 124.812706, 121.794745, 118.758667, 115.705557, 112.636502, 109.552586, 106.454897, 103.344519, 100.215191, 97.0608271, 93.8976677, 90.7272747, 87.5512101, 84.3710362, 81.188315, 78.0046086, 74.8214793, 71.6404891, 68.4632001, 65.2911746, 62.1259745, 58.9691621, 55.8222995, 52.6869488, 49.5646721, 46.6908845, 43.9453028, 41.2048332, 38.4634489, 35.7151231, 32.9538289, 30.1735394, 27.3682278, 24.5318673, 21.658431, 18.7418921, 15.7762237, 12.755399, 9.67339111, 6.52417321, 3.30171846, 0, -4.50141308, -8.90731024, -13.2328347, -17.4931297, -21.7033384, -25.8786042, -30.0340701, -34.1848794, -38.3461754, -42.5331013, -46.7608003, -51.0444156, -55.3990904, -59.8399681, -64.3821917, -69.0409045, -73.8651193, -78.8748178, -84.0085336, -89.2886677, -94.7376211, -100.377795, -106.231589, -112.321406, -118.669646, -125.298711, -132.231, -139.488916, -147.094858, -155.068997, -163.436783, -172.221287, -181.44558, -191.15985, -201.373609, -212.089137, -223.32137, -235.085243, -247.395691, -260.267649, -273.611527, -287.578598, -302.204248, -317.523863, -333.572828, -349.877313, -367.109225, -385.431253, -405.006087, -425.996417, -449.938099, -473.942343, -497.655258, -520.368408, -540.64949, -557.3912, -569.486236, -574.524264, -572.997156, -566.818272, -559.242354, -544.282398, -523.810288, -494.332455, -443.263871, -362.192635, -241.087838, -71.1178379, 150.612639, 430.356813, 709.417648, 1068.99096, 1537.81277, 2144.61905, 2918.14582, 4555.20552, 6249.43859, 7862.56193, 9256.29242, 10292.3469, 10289.0334, 10013.7503, 9561.10424, 9302.27864, 9222.40158, 9191.36064, 9200.0459, 9239.3474, 9300.15522, 9373.3594, 8833.68399, 8739.45127, 8938.36811, 9278.1414, 9606.47802, 9771.08483, 9619.66874, 9492.27298, 9145.72712},
        {240.493549, 239.529841, 238.512237, 237.441726, 236.319297, 235.145937, 233.922636, 232.650382, 231.330164, 229.962971, 228.549791, 227.091612, 225.589424, 224.044215, 222.456973, 220.828687, 219.160346, 217.452938, 215.707452, 213.924877, 212.106201, 210.252413, 208.364501, 206.443454, 204.490261, 202.50591, 200.49139, 198.447689, 196.375796, 194.2767, 192.151389, 190.000852, 187.826078, 185.628054, 183.457409, 181.289268, 179.098159, 176.88346, 174.644548, 172.380802, 170.091599, 167.776316, 165.434332, 163.065024, 160.66777, 158.241948, 155.786934, 153.302109, 150.786847, 148.240529, 145.662531, 142.961064, 140.228649, 137.46664, 134.676392, 131.859257, 129.016591, 126.149747, 123.26008, 120.348942, 117.417689, 114.467675, 111.500253, 108.516777, 105.518602, 102.507081, 99.4835694, 96.4468116, 93.395616, 90.3367166, 87.2716366, 84.201899, 81.129027, 78.0545438, 74.9799725, 71.9068362, 68.8366581, 65.7709613, 62.711269, 59.6591042, 56.6159902, 53.58345, 50.5630069, 47.5561839, 44.8019238, 42.178046, 39.5595166, 36.9401539, 34.3137765, 31.6742028, 29.0152511, 26.3307399, 23.6144876, 20.8603126, 18.0620333, 15.2134682, 12.3084357, 9.34075424, 6.30424218, 3.19271795, 0, -4.50040299, -8.89113704, -13.1880877, -17.4071405, -21.564181, -25.6750948, -29.7557673, -33.8220841, -37.8899309, -41.975193, -46.0937561, -50.2615058, -54.4943274, -58.8081067, -63.2187291, -67.7420802, -72.3855315, -77.1637523, -82.111886, -87.2439937, -92.5741366, -98.1163757, -103.884772, -109.893387, -116.156282, -122.687517, -129.501154, -136.611254, -144.031878, -151.718133, -159.764651, -168.203178, -177.065464, -186.49986, -196.47981, -206.947665, -217.90019, -229.33415, -241.246311, -253.633438, -266.069205, -279.102236, -292.812076, -307.278269, -322.58036, -339.297013, -356.883872, -375.295702, -394.487267, -414.413331, -435.834793, -456.915009, -477.491233, -497.027466, -514.225642, -528.129849, -537.784173, -542.507328, -541.006358, -531.688712, -513.246879, -484.536787, -451.921731, -413.518422, -353.470191, -262.334379, -128.094649, 59.3589065, 300.699378, 601.937853, 910.679492, 1300.32684, 1797.10842, 2427.25275, 3216.98837, 4813.31265, 6466.49305, 8047.56587, 9427.56742, 10477.534, 10564.732, 10399.8142, 10069.7186, 9918.26184, 9930.70042, 9985.25782, 10072.5421, 10183.1614, 10307.7237, 10436.8371, 9815.9101, 9723.07286, 9979.21081, 10405.2094, 10821.954, 11050.3301, 10911.2232, 10776.2024, 10388.1803},
        {229.479402, 228.563985, 227.600691, 226.590305, 225.533613, 224.431399, 223.284448, 222.093546, 220.859478, 219.583028, 218.264983, 216.906126, 215.507243, 214.069119, 212.592539, 211.078288, 209.527151, 207.939914, 206.317361, 204.660277, 202.969448, 201.245658, 199.489693, 197.702338, 195.884377, 194.036597, 192.159781, 190.254715, 188.322184, 186.362973, 184.377868, 182.367652, 180.333112, 178.275033, 176.239632, 174.204024, 172.144773, 170.061192, 167.95259, 165.818278, 163.657567, 161.469768, 159.254191, 157.010146, 154.736946, 152.433899, 150.100317, 147.73551, 145.33879, 142.909466, 140.44685, 137.850342, 135.221328, 132.561285, 129.87169, 127.154019, 124.409749, 121.640355, 118.847316, 116.032107, 113.196204, 110.341085, 107.468226, 104.579102, 101.675192, 98.7579711, 95.828916, 92.8886438, 89.9377922, 86.9796104, 84.0156304, 81.0473847, 78.0764053, 75.1042245, 72.1323746, 69.1623878, 66.1957963, 63.2341324, 60.2789282, 57.3317161, 54.3940282, 51.4673967, 48.553354, 45.6534323, 43.0112018, 40.5020848, 37.9985944, 35.4944083, 32.9832041, 30.4586594, 27.914452, 25.3442593, 22.741759, 20.1006288, 17.4145463, 14.6771891, 11.8822348, 9.02336116, 6.09424572, 3.08856612, 0, -4.46841451, -8.81932051, -13.0689153, -17.233396, -21.3289601, -25.3718047, -29.3781271, -33.3641246, -37.3459945, -41.339934, -45.3621404, -49.428811, -53.556143, -57.7603337, -62.0575804, -66.4640803, -70.9486923, -75.5208508, -80.3038285, -85.3036785, -90.5264543, -95.978209, -101.664996, -107.592869, -113.76788, -120.196084, -126.883533, -133.836281, -141.060381, -148.443803, -156.153981, -164.232394, -172.720519, -181.852204, -191.572744, -201.776133, -212.446142, -223.566537, -235.121087, -247.093561, -258.884139, -271.237791, -284.252462, -298.026101, -312.656656, -329.353554, -346.825393, -364.892253, -383.374211, -402.091347, -421.336848, -439.879437, -457.500189, -473.734487, -487.616088, -498.403969, -505.357105, -509.361292, -507.678957, -495.797818, -469.039771, -429.522732, -385.976974, -337.560341, -267.210719, -165.311224, -19.7234135, 179.823336, 434.353946, 750.123232, 1085.22463, 1501.48377, 2021.82671, 2669.17952, 3466.46826, 5000.49127, 6589.33425, 8114.95519, 9459.31208, 10504.3629, 10672.8369, 10612.0733, 10400.8424, 10353.4015, 10455.3072, 10593.997, 10759.9334, 10943.5792, 11135.3973, 11325.8504, 10740.5853, 10690.3096, 10992.4651, 11464.4934, 11923.8362, 12187.9353, 12074.2322, 11941.0787, 11546.0614},
        {218.388245, 217.46676, 216.511753, 215.52343, 214.501998, 213.447661, 212.360627, 211.241101, 210.089289, 208.905398, 207.689633, 206.4422, 205.163305, 203.853155, 202.511955, 201.139912, 199.737231, 198.304119, 196.840781, 195.347423, 193.824252, 192.271473, 190.689293, 189.077918, 187.437553, 185.768404, 184.070678, 182.344581, 180.590319, 178.808097, 176.998121, 175.160599, 173.295735, 171.403736, 169.503484, 167.585456, 165.6399, 163.666417, 161.664605, 159.634067, 157.5744, 155.485205, 153.366083, 151.216632, 149.036453, 146.825146, 144.58231, 142.307546, 140.000453, 137.660631, 135.287681, 132.797638, 130.275477, 127.72261, 125.140449, 122.530404, 119.893887, 117.23231, 114.547083, 111.839619, 109.111329, 106.363623, 103.597914, 100.815614, 98.018132, 95.2068812, 92.3832724, 89.5457352, 86.6927706, 83.8319401, 80.9648484, 78.0931001, 75.2183001, 72.342053, 69.4659635, 66.5916363, 63.7206762, 60.8546879, 57.9952761, 55.1440454, 52.3026006, 49.4725465, 46.6554876, 43.8530288, 41.3148551, 38.9133129, 36.5177671, 34.1217717, 31.7188807, 29.3026481, 26.866628, 24.4043745, 21.9094414, 19.375383, 16.7957531, 14.1641059, 11.4739953, 8.71897536, 5.89260016, 2.9884237, 0, -4.38539769, -8.65529489, -12.8255222, -16.9119102, -20.9302895, -24.8964907, -28.8263445, -32.7356814, -36.640332, -40.5561269, -44.4988967, -48.4844721, -52.5286836, -56.6473619, -60.8563375, -65.171441, -69.5328422, -73.945563, -78.5965716, -83.4854857, -88.6119228, -93.9755005, -99.5758363, -105.412548, -111.485252, -117.793568, -124.337112, -131.115501, -138.128354, -145.204502, -152.57697, -160.296611, -168.41428, -177.215115, -186.632826, -196.53865, -206.913153, -217.736906, -228.990475, -240.654428, -252.237173, -264.335141, -277.021279, -290.368534, -304.449855, -320.385721, -336.939664, -353.922749, -371.146043, -388.420609, -406.091689, -422.783289, -437.88453, -450.843045, -461.225932, -468.546651, -472.318661, -474.473751, -471.557434, -457.477046, -426.523047, -381.109632, -328.757225, -268.596009, -185.539374, -71.0264496, 84.5368248, 289.979183, 550.340986, 872.805511, 1230.45262, 1669.81568, 2209.88302, 2869.64293, 3668.08373, 5125.54495, 6634.32587, 8088.07699, 9380.44881, 10405.0918, 10642.8472, 10675.381, 10573.262, 10620.0068, 10801.8922, 11016.467, 11254.4448, 11506.5387, 11763.4623, 12015.929, 11627.3566, 11673.0977, 12000.1697, 12455.5902, 12886.3765, 13139.5461, 13062.1166, 12940.7268, 12582.045},
        {205.34996, 204.301285, 203.257605, 202.217431, 201.179274, 200.141643, 199.10305, 198.062003, 197.017014, 195.966593, 194.90925, 193.843496, 192.767841, 191.680795, 190.580868, 189.466571, 188.336414, 187.188908, 186.022562, 184.835888, 183.627395, 182.395593, 181.138994, 179.856107, 178.545443, 177.205512, 175.834824, 174.43189, 172.99522, 171.523324, 170.014712, 168.467896, 166.881385, 165.253689, 163.507885, 161.681787, 159.814115, 157.905828, 155.957884, 153.971242, 151.946859, 149.885694, 147.788706, 145.656853, 143.491092, 141.292383, 139.061684, 136.799953, 134.508147, 132.187227, 129.838149, 127.467865, 125.071211, 122.649015, 120.202105, 117.73131, 115.237459, 112.721379, 110.183899, 107.625848, 105.048053, 102.451344, 99.8365491, 97.204496, 94.5560135, 91.8919299, 89.2130737, 86.50447, 83.7515233, 80.9876552, 78.2147185, 75.4345661, 72.6490509, 69.8600258, 67.0693437, 64.2788573, 61.4904197, 58.7058836, 55.927102, 53.1559277, 50.3942136, 47.6438125, 44.9065774, 42.1843612, 39.7368948, 37.4316679, 35.1330707, 32.8345875, 30.5297024, 28.2118998, 25.8746638, 23.5114786, 21.1158286, 18.6811979, 16.2010708, 13.6689315, 11.0782642, 8.42255315, 5.69528262, 2.88993683, 0, -4.17554892, -8.25891319, -12.2642416, -16.2056831, -20.0973864, -23.9535003, -27.7881738, -31.6155555, -35.4497945, -39.3050393, -43.195439, -47.1351423, -51.1382981, -55.2190551, -59.3915622, -63.6699683, -67.9540322, -72.2415612, -76.8054435, -81.6353158, -86.7208146, -92.0515768, -97.617239, -103.407438, -109.41181, -115.619993, -122.021622, -128.606334, -135.363767, -142.014307, -148.915565, -156.137954, -163.751883, -172.016262, -180.907252, -190.350754, -200.360628, -210.950738, -222.134947, -233.927117, -246.970366, -260.45779, -274.280136, -288.328151, -302.492584, -315.161621, -328.10421, -341.586739, -355.875595, -371.237166, -390.185649, -407.992298, -421.589326, -429.54062, -433.741409, -434.591219, -432.489572, -430.291207, -425.382432, -412.471146, -390.912181, -354.92582, -300.189589, -222.573536, -118.273914, 16.6627552, 181.300076, 378.323784, 628.34916, 941.849521, 1321.09204, 1786.01865, 2349.94327, 3026.1798, 3828.04215, 5197.18483, 6611.49599, 7977.20437, 9200.53874, 10187.7278, 10481.3556, 10593.7255, 10586.9141, 10707.6397, 10944.8917, 11210.7249, 11497.9752, 11799.4786, 12108.0707, 12416.5876, 12729.788, 13035.8547, 13314.3921, 13545.0047, 13707.297, 13780.8734, 13745.3383, 13662.5604, 13465.4794},
        {194.679794, 193.625398, 192.592763, 191.579663, 190.583873, 189.603167, 188.63532, 187.678105, 186.729297, 185.786671, 184.848, 183.911058, 182.973621, 182.033462, 181.088356, 180.136077, 179.174399, 178.201096, 177.213944, 176.210715, 175.189185, 174.147127, 173.082316, 171.992527, 170.875533, 169.729109, 168.551029, 167.339067, 166.090998, 164.804596, 163.477635, 162.107889, 160.693133, 159.231142, 157.6039, 155.869511, 154.087472, 152.259314, 150.38657, 148.47077, 146.513448, 144.516134, 142.480362, 140.407661, 138.299565, 136.157606, 133.983314, 131.778223, 129.543863, 127.281767, 124.993466, 122.724136, 120.430719, 118.113801, 115.773967, 113.411804, 111.027897, 108.622832, 106.197196, 103.751572, 101.286548, 98.8027095, 96.3006416, 93.7809303, 91.2441616, 88.690921, 86.1217945, 83.5157155, 80.8521374, 78.1763022, 75.4901992, 72.7958178, 70.0951474, 67.3901774, 64.6828972, 61.9752963, 59.2693639, 56.5670895, 53.8704625, 51.1814723, 48.5021083, 45.8343599, 43.1802164, 40.5416674, 38.1863874, 35.9779349, 33.7766749, 31.575975, 29.3692024, 27.1497246, 24.910909, 22.6461229, 20.3487339, 18.0121091, 15.6296162, 13.1946224, 10.7004952, 8.14060196, 5.50831008, 2.79698696, 0, -3.99374777, -7.91853163, -11.786486, -15.6097451, -19.4004434, -23.1707153, -26.9326951, -30.6985171, -34.4803158, -38.2902254, -42.1403804, -46.042915, -50.0099638, -54.0536609, -58.1861408, -62.4195379, -66.6954834, -71.0160397, -75.5528126, -80.3028286, -85.2631142, -90.4306961, -95.8026007, -101.375855, -107.147484, -113.114517, -119.273978, -125.622894, -132.158293, -138.661736, -145.424717, -152.508904, -159.975961, -168.028416, -176.657503, -185.816897, -195.526004, -205.804232, -216.670989, -228.145683, -241.197273, -254.606622, -268.207355, -281.833096, -295.317472, -306.445221, -317.611078, -329.160891, -341.440507, -354.795776, -371.923846, -387.94545, -400.286467, -407.57494, -410.89333, -410.222116, -405.541775, -397.817142, -385.82062, -367.250771, -345.585651, -309.935989, -249.572156, -158.577374, -39.1899717, 110.095051, 283.128204, 479.429991, 726.587373, 1036.31134, 1425.29034, 1899.49213, 2468.90472, 3143.51611, 3933.31433, 5224.50432, 6556.80293, 7846.14393, 9008.46111, 9959.68824, 10288.9621, 10456.8782, 10519.426, 10699.7798, 10987.7518, 11303.4945, 11640.3033, 11991.4737, 12350.3012, 12710.0814, 13338.0194, 13773.4265, 14052.7177, 14212.3082, 14288.6129, 14318.0471, 14337.0257, 14268.7911, 14139.4658},
        {186.065989, 185.10186, 184.158104, 183.232569, 182.323103, 181.427556, 180.543775, 179.669609, 178.802908, 177.941518, 177.08329, 176.226071, 175.36771, 174.506056, 173.638956, 172.764261, 171.879818, 170.983475, 170.073082, 169.146487, 168.201538, 167.236085, 166.247975, 165.235057, 164.195179, 163.126191, 162.025941, 160.892278, 159.723049, 158.516103, 157.26929, 155.980457, 154.647454, 153.268128, 151.730168, 150.08882, 148.400653, 146.667091, 144.889558, 143.069477, 141.208271, 139.307365, 137.36818, 135.392142, 133.380673, 131.335197, 129.257137, 127.147917, 125.00896, 122.84169, 120.64753, 118.464225, 116.25609, 114.023761, 111.767874, 109.489065, 107.187971, 104.865226, 102.521468, 100.157332, 97.7734547, 95.3704719, 92.9490199, 90.5097346, 88.0532522, 85.5802087, 83.0912403, 80.5654338, 77.9823936, 75.3871993, 72.7818839, 70.1684806, 67.5490222, 64.925542, 62.3000728, 59.6746478, 57.0512999, 54.4320623, 51.8189679, 49.2140497, 46.6193409, 44.0368745, 41.4686835, 38.9168009, 36.6551097, 34.544, 32.4405944, 30.3381039, 28.2297395, 26.1087122, 23.968233, 21.8015128, 19.6017626, 17.3621934, 15.0760162, 12.736442, 10.3366817, 7.86994644, 5.32944705, 2.70839458, 0, -3.87171929, -7.68995368, -11.4651561, -15.2077793, -18.9282764, -22.6371, -26.3447032, -30.0615388, -33.7980596, -37.5647186, -41.3719687, -45.2302627, -49.1500535, -53.141794, -57.215937, -61.3829356, -65.6694155, -70.0881402, -74.6129805, -79.257855, -84.0366822, -88.9633807, -94.051869, -99.3160657, -104.769889, -110.427258, -116.302091, -122.408307, -128.759823, -135.299344, -142.138115, -149.31142, -156.854542, -164.91161, -173.463484, -182.461999, -191.909785, -201.80947, -212.163686, -222.97506, -234.890869, -247.0729, -259.397656, -271.741639, -283.981354, -294.947963, -305.824645, -316.749237, -327.859578, -339.293506, -352.480618, -364.688177, -375.784699, -385.112611, -390.940236, -392.018149, -387.096923, -373.873967, -352.392376, -323.844157, -293.720633, -249.938777, -181.156861, -81.3246551, 46.6419152, 203.942537, 384.371019, 587.207003, 838.885028, 1150.47929, 1540.19025, 2010.86949, 2571.12439, 3229.56238, 3994.79084, 5216.37745, 6476.72927, 7699.21366, 8807.19795, 9724.04948, 10073.6765, 10278.5448, 10390.3614, 10618.7251, 10952.844, 11316.1379, 11701.1046, 12100.2416, 12506.0464, 12911.0165, 13500.0261, 13947.2568, 14274.9575, 14505.3768, 14660.7634, 14763.3661, 14835.4335, 14775.9727, 14636.3512},
        {178.669995, 177.838483, 177.016821, 176.203346, 175.396397, 174.594312, 173.795428, 172.998084, 172.200618, 171.401368, 170.598671, 169.790866, 168.97629, 168.153282, 167.32018, 166.475321, 165.617044, 164.743687, 163.853587, 162.945082, 162.016511, 161.066212, 160.092522, 159.093779, 158.068322, 157.014489, 155.930617, 154.815044, 153.666109, 152.482149, 151.261502, 150.002507, 148.703501, 147.362823, 145.898119, 144.349772, 142.75913, 141.12715, 139.454788, 137.743002, 135.992747, 134.204981, 132.380659, 130.520738, 128.626176, 126.697928, 124.73695, 122.7442, 120.720634, 118.667209, 116.58488, 114.480446, 112.348894, 110.191057, 108.007762, 105.799841, 103.568123, 101.313438, 99.0366164, 96.7384876, 94.4198816, 92.0816285, 89.7245581, 87.3495004, 84.9572853, 82.5487427, 80.1247026, 77.6673756, 75.1594194, 72.6400647, 70.1113487, 67.575308, 65.0339798, 62.4894009, 59.9436084, 57.398639, 54.8565298, 52.3193177, 49.7890397, 47.2677326, 44.7574335, 42.2601792, 39.7780067, 37.3129529, 35.1444658, 33.1300418, 31.1238432, 29.1189043, 27.1082596, 25.0849436, 23.0419906, 20.972435, 18.8693114, 16.725654, 14.5344973, 12.2888758, 9.98182385, 7.60637587, 5.15556629, 2.62242952, 0, -3.82398765, -7.59684398, -11.3282454, -15.0278685, -18.7053895, -22.3704851, -26.0328315, -29.7021054, -33.387983, -37.100141, -40.8482557, -44.6420035, -48.491061, -52.4051045, -56.3938106, -60.4668557, -64.7276679, -69.1993169, -73.6794216, -78.197748, -82.7840622, -87.4681304, -92.2797187, -97.2485932, -102.40452, -107.777265, -113.396595, -119.292276, -125.494073, -132.108565, -139.060541, -146.356725, -154.003839, -162.108412, -170.627263, -179.490598, -188.675197, -198.157841, -207.915311, -217.924388, -228.247514, -238.749738, -249.391081, -260.131564, -270.931209, -282.256264, -293.433966, -304.297777, -314.681161, -324.417581, -333.258068, -341.219273, -350.937199, -362.605556, -371.43169, -374.861725, -370.341783, -352.712471, -322.617588, -283.54331, -240.525507, -183.679748, -105.315763, -0.528599903, 130.867139, 291.223206, 477.70916, 691.33322, 952.104979, 1269.28577, 1654.92279, 2115.06294, 2658.00322, 3292.04065, 4025.47224, 5181.64895, 6375.05035, 7535.20996, 8591.6613, 9473.93789, 9834.3998, 10064.5363, 10212.6696, 10479.2246, 10852.3161, 11256.6827, 11683.608, 12124.3755, 12570.2688, 13012.5716, 13347.673, 13737.5138, 14145.062, 14533.2853, 14865.1517, 15103.629, 15211.685, 15180.2208, 14994.1635},
        {170.921216, 170.169985, 169.424261, 168.682627, 167.943669, 167.205968, 166.46811, 165.728678, 164.986256, 164.239427, 163.486776, 162.726887, 161.958342, 161.179727, 160.389625, 159.586619, 158.769294, 157.936233, 157.08602, 156.217239, 155.328474, 154.418309, 153.485328, 152.528113, 151.54525, 150.535322, 149.496913, 148.428606, 147.328986, 146.196636, 145.03014, 143.828083, 142.589047, 141.311616, 139.930934, 138.47864, 136.987134, 135.457058, 133.889057, 132.283772, 130.641845, 128.96392, 127.250639, 125.502645, 123.720581, 121.905088, 120.05681, 118.176389, 116.264467, 114.321688, 112.348695, 110.333292, 108.289238, 106.217453, 104.118859, 101.994375, 99.8449239, 97.6714255, 95.4748009, 93.2559709, 91.0158565, 88.7553785, 86.4754578, 84.1770152, 81.8609718, 79.5282483, 77.1797656, 74.7982176, 72.366736, 69.9238332, 67.4716118, 65.0121743, 62.547623, 60.0800604, 57.6115891, 55.1443114, 52.6803298, 50.2217468, 47.7706648, 45.3291863, 42.8994138, 40.4834496, 38.0833964, 35.7013565, 33.623784, 31.7046247, 29.7944066, 27.8860042, 25.9722923, 24.0461457, 22.1004389, 20.1280468, 18.1218439, 16.0747051, 13.9795049, 11.8291182, 9.61641956, 7.33428372, 4.97558538, 2.53319924, 0, -3.98218802, -7.86972709, -11.6753139, -15.4116453, -19.091418, -22.7273286, -26.332074, -29.9183508, -33.4988559, -37.0862859, -40.6933376, -44.3327078, -48.0170931, -51.7591903, -55.5716961, -59.4673073, -63.4542974, -67.5447312, -71.7613104, -76.1157839, -80.6199007, -85.2854096, -90.1240594, -95.1475992, -100.367778, -105.796344, -111.445047, -117.325635, -123.449858, -129.803061, -136.433077, -143.359576, -150.602229, -158.366473, -166.579094, -175.117341, -183.945153, -193.026469, -202.32523, -211.805376, -221.386365, -231.090155, -240.889389, -250.756708, -260.664754, -271.099735, -281.392337, -291.38681, -300.927408, -309.858379, -318.237543, -325.434558, -332.992673, -340.818361, -345.476336, -345.031697, -337.549538, -321.115502, -293.76947, -253.528911, -198.210119, -126.340896, -38.8195926, 66.2884625, 195.718341, 354.001652, 543.694908, 768.817656, 1040.63146, 1366.30141, 1751.85802, 2205.77545, 2735.6115, 3348.92394, 4053.27059, 5142.21016, 6265.79928, 7360.09551, 8361.1564, 9205.03951, 9575.10795, 9828.5767, 10009.801, 10304.6443, 10701.4482, 11129.5922, 11580.3659, 12045.0588, 12514.9604, 12981.3603, 12997.9212, 13325.7595, 13842.7781, 14426.8799, 14955.9678, 15307.9449, 15360.714, 15492.0341, 15291.5177},
        {163.185019, 162.475593, 161.772181, 161.073407, 160.377897, 159.684275, 158.991165, 158.297192, 157.60098, 156.901154, 156.196337, 155.485155, 154.766232, 154.038192, 153.29966, 152.54926, 151.785617, 151.007355, 150.213098, 149.401472, 148.5711, 147.720606, 146.848616, 145.953754, 145.034644, 144.08991, 143.118178, 142.118071, 141.088214, 140.027231, 138.933747, 137.806386, 136.643773, 135.444533, 134.14963, 132.787731, 131.388197, 129.951522, 128.478202, 126.968733, 125.423611, 123.84333, 122.228387, 120.579276, 118.896495, 117.180537, 115.4319, 113.651077, 111.838566, 109.994861, 108.120458, 106.196321, 104.242901, 102.261117, 100.251887, 98.2161311, 96.1547673, 94.068715, 91.9588931, 89.8262206, 87.6716165, 85.4959997, 83.3002892, 81.0854039, 78.852263, 76.6017852, 74.3348897, 72.0326809, 69.6767391, 67.3088492, 64.9312144, 62.5460383, 60.1555243, 57.7618759, 55.3672966, 52.9739898, 50.5841589, 48.2000075, 45.8237389, 43.4575567, 41.1036643, 38.7642651, 36.4415627, 34.1377604, 32.1472195, 30.3199193, 28.5023279, 26.6871676, 24.8671608, 23.0350299, 21.1834973, 19.3052853, 17.3931164, 15.4397129, 13.4377972, 11.3800916, 9.25931868, 7.06820065, 4.79945994, 2.44581893, 0, -4.1682464, -8.19207035, -12.0877428, -15.8715349, -19.5597176, -23.1685618, -26.7143387, -30.2133192, -33.6817743, -37.1359752, -40.5921926, -44.0666978, -47.5757617, -51.1356553, -54.7626496, -58.4730157, -62.1415327, -65.7642586, -69.6775053, -73.8672243, -78.3193668, -83.0198842, -87.9547278, -93.109849, -98.4711991, -104.02473, -109.756391, -115.652136, -121.697916, -127.688971, -133.87189, -140.289837, -146.985976, -154.271202, -162.054813, -170.174715, -178.593753, -187.274772, -196.180616, -205.274133, -214.630746, -224.066459, -233.522088, -242.938453, -252.256373, -261.184721, -269.954248, -278.563757, -287.012056, -295.297947, -304.579375, -312.279284, -317.575511, -319.688749, -317.927176, -311.559689, -299.855187, -285.615463, -263.77359, -225.408575, -160.4684, -76.0825731, 18.4827024, 122.255999, 248.320921, 403.309585, 594.954942, 830.174506, 1111.8494, 1445.14395, 1831.24589, 2279.90795, 2797.67097, 3391.07582, 4066.66335, 5085.44973, 6135.88167, 7160.88118, 8103.37027, 8906.27098, 9286.35649, 9563.46354, 9777.43507, 10095.5612, 10507.2372, 10948.7549, 11412.1647, 11889.5168, 12372.8614, 12854.2488, 12607.1463, 12889.8696, 13506.0778, 14259.4302, 14953.5858, 15392.2038, 15378.9433, 15683.8736, 15499.6932},
        {155.621618, 154.914842, 154.220455, 153.536852, 152.862429, 152.195583, 151.534709, 150.878204, 150.224464, 149.571884, 148.91886, 148.263789, 147.605066, 146.941088, 146.270251, 145.59095, 144.901582, 144.200542, 143.486227, 142.757033, 142.011355, 141.24759, 140.464134, 139.659382, 138.831731, 137.979576, 137.101315, 136.195342, 135.260054, 134.293847, 133.295116, 132.262258, 131.193669, 130.087745, 128.874526, 127.588023, 126.26307, 124.900282, 123.500272, 122.063655, 120.591046, 119.083057, 117.540305, 115.963402, 114.352964, 112.709604, 111.033938, 109.326578, 107.58814, 105.819238, 104.020485, 102.18434, 100.319751, 98.4275077, 96.5084024, 94.563226, 92.5927695, 90.5978242, 88.5791811, 86.5376314, 84.4739661, 82.3889764, 80.2834535, 78.1581884, 76.0139723, 73.8515963, 71.6718515, 69.4517877, 67.1690246, 64.8733182, 62.5669986, 60.2523961, 57.9318408, 55.6076629, 53.2821926, 50.95776, 48.6366954, 46.3213289, 44.0139908, 41.7170111, 39.4327201, 37.1634479, 34.9115248, 32.6792809, 30.7697101, 29.0284868, 27.2976714, 25.5698369, 23.8375565, 22.0934031, 20.3299499, 18.5397699, 16.7154362, 14.8495218, 12.9345998, 10.9632433, 8.92802535, 6.82151896, 4.63629723, 2.36493323, 0, -4.21167421, -8.26038481, -12.1634844, -15.9383257, -19.6022613, -23.1726439, -26.666826, -30.1021603, -33.4959996, -36.8656963, -40.2286032, -43.6020728, -47.0034579, -50.4501111, -53.9593849, -57.5486321, -61.0406777, -64.4250847, -68.1592085, -72.2187175, -76.5792797, -81.2165632, -86.1062363, -91.2239671, -96.5454237, -102.046274, -107.702187, -113.48883, -119.381872, -125.108665, -130.984242, -137.058765, -143.382397, -150.249014, -157.586921, -165.259434, -173.243601, -181.516468, -190.055085, -198.836499, -208.377399, -217.950954, -227.42863, -236.681893, -245.582209, -252.739585, -259.602311, -266.357217, -273.191134, -280.290891, -289.831371, -297.581514, -301.214438, -299.45957, -293.202958, -282.362371, -266.855579, -251.555652, -230.298852, -191.515657, -122.512131, -33.2415201, 64.2113021, 167.947879, 293.330071, 447.796971, 641.533717, 882.691342, 1169.35203, 1505.29348, 1890.22848, 2332.625, 2837.66782, 3410.54171, 4056.43146, 4999.69298, 5971.54713, 6922.38589, 7802.60127, 8562.58523, 8956.38347, 9261.63184, 9512.87075, 9854.44357, 10277.427, 10727.1467, 11196.9668, 11680.2518, 12170.3658, 12660.6732, 12348.7517, 12634.1046, 13297.6371, 14120.2547, 14882.8626, 15366.3662, 15351.6709, 15723.2965, 15602.1655},
        {148.24968, 147.495725, 146.769026, 146.06735, 145.388461, 144.730123, 144.0901, 143.466157, 142.856059, 142.25757, 141.668454, 141.086477, 140.509402, 139.934995, 139.361019, 138.785239, 138.205421, 137.619327, 137.024723, 136.419373, 135.801042, 135.167495, 134.516495, 133.845808, 133.153197, 132.436428, 131.693264, 130.921471, 130.118813, 129.283054, 128.411959, 127.503292, 126.554819, 125.564303, 124.425001, 123.189129, 121.910161, 120.589252, 119.227559, 117.826237, 116.386443, 114.909332, 113.396061, 111.847786, 110.265661, 108.650845, 107.004491, 105.327758, 103.621799, 101.887773, 100.126833, 98.3715672, 96.5910192, 94.7856641, 92.9559764, 91.102431, 89.2255024, 87.3256654, 85.4033946, 83.4591647, 81.4934503, 79.5067262, 77.4994669, 75.4721472, 73.4252418, 71.3592253, 69.2745723, 67.1407596, 64.9280079, 62.7007232, 60.4613895, 58.212491, 55.9565118, 53.6959362, 51.4332482, 49.170932, 46.9114718, 44.6573516, 42.4110557, 40.1750682, 37.9518733, 35.7439551, 33.5537977, 31.3838853, 29.5466474, 27.8830763, 26.2304388, 24.5811606, 22.9276677, 21.2623858, 19.5777408, 17.8661585, 16.1200647, 14.3318852, 12.494046, 10.5989727, 8.6390912, 6.60682737, 4.49460701, 2.29485595, 0, -3.9474399, -7.78057898, -11.5124515, -15.1560918, -18.7245341, -22.2308127, -25.6879619, -29.1090159, -32.5070091, -35.8949758, -39.2859501, -42.6929665, -46.1290592, -49.6072624, -53.1406105, -56.7421377, -60.3744991, -64.043532, -67.8762235, -71.8748126, -76.0415379, -80.3786381, -84.8883521, -89.5729184, -94.4345759, -99.4755632, -104.698119, -110.104482, -115.696892, -121.430515, -127.37192, -133.537469, -139.943522, -146.616316, -153.567273, -160.80518, -168.343433, -176.195429, -184.374566, -192.894238, -203.072873, -213.221655, -223.098649, -232.461921, -241.069536, -246.351047, -250.975158, -255.282063, -259.611956, -264.305028, -271.81163, -277.782719, -281.064998, -280.582254, -275.415651, -264.575694, -247.072887, -223.450967, -190.838745, -144.692411, -80.604221, 0.347299362, 94.2468368, 202.178081, 333.695253, 494.464462, 691.68693, 931.426759, 1214.1193, 1543.38384, 1923.81443, 2357.90565, 2848.93935, 3400.19741, 4014.9617, 4873.93554, 5758.62399, 6627.95356, 7440.85077, 8156.24211, 8571.86121, 8915.2894, 9214.53586, 9585.06888, 10020.2784, 10477.1131, 10950.785, 11436.506, 11929.4883, 12424.9437, 12407.7012, 12780.6978, 13396.9952, 14109.6553, 14771.7398, 15236.3107, 15356.4297, 15574.5559, 15590.8178},
        {141.68755, 140.938732, 140.222888, 139.537567, 138.880316, 138.248683, 137.640217, 137.052464, 136.482974, 135.929294, 135.388972, 134.859555, 134.338593, 133.823633, 133.312223, 132.80191, 132.290243, 131.77477, 131.253039, 130.722597, 130.180993, 129.625774, 129.054489, 128.464686, 127.853912, 127.219715, 126.559644, 125.871246, 125.152069, 124.399661, 123.611571, 122.785346, 121.918534, 121.008683, 119.938722, 118.76592, 117.548922, 116.288996, 114.987409, 113.645429, 112.264324, 110.84536, 109.389807, 107.89893, 106.373998, 104.816278, 103.227038, 101.607545, 99.9590673, 98.2828718, 96.5802263, 94.8909721, 93.1769667, 91.4386415, 89.6764279, 87.8907572, 86.0820608, 84.25077, 82.3973162, 80.5221307, 78.6256449, 76.7082902, 74.7704977, 72.8126991, 70.8353255, 68.8388083, 66.8235789, 64.7575877, 62.6095651, 60.4469319, 58.2722249, 56.087981, 53.8967374, 51.7010308, 49.5033983, 47.3063767, 45.1125032, 42.9243145, 40.7443477, 38.5751398, 36.4192276, 34.2791481, 32.1574382, 30.056635, 28.2960305, 26.7131216, 25.1415979, 23.5737172, 22.001737, 20.4179151, 18.814509, 17.1837764, 15.5179749, 13.8093621, 12.0501958, 10.2327335, 8.3492329, 6.3919516, 4.35314726, 2.22507751, 0, -3.75096184, -7.42177719, -11.0224829, -14.5631159, -18.053713, -21.5043111, -24.924947, -28.3256577, -31.7164799, -35.1074506, -38.5086066, -41.9299848, -45.381622, -48.8735552, -52.4158211, -56.0184567, -59.7190904, -63.5317008, -67.403915, -71.3516826, -75.3909529, -79.5376751, -83.8077987, -88.217273, -92.7820474, -97.5180711, -102.441294, -107.567664, -112.913132, -118.552004, -124.420474, -130.516985, -136.83998, -143.373002, -150.121942, -157.096667, -164.300088, -171.735118, -179.404668, -187.311652, -196.557437, -205.712166, -214.563839, -222.900451, -230.509999, -235.666854, -240.051048, -243.828982, -247.167062, -250.231692, -254.676794, -257.363174, -258.238376, -256.562996, -250.195097, -237.622454, -217.332838, -186.947104, -146.016969, -95.0398732, -36.1023132, 37.0472045, 127.415669, 238.655351, 375.511913, 542.229132, 744.032534, 985.420423, 1267.29137, 1592.58015, 1966.54012, 2389.43079, 2863.38436, 3390.53302, 3973.00898, 4758.74424, 5567.62124, 6365.32222, 7117.52942, 7789.92509, 8213.35782, 8578.23262, 8908.01758, 9300.12954, 9748.76522, 10217.3094, 10701.4993, 11197.0724, 11699.7659, 12205.3172, 12506.713, 12968.5986, 13521.9291, 14097.66, 14626.7464, 15040.1437, 15268.8072, 15426.1348, 15358.9927},
        {135.976907, 135.300996, 134.652172, 134.028292, 133.427212, 132.84679, 132.284882, 131.739345, 131.208035, 130.68881, 130.179526, 129.678041, 129.18221, 128.689891, 128.198941, 127.707216, 127.212573, 126.71287, 126.205962, 125.689706, 125.16196, 124.620581, 124.063424, 123.488347, 122.893207, 122.27586, 121.634164, 120.965974, 120.269149, 119.541544, 118.781016, 117.985423, 117.152621, 116.280467, 115.274324, 114.180243, 113.045382, 111.8706, 110.656756, 109.404709, 108.115315, 106.789435, 105.427925, 104.031646, 102.601454, 101.138208, 99.6427676, 98.11599, 96.5587337, 94.9718573, 93.3562192, 91.7207525, 90.0580657, 88.3688422, 86.6537652, 84.9135182, 83.1487844, 81.3602471, 79.5485898, 77.7144957, 75.8586481, 73.9817304, 72.0844259, 70.167418, 68.2313899, 66.2770251, 64.3050067, 62.2881345, 60.1998776, 58.0984272, 55.9862745, 53.8659102, 51.7398252, 49.6105106, 47.4804572, 45.3521559, 43.2280977, 41.1107734, 39.002674, 36.9062905, 34.8241136, 32.7586344, 30.7123438, 28.6877326, 27.0082959, 25.5092706, 24.0220356, 22.5386645, 21.0512311, 19.5518091, 18.0324721, 16.4852939, 14.9023483, 13.2757089, 11.5974495, 9.85964368, 8.0543653, 6.17368802, 4.20968557, 2.15443166, 0, -3.69846337, -7.32007781, -10.8744963, -14.3713719, -17.8203575, -21.2311061, -24.6132707, -27.9765043, -31.3304599, -34.6847904, -38.0491489, -41.4331884, -44.8465617, -48.298922, -51.7999221, -55.3592152, -58.9837431, -62.6827717, -66.4720858, -70.3607575, -74.3578588, -78.4724619, -82.7136387, -87.0904614, -91.612002, -96.2873326, -101.125525, -106.135652, -111.326785, -116.724154, -122.314749, -128.102795, -134.092516, -140.483405, -147.182054, -154.042981, -161.011829, -168.034244, -175.05587, -182.02235, -188.814197, -195.46201, -201.924175, -208.15908, -214.125112, -220.819072, -226.90133, -232.07067, -236.025874, -238.465726, -239.429008, -237.858953, -234.109532, -227.750847, -216.752604, -199.803051, -175.590442, -141.700307, -98.1742334, -46.2567737, 10.0096603, 79.629272, 168.463968, 281.275037, 420.958836, 591.267754, 797.562408, 1044.01214, 1328.88945, 1653.80244, 2019.70721, 2428.96415, 2883.40702, 3384.86961, 3935.18568, 4662.94459, 5411.53565, 6151.10376, 6851.79378, 7483.75062, 7897.407, 8262.42807, 8600.2639, 9003.58137, 9465.64058, 9950.02801, 10451.4875, 10964.7627, 11484.5976, 12005.7359, 12524.3225, 13044.6811, 13549.874, 14022.9635, 14447.0119, 14805.0813, 15080.2341, 15310.6306, 14964.0243},
        {130.320679, 129.732151, 129.164014, 128.614443, 128.081613, 127.563697, 127.05887, 126.565307, 126.081182, 125.604669, 125.133944, 124.66718, 124.202552, 123.738234, 123.272401, 122.803228, 122.328889, 121.847557, 121.357409, 120.856618, 120.343358, 119.815805, 119.272132, 118.710515, 118.129127, 117.526143, 116.899737, 116.248085, 115.56936, 114.861736, 114.12339, 113.352494, 112.547223, 111.705752, 110.751864, 109.722494, 108.65547, 107.551382, 106.410818, 105.234367, 104.022616, 102.776156, 101.495573, 100.181458, 98.8343985, 97.4549831, 96.0438006, 94.6014396, 93.1284886, 91.6255362, 90.0931712, 88.525338, 86.9294133, 85.3061298, 83.6562201, 81.9804168, 80.2794526, 78.5540602, 76.8049721, 75.032921, 73.2386396, 71.4228604, 69.5863161, 67.7297394, 65.853863, 63.9594193, 62.0471411, 60.0899361, 58.0613801, 56.0195926, 53.9671098, 51.9064684, 49.8402046, 47.7708549, 45.7009557, 43.6330434, 41.5696545, 39.5133253, 37.4665922, 35.4319918, 33.4120603, 31.4093343, 29.4263501, 27.4656441, 25.8565807, 24.4314079, 23.0184457, 21.6096241, 20.1968735, 18.7721241, 17.327306, 15.8543496, 14.3451849, 12.7917423, 11.185952, 9.51974415, 7.78504902, 5.97379682, 4.07791777, 2.08934209, 0, -3.73371084, -7.37627948, -10.9382349, -14.430106, -17.8624219, -21.2457115, -24.5905037, -27.9073276, -31.2067121, -34.4991862, -37.7952788, -41.105519, -44.4404358, -47.810558, -51.2264147, -54.6985348, -58.1864913, -61.6935338, -65.3454489, -69.1418463, -73.082336, -77.1665277, -81.3940313, -85.7644566, -90.2774134, -94.9325116, -99.7293611, -104.667572, -109.746753, -114.934137, -120.273583, -125.774416, -131.445957, -137.556192, -143.985113, -150.543736, -157.163787, -163.776991, -170.315071, -176.709754, -182.545723, -188.207364, -193.694305, -199.006167, -204.142578, -210.909301, -217.048285, -222.107619, -225.635393, -227.179696, -226.471708, -222.652649, -215.402812, -204.176476, -187.96648, -165.972842, -137.395578, -101.14744, -56.7809961, -4.16219847, 53.5591673, 124.661965, 214.822486, 328.589222, 468.599673, 638.368517, 843.160605, 1086.94437, 1367.27091, 1685.32135, 2042.54383, 2439.81134, 2878.21254, 3358.83609, 3882.77068, 4560.63359, 5256.60271, 5944.38456, 6597.68563, 7190.21246, 7587.20882, 7943.15244, 8278.23384, 8682.80939, 9149.29566, 9640.87343, 10151.4211, 10674.8169, 11204.9393, 11735.6666, 12155.7615, 12649.9918, 13179.02, 13703.5082, 14184.1188, 14581.5142, 14856.3566, 15113.7516, 14791.9248},
        {124.827823, 124.304253, 123.797584, 123.306154, 122.828303, 122.362369, 121.90669, 121.459605, 121.019452, 120.584571, 120.1533, 119.723977, 119.294941, 118.86453, 118.431084, 117.992941, 117.548439, 117.095917, 116.633713, 116.160167, 115.673617, 115.172401, 114.654858, 114.119326, 113.564145, 112.987652, 112.388187, 111.764088, 111.113694, 110.435342, 109.727373, 108.988123, 108.215933, 107.40914, 106.497626, 105.515396, 104.497283, 103.443846, 102.355645, 101.233241, 100.077194, 98.888064, 97.6664112, 96.412796, 95.1277785, 93.8119189, 92.4657774, 91.0899142, 89.6848895, 88.2512636, 86.7895966, 85.3008291, 83.7851327, 82.2430594, 80.6751611, 79.0819897, 77.4640974, 75.8220359, 74.1563573, 72.4676135, 70.7563565, 69.0231383, 67.2685108, 65.4930259, 63.6972357, 61.881692, 60.0469469, 58.1608379, 56.191988, 54.2084202, 52.2128072, 50.2078216, 48.1961363, 46.1804239, 44.163357, 42.1476085, 40.135851, 38.1307572, 36.1349998, 34.1512515, 32.182185, 30.2304731, 28.2987884, 26.3898036, 24.8405189, 23.4794112, 22.130939, 20.7869252, 19.4391932, 18.0795659, 16.6998667, 15.2919185, 13.8475445, 12.3585678, 10.8168116, 9.21409908, 7.54225325, 5.7930973, 3.95845435, 2.03014754, 0, -3.79603718, -7.48203918, -11.0696278, -14.5704247, -17.9960517, -21.3581306, -24.6682832, -27.9381312, -31.1792964, -34.4034005, -37.6220654, -40.8469127, -44.0895643, -47.361642, -50.6747674, -54.0405624, -57.3888452, -60.7195513, -64.2293343, -67.9122867, -71.7625009, -75.7740693, -79.9410844, -84.2576386, -88.7178243, -93.3157341, -98.0454602, -102.901095, -107.876732, -112.92603, -118.09834, -123.399882, -128.83688, -134.554835, -140.490328, -146.542801, -152.676691, -158.856434, -165.04647, -171.211235, -177.59965, -183.805088, -189.736326, -195.302144, -200.411318, -205.796812, -210.337171, -213.735129, -215.693417, -215.914767, -214.962926, -210.627257, -201.318698, -186.086216, -165.28142, -138.67106, -106.021884, -68.0599261, -23.3746329, 30.4910388, 92.9304579, 169.265592, 263.006275, 377.581354, 516.282444, 682.464153, 880.803615, 1114.99831, 1383.89647, 1689.08932, 2036.99069, 2423.6273, 2848.92102, 3312.79373, 3815.16732, 4448.57273, 5097.17051, 5737.73025, 6347.02157, 6901.81407, 7280.07962, 7623.25073, 7950.4389, 8348.95158, 8810.90603, 9300.00456, 9809.78516, 10333.7859, 10865.5447, 11398.5996, 11508.9095, 11921.4453, 12524.6809, 13207.0901, 13857.1469, 14363.3251, 14614.0985, 14843.3321, 14835.0579},
        {120.365703, 119.812744, 119.28103, 118.76877, 118.274177, 117.795462, 117.330835, 116.878508, 116.436691, 116.003595, 115.577433, 115.156413, 114.738749, 114.32265, 113.906327, 113.487992, 113.065856, 112.63813, 112.203024, 111.75875, 111.303519, 110.835541, 110.353028, 109.854191, 109.337241, 108.800389, 108.241845, 107.659822, 107.052529, 106.418178, 105.75498, 105.061146, 104.334887, 103.574414, 102.702566, 101.756825, 100.775579, 99.7594866, 98.7092035, 97.625387, 96.5086941, 95.3597819, 94.1793072, 92.9679271, 91.7262986, 90.4550787, 89.1549244, 87.8264926, 86.4704403, 85.0874246, 83.6781025, 82.2565111, 80.8096373, 79.3378478, 77.8415096, 76.3209897, 74.776655, 73.2088724, 71.6180088, 70.0044313, 68.3685066, 66.7106018, 65.0310837, 63.3303194, 61.6086757, 59.8665196, 58.104218, 56.2848837, 54.3725241, 52.4443397, 50.5031127, 48.551625, 46.5926585, 44.6289952, 42.6634171, 40.6987061, 38.7376442, 36.7830134, 34.8375957, 32.9041729, 30.9855271, 29.0844402, 27.2036942, 25.3460711, 23.8557372, 22.5576016, 21.2724705, 19.9920472, 18.7080352, 17.4121378, 16.0960584, 14.7515003, 13.3701671, 11.9437621, 10.4639886, 8.92255002, 7.3111498, 5.6214913, 3.84527791, 1.97421301, 0, -3.72952917, -7.35811904, -10.8962804, -14.3545239, -17.7433604, -21.0733007, -24.3548554, -27.5985354, -30.8148514, -34.0143142, -37.2074346, -40.4047232, -43.6166909, -46.8538484, -50.1267064, -53.4457759, -56.7819624, -60.1401191, -63.6203391, -67.2246466, -70.9550655, -74.8136198, -78.8023334, -82.9232305, -87.1783348, -91.5696706, -96.0992616, -100.769132, -105.581306, -110.596185, -115.736011, -120.985293, -126.32854, -131.701145, -137.112177, -142.583804, -148.115274, -153.705832, -159.354724, -165.061197, -171.568302, -177.905104, -183.925324, -189.482679, -194.430891, -199.209192, -202.939408, -205.32888, -206.084946, -204.914948, -202.376927, -196.287775, -184.929876, -167.348121, -144.144196, -115.220819, -80.4807065, -41.3706635, 4.10160797, 59.6128636, 126.328434, 207.350938, 304.564147, 420.509036, 558.840174, 722.700974, 916.309664, 1143.08831, 1402.51799, 1696.30901, 2033.31484, 2407.00372, 2816.61331, 3261.38131, 3740.5454, 4332.22592, 4937.05724, 5534.5564, 6104.24041, 6625.62631, 6988.9477, 7322.5273, 7644.14645, 8035.54468, 8489.18655, 8970.88739, 9474.44421, 9993.65403, 10522.3139, 11054.2208, 11091.4288, 11487.7958, 12114.6724, 12843.4092, 13545.357, 14091.8665, 14354.2882, 14585.4656, 14827.8889},
        {117.105437, 116.413492, 115.755879, 115.130368, 114.534726, 113.966722, 113.424125, 112.904703, 112.406224, 111.926457, 111.46317, 111.014132, 110.577111, 110.149875, 109.730194, 109.315835, 108.904567, 108.494158, 108.082377, 107.666993, 107.245773, 106.816486, 106.376901, 105.924786, 105.457909, 104.974039, 104.470945, 103.946394, 103.398155, 102.823998, 102.221689, 101.588998, 100.923693, 100.223542, 99.3920303, 98.476051, 97.523631, 96.5355984, 95.5127814, 94.4560081, 93.3661066, 92.2439051, 91.0902317, 89.9059145, 88.6917818, 87.4486616, 86.1773821, 84.8787714, 83.5536576, 82.202869, 80.8272336, 79.4538378, 78.0566823, 76.636026, 75.1921279, 73.7252467, 72.2356416, 70.7235712, 69.1892947, 67.6330708, 66.0551586, 64.4558168, 62.8353045, 61.1938805, 59.5318038, 57.8493332, 56.1467277, 54.3846838, 52.5248476, 50.6490728, 48.760183, 46.8610017, 44.9543526, 43.0430593, 41.1299453, 39.2178341, 37.3095495, 35.407915, 33.5157542, 31.6358906, 29.7711479, 27.9243497, 26.0983194, 24.2958808, 22.8670036, 21.6336364, 20.4135557, 19.1983195, 17.9794858, 16.7486126, 15.497258, 14.21698, 12.8993367, 11.535886, 10.118186, 8.63779467, 7.08627012, 5.45517036, 3.73605339, 1.92047726, 0, -3.50308491, -6.94876431, -10.3437671, -13.6948221, -17.0086583, -20.2920044, -23.5515895, -26.7941424, -30.0263919, -33.255067, -36.4868965, -39.7286094, -42.9869344, -46.2686006, -49.5803367, -52.9288717, -56.4022843, -60.0189247, -63.6015163, -67.1742201, -70.761197, -74.386608, -78.0746142, -81.8493765, -85.7350559, -89.7558135, -93.9358101, -98.2992069, -102.870165, -107.935036, -113.159653, -118.48952, -123.870142, -129.002902, -133.955364, -138.86019, -143.736117, -148.601885, -153.476233, -158.3779, -164.029646, -169.531922, -174.765724, -179.612047, -183.951887, -189.370239, -193.618099, -196.150465, -196.422331, -193.888693, -187.803832, -178.068779, -164.414146, -146.444495, -123.507042, -95.0645476, -60.5797698, -20.5697689, 26.7966099, 84.5006675, 153.706155, 237.222535, 336.861086, 454.950777, 594.697767, 758.905571, 951.567664, 1175.79607, 1430.33962, 1716.41523, 2040.8726, 2398.42924, 2788.35223, 3209.90863, 3662.36553, 4213.64021, 4777.187, 5335.11041, 5869.51497, 6362.50521, 6717.10654, 7047.22201, 7368.84717, 7754.93682, 8198.94018, 8670.57345, 9164.47249, 9675.27317, 10197.6113, 10726.1229, 11050.13, 11528.3805, 12099.3689, 12701.5894, 13273.5364, 13753.7043, 14080.5875, 14364.6513, 14678.0166},
        {114.315772, 113.458865, 112.650949, 111.889304, 111.171206, 110.493936, 109.854771, 109.25099, 108.679872, 108.138696, 107.62474, 107.135283, 106.667603, 106.218979, 105.786689, 105.368013, 104.960229, 104.560615, 104.166451, 103.775014, 103.383583, 102.989438, 102.589856, 102.182117, 101.763498, 101.331279, 100.882738, 100.415153, 99.9258046, 99.4119697, 98.8709274, 98.2999563, 97.6963352, 97.0573425, 96.2665801, 95.3805559, 94.4571446, 93.4973136, 92.5020308, 91.4722637, 90.40898, 89.3131473, 88.1857334, 87.0277059, 85.8400325, 84.6236807, 83.3796184, 82.108813, 80.8122324, 79.4908441, 78.1456158, 76.8123068, 75.4563389, 74.0779256, 72.6772803, 71.2546165, 69.8101475, 68.3440869, 66.8566481, 65.3480445, 63.8184896, 62.2681969, 60.6973797, 59.1062515, 57.4950258, 55.863916, 54.2131355, 52.5026026, 50.6932027, 48.8682554, 47.0305864, 45.1830212, 43.3283855, 41.4695049, 39.6092051, 37.7503115, 35.89565, 34.0480461, 32.2103255, 30.3853137, 28.5758364, 26.7847192, 25.0147878, 23.2688678, 21.901387, 20.7319746, 19.5760335, 18.4249791, 17.2702269, 16.1031925, 14.9152911, 13.6979383, 12.4425496, 11.1405404, 9.7833261, 8.36232224, 6.86894424, 5.29460756, 3.63072765, 1.86871998, 0, -3.2861084, -6.55690629, -9.81544884, -13.0647912, -16.3079885, -19.548096, -22.7881687, -26.0312619, -29.2804307, -32.5387303, -35.8092158, -39.0949424, -42.3989652, -45.7243394, -49.0741202, -52.4513627, -56.0487257, -59.8963508, -63.5684274, -67.1086402, -70.5606738, -73.9682125, -77.3749409, -80.8245436, -84.3607051, -88.0271098, -91.8674424, -95.9253872, -100.244629, -105.297416, -110.541728, -115.892682, -121.265393, -126.214471, -130.835284, -135.319336, -139.689892, -143.970219, -148.183584, -152.353255, -157.047189, -161.578186, -165.862943, -169.818157, -173.360525, -179.156468, -183.685528, -186.17697, -185.860058, -181.96406, -172.500654, -159.404852, -144.026184, -126.55103, -104.791022, -77.6240031, -43.9278195, -2.79380353, 47.1622068, 107.557268, 178.924071, 264.379865, 366.103254, 486.399701, 627.789624, 792.694775, 984.860283, 1207.051, 1457.17938, 1735.90267, 2047.20029, 2387.89621, 2757.49769, 3155.512, 3581.44639, 4095.17467, 4620.74593, 5142.5758, 5645.07991, 6112.67389, 6460.23269, 6788.07308, 7110.41383, 7491.36612, 7925.39966, 8386.19565, 8869.26321, 9370.11148, 9884.24958, 10407.1866, 11070.5347, 11642.2611, 12141.2624, 12586.4351, 12996.6758, 13390.8811, 13787.9475, 14136.1871, 14451.4613},
        {111.077693, 110.134813, 109.247568, 108.413072, 107.628445, 106.890803, 106.197262, 105.544941, 104.930955, 104.352423, 103.806461, 103.290186, 102.800715, 102.335166, 101.890656, 101.464301, 101.053218, 100.654526, 100.26534, 99.8827777, 99.5039566, 99.1259934, 98.7460053, 98.3611093, 97.9684225, 97.565062, 97.1481447, 96.7147879, 96.2621086, 95.7872237, 95.2872505, 94.759306, 94.2005072, 93.6079712, 92.8643523, 92.0264064, 91.1522655, 90.2427611, 89.2987248, 88.3209883, 87.3103831, 86.267741, 85.1938934, 84.089672, 82.9559084, 81.7934342, 80.6030811, 79.3856805, 78.1420643, 76.8730639, 75.5795109, 74.2852481, 72.9675972, 71.6268909, 70.263462, 68.8776433, 67.4697675, 66.0401674, 64.5891759, 63.1171256, 61.6243493, 60.1111799, 58.57795, 57.0249925, 55.4526402, 53.8612257, 52.251082, 50.5850176, 48.8267421, 47.0539791, 45.2694937, 43.4760513, 41.6764173, 39.8733569, 38.0696354, 36.2680182, 34.4712705, 32.6821577, 30.9034451, 29.137898, 27.3882817, 25.6573615, 23.9479027, 22.2626706, 20.9574457, 19.8510605, 18.7581046, 17.6698873, 16.5777178, 15.4729055, 14.3467597, 13.1905896, 11.9957046, 10.7534138, 9.45502668, 8.09185246, 6.65520042, 5.13637987, 3.52670009, 1.81747037, 0, -3.34355867, -6.65677657, -9.94344972, -13.2073742, -16.4523459, -19.6821611, -22.9006156, -26.1115055, -29.3186269, -32.5257758, -35.7367481, -38.9553401, -42.1853476, -45.4305667, -48.6947935, -51.981824, -55.4238372, -59.0429696, -62.552627, -65.9841162, -69.3687439, -72.7378166, -76.1226412, -79.5545241, -83.0647723, -86.6846922, -90.4455907, -94.3787744, -98.5155499, -103.175923, -107.996645, -112.922413, -117.897924, -122.731474, -127.435961, -132.060655, -136.591172, -141.013129, -145.312143, -149.47383, -153.984082, -158.175983, -161.937271, -165.155685, -167.71896, -171.458039, -173.831654, -174.241739, -172.090232, -166.779068, -156.890968, -143.648346, -128.591829, -111.846445, -90.6470496, -63.5261249, -29.0161552, 14.9235631, 69.0570893, 133.523187, 207.939312, 295.772184, 399.43166, 521.144432, 662.826829, 826.537641, 1015.46933, 1231.97461, 1474.24938, 1742.84087, 2041.38602, 2366.867, 2718.7615, 3096.54724, 3499.70191, 3980.08848, 4471.70309, 4960.92713, 5434.14198, 5877.72905, 6216.14444, 6538.97835, 6859.15158, 7235.94866, 7663.93809, 8118.37013, 8594.7792, 9088.69969, 9595.66601, 10111.2126, 10839.531, 11419.7245, 11890.773, 12291.6559, 12661.353, 13038.8438, 13463.1081, 13853.9172, 14128.7453},
        {107.896638, 106.893608, 105.949962, 105.062774, 104.229118, 103.446068, 102.710698, 102.020081, 101.371292, 100.761404, 100.187491, 99.6466264, 99.135885, 98.6523402, 98.1930659, 97.7551358, 97.3356239, 96.9316041, 96.54015, 96.1583357, 95.7832349, 95.4119214, 95.0414693, 94.6689521, 94.291444, 93.9060185, 93.5097497, 93.0997114, 92.6729774, 92.2266215, 91.7577177, 91.2633397, 90.7405614, 90.1864566, 89.4894164, 88.7031419, 87.8826408, 87.0285139, 86.141362, 85.2217859, 84.2703864, 83.2877643, 82.2745206, 81.2312559, 80.158571, 79.0570669, 77.9273443, 76.7700041, 75.5856469, 74.3748738, 73.1382854, 71.8808767, 70.5987592, 69.2924384, 67.9624199, 66.6092091, 65.2333118, 63.8352335, 62.4154796, 60.9745559, 59.5129677, 58.0312208, 56.5298206, 55.0092727, 53.4700827, 51.9127562, 50.3377986, 48.7120907, 47.0033204, 45.2813421, 43.5488412, 41.8085029, 40.0630127, 38.3150558, 36.5673175, 34.8224832, 33.0832383, 31.352268, 29.6322576, 27.9258925, 26.2358581, 24.5648396, 22.9155223, 21.2905917, 20.0461304, 19.0006918, 17.9685818, 16.9410171, 15.9092148, 14.8643918, 13.7977648, 12.7005509, 11.5639669, 10.3792297, 9.13755623, 7.83016339, 6.44826807, 4.98308718, 3.4258376, 1.76773624, 0, -3.4801417, -6.89829272, -10.260312, -13.5720584, -16.8393909, -20.0681684, -23.2642499, -26.4334941, -29.5817601, -32.7149068, -35.838793, -38.9592777, -42.0822199, -45.2134784, -48.3589121, -51.52438, -54.730455, -57.9850981, -61.2608864, -64.5668318, -67.9119462, -71.3052416, -74.7557299, -78.2724231, -81.864333, -85.5404715, -89.3098507, -93.1814825, -97.1643787, -101.322541, -105.589829, -109.958757, -114.42184, -119.166639, -124.088146, -129.029341, -133.924228, -138.706806, -143.311079, -147.671047, -152.190664, -156.190952, -159.513965, -162.001759, -163.496387, -164.369009, -163.8003, -161.500038, -157.178001, -150.543969, -141.406976, -129.256232, -115.19531, -99.0300374, -77.9166749, -50.2010854, -14.2291312, 32.8662747, 91.2506501, 159.766295, 237.113527, 327.117414, 432.446793, 555.253004, 696.810528, 858.796338, 1043.69799, 1253.40262, 1486.8252, 1744.56191, 2030.69612, 2341.80037, 2677.26375, 3036.47537, 3418.82431, 3868.91824, 4329.62305, 4789.0232, 5235.20313, 5656.24729, 5984.73351, 6301.2573, 6617.69168, 6990.19214, 7413.1032, 7862.31838, 8333.10476, 8820.72945, 9320.45951, 9827.56204, 10512.2517, 11068.1016, 11529.6515, 11931.441, 12308.0099, 12693.8979, 13123.6449, 13546.3591, 13756.6125},
        {104.915139, 103.8649, 102.87628, 101.946368, 101.072252, 100.25102, 99.4797601, 98.7555603, 98.075509, 97.4366943, 96.8362045, 96.2711277, 95.738552, 95.2355657, 94.7592569, 94.3067138, 93.8750246, 93.4612775, 93.0625605, 92.675962, 92.2985701, 91.9274729, 91.5597587, 91.1925156, 90.8228318, 90.4477954, 90.0644947, 89.6700178, 89.2614529, 88.8358881, 88.3904118, 87.9221119, 87.4280767, 86.9053945, 86.2507059, 85.5134358, 84.7442158, 83.9433941, 83.1113183, 82.2483365, 81.3547964, 80.431046, 79.4774332, 78.4943058, 77.4820117, 76.4408989, 75.3713151, 74.2736083, 73.1481264, 71.9952172, 70.8152287, 69.5934245, 68.3455637, 67.0723212, 65.7743719, 64.4523906, 63.1070522, 61.7390316, 60.3490036, 58.9376432, 57.5056252, 56.0536245, 54.5823159, 53.0923745, 51.5844749, 50.0592922, 48.5175012, 46.9298996, 45.2680028, 43.5941046, 41.9108169, 40.2207513, 38.5265195, 36.8307331, 35.136004, 33.4449437, 31.7601639, 30.0842764, 28.4198928, 26.7696248, 25.1360841, 23.5218824, 21.9296314, 20.3619427, 19.1752346, 18.1877736, 17.213574, 16.2437655, 15.2694782, 14.2818422, 13.2719873, 12.2310436, 11.150141, 10.0204096, 8.83297931, 7.5789801, 6.24954197, 4.83579492, 3.32886891, 1.71989395, 0, -3.61782163, -7.14138195, -10.5787172, -13.9378637, -17.2268577, -20.4537354, -23.6265332, -26.7532872, -29.8420338, -32.9008092, -35.9376498, -38.9605917, -41.9776712, -44.9969246, -48.0263882, -51.0740983, -54.0439583, -56.9291285, -59.9731834, -63.1618451, -66.4808358, -69.9158775, -73.4526923, -77.0770025, -80.77453, -84.5309971, -88.3321258, -92.1636383, -96.0112566, -99.677937, -103.399182, -107.215543, -111.167571, -115.822613, -120.957824, -126.20988, -131.461295, -136.594584, -141.49226, -146.036836, -150.549155, -154.339998, -157.206121, -158.944277, -159.351221, -157.205334, -153.576337, -148.515576, -142.074399, -134.304153, -126.478683, -115.932673, -102.893179, -86.7066276, -64.921493, -35.8934937, 2.02165204, 52.0016641, 113.808877, 185.532785, 265.252996, 356.994874, 463.689308, 587.397342, 728.706123, 888.879341, 1069.58922, 1272.20538, 1496.59944, 1743.49038, 2017.76568, 2315.33449, 2635.47284, 2977.45676, 3340.56227, 3762.89416, 4195.19251, 4627.02615, 5047.96391, 5447.57464, 5765.36339, 6074.33862, 6385.52508, 6752.95677, 7170.63391, 7614.63624, 8079.86003, 8561.20153, 9053.55702, 9551.82277, 10124.5357, 10637.5195, 11104.282, 11538.3312, 11953.1748, 12362.3208, 12779.2772, 13223.0059, 13364.9356},
        {102.481655, 101.374448, 100.332033, 99.3514539, 98.4297565, 97.5639858, 96.7511869, 95.988405, 95.2726853, 94.6010729, 93.9706129, 93.3783505, 92.8213308, 92.296599, 91.8012001, 91.3321794, 90.886582, 90.461453, 90.0538376, 89.6607809, 89.279328, 88.9065242, 88.5394144, 88.175044, 87.8104579, 87.4427015, 87.0688197, 86.6858577, 86.2908608, 85.880874, 85.4529424, 85.0041112, 84.5314256, 84.0319307, 83.4038956, 82.695831, 81.957435, 81.1889581, 80.3906511, 79.5627645, 78.7055489, 77.819255, 76.9041334, 75.9604346, 74.9884093, 73.9883082, 72.9603817, 71.9048807, 70.8220555, 69.712157, 68.5754357, 67.3925141, 66.1836965, 64.9496589, 63.6910774, 62.4086282, 61.1029874, 59.7748311, 58.4248353, 57.0536763, 55.66203, 54.2505727, 52.8199805, 51.3709294, 49.9040956, 48.4201551, 46.9197841, 45.3735324, 43.7526728, 42.1200151, 40.4781882, 38.8298213, 37.1775434, 35.5239837, 33.8717712, 32.223535, 30.5819041, 28.9495078, 27.328975, 25.7229348, 24.1340164, 22.5648487, 21.018061, 19.4962822, 18.3597072, 17.4246612, 16.503172, 15.5862649, 14.6649653, 13.7302986, 12.7732903, 11.7849656, 10.7563501, 9.67846902, 8.54234783, 7.3390119, 6.05948664, 4.69479742, 3.23596965, 1.67402871, 0, -3.5132274, -6.94834573, -10.3120406, -13.6109978, -16.8519028, -20.0414413, -23.1862989, -26.2931614, -29.3687143, -32.4196433, -35.452634, -38.4743721, -41.4915433, -44.5108331, -47.5389272, -50.5825113, -53.563662, -56.4769781, -59.5205224, -62.6828503, -65.9525168, -69.3180771, -72.7680863, -76.2910997, -79.8756723, -83.5103593, -87.183716, -90.8842974, -94.6006587, -98.1697268, -101.787282, -105.487368, -109.304029, -113.812858, -118.777124, -123.815682, -128.80011, -133.601988, -138.092895, -142.144408, -145.993361, -149.034914, -151.069182, -151.896282, -151.316328, -147.523357, -142.325084, -135.923146, -128.519178, -120.314817, -113.513156, -103.868145, -90.1525567, -71.2232134, -46.1085428, -13.7599255, 26.8712581, 77.3880207, 138.15859, 208.9464, 288.901543, 381.240132, 489.009722, 614.030402, 756.042394, 915.740613, 1093.75371, 1290.75942, 1507.67844, 1745.29406, 2009.03083, 2294.31668, 2600.32828, 2926.24229, 3271.2354, 3667.81599, 4073.49606, 4479.11936, 4875.5296, 5253.57053, 5558.42373, 5857.03651, 6159.82208, 6518.70542, 6927.668, 7363.39476, 7820.73202, 8294.52608, 8779.62327, 9270.8699, 9692.81669, 10162.8383, 10658.6961, 11158.1515, 11638.9658, 12078.9004, 12455.7166, 12901.0063, 13032.1763},
        {100.344301, 99.1818033, 98.0870775, 97.0571242, 96.0889439, 95.1795371, 94.3259045, 93.5250464, 92.7739636, 92.0696564, 91.4091254, 90.7893711, 90.2073941, 89.6601949, 89.144774, 88.658132, 88.1972693, 87.7591866, 87.3408843, 86.9393629, 86.551623, 86.1746652, 85.8054899, 85.4410977, 85.078489, 84.7146646, 84.3466247, 83.9713701, 83.5859012, 83.1872185, 82.7723226, 82.3382141, 81.8818933, 81.4003609, 80.7920686, 80.105364, 79.3897796, 78.6455139, 77.8727655, 77.0717332, 76.2426155, 75.3856111, 74.5009186, 73.5887366, 72.6492637, 71.6826986, 70.68924, 69.6690863, 68.6224364, 67.5494887, 66.450442, 65.3057724, 64.1358286, 62.9412367, 61.7226228, 60.4806132, 59.215834, 57.9289114, 56.6204715, 55.2911406, 53.9415448, 52.5723103, 51.1840632, 49.7774297, 48.353036, 46.9115083, 45.4534727, 43.9479979, 42.3649101, 40.7699207, 39.1657015, 37.5549245, 35.9402615, 34.3243845, 32.7099654, 31.0996762, 29.4961887, 27.9021748, 26.3203066, 24.7532558, 23.2036944, 21.6742944, 20.1677277, 18.6866661, 17.5959015, 16.7094292, 15.836892, 14.9692103, 14.0973047, 13.2120954, 12.304503, 11.3654479, 10.3858504, 9.35663111, 8.26871038, 7.11300864, 5.88044633, 4.56194388, 3.14842172, 1.63080028, 0, -3.36328059, -6.67381515, -9.93629037, -13.155393, -16.3358097, -19.4822271, -22.5993321, -25.6918112, -28.7643513, -31.821639, -34.8683609, -37.9092039, -40.9488547, -43.9919998, -47.0433261, -50.1075202, -53.1493588, -56.1678272, -59.2578847, -62.4156661, -65.6373058, -68.9189383, -72.2566983, -75.6467203, -79.0851388, -82.5680883, -86.0917034, -89.6521187, -93.2454686, -96.8032708, -100.40997, -104.081085, -107.832136, -112.155087, -116.827235, -121.498826, -126.042447, -130.330684, -134.236124, -137.631353, -140.681708, -142.877926, -144.035318, -143.969193, -142.494859, -137.559082, -131.312851, -124.038612, -116.01881, -107.535891, -101.480994, -92.3394652, -77.3825848, -55.0123308, -25.9391916, 10.1628181, 53.6196839, 104.048074, 162.6445, 231.379272, 310.829229, 403.21476, 511.669768, 637.761105, 780.400361, 939.717934, 1115.3244, 1307.21538, 1517.29254, 1746.37937, 2000.34172, 2274.18498, 2566.98724, 2877.82655, 3205.78099, 3578.25723, 3958.92259, 4339.77297, 4712.80429, 5070.01247, 5361.61926, 5649.24417, 5942.78635, 6292.22991, 6691.60701, 7118.24772, 7567.01928, 8032.78896, 8510.42402, 8994.7917, 9253.32854, 9679.32028, 10210.3618, 10784.0479, 11337.9735, 11809.7334, 12136.9225, 12573.8648, 12711.616},
        {98.4644962, 97.2476911, 96.1015916, 95.0231471, 94.0093071, 93.0570211, 92.1632385, 91.3249088, 90.5389815, 89.802406, 89.1121319, 88.4651085, 87.8582854, 87.2886121, 86.7530379, 86.2485125, 85.7719852, 85.3204055, 84.8907229, 84.4798868, 84.0848468, 83.7025523, 83.3299527, 82.9639976, 82.6016364, 82.2398186, 81.8754937, 81.5056111, 81.1271203, 80.7369707, 80.3321119, 79.9094933, 79.4660644, 78.9987746, 78.404904, 77.7333332, 77.0341407, 76.3075105, 75.5536265, 74.7726726, 73.9648329, 73.1302911, 72.2692314, 71.3818375, 70.4682935, 69.5287832, 68.5634907, 67.5725998, 66.5562945, 65.5147588, 64.4481765, 63.3402171, 62.207937, 61.0518782, 59.8725826, 58.6705922, 57.4464489, 56.2006946, 54.9338713, 53.6465209, 52.3391854, 51.0124066, 49.6667266, 48.3026872, 46.9208305, 45.5216982, 44.1058325, 42.6400534, 41.0919906, 39.5317349, 37.9620142, 36.3855565, 34.8050897, 33.223342, 31.6430413, 30.0669155, 28.4976927, 26.9381008, 25.3908678, 23.8587218, 22.3443907, 20.8506024, 19.3800851, 17.9355666, 16.886778, 16.0452262, 15.2180074, 14.3959396, 13.5698407, 12.7305288, 11.868822, 10.9755383, 10.0414957, 9.05751233, 8.01440611, 6.90299513, 5.71409742, 4.43853102, 3.06711396, 1.59066427, 0, -3.20846442, -6.39090225, -9.54991527, -12.6881053, -15.808074, -18.9124232, -22.0037547, -25.0846703, -28.1577716, -31.2256606, -34.2909389, -37.3562084, -40.4240708, -43.4971278, -46.5779813, -49.6692331, -52.7814756, -55.9184524, -59.0646905, -62.2245042, -65.4022074, -68.6021144, -71.8285391, -75.0857957, -78.3781984, -81.710061, -85.0856979, -88.509423, -91.9855505, -95.5563064, -99.1741917, -102.832147, -106.523113, -110.611644, -114.904874, -119.110838, -123.110994, -126.786796, -130.019702, -132.691168, -134.916285, -136.271768, -136.593361, -135.716807, -133.477853, -127.885805, -121.059454, -113.291156, -104.873264, -96.098132, -90.2184849, -80.9480773, -64.3023947, -38.3442904, -5.31666128, 34.4143501, 80.4826011, 130.570243, 186.706408, 253.049366, 331.593899, 423.606302, 532.297933, 659.011126, 801.921261, 960.657404, 1133.95929, 1321.22541, 1525.11514, 1746.44334, 1991.30975, 2254.43123, 2534.79323, 2831.38118, 3143.18054, 3492.78016, 3849.66121, 4206.9083, 4557.60601, 4894.83893, 5173.60866, 5450.47146, 5734.83319, 6074.65215, 6464.07389, 6881.18042, 7320.92046, 7778.24271, 8248.09588, 8725.42869, 8835.27002, 9222.08476, 9786.18519, 10427.8835, 11047.4921, 11545.323, 11821.6887, 12243.487, 12401.0133},
        {96.7807852, 95.5062055, 94.30594, 93.1768516, 92.1158028, 91.1196564, 90.1852749, 89.3095212, 88.4892577, 87.7213472, 87.0026524, 86.3300359, 85.7003603, 85.1104885, 84.5572829, 84.0376063, 83.5483213, 83.0862906, 82.6483769, 82.2314429, 81.8323511, 81.4479643, 81.0751451, 80.7107562, 80.3516602, 79.9947199, 79.6367978, 79.2747567, 78.9054592, 78.5257679, 78.1325456, 77.722655, 77.2929585, 76.840319, 76.2584844, 75.5980436, 74.9108248, 74.197037, 73.456889, 72.69059, 71.8983487, 71.0803742, 70.2368753, 69.368061, 68.4741402, 67.5553218, 66.6118149, 65.6438283, 64.651571, 63.6352519, 62.5950799, 61.520134, 60.4219945, 59.3011114, 58.157935, 56.9929156, 55.8065032, 54.5991483, 53.3713008, 52.1234111, 50.8559294, 49.5693059, 48.2639907, 46.9404342, 45.5990865, 44.2403977, 42.8648183, 41.4370112, 39.9224992, 38.3955402, 36.8589045, 35.3153621, 33.7676832, 32.2186379, 30.6709965, 29.1275289, 27.5910053, 26.0641959, 24.5498709, 23.0508002, 21.5697542, 20.1095028, 18.6728163, 17.2624648, 16.2524794, 15.4521907, 14.6664672, 13.8860309, 13.1016037, 12.3039077, 11.4836649, 10.6315972, 9.73842662, 8.79487524, 7.79166501, 6.71951794, 5.56915603, 4.33130129, 2.9966757, 1.55600127, 0, -3.226511, -6.42032964, -9.58443916, -12.7218228, -15.8354638, -18.9283455, -22.0034509, -25.0637635, -28.1122665, -31.151943, -34.1857764, -37.2167498, -40.2478466, -43.2820499, -46.3223431, -49.3717093, -52.3899841, -55.3739868, -58.4242966, -61.5346509, -64.6987868, -67.9104418, -71.1633532, -74.4512582, -77.7678942, -81.1069984, -84.4623083, -87.8275611, -91.1964941, -94.5461188, -97.8930312, -101.235986, -104.573739, -108.287504, -112.184806, -115.971182, -119.530649, -122.747225, -125.504925, -127.687769, -129.352546, -130.157917, -129.954096, -128.591296, -125.919731, -120.585094, -113.94325, -106.145542, -97.3433141, -87.6879102, -79.8457808, -68.3791436, -49.5477313, -21.5800529, 13.2757982, 54.5764402, 101.878491, 152.719393, 209.133843, 275.359277, 353.311295, 444.414439, 552.260761, 678.248206, 820.056929, 977.073612, 1147.75372, 1331.2425, 1530.09969, 1744.95357, 1981.41232, 2234.35825, 2502.69587, 2785.32969, 3081.16422, 3407.33196, 3739.95243, 4073.37314, 4401.94161, 4720.00535, 4988.46199, 5257.40884, 5535.52907, 5867.86007, 6248.6948, 6657.17864, 7088.36192, 7537.29499, 7999.02819, 8468.61186, 8538.7192, 8912.76536, 9481.00159, 10133.6791, 10761.0493, 11253.3632, 11500.8722, 11912.7303, 12102.4933},
        {95.2597827, 93.9241211, 92.6670707, 91.4853722, 90.3757664, 89.3349941, 88.3597961, 87.4469132, 86.5930862, 85.7950559, 85.049563, 84.3533485, 83.7031529, 83.0957173, 82.5277823, 81.9960887, 81.4973773, 81.028389, 80.5858645, 80.1665446, 79.7671701, 79.3844818, 79.0152204, 78.6561269, 78.3039419, 77.9554063, 77.6072608, 77.2562463, 76.8991036, 76.5325733, 76.1533964, 75.7583136, 75.3440658, 74.9073936, 74.3364932, 73.6846608, 73.0064973, 72.302266, 71.5722301, 70.816653, 70.0357978, 69.229928, 68.3993066, 67.5441972, 66.6648628, 65.7615668, 64.8345725, 63.8841431, 62.9105419, 61.9140322, 60.8948773, 59.8486196, 58.7803455, 57.6904208, 56.5792109, 55.4470815, 54.2943983, 53.1215268, 51.9288327, 50.7166816, 49.4854391, 48.2354708, 46.9671423, 45.6808193, 44.3768674, 43.0556521, 41.7175392, 40.3255027, 38.8434145, 37.3487575, 35.844321, 34.3328948, 32.8172683, 31.300231, 29.7845726, 28.2730826, 26.7685505, 25.2737659, 23.7915183, 22.3245973, 20.8757925, 19.4478934, 18.0436896, 16.6659706, 15.6920643, 14.9296132, 14.1817581, 13.439134, 12.6923761, 11.9321193, 11.1489988, 10.3336495, 9.47670662, 8.56880516, 7.60058019, 6.56266676, 5.44569994, 4.24031479, 2.93714637, 1.52682976, 0, -3.42673661, -6.77898114, -10.0626751, -13.2837601, -16.4481775, -19.561869, -22.6307761, -25.6608403, -28.6580031, -31.628206, -34.5773907, -37.5114985, -40.4364711, -43.35825, -46.2827766, -49.2159926, -51.9667685, -54.5128928, -57.3060628, -60.3099906, -63.4883882, -66.8049676, -70.2234409, -73.7075201, -77.2209172, -80.7273443, -84.1905134, -87.5741365, -90.8419257, -93.7187564, -96.4947507, -99.2052715, -101.885682, -105.094289, -108.604983, -112.052204, -115.314432, -118.270146, -120.797826, -122.775951, -124.212337, -124.816766, -124.442411, -122.94245, -120.170056, -115.887742, -110.062011, -102.568705, -93.2836637, -82.082728, -70.1347103, -54.4421814, -33.0837451, -4.96934418, 29.2937631, 69.8603785, 116.885304, 169.725125, 229.514273, 298.257957, 376.18549, 466.028749, 572.01914, 695.883559, 835.105155, 989.115024, 1156.72054, 1337.1911, 1532.08308, 1741.6592, 1970.29127, 2213.50079, 2470.12811, 2739.01362, 3018.99767, 3321.06873, 3628.88203, 3938.24093, 4244.94875, 4544.80883, 4805.87994, 5070.20638, 5345.48204, 5672.8016, 6046.63778, 6447.54631, 6870.69578, 7311.25476, 7764.39183, 8225.27556, 8376.77937, 8766.83548, 9306.29159, 9905.99537, 10476.7945, 10929.5367, 11175.0695, 11583.284, 11812.823},
        {93.9037119, 92.5096718, 91.1983404, 89.9663403, 88.810294, 87.7268242, 86.7125535, 85.7641043, 84.8780993, 84.0511612, 83.2799123, 82.5609754, 81.8909731, 81.2665278, 80.6842623, 80.140799, 79.6327605, 79.1567695, 78.7094485, 78.2874201, 77.8873069, 77.5057314, 77.1393163, 76.7846841, 76.4384575, 76.0972589, 75.757711, 75.4164363, 75.0700575, 74.7151971, 74.3484777, 73.9665219, 73.5659523, 73.1433914, 72.5815128, 71.9363102, 71.2651473, 70.5683444, 69.846222, 69.0991006, 68.3273006, 67.5311425, 66.7109467, 65.8670336, 64.9997237, 64.1093375, 63.1961955, 62.2606179, 61.3029254, 60.3234384, 59.3224772, 58.3018891, 57.2604347, 56.1984014, 55.1160765, 54.0137474, 52.8917014, 51.750226, 50.5896083, 49.4101359, 48.212096, 46.995776, 45.7614633, 44.5094452, 43.240009, 41.9534422, 40.6500321, 39.291295, 37.8396779, 36.3754308, 34.9013543, 33.4202492, 31.9349161, 30.4481559, 28.9627692, 27.4815567, 26.0073192, 24.5428575, 23.0909721, 21.6544639, 20.2361337, 18.838782, 17.4652096, 16.1182173, 15.1778441, 14.4503369, 13.7374083, 13.0296171, 12.3175219, 11.5916812, 10.8426538, 10.0609981, 9.23727283, 8.36203662, 7.42584806, 6.41926577, 5.33284837, 4.15715448, 2.88274271, 1.50017168, 0, -3.60480669, -7.09829349, -10.4889496, -13.785264, -16.995726, -20.1288248, -23.1930493, -26.1968889, -29.1488326, -32.0573695, -34.9309889, -37.7781799, -40.6074316, -43.4272331, -46.2460736, -49.0724423, -51.5822552, -53.7364909, -56.2958874, -59.1976681, -62.3790565, -65.7772761, -69.3295503, -72.9731027, -76.6451566, -80.2829356, -83.8236632, -87.2045629, -90.362858, -92.8030376, -95.053729, -97.181976, -99.2548226, -101.966926, -105.071524, -108.15449, -111.094582, -113.770561, -116.061187, -117.845219, -119.167441, -119.690059, -119.259351, -117.721593, -114.923063, -111.67666, -106.620383, -99.3588525, -89.49669, -76.6385167, -60.4152238, -40.3730551, -16.5620581, 11.3792214, 44.6523868, 84.0818317, 130.49195, 185.212369, 248.447423, 319.850277, 397.995432, 486.796244, 590.885048, 712.25668, 848.436896, 999.00283, 1163.30514, 1340.86224, 1532.02298, 1736.66649, 1957.79704, 2191.68779, 2437.13605, 2692.93917, 2957.89447, 3237.21734, 3521.68253, 3808.48284, 4094.81106, 4377.86002, 4632.70622, 4893.40295, 5166.62145, 5489.23219, 5855.95253, 6248.91675, 6663.47136, 7094.96282, 7538.73765, 7990.14231, 8252.21837, 8665.09472, 9167.79282, 9699.33417, 10198.7403, 10605.0325, 10857.2325, 11255.3556, 11521.9724},
        {92.6925216, 91.2451572, 89.8840757, 88.6057999, 87.4068528, 86.2837575, 85.2330368, 84.2512137, 83.3348112, 82.4803523, 81.6843599, 80.9433569, 80.2538664, 79.6124113, 79.0155146, 78.4596992, 77.941488, 77.4574041, 77.0039705, 76.57771, 76.1751456, 75.7928004, 75.4271972, 75.074859, 74.7323088, 74.3960696, 74.0626643, 73.7286159, 73.3904473, 73.0446815, 72.6878414, 72.3164501, 71.9270305, 71.5161056, 70.9618357, 70.3224148, 69.6574593, 68.9673331, 68.2524005, 67.5130253, 66.7495717, 65.9624037, 65.1518855, 64.3183809, 63.4622541, 62.5838692, 61.6835901, 60.761781, 59.8188059, 58.8550288, 57.8708139, 56.8729729, 55.8552825, 54.8179667, 53.7612499, 52.6853563, 51.5905103, 50.4769361, 49.344858, 48.1945002, 47.026087, 45.8398428, 44.6359917, 43.4147581, 42.1763663, 40.9210404, 39.6490048, 38.3206958, 36.8975051, 35.4617165, 34.0161336, 32.5635601, 31.1067994, 29.6486551, 28.1919308, 26.7394301, 25.2939565, 23.8583136, 22.435305, 21.0277341, 19.6384047, 18.2701202, 16.9256843, 15.6079004, 14.6988615, 14.0037487, 13.3231842, 12.6476629, 11.9676798, 11.2737297, 10.5563075, 9.80590822, 9.01302662, 8.16815766, 7.26179622, 6.28443721, 5.22657553, 4.07870606, 2.83132372, 1.4749234, 0, -3.69003281, -7.25081419, -10.6919206, -14.0229285, -17.2534143, -20.3929545, -23.4511255, -26.4375037, -29.3616657, -32.2331878, -35.0616466, -37.8566184, -40.6276798, -43.3844071, -46.1363768, -48.8931654, -51.2577397, -53.181589, -55.5940021, -58.4174247, -61.5743026, -64.9870817, -68.5782078, -72.2701267, -75.9852842, -79.6461262, -83.1750984, -86.4946468, -89.5272171, -91.6636657, -93.552944, -95.2769746, -96.9176803, -99.2002308, -101.884925, -104.56053, -107.115994, -109.440265, -111.422293, -112.951025, -114.233723, -114.744145, -114.308961, -112.754842, -109.908456, -107.303444, -102.632764, -95.2963421, -84.6941071, -70.225986, -50.2936822, -26.5153988, -0.579344635, 27.2687592, 59.7282703, 98.1762661, 143.989824, 200.056621, 265.899818, 339.394644, 418.037068, 506.341633, 608.889146, 727.751927, 860.751806, 1007.66166, 1168.43125, 1342.87932, 1530.176, 1729.85829, 1943.64958, 2168.56416, 2403.38224, 2646.88401, 2897.84969, 3156.45827, 3419.74145, 3686.12973, 3954.05362, 4221.9436, 4471.31635, 4728.79209, 5000.05011, 5317.65324, 5676.63937, 6060.84491, 6465.86782, 6887.30605, 7320.75757, 7761.82034, 8136.82278, 8572.58589, 9037.38888, 9499.511, 9927.23146, 10288.8295, 10552.5844, 10930.0793, 11224.1632},
        {91.5947241, 90.1046634, 88.7032256, 87.3868841, 86.1521123, 84.9953836, 83.9131715, 82.9019495, 81.9581909, 81.0783691, 80.2589577, 79.4964299, 78.7872594, 78.1279194, 77.5148834, 76.9446249, 76.4136172, 75.9183338, 75.4552482, 75.0208337, 74.6115638, 74.2239119, 73.8543515, 73.4993559, 73.1553987, 72.8189531, 72.4864928, 72.154491, 71.8194212, 71.4777569, 71.1259715, 70.7605383, 70.3779309, 69.9746227, 69.4272468, 68.7947175, 68.1373895, 67.455625, 66.7497868, 66.0202371, 65.2673385, 64.4914535, 63.6929446, 62.8721741, 62.0295046, 61.1652986, 60.2799185, 59.3737268, 58.447086, 57.5003585, 56.5339069, 55.5558575, 54.5586405, 53.5424502, 52.5074806, 51.4539261, 50.3819807, 49.2918386, 48.183694, 47.057741, 45.9141738, 44.7531867, 43.5749737, 42.379729, 41.1676468, 39.9389212, 38.6937465, 37.3921892, 35.9952794, 34.5859713, 33.1670605, 31.7413423, 30.3116123, 28.8806659, 27.4512987, 26.0263061, 24.6084835, 23.2006265, 21.8055305, 20.425991, 19.0648035, 17.7247634, 16.4086662, 15.1193075, 14.2396953, 13.5745216, 12.9239083, 12.2783123, 11.6281904, 10.9639995, 10.2761964, 9.55523798, 8.79158102, 7.97568239, 7.09799893, 6.14898749, 5.11910491, 3.99880802, 2.77855368, 1.44879873, 0, -3.52385036, -6.95038978, -10.2864934, -13.5390362, -16.7148934, -19.8209401, -22.8640513, -25.8511022, -28.7889679, -31.6845235, -34.5446441, -37.3762048, -40.1860807, -42.9811468, -45.7682784, -48.5543505, -51.018481, -53.1207224, -55.6093056, -58.4208721, -61.4920631, -64.7595202, -68.1598847, -71.629798, -75.1059015, -78.5248367, -81.8232449, -84.9377676, -87.8050462, -89.9674114, -91.9003959, -93.6589342, -95.2979609, -97.3194495, -99.5548151, -101.716262, -103.724614, -105.500694, -106.965323, -108.039326, -109.351731, -109.899615, -109.465236, -107.830857, -104.778738, -101.693141, -96.3538251, -88.1425501, -76.4410764, -60.6311639, -38.8325723, -13.2315077, 14.0320625, 42.5220584, 75.2811704, 113.790192, 159.529917, 215.493623, 281.305873, 354.941247, 434.467473, 523.493398, 625.600748, 742.477473, 872.602621, 1015.92818, 1172.86985, 1343.49985, 1526.19008, 1720.27424, 1926.62439, 2142.79919, 2367.59983, 2599.82751, 2838.28341, 3079.49577, 3325.10697, 3574.48645, 3827.00364, 4082.02796, 4326.25289, 4580.17445, 4848.59463, 5159.7931, 5509.2786, 5882.77746, 6276.28752, 6685.80662, 7107.33259, 7536.86327, 7991.58605, 8441.75026, 8877.31793, 9288.25109, 9664.51175, 9996.06195, 10272.8637, 10610.7827, 10910.241},
        {90.5861288, 89.0675171, 87.6381703, 86.2945832, 85.0332507, 83.8506677, 82.7433291, 81.7077298, 80.7403648, 79.8377289, 78.996317, 78.2126241, 77.4831451, 76.8043748, 76.1728082, 75.5849401, 75.0372656, 74.5262794, 74.0484765, 73.6003518, 73.1784003, 72.7791167, 72.3989961, 72.0345333, 71.6822232, 71.3385607, 71.0000408, 70.6631584, 70.3244083, 69.9802855, 69.6272848, 69.2619013, 68.8806297, 68.479965, 67.9393397, 67.3162421, 66.6695676, 65.9996101, 65.3066633, 64.591021, 63.852977, 63.0928252, 62.3108593, 61.5073732, 60.6826607, 59.8370155, 58.9707316, 58.0841025, 57.1774223, 56.2509847, 55.3050835, 54.3439191, 53.363794, 52.3649173, 51.3474982, 50.3117459, 49.2578693, 48.1860776, 47.0965801, 45.9895857, 44.8653036, 43.7239429, 42.5657127, 41.3908222, 40.1994805, 38.9918967, 37.7682799, 36.4892511, 35.1163818, 33.7315286, 32.3374671, 30.9369726, 29.5328207, 28.1277868, 26.7246463, 25.3261748, 23.9351476, 22.5543403, 21.1865282, 19.8344869, 18.5009919, 17.1888184, 15.9007421, 14.6395384, 13.7877724, 13.1503363, 12.5275576, 11.909887, 11.2877749, 10.6516719, 9.99202861, 9.29929555, 8.56392327, 7.77636234, 6.9270633, 6.00647672, 5.00505314, 3.91324312, 2.72149722, 1.420266, 0, -3.0037641, -6.01212586, -9.02395793, -12.038133, -15.0535236, -18.0690025, -21.0834422, -24.0957155, -27.104695, -30.1092532, -33.108263, -36.1005968, -39.0851274, -42.0607274, -45.0262694, -47.9806261, -50.8872092, -53.7398255, -56.6175569, -59.5116772, -62.4134604, -65.3141803, -68.2051107, -71.0775256, -73.9226988, -76.7319043, -79.4964157, -82.2075072, -84.8564524, -87.5073364, -90.0519246, -92.4596474, -94.6999355, -96.7028019, -98.4573858, -99.9633378, -101.201914, -102.154369, -102.80196, -103.125942, -104.51007, -105.106252, -104.621344, -102.762199, -99.2356698, -94.0429575, -86.5229829, -76.3090135, -63.0343164, -46.3321591, -25.4413997, -0.871770209, 26.5901923, 56.9493555, 91.8263718, 132.116439, 178.714755, 232.529261, 294.439516, 365.311177, 446.17038, 537.588055, 640.883532, 756.717711, 884.635815, 1024.69518, 1177.48938, 1343.21478, 1520.10154, 1707.49201, 1906.09957, 2113.68463, 2329.11497, 2551.25837, 2778.98259, 3007.01892, 3239.40575, 3476.04497, 3716.83848, 3961.68817, 4200.67573, 4450.07007, 4713.98175, 5016.59727, 5354.08046, 5714.23017, 6093.62847, 6488.85744, 6896.49916, 7313.13569, 7784.38817, 8233.06952, 8655.9724, 9049.88944, 9411.61328, 9737.93657, 10025.6519, 10299.3939, 10573.2056},
        {89.7128469, 88.1635362, 86.7043451, 85.3317802, 84.0423478, 82.8325545, 81.6989067, 80.637911, 79.6460738, 78.7199017, 77.8559011, 77.0505785, 76.3004403, 75.6019932, 74.9517435, 74.3461978, 73.7818625, 73.2552441, 72.7628492, 72.3011842, 71.8667555, 71.4560698, 71.0656334, 70.6919528, 70.3315346, 69.9808853, 69.6365112, 69.294919, 68.9526151, 68.6061059, 68.251898, 67.8864979, 67.506412, 67.1081469, 66.5734803, 65.9586962, 65.3214576, 64.6619941, 63.9805354, 63.2773112, 62.552551, 61.8064846, 61.0393416, 60.2513517, 59.4427445, 58.6137496, 57.7645968, 56.8955156, 56.0067358, 55.098487, 54.1709988, 53.2246571, 52.2595319, 51.2758495, 50.2738363, 49.2537184, 48.2157222, 47.1600739, 46.0869998, 44.9967262, 43.8894793, 42.7654854, 41.6249707, 40.4681617, 39.2952844, 38.1065652, 36.9022303, 35.6434521, 34.2923404, 32.9296675, 31.5581914, 30.18067, 28.7998613, 27.4185233, 26.0394141, 24.6652915, 23.2989136, 21.9430384, 20.6004237, 19.2738277, 17.9660083, 16.6797235, 15.4177312, 14.1827895, 13.3579609, 12.7478519, 12.1524981, 11.5623157, 10.9677213, 10.359131, 9.72696144, 9.06162881, 8.35354953, 7.59313997, 6.77081649, 5.87699546, 4.90209326, 3.83652624, 2.67071078, 1.39506324, 0, -2.5674759, -5.22670302, -7.96959841, -10.7880791, -13.6740622, -16.6194647, -19.6162037, -22.6561962, -25.7313593, -28.83361, -31.9548654, -35.0870426, -38.2220585, -41.3518303, -44.468275, -47.5633096, -50.8622899, -54.3904815, -57.611785, -60.5681402, -63.3014868, -65.8537642, -68.2669123, -70.5828706, -72.8435787, -75.0909763, -77.3670031, -79.7135987, -82.1727028, -85.2932437, -88.4242765, -91.455644, -94.2771894, -96.3119747, -97.6832333, -98.6386736, -99.2081728, -99.4216083, -99.3088572, -98.8997968, -100.160363, -100.59514, -99.8552091, -97.591655, -93.4555605, -86.1721788, -76.5498806, -64.4712063, -49.8186967, -32.4748924, -12.8730016, 10.3281406, 37.8617248, 70.2399463, 107.523802, 149.976869, 197.862722, 249.9549, 308.345662, 376.752762, 459.03047, 552.856715, 657.437192, 772.454797, 898.401013, 1035.39617, 1184.09748, 1344.76446, 1515.68805, 1696.27272, 1887.25902, 2086.50981, 2292.96705, 2505.57275, 2723.26889, 2939.57045, 3160.20317, 3385.46578, 3615.65701, 3851.0756, 4084.42326, 4328.66043, 4587.34171, 4881.46877, 5207.69008, 5555.3845, 5921.56399, 6303.2405, 6697.42597, 7101.13237, 7549.45708, 7984.49031, 8401.03761, 8793.90451, 9157.89656, 9487.81931, 9778.47829, 9984.36029, 10226.8112},
        {89.0085334, 87.4127432, 85.9096326, 84.495649, 83.1672399, 81.9208525, 80.7529342, 79.6599324, 78.6382945, 77.6844678, 76.7948997, 75.9660375, 75.1943287, 74.4762206, 73.8081606, 73.186596, 72.6079742, 72.0687426, 71.5653486, 71.0942395, 70.6518626, 70.2346654, 69.8390953, 69.4615995, 69.0986255, 68.7466206, 68.4020322, 68.0613077, 67.7208944, 67.3772398, 67.0267911, 66.6659957, 66.2913011, 65.8991545, 65.3698948, 64.7604657, 64.1292068, 63.4763333, 62.8020605, 62.1066037, 61.3901784, 60.6529997, 59.895283, 59.1172436, 58.3190969, 57.5010581, 56.6633426, 55.8061656, 54.9297426, 54.0342888, 53.1200195, 52.1869807, 51.2355607, 50.2659785, 49.2784532, 48.2732037, 47.2504489, 46.210408, 45.1532998, 44.0793433, 42.9887576, 41.8817617, 40.7585744, 39.6194149, 38.464502, 37.2940549, 36.1082924, 34.8682603, 33.5359449, 32.192357, 30.8402552, 29.4823977, 28.1215431, 26.7604498, 25.4018762, 24.0485808, 22.7033219, 21.368858, 20.0479475, 18.743349, 17.4578207, 16.1941211, 14.9550087, 13.7432419, 12.9463638, 12.3666471, 11.8017396, 11.2419127, 10.6774379, 10.0985866, 9.4956302, 8.85884013, 8.17848782, 7.4448447, 6.64818219, 5.77877173, 4.82688474, 3.78279264, 2.63676687, 1.37907885, 0, -2.54883964, -5.19913068, -7.94153887, -10.76673, -13.6653698, -16.6281241, -19.6456586, -22.7086391, -25.8077314, -28.9336012, -32.0769143, -35.2283365, -38.3785335, -41.5181711, -44.6379151, -47.7284313, -51.0953457, -54.7743183, -58.0436049, -60.9613626, -63.5857486, -65.9749204, -68.187035, -70.2802499, -72.3127221, -74.3426091, -76.428068, -78.6272562, -80.9983308, -84.2569055, -87.5626138, -90.7763763, -93.7591132, -95.8080306, -97.065906, -97.8258408, -98.1178698, -97.9720276, -97.4183489, -96.4868685, -97.064095, -96.7585752, -95.2371208, -92.1665432, -87.2136539, -78.8343283, -68.2090481, -55.3073588, -40.098806, -22.5529351, -3.71372393, 18.8369089, 46.8398267, 81.1522297, 120.827173, 165.727737, 215.717001, 268.444044, 326.489128, 394.847786, 478.577551, 574.001384, 679.323475, 793.7276, 918.05738, 1052.39454, 1197.26113, 1352.85303, 1517.75159, 1691.45144, 1874.85228, 2065.82749, 2263.3853, 2466.53397, 2674.28174, 2879.21246, 3088.36489, 3302.35335, 3521.79219, 3747.29577, 3973.33642, 4210.76516, 4462.82862, 4749.73763, 5068.02383, 5407.15043, 5764.00246, 6135.46497, 6518.42302, 6909.76164, 7258.8044, 7653.40777, 8070.82763, 8488.3198, 8883.14014, 9232.5445, 9513.78873, 9652.05376, 9884.52983},
        {88.4407418, 86.7898267, 85.2351103, 83.7729399, 82.3996627, 81.1116261, 79.9051774, 78.7766638, 77.7224327, 76.7388314, 75.8222071, 74.9689073, 74.1752791, 73.4376698, 72.7524269, 72.1158975, 71.524429, 70.9743687, 70.4620639, 69.9838618, 69.5361099, 69.1151553, 68.7173454, 68.3390274, 67.9765488, 67.6262568, 67.2844986, 66.9476216, 66.6119731, 66.2739004, 65.9297508, 65.5758716, 65.2086101, 64.8243136, 64.2996011, 63.6932024, 63.0652863, 62.4160855, 61.7458327, 61.0547606, 60.3431019, 59.6110894, 58.8589558, 58.0869337, 57.2952559, 56.4841551, 55.653864, 54.8046154, 53.9366418, 53.0501762, 52.145451, 51.2243958, 50.2855097, 49.3289888, 48.3550289, 47.3638261, 46.3555762, 45.3304752, 44.288719, 43.2305035, 42.1560247, 41.0654786, 39.9590609, 38.8369678, 37.6993951, 36.5465387, 35.3785946, 34.1560755, 32.840447, 31.5137485, 30.1787485, 28.8382155, 27.4949179, 26.1516241, 24.8111026, 23.4761218, 22.1494501, 20.833856, 19.5321079, 18.2469742, 16.9812234, 15.7376239, 14.5189441, 13.3279525, 12.5583308, 12.0091684, 11.4748563, 10.9454771, 10.4111134, 9.86184764, 9.28776242, 8.67894027, 8.02546375, 7.31741541, 6.54487779, 5.69793345, 4.76666492, 3.74115477, 2.61148553, 1.36773976, 0, -2.79429351, -5.64943958, -8.55857438, -11.5148341, -14.5113548, -17.5412728, -20.5977242, -23.6738451, -26.7627718, -29.8576403, -32.9515869, -36.0377477, -39.1092589, -42.1592567, -45.1808771, -48.1672565, -51.3854445, -54.8677079, -57.9876163, -60.7970016, -63.3476958, -65.6915308, -67.8803385, -69.965951, -72.0002001, -74.0349178, -76.121936, -78.3130867, -80.6602019, -83.8392876, -87.0491378, -90.154332, -93.0194499, -95.0535816, -96.3490516, -97.119648, -97.3665975, -97.0911264, -96.2944613, -94.9778288, -94.5594585, -93.1923119, -90.6003756, -86.5076363, -80.6380808, -71.8330148, -60.9197762, -47.8430222, -32.5474096, -14.9775954, 3.66728857, 26.230307, 54.979365, 90.9590382, 132.716271, 179.919395, 232.236739, 286.996859, 346.739627, 416.557393, 501.526324, 597.748608, 703.237218, 817.035567, 939.933094, 1071.91778, 1213.28867, 1364.1144, 1523.32304, 1690.48784, 1866.67175, 2049.72911, 2238.7175, 2432.69448, 2630.71764, 2824.37984, 3022.06952, 3224.71044, 3433.22635, 3648.54098, 3866.60006, 4096.62405, 4341.67011, 4622.33682, 4934.7379, 5267.67815, 5617.63673, 5981.09282, 6354.52558, 6734.41416, 6949.26418, 7292.84232, 7716.30057, 8170.79091, 8607.46535, 8977.47586, 9231.97444, 9312.77506, 9531.76013},
        {87.976596, 86.2690309, 84.6614017, 83.1499441, 81.730894, 80.400487, 79.154959, 77.9905456, 76.9034826, 75.8900057, 74.9463507, 74.0687533, 73.2534493, 72.4966744, 71.7946644, 71.1436549, 70.5398818, 69.9795808, 69.4589876, 68.974338, 68.5218677, 68.0978125, 67.698408, 67.3198901, 66.9584945, 66.610457, 66.2720132, 65.9393989, 65.6088499, 65.2766019, 64.9388906, 64.5919519, 64.2320214, 63.8553348, 63.33399, 62.7289027, 62.1024808, 61.4549885, 60.7866902, 60.0978501, 59.3887325, 58.6596015, 57.9107216, 57.1423569, 56.3547717, 55.5482302, 54.7229968, 53.8793356, 53.017511, 52.1377872, 51.2404285, 50.33026, 49.4028863, 48.4584727, 47.4971845, 46.5191873, 45.5246463, 44.513727, 43.4865948, 42.443415, 41.384353, 40.3095743, 39.2192442, 38.113528, 36.9925913, 35.8565993, 34.7057175, 33.4997846, 32.1996076, 30.8885225, 29.569309, 28.2447467, 26.9176151, 25.5906938, 24.2667625, 22.9486008, 21.6389883, 20.3407045, 19.0565291, 17.7892416, 16.5416217, 15.316449, 14.116503, 12.9445635, 12.1997539, 11.6783454, 11.1718452, 10.6701708, 10.1632396, 9.64096918, 9.09327694, 8.51008037, 7.88129695, 7.19684415, 6.44663946, 5.62060034, 4.70864427, 3.70068873, 2.58665119, 1.35644912, 0, -3.15339913, -6.3032469, -9.44693774, -12.5818661, -15.7054264, -18.8150131, -21.9080207, -24.9818435, -28.0338761, -31.0615128, -34.0621481, -37.0331765, -39.9719924, -42.8759902, -45.7425643, -48.5691093, -51.5265667, -54.6371234, -57.5056267, -60.1666597, -62.6548056, -65.0046474, -67.2507683, -69.4277514, -71.5701798, -73.7126366, -75.889705, -78.135968, -80.4860089, -83.4792714, -86.4603624, -89.3124066, -91.9185289, -93.8965574, -95.2622653, -96.1021718, -96.3789909, -96.0554364, -95.0942224, -93.4580627, -91.9776123, -89.4834883, -85.7685898, -80.6258164, -73.8480674, -64.9771832, -54.1198868, -41.1318425, -25.8687142, -8.18616612, 10.902989, 34.1145286, 63.7768239, 100.962599, 144.180966, 193.092046, 247.35596, 304.597691, 366.97503, 438.865922, 524.599631, 620.808594, 725.837108, 838.833613, 960.309129, 1090.14923, 1228.45489, 1375.16754, 1529.43878, 1690.86698, 1860.54781, 2036.352, 2217.35962, 2402.65076, 2591.30549, 2773.49404, 2959.43381, 3150.43232, 3347.79712, 3552.83574, 3763.24879, 3986.35536, 4225.00885, 4500.14004, 4807.44666, 5135.09418, 5479.15694, 5835.70928, 6200.82554, 6570.58006, 6657.86115, 6956.19178, 7388.97671, 7879.6207, 8351.52851, 8728.10491, 8932.75466, 8977.26418, 9153.63919},
        {87.5465505, 85.7931714, 84.1425406, 82.5908136, 81.134146, 79.7686933, 78.490611, 77.2960546, 76.1811796, 75.1421415, 74.1750958, 73.2761981, 72.4416039, 71.6674686, 70.9499478, 70.285197, 69.6693718, 69.0986275, 68.5691198, 68.0770041, 67.6184359, 67.1895708, 66.7865643, 66.4055719, 66.0427491, 65.6942514, 65.3562343, 65.0248534, 64.696264, 64.3666219, 64.0320824, 63.6888011, 63.3329335, 62.9606351, 62.4404937, 61.835132, 61.2087056, 60.56151, 59.8938405, 59.2059924, 58.4982611, 57.7709419, 57.02433, 56.2587209, 55.4744099, 54.6716922, 53.8508633, 53.0122184, 52.1560529, 51.2826621, 50.3923413, 49.4930253, 48.5772043, 47.6450082, 46.6965666, 45.7320093, 44.7514659, 43.7550662, 42.74294, 41.7152169, 40.6720266, 39.6134989, 38.5397636, 37.4509502, 36.3471886, 35.2286085, 34.0953396, 32.9066965, 31.6229738, 30.3284793, 29.0259886, 27.7182772, 26.4081209, 25.0982952, 23.7915756, 22.4907378, 21.1985574, 19.91781, 18.6512712, 17.4017166, 16.1719217, 14.9646622, 13.7827137, 12.6288518, 11.9018448, 11.3981428, 10.9094362, 10.4256501, 9.93670912, 9.43253829, 8.90306245, 8.33820646, 7.7278952, 7.06205355, 6.33060636, 5.52347851, 4.63059489, 3.64188035, 2.54725978, 1.33665804, 0, -3.4687308, -6.86993848, -10.2052972, -13.4764812, -16.6851645, -19.8330214, -22.921726, -25.9529525, -28.9283751, -31.8496679, -34.7185051, -37.5365609, -40.3055094, -43.0270248, -45.7027813, -48.3344531, -51.0001121, -53.7123465, -56.3000259, -58.7811953, -61.1738999, -63.4961848, -65.7660952, -68.0016762, -70.220973, -72.4420306, -74.6828942, -76.961609, -79.2962201, -82.1072237, -84.8626484, -87.4598042, -89.7960008, -91.6272589, -92.9215329, -93.6844544, -93.8557199, -93.3750261, -92.1820695, -90.2165467, -87.9530759, -84.6336298, -80.0932467, -74.1669644, -66.6898211, -58.0307132, -47.3573558, -34.3713222, -18.7741859, -0.267520317, 20.9057794, 46.247077, 77.1662621, 114.40604, 157.34695, 205.981115, 260.300659, 319.095778, 383.833689, 457.292797, 542.370275, 636.954031, 740.187202, 851.582338, 971.277927, 1099.12514, 1235.29606, 1379.72505, 1531.16982, 1689.05367, 1853.93551, 2023.92831, 2198.06226, 2375.36757, 2554.87443, 2723.39055, 2895.22424, 3072.46131, 3257.18759, 3451.48887, 3656.9738, 3876.52351, 4112.42833, 4383.12276, 4684.50674, 5005.75345, 5343.09915, 5692.7801, 6051.03257, 6414.09282, 6447.91642, 6739.67263, 7190.08576, 7699.8801, 8169.77995, 8500.50963, 8592.79343, 8681.40759, 8716.4231},
        {87.2064897, 85.408796, 83.7164755, 82.1256096, 80.6322801, 79.2325685, 77.9225564, 76.6983255, 75.5559572, 74.4915331, 73.501135, 72.5808444, 71.7267428, 70.9349119, 70.2014332, 69.5223884, 68.8938591, 68.3119268, 67.7726731, 67.2721797, 66.8065281, 66.3717999, 65.9640767, 65.5794402, 65.2139718, 64.8637533, 64.5248661, 64.1933919, 63.8654123, 63.5370089, 63.2042632, 62.863257, 62.5100716, 62.1407889, 61.6207589, 61.0141788, 60.3868168, 59.738997, 59.0710435, 58.3832804, 57.6760318, 56.9496216, 56.2043741, 55.4406133, 54.6586632, 53.858848, 53.0414917, 52.2069184, 51.3554521, 50.4874171, 49.6031372, 48.7134476, 47.8079335, 46.8866912, 45.9498169, 44.9974067, 44.0295569, 43.0463638, 42.0479235, 41.0343323, 40.0056864, 38.962082, 37.9036153, 36.8303826, 35.74248, 34.6400039, 33.5230503, 32.3504641, 31.0820804, 29.8030735, 28.5162138, 27.2242717, 25.9300177, 24.636222, 23.3456551, 22.0610873, 20.785289, 19.5210307, 18.2710827, 17.0382154, 15.8251992, 14.6348045, 13.4698016, 12.332961, 11.6213303, 11.1322481, 10.6583142, 10.1895039, 9.71579266, 9.22715573, 8.71356846, 8.16500617, 7.5714442, 6.92285788, 6.20922256, 5.42051355, 4.5467062, 3.57777584, 2.50369779, 1.3144474, 0, -3.76491022, -7.39943214, -10.9095142, -14.3011049, -17.5801528, -20.7526062, -23.8244138, -26.8015238, -29.6898849, -32.4954455, -35.2241541, -37.8819591, -40.474809, -43.0086523, -45.4894376, -47.9231131, -50.2864339, -52.5811778, -54.8793265, -57.1805728, -59.4846093, -61.7911288, -64.0998239, -66.4103874, -68.7225119, -71.0358901, -73.3502147, -75.6651784, -77.9804739, -80.5841447, -83.0818041, -85.3866395, -87.4118384, -89.0414634, -90.2032644, -90.8327578, -90.8518687, -90.1825219, -88.7466423, -86.4661549, -83.5590577, -79.5610933, -74.3362593, -67.7485531, -59.6619725, -51.2009605, -40.6539577, -27.5698502, -11.4975242, 8.01413415, 31.5751923, 59.283534, 91.3874592, 128.221352, 170.295353, 218.040691, 271.888595, 332.098881, 399.31315, 474.360001, 558.410546, 650.898259, 751.942698, 861.518113, 979.352536, 1105.28702, 1239.64232, 1382.38388, 1531.71814, 1686.84654, 1847.66565, 2012.57481, 2180.53483, 2350.50651, 2521.45066, 2676.58573, 2834.55047, 2998.24128, 3170.55454, 3354.38666, 3555.42187, 3771.91016, 4005.55313, 4271.92741, 4567.24239, 4881.92624, 5212.47744, 5555.39446, 5907.17578, 6264.31986, 6287.00065, 6586.89604, 7049.8119, 7561.5541, 8007.9285, 8274.74097, 8247.79738, 8382.20452, 8248.96581},
        {86.969027, 85.1266671, 83.3922059, 81.761656, 80.2310297, 78.7963394, 77.4535974, 76.1988161, 75.028008, 73.9371854, 72.9223606, 71.9795461, 71.1047542, 70.2939973, 69.5432879, 68.8486382, 68.2060606, 67.6115676, 67.0611715, 66.5508846, 66.0767195, 65.6346884, 65.2208037, 64.8310778, 64.4615231, 64.108152, 63.7669768, 63.4340099, 63.1052637, 62.7767506, 62.444483, 62.1044732, 61.7527337, 61.3852767, 60.8646287, 60.2563528, 59.6276164, 58.9787638, 58.3101393, 57.622087, 56.9149513, 56.1890764, 55.4448066, 54.6824861, 53.9024592, 53.1050702, 52.2906632, 51.4595826, 50.6121726, 49.7487775, 48.8697415, 47.9880383, 47.091109, 46.1790239, 45.2518537, 44.3096688, 43.3525396, 42.3805367, 41.3937304, 40.3921914, 39.37599, 38.3451968, 37.2998821, 36.2401166, 35.1659706, 34.0775147, 32.9748192, 31.816351, 30.5615757, 29.2963683, 28.0234963, 26.7457272, 25.4658285, 24.1865677, 22.9107122, 21.6410295, 20.3802871, 19.1312525, 17.8966931, 16.6793764, 15.48207, 14.3075412, 13.1585576, 12.0378866, 11.340633, 10.8651433, 10.4050485, 9.95038402, 9.49118519, 9.01748742, 8.51932608, 7.98673654, 7.4097542, 6.77841443, 6.08275262, 5.31280413, 4.45860436, 3.51018868, 2.45759247, 1.29085112, 0, -4.00375623, -7.82416866, -11.4708277, -14.9533237, -18.281247, -21.464188, -24.5117372, -27.4334847, -30.2390212, -32.9379368, -35.5398221, -38.0542673, -40.4908628, -42.8591991, -45.1688664, -47.4294553, -49.5212656, -51.4354176, -53.4739444, -55.6187311, -57.8516632, -60.1546258, -62.5095042, -64.8981839, -67.3025499, -69.7044878, -72.0858826, -74.4286198, -76.7145846, -79.1050034, -81.3366621, -83.3376435, -85.0360307, -86.4155877, -87.3765571, -87.8043328, -87.6102936, -86.7058181, -85.0022849, -82.4110727, -79.0274656, -74.5229656, -68.7729696, -61.6528749, -53.0380786, -44.5422899, -33.8680158, -20.4560753, -3.74728751, 16.8175286, 42.5401038, 72.3295082, 105.334573, 141.517745, 182.502599, 229.166897, 282.388401, 343.91814, 413.56214, 490.173771, 573.188083, 663.580306, 762.415091, 870.12493, 986.071171, 1110.10684, 1242.73772, 1383.98609, 1531.48069, 1684.20416, 1841.40445, 2001.7581, 2164.15594, 2327.48877, 2490.64742, 2633.63614, 2778.95396, 2930.21332, 3091.02669, 3265.00653, 3461.42536, 3674.46219, 3905.30372, 4166.96629, 4455.95406, 4763.84214, 5087.38113, 5423.32162, 5768.4142, 6119.40946, 6175.71658, 6494.11909, 6957.34676, 7448.12937, 7849.19669, 8043.2785, 7913.10457, 8063.86881, 7758.52621},
        {86.8415902, 84.949902, 83.1686922, 81.4939093, 79.9215021, 78.4474193, 77.0676096, 75.7780216, 74.5746041, 73.4533057, 72.4100752, 71.4408613, 70.5416126, 69.7082778, 68.9368057, 68.2231449, 67.5632441, 66.9530521, 66.3885175, 65.865589, 65.3802153, 64.9283451, 64.5059271, 64.10891, 63.7332425, 63.3748733, 63.0297511, 62.6938246, 62.3630424, 62.0333533, 61.700706, 61.3610491, 61.0103314, 60.6445015, 60.12467, 59.5170407, 58.8894382, 58.2421871, 57.5756118, 56.8900367, 56.1857864, 55.4631852, 54.7225576, 53.964228, 53.188521, 52.395761, 51.5862724, 50.7603797, 49.9184074, 49.0606798, 48.1875216, 47.31071, 46.4188683, 45.5120726, 44.5903992, 43.6539243, 42.702724, 41.7368744, 40.7564519, 39.7615325, 38.7521924, 37.7285078, 36.690555, 35.6384099, 34.572149, 33.4918482, 32.3975839, 31.2475428, 30.0009176, 28.744178, 27.4801157, 26.2115225, 24.9411899, 23.6719098, 22.4064737, 21.1476735, 19.8983007, 18.6611471, 17.4390043, 16.2346641, 15.0509182, 13.8905582, 12.7563758, 11.6511628, 10.9723507, 10.5171106, 10.0778703, 9.64461444, 9.20732806, 8.75599599, 8.2806031, 7.77113425, 7.21757432, 6.60990817, 5.93812068, 5.1921967, 4.36212111, 3.43787878, 2.40945457, 1.26683335, 0, -4.01857318, -7.84648112, -11.4938938, -14.9709813, -18.2879135, -21.4548604, -24.4819921, -27.3794785, -30.1574896, -32.8261954, -35.3957659, -37.876371, -40.2781809, -42.6113654, -44.8860946, -47.1125385, -49.1207803, -50.8952635, -52.8534968, -54.9670603, -57.207534, -59.5464978, -61.9555317, -64.4062158, -66.87013, -69.3188544, -71.7239689, -74.0570535, -76.2896882, -78.5248948, -80.5546163, -82.311, -83.7261935, -84.7583855, -85.3267029, -85.3433282, -84.7325964, -83.4188428, -81.3264025, -78.3796105, -74.9557256, -70.388313, -64.5130922, -57.165783, -48.1821049, -38.4816477, -26.5452933, -11.9377939, 5.77609864, 27.0316321, 52.6948189, 82.2436527, 114.814374, 150.442628, 191.000348, 237.535011, 291.094094, 354.177747, 426.050621, 504.392638, 587.634099, 677.450129, 775.137449, 881.279282, 995.249858, 1116.97835, 1247.16049, 1385.92417, 1530.58666, 1680.05506, 1833.28013, 1989.11865, 2146.46264, 2304.20415, 2461.23518, 2597.01726, 2734.73055, 2878.12469, 3030.94934, 3196.95413, 3385.97571, 3591.61875, 3815.02419, 4069.35823, 4351.0018, 4651.32533, 4966.96027, 5294.53808, 5630.69024, 5972.0482, 6171.21927, 6517.54283, 6930.97868, 7331.48659, 7639.02635, 7773.55775, 7655.04058, 7658.29655, 7234.93224},
        {86.8441654, 84.9001317, 83.0692068, 81.3472759, 79.7302246, 78.2139384, 76.7943027, 75.4672029, 74.2285246, 73.0741532, 71.9999741, 71.0018728, 70.0757348, 69.2174455, 68.4228904, 67.687955, 67.0085246, 66.3804849, 65.7997211, 65.2621189, 64.7635636, 64.2999407, 63.8671356, 63.4610339, 63.077521, 62.7124824, 62.3618034, 62.0213696, 61.6870665, 61.3547794, 61.0203939, 60.6797954, 60.3288694, 59.9635013, 59.4437223, 58.8362026, 58.2092451, 57.5631439, 56.8981934, 56.2146876, 55.5129208, 54.7931871, 54.0557808, 53.300996, 52.529127, 51.7404679, 50.9353129, 50.1139562, 49.276692, 48.4238145, 47.5556179, 46.6816687, 45.7927876, 44.8890681, 43.9706032, 43.0374861, 42.08981, 41.1276681, 40.1511536, 39.1603597, 38.1553795, 37.1363063, 36.1032333, 35.0562535, 33.9954603, 32.9209468, 31.8328062, 30.6891259, 29.4490021, 28.1991615, 26.9424203, 25.6815948, 24.4195013, 23.1589559, 21.9027749, 20.6537746, 19.4147712, 18.188581, 16.9780202, 15.785905, 14.6150517, 13.4682765, 12.3483957, 11.2582255, 10.5981768, 10.1639056, 9.74628931, 9.33524135, 8.92067532, 8.49250474, 8.04064316, 7.5550041, 7.02550111, 6.44204773, 5.79455748, 5.07294392, 4.26712056, 3.36700095, 2.36249863, 1.24352714, 0, -3.95389012, -7.72595663, -11.325823, -14.7631129, -18.0474497, -21.188457, -24.1957583, -27.0789772, -29.8477371, -32.5116616, -35.0803743, -37.5634985, -39.970658, -42.3114762, -44.5955765, -46.8325827, -48.8220079, -50.5434601, -52.4818118, -54.6016631, -56.8676139, -59.244264, -61.6962135, -64.1880621, -66.6844099, -69.1498567, -71.5490025, -73.8464472, -76.0067906, -78.1002802, -79.9471311, -81.4802489, -82.6325393, -83.3051037, -83.4467499, -83.014767, -81.951602, -80.1997018, -77.7015135, -74.3994841, -71.0560844, -66.5441651, -60.6457337, -53.1427978, -43.8173649, -32.5436966, -18.9884829, -2.91066778, 15.9308046, 37.7769903, 62.7125607, 91.326093, 123.110388, 158.316035, 198.740773, 245.487702, 299.659924, 364.165486, 437.892325, 517.759342, 601.557165, 691.298649, 788.262836, 893.063725, 1005.18843, 1124.64132, 1252.25778, 1388.25764, 1529.81366, 1675.8222, 1825.14552, 1976.71932, 2129.45178, 2282.25107, 2434.02535, 2564.55753, 2696.66235, 2834.0293, 2980.34784, 3139.30747, 3320.30546, 3517.5183, 3732.18926, 3978.51114, 4252.60138, 4545.18178, 4852.63955, 5171.36192, 5497.73609, 5828.14929, 6193.7105, 6568.20145, 6918.53219, 7211.61278, 7414.35326, 7493.66367, 7416.45408, 7224.08213, 6710.22452},
        {86.9930403, 84.9935347, 83.1098621, 81.3378418, 79.6732932, 78.1120355, 76.6498882, 75.2826704, 74.0062017, 72.8163013, 71.7087885, 70.6794826, 69.7242031, 68.8387692, 68.0190003, 67.2607156, 66.5597346, 65.9118765, 65.3129607, 64.7588066, 64.2452333, 63.7680604, 63.3231071, 62.9061927, 62.5131366, 62.139758, 61.7818765, 61.4353112, 61.0958814, 60.7594066, 60.4217061, 60.0785992, 59.7259052, 59.3594434, 58.8380396, 58.2288915, 57.6008431, 56.9541595, 56.2891058, 55.6059468, 54.9049478, 54.1863737, 53.4504896, 52.6975605, 51.9278515, 51.1416275, 50.3391536, 49.5206949, 48.6865163, 47.836883, 46.9720599, 46.099267, 45.2116636, 44.3093641, 43.3924826, 42.4611335, 41.5154311, 40.5554895, 39.5814231, 38.5933462, 37.5913729, 36.5756176, 35.5461946, 34.503218, 33.4468022, 32.3770615, 31.2941101, 30.1561766, 28.9224952, 27.6795663, 26.4302195, 25.1772842, 23.92359, 22.6719665, 21.4252432, 20.1862497, 18.9578155, 17.7427701, 16.543943, 15.3641639, 14.2062623, 13.0730678, 11.9674097, 10.8921178, 10.2500051, 9.83549613, 9.43820822, 9.04799064, 8.65469272, 8.24816376, 7.81825307, 7.35480997, 6.84768377, 6.28672378, 5.66177932, 4.96269969, 4.17933421, 3.30153219, 2.31914294, 1.22201577, 0, -3.86257885, -7.55806812, -11.0950692, -14.4821833, -17.728012, -20.8411564, -23.8302181, -26.7037982, -29.4704982, -32.1389195, -34.7176633, -37.215331, -39.640524, -42.0018436, -44.3078911, -46.567268, -48.5619894, -50.2682872, -52.207279, -54.3390121, -56.6235336, -59.0208906, -61.4911304, -63.9943001, -66.490447, -68.939618, -71.3018606, -73.5372217, -75.6057486, -77.5584267, -79.2310209, -80.556297, -81.4670206, -81.813296, -81.5692198, -80.7309312, -79.2559944, -77.1019735, -74.2264326, -70.5869358, -67.3014785, -62.8140188, -56.8550799, -49.1551851, -39.4448575, -26.5188263, -11.2773574, 6.31507756, 26.2940068, 48.6949585, 72.7426514, 100.274413, 131.3042, 166.345625, 206.932424, 254.140321, 309.045037, 374.679208, 449.716889, 530.697319, 615.096473, 704.994393, 801.528753, 905.267158, 1015.81127, 1133.20612, 1258.32835, 1391.43858, 1529.74834, 1672.1939, 1817.75653, 1965.32059, 2113.80677, 2262.13576, 2409.22828, 2535.45396, 2662.92232, 2795.19182, 2935.82091, 3088.36807, 3261.46624, 3450.21639, 3656.00221, 3894.33684, 4161.01105, 4445.92833, 4745.17379, 5054.83257, 5370.9898, 5689.73061, 6203.72547, 6599.8815, 6888.5838, 7080.21751, 7185.16775, 7213.81965, 7176.55834, 6782.50665, 6204.01184},
        {87.3325968, 85.2748441, 83.3357921, 81.5111844, 79.7967651, 78.188278, 76.6814672, 75.2720764, 73.9558496, 72.7285307, 71.5858637, 70.5235923, 69.5374607, 68.6232125, 67.7765919, 66.9933426, 66.2692086, 65.5999339, 64.9812622, 64.4089376, 63.8787039, 63.3863051, 62.927485, 62.4979877, 62.0935569, 61.7099366, 61.3428707, 60.9881032, 60.6413779, 60.2984387, 59.9550296, 59.6068945, 59.2497773, 58.8794218, 58.3522595, 57.7366206, 57.1025103, 56.4501932, 55.7799345, 55.0919988, 54.3866511, 53.6641562, 52.9247791, 52.1687846, 51.3964376, 50.6080029, 49.8037455, 48.9839301, 48.1488218, 47.2986853, 46.4337856, 45.5609465, 44.6737318, 43.772264, 42.8566659, 41.9270602, 40.9835695, 40.0263166, 39.0554241, 38.0710147, 37.0732111, 36.062136, 35.0379121, 34.0006621, 32.9505086, 31.8875744, 30.8119821, 29.6824919, 28.4588571, 27.2265075, 25.9882471, 24.7468799, 23.50521, 22.2660415, 21.0321783, 19.8064246, 18.5915844, 17.3904618, 16.2058607, 15.0405853, 13.8974396, 12.7792277, 11.6887535, 10.6288212, 10.0015533, 9.60169552, 9.21919397, 8.84389405, 8.46564117, 8.07428073, 7.65965812, 7.21161874, 6.72000798, 6.17467124, 5.56545393, 4.88220142, 4.11475913, 3.25297244, 2.28668676, 1.20574748, 0, -3.84378736, -7.52274548, -11.0450663, -14.4189419, -17.6525641, -20.7541249, -23.7318164, -26.5938304, -29.3483589, -32.003594, -34.5677275, -37.0489515, -39.4554579, -41.7954387, -44.0770858, -46.3085913, -48.2668191, -49.9269146, -51.8203111, -53.9056305, -56.1414944, -58.4865247, -60.8993429, -63.3385707, -65.76283, -68.1307424, -70.4009296, -72.5320133, -74.4826152, -76.2912552, -77.8073609, -78.9655845, -79.7005784, -79.8963416, -79.512853, -78.523599, -76.8784277, -74.5271875, -71.4197265, -67.505893, -63.9146557, -59.0578796, -52.6547153, -44.4243138, -34.0858258, -19.9884572, -3.56378975, 15.1265397, 36.0208941, 59.0576367, 83.0431154, 110.431282, 141.599145, 177.204629, 218.479199, 266.396811, 321.931424, 387.885375, 462.9887, 543.976836, 628.472148, 718.361842, 814.719503, 918.025944, 1027.75756, 1143.85178, 1266.91521, 1397.05876, 1531.93979, 1670.60356, 1812.57213, 1956.34053, 2100.78897, 2244.79761, 2387.24663, 2508.14301, 2629.95844, 2756.29139, 2890.74036, 3036.90382, 3203.62082, 3385.75557, 3584.66059, 3816.42792, 4076.57568, 4354.48056, 4645.96776, 4946.86249, 5252.98995, 5560.17536, 6099.57424, 6491.19747, 6753.68911, 6905.6932, 6965.8538, 6952.81497, 6885.22076, 6367.12838, 5757.95917},
        {87.8762463, 85.7571093, 83.7597316, 81.8797697, 80.1128797, 78.4547181, 76.9009411, 75.4472052, 74.0891666, 72.8224818, 71.6428069, 70.5457984, 69.5271127, 68.5824059, 67.7073346, 66.8975549, 66.1487233, 65.4564961, 64.8165297, 64.2244802, 63.6760042, 63.1667579, 62.6923977, 62.2485799, 61.8309608, 61.4351969, 61.0569443, 60.6918595, 60.3355988, 59.9838185, 59.6321749, 59.2763245, 58.9119235, 58.5346283, 57.9969494, 57.3691269, 56.7231215, 56.0592349, 55.3777686, 54.6790245, 53.9633042, 53.2309094, 52.4821418, 51.7173031, 50.936695, 50.1406191, 49.3293772, 48.503271, 47.6626022, 46.8076723, 45.9387833, 45.064913, 44.1774988, 43.2766542, 42.3624928, 41.4351283, 40.4946743, 39.5412443, 38.574952, 37.595911, 36.6042349, 35.6000372, 34.5834317, 33.5545318, 32.5134512, 31.4603036, 30.3952025, 29.2778795, 28.0690361, 26.8520704, 25.6297138, 24.4046977, 23.1797536, 21.9576128, 20.7410067, 19.5326667, 18.3353242, 17.1517107, 15.9845575, 14.836596, 13.7105576, 12.6091737, 11.5351758, 10.4912951, 9.87504371, 9.483459, 9.1088335, 8.74108721, 8.37014015, 7.98591233, 7.57832377, 7.13729447, 6.65274446, 6.11459374, 5.51276233, 4.83717023, 4.07773747, 3.22438406, 2.26703, 1.19559531, 0, -3.93523762, -7.68819756, -11.2677313, -14.6826901, -17.9419257, -21.0542894, -24.0286326, -26.8738068, -29.5986634, -32.2120539, -34.7228298, -37.1398425, -39.4719434, -41.727984, -43.9168157, -46.04729, -47.8988228, -49.447489, -51.2211015, -53.1793468, -55.2819117, -57.4884829, -59.7587471, -62.0523909, -64.3291011, -66.5485643, -68.6704673, -70.6544967, -72.4603392, -74.1122831, -75.4817262, -76.5089749, -77.1343353, -77.3878035, -77.1648407, -76.3369912, -74.8176543, -72.5202293, -69.3581156, -65.2447125, -60.8787042, -55.1492055, -47.815973, -38.6387633, -27.3773331, -12.7933603, 4.10580011, 23.3108718, 44.8125783, 68.6016435, 93.690135, 122.243567, 154.711987, 191.717097, 234.23107, 283.068723, 339.044877, 404.404034, 478.2064, 557.952523, 641.856629, 731.404458, 827.761355, 931.306622, 1041.08765, 1156.7632, 1278.32545, 1405.5196, 1536.86831, 1671.58577, 1810.15401, 1950.32434, 2090.87206, 2230.5725, 2368.20099, 2482.05391, 2596.50525, 2715.45007, 2842.78344, 2982.40041, 3144.77843, 3322.84191, 3517.66487, 3744.78503, 3999.55122, 4271.2857, 4555.63863, 4848.26014, 5144.80039, 5440.90952, 5852.07347, 6207.62976, 6489.78744, 6680.75561, 6762.74336, 6717.95976, 6528.61389, 5993.62758, 5388.25787},
        {88.5777614, 86.392366, 84.3320258, 82.3923013, 80.5687529, 78.8569413, 77.252427, 75.7507705, 74.3475324, 73.0382732, 71.8185534, 70.6839336, 69.6299743, 68.652236, 67.7462793, 66.9076647, 66.1319528, 65.4147041, 64.7514791, 64.1378384, 63.5693425, 63.0415519, 62.5500272, 62.0903289, 61.6580176, 61.2486538, 60.857798, 60.4810107, 60.1138526, 59.7518841, 59.3906658, 59.0257582, 58.6527218, 58.2671173, 57.7170687, 57.074956, 56.414931, 55.7373388, 55.0425246, 54.3308335, 53.6026107, 52.8582012, 52.0979502, 51.3222029, 50.5313043, 49.7255996, 48.9054338, 48.0711523, 47.2231, 46.3616221, 45.4870637, 44.6108436, 43.7219933, 42.8206177, 41.906822, 40.9807113, 40.0423907, 39.0919651, 38.1295397, 37.1552195, 36.1691097, 35.1713153, 34.1619414, 33.141093, 32.1088753, 31.0653932, 30.010752, 28.9059503, 27.7129262, 26.5124381, 25.3071264, 24.0996311, 22.8925925, 21.6886506, 20.4904458, 19.3006181, 18.1218077, 16.9566549, 15.8077996, 14.6778823, 13.5695429, 12.4854217, 11.4281589, 10.4003946, 9.79353317, 9.40765559, 9.03817014, 8.67510083, 8.30847168, 7.92830674, 7.52463001, 7.08746552, 6.6068373, 6.07276936, 5.47528574, 4.80441047, 4.05016755, 3.20258102, 2.2516749, 1.18747322, 0, -4.05636366, -7.90749239, -11.5632041, -15.0333168, -18.3276484, -21.4560169, -24.4282401, -27.2541361, -29.9435228, -32.5062181, -34.9520401, -37.2908065, -39.5323355, -41.6864448, -43.7629526, -45.7716767, -47.4900961, -48.8948376, -50.5212476, -52.3293571, -54.279197, -56.3307986, -58.4441928, -60.5794106, -62.6964832, -64.7554415, -66.7163166, -68.5391395, -70.1839413, -71.6468944, -72.8386366, -73.7083566, -74.2052428, -74.545252, -74.5441887, -73.9467193, -72.6220022, -70.4391954, -67.2674571, -62.9759457, -57.7152003, -50.9873603, -42.6065312, -32.3868182, -20.1423267, -5.27373807, 11.8880623, 31.425613, 53.4214526, 77.9581198, 104.409721, 134.433088, 168.496803, 207.14611, 251.082774, 300.938285, 357.344135, 421.877297, 494.008896, 572.17862, 655.330525, 744.470377, 840.853971, 944.748708, 1054.74721, 1170.21086, 1290.48609, 1414.93043, 1542.95608, 1673.94432, 1809.33613, 1946.1362, 2083.01281, 2218.63424, 2351.6688, 2458.52768, 2565.70053, 2677.41992, 2797.91841, 2931.42855, 3090.35536, 3265.31066, 3457.02456, 3680.19471, 3930.02238, 4195.88901, 4473.28683, 4757.70811, 5044.64509, 5329.59002, 5571.66562, 5881.93604, 6194.87592, 6444.95991, 6566.66264, 6494.45878, 6162.82295, 5643.25699, 5062.04475},
        {89.4250614, 87.1675839, 85.0387414, 83.0339905, 81.1487876, 79.3785894, 77.7188524, 76.1650331, 74.7125881, 73.356974, 72.0936473, 70.9180645, 69.8256822, 68.811957, 67.8723454, 67.002304, 66.1972893, 65.4527579, 64.7641664, 64.1269712, 63.5366289, 62.9885961, 62.4783294, 62.0012853, 61.5529203, 61.128691, 60.724054, 60.3344657, 59.9553829, 59.5822619, 59.2105594, 58.835732, 58.4532361, 58.0585283, 57.4951412, 56.8376889, 56.1626114, 55.4702954, 54.7611279, 54.0354956, 53.2937851, 52.5363835, 51.7636773, 50.9760534, 50.1738986, 49.3575996, 48.5275432, 47.6841163, 46.8277055, 45.9586977, 45.0774797, 44.1975731, 43.3059451, 42.4026976, 41.4879328, 40.5617527, 39.6242594, 38.6755547, 37.7157409, 36.7449199, 35.7631938, 34.7706647, 33.7674345, 32.7536052, 31.7292791, 30.694558, 29.649544, 28.5567655, 27.3796528, 26.195801, 25.007748, 23.8180314, 22.6291891, 21.4437589, 20.2642786, 19.0932859, 17.9333187, 16.7869147, 15.6566117, 14.5449476, 13.4544601, 12.387687, 11.3471662, 10.3354353, 9.73685641, 9.35502407, 8.98894355, 8.62876171, 8.26462541, 7.88668153, 7.48507691, 7.04995842, 6.57147292, 6.03976727, 5.44498834, 4.77728298, 4.02679806, 3.18368044, 2.23807699, 1.18013455, 0, -4.19432618, -8.15687856, -11.8986841, -15.4307697, -18.7641623, -21.9098889, -24.8789764, -27.6824518, -30.3313421, -32.8366741, -35.2094748, -37.4607712, -39.6015902, -41.6429587, -43.5959038, -45.4714523, -47.0327875, -48.2555301, -49.7113059, -51.3580325, -53.1536275, -55.0560084, -57.0230929, -59.0127984, -60.9830427, -62.8917432, -64.6968176, -66.3561835, -67.8277583, -69.0556327, -70.0166212, -70.6727895, -70.9862034, -71.3834887, -71.5944311, -71.2249334, -70.0976933, -68.0354086, -64.8607768, -60.3964956, -54.2099624, -46.4568752, -37.0098817, -25.7416297, -12.5247668, 2.51616025, 19.8233775, 39.587212, 61.9979909, 87.2460411, 115.128377, 146.709354, 182.446726, 222.793609, 268.193649, 319.094744, 375.944793, 439.652316, 510.099901, 586.667637, 669.027025, 757.710391, 854.040478, 958.178959, 1068.32014, 1183.56137, 1302.6869, 1424.71291, 1549.80366, 1677.47398, 1809.94168, 1943.60279, 2077.0365, 2208.82199, 2337.53847, 2437.81073, 2538.16095, 2643.15693, 2757.36644, 2885.35727, 3041.52691, 3214.06032, 3403.31259, 3622.99352, 3868.17194, 4128.35217, 4398.88123, 4675.10617, 4952.37401, 5226.03178, 5288.84932, 5551.09811, 5897.06654, 6211.04298, 6377.3158, 6280.17339, 5803.90412, 5316.98108, 4774.54318},
        {90.3986141, 88.0600549, 85.8541637, 83.7762845, 81.8217615, 79.9859386, 78.2641599, 76.6517694, 75.1441112, 73.7365293, 72.4243676, 71.2029704, 70.0676815, 69.0138451, 68.0368051, 67.1319057, 66.2944908, 65.5199044, 64.8034907, 64.1405937, 63.5265573, 62.9567257, 62.4264429, 61.9310528, 61.4658996, 61.0263273, 60.6076798, 60.2053013, 59.8145359, 59.4307274, 59.04922, 58.6653577, 58.2744845, 57.8719445, 57.2973734, 56.6272445, 55.9399026, 55.2357448, 54.5151681, 53.7785697, 53.0263465, 52.2588957, 51.4766143, 50.6798994, 49.8691481, 49.0447574, 48.2071244, 47.3566462, 46.4937199, 45.6187425, 44.732111, 43.8467876, 42.9503319, 42.0428686, 41.1245225, 40.1954182, 39.2556804, 38.3054337, 37.344803, 36.3739128, 35.3928878, 34.4018528, 33.4009324, 32.3902513, 31.3699342, 30.3401058, 29.3008908, 28.2166708, 27.0526857, 25.8827775, 24.7093879, 23.5349587, 22.3619317, 21.1927485, 20.0298511, 18.8756811, 17.7326802, 16.6032904, 15.4899532, 14.3951106, 13.3212041, 12.2706757, 11.2459671, 10.24952, 9.65928083, 9.28204083, 8.9200779, 8.56364794, 8.20300688, 7.82841062, 7.43011508, 6.99837617, 6.52344981, 5.9955919, 5.40505836, 4.7421051, 3.99698804, 3.15996308, 2.22128615, 1.17121315, 0, -4.32404258, -8.38967565, -12.2093551, -15.7955367, -19.1606765, -22.3172302, -25.2776538, -28.054403, -30.6599339, -33.1067022, -35.4071637, -37.5737745, -39.6189903, -41.555267, -43.3950606, -45.1508267, -46.5226468, -47.4783516, -48.7369589, -50.2439869, -51.9449541, -53.7853791, -55.7107802, -57.6666759, -59.5985846, -61.4520248, -63.172515, -64.7055735, -65.9967188, -66.8566518, -67.4151417, -67.658152, -67.5716466, -67.8002389, -68.0005676, -67.6536318, -66.5478002, -64.4714416, -61.2129247, -56.5606184, -49.6972282, -41.2011184, -30.9791572, -18.9382127, -4.98515306, 10.3719399, 27.9777151, 48.0756079, 70.9090535, 96.7214873, 125.503098, 158.060091, 194.884082, 236.396153, 282.873382, 334.6575, 392.09024, 455.771489, 525.726154, 601.697519, 683.610905, 771.815258, 867.577178, 971.015145, 1080.30221, 1194.50439, 1312.37596, 1432.90211, 1556.21124, 1681.78507, 1811.66971, 1942.38802, 2072.53407, 2200.70194, 2325.4857, 2420.76739, 2516.03112, 2616.04899, 2725.59307, 2849.43545, 3003.46075, 3173.92017, 3360.82325, 3576.93331, 3817.21155, 4071.30513, 4334.44097, 4601.84598, 4868.74708, 5130.37117, 5088.12788, 5316.77254, 5672.72959, 6012.42347, 6192.27863, 6068.71952, 5498.17058, 5017.12271, 4518.20335},
        {91.488041, 89.0589591, 86.7670507, 84.6075389, 82.5756468, 80.6665974, 78.875614, 77.1979196, 75.6287373, 74.1632903, 72.7968017, 71.5244946, 70.3415921, 69.2433173, 68.2248934, 67.2815434, 66.4084906, 65.6009579, 64.8541685, 64.1633455, 63.5237121, 62.9304914, 62.3789064, 61.8641804, 61.3815363, 60.9261973, 60.4933866, 60.0783273, 59.6762424, 59.2823552, 58.8918886, 58.5000658, 58.10211, 57.6932443, 57.1100165, 56.430325, 55.733975, 55.0213391, 54.2927897, 53.5486991, 52.7894398, 52.0153843, 51.226905, 50.4243743, 49.6081646, 48.7786484, 47.9361981, 47.0811862, 46.213985, 45.334967, 44.4445046, 43.5521017, 42.6488013, 41.7347779, 40.8102061, 39.8752602, 38.9301148, 37.9749443, 37.0099232, 36.0352261, 35.0510273, 34.0575014, 33.0548228, 32.043166, 31.0227056, 29.9936159, 28.9560715, 27.8765762, 26.7224424, 25.5632876, 24.4014691, 23.2393439, 22.0792694, 20.9236027, 19.7747011, 18.6349217, 17.5066217, 16.3921584, 15.293889, 14.2141707, 13.1553607, 12.1198162, 11.1098943, 10.1279524, 9.5464817, 9.1749871, 8.81852476, 8.46744033, 8.11207945, 7.74278776, 7.3499109, 6.92379452, 6.45478427, 5.93322579, 5.34946472, 4.6938467, 3.95671738, 3.12842241, 2.19930743, 1.15971808, 0, -4.43547038, -8.58737008, -12.4697297, -16.0965799, -19.4819512, -22.6398743, -25.5843798, -28.3294982, -30.8892603, -33.2776965, -35.5088376, -37.5967141, -39.5553566, -41.3987958, -43.1410623, -44.7961866, -45.9509996, -46.5585033, -47.5990137, -48.9950185, -50.6690055, -52.5434624, -54.5408772, -56.5837376, -58.5945313, -60.4957463, -62.2098702, -63.659391, -64.7667964, -65.1288036, -65.1131202, -64.7399652, -64.0295576, -63.8370488, -63.7651916, -63.1940899, -61.8934829, -59.6331098, -56.1827097, -51.3120217, -44.0659741, -35.1597117, -24.5047846, -12.0127428, 2.40486364, 18.2736505, 36.3856105, 56.969902, 80.2556836, 106.472114, 135.521956, 168.359691, 205.569457, 247.613275, 294.703839, 347.165786, 405.323753, 469.880023, 540.695757, 617.220139, 699.093918, 786.812419, 881.458502, 983.177449, 1090.52587, 1202.78961, 1319.26374, 1439.23648, 1561.96223, 1686.71451, 1814.37177, 1942.35573, 2069.38435, 2194.17562, 2315.44752, 2407.5029, 2499.57867, 2596.49656, 2703.07834, 2824.14579, 2976.4202, 3144.89076, 3329.27939, 3541.57046, 3776.63206, 4024.2229, 4279.46655, 4537.48653, 4793.40639, 5042.34968, 4987.66325, 5201.16695, 5538.95304, 5857.11379, 6011.74146, 5858.92833, 5254.76666, 4744.33607, 4288.88213},
        {92.6936721, 90.1673848, 87.7830924, 85.5358893, 83.4208702, 81.4331294, 79.5677614, 77.8198608, 76.1845219, 74.6568393, 73.2319074, 71.9048207, 70.6706737, 69.5245609, 68.4615767, 67.4768157, 66.5653722, 65.7223408, 64.942816, 64.2218921, 63.5546638, 62.9362255, 62.3616716, 61.8260967, 61.3245951, 60.8522615, 60.4041902, 59.9754758, 59.5612126, 59.1564953, 58.7564182, 58.3560759, 57.9505628, 57.5349734, 56.9427714, 56.2532663, 55.5477104, 54.8264434, 54.089805, 53.3381349, 52.5717728, 51.7910584, 50.9963314, 50.1879316, 49.3661986, 48.5314722, 47.684092, 46.8243978, 45.9527293, 45.0694261, 44.174828, 43.2742904, 42.3630286, 41.4412735, 40.5092562, 39.5672075, 38.6153585, 37.6539402, 36.6831835, 35.7033195, 34.714579, 33.7171931, 32.7113928, 31.6974089, 30.6754726, 29.6458147, 28.6086663, 27.5327307, 26.3874682, 25.2381332, 24.0870004, 22.9363447, 21.7884407, 20.6455634, 19.5099875, 18.3839877, 17.269839, 16.1698159, 15.0861934, 14.0212462, 12.9772492, 11.956477, 10.9612045, 9.99370646, 9.42087667, 9.05501115, 8.70402347, 8.35835583, 8.00845042, 7.64474945, 7.25769512, 6.83772961, 6.37529513, 5.86083387, 5.28478804, 4.63759982, 3.90971143, 3.09156505, 2.17360289, 1.14626714, 0, -4.53616176, -8.76439912, -12.7003289, -16.3595678, -19.7577328, -22.9104406, -25.8333079, -28.5419517, -31.0519887, -33.3790356, -35.5387094, -37.5466267, -39.4184044, -41.1696594, -42.8160083, -44.373068, -45.3039959, -45.5454859, -46.3627656, -47.6530674, -49.3136242, -51.2416684, -53.3344328, -55.48915, -57.6030525, -59.5733731, -61.2973444, -62.672199, -63.5951696, -63.4370243, -62.8144972, -61.7827603, -60.3969855, -59.7037135, -59.2624318, -58.3682633, -56.7789694, -54.2523116, -50.5460511, -45.4179495, -37.8541358, -28.6188484, -17.6208203, -4.76878457, 10.0285259, 26.4171135, 45.044826, 66.1142467, 89.8279587, 116.388545, 145.560326, 178.519803, 215.945981, 258.35409, 305.924992, 358.989673, 417.87912, 483.498162, 555.473523, 632.80592, 714.723877, 801.953341, 895.372537, 995.173802, 1100.38728, 1210.57659, 1325.70727, 1445.44716, 1567.99027, 1692.36427, 1818.07781, 1943.60362, 2067.8029, 2189.53683, 2307.66663, 2397.23456, 2486.87547, 2581.40528, 2685.63991, 2804.39529, 2954.9312, 3121.32381, 3303.05595, 3511.55337, 3741.57153, 3982.88965, 4230.51199, 4479.44281, 4724.68635, 4961.24689, 4928.00813, 5133.5459, 5443.47368, 5723.40498, 5838.95328, 5655.73208, 5039.35489, 4497.32324, 4086.42167},
        {94.0084985, 91.3788883, 88.8963744, 86.5559154, 84.3524696, 82.2809954, 80.3364513, 78.5137957, 76.8079871, 75.2139837, 73.7267442, 72.3412268, 71.05239, 69.8551922, 68.7445919, 67.7155474, 66.7630173, 65.8819598, 65.0673335, 64.3140968, 63.617208, 62.9716256, 62.372308, 61.8142137, 61.292301, 60.8015285, 60.3368544, 59.8932373, 59.4656355, 59.0490074, 58.6383116, 58.2285064, 57.8145502, 57.3914015, 56.7892679, 56.0889486, 55.3732228, 54.6423954, 53.8967713, 53.1366555, 52.3623527, 51.574168, 50.7724062, 49.9573723, 49.129371, 48.2887074, 47.4356863, 46.5706127, 45.6937913, 44.8055272, 43.9061252, 42.9966352, 42.076601, 41.1463114, 40.206055, 39.2561206, 38.2967971, 37.328373, 36.3511372, 35.3653784, 34.3713854, 33.3694469, 32.3598517, 31.3428885, 30.3188461, 29.2880132, 28.2506785, 27.1777241, 26.040738, 24.9006582, 23.7596797, 22.6199977, 21.4838071, 20.3533031, 19.2306807, 18.118135, 17.0178611, 15.932054, 14.8629088, 13.8126206, 12.7833845, 11.7773955, 10.7968486, 9.84393904, 9.27970224, 8.91936534, 8.57384246, 8.23368363, 7.88943885, 7.53165812, 7.15089145, 6.73768886, 6.28260035, 5.77617592, 5.20896559, 4.57151936, 3.85438723, 3.04811923, 2.14326535, 1.13037561, 0, -4.6216105, -8.91261736, -12.8901419, -16.5713056, -19.9732297, -23.1130357, -26.0078449, -28.6747786, -31.1309582, -33.3935052, -35.4795408, -37.4061865, -39.1905636, -40.8497934, -42.4009974, -43.8612969, -44.5713327, -44.4515862, -45.0493132, -46.236675, -47.8858329, -49.8689481, -52.0581818, -54.3256954, -56.54365, -58.584207, -60.3195275, -61.6217729, -62.3631043, -61.6984588, -60.4802042, -58.7956689, -56.7321814, -55.4947197, -54.6117875, -53.313849, -51.3529379, -48.4810878, -44.4503326, -39.0127059, -31.1548889, -21.6272013, -10.3314198, 2.83067889, 17.9573182, 34.8502725, 53.9783263, 75.5138151, 99.6290746, 126.49644, 155.728146, 188.741197, 226.30366, 268.985975, 316.955096, 370.559136, 430.146205, 496.861106, 570.074194, 648.28671, 730.272155, 817.021301, 909.217323, 1007.10933, 1110.21091, 1218.37327, 1332.24978, 1451.89962, 1574.4407, 1698.65468, 1822.69551, 1946.06914, 2067.7745, 2186.81056, 2302.17624, 2389.7882, 2477.49825, 2570.07593, 2672.29075, 2788.91225, 2937.45175, 3101.44243, 3280.19301, 3484.89409, 3710.15118, 3945.61903, 4186.15067, 4426.59918, 4661.81758, 4886.65894, 4902.90421, 5106.87187, 5381.55769, 5609.95742, 5675.06684, 5459.88171, 4847.39778, 4276.1242, 3907.9328},
        {95.4255111, 92.687026, 90.1009827, 87.6621965, 85.3654826, 83.2056562, 81.1775326, 79.2759271, 77.4956549, 75.8315313, 74.2783715, 72.8309908, 71.4842044, 70.2328276, 69.0716757, 67.9955639, 66.9993074, 66.0777215, 65.2256215, 64.4378226, 63.7091401, 63.0343892, 62.4083852, 61.8259433, 61.2818789, 60.771007, 60.2881431, 59.8281023, 59.3856999, 58.9557512, 58.5330714, 58.1124758, 57.6887796, 57.2567981, 56.643136, 55.9302519, 55.2026262, 54.4605328, 53.7042459, 52.9340393, 52.1501872, 51.3529634, 50.5426421, 49.7194972, 48.8838028, 48.0358328, 47.1758613, 46.3041622, 45.4210096, 44.5266775, 43.6214399, 42.7024175, 41.7731059, 40.8338476, 39.8849849, 38.9268601, 37.9598157, 36.984194, 36.0003374, 35.0085882, 34.0092889, 33.0027817, 31.989409, 30.9695133, 29.9434369, 28.9115221, 27.8741113, 26.8041463, 25.6752268, 24.5442068, 23.4132049, 22.2843397, 21.1597299, 20.0414941, 18.931751, 17.8326192, 16.7462172, 15.6746639, 14.6200777, 13.5845774, 12.5702815, 11.5793088, 10.6137777, 9.67580705, 9.12019491, 8.76530202, 8.42525019, 8.09071293, 7.75236375, 7.40087615, 7.02692364, 6.62117973, 6.17431792, 5.67701172, 5.11993464, 4.49376018, 3.78916185, 2.99681317, 2.10738762, 1.11155873, 0, -4.68731041, -9.02387942, -13.0281583, -16.7185982, -20.1136505, -23.2317664, -26.0913972, -28.7109941, -31.1090083, -33.3038911, -35.3140938, -37.1580677, -38.8542638, -40.4211337, -41.8771283, -43.2406991, -43.7427072, -43.289091, -43.6797555, -44.7643825, -46.3926542, -48.4142524, -50.6788591, -53.0361562, -55.3358257, -57.4275495, -59.1610096, -60.3858879, -60.9518664, -59.8302523, -58.071173, -55.7878229, -53.0933962, -51.3045543, -49.9327581, -48.168544, -45.7640663, -42.4714794, -38.0429376, -32.2305951, -24.0614089, -14.2334431, -2.64073834, 10.8226645, 26.2627246, 43.621071, 63.2090757, 85.1737804, 109.662227, 136.821457, 166.135354, 199.224645, 236.932505, 279.876308, 328.212408, 382.304148, 442.51487, 510.204054, 584.51251, 663.494357, 745.510126, 831.799574, 922.890898, 1019.08914, 1120.32122, 1226.68762, 1339.43448, 1458.9593, 1581.45885, 1705.50605, 1828.13255, 1949.68969, 2069.28394, 2186.02175, 2299.00957, 2384.98967, 2471.02375, 2561.80931, 2662.04387, 2776.42493, 2922.43987, 3083.46974, 3258.73067, 3459.60467, 3680.49223, 3910.72469, 4144.95597, 4377.84, 4604.0307, 4818.18201, 4906.09317, 5114.10735, 5348.47123, 5515.43152, 5521.23492, 5272.12813, 4674.35785, 4080.77889, 3750.5264},
        {96.8658079, 94.0016932, 91.2969507, 88.7461885, 86.3440145, 84.0850367, 81.963863, 79.9751016, 78.1133602, 76.373247, 74.7493699, 73.2363368, 71.8287558, 70.5212348, 69.3083818, 68.1848048, 67.1451118, 66.1839107, 65.2958095, 64.4754162, 63.7173388, 63.0161853, 62.3665636, 61.7630817, 61.2003476, 60.6729692, 60.1755546, 59.7027118, 59.2490486, 58.8091731, 58.3776933, 57.9492171, 57.5183526, 57.0797076, 56.4521904, 55.7219545, 54.977265, 54.2184319, 53.4457649, 52.6595737, 51.8601681, 51.047858, 50.222953, 49.385763, 48.5365977, 47.675767, 46.8035805, 45.9203481, 45.0263795, 44.1219846, 43.207473, 42.2818098, 41.3466787, 40.4024186, 39.4493684, 38.4878672, 37.5182537, 36.540867, 35.556046, 34.5641296, 33.5654567, 32.5603663, 31.5491972, 30.5322885, 29.509979, 28.4826077, 27.4505135, 26.3872765, 25.2671193, 24.1455689, 23.0246988, 21.9065827, 20.7932942, 19.6869069, 18.5894943, 17.50313, 16.4298877, 15.3718409, 14.3310633, 13.3096284, 12.3096098, 11.3330811, 10.3821159, 9.45878779, 8.91199972, 8.56311552, 8.22933041, 7.9014627, 7.57033072, 7.22675276, 6.86154716, 6.46553223, 6.02952628, 5.54434763, 5.00081459, 4.38974549, 3.70195863, 2.92827234, 2.05950493, 1.08647471, 0, -4.78291077, -9.18779278, -13.2347925, -16.9440563, -20.3357308, -23.4299623, -26.2468974, -28.8066825, -31.129464, -33.2353884, -35.1446023, -36.8772519, -38.4534839, -39.8934446, -41.2172806, -42.4451382, -42.7807946, -42.127772, -42.3527666, -43.3009888, -44.8176486, -46.7479561, -48.9371216, -51.2303553, -53.4728671, -55.5098674, -57.1865663, -58.348174, -58.8399005, -57.6559562, -55.8045844, -53.3862953, -50.5015989, -48.3770443, -46.5501221, -44.2580461, -41.2635148, -37.3292266, -32.21788, -25.6921736, -16.7036202, -6.07298652, 6.27831874, 20.4288868, 36.4573091, 53.8701822, 73.4610907, 95.4516243, 120.063373, 147.517926, 177.570302, 211.478918, 250.076227, 293.928181, 343.056622, 397.727689, 458.207517, 525.51276, 598.988466, 677.16094, 758.733832, 844.74827, 935.841684, 1032.28054, 1133.70276, 1240.03474, 1352.1016, 1470.06275, 1590.78247, 1712.98898, 1834.9134, 1955.85747, 2074.72148, 2190.40569, 2301.81036, 2384.13036, 2465.89771, 2551.93905, 2647.081, 2756.15018, 2896.73652, 3052.39447, 3222.48067, 3420.45704, 3640.09687, 3869.29628, 4102.1224, 4332.64242, 4554.92348, 4763.03275, 4960.30058, 5183.93301, 5370.96352, 5458.42562, 5383.3528, 5082.77856, 4493.7364, 3930.49077, 3607.10555},
        {98.4092025, 95.4174809, 92.5919478, 89.9270095, 87.4170719, 85.0565412, 82.8398235, 80.7613248, 78.8154512, 76.9966089, 75.299204, 73.7176425, 72.2463305, 70.8796742, 69.6120796, 68.4379529, 67.3517001, 66.3477272, 65.4204406, 64.5642461, 63.7735499, 63.0427582, 62.3662769, 61.7385123, 61.1538703, 60.6067572, 60.0915789, 59.6027416, 59.1346514, 58.6817144, 58.2383367, 57.7989243, 57.3578834, 56.9096201, 56.2653901, 55.5150271, 54.7505659, 53.9723561, 53.1807476, 52.3760899, 51.5587327, 50.7290257, 49.8873186, 49.0339611, 48.1693027, 47.2936933, 46.4074824, 45.5110198, 44.6046552, 43.6887381, 42.7636183, 41.8305376, 40.8889342, 39.9391384, 38.9814806, 38.016291, 37.0439002, 36.0646382, 35.0788356, 34.0868227, 33.0889297, 32.085487, 31.076825, 30.063274, 29.0451643, 28.0228263, 26.9965903, 25.9406099, 24.8296673, 23.7180803, 22.6078762, 21.5010822, 20.3997255, 19.3058335, 18.2214333, 17.1485523, 16.0892178, 15.0454569, 14.0192969, 13.0127652, 12.027889, 11.0666955, 10.131212, 9.22346574, 8.68567433, 8.34261442, 8.01497384, 7.69374004, 7.36990046, 7.03444251, 6.67835365, 6.29262131, 5.86823291, 5.39617589, 4.86743769, 4.27300575, 3.60386749, 2.85101035, 2.00542176, 1.05808917, 0, -4.83850383, -9.27815957, -13.3403166, -17.0463243, -20.4175321, -23.4752894, -26.2409455, -28.7358498, -30.9813518, -32.9988007, -34.8095461, -36.4349372, -37.8963235, -39.2150543, -40.412479, -41.5099471, -41.7079252, -40.9104937, -40.9957605, -41.8091716, -43.1961729, -45.0022103, -47.0727299, -49.2531775, -51.3889991, -53.3256406, -54.9085481, -55.9831675, -56.3949446, -55.2057449, -53.3319078, -50.8539535, -47.8524021, -45.4186616, -43.1278083, -40.2853483, -36.6685353, -32.0546232, -26.2208658, -18.9445169, -9.1303184, 2.30641696, 15.4177266, 30.2556479, 46.8722183, 64.3259014, 83.9107018, 105.92705, 130.675377, 158.456114, 189.323634, 224.125162, 263.665421, 308.446684, 358.353731, 413.548582, 474.193259, 540.908344, 613.293079, 690.446457, 771.49635, 857.248605, 948.510705, 1045.50591, 1147.46754, 1254.08332, 1365.60502, 1481.8666, 1600.63381, 1720.8423, 1842.20087, 1962.75305, 2081.16688, 2196.11036, 2306.25151, 2385.26624, 2463.0627, 2544.55693, 2634.66498, 2738.30288, 2872.82917, 3022.4224, 3186.66614, 3381.45287, 3599.91052, 3828.37161, 4060.30385, 4289.17491, 4508.4525, 4711.60431, 5024.26541, 5266.17583, 5406.27987, 5413.52179, 5256.84588, 4905.19641, 4327.51766, 3802.15674, 3479.86266},
        {100.067663, 96.9500297, 94.0048871, 91.2264607, 88.6089755, 86.1466565, 83.8337288, 81.6644175, 79.6329475, 77.7335439, 75.9604318, 74.3078361, 72.769982, 71.3410944, 70.0153984, 68.787119, 67.6504813, 66.5997102, 65.629031, 64.7326684, 63.9048477, 63.1397939, 62.4317319, 61.7748868, 61.1634837, 60.5917476, 60.0539035, 59.5441765, 59.0567916, 58.5859739, 58.1259483, 57.6709399, 57.2151738, 56.752875, 56.0884587, 55.3150472, 54.5280444, 53.7278348, 52.9148032, 52.0893343, 51.251813, 50.4026238, 49.5421516, 48.6707811, 47.7888969, 46.8968839, 45.9951268, 45.0840102, 44.163919, 43.2352378, 42.2983514, 41.3564603, 40.4070723, 39.4505113, 38.4871008, 37.5171647, 36.5410264, 35.5590099, 34.5714387, 33.5786365, 32.580927, 31.578634, 30.5720811, 29.561592, 28.5474903, 27.5300999, 26.5097443, 25.4613059, 24.3602781, 23.2594551, 22.1608101, 21.0663159, 19.9779455, 18.8976719, 17.8274679, 16.7693066, 15.7251608, 14.6970036, 13.6868079, 12.6965465, 11.7281926, 10.7837189, 9.86509857, 8.97430441, 8.44573855, 8.10816718, 7.78636091, 7.47150474, 7.15478364, 6.82738261, 6.48048661, 6.10528064, 5.69294967, 5.23467869, 4.72165268, 4.14505663, 3.4960755, 2.7658943, 1.94569799, 1.02667156, 0, -4.83528792, -9.26103054, -13.2990277, -16.9710792, -20.298985, -23.3045449, -26.0095587, -28.4358262, -30.6051474, -32.5393221, -34.2601501, -35.7894312, -37.1489654, -38.3605524, -39.4459921, -40.4270844, -40.5209291, -39.6343688, -39.6093583, -40.2952617, -41.5414431, -43.1972666, -45.1120964, -47.1352966, -49.1162313, -50.9042646, -52.3487607, -53.2990836, -53.6045976, -52.4214694, -50.5464328, -48.0368111, -44.9499276, -42.2205014, -39.4671579, -36.0745503, -31.8367835, -26.5479618, -20.0021901, -11.9935728, -1.42497856, 10.7470062, 24.5539045, 40.0272395, 57.1985343, 74.747724, 94.3958166, 116.512232, 141.466391, 169.627713, 201.32621, 237.018876, 277.474376, 323.129844, 373.745538, 429.38562, 490.114252, 556.163911, 627.392189, 703.473061, 783.958499, 869.433132, 960.914686, 1058.58998, 1161.23788, 1268.28354, 1379.35367, 1493.92567, 1610.73786, 1728.94665, 1849.73982, 1969.92985, 2087.96993, 2202.31329, 2311.41314, 2387.83631, 2462.39398, 2540.01096, 2625.61204, 2724.12204, 2852.24237, 2995.27014, 3153.08302, 3344.09923, 3560.85749, 3788.17128, 4018.96557, 4246.16535, 4462.69557, 4661.48123, 5084.07846, 5344.36978, 5441.37587, 5374.1174, 5141.61502, 4742.88942, 4176.96126, 3690.75814, 3367.01932},
        {101.85316, 98.6149803, 95.5546814, 92.6663431, 89.9440457, 87.3818693, 84.9738942, 82.7142006, 80.5968688, 78.6159789, 76.7656113, 75.039846, 73.4327635, 71.9384438, 70.5509673, 69.2644141, 68.0728645, 66.9703988, 65.951097, 65.0090395, 64.1383066, 63.3329783, 62.5871351, 61.894857, 61.2502243, 60.6473173, 60.0802161, 59.5430011, 59.0297524, 58.5345502, 58.0514749, 57.5746066, 57.0980255, 56.6158119, 55.9271198, 55.1275924, 54.3152162, 53.4903969, 52.6535406, 51.8050533, 50.9453409, 50.0748093, 49.1938646, 48.3029126, 47.4023594, 46.492611, 45.5740732, 44.647152, 43.7122534, 42.7697834, 41.8201479, 40.8674372, 39.908293, 38.9430415, 37.9720088, 36.995521, 36.013904, 35.0274842, 34.0365875, 33.04154, 32.0426678, 31.0402971, 30.0347539, 29.0263643, 28.0154544, 27.0023503, 25.9873781, 24.9465239, 23.8563585, 22.7674071, 21.6815736, 20.600762, 19.5268762, 18.4618202, 17.4074979, 16.3658133, 15.3386704, 14.327973, 13.3356252, 12.3635308, 11.4135939, 10.4877184, 9.58780818, 8.71576727, 8.1967122, 7.86414228, 7.54767205, 7.23871657, 6.92869094, 6.60901024, 6.27108953, 5.9063439, 5.50618844, 5.06203821, 4.5653083, 4.00741379, 3.37976976, 2.67379129, 1.88089345, 0.992491327, 0, -4.75446141, -9.10245642, -13.0652227, -16.663998, -19.9200198, -22.8545259, -25.488754, -27.8439417, -29.9413266, -31.8021465, -33.447639, -34.8990418, -36.1775925, -37.3045288, -38.3010883, -39.1885088, -39.2166364, -38.2965098, -38.1941811, -38.7655897, -39.866675, -41.3533766, -43.0816338, -44.9073863, -46.6865735, -48.2751347, -49.5290096, -50.3041376, -50.4564582, -49.2449806, -47.3414489, -44.7808817, -41.5982975, -38.5736587, -35.369512, -31.4497522, -26.6259147, -20.7095348, -13.5121479, -4.84528947, 6.32892423, 19.09102, 33.4632717, 49.4679535, 67.1273391, 84.8951454, 104.754343, 127.119345, 152.404564, 181.024415, 213.508885, 250.015554, 291.277381, 337.675683, 388.873848, 444.857597, 505.612655, 571.052567, 641.251637, 716.362908, 796.281096, 881.434404, 973.070352, 1071.35748, 1174.63607, 1282.0856, 1392.75648, 1505.79473, 1620.81961, 1737.18266, 1857.27513, 1976.94127, 2094.48044, 2208.19202, 2316.37537, 2391.27956, 2463.76683, 2538.64914, 2620.73844, 2714.84667, 2836.5007, 2972.65428, 3123.52725, 3309.90318, 3523.8621, 3748.91585, 3977.57286, 4202.34161, 4415.73053, 4610.24808, 5125.83054, 5402.04879, 5463.20712, 5333.6098, 5037.56112, 4599.36537, 4043.32682, 3591.27635, 3266.79711},
        {103.777659, 100.427973, 97.2602437, 94.268458, 91.4466032, 88.7886663, 86.2886345, 83.9404949, 81.7382346, 79.6758407, 77.7473003, 75.9466005, 74.2677284, 72.704671, 71.2514156, 69.9019492, 68.6502589, 67.4903318, 66.416155, 65.4217157, 64.5010008, 63.6479976, 62.8566931, 62.1210744, 61.4351286, 60.7928429, 60.1882043, 59.6152, 59.0678169, 58.5400423, 58.0258633, 57.5192669, 57.0142402, 56.5047704, 55.787097, 54.9582402, 54.117597, 53.2655718, 52.4025691, 51.5289931, 50.6452484, 49.7517393, 48.8488701, 47.9370454, 47.0166694, 46.0881465, 45.1518813, 44.2082779, 43.2577409, 42.3006746, 41.3374835, 40.3713278, 39.3997962, 38.4232334, 37.4419841, 36.4563929, 35.4668045, 34.4735634, 33.4770144, 32.4775021, 31.4753712, 30.4709662, 29.4646319, 28.4567129, 27.4475539, 26.4374994, 25.4268942, 24.3934233, 23.3153158, 22.2396501, 21.1682397, 20.1028983, 19.0454393, 17.9976764, 16.9614231, 15.9384931, 14.9306998, 13.939857, 12.967778, 12.0162766, 11.0871663, 10.1822607, 9.30337334, 8.45231781, 7.9431151, 7.61490822, 7.30308768, 6.99933534, 6.69533303, 6.3827626, 6.0533059, 5.69864478, 5.31046107, 4.88043663, 4.4002533, 3.86159292, 3.25613735, 2.57556843, 1.811568, 0.955817906, 0, -4.57722263, -8.76848797, -12.5931987, -16.0707575, -19.220567, -22.0620298, -24.6145487, -26.8975262, -28.9303651, -30.7324681, -32.3232377, -33.7220766, -34.9483875, -36.0215731, -36.961036, -37.7861789, -37.791877, -36.8940293, -36.7508501, -37.2264862, -38.1850846, -39.4907918, -41.0077549, -42.6001205, -44.1320354, -45.4676466, -46.4711007, -47.0065447, -46.9381252, -45.6181294, -43.6102458, -40.9321789, -37.6016336, -34.2692287, -30.636212, -26.2350537, -20.8935844, -14.4396346, -6.70103451, 2.49438527, 14.0479149, 27.1806971, 41.9222475, 58.3020815, 76.3497148, 94.527661, 114.824188, 137.660562, 163.458048, 192.637913, 225.802519, 262.970696, 304.848726, 351.782228, 403.380464, 459.583305, 520.330623, 585.347415, 654.837262, 729.238152, 808.624959, 893.384974, 984.994428, 1083.63317, 1187.28442, 1294.93969, 1405.22239, 1517.0286, 1630.60404, 1745.43099, 1864.55167, 1983.34074, 2100.04821, 2212.92408, 2320.21835, 2395.03497, 2467.05651, 2540.81948, 2620.86041, 2711.7158, 2827.1287, 2956.29141, 3099.79479, 3280.37179, 3489.84869, 3710.8259, 3935.591, 4156.43157, 4365.63518, 4555.48941, 5135.61248, 5422.74683, 5458.72921, 5285.39637, 4944.58505, 4478.132, 3927.87396, 3498.69269, 3177.4176},
        {105.942093, 102.468693, 99.181958, 96.0757651, 93.1439897, 90.3805081, 87.7791961, 85.3339298, 83.0385853, 80.8870387, 78.8731658, 76.9908429, 75.2339458, 73.5963508, 72.0719337, 70.6545707, 69.3381377, 68.1165109, 66.9835662, 65.9331797, 64.9592275, 64.0555855, 63.2161298, 62.4347364, 61.7052815, 61.0216409, 60.3776909, 59.7673073, 59.1843662, 58.6227438, 58.0763159, 57.5389587, 57.0045482, 56.4669605, 55.7195822, 54.8607511, 53.9912146, 53.111355, 52.2215546, 51.3221958, 50.4136608, 49.4963319, 48.5705915, 47.6368217, 46.695405, 45.7467236, 44.7911599, 43.829096, 42.8609144, 41.8869974, 40.9077272, 39.9221221, 38.9319581, 37.9376469, 36.9396005, 35.9382307, 34.9339493, 33.9271683, 32.9182996, 31.9077549, 30.8959462, 29.8832853, 28.8701841, 27.8570544, 26.8443082, 25.8323573, 24.8216136, 23.7928812, 22.7274351, 21.6661265, 20.6106385, 19.562654, 18.5238558, 17.4959271, 16.4805507, 15.4794096, 14.4941867, 13.5265651, 12.5782276, 11.6508572, 10.7461369, 9.86574969, 9.01137843, 8.18470612, 7.68508745, 7.35910839, 7.04977711, 6.74911654, 6.44914963, 6.1418993, 5.8193885, 5.47364015, 5.0966772, 4.68052257, 4.21719921, 3.69873005, 3.11713803, 2.46444607, 1.73267713, 0.913854125, 0, -4.16306344, -8.0031771, -11.5349233, -14.7728844, -17.7316426, -20.4257803, -22.8698798, -25.0785234, -27.0662934, -28.8477722, -30.437542, -31.8501851, -33.1002839, -34.2024207, -35.1711778, -36.0211376, -36.0886063, -35.2912699, -35.177906, -35.6177521, -36.4800458, -37.6340245, -38.9489256, -40.2939868, -41.5384453, -42.5515388, -43.2025048, -43.3605806, -42.8950037, -41.1552327, -38.7208697, -35.6170858, -31.8690523, -28.1281426, -24.1064268, -19.3489876, -13.6931356, -6.97618142, 0.964564512, 10.2917916, 21.5581795, 34.4177348, 48.9568448, 65.2618964, 83.4192768, 102.653416, 124.128148, 148.145348, 175.006892, 205.014658, 238.397138, 275.619281, 317.436775, 364.245403, 415.706157, 471.809933, 532.54763, 597.899175, 667.86893, 742.473228, 821.318076, 905.256751, 996.103247, 1094.19057, 1197.34225, 1304.53373, 1414.42073, 1525.89581, 1639.0237, 1753.20607, 1870.37031, 1987.00383, 2101.63399, 2212.7882, 2318.99384, 2397.72092, 2473.81855, 2551.07848, 2633.29245, 2724.25218, 2834.21242, 2956.19323, 3091.21651, 3262.75295, 3462.93843, 3674.55519, 3890.20679, 4102.49681, 4304.02883, 4487.40641, 5058.77504, 5344.6767, 5383.46371, 5213.48837, 4873.10299, 4400.65987, 3834.51133, 3389.87171, 3095.36968},
        {108.201916, 104.640105, 101.265776, 98.0728404, 95.0552106, 92.2067986, 89.5215167, 86.9932768, 84.615991, 82.3835715, 80.2899304, 78.3289797, 76.4946315, 74.780798, 73.1813911, 71.6903231, 70.3015059, 69.0088517, 67.8062726, 66.6876807, 65.646988, 64.6781067, 63.7749488, 62.9314265, 62.1414518, 61.3989367, 60.6977935, 60.0319342, 59.3952709, 58.7817156, 58.1851805, 57.5995777, 57.0188192, 56.4368172, 55.6503055, 54.7559327, 53.8527173, 52.9409702, 52.0210023, 51.0931245, 50.1576478, 49.2148829, 48.2651409, 47.3087326, 46.345969, 45.3771608, 44.4026192, 43.4226548, 42.4375787, 41.4477018, 40.4533349, 39.4465205, 38.4360172, 37.4223151, 36.4059045, 35.3872754, 34.366918, 33.3453224, 32.3229788, 31.3003774, 30.2780082, 29.2563615, 28.2359274, 27.2171961, 26.2006576, 25.1868021, 24.1761198, 23.1531093, 22.1026532, 21.058223, 20.0213454, 18.9935472, 17.9763554, 16.9712967, 15.9798978, 15.0036857, 14.0441872, 13.102929, 12.1814379, 11.2812409, 10.4038646, 9.55083597, 8.72368175, 7.92392876, 7.43485424, 7.11118356, 6.80425565, 6.50645398, 6.21016203, 5.90776327, 5.59164117, 5.25417919, 4.88776082, 4.48476952, 4.03758876, 3.53860202, 2.98019276, 2.35474446, 1.65464059, 0.872264608, 0, -3.70730377, -7.16314319, -10.3765899, -13.3567153, -16.1125912, -18.6532891, -20.9878806, -23.1254373, -25.0750306, -26.8457324, -28.446614, -29.8867472, -31.1752034, -32.3210544, -33.3333716, -34.2212266, -34.3761299, -33.7189299, -33.6555629, -34.062766, -34.8172762, -35.7958306, -36.8751661, -37.9320198, -38.8431287, -39.48523, -39.7350605, -39.4693575, -38.5648578, -36.3743812, -33.4906852, -29.9476818, -25.7792835, -21.6483316, -17.2742737, -12.2088431, -6.29727313, 0.615202598, 8.68335066, 18.0619376, 28.9351786, 41.4194289, 55.6636932, 71.8169764, 90.0282832, 110.353777, 133.058515, 158.314711, 186.294582, 217.170341, 250.767636, 288.028835, 329.741091, 376.350321, 427.605742, 483.569375, 544.303241, 610.103732, 680.745231, 755.746445, 834.22716, 917.464916, 1007.46982, 1104.70384, 1207.00572, 1313.41832, 1422.82095, 1534.21409, 1647.19775, 1761.03282, 1876.1367, 1990.43581, 2102.79068, 2212.06185, 2317.10984, 2400.22014, 2480.9721, 2562.37003, 2647.41823, 2739.121, 2844.21397, 2959.48256, 3086.33195, 3248.45372, 3438.43132, 3639.5622, 3844.8941, 4047.47477, 4240.35196, 4416.57341, 4953.0841, 5231.10707, 5279.66896, 5127.79638, 4804.51595, 4338.8543, 3759.83803, 3291.66818, 3021.92332},
        {110.497277, 106.907259, 103.49924, 100.267403, 97.2059319, 94.3090101, 91.5708214, 88.9855492, 86.5473773, 84.2504892, 82.0890684, 80.0572987, 78.1493636, 76.3594466, 74.6817315, 73.1104017, 71.6396409, 70.2636328, 68.9765608, 67.7726085, 66.6459597, 65.5907978, 64.6013065, 63.6716694, 62.79607, 61.968692, 61.183719, 60.4353345, 59.7177222, 59.0250656, 58.3515484, 57.6913542, 57.0386665, 56.3876689, 55.5480515, 54.6101253, 53.6664186, 52.7171022, 51.7623469, 50.8023235, 49.8372027, 48.8671553, 47.892352, 46.9129637, 45.9291611, 44.9411149, 43.948996, 42.9529751, 41.9532229, 40.9499103, 39.943208, 38.9150413, 37.884222, 36.8513163, 35.8168906, 34.7815111, 33.7457441, 32.710156, 31.6753131, 30.6417815, 29.6101277, 28.5809178, 27.5547183, 26.5320954, 25.5136153, 24.4998445, 23.4913491, 22.4766494, 21.4445551, 20.4204763, 19.4057602, 18.4017541, 17.4098051, 16.4312603, 15.4674671, 14.5197725, 13.5895238, 12.6780681, 11.7867527, 10.9169247, 10.0699313, 9.24711971, 8.4498371, 7.67943066, 7.20246043, 6.88208877, 6.5782904, 6.28380584, 5.9913756, 5.69374022, 5.38364021, 5.05381609, 4.69700839, 4.30595761, 3.87340429, 3.39208894, 2.85475208, 2.25413424, 1.58297593, 0.834017679, 0, -3.29035333, -6.3954964, -9.31921127, -12.06528, -14.6374846, -17.0396073, -19.2754299, -21.3487347, -23.2633035, -25.0229186, -26.631362, -28.0924156, -29.4098617, -30.5874821, -31.6290591, -32.5383746, -32.7825927, -32.2888357, -32.2746646, -32.6288717, -33.2402497, -33.9975909, -34.7896878, -35.5053331, -36.033319, -36.2624381, -36.081483, -35.379246, -34.0445197, -31.4782147, -28.2358953, -24.3527187, -19.8638418, -15.411207, -10.7265786, -5.37991179, 0.775672022, 7.88705132, 16.1011046, 25.5647104, 36.1531481, 48.367556, 62.4079514, 78.4743519, 96.7667748, 117.971343, 141.680442, 167.972563, 196.926198, 228.619838, 262.613576, 300.1379, 341.952671, 388.519289, 439.689789, 495.589796, 556.344939, 622.488408, 693.645637, 768.997438, 847.407815, 930.308502, 1019.63024, 1115.86813, 1217.08462, 1322.4588, 1431.22506, 1542.57681, 1655.50468, 1769.11399, 1882.38252, 1994.56278, 2104.80426, 2212.25644, 2316.0688, 2402.58733, 2486.96589, 2571.55483, 2658.70451, 2750.76529, 2851.65808, 2961.11565, 3080.81525, 3234.32284, 3414.6514, 3605.69363, 3800.91854, 3993.7951, 4177.79234, 4346.37924, 4836.31658, 5100.54475, 5158.12369, 5028.11336, 4729.5737, 4281.56464, 3703.14614, 3213.90036, 2956.50718},
        {112.72494, 109.236657, 105.911707, 102.745036, 99.7315884, 96.8663101, 94.1441459, 91.560041, 89.1089406, 86.7857898, 84.5855338, 82.5031178, 80.5334869, 78.6715862, 76.912361, 75.2507564, 73.6817175, 72.2001896, 70.8011177, 69.479447, 68.2301227, 67.0480899, 65.9282939, 64.8656797, 63.8551925, 62.8917775, 61.9703799, 61.0859447, 60.2334172, 59.4077425, 58.6038658, 57.8167322, 57.0412869, 56.2724751, 55.3526537, 54.3562715, 53.3596112, 52.3625697, 51.3650439, 50.3669309, 49.3681275, 48.3685307, 47.3680374, 46.3665446, 45.3639492, 44.3601481, 43.3550383, 42.3485167, 41.3404802, 40.3308259, 39.3194505, 38.2728393, 37.2250254, 36.1766298, 35.128274, 34.0805791, 33.0341663, 31.9896569, 30.9476722, 29.9088334, 28.8737617, 27.8430784, 26.8174047, 25.7973618, 24.783571, 23.7766536, 22.7772307, 21.7782326, 20.7727743, 19.777339, 18.7930465, 17.8210168, 16.8623696, 15.9182247, 14.9897019, 14.0779211, 13.1840021, 12.3090646, 11.4542286, 10.6206138, 9.80934002, 9.0215271, 8.25829486, 7.5207631, 7.05855362, 6.74330834, 6.44368239, 6.15273121, 5.86351026, 5.56907498, 5.26248083, 4.93678325, 4.58503769, 4.2002996, 3.77562444, 3.30406765, 2.77868468, 2.19253098, 1.538662, 0.81013319, 0, -3.07995726, -6.00938029, -8.78857963, -11.4178658, -13.8975494, -16.2279409, -18.4093509, -20.4420899, -22.3264684, -24.062797, -25.6513863, -27.0925466, -28.3865887, -29.533823, -30.5345601, -31.3891105, -31.6961637, -31.3986559, -31.405326, -31.630423, -31.9881957, -32.3928931, -32.758764, -33.0000574, -33.0310221, -32.7659069, -32.1189609, -31.0044328, -29.3365717, -26.7219765, -23.4953509, -19.6632386, -15.2321835, -10.6787459, -5.77446122, -0.165527406, 6.2825167, 13.7041323, 22.2337804, 32.0059222, 42.7829309, 55.1845998, 69.4181898, 85.6909619, 104.210177, 125.524529, 149.414487, 176.001956, 205.408836, 237.757032, 272.859805, 311.524927, 354.286898, 401.496366, 453.128604, 509.327423, 570.236631, 636.310531, 707.311872, 782.664686, 861.708536, 945.246067, 1034.43202, 1129.5292, 1229.289, 1333.15661, 1440.94941, 1552.20907, 1665.11241, 1778.60828, 1890.82634, 2001.66067, 2110.34367, 2216.10775, 2318.18534, 2403.90127, 2487.37243, 2570.80813, 2656.41766, 2746.41036, 2844.31502, 2950.14178, 3065.53445, 3213.7191, 3387.79656, 3571.87503, 3759.50495, 3944.23679, 4119.62099, 4279.20801, 4710.69517, 4946.07358, 4999.23272, 4884.06206, 4614.45105, 4204.28916, 3667.46587, 3181.40427, 2893.75773},
        {114.960181, 111.627562, 108.434017, 105.37546, 102.447806, 99.6469709, 96.9688697, 94.4094174, 91.9645293, 89.6301204, 87.402106, 85.2764011, 83.248921, 81.3155807, 79.4722955, 77.7149805, 76.0395508, 74.4419217, 72.9180082, 71.4637255, 70.0749888, 68.7477133, 67.477814, 66.2612061, 65.0938049, 63.9715253, 62.8902827, 61.8459922, 60.8345688, 59.8519278, 58.8939843, 57.9566535, 57.0358505, 56.1274905, 55.11504, 54.0530036, 52.9972377, 51.9473065, 50.9027743, 49.8632056, 48.8281645, 47.7972155, 46.7699229, 45.7458511, 44.7245642, 43.7056268, 42.688603, 41.6730573, 40.6585539, 39.6446572, 38.6309316, 37.5660378, 36.5015473, 35.4381277, 34.3764469, 33.3171728, 32.2609731, 31.2085156, 30.1604682, 29.1174987, 28.0802749, 27.0494647, 26.0257357, 25.0097559, 24.0021931, 23.0037151, 22.0149897, 21.0332506, 20.0558138, 19.09043, 18.1379894, 17.1993827, 16.2755001, 15.3672322, 14.4754693, 13.601102, 12.7450207, 11.9081157, 11.0912776, 10.2953968, 9.5213637, 8.77006876, 8.04240242, 7.33925513, 6.89372018, 6.58570252, 6.29231616, 6.00691613, 5.72285742, 5.43349505, 5.13218402, 4.81227936, 4.46713605, 4.09010913, 3.67455359, 3.21382445, 2.70127671, 2.13026539, 1.49414549, 0.786272021, 0, -2.91766074, -5.71093841, -8.37723991, -10.9139721, -13.3185421, -15.5883565, -17.7208225, -19.7133468, -21.5633365, -23.2681983, -24.8253393, -26.2321664, -27.4860864, -28.5845062, -29.5248329, -30.3044732, -30.6537635, -30.5319577, -30.5405512, -30.6197212, -30.7096455, -30.7505014, -30.6824664, -30.4457181, -29.9804338, -29.2267913, -28.1249678, -26.6151409, -24.6374881, -22.017462, -18.8520307, -15.1157894, -10.783333, -6.15841662, -1.05105526, 4.81651223, 11.5684389, 19.3288778, 28.221982, 38.3719045, 49.5248425, 62.297935, 76.889283, 93.4969873, 112.319149, 133.547048, 157.387311, 184.039745, 213.704157, 246.580351, 282.850439, 322.753551, 366.583675, 414.561569, 466.758475, 523.312765, 584.362811, 650.155717, 720.69641, 795.871199, 875.751773, 960.246598, 1049.53397, 1143.55447, 1241.84593, 1344.13107, 1450.79797, 1561.74186, 1674.41815, 1787.66233, 1899.1642, 2009.08123, 2116.6455, 2221.08911, 2321.64416, 2405.6252, 2487.16123, 2568.46372, 2651.74416, 2739.21401, 2834.69103, 2937.70956, 3049.79263, 3193.54357, 3362.14118, 3539.98079, 3720.68038, 3897.8579, 4065.13131, 4216.11859, 4600.45492, 4809.05082, 4854.20585, 4748.21956, 4503.39151, 4132.02125, 3646.40835, 3164.5494, 2841.34803},
        {117.215076, 114.067641, 111.031888, 108.104819, 105.283435, 102.564737, 99.9457289, 97.4234111, 94.9947858, 92.6568547, 90.4066197, 88.2410826, 86.1572451, 84.1521091, 82.2226763, 80.3659487, 78.5789279, 76.8586159, 75.2020144, 73.6061251, 72.06795, 70.5844908, 69.1527494, 67.7697275, 66.4324269, 65.1378495, 63.882997, 62.6648713, 61.4804742, 60.3268074, 59.2008729, 58.0996723, 57.0202075, 55.9594803, 54.8468068, 53.7144552, 52.5955076, 51.4891623, 50.3946175, 49.3110716, 48.237723, 47.1737701, 46.118411, 45.0708443, 44.0302681, 42.995881, 41.9668811, 40.9424669, 39.9218367, 38.9041888, 37.8887215, 36.8050135, 35.723392, 34.6445646, 33.5692389, 32.4981225, 31.4319228, 30.3713476, 29.3171042, 28.2699003, 27.2304435, 26.1994412, 25.1776012, 24.1656308, 23.1642377, 22.1741295, 21.1960137, 20.2310228, 19.2802791, 18.3436137, 17.4217068, 16.5152382, 15.624888, 14.7513362, 13.8952628, 13.0573478, 12.2382713, 11.4387132, 10.6593536, 9.90087252, 9.1639499, 8.44926577, 7.75750017, 7.08933309, 6.66229487, 6.36407919, 6.0797734, 5.80302029, 5.52746268, 5.24674336, 4.95450514, 4.64439082, 4.31004321, 3.9451051, 3.5432193, 3.09802861, 2.60317584, 2.0523038, 1.43905527, 0.75707307, 0, -2.76049588, -5.41979872, -7.9727375, -10.4141412, -12.7388388, -14.9416592, -17.0174316, -18.9609847, -20.7671477, -22.4307495, -23.946619, -25.3095853, -26.5144774, -27.5561242, -28.4293546, -29.1289978, -29.493686, -29.4959346, -29.4838741, -29.418863, -29.2622594, -28.9754217, -28.5197081, -27.8564771, -26.9470867, -25.7528954, -24.2352614, -22.355543, -20.0750985, -17.3980597, -14.2273278, -10.5114289, -6.19888948, -1.46207859, 3.86239896, 9.99762981, 17.0622406, 25.174858, 34.4541085, 45.0186188, 56.6793231, 69.956186, 85.0280349, 102.073697, 121.272, 142.390178, 166.121549, 192.74784, 222.550775, 255.81208, 293.073086, 334.038617, 378.815839, 427.513243, 480.24202, 537.112146, 598.2336, 663.555466, 733.385183, 808.205713, 888.948077, 974.514519, 1064.07783, 1157.21985, 1254.21559, 1355.02188, 1460.4642, 1570.72459, 1682.80003, 1795.48916, 1906.51624, 2015.91955, 2122.86962, 2226.53696, 2326.09209, 2408.37815, 2487.97489, 2567.1347, 2648.10993, 2733.15297, 2826.53491, 2927.14358, 3036.36617, 3175.90285, 3339.10026, 3510.84134, 3684.89591, 3855.03383, 4015.02493, 4158.63905, 4522.03768, 4713.74311, 4749.18525, 4643.79403, 4412.99937, 4072.2312, 3636.91943, 3152.98842, 2802.33571},
        {119.53678, 116.537329, 113.625664, 110.799745, 108.057531, 105.396983, 102.816059, 100.312719, 97.8849222, 95.5306281, 93.2477962, 91.034386, 88.8883567, 86.807668, 84.7902792, 82.8341498, 80.9372391, 79.0975068, 77.3129121, 75.5814145, 73.9009735, 72.2695485, 70.685099, 69.1455843, 67.648964, 66.1931975, 64.7762441, 63.3960634, 62.0506148, 60.7378577, 59.4557516, 58.2022559, 56.97533, 55.7729334, 54.564929, 53.3639155, 52.1828293, 51.0205415, 49.8759234, 48.7478462, 47.6351812, 46.5367996, 45.4515725, 44.3783713, 43.3160671, 42.2635313, 41.2196349, 40.1832492, 39.1532455, 38.1284949, 37.1078688, 36.0052016, 34.9062446, 33.8117125, 32.7223201, 31.6387819, 30.5618126, 29.4921269, 28.4304394, 27.3774648, 26.3339177, 25.3005128, 24.2779647, 23.2669881, 22.2682977, 21.2826081, 20.310634, 19.3548454, 18.4176705, 17.4962033, 16.5910447, 15.7027954, 14.8320564, 13.9794285, 13.1455126, 12.3309095, 11.5362201, 10.7620453, 10.0089858, 9.27764257, 8.56861643, 7.88250824, 7.21991886, 6.58144914, 6.17425069, 5.89130447, 5.62310886, 5.36356174, 5.106561, 4.84600453, 4.57579021, 4.28981592, 3.98197955, 3.64617899, 3.27631212, 2.86627683, 2.409971, 1.90129251, 1.33413926, 0.70240913, 0, -2.40431622, -4.75752363, -7.0506962, -9.27490792, -11.4212328, -13.4807448, -15.4445179, -17.3036261, -19.0491434, -20.6721438, -22.1637012, -23.5148897, -24.7167833, -25.7604559, -26.6369815, -27.337434, -27.696871, -27.6840783, -27.6430246, -27.5313516, -27.3067014, -26.9267157, -26.3490364, -25.5313053, -24.4311644, -23.0062555, -21.2142204, -19.012701, -16.3593392, -13.1867453, -9.48677086, -5.22456731, -0.365285918, 4.80205784, 10.4742447, 16.9344192, 24.3145893, 32.7467628, 42.3629476, 53.2951518, 65.513631, 79.3613743, 95.0020367, 112.599273, 132.316739, 153.900023, 178.035362, 204.99093, 235.034897, 268.435436, 305.652734, 346.52826, 391.339336, 440.241363, 493.140821, 550.055952, 611.004996, 675.477631, 744.140789, 818.23802, 899.613071, 986.184606, 1076.39155, 1169.46768, 1265.99368, 1365.93198, 1469.95282, 1578.20213, 1688.23054, 1799.05673, 1909.55358, 2018.90798, 2126.18907, 2230.466, 2330.8079, 2414.34031, 2495.06187, 2575.02762, 2656.29259, 2740.91181, 2832.47579, 2930.48046, 3036.32279, 3170.02707, 3325.61698, 3489.35479, 3655.65223, 3818.92103, 3973.57294, 4114.01968, 4556.24183, 4774.79201, 4801.58779, 4668.54674, 4407.58646, 4050.62451, 3629.57849, 3100.08847, 2784.8712},
        {121.8803, 119.036053, 116.254884, 113.53574, 110.877567, 108.279313, 105.739923, 103.258345, 100.833524, 98.4644087, 96.1499443, 93.8890779, 91.6807561, 89.5239256, 87.4175329, 85.3605248, 83.351848, 81.3904489, 79.4752744, 77.6052711, 75.7793855, 73.9965644, 72.2557544, 70.5559021, 68.8959542, 67.2748574, 65.6915582, 64.1450034, 62.6341396, 61.1579134, 59.7152715, 58.3051605, 56.9265271, 55.578318, 54.2722526, 53.0006226, 51.7555659, 50.5356146, 49.339301, 48.1651571, 47.0117152, 45.8775073, 44.7610655, 43.6609222, 42.5756092, 41.5036589, 40.4436034, 39.3939748, 38.3533052, 37.3201268, 36.2929718, 35.1687488, 34.0498165, 32.9369099, 31.8307642, 30.7321146, 29.6416961, 28.5602439, 27.4884932, 26.427179, 25.3770366, 24.3388011, 23.3132075, 22.3009911, 21.302887, 20.3196302, 19.3519561, 18.4038368, 17.479167, 16.5720054, 15.6828772, 14.8123078, 13.9608224, 13.1289463, 12.3172048, 11.5261232, 10.7562267, 10.0080407, 9.28209038, 8.57890109, 7.89899811, 7.24290672, 6.61115219, 6.00425982, 5.61878304, 5.35310988, 5.10327224, 4.86343392, 4.62775871, 4.39041041, 4.14555281, 3.8873497, 3.60996489, 3.30756215, 2.97430529, 2.6043581, 2.19188437, 1.73104791, 1.21601249, 0.640941927, 0, -2.06837363, -4.12866099, -6.16881259, -8.17677896, -10.1405106, -12.0479581, -13.8870719, -15.6458026, -17.3121007, -18.8739166, -20.319201, -21.6359043, -22.8119771, -23.8353699, -24.6940332, -25.3759175, -25.6829889, -25.5766286, -25.4654664, -25.2975991, -25.0211233, -24.5841358, -23.9347331, -23.0210121, -21.7910694, -20.1930017, -18.1749057, -15.6848781, -12.6710155, -8.94578135, -4.64263791, 0.249628066, 5.74222984, 11.3893067, 17.430609, 24.2277733, 31.929135, 40.6830296, 50.6377926, 61.9417594, 74.7552998, 89.2110522, 105.454998, 123.633117, 143.891391, 165.995943, 190.567577, 217.847236, 248.075868, 281.494416, 318.413145, 358.918958, 403.503726, 452.395197, 505.281867, 562.094347, 622.763247, 686.370706, 753.888639, 827.214565, 908.938636, 996.201484, 1086.94087, 1180.11654, 1276.41988, 1375.74743, 1478.45086, 1584.54469, 1692.37467, 1801.23051, 1911.38069, 2020.98557, 2128.99602, 2234.36295, 2336.03725, 2421.59094, 2504.19851, 2585.65556, 2667.75772, 2752.30059, 2841.81512, 2936.87136, 3038.94979, 3166.43361, 3314.11262, 3469.73381, 3628.43911, 3785.37047, 3935.66982, 4074.47909, 4614.80826, 4869.43419, 4889.06184, 4724.39617, 4426.14215, 4045.00475, 3631.68892, 3054.53477, 2783.72667},
        {124.234194, 121.556326, 118.915687, 116.312251, 113.745993, 111.216889, 108.724913, 106.270039, 103.852244, 101.4715, 99.1277841, 96.82107, 94.5513329, 92.3185475, 90.1226886, 87.9637311, 85.8416498, 83.7564195, 81.708015, 79.6964112, 77.7215828, 75.7835048, 73.8821518, 72.0174988, 70.1895205, 68.3981918, 66.6434874, 64.9253823, 63.2438513, 61.598869, 59.9904105, 58.4184504, 56.8829637, 55.3839251, 53.9767683, 52.6325722, 51.3217492, 50.0424743, 48.7929225, 47.5712689, 46.3756887, 45.2043568, 44.0554484, 42.9271385, 41.8176022, 40.7250146, 39.6475507, 38.5833856, 37.5306944, 36.4876521, 35.4524339, 34.3026408, 33.1596357, 32.0242074, 30.8971449, 29.779237, 28.6712726, 27.5740406, 26.4883299, 25.4149293, 24.3546277, 23.3082141, 22.2764772, 21.2602061, 20.2601895, 19.2772163, 18.3120754, 17.3713539, 16.4614994, 15.5713427, 14.7012968, 13.8517747, 13.0231895, 12.2159542, 11.4304817, 10.6671851, 9.92647745, 9.20877169, 8.51448085, 7.84401796, 7.19779603, 6.57622807, 5.9797271, 5.40870612, 5.04700108, 4.79945548, 4.56870821, 4.34921986, 4.13545101, 3.92186225, 3.70291417, 3.47306736, 3.22678241, 2.95851989, 2.66274041, 2.33390454, 1.96647288, 1.554906, 1.09366451, 0.577208977, 0, -1.81774798, -3.65265025, -5.49076821, -7.31816328, -9.12089687, -10.8850304, -12.5966253, -14.2417429, -15.8064448, -17.2767922, -18.6388466, -19.8786694, -20.9823221, -21.935866, -22.7253626, -23.3368732, -23.5298901, -23.2581077, -23.0200653, -22.7532738, -22.3952442, -21.8834872, -21.1555137, -20.1488346, -18.8009608, -17.0494031, -14.8316723, -12.0852795, -8.7477353, -4.51442896, 0.346228958, 5.82409105, 11.9090099, 18.0150018, 24.4198374, 31.5548441, 39.5826252, 48.6657843, 58.966925, 70.6486508, 84.0566525, 99.1147241, 115.949648, 134.688205, 155.457179, 178.027145, 202.970143, 230.502007, 260.838569, 294.195663, 330.744286, 370.79991, 415.025024, 463.711661, 516.395502, 572.951816, 633.255869, 696.144042, 762.7666, 835.407144, 917.069323, 1004.48663, 1095.41312, 1188.66593, 1284.86349, 1383.7974, 1485.4431, 1589.63986, 1695.55284, 1802.7285, 1912.7595, 2022.83133, 2131.78219, 2238.45027, 2341.67374, 2429.70255, 2514.6102, 2597.88193, 2681.003, 2765.45866, 2852.34864, 2943.80073, 3041.46538, 3162.4537, 3302.29633, 3450.08198, 3601.66337, 3752.89325, 3899.62435, 4037.70939, 4671.03674, 4961.35253, 4976.56266, 4784.57301, 4453.28949, 4050.61798, 3644.46439, 3030.30438, 2798.54574},
        {126.503785, 124.022677, 121.549989, 119.086836, 116.63433, 114.193587, 111.765718, 109.351838, 106.953061, 104.570499, 102.205268, 99.85848, 97.5312493, 95.2246893, 92.9399138, 90.6780364, 88.4401708, 86.2274306, 84.0409295, 81.8817812, 79.7510992, 77.6499973, 75.5795891, 73.5409883, 71.5353085, 69.5636634, 67.6271666, 65.7269318, 63.8640727, 62.0397029, 60.254936, 58.5108858, 56.8086659, 55.1493899, 53.6378325, 52.2210966, 50.8450388, 49.507409, 48.2059568, 46.9384317, 45.7025836, 44.496162, 43.3169165, 42.1625969, 41.0309529, 39.919734, 38.8266899, 37.7495703, 36.6861248, 35.6341032, 34.591255, 33.4072235, 32.2310754, 31.0637711, 29.9062708, 28.7595349, 27.6245237, 26.5021974, 25.3935165, 24.2994413, 23.2209319, 22.1589489, 21.1144524, 20.0884028, 19.0817604, 18.0954855, 17.1305384, 16.2007712, 15.3197262, 14.4617757, 13.6270443, 12.8156566, 12.0277372, 11.2634107, 10.5228016, 9.80603469, 9.11323446, 8.44452552, 7.80003248, 7.17987995, 6.58419251, 6.01309477, 5.46671133, 4.9451668, 4.61095508, 4.37960089, 4.16467742, 3.96104006, 3.76354422, 3.5670453, 3.36639872, 3.15645987, 2.93208416, 2.68812699, 2.41944377, 2.1208899, 1.78732079, 1.41359184, 0.994558452, 0.525076039, 0, -1.8967898, -3.77707958, -5.62863254, -7.43921186, -9.19658075, -10.8885024, -12.50274, -14.0270567, -15.4492158, -16.7569803, -17.9381136, -18.9803788, -19.8715392, -20.5993578, -21.1515979, -21.5160226, -21.4658168, -20.9580894, -20.4659598, -19.9312102, -19.2956225, -18.500979, -17.4890616, -16.2016525, -14.5805337, -12.5674874, -10.1042955, -7.13274016, -3.59460349, 0.752428998, 5.71497661, 11.2960283, 17.498573, 23.8073511, 30.4844758, 37.9302601, 46.3031674, 55.7616614, 66.4642055, 78.5692634, 92.5072184, 108.081856, 125.398438, 144.562226, 165.678483, 188.287625, 213.200969, 240.66499, 270.926158, 304.230948, 340.884472, 381.002892, 425.074997, 473.339434, 525.524141, 581.586352, 641.483302, 704.401139, 771.243431, 843.754938, 924.297847, 1010.31965, 1100.1888, 1192.97499, 1288.93615, 1387.78479, 1489.32961, 1593.30804, 1699.10488, 1806.30442, 1916.74617, 2027.4122, 2137.1061, 2244.63149, 2348.79198, 2439.20179, 2526.15126, 2610.74132, 2694.07294, 2777.24705, 2859.18494, 2944.62034, 3035.58835, 3149.66452, 3282.7325, 3424.2008, 3570.27684, 3717.16806, 3861.08189, 3998.22577, 4659.25912, 4960.54376, 4976.45474, 4781.36713, 4449.65598, 4055.69634, 3673.86328, 3073.16342, 2828.78909},
        {128.77986, 126.511111, 124.220518, 121.910385, 119.583015, 117.240711, 114.885775, 112.520511, 110.147222, 107.768209, 105.385778, 103.002229, 100.619867, 98.2409934, 95.8679123, 93.5029261, 91.1483379, 88.8064505, 86.4795669, 84.16999, 81.8800228, 79.6119681, 77.3681289, 75.1508082, 72.9623088, 70.8049337, 68.6809858, 66.5927681, 64.5425835, 62.5327348, 60.5655251, 58.6432573, 56.7682342, 54.9427589, 53.3231651, 51.8315005, 50.3879615, 48.9898522, 47.6344772, 46.3191406, 45.0411469, 43.7978004, 42.5864053, 41.4042661, 40.2486871, 39.1169725, 38.0064268, 36.9143542, 35.8380591, 34.7748459, 33.7220188, 32.4976365, 31.2821348, 30.0767036, 28.8825329, 27.7008129, 26.5327334, 25.3794846, 24.2422564, 23.1222389, 22.0206221, 20.938596, 19.8773507, 18.8380761, 17.8219624, 16.8301994, 15.8639774, 14.9468306, 14.1017571, 13.2838632, 12.4928903, 11.7285802, 10.9906742, 10.2789139, 9.59304088, 8.93279658, 8.29792257, 7.68816036, 7.10325148, 6.54293747, 6.00695984, 5.49506013, 5.00697986, 4.54246056, 4.23778559, 4.02155359, 3.82072207, 3.6306015, 3.44650236, 3.26373514, 3.07761032, 2.88343838, 2.67652981, 2.45219507, 2.20574467, 1.93248906, 1.62773875, 1.2868042, 0.90499591, 0.477624348, 0, -2.15041158, -4.21837964, -6.19540909, -8.07300485, -9.84267184, -11.495915, -13.0242392, -14.4191494, -15.6721504, -16.7747473, -17.718445, -18.4947483, -19.0951622, -19.5111915, -19.7343413, -19.7561164, -19.3839191, -18.5829541, -17.7611487, -16.8705573, -15.8632343, -14.691234, -13.3066109, -11.6614192, -9.7077133, -7.39754762, -4.68297648, -1.51605424, 2.15116473, 6.42065301, 11.2665194, 16.7205016, 22.8143369, 29.2027539, 36.1059947, 43.8448372, 52.5641217, 62.4086887, 73.5233784, 86.0530312, 100.450504, 116.458877, 134.162725, 153.646625, 174.995152, 197.422519, 222.101257, 249.333534, 279.421518, 312.667375, 349.638048, 390.047316, 434.038067, 481.737867, 533.242999, 588.66379, 648.110568, 711.426965, 779.050619, 851.710108, 930.586791, 1014.41032, 1102.56014, 1194.55783, 1290.16583, 1389.03602, 1490.92732, 1595.51938, 1702.09928, 1810.17617, 1921.28494, 2032.59731, 2142.92119, 2251.06449, 2355.83512, 2448.4894, 2537.27474, 2622.88692, 2706.02177, 2787.37508, 2863.41997, 2941.89007, 3025.29091, 3132.46405, 3259.44869, 3395.47732, 3536.92944, 3680.18458, 3821.62225, 3957.62197, 4600.6328, 4896.62427, 4918.28155, 4738.28982, 4429.33423, 4064.09996, 3715.27217, 3158.0479, 2874.32496},
        {131.136568, 129.084043, 126.979161, 124.825416, 122.626302, 120.385313, 118.105944, 115.791688, 113.446039, 111.072492, 108.67454, 106.255678, 103.819399, 101.369198, 98.9085691, 96.4410057, 93.9700023, 91.4990527, 89.0316512, 86.5712917, 84.1214684, 81.6856753, 79.2674064, 76.8701559, 74.4974178, 72.1526861, 69.8394549, 67.5612184, 65.3214704, 63.1237052, 60.9714167, 58.8680991, 56.8172464, 54.8223527, 53.0907643, 51.5197615, 50.0041742, 48.540881, 47.1267606, 45.7586917, 44.4335529, 43.148223, 41.8995805, 40.6845041, 39.4998726, 38.3425644, 37.2094584, 36.0974332, 35.0033674, 33.9241397, 32.8566287, 31.5876925, 30.3287832, 29.0813323, 27.8467713, 26.6265318, 25.4220452, 24.234743, 23.0660568, 21.917418, 20.7902581, 19.6860087, 18.6061012, 17.5519671, 16.525038, 15.5267453, 14.5585205, 13.6542829, 12.8471717, 12.0716148, 11.3269378, 10.6124661, 9.92752505, 9.27144025, 8.64353708, 8.04314102, 7.46957752, 6.92217202, 6.40024998, 5.90313687, 5.43015812, 4.98063919, 4.55390555, 4.14928264, 3.87481024, 3.67316999, 3.48573685, 3.30830828, 3.13668176, 2.96665475, 2.79402471, 2.6145891, 2.4241454, 2.21849106, 1.99342355, 1.74474034, 1.46823889, 1.15971667, 0.814971136, 0.429799758, 0, -2.45936011, -4.75857448, -6.89372931, -8.86091079, -10.6562051, -12.2756985, -13.7154771, -14.9716271, -16.0402348, -16.9173862, -17.5991677, -18.0816654, -18.3609654, -18.4331541, -18.2943175, -17.9405419, -17.2074093, -16.0680765, -14.8816752, -13.6098978, -12.2144367, -10.6569844, -8.89923326, -6.90287569, -4.62960412, -2.04111099, 0.90091129, 4.2347703, 7.99877361, 12.1529255, 16.8425481, 22.12944, 28.0753995, 34.5202494, 41.6367761, 49.6569599, 58.7091922, 68.9218642, 80.4233675, 93.3420933, 108.11752, 124.472274, 142.47388, 162.189867, 183.687759, 205.866042, 230.253545, 257.210054, 287.095358, 320.269243, 357.579156, 398.301195, 442.189868, 489.248727, 539.98978, 594.696748, 653.653355, 717.445355, 785.985792, 858.858216, 935.928354, 1017.34008, 1103.55814, 1194.60068, 1289.72917, 1388.55239, 1490.83635, 1596.23058, 1703.80826, 1812.9686, 1924.77863, 2036.71315, 2147.59402, 2256.24315, 2361.48241, 2456.15684, 2546.55938, 2633.00612, 2715.81313, 2795.29652, 2865.61257, 2937.34372, 3013.44595, 3114.10013, 3235.44409, 3366.43866, 3503.58051, 3643.36635, 3782.29285, 3916.85671, 4508.52589, 4788.47916, 4821.07421, 4670.66873, 4401.62042, 4078.28696, 3765.02606, 3266.24815, 2934.98504},
        {133.721218, 131.882962, 129.961276, 127.960876, 125.88648, 123.742803, 121.534561, 119.266471, 116.94325, 114.569613, 112.150277, 109.689959, 107.193375, 104.665242, 102.110275, 99.5331905, 96.9387059, 94.3315371, 91.7164006, 89.0980127, 86.4810897, 83.8703482, 81.2705043, 78.6862747, 76.1223755, 73.5835233, 71.0744344, 68.5998251, 66.1644119, 63.7729112, 61.4300393, 59.1405127, 56.9090476, 54.7403606, 52.8947828, 51.2438468, 49.6560126, 48.1277014, 46.6553344, 45.2353328, 43.8641178, 42.5381106, 41.2537323, 40.0074041, 38.7955473, 37.6145829, 36.4609323, 35.3310166, 34.221257, 33.1280746, 32.0478907, 30.7351317, 29.4334595, 28.1445412, 26.8700441, 25.6116353, 24.370982, 23.1497512, 21.9496103, 20.7722262, 19.6192663, 18.4923976, 17.3932873, 16.3236026, 15.2850107, 14.2791786, 13.3077736, 12.4146714, 11.6427343, 10.9065774, 10.2051317, 9.53732819, 8.90209781, 8.29837151, 7.72508022, 7.18115492, 6.66552654, 6.17712605, 5.71488439, 5.27773251, 4.86460137, 4.47442193, 4.10612512, 3.75864191, 3.51112719, 3.32071252, 3.14356528, 2.97601298, 2.81438316, 2.65500335, 2.49420108, 2.32830388, 2.15363928, 1.96653482, 1.76331803, 1.54031643, 1.29385757, 1.02026896, 0.715878139, 0.377012643, 0, -2.686425, -5.14541502, -7.37725671, -9.38223673, -11.1606417, -12.7127584, -14.0388734, -15.1392733, -16.0142448, -16.6640746, -17.0890493, -17.2894557, -17.2655803, -17.0177097, -16.5461308, -15.85113, -14.744669, -13.2001306, -11.6437754, -10.0355344, -8.33533899, -6.50312043, -4.49880994, -2.28233881, 0.18636173, 2.94736041, 6.04072597, 9.50652716, 13.3848327, 17.6110655, 22.3683102, 27.7280295, 33.7616859, 40.3865805, 47.7512564, 56.0453667, 65.3866222, 75.892734, 87.681413, 100.87037, 115.86436, 132.406689, 150.558906, 170.382564, 191.93921, 213.960107, 238.169666, 264.962011, 294.731263, 327.871546, 365.422822, 406.344015, 450.077646, 496.529387, 546.550498, 600.567689, 659.00767, 723.029294, 792.160732, 865.13145, 940.881476, 1020.51306, 1105.18015, 1195.22975, 1289.64166, 1388.02325, 1490.18066, 1595.77281, 1703.72974, 1813.3938, 1925.68519, 2038.12563, 2149.51128, 2258.63827, 2364.30278, 2460.15098, 2551.41649, 2638.18293, 2720.53394, 2798.55317, 2865.07398, 2932.26379, 3003.31351, 3098.75015, 3214.90238, 3340.87553, 3473.35943, 3609.04396, 3744.61896, 3876.77428, 4362.30746, 4611.09989, 4666.48838, 4571.80973, 4370.40076, 4105.59826, 3820.73906, 3380.67922, 3005.28501},
        {136.633402, 135.000954, 133.25366, 131.397489, 129.43841, 127.382393, 125.235408, 123.003424, 120.69241, 118.308338, 115.857175, 113.344893, 110.777459, 108.160845, 105.501019, 102.803952, 100.075612, 97.3219694, 94.548994, 91.7626554, 88.968923, 86.1737665, 83.3831554, 80.6030595, 77.8394484, 75.0982916, 72.3855587, 69.7072194, 67.0692433, 64.4776, 61.9382592, 59.4571904, 57.0403632, 54.6937473, 52.7330362, 51.0028231, 49.3439781, 47.7524198, 46.2240667, 44.7548372, 43.3406499, 41.9774231, 40.6610754, 39.3875252, 38.1526911, 36.9524914, 35.7828447, 34.6396695, 33.5188841, 32.4164072, 31.3281571, 29.9752976, 28.6344581, 27.3075131, 25.9963373, 24.7028054, 23.4287919, 22.1761715, 20.9468188, 19.7426084, 18.5654149, 17.417113, 16.2995772, 15.2146822, 14.1643027, 13.1503132, 12.1745883, 11.2894899, 10.5461674, 9.84236967, 9.17669834, 8.54775518, 7.95414194, 7.39446036, 6.86731219, 6.37129916, 5.90502303, 5.46708553, 5.05608842, 4.67063344, 4.30932233, 3.97075684, 3.65353871, 3.35626968, 3.13001021, 2.94602928, 2.77496538, 2.61371787, 2.45918611, 2.30826946, 2.15786727, 2.00487891, 1.84620373, 1.67874109, 1.49939036, 1.30505089, 1.09262204, 0.859003161, 0.601093625, 0.315792785, 0, -2.72902945, -5.19073252, -7.38825799, -9.32475461, -11.0033711, -12.4272563, -13.599559, -14.5234278, -15.2020115, -15.638459, -15.8359189, -15.7975401, -15.5264712, -15.0258611, -14.2988585, -13.3486122, -11.8818244, -9.85929452, -7.95470484, -6.1076799, -4.25784422, -2.34482232, -0.308238743, 1.912282, 4.37711539, 7.1466369, 10.281222, 13.8412462, 17.8870849, 22.5272113, 27.7562675, 33.6201994, 40.1649534, 47.193592, 54.8735031, 63.4368437, 73.0024249, 83.6890581, 95.6155545, 108.900725, 123.904857, 140.431793, 158.553099, 178.340342, 199.865088, 221.939402, 246.209226, 273.061001, 302.88117, 336.056174, 373.649127, 414.542753, 458.037549, 503.998305, 553.441905, 606.867965, 664.776102, 728.537232, 797.581648, 870.389139, 945.749884, 1024.83926, 1108.76331, 1197.87578, 1291.25577, 1388.56746, 1489.67205, 1594.28479, 1701.39861, 1810.41504, 1922.76674, 2035.51161, 2147.34806, 2256.97452, 2363.08942, 2458.87097, 2549.91787, 2636.30859, 2718.12163, 2795.43547, 2861.33652, 2927.5567, 2997.17109, 3089.30748, 3200.68673, 3321.34113, 3448.32597, 3578.69659, 3709.50829, 3837.8164, 4152.31631, 4353.15272, 4446.83144, 4439.85829, 4338.73907, 4149.9796, 3880.0857, 3488.05054, 3082.30861},
        {139.66004, 138.213955, 136.624909, 134.899963, 133.046175, 131.070606, 128.980315, 126.78236, 124.483803, 122.091702, 119.613117, 117.055106, 114.424731, 111.72905, 108.975123, 106.170009, 103.320768, 100.43446, 97.518143, 94.5788775, 91.6237228, 88.6597385, 85.693984, 82.7335189, 79.7854026, 76.8566947, 73.9544546, 71.0857419, 68.257616, 65.4771365, 62.7513629, 60.0873546, 57.4921712, 54.9728722, 52.8894459, 51.0650639, 49.3186561, 47.6458288, 46.0421882, 44.5033408, 43.0248926, 41.6024501, 40.2316196, 38.9080072, 37.6272193, 36.3848622, 35.1765422, 33.9978655, 32.8444385, 31.7118673, 30.5957585, 29.196777, 27.8118641, 26.4430197, 25.092244, 23.7615369, 22.4528984, 21.1683286, 19.9098276, 18.6793953, 17.4790318, 16.3107371, 15.1765112, 14.0783543, 13.0182662, 11.9982471, 11.020297, 10.1433306, 9.42489628, 8.74961162, 8.1157871, 7.52173321, 6.96576043, 6.44617923, 5.96130011, 5.50943352, 5.08888997, 4.69797992, 4.33501385, 3.99830225, 3.6861556, 3.39688437, 3.12879905, 2.88021011, 2.67656813, 2.50204269, 2.34047716, 2.18930127, 2.04594475, 1.90783734, 1.77240877, 1.63708876, 1.49930706, 1.35649339, 1.20607747, 1.04548906, 0.872157868, 0.683513636, 0.476986092, 0.25000497, 0, -2.7115359, -5.12625013, -7.24957975, -9.0869618, -10.6438333, -11.9256314, -12.9377931, -13.6857555, -14.1749556, -14.4108304, -14.398817, -14.1443526, -13.652874, -12.9298184, -11.9806229, -10.8107244, -9.01442867, -6.53843962, -4.31813448, -2.27085069, -0.313925728, 1.63530295, 3.65949789, 5.84132162, 8.26343668, 11.0085056, 14.1591909, 17.7981552, 22.008061, 27.094278, 32.8351022, 39.2463841, 46.343974, 53.7939394, 61.7870221, 70.6072394, 80.3753766, 91.212219, 103.238552, 116.57516, 131.520672, 147.963904, 165.990846, 185.687489, 207.139823, 229.336006, 253.734318, 280.695206, 310.579119, 343.746505, 381.198638, 421.871907, 465.022503, 510.50843, 559.416392, 612.281429, 669.638585, 732.921842, 801.562996, 874.013181, 949.165003, 1027.99093, 1111.31883, 1199.41238, 1291.57928, 1387.56601, 1487.32133, 1590.64418, 1696.59194, 1804.64145, 1916.77856, 2029.58513, 2141.66949, 2251.64, 2358.105, 2453.67949, 2544.4635, 2630.56374, 2712.08688, 2789.13962, 2855.38137, 2921.66428, 2990.85816, 3080.25186, 3187.00765, 3302.27989, 3423.53123, 3548.22435, 3673.82193, 3797.78662, 3940.77142, 4092.90931, 4224.20718, 4304.67192, 4304.31045, 4193.12965, 3941.13643, 3599.93455, 3179.96684},
        {142.743311, 141.455381, 140.001623, 138.389926, 136.628178, 134.724267, 132.686082, 130.52151, 128.238439, 125.844758, 123.348355, 120.757117, 118.078934, 115.321693, 112.493281, 109.601589, 106.654502, 103.659911, 100.625701, 97.5597632, 94.4699838, 91.3642515, 88.2504544, 85.1364806, 82.0302181, 78.9395552, 75.8723799, 72.8365804, 69.8400447, 66.8906611, 63.9963175, 61.1649022, 58.4043033, 55.7224088, 53.5058664, 51.5652176, 49.706452, 47.925166, 46.2169563, 44.5774196, 43.0021526, 41.4867518, 40.026814, 38.6179357, 37.2557137, 35.9357445, 34.6536248, 33.4049513, 32.1853206, 30.9903294, 29.8155742, 28.3610292, 26.924322, 25.5074579, 24.1124421, 22.7412797, 21.3959759, 20.0785358, 18.7909647, 17.5352676, 16.3134499, 15.1275165, 13.9794728, 12.8713238, 11.8050747, 10.7827308, 9.80629708, 8.93855805, 8.24083877, 7.58979184, 6.98348235, 6.41997541, 5.89733611, 5.41362956, 4.96692084, 4.55527507, 4.17675733, 3.82943273, 3.51136637, 3.22062334, 2.95526875, 2.71336769, 2.49298526, 2.29218656, 2.11486072, 1.95603831, 1.81067994, 1.67666168, 1.55185965, 1.43414993, 1.32140863, 1.21151185, 1.10233567, 0.991756203, 0.877649543, 0.757891788, 0.630359034, 0.492927381, 0.343472926, 0.179871766, 0, -2.64561941, -4.97519236, -6.99577981, -8.71444274, -10.1382421, -11.274239, -12.1294943, -12.711069, -13.026024, -13.0814205, -12.8843192, -12.4417813, -11.7608677, -10.8486394, -9.71215736, -8.35848255, -6.28507722, -3.42620255, -0.941583304, 1.27091927, 3.31344392, 5.28812938, 7.29711442, 9.44253777, 11.8265382, 14.5512544, 17.7188252, 21.4313893, 25.7910855, 31.2763803, 37.4750977, 44.3764783, 51.9697624, 59.8045674, 68.0899452, 77.1521805, 87.1124008, 98.0917333, 110.211305, 123.592244, 138.456339, 154.793419, 172.704917, 192.292266, 213.656899, 236.030881, 260.602353, 287.690091, 317.612867, 350.689457, 387.776372, 427.997415, 470.672278, 515.690068, 564.102446, 616.439108, 673.229752, 735.83384, 803.762721, 875.622547, 950.588644, 1029.24802, 1112.00598, 1198.97603, 1289.77733, 1384.25594, 1482.48051, 1584.35483, 1688.96653, 1795.86492, 1907.48622, 2020.04865, 2132.10219, 2242.19682, 2348.8825, 2444.46766, 2535.30422, 2621.50253, 2703.17296, 2780.42589, 2847.47029, 2914.25215, 2983.411, 3070.24085, 3172.37216, 3282.16267, 3397.50647, 3516.29765, 3636.4303, 3755.79849, 3754.19944, 3864.14746, 4026.06174, 4180.36145, 4267.46579, 4227.79394, 4001.76508, 3715.55787, 3303.8151},
        {145.807159, 144.639321, 143.290345, 141.768569, 140.082329, 138.239964, 136.24981, 134.120207, 131.85949, 129.475998, 126.978068, 124.374037, 121.672244, 118.881026, 116.00872, 113.063664, 110.054196, 106.988652, 103.875371, 100.72269, 97.5389469, 94.3324788, 91.1116234, 87.8847182, 84.6601009, 81.4461089, 78.2510798, 75.0833512, 71.9512606, 68.8631456, 65.8273436, 62.8521923, 59.9460292, 57.1171918, 54.7532822, 52.6650283, 50.6585997, 48.7300261, 46.8753371, 45.0905625, 43.3717319, 41.714875, 40.1160214, 38.571201, 37.0764433, 35.627778, 34.2212349, 32.8528436, 31.5186338, 30.2146352, 28.9368775, 27.413408, 25.9140482, 24.4406369, 22.9950133, 21.5790162, 20.1944848, 18.843258, 17.5271749, 16.2480745, 15.0077958, 13.8081777, 12.6510594, 11.5382799, 10.4716781, 9.45309305, 8.4843638, 7.62844599, 6.94682791, 6.31530006, 5.73173953, 5.19402338, 4.70002869, 4.24763252, 3.83471195, 3.45914406, 3.1188059, 2.81157456, 2.53532711, 2.28794062, 2.06729216, 1.8712588, 1.69771761, 1.54454567, 1.40018574, 1.26743047, 1.14918594, 1.04363538, 0.948962011, 0.863349075, 0.784979803, 0.712037429, 0.642705187, 0.57516631, 0.507604032, 0.438201585, 0.365142204, 0.286609121, 0.200785571, 0.105854786, 0, -2.55323509, -4.78000843, -6.68831698, -8.28615769, -9.58152751, -10.5824234, -11.2968423, -11.7327811, -11.8982369, -11.8012065, -11.4496869, -10.851675, -10.0151679, -8.94816244, -7.65865557, -6.15464426, -3.87988931, -0.761782494, 1.91543079, 4.26466135, 6.39881998, 8.43081746, 10.4735646, 12.6399722, 15.0429511, 17.795412, 21.0102658, 24.8004232, 29.278795, 35.0129549, 41.4944411, 48.6997657, 56.6054404, 64.7226689, 73.2606172, 82.5525339, 92.7145234, 103.86269, 116.113139, 129.581974, 144.393931, 160.653856, 178.476166, 197.975276, 219.265602, 241.869057, 266.640685, 293.843028, 323.738628, 356.590025, 393.018339, 432.489275, 474.517615, 519.070329, 567.037596, 618.895095, 675.118503, 736.856049, 803.758007, 874.740957, 949.374742, 1027.77022, 1109.8484, 1195.5496, 1284.8469, 1377.69831, 1474.33492, 1574.78553, 1678.07788, 1783.80596, 1894.56687, 2006.48949, 2118.1283, 2228.03776, 2334.77234, 2431.02213, 2522.67206, 2609.74269, 2692.25458, 2770.22829, 2838.00007, 2905.06432, 2973.87772, 3057.89297, 3155.21416, 3259.36371, 3368.67478, 3481.48056, 3596.11422, 3710.90893, 3627.60946, 3711.32412, 3888.49355, 4085.5584, 4228.95929, 4245.13688, 4060.5318, 3835.78428, 3464.224},
        {148.599376, 147.49315, 146.203881, 144.739666, 143.1086, 141.318782, 139.378308, 137.295275, 135.077779, 132.733918, 130.271789, 127.699487, 125.025111, 122.256757, 119.402522, 116.470502, 113.468795, 110.405497, 107.288705, 104.126517, 100.927028, 97.6983359, 94.4485375, 91.1857295, 87.9180088, 84.6534723, 81.4002166, 78.1663388, 74.9599357, 71.7891041, 68.6619408, 65.5865428, 62.5710068, 59.6234296, 57.0840858, 54.7879965, 52.5661903, 50.4159842, 48.3346953, 46.3196406, 44.3681372, 42.4775022, 40.6450525, 38.8681053, 37.1439775, 35.4699862, 33.8434485, 32.2616814, 30.7220019, 29.2217272, 27.7581741, 26.1436821, 24.5645559, 23.0221224, 21.5177087, 20.0526418, 18.6282488, 17.2458565, 15.906792, 14.6123823, 13.3639544, 12.1628354, 11.0103521, 9.90783164, 8.856601, 7.85798718, 6.91331716, 6.07844915, 5.40793206, 4.79062641, 4.22432417, 3.7068173, 3.23589778, 2.80935757, 2.42498864, 2.08058296, 1.77393248, 1.50282919, 1.26506505, 1.05843202, 0.880722074, 0.729727177, 0.603239297, 0.499050401, 0.403195969, 0.319593494, 0.252301773, 0.199494293, 0.159344542, 0.130026008, 0.109712179, 0.0965765436, 0.0887925885, 0.0845338021, 0.0819736721, 0.0792856865, 0.0746433332, 0.0662201, 0.0521894748, 0.0307249455, 0, -2.55564835, -4.76886784, -6.64852686, -8.20349381, -9.44263711, -10.3748251, -11.0089263, -11.353809, -11.4183417, -11.2113927, -10.7418304, -10.0185233, -9.05033967, -7.84614803, -6.41481673, -4.76521417, -2.40544438, 0.727162019, 3.49105528, 5.98467366, 8.3064554, 10.5548388, 12.828262, 15.2251634, 17.8439811, 20.7831535, 24.1411187, 28.0163151, 32.5071809, 38.0050349, 44.2080452, 51.126786, 58.7718314, 66.8885745, 75.62018, 85.1805273, 95.6597448, 107.147961, 119.735304, 133.511902, 148.412108, 164.729235, 182.583892, 202.096683, 223.388216, 246.348436, 271.386275, 298.680005, 328.407897, 360.748223, 395.893915, 433.990664, 475.035983, 519.17937, 566.880632, 618.460253, 674.238716, 734.915416, 800.345813, 869.972009, 943.835095, 1021.55514, 1102.55958, 1186.63213, 1274.16024, 1365.25424, 1460.45998, 1560.0007, 1662.50272, 1767.49566, 1876.84354, 1987.38247, 2097.83401, 2206.91973, 2313.36121, 2412.12011, 2506.61787, 2596.51604, 2681.47616, 2761.15977, 2828.67221, 2894.60204, 2961.42059, 3041.4496, 3133.26129, 3231.32359, 3334.41395, 3441.30982, 3550.78866, 3661.62792, 3677.95512, 3782.06182, 3931.18628, 4082.56674, 4193.44146, 4221.04869, 4122.62667, 3977.29319, 3718.08175},
        {151.484769, 150.39826, 149.133657, 147.698492, 146.100296, 144.346599, 142.444934, 140.402831, 138.227821, 135.927436, 133.509207, 130.980664, 128.349341, 125.622766, 122.808472, 119.913991, 116.946852, 113.914588, 110.824729, 107.684806, 104.502352, 101.284897, 98.039972, 94.7751086, 91.497838, 88.2156912, 84.9361996, 81.6668944, 78.4153066, 75.1889676, 71.9954086, 68.8421607, 65.7367551, 62.6867231, 59.9649692, 57.4407562, 54.9801573, 52.5820917, 50.2454788, 47.9692379, 45.7522883, 43.5935494, 41.4919405, 39.446381, 37.45579, 35.5190871, 33.6351914, 31.8030224, 30.0214993, 28.2895415, 26.6060683, 24.8865959, 23.2152548, 21.5927723, 20.0198758, 18.4972926, 17.02575, 15.6059754, 14.2386961, 12.9246396, 11.664533, 10.4591038, 9.30907936, 8.21518694, 7.17815392, 6.19870764, 5.27757544, 4.46163018, 3.79663677, 3.1881509, 2.6339085, 2.13164549, 1.67909777, 1.27400127, 0.914091909, 0.597105599, 0.320778261, 0.0828458123, -0.118955829, -0.286890746, -0.42322302, -0.530216733, -0.610135968, -0.665244806, -0.708966023, -0.737749645, -0.747911583, -0.741353797, -0.719978248, -0.685686897, -0.640381706, -0.585964635, -0.524337646, -0.457402699, -0.387061756, -0.315216778, -0.243769726, -0.174622561, -0.109677244, -0.050835737, 0, -2.42626594, -4.51940952, -6.28746317, -7.73845929, -8.88043031, -9.72140864, -10.2694267, -10.5325169, -10.5187117, -10.2360435, -9.69254461, -8.89624759, -7.8551848, -6.57738867, -5.07089161, -3.34372603, -0.995557218, 2.02392058, 4.78298736, 7.36111794, 9.83778716, 12.2924698, 14.8046408, 17.4537749, 20.3193469, 23.4808317, 27.0177042, 31.009439, 35.5355111, 40.7887205, 46.6936642, 53.2958195, 60.6406637, 68.6948074, 77.5431612, 87.2916668, 98.0094616, 109.765683, 122.629467, 136.669953, 151.703127, 168.128323, 186.064206, 205.629442, 226.942699, 250.182816, 275.393243, 302.677603, 332.139519, 363.882614, 397.707528, 434.391178, 474.394194, 518.03941, 565.368335, 616.548788, 671.748589, 731.236257, 795.056024, 863.142263, 935.931843, 1012.67078, 1092.44216, 1174.92015, 1260.78054, 1350.23939, 1444.07027, 1542.63378, 1644.24614, 1748.37999, 1856.13106, 1965.09917, 2074.19512, 2182.32969, 2288.41367, 2390.07456, 2487.82727, 2580.90341, 2668.53461, 2749.95248, 2816.61824, 2880.71416, 2944.80202, 3020.47313, 3106.71508, 3198.75317, 3295.76312, 3396.92063, 3501.40143, 3608.38122, 3735.15925, 3862.62039, 3981.04592, 4080.71706, 4151.91509, 4184.92125, 4170.01679, 4101.36589, 3959.33732},
        {154.592035, 153.474949, 152.191362, 150.747946, 149.151371, 147.408309, 145.52543, 143.509405, 141.366905, 139.104601, 136.729163, 134.247262, 131.665569, 128.990755, 126.229491, 123.388447, 120.474295, 117.493704, 114.453347, 111.359893, 108.220014, 105.04038, 101.827662, 98.5885316, 95.3296586, 92.0577142, 88.7793692, 85.5012944, 82.2301606, 78.9726387, 75.7353993, 72.5251134, 69.3484518, 66.2120853, 63.3048272, 60.538446, 57.8225209, 55.1578117, 52.5450784, 49.9850809, 47.4785792, 45.0263332, 42.6291028, 40.287648, 38.0027287, 35.7751048, 33.6055362, 31.494783, 29.4436049, 27.452762, 25.5230142, 23.6868314, 21.9125762, 20.200321, 18.5501384, 16.9621009, 15.436281, 13.9727512, 12.5715842, 11.2328522, 9.95662804, 8.74298405, 7.5919928, 6.50372682, 5.47825863, 4.51566073, 3.61600567, 2.81600633, 2.1514959, 1.54697887, 1.0001525, 0.508714066, 0.0703608501, -0.317209877, -0.656300843, -0.949214772, -1.19825439, -1.40572243, -1.5739216, -1.70515465, -1.80172429, -1.86593325, -1.90008426, -1.90648004, -1.89798563, -1.87140043, -1.82339691, -1.75593502, -1.67097472, -1.57047597, -1.45639872, -1.33070294, -1.19534858, -1.0522956, -0.903503966, -0.75093363, -0.596544552, -0.442296692, -0.290150007, -0.142064457, 0, -2.03730601, -3.79612762, -5.28072117, -6.49534301, -7.44424947, -8.1316969, -8.56194164, -8.73924003, -8.66784841, -8.35202312, -7.79602051, -7.00409692, -5.98050869, -4.72951215, -3.25536366, -1.56231955, 0.645128979, 3.40554917, 6.03664449, 8.59839398, 11.1507767, 13.7537716, 16.4673579, 19.3515146, 22.4662206, 25.8714551, 29.627197, 33.7934255, 38.4301196, 43.5689836, 49.3086387, 55.7175463, 62.8641679, 70.8328497, 79.6841108, 89.4742343, 100.266916, 112.125853, 125.114739, 139.297273, 154.53949, 171.162902, 189.269878, 208.962787, 230.343997, 253.68287, 278.873033, 305.975103, 335.049702, 366.157448, 398.857263, 434.324652, 473.325981, 516.29759, 563.002009, 613.504297, 667.869511, 726.09351, 788.326276, 854.793281, 926.087209, 1001.37974, 1079.75231, 1160.87806, 1245.43264, 1333.63156, 1426.25845, 1523.6761, 1624.16405, 1727.18025, 1833.24819, 1940.59635, 2048.31382, 2155.48971, 2261.21313, 2365.91678, 2467.01026, 2563.24677, 2653.37953, 2736.16173, 2801.03783, 2862.27561, 2922.61777, 2993.47476, 3074.15827, 3160.36837, 3251.59733, 3347.33739, 3447.0808, 3550.31982, 3747.84327, 3888.94357, 3986.36793, 4052.86357, 4101.1777, 4144.05752, 4194.25026, 4192.32518, 4149.84044},
        {158.269136, 157.026245, 155.637794, 154.109219, 152.445957, 150.653445, 148.737117, 146.702411, 144.554762, 142.299606, 139.94238, 137.48852, 134.943462, 132.312642, 129.601496, 126.81546, 123.959971, 121.040464, 118.062376, 115.031143, 111.952201, 108.830986, 105.672935, 102.483483, 99.2680664, 96.0321217, 92.7810848, 89.5203918, 86.255479, 82.9917825, 79.7347384, 76.4897829, 73.262352, 70.0578819, 66.9606355, 63.9349774, 60.9443252, 57.9915567, 55.0795502, 52.2111837, 49.3893351, 46.6168827, 43.8967045, 41.2316784, 38.6246826, 36.0785952, 33.5962942, 31.1806576, 28.8345636, 26.5608902, 24.3625154, 22.4086222, 20.5321787, 18.7324576, 17.008732, 15.3602747, 13.7863586, 12.2862566, 10.8592417, 9.50458659, 8.22156433, 7.00944778, 5.86750984, 4.7950234, 3.79126136, 2.85549663, 1.98700209, 1.20934912, 0.545525657, -0.0553091332, -0.595457525, -1.07722179, -1.5029042, -1.87480702, -2.19523254, -2.466483, -2.6908607, -2.87066791, -3.00820688, -3.1057799, -3.16568923, -3.19023715, -3.18172592, -3.14245783, -3.09073681, -3.02052772, -2.92560311, -2.80774594, -2.66873922, -2.51036593, -2.33440905, -2.14265157, -1.93687647, -1.71886675, -1.49040539, -1.25327537, -1.00925969, -0.760141324, -0.507703263, -0.253728493, 0, -0.979967646, -1.8474724, -2.59591928, -3.21871331, -3.70925953, -4.06096294, -4.26722859, -4.32146149, -4.21706666, -3.94744914, -3.50601394, -2.88616609, -2.08131062, -1.08485255, 0.1098031, 1.5092513, 3.30200983, 5.52066265, 7.76031266, 10.0665383, 12.484918, 15.0610302, 17.8404533, 20.8687658, 24.1915461, 27.8543726, 31.9028237, 36.382478, 41.3389138, 46.7943948, 52.8263629, 59.4873911, 66.8300523, 74.7438311, 83.3628444, 92.8646996, 103.350896, 114.922933, 127.68231, 141.730526, 157.435526, 174.551272, 193.127133, 213.212477, 234.856673, 257.876554, 282.612159, 309.17099, 337.660549, 368.18834, 400.512194, 435.516659, 473.778285, 515.663903, 561.112169, 610.253984, 663.220245, 720.256039, 781.352126, 846.374696, 915.363976, 988.253659, 1065.04445, 1145.70093, 1230.12645, 1318.25249, 1410.28612, 1506.23025, 1605.07721, 1706.39098, 1810.75977, 1916.56572, 2023.01826, 2129.32681, 2234.70079, 2342.2618, 2446.32905, 2545.1339, 2636.90774, 2719.88195, 2780.97252, 2837.2698, 2891.85463, 2956.93096, 3031.96839, 3112.6352, 3198.59107, 3289.49565, 3385.00861, 3484.78961, 3603.30476, 3720.89982, 3834.33867, 3940.3852, 4035.80328, 4117.35681, 4181.80965, 4223.26481, 4219.01046},
        {161.951839, 160.594966, 159.113403, 157.511355, 155.793025, 153.962617, 152.024337, 149.982387, 147.840971, 145.604295, 143.276563, 140.861977, 138.364743, 135.789065, 133.139146, 130.419191, 127.633405, 124.78599, 121.881152, 118.923094, 115.916021, 112.864136, 109.771644, 106.64275, 103.481656, 100.292568, 97.0796889, 93.8472235, 90.5993758, 87.3403499, 84.0743499, 80.80558, 77.5382443, 74.276547, 71.0053704, 67.7389859, 64.4918972, 61.2689355, 58.074932, 54.9147178, 51.793124, 48.714982, 45.6851227, 42.7083774, 39.7895773, 36.9335535, 34.1451372, 31.4291596, 28.7904517, 26.2338449, 23.7641702, 21.6669244, 19.6601888, 17.7427104, 15.9132359, 14.1705122, 12.5132864, 10.9403052, 9.45031547, 8.0420642, 6.71429824, 5.46576447, 4.29520978, 3.20138106, 2.18302519, 1.23888906, 0.367719565, -0.414357123, -1.09163209, -1.70120194, -2.24544642, -2.72674526, -3.1474782, -3.510025, -3.81676539, -4.07007911, -4.2723459, -4.42594551, -4.53325768, -4.59666215, -4.61853866, -4.60126696, -4.54722678, -4.45879786, -4.35892022, -4.23926101, -4.09124026, -3.91657049, -3.71696422, -3.49413397, -3.24979227, -2.98565163, -2.70342459, -2.40482365, -2.09156135, -1.7653502, -1.42790272, -1.08093144, -0.72614888, -0.365267559, 0, -0.0145795258, -0.068802321, -0.146086994, -0.229852155, -0.303516411, -0.350498371, -0.354216644, -0.298089839, -0.165536564, 0.060024572, 0.39517496, 0.856495991, 1.46056906, 2.22397555, 3.16329686, 4.29511438, 5.72176203, 7.47207156, 9.36866031, 11.4464852, 13.7405032, 16.2856712, 19.1169462, 22.269285, 25.7776447, 29.6769821, 34.0022542, 38.7884179, 44.0704301, 49.9000763, 56.2893144, 63.268053, 70.8662004, 78.7501063, 87.1314585, 96.3188938, 106.451388, 117.667918, 130.107459, 143.908988, 159.899464, 177.320494, 196.176447, 216.471694, 238.210605, 260.807694, 285.004653, 310.953316, 338.805517, 368.71309, 400.633841, 435.150782, 472.650708, 513.42654, 557.579546, 605.297044, 656.76635, 712.437814, 772.175939, 835.558288, 902.15997, 972.578354, 1047.63224, 1127.54404, 1211.52559, 1299.2526, 1390.39139, 1484.61524, 1581.63185, 1681.12947, 1783.96653, 1888.48104, 1993.95615, 2099.67504, 2204.92087, 2315.13578, 2421.90422, 2522.9696, 2616.07533, 2698.96485, 2756.12179, 2807.3892, 2856.19324, 2915.78339, 2985.69743, 3061.38011, 3142.57389, 3229.0212, 3320.46448, 3416.64617, 3443.96817, 3534.85511, 3666.37808, 3815.60815, 3959.61638, 4075.47384, 4140.25162, 4213.99502, 4231.18751},
        {165.33743, 163.940135, 162.433194, 160.819818, 159.103214, 157.286593, 155.373164, 153.366138, 151.268722, 149.084127, 146.815563, 144.466238, 142.039363, 139.538146, 136.965799, 134.325529, 131.620546, 128.85406, 126.029281, 123.149418, 120.217681, 117.237278, 114.21142, 111.143316, 108.036176, 104.893209, 101.717625, 98.5126326, 95.2814421, 92.0272628, 88.7533043, 85.462776, 82.1588875, 78.8448484, 75.4268962, 71.958766, 68.4953586, 65.0430304, 61.6081379, 58.1970374, 54.8160854, 51.4716384, 48.1700528, 44.9176851, 41.7208916, 38.5860289, 35.5194534, 32.5275215, 29.6165897, 26.7930145, 24.0631522, 21.785467, 19.6065746, 17.5251986, 15.5400623, 13.6498891, 11.8534026, 10.1493261, 8.53638313, 7.01329706, 5.57879136, 4.23158946, 2.9704148, 1.79399084, 0.701040991, -0.30971129, -1.23954257, -2.06913915, -2.77968171, -3.41491266, -3.97744334, -4.4698851, -4.89484927, -5.2549472, -5.55279024, -5.79098972, -5.972157, -6.0989034, -6.17384028, -6.19957898, -6.17873084, -6.11390721, -6.00771942, -5.86277882, -5.70344619, -5.52100068, -5.30646004, -5.0617298, -4.78871549, -4.48932265, -4.16545679, -3.81902347, -3.4519282, -3.06607652, -2.66337397, -2.24572607, -1.81503836, -1.37321636, -0.922165614, -0.46379165, 0, 0.432124521, 0.757333341, 0.996837286, 1.17184718, 1.30357384, 1.4132281, 1.52202078, 1.6511627, 1.82186469, 2.05533757, 2.37279216, 2.7954393, 3.34448979, 4.04115447, 4.90664416, 5.96216968, 7.27123135, 8.86108135, 10.6572757, 12.6900873, 14.9897889, 17.5866534, 20.5109538, 23.7929627, 27.4629532, 31.551198, 36.0879701, 41.1035422, 46.6281872, 52.7520626, 59.4235989, 66.6551037, 74.4588844, 82.3771036, 90.6571412, 99.6717491, 109.574278, 120.51808, 132.656504, 146.142903, 161.929378, 179.127431, 197.734135, 217.746565, 239.161793, 261.419058, 285.212728, 310.679336, 337.955413, 367.177493, 398.242401, 431.819352, 468.384086, 508.252875, 551.416403, 598.011536, 648.17514, 702.217132, 760.061998, 821.445441, 885.97798, 954.331047, 1027.49551, 1105.7455, 1188.14072, 1274.29832, 1363.70624, 1455.94812, 1551.0814, 1648.89549, 1750.62729, 1854.39609, 1959.49027, 2065.19822, 2170.80833, 2283.33723, 2392.41299, 2495.39195, 2589.63042, 2672.48473, 2726.52106, 2773.74598, 2817.85442, 2873.06375, 2938.90381, 3010.5201, 3087.66168, 3170.0776, 3257.51692, 3349.72869, 3331.79808, 3406.96446, 3543.48239, 3709.60647, 3873.59127, 4003.69137, 4068.16135, 4166.29693, 4201.61824},
        {168.065254, 166.771321, 165.370561, 163.865684, 162.259398, 160.554412, 158.753433, 156.859171, 154.874334, 152.801629, 150.643766, 148.403453, 146.083398, 143.68631, 141.214897, 138.671868, 136.05993, 133.381793, 130.640164, 127.837752, 124.977266, 122.061414, 119.092904, 116.074445, 113.008744, 109.898512, 106.746455, 103.555283, 100.327703, 97.066424, 93.7741547, 90.4536034, 87.1074784, 83.7384881, 80.2112924, 76.6005278, 72.9824896, 69.3643661, 65.7533459, 62.1566171, 58.5813681, 55.0347872, 51.5240628, 48.0563833, 44.6389368, 41.2789119, 37.9834968, 34.7598799, 31.6152494, 28.5567938, 25.5917014, 23.0848426, 20.6779698, 18.3705175, 16.1619202, 14.0516123, 12.0390284, 10.123603, 8.30477042, 6.5819653, 4.95462209, 3.42217527, 1.98405934, 0.639708786, -0.611441909, -1.76995825, -2.83640575, -3.77254016, -4.54104902, -5.22254008, -5.82009473, -6.33679436, -6.77572036, -7.13995413, -7.43257704, -7.65667049, -7.81531588, -7.91159459, -7.94858801, -7.92937754, -7.85704456, -7.73467046, -7.56533664, -7.35212448, -7.11485134, -6.84787861, -6.54536652, -6.20985336, -5.8438774, -5.44997691, -5.03069017, -4.58855547, -4.12611106, -3.64589524, -3.15044628, -2.64230245, -2.12400203, -1.5980833, -1.06708454, -0.53354401, 0, -0.141915049, -0.289791222, -0.428413625, -0.542567363, -0.617037541, -0.636609265, -0.586067641, -0.450197773, -0.213784766, 0.138386273, 0.621530239, 1.25086203, 2.04159653, 3.00894865, 4.16813327, 5.5343653, 7.2116231, 9.22780207, 11.3973426, 13.7544802, 16.3334507, 19.1684896, 22.2938326, 25.7437153, 29.5523734, 33.7540424, 38.3829582, 43.4733562, 49.0594721, 55.2458773, 61.970682, 69.2470211, 77.0880295, 85.1587396, 93.6463376, 102.830837, 112.829804, 123.760804, 135.741402, 148.889165, 163.594117, 179.618442, 197.0264, 215.882247, 236.250243, 258.492892, 282.301644, 307.666198, 334.57625, 363.021497, 392.247334, 423.897463, 459.161634, 498.603954, 541.511161, 587.743506, 637.161241, 689.225804, 744.286897, 803.129287, 866.379508, 933.623131, 1004.74284, 1079.65851, 1158.35308, 1240.78053, 1326.99462, 1416.97518, 1510.33633, 1606.89905, 1708.29053, 1812.24769, 1917.96629, 2024.64209, 2131.47086, 2245.83684, 2356.70021, 2461.2096, 2556.51367, 2639.76107, 2692.59834, 2738.01096, 2779.79135, 2832.68722, 2896.15452, 2965.08443, 3039.16111, 3118.06872, 3201.49143, 3289.11338, 3340.07222, 3427.1522, 3537.75714, 3659.29091, 3779.15733, 3884.76028, 3963.50357, 4081.81906, 4147.15653},
        {170.892876, 169.753852, 168.504458, 167.147246, 165.684767, 164.119573, 162.454215, 160.691246, 158.833216, 156.882678, 154.842184, 152.714284, 150.501531, 148.206476, 145.83167, 143.379667, 140.853016, 138.254271, 135.585982, 132.850701, 130.05098, 127.189371, 124.268425, 121.290693, 118.258728, 115.175082, 112.042305, 108.86295, 105.639568, 102.37471, 99.0709297, 95.7307771, 92.3568043, 88.951563, 85.3774189, 81.7099651, 78.02648, 74.3340646, 70.6398197, 66.9508465, 63.2742458, 59.6171186, 55.9865659, 52.3896886, 48.8335878, 45.3253644, 41.8721194, 38.4809538, 35.1589684, 31.9132644, 28.7509426, 25.9572206, 23.2550539, 20.6455143, 18.1296738, 15.7086043, 13.3833777, 11.1550661, 9.02474136, 6.99347541, 5.06234023, 3.23240775, 1.50474993, -0.119561298, -1.63945397, -3.05385615, -4.36169589, -5.48783052, -6.35889576, -7.126585, -7.79462797, -8.36675445, -8.84669419, -9.23817693, -9.54493245, -9.77069049, -9.91918082, -9.99413318, -9.99927734, -9.93834305, -9.81506007, -9.63315815, -9.39636706, -9.10841654, -8.78289047, -8.41838429, -8.01337497, -7.5712725, -7.09548684, -6.58942797, -6.05650586, -5.50013047, -4.92371178, -4.33065976, -3.72438439, -3.10829562, -2.48580344, -1.86031782, -1.23524873, -0.614006128, 0, -0.78365739, -1.46162142, -2.02557829, -2.46721419, -2.77821532, -2.95026787, -2.97505804, -2.84427202, -2.54959601, -2.0827162, -1.43531878, -0.599089963, 0.434284071, 1.67311712, 3.125723, 4.8004155, 6.84801785, 9.29720232, 11.83394, 14.4970825, 17.3254813, 20.3579879, 23.6334539, 27.1907307, 31.0686701, 35.3061234, 39.9419422, 45.014978, 50.5640825, 56.7027596, 63.3678358, 70.5757668, 78.3430083, 86.4493656, 95.0296198, 104.281659, 114.292933, 125.150894, 136.942992, 149.756678, 163.588809, 178.645002, 195.030434, 212.85028, 232.209716, 253.922315, 277.207757, 301.99412, 328.209481, 355.781918, 383.699941, 413.979552, 448.179947, 486.967218, 529.184034, 574.491744, 622.551698, 672.076516, 723.891895, 779.858507, 841.798562, 907.995775, 977.024437, 1048.37009, 1123.06231, 1201.42197, 1284.13221, 1371.60783, 1462.9351, 1557.95178, 1658.57537, 1762.24392, 1868.0753, 1975.18733, 2082.69789, 2198.2903, 2310.37554, 2415.93011, 2511.93046, 2595.35308, 2647.16811, 2691.02924, 2730.7728, 2781.46292, 2842.49159, 2908.57255, 2979.33171, 3054.39497, 3133.38823, 3215.93741, 3323.42736, 3419.24193, 3506.61063, 3588.763, 3668.92858, 3750.33689, 3836.21748, 3964.7291, 4053.71547},
        {174.307038, 173.378106, 172.324775, 171.150052, 169.856946, 168.448466, 166.92762, 165.297416, 163.560863, 161.720969, 159.780743, 157.743193, 155.611328, 153.388156, 151.076685, 148.679924, 146.200882, 143.642566, 141.007985, 138.300148, 135.522064, 132.676739, 129.767184, 126.796406, 123.767415, 120.683217, 117.546822, 114.361238, 111.129474, 107.854538, 104.539439, 101.187184, 97.8007832, 94.3832439, 90.867116, 87.2921193, 83.6988207, 80.092515, 76.4784973, 72.8620626, 69.2485057, 65.6431217, 62.0512055, 58.4780522, 54.9289566, 51.4092137, 47.9241185, 44.478966, 41.0790511, 37.7296688, 34.4361141, 31.2373595, 28.1042924, 25.0414776, 22.05348, 19.1448645, 16.3201962, 13.5840397, 10.9409601, 8.39552231, 5.95229115, 3.61583155, 1.39070842, -0.718513347, -2.70726884, -4.57099316, -6.30512139, -7.75702259, -8.7776208, -9.66772624, -10.4323725, -11.0765933, -11.6054222, -12.0238928, -12.3370386, -12.5498935, -12.6674909, -12.6948644, -12.6370478, -12.4990746, -12.2859784, -12.0027928, -11.6545515, -11.2462881, -10.8082945, -10.332444, -9.81030607, -9.24609465, -8.64402368, -8.00830712, -7.34315888, -6.6527929, -5.94142311, -5.21326346, -4.47252787, -3.72343028, -2.97018462, -2.21700483, -1.46810484, -0.727698584, 0, -0.833490275, -1.54528201, -2.128481, -2.57619306, -2.88152399, -3.03757958, -3.03746565, -2.87428799, -2.54115241, -2.03116471, -1.3374307, -0.453056163, 0.628853081, 1.91519123, 3.41285249, 5.12873106, 7.20516038, 9.66838312, 12.2189428, 14.8927563, 17.7257404, 20.7538122, 24.0128885, 27.5388862, 31.3677221, 35.5353133, 40.0775765, 45.0304287, 50.4297868, 56.3672216, 62.8025896, 69.7551115, 77.2440081, 84.9913653, 93.1649713, 102.01185, 111.640362, 122.15887, 133.675733, 146.299313, 160.613897, 176.107072, 192.794086, 210.690183, 229.810608, 249.387137, 270.414353, 293.103368, 317.665293, 344.311242, 373.333152, 404.762524, 439.154075, 476.709199, 516.907924, 559.554159, 604.45181, 650.191333, 698.065871, 750.692338, 811.17452, 876.405564, 943.706097, 1012.08616, 1083.41843, 1158.26158, 1237.69536, 1322.41352, 1411.19927, 1503.91651, 1602.60334, 1704.61502, 1809.06286, 1915.05817, 2021.71229, 2137.49414, 2249.81803, 2355.45585, 2451.17952, 2533.76095, 2583.34226, 2624.41168, 2660.86819, 2707.97494, 2765.13258, 2827.00669, 2893.23466, 2963.45389, 3037.3018, 3114.41578, 3176.42255, 3256.0398, 3346.84068, 3442.39831, 3536.28584, 3622.0764, 3693.34312, 3816.92893, 3917.85728},
        {178.020724, 177.235703, 176.318616, 175.272747, 174.101378, 172.807795, 171.39528, 169.867117, 168.226591, 166.476984, 164.621581, 162.663665, 160.606519, 158.453428, 156.207676, 153.872545, 151.451321, 148.947285, 146.363723, 143.703918, 140.971153, 138.168712, 135.299879, 132.367938, 129.376173, 126.327866, 123.226302, 120.074765, 116.876538, 113.634905, 110.353149, 107.034555, 103.682406, 100.299986, 96.8897042, 93.4553005, 90.000524, 86.5286866, 83.0431006, 79.5470778, 76.0439304, 72.5369704, 69.02951, 65.524861, 62.0263357, 58.537246, 55.0609041, 51.6006219, 48.1597116, 44.7414852, 41.3492547, 37.7786391, 34.245146, 30.7565897, 27.3207848, 23.9455457, 20.6386867, 17.4080224, 14.2613672, 11.2065356, 8.25134185, 5.40360055, 2.67112609, 0.0617329193, -2.41676453, -4.75655181, -6.94981449, -8.78396512, -10.0513327, -11.1664319, -12.1347228, -12.9616656, -13.6527204, -14.2133473, -14.6490067, -14.9651586, -15.1672631, -15.2607805, -15.251171, -15.1438946, -14.9444115, -14.6581819, -14.290666, -13.8473239, -13.325216, -12.734179, -12.0841506, -11.3808636, -10.6300507, -9.83744474, -9.00877838, -8.14978441, -7.26619557, -6.36374462, -5.44816431, -4.5251874, -3.60054662, -2.67997474, -1.76920451, -0.87396868, 0, -0.886406705, -1.64028383, -2.25590555, -2.72754604, -3.04947944, -3.21597995, -3.22132171, -3.05977891, -2.72562571, -2.21313627, -1.51658477, -0.630245363, 0.451607771, 1.73470047, 3.22475856, 4.92750788, 6.99183904, 9.44392999, 11.9656815, 14.5934974, 17.3637819, 20.3129389, 23.4773725, 26.8934865, 30.5976851, 34.6263721, 39.0159517, 43.8028278, 49.0234044, 54.7876456, 61.0314232, 67.7690734, 75.0149321, 82.4045314, 90.1416092, 98.5309179, 107.700435, 117.778136, 128.892, 141.170004, 155.69552, 171.350358, 188.075568, 205.812202, 224.501313, 242.145925, 261.109623, 281.817966, 304.696511, 330.170816, 359.556257, 391.301021, 425.052296, 460.621855, 498.157506, 537.656185, 579.114826, 621.334554, 665.77989, 715.219878, 773.251167, 836.256945, 901.149509, 966.767794, 1035.21531, 1107.09707, 1183.56253, 1265.35787, 1351.23295, 1441.06689, 1536.6283, 1635.61605, 1737.25517, 1840.77067, 1945.38758, 2061.119, 2173.70486, 2279.67317, 2375.55194, 2457.86919, 2505.99683, 2545.05634, 2578.92833, 2622.59013, 2675.52989, 2732.61215, 2793.55625, 2858.08153, 2925.90734, 2996.75302, 2999.63806, 3059.43678, 3156.76334, 3272.23187, 3386.45651, 3480.05142, 3533.63073, 3645.53178, 3744.29914},
        {182.127512, 180.879312, 179.561161, 178.173858, 176.7182, 175.194983, 173.605007, 171.949066, 170.22796, 168.442485, 166.593439, 164.681619, 162.707822, 160.672846, 158.577488, 156.422545, 154.208815, 151.937095, 149.608183, 147.222875, 144.781969, 142.286263, 139.736553, 137.133638, 134.478314, 131.771378, 129.013629, 126.205863, 123.348878, 120.443471, 117.490439, 114.49058, 111.444691, 108.353569, 105.150964, 101.872607, 98.555035, 95.2012225, 91.8141422, 88.3967674, 84.9520712, 81.4830267, 77.9926073, 74.4837861, 70.9595362, 67.4228309, 63.8766434, 60.3239469, 56.7677145, 53.2109194, 49.6565349, 45.7543396, 41.8681578, 38.0086192, 34.1863536, 30.4119908, 26.6961605, 23.0494924, 19.4826164, 16.0061622, 12.6307596, 9.36703821, 6.22562792, 3.21715845, 0.352259561, -2.35843899, -4.90430744, -6.99980425, -8.36598836, -9.55921499, -10.5866758, -11.4555624, -12.1730665, -12.7463798, -13.1826938, -13.4892003, -13.6730909, -13.7415572, -13.701791, -13.5609838, -13.3263274, -13.0050133, -12.6042333, -12.131179, -11.6441231, -11.1236423, -10.5496994, -9.92782831, -9.26356303, -8.56243755, -7.82998582, -7.07174183, -6.29323955, -5.50001295, -4.69759602, -3.89152273, -3.08732705, -2.29054297, -1.50670444, -0.741345463, 0, -2.56355904, -4.69687761, -6.41514243, -7.73354025, -8.66725779, -9.23148179, -9.44139899, -9.31219611, -8.85905989, -8.09717707, -7.04173437, -5.70791854, -4.1109163, -2.26591439, -0.188099537, 2.10734152, 5.20144911, 9.16421182, 12.631834, 15.716892, 18.5319619, 21.1896199, 23.8024423, 26.4830052, 29.343885, 32.4976577, 36.0568996, 40.134187, 44.8420959, 50.9077343, 57.6038185, 64.8585653, 72.6001912, 80.0518307, 87.4942416, 95.3962039, 103.897459, 113.137749, 123.256815, 134.394398, 147.220928, 161.183946, 176.31936, 192.663083, 210.251026, 230.161444, 251.127317, 272.923971, 295.32673, 318.110919, 339.596131, 362.792652, 389.774116, 421.433872, 456.25552, 493.804585, 533.646596, 574.155571, 616.359344, 662.585581, 714.771022, 771.485203, 832.103486, 896.29093, 964.203481, 1035.77176, 1110.87304, 1189.42412, 1271.53738, 1357.21457, 1447.85721, 1541.85193, 1638.71592, 1737.96638, 1839.12052, 1955.223, 2068.8817, 2176.23197, 2273.40915, 2356.54858, 2403.67911, 2441.11356, 2472.7472, 2513.96793, 2564.12184, 2617.76644, 2674.48722, 2733.86972, 2795.49946, 2858.96195, 2841.54564, 2884.83073, 2969.85368, 3077.65095, 3189.25899, 3285.71425, 3348.05319, 3453.67784, 3500.31941},
        {189.757548, 188.974412, 188.068227, 187.042066, 185.899003, 184.642109, 183.274459, 181.799126, 180.219183, 178.537702, 176.757759, 174.882424, 172.914772, 170.857876, 168.71481, 166.488645, 164.182456, 161.799315, 159.342296, 156.814472, 154.218916, 151.558702, 148.836902, 146.05659, 143.220839, 140.332722, 137.395312, 134.411683, 131.384907, 128.318058, 125.214209, 122.076434, 118.907805, 115.711395, 112.53458, 109.357349, 106.159161, 102.941651, 99.7064564, 96.4552109, 93.1895507, 89.9111112, 86.621528, 83.3224365, 80.0154724, 76.7022712, 73.3844684, 70.0636994, 66.7416, 63.4198055, 60.0999516, 56.577289, 53.064312, 49.5671302, 46.091853, 42.6445901, 39.231451, 35.8585453, 32.5319825, 29.2578721, 26.0423237, 22.8914468, 19.811351, 16.8081459, 13.887941, 11.0568458, 8.32096987, 5.7383315, 3.3657026, 1.10213493, -1.04962699, -3.08683865, -5.00675553, -6.80663311, -8.48372686, -10.0352923, -11.4585848, -12.75086, -13.9093733, -14.9313802, -15.8141361, -16.5548966, -17.1509171, -17.5994532, -17.3614263, -16.7135401, -15.9389443, -15.0522994, -14.068266, -13.0015046, -11.8666758, -10.6784402, -9.45145819, -8.20039042, -6.93989741, -5.68463972, -4.44927788, -3.24847244, -2.09688395, -1.00917295, 0, -0.464272642, -1.06208401, -1.76680075, -2.55178949, -3.39041686, -4.2560495, -5.12205404, -5.96179713, -6.74864539, -7.45596545, -8.05712397, -8.52548756, -8.83442287, -8.95729652, -8.86747516, -8.53832542, -8.23849337, -7.98352841, -7.07890323, -5.56125863, -3.4672354, -0.833474354, 2.30338372, 5.90669802, 9.93982775, 14.3661321, 19.1489703, 24.2517015, 29.637685, 35.0338828, 40.7267301, 46.7505054, 53.1394868, 60.0336828, 67.4145066, 75.235177, 83.4982533, 92.2062948, 101.361861, 110.967511, 120.325326, 130.351533, 141.185741, 152.967559, 165.836597, 178.49663, 192.882059, 209.491454, 228.823382, 251.376412, 279.686181, 309.724435, 338.595908, 365.214922, 392.190382, 420.472399, 451.011086, 485.355707, 523.721053, 565.668289, 612.134435, 662.043424, 714.144614, 768.076567, 824.984553, 885.322242, 950.398908, 1020.89004, 1094.33391, 1170.04335, 1243.08663, 1317.67421, 1394.58825, 1474.61088, 1558.52424, 1666.26664, 1774.67502, 1879.74249, 1977.46216, 2063.82712, 2119.85835, 2166.50251, 2206.16932, 2243.16484, 2279.14817, 2314.98826, 2352.38373, 2393.03318, 2438.63521, 2490.88842, 2394.56646, 2444.55019, 2584.74341, 2759.04992, 2911.37352, 2985.618, 2925.68715, 3056.31619, 3112.65579},
        {192.62829, 192.383279, 191.979674, 191.421691, 190.713545, 189.85945, 188.863624, 187.73028, 186.463635, 185.067905, 183.547303, 181.906047, 180.148351, 178.278431, 176.300503, 174.218781, 172.037482, 169.76082, 167.393012, 164.938272, 162.400816, 159.78486, 157.094619, 154.334308, 151.508144, 148.620341, 145.675114, 142.676681, 139.629254, 136.537052, 133.404288, 130.235178, 127.033937, 123.804782, 120.650349, 117.523573, 114.376205, 111.209267, 108.02378, 104.820765, 101.601245, 98.3662403, 95.1167727, 91.8538638, 88.578535, 85.2918078, 81.9947037, 78.6882441, 75.3734506, 72.0513446, 68.7229477, 65.2631982, 61.8019339, 58.3429097, 54.8898801, 51.4465999, 48.0168238, 44.6043066, 41.212803, 37.8460676, 34.5078553, 31.2019207, 27.9320185, 24.7019036, 21.5153305, 18.376054, 15.2878289, 12.146028, 8.84862633, 5.62290814, 2.47965406, -0.570355221, -3.51633903, -6.34751669, -9.05310753, -11.6223309, -14.044406, -16.3085524, -18.4039891, -20.3199357, -22.0456114, -23.5702356, -24.8830275, -25.9732065, -25.8249983, -24.9512576, -23.8769187, -22.6238147, -21.2137791, -19.668645, -18.0102459, -16.2604149, -14.4409855, -12.5737908, -10.6806643, -8.78343911, -6.90394867, -5.06402624, -3.28550511, -1.5902186, 0, -1.07318919, -2.17789359, -3.29368944, -4.400153, -5.4768605, -6.5033882, -7.45931233, -8.32420916, -9.07765492, -9.69922586, -10.1684982, -10.4650483, -10.5684522, -10.4582864, -10.1141269, -9.51555012, -8.98976145, -8.56599847, -7.43753325, -5.65843399, -3.28276893, -0.364606292, 3.04198571, 6.88293885, 11.1041849, 15.6516557, 20.4712829, 25.5089984, 30.7107339, 35.6010443, 40.7017431, 46.0851753, 51.8236856, 58.2888541, 65.4034081, 73.0102788, 81.0920404, 89.6312673, 98.6105336, 108.012414, 117.135959, 126.855295, 137.286729, 148.546569, 160.75112, 172.779886, 186.295179, 201.722507, 219.48738, 240.015304, 265.35449, 292.324443, 318.540363, 343.153902, 368.453633, 395.319716, 424.632311, 458.27741, 495.900743, 536.05077, 578.186392, 622.784759, 669.748774, 719.254349, 771.939999, 828.231896, 889.175185, 955.356513, 1025.09296, 1097.98541, 1171.11729, 1246.99422, 1326.08851, 1408.87246, 1495.81836, 1605.93124, 1716.5175, 1823.41628, 1922.46669, 2009.50787, 2064.6474, 2109.94364, 2147.9778, 2185.3915, 2223.15872, 2260.56159, 2298.65869, 2338.50858, 2381.16984, 2427.70102, 2434.03458, 2490.19719, 2575.96849, 2671.12815, 2755.45581, 2808.73114, 2810.73379, 2876.42573, 2952.08091},
        {202.644418, 202.442402, 202.111105, 201.653542, 201.07273, 200.371684, 199.55342, 198.620954, 197.577301, 196.425478, 195.1685, 193.809383, 192.351142, 190.796794, 189.149355, 187.411839, 185.587263, 183.678643, 181.688995, 179.621334, 177.478676, 175.264036, 172.980432, 170.630877, 168.218389, 165.745983, 163.216675, 160.633481, 157.999416, 155.317496, 152.590737, 149.822155, 147.014765, 144.171584, 141.343869, 138.509501, 135.645779, 132.754153, 129.836075, 126.892994, 123.926361, 120.937625, 117.928237, 114.899647, 111.853305, 108.790662, 105.713168, 102.622273, 99.5194277, 96.4060817, 93.2836854, 90.170004, 87.0498192, 83.9242276, 80.7943257, 77.6612102, 74.5259775, 71.3897242, 68.2535469, 65.1185422, 61.9858067, 58.8564368, 55.7315292, 52.6121805, 49.4994871, 46.3945457, 43.2984528, 40.0494557, 36.4897114, 32.9561812, 29.4605185, 26.0143766, 22.6294091, 19.3172693, 16.0896106, 12.9580866, 9.93435055, 7.03005597, 4.25685626, 1.62640485, -0.84964483, -3.15963936, -5.29192532, -7.23484927, -7.98347051, -8.04367171, -7.9332741, -7.67285834, -7.28300508, -6.78429497, -6.19730869, -5.54262687, -4.84083018, -4.11249927, -3.3782148, -2.65855743, -1.97410782, -1.34544661, -0.793154464, -0.337812043, 0, -5.27410188, -10.0410472, -14.3146798, -18.1088436, -21.4373827, -24.314141, -26.7529624, -28.7676908, -30.3721702, -31.5802445, -32.4057577, -32.8625537, -32.9644764, -32.7253699, -32.1590779, -31.2794446, -30.1324341, -28.736479, -27.0327701, -25.0420343, -22.7849984, -20.2823894, -17.5549341, -14.6233594, -11.5083921, -8.2307591, -4.81118728, -1.2704035, 2.37086536, 5.87102122, 9.51119455, 13.3369198, 17.3937316, 21.7983637, 26.5607509, 31.6718416, 37.1558104, 43.0368321, 49.3390813, 56.0867327, 63.1527035, 70.7584608, 78.9577731, 87.8044091, 97.3521373, 107.593928, 118.659547, 130.617963, 143.538145, 157.489061, 172.687078, 188.873609, 205.764525, 223.302211, 241.89153, 261.729701, 283.013946, 305.83741, 330.52504, 357.515323, 387.758191, 420.194739, 453.395946, 487.033836, 522.646088, 560.914013, 603.076466, 649.959307, 700.344075, 754.168692, 809.919243, 869.208849, 932.477992, 1000.16715, 1072.71681, 1168.19415, 1265.00627, 1359.187, 1446.77014, 1523.78951, 1571.40233, 1610.43672, 1643.30216, 1677.23141, 1712.72475, 1748.27276, 1784.4762, 1821.93583, 1861.2524, 1903.02668, 1955.09636, 2011.60688, 2068.75893, 2122.75319, 2169.79035, 2206.07109, 2227.7961, 2280.07228, 2343.36149}
    };
    
    const double a4tab[98][200] = {{-1766.7572, -1749.18855, -1731.40558, -1713.41905, -1695.2397, -1676.87827, -1658.34553, -1639.65221, -1620.80907, -1601.82685, -1582.7163, -1563.48817, -1544.15322, -1524.72217, -1505.2058, -1485.61483, -1465.96003, -1446.25214, -1426.50191, -1406.72008, -1386.91741, -1367.10465, -1347.29253, -1327.49182, -1307.71325, -1287.96759, -1268.26556, -1248.61793, -1229.03545, -1209.52885, -1190.10889, -1170.78632, -1151.57189, -1132.47634, -1114.02103, -1095.95066, -1078.00381, -1060.17465, -1042.45736, -1024.84611, -1007.33509, -989.918464, -972.590423, -955.34514, -938.176794, -921.079562, -904.047622, -887.075153, -870.156333, -853.28534, -836.456351, -819.512409, -802.602104, -785.722891, -768.872224, -752.047557, -735.246345, -718.466043, -701.704105, -684.957985, -668.225137, -651.503016, -634.789077, -618.080774, -601.375561, -584.670893, -567.964224, -551.188272, -534.277311, -517.362308, -500.444913, -483.526779, -466.609555, -449.694892, -432.784442, -415.879856, -398.982785, -382.094879, -365.21779, -348.353168, -331.502665, -314.667931, -297.850618, -281.052376, -264.574224, -248.261831, -231.95727, -215.652479, -199.339392, -183.009945, -166.656076, -150.269718, -133.84281, -117.367285, -100.835081, -84.2381337, -67.5683784, -50.8177515, -33.9781888, -17.0416263, 0, 19.1746688, 37.9365544, 56.3501986, 74.4801434, 92.3909305, 110.147102, 127.813199, 145.453764, 163.133338, 180.916464, 198.867684, 217.051538, 235.532569, 254.375319, 273.64433, 293.404142, 313.773689, 334.825281, 356.500436, 378.875352, 402.026224, 426.02925, 450.960626, 476.896549, 503.913216, 532.086823, 561.493567, 592.209645, 624.311253, 657.696936, 692.685682, 729.406984, 767.990335, 808.116944, 850.140446, 894.534017, 941.561637, 991.487282, 1044.57493, 1101.08856, 1160.52898, 1224.15561, 1292.38174, 1365.62067, 1444.28569, 1525.22296, 1613.30469, 1709.83599, 1816.12193, 1933.46758, 2065.77763, 2208.5803, 2357.43451, 2511.46638, 2677.08511, 2857.42995, 3055.64015, 3276.90575, 3521.84913, 3788.85541, 4076.91686, 4391.02783, 4732.56227, 5104.7603, 5514.02414, 5965.30455, 6462.33843, 7009.76183, 7616.66165, 8289.60712, 9021.112, 9833.96341, 10739.596, 11749.4442, 12874.9429, 14600.5298, 16346.3855, 18005.6936, 19471.638, 20637.4024, 21117.3678, 21269.3892, 21106.1372, 20526.6706, 19588.6315, 18397., 17007.0512, 15474.0602, 13853.3021, 12200.0518, 10554.486, 9059.9718, 7706.33993, 6483.42127, 5381.04667, 4389.04698, 3497.25303, 2689.0974, 1951.75256},
        {-1661.23266, -1645.23548, -1628.96586, -1612.43781, -1595.66538, -1578.6626, -1561.44351, -1544.02213, -1526.41251, -1508.62869, -1490.68468, -1472.59454, -1454.37229, -1436.03197, -1417.58761, -1399.05326, -1380.44293, -1361.77068, -1343.05053, -1324.29651, -1305.52267, -1286.74304, -1267.97165, -1249.22254, -1230.50974, -1211.84729, -1193.24922, -1174.72956, -1156.30236, -1137.98165, -1119.78145, -1101.71581, -1083.79877, -1066.04434, -1049.34448, -1033.25578, -1017.32434, -1001.53568, -985.875372, -970.328944, -954.881947, -939.519926, -924.228426, -908.992994, -893.799174, -878.632512, -863.478554, -848.322846, -833.150932, -817.948359, -802.700672, -786.50964, -770.263744, -753.967688, -737.626175, -721.243911, -704.825599, -688.375943, -671.899648, -655.401417, -638.885954, -622.357965, -605.822152, -589.28322, -572.745873, -556.214816, -539.694752, -523.25083, -506.946751, -490.662552, -474.39902, -458.15694, -441.937098, -425.74028, -409.567272, -393.418858, -377.295825, -361.198959, -345.129045, -329.086868, -313.073215, -297.088871, -281.134621, -265.211252, -249.627697, -234.224186, -218.837246, -203.457662, -188.07622, -172.683706, -157.270906, -141.828605, -126.347589, -110.818644, -95.2325543, -79.5801071, -63.8520876, -48.0392814, -32.1324743, -16.1224519, 0, 18.631188, 36.8179774, 54.6249164, 72.1165534, 89.3574366, 106.412114, 123.345135, 140.221047, 157.104398, 174.059736, 191.151611, 208.44457, 226.003161, 243.891934, 262.175435, 280.918214, 300.160596, 319.963671, 340.446773, 361.669261, 383.690491, 406.569824, 430.366615, 455.140223, 480.950007, 507.855323, 535.91553, 565.189986, 595.738049, 627.273087, 660.327311, 695.063874, 731.645931, 770.284321, 811.118356, 854.274634, 899.902003, 948.149312, 999.165412, 1053.09915, 1108.35958, 1167.36484, 1230.6042, 1298.56689, 1371.74215, 1449.63621, 1533.96709, 1625.46979, 1724.87933, 1832.9307, 1951.38092, 2078.69385, 2212.11928, 2351.39934, 2501.36464, 2664.56116, 2843.53489, 3042.31182, 3261.62156, 3500.57918, 3764.46999, 4044.8957, 4323.56168, 4597.71899, 4890.95998, 5214.78592, 5592.15334, 6037.53338, 6523.39458, 7045.96454, 7591.63185, 8167.97679, 8774.63277, 9411.23318, 10077.4114, 10743.1857, 11445.2084, 12190.5167, 12986.1479, 13839.1392, 14987.8517, 16053.7826, 16944.8303, 17121.4612, 16668.6822, 15857.9298, 14764.8894, 13465.2462, 12034.6854, 10548.8923, 9205.32317, 7979.00757, 6863.62985, 5852.87439, 4940.42557, 4119.96776, 3385.18533, 2733.65426, 2184.39136},
        {-1563.20972, -1547.11804, -1530.91591, -1514.61038, -1498.20846, -1481.7172, -1465.1436, -1448.49471, -1431.77754, -1414.99912, -1398.16649, -1381.28667, -1364.36669, -1347.41357, -1330.43434, -1313.43602, -1296.42566, -1279.41026, -1262.39687, -1245.39251, -1228.40419, -1211.43896, -1194.50384, -1177.60586, -1160.75204, -1143.94941, -1127.205, -1110.52583, -1093.91894, -1077.39134, -1060.95008, -1044.60216, -1028.35463, -1012.21451, -996.747581, -981.66974, -966.690168, -951.797759, -936.98141, -922.230013, -907.532465, -892.877658, -878.254489, -863.65185, -849.058638, -834.463746, -819.856069, -805.224502, -790.557939, -775.845274, -761.075403, -745.413129, -729.689303, -713.910682, -698.084027, -682.216097, -666.313651, -650.383448, -634.432247, -618.466808, -602.49389, -586.520252, -570.552653, -554.597853, -538.662611, -522.753686, -506.877837, -491.158578, -475.716618, -460.317921, -444.961677, -429.647076, -414.373309, -399.139566, -383.945039, -368.788917, -353.670392, -338.588653, -323.542891, -308.532297, -293.556061, -278.613374, -263.703426, -248.825408, -234.274928, -219.896732, -205.532006, -191.17032, -176.801246, -162.414355, -147.999218, -133.545407, -119.042492, -104.480045, -89.8476376, -75.1348401, -60.3312241, -45.4263609, -30.4098216, -15.2711776, 0, 17.8774088, 35.3531331, 52.4868186, 69.3381108, 85.9666554, 102.432098, 118.794084, 135.11226, 151.446271, 167.855762, 184.400379, 201.139768, 218.133575, 235.441445, 253.123023, 271.237956, 289.833504, 308.967542, 328.727732, 349.171064, 370.354531, 392.335123, 415.169833, 438.915653, 463.629573, 489.368586, 516.189683, 544.149855, 573.306095, 603.279032, 634.722019, 667.822956, 702.769743, 739.962724, 779.483578, 821.357331, 865.708149, 912.6602, 962.33765, 1014.86467, 1069.16005, 1126.92018, 1188.50507, 1254.2747, 1324.58908, 1401.10296, 1482.5579, 1568.99021, 1660.4362, 1756.93218, 1855.8116, 1963.11714, 2082.02853, 2214.53306, 2360.18338, 2519.62528, 2693.50449, 2884.50244, 3090.76658, 3308.22368, 3533.93571, 3769.24859, 4012.37062, 4263.44597, 4525.89905, 4801.64861, 5098.36574, 5419.46052, 5747.25106, 6075.98628, 6385.50649, 6686.68594, 6978.76116, 7260.96866, 7532.54498, 7562.23654, 7637.39249, 7814.87186, 8151.53369, 8704.23704, 9951.09184, 11246.8723, 12467.9014, 12983.2468, 12870.02, 12416.6889, 11689.797, 10755.888, 9681.50566, 8533.19352, 7559.96123, 6605.00581, 5684.51861, 4814.69096, 4011.71421, 3291.7797, 2671.07878, 2088.65773, 1581.60906},
        {-1464.60418, -1448.00865, -1431.59044, -1415.3427, -1399.2586, -1383.33132, -1367.55401, -1351.91984, -1336.42199, -1321.05361, -1305.80788, -1290.67795, -1275.65701, -1260.7382, -1245.91471, -1231.17969, -1216.52632, -1201.94775, -1187.43717, -1172.98772, -1158.59259, -1144.24493, -1129.93791, -1115.6647, -1101.41847, -1087.19238, -1072.9796, -1058.7733, -1044.56664, -1030.35279, -1016.12491, -1001.87618, -987.59975, -973.288799, -958.627649, -943.770383, -928.874796, -913.944078, -898.981417, -883.990002, -868.973024, -853.93367, -838.87513, -823.800593, -808.713249, -793.616285, -778.512892, -763.406259, -748.299574, -733.196027, -718.098806, -703.132353, -688.175977, -673.230237, -658.295695, -643.372912, -628.462447, -613.564862, -598.680717, -583.810573, -568.95499, -554.114529, -539.289751, -524.481216, -509.689484, -494.915117, -480.158675, -465.397753, -450.610499, -435.844837, -421.102815, -406.386484, -391.697893, -377.039091, -362.412128, -347.819052, -333.261913, -318.742761, -304.263645, -289.826614, -275.433717, -261.087004, -246.788524, -232.540327, -218.818626, -205.378417, -191.968991, -178.577013, -165.189142, -151.792041, -138.372371, -124.916795, -111.411974, -97.8445697, -84.2012441, -70.4686589, -56.6334758, -42.6823566, -28.6019631, -14.378957, 0, 17.3015483, 34.2193405, 50.8106714, 67.1328354, 83.2431274, 99.1988419, 115.057274, 130.875717, 146.711467, 162.621818, 178.664064, 194.895501, 211.373423, 228.155125, 245.297901, 262.859046, 280.968778, 299.694809, 318.929488, 338.745736, 359.216474, 380.414624, 402.413107, 425.284842, 449.102753, 473.939759, 499.868781, 526.962742, 555.294561, 584.716915, 615.603726, 648.09399, 682.326701, 718.271633, 756.152392, 796.237707, 838.71734, 883.781051, 931.618603, 982.419756, 1036.39799, 1093.71214, 1154.5473, 1219.0886, 1287.52117, 1359.51749, 1435.90346, 1516.99236, 1603.09747, 1694.53206, 1790.1843, 1893.53437, 2008.86511, 2137.77539, 2276.38419, 2423.27085, 2577.01471, 2733.88909, 2895.30344, 3065.18285, 3247.21915, 3436.94269, 3627.65482, 3817.92145, 4015.22897, 4222.96911, 4459.90867, 4733.42549, 4994.5223, 5226.09079, 5368.71106, 5454.09585, 5479.78315, 5443.31098, 5342.21734, 4754.77945, 4202.61128, 3788.06601, 3613.49685, 3781.25697, 4961.8822, 6310.75469, 7686.72079, 8443.82634, 8640.73456, 8546.44224, 8209.09595, 7676.84223, 6997.82766, 6220.19878, 5477.91357, 4716.74936, 3958.50481, 3224.97858, 2537.96934, 1919.27574, 1390.69646, 852.457114, 367.749927},
        {-1373.5761, -1358.8443, -1344.26317, -1329.82561, -1315.52454, -1301.35284, -1287.30343, -1273.36919, -1259.54303, -1245.81785, -1232.18655, -1218.64204, -1205.1772, -1191.78494, -1178.45817, -1165.18977, -1151.97266, -1138.79973, -1125.66389, -1112.55802, -1099.47504, -1086.40784, -1073.34932, -1060.29239, -1047.22994, -1034.15488, -1021.0601, -1007.93851, -994.782998, -981.586475, -968.341837, -955.041985, -941.679819, -928.24824, -914.31163, -900.086161, -885.793156, -871.439422, -857.031766, -842.576994, -828.081912, -813.553328, -798.998047, -784.422876, -769.834622, -755.240091, -740.64609, -726.059425, -711.486903, -696.93533, -682.411513, -668.363353, -654.347, -640.359698, -626.398691, -612.461225, -598.544543, -584.645891, -570.762512, -556.891651, -543.030553, -529.176461, -515.326621, -501.478277, -487.628673, -473.775053, -459.914663, -445.927988, -431.698315, -417.463697, -403.228948, -388.998879, -374.778306, -360.572042, -346.384899, -332.221692, -318.087234, -303.986339, -289.923819, -275.904489, -261.933162, -248.014651, -234.153771, -220.355333, -207.23224, -194.472469, -181.756694, -169.069993, -156.397446, -143.724133, -131.035134, -118.31553, -105.550399, -92.7248221, -79.8238788, -66.8326491, -53.7362128, -40.5196498, -27.1680401, -13.6664636, 0, 16.8451038, 33.293533, 49.4028704, 65.2306982, 80.8345993, 96.272156, 111.600951, 126.878567, 142.162586, 157.510591, 172.980164, 188.628888, 204.514346, 220.69412, 237.225792, 254.166946, 271.727962, 289.98825, 308.659778, 327.832872, 347.597857, 368.045056, 389.264797, 411.347402, 434.383198, 458.46251, 483.675662, 510.112979, 537.864787, 567.094841, 597.79311, 630.027892, 663.867482, 698.83079, 735.260806, 773.647021, 814.222546, 857.220492, 902.873972, 951.416095, 1003.9063, 1059.49989, 1118.26829, 1180.28294, 1245.61528, 1312.02187, 1382.46775, 1457.60308, 1538.07802, 1624.54273, 1718.00101, 1818.31715, 1929.11705, 2051.17194, 2179.42477, 2311.43527, 2444.76316, 2571.72603, 2696.31714, 2828.24841, 2977.41886, 3132.91993, 3284.18792, 3427.50539, 3570.75552, 3716.49661, 3886.48345, 4088.25121, 4258.94795, 4375.53668, 4387.67093, 4303.82462, 4110.41401, 3793.8554, 3340.56505, 2006.30315, 690.806083, -436.845851, -1207.57236, -1452.29313, -180.708766, 1407.56251, 3129.64993, 4201.00234, 4676.91414, 4863.3799, 4803.15903, 4539.01091, 4113.69495, 3569.97054, 2980.82129, 2368.8479, 1751.63215, 1146.75581, 571.80066, 44.3484677, -418.018991, -857.938147, -1285.90712},
        {-1295.44714, -1284.44475, -1273.24877, -1261.86669, -1250.30599, -1238.57416, -1226.67866, -1214.627, -1202.42665, -1190.08509, -1177.6098, -1165.00827, -1152.28799, -1139.45643, -1126.52108, -1113.48941, -1100.36892, -1087.16708, -1073.89138, -1060.5493, -1047.14832, -1033.69593, -1020.1996, -1006.66683, -993.105088, -979.521864, -965.924639, -952.320896, -938.718116, -925.123784, -911.545381, -897.990391, -884.466296, -870.980578, -857.932871, -845.126334, -832.359413, -819.626864, -806.923444, -794.24391, -781.583018, -768.935525, -756.296188, -743.659763, -731.021007, -718.374676, -705.715528, -693.038319, -680.337805, -667.608744, -654.845891, -641.709563, -628.536208, -615.327831, -602.08644, -588.814041, -575.512641, -562.184247, -548.830864, -535.454501, -522.057163, -508.640858, -495.207591, -481.75937, -468.298202, -454.826092, -441.345048, -427.830424, -414.258216, -400.685396, -387.115701, -373.552863, -360.000618, -346.4627, -332.942844, -319.444783, -305.972252, -292.528985, -279.118718, -265.745184, -252.412117, -239.123253, -225.882326, -212.693069, -200.098419, -187.821169, -175.577629, -163.354036, -151.136626, -138.911636, -126.665301, -114.383859, -102.053544, -89.6605947, -77.1912457, -64.6317337, -51.9682952, -39.1871663, -26.2745832, -13.2167824, 0, 16.2259008, 32.0707647, 47.5903806, 62.8405371, 77.8770231, 92.7556272, 107.532138, 122.262345, 137.002036, 151.807, 166.733026, 181.835903, 197.171419, 212.795363, 228.763524, 245.13169, 262.11871, 279.803666, 297.873523, 316.419011, 335.53086, 355.2998, 375.816561, 397.171873, 419.456466, 442.761069, 467.176413, 492.793227, 519.702242, 548.099557, 577.931896, 609.25838, 642.138128, 676.250712, 711.845025, 749.271173, 788.702138, 830.310905, 874.270456, 920.753775, 970.473381, 1022.89852, 1078.0966, 1136.13506, 1197.08132, 1260.18397, 1326.53396, 1396.40344, 1470.06451, 1547.78931, 1628.53804, 1715.49822, 1813.33151, 1923.57519, 2041.38766, 2164.79129, 2291.80846, 2419.05471, 2546.27902, 2674.76512, 2803.12109, 2932.93776, 3077.72439, 3236.953, 3386.31019, 3512.40048, 3602.08648, 3642.03961, 3617.98486, 3516.18256, 3378.91816, 3127.80764, 2729.71805, 2151.51642, 1360.06979, -815.413546, -2985.99312, -4900.38721, -6307.3141, -6955.49207, -5430.83382, -3420.06692, -1170.25489, 273.350942, 979.636388, 1350.06553, 1436.89527, 1292.38252, 968.784185, 518.357177, 139.783921, -307.164391, -797.383413, -1305.7688, -1807.21621, -2276.62129, -2688.8797, -2974.60317, -3327.04718},
        {-1218.35287, -1208.13344, -1197.75858, -1187.23392, -1176.56509, -1165.75772, -1154.81744, -1143.74987, -1132.56065, -1121.25541, -1109.83977, -1098.31937, -1086.69984, -1074.9868, -1063.18589, -1051.30273, -1039.34295, -1027.31219, -1015.21607, -1003.06023, -990.850284, -978.591874, -966.290626, -953.952168, -941.582132, -929.186147, -916.769841, -904.338846, -891.898789, -879.455302, -867.014014, -854.580553, -842.160551, -829.759635, -817.695003, -805.809948, -793.944017, -782.092729, -770.251604, -758.416159, -746.581914, -734.744386, -722.899096, -711.041561, -699.1673, -687.271833, -675.350676, -663.39935, -651.413373, -639.388263, -627.31954, -614.907329, -602.448946, -589.946313, -577.40135, -564.815982, -552.19213, -539.531715, -526.836661, -514.10889, -501.350323, -488.562883, -475.748493, -462.909073, -450.046548, -437.162838, -424.259866, -411.298978, -398.242495, -385.176024, -372.104117, -359.031327, -345.962207, -332.901308, -319.853184, -306.822386, -293.813469, -280.830984, -267.879483, -254.96352, -242.087646, -229.256416, -216.47438, -203.746091, -191.650816, -179.893662, -168.172829, -156.47422, -144.783735, -133.087277, -121.370748, -109.62005, -97.8210857, -85.9597561, -74.0219636, -61.9936102, -49.8605979, -37.6088287, -25.2242047, -12.6926278, 0, 15.3638125, 30.4250209, 45.2331521, 59.837733, 74.2882905, 88.6343516, 102.925443, 117.211092, 131.540825, 145.964169, 160.530652, 175.289799, 190.291138, 205.584196, 221.2185, 237.243576, 253.817501, 271.005309, 288.610998, 306.707357, 325.367172, 344.663231, 364.668322, 385.455232, 407.096748, 429.665657, 453.234748, 477.876808, 503.664624, 530.520586, 558.723024, 588.389846, 619.638958, 652.789767, 687.859431, 724.81137, 763.70304, 804.591899, 847.535403, 892.591006, 939.002168, 987.888081, 1039.46546, 1093.95103, 1151.56149, 1211.97529, 1276.08199, 1344.23289, 1416.77927, 1494.07242, 1577.0968, 1664.79665, 1757.42602, 1854.95017, 1956.74481, 2062.45032, 2171.70711, 2286.77121, 2404.07293, 2517.18914, 2617.48789, 2713.36818, 2829.0993, 2965.72595, 3084.93994, 3166.49673, 3184.90334, 3118.55449, 2965.08912, 2711.26057, 2467.53446, 2077.91915, 1486.34449, 636.74031, -526.963538, -3668.66433, -6834.64834, -9679.02894, -11855.9195, -13019.4335, -11282.2827, -8867.58299, -6090.04842, -4259.50791, -3296.7758, -2708.03522, -2434.83207, -2418.71226, -2601.22167, -2923.90624, -3075.87125, -3359.37111, -3733.90412, -4158.96859, -4594.06282, -4998.68511, -5332.33377, -5490.4468, -5762.20713},
        {-1139.48068, -1126.70171, -1114.26581, -1102.15652, -1090.35734, -1078.85181, -1067.62343, -1056.65574, -1045.93224, -1035.43646, -1025.15192, -1015.06214, -1005.15064, -995.400931, -985.796544, -976.320995, -966.957804, -957.69049, -948.502573, -939.377573, -930.29901, -921.250404, -912.215274, -903.177141, -894.119524, -885.025943, -875.879917, -866.664968, -857.364613, -847.962374, -838.44177, -828.786321, -818.979547, -809.004967, -797.920786, -786.192645, -774.297305, -762.248314, -750.05922, -737.743571, -725.314916, -712.786802, -700.172778, -687.486391, -674.74119, -661.950723, -649.128537, -636.288182, -623.443204, -610.607152, -597.793575, -585.881805, -574.000838, -562.145451, -550.310425, -538.49054, -526.680574, -514.875308, -503.06952, -491.257991, -479.4355, -467.596826, -455.73675, -443.850049, -431.931505, -419.975897, -407.978003, -395.724741, -383.008016, -370.251311, -357.462881, -344.650981, -331.823864, -318.989785, -306.157, -293.333761, -280.528324, -267.748943, -255.003873, -242.301368, -229.649683, -217.057071, -204.531789, -192.082089, -180.487834, -169.355244, -158.281268, -147.249119, -136.242011, -125.24316, -114.235778, -103.203082, -92.1282851, -80.9946015, -69.7852457, -58.4834321, -47.0723751, -35.535289, -23.8553883, -12.0158871, 0, 14.3703579, 28.56094, 42.6118138, 56.5630469, 70.4547067, 84.3268608, 98.2195768, 112.172922, 126.226964, 140.421771, 154.797409, 169.393947, 184.251452, 199.409992, 214.909633, 230.790444, 247.124874, 263.957615, 281.255493, 299.065512, 317.43468, 336.410003, 356.038487, 376.36714, 397.442966, 419.312973, 442.024167, 465.623554, 490.158141, 515.127672, 541.327079, 568.967546, 598.260259, 630.311728, 664.885475, 701.506271, 740.116703, 780.659358, 823.076825, 867.311691, 910.793582, 956.742862, 1005.59379, 1057.78061, 1113.73758, 1171.0638, 1233.73746, 1302.90162, 1379.69932, 1465.2736, 1567.85749, 1672.83851, 1770.89827, 1857.81597, 1939.77835, 2018.29935, 2094.89294, 2174.52151, 2254.46685, 2328.24883, 2392.38719, 2447.27458, 2502.83026, 2557.23651, 2588.78796, 2584.90802, 2543.30606, 2454.07223, 2269.58135, 1963.54219, 1658.75407, 1156.89813, 380.074836, -749.615343, -2310.07194, -6445.9487, -10651.7024, -14488.544, -17517.6846, -19300.335, -17510.9933, -14855.3922, -11703.3339, -9556.79583, -8337.42759, -7495.13942, -6975.16804, -6722.75016, -6683.12252, -6801.52182, -6735.59458, -6869.4191, -7140.77584, -7487.44526, -7847.20782, -8157.84396, -8357.13416, -8428.71785, -8595.6289},
        {-1068.67232, -1056.21583, -1044.21776, -1032.65596, -1021.50823, -1010.75241, -1000.36632, -990.327787, -980.614635, -971.204689, -962.075775, -953.205716, -944.572339, -936.153468, -927.926929, -919.870546, -911.962144, -904.179549, -896.500585, -888.903078, -881.364853, -873.863734, -866.377547, -858.884117, -851.361268, -843.786826, -836.138616, -828.394463, -820.532192, -812.529628, -804.364596, -796.014921, -787.458428, -778.672942, -768.382833, -757.219019, -745.827481, -734.226723, -722.435246, -710.471553, -698.354146, -686.101526, -673.732196, -661.264659, -648.717415, -636.108967, -623.457817, -610.782467, -598.10142, -585.433177, -572.796241, -561.360839, -549.96878, -538.6136, -527.288833, -515.988015, -504.704679, -493.432362, -482.164597, -470.894921, -459.616867, -448.323971, -437.009768, -425.667793, -414.29158, -402.874664, -391.410581, -379.650984, -367.353332, -355.010024, -342.630275, -330.223301, -317.798317, -305.364537, -292.931177, -280.507452, -268.102577, -255.725768, -243.386238, -231.093205, -218.855882, -206.683485, -194.585229, -182.570329, -171.483803, -160.899385, -150.380763, -139.910028, -129.469272, -119.040588, -108.606067, -98.1478005, -87.6478804, -77.0883984, -66.4514464, -55.719116, -44.873499, -33.8966871, -22.7707722, -11.4778459, 0, 13.6904693, 27.2623608, 40.7505076, 54.1897431, 67.6149004, 81.0608127, 94.5623133, 108.154235, 121.871412, 135.748677, 149.820863, 164.122803, 178.689331, 193.55528, 208.755482, 224.324772, 240.347877, 256.866757, 273.80339, 291.203301, 309.112015, 327.575056, 346.63795, 366.346221, 386.745393, 407.880993, 429.798544, 452.543571, 476.161599, 500.079406, 525.188137, 551.718942, 579.902971, 611.025508, 644.790635, 680.62133, 718.432499, 758.139052, 799.655897, 842.89794, 885.196909, 929.837079, 977.238765, 1027.82228, 1082.00793, 1135.32869, 1194.31406, 1260.60619, 1335.84722, 1421.6793, 1529.96781, 1639.63664, 1737.80595, 1818.44615, 1889.51351, 1952.68493, 2009.63732, 2063.51179, 2114.18826, 2159.94927, 2206.23222, 2241.63563, 2264.93208, 2269.64007, 2240.37532, 2165.84008, 2048.94342, 1882.07084, 1605.51612, 1189.03907, 787.170346, 156.472612, -797.253474, -2168.20725, -4050.58807, -8903.25018, -13864.5743, -18437.596, -22125.3508, -24430.8745, -22769.6527, -20123.9709, -16891.5289, -14665.1725, -13369.5229, -12431.224, -11799.7961, -11424.7594, -11255.6339, -11241.94, -10983.8071, -10994.174, -11183.2284, -11461.1579, -11738.15, -11924.3923, -11930.0724, -11871.9801, -11839.2614},
        {-1003.72827, -994.401006, -985.297653, -976.404899, -967.70943, -959.197932, -950.857092, -942.673595, -934.634127, -926.725374, -918.934022, -911.246758, -903.650266, -896.131233, -888.676345, -881.272289, -873.905749, -866.563412, -859.231963, -851.89809, -844.548477, -837.169811, -829.748778, -822.272063, -814.726353, -807.098334, -799.374691, -791.54211, -783.587278, -775.496881, -767.257604, -758.856133, -750.279155, -741.513355, -731.729929, -721.340462, -710.766338, -700.020707, -689.11672, -678.067526, -666.886277, -655.586122, -644.180211, -632.681694, -621.103722, -609.459445, -597.762013, -586.024576, -574.260284, -562.482288, -550.703737, -539.69608, -528.697731, -517.705402, -506.715803, -495.725648, -484.731647, -473.730512, -462.718956, -451.693689, -440.651424, -429.588873, -418.502747, -407.389758, -396.246618, -385.070038, -373.85673, -362.427721, -350.608255, -338.757381, -326.883201, -314.993816, -303.097325, -291.20183, -279.315432, -267.446231, -255.602328, -243.791824, -232.02282, -220.303416, -208.641712, -197.045811, -185.523812, -174.083816, -163.5412, -153.483447, -143.488337, -133.537772, -123.613656, -113.697893, -103.772384, -93.8190331, -83.8197436, -73.7564183, -63.6109605, -53.365273, -43.001259, -32.5008217, -21.845864, -11.0182891, 0, 13.4518197, 26.7619843, 39.9658862, 53.098918, 66.1964723, 79.2939416, 92.4267184, 105.630195, 118.939765, 132.39082, 146.018753, 159.858955, 173.946821, 188.317743, 203.007112, 218.050322, 233.690463, 249.992601, 266.522332, 283.359555, 300.584171, 318.276078, 336.515176, 355.381365, 374.954543, 395.314611, 416.541468, 438.715014, 461.915147, 486.019018, 511.383618, 538.14967, 566.457898, 596.811604, 629.170222, 663.396503, 699.522399, 737.57986, 777.60084, 819.61729, 863.418509, 909.352952, 957.500046, 1007.93922, 1060.74989, 1108.15747, 1160.05892, 1218.49716, 1285.51512, 1363.15575, 1463.97259, 1566.65164, 1661.56911, 1743.24313, 1814.64816, 1874.96196, 1923.36229, 1952.49464, 1969.55361, 1988.85983, 2036.05311, 2076.00299, 2089.68503, 2061.59838, 1982.21276, 1839.25727, 1632.63347, 1353.22631, 947.288451, 386.319114, -150.250125, -930.842022, -2055.93396, -3626.00331, -5741.52747, -10926.454, -16251.9225, -21212.5429, -25302.925, -28017.6786, -26779.7768, -24536.5574, -21670.4746, -19689.9552, -18531.7559, -17663.4782, -17045.3365, -16637.5453, -16400.319, -16293.8721, -15893.5914, -15818.9271, -15948.0993, -16159.3283, -16330.8343, -16340.8376, -16067.5585, -15846.3563, -15497.7201},
        {-946.797945, -941.201345, -935.487081, -929.655532, -923.707072, -917.642078, -911.460926, -905.163993, -898.751654, -892.224286, -885.582265, -878.825967, -871.955768, -864.972045, -857.875173, -850.66553, -843.34349, -835.909431, -828.363728, -820.706758, -812.938896, -805.06052, -797.072005, -788.973727, -780.766063, -772.449388, -764.02408, -755.490514, -746.849066, -738.100112, -729.24403, -720.281194, -711.211981, -702.036768, -692.654321, -683.117959, -673.48222, -663.750778, -653.927307, -644.015481, -634.018972, -623.941455, -613.786603, -603.55809, -593.259588, -582.894773, -572.467318, -561.980895, -551.439179, -540.845843, -530.204561, -519.619415, -508.991493, -498.322292, -487.61331, -476.866043, -466.081988, -455.262642, -444.409502, -433.524065, -422.607828, -411.662288, -400.688942, -389.689287, -378.664819, -367.617035, -356.547434, -345.38383, -334.055812, -322.716834, -311.37317, -300.031094, -288.696879, -277.376798, -266.077124, -254.804131, -243.564093, -232.363281, -221.20797, -210.104434, -199.058944, -188.077776, -177.167201, -166.333493, -156.328428, -146.769848, -137.266632, -127.800863, -118.35462, -108.909984, -99.4490342, -89.9538522, -80.4065178, -70.7891114, -61.0837133, -51.272404, -41.3372637, -31.2603729, -21.0238117, -10.6096606, 0, 13.5090309, 26.7999574, 39.912497, 52.8863671, 65.7612854, 78.5769692, 91.3731361, 104.189504, 117.065789, 130.04171, 143.156984, 156.451329, 169.964461, 183.736099, 197.80596, 212.213761, 227.328319, 243.236365, 259.233227, 275.429144, 291.934353, 308.859095, 326.313607, 344.408128, 363.252896, 382.958151, 403.63413, 425.391074, 448.339219, 472.833743, 498.650136, 525.825154, 554.395555, 584.02927, 614.94747, 647.469676, 681.743291, 717.915721, 756.13437, 796.546641, 841.124701, 887.635829, 935.870412, 985.618834, 1036.67148, 1080.27848, 1126.90554, 1178.47811, 1236.92165, 1304.16159, 1390.52813, 1479.26954, 1566.63735, 1648.946, 1722.63834, 1784.09955, 1829.71483, 1843.40588, 1836.854, 1835.337, 1876.0452, 1910.6562, 1914.18923, 1865.88199, 1752.12025, 1556.00869, 1265.60546, 865.299484, 321.316942, -389.842119, -1071.94312, -2002.02584, -3279.6522, -5004.3841, -7275.78343, -12531.8383, -17949.0778, -23042.4574, -27326.9324, -30317.4583, -29583.7441, -27882.8222, -25563.3249, -23985.1548, -23096.6495, -22424.7843, -21938.9652, -21608.598, -21403.0888, -21291.8434, -20551.688, -20367.1114, -20531.3063, -20837.4654, -21078.7812, -21048.4466, -20539.6542, -20164.6455, -19457.8725},
        {-896.802199, -892.266408, -887.556889, -882.676308, -877.627333, -872.412629, -867.034864, -861.496704, -855.800815, -849.949864, -843.946518, -837.793443, -831.493306, -825.048773, -818.462511, -811.737187, -804.875466, -797.880016, -790.753503, -783.498594, -776.117955, -768.614253, -760.990154, -753.248325, -745.391432, -737.422143, -729.343123, -721.157039, -712.866558, -704.474347, -695.983071, -687.395397, -678.713993, -669.941524, -661.120067, -652.231754, -643.258246, -634.200928, -625.06119, -615.840418, -606.54, -597.161324, -587.705778, -578.174749, -568.569625, -558.891793, -549.142642, -539.323558, -529.435929, -519.481144, -509.460589, -499.296259, -489.070656, -478.786889, -468.448066, -458.057297, -447.617689, -437.132353, -426.604395, -416.036926, -405.433054, -394.795887, -384.128535, -373.434106, -362.715709, -351.976452, -341.219445, -330.404866, -319.493925, -308.57827, -297.663792, -286.756385, -275.861938, -264.986345, -254.135496, -243.315283, -232.531599, -221.790334, -211.09738, -200.45863, -189.879975, -179.367306, -168.926515, -158.563495, -149.033086, -139.950954, -130.923758, -121.933086, -112.960524, -103.98766, -94.9960795, -85.9673701, -76.8831185, -67.7249114, -58.4743356, -49.112978, -39.6224253, -29.9842643, -20.1800819, -10.1914649, 0, 13.4738804, 26.6797117, 39.6598261, 52.4565559, 65.1122333, 77.6691906, 90.1697601, 102.656274, 115.171065, 127.756464, 140.454805, 153.30842, 166.35964, 179.650798, 193.224227, 207.122258, 221.624983, 236.808698, 252.17795, 267.826019, 283.846185, 300.331729, 317.375932, 335.072073, 353.513434, 372.793295, 393.004936, 414.241638, 436.596682, 460.328017, 485.303876, 511.568137, 539.164681, 567.947223, 598.054724, 629.676856, 662.914548, 697.86873, 734.64033, 773.330277, 814.90825, 858.342026, 903.562561, 950.500809, 999.087728, 1043.85634, 1091.48501, 1143.25419, 1200.44432, 1264.33583, 1342.77928, 1422.45487, 1501.00499, 1575.41247, 1641.31344, 1694.94868, 1732.55894, 1741.27998, 1728.52692, 1711.64763, 1716.32784, 1706.92925, 1672.92886, 1597.69118, 1454.22346, 1220.28711, 873.087178, 390.240819, -248.594806, -1064.91647, -1858.0061, -2905.76223, -4306.60216, -6158.9432, -8561.20266, -13902.5616, -19417.9826, -24633.1921, -29073.9164, -32265.8819, -31853.5061, -30498.0302, -28531.4559, -27250.4472, -26605.1602, -26143.8085, -25836.6452, -25653.9237, -25565.8973, -25542.8192, -24207.1952, -23870.9589, -24184.3044, -24797.4259, -25360.5175, -25523.7735, -24937.3878, -24469.7625, -23487.3323},
        {-851.022377, -847.220165, -843.211925, -839.001789, -834.593885, -829.992343, -825.201293, -820.224864, -815.067187, -809.732391, -804.224607, -798.547962, -792.706588, -786.704614, -780.54617, -774.235385, -767.77639, -761.173313, -754.430285, -747.551436, -740.540894, -733.402791, -726.141255, -718.760416, -711.264405, -703.65735, -695.943381, -688.126629, -680.211223, -672.201293, -664.100968, -655.914378, -647.645652, -639.298922, -631.02294, -622.746482, -614.396585, -605.972687, -597.474223, -588.900629, -580.251343, -571.5258, -562.723437, -553.843691, -544.885997, -535.849793, -526.734513, -517.539596, -508.264477, -498.908593, -489.47138, -479.719483, -469.890176, -459.987942, -450.017264, -439.982626, -429.888509, -419.739397, -409.539772, -399.294118, -389.006918, -378.682653, -368.325807, -357.940864, -347.532305, -337.104614, -326.662273, -316.192819, -305.684194, -295.175833, -284.673318, -274.182229, -263.708149, -253.256658, -242.833339, -232.443772, -222.093539, -211.788222, -201.533402, -191.334661, -181.19758, -171.12774, -161.130723, -151.21211, -142.131235, -133.500949, -124.925044, -116.38464, -107.860859, -99.3348218, -90.7876481, -82.2004592, -73.5543758, -64.8305188, -56.0100088, -47.0739666, -38.003513, -28.7797688, -19.3838547, -9.79689152, 0, 13.4710037, 26.6212654, 39.4959693, 52.1402999, 64.5994415, 76.9185783, 89.1428948, 101.317575, 113.487804, 125.698765, 137.995644, 150.423624, 163.027889, 175.853624, 188.946014, 202.350242, 216.222973, 230.625316, 245.3503, 260.466998, 276.044482, 292.151826, 308.858102, 326.232383, 344.343741, 363.26125, 383.053982, 403.791009, 425.541405, 448.390927, 472.385846, 497.590228, 524.068141, 551.961252, 581.294828, 612.073444, 644.337888, 678.128945, 713.487405, 750.454054, 788.751721, 828.835923, 870.809655, 914.775914, 960.837698, 1007.38242, 1056.65755, 1109.19499, 1165.52663, 1226.18436, 1296.15515, 1366.07072, 1434.73236, 1499.83965, 1556.84273, 1602.20172, 1632.37671, 1639.53803, 1625.41049, 1596.39866, 1563.07461, 1506.36628, 1432.75939, 1331.14816, 1160.61897, 893.940473, 498.305024, -54.9644263, -774.098438, -1678.89326, -2575.90938, -3730.98431, -5237.72669, -7189.74515, -9680.64833, -14988.8986, -20477.0374, -25692.46, -30182.5615, -33494.7371, -33396.0095, -32401.0614, -30820.3048, -29872.2764, -29507.9223, -29299.8029, -29217.7834, -29231.729, -29311.5047, -29426.9759, -27588.7649, -27155.5433, -27647.2339, -28583.7595, -29485.043, -29871.0072, -29261.575, -28728.4274, -27521.1158},
        {-807.945011, -804.590296, -801.023677, -797.249715, -793.272976, -789.098023, -784.729419, -780.171727, -775.429512, -770.507337, -765.409765, -760.14136, -754.706685, -749.110305, -743.356782, -737.450681, -731.396564, -725.198996, -718.862539, -712.391758, -705.791216, -699.065476, -692.219102, -685.256658, -678.182708, -671.001814, -663.71854, -656.337451, -648.863108, -641.300077, -633.65292, -625.926202, -618.124485, -610.252333, -602.511639, -594.804151, -587.029246, -579.185084, -571.269825, -563.281629, -555.218655, -547.079063, -538.861014, -530.562667, -522.182181, -513.717718, -505.167436, -496.529496, -487.802057, -478.98328, -470.071323, -460.727272, -451.293669, -441.775981, -432.179674, -422.510217, -412.773076, -402.973718, -393.11761, -383.210219, -373.257012, -363.263456, -353.235018, -343.177165, -333.095364, -322.995082, -312.881786, -302.762012, -292.642271, -282.525824, -272.41807, -262.324406, -252.250229, -242.200938, -232.18193, -222.198603, -212.256355, -202.360583, -192.516685, -182.730058, -173.006101, -163.350211, -153.767786, -144.264224, -135.606278, -127.402655, -119.252909, -111.137729, -103.037806, -94.9338298, -86.8064911, -78.6364799, -70.4044864, -62.0912008, -53.6773133, -45.1435141, -36.4704934, -27.6389415, -18.6295485, -9.42300456, 0, 13.43669, 26.5084284, 39.2627002, 51.7469901, 64.0087831, 76.095564, 88.0548177, 99.934029, 111.780683, 123.642264, 135.566257, 147.600148, 159.79142, 172.187559, 184.83605, 197.784377, 211.052726, 224.684681, 238.789477, 253.408749, 268.584132, 284.357261, 300.76977, 317.863294, 335.679469, 354.259929, 373.646309, 393.880244, 415.003368, 436.886906, 459.805387, 483.851571, 509.118215, 536.067218, 564.606768, 594.546616, 625.868779, 658.555272, 692.588113, 727.949316, 763.257353, 800.272778, 839.244389, 880.420981, 924.051354, 971.971689, 1022.44655, 1075.32789, 1130.46766, 1187.71781, 1249.51562, 1309.96785, 1368.51244, 1423.42326, 1470.59769, 1507.00008, 1529.59483, 1536.22758, 1522.78114, 1484.17686, 1416.03225, 1315.17628, 1202.87846, 1073.32539, 874.813181, 576.702416, 139.663977, -469.194434, -1250.90283, -2224.51434, -3216.72363, -4468.53589, -6065.58986, -8093.52428, -10637.9779, -15818.126, -21177.6866, -26293.9144, -30744.0639, -34105.3898, -34302.746, -33667.3883, -32484.7429, -31883.8503, -31816.2062, -31881.8004, -32049.704, -32288.9879, -32568.7231, -32857.9807, -30751.7383, -30315.9514, -30983.9466, -32189.0503, -33364.5891, -33943.8894, -33360.2778, -32790.9371, -31438.7823},
        {-766.056631, -762.904483, -759.563627, -756.03783, -752.330858, -748.446476, -744.38845, -740.160547, -735.766532, -731.21017, -726.495229, -721.625474, -716.60467, -711.436585, -706.124983, -700.67363, -695.086293, -689.366738, -683.51873, -677.546036, -671.45242, -665.24165, -658.917491, -652.483709, -645.94407, -639.30234, -632.562284, -625.72767, -618.802262, -611.789827, -604.69413, -597.518937, -590.268015, -582.94513, -575.734864, -568.546773, -561.288235, -553.95715, -546.551416, -539.068929, -531.507589, -523.865294, -516.139941, -508.329428, -500.431654, -492.444516, -484.365913, -476.193742, -467.925902, -459.560291, -451.094806, -442.157812, -433.124751, -424.001533, -414.794066, -405.508258, -396.15002, -386.725259, -377.239884, -367.699805, -358.110929, -348.479166, -338.810425, -329.110615, -319.385643, -309.64142, -299.883853, -290.12677, -280.383807, -270.644542, -260.914373, -251.198693, -241.502898, -231.832385, -222.192548, -212.588783, -203.026486, -193.511051, -184.047876, -174.642355, -165.299883, -156.025856, -146.825671, -137.704721, -129.441622, -121.638896, -113.889772, -106.174553, -98.4735428, -90.7670439, -83.0353595, -75.2587929, -67.4176469, -59.4922249, -51.4628299, -43.3097649, -35.0133331, -26.5538376, -17.9115816, -9.06686798, 0, 13.3072284, 26.2250108, 38.8017921, 51.0860174, 63.1261315, 74.9705796, 86.6678067, 98.2662577, 109.814378, 121.360611, 132.953404, 144.641201, 156.472447, 168.495586, 180.759065, 193.311327, 206.044678, 218.985258, 232.534685, 246.707945, 261.520025, 276.985911, 293.12059, 309.939048, 327.456271, 345.687247, 364.646961, 384.3504, 404.81255, 425.680385, 447.471842, 470.312312, 494.327186, 520.260982, 547.929527, 576.983546, 607.362691, 639.006615, 671.854969, 705.847407, 739.027385, 773.807853, 810.499457, 849.412843, 890.858659, 938.739101, 988.875377, 1040.68025, 1093.56647, 1146.94679, 1201.72058, 1253.99705, 1303.03819, 1347.35906, 1383.94978, 1410.48516, 1424.63998, 1429.39619, 1415.91525, 1369.569, 1274.94751, 1139.56824, 992.483685, 831.295308, 600.312503, 268.306594, -204.430981, -851.327389, -1674.71686, -2694.52156, -3771.51955, -5109.26078, -6782.75574, -8867.01493, -11437.0488, -16417.5203, -21571.3744, -26511.2085, -30849.62, -34199.2064, -34665.2077, -34372.4835, -33579.6416, -33318.3764, -33541.2822, -33878.8337, -34299.731, -34772.6746, -35266.3644, -35749.5007, -33751.4564, -33447.27, -34258.2943, -35605.8825, -36911.3873, -37596.1616, -37081.5586, -36507.5883, -35119.8911},
        {-723.843768, -720.690405, -717.403263, -713.983876, -710.433779, -706.754507, -702.947594, -699.014576, -694.956987, -690.776362, -686.474234, -682.05214, -677.511614, -672.85419, -668.081404, -663.194789, -658.195881, -653.086215, -647.867324, -642.540744, -637.108009, -631.570655, -625.930215, -620.188225, -614.346219, -608.405732, -602.368299, -596.235454, -590.008732, -583.689668, -577.279796, -570.780651, -564.193768, -557.520682, -550.841314, -544.116356, -537.30556, -530.407915, -523.422413, -516.348045, -509.183801, -501.928673, -494.581651, -487.141727, -479.607891, -471.979134, -464.254447, -456.432821, -448.513247, -440.494715, -432.376218, -423.849288, -415.22704, -406.515128, -397.719208, -388.844936, -379.897969, -370.883962, -361.808572, -352.677453, -343.496262, -334.270656, -325.006289, -315.708819, -306.3839, -297.037189, -287.674342, -278.301415, -268.924452, -259.548287, -250.178549, -240.820868, -231.480874, -222.164197, -212.876468, -203.623316, -194.410371, -185.243264, -176.127624, -167.069082, -158.073267, -149.145809, -140.29234, -131.518487, -123.620671, -116.192493, -108.818052, -101.477315, -94.1502502, -86.816824, -79.4570043, -72.0507585, -64.5780541, -57.0188584, -49.353139, -41.5608633, -33.6219986, -25.5165125, -17.2243725, -8.72554579, 0, 13.0189083, 25.6548225, 37.9550186, 49.9667724, 61.7373599, 73.3140569, 84.7441394, 96.0748833, 107.353565, 118.627459, 129.943843, 141.349991, 152.893181, 164.620687, 176.579787, 188.817755, 201.129265, 213.525509, 226.625123, 240.421253, 254.907047, 270.075652, 285.920214, 302.433881, 319.609799, 337.441117, 355.920979, 375.042535, 394.79893, 414.635796, 435.294551, 456.932597, 479.707337, 504.538406, 531.202088, 559.271407, 588.675093, 619.341877, 651.200492, 684.179668, 716.664057, 750.596408, 786.207555, 823.728333, 863.389575, 908.799611, 955.967406, 1004.27942, 1053.12211, 1101.88195, 1151.62992, 1198.00907, 1239.00258, 1272.84276, 1298.27049, 1313.79834, 1317.93885, 1317.0914, 1300.08925, 1247.16183, 1139.56714, 985.751153, 810.772702, 612.13035, 340.623336, -31.5133421, -535.574867, -1200.24147, -2041.24941, -3081.65678, -4231.36782, -5644.00277, -7381.7884, -9506.95143, -12081.7186, -16814.3582, -21709.5451, -26417.9956, -30590.4262, -33877.5531, -34574.8862, -34591.8195, -34159.8722, -34209.0626, -34694.4207, -35279.9354, -35935.1888, -36629.7631, -37333.2405, -38015.2031, -36643.2604, -36644.5857, -37534.129, -38826.8401, -40037.669, -40681.5655, -40273.4795, -39728.6776, -38444.0017},
        {-673.679382, -670.093966, -666.520683, -662.954605, -659.390801, -655.824344, -652.250303, -648.66375, -645.059755, -641.433389, -637.779723, -634.093827, -630.370773, -626.605631, -622.793471, -618.929365, -615.008384, -611.025597, -606.976077, -602.854893, -598.657117, -594.377819, -590.01207, -585.55494, -581.001502, -576.346824, -571.585979, -566.714036, -561.726067, -556.617143, -551.382333, -546.01671, -540.515343, -534.873303, -528.806297, -522.454952, -515.963257, -509.335349, -502.575364, -495.687438, -488.675709, -481.544314, -474.297388, -466.939069, -459.473494, -451.904799, -444.23712, -436.474596, -428.621361, -420.681554, -412.65931, -404.588301, -396.44249, -388.225372, -379.940443, -371.591201, -363.181142, -354.713762, -346.192558, -337.621026, -329.002663, -320.340965, -311.639428, -302.90155, -294.130826, -285.330753, -276.504828, -267.609918, -258.604011, -249.586771, -240.564717, -231.544369, -222.532246, -213.534867, -204.558751, -195.610417, -186.696385, -177.823174, -168.997302, -160.22529, -151.513656, -142.86892, -134.297601, -125.806217, -118.224707, -111.130561, -104.091373, -97.0869405, -90.0970609, -83.101532, -76.0801515, -69.0127171, -61.8790265, -54.6588774, -47.3320675, -39.8783943, -32.2776557, -24.5096493, -16.5541727, -8.39102374, 0, 12.3324083, 24.3548809, 36.109451, 47.638152, 58.983017, 70.1860793, 81.2893722, 92.3349289, 103.364783, 114.420967, 125.545515, 136.780459, 148.167834, 159.749672, 171.568007, 183.664871, 195.692219, 207.636359, 220.421645, 234.006524, 248.349439, 263.408836, 279.143159, 295.510854, 312.470364, 329.980135, 347.998611, 366.484238, 385.39546, 403.802758, 422.878127, 442.846402, 463.932416, 486.994034, 511.939574, 538.508548, 566.735882, 596.656501, 628.305331, 661.717296, 698.955603, 737.409594, 776.717357, 816.516981, 856.446555, 891.15663, 926.519714, 963.42078, 1002.7448, 1045.37674, 1099.51294, 1149.79091, 1186.14508, 1203.91104, 1209.45174, 1204.17906, 1189.50488, 1174.80738, 1151.72162, 1103.19216, 1027.19124, 905.188349, 723.284303, 468.255788, 128.024685, -310.012789, -842.412343, -1477.61715, -2282.91332, -3292.30207, -4512.11948, -6008.28808, -7824.38549, -10003.9893, -12590.6771, -17033.5445, -21619.2717, -26040.0567, -29988.0977, -33155.5928, -34042.3568, -34327.8934, -34215.4222, -34512.6899, -35183.6241, -35940.266, -36759.1375, -37616.7605, -38489.6568, -39354.3484, -40226.8493, -41063.2252, -41801.216, -42378.5613, -42733.0009, -42802.2747, -42524.1223, -42099.0796, -41316.6513},
        {-633.134032, -629.561437, -626.064089, -622.634261, -619.26422, -615.946239, -612.672586, -609.435532, -606.227347, -603.040302, -599.866666, -596.698709, -593.528702, -590.348915, -587.151618, -583.929082, -580.673575, -577.377369, -574.032734, -570.631939, -567.167256, -563.630953, -560.015302, -556.312572, -552.515033, -548.614957, -544.604612, -540.476269, -536.222198, -531.83467, -527.305953, -522.62832, -517.794039, -512.795382, -507.191779, -501.201024, -495.046114, -488.733364, -482.269092, -475.659613, -468.911245, -462.030303, -455.023106, -447.895968, -440.655207, -433.307139, -425.858081, -418.314349, -410.682261, -402.968132, -395.178279, -387.491174, -379.737246, -371.91908, -364.03926, -356.100372, -348.104999, -340.055726, -331.955139, -323.805821, -315.610357, -307.371332, -299.09133, -290.772936, -282.418734, -274.03131, -265.613247, -257.099235, -248.425594, -239.734937, -231.03425, -222.330519, -213.63073, -204.941869, -196.270921, -187.624873, -179.01071, -170.43542, -161.905986, -153.429397, -145.012636, -136.662692, -128.386548, -120.191192, -112.931515, -106.172714, -99.4697992, -92.8022386, -86.1495027, -79.4910612, -72.8063839, -66.0749406, -59.2762011, -52.3896353, -45.3947129, -38.2709038, -30.9976777, -23.5545045, -15.9208539, -8.07619581, 0, 11.7350793, 23.2342279, 34.533034, 45.6670856, 56.6719707, 67.5832773, 78.4365934, 89.2675071, 100.111606, 111.004479, 121.981714, 133.078898, 144.33162, 155.775468, 167.446029, 179.378892, 191.359811, 203.388684, 216.066198, 229.374407, 243.295361, 257.811114, 272.903717, 288.555224, 304.747685, 321.463153, 338.683681, 356.391321, 374.568125, 392.479223, 411.086461, 430.586967, 451.17787, 473.537161, 497.621537, 523.259466, 550.503816, 579.407459, 610.023263, 642.404097, 679.669643, 717.872581, 756.465755, 794.902005, 832.634174, 862.327761, 891.919786, 922.559923, 955.397851, 991.583246, 1039.92027, 1084.54863, 1117.02016, 1132.88214, 1135.83913, 1125.93327, 1103.20666, 1070.82625, 1024.99913, 958.523486, 882.977567, 762.969482, 563.469618, 265.356837, -123.533967, -607.741162, -1166.58825, -1798.07641, -2592.66133, -3588.50117, -4839.57917, -6365.79385, -8199.82548, -10374.3543, -12922.0607, -17109.7834, -21427.5046, -25599.3649, -29349.5049, -32402.0652, -33410.4617, -33883.3765, -34003.8323, -34501.7097, -35343.5669, -36268.1047, -37253.2744, -38277.0269, -39317.3134, -40352.085, -42236.2126, -43491.1468, -44235.9598, -44589.7233, -44671.5093, -44600.3897, -44495.4364, -44113.7306, -43540.2346},
        {-600.947329, -597.736631, -594.595268, -591.515799, -588.490786, -585.512791, -582.574376, -579.668102, -576.78653, -573.922222, -571.06774, -568.215644, -565.358497, -562.488859, -559.599294, -556.682361, -553.730622, -550.736639, -547.692974, -544.592187, -541.426841, -538.189497, -534.872716, -531.46906, -527.97109, -524.371368, -520.662456, -516.836914, -512.887304, -508.806189, -504.586129, -500.219685, -495.69942, -491.017895, -485.756205, -480.121301, -474.325075, -468.373443, -462.272318, -456.027616, -449.64525, -443.131134, -436.491184, -429.731313, -422.857436, -415.875467, -408.79132, -401.61091, -394.340151, -386.984958, -379.551245, -372.190906, -364.760711, -357.26341, -349.701752, -342.078488, -334.396366, -326.658137, -318.866551, -311.024357, -303.134305, -295.199144, -287.221626, -279.204498, -271.150512, -263.062416, -254.942961, -246.727777, -238.354104, -229.963123, -221.561934, -213.157638, -204.757336, -196.368129, -187.997118, -179.651403, -171.338086, -163.064267, -154.837048, -146.663528, -138.55081, -130.505994, -122.53618, -114.64847, -107.715576, -101.293586, -94.9281844, -88.5983813, -82.2831867, -75.9616109, -69.6126638, -63.2153555, -56.7486963, -50.1916962, -43.5233653, -36.7227137, -29.7687516, -22.6404891, -15.3169362, -7.77710316, 0, 11.3305105, 22.4749012, 33.4633079, 44.3258661, 55.0927114, 65.7939795, 76.459806, 87.1203265, 97.8056767, 108.545992, 119.371408, 130.312061, 141.398086, 152.659618, 164.126794, 175.829749, 187.834614, 200.176669, 212.804627, 225.756339, 239.069654, 252.78242, 266.932486, 281.557703, 296.695918, 312.384981, 328.662741, 345.567048, 363.13575, 381.166713, 400.025763, 419.822745, 440.667502, 463.036964, 486.857432, 511.957319, 538.336344, 565.994225, 594.930682, 625.145434, 658.71456, 692.929482, 727.383674, 761.670612, 795.38377, 824.608671, 853.323727, 881.999402, 911.106158, 941.114457, 976.674275, 1008.96827, 1037.55224, 1060.31829, 1071.76171, 1067.9029, 1044.76224, 994.804369, 918.413557, 819.853054, 717.358971, 571.53771, 345.261775, 18.9352595, -397.331446, -907.063629, -1489.34357, -2141.35947, -2949.92392, -3951.15287, -5204.01803, -6718.26336, -8522.09321, -10643.7119, -13111.3239, -17071.7657, -21154.9513, -25109.4272, -28683.7397, -31626.435, -32705.2437, -33303.4051, -33589.815, -34249.8017, -35247.8399, -36333.2242, -37481.1886, -38666.9667, -39865.7924, -41052.8995, -42819.4935, -44118.4321, -45023.3676, -45607.9521, -45945.8379, -46110.6773, -46176.1226, -45820.2598, -45211.5649},
        {-573.713799, -571.017105, -568.349071, -565.704136, -563.076741, -560.461323, -557.852324, -555.244182, -552.631336, -550.008227, -547.369293, -544.708975, -542.021711, -539.301942, -536.544106, -533.742643, -530.891992, -527.986594, -525.020887, -521.989311, -518.886306, -515.70631, -512.443764, -509.093106, -505.648777, -502.105216, -498.456862, -494.698154, -490.823533, -486.827438, -482.704308, -478.448582, -474.054701, -469.517103, -464.531059, -459.246887, -453.818497, -448.250038, -442.545659, -436.709507, -430.74573, -424.658476, -418.451894, -412.130131, -405.697335, -399.157656, -392.515239, -385.774235, -378.938791, -372.013054, -365.001173, -357.939653, -350.799584, -343.584412, -336.297583, -328.942546, -321.522745, -314.041628, -306.502642, -298.909233, -291.264848, -283.572933, -275.836935, -268.0603, -260.246476, -252.398909, -244.521045, -236.560542, -228.466394, -220.357113, -212.23976, -204.121399, -196.009094, -187.909908, -179.830904, -171.779145, -163.761695, -155.785616, -147.857973, -139.985828, -132.176245, -124.436287, -116.773018, -109.1935, -102.585481, -96.4971598, -90.4661465, -84.4709248, -78.4899779, -72.5017894, -66.4848424, -60.4176204, -54.2786068, -48.0462848, -41.6991379, -35.2156493, -28.5743026, -21.7535809, -14.7319677, -7.48794625, 0, 11.1683141, 22.1578673, 32.9962657, 43.7111154, 54.3300227, 64.8805938, 75.3904347, 85.8871518, 96.3983511, 106.951639, 117.574621, 128.294904, 139.140095, 150.137798, 161.315621, 172.701169, 184.616126, 197.130109, 209.605549, 222.133071, 234.803295, 247.706845, 260.934344, 274.576414, 288.723678, 303.466759, 318.896279, 335.102861, 352.177128, 370.462505, 389.704118, 409.916748, 431.115179, 453.637957, 477.337982, 501.981815, 527.487109, 553.771515, 580.752685, 608.348273, 636.736938, 665.495886, 694.491704, 723.590976, 752.660287, 783.134822, 812.920418, 841.491509, 868.32253, 892.887918, 914.33157, 932.862449, 957.107484, 987.753929, 1009.27886, 1013.43741, 991.984674, 928.039572, 823.956205, 691.509994, 547.472167, 359.601754, 102.699898, -239.102066, -665.863317, -1184.89985, -1786.37804, -2473.1677, -3311.02051, -4330.11861, -5569.12788, -7048.44151, -8795.30497, -10836.9637, -13200.6632, -16947.4125, -20812.2525, -24563.9878, -27971.423, -30803.3626, -31921.07, -32605.9183, -33015.6583, -33805.6105, -34936.5708, -36162.0148, -37453.1046, -38781.0024, -40116.8703, -41431.8705, -42391.9119, -43513.8023, -44680.938, -45776.7154, -46684.531, -47287.781, -47469.8618, -47203.234, -46446.9931},
        {-545.055603, -542.676763, -540.309464, -537.949096, -535.591049, -533.230712, -530.863474, -528.484725, -526.089855, -523.674252, -521.233306, -518.762406, -516.256943, -513.712305, -511.123882, -508.487063, -505.797237, -503.049795, -500.240125, -497.363617, -494.41566, -491.391644, -488.286959, -485.096993, -481.817136, -478.442777, -474.969307, -471.392113, -467.706587, -463.908116, -459.992091, -455.953901, -451.788936, -447.492584, -442.827564, -437.910495, -432.860792, -427.681396, -422.375247, -416.945284, -411.394448, -405.725679, -399.941917, -394.046102, -388.041174, -381.930073, -375.71574, -369.401114, -362.989135, -356.482743, -349.884879, -343.159609, -336.349589, -329.458602, -322.490431, -315.448859, -308.337668, -301.160642, -293.921562, -286.624213, -279.272376, -271.869834, -264.420371, -256.927768, -249.39581, -241.828278, -234.228955, -226.548131, -218.73738, -210.91081, -203.075672, -195.239216, -187.408694, -179.591354, -171.794449, -164.025228, -156.290943, -148.598843, -140.956179, -133.370202, -125.848162, -118.39731, -111.024896, -103.738171, -97.4453751, -91.6843133, -85.9819606, -80.3163285, -74.6654287, -69.0072729, -63.3198728, -57.5812399, -51.7693861, -45.862323, -39.8380622, -33.6746154, -27.3499943, -20.8422105, -14.1292758, -7.1892017, 0, 11.6890108, 23.0547206, 34.1347803, 44.9668406, 55.5885524, 66.0375664, 76.3515335, 86.5681044, 96.72493, 106.859661, 117.009948, 127.213443, 137.507795, 147.930656, 158.519677, 169.312507, 180.31678, 191.565856, 203.165288, 215.146294, 227.540092, 240.377901, 253.690938, 267.510421, 281.867568, 296.793598, 312.319728, 328.477176, 345.297161, 362.722745, 380.905625, 399.903466, 419.77393, 441.178331, 463.872509, 487.451331, 511.791367, 536.769188, 562.261362, 588.144461, 614.132587, 640.314223, 666.597725, 692.891452, 719.103759, 746.720765, 773.678625, 799.491255, 823.672572, 845.736492, 865.835757, 882.064672, 899.483155, 917.842781, 926.057257, 917.906269, 887.169507, 827.588747, 732.990204, 597.24145, 413.454628, 176.96986, -109.197671, -451.178923, -870.602484, -1381.98327, -1993.23825, -2717.17176, -3590.78066, -4637.37715, -5875.96434, -7334.82647, -9038.76736, -11012.5909, -13281.1008, -16806.3395, -20441.5627, -23977.2644, -27203.9389, -29912.0802, -31066.2828, -31833.5404, -32358.3043, -33247.054, -34461.3545, -35770.2591, -37144.9585, -38556.6435, -39976.5047, -41375.7327, -41307.3732, -42234.2688, -43763.176, -45500.8507, -47054.0494, -48029.5281, -48034.0432, -48293.7834, -47490.6433},
        {-516.455553, -514.250244, -512.057642, -509.873314, -507.692826, -505.511744, -503.325634, -501.130061, -498.920592, -496.692792, -494.442227, -492.164464, -489.855069, -487.509607, -485.123644, -482.692746, -480.212479, -477.67841, -475.086103, -472.431125, -469.709043, -466.915421, -464.045826, -461.095824, -458.06098, -454.936861, -451.719033, -448.403061, -444.984512, -441.458951, -437.821944, -434.069058, -430.195858, -426.19791, -421.861724, -417.291792, -412.595116, -407.774048, -402.830938, -397.768135, -392.587991, -387.292854, -381.885076, -376.367007, -370.740996, -365.009395, -359.174553, -353.23882, -347.204547, -341.074085, -334.849782, -328.467863, -321.998239, -315.444692, -308.811007, -302.100967, -295.318356, -288.466957, -281.550555, -274.572933, -267.537874, -260.449163, -253.310583, -246.125917, -238.89895, -231.633465, -224.333246, -216.943682, -209.411564, -201.86111, -194.29989, -186.735472, -179.175426, -171.627321, -164.098726, -156.597211, -149.130344, -141.705695, -134.330833, -127.013327, -119.760747, -112.580661, -105.480638, -98.4682489, -92.4756217, -87.0285989, -81.6419103, -76.2931213, -70.9597974, -65.6195039, -60.2498064, -54.8282701, -49.3324607, -43.7399434, -38.0282837, -32.1750471, -26.157799, -19.9541048, -13.54153, -6.89763988, 0, 12.2993335, 24.1106363, 35.4834103, 46.4671574, 57.1113796, 67.4655788, 77.5792569, 87.5019159, 97.2830577, 106.972184, 116.618797, 126.272399, 135.982492, 145.798576, 155.770156, 165.946731, 175.894917, 185.595232, 196.189423, 207.623516, 219.843538, 232.795515, 246.425471, 260.679435, 275.50343, 290.843484, 306.645623, 322.855872, 339.420257, 355.652118, 372.362153, 389.686193, 407.760072, 427.595522, 448.890426, 471.10909, 494.124575, 517.809944, 542.038257, 566.682577, 591.960509, 617.295709, 642.493826, 667.360512, 691.701417, 714.451423, 736.504641, 757.884412, 778.614079, 798.716984, 821.950352, 840.040008, 850.312388, 850.278506, 837.826224, 810.674209, 766.541125, 714.61582, 638.539917, 509.442115, 294.616512, 17.6509361, -290.893771, -627.603664, -1035.06371, -1534.7446, -2151.65897, -2908.1958, -3813.75707, -4885.09095, -6125.44617, -7567.14718, -9231.61503, -11140.2708, -13314.5355, -16607.8984, -20001.6954, -23309.3303, -26344.2071, -28919.7299, -30112.2657, -30963.6133, -31603.0497, -32574.6348, -33843.4605, -35201.7927, -36623.3648, -38081.9101, -39551.1621, -41004.854, -40081.9331, -40867.8102, -42726.7205, -45022.8989, -47120.5807, -48384.0007, -48177.3942, -48999.8973, -48248.7344},
        {-488.567494, -486.38975, -484.24778, -482.13632, -480.050106, -477.983874, -475.932359, -473.890296, -471.852423, -469.813473, -467.768184, -465.71129, -463.637527, -461.541632, -459.418338, -457.262384, -455.068503, -452.831431, -450.545905, -448.20666, -445.808431, -443.345955, -440.813967, -438.207202, -435.520396, -432.748286, -429.885606, -426.927092, -423.86748, -420.701506, -417.423906, -414.029414, -410.512767, -406.8687, -402.84477, -398.564502, -394.15439, -389.617192, -384.955664, -380.172564, -375.270649, -370.252675, -365.1214, -359.879581, -354.529976, -349.07534, -343.518431, -337.862007, -332.108825, -326.26164, -320.323212, -314.270419, -308.132457, -301.912643, -295.614297, -289.240735, -282.795276, -276.281238, -269.70194, -263.060698, -256.360831, -249.605658, -242.798496, -235.942663, -229.041477, -222.098257, -215.11632, -208.026964, -200.763214, -193.476927, -186.176089, -178.868688, -171.56271, -164.266142, -156.986972, -149.733186, -142.51277, -135.333713, -128.204, -121.131619, -114.124556, -107.190799, -100.338334, -93.5751478, -87.8600295, -82.705566, -77.6129169, -72.5592136, -67.5215876, -62.4771707, -57.4030942, -52.2764899, -47.0744891, -41.7742235, -36.3528247, -30.7874241, -25.0551534, -19.1331441, -12.9985277, -6.62843581, 0, 12.4309143, 24.3135483, 35.7009447, 46.6461462, 57.2021955, 67.4221354, 77.3590085, 87.0658576, 96.5957254, 106.001655, 115.336688, 124.653868, 134.006238, 143.44684, 153.028717, 162.804912, 172.170908, 181.085811, 191.09, 202.095612, 214.014785, 226.759655, 240.24236, 254.375037, 269.069823, 284.238855, 299.794271, 315.648206, 331.712799, 347.074727, 362.774254, 378.971158, 395.825213, 414.298451, 434.149519, 454.92313, 476.538381, 498.914371, 521.970197, 545.62496, 571.522115, 597.331598, 622.635131, 647.014438, 670.051244, 687.105598, 703.036316, 718.480539, 734.075408, 750.458065, 774.699485, 793.139402, 798.18881, 785.758192, 758.902825, 717.470121, 661.30749, 606.416351, 532.818743, 402.914153, 175.263342, -117.200052, -434.541936, -770.280458, -1174.61672, -1671.73343, -2294.80533, -3070.34639, -3991.89974, -5071.6587, -6308.1264, -7729.2925, -9352.08918, -11193.4487, -13270.3031, -16314.197, -19448.7972, -22512.3829, -25343.2334, -27779.6277, -29022.216, -29971.9926, -30740.5063, -31795.2193, -33106.7981, -34497.6119, -35945.7418, -37429.2691, -38926.2752, -40414.8414, -39283.8817, -40084.2922, -42105.0901, -44635.2925, -46963.9168, -48379.9801, -48172.4996, -49213.5653, -48665.6033},
        {-461.473317, -459.138735, -456.894623, -454.73338, -452.647403, -450.629092, -448.670844, -446.765059, -444.904134, -443.080469, -441.286461, -439.51451, -437.757014, -436.006371, -434.25498, -432.495239, -430.719548, -428.920303, -427.089905, -425.22075, -423.305239, -421.335769, -419.304739, -417.204548, -415.027593, -412.766274, -410.412989, -407.960137, -405.400115, -402.725323, -399.928159, -397.001021, -393.936308, -390.726419, -386.982894, -382.896571, -378.662867, -374.286539, -369.772346, -365.125045, -360.349394, -355.450153, -350.432078, -345.299929, -340.058462, -334.712437, -329.26661, -323.725741, -318.094588, -312.377907, -306.580459, -300.827456, -295.000589, -289.102006, -283.133853, -277.098276, -270.997424, -264.833442, -258.608478, -252.324677, -245.984188, -239.589157, -233.141731, -226.644056, -220.09828, -213.506548, -206.871009, -200.0956, -193.086614, -186.04875, -178.990521, -171.92044, -164.84702, -157.778775, -150.724218, -143.691862, -136.690221, -129.727808, -122.813137, -115.95472, -109.161072, -102.440705, -95.8021334, -89.2538699, -83.7844925, -78.8918681, -74.0620837, -69.2718478, -64.4978687, -59.7168547, -54.9055143, -50.0405559, -45.0986878, -40.0566184, -34.8910561, -29.5787092, -24.0962862, -18.4204954, -12.5280451, -6.39564389, 0, 11.5333141, 22.6822278, 33.4853598, 43.9813288, 54.2087535, 64.2062526, 74.0124447, 83.6659486, 93.2053829, 102.669366, 112.096518, 121.525455, 130.994798, 140.543165, 150.209174, 160.031445, 169.868954, 179.734656, 190.073503, 200.88562, 212.171129, 223.930156, 236.162824, 248.869257, 262.049579, 275.703914, 289.832387, 304.43512, 319.512239, 334.900371, 350.823085, 367.329553, 384.468947, 402.33036, 420.943003, 440.325446, 460.514884, 481.548516, 503.463536, 526.297144, 554.283286, 581.985135, 608.618786, 633.400331, 655.545864, 666.624848, 675.411664, 683.034062, 690.619793, 699.296607, 716.986666, 729.719031, 733.758151, 725.665093, 702.606519, 661.477191, 599.171872, 517.591117, 407.486216, 254.147591, 43.1598896, -221.508882, -526.584512, -875.326013, -1299.00703, -1816.04202, -2449.90632, -3220.32646, -4128.47251, -5186.01114, -6407.44219, -7801.16326, -9377.86027, -11148.2192, -13122.9259, -15889.3237, -18737.277, -21533.3071, -24143.9359, -26435.6849, -27753.7902, -28833.5826, -29764.9918, -30920.0814, -32277.1031, -33696.7527, -35163.1596, -36660.4532, -38172.763, -39684.2185, -39516.9306, -40609.3208, -42484.254, -44664.5952, -46673.2094, -48032.9616, -48266.7168, -48815.9955, -48711.8645},
        {-437.574084, -435.272926, -433.082772, -430.995238, -429.001936, -427.094482, -425.26449, -423.503575, -421.80335, -420.15543, -418.55143, -416.982964, -415.441646, -413.91909, -412.406911, -410.896724, -409.380142, -407.848781, -406.294254, -404.708176, -403.082161, -401.407823, -399.676778, -397.880639, -396.01102, -394.059537, -392.017803, -389.877433, -387.630041, -385.267242, -382.78065, -380.161879, -377.402544, -374.494259, -371.011308, -367.162748, -363.162653, -359.016181, -354.728489, -350.304735, -345.750076, -341.069671, -336.268676, -331.35225, -326.325549, -321.193731, -315.961955, -310.635377, -305.219154, -299.718446, -294.138408, -288.631012, -283.051419, -277.401605, -271.683544, -265.899212, -260.050583, -254.139633, -248.168337, -242.138668, -236.052604, -229.912117, -223.719184, -217.475779, -211.183877, -204.845454, -198.462483, -191.933773, -185.160609, -178.357739, -171.533828, -164.697538, -157.857531, -151.022472, -144.201022, -137.401845, -130.633604, -123.904961, -117.224579, -110.601121, -104.043251, -97.5596307, -91.1589233, -84.8497918, -79.6405675, -75.0190523, -70.4610331, -65.9427315, -61.4403693, -56.9301681, -52.3883497, -47.7911358, -43.1147482, -38.3354086, -33.4293387, -28.3727602, -23.1418949, -17.7129645, -12.0621908, -6.16579535, 0, 10.8735779, 21.4764586, 31.8373473, 41.9849491, 51.9479694, 61.7551132, 71.4350857, 81.0165921, 90.5283377, 99.9990276, 109.457367, 118.932061, 128.451815, 138.045334, 147.741323, 157.568488, 167.634686, 177.979929, 188.453888, 199.102229, 209.970619, 221.104725, 232.550211, 244.352745, 256.557994, 269.211623, 282.359299, 296.046689, 310.319458, 325.408338, 341.106073, 357.402811, 374.288698, 391.707901, 409.673557, 428.211065, 447.324367, 467.017404, 487.294116, 508.158446, 533.130532, 557.62797, 580.966752, 602.462866, 621.432303, 632.21862, 640.353347, 646.395582, 650.904423, 654.438968, 662.311822, 664.518733, 660.925537, 649.19741, 622.506509, 576.042264, 504.994105, 401.719709, 264.883839, 96.2404472, -97.4057206, -336.046326, -629.006938, -987.947631, -1428.48416, -1964.41657, -2612.84711, -3388.43194, -4293.71909, -5338.10562, -6538.1637, -7895.01146, -9415.56226, -11106.7295, -12975.4266, -15500.6169, -18099.1514, -20659.9307, -23071.8559, -25223.8277, -26570.737, -27724.8347, -28761.0259, -29991.8429, -31398.1302, -32861.7205, -34368.4089, -35903.99, -37454.2589, -39005.0104, -39880.3102, -41268.4061, -42945.2337, -44686.7287, -46268.8268, -47467.4635, -48058.5747, -48417.2429, -48053.1288},
        {-417.028928, -415.008631, -413.076575, -411.225555, -409.448366, -407.737805, -406.086668, -404.48775, -402.933847, -401.417755, -399.93227, -398.470187, -397.024303, -395.587413, -394.152313, -392.711799, -391.258667, -389.785712, -388.285731, -386.751518, -385.175871, -383.551585, -381.871455, -380.128277, -378.314848, -376.423963, -374.448418, -372.381009, -370.214531, -367.941781, -365.555554, -363.048646, -360.413853, -357.64397, -354.398421, -350.843699, -347.150306, -343.321857, -339.361966, -335.274247, -331.062315, -326.729785, -322.280271, -317.717387, -313.044747, -308.265967, -303.38466, -298.404441, -293.328925, -288.161725, -282.906457, -277.600241, -272.21246, -266.746, -261.20375, -255.588598, -249.903433, -244.151141, -238.334612, -232.456733, -226.520393, -220.52848, -214.483881, -208.389484, -202.248179, -196.062852, -189.836393, -183.486, -176.930934, -170.350805, -163.754056, -157.149131, -150.544472, -143.948521, -137.369723, -130.816519, -124.297352, -117.820666, -111.394903, -105.028506, -98.7299178, -92.5075814, -86.3699395, -80.325435, -75.3918888, -71.0518238, -66.775631, -62.5389969, -58.3176082, -54.0871514, -49.8233132, -45.5017801, -41.0982387, -36.5883756, -31.9478774, -27.1524307, -22.1777221, -16.9994381, -11.5932654, -5.93489047, 0, 10.705791, 21.1500442, 31.3603525, 41.3643087, 51.1895056, 60.863536, 70.4139928, 79.8684689, 89.2545571, 98.5998501, 107.931941, 117.278422, 126.666887, 136.124928, 145.680138, 155.36011, 165.172138, 175.140914, 185.339947, 195.792478, 206.521752, 217.55101, 228.903497, 240.602455, 252.671128, 265.132757, 278.010587, 291.327861, 305.10782, 319.422831, 334.229003, 349.534843, 365.348857, 382.309775, 400.110991, 418.277843, 436.62977, 454.986213, 473.166612, 490.990407, 508.010264, 524.39359, 540.01202, 554.737187, 568.440728, 584.314478, 598.079821, 608.778341, 615.451622, 617.141249, 613.922313, 602.539717, 584.177502, 557.502855, 516.044389, 455.637832, 372.118914, 257.69243, 112.650253, -58.7547267, -243.270766, -469.933256, -757.388589, -1120.93843, -1570.217, -2117.45997, -2780.31942, -3572.43529, -4487.58728, -5530.78917, -6704.42647, -8016.54556, -9473.07, -11079.9233, -12843.0291, -15176.999, -17576.8965, -19946.473, -22189.4801, -24209.6694, -25527.0836, -26684.9889, -27751.5835, -29023.7377, -30479.1335, -32000.1733, -33569.3035, -35168.9702, -36781.6199, -38389.6986, -39977.5224, -41561.6847, -43087.6802, -44501.0034, -45747.149, -46771.6115, -47519.8856, -48121.3157, -46871.7893},
        {-396.733209, -395.049481, -393.428185, -391.863328, -390.348918, -388.878961, -387.447465, -386.048438, -384.675885, -383.323815, -381.986235, -380.657151, -379.330572, -378.000503, -376.660954, -375.30593, -373.929438, -372.525487, -371.088083, -369.611234, -368.088946, -366.515227, -364.884084, -363.189524, -361.425555, -359.586183, -357.665417, -355.657262, -353.555727, -351.354818, -349.048542, -346.630908, -344.095921, -341.43759, -338.385487, -335.0714, -331.630291, -328.06475, -324.377365, -320.570725, -316.647419, -312.610035, -308.461161, -304.203387, -299.839301, -295.371491, -290.802546, -286.135055, -281.371606, -276.514788, -271.567189, -266.509186, -261.366061, -256.140884, -250.836725, -245.456655, -240.003743, -234.48106, -228.891676, -223.238661, -217.525084, -211.754016, -205.928528, -200.051688, -194.126568, -188.156238, -182.143767, -176.007481, -169.667743, -163.302399, -156.920014, -150.529151, -144.138373, -137.756245, -131.39133, -125.052191, -118.747393, -112.4855, -106.275074, -100.12468, -94.0428807, -88.0382407, -82.1193234, -76.2946924, -71.5987095, -67.505463, -63.4767121, -59.4877314, -55.5137954, -51.5301787, -47.5121559, -43.4350016, -39.2739904, -35.0043968, -30.6014954, -26.0405608, -21.2968676, -16.3456903, -11.1623036, -5.72198194, 0, 10.8216009, 21.3338573, 31.5673352, 41.5526003, 51.3202188, 60.9007564, 70.324779, 79.6228527, 88.8255434, 97.9634168, 107.067039, 116.166976, 125.293793, 134.478057, 143.750334, 153.141189, 162.503278, 171.841753, 181.589594, 191.739241, 202.283139, 213.213729, 224.523455, 236.204758, 248.250081, 260.651866, 273.402557, 286.494595, 299.920423, 313.563247, 327.5648, 341.950295, 356.744945, 372.807845, 389.747267, 406.949117, 424.188443, 441.240294, 457.879719, 473.881768, 487.836026, 501.063798, 513.572071, 525.367833, 536.45807, 552.664191, 566.725157, 577.194349, 582.625149, 581.570939, 573.113168, 554.631735, 525.125423, 482.884173, 424.750711, 348.217534, 250.777137, 128.904781, -18.6586129, -192.062411, -380.859544, -611.829652, -903.103428, -1269.32511, -1719.23134, -2264.27043, -2921.74372, -3704.61697, -4604.39493, -5624.7219, -6769.88913, -8042.79446, -9446.85826, -10985.5009, -12662.1427, -14831.1299, -17057.2255, -19256.1186, -21343.4981, -23235.053, -24500.5668, -25632.2375, -26693.9993, -27974.7729, -29449.4352, -30999.119, -32603.3881, -34241.807, -35893.9396, -37539.3503, -38806.8596, -40310.4902, -41921.4629, -43510.9989, -44950.3192, -46110.6447, -46863.1966, -47554.8335, -46391.7571},
        {-377.094882, -375.664048, -374.281869, -372.942983, -371.642033, -370.373659, -369.132501, -367.913199, -366.710394, -365.518727, -364.332838, -363.147367, -361.956956, -360.756244, -359.539872, -358.302481, -357.038711, -355.743202, -354.410596, -353.035532, -351.612651, -350.136594, -348.602, -347.003512, -345.335768, -343.59341, -341.771079, -339.863413, -337.865055, -335.770644, -333.574822, -331.272228, -328.857503, -326.325288, -323.429303, -320.289716, -317.02959, -313.651386, -310.15756, -306.550572, -302.832882, -299.006946, -295.075226, -291.040178, -286.904263, -282.669939, -278.339664, -273.915898, -269.401099, -264.797726, -260.108238, -255.336981, -250.484486, -245.553171, -240.545453, -235.46375, -230.310481, -225.088062, -219.798913, -214.445451, -209.030093, -203.555258, -198.023363, -192.436827, -186.798068, -181.109502, -175.373549, -169.49073, -163.364015, -157.205973, -151.025627, -144.832001, -138.634119, -132.441003, -126.261676, -120.105163, -113.980487, -107.896671, -101.862738, -95.8877122, -89.9806162, -84.1504737, -78.4063081, -72.7571426, -68.2612044, -64.3808464, -60.5658109, -56.7910725, -53.0316061, -49.2623862, -45.4583875, -41.5945847, -37.6459526, -33.5874658, -29.3940989, -25.0408268, -20.5026239, -15.7544652, -10.7713251, -5.52817852, 0, 11.0171597, 21.6636735, 31.9736873, 41.9813469, 51.7207981, 61.2261868, 70.5316588, 79.6713598, 88.6794357, 97.5900323, 106.437295, 115.255371, 124.078405, 132.940542, 141.87593, 150.918713, 159.826078, 168.592606, 177.878897, 187.659749, 197.90996, 208.604327, 219.717647, 231.224717, 243.100336, 255.3193, 267.856408, 280.686455, 293.784241, 306.988626, 320.460186, 334.214501, 348.267147, 363.080506, 378.446753, 394.038919, 409.738542, 425.427158, 440.986307, 456.297524, 472.106208, 487.167122, 501.192788, 513.895728, 524.988463, 536.808179, 545.786567, 550.979983, 551.444785, 546.237326, 537.141925, 517.1528, 481.155963, 426.127827, 353.312711, 262.038733, 151.634009, 24.4281912, -123.934306, -301.083471, -504.750409, -752.199652, -1054.64127, -1423.11001, -1868.34351, -2401.21581, -3036.95805, -3787.57388, -4649.09083, -5626.57341, -6741.14841, -7979.34698, -9340.67365, -10824.6329, -12430.7293, -14452.4262, -16521.7795, -18564.8041, -20507.5148, -22275.9265, -23482.8953, -24576.3671, -25616.5677, -26881.7888, -28345.9382, -29892.0629, -31498.6138, -33144.0416, -34806.7974, -36465.3319, -36721.6607, -37963.2945, -39824.2773, -41938.653, -43940.4656, -45463.759, -46142.5771, -46743.6725, -46589.0881},
        {-361.453756, -359.913943, -358.438598, -357.021901, -355.65803, -354.341167, -353.06549, -351.825179, -350.614415, -349.427377, -348.258245, -347.101199, -345.950417, -344.800081, -343.644371, -342.477464, -341.293543, -340.086786, -338.851373, -337.581483, -336.271298, -334.914996, -333.506758, -332.040762, -330.511189, -328.912219, -327.238032, -325.482807, -323.640724, -321.705962, -319.672702, -317.535124, -315.287407, -312.923731, -310.17307, -307.167786, -304.043427, -300.802778, -297.448626, -293.983756, -290.410956, -286.733011, -282.952708, -279.072832, -275.096171, -271.025509, -266.863633, -262.613331, -258.277386, -253.858587, -249.359719, -244.831531, -240.227807, -235.550293, -230.800736, -225.980882, -221.092477, -216.137268, -211.117001, -206.033423, -200.888281, -195.68332, -190.420287, -185.100929, -179.726992, -174.300222, -168.822366, -163.177379, -157.252046, -151.291047, -145.303764, -139.299581, -133.287879, -127.278041, -121.279449, -115.301486, -109.353533, -103.444973, -97.5851893, -91.783563, -86.0494769, -80.3923133, -74.8214546, -69.3462831, -65.0467626, -61.3744665, -57.7681006, -54.2023048, -50.651719, -47.0909832, -43.4947372, -39.8376209, -36.0942743, -32.2393374, -28.2474499, -24.0932518, -19.7513831, -15.1964837, -10.4031934, -5.3461522, 0, 10.7815877, 21.2246582, 31.3596049, 41.2168216, 50.8267017, 60.219639, 69.426027, 78.4762594, 87.4007296, 96.2298315, 104.993958, 113.723504, 122.448862, 131.200426, 140.00859, 148.903747, 157.781428, 166.65276, 175.853185, 185.384195, 195.247287, 205.443953, 215.97569, 226.84399, 238.050349, 249.596261, 261.48322, 273.712721, 286.286259, 299.394423, 312.780277, 326.388587, 340.164117, 353.8823, 367.572569, 381.30951, 395.088689, 408.905672, 422.756023, 436.63531, 452.900298, 468.466725, 482.868181, 495.638258, 506.310547, 516.259337, 522.717347, 524.757992, 521.454689, 511.880855, 497.803918, 472.310601, 429.876002, 367.46616, 287.132795, 188.644259, 71.7689045, -58.8282221, -209.388346, -391.494539, -608.611644, -870.831695, -1184.14101, -1556.70452, -2000.37871, -2525.32558, -3145.22422, -3871.14848, -4701.27615, -5641.07973, -6719.86786, -7915.60985, -9225.52729, -10646.8417, -12176.7747, -14060.4663, -15985.24, -17886.3377, -19699.0014, -21358.4732, -22518.8268, -23583.9177, -24609.4883, -25855.1313, -27295.8984, -28821.9041, -30412.447, -32046.8259, -33704.3394, -35364.2865, -35386.4649, -36581.0446, -38525.7043, -40798.1228, -42975.979, -44636.9516, -45358.7194, -45970.3541, -46623.0292},
        {-350.461653, -348.392172, -346.436031, -344.585767, -342.833919, -341.173024, -339.595618, -338.094241, -336.661429, -335.289721, -333.971653, -332.699763, -331.46659, -330.264669, -329.08654, -327.92474, -326.771805, -325.620275, -324.462685, -323.291575, -322.099482, -320.878942, -319.622494, -318.322676, -316.972024, -315.563077, -314.088372, -312.540447, -310.911839, -309.195085, -307.382725, -305.467294, -303.44133, -301.297372, -298.693744, -295.797119, -292.778188, -289.640335, -286.386943, -283.021396, -279.547077, -275.96737, -272.285658, -268.505324, -264.629752, -260.662326, -256.606428, -252.465443, -248.242754, -243.941744, -239.565796, -235.2122, -230.788398, -226.295738, -221.735567, -217.109233, -212.418084, -207.663467, -202.846731, -197.969224, -193.032292, -188.037284, -182.985548, -177.87843, -172.71728, -167.503445, -162.238272, -156.797492, -151.05985, -145.285773, -139.48475, -133.666272, -127.839831, -122.014918, -116.201023, -110.407639, -104.644255, -98.920363, -93.245454, -87.6290189, -82.0805488, -76.6095348, -71.2254678, -65.9378389, -61.8426357, -58.383248, -54.9900067, -51.6371439, -48.2988918, -44.9494826, -41.5631483, -38.1141211, -34.5766332, -30.9249167, -27.1332037, -23.1757264, -19.0267169, -14.6604074, -10.05103, -5.1728168, 0, 10.0140728, 19.8361513, 29.4840259, 38.9754871, 48.3283254, 57.5603313, 66.6892953, 75.7330078, 84.7092593, 93.6358404, 102.530541, 111.411153, 120.295465, 129.201269, 138.146355, 147.148513, 156.492532, 166.234343, 175.787816, 185.227954, 194.629762, 204.068243, 213.618402, 223.355243, 233.353771, 243.688988, 254.435901, 265.669513, 277.464827, 290.757399, 304.446147, 318.347912, 332.279532, 345.244526, 357.466395, 369.385522, 381.062744, 392.558896, 403.934813, 415.251331, 428.814163, 441.756043, 453.698591, 464.263428, 473.072173, 485.202333, 493.45567, 496.089833, 491.362472, 477.531234, 452.13622, 415.029631, 365.389963, 302.002395, 222.84908, 126.272715, 10.6159958, -122.446619, -278.661416, -467.409329, -692.197972, -962.142887, -1282.97509, -1662.17126, -2110.16598, -2636.0361, -3252.82415, -3970.63514, -4785.03321, -5699.80766, -6737.35145, -7879.98807, -9125.06706, -10469.938, -11911.9503, -13662.4202, -15451.239, -17222.2646, -18919.355, -20486.368, -21619.8774, -22675.8813, -23704.2167, -24935.548, -26348.1345, -27844.8904, -29407.8422, -31019.0164, -32660.4393, -34314.1375, -35282.76, -36751.5902, -38518.4636, -40381.2157, -42137.6822, -43585.6984, -44523.1, -45315.3681, -46200.3479},
        {-341.290358, -338.59135, -336.060571, -333.688735, -331.466554, -329.384742, -327.434013, -325.60508, -323.888657, -322.275458, -320.756195, -319.321583, -317.962336, -316.669165, -315.432786, -314.243912, -313.093256, -311.971532, -310.869453, -309.777733, -308.687086, -307.588224, -306.471863, -305.328714, -304.149492, -302.924911, -301.645683, -300.302523, -298.886143, -297.387258, -295.796581, -294.104826, -292.302705, -290.380934, -287.924749, -285.136132, -282.221935, -279.186029, -276.032288, -272.764583, -269.386786, -265.902769, -262.316404, -258.631563, -254.852118, -250.981941, -247.024904, -242.984879, -238.865738, -234.671353, -230.405596, -226.197085, -221.922242, -217.582235, -213.17823, -208.711397, -204.182902, -199.593914, -194.945599, -190.239127, -185.475664, -180.656379, -175.782439, -170.855011, -165.875265, -160.844366, -155.763484, -150.505872, -144.947856, -139.354416, -133.735011, -128.099101, -122.456146, -116.815605, -111.18694, -105.579608, -100.003071, -94.4667878, -88.9802184, -83.5528228, -78.1940607, -72.9133919, -67.7202764, -62.6241739, -58.7322311, -55.481976, -52.2977434, -49.153371, -46.0226966, -42.8795579, -39.6977929, -36.4512391, -33.1137345, -29.6591169, -26.0612238, -22.2938933, -18.330963, -14.1462708, -9.71365432, -5.00695147, 0, 9.27817866, 18.5061114, 27.6893474, 36.8334359, 45.9439262, 55.0263675, 64.0863091, 73.1293001, 82.1608899, 91.1866277, 100.212063, 109.242744, 118.284221, 127.342043, 136.421759, 145.528919, 155.296496, 165.819672, 175.684821, 185.03194, 194.001027, 202.732078, 211.365092, 220.040066, 228.896997, 238.075883, 247.716721, 257.95951, 268.944245, 282.221083, 296.002806, 310.006363, 323.948704, 336.345196, 347.513579, 358.09202, 368.157942, 377.788772, 387.061935, 396.054855, 406.583882, 416.458279, 425.415249, 433.191991, 439.525707, 452.986263, 462.270029, 464.90604, 458.423329, 440.350934, 404.204997, 356.432087, 301.461655, 239.967457, 164.955373, 72.8740056, -39.8280462, -176.065474, -340.17109, -537.1723, -768.596343, -1044.5151, -1371.82143, -1757.9103, -2211.02752, -2739.02834, -3354.2476, -4065.70195, -4865.98286, -5756.97283, -6751.2514, -7838.35828, -9016.47312, -10283.7756, -11638.4453, -13265.2186, -14929.0793, -16581.5679, -18174.2249, -19658.5908, -20770.5072, -21821.0127, -22854.0901, -24071.3891, -25454.4827, -26919.2494, -28450.547, -30033.2331, -31652.1655, -33292.2021, -35380.191, -37161.0688, -38696.6286, -40048.6634, -41278.9665, -42449.3309, -43621.5498, -44632.1762, -45527.6433},
        {-330.401283, -327.37474, -324.541316, -321.891116, -319.414244, -317.100807, -314.940909, -312.924655, -311.04215, -309.283499, -307.638807, -306.09818, -304.651721, -303.289537, -302.001733, -300.778413, -299.609682, -298.485646, -297.396409, -296.332077, -295.282754, -294.238546, -293.189558, -292.125895, -291.037661, -289.914962, -288.747903, -287.526588, -286.241124, -284.881614, -283.438165, -281.900881, -280.259866, -278.505227, -276.218092, -273.601656, -270.864031, -268.008593, -265.03872, -261.957788, -258.769175, -255.476257, -252.082413, -248.591019, -245.005451, -241.329088, -237.565307, -233.717483, -229.788995, -225.78322, -221.703534, -217.63692, -213.501336, -209.298349, -205.029523, -200.696423, -196.300613, -191.843658, -187.327123, -182.752573, -178.121572, -173.435684, -168.696476, -163.905511, -159.064355, -154.174571, -149.237725, -144.137279, -138.75953, -133.349621, -127.916773, -122.470207, -117.019143, -111.572803, -106.140406, -100.731175, -95.3543298, -90.0190913, -84.7346803, -79.5103177, -74.3552244, -69.2786212, -64.289729, -59.3977685, -55.7107286, -52.6653349, -49.6851071, -46.7436087, -43.8144035, -40.8710552, -37.8871274, -34.8361841, -31.6917887, -28.4275051, -25.0168969, -21.4335279, -17.6509617, -13.6427621, -9.38249285, -4.84371757, 0, 9.45717392, 18.8158351, 28.0839451, 37.2694656, 46.3803582, 55.4245846, 64.4101063, 73.3448849, 82.2368821, 91.0940596, 99.9243788, 108.735802, 117.536289, 126.333804, 135.136307, 143.951759, 153.219103, 163.007869, 172.351183, 181.349359, 190.102711, 198.711554, 207.276204, 215.896973, 224.674176, 233.708129, 243.099145, 252.947539, 263.353625, 275.374988, 287.803673, 300.452814, 313.135545, 325.190582, 336.668269, 347.745456, 358.377602, 368.520166, 378.128605, 387.158377, 397.155006, 405.999953, 413.337575, 418.812232, 422.068282, 428.947975, 431.348305, 427.364158, 415.09042, 392.621975, 355.353013, 307.379969, 253.730143, 194.869102, 121.948875, 30.3030811, -84.7346626, -229.727342, -407.013689, -616.863412, -857.858614, -1141.20077, -1474.58325, -1865.14616, -2319.09219, -2843.0543, -3447.48347, -4140.00251, -4914.23497, -5771.72319, -6724.05677, -7761.18503, -8881.17236, -10082.0832, -11361.9819, -12878.6765, -14430.5519, -15975.7367, -17472.3595, -18878.5489, -19962.1642, -20998.4493, -22027.0768, -23233.8022, -24600.4736, -26048.0709, -27561.4855, -29125.6088, -30725.3321, -32345.5468, -34654.7719, -36468.5316, -37915.643, -39124.9231, -40225.1888, -41345.2571, -42613.9452, -43769.966, -44537.2317},
        {-319.712611, -316.457678, -313.409907, -310.559234, -307.895594, -305.408923, -303.089157, -300.92623, -298.910078, -297.030637, -295.277841, -293.641627, -292.111929, -290.678683, -289.331825, -288.06129, -286.857013, -285.70893, -284.606976, -283.541086, -282.501197, -281.477243, -280.459159, -279.436882, -278.400347, -277.339488, -276.244242, -275.104543, -273.910328, -272.651532, -271.318089, -269.899936, -268.387008, -266.76924, -264.648022, -262.215735, -259.669429, -257.01165, -254.244941, -251.371847, -248.394913, -245.316682, -242.1397, -238.86651, -235.499658, -232.041688, -228.495144, -224.862571, -221.146513, -217.349514, -213.47412, -209.541079, -205.534336, -201.456041, -197.308344, -193.093395, -188.813344, -184.47034, -180.066534, -175.604075, -171.085113, -166.511798, -161.88628, -157.210708, -152.487233, -147.718004, -142.905171, -137.946507, -132.74229, -127.50994, -122.258374, -116.996507, -111.733257, -106.477538, -101.238268, -96.0243624, -90.8447373, -85.708309, -80.6239938, -75.6007077, -70.6473669, -65.7728878, -60.9861863, -56.2961788, -52.8082156, -49.9599339, -47.1757924, -44.4291257, -41.6932687, -38.9415562, -36.1473229, -33.2839036, -30.3246331, -27.2428463, -24.0118778, -20.6050624, -16.9957349, -13.1572302, -9.06288294, -4.68602795, 0, 9.90104658, 19.5998281, 29.1111246, 38.449716, 47.6303822, 56.6679033, 65.5770591, 74.3726298, 83.0693952, 91.6821354, 100.22563, 108.71466, 117.164004, 125.588443, 134.002756, 142.421724, 150.921558, 159.525815, 168.110321, 176.70302, 185.331856, 194.024773, 202.809715, 211.714626, 220.76745, 229.996131, 239.428613, 249.09284, 259.016756, 269.424233, 280.075447, 290.939563, 301.985745, 313.788476, 326.014262, 338.168194, 350.037841, 361.410773, 372.074561, 381.816776, 391.906195, 400.198378, 406.191096, 409.382116, 409.269207, 406.943992, 399.911924, 387.272306, 368.124446, 341.567648, 307.004002, 262.859959, 212.716121, 155.990286, 83.6012003, -9.71648985, -129.228138, -284.154341, -474.906289, -697.580357, -947.801197, -1237.93196, -1576.57283, -1970.61898, -2424.0766, -2942.27795, -3533.24095, -4202.99414, -4947.71881, -5769.1664, -6680.68679, -7670.64864, -8736.78858, -9876.84323, -11088.5492, -12505.1523, -13955.0028, -15401.9598, -16809.8827, -18142.631, -19195.3166, -20213.0443, -21231.9935, -22427.5858, -23780.8839, -25214.9335, -26713.6982, -28261.1419, -29841.2285, -31437.9216, -33612.1079, -35350.7807, -36769.4901, -37983.7862, -39109.2194, -40261.3397, -41555.6975, -42821.5842, -43382.4601},
        {-309.752308, -306.319341, -303.101831, -300.089759, -297.273108, -294.641859, -292.185994, -289.895495, -287.760345, -285.770524, -283.916015, -282.1868, -280.572861, -279.064178, -277.650736, -276.322514, -275.069496, -273.881662, -272.748996, -271.661478, -270.609092, -269.581817, -268.569637, -267.562534, -266.550489, -265.523484, -264.471501, -263.384521, -262.252528, -261.065503, -259.813427, -258.486282, -257.074051, -255.566715, -253.595247, -251.336684, -248.97238, -246.503966, -243.933076, -241.261342, -238.490396, -235.621871, -232.657399, -229.598613, -226.447145, -223.204627, -219.872692, -216.452973, -212.947101, -209.35671, -205.683431, -201.87852, -197.995078, -194.03583, -190.003502, -185.900816, -181.730499, -177.495273, -173.197865, -168.840998, -164.427397, -159.959787, -155.440892, -150.873437, -146.260145, -141.603743, -136.906954, -132.081336, -127.040638, -121.975606, -116.894876, -111.807081, -106.720856, -101.644837, -96.5876558, -91.5579487, -86.5643496, -81.615493, -76.7200135, -71.8865453, -67.1237232, -62.4401813, -57.8445544, -53.3454767, -50.0458863, -47.3842487, -44.7858767, -42.2238925, -39.6714184, -37.1015766, -34.4874892, -31.8022786, -29.0190669, -26.1109763, -23.0511291, -19.8126474, -16.3686535, -12.6922696, -8.75661791, -4.53482064, 0, 10.350417, 20.3923933, 30.1479375, 39.6390581, 48.8877635, 57.9160624, 66.7459632, 75.3994744, 83.8986046, 92.2653622, 100.521756, 108.689794, 116.791485, 124.848837, 132.88386, 140.918561, 148.648409, 156.048766, 163.88024, 172.094866, 180.644681, 189.48172, 198.55802, 207.825615, 217.236543, 226.742838, 236.296537, 245.849675, 255.354288, 264.181409, 273.077112, 282.167733, 291.579609, 303.127203, 316.09279, 329.308475, 342.394159, 354.96974, 366.655118, 377.07019, 387.201143, 394.885763, 399.476634, 400.326337, 396.787456, 384.774745, 367.938071, 346.489474, 320.640993, 290.604667, 260.561066, 221.903275, 175.339662, 118.785692, 44.4546614, -52.8799768, -178.444769, -342.443146, -543.993659, -776.785535, -1034.45309, -1329.99976, -1672.93422, -2069.85444, -2522.42637, -3034.57984, -3611.53622, -4257.56017, -4972.18074, -5757.60568, -6630.15539, -7575.75385, -8591.72371, -9675.3876, -10824.0682, -12148.9045, -13504.9487, -14861.0693, -16186.1347, -17449.0136, -18468.1347, -19463.0994, -20467.2503, -21648.965, -22988.1349, -24408.3867, -25892.4248, -27422.9539, -28982.6785, -30554.303, -32368.1122, -33970.5578, -35409.2269, -36731.7066, -37985.5838, -39218.4455, -40477.8787, -41817.8007, -42160.3638},
        {-301.815533, -298.169012, -294.749543, -291.546948, -288.551054, -285.751685, -283.138665, -280.70182, -278.430973, -276.315949, -274.346574, -272.512672, -270.804067, -269.210584, -267.722048, -266.328283, -265.019115, -263.784367, -262.613865, -261.497432, -260.424895, -259.386077, -258.370804, -257.368899, -256.370187, -255.364494, -254.341643, -253.29146, -252.203769, -251.068395, -249.875163, -248.613896, -247.27442, -245.84656, -243.967327, -241.810374, -239.553593, -237.198258, -234.745643, -232.197022, -229.553669, -226.816859, -223.987865, -221.067962, -218.058424, -214.960524, -211.775538, -208.504738, -205.1494, -201.710797, -198.190203, -194.522447, -190.776689, -186.955644, -183.062025, -179.098549, -175.067928, -170.972878, -166.816113, -162.600349, -158.328298, -154.002676, -149.626198, -145.201577, -140.731529, -136.218768, -131.666008, -126.984184, -122.086433, -117.16476, -112.227829, -107.284305, -102.342852, -97.4121345, -92.5008163, -87.6175619, -82.7710356, -77.9699018, -73.2228246, -68.5384685, -63.9254976, -59.3925764, -54.948369, -50.6015398, -47.4644221, -44.9706315, -42.5405183, -40.1469306, -37.7627165, -35.3607242, -32.9138016, -30.394797, -27.7765584, -25.031934, -22.1337718, -19.05492, -15.7682266, -12.2465399, -8.46270776, -4.38957843, 0, 9.99839399, 19.7419315, 29.2482436, 38.5349616, 47.6197167, 56.5201401, 65.2538629, 73.8385163, 82.2917316, 90.63114, 98.8743726, 107.039061, 115.142835, 123.203328, 131.238169, 139.264991, 147.036254, 154.53171, 162.368776, 170.508263, 178.910979, 187.537734, 196.349336, 205.306595, 214.370321, 223.501321, 232.660406, 241.808384, 250.906066, 259.427741, 267.999127, 276.726989, 285.71809, 296.819863, 309.268738, 321.836969, 334.10912, 345.669755, 356.103438, 364.994734, 373.052441, 378.394732, 380.386212, 378.391486, 371.77516, 354.562831, 332.792865, 307.16462, 278.377455, 247.130728, 220.622077, 185.110235, 136.572125, 71.3315561, -13.5794365, -121.446796, -255.556467, -420.959843, -618.776179, -848.198783, -1106.41082, -1403.72264, -1749.99118, -2151.01147, -2605.6959, -3116.11612, -3684.08369, -4311.60284, -5001.63138, -5756.58775, -6594.60355, -7499.96659, -8469.65608, -9500.65126, -10589.9314, -11829.7493, -13097.9922, -14367.8208, -15612.396, -16804.8784, -17782.7737, -18745.3348, -19723.8602, -20879.6775, -22192.6157, -23588.3252, -25049.3442, -26558.2106, -28097.4625, -29649.6379, -30969.4484, -32434.4181, -33973.8384, -35517.0012, -36993.198, -38331.7206, -39461.8604, -40813.5098, -41124.6269},
        {-294.951689, -291.099592, -287.485268, -284.098385, -280.928612, -277.965615, -275.199063, -272.618623, -270.213964, -267.974754, -265.890659, -263.951349, -262.14649, -260.465752, -258.898801, -257.435305, -256.064933, -254.777352, -253.56223, -252.409235, -251.308034, -250.248297, -249.21969, -248.211881, -247.214539, -246.21733, -245.209924, -244.181987, -243.123188, -242.023195, -240.871675, -239.658296, -238.372726, -237.004634, -235.192162, -233.108135, -230.929601, -228.657635, -226.293314, -223.837713, -221.291907, -218.656972, -215.933982, -213.124015, -210.228145, -207.247447, -204.182997, -201.035872, -197.807145, -194.497893, -191.109191, -187.57511, -183.965182, -180.281936, -176.5279, -172.705602, -168.817569, -164.86633, -160.854412, -156.784345, -152.658655, -148.479871, -144.25052, -139.973132, -135.650233, -131.284352, -126.878017, -122.337269, -117.570467, -112.779134, -107.972053, -103.158009, -98.3457819, -93.5441563, -88.7619146, -84.0078397, -79.2907144, -74.6193216, -70.002444, -65.4488647, -60.9673663, -56.5667319, -52.2557441, -48.0431859, -45.0537893, -42.71489, -40.4404111, -38.2029205, -35.9749866, -33.7291773, -31.438061, -29.0742057, -26.6101796, -24.0185508, -21.2718876, -18.342758, -15.2037302, -11.8273723, -8.18625257, -4.25293908, 0, 9.49668366, 18.8224688, 27.9884806, 37.0058443, 45.8856852, 54.6391286, 63.2772996, 71.8113236, 80.2523258, 88.6114314, 96.8997656, 105.128454, 113.308621, 121.451393, 129.567894, 137.669251, 145.64445, 153.487168, 161.484796, 169.622284, 177.884586, 186.256656, 194.723445, 203.269908, 211.880996, 220.541664, 229.236863, 237.951546, 246.670668, 255.168177, 263.717397, 272.366582, 281.163987, 291.691439, 303.230405, 314.653399, 325.548601, 335.504195, 344.10836, 350.949278, 356.501029, 359.196276, 358.449872, 353.67667, 344.291525, 323.531223, 298.533202, 270.256831, 239.661482, 207.706524, 183.815004, 150.138123, 97.7763289, 21.5780455, -75.9561907, -195.761486, -338.772946, -503.557048, -693.955852, -916.395379, -1172.76846, -1470.07895, -1818.41292, -2222.71995, -2679.24608, -3188.23241, -3748.19776, -4358.9367, -5026.55894, -5753.602, -6559.63458, -7427.54283, -8353.96884, -9335.55472, -10368.9426, -11529.1699, -12714.8845, -13903.1297, -15070.9488, -16195.385, -17130.903, -18058.8438, -19008.7842, -20135.5266, -21419.0796, -22787.2747, -24222.7293, -25708.0611, -27225.8874, -28758.8259, -29542.071, -30865.8466, -32526.2983, -34319.5719, -36041.8129, -37489.1672, -38457.7803, -39788.5892, -40124.8977},
        {-289.008112, -284.955944, -281.151926, -277.585552, -274.246316, -271.123712, -268.207234, -265.486376, -262.950631, -260.589494, -258.392459, -256.349018, -254.448667, -252.680899, -251.035208, -249.501088, -248.068032, -246.725536, -245.463092, -244.270194, -243.136337, -242.051015, -241.00372, -239.983948, -238.981192, -237.984946, -236.984703, -235.969959, -234.930206, -233.854938, -232.73365, -231.555835, -230.310988, -228.988602, -227.22338, -225.189676, -223.066103, -220.85367, -218.553385, -216.166254, -213.693286, -211.135487, -208.493866, -205.769429, -202.963185, -200.076141, -197.109305, -194.063683, -190.940283, -187.740114, -184.464182, -181.057375, -177.578036, -174.028391, -170.410663, -166.727076, -162.979855, -159.171224, -155.303407, -151.378627, -147.39911, -143.367078, -139.284757, -135.154371, -130.978142, -126.758297, -122.497058, -118.092912, -113.446835, -108.775003, -104.086366, -99.3898731, -94.6944727, -90.0091141, -85.3427463, -80.7043182, -76.1027791, -71.5470779, -67.0461636, -62.6089853, -58.2444921, -53.9616328, -49.7693567, -45.6766128, -42.8219737, -40.6258165, -38.4949412, -36.4016382, -34.3181982, -32.2169118, -30.0700695, -27.8499618, -25.5288795, -23.079113, -20.472953, -17.6826899, -14.6806145, -11.4390172, -7.9301886, -4.12641935, 0, 8.97874449, 17.8751324, 26.6934949, 35.4381632, 44.1134686, 52.7237422, 61.2733153, 69.7665192, 78.2076849, 86.6011437, 94.9512269, 103.262266, 111.538591, 119.784535, 128.004427, 136.202601, 144.415101, 152.650791, 160.842263, 169.000646, 177.137066, 185.262651, 193.388529, 201.525827, 209.685671, 217.879191, 226.117513, 234.411764, 242.773073, 251.328175, 259.930199, 268.55559, 277.180792, 286.978768, 297.327704, 307.286713, 316.473284, 324.504906, 330.999067, 335.573257, 338.537581, 338.606114, 335.260835, 327.983721, 316.256747, 293.53765, 266.838708, 237.147961, 205.453444, 172.743195, 149.599564, 115.689896, 58.0441705, -29.5859884, -139.725594, -271.061168, -422.279233, -585.639589, -767.7161, -982.093604, -1235.33723, -1531.28526, -1880.2186, -2286.33635, -2743.53229, -3250.4305, -3802.75616, -4398.38176, -5045.80925, -5747.52796, -6523.83379, -7356.67003, -8242.36221, -9177.23585, -10157.6165, -11242.2693, -12349.4702, -13459.9348, -14554.3787, -15613.5175, -16507.9176, -17401.8765, -18323.3168, -19420.212, -20672.9643, -22011.9545, -23420.0859, -24880.2615, -26375.3848, -27888.3588, -28183.7617, -29382.3132, -31156.5875, -33179.1593, -35122.603, -36659.4931, -37462.4041, -38749.5705, -39154.3807},
        {-283.726184, -279.463698, -275.461963, -271.710178, -268.197543, -264.913259, -261.846525, -258.98654, -256.322506, -253.84362, -251.539084, -249.398097, -247.409859, -245.563569, -243.848428, -242.253635, -240.76839, -239.381893, -238.083343, -236.861941, -235.706886, -234.607378, -233.552617, -232.531803, -231.534135, -230.548813, -229.565037, -228.572007, -227.558922, -226.514983, -225.429389, -224.29134, -223.090036, -221.814676, -220.088596, -218.091623, -216.007982, -213.838747, -211.58499, -209.247783, -206.8282, -204.327313, -201.746196, -199.08592, -196.34756, -193.532187, -190.640874, -187.674695, -184.634722, -181.522028, -178.337686, -175.044612, -171.682862, -168.254337, -164.760936, -161.20456, -157.587108, -153.910481, -150.176578, -146.387299, -142.544545, -138.650214, -134.706208, -130.714427, -126.676769, -122.595135, -118.471426, -114.196904, -109.665491, -105.107265, -100.531298, -95.9466613, -91.3624276, -86.787669, -82.2314574, -77.7028649, -73.2109635, -68.7648253, -64.3735223, -60.0461266, -55.7917103, -51.6193453, -47.5381037, -43.5570577, -40.8269694, -38.762052, -36.7627962, -34.8012239, -32.8493572, -30.879218, -28.8628284, -26.7722104, -24.579386, -22.2563771, -19.7752059, -17.1078942, -14.2264642, -11.1029377, -7.70933685, -4.01768362, 0, 9.02323148, 17.9444341, 26.7691271, 35.5028297, 44.1510613, 52.7193409, 61.213188, 69.6381217, 77.9996613, 86.303326, 94.5546351, 102.759108, 110.922263, 119.049621, 127.1467, 135.21902, 143.139616, 150.89508, 158.790599, 166.803302, 174.910319, 183.088779, 191.315813, 199.568551, 207.824121, 216.059655, 224.252282, 232.379131, 240.417333, 248.282068, 256.03513, 263.672234, 271.189094, 279.811332, 288.919708, 297.56701, 305.379978, 311.985356, 317.009885, 320.080309, 321.319223, 319.706604, 314.77218, 306.04568, 293.056832, 271.349692, 245.436077, 215.842132, 183.094002, 147.717833, 118.372606, 77.5114979, 12.9824767, -80.9164571, -196.718482, -332.86931, -487.814649, -653.362433, -836.104736, -1049.87485, -1300.9868, -1593.78371, -1939.76984, -2343.32141, -2796.73658, -3297.85753, -3841.50609, -4424.74139, -5055.69723, -5736.24293, -6485.11289, -7284.71676, -8131.08598, -9020.25198, -9948.24618, -10955.2818, -11982.163, -13013.8758, -14035.4064, -15031.7405, -15893.3297, -16762.7176, -17665.4052, -18739.5845, -19966.1495, -21279.0393, -22661.4955, -24096.7598, -25568.0738, -27058.6791, -27231.0654, -28392.489, -30183.3046, -32243.8671, -34214.5313, -35735.6521, -36447.5843, -37705.4466, -38220.5701},
        {-278.9773, -274.494794, -270.287924, -266.34548, -262.656247, -259.209013, -255.992566, -252.995693, -250.207182, -247.61582, -245.210393, -242.979691, -240.912499, -238.997605, -237.223798, -235.579863, -234.054589, -232.636763, -231.315172, -230.078604, -228.915845, -227.815684, -226.766908, -225.758304, -224.77866, -223.816762, -222.861399, -221.901358, -220.925425, -219.922389, -218.881037, -217.790156, -216.638533, -215.414957, -213.724638, -211.756222, -209.702959, -207.566086, -205.346837, -203.046448, -200.666154, -198.20719, -195.670791, -193.058193, -190.370629, -187.609336, -184.775549, -181.870502, -178.895431, -175.851571, -172.740157, -169.545312, -166.285755, -162.963092, -159.578927, -156.134868, -152.632521, -149.073491, -145.459385, -141.791808, -138.072367, -134.302668, -130.484317, -126.61892, -122.708083, -118.753412, -114.756513, -110.603051, -106.181474, -101.732509, -97.2652774, -92.7889018, -88.3125041, -83.8452063, -79.3961306, -74.974399, -70.5891335, -66.2494563, -61.9644894, -57.7433549, -53.5951748, -49.5290713, -45.5541664, -41.6795821, -39.0654062, -37.1210689, -35.2421668, -33.4004708, -31.567752, -29.7157814, -27.81633, -25.841169, -23.7620692, -21.5508017, -19.1791375, -16.6188477, -13.8417033, -10.8194753, -7.52393475, -3.92685263, 0, 9.66250643, 19.0890658, 28.294748, 37.2946228, 46.1037602, 54.7372298, 63.2101016, 71.5374454, 79.7343311, 87.8158284, 95.7970073, 103.692937, 111.518689, 119.289331, 127.019935, 134.725569, 141.794321, 148.151692, 155.230927, 162.912741, 171.077853, 179.606977, 188.38083, 197.280129, 206.18559, 214.97793, 223.537865, 231.746112, 239.483387, 245.853383, 251.798749, 257.433306, 262.87088, 269.904513, 277.80842, 285.409019, 292.316368, 298.140522, 302.49154, 304.979478, 305.571747, 303.41229, 298.041244, 288.998751, 275.824949, 257.710131, 234.631746, 206.217392, 172.094671, 131.891184, 89.3810411, 34.9773764, -37.5254536, -131.605178, -245.170033, -378.628843, -532.390434, -704.230899, -897.790142, -1119.58014, -1370.37967, -1658.83132, -1998.55386, -2395.0079, -2839.83055, -3331.00413, -3864.49035, -4437.7477, -5055.64345, -5718.85404, -6442.21018, -7210.05645, -8018.16975, -8862.32696, -9738.30496, -10665.3244, -11609.8575, -12561.8201, -13511.1282, -14447.6978, -15286.1411, -16141.8802, -17037.104, -18096.8552, -19302.6097, -20592.9878, -21951.6082, -23362.0896, -24808.051, -26273.1108, -26727.545, -27947.7169, -29644.5152, -31528.8285, -33311.5453, -34703.5543, -35415.744, -36662.1902, -37313.4254},
        {-274.797779, -270.106773, -265.705458, -261.582224, -257.725463, -254.123567, -250.764927, -247.637935, -244.730981, -242.032458, -239.530757, -237.214269, -235.071387, -233.0905, -231.260001, -229.568281, -228.003733, -226.554746, -225.209713, -223.957024, -222.785073, -221.682249, -220.636945, -219.637552, -218.672461, -217.730063, -216.798752, -215.866916, -214.922949, -213.955242, -212.952186, -211.902172, -210.793593, -209.614839, -207.953172, -206.005986, -203.975495, -201.863109, -199.670236, -197.398286, -195.048667, -192.622788, -190.122057, -187.547885, -184.901679, -182.184849, -179.398804, -176.544952, -173.624703, -170.639464, -167.590646, -164.483127, -161.314771, -158.086911, -154.800882, -151.458016, -148.059647, -144.607109, -141.101736, -137.544861, -133.937818, -130.281941, -126.578563, -122.829017, -119.034639, -115.196761, -111.316716, -107.275317, -102.956267, -98.6094687, -94.2440696, -89.8692161, -85.4940548, -81.1277324, -76.7793954, -72.4581907, -68.1732647, -63.9337642, -59.7488359, -55.6276263, -51.5792823, -47.6129503, -43.7377771, -39.9629093, -37.4560671, -35.6227414, -33.8544978, -32.1228851, -30.3994521, -28.6557473, -26.8633197, -24.9937178, -23.0184904, -20.9091863, -18.6373541, -16.1745425, -13.4923004, -10.5621763, -7.35571909, -3.84447741, 0, 10.2359086, 20.1171173, 29.6669809, 38.9088542, 47.8660918, 56.5620486, 65.0200791, 73.2635382, 81.3157807, 89.2001611, 96.9400343, 104.558755, 112.079678, 119.526158, 126.921549, 134.289207, 140.590131, 145.69591, 152.032846, 159.396646, 167.583018, 176.387667, 185.606303, 195.034631, 204.468358, 213.703193, 222.534842, 230.759012, 238.17141, 243.166899, 247.455674, 251.253695, 254.776922, 260.254148, 266.894917, 273.372018, 279.297563, 284.283661, 287.942424, 289.885962, 290.205167, 287.887653, 282.451856, 273.416213, 260.299159, 245.718744, 225.318888, 197.843123, 162.034982, 116.637997, 60.4129821, -7.93492992, -88.1712277, -181.396954, -291.439905, -420.903623, -572.391646, -750.149254, -954.765126, -1185.03695, -1436.30395, -1721.10724, -2054.41136, -2442.56622, -2877.32462, -3357.13967, -3879.73829, -4443.38532, -5049.0083, -5696.02859, -6394.91257, -7132.33735, -7903.90102, -8705.20166, -9531.83737, -10383.3203, -11250.356, -12127.5639, -13009.5637, -13890.9749, -14710.0776, -15555.3906, -16445.6787, -17491.8864, -18676.2911, -19942.762, -21275.4968, -22658.6931, -24076.5489, -25513.2616, -26345.0891, -27646.0361, -29222.4034, -30880.4917, -32426.6015, -33667.0335, -34408.0884, -35621.9417, -36400.6617},
        {-271.120774, -266.241078, -261.663163, -257.375093, -253.364931, -249.62074, -246.130586, -242.88253, -239.864637, -237.064971, -234.471595, -232.072573, -229.855968, -227.809845, -225.922266, -224.181296, -222.574998, -221.091435, -219.718673, -218.444773, -217.257801, -216.145818, -215.09689, -214.09908, -213.140452, -212.209068, -211.292993, -210.380291, -209.459025, -208.517259, -207.543056, -206.524481, -205.449596, -204.306466, -202.667901, -200.738326, -198.727156, -196.63593, -194.466189, -192.21947, -189.897315, -187.501261, -185.032849, -182.493619, -179.885108, -177.208858, -174.466406, -171.659293, -168.789059, -165.857242, -162.865382, -159.834778, -156.746781, -153.602503, -150.403055, -147.149547, -143.843091, -140.484797, -137.075777, -133.617142, -130.110003, -126.555471, -122.954656, -119.308671, -115.618626, -111.885632, -108.1108, -104.171311, -99.9473201, -95.6955369, -91.4251059, -87.145172, -82.8648803, -78.5933758, -74.3398033, -70.1133078, -65.9230344, -61.778128, -57.6877336, -53.6609961, -49.7070605, -45.8350717, -42.0541748, -38.3735147, -35.9664309, -34.2354839, -32.5692413, -30.9390678, -29.316328, -27.6723865, -25.9786079, -24.206357, -22.3269981, -20.3118961, -18.1324155, -15.7599209, -13.165777, -10.3213483, -7.19799947, -3.76709518, 0, 10.5162872, 20.6189741, 30.3350672, 39.6915729, 48.7154978, 57.4338484, 65.873631, 74.0618524, 82.0255188, 89.7916369, 97.3872132, 104.839254, 112.174766, 119.420756, 126.604229, 133.752193, 139.591217, 143.962531, 149.834606, 156.955784, 165.074405, 173.938812, 183.297346, 192.898348, 202.49016, 211.821122, 220.639577, 228.693865, 235.732328, 239.784487, 242.947737, 245.486067, 247.663465, 251.80403, 257.141695, 262.361027, 267.107978, 271.028504, 273.768557, 274.974091, 275.265328, 273.017437, 267.685754, 258.725614, 245.592354, 233.256275, 214.279006, 186.73714, 148.707271, 98.2659942, 30.1690715, -50.127278, -137.040041, -229.690676, -336.805609, -462.798831, -612.084334, -794.014716, -1006.94295, -1243.83444, -1496.47346, -1779.4053, -2107.46408, -2487.26832, -2911.51553, -3379.2932, -3890.28339, -4443.72774, -5036.68769, -5667.45793, -6342.35102, -7050.41024, -7787.15486, -8548.10415, -9328.77736, -10111.5564, -10908.3823, -11718.0585, -12539.3888, -13371.1766, -14173.3993, -15009.5709, -15895.1351, -16926.6354, -18087.4131, -19327.0178, -20630.4803, -21982.8315, -23369.1023, -24774.3234, -25988.0818, -27369.0434, -28821.77, -30250.8234, -31560.7654, -32656.158, -33441.5629, -34589.2386, -35463.9089},
        {-267.841247, -262.812753, -258.093655, -253.671868, -249.535306, -245.671883, -242.069514, -238.716112, -235.599592, -232.707867, -230.028853, -227.550462, -225.26061, -223.147211, -221.198178, -219.401425, -217.744868, -216.21642, -214.803994, -213.495507, -212.27887, -211.142, -210.072809, -209.059212, -208.089124, -207.150458, -206.231128, -205.319049, -204.402134, -203.468299, -202.505456, -201.501521, -200.444408, -199.322029, -197.702995, -195.79369, -193.805542, -191.740073, -189.598803, -187.383255, -185.094949, -182.735407, -180.306151, -177.808702, -175.244582, -172.615312, -169.922413, -167.167406, -164.351815, -161.477159, -158.54496, -155.581035, -152.562084, -149.489101, -146.363081, -143.185019, -139.95591, -136.676748, -133.348528, -129.972245, -126.548893, -123.079469, -119.564965, -116.006378, -112.404701, -108.760931, -105.07606, -101.225931, -97.0893906, -92.9255522, -88.743524, -84.5524142, -80.3613306, -76.1793812, -72.0156739, -67.8793169, -63.7794179, -59.7250851, -55.7254263, -51.7895495, -47.9265628, -44.145574, -40.4556911, -36.8660222, -34.551833, -32.9146242, -31.3418797, -29.8048634, -28.2748394, -26.7230714, -25.1208234, -23.4393592, -21.6499428, -19.7238382, -17.632309, -15.3466194, -12.8380332, -10.0778142, -7.03722641, -3.68753371, 0, 9.996635, 19.6795971, 29.06747, 38.1788374, 47.032283, 55.6463905, 64.0397435, 72.2309257, 80.2385209, 88.0811127, 95.7772848, 103.345621, 110.804705, 118.17312, 125.46945, 132.712279, 138.872137, 143.817885, 149.938706, 157.028598, 164.881563, 173.291603, 182.052717, 190.958907, 199.804173, 208.382516, 216.487937, 223.914437, 230.456017, 234.628771, 237.973172, 240.666592, 242.886404, 246.236688, 250.181464, 253.804293, 256.854537, 259.081556, 260.23471, 260.063359, 260.555467, 258.540477, 253.329761, 244.234691, 230.56664, 216.811592, 195.812656, 165.58755, 124.153993, 69.5297063, -4.4464583, -90.4704117, -181.48693, -275.99776, -383.850138, -509.797404, -658.592903, -839.934986, -1052.50812, -1289.60221, -1544.85775, -1829.88702, -2156.28559, -2529.42911, -2944.15401, -3400.13437, -3898.59655, -4439.6171, -5017.58096, -5630.09258, -6280.55161, -6959.8757, -7663.66326, -8387.51269, -9127.0224, -9852.33945, -10590.8764, -11344.5946, -12115.4551, -12905.4193, -13691.8484, -14517.7036, -15395.489, -16407.3742, -17538.2873, -18744.1292, -20011.2791, -21326.1167, -22675.0212, -24044.3722, -25522.724, -26953.652, -28313.0774, -29576.9214, -30721.1053, -31721.5504, -32554.1778, -33576.3224, -34474.0904},
        {-264.87853, -259.753684, -254.939747, -250.424734, -246.19666, -242.243541, -238.553393, -235.114232, -231.914074, -228.940933, -226.182827, -223.62777, -221.263779, -219.078869, -217.061057, -215.198356, -213.478784, -211.890357, -210.421089, -209.058997, -207.792097, -206.608403, -205.495933, -204.442701, -203.436724, -202.466017, -201.518595, -200.582476, -199.645673, -198.696204, -197.722084, -196.711328, -195.651953, -194.531974, -192.930609, -191.048871, -189.092687, -187.063339, -184.962109, -182.790279, -180.54913, -178.239946, -175.864007, -173.422596, -170.916995, -168.348486, -165.718351, -163.027872, -160.27833, -157.471009, -154.60719, -151.700087, -148.738791, -145.724325, -142.657714, -139.53998, -136.372147, -133.155238, -129.890277, -126.578286, -123.220289, -119.817309, -116.370371, -112.880496, -109.348709, -105.776033, -102.163491, -98.388568, -94.3317147, -90.2487435, -86.1486865, -82.0405753, -77.933442, -73.8363182, -69.7582359, -65.708227, -61.6953232, -57.7285565, -53.8169586, -49.9695615, -46.195397, -42.5034969, -38.902893, -35.4026173, -33.1753569, -31.6238399, -30.1368086, -28.6855319, -27.2412788, -25.7753183, -24.2589194, -22.6633512, -20.9598825, -19.1197825, -17.1143201, -14.9147644, -12.4923842, -9.81844873, -6.86422685, -3.60098761, 0, 8.34850691, 16.7064331, 25.0670228, 33.4235203, 41.7691699, 50.0972159, 58.4009026, 66.6734743, 74.9081753, 83.0982499, 91.2369422, 99.3174968, 107.333158, 115.277169, 123.142776, 130.923222, 138.500819, 145.852963, 153.223819, 160.582859, 167.899556, 175.143383, 182.283813, 189.290319, 196.132374, 202.779452, 209.201025, 215.366565, 221.245547, 227.030246, 232.385637, 237.214354, 241.419028, 244.763301, 247.219302, 248.79622, 249.438386, 249.090129, 247.695778, 245.199662, 246.033617, 244.288703, 239.031259, 229.327624, 214.244137, 193.76165, 165.80336, 129.206978, 82.8102128, 25.4507754, -45.3831763, -127.854927, -219.332808, -319.757825, -434.32352, -565.865153, -717.217987, -891.245416, -1090.74817, -1318.49629, -1577.81539, -1870.39349, -2200.43725, -2569.97415, -2977.33916, -3422.56216, -3907.50766, -4432.68121, -4991.86136, -5582.63184, -6207.52148, -6858.40809, -7531.16363, -8221.66006, -8925.76933, -9607.96347, -10303.3643, -11015.6938, -11748.6739, -12506.0265, -13276.3933, -14088.63, -14952.9215, -15937.6217, -17029.8903, -18192.6498, -19414.2715, -20683.1269, -21987.5874, -23316.0246, -24839.5793, -26265.2854, -27588.6977, -28805.3708, -29910.8595, -30900.7187, -31770.503, -32590.519, -33408.961},
        {-262.386706, -257.158019, -252.24252, -247.628289, -243.303409, -239.25596, -235.474025, -231.945684, -228.659018, -225.60211, -222.763041, -220.129892, -217.690745, -215.433681, -213.346782, -211.418128, -209.635802, -207.987885, -206.462458, -205.047603, -203.731401, -202.501934, -201.347283, -200.255529, -199.214754, -198.21304, -197.238467, -196.279118, -195.323074, -194.358416, -193.373225, -192.355583, -191.293572, -190.175273, -188.588876, -186.73082, -184.802348, -182.804521, -180.738395, -178.605031, -176.405486, -174.140819, -171.81209, -169.420356, -166.966677, -164.45211, -161.877716, -159.244551, -156.553676, -153.806148, -151.003027, -148.145227, -145.233954, -142.27027, -139.255236, -136.189914, -133.075367, -129.912655, -126.70284, -123.446985, -120.146152, -116.801401, -113.413795, -109.984396, -106.514265, -103.004464, -99.4560548, -95.7482135, -91.763042, -87.7529829, -83.7269995, -79.6940549, -75.6631123, -71.6431348, -67.6430857, -63.6719281, -59.7386252, -55.8521401, -52.021436, -48.2554761, -44.5632236, -40.9536416, -37.4356933, -34.0183418, -31.8747353, -30.4072574, -29.0042991, -27.6370436, -26.276674, -24.8943735, -23.4613251, -21.948712, -20.3277174, -18.5695244, -16.645316, -14.5262755, -12.1835859, -9.58843047, -6.71199224, -3.52545437, 0, 6.96380804, 14.2139024, 21.7214734, 29.4577112, 37.3938061, 45.5009483, 53.750328, 62.1131355, 70.5605611, 79.0637949, 87.5940272, 96.1224482, 104.620248, 113.058618, 121.408746, 129.641825, 138.481357, 148.006008, 156.485304, 164.051644, 170.837431, 176.975064, 182.596943, 187.83547, 192.823044, 197.692066, 202.574936, 207.604056, 212.911825, 220.249016, 227.536255, 234.420429, 240.548428, 244.05389, 245.340328, 245.214791, 243.778143, 241.131247, 237.374967, 232.610167, 233.149201, 230.990989, 225.021105, 214.12512, 197.188609, 170.039634, 135.385654, 92.8766196, 42.1624793, -17.1068166, -83.5476633, -161.362678, -252.887058, -359.718776, -481.95064, -620.351247, -775.689191, -943.840199, -1131.57775, -1351.01301, -1614.72196, -1914.94721, -2248.99732, -2615.71224, -3016.52677, -3451.68463, -3923.30472, -4432.11693, -4971.97545, -5540.62376, -6140.35439, -6763.66191, -7406.71513, -8065.68285, -8736.73388, -9380.6014, -10037.7487, -10713.2036, -11411.9938, -12139.1469, -12892.0061, -13688.4067, -14536.67, -15494.372, -16550.337, -17672.8665, -18851.8111, -20077.0211, -21338.3472, -22625.6397, -24063.9334, -25444.0472, -26752.087, -27974.1584, -29096.3671, -30104.8188, -30985.6192, -31591.7965, -32312.0888},
        {-260.487583, -255.096982, -250.028034, -245.268632, -240.806669, -236.63004, -232.726637, -229.084354, -225.691084, -222.534721, -219.603158, -216.884288, -214.366005, -212.036203, -209.882775, -207.893613, -206.056613, -204.359666, -202.790667, -201.337509, -199.988084, -198.730288, -197.552013, -196.441152, -195.385599, -194.373248, -193.391991, -192.429723, -191.474336, -190.513725, -189.535781, -188.5284, -187.479474, -186.376897, -184.804632, -182.961033, -181.049311, -179.070467, -177.025502, -174.915419, -172.741219, -170.503903, -168.204474, -165.843933, -163.423282, -160.943522, -158.405656, -155.810685, -153.15961, -150.453434, -147.693159, -144.878588, -142.011947, -139.094263, -136.126564, -133.109878, -130.045231, -126.933653, -123.77617, -120.57381, -117.327602, -114.038571, -110.707747, -107.336157, -103.924828, -100.474788, -96.9870653, -93.3404708, -89.4167504, -85.4689935, -81.5061505, -77.5371719, -73.5710081, -69.6166096, -65.6829267, -61.7789099, -57.9135096, -54.0956764, -50.3343605, -46.6385125, -43.0170828, -39.4790218, -36.0332799, -32.6888077, -30.6320319, -29.2583981, -27.9491996, -26.6751754, -25.4070647, -24.1156064, -22.7715397, -21.3456035, -19.808537, -18.1310792, -16.2839692, -14.237946, -11.9637488, -9.43211647, -6.6137882, -3.47950303, 0, 6.89766256, 14.1141147, 21.6165163, 29.3720274, 37.3478079, 45.5110177, 53.8288169, 62.2683653, 70.796823, 79.3813498, 87.9891057, 96.5872507, 105.142945, 113.623348, 121.995619, 130.22692, 139.301863, 149.332959, 157.985984, 165.446125, 171.898568, 177.528497, 182.521099, 187.06156, 191.335064, 195.526798, 199.821947, 204.405697, 209.463234, 217.283126, 225.175935, 232.695834, 239.396992, 243.012889, 244.008042, 243.332486, 241.0866, 237.370761, 232.28535, 225.930745, 224.343837, 219.881727, 211.483303, 198.08745, 178.633057, 148.088324, 110.355496, 65.3661306, 13.0517876, -46.6559748, -110.379794, -185.845454, -278.680953, -391.61442, -521.45341, -667.663697, -829.71105, -999.818437, -1186.34053, -1405.53323, -1673.8712, -1979.20509, -2315.60469, -2680.29089, -3075.82397, -3502.31333, -3961.43142, -4453.69288, -4973.88121, -5520.02178, -6094.81025, -6690.88316, -7304.64923, -7932.5172, -8570.89577, -9177.50993, -9797.12307, -10435.8148, -11099.6649, -11794.7529, -12524.3232, -13299.181, -14126.6208, -15061.4985, -16092.2827, -17187.6246, -18336.9835, -19529.8188, -20755.5899, -22003.7559, -23110.1707, -24357.3751, -25672.1552, -26981.2975, -28211.5882, -29289.8137, -30142.7602, -30534.3348, -31226.0044},
        {-259.058751, -253.47454, -248.223926, -243.294467, -238.673721, -234.349248, -230.308606, -226.539353, -223.029047, -219.765248, -216.735514, -213.927404, -211.328475, -208.926286, -206.708396, -204.662364, -202.775747, -201.036105, -199.430996, -197.947978, -196.57461, -195.298451, -194.107058, -192.987991, -191.928808, -190.917067, -189.940327, -188.986147, -188.042085, -187.095699, -186.134548, -185.146191, -184.118185, -183.03809, -181.477703, -179.641208, -177.737786, -175.768488, -173.734364, -171.636466, -169.475844, -167.253549, -164.970632, -162.628143, -160.227133, -157.768654, -155.253755, -152.683488, -150.058903, -147.381052, -144.650984, -141.874549, -139.047896, -136.171972, -133.247721, -130.276093, -127.258033, -124.194487, -121.086404, -117.934729, -114.740409, -111.504391, -108.227622, -104.911048, -101.555616, -98.1622735, -94.7319662, -91.1419153, -87.2723123, -83.3792792, -79.4717834, -75.5587925, -71.6492736, -67.7521943, -63.8765219, -60.0312237, -56.2252672, -52.4676197, -48.7672487, -45.1331215, -41.5742054, -38.0994679, -34.7178764, -31.4383982, -29.4661148, -28.1867779, -26.9717628, -25.7912206, -24.6153027, -23.41416, -22.1579439, -20.8168055, -19.3608959, -17.7603664, -15.9853681, -14.0060522, -11.7925699, -9.31507232, -6.54371071, -3.44863621, 0, 7.66821068, 15.5284445, 23.5556325, 31.7247058, 40.0105954, 48.3882324, 56.8325478, 65.3184727, 73.8209381, 82.3148751, 90.7752148, 99.1768882, 107.494826, 115.70396, 123.779221, 131.69554, 140.318822, 149.751281, 157.952548, 165.088477, 171.324923, 176.82774, 181.762782, 186.295903, 190.592957, 194.819799, 199.142281, 203.72626, 208.737588, 216.341223, 223.970912, 231.192777, 237.572942, 241.210187, 242.404308, 241.846388, 239.542755, 235.499734, 229.72365, 222.22083, 217.489488, 209.676964, 197.910737, 181.318283, 159.027081, 127.282725, 88.8150466, 43.4719968, -8.89847579, -68.4484221, -131.290724, -206.553363, -301.586694, -419.736159, -556.170513, -709.729676, -879.253571, -1055.9575, -1248.03888, -1472.01285, -1744.36149, -2052.26165, -2389.20224, -2751.95835, -3142.87329, -3561.73447, -4009.45474, -4486.11329, -4987.66266, -5512.38967, -6063.54003, -6633.67874, -7219.37606, -7817.20223, -8423.72753, -8993.29816, -9575.26445, -10176.7527, -10804.8891, -11466.8, -12170.6808, -12921.8759, -13727.0532, -14642.3201, -15654.7749, -16730.9163, -17858.873, -19026.7732, -20222.7455, -21434.9182, -22098.798, -23179.1132, -24516.3251, -25950.8947, -27323.2831, -28473.9515, -29243.3609, -29452.5864, -30104.5765},
        {-257.975999, -252.192794, -246.75592, -241.652561, -236.869898, -232.395115, -228.215393, -224.317916, -220.689866, -217.318424, -214.190775, -211.2941, -208.615582, -206.142403, -203.861746, -201.760793, -199.826727, -198.046731, -196.407986, -194.897676, -193.502982, -192.211088, -191.009176, -189.884428, -188.824027, -187.815155, -186.844995, -185.90073, -184.969541, -184.038612, -183.095125, -182.126262, -181.119206, -180.061139, -178.508929, -176.673966, -174.772815, -172.806622, -170.776534, -168.683699, -166.529263, -164.314374, -162.040178, -159.707823, -157.318456, -154.873223, -152.373272, -149.819751, -147.213805, -144.556583, -141.84923, -139.106895, -136.316421, -133.478651, -130.594429, -127.664599, -124.690004, -121.671487, -118.609893, -115.506064, -112.360845, -109.175079, -105.949609, -102.68528, -99.382934, -96.0434154, -92.6675675, -89.1306043, -85.3107559, -81.4679677, -77.6112274, -73.7495227, -69.8918413, -66.0471706, -62.2244985, -58.4328125, -54.6811003, -50.9783495, -47.3335477, -43.7556826, -40.2537418, -36.836713, -33.5135838, -30.2933419, -28.3975632, -27.2034362, -26.0735746, -24.9776153, -23.885195, -22.7659507, -21.589519, -20.3255369, -18.9436411, -17.4134684, -15.7046557, -13.7868398, -11.6296574, -9.20274542, -6.47574066, -3.41827991, 0, 8.80359995, 17.5961659, 26.3661352, 35.1019456, 43.7920345, 52.4248395, 60.9887981, 69.4723477, 77.863926, 86.1519705, 94.3249186, 102.371208, 110.279276, 118.03756, 125.634498, 133.058528, 140.872471, 149.146821, 156.57081, 163.255959, 169.313788, 174.855817, 179.993566, 184.838554, 189.502302, 194.09633, 198.732158, 203.521307, 208.575295, 215.625316, 222.569338, 229.032977, 234.641852, 238.168944, 239.666189, 239.412894, 237.290467, 233.180319, 226.963857, 218.522491, 210.435113, 199.064676, 183.764821, 163.889187, 138.791415, 107.015174, 68.9265691, 24.0817335, -27.9631994, -87.6520966, -151.699379, -228.83657, -326.612916, -448.471813, -589.473875, -748.443415, -924.204746, -1108.98284, -1309.7012, -1540.4826, -1815.31685, -2123.15711, -2458.61376, -2818.81607, -3205.16144, -3617.08813, -4054.82998, -4518.03153, -5003.42007, -5509.37307, -6039.3203, -6585.8096, -7145.46954, -7714.92868, -8290.8156, -8822.53169, -9365.73948, -9928.87431, -10520.3716, -11148.6666, -11827.8963, -12556.9934, -13341.9498, -14239.9476, -15236.7157, -16296.654, -17406.5579, -18553.223, -19723.4447, -20904.0185, -21151.1204, -22084.5621, -23454.0521, -25009.2991, -26500.0116, -27675.8984, -28286.6679, -28382.4257, -28900.8301},
        {-256.970622, -251.028351, -245.441801, -240.197889, -235.283537, -230.685663, -226.391187, -222.387028, -218.660106, -215.197341, -211.985652, -209.011958, -206.263179, -203.726234, -201.388043, -199.235526, -197.255601, -195.435189, -193.761208, -192.220579, -190.800221, -189.487053, -188.267995, -187.129966, -186.059886, -185.044674, -184.07125, -183.126533, -182.197443, -181.270899, -180.333821, -179.373128, -178.37574, -177.328576, -175.777249, -173.938613, -172.034828, -170.067136, -168.036777, -165.944991, -163.79302, -161.582104, -159.313485, -156.988402, -154.608097, -152.17381, -149.686783, -147.148255, -144.559468, -141.921662, -139.236079, -136.52786, -133.773827, -130.974701, -128.131207, -125.244066, -122.314002, -119.341736, -116.327991, -113.273491, -110.178957, -107.045113, -103.872681, -100.662383, -97.4149422, -94.1310815, -90.8115232, -87.329928, -83.563007, -79.773539, -75.9704835, -72.1628001, -68.3594482, -64.5693875, -60.8015774, -57.0649776, -53.3685476, -49.721247, -46.1320353, -42.609872, -39.1637167, -35.802529, -32.5352683, -29.3708944, -27.52898, -26.3877144, -25.3107363, -24.2677184, -23.2283335, -22.1622545, -21.0391543, -19.8287055, -18.5005811, -17.0244538, -15.3699965, -13.506882, -11.4047831, -9.03337265, -6.36232343, -3.36130827, 0, 9.81656342, 19.4172079, 28.8041679, 37.9796778, 46.9459722, 55.7052853, 64.2598517, 72.6119058, 80.7636821, 88.7174149, 96.4753388, 104.039688, 111.412697, 118.596601, 125.593633, 132.406029, 139.303146, 146.325381, 152.870757, 158.99875, 164.768836, 170.24049, 175.473186, 180.526402, 185.459611, 190.332289, 195.203912, 200.133955, 205.181894, 211.706196, 217.991047, 223.706224, 228.521502, 231.644434, 232.975909, 232.540078, 230.145387, 225.600281, 218.713205, 209.292604, 198.775633, 184.846333, 166.994489, 144.709887, 117.48231, 86.5367109, 49.1939155, 4.50991755, -48.4592898, -110.657713, -181.297783, -265.171453, -366.83602, -488.663085, -628.561793, -786.44484, -962.224923, -1151.94238, -1360.26235, -1596.07435, -1868.68509, -2171.15389, -2500.70022, -2855.6448, -3236.17436, -3641.61912, -4072.45974, -4528.32471, -5004.62408, -5499.15407, -6013.65295, -6541.36845, -7078.73223, -7622.17598, -8168.13138, -8654.38122, -9150.66828, -9668.08646, -10217.7296, -10810.6917, -11474.391, -12192.7141, -12969.7586, -13854.2933, -14832.7336, -15872.8821, -16962.0849, -18087.688, -19237.0377, -20397.4801, -20480.1244, -21396.5507, -22824.2262, -24440.6185, -25923.1951, -26949.4233, -27196.7706, -27441.9376, -27507.4317},
        {-256.262747, -250.167433, -244.43654, -239.056746, -234.014731, -229.297174, -224.890754, -220.782151, -216.958043, -213.40511, -210.11003, -207.059485, -204.240151, -201.63871, -199.241839, -197.036219, -195.008528, -193.145445, -191.433651, -189.859824, -188.410642, -187.072787, -185.832936, -184.677769, -183.593966, -182.568205, -181.587165, -180.637527, -179.705968, -178.779169, -177.843808, -176.886566, -175.89412, -174.85315, -173.298964, -171.453421, -169.543805, -167.571443, -165.537662, -163.44379, -161.291153, -159.081079, -156.814895, -154.493927, -152.119504, -149.692951, -147.215596, -144.688767, -142.113791, -139.491993, -136.824703, -134.14637, -131.42448, -128.659642, -125.852464, -123.003557, -120.113528, -117.182988, -114.212544, -111.202807, -108.154385, -105.067887, -101.943922, -98.7830991, -95.5860278, -92.3533169, -89.0855753, -85.6550894, -81.9372263, -78.1972513, -74.444092, -70.6866762, -66.9339315, -63.1947857, -59.4781664, -55.7930013, -52.1482181, -48.5527446, -45.0155084, -41.5454371, -38.1514586, -34.8425004, -31.6274904, -28.515356, -26.7193896, -25.6210043, -24.587138, -23.5876343, -22.5923371, -21.5710902, -20.4937371, -19.3301218, -18.050088, -16.6234793, -15.0201397, -13.2099127, -11.1626422, -8.84817195, -6.23634565, -3.29700709, 0, 10.7752239, 21.1319596, 31.0863138, 40.654393, 49.8523038, 58.6961528, 67.2020465, 75.3860916, 83.2643946, 90.853062, 98.1682005, 105.225917, 112.042317, 118.633508, 125.015597, 131.204689, 137.147978, 142.851724, 148.487912, 154.057882, 159.562973, 165.004524, 170.383873, 175.702361, 180.961327, 186.162109, 191.306046, 196.394479, 201.428745, 207.35015, 212.875414, 217.723886, 221.614916, 224.156023, 225.122472, 224.319349, 221.499552, 216.415982, 208.821535, 198.469112, 185.984737, 169.982447, 150.044314, 125.752406, 96.6887935, 66.518637, 29.7201419, -15.1453951, -69.5166772, -134.832408, -213.073742, -304.473935, -409.801878, -530.100398, -666.971605, -821.766503, -995.836095, -1190.02773, -1406.31054, -1647.20307, -1916.37616, -2211.93654, -2534.28519, -2883.25161, -3257.69698, -3656.92696, -4081.9321, -4532.45488, -5002.05982, -5487.80603, -5989.3314, -6500.71955, -7018.13697, -7537.75012, -8055.72549, -8497.44445, -8947.55485, -9419.91945, -9928.40098, -10486.8622, -11136.57, -11845.7135, -12616.411, -13487.8138, -14447.3809, -15467.1411, -16535.3251, -17640.1635, -18769.8871, -19912.7266, -19971.5883, -20919.7986, -22388.479, -24008.7508, -25411.7355, -26228.5544, -26090.3291, -26490.6239, -26016.7995},
        {-255.903863, -249.654952, -243.778789, -238.261827, -233.090525, -228.251336, -223.730716, -219.515123, -215.59101, -211.944835, -208.563052, -205.432118, -202.538489, -199.868619, -197.408965, -195.145982, -193.066127, -191.155855, -189.401622, -187.789883, -186.307095, -184.939712, -183.674191, -182.496988, -181.394558, -180.353358, -179.359842, -178.400466, -177.461687, -176.52996, -175.591741, -174.633485, -173.641649, -172.602687, -171.042991, -169.188724, -167.271583, -165.292954, -163.254222, -161.156773, -159.001993, -156.791268, -154.525983, -152.207524, -149.837277, -147.416628, -144.946963, -142.429667, -139.866126, -137.257726, -134.605853, -131.951794, -129.256168, -126.519496, -123.742299, -120.925098, -118.068413, -115.172766, -112.238677, -109.266666, -106.257256, -103.210966, -100.128318, -97.0098314, -93.8560281, -90.6674286, -87.4445537, -84.0586143, -80.3839259, -76.6877014, -72.9788442, -69.2662577, -65.5588454, -61.8655108, -58.1951572, -54.5566881, -50.9590071, -47.4110174, -43.9216226, -40.499726, -37.1542312, -33.8940416, -30.7280606, -27.6651916, -25.9118465, -24.8529865, -23.8591798, -22.9004684, -21.9468943, -20.9684995, -19.935326, -18.8174158, -17.5848108, -16.2075531, -14.6556846, -12.8992474, -10.9082835, -8.65283475, -6.10294327, -3.22865103, 0, 11.5561054, 22.5218714, 32.9253326, 42.7945236, 52.1574791, 61.0422335, 69.4768214, 77.4892775, 85.1076363, 92.3599323, 99.2742002, 105.878475, 112.20079, 118.269181, 124.111682, 129.756328, 134.841264, 139.338826, 144.158941, 149.246097, 154.544779, 159.999475, 165.554671, 171.154855, 176.744514, 182.268134, 187.670202, 192.895205, 197.88763, 203.186138, 207.923178, 211.864985, 214.777792, 216.579555, 216.960649, 215.570988, 212.131292, 206.362277, 197.984663, 186.719169, 172.81809, 155.308782, 133.807959, 107.932334, 77.2986213, 47.1422901, 10.0566097, -35.7463952, -92.0546995, -160.656279, -245.777696, -343.787841, -451.863903, -569.836459, -702.953415, -854.030409, -1025.88308, -1224.24116, -1448.34389, -1694.25149, -1959.96365, -2248.62467, -2563.71594, -2906.56538, -3274.852, -3667.92973, -4087.4195, -4533.26297, -4997.08945, -5475.23018, -5965.25371, -6462.06363, -6961.5629, -7459.65448, -7952.2413, -8353.61713, -8761.69641, -9192.78441, -9663.18636, -10189.2075, -10824.2846, -11522.8371, -12285.2568, -13142.0877, -14081.7864, -15080.2459, -16126.5509, -17209.7863, -18319.0366, -19443.3867, -19624.1708, -20637.7635, -22107.5852, -23657.0562, -24909.5971, -25488.6282, -25017.5701, -25477.1926, -24452.441},
        {-255.940798, -249.522074, -243.484644, -237.814757, -232.498661, -227.522604, -222.872835, -218.535604, -214.497157, -210.743744, -207.261612, -204.037012, -201.056191, -198.305397, -195.770879, -193.438886, -191.295666, -189.327468, -187.52054, -185.86113, -184.335488, -182.929861, -181.630498, -180.423648, -179.295559, -178.23248, -177.220658, -176.246344, -175.295784, -174.355228, -173.410924, -172.449121, -171.456067, -170.41801, -168.856897, -167.000892, -165.083744, -163.106767, -161.071278, -158.978592, -156.830026, -154.626896, -152.370518, -150.062207, -147.703281, -145.295054, -142.838843, -140.335964, -137.787732, -135.195465, -132.560478, -129.920124, -127.238901, -124.517343, -121.755986, -118.955364, -116.116012, -113.238465, -110.323257, -107.370924, -104.382, -101.357019, -98.2965174, -95.2010292, -92.0710892, -88.9072323, -85.7099933, -82.3498254, -78.7003052, -75.0302508, -71.3486297, -67.6644092, -63.9865569, -60.3240401, -56.6858264, -53.0808832, -49.5181779, -46.006678, -42.555351, -39.1731642, -35.8690852, -32.6520814, -29.5311201, -26.515169, -24.8177101, -23.8201181, -22.8892906, -21.9951061, -21.1074432, -20.1961807, -19.231197, -18.1823709, -17.019581, -15.7127059, -14.2316243, -12.5462147, -10.6263558, -8.44192628, -5.96280471, -3.15886974, 0, 11.6265849, 22.6357423, 33.0577523, 42.922895, 52.2614506, 61.1036992, 69.4799211, 77.4203962, 84.9554048, 92.1152271, 98.9301431, 105.430433, 111.646377, 117.608255, 123.346348, 128.890935, 133.711669, 137.758741, 142.330517, 147.337142, 152.688763, 158.295524, 164.067571, 169.915051, 175.748109, 181.47689, 187.011541, 192.262206, 197.139033, 201.983138, 206.115673, 209.317491, 211.369446, 212.111519, 211.295003, 208.655418, 203.95588, 196.959507, 187.429412, 175.128713, 161.23821, 143.671863, 121.915417, 95.454612, 63.7751916, 29.8389536, -11.213429, -60.7652275, -120.199713, -190.900158, -275.673818, -372.739553, -479.206962, -595.14062, -726.638183, -877.088575, -1049.88072, -1253.24555, -1484.62953, -1736.19696, -2002.61682, -2289.36247, -2600.69782, -2938.4559, -3300.35077, -3685.98713, -4097.61822, -4535.5354, -4990.31892, -5458.04219, -5935.2591, -6417.48865, -6900.63793, -7380.61402, -7853.32402, -8231.27839, -8614.62999, -9020.13504, -9464.5498, -9964.63049, -10575.6241, -11250.1357, -11988.4253, -12821.5274, -13737.5206, -14711.6613, -15732.6692, -16789.2642, -17870.1663, -18964.0952, -19612.9685, -20719.0696, -22027.7583, -23284.3946, -24234.3384, -24622.9497, -24195.5884, -24180.6015, -22781.8064},
        {-256.431669, -249.833255, -243.624817, -237.792402, -232.322052, -227.199811, -222.411723, -217.943833, -213.782183, -209.912818, -206.321781, -202.995117, -199.918869, -197.079082, -194.461798, -192.053063, -189.838919, -187.80541, -185.938581, -184.224476, -182.649137, -181.19861, -179.858938, -178.616164, -177.456333, -176.365488, -175.329674, -174.334934, -173.367312, -172.412852, -171.457598, -170.487593, -169.488882, -168.447508, -166.882115, -165.022278, -163.10319, -161.126063, -159.092109, -157.00254, -154.858567, -152.661403, -150.412259, -148.112347, -145.762879, -143.365068, -140.920124, -138.42926, -135.893688, -133.31462, -130.693267, -128.059746, -125.385739, -122.671829, -119.918603, -117.126645, -114.296541, -111.428876, -108.524236, -105.583205, -102.60637, -99.594315, -96.5476256, -93.4668872, -90.352685, -87.2056043, -84.0262305, -80.6848412, -77.0548427, -73.40557, -69.7460557, -66.0853324, -62.4324327, -58.7963892, -55.1862345, -51.6110013, -48.079722, -44.6014294, -41.185156, -37.8399345, -34.5747974, -31.3987773, -28.3209069, -25.3502188, -23.7094376, -22.7752215, -21.9096411, -21.0823423, -20.2629712, -19.4211735, -18.5265954, -17.5488828, -16.4576817, -15.2226379, -13.8133975, -12.1996065, -10.3509107, -8.23695625, -5.82738899, -3.09185492, 0, 11.4448745, 22.2966395, 32.5842498, 42.3366597, 51.582824, 60.3516971, 68.6722336, 76.5733882, 84.0841153, 91.2333696, 98.0501056, 104.563278, 110.801841, 116.79475, 122.570958, 128.159422, 132.921962, 136.792229, 141.308167, 146.355774, 151.821047, 157.589984, 163.548582, 169.58284, 175.578755, 181.422325, 186.999547, 192.196419, 196.898939, 201.327626, 204.911298, 207.435598, 208.686165, 208.328483, 206.208272, 202.203295, 196.135242, 187.825801, 177.096661, 163.769513, 150.284969, 133.04873, 111.370085, 84.5583256, 51.9227412, 13.0129065, -33.1622429, -87.3534875, -150.311608, -222.787385, -305.047297, -398.91835, -502.641188, -616.974681, -747.81908, -898.766258, -1073.40809, -1281.36909, -1518.83745, -1775.42029, -2043.64068, -2330.12607, -2639.11708, -2972.56239, -3328.52715, -3706.85913, -4110.26914, -4539.3472, -4984.18549, -5440.81435, -5905.43783, -6373.88586, -6842.12867, -7306.13649, -7761.87952, -8123.10651, -8488.56454, -8874.77923, -9298.27618, -9775.58099, -10360.2942, -11007.1498, -11716.7842, -12523.6832, -13415.0843, -14364.1212, -15358.7369, -16386.8742, -17436.4758, -18495.4846, -19685.797, -20883.9209, -21988.3037, -22897.3929, -23509.6359, -23723.4802, -23437.3733, -22790.8939, -21108.1006},
        {-257.431268, -250.643145, -244.253935, -238.249471, -232.615583, -227.338104, -222.402867, -217.795703, -213.502444, -209.508923, -205.800971, -202.364421, -199.185104, -196.248853, -193.541499, -191.048876, -188.756814, -186.651146, -184.717704, -182.94232, -181.310826, -179.809054, -178.422837, -177.138005, -175.940392, -174.81583, -173.75015, -172.729184, -171.738765, -170.764724, -169.792895, -168.809108, -167.799195, -166.74899, -165.173534, -163.303955, -161.377044, -159.393909, -157.35566, -155.263409, -153.118264, -150.921336, -148.673736, -146.376572, -144.030956, -141.637998, -139.198807, -136.714493, -134.186167, -131.61494, -129.00192, -126.369508, -123.697063, -120.985233, -118.234667, -115.446013, -112.619921, -109.757038, -106.858013, -103.923494, -100.954131, -97.9505723, -94.9134657, -91.8434601, -88.741204, -85.6073462, -82.4425352, -79.1176469, -75.5066732, -71.8779088, -68.240415, -64.603253, -60.9754839, -57.366169, -53.7843695, -50.2391466, -46.7395615, -43.2946754, -39.9135495, -36.6052451, -33.3788233, -30.2433453, -27.2078724, -24.2814657, -22.6942401, -21.8190467, -21.0140607, -20.2487177, -19.4924533, -18.7147033, -17.8849032, -16.9724888, -15.9468958, -14.7775598, -13.4339164, -11.8854014, -10.1014504, -8.05149909, -5.70498314, -3.03133822, 0, 11.1773799, 21.8050992, 31.909248, 41.5159168, 50.651196, 59.3411759, 67.6119469, 75.4895994, 83.0002237, 90.1699101, 97.0247491, 103.590831, 109.894246, 115.961085, 121.817438, 127.489395, 132.272778, 136.089352, 140.617008, 145.72535, 151.283981, 157.162506, 163.230529, 169.357654, 175.413485, 181.267627, 186.789682, 191.849256, 196.315953, 200.33235, 203.394987, 205.291577, 205.809831, 204.462108, 201.173796, 195.943712, 188.642175, 179.1395, 167.306008, 153.012015, 139.858976, 122.85051, 101.126929, 73.8285458, 40.0956741, -4.04010371, -55.5525618, -114.524205, -181.037537, -255.175063, -334.410101, -424.623347, -525.724451, -639.322896, -770.498656, -922.773525, -1099.6693, -1311.27151, -1553.04645, -1813.29997, -2083.47831, -2370.47518, -2678.10777, -3008.18753, -3359.12729, -3730.89989, -4126.33954, -4546.16092, -4980.58707, -5425.77568, -5878.27217, -6333.78689, -6788.34333, -7237.965, -7678.6754, -8026.40373, -8377.29136, -8747.38535, -9152.73278, -9609.38072, -10168.1397, -10787.1177, -11467.4632, -12248.1265, -13115.293, -14039.3969, -15007.4109, -16006.3077, -17023.0601, -18044.6408, -19716.3085, -20984.5838, -21889.4624, -22470.9401, -22769.0127, -22823.6759, -22674.9255, -21377.0466, -19494.5479},
        {-259.094316, -252.10829, -245.53043, -239.346333, -233.541595, -228.101813, -223.012583, -218.259503, -213.828167, -209.704172, -205.873116, -202.320594, -199.032203, -195.993539, -193.190198, -190.607778, -188.231875, -186.048084, -184.042003, -182.199227, -180.505354, -178.945979, -177.5067, -176.173112, -174.930812, -173.765396, -172.662461, -171.607604, -170.58642, -169.584506, -168.587459, -167.580875, -166.55035, -165.481481, -163.882178, -161.986928, -160.035959, -158.03037, -155.97126, -153.859725, -151.696865, -149.483778, -147.221562, -144.911316, -142.554137, -140.151125, -137.703376, -135.21199, -132.678065, -130.102699, -127.48699, -124.851801, -122.178037, -119.466369, -116.717465, -113.931997, -111.110633, -108.254044, -105.362899, -102.437869, -99.4796221, -96.4888295, -93.4661606, -90.4122853, -87.3278733, -84.2135945, -81.0701187, -77.7701747, -74.189563, -70.5928223, -66.9889163, -63.3868089, -59.7954637, -56.2238445, -52.6809149, -49.1756388, -45.7169798, -42.3139016, -38.9753681, -35.7103429, -32.5277897, -29.4366723, -26.4459545, -23.5645998, -22.0199951, -21.1862647, -20.4228884, -19.6992898, -18.9848924, -18.2491199, -17.4613958, -16.5911437, -15.6077871, -14.4807495, -13.1794546, -11.6733259, -9.93178696, -7.92426133, -5.62017259, -2.98894429, 0, 11.1347384, 21.7250013, 31.7958475, 41.3723358, 50.4795251, 59.1424742, 67.3862419, 75.2358871, 82.7164686, 89.8530452, 96.6706758, 103.194419, 109.449334, 115.46048, 121.252914, 126.851697, 131.53102, 135.208674, 139.608109, 144.593484, 150.028957, 155.778687, 161.706832, 167.677552, 173.555004, 179.203349, 184.486744, 189.269348, 193.415321, 197.018571, 199.629265, 201.042635, 201.053915, 199.291848, 195.634912, 190.005982, 182.25024, 172.212865, 159.739038, 144.673938, 130.658886, 112.587575, 89.5624635, 60.6860086, 25.060668, -22.7223488, -77.9215247, -140.30659, -209.647276, -285.713313, -364.630264, -454.26601, -555.696937, -670.997468, -804.279675, -958.740766, -1137.57795, -1350.13795, -1592.07085, -1852.31815, -2122.79991, -2409.7946, -2716.94444, -3045.78366, -3394.27461, -3762.01921, -4150.92881, -4561.2041, -4984.57764, -5417.57204, -5858.54229, -6301.8967, -6743.52359, -7179.31126, -7605.14804, -7935.88031, -8268.6988, -8619.75228, -9005.18954, -9441.15938, -9978.94326, -10576.1355, -11233.8281, -11993.0119, -12838.9856, -13740.4231, -14683.4541, -15654.2086, -16638.8163, -17623.407, -19376.5516, -20629.7666, -21450.1334, -21904.7335, -22060.6484, -21984.9595, -21744.7484, -20047.6994, -18075.8632},
        {-261.466571, -254.273347, -247.498006, -241.125875, -235.142283, -229.532558, -224.282028, -219.376022, -214.799868, -210.538894, -206.578429, -202.9038, -199.500336, -196.353365, -193.448215, -190.770215, -188.304694, -186.036978, -183.952397, -182.036278, -180.273951, -178.650742, -177.151982, -175.762997, -174.469115, -173.255667, -172.107978, -171.011379, -169.951196, -168.912759, -167.881395, -166.842433, -165.781201, -164.683028, -163.044103, -161.104613, -159.110622, -157.063331, -154.963943, -152.813659, -150.61368, -148.365208, -146.069446, -143.727595, -141.340856, -138.910431, -136.437522, -133.92333, -131.369058, -128.775907, -126.145079, -123.504005, -120.827087, -118.114961, -115.368258, -112.587613, -109.773657, -106.927024, -104.048348, -101.138261, -98.1973959, -95.2263868, -92.2258663, -89.1964677, -86.1388239, -83.0535681, -79.9413334, -76.6781139, -73.1428962, -69.5933721, -66.0382545, -62.4862564, -58.9460906, -55.4264703, -51.9361083, -48.4837175, -45.078011, -41.7277017, -38.4415025, -35.2281264, -32.0962864, -29.0546953, -26.1120662, -23.2771119, -21.761658, -20.9475553, -20.2021942, -19.4952443, -18.7963752, -18.0752566, -17.3015581, -16.4449494, -15.4751, -14.3616797, -13.0743581, -11.5828047, -9.85668935, -7.86568154, -5.57945095, -2.96766723, 0, 11.4353913, 22.2705549, 32.5327715, 42.2493214, 51.4474853, 60.1545437, 68.397777, 76.2044659, 83.6018909, 90.6173325, 97.2780713, 103.611388, 109.644562, 115.404876, 120.919608, 126.216041, 130.577858, 133.926113, 137.970028, 142.577542, 147.616593, 152.95512, 158.461062, 164.002357, 169.446944, 174.662762, 179.517749, 183.879845, 187.616987, 190.779023, 192.985284, 194.049134, 193.783941, 192.289824, 189.236772, 184.218306, 176.961766, 167.194492, 154.643823, 139.0371, 122.621134, 101.836998, 75.9190899, 44.1018109, 5.61956032, -43.5693232, -100.135364, -164.025146, -235.185256, -313.562278, -395.973225, -489.319287, -594.91601, -714.716648, -851.976434, -1009.36604, -1189.55613, -1400.03435, -1637.55962, -1893.63604, -2162.1674, -2448.09521, -2755.38763, -3085.2434, -3434.16858, -3800.82327, -4185.04005, -4585.78301, -4997.71695, -5417.93859, -5848.09297, -6280.02873, -6709.26814, -7131.33348, -7541.74705, -7849.62504, -8158.49733, -8485.48774, -8847.72007, -9262.31813, -9785.86948, -10369.725, -11014.0954, -11758.2789, -12587.0141, -13468.7533, -14389.0495, -15333.4557, -16287.5248, -17236.8099, -18572.9003, -19708.8378, -20593.6321, -21176.2929, -21405.8299, -21231.2528, -20601.5715, -18853.6713, -16904.2323},
        {-264.383961, -256.967526, -249.979444, -243.40475, -237.228477, -231.435661, -226.011334, -220.940532, -216.208288, -211.799637, -207.699613, -203.89325, -200.365582, -197.101644, -194.086469, -191.305092, -188.742547, -186.383868, -184.21409, -182.218246, -180.381371, -178.688499, -177.124664, -175.674901, -174.324242, -173.057724, -171.86038, -170.717243, -169.613349, -168.533731, -167.463424, -166.387462, -165.290879, -164.158709, -162.473965, -160.483252, -158.439207, -156.343155, -154.196424, -152.000339, -149.756225, -147.465408, -145.129213, -142.748968, -140.325997, -137.861626, -135.357181, -132.813988, -130.233372, -127.616659, -124.965175, -122.313844, -119.629664, -116.913234, -114.165151, -111.386013, -108.576416, -105.736958, -102.868237, -99.9708496, -97.0453937, -94.0924667, -91.1126658, -88.1065886, -85.0748325, -82.0179948, -78.936673, -75.7110768, -72.2243063, -68.7252496, -65.2223083, -61.7238842, -58.2383788, -54.7741939, -51.3397311, -47.943392, -44.5935784, -41.2986918, -38.067134, -34.9073067, -31.8276113, -28.8364498, -25.9422236, -23.1533345, -21.6611886, -20.8578025, -20.1209738, -19.4207135, -18.7270323, -18.0099413, -17.2394512, -16.385573, -15.4183175, -14.3076957, -13.0237184, -11.5363966, -9.81574108, -7.83176277, -5.55447257, -2.95388134, 0, 11.8276695, 22.9826085, 33.4952365, 43.3959729, 52.7152373, 61.4834492, 69.731028, 77.4883933, 84.7859645, 91.6541612, 98.1234028, 104.224109, 109.986699, 115.441592, 120.619209, 125.549968, 129.514511, 132.436146, 136.04122, 140.199484, 144.780692, 149.654596, 154.69095, 159.759505, 164.730016, 169.472234, 173.855913, 177.750806, 181.026665, 183.64748, 185.354213, 185.988347, 185.391363, 184.262023, 182.013169, 177.829034, 171.293916, 161.992113, 149.507921, 133.425639, 114.214356, 90.3042936, 61.1066374, 26.0325737, -15.5067115, -65.4473714, -122.694046, -187.498713, -260.113353, -340.789943, -427.544009, -525.597429, -636.336191, -761.463584, -903.330711, -1063.99782, -1245.52516, -1453.16846, -1685.06624, -1935.87106, -2201.93423, -2486.5935, -2794.13375, -3125.38159, -3475.3037, -3841.55658, -4221.78819, -4613.65282, -5014.83596, -5423.00542, -5843.11631, -6264.42807, -6682.08603, -7091.2355, -7487.02181, -7772.23266, -8057.46039, -8360.93972, -8700.90537, -9095.59207, -9607.85262, -10181.5583, -10815.8217, -11547.3045, -12360.2561, -13223.281, -14121.4108, -15039.6776, -15963.1132, -16876.7494, -17662.9683, -18651.6324, -19636.9178, -20413.0004, -20774.0562, -20514.2609, -19427.7906, -17734.4354, -15872.9148},
        {-267.802551, -260.14339, -252.924016, -246.129141, -239.743478, -233.751738, -228.138633, -222.888877, -217.987179, -213.418254, -209.166813, -205.217567, -201.555229, -198.164512, -195.030127, -192.136786, -189.469201, -187.012085, -184.750149, -182.668106, -180.750668, -178.982546, -177.348453, -175.833101, -174.421203, -173.097469, -171.846612, -170.653345, -169.502379, -168.378427, -167.2662, -166.15041, -165.015771, -163.846993, -162.113184, -160.067705, -157.97011, -155.821846, -153.624355, -151.379084, -149.087475, -146.750973, -144.371024, -141.94907, -139.486558, -136.98493, -134.445632, -131.870108, -129.259803, -126.61616, -123.940625, -121.274571, -118.578648, -115.853434, -113.099508, -110.317449, -107.507837, -104.671249, -101.808264, -98.9194625, -96.0054218, -93.0667212, -90.1039396, -87.1176557, -84.1084484, -81.0768966, -78.023579, -74.8337642, -71.3954892, -67.9471519, -64.496806, -61.0525056, -57.6223046, -54.2142569, -50.8364164, -47.4968372, -44.203573, -40.9646779, -37.7882058, -34.6822107, -31.6547463, -28.7138668, -25.867626, -23.1240779, -21.6512059, -20.8527016, -20.1183152, -19.4184587, -18.7235441, -18.0039833, -17.2301884, -16.3725714, -15.4015442, -14.2875188, -13.0009073, -11.5121216, -9.79157367, -7.80967557, -5.53683926, -2.94347674, 0, 12.2718137, 23.7875483, 34.5814944, 44.6879428, 54.1411838, 62.9755081, 71.2252064, 78.9245691, 86.1078869, 92.8094503, 99.06355, 104.904477, 110.36652, 115.483972, 120.291123, 124.822263, 128.316205, 130.693602, 133.788036, 137.463338, 141.583341, 146.011874, 150.612771, 155.249861, 159.786976, 164.087947, 168.016605, 171.436783, 174.21231, 176.145566, 177.184368, 177.210981, 176.107673, 175.250718, 173.77538, 170.418518, 164.614199, 155.796485, 143.399443, 126.857135, 104.747587, 77.6214374, 45.0802358, 6.72553289, -37.8411209, -88.1785226, -145.734687, -211.117978, -284.936757, -367.799388, -459.112043, -562.154617, -678.295673, -808.940447, -955.569058, -1119.628, -1302.56378, -1507.39244, -1733.63421, -1979.09709, -2242.57323, -2525.81569, -2833.34338, -3165.63017, -3516.29577, -3882.10608, -4258.80459, -4642.89214, -5034.61822, -5432.11187, -5843.04088, -6254.52802, -6661.40043, -7058.48527, -7440.6097, -7704.57291, -7967.73699, -8249.43608, -8569.00431, -8945.77583, -9449.07234, -10014.9154, -10641.2158, -11361.4975, -12159.5746, -13004.4225, -13880.597, -14772.6543, -15665.1504, -16542.6412, -16746.0083, -17578.5372, -18671.5662, -19656.4336, -20164.4778, -19827.037, -18275.4495, -16693.2166, -14966.1649},
        {-271.647706, -263.714198, -256.233634, -249.190364, -242.568741, -236.353118, -230.527847, -225.07728, -219.98577, -215.237669, -210.81733, -206.709104, -202.897345, -199.366404, -196.100634, -193.084388, -190.302017, -187.737874, -185.376311, -183.201681, -181.198337, -179.350629, -177.642912, -176.059537, -174.584856, -173.203222, -171.898988, -170.656505, -169.460127, -168.294205, -167.143091, -165.991139, -164.8227, -163.622127, -161.846404, -159.754659, -157.612361, -155.420976, -153.18197, -150.896811, -148.566963, -146.193895, -143.779073, -141.323962, -138.830029, -136.298742, -133.731566, -131.129968, -128.495414, -125.829371, -123.133306, -120.446681, -117.732144, -114.990335, -112.221899, -109.427478, -106.607715, -103.763252, -100.894733, -98.0027999, -95.0880956, -92.151263, -89.1929448, -86.213784, -83.2144231, -80.1955051, -77.1576727, -73.9923016, -70.5933912, -67.1869392, -63.7806716, -60.3823148, -56.9995947, -53.6402374, -50.3119692, -47.0225161, -43.7796042, -40.5909597, -37.4643086, -34.4073772, -31.4278915, -28.5335776, -25.7321616, -23.0313697, -21.5778798, -20.7859857, -20.0562921, -19.359564, -18.6665661, -17.9480636, -17.1748213, -16.3176043, -15.3471773, -14.2343055, -12.9497536, -11.4642868, -9.7486698, -7.7736677, -5.5100454, -2.92856786, 0, 12.6900572, 24.5404796, 35.5901342, 45.8778878, 55.4426074, 64.3231599, 72.5584123, 80.1872314, 87.2484843, 93.7810379, 99.8237591, 105.415515, 110.595172, 115.401598, 119.873659, 124.050222, 126.969764, 128.52824, 131.027302, 134.29145, 138.14518, 142.412991, 146.919379, 151.488843, 155.945879, 160.114987, 163.820662, 166.887404, 169.139708, 169.953714, 169.766678, 168.537604, 166.225498, 164.909265, 163.487961, 160.295336, 154.654425, 145.888263, 133.319888, 116.272333, 92.0810272, 62.6615385, 27.7257832, -13.0143231, -59.8468644, -111.087862, -169.490479, -235.835815, -310.90497, -395.479044, -489.58189, -595.67738, -715.272603, -849.678064, -999.80291, -1166.73649, -1351.56816, -1556.27342, -1780.85406, -2024.34514, -2286.39516, -2568.14277, -2873.92658, -3204.07711, -3552.1333, -3914.75154, -4287.55553, -4666.93392, -5053.06188, -5443.9727, -5846.85818, -6249.18412, -6645.81358, -7031.60963, -7401.43531, -7649.6767, -7896.79308, -8162.76675, -8467.58, -8831.21515, -9327.74833, -9887.00543, -10505.7857, -11214.5195, -11996.6937, -12821.7823, -13673.9728, -14537.4527, -15396.4097, -16235.0312, -16097.3152, -16820.8832, -17946.7733, -19016.0235, -19569.6721, -19148.7571, -17294.3165, -15738.2395, -14159.0129},
        {-275.882505, -267.641496, -259.868388, -252.547136, -245.661694, -239.196017, -233.13406, -227.459776, -222.157121, -217.21005, -212.602516, -208.318475, -204.341881, -200.656688, -197.246852, -194.096327, -191.189066, -188.509026, -186.040161, -183.766424, -181.671772, -179.740157, -177.955536, -176.301862, -174.763091, -173.323176, -171.966072, -170.675734, -169.436117, -168.231175, -167.044862, -165.861134, -164.663944, -163.437248, -161.628013, -159.499981, -157.32335, -155.0995, -152.829814, -150.515671, -148.158453, -145.75954, -143.320313, -140.842154, -138.326442, -135.774559, -133.187886, -130.567804, -127.915693, -125.232935, -122.52091, -119.808067, -117.068132, -114.3019, -111.510164, -108.693718, -105.853357, -102.989875, -100.104065, -97.1967222, -94.2686402, -91.3206132, -88.3534353, -85.3679005, -82.3648029, -79.3449366, -76.3090958, -73.1556334, -69.7853622, -66.410353, -63.0380409, -59.6758607, -56.3312474, -53.0116359, -49.7244611, -46.477158, -43.2771613, -40.1319062, -37.0488274, -34.03536, -31.0989388, -28.2469987, -25.4869747, -22.8263017, -21.3935776, -20.6120736, -19.8915855, -19.2031644, -18.5178613, -17.8067272, -17.0408131, -16.1911701, -15.2288492, -14.1249013, -12.8503777, -11.3763292, -9.67380684, -7.71386174, -5.46754488, -2.90590729, 0, 13.051323, 25.1840721, 36.4421743, 46.8695571, 56.5101476, 65.4078732, 73.6066609, 81.1504382, 88.0831323, 94.4486703, 100.29098, 105.653987, 110.581621, 115.117808, 119.306474, 123.191549, 125.447767, 125.923458, 127.759528, 130.706506, 134.514921, 138.935304, 143.718182, 148.614085, 153.373542, 157.747083, 161.485237, 164.338532, 166.057499, 165.328835, 163.356972, 160.211589, 155.962364, 153.368779, 151.15561, 147.330356, 141.155754, 131.894542, 118.809459, 101.163242, 75.8546804, 45.2299246, 9.01422461, -33.0671697, -81.2890084, -134.110939, -194.07659, -261.914486, -338.353153, -424.121117, -518.920546, -625.76076, -746.486952, -882.559935, -1034.65574, -1203.80275, -1391.02934, -1598.64237, -1826.10119, -2071.47047, -2333.45299, -2613.67725, -2915.87272, -3240.46135, -3582.2607, -3938.6608, -4307.08072, -4684.91802, -5069.46395, -5458.06991, -5854.09663, -6247.96164, -6634.93618, -7010.2915, -7369.29884, -7607.92004, -7845.56308, -8102.32655, -8398.30901, -8753.60907, -9244.85295, -9797.92649, -10408.7021, -11104.9803, -11869.9952, -12673.6779, -13499.928, -14332.6449, -15155.7281, -15953.077, -15775.9273, -16450.8696, -17518.1081, -18517.8472, -18990.2913, -18475.6444, -16514.1111, -14871.5871, -13437.8199},
        {-280.514156, -271.943319, -263.856434, -256.237031, -249.068639, -242.334787, -236.019004, -230.104818, -224.57576, -219.415357, -214.607139, -210.134634, -205.981373, -202.130883, -198.566694, -195.272334, -192.231334, -189.42722, -186.843524, -184.463773, -182.271497, -180.250224, -178.383484, -176.654805, -175.047717, -173.545748, -172.132428, -170.791285, -169.505849, -168.259648, -167.036212, -165.819069, -164.591748, -163.337779, -161.493993, -159.328767, -157.117048, -154.860106, -152.559212, -150.215637, -147.830651, -145.405524, -142.941526, -140.439927, -137.902, -135.329012, -132.722235, -130.08294, -127.412396, -124.711875, -121.982645, -119.239986, -116.470856, -113.676222, -110.857051, -108.01431, -105.148965, -102.261982, -99.3543296, -96.4269731, -93.4808796, -90.5170157, -87.5363483, -84.539844, -81.5284695, -78.5031915, -75.4649767, -72.3193984, -68.9743202, -65.6274507, -62.2859405, -58.9569401, -55.6476004, -52.365072, -49.1165054, -45.9090515, -42.7498609, -39.6460841, -36.604872, -33.6333752, -30.7387442, -27.9281299, -25.2086829, -22.5875538, -21.1749006, -20.4031702, -19.6915518, -19.0114014, -18.3340751, -17.6309288, -16.8733183, -16.0325997, -15.080129, -13.9872621, -12.7253549, -11.2657634, -9.57984354, -7.63895133, -5.4144427, -2.8776736, 0, 13.379153, 25.7634313, 37.2018575, 47.7434545, 57.4372451, 66.332252, 74.477498, 81.922006, 88.7147986, 94.9048988, 100.541329, 105.673113, 110.349272, 114.61883, 118.530809, 122.134232, 123.706132, 123.042388, 124.200099, 126.848576, 130.657128, 135.295067, 140.431702, 145.736344, 150.878303, 155.526891, 159.351416, 162.02119, 163.205523, 160.862539, 157.00017, 151.801082, 145.447941, 141.319278, 137.999825, 133.222089, 126.209976, 116.187394, 102.37825, 84.0064511, 57.7906485, 26.2224752, -10.9840027, -54.1147189, -103.455607, -157.897454, -219.470128, -288.808349, -366.546838, -453.320316, -548.337895, -655.402315, -776.705892, -913.901217, -1067.53895, -1238.66449, -1428.32324, -1639.48818, -1870.83905, -2118.95279, -2381.15837, -2659.85613, -2958.09404, -3276.45326, -3611.33514, -3961.05965, -4325.27959, -4702.66049, -5086.9807, -5474.78303, -5864.9128, -6251.25608, -6629.55865, -6995.56629, -7345.02479, -7576.5731, -7807.34054, -8057.3496, -8346.62276, -8695.18253, -9181.18188, -9727.75914, -10330.0237, -11013.8819, -11762.2527, -12545.2231, -13346.3536, -14149.2048, -14937.3374, -15694.3118, -15587.896, -16237.5508, -17214.6256, -18090.4699, -18436.4333, -17823.8651, -15824.1149, -14088.1572, -12802.2021},
        {-285.519624, -276.598986, -268.179278, -260.24358, -252.774968, -245.756523, -239.171321, -233.002443, -227.232966, -221.845968, -216.824528, -212.151725, -207.810637, -203.784343, -200.05592, -196.608448, -193.425005, -190.488668, -187.782518, -185.289632, -182.993088, -180.875965, -178.921342, -177.112297, -175.431908, -173.863255, -172.389414, -170.993465, -169.658487, -168.367557, -167.103755, -165.850157, -164.589844, -163.305894, -161.424407, -159.218667, -156.96865, -154.675508, -152.340397, -149.964468, -147.548876, -145.094774, -142.603315, -140.075653, -137.512941, -134.916333, -132.286982, -129.626041, -126.934664, -124.214005, -121.465216, -118.689925, -115.888801, -113.062987, -110.213627, -107.341864, -104.448841, -101.535701, -98.6035876, -95.6536433, -92.6870115, -89.7048355, -86.7082584, -83.6984233, -80.6764734, -77.6435519, -74.6008019, -71.4609403, -68.1388072, -64.8179181, -61.5051485, -58.207374, -54.93147, -51.684312, -48.4727754, -45.3037357, -42.1840685, -39.120649, -36.120353, -33.1900557, -30.3366327, -27.5669595, -24.8879115, -22.3063641, -20.9133073, -20.1507564, -19.4476966, -18.7758218, -18.1068258, -17.4124027, -16.6642462, -15.8340502, -14.8935086, -13.8143153, -12.5681642, -11.1267491, -9.46176394, -7.54490256, -5.34785886, -2.84232671, 0, 13.6596563, 26.253459, 37.8352687, 48.4589461, 58.1783519, 67.0473468, 75.1197913, 82.4495463, 89.0904723, 95.09643, 100.52128, 105.418883, 109.8431, 113.847791, 117.486818, 120.81404, 121.712194, 119.924981, 120.417137, 122.777281, 126.594034, 131.456015, 136.951844, 142.670141, 148.199526, 153.128618, 157.046038, 159.540405, 160.200339, 156.287741, 150.571081, 143.336992, 134.87211, 129.067738, 124.408177, 118.416484, 110.297893, 99.2576403, 84.5009599, 65.2330868, 38.1840813, 5.78766701, -32.2666477, -76.2893548, -126.590946, -182.607669, -245.74282, -316.525453, -395.484619, -483.149373, -578.192015, -685.267712, -806.894855, -944.923831, -1099.84119, -1272.7458, -1464.73656, -1679.57943, -1915.10006, -2166.21454, -2428.73314, -2705.94322, -3000.241, -3312.3995, -3640.44128, -3983.65139, -4343.97963, -4721.4019, -5106.12348, -5493.87645, -5879.03278, -6258.89396, -6629.66289, -6987.54248, -7328.73564, -7555.03459, -7780.65561, -8025.40427, -8309.08618, -8651.5069, -9131.35876, -9670.30279, -10262.9074, -10934.2939, -11666.937, -12430.5806, -13208.3358, -13983.3138, -14738.6255, -15457.3819, -15512.6556, -16157.7395, -17020.6206, -17729.2859, -17911.7225, -17195.9175, -15209.8577, -13387.7722, -12242.6913},
        {-290.875875, -281.587819, -272.81843, -264.550313, -256.766074, -249.448319, -242.579655, -236.142687, -230.12002, -224.494261, -219.248016, -214.363891, -209.824492, -205.612423, -201.710292, -198.100705, -194.766266, -191.689583, -188.853261, -186.239906, -183.832123, -181.612519, -179.5637, -177.668272, -175.90884, -174.26801, -172.728389, -171.272581, -169.883194, -168.542833, -167.234104, -165.939613, -164.641965, -163.323767, -161.399319, -159.147332, -156.853351, -154.518423, -152.143597, -149.729922, -147.278447, -144.790219, -142.266287, -139.7077, -137.115507, -134.490755, -131.834494, -129.147772, -126.431638, -123.687139, -120.915325, -118.105368, -115.27045, -112.411878, -109.530956, -106.628992, -103.70729, -100.767158, -97.8099001, -94.8368236, -91.849234, -88.8484375, -85.8357398, -82.8124471, -79.7798652, -76.7393003, -73.6920582, -70.5576033, -67.2573647, -63.9614409, -60.6764434, -57.4089835, -54.1656727, -50.9531225, -47.7779444, -44.6467497, -41.5661499, -38.5427565, -35.5831809, -32.6940346, -29.881929, -27.1534755, -24.5152857, -21.973971, -20.6002559, -19.8463131, -19.1515255, -18.4879721, -17.8277317, -17.1428834, -16.405506, -15.5876786, -14.6614799, -13.598989, -12.3722848, -10.9534462, -9.31455209, -7.42768151, -5.26491335, -2.79832653, 0, 13.8789418, 26.6290568, 38.3084928, 48.9753979, 58.6879203, 67.5042078, 75.4824086, 82.6806707, 89.1571421, 94.9699708, 100.177305, 104.837293, 109.008082, 112.74782, 116.114656, 119.166738, 119.433291, 116.611189, 116.478762, 118.552244, 122.347872, 127.381883, 133.170512, 139.229996, 145.076571, 150.226473, 154.195938, 156.501203, 156.658503, 151.337355, 143.944513, 134.850227, 124.424752, 116.921137, 110.768236, 103.359488, 93.9003053, 81.5961005, 65.6522861, 45.2742744, 17.3301282, -15.9260231, -54.8314596, -99.7234613, -150.939308, -208.401844, -272.966394, -345.073848, -425.165094, -513.681021, -608.840986, -716.022617, -838.019274, -976.849696, -1132.9511, -1307.47079, -1501.55605, -1719.68474, -1958.91663, -2212.67814, -2475.39917, -2751.20229, -3041.96404, -3348.64676, -3670.66376, -4008.13936, -4365.00831, -4742.38283, -5127.40365, -5515.11458, -5896.18265, -6270.70177, -6635.2308, -6986.32857, -7320.55391, -7542.70326, -7764.03846, -8004.05895, -8282.26418, -8618.15357, -9090.00725, -9619.35686, -10200.5101, -10859.2858, -11577.5185, -12323.9133, -13080.9611, -13831.1531, -14556.9803, -15240.9338, -15529.6404, -16188.2484, -16920.3879, -17429.6894, -17419.7833, -16594.2999, -14656.8696, -12770.2539, -11749.8193},
        {-296.320558, -286.612141, -277.44513, -268.801454, -260.663047, -253.011838, -245.829759, -239.098741, -232.800714, -226.91761, -221.43136, -216.323895, -211.577146, -207.173044, -203.093519, -199.320503, -195.835928, -192.621723, -189.659821, -186.932151, -184.420646, -182.107235, -179.973851, -178.002424, -176.174885, -174.473165, -172.879196, -171.374908, -169.942232, -168.563099, -167.219441, -165.893188, -164.566272, -163.220622, -161.245808, -158.932401, -156.578094, -154.18404, -151.751394, -149.281311, -146.774945, -144.23345, -141.657981, -139.049691, -136.409736, -133.73927, -131.039446, -128.311421, -125.556347, -122.775379, -119.969671, -117.134286, -114.276601, -111.397905, -108.499482, -105.582621, -102.648607, -99.6987259, -96.7342653, -93.7565112, -90.7667501, -87.7662685, -84.7563528, -81.7382895, -78.713365, -75.6828658, -72.6480783, -69.5306512, -66.254145, -62.9840934, -59.7269455, -56.4891502, -53.2771566, -50.0974139, -46.9563709, -43.8604769, -40.8161808, -37.8299317, -34.9081787, -32.0573708, -29.283957, -26.5943865, -23.9951082, -21.4925713, -20.1384565, -19.394612, -18.7103229, -18.0581187, -17.4105286, -16.7400818, -16.0193076, -15.2207353, -14.3168942, -13.2803135, -12.0835225, -10.6990505, -9.09942681, -7.25718062, -5.14484124, -2.73493794, 0, 14.2004791, 27.1852282, 39.0179267, 49.7622542, 59.4818901, 68.2405141, 76.1018056, 83.1294441, 89.3871091, 94.9384801, 99.8472367, 104.177058, 107.991625, 111.354615, 114.329709, 116.980586, 116.749698, 113.326265, 112.702932, 114.381685, 117.864514, 122.653407, 128.250351, 134.157334, 139.876345, 144.909372, 148.758402, 150.925424, 150.912426, 145.46553, 137.853074, 128.403805, 117.446471, 108.941618, 101.402093, 92.3722669, 81.0913477, 66.7985437, 48.7330632, 26.1341147, -4.38324174, -40.1569958, -81.4345188, -128.463182, -181.490357, -238.986037, -303.419316, -375.48191, -455.865535, -545.261907, -642.802119, -752.645939, -877.540851, -1019.3318, -1178.02107, -1354.43824, -1549.41293, -1766.29532, -2002.82156, -2253.97801, -2515.32736, -2790.34341, -3081.26689, -3388.93769, -3711.82211, -4049.47594, -4404.45068, -4777.07879, -5156.70862, -5538.90177, -5918.39143, -6291.69503, -6654.6609, -7003.13741, -7332.97287, -7543.83733, -7751.80202, -7976.75986, -8238.60376, -8557.22666, -9010.9432, -9522.27674, -10086.0823, -10735.4866, -11450.1954, -12193.968, -12947.3077, -13690.7176, -14404.7007, -15069.7604, -15712.4641, -16421.0343, -16999.1583, -17250.5234, -16978.8166, -15987.7253, -14080.9367, -12295.5646, -11299.3173},
        {-302.119652, -291.985791, -282.415583, -273.390304, -264.891233, -256.899648, -249.396828, -242.36405, -235.782592, -229.633731, -223.898748, -218.558918, -213.59552, -208.989833, -204.723133, -200.7767, -197.131811, -193.769745, -190.671778, -187.81919, -185.193258, -182.77526, -180.546474, -178.488179, -176.581652, -174.808171, -173.149014, -171.585459, -170.098785, -168.670269, -167.281189, -165.912824, -164.546451, -163.163348, -161.128689, -158.744763, -156.321265, -153.859466, -151.360637, -148.826051, -146.256979, -143.654694, -141.020466, -138.355568, -135.661271, -132.938848, -130.189569, -127.414707, -124.615534, -121.793321, -118.94934, -116.085878, -113.203169, -110.302464, -107.385012, -104.452063, -101.504866, -98.5446717, -95.5727295, -92.5902892, -89.5986006, -86.5989134, -83.5924774, -80.5805424, -77.564358, -74.5451741, -71.5242405, -68.4251535, -65.1733741, -61.9303062, -58.7022334, -55.4954395, -52.3162082, -49.1708232, -46.0655682, -43.0067268, -40.0005829, -37.0534202, -34.1715222, -31.3611728, -28.6286557, -25.9802545, -23.422253, -20.9609349, -19.6265452, -18.8918493, -18.2173054, -17.5759677, -16.9408903, -16.2851276, -15.5817336, -14.8037626, -13.9242688, -12.9163063, -11.7529294, -10.4071923, -8.85214913, -7.06085412, -5.00636147, -2.66172536, 0, 14.3961857, 27.5102999, 39.4099838, 50.1628788, 59.836626, 68.4988667, 76.2172421, 83.0593936, 89.0929623, 94.3855895, 99.0049164, 103.018584, 106.494235, 109.499508, 112.102047, 114.369491, 113.733893, 109.886381, 108.856049, 110.145769, 113.258414, 117.696854, 122.963963, 128.562612, 133.995674, 138.76602, 142.376522, 144.330052, 144.129483, 138.741833, 131.13564, 121.574533, 110.322139, 100.901184, 91.9457472, 81.2208151, 68.0122866, 51.6060604, 31.2880353, 6.34410998, -26.7729735, -65.0818945, -108.742441, -157.914401, -212.757562, -270.238386, -334.508318, -406.52548, -487.24799, -577.633969, -677.825869, -790.594403, -918.579717, -1063.41018, -1224.64846, -1402.78466, -1598.30892, -1813.26878, -2046.24297, -2294.1112, -2553.8394, -2828.12973, -3119.78866, -3429.51778, -3754.45218, -4093.36404, -4446.91588, -4814.36998, -5188.05731, -5564.2296, -5942.60401, -6315.43366, -6677.77067, -7024.66717, -7351.17529, -7551.96691, -7747.56947, -7958.13016, -8203.79617, -8504.71471, -8938.07229, -9428.95054, -9973.05122, -10612.0501, -11323.451, -12065.6017, -12816.9636, -13555.9983, -14261.1672, -14910.9319, -15926.8242, -16693.8547, -17119.3303, -17110.5575, -16574.843, -15419.4936, -13551.8159, -11890.8976, -10899.5913},
        {-308.313295, -297.761209, -287.793198, -278.389958, -269.532189, -261.200588, -253.375854, -246.038684, -239.169777, -232.74983, -226.759543, -221.179613, -215.990738, -211.173616, -206.708945, -202.577425, -198.759752, -195.236624, -191.988741, -188.9968, -186.241499, -183.703536, -181.36361, -179.202419, -177.20066, -175.339032, -173.598232, -171.95896, -170.401914, -168.90779, -167.457288, -166.031106, -164.609941, -163.174492, -161.068406, -158.604306, -156.102473, -153.564277, -150.99109, -148.384285, -145.745232, -143.075304, -140.375871, -137.648307, -134.893981, -132.114267, -129.310535, -126.484157, -123.636506, -120.768951, -117.882867, -114.986694, -112.074581, -109.147745, -106.207405, -103.25478, -100.291087, -97.317545, -94.3353726, -91.3457879, -88.3500093, -85.3492552, -82.3447439, -79.3376937, -76.3293231, -73.3208504, -70.313494, -67.2332576, -64.0079511, -60.793917, -57.5972495, -54.424043, -51.2803915, -48.1723893, -45.1061306, -42.0877097, -39.1232206, -36.2187576, -33.380415, -30.614287, -27.9264677, -25.3230513, -22.8101321, -20.3938043, -19.0794072, -18.3523946, -17.6862104, -17.0545212, -16.4309939, -15.7892952, -15.1030919, -14.3460509, -13.4918388, -12.5141225, -11.3865688, -10.0828444, -8.57661614, -6.84155079, -4.85131512, -2.57957593, 0, 14.4053528, 27.4946762, 39.3371597, 50.0019924, 59.5583639, 68.0754634, 75.6224803, 82.2686039, 88.0830237, 93.134929, 97.4935091, 101.227953, 104.407451, 107.101192, 109.378365, 111.308159, 110.376193, 106.282574, 104.940285, 105.864891, 108.571961, 112.57706, 117.395756, 122.543616, 127.536206, 131.889094, 135.117846, 136.73803, 136.265211, 130.972765, 123.440588, 113.856904, 102.409939, 92.1155612, 81.7481729, 69.3281385, 54.1953893, 35.6898567, 13.1514721, -14.079833, -49.5684191, -90.1861481, -136.024858, -187.176385, -243.732568, -301.383168, -365.722618, -437.943274, -519.237494, -610.797635, -713.700341, -829.395107, -960.386485, -1108.0812, -1271.64459, -1451.24831, -1647.06405, -1859.8631, -2089.08122, -2333.50005, -2591.49043, -2865.02709, -3157.61446, -3469.83726, -3797.33289, -4138.00962, -4490.46783, -4852.79523, -5220.54195, -5590.69353, -5967.93856, -6340.37444, -6702.32027, -7048.09513, -7372.01813, -7565.20794, -7750.98418, -7949.46605, -8180.77275, -8465.02348, -8876.7902, -9345.43751, -9867.7477, -10494.305, -11200.595, -11939.6862, -12688.1849, -13422.6979, -14119.8316, -14756.1927, -16127.2945, -16953.1277, -17238.5545, -16988.4375, -16207.6389, -14901.0214, -13073.4472, -11540.1998, -10545.0803},
        {-314.941622, -303.990833, -293.641384, -283.873516, -274.667472, -266.003496, -257.861829, -250.222714, -243.066395, -236.373113, -230.123112, -224.296634, -218.873922, -213.835218, -209.160765, -204.830805, -200.825582, -197.125339, -193.710317, -190.560759, -187.656908, -184.979007, -182.507299, -180.222025, -178.103429, -176.131754, -174.287242, -172.550135, -170.900677, -169.319109, -167.785676, -166.280619, -164.784181, -163.276604, -161.085405, -158.530918, -155.941325, -153.31805, -150.662519, -147.976155, -145.260385, -142.516633, -139.746324, -136.950883, -134.131734, -131.290303, -128.428015, -125.546295, -122.646566, -119.730256, -116.798787, -113.863284, -110.915262, -107.955937, -104.986523, -102.008235, -99.0222876, -96.0298959, -93.0322742, -90.0306373, -87.0262, -84.0201768, -81.0137825, -78.0082318, -75.0047394, -72.00452, -69.0087882, -65.9471108, -62.7507748, -59.5687634, -56.406936, -53.2711519, -50.1672705, -47.101151, -44.0786529, -41.1056355, -38.1879581, -35.33148, -32.5420607, -29.8255595, -27.1878356, -24.6347485, -22.1721574, -19.8059218, -18.5119275, -17.7906174, -17.1307754, -16.5067816, -15.8930161, -15.263859, -14.5936904, -13.8568905, -13.0278394, -12.0809171, -10.9905039, -9.73097978, -8.2767249, -6.60211936, -4.68154329, -2.48937679, 0, 14.1672716, 27.0287614, 38.6519495, 49.1043156, 58.4533398, 66.7665019, 74.111282, 80.5551598, 86.1656154, 91.0101287, 95.1561796, 98.671248, 101.622814, 104.078357, 106.105358, 107.771296, 106.666917, 102.505884, 100.95781, 101.559447, 103.847546, 107.358859, 111.630137, 116.198131, 120.599592, 124.371273, 127.049925, 128.172298, 127.275144, 121.964824, 114.41629, 104.745411, 93.068055, 81.9004737, 70.1583444, 56.1172428, 39.1729228, 18.7211384, -5.8423564, -35.1218076, -72.4989307, -114.955185, -162.5514, -215.348402, -273.40702, -331.64466, -396.551428, -469.474007, -551.759081, -644.753333, -750.21364, -868.575147, -1002.21176, -1152.34123, -1317.82079, -1498.56749, -1694.49839, -1905.33627, -2131.23662, -2372.56688, -2628.83563, -2901.50132, -3194.82942, -3509.34638, -3839.24312, -4181.61864, -4533.17042, -4890.89338, -5253.25478, -5617.88906, -5993.51324, -6364.97413, -6726.06986, -7070.59855, -7392.35836, -7581.67639, -7761.68955, -7952.06374, -8172.46483, -8442.55872, -8832.49265, -9277.79692, -9776.50257, -10387.5795, -11084.937, -11817.0934, -12559.2277, -13286.5188, -13974.1457, -14597.2873, -16268.449, -17145.2712, -17314.4818, -16862.809, -15876.9808, -14443.7254, -12649.7709, -11227.418, -10230.223},
        {-322.044773, -310.727104, -300.02355, -289.914074, -280.378639, -271.397209, -262.949745, -255.016212, -247.576571, -240.610786, -234.09882, -228.020636, -222.356196, -217.085463, -212.1884, -207.644971, -203.435138, -199.538864, -195.936112, -192.606844, -189.531024, -186.688615, -184.05958, -181.623881, -179.361481, -177.252344, -175.276431, -173.413707, -171.644134, -169.947674, -168.304292, -166.693949, -165.096608, -163.492233, -161.200131, -158.544487, -155.85743, -153.140363, -150.394687, -147.621806, -144.823121, -142.000036, -139.153952, -136.286272, -133.398399, -130.491735, -127.567683, -124.627644, -121.673022, -118.705219, -115.725638, -112.742196, -109.749639, -106.749228, -103.742225, -100.729891, -97.7134871, -94.6942744, -91.6735142, -88.6524678, -85.6323965, -82.6145613, -79.6002237, -76.5906447, -73.5870857, -70.5908079, -67.6030724, -64.5588605, -61.3947444, -58.2486833, -55.1262351, -52.0329578, -48.9744091, -45.9561471, -42.9837295, -40.0627145, -37.1986597, -34.3971232, -31.6636628, -29.0038365, -26.4232021, -23.9273176, -21.5217409, -19.2120298, -17.9389914, -17.2208872, -16.564738, -15.9457512, -15.3391338, -14.7200933, -14.063837, -13.345572, -12.5405056, -11.6238452, -10.5707981, -9.35657138, -7.95637247, -6.34540859, -4.49888704, -2.39201508, 0, 13.6212333, 26.0029598, 37.2068487, 47.294569, 56.3277898, 64.3681802, 71.4774092, 77.7171459, 83.1490594, 87.8348187, 91.8360929, 95.2145511, 98.0318624, 100.349696, 102.22972, 103.733605, 102.596384, 98.547348, 96.9107961, 97.2498323, 99.1275606, 102.107086, 105.751511, 109.623942, 113.287482, 116.305235, 118.240306, 118.655799, 117.114819, 111.524511, 103.711123, 93.7345455, 81.6546702, 69.5716461, 56.525236, 41.011134, 22.4771544, 0.371111502, -25.8591805, -56.7659073, -95.2938601, -138.874435, -187.591699, -241.529719, -300.772563, -360.24714, -426.483964, -500.856392, -584.737783, -679.501492, -787.153872, -907.661623, -1043.30616, -1195.18663, -1361.98837, -1543.48046, -1739.43198, -1948.9463, -2172.60953, -2411.73404, -2666.43013, -2938.01825, -3231.51865, -3547.49538, -3878.96177, -4222.39704, -4573.08755, -4927.20327, -5285.28806, -5645.41166, -6018.44623, -6387.68949, -6746.77961, -7089.35471, -7409.05295, -7599.48822, -7779.32898, -7967.21942, -8181.80376, -8441.7262, -8810.57531, -9232.08802, -9705.64668, -10297.202, -10979.7864, -11698.6954, -12428.348, -13143.1636, -13817.5613, -14425.9602, -16304.8615, -17216.7032, -17304.7628, -16712.3177, -15582.6451, -14059.0225, -12284.7272, -10936.4993, -9949.45818},
        {-329.973393, -318.249723, -307.154904, -296.668566, -286.770335, -277.439841, -268.656712, -260.400577, -252.651063, -245.387799, -238.590413, -232.238535, -226.311791, -220.789811, -215.652223, -210.878655, -206.448736, -202.342093, -198.538356, -195.017153, -191.758112, -188.740861, -185.945029, -183.350244, -180.936134, -178.682329, -176.568456, -174.574143, -172.679019, -170.862713, -169.104852, -167.385065, -165.682981, -163.978228, -161.582178, -158.822644, -156.03547, -153.22197, -150.383458, -147.521249, -144.636658, -141.730999, -138.805587, -135.861737, -132.900762, -129.923977, -126.932697, -123.928237, -120.91191, -117.885032, -114.848917, -111.798193, -108.741006, -105.678816, -102.613081, -99.5452621, -96.4768181, -93.4092086, -90.3438931, -87.282331, -84.2259819, -81.1763052, -78.1347605, -75.1028071, -72.0819047, -69.0735126, -66.0790905, -63.0434589, -59.9127985, -56.805382, -53.7263405, -50.6808052, -47.6739073, -44.7107779, -41.7965481, -38.9363491, -36.1353121, -33.3985681, -30.7312485, -28.1384842, -25.6254065, -23.1971465, -20.8588353, -18.6156041, -17.3620411, -16.6398404, -15.9800055, -15.358803, -14.7524994, -14.1373612, -13.4896552, -12.7856477, -12.0016054, -11.1137948, -10.0984825, -8.93193501, -7.59041896, -6.05020087, -4.28754731, -2.27872483, 0, 12.3153527, 23.5943358, 33.8831732, 43.2280884, 51.6753054, 59.2710478, 66.0615393, 72.0930037, 77.4116647, 82.063746, 86.0954714, 89.5530646, 92.4827493, 94.9307493, 96.9432883, 98.56659, 97.6672392, 93.9800827, 92.4814799, 92.750589, 94.3665684, 96.9085762, 99.9557706, 103.08731, 105.882353, 107.920057, 108.77958, 108.040082, 105.280719, 98.3950901, 89.2659526, 77.9781334, 64.6164592, 51.2902662, 37.0721262, 20.4947417, 1.03558618, -21.8278665, -48.6181428, -79.8577691, -117.357185, -159.959029, -207.933846, -261.552179, -321.084571, -383.976111, -454.02916, -532.220626, -619.527415, -716.926433, -825.165002, -945.73022, -1081.09468, -1232.53753, -1398.90129, -1580.12248, -1776.1376, -1986.88986, -2212.30756, -2452.31176, -2705.46751, -2974.62726, -3265.9199, -3580.41185, -3910.59143, -4252.88446, -4602.68243, -4956.14299, -5313.21636, -5671.70746, -6038.66708, -6401.23185, -6754.00624, -7091.59468, -7408.60163, -7614.12387, -7809.65042, -8011.16264, -8234.6419, -8496.06954, -8846.49952, -9243.45889, -9687.8979, -10248.6282, -10899.7382, -11587.0408, -12286.0382, -12972.2328, -13621.1266, -14208.2221, -16058.3898, -16967.0902, -17065.8148, -16486.0547, -15359.3011, -13817.0454, -11990.7787, -10593.8263, -9693.17224},
        {-338.221334, -326.211299, -314.831791, -304.062595, -293.883498, -284.274286, -275.214745, -266.684661, -258.663819, -251.132006, -244.069009, -237.454612, -231.268602, -225.490765, -220.100888, -215.078755, -210.404154, -206.05687, -202.016689, -198.263397, -194.77678, -191.536625, -188.522717, -185.714843, -183.092788, -180.636338, -178.32528, -176.139399, -174.058482, -172.062315, -170.130683, -168.243372, -166.38017, -164.520861, -161.989591, -159.107756, -156.204635, -153.281289, -150.338783, -147.378178, -144.400538, -141.406925, -138.398402, -135.376033, -132.340879, -129.294004, -126.236471, -123.169342, -120.093681, -117.01055, -113.921012, -110.797069, -107.669475, -104.539923, -101.410105, -98.2817143, -95.1564433, -92.0359851, -88.9220323, -85.8162779, -82.7204147, -79.6361354, -76.5651328, -73.5090999, -70.4697293, -67.4487139, -64.4477466, -61.423305, -58.3329528, -55.2716352, -52.243976, -49.2545992, -46.3081288, -43.4091885, -40.5624024, -37.7723944, -35.0437883, -32.381208, -29.7892776, -27.2726208, -24.8358616, -22.483624, -20.2205318, -18.0512089, -16.8196122, -16.0930449, -15.428722, -14.8040296, -14.1963535, -13.5830797, -12.9415942, -12.2492828, -11.4835316, -10.6217265, -9.64125336, -8.51949821, -7.23384697, -5.76168558, -4.08039999, -2.16737615, 0, 10.8771282, 20.948292, 30.2420411, 38.7869252, 46.6114943, 53.7442979, 60.213886, 66.0488081, 71.2776143, 75.9288541, 80.0310773, 83.6128338, 86.7026733, 89.3291455, 91.5208003, 93.3061873, 92.7277753, 89.5303882, 88.234902, 88.4442777, 89.7614762, 91.7894584, 94.1311854, 96.3896181, 98.1677175, 99.0684446, 98.6947603, 96.6496255, 92.5360013, 84.2536397, 73.7332206, 61.0886674, 46.4339038, 31.9219221, 16.6471116, -0.83989005, -21.0368799, -44.441655, -71.5520122, -102.865749, -138.990697, -180.28113, -227.213314, -280.263519, -339.908013, -406.34645, -480.400865, -562.616681, -653.539318, -753.714198, -862.529397, -983.102216, -1118.03978, -1268.80242, -1434.50911, -1615.33005, -1811.43544, -2023.81078, -2251.62568, -2493.16028, -2745.37973, -3012.52689, -3301.36844, -3613.39165, -3941.13816, -4281.25579, -4629.87244, -4983.50113, -5340.56131, -5698.39403, -6058.90375, -6414.16608, -6759.94247, -7091.99435, -7406.08316, -7628.33606, -7841.55733, -8058.91698, -8293.58501, -8558.73143, -8892.76568, -9266.79204, -9683.16361, -10211.7762, -10828.2473, -11480.0515, -12144.2135, -12797.7577, -13417.7089, -13981.0915, -15717.0985, -16601.0345, -16733.3916, -16214.6619, -15145.3373, -13625.91, -11756.8718, -10284.6203, -9462.93352},
        {-346.578053, -334.484553, -323.002197, -312.111716, -301.793837, -292.02929, -282.798804, -274.083108, -265.86293, -258.119, -250.832047, -243.982799, -237.551986, -231.520337, -225.86858, -220.577445, -215.62766, -210.999955, -206.675059, -202.6337, -198.856607, -195.324511, -192.018138, -188.918219, -186.005483, -183.260657, -180.664473, -178.197657, -175.84094, -173.575051, -171.380717, -169.238669, -167.129636, -165.034345, -162.321761, -159.291363, -156.249983, -153.198204, -150.136607, -147.065775, -143.986288, -140.898731, -137.803683, -134.701728, -131.593448, -128.479423, -125.360236, -122.23647, -119.108706, -115.977526, -112.843512, -109.645497, -106.447151, -103.250394, -100.057146, -96.8693291, -93.6888627, -90.5176676, -87.3576644, -84.2107737, -81.0789159, -77.9640116, -74.8679815, -71.792746, -68.7402258, -65.7123412, -62.711013, -59.7054402, -56.6656074, -53.6609209, -50.6954224, -47.7731536, -44.8981563, -42.0744723, -39.3061434, -36.5972112, -33.9517176, -31.3737043, -28.867213, -26.4362855, -24.0849636, -21.8172891, -19.6373036, -17.5490489, -16.3438273, -15.6154912, -14.948431, -14.3211404, -13.712113, -13.0998425, -12.4628224, -11.7795465, -11.0285083, -10.1882016, -9.23711987, -8.15375688, -6.91660624, -5.50416159, -3.89491657, -2.06736483, 0, 9.56479469, 18.5371489, 26.9286502, 34.7508859, 42.0154434, 48.7339102, 54.9178737, 60.5789213, 65.7286405, 70.3786187, 74.5404433, 78.2257017, 81.4459814, 84.2128698, 86.5379543, 88.4328224, 88.1821405, 85.5507932, 84.4564983, 84.540789, 85.4451981, 86.8112587, 88.2805035, 89.4944657, 90.0946781, 89.7226736, 88.0199853, 84.6281461, 79.1886889, 69.7531933, 58.1361286, 44.4560136, 28.8313674, 13.3519364, -2.84937421, -21.1653198, -42.0687496, -66.0325128, -93.5294588, -125.032437, -160.137362, -200.460911, -246.647506, -299.341571, -359.18753, -428.474456, -505.790959, -591.370299, -685.445738, -788.250536, -898.273452, -1019.62442, -1154.84226, -1305.4502, -1470.8979, -1651.56598, -1847.83509, -2061.47089, -2291.15422, -2534.05503, -2786.30485, -3052.64092, -3339.55568, -3648.63968, -3973.17764, -4310.26678, -4657.20548, -5011.14311, -5368.49155, -5726.07984, -6080.92778, -6429.64432, -6768.99245, -7095.7352, -7406.63557, -7642.20155, -7869.51492, -8099.40247, -8342.69094, -8610.2071, -8930.18278, -9284.43632, -9676.33518, -10175.6429, -10759.4451, -11377.1626, -12007.2641, -12628.2188, -13218.4954, -13756.5629, -15338.647, -16177.9969, -16341.6662, -15896.7085, -14910.1773, -13449.1261, -11580.6085, -10039.9739, -9255.77765},
        {-354.681743, -342.941921, -331.749773, -321.08861, -310.941738, -301.292466, -292.124102, -283.419954, -275.163332, -267.337542, -259.925894, -252.911696, -246.278255, -240.008881, -234.086881, -228.495564, -223.218238, -218.238211, -213.538792, -209.103289, -204.91501, -200.957263, -197.213358, -193.666601, -190.300301, -187.097767, -184.042307, -181.117229, -178.305841, -175.591452, -172.95737, -170.386904, -167.86336, -165.370049, -162.38665, -159.158879, -155.938504, -152.725177, -149.51855, -146.318275, -143.124004, -139.935389, -136.752082, -133.573734, -130.399999, -127.230527, -124.064971, -120.902983, -117.744215, -114.588318, -111.434945, -108.171497, -104.91231, -101.659469, -98.4150601, -95.1811678, -91.9598777, -88.7532751, -85.5634454, -82.3924738, -79.2424458, -76.1154465, -73.0135614, -69.9388757, -66.8934749, -63.8794441, -60.8988688, -57.9350235, -54.9716344, -52.0495827, -49.172173, -46.3427101, -43.5644987, -40.8408437, -38.1750498, -35.5704216, -33.030264, -30.5578818, -28.1565796, -25.8296622, -23.5804343, -21.4122008, -19.3282663, -17.3319357, -16.1610158, -15.436111, -14.7689976, -14.1391443, -13.5260199, -12.9090931, -12.2678327, -11.5817073, -10.8301857, -9.99273674, -9.04882906, -7.97793144, -6.7595126, -5.3730413, -3.79798627, -2.01381626, 0, 8.91931416, 17.3575728, 25.3152923, 32.7929887, 39.7911784, 46.3103777, 52.3511028, 57.9138699, 62.9991953, 67.6075953, 71.7395861, 75.3956841, 78.5764053, 81.2822661, 83.5137828, 85.2714717, 85.2633989, 83.305445, 82.3215251, 82.0352019, 82.170038, 82.4495961, 82.597439, 82.3371292, 81.3922296, 79.4863028, 76.3429114, 71.6856182, 65.2379859, 55.7165431, 44.2211323, 30.7774266, 15.4110987, -0.326036484, -17.1713771, -36.2692921, -58.0519513, -82.9515241, -111.40018, -143.830089, -179.456356, -220.298626, -267.02719, -320.312339, -380.824364, -450.405218, -528.260614, -614.767929, -710.304537, -815.247816, -928.927206, -1054.04882, -1192.39675, -1345.12565, -1512.10518, -1693.78191, -1890.60246, -2104.07365, -2333.34088, -2576.39288, -2830.95921, -3099.67355, -3386.44891, -3692.10039, -4012.19399, -4344.70515, -4688.85318, -5042.93601, -5400.6906, -5758.43386, -6110.39043, -6455.29135, -6790.17776, -7112.09083, -7418.07168, -7652.17421, -7877.67363, -8104.85791, -8344.015, -8605.43285, -8915.68959, -9257.92292, -9635.43928, -10116.993, -10679.984, -11274.5425, -11879.432, -12473.416, -13035.258, -13543.7216, -14926.0254, -15669.2172, -15821.0937, -15429.4514, -14542.0867, -13206.7962, -11471.3765, -9938.78231, -9050.88091},
        {-362.793212, -351.588052, -340.848519, -330.561191, -320.712646, -311.289461, -302.278213, -293.66548, -285.437839, -277.581869, -270.084146, -262.931248, -256.109752, -249.606236, -243.407278, -237.499455, -231.869345, -226.503524, -221.388571, -216.511063, -211.857577, -207.414692, -203.168984, -199.107031, -195.215411, -191.4807, -187.889477, -184.42832, -181.083804, -177.842509, -174.691012, -171.615889, -168.603719, -165.641079, -162.346288, -158.897799, -155.47849, -152.08689, -148.721527, -145.380929, -142.063623, -138.76814, -135.493006, -132.23675, -128.9979, -125.774984, -122.56653, -119.371068, -116.187124, -113.013227, -109.847905, -106.519177, -103.199778, -99.8919305, -96.5978602, -93.3197915, -90.0599489, -86.8205569, -83.6038399, -80.4120226, -77.2473294, -74.1119849, -71.0082136, -67.93824, -64.9042885, -61.9085839, -58.9533504, -56.0356111, -53.1525136, -50.3170106, -47.5316637, -44.7990346, -42.1216852, -39.502177, -36.9430719, -34.4469315, -32.0163175, -29.6537918, -27.361916, -25.1432518, -23.0003609, -20.9358052, -18.9521462, -17.0519458, -15.9208271, -15.2054539, -14.5442799, -13.9177002, -13.30611, -12.6899041, -12.0494778, -11.365226, -10.6175438, -9.78682628, -8.85346853, -7.79786558, -6.60041249, -5.24150432, -3.70153615, -1.96090302, 0, 8.42728518, 16.4578245, 24.0828633, 31.2936465, 38.0814195, 44.4374274, 50.3529154, 55.8191287, 60.8273125, 65.3687119, 69.4345722, 73.0161384, 76.104656, 78.6913699, 80.7675253, 82.3243676, 82.4944641, 81.1463917, 80.2136429, 79.5034605, 78.8230874, 77.9797664, 76.7807402, 75.0332516, 72.5445436, 69.1218588, 64.5724402, 58.7035304, 51.3223724, 41.8545226, 30.6288617, 17.5671386, 2.59110186, -13.3110931, -30.750901, -50.6241514, -73.3290177, -99.263673, -128.826291, -162.415044, -199.179478, -241.146411, -288.958314, -343.257657, -404.686911, -473.911946, -551.545982, -638.225639, -734.587537, -841.268296, -958.8283, -1088.07358, -1229.97472, -1385.25063, -1554.10635, -1736.97766, -1934.30032, -2146.90015, -2374.73422, -2617.33406, -2874.87537, -3147.01391, -3434.42173, -3736.84267, -4052.4479, -4380.13046, -4720.99289, -5074.50107, -5432.01946, -5789.49515, -6139.73443, -6482.30839, -6814.2513, -7132.59744, -7434.38112, -7663.79716, -7883.92916, -8105.02124, -8337.31755, -8591.06224, -8893.53159, -9226.58284, -9593.16061, -10060.059, -10604.949, -11178.8975, -11760.9056, -12329.9744, -12865.1049, -13345.2984, -14565.6474, -15219.5313, -15347.4354, -14989.8454, -14187.2466, -12980.1247, -11408.9651, -9886.62934, -8877.39399},
        {-370.95284, -360.380787, -350.182062, -340.346889, -330.865496, -321.728109, -312.924956, -304.446262, -296.282253, -288.423158, -280.859202, -273.580612, -266.577615, -259.840436, -253.359304, -247.124444, -241.126083, -235.354447, -229.799764, -224.452259, -219.302161, -214.339694, -209.555086, -204.938563, -200.480352, -196.170679, -191.999772, -187.957856, -184.035159, -180.221906, -176.508326, -172.884643, -169.341085, -165.867879, -162.237353, -158.553268, -154.921998, -151.340843, -147.8071, -144.318067, -140.871042, -137.463322, -134.092205, -130.754989, -127.448972, -124.171451, -120.919724, -117.69109, -114.482845, -111.292288, -108.116716, -104.720741, -101.339391, -97.9750073, -94.6299332, -91.3065101, -88.00708, -84.7339849, -81.4895668, -78.2761676, -75.0961293, -71.9517939, -68.8455034, -65.7795997, -62.7564248, -59.7783206, -56.8476293, -53.9738111, -51.1661557, -48.4123236, -45.7141955, -43.0736518, -40.492573, -37.9728396, -35.5163322, -33.1249312, -30.8005172, -28.5449706, -26.3601721, -24.2480021, -22.210341, -20.2490695, -18.366068, -16.5632171, -15.4770767, -14.7788381, -14.1320639, -13.5180381, -12.9180446, -12.3133674, -11.6852904, -11.0150976, -10.2840729, -9.47350029, -8.56466363, -7.53884689, -6.37733402, -5.06140895, -3.57235563, -1.891458, 0, 7.94859018, 15.5759729, 22.8651133, 29.7989768, 36.3605286, 42.5327341, 48.2985584, 53.6409669, 58.5429249, 62.9873976, 66.9573503, 70.4357483, 73.4055569, 75.8497414, 77.7512669, 79.0930989, 79.3577437, 78.4566724, 77.5048406, 76.3779727, 74.9517925, 73.1020241, 70.7043917, 67.6346194, 63.7684312, 58.9815511, 53.1497034, 46.148612, 37.854001, 28.2716401, 17.0995245, 4.17436446, -10.6671295, -26.8663881, -44.9476302, -65.6275708, -89.2860572, -116.302937, -147.058056, -181.931264, -220.273685, -263.806977, -313.112259, -368.77065, -431.363269, -500.156033, -577.374062, -663.927275, -760.725593, -868.678933, -989.550195, -1122.35379, -1267.42232, -1425.09608, -1595.73111, -1779.67635, -1977.28078, -2188.39033, -2413.97132, -2655.53875, -2916.12921, -3192.05708, -3480.66041, -3780.49385, -4092.17355, -4415.3694, -4752.63224, -5104.37887, -5460.46262, -5816.71211, -6166.07885, -6507.71121, -6838.43373, -7155.07096, -7454.44743, -7679.39559, -7894.2301, -8109.27355, -8334.8485, -8581.27751, -8877.04443, -9202.20305, -9559.29253, -10012.2684, -10539.3592, -11093.2032, -11653.3087, -12199.1835, -12710.3358, -13166.2736, -14314.3918, -14912.3154, -15009.6051, -14655.8219, -13900.5265, -12793.2799, -11383.6428, -9850.4576, -8746.77478},
        {-379.327632, -369.26209, -359.489632, -350.003689, -340.797687, -331.865057, -323.199228, -314.793628, -306.641687, -298.736834, -291.072498, -283.642107, -276.439091, -269.456879, -262.688899, -256.128582, -249.769356, -243.604649, -237.627892, -231.832512, -226.21194, -220.759603, -215.468932, -210.333355, -205.346301, -200.5012, -195.791479, -191.210569, -186.751899, -182.408896, -178.174991, -174.043613, -170.00819, -166.062151, -162.113397, -158.19992, -154.360741, -150.592064, -146.890093, -143.251034, -139.671091, -136.146469, -132.673372, -129.248005, -125.866572, -122.525278, -119.220328, -115.947926, -112.704278, -109.485586, -106.288057, -102.824216, -99.3800966, -95.958052, -92.5604367, -89.189605, -85.8479111, -82.5377092, -79.2613534, -76.0211981, -72.8195973, -69.6589053, -66.5414763, -63.4696645, -60.4458241, -57.4723093, -54.5514743, -51.6969907, -48.9222584, -46.2062904, -43.550707, -40.957129, -38.4271768, -35.9624709, -33.564632, -31.2352805, -28.976037, -26.7885221, -24.6743562, -22.6351599, -20.6725538, -18.7881584, -16.9835943, -15.2604819, -14.2229458, -13.5586816, -12.9481435, -12.3733996, -11.8165181, -11.2595669, -10.6846144, -10.0737285, -9.40897757, -8.67242957, -7.84615268, -6.91221503, -5.85268475, -4.64962998, -3.28511884, -1.74121947, 0, 6.81548975, 13.4748805, 19.9488505, 26.2080782, 32.2232419, 37.9650198, 43.4040902, 48.5111316, 53.2568222, 57.6118403, 61.5468643, 65.0325724, 68.039643, 70.5387544, 72.5005849, 73.8958129, 74.197836, 73.3062925, 72.316662, 71.0930628, 69.499613, 67.4004309, 64.6596345, 61.1413421, 56.7096719, 51.2287421, 44.5626709, 36.5755764, 27.1315768, 16.0085066, 3.18840503, -11.4387244, -27.9828785, -45.5068479, -64.6442322, -86.307886, -110.921968, -138.910636, -170.698048, -206.708364, -246.819167, -292.167538, -343.284575, -400.701375, -464.949035, -535.210007, -613.701195, -701.290855, -798.847245, -907.238623, -1027.96813, -1160.49317, -1305.7041, -1464.09538, -1635.3533, -1819.52701, -2016.66565, -2225.10503, -2446.99703, -2686.3626, -2949.24572, -3228.72374, -3519.71203, -3819.62168, -4130.09256, -4450.82346, -4783.87466, -5129.55719, -5479.52332, -5830.3232, -6178.601, -6520.7984, -6853.43303, -7173.02247, -7476.08436, -7707.12989, -7927.6847, -8147.26798, -8375.39897, -8621.59689, -8911.34854, -9227.56048, -9572.90866, -10006.1928, -10507.8731, -11035.0251, -11569.2784, -12092.2624, -12585.6068, -13030.9411, -14450.9473, -15140.8285, -15207.0619, -14756.1251, -13894.4954, -12728.6504, -11365.0674, -9679.99345, -8690.73196},
        {-387.751961, -378.215633, -368.889776, -359.771104, -350.856331, -342.142174, -333.625346, -325.302564, -317.170541, -309.225994, -301.465637, -293.886186, -286.484355, -279.256859, -272.200414, -265.311734, -258.587534, -252.024531, -245.619438, -239.36897, -233.269844, -227.318773, -221.512473, -215.847659, -210.321046, -204.929349, -199.669283, -194.537563, -189.530904, -184.646021, -179.87963, -175.228445, -170.689181, -166.258554, -161.983642, -157.834919, -153.782254, -149.820727, -145.945419, -142.15141, -138.433782, -134.787614, -131.207988, -127.689984, -124.228682, -120.819164, -117.45651, -114.135801, -110.852117, -107.600538, -104.376147, -100.836056, -97.3206389, -93.8323036, -90.3734568, -86.9465058, -83.5538575, -80.1979191, -76.8810978, -73.6058005, -70.3744345, -67.1894068, -64.0531246, -60.9679949, -57.9364249, -54.9608217, -52.0435923, -49.2030608, -46.4571693, -43.7754971, -41.1594196, -38.6103121, -36.1295499, -33.7185082, -31.3785623, -29.1110875, -26.917459, -24.7990521, -22.7572422, -20.7934045, -18.9089142, -17.1051466, -15.3834771, -13.7452808, -12.7616515, -12.1371218, -11.5693772, -11.0413045, -10.5357904, -10.0357218, -9.5239855, -8.98346834, -8.39705712, -7.74763867, -7.01809981, -6.19132736, -5.25020814, -4.17762898, -2.95647668, -1.56963808, 0, 5.746249, 11.4791692, 17.1591515, 22.7465869, 28.2018661, 33.48538, 38.5575197, 43.3786758, 47.9092394, 52.1096014, 55.9401525, 59.3612837, 62.333386, 64.81685, 66.7720669, 68.1594274, 68.3482393, 67.2144532, 66.0554335, 64.7049104, 62.9966142, 60.764275, 57.8416232, 54.0623889, 49.2603024, 43.2690939, 35.9224936, 27.0542319, 16.4980389, 3.6466943, -11.0634392, -27.6663462, -46.1960113, -65.2053515, -85.468885, -108.156082, -133.745246, -162.714684, -195.542699, -232.707597, -274.717996, -322.012661, -375.063966, -434.344286, -500.325994, -572.245379, -652.119921, -740.731017, -838.860062, -947.288452, -1067.03015, -1198.34974, -1342.84662, -1501.25528, -1672.54422, -1856.47483, -2052.80851, -2258.54129, -2476.82839, -2713.84184, -2978.07771, -3260.06144, -3553.05654, -3853.57728, -4163.64615, -4482.75726, -4811.94501, -5151.10286, -5494.47658, -5839.50659, -6187.34025, -6531.14075, -6867.06535, -7191.27133, -7499.91593, -7739.832, -7968.83232, -8195.40525, -8428.03916, -8675.22241, -8958.06922, -9264.02485, -9596.16658, -10008.352, -10483.5579, -10983.6024, -11492.5203, -11994.3462, -12473.1147, -12912.8606, -14671.3322, -15484.3022, -15523.4966, -14960.6415, -13967.4627, -12715.6863, -11377.0381, -9532.87726, -8687.77764},
        {-396.18132, -387.209904, -378.362918, -369.640483, -361.042721, -352.569752, -344.221697, -335.998677, -327.900813, -319.928225, -312.081035, -304.359363, -296.763331, -289.293059, -281.948667, -274.730278, -267.638011, -260.671988, -253.832329, -247.119156, -240.532589, -234.072749, -227.739757, -221.533735, -215.454801, -209.503079, -203.678688, -197.981749, -192.412384, -186.970712, -181.656856, -176.470935, -171.413072, -166.483386, -161.873452, -157.483637, -153.212008, -149.052471, -144.998936, -141.045309, -137.185498, -133.413412, -129.722958, -126.108044, -122.562578, -119.080468, -115.655621, -112.281945, -108.953349, -105.663739, -102.407025, -98.7778505, -95.1780419, -91.6101621, -88.0767741, -84.580441, -81.1237259, -77.7091918, -74.3394018, -71.0169189, -67.7443062, -64.5241267, -61.3589436, -58.2513199, -55.2038186, -52.2190028, -49.2994356, -46.471528, -43.7611187, -41.1215858, -38.5539464, -36.0592177, -33.6384166, -31.2925603, -29.0226659, -26.8297504, -24.714831, -22.6789247, -20.7230486, -18.8482198, -17.0554554, -15.3457725, -13.7201881, -12.1797193, -11.2558005, -10.6730835, -10.149845, -9.66988898, -9.21701947, -8.7750405, -8.32775612, -7.85897034, -7.35248718, -6.79211069, -6.16164489, -5.4448938, -4.62566145, -3.68775186, -2.61496908, -1.39111711, 0, 4.95444735, 9.98064771, 15.0327035, 20.0647171, 25.0307909, 29.8850274, 34.5815288, 39.0743977, 43.3177364, 47.2656473, 50.8722329, 54.0915954, 56.8778374, 59.1850612, 60.9673692, 62.1788639, 62.0541041, 60.4444005, 58.931394, 57.3149992, 55.3951306, 52.9717028, 49.8446302, 45.8138275, 40.679209, 34.2406894, 26.2981832, 16.6516049, 5.10086899, -9.33681482, -25.7918354, -44.2294667, -64.6149828, -85.0451928, -106.419603, -130.135978, -156.72013, -186.697873, -220.595021, -258.937386, -302.850141, -352.077328, -407.027494, -468.109188, -535.730957, -609.131302, -690.18133, -779.582101, -878.034674, -986.240109, -1104.75562, -1234.60192, -1377.94627, -1535.73788, -1706.43912, -1889.62891, -2084.88616, -2288.39871, -2503.90725, -2738.85186, -3003.08147, -3285.79439, -3579.64943, -3880.6983, -4190.74178, -4508.94176, -4835.1163, -5168.59739, -5506.31097, -5846.54395, -6194.73946, -6540.92731, -6880.91729, -7210.5192, -7525.54283, -7775.91539, -8014.79992, -8249.47688, -8487.22672, -8735.32989, -9009.23484, -9302.60869, -9619.23132, -10009.3174, -10458.2664, -10932.1299, -11417.2748, -11900.0678, -12366.8759, -12804.0658, -14883.7322, -15818.3799, -15839.6459, -15179.1676, -14068.5822, -12739.5271, -11423.6395, -9454.6287, -8735.79435},
        {-404.288181, -395.981464, -387.703721, -379.458826, -371.250655, -363.083083, -354.959983, -346.885232, -338.862704, -330.896273, -322.989815, -315.147204, -307.372315, -299.669023, -292.041202, -284.492728, -277.027476, -269.649319, -262.362133, -255.169793, -248.076174, -241.085149, -234.200595, -227.426385, -220.766395, -214.2245, -207.804574, -201.510492, -195.346129, -189.315359, -183.422058, -177.670101, -172.063361, -166.605714, -161.650682, -157.019961, -152.531046, -148.176465, -143.948747, -139.840419, -135.844009, -131.952046, -128.157058, -124.451572, -120.828118, -117.279222, -113.797413, -110.375219, -107.005169, -103.67979, -100.39161, -96.6461314, -92.9334658, -89.2566994, -85.6189182, -82.0232081, -78.4726549, -74.9703445, -71.5193628, -68.1227957, -64.783729, -61.5052487, -58.2904406, -55.1423907, -52.0641847, -49.0589086, -46.1296483, -43.3254625, -40.6943063, -38.1444498, -35.6759988, -33.2890588, -30.9837356, -28.7601347, -26.6183618, -24.5585225, -22.5807226, -20.6850676, -18.8716633, -17.1406151, -15.4920289, -13.9260102, -12.4426647, -11.042098, -10.1888806, -9.64111616, -9.15132626, -8.70454373, -8.28580144, -7.88013228, -7.47256911, -7.04814482, -6.59189228, -6.08884437, -5.52403395, -4.88249391, -4.14925712, -3.30935646, -2.3478248, -1.24969502, 0, 5.23966661, 10.4435192, 15.571137, 20.5820994, 25.4359855, 30.0923747, 34.5108461, 38.6509791, 42.4723528, 45.9345467, 48.9971399, 51.6197116, 53.7618412, 55.3831079, 56.443091, 56.9013697, 56.0368492, 53.7118695, 51.4256299, 48.9918511, 46.2242535, 42.9365576, 38.9424839, 34.055753, 28.0900854, 20.8592016, 12.176822, 1.85666732, -10.2875421, -25.0360148, -41.7633271, -60.4775798, -81.1868734, -102.216309, -124.415486, -149.082807, -176.731272, -207.873881, -243.023636, -282.693538, -328.290462, -379.161484, -435.644718, -498.078275, -566.800267, -640.290229, -721.209495, -810.360822, -908.546966, -1016.57069, -1135.42842, -1265.49251, -1408.36598, -1564.82573, -1733.96202, -1915.6224, -2109.6544, -2313.39313, -2529.76957, -2764.45556, -3025.18931, -3303.52384, -3594.13542, -3893.91115, -4203.48421, -4521.76826, -4848.04304, -5181.31711, -5519.25683, -5860.28777, -6210.74633, -6559.93025, -6903.52674, -7237.22305, -7556.7064, -7817.00746, -8064.63417, -8305.43789, -8545.26999, -8789.98185, -9046.93027, -9319.45755, -9612.76988, -9979.27434, -10405.4934, -10858.2825, -11325.1476, -11793.5944, -12251.1288, -12685.2566, -14860.175, -15832.3784, -15855.3471, -15182.5613, -14067.5016, -12763.6482, -11524.4814, -9593.9521, -8829.53039},
        {-412.375627, -404.78363, -397.12075, -389.394762, -381.613444, -373.784574, -365.915928, -358.015283, -350.090417, -342.149107, -334.19913, -326.248262, -318.304282, -310.374966, -302.468092, -294.591436, -286.752776, -278.959889, -271.220551, -263.542541, -255.933635, -248.40161, -240.954244, -233.599314, -226.344596, -219.197868, -212.166907, -205.25949, -198.483395, -191.846398, -185.356277, -179.020808, -172.84777, -166.844938, -161.534136, -156.655306, -151.942212, -147.385948, -142.97761, -138.708294, -134.569095, -130.551108, -126.645429, -122.843152, -119.135374, -115.51319, -111.967695, -108.489984, -105.071153, -101.702297, -98.3745112, -94.4932676, -90.6479804, -86.8424402, -83.0804375, -79.3657628, -75.7022067, -72.0935597, -68.5436125, -65.0561554, -61.6349792, -58.2838742, -55.0066311, -51.8070404, -48.6888927, -45.6559785, -42.7120883, -39.9364945, -37.4066573, -34.9706915, -32.6274944, -30.3759636, -28.2149963, -26.1434898, -24.1603417, -22.2644491, -20.4547095, -18.7300203, -17.0892787, -15.5313823, -14.0552283, -12.659714, -11.343737, -10.1061945, -9.32941571, -8.81251719, -8.35000678, -7.92833875, -7.53396736, -7.15334687, -6.77293155, -6.37917565, -5.95853344, -5.49745919, -4.98240714, -4.39983157, -3.73618674, -2.9779269, -2.11150632, -1.12337927, 0, 6.0958, 11.9425542, 17.5120007, 22.775878, 27.7059242, 32.2738776, 36.4514765, 40.2104592, 43.5225638, 46.3595287, 48.6930922, 50.4949924, 51.7369677, 52.3907564, 52.4280966, 51.8207266, 49.9567062, 46.7243909, 43.4157444, 39.877431, 35.9561149, 31.4984604, 26.3511318, 20.3607935, 13.3741095, 5.23774436, -4.20163781, -15.0973727, -27.602796, -42.0445637, -58.3391405, -76.5878661, -96.8920802, -118.127279, -141.007723, -166.574566, -195.296902, -227.643822, -264.084421, -305.087789, -352.135835, -404.376589, -462.080985, -525.519959, -594.964442, -667.824312, -747.946824, -836.318176, -933.924567, -1041.75219, -1161.65553, -1292.69128, -1435.33102, -1589.98925, -1756.96387, -1936.60514, -2129.26331, -2334.43239, -2553.51347, -2788.84174, -3044.27285, -3315.58121, -3600.76416, -3898.19909, -4206.90734, -4525.61467, -4853.44226, -5189.21838, -5530.32141, -5874.94988, -6228.58686, -6581.02563, -6927.94313, -7265.01631, -7587.92213, -7857.57459, -8113.10431, -8358.87898, -8599.26629, -8838.63391, -9075.39578, -9323.8425, -9590.89335, -9933.83069, -10339.6384, -10774.3159, -11225.8962, -11682.4125, -12131.8979, -12562.3856, -14676.0666, -15630.7739, -15673.5039, -15051.2532, -14011.0182, -12799.7956, -11664.5818, -9869.64311, -8970.7176},
        {-420.690164, -413.82336, -406.785439, -399.588068, -392.242914, -384.761643, -377.155923, -369.437419, -361.617798, -353.708727, -345.721872, -337.6689, -329.561477, -321.411271, -313.229947, -305.029172, -296.820613, -288.615936, -280.426808, -272.264896, -264.141865, -256.069384, -248.059117, -240.122732, -232.271896, -224.518274, -216.873534, -209.349342, -201.957364, -194.709268, -187.61672, -180.691386, -173.944933, -167.389027, -161.711135, -156.570809, -151.619479, -146.84688, -142.242747, -137.796817, -133.498825, -129.338506, -125.305596, -121.389831, -117.580946, -113.868677, -110.242759, -106.692928, -103.20892, -99.7804694, -96.3973127, -92.3666191, -88.375487, -84.4284487, -80.5300364, -76.6847825, -72.8972191, -69.1718787, -65.5132934, -61.9259956, -58.4145174, -54.9833913, -51.6371494, -48.380324, -45.2174474, -42.153052, -39.1916698, -36.444984, -34.0221048, -31.7065743, -29.4959786, -27.3879038, -25.379936, -23.4696614, -21.6546662, -19.9325364, -18.3008582, -16.7572177, -15.299201, -13.9243944, -12.6303838, -11.4147555, -10.2750956, -9.20899013, -8.51007081, -8.02167408, -7.58357818, -7.183762, -6.81020443, -6.45088435, -6.09378066, -5.72687223, -5.33813797, -4.91555674, -4.44710745, -3.92076898, -3.32452022, -2.64634005, -1.87420737, -0.996101056, 0, 7.13387633, 13.7668832, 19.8856696, 25.4768844, 30.5271767, 35.0231955, 38.9515896, 42.2990081, 45.0520999, 47.197514, 48.7218994, 49.611905, 49.8541799, 49.4353729, 48.3421331, 46.5611094, 43.5691081, 39.2799434, 34.8334801, 30.107115, 24.9782452, 19.3242676, 13.0225791, 5.95057673, -2.01434257, -10.9947818, -21.1133441, -32.4926324, -45.2552499, -59.2701406, -75.0065748, -92.663253, -112.438876, -133.808221, -157.33195, -183.763772, -213.519564, -247.015206, -284.666574, -326.889546, -375.122316, -428.447307, -487.080377, -551.237385, -621.134192, -693.146744, -772.290791, -859.74217, -956.676718, -1064.27027, -1185.29727, -1317.38111, -1459.72014, -1612.32317, -1776.85363, -1954.23208, -2145.37905, -2352.22767, -2574.45576, -2810.63649, -3060.29888, -3323.87358, -3602.92362, -3897.46374, -4204.88522, -4523.78357, -4853.30769, -5192.19654, -5537.15996, -5886.05556, -6242.98719, -6598.6053, -6948.60535, -7288.68282, -7614.53317, -7892.72553, -8155.36331, -8405.42354, -8645.88329, -8879.71961, -9096.88404, -9322.06283, -9563.81538, -9884.51947, -10271.4314, -10689.3832, -11126.7524, -11571.9166, -12013.2532, -12439.1398, -14379.5967, -15280.7979, -15360.8605, -14837.9013, -13930.0371, -12855.385, -11832.0618, -10221.1229, -9160.46137},
        {-429.7245, -423.570124, -417.143523, -410.460313, -403.536112, -396.386538, -389.027207, -381.473737, -373.741745, -365.846848, -357.804664, -349.630811, -341.340904, -332.950562, -324.475402, -315.931042, -307.333098, -298.697187, -290.038928, -281.373937, -272.717831, -264.086229, -255.494747, -246.959002, -238.494613, -230.117195, -221.842367, -213.685745, -205.662948, -197.789592, -190.081294, -182.553672, -175.222344, -168.102926, -162.052964, -156.649402, -151.459065, -146.470249, -141.671249, -137.050359, -132.595875, -128.296092, -124.139305, -120.113809, -116.207899, -112.40987, -108.708018, -105.090638, -101.546024, -98.0624712, -94.6282756, -90.4493149, -86.3132626, -82.2253748, -78.1909082, -74.2151189, -70.3032634, -66.460598, -62.6923791, -59.0038632, -55.4003065, -51.8869654, -48.4690964, -45.1519557, -41.9407998, -38.840885, -35.8574678, -33.1333086, -30.8078668, -28.6028065, -26.5144703, -24.5392005, -22.6733399, -20.9132309, -19.255216, -17.6956377, -16.2308387, -14.8571614, -13.5709484, -12.3685421, -11.2462852, -10.2005201, -9.22758946, -8.32383567, -7.69206376, -7.22104237, -6.79705865, -6.40975501, -6.04877389, -5.70375771, -5.3643489, -5.02018988, -4.66092309, -4.27619094, -3.85563588, -3.38890031, -2.86562668, -2.27545741, -1.60803492, -0.85300164, 0, 7.90796871, 15.096832, 21.5669213, 27.3185682, 32.3521042, 36.6678608, 40.2661695, 43.1473618, 45.3117693, 46.7597235, 47.4915558, 47.5075977, 46.8081809, 45.3936368, 43.2642968, 40.4204926, 36.2593911, 30.6951572, 25.0924236, 19.3222721, 13.2557848, 6.76404343, -0.281869906, -8.01087329, -16.5518848, -26.0338224, -36.5856042, -48.3361483, -61.4143727, -75.6131509, -91.5206619, -109.366637, -129.380807, -151.28726, -175.568549, -202.842066, -233.489236, -267.891483, -306.430233, -349.486909, -398.385676, -452.278299, -511.361753, -575.833016, -645.889063, -717.355495, -795.893506, -882.792917, -979.343546, -1086.83521, -1208.67864, -1341.45053, -1483.3186, -1633.96893, -1796.18767, -1971.36909, -2160.90748, -2368.62805, -2592.94166, -2829.60725, -3075.10396, -3333.02388, -3607.19512, -3898.71934, -4204.07844, -4521.87487, -4851.39975, -5191.43408, -5538.23367, -5889.48272, -6248.90262, -6607.21158, -6960.00399, -7302.87427, -7631.4168, -7915.25549, -8182.44787, -8435.08097, -8675.24183, -8905.01746, -9109.29763, -9318.83148, -9543.07632, -9846.09757, -10215.7368, -10617.0822, -11039.1129, -11470.8083, -11901.1475, -12319.1098, -13907.2374, -14705.292, -14861.1258, -14522.5911, -13837.5401, -12953.8251, -12019.2982, -10592.9555, -9382.69215},
        {-439.810344, -434.33181, -428.48006, -422.274709, -415.735373, -408.88167, -401.733215, -394.309624, -386.630514, -378.715501, -370.584201, -362.256231, -353.751206, -345.088744, -336.288459, -327.369969, -318.35289, -309.256838, -300.101429, -290.90628, -281.691006, -272.475224, -263.278551, -254.120602, -245.020994, -235.999343, -227.075266, -218.268377, -209.598295, -201.084635, -192.747013, -184.605046, -176.678349, -168.98654, -162.562416, -156.897692, -151.471904, -146.271789, -141.284084, -136.495524, -131.892847, -127.46279, -123.192087, -119.067477, -115.075696, -111.203479, -107.437565, -103.764689, -100.171588, -96.6449978, -93.1716558, -88.8546067, -84.5834352, -80.3640348, -76.2022986, -72.1041199, -68.0753919, -64.1220079, -60.2498612, -56.4648449, -52.7728524, -49.1797768, -45.6915115, -42.3139497, -39.0529846, -35.9145095, -32.9044176, -30.1919606, -27.9424678, -25.8249182, -23.8346152, -21.9668621, -20.2169624, -18.5802194, -17.0519365, -15.6274172, -14.3019647, -13.0708826, -11.9294742, -10.8730428, -9.896892, -8.99632502, -8.16664531, -7.40315626, -6.82017772, -6.35100151, -5.92748894, -5.54108109, -5.18321905, -4.84534389, -4.51889669, -4.19531854, -3.8660505, -3.52253367, -3.15620911, -2.7585179, -2.32090114, -1.83479988, -1.29165522, -0.682908236, 0, 8.08451029, 15.3205798, 21.7178708, 27.2860457, 32.0347669, 35.9736967, 39.1124974, 41.4608314, 43.0283611, 43.8247488, 43.8596568, 43.1427476, 41.6836835, 39.4921267, 36.5777397, 32.9501849, 27.6622427, 20.586878, 13.8981289, 7.40061159, 0.898942227, -5.80226289, -12.8983875, -20.5848154, -29.0569303, -38.510116, -49.1397562, -61.1412346, -74.7099351, -90.2042497, -107.596784, -127.03402, -148.662437, -171.83223, -197.088024, -225.186787, -256.513888, -291.454693, -330.394572, -373.71889, -422.606117, -476.407142, -535.35216, -599.671369, -669.594963, -741.204999, -819.916848, -906.997741, -1003.71491, -1111.33559, -1233.35878, -1366.09221, -1507.2191, -1656.28591, -1816.64326, -1989.93378, -2177.80009, -2384.78531, -2608.97238, -2845.28009, -3089.68082, -3346.02323, -3617.97942, -3906.68019, -4208.95028, -4523.59096, -4850.09448, -5187.44125, -5532.07821, -5881.88539, -6242.2596, -6602.40763, -6957.60754, -7303.13738, -7634.27521, -7919.55258, -8187.68066, -8440.62415, -8680.3477, -8908.81601, -9111.22724, -9317.49025, -9536.75536, -9828.80671, -10182.7521, -10566.6018, -10970.5186, -11384.665, -11799.2037, -12204.2975, -13230.3355, -13870.6783, -14151.9507, -14100.7774, -13743.7832, -13107.5929, -12218.8313, -10942.085, -9629.27022},
        {-450.224459, -445.357494, -440.027253, -434.256814, -428.069258, -421.487663, -414.535108, -407.234672, -399.609434, -391.682474, -383.47687, -375.015701, -366.322047, -357.418986, -348.329597, -339.07696, -329.684154, -320.174257, -310.570349, -300.895509, -291.172815, -281.425348, -271.676185, -261.948405, -252.265089, -242.649315, -233.124162, -223.712709, -214.438035, -205.323219, -196.391341, -187.665478, -179.168711, -170.924119, -164.104014, -158.132687, -152.420614, -146.953578, -141.717363, -136.697754, -131.880536, -127.251491, -122.796405, -118.501061, -114.351244, -110.332737, -106.431326, -102.632793, -98.9229239, -95.2875021, -91.7123118, -87.2378662, -82.8157122, -78.4521254, -74.1533817, -69.9257568, -65.7755265, -61.7089667, -57.732353, -53.8519613, -50.0740673, -46.4049469, -42.8508758, -39.4181298, -36.1129846, -32.9417161, -29.9106001, -27.2093469, -25.023263, -22.9803038, -21.0748537, -19.3012972, -17.6540187, -16.1274027, -14.7158337, -13.4136962, -12.2153744, -11.1152531, -10.1077165, -9.1871491, -8.34793543, -7.58445991, -6.891107, -6.26226115, -5.73150573, -5.27280125, -4.85963707, -4.48512555, -4.14237909, -3.82451005, -3.5246308, -3.23585374, -2.95129122, -2.66405563, -2.36725934, -2.05401473, -1.71743417, -1.35063005, -0.946714726, -0.498800585, 0, 8.06618287, 15.1882845, 21.383434, 26.6687606, 31.0613933, 34.5784612, 37.2370935, 39.0544192, 40.0475676, 40.2336677, 39.6298487, 38.2532396, 36.1209696, 33.2501678, 29.6579633, 25.3614853, 19.0462444, 10.5391385, 2.86929094, -4.23151614, -11.0315004, -17.7988794, -24.801871, -32.3086928, -40.5875624, -49.9066976, -60.5343161, -72.7386356, -86.7878736, -103.682825, -122.690519, -143.859399, -167.237912, -191.729085, -217.954071, -246.839467, -278.777342, -314.159765, -353.378807, -396.826536, -445.478723, -498.966087, -557.566498, -621.557821, -691.217924, -763.198654, -842.310403, -929.737544, -1026.66445, -1134.27549, -1255.87868, -1387.93919, -1527.97635, -1675.50887, -1834.13744, -2005.63001, -2191.75454, -2397.28204, -2620.29491, -2855.59951, -3099.49283, -3355.17021, -3625.38899, -3910.98986, -4209.56581, -4520.2006, -4842.69651, -5176.32357, -5517.71718, -5865.00305, -6225.34806, -6586.44576, -6943.29222, -7290.88349, -7624.21563, -7910.44265, -8179.36316, -8432.93373, -8673.11091, -8901.85129, -9105.77269, -9312.39623, -9530.2522, -9813.30219, -10152.0933, -10518.2429, -10903.2445, -11298.5918, -11695.7783, -12086.2976, -12548.7559, -13028.9453, -13433.3919, -13668.6217, -13641.161, -13257.5358, -12424.2722, -11306.5776, -9944.01651},
        {-460.767447, -456.422007, -451.540103, -446.147438, -440.269719, -433.93265, -427.161935, -419.983281, -412.422391, -404.504972, -396.256727, -387.703362, -378.870582, -369.784092, -360.469597, -350.952801, -341.25941, -331.415128, -321.445661, -311.376714, -301.233991, -291.043198, -280.830039, -270.62022, -260.439445, -250.313419, -240.267848, -230.328436, -220.520889, -210.87091, -201.404206, -192.146481, -183.12344, -174.360788, -167.114388, -160.76899, -154.694561, -148.876886, -143.301747, -137.954929, -132.822217, -127.889394, -123.142244, -118.566551, -114.148099, -109.872672, -105.726054, -101.69403, -97.7623822, -93.916896, -90.143355, -85.4818271, -80.8843137, -76.3571, -71.9064713, -67.5387128, -63.26011, -59.0769479, -54.9955119, -51.0220873, -47.1629594, -43.4244134, -39.8127346, -36.3342083, -32.9951197, -29.8017542, -26.760397, -24.072842, -21.9361891, -19.9535017, -18.1183909, -16.424468, -14.8653443, -13.434631, -12.1259394, -10.9328807, -9.84906611, -8.86810693, -7.9836144, -7.18919975, -6.47847423, -5.84504909, -5.28253558, -4.78454493, -4.3166147, -3.88698347, -3.50422725, -3.16286835, -2.85742904, -2.58243162, -2.33239838, -2.10185161, -1.88531361, -1.67730666, -1.47235306, -1.26497509, -1.04969505, -0.82103523, -0.573517921, -0.301665414, 0, 7.89008169, 14.7738094, 20.6736163, 25.6119355, 29.6112004, 32.6938441, 34.8822998, 36.1990007, 36.66638, 36.3068709, 35.1429066, 33.1969204, 30.4913453, 27.0486147, 22.8911617, 18.0414195, 10.8686626, 1.15915857, -7.32537164, -14.9167432, -21.9467711, -28.7472706, -35.6500567, -42.9869445, -51.0897492, -60.2902858, -70.9203694, -83.3118151, -97.7964381, -115.939098, -136.386449, -159.100393, -184.042831, -209.737545, -236.865497, -266.491147, -299.007832, -334.808889, -374.287658, -417.837474, -466.180846, -519.281759, -577.469147, -641.071946, -710.41909, -782.957617, -862.618834, -950.452147, -1047.50697, -1154.8327, -1275.2711, -1405.88859, -1544.39843, -1690.41723, -1847.44751, -2017.24711, -2201.57386, -2404.96968, -2625.77556, -2859.29532, -3102.749, -3358.07284, -3626.63121, -3908.77532, -4203.13959, -4509.14652, -4827.01493, -5156.37376, -5493.9319, -5838.04997, -6197.29812, -6558.25677, -6915.75249, -7264.61185, -7599.66141, -7887.56868, -8158.35906, -8413.89889, -8656.05451, -8886.69226, -9093.79479, -9302.36791, -9520.22812, -9794.93053, -10118.5618, -10466.6507, -10832.1117, -11207.8595, -11586.8083, -11961.8728, -11948.8922, -12290.2826, -12795.0698, -13272.2794, -13530.9372, -13380.0689, -12628.7001, -11684.2181, -10344.9507},
        {-471.178999, -467.236262, -462.707805, -457.620759, -452.002255, -445.879425, -439.2794, -432.229311, -424.75629, -416.887468, -408.649976, -400.070945, -391.177508, -381.996794, -372.555937, -362.882066, -353.002313, -342.943809, -332.733687, -322.399076, -311.967109, -301.464917, -290.919631, -280.358382, -269.808302, -259.296522, -248.850173, -238.496387, -228.262295, -218.175028, -208.261718, -198.549496, -189.065493, -179.83684, -172.122046, -165.306654, -158.761405, -152.473471, -146.430022, -140.618228, -135.025262, -129.638292, -124.444491, -119.43103, -114.585077, -109.893806, -105.344385, -100.923987, -96.6197819, -92.4189401, -88.3086328, -83.4183989, -78.6116328, -73.8940969, -69.2715538, -64.749766, -60.3344962, -56.0315068, -51.8465604, -47.7854195, -43.8538467, -40.0576045, -36.4024556, -32.8941623, -29.5384873, -26.3411931, -23.3080423, -20.6413546, -18.5387308, -16.6005487, -14.8198287, -13.1895915, -11.7028575, -10.3526472, -9.13198135, -8.03388033, -7.05136469, -6.17745499, -5.40517176, -4.72753553, -4.13756683, -3.6282862, -3.19271419, -2.82387131, -2.43849321, -2.06934742, -1.75011837, -1.47630218, -1.24339498, -1.04689289, -0.882292049, -0.745088578, -0.630778605, -0.534858256, -0.45282366, -0.380170943, -0.312396233, -0.244995657, -0.173465341, -0.0933014128, 0, 7.62667166, 14.2134172, 19.7857303, 24.3691046, 27.9890337, 30.6710112, 32.4405307, 33.3230859, 33.3441704, 32.5292778, 30.9039017, 28.4935357, 25.3236736, 21.4198089, 16.8074352, 11.5120461, 3.72760157, -6.78205251, -15.8486676, -23.8392215, -31.1206922, -38.0600575, -45.0242955, -52.3803838, -60.4953006, -69.7360236, -80.4695307, -93.0627999, -107.882809, -126.784, -148.100483, -171.752997, -197.662281, -224.228346, -252.132294, -282.460755, -315.590687, -351.899046, -391.76279, -435.558876, -483.689192, -536.498176, -594.357907, -657.640465, -726.717929, -799.989132, -880.292712, -968.494058, -1065.45856, -1172.05161, -1290.34454, -1418.52287, -1554.93498, -1699.45195, -1855.0538, -2023.32471, -2205.84884, -2406.47907, -2624.01527, -2854.78237, -3097.30188, -3351.94248, -3618.46644, -3896.6564, -4186.32947, -4487.28695, -4800.30221, -5125.42857, -5459.15505, -5799.98682, -6156.92643, -6516.36672, -6873.17919, -7222.23532, -7558.40659, -7850.20798, -8125.45662, -8385.61312, -8632.13809, -8866.49214, -9076.61777, -9286.50579, -9503.41031, -9768.93161, -10076.7284, -10406.1486, -10751.5663, -11107.356, -11467.8919, -11827.5482, -11544.7608, -11799.692, -12354.8358, -12972.686, -13415.7365, -13446.4812, -12827.4138, -12078.166, -10865.7065},
        {-480.610388, -476.893666, -472.583828, -467.707271, -462.29039, -456.359582, -449.941244, -443.061771, -435.747561, -428.025009, -419.920512, -411.460466, -402.671268, -393.579314, -384.211001, -374.592724, -364.750881, -354.711867, -344.502078, -334.147913, -323.675766, -313.112034, -302.483113, -291.8154, -281.135292, -270.469184, -259.843473, -249.284555, -238.818827, -228.472685, -218.272526, -208.244745, -198.41574, -188.811906, -180.543122, -173.071253, -165.845143, -158.856028, -152.095142, -145.553721, -139.223, -133.094215, -127.158601, -121.407393, -115.831826, -110.423136, -105.172559, -100.071329, -95.1106814, -90.2818522, -85.5760765, -80.3885077, -75.3193849, -70.3728653, -65.553106, -60.8642642, -56.3104973, -51.8959623, -47.6248164, -43.501217, -39.5293211, -35.713286, -32.0572689, -28.565427, -25.2419175, -22.0908975, -19.1165244, -16.498802, -14.4135121, -12.5021414, -10.7574476, -9.17218864, -7.73912215, -6.45100594, -5.30059781, -4.28065552, -3.38393685, -2.60319959, -1.93120151, -1.36070039, -0.884454002, -0.495220136, -0.185756564, 0.051178935, 0.306290266, 0.543333305, 0.725060782, 0.856006427, 0.940703969, 0.983687137, 0.989489663, 0.962645273, 0.907687699, 0.82915067, 0.731567916, 0.619473165, 0.497400147, 0.369882592, 0.24145423, 0.116648789, 0, 7.66878541, 14.2461831, 19.760525, 24.2401429, 27.7133689, 30.2085348, 31.7539726, 32.378014, 32.1089912, 30.9752359, 29.0050801, 26.2268556, 22.6688945, 18.3595285, 13.3270897, 7.59990996, -0.418268244, -10.9311972, -20.2358785, -28.6521065, -36.4996757, -44.0983803, -51.7680149, -59.8283738, -68.5992513, -78.4004419, -89.5517399, -102.37294, -117.183836, -135.26246, -155.619016, -178.285825, -203.295212, -229.814411, -258.308288, -289.472402, -323.598601, -360.978737, -401.904657, -446.668213, -495.042095, -547.995316, -605.921299, -669.213468, -738.265248, -812.69997, -893.873674, -982.372306, -1078.78181, -1183.68814, -1297.74074, -1421.38444, -1554.57306, -1697.78468, -1852.56779, -2019.99031, -2201.12012, -2398.31001, -2611.05099, -2837.43241, -3077.56062, -3330.15873, -3593.33671, -3866.35353, -4150.41515, -4445.83382, -4754.43404, -5076.92011, -5408.45153, -5747.32424, -6100.02122, -6455.39186, -6808.89816, -7156.00211, -7492.16571, -7794.10555, -8080.71536, -8352.14349, -8608.53826, -8850.04803, -9060.0428, -9266.63478, -9477.16303, -9729.51495, -10018.9375, -10327.9493, -10652.4336, -10988.2739, -11331.3536, -11677.5561, -11717.23, -12038.4779, -12503.2662, -12973.5615, -13311.3301, -13378.5384, -13037.1529, -12543.5046, -11690.7554},
        {-490.273963, -486.650583, -482.448426, -477.692128, -472.406325, -466.615655, -460.344754, -453.618259, -446.460806, -438.897031, -430.951573, -422.649067, -414.014149, -405.071457, -395.845628, -386.361297, -376.643101, -366.715678, -356.603663, -346.331694, -335.924406, -325.406437, -314.802424, -304.137002, -293.434809, -282.720481, -272.018655, -261.353967, -250.751055, -240.234554, -229.829101, -219.559334, -209.449888, -199.5254, -190.684635, -182.496779, -174.520512, -166.752104, -159.187826, -151.823945, -144.656732, -137.682457, -130.897388, -124.297796, -117.87995, -111.640119, -105.574573, -99.6795817, -93.9514143, -88.3863405, -82.9806298, -77.4528694, -72.0830308, -66.8734031, -61.8262754, -56.9439368, -52.2286765, -47.6827837, -43.3085474, -39.1082568, -35.0842011, -31.2386693, -27.5739506, -24.0923342, -20.7961091, -17.6875646, -14.7689897, -12.1923264, -10.1059236, -8.20342268, -6.4774114, -4.92047749, -3.52520872, -2.28419284, -1.19001759, -0.23527073, 0.587459987, 1.28558681, 1.86652199, 2.33767777, 2.7064664, 2.98030014, 3.16659122, 3.27275189, 3.38796571, 3.47703931, 3.50379653, 3.47299597, 3.38939626, 3.25775599, 3.08283379, 2.86938826, 2.62217802, 2.34596169, 2.04549786, 1.72554517, 1.39086221, 1.0462076, 0.696339953, 0.346017883, 0, 7.28008396, 13.4996853, 18.6844092, 22.8598606, 26.0516447, 28.2853665, 29.5866311, 29.9810435, 29.4942089, 28.1517322, 25.9792187, 23.0022733, 19.2465011, 14.7375072, 9.50089671, 3.56227466, -4.3759908, -14.4773284, -23.7230973, -32.371243, -40.6797113, -48.906448, -57.3093986, -66.1465089, -75.6757245, -86.1549913, -97.8422548, -110.995461, -125.872555, -143.10251, -162.436201, -184.020267, -208.001343, -234.271311, -263.104187, -294.841922, -329.707581, -367.92423, -409.714933, -455.302756, -504.067504, -557.332145, -615.48473, -678.913307, -748.005928, -823.351572, -905.087127, -993.550412, -1089.07924, -1192.01144, -1301.69094, -1420.66418, -1550.42825, -1692.03874, -1845.64984, -2011.82047, -2191.10952, -2384.42679, -2591.90056, -2813.27628, -3049.99573, -3299.52286, -3558.81004, -3826.74847, -4105.51453, -4395.77667, -4700.1197, -5019.70889, -5348.68279, -5685.15468, -6032.97844, -6383.64355, -6733.27612, -7078.00229, -7413.94818, -7727.37922, -8026.74742, -8310.64406, -8577.66046, -8826.38791, -9034.18947, -9235.03685, -9436.61917, -9674.49909, -9945.33244, -10234.1611, -10538.2217, -10854.7507, -11180.9846, -11514.1601, -11912.4545, -12309.9626, -12675.6516, -12978.4883, -13187.4399, -13271.4732, -13199.5554, -12953.442, -12476.2222},
        {-500.59357, -496.901677, -492.666538, -487.910099, -482.654305, -476.921105, -470.732442, -464.110265, -457.076519, -449.653151, -441.862107, -433.725333, -425.264775, -416.502381, -407.460095, -398.159865, -388.623636, -378.873356, -368.93097, -358.818424, -348.557666, -338.17064, -327.679294, -317.105575, -306.471427, -295.798797, -285.109633, -274.425879, -263.769483, -253.16239, -242.626547, -232.183901, -221.856397, -211.665982, -202.248666, -193.306447, -184.533944, -175.933175, -167.506158, -159.254912, -151.181457, -143.287811, -135.575992, -128.048019, -120.705912, -113.551688, -106.587367, -99.8149675, -93.2365078, -86.8540069, -80.6694835, -74.7655813, -69.0619462, -63.5588493, -58.2565616, -53.1553539, -48.2554973, -43.5572626, -39.0609209, -34.766743, -30.675, -26.7859628, -23.0999023, -19.6170895, -16.3377953, -13.2622907, -10.3908467, -7.84406063, -5.73964085, -3.8296941, -2.10669108, -0.563102449, 0.8086011, 2.01594889, 3.06647025, 3.96769449, 4.72715093, 5.3523689, 5.85087771, 6.23020669, 6.49788515, 6.66144242, 6.72840782, 6.70631066, 6.68342127, 6.62520019, 6.49613751, 6.30114237, 6.04512387, 5.73299114, 5.36965329, 4.96001945, 4.50899874, 4.02150028, 3.50243318, 2.95670657, 2.38922956, 1.80491128, 1.20866085, 0.605387382, 0, 6.04561945, 11.209162, 15.5039382, 18.9432585, 21.5404333, 23.3087732, 24.2615887, 24.4121902, 23.7738881, 22.3599931, 20.1838156, 17.258666, 13.5978549, 9.21469264, 4.12248984, -1.66544306, -9.1049758, -18.3212522, -27.1087685, -35.6618955, -44.1750045, -52.8424662, -61.8586517, -71.4179321, -81.7146782, -92.9432611, -105.298052, -118.973421, -134.16374, -150.971375, -169.716436, -190.620898, -213.906731, -239.851611, -268.64966, -300.480145, -335.548328, -374.059472, -416.218838, -462.231689, -511.638248, -565.511219, -624.185981, -687.997911, -757.28239, -832.93972, -914.599123, -1002.45475, -1096.70074, -1197.53124, -1303.48212, -1418.4326, -1544.90231, -1684.32646, -1835.92632, -1999.91717, -2176.51431, -2365.71093, -2567.99488, -2784.09628, -3015.99051, -3260.88266, -3515.70838, -3779.35308, -4054.00586, -4340.33938, -4640.97657, -4957.09574, -5282.72407, -5615.93383, -5958.61432, -6304.43352, -6650.14241, -6992.49195, -7328.23312, -7653.68358, -7966.13595, -8262.44953, -8539.48364, -8794.09757, -8996.29219, -9187.69087, -9376.66367, -9598.43232, -9850.73808, -10120.1341, -10404.921, -10703.3995, -11013.8703, -11334.634, -11963.1717, -12405.5115, -12703.4562, -12898.8088, -13033.3721, -13148.949, -13287.3423, -13256.984, -13098.4345},
        {-512.726318, -508.65409, -504.104323, -499.095084, -493.64444, -487.770456, -481.491201, -474.824739, -467.789138, -460.402465, -452.682786, -444.648168, -436.316678, -427.706381, -418.835345, -409.721636, -400.383321, -390.838466, -381.105139, -371.201405, -361.145332, -350.954985, -340.648433, -330.24374, -319.758974, -309.212202, -298.62149, -288.004904, -277.380512, -266.766379, -256.180574, -245.641161, -235.166208, -224.773781, -214.7747, -205.034498, -195.415188, -185.925336, -176.573509, -167.368273, -158.318194, -149.431838, -140.717772, -132.184563, -123.840775, -115.694975, -107.75573, -100.031606, -92.5311692, -85.2629855, -78.2356212, -71.9513915, -65.9144103, -60.1225402, -54.573644, -49.2655845, -44.1962244, -39.3634263, -34.765053, -30.3989673, -26.2630318, -22.3551093, -18.6730625, -15.2147541, -11.9780469, -8.96080359, -6.16088686, -3.65950244, -1.535855, 0.384081375, 2.10784675, 3.6429812, 4.99702476, 6.17751751, 7.19199951, 8.04801082, 8.7530915, 9.31478161, 9.74062122, 10.0381504, 10.2149092, 10.2784376, 10.2362758, 10.0959639, 9.96535563, 9.79972432, 9.55313736, 9.22987945, 8.83423526, 8.37048948, 7.84292678, 7.25583186, 6.6134894, 5.92018407, 5.18020056, 4.39782356, 3.57733774, 2.7230278, 1.8391784, 0.930074242, 0, 2.64118845, 4.94021312, 6.87512108, 8.42395936, 9.56477502, 10.2756151, 10.5345267, 10.3195568, 9.60875246, 8.38016077, 6.61182876, 4.28180348, 1.36813198, -2.15113868, -6.29796147, -11.0942893, -17.1480881, -24.565027, -32.0415534, -39.7251945, -47.7634774, -56.3039293, -65.4940773, -75.4814486, -86.4135704, -98.4379698, -111.702174, -126.35371, -142.540105, -160.333296, -179.984117, -201.662772, -225.539465, -251.254161, -279.242185, -310.080256, -344.097649, -381.623642, -422.987509, -468.518526, -519.40157, -574.849916, -635.025437, -700.09001, -770.20551, -844.790789, -924.936502, -1010.99028, -1103.29975, -1202.21255, -1306.91426, -1420.33484, -1544.37328, -1680.23819, -1827.7287, -1987.27677, -2159.31434, -2344.65733, -2543.26648, -2754.68369, -2979.04917, -3216.10364, -3465.83254, -3728.093, -4002.52476, -4288.86737, -4587.84226, -4899.44351, -5220.06494, -5548.13687, -5885.71466, -6227.04593, -6569.30619, -6909.67096, -7245.31576, -7582.31514, -7906.72082, -8213.48358, -8497.55417, -8753.88337, -8943.1255, -9117.39203, -9286.33219, -9488.47561, -9722.03969, -9973.20833, -10240.8006, -10523.6356, -10820.5323, -11130.3099, -11502.6095, -11868.9481, -12219.2137, -12543.2943, -12831.0781, -13072.4531, -13257.3074, -13366.9625, -13328.691},
        {-524.778692, -520.366852, -515.542709, -510.320471, -504.714344, -498.738537, -492.407257, -485.734709, -478.735103, -471.422645, -463.811542, -455.916001, -447.75023, -439.328437, -430.664827, -421.773609, -412.668989, -403.365175, -393.876374, -384.216794, -374.400641, -364.442123, -354.355447, -344.15482, -333.854449, -323.468543, -313.011307, -302.496949, -291.939676, -281.353697, -270.753217, -260.152444, -249.565585, -239.006848, -228.478664, -218.001376, -207.595469, -197.275531, -187.056151, -176.95192, -166.977426, -157.147259, -147.476008, -137.978263, -128.668613, -119.561647, -110.671954, -102.014124, -93.6027472, -85.4524115, -77.5777069, -70.8373103, -64.3834252, -58.2123429, -52.3203545, -46.7037512, -41.3588242, -36.2818647, -31.4691638, -26.9170129, -22.621703, -18.5795253, -14.7867711, -11.2397315, -7.93469767, -4.86796085, -2.03581221, 0.502676175, 2.68993956, 4.65916765, 6.41813908, 7.97463249, 9.33642652, 10.5112998, 11.507031, 12.3313988, 12.9921817, 13.4971585, 13.8541078, 14.0708082, 14.1550383, 14.1145769, 13.9572025, 13.6906937, 13.4399292, 13.1516745, 12.7712879, 12.302748, 11.7500331, 11.117122, 10.407993, 9.62662477, 8.77699578, 7.86308456, 6.88886964, 5.85832956, 4.77544283, 3.64418798, 2.46854354, 1.25248804, 0, -0.474617309, -0.795330739, -1.01662598, -1.19298874, -1.37890469, -1.62885954, -1.99733897, -2.53882868, -3.30781436, -4.35878171, -5.74621641, -7.52460416, -9.74843066, -12.4721816, -15.7503426, -19.6373995, -24.4605105, -30.3131145, -36.6329378, -43.5328959, -51.1259043, -59.5248787, -68.8427345, -79.1923872, -90.6867525, -103.438746, -117.561283, -133.167278, -150.369649, -169.336673, -190.105603, -212.772745, -237.434406, -263.000472, -290.16046, -319.920264, -352.732116, -389.048248, -429.320892, -474.00228, -525.782898, -582.195516, -643.254448, -708.974006, -779.368504, -852.535144, -930.884627, -1014.91054, -1105.10648, -1201.96603, -1305.33235, -1417.14443, -1538.67661, -1670.89465, -1814.13438, -1969.01447, -2136.15357, -2317.04858, -2511.24033, -2717.31159, -2933.85225, -3162.81725, -3406.91087, -3666.85188, -3939.99483, -4225.23849, -4521.51595, -4827.73487, -5142.67721, -5465.19602, -5798.22988, -6135.91781, -6475.69865, -6815.01125, -7151.29446, -7499.09814, -7834.47237, -8150.57825, -8440.57687, -8697.62933, -8873.37715, -9030.1807, -9178.99519, -9362.52352, -9579.15408, -9814.04705, -10066.2522, -10334.8194, -10618.7983, -10917.2389, -10994.0193, -11273.9483, -11683.0741, -12147.4447, -12593.1085, -12946.1134, -13132.5077, -13346.7033, -13375.2461},
        {-535.733965, -531.226259, -526.352277, -521.123142, -515.54998, -509.643915, -503.416073, -496.877578, -490.039554, -482.913128, -475.509423, -467.839565, -459.914679, -451.745889, -443.34432, -434.721097, -425.887346, -416.85419, -407.632755, -398.234165, -388.669546, -378.950023, -369.086719, -359.090761, -348.973272, -338.745379, -328.418205, -318.002876, -307.510516, -296.95225, -286.339204, -275.682502, -264.993268, -254.282629, -243.310273, -232.218332, -221.151958, -210.130437, -199.173052, -188.299088, -177.527829, -166.87856, -156.370565, -146.023129, -135.855535, -125.88707, -116.137016, -106.624658, -97.3692812, -88.3901693, -79.706607, -72.4008717, -65.4062112, -58.7188663, -52.3350777, -46.2510864, -40.463133, -34.9674585, -29.7603037, -24.8379093, -20.1965162, -15.8323651, -11.741697, -7.92075251, -4.36577258, -1.07299799, 1.96133046, 4.66831316, 6.98079501, 9.0523885, 10.891563, 12.5067879, 13.9065325, 15.0992662, 16.0934583, 16.8975783, 17.5200955, 17.9694792, 18.2541988, 18.3827236, 18.3635231, 18.2050665, 17.9158234, 17.5042629, 17.1013312, 16.6516832, 16.0985016, 15.4462813, 14.6995169, 13.8627032, 12.940335, 11.936907, 10.856914, 9.7048507, 8.48521193, 7.20249241, 5.86118689, 4.46579014, 3.0207969, 1.53070194, 0, -1.92408809, -3.47109889, -4.71080509, -5.71297938, -6.54739446, -7.28382303, -7.99203778, -8.74181141, -9.60291661, -10.6451261, -11.9382125, -13.5519486, -15.556107, -18.0204605, -21.0147817, -24.6088434, -29.0014498, -34.2808066, -40.2248295, -46.9309407, -54.4965627, -63.0191177, -72.596028, -83.324716, -95.3026039, -108.627114, -123.395669, -139.705691, -157.654601, -177.53442, -199.176621, -222.620247, -247.904341, -273.533178, -300.313186, -329.460063, -361.473284, -396.852324, -436.096655, -479.705753, -530.782937, -586.43136, -646.641047, -711.402025, -780.704319, -852.719685, -929.710985, -1012.12281, -1100.39977, -1194.98643, -1295.53054, -1404.2475, -1522.7008, -1651.9316, -1791.91452, -1943.10307, -2105.95073, -2281.49373, -2469.47035, -2668.98313, -2878.73226, -3100.9302, -3338.8598, -3593.41708, -3861.45374, -4141.67787, -4432.42286, -4732.29964, -5041.29324, -5358.61139, -5688.49596, -6024.34614, -6363.62714, -6703.80415, -7042.34238, -7399.62819, -7744.47531, -8068.61866, -8363.79314, -8621.73365, -8786.94825, -8929.88326, -9062.51367, -9231.31245, -9434.59913, -9656.31907, -9895.51051, -10151.2117, -10422.4608, -10708.2962, -10639.2606, -10868.4924, -11293.6728, -11812.4831, -12322.6044, -12721.7179, -12907.5048, -13201.3658, -13287.4471},
        {-544.3831, -540.254574, -535.765694, -530.926108, -525.745464, -520.23341, -514.399595, -508.253666, -501.805273, -495.064063, -488.039685, -480.741786, -473.180015, -465.364021, -457.303451, -449.007953, -440.487177, -431.750769, -422.808379, -413.669654, -404.344243, -394.841794, -385.171955, -375.344375, -365.368701, -355.254583, -345.011667, -334.649602, -324.178037, -313.60662, -302.944999, -292.202822, -281.389738, -270.515394, -259.213877, -247.690516, -236.155153, -224.629624, -213.135765, -201.695412, -190.330401, -179.062567, -167.913748, -156.905778, -146.060494, -135.399731, -124.945326, -114.719114, -104.742931, -95.0386142, -85.627998, -77.6136501, -69.9132471, -62.5251967, -55.4479069, -48.6797853, -42.2192399, -36.0646783, -30.2145085, -24.6671381, -19.420975, -14.4744269, -9.82590178, -5.4738073, -1.41655129, 2.34745846, 5.81981414, 8.87453925, 11.3887201, 13.6270499, 15.5993905, 17.315604, 18.7855523, 20.0190972, 21.026101, 21.8164253, 22.3999324, 22.786484, 22.9859422, 23.008169, 22.8630262, 22.5603759, 22.11008, 21.5220005, 20.9126372, 20.2362901, 19.4459776, 18.5481011, 17.5490619, 16.4552615, 15.273101, 14.008982, 12.6693058, 11.2604738, 9.78888724, 8.26094757, 6.68305613, 5.0616143, 3.40302342, 1.71368487, 0, -0.0868054407, -0.115489755, -0.136851141, -0.201687798, -0.360797923, -0.664979716, -1.16503137, -1.91175109, -2.95593708, -4.34838752, -6.13990062, -8.38127458, -11.1233076, -14.4167979, -18.3125436, -22.861343, -28.3888124, -34.98501, -42.0791211, -49.7808335, -58.1998351, -67.4458137, -77.6284571, -88.8574531, -101.24249, -114.893254, -129.919435, -146.43072, -164.536797, -184.57235, -206.339571, -229.880651, -255.237776, -281.319399, -308.734577, -338.394698, -370.682074, -405.979014, -444.66783, -487.130832, -534.633484, -586.406158, -642.658372, -703.599646, -769.4395, -841.366858, -918.366984, -1000.40454, -1087.44421, -1179.45064, -1273.94101, -1376.31888, -1490.49116, -1618.31092, -1757.43795, -1907.41475, -2067.78381, -2236.7787, -2415.5483, -2606.66941, -2812.21173, -3030.79001, -3262.01953, -3505.62333, -3761.50701, -4029.49235, -4309.75832, -4602.2193, -4905.48012, -5218.88634, -5548.11749, -5885.21076, -6227.31922, -6571.59594, -6915.19398, -7280.04418, -7632.3274, -7963.00227, -8263.0274, -8523.36144, -8685.03964, -8822.22623, -8947.27555, -9108.63645, -9304.37586, -9517.52135, -9746.89185, -9991.30625, -10249.5835, -10520.5424, -10677.2491, -10945.6197, -11286.0051, -11658.7566, -12024.2251, -12342.7618, -12574.7176, -12935.6241, -13119.9049},
        {-553.228513, -549.644199, -545.685461, -541.36154, -536.681678, -531.655117, -526.291099, -520.598866, -514.58766, -508.266723, -501.645296, -494.732622, -487.537943, -480.0705, -472.339536, -464.354292, -456.124011, -447.657933, -438.965302, -430.055359, -420.937346, -411.620505, -402.114078, -392.427307, -382.569434, -372.5497, -362.377348, -352.06162, -341.611757, -331.037002, -320.346596, -309.549782, -298.6558, -287.673894, -276.232023, -264.538089, -252.804579, -241.053107, -229.305289, -217.58274, -205.907076, -194.299911, -182.782861, -171.377541, -160.105567, -148.988553, -138.048116, -127.305869, -116.78343, -106.502412, -96.4844314, -87.5953294, -78.9941939, -70.6843391, -62.669079, -54.9517278, -47.5355996, -40.4240085, -33.6202685, -27.1276939, -20.9495987, -15.0892969, -9.55010284, -4.33533048, 0.551706029, 5.10769259, 9.32931508, 12.9808695, 15.8322313, 18.3593733, 20.5740463, 22.4880009, 24.112988, 25.4607584, 26.5430628, 27.3716521, 27.9582769, 28.3146882, 28.4526367, 28.3838732, 28.1201484, 27.6732132, 27.0548183, 26.2767145, 25.4348235, 24.4970401, 23.4302477, 22.2434656, 20.9457129, 19.546009, 18.0533732, 16.4768248, 14.8253829, 13.108067, 11.3338962, 9.51188993, 7.65106736, 5.76044781, 3.84905055, 1.92589485, 0, 1.96933414, 3.6425333, 4.99077029, 5.98521792, 6.597049, 6.79743635, 6.55755276, 5.84857106, 4.64166405, 2.90800455, 0.618765361, -2.2548807, -5.74176083, -9.87070221, -14.670532, -20.1700775, -26.8436512, -34.7837211, -43.0114688, -51.6511826, -60.827151, -70.6636622, -81.2850046, -92.8154665, -105.379336, -119.100903, -134.104453, -150.514277, -168.454662, -188.286605, -209.810892, -233.080801, -258.149607, -284.30271, -311.977324, -341.81543, -374.100666, -409.116671, -447.147084, -488.475544, -533.081156, -581.644776, -634.509624, -692.018921, -754.515888, -824.662628, -899.90376, -980.002783, -1064.7232, -1153.82851, -1243.99738, -1341.84849, -1452.50231, -1578.14559, -1714.97548, -1861.87836, -2017.7406, -2178.31985, -2346.34225, -2527.94706, -2729.16385, -2944.29909, -3168.6402, -3400.47366, -3643.16769, -3897.75792, -4166.48507, -4450.6972, -4747.32366, -5055.79328, -5382.85206, -5719.48592, -6062.59079, -6409.06262, -6755.79735, -7126.02123, -7483.71733, -7819.19901, -8122.77965, -8384.77261, -8543.63923, -8676.11293, -8794.72939, -8949.08988, -9137.05828, -9341.05444, -9559.71443, -9791.67432, -10035.5702, -10290.0381, -10630.6942, -10927.5199, -11192.3357, -11436.9623, -11673.2204, -11912.9305, -12167.9134, -12563.3814, -12825.8942},
        {-563.878466, -561.004309, -557.709673, -554.005266, -549.901791, -545.409955, -540.540464, -535.304022, -529.711335, -523.773109, -517.500049, -510.902861, -503.99225, -496.778922, -489.273583, -481.486937, -473.42969, -465.112549, -456.546218, -447.741403, -438.708809, -429.459142, -420.003108, -410.351412, -400.514759, -390.503856, -380.329407, -370.002118, -359.532694, -348.931842, -338.210267, -327.378674, -316.447768, -305.428256, -294.160241, -282.743318, -271.279133, -259.783926, -248.27394, -236.765417, -225.274598, -213.817725, -202.41104, -191.070786, -179.813203, -168.654534, -157.611021, -146.698905, -135.934428, -125.333832, -114.913359, -104.806923, -94.9105418, -85.2379076, -75.8027112, -66.6186435, -57.6993955, -49.0586582, -40.7101226, -32.6674795, -24.9444201, -17.5546353, -10.5118161, -3.82965337, 2.47816183, 8.39793855, 13.9159858, 18.5679496, 21.9002962, 24.8287933, 27.3689645, 29.5363336, 31.3464241, 32.8147598, 33.9568644, 34.7882615, 35.3244747, 35.5810279, 35.5734445, 35.3172484, 34.8279631, 34.1211124, 33.21222, 32.1168094, 30.9779512, 29.7447129, 28.364629, 26.849084, 25.2094623, 23.4571487, 21.6035275, 19.6599833, 17.6379007, 15.5486642, 13.4036584, 11.2142676, 8.99187659, 6.74786977, 4.49363171, 2.24054694, 0, 2.13009495, 3.91497569, 5.33035825, 6.35195869, 6.95549305, 7.11667739, 6.81122776, 6.0148602, 4.70329077, 2.85223552, 0.437410487, -2.56546827, -6.1806847, -10.4325228, -15.3452664, -20.9431995, -27.6751224, -35.6259641, -43.8597932, -52.4918617, -61.637421, -71.411723, -81.9300194, -93.3075618, -105.659602, -119.101392, -133.748183, -149.715226, -167.117775, -186.249708, -206.982152, -229.37677, -253.495226, -278.43124, -304.730446, -333.196597, -364.181741, -398.037923, -435.117189, -475.771586, -521.900452, -571.837626, -625.632424, -683.334161, -744.992152, -808.092449, -875.888447, -949.070276, -1028.32807, -1114.35195, -1208.09148, -1309.66029, -1420.8864, -1542.43264, -1672.58292, -1810.68925, -1956.10364, -2104.1721, -2259.16308, -2429.71516, -2626.08129, -2837.98911, -3056.59208, -3278.61271, -3510.20982, -3753.21081, -4011.17371, -4286.37462, -4574.74387, -4875.80134, -5196.63586, -5528.03383, -5866.89508, -6210.11941, -6554.60664, -6926.64272, -7286.3948, -7623.41616, -7927.26008, -8187.47985, -8339.43577, -8463.00275, -8570.95964, -8713.62962, -8888.93023, -9079.13557, -9282.94536, -9499.05934, -9726.17722, -9962.99874, -10151.0832, -10394.2996, -10671.8891, -10963.0926, -11247.1514, -11503.3065, -11710.7989, -12090.609, -12393.3402},
        {-575.353228, -572.975322, -570.151657, -566.893798, -563.213307, -559.121746, -554.63068, -549.75167, -544.496281, -538.876075, -532.902614, -526.587463, -519.942183, -512.978339, -505.707492, -498.141207, -490.291045, -482.168571, -473.785346, -465.152935, -456.282899, -447.186803, -437.876208, -428.362679, -418.657777, -408.773066, -398.72011, -388.51047, -378.155711, -367.667394, -357.057083, -346.336342, -335.516732, -324.609818, -313.663779, -302.6711, -291.623824, -280.532327, -269.406982, -258.258165, -247.09625, -235.931612, -224.774625, -213.635665, -202.525105, -191.453322, -180.430688, -169.467579, -158.57437, -147.761436, -137.03915, -125.822214, -114.729589, -103.784563, -93.0104242, -82.43046, -72.067958, -61.946206, -52.0884918, -42.518103, -33.2583274, -24.3324528, -15.7637668, -7.57555713, 0.208888402, 7.56628211, 14.4733363, 20.2906323, 24.3935446, 28.0295093, 31.2151797, 33.9672095, 36.3022521, 38.2369611, 39.7879901, 40.9719925, 41.8056219, 42.3055318, 42.4883757, 42.3708073, 41.9694799, 41.3010472, 40.3821626, 39.2294797, 37.8840623, 36.3498453, 34.6304703, 32.7417987, 30.6996918, 28.5200111, 26.2186179, 23.8113735, 21.3141394, 18.742777, 16.1131475, 13.4411123, 10.7425329, 8.03327061, 5.32918677, 2.64614278, 0, 2.28896277, 4.20154882, 5.71715375, 6.81517314, 7.47500258, 7.67603767, 7.39767399, 6.61930713, 5.32033269, 3.48014625, 1.0781434, -1.90628026, -5.49372916, -9.70480769, -14.5601203, -20.0802713, -26.7363319, -34.6132589, -42.7127439, -51.1519199, -60.0479198, -69.5178767, -79.6789236, -90.6481933, -102.542819, -115.479934, -129.57667, -144.950162, -161.717541, -180.234638, -200.292367, -221.936251, -245.211816, -268.929587, -293.752587, -320.673171, -350.107363, -382.471186, -418.180662, -457.651815, -504.414991, -554.824052, -608.685698, -665.806627, -725.993538, -782.7123, -843.69565, -910.335493, -984.023737, -1066.15229, -1161.02387, -1263.56191, -1372.61793, -1487.57665, -1608.91135, -1736.60659, -1870.64691, -2007.06685, -2150.69867, -2310.68372, -2498.89184, -2703.38191, -2913.96668, -3126.81056, -3348.84054, -3582.04339, -3830.21929, -4095.82515, -4374.66882, -4666.31924, -4976.86483, -5298.35204, -5628.09307, -5963.4001, -6301.58532, -6674.56111, -7036.38942, -7375.73242, -7681.25223, -7941.61101, -8089.26902, -8206.55821, -8306.22723, -8437.67402, -8599.14025, -8773.59702, -8960.04953, -9157.50295, -9364.96248, -9581.43328, -9573.15567, -9751.24588, -10051.7454, -10410.6955, -10764.1379, -11048.1138, -11198.6648, -11540.8206, -11837.8312},
        {-587.760262, -583.93655, -579.864933, -575.549007, -570.99237, -566.198619, -561.171351, -555.914164, -550.430653, -544.724416, -538.799051, -532.658154, -526.305323, -519.744154, -512.978245, -506.011192, -498.846593, -491.488045, -483.939145, -476.203491, -468.284678, -460.186304, -451.911967, -443.465263, -434.849789, -426.069143, -417.126921, -408.026721, -398.77214, -389.366775, -379.814223, -370.11808, -360.281945, -350.309414, -340.014969, -329.500743, -318.87114, -308.135896, -297.304745, -286.38742, -275.393656, -264.333187, -253.215747, -242.05107, -230.848892, -219.618945, -208.370965, -197.114685, -185.85984, -174.616163, -163.39339, -151.206006, -139.080568, -127.048387, -115.14077, -103.389028, -91.8244691, -80.4784031, -69.3821391, -58.5669862, -48.0642539, -37.9052513, -28.1212876, -18.7436722, -9.80371419, -1.33272288, 6.63799251, 13.265548, 17.7265451, 21.6634866, 25.0976745, 28.0504105, 30.5429965, 32.5967344, 34.2329259, 35.472873, 36.3378775, 36.8492412, 37.028266, 36.8962537, 36.4745062, 35.7843252, 34.8470128, 33.6838706, 32.49401, 31.206088, 29.7466247, 28.1311518, 26.3752009, 24.4943035, 22.5039912, 20.4197956, 18.2572483, 16.0318809, 13.759225, 11.4548122, 9.13417403, 6.81284215, 4.50634812, 2.23022354, 0, 7.62552032, 13.9280339, 18.953679, 22.7485943, 25.358918, 26.8307886, 27.2103446, 26.5437243, 24.8770662, 22.2565087, 18.7281903, 14.3382493, 9.13282418, 3.1580534, -3.53992465, -10.9149715, -20.8226815, -33.4885921, -44.5650269, -54.4133617, -63.394972, -71.8712336, -80.2035223, -88.7532137, -97.8816836, -107.950308, -119.320462, -132.353521, -147.410862, -166.827047, -188.266763, -211.499429, -236.294465, -260.14921, -283.969125, -309.265555, -336.489543, -366.092133, -398.524366, -434.237286, -475.386187, -520.199175, -568.793854, -621.287825, -677.798692, -741.846063, -809.295032, -879.412701, -951.466169, -1024.72254, -1093.69998, -1168.21877, -1255.04625, -1357.09707, -1469.41993, -1590.59519, -1719.20318, -1849.91458, -1986.10799, -2135.42715, -2304.24122, -2487.84494, -2684.1756, -2892.13173, -3112.24066, -3344.28206, -3587.90045, -3842.84046, -4109.34235, -4387.36605, -4681.57344, -4986.49907, -5300.47524, -5621.83426, -5948.90843, -6324.2539, -6690.92316, -7036.19255, -7347.33843, -7611.63711, -7756.79216, -7869.36791, -7962.1719, -8085.82515, -8238.16738, -8401.28266, -8573.7985, -8754.34237, -8941.54177, -9134.02418, -9063.41901, -9189.26179, -9449.76909, -9783.15749, -10127.6435, -10421.4438, -10602.7749, -10925.7305, -11053.1616},
        {-612.03065, -609.696681, -606.946018, -603.789452, -600.237775, -596.301779, -591.992255, -587.319996, -582.295791, -576.930435, -571.234716, -565.219429, -558.895363, -552.273312, -545.364066, -538.178417, -530.727156, -523.021076, -515.070968, -506.887624, -498.481835, -489.864393, -481.04609, -472.037716, -462.850065, -453.493927, -443.980095, -434.319359, -424.522512, -414.600344, -404.563649, -394.423217, -384.18984, -373.874309, -363.649976, -353.442932, -343.177318, -332.858649, -322.492443, -312.084214, -301.639481, -291.163758, -280.662563, -270.141411, -259.605819, -249.061303, -238.513379, -227.967565, -217.429375, -206.904327, -196.397936, -185.343929, -174.332007, -163.380083, -152.506067, -141.727872, -131.063409, -120.530589, -110.147326, -99.9315286, -89.9011102, -80.0739821, -70.4680558, -61.1012429, -51.9914552, -43.1566041, -34.6146013, -26.5376975, -19.0944375, -11.9844204, -5.21555274, 1.20425906, 7.26710856, 12.9650893, 18.2902948, 23.2348187, 27.7907544, 31.9501956, 35.7052357, 39.0479684, 41.9704871, 44.4648855, 46.523257, 48.1376952, 47.7534872, 46.1686648, 44.1998511, 41.8893366, 39.2794116, 36.4123666, 33.3304919, 30.0760779, 26.6914151, 23.2187938, 19.7005043, 16.1788372, 12.6960828, 9.2945315, 6.01647366, 2.9041997, 0, 0.601879743, 1.7268157, 3.28495693, 5.18645248, 7.34145141, 9.66010278, 12.0525556, 14.428959, 16.699462, 18.7742136, 20.5633629, 21.977059, 22.9254509, 23.3186877, 23.0669183, 22.080292, 21.2454576, 20.6120643, 17.8815111, 13.1731972, 6.60652147, -1.69911691, -11.6243189, -23.0496856, -35.8558177, -49.9233164, -65.1327825, -81.3648171, -98.500021, -115.650236, -133.7467, -152.900643, -173.223294, -195.173179, -218.687879, -243.61236, -269.953661, -297.718821, -326.914879, -357.548874, -387.352463, -419.300576, -453.845435, -491.439262, -532.534281, -572.931856, -618.897782, -672.046996, -733.994434, -806.355034, -897.342064, -993.907502, -1086.6618, -1172.08705, -1258.65328, -1349.44819, -1447.55944, -1558.01636, -1681.52374, -1816.66825, -1966.49037, -2127.52203, -2295.77809, -2470.09666, -2654.09964, -2849.2131, -3059.61093, -3287.43158, -3524.73814, -3769.29291, -4005.15923, -4245.90586, -4494.03702, -4752.05695, -5022.46987, -5370.67435, -5720.55669, -6058.89753, -6372.47753, -6648.07732, -6823.13223, -6966.6651, -7086.60455, -7197.18389, -7303.83616, -7409.36742, -7519.34206, -7639.32443, -7774.87891, -7931.56989, -7603.82024, -7752.52027, -8194.62063, -8747.07201, -9226.82508, -9450.83053, -9236.03904, -9642.32911, -9803.53364},
        {-620.114449, -619.577434, -618.50628, -616.915517, -614.819675, -612.233285, -609.170876, -605.64698, -601.676126, -597.272845, -592.451666, -587.227122, -581.613741, -575.626054, -569.278592, -562.585884, -555.562461, -548.222854, -540.581593, -532.653207, -524.452228, -515.993186, -507.290611, -498.359033, -489.212983, -479.866991, -470.335587, -460.633302, -450.774665, -440.774209, -430.646461, -420.405954, -410.067217, -399.644781, -389.488684, -379.438645, -369.330352, -359.167446, -348.95357, -338.692367, -328.387479, -318.042547, -307.661216, -297.247127, -286.803922, -276.335245, -265.844737, -255.336041, -244.812799, -234.278654, -223.737248, -212.847611, -201.965469, -191.101934, -180.26812, -169.47514, -158.734106, -148.056132, -137.45233, -126.933814, -116.511696, -106.197089, -96.0011071, -85.9348624, -76.009468, -66.236037, -56.6256823, -46.877144, -36.6966623, -26.7395956, -17.0373068, -7.62115868, 1.47748592, 10.2272641, 18.5968132, 26.5547702, 34.0697723, 41.1104568, 47.6454607, 53.6434212, 59.0729756, 63.902761, 68.1014144, 71.6375732, 71.5679686, 69.3784388, 66.5898206, 63.2652483, 59.4678559, 55.2607776, 50.7071475, 45.8700996, 40.8127681, 35.5982872, 30.2897908, 24.9504132, 19.6432884, 14.4315506, 9.37833385, 4.54677228, 0, 2.57954673, 5.35478685, 8.25590749, 11.2130958, 14.1565389, 17.016424, 19.7229382, 22.2062686, 24.3966024, 26.2241267, 27.6190287, 28.5114955, 28.8317142, 28.509872, 27.4761561, 25.6607535, 24.1351604, 22.9926078, 19.5817512, 14.0773436, 6.65413861, -2.51311061, -13.2496507, -25.3807283, -38.7315901, -53.1274827, -68.3936529, -84.3553473, -100.837813, -116.309138, -132.449351, -149.490848, -167.666021, -188.171407, -210.757328, -234.917004, -260.593587, -287.730228, -316.270079, -346.156291, -375.122949, -405.994597, -439.146597, -474.954308, -513.793091, -552.017411, -595.028749, -644.20769, -700.93482, -766.590723, -847.823853, -934.308421, -1018.29153, -1097.01725, -1177.93171, -1263.90055, -1357.78939, -1465.7404, -1586.59798, -1715.63206, -1851.06285, -1994.47649, -2145.61364, -2305.05986, -2474.83237, -2656.29122, -2852.78914, -3066.20278, -3291.10247, -3526.19142, -3762.01846, -4006.69569, -4261.74898, -4528.7042, -4809.08723, -5165.71166, -5523.4937, -5868.63731, -6187.34643, -6465.82499, -6638.18298, -6777.44761, -6892.1488, -7004.3131, -7117.12803, -7228.1575, -7340.87021, -7458.73485, -7585.2201, -7723.79467, -7729.89173, -7898.79706, -8164.21554, -8459.85202, -8719.41138, -8876.59847, -8865.11817, -9061.2064, -9288.13446},
        {-650.766465, -650.403575, -649.598406, -648.361621, -646.703883, -644.635856, -642.168203, -639.311588, -636.076673, -632.474123, -628.5146, -624.208768, -619.56729, -614.60083, -609.32005, -603.735615, -597.858187, -591.69843, -585.267007, -578.574582, -571.631818, -564.449378, -557.037926, -549.408125, -541.570638, -533.536129, -525.315261, -516.918698, -508.357102, -499.641137, -490.781466, -481.788754, -472.673662, -463.446855, -454.288682, -445.121395, -435.865205, -426.525268, -417.106741, -407.614781, -398.054544, -388.431188, -378.749868, -369.015742, -359.233966, -349.409697, -339.548092, -329.654307, -319.733499, -309.790825, -299.831441, -289.954177, -280.068487, -270.177496, -260.28433, -250.392115, -240.503977, -230.623043, -220.752438, -210.895287, -201.054719, -191.233857, -181.435829, -171.66376, -161.920777, -152.210005, -142.534571, -132.420842, -121.406633, -110.478349, -99.6700201, -89.0156801, -78.5493609, -68.3050948, -58.3169141, -48.6188511, -39.244938, -30.2292072, -21.605691, -13.4084217, -5.67143158, 1.57124707, 8.28558193, 14.4375407, 17.1035912, 17.7552278, 17.8986744, 17.5936687, 16.899948, 15.87725, 14.5853122, 13.0838721, 11.4326672, 9.6914351, 7.91991328, 6.17783927, 4.52495061, 3.02098482, 1.72567944, 0.698771989, 0, 15.6624449, 29.8387424, 42.5670723, 53.8856148, 63.8325498, 72.4460574, 79.7643176, 85.8255104, 90.6678158, 94.3294139, 96.8484846, 98.2632081, 98.6117642, 97.9323331, 96.2630947, 93.6422291, 90.2428625, 86.1224528, 81.0139448, 74.9844355, 68.101022, 60.4308013, 52.0408705, 42.9983266, 33.3702667, 23.2237878, 12.625987, 1.64396138, -9.65519208, -20.4974205, -31.7817998, -43.6533195, -56.2569695, -69.966155, -84.811658, -100.763349, -117.897694, -136.291157, -156.020204, -177.161299, -199.297568, -223.148962, -248.888469, -276.689077, -306.723772, -338.983303, -373.868457, -411.597784, -452.389833, -496.46315, -544.505388, -595.692644, -649.112253, -704.579087, -763.393398, -826.188532, -893.597834, -965.913192, -1044.18701, -1129.84421, -1225.96334, -1329.10873, -1434.64176, -1541.49727, -1654.66501, -1776.35544, -1910.58953, -2060.04712, -2220.76948, -2392.55304, -2570.44843, -2759.72799, -2961.8309, -3178.19635, -3410.26354, -3717.39822, -4028.63138, -4330.92057, -4611.22336, -4856.49729, -5004.82435, -5124.62137, -5223.79258, -5326.12107, -5433.22568, -5540.10906, -5648.72086, -5761.01074, -5878.92835, -6004.42335, -6164.38028, -6337.30453, -6511.18767, -6674.02129, -6813.79695, -6918.50624, -6976.14074, -7131.97789, -7322.73834}
    };
    
    const double a5tab[98][200] = {{1623.72041, 1605.90974, 1587.91237, 1569.73931, 1551.40152, 1532.91, 1514.27573, 1495.5097, 1476.62289, 1457.62629, 1438.53089, 1419.34766, 1400.0876, 1380.76169, 1361.38092, 1341.95627, 1322.49873, 1303.01928, 1283.52891, 1264.03861, 1244.55935, 1225.10213, 1205.67793, 1186.29774, 1166.97254, 1147.71332, 1128.53106, 1109.43676, 1090.44138, 1071.55593, 1052.79138, 1034.15873, 1015.66895, 997.333034, 979.683307, 962.459118, 945.393551, 928.480673, 911.714551, 895.089253, 878.598845, 862.237394, 845.998969, 829.877635, 813.867459, 797.96251, 782.156854, 766.444558, 750.81969, 735.276315, 719.808503, 704.235674, 688.730327, 673.290314, 657.913489, 642.597705, 627.340815, 612.140672, 596.995129, 581.902039, 566.859255, 551.864631, 536.916018, 522.011271, 507.148242, 492.324785, 477.538752, 462.727194, 447.828622, 432.966289, 418.14199, 403.35752, 388.614672, 373.915242, 359.261024, 344.653812, 330.095402, 315.587588, 301.132165, 286.730927, 272.385669, 258.098186, 243.870271, 229.703721, 215.852956, 202.188146, 188.576419, 175.011372, 161.486601, 147.995704, 134.532276, 121.089913, 107.662213, 94.2427704, 80.8251827, 67.403046, 53.9699567, 40.5195111, 27.0453055, 13.5409364, 0, -15.1469622, -29.9245654, -44.3840485, -58.5766505, -72.5536103, -86.3661666, -100.065559, -113.703025, -127.329805, -140.997137, -154.75626, -168.658413, -182.754835, -197.096765, -211.735442, -226.722104, -242.134493, -258.027633, -274.392818, -291.286967, -308.766998, -326.889827, -345.712374, -365.291555, -385.684289, -406.947494, -429.138087, -452.312986, -476.52911, -501.70482, -528.086393, -555.772315, -584.861071, -615.101786, -646.767625, -680.224913, -715.676944, -753.327009, -793.378402, -836.034413, -880.922158, -928.996465, -980.573357, -1035.96886, -1095.49899, -1156.62173, -1223.22565, -1296.3413, -1376.9992, -1466.22988, -1567.23867, -1676.22324, -1789.51716, -1906.34031, -2031.8055, -2168.37975, -2318.53006, -2486.4335, -2672.45837, -2875.10747, -3093.43215, -3331.35564, -3590.05281, -3872.00567, -4181.91112, -4523.44937, -4899.57496, -5313.77999, -5772.21732, -6279.53474, -6829.87618, -7440.00924, -8118.21763, -8872.7851, -9711.99536, -11017.3826, -12330.6674, -13566.821, -14640.8145, -15467.619, -15725.7937, -15724.3294, -15479.5168, -14935.8768, -14138.1085, -13160.8153, -12047.2012, -10840.4702, -9583.82634, -8320.47376, -7104.01992, -6014.93509, -5043.55193, -4180.20311, -3415.22129, -2738.93915, -2141.68936, -1611.62709, -1138.33122},
        {1520.58131, 1504.40784, 1487.98749, 1471.33453, 1454.46324, 1437.38789, 1420.12276, 1402.68212, 1385.08023, 1367.33138, 1349.44983, 1331.44987, 1313.34576, 1295.15177, 1276.88219, 1258.55127, 1240.1733, 1221.76255, 1203.33329, 1184.8998, 1166.47634, 1148.07719, 1129.71663, 1111.40892, 1093.16835, 1075.00918, 1056.94568, 1038.99213, 1021.16281, 1003.47198, 985.933919, 968.5629, 951.373196, 934.37908, 918.480071, 903.229199, 888.168854, 873.284582, 858.561927, 843.986435, 829.543651, 815.21912, 800.998387, 786.866998, 772.810497, 758.81443, 744.864341, 730.945777, 717.044282, 703.145401, 689.23468, 674.401031, 659.546069, 644.674777, 629.792136, 614.903129, 600.01274, 585.125949, 570.247739, 555.383094, 540.536994, 525.714423, 510.920363, 496.159796, 481.437705, 466.759072, 452.128879, 437.615746, 423.286764, 409.015669, 394.803318, 380.650569, 366.558278, 352.527303, 338.5585, 324.652727, 310.810841, 297.033698, 283.322157, 269.677073, 256.099305, 242.589709, 229.149142, 215.778461, 202.732087, 189.878762, 177.08418, 164.340969, 151.641759, 138.979177, 126.345853, 113.734413, 101.137488, 88.5477048, 75.9576926, 63.3600796, 50.7474944, 38.1125654, 25.447921, 12.7461897, 0, -14.6954389, -28.9918808, -42.940874, -56.5939665, -70.0027067, -83.2186427, -96.2933228, -109.278295, -122.225108, -135.185309, -148.210447, -161.352071, -174.661727, -188.190965, -201.991333, -216.114378, -230.568673, -245.399627, -260.755995, -276.680118, -293.214334, -310.400982, -328.282401, -346.900929, -366.298907, -386.518673, -407.602565, -429.592924, -452.532087, -476.177732, -500.961236, -527.010336, -554.45277, -583.482013, -614.192933, -646.66287, -680.999841, -717.31186, -755.706946, -796.293115, -837.772362, -882.086644, -929.619067, -980.75274, -1035.87077, -1094.64435, -1158.34648, -1227.53825, -1302.78073, -1384.63502, -1474.5035, -1571.07769, -1672.11545, -1777.36784, -1890.65534, -2013.97135, -2149.30926, -2299.87525, -2466.1743, -2647.38835, -2847.64345, -3060.20496, -3270.57293, -3476.48848, -3696.43455, -3939.37324, -4223.40519, -4559.86173, -4926.56629, -5320.29625, -5730.12631, -6161.7216, -6614.52583, -7087.98271, -7581.53596, -8079.00552, -8599.36482, -9145.96351, -9722.15124, -10331.2777, -11140.1695, -11879.7146, -12483.2011, -12547.82, -12139.8978, -11465.8015, -10584.8556, -9556.38464, -8439.71315, -7294.16565, -6292.48754, -5389.3908, -4578.68114, -3854.16429, -3209.64598, -2638.93193, -2135.82787, -1696.73586, -1332.38619},
        {1424.81636, 1408.58681, 1392.27309, 1375.88235, 1359.42178, 1342.89854, 1326.3198, 1309.69272, 1293.02448, 1276.32224, 1259.59317, 1242.84445, 1226.08323, 1209.31668, 1192.55198, 1175.79629, 1159.05679, 1142.34063, 1125.65499, 1109.00703, 1092.40393, 1075.85285, 1059.36096, 1042.93543, 1026.58342, 1010.31212, 994.128668, 978.040253, 962.054037, 946.177189, 930.416878, 914.78027, 899.274535, 883.906841, 869.241922, 854.996436, 840.880338, 826.882704, 812.992606, 799.19912, 785.491319, 771.858278, 758.28907, 744.772771, 731.298453, 717.855192, 704.432061, 691.018135, 677.602488, 664.174193, 650.722325, 636.419198, 622.088352, 607.736568, 593.370624, 578.997301, 564.623379, 550.255638, 535.900857, 521.565818, 507.257299, 492.98208, 478.746942, 464.558664, 450.424027, 436.34981, 422.342793, 408.5247, 395.014496, 381.581897, 368.226231, 354.946828, 341.743014, 328.61412, 315.559474, 302.578404, 289.670239, 276.834307, 264.069937, 251.376458, 238.753198, 226.199485, 213.714649, 201.298017, 189.185177, 177.252358, 165.372952, 153.538619, 141.741021, 129.97182, 118.222678, 106.485255, 94.7512139, 83.0122156, 71.2599218, 59.4859941, 47.6820939, 35.8398829, 23.9510225, 12.0071744, 0, -14.0474162, -27.7323223, -41.1021669, -54.2043988, -67.0864666, -79.795819, -92.3799046, -104.886172, -117.36207, -129.855048, -142.412554, -155.082036, -167.910944, -180.946725, -194.23683, -207.828706, -221.740556, -236.015652, -250.767592, -266.037559, -281.866734, -298.296299, -315.367436, -333.121326, -351.599152, -370.842095, -390.891337, -411.788059, -433.573443, -455.932431, -479.393067, -504.103403, -530.211495, -558.066658, -587.716315, -619.154219, -652.468046, -687.745471, -725.074169, -764.541817, -805.259254, -848.588289, -894.807717, -944.196332, -997.032931, -1054.74602, -1116.17726, -1181.318, -1250.15963, -1322.6935, -1396.67705, -1477.06595, -1566.50144, -1666.59314, -1776.84444, -1897.70439, -2029.62203, -2174.70538, -2331.36747, -2496.21157, -2666.68672, -2843.98267, -3026.81563, -3215.30179, -3411.92952, -3618.09833, -3839.60307, -4078.98278, -4322.66024, -4566.1745, -4794.71968, -5015.77129, -5228.44931, -5431.87374, -5625.16456, -5630.74346, -5668.60331, -5782.03865, -6014.34406, -6408.81408, -7328.56382, -8283.85361, -9180.9121, -9535.14268, -9407.47562, -9021.68505, -8430.55905, -7686.8857, -6843.45308, -5953.0493, -5221.07624, -4513.83935, -3841.68862, -3214.97405, -2644.04564, -2139.25339, -1710.9473, -1314.44958, -974.411877},
        {1327.80903, 1311.14287, 1294.67866, 1278.40971, 1262.32934, 1246.43085, 1230.70754, 1215.15274, 1199.75973, 1184.52184, 1169.43237, 1154.48463, 1139.67192, 1124.98756, 1110.42485, 1095.9771, 1081.63762, 1067.39972, 1053.2567, 1039.20188, 1025.22855, 1011.33003, 997.499633, 983.730657, 970.016415, 956.350213, 942.72536, 929.135163, 915.572931, 902.031972, 888.505594, 874.987105, 861.469813, 847.947025, 834.103122, 820.092374, 806.072763, 792.047624, 778.020289, 763.994092, 749.972366, 735.958445, 721.955661, 707.967348, 693.99684, 680.04747, 666.122571, 652.225476, 638.359519, 624.528033, 610.734352, 597.102396, 583.512297, 569.964775, 556.460548, 543.000335, 529.584857, 516.214832, 502.890979, 489.614018, 476.384667, 463.203647, 450.071676, 436.989474, 423.957759, 410.977252, 398.04867, 385.154436, 372.277409, 359.456048, 346.692258, 333.987944, 321.345012, 308.765367, 296.250915, 283.803561, 271.42521, 259.117769, 246.883142, 234.723235, 222.639953, 210.635202, 198.710887, 186.868914, 175.503884, 164.413097, 153.389127, 142.421137, 131.498289, 120.609745, 109.744665, 98.8922114, 88.0415463, 77.181831, 66.3022273, 55.3918968, 44.4400011, 33.435702, 22.368161, 11.2265397, 0, -13.5246175, -26.7066093, -39.5911924, -52.2235839, -64.6490007, -76.9126599, -89.0597785, -101.135574, -113.185262, -125.254062, -137.387188, -149.62986, -162.027293, -174.624705, -187.467312, -200.600333, -214.115772, -228.06553, -242.388994, -257.141406, -272.37801, -288.154048, -304.524764, -321.545401, -339.271203, -357.757411, -377.059271, -397.232024, -418.330914, -440.239374, -463.247455, -487.461943, -512.989624, -539.816087, -568.108718, -598.067221, -629.83474, -663.554423, -699.369413, -737.422856, -777.898124, -820.885893, -866.52144, -914.94004, -966.276968, -1020.28272, -1077.57355, -1138.38092, -1202.9363, -1271.47117, -1343.06206, -1420.50696, -1507.28999, -1604.68842, -1709.47394, -1820.44112, -1936.38458, -2054.21492, -2175.03891, -2302.01852, -2438.07817, -2579.62222, -2721.21321, -2861.68611, -3007.11526, -3160.25202, -3336.11742, -3540.64385, -3734.77487, -3904.90225, -4003.38626, -4055.88582, -4060.57285, -4015.61928, -3919.19702, -3442.28423, -2992.04505, -2648.44985, -2491.46899, -2601.07285, -3499.70325, -4529.87812, -5581.93721, -6150.02688, -6281.31342, -6188.0437, -5909.12997, -5483.4845, -4950.01955, -4347.64738, -3789.32258, -3227.28253, -2676.05088, -2150.15129, -1664.10742, -1232.44291, -869.681441, -504.627123, -181.241274},
        {1238.0202, 1223.30534, 1208.76191, 1194.38304, 1180.16186, 1166.09149, 1152.16505, 1138.37569, 1124.71652, 1111.18068, 1097.76128, 1084.45146, 1071.24435, 1058.13306, 1045.11074, 1032.17051, 1019.30549, 1006.50881, 993.773599, 981.092987, 968.460099, 955.868062, 943.310005, 930.779055, 918.268338, 905.770983, 893.280116, 880.788866, 868.29036, 855.777725, 843.244088, 830.682578, 818.086321, 805.448444, 792.338216, 778.969609, 765.56169, 752.121342, 738.655447, 725.170888, 711.674548, 698.173308, 684.674052, 671.183661, 657.709019, 644.257008, 630.834511, 617.448409, 604.105586, 590.812923, 577.577305, 564.834979, 552.154154, 539.532404, 526.967304, 514.456429, 501.997353, 489.587651, 477.224898, 464.906669, 452.630537, 440.394078, 428.194867, 416.030477, 403.898485, 391.796464, 379.721989, 367.566886, 355.225523, 342.91357, 330.635458, 318.395615, 306.198473, 294.04846, 281.950007, 269.907544, 257.925501, 246.008307, 234.160392, 222.386186, 210.69012, 199.076622, 187.550123, 176.115053, 165.287675, 154.805737, 144.400834, 134.060785, 123.773412, 113.526533, 103.307969, 93.1055386, 82.9070629, 72.7003615, 62.4732542, 52.2135609, 41.9091014, 31.5476957, 21.1171636, 10.6053251, 0, -13.098253, -25.8483972, -38.295585, -50.4849692, -62.4617024, -74.2709371, -85.9578261, -97.567522, -109.145177, -120.735945, -132.384977, -144.137426, -156.038446, -168.133188, -180.466805, -193.08445, -206.148486, -219.720809, -233.581454, -247.80069, -262.448787, -277.596013, -293.312637, -309.668929, -326.735158, -344.581591, -363.2785, -382.896152, -403.504816, -425.249997, -448.099143, -472.099952, -497.300123, -523.303824, -550.38052, -578.917949, -609.096868, -641.098037, -675.102212, -711.290152, -750.554404, -792.147305, -836.110349, -882.485032, -931.312847, -980.726665, -1033.15376, -1089.11279, -1149.12239, -1213.70123, -1283.69354, -1358.89444, -1442.20888, -1534.22769, -1630.81713, -1729.9647, -1829.65789, -1923.59159, -2015.02151, -2111.88617, -2222.22061, -2337.19942, -2448.23872, -2552.34353, -2655.98901, -2761.30336, -2885.83089, -3035.69658, -3160.4997, -3241.81365, -3237.60704, -3156.68958, -2988.80555, -2723.69925, -2351.11496, -1286.35104, -237.209208, 662.954777, 1280.78514, 1482.92611, 488.545296, -756.585381, -2108.33172, -2945.00299, -3311.49777, -3451.69655, -3400.54883, -3193.00411, -2864.01188, -2448.52165, -2012.40373, -1567.32092, -1125.12821, -697.680544, -296.832906, 65.5597377, 377.642418, 669.76026, 950.590204},
        {1161.2229, 1150.30802, 1139.21718, 1127.95814, 1116.53866, 1104.9665, 1093.24943, 1081.39521, 1069.4116, 1057.30636, 1045.08726, 1032.76205, 1020.33851, 1007.82439, 995.227451, 982.555463, 969.816184, 957.017379, 944.166808, 931.272235, 918.341421, 905.38213, 892.402124, 879.409164, 866.411014, 853.415436, 840.430192, 827.463045, 814.521756, 801.61409, 788.747807, 775.930671, 763.170443, 750.474887, 738.247512, 726.289882, 714.398806, 702.569204, 690.795995, 679.074098, 667.398434, 655.763922, 644.165481, 632.598032, 621.056494, 609.535787, 598.030829, 586.536542, 575.047844, 563.559656, 552.066896, 540.228709, 528.38307, 516.532176, 504.678227, 492.82342, 480.969955, 469.12003, 457.275844, 445.439594, 433.613481, 421.799702, 410.000456, 398.217942, 386.454358, 374.711902, 362.992774, 351.279581, 339.555399, 327.862834, 316.205354, 304.586428, 293.009525, 281.478112, 269.995659, 258.565635, 247.191507, 235.876746, 224.624818, 213.439194, 202.323341, 191.280728, 180.314824, 169.429097, 159.075153, 149.022966, 139.037125, 129.106554, 119.22018, 109.366928, 99.535725, 89.7154954, 79.8951654, 70.0636608, 60.2099072, 50.3228305, 40.3913563, 30.4044104, 20.3509186, 10.2198065, 0, -12.5494831, -24.7666133, -36.6948906, -48.3778155, -59.858888, -71.1816084, -82.3894769, -93.5259937, -104.634659, -115.758973, -126.942437, -138.228549, -149.660811, -161.282722, -173.137784, -185.269495, -197.84948, -210.939542, -224.293379, -237.981946, -252.076198, -266.647091, -281.765579, -297.502618, -313.929163, -331.116169, -349.134592, -368.055385, -387.949506, -408.990131, -431.108511, -454.344935, -478.739691, -504.027349, -530.401056, -558.135487, -587.362645, -618.214535, -650.823161, -685.320527, -722.316507, -761.319798, -802.368908, -845.502347, -890.758623, -937.490827, -986.594239, -1038.27872, -1092.75414, -1150.23035, -1209.85169, -1274.19587, -1347.05948, -1429.68235, -1518.08417, -1610.62834, -1705.67831, -1800.41473, -1894.65258, -1989.49709, -2083.78402, -2178.91972, -2285.90755, -2404.38913, -2514.75427, -2606.22967, -2668.04175, -2689.41714, -2659.58356, -2567.76814, -2447.12165, -2236.19004, -1908.99609, -1439.56255, -801.912187, 932.863215, 2662.07716, 4185.97414, 5304.79863, 5818.79514, 4603.57132, 3000.43303, 1205.89769, 54.0737338, -510.901207, -810.552443, -887.459202, -784.200714, -543.356207, -207.504912, 60.6260937, 374.125332, 714.967596, 1065.12768, 1406.58038, 1721.30048, 1991.26278, 2167.67433, 2390.07201},
        {1085.38834, 1075.30473, 1065.08213, 1054.72643, 1044.24355, 1033.63939, 1022.91984, 1012.09081, 1001.15821, 990.127942, 979.0059, 967.797992, 956.510121, 945.148189, 933.7181, 922.225757, 910.677063, 899.07792, 887.434233, 875.751903, 864.036834, 852.294928, 840.53209, 828.754221, 816.967225, 805.177005, 793.389464, 781.610505, 769.846031, 758.101944, 746.384148, 734.698546, 723.051041, 711.447536, 700.208465, 689.175849, 678.18793, 667.240405, 656.328967, 645.449315, 634.597143, 623.768147, 612.958024, 602.162469, 591.377177, 580.597846, 569.82017, 559.039846, 548.25257, 537.454036, 526.639942, 515.50849, 504.359318, 493.19457, 482.016393, 470.82693, 459.628327, 448.422729, 437.212281, 425.999128, 414.785414, 403.573284, 392.364885, 381.162359, 369.967854, 358.783512, 347.61148, 336.42192, 325.185762, 313.971112, 302.782189, 291.623211, 280.498397, 269.411964, 258.368131, 247.371116, 236.425137, 225.534413, 214.703162, 203.935602, 193.235951, 182.608428, 172.057251, 161.586637, 151.678879, 142.089102, 132.566687, 123.100336, 113.678756, 104.290648, 94.9247173, 85.5696676, 76.2142026, 66.8470263, 57.4568426, 48.0323552, 38.5622681, 29.0352852, 19.4401103, 9.76544725, 0, -11.8049567, -23.3441546, -34.6558498, -45.7782981, -56.7497558, -67.6084787, -78.3927228, -89.1407443, -99.8907991, -110.681143, -121.550033, -132.535723, -143.676472, -155.010533, -166.576164, -178.41162, -190.635736, -203.298278, -216.255242, -229.562153, -243.274532, -257.447903, -272.137787, -287.399708, -303.289188, -319.86175, -337.172916, -355.278209, -374.233152, -393.977559, -414.725088, -436.565973, -459.590451, -484.069039, -510.001831, -537.340845, -566.122233, -596.382144, -628.156728, -661.482137, -695.744136, -731.827204, -769.894738, -810.11014, -852.636809, -897.183388, -944.481722, -994.808899, -1048.44201, -1105.65814, -1167.25928, -1232.35607, -1301.1275, -1373.52868, -1449.05779, -1527.41815, -1608.31308, -1693.5495, -1780.24908, -1863.2387, -1935.44503, -2003.88058, -2088.34246, -2189.79165, -2277.27229, -2334.47888, -2340.60998, -2278.19449, -2146.24631, -1934.45452, -1731.42954, -1412.72037, -933.774029, -250.037551, 683.042033, 3196.00852, 5725.92634, 7995.85077, 9728.83706, 10647.9405, 9241.12618, 7289.93293, 5046.74036, 3564.33272, 2778.26277, 2290.58131, 2053.51619, 2019.29526, 2140.14635, 2368.29733, 2454.55912, 2635.42004, 2880.97284, 3161.31027, 3446.52505, 3706.70994, 3911.95768, 3988.57561, 4146.65791},
        {1007.70859, 995.099496, 982.850043, 970.943981, 959.365064, 948.097046, 937.12368, 926.42872, 915.995919, 905.80903, 895.851808, 886.108005, 876.561376, 867.195673, 857.99465, 848.942061, 840.021659, 831.217198, 822.512431, 813.891111, 805.336993, 796.833829, 788.365374, 779.91538, 771.467601, 763.005791, 754.513702, 745.97509, 737.373706, 728.693306, 719.917641, 711.030466, 702.015534, 692.856598, 682.616943, 671.759918, 660.759934, 649.630617, 638.385591, 627.038481, 615.602912, 604.092506, 592.52089, 580.901688, 569.248524, 557.575023, 545.894809, 534.221507, 522.568741, 510.950136, 499.379316, 488.716189, 478.109752, 467.555281, 457.048057, 446.583357, 436.156461, 425.762647, 415.397195, 405.055382, 394.732487, 384.423791, 374.12457, 363.830104, 353.535671, 343.236551, 332.928022, 322.415871, 311.510437, 300.597809, 289.685549, 278.781221, 267.892386, 257.026608, 246.19145, 235.394473, 224.64324, 213.945315, 203.30826, 192.739638, 182.247011, 171.837942, 161.519993, 151.300728, 141.840296, 132.80624, 123.858259, 114.982738, 106.166063, 97.3946173, 88.6547868, 79.9329563, 71.2155107, 62.488835, 53.7393142, 44.9533333, 36.1172771, 27.2175308, 18.2404791, 9.17250724, 0, -10.9545357, -21.7462535, -32.4055766, -42.9629283, -53.4487316, -63.8934099, -74.3273864, -84.7810842, -95.2849267, -105.869337, -116.564738, -127.401554, -138.410207, -149.621121, -161.064719, -172.771424, -184.783224, -197.132193, -209.822599, -222.887341, -236.359322, -250.271443, -264.656605, -279.547709, -294.977657, -310.979349, -327.585688, -344.829573, -362.743908, -380.906826, -399.972743, -420.11099, -441.490898, -465.046339, -490.564374, -517.628186, -546.177744, -576.153016, -607.493971, -640.140578, -671.958177, -705.592775, -741.390245, -779.696461, -820.857299, -862.871876, -909.019512, -960.23277, -1017.44421, -1081.58641, -1159.44059, -1238.94226, -1312.34902, -1376.16366, -1435.55626, -1491.80548, -1546.18995, -1602.79783, -1659.45975, -1710.94142, -1754.37317, -1790.29039, -1826.87536, -1862.78637, -1880.54388, -1870.07585, -1829.44831, -1750.6991, -1596.02658, -1344.508, -1094.57459, -687.587484, -61.3091052, 846.49813, 2098.0718, 5418.46803, 8791.40122, 11863.4043, 14281.0103, 15690.752, 14217.0462, 12043.2863, 9468.34081, 7708.89194, 6700.50695, 5996.33062, 5551.05128, 5319.35726, 5255.93688, 5315.47847, 5224.92699, 5281.61598, 5439.2036, 5651.34801, 5871.70736, 6053.9398, 6151.7035, 6154.7057, 6228.13996},
        {938.236429, 926.002535, 914.241285, 902.930781, 892.049127, 881.574427, 871.484784, 861.758302, 852.373084, 843.307233, 834.538854, 826.04605, 817.806923, 809.799579, 802.00212, 794.392649, 786.949271, 779.650089, 772.473206, 765.396727, 758.398753, 751.45739, 744.55074, 737.656907, 730.753994, 723.820106, 716.833345, 709.771815, 702.61362, 695.336863, 687.919647, 680.340077, 672.576255, 664.606286, 655.161155, 644.868757, 634.371973, 623.68938, 612.839551, 601.841062, 590.712487, 579.472402, 568.139381, 556.731998, 545.26883, 533.768451, 522.249435, 510.730357, 499.229793, 487.766317, 476.358504, 466.15516, 456.020127, 445.947479, 435.931291, 425.965635, 416.044585, 406.162215, 396.312599, 386.48981, 376.687923, 366.90101, 357.123145, 347.348402, 337.570855, 327.784577, 317.983642, 307.941269, 297.43598, 286.917343, 276.393752, 265.873595, 255.365264, 244.87715, 234.417644, 223.995136, 213.618017, 203.294678, 193.03351, 182.842903, 172.731249, 162.706938, 152.778361, 142.953909, 133.948364, 125.402064, 116.946854, 108.568203, 100.251577, 91.9824421, 83.7462666, 75.5285171, 67.3146606, 59.0901641, 50.8404948, 42.5511195, 34.2075053, 25.7951192, 17.2994283, 8.70589954, 0, -10.3687046, -20.6265197, -30.7994845, -40.9136387, -50.9950217, -61.0696728, -71.1636316, -81.3029374, -91.5136298, -101.821748, -112.253332, -122.834421, -133.591054, -144.54927, -155.73511, -167.174613, -178.918488, -190.996297, -203.378282, -216.095767, -229.180077, -242.66254, -256.57448, -270.947224, -285.812097, -301.200426, -317.143535, -333.672751, -350.819399, -368.097868, -386.245964, -405.450096, -425.896671, -448.670088, -473.50776, -499.90763, -527.786709, -557.062006, -587.650529, -619.469288, -650.290706, -682.829079, -717.421008, -754.403096, -794.111945, -832.854571, -876.004559, -924.905908, -980.902616, -1045.33868, -1127.97523, -1211.45154, -1285.0903, -1343.91191, -1394.56947, -1438.49326, -1477.11354, -1513.04992, -1546.27299, -1575.45586, -1605.02831, -1625.91522, -1637.22405, -1633.82346, -1603.39966, -1536.93569, -1436.74701, -1296.75468, -1069.32779, -730.339598, -404.516463, 104.021968, 870.928532, 1971.85606, 3482.45739, 7389.11142, 11379.0634, 15050.2847, 18000.7465, 19828.4202, 18439.6458, 16251.7801, 13587.7795, 11743.7885, 10657.5428, 9861.28324, 9313.02046, 8970.76507, 8792.52768, 8736.31891, 8486.01633, 8436.23775, 8519.58616, 8668.66459, 8816.07604, 8894.42352, 8836.31005, 8731.67937, 8646.91163},
        {874.787044, 865.762971, 856.97142, 848.399514, 840.034377, 831.863133, 823.872905, 816.050816, 808.383992, 800.859555, 793.46463, 786.18634, 779.011808, 771.928159, 764.922517, 757.982004, 751.093745, 744.244864, 737.422484, 730.613729, 723.805723, 716.98559, 710.140453, 703.257436, 696.323662, 689.326257, 682.252342, 675.089043, 667.823482, 660.442784, 652.934072, 645.28447, 637.481102, 629.511092, 620.555361, 611.021091, 601.325154, 591.480838, 581.501428, 571.400212, 561.190475, 550.885504, 540.498586, 530.043007, 519.532054, 508.979014, 498.397173, 487.799817, 477.200233, 466.611708, 456.047528, 446.267543, 436.522292, 426.808878, 417.124404, 407.465971, 397.830683, 388.215642, 378.61795, 369.03471, 359.463025, 349.899997, 340.342729, 330.788323, 321.233881, 311.676507, 302.113303, 292.384125, 282.332607, 272.280158, 262.234073, 252.20165, 242.190184, 232.206971, 222.259308, 212.35449, 202.499814, 192.702576, 182.970072, 173.309598, 163.728449, 154.233924, 144.833316, 135.533923, 127.018152, 118.941542, 110.951522, 103.033477, 95.1727953, 87.3548646, 79.5650723, 71.7888058, 64.0114525, 56.2184, 48.3950356, 40.526747, 32.5989214, 24.5969465, 16.5062096, 8.31209831, 0, -10.1509217, -20.1738474, -30.0951292, -39.9411187, -49.7381679, -59.5126284, -69.2908523, -79.0991912, -88.963997, -98.9116216, -108.968417, -119.160734, -129.514926, -140.057344, -150.81434, -161.812265, -173.237583, -185.139519, -197.182267, -209.426491, -221.93285, -234.762006, -247.974621, -261.631356, -275.792873, -290.519832, -305.872895, -321.912723, -338.699978, -356.130176, -374.489677, -393.888685, -414.437407, -436.562258, -460.215337, -485.264422, -511.724854, -539.611975, -568.941127, -599.727651, -631.775938, -665.376484, -700.585904, -737.460812, -776.057824, -809.967929, -847.329774, -889.816381, -939.100771, -996.855967, -1073.38623, -1151.18405, -1222.26943, -1282.12804, -1333.32119, -1375.23339, -1407.24915, -1423.3852, -1429.61374, -1437.76269, -1468.80926, -1494.2716, -1498.68504, -1469.55195, -1399.40228, -1278.45821, -1106.64968, -876.715709, -545.800336, -91.1821388, 341.294329, 968.849697, 1872.55515, 3133.48187, 4832.70105, 9017.1922, 13310.6411, 17302.6418, 20582.7885, 22740.6753, 21680.7911, 19801.2386, 17413.7997, 15750.1296, 14757.8938, 14001.4774, 13447.7097, 13063.4203, 12815.4388, 12670.5946, 12296.2055, 12167.1711, 12191.5185, 12277.2749, 12332.4674, 12265.123, 11983.269, 11743.8397, 11410.9879},
        {819.559987, 814.333273, 808.992681, 803.539157, 797.973645, 792.29709, 786.510437, 780.61463, 774.610614, 768.499334, 762.281735, 755.958761, 749.531357, 743.000469, 736.367039, 729.632015, 722.796339, 715.860957, 708.826813, 701.694853, 694.466021, 687.141262, 679.72152, 672.207741, 664.600869, 656.901848, 649.111624, 641.231141, 633.261344, 625.203178, 617.057587, 608.825516, 600.507911, 592.105715, 583.528062, 574.823733, 566.042614, 557.188629, 548.265702, 539.277757, 530.228719, 521.122512, 511.963059, 502.754286, 493.500115, 484.204472, 474.871281, 465.504466, 456.107951, 446.68566, 437.241518, 427.883887, 418.509989, 409.121484, 399.720032, 390.307294, 380.884929, 371.454598, 362.01796, 352.576677, 343.132409, 333.686814, 324.241554, 314.798289, 305.358679, 295.924384, 286.497065, 277.018312, 267.431159, 257.861154, 248.313851, 238.794805, 229.309569, 219.863697, 210.462744, 201.112265, 191.817812, 182.584941, 173.419205, 164.326159, 155.311357, 146.380354, 137.538702, 128.791957, 120.759281, 113.126517, 105.572134, 98.0817731, 90.6410768, 83.235686, 75.851242, 68.4733865, 61.0877607, 53.6800062, 46.2357643, 38.7406766, 31.1803844, 23.5405292, 15.8067525, 7.96469557, 0, -10.1747141, -20.1611166, -29.9889125, -39.6878065, -49.2875035, -58.8177085, -68.3081262, -77.7884614, -87.2884192, -96.8377043, -106.466022, -116.203076, -126.078572, -136.122215, -146.36371, -156.832761, -167.826926, -179.414176, -191.018357, -202.72657, -214.625917, -226.803501, -239.346423, -252.341786, -265.876691, -280.038241, -294.913537, -310.589681, -327.153776, -344.909269, -363.64759, -383.390936, -404.161506, -425.678138, -448.11471, -471.725995, -496.625202, -522.925535, -550.740202, -580.182408, -612.873712, -646.959904, -682.259078, -718.589328, -755.768748, -786.58192, -819.638827, -856.515941, -898.789732, -948.036673, -1012.70963, -1079.10419, -1144.09916, -1204.67414, -1258.01456, -1301.21346, -1331.36386, -1335.3245, -1322.74866, -1314.45428, -1340.8718, -1362.74588, -1359.74227, -1315.07399, -1217.96472, -1054.87915, -816.127412, -489.17115, -47.3720739, 527.932381, 1076.30076, 1823.05412, 2848.85257, 4234.3562, 6060.22515, 10310.4316, 14686.4955, 18793.249, 22235.5243, 24618.1533, 23958.7332, 22507.4878, 20549.4926, 19197.8906, 18409.9808, 17798.0341, 17336.6009, 17000.2315, 16763.4764, 16600.886, 15943.081, 15714.5216, 15755.7427, 15907.2794, 16009.6666, 15903.4394, 15429.1328, 15060.6602, 14443.3343},
        {771.4088, 767.256069, 762.933322, 758.443761, 753.790586, 748.977001, 744.006207, 738.881406, 733.605799, 728.182589, 722.614976, 716.906164, 711.059354, 705.077747, 698.964546, 692.722952, 686.356167, 679.867393, 673.259831, 666.536685, 659.701154, 652.756442, 645.705749, 638.552278, 631.299231, 623.94981, 616.507215, 608.974649, 601.355315, 593.652413, 585.869145, 578.008713, 570.07432, 562.069167, 554.041682, 545.971503, 537.837723, 529.642077, 521.386297, 513.07212, 504.701277, 496.275505, 487.796536, 479.266105, 470.685946, 462.057794, 453.383381, 444.664443, 435.902713, 427.099925, 418.257815, 409.312412, 400.332578, 391.321472, 382.282252, 373.218076, 364.132103, 355.027491, 345.907399, 336.774985, 327.633408, 318.485825, 309.335395, 300.185278, 291.03863, 281.898611, 272.768379, 263.620154, 254.426899, 245.255581, 236.111362, 226.999408, 217.924883, 208.89295, 199.908773, 190.977517, 182.104345, 173.294422, 164.552911, 155.884977, 147.295783, 138.790494, 130.374273, 122.052285, 114.441761, 107.228957, 100.092774, 93.0185111, 85.9914712, 78.996955, 72.0202637, 65.0466987, 58.0615612, 51.0501524, 43.9977736, 36.889726, 29.7113109, 22.4478294, 15.084583, 7.60687277, 0, -10.1173426, -20.0081234, -29.7040189, -39.2367053, -48.6378591, -57.9391565, -67.1722739, -76.3688875, -85.5606738, -94.7793091, -104.05647, -113.423832, -122.913072, -132.555867, -142.383892, -152.428824, -162.917562, -173.909672, -184.995254, -196.247818, -207.740874, -219.547933, -231.742502, -244.398094, -257.588216, -271.386379, -285.866093, -301.100867, -317.164211, -334.282757, -352.320748, -371.305757, -391.265358, -412.064643, -433.812426, -456.660849, -480.68623, -505.964887, -532.573137, -560.587298, -590.813144, -622.375527, -655.208047, -689.244301, -724.417886, -756.203127, -790.107715, -827.180066, -868.468597, -915.021724, -973.247793, -1032.28427, -1090.16025, -1144.41243, -1191.57213, -1228.62206, -1252.54494, -1252.84626, -1235.68532, -1215.37838, -1212.908, -1199.46483, -1166.63687, -1101.04362, -980.885744, -788.228288, -504.461219, -111.474565, 406.36635, 1066.09637, 1702.98045, 2543.78793, 3668.47398, 5156.9938, 7089.30257, 11414.0753, 15875.3674, 20085.9541, 23658.6107, 26206.1124, 25803.0672, 24625.8625, 22946.4877, 21828.0138, 21229.2942, 20779.5653, 20454.1614, 20228.4168, 20077.6659, 19977.2429, 18834.465, 18475.9339, 18627.7423, 19015.9832, 19366.7491, 19406.1328, 18860.2272, 18408.5688, 17564.9181},
        {727.568456, 724.15361, 720.536816, 716.722695, 712.715867, 708.520954, 704.142576, 699.585356, 694.853913, 689.952869, 684.886845, 679.660462, 674.278341, 668.745103, 663.06537, 657.243761, 651.284899, 645.193404, 638.973898, 632.631, 626.169334, 619.593519, 612.908176, 606.117928, 599.227394, 592.241195, 585.163954, 578.00029, 570.754826, 563.432181, 556.036977, 548.573836, 541.047377, 533.462223, 525.969474, 518.497431, 510.972633, 503.394947, 495.76424, 488.080381, 480.343236, 472.552674, 464.708561, 456.810765, 448.859155, 440.853596, 432.793957, 424.680105, 416.511907, 408.289232, 400.011946, 391.469479, 382.876699, 374.238034, 365.557915, 356.84077, 348.091029, 339.313122, 330.511477, 321.690524, 312.854693, 304.008412, 295.156112, 286.302221, 277.451169, 268.607386, 259.7753, 250.952911, 242.138373, 233.349377, 224.590769, 215.867394, 207.1841, 198.545732, 189.957137, 181.423161, 172.948649, 164.538449, 156.197406, 147.930367, 139.742178, 131.637684, 123.621733, 115.69917, 108.486848, 101.670737, 94.9294512, 88.2479768, 81.6112989, 75.0044032, 68.4122749, 61.8198995, 55.2122625, 48.5743493, 41.8911454, 35.1476361, 28.3288069, 21.4196433, 14.4051306, 7.27025438, 0, -10.0883999, -19.9094955, -29.4971498, -38.8852257, -48.107586, -57.1980935, -66.1906113, -75.119002, -84.0171287, -92.9188542, -101.858041, -110.868553, -119.984252, -129.239001, -138.666664, -148.301102, -158.269217, -168.618163, -179.17136, -189.982607, -201.105704, -212.594451, -224.502648, -236.884092, -249.792585, -263.281926, -277.405914, -292.218348, -307.773029, -324.15656, -341.377909, -359.481032, -378.509888, -398.561707, -419.653809, -441.789314, -464.996198, -489.302437, -514.736008, -541.324889, -568.856456, -597.672513, -627.848109, -659.458295, -692.578123, -725.840327, -761.122851, -798.861326, -839.491381, -883.448645, -934.783975, -985.89916, -1035.78246, -1082.56633, -1122.63595, -1153.16099, -1171.31112, -1170.72839, -1152.91181, -1123.20873, -1090.20198, -1038.69911, -974.295668, -888.161966, -747.025576, -528.833689, -206.743475, 242.539733, 824.747186, 1555.545, 2275.60995, 3202.51893, 4412.51126, 5981.8263, 7986.70338, 12289.1341, 16733.1674, 20948.6047, 24565.2473, 27212.8964, 27063.453, 26176.5522, 24806.8097, 23954.1533, 23578.1795, 23327.9374, 23178.5508, 23105.1431, 23082.838, 23086.7589, 21534.0448, 21090.3712, 21377.2867, 22016.3397, 22629.0787, 22837.0522, 22261.8089, 21747.9492, 20716.2467},
        {686.487617, 683.523078, 680.350851, 676.975962, 673.403439, 669.638306, 665.685589, 661.550315, 657.23751, 652.752198, 648.099407, 643.284162, 638.311489, 633.186414, 627.913963, 622.499162, 616.947036, 611.262613, 605.450916, 599.516974, 593.465811, 587.302453, 581.031926, 574.659257, 568.189471, 561.627594, 554.978652, 548.24767, 541.439676, 534.559695, 527.612752, 520.603874, 513.538086, 506.420415, 499.452483, 492.536883, 485.573844, 478.562011, 471.500032, 464.386552, 457.220216, 449.999672, 442.723564, 435.390539, 427.999243, 420.548322, 413.036422, 405.462188, 397.824267, 390.121305, 382.351947, 374.206299, 365.998237, 357.733094, 349.416205, 341.052906, 332.648529, 324.208411, 315.737885, 307.242286, 298.726948, 290.197207, 281.658396, 273.115851, 264.574905, 256.040893, 247.51915, 239.025567, 230.575784, 222.15336, 213.762946, 205.409192, 197.096749, 188.830266, 180.614394, 172.453783, 164.353083, 156.316944, 148.350016, 140.45695, 132.642395, 124.911003, 117.267422, 109.716303, 102.877115, 96.4341674, 90.0643787, 83.7524473, 77.4830714, 71.240949, 65.0107783, 58.7772575, 52.5250845, 46.2389576, 39.9035748, 33.5036343, 27.0238342, 20.4488726, 13.7634476, 6.95225738, 0, -10.0374334, -19.7732258, -29.2430152, -38.4824397, -47.5271376, -56.4127471, -65.1749061, -73.849253, -82.4714258, -91.0770626, -99.7018018, -108.381281, -117.151139, -126.047014, -135.104544, -144.359366, -153.826537, -163.538754, -173.578211, -183.976137, -194.763759, -205.972303, -217.632999, -229.777073, -242.435753, -255.640266, -269.42184, -283.811702, -298.841081, -314.419749, -330.744922, -347.884261, -365.905432, -385.166957, -405.59107, -427.022442, -449.441479, -472.828586, -497.164171, -522.428638, -547.504196, -573.803682, -601.522366, -630.85552, -661.998413, -696.409715, -732.705448, -770.765032, -810.467888, -851.693437, -896.395083, -939.843393, -981.556432, -1020.14444, -1052.36419, -1075.80462, -1088.05467, -1087.41363, -1070.59865, -1033.55193, -972.598618, -885.600308, -790.318829, -682.273353, -519.29411, -276.511173, 78.2761651, 571.838402, 1204.06652, 1990.05608, 2786.55696, 3791.37127, 5074.40653, 6705.57027, 8754.77002, 12956.7061, 17300.295, 21439.2461, 25027.2687, 27718.0721, 27810.7364, 27216.6863, 26170.1866, 25597.9951, 25460.2648, 25428.6097, 25477.6097, 25581.8446, 25715.8942, 25854.3384, 24080.4945, 23628.6904, 24050.6517, 24898.1038, 25722.7721, 26076.3822, 25510.6594, 24958.1238, 23800.7623},
        {646.614948, 643.861653, 640.923113, 637.803567, 634.507254, 631.038416, 627.401292, 623.600121, 619.639144, 615.5226, 611.25473, 606.839773, 602.281969, 597.585558, 592.75478, 587.793874, 582.707081, 577.498641, 572.172793, 566.733777, 561.185834, 555.533202, 549.780122, 543.930834, 537.989578, 531.960593, 525.848119, 519.656397, 513.389666, 507.052166, 500.648136, 494.181818, 487.65745, 481.079273, 474.631752, 468.225223, 461.767856, 455.258044, 448.694176, 442.074643, 435.397839, 428.662152, 421.865975, 415.007699, 408.085715, 401.098415, 394.044189, 386.921428, 379.728524, 372.463869, 365.125852, 357.374082, 349.553077, 341.668574, 333.726307, 325.732013, 317.691425, 309.610281, 301.494315, 293.349263, 285.18086, 276.994841, 268.796942, 260.592898, 252.388445, 244.189318, 236.001252, 227.847108, 219.749332, 211.678344, 203.638769, 195.635232, 187.672358, 179.754773, 171.887103, 164.073971, 156.320004, 148.629827, 141.008066, 133.459345, 125.98829, 118.599526, 111.297678, 104.087373, 97.5951359, 91.5015617, 85.4797686, 79.5142002, 73.5892999, 67.689511, 61.7992769, 55.903041, 49.9852468, 44.0303376, 38.0227569, 31.946948, 25.7873543, 19.5284193, 13.1545863, 6.65029872, 0, -9.91399075, -19.507307, -28.8163247, -37.8774196, -46.7269676, -55.4013445, -63.9369261, -72.3700883, -80.7372069, -89.0746578, -97.4188168, -105.80606, -114.272762, -122.855301, -131.59005, -140.513387, -149.534167, -158.670548, -168.247345, -178.273608, -188.758388, -199.710734, -211.139697, -223.054326, -235.463672, -248.376783, -261.802712, -275.750506, -290.229217, -304.961392, -320.347634, -336.482944, -353.462324, -371.878022, -391.576416, -412.271282, -433.908447, -456.433738, -479.792981, -503.932003, -527.256908, -551.721855, -577.573921, -605.060182, -634.427713, -668.827491, -704.876717, -742.096494, -780.007923, -818.132105, -857.158032, -894.011494, -928.072809, -958.155353, -981.914997, -997.52731, -1003.16786, -1001.34474, -984.916323, -942.014608, -859.942798, -745.320013, -622.306527, -489.22211, -300.601239, -31.0765542, 351.848043, 875.388102, 1540.62045, 2363.40744, 3228.18917, 4302.46909, 5647.6019, 7324.94229, 9395.84495, 13437.8896, 17617.1496, 21615.9235, 25116.5097, 27801.2066, 28115.7633, 27803.3939, 27076.3466, 26781.2252, 26879.1787, 27067.0416, 27319.1788, 27609.9554, 27913.7361, 28204.8861, 26512.488, 26161.7481, 26694.1133, 27651.0303, 28573.946, 29004.3071, 28483.5605, 27918.4149, 26721.9073},
        {606.399111, 603.666519, 600.80129, 597.805511, 594.681266, 591.430643, 588.055729, 584.558609, 580.94137, 577.206098, 573.35488, 569.389802, 565.312951, 561.126412, 556.832272, 552.432618, 547.929536, 543.325112, 538.621432, 533.820584, 528.924653, 523.935725, 518.855888, 513.687227, 508.43183, 503.091781, 497.669168, 492.166077, 486.584594, 480.926806, 475.194799, 469.390659, 463.516474, 457.574328, 451.648325, 445.697817, 439.681172, 433.597816, 427.447173, 421.228667, 414.941724, 408.585768, 402.160225, 395.664519, 389.098074, 382.460317, 375.75067, 368.96856, 362.113411, 355.184648, 348.181696, 340.824036, 333.397105, 325.906397, 318.357404, 310.75562, 303.106539, 295.415655, 287.688459, 279.930447, 272.147111, 264.343945, 256.526442, 248.700095, 240.870399, 233.042846, 225.22293, 217.426516, 209.669221, 201.935146, 194.229113, 186.555943, 178.920457, 171.327476, 163.781821, 156.288313, 148.851773, 141.477022, 134.168881, 126.932172, 119.771715, 112.692331, 105.698842, 98.7960685, 92.623483, 86.8552319, 81.1578334, 75.5155129, 69.9124958, 64.3330075, 58.7612733, 53.1815188, 47.5779693, 41.9348501, 36.2363867, 30.4668045, 24.6103289, 18.6511852, 12.5735989, 6.36179537, 0, -9.66761922, -19.0197322, -28.0917884, -36.9192376, -45.5375294, -53.9821133, -62.288439, -70.4919561, -78.6281142, -86.7323629, -94.8401519, -102.986931, -111.208149, -119.539256, -128.015702, -136.672936, -145.336752, -154.012647, -163.210295, -172.92022, -183.132944, -193.83899, -205.028882, -216.693143, -228.822296, -241.406864, -254.43737, -267.904338, -281.798291, -295.670557, -310.111893, -325.244578, -341.190895, -358.692528, -377.562055, -397.446885, -418.283478, -440.008295, -462.557795, -485.868442, -508.615139, -532.379855, -557.345876, -583.696488, -611.61498, -644.009851, -677.65787, -712.061022, -746.721288, -781.140653, -816.149732, -848.297991, -875.922238, -897.607656, -912.446514, -919.303412, -917.042949, -910.964446, -892.03534, -844.203366, -752.079397, -623.009814, -477.858932, -314.852565, -93.8568517, 207.654351, 615.222501, 1152.15549, 1830.70505, 2669.37689, 3592.87428, 4727.93654, 6125.53949, 7836.65891, 9912.27063, 13753.7827, 17724.1306, 21536.682, 24904.8048, 27541.8667, 28049.3796, 27993.8043, 27565.018, 27525.5294, 27838.5494, 28228.6923, 28671.0988, 29140.9096, 29613.2652, 30063.3064, 28868.6993, 28760.4011, 29353.9474, 30264.8741, 31108.7169, 31501.0116, 31057.2941, 30508.1452, 29383.124},
        {558.061705, 554.971736, 551.887198, 548.804019, 545.718126, 542.625445, 539.521902, 536.403426, 533.265941, 530.105376, 526.917656, 523.698709, 520.444462, 517.15084, 513.813771, 510.429181, 506.992998, 503.501147, 499.949556, 496.334152, 492.65086, 488.895608, 485.064323, 481.152931, 477.15736, 473.073534, 468.897383, 464.624831, 460.251807, 455.774236, 451.188045, 446.489161, 441.673512, 436.737023, 431.417555, 425.845496, 420.154336, 414.348376, 408.431918, 402.409263, 396.284712, 390.062567, 383.74713, 377.342701, 370.853583, 364.284077, 357.638484, 350.921106, 344.136243, 337.288199, 330.381274, 323.456823, 316.481291, 309.458177, 302.390977, 295.283191, 288.138317, 280.959852, 273.751296, 266.516145, 259.257899, 251.980056, 244.686114, 237.379571, 230.063925, 222.742674, 215.419317, 208.064529, 200.649773, 193.246242, 185.859562, 178.49536, 171.159261, 163.856891, 156.593877, 149.375844, 142.208419, 135.097226, 128.047893, 121.066046, 114.157309, 107.327311, 100.581675, 93.9260286, 88.0283655, 82.5491956, 77.1410679, 71.7881132, 66.4744624, 61.1842463, 55.901596, 50.6106422, 45.2955159, 39.9403479, 34.5292691, 29.0464105, 23.4759028, 17.8018771, 12.0084641, 6.07979477, 0, -9.11031261, -17.9620598, -26.5866301, -35.0154122, -43.2797945, -51.4111657, -59.4409144, -67.4004292, -75.3210985, -83.2343111, -91.1714554, -99.1639202, -107.243094, -115.440365, -123.787122, -132.314754, -140.723447, -148.997276, -157.916776, -167.442363, -177.534456, -188.153469, -199.259821, -210.813928, -222.776206, -235.107073, -247.766945, -260.716238, -273.91537, -286.591886, -299.707793, -313.443369, -327.978893, -344.02555, -361.498165, -380.169985, -400.062017, -421.195266, -443.590737, -467.269436, -493.887655, -521.333418, -549.307781, -577.511803, -605.646542, -629.277962, -653.275988, -678.375451, -705.311183, -734.818014, -773.579289, -809.110922, -833.156739, -841.927526, -840.754119, -830.872639, -813.519207, -796.396113, -772.803723, -728.988405, -663.350272, -561.032881, -410.859505, -202.240631, 74.4182491, 429.168366, 858.897346, 1370.24363, 2018.11223, 2830.10582, 3810.58591, 5013.97198, 6475.91178, 8232.05304, 10318.0435, 13921.8084, 17638.6486, 21216.1425, 24401.8685, 26943.4049, 27611.3388, 27781.5675, 27620.3801, 27788.653, 28256.9011, 28789.5609, 29367.4512, 29971.3906, 30582.1979, 31180.6917, 31780.0938, 32340.8679, 32815.583, 33156.808, 33317.1119, 33249.0637, 32905.2324, 32448.0997, 31712.4113},
        {519.490016, 516.447316, 513.469402, 510.549537, 507.680981, 504.856997, 502.070846, 499.315789, 496.585089, 493.872008, 491.169806, 488.471745, 485.771087, 483.061095, 480.335029, 477.58615, 474.807722, 471.993005, 469.135261, 466.227752, 463.263739, 460.236485, 457.13925, 453.965297, 450.707886, 447.360281, 443.915742, 440.367531, 436.708909, 432.933139, 429.033483, 425.003201, 420.835555, 416.523807, 411.657151, 406.43938, 401.079147, 395.582826, 389.956792, 384.207419, 378.341082, 372.364155, 366.283013, 360.104029, 353.833579, 347.478037, 341.043778, 334.537176, 327.964605, 321.33244, 314.647055, 308.086904, 301.482552, 294.836643, 288.151821, 281.430731, 274.676016, 267.890321, 261.07629, 254.236567, 247.373796, 240.490621, 233.589687, 226.673638, 219.745117, 212.806769, 205.861238, 198.858917, 191.751455, 184.649258, 177.558359, 170.484788, 163.434577, 156.413756, 149.428359, 142.484415, 135.587956, 128.745014, 121.961619, 115.243804, 108.5976, 102.029037, 95.5441475, 89.1489626, 83.5312924, 78.3419386, 73.2235084, 68.1599081, 63.135044, 58.1328225, 53.1371498, 48.1319323, 43.1010762, 38.028488, 32.898074, 27.6937403, 22.3993935, 16.9989397, 11.4762853, 5.81533664, 0, -8.62332975, -17.0468977, -25.2969498, -33.3997322, -41.3814907, -49.2684714, -57.0869203, -64.8630834, -72.6232067, -80.3935362, -88.2003178, -96.0697977, -104.028222, -112.101836, -120.316887, -128.699619, -137.056881, -145.383575, -154.202207, -163.49201, -173.232215, -183.402053, -193.980758, -204.94756, -216.281692, -227.962385, -239.968872, -252.280385, -264.876154, -277.138348, -289.862186, -303.206021, -317.328205, -332.796733, -349.565136, -367.477703, -386.569891, -406.877159, -428.434964, -451.278765, -477.92144, -505.167028, -532.566274, -559.669923, -586.028719, -605.575969, -624.884217, -644.908567, -666.604124, -690.925995, -725.062772, -756.117362, -777.157832, -784.56903, -781.507562, -768.089654, -744.431529, -713.130914, -671.258555, -613.178106, -548.514443, -448.500103, -284.534351, -41.1608346, 274.804373, 666.768607, 1117.37729, 1624.65503, 2262.75347, 3062.76251, 4068.40209, 5296.18896, 6772.84104, 8525.07628, 10579.6126, 13974.5718, 17472.417, 20847.0152, 23872.2334, 26321.9386, 27093.4602, 27421.561, 27455.7673, 27792.5357, 28404.4821, 29078.0012, 29795.0199, 30537.4651, 31287.2638, 32026.3428, 33438.2553, 34337.1303, 34820.2534, 34984.91, 34928.3857, 34747.9659, 34540.9362, 34115.1494, 33539.609},
        {489.408433, 486.727803, 484.10577, 481.53588, 479.011678, 476.526707, 474.074513, 471.64864, 469.242633, 466.850036, 464.464395, 462.079254, 459.688157, 457.28465, 454.862276, 452.414581, 449.935109, 447.417405, 444.855014, 442.241479, 439.570347, 436.835161, 434.029467, 431.146808, 428.18073, 425.124777, 421.972493, 418.717425, 415.353115, 411.873109, 408.270952, 404.540188, 400.674362, 396.667018, 392.127836, 387.250366, 382.232774, 377.081063, 371.801234, 366.399289, 360.88123, 355.253059, 349.520777, 343.690387, 337.76789, 331.759289, 325.670584, 319.507779, 313.276874, 306.983872, 300.634775, 294.384282, 288.086473, 281.744128, 275.360024, 268.93694, 262.477655, 255.984946, 249.461592, 242.910372, 236.334064, 229.735445, 223.117296, 216.482394, 209.833517, 203.173443, 196.504952, 189.779527, 182.949882, 176.124588, 169.309749, 162.511468, 155.735849, 148.988995, 142.277011, 135.605999, 128.982064, 122.411309, 115.899837, 109.453753, 103.07916, 96.7821619, 90.5688617, 84.4453634, 79.1120067, 74.2131749, 69.384907, 64.6107774, 59.8743605, 55.1592304, 50.4489617, 45.7271287, 40.9773055, 36.1830667, 31.3279866, 26.3956394, 21.3695995, 16.2334413, 10.9707391, 5.56506722, 0, -8.29121949, -16.422682, -24.4162191, -32.2936627, -40.0768443, -47.7875956, -55.4477484, -63.0791344, -70.7035851, -78.3429324, -86.0190079, -93.7536432, -101.56867, -109.48592, -117.527226, -125.714417, -134.085481, -142.664555, -151.436934, -160.427912, -169.662783, -179.166837, -188.96537, -199.083674, -209.547042, -220.380767, -231.610142, -243.26046, -255.357015, -267.722295, -280.658758, -294.252541, -308.589775, -324.065858, -340.612292, -358.078112, -376.456674, -395.741335, -415.925451, -437.002377, -460.638952, -484.64573, -508.688646, -532.433636, -555.546636, -574.755408, -593.398605, -611.876707, -630.590194, -649.939545, -673.709849, -694.780232, -712.775545, -726.004393, -730.088039, -721.854308, -698.131024, -652.750299, -586.216514, -502.302098, -416.137404, -295.601563, -110.297486, 155.651153, 493.516051, 905.862833, 1375.09282, 1898.54385, 2547.48956, 3351.38132, 4358.05196, 5575.74347, 7027.54439, 8736.54327, 10725.8287, 13935.2402, 17241.9275, 20439.7915, 23322.733, 25684.6528, 26516.8582, 26950.9057, 27124.6647, 27598.02, 28341.7932, 29150.9661, 30005.1524, 30883.9657, 31767.0195, 32633.9274, 33957.2201, 34895.8322, 35510.6807, 35862.6826, 36012.755, 36021.8149, 35950.7793, 35543.1358, 34936.88},
        {464.357294, 462.173788, 460.009665, 457.860265, 455.720932, 453.58701, 451.45384, 449.316765, 447.171129, 445.012273, 442.835542, 440.636277, 438.409821, 436.151518, 433.856709, 431.520739, 429.138948, 426.706681, 424.219281, 421.672089, 419.060448, 416.379703, 413.625194, 410.792266, 407.87626, 404.87252, 401.776388, 398.583208, 395.288321, 391.887071, 388.374801, 384.746853, 380.99857, 377.125296, 372.845392, 368.298517, 363.62766, 358.837151, 353.931322, 348.914504, 343.791029, 338.565227, 333.24143, 327.82397, 322.317177, 316.725383, 311.05292, 305.304118, 299.48331, 293.594825, 287.642996, 281.67491, 275.651215, 269.575316, 263.450618, 257.280524, 251.068439, 244.817767, 238.531914, 232.214283, 225.868279, 219.497307, 213.10477, 206.694074, 200.268622, 193.83182, 187.387071, 180.89728, 174.326326, 167.761139, 161.207749, 154.672187, 148.160482, 141.678664, 135.232762, 128.828808, 122.47283, 116.170858, 109.928923, 103.753054, 97.6492815, 91.6236346, 85.6821435, 79.8308381, 74.779627, 70.1677956, 65.6261691, 61.1379352, 56.6862815, 52.2543957, 47.8254654, 43.3826783, 38.909222, 34.3882842, 29.8030526, 25.1367147, 20.3724583, 15.493471, 10.4829404, 5.32405417, 0, -8.15644, -16.158845, -24.0269888, -31.7806452, -39.4395878, -47.0235905, -54.552427, -62.0458711, -69.5236965, -77.005677, -84.5115864, -92.0611984, -99.6742867, -107.370625, -115.169988, -123.092148, -131.387588, -140.10904, -148.754437, -157.392991, -166.093912, -174.926412, -183.959702, -193.262994, -202.905498, -212.956427, -223.484992, -234.560403, -246.251873, -258.835952, -272.098488, -286.046489, -300.686966, -316.289529, -332.729886, -349.81372, -367.469259, -385.624732, -404.208367, -423.148394, -442.573083, -462.14974, -481.767454, -501.315316, -520.682416, -540.972454, -560.55626, -579.01927, -595.946925, -610.924661, -623.222203, -633.126577, -647.695988, -667.533272, -680.061549, -678.62136, -656.553248, -600.044181, -511.214088, -399.986802, -280.31622, -125.897951, 83.9239918, 361.99354, 707.917523, 1127.37882, 1611.87743, 2163.41919, 2836.01381, 3654.25464, 4649.21847, 5838.01781, 7243.00177, 8886.51948, 10790.92, 13825.7827, 16954.4189, 19987.3702, 22735.1783, 25008.3846, 25875.1884, 26383.3687, 26661.6139, 27245.2564, 28102.0214, 29030.1991, 30005.9864, 31005.5804, 32005.1779, 32980.9759, 33663.6116, 34465.5205, 35295.0006, 36060.35, 36669.8667, 37031.8489, 37054.5946, 36717.3372, 35993.1069},
        {437.858147, 435.988536, 434.121511, 432.253331, 430.380252, 428.498532, 426.60443, 424.694201, 422.764104, 420.810396, 418.829335, 416.817178, 414.770182, 412.684606, 410.556706, 408.382741, 406.158967, 403.881642, 401.547024, 399.151371, 396.690938, 394.161985, 391.560769, 388.883547, 386.126576, 383.286115, 380.35842, 377.339749, 374.22636, 371.01451, 367.700457, 364.280457, 360.75077, 357.107651, 353.134358, 348.938127, 344.628759, 340.209422, 335.683288, 331.053525, 326.323303, 321.495793, 316.574164, 311.561586, 306.461228, 301.276261, 296.009854, 290.665177, 285.245401, 279.753694, 274.193227, 268.541944, 262.828788, 257.057474, 251.23172, 245.355241, 239.431755, 233.464979, 227.458629, 221.416421, 215.342072, 209.2393, 203.111819, 196.963348, 190.797603, 184.618301, 178.429157, 172.19617, 165.886244, 159.580887, 153.286261, 147.008527, 140.753849, 134.528387, 128.338303, 122.18976, 116.088918, 110.04194, 104.054988, 98.1342228, 92.285807, 86.5159021, 80.83067, 75.2362725, 70.4573787, 66.1257839, 61.8646476, 57.6568145, 53.4851296, 49.3324378, 45.1815839, 41.0154127, 36.8167692, 32.5684982, 28.2534445, 23.8544531, 19.3543688, 14.7360365, 9.98230102, 5.07600723, 0, -8.58777823, -16.9014908, -24.9692633, -32.8192215, -40.479491, -47.9781975, -55.3434667, -62.6034243, -69.7861959, -76.9199073, -84.0326841, -91.1526519, -98.3079365, -105.526664, -112.836959, -120.266948, -127.807921, -135.482743, -143.402856, -151.588492, -160.059883, -168.837263, -177.940863, -187.390917, -197.207656, -207.411314, -218.022123, -229.060314, -240.546122, -252.425613, -264.820378, -277.772901, -291.325663, -306.011654, -321.628102, -337.841434, -354.54698, -371.64007, -389.016034, -406.5702, -424.053878, -441.550252, -458.982831, -476.275123, -493.350636, -511.343077, -528.663207, -544.931985, -559.77037, -572.799321, -584.117655, -592.284425, -601.44678, -611.445417, -613.328677, -602.092491, -572.732788, -520.13116, -439.42386, -325.871875, -174.041657, 19.7585397, 253.009577, 530.57242, 869.813207, 1282.358, 1774.33065, 2355.93074, 3057.53253, 3898.09837, 4892.64159, 6064.68102, 7434.54595, 9022.56565, 10849.0694, 13701.3883, 16641.0993, 19496.7813, 22097.0131, 24270.3735, 25171.2263, 25751.8421, 26129.749, 26799.2761, 27728.8013, 28729.7018, 29778.2056, 30850.5403, 31922.9338, 32971.614, 32823.8402, 33472.1276, 34599.8692, 35890.4581, 37027.2873, 37693.7496, 37573.2382, 37660.7647, 36897.7784},
        {411.415967, 409.725115, 408.037275, 406.348891, 404.656405, 402.956262, 401.244905, 399.51878, 397.774328, 396.007995, 394.216225, 392.39546, 390.542145, 388.652724, 386.72364, 384.751338, 382.732262, 380.662854, 378.53956, 376.358823, 374.117086, 371.810794, 369.436391, 366.99032, 364.469026, 361.868951, 359.186541, 356.418239, 353.560488, 350.609733, 347.562417, 344.414985, 341.16388, 337.805546, 334.147319, 330.284173, 326.313359, 322.237456, 318.059046, 313.780709, 309.405027, 304.934579, 300.371947, 295.719711, 290.980453, 286.156753, 281.251192, 276.266351, 271.20481, 266.069151, 260.861953, 255.532925, 250.138667, 244.682906, 239.169369, 233.601783, 227.983875, 222.319371, 216.612, 210.865488, 205.083562, 199.269949, 193.428376, 187.56257, 181.676258, 175.773167, 169.857024, 163.890082, 157.835593, 151.782817, 145.73817, 139.708068, 133.698927, 127.717162, 121.769188, 115.861421, 110.000277, 104.19217, 98.443518, 92.7607347, 87.1502361, 81.6184379, 76.1717555, 70.8166045, 66.2956387, 62.2316687, 58.2386561, 54.2991242, 50.3955961, 46.510595, 42.6266442, 38.7262667, 34.7919858, 30.8063248, 26.7518067, 22.6109548, 18.3662923, 14.0003424, 9.49562822, 4.83467303, 0, -9.09101917, -17.7716702, -26.0799016, -34.0536619, -41.7308997, -49.1495636, -56.3476021, -63.3629637, -70.2335971, -76.9974507, -83.6924731, -90.3566129, -97.0278187, -103.744039, -110.543222, -117.463317, -124.131322, -130.526478, -137.616264, -145.350568, -153.679278, -162.552281, -171.919465, -181.730719, -191.935929, -202.484984, -213.327771, -224.414179, -235.694095, -246.592176, -257.776126, -269.353401, -281.431459, -294.83442, -309.311409, -324.420442, -340.053978, -356.104474, -372.464389, -389.02618, -405.947825, -422.775452, -439.349573, -455.510694, -471.099325, -485.155842, -498.520921, -511.235103, -523.338931, -534.872947, -548.888216, -558.73523, -562.239745, -557.412202, -542.640106, -516.141669, -476.1351, -430.152155, -364.980786, -257.248717, -80.3386097, 146.244443, 397.308067, 669.873633, 998.602474, 1400.87856, 1896.94181, 2504.91662, 3232.45562, 4093.13488, 5089.1575, 6247.22217, 7584.84176, 9119.5291, 10868.7971, 13529.3937, 16269.7878, 18937.6835, 21380.7848, 23446.7958, 24383.6288, 25038.6407, 25516.5891, 26259.3114, 27238.1379, 28283.9489, 29375.0971, 30489.9352, 31606.816, 32704.0921, 31863.5543, 32401.3921, 33802.9644, 35553.6301, 37138.748, 38043.677, 37753.776, 38296.6427, 37574.3219},
        {385.697795, 384.049735, 382.425538, 380.820892, 379.231488, 377.653013, 376.081157, 374.511607, 372.940054, 371.362186, 369.773691, 368.170258, 366.547576, 364.901335, 363.227222, 361.520927, 359.778138, 357.994544, 356.165834, 354.287697, 352.355822, 350.365897, 348.313611, 346.194653, 344.004711, 341.739475, 339.394634, 336.965876, 334.448889, 331.839364, 329.132987, 326.32545, 323.412439, 320.389644, 317.029723, 313.444572, 309.748765, 305.945231, 302.036895, 298.026684, 293.917524, 289.712343, 285.414067, 281.025621, 276.549934, 271.989931, 267.34854, 262.628686, 257.833296, 252.965297, 248.027615, 243.005714, 237.920362, 232.774865, 227.572526, 222.316653, 217.010549, 211.65752, 206.260871, 200.823907, 195.349933, 189.842254, 184.304177, 178.739005, 173.150044, 167.540598, 161.913974, 156.220118, 150.410258, 144.597747, 138.789348, 132.991827, 127.211946, 121.45647, 115.732164, 110.045792, 104.404116, 98.8139033, 93.281916, 87.8149189, 82.419676, 77.1029515, 71.8715095, 66.7321142, 62.4480651, 58.6319459, 54.8872474, 51.1961829, 47.5409653, 43.903808, 40.2669238, 36.6125261, 32.9228279, 29.1800424, 25.3663826, 21.4640617, 17.4552928, 13.3222891, 9.04726362, 4.61242955, 0, -9.19240262, -17.9255187, -26.2401917, -34.1772649, -41.7775819, -49.0819858, -56.1313203, -62.9664286, -69.6281541, -76.1573403, -82.5948305, -88.9814682, -95.3580967, -101.765559, -108.2447, -114.836361, -121.026589, -126.776969, -133.383247, -140.767379, -148.851324, -157.55704, -166.806485, -176.521617, -186.624394, -197.036775, -207.680717, -218.478179, -229.351118, -239.535129, -249.890201, -260.544201, -271.624994, -283.921002, -297.229816, -311.172877, -325.679887, -340.680546, -356.104555, -371.881616, -389.321746, -406.554232, -423.238713, -439.034828, -453.602216, -463.073647, -471.517346, -479.474668, -487.48697, -496.095609, -511.051249, -521.319003, -520.709697, -505.92985, -479.59803, -441.678415, -392.135184, -344.101939, -281.380385, -173.40558, 13.6576259, 252.57023, 510.348045, 781.473651, 1106.97054, 1506.49916, 2007.09021, 2630.31497, 3370.7204, 4238.13995, 5230.95885, 6372.22003, 7675.71975, 9155.25432, 10824.62, 13279.1249, 15805.6756, 18272.6903, 20548.5874, 22501.7853, 23483.8403, 24224.6073, 24814.0169, 25629.9543, 26648.3681, 27725.2256, 28842.4746, 29982.0627, 31125.9374, 32256.0463, 31248.3441, 31802.4825, 33341.6946, 35289.2131, 37068.2712, 38102.1016, 37813.9373, 38535.1623, 37975.9562},
        {360.792502, 359.016758, 357.315149, 355.681202, 354.108442, 352.590396, 351.120589, 349.692548, 348.299799, 346.935867, 345.594279, 344.268561, 342.952239, 341.638838, 340.321886, 338.994908, 337.65143, 336.284978, 334.889079, 333.457258, 331.983041, 330.459955, 328.881525, 327.241278, 325.532739, 323.749436, 321.884892, 319.932636, 317.886193, 315.739088, 313.484849, 311.117, 308.629069, 306.014581, 302.920439, 299.520771, 295.993873, 292.344518, 288.577481, 284.697538, 280.709462, 276.618028, 272.428011, 268.144186, 263.771326, 259.314208, 254.777605, 250.166292, 245.485044, 240.738636, 235.931841, 231.187145, 226.38906, 221.53981, 216.641618, 211.696705, 206.707297, 201.675614, 196.603881, 191.49432, 186.349154, 181.170607, 175.9609, 170.722257, 165.456901, 160.167055, 154.854941, 149.4457, 143.866322, 138.278009, 132.687979, 127.103453, 121.531652, 115.979794, 110.4551, 104.964789, 99.5160827, 94.1161995, 88.7723597, 83.4917834, 78.2816903, 73.1493005, 68.1018338, 63.1465102, 59.0703376, 55.4742402, 51.9497696, 48.4788401, 45.0433659, 41.6252614, 38.2064409, 34.7688187, 31.294309, 27.7648263, 24.1622846, 20.4685985, 16.6656821, 12.7354497, 8.65981573, 4.42069439, 0, -8.43291001, -16.5444991, -24.3635621, -31.918894, -39.2392896, -46.3535438, -53.2904516, -60.0788076, -66.747407, -73.3250445, -79.840515, -86.3226134, -92.8001346, -99.3018735, -105.856625, -112.493184, -119.081671, -125.628213, -132.520512, -139.75336, -147.32155, -155.219876, -163.443131, -171.986109, -180.843602, -190.010405, -199.481309, -209.25111, -219.314599, -229.524715, -240.070121, -250.988165, -262.316197, -274.130838, -286.449803, -299.280333, -312.647995, -326.578358, -341.096991, -356.22946, -375.378396, -394.164503, -411.952623, -428.107594, -441.994254, -446.702732, -449.441255, -451.143341, -452.742507, -455.172271, -464.839368, -470.514606, -469.17106, -458.058807, -434.993478, -397.53678, -343.250421, -273.783723, -181.681778, -55.0304622, 117.965413, 333.708319, 580.970467, 862.190915, 1202.96643, 1618.26403, 2127.23033, 2745.91593, 3475.04617, 4324.01522, 5304.30458, 6422.89997, 7688.473, 9109.69532, 10695.2385, 12921.3454, 15212.2237, 17459.6522, 19555.4098, 21391.2753, 22438.416, 23289.63, 24016.9582, 24919.5089, 25979.4422, 27084.3068, 28220.9933, 29376.3925, 30537.395, 31690.8914, 31470.8523, 32267.71, 33694.1937, 35363.0329, 36886.9569, 37878.6951, 37950.9767, 38277.7327, 38076.4396},
        {339.029877, 337.298818, 335.660318, 334.107193, 332.632259, 331.22833, 329.888224, 328.604755, 327.37074, 326.178993, 325.022331, 323.89357, 322.785525, 321.691012, 320.602847, 319.513844, 318.416821, 317.304593, 316.169975, 315.005783, 313.804834, 312.559942, 311.263923, 309.909593, 308.489768, 306.997264, 305.424895, 303.765479, 302.01183, 300.156765, 298.193098, 296.113647, 293.911226, 291.578651, 288.729224, 285.5535, 282.246597, 278.813645, 275.259776, 271.590122, 267.809815, 263.923986, 259.937767, 255.85629, 251.684686, 247.428087, 243.091624, 238.68043, 234.199636, 229.654373, 225.049774, 220.532901, 215.963878, 211.34476, 206.677601, 201.964457, 197.207382, 192.408432, 187.569661, 182.693123, 177.780875, 172.83497, 167.857464, 162.850412, 157.815868, 152.755887, 147.672524, 142.486527, 137.120594, 131.744471, 126.365485, 120.99096, 115.628224, 110.284601, 104.967418, 99.6839996, 94.4416722, 89.2477613, 84.1095928, 79.0344923, 74.0297856, 69.1027985, 64.2608568, 59.511286, 55.6551138, 52.2861205, 48.988547, 45.7439617, 42.5339334, 39.3400305, 36.1438217, 32.9268756, 29.6707608, 26.3570458, 22.9672994, 19.48309, 15.8859863, 12.1575568, 8.27937022, 4.23299508, 0, -7.88151716, -15.5363533, -22.9851023, -30.2483579, -37.3467139, -44.3007642, -51.1311024, -57.8583225, -64.5030183, -71.0857834, -77.6272118, -84.1478973, -90.6684336, -97.2094146, -103.791434, -110.435086, -117.216965, -124.165667, -131.175114, -138.2779, -145.506618, -152.893865, -160.472233, -168.274317, -176.332711, -184.680009, -193.348807, -202.371697, -211.781274, -221.756418, -232.1298, -242.890128, -254.026109, -265.491693, -277.292968, -289.445292, -301.9478, -314.799629, -327.999915, -341.547795, -358.258815, -374.458534, -389.595049, -403.116462, -414.47087, -419.026597, -421.331464, -421.853514, -421.060791, -419.421339, -421.203648, -418.430324, -411.020054, -397.127723, -371.307131, -329.728889, -268.563612, -181.669306, -68.0547831, 70.7483046, 229.19554, 423.2741, 660.207457, 949.305713, 1303.41465, 1733.75711, 2254.33959, 2877.10663, 3603.79603, 4441.91908, 5404.57369, 6492.82509, 7712.25078, 9068.42825, 10566.935, 12594.417, 14680.1162, 16734.3428, 18667.4072, 20389.6198, 21462.92, 22378.2363, 23195.1809, 24163.7886, 25268.296, 26413.5968, 27587.8946, 28779.393, 29976.2957, 31166.8062, 31799.3112, 32841.6629, 34112.142, 35429.0289, 36610.6045, 37475.1492, 37840.9436, 38017.8981, 37613.5646},
        {320.562039, 319.100885, 317.710356, 316.384393, 315.116939, 313.901937, 312.73333, 311.605061, 310.511072, 309.445307, 308.401707, 307.374218, 306.356779, 305.343336, 304.32783, 303.304205, 302.266403, 301.208367, 300.12404, 299.007365, 297.852284, 296.652741, 295.402678, 294.096038, 292.726764, 291.288799, 289.776086, 288.182567, 286.502185, 284.728883, 282.856605, 280.879292, 278.790887, 276.585335, 273.956633, 271.055007, 268.034283, 264.898139, 261.65025, 258.294295, 254.833948, 251.272886, 247.614787, 243.863327, 240.022182, 236.095029, 232.085544, 227.997404, 223.834286, 219.599866, 215.29782, 210.968409, 206.577934, 202.129276, 197.62532, 193.068949, 188.463047, 183.810497, 179.114182, 174.376986, 169.601793, 164.791486, 159.948948, 155.077063, 150.178715, 145.256787, 140.314162, 135.288878, 130.12053, 124.945741, 119.771598, 114.605188, 109.453598, 104.323916, 99.2232281, 94.1586221, 89.137185, 84.166004, 79.2521663, 74.402759, 69.6248692, 64.9255842, 60.3119911, 55.7911771, 52.1683483, 49.0347451, 45.9720149, 42.9613444, 39.9839204, 37.0209295, 34.0535583, 31.0629935, 28.0304218, 24.9370299, 21.7640044, 18.492532, 15.1037994, 11.5789932, 7.89930015, 4.04590685, 0, -7.74996501, -15.2793615, -22.6079905, -29.7556527, -36.7421492, -43.5872808, -50.3108484, -56.9326529, -63.4724952, -69.9501763, -76.3854969, -82.798258, -89.2082606, -95.6353055, -102.099194, -108.619726, -115.190983, -121.829091, -128.612028, -135.554084, -142.669547, -149.972708, -157.477855, -165.199279, -173.151267, -181.348111, -189.804099, -198.533521, -207.550666, -216.906631, -226.565402, -236.530226, -246.804352, -257.900332, -269.566763, -281.41642, -293.299759, -305.067236, -316.569304, -327.656419, -337.920168, -347.54866, -356.442997, -364.504283, -371.63362, -380.390918, -387.353772, -391.758583, -392.841752, -389.839679, -382.773189, -369.135523, -349.915131, -324.078246, -286.462394, -233.758806, -162.658714, -66.8637881, 53.2657348, 194.108278, 344.805706, 528.774519, 760.847987, 1053.32592, 1414.21535, 1853.49374, 2385.70493, 3022.01027, 3756.82781, 4594.04655, 5535.05016, 6586.61842, 7753.50757, 9040.47383, 10452.2735, 12322.0152, 14244.0147, 16140.94, 17935.4591, 19550.2403, 20601.4326, 21522.5691, 22367.683, 23373.8812, 24522.7304, 25719.5168, 26949.6211, 28198.4242, 29451.3069, 30693.6499, 31910.1282, 33116.3789, 34268.5851, 35322.9303, 36235.5976, 36962.7705, 37460.6322, 37838.5793, 36732.3546},
        {302.36832, 301.230796, 300.139075, 299.088253, 298.073426, 297.08969, 296.13214, 295.195872, 294.275981, 293.367564, 292.465717, 291.565534, 290.662112, 289.750547, 288.825934, 287.883368, 286.917947, 285.924764, 284.898917, 283.835501, 282.729612, 281.576345, 280.370796, 279.108061, 277.783236, 276.391416, 274.927697, 273.387175, 271.764945, 270.056104, 268.255747, 266.35897, 264.360869, 262.256538, 259.806681, 257.12852, 254.342095, 251.450108, 248.455262, 245.360258, 242.1678, 238.88059, 235.501329, 232.03272, 228.477466, 224.838268, 221.117829, 217.318851, 213.444036, 209.496087, 205.477706, 201.375439, 197.208495, 192.979927, 188.692785, 184.350125, 179.954997, 175.510454, 171.019549, 166.485334, 161.910862, 157.299185, 152.653356, 147.976428, 143.271452, 138.541481, 133.789569, 128.955338, 123.979936, 118.997231, 114.014389, 109.038573, 104.076948, 99.1366779, 94.2249271, 89.3488599, 84.5156407, 79.7324338, 75.0064035, 70.3447142, 65.7545301, 61.2430156, 56.8173349, 52.4846525, 49.0615522, 46.1334635, 43.2760873, 40.4703206, 37.6970605, 34.9372042, 32.1716488, 29.3812912, 26.5470286, 23.6497582, 20.6703769, 17.589782, 14.3888704, 11.0485394, 7.54968588, 3.87320706, 0, -7.84710341, -15.4336768, -22.7820366, -29.9144995, -36.853382, -43.6210005, -50.2396717, -56.7317121, -63.1194382, -69.4251665, -75.6712136, -81.8798961, -88.0735304, -94.2744332, -100.504921, -106.78731, -112.989396, -119.111421, -125.525214, -132.219979, -139.184921, -146.409246, -153.882157, -161.592861, -169.530561, -177.684463, -186.043771, -194.597691, -203.335427, -212.153867, -221.168383, -230.395875, -239.853243, -250.23036, -261.207639, -272.286036, -283.280562, -294.006226, -304.278036, -313.911002, -321.71295, -328.812606, -335.222038, -340.953312, -346.018495, -355.114358, -362.397088, -366.707576, -366.88671, -361.775384, -350.592043, -331.338564, -303.230351, -264.926171, -213.948318, -148.329339, -66.1017787, 35.5966289, 157.635157, 299.906916, 453.756076, 640.846335, 875.661944, 1169.99514, 1531.07672, 1968.23123, 2495.68397, 3124.03006, 3845.89533, 4664.07004, 5581.73815, 6601.23569, 7725.21668, 8956.3351, 10297.245, 12030.5152, 13808.9061, 15565.0931, 17231.7513, 18741.5561, 19752.454, 20655.0015, 21499.6148, 22516.5387, 23685.0073, 24909.326, 26172.4751, 27457.4353, 28747.187, 30024.7105, 30980.4679, 32124.5812, 33351.6671, 34556.3421, 35633.2229, 36476.9261, 36982.0684, 37434.8729, 36403.1006},
        {284.82981, 283.938005, 283.078457, 282.246869, 281.438947, 280.650395, 279.87692, 279.114225, 278.358017, 277.603999, 276.847877, 276.085356, 275.312141, 274.523937, 273.716449, 272.885382, 272.026441, 271.135332, 270.207758, 269.239426, 268.22604, 267.163305, 266.046927, 264.872609, 263.636058, 262.332979, 260.959076, 259.510054, 257.981618, 256.369475, 254.669327, 252.876881, 250.987842, 248.997914, 246.691551, 244.174528, 241.555157, 238.835999, 236.019615, 233.108564, 230.105409, 227.012709, 223.833024, 220.568916, 217.222945, 213.797672, 210.295656, 206.719459, 203.071641, 199.354762, 195.571384, 191.728486, 187.824114, 183.860731, 179.840804, 175.766796, 171.641173, 167.466399, 163.244939, 158.979257, 154.671819, 150.325089, 145.941532, 141.523613, 137.073796, 132.594546, 128.088328, 123.479069, 118.69258, 113.893305, 109.088802, 104.286624, 99.4943298, 94.7194739, 89.9696128, 85.2523024, 80.5750989, 75.9455583, 71.3712367, 66.85969, 62.4184744, 58.0551459, 53.7772606, 49.5923744, 46.3358365, 43.5838948, 40.902848, 38.2733886, 35.6762094, 33.0920027, 30.5014613, 27.8852776, 25.2241443, 22.4987539, 19.689799, 16.7779723, 13.7439662, 10.5684733, 7.23218629, 3.71579766, 0, -8.00176532, -15.6933236, -23.0999156, -30.246782, -37.1591636, -43.8623012, -50.3814354, -56.7418071, -62.9686569, -69.0872257, -75.1227541, -81.1004829, -87.0456528, -92.9835047, -98.9392792, -104.938217, -110.771539, -116.431053, -122.471333, -128.867472, -135.594564, -142.627702, -149.94198, -157.512491, -165.314329, -173.322586, -181.512358, -189.858736, -198.336816, -206.807368, -215.401726, -224.129279, -232.999417, -242.380716, -252.102974, -261.900203, -271.674037, -281.326111, -290.758059, -299.871514, -309.224214, -317.862005, -325.558154, -332.085926, -337.218588, -342.822137, -346.053924, -346.164032, -342.402542, -334.019538, -322.428565, -302.072011, -268.831596, -220.301089, -157.569689, -80.1572171, 12.4165036, 118.285742, 240.809747, 385.906942, 551.49977, 751.592389, 995.140527, 1291.02521, 1648.00087, 2074.88006, 2584.07291, 3185.32469, 3875.18945, 4657.68293, 5550.24509, 6541.4022, 7630.52319, 8816.97695, 10100.1324, 11711.3045, 13359.9295, 14987.39, 16535.0683, 17944.3468, 18909.6485, 19783.9548, 20614.4879, 21622.2098, 22785.5705, 24010.6288, 25279.4542, 26574.1166, 27876.6857, 29169.2313, 29299.6661, 30233.537, 31670.672, 33310.8993, 34854.0468, 35999.9427, 36448.4151, 36828.1002, 36607.012},
        {271.159829, 270.16653, 269.219911, 268.315261, 267.447868, 266.613017, 265.805999, 265.022099, 264.256606, 263.504808, 262.761992, 262.023446, 261.284457, 260.540314, 259.786303, 259.017713, 258.229832, 257.417946, 256.577344, 255.703313, 254.791141, 253.836116, 252.833525, 251.778657, 250.666798, 249.493236, 248.25326, 246.942156, 245.555213, 244.087718, 242.534959, 240.892223, 239.154798, 237.317973, 235.144439, 232.750674, 230.255951, 227.663106, 224.974975, 222.194394, 219.324198, 216.367224, 213.326307, 210.204282, 207.003987, 203.728256, 200.379926, 196.961832, 193.47681, 189.927696, 186.317327, 182.693721, 179.013552, 175.278675, 171.490947, 167.652223, 163.764361, 159.829216, 155.848645, 151.824504, 147.75865, 143.652938, 139.509226, 135.329369, 131.115223, 126.868645, 122.591492, 118.193089, 113.584984, 108.95987, 104.325601, 99.6900331, 95.06102, 90.4464168, 85.8540779, 81.2918581, 76.7676121, 72.2891946, 67.8644602, 63.5012636, 59.2074595, 54.9909025, 50.8594474, 46.8209488, 43.7272667, 41.1465005, 38.6366055, 36.1780469, 33.7512896, 31.3367988, 28.9150395, 26.4664769, 23.971576, 21.4108019, 18.7646196, 16.0134943, 13.137891, 10.1182749, 6.93511094, 3.56886429, 0, -7.79579543, -15.309652, -22.5636365, -29.5798157, -36.3802564, -42.9870253, -49.4221893, -55.7078151, -61.8659695, -67.9187193, -73.8881313, -79.7962722, -85.6652088, -91.517008, -97.3737364, -103.257461, -109.075753, -114.834322, -120.814214, -127.01296, -133.428093, -140.057144, -146.897645, -153.947129, -161.203127, -168.663172, -176.324796, -184.18553, -192.242907, -200.647593, -209.187837, -217.815229, -226.481361, -234.993153, -243.374532, -251.688005, -259.928564, -268.091203, -276.170915, -284.162693, -293.938559, -303.045207, -311.110386, -317.761844, -322.627331, -326.783158, -328.046368, -325.68257, -318.95737, -307.136377, -291.620895, -266.930537, -228.618031, -174.270843, -105.630696, -22.5741378, 75.0222854, 183.40004, 307.446835, 456.285277, 632.478042, 844.231348, 1096.30411, 1395.25702, 1750.70378, 2170.85673, 2666.81469, 3247.53833, 3911.4045, 4662.77686, 5525.95199, 6482.29404, 7529.2668, 8664.3341, 9884.95974, 11381.9249, 12911.0467, 14421.4595, 15862.2979, 17182.6963, 18111.7502, 18965.3257, 19786.8595, 20781.5869, 21928.8996, 23140.7732, 24399.9703, 25689.2533, 26991.3847, 28289.1269, 28232.0522, 29131.9953, 30642.4319, 32416.8375, 34108.6874, 35371.4573, 35858.6227, 36249.9159, 36678.4745},
        {261.977625, 260.47887, 259.072498, 257.752273, 256.511954, 255.345303, 254.246082, 253.208051, 252.224973, 251.290607, 250.398716, 249.543061, 248.717403, 247.915504, 247.131124, 246.358025, 245.589969, 244.820716, 244.044029, 243.253667, 242.443393, 241.606968, 240.738153, 239.830709, 238.878398, 237.874981, 236.81422, 235.689875, 234.495707, 233.22548, 231.872952, 230.431887, 228.896044, 227.259186, 225.21944, 222.924602, 220.526022, 218.027055, 215.431057, 212.741383, 209.961387, 207.094425, 204.143853, 201.113025, 198.005296, 194.824023, 191.572559, 188.25426, 184.872482, 181.430579, 177.931907, 174.465891, 170.947951, 167.379575, 163.762253, 160.097474, 156.386728, 152.631504, 148.833291, 144.993578, 141.113856, 137.195613, 133.240339, 129.249522, 125.224653, 121.167221, 117.078715, 112.861378, 108.419838, 103.960271, 99.4905985, 95.0187446, 90.5526321, 86.1001841, 81.6693236, 77.2679736, 72.9040571, 68.5854971, 64.3202168, 60.1161391, 55.9811871, 51.9232837, 47.950352, 44.0703151, 41.1457709, 38.739328, 36.4033225, 34.1179417, 31.8633726, 29.6198026, 27.3674187, 25.0864081, 22.7569581, 20.3592559, 17.8734885, 15.2798433, 12.5585073, 9.68966779, 6.65351196, 3.43022696, 0, -7.14763406, -14.1365626, -20.9783466, -27.6845471, -34.2667251, -40.7364418, -47.105258, -53.3847349, -59.5864335, -65.7219148, -71.8027399, -77.8404697, -83.8466654, -89.8328879, -95.8106984, -101.791658, -108.006535, -114.498207, -120.782407, -126.917668, -132.962526, -138.975514, -145.015167, -151.140018, -157.408602, -163.879453, -170.611106, -177.662094, -185.090952, -193.662531, -202.470066, -211.360195, -220.179555, -228.098346, -235.301425, -242.154038, -248.705756, -255.006151, -261.104793, -267.051256, -274.688421, -281.726759, -287.864974, -292.801773, -296.235862, -302.237072, -305.040203, -303.251179, -295.475923, -280.32036, -255.757322, -221.799606, -177.820351, -122.885893, -55.4361848, 25.8075863, 122.124234, 232.161445, 360.427143, 514.299572, 696.413295, 914.162954, 1172.08452, 1476.18027, 1834.937, 2255.70107, 2749.13203, 3323.43522, 3974.66763, 4705.75807, 5535.04831, 6447.68292, 7441.25543, 8513.35937, 9661.58828, 11048.6669, 12465.7747, 13869.2224, 15215.3208, 16460.3807, 17368.8215, 18216.7726, 19042.7837, 20028.2652, 21155.2179, 22346.1172, 23585.9402, 24859.6639, 26152.2654, 27448.7216, 28172.354, 29301.6273, 30670.5142, 32112.9874, 33463.0196, 34554.5835, 35221.6518, 35766.3705, 36384.7761},
        {254.546663, 252.446565, 250.490313, 248.669968, 246.977589, 245.405238, 243.944973, 242.588855, 241.328945, 240.157302, 239.065987, 238.047061, 237.092582, 236.194613, 235.345212, 234.53644, 233.760357, 233.009024, 232.2745, 231.548846, 230.824122, 230.092389, 229.345706, 228.576134, 227.775733, 226.936563, 226.050684, 225.110157, 224.107042, 223.033399, 221.881288, 220.64277, 219.309904, 217.874752, 215.968189, 213.770465, 211.46617, 209.059085, 206.552991, 203.95167, 201.258903, 198.478471, 195.614156, 192.669739, 189.649001, 186.555724, 183.393688, 180.166676, 176.878469, 173.532847, 170.133592, 166.798328, 163.414527, 159.9835, 156.506563, 152.985027, 149.420207, 145.813416, 142.165968, 138.479175, 134.754352, 130.992811, 127.195867, 123.364832, 119.50102, 115.605744, 111.680319, 107.624945, 103.342252, 99.0420883, 94.7323228, 90.4208231, 86.1154573, 81.8240934, 77.5545997, 73.3148442, 69.1126948, 64.9560198, 60.8526873, 56.8105652, 52.8375217, 48.9414248, 45.1301427, 41.4115434, 38.654598, 36.418511, 34.252137, 32.1353995, 30.0482224, 27.9705291, 25.8822434, 23.7632888, 21.5935891, 19.3530677, 17.0216485, 14.5792549, 12.0058107, 9.2812395, 6.38546488, 3.2984105, 0, -6.52585172, -13.0122542, -19.4605657, -25.8721443, -32.2483483, -38.5905357, -44.9000648, -51.1782938, -57.4265808, -63.6462841, -69.8387617, -76.0053719, -82.1474728, -88.2664227, -94.3635796, -100.440302, -107.017042, -114.169314, -120.724334, -126.794696, -132.492992, -137.931815, -143.223757, -148.481411, -153.817369, -159.344225, -165.174571, -171.420999, -178.196102, -186.772793, -195.677894, -204.675901, -213.531311, -221.008447, -227.371892, -233.152943, -238.416148, -243.226056, -247.647215, -251.744173, -256.97345, -261.58398, -265.367972, -268.117632, -269.625166, -276.781223, -280.504958, -278.813967, -269.725846, -251.258191, -218.124261, -175.684626, -127.572173, -74.3865451, -10.5345189, 66.7966582, 160.419739, 272.677116, 406.958801, 567.167925, 754.3847, 976.747126, 1239.71053, 1549.20699, 1911.97637, 2334.38775, 2826.60778, 3395.99012, 4035.96388, 4747.83467, 5541.5686, 6408.4784, 7346.87182, 8355.05661, 9431.34052, 10716.4665, 12030.6983, 13336.7349, 14597.2752, 15775.0181, 16666.936, 17511.9388, 18343.7515, 19319.982, 20425.3179, 21592.8287, 22809.7829, 24063.449, 25341.0954, 26629.9907, 28276.5752, 29665.9983, 30848.8896, 31875.879, 32797.5963, 33664.6713, 34527.7338, 35257.3846, 35887.4434},
        {245.454061, 243.041826, 240.796782, 238.710415, 236.774211, 234.979657, 233.318237, 231.781439, 230.360748, 229.04765, 227.833633, 226.710181, 225.66878, 224.700918, 223.79808, 222.951752, 222.15342, 221.394571, 220.66669, 219.961263, 219.269777, 218.583718, 217.894572, 217.193824, 216.472962, 215.723471, 214.936837, 214.104546, 213.218085, 212.268939, 211.248595, 210.148539, 208.960256, 207.675234, 205.919968, 203.876116, 201.729724, 199.484122, 197.14264, 194.70861, 192.185362, 189.576227, 186.884535, 184.113617, 181.266805, 178.347428, 175.358817, 172.304303, 169.187217, 166.01089, 162.778651, 159.571696, 156.313803, 153.006616, 149.651777, 146.25093, 142.805716, 139.317778, 135.78876, 132.220304, 128.614053, 124.971649, 121.294736, 117.584957, 113.843953, 110.073368, 106.274845, 102.3575, 98.2326734, 94.0928336, 89.9456221, 85.7986797, 81.6596472, 77.5361655, 73.4358753, 69.3664177, 65.3354333, 61.3505631, 57.4194479, 53.5497285, 49.7490459, 46.0250407, 42.385354, 38.8376264, 36.2486889, 34.1785153, 32.1767529, 30.2231601, 28.2974955, 26.3795174, 24.4489845, 22.485655, 20.4692877, 18.3796408, 16.196473, 13.8995426, 11.4686082, 8.88342831, 6.1237613, 3.16936571, 0, -6.66650864, -13.2548003, -19.7681911, -26.2099971, -32.5835344, -38.8921189, -45.1390668, -51.3276942, -57.4613171, -63.5432515, -69.5768135, -75.5653193, -81.5120848, -87.4204261, -93.2936593, -99.1351004, -105.309482, -111.871751, -118.007734, -123.798194, -129.323892, -134.665591, -139.904054, -145.120043, -150.394321, -155.80765, -161.440792, -167.37451, -173.689567, -181.260358, -189.083014, -197.000206, -204.854608, -212.078862, -218.720656, -224.937018, -230.693629, -235.95617, -240.690323, -244.861769, -249.703562, -253.528287, -256.053664, -256.997409, -256.07724, -257.956702, -256.171227, -249.202075, -235.530506, -213.63778, -179.781522, -137.384403, -90.5459818, -39.689483, 22.2677124, 99.0382526, 194.334786, 313.439433, 458.13867, 628.506825, 823.254222, 1051.43172, 1319.18627, 1632.24343, 1995.61483, 2414.63979, 2897.87941, 3451.5083, 4069.88795, 4754.06203, 5513.29849, 6339.15143, 7229.81765, 8183.49395, 9198.37711, 10392.5585, 11614.3666, 12832.0248, 14013.7562, 15127.7839, 15997.8569, 16832.9887, 17663.32, 18633.3588, 19727.9755, 20884.3871, 22089.8468, 23331.6072, 24596.9215, 25873.0425, 27707.3864, 29127.717, 30240.5316, 31152.3271, 31969.6007, 32798.8496, 33746.5709, 34599.7427, 35127.9387},
        {236.525754, 233.89683, 231.448249, 229.171328, 227.057389, 225.09775, 223.283732, 221.606654, 220.057835, 218.628597, 217.310257, 216.094137, 214.971556, 213.933833, 212.972289, 212.078242, 211.243014, 210.457923, 209.714289, 209.003432, 208.316673, 207.645329, 206.980722, 206.314171, 205.636995, 204.940515, 204.21605, 203.45492, 202.648445, 201.787944, 200.864737, 199.870144, 198.795484, 197.632078, 196.024281, 194.144191, 192.168081, 190.098528, 187.938113, 185.689415, 183.355013, 180.937487, 178.439416, 175.86338, 173.211957, 170.487728, 167.693272, 164.831168, 161.903996, 158.914335, 155.864764, 152.778236, 149.636517, 146.441743, 143.196052, 139.901581, 136.56047, 133.174854, 129.746873, 126.278663, 122.772363, 119.230111, 115.654043, 112.046298, 108.409014, 104.744328, 101.054378, 97.2607587, 93.2869993, 89.301351, 85.3111725, 81.323823, 77.3466614, 73.3870468, 69.4523381, 65.5498944, 61.6870747, 57.8712379, 54.1097431, 50.4099492, 46.7792153, 43.2249004, 39.7543635, 36.3749636, 33.9485256, 32.037203, 30.1928793, 28.3951845, 26.6237484, 24.8582011, 23.0781724, 21.2632924, 19.3931909, 17.4474979, 15.4058433, 13.2478572, 10.9531693, 8.50140974, 5.87220836, 3.04519513, 0, -7.02867421, -13.8940542, -20.6050842, -27.1707084, -33.5998711, -39.9015165, -46.084589, -52.1580326, -58.1307917, -64.0118105, -69.8100332, -75.5344041, -81.1938675, -86.7973675, -92.3538484, -97.8722545, -103.422422, -109.021993, -114.542182, -120.00498, -125.43238, -130.846375, -136.268957, -141.722118, -147.227851, -152.808148, -158.485003, -164.280407, -170.216353, -176.488266, -182.881114, -189.364859, -195.909465, -202.954612, -210.235404, -217.361687, -224.162506, -230.46691, -236.103946, -240.902661, -245.862565, -249.286014, -250.773053, -249.923724, -246.338071, -240.813021, -231.452514, -217.557373, -198.42842, -173.366479, -141.904977, -102.827838, -59.0368906, -10.095229, 51.2506305, 129.193667, 227.92686, 354.868691, 510.253567, 690.797166, 892.82557, 1126.33052, 1398.24998, 1714.10985, 2077.0434, 2491.28216, 2963.28984, 3497.87667, 4091.66829, 4745.92, 5471.59174, 6258.74118, 7105.26432, 8009.05717, 8968.01576, 10079.4256, 11216.9458, 12353.625, 13462.5119, 14516.6552, 15362.6958, 16184.3618, 17008.8766, 17972.1971, 19058.5063, 20206.6816, 21403.1712, 22634.4234, 23886.8865, 25147.0088, 26876.8575, 28239.2716, 29330.8802, 30248.3123, 31088.1968, 31947.163, 32921.8397, 33869.0116, 34232.008},
        {228.252986, 225.456539, 222.848117, 220.419076, 218.160773, 216.064565, 214.121808, 212.32386, 210.662077, 209.127816, 207.712434, 206.407287, 205.203733, 204.093128, 203.066828, 202.116191, 201.232574, 200.407333, 199.631825, 198.897406, 198.195435, 197.517266, 196.854258, 196.197766, 195.539149, 194.869762, 194.180962, 193.464106, 192.710551, 191.911653, 191.05877, 190.143259, 189.156475, 188.089776, 186.614035, 184.887591, 183.072636, 181.170927, 179.184222, 177.114277, 174.962849, 172.731695, 170.422573, 168.037239, 165.577452, 163.044967, 160.441542, 157.768933, 155.028899, 152.223196, 149.353582, 146.381765, 143.35042, 140.26217, 137.119641, 133.925459, 130.682247, 127.392633, 124.059241, 120.684696, 117.271623, 113.822648, 110.340395, 106.827491, 103.286561, 99.7202285, 96.1311201, 92.4528676, 88.6207595, 84.7797143, 80.9368298, 77.0992038, 73.273934, 69.4681183, 65.6888543, 61.9432399, 58.2383729, 54.581351, 50.9792721, 47.4392338, 43.9683341, 40.5736706, 37.2623411, 34.0414434, 31.7680832, 30.0064755, 28.310619, 26.6600271, 25.0342131, 23.4126905, 21.7749726, 20.1005728, 18.3690045, 16.5597812, 14.6524162, 12.6264228, 10.4613145, 8.13660471, 5.63180676, 2.92643406, 0, -7.39675546, -14.5428365, -21.4531796, -28.1427211, -34.6263973, -40.9191446, -47.0358995, -52.9915983, -58.8011773, -64.479573, -70.0417218, -75.5025599, -80.8770239, -86.18005, -91.4265747, -96.6315343, -101.553965, -106.172245, -111.080134, -116.237733, -121.605142, -127.142461, -132.809792, -138.567234, -144.374889, -150.192856, -155.981237, -161.700131, -167.309641, -172.308442, -177.287247, -182.344583, -187.578978, -194.442628, -202.357226, -210.383486, -218.213837, -225.540704, -232.056516, -237.453699, -242.491746, -245.471258, -245.87589, -243.189293, -236.895122, -223.582159, -206.352645, -185.413951, -160.973448, -133.238507, -105.640931, -71.2226857, -30.5315343, 18.0981055, 80.8507311, 161.881939, 265.347327, 399.441703, 563.363449, 751.904536, 959.796061, 1197.54025, 1472.86027, 1791.03671, 2153.21137, 2562.42583, 3022.74748, 3537.48384, 4106.18114, 4730.51321, 5424.15461, 6174.93219, 6980.36569, 7837.97484, 8745.2794, 9780.76125, 10840.7374, 11902.4872, 12943.2897, 13940.4241, 14760.1825, 15564.8227, 16379.1898, 17333.3736, 18410.56, 19550.0826, 20737.32, 21957.6504, 23196.4522, 24439.1039, 25878.7503, 27132.812, 28242.9753, 29250.9264, 30198.3517, 31126.9375, 32078.3701, 33090.1144, 33278.4476},
        {221.842616, 218.846475, 216.048932, 213.441202, 211.014498, 208.760033, 206.66902, 204.732675, 202.94221, 201.288838, 199.763774, 198.358231, 197.063423, 195.870563, 194.770865, 193.755542, 192.815808, 191.942877, 191.127962, 190.362278, 189.637036, 188.943452, 188.272738, 187.616109, 186.964778, 186.309958, 185.642863, 184.954708, 184.236704, 183.480066, 182.676008, 181.815743, 180.890485, 179.891447, 178.495121, 176.856701, 175.135164, 173.331937, 171.448448, 169.486123, 167.446389, 165.330674, 163.140404, 160.877006, 158.541908, 156.136535, 153.662315, 151.120676, 148.513043, 145.840845, 143.105507, 140.254146, 137.343677, 134.376704, 131.355833, 128.283666, 125.162809, 121.995866, 118.785442, 115.534139, 112.244564, 108.919319, 105.56101, 102.172241, 98.7556165, 95.3137401, 91.8492165, 88.2952619, 84.5867587, 80.8694188, 77.1503446, 73.4366388, 69.7354037, 66.0537418, 62.3987557, 58.7775477, 55.1972204, 51.6648763, 48.1876177, 44.7725472, 41.4267673, 38.1573804, 34.9714891, 31.8761957, 29.7339862, 28.1062794, 26.5442134, 25.0271317, 23.5343779, 22.0452958, 20.5392289, 18.9955209, 17.3935155, 15.7125562, 13.9319867, 12.0311507, 9.98939184, 7.78605367, 5.40047988, 2.81201411, 0, -7.10174147, -13.9972501, -20.697905, -27.2150852, -33.5601697, -39.7445374, -45.7795675, -51.676639, -57.4471309, -63.1024221, -68.6538919, -74.1129191, -79.4908829, -84.7991622, -90.049136, -95.2521835, -100.211942, -104.910113, -109.827967, -114.932368, -120.190179, -125.568261, -131.033479, -136.552696, -142.092773, -147.620575, -153.102964, -158.506803, -163.798955, -168.556289, -173.27866, -178.049928, -182.953956, -189.474749, -196.996096, -204.528415, -211.735522, -218.281236, -223.829374, -228.043755, -231.454654, -232.595727, -230.961266, -226.045567, -217.342922, -199.915123, -178.797091, -154.591244, -127.900001, -99.3257809, -74.749549, -43.043619, -0.906554162, 54.6210165, 125.80008, 215.205185, 325.410878, 460.39715, 621.013686, 806.576961, 1014.759, 1253.82109, 1531.74308, 1853.13624, 2216.90393, 2624.56954, 3077.41479, 3576.90042, 4125.37322, 4724.67879, 5390.01735, 6107.82529, 6875.32544, 7689.74062, 8548.29365, 9513.51078, 10500.9856, 11491.615, 12466.296, 13405.9256, 14192.8551, 14972.2242, 15767.1638, 16701.8507, 17759.4182, 18880.8976, 20051.5274, 21256.5464, 22481.1931, 23710.7061, 24745.5897, 25889.588, 27086.5358, 28280.2682, 29414.6199, 30433.426, 31280.5213, 32306.9729, 32472.326},
        {216.397646, 213.211022, 210.232602, 207.453465, 204.864686, 202.457343, 200.222514, 198.151273, 196.2347, 194.463869, 192.829859, 191.323746, 189.936608, 188.65952, 187.48356, 186.399805, 185.399331, 184.473216, 183.612536, 182.808369, 182.05179, 181.333878, 180.645709, 179.978359, 179.322906, 178.670427, 178.011998, 177.338696, 176.641599, 175.911783, 175.140324, 174.318301, 173.43679, 172.486867, 171.146506, 169.569921, 167.915091, 166.183252, 164.375643, 162.493501, 160.538063, 158.510568, 156.412253, 154.244356, 152.008115, 149.704766, 147.335549, 144.9017, 142.404457, 139.845059, 137.224742, 134.48968, 131.69737, 128.850241, 125.950726, 123.001255, 120.004261, 116.962175, 113.877427, 110.75245, 107.589675, 104.391532, 101.160455, 97.8988726, 94.6092178, 91.2939216, 87.9554154, 84.5228139, 80.9269924, 77.3215921, 73.7137971, 70.1107916, 66.5197599, 62.947886, 59.4023543, 55.8903489, 52.4190539, 48.9956536, 45.6273321, 42.3212737, 39.0846624, 35.9246826, 32.8485184, 29.8633539, 27.8397634, 26.3350758, 24.8962427, 23.5024297, 22.1328023, 20.766526, 19.3827664, 17.9606889, 16.4794592, 14.9182426, 13.2562047, 11.4725111, 9.54632727, 7.45681871, 5.18315097, 2.70448956, 0, -6.6829892, -13.2292378, -19.6448188, -25.9358054, -32.1082704, -38.168287, -44.1219283, -49.9752672, -55.7343768, -61.4053302, -66.9942003, -72.5070603, -77.9499832, -83.3290419, -88.6503097, -93.9198595, -99.0504315, -104.034766, -109.090045, -114.202342, -119.357731, -124.542283, -129.742074, -134.943175, -140.13166, -145.293603, -150.415075, -155.482152, -160.480904, -165.225524, -169.93699, -174.652941, -179.411015, -185.484366, -192.292875, -198.926951, -205.053577, -210.339737, -214.452414, -217.058592, -218.495623, -217.556097, -213.775836, -206.690667, -195.836412, -175.64627, -152.034347, -125.812124, -97.7910823, -68.7827014, -46.4672231, -16.3922152, 28.7037862, 92.9766709, 174.241783, 273.161241, 390.397164, 524.635673, 678.963977, 858.624926, 1065.17722, 1304.1366, 1583.6256, 1907.54946, 2272.6674, 2679.01877, 3125.22367, 3610.95349, 4141.08361, 4717.54576, 5356.70861, 6043.84583, 6775.89174, 7549.78063, 8362.44682, 9261.87098, 10181.1794, 11104.5449, 12016.1401, 12900.1377, 13653.1868, 14405.3328, 15178.2587, 16090.6363, 17125.7569, 18226.5001, 19378.1777, 20566.1019, 21775.5848, 22991.9382, 23586.3167, 24617.052, 25917.6729, 27321.7081, 28662.6863, 29774.1362, 30489.5867, 31504.0105, 31692.6514},
        {211.768591, 208.398473, 205.245681, 202.301146, 199.5558, 197.000572, 194.626393, 192.424195, 190.384907, 188.499461, 186.758786, 185.153814, 183.675476, 182.314701, 181.062421, 179.909567, 178.847068, 177.865856, 176.956862, 176.111015, 175.319247, 174.572489, 173.86167, 173.177722, 172.511575, 171.85416, 171.196408, 170.529249, 169.843614, 169.130433, 168.380638, 167.585159, 166.734926, 165.820871, 164.518619, 162.983386, 161.374175, 159.69215, 157.938475, 156.11431, 154.220821, 152.259169, 150.230517, 148.136029, 145.976866, 143.754193, 141.469172, 139.122965, 136.716737, 134.251648, 131.728864, 129.103637, 126.424035, 123.692215, 120.910336, 118.080556, 115.205033, 112.285925, 109.32539, 106.325587, 103.288673, 100.216807, 97.1121463, 93.9768496, 90.8130748, 87.62298, 84.4087234, 81.093017, 77.6004802, 74.0971228, 70.5902531, 67.0871793, 63.5952097, 60.1216527, 56.6738164, 53.2590092, 49.8845394, 46.5577152, 43.2858449, 40.0762368, 36.9361992, 33.8730404, 30.8940685, 28.0065921, 26.0907028, 24.6989595, 23.3734305, 22.0931006, 20.8369546, 19.5839772, 18.3131532, 17.0034673, 15.6339044, 14.1834492, 12.6310866, 10.9558011, 9.13657775, 7.15240117, 4.98225618, 2.60512753, 0, -6.2505159, -12.4376544, -18.5619409, -24.6239007, -30.624059, -36.5629411, -42.4410723, -48.2589779, -54.0171833, -59.7162136, -65.3565942, -70.9388505, -76.4635076, -81.9310908, -87.3421256, -92.6971371, -98.0268551, -103.33612, -108.557137, -113.696904, -118.762419, -123.76068, -128.698683, -133.583428, -138.42191, -143.221129, -147.988082, -152.729766, -157.45318, -162.253612, -167.017394, -171.725039, -176.357055, -181.8585, -187.727609, -193.20541, -197.983048, -201.751671, -204.202425, -205.026457, -204.427067, -201.427375, -195.618325, -186.590859, -173.935919, -152.281144, -127.421607, -100.189077, -71.4153223, -41.9321118, -20.3482781, 9.78677427, 59.0332849, 132.433894, 223.766145, 331.865378, 455.566936, 588.415959, 735.740309, 908.638973, 1112.51973, 1350.27196, 1630.12844, 1955.36694, 2320.86928, 2725.37116, 3165.25386, 3638.64268, 4152.29575, 4708.088, 5322.94972, 5981.38423, 6680.05498, 7415.6254, 8184.75893, 9021.67434, 9876.09092, 10735.2833, 11586.526, 12417.0937, 13137.2268, 13862.5903, 14613.4733, 15502.7737, 16514.1232, 17592.5664, 18723.6645, 19892.9784, 21086.069, 22288.4975, 22482.7212, 23413.3467, 24811.5448, 26408.4863, 27935.342, 29123.2829, 29703.4797, 30686.7909, 30934.283},
        {207.691201, 204.130615, 200.798283, 197.684887, 194.781114, 192.077646, 189.565168, 187.234364, 185.075919, 183.080516, 181.238841, 179.541577, 177.979408, 176.543019, 175.223094, 174.010318, 172.895374, 171.868946, 170.92172, 170.044379, 169.227607, 168.46209, 167.73851, 167.047552, 166.379902, 165.726241, 165.077256, 164.42363, 163.756048, 163.065193, 162.341751, 161.576405, 160.759839, 159.882738, 158.611752, 157.106399, 155.530089, 153.884022, 152.169398, 150.387417, 148.539278, 146.626181, 144.649326, 142.609913, 140.509142, 138.348211, 136.128322, 133.850673, 131.516465, 129.126897, 126.683169, 124.155446, 121.576636, 118.948611, 116.273242, 113.552404, 110.787968, 107.981807, 105.135794, 102.251801, 99.3317001, 96.3773644, 93.3906664, 90.3734785, 87.3276734, 84.2551234, 81.1577012, 77.9520342, 74.5567964, 71.149672, 67.7380598, 64.3293583, 60.9309661, 57.5502819, 54.1947043, 50.871632, 47.5884634, 44.3525974, 41.1714324, 38.0523672, 35.0028003, 32.0301304, 29.1417561, 26.3450761, 24.5286677, 23.2408057, 22.0191739, 20.842575, 19.6898117, 18.5396866, 17.3710024, 16.1625618, 14.8931675, 13.5416221, 12.0867283, 10.5072889, 8.78210641, 6.88998361, 4.80972316, 2.52012773, 0, -6.2754785, -12.4732133, -18.5946426, -24.6412046, -30.6143375, -36.5154795, -42.3460688, -48.1075436, -53.8013421, -59.4289024, -64.9916629, -70.4910617, -75.928537, -81.305527, -86.6234699, -91.8838039, -96.9862753, -101.917795, -106.909818, -111.941993, -116.993965, -122.045382, -127.075891, -132.065139, -136.992774, -141.838442, -146.58179, -151.202466, -155.680117, -159.938402, -164.033484, -167.961807, -171.719815, -176.294177, -181.186224, -185.633225, -189.334555, -191.989589, -193.297701, -192.958269, -191.023576, -186.732681, -179.71591, -169.603591, -156.026053, -135.321652, -111.235679, -84.2214521, -54.7322924, -23.2215193, 3.2805837, 38.9361713, 93.6203636, 171.924905, 267.654648, 379.45754, 505.981526, 640.419905, 788.114966, 960.359525, 1162.36203, 1397.41387, 1674.72999, 1997.7287, 2360.00627, 2759.66743, 3192.37589, 3655.60352, 4155.77272, 4694.24294, 5286.70867, 5917.98899, 6584.48127, 7282.58285, 8008.69109, 8781.23611, 9569.07432, 10363.0949, 11154.187, 11933.2398, 12627.2883, 13333.6453, 14070.5661, 14942.743, 15934.2165, 16993.0119, 18104.9703, 19255.9332, 20431.7418, 21618.2375, 21717.4294, 22621.3687, 24035.4351, 25665.0085, 27215.4687, 28392.1955, 28900.5686, 29862.3561, 30203.7779},
        {204.04205, 200.28447, 196.767946, 193.482817, 190.419421, 187.568097, 184.919184, 182.46302, 180.189944, 178.090295, 176.154411, 174.37263, 172.735293, 171.232737, 169.8553, 168.593322, 167.437142, 166.377097, 165.403527, 164.50677, 163.677165, 162.905051, 162.180766, 161.494649, 160.837038, 160.198272, 159.56869, 158.938631, 158.298433, 157.638435, 156.948975, 156.220392, 155.443026, 154.607213, 153.365013, 151.883019, 150.331891, 148.712947, 147.027504, 145.276876, 143.462382, 141.585338, 139.647059, 137.648864, 135.592067, 133.477987, 131.307939, 129.083239, 126.805205, 124.475153, 122.0944, 119.650349, 117.158531, 114.620564, 112.038067, 109.412657, 106.745954, 104.039574, 101.295136, 98.5142586, 95.6985595, 92.849657, 89.9691691, 87.0587141, 84.11991, 81.1543751, 78.1637274, 75.0599317, 71.7571052, 68.4417695, 65.1213547, 61.8032909, 58.4950079, 55.2039359, 51.9375049, 48.7031447, 45.5082856, 42.3603575, 39.2667903, 36.2350142, 33.2724592, 30.3865551, 27.5847322, 24.8744203, 23.1506402, 21.9583571, 20.8318685, 19.7498003, 18.6907786, 17.6334292, 16.5563783, 15.4382517, 14.2576756, 12.9932758, 11.6236785, 10.1275095, 8.48339485, 6.6699606, 4.66583271, 2.44963718, 0, -6.78588548, -13.3867432, -19.8117119, -26.0699305, -32.1705378, -38.1226725, -43.9354734, -49.6180795, -55.1796294, -60.6292621, -65.9761162, -71.2293306, -76.3980442, -81.4913956, -86.5185238, -91.4885675, -95.9116282, -99.7255535, -104.068257, -108.841942, -113.94881, -119.291063, -124.770903, -130.290534, -135.752156, -141.057972, -146.110184, -150.810995, -155.062607, -158.135523, -160.795268, -163.133552, -165.242087, -168.562188, -172.510766, -176.144836, -179.151231, -181.216781, -182.028317, -181.272671, -178.879071, -174.218178, -166.929397, -156.652133, -143.025793, -125.365833, -103.716595, -77.7984691, -47.3318483, -12.0371246, 25.040432, 71.5690465, 132.561937, 210.794967, 304.476063, 413.864063, 539.217803, 678.626763, 835.012174, 1013.66184, 1215.2406, 1446.57509, 1718.62901, 2035.7141, 2390.87071, 2782.31292, 3206.62534, 3661.59958, 4151.00215, 4675.2198, 5246.87788, 5852.24173, 6487.46172, 7148.68821, 7832.07153, 8538.09585, 9257.49426, 9985.33367, 10716.681, 11446.6031, 12122.5398, 12818.9365, 13551.2718, 14413.2611, 15389.4141, 16431.6419, 17526.0842, 18658.8807, 19816.1711, 20984.0951, 21326.6246, 22283.6809, 23620.8737, 25103.8128, 26498.1079, 27569.3688, 28083.2052, 29035.934, 29493.3389},
        {200.87636, 196.933985, 193.244602, 189.798224, 186.584862, 183.594528, 180.817231, 178.242985, 175.861799, 173.663686, 171.638655, 169.77672, 168.067891, 166.502178, 165.069594, 163.76015, 162.563857, 161.470726, 160.470768, 159.553995, 158.710418, 157.930048, 157.202896, 156.518975, 155.868294, 155.240865, 154.6267, 154.015809, 153.398205, 152.763897, 152.102898, 151.405219, 150.660871, 149.859864, 148.639889, 147.174527, 145.641596, 144.042543, 142.378811, 140.651846, 138.863091, 137.013993, 135.105995, 133.140543, 131.119081, 129.043055, 126.913908, 124.733086, 122.502034, 120.222197, 117.895019, 115.524838, 113.110144, 110.652318, 108.152742, 105.612798, 103.03387, 100.417337, 97.7645841, 95.0769915, 92.3559418, 89.6028171, 86.8189995, 84.005871, 81.1648138, 78.29721, 75.4044416, 72.3944133, 69.1772738, 65.9471958, 62.711621, 59.4779914, 56.2537487, 53.0463348, 49.8631916, 46.711761, 43.5994847, 40.5338047, 37.5221628, 34.5720009, 31.6907608, 28.8858843, 26.1648134, 23.5349899, 21.8966673, 20.7921025, 19.7528215, 18.7572928, 17.7839848, 16.8113659, 15.8179044, 14.7820689, 13.6823276, 12.497149, 11.2050015, 9.7843535, 8.21367339, 6.47142959, 4.53609049, 2.38612449, 0, -7.24757117, -14.2143469, -20.9161936, -27.3689777, -33.5885655, -39.5908236, -45.3916182, -51.0068159, -56.452283, -61.7438859, -66.897491, -71.9289648, -76.8541736, -81.6889839, -86.449262, -91.1508744, -94.9607547, -97.7734931, -101.525161, -106.049711, -111.181096, -116.753267, -122.600176, -128.555776, -134.45402, -140.128859, -145.414245, -150.144132, -154.15247, -156.139612, -157.488764, -158.373959, -158.969228, -161.064074, -164.024794, -166.786895, -169.039769, -170.472808, -170.775405, -169.636951, -167.089686, -162.37581, -155.117636, -144.937478, -131.457648, -116.790726, -97.4461934, -72.4237968, -40.7232836, -1.34440076, 46.7546709, 104.481815, 171.589057, 248.913327, 339.506564, 445.426215, 568.729725, 712.81033, 878.086097, 1063.51787, 1265.29945, 1493.47337, 1760.13237, 2070.32559, 2417.16725, 2799.24989, 3214.59011, 3661.63115, 4140.92802, 4651.84121, 5203.52248, 5784.03465, 6389.34893, 7015.4365, 7658.26855, 8301.49501, 8956.48863, 9622.30092, 10297.9834, 10982.5875, 11643.5326, 12332.5905, 13062.3692, 13914.8083, 14875.0721, 15899.4371, 16974.512, 18086.9056, 19223.2269, 20370.0845, 21033.8552, 22061.8457, 23300.3151, 24595.5226, 25793.7274, 26741.1886, 27284.1655, 28210.4506, 28776.9591},
        {198.137604, 194.029921, 190.185322, 186.593553, 183.24436, 180.12749, 177.23269, 174.549707, 172.068286, 169.778175, 167.66912, 165.730867, 163.953164, 162.325757, 160.838393, 159.480817, 158.242777, 157.11402, 156.084292, 155.143339, 154.280908, 153.486746, 152.7506, 152.062215, 151.41134, 150.787719, 150.1811, 149.581229, 148.977854, 148.36072, 147.719574, 147.044163, 146.324233, 145.549532, 144.346422, 142.893984, 141.375668, 139.793014, 138.147559, 136.440844, 134.674407, 132.849788, 130.968525, 129.032158, 127.042226, 125.000268, 122.907822, 120.766429, 118.577627, 116.342955, 114.063953, 111.758412, 109.411266, 107.023702, 104.596907, 102.132068, 99.6303708, 97.0930027, 94.5211503, 91.9160006, 89.2787401, 86.6105558, 83.9126342, 81.1861623, 78.4323268, 75.6523144, 72.8473118, 69.9221498, 66.7839724, 63.6326937, 60.4757468, 57.3205649, 54.174581, 51.0452284, 47.9399401, 44.8661494, 41.8312893, 38.842793, 35.9080936, 33.0346243, 30.2298181, 27.5011083, 24.855928, 22.3017103, 20.7424111, 19.7183449, 18.7590509, 17.8428685, 16.9481367, 16.0531949, 15.1363824, 14.1760383, 13.150502, 12.0381126, 10.8172095, 9.4661319, 7.96321902, 6.28681014, 4.41524451, 2.32686138, 0, -7.47802968, -14.6268742, -21.465453, -28.0126856, -34.2874914, -40.30879, -46.0955007, -51.666543, -57.0408365, -62.2373005, -67.2748546, -72.1724181, -76.9489107, -81.6232517, -86.2143606, -90.7411568, -94.1820866, -96.4074304, -99.7895592, -104.124434, -109.208015, -114.836264, -120.805142, -126.910609, -132.948626, -138.715154, -144.006155, -148.617589, -152.345417, -153.596224, -154.064784, -153.963872, -153.506261, -154.555747, -156.499593, -158.284789, -159.6288, -160.249094, -159.863137, -158.188397, -155.688473, -151.107615, -144.017307, -133.989033, -120.594277, -107.864825, -89.7967838, -64.8465624, -31.4705692, 11.8747872, 69.4946021, 136.795791, 209.163776, 285.821461, 373.826607, 476.719286, 598.039569, 745.366499, 917.283225, 1107.96675, 1310.68131, 1537.13871, 1799.35592, 2102.61373, 2440.7788, 2812.95705, 3218.75942, 3657.42276, 4126.33227, 4623.92056, 5155.97403, 5712.44014, 6289.20595, 6882.15853, 7487.18494, 8073.35512, 8670.07753, 9279.94352, 9905.54443, 10549.4716, 11197.4425, 11880.1717, 12607.4819, 13449.2641, 14391.5643, 15395.4005, 16448.0703, 17536.8712, 18649.1008, 19772.0567, 20758.2338, 21855.7547, 22993.2959, 24099.534, 25103.1453, 25932.8065, 26517.194, 27390.2716, 28040.0225},
        {195.736755, 191.501337, 187.535024, 183.827457, 180.368276, 177.147124, 174.15364, 171.377467, 168.808245, 166.435614, 164.249217, 162.238694, 160.393686, 158.703835, 157.158781, 155.748165, 154.461629, 153.288813, 152.219359, 151.242907, 150.349099, 149.527576, 148.767978, 148.059947, 147.393124, 146.75715, 146.141666, 145.536313, 144.930731, 144.314563, 143.677449, 143.00903, 142.298948, 141.536842, 140.346551, 138.908311, 137.406756, 135.843399, 134.219752, 132.537326, 130.797635, 129.002189, 127.152502, 125.250084, 123.296449, 121.293109, 119.241575, 117.14336, 114.999975, 112.812933, 110.583746, 108.334017, 106.04473, 103.716964, 101.351795, 98.9502994, 96.5135537, 94.0426345, 91.5386186, 89.0025825, 86.4356028, 83.838756, 81.2131189, 78.5597679, 75.8797797, 73.1742308, 70.4441979, 67.5932697, 64.5273757, 61.4486538, 58.3645002, 55.2823113, 52.2094835, 49.1534131, 46.1214964, 43.1211297, 40.1597094, 37.2446319, 34.3832933, 31.5830902, 28.8514187, 26.1956753, 23.6232563, 21.141558, 19.655272, 18.7042736, 17.8176533, 16.9736885, 16.1506566, 15.3268349, 14.4805007, 13.5899316, 12.6334048, 11.5891977, 10.4355878, 9.15085228, 7.71326863, 6.10111421, 4.29266638, 2.26620252, 0, -7.07197789, -13.8928293, -20.4749056, -26.8305583, -32.9721388, -38.9119984, -44.6624886, -50.2359608, -55.6447664, -60.9012569, -66.0177835, -71.0066979, -75.8803513, -80.6510952, -85.331281, -89.9332601, -93.6308219, -96.3165234, -99.8994632, -104.212301, -109.087696, -114.358308, -119.856797, -125.415822, -130.868042, -136.046118, -140.782709, -144.910474, -148.262073, -149.634938, -150.27654, -150.330106, -149.938864, -150.38559, -151.243738, -151.782881, -151.804383, -151.109606, -149.499913, -146.776667, -144.513093, -140.19943, -133.290373, -123.240615, -109.504851, -95.7207705, -76.1143242, -49.0944578, -13.0701171, 33.5497522, 95.8150224, 167.630486, 243.182992, 321.191238, 409.626433, 512.305599, 633.045755, 779.707572, 951.145409, 1141.80237, 1346.4503, 1574.42481, 1835.11648, 2132.80074, 2463.08564, 2825.5577, 3221.10877, 3649.66379, 4106.36147, 4589.04788, 5101.01435, 5633.82395, 6183.43785, 6745.81721, 7316.92321, 7855.56231, 8404.13906, 8967.90329, 9552.10484, 10161.9936, 10797.9042, 11473.2785, 12195.4725, 13022.3021, 13941.1501, 14918.3743, 15942.4087, 17001.6872, 18084.6438, 19179.7125, 20385.3235, 21525.9682, 22589.0285, 23561.8865, 24431.9242, 25186.5234, 25813.0663, 26586.4858, 27259.4193},
        {193.605523, 189.291138, 185.24638, 181.461008, 177.924779, 174.627451, 171.558782, 168.70853, 166.066453, 163.622309, 161.365855, 159.28685, 157.375051, 155.620216, 154.012104, 152.540472, 151.195077, 149.965679, 148.842034, 147.813901, 146.871038, 146.003202, 145.200151, 144.451644, 143.747437, 143.07729, 142.43096, 141.798204, 141.168782, 140.532449, 139.878966, 139.198089, 138.479576, 137.713185, 136.532907, 135.113867, 133.635465, 132.099004, 130.505788, 128.85712, 127.154302, 125.398639, 123.591433, 121.733987, 119.827606, 117.873591, 115.873247, 113.827876, 111.738782, 109.607268, 107.434637, 105.232399, 102.991429, 100.712809, 98.3976213, 96.046948, 93.6618708, 91.2434717, 88.7928327, 86.3110357, 83.7991627, 81.2582957, 78.6895165, 76.0939073, 73.4725499, 70.8265263, 68.1569184, 65.3684563, 62.3681833, 59.3559, 56.3389347, 53.3246152, 50.3202697, 47.3332263, 44.3708131, 41.440358, 38.5491893, 35.7046348, 32.9140228, 30.1846812, 27.5239382, 24.9391219, 22.4375602, 20.0265812, 18.6079218, 17.7228929, 16.9020555, 16.1237125, 15.3661664, 14.6077202, 13.8266767, 13.0013386, 12.1100087, 11.1309899, 10.0425848, 8.82309627, 7.45082714, 5.90408015, 4.16115811, 2.2003638, 0, -5.76627557, -11.5374328, -17.3057578, -23.0635368, -28.803056, -34.5166014, -40.1964594, -45.834916, -51.4242575, -56.95677, -62.4247397, -67.8204527, -73.1361952, -78.3642535, -83.4969136, -88.5264617, -93.3578413, -97.970861, -102.555368, -107.084932, -111.533123, -115.873511, -120.079666, -124.125157, -127.983554, -131.628428, -135.033349, -138.171885, -141.017607, -143.714554, -146.003321, -147.806338, -149.046033, -149.52396, -149.222985, -148.15821, -146.288326, -143.572025, -139.967999, -135.434941, -133.525876, -129.511235, -122.646469, -112.187032, -97.3883772, -78.215935, -53.037685, -20.9315852, 19.0244064, 67.7523318, 127.324185, 196.106559, 271.9042, 354.61133, 448.388185, 555.479649, 678.130606, 818.597333, 979.11073, 1161.88926, 1369.63487, 1603.57397, 1867.05972, 2161.63737, 2485.78877, 2839.40191, 3223.94144, 3639.70412, 4081.20565, 4546.2318, 5037.05209, 5546.27879, 6070.14561, 6604.88628, 7146.73454, 7650.0253, 8163.3658, 8693.46446, 9247.0297, 9830.76995, 10454.4313, 11119.6594, 11831.8608, 12637.1752, 13524.8858, 14467.3033, 15454.5595, 16476.786, 17524.1145, 18586.6766, 19822.0936, 20958.1417, 21996.0737, 22937.142, 23782.5993, 24533.6981, 25191.691, 25805.8923, 26417.4592},
        {191.875385, 187.475663, 183.346779, 179.478573, 175.860889, 172.48357, 169.336456, 166.409392, 163.692218, 161.174777, 158.846913, 156.698466, 154.71928, 152.899197, 151.228058, 149.695708, 148.291987, 147.006738, 145.829804, 144.751026, 143.760248, 142.847312, 142.002059, 141.214333, 140.473976, 139.770829, 139.094736, 138.435538, 137.783079, 137.1272, 136.457744, 135.764553, 135.037469, 134.266335, 133.093818, 131.69065, 130.231736, 128.718186, 127.151107, 125.53161, 123.860802, 122.139794, 120.369693, 118.551608, 116.686649, 114.775923, 112.820541, 110.821611, 108.780242, 106.697542, 104.574621, 102.413096, 100.213555, 97.9770979, 95.7048215, 93.397824, 91.0572031, 88.6840567, 86.2794827, 83.8445789, 81.3804432, 78.8881733, 76.3688673, 73.8236228, 71.2535379, 68.6597102, 66.0432378, 63.3100836, 60.368495, 57.4157781, 54.4591979, 51.5060195, 48.5635078, 45.6389281, 42.7395452, 39.8726242, 37.0454303, 34.2652284, 31.5392836, 28.8748609, 26.2792254, 23.7596422, 21.3233763, 18.9776926, 17.623818, 16.8032333, 16.0466749, 15.3323969, 14.6386536, 13.9436993, 13.2257884, 12.463175, 11.6341134, 10.7168579, 9.68966287, 8.53078246, 7.218471, 5.73098274, 4.04657198, 2.14349297, 0, -4.66725051, -9.55922514, -14.6507196, -19.9165297, -25.331451, -30.8702794, -36.5078105, -42.2188401, -47.9781639, -53.7605777, -59.5408771, -65.2938579, -70.9943158, -76.6170466, -82.1368459, -87.5285096, -93.3724327, -99.7299253, -105.205975, -109.905148, -113.932012, -117.391134, -120.387082, -123.024421, -125.407719, -127.641544, -129.830461, -132.079039, -134.491844, -138.465717, -142.339117, -145.828931, -148.652042, -149.299154, -148.100242, -145.712266, -142.219967, -137.708083, -132.261356, -125.964525, -123.890276, -119.617333, -112.254535, -100.910719, -84.6947236, -60.1900644, -29.6622323, 7.14860435, 50.5022768, 100.658617, 156.510922, 221.355765, 297.059049, 384.868855, 484.771878, 597.321148, 723.069697, 858.551436, 1009.25195, 1185.04133, 1396.18829, 1636.20637, 1902.81678, 2194.96638, 2513.67888, 2859.02466, 3232.71842, 3635.25687, 4061.10759, 4508.14855, 4978.16998, 5464.53576, 5963.76966, 6472.39544, 6986.93686, 7458.87021, 7941.0286, 8441.19763, 8967.16295, 9526.71017, 10136.8319, 10789.9688, 11489.9607, 12273.8117, 13132.234, 14042.1213, 14994.8767, 15981.9032, 16994.604, 18024.3821, 19181.7507, 20279.7111, 21309.4032, 22261.9667, 23128.5416, 23900.2674, 24568.2841, 25013.9636, 25548.6648},
        {190.658088, 186.119716, 181.859002, 177.865646, 174.129347, 170.639803, 167.386713, 164.359778, 161.548695, 158.943163, 156.532883, 154.307552, 152.256869, 150.370535, 148.638247, 147.049704, 145.594607, 144.262653, 143.043541, 141.926972, 140.902643, 139.960254, 139.089503, 138.28009, 137.521714, 136.804074, 136.116868, 135.449797, 134.792557, 134.13485, 133.466373, 132.776827, 132.055908, 131.293318, 130.128873, 128.734573, 127.286622, 125.786072, 124.233978, 122.631392, 120.979369, 119.27896, 117.53122, 115.737202, 113.897959, 112.014544, 110.088012, 108.119414, 106.109805, 104.060238, 101.971766, 99.8450684, 97.6815811, 95.482365, 93.2484815, 90.980992, 88.680958, 86.3494409, 83.9875021, 81.5962029, 79.1766049, 76.7297693, 74.2567577, 71.7586314, 69.2364519, 66.6912805, 64.1241787, 61.4408214, 58.5491735, 55.6470241, 52.7416182, 49.8402006, 46.9500163, 44.0783102, 41.2323273, 38.4193125, 35.6465107, 32.9211669, 30.2505259, 27.6418328, 25.1023325, 22.6392698, 20.2598897, 17.9714372, 16.6848923, 15.9366235, 15.2521371, 14.6093501, 13.9861796, 13.3605425, 12.7103558, 12.0135366, 11.2480017, 10.3916682, 9.4224531, 8.31827337, 7.057046, 5.61668799, 3.97511633, 2.110248, 0, -4.60808665, -9.46811088, -14.5516083, -19.8301144, -25.2751649, -30.8582953, -36.5510411, -42.3249381, -48.1515218, -54.0023277, -59.8488914, -65.6627486, -71.4154348, -77.0784856, -82.6234365, -88.0218232, -94.066421, -100.846085, -106.474786, -111.100038, -114.869358, -117.93026, -120.43026, -122.516875, -124.337619, -126.040007, -127.771556, -129.679781, -131.912197, -136.299607, -140.689034, -144.723008, -148.044057, -148.824696, -147.442461, -144.666894, -140.581528, -135.269897, -128.815536, -121.301978, -117.563917, -111.487722, -102.227351, -88.9367645, -70.7699211, -43.6245261, -10.7248559, 27.9610668, 72.4652195, 122.819579, 176.290613, 239.055878, 315.67131, 408.310369, 514.282296, 633.080266, 764.197455, 901.201051, 1050.85703, 1226.39609, 1441.24426, 1685.35624, 1953.83487, 2244.31389, 2558.71563, 2896.99388, 3260.49515, 3649.53431, 4059.31963, 4487.94793, 4937.43635, 5401.35833, 5876.45371, 6359.46231, 6847.12395, 7288.47374, 7739.88241, 8210.01598, 8707.54045, 9241.12184, 9832.69282, 10468.8083, 11152.4486, 11918.2667, 12756.7059, 13645.0123, 14574.2888, 15535.6386, 16520.1646, 17518.9697, 18399.0187, 19387.6663, 20425.9712, 21454.9923, 22415.7883, 23249.4181, 23896.9405, 24172.2345, 24686.6078},
        {189.838058, 185.132123, 180.71382, 176.572573, 172.697805, 169.078942, 165.705406, 162.566622, 159.652013, 156.951004, 154.453018, 152.14748, 150.023813, 148.071441, 146.279789, 144.638279, 143.136337, 141.763386, 140.508849, 139.362152, 138.312717, 137.349969, 136.463331, 135.642229, 134.876084, 134.154323, 133.466367, 132.801642, 132.149572, 131.499579, 130.841089, 130.163525, 129.456311, 128.70887, 127.551256, 126.160136, 124.716499, 123.221431, 121.676019, 120.081349, 118.438506, 116.748578, 115.01265, 113.231809, 111.407141, 109.539733, 107.630669, 105.681038, 103.691925, 101.664416, 99.5995974, 97.502937, 95.3710447, 93.2049117, 91.0055294, 88.773889, 86.5109818, 84.2177991, 81.8953322, 79.5445723, 77.1665107, 74.7621387, 72.3324476, 69.8784286, 67.4010731, 64.9013723, 62.3803175, 59.7423937, 56.8944031, 54.0363738, 51.1755532, 48.3191887, 45.4745276, 42.6488174, 39.8493054, 37.083239, 34.3578656, 31.6804326, 29.0581873, 26.4983772, 24.0082496, 21.5950519, 19.2660315, 17.0284358, 15.8073234, 15.1317312, 14.5196646, 13.9485862, 13.3959588, 12.8392448, 12.255907, 11.6234081, 10.9192105, 10.1207771, 9.20557045, 8.15105317, 6.9346879, 5.53393731, 3.92626404, 2.08913072, 0, -5.21120401, -10.5755188, -16.0705728, -21.6739947, -27.363413, -33.1164562, -38.9107529, -44.7239316, -50.5336208, -56.3174492, -62.0530452, -67.7180374, -73.2900544, -78.7467246, -84.0656767, -89.2245392, -94.9249787, -101.248058, -106.5337, -110.914684, -114.523791, -117.493799, -119.957488, -122.047638, -123.897028, -125.638439, -127.404648, -129.328437, -131.542584, -135.781581, -139.989201, -143.817711, -146.919376, -147.764948, -146.59745, -143.974977, -139.904251, -134.391992, -127.444921, -119.06976, -112.836542, -104.10419, -92.1822524, -76.3802819, -56.007829, -28.0224346, 5.3263376, 44.1409341, 88.5238014, 138.577386, 191.151193, 253.576428, 331.814504, 428.546533, 539.756341, 664.434698, 801.572372, 943.947242, 1098.17499, 1277.649, 1495.75545, 1741.93916, 2010.86609, 2299.81664, 2610.50126, 2942.59695, 3296.80275, 3673.06062, 4067.56501, 4478.63018, 4908.73815, 5351.3942, 5803.4699, 6261.83681, 6723.36649, 7134.19576, 7554.6146, 7994.17828, 8462.44204, 8968.96112, 9539.85088, 10157.0664, 10823.0657, 11573.4992, 12397.6862, 13271.1161, 14183.8, 15125.7488, 16086.9735, 17057.4851, 17572.5258, 18424.6002, 19483.3789, 20618.5322, 21699.7306, 22596.6446, 23178.9446, 23309.4829, 23794.8419},
        {189.297894, 184.419818, 179.840056, 175.547722, 171.531928, 167.781785, 164.286406, 161.034904, 158.016391, 155.219978, 152.634779, 150.249906, 148.05447, 146.037585, 144.188363, 142.495915, 140.949354, 139.537793, 138.250344, 137.076118, 136.004229, 135.023789, 134.12391, 133.293704, 132.522284, 131.798761, 131.112249, 130.451859, 129.806705, 129.165897, 128.518548, 127.853772, 127.160679, 126.428383, 125.274788, 123.882418, 122.438258, 120.943468, 119.399206, 117.80663, 116.1669, 114.481174, 112.75061, 110.976368, 109.159607, 107.301484, 105.403158, 103.465789, 101.490535, 99.4785543, 97.4310061, 95.3607458, 93.2569816, 91.1206187, 88.9525623, 86.7537174, 84.5249893, 82.2672831, 79.9815039, 77.6685567, 75.3293469, 72.9647794, 70.5757595, 68.1631923, 65.7279829, 63.2710364, 60.793258, 58.1976469, 55.389553, 52.571805, 49.751655, 46.9363548, 44.1331566, 41.3493122, 38.5920736, 35.8686928, 33.1864219, 30.5525127, 27.9742172, 25.4587875, 23.0134755, 20.6455332, 18.3622125, 16.1707655, 15.0086412, 14.3984276, 13.8515325, 13.3450211, 12.8559587, 12.3614106, 11.8384421, 11.2641185, 10.6155051, 9.8696672, 9.00367007, 7.994579, 6.8194593, 5.45537626, 3.87939517, 2.06858132, 0, -6.10704066, -12.2072193, -18.2888584, -24.3402808, -30.3498088, -36.3057652, -42.1964726, -48.0102534, -53.7354305, -59.3603262, -64.8732633, -70.2625643, -75.5165519, -80.6235485, -85.5718769, -90.3498596, -95.4202995, -100.839302, -105.52196, -109.558272, -113.038234, -116.051843, -118.689095, -121.039989, -123.19452, -125.242685, -127.274483, -129.379909, -131.64896, -135.471705, -139.161376, -142.417949, -144.941401, -145.747094, -144.877312, -142.556897, -138.691208, -133.185605, -125.945446, -116.87609, -107.979109, -96.4256703, -81.7110054, -63.330344, -40.7789164, -12.9009902, 19.9945009, 58.2495857, 102.206293, 152.206653, 205.587689, 269.369218, 349.673977, 449.267039, 564.058244, 693.035291, 835.185879, 984.148342, 1145.47551, 1330.55588, 1550.69468, 1796.82561, 2064.56445, 2351.49038, 2658.50936, 2985.00028, 3331.07938, 3696.31674, 4077.57885, 4473.26149, 4886.06611, 5309.42042, 5740.22979, 6175.39959, 6611.83517, 6991.36093, 7379.73345, 7787.62834, 8225.72119, 8704.68763, 9255.69583, 9855.2671, 10505.2945, 11242.2862, 12054.4333, 12915.5044, 13814.4118, 14740.068, 15681.3854, 16627.2765, 16801.6959, 17534.4768, 18621.0475, 19856.8365, 21037.2723, 21957.7832, 22413.7978, 22455.6388, 22836.2406},
        {188.77964, 183.766421, 179.059159, 174.646756, 170.518117, 166.662143, 163.067738, 159.723804, 156.619245, 153.742964, 151.083864, 148.630846, 146.372816, 144.298674, 142.397325, 140.657671, 139.068616, 137.619061, 136.297911, 135.094068, 133.996434, 132.993914, 132.07541, 131.229825, 130.446061, 129.713022, 129.019611, 128.354731, 127.707284, 127.066174, 126.420303, 125.758575, 125.069892, 124.343158, 123.187473, 121.78963, 120.340986, 118.84277, 117.29621, 115.702534, 114.06297, 112.378747, 110.651094, 108.881237, 107.070406, 105.219828, 103.330732, 101.404347, 99.4418999, 97.4446197, 95.4137344, 93.3700648, 91.294822, 89.1888095, 87.0528307, 84.8876893, 82.6941888, 80.4731327, 78.2253244, 75.9515677, 73.6526658, 71.3294225, 68.9826413, 66.6131255, 64.2216789, 61.8091049, 59.376207, 56.8248356, 54.0592169, 51.2842379, 48.5071169, 45.7350721, 42.9753217, 40.2350841, 37.5215775, 34.8420201, 32.2036302, 29.6136261, 27.0792261, 24.6076483, 22.2061111, 19.8818327, 17.6420313, 15.4939253, 14.3726077, 13.801843, 13.2942427, 12.826914, 12.3769636, 11.9214987, 11.437626, 10.9024525, 10.2930852, 9.58663094, 8.76019672, 7.79088945, 6.65581607, 5.3320835, 3.79679867, 2.02706853, 0, -6.91953016, -13.668077, -20.2450749, -26.649958, -32.8821606, -38.9411169, -44.8262612, -50.5370278, -56.0728508, -61.4331645, -66.6174033, -71.6250012, -76.4553927, -81.1080118, -85.582293, -89.8776704, -94.2251199, -98.657153, -102.649479, -106.251148, -109.511211, -112.478718, -115.202719, -117.732264, -120.116404, -122.404189, -124.64467, -126.886896, -129.179918, -132.621978, -135.828231, -138.532969, -140.470486, -140.997702, -140.037596, -137.613779, -133.573756, -127.765031, -120.03511, -110.231496, -99.4398193, -85.8926385, -69.195216, -48.9528141, -24.7706955, 2.333288, 34.5196105, 72.5361569, 117.130812, 169.05146, 227.661972, 296.785821, 380.107762, 479.522157, 593.267977, 721.225391, 863.274563, 1016.17891, 1183.6437, 1372.77431, 1591.04569, 1832.74945, 2095.63113, 2378.24437, 2680.51216, 3001.72892, 3342.225, 3701.56341, 4075.50878, 4461.97429, 4862.33466, 5270.50905, 5683.21244, 6097.15983, 6509.0662, 6850.70673, 7199.97118, 7569.80947, 7973.17156, 8423.00737, 8961.79444, 9553.27006, 10197.7295, 10924.5771, 11722.5859, 12568.3998, 13451.3982, 14360.9604, 15286.4657, 16217.2935, 16265.5747, 16988.0358, 18122.5813, 19407.1161, 20579.5449, 21377.7724, 21539.7034, 21707.8462, 21725.4226},
        {188.497448, 183.355267, 178.526043, 173.998492, 169.761327, 165.803264, 162.113016, 158.679298, 155.490825, 152.53631, 149.804469, 147.284015, 144.963664, 142.832129, 140.878125, 139.090367, 137.457568, 135.968444, 134.611709, 133.376077, 132.250262, 131.22298, 130.282944, 129.418869, 128.61947, 127.87346, 127.169554, 126.496467, 125.842914, 125.197608, 124.549263, 123.886596, 123.198319, 122.473147, 121.312217, 119.906183, 118.450362, 116.946043, 115.39452, 113.797085, 112.15503, 110.469646, 108.742227, 106.974063, 105.166448, 103.320673, 101.43803, 99.5198116, 97.56731, 95.581817, 93.5646247, 91.5439614, 89.4935992, 87.4142462, 85.3066105, 83.1714005, 81.0093243, 78.82109, 76.6074058, 74.3689799, 72.1065205, 69.8207357, 67.5123338, 65.182023, 62.8305113, 60.458507, 58.0667182, 55.555995, 52.8295849, 50.0941461, 47.3568601, 44.6249086, 41.9054731, 39.2057353, 36.5328766, 33.8940787, 31.2965232, 28.7473915, 26.2538654, 23.8231264, 21.462356, 19.1787359, 16.9794476, 14.8716727, 13.7847928, 13.2454922, 12.7693711, 12.333684, 11.9156853, 11.4926293, 11.0417706, 10.5403634, 9.96566228, 9.29492159, 8.50539575, 7.57433919, 6.47900632, 5.19665154, 3.70452929, 1.97989397, 0, -7.69421501, -15.0541537, -22.0904992, -28.8139347, -35.2351435, -41.3648086, -47.2136134, -52.7922409, -58.1113743, -63.1816969, -68.0138919, -72.6186423, -77.0066315, -81.1885426, -85.1750587, -88.9768631, -92.568755, -95.9562913, -99.2313213, -102.396839, -105.455837, -108.411311, -111.266253, -114.023658, -116.686519, -119.25783, -121.740585, -124.137778, -126.452402, -129.454286, -132.098416, -134.157736, -135.405188, -135.509842, -134.296578, -131.617976, -127.278141, -121.081178, -112.831194, -102.332294, -90.0289952, -74.8900826, -56.5943636, -34.8206455, -9.24773565, 17.1363182, 48.7892594, 86.8595906, 132.495815, 186.846435, 251.522381, 326.644539, 412.807759, 510.824995, 621.9545, 747.254593, 887.783599, 1044.23716, 1218.11871, 1411.3273, 1626.73635, 1862.75818, 2119.72883, 2397.43423, 2694.72821, 3010.89246, 3346.69424, 3701.80038, 4070.43061, 4449.88583, 4839.88225, 5234.93392, 5631.50571, 6026.06247, 6415.06906, 6719.57661, 7030.31715, 7362.60896, 7731.77036, 8153.11962, 8680.92276, 9265.58522, 9905.7334, 10622.6661, 11406.0352, 12236.0458, 13102.8224, 13996.4891, 14907.1702, 15824.99, 15864.0701, 16616.685, 17784.8129, 19070.432, 20175.5205, 20802.0566, 20652.0185, 20951.2976, 20536.414},
        {188.502602, 183.231796, 178.280587, 173.637514, 169.291119, 165.229944, 161.44253, 157.917418, 154.64315, 151.608267, 148.80131, 146.21082, 143.825339, 141.633408, 139.623569, 137.784362, 136.104329, 134.572012, 133.175951, 131.904688, 130.746764, 129.690721, 128.7251, 127.838442, 127.019288, 126.25618, 125.537659, 124.852267, 124.188544, 123.535032, 122.880273, 122.212807, 121.521176, 120.793921, 119.625387, 118.209504, 116.744943, 115.233036, 113.675117, 112.072521, 110.42658, 108.738629, 107.010002, 105.242032, 103.436053, 101.593399, 99.7154034, 97.8034001, 95.8587228, 93.8827053, 91.8766814, 89.8742874, 87.8438542, 85.7860154, 83.7014045, 81.5906549, 79.4544003, 77.293274, 75.1079097, 72.8989407, 70.6670007, 68.4127231, 66.1367414, 63.8396892, 61.5221999, 59.1849071, 56.8284442, 54.3529043, 51.6607943, 48.960105, 46.2579875, 43.5615929, 40.8780724, 38.2145771, 35.5782582, 32.9762667, 30.4157539, 27.9038709, 25.4477688, 23.0545988, 20.7315119, 18.4856594, 16.3241924, 14.2542619, 13.1991386, 12.6886445, 12.2415908, 11.8353986, 11.4474893, 11.0552843, 10.6362048, 10.1676721, 9.62710751, 8.99193237, 8.23956797, 7.3474356, 6.29295658, 5.05355221, 3.60664378, 1.92965261, 0, -8.33119073, -16.188563, -23.5925594, -30.5636224, -37.1221946, -43.2887187, -49.0836372, -54.5273925, -59.6404274, -64.4431844, -68.956106, -73.1996349, -77.1942135, -80.9602845, -84.5182904, -87.8886738, -90.7985494, -93.2264558, -95.8563153, -98.6457144, -101.55224, -104.533478, -107.547015, -110.550439, -113.501335, -116.357291, -119.075892, -121.614726, -123.931378, -126.476, -128.533007, -129.912218, -130.42345, -129.977213, -128.332979, -125.223367, -120.427989, -113.726454, -104.898371, -93.7233522, -80.362149, -64.0972288, -44.6336302, -21.6763916, 5.06944819, 31.356456, 63.1575867, 101.9034, 149.024457, 205.951318, 276.117157, 356.502279, 444.767777, 540.738648, 648.657848, 770.784761, 909.378769, 1069.13233, 1249.31879, 1446.55719, 1659.08928, 1889.44168, 2140.45923, 2413.15231, 2705.42285, 3016.5996, 3347.98608, 3699.42302, 4063.51038, 4436.94389, 4817.79622, 5201.17437, 5583.29747, 5960.38466, 6328.65504, 6599.58527, 6875.32259, 7173.27174, 7510.83746, 7905.42452, 8421.64044, 8998.21867, 9632.3342, 10338.0382, 11105.8255, 11919.1592, 12768.8857, 13645.8515, 14540.9029, 15444.8864, 15593.3871, 16403.3934, 17572.5016, 18798.3084, 19778.4102, 20210.4036, 19791.885, 20144.3048, 19288.2486},
        {188.855419, 183.442128, 178.355277, 173.583244, 169.114411, 164.937157, 161.039863, 157.410909, 154.038676, 150.911543, 148.017891, 145.3461, 142.884551, 140.621624, 138.545698, 136.645155, 134.908374, 133.323737, 131.879622, 130.564411, 129.366484, 128.274221, 127.276002, 126.360207, 125.515218, 124.729413, 123.991174, 123.288881, 122.610914, 121.945652, 121.281478, 120.60677, 119.90991, 119.179276, 118.005974, 116.585378, 115.117637, 113.604024, 112.045812, 110.444274, 108.800682, 107.116309, 105.392427, 103.63031, 101.831229, 99.9964584, 98.1272699, 96.2249362, 94.2907302, 92.3259244, 90.3317916, 88.3387146, 86.318225, 84.2709641, 82.1975738, 80.0986955, 77.9749709, 75.8270417, 73.6555494, 71.4611357, 69.2444421, 67.0061103, 64.746782, 62.4670987, 60.1677021, 57.8492337, 55.5123352, 53.0565987, 50.3840425, 47.7037153, 45.0228092, 42.3485165, 39.6880295, 37.0485403, 34.4372413, 31.8613247, 29.3279826, 26.8444075, 24.4177915, 22.0553268, 19.7642058, 17.5516206, 15.4247635, 13.3908268, 12.3783277, 11.9144171, 11.5151723, 11.1578868, 10.819854, 10.4783675, 10.1107206, 9.69420695, 9.20611992, 8.62375302, 7.9243997, 7.08535347, 6.08390778, 4.89735611, 3.50299194, 1.87810875, 0, -8.40481087, -16.3110508, -23.7412706, -30.7180212, -37.2638534, -43.401318, -49.1529658, -54.5413479, -59.5890148, -64.3185177, -68.7524071, -72.9132341, -76.8235494, -80.505904, -83.9828485, -87.276934, -89.9739769, -92.0341374, -94.4678179, -97.2039832, -100.171598, -103.299628, -106.517037, -109.752791, -112.935855, -115.995192, -118.859769, -121.45855, -123.720501, -125.928363, -127.527606, -128.341061, -128.191558, -126.93039, -124.366157, -120.299873, -114.545831, -106.918326, -97.2316521, -85.3001023, -72.0486476, -55.8428732, -36.2797667, -12.9563157, 14.5304921, 43.7957465, 78.7273631, 120.425335, 169.989655, 228.520317, 298.294055, 377.795881, 464.611553, 558.753265, 665.187185, 786.655248, 925.899389, 1089.69585, 1275.83537, 1477.70763, 1690.79729, 1919.56566, 2167.51617, 2436.08134, 2723.18468, 3028.36041, 3353.4355, 3698.5386, 4055.39146, 4420.47125, 4790.99312, 5162.58272, 5531.46175, 5893.85192, 6245.97494, 6497.82713, 6753.41195, 7030.50742, 7346.89162, 7720.34258, 8216.93399, 8773.9512, 9388.47415, 10074.9466, 10823.5752, 11617.3319, 12446.7852, 13302.5036, 14175.0556, 15055.0097, 15587.1553, 16473.792, 17512.4258, 18500.5631, 19235.7098, 19515.3722, 19137.0564, 19107.4115, 17954.8506},
        {189.594339, 184.03074, 178.800493, 173.891822, 169.292949, 164.992098, 160.977489, 157.237348, 153.759896, 150.533356, 147.545952, 144.785905, 142.24144, 139.900778, 137.752142, 135.783756, 133.983842, 132.340623, 130.842322, 129.477161, 128.233364, 127.099154, 126.062752, 125.112383, 124.236269, 123.422632, 122.659696, 121.935683, 121.238816, 120.557319, 119.879413, 119.193322, 118.487269, 117.749476, 116.568786, 115.141513, 113.66877, 112.151741, 110.591608, 108.989556, 107.346768, 105.664427, 103.943717, 102.185821, 100.391923, 98.5632065, 96.7008547, 94.8060512, 92.8799794, 90.9238229, 88.9387651, 86.9492549, 84.932706, 82.8897975, 80.8212086, 78.7276185, 76.6097062, 74.4681508, 72.3036316, 70.1168277, 67.9084181, 65.679082, 63.4294985, 61.1603468, 58.872306, 56.5660553, 54.2422737, 51.8005204, 49.1427823, 46.478291, 43.814281, 41.1579864, 38.5166416, 35.8974809, 33.3077388, 30.7546495, 28.2454473, 25.7873666, 23.3876417, 21.0535069, 18.7921966, 16.6109451, 14.5169868, 12.5175559, 11.5482181, 11.132516, 10.7828341, 10.4762806, 10.1899637, 9.90099167, 9.58647261, 9.22351476, 8.78922629, 8.26071542, 7.61509032, 6.8294592, 5.88093025, 4.74661166, 3.40361162, 1.82903834, 0, -8.27828127, -16.0741106, -23.4093162, -30.3057263, -36.7851691, -42.8694729, -48.5804658, -53.9399762, -58.9698322, -63.691862, -68.1278939, -72.2997561, -76.2292768, -79.9382842, -83.4486067, -86.7820723, -89.4299413, -91.3382466, -93.7289106, -96.5100684, -99.589855, -102.876406, -106.277855, -109.702339, -113.057992, -116.25295, -119.195346, -121.793317, -123.954998, -125.853682, -127.03512, -127.327901, -126.560613, -124.451569, -120.884495, -115.77252, -108.977314, -100.360547, -89.7838895, -77.1090099, -64.2910316, -48.461033, -29.0710951, -5.57329893, 22.5802745, 55.7900943, 94.7883919, 140.159949, 192.489546, 252.361966, 319.987597, 396.783205, 481.232706, 573.909315, 679.651895, 801.384118, 942.029657, 1109.54884, 1300.68398, 1506.6829, 1721.23502, 1949.78316, 2195.83998, 2460.94409, 2743.2892, 3042.6044, 3361.08782, 3699.10868, 4047.98296, 4404.14766, 4764.48994, 5124.92739, 5481.74117, 5831.21246, 6169.62244, 6408.12635, 6649.41278, 6911.04438, 7210.58381, 7565.59373, 8040.63515, 8574.94013, 9165.92916, 9830.94121, 10559.5124, 11332.8794, 12140.9935, 12973.8056, 13821.2668, 14673.3283, 15646.2022, 16608.0257, 17481.1722, 18188.015, 18650.9272, 18792.2822, 18534.4532, 17994.7952, 16619.3861},
        {190.764245, 185.042591, 179.661346, 174.608573, 169.872334, 165.44069, 161.301704, 157.443437, 153.853952, 150.521311, 147.433575, 144.578807, 141.945068, 139.520421, 137.292928, 135.25065, 133.38165, 131.67399, 130.115731, 128.694936, 127.399666, 126.217984, 125.137951, 124.14763, 123.235083, 122.388371, 121.595557, 120.844702, 120.123869, 119.42112, 118.724516, 118.022119, 117.301992, 116.552197, 115.359182, 113.920263, 112.437584, 110.912239, 109.345324, 107.737933, 106.091163, 104.406107, 102.683862, 100.925522, 99.1321829, 97.3049391, 95.444886, 93.5531187, 91.6307323, 89.678822, 87.698483, 85.7078065, 83.690523, 81.6473593, 79.5790419, 77.4862975, 75.3698529, 73.2304348, 71.0687697, 68.8855844, 66.6816056, 64.4575598, 62.214174, 59.9521746, 57.6722884, 55.3752421, 53.0617623, 50.6319961, 47.9885055, 45.339455, 42.6920915, 40.0536619, 37.4314129, 34.8325915, 32.2644444, 29.7342186, 27.2491609, 24.816518, 22.443537, 20.1374645, 17.9055475, 15.7550328, 13.6931673, 11.7271978, 10.7982606, 10.4270148, 10.1228946, 9.86284053, 9.62379279, 9.38269176, 9.11647779, 8.80209119, 8.41647233, 7.93656152, 7.33929912, 6.60162545, 5.70048086, 4.61280568, 3.31554025, 1.78562491, 0, -8.08248674, -15.7141745, -22.9149239, -29.7045958, -36.1030508, -42.1301498, -47.8057534, -53.1497224, -58.1819174, -62.9221993, -67.3904287, -71.6064664, -75.5901731, -79.3614096, -82.9400365, -86.3459146, -89.0092413, -90.8656396, -93.2687327, -96.1120249, -99.2890204, -102.693224, -106.218139, -109.75727, -113.204122, -116.452199, -119.395005, -121.926044, -123.938821, -125.530869, -126.316853, -126.129068, -124.799809, -121.933017, -117.475166, -111.433622, -103.709188, -94.2026662, -82.8148579, -69.4465655, -57.0006641, -41.4622081, -22.1446375, 1.63860808, 30.5740886, 67.9291695, 111.164404, 160.321152, 215.440771, 276.564621, 341.634228, 415.357244, 497.555273, 589.489601, 695.360866, 818.049995, 960.437918, 1130.90183, 1325.57723, 1534.60374, 1750.75434, 1979.72547, 2224.71601, 2487.16363, 2765.52168, 3059.61338, 3371.72422, 3702.31578, 4042.82123, 4389.78072, 4740.32903, 5090.32019, 5436.0885, 5773.96826, 6100.29379, 6328.21525, 6558.04712, 6806.91973, 7091.96344, 7430.30858, 7883.99702, 8395.30656, 8962.11974, 9605.54627, 10314.2746, 11067.3152, 11853.8405, 12663.0228, 13484.0345, 14306.0479, 15669.367, 16687.9868, 17399.5596, 17841.7377, 18052.1735, 18068.5194, 17928.4277, 16862.3765, 15332.8491},
        {192.497459, 186.61208, 181.074284, 175.871957, 170.992988, 166.425264, 162.156674, 158.175104, 154.468444, 151.024581, 147.831402, 144.876797, 142.148651, 139.634854, 137.323294, 135.201857, 133.258432, 131.480908, 129.857171, 128.375109, 127.022611, 125.787565, 124.657857, 123.621377, 122.666012, 121.779649, 120.950177, 120.165484, 119.413457, 118.681985, 117.958954, 117.232254, 116.489771, 115.719394, 114.502622, 113.039085, 111.533297, 109.986335, 108.399274, 106.773191, 105.109161, 103.408261, 101.671565, 99.9001516, 98.0950951, 96.2574718, 94.3883577, 92.4888288, 90.5599611, 88.6028305, 86.618513, 84.6236047, 82.6033249, 80.5584133, 78.4896094, 76.3976526, 74.2832824, 72.1472386, 69.9902604, 67.8130876, 65.6164595, 63.4011157, 61.1677958, 58.9172393, 56.6501856, 54.3673744, 52.0695451, 49.6584363, 47.0381635, 44.4136483, 41.792048, 39.1805202, 36.586222, 34.0163108, 31.4779441, 28.9782792, 26.5244734, 24.123684, 21.7830685, 19.5097842, 17.3109885, 15.1938387, 13.1654921, 11.2331062, 10.3351598, 9.99356038, 9.71902173, 9.488478, 9.27886337, 9.06711203, 8.83015815, 8.54493592, 8.18837952, 7.73742313, 7.16900094, 6.46004712, 5.58749585, 4.52828133, 3.25933772, 1.75759922, 0, -8.060563, -15.6727343, -22.8557667, -29.6289128, -36.0114253, -42.0225568, -47.68156, -53.0076877, -58.0201924, -62.7383269, -67.1813438, -71.3684958, -75.3190356, -79.0522159, -82.5872892, -85.9435084, -88.5295689, -90.2775008, -92.5875792, -95.348223, -98.4478511, -101.774883, -105.217737, -108.664832, -112.004587, -115.125422, -117.915756, -120.264006, -122.058593, -123.350977, -123.806753, -123.265428, -121.566508, -118.413372, -113.71359, -107.411033, -99.3860465, -89.5189742, -77.6901615, -63.7799529, -50.7269981, -34.4225486, -14.1485853, 10.812911, 41.1799595, 81.3844195, 127.50201, 179.32229, 236.63482, 299.229157, 363.96123, 437.13978, 519.524134, 612.750054, 720.242677, 844.623749, 988.515013, 1159.70009, 1354.4657, 1563.4674, 1779.86185, 2008.85526, 2253.53357, 2515.11326, 2791.64304, 3082.62566, 3389.55571, 3712.45237, 4044.03137, 4381.13971, 4722.43328, 5062.67183, 5398.07616, 5724.8671, 6039.26543, 6253.61504, 6468.9829, 6702.55905, 6971.53355, 7293.09644, 7729.50913, 8222.84271, 8770.97091, 9396.79615, 10088.285, 10822.9754, 11589.3553, 12375.9127, 13171.1357, 13963.5121, 15392.0723, 16398.13, 17041.534, 17382.1333, 17479.7767, 17394.313, 17185.5911, 15798.1249, 14204.1909},
        {194.832096, 188.776504, 183.075888, 177.717942, 172.690359, 167.980833, 163.577058, 159.466728, 155.637537, 152.07718, 148.773349, 145.713739, 142.886043, 140.277956, 137.877171, 135.671383, 133.648285, 131.795572, 130.100936, 128.552073, 127.136675, 125.842437, 124.657053, 123.568217, 122.563622, 121.630963, 120.757933, 119.932226, 119.141536, 118.373558, 117.615985, 116.85651, 116.082829, 115.282634, 114.029097, 112.525891, 110.981675, 109.397593, 107.774791, 106.114416, 104.417614, 102.68553, 100.91931, 99.1201013, 97.2890487, 95.4272986, 93.535997, 91.6162897, 89.6693228, 87.6962422, 85.698194, 83.6966837, 81.6720563, 79.6250165, 77.556269, 75.4665182, 73.3564689, 71.2268256, 69.0782931, 66.9115759, 64.7273787, 62.526406, 60.3093625, 58.0769529, 55.8298817, 53.5688536, 51.2945732, 48.9115002, 46.326405, 43.7384902, 41.1546998, 38.5819774, 36.0272669, 33.497512, 30.9996566, 28.5406444, 26.1274193, 23.7669251, 21.4661054, 19.2319042, 17.0712653, 14.9911324, 12.9984492, 11.1001598, 10.2216415, 9.89130344, 9.62651543, 9.40441637, 9.20214512, 8.99684051, 8.7656414, 8.48568664, 8.13411508, 7.68806558, 7.12467698, 6.42108814, 5.55443789, 4.50186511, 3.24050863, 1.74750731, 0, -8.3054306, -16.1178844, -23.4584701, -30.3482963, -36.8084714, -42.8601041, -48.524303, -53.8221767, -58.7748337, -63.4033827, -67.7289322, -71.7725908, -75.5554671, -79.0986696, -82.4233071, -85.550488, -87.8976964, -89.3998088, -91.443275, -93.9198555, -96.7213107, -99.7394012, -102.865887, -105.99253, -109.011088, -111.813324, -114.290998, -116.33587, -117.8397, -118.820072, -118.996788, -118.223863, -116.355309, -113.474859, -109.321668, -103.573632, -96.0158486, -86.4334161, -74.6114326, -60.3349964, -45.4125529, -26.9900516, -4.45671816, 22.7982215, 55.3855416, 96.6044673, 143.705209, 196.626428, 255.306786, 319.684945, 387.197633, 463.343361, 549.076032, 645.929736, 756.623123, 883.343002, 1028.27618, 1197.66001, 1388.71018, 1594.22357, 1809.01498, 2037.18078, 2282.09775, 2544.70598, 2821.81675, 3112.13652, 3415.40023, 3730.58096, 4052.88083, 4379.63502, 4712.32067, 5043.4938, 5369.05587, 5684.90832, 5986.95259, 6182.72455, 6378.58262, 6592.51965, 6842.52846, 7146.60192, 7571.30135, 8053.67209, 8590.89193, 9204.58385, 9882.2494, 10601.2046, 11349.4602, 12115.0271, 12885.916, 13650.1378, 14739.1959, 15649.7785, 16345.9799, 16791.8948, 16951.6177, 16789.2431, 16268.8656, 14843.2369, 13275.4857},
        {197.624867, 191.386169, 185.510427, 179.985123, 174.797735, 169.935744, 165.386631, 161.137874, 157.176955, 153.491352, 150.068547, 146.896019, 143.961248, 141.251715, 138.754899, 136.458281, 134.34934, 132.415556, 130.644411, 129.023383, 127.539952, 126.181599, 124.935805, 123.790048, 122.731808, 121.748567, 120.827804, 119.956999, 119.123632, 118.315183, 117.519132, 116.72296, 115.914146, 115.08017, 113.78475, 112.236161, 110.64777, 109.02081, 107.356511, 105.656107, 103.920829, 102.151909, 100.35058, 98.5180718, 96.6556179, 94.76445, 92.8458, 90.9008998, 88.9309814, 86.9372767, 84.9210177, 82.9093594, 80.8770488, 78.8247557, 76.7531503, 74.6629026, 72.5546824, 70.4291598, 68.2870047, 66.1288873, 63.9554774, 61.7674451, 59.5654603, 57.3501931, 55.1223134, 52.8824912, 50.6313966, 48.2769725, 45.7293884, 43.1805566, 40.6371582, 38.1058743, 35.593386, 33.1063744, 30.6515207, 28.2355058, 25.865011, 23.5467172, 21.2873057, 19.0934575, 16.9718537, 14.9291754, 12.9721037, 11.1073198, 10.2431727, 9.91654415, 9.65347854, 9.43139609, 9.22771703, 9.0198616, 8.78525004, 8.5013026, 8.14543951, 7.695081, 7.12764732, 6.42055871, 5.5512354, 4.49709764, 3.23556566, 1.7440597, 0, -8.62074126, -16.6912379, -24.2351348, -31.2760767, -37.8377085, -43.9436751, -49.6176213, -54.8831919, -59.7640318, -64.2837858, -68.4660988, -72.3346155, -75.912981, -79.2248399, -82.2938371, -85.1436176, -87.1919372, -88.3758854, -90.0895697, -92.2268016, -94.6813927, -97.3471546, -100.117899, -102.887436, -105.54958, -107.99814, -110.126928, -111.829756, -113.000436, -113.5914, -113.416345, -112.351494, -110.273072, -107.74649, -104.30338, -99.2915915, -92.3805937, -83.2398559, -71.5388475, -56.9470377, -39.8310783, -18.9510703, 6.15992232, 35.9688354, 70.9426049, 112.645996, 160.173657, 213.718069, 273.471709, 339.627056, 410.612973, 490.541086, 580.429242, 681.606735, 796.038726, 925.404882, 1071.38488, 1238.3544, 1424.68436, 1625.80456, 1838.57643, 2065.76093, 2311.0143, 2574.96834, 2853.13012, 3143.35905, 3443.5416, 3751.54418, 4065.13402, 4382.13445, 4706.87807, 5029.67752, 5346.10292, 5651.72439, 5942.11204, 6119.48405, 6296.10048, 6490.86945, 6722.69907, 7010.49744, 7425.91937, 7899.96183, 8428.92721, 9032.38137, 9697.41824, 10401.4387, 11132.0213, 11876.7446, 12623.1873, 13358.928, 14000.4493, 14791.1348, 15569.402, 16173.6685, 16442.3518, 16213.8693, 15326.6386, 13948.8369, 12460.0222},
        {200.836347, 194.398442, 188.332279, 182.625103, 177.264154, 172.236675, 167.52991, 163.131101, 159.027489, 155.206318, 151.654831, 148.360268, 145.309875, 142.490891, 139.890561, 137.496127, 135.294831, 133.273915, 131.420623, 129.722197, 128.165879, 126.738911, 125.428537, 124.221999, 123.106539, 122.0694, 121.097824, 120.179054, 119.300333, 118.448902, 117.612005, 116.776884, 115.930781, 115.060939, 113.72099, 112.124101, 110.488653, 108.815961, 107.107337, 105.364097, 103.587554, 101.779021, 99.9398145, 98.0712468, 96.1746323, 94.2512851, 92.3025192, 90.3296486, 88.3339873, 86.3168495, 84.2795491, 82.2540924, 80.210436, 78.1492283, 76.071118, 73.976754, 71.8667847, 69.741859, 67.6026255, 65.4497328, 63.2838297, 61.1055648, 58.9155869, 56.7145444, 54.5030863, 52.2818611, 50.0515175, 47.7241595, 45.214017, 42.7043554, 40.2015635, 37.7120299, 35.2421433, 32.7982923, 30.3868657, 28.0142521, 25.6868402, 23.4110186, 21.193176, 19.0397011, 16.9569825, 14.951409, 13.0293692, 11.1972517, 10.3439668, 10.0160798, 9.74954857, 9.52212236, 9.31155029, 9.09558153, 8.85196522, 8.55845051, 8.19278655, 7.7327225, 7.15600749, 6.44039069, 5.56362125, 4.5034483, 3.23762101, 1.74388853, 0, -8.97576982, -17.3358487, -25.106968, -32.3158592, -38.9892538, -45.1538832, -50.836479, -56.0637725, -60.8624952, -65.2593786, -69.2811541, -72.9545532, -76.3063073, -79.3631479, -82.1518065, -84.6990145, -86.3923876, -87.1673549, -88.4962671, -90.2689021, -92.3750381, -94.704453, -97.146925, -99.5922322, -101.930153, -104.050464, -105.842945, -107.197374, -108.003528, -108.088826, -107.42827, -105.930348, -103.503545, -101.25793, -98.5011985, -94.2206226, -87.9642142, -79.2799846, -67.7159454, -52.820108, -33.4255122, -10.0127412, 17.7303077, 50.1157372, 87.4556502, 129.362653, 177.023219, 230.924324, 291.552947, 359.396062, 434.026151, 517.962404, 612.221311, 717.890606, 836.203486, 968.32784, 1115.43156, 1280.01872, 1461.61734, 1658.29775, 1868.96316, 2095.05306, 2340.43054, 2605.43441, 2884.43172, 3174.53067, 3472.0048, 3773.74589, 4079.70617, 4388.10668, 4705.64395, 5020.75967, 5328.74111, 5624.87551, 5904.45013, 6064.66101, 6223.40944, 6400.50545, 6615.75911, 6888.98048, 7297.07454, 7764.69313, 8287.17705, 8881.61152, 9534.73016, 10224.2065, 10937.227, 11660.9785, 12382.6476, 13089.4209, 13256.6554, 13920.2628, 14786.4491, 15561.4206, 15951.3835, 15662.5438, 14401.1079, 13117.621, 11744.8263},
        {204.39771, 197.733075, 191.450578, 185.537183, 179.979857, 174.765567, 169.881277, 165.313954, 161.050564, 157.078074, 153.383448, 149.953653, 146.775654, 143.836419, 141.122912, 138.6221, 136.320949, 134.206425, 132.265493, 130.48512, 128.852271, 127.353914, 125.977012, 124.708533, 123.535443, 122.444707, 121.423292, 120.458163, 119.536287, 118.644628, 117.770155, 116.899831, 116.020624, 115.119499, 113.74103, 112.102683, 110.427245, 108.716036, 106.97038, 105.191599, 103.381016, 101.539952, 99.6697316, 97.7716757, 95.8471073, 93.8973487, 91.9237225, 89.9275512, 87.9101571, 85.8728628, 83.8169907, 81.7728756, 79.7121986, 77.6356535, 75.5439337, 73.4377328, 71.3177442, 69.1846616, 67.0391785, 64.8819885, 62.713785, 60.5352616, 58.3471119, 56.1500294, 53.9447076, 51.7318401, 49.5121204, 47.2026637, 44.7225922, 42.2449731, 39.775918, 37.3215387, 34.8879465, 32.4812532, 30.1075703, 27.7730094, 25.483682, 23.2456997, 21.0651742, 18.9482169, 16.9009395, 14.9294536, 13.0398708, 11.2383025, 10.3959729, 10.0682408, 9.80006719, 9.56948899, 9.35454323, 9.13326695, 8.88369716, 8.58387088, 8.21182512, 7.7455969, 7.16322324, 6.44274116, 5.56218767, 4.49959979, 3.23301454, 1.74046894, 0, -9.31034946, -17.9394536, -25.9176986, -33.2754708, -40.0431565, -46.2511418, -51.9298132, -57.1095568, -61.8207589, -66.0938058, -69.9590837, -73.446979, -76.5878779, -79.4121666, -81.9502314, -84.2324586, -85.4875037, -85.6312199, -86.5074323, -87.9747276, -89.8916927, -92.1169143, -94.5089793, -96.9264744, -99.2279866, -101.272103, -102.917409, -104.022493, -104.445941, -103.674619, -102.075133, -99.6175856, -96.2720801, -93.714994, -91.0632933, -86.9789378, -80.9201482, -72.3451456, -60.7121508, -45.4793848, -24.4760517, 0.714823058, 30.316298, 64.5514315, 103.643282, 146.197697, 194.459248, 249.055297, 310.613206, 379.760335, 456.562801, 542.895177, 639.908981, 748.622405, 869.781429, 1004.25425, 1152.90908, 1317.37567, 1497.58758, 1692.64752, 1902.17744, 2127.10498, 2371.1502, 2634.53369, 2911.54639, 3199.20104, 3493.65511, 3791.69959, 4093.26158, 4396.49422, 4707.77489, 5015.76722, 5315.77752, 5603.11206, 5873.07715, 6020.91394, 6167.01013, 6331.68831, 6535.27106, 6798.08093, 7200.81545, 7663.1723, 8179.59908, 8764.6584, 9404.83567, 10078.2351, 10771.7611, 11472.3182, 12166.8108, 12842.1435, 12732.2398, 13307.4029, 14200.6099, 15044.838, 15473.0646, 15118.2668, 13613.4219, 12356.8185, 11109.2651},
        {208.276248, 201.356031, 194.830036, 188.684913, 182.907312, 177.483881, 172.40127, 167.646128, 163.205105, 159.06485, 155.212011, 151.633239, 148.315183, 145.244491, 142.407814, 139.7918, 137.383099, 135.16836, 133.134233, 131.267366, 129.55441, 127.982012, 126.536823, 125.205492, 123.974668, 122.831001, 121.761139, 120.751733, 119.78943, 118.860882, 117.952736, 117.051642, 116.14425, 115.217209, 113.807286, 112.135535, 110.428414, 108.687174, 106.913063, 105.107329, 103.271222, 101.40599, 99.5128815, 97.5931464, 95.6480328, 93.6787898, 91.6866661, 89.6729104, 87.6387717, 85.5854987, 83.5143402, 81.4468544, 79.3635405, 77.265207, 75.1526625, 73.0267155, 70.8881746, 68.7378483, 66.5765451, 64.4050736, 62.2242424, 60.0348599, 57.8377347, 55.6336754, 53.4234906, 51.2079886, 48.9879782, 46.6862926, 44.2276372, 41.7736386, 39.3301604, 36.9030657, 34.498218, 32.1214807, 29.778717, 27.4757904, 25.2185642, 23.0129018, 20.8646665, 18.7797217, 16.7639308, 14.8231571, 12.963264, 11.1901148, 10.3599125, 10.035465, 9.76935951, 9.53986436, 9.32524774, 9.10377784, 8.85372288, 8.55335105, 8.18093057, 7.71472964, 7.13301648, 6.41405928, 5.53612625, 4.4774856, 3.21640554, 1.73115427, 0, -9.60048058, -18.4577431, -26.606233, -34.0803959, -40.9146772, -47.1435225, -52.8013773, -57.9226871, -62.5418974, -66.6934537, -70.4118016, -73.7313865, -76.686654, -79.3120495, -81.6420187, -83.7110069, -84.4554721, -83.7530044, -84.1215928, -85.3603998, -87.2685877, -89.6453191, -92.2897564, -95.001062, -97.5783984, -99.820928, -101.527813, -102.498217, -102.531301, -100.558276, -97.5645062, -93.6095382, -88.7529202, -85.2213269, -81.9907428, -77.4592516, -71.0362632, -62.1311872, -50.1534336, -34.512412, -12.6912484, 13.3881034, 43.9393518, 79.1762056, 119.312373, 163.098524, 212.577165, 268.327766, 330.929795, 400.962721, 478.200192, 565.012387, 662.856411, 772.88677, 895.640162, 1031.93067, 1182.57238, 1349.46318, 1532.08702, 1728.74549, 1938.27405, 2162.01072, 2403.17066, 2662.0526, 2934.01344, 3216.67847, 3507.69617, 3804.69785, 4105.22968, 4406.88596, 4712.89658, 5014.3533, 5306.89967, 5586.17925, 5847.8356, 5988.57831, 6127.71839, 6285.63288, 6482.69883, 6739.29329, 7138.03516, 7595.56505, 8105.58492, 8680.44086, 9306.45485, 9962.18051, 10634.3282, 11309.6084, 11974.7316, 12616.408, 12475.2215, 13011.2874, 13857.1012, 14645.1582, 15007.954, 14577.9839, 12987.7432, 11668.0896, 10542.1139},
        {212.481508, 205.287329, 198.500391, 192.107004, 186.093475, 180.446112, 175.151223, 170.195115, 165.564098, 161.244478, 157.222564, 153.484663, 150.017085, 146.806135, 143.838123, 141.099357, 138.576144, 136.254792, 134.121609, 132.162904, 130.364983, 128.714156, 127.19673, 125.799012, 124.507311, 123.307936, 122.187193, 121.13139, 120.126836, 119.159839, 118.216706, 117.283746, 116.347266, 115.393575, 113.951431, 112.245541, 110.50611, 108.734295, 106.931251, 105.098136, 103.236106, 101.346316, 99.4299228, 97.4880832, 95.5219532, 93.532689, 91.5214468, 89.4893831, 87.437654, 85.3674158, 83.2798248, 81.185993, 79.0769051, 76.9535016, 74.8167229, 72.6675095, 70.5068018, 68.3355403, 66.1546655, 63.9651177, 61.7678375, 59.5637653, 57.3538416, 55.1390067, 52.9202013, 50.6983656, 48.4744402, 46.1771692, 43.73703, 41.3038628, 38.883288, 36.4809265, 34.1023988, 31.7533256, 29.4393275, 27.1660252, 24.9390394, 22.7639906, 20.6464996, 18.592187, 16.6066734, 14.6955795, 12.8645259, 11.1191334, 10.3003578, 9.97858218, 9.71414904, 9.48557084, 9.27136, 9.05002892, 8.80009, 8.50005566, 8.12843831, 7.66375035, 7.08450419, 6.36921225, 5.49638692, 4.44454063, 3.19218577, 1.71783476, 0, -9.86447367, -18.9259029, -27.2228206, -34.7937598, -41.6772535, -47.9118347, -53.5360363, -58.5883915, -63.1074332, -67.1316943, -70.6997079, -73.850007, -76.6211245, -79.0515934, -81.1799469, -83.0447177, -83.2624651, -81.668583, -81.5179737, -82.5444617, -84.4818712, -87.0640265, -90.024752, -93.0978719, -96.0172106, -98.5165923, -100.329841, -101.190782, -100.833239, -97.6007178, -93.1271447, -87.563439, -81.0605206, -76.3459783, -72.2823972, -67.0452506, -60.0124573, -50.5619362, -38.0716062, -21.9193862, 0.549391374, 27.3056079, 58.5736644, 94.5779621, 135.542902, 180.6023, 231.343789, 288.264417, 351.861229, 422.631275, 499.911877, 586.777246, 684.998831, 795.902523, 919.90453, 1057.82949, 1210.50204, 1380.36733, 1566.26119, 1765.25161, 1975.02954, 2197.57865, 2435.54601, 2689.36495, 2955.71762, 3233.00802, 3520.74655, 3817.62405, 4118.27458, 4419.62695, 4721.18346, 5016.9035, 5302.82503, 5574.98604, 5829.4245, 5965.27207, 6099.69962, 6252.97172, 6445.35291, 6697.10775, 7091.89707, 7544.32497, 8047.4862, 8612.12544, 9224.35098, 9862.91697, 10514.2994, 11164.9743, 11801.4176, 12410.1054, 12327.2508, 12843.2164, 13615.9859, 14303.5429, 14563.8712, 14054.9542, 12434.7757, 11046.5794, 10043.1546},
        {216.994077, 209.50994, 202.446817, 195.790654, 189.527398, 183.642994, 178.123389, 172.954528, 168.122357, 163.612823, 159.411872, 155.505448, 151.879499, 148.519971, 145.412809, 142.54396, 139.899369, 137.464982, 135.226746, 133.170606, 131.282509, 129.548401, 127.954226, 126.485933, 125.129466, 123.870771, 122.695795, 121.590484, 120.540783, 119.532638, 118.551996, 117.584803, 116.617004, 115.634546, 114.157672, 112.414974, 110.640644, 108.835739, 107.001319, 105.138442, 103.248168, 101.331556, 99.3896643, 97.4235521, 95.4342783, 93.4229019, 91.3904819, 89.3380772, 87.2667467, 85.1775495, 83.0715443, 80.9490391, 78.8118602, 76.6610829, 74.4977823, 72.3230339, 70.1379127, 67.9434941, 65.7408532, 63.5310653, 61.3152057, 59.0943495, 56.869572, 54.6419485, 52.4125541, 50.1824642, 47.952754, 45.6580108, 43.2344187, 40.8201787, 38.4206763, 36.0412969, 33.6874257, 31.3644483, 29.0777499, 26.8327159, 24.6347317, 22.4891826, 20.401454, 18.3769314, 16.4209999, 14.5390451, 12.7364523, 11.0186068, 10.2107078, 9.89098992, 9.62783465, 9.40002279, 9.18633517, 8.96555264, 8.71645604, 8.41782619, 8.04844395, 7.58709013, 7.01254559, 6.30359117, 5.43900769, 4.39757599, 3.15807692, 1.69929131, 0, -10.0916423, -19.324637, -27.7414023, -35.3843563, -42.2959171, -48.5185028, -54.0945316, -59.0664214, -63.4765904, -67.3674568, -70.7814386, -73.760954, -76.348421, -78.5862578, -80.5168825, -82.1827131, -81.882889, -79.4107884, -78.7519602, -79.5752625, -81.549554, -84.3436927, -87.6265373, -91.0669459, -94.333777, -97.095889, -99.0221401, -99.7813888, -99.0424934, -94.5871013, -88.6632592, -81.5059889, -73.3503119, -67.3390362, -62.2532904, -56.0987933, -48.2382307, -38.0342884, -24.8496523, -8.04700805, 15.0115394, 42.3538509, 74.2258226, 110.873351, 152.542332, 198.841999, 250.814076, 308.863627, 373.395712, 444.815394, 521.988612, 608.740701, 707.138717, 818.684463, 943.725677, 1083.12825, 1237.75807, 1410.71449, 1600.12233, 1801.66994, 2011.78185, 2233.18361, 2467.97682, 2716.75663, 2977.56554, 3249.61777, 3534.34304, 3831.52995, 4132.84354, 4434.544, 4732.43389, 5023.30142, 5303.5652, 5569.64381, 5817.95585, 5950.47561, 6081.67707, 6231.58989, 6420.24375, 6667.66831, 7057.71547, 7504.04453, 7999.3328, 8553.67443, 9152.85254, 9775.39322, 10407.4424, 11035.1462, 11644.6504, 12222.101, 12271.4115, 12784.0659, 13464.2209, 14016.0332, 14143.6596, 13551.2567, 11942.9813, 10491.8984, 9604.62562},
        {221.794543, 214.006835, 206.654486, 199.723063, 193.198133, 187.065264, 181.310024, 175.917979, 170.874699, 166.16575, 161.7767, 157.693116, 153.900566, 150.384618, 147.130839, 144.124797, 141.352059, 138.798193, 136.448767, 134.289348, 132.305503, 130.482801, 128.806809, 127.263094, 125.837224, 124.514767, 123.28129, 122.12236, 121.023546, 119.970415, 118.948535, 117.943472, 116.940795, 115.926071, 114.410217, 112.626109, 110.812329, 108.969848, 107.099637, 105.202668, 103.27991, 101.332336, 99.3609147, 97.3666186, 95.3504182, 93.3132846, 91.2561885, 89.1801009, 87.0859928, 84.9748351, 82.8475988, 80.6947406, 78.5279737, 76.3484967, 74.1575085, 71.9562081, 69.7457943, 67.5274659, 65.3024218, 63.0718609, 60.8369819, 58.5989839, 56.3590656, 54.1184258, 51.8782636, 49.6397776, 47.4041669, 45.1115345, 42.7034509, 40.3071197, 37.9277005, 35.5703527, 33.2402357, 30.942509, 28.6823321, 26.4648643, 24.295265, 22.1786939, 20.1203101, 18.1252733, 16.1987429, 14.3458782, 12.5718387, 10.8817839, 10.0843611, 9.76608561, 9.50381518, 9.27663454, 9.06362841, 8.84388149, 8.5964785, 8.30050413, 7.9350431, 7.47918011, 6.91199988, 6.21258712, 5.36002652, 4.3334028, 3.11180067, 1.67430483, 0, -10.2712999, -19.6346497, -28.1359195, -35.8209796, -42.7357003, -48.9259518, -54.4376043, -59.3165281, -63.6085935, -67.3596707, -70.6156299, -73.4223413, -75.8256753, -77.8715021, -79.6056919, -81.0741149, -80.2911498, -77.0124536, -75.8789373, -76.5011516, -78.489647, -81.4549741, -85.0076835, -88.7583259, -92.3174517, -95.2956117, -97.3033564, -97.9512364, -96.8498023, -91.3025842, -84.0730606, -75.4638886, -65.7777247, -58.4505887, -52.2184558, -44.9817381, -36.1030835, -24.9451401, -10.8705556, 6.75802191, 30.4608675, 58.4193466, 90.9024134, 128.179022, 170.518126, 217.950591, 271.042982, 330.123775, 395.521449, 467.564478, 544.721152, 631.453697, 730.078544, 842.247387, 968.254747, 1109.00449, 1265.40047, 1441.13101, 1633.6827, 1837.50453, 2047.86889, 2268.20043, 2500.16367, 2744.51355, 3000.46386, 3267.9358, 3550.02243, 3847.4673, 4149.38378, 4451.46392, 4746.44621, 5033.43068, 5309.13177, 5570.26395, 5813.54166, 5943.66934, 6072.37395, 6219.37245, 6404.3818, 6647.11894, 7030.80468, 7469.31622, 7955.15459, 8499.0501, 9086.28801, 9694.55803, 10309.5252, 10916.8545, 11502.211, 12051.2597, 12290.7874, 12814.712, 13388.7631, 13778.67, 13750.1625, 13068.97, 11500.822, 10003.6567, 9218.76563},
        {226.661734, 218.529844, 210.851508, 203.61176, 196.795632, 190.388156, 184.374366, 178.739295, 173.467976, 168.545441, 163.956724, 159.686858, 155.720874, 152.043807, 148.640689, 145.496553, 142.596433, 139.92536, 137.468368, 135.21049, 133.136758, 131.232206, 129.481867, 127.870773, 126.383958, 125.006454, 123.723294, 122.519511, 121.380138, 120.290208, 119.234754, 118.198809, 117.167405, 116.125577, 114.56484, 112.727367, 110.861259, 108.967566, 107.047336, 105.101617, 103.131457, 101.137906, 99.1220117, 97.0848221, 95.0273859, 92.9507516, 90.8559677, 88.7440825, 86.6161446, 84.4732023, 82.316304, 80.1405699, 77.9531052, 75.7550869, 73.5476921, 71.3320976, 69.1094805, 66.8810176, 64.647886, 62.4112625, 60.1723243, 57.9322481, 55.692211, 53.4533899, 51.2169618, 48.9841037, 46.7559925, 44.4745891, 42.0832764, 39.7053598, 37.345855, 35.0097776, 32.7021433, 30.4279679, 28.1922671, 26.0000564, 23.8563517, 21.7661685, 19.7345227, 17.7664298, 15.8669056, 14.0409657, 12.2936258, 10.6299017, 9.84296829, 9.52718375, 9.26743319, 9.04315503, 8.83378765, 8.61876947, 8.37753889, 8.08953431, 7.73419415, 7.29095679, 6.73926065, 6.05854414, 5.22824564, 4.22780358, 3.03665635, 1.63424235, 0, -10.5379488, -20.098656, -28.7324995, -36.4898577, -43.4211086, -49.5766305, -55.0068015, -59.7619997, -63.8926033, -67.4489906, -70.4815395, -73.0406284, -75.1766354, -76.9399386, -78.3809162, -79.5499464, -78.3918991, -74.6563655, -73.156183, -73.4905494, -75.2586624, -78.05972, -81.4929199, -85.15746, -88.652538, -91.577352, -93.5310995, -94.1129786, -92.922187, -87.3262431, -79.9743066, -71.1350792, -61.0772625, -52.9994044, -45.7052834, -37.2173858, -26.9254593, -14.2192517, 1.51148915, 20.8770155, 46.609873, 76.5521051, 110.898733, 149.844776, 193.585257, 240.937528, 293.818696, 352.768196, 418.325468, 491.029949, 570.116568, 659.02367, 760.021152, 874.613766, 1002.74409, 1145.0561, 1302.19375, 1476.91679, 1667.27255, 1869.0002, 2078.30877, 2298.09154, 2530.31506, 2775.63669, 3032.49544, 3300.34855, 3581.15313, 3875.01453, 4172.87209, 4470.84998, 4765.13449, 5051.7781, 5326.88346, 5586.55322, 5826.89006, 5948.35374, 6066.60052, 6201.64376, 6373.49684, 6602.17312, 6970.66206, 7394.0169, 7865.26129, 8401.53417, 8986.31354, 9593.0293, 10205.6619, 10808.1916, 11384.5988, 11918.864, 12445.028, 13008.8734, 13457.4876, 13637.958, 13397.3721, 12582.8174, 11041.3813, 9628.65598, 8864.87599},
        {231.819896, 223.34108, 215.333505, 207.781691, 200.670158, 193.983424, 187.706011, 181.822438, 176.317224, 171.174889, 166.379954, 161.916939, 157.770362, 153.924743, 150.364604, 147.074462, 144.038839, 141.242254, 138.669227, 136.304277, 134.131925, 132.13669, 130.303091, 128.61565, 127.058886, 125.617317, 124.275466, 123.01785, 121.82899, 120.693406, 119.595617, 118.520144, 117.451506, 116.374223, 114.760854, 112.862667, 110.937085, 108.985242, 107.00827, 105.007306, 102.983481, 100.937931, 98.8717889, 96.7861888, 94.6822646, 92.5611504, 90.42398, 88.2718873, 86.1060062, 83.9274707, 81.7374148, 79.5366542, 77.3266478, 75.1085366, 72.8834612, 70.6525625, 68.4169813, 66.1778584, 63.9363347, 61.6935509, 59.450648, 57.2087666, 54.9690476, 52.7326319, 50.5006602, 48.2742734, 46.0546123, 43.7852852, 41.4112818, 39.0523994, 36.7135083, 34.399479, 32.1151818, 29.8654872, 27.6552656, 25.4893875, 23.3727231, 21.310143, 19.3065175, 17.3667171, 15.4956122, 13.6980732, 11.9789705, 10.3431744, 9.56657614, 9.2523183, 8.99427657, 8.77230028, 8.56623873, 8.35594124, 8.12125711, 7.84203568, 7.49812623, 7.0693781, 6.53564058, 5.876763, 5.07259466, 4.10298488, 2.94778297, 1.58683824, 0, -10.7055637, -20.3809392, -29.0797613, -36.855665, -43.762285, -49.8532563, -55.1822136, -59.8027918, -63.7686259, -67.1333505, -69.9506006, -72.274011, -74.1572166, -75.6538523, -76.8175528, -77.701953, -76.2437831, -72.194263, -70.3952643, -70.4467995, -71.948881, -74.5015212, -77.7047325, -81.1585273, -84.462918, -87.2179171, -89.023537, -89.47979, -88.1866886, -82.6921112, -75.4006533, -66.5279676, -56.2897067, -47.5296413, -39.1493649, -29.3496398, -17.5576831, -3.20071218, 14.2940558, 35.4994036, 63.2884447, 95.2335306, 131.457379, 172.082708, 217.232236, 264.461967, 317.103012, 375.919766, 441.676627, 515.137991, 596.392889, 687.706529, 791.247198, 908.335495, 1038.56133, 1182.29166, 1339.89345, 1513.05415, 1700.5201, 1899.59706, 2107.65631, 2326.95663, 2559.93347, 2807.13305, 3065.90044, 3335.05009, 3614.98334, 3904.92606, 4198.28553, 4491.76038, 4785.74202, 5072.66812, 5347.95902, 5607.03503, 5845.31649, 5959.15429, 6067.80555, 6191.45795, 6350.29918, 6564.51693, 6915.85668, 7321.90978, 7776.48198, 8304.21732, 8886.71475, 9492.72394, 10104.4863, 10704.2432, 11274.2361, 11796.7064, 12624.7516, 13235.3037, 13559.6295, 13528.9953, 13074.6679, 12127.9139, 10619.9997, 9309.51523, 8550.81911},
        {237.302875, 228.48453, 220.153508, 212.293881, 204.889717, 197.925088, 191.384063, 185.250713, 179.509109, 174.14332, 169.137417, 164.47547, 160.14155, 156.119726, 152.39407, 148.948651, 145.76754, 142.834808, 140.134523, 137.650758, 135.367581, 133.269064, 131.339277, 129.562289, 127.922172, 126.402996, 124.98883, 123.663746, 122.411813, 121.217103, 120.063684, 118.935628, 117.817005, 116.691885, 115.016368, 113.049616, 111.057133, 109.040125, 106.999795, 104.937346, 102.853984, 100.750912, 98.6293338, 96.4904535, 94.3354751, 92.1656026, 89.9820399, 87.7859909, 85.5786597, 83.3612502, 81.1349664, 78.9054311, 76.6693336, 74.4277821, 72.1818847, 69.9327496, 67.6814849, 65.4291989, 63.1769997, 60.9259954, 58.6772943, 56.4320045, 54.1912341, 51.9560914, 49.7276845, 47.5071216, 45.2955108, 43.0384478, 40.6828611, 38.3443492, 36.027619, 33.7373773, 31.4783309, 29.2551866, 27.0726512, 24.9354316, 22.8482345, 20.8157669, 18.8427354, 16.9338469, 15.0938083, 13.3273263, 11.6391078, 10.0338596, 9.26752306, 8.9533865, 8.695709, 8.47481838, 8.27104244, 8.064709, 7.83614587, 7.56568086, 7.23364178, 6.82035644, 6.30615265, 5.67135823, 4.89630099, 3.96130873, 2.84670927, 1.53283042, 0, -10.7251324, -20.3930402, -29.0586791, -36.7770051, -43.6029739, -49.5915415, -54.7976636, -59.2762961, -63.0823948, -66.2709156, -68.8968144, -71.0150468, -72.6805689, -73.9483365, -74.8733053, -75.5104313, -73.8396212, -69.6193952, -67.5982726, -67.3865559, -68.5945474, -70.8325498, -73.7108654, -76.8397967, -79.8296463, -82.2907165, -83.8333098, -84.0677288, -82.6042758, -77.2397451, -70.0629004, -61.2280968, -50.8896894, -41.4813468, -32.0177671, -20.9058315, -7.61610093, 8.38086388, 27.614502, 50.6142526, 80.2762909, 114.043009, 151.980789, 194.156012, 240.635062, 287.897574, 340.493363, 399.385496, 465.537042, 539.911069, 623.387813, 717.114413, 823.133622, 942.577556, 1074.71913, 1219.66599, 1377.52581, 1548.93579, 1733.35368, 1929.65952, 2136.38973, 2355.2033, 2589.11332, 2838.57182, 3099.69119, 3370.5758, 3649.92668, 3936.00286, 4224.87568, 4513.85394, 4807.50867, 5094.76924, 5370.43064, 5629.28787, 5866.13591, 5974.49021, 6075.74519, 6190.01575, 6337.41675, 6538.06309, 6871.15746, 7258.33504, 7694.38736, 8211.80029, 8790.44467, 9394.48538, 10004.5786, 10601.3804, 11165.5469, 11677.7344, 12792.8308, 13450.3716, 13660.7857, 13434.5022, 12781.9503, 11713.5592, 10239.7578, 9033.38751, 8272.24862},
        {243.144518, 234.00418, 225.364549, 217.209352, 209.522315, 202.287164, 195.487627, 189.107429, 183.130297, 177.539957, 172.320137, 167.454562, 162.926959, 158.721055, 154.820575, 151.209248, 147.870797, 144.788952, 141.947437, 139.32998, 136.920307, 134.702143, 132.659217, 130.775255, 129.033982, 127.419125, 125.914411, 124.503567, 123.170318, 121.898392, 120.671514, 119.473412, 118.287811, 117.098439, 115.349495, 113.305818, 111.23873, 109.149466, 107.039261, 104.90935, 102.760969, 100.595352, 98.4137339, 96.2173511, 94.0074381, 91.78523, 89.5519621, 87.3088695, 85.0571872, 82.7981504, 80.5329943, 78.2693381, 76.0018944, 73.73176, 71.4600315, 69.1878059, 66.9161797, 64.6462497, 62.3791127, 60.1158654, 57.8576046, 55.6054269, 53.3604292, 51.1237081, 48.8963605, 46.679483, 44.4741724, 42.2289018, 39.8934082, 37.5773201, 35.2851457, 33.0213931, 30.7905705, 28.5971859, 26.4457474, 24.3407631, 22.2867412, 20.2881898, 18.3496169, 16.4755307, 14.6704393, 12.9388507, 11.2852732, 9.71421474, 8.95814741, 8.64228561, 8.38309419, 8.16145742, 7.95825959, 7.75438498, 7.53071785, 7.2681425, 6.94754319, 6.54980421, 6.05580985, 5.44644436, 4.70259204, 3.80513717, 2.73496402, 1.47295687, 0, -10.5476431, -20.0465, -28.5502273, -36.1124817, -42.7869199, -48.6271986, -53.6869744, -58.019904, -61.679644, -64.7198512, -67.1941822, -69.1562936, -70.6598422, -71.7584847, -72.5058776, -72.9556777, -71.1722325, -66.9250116, -64.7672995, -64.3264725, -65.2299066, -67.104978, -69.5790629, -72.2795373, -74.8337773, -76.8691592, -78.013059, -77.8928529, -76.135917, -70.8087012, -63.6718475, -54.8210098, -44.3518424, -34.2945682, -23.7775568, -11.4132925, 3.28294146, 20.7958618, 41.6101854, 66.2106288, 97.3531198, 132.559926, 171.871397, 215.327882, 262.96973, 310.618016, 363.587183, 422.972398, 489.868831, 565.371648, 650.93904, 746.859461, 855.057362, 976.50493, 1110.23014, 1256.13389, 1414.11706, 1583.95439, 1765.70159, 1959.55198, 2164.98727, 2383.23914, 2617.94905, 2869.52218, 3132.88, 3405.46107, 3684.39674, 3967.04591, 4251.89408, 4536.78949, 4829.67431, 5116.74992, 5392.37049, 5650.89021, 5886.66324, 5992.78069, 6090.17558, 6198.51785, 6337.47747, 6526.72437, 6841.33332, 7208.6329, 7624.54815, 8128.98383, 8700.45631, 9299.15709, 9904.51878, 10495.974, 11052.9554, 11554.8956, 12912.1385, 13610.4453, 13726.553, 13337.1988, 12519.1198, 11349.0531, 9903.73575, 8787.42593, 8024.81818},
        {249.378672, 239.944017, 231.019659, 222.589129, 214.635957, 207.143673, 200.095808, 193.475892, 187.267455, 181.454029, 176.019142, 170.946327, 166.219112, 161.821029, 157.735608, 153.946378, 150.436872, 147.190618, 144.191148, 141.421991, 138.866679, 136.508741, 134.331708, 132.319111, 130.454479, 128.721343, 127.103233, 125.58368, 124.146215, 122.774367, 121.451667, 120.161645, 118.887832, 117.613759, 115.778345, 113.648881, 111.499203, 109.330516, 107.144024, 104.940929, 102.722437, 100.489752, 98.2440768, 95.9866162, 93.7185741, 91.4411545, 89.1555613, 86.8629987, 84.5646706, 82.2617811, 79.9555341, 77.6508127, 75.3450621, 73.0394067, 70.7349707, 68.4328782, 66.1342537, 63.8402214, 61.5519055, 59.2704303, 56.99692, 54.732499, 52.4782915, 50.2354217, 48.0050139, 45.7881925, 43.5860815, 41.3514725, 39.0383171, 36.7474228, 34.4830468, 32.2494469, 30.0508803, 27.8916047, 25.7758776, 23.7079563, 21.6920984, 19.7325614, 17.8336028, 15.99948, 14.2344506, 12.542772, 10.9287018, 9.39649733, 8.65078756, 8.33091288, 8.06779582, 7.84296549, 7.63795098, 7.43428136, 7.21348575, 6.95709323, 6.6466329, 6.26363384, 5.78962515, 5.20613593, 4.49469526, 3.63683224, 2.61407596, 1.40795552, 0, -10.1240838, -19.2528596, -27.4353802, -34.7206986, -41.1578676, -46.79594, -51.6839689, -55.871007, -59.4061071, -62.3383223, -64.7167053, -66.5903091, -68.0081865, -69.0193903, -69.6729735, -70.017989, -68.2344365, -64.1043614, -61.9044364, -61.2832032, -61.8892033, -63.370978, -65.377069, -67.5560178, -69.5563658, -71.0266546, -71.6154256, -70.9712205, -68.7425807, -63.2385361, -55.9382941, -46.8922496, -36.1507976, -25.409353, -13.8958008, -0.399354103, 15.5230982, 34.3146671, 56.4184636, 82.2775986, 114.29864, 150.363668, 190.531639, 234.86151, 283.412238, 331.996959, 385.981904, 446.487484, 514.634111, 591.542195, 678.884272, 776.553811, 886.395357, 1009.2826, 1144.10705, 1290.65017, 1448.69345, 1617.50265, 1797.49215, 1989.63884, 2193.92716, 2411.47173, 2646.53508, 2899.55335, 3164.4792, 3438.24129, 3716.80713, 3996.85618, 4278.59232, 4560.22582, 4851.4788, 5137.27864, 5411.85076, 5669.42055, 5904.21343, 6012.44495, 6110.85284, 6218.16498, 6353.10924, 6534.4135, 6831.15316, 7178.14355, 7572.53505, 8060.46868, 8619.70268, 9207.5825, 9802.88702, 10384.3951, 10930.8856, 11421.1375, 12945.5473, 13671.8934, 13722.5285, 13219.8054, 12286.0767, 11043.6954, 9615.014, 8558.78359, 7804.18143},
        {256.310456, 246.55014, 237.311236, 228.577037, 220.330834, 212.55592, 205.235587, 198.353127, 191.891833, 185.834996, 180.16591, 174.867866, 169.924157, 165.318074, 161.032911, 157.051959, 153.35851, 149.935857, 146.767293, 143.836109, 141.125597, 138.61905, 136.299761, 134.151021, 132.156122, 130.298357, 128.561019, 126.927398, 125.380789, 123.904482, 122.481771, 121.095946, 119.730302, 118.368129, 116.44334, 114.225472, 111.990667, 109.740045, 107.474727, 105.195835, 102.904488, 100.601807, 98.288914, 95.9669285, 93.6369716, 91.300164, 88.9576264, 86.6104795, 84.2598442, 81.906841, 79.5525907, 77.1913325, 74.8312178, 72.4735166, 70.1194987, 67.7704341, 65.4275927, 63.0922443, 60.7656589, 58.4491064, 56.1438567, 53.8511797, 51.5723453, 49.3086234, 47.0612839, 44.8315968, 42.6208318, 40.3895442, 38.0992664, 35.8352398, 33.6013735, 31.4015769, 29.2397593, 27.1198299, 25.0456979, 23.0212728, 21.0504638, 19.1371801, 17.285331, 15.4988259, 13.7815739, 12.1374845, 10.5704668, 9.08443011, 8.34747162, 8.01743636, 7.74418637, 7.51007675, 7.29746261, 7.08869906, 6.86614121, 6.61214417, 6.30906304, 5.93925294, 5.48506897, 4.92886624, 4.25299986, 3.43982494, 2.47169658, 1.3309699, 0, -9.09106002, -17.3505096, -24.815023, -31.5212745, -37.5059384, -42.805689, -47.4572005, -51.4971472, -54.9622034, -57.8890434, -60.3143414, -62.2747718, -63.8070087, -64.9477265, -65.7335995, -66.2013019, -64.6352781, -60.8218839, -58.7611706, -58.1143347, -58.5425726, -59.7070808, -61.2690558, -62.8896939, -64.2301916, -64.9517453, -64.7155515, -63.1828067, -60.0147072, -53.5057286, -45.1849192, -35.1234919, -23.3926593, -11.7007601, 0.699556008, 14.9922059, 31.5971146, 50.9342071, 73.4234083, 99.4846432, 130.601308, 165.806192, 205.311148, 249.328032, 298.068697, 349.428469, 406.514864, 470.118869, 541.031471, 620.043657, 707.767293, 805.391412, 914.941816, 1037.45284, 1171.93457, 1318.30592, 1476.48584, 1646.43287, 1828.01732, 2021.06624, 2224.28906, 2440.06067, 2673.55169, 2925.61054, 3189.76188, 3462.8921, 3741.05921, 4020.93481, 4302.22794, 4582.9295, 4869.47371, 5150.10918, 5419.90412, 5673.92673, 5907.24523, 6029.80008, 6143.06917, 6263.40267, 6407.15071, 6590.66346, 6871.77112, 7197.6904, 7566.99428, 8028.56929, 8561.07262, 9122.19751, 9691.64935, 10249.1335, 10774.3555, 11247.0206, 12747.6884, 13472.1267, 13532.6972, 13041.762, 12111.6827, 10854.8213, 9383.53956, 8288.86497, 7603.97683},
        {263.496459, 253.510953, 244.047121, 235.088425, 226.618327, 218.620288, 211.077772, 203.97424, 197.293153, 191.017974, 185.132165, 179.619188, 174.462505, 169.645577, 165.151867, 160.964837, 157.067948, 153.444664, 150.078444, 146.952753, 144.051051, 141.3568, 138.853463, 136.524502, 134.353378, 132.323554, 130.418491, 128.621652, 126.916498, 125.286492, 123.715095, 122.185769, 120.681977, 119.18718, 117.147403, 114.826132, 112.493311, 110.149842, 107.796627, 105.434571, 103.064575, 100.687543, 98.3043769, 95.9159801, 93.5232554, 91.1271056, 88.7284334, 86.3281417, 83.9271333, 81.5263111, 79.1265779, 76.7039045, 74.2846662, 71.8703063, 69.4622682, 67.0619951, 64.6709304, 62.2905174, 59.9221993, 57.5674196, 55.2276215, 52.9042483, 50.5987433, 48.3125498, 46.0471112, 43.8038707, 41.5842717, 39.3580695, 37.0957801, 34.8642012, 32.66683, 30.5071642, 28.3887012, 26.3149387, 24.289374, 22.3155047, 20.3968283, 18.5368423, 16.7390443, 15.0069317, 13.3440021, 11.753753, 10.2396818, 8.80528613, 8.07867708, 7.73790745, 7.45352266, 7.20875032, 6.98681802, 6.77095337, 6.54438396, 6.29033738, 5.99204125, 5.63272316, 5.19561071, 4.6639315, 4.02091312, 3.24978319, 2.33376929, 1.25609903, 0, -7.95266151, -15.25935, -21.9425623, -28.024795, -33.5285448, -38.4763085, -42.8905828, -46.7938644, -50.20865, -53.1574363, -55.6627201, -57.7469981, -59.4327669, -60.7425233, -61.6987639, -62.3239857, -61.0427614, -57.649313, -55.7804886, -55.1163725, -55.3370492, -56.1226033, -57.1531193, -58.1086816, -58.6693747, -58.5152831, -57.3264913, -54.7830839, -50.5651452, -42.9682715, -33.5646813, -22.4498056, -9.7190754, 2.87850845, 16.0742999, 31.0406047, 48.1780628, 67.8873142, 90.5689988, 116.623756, 146.553239, 180.626332, 219.223912, 262.726856, 311.516039, 365.76898, 426.120754, 493.003078, 566.847669, 648.086244, 736.183567, 833.720137, 942.85822, 1064.79483, 1198.75622, 1344.8535, 1503.19775, 1674.60713, 1858.32497, 2052.82325, 2255.49308, 2469.84986, 2701.58125, 2951.87937, 3214.31888, 3485.97098, 3763.50509, 4043.88824, 4325.56063, 4606.12914, 4887.62853, 5162.55661, 5426.98753, 5676.99544, 5908.6545, 6046.9636, 6176.84098, 6312.12963, 6466.67252, 6654.31263, 6921.53586, 7227.78032, 7572.89916, 8006.99814, 8510.03876, 9041.04393, 9581.02208, 10110.9816, 10611.931, 11064.8787, 12471.8669, 13176.7991, 13266.238, 12826.7461, 11944.8861, 10707.2205, 9200.31202, 8045.30998, 7423.40339},
        {270.751476, 260.710374, 251.173764, 242.125924, 233.551136, 225.433681, 217.75784, 210.507893, 203.668121, 197.222805, 191.156225, 185.452662, 180.096397, 175.071711, 170.362884, 165.954198, 161.829932, 157.974368, 154.371786, 151.006467, 147.862692, 144.924741, 142.176896, 139.603437, 137.188644, 134.916799, 132.772182, 130.739074, 128.801755, 126.944507, 125.15161, 123.407345, 121.695993, 120.001833, 117.809478, 115.363553, 112.914688, 110.463381, 108.010126, 105.55542, 103.09976, 100.64364, 98.1875563, 95.7320059, 93.2774844, 90.8244876, 88.3735116, 85.9250523, 83.4796058, 81.037668, 78.599735, 76.1147647, 73.6359083, 71.1647791, 68.7029902, 66.252155, 63.8138865, 61.3897982, 58.9815031, 56.5906145, 54.2187456, 51.8675097, 49.53852, 47.2333896, 44.9537319, 42.7011601, 40.4772873, 38.2619356, 36.03545, 33.8443867, 31.6917715, 29.5806304, 27.5139891, 25.4948736, 23.5263096, 21.6113232, 19.7529401, 17.9541862, 16.2180874, 14.5476695, 12.9459584, 11.41598, 9.96076012, 8.58332466, 7.87019175, 7.52036813, 7.22584745, 6.97072011, 6.73907653, 6.51500714, 6.28260235, 6.02595258, 5.72914826, 5.3762798, 4.95143763, 4.43871215, 3.8221938, 3.08597299, 2.21414014, 1.19078567, 0, -6.91626146, -13.3585771, -19.3358386, -24.8569373, -29.9307649, -34.5662128, -38.7721726, -42.5575358, -45.931194, -48.9020386, -51.4789612, -53.6708534, -55.4866066, -56.9351124, -58.0252624, -58.765948, -57.775652, -54.8646362, -53.1867642, -52.4529829, -52.3742391, -52.6614795, -53.025651, -53.1777003, -52.8285743, -51.6892197, -49.4705833, -45.8836119, -40.6392524, -32.1528659, -21.9060328, -9.99837559, 3.47048336, 16.799163, 30.6889201, 46.2681473, 63.9177496, 84.0186323, 106.9517, 133.097859, 162.12971, 195.352032, 233.284312, 276.446037, 325.356693, 381.927394, 444.938092, 514.560366, 590.965796, 674.325961, 763.344047, 861.454729, 970.781734, 1092.58371, 1226.35308, 1372.37527, 1530.93571, 1703.49635, 1888.89869, 2084.70077, 2287.61303, 2501.54776, 2731.95501, 2980.10744, 3240.19941, 3509.67434, 3786.17015, 4067.18071, 4349.48665, 4630.27207, 4907.41653, 5177.30161, 5436.87066, 5683.067, 5912.83395, 6063.90294, 6207.23219, 6354.56802, 6517.65674, 6708.24466, 6963.63477, 7252.97892, 7577.07057, 7986.16391, 8461.47418, 8963.60798, 9474.81095, 9977.32877, 10453.4071, 10885.2916, 12164.9223, 12833.7508, 12950.2752, 12572.994, 11760.4051, 10571.0069, 9063.29766, 7852.74255, 7259.15085},
        {277.756863, 268.025127, 258.744131, 249.900329, 241.480175, 233.470124, 225.85663, 218.626147, 211.765129, 205.260031, 199.097307, 193.263411, 187.744798, 182.527922, 177.599236, 172.945196, 168.552256, 164.40687, 160.495491, 156.804576, 153.320576, 150.029948, 146.919145, 143.974622, 141.182833, 138.530231, 136.003272, 133.58841, 131.272098, 129.040792, 126.880945, 124.779012, 122.721447, 120.694704, 118.276571, 115.666432, 113.068581, 110.482735, 107.908609, 105.345922, 102.794387, 100.253723, 97.7236446, 95.2038687, 92.6941116, 90.1940894, 87.7035186, 85.2221153, 82.7495958, 80.2856765, 77.8300735, 75.2893062, 72.7583082, 70.2388162, 67.7325666, 65.2412962, 62.7667416, 60.3106394, 57.8747261, 55.4607384, 53.070413, 50.7054864, 48.3676952, 46.0587761, 43.7804656, 41.5345004, 39.3226171, 37.1358791, 34.9656058, 32.8355468, 30.7481306, 28.7057857, 26.7109407, 24.766024, 22.8734642, 21.0356897, 19.2551292, 17.5342111, 15.875364, 14.2810163, 12.7535966, 11.2955334, 9.90925528, 8.59719067, 7.90347557, 7.54829068, 7.24530582, 6.97937005, 6.73533245, 6.49804208, 6.25234802, 5.98309934, 5.67514511, 5.3133344, 4.88251628, 4.36753983, 3.75325411, 3.0245082, 2.16615116, 1.16303207, 0, -6.41807523, -12.4516376, -18.1007465, -23.3654616, -28.2458422, -32.7419481, -36.8538386, -40.5815733, -43.9252117, -46.8848134, -49.4604377, -51.6521443, -53.4599927, -54.8840423, -55.9243527, -56.5809835, -55.8133002, -53.4726917, -51.9131682, -50.9117832, -50.2455904, -49.6916435, -49.0269963, -48.0287023, -46.4738152, -44.1393889, -40.8024768, -36.2401328, -30.2294105, -21.7232474, -11.6249893, 0.0410752582, 13.2506578, 26.739999, 41.1045458, 57.2702705, 75.5847258, 96.3954644, 120.050039, 146.896002, 176.284952, 209.863513, 248.174098, 291.75912, 341.160993, 397.926746, 461.343022, 531.701081, 609.292182, 694.407584, 786.449154, 887.68458, 999.604526, 1123.16057, 1258.20363, 1405.07882, 1564.13123, 1736.61106, 1921.75261, 2117.80275, 2322.8146, 2538.94142, 2769.48911, 3015.08294, 3271.80609, 3537.82661, 3812.36453, 4093.86063, 4376.8986, 4658.24391, 4933.4736, 5200.72444, 5457.1733, 5699.99705, 5926.37256, 6077.23422, 6220.06199, 6366.09336, 6526.56581, 6712.71683, 6960.17223, 7239.52228, 7551.55235, 7945.7775, 8403.69423, 8886.49498, 9376.69146, 9856.79536, 10309.3184, 10716.7722, 11825.9143, 12414.5585, 12523.5457, 12193.717, 11465.9133, 10380.9756, 8979.74498, 7773.8466, 7091.52596},
        {284.741703, 275.469862, 266.579733, 258.060508, 249.901383, 242.091548, 234.620199, 227.476529, 220.64973, 214.128997, 207.903523, 201.9625, 196.295123, 190.890584, 185.738078, 180.826797, 176.145935, 171.684686, 167.432242, 163.377797, 159.510545, 155.819678, 152.29439, 148.923875, 145.697326, 142.603936, 139.632899, 136.773408, 134.014656, 131.345837, 128.756144, 126.234771, 123.77091, 121.353756, 118.677131, 115.884126, 113.121383, 110.387681, 107.681799, 105.002514, 102.348607, 99.7188545, 97.1120368, 94.5269321, 91.9623192, 89.4169768, 86.8896836, 84.3792184, 81.8843598, 79.4038866, 76.9365776, 74.3399733, 71.7571524, 69.1899553, 66.6402225, 64.1097944, 61.6005116, 59.1142146, 56.6527438, 54.2179397, 51.8116428, 49.4356936, 47.0919326, 44.7822002, 42.508337, 40.2721835, 38.07558, 35.920582, 33.8092394, 31.7429498, 29.7235398, 27.7528359, 25.8326647, 23.9648528, 22.1512268, 20.3936131, 18.6938384, 17.0537293, 15.4751122, 13.9598138, 12.5096606, 11.1264791, 9.81209604, 8.56833787, 7.89800236, 7.54189323, 7.2347929, 6.96227046, 6.70989499, 6.46323558, 6.2078613, 5.92934125, 5.61324452, 5.24514017, 4.81059731, 4.29518502, 3.68447237, 2.96402846, 2.11942238, 1.13622319, 0, -6.04220097, -11.7678362, -17.1695595, -22.2400245, -26.971885, -31.3577946, -35.3904072, -39.0623765, -42.366356, -45.2949997, -47.8409612, -49.9968943, -51.7554526, -53.1092899, -54.0510599, -54.5734164, -53.9780781, -52.1589938, -50.6716462, -49.3606316, -48.0705461, -46.6459861, -44.9315478, -42.7718274, -40.0114212, -36.4949253, -32.0669361, -26.5720497, -19.8548625, -11.4427691, -1.61387445, 9.69206469, 22.5352917, 36.1118884, 50.914179, 67.6649304, 86.6836343, 108.289782, 132.802866, 160.542378, 190.795867, 225.228837, 264.36268, 308.71879, 358.81856, 415.240706, 478.434969, 548.908409, 627.168091, 713.721076, 808.997153, 913.675105, 1028.55677, 1154.22776, 1290.83221, 1438.71247, 1598.21089, 1770.01832, 1954.04941, 2149.83867, 2357.48322, 2576.65607, 2807.97348, 3051.16991, 3304.48572, 3566.84709, 3839.02625, 4120.428, 4403.68825, 4685.27179, 4959.61689, 5225.5266, 5480.16714, 5720.70477, 5944.30574, 6092.17566, 6231.43157, 6373.22987, 6528.72698, 6709.07929, 6950.26663, 7222.07305, 7525.01633, 7907.07046, 8349.9685, 8815.67008, 9286.89692, 9746.37075, 10176.8133, 10560.9462, 11531.2933, 12045.4336, 12136.5286, 11837.7395, 11182.2276, 10203.1543, 8933.68093, 7733.36553, 6947.66278},
        {291.741164, 283.010399, 274.584023, 266.454275, 258.613391, 251.05361, 243.76717, 236.74631, 229.983267, 223.47028, 217.199586, 211.163423, 205.354031, 199.763646, 194.384507, 189.208852, 184.228919, 179.436946, 174.825172, 170.385834, 166.11117, 161.99342, 158.024819, 154.197608, 150.504023, 146.936303, 143.486686, 140.147411, 136.910714, 133.768835, 130.714012, 127.738482, 124.834483, 121.994254, 119.039967, 116.052469, 113.114653, 110.224277, 107.379098, 104.576872, 101.815356, 99.092308, 96.4054848, 93.7526433, 91.1315407, 88.539934, 85.9755803, 83.4362367, 80.9196602, 78.423608, 75.9458371, 73.2917449, 70.655618, 68.0393837, 65.4449689, 62.8743007, 60.3293062, 57.8119124, 55.3240463, 52.8676351, 50.4446057, 48.0568852, 45.7064008, 43.3950794, 41.124848, 38.8976339, 36.7153639, 34.590025, 32.5333627, 30.5265562, 28.5708804, 26.6676104, 24.8180209, 23.023387, 21.2849835, 19.6040853, 17.9819674, 16.4199047, 14.9191721, 13.4810445, 12.1067969, 10.7977041, 9.55504102, 8.38008265, 7.73680619, 7.3854031, 7.08050346, 6.80836662, 6.55525191, 6.30741867, 6.05112623, 5.77263394, 5.45820112, 5.09408711, 4.66655125, 4.16185287, 3.56625131, 2.86600591, 2.047376, 1.09662092, 0, -5.67441797, -11.0937827, -16.2440937, -21.1113503, -25.6815519, -29.9406979, -33.8747877, -37.4698207, -40.7117963, -43.5867139, -46.0805729, -48.1793726, -49.8691125, -51.135792, -51.9654105, -52.3439673, -51.8560086, -50.4301835, -48.9605401, -47.3470523, -45.4896938, -43.2884382, -40.6432593, -37.4541307, -33.6210262, -29.0439195, -23.6227842, -17.2575941, -9.84832292, -1.39355285, 8.3415074, 19.4864667, 32.1709339, 45.9422979, 61.2212775, 78.5838502, 98.334291, 120.776875, 146.215876, 174.95557, 206.440025, 242.095525, 282.394645, 327.809961, 378.81405, 434.829367, 497.641139, 567.984471, 646.59447, 734.206241, 832.256634, 939.921326, 1057.45705, 1185.12992, 1323.22516, 1472.01943, 1631.78938, 1802.42045, 1984.66939, 2179.71977, 2390.05077, 2612.5606, 2845.10176, 3086.4229, 3336.79119, 3595.77673, 3865.34932, 4145.7012, 4428.22596, 4709.29422, 4983.49379, 5249.25236, 5503.55764, 5743.39739, 5965.75931, 6110.90003, 6246.72117, 6384.39325, 6535.08679, 6709.97229, 6945.77058, 7211.06832, 7506.08592, 7876.57451, 8304.73188, 8753.7847, 9206.87932, 9647.1621, 10057.7794, 10421.8776, 11330.1104, 11797.8413, 11864.7384, 11570.47, 10954.7042, 10057.1094, 8917.35375, 7704.87883, 6838.11677},
        {298.907735, 290.610145, 282.550015, 274.722244, 267.121727, 259.743361, 252.582043, 245.63267, 238.890138, 232.349345, 226.005186, 219.852559, 213.886361, 208.101487, 202.492835, 197.055302, 191.783784, 186.673178, 181.718381, 176.914289, 172.255799, 167.737808, 163.355213, 159.102911, 154.975797, 150.968769, 147.076724, 143.294558, 139.617168, 136.039451, 132.556303, 129.162622, 125.853303, 122.623244, 119.406981, 116.230861, 113.121955, 110.077119, 107.093209, 104.16708, 101.295587, 98.4755871, 95.7039348, 92.9774858, 90.2930957, 87.64762, 85.0379142, 82.4608339, 79.9132347, 77.391972, 74.8939015, 72.1818534, 69.4917817, 66.8256152, 64.1852826, 61.5727127, 58.9898341, 56.4385757, 53.9208663, 51.4386344, 48.993809, 46.5883187, 44.2240924, 41.9030587, 39.6271464, 37.3982842, 35.218401, 33.102807, 31.0664914, 29.0837842, 27.1557469, 25.2834406, 23.4679267, 21.7102663, 20.0115209, 18.3727516, 16.7950198, 15.2793867, 13.8269136, 12.4386617, 11.1156923, 9.85906681, 8.66984636, 7.54909227, 6.93530103, 6.60161545, 6.31592068, 6.0650824, 5.83596626, 5.61543793, 5.39036306, 5.14760731, 4.87403634, 4.55651582, 4.1819114, 3.73708874, 3.2089135, 2.58425135, 1.84996795, 0.992928941, 0, -4.76866215, -9.418325, -13.9249312, -18.2644233, -22.4127439, -26.3458358, -30.0396414, -33.4701034, -36.6131644, -39.4447671, -41.940854, -44.0773677, -45.830251, -47.1754463, -48.0888964, -48.5465438, -48.1276119, -46.7513692, -45.2910994, -43.6377338, -41.6822037, -39.3154406, -36.4283757, -32.9119405, -28.6570663, -23.5546844, -17.4957263, -10.3711233, -2.07180671, 7.58522351, 18.6079964, 31.083401, 45.0983267, 59.8928059, 75.9771562, 94.0875236, 114.564854, 137.750094, 163.98419, 193.608087, 226.500794, 263.605786, 305.354387, 352.177922, 404.507718, 461.68813, 525.509195, 596.673982, 675.885559, 763.846994, 861.786564, 969.240206, 1086.92242, 1215.22632, 1353.8888, 1502.9414, 1662.41564, 1830.95581, 2010.29596, 2203.68348, 2416.07411, 2641.70374, 2876.42545, 3118.04903, 3367.69987, 3624.98146, 3891.47899, 4167.30974, 4445.32463, 4722.48485, 4996.33237, 5263.15824, 5519.72259, 5762.78555, 5989.10725, 6141.0075, 6283.29685, 6426.34553, 6580.52377, 6756.20178, 6986.94741, 7244.46819, 7529.1925, 7882.95856, 8289.80336, 8716.50982, 9147.97765, 9569.10657, 9964.79627, 10319.9465, 11462.1253, 12008.43, 12047.5601, 11668.2152, 10959.0947, 10008.8982, 8906.32512, 7567.22051, 6792.93053},
        {306.085713, 298.238844, 290.560839, 283.049311, 275.701872, 268.516132, 261.489704, 254.6202, 247.905231, 241.342409, 234.929346, 228.663653, 222.542943, 216.564826, 210.726916, 205.026823, 199.46216, 194.030537, 188.729568, 183.556863, 178.510034, 173.586694, 168.784453, 164.100924, 159.533718, 155.080447, 150.738724, 146.506158, 142.380364, 138.358951, 134.439532, 130.619719, 126.897123, 123.269357, 119.785678, 116.41679, 113.132773, 109.929562, 106.803094, 103.749304, 100.764128, 97.8435006, 94.9833588, 92.1796377, 89.4282732, 86.7252009, 84.0663564, 81.4476756, 78.8650941, 76.3145475, 73.7919716, 71.0153426, 68.2645812, 65.5416489, 62.8485068, 60.1871163, 57.5594388, 54.9674355, 52.4130677, 49.8982968, 47.4250841, 44.9953908, 42.6111783, 40.2744079, 37.9870409, 35.7510386, 33.5683624, 31.4579337, 29.438267, 27.4763443, 25.5730275, 23.7291783, 21.9456587, 20.2233305, 18.5630555, 16.9656956, 15.4321127, 13.9631685, 12.5597249, 11.2226438, 9.95278701, 8.75101637, 7.61819375, 6.55518097, 5.97459845, 5.66276796, 5.40106226, 5.17698085, 4.97802321, 4.79168883, 4.60547721, 4.40688783, 4.18342018, 3.92257374, 3.61184802, 3.2387425, 2.79075667, 2.25539001, 1.62014202, 0.872512188, 0, -3.9132745, -7.82594318, -11.7054748, -15.5193383, -19.2350024, -22.8199358, -26.2416075, -29.4674863, -32.4650408, -35.2017401, -37.6450527, -39.7624477, -41.5213937, -42.8893596, -43.8338142, -44.3222263, -43.8520634, -42.3236511, -40.7675567, -39.0505344, -37.0393384, -34.6007229, -31.6014422, -27.9082505, -23.387902, -17.9071509, -11.3327516, -3.53145812, 5.6299752, 16.6430846, 29.1514522, 43.1808368, 58.7569968, 74.7054687, 91.6521222, 110.542886, 131.763587, 155.700049, 182.738097, 213.263559, 247.679089, 286.34856, 329.654503, 377.979451, 431.705935, 490.21208, 555.135927, 627.111113, 706.77127, 794.750034, 891.875572, 998.349223, 1115.47244, 1243.83779, 1382.59049, 1531.52559, 1690.43815, 1856.87111, 2033.38346, 2224.99106, 2438.66242, 2666.54238, 2903.12316, 3145.48159, 3395.07408, 3651.34672, 3915.06164, 4186.00614, 4459.14212, 4732.161, 5006.22352, 5275.01804, 5535.03504, 5782.76504, 6014.69851, 6175.84101, 6327.03923, 6477.6549, 6637.04977, 6814.58557, 7039.42851, 7287.93288, 7560.97278, 7896.73949, 8281.33575, 8685.31545, 9095.55878, 9498.94591, 9882.35701, 10232.6722, 11666.1275, 12317.2404, 12331.22, 11853.2757, 11028.6166, 10002.4518, 8919.99047, 7448.1105, 6790.95605},
        {313.231616, 305.863933, 298.593985, 291.422184, 284.348943, 277.374674, 270.499791, 263.724705, 257.04983, 250.475578, 244.002362, 237.630594, 231.360687, 225.193054, 219.128108, 213.16626, 207.307924, 201.553513, 195.903438, 190.358113, 184.917951, 179.583363, 174.354763, 169.232563, 164.217176, 159.309015, 154.508492, 149.81602, 145.232011, 140.756878, 136.391035, 132.134892, 127.988864, 123.953363, 120.196088, 116.63029, 113.167209, 109.801829, 106.529133, 103.344105, 100.24173, 97.2169906, 94.2648712, 91.3803556, 88.5584276, 85.7940711, 83.0822699, 80.418008, 77.7962692, 75.2120373, 72.6602962, 69.8089768, 66.9872058, 64.1970569, 61.4406039, 58.7199204, 56.0370803, 53.3941573, 50.7932251, 48.2363573, 45.7256279, 43.2631104, 40.8508786, 38.4910063, 36.1855671, 33.9366348, 31.7462832, 29.6396976, 27.6415083, 25.7061231, 23.8341177, 22.0260675, 20.2825479, 18.6041346, 16.9914029, 15.4449285, 13.9652867, 12.5530532, 11.2088034, 9.93311275, 8.72655684, 7.58971113, 6.52315114, 5.52745237, 4.98417064, 4.69537789, 4.45856696, 4.26194979, 4.09373835, 3.94214459, 3.79538047, 3.64165794, 3.46918897, 3.26618551, 3.02085952, 2.72142296, 2.35608777, 1.91306593, 1.38056938, 0.746810085, 0, -3.28352521, -6.63803509, -10.0257651, -13.4089509, -16.7498277, -20.0106313, -23.1535969, -26.1409603, -28.9349567, -31.4978218, -33.7917911, -35.7790999, -37.4219839, -38.6826785, -39.5234193, -39.9064416, -39.2223991, -37.3518729, -35.549962, -33.6564203, -31.5110013, -28.9534587, -25.8235463, -21.9610176, -17.2056263, -11.3971261, -4.37527063, 4.02018641, 13.9494914, 26.205409, 40.0837438, 55.5549866, 72.5896282, 89.6419473, 107.440541, 127.118329, 149.100666, 173.812907, 201.680405, 233.128516, 269.072692, 309.299029, 354.136993, 403.91605, 458.965665, 518.656194, 584.515991, 657.1143, 737.020366, 824.803433, 920.916152, 1026.18686, 1142.39043, 1270.29915, 1408.63838, 1557.05254, 1715.186, 1879.91707, 2054.2849, 2244.3462, 2458.17485, 2686.83587, 2924.3245, 3167.34329, 3417.18214, 3673.0253, 3934.65355, 4201.40576, 4470.43354, 4740.12574, 5015.11786, 5286.59109, 5550.77197, 5803.88706, 6042.16291, 6213.935, 6375.29372, 6534.43838, 6699.56829, 6878.88277, 7096.02099, 7333.44917, 7592.74954, 7909.6091, 8272.11028, 8654.1252, 9044.44679, 9431.86803, 9805.18184, 10153.1812, 11863.325, 12617.9689, 12614.3443, 12049.6822, 11121.214, 10026.1709, 8961.78407, 7384.64192, 6829.68535},
        {320.057931, 313.251938, 306.46504, 299.700724, 292.962478, 286.253789, 279.578145, 272.939034, 266.339942, 259.784357, 253.275767, 246.817659, 240.413521, 234.06684, 227.781103, 221.559799, 215.406414, 209.324437, 203.317354, 197.388653, 191.541822, 185.780348, 180.107719, 174.527421, 169.042943, 163.657772, 158.375396, 153.199301, 148.132976, 143.179908, 138.343585, 133.627493, 129.035121, 124.569956, 120.531659, 116.769402, 113.128794, 109.603714, 106.188035, 102.875635, 99.6603904, 96.5361762, 93.4968687, 90.5363442, 87.6484787, 84.8271482, 82.0662288, 79.3595967, 76.7011279, 74.0846985, 71.5041845, 68.556925, 65.6419293, 62.7616696, 59.918618, 57.1152469, 54.3540284, 51.6374348, 48.9679382, 46.348011, 43.7801252, 41.2667533, 38.8103673, 36.4134394, 34.078442, 31.8078473, 29.6041274, 27.5101379, 25.5677649, 23.6961929, 21.8952762, 20.1648692, 18.5048262, 16.9150015, 15.3952495, 13.9454245, 12.5653809, 11.2549729, 10.0140549, 8.84248124, 7.74010626, 6.70678428, 5.74236964, 4.84671666, 4.34883967, 4.07708956, 3.855861, 3.67432643, 3.52165833, 3.38702915, 3.25961134, 3.12857737, 2.9830997, 2.81235078, 2.60550308, 2.35172904, 2.04020114, 1.66009182, 1.20057356, 0.650818796, 0, -3.53370193, -7.0523968, -10.5227267, -13.9113339, -17.1848605, -20.3099487, -23.2532406, -25.9813784, -28.4610043, -30.6587604, -32.5412888, -34.0752318, -35.2272315, -35.9639301, -36.2519697, -36.0579925, -34.8086024, -32.3932932, -30.0002222, -27.4803091, -24.6844735, -21.4636352, -17.6687138, -13.1506289, -7.76030038, -1.34864776, 6.23340924, 15.1349509, 25.5050576, 37.971694, 52.0294651, 67.683786, 84.9400716, 102.436979, 120.863303, 141.272305, 164.079428, 189.700114, 218.549806, 251.043946, 288.332223, 329.872366, 375.936161, 426.795393, 482.721849, 542.459694, 608.190238, 680.56717, 760.244179, 847.874957, 944.272048, 1049.73613, 1165.58362, 1292.4477, 1429.56665, 1576.80501, 1734.02737, 1899.05288, 2074.25636, 2264.24391, 2475.35286, 2700.59028, 2935.60908, 3177.80283, 3427.51459, 3683.73355, 3945.80303, 4212.80403, 4482.51913, 4753.46538, 5031.09071, 5305.91499, 5574.05611, 5831.63194, 6074.76038, 6256.52369, 6426.83429, 6592.56898, 6760.60455, 6937.81779, 7140.36886, 7359.66229, 7598.59627, 7895.24663, 8238.5871, 8602.93408, 8977.9942, 9353.47406, 9719.08026, 10064.5194, 11856.3991, 12643.0937, 12640.2905, 12063.6771, 11128.941, 10051.7695, 9047.85031, 7497.72046, 6901.99489},
        {326.820878, 320.617041, 314.350331, 308.027421, 301.65498, 295.23968, 288.788193, 282.307188, 275.803337, 269.283311, 262.753781, 256.221418, 249.692893, 243.174876, 236.67404, 230.197054, 223.750591, 217.34132, 210.975913, 204.661041, 198.403375, 192.209585, 186.086344, 180.040321, 174.078188, 168.206616, 162.432276, 156.761838, 151.201974, 145.759355, 140.440652, 135.252536, 130.201677, 125.294747, 120.967879, 117.003979, 113.180793, 109.491054, 105.927498, 102.482857, 99.1498667, 95.9212606, 92.789773, 89.748138, 86.7890898, 83.9053624, 81.08969, 78.3348068, 75.6334468, 72.9783443, 70.3622333, 67.3037021, 64.2799093, 61.2938676, 58.3485896, 55.4470881, 52.5923757, 49.7874651, 47.0353691, 44.3391002, 41.7016713, 39.126095, 36.6153839, 34.1725508, 31.8006084, 29.5025693, 27.2814463, 25.2036896, 23.3342266, 21.5452345, 19.8356136, 18.2042642, 16.6500865, 15.1719809, 13.7688476, 12.4395869, 11.1830991, 9.99828447, 8.88404331, 7.83927591, 6.86288253, 5.95376348, 5.11081902, 4.33294945, 3.8844753, 3.62600397, 3.41495272, 3.24160843, 3.09625799, 2.9691883, 2.85068622, 2.73103866, 2.6005325, 2.44945462, 2.26809192, 2.04673127, 1.77565958, 1.44516371, 1.04553057, 0.567047039, 0, -4.25077986, -8.31409646, -12.1664724, -15.7844302, -19.1444925, -22.2231819, -24.9970209, -27.4425321, -29.5362382, -31.2546616, -32.574325, -33.471751, -33.9234621, -33.9059809, -33.39583, -32.369532, -30.3407695, -27.2199449, -24.030481, -20.6497204, -16.9550055, -12.8236791, -8.13308367, -2.76056175, 3.41654405, 10.5208912, 18.675137, 28.0019391, 38.6239547, 50.8025368, 64.4707924, 79.7097702, 96.600519, 114.227484, 133.170016, 154.273226, 177.917144, 204.481801, 234.347225, 267.893447, 306.332766, 348.959643, 395.991273, 447.64485, 504.13757, 563.335667, 628.395037, 700.120612, 779.317329, 866.790122, 964.055451, 1070.33708, 1186.02533, 1311.45742, 1446.86216, 1592.51706, 1748.69959, 1915.0014, 2092.5417, 2283.1879, 2490.08609, 2709.71822, 2940.46638, 3180.95295, 3430.20752, 3687.07279, 3950.76055, 4220.20918, 4493.00364, 4767.49446, 5048.5708, 5327.03861, 5598.98506, 5860.49729, 6107.66246, 6298.82581, 6477.2519, 6648.46337, 6817.98284, 6991.33297, 7176.53269, 7374.94413, 7591.02009, 7867.46556, 8193.59183, 8542.77179, 8905.11366, 9270.72566, 9629.716, 9972.19292, 11712.1234, 12484.308, 12498.3744, 11963.9503, 11090.6637, 10088.1423, 9166.01377, 7721.50665, 7011.04519},
        {333.727898, 328.133263, 322.393822, 316.519407, 310.519845, 304.404967, 298.184603, 291.868581, 285.466732, 278.988885, 272.444869, 265.844515, 259.197652, 252.51411, 245.803717, 239.076304, 232.341701, 225.609736, 218.89024, 212.193042, 205.527972, 198.904859, 192.333533, 185.823823, 179.38556, 173.028572, 166.762689, 160.597741, 154.543558, 148.609969, 142.806804, 137.143892, 131.631063, 126.278146, 121.654932, 117.479493, 113.463168, 109.597616, 105.874495, 102.285462, 98.8221749, 95.4762918, 92.2394703, 89.1033683, 86.0596436, 83.0999539, 80.215957, 77.3993107, 74.6416728, 71.9347011, 69.2700534, 66.0893064, 62.9461237, 59.8440876, 56.7867807, 53.7777854, 50.8206843, 47.9190597, 45.0764942, 42.2965701, 39.5828701, 36.9389764, 34.3684717, 31.8749384, 29.4619589, 27.1331158, 24.8919914, 22.8303505, 21.0378406, 19.336175, 17.7232194, 16.19684, 14.7549028, 13.3952736, 12.1158184, 10.9144033, 9.78889418, 8.73715702, 7.7570578, 6.84646249, 6.00323706, 5.22524748, 4.51035973, 3.85643977, 3.45798901, 3.21041975, 3.00678086, 2.83855713, 2.69723334, 2.57429424, 2.46122463, 2.34950926, 2.23063292, 2.09608038, 1.93733642, 1.7458858, 1.5132133, 1.2308037, 0.890141764, 0.482712274, 0, -5.11748461, -9.84338483, -14.1663591, -18.0750657, -21.5581632, -24.6043098, -27.2021641, -29.3403844, -31.007629, -32.1925565, -32.8838252, -33.0700935, -32.7400197, -31.8822624, -30.4854799, -28.5383305, -25.6243445, -21.6743047, -17.5932311, -13.2829689, -8.64536338, -3.58225964, 2.0044971, 8.21306163, 15.1415888, 22.8882333, 31.5511501, 41.2284938, 52.0184194, 63.8132171, 76.9923898, 91.7158518, 108.143517, 125.845166, 145.275779, 167.047707, 191.497905, 218.963327, 249.780929, 284.287663, 323.660134, 367.140102, 414.900243, 467.113232, 523.951746, 582.434361, 646.676377, 717.638993, 796.28341, 883.570828, 981.773139, 1088.9389, 1204.41383, 1328.20294, 1461.65723, 1605.52337, 1760.54803, 1928.32724, 2108.56528, 2300.03994, 2502.34058, 2715.78182, 2941.67625, 3179.99603, 3428.44199, 3685.75754, 3951.17529, 4223.56551, 4500.00457, 4778.58359, 5063.21109, 5345.26575, 5620.82493, 5885.96599, 6136.76629, 6336.59975, 6522.42306, 6698.48943, 6869.05211, 7038.3643, 7206.74145, 7384.99978, 7579.08017, 7836.47886, 8146.68771, 8481.88382, 8832.42479, 9188.66819, 9540.97163, 9879.69269, 11473.5791, 12201.0794, 12246.875, 11795.6473, 11032.0779, 10140.8482, 9306.63961, 8006.91421, 7159.07415},
        {341.193481, 336.192556, 330.964924, 325.523594, 319.881571, 314.051862, 308.047473, 301.881413, 295.566686, 289.116299, 282.54326, 275.860574, 269.081249, 262.218291, 255.284706, 248.293501, 241.257683, 234.190259, 227.104234, 220.012616, 212.92841, 205.864625, 198.834266, 191.850339, 184.925853, 178.073812, 171.307224, 164.639095, 158.082432, 151.650242, 145.355531, 139.211305, 133.230572, 127.426337, 122.504137, 118.116146, 113.906217, 109.864884, 105.982677, 102.250132, 98.6577792, 95.1961525, 91.8557843, 88.6272075, 85.5009546, 82.4675585, 79.5175518, 76.6414674, 73.8298378, 71.0731959, 68.3620743, 65.0593595, 61.7968366, 58.5786445, 55.4089221, 52.2918083, 49.2314419, 46.2319618, 43.2975069, 40.432216, 37.640228, 34.9256817, 32.2927162, 29.7454701, 27.2880824, 24.924692, 22.6594376, 20.6083578, 18.8848044, 17.2621317, 15.7372245, 14.3069679, 12.9682466, 11.7179456, 10.5529496, 9.47014367, 8.46641257, 7.53864121, 6.68371445, 5.89851719, 5.17993428, 4.52485061, 3.93015105, 3.39272048, 3.03548317, 2.78965322, 2.58492986, 2.41410773, 2.26998148, 2.14534578, 2.03299527, 1.92572462, 1.81632847, 1.69760147, 1.5623383, 1.40333359, 1.21338201, 0.985278203, 0.71181683, 0.385792544, 0, -5.77129366, -10.9739948, -15.6079063, -19.6728313, -23.1685725, -26.094933, -28.4517158, -30.2387237, -31.4557598, -32.1026269, -32.1791281, -31.6850663, -30.6202444, -28.9844655, -26.7775323, -23.999248, -20.1660902, -15.2088154, -10.2204622, -5.0972639, 0.264546265, 5.96873503, 12.1190691, 18.8193153, 26.1732404, 34.284611, 43.2571939, 53.1947559, 64.2010637, 76.1096908, 89.3936681, 104.23782, 120.826972, 138.931524, 158.943514, 181.365491, 206.506606, 234.676011, 266.182859, 301.336301, 341.219373, 385.131812, 433.23136, 485.675754, 542.622734, 600.637193, 664.36793, 734.870897, 813.202044, 900.417325, 999.315151, 1107.07934, 1222.20373, 1344.42703, 1476.02953, 1618.1504, 1771.92879, 1940.5223, 2122.59294, 2314.60084, 2513.62377, 2722.5836, 2944.65524, 3180.70904, 3427.71033, 3684.41692, 3950.18324, 4223.92178, 4502.35731, 4783.45208, 5070.89596, 5356.04243, 5634.87111, 5903.36163, 6157.49361, 6363.53225, 6554.6002, 6734.1057, 6905.45699, 7072.06229, 7229.43701, 7394.14409, 7572.9744, 7815.29535, 8111.0446, 8432.41957, 8770.24496, 9115.34542, 9458.54563, 9790.67027, 11092.0215, 11734.0575, 11842.7534, 11544.0837, 10964.0233, 10228.5471, 9463.62979, 8308.92024, 7334.43786},
        {349.495856, 345.05152, 340.299656, 335.256447, 329.938074, 324.36072, 318.540565, 312.493793, 306.236584, 299.785121, 293.155585, 286.364159, 279.427023, 272.360361, 265.180354, 257.903184, 250.545032, 243.12208, 235.650511, 228.146506, 220.626247, 213.105915, 205.601694, 198.129764, 190.706307, 183.347506, 176.069542, 168.888596, 161.820852, 154.88249, 148.089693, 141.458642, 135.005519, 128.746506, 123.525122, 118.926478, 114.525742, 110.312246, 106.275325, 102.404312, 98.6885389, 95.1173397, 91.6800474, 88.3659952, 85.1645164, 82.0649441, 79.0566115, 76.1288518, 73.2709982, 70.4723838, 67.7223419, 64.3046735, 60.9295385, 57.6015646, 54.3253796, 51.1056111, 47.946887, 44.853835, 41.8310828, 38.8832581, 36.0149887, 33.2309023, 30.5356267, 27.9337896, 25.4300188, 23.0289419, 20.7351868, 18.6854816, 17.0133828, 15.451071, 13.9946103, 12.6400648, 11.3834989, 10.2209767, 9.14856238, 8.1623202, 7.25831434, 6.432609, 5.6812684, 5.00035672, 4.38593819, 3.834077, 3.34083735, 2.90228345, 2.57149898, 2.31483908, 2.09797355, 1.91511621, 1.76048086, 1.62828133, 1.51273141, 1.40804493, 1.30843569, 1.20811751, 1.10130419, 0.982209557, 0.845047414, 0.684031575, 0.493375852, 0.267294056, 0, -5.94089225, -11.2083525, -15.8097884, -19.7526076, -23.0442179, -25.6920269, -27.7034423, -29.0858718, -29.8467231, -29.9934039, -29.5333219, -28.4738847, -26.8225, -24.5865756, -21.7735191, -18.3907382, -13.6729344, -7.51712874, -1.6785168, 4.00107361, 9.67981472, 15.5158787, 21.6674379, 28.2926644, 35.5497305, 43.5968083, 52.5920702, 62.6936883, 74.0598349, 86.9863264, 101.443221, 117.547398, 135.415737, 154.512501, 175.280878, 198.338083, 223.996779, 252.56963, 284.369298, 319.708446, 359.550896, 403.359974, 451.320942, 503.619062, 560.439597, 618.550454, 682.408588, 753.053602, 831.525094, 918.862668, 1017.94741, 1125.72673, 1240.28194, 1361.23269, 1491.33938, 1631.9523, 1784.4217, 1952.51171, 2134.61015, 2326.47159, 2524.74598, 2732.57967, 2953.0208, 3186.9697, 3431.68741, 3686.10544, 3949.76068, 4221.74159, 4498.91705, 4779.41146, 5068.28065, 5355.65123, 5637.24827, 5908.79685, 6166.02203, 6374.71067, 6568.01061, 6749.13147, 6921.28287, 7087.67445, 7243.58723, 7405.44514, 7579.85579, 7812.96005, 8095.71693, 8402.6132, 8725.43147, 9055.95433, 9385.96439, 9707.24424, 10546.4169, 11058.6158, 11270.0122, 11206.777, 10895.0815, 10361.0969, 9630.9943, 8592.61333, 7531.60126},
        {358.020359, 354.079849, 349.75998, 345.079673, 340.057848, 334.713422, 329.065318, 323.132453, 316.933747, 310.48812, 303.814492, 296.931782, 289.858909, 282.614794, 275.218355, 267.688513, 260.044187, 252.304296, 244.48776, 236.613499, 228.700432, 220.767479, 212.833559, 204.917592, 197.038497, 189.215194, 181.466603, 173.811643, 166.269234, 158.858295, 151.597746, 144.506506, 137.603495, 130.907633, 125.371709, 120.528066, 115.89782, 111.469585, 107.231975, 103.173603, 99.2830832, 95.5490293, 91.9600547, 88.5047733, 85.1717986, 81.9497443, 78.8272241, 75.7928516, 72.8352406, 69.9430047, 67.1047575, 63.5569172, 60.0565991, 56.608723, 53.2182087, 49.889976, 46.6289448, 43.4400348, 40.3281657, 37.2982575, 34.3552299, 31.5040027, 28.7494956, 26.0966286, 23.5503213, 21.1154937, 18.7970654, 16.7477487, 15.1167075, 13.6040507, 12.2051172, 10.9152462, 9.72977667, 8.64404753, 7.65339785, 6.75316664, 5.9386929, 5.20531565, 4.54837391, 3.96320669, 3.44515299, 2.98955184, 2.59174225, 2.24706322, 1.94482223, 1.68350092, 1.46165345, 1.27481456, 1.11851899, 0.988301502, 0.879696835, 0.78823974, 0.709464964, 0.638907255, 0.572101361, 0.50458203, 0.431884011, 0.34954205, 0.253090896, 0.138065297, 0, -5.95261228, -11.1542613, -15.6184454, -19.358663, -22.3884126, -24.7211926, -26.3705015, -27.3498375, -27.6726993, -27.3525852, -26.4029936, -24.8374229, -22.6693716, -19.9123382, -16.5798209, -12.6853183, -7.16340176, 0.126562851, 6.73064673, 12.8665501, 18.7519732, 24.6046164, 30.6421797, 37.0823635, 44.1428679, 52.0413933, 60.9956399, 71.2233078, 82.9420974, 96.9720799, 112.707715, 130.185992, 149.443899, 169.580538, 191.10184, 214.763844, 240.884904, 269.783376, 301.777615, 337.185978, 376.805826, 420.330723, 467.985307, 519.994213, 576.582078, 634.977636, 699.150404, 770.073992, 848.722014, 936.068082, 1034.8463, 1142.11808, 1255.81538, 1375.5306, 1504.24622, 1643.42263, 1794.52023, 1961.50676, 2142.76543, 2333.94413, 2531.94984, 2739.40365, 2958.60163, 3190.211, 3432.13904, 3683.55987, 3944.28503, 4213.65389, 4488.66873, 4767.6538, 5057.07182, 5345.85654, 5629.51508, 5903.55455, 6163.48206, 6374.40035, 6569.82201, 6752.85524, 6926.60826, 7094.18928, 7251.6065, 7413.80147, 7586.92526, 7812.54488, 8082.79258, 8375.04399, 8682.16919, 8997.03828, 9312.52134, 9621.48847, 9997.3111, 10377.6569, 10689.8463, 10861.2, 10819.0383, 10490.6819, 9803.45114, 8889.33854, 7785.05835},
        {366.595695, 363.087632, 359.141874, 354.779412, 350.021236, 344.888337, 339.401703, 333.582327, 327.451198, 321.029306, 314.337642, 307.397197, 300.22896, 292.853923, 285.293074, 277.567405, 269.697907, 261.705568, 253.61138, 245.436334, 237.201418, 228.927625, 220.635943, 212.347364, 204.082878, 195.863474, 187.710144, 179.643878, 171.685666, 163.856499, 156.177366, 148.669259, 141.353167, 134.250081, 128.377844, 123.238053, 118.320323, 113.613293, 109.105606, 104.7859, 100.642817, 96.6649976, 92.8410818, 89.1597104, 85.6095238, 82.1791626, 78.8572674, 75.6324787, 72.493437, 69.428783, 66.427157, 62.726093, 59.0816208, 55.4986633, 51.9821437, 48.536985, 45.1681103, 41.8804426, 38.678905, 35.5684206, 32.5539123, 29.6403034, 26.8325168, 24.1354757, 21.5541029, 19.0933218, 16.7580552, 14.7104661, 13.1089422, 11.6341112, 10.2807028, 9.04344686, 7.9170733, 6.89631192, 5.97589256, 5.15054507, 4.41499929, 3.76398506, 3.19223221, 2.6944706, 2.26543006, 1.89984043, 1.59243156, 1.33793328, 1.07185529, 0.819783106, 0.60801406, 0.433199796, 0.291991963, 0.181042212, 0.0970021915, 0.0365235505, -0.00374206147, -0.0271429952, -0.0370276014, -0.0367442307, -0.029641234, -0.0190669618, -0.00836976491, -0.000897994084, 0, -5.83602233, -10.870646, -15.1217005, -18.6070155, -21.3444205, -23.3517451, -24.6468188, -25.2474712, -25.1715319, -24.4368305, -23.0611964, -21.0624593, -18.4584488, -15.2669944, -11.5059256, -7.19307211, -1.00473934, 7.23288939, 14.4675849, 20.968987, 27.0067357, 32.8504708, 38.7698321, 45.0344596, 51.9139931, 59.6780725, 68.5963376, 78.9384284, 90.9739846, 105.982869, 122.854084, 141.554202, 162.049796, 183.130923, 205.352415, 229.582839, 256.141725, 285.3486, 317.522991, 352.984428, 392.321693, 435.503112, 482.795531, 534.465797, 590.780758, 649.61742, 714.229931, 785.4826, 864.239733, 951.36564, 1049.21885, 1155.34318, 1267.81815, 1386.31319, 1513.74373, 1651.56781, 1801.24348, 1966.56286, 2146.11944, 2335.96043, 2533.74609, 2741.06998, 2959.08016, 3188.0454, 3426.74453, 3674.64028, 3931.90695, 4198.19176, 4470.53287, 4747.44425, 5036.46782, 5325.70074, 5610.53193, 5886.35032, 6148.54485, 6362.30928, 6560.7765, 6746.88423, 6923.57021, 7093.77215, 7254.19783, 7418.16822, 7591.29102, 7810.01877, 8067.74975, 8345.03023, 8635.90106, 8934.40315, 9234.57735, 9530.46456, 9515.08587, 9781.10292, 10175.4655, 10545.1234, 10737.0264, 10598.1242, 9975.36676, 9197.54105, 8109.42657},
        {374.999505, 371.831951, 368.187213, 364.087409, 359.554657, 354.611072, 349.278772, 343.579875, 337.536498, 331.170757, 324.50477, 317.560654, 310.360526, 302.926504, 295.280704, 287.445244, 279.442241, 271.293812, 263.022074, 254.649145, 246.197141, 237.68818, 229.144378, 220.587854, 212.040724, 203.525105, 195.063114, 186.67687, 178.388488, 170.220086, 162.193781, 154.331691, 146.655932, 139.188622, 132.949278, 127.440338, 122.152252, 117.074771, 112.197642, 107.510617, 103.003444, 98.6658736, 94.4876549, 90.4585375, 86.5682712, 82.8066056, 79.1632901, 75.6280744, 72.1907082, 68.8409409, 65.5685223, 61.6821406, 57.8673708, 54.1287266, 50.4707216, 46.8978695, 43.4146841, 40.0256789, 36.7353676, 33.548264, 30.4688817, 27.5017343, 24.6513356, 21.9221992, 19.3188389, 16.8457682, 14.5075008, 12.4666275, 10.8819432, 9.43194008, 8.11088419, 6.91304176, 5.83267898, 4.86406203, 4.00145712, 3.23913043, 2.57134817, 1.99237651, 1.49648166, 1.0779298, 0.730987132, 0.449919845, 0.228994132, 0.0624761858, -0.152737062, -0.371646758, -0.548083743, -0.684622005, -0.783835532, -0.848298311, -0.88058433, -0.883267576, -0.858922038, -0.810121703, -0.739440559, -0.649452592, -0.542731792, -0.421852144, -0.289387638, -0.147912261, 0, -5.6478044, -10.4671284, -14.4783046, -17.7016655, -20.1575437, -21.8662716, -22.848182, -23.1236072, -22.7128798, -21.6363324, -19.9142975, -17.5671077, -14.6150955, -11.0785935, -6.9779341, -2.33344998, 4.32171411, 13.179681, 20.8562411, 27.649745, 33.8585432, 39.7809863, 45.7154247, 51.9602091, 58.81369, 66.574218, 75.5401435, 86.0098172, 98.2815895, 113.870786, 131.412557, 150.840163, 172.086861, 193.842834, 216.662877, 241.433274, 268.460205, 298.049852, 330.508396, 366.142017, 405.27478, 448.18954, 495.188978, 546.575776, 602.652615, 662.078182, 727.210153, 798.76221, 877.448033, 963.981303, 1060.08976, 1164.23362, 1275.00957, 1392.29512, 1518.58289, 1655.19136, 1803.43902, 1966.55601, 2143.5147, 2331.20202, 2528.35113, 2735.26625, 2951.77068, 3177.66701, 3412.72284, 3656.72184, 3910.31271, 4173.50338, 4443.12997, 4717.82275, 5005.38934, 5293.87077, 5578.72057, 5855.39224, 6119.33929, 6337.82552, 6541.54158, 6732.9884, 6914.66693, 7089.0781, 7252.46061, 7417.75247, 7590.13843, 7801.28156, 8045.88562, 8307.62173, 8581.74525, 8863.51159, 9148.17611, 9430.9942, 9192.62188, 9387.26134, 9823.12019, 10308.4061, 10651.3265, 10660.0893, 10142.9019, 9520.079, 8531.98624},
        {382.516148, 379.557799, 376.115205, 372.20993, 367.863535, 363.097586, 357.933643, 352.39327, 346.498031, 340.269488, 333.729203, 326.89874, 319.799662, 312.453532, 304.881912, 297.106366, 289.148456, 281.029746, 272.771798, 264.396175, 255.92444, 247.378156, 238.778887, 230.148194, 221.507641, 212.878791, 204.283206, 195.74245, 187.278086, 178.911676, 170.664784, 162.558972, 154.615803, 146.85684, 140.185429, 134.163694, 128.343162, 122.71678, 117.277494, 112.018254, 106.932005, 102.011695, 97.2502717, 92.6406825, 88.1758748, 83.8487957, 79.6523929, 75.5796135, 71.623405, 67.7767147, 64.0324901, 59.9079717, 55.882126, 51.9582129, 48.1394921, 44.4292234, 40.8306666, 37.3470814, 33.9817276, 30.737865, 27.6187533, 24.6276523, 21.7678218, 19.0425216, 16.4550114, 14.008551, 11.7064001, 9.69366921, 8.11206287, 6.67228461, 5.36839859, 4.19446899, 3.14455997, 2.21273571, 1.39306038, 0.679598146, 0.0664131825, -0.45243034, -0.88286825, -1.23083638, -1.50227055, -1.70310659, -1.83928034, -1.91672762, -2.04428599, -2.17427601, -2.25778146, -2.29739879, -2.29572448, -2.25535501, -2.17888683, -2.06891643, -1.92804026, -1.75885481, -1.56395653, -1.34594191, -1.10740741, -0.850949494, -0.579164642, -0.294649321, 0, -5.70656944, -10.5430877, -14.5322038, -17.6965667, -20.0588257, -21.6416297, -22.4676278, -22.559469, -21.9398024, -20.6312771, -18.656542, -16.0382464, -12.7990392, -8.96156953, -4.54848639, 0.417561145, 7.23243831, 16.0618552, 23.9008091, 31.0091896, 37.6468864, 44.0737893, 50.549788, 57.3347723, 64.6886318, 72.8712562, 82.1425353, 92.7623589, 104.990617, 119.871331, 136.592744, 155.179505, 175.656263, 197.34165, 220.613325, 246.037218, 273.849784, 304.287478, 337.586756, 373.984073, 413.283613, 456.285663, 503.311252, 554.68141, 610.717168, 671.095953, 736.943296, 808.74113, 886.971382, 972.115985, 1064.72191, 1165.12655, 1273.26128, 1389.50832, 1515.17036, 1651.1368, 1798.29706, 1958.62913, 2131.68644, 2315.83485, 2511.13781, 2716.52241, 2930.43077, 3152.22852, 3382.84601, 3622.49536, 3872.70207, 4134.01879, 4402.1823, 4675.65349, 4960.07428, 5245.61977, 5528.26513, 5803.98552, 6068.75611, 6297.24659, 6512.06395, 6714.50972, 6905.88542, 7087.49258, 7251.32339, 7414.19492, 7581.39841, 7781.56553, 8010.74859, 8255.27471, 8511.68248, 8776.51045, 9046.29721, 9317.58131, 9340.14893, 9588.6543, 9951.64895, 10317.6844, 10575.3121, 10613.0836, 10319.5505, 9902.44143, 9202.34715},
        {390.157831, 387.303882, 383.975824, 380.193859, 375.97819, 371.349018, 366.326545, 360.930975, 355.182507, 349.101345, 342.707691, 336.021747, 329.063715, 321.853796, 314.412193, 306.759109, 298.914744, 290.899301, 282.732983, 274.435991, 266.028527, 257.530793, 248.962991, 240.345324, 231.697994, 223.041202, 214.395151, 205.780042, 197.216078, 188.723461, 180.322392, 172.033074, 163.87571, 155.8705, 148.755973, 142.178059, 135.774054, 129.54085, 123.475337, 117.574406, 111.834949, 106.253856, 100.828019, 95.554328, 90.4296746, 85.4509495, 80.6150439, 75.9188488, 71.3592551, 66.9331538, 62.637436, 58.2421575, 53.9759618, 49.8406573, 45.8380523, 41.9699551, 38.2381741, 34.6445176, 31.1907941, 27.8788118, 24.7103791, 21.6873043, 18.8113958, 16.0844619, 13.508311, 11.0847514, 8.81559144, 6.82406946, 5.23050819, 3.78627589, 2.4853091, 1.32154435, 0.288918178, -0.618632868, -1.40717225, -2.08276343, -2.65146987, -3.11935504, -3.49248238, -3.77691536, -3.97871745, -4.1039521, -4.15868277, -4.14897292, -4.18256058, -4.21253333, -4.19075673, -4.11999469, -4.00301114, -3.84257002, -3.64143523, -3.40237072, -3.1281404, -2.82150821, -2.48523806, -2.12209388, -1.7348396, -1.32623915, -0.89905645, -0.456055424, 0, -5.41345824, -9.9824082, -13.7272801, -16.6685041, -18.8265104, -20.2217292, -20.8745907, -20.8055251, -20.0349627, -18.5833335, -16.4710678, -13.7185958, -10.3463477, -6.37475375, -1.82424404, 3.28475119, 10.004573, 18.4680443, 26.2282284, 33.4945745, 40.4765322, 47.3835507, 54.4250795, 61.8105678, 69.7494651, 78.4512207, 88.125284, 98.9811044, 111.228131, 125.379765, 141.230055, 158.897266, 178.49966, 199.94942, 223.467851, 249.331213, 277.719595, 308.813087, 342.791777, 379.835753, 419.423136, 462.649627, 509.832656, 561.289653, 617.338049, 678.462372, 744.771181, 816.540133, 894.044884, 977.561092, 1066.5493, 1163.09653, 1268.4196, 1383.3818, 1508.12446, 1643.11306, 1788.81308, 1945.99571, 2114.75123, 2294.83643, 2487.43916, 2690.42881, 2901.27617, 3119.03872, 3345.46259, 3581.05979, 3827.98978, 4087.1917, 4353.56391, 4625.4218, 4906.15751, 5188.22858, 5468.19302, 5742.60885, 6008.03406, 6247.69545, 6474.81507, 6689.28375, 6890.99232, 7079.83161, 7242.3021, 7400.61185, 7560.39041, 7748.44126, 7962.01259, 8189.53011, 8428.68313, 8677.16097, 8932.65292, 9192.84832, 9506.57833, 9817.29985, 10100.2218, 10330.5531, 10483.5027, 10534.2795, 10458.0924, 10240.7921, 9841.63017},
        {398.27571, 395.395684, 392.068618, 388.312617, 384.145785, 379.586228, 374.652051, 369.361359, 363.732256, 357.782848, 351.53124, 344.995537, 338.193843, 331.144264, 323.864906, 316.373872, 308.689268, 300.829199, 292.81177, 284.655086, 276.377251, 267.996372, 259.530553, 250.997899, 242.416515, 233.804506, 225.179977, 216.561033, 207.965779, 199.412321, 190.918762, 182.503209, 174.183766, 165.978538, 158.421235, 151.261315, 144.242038, 137.364776, 130.630903, 124.04179, 117.598812, 111.303339, 105.156745, 99.1604029, 93.3156849, 87.6239636, 82.0866118, 76.705002, 71.4805071, 66.4144995, 61.508352, 56.8150168, 52.2831688, 47.9130626, 43.7049525, 39.6590932, 35.775739, 32.0551445, 28.4975643, 25.1032527, 21.8724644, 18.8054538, 15.9024753, 13.1637836, 10.5896331, 8.18027832, 5.93597374, 3.95577313, 2.33635811, 0.874217137, -0.436800042, -1.60284368, -2.63006402, -3.52461132, -4.29263584, -4.94028781, -5.47371751, -5.89907517, -6.22251105, -6.45017541, -6.58821848, -6.64279054, -6.62004182, -6.52612258, -6.4686354, -6.40087047, -6.27489856, -6.09357761, -5.85976555, -5.57632031, -5.2460998, -4.87196197, -4.45676475, -4.00336605, -3.51462381, -2.99339596, -2.44254042, -1.86491513, -1.26337801, -0.640786989, 0, -4.43135683, -8.16377666, -11.2076838, -13.5735024, -15.271657, -16.3125717, -16.7066709, -16.4643789, -15.5961199, -14.1123183, -12.0233983, -9.33978431, -6.07190053, -2.23017129, 2.17497912, 7.13312639, 13.4179694, 21.1311016, 28.4884865, 35.6477263, 42.766423, 50.0021789, 57.512596, 65.4552763, 73.9878221, 83.2678354, 93.4529183, 104.700673, 117.168701, 130.939915, 146.273994, 163.350946, 182.350781, 203.501836, 226.959957, 252.868101, 281.391779, 312.696501, 346.94778, 384.311125, 424.393016, 468.088135, 515.67137, 567.417606, 623.60173, 684.975764, 751.218176, 822.484568, 898.930545, 980.711707, 1066.6131, 1159.83602, 1262.45192, 1375.64146, 1498.76656, 1632.00573, 1775.53745, 1929.36349, 2093.87925, 2269.67291, 2458.38784, 2657.66946, 2864.96563, 3079.32745, 3302.52251, 3535.07148, 3779.16921, 4035.77039, 4299.69117, 4569.22, 4846.06254, 5124.56434, 5401.83101, 5674.96812, 5941.08128, 6192.40812, 6432.13918, 6658.59705, 6870.10432, 7064.98357, 7223.01557, 7373.42593, 7522.48373, 7696.95, 7894.98241, 8106.20034, 8329.18836, 8562.53109, 8804.8131, 9054.619, 9555.73112, 9903.24636, 10131.4404, 10274.5889, 10366.9676, 10442.852, 10536.5179, 10493.6484, 10349.5618},
        {407.835532, 404.672931, 401.114878, 397.17643, 392.872642, 388.218569, 383.229267, 377.919792, 372.305197, 366.40054, 360.220875, 353.781259, 347.096745, 340.18239, 333.053249, 325.724377, 318.210831, 310.527664, 302.689934, 294.712694, 286.611001, 278.39991, 270.094476, 261.709756, 253.260803, 244.762674, 236.230424, 227.679108, 219.123782, 210.579502, 202.061322, 193.584298, 185.163486, 176.813941, 168.816426, 161.047555, 153.380746, 145.822432, 138.379045, 131.057017, 123.862782, 116.802772, 109.88342, 103.111158, 96.4924197, 90.0336369, 83.7412425, 77.621669, 71.6813493, 65.926716, 60.3642016, 55.3690052, 50.570799, 45.9680214, 41.5591111, 37.3425065, 33.3166461, 29.4799685, 25.8309123, 22.3679158, 19.0894177, 15.9938564, 13.0796705, 10.3452985, 7.7891789, 5.40975026, 3.20545108, 1.24573262, -0.401658862, -1.88474328, -3.20968557, -4.38265069, -5.40980358, -6.29730919, -7.05133246, -7.67803834, -8.18359178, -8.57415771, -8.85590109, -9.03498687, -9.11757998, -9.10984537, -9.017948, -8.8480528, -8.72500428, -8.59313133, -8.39533644, -8.13393315, -7.81123505, -7.42955569, -6.99120865, -6.49850748, -5.95376576, -5.35929705, -4.71741492, -4.03043294, -3.30066467, -2.53042367, -1.72202351, -0.877777769, 0, -1.68716476, -3.11445249, -4.26362962, -5.11646258, -5.65471779, -5.86016169, -5.71456071, -5.19968127, -4.29728981, -2.98915276, -1.25703655, 0.917292398, 3.55206765, 6.66552277, 10.2758913, 14.4014069, 19.532754, 25.7556592, 32.019717, 38.4444005, 45.1491827, 52.2535367, 59.8769356, 68.1388523, 77.1587601, 87.0561318, 97.9504407, 109.96116, 123.207762, 137.748476, 153.786475, 171.459607, 190.905717, 211.83145, 234.590254, 259.650563, 287.279583, 317.744522, 351.312587, 388.250984, 429.513337, 474.471527, 523.258461, 576.007048, 632.850195, 693.3277, 758.313858, 828.089855, 902.936876, 983.136106, 1068.00317, 1159.96494, 1260.58547, 1370.86065, 1490.62647, 1620.23967, 1760.05701, 1910.75865, 2072.30446, 2244.3015, 2426.87093, 2619.76561, 2822.96087, 3036.31717, 3259.50034, 3492.26555, 3735.24097, 3988.4081, 4248.54744, 4514.25016, 4787.31642, 5062.6347, 5337.68535, 5609.94874, 5876.90522, 6139.64017, 6391.12767, 6627.94683, 6846.67674, 7043.89651, 7190.76966, 7326.23459, 7457.7628, 7615.463, 7797.84612, 7993.99091, 8202.87989, 8423.49561, 8654.8206, 8895.83739, 9189.11498, 9475.11249, 9745.95618, 9993.7723, 10210.6871, 10388.8269, 10520.3179, 10590.3857, 10540.7864},
        {417.246149, 413.835941, 410.081312, 405.994293, 401.586912, 396.871202, 391.859191, 386.56291, 380.99439, 375.165661, 369.088753, 362.775696, 356.23852, 349.489257, 342.539935, 335.402586, 328.089239, 320.611925, 312.982675, 305.213518, 297.316485, 289.303605, 281.18691, 272.97843, 264.690194, 256.334233, 247.922578, 239.467258, 230.980304, 222.473746, 213.959615, 205.449941, 196.956753, 188.492083, 180.097278, 171.769093, 163.503931, 155.31287, 147.206989, 139.197367, 131.295083, 123.511215, 115.856842, 108.343044, 100.980899, 93.7814855, 86.755883, 79.91517, 73.2704254, 66.8327281, 60.6131569, 55.2602132, 50.1337353, 45.2309838, 40.5492195, 36.0857033, 31.8376959, 27.802458, 23.9772505, 20.359334, 16.9459694, 13.7344175, 10.7219389, 7.90579455, 5.28324512, 2.85155139, 0.607974132, -1.39458968, -3.10457888, -4.63947818, -6.0056335, -7.20939072, -8.25709576, -9.1550945, -9.90973285, -10.5273567, -11.014312, -11.3769445, -11.6216003, -11.7546252, -11.7823651, -11.7111659, -11.5473735, -11.2973338, -11.1010822, -10.8953079, -10.6150932, -10.2624455, -9.8393722, -9.3478808, -8.78997868, -8.16767326, -7.48297196, -6.73788219, -5.93441137, -5.07456693, -4.16035629, -3.19378685, -2.17686605, -1.11160129, 0, 0.830278192, 1.51593359, 2.1017009, 2.63231485, 3.15251016, 3.70702154, 4.34058372, 5.09793141, 6.02379932, 7.16292219, 8.56003472, 10.2598716, 12.3071677, 14.7466575, 17.6230759, 20.9811575, 25.0826969, 30.0034374, 35.297145, 41.0550673, 47.3684518, 54.3285461, 62.0265977, 70.5538541, 80.0015629, 90.4609716, 102.023328, 114.779879, 128.821872, 144.286015, 161.201427, 179.645717, 199.696497, 220.463236, 242.507613, 266.649479, 293.256886, 322.697884, 355.340525, 391.552861, 433.523038, 479.245071, 528.730907, 581.992492, 639.041772, 698.333237, 761.825654, 829.920334, 903.018585, 981.521719, 1065.28603, 1155.92397, 1254.48537, 1361.76654, 1478.04634, 1603.83593, 1739.6465, 1886.72228, 2044.67479, 2212.31586, 2388.47674, 2574.75509, 2773.38305, 2984.94037, 3207.20694, 3439.24784, 3680.21367, 3929.19171, 4184.95555, 4446.45624, 4716.21374, 4989.06107, 5262.71387, 5534.88776, 5803.29839, 6076.76426, 6339.12239, 6585.31271, 6810.27513, 7008.94954, 7144.17227, 7264.38922, 7378.299, 7520.10097, 7688.39969, 7871.17437, 8067.56082, 8276.69485, 8497.71226, 8729.74886, 8783.51328, 8999.52927, 9318.18548, 9679.87053, 10024.9731, 10293.8818, 10426.9852, 10581.6194, 10583.9178},
        {425.6531, 422.197683, 418.433008, 414.368717, 410.014451, 405.379852, 400.474562, 395.308222, 389.890474, 384.230958, 378.339318, 372.225194, 365.898228, 359.368062, 352.644337, 345.736695, 338.654777, 331.408225, 324.006681, 316.459785, 308.77718, 300.968508, 293.043409, 285.011526, 276.8825, 268.665972, 260.371584, 252.008979, 243.587796, 235.117678, 226.608267, 218.069204, 209.51013, 200.940688, 192.214765, 183.423156, 174.658526, 165.935573, 157.268992, 148.67348, 140.163732, 131.754444, 123.460314, 115.296035, 107.276306, 99.4158212, 91.7292775, 84.2313708, 76.9367972, 69.8602528, 63.0164336, 57.225611, 51.6794426, 46.375161, 41.3099992, 36.4811898, 31.8859657, 27.5215596, 23.3852044, 19.4741328, 15.7855776, 12.3167716, 9.06494765, 6.02733844, 3.2011768, 0.583695515, -1.82787263, -3.97318933, -5.79343143, -7.42151652, -8.86430269, -10.128648, -11.2214106, -12.1494486, -12.9196199, -13.5387828, -14.0137953, -14.3515155, -14.5588014, -14.6425112, -14.609503, -14.4666348, -14.2207647, -13.8787508, -13.5865259, -13.2784855, -12.8873535, -12.4154747, -11.865194, -11.2388563, -10.5388063, -9.76738895, -8.92694905, -8.01983147, -7.04838104, -6.01494261, -4.92186103, -3.77148113, -2.56614776, -1.30820577, 0, 2.00832054, 3.68713201, 5.09378085, 6.28561352, 7.31997647, 8.25421616, 9.14567903, 10.0517115, 11.0296602, 12.1368713, 13.4306915, 14.9684671, 16.8075446, 19.0052705, 21.6189912, 24.7060532, 28.422196, 32.8388224, 37.7907227, 43.3563277, 49.6140682, 56.6423747, 64.5196781, 73.3244089, 83.134998, 94.029876, 106.087474, 119.386221, 134.00455, 150.178909, 167.771769, 186.814158, 207.337101, 228.118824, 249.816752, 273.42239, 299.342602, 327.984253, 359.754207, 395.05933, 436.428255, 481.500322, 530.267269, 582.72083, 638.852743, 697.173237, 759.525931, 826.272939, 897.776372, 974.398345, 1055.83871, 1143.93127, 1239.96078, 1344.78431, 1458.38578, 1581.14112, 1713.42627, 1856.10729, 2008.95862, 2171.21999, 2341.80913, 2522.5463, 2716.14909, 2923.34869, 3141.51048, 3369.5448, 3606.09842, 3850.01336, 4101.09804, 4358.61426, 4626.20274, 4898.07262, 5171.96986, 5445.64045, 5716.83037, 5999.7593, 6271.58108, 6525.92327, 6756.41342, 6956.67908, 7082.93851, 7191.16808, 7291.02788, 7420.34565, 7577.6318, 7749.66023, 7935.52816, 8134.33283, 8345.17148, 8567.14134, 8503.35427, 8678.40832, 9009.87797, 9415.3377, 9812.36199, 10118.5253, 10251.4022, 10471.0332, 10518.914},
        {432.041947, 428.934754, 425.520491, 421.80772, 417.805003, 413.520902, 408.963979, 404.142795, 399.065913, 393.741893, 388.179299, 382.386691, 376.372632, 370.145684, 363.714407, 357.087365, 350.273119, 343.28023, 336.117262, 328.792774, 321.31533, 313.693491, 305.935818, 298.050875, 290.047222, 281.933421, 273.718034, 265.409624, 257.016751, 248.547978, 240.011867, 231.416979, 222.771876, 214.08512, 205.115796, 196.002451, 186.886632, 177.784997, 168.714203, 159.690909, 150.731772, 141.853449, 133.0726, 124.405881, 115.869951, 107.481467, 99.2570869, 91.213469, 83.3672709, 75.7351505, 68.3337655, 62.000012, 55.9125284, 50.0701916, 44.4718781, 39.1164646, 34.0028278, 29.1298444, 24.4963911, 20.1013445, 15.9435813, 12.0219782, 8.33541192, 4.88275905, 1.66289631, -1.32529962, -4.08295207, -6.50686037, -8.49632865, -10.2676406, -11.8286825, -13.1873405, -14.3515008, -15.3290497, -16.1278734, -16.7558581, -17.22089, -17.5308554, -17.6936404, -17.7171314, -17.6092144, -17.3777758, -17.0307017, -16.5758785, -16.147601, -15.6878924, -15.1372562, -14.4994764, -13.7783371, -12.9776222, -12.1011158, -11.1526019, -10.1358644, -9.05468749, -7.91285507, -6.71415118, -5.46235983, -4.16126502, -2.81465078, -1.4263011, 0, 0.538709448, 1.00007119, 1.42640628, 1.86003574, 2.34328061, 2.91846194, 3.62790076, 4.51391811, 5.61883504, 6.98497257, 8.65465175, 10.6701936, 13.0739192, 15.9081496, 19.2152057, 23.0374087, 27.6299027, 33.0654119, 38.904872, 45.2362088, 52.1473485, 59.7262169, 68.06074, 77.2388438, 87.3484544, 98.4774977, 110.7139, 124.145586, 138.860483, 155.126498, 172.785583, 191.871669, 212.418687, 233.537102, 255.722579, 279.717041, 305.83146, 334.376808, 365.664056, 400.004177, 438.42372, 480.300293, 525.804865, 575.108401, 628.381868, 686.600228, 748.929456, 815.33952, 885.800388, 960.28203, 1036.7422, 1119.62245, 1212.1305, 1315.78848, 1428.67718, 1550.4225, 1680.65031, 1817.9126, 1963.15323, 2118.48761, 2285.62563, 2463.40943, 2651.5241, 2849.73058, 3057.91834, 3275.91786, 3503.87883, 3741.71447, 3988.16758, 4242.64303, 4510.09615, 4783.52739, 5060.42031, 5338.25843, 5614.52528, 5905.02929, 6184.3479, 6445.38342, 6681.03817, 6884.21447, 7007.8563, 7111.46321, 7205.06254, 7328.32192, 7479.53605, 7644.64952, 7822.57455, 8012.22335, 8212.50814, 8422.34113, 8543.41659, 8750.53124, 9012.53971, 9298.29665, 9576.65666, 9816.4744, 9986.60449, 10261.871, 10390.0449},
        {438.485846, 435.859138, 432.911824, 429.65222, 426.088639, 422.229397, 418.08281, 413.657191, 408.960856, 404.00212, 398.789298, 393.330704, 387.634654, 381.709462, 375.563444, 369.204914, 362.642187, 355.883578, 348.937403, 341.811975, 334.51561, 327.056624, 319.44333, 311.684044, 303.78708, 295.760754, 287.613381, 279.353275, 270.988751, 262.528125, 253.979711, 245.351824, 236.652779, 227.890891, 218.821435, 209.584569, 200.32349, 191.054726, 181.794803, 172.560247, 163.367585, 154.233344, 145.17405, 136.206229, 127.346409, 118.611116, 110.016876, 101.580216, 93.3176623, 85.2457417, 77.3809806, 70.384657, 63.6145691, 57.0732666, 50.763299, 44.6872159, 38.847567, 33.2469018, 27.88777, 22.7727212, 17.904305, 13.2850709, 8.91756863, 4.80434776, 0.947957904, -2.64905134, -5.98413037, -8.8725496, -11.1339535, -13.1415251, -14.9045246, -16.4322125, -17.7338491, -18.8186949, -19.6960101, -20.3750553, -20.8650906, -21.1753767, -21.3151737, -21.2937421, -21.1203423, -20.8042346, -20.3546795, -19.7809372, -19.2003349, -18.5658266, -17.8290674, -16.9958106, -16.0718098, -15.0628182, -13.9745894, -12.8128767, -11.5834335, -10.2920133, -8.9443694, -7.54625525, -6.10342426, -4.62162984, -3.10662538, -1.5641643, 0, -1.10860957, -2.013216, -2.68900653, -3.11116841, -3.25488888, -3.09535518, -2.60775455, -1.76727424, -0.549101482, 1.07157648, 3.1195724, 5.61969903, 8.59676915, 12.0755955, 16.0809908, 20.6377679, 26.1192192, 32.5999403, 39.316516, 46.3684333, 53.8551793, 61.8762409, 70.531105, 79.9192587, 90.1401889, 101.293383, 113.478327, 126.794508, 141.341414, 157.406277, 174.831999, 193.661742, 213.93867, 235.082651, 257.448496, 281.557227, 307.638998, 335.923958, 366.642262, 400.024059, 436.043968, 475.265446, 517.96864, 564.4337, 614.940772, 671.667414, 732.522012, 797.31036, 865.838255, 937.911492, 1010.80355, 1089.94759, 1179.54799, 1281.3999, 1392.37969, 1511.57219, 1638.06219, 1768.35545, 1904.70199, 2052.16532, 2215.73258, 2390.68322, 2573.12502, 2761.63174, 2958.95562, 3165.93091, 3384.39412, 3615.43935, 3856.48569, 4107.03105, 4373.00353, 4646.48152, 4924.73706, 5205.04217, 5484.66891, 5780.78717, 6065.79666, 6331.99492, 6571.67953, 6777.14805, 6898.89715, 6998.89256, 7087.34651, 7204.98445, 7349.94048, 7507.63485, 7676.83733, 7856.3177, 8044.84576, 8241.19127, 8511.83997, 8742.05087, 8942.5306, 9123.98579, 9297.12308, 9472.64911, 9661.27051, 9965.73757, 10158.5614},
        {446.318603, 444.296053, 441.915025, 439.18501, 436.115502, 432.715993, 428.995975, 424.964941, 420.632382, 416.007791, 411.100661, 405.920483, 400.47675, 394.778954, 388.836588, 382.659144, 376.256114, 369.636991, 362.811266, 355.788433, 348.577983, 341.189409, 333.632202, 325.915857, 318.049864, 310.043716, 301.906905, 293.648924, 285.279265, 276.807421, 268.242882, 259.595143, 250.873695, 242.088031, 233.154201, 224.130383, 215.07588, 206.003216, 196.924916, 187.853504, 178.801505, 169.781444, 160.805846, 151.887234, 143.038135, 134.271072, 125.59857, 117.033153, 108.587348, 100.273677, 92.1046661, 84.1970627, 76.4569092, 68.8944707, 61.5200128, 54.3438006, 47.3760995, 40.6271749, 34.107292, 27.8267162, 21.7957128, 16.0245471, 10.5234844, 5.3027901, 0.37272949, -4.25643211, -8.57442938, -12.2276316, -14.870652, -17.2013908, -19.2318416, -20.9739981, -22.4398538, -23.6414026, -24.590638, -25.2995537, -25.7801433, -26.0444006, -26.1043192, -25.9718927, -25.6591148, -25.1779792, -24.5404795, -23.7586094, -22.9830679, -22.1535785, -21.2081976, -20.1544175, -18.9997307, -17.7516296, -16.4176066, -15.0051541, -13.5217645, -11.9749301, -10.3721435, -8.72089695, -7.02868293, -5.30299382, -3.55132203, -1.78115995, 0, -1.23890736, -2.23588075, -2.96976249, -3.41939489, -3.56362025, -3.3812809, -2.85121914, -1.95227729, -0.663297652, 1.03687746, 3.16940572, 5.75544484, 8.81615249, 12.3726864, 16.4462042, 21.0578635, 26.5619357, 33.0271659, 39.7212409, 46.7366999, 54.1660821, 62.1019265, 70.6367724, 79.8631588, 89.873625, 100.76071, 112.616953, 125.534893, 139.607069, 155.069497, 171.818631, 189.903968, 209.375006, 229.492573, 250.7005, 273.652928, 298.635953, 325.93567, 355.838178, 388.629571, 425.853745, 466.156189, 509.57691, 556.155913, 605.933205, 656.851939, 711.573185, 770.661162, 834.680088, 904.194182, 979.975506, 1062.1264, 1152.15322, 1250.6017, 1356.0563, 1467.98208, 1585.84406, 1705.80136, 1831.37629, 1969.69764, 2129.2325, 2301.4904, 2479.17015, 2659.55927, 2847.72057, 3045.14785, 3254.77247, 3478.46093, 3712.80881, 3957.39322, 4218.37587, 4487.73626, 4762.77223, 5040.7816, 5319.06222, 5617.7373, 5905.57295, 6174.16063, 6415.09184, 6619.95806, 6736.35257, 6829.1972, 6908.9398, 7016.97061, 7151.48946, 7297.80355, 7454.75734, 7621.1953, 7795.96191, 7977.90163, 8120.61656, 8306.40351, 8518.51817, 8740.21628, 8954.75354, 9145.38566, 9295.36838, 9587.32525, 9813.9132},
        {454.702129, 453.110811, 451.140099, 448.800158, 446.101153, 443.053249, 439.666609, 435.951399, 431.917783, 427.575926, 422.935992, 418.008146, 412.802554, 407.329378, 401.598785, 395.620938, 389.406003, 382.964143, 376.305524, 369.44031, 362.378667, 355.130757, 347.706747, 340.1168, 332.371082, 324.479757, 316.45299, 308.300945, 300.033786, 291.66168, 283.194789, 274.64328, 266.017316, 257.327062, 248.643834, 239.945934, 231.210931, 222.447005, 213.662336, 204.865104, 196.063489, 187.265672, 178.479832, 169.71415, 160.976806, 152.27598, 143.619852, 135.016603, 126.474412, 118.00146, 109.605928, 100.872475, 92.2419818, 83.7318107, 75.3593224, 67.1418782, 59.0968393, 51.2415668, 43.5934222, 36.1697665, 28.987961, 22.0653671, 15.4193458, 9.0672584, 3.0264662, -2.68566959, -8.05178775, -12.5902465, -15.8305512, -18.7128844, -21.2499824, -23.4545812, -25.3394172, -26.9172265, -28.2007453, -29.2027097, -29.935856, -30.4129203, -30.6466388, -30.6497477, -30.4349832, -30.0150815, -29.4027787, -28.610811, -27.7117412, -26.6871337, -25.5178342, -24.2146372, -22.7883375, -21.2497298, -19.6096087, -17.8787691, -16.0680055, -14.1881127, -12.2498854, -10.2641183, -8.2416061, -6.19314345, -4.12952508, -2.0615457, 0, -1.35910227, -2.45369005, -3.26552453, -3.77636692, -3.96797844, -3.8221203, -3.32055369, -2.44503984, -1.17733994, 0.500784782, 2.60757313, 5.1612639, 8.18009587, 11.6823078, 15.6861386, 20.2098269, 25.6265256, 32.0051753, 38.5612334, 45.3889916, 52.5827417, 60.2367756, 68.4453849, 77.3028614, 86.903497, 97.3415834, 108.711412, 121.107276, 134.623465, 149.547984, 165.710385, 183.146846, 201.893546, 220.979757, 240.945111, 262.597747, 286.275916, 312.317868, 341.061854, 372.846125, 410.547048, 451.192285, 494.623498, 540.682349, 589.210501, 634.862773, 683.96438, 737.653694, 797.06909, 863.348939, 940.012012, 1022.90691, 1111.09928, 1204.08658, 1302.2479, 1405.5665, 1514.02564, 1624.34688, 1740.51646, 1870.07885, 2022.82777, 2188.90701, 2359.91361, 2532.67989, 2712.90901, 2902.2323, 3103.79189, 3319.61081, 3546.17226, 3783.09308, 4035.61258, 4296.86012, 4564.50632, 4836.22175, 5109.67702, 5410.11964, 5700.74907, 5972.34165, 6215.67377, 6421.52178, 6534.75022, 6622.65517, 6695.6895, 6794.5105, 6917.61511, 7050.91508, 7193.53662, 7344.60591, 7503.24915, 7668.59254, 7648.22591, 7780.44052, 8012.50204, 8291.67614, 8565.2285, 8780.4248, 8884.5307, 9146.07088, 9369.0033},
        {463.557869, 460.843628, 457.907647, 454.753706, 451.385587, 447.807068, 444.02193, 440.033952, 435.846915, 431.464598, 426.890782, 422.129246, 417.183771, 412.058136, 406.756122, 401.281507, 395.638073, 389.8296, 383.859866, 377.732653, 371.45174, 365.020906, 358.443933, 351.724601, 344.866688, 337.873975, 330.750242, 323.499268, 316.124835, 308.630722, 301.020708, 293.298574, 285.4681, 277.533066, 269.367426, 261.042604, 252.631583, 244.142357, 235.582918, 226.96126, 218.285375, 209.563257, 200.802898, 192.012293, 183.199432, 174.372311, 165.538922, 156.707257, 147.88531, 139.081075, 130.302543, 120.8618, 111.477834, 102.173722, 92.9725447, 83.8973799, 74.9713069, 66.2174047, 57.6587523, 49.3184285, 41.2195123, 33.3850828, 25.8382188, 18.6019994, 11.6995034, 5.15380988, -1.01200226, -6.17454902, -9.72485951, -12.8779388, -15.6496234, -18.0557495, -20.1121536, -21.834672, -23.2391412, -24.3413974, -25.1572771, -25.7026166, -25.9932524, -26.0450207, -25.873758, -25.4953006, -24.9254849, -24.1801472, -23.4319712, -22.6150705, -21.6616737, -20.5825272, -19.3883773, -18.0899704, -16.6980529, -15.2233712, -13.6766717, -12.0687008, -10.4102048, -8.71193013, -6.98462316, -5.23903028, -3.48589787, -1.73597232, 0, -5.60719688, -10.196777, -13.8037848, -16.4632645, -18.2102605, -19.0798171, -19.1069787, -18.3267897, -16.7742943, -14.4845368, -11.4925618, -7.83341337, -3.54213601, 1.346226, 6.7966283, 12.7740265, 20.7605339, 30.9378429, 39.8392433, 47.754796, 54.9745617, 61.7886014, 68.4869758, 75.3597459, 82.6969725, 90.7887163, 99.9250384, 110.395999, 122.49166, 138.085904, 155.304235, 173.961567, 193.872813, 213.022391, 232.140463, 252.445324, 274.301039, 298.071669, 324.121278, 352.81393, 385.8818, 421.904456, 460.978288, 503.199685, 548.665035, 600.245918, 654.569735, 711.039077, 769.056536, 828.024703, 883.473338, 943.411325, 1013.35993, 1095.69661, 1186.38021, 1284.25139, 1388.15084, 1493.71204, 1603.71174, 1724.42544, 1861.08888, 2009.83505, 2168.96549, 2337.57568, 2516.10633, 2704.38067, 2902.14205, 3109.19296, 3325.62884, 3551.37942, 3790.33016, 4037.84649, 4292.48885, 4552.81769, 4817.39345, 5120.82123, 5416.60565, 5694.29598, 5943.44148, 6153.59143, 6265.37457, 6349.8744, 6417.60613, 6510.14467, 6625.71077, 6749.58365, 6880.63114, 7017.72108, 7159.72129, 7305.49962, 7237.32894, 7328.16975, 7527.68963, 7785.55618, 8051.43701, 8274.9997, 8405.91185, 8652.74701, 8737.68932},
        {482.446566, 480.928505, 479.054361, 476.83361, 474.275733, 471.390209, 468.186517, 464.674136, 460.862545, 456.761223, 452.37965, 447.727304, 442.813665, 437.648211, 432.240423, 426.599778, 420.735757, 414.657838, 408.375501, 401.898224, 395.235487, 388.396769, 381.391548, 374.229305, 366.919519, 359.471667, 351.895231, 344.199688, 336.394517, 328.489199, 320.493212, 312.416035, 304.267147, 296.056028, 287.940759, 279.853392, 271.724193, 263.557819, 255.358928, 247.132175, 238.882217, 230.613712, 222.331315, 214.039684, 205.743476, 197.447346, 189.155951, 180.873949, 172.605996, 164.356749, 156.130865, 147.541662, 138.993619, 130.499875, 122.073571, 113.727846, 105.475842, 97.330697, 89.3055525, 81.4135483, 73.6678245, 66.0815212, 58.6677786, 51.4397368, 44.4105359, 37.5933161, 31.0012175, 24.7627925, 19.0038225, 13.4994186, 8.25523904, 3.27694235, -1.42981307, -5.8593688, -10.0060664, -13.8642474, -17.4282535, -20.6924262, -23.651107, -26.2986376, -28.6293596, -30.6376144, -32.3177436, -33.664089, -33.5593186, -32.5769932, -31.3040352, -29.770862, -28.0078912, -26.0455403, -23.9142267, -21.6443681, -19.2663818, -16.8106855, -14.3076966, -11.7878326, -9.28151114, -6.81914962, -4.43116558, -2.14797653, 0, 0.237163541, -0.0161211039, -0.68448259, -1.69254957, -2.96495071, -4.42631465, -6.00127007, -7.6144456, -9.19046991, -10.6539716, -11.9295795, -12.9419221, -13.615628, -13.8753261, -13.6456448, -12.8512129, -12.2221394, -11.7981215, -9.68186808, -5.97061074, -0.761581066, 5.84798937, 13.760869, 22.8798262, 33.1076294, 44.347047, 56.5008475, 69.4717992, 83.1626706, 96.8511261, 111.294243, 126.582321, 142.805659, 160.33955, 179.131799, 199.054208, 220.11158, 242.308716, 265.650419, 290.141489, 313.939176, 339.458132, 367.064637, 397.12497, 430.00541, 462.304784, 499.098687, 541.695261, 591.402648, 649.528991, 722.72609, 800.427292, 875.019579, 943.653241, 1013.20365, 1086.17983, 1165.09078, 1254.01888, 1353.54219, 1462.52239, 1583.42603, 1713.45873, 1849.44345, 1990.4467, 2139.33648, 2297.23581, 2467.47979, 2651.76497, 2843.67702, 3041.38957, 3232.01103, 3426.48262, 3626.80824, 3834.9918, 4053.0372, 4334.52855, 4616.99452, 4889.54395, 5141.2857, 5361.32864, 5498.16389, 5608.59652, 5699.14296, 5781.53759, 5860.23144, 5937.50138, 6017.90713, 6106.00843, 6206.36499, 6323.53653, 6045.90001, 6156.16219, 6505.00993, 6943.13006, 7321.20942, 7489.93486, 7299.99322, 7616.11684, 7730.27552},
        {487.525079, 487.514284, 487.050166, 486.145273, 484.812152, 483.063351, 480.911417, 478.368899, 475.448344, 472.162301, 468.523315, 464.543937, 460.236712, 455.61419, 450.688917, 445.473442, 439.980312, 434.222075, 428.211279, 421.960471, 415.4822, 408.789012, 401.893457, 394.808081, 387.545432, 380.118058, 372.538507, 364.819327, 356.973065, 349.012268, 340.949486, 332.797265, 324.568154, 316.2747, 308.216679, 300.256984, 292.255054, 284.214116, 276.137396, 268.028121, 259.889518, 251.724813, 243.537233, 235.330004, 227.106353, 218.869508, 210.622693, 202.369137, 194.112065, 185.854704, 177.600281, 169.121423, 160.656955, 152.215103, 143.804092, 135.432147, 127.107495, 118.838361, 110.632971, 102.499549, 94.4463223, 86.4815156, 78.6133547, 70.8500651, 63.1998724, 55.6710023, 48.2716802, 40.7852431, 33.0004275, 25.3892749, 17.9745894, 10.7791752, 3.82583652, -2.86262249, -9.26339763, -15.3536847, -21.1106796, -26.511578, -31.5335758, -36.1538688, -40.3496528, -44.0981237, -47.3764772, -50.1619091, -50.3239601, -48.9379851, -47.1046719, -44.8696136, -42.2784035, -39.376635, -36.2099013, -32.8237956, -29.2639113, -25.5758417, -21.80518, -17.9975195, -14.1984535, -10.4535753, -6.80847807, -3.30875523, 0, -1.36852174, -2.9645561, -4.72891623, -6.6024153, -8.52586646, -10.4400829, -12.2858778, -14.0040642, -15.5354554, -16.8208645, -17.8011047, -18.4169891, -18.6093309, -18.3189433, -17.4866395, -16.0532325, -14.8945264, -14.0849046, -11.4503199, -7.1319405, -1.27093488, 5.99152863, 14.5142817, 24.1561559, 34.775983, 46.2325945, 58.3848222, 71.0914976, 84.2114525, 96.5106437, 109.341498, 122.89071, 137.344974, 153.667539, 171.656821, 190.904155, 211.363268, 232.987889, 255.731743, 279.548558, 302.607199, 327.189474, 353.598325, 382.136691, 413.107511, 443.54481, 477.837672, 517.106266, 562.470759, 615.051321, 680.243949, 749.66697, 817.018789, 880.061111, 944.85154, 1013.72299, 1089.00838, 1175.70947, 1272.88377, 1376.67728, 1485.6271, 1601.04968, 1722.77581, 1851.29226, 1988.19734, 2134.57914, 2293.13315, 2465.3642, 2646.88328, 2836.63525, 3026.95805, 3224.41994, 3430.25282, 3645.68858, 3871.95911, 4160.93835, 4450.55562, 4729.38231, 4985.9898, 5208.94945, 5343.72532, 5450.73433, 5537.02244, 5620.84205, 5704.80323, 5786.84665, 5869.81583, 5956.55429, 6049.90554, 6152.71311, 6146.46838, 6273.20767, 6478.6099, 6708.35396, 6908.11876, 7023.5832, 7000.42619, 7146.4006, 7316.27282},
        {510.463075, 510.622715, 510.401023, 509.807437, 508.851394, 507.54233, 505.889684, 503.902892, 501.591392, 498.96462, 496.032014, 492.80301, 489.287047, 485.493561, 481.431989, 477.111768, 472.542336, 467.73313, 462.693587, 457.433144, 451.961238, 446.287307, 440.420787, 434.371115, 428.14773, 421.760067, 415.217564, 408.529659, 401.705788, 394.755389, 387.687898, 380.512753, 373.239392, 365.87725, 358.58659, 351.298265, 343.941314, 336.52028, 329.039704, 321.504131, 313.918102, 306.286159, 298.612847, 290.902706, 283.16028, 275.390112, 267.596743, 259.784717, 251.958576, 244.122863, 236.28212, 228.548036, 220.815685, 213.087287, 205.365061, 197.651227, 189.948006, 182.257616, 174.582279, 166.924214, 159.28564, 151.668778, 144.075847, 136.509067, 128.970659, 121.462842, 113.987836, 106.1987, 97.7568772, 89.3847046, 81.1070367, 72.9487278, 64.9346324, 57.0896052, 49.4385006, 42.0061731, 34.8174773, 27.8972676, 21.2703986, 14.9617247, 8.99610055, 3.39838059, -1.80658067, -6.59392871, -8.83630043, -9.60431906, -9.99387834, -10.0483542, -9.81112259, -9.32555944, -8.63504067, -7.78294222, -6.81264004, -5.76751005, -4.69092818, -3.62627038, -2.61691257, -1.7062307, -0.937600686, -0.354398476, 0, -11.580128, -22.0715197, -31.5002632, -39.8924466, -47.274158, -53.6714854, -59.1105171, -63.617341, -67.2180453, -69.938718, -71.8054473, -72.8443212, -73.0814278, -72.5428553, -71.2546917, -69.2430251, -66.6653996, -63.5666827, -59.6756216, -55.0464735, -49.7334956, -43.7909452, -37.2730796, -30.2341558, -22.7284313, -14.8101632, -6.5336087, 2.0469749, 10.8773304, 19.3373886, 28.1461686, 37.4191568, 47.2718394, 58.0028926, 69.6362079, 82.1468262, 95.5952769, 110.042089, 125.547793, 142.172918, 159.57604, 178.341975, 198.609896, 220.518974, 244.208383, 269.680647, 297.245748, 327.07702, 359.347798, 394.231415, 432.274729, 472.821024, 515.137216, 559.074579, 605.677457, 655.454532, 708.914483, 766.285867, 828.421154, 896.4784, 972.952521, 1055.0564, 1139.02538, 1223.99464, 1314.01301, 1410.8739, 1517.84067, 1637.0878, 1765.39992, 1902.61048, 2044.67437, 2195.90028, 2357.46421, 2530.54216, 2716.31015, 2963.53023, 3213.89584, 3456.6865, 3681.18171, 3876.66095, 3992.26109, 4084.16604, 4158.85929, 4235.89211, 4316.57532, 4396.77487, 4478.07385, 4562.05535, 4650.30245, 4744.39824, 4867.3252, 4999.63209, 5131.83724, 5254.45897, 5358.01561, 5433.02549, 5470.00692, 5585.93739, 5729.57162}
    };
    
    const double b3tab[98][200] = {{98.4097635, 97.7327282, 97.0246638, 96.2867998, 95.5203657, 94.7265909, 93.9067049, 93.0619371, 92.1935171, 91.3026744, 90.3906383, 89.4586384, 88.5079042, 87.539665, 86.5551505, 85.55559, 84.5422131, 83.5162492, 82.4789278, 81.4314783, 80.3751303, 79.3111132, 78.2406564, 77.1649896, 76.085342, 75.0029433, 73.9190229, 72.8348102, 71.7515347, 70.670426, 69.5927134, 68.5196265, 67.4523947, 66.3922475, 65.3869471, 64.413478, 63.448265, 62.4910277, 61.5414854, 60.5993576, 59.6643635, 58.7362226, 57.8146543, 56.899378, 55.9901132, 55.086579, 54.1884951, 53.2955807, 52.4075554, 51.5241383, 50.6450491, 49.7826258, 48.9236956, 48.0677042, 47.2140974, 46.3623213, 45.5118214, 44.6620438, 43.8124343, 42.9624387, 42.1115028, 41.2590725, 40.4045937, 39.5475121, 38.6872737, 37.8233242, 36.9551095, 36.0852121, 35.2161389, 34.3408671, 33.4586389, 32.568697, 31.6702839, 30.7626421, 29.8450141, 28.9166423, 27.9767694, 27.0246379, 26.0594902, 25.0805688, 24.0871163, 23.0783752, 22.053588, 21.0119973, 19.8400545, 18.5957701, 17.3385101, 16.0711773, 14.7966746, 13.5179046, 12.2377703, 10.9591743, 9.68501957, 8.41820881, 7.16164483, 5.91823044, 4.69086842, 3.48246157, 2.29591268, 1.13412456, 0, -0.492508072, -1.02763148, -1.60286055, -2.21568561, -2.86359697, -3.54408496, -4.25463991, -4.99275214, -5.75591197, -6.54160973, -7.34733574, -8.17058032, -9.0088338, -9.8595865, -10.7203287, -11.5885509, -12.4624557, -13.3396354, -14.2159686, -15.0890983, -15.9566674, -16.816319, -17.6656962, -18.5024419, -19.3241991, -20.1286109, -20.9133202, -21.6759701, -22.4142037, -23.2119093, -23.9488612, -24.5968287, -25.1275812, -25.5916498, -25.921423, -26.0282861, -25.8603797, -25.3658447, -24.4928219, -23.189452, -22.0037422, -20.1013989, -17.3131977, -13.4699143, -8.40232424, -2.29175632, 5.46920525, 15.1374231, 26.96976, 41.2230784, 56.8581817, 77.0120645, 102.512478, 134.447753, 174.438233, 223.865399, 284.110732, 353.402663, 436.992326, 543.570539, 675.918512, 843.228763, 1045.82401, 1291.47771, 1600.58816, 1987.75864, 2452.12391, 3004.27689, 3711.52847, 4609.10665, 5449.80206, 6594.73195, 8176.89113, 10329.2744, 13184.8765, 20562.1728, 27987.3073, 34671.9048, 39827.5898, 42665.987, 39060.0148, 33785.8082, 27485.8661, 22645.5703, 19177.945, 16228.1465, 13747.5924, 11687.7002, 9999.88758, 8635.57211, 7352.59002, 6409.80526, 5741.56114, 5282.20096, 4966.06801, 4727.50559, 4500.85699, 4398.38134, 4348.23251},
        {88.9798529, 88.2951217, 87.5849799, 86.8505661, 86.0930192, 85.3134776, 84.5130801, 83.6929653, 82.8542718, 81.9981384, 81.1257036, 80.2381062, 79.3364847, 78.4219779, 77.4957243, 76.5588627, 75.6125317, 74.65787, 73.6960161, 72.7281088, 71.7552867, 70.7786885, 69.7994529, 68.8187183, 67.8376237, 66.8573075, 65.8789084, 64.9035652, 63.9324164, 62.9666007, 62.0072567, 61.0555232, 60.1125387, 59.179442, 58.3192783, 57.5009308, 56.6925387, 55.8932316, 55.1021392, 54.3183912, 53.5411172, 52.7694469, 52.0025099, 51.239436, 50.4793547, 49.7213958, 48.9646889, 48.2083637, 47.4515499, 46.693377, 45.9329748, 45.1284266, 44.3207982, 43.5101091, 42.6963787, 41.8796265, 41.059872, 40.2371346, 39.4114339, 38.5827893, 37.7512202, 36.9167461, 36.0793866, 35.239161, 34.3960888, 33.5501896, 32.7014827, 31.8576977, 31.0263788, 30.1916443, 29.3530139, 28.5100071, 27.6621437, 26.8089433, 25.9499256, 25.0846103, 24.2125169, 23.3331652, 22.4460749, 21.5507655, 20.6467568, 19.7335684, 18.81072, 17.8777313, 16.8725184, 15.8266986, 14.7726294, 13.7118295, 12.6458177, 11.5761129, 10.5042339, 9.43169952, 8.36002851, 7.2907397, 6.22535192, 5.16538397, 4.11235466, 3.0677828, 2.03318719, 1.01008666, 0, -0.487537316, -1.03385848, -1.63540289, -2.28860991, -2.98991894, -3.73576936, -4.52260055, -5.3468519, -6.2049628, -7.09337262, -8.00852076, -8.94684659, -9.90478951, -10.8787889, -11.8652841, -12.8607146, -13.8952686, -14.9702068, -16.0056317, -17.0052147, -17.972627, -18.91154, -19.8256249, -20.718553, -21.5939957, -22.4556241, -23.3071097, -24.1521236, -24.9943372, -25.9490019, -26.8672961, -27.7194172, -28.4755625, -29.1827966, -29.7728832, -30.1570877, -30.2825473, -30.096399, -29.5457798, -28.5778267, -27.838004, -26.3625871, -23.9620841, -20.4470026, -15.6278506, -9.26070561, -1.22411326, 8.65781103, 20.5609519, 34.6611938, 49.2274551, 68.6733235, 94.8006322, 129.095981, 172.402368, 225.851755, 290.576105, 363.790904, 451.434698, 563.718552, 703.320691, 881.487571, 1099.68614, 1367.07779, 1705.86178, 2132.25278, 2645.23713, 3256.56291, 4041.14864, 5038.18004, 6076.69069, 7438.3496, 9235.08747, 11578.835, 14581.523, 21576.8527, 28649.5416, 35106.0777, 40252.949, 43396.6436, 41056.7545, 37184.5947, 32281.036, 28102.5778, 24653.0724, 21413.1937, 18412.9528, 15682.3611, 13251.4295, 11150.1694, 9897.34365, 8812.39826, 7902.78179, 7175.9428, 6639.32987, 6300.3916, 6166.57655, 6185.58472, 6453.05986},
        {81.2404201, 80.4276522, 79.6081486, 78.7823681, 77.9507692, 77.1138106, 76.2719509, 75.4256488, 74.5753629, 73.7215518, 72.8646742, 72.0051888, 71.143554, 70.2802287, 69.4156714, 68.5503408, 67.6846955, 66.8191942, 65.9542954, 65.0904578, 64.2281402, 63.367801, 62.5098989, 61.6548927, 60.8032408, 59.9554021, 59.111835, 58.2729982, 57.4393504, 56.6113502, 55.7894563, 54.9741272, 54.1658217, 53.3649984, 52.6151191, 51.8942364, 51.1798859, 50.4711307, 49.7670338, 49.0666584, 48.3690676, 47.6733244, 46.9784921, 46.2836336, 45.587812, 44.8900905, 44.1895322, 43.4852002, 42.7761575, 42.0614672, 41.3401925, 40.5353818, 39.7237607, 38.9060401, 38.082931, 37.2551444, 36.4233911, 35.5883822, 34.7508286, 33.9114413, 33.0709312, 32.2300092, 31.3893862, 30.5497734, 29.7118815, 28.8764216, 28.0441045, 27.2379374, 26.4803921, 25.7261953, 24.9746128, 24.2249099, 23.4763524, 22.7282057, 21.9797356, 21.2302075, 20.478887, 19.7250398, 18.9679314, 18.2068274, 17.4409933, 16.6696948, 15.8921974, 15.1077668, 14.2814371, 13.4303097, 12.5718972, 11.7065761, 10.8347227, 9.95671367, 9.07292531, 8.1837341, 7.28951651, 6.39064899, 5.48750797, 4.58046993, 3.6699113, 2.75620854, 1.83973811, 0.920876442, 0, -0.61151956, -1.27769034, -1.9953361, -2.76128062, -3.57234765, -4.42536096, -5.31714432, -6.24452148, -7.20431623, -8.19335231, -9.2084535, -10.2464436, -11.3041463, -12.3783854, -13.4659846, -14.5637678, -15.6701332, -16.7821293, -17.8930188, -18.9999628, -20.1001225, -21.190659, -22.2687334, -23.3315069, -24.3761406, -25.3997957, -26.3996332, -27.3728145, -28.3165005, -29.2859219, -30.1988782, -31.0351098, -31.7743571, -32.4801404, -33.09031, -33.5203748, -33.7249411, -33.6586152, -33.2760033, -32.5317118, -32.2337918, -31.2236606, -29.2889462, -26.2172765, -21.7962795, -15.9210933, -8.24495819, 1.47137557, 13.4671576, 27.9816374, 43.4440488, 64.1158989, 91.289527, 126.393795, 171.136296, 227.099481, 295.865799, 376.246664, 473.225341, 594.808041, 740.480212, 925.818566, 1152.86534, 1432.84226, 1792.52516, 2251.5503, 2809.42907, 3480.58017, 4353.21336, 5473.79799, 6657.75087, 8218.14044, 10281.4158, 12974.026, 16422.4202, 24433.5609, 32533.2553, 39927.8242, 45823.5882, 49426.8679, 46729.8717, 42295.7743, 36708.3731, 32078.1236, 28384.5212, 24970.9535, 21848.7211, 19029.1246, 16523.4645, 14343.0416, 13198.4729, 11967.8175, 10746.6429, 9630.51633, 8715.00515, 8095.67668, 7868.09825, 7677.20175, 7788.23028},
        {74.8104589, 73.8658782, 72.9408945, 72.034769, 71.1467631, 70.2761381, 69.4221554, 68.5840762, 67.7611619, 66.9526738, 66.1578733, 65.3760217, 64.6063802, 63.8482104, 63.1007734, 62.3633306, 61.6351433, 60.9154729, 60.2035807, 59.498728, 58.8001761, 58.1071864, 57.4190203, 56.7349389, 56.0542038, 55.3760761, 54.6998173, 54.0246886, 53.3499514, 52.674867, 51.9986967, 51.320702, 50.640144, 49.9562842, 49.2353111, 48.493718, 47.7483959, 46.9996792, 46.2479027, 45.493401, 44.7365085, 43.97756, 43.21689, 42.4548331, 41.691724, 40.9278972, 40.1636873, 39.399429, 38.6354569, 37.8721055, 37.1097095, 36.3687079, 35.6288952, 34.8901699, 34.1524309, 33.415577, 32.6795069, 31.9441195, 31.2093134, 30.4749876, 29.7410406, 29.0073714, 28.2738787, 27.5404613, 26.8070179, 26.0734474, 25.3396484, 24.60396, 23.8647584, 23.1251596, 22.3851635, 21.6447699, 20.9039787, 20.1627899, 19.4212033, 18.6792189, 17.9368365, 17.1940561, 16.4508774, 15.7073005, 14.9633252, 14.2189515, 13.4741791, 12.7290081, 12.0121767, 11.3087112, 10.6032924, 9.89498743, 9.18286367, 8.46598838, 7.74342883, 7.01425229, 6.27752605, 5.53231739, 4.77769357, 4.01272188, 3.23646959, 2.44800398, 1.64639232, 0.830701907, 0, -0.928631309, -1.87062187, -2.82548311, -3.79272648, -4.77186341, -5.76240534, -6.7638637, -7.77574993, -8.79757547, -9.82885175, -10.8690902, -11.9178023, -12.9744994, -14.0386931, -15.1098946, -16.1876155, -17.2410416, -18.265352, -19.3286519, -20.4239545, -21.5442727, -22.6826198, -23.8320086, -24.9854525, -26.1359643, -27.2765572, -28.4002444, -29.5000388, -30.5689535, -31.6352894, -32.6438329, -33.5770109, -34.41725, -35.2277441, -35.9505363, -36.5061321, -36.8527282, -36.9485213, -36.7517081, -36.2204852, -36.2230976, -35.5307225, -33.9235037, -31.1815848, -27.0851096, -22.4137679, -15.6982708, -6.46887593, 5.74415961, 21.4105785, 40.1101974, 64.2903723, 93.0115871, 127.114364, 171.073469, 227.731967, 299.932922, 389.504433, 500.535206, 638.221185, 798.467064, 1002.46373, 1257.69936, 1577.83352, 1986.98286, 2504.46407, 3119.08644, 3842.25734, 4797.24453, 6044.04119, 7401.65425, 9208.13798, 11606.9039, 14741.3635, 18754.9283, 28040.0194, 37428.7863, 46002.3882, 52841.9843, 57028.7336, 53900.9483, 48777.8658, 42344.7226, 37137.4751, 33108.7829, 29440.1725, 26122.8601, 23148.0622, 20506.9951, 18190.8752, 16228.3059, 14583.1018, 13217.8003, 12094.9391, 11177.0555, 10426.6871, 9806.37138, 9414.92513, 9198.11435},
        {70.5057947, 69.6591428, 68.8282526, 68.0124611, 67.2111057, 66.4235234, 65.6490515, 64.8870271, 64.1367874, 63.3976697, 62.669011, 61.9501486, 61.2404197, 60.5391615, 59.8457111, 59.1594057, 58.4795825, 57.8055787, 57.1367314, 56.472378, 55.8118554, 55.154501, 54.499652, 53.8466454, 53.1948185, 52.5435085, 51.8920525, 51.2397878, 50.5860515, 49.9301808, 49.2715129, 48.609385, 47.9431343, 47.2720979, 46.5531756, 45.8078159, 45.0579774, 44.3043744, 43.5477214, 42.7887327, 42.0281226, 41.2666056, 40.5048959, 39.7437079, 38.9837561, 38.2257547, 37.4704181, 36.7184607, 35.9705968, 35.2275409, 34.4900072, 33.82473, 33.1649725, 32.5100181, 31.8591498, 31.2116509, 30.5668046, 29.9238939, 29.2822022, 28.6410125, 27.9996081, 27.3572721, 26.7132877, 26.0669382, 25.4175067, 24.7642763, 24.1065302, 23.423354, 22.6943181, 21.9603619, 21.2220778, 20.4800585, 19.7348963, 18.9871839, 18.2375138, 17.4864784, 16.7346703, 15.982682, 15.231106, 14.4805348, 13.731561, 12.984777, 12.2407755, 11.5001487, 10.8313047, 10.1995019, 9.56918361, 8.93874163, 8.30656775, 7.67105375, 7.03059139, 6.38357247, 5.72838876, 5.06343203, 4.38709407, 3.69776666, 2.99384157, 2.27371058, 1.53576547, 0.778398015, 0, -1.16977963, -2.31275573, -3.43134998, -4.52798409, -5.60507975, -6.66505866, -7.71034252, -8.74335302, -9.76651187, -10.7822408, -11.7929614, -12.8010954, -13.8090646, -14.8192907, -15.8341952, -16.8562, -17.8338322, -18.7618145, -19.7644725, -20.8326791, -21.9573073, -23.1292299, -24.3393198, -25.5784498, -26.837493, -28.1073222, -29.3788102, -30.6428301, -31.8902546, -33.1537261, -34.3670326, -35.5085161, -36.5565188, -37.5945106, -38.5482697, -39.3155399, -39.8431249, -40.0778283, -39.9664538, -39.4558051, -39.2850931, -38.3675467, -36.4949334, -33.4590205, -29.0515755, -24.864798, -18.4399152, -9.11858651, 3.75752865, 20.8467709, 42.383223, 69.9680207, 101.546673, 137.727506, 184.555435, 245.634456, 324.568566, 423.835643, 548.42174, 704.541408, 888.585781, 1123.47598, 1417.36102, 1785.29619, 2254.03904, 2844.97559, 3541.74485, 4355.94656, 5437.58623, 6858.37923, 8342.454, 10362.5878, 13110.8145, 16779.1682, 21559.6826, 33030.9841, 44651.8663, 55267.715, 63723.9162, 68865.8557, 64792.3637, 58259.7523, 50147.6457, 43811.0798, 39149.8281, 35032.2428, 31409.6686, 28233.4501, 25454.932, 23025.459, 20365.0879, 18306.9473, 16717.5299, 15463.3285, 14410.8358, 13426.5447, 12376.9478, 11664.0261, 11192.1058},
        {68.2331449, 67.6536545, 67.0564728, 66.4424133, 65.8122896, 65.1669154, 64.5071041, 63.8336695, 63.1474251, 62.4491845, 61.7397612, 61.019969, 60.2906214, 59.552532, 58.8065144, 58.0533821, 57.2939489, 56.5290282, 55.7594337, 54.985979, 54.2094776, 53.4307433, 52.6505895, 51.8698298, 51.089278, 50.3097475, 49.5320519, 48.7570049, 47.9854201, 47.218111, 46.4558912, 45.6995744, 44.9499741, 44.207904, 43.5194208, 42.8617648, 42.2116327, 41.5683699, 40.9313217, 40.2998335, 39.6732507, 39.0509186, 38.4321826, 37.816388, 37.2028803, 36.5910048, 35.9801069, 35.3695319, 34.7586253, 34.1467323, 33.5331984, 32.8831009, 32.230796, 31.5763721, 30.9199172, 30.2615198, 29.6012679, 28.9392498, 28.2755538, 27.6102681, 26.943481, 26.2752806, 25.6057552, 24.934993, 24.2630823, 23.5901113, 22.9161682, 22.2415203, 21.5664304, 20.8906177, 20.2141588, 19.5371303, 18.8596089, 18.1816712, 17.5033938, 16.8248533, 16.1461264, 15.4672898, 14.78842, 14.1095937, 13.4308875, 12.752378, 12.0741419, 11.3962558, 10.7641422, 10.154251, 9.54248703, 8.92745547, 8.30776136, 7.68200978, 7.0488058, 6.40675448, 5.75446092, 5.09053017, 4.41356731, 3.72217742, 3.01496557, 2.29053683, 1.54749627, 0.784448969, 0, -1.15112045, -2.2738219, -3.3712812, -4.44667517, -5.50318068, -6.54397458, -7.57223369, -8.59113488, -9.60385499, -10.6135709, -11.6234593, -12.6366973, -13.6564615, -14.6859289, -15.7282763, -16.7866805, -17.7950136, -18.7465517, -19.8012325, -20.947382, -22.1733259, -23.46739, -24.8179002, -26.2131822, -27.6415618, -29.0913649, -30.5509173, -32.0085447, -33.452573, -34.9148531, -36.3242265, -37.6559614, -38.8853264, -40.1479529, -41.3389276, -42.3105735, -42.9900499, -43.3045162, -43.1811315, -42.5470552, -41.9427442, -40.4954043, -38.0122017, -34.3003027, -29.1668734, -24.4056033, -17.3405048, -7.28211317, 6.45903596, 24.5724072, 46.7507504, 75.8984515, 111.149647, 153.611632, 208.420237, 278.902564, 368.385712, 476.972705, 611.947467, 784.111096, 993.903249, 1261.89239, 1588.25271, 1988.41983, 2503.68906, 3163.4857, 3960.55689, 4914.81877, 6180.67404, 7836.4522, 9319.36572, 11447.4945, 14521.0931, 18840.4163, 24705.7185, 39993.8626, 55534.343, 69733.2622, 80996.7227, 87730.8271, 81707.9447, 72390.4001, 61027.3244, 52547.8012, 46744.3138, 41876.0315, 37812.1032, 34421.6778, 31573.9039, 29137.9303, 26552.8893, 24218.3982, 22118.1623, 20235.8866, 18555.2763, 17060.0366, 15733.8726, 14468.0681, 13802.1358},
        {66.0053166, 65.4326416, 64.843393, 64.2383484, 63.6182855, 62.9839822, 62.3362161, 61.6757651, 61.0034067, 60.3199188, 59.626079, 58.9226652, 58.2104549, 57.4902261, 56.7627564, 56.0288235, 55.2892052, 54.5446791, 53.7960232, 53.0440149, 52.2894322, 51.5330527, 50.7756542, 50.0180143, 49.2609109, 48.5051216, 47.7514242, 47.0005965, 46.2534161, 45.5106608, 44.7731083, 44.0415363, 43.3167227, 42.599445, 41.934605, 41.2999904, 40.6728585, 40.0525551, 39.4384261, 38.8298172, 38.2260744, 37.6265433, 37.0305699, 36.4375, 35.8466794, 35.257454, 34.6691694, 34.0811717, 33.4928066, 32.9034199, 32.3123574, 31.6833824, 31.0521947, 30.4189114, 29.7836497, 29.1465269, 28.5076601, 27.8671666, 27.2251634, 26.5817678, 25.9370969, 25.291268, 24.6443983, 23.9966049, 23.348005, 22.6987158, 22.0488545, 21.4000373, 20.7538445, 20.1073015, 19.4604284, 18.8132452, 18.165772, 17.5180287, 16.8700353, 16.2218119, 15.5733785, 14.9247551, 14.2759616, 13.6270182, 12.9779448, 12.3287614, 11.6794881, 11.0301448, 10.4311763, 9.85632968, 9.2787458, 8.69680836, 8.10890098, 7.51340729, 6.90871094, 6.29319555, 5.66524476, 5.02324221, 4.36557151, 3.69061632, 2.99676026, 2.28238696, 1.54588006, 0.785623199, 0, -1.06562529, -2.12669825, -3.1850093, -4.24234891, -5.30050752, -6.36127558, -7.42644353, -8.49780181, -9.57714089, -10.6662512, -11.7669232, -12.8809473, -14.010114, -15.1562137, -16.3210369, -17.506374, -18.6816647, -19.8440777, -21.0685781, -22.3500241, -23.6832737, -25.0631851, -26.4846163, -27.9424256, -29.431471, -30.9466107, -32.4827027, -34.0346052, -35.5971764, -37.2427523, -38.8603045, -40.4214477, -41.8977965, -43.4301276, -44.9054748, -46.165762, -47.1318554, -47.7246214, -47.8649259, -47.4736353, -47.0545506, -45.7681882, -43.4213617, -39.8208848, -34.7735711, -30.2745105, -23.3951715, -13.395299, 0.465362335, 18.9270678, 41.4166233, 71.5930605, 109.013426, 155.064849, 214.870872, 291.877466, 389.530601, 507.025708, 653.025326, 840.828946, 1071.95339, 1368.55179, 1729.1479, 2170.87994, 2742.42736, 3477.99168, 4369.43794, 5439.99154, 6868.11157, 8744.44816, 10154.2691, 12348.2811, 15765.1513, 20843.5466, 28022.1343, 47917.0559, 68245.1354, 86900.6713, 101777.962, 110771.306, 103058.187, 91060.9285, 76409.6053, 65557.2936, 58224.9653, 52124.0089, 47075.8756, 42902.0162, 39423.882, 36462.9239, 33178.0283, 30234.406, 27603.5953, 25257.1343, 23166.5613, 21303.4144, 19639.2318, 17920.9366, 16933.8891},
        {63.1686145, 62.3035731, 61.4654801, 60.653184, 59.8655334, 59.1013769, 58.359563, 57.6389402, 56.9383572, 56.2566624, 55.5927045, 54.9453319, 54.3133932, 53.6957371, 53.0912119, 52.4986664, 51.916949, 51.3449082, 50.7813928, 50.2252511, 49.6753318, 49.1304833, 48.5895544, 48.0513934, 47.514849, 46.9787698, 46.4420042, 45.9034008, 45.3618082, 44.816075, 44.2650496, 43.7075807, 43.1425167, 42.5687064, 41.9121771, 41.2097197, 40.4990001, 39.7812301, 39.0576214, 38.3293858, 37.5977349, 36.8638805, 36.1290344, 35.3944082, 34.6612138, 33.9306628, 33.203967, 32.482338, 31.7669878, 31.0591279, 30.3599701, 29.7743116, 29.1975331, 28.6286009, 28.0664811, 27.51014, 26.9585438, 26.4106586, 25.8654508, 25.3218864, 24.7789317, 24.2355529, 23.6907162, 23.1433878, 22.5925339, 22.0371207, 21.4761144, 20.8808014, 20.2231326, 19.5591618, 18.8896497, 18.2153567, 17.5370435, 16.8554706, 16.1713986, 15.4855881, 14.7987997, 14.111794, 13.4253314, 12.7401727, 12.0570783, 11.3768089, 10.7001251, 10.0277873, 9.46253731, 8.95200005, 8.44257483, 7.93171275, 7.41686491, 6.89548239, 6.3650163, 5.82291773, 5.26663778, 4.69362755, 4.10133813, 3.48722061, 2.8487261, 2.18330568, 1.48841046, 0.761491538, 0, -1.01329834, -2.05467429, -3.12344209, -4.21891601, -5.34041032, -6.48723927, -7.65871712, -8.85415813, -10.0728766, -11.3141867, -12.5774027, -13.861839, -15.1668097, -16.4916291, -17.8356116, -19.1980712, -20.6253648, -22.1235269, -23.5854663, -25.0205778, -26.4382562, -27.8478963, -29.2588928, -30.6806405, -32.1225342, -33.5939686, -35.1043386, -36.6630389, -38.2794643, -40.0999002, -41.9466575, -43.7880638, -45.5924465, -47.4522551, -49.2737561, -50.9301171, -52.3524291, -53.4717832, -54.2192703, -54.5259815, -55.0757017, -54.8177473, -53.535943, -51.0141136, -47.036084, -43.8959314, -38.2396649, -29.2235465, -16.0038378, 2.26319956, 25.330115, 56.4655108, 94.1158101, 139.505013, 199.528008, 278.533567, 380.870467, 507.547101, 667.011801, 871.356952, 1118.57017, 1438.08402, 1835.90811, 2332.68959, 2973.87806, 3793.53842, 4768.37, 5917.56513, 7470.65642, 9538.19612, 10867.0537, 13141.2616, 16943.416, 22856.1132, 31461.9495, 56057.2094, 81332.3789, 104691.632, 123539.144, 135279.087, 126750.848, 112966.582, 95858.2317, 82955.3352, 73974.118, 66298.4736, 59761.2441, 54195.2717, 49433.3986, 45308.467, 40246.6708, 36263.9353, 33119.992, 30574.5723, 28387.4076, 26318.2295, 24126.7694, 22079.3879, 20549.4139},
        {60.9968525, 60.0886385, 59.2176126, 58.3821511, 57.5806308, 56.8114283, 56.0729202, 55.363483, 54.6814934, 54.025328, 53.3933634, 52.7839762, 52.195543, 51.6264404, 51.0750451, 50.5397336, 50.0188825, 49.5108685, 49.0140681, 48.526858, 48.0476148, 47.574715, 47.1065353, 46.6414524, 46.1778427, 45.7140829, 45.2485496, 44.7796195, 44.305669, 43.8250749, 43.3362138, 42.8374622, 42.3271967, 41.803794, 41.1642652, 40.4598017, 39.7428132, 39.014966, 38.2779261, 37.5333596, 36.7829327, 36.0283114, 35.2711619, 34.5131503, 33.7559426, 33.0012051, 32.2506038, 31.5058049, 30.7684744, 30.0402784, 29.3228831, 28.7529911, 28.1943047, 27.6455626, 27.1055036, 26.5728666, 26.0463904, 25.5248136, 25.0068751, 24.4913138, 23.9768682, 23.4622774, 22.94628, 22.4276148, 21.9050207, 21.3772363, 20.8430006, 20.2671836, 19.6154687, 18.9564461, 18.2910502, 17.6202155, 16.9448762, 16.2659667, 15.5844213, 14.9011744, 14.2171604, 13.5333136, 12.8505684, 12.169859, 11.49212, 10.8182855, 10.1492901, 9.48606794, 8.94958697, 8.47824003, 8.00897728, 7.53883777, 7.06486052, 6.58408459, 6.09354899, 5.59029278, 5.07135498, 4.53377464, 3.97459078, 3.39084246, 2.77956869, 2.13780853, 1.462601, 0.750985144, 0, -0.991277573, -2.02291387, -3.09370956, -4.20246529, -5.34798172, -6.5290595, -7.74449929, -8.99310174, -10.2736675, -11.5849973, -12.9258917, -14.2951513, -15.691577, -17.1139692, -18.5611287, -20.0318561, -21.6000208, -23.2751474, -24.8862386, -26.4481809, -27.9758614, -29.4841666, -30.9879836, -32.502199, -34.0416996, -35.6213724, -37.256104, -38.9607813, -40.7502911, -42.7904121, -44.8898121, -47.0181105, -49.1449263, -51.3615716, -53.5768191, -55.6669905, -57.5651971, -59.2045505, -60.5181621, -61.4391433, -62.8316123, -63.4143239, -62.9382361, -61.154307, -57.8134944, -55.0126901, -49.5704348, -40.6512033, -27.4194699, -9.03970929, 13.8719585, 45.3769378, 83.7094448, 130.276205, 192.96115, 276.740079, 386.588789, 524.722057, 699.504209, 922.311595, 1188.5026, 1533.12812, 1964.18266, 2505.39921, 3207.17886, 4107.68161, 5178.34659, 6440.03578, 8158.25444, 10460.1238, 11794.9282, 14225.7545, 18460.5133, 25207.1154, 35173.4718, 63910.2924, 93571.9889, 121155.772, 143658.854, 158078.444, 149378.995, 134612.316, 115932.515, 101626.753, 91421.4712, 82487.6695, 74679.5604, 67851.3558, 61857.268, 56551.509, 50097.1305, 45080.5916, 41160.5979, 37995.8552, 35245.0694, 32566.946, 29620.1909, 26937.7104, 24671.3239},
        {58.9297241, 58.2171467, 57.5225013, 56.8449496, 56.1836532, 55.5377739, 54.9064732, 54.2889128, 53.6842543, 53.0916595, 52.51029, 51.9393073, 51.3778733, 50.8251495, 50.2802976, 49.7424792, 49.210856, 48.6845897, 48.1628419, 47.6447742, 47.1295484, 46.616326, 46.1042688, 45.5925383, 45.0802963, 44.5667043, 44.0509241, 43.5321174, 43.0094456, 42.4820706, 41.9491539, 41.4098572, 40.8633423, 40.3087706, 39.6879727, 39.0299816, 38.3645197, 37.6926091, 37.015272, 36.3335305, 35.6484068, 34.960923, 34.2721014, 33.5829641, 32.8945332, 32.2078309, 31.5238794, 30.8437009, 30.1683174, 29.4987513, 28.8360246, 28.2674027, 27.705795, 27.1503541, 26.6002325, 26.0545829, 25.5125577, 24.9733096, 24.4359912, 23.899755, 23.3637536, 22.8271396, 22.2890655, 21.748684, 21.2051477, 20.657609, 20.1052206, 19.5217714, 18.8816592, 18.2363474, 17.5866326, 16.9333118, 16.2771817, 15.6190391, 14.9596809, 14.2999039, 13.6405047, 12.9822804, 12.3260276, 11.6725431, 11.0226238, 10.3770665, 9.73666791, 9.10222492, 8.59325306, 8.14869265, 7.70605715, 7.2622907, 6.81433746, 6.35914157, 5.89364719, 5.41479848, 4.91953958, 4.40481465, 3.86756783, 3.30474329, 2.71328517, 2.09013763, 1.43224482, 0.736550895, 0, -1.09458325, -2.20413823, -3.33006327, -4.47375669, -5.63661683, -6.820042, -8.02543054, -9.25418076, -10.507691, -11.7873596, -13.0945848, -14.4307651, -15.7972986, -17.1955838, -18.627019, -20.0930025, -21.6326713, -23.2528152, -24.8694708, -26.4921234, -28.1302582, -29.7933605, -31.4909153, -33.232408, -35.0273237, -36.8851476, -38.815365, -40.827461, -42.9309208, -45.2448663, -47.628946, -50.059754, -52.5138847, -55.1274367, -57.797252, -60.377638, -62.7973376, -64.9850937, -66.8696492, -68.3797468, -70.6016068, -71.954219, -72.1398634, -70.86082, -67.8193686, -64.4832186, -58.3478632, -48.674225, -34.7232268, -15.755791, 6.78754148, 38.4894441, 77.6575559, 126.05242, 192.484259, 282.598134, 402.039104, 554.653115, 748.293232, 992.775194, 1280.43909, 1651.60903, 2112.72981, 2690.54987, 3446.221, 4425.1033, 5600.36774, 6998.65872, 8911.31404, 11479.9459, 12973.1565, 15693.7235, 20431.6089, 27976.7745, 39119.1823, 70744.5456, 103523.137, 134220.98, 159604.1, 176438.519, 168786.688, 154587.255, 136050.632, 121530.785, 110806.303, 101095.964, 92306.3473, 84344.0344, 77115.6057, 70527.6419, 62848.5673, 56732.1329, 51794.9837, 47653.7644, 43925.1198, 40225.6948, 36172.134, 32549.6335, 29264.6017},
        {57.0382514, 56.5986312, 56.1466751, 55.6828189, 55.2074982, 54.7211486, 54.2242059, 53.7171057, 53.2002838, 52.6741758, 52.1392174, 51.5958442, 51.044492, 50.4855964, 49.9195931, 49.3469179, 48.7680063, 48.183294, 47.5932168, 46.9982104, 46.3987103, 45.7951523, 45.1879721, 44.5776054, 43.9644877, 43.3490549, 42.7317426, 42.1129864, 41.4932221, 40.8728854, 40.2524118, 39.6322372, 39.0127972, 38.3945274, 37.7940709, 37.2034188, 36.6143674, 36.0268265, 35.4407057, 34.8559148, 34.2723634, 33.6899614, 33.1086185, 32.5282443, 31.9487486, 31.3700411, 30.7920315, 30.2146296, 29.637745, 29.0612876, 28.4851669, 27.9124622, 27.339845, 26.7671564, 26.1942373, 25.6209289, 25.0470721, 24.4725079, 23.8970774, 23.3206216, 22.7429814, 22.163998, 21.5835122, 21.0013652, 20.417398, 19.8314515, 19.2433668, 18.6424595, 18.0182979, 17.3924308, 16.7653817, 16.1376739, 15.5098306, 14.8823753, 14.2558313, 13.6307219, 13.0075705, 12.3869004, 11.7692349, 11.1550973, 10.5450111, 9.93949953, 9.33908592, 8.74429363, 8.26691067, 7.84948773, 7.43323821, 7.01507467, 6.59190972, 6.16065592, 5.71822587, 5.26153215, 4.78748735, 4.29300404, 3.77499482, 3.23037227, 2.65604897, 2.04893751, 1.40595047, 0.72400044, 0, -1.30151831, -2.57096434, -3.8139484, -5.03608079, -6.2429718, -7.44023174, -8.6334709, -9.8282996, -11.0303281, -12.2451668, -13.4784258, -14.7357157, -16.0226465, -17.3448287, -18.7078725, -20.1173882, -21.5464503, -22.9960212, -24.5453041, -26.1929372, -27.937559, -29.7778078, -31.7123219, -33.7397396, -35.8586994, -38.0678395, -40.3657983, -42.751214, -45.2227252, -47.8152888, -50.4779076, -53.1983242, -55.9642813, -59.0071187, -62.1927805, -65.3222517, -68.310196, -71.071277, -73.5201582, -75.5715033, -78.387236, -80.2551589, -80.8459064, -79.8301129, -76.8784129, -72.9777052, -66.1532938, -55.746747, -41.0996332, -21.5535207, 0.657508905, 32.6196944, 72.7878851, 123.289752, 193.751644, 290.433157, 419.593885, 586.097706, 797.917143, 1064.54732, 1377.01415, 1779.00617, 2273.11369, 2888.48185, 3699.25147, 4758.90958, 6045.34503, 7592.44534, 9711.29148, 12556.1681, 14315.649, 17412.1469, 22680.3846, 30955.0854, 43070.9722, 76566.6735, 111397.031, 144220.79, 171696.699, 190483.503, 184362.797, 171455.248, 153938.385, 139904.804, 129190.651, 119167.456, 109794.597, 101031.449, 92837.3868, 85171.7866, 75978.921, 68705.4137, 62846.0463, 57895.6006, 53348.8584, 48700.6012, 43445.6109, 38956.5235, 34431.1004},
        {55.3408179, 54.981749, 54.6020729, 54.202585, 53.7840803, 53.3473541, 52.8932017, 52.4224181, 51.9357985, 51.4341383, 50.9182326, 50.3888765, 49.8468653, 49.2929941, 48.7280582, 48.1528528, 47.568173, 46.9748141, 46.3735712, 45.7652396, 45.1506144, 44.5304908, 43.9056641, 43.2769294, 42.6450819, 42.0109169, 41.3752294, 40.7388148, 40.1024682, 39.4669848, 38.8331598, 38.2017885, 37.5736659, 36.9495873, 36.3691196, 35.8128568, 35.2609264, 34.7128656, 34.1682114, 33.6265006, 33.0872702, 32.5500573, 32.0143989, 31.4798318, 30.9458932, 30.4121198, 29.8780489, 29.3432172, 28.8071618, 28.2694197, 27.7295279, 27.16037, 26.5887141, 26.0146749, 25.4383673, 24.8599061, 24.2794058, 23.6969815, 23.1127477, 22.5268194, 21.9393112, 21.3503379, 20.7600143, 20.1684552, 19.5757753, 18.9820894, 18.3875123, 17.787031, 17.1757557, 16.5643767, 15.9533412, 15.3430963, 14.7340893, 14.1267673, 13.5215774, 12.918967, 12.3193831, 11.7232729, 11.1310836, 10.5432624, 9.96025648, 9.38251301, 8.81047916, 8.24460212, 7.80441652, 7.42832097, 7.05328273, 6.67588437, 6.29270844, 5.90033752, 5.49535415, 5.07434091, 4.63388035, 4.17055505, 3.68094754, 3.16164042, 2.60921622, 2.02025752, 1.39134687, 0.719066845, 0, -1.49570059, -2.92845702, -4.30656106, -5.63830453, -6.93197921, -8.19587689, -9.43828938, -10.6675085, -11.8918259, -13.1195336, -14.3589232, -15.6182866, -16.9059156, -18.2301019, -19.5991374, -21.0213138, -22.4369579, -23.8446521, -25.4064189, -27.115986, -28.9670812, -30.9534325, -33.0687678, -35.3068147, -37.6613012, -40.1259552, -42.6945044, -45.3606767, -48.1182, -50.9464287, -53.8587343, -56.8531567, -59.9277358, -63.416163, -67.1486524, -70.8659111, -74.4652835, -77.8441141, -80.8997472, -83.5295274, -86.769826, -89.0323002, -89.9914412, -89.3217403, -86.697689, -83.2795221, -76.8845515, -66.8158326, -52.3764209, -32.8693719, -10.7612274, 21.6749271, 62.7689001, 114.82928, 188.287996, 289.929765, 426.539301, 602.869449, 828.198584, 1114.02167, 1452.28351, 1888.53504, 2423.32608, 3087.69148, 3964.32138, 5112.19527, 6508.94986, 8192.47578, 10498.9207, 13595.7211, 15613.4105, 19069.6985, 24837.1034, 33788.1439, 46795.3385, 82042.4919, 118763.015, 153501.604, 182802.957, 203211.77, 198112.9, 185984.111, 169010.032, 155236.713, 144528.636, 134308.025, 124561.478, 115275.589, 106436.955, 98032.1708, 86847.3374, 78432.6015, 72011.7522, 66808.5786, 62046.8697, 56950.4146, 50743.0025, 45802.2026, 40047.5357},
        {53.6791712, 53.3803619, 53.0547029, 52.7032689, 52.3271347, 51.9273751, 51.5050647, 51.0612785, 50.5970911, 50.1135774, 49.611812, 49.0928698, 48.5578255, 48.0077539, 47.4437297, 46.8668278, 46.2781229, 45.6786897, 45.069603, 44.4519376, 43.8267683, 43.1951697, 42.5582168, 41.9169842, 41.2725468, 40.6259792, 39.9783563, 39.3307528, 38.6842434, 38.039903, 37.3988064, 36.7620282, 36.1306433, 35.5057264, 34.945473, 34.421196, 33.9035219, 33.3916717, 32.8848666, 32.3823276, 31.8832758, 31.3869323, 30.8925182, 30.3992545, 29.9063625, 29.413063, 28.9185774, 28.4221265, 27.9229316, 27.4202137, 26.9131939, 26.3485192, 25.7791245, 25.2053705, 24.6276181, 24.0462281, 23.4615612, 22.8739783, 22.2838401, 21.6915074, 21.0973411, 20.5017019, 19.9049505, 19.3074479, 18.7095548, 18.1116319, 17.5140401, 16.9170368, 16.3208818, 15.7261492, 15.1332065, 14.5424212, 13.9541607, 13.3687926, 12.7866844, 12.2082036, 11.6337176, 11.0635939, 10.4982001, 9.93790369, 9.38307209, 8.83407282, 8.29127338, 7.75504128, 7.35278071, 7.01866876, 6.68535566, 6.3490863, 6.00610561, 5.65265848, 5.28498984, 4.89934458, 4.49196762, 4.05910388, 3.59699825, 3.10189566, 2.57004101, 1.9976792, 1.38105516, 0.71641379, 0, -1.68525573, -3.27859237, -4.790844, -6.23284468, -7.6154285, -8.94942952, -10.2456818, -11.5150194, -12.7682765, -14.016287, -15.2698851, -16.5399048, -17.8371802, -19.1725453, -20.5568343, -22.0008812, -23.4166662, -24.8009012, -26.3880186, -28.1676695, -30.129505, -32.2631761, -34.5583339, -37.0046296, -39.5917141, -42.3092386, -45.1468541, -48.0942118, -51.1409627, -54.2178447, -57.3950236, -60.6798243, -64.0795719, -68.0241937, -72.3097135, -76.6194613, -80.8339811, -84.8338174, -88.4995143, -91.7116163, -95.363291, -98.0142695, -99.346974, -99.0438269, -96.7872504, -94.0438099, -88.2657489, -78.6894538, -64.5513112, -45.0877075, -22.8879529, 10.2625074, 52.5988259, 106.590594, 183.352714, 290.118521, 434.121351, 619.634141, 857.523439, 1161.88532, 1526.08161, 1997.22023, 2574.22501, 3289.92856, 4234.6201, 5472.21531, 6979.57986, 8798.0609, 11288.1891, 14629.9469, 16950.1148, 20797.7544, 27063.5716, 36638.2727, 50412.5638, 86894.049, 124952.312, 161073.834, 191745.097, 213452.583, 209441.387, 198266.967, 182090.685, 168812.889, 158323.257, 148120.225, 138213.032, 128610.919, 119323.128, 110358.899, 97567.6836, 88194.1391, 81251.7742, 75754.0983, 70714.6204, 65146.8497, 58064.2953, 52699.2217, 45800.097},
        {51.9804494, 51.7162513, 51.4220672, 51.0991201, 50.7486332, 50.3718297, 49.9699327, 49.5441653, 49.0957507, 48.625912, 48.1358723, 47.626855, 47.1000829, 46.5567794, 45.9981676, 45.4254705, 44.8399114, 44.2427134, 43.6350997, 43.0182933, 42.3935175, 41.7619954, 41.1249501, 40.4836047, 39.8391825, 39.1929066, 38.546, 37.8996861, 37.2551878, 36.6137284, 35.976531, 35.3448187, 34.7198147, 34.1027422, 33.5626695, 33.0654702, 32.5762022, 32.093887, 31.617546, 31.1462007, 30.6788726, 30.2145829, 29.7523532, 29.291205, 28.8301595, 28.3682384, 27.904463, 27.4378547, 26.967435, 26.4922254, 26.0112472, 25.4533496, 24.8892474, 24.3194835, 23.7446004, 23.1651407, 22.581647, 21.9946621, 21.4047286, 20.812389, 20.218186, 19.6226623, 19.0263605, 18.4298232, 17.833593, 17.2382127, 16.6442248, 16.0560089, 15.4778525, 14.9023848, 14.3298997, 13.760691, 13.1950528, 12.6332789, 12.0756631, 11.5224995, 10.9740818, 10.4307041, 9.89266015, 9.36024391, 8.83374926, 8.31347012, 7.79970037, 7.29273394, 6.92763004, 6.63446531, 6.3416966, 6.04524441, 5.74102925, 5.42497162, 5.09299204, 4.741011, 4.36494903, 3.96072661, 3.52426426, 3.05148249, 2.5383018, 1.9806427, 1.3744257, 0.715571295, 0, -1.83796438, -3.56222157, -5.18565882, -6.72116337, -8.18162245, -9.57992329, -10.9289531, -12.2415992, -13.5307488, -14.809289, -16.0901072, -17.3860906, -18.7101264, -20.0751019, -21.4939042, -22.9794207, -24.4267234, -25.831869, -27.474231, -29.3414507, -31.4211693, -33.7010278, -36.1686674, -38.8117292, -41.6178545, -44.5746843, -47.6698598, -50.8910221, -54.2258125, -57.5745642, -61.0442391, -64.6486706, -68.4016921, -72.8041262, -77.6263118, -82.5087241, -87.3190999, -91.9251757, -96.1946885, -99.995375, -104.1185, -107.227198, -109.008516, -109.1495, -107.337198, -105.41794, -100.379666, -91.3696015, -77.5349712, -58.0229999, -35.4635445, -1.26464833, 42.7829361, 99.3020403, 179.926563, 292.244618, 443.844318, 638.271747, 888.075692, 1210.21442, 1599.4584, 2104.72052, 2724.29874, 3492.91408, 4507.00448, 5834.6788, 7451.11753, 9400.33697, 12067.0909, 15643.3928, 18299.5928, 22560.3926, 29317.6121, 39463.0711, 53888.5894, 91158.0317, 130073.162, 167107.789, 198735.721, 221430.767, 218499.365, 208360.942, 193132.463, 180504.119, 170386.81, 160369.26, 150478.479, 140741.475, 131185.258, 121836.838, 108186.52, 98152.4805, 90675.6158, 84696.8229, 79156.9989, 72997.0406, 65157.8452, 59333.71, 51492.4318},
        {50.1717908, 49.9111986, 49.621668, 49.304388, 48.9605478, 48.5913364, 48.197943, 47.7815566, 47.3433662, 46.8845611, 46.4063303, 45.9098628, 45.3963478, 44.8669744, 44.3229316, 43.7654086, 43.1955943, 42.614678, 42.0238488, 41.4242956, 40.8172076, 40.2037739, 39.5851836, 38.9626257, 38.3372894, 37.7103637, 37.0830377, 36.4565006, 35.8319414, 35.2105492, 34.5935131, 33.9820222, 33.3772656, 32.7804323, 32.260248, 31.7827129, 31.3130157, 30.8501537, 30.3931243, 29.9409249, 29.492553, 29.0470058, 28.603281, 28.1603757, 27.7172875, 27.2730137, 26.8265517, 26.376899, 25.9230529, 25.4640109, 24.9987702, 24.451301, 23.8972544, 23.3372544, 22.7719247, 22.2018893, 21.627772, 21.0501966, 20.469787, 19.8871671, 19.3029606, 18.7177915, 18.1322836, 17.5470607, 16.9627466, 16.3799654, 15.7993407, 15.2274792, 14.670844, 14.1177198, 13.5683426, 13.0229485, 12.4817734, 11.9450534, 11.4130244, 10.8859225, 10.3639837, 9.84744393, 9.33653929, 8.83150576, 8.33257935, 7.83999607, 7.35399192, 6.87480291, 6.54459133, 6.28964481, 6.03454516, 5.77492262, 5.5064074, 5.22462975, 4.92521988, 4.60380803, 4.25602443, 3.87749931, 3.46386289, 3.0107454, 2.51377708, 1.96858815, 1.37080884, 0.71606938, 0, -1.92160722, -3.72019581, -5.40986717, -7.00472271, -8.51886381, -9.96639187, -11.3614083, -12.7180144, -14.0503118, -15.3724016, -16.6983854, -18.0423646, -19.4184404, -20.8407144, -22.3232879, -23.8802624, -25.408278, -26.9046561, -28.6491839, -30.6307925, -32.8384131, -35.2609769, -37.887415, -40.7066585, -43.7076387, -46.8792867, -50.2105336, -53.6903106, -57.307549, -60.9616146, -64.7638444, -68.730039, -72.875999, -77.728876, -83.0487952, -88.4595217, -93.8234508, -99.0029781, -103.860499, -108.258409, -112.986322, -116.697217, -119.072079, -119.791889, -118.537631, -117.549283, -113.309083, -104.858267, -91.238068, -71.4897213, -48.228879, -12.5536238, 33.8265043, 93.6919666, 178.990309, 297.553243, 457.212482, 660.662234, 922.039328, 1261.08514, 1673.46379, 2210.69469, 2872.03554, 3694.36903, 4778.33137, 6195.29486, 7917.44538, 9990.44026, 12823.6203, 16620.6063, 19635.6755, 24321.6914, 31557.0477, 42220.1382, 57189.3567, 94871.1268, 134233.805, 171773.777, 203987.428, 227371.146, 225437.942, 216323.159, 202087.488, 190181.192, 180531.589, 170820.338, 161087.039, 151371.293, 141712.699, 132150.855, 118750.408, 108470.08, 100392.78, 93601.4154, 87178.8933, 80208.1217, 71772.008, 65391.7972, 56928.1879},
        {48.1803337, 47.8869851, 47.5710074, 47.233322, 46.8748503, 46.4965136, 46.0992332, 45.6839306, 45.251527, 44.8029438, 44.3391024, 43.8609242, 43.3693304, 42.8652424, 42.3495817, 41.8232695, 41.2872272, 40.7423761, 40.1896376, 39.6299332, 39.064184, 38.4933115, 37.918237, 37.3398819, 36.7591675, 36.1770152, 35.5943463, 35.0120823, 34.4311443, 33.8524539, 33.2769323, 32.705501, 32.1390812, 31.5785943, 31.0777471, 30.6099579, 30.1480108, 29.691114, 29.2384759, 28.7893049, 28.3428094, 27.8981977, 27.4546783, 27.0114593, 26.5677493, 26.1227566, 25.6756896, 25.2257566, 24.772166, 24.3141261, 23.8508453, 23.3188133, 22.7813167, 22.2389236, 21.692202, 21.1417197, 20.588045, 20.0317456, 19.4733896, 18.913545, 18.3527797, 17.7916619, 17.2307593, 16.6706401, 16.1118722, 15.5550236, 15.0006623, 14.4549799, 13.9240328, 13.3967908, 12.8734575, 12.3542362, 11.8393304, 11.3289435, 10.823279, 10.3225403, 9.82693079, 9.33665394, 8.85191319, 8.37291196, 7.8998537, 7.43294183, 6.9723798, 6.51837103, 6.21929137, 5.99814144, 5.77614096, 5.54868486, 5.31116811, 5.05898567, 4.78753249, 4.49220352, 4.16839372, 3.81149804, 3.41691144, 2.98002888, 2.4962453, 1.96095567, 1.36955494, 0.717438066, 0, -1.90396492, -3.69336626, -5.38233067, -6.98498482, -8.51545535, -9.98786893, -11.4163522, -12.8150319, -14.1980345, -15.5794868, -16.9735155, -18.3942471, -19.8558084, -21.372326, -22.9579265, -24.6267366, -26.3024783, -27.9863633, -29.8970051, -32.0291577, -34.3775756, -36.9370128, -39.7022237, -42.6679624, -45.8289832, -49.1800403, -52.7158878, -56.4312802, -60.3209715, -64.3240232, -68.5113031, -72.8942731, -77.4843952, -82.7713586, -88.5275117, -94.3976759, -100.249845, -105.952013, -111.372174, -116.378322, -121.917625, -126.450459, -129.633674, -131.124121, -130.578652, -130.585209, -127.13678, -119.15744, -105.571269, -85.3023439, -60.9248332, -23.2515027, 26.2348043, 90.4887193, 181.524718, 307.289589, 475.730125, 688.685566, 961.598331, 1316.57362, 1749.1477, 2314.80151, 3015.92366, 3892.0144, 5045.45763, 6549.77259, 8372.44594, 10559.507, 13545.7714, 17546.1351, 20932.1937, 26045.7286, 33739.7013, 44867.0734, 60280.8068, 98070.0213, 137542.481, 175242.107, 207712.822, 231498.548, 230408.225, 222210.745, 208907.88, 197714.895, 188569.887, 179238.663, 169767.933, 160204.409, 150594.8, 140985.819, 129305.907, 119309.393, 110512.771, 102432.539, 94585.1922, 86487.227, 77655.1396, 70559.6126, 61911.0129},
        {45.5932658, 45.1466014, 44.7008794, 44.2559867, 43.8118099, 43.368236, 42.9251517, 42.4824437, 42.0399988, 41.5977039, 41.1554456, 40.7131109, 40.2705864, 39.8277589, 39.3845152, 38.9407422, 38.4963265, 38.051155, 37.6051144, 37.1580916, 36.7099732, 36.2606461, 35.8099971, 35.3579129, 34.9042804, 34.4489862, 33.9919173, 33.5329602, 33.072002, 32.6089292, 32.1436287, 31.6759874, 31.2058918, 30.733229, 30.2488147, 29.757262, 29.2632929, 28.7670885, 28.2688301, 27.7686986, 27.2668753, 26.7635414, 26.2588779, 25.753066, 25.2462868, 24.7387216, 24.2305514, 23.7219574, 23.2131208, 22.7042226, 22.1954441, 21.6906822, 21.1863218, 20.6824633, 20.1792074, 19.6766546, 19.1749055, 18.6740608, 18.1742209, 17.6754866, 17.1779583, 16.6817368, 16.1869225, 15.693616, 15.201918, 14.7119291, 14.2237498, 13.7355502, 13.2455466, 12.7578214, 12.2726001, 11.7901086, 11.3105726, 10.8342179, 10.3612701, 9.89195498, 9.42649835, 8.96512592, 8.50806342, 8.0555366, 7.60777119, 7.16499294, 6.72742758, 6.29530086, 6.02020839, 5.8235071, 5.62473459, 5.41920437, 5.20222992, 4.96912474, 4.71520234, 4.4357762, 4.12615983, 3.78166672, 3.39761037, 2.96930427, 2.49206193, 1.96119683, 1.37202248, 0.719852371, 0, -1.62166825, -3.18354118, -4.69663879, -6.17198106, -7.62058798, -9.05347956, -10.4816758, -11.9161966, -13.3680621, -14.8482922, -16.3679068, -17.9379261, -19.56937, -21.2732584, -23.0606114, -24.942449, -26.9383604, -29.06059, -31.3007745, -33.6717704, -36.1864339, -38.8576212, -41.6981887, -44.7209927, -47.9388893, -51.3647349, -55.0113858, -58.8916982, -63.0185284, -67.4869729, -72.1974929, -77.1382729, -82.297497, -87.9133726, -93.8490723, -99.9010961, -105.982621, -112.006825, -117.886885, -123.535978, -130.639963, -136.799824, -141.58191, -144.552569, -145.27815, -146.328207, -143.515081, -135.65432, -121.56147, -100.052079, -73.8993189, -33.1240154, 22.1477925, 95.3002064, 196.883863, 334.231773, 514.676943, 738.840115, 1022.29208, 1387.92611, 1826.23737, 2403.65346, 3134.89997, 4064.74799, 5288.879, 6879.60576, 8790.68907, 11063.706, 14174.9242, 18354.7258, 22143.5257, 27721.668, 35904.5235, 47507.4628, 63345.8568, 101156.452, 140602.899, 178270.225, 210743.456, 234607.62, 233859.064, 226063.947, 213202.445, 202421.874, 193657.084, 184689.945, 175562.956, 166318.613, 156999.413, 147647.853, 144632.1, 137130.292, 126560.78, 114341.913, 101892.04, 90629.5118, 81972.6767, 73776.7425, 66334.0678},
        {43.3985232, 42.9183858, 42.4470442, 41.9840353, 41.528896, 41.081163, 40.6403733, 40.2060638, 39.7777712, 39.3550325, 38.9373845, 38.524364, 38.115508, 37.7103532, 37.3084366, 36.909295, 36.5124652, 36.1174842, 35.7238888, 35.3312157, 34.939002, 34.5467845, 34.1540999, 33.7604852, 33.3654773, 32.968613, 32.5694291, 32.1674625, 31.7622501, 31.3533287, 30.9402352, 30.5225065, 30.0996794, 29.6712907, 29.2066647, 28.72108, 28.2301784, 27.7344774, 27.2344941, 26.7307459, 26.2237501, 25.714024, 25.2020851, 24.6884505, 24.1736376, 23.6581637, 23.1425462, 22.6273023, 22.1129494, 21.6000048, 21.0889858, 20.6076081, 20.1286011, 19.6518924, 19.1774098, 18.705081, 18.2348336, 17.7665955, 17.3002943, 16.8358577, 16.3732134, 15.9122892, 15.4530127, 14.9953117, 14.5391139, 14.0843469, 13.6309386, 13.1738428, 12.708133, 12.2439948, 11.7816784, 11.3214338, 10.8635114, 10.4081611, 9.9556332, 9.50617774, 9.06004489, 8.61748478, 8.17874756, 7.74408337, 7.31374234, 6.88797461, 6.46703032, 6.05115962, 5.79736952, 5.62423456, 5.44844515, 5.26516439, 5.06955533, 4.85678106, 4.62200466, 4.3603892, 4.06709775, 3.73729341, 3.36613923, 2.94879829, 2.48043368, 1.95620847, 1.37128574, 0.720828555, 0, -1.45241016, -2.8787161, -4.2883278, -5.69065522, -7.09510833, -8.51109708, -9.94803145, -11.4153214, -12.9223769, -14.4786079, -16.0934244, -17.7762363, -19.5364537, -21.3834864, -23.3267444, -25.3756378, -27.6082596, -30.0438316, -32.5364091, -35.1101199, -37.7890918, -40.5974525, -43.5593298, -46.6988515, -50.0401455, -53.6073394, -57.424561, -61.5159381, -65.9055986, -70.771022, -75.9267552, -81.3509205, -87.0216401, -93.1009918, -99.4751201, -105.981114, -112.541911, -119.080444, -125.51965, -131.782464, -139.801469, -146.878322, -152.542767, -156.324545, -157.753403, -159.768268, -157.637403, -150.038254, -135.648268, -113.144893, -85.0217559, -41.4759018, 16.9498561, 93.6322486, 199.950408, 343.690552, 532.638898, 767.447649, 1064.6584, 1448.59532, 1909.85185, 2517.51532, 3288.33576, 4268.91261, 5556.42356, 7224.83005, 9226.33242, 11603.3244, 14844.6572, 19186.6402, 23304.8968, 29236.5283, 37758.082, 49646.1048, 65677.144, 103033.641, 141984.775, 179205.619, 211371.247, 235156.733, 235023.643, 228002.896, 216003.926, 205910.807, 197663.845, 189130.579, 180354.952, 171380.907, 162252.39, 153013.343, 150057.026, 142504.923, 131799.463, 119383.076, 106698.193, 95187.2438, 86292.6587, 77064.0926, 69728.3018},
        {41.5158516, 41.1006997, 40.689225, 40.2811795, 39.876315, 39.4743833, 39.0751363, 38.6783257, 38.2837034, 37.8910211, 37.5000309, 37.1104843, 36.7221333, 36.3347297, 35.9480253, 35.561772, 35.1757215, 34.7896257, 34.4032364, 34.0163054, 33.6285846, 33.2398257, 32.8497807, 32.4582012, 32.0648392, 31.6694464, 31.2717748, 30.871576, 30.468602, 30.0626045, 29.6533354, 29.2405465, 28.8239896, 28.4034166, 27.9607412, 27.5050094, 27.0454818, 26.582489, 26.1163617, 25.6474306, 25.1760264, 24.7024798, 24.2271216, 23.7502823, 23.2722928, 22.7934837, 22.3141857, 21.8347295, 21.3554458, 20.8766653, 20.3987188, 19.9373149, 19.4770729, 19.0179903, 18.5600643, 18.1032922, 17.6476714, 17.1931993, 16.7398731, 16.2876902, 15.836648, 15.3867437, 14.9379748, 14.4903385, 14.0438321, 13.5984531, 13.1541987, 12.7067337, 12.2518267, 11.7984108, 11.3467644, 10.8971654, 10.4498922, 10.005223, 9.56343591, 9.12480918, 8.68962102, 8.25814962, 7.83067319, 7.40746992, 6.98881803, 6.57499571, 6.16628116, 5.7629526, 5.52977149, 5.38131473, 5.23018223, 5.07131438, 4.8996516, 4.71013428, 4.49770283, 4.25729766, 3.98385916, 3.67232774, 3.31764381, 2.91474777, 2.45858002, 1.94408096, 1.366191, 0.719850548, 0, -1.39813641, -2.78255867, -4.16257762, -5.54750412, -6.94664902, -8.36932319, -9.82483748, -11.3225027, -12.8716298, -14.4815296, -16.1615129, -17.9208906, -19.7689736, -21.7150727, -23.7684987, -25.9385625, -28.3257449, -30.9523808, -33.6215637, -36.3621409, -39.2029597, -42.1728672, -45.3007106, -48.6153374, -52.1455946, -55.9203295, -59.9683894, -64.3186216, -68.9998732, -74.2054471, -79.7394345, -85.5813461, -91.7106923, -98.3561385, -105.372618, -112.548622, -119.788916, -126.998263, -134.081428, -140.943174, -149.371318, -156.814469, -162.808969, -166.891158, -168.597378, -171.495843, -170.083052, -162.887379, -148.437196, -125.260876, -95.3116485, -49.5070933, 10.0730887, 86.5354773, 193.575308, 340.133727, 535.15188, 779.917493, 1092.76454, 1500.37619, 1997.77132, 2649.96479, 3467.28697, 4495.42793, 5839.99268, 7579.08428, 9673.45239, 12170.7751, 15549.0264, 20042.7805, 24428.2335, 30622.7037, 39365.5887, 51396.2859, 67454.1929, 104060.547, 142227.446, 178748.827, 210418.628, 234030.786, 234680.165, 228659.16, 217775.597, 208525.371, 200860.616, 192780.103, 184333.634, 175571.012, 166542.037, 157296.511, 148457.518, 139093.561, 129375.818, 119475.465, 109563.678, 99811.6358, 90390.5136, 80382.1402, 72267.4788},
        {39.7446927, 39.4430163, 39.133607, 38.8167032, 38.4925435, 38.1613667, 37.8234111, 37.4789156, 37.1281185, 36.7712586, 36.4085743, 36.0403044, 35.6666873, 35.2879616, 34.9043661, 34.5161392, 34.1235194, 33.7267456, 33.3260561, 32.9216896, 32.5138847, 32.1028799, 31.6889139, 31.2722252, 30.8530525, 30.4316343, 30.0082092, 29.5830158, 29.1562926, 28.7282783, 28.2992115, 27.8693307, 27.4388746, 27.0080817, 26.5875428, 26.1721028, 25.7564818, 25.3405825, 24.9243075, 24.5075595, 24.0902412, 23.672255, 23.2535038, 22.83389, 22.4133164, 21.9916856, 21.5689003, 21.1448629, 20.7194763, 20.292643, 19.8642657, 19.4217878, 18.9778412, 18.5325987, 18.086233, 17.6389167, 17.1908228, 16.7421238, 16.2929925, 15.8436016, 15.3941238, 14.944732, 14.4955988, 14.0468968, 13.598799, 13.151478, 12.7051064, 12.2576137, 11.8069831, 11.3580141, 10.9110248, 10.4663335, 10.0242582, 9.58511717, 9.14922854, 8.71691047, 8.2884811, 7.86425861, 7.44456114, 7.02970686, 6.62001392, 6.21580048, 5.8173847, 5.42508474, 5.21331085, 5.09167336, 4.96769011, 4.83602965, 4.6913605, 4.52835121, 4.34167031, 4.12598632, 3.87596779, 3.58628325, 3.25160123, 2.86659027, 2.42591889, 1.92425565, 1.35626906, 0.716627666, 0, -1.41614773, -2.81738934, -4.21391703, -5.61592298, -7.03359938, -8.47713842, -9.95673229, -11.4825732, -13.0648533, -14.7137647, -16.4394998, -18.2522507, -20.1622094, -22.1795684, -24.3145197, -26.5772555, -29.0682841, -31.8107002, -34.6104092, -37.4969569, -40.4998889, -43.6487509, -46.9730885, -50.5024473, -54.2663731, -58.2944114, -62.6161079, -67.2610083, -72.2586583, -77.7812596, -83.6633944, -89.8918116, -96.4532599, -103.706209, -111.451548, -119.391038, -127.399912, -135.353404, -143.126745, -150.595167, -159.240758, -166.842853, -172.9623, -177.159947, -178.996643, -182.706538, -182.008852, -175.296105, -160.96082, -137.395518, -106.063735, -58.5335188, 1.47514655, 76.7976582, 183.653318, 332.25233, 532.804894, 786.985556, 1115.4801, 1548.28593, 2087.15314, 2789.68584, 3655.55608, 4728.24716, 6125.47813, 7930.89259, 10119.3202, 12746.1936, 16265.9299, 20907.2578, 25520.4171, 31924.7012, 40826.0083, 52930.2365, 68943.2843, 104740.321, 142065.655, 177832.869, 208955.541, 232347.253, 233741.829, 228685.776, 218883.999, 210436.293, 203307.919, 195632.931, 187468.146, 178870.386, 169896.468, 160603.212, 144515.389, 132945.511, 124492.777, 117756.388, 111335.543, 103829.44, 93837.2792, 83441.0431, 74174.6163},
        {37.6880581, 37.4607203, 37.2196007, 36.9651831, 36.6979516, 36.4183901, 36.1269824, 35.8242127, 35.5105648, 35.1865227, 34.8525704, 34.5091917, 34.1568707, 33.7960913, 33.4273374, 33.0510931, 32.6678422, 32.2780687, 31.8822566, 31.4808898, 31.0744523, 30.663428, 30.2483009, 29.8295549, 29.407674, 28.9831422, 28.5564433, 28.1280613, 27.6984803, 27.2681841, 26.8376567, 26.4073821, 25.9778441, 25.5495269, 25.1456564, 24.7548672, 24.3655205, 23.9773622, 23.5901382, 23.2035944, 22.8174769, 22.4315314, 22.0455039, 21.6591404, 21.2721868, 20.884389, 20.4954928, 20.1052444, 19.7133895, 19.3196741, 18.9238441, 18.5052402, 18.084456, 17.6616796, 17.2370994, 16.8109037, 16.3832806, 15.9544185, 15.5245057, 15.0937303, 14.6622808, 14.2303453, 13.7981121, 13.3657695, 12.9335058, 12.5015093, 12.0699681, 11.6357813, 11.1959266, 10.7573764, 10.3205323, 9.88579572, 9.45356822, 9.02425129, 8.59824643, 8.17595517, 7.75777899, 7.34411942, 6.93537796, 6.53195612, 6.1342554, 5.74267732, 5.35762338, 4.97949509, 4.79409018, 4.70521248, 4.61443753, 4.51615035, 4.40473599, 4.2745795, 4.12006593, 3.9355803, 3.71550768, 3.4542331, 3.14614161, 2.78561826, 2.36704808, 1.88481612, 1.33330742, 0.706907034, 0, -1.40955265, -2.80589767, -4.20001709, -5.60289292, -7.02550719, -8.47884192, -9.97387914, -11.5216009, -13.1329891, -14.8190259, -16.5906933, -18.4589733, -20.4348479, -22.5292992, -24.7533091, -27.1178597, -29.7247609, -32.5979702, -35.5430254, -38.5903719, -41.7704547, -45.1137189, -48.6506097, -52.4115723, -56.4270518, -60.7274932, -65.3433419, -70.3050428, -75.6430411, -81.5217708, -87.788559, -94.4336542, -101.447305, -109.235284, -117.580078, -126.153366, -134.82074, -143.447791, -151.900111, -160.04329, -169.280555, -177.471887, -184.182033, -188.975744, -191.417769, -196.142385, -196.377432, -190.420276, -176.568284, -153.118824, -121.838159, -73.3149971, -10.1174257, 71.0383061, 185.38345, 342.785075, 553.110247, 814.399218, 1151.03379, 1600.29773, 2172.11059, 2916.77606, 3823.5491, 4934.26563, 6379.85723, 8250.3591, 10534.2872, 13295.3568, 16969.5209, 21782.1737, 26629.7753, 33271.1058, 42391.5751, 54676.5934, 70811.5706, 106394.463, 143469.998, 178995.448, 209928.089, 233225.193, 234834.884, 230063.018, 220585.077, 212383.826, 205429.779, 197898.748, 189850.983, 181346.733, 172446.246, 163209.774, 150537.596, 139931.304, 130749.758, 122351.815, 114096.336, 105342.18, 95448.2043, 85671.7532, 75927.6296},
        {35.468434, 35.2882626, 35.092123, 34.8805719, 34.6541661, 34.4134626, 34.1590182, 33.8913897, 33.6111339, 33.3188077, 33.0149679, 32.7001714, 32.374975, 32.0399355, 31.6956098, 31.3425547, 30.9813271, 30.6124838, 30.2365816, 29.8541774, 29.465828, 29.0720903, 28.6735211, 28.2706772, 27.8641155, 27.4543928, 27.0420659, 26.6276918, 26.2118272, 25.7950289, 25.3778539, 24.9608589, 24.5446008, 24.1296364, 23.741038, 23.366589, 22.9937783, 22.6223672, 22.2521169, 21.8827887, 21.5141439, 21.1459437, 20.7779495, 20.4099224, 20.0416238, 19.6728149, 19.303257, 18.9327114, 18.5609393, 18.187702, 17.8127607, 17.4193663, 17.0241484, 16.6272262, 16.2287189, 15.8287457, 15.4274257, 15.0248782, 14.6212223, 14.2165773, 13.8110622, 13.4047963, 12.9978988, 12.5904889, 12.1826857, 11.7746084, 11.3663762, 10.9522617, 10.5266778, 10.1018019, 9.67813226, 9.256167, 8.83640436, 8.41934253, 8.00547972, 7.59531412, 7.18934394, 6.78806736, 6.3919826, 6.00158784, 5.61738129, 5.23986114, 4.86952561, 4.50687287, 4.34989609, 4.29619195, 4.2409834, 4.17835953, 4.10240946, 4.00722229, 3.88688714, 3.73549311, 3.5471293, 3.31588483, 3.03584881, 2.70111033, 2.30575852, 1.84388247, 1.30957129, 0.696914103, 0, -1.42477514, -2.83425497, -4.24047987, -5.65549022, -7.09132639, -8.56002874, -10.0736377, -11.6441935, -13.2837367, -15.0043075, -16.8179464, -18.7366938, -20.7725899, -22.9376752, -25.24399, -27.7035748, -30.4183269, -33.4131235, -36.4967568, -39.7005223, -43.0557154, -46.5936315, -50.3455663, -54.342815, -58.6166731, -63.1984362, -68.1193997, -73.410859, -79.1041095, -85.3648727, -92.0407286, -99.1226449, -106.601589, -114.877538, -123.736956, -132.857235, -142.106641, -151.353439, -160.465894, -169.31227, -179.415647, -188.485836, -196.067336, -201.704643, -204.942255, -210.455779, -211.375824, -205.964111, -192.482359, -169.192288, -138.559444, -89.5037115, -22.7223132, 65.7435755, 189.358879, 357.320475, 578.825245, 848.206849, 1192.90339, 1656.56739, 2254.80723, 3034.04048, 3977.09997, 5125.4896, 6620.12808, 8556.3011, 10933.4506, 13822.014, 17643.8554, 22622.0521, 27695.0274, 34568.1516, 43909.1896, 56385.9067, 72666.0679, 108180.285, 145142.765, 180530.561, 211320.726, 234490.315, 236055.283, 231261.18, 221782.67, 213602.062, 206688.915, 199217.935, 191248.422, 182839.678, 174051.002, 164941.696, 158665.81, 149948.463, 139541.012, 128194.812, 116661.22, 105691.591, 96037.2813, 87243.6036, 77330.3376},
        {33.1454633, 32.9844054, 32.8092957, 32.6205839, 32.4187199, 32.2041536, 31.9773349, 31.7387137, 31.4887397, 31.227863, 30.9565333, 30.6752006, 30.3843147, 30.0843255, 29.7756828, 29.4588366, 29.1342367, 28.802333, 28.4635753, 28.1184135, 27.7672976, 27.4106773, 27.0490026, 26.6827233, 26.3122893, 25.9381504, 25.5607566, 25.1805577, 24.7980036, 24.4135442, 24.0276293, 23.6407088, 23.2532326, 22.8656505, 22.4946485, 22.1322169, 21.770151, 21.4083736, 21.0468078, 20.6853766, 20.324003, 19.9626099, 19.6011203, 19.2394572, 18.8775436, 18.5153025, 18.1526568, 17.7895296, 17.4258438, 17.0615224, 16.6964884, 16.3269479, 15.9566214, 15.5855124, 15.2136244, 14.840961, 14.4675257, 14.0933221, 13.7183538, 13.3426243, 12.9661372, 12.588896, 12.2109042, 11.8321655, 11.4526834, 11.0724614, 10.6915031, 10.300903, 9.89196948, 9.48308034, 9.07481671, 8.66775969, 8.26249037, 7.85958984, 7.4596392, 7.06321955, 6.67091198, 6.28329758, 5.90095744, 5.52447267, 5.15442436, 4.7913936, 4.43596149, 4.08870911, 3.9596139, 3.94015441, 3.91929242, 3.89081371, 3.84850407, 3.78614928, 3.69753512, 3.57644737, 3.41667181, 3.21199422, 2.95620039, 2.64307609, 2.26640711, 1.81997923, 1.29757823, 0.692989896, 0, -1.50395465, -2.98080444, -4.44422663, -5.90789852, -7.38549737, -8.89070045, -10.437185, -12.0386284, -13.7087079, -15.4611006, -17.309484, -19.2675353, -21.3489317, -23.5673505, -25.9364691, -28.4699646, -31.2685967, -34.358483, -37.5563288, -40.8944719, -44.4052502, -48.1210013, -52.074063, -56.2967732, -60.8214695, -65.6804897, -70.9061717, -76.5308531, -82.5868717, -89.2534454, -96.3621759, -103.901337, -111.859202, -120.582891, -129.881254, -139.46745, -149.222101, -159.025825, -168.759243, -178.302975, -189.541763, -199.742155, -208.392662, -214.981792, -218.998055, -224.602728, -225.443362, -219.840273, -206.113779, -182.584198, -152.533986, -103.256486, -34.6760711, 57.6492396, 187.076064, 362.958529, 594.650762, 876.478122, 1236.23914, 1718.12735, 2336.54992, 3141.56027, 4118.34044, 5310.46815, 6860.46941, 8865.45136, 11329.525, 14325.6875, 18267.9213, 23377.3285, 28654.2761, 35728.4062, 45243.3826, 57842.8694, 74170.5304, 109377.756, 145973.551, 180974.65, 211397.784, 234259.687, 235728.262, 230901.622, 221445.996, 213331.853, 206521.662, 199184.457, 191372.38, 183137.57, 174532.166, 165608.311, 162126.392, 154316.514, 143490.479, 130960.089, 118037.144, 106033.448, 96260.8008, 88253.6243, 78294.0813},
        {30.7355012, 30.5557814, 30.3694662, 30.1766586, 29.9774619, 29.7719793, 29.5603138, 29.3425687, 29.1188471, 28.8892522, 28.6538872, 28.4128553, 28.1662595, 27.9142032, 27.6567894, 27.3941214, 27.1263023, 26.8534352, 26.5756234, 26.29297, 26.0055782, 25.7135511, 25.416992, 25.116004, 24.8106902, 24.5011539, 24.1874982, 23.8698263, 23.5482414, 23.2228465, 22.893745, 22.5610399, 22.2248345, 21.8852319, 21.5376231, 21.1845665, 20.828677, 20.4702108, 20.109424, 19.7465727, 19.3819129, 19.0157008, 18.6481923, 18.2796437, 17.910311, 17.5404503, 17.1703177, 16.8001693, 16.4302612, 16.0608494, 15.69219, 15.3430942, 14.9948607, 14.6473435, 14.3003964, 13.9538733, 13.607628, 13.2615143, 12.9153862, 12.5690974, 12.2225019, 11.8754535, 11.527806, 11.1794134, 10.8301294, 10.4798079, 10.1283027, 9.76357057, 9.37385359, 8.98354287, 8.59326353, 8.20364067, 7.8152994, 7.42886482, 7.04496204, 6.66421616, 6.2872523, 5.91469556, 5.54717104, 5.18530385, 4.8297191, 4.4810419, 4.13989734, 3.80691055, 3.70275343, 3.71339929, 3.72217667, 3.72256971, 3.70806259, 3.67213946, 3.6082845, 3.50998187, 3.37071573, 3.18397025, 2.94322958, 2.6419779, 2.27369936, 1.83187814, 1.3099984, 0.701544298, 0, -1.68617805, -3.31831038, -4.91256437, -6.48510739, -8.05210684, -9.62973008, -11.2341445, -12.8815175, -14.5880164, -16.3698086, -18.2430615, -20.2239425, -22.328619, -24.5732583, -26.9740277, -29.5470948, -32.3925556, -35.5385674, -38.8114585, -42.2453813, -45.8744877, -49.73293, -53.8548601, -58.2744302, -63.0257925, -68.143099, -73.6605019, -79.6121532, -86.0322052, -93.1338464, -100.706546, -108.730744, -117.186883, -126.312121, -135.958542, -145.909767, -156.069224, -166.340337, -176.626532, -186.831234, -199.534692, -211.148823, -221.053327, -228.627904, -233.252254, -237.749862, -237.195696, -230.108507, -215.007049, -190.410075, -160.336363, -111.082388, -44.5354804, 43.6705953, 170.669768, 347.863534, 586.653386, 888.696622, 1277.35353, 1786.61442, 2418.05443, 3237.54025, 4246.84757, 5495.35346, 7113.09283, 9193.07729, 11733.5596, 14802.9843, 18817.2451, 23995.6989, 29445.3071, 36668.9597, 46271.0732, 58856.0645, 75028.3503, 109352.411, 144982.583, 179033.269, 208618.868, 230853.781, 232339.46, 227711.888, 218591.015, 210820.07, 204347.289, 197361.211, 189898.057, 181994.049, 173685.411, 165008.363, 154903.38, 145347.845, 136080.44, 126839.849, 117364.752, 107393.832, 96665.7725, 88748.9103, 78804.0636},
        {28.6653696, 28.5095938, 28.3486249, 28.1824777, 28.0111669, 27.8347072, 27.6531133, 27.4663998, 27.2745815, 27.0776731, 26.8756892, 26.6686445, 26.4565538, 26.2394317, 26.0172929, 25.7901521, 25.558024, 25.3209234, 25.0788648, 24.831863, 24.5799327, 24.3230886, 24.0613453, 23.7947176, 23.5232202, 23.2468677, 22.9656748, 22.6796563, 22.3888267, 22.093201, 21.7927936, 21.4876193, 21.1776928, 20.8630289, 20.5328098, 20.1926942, 19.8484711, 19.5005067, 19.1491672, 18.7948188, 18.4378277, 18.0785602, 17.7173826, 17.3546609, 16.9907615, 16.6260505, 16.2608943, 15.8956589, 15.5307107, 15.1664158, 14.8031405, 14.4660861, 14.1302452, 13.7954459, 13.461516, 13.1282832, 12.7955755, 12.4632207, 12.1310466, 11.7988811, 11.4665521, 11.1338873, 10.8007147, 10.4668621, 10.1321573, 9.79642816, 9.45950258, 9.10809022, 8.72921571, 8.34976214, 7.97040777, 7.59183082, 7.21470956, 6.83972223, 6.46754706, 6.09886232, 5.73434624, 5.37467708, 5.02053307, 4.67259247, 4.33153351, 3.99803445, 3.67277354, 3.35642901, 3.28243622, 3.33019913, 3.37632414, 3.41393608, 3.43615979, 3.43612008, 3.4069418, 3.34174978, 3.23366884, 3.07582381, 2.86133953, 2.58334082, 2.23495252, 1.80929946, 1.29950646, 0.698698368, 0, -1.75106259, -3.441761, -5.08940933, -6.71132167, -8.3248121, -9.94719473, -11.5957836, -13.2878929, -15.0408367, -16.871929, -18.7984839, -20.8378156, -23.0072381, -25.3240655, -27.8056119, -30.4691914, -33.4192945, -36.6856891, -40.0885044, -43.6637354, -47.4473767, -51.4754231, -55.7838693, -60.4087101, -65.3859402, -70.7515543, -76.5415472, -82.7919137, -89.5386485, -96.9964814, -104.957136, -113.402986, -122.316406, -131.951784, -142.155485, -152.701343, -163.490128, -174.42261, -185.39956, -196.321748, -210.015147, -222.565047, -233.299895, -241.54814, -246.638229, -251.309832, -250.62737, -243.066488, -227.102828, -201.212034, -169.483392, -117.917781, -49.2756043, 40.8180145, 171.305813, 354.589857, 603.07221, 922.690866, 1333.78107, 1863.72978, 2503.99824, 3334.02425, 4368.37605, 5661.85135, 7337.41448, 9486.74018, 12108.0076, 15268.652, 19378.9254, 24655.1633, 30305.3784, 37709.3554, 47432.4644, 60040.0756, 76097.559, 109595.31, 144317.418, 177472.996, 206271.157, 227921.015, 229449.6, 225036.166, 216253.578, 208804.083, 202625.998, 195937.066, 188761.634, 181124.047, 173048.652, 164559.797, 152610.452, 142492.524, 133569.863, 125206.318, 116765.738, 107611.973, 97108.8724, 88785.5659, 77156.924},
        {26.9443745, 26.8663857, 26.7768867, 26.6761355, 26.5643899, 26.441908, 26.3089475, 26.1657665, 26.0126228, 25.8497743, 25.6774789, 25.4959946, 25.3055793, 25.1064908, 24.8989871, 24.6833261, 24.4597657, 24.2285638, 23.9899784, 23.7442672, 23.4916884, 23.2324996, 22.966959, 22.6953243, 22.4178534, 22.1348044, 21.8464351, 21.5530034, 21.2547672, 20.9519845, 20.6449131, 20.333811, 20.018936, 19.7005461, 19.3806333, 19.058552, 18.7336357, 18.406086, 18.0761047, 17.7438932, 17.4096534, 17.0735867, 16.7358948, 16.3967794, 16.0564421, 15.7150845, 15.3729083, 15.0301151, 14.6869066, 14.3434843, 14.0000499, 13.669414, 13.3388958, 13.0084239, 12.6779264, 12.3473316, 12.0165678, 11.6855634, 11.3542467, 11.0225459, 10.6903893, 10.3577053, 10.0244221, 9.69046801, 9.35577139, 9.02026052, 8.68386368, 8.33387185, 7.9578794, 7.58187819, 7.20661575, 6.8328396, 6.46129727, 6.09273629, 5.72790417, 5.36754846, 5.01241667, 4.66325633, 4.32081497, 3.98584012, 3.65907929, 3.34128002, 3.03318984, 2.73555626, 2.69658262, 2.78808364, 2.87889982, 2.96174827, 3.02934611, 3.07441043, 3.08965836, 3.06780701, 3.00157347, 2.88367487, 2.70682831, 2.4637509, 2.14715975, 1.74977198, 1.26430469, 0.683474995, 0, -1.68203593, -3.32030864, -4.93182345, -6.53358565, -8.14260055, -9.77587346, -11.4504097, -13.1832145, -14.9912933, -16.8916514, -18.9012939, -21.0372263, -23.3164539, -25.7559819, -28.3728158, -31.1839606, -34.3015112, -37.7560568, -41.3495201, -45.1192825, -49.1027257, -53.3372312, -57.8601805, -62.7089554, -67.9209373, -73.5335078, -79.5840485, -86.109941, -93.1485669, -100.879744, -109.146191, -117.942558, -127.263497, -137.514902, -148.485803, -159.855561, -171.495456, -183.276764, -195.070762, -206.748729, -220.858698, -233.780525, -244.861775, -253.450012, -258.892801, -265.257617, -265.992135, -259.263944, -243.240628, -216.089775, -181.339394, -125.245022, -49.7348004, 49.78239, 191.207832, 386.466815, 647.484632, 979.428734, 1402.51812, 1944.34399, 2592.92094, 3434.58118, 4488.25256, 5812.26818, 7531.73123, 9741.09632, 12446.5401, 15719.6304, 19955.6207, 25366.6697, 31247.8661, 38869.5261, 48761.2553, 61452.6596, 77473.3449, 110305.067, 144287.243, 176711.444, 204869.236, 226052.188, 227667.577, 223480.791, 215019.522, 207824.78, 201835.652, 195319.016, 188297.571, 180794.017, 172831.056, 164431.388, 156959.624, 147930.794, 137862.401, 127271.953, 116676.954, 106594.911, 97543.3286, 88478.7336, 73934.6296},
        {24.9812298, 24.9834837, 24.9675051, 24.9338189, 24.8829502, 24.8154238, 24.7317647, 24.6324977, 24.518148, 24.3892403, 24.2462997, 24.089851, 23.9204192, 23.7385293, 23.5447061, 23.3394746, 23.1233598, 22.8968866, 22.6605799, 22.4149646, 22.1605658, 21.8979082, 21.627517, 21.3499169, 21.065633, 20.7751902, 20.4791134, 20.1779275, 19.8721575, 19.5623284, 19.248965, 18.9325923, 18.6137353, 18.2929188, 17.9851194, 17.6833322, 17.3803789, 17.0763152, 16.7711971, 16.4650807, 16.1580219, 15.8500766, 15.5413008, 15.2317504, 14.9214814, 14.6105498, 14.2990115, 13.9869224, 13.6743386, 13.3613159, 13.0479104, 12.7396315, 12.4309635, 12.121844, 11.8122108, 11.5020017, 11.1911542, 10.8796063, 10.5672956, 10.2541598, 9.94013671, 9.62516402, 9.30917948, 8.99212082, 8.67392576, 8.35453205, 8.03387741, 7.69836092, 7.33470672, 6.970775, 6.60738115, 6.24534055, 5.8854686, 5.52858068, 5.17549218, 4.82701849, 4.483975, 4.14717709, 3.81744015, 3.49557958, 3.18241076, 2.87874908, 2.58540993, 2.30320869, 2.29291248, 2.41989309, 2.54639789, 2.66480634, 2.76749789, 2.846852, 2.89524812, 2.9050657, 2.86868421, 2.77848309, 2.6268418, 2.4061398, 2.10875654, 1.72707147, 1.25346406, 0.680313749, 0, -1.67334285, -3.30884859, -4.92398155, -6.53620608, -8.16298653, -9.82178724, -11.5300726, -13.3053069, -15.1649545, -17.1264797, -19.2073469, -21.4250205, -23.7969648, -26.3406441, -29.0735228, -32.0130653, -35.2684787, -38.8703334, -42.6285803, -46.5803429, -50.7627448, -55.2129093, -59.9679601, -65.0650208, -70.5412147, -76.4336655, -82.7794967, -89.6158318, -96.9797943, -105.031905, -113.640645, -122.806117, -132.528428, -143.275481, -154.813481, -166.783884, -179.046455, -191.460959, -203.887158, -216.184818, -230.819538, -244.252167, -255.83263, -264.910853, -270.836763, -278.377516, -280.111502, -274.034339, -258.141647, -230.429045, -194.395938, -135.807311, -55.3188249, 52.0398426, 202.725399, 408.0374, 679.275401, 1018.62592, 1448.5727, 2001.54095, 2669.43591, 3536.36112, 4618.24076, 5973.3738, 7731.86084, 9990.84436, 12763.2794, 16124.482, 20458.4556, 25974.5926, 32058.122, 39869.394, 49903.0882, 62653.8843, 78616.4621, 110745.947, 143961.462, 175642.574, 203168.852, 223919.865, 225740.603, 221901.596, 213868.66, 206896.858, 200952.095, 194421.423, 187349.689, 179781.74, 171762.424, 163336.589, 158295.264, 150070.198, 139700.063, 128223.533, 116679.281, 106105.98, 97542.3027, 88162.4369, 74083.4278},
        {22.7887268, 22.8562649, 22.900777, 22.9229781, 22.9235829, 22.9033065, 22.8628637, 22.8029695, 22.7243388, 22.6276864, 22.5137274, 22.3831765, 22.2367488, 22.0751591, 21.8991224, 21.7093535, 21.5065674, 21.2914789, 21.0648031, 20.8272548, 20.5795489, 20.3224003, 20.056524, 19.7826348, 19.5014477, 19.2136776, 18.9200394, 18.621248, 18.3180183, 18.0110653, 17.7011038, 17.3888488, 17.0750151, 16.7603177, 16.4674927, 16.1857813, 15.9041598, 15.6226285, 15.3411878, 15.0598379, 14.7785791, 14.4974117, 14.216336, 13.9353522, 13.6544607, 13.3736616, 13.0929554, 12.8123423, 12.5318225, 12.2513964, 11.9710642, 11.6979679, 11.4248113, 11.1514399, 10.877699, 10.6034343, 10.3284911, 10.0527148, 9.77595108, 9.49804526, 9.21884284, 8.93818928, 8.65593005, 8.37191061, 8.08597643, 7.79797297, 7.50774569, 7.19919134, 6.85658958, 6.51267891, 6.16833871, 5.82444832, 5.48188709, 5.1415344, 4.80426958, 4.470972, 4.14252101, 3.81979597, 3.50367624, 3.19504116, 2.8947701, 2.60374241, 2.32283745, 2.05293457, 2.06537551, 2.22011963, 2.37387553, 2.51874554, 2.64683198, 2.75023719, 2.8210635, 2.85141325, 2.83338875, 2.75909235, 2.62062638, 2.41009316, 2.11959503, 1.74123432, 1.26711335, 0.68933447, 0, -1.75819133, -3.46799213, -5.14845619, -6.8186373, -8.49758922, -10.2043657, -11.9580206, -13.7776077, -15.6821807, -17.6907934, -19.8224996, -22.0963531, -24.5314077, -27.146717, -29.961335, -32.9943154, -36.3390538, -40.0252242, -43.9037262, -48.0095442, -52.3776623, -57.0430645, -62.0407352, -67.4056584, -73.1728183, -79.3771991, -86.0537849, -93.2375599, -100.963508, -109.388742, -118.381338, -127.93964, -138.061996, -149.186551, -161.091751, -173.438761, -186.093986, -198.923832, -211.794704, -224.573007, -239.879342, -253.987686, -266.22558, -275.920563, -282.400178, -290.400529, -292.488452, -286.639347, -270.828615, -243.031656, -207.202762, -148.030905, -65.0334738, 47.1858477, 204.055519, 416.499773, 695.442841, 1039.4716, 1474.65158, 2040.50724, 2737.60267, 3641.36192, 4760.26508, 6149.83758, 7945.32056, 10245.3643, 13066.4074, 16485.7859, 20882.2793, 26464.1539, 32709.7567, 40669.5681, 50804.5184, 63575.5379, 79443.5572, 110824.771, 143236.03, 174149.446, 201037.135, 221371.211, 223477.266, 220071.617, 212546.317, 205787.554, 199804.286, 193149.579, 185905.184, 178152.853, 169974.338, 161451.388, 155601.075, 147507.796, 137847.679, 127296.851, 116531.439, 106227.568, 97061.3665, 87825.7181, 77184.0611},
        {20.779917, 20.8811266, 20.9584887, 21.0127358, 21.0446007, 21.0548158, 21.0441138, 21.0132272, 20.9628887, 20.8938307, 20.8067859, 20.7024869, 20.5816663, 20.4450566, 20.2933904, 20.1274003, 19.9478189, 19.7553788, 19.5508126, 19.3348528, 19.108232, 18.8716828, 18.6259379, 18.3717297, 18.1097909, 17.840854, 17.5656517, 17.2849165, 16.999381, 16.7097778, 16.4168394, 16.1212985, 15.8238877, 15.5253395, 15.2470647, 14.9790218, 14.7109209, 14.4428235, 14.1747912, 13.9068854, 13.6391677, 13.3716997, 13.1045429, 12.8377588, 12.571409, 12.3055549, 12.0402582, 11.7755804, 11.511583, 11.2483275, 10.9858755, 10.7387837, 10.4923043, 10.2461845, 10.0001716, 9.75401301, 9.50745598, 9.26024783, 9.01213586, 8.7628674, 8.51218974, 8.2598502, 8.00559608, 7.7491747, 7.49033337, 7.2288194, 6.96438009, 6.67819905, 6.3519056, 6.02353329, 5.69403284, 5.36435496, 5.03545039, 4.70826983, 4.383764, 4.06288363, 3.74657944, 3.43580213, 3.13150244, 2.83463107, 2.54613876, 2.26697621, 1.99809415, 1.7404433, 1.77760685, 1.96327448, 2.14773887, 2.32277876, 2.4801729, 2.61170005, 2.70913895, 2.76426837, 2.76886704, 2.71471373, 2.59358718, 2.39726614, 2.11752937, 1.74615562, 1.27492364, 0.695612185, 0, -1.91577389, -3.75959966, -5.5529577, -7.31732839, -9.07419212, -10.8450293, -12.6513202, -14.5145453, -16.456185, -18.4977197, -20.6606297, -22.9663954, -25.4364972, -28.0924155, -30.9556306, -34.047623, -37.4335716, -41.1411994, -45.0931448, -49.3202523, -53.8533661, -58.7233306, -63.9609902, -69.5971892, -75.6627721, -82.188583, -89.2054665, -96.7442668, -104.835828, -113.615515, -122.971328, -132.902756, -143.409286, -154.949391, -167.293068, -180.087917, -193.195732, -206.478305, -219.79743, -233.014899, -248.776625, -263.312941, -275.940924, -285.977646, -292.740184, -301.27755, -303.741896, -298.017311, -281.987883, -253.537704, -216.647993, -155.653661, -69.8403881, 46.3651189, 208.456553, 427.473549, 714.455743, 1066.299, 1511.40122, 2093.58987, 2819.36517, 3757.98701, 4913.78464, 6340.13638, 8173.53146, 10512.3098, 13372.3135, 16830.4941, 21266.2956, 26888.0551, 33267.387, 41347.3085, 51556.5608, 64323.8851, 80078.0225, 110723.418, 142344.183, 172500.133, 198751.082, 218656.847, 220982.467, 217945.715, 210875.596, 204383.083, 198498.066, 191882.951, 184636.003, 176855.487, 168639.669, 160086.815, 151800.131, 143194.742, 134294.644, 125123.83, 115706.295, 106066.032, 96227.0365, 87154.0764, 78897.211},
        {19.071373, 19.1710306, 19.2503763, 19.3099708, 19.3503746, 19.3721484, 19.3758527, 19.3620482, 19.3312954, 19.284155, 19.2211875, 19.1429536, 19.0500138, 18.9429287, 18.822259, 18.6885652, 18.542408, 18.3843479, 18.2149455, 18.0347614, 17.8443563, 17.6442907, 17.4351253, 17.2174205, 16.9917371, 16.7586356, 16.5186767, 16.2724208, 16.0204287, 15.7632609, 15.501478, 15.2356406, 14.9663094, 14.6940448, 14.4295436, 14.1680852, 13.9048271, 13.6400009, 13.3738384, 13.1065711, 12.8384307, 12.5696489, 12.3004574, 12.0310878, 11.7617718, 11.492741, 11.2242272, 10.9564619, 10.6896768, 10.4241037, 10.1599741, 9.92271283, 9.68681231, 9.45195807, 9.21783563, 8.98413052, 8.75052827, 8.51671439, 8.28237443, 8.0471939, 7.81085832, 7.57305323, 7.33346416, 7.09177662, 6.84767614, 6.60084826, 6.35097849, 6.07709443, 5.75871967, 5.43814516, 5.1163956, 4.79449568, 4.4734701, 4.15434356, 3.83814075, 3.52588637, 3.21860511, 2.91732167, 2.62306074, 2.33684702, 2.0597052, 1.79265998, 1.53673606, 1.29295813, 1.35950412, 1.58257188, 1.80478773, 2.01753318, 2.21218971, 2.38013884, 2.51276206, 2.60144089, 2.63755681, 2.61249133, 2.51762596, 2.34434219, 2.08402153, 1.72804548, 1.26779554, 0.694653213, 0, -2.1277112, -4.1501564, -6.0918437, -7.97728123, -9.8309771, -11.6774394, -13.5411763, -15.4466959, -17.4185063, -19.4811156, -21.659032, -23.9767635, -26.4588183, -29.1297045, -32.0139301, -35.1360035, -38.5229257, -42.1995611, -46.1847786, -50.5036204, -55.1811289, -60.2423464, -65.7123155, -71.6160784, -77.9786775, -84.8251552, -92.1805538, -100.069916, -108.518283, -117.613542, -127.294848, -137.568392, -148.440364, -160.47468, -173.398667, -186.790922, -200.490317, -214.335722, -228.16601, -241.820051, -257.766581, -272.414213, -285.087279, -295.110111, -301.807042, -311.121592, -314.104107, -308.424123, -291.751174, -261.754795, -221.778984, -156.883358, -67.3223467, 52.3099497, 218.975938, 444.449389, 740.504074, 1105.28865, 1567.29862, 2169.89324, 2919.99767, 3886.49022, 5075.44469, 6539.94447, 8412.72814, 10789.9712, 13683.3351, 17167.0841, 21625.3682, 27267.0488, 33750.2598, 41919.3169, 52173.7175, 64912.9592, 80536.5395, 110491.225, 141367.426, 170802.825, 196435.1, 215901.933, 218331.869, 215544.479, 208824.217, 202652.817, 197049.141, 190699.848, 183690.411, 176106.301, 168032.99, 159555.949, 147915.12, 138245.27, 129850.87, 122036.391, 114106.303, 105365.077, 95117.1834, 86098.7844, 77898.161},
        {17.4690024, 17.5594253, 17.6336694, 17.6921031, 17.7350953, 17.7630146, 17.7762297, 17.7751092, 17.7600217, 17.731336, 17.6894207, 17.6346445, 17.567376, 17.4879839, 17.3968368, 17.2943035, 17.1807525, 17.0565526, 16.9220724, 16.7776805, 16.6237457, 16.4606366, 16.2887218, 16.1083701, 15.91995, 15.7238303, 15.5203795, 15.3099665, 15.0929598, 14.869728, 14.6406399, 14.4060641, 14.1663694, 13.9219242, 13.6727799, 13.4194704, 13.1625337, 12.9023489, 12.6392947, 12.3737502, 12.1060944, 11.8367063, 11.5659647, 11.2942487, 11.0219372, 10.7494093, 10.4770437, 10.2052196, 9.93431595, 9.6647116, 9.39678558, 9.16321517, 8.93138084, 8.70096141, 8.47163567, 8.24308243, 8.01498049, 7.78700865, 7.55884571, 7.33017048, 7.10066175, 6.86999832, 6.63785901, 6.40392261, 6.16786792, 5.92937374, 5.68811888, 5.42228904, 5.11058597, 4.79701637, 4.48265235, 4.16856601, 3.85582948, 3.54551487, 3.23869428, 2.93643983, 2.63982363, 2.34991779, 2.06779442, 1.79452564, 1.53118355, 1.27884027, 1.03856792, 0.811438587, 0.909612424, 1.17207421, 1.43406968, 1.68657556, 1.92056854, 2.12702533, 2.29692265, 2.42123719, 2.49094567, 2.49702479, 2.43045126, 2.28220178, 2.04325307, 1.70458183, 1.25716477, 0.691978587, 0, -2.32987743, -4.52294719, -6.60661944, -8.60830434, -10.5554121, -12.4753528, -14.3955366, -16.3433738, -18.3462744, -20.4316487, -22.6269068, -24.9594589, -27.4567151, -30.1460856, -33.0549806, -36.2108102, -39.6021811, -43.2509602, -47.2723273, -51.6853776, -56.5092063, -61.7629084, -67.4655791, -73.6363136, -80.2942069, -87.4583542, -95.1478507, -103.381791, -112.179272, -121.57734, -131.570555, -142.172626, -153.397261, -165.921972, -179.428567, -193.421839, -207.716357, -222.12669, -236.467405, -250.553071, -266.697268, -281.454983, -294.151849, -304.113496, -310.665557, -320.633142, -323.967533, -318.119491, -300.539778, -268.679156, -225.098631, -155.892864, -62.8516214, 59.1205169, 229.176229, 460.15677, 764.903397, 1144.20619, 1625.69649, 2250.15235, 3022.90945, 4014.21071, 5232.9798, 6732.50431, 8641.24413, 11053.1538, 13976.8408, 17484.1327, 21959.7964, 27611.5829, 34158.4011, 42372.0499, 52629.4946, 65307.7004, 80783.6324, 110130.132, 140354.319, 169159.19, 194247.741, 213322.969, 215814.953, 213214.883, 206771.29, 200873.37, 195526.474, 189427.345, 182646.768, 175255.526, 167324.401, 158924.177, 144284.932, 133632.593, 125643.856, 118995.421, 112363.985, 104426.245, 93858.8992, 84942.2041, 76163.6073},
        {15.7523547, 15.8651138, 15.9609996, 16.0404158, 16.103766, 16.1514542, 16.183884, 16.2014591, 16.2045834, 16.1936605, 16.1690942, 16.1312882, 16.0806463, 16.0175722, 15.9424697, 15.8557425, 15.7577943, 15.6490289, 15.52985, 15.4006615, 15.2618669, 15.1138701, 14.9570748, 14.7918848, 14.6187037, 14.4379354, 14.2499835, 14.0552519, 13.8541442, 13.6470642, 13.4344157, 13.2166023, 12.9940279, 12.7670962, 12.540445, 12.312272, 12.0807238, 11.8460669, 11.6085676, 11.3684921, 11.1261069, 10.8816783, 10.6354726, 10.3877562, 10.1387953, 9.88885641, 9.63820576, 9.38710971, 9.1358346, 8.88464677, 8.63381256, 8.40325325, 8.17315414, 7.9433555, 7.71369759, 7.48402066, 7.25416496, 7.02397077, 6.79327832, 6.56192788, 6.32975971, 6.09661405, 5.86233118, 5.62675134, 5.3897148, 5.1510618, 4.91063261, 4.65007022, 4.35145455, 4.05215633, 3.75319546, 3.45559186, 3.16036545, 2.86853613, 2.58112381, 2.29914841, 2.02362983, 1.75558799, 1.49604279, 1.24601415, 1.00652198, 0.778586195, 0.563226698, 0.361463402, 0.494029132, 0.79536261, 1.09605986, 1.38676549, 1.65812412, 1.90078035, 2.1053788, 2.26256408, 2.3629808, 2.39727358, 2.35608703, 2.23006576, 2.00985437, 1.68609749, 1.24943973, 0.690525696, 0, -2.39613108, -4.6477676, -6.78371229, -8.8327679, -10.8237372, -12.7854228, -14.7466276, -16.7361542, -18.7828055, -20.9153841, -23.1626927, -25.5535342, -28.1167112, -30.8810266, -33.8752829, -37.1282831, -40.6023631, -44.3168306, -48.4508293, -53.0189192, -58.0356602, -63.5156121, -69.4733349, -75.9233884, -82.8803325, -90.3587272, -98.3731324, -106.938108, -116.068214, -125.761483, -136.055062, -146.968469, -158.521221, -171.436026, -185.380808, -199.835972, -214.610078, -229.511686, -244.349359, -258.931656, -275.735125, -291.088344, -304.277877, -314.590287, -321.312137, -331.526448, -334.774213, -328.392879, -309.719896, -276.092714, -229.484027, -156.930737, -62.311939, 59.0548808, 229.332188, 462.834306, 773.87556, 1167.04458, 1668.59177, 2315.37737, 3111.13324, 4126.19568, 5367.52308, 6888.09446, 8818.86823, 11255.009, 14213.8946, 17770.5227, 22285.109, 27956.5338, 34478.9098, 42633.5413, 52794.036, 65334.0012, 80627.0445, 109580.649, 139401.078, 167828.469, 192602.962, 211464.695, 214045.093, 211598.818, 205341.169, 199484.697, 194058.291, 187841.989, 180927.126, 173405.034, 165367.048, 156904.5, 142625.849, 132276.24, 124517.508, 118011.489, 111420.018, 103404.93, 92628.0601, 84104.4639, 75188.6224},
        {14.0891635, 14.2319421, 14.3555862, 14.4606111, 14.547532, 14.6168639, 14.6691221, 14.7048217, 14.7244778, 14.7286056, 14.7177203, 14.692337, 14.652971, 14.6001372, 14.534351, 14.4561274, 14.3659816, 14.2644288, 14.1519841, 14.0291627, 13.8964798, 13.7544505, 13.6035899, 13.4444132, 13.2774356, 13.1031723, 12.9221383, 12.7348489, 12.5418192, 12.3435643, 12.1405995, 11.9334398, 11.7226005, 11.5085966, 11.3058269, 11.1075727, 10.9069486, 10.7040191, 10.4988489, 10.2915026, 10.0820448, 9.87054012, 9.65705315, 9.44164852, 9.22439084, 9.00534472, 8.78457476, 8.56214559, 8.33812182, 8.11256805, 7.8855489, 7.6575731, 7.42825151, 7.19763912, 6.96579092, 6.73276188, 6.49860699, 6.26338124, 6.0271396, 5.78993707, 5.55182863, 5.31286926, 5.07311395, 4.83261767, 4.59143542, 4.34962218, 4.10723294, 3.85071431, 3.56683969, 3.28373023, 3.00232312, 2.7235555, 2.44836452, 2.17768736, 1.91246117, 1.65362311, 1.40211032, 1.15885998, 0.924809244, 0.700895264, 0.4880552, 0.287226211, 0.0993454559, -0.0746499083, 0.0929647476, 0.432859312, 0.771775939, 1.1000469, 1.40800449, 1.68598096, 1.9243086, 2.11331968, 2.24334649, 2.30472129, 2.28777636, 2.18284398, 1.98025643, 1.67034598, 1.24344491, 0.689885488, 0, -2.41271441, -4.6825496, -6.83906381, -8.91181532, -10.9303624, -12.9242633, -14.9230763, -16.9563597, -19.0536716, -21.2445705, -23.5586145, -26.025362, -28.6743711, -31.5352002, -34.6374075, -38.0105513, -41.6007601, -45.4256738, -49.7035606, -54.4461008, -59.6649751, -65.3718638, -71.5784475, -78.2964067, -85.5374218, -93.3131732, -101.635342, -110.515607, -119.965651, -129.960254, -140.561525, -151.792215, -163.675074, -176.946509, -191.272442, -206.128486, -221.323293, -236.665516, -251.963809, -267.026825, -284.612985, -300.683417, -314.469646, -325.203196, -332.11559, -342.207978, -344.999853, -337.780334, -317.838538, -282.463584, -233.278153, -157.941223, -62.2528514, 58.1202199, 228.200109, 463.720057, 780.413308, 1186.70827, 1707.30321, 2374.8653, 3191.07477, 4226.82356, 5486.94949, 7024.32552, 8973.2174, 11430.5303, 14424.638, 18034.6784, 22591.0703, 28282.0939, 34753.0635, 42818.758, 52855.4074, 65239.2413, 80346.4896, 108995.797, 138510.375, 166655.849, 191197.845, 209901.989, 212552.116, 210216.836, 204082.544, 198176.314, 192560.104, 186112.252, 178953.896, 171206.175, 162990.224, 154427.183, 141675.944, 132043.215, 124407.93, 117649.024, 110645.432, 102276.09, 91419.9321, 83299.147, 74376.9383},
        {12.5482653, 12.7209452, 12.8717112, 13.0012167, 13.1101152, 13.1990604, 13.2687057, 13.3197047, 13.352711, 13.3683779, 13.3673592, 13.3503082, 13.3178786, 13.2707239, 13.2094976, 13.1348532, 13.0474443, 12.9479244, 12.836947, 12.7151658, 12.5832341, 12.4418056, 12.2915337, 12.133072, 11.9670741, 11.7941935, 11.6150836, 11.4303981, 11.2407905, 11.0469142, 10.8494229, 10.6489701, 10.4462092, 10.2417939, 10.0614252, 9.89270601, 9.72293821, 9.55196251, 9.37961958, 9.20575013, 9.03019485, 8.85279442, 8.67338955, 8.49182092, 8.30792922, 8.12155516, 7.93253941, 7.74072268, 7.54594565, 7.34804902, 7.14687348, 6.92263255, 6.69521956, 6.4649007, 6.23194214, 5.99661005, 5.75917059, 5.51988994, 5.27903427, 5.03686976, 4.79366257, 4.54967887, 4.30518483, 4.06044664, 3.81573045, 3.57130244, 3.32742877, 3.07526603, 2.80618948, 2.53925318, 2.27531383, 2.01522816, 1.75985286, 1.51004465, 1.26666024, 1.03055633, 0.802589646, 0.583616886, 0.374494762, 0.176079986, -0.0107707337, -0.185200687, -0.346353163, -0.493371453, -0.291488916, 0.086172847, 0.462478894, 0.827449914, 1.1711066, 1.48346964, 1.75455973, 1.97439755, 2.1330038, 2.22039916, 2.22660434, 2.14164001, 1.95552687, 1.65828561, 1.23993691, 0.690501482, 0, -2.41176091, -4.68634003, -6.85371174, -8.9438504, -10.9867304, -13.012326, -15.0506117, -17.1315617, -19.2851505, -21.5413524, -23.9301418, -26.4814931, -29.2253805, -32.1917785, -35.4106614, -38.9120036, -42.6373972, -46.6041907, -51.0422704, -55.9626717, -61.3764298, -67.2945801, -73.7281579, -80.6881986, -88.1857375, -96.2318098, -104.837451, -114.013696, -123.771581, -134.085204, -145.016081, -156.586327, -168.81806, -182.437885, -197.115675, -212.333436, -227.901938, -243.63195, -259.33424, -274.819579, -293.174147, -309.936438, -324.276379, -335.363899, -342.368926, -351.926919, -353.875891, -345.519389, -324.160956, -287.104138, -235.979142, -158.47471, -61.6019682, 58.8208642, 229.827763, 468.192536, 790.68899, 1208.73599, 1745.67932, 2430.70651, 3263.18504, 4315.54777, 5592.321, 7146.74935, 9114.62807, 11593.8603, 14621.6138, 18280.0456, 22874.0089, 28582.3875, 35008.3368, 42995.0422, 52920.678, 65163.4186, 80101.4381, 108474.457, 137708.717, 165592.005, 189912.11, 208456.819, 211141.899, 208875.172, 202818.149, 196828.927, 191001.617, 184306.754, 176894.63, 168915.536, 160519.762, 151857.6, 140932.047, 132087.748, 124501.582, 117350.426, 109811.158, 101060.655, 90275.7952, 82419.8425, 73665.0952},
        {11.3204304, 11.5067041, 11.6698094, 11.8104682, 11.9294021, 12.0273326, 12.1049816, 12.1630706, 12.2023213, 12.2234553, 12.2271944, 12.2142603, 12.1853744, 12.1412586, 12.0826345, 12.0102237, 11.924748, 11.8269289, 11.7174882, 11.5971475, 11.4666284, 11.3266527, 11.1779419, 11.0212179, 10.8572021, 10.6866164, 10.5101822, 10.3286214, 10.1426556, 9.95300639, 9.76039549, 9.56554455, 9.36917524, 9.17200922, 9.00475272, 8.85250443, 8.70000265, 8.546996, 8.39323309, 8.23846251, 8.08243287, 7.92489279, 7.76559086, 7.6042757, 7.4406959, 7.27460009, 7.10573685, 6.93385481, 6.75870256, 6.58002872, 6.39758188, 6.18499579, 5.96870004, 5.74900935, 5.52623846, 5.30070209, 5.07271496, 4.84259181, 4.61064735, 4.37719631, 4.14255342, 3.90703341, 3.670951, 3.43462091, 3.19835787, 2.96247661, 2.72729186, 2.48443038, 2.22572755, 1.96941635, 1.71637469, 1.46748052, 1.22361175, 0.98564633, 0.754462179, 0.530937228, 0.31594941, 0.110376654, -0.0849031097, -0.269011951, -0.441071939, -0.600205144, -0.745533636, -0.876179484, -0.645286056, -0.23224197, 0.17940576, 0.57930739, 0.957113177, 1.30247338, 1.60503824, 1.85445803, 2.04038301, 2.15246341, 2.18034951, 2.11369156, 1.94213982, 1.65534453, 1.24295596, 0.694624367, 0, -2.47413644, -4.80870957, -7.03439701, -9.18187638, -11.2818253, -13.3649214, -15.4618423, -17.6032656, -19.8198689, -22.1423299, -24.6013262, -27.2275354, -30.0516351, -33.1043029, -36.4162165, -40.0180536, -43.8781091, -48.018149, -52.6099541, -57.6708345, -63.2180999, -69.2690604, -75.8410259, -82.9513063, -90.6172117, -98.8560519, -107.685137, -117.121777, -127.183281, -137.914047, -149.294365, -161.333421, -174.040397, -188.10135, -203.187028, -218.787679, -234.709426, -250.758392, -266.740697, -282.462466, -301.235894, -318.293963, -332.756825, -343.744631, -350.377533, -358.445592, -358.731572, -348.688146, -325.767987, -287.423768, -236.463425, -158.43905, -58.4262609, 68.4487633, 247.373276, 494.414213, 825.638512, 1250.91143, 1793.91106, 2485.08094, 3321.77459, 4381.42206, 5678.16312, 7267.38633, 9274.37398, 11792.326, 14848.1153, 18517.7459, 23120.4209, 28837.7761, 35345.3438, 43408.7256, 53384.7472, 65630.2341, 80502.0122, 108467.285, 137244.905, 164664.105, 188554.114, 206744.165, 209327.501, 207026.666, 200986.314, 195007.199, 189182.603, 182499.1, 175105.306, 167149.839, 158781.316, 150148.354, 138944.614, 129999.724, 122463.724, 115486.656, 108218.559, 99809.4739, 89409.4408, 81073.2373, 73156.7751},
        {10.2687401, 10.4608772, 10.629148, 10.7743185, 10.8971544, 10.9984217, 11.0788861, 11.1393136, 11.1804701, 11.2031214, 11.2080333, 11.1959717, 11.1677025, 11.1239916, 11.0656048, 10.993308, 10.907867, 10.8100477, 10.700616, 10.5803378, 10.4499788, 10.310305, 10.1620823, 10.0060764, 9.84305336, 9.6737789, 9.49901893, 9.31953932, 9.13610593, 8.94948463, 8.76044128, 8.56974174, 8.37815188, 8.18643757, 8.0282007, 7.88709847, 7.74639339, 7.60578573, 7.46497577, 7.32366377, 7.18155001, 7.03833475, 6.89371828, 6.74740086, 6.59908276, 6.44846426, 6.29524563, 6.13912714, 5.97980906, 5.81699166, 5.65037522, 5.45151186, 5.2488602, 5.0427307, 4.83343384, 4.62128008, 4.40657988, 4.18964373, 3.97078208, 3.7503054, 3.52852417, 3.30574885, 3.08228991, 2.85845782, 2.63456304, 2.41091605, 2.18782732, 1.95611682, 1.70683242, 1.45985798, 1.2161192, 0.976541777, 0.742051417, 0.513573813, 0.292034662, 0.0783596632, -0.126525486, -0.321695089, -0.506223446, -0.67918486, -0.839653634, -0.98670407, -1.11941047, -1.23684713, -0.97894171, -0.531895995, -0.086228451, 0.347331619, 0.758054912, 1.13521213, 1.46807396, 1.74591112, 1.95799429, 2.09359417, 2.14198147, 2.09242687, 1.93420109, 1.65657481, 1.24881874, 0.700203569, 0, -2.54624828, -4.94985218, -7.24214846, -9.45447386, -11.6181651, -13.764559, -15.9249922, -18.1308015, -20.4133237, -22.8038955, -25.3338536, -28.0345348, -30.9372758, -34.0734134, -37.4742843, -41.1712253, -45.1661442, -49.4861738, -54.2292159, -59.4203011, -65.0844601, -71.2467233, -77.9321214, -85.1656849, -92.9724443, -101.37743, -110.405673, -120.082204, -130.432053, -141.586263, -153.42498, -165.941434, -179.12885, -193.625822, -209.102894, -225.058513, -241.292294, -257.603856, -273.792814, -289.658788, -308.706842, -325.903399, -340.323097, -351.040574, -357.130466, -363.43515, -361.81959, -349.916492, -325.358557, -285.77849, -235.434146, -157.235669, -53.50808, 81.5205648, 270.153513, 527.271794, 867.756438, 1298.72541, 1844.58714, 2538.21884, 3375.03639, 4438.24809, 5753.66543, 7379.24119, 9427.38976, 11985.5267, 15066.5548, 18738.5714, 23342.8881, 29066.271, 35679.0457, 43848.4869, 53905.5131, 66181.0431, 81005.9956, 108534.942, 136819.236, 163733.881, 187153.884, 204954.251, 207418.155, 205073.654, 199049.394, 193109.527, 187339.479, 180726.55, 173411.074, 165533.382, 157233.808, 148652.685, 136805.297, 127577.514, 120051.704, 113310.237, 106435.483, 98509.8111, 88615.5903, 79624.4246, 72666.5971},
        {9.37270529, 9.56275306, 9.72889437, 9.87191067, 9.99258342, 10.0916941, 10.1700241, 10.2283549, 10.2674681, 10.288145, 10.291167, 10.2773158, 10.2473726, 10.2021191, 10.1423366, 10.0688065, 9.98231045, 9.88362979, 9.77354601, 9.65284055, 9.52229489, 9.38269047, 9.23480876, 9.07943122, 8.91733931, 8.74931448, 8.57613819, 8.3985919, 8.21745708, 8.03351517, 7.84754764, 7.66033595, 7.47266156, 7.28530592, 7.13257177, 6.99777485, 6.86382802, 6.7304249, 6.59725912, 6.46402431, 6.33041409, 6.1961221, 6.06084196, 5.9242673, 5.78609176, 5.64600896, 5.50371253, 5.3588961, 5.21125329, 5.06047775, 4.90626308, 4.72182248, 4.53390407, 4.34277551, 4.14870449, 3.95195868, 3.75280574, 3.55151336, 3.34834921, 3.14358096, 2.93747628, 2.73030285, 2.52232834, 2.31382042, 2.10504677, 1.89627506, 1.68777296, 1.46871361, 1.22853654, 0.990391048, 0.755264027, 0.524142364, 0.298012944, 0.0778626553, -0.135321616, -0.340552984, -0.536844562, -0.723209462, -0.898660799, -1.06221168, -1.21287523, -1.34966456, -1.47159277, -1.57767299, -1.29424432, -0.81449556, -0.336094794, 0.12985088, 0.57223436, 0.979948545, 1.34188633, 1.64694062, 1.88400431, 2.0419703, 2.10973148, 2.07618076, 1.93021103, 1.66071519, 1.25658614, 0.706716776, 0, -2.62227754, -5.09856672, -7.46082832, -9.74102311, -11.9711119, -14.1830554, -16.4088145, -18.6803498, -21.0296223, -23.4885926, -26.0892216, -28.86347, -31.8432985, -35.0606681, -38.5475394, -42.3358732, -46.4593677, -50.9502317, -55.8364965, -61.150495, -66.9245604, -73.1910258, -79.9822241, -87.3304884, -95.2681518, -103.827547, -113.041008, -122.940867, -133.559457, -145.112794, -157.382178, -170.344838, -183.978002, -198.878682, -214.714215, -230.986663, -247.487318, -264.007475, -280.338426, -296.271466, -315.463158, -332.663139, -346.906453, -357.228146, -362.663261, -367.177913, -363.643311, -349.861732, -323.635455, -282.766757, -232.945413, -154.445929, -46.6649871, 97.0920508, 295.95627, 563.475044, 913.195747, 1349.14259, 1896.55046, 2591.04318, 3426.2309, 4491.03445, 5822.80725, 7482.07505, 9568.62765, 12164.232, 15268.0822, 18940.5373, 23544.724, 29272.5065, 35990.3674, 44265.6113, 54402.7346, 66706.2335, 81480.6044, 108554.312, 136326.891, 162721.848, 185662.687, 203072.911, 205440.209, 203081.112, 197105.904, 191238.077, 185553.519, 179039.28, 171825.688, 164043.074, 155821.766, 147292.094, 134702.849, 125124.192, 117568.928, 111049.859, 104579.79, 97171.5233, 87837.8629, 78169.032, 72111.886},
        {8.60440315, 8.78278689, 8.93841982, 9.07204161, 9.18439191, 9.27621039, 9.34823671, 9.40121054, 9.43587153, 9.45295937, 9.4532137, 9.43737419, 9.40618051, 9.36037231, 9.30068927, 9.22787104, 9.14265729, 9.04578769, 8.93800189, 8.82003957, 8.69264037, 8.55654398, 8.41249004, 8.26121823, 8.10346821, 7.93997965, 7.77149219, 7.59874552, 7.42247929, 7.24343317, 7.06234682, 6.8799599, 6.69701208, 6.51424303, 6.36258922, 6.22705677, 6.09228883, 5.95804512, 5.82408535, 5.69016926, 5.55605654, 5.42150693, 5.28628014, 5.15013589, 5.0128339, 4.87413389, 4.73379557, 4.59157867, 4.4472429, 4.30054798, 4.15125363, 3.97908431, 3.80426932, 3.62700271, 3.44747852, 3.26589079, 3.08243356, 2.89730089, 2.7106868, 2.52278536, 2.33379059, 2.14389655, 1.95329727, 1.76218681, 1.5707592, 1.37920849, 1.18772872, 0.983536119, 0.754158512, 0.526555709, 0.301763053, 0.0808158861, -0.13525045, -0.345400612, -0.548599257, -0.743811044, -0.93000063, -1.10613267, -1.27117183, -1.42408276, -1.56383011, -1.68937856, -1.79969274, -1.89373733, -1.58555773, -1.07467447, -0.565234541, -0.0686939612, 0.403491239, 0.839865032, 1.22897139, 1.55935429, 1.81955769, 1.99812558, 2.08360193, 2.0645307, 1.92945587, 1.66692141, 1.2654713, 0.713649505, 0, -2.73155808, -5.306088, -7.75678567, -10.116847, -12.4194679, -14.6978444, -16.9851722, -19.3146474, -21.7194659, -24.2328235, -26.8879163, -29.71794, -32.7560907, -36.0355643, -39.5895566, -43.4512636, -47.6420038, -52.1932766, -57.165143, -62.588254, -68.4932602, -74.9108125, -81.8715616, -89.4061584, -97.5452535, -106.319498, -115.759542, -125.896037, -136.759634, -148.569912, -161.099319, -174.321827, -188.211408, -203.395649, -219.521714, -236.062471, -252.795805, -269.499605, -285.951758, -301.93015, -321.148055, -338.250248, -352.24465, -362.139183, -366.941765, -370.378001, -365.558708, -350.312386, -322.467536, -279.852658, -228.685168, -148.151531, -36.4467788, 113.531072, 319.698741, 595.117351, 952.848025, 1396.31192, 1950.40102, 2650.52367, 3487.78743, 4555.34805, 5895.49172, 7570.57702, 9676.97408, 12295.4412, 15426.6505, 19130.5974, 23760.9257, 29505.1753, 36249.1354, 44528.519, 54635.318, 66861.5238, 81499.1282, 108123.733, 135422.818, 161367.47, 183928.78, 201077.837, 203543.358, 201367.05, 195623.876, 189889.877, 184250.005, 177747.094, 170518.202, 162700.387, 154430.705, 145846.214, 133223.038, 123565.19, 115934.17, 109391.482, 102998.627, 95817.1073, 86908.4252, 76990.1546, 71151.4998},
        {7.94503275, 8.10227307, 8.23911643, 8.35620411, 8.45417738, 8.53367753, 8.59534582, 8.63982353, 8.66775193, 8.6797723, 8.67652592, 8.65865405, 8.62679797, 8.58159896, 8.5236983, 8.45373725, 8.37235709, 8.28019909, 8.17790454, 8.0661147, 7.94547085, 7.81661427, 7.68018623, 7.536828, 7.38718085, 7.23188607, 7.07158493, 6.9069187, 6.73852866, 6.56705608, 6.39314224, 6.21742841, 6.04055586, 5.86316587, 5.70908202, 5.5668668, 5.42480413, 5.28278295, 5.14069224, 4.99842095, 4.85585806, 4.71289253, 4.56941332, 4.4253094, 4.28046973, 4.13478327, 3.988139, 3.84042587, 3.69153286, 3.54134892, 3.38976302, 3.22668255, 3.06219442, 2.89640399, 2.72941661, 2.56133761, 2.39227235, 2.22232616, 2.05160441, 1.88021242, 1.70825556, 1.53583915, 1.36306856, 1.19004912, 1.01688618, 0.843685094, 0.670551196, 0.482849562, 0.266299169, 0.0514060764, -0.160768819, -0.369164618, -0.572720424, -0.770375338, -0.961068463, -1.1437389, -1.31732575, -1.48076812, -1.63300511, -1.77297582, -1.89961935, -2.01187481, -2.10868129, -2.1889779, -1.85635639, -1.31574512, -0.77682426, -0.25136269, 0.248870706, 0.712107044, 1.12657744, 1.48051301, 1.76214486, 1.95970412, 2.0614219, 2.05552932, 1.93025749, 1.67383753, 1.27450055, 0.720477666, 0, -2.86028845, -5.54703672, -8.09515135, -10.5395388, -12.9151058, -15.2567586, -17.5994039, -19.9779481, -22.4272978, -24.9823596, -27.6780399, -30.5492453, -33.6308823, -36.9578573, -40.5650771, -44.4874479, -48.6886404, -53.1933843, -58.1977155, -63.7212758, -69.7837068, -76.40465, -83.6037472, -91.4006399, -99.8149698, -108.866379, -118.574508, -128.958999, -140.039494, -151.953724, -164.559942, -177.842362, -191.785198, -207.119418, -223.45586, -240.206229, -257.130711, -273.989498, -290.542776, -306.550734, -325.690597, -342.613377, -356.312885, -365.782935, -370.017338, -373.221952, -367.875533, -351.66888, -322.292794, -277.438076, -222.790665, -138.274385, -22.9589047, 130.049629, 339.820587, 619.956783, 984.061028, 1437.83746, 2004.58271, 2716.21018, 3560.88433, 4633.79987, 5974.35299, 7645.65998, 9751.02955, 12375.3214, 15537.3147, 19304.7371, 23987.6119, 29758.9065, 36438.7945, 44607.3127, 54559.549, 66590.5917, 80995.5288, 107183.342, 134056.752, 159632.374, 181926.823, 198956.713, 201725.94, 199938.984, 194617.914, 189077.428, 183432.093, 176841.21, 169467.106, 161472.112, 153018.554, 144268.763, 132464.462, 123062.253, 115311.098, 108459.956, 101757.786, 94453.5494, 85796.2055, 76140.8492, 69761.4567},
        {7.38736532, 7.5162566, 7.62762565, 7.7219971, 7.79989555, 7.86184561, 7.90837188, 7.93999899, 7.95725153, 7.96065413, 7.95073138, 7.9280079, 7.8930083, 7.84625718, 7.78827916, 7.71959885, 7.64074086, 7.55222978, 7.45459025, 7.34834686, 7.23402422, 7.11214695, 6.98323965, 6.84782694, 6.70643341, 6.5595837, 6.40780239, 6.2516141, 6.09154345, 5.92811504, 5.76185347, 5.59328337, 5.42292934, 5.25131599, 5.0945092, 4.94493606, 4.79483686, 4.64423186, 4.49314132, 4.34158551, 4.1895847, 4.03715915, 3.88432912, 3.73111489, 3.57753671, 3.42361486, 3.26936959, 3.11482118, 2.95998988, 2.80489597, 2.6495597, 2.49366414, 2.33757406, 2.18131705, 2.02492068, 1.86841253, 1.71182016, 1.55517117, 1.39849311, 1.24181357, 1.08516012, 0.92856034, 0.772041803, 0.615632085, 0.45935876, 0.303249404, 0.147331593, -0.0245040293, -0.227979378, -0.429785667, -0.62884923, -0.824096399, -1.01445351, -1.19884689, -1.37620288, -1.5454478, -1.705508, -1.8553098, -1.99377954, -2.11984355, -2.23242816, -2.33045971, -2.41286453, -2.47856895, -2.12167232, -1.55202945, -0.984360878, -0.430730344, 0.096798403, 0.586161616, 1.02529555, 1.40213645, 1.70462058, 1.92068418, 2.03826352, 2.04529484, 1.92971439, 1.67945843, 1.28246321, 0.726664981, 0, -2.9399461, -5.69884162, -8.31265167, -10.8173414, -13.2488758, -15.6432202, -18.0363395, -20.464199, -22.9627636, -25.5679987, -28.3158691, -31.2423402, -34.383377, -37.7749445, -41.453008, -45.4535326, -49.6884992, -54.1761611, -59.2329235, -64.8681833, -71.0913378, -77.9117841, -85.3389192, -93.3821403, -102.050845, -111.354429, -121.302291, -131.903827, -143.168434, -155.158644, -167.811237, -181.119671, -195.077402, -210.508972, -226.992295, -243.883664, -260.927211, -277.867068, -294.447367, -310.412239, -329.38248, -346.045704, -359.387566, -368.393718, -372.049816, -375.239714, -369.576315, -352.570724, -321.734045, -274.577382, -215.969146, -127.07087, -8.28030258, 146.9306, 359.230037, 642.937569, 1012.37275, 1475.70636, 2054.80419, 2778.23981, 3632.42716, 4712.56509, 6052.80386, 7716.13215, 9814.40367, 12437.0423, 15621.8991, 19445.0274, 24171.5802, 29959.745, 36560.0659, 44601.0883, 54382.107, 66202.4169, 80361.3129, 106106.865, 132552.4, 157760.017, 179791.818, 196709.905, 199794.278, 198410.538, 193527.399, 188198.562, 182567.435, 175908.681, 168409.146, 160255.675, 151635.112, 142734.303, 131907.689, 122808.327, 114905.456, 107668.312, 100566.134, 93068.1596, 84643.6269, 75382.347, 68320.1215},
        {6.91693331, 7.01118101, 7.09104993, 7.15694458, 7.20926948, 7.24842915, 7.2748281, 7.28887085, 7.2909619, 7.28150578, 7.260907, 7.22957007, 7.18789952, 7.13629984, 7.07517557, 7.00493121, 6.92597128, 6.83870029, 6.74352276, 6.64084321, 6.53106614, 6.41459607, 6.29183752, 6.16319501, 6.02907304, 5.88987613, 5.7460088, 5.59787556, 5.44588093, 5.29042942, 5.13192554, 4.97077382, 4.80737876, 4.64214488, 4.48409762, 4.32914969, 4.17311022, 4.01610394, 3.85825561, 3.69968998, 3.5405318, 3.38090581, 3.22093677, 3.06074942, 2.90046852, 2.74021882, 2.58012506, 2.42031199, 2.26090436, 2.10202693, 1.94380444, 1.79302371, 1.64300299, 1.49372263, 1.34516294, 1.19730426, 1.05012692, 0.90361124, 0.757737557, 0.612486197, 0.46783749, 0.323771764, 0.180269348, 0.0373105718, -0.105124236, -0.247054747, -0.388500631, -0.546307357, -0.736916412, -0.925645524, -1.11142361, -1.29317961, -1.46984243, -1.64034099, -1.80360423, -1.95856106, -2.10414041, -2.2392712, -2.36288236, -2.4739028, -2.57126145, -2.65388723, -2.72070907, -2.77065589, -2.38930787, -1.79096255, -1.19488558, -0.613419837, -0.058908218, 0.456306398, 0.919881125, 1.31947308, 1.64273938, 1.87733714, 2.01092347, 2.03115549, 1.92569032, 1.68218507, 1.28829685, 0.731682793, 0, -2.93623847, -5.69972842, -8.3263345, -10.8519213, -13.3123536, -15.7434959, -18.1812129, -20.6613693, -23.2198296, -25.8924586, -28.7151208, -31.723681, -34.9540037, -38.4419536, -42.2233954, -46.3341937, -50.6590946, -55.2123745, -60.3717136, -66.140594, -72.5224978, -79.5209071, -87.1393041, -95.3811709, -104.24999, -113.749242, -123.882411, -134.652978, -146.064425, -158.142697, -170.860578, -184.21481, -198.202137, -213.700337, -230.265635, -247.219316, -264.293813, -281.22156, -297.73499, -313.566536, -332.288561, -348.624894, -361.55668, -370.065061, -373.131181, -376.371953, -370.473808, -352.758944, -320.549562, -271.167864, -208.571306, -115.337071, 6.74457645, 163.786991, 378.040072, 664.50874, 1038.19792, 1509.16401, 2098.48511, 2832.63768, 3698.2496, 4788.10003, 6130.02542, 7785.86291, 9875.06197, 12490.6271, 15685.483, 19542.2432, 24290.48, 30076.6374, 36602.8637, 44528.2521, 54153.0412, 65777.4699, 79701.7769, 104996.805, 130999.537, 155817.563, 177558.469, 194329.844, 197674.808, 196641.728, 192151.536, 187035.609, 181458.66, 174789.382, 167232.288, 158991.893, 150272.709, 141279.25, 131419.887, 122602.376, 114522.201, 106874.85, 99355.8062, 91660.5562, 83484.5854, 74658.4266, 66954.1771},
        {6.51856216, 6.57147479, 6.61312072, 6.64380295, 6.66382448, 6.67348833, 6.6730975, 6.662955, 6.64336384, 6.61462702, 6.57704756, 6.53092846, 6.47657272, 6.41428336, 6.34436339, 6.26711581, 6.18284363, 6.09184985, 5.99443749, 5.89090955, 5.78156905, 5.66671898, 5.54666235, 5.42170219, 5.29214148, 5.15828324, 5.02043048, 4.8788862, 4.73395342, 4.58593513, 4.43513436, 4.2818541, 4.12639736, 3.96906715, 3.81496545, 3.66189484, 3.50760022, 3.35222887, 3.19592806, 3.03884506, 2.88112714, 2.72292158, 2.56437564, 2.40563661, 2.24685174, 2.08816832, 1.92973362, 1.7716949, 1.61419944, 1.45739452, 1.30142739, 1.15300411, 1.005571, 0.859133141, 0.713695632, 0.569263563, 0.425842025, 0.283436107, 0.142050901, 0.00169149751, -0.137637012, -0.275929537, -0.413180987, -0.549386271, -0.684540297, -0.818637976, -0.951674216, -1.09966085, -1.27822517, -1.45432836, -1.62692702, -1.79497774, -1.95743713, -2.11326177, -2.26140827, -2.40083323, -2.53049323, -2.64934488, -2.75634478, -2.85044951, -2.93061568, -2.99579989, -3.04495873, -3.07704879, -2.670592, -2.04360557, -1.41916001, -0.809855838, -0.228293562, 0.312926308, 0.801203261, 1.22393679, 1.56852637, 1.82237151, 1.97287169, 2.0074264, 1.91343512, 1.67829736, 1.28941259, 0.734180306, 0, -2.75129473, -5.37279982, -7.8977519, -10.3593876, -12.7909436, -15.2256564, -17.6967629, -20.2374994, -22.8811028, -25.6608097, -28.6098566, -31.7614802, -35.1489172, -38.8054042, -42.7641778, -47.0584747, -51.6117383, -56.4415205, -61.8293997, -67.7850854, -74.3182871, -81.4387143, -89.1560765, -97.4800834, -106.420444, -115.986869, -126.189067, -137.036748, -148.53962, -160.793644, -173.690654, -187.214486, -201.348974, -216.922466, -233.496541, -250.407573, -267.386043, -284.162434, -300.467224, -316.030897, -334.417558, -350.357308, -362.830569, -370.817764, -373.299318, -376.537853, -370.411044, -352.078765, -318.700889, -267.437292, -201.458898, -104.567688, 20.4993322, 179.524711, 395.685092, 684.347525, 1060.87906, 1535.41749, 2129.7476, 2871.35877, 3750.64252, 4854.41376, 6205.00647, 7861.62818, 9945.96663, 12551.0301, 15731.6121, 19570.8132, 24291.8556, 30038.7009, 36534.5787, 44408.9106, 53951.2959, 65451.3342, 79198.625, 104028.087, 129547.671, 153910.647, 175270.283, 191779.85, 195206.644, 194347.223, 190091.703, 185161.549, 179721.968, 173175.102, 165724.31, 157572.951, 148924.385, 139981.969, 130780.126, 122092.238, 113805.047, 105805.295, 97979.7225, 90215.0706, 82398.0809, 73907.709, 65893.7055},
        {6.1775284, 6.18285213, 6.17972035, 6.16837057, 6.14904034, 6.12196717, 6.08738861, 6.04554218, 5.99666541, 5.94099584, 5.87877098, 5.81022838, 5.73560556, 5.65514006, 5.5690694, 5.47763111, 5.38106273, 5.27960178, 5.17348579, 5.06295231, 4.94823884, 4.82958294, 4.70722212, 4.58139391, 4.45233585, 4.32028547, 4.1854803, 4.04815786, 3.90855569, 3.76691132, 3.62346228, 3.4784461, 3.3321003, 3.18466243, 3.04216559, 2.90182763, 2.76079671, 2.6191223, 2.47685383, 2.33404077, 2.19073256, 2.04697866, 1.90282853, 1.7583316, 1.61353734, 1.4684952, 1.32325462, 1.17786507, 1.03237599, 0.886836838, 0.741297064, 0.592119083, 0.443119314, 0.294427134, 0.146171926, -0.00151693264, -0.14851006, -0.294678078, -0.439891605, -0.584021261, -0.726937667, -0.868511443, -1.00861321, -1.14711358, -1.28388319, -1.41879264, -1.55171257, -1.69571978, -1.86357404, -2.02790915, -2.18773964, -2.34208002, -2.4899448, -2.6303485, -2.76230563, -2.88483072, -2.99693827, -3.0976428, -3.18595884, -3.26090088, -3.32148345, -3.36672107, -3.39562825, -3.4072195, -2.97460351, -2.31870605, -1.66557199, -1.02803731, -0.418937977, 0.148890022, 0.662610709, 1.1093881, 1.47638623, 1.75076911, 1.91970075, 1.97034519, 1.88986645, 1.66542853, 1.28419547, 0.733331288, 0, -2.32781012, -4.61461268, -6.88766993, -9.17424412, -11.5015975, -13.8969924, -16.3876909, -19.0009554, -21.764048, -24.7042312, -27.848767, -31.2249178, -34.8599458, -38.7811132, -43.0156824, -47.5909155, -52.5616994, -57.9592429, -63.7483238, -69.9621239, -76.633825, -83.796609, -91.4836577, -99.7281529, -108.563276, -118.02221, -128.138136, -138.944235, -150.47369, -163.044077, -176.299906, -190.189039, -204.65934, -220.345022, -236.850774, -253.598257, -270.329429, -286.78625, -302.710676, -317.844667, -335.814012, -351.294802, -363.269116, -370.719032, -372.626629, -375.670305, -369.211739, -350.308929, -316.019873, -263.402572, -195.167594, -95.8116676, 31.8601062, 193.500834, 412.032468, 702.623601, 1080.44283, 1552.97443, 2144.81775, 2888.95499, 3784.15727, 4907.07681, 6276.85849, 7948.34935, 10036.8917, 12630.0596, 15764.8105, 19515.5958, 24142.46, 29800.466, 36336.9728, 44262.0854, 53837.3788, 65324.4284, 78984.8093, 103329.429, 128308.197, 152120.356, 172965.148, 189041.816, 192284.614, 191334.43, 187075.501, 182282.94, 177092.511, 170852.116, 163736.914, 155922.065, 147582.727, 138894.06, 129823.684, 121022.178, 112500.588, 104269.962, 96341.3463, 88725.7882, 81434.3346, 73072.1057, 65302.8149},
        {5.88345485, 5.84341415, 5.79745318, 5.74575339, 5.68849621, 5.62586309, 5.55803547, 5.48519479, 5.40752249, 5.32520002, 5.23840881, 5.1473303, 5.05214594, 4.95303717, 4.85018543, 4.74377216, 4.6339788, 4.52098679, 4.40497758, 4.28613261, 4.16463331, 4.04066114, 3.91439752, 3.7860239, 3.65572173, 3.52367244, 3.39005748, 3.25505829, 3.1188563, 2.98163296, 2.84356972, 2.704848, 2.56564926, 2.42615494, 2.29354193, 2.16434682, 2.03502209, 1.90552216, 1.77580146, 1.64581441, 1.51551544, 1.38485898, 1.25379944, 1.12229126, 0.990288854, 0.857746653, 0.72461908, 0.590860561, 0.456425521, 0.321268384, 0.185343575, 0.0351475348, -0.115615585, -0.266699615, -0.417858388, -0.568845737, -0.719415494, -0.86932149, -1.01831756, -1.16615753, -1.31259524, -1.45738452, -1.6002792, -1.74103311, -1.87940009, -2.01513396, -2.14798857, -2.28836095, -2.44639261, -2.59988655, -2.74790666, -2.8895168, -3.02378086, -3.1497627, -3.26652622, -3.37313527, -3.46865374, -3.5521455, -3.62267443, -3.67930441, -3.72109931, -3.74712301, -3.75643938, -3.7481123, -3.28874226, -2.60272111, -1.91963859, -1.25259283, -0.614681994, -0.0190042146, 0.52134236, 0.993259584, 1.38364931, 1.67941341, 1.86745371, 1.93467209, 1.8679704, 1.65425049, 1.28041421, 0.733363433, 0, -1.99931152, -4.02941447, -6.11272686, -8.2716667, -10.528652, -12.9061008, -15.4264312, -18.112061, -20.9854084, -24.0688914, -27.3849279, -30.955936, -34.8043338, -38.9525392, -43.4229702, -48.2380448, -53.5623751, -59.4386923, -65.5677855, -72.0025428, -78.7958524, -86.0006024, -93.6696809, -101.855976, -110.612376, -119.991769, -130.047043, -140.831086, -152.396787, -165.258538, -178.838504, -193.051122, -207.810828, -223.566423, -239.965161, -256.511801, -272.960467, -289.065289, -304.580392, -319.259904, -336.867549, -351.927545, -363.409533, -370.283154, -371.518047, -373.834284, -366.513468, -346.587632, -311.088812, -257.04904, -187.091453, -85.8234105, 44.4484045, 209.419133, 431.12097, 724.251108, 1103.50674, 1572.87224, 2160.19235, 2904.99849, 3815.12394, 4956.7039, 6345.87502, 8033.08079, 10126.8957, 12709.211, 15799.0283, 19461.934, 23994.1113, 29561.3049, 36130.2649, 44096.3451, 53694.1731, 65158.3763, 78723.5823, 102584.311, 127025.325, 150291.277, 170626.823, 186276.616, 189340.33, 188304.254, 184046.605, 179393.249, 174451.459, 168516.987, 161737.683, 154261.394, 146235.969, 137809.257, 128927.402, 120065.468, 111346.646, 102894.128, 94831.1037, 87280.7654, 80366.3036, 72125.2321, 64709.9922},
        {5.60830886, 5.53462908, 5.45696545, 5.37544773, 5.2902057, 5.20136914, 5.10906783, 5.01343152, 4.91459001, 4.81267306, 4.70781044, 4.60013194, 4.48976732, 4.37684636, 4.26149883, 4.14385451, 4.02404317, 3.90219459, 3.77843853, 3.65290478, 3.52572311, 3.39702329, 3.26693509, 3.13558829, 3.00311267, 2.86963799, 2.73529404, 2.60021058, 2.46451739, 2.32834425, 2.19182092, 2.05507719, 1.91824282, 1.78144759, 1.65069462, 1.52305346, 1.3955231, 1.26804271, 1.14055145, 1.01298851, 0.885293056, 0.757404257, 0.62926129, 0.500803326, 0.371969539, 0.242699102, 0.112931188, -0.0173950317, -0.148340383, -0.279965692, -0.412331788, -0.561276441, -0.710741521, -0.860445841, -1.01010822, -1.15944746, -1.30818239, -1.45603181, -1.60271454, -1.7479494, -1.89145519, -2.03295074, -2.17215485, -2.30878634, -2.44256403, -2.57320672, -2.70043323, -2.8339112, -2.98306938, -3.1271079, -3.26510063, -3.39612145, -3.51924423, -3.63354284, -3.73809116, -3.83196305, -3.91423239, -3.98397305, -4.04025891, -4.08216383, -4.1087617, -4.11912637, -4.11233173, -4.08745165, -3.60070141, -2.88189912, -2.16618569, -1.46700661, -0.797807363, -0.172033434, 0.396869692, 0.895456529, 1.31028159, 1.6278994, 1.83486445, 1.91773128, 1.8630544, 1.65738831, 1.28728754, 0.739306598, 0, -2.07840002, -4.18461635, -6.34142038, -8.57158347, -10.897877, -13.3430723, -15.9299408, -18.6812538, -21.6197827, -24.7682988, -28.1495737, -31.7863785, -35.7014846, -39.9176635, -44.4576866, -49.3443251, -54.7422729, -60.6945761, -66.9029908, -73.4207004, -80.3008882, -87.5967376, -95.3614318, -103.648154, -112.510088, -122.000417, -132.172324, -143.078993, -154.773606, -167.820446, -181.574194, -195.934705, -210.801834, -226.535019, -242.804322, -259.157248, -275.355774, -291.161879, -306.33754, -320.644736, -338.208717, -353.100236, -364.227587, -370.499067, -370.822968, -370.88991, -361.130281, -338.756795, -300.982166, -245.019108, -174.440202, -72.3251189, 60.8551914, 231.294801, 458.795533, 757.049607, 1139.74924, 1606.19826, 2187.74738, 2931.44374, 3853.85429, 5011.84278, 6417.77069, 8118.56009, 10219.6941, 12799.7754, 15870.9477, 19494.5834, 23975.7381, 29481.6264, 36026.7043, 43954.4465, 53484.9456, 64838.2944, 78234.5855, 101613.481, 125545.612, 148321.178, 168230.38, 183563.419, 186624.247, 185680.143, 181586.842, 177110.726, 172351.231, 166611.693, 160031.355, 152749.458, 144905.246, 136637.961, 128485.219, 119867.836, 111045.92, 102279.582, 93828.9325, 85954.0806, 78915.1364, 70911.7555, 63869.1214},
        {5.36253757, 5.26271542, 5.1603926, 5.05565316, 4.94858116, 4.83926067, 4.72777576, 4.61421047, 4.49864888, 4.38117505, 4.26187303, 4.14082689, 4.0181207, 3.89383852, 3.7680644, 3.64088241, 3.51237661, 3.38263106, 3.25172983, 3.11975698, 2.98679657, 2.85293266, 2.71824932, 2.5828306, 2.44676057, 2.31012329, 2.17300282, 2.03548323, 1.89764857, 1.75958292, 1.62137033, 1.48309485, 1.34484057, 1.20669153, 1.07215368, 0.939528153, 0.807074987, 0.674767194, 0.542577789, 0.410479789, 0.278446209, 0.146450065, 0.0144643731, -0.117537851, -0.249583592, -0.381699833, -0.513913559, -0.646251754, -0.778741402, -0.911409488, -1.04428299, -1.19082051, -1.33732625, -1.48353602, -1.62918564, -1.77401091, -1.91774765, -2.06013169, -2.20089882, -2.33978486, -2.47652562, -2.61085693, -2.74251458, -2.8712344, -2.9967522, -3.11880379, -3.23712499, -3.36196129, -3.50330585, -3.63921906, -3.76875542, -3.89096946, -4.00491568, -4.1096486, -4.20422273, -4.28769258, -4.35911267, -4.4175375, -4.46202161, -4.49161948, -4.50538565, -4.50237462, -4.4816409, -4.44223901, -3.92750157, -3.17392954, -2.42350129, -1.6900604, -0.987450423, -0.329514931, 0.269902513, 0.796958344, 1.237809, 1.57861091, 1.80552051, 1.90469424, 1.86228852, 1.66445981, 1.29736453, 0.747159114, 0, -2.38837918, -4.75898505, -7.13789443, -9.55118413, -12.024931, -14.5852117, -17.2581032, -20.0696822, -23.0460256, -26.2132101, -29.5973126, -33.2244099, -37.1205788, -41.311896, -45.8244385, -50.6842829, -55.9876739, -61.770712, -67.9007617, -74.4189355, -81.3663463, -88.7841067, -96.7133296, -105.195128, -114.270613, -123.9809, -134.3671, -145.470326, -157.331691, -170.466553, -184.267889, -198.634539, -213.46534, -229.109283, -245.240131, -261.411608, -277.387507, -292.931622, -307.807747, -321.779675, -339.382122, -354.15594, -364.931482, -370.539101, -369.809149, -367.030533, -354.210414, -328.814505, -288.308522, -230.158178, -159.245389, -56.5554238, 79.6879189, 256.329478, 490.562561, 794.934221, 1181.99152, 1646.53625, 2222.89373, 2964.74775, 3895.62648, 5065.38688, 6482.95469, 8192.76899, 10299.4442, 12879.9734, 15946.8721, 19560.417, 24017.3016, 29480.4884, 35966.6029, 43807.4586, 53216.6733, 64407.865, 77594.6515, 100509.156, 123966.865, 146301.77, 165847.862, 180939.134, 184099.029, 183345.785, 179503.631, 175214.676, 170583.568, 164957.505, 158479.007, 151290.594, 143534.787, 135354.107, 128096.869, 119855.99, 111014.903, 101957.041, 93065.8355, 84724.72, 77317.1269, 69557.4381, 62847.4338},
        {5.15657962, 5.03398129, 4.91005219, 4.78483823, 4.65838528, 4.53073925, 4.40194602, 4.27205149, 4.14110154, 4.00914206, 3.87621896, 3.74237811, 3.6076654, 3.47212674, 3.33580801, 3.19875509, 3.06101389, 2.92263029, 2.78365018, 2.64411945, 2.504084, 2.36358971, 2.22268248, 2.0814082, 1.93981275, 1.79794203, 1.65584193, 1.51355834, 1.37113715, 1.22862425, 1.08606553, 0.943506891, 0.80099421, 0.658573384, 0.516902551, 0.375708599, 0.23471106, 0.0939359548, -0.0465906932, -0.186842862, -0.32679453, -0.466419675, -0.605692274, -0.744586306, -0.883075748, -1.02113458, -1.15873677, -1.29585631, -1.43246717, -1.56854333, -1.70405877, -1.84812802, -1.99138636, -2.13360961, -2.2745736, -2.41405415, -2.5518271, -2.68766828, -2.82135351, -2.95265863, -3.08135945, -3.20723181, -3.33005154, -3.44959447, -3.56563642, -3.67795322, -3.7863207, -3.90224461, -4.03694907, -4.16601782, -4.2884663, -4.40330992, -4.50956411, -4.60624429, -4.6923659, -4.76694436, -4.82899508, -4.87753351, -4.91157505, -4.93013515, -4.93222922, -4.91687268, -4.88308098, -4.82986952, -4.28655415, -3.4968796, -2.710236, -1.94088143, -1.20307396, -0.511071689, 0.120867317, 0.678484973, 1.1475232, 1.51372392, 1.76282905, 1.88058051, 1.85272023, 1.66499011, 1.30313209, 0.752888078, 0, -2.7541978, -5.43424266, -8.07024831, -10.6923285, -13.3305968, -16.0151671, -18.7761531, -21.6436684, -24.6478268, -27.8187421, -31.1865278, -34.7812978, -38.6331657, -42.7722453, -47.2286503, -52.0324944, -57.1832056, -62.7065137, -68.701941, -75.1930257, -82.2033058, -89.7563197, -97.8756054, -106.584701, -115.907145, -125.866476, -136.486231, -147.789949, -159.801168, -172.933835, -186.67793, -200.939868, -215.626063, -231.139016, -247.137099, -263.146398, -278.924502, -294.229001, -308.817487, -322.447549, -339.93101, -354.432983, -364.722186, -369.567339, -367.73716, -362.461866, -346.933303, -318.804813, -275.729742, -215.361432, -143.558048, -39.7399997, 99.5804775, 281.738945, 521.926888, 831.80864, 1223.04854, 1687.53388, 2261.20082, 3001.56046, 3935.78413, 5110.11387, 6531.53509, 8243.23901, 10349.7656, 12927.4948, 15992.7828, 19606.4478, 24049.3947, 29478.0348, 35891.2925, 43627.295, 52896.9682, 63911.2384, 76881.0316, 99364.583, 122388.682, 144327.428, 163554.92, 178445.257, 181733.061, 181191.558, 177611.765, 173491.258, 168945.54, 163379.05, 156942.847, 149787.989, 142065.532, 133926.535, 127355.149, 119449.232, 110663.403, 101452.278, 92270.4741, 83572.6088, 75813.2992, 68188.9747, 61711.3909},
        {5.01139201, 4.86548005, 4.71913054, 4.57236433, 4.42520223, 4.27766511, 4.12977379, 3.98154911, 3.83301191, 3.68418303, 3.53508331, 3.38573359, 3.2361547, 3.08636748, 2.93639277, 2.78625141, 2.63596424, 2.48555209, 2.33503581, 2.18443623, 2.03377419, 1.88307053, 1.73234609, 1.5816217, 1.43091821, 1.28025645, 1.12965726, 0.979141473, 0.828729939, 0.67844349, 0.528302965, 0.378329202, 0.228543039, 0.0789653155, -0.0718723347, -0.223173148, -0.374122462, -0.52465111, -0.674689927, -0.824169746, -0.973021403, -1.12117573, -1.26856356, -1.41511574, -1.56076308, -1.70543644, -1.84906663, -1.9915845, -2.13292088, -2.27300661, -2.41177252, -2.55463842, -2.69592717, -2.83545063, -2.97302063, -3.10844902, -3.24154764, -3.37212834, -3.50000295, -3.62498333, -3.74688131, -3.86550874, -3.98067747, -4.09219934, -4.19988618, -4.30354985, -4.4030022, -4.51116715, -4.64065384, -4.7642314, -4.88086166, -4.98950645, -5.08912763, -5.17868703, -5.25714648, -5.32346784, -5.37661294, -5.41554361, -5.4392217, -5.44660905, -5.4366675, -5.40835889, -5.36064505, -5.29248783, -4.71991511, -3.89351358, -3.06967442, -2.26303159, -1.48821907, -0.759870834, -0.0926208618, 0.498896874, 1.0000484, 1.39619974, 1.67271692, 1.81496596, 1.80831289, 1.63812373, 1.28976452, 0.748601264, 0, -2.93141128, -5.76471689, -8.53183584, -11.2646872, -13.9951899, -16.7552629, -19.5768254, -22.4917964, -25.5320948, -28.7296396, -32.1163499, -35.7241448, -39.5849431, -43.730664, -48.1932265, -53.0045495, -58.0758418, -63.4217781, -69.3473125, -75.8584977, -82.9613861, -90.6620303, -98.9664828, -107.880796, -117.411023, -127.563215, -138.343426, -149.757708, -161.812113, -174.785893, -188.311729, -202.313713, -216.715939, -232.085606, -248.025255, -263.965928, -279.638788, -294.774994, -309.105708, -322.36209, -339.034745, -352.646863, -361.99841, -365.889351, -363.11965, -358.001154, -342.443977, -313.870116, -269.701566, -207.360321, -131.607361, -23.5558309, 119.147803, 303.405632, 545.406377, 857.169252, 1250.71347, 1720.56954, 2300.22009, 3042.22633, 3968.16313, 5129.53458, 6536.69089, 8234.48878, 10326.2543, 12888.4751, 15941.291, 19546.5808, 23972.8326, 29371.6647, 35736.2166, 43400.9665, 52572.3908, 63456.9658, 76261.1677, 98392.7756, 121056.637, 142658.904, 161605.725, 176303.252, 179652.721, 179235.141, 175815.91, 171776.368, 167242.272, 161666.07, 155207.185, 148025.038, 140279.051, 132128.647, 125338.126, 117412.325, 108757.933, 99781.6396, 90890.1335, 82490.1039, 74988.2398, 67051.5105, 60583.7949},
        {4.90545191, 4.73571817, 4.56633556, 4.39730513, 4.22862795, 4.06030507, 3.89233754, 3.72472643, 3.55747279, 3.39057768, 3.22404215, 3.05786727, 2.89205408, 2.72660365, 2.56151703, 2.39679527, 2.23243944, 2.0684506, 1.90482979, 1.74157807, 1.57869651, 1.41618616, 1.25404807, 1.09228331, 0.930892922, 0.769877971, 0.609239514, 0.448978605, 0.289096302, 0.129593663, -0.0295282575, -0.188268401, -0.346625712, -0.504599134, -0.665443779, -0.827462017, -0.98891708, -1.14970224, -1.30971078, -1.46883596, -1.62697107, -1.78400937, -1.93984414, -2.09436866, -2.24747619, -2.39906002, -2.54901342, -2.69722965, -2.84360201, -2.98802375, -3.13038816, -3.27277446, -3.41284259, -3.55043843, -3.68540787, -3.8175968, -3.9468511, -4.07301666, -4.19593937, -4.31546512, -4.43143978, -4.54370925, -4.65211942, -4.75651617, -4.85674539, -4.95265297, -5.04408479, -5.14551196, -5.27105447, -5.39039475, -5.50243061, -5.60605983, -5.70018021, -5.78368954, -5.85548561, -5.91446622, -5.95952916, -5.98957222, -6.00349319, -6.00018987, -5.97856006, -5.93750153, -5.8759121, -5.79268954, -5.19037065, -4.326474, -3.46448407, -2.6194065, -1.80624696, -1.04001112, -0.33570463, 0.29166684, 0.827097629, 1.25558208, 1.56211452, 1.7316893, 1.74930074, 1.5999432, 1.268611, 0.74029849, 0, -3.03424474, -5.95998688, -8.80995255, -11.6168679, -14.4134589, -17.2324519, -20.1065728, -23.0685479, -26.1511033, -29.386965, -32.8088592, -36.449512, -40.3416495, -44.5179979, -49.0112833, -53.8542318, -58.8814405, -64.0973315, -69.9827797, -76.528055, -83.7234273, -91.5591665, -100.025543, -109.112825, -118.811285, -129.111191, -140.002814, -151.476423, -163.522289, -176.290658, -189.553164, -203.252086, -217.329699, -232.535762, -248.40881, -264.272051, -279.825519, -294.769246, -308.803264, -321.627606, -337.260586, -349.769696, -358.010086, -360.836908, -357.105312, -352.630024, -337.566728, -309.030679, -264.137137, -200.001358, -119.873783, -7.23592799, 138.385063, 323.241608, 565.386073, 877.572887, 1272.55648, 1749.25133, 2337.39727, 3081.83222, 3995.85865, 5138.5079, 6524.44161, 8200.85188, 10270.0519, 12809.0533, 15842.5577, 19434.0147, 23838.6784, 29203.3075, 35510.6914, 43096.784, 52165.1026, 62919.1645, 75562.4869, 97378.5674, 119720.948, 141023.15, 159718.697, 174241.111, 177634.669, 177314.968, 174025.205, 170045.053, 165509.523, 159913.429, 153423.793, 146207.637, 138431.981, 130263.847, 122840.015, 114761.95, 106307.593, 97754.885, 89381.7653, 81466.1749, 74286.0539, 65962.517, 59324.1932},
        {4.83189122, 4.63650004, 4.44226951, 4.24918299, 4.05722379, 3.86637525, 3.67662069, 3.48794345, 3.30032686, 3.11375424, 2.92820893, 2.74367425, 2.56013355, 2.37757013, 2.19596734, 2.01530851, 1.83557696, 1.65675603, 1.47882905, 1.30177934, 1.12559023, 0.950245059, 0.775727154, 0.602019844, 0.429106458, 0.256970328, 0.0855947818, -0.0850368494, -0.254941236, -0.424135048, -0.592634956, -0.760457629, -0.927619738, -1.09413795, -1.26459451, -1.43662726, -1.60781919, -1.77803881, -1.94715463, -2.11503516, -2.2815489, -2.44656437, -2.60995007, -2.77157451, -2.9313062, -3.08901365, -3.24456537, -3.39782986, -3.54867564, -3.69697122, -3.8425851, -3.98531855, -4.12510878, -4.26182575, -4.39533944, -4.5255198, -4.65223681, -4.77536042, -4.89476062, -5.01030735, -5.12187059, -5.22932031, -5.33252647, -5.43135904, -5.52568798, -5.61538326, -5.70031485, -5.79643011, -5.9192904, -6.03560726, -6.14420841, -6.24392159, -6.33357453, -6.41199498, -6.47801065, -6.53044929, -6.56813862, -6.58990639, -6.59458032, -6.58098816, -6.54795762, -6.49431646, -6.41889239, -6.32051316, -5.68805352, -4.78604385, -3.88514381, -3.00073009, -2.14817938, -1.34286836, -0.600173725, 0.0645278385, 0.635859641, 1.09844499, 1.43690721, 1.63586959, 1.67995546, 1.55378813, 1.2419909, 0.729187085, 0, -3.05318913, -6.00358263, -8.88350123, -11.7252657, -14.5611967, -17.423615, -20.3448413, -23.3571964, -26.493001, -29.7845758, -33.2642416, -36.9643191, -40.917129, -45.154992, -49.710229, -54.6151606, -59.6541038, -64.8239504, -70.7179818, -77.3153749, -84.5953069, -92.536955, -101.119496, -110.322107, -120.123966, -130.504249, -141.442133, -152.916796, -164.907415, -177.47308, -190.483753, -203.894637, -217.660936, -232.687977, -248.455901, -264.191482, -279.564883, -294.24627, -307.905808, -320.213662, -334.696998, -345.99511, -353.023533, -354.697801, -349.93345, -345.963357, -331.306377, -302.79871, -257.276555, -191.57611, -107.510933, 9.14357436, 156.811268, 341.078836, 582.157074, 893.69085, 1289.32504, 1772.96832, 2370.21452, 3117.27874, 4018.77162, 5141.07826, 6502.57779, 8152.57711, 10192.7885, 12700.8692, 15706.1916, 19273.8948, 23646.1626, 28965.0323, 35198.5486, 42689.5094, 51640.1804, 62252.8271, 74729.7149, 96254.5058, 118302.72, 139331.273, 157797.083, 172157.064, 175576.209, 175331.309, 172147.25, 168224.358, 163703.487, 158110.729, 151617.676, 144395.924, 136617.067, 128452.699, 120062.992, 111737.922, 103547.835, 95563.08, 87854.0056, 80490.9599, 73544.2909, 64887.0049, 57882.9509},
        {4.76909343, 4.54279004, 4.31864637, 4.09663245, 3.8767183, 3.65887394, 3.44306939, 3.22927467, 3.01745981, 2.80759482, 2.59964973, 2.39359456, 2.18939933, 1.98703407, 1.78646878, 1.5876735, 1.39061824, 1.19527304, 1.0016079, 0.809592852, 0.619197916, 0.430393113, 0.243148465, 0.0574339936, -0.126780281, -0.309524336, -0.49082815, -0.670721703, -0.849234972, -1.02639794, -1.20224057, -1.37679287, -1.55008479, -1.72214632, -1.89698937, -2.07256921, -2.24679321, -2.41956216, -2.59077679, -2.76033787, -2.92814615, -3.0941024, -3.25810736, -3.4200618, -3.57986648, -3.73742214, -3.89262955, -4.04538946, -4.19560263, -4.34316982, -4.48799179, -4.6322457, -4.77350655, -4.91162576, -5.04645472, -5.17784485, -5.30564755, -5.42971424, -5.54989633, -5.66604522, -5.77801232, -5.88564905, -5.9888068, -6.08733699, -6.18109104, -6.26992034, -6.3536763, -6.4492762, -6.57322754, -6.6901847, -6.79889277, -6.89809685, -6.98654203, -7.06297343, -7.12613614, -7.17477525, -7.20763587, -7.2234631, -7.22100203, -7.19899776, -7.1561954, -7.09134003, -7.00317677, -6.8904507, -6.22637368, -5.28351131, -4.34074572, -3.41387656, -2.51870349, -1.67102619, -0.886644318, -0.181357533, 0.4290345, 0.928732116, 1.30193565, 1.53284544, 1.60566182, 1.50458514, 1.21381571, 0.717553887, 0, -2.88559612, -5.71333329, -8.51215243, -11.3109944, -14.1388001, -17.0245105, -19.9970664, -23.0854087, -26.3184783, -29.7252161, -33.334563, -37.1754599, -41.2768477, -45.6676673, -50.3768595, -55.4333653, -60.6705176, -66.0893134, -72.16114, -78.8730223, -86.2119853, -94.1650538, -102.719253, -111.861607, -121.579142, -131.858881, -142.687851, -154.053076, -165.941581, -178.503826, -191.503474, -204.878521, -218.566959, -233.419558, -248.917924, -264.30026, -279.230726, -293.373487, -306.392703, -317.952537, -331.508075, -341.778796, -347.68716, -348.155629, -342.106661, -336.379812, -320.001173, -289.913931, -243.061271, -176.386379, -91.9830295, 24.6513417, 171.33872, 353.606472, 592.713783, 902.856573, 1298.23076, 1785.50932, 2388.12087, 3137.70093, 4036.52898, 5150.12913, 6494.16181, 8116.7315, 10120.9175, 12584.5646, 15540.1556, 19053.7747, 23357.8349, 28590.6642, 34697.7954, 42040.0126, 50822.381, 61249.9654, 73527.8307, 94790.2334, 116580.749, 137372.145, 155637.188, 169848.645, 173241.089, 173017.612, 169895.92, 166075.013, 161686.256, 156243.804, 149909.881, 142846.71, 135216.514, 127181.516, 117170.951, 108451.985, 100681.381, 93515.8993, 86612.3013, 79627.3479, 72217.8005, 63801.5855, 56056.9625},
        {4.73115839, 4.47033419, 4.21292215, 3.95887226, 3.7081345, 3.46065886, 3.21639531, 2.97529385, 2.73730444, 2.50237708, 2.27046174, 2.04150842, 1.81546708, 1.59228772, 1.37192031, 1.15431483, 0.939421283, 0.727189633, 0.517569867, 0.310511969, 0.105965919, -0.0961182995, -0.295790704, -0.493101313, -0.688100144, -0.880837214, -1.07136254, -1.25972614, -1.44597804, -1.63016824, -1.81234678, -1.99256366, -2.1708689, -2.34731252, -2.52519253, -2.70286663, -2.8786535, -3.05249774, -3.22434397, -3.3941368, -3.56182086, -3.72734074, -3.89064108, -4.05166647, -4.21036154, -4.3666709, -4.52053917, -4.67191095, -4.82073087, -4.96694353, -5.11049355, -5.25719794, -5.4010016, -5.54172187, -5.67917606, -5.81318146, -5.94355541, -6.0701152, -6.19267816, -6.3110616, -6.42508282, -6.53455914, -6.63930787, -6.73914633, -6.83389183, -6.92336168, -7.00737319, -7.10344214, -7.22865938, -7.34634046, -7.45515538, -7.55377412, -7.64086668, -7.71510306, -7.77515323, -7.81968721, -7.84737498, -7.85688653, -7.84689185, -7.81606095, -7.76306381, -7.68657042, -7.58525078, -7.45777488, -6.76237045, -5.77929609, -4.79530611, -3.82660887, -2.88941277, -1.9999262, -1.17435753, -0.428915147, 0.220192562, 0.756757215, 1.16457043, 1.42742382, 1.529109, 1.4534176, 1.18414122, 0.705071479, 0, -2.65437361, -5.30834721, -7.98649343, -10.7133849, -13.5135943, -16.4116941, -19.4322572, -22.599856, -25.9390632, -29.4744515, -33.2305934, -37.2320616, -41.5034288, -46.0692676, -50.9541505, -56.1826503, -61.6683342, -67.419917, -73.7130548, -80.5485333, -87.9271383, -95.8496557, -104.316871, -113.32957, -122.888539, -132.994564, -143.648429, -154.850921, -166.602827, -179.200682, -192.241079, -205.636079, -219.297741, -233.963926, -249.133794, -264.086292, -278.485739, -291.996457, -304.282765, -315.008983, -327.674751, -336.941798, -341.724057, -340.935458, -333.489933, -325.251816, -306.447036, -274.251923, -225.842808, -158.396022, -74.8455239, 40.4270848, 185.19242, 365.124628, 602.034231, 910.486847, 1305.0481, 1794.22536, 2400.01936, 3151.03982, 4049.42111, 5157.13835, 6485.27212, 8079.65314, 10044.9943, 12458.9807, 15356.3981, 18803.8095, 23025.075, 28155.0786, 34114.2808, 41284.9772, 49876.2754, 60097.283, 72157.1075, 93149.901, 114678.465, 135230.647, 153294.292, 167357.246, 170716.495, 170511.322, 167454.226, 163760.508, 159547.811, 154307.614, 148188.869, 141340.529, 133911.545, 126050.872, 114420.496, 105326.029, 97990.2114, 91635.782, 85485.4807, 78762.0467, 70688.2197, 62654.7769, 54014.7837},
        {4.7216673, 4.42286809, 4.12908159, 3.84022385, 3.55621091, 3.27695882, 3.00238362, 2.73240135, 2.46692806, 2.20587979, 1.94917258, 1.69672247, 1.44844551, 1.20425774, 0.964075205, 0.727813946, 0.495390007, 0.26671943, 0.0417182591, -0.179697463, -0.397611694, -0.612108389, -0.823271507, -1.031185, -1.23593284, -1.43759896, -1.63626734, -1.83202192, -2.02494667, -2.21512554, -2.40264248, -2.58758147, -2.77002644, -2.95006136, -3.13094063, -3.3110963, -3.4889223, -3.66439972, -3.83750963, -4.00823308, -4.17655115, -4.34244492, -4.50589545, -4.66688381, -4.82539107, -4.98139831, -5.13488659, -5.28583698, -5.43423055, -5.58004838, -5.72327152, -5.87328681, -6.02046565, -6.16458524, -6.30542273, -6.4427553, -6.57636013, -6.70601438, -6.83149523, -6.95257984, -7.0690454, -7.18066906, -7.28722801, -7.38849942, -7.48426046, -7.57428829, -7.6583601, -7.75418579, -7.8790447, -7.99572909, -8.10285363, -8.19903297, -8.2828818, -8.35301477, -8.40804655, -8.4465918, -8.46726519, -8.46868139, -8.44945506, -8.40820087, -8.34353349, -8.25406757, -8.13841779, -7.99519881, -7.26968459, -6.24879892, -5.22611633, -4.21820847, -3.24164695, -2.31300343, -1.44884954, -0.665756906, 0.0197028274, 0.590958025, 1.03143705, 1.32456827, 1.45378005, 1.40250074, 1.15415873, 0.692182357, 0, -2.41600204, -4.88955393, -7.44075383, -10.0896999, -12.8564904, -15.7612234, -18.8239971, -22.0649097, -25.5040593, -29.1615442, -33.0574626, -37.2119125, -41.6449922, -46.3767998, -51.4274336, -56.8169916, -62.5439541, -68.6253308, -75.130118, -82.0737815, -89.4717868, -97.3395998, -105.692686, -114.546512, -123.916542, -133.818243, -144.26708, -155.27852, -166.868027, -179.485359, -192.552451, -205.954479, -219.576623, -234.038637, -248.85841, -263.357946, -277.202051, -290.05553, -301.583189, -311.449834, -323.231901, -331.492069, -335.129822, -333.044649, -324.136034, -313.070116, -291.538067, -256.997711, -206.906871, -138.723371, -56.4489777, 57.000581, 199.738247, 377.691953, 612.74523, 919.617874, 1313.02968, 1802.47333, 2409.08371, 3159.69798, 4057.54304, 5159.8878, 6472.87912, 8039.675, 9966.14666, 12329.379, 15165.6938, 18541.6079, 22673.1018, 27691.6109, 33492.6433, 40479.8581, 48866.3578, 58865.245, 70689.622, 91381.4746, 112617.801, 132904.484, 150747.404, 164652.443, 167994.358, 167830.906, 164867.02, 161325.716, 157311.022, 152294.432, 146411.599, 139798.179, 132589.826, 124922.193, 112040.929, 102690.603, 95732.9639, 90029.7597, 84442.7389, 77833.6501, 69064.2418, 61403.3708, 51815.5142},
        {4.74779677, 4.40607548, 4.07178501, 3.74476648, 3.42486104, 3.11190979, 2.80575388, 2.50623444, 2.21319258, 1.92646945, 1.64590616, 1.37134386, 1.10262366, 0.839586704, 0.58207411, 0.329927011, 0.0829865361, -0.158906186, -0.395910027, -0.628183856, -0.855886546, -1.07917697, -1.29821399, -1.51315649, -1.72416333, -1.93139339, -2.13500553, -2.33515863, -2.53201156, -2.72572319, -2.91645239, -3.10435803, -3.28959898, -3.47233412, -3.65884979, -3.84611225, -4.03101409, -4.21351534, -4.39357601, -4.57115613, -4.74621572, -4.91871481, -5.08861342, -5.25587156, -5.42044927, -5.58230656, -5.74140346, -5.89769998, -6.05115616, -6.20173201, -6.34938756, -6.50370551, -6.65481459, -6.80246624, -6.94641187, -7.08640291, -7.22219077, -7.35352689, -7.48016267, -7.60184954, -7.71833893, -7.82938225, -7.93473093, -8.03413639, -8.12735004, -8.21412332, -8.29420764, -8.3848884, -8.50302999, -8.61222135, -8.71107724, -8.79821243, -8.87224168, -8.93177977, -8.97544145, -9.00184149, -9.00959466, -8.99731572, -8.96361944, -8.90712059, -8.82643392, -8.72017422, -8.58695623, -8.42539473, -7.67347849, -6.62192727, -5.56819915, -4.52910231, -3.52144488, -2.56203503, -1.66768091, -0.855190694, -0.141372527, 0.45696543, 0.923015019, 1.23996808, 1.39101646, 1.359352, 1.12816655, 0.68065193, 0, -2.28782621, -4.66657186, -7.15357099, -9.76615763, -12.5216658, -15.4374296, -18.5307829, -21.8190599, -25.3195945, -29.0497208, -33.0267728, -37.2680846, -41.7909902, -46.6128235, -51.7509187, -57.2226098, -63.0646414, -69.2971205, -75.8934761, -82.8752018, -90.2637908, -98.0807368, -106.347533, -115.085673, -124.316651, -134.061959, -144.343091, -155.181541, -166.598801, -179.112748, -192.066487, -205.332595, -218.783653, -233.023284, -247.558544, -261.701546, -275.105553, -287.423832, -298.309648, -307.416266, -318.292622, -325.510672, -327.961487, -324.536135, -314.125687, -300.511168, -276.471201, -239.674366, -187.789244, -118.484413, -36.593175, 76.1374902, 218.620541, 397.053327, 632.5055, 939.369349, 1332.03717, 1819.56032, 2422.88586, 3168.78734, 4061.75045, 5155.70766, 6454.76807, 8000.23021, 9896.40803, 12218.6904, 15003.4416, 18315.933, 22364.5259, 27276.6418, 32927.7789, 39736.0395, 47916.0493, 57682.434, 69249.8194, 89546.3632, 110394.776, 130331.301, 147892.181, 161613.657, 165020.816, 165001.83, 162240.581, 158895.287, 155066.241, 150239.424, 144545.842, 138116.501, 131082.408, 123574.566, 110889.44, 101707.269, 94861.6404, 89186.1379, 83514.347, 76679.8527, 67516.24, 59933.2881, 49534.0311},
        {4.81153752, 4.42206221, 4.04331951, 3.67502967, 3.31691297, 2.96868969, 2.6300801, 2.30080447, 1.98058307, 1.66913619, 1.3661841, 1.07144706, 0.784645351, 0.505499251, 0.233729031, -0.0309450361, -0.288802675, -0.540123614, -0.785187577, -1.02427429, -1.25766348, -1.48563488, -1.7084682, -1.92644318, -2.13983955, -2.34893702, -2.55401533, -2.75535419, -2.95323335, -3.14793252, -3.33973142, -3.52890979, -3.71574736, -3.90052384, -4.0961952, -4.29643642, -4.49477013, -4.69106467, -4.88518841, -5.07700972, -5.26639693, -5.45321842, -5.63734254, -5.81863765, -5.99697211, -6.17221427, -6.3442325, -6.51289514, -6.67807057, -6.83962713, -6.99743319, -7.15697977, -7.31239067, -7.46341237, -7.60979133, -7.75127402, -7.88760691, -8.01853647, -8.14380917, -8.26317149, -8.37636988, -8.48315083, -8.58326079, -8.67644625, -8.76245366, -8.84102951, -8.91192025, -8.99133031, -9.09506899, -9.18893943, -9.27162122, -9.34179391, -9.39813707, -9.43933027, -9.46405307, -9.47098505, -9.45880577, -9.4261948, -9.37183171, -9.29439605, -9.19256741, -9.06502534, -8.91044942, -8.72751921, -7.95359934, -6.87983335, -5.80411135, -4.74332867, -3.71438061, -2.73416249, -1.81956964, -0.987497361, -0.254840976, 0.3615042, 0.84464285, 1.17767966, 1.34371931, 1.32586648, 1.10722587, 0.670902147, 0, -2.31298721, -4.71656388, -7.22767743, -9.86327528, -12.6403049, -15.5757136, -18.6864489, -21.9894582, -25.501689, -29.2400886, -33.2216044, -37.463184, -41.9817747, -46.794324, -51.9177792, -57.3690878, -63.1572659, -69.2981278, -75.826561, -82.7578132, -90.1071324, -97.8897663, -106.120963, -114.81597, -123.990035, -133.658407, -143.836332, -154.53906, -165.781837, -178.031612, -190.68631, -203.625666, -216.72942, -230.72352, -245.064598, -258.98363, -272.106489, -284.059051, -294.46719, -302.95678, -312.891597, -319.021992, -320.242509, -315.447695, -303.532095, -287.942389, -261.882956, -223.110307, -169.380955, -98.4514111, -15.5296156, 98.1866458, 242.782192, 424.676321, 663.221215, 971.962458, 1364.44563, 1847.84334, 2443.52255, 3179.80255, 4062.04889, 5143.18599, 6429.16445, 7960.671, 9837.2508, 12131.432, 14878.1469, 18140.1766, 22118.1538, 26934.7346, 32452.5376, 39094.2621, 47072.577, 56600.1514, 67889.6543, 87677.2335, 108021.21, 127503.382, 144705.549, 158209.511, 161780.162, 162027.477, 159597.64, 156492.774, 152822.34, 148130.831, 142555.958, 136235.432, 129306.967, 121908.274, 111137.787, 102623.703, 95572.165, 89189.3185, 82681.3072, 75254.2753, 66114.367, 58208.4472, 47214.7519},
        {4.9048938, 4.46604417, 4.04140205, 3.63055851, 3.23310464, 2.84863152, 2.47673025, 2.1169919, 1.76900756, 1.43236831, 1.10666525, 0.791489446, 0.486431995, 0.191083979, -0.0949635173, -0.372119409, -0.640792611, -0.901392039, -1.15432661, -1.40000523, -1.63883682, -1.8712303, -2.09759458, -2.31833858, -2.5338712, -2.74460138, -2.95093801, -3.15329002, -3.35206631, -3.54767582, -3.74052744, -3.9310301, -4.11959272, -4.30662419, -4.51240045, -4.72697902, -4.94017867, -5.15176358, -5.36149796, -5.56914598, -5.77447185, -5.97723976, -6.1772139, -6.37415846, -6.56783764, -6.75801563, -6.94445662, -7.1269248, -7.30518437, -7.47899952, -7.64813445, -7.81339025, -7.97347173, -8.12812061, -8.2770786, -8.42008741, -8.55688876, -8.68722436, -8.81083593, -8.92746518, -9.03685383, -9.13874359, -9.23287618, -9.31899331, -9.39683669, -9.46614805, -9.52666909, -9.59316881, -9.6800554, -9.75607795, -9.82000401, -9.87060114, -9.90663688, -9.92687881, -9.93009447, -9.91505143, -9.88051724, -9.82525945, -9.74804563, -9.64764333, -9.52282011, -9.37234352, -9.19498113, -8.98950049, -8.1952007, -7.1023512, -6.00794584, -4.928908, -3.88216106, -2.88462841, -1.95323342, -1.10489949, -0.356550001, 0.274891667, 0.772502127, 1.11935799, 1.29853589, 1.29311242, 1.0861642, 0.660767858, 0, -2.39173242, -4.86162368, -7.4270662, -10.1054524, -12.9141747, -15.8706256, -18.9921975, -22.2962828, -25.8002739, -29.5215633, -33.4775434, -37.6856066, -42.1631453, -46.927552, -51.9962192, -57.3865392, -63.0560491, -69.0135905, -75.4119431, -82.2556731, -89.5493468, -97.2975305, -105.50479, -114.175693, -123.314804, -132.92669, -143.015918, -153.587054, -164.644663, -176.564054, -188.843113, -201.375184, -214.05361, -227.765454, -241.907202, -255.610344, -268.470109, -280.081724, -290.040418, -297.941418, -306.912519, -311.941254, -311.931694, -305.787914, -292.413987, -275.13609, -247.330666, -206.796265, -151.331433, -78.7347185, 5.62033274, 120.768503, 267.96393, 453.791697, 695.720901, 1006.33394, 1398.21321, 1876.29578, 2462.54695, 3187.2724, 4056.80678, 5123.33308, 6395.14151, 7912.491, 9769.80971, 12036.6613, 14746.527, 17959.6157, 21868.4382, 26590.6677, 31980.2877, 38459.2289, 46236.4419, 55520.8773, 66521.4856, 85751.4035, 105539.349, 124518.226, 141320.938, 154580.389, 158330.169, 158868.706, 156800.04, 153947.039, 150431.29, 145866.532, 140399.744, 134177.904, 127347.991, 120056.982, 111637.8, 103927.71, 96604.4864, 89345.9061, 81829.7446, 73733.7778, 64735.7814, 56342.966, 44843.4377},
        {5.0255877, 4.53718214, 4.06636626, 3.61260685, 3.17537071, 2.75412462, 2.3483354, 1.95746982, 1.58099468, 1.21837678, 0.869082909, 0.532579863, 0.208334435, -0.10418658, -0.405516389, -0.696188197, -0.976735212, -1.24769064, -1.50958768, -1.76295955, -2.00833945, -2.24626058, -2.47725615, -2.70185937, -2.92060345, -3.13402159, -3.34264699, -3.54701286, -3.74765241, -3.94509885, -4.13988537, -4.33254519, -4.52361151, -4.71361754, -4.92985441, -5.15891355, -5.38706477, -5.61397294, -5.8393029, -6.06271953, -6.28388767, -6.5024722, -6.71813797, -6.93054984, -7.13937267, -7.34427132, -7.54491065, -7.74095552, -7.93207079, -8.11792131, -8.29817196, -8.46938953, -8.63440411, -8.79294769, -8.94475232, -9.08955, -9.22707276, -9.35705262, -9.47922159, -9.59331169, -9.69905495, -9.79618338, -9.88442901, -9.96352384, -10.0331999, -10.0931892, -10.1432238, -10.1963902, -10.2654543, -10.3226054, -10.3667099, -10.396634, -10.411244, -10.4094063, -10.3899871, -10.3518528, -10.2938696, -10.2149039, -10.1138219, -9.98948998, -9.84077439, -9.66654146, -9.46565748, -9.23698875, -8.42314214, -7.31281383, -6.20138583, -5.10577612, -4.0429027, -3.02968355, -2.08303666, -1.21988003, -0.457131636, 0.188290523, 0.69946846, 1.05948419, 1.25141971, 1.25835704, 1.0633782, 0.649565177, 0, -2.50612413, -5.06988679, -7.70967826, -10.4438888, -13.2909087, -16.2691283, -19.3969378, -22.6927274, -26.1748875, -29.8618084, -33.7718803, -37.9234935, -42.3350382, -47.0249048, -52.0114835, -57.3131646, -62.8192359, -68.5296446, -74.7547982, -81.4854223, -88.7122423, -96.4259837, -104.617372, -113.277133, -122.395991, -131.964673, -141.973903, -152.414408, -163.276912, -174.818418, -186.665739, -198.729719, -210.921201, -224.305437, -238.216064, -251.678879, -264.2584, -275.519147, -285.02564, -292.3424, -300.325532, -304.246183, -303.024865, -295.582092, -280.838379, -262.205312, -232.989562, -190.988876, -134.000997, -59.8236706, 26.2333518, 143.065672, 292.885021, 482.346266, 727.175677, 1039.02665, 1429.55257, 1901.55391, 2477.54902, 3189.71401, 4045.32988, 5095.77234, 6351.28708, 7851.68331, 9686.86397, 11923.7382, 14594.9506, 17758.527, 21598.1292, 26226.3521, 31492.8075, 37813.5323, 45392.1542, 54432.3005, 65137.5988, 83777.8816, 102974.02, 121413.091, 137782.17, 150768.333, 154686.537, 155510.725, 153801.74, 151198.418, 147834.895, 143396.143, 138038.757, 131919.329, 125194.455, 118020.729, 111997.044, 105114.724, 97532.1832, 89407.8391, 80900.107, 72167.4028, 63368.1424, 54364.9642, 42439.4338},
        {5.16508465, 4.63260642, 4.12007079, 3.62687075, 3.15239932, 2.69604949, 2.25721427, 1.83528666, 1.42965966, 1.03972629, 0.664879525, 0.304512386, -0.0419821265, -0.375211011, -0.695781265, -1.00429988, -1.30137387, -1.58761021, -1.86361591, -2.12999796, -2.38736337, -2.63631912, -2.87747222, -3.11142966, -3.33879845, -3.56018556, -3.77619802, -3.9874428, -4.19452691, -4.39805735, -4.5986411, -4.79688518, -4.99339658, -5.18878228, -5.41377636, -5.65328858, -5.89186664, -6.12913986, -6.36473754, -6.59828901, -6.82942358, -7.05777057, -7.28295929, -7.50461906, -7.7223792, -7.93586902, -8.14471784, -8.34855497, -8.54700973, -8.73971144, -8.92628942, -9.10330339, -9.2735188, -9.4366315, -9.59233736, -9.74033224, -9.88031199, -10.0119725, -10.1350095, -10.2491191, -10.3539969, -10.4493389, -10.5348409, -10.6101988, -10.6751085, -10.7292657, -10.7723664, -10.8157271, -10.8703852, -10.9120699, -10.9397237, -10.9522893, -10.948709, -10.9279256, -10.8888815, -10.8305192, -10.7517814, -10.6516104, -10.528949, -10.3827396, -10.2119248, -10.015447, -9.79224893, -9.54127303, -8.70642397, -7.57583772, -6.44432136, -5.32882948, -4.24631666, -3.2137375, -2.24804658, -1.36619849, -0.585147812, 0.0781508561, 0.606742931, 0.983673825, 1.19198895, 1.21473372, 1.03495354, 0.63569383, 0, -2.62233882, -5.28196043, -7.99823178, -10.7905198, -13.6781915, -16.6806137, -19.8171534, -23.1071776, -26.5700532, -30.2251472, -34.0918264, -38.1894579, -42.5374086, -47.1550454, -52.0617353, -57.2768452, -62.6160012, -68.0694644, -74.1174442, -80.7356488, -87.8997863, -95.585565, -103.768693, -112.424879, -121.52983, -131.059255, -140.988862, -151.29436, -161.951456, -173.098421, -184.488795, -196.049517, -207.707526, -220.672151, -234.229137, -247.322257, -259.493735, -270.285793, -279.240652, -285.900536, -292.82638, -295.622954, -293.241865, -284.634715, -268.753111, -249.434219, -219.522692, -176.748741, -118.842583, -43.534429, 44.3931759, 162.881862, 314.036871, 504.288865, 748.89957, 1059.16574, 1446.38415, 1912.90358, 2481.00238, 3182.72031, 4026.06288, 5060.05821, 6292.79361, 7762.77984, 9560.07842, 11751.0882, 14371.2027, 17478.7823, 21245.874, 25780.2172, 30935.1427, 37111.9384, 44506.4084, 53314.3568, 63731.5875, 81780.7607, 100374.11, 118250.726, 134149.699, 146810.117, 150801.056, 151811.632, 150368.085, 147974.702, 144770.58, 140487.296, 135284.32, 129321.125, 122757.183, 115751.965, 111007.557, 104645.696, 97049.6692, 88602.7664, 79688.2763, 70689.4875, 61989.6889, 52296.1081, 39992.5847},
        {5.322536, 4.75194165, 4.2025009, 3.6735893, 3.1645824, 2.67485575, 2.2037849, 1.7507454, 1.31511279, 0.896262624, 0.493570451, 0.106411818, -0.265837725, -0.623802631, -0.968107349, -1.29937633, -1.61823403, -1.9253049, -2.22121339, -2.50658394, -2.78204102, -3.04820907, -3.30571255, -3.5551759, -3.79722358, -4.03248004, -4.26156973, -4.4851171, -4.7037466, -4.91808269, -5.12874981, -5.33637242, -5.54157497, -5.74498191, -5.97677109, -6.22216865, -6.46604552, -6.70806707, -6.9478987, -7.18520579, -7.41965373, -7.65090791, -7.87863371, -8.10249652, -8.32216173, -8.53729471, -8.74756087, -8.95262558, -9.15215424, -9.34581222, -9.53326492, -9.71572525, -9.89127751, -10.0595536, -10.2201852, -10.3728044, -10.5170428, -10.6525324, -10.7789049, -10.8957923, -11.0028263, -11.0996388, -11.1858616, -11.2611267, -11.3250657, -11.3773106, -11.4174932, -11.4551679, -11.4996508, -11.5301093, -11.545532, -11.5449074, -11.5272242, -11.4914709, -11.4366362, -11.3617087, -11.2656769, -11.1475294, -11.006255, -10.840842, -10.6502793, -10.4335552, -10.1896586, -9.91757787, -9.05982238, -7.90532517, -6.74970775, -5.61001756, -4.503302, -3.44660851, -2.45698452, -1.55147744, -0.747134703, -0.0610037306, 0.489868052, 0.888433221, 1.11764435, 1.16045402, 0.999814799, 0.618679267, 0, -2.7260677, -5.47226863, -8.25871391, -11.1055146, -14.032782, -17.060627, -20.2091608, -23.4984945, -26.9487393, -30.5800062, -34.4124063, -38.4660509, -42.761051, -47.3175176, -52.155562, -57.2952952, -62.4792493, -67.6878813, -73.5691596, -80.0837142, -87.192175, -94.8551718, -103.033335, -111.687293, -120.777677, -130.265117, -140.110242, -150.273683, -160.716069, -171.467779, -182.394119, -193.434795, -204.529512, -216.98258, -230.051401, -242.629484, -254.247153, -264.434731, -272.722542, -278.640909, -284.44789, -286.115894, -282.64156, -273.021524, -256.252423, -236.896765, -206.993848, -164.148841, -105.966912, -30.053232, 59.7718447, 179.714675, 330.647674, 518.479589, 759.401678, 1064.98856, 1446.81486, 1908.66019, 2471.66576, 3165.47642, 3998.50826, 5015.82986, 6219.0246, 7644.39189, 9387.05559, 11515.1382, 14070.4663, 17114.345, 20804.4973, 25244.1068, 30297.4628, 36343.3961, 43567.7846, 52156.5061, 62295.4383, 79765.6034, 97758.9488, 115062.566, 130463.548, 142748.984, 146704.026, 147785.669, 146496.121, 144264.779, 141226.84, 137131.177, 132132.307, 126384.744, 120043.004, 113261.602, 108463.202, 102252.374, 94931.3573, 86802.3939, 78167.7245, 69329.5901, 60590.2316, 50165.9918, 37528.9748},
        {5.5060844, 4.89773039, 4.31145025, 3.74661772, 3.20260656, 2.67879051, 2.17454333, 1.68923875, 1.22225054, 0.772952433, 0.340718185, -0.0750784569, -0.475063744, -0.859863928, -1.23010526, -1.58641399, -1.92941637, -2.25973865, -2.57800709, -2.88484793, -3.18088742, -3.46675182, -3.74306738, -4.01046035, -4.26955698, -4.52098352, -4.76536623, -5.00333135, -5.23550513, -5.46251384, -5.68498371, -5.903541, -6.11881196, -6.33142284, -6.56989264, -6.82031454, -7.0684465, -7.31400959, -7.5567249, -7.79631349, -8.03249645, -8.26499485, -8.49352978, -8.71782231, -8.93759353, -9.1525645, -9.3624563, -9.56699002, -9.76588674, -9.95886753, -10.1456535, -10.3332983, -10.5140315, -10.6874152, -10.8530114, -11.0103824, -11.1590902, -11.2986969, -11.4287647, -11.5488557, -11.658532, -11.7573557, -11.844889, -11.920694, -11.9843327, -12.0353674, -12.0733601, -12.1061319, -12.1413055, -12.1614097, -12.1654712, -12.1525167, -12.1215728, -12.0716664, -12.0018242, -11.9110728, -11.7984391, -11.6629497, -11.5036314, -11.3195108, -11.1096148, -10.87297, -10.6086031, -10.315541, -9.43467967, -8.25585068, -7.07559391, -5.9110485, -4.77935359, -3.69764833, -2.68307183, -1.75276326, -0.923861733, -0.213506398, 0.361163607, 0.783009145, 1.03489108, 1.09967027, 0.960207578, 0.599363868, 0, -2.82583117, -5.65479586, -8.50772986, -11.4054689, -14.3688488, -17.4187052, -20.575874, -23.861191, -27.2954918, -30.8996122, -34.694388, -38.700655, -42.939249, -47.4310056, -52.1967607, -57.25735, -62.282469, -67.2427905, -72.9533962, -79.3598776, -86.4078262, -94.0428332, -102.21049, -110.856388, -119.926119, -129.365274, -139.119445, -149.134222, -159.355198, -169.709435, -180.167847, -190.681584, -201.201796, -213.106048, -225.632283, -237.638065, -248.646618, -258.181168, -265.76494, -270.921158, -275.617225, -276.188308, -271.679425, -261.135591, -243.601824, -224.420989, -194.765791, -152.106783, -93.9145206, -17.6595566, 73.8994707, 195.043305, 345.147247, 529.523962, 765.608232, 1065.39226, 1440.86823, 1897.70235, 2455.65055, 3141.36986, 3963.40591, 4962.8256, 6134.22301, 7511.60149, 9195.74108, 11256.7702, 13743.3729, 16720.1178, 20330.2036, 24672.3513, 29622.7256, 35536.8435, 42590.465, 50959.3498, 60819.258, 77716.2942, 95113.2872, 111843.411, 126739.839, 138635.745, 142517.598, 143629.747, 142453.952, 140362.971, 137485.881, 133580.507, 128794.488, 123275.462, 117171.067, 110628.941, 105329.821, 99153.1408, 92215.094, 84631.8746, 76519.6764, 67994.6932, 59173.1188, 48013.4844, 35117.0834},
        {5.71771039, 5.07051544, 4.44621451, 3.84418639, 3.26380988, 2.7044638, 2.16552693, 1.64637809, 1.14639607, 0.664959685, 0.201447728, -0.244760994, -0.674287681, -1.08775353, -1.48577974, -1.8689875, -2.23799801, -2.59343248, -2.9359121, -3.26605806, -3.58449157, -3.89183382, -4.188706, -4.47572932, -4.75352498, -5.02271417, -5.28391808, -5.53775792, -5.78485489, -6.02583017, -6.26130498, -6.49190049, -6.71823793, -6.94093847, -7.18637846, -7.4417598, -7.69397484, -7.94280897, -8.18804758, -8.42947607, -8.66687983, -8.90004426, -9.12875474, -9.35279668, -9.57195547, -9.7860165, -9.99476516, -10.1979868, -10.395467, -10.5869909, -10.772344, -10.9648127, -11.1503888, -11.3285648, -11.4988337, -11.660688, -11.8136205, -11.9571239, -12.090691, -12.2138144, -12.3259869, -12.4267012, -12.51545, -12.591726, -12.655022, -12.7048306, -12.7406447, -12.7686526, -12.7948822, -12.8050165, -12.7981143, -12.7732342, -12.7294349, -12.6657751, -12.5813133, -12.4751082, -12.3462186, -12.193703, -12.0166201, -11.8140287, -11.5849872, -11.3285545, -11.0437891, -10.7297497, -9.82581214, -8.62264866, -7.41765848, -6.22806304, -5.07108378, -3.96394214, -2.92385957, -1.96805751, -1.1137574, -0.37818069, 0.221451183, 0.667916775, 0.943994642, 1.03246334, 0.916101428, 0.577687463, 0, -2.91450377, -5.8164141, -8.72719407, -11.6683068, -14.6612152, -17.7273826, -20.8882719, -24.1653462, -27.5800686, -31.1539022, -34.9083101, -38.8647553, -43.044701, -47.4696101, -52.1609458, -57.1401712, -62.0084034, -66.727056, -72.2683746, -78.5637482, -85.5445657, -93.142216, -101.288088, -109.913571, -118.950054, -128.328925, -137.981574, -147.839389, -157.83376, -167.798758, -177.798773, -187.794388, -197.746188, -209.081612, -221.027818, -232.420798, -242.78108, -251.629191, -258.485658, -262.871007, -266.491294, -266.016356, -260.539119, -249.152506, -230.949441, -212.003464, -182.681728, -140.332003, -82.3020593, -5.93966653, 87.0327482, 208.988876, 357.771197, 538.089915, 768.676832, 1061.96834, 1430.40083, 1881.66925, 2434.02904, 3110.90808, 3920.64767, 4900.71577, 6039.28772, 7368.20256, 8993.28505, 10986.3993, 13402.5528, 16309.2393, 19835.7605, 24076.4563, 28917.7109, 34693.8126, 41571.0854, 49715.8537, 59294.4417, 75631.5744, 92445.5749, 108613.167, 123011.076, 134516.024, 138300.974, 139415.586, 138323.912, 136354.274, 133628.615, 129908.765, 125334.671, 120046.282, 114183.547, 107886.414, 101770.047, 95547.5237, 89072.8399, 82199.9916, 74782.9749, 66675.7859, 57732.4206, 45871.0987, 32796.3346},
        {5.95939447, 5.27083957, 4.60608933, 3.96452565, 3.34553045, 2.74848563, 2.1727731, 1.61777477, 1.08287255, 0.56744835, 0.0708840805, -0.407438351, -0.868137034, -1.31183006, -1.73913552, -2.1506715, -2.5470561, -2.92890741, -3.2968435, -3.65148249, -3.99344246, -4.32334149, -4.64179768, -4.94942912, -5.2468539, -5.53469011, -5.81355584, -6.08406919, -6.34684824, -6.60251108, -6.8516758, -7.0949605, -7.33298326, -7.56636218, -7.81946602, -8.08053797, -8.33753576, -8.59030672, -8.83869818, -9.08255749, -9.32173198, -9.55606898, -9.78541583, -10.0096199, -10.2285284, -10.4419888, -10.6498485, -10.8519546, -11.0481546, -11.2382958, -11.4222255, -11.6190587, -11.8089573, -11.9913512, -12.1656698, -12.331343, -12.4878003, -12.6344714, -12.7707859, -12.8961735, -13.0100639, -13.1118867, -13.2010715, -13.2770481, -13.339246, -13.387095, -13.4200246, -13.4427636, -13.4599134, -13.4599749, -13.4420342, -13.4051773, -13.3484904, -13.2710597, -13.1719713, -13.0503113, -12.9051659, -12.7356211, -12.5407632, -12.3196783, -12.0714525, -11.795172, -11.4899228, -11.1547912, -10.2280361, -9.00095356, -7.77158014, -6.5572019, -5.37510489, -4.24257517, -3.17689877, -2.19536177, -1.31525021, -0.553850147, 0.0715523699, 0.543671286, 0.845220549, 0.958914104, 0.8674659, 0.553589883, 0, -2.98496007, -5.94399532, -8.899021, -11.8719523, -14.8847046, -17.9591931, -21.1173329, -24.3810394, -27.7722278, -31.3128133, -35.0247113, -38.9298368, -43.0501053, -47.4074318, -52.0237318, -56.9209203, -61.6397956, -66.1335419, -71.5123155, -77.6949351, -84.6002195, -92.1469873, -100.254057, -108.840248, -117.824379, -127.125268, -136.661734, -146.352595, -156.116672, -165.711116, -175.27569, -184.777711, -194.184498, -204.948328, -216.29404, -227.050484, -236.739489, -244.882886, -251.002506, -254.620179, -257.227004, -255.776199, -249.404306, -237.247862, -218.443407, -199.640765, -170.584868, -128.533934, -70.7461806, 5.52017483, 99.4283712, 221.672509, 368.755129, 544.84538, 769.765083, 1056.30833, 1417.2692, 1862.20008, 2407.87352, 3074.59852, 3870.12536, 4829.17075, 5935.11758, 7217.98903, 8786.83768, 10714.4404, 13060.6362, 15894.8477, 19333.9357, 23467.9272, 28189.1984, 33815.835, 40506.2818, 48418.9833, 57712.3844, 73510.1854, 89764.2612, 105391.743, 119309.76, 130435.445, 134113.355, 135214.908, 134188.338, 132323.681, 129735.956, 126189.431, 121816.662, 116750.204, 111122.614, 105066.449, 97946.5129, 91635.0494, 85676.5554, 79615.5277, 72996.4628, 65363.8572, 56262.2077, 43771.3472, 30606.1525},
        {6.22604782, 5.47992078, 4.76026967, 4.06640436, 3.3976347, 2.75327056, 2.13262181, 1.53499829, 0.959709883, 0.406066439, -0.126622177, -0.639046105, -1.13189548, -1.60586045, -2.06163114, -2.49989769, -2.92135025, -3.32667895, -3.71657393, -4.09172533, -4.45282329, -4.80055793, -5.13561942, -5.45869787, -5.77048344, -6.07166625, -6.36293645, -6.64498418, -6.91849956, -7.18417276, -7.44269389, -7.6947531, -7.94104052, -8.18224631, -8.4456661, -8.71812767, -8.98613902, -9.24952687, -9.50811797, -9.76173906, -10.0102169, -10.2533781, -10.4910496, -10.723058, -10.9492301, -11.1693926, -11.3833722, -11.5909958, -11.7920899, -11.9864815, -12.1739971, -12.3728576, -12.5640969, -12.747143, -12.9214239, -13.0863677, -13.2414023, -13.3859556, -13.5194558, -13.6413307, -13.7510085, -13.847917, -13.9314842, -14.0011382, -14.056307, -14.0964185, -14.1209007, -14.1344121, -14.1414855, -14.1307615, -14.101329, -14.052277, -13.9826945, -13.8916702, -13.7782932, -13.6416524, -13.4808367, -13.294935, -13.0830363, -12.8442295, -12.5776034, -12.2822472, -11.9572496, -11.6016995, -10.6527324, -9.40223384, -8.14884378, -6.90988761, -5.70269074, -4.54457857, -3.45287651, -2.44490995, -1.5380043, -0.749484959, -0.0966773285, 0.403093188, 0.732501189, 0.874221273, 0.810928037, 0.52529608, 0, -3.05476635, -6.06866516, -9.06415155, -12.0636806, -15.0897074, -18.1646871, -21.3110748, -24.5513256, -27.9078946, -31.4032368, -35.0598074, -38.9000615, -42.9464543, -47.2214407, -51.7474759, -56.547015, -61.1293822, -65.4437282, -70.6731613, -76.7301801, -83.5272829, -90.9769682, -98.9917347, -107.484081, -116.366505, -125.551505, -134.95158, -144.479229, -154.04695, -163.397229, -172.674914, -181.843509, -190.866514, -201.173044, -211.993795, -222.168634, -231.221381, -238.675856, -244.055878, -246.885265, -248.57873, -246.193711, -238.884069, -225.803666, -206.106365, -185.872712, -155.598214, -112.70506, -54.6154433, 21.2484469, 113.520643, 233.542901, 378.689027, 553.295659, 775.915224, 1058.71755, 1413.87247, 1850.22489, 2384.02522, 3035.15104, 3809.49941, 4744.08151, 5828.50989, 7093.30475, 8638.30199, 10531.5205, 12828.7567, 15594.8607, 18939.4979, 22946.9835, 27470.932, 32861.852, 39283.9168, 46901.2996, 55878.1738, 71272.7317, 87131.6227, 102395.516, 116005.079, 126900.98, 130543.143, 131673.479, 130724.406, 128917.014, 126371.536, 122879.595, 118577.854, 113602.978, 108091.633, 102180.482, 93884.6567, 87511.2567, 82208.814, 77125.8605, 71410.9279, 64212.5478, 54679.2521, 41858.7585, 28748.1304},
        {6.52818835, 5.72163965, 4.94440268, 4.19570906, 3.47479044, 2.78087846, 2.11320476, 1.47100098, 0.85349876, 0.259929741, -0.310474437, -0.858482133, -1.38486171, -1.89038152, -2.37580993, -2.8419153, -3.28946598, -3.71923034, -4.13197674, -4.52847353, -4.90948908, -5.27579174, -5.62814988, -5.96733185, -6.29410601, -6.60924073, -6.91350437, -7.20766527, -7.49249181, -7.76875234, -8.03721522, -8.29864882, -8.55382148, -8.80350158, -9.07841635, -9.36372458, -9.64422494, -9.91971357, -10.1899866, -10.4548402, -10.7140705, -10.9674736, -11.2148456, -11.4559827, -11.6906811, -11.9187368, -12.139946, -12.3541049, -12.5610095, -12.7604561, -12.9522407, -13.1528408, -13.3450096, -13.5281816, -13.7017914, -13.8652734, -14.0180621, -14.1595922, -14.2892981, -14.4066142, -14.5109752, -14.6018156, -14.6785698, -14.7406724, -14.787558, -14.8186609, -14.8334158, -14.8364898, -14.8324245, -14.8098625, -14.767899, -14.7056292, -14.6221486, -14.5165524, -14.3879359, -14.2353944, -14.0580233, -13.8549178, -13.6251732, -13.3678848, -13.082148, -12.767058, -12.4217102, -12.0451998, -11.074713, -9.80211869, -8.52603946, -7.26380564, -6.03274756, -4.85019555, -3.73347995, -2.69993109, -1.76687931, -0.951654944, -0.271588322, 0.255990218, 0.613750343, 0.784361719, 0.750494009, 0.494816881, 0, -3.08697788, -6.1238521, -9.13321613, -12.1376634, -15.1597873, -18.2221814, -21.3474391, -24.5581538, -27.8769189, -31.326328, -34.9289744, -38.7074517, -42.6843531, -46.8822723, -51.3238026, -56.0315375, -60.4961443, -64.664227, -69.7615453, -75.6967084, -82.3783257, -89.7150064, -97.6153598, -105.987995, -114.741522, -123.784549, -133.025686, -142.373542, -151.736726, -160.86223, -169.879541, -178.745753, -187.417961, -197.278187, -207.571061, -217.161235, -225.578324, -232.351946, -237.011716, -239.087252, -239.981842, -236.781183, -228.648302, -214.746228, -194.23799, -172.455147, -140.850064, -97.0436374, -38.6567621, 36.6896656, 126.93644, 244.325576, 387.380527, 560.6821, 781.178202, 1060.26417, 1409.33533, 1836.16263, 2356.58952, 2990.41331, 3741.35645, 4649.68263, 5713.24017, 6963.46497, 8488.70517, 10351.1853, 12600.5857, 15295.8788, 18542.0318, 22417.9945, 26737.1279, 31884.8307, 38031.9875, 45349.4829, 54008.2015, 69012.3176, 84491.1036, 99407.1216, 112722.934, 123401.102, 127010.446, 128169.765, 127296.085, 125541.975, 123033.306, 119589.468, 115351.656, 110461.061, 105058.879, 99286.3009, 89912.7975, 83501.2548, 78828.7593, 74672.3972, 69809.2548, 63016.4184, 53070.9743, 40030.5679, 27065.8471},
        {6.86966265, 6.00164014, 5.1657297, 4.36108679, 3.58686686, 2.84222534, 2.12631769, 1.43829934, 0.777325761, 0.142552381, -0.466865348, -1.05177198, -1.61301207, -2.15143017, -2.66787083, -3.16317862, -3.63819807, -4.09377375, -4.53075021, -4.949972, -5.35228368, -5.73852979, -6.1095549, -6.46620356, -6.80932032, -7.13974973, -7.45833636, -7.76592474, -8.06335944, -8.35148501, -8.631146, -8.90318697, -9.16845247, -9.42778705, -9.7153696, -10.0146764, -10.3087829, -10.5974521, -10.8804466, -11.1575294, -11.4284631, -11.6930105, -11.9509345, -12.2019979, -12.4459634, -12.6825938, -12.9116519, -13.1329005, -13.3461024, -13.5510205, -13.7474173, -13.9498402, -14.142947, -14.3261801, -14.4989817, -14.6607942, -14.8110598, -14.9492209, -15.0747197, -15.1869986, -15.2854998, -15.3696656, -15.4389383, -15.4927603, -15.5305737, -15.551821, -15.5559444, -15.5474689, -15.5307974, -15.4948896, -15.4388583, -15.3618162, -15.2628763, -15.1411513, -14.9957539, -14.8257971, -14.6303936, -14.4086562, -14.1596977, -13.882631, -13.5765688, -13.240624, -12.8739093, -12.4755376, -11.4844216, -10.1915572, -8.89467348, -7.61106278, -6.35801749, -5.15282998, -4.01279262, -2.95519779, -1.99733787, -1.15650523, -0.449992243, 0.104908718, 0.490905276, 0.690705059, 0.687015689, 0.462544795, 0, -3.06795131, -6.08474645, -9.07254794, -12.0535183, -15.0498202, -18.083616, -21.1770683, -24.3523396, -27.6315926, -31.0369897, -34.5906934, -38.3148664, -42.2316711, -46.3632701, -50.7318259, -55.359501, -59.730746, -63.7909074, -68.7784011, -74.6001658, -81.1631398, -88.3742618, -96.1404704, -104.368704, -112.965901, -121.839, -130.89494, -140.040658, -149.183095, -158.087995, -166.855261, -175.434187, -183.774072, -193.189887, -202.948092, -211.950962, -219.738089, -225.849068, -229.82349, -231.200949, -231.435533, -227.569667, -218.768368, -204.196653, -183.019539, -159.769186, -126.901684, -82.2402644, -23.6081601, 51.1713973, 139.357634, 254.055189, 394.97169, 566.923654, 785.158251, 1060.2395, 1402.73142, 1819.19997, 2325.11589, 2940.31139, 3666.17732, 4546.64213, 5588.65027, 6824.72725, 8330.74426, 10162.6761, 12362.7516, 14983.3818, 18127.3796, 21869.0479, 25987.4479, 30897.5456, 36775.8427, 43798.841, 52143.0422, 66748.3103, 81836.9443, 96394.6052, 109406.954, 119859.653, 123434.773, 124623.958, 123828.691, 122135.518, 119674.754, 116291.271, 112129.655, 107334.494, 102050.373, 96421.8783, 86229.7839, 79832.0204, 75705.5502, 72327.3356, 68174.3389, 61723.5227, 51451.8493, 38289.7192, 25555.9188},
        {7.25431735, 6.32556618, 5.4314921, 4.5711846, 3.74373314, 2.94822722, 2.18375632, 1.44940991, 0.744277469, 0.0674484867, -0.581987561, -1.20494119, -1.80232293, -2.37504329, -2.92401279, -3.45014196, -3.95434131, -4.43752136, -4.90059264, -5.34446565, -5.77005093, -6.17825899, -6.57000035, -6.94618553, -7.30772505, -7.65552943, -7.99050918, -8.31357484, -8.62563692, -8.92760593, -9.22039241, -9.50490686, -9.78205981, -10.0527618, -10.3541787, -10.6683307, -10.9768024, -11.2793277, -11.5756406, -11.865475, -12.1485649, -12.4246443, -12.6934471, -12.9547073, -13.2081588, -13.4535356, -13.6905718, -13.9190011, -14.1385577, -14.3489755, -14.5499883, -14.7546878, -14.9491607, -15.1328515, -15.3052046, -15.4656644, -15.6136753, -15.7486817, -15.8701279, -15.9774584, -16.0701176, -16.1475499, -16.2091997, -16.2545113, -16.2829293, -16.2938978, -16.2868615, -16.2658218, -16.234671, -16.1834545, -16.1113213, -16.0174205, -15.900901, -15.7609117, -15.5966017, -15.40712, -15.1916154, -14.949237, -14.6791337, -14.3804546, -14.0523485, -13.6939646, -13.3044516, -12.8829587, -11.872302, -10.5614985, -9.24625211, -7.94376581, -6.67124267, -5.44588574, -4.28489805, -3.20548266, -2.22484261, -1.36018095, -0.628700723, -0.0476049724, 0.365903254, 0.59462091, 0.62134495, 0.428872329, 0, -2.98404332, -5.9265385, -8.84848019, -11.770863, -14.7146817, -17.7009308, -20.750605, -23.8846989, -27.1242072, -30.4901246, -34.0034456, -37.685165, -41.5562773, -45.6377773, -49.9506595, -54.5159186, -58.8238514, -62.8196385, -67.7246626, -73.4461976, -79.891517, -86.9678945, -94.5826037, -102.642918, -111.056112, -119.729459, -128.570232, -137.485704, -146.383151, -155.056398, -163.567763, -171.858555, -179.87008, -188.834277, -198.047139, -206.460491, -213.62845, -219.105136, -222.444668, -223.201164, -222.938998, -218.590217, -209.315629, -194.276043, -172.632265, -148.195941, -114.314334, -68.9855402, -10.20766, 64.0212084, 150.466094, 262.766392, 401.604576, 571.939275, 787.459603, 1057.93487, 1393.13437, 1798.52359, 2289.15379, 2884.77136, 3584.44288, 4435.62802, 5454.08202, 6673.34918, 8157.11628, 9955.23418, 12101.8826, 14642.8495, 17681.3839, 21288.2315, 25221.554, 29912.7713, 35540.8311, 42284.6816, 50323.2706, 64500.077, 79163.3852, 93326.0105, 106000.768, 116200.473, 119735.635, 120956.246, 120247.538, 118634.593, 116249.369, 112957.223, 108903.442, 104233.317, 99092.1355, 93625.188, 83034.4649, 76730.5301, 73008.3457, 70162.8736, 66489.0759, 60281.9146, 49836.3517, 36639.1563, 24214.9619},
        {7.68599905, 6.69906173, 5.74893123, 4.8346495, 3.95525849, 3.10980014, 2.29731639, 1.51684918, 0.767440467, 0.0481321837, -0.642033725, -1.30401532, -1.93877064, -2.54725777, -3.13043474, -3.68925963, -4.22469048, -4.73768536, -5.22920231, -5.7001994, -6.15163468, -6.58446621, -6.99965205, -7.39815026, -7.78091888, -8.14891598, -8.50309962, -8.84442784, -9.17385871, -9.49235029, -9.80086063, -10.1003478, -10.3917698, -10.6760848, -10.9924963, -11.3220353, -11.6452727, -11.9619257, -12.2717109, -12.5743455, -12.8695462, -13.15703, -13.4365139, -13.7077147, -13.9703493, -14.2241346, -14.4687877, -14.7040253, -14.9295644, -15.1451219, -15.3504148, -15.5582154, -15.7549021, -15.9399089, -16.1126696, -16.2726181, -16.4191884, -16.5518142, -16.6699296, -16.7729683, -16.8603643, -16.9315516, -16.9859639, -17.0230352, -17.0421993, -17.0428903, -17.0245419, -16.990021, -16.9421121, -16.8731688, -16.7824025, -16.6690245, -16.5322461, -16.3712788, -16.1853339, -15.9736227, -15.7353566, -15.4697468, -15.1760049, -14.8533421, -14.5009697, -14.1180992, -13.7039418, -13.257709, -12.2287979, -10.9028917, -9.57228165, -8.25402153, -6.96516526, -5.72276672, -4.54387977, -3.4455583, -2.44485616, -1.55882723, -0.804525393, -0.199004511, 0.24068154, 0.49747889, 0.554333664, 0.394191992, 0, -2.82161056, -5.62441858, -8.42734609, -11.2493151, -14.1092477, -17.0260659, -20.0186917, -23.1060472, -26.3070544, -29.6406354, -33.1257121, -36.7812066, -40.6260409, -44.6791372, -48.9594173, -53.4858034, -57.7661245, -61.7462893, -66.6012636, -72.2404494, -78.5732489, -85.5090643, -92.9572975, -100.827351, -109.028626, -117.470525, -126.062451, -134.713805, -143.33399, -151.749313, -159.982739, -167.968598, -175.64122, -184.137487, -192.790455, -200.612497, -207.177179, -212.058064, -214.828718, -215.062706, -214.49143, -209.873885, -200.361449, -185.105502, -163.257424, -138.116529, -103.649277, -57.9700639, 0.806715335, 74.566665, 159.943691, 270.493839, 407.421245, 575.647912, 787.686492, 1052.64159, 1379.61781, 1773.32014, 2248.25271, 2823.71928, 3496.63399, 4317.3083, 5308.87727, 6505.58834, 7960.51826, 9718.10081, 11804.607, 14259.7617, 17189.8868, 20663.633, 24439.1083, 28943.2823, 34352.3015, 40842.3124, 48589.4615, 62286.9848, 76464.667, 90169.382, 102448.004, 112347.407, 115832.54, 117086.819, 116477.944, 114976.153, 112710.642, 109559.543, 105664.604, 101167.57, 96210.188, 90934.2034, 80525.6891, 74423.7604, 70906.3047, 68251.2091, 64736.361, 58639.6477, 48238.9565, 35081.8228, 23039.5929},
        {8.1752324, 7.12661618, 6.11717806, 5.14590617, 4.21178863, 3.31381357, 2.45096912, 1.6222434, 0.826624545, 0.0631006752, -0.66934008, -1.37170959, -2.04501974, -2.69028239, -3.30850943, -3.90071271, -4.46790412, -5.01109553, -5.53129881, -6.02952584, -6.50678848, -6.96409862, -7.40246812, -7.82290886, -8.22643271, -8.61405155, -8.98677725, -9.34562168, -9.69159671, -10.0257142, -10.3489861, -10.6624242, -10.9670404, -11.2638465, -11.5936321, -11.9366835, -12.2728091, -12.6017299, -12.9231668, -13.236841, -13.5424734, -13.839785, -14.1284968, -14.4083299, -14.6790053, -14.9402439, -15.1917669, -15.4332952, -15.6645498, -15.8852517, -16.0951221, -16.3097828, -16.5127092, -16.7032776, -16.8808643, -17.0448456, -17.1945978, -17.3294973, -17.4489204, -17.5522433, -17.6388423, -17.7080939, -17.7593742, -17.7920597, -17.8055265, -17.7991511, -17.7723097, -17.7251428, -17.6577723, -17.5679988, -17.4551489, -17.3185496, -17.1575275, -16.9714095, -16.7595222, -16.5211926, -16.2557474, -15.9625133, -15.6408171, -15.2899857, -14.9093457, -14.4982241, -14.0559474, -13.5818427, -12.5373296, -11.2020025, -9.8619212, -8.53386845, -7.23462717, -5.98098026, -4.78971062, -3.67760113, -2.66143469, -1.75799418, -0.984062514, -0.356422572, 0.108142752, 0.392850564, 0.480917971, 0.355562081, 0, -2.46518721, -4.96659971, -7.51892654, -10.1368567, -12.8350793, -15.6282833, -18.5311578, -21.5583918, -24.7246743, -28.0446944, -31.5331411, -35.2047034, -39.0740704, -43.1559311, -47.4649745, -52.0158897, -56.3449481, -60.3984932, -65.2733494, -70.8816875, -77.1356786, -83.9474936, -91.2293035, -98.8932792, -106.851592, -115.016412, -123.299911, -131.61426, -139.871629, -147.913546, -155.748728, -163.31054, -170.532346, -178.437984, -186.415581, -193.562471, -199.484876, -203.789017, -206.081116, -205.967395, -205.514394, -201.119224, -191.906739, -177.001796, -155.52925, -130.997909, -97.0526874, -51.72245, 6.96393636, 80.9776051, 168.105656, 279.617076, 415.486152, 579.941576, 785.898103, 1042.37062, 1358.37403, 1739.49315, 2198.9525, 2753.71844, 3398.25478, 4184.51748, 5139.34989, 6297.03499, 7704.46172, 9402.7327, 11410.7661, 13763.913, 16578.8674, 19926.3112, 23627.6741, 28041.2367, 33323.9603, 39632.8065, 47124.7366, 60165.4044, 73650.906, 86686.0298, 98375.5641, 107824.297, 111246.645, 112564.684, 112121.932, 110841.514, 108837.522, 105982.549, 102402.762, 98224.3305, 93573.4229, 88576.2075, 79915.0663, 74304.3437, 70447.7118, 67048.8429, 62811.4092, 56439.0828, 46635.536, 33510.7789, 21959.311},
        {8.71411517, 7.61590471, 6.55774275, 5.53862344, 4.5575409, 3.61348927, 2.70546266, 1.83245522, 0.99346106, 0.187474324, -0.586510865, -1.32950038, -2.04250008, -2.72651585, -3.38255355, -4.01161906, -4.61471825, -5.19285698, -5.74704113, -6.27827656, -6.78756916, -7.27592479, -7.74434931, -8.19384861, -8.62542855, -9.040095, -9.43885384, -9.82271093, -10.1926721, -10.5497434, -10.8949304, -11.2292392, -11.5536757, -11.8692456, -12.21566, -12.5737581, -12.9239139, -13.2658771, -13.5993976, -13.9242253, -14.2401098, -14.5468011, -14.844049, -15.1316031, -15.4092135, -15.6766298, -15.933602, -16.1798797, -16.4152129, -16.6393513, -16.8520447, -17.0731851, -17.2819434, -17.477633, -17.659567, -17.8270585, -17.9794208, -18.115967, -18.2360103, -18.3388638, -18.4238407, -18.4902543, -18.5374176, -18.5646439, -18.5712463, -18.556538, -18.5198322, -18.4580409, -18.3681338, -18.2543763, -18.1162371, -17.9531851, -17.7646891, -17.550218, -17.3092406, -17.0412257, -16.7456421, -16.4219588, -16.0696444, -15.6881679, -15.276998, -14.8356037, -14.3634537, -13.8600168, -12.8006975, -11.4580198, -10.1108065, -8.77553346, -7.46867636, -6.20671093, -5.0061129, -3.88335799, -2.85492195, -1.93728049, -1.14690935, -0.500284258, -0.0138809385, 0.295824876, 0.412357457, 0.319241075, 0, -2.08026903, -4.25722647, -6.54091427, -8.94137436, -11.4686487, -14.1327792, -16.9438078, -19.9117765, -23.0467272, -26.3587018, -29.8577423, -33.5538907, -37.4571888, -41.5776786, -45.9254021, -50.5104011, -54.9086246, -59.0681012, -63.98075, -69.5635931, -75.7336524, -82.40795, -89.5035079, -96.9373479, -104.626492, -112.487963, -120.438781, -128.39597, -136.276551, -143.944534, -151.389391, -158.544048, -165.341429, -172.695794, -180.0494, -186.582821, -191.934581, -195.743208, -197.647228, -197.285167, -197.023323, -192.94226, -184.146807, -169.741796, -148.832059, -125.310838, -92.2974507, -47.6996263, 10.5749064, 84.6184183, 173.519568, 286.045321, 420.799909, 581.267961, 780.898256, 1028.66587, 1333.54589, 1702.00543, 2146.09752, 2680.61115, 3298.54179, 4051.70528, 4967.76219, 6080.30729, 7432.99058, 9064.84654, 10989.3018, 13238.7517, 15939.4872, 19164.6866, 22808.1988, 27150.1206, 32325.7061, 38470.2092, 45718.8836, 58089.6017, 70864.3445, 83207.7113, 94284.3015, 103258.714, 106598.292, 107963.063, 107673.592, 106617.375, 104890.569, 102353.11, 99112.9676, 95278.1089, 90956.5023, 86256.1161, 79619.667, 74548.4863, 70255.0561, 65951.8584, 60851.375, 54166.088, 45108.4794, 32122.2899, 21088.8134},
        {9.30106999, 8.17374939, 7.08490683, 6.03363832, 5.01903986, 4.04020745, 3.09623709, 2.18622479, 1.30926656, 0.464458378, -0.349103734, -1.13232378, -1.88610576, -2.61135366, -3.30897149, -3.97986325, -4.62493292, -5.24508452, -5.84122203, -6.41424945, -6.96507079, -7.49459003, -8.00371118, -8.49333824, -8.96437519, -9.41772605, -9.8542948, -10.2749854, -10.680702, -11.0723484, -11.4508287, -11.8170469, -12.171907, -12.5163129, -12.8848994, -13.2609957, -13.6275254, -13.984298, -14.3311228, -14.6678092, -14.9941667, -15.3100045, -15.615132, -15.9093586, -16.1924937, -16.4643467, -16.7247269, -16.9734437, -17.2103064, -17.4351245, -17.6477073, -17.8732232, -18.0855728, -18.2840157, -18.4678117, -18.6362203, -18.7885012, -18.923914, -19.0417184, -19.141174, -19.2215405, -19.2820775, -19.3220445, -19.3407014, -19.3373077, -19.311123, -19.261407, -19.1817009, -19.065683, -18.9244071, -18.7575033, -18.5646022, -18.345334, -18.099329, -17.8262177, -17.5256304, -17.1971974, -16.840549, -16.4553156, -16.0411276, -15.5976153, -15.124409, -14.6211392, -14.0874361, -13.0133524, -11.6640258, -10.3107182, -8.96959673, -7.65682861, -6.38858091, -5.18102075, -4.05031525, -3.01263152, -2.0841367, -1.28099789, -0.619382214, -0.115456792, 0.21461126, 0.354654821, 0.288506774, 0, -1.73614712, -3.62385198, -5.66880492, -7.87669627, -10.2532164, -12.8040556, -15.5349042, -18.4514525, -21.559391, -24.86441, -28.3721997, -32.0884505, -36.0188528, -40.1690968, -44.5448731, -49.1518717, -53.6175047, -57.8934224, -62.8409454, -68.3847042, -74.4493295, -80.9594519, -87.8397021, -95.0147107, -102.409108, -109.947526, -117.554594, -125.154943, -132.673203, -140.012594, -147.127008, -153.947502, -160.405128, -167.320078, -174.178837, -180.23079, -185.14025, -188.571531, -190.188947, -189.656811, -189.507194, -185.663857, -177.230032, -163.308949, -143.00384, -120.591673, -88.7085076, -45.164139, 12.2316362, 85.6690217, 175.258526, 287.811898, 421.035786, 577.854633, 771.845876, 1011.80398, 1306.52339, 1662.72987, 2091.75675, 2607.19413, 3202.2625, 3926.22181, 4804.25959, 5868.37247, 7162.09484, 8723.66505, 10564.0269, 12714.3423, 15305.8535, 18413.1919, 21990.6974, 26251.7987, 31312.7662, 37289.8705, 44299.382, 56051.9746, 68170.9144, 79873.8709, 90378.5137, 98902.5123, 102132.794, 103504.932, 103321.2, 102448.415, 100965.386, 98715.693, 95789.9055, 92278.5928, 88272.3242, 83861.6688, 79027.0347, 74449.2657, 69793.8982, 64726.4689, 58912.5141, 52017.5703, 43707.174, 31013.4275, 20482.9922},
        {9.93030726, 8.81476654, 7.73137648, 6.6795206, 5.65858242, 4.66794546, 3.70699326, 2.77510933, 1.87167718, 0.996080354, 0.147702363, -0.674073271, -1.46986302, -2.24028337, -2.9859508, -3.70748177, -4.40549278, -5.08060029, -5.7334208, -6.36457076, -6.97466667, -7.56432499, -8.13416222, -8.68479481, -9.21683927, -9.73091205, -10.2276296, -10.7076085, -11.1714652, -11.619816, -12.0532776, -12.4724664, -12.8779989, -13.2704915, -13.6720375, -14.0720633, -14.459737, -14.8349782, -15.1977065, -15.5478412, -15.8853021, -16.2100085, -16.5218799, -16.820836, -17.1067962, -17.37968, -17.639407, -17.8858966, -18.1190685, -18.338842, -18.5451367, -18.7692243, -18.9789925, -19.1736811, -19.3525301, -19.5147792, -19.6596684, -19.7864375, -19.8943264, -19.9825749, -20.050423, -20.0971104, -20.121877, -20.1239628, -20.1026075, -20.057051, -19.9865333, -19.8813296, -19.7319306, -19.5560647, -19.3535529, -19.1242163, -18.8678758, -18.5843523, -18.273467, -17.9350408, -17.5688946, -17.1748495, -16.7527266, -16.3023466, -15.8235308, -15.3161, -14.7798753, -14.2146776, -13.1242765, -11.7669525, -10.4069483, -9.06018257, -7.74257426, -6.47004214, -5.25850504, -4.12388178, -3.08209117, -2.14905203, -1.34068318, -0.672903437, -0.161631622, 0.177213447, 0.327712953, 0.273948077, 0, -1.57317772, -3.32657792, -5.26304486, -7.38542281, -9.69655601, -12.1992887, -14.8964652, -17.7909298, -20.8855266, -24.1831001, -27.6864943, -31.3985536, -35.3221222, -39.4600445, -43.8151646, -48.3903268, -52.8947249, -57.289253, -62.2409649, -67.6897798, -73.5756168, -79.838395, -86.4180337, -93.2544519, -100.287569, -107.457304, -114.703575, -121.966303, -129.185406, -136.363293, -143.354481, -150.080142, -156.461448, -163.25878, -169.973706, -175.884007, -180.659092, -183.968371, -185.481256, -184.867156, -184.425437, -180.395132, -171.931095, -158.18818, -138.32124, -116.784745, -86.1090284, -44.1240407, 11.3402684, 82.4539492, 169.22327, 278.626685, 409.226292, 564.073499, 755.385387, 991.263834, 1279.81072, 1626.18712, 2042.10409, 2542.48083, 3124.5086, 3829.75194, 4673.73244, 5684.97987, 6914.06574, 8401.44403, 10164.2687, 12236.953, 14739.3435, 17742.9611, 21208.2904, 25334.9704, 30232.9573, 36012.2074, 42782.6766, 54112.2667, 65788.502, 77056.8523, 87162.7874, 95351.7773, 98412.9665, 99685.7, 99468.1584, 98643.3817, 97278.0436, 95195.1263, 92473.488, 89191.9871, 85429.4816, 81264.83, 77217.2586, 72942.5752, 68282.1362, 63077.2981, 57169.4175, 50399.8507, 42609.9543, 30493.9649, 20349.6522},
        {10.6122373, 9.53007264, 8.47119606, 7.43537758, 6.42238728, 5.43199519, 4.4639714, 3.51808595, 2.5941089, 1.69181032, 0.81096026, -0.0486712166, -0.887314052, -1.70519819, -2.50255356, -3.27961012, -4.03659779, -4.77374653, -5.49128627, -6.18944695, -6.86845852, -7.52855091, -8.16995407, -8.79289794, -9.39761245, -9.98432755, -10.5532732, -11.1046793, -11.6387758, -12.1557926, -12.6559598, -13.1395072, -13.6066647, -14.0576624, -14.4979369, -14.9250053, -15.336322, -15.731948, -16.1119443, -16.4763717, -16.8252914, -17.1587643, -17.4768513, -17.7796134, -18.0671115, -18.3394067, -18.59656, -18.8386322, -19.0656844, -19.2777775, -19.4749725, -19.6953991, -19.9002242, -20.0886836, -20.260013, -20.4134482, -20.5482248, -20.6635786, -20.7587452, -20.8329605, -20.8854601, -20.9154797, -20.9222551, -20.9050219, -20.8630159, -20.7954728, -20.7016283, -20.5687383, -20.3843464, -20.1723956, -19.9328984, -19.6658671, -19.371314, -19.0492514, -18.6996916, -18.322647, -17.9181299, -17.4861526, -17.0267274, -16.5398667, -16.0255828, -15.4838879, -14.9147945, -14.3183148, -13.2115871, -11.8462174, -10.4796154, -9.12746352, -7.80544439, -6.52924048, -5.31453429, -4.17700835, -3.13234515, -2.19622722, -1.38433706, -0.712357184, -0.195970106, 0.149141662, 0.307295611, 0.262809227, 0, -1.46467441, -3.1291251, -4.99397083, -7.05983037, -9.32732248, -11.7970659, -14.4696795, -17.3457819, -20.4259919, -23.7109284, -27.20121, -30.8974555, -34.8002837, -38.9103134, -43.2281633, -47.7544522, -52.2760979, -56.7631906, -61.6997201, -67.040512, -72.740392, -78.754186, -85.0367196, -91.5428185, -98.2273085, -105.045015, -111.950765, -118.899382, -125.845693, -132.885862, -139.781553, -146.445189, -152.789196, -159.541802, -166.207529, -172.073351, -176.806953, -180.076016, -181.548224, -180.891259, -180.060699, -175.740017, -167.149265, -153.508496, -134.03776, -113.26701, -83.7789218, -43.4660734, 9.77895842, 78.0635967, 160.9738, 266.220246, 393.688033, 546.731861, 735.790194, 968.12103, 1250.98237, 1587.24356, 1989.54868, 2475.32938, 3047.25097, 3737.45971, 4550.34578, 5510.34565, 6675.86319, 8089.71069, 9776.15263, 11773.1926, 14186.8428, 17084.6457, 20419.4191, 24391.0806, 29106.8854, 34674.0885, 41199.9452, 52154.3312, 63441.2258, 74327.2291, 84078.941, 91962.9615, 94841.2354, 95988.1213, 95700.7858, 94882.6166, 93590.8177, 91630.0005, 89069.9984, 85980.6449, 82431.7734, 78493.2171, 74756.5705, 70684.5182, 66171.401, 61111.5599, 55399.3355, 48929.0688, 41595.1004, 30187.0636, 20409.8547},
        {11.3515349, 10.319104, 9.2991784, 8.29198963, 7.29776913, 6.31674832, 5.34915865, 4.39523155, 3.45519847, 2.52929083, 1.61774009, 0.720777678, -0.161364968, -1.02845641, -1.8802652, -2.71655992, -3.53710911, -4.34168134, -5.13004517, -5.90196916, -6.65722188, -7.39557188, -8.11678773, -8.82063799, -9.50689121, -10.175316, -10.8256808, -11.4577543, -12.071305, -12.6661015, -13.2419124, -13.7985061, -14.3356513, -14.8531165, -15.3365035, -15.7929622, -16.2298129, -16.6472837, -17.0456032, -17.4249995, -17.7857009, -18.1279358, -18.4519324, -18.7579191, -19.0461241, -19.3167758, -19.5701024, -19.8063323, -20.0256937, -20.228415, -20.4147245, -20.6302541, -20.8288442, -21.0097388, -21.172182, -21.3154179, -21.4386904, -21.5412435, -21.6223215, -21.6811681, -21.7170276, -21.729144, -21.7167612, -21.6791234, -21.6154745, -21.5250587, -21.4071198, -21.2463195, -21.0276693, -20.7804887, -20.504967, -20.2012937, -19.869658, -19.5102495, -19.1232574, -18.7088711, -18.2672799, -17.7986733, -17.3032406, -16.7811712, -16.2326544, -15.6578796, -15.0570362, -14.4303136, -13.3076829, -11.934333, -10.5611044, -9.20344601, -7.87680663, -6.59663512, -5.37838035, -4.23749115, -3.18941639, -2.24960491, -1.43350556, -0.756567201, -0.234238678, 0.118031154, 0.284793446, 0.250599345, 0, -1.37849383, -2.97095163, -4.77624403, -6.79324163, -9.02081505, -11.4578349, -14.1031718, -16.9556963, -20.0142792, -23.2777909, -26.7451021, -30.4150834, -34.2866054, -38.3585388, -42.6297541, -47.099122, -51.6115822, -56.1440151, -61.0434684, -66.2758276, -71.806978, -77.6028051, -83.6291943, -89.852031, -96.2372006, -102.750589, -109.35808, -116.025561, -122.718917, -129.595688, -136.359831, -142.919734, -149.183787, -155.8781, -162.502201, -168.337558, -173.047243, -176.294327, -177.741882, -177.052981, -175.820328, -171.190083, -162.447778, -148.878951, -129.769134, -109.616296, -81.1904312, -42.4739645, 8.55067729, 73.9010679, 152.65038, 253.3593, 377.326475, 528.234427, 714.632762, 942.885864, 1219.35812, 1544.18355, 1931.37289, 2401.73364, 2964.18649, 3640.97928, 4426.2233, 5340.09832, 6446.9557, 7790.8713, 9401.09912, 11317.8723, 13635.771, 16418.6334, 19608.9642, 23410.6023, 27929.3862, 33271.1545, 39541.7456, 50124.293, 61028.0166, 71539.4313, 80945.052, 88531.3936, 91230.8482, 92253.4684, 91894.6789, 91068.9411, 89826.6442, 87960.2452, 85533.0296, 82608.2832, 79249.2917, 75519.3407, 71685.0523, 67722.1797, 63476.5364, 58793.936, 53520.1919, 47501.1177, 40582.5268, 29958.971, 20556.9862},
        {12.174701, 11.1932066, 10.2131885, 9.23534946, 8.260392, 7.28901882, 6.32193257, 5.3598359, 4.40343147, 3.45342192, 2.51050992, 1.5753981, 0.648789132, -0.26861434, -1.17610966, -2.07299418, -2.95856524, -3.8321202, -4.69295639, -5.54037117, -6.37366189, -7.19212588, -7.9950605, -8.7817631, -9.55153103, -10.3036616, -11.0374522, -11.7522002, -12.4472029, -13.1217576, -13.7751618, -14.4067127, -15.0157077, -15.6014442, -16.1291749, -16.6159356, -17.0791711, -17.5192837, -17.9366754, -18.3317484, -18.7049049, -19.056547, -19.3870769, -19.6968967, -19.9864087, -20.2560149, -20.5061175, -20.7371187, -20.9494206, -21.1434254, -21.3195353, -21.5298176, -21.7218893, -21.8950325, -22.0485294, -22.1816621, -22.2937128, -22.3839636, -22.4516968, -22.4961944, -22.5167386, -22.5126115, -22.4830954, -22.4274724, -22.3450246, -22.2350341, -22.0967833, -21.9144131, -21.6724283, -21.4013382, -21.1014065, -20.7728969, -20.4160731, -20.0311988, -19.6185376, -19.1783532, -18.7109094, -18.2164698, -17.6952981, -17.147658, -16.5738131, -15.9740272, -15.348564, -14.6976871, -13.5607496, -12.1781652, -10.7958873, -9.42913689, -8.09313519, -6.8031033, -5.57426234, -4.42183342, -3.36103767, -2.4070962, -1.57523014, -0.880660606, -0.338608714, 0.0357044119, 0.227057653, 0.220229889, 0, -1.18254476, -2.60490806, -4.26397619, -6.15663544, -8.27977211, -10.6302725, -13.2050229, -16.0009095, -19.0148188, -22.2436369, -25.6842502, -29.333545, -33.1884075, -37.2457241, -41.502381, -45.9552646, -50.4198513, -54.8671118, -59.7042639, -64.8893205, -70.3802944, -76.1351982, -82.1120448, -88.2688469, -94.5636173, -100.954369, -107.399114, -113.855866, -120.282637, -126.740143, -133.046035, -139.127518, -144.911793, -151.202892, -157.489602, -163.02689, -167.47891, -170.509815, -171.783758, -170.964892, -169.543823, -164.802377, -156.047357, -142.585569, -123.723817, -103.704171, -75.664353, -37.6773513, 12.1838464, 75.846252, 152.378924, 250.059882, 370.79017, 518.187517, 699.374551, 919.900273, 1185.31368, 1492.23416, 1857.16978, 2306.37005, 2856.08916, 3519.55229, 4287.96758, 5177.61313, 6247.24676, 7536.12737, 9063.85924, 10864.6056, 13044.5969, 15669.2982, 18710.8768, 22342.4493, 26661.7761, 31766.6179, 37754.7352, 47824.7613, 58198.3659, 68198.0913, 77146.48, 84366.0742, 86944.0668, 87928.5829, 87600.1718, 86826.9675, 85656.9337, 83893.2074, 81595.9936, 78825.4975, 75641.9241, 72105.4785, 67737.8054, 63765.7263, 59890.6705, 55814.0678, 51237.3477, 45861.9397, 39389.2734, 29458.3696, 20478.1177},
        {13.0509021, 12.1324409, 11.2030737, 10.2640264, 9.31652496, 8.36179539, 7.40106361, 6.4355556, 5.46649729, 4.49511465, 3.52263364, 2.5502802, 1.5792803, 0.610859883, -0.353755084, -1.31333865, -2.26666485, -3.21250775, -4.14964137, -5.07683977, -5.99287698, -6.89652706, -7.78656405, -8.66176199, -9.52089492, -10.3627369, -11.1860619, -11.9896441, -12.7722575, -13.5326761, -14.2696739, -14.982025, -15.6685035, -16.3278834, -16.9026798, -17.4209795, -17.9115456, -18.3749779, -18.811876, -19.2228398, -19.6084689, -19.9693633, -20.3061225, -20.6193463, -20.9096346, -21.1775871, -21.4238034, -21.6488835, -21.8534269, -22.0380336, -22.2033031, -22.409166, -22.5956051, -22.7619341, -22.9074665, -23.0315159, -23.133396, -23.2124202, -23.2679023, -23.2991557, -23.305494, -23.2862309, -23.2406798, -23.1681545, -23.0679685, -22.9394353, -22.7818686, -22.5787504, -22.3139427, -22.0194106, -21.695494, -21.3425328, -20.9608668, -20.5508359, -20.1127801, -19.6470391, -19.1539528, -18.6338613, -18.0871042, -17.5140215, -16.9149531, -16.2902389, -15.6402187, -14.9652324, -13.8148133, -12.4241865, -11.0339959, -9.65921083, -8.31480077, -7.01573512, -5.77698331, -4.61351476, -3.54029891, -2.57230517, -1.72450298, -1.01186175, -0.449350913, -0.0519398948, 0.165401881, 0.187704987, 0, -1.00728793, -2.27404243, -3.79571382, -5.56775243, -7.5856086, -9.84473264, -12.3405749, -15.0685857, -18.0242153, -21.2029142, -24.6001325, -28.2113208, -32.0319292, -36.0574081, -40.2832078, -44.7047787, -49.0898275, -53.4012697, -58.1496899, -63.2817365, -68.7440574, -74.4833007, -80.4461147, -86.5791475, -92.829047, -99.1424616, -105.466039, -111.746428, -117.930276, -123.957198, -129.783455, -135.357804, -140.629005, -146.483931, -152.402282, -157.613595, -161.785193, -164.584399, -165.678536, -164.734928, -163.189249, -158.402277, -149.695355, -136.389824, -117.807026, -97.8843617, -70.1730985, -32.8405634, 15.9459167, 78.0190148, 152.561697, 247.294875, 364.387438, 507.279613, 682.007273, 893.4409, 1146.45097, 1434.65358, 1776.7314, 2203.64469, 2737.67996, 3384.97413, 4136.68454, 5005.91554, 6042.95939, 7281.03851, 8727.89713, 10410.1522, 12447.8406, 14908.1556, 17792.894, 21244.7065, 25353.4236, 30208.8758, 35900.8936, 45435.3455, 55257.0145, 64726.7216, 73205.2879, 80053.5344, 82527.8421, 83496.432, 83222.0278, 82512.1318, 81413.8265, 79747.203, 77570.6934, 74942.7295, 71921.7433, 68566.1666, 63451.0819, 59410.3724, 55952.0428, 52584.0981, 48814.5428, 44151.3818, 38102.6198, 28905.4614, 20402.2886},
        {13.9701823, 13.1282594, 12.2616006, 11.3720102, 10.4612929, 9.53125311, 8.5836955, 7.62042456, 6.64324484, 5.65396089, 4.65437726, 3.64629848, 2.6315291, 1.61187368, 0.589136746, -0.434877147, -1.45836346, -2.47951763, -3.49653514, -4.50761142, -5.51094194, -6.50472214, -7.48714749, -8.45641344, -9.41071544, -10.3482489, -11.2672094, -12.1657923, -13.0421931, -13.8946071, -14.72123, -15.5202571, -16.2898839, -17.0283058, -17.6526897, -18.2034495, -18.7219735, -19.2090865, -19.6656135, -20.0923796, -20.4902098, -20.859929, -21.2023621, -21.5183344, -21.8086706, -22.0741958, -22.315735, -22.5341132, -22.7301554, -22.9046865, -23.0585316, -23.2618743, -23.4446774, -23.6062623, -23.7459505, -23.8630633, -23.9569223, -24.0268488, -24.0721642, -24.0921901, -24.0862478, -24.0536588, -23.9937445, -23.9058263, -23.7892256, -23.643264, -23.4672628, -23.2430994, -22.9530701, -22.6324733, -22.2817612, -21.9013861, -21.4918003, -21.0534559, -20.5868054, -20.0923009, -19.5703947, -19.021539, -18.4461861, -17.8447883, -17.2177978, -16.5656669, -15.8888478, -15.1877928, -14.0246535, -12.6282515, -11.2326766, -9.85260462, -8.50271158, -7.19767344, -5.95216614, -4.78086565, -3.69844791, -2.71958889, -1.85896453, -1.1312508, -0.551123656, -0.133259047, 0.10766707, 0.156978737, 0, -0.887581059, -2.04290722, -3.46078026, -5.13600194, -7.06337403, -9.23769829, -11.6537765, -14.3064104, -17.1904018, -20.3005525, -23.6316641, -27.1785385, -30.9359775, -34.8987828, -39.0617562, -43.4196994, -47.6935495, -51.8389846, -56.4702626, -61.5234999, -66.9348131, -72.6403185, -78.5761326, -84.678372, -90.8831529, -97.1265919, -103.344805, -109.47391, -115.450022, -121.106728, -126.520269, -131.65752, -136.485356, -141.948862, -147.525809, -152.433112, -156.344183, -158.932434, -159.871277, -158.834126, -157.214976, -152.443, -143.854974, -130.787674, -112.577877, -92.8785077, -65.6311567, -29.0935626, 18.4765354, 78.8213982, 151.326417, 242.971341, 355.833082, 493.087748, 660.15564, 861.449466, 1101.38193, 1371.77563, 1692.49476, 2097.1382, 2611.31875, 3237.18638, 3968.19902, 4815.44096, 5819.8633, 7008.70779, 8378.99333, 9950.12611, 11852.3294, 14153.1422, 16874.0371, 20135.7063, 24021.6346, 28615.307, 34000.2082, 42998.3778, 52270.1073, 61214.243, 69229.6312, 75715.1181, 78097.7442, 79062.6985, 78853.8869, 78213.7555, 77188.8712, 75618.0758, 73558.3225, 71066.5645, 68199.755, 65014.8472, 59069.4106, 54929.0601, 51891.1625, 49253.0851, 46312.1947, 42365.8585, 36711.4434, 28312.3408, 20361.1915},
        {14.8745343, 14.128123, 13.3413225, 12.5165854, 11.6563644, 10.7631122, 9.83928153, 8.88732496, 7.9096952, 6.90884494, 5.88722684, 4.84729359, 3.79149785, 2.7222923, 1.64212962, 0.553462483, -0.541256439, -1.63957447, -2.73903893, -3.83719715, -4.93159646, -6.01978417, -7.09930762, -8.16771413, -9.22255102, -10.2613656, -11.2817053, -12.2811172, -13.2571489, -14.2073476, -15.1292606, -16.0204353, -16.878419, -17.7007589, -18.3753575, -18.9568907, -19.5013525, -20.0098484, -20.483484, -20.9233647, -21.3305961, -21.7062837, -22.051533, -22.3674496, -22.6551389, -22.9157065, -23.1502579, -23.3598986, -23.5457341, -23.7088699, -23.8504116, -24.0579764, -24.2442826, -24.4085605, -24.5500401, -24.6679515, -24.7615249, -24.8299903, -24.8725779, -24.8885179, -24.8770402, -24.8373751, -24.7687527, -24.670403, -24.5415562, -24.3814423, -24.1892916, -23.9413341, -23.6143522, -23.2550117, -22.8640338, -22.4421396, -21.9900501, -21.5084865, -20.99817, -20.4598215, -19.8941623, -19.3019134, -18.683796, -18.0405312, -17.37284, -16.6814436, -15.9670631, -15.2304196, -14.054362, -12.657337, -11.2628135, -9.88507353, -8.53839888, -7.23707145, -5.9953731, -4.8275857, -3.74799111, -2.77087121, -1.91050786, -1.18118293, -0.597178293, -0.172775812, 0.0777426426, 0.140095203, 0, -0.954885101, -2.15428036, -3.59408989, -5.27021781, -7.17856825, -9.31504532, -11.6755531, -14.2559958, -17.0522775, -20.0603023, -23.2759743, -26.6951976, -30.3138765, -34.1279148, -38.1332169, -42.3256869, -46.4487902, -50.4623685, -54.9333177, -59.8034479, -65.0145692, -70.5084919, -76.2270261, -82.1119819, -88.1051695, -94.1483991, -100.183481, -106.152225, -111.996441, -117.602864, -122.988575, -128.111905, -132.931189, -138.311733, -143.758383, -148.534126, -152.325201, -154.817851, -155.698315, -154.652835, -152.93816, -148.192042, -139.793446, -127.12134, -109.554691, -90.7241707, -64.6941155, -29.7805669, 15.7004342, 73.4328467, 143.066723, 230.805814, 337.989782, 467.46961, 625.18189, 815.677837, 1043.50867, 1302.76177, 1610.83033, 1996.52263, 2480.02008, 3068.25659, 3759.52126, 4566.13219, 5524.17888, 6658.83939, 7968.63206, 9471.82317, 11284.4312, 13467.1804, 16012.0838, 19059.1479, 22692.8818, 26997.7946, 32058.3953, 40578.531, 49368.5382, 57858.0911, 65476.8644, 71654.5324, 73957.0645, 74920.3104, 74772.6756, 74211.4634, 73279.4736, 71824.1319, 69898.0068, 67553.6669, 64843.6807, 61820.6168, 55136.7727, 50878.1849, 48164.3606, 46114.8068, 43849.0304, 40486.5386, 35146.8384, 27750.5592, 20540.5813},
        {15.8130147, 15.1775115, 14.4843935, 13.7368206, 12.9379529, 12.0909503, 11.1989728, 10.2651803, 9.29273279, 8.28479023, 7.24451258, 6.17505981, 5.07959188, 3.96126875, 2.82325039, 1.66869676, 0.500767817, -0.677376469, -1.86257614, -3.05167123, -4.24150177, -5.42890781, -6.61072939, -7.78380653, -8.94497927, -10.0910877, -11.2189717, -12.3254715, -13.4074271, -14.4616784, -15.4850656, -16.4744286, -17.4266075, -18.3384424, -19.0657637, -19.6788822, -20.2498035, -20.7799436, -21.2707186, -21.7235445, -22.1398376, -22.5210137, -22.868489, -23.1836796, -23.4680014, -23.7228707, -23.9497034, -24.1499156, -24.3249235, -24.4761429, -24.6049901, -24.8199591, -25.0130667, -25.1834078, -25.3300771, -25.4521696, -25.5487799, -25.619003, -25.6619337, -25.6766668, -25.6622971, -25.6179195, -25.5426287, -25.4355197, -25.2956872, -25.1222261, -24.9142311, -24.6402489, -24.2695594, -23.8642611, -23.4254289, -22.9541382, -22.451464, -21.9184815, -21.3562658, -20.7658921, -20.1484355, -19.5049712, -18.8365744, -18.1443201, -17.4292835, -16.6925398, -15.9351641, -15.1582315, -13.970095, -12.5764511, -11.1877168, -9.81771191, -8.48025605, -7.18916902, -5.95827054, -4.80138038, -3.73231825, -2.76490392, -1.91295712, -1.19029759, -0.610745073, -0.188119314, 0.0637599451, 0.131072964, 0, -1.12541657, -2.45383058, -3.9833487, -5.71207764, -7.63812409, -9.75959474, -12.0745963, -14.5812354, -17.2776188, -20.1618532, -23.2320452, -26.4863016, -29.9227291, -33.5394342, -37.3345239, -41.3061046, -45.248608, -49.1310444, -53.4122141, -58.0465793, -62.9886018, -68.1927438, -73.6134673, -79.2052343, -84.9225068, -90.7197468, -96.5514163, -102.371977, -108.135892, -113.852907, -119.401929, -124.720833, -129.747498, -135.202826, -140.633181, -145.376119, -149.134607, -151.611615, -152.51011, -151.533062, -149.762477, -145.102577, -136.986519, -124.84746, -108.118555, -90.5268676, -66.1381715, -33.3121471, 9.59152538, 64.2131659, 130.496256, 213.851867, 314.53138, 435.026322, 582.401853, 761.669664, 977.841449, 1229.38564, 1529.34429, 1897.89756, 2345.33977, 2887.7967, 3529.58282, 4285.87351, 5190.24743, 6267.83582, 7524.71416, 8981.07331, 10726.9753, 12812.9592, 15178.4582, 18002.2102, 21373.2136, 25380.4669, 30112.9688, 38190.7004, 46538.9319, 54613.9165, 61871.9069, 67769.1561, 69990.8274, 70944.9898, 70846.9348, 70360.8596, 69524.3638, 68188.0028, 66398.7283, 64203.4921, 61649.2459, 58782.9414, 51529.9316, 47182.5717, 44724.8081, 43140.5873, 41413.8559, 38528.5603, 33468.6471, 27159.1907, 20782.9421},
        {16.8250749, 16.3131112, 15.7249293, 15.0644475, 14.3355842, 13.5422574, 12.6883854, 11.7778866, 10.814679, 9.80268099, 8.74581076, 7.64798655, 6.51312661, 5.34514915, 4.14797243, 2.92551468, 1.68169414, 0.420429035, -0.854362385, -2.13876189, -3.42885124, -4.72071219, -6.01042652, -7.29407599, -8.56774235, -9.82750737, -11.0694528, -12.2896604, -13.484212, -14.6491893, -15.7806741, -16.8747481, -17.9274931, -18.9349909, -19.7190851, -20.3666415, -20.9666199, -21.5207564, -22.0307871, -22.4984481, -22.9254757, -23.3136058, -23.6645747, -23.9801184, -24.2619731, -24.5118749, -24.7315598, -24.9227642, -25.087224, -25.2266754, -25.3428544, -25.5656038, -25.765776, -25.9423302, -26.0942252, -26.2204201, -26.319874, -26.3915459, -26.4343948, -26.4473798, -26.4294598, -26.3795939, -26.2967411, -26.1798605, -26.027911, -25.8398517, -25.6146416, -25.312656, -24.8951964, -24.4407981, -23.9509214, -23.4270264, -22.8705734, -22.2830228, -21.6658347, -21.0204694, -20.3483872, -19.6510483, -18.9299129, -18.1864414, -17.422094, -16.638331, -15.8366125, -15.0183989, -13.8198809, -12.4328365, -11.0534339, -9.69498704, -8.37080987, -7.09421636, -5.87852046, -4.73703612, -3.6830773, -2.72995796, -1.89099204, -1.17949351, -0.608776316, -0.192154414, 0.0570582431, 0.125547699, 0, -1.33470195, -2.82285198, -4.46535371, -6.26311076, -8.21702673, -10.3280052, -12.5969499, -15.0247643, -17.6123521, -20.3606168, -23.2704622, -26.3427918, -29.5785091, -32.9785179, -36.5437218, -40.2750243, -44.0151919, -47.7425372, -51.815655, -56.2015624, -60.8672764, -65.7798142, -70.9061928, -76.2134293, -81.6685407, -87.2385442, -92.8904568, -98.5912956, -104.308078, -110.175057, -115.930694, -121.491834, -126.775323, -132.34965, -137.80584, -142.561146, -146.336919, -148.854513, -149.835281, -149.000575, -147.263262, -142.790547, -135.070659, -123.591831, -107.842291, -91.6551547, -69.087548, -38.5414804, 1.58103845, 52.8779988, 115.582255, 194.325433, 288.161829, 399.215768, 535.879559, 703.731311, 908.349137, 1153.18653, 1446.42909, 1798.58017, 2208.96451, 2703.84471, 3293.55445, 3996.45409, 4844.50476, 5863.42464, 7068.285, 8482.04371, 10166.6972, 12162.1593, 14352.3316, 16957.1426, 20070.5104, 23786.3536, 28198.5904, 35854.2466, 43780.8562, 51459.0606, 58369.5015, 63992.8202, 66121.5805, 67049.8868, 66981.8423, 66555.588, 65803.4206, 64576.788, 62917.0294, 60865.484, 58463.4911, 55752.3899, 48185.5085, 43823.7044, 41570.8153, 40330.6788, 39007.1326, 36504.0144, 31725.1618, 26489.1943, 20961.5535},
        {18.0009895, 17.6273582, 17.1563057, 16.5926249, 15.9411086, 15.2065498, 14.3937412, 13.5074758, 12.5525465, 11.533746, 10.4558674, 9.32370341, 8.14204697, 6.91569094, 5.64942821, 4.34805164, 3.01635411, 1.65912849, 0.281167666, -1.1127355, -2.51778812, -3.92919733, -5.34217026, -6.75191402, -8.15363574, -9.54254255, -10.9138416, -12.2627399, -13.5844448, -14.8741632, -16.1271023, -17.3384693, -18.5034712, -19.6173152, -20.4619656, -21.1437362, -21.7715038, -22.3473957, -22.8735392, -23.3520615, -23.7850899, -24.1747517, -24.5231741, -24.8324844, -25.1048098, -25.3422777, -25.5470152, -25.7211497, -25.8668085, -25.9861187, -26.0812077, -26.3056059, -26.5067553, -26.683501, -26.8346881, -26.9591618, -27.0557673, -27.1233496, -27.1607539, -27.1668253, -27.140409, -27.0803501, -26.9854938, -26.8546852, -26.6867694, -26.4805916, -26.2349969, -25.9028024, -25.4379301, -24.9341547, -24.3933053, -23.8172108, -23.2077001, -22.5666021, -21.895746, -21.1969605, -20.4720747, -19.7229175, -18.9513178, -18.1591047, -17.3481071, -16.5201539, -15.6770742, -14.8206967, -13.6161699, -12.2407446, -10.8757185, -9.53384629, -8.22788304, -6.97058359, -5.77470283, -4.65299566, -3.61821698, -2.68312166, -1.86046462, -1.16300073, -0.603484901, -0.19467202, 0.0506830197, 0.119825324, 0, -1.51423664, -3.13383411, -4.86272291, -6.70483352, -8.66409641, -10.7444421, -12.949801, -15.2841036, -17.7512805, -20.355262, -23.0999787, -25.9893611, -29.0273396, -32.2178448, -35.564807, -39.0721568, -42.5756661, -46.0552427, -49.8951758, -54.0633619, -58.5276975, -63.2560791, -68.2164033, -73.3765664, -78.7044649, -84.1679955, -89.7350546, -95.3735386, -101.051344, -106.915695, -112.689406, -118.286577, -123.621307, -129.197994, -134.635586, -139.395619, -143.215098, -145.831033, -146.980432, -146.400302, -144.885604, -140.793408, -133.653732, -122.996592, -108.352007, -93.5647996, -72.7714799, -44.4233633, -6.97176525, 41.1319984, 100.383045, 174.671319, 262.01846, 364.205595, 490.689471, 647.480638, 840.589646, 1077.1894, 1361.86393, 1697.92922, 2077.31576, 2532.22047, 3076.36408, 3729.64637, 4522.43719, 5480.30044, 6623.25684, 7975.43317, 9581.28127, 11473.756, 13506.3874, 15919.159, 18807.134, 22265.3759, 26388.9479, 33636.7736, 41149.0912, 48429.9993, 54983.5962, 60313.9804, 62311.5527, 63169.9073, 63084.7275, 62672.272, 61961.746, 60807.1147, 59246.3452, 57317.4047, 55058.2602, 52506.8789, 45249.9117, 41092.4356, 38982.2821, 37867.2829, 36695.2695, 34414.0736, 29971.5268, 25670.2603, 20900.2671},
        {19.3992951, 19.1791322, 18.8371741, 18.3792356, 17.8111318, 17.1386776, 16.3676879, 15.5039777, 14.553362, 13.5216556, 12.4146736, 11.238231, 9.99814257, 8.70022338, 7.35028837, 5.95415248, 4.51763068, 3.0465379, 1.54668912, 0.023899273, -1.51601668, -3.06724378, -4.62396707, -6.18037162, -7.73064245, -9.26896461, -10.7895232, -12.2865031, -13.7540896, -15.1864676, -16.5778221, -17.9223383, -19.2142011, -20.4475956, -21.3569567, -22.0718081, -22.7247953, -23.3185328, -23.855635, -24.3387163, -24.7703912, -25.1532741, -25.4899795, -25.7831217, -26.0353153, -26.2491747, -26.4273142, -26.5723484, -26.6868917, -26.7735585, -26.8349633, -27.0507219, -27.2426104, -27.409406, -27.5498862, -27.6628283, -27.7470098, -27.801208, -27.8242004, -27.8147643, -27.7716771, -27.6937162, -27.579659, -27.4282829, -27.2383653, -27.0086835, -26.7380151, -26.3732205, -25.8624071, -25.3114265, -24.7224217, -24.0975356, -23.4389112, -22.7486914, -22.0290191, -21.2820373, -20.5098889, -19.7147169, -18.8986642, -18.0638737, -17.2124885, -16.3466513, -15.4685052, -14.5801931, -13.3756891, -12.0177024, -10.6726463, -9.35266442, -8.06990042, -6.83649787, -5.6646004, -4.56635162, -3.55389514, -2.63937458, -1.83493356, -1.15271568, -0.604864562, -0.203523821, 0.0391629287, 0.111052074, 0, -1.61142374, -3.28928239, -5.04043434, -6.87173799, -8.79005174, -10.802234, -12.9151432, -15.1356377, -17.4705759, -19.9268162, -22.5112171, -25.2306369, -28.091934, -31.1019668, -34.2675938, -37.5956733, -40.8245809, -43.9228203, -47.5045325, -51.519044, -55.915681, -60.6437699, -65.6526371, -70.8916089, -76.3100118, -81.857172, -87.482416, -93.1350701, -98.7644607, -104.343124, -109.788665, -115.04345, -120.049841, -125.401329, -130.714713, -135.433161, -139.303697, -142.073346, -143.489136, -143.29809, -142.262084, -138.804426, -132.473585, -122.818028, -109.386223, -95.8589015, -76.6192019, -50.1825264, -15.0642771, 30.2201439, 86.3736472, 156.617914, 238.326911, 332.991724, 450.509211, 597.002211, 778.593563, 1003.52496, 1275.31882, 1595.18611, 1954.59871, 2383.80969, 2895.46266, 3507.9503, 4249.40683, 5143.49493, 6207.19363, 7462.43284, 8955.64917, 10719.4168, 12622.0681, 14885.6871, 17599.7012, 20853.5375, 24736.6234, 31585.5645, 38680.8149, 45550.0074, 51720.7746, 56720.7492, 58530.6955, 59256.3599, 59088.3188, 58623.3036, 57891.1207, 56752.3452, 55245.2448, 53408.0877, 51279.1417, 48896.6748, 42810.3767, 39169.5362, 37134.2677, 35864.6856, 34520.9046, 32263.0391, 28251.2036, 24651.0392, 20469.2895},
        {20.8705639, 20.8074309, 20.5986508, 20.2510145, 19.7713127, 19.1663363, 18.442876, 17.6077227, 16.6676672, 15.6295002, 14.5000126, 13.2859951, 11.9942386, 10.6315338, 9.20467162, 7.72044278, 6.1856381, 4.60704837, 2.99146441, 1.345677, -0.323523039, -2.00934492, -3.70499784, -5.403691, -7.09863359, -8.78303483, -10.4501039, -12.09305, -13.7050823, -15.2794101, -16.8092425, -18.2877888, -19.7082581, -21.0638596, -22.0466407, -22.8067262, -23.4978602, -24.1230254, -24.6852042, -25.187379, -25.6325324, -26.0236468, -26.3637047, -26.6556885, -26.9025809, -27.1073642, -27.2730209, -27.4025334, -27.4988844, -27.5650562, -27.6040314, -27.8121083, -27.9947629, -28.1507868, -28.2789718, -28.3781096, -28.4469922, -28.4844111, -28.4891581, -28.4600252, -28.3958039, -28.2952861, -28.1572635, -27.9805279, -27.7638711, -27.5060848, -27.2059608, -26.8062601, -26.2510887, -25.654798, -25.0198119, -24.3485544, -23.6434495, -22.9069214, -22.1413939, -21.3492911, -20.533037, -19.6950557, -18.8377711, -17.9636073, -17.0749882, -16.174338, -15.2640806, -14.3466401, -13.1431919, -11.8027, -10.4775087, -9.1791692, -7.91923246, -6.7092496, -5.56077168, -4.48534976, -3.49453494, -2.59987826, -1.81293082, -1.14524368, -0.608367907, -0.213854582, 0.0267452256, 0.101880444, 0, -1.67790276, -3.38838365, -5.14103374, -6.94544407, -8.8112057, -10.7479097, -12.765147, -14.8725089, -17.0795862, -19.3959701, -21.8312516, -24.3950217, -27.0968715, -29.9463922, -32.9531746, -36.1268099, -39.0983005, -41.8231536, -45.1672912, -49.0591784, -53.4272801, -58.2000613, -63.3059868, -68.6735216, -74.2311306, -79.9072788, -85.6304312, -91.3290525, -96.9316079, -102.214056, -107.313287, -112.203517, -116.858964, -121.973976, -127.162703, -131.847262, -135.785832, -138.736589, -140.45771, -140.707375, -140.253856, -137.537815, -132.1198, -123.560362, -111.42005, -99.1059602, -81.3704604, -56.8124643, -24.0308857, 18.3753619, 71.2069687, 137.199235, 213.338865, 300.892249, 409.863467, 546.333555, 716.383546, 928.464295, 1185.74842, 1488.8233, 1830.49596, 2237.16662, 2719.30761, 3293.10668, 3984.43608, 4814.72271, 5796.06156, 6947.46014, 8322.14301, 9953.97961, 11738.2098, 13868.6603, 16427.1877, 19495.6486, 23155.8997, 29614.7477, 36297.8617, 42755.8606, 48539.3635, 53198.9897, 54799.5676, 55368.0339, 55091.775, 54555.4502, 53789.1922, 52657.6031, 51199.0922, 49452.0686, 47454.9414, 45246.1198, 40566.4032, 37514.4598, 35514.8542, 33992.1512, 32370.9153, 30075.7113, 26531.1038, 23504.4174, 19842.5723},
        {22.3686335, 22.4604349, 22.3848354, 22.1494735, 21.7619875, 21.2300157, 20.5611965, 19.7631682, 18.8435692, 17.8100377, 16.6702122, 15.431731, 14.1022324, 12.6893547, 11.2007364, 9.64401566, 8.02683096, 6.35682059, 4.64162289, 2.88887621, 1.10621889, -0.698710735, -2.51827431, -4.3448335, -6.17074997, -7.98838537, -9.79010135, -11.5682596, -13.3152217, -15.0233494, -16.6850043, -18.2925481, -19.8383424, -21.314749, -22.3821369, -23.2058471, -23.9552107, -24.6333634, -25.2434404, -25.7885772, -26.2719094, -26.6965723, -27.0657014, -27.3824322, -27.6499001, -27.8712405, -28.049589, -28.1880809, -28.2898518, -28.358037, -28.395772, -28.600723, -28.7772776, -28.9243544, -29.0408716, -29.1257478, -29.1779012, -29.1962504, -29.1797137, -29.1272096, -29.0376564, -28.9099725, -28.7430764, -28.5358865, -28.2873211, -27.9962988, -27.6617378, -27.2247752, -26.6279351, -25.989306, -25.3115519, -24.5973371, -23.8493257, -23.0701817, -22.2625694, -21.4291529, -20.5725964, -19.6955639, -18.8007197, -17.8907279, -16.9682526, -16.0359579, -15.0965081, -14.1525672, -12.9495152, -11.62407, -10.3160418, -9.03644411, -7.79629044, -6.60659433, -5.47836928, -4.4226288, -3.45038641, -2.57265562, -1.80044995, -1.1447829, -0.616667987, -0.227118729, 0.0128513639, 0.0922287778, 0, -1.71355425, -3.43266197, -5.16923977, -6.93520427, -8.74247209, -10.6029599, -12.5285842, -14.5312617, -16.622909, -18.8154427, -21.1207795, -23.5508359, -26.1175286, -28.8327743, -31.7084894, -34.7565907, -37.5138053, -39.9241657, -43.0744216, -46.8746635, -51.2349816, -56.0654662, -61.2762076, -66.7772963, -72.4788224, -78.2908763, -84.1235483, -89.8869288, -95.491108, -100.539236, -105.360888, -109.958237, -114.333454, -119.254522, -124.340708, -129.007065, -133.026021, -136.170007, -138.211453, -138.922788, -139.112024, -137.200831, -132.759026, -125.356424, -114.562841, -103.395103, -87.1142632, -64.4283865, -34.0455364, 5.32622319, 54.4337993, 115.780304, 186.272389, 267.037716, 367.821847, 494.501081, 652.951721, 851.188113, 1092.46259, 1377.69507, 1701.90135, 2086.89979, 2540.68424, 3077.02645, 3719.48855, 4487.13871, 5386.4823, 6433.33049, 7689.55807, 9190.98365, 10865.1498, 12872.8091, 15287.2601, 18181.8017, 21629.7325, 27704.5703, 33979.3401, 40027.286, 45421.6519, 49735.682, 51113.5988, 51510.3486, 51111.6134, 50499.0994, 49701.7286, 48583.413, 47181.325, 45532.6371, 43674.5221, 41644.1523, 38410.0222, 35944.8324, 33953.2446, 32139.9207, 30209.5224, 27866.7116, 24816.1501, 22241.5915, 19043.2416},
        {23.831104, 24.0695332, 24.1227838, 23.999114, 23.7067822, 23.2540466, 22.6491657, 21.9003977, 21.0160009, 20.0042336, 18.8733542, 17.6316209, 16.2872922, 14.8486262, 13.3238813, 11.7213159, 10.0491882, 8.31575656, 6.52927929, 4.6980147, 2.83022111, 0.934156837, -0.981919811, -2.90975052, -4.84107696, -6.76764084, -8.68118383, -10.5734476, -12.4361739, -14.2611043, -16.0399806, -17.7645445, -19.4265375, -21.0177015, -22.1843171, -23.0981229, -23.9350094, -24.6979291, -25.3898348, -26.0136789, -26.572414, -27.0689926, -27.5063673, -27.8874907, -28.2153153, -28.4927937, -28.7228784, -28.908522, -29.0526771, -29.1582961, -29.2283318, -29.4388996, -29.6160352, -29.7589374, -29.8668049, -29.9388364, -29.9742307, -29.9721864, -29.9319022, -29.852577, -29.7334094, -29.5735981, -29.3723419, -29.1288395, -28.8422896, -28.5118909, -28.1368422, -27.6600222, -27.0256622, -26.3491163, -25.6332342, -24.8808656, -24.0948603, -23.2780679, -22.4333381, -21.5635207, -20.6714653, -19.7600218, -18.8320397, -17.8903688, -16.9378588, -15.9773594, -15.0117203, -14.0437912, -12.8381032, -11.5216779, -10.2244015, -8.95685784, -7.72963035, -6.55330268, -5.43845846, -4.39568129, -3.43555478, -2.56866255, -1.8055882, -1.15691535, -0.633227614, -0.245108598, -0.00314191629, 0.0820888203, 0, -1.72951047, -3.44465194, -5.1591289, -6.88664586, -8.64090731, -10.4356178, -12.2844817, -14.2012037, -16.1994881, -18.2930396, -20.4955626, -22.8207615, -25.282341, -27.8940055, -30.6694595, -33.6224075, -36.23599, -38.4481167, -41.4725773, -45.209384, -49.558549, -54.4200845, -59.6940028, -65.280316, -71.0790365, -76.9901763, -82.9137478, -88.7497631, -94.3982345, -99.3709194, -104.098445, -108.5973, -112.883972, -117.719779, -122.748795, -127.416474, -131.515856, -134.839982, -137.181891, -138.334624, -139.180999, -138.092607, -134.64927, -128.430811, -119.017053, -108.893185, -94.0073222, -73.212944, -45.3635317, -9.31256596, 35.4144542, 91.4574217, 156.033295, 230.25021, 323.160783, 440.25072, 587.005728, 770.574511, 994.401825, 1260.11824, 1564.76254, 1926.21929, 2350.61138, 2849.65691, 3444.55475, 4152.15194, 4974.00483, 5922.99623, 7068.07476, 8446.47508, 10015.7588, 11905.0909, 14179.2852, 16903.1554, 20141.5154, 25836.0513, 31705.4863, 37345.4161, 42351.4363, 46319.1427, 47468.1956, 47687.4161, 47162.0865, 46484.0799, 45677.7364, 44597.9828, 43277.63, 41749.4883, 40046.3685, 38201.0811, 36232.4146, 34265.008, 32260.8765, 30182.0357, 27990.5009, 25648.2876, 23117.4113, 20873.6488, 18096.3295},
        {25.0387119, 25.4038978, 25.5748997, 25.560103, 25.367893, 25.0066551, 24.4847747, 23.8106371, 22.9926278, 22.0391321, 20.9585354, 19.759223, 18.4495804, 17.0379929, 15.5328458, 13.9425246, 12.2754147, 10.5399014, 8.74437002, 6.89720603, 5.00679477, 3.08152163, 1.12977197, -0.840068835, -2.8196154, -4.80048236, -6.77428433, -8.73263594, -10.6671518, -12.5694466, -14.4311349, -16.2438313, -17.9991505, -19.688707, -20.9818488, -22.0381022, -23.0208672, -23.9322166, -24.7742232, -25.5489597, -26.2584989, -26.9049138, -27.490277, -28.0166614, -28.4861398, -28.900785, -29.2626698, -29.573867, -29.8364493, -30.0524897, -30.2240609, -30.4588685, -30.6510626, -30.8004262, -30.906742, -30.9697929, -30.9893619, -30.9652319, -30.8971857, -30.7850061, -30.6284762, -30.4273788, -30.1814968, -29.8906131, -29.5545105, -29.172972, -28.7457804, -28.2244297, -27.5615733, -26.8565863, -26.1123821, -25.3318738, -24.5179747, -23.6735982, -22.8016574, -21.9050656, -20.9867361, -20.0495822, -19.0965171, -18.130454, -17.1543063, -16.1709872, -15.18341, -14.1944879, -12.9741899, -11.6468011, -10.3394064, -9.06247856, -7.82649027, -6.6419143, -5.51922342, -4.46889038, -3.50138793, -2.62718882, -1.85676582, -1.20059166, -0.669139106, -0.272880911, -0.022289826, 0.0721613953, 0, -1.84560019, -3.64986032, -5.42838894, -7.19679461, -8.9706859, -10.7656714, -12.5973596, -14.481359, -16.4332784, -18.4687262, -20.6033109, -22.8526412, -25.2323255, -27.7579726, -30.4451908, -33.3095888, -35.8926306, -38.1421898, -41.1463448, -44.8191017, -49.0744668, -53.8264462, -58.9890459, -64.4762723, -70.2021313, -76.0806293, -82.0257723, -87.9515664, -93.7720179, -99.162374, -104.362945, -109.359364, -114.137265, -119.181865, -124.229005, -128.881308, -132.974531, -136.344433, -138.826775, -140.257314, -141.706609, -141.399811, -138.93109, -133.894612, -125.884547, -116.517965, -102.860405, -84.0003107, -59.0261232, -27.0262852, 11.669165, 60.7308874, 118.510848, 186.371965, 271.824523, 379.618763, 514.504929, 682.578279, 886.938352, 1129.21739, 1405.88651, 1734.82546, 2123.04284, 2581.9756, 3130.56429, 3784.30519, 4543.8151, 5420.73266, 6481.25461, 7760.71654, 9227.38098, 10993.9755, 13119.0534, 15661.1678, 18678.8719, 23997.6712, 29467.4285, 34704.959, 39327.0779, 42950.6002, 43858.5428, 43890.7181, 43235.5692, 42534.8246, 41797.5019, 40843.7496, 39692.0288, 38360.8003, 36868.5253, 35233.6647, 33916.5489, 32151.1272, 30077.5681, 27836.0403, 25566.7123, 23409.7528, 21505.3304, 19410.6179, 17045.2807},
        {26.3216461, 26.8009052, 27.0803204, 27.1682276, 27.0729628, 26.8028618, 26.3662605, 25.7714949, 25.0269008, 24.1408142, 23.121571, 21.9775072, 20.7169585, 19.3482609, 17.8797504, 16.3197628, 14.6766341, 12.9587002, 11.1742969, 9.33176025, 7.43942609, 5.50563035, 3.53870894, 1.54699777, -0.461167248, -2.4774502, -4.49351517, -6.50102626, -8.49164755, -10.4570431, -12.3888771, -14.2788135, -16.1185165, -17.8996501, -19.3269173, -20.5414987, -21.6880237, -22.7674689, -23.7808107, -24.7290258, -25.6130906, -26.4339818, -27.1926757, -27.8901489, -28.527378, -29.1053395, -29.6250099, -30.0873657, -30.4933835, -30.8440398, -31.1403111, -31.4079204, -31.6225612, -31.7846738, -31.8946981, -31.9530743, -31.9602424, -31.9166425, -31.8227147, -31.678899, -31.4856355, -31.2433644, -30.9525255, -30.6135592, -30.2269053, -29.793004, -29.3122954, -28.7461017, -28.0566845, -27.3251613, -26.554508, -25.7477006, -24.9077149, -24.037527, -23.1401126, -22.2184479, -21.2755085, -20.3142706, -19.33771, -18.3488026, -17.3505243, -16.3458512, -15.337759, -14.3292237, -13.0964095, -11.7604687, -10.4452801, -9.16118009, -7.918505, -6.72759126, -5.59877527, -4.54239343, -3.56878214, -2.68827782, -1.91121684, -1.24793563, -0.70877058, -0.304058089, -0.0441345611, 0.0606636024, 0, -1.800916, -3.56283431, -5.30128979, -7.0318173, -8.7699517, -10.5312279, -12.3311806, -14.1853449, -16.1092555, -18.1184473, -20.2284552, -22.4548141, -24.8130587, -27.318724, -29.9873448, -32.834456, -35.4901497, -37.9148975, -40.9960691, -44.6666045, -48.8594436, -53.5075265, -58.543793, -63.9011834, -69.5126374, -75.3110951, -81.2294964, -87.2007814, -93.15789, -98.9622227, -104.64449, -110.159094, -115.460437, -120.775655, -125.922782, -130.647124, -134.821262, -138.317778, -141.009253, -142.768267, -144.777255, -145.200295, -143.653694, -139.753758, -133.116793, -124.556985, -112.19329, -95.3425438, -73.3215834, -45.4472445, -12.8364947, 29.1941221, 80.1605081, 141.682876, 219.67774, 318.132661, 441.035202, 593.393811, 777.943144, 996.304167, 1244.09762, 1539.73979, 1891.70685, 2311.40624, 2815.21225, 3417.21931, 4117.39804, 4926.18114, 5905.78856, 7089.14649, 8451.6353, 10092.5801, 12064.8268, 14421.2212, 17214.6091, 22151.7754, 27218.1422, 32053.0706, 36295.9216, 39586.0562, 40280.6026, 40155.9768, 39401.0685, 38694.6677, 38031.7451, 37203.146, 36214.0472, 35069.6256, 33775.058, 32335.5214, 31374.16, 29728.4296, 27644.2769, 25367.6487, 23144.4916, 21220.7524, 19842.3779, 17866.2346, 15869.0451},
        {27.8139918, 28.3862923, 28.756121, 28.9316251, 28.9209518, 28.7322482, 28.3736613, 27.8533383, 27.1794263, 26.3600725, 25.4034239, 24.3176277, 23.110831, 21.7911809, 20.3668246, 18.8459091, 17.2365816, 15.5469892, 13.7852789, 11.9595981, 10.0780937, 8.1489128, 6.18020265, 4.18011031, 2.15678291, 0.118367559, -1.92698862, -3.9711385, -6.00593498, -8.02323092, -10.0148792, -11.9727327, -13.8886443, -15.754467, -17.3185629, -18.6996517, -20.0193798, -21.2775016, -22.4737719, -23.6079453, -24.6797763, -25.6890195, -26.6354297, -27.5187614, -28.3387692, -29.0952078, -29.7878317, -30.4163956, -30.9806542, -31.4803619, -31.9152735, -32.2219541, -32.4647177, -32.6446885, -32.7629912, -32.82075, -32.8190896, -32.7591342, -32.6420083, -32.4688364, -32.240743, -31.9588523, -31.624289, -31.2381774, -30.801642, -30.3158072, -29.7817974, -29.1707154, -28.4543846, -27.6958471, -26.8981735, -26.0644345, -25.1977006, -24.3010425, -23.3775308, -22.4302363, -21.4622293, -20.4765807, -19.4763611, -18.464641, -17.444491, -16.418982, -15.3911843, -14.3641687, -13.1258519, -11.7914068, -10.4785927, -9.19746221, -7.95806821, -6.77046346, -5.64470077, -4.59083292, -3.61891268, -2.73899287, -1.96112626, -1.29536563, -0.751763792, -0.340373519, -0.0712476039, 0.0455611654, 0, -1.44348336, -2.90377278, -4.3928075, -5.92252679, -7.5048699, -9.15177607, -10.8751846, -12.6870346, -14.5992655, -16.6238165, -18.7726268, -21.0576357, -23.4907824, -26.0840063, -28.8492464, -31.7984422, -34.6712556, -37.4407291, -40.7346671, -44.5066639, -48.7103134, -53.2992098, -58.2269474, -63.4471201, -68.9133222, -74.5791477, -80.3981908, -86.3240457, -92.3103064, -98.3700691, -104.375609, -110.262668, -115.966992, -121.572016, -126.939636, -131.892367, -136.321643, -140.1189, -143.175574, -145.383101, -147.851738, -148.883154, -148.130318, -145.246199, -139.883769, -132.299855, -121.392606, -106.664027, -87.6161251, -63.7509056, -36.7497381, -1.27159898, 43.0797629, 98.0825112, 168.336215, 257.17369, 367.927753, 504.795085, 670.048303, 865.01762, 1084.18129, 1346.88436, 1663.75645, 2046.30348, 2507.51616, 3059.70366, 3701.65109, 4442.16139, 5339.62567, 6424.38515, 7676.38337, 9184.11275, 10995.3599, 13157.9114, 15719.5538, 20260.2605, 24912.5847, 29341.2661, 33211.0447, 36186.6603, 36723.1957, 36501.4861, 35704.6957, 35013.0015, 34412.8337, 33683.5342, 32821.8877, 31824.679, 30688.6929, 29410.714, 28421.1614, 26826.406, 24863.1951, 22768.2758, 20778.3956, 19130.3017, 18060.7415, 16247.9024, 14572.1596},
        {29.9100206, 30.5142798, 30.9180388, 31.1291009, 31.1552693, 31.0043472, 30.6841379, 30.2024446, 29.5670705, 28.7858189, 27.866493, 26.816896, 25.6448312, 24.3581018, 22.964511, 21.4718621, 19.8879583, 18.2206028, 16.4775988, 14.6667497, 12.7958586, 10.8727287, 8.90516335, 6.90096569, 4.867939, 2.81388649, 0.746611399, -1.32608304, -3.39639359, -5.45651703, -7.49865011, -9.51498962, -11.4977323, -13.439075, -15.1373779, -16.6858333, -18.1799629, -19.6182538, -20.9991931, -22.3212681, -23.5829657, -24.7827732, -25.9191777, -26.9906663, -27.9957261, -28.9328443, -29.800508, -30.5972044, -31.3214204, -31.9716434, -32.5463604, -32.8896395, -33.1577344, -33.3524797, -33.4757101, -33.5292601, -33.5149645, -33.4346578, -33.2901747, -33.0833498, -32.8160178, -32.4900132, -32.1071708, -31.6693251, -31.1783108, -30.6359625, -30.0441149, -29.3831726, -28.6340551, -27.8427943, -27.0126143, -26.1467387, -25.2483916, -24.3207966, -23.3671777, -22.3907587, -21.3947635, -20.3824158, -19.3569396, -18.3215587, -17.2794969, -16.233978, -15.188226, -14.1454647, -12.9201826, -11.6139411, -10.3305697, -9.07941731, -7.86983247, -6.71116386, -5.61276017, -4.58397005, -3.63414217, -2.77262521, -2.00876782, -1.35191868, -0.811426445, -0.396639795, -0.116907393, 0.0184220947, 0, -0.317338927, -0.835498835, -1.55447855, -2.47427689, -3.59489268, -4.91632474, -6.43857189, -8.16163297, -10.0855068, -12.2101922, -14.5356879, -17.0619929, -19.7891059, -22.7170258, -25.8457513, -29.1752813, -32.575273, -36.0271049, -39.8255963, -43.9428157, -48.3508315, -53.0217123, -57.9275264, -63.0403424, -68.3322287, -73.7752537, -79.3414859, -85.0029937, -90.7318457, -96.5482251, -102.358443, -108.120135, -113.790933, -119.650633, -125.495788, -131.037042, -136.135383, -140.651795, -144.447265, -147.382779, -150.011091, -151.29088, -150.947786, -148.70745, -144.29551, -138.081286, -128.98582, -116.57383, -100.410037, -80.059161, -57.0992777, -26.6209835, 11.6766294, 59.4070105, 120.863382, 199.135801, 297.314329, 419.890494, 568.234368, 742.187142, 934.099206, 1165.26914, 1450.59775, 1801.91711, 2225.85947, 2731.44393, 3313.17071, 3976.29476, 4779.30666, 5750.58391, 6874.80589, 8230.77099, 9861.98295, 11811.9455, 14124.1625, 18227.7811, 22434.2507, 26440.6642, 29944.1146, 32641.6948, 33144.0129, 32958.9696, 32249.2941, 31606.927, 31024.7011, 30316.6114, 29484.4328, 28529.9402, 27454.9083, 26261.1122, 24494.4755, 22865.5696, 21351.1497, 19927.9707, 18572.7878, 17262.3562, 15973.431, 14541.7944, 13140.4064},
        {31.9215462, 32.5885224, 33.0553125, 33.3294254, 33.41837, 33.329655, 33.0707893, 32.6492817, 32.072641, 31.348376, 30.4839956, 29.4870086, 28.3649238, 27.12525, 25.7754961, 24.3231708, 22.7757831, 21.1408416, 19.4258553, 17.6383329, 15.7857833, 13.8757154, 11.9156378, 9.91305951, 7.87548927, 5.81043592, 3.72540828, 1.62791518, -0.474534562, -2.57443213, -4.6642687, -6.73653544, -8.78372354, -10.7983242, -12.6246709, -14.33245, -15.9931286, -17.6040057, -19.1623805, -20.6655523, -22.1108202, -23.4954834, -24.816841, -26.0721923, -27.2588364, -28.3740725, -29.4151998, -30.3795174, -31.2643245, -32.0669204, -32.7846041, -33.1801032, -33.4903737, -33.7177997, -33.8647654, -33.9336551, -33.926853, -33.8467433, -33.6957103, -33.4761382, -33.1904112, -32.8409136, -32.4300296, -31.9601434, -31.4336392, -30.8529014, -30.220314, -29.5212192, -28.7413681, -27.9182933, -27.0554837, -26.1564284, -25.2246163, -24.2635365, -23.2766779, -22.2675296, -21.2395806, -20.1963198, -19.1412364, -18.0778192, -17.0095573, -15.9399397, -14.8724554, -13.8105935, -12.6009341, -11.3287212, -10.0810842, -8.86660421, -7.69386216, -6.57143906, -5.50791591, -4.51187371, -3.59189343, -2.75655608, -2.01444265, -1.37413413, -0.844211522, -0.43325581, -0.14984799, -0.00256905562, 0, 0.768789563, 1.1603022, 1.1860066, 0.857371441, 0.185865399, -0.817042845, -2.13988461, -3.77119122, -5.69949398, -7.91332423, -10.4012133, -13.1516924, -16.1532931, -19.3945464, -22.8639839, -26.5501367, -30.435763, -34.5085694, -38.7701456, -43.207786, -47.8087846, -52.5604358, -57.4500336, -62.4648723, -67.5922461, -72.8194491, -78.1337756, -83.5225197, -88.9729757, -94.4782778, -100.017739, -105.576901, -111.141306, -117.214826, -123.523839, -129.6565, -135.442852, -140.71294, -145.296805, -149.024492, -151.907982, -153.540007, -153.715014, -152.227451, -148.871762, -144.208684, -137.074803, -127.072993, -113.806131, -96.8770902, -77.6473101, -51.811745, -19.2534257, 21.4357566, 74.2999879, 142.199909, 227.996162, 336.236838, 467.711616, 621.369319, 788.598853, 991.057603, 1245.94279, 1564.81086, 1949.66025, 2406.87669, 2928.20232, 3516.22636, 4227.23121, 5087.12738, 6081.16109, 7282.16318, 8730.12024, 10465.0189, 12526.8457, 16197.1362, 19963.9408, 23556.859, 26705.4901, 29139.4334, 29630.2566, 29504.2786, 28901.6839, 28311.8546, 27736.751, 27032.8343, 26209.34, 25275.5031, 24240.5591, 23113.7432, 20607.7498, 18967.9112, 17912.177, 17158.4966, 16424.8196, 15429.0957, 13889.2742, 12822.4102, 11637.5931},
        {33.5062484, 34.3170789, 34.9224934, 35.3298945, 35.5466844, 35.5802656, 35.4380406, 35.1274116, 34.6557811, 34.0305515, 33.2591251, 32.3489045, 31.307292, 30.1416899, 28.8595007, 27.4681268, 25.9749706, 24.3874345, 22.7129209, 20.9588322, 19.1325707, 17.2415389, 15.2931392, 13.294774, 11.2538457, 9.17775667, 7.07390931, 4.94970603, 2.81254923, 0.669841314, -1.47101533, -3.60261829, -5.71756517, -7.80845357, -9.75408521, -11.6063492, -13.4185023, -15.186835, -16.9076379, -18.5772014, -20.1918161, -21.7477724, -23.2413608, -24.6688718, -26.026596, -27.3108238, -28.5178456, -29.6439521, -30.6854337, -31.6385809, -32.4996841, -32.9713081, -33.3498365, -33.6379273, -33.8382384, -33.9534277, -33.986153, -33.9390724, -33.8148436, -33.6161246, -33.3455732, -33.0058475, -32.5996052, -32.1295043, -31.5982027, -31.0083582, -30.3626288, -29.6441258, -28.83643, -27.9825123, -27.086298, -26.151712, -25.1826793, -24.1831249, -23.1569739, -22.1081513, -21.0405821, -19.9581914, -18.8649041, -17.7646452, -16.6613399, -15.5589131, -14.4612898, -13.3723951, -12.1738286, -10.9332509, -9.71979307, -8.54134971, -7.40581555, -6.32108532, -5.29505372, -4.33561548, -3.4506653, -2.6480979, -1.935808, -1.3216903, -0.813639536, -0.419550407, -0.147317631, -0.00483592427, 0, 1.38064176, 2.28846345, 2.74117339, 2.75647991, 2.35209134, 1.54571602, 0.355062261, -1.20216159, -3.10824721, -5.34548627, -7.89617044, -10.7425914, -13.8670408, -17.2518103, -20.8791916, -24.7314764, -28.8568988, -33.2471708, -37.7354285, -42.3180941, -46.9915898, -51.7523378, -56.5967603, -61.5212794, -66.5223175, -71.5962967, -76.7396392, -81.9487672, -87.220103, -92.5194747, -97.8851164, -103.322628, -108.837611, -115.041311, -121.636507, -128.164468, -134.449102, -140.314315, -145.584013, -150.082104, -153.580177, -155.970377, -157.086846, -156.763728, -154.835165, -151.813294, -146.684762, -139.114215, -128.766295, -115.305645, -100.03163, -78.9761801, -51.8842947, -17.4608222, 27.7130329, 86.1025934, 160.173182, 253.563823, 367.299388, 501.124345, 648.230104, 826.930537, 1051.69914, 1332.47871, 1671.53506, 2074.65784, 2535.30619, 3056.07296, 3684.76304, 4443.60683, 5310.75949, 6359.30749, 7626.73804, 9150.53837, 10968.1957, 14231.773, 17585.5379, 20788.3335, 23599.0032, 25776.3903, 26234.0766, 26139.6744, 25614.282, 25061.9496, 24491.7736, 23794.9536, 22986.4811, 22081.3471, 21094.5429, 20041.0598, 17310.0201, 15768.2572, 15017.7234, 14660.3713, 14298.1531, 13533.0215, 11966.9287, 11139.2645, 10095.6432},
        {34.2768098, 35.3689484, 36.2405986, 36.8993948, 37.3529715, 37.6089633, 37.6750047, 37.5587304, 37.2677747, 36.8097722, 36.1923576, 35.4231652, 34.5098298, 33.4599857, 32.2812676, 30.9813099, 29.5677473, 28.0482142, 26.4303453, 24.7217749, 22.9301378, 21.0630683, 19.1282011, 17.1331707, 15.0856117, 12.9931585, 10.8634457, 8.70410786, 6.52277951, 4.32709518, 2.12468942, -0.0768032308, -2.26974824, -4.44651107, -6.50116108, -8.4772793, -10.4196282, -12.3237632, -14.1852396, -15.9996127, -17.762438, -19.4692708, -21.1156664, -22.6971802, -24.2093676, -25.6477839, -27.0079845, -28.2855248, -29.47596, -30.5748456, -31.5777369, -32.1579621, -32.6402891, -33.0272584, -33.3214105, -33.5252862, -33.641426, -33.6723705, -33.6206604, -33.4888362, -33.2794386, -32.9950081, -32.6380855, -32.2112112, -31.716926, -31.1577703, -30.5362849, -29.8232214, -28.9900949, -28.105008, -27.1725623, -26.1973589, -25.1839993, -24.1370849, -23.0612169, -21.9609968, -20.8410259, -19.7059055, -18.560237, -17.4086217, -16.2556611, -15.1059564, -13.964109, -12.8347203, -11.6344922, -10.4138246, -9.22417391, -8.07299398, -6.96773868, -5.91586185, -4.92481736, -4.00205907, -3.15504083, -2.39121651, -1.71803996, -1.14296505, -0.67344562, -0.316935544, -0.0808886755, 0.0272411253, 0, 1.01157298, 1.6210431, 1.84161082, 1.68647661, 1.16884093, 0.301904258, -0.901132947, -2.42707022, -4.26270708, -6.39484308, -8.81027775, -11.4958106, -14.4382412, -17.6243691, -21.0409937, -24.6749147, -28.5564993, -32.6787709, -36.9299831, -41.3062714, -45.8037712, -50.418618, -55.1469472, -59.9848944, -64.9285951, -69.9741846, -75.1177984, -80.3555721, -85.6836412, -91.079241, -96.5643371, -102.140735, -107.81024, -114.005315, -120.512438, -127.003242, -133.350335, -139.426326, -145.103823, -150.255434, -155.05538, -158.982862, -161.851474, -163.474815, -163.666481, -162.253823, -159.033246, -153.814905, -146.408959, -136.625567, -126.188048, -110.655088, -88.8167134, -59.2360088, -20.0127117, 30.5454076, 94.1305785, 171.637244, 265.736735, 379.970696, 514.050236, 676.648113, 870.738693, 1102.40211, 1382.98108, 1721.40268, 2120.97869, 2589.18039, 3154.06857, 3832.05746, 4581.83025, 5486.41384, 6582.20624, 7905.60545, 9493.00948, 12402.8585, 15393.9978, 18247.3148, 20743.6972, 22664.0323, 23017.2387, 22870.8187, 22336.5036, 21785.5183, 21226.878, 20561.4752, 19803.7312, 18968.0673, 18068.9045, 17120.664, 15242.0684, 14009.9405, 13219.848, 12667.3583, 12148.039, 11457.4578, 10391.1823, 9547.7087, 8545.07308},
        {34.7165028, 36.1706139, 37.3828023, 38.3611984, 39.1139323, 39.6491346, 39.9749353, 40.0994649, 40.0308537, 39.7772319, 39.3467298, 38.7474779, 37.9876063, 37.0752453, 36.0185254, 34.8255767, 33.5045296, 32.0635144, 30.5106614, 28.8541009, 27.1019632, 25.2623786, 23.3434774, 21.3533899, 19.3002465, 17.1921773, 15.0373128, 12.8437833, 10.6197189, 8.37325016, 6.11250721, 3.84562042, 1.58072008, -0.674063498, -2.80484284, -4.85859069, -6.88355091, -8.87502521, -10.8283153, -12.738723, -14.6015498, -16.4120977, -18.1656682, -19.8575631, -21.4830842, -23.0375331, -24.5162115, -25.9144213, -27.227464, -28.4506414, -29.5792553, -30.3040436, -30.9314741, -31.463451, -31.9018781, -32.2486597, -32.5056997, -32.6749023, -32.7581715, -32.7574114, -32.674526, -32.5114195, -32.2699957, -31.9521589, -31.5598132, -31.0948624, -30.5592108, -29.9010341, -29.0697979, -28.1782171, -27.2316787, -26.2355696, -25.195277, -24.1161879, -23.0036894, -21.8631685, -20.7000122, -19.5196077, -18.327342, -17.1286021, -15.928775, -14.7332479, -13.5474078, -12.3766417, -11.174975, -9.97455596, -8.80815007, -7.68281118, -6.60559315, -5.5835498, -4.62373498, -3.73320251, -2.91900625, -2.18820003, -1.54783769, -1.00497306, -0.566659987, -0.239952306, -0.0319038554, 0.050431526, 0, 0.550622004, 0.790309936, 0.726221735, 0.365515338, -0.284651318, -1.2171203, -2.42473366, -3.90033346, -5.63676178, -7.62686067, -9.86347219, -12.3394384, -15.0476014, -17.9808032, -21.1318859, -24.4936915, -28.0553927, -31.8093072, -35.7565772, -39.8892584, -44.1994066, -48.6790774, -53.3203267, -58.1152101, -63.0557835, -68.1341026, -73.342223, -78.6722007, -84.1160913, -89.6412356, -95.2734665, -101.012254, -106.857069, -113.090525, -119.570522, -126.079451, -132.531838, -138.842212, -144.925099, -150.695025, -156.570435, -161.808573, -166.22537, -169.636763, -171.858683, -172.506459, -171.646783, -169.145741, -164.869419, -158.683904, -152.315142, -141.496195, -124.385449, -99.5571052, -66.4343266, -24.0591102, 28.5265469, 89.5093996, 163.248528, 257.126194, 377.193055, 522.995705, 687.192579, 872.214078, 1097.04865, 1373.08433, 1711.49343, 2123.60806, 2621.55128, 3216.99877, 3856.01718, 4625.98486, 5561.28826, 6696.31385, 8065.4481, 10633.366, 13271.5934, 15781.9445, 17966.2337, 19626.2753, 19856.3287, 19637.4665, 19074.7408, 18522.7464, 17987.7581, 17372.0749, 16687.1707, 15944.519, 15155.5938, 14331.8688, 13273.2209, 12340.9175, 11514.0339, 10771.6456, 10092.8283, 9456.6573, 8842.20817, 7990.52907, 7044.53514},
        {35.1893876, 37.1183168, 38.7719593, 40.1594877, 41.2900746, 42.1728924, 42.8171139, 43.2319114, 43.4264576, 43.409925, 43.1914862, 42.7803137, 42.18558, 41.4164577, 40.4821194, 39.3917375, 38.1544847, 36.7795335, 35.2760565, 33.6532261, 31.920215, 30.0861957, 28.1603407, 26.1518226, 24.069814, 21.9234873, 19.7220152, 17.4745702, 15.1903248, 12.8784516, 10.5481231, 8.20851192, 5.86879056, 3.53813158, 1.38664038, -0.660361932, -2.67948852, -4.6667894, -6.61831462, -8.53011421, -10.3982382, -12.2187366, -13.9876595, -15.7010568, -17.3549787, -18.9454752, -20.4685962, -21.9203918, -23.2969121, -24.5942071, -25.8083267, -26.7386017, -27.582066, -28.3390341, -29.0098204, -29.5947396, -30.0941062, -30.5082346, -30.8374394, -31.0820351, -31.2423363, -31.3186575, -31.3113131, -31.2206177, -31.0468859, -30.7904322, -30.451571, -29.9377993, -29.1588424, -28.3064443, -27.3869366, -26.4066507, -25.3719181, -24.2890703, -23.1644389, -22.0043554, -20.8151513, -19.6031581, -18.3747073, -17.1361304, -15.8937589, -14.6539244, -13.4229584, -12.2071923, -11.0058609, -9.82899329, -8.68670408, -7.58555503, -6.53210797, -5.53292472, -4.59456709, -3.7235969, -2.92657597, -2.21006612, -1.58062916, -1.04482691, -0.609221199, -0.280373831, -0.0648466306, 0.0307985858, 0, 0.497386476, 0.718422792, 0.667260281, 0.348050275, -0.235055894, -1.07790689, -2.17635139, -3.52623806, -5.12341556, -6.96373256, -9.04303774, -11.3571797, -13.9020073, -16.673369, -19.6671135, -22.8790895, -26.249179, -29.7652353, -33.5496986, -37.5864247, -41.8592693, -46.3520882, -51.0487373, -55.9330724, -60.9889491, -66.2002234, -71.550751, -77.0243877, -82.6049894, -88.1626507, -93.836701, -99.6451241, -105.605904, -112.13806, -119.059059, -126.079424, -133.09683, -140.008949, -146.713456, -153.108024, -158.937132, -164.298273, -169.119093, -173.327238, -176.850356, -180.90903, -183.814734, -185.17188, -184.584879, -181.658142, -176.586147, -167.662048, -153.742524, -133.999236, -108.246933, -76.0116279, -36.819337, 5.69526657, 57.0486125, 127.239304, 225.734308, 347.823762, 484.293802, 635.764902, 819.521277, 1045.19821, 1321.24237, 1656.98088, 2066.09913, 2559.81726, 3097.86276, 3750.8703, 4547.88434, 5517.94938, 6690.10987, 8914.7478, 11194.2358, 13352.2839, 15212.6022, 16598.9009, 16689.2654, 16383.4468, 15781.8517, 15237.1676, 14749.9396, 14215.5957, 13639.937, 13028.7643, 12387.8787, 11723.0812, 10665.2442, 9860.50468, 9236.72051, 8721.74961, 8243.44988, 7729.67922, 7108.29554, 6440.75253, 5687.00314},
        {35.0722149, 37.4406031, 39.5075561, 41.2830007, 42.7768634, 43.9990711, 44.9595505, 45.6682283, 46.1350313, 46.3698861, 46.3827196, 46.1834584, 45.7820292, 45.1883588, 44.412374, 43.4640014, 42.3531678, 41.0897999, 39.6838244, 38.1451681, 36.4837576, 34.7095198, 32.8323814, 30.862269, 28.8091095, 26.6828295, 24.4933557, 22.2506149, 19.9645339, 17.6450393, 15.3020579, 12.9455165, 10.5853416, 8.23146016, 6.10150786, 4.09718803, 2.11770751, 0.166252439, -1.753991, -3.63983663, -5.48809829, -7.2955898, -9.05912501, -10.7755177, -12.4415818, -14.054131, -15.6099793, -17.1059404, -18.5388281, -19.9054564, -21.2026389, -22.3427699, -23.4089129, -24.3997118, -25.3138105, -26.1498529, -26.9064829, -27.5823444, -28.1760812, -28.6863373, -29.1117565, -29.4509827, -29.7026597, -29.8654316, -29.9379421, -29.9188352, -29.8067546, -29.4703767, -28.7814981, -28.0068113, -27.1533854, -26.2282896, -25.2385931, -24.191365, -23.0936746, -21.9525909, -20.7751831, -19.5685205, -18.339672, -17.0957071, -15.8436947, -14.590704, -13.3438043, -12.1100646, -10.92415, -9.77875119, -8.66622741, -7.59275226, -6.56449936, -5.58764234, -4.66835482, -3.81281043, -3.02718279, -2.31764554, -1.69037229, -1.15153667, -0.707312316, -0.363872845, -0.127391884, -0.00404306057, 0, 0.601704283, 0.934253263, 1.000716, 0.804161553, 0.347658983, -0.365722649, -1.33291428, -2.55084686, -4.01645132, -5.7266586, -7.67839964, -9.86860539, -12.2942068, -14.9521347, -17.8393202, -20.9526942, -24.2144596, -27.6108719, -31.3078897, -35.2864307, -39.5274127, -44.0117536, -48.7203712, -53.6341832, -58.7341074, -64.0010617, -69.4159639, -74.9597318, -80.6132831, -86.1906018, -91.9007487, -97.7747217, -103.843519, -110.620605, -117.895746, -125.330046, -132.809764, -140.221158, -147.450486, -154.384005, -160.194802, -165.699359, -170.923465, -175.892913, -180.633494, -187.570733, -193.730753, -198.539412, -201.422568, -201.806078, -198.686305, -192.44354, -182.7357, -169.268943, -151.84792, -130.233062, -104.1848, -77.8136886, -45.5413678, 2.95611605, 77.5562868, 172.760674, 281.080244, 402.534608, 549.866364, 729.978052, 946.535935, 1205.60352, 1525.11065, 1916.27489, 2356.45185, 2895.93061, 3557.65017, 4364.54956, 5339.5678, 7204.11832, 9108.04712, 10899.6746, 12427.3212, 13539.3073, 13512.8633, 13148.1263, 12538.1698, 12023.0528, 11598.0833, 11155.659, 10696.2334, 10220.2604, 9728.19355, 9220.48664, 8059.35126, 7353.63717, 6952.76517, 6706.15604, 6463.23058, 6073.40958, 5386.11382, 4936.78504, 4373.19748},
        {33.2162295, 35.1970168, 36.9500885, 38.4821369, 39.7998539, 40.9099318, 41.8190627, 42.5339386, 43.0612517, 43.4076941, 43.579958, 43.5847354, 43.4287185, 43.1185994, 42.6610702, 42.062823, 41.33055, 40.4709432, 39.4906949, 38.396497, 37.1950418, 35.8930214, 34.4971278, 33.0140532, 31.4504897, 29.8131295, 28.1086646, 26.3437872, 24.5251893, 22.6595632, 20.7536009, 18.8139945, 16.8474363, 14.8606182, 12.9302648, 11.026579, 9.11892189, 7.21171299, 5.3093717, 3.41631745, 1.53696968, -0.324252184, -2.1629287, -3.97464043, -5.75496795, -7.49949182, -9.2037926, -10.8634509, -12.4740472, -14.0311621, -15.5303762, -16.8862858, -18.1772112, -19.4004887, -20.5534544, -21.6334444, -22.6377949, -23.563842, -24.4089219, -25.1703707, -25.8455245, -26.4317196, -26.926292, -27.3265779, -27.6299134, -27.8336347, -27.935078, -27.7700066, -27.1780633, -26.4898158, -25.7130745, -24.8556494, -23.925351, -22.9299894, -21.8773749, -20.7753177, -19.6316281, -18.4541163, -17.2505926, -16.0288673, -14.7967505, -13.5620526, -12.3325837, -11.1161542, -9.99670926, -8.94239024, -7.92042341, -6.9361483, -5.99490442, -5.1020313, -4.26286846, -3.48275541, -2.76703169, -2.12103681, -1.55011029, -1.05959166, -0.654820431, -0.34113613, -0.123878277, -0.00838639314, 0, 0.0509523361, -0.0273286456, -0.241888344, -0.599772159, -1.10802549, -1.77369373, -2.60382229, -3.60545656, -4.78564195, -6.15142384, -7.70984765, -9.46795876, -11.4328026, -13.6114245, -16.01087, -18.6381843, -21.3161336, -24.0254378, -27.1899643, -30.7772703, -34.7549126, -39.0904483, -43.7514343, -48.7054276, -53.9199852, -59.3626641, -65.0010212, -70.8026135, -76.7349979, -82.6585143, -88.6872497, -94.8209265, -101.059267, -107.467663, -114.013003, -120.64466, -127.342657, -134.087013, -140.85775, -147.63489, -154.585718, -161.445996, -168.159107, -174.668432, -180.917355, -186.457354, -191.72169, -196.75172, -201.588802, -206.274294, -212.19009, -216.398576, -217.520686, -214.645561, -207.818257, -196.654644, -180.770591, -161.826571, -136.929168, -100.954493, -49.7856387, 15.9727821, 93.4666584, 183.934752, 295.550972, 433.30588, 599.130492, 797.222152, 1042.99655, 1345.52365, 1692.3523, 2117.38451, 2637.13986, 3268.13791, 4026.89825, 5481.44959, 6958.92508, 8337.96701, 9497.21768, 10315.3194, 10199.7139, 9814.37765, 9239.89614, 8784.04502, 8437.48019, 8096.1949, 7755.57818, 7411.01905, 7057.90653, 6691.62964, 5770.99575, 5202.42885, 4875.15549, 4678.40217, 4501.39545, 4233.36185, 3763.5279, 3524.68665, 2959.94501},
        {36.6432225, 40.1211924, 43.2291533, 45.9790412, 48.3827919, 50.4523411, 52.1996248, 53.6365786, 54.7751386, 55.6272404, 56.20482, 56.5198131, 56.5841555, 56.4097832, 56.0086319, 55.3926374, 54.5737356, 53.5638622, 52.3749532, 51.0189444, 49.5077715, 47.8533705, 46.067677, 44.162627, 42.1501563, 40.0422007, 37.850696, 35.5875781, 33.2647827, 30.8942458, 28.4879031, 26.0576904, 23.6155437, 21.1733986, 19.0503294, 17.0982424, 15.1653526, 13.2536285, 11.3650388, 9.50155194, 7.66513665, 5.8577615, 4.08139509, 2.338006, 0.629562834, -1.0419658, -2.67461132, -4.26640511, -5.81537859, -7.31956314, -8.77699019, -10.2239373, -11.6193606, -12.9604624, -14.244445, -15.4685107, -16.6298617, -17.7257005, -18.7532292, -19.7096502, -20.5921657, -21.3979782, -22.1242898, -22.7683029, -23.3272198, -23.7982428, -24.1785741, -24.3427077, -24.1680835, -23.9049809, -23.5585568, -23.1339682, -22.6363721, -22.0709255, -21.4427853, -20.7571086, -20.0190523, -19.2337734, -18.4064289, -17.5421759, -16.6461712, -15.7235719, -14.779535, -13.8192174, -12.7324035, -11.5843632, -10.4417533, -9.31347494, -8.20842909, -7.13551683, -6.1036392, -5.12169727, -4.19859209, -3.34322471, -2.56449618, -1.87130756, -1.27255991, -0.77715427, -0.393991703, -0.131973262, 0, 3.16757266, 5.5149434, 7.08784394, 7.93200596, 8.09316119, 7.61704131, 6.54937804, 4.93590308, 2.82234813, 0.254444896, -2.72207491, -6.0614796, -9.71803745, -13.6460168, -17.7996859, -22.133313, -27.4443383, -33.8074833, -39.2698417, -43.9663616, -48.031991, -51.6016779, -54.8103701, -57.7930157, -60.6845627, -63.619959, -66.7341527, -70.1620916, -74.0387237, -79.1020911, -84.6629132, -90.6752099, -97.0930008, -104.003981, -111.295332, -118.818591, -126.487674, -134.216498, -141.918981, -149.509041, -156.000845, -162.481897, -169.042152, -175.771564, -182.76009, -191.364154, -200.090624, -208.712839, -217.004137, -224.737855, -231.684583, -237.623768, -243.208534, -248.311602, -251.212371, -250.905608, -246.386081, -235.811011, -219.203063, -197.498594, -172.740926, -140.263801, -98.3600274, -44.4232873, 25.6762647, 115.36915, 227.842523, 366.46381, 535.492787, 738.684466, 949.84036, 1207.27721, 1525.11856, 1917.48792, 2398.50884, 3358.97737, 4342.17638, 5268.06128, 6056.58748, 6627.71038, 6604.01804, 6401.0781, 6066.28911, 5761.3512, 5488.41854, 5202.01966, 4906.69003, 4606.9651, 4307.38034, 4012.47122, 3274.7691, 2920.72955, 2810.97408, 2806.12418, 2766.80134, 2553.62707, 2027.22285, 1948.24227, 1580.50333},
        {32.139939, 36.4825539, 40.4069449, 43.9263424, 47.0539767, 49.8030783, 52.1868775, 54.2186047, 55.9114902, 57.2787644, 58.3336578, 59.0894006, 59.5592233, 59.7563562, 59.6940297, 59.3854741, 58.84392, 58.0825975, 57.1147372, 55.9535693, 54.6123243, 53.1042325, 51.4425243, 49.64043, 47.7111802, 45.668005, 43.5241349, 41.2928003, 38.9872316, 36.620659, 34.2063131, 31.7574241, 29.2872225, 26.8089386, 24.6756205, 22.7234421, 20.7844931, 18.8609759, 16.9550933, 15.0690479, 13.2050422, 11.3652791, 9.55196099, 7.76729063, 6.01347063, 4.29270361, 2.60719223, 0.959139104, -0.649253123, -2.21578182, -3.73824434, -5.22618186, -6.66539335, -8.05342161, -9.38780941, -10.6660995, -11.8858348, -13.0445579, -14.1398117, -15.169139, -16.1300825, -17.020185, -17.8369893, -18.5780382, -19.2408745, -19.823041, -20.3220803, -20.6365226, -20.6672751, -20.6160871, -20.4869198, -20.2837346, -20.0104931, -19.6711564, -19.269686, -18.8100433, -18.2961897, -17.7320864, -17.121695, -16.4689768, -15.7778932, -15.0524054, -14.296475, -13.5140633, -12.5629525, -11.5232683, -10.4768932, -9.43253238, -8.3988911, -7.3846745, -6.39858774, -5.44933601, -4.54562448, -3.69615831, -2.90964268, -2.19478277, -1.56028375, -1.01485079, -0.567189058, -0.226003738, 0, 2.37201249, 4.12845071, 5.30012939, 5.91786328, 6.01246711, 5.61475562, 4.75554353, 3.4656456, 1.77587656, -0.282948868, -2.68001593, -5.3845099, -8.36561604, -11.5925196, -15.0344059, -18.6604601, -22.9742865, -28.0214159, -32.5622287, -36.6804287, -40.4597194, -43.9838045, -47.3363874, -50.6011719, -53.8618616, -57.20216, -60.7057708, -64.4563976, -68.5377439, -73.3090368, -78.4774312, -84.0439739, -90.0097115, -96.5764316, -103.64481, -111.061993, -118.768803, -126.706067, -134.814607, -143.035249, -150.705286, -158.552756, -166.636565, -175.015621, -183.74883, -194.168467, -204.741727, -215.209175, -225.311375, -234.788891, -242.811323, -250.388046, -258.637576, -267.791933, -276.273204, -283.316095, -288.155316, -289.104586, -286.528919, -281.798048, -277.011511, -268.422776, -254.440309, -233.150884, -202.096178, -159.06808, -101.83306, -28.1764156, 64.02337, 176.940525, 290.710345, 432.936596, 613.422368, 841.97075, 1128.38483, 1721.13692, 2331.69358, 2910.1906, 3406.76377, 3771.54887, 3774.38936, 3665.90823, 3473.50941, 3284.58555, 3102.79503, 2906.80094, 2701.51142, 2491.83462, 2282.6787, 2078.95181, 1779.25092, 1603.66348, 1501.38509, 1421.61137, 1313.53791, 1126.36034, 809.274263, 686.263591, 525.855419},
        {31.6116123, 36.0544624, 40.1409146, 43.8808894, 47.2843074, 50.3610889, 53.1211545, 55.5744247, 57.7308199, 59.6002607, 61.1926676, 62.5179609, 63.5860613, 64.4068892, 64.990365, 65.3464094, 65.4849428, 65.4158856, 65.1491583, 64.6946815, 64.0623757, 63.2621613, 62.3039587, 61.1976886, 59.9532714, 58.5806276, 57.0896776, 55.490342, 53.7925413, 52.0061959, 50.1412263, 48.2075531, 46.2150967, 44.1737775, 42.263434, 40.4054536, 38.5191817, 36.6090246, 34.6793887, 32.7346804, 30.7793058, 28.8176714, 26.8541835, 24.8932485, 22.9392726, 20.9966621, 19.0698236, 17.1631631, 15.2810872, 13.428002, 11.608314, 9.952898, 8.3369502, 6.76213535, 5.23011822, 3.74256353, 2.30113603, 0.907500469, -0.43667842, -1.72973589, -2.9700072, -4.15582761, -5.28553237, -6.35745674, -7.36993599, -8.32130536, -9.20990011, -10.0013518, -10.6620772, -11.2578605, -11.789157, -12.256422, -12.6601108, -13.0006787, -13.278581, -13.4942731, -13.6482101, -13.7408475, -13.7726405, -13.7440445, -13.6555147, -13.5075065, -13.3004751, -13.0348759, -12.4043752, -11.5692757, -10.6935675, -9.78766167, -8.86196948, -7.92690208, -6.9928707, -6.07028651, -5.16956072, -4.30110452, -3.4753291, -2.70264567, -1.9934654, -1.35819951, -0.807259183, -0.351055614, 0, 0.0799160617, -0.0124712986, -0.274033511, -0.701642006, -1.29216821, -2.04248356, -2.94945949, -4.00996742, -5.22087878, -6.579065, -8.08139752, -9.72474777, -11.5059872, -13.4219872, -15.4696192, -17.6457546, -20.1091727, -22.8798747, -25.5885119, -28.2666501, -30.9458555, -33.6576939, -36.4337314, -39.3055338, -42.3046673, -45.4626976, -48.8111908, -52.3817129, -56.2058297, -60.4266824, -64.9233509, -69.6939289, -74.7365097, -80.1101235, -85.7823954, -91.7047008, -97.8568522, -104.218662, -110.769943, -117.490507, -123.882493, -130.548767, -137.562598, -144.997258, -152.926016, -162.029022, -171.620947, -181.623343, -191.95776, -202.54575, -213.077823, -223.988952, -235.60093, -248.016692, -260.89234, -274.084597, -287.450184, -300.710732, -313.888756, -327.154144, -340.620839, -354.057548, -367.668083, -381.30143, -394.205347, -405.903564, -416.059574, -424.233344, -429.47238, -431.114062, -433.151868, -429.550764, -418.036406, -396.334453, -362.17056, -263.305271, -159.920637, -62.2335944, 19.5389208, 75.1799722, 57.6226856, 18.0666877, -37.9121271, -92.4075556, -144.724451, -199.305295, -255.19806, -311.450716, -367.111237, -421.227594, -533.986184, -592.745779, -616.535913, -624.386117, -635.325925, -668.38487, -742.592486, -785.936348, -786.09657}
    };
    
    const double b4tab[98][200] = {{-186.925396, -185.428619, -183.87549, -182.268397, -180.609732, -178.901884, -177.147243, -175.348199, -173.507143, -171.626463, -169.70855, -167.755794, -165.770585, -163.755313, -161.712368, -159.64414, -157.553018, -155.441394, -153.311656, -151.166195, -149.0074, -146.837662, -144.659371, -142.474916, -140.286688, -138.097077, -135.908472, -133.723263, -131.543841, -129.372595, -127.211916, -125.064193, -122.931816, -120.817176, -118.813762, -116.876499, -114.959215, -113.061343, -111.182317, -109.321571, -107.478538, -105.652652, -103.843346, -102.050055, -100.27221, -98.5092469, -96.7605982, -95.0256978, -93.3039792, -91.594876, -89.8978218, -88.2296282, -86.571974, -84.9239161, -83.2845113, -81.6528165, -80.0278885, -78.4087842, -76.7945604, -75.184274, -73.5769818, -71.9717407, -70.3676076, -68.7636393, -67.1588926, -65.5524245, -63.9432917, -62.3335941, -60.7253585, -59.1113659, -57.4904757, -55.8615477, -54.2234413, -52.5750162, -50.9151319, -49.2426481, -47.5564244, -45.8553203, -44.1381954, -42.4039093, -40.6513216, -38.8792919, -37.0866797, -35.2723448, -33.2511509, -31.1178257, -28.9693081, -26.8104287, -24.6460181, -22.480907, -20.319926, -18.1679055, -16.0296763, -13.910069, -11.8139141, -9.74604227, -7.71128409, -5.71447017, -3.76043111, -1.85399752, 0, 0.789810703, 1.64166074, 2.55162082, 3.51576165, 4.53015392, 5.59086834, 6.69397562, 7.83554645, 9.01165154, 10.2183616, 11.4517473, 12.7078794, 13.9828286, 15.2726656, 16.573461, 17.8812856, 19.1938099, 20.5073332, 21.8143075, 23.1111464, 24.3942633, 25.6600718, 26.9049855, 28.1254177, 29.3177821, 30.4784922, 31.6039615, 32.6906035, 33.7348317, 34.8648818, 35.8970104, 36.7880844, 37.4949707, 38.0937281, 38.4816274, 38.5241553, 38.142421, 37.2575338, 35.7906031, 33.6627382, 31.7098614, 28.6598477, 24.2548214, 18.2369067, 10.3482278, 0.852328441, -11.1604416, -26.078313, -44.2895165, -66.1822827, -90.1603839, -121.021959, -160.060796, -208.94172, -270.087083, -345.579122, -437.500075, -543.112905, -670.414413, -832.658788, -1034.17891, -1288.75743, -1596.96821, -1970.51462, -2439.95837, -3027.20486, -3731.23357, -4568.00615, -5637.54608, -6992.32673, -8265.8317, -9993.98406, -12374.2921, -15604.2641, -19881.4083, -30975.1018, -42118.0169, -52114.6946, -59769.676, -63887.5022, -58174.8196, -49932.6602, -40150.377, -32679.1088, -27375.4184, -22903.4579, -19179.4107, -16119.4601, -13639.7894, -11656.5821, -9802.55027, -8428.98912, -7442.177, -6748.39222, -6253.91311, -5865.018, -5487.98523, -5272.30142, -5113.43163},
        {-168.555271, -167.062415, -165.523968, -163.942134, -162.319116, -160.657119, -158.958346, -157.225003, -155.459291, -153.663416, -151.839581, -149.98999, -148.116848, -146.222358, -144.308723, -142.378149, -140.432839, -138.474996, -136.506826, -134.530531, -132.548316, -130.562384, -128.57494, -126.588187, -124.60433, -122.625573, -120.654118, -118.692171, -116.741935, -114.805614, -112.885412, -110.983533, -109.102182, -107.24356, -105.52924, -103.899231, -102.292108, -100.706202, -99.1398436, -97.5913624, -96.0590889, -94.5413534, -93.0364862, -91.5428176, -90.0586778, -88.5823972, -87.1123061, -85.6467346, -84.1840132, -82.722472, -81.2604415, -79.7134777, -78.1644795, -76.6135715, -75.0608783, -73.5065246, -71.9506351, -70.3933344, -68.8347472, -67.2749981, -65.7142119, -64.1525131, -62.5900265, -61.0268766, -59.4631883, -57.899086, -56.3346945, -54.7830113, -53.2567248, -51.7294103, -50.2003582, -48.6688584, -47.1342014, -45.5956771, -44.0525758, -42.5041876, -40.9498027, -39.3887113, -37.8202035, -36.2435696, -34.6580996, -33.0630838, -31.4578124, -29.8415754, -28.1133874, -26.3247857, -24.5285125, -22.7271121, -20.923129, -19.1191073, -17.3175915, -15.5211257, -13.7322544, -11.9535218, -10.1874723, -8.43665009, -6.70359954, -4.99086494, -3.30099062, -1.63652087, 0, 0.788871434, 1.66351404, 2.61841948, 3.6480794, 4.74698546, 5.9096293, 7.13050258, 8.40409695, 9.72490407, 11.0874156, 12.4861231, 13.9155184, 15.370093, 16.8443387, 18.332747, 19.8298096, 21.3822195, 22.9919258, 24.5353455, 26.0181564, 27.446036, 28.824662, 30.1597122, 31.4568641, 32.7217954, 33.9601838, 35.1777069, 36.3800425, 37.572868, 38.9324269, 40.2312905, 41.4239667, 42.4649636, 43.4260439, 44.2030881, 44.6607091, 44.7182384, 44.2950075, 43.3103479, 41.6835914, 40.3986326, 37.9862424, 34.1574683, 28.6233578, 21.0949583, 11.1902133, -1.26344967, -16.5317072, -34.8802358, -56.574712, -78.9590514, -108.791732, -148.859641, -201.440001, -267.769481, -349.551935, -448.49122, -560.290616, -694.018322, -865.288623, -1078.25732, -1349.9475, -1682.70625, -2090.42449, -2606.55368, -3255.56664, -4036.14928, -4966.08894, -6157.72514, -7669.91336, -9249.63177, -11315.5943, -14034.6138, -17573.5033, -22099.0759, -32687.204, -43368.3763, -53082.1407, -60768.0454, -65365.6386, -61514.4847, -55320.7711, -47566.8962, -41028.7966, -35699.7615, -30742.4393, -26191.6512, -22082.2187, -18448.963, -15326.7055, -13453.3667, -11842.239, -10495.024, -9413.42304, -8599.13781, -8053.86979, -7779.32052, -7696.6568, -7916.77787},
        {-153.482925, -151.757938, -150.023487, -148.280448, -146.529697, -144.772112, -143.008569, -141.239946, -139.467118, -137.690963, -135.912358, -134.13218, -132.351305, -130.57061, -128.790972, -127.013267, -125.238374, -123.467168, -121.700526, -119.939325, -118.184442, -116.436754, -114.697138, -112.96647, -111.245627, -109.535486, -107.836924, -106.150818, -104.478045, -102.819481, -101.176003, -99.5484876, -97.9378125, -96.344854, -94.8511128, -93.4154581, -91.9957898, -90.5903684, -89.1974544, -87.8153082, -86.4421901, -85.0763608, -83.7160804, -82.3596096, -81.0052087, -79.6511382, -78.2956584, -76.9370299, -75.573513, -74.2033682, -72.8248559, -71.2943193, -69.7550125, -68.2082726, -66.6554363, -65.0978406, -63.5368224, -61.9737186, -60.4098662, -58.8466021, -57.2852631, -55.7271863, -54.1737084, -52.6261665, -51.0858975, -49.5542382, -48.0325256, -46.5612748, -45.1800602, -43.8094166, -42.4481413, -41.0950315, -39.7488841, -38.4084965, -37.0726657, -35.7401889, -34.4098633, -33.080486, -31.7508542, -30.419765, -29.0860156, -27.7484031, -26.4057248, -25.0567776, -23.6417038, -22.1898618, -20.731315, -19.2667639, -17.7969092, -16.3224516, -14.8440916, -13.3625299, -11.8784671, -10.3926038, -8.90564066, -7.41827831, -5.93121735, -4.44515842, -2.96080214, -1.47884912, 0, 0.987987976, 2.05400637, 3.1931904, 4.40067528, 5.67159622, 7.00108844, 8.38428716, 9.81632759, 11.2923449, 12.8074744, 14.3568513, 15.9356107, 17.5388879, 19.1618182, 20.7995366, 22.4471785, 24.1025717, 25.7612362, 27.4122165, 29.0512249, 30.6739734, 32.2761745, 33.8535403, 35.401783, 36.9166148, 38.3937481, 39.828895, 41.2177677, 42.5560784, 43.9278284, 45.2080683, 46.3660237, 47.3709201, 48.3208169, 49.1205228, 49.6404906, 49.8112957, 49.5635135, 48.8277193, 47.5344884, 46.9169366, 45.2066736, 42.0794299, 37.2109362, 30.2769229, 21.1169577, 9.20197479, -5.83325462, -24.3539593, -46.7253679, -70.5469759, -102.330086, -144.061192, -197.924616, -266.508582, -352.219968, -457.465655, -580.412688, -728.67138, -914.47732, -1137.06446, -1420.17539, -1766.99466, -2194.62664, -2743.7621, -3444.26525, -4295.58415, -5319.69737, -6650.10934, -8357.23917, -10165.8875, -12545.2639, -15685.4247, -19776.4261, -25008.3244, -37214.9469, -49531.636, -60737.5053, -69611.6683, -74933.2386, -70499.0452, -63392.009, -54526.5083, -47235.1107, -41474.9945, -36195.7591, -31404.9619, -27110.16, -23318.9108, -20038.7716, -18254.2201, -16392.3622, -14575.8259, -12927.2389, -11569.2291, -10624.4245, -10215.4528, -9851.4599, -9878.1402},
        {-140.878578, -138.916758, -136.99567, -135.113903, -133.270049, -131.462697, -129.690437, -127.95186, -126.245556, -124.570115, -122.924127, -121.306182, -119.714871, -118.148783, -116.606509, -115.08664, -113.587764, -112.108473, -110.647356, -109.203004, -107.774007, -106.358955, -104.956439, -103.565047, -102.183371, -100.810001, -99.4435272, -98.0825392, -96.7256274, -95.371382, -94.0183933, -92.6652513, -91.3105463, -89.9528685, -88.5271908, -87.0652502, -85.5995481, -84.1307392, -82.659478, -81.1864193, -79.7122178, -78.2375281, -76.7630049, -75.2893028, -73.8170766, -72.3469809, -70.8796703, -69.4157996, -67.9560234, -66.5009963, -65.0513731, -63.6447044, -62.2439489, -60.8489617, -59.4595975, -58.0757112, -56.6971575, -55.3237914, -53.9554677, -52.5920412, -51.2333668, -49.8792992, -48.5296934, -47.1844042, -45.8432863, -44.5061947, -43.1729842, -41.8405662, -40.5059227, -39.1749793, -37.8477815, -36.524375, -35.2048054, -33.8891185, -32.5773598, -31.2695751, -29.9658099, -28.6661099, -27.3705207, -26.0790881, -24.7918576, -23.5088749, -22.2301857, -20.9558356, -19.731786, -18.5341591, -17.3385247, -16.1434386, -14.9474562, -13.7491333, -12.5470253, -11.339688, -10.1256767, -8.90354729, -7.67185518, -6.42915601, -5.17400538, -3.90495887, -2.62057208, -1.31940059, 0, 1.48262623, 2.97832256, 4.48643111, 6.00629395, 7.53725318, 9.07865088, 10.6298292, 12.1901301, 13.7588958, 15.3354683, 16.9191897, 18.5094022, 20.1054478, 21.7066686, 23.3124067, 24.9220042, 26.4881129, 28.0034048, 29.5728311, 31.1857286, 32.8314345, 34.4992858, 36.1786195, 37.8587727, 39.5290824, 41.1788856, 42.7975193, 44.3743207, 45.8986267, 47.4134394, 48.8347537, 50.135807, 51.2898368, 52.3936627, 53.3587313, 54.0635337, 54.4442327, 54.4369911, 53.9779719, 53.0033379, 52.8404883, 51.6109299, 48.978627, 44.6075439, 38.1616452, 30.8355043, 20.3798242, 6.07591719, -12.7949047, -36.9513292, -65.7540648, -102.939531, -147.054274, -199.380359, -266.784879, -353.627363, -464.267341, -601.536259, -771.669022, -982.567531, -1227.89859, -1540.17385, -1931.05681, -2421.47106, -3048.03093, -3840.14844, -4780.5206, -5886.44816, -7346.52013, -9252.43521, -11333.1233, -14098.8623, -17766.925, -22554.5836, -28679.1108, -42909.4015, -57273.2, -70349.8734, -80718.7884, -86959.3119, -81826.6949, -73607.1641, -63376.1363, -55135.0026, -48800.9814, -43072.1348, -37926.6389, -33342.6699, -29298.4036, -25772.0163, -22796.3174, -20305.0625, -18238.898, -16538.4706, -15144.4268, -13997.4132, -13038.0762, -12400.6408, -12000.1893},
        {-132.327653, -130.565925, -128.837385, -127.140774, -125.474831, -123.838296, -122.229907, -120.648405, -119.092529, -117.561019, -116.052613, -114.566052, -113.100075, -111.653421, -110.22483, -108.813042, -107.416795, -106.03483, -104.665886, -103.308701, -101.962017, -100.624572, -99.2951055, -97.9723573, -96.6550669, -95.3419737, -94.0318172, -92.7233369, -91.4152722, -90.1063627, -88.7953478, -87.480967, -86.1619598, -84.8370657, -83.424717, -81.9654961, -80.5009502, -79.0324249, -77.5612659, -76.0888188, -74.6164292, -73.1454426, -71.6772047, -70.213061, -68.7543572, -67.3024389, -65.8586517, -64.4243411, -63.0008529, -61.5895325, -60.1917256, -58.9283391, -57.6785654, -56.4411582, -55.2148715, -53.9984587, -52.7906739, -51.5902705, -50.3960025, -49.2066236, -48.0208874, -46.8375479, -45.6553586, -44.4730734, -43.289446, -42.1032301, -40.9131795, -39.6823008, -38.3744589, -37.0621333, -35.7463952, -34.4283155, -33.1089654, -31.789416, -30.4707384, -29.1540036, -27.8402828, -26.530647, -25.2261674, -23.927915, -22.6369608, -21.3543761, -20.0812319, -18.8185992, -17.6792101, -16.6059566, -15.5403885, -14.4799534, -13.4220988, -12.3642721, -11.3039208, -10.2384924, -9.16543431, -8.08219408, -6.98621916, -5.87495703, -4.74585518, -3.59636107, -2.42392218, -1.225986, 0, 1.85063991, 3.65235287, 5.40896889, 7.12431796, 8.80223008, 10.4465353, 12.0610635, 13.6496447, 15.216109, 16.7642863, 18.2980067, 19.8211001, 21.3373965, 22.850726, 24.3649184, 25.8838039, 27.3283841, 28.6906565, 30.1618007, 31.7278977, 33.3750287, 35.0892747, 36.8567169, 38.6634362, 40.4955139, 42.3390309, 44.1800683, 46.0047073, 47.7990289, 49.6133342, 51.3459369, 52.9636521, 54.4332949, 55.8813098, 57.1946971, 58.2178895, 58.8698131, 59.0693943, 58.7355592, 57.7872343, 57.3578928, 55.7822691, 52.7416607, 47.9173652, 40.9906801, 34.3979227, 24.3766157, 9.91930153, -9.98147735, -36.3331784, -69.4929654, -111.913393, -160.423571, -215.952054, -287.797103, -381.49915, -502.59863, -654.932634, -846.13206, -1085.686, -1367.98541, -1728.30247, -2179.32781, -2744.1963, -3463.74006, -4370.66787, -5439.55038, -6688.02352, -8346.89649, -10526.3956, -12804.5624, -15905.6826, -20124.2004, -25754.5602, -33091.2062, -50791.5592, -68696.3425, -85009.2563, -97934.0005, -105674.275, -99009.8931, -88517.6996, -75583.0609, -65509.1954, -58130.6526, -51649.5432, -45982.0385, -41044.3099, -36752.5288, -33022.8663, -29010.161, -25879.2976, -23439.7331, -21500.925, -19872.3305, -18363.4068, -16783.6111, -15683.0225, -14926.393},
        {-127.695887, -126.450394, -125.174478, -123.869678, -122.537533, -121.179584, -119.797371, -118.392433, -116.966311, -115.520545, -114.056674, -112.576239, -111.08078, -109.571836, -108.050947, -106.519654, -104.979496, -103.432014, -101.878747, -100.321235, -98.7610189, -97.1996377, -95.6386318, -94.079541, -92.5239054, -90.9732648, -89.4291594, -87.8931291, -86.3667137, -84.8514534, -83.348888, -81.8605575, -80.3880019, -78.9327612, -77.581541, -76.2915072, -75.0188018, -73.762201, -72.5204811, -71.2924181, -70.0767882, -68.8723677, -67.6779327, -66.4922593, -65.3141238, -64.1423024, -62.9755711, -61.8127062, -60.6524839, -59.4936803, -58.3350716, -57.1093067, -55.8827225, -54.6555287, -53.4279351, -52.2001513, -50.9723871, -49.7448521, -48.517756, -47.2913084, -46.0657192, -44.8411979, -43.6179543, -42.396198, -41.1761387, -39.9579862, -38.74195, -37.5287503, -36.3190948, -35.1121406, -33.9080644, -32.7070428, -31.5092523, -30.3148696, -29.1240714, -27.9370341, -26.7539344, -25.574949, -24.4002543, -23.2300271, -22.0644439, -20.9036814, -19.7479161, -18.5973246, -17.5243323, -16.4914718, -15.4604072, -14.4289707, -13.3949941, -12.3563095, -11.3107488, -10.2561441, -9.19032742, -8.11113067, -7.01638587, -5.90392503, -4.77158014, -3.61718319, -2.43856619, -1.23356113, 0, 1.81096445, 3.57151947, 5.28659702, 6.96112905, 8.60004751, 10.2082844, 11.7907716, 13.3524411, 14.8982249, 16.4330549, 17.961863, 19.4895814, 21.0211418, 22.5614762, 24.1155167, 25.6881951, 27.1785391, 28.5763514, 30.1261095, 31.8100517, 33.610416, 35.5094407, 37.4893639, 39.5324237, 41.6208584, 43.7369061, 45.862805, 47.9807932, 50.073109, 52.188949, 54.2190414, 56.1255368, 57.8705859, 59.6606566, 61.335741, 62.6706795, 63.5543275, 63.8755403, 63.5231733, 62.386082, 61.2957633, 58.9115399, 54.9378374, 49.0790812, 41.0396967, 33.5662299, 22.5604557, 6.9662694, -14.2724336, -42.211758, -76.3777648, -121.226655, -175.411611, -240.628437, -324.784963, -432.999949, -570.392152, -737.118219, -944.386772, -1208.81965, -1531.08872, -1942.85475, -2444.3232, -3059.25793, -3851.34115, -4865.93191, -6091.71584, -7559.35889, -9507.32939, -12056.5509, -14332.6239, -17604.9217, -22338.9025, -29000.0249, -38053.7473, -61802.9619, -85916.3349, -107899.966, -125259.956, -135502.404, -125711.215, -110762.814, -92629.9572, -79114.1999, -69881.0637, -62167.403, -55760.1725, -50446.3271, -46012.8214, -42246.6103, -38232.4442, -34627.6603, -31402.1374, -28525.7545, -25968.3903, -23699.9237, -21690.2337, -19790.0046, -18741.5427},
        {-123.130142, -121.902399, -120.646649, -119.36435, -118.056962, -116.725941, -115.372746, -113.998836, -112.605669, -111.194703, -109.767396, -108.325207, -106.869594, -105.402015, -103.923928, -102.436792, -100.942065, -99.4412048, -97.9356704, -96.4269199, -94.9164114, -93.4056032, -91.8959536, -90.3889209, -88.8859634, -87.3885392, -85.8981068, -84.4161243, -82.94405, -81.4833423, -80.0354593, -78.6018594, -77.1840008, -75.7833417, -74.4840079, -73.2443851, -72.0218655, -70.8152247, -69.6232382, -68.4446816, -67.2783304, -66.1229603, -64.9773467, -63.8402653, -62.7104916, -61.586801, -60.4679693, -59.352772, -58.2399845, -57.1283826, -56.0167417, -54.8352678, -53.6527925, -52.4695778, -51.2858858, -50.1019784, -48.9181177, -47.7345656, -46.5515842, -45.3694354, -44.1883814, -43.0086841, -41.8306055, -40.6544076, -39.4803525, -38.3087021, -37.1397185, -35.9765017, -34.8220837, -33.6708732, -32.5229483, -31.3783868, -30.237267, -29.0996668, -27.9656642, -26.8353374, -25.7087642, -24.5860227, -23.467191, -22.3523472, -21.2415691, -20.1349349, -19.0325226, -17.9344103, -16.920217, -15.9485772, -14.9771693, -14.00349, -13.0250363, -12.0393047, -11.0437922, -10.0359953, -9.01341106, -7.97353608, -6.91386715, -5.83190106, -4.72513457, -3.59106445, -2.42718747, -1.2310004, 0, 1.67556303, 3.33749321, 4.98859319, 6.63166561, 8.26951311, 9.90493831, 11.5407439, 13.1797324, 14.8247066, 16.478469, 18.1438223, 19.8235691, 21.5205122, 23.237454, 24.9771973, 26.7425447, 28.4867855, 30.2056492, 32.0139325, 33.9038282, 35.8675288, 37.8972271, 39.9851156, 42.1233872, 44.3042343, 46.5198497, 48.7624261, 51.024156, 53.2972323, 55.6916812, 58.0386561, 60.2949992, 62.4175532, 64.6228629, 66.7379196, 68.5204608, 69.8494181, 70.6037236, 70.6623089, 69.9041059, 69.1010294, 66.9672508, 63.2069879, 57.5244585, 49.6238802, 42.5598703, 31.8486472, 16.3568291, -5.04896624, -33.5021208, -68.107809, -114.506541, -172.012252, -242.758065, -334.632941, -452.941587, -602.988711, -783.508256, -1007.86905, -1296.60802, -1652.0642, -2108.42564, -2663.274, -3343.07014, -4223.20921, -5356.62473, -6730.45394, -8380.57208, -10584.1073, -13481.7215, -15639.0985, -19014.1825, -24292.82, -32160.8574, -43304.1414, -74407.5516, -106158.143, -135242.004, -158345.223, -172153.887, -159586.262, -140274.808, -116806.111, -99439.3435, -87724.0228, -78012.7564, -70014.908, -63439.8418, -57996.9218, -53395.512, -48257.7211, -43686.4284, -39628.3655, -36030.264, -32838.8553, -30000.8712, -27463.0431, -24884.1289, -23353.191},
        {-117.374593, -115.594778, -113.870151, -112.198489, -110.577567, -109.00516, -107.479045, -105.996996, -104.55679, -103.156202, -101.793008, -100.464984, -99.1699045, -97.9055461, -96.6696842, -95.4600945, -94.2745527, -93.1108343, -91.9667152, -90.8399709, -89.728377, -88.6297093, -87.5417435, -86.4622551, -85.3890199, -84.3198135, -83.2524116, -82.1845898, -81.1141238, -80.0387893, -78.9563619, -77.8646173, -76.7613312, -75.6442791, -74.3727433, -73.0166592, -71.6476269, -70.2679163, -68.8797975, -67.4855406, -66.0874157, -64.6876927, -63.2886417, -61.8925327, -60.5016359, -59.1182212, -57.7445586, -56.3829184, -55.0355703, -53.7047846, -52.3928313, -51.2904778, -50.2074105, -49.1418132, -48.0918696, -47.0557637, -46.031679, -45.0177996, -44.012309, -43.0133912, -42.01923, -41.028009, -40.0379121, -39.0471231, -38.0538258, -37.056204, -36.0524414, -34.9917516, -33.8245238, -32.6499393, -31.4693566, -30.2841338, -29.0956295, -27.9052018, -26.7142091, -25.5240098, -24.3359622, -23.1514247, -21.9717555, -20.798313, -19.6324555, -18.4755415, -17.3289291, -16.1939768, -15.2378751, -14.3755782, -13.5200471, -12.6672584, -11.8131891, -10.9538157, -10.0851152, -9.2030642, -8.3036395, -7.38281788, -6.43657607, -5.46089085, -4.45173896, -3.40509718, -2.31694225, -1.18325094, 0, 1.60046196, 3.236288, 4.90655574, 6.6103428, 8.34672681, 10.1147854, 11.9135961, 13.7422367, 15.5997847, 17.4853177, 19.3979135, 21.3366495, 23.3006034, 25.2888528, 27.3004754, 29.3345488, 31.4612275, 33.689743, 35.8584037, 37.9815179, 40.0733941, 42.1483406, 44.2206659, 46.3046783, 48.4146861, 50.5649979, 52.7699221, 55.0437669, 57.4008408, 60.0624164, 62.7599511, 65.4456638, 68.0717738, 70.7840318, 73.4378912, 75.8371966, 77.8761078, 79.4487842, 80.4493856, 80.7720714, 81.4564459, 80.9026105, 78.780616, 74.7605135, 68.5123537, 63.5450777, 54.7301237, 40.7778199, 20.3984948, -7.69752335, -43.0893195, -90.8678692, -148.659757, -218.361326, -310.586342, -432.034627, -589.406002, -784.216483, -1029.52783, -1344.05688, -1724.6412, -2216.88704, -2830.01296, -3596.00521, -4585.42864, -5851.13989, -7356.36497, -9130.72295, -11532.4785, -14733.9959, -16755.975, -20253.4694, -26143.6501, -35343.6882, -48770.7546, -87445.2835, -127155.367, -163792.36, -193247.618, -211412.496, -197395.306, -175059.142, -147488.093, -126729.833, -112320.366, -100060.863, -89674.0687, -80882.7251, -73409.5754, -66977.3625, -59171.0058, -52999.2412, -48105.554, -44133.4293, -40726.3523, -37527.8082, -34181.2822, -31072.1228, -28733.4857},
        {-112.925689, -111.070608, -109.290245, -107.581475, -105.941169, -104.366202, -102.853447, -101.399776, -100.002063, -98.6571806, -97.3620024, -96.1134014, -94.9082508, -93.7434236, -92.615793, -91.5222322, -90.4596143, -89.4248125, -88.4146998, -87.4261494, -86.4560345, -85.5012282, -84.5586037, -83.625034, -82.6973924, -81.7725519, -80.8473858, -79.9187671, -78.983569, -78.0386646, -77.0809271, -76.1072296, -75.1144453, -74.0994473, -72.8663967, -71.512576, -70.1375842, -68.7445483, -67.3365953, -65.9168522, -64.488446, -63.0545038, -61.6181525, -60.1825192, -58.7507309, -57.3259147, -55.9111975, -54.5097063, -53.1245682, -51.7589103, -50.4158594, -49.3455145, -48.2986769, -47.2731197, -46.2666161, -45.2769393, -44.3018624, -43.3391585, -42.3866008, -41.4419625, -40.5030168, -39.5675367, -38.6332954, -37.6980661, -36.759622, -35.8157361, -34.8641817, -33.8429186, -32.691343, -31.5305882, -30.3623048, -29.1881435, -28.0097548, -26.8287893, -25.6468977, -24.4657305, -23.2869383, -22.1121718, -20.9430815, -19.7813181, -18.6285321, -17.4863741, -16.3564948, -15.2405447, -14.3352563, -13.5406358, -12.7543452, -11.9717041, -11.1880324, -10.3986499, -9.59887642, -8.78403183, -7.94943592, -7.09040852, -6.20226945, -5.28033853, -4.31993559, -3.31638044, -2.26499291, -1.16109282, 0, 1.56417819, 3.1825465, 4.85340408, 6.5750501, 8.34578373, 10.1639041, 12.0277105, 13.9355019, 15.8855776, 17.8762368, 19.9057785, 21.9725021, 24.0747065, 26.2106911, 28.3787549, 30.5771972, 32.9182781, 35.4165771, 37.8126241, 40.1291385, 42.3888398, 44.6144473, 46.8286805, 49.0542586, 51.3139013, 53.6303278, 56.0262575, 58.5244098, 61.1475042, 64.1462268, 67.2317424, 70.3583805, 73.4804703, 76.7425193, 80.0037679, 83.0727423, 85.8467186, 88.2229728, 90.098781, 91.3714191, 93.3550194, 94.0987846, 93.2227796, 90.3470695, 85.0917193, 80.6624915, 72.1973292, 58.4198731, 38.0537638, 9.82264203, -25.2740017, -73.569971, -132.363147, -203.836489, -300.146652, -428.972305, -597.992119, -810.592248, -1079.71945, -1423.00322, -1833.15247, -2364.64111, -3029.73681, -3865.31133, -4949.92668, -6343.00838, -7999.31079, -9951.12246, -12614.5261, -16188.5088, -18212.0639, -21953.402, -28532.2769, -39068.4428, -54681.6536, -100061.457, -146865.366, -190320.684, -225654.719, -248094.774, -233673.923, -209609.86, -179355.992, -156229.157, -139778.485, -125443.344, -112978.356, -102138.138, -92677.3099, -84350.4919, -74365.0189, -66557.5268, -60421.5448, -55450.6022, -51138.2282, -46977.9521, -42463.3031, -38369.1972, -34919.4995},
        {-108.709406, -107.238695, -105.805603, -104.408507, -103.04578, -101.715801, -100.416945, -99.1475868, -97.9061037, -96.6908711, -95.5002652, -94.3326618, -93.1864369, -92.0599664, -90.9516264, -89.8597928, -88.7828416, -87.7191486, -86.66709, -85.6250415, -84.5913793, -83.5644792, -82.5427172, -81.5244694, -80.5081115, -79.4920197, -78.4745698, -77.4541379, -76.4290998, -75.3978316, -74.3587092, -73.3101085, -72.2504056, -71.1779764, -69.9816561, -68.7168951, -67.4404603, -66.1542826, -64.8602927, -63.5604213, -62.2565993, -60.9507573, -59.6448262, -58.3407368, -57.0404197, -55.7458058, -54.4588258, -53.1814105, -51.9154907, -50.6629972, -49.4258606, -48.3632378, -47.3164251, -46.2839451, -45.2643201, -44.2560725, -43.2577248, -42.2677994, -41.2848187, -40.307305, -39.3337809, -38.3627688, -37.392791, -36.4223699, -35.4500281, -34.4742879, -33.4936717, -32.4623916, -31.3357236, -30.2035767, -29.067346, -27.9284262, -26.7882123, -25.6480992, -24.5094816, -23.3737546, -22.2423129, -21.1165514, -19.997865, -18.8876487, -17.7872972, -16.6982054, -15.6217683, -14.5593807, -13.7038653, -12.9568766, -12.2177684, -11.4817233, -10.743924, -9.99955327, -9.24379365, -8.47182786, -7.6788386, -6.86000852, -6.0105203, -5.12555663, -4.20030017, -3.2299336, -2.2096396, -1.13460084, 0, 1.71413763, 3.44428947, 5.1926992, 6.96161049, 8.75326704, 10.5699125, 12.4137906, 14.2871451, 16.1922195, 18.1312575, 20.106503, 22.1201994, 24.1745906, 26.2719202, 28.4144319, 30.6043694, 32.9017118, 35.3169507, 37.7217377, 40.1306884, 42.5584183, 45.0195429, 47.5286779, 50.1004388, 52.7494413, 55.4903008, 58.337633, 61.3060534, 64.4101776, 67.8309051, 71.3555968, 74.948983, 78.5757942, 82.4457379, 86.401056, 90.2186628, 93.7897956, 97.0056917, 99.7575881, 101.936722, 105.203833, 107.142112, 107.296589, 105.212294, 100.434258, 95.2049946, 85.6976792, 70.7829718, 49.3315318, 20.2140188, -14.3057713, -62.8974823, -122.94115, -197.160273, -299.18792, -437.758986, -621.608364, -856.685203, -1155.14327, -1532.17532, -1975.71716, -2548.62429, -3260.70495, -4153.63194, -5323.07084, -6839.90366, -8661.30943, -10828.6916, -13800.3646, -17798.8144, -20063.4255, -24258.7241, -31639.8082, -43461.7759, -60979.7253, -111110.822, -163032.58, -211584.58, -251606.403, -277937.629, -265074.595, -241762.291, -211558.831, -187940.184, -170538.646, -154854.078, -140725.4, -127991.531, -116491.391, -106063.901, -94085.0348, -84501.0421, -76737.9308, -70221.7093, -64378.3862, -58633.9698, -52414.4686, -46877.6453, -41886.5261},
        {-104.863058, -103.921983, -102.96067, -101.979918, -100.980528, -99.9632975, -98.9290265, -97.8785141, -96.8125593, -95.7319614, -94.6375197, -93.5300333, -92.4103014, -91.2791232, -90.137298, -88.9856249, -87.8249032, -86.6559319, -85.4795105, -84.2964379, -83.1075136, -81.9135366, -80.7153061, -79.5136214, -78.3092817, -77.1030861, -75.8958339, -74.6883243, -73.4813565, -72.2757296, -71.072243, -69.8716957, -68.674887, -67.4826161, -66.3243922, -65.1860557, -64.0531018, -62.9253979, -61.8028116, -60.6852105, -59.5724619, -58.4644335, -57.3609927, -56.262007, -55.1673439, -54.0768711, -52.9904558, -51.9079658, -50.8292684, -49.7542312, -48.6827216, -47.619479, -46.5593934, -45.5022268, -44.4477411, -43.3956981, -42.3458599, -41.2979883, -40.2518452, -39.2071925, -38.1637921, -37.1214059, -36.0797958, -35.0387236, -33.9979514, -32.957241, -31.9163543, -30.8574645, -29.7631671, -28.6694993, -27.5773633, -26.4876612, -25.4012951, -24.3191671, -23.2421793, -22.1712338, -21.1072326, -20.051078, -19.0036721, -17.9659168, -16.9387144, -15.9229669, -14.9195764, -13.929445, -13.1304803, -12.4313586, -11.7386287, -11.0474485, -10.3529761, -9.65036938, -8.93478625, -8.20138466, -7.44532255, -6.66175788, -5.84584856, -4.99275256, -4.09762779, -3.15563222, -2.16192377, -1.11166038, 0, 2.01888409, 3.98302187, 5.90103286, 7.78153662, 9.63315269, 11.4645006, 13.2841999, 15.1008702, 16.9231309, 18.7596016, 20.6189019, 22.5096513, 24.4404693, 26.4199756, 28.4567895, 30.5595307, 32.6888088, 34.8463847, 37.149472, 39.5964022, 42.1855072, 44.9151186, 47.783568, 50.7891873, 53.9303081, 57.205262, 60.6123808, 64.1499962, 67.8164399, 71.6669644, 75.6221095, 79.6631307, 83.7712835, 88.2962374, 93.0350408, 97.6864983, 102.121341, 106.2103, 109.824106, 112.833492, 117.026067, 119.772285, 120.567835, 118.908406, 114.289689, 108.220989, 97.6809734, 81.6619281, 59.1561373, 29.1558861, -4.86614813, -53.8741087, -115.43529, -192.830187, -301.00374, -449.663792, -648.518186, -905.117144, -1231.8165, -1643.32585, -2125.48912, -2746.54868, -3510.2047, -4461.96591, -5718.18324, -7362.68989, -9359.78857, -11762.3231, -15061.4157, -19500.2736, -22178.3582, -26966.2756, -35195.4527, -48197.3163, -67303.2935, -120569.316, -175921.18, -228009.185, -271483.632, -300994.822, -290503.69, -269142.811, -240429.355, -217457.307, -199953.016, -183652.489, -168481.594, -154366.194, -141232.156, -129005.345, -114541.89, -103069.416, -93816.6309, -86012.246, -78884.9712, -71663.5166, -63576.5922, -56700.3034, -49806.6676},
        {-101.405685, -100.619739, -99.7983489, -98.9429703, -98.0550593, -97.1360719, -96.1874638, -95.2106911, -94.2072096, -93.1784752, -92.1259438, -91.0510713, -89.9553137, -88.8401267, -87.7069664, -86.5572885, -85.3925491, -84.214204, -83.023709, -81.8225202, -80.6120933, -79.3938844, -78.1693492, -76.9399438, -75.7071239, -74.4723455, -73.2370645, -72.0027368, -70.7708183, -69.5427649, -68.3200324, -67.1040768, -65.896354, -64.6983199, -63.5809826, -62.509559, -61.4484298, -60.396794, -59.3538504, -58.3187976, -57.2908346, -56.26916, -55.2529728, -54.2414716, -53.2338553, -52.2293228, -51.2270727, -50.2263038, -49.2262151, -48.2260052, -47.2248729, -46.1738813, -45.1214084, -44.0676966, -43.012988, -41.9575251, -40.90155, -39.845305, -38.7890325, -37.7329748, -36.677374, -35.6224726, -34.5685127, -33.5157367, -32.4643868, -31.4147054, -30.3669347, -29.3133519, -28.2464255, -27.1828251, -26.1233094, -25.0686369, -24.0195664, -22.9768565, -21.9412656, -20.9135526, -19.8944761, -18.8847946, -17.8852667, -16.8966512, -15.9197066, -14.9551916, -14.0038648, -13.0664848, -12.3318646, -11.7027801, -11.0797472, -10.4574217, -9.83045944, -9.19351643, -8.54124854, -7.86831167, -7.16936173, -6.43905462, -5.67204624, -4.8629925, -4.0065493, -3.09737255, -2.13011814, -1.09944199, 0, 2.30610506, 4.51135415, 6.62839952, 8.66989342, 10.6484881, 12.5768358, 14.4675888, 16.3333993, 18.1869196, 20.040802, 21.9076986, 23.8002618, 25.7311437, 27.7129967, 29.758473, 31.8802248, 33.9904844, 36.0875583, 38.4112392, 40.952661, 43.7029571, 46.6532613, 49.7947073, 53.1184288, 56.6155594, 60.2772328, 64.0945827, 68.0587427, 72.1608466, 76.3736826, 80.7134565, 85.1768054, 89.7603667, 94.9660129, 100.537764, 106.084909, 111.452516, 116.485651, 121.02938, 124.928771, 129.786185, 133.154565, 134.535158, 133.429213, 129.337976, 124.041613, 114.192726, 98.7228338, 76.5634561, 46.6461116, 12.7970689, -36.9293741, -99.8903621, -179.642598, -292.400214, -448.694414, -659.056405, -930.8336, -1278.46458, -1719.86116, -2242.32868, -2917.0839, -3744.47741, -4773.04207, -6132.95297, -7916.91008, -10088.624, -12707.3529, -16305.3163, -21147.2001, -24230.4197, -29589.5883, -38620.3701, -52718.4293, -73279.43, -129463.447, -187960.631, -243225.543, -289712.746, -321876.8, -312981.292, -292799.076, -264866.559, -242227.644, -224655.356, -207961.263, -192116.461, -177092.048, -162859.121, -149388.776, -131730.358, -118404.821, -108216.035, -99967.8716, -92464.1987, -84508.8867, -74905.8051, -67294.0623, -58480.3337},
        {-98.0372641, -97.3673233, -96.650595, -95.8890407, -95.0846216, -94.2392991, -93.3550346, -92.4337894, -91.4775248, -90.4882023, -89.4677831, -88.4182287, -87.3415003, -86.2395593, -85.1143671, -83.967885, -82.8020744, -81.6188966, -80.420313, -79.2082848, -77.9847736, -76.7517406, -75.5111471, -74.2649546, -73.0151244, -71.7636177, -70.5123961, -69.2634208, -68.0186531, -66.7800545, -65.5495863, -64.3292098, -63.1208864, -61.9265774, -60.8507795, -59.8420297, -58.8476326, -57.8662221, -56.8964321, -55.9368964, -54.986249, -54.0431237, -53.1061544, -52.173975, -51.2452193, -50.3185213, -49.3925149, -48.4658339, -47.5371121, -46.6049836, -45.6680821, -44.6310488, -43.5885479, -42.5412508, -41.4898292, -40.4349543, -39.3772977, -38.3175308, -37.2563251, -36.1943521, -35.1322832, -34.07079, -33.0105437, -31.952216, -30.8964783, -29.844002, -28.7954586, -27.7524503, -26.7165568, -25.6865302, -24.6629816, -23.6465221, -22.6377629, -21.637315, -20.6457897, -19.663798, -18.691951, -17.73086, -16.7811361, -15.8433903, -14.9182338, -14.0062777, -13.1081332, -12.2244114, -11.5549829, -10.9966379, -10.4437717, -9.89052914, -9.33105498, -8.75949401, -8.16999099, -7.55669071, -6.91373794, -6.23527747, -5.51545408, -4.74841255, -3.92829765, -3.04925416, -2.10542687, -1.09096056, 0, 2.58686782, 5.02954544, 7.3445123, 9.54824786, 11.6572316, 13.6879429, 15.6568613, 17.5804663, 19.4752372, 21.3576536, 23.2441948, 25.1513404, 27.0955699, 29.0933626, 31.161198, 33.3155556, 35.4267956, 37.4905234, 39.853726, 42.5015717, 45.4192287, 48.5918652, 52.0046494, 55.6427495, 59.4913336, 63.53557, 67.760627, 72.1516726, 76.6938751, 81.2877721, 86.0341937, 90.9436971, 96.0268397, 101.928925, 108.343138, 114.793397, 121.099835, 127.082585, 132.56178, 137.357555, 142.859044, 146.839857, 148.813148, 148.292074, 144.789788, 140.562053, 131.693765, 117.012428, 95.3455431, 65.5206151, 31.5496888, -19.2609372, -84.108006, -166.794585, -284.611655, -448.794397, -670.577995, -956.528817, -1323.61195, -1793.88574, -2356.89509, -3086.36171, -3979.93777, -5089.04199, -6556.19757, -8482.10926, -10829.0693, -13661.8977, -17552.927, -22787.412, -26347.98, -32330.6097, -42163.7728, -57275.9412, -79095.5868, -137367.44, -198124.649, -255716.622, -304492.766, -338802.488, -331631.954, -312935.973, -286219.628, -264309.821, -247021.748, -230286.51, -214113.507, -198512.138, -183491.803, -169061.9, -148806.938, -133909.925, -122843.988, -114082.254, -106097.848, -97363.8967, -86353.527, -78038.7532, -67422.3408},
        {-94.6155069, -94.0137302, -93.3595706, -92.6552533, -91.9030033, -91.1050458, -90.2636059, -89.3809086, -88.4591791, -87.5006425, -86.5075239, -85.4820483, -84.426441, -83.3429269, -82.2337313, -81.1010791, -79.9471955, -78.7743057, -77.5846347, -76.3804076, -75.1638495, -73.9371856, -72.7026408, -71.4624405, -70.2188095, -68.9739731, -67.7301564, -66.4895844, -65.2544823, -64.0270751, -62.809588, -61.604246, -60.4132744, -59.2388981, -58.2049725, -57.2503494, -56.3124184, -55.3894574, -54.4797446, -53.5815578, -52.6931749, -51.8128741, -50.9389332, -50.0696302, -49.2032431, -48.3380498, -47.4723283, -46.6043566, -45.7324127, -44.8547745, -43.96972, -42.9505939, -41.9233157, -40.8888717, -39.8482483, -38.8024315, -37.7524078, -36.6991635, -35.6436847, -34.5869578, -33.529969, -32.4737046, -31.419151, -30.3672943, -29.3191208, -28.2756169, -27.2377688, -26.2144201, -25.2142257, -24.2219669, -23.2381204, -22.2631633, -21.2975724, -20.3418246, -19.3963969, -18.4617662, -17.5384094, -16.6268034, -15.7274252, -14.8407516, -13.9672595, -13.107426, -12.2617278, -11.430642, -10.824822, -10.3352357, -9.85032478, -9.36374545, -8.86915403, -8.3602068, -7.83056003, -7.27387001, -6.68379303, -6.05398535, -5.37810327, -4.64980307, -3.86274102, -3.0105734, -2.08695651, -1.08554661, 0, 2.81302897, 5.44918581, 7.92805186, 10.2692085, 12.492237, 14.6167188, 16.6622353, 18.6483677, 20.5946975, 22.520806, 24.4462745, 26.3906844, 28.373617, 30.4146538, 32.533376, 34.749365, 36.9080039, 39.0039885, 41.4509202, 44.2310522, 47.3266374, 50.7199292, 54.3931806, 58.3286448, 62.508575, 66.9152242, 71.5308457, 76.3376925, 81.3180179, 86.3268592, 91.520331, 96.9188514, 102.542838, 109.143706, 116.376374, 123.700831, 130.918193, 137.829582, 144.236114, 149.938911, 156.174882, 160.872374, 163.55159, 163.732733, 160.936005, 158.005664, 150.306845, 136.528737, 115.36053, 85.491413, 50.9954344, -1.40456837, -68.8549905, -155.394957, -279.136248, -451.87231, -685.396588, -985.112659, -1370.65608, -1868.63317, -2470.80039, -3253.7957, -4214.16606, -5406.36744, -6982.97937, -9051.54108, -11571.4603, -14611.8717, -18785.0589, -24396.0698, -28489.2754, -35132.043, -45758.1162, -61801.2388, -84695.1544, -144334.223, -206576.419, -265740.33, -316144.546, -352107.657, -346669.154, -329612.791, -304376.769, -283461.081, -266716.971, -250219.938, -234010.012, -218127.224, -202611.606, -187503.189, -165819.874, -149817.964, -137854.413, -128286.175, -119470.205, -109763.458, -97522.887, -88445.2724, -76328.7246},
        {-90.998127, -90.4079557, -89.7674386, -89.0787322, -88.3439933, -87.5653785, -86.7450447, -85.8851484, -84.9878464, -84.0552954, -83.0896521, -82.0930732, -81.0677154, -80.0157354, -78.9392898, -77.8405355, -76.721629, -75.5847272, -74.4319866, -73.2655641, -72.0876162, -70.9002998, -69.7057715, -68.5061879, -67.3037059, -66.1004821, -64.8986732, -63.7004359, -62.507927, -61.323303, -60.1487208, -58.986337, -57.8383083, -56.7067914, -55.7147512, -54.8013993, -53.9044952, -53.0222752, -52.1529755, -51.2948324, -50.4460821, -49.6049608, -48.7697048, -47.9385504, -47.1097338, -46.2814912, -45.4520588, -44.619673, -43.78257, -42.9389859, -42.0871571, -41.092129, -40.0882159, -39.0765414, -38.0582291, -37.0344025, -36.0061853, -34.9747009, -33.9410729, -32.9064249, -31.8718805, -30.8385632, -29.8075965, -28.780104, -27.7572094, -26.7400361, -25.7297077, -24.7389215, -23.7800967, -22.8304872, -21.8904662, -20.960407, -20.0406829, -19.1316671, -18.233733, -17.3472539, -16.472603, -15.6101537, -14.7602791, -13.9233527, -13.0997476, -12.2898373, -11.4939948, -10.7125937, -10.1663688, -9.74087772, -9.3190291, -8.89404511, -8.45914787, -8.00755949, -7.53250211, -7.02719785, -6.48486883, -5.89873717, -5.26202499, -4.56795442, -3.80974759, -2.9806266, -2.07381359, -1.08253069, 0, 2.93644511, 5.68186537, 8.25769889, 10.6853838, 12.9863583, 15.1820605, 17.2939285, 19.3434004, 21.3519145, 23.3409087, 25.3318214, 27.3460905, 29.4051544, 31.530451, 33.7434185, 36.0654951, 38.3443707, 40.5766624, 43.17681, 46.1290199, 49.4174985, 53.0264521, 56.9400873, 61.1426103, 65.6182276, 70.3511456, 75.3255707, 80.5257092, 85.9357676, 91.40857, 97.1078786, 103.057314, 109.280498, 116.569086, 124.562681, 132.696078, 140.762488, 148.555122, 155.86719, 162.491902, 169.663938, 175.296332, 178.900269, 179.986934, 178.06751, 176.5958, 170.154716, 157.269789, 136.466547, 106.270522, 70.7380283, 16.1039622, -54.898084, -146.55252, -277.472179, -459.836723, -705.825816, -1019.49499, -1422.99443, -1947.33702, -2585.6566, -3418.79946, -4444.74257, -5721.42014, -7408.36067, -9618.4592, -12306.1326, -15543.189, -19982.5232, -25948.3345, -30612.5421, -37936.5916, -49335.8561, -66225.7088, -90021.5229, -150416.724, -213479.119, -273554.57, -324988.935, -362128.075, -358306.373, -342888.818, -319226.188, -299438.666, -283405.809, -267353.255, -251343.256, -235438.065, -219699.935, -204191.117, -182817.409, -166362.171, -153401.232, -142510.418, -132265.559, -121242.481, -108017.014, -98024.5162, -84895.5208},
        {-87.0428371, -86.3989955, -85.7163616, -84.9966013, -84.2413801, -83.4523638, -82.631218, -81.7796084, -80.8992008, -79.9916608, -79.0586541, -78.1018464, -77.1229033, -76.1234906, -75.1052739, -74.0699189, -73.0190914, -71.9544569, -70.8776812, -69.79043, -68.694369, -67.5911638, -66.4824801, -65.3699836, -64.2553401, -63.1402151, -62.0262744, -60.9151837, -59.8086086, -58.7082148, -57.6156681, -56.532634, -55.4607784, -54.4017668, -53.4513051, -52.5620608, -51.6855714, -50.8204507, -49.9653126, -49.1187711, -48.2794401, -47.4459334, -46.616865, -45.7908488, -44.9664986, -44.1424285, -43.3172522, -42.4895837, -41.658037, -40.8212258, -39.9777642, -39.0152665, -38.0457524, -37.070242, -36.0897555, -35.105313, -34.1179348, -33.1286409, -32.1384515, -31.1483868, -30.159467, -29.1727121, -28.1891424, -27.2097779, -26.235639, -25.2677456, -24.307118, -23.3656148, -22.4548344, -21.5534433, -20.6617593, -19.7800997, -18.9087822, -18.0481244, -17.1984436, -16.3600576, -15.5332838, -14.7184398, -13.9158431, -13.1258112, -12.3486618, -11.5847124, -10.8342805, -10.0976836, -9.60461017, -9.23586784, -8.86950732, -8.49840263, -8.11542778, -7.71345679, -7.28536368, -6.82402247, -6.32230718, -5.77309182, -5.16925041, -4.50365697, -3.76918553, -2.95871009, -2.06510468, -1.08124331, 0, 2.90897285, 5.6391742, 8.21213406, 10.6493824, 12.9724493, 15.2028647, 17.3621586, 19.471861, 21.553502, 23.6286115, 25.7187195, 27.845356, 30.0300511, 32.2943348, 34.6597369, 37.1477876, 39.6461574, 42.1572535, 45.0053832, 48.1833925, 51.6841272, 55.500433, 59.6251559, 64.0511416, 68.7712359, 73.7782847, 79.0651339, 84.6246292, 90.4496164, 96.450531, 102.732846, 109.314132, 116.211955, 124.163796, 132.827268, 141.668009, 150.487617, 159.087686, 167.269814, 174.835596, 183.25645, 190.155944, 195.00897, 197.29042, 196.475185, 196.555817, 191.360133, 179.23361, 158.521726, 127.569958, 90.3811935, 32.7288841, -43.0040548, -141.376082, -281.117631, -474.596205, -734.179312, -1062.58567, -1484.02448, -2033.23087, -2703.07575, -3580.7866, -4669.24757, -6030.60185, -7827.40377, -10176.1173, -13023.422, -16441.7637, -21126.1309, -27419.3666, -32676.0162, -40686.9586, -52829.4479, -70480.7379, -95018.0827, -155667.868, -218995.934, -279417.249, -331346.785, -369199.514, -366757.089, -352823.342, -330656.092, -311999.819, -296753.041, -281278.167, -265650.518, -249945.419, -234238.192, -218604.158, -199847.786, -183775.783, -169638.369, -156685.766, -144168.196, -131335.879, -117439.037, -106287.381, -92818.7653},
        {-81.9500307, -81.0372164, -80.1282803, -79.2230153, -78.3212142, -77.4226697, -76.5271747, -75.6345218, -74.7445039, -73.8569138, -72.9715442, -72.0881879, -71.2066377, -70.3266863, -69.4481266, -68.5707513, -67.6943532, -66.8187251, -65.9436598, -65.06895, -64.1943885, -63.3197681, -62.4448815, -61.5695216, -60.6934811, -59.8165529, -58.9385296, -58.059204, -57.178369, -56.2958174, -55.4113418, -54.5247351, -53.63579, -52.7442994, -51.8329775, -50.9105155, -49.98581, -49.059208, -48.1310565, -47.2017025, -46.271493, -45.3407751, -44.4098957, -43.4792018, -42.5490404, -41.6197587, -40.6917034, -39.7652218, -38.8406607, -37.9183673, -36.9986884, -36.0890022, -35.1824722, -34.279293, -33.3796592, -32.4837653, -31.591806, -30.7039758, -29.8204694, -28.9414812, -28.067206, -27.1978383, -26.3335726, -25.4746036, -24.6211259, -23.773334, -22.9314226, -22.0929027, -21.2553501, -20.4244937, -19.6007019, -18.7843433, -17.9757866, -17.1754001, -16.3835525, -15.6006123, -14.826948, -14.0629282, -13.3089215, -12.5652963, -11.8324212, -11.1106648, -10.4003956, -9.70198211, -9.25007025, -8.92296217, -8.59614421, -8.26238216, -7.91444183, -7.545089, -7.14708946, -6.71320902, -6.23621346, -5.70886858, -5.12394016, -4.47419401, -3.75239592, -2.95131168, -2.06370709, -1.08234793, 0, 2.48682242, 4.87601733, 7.18455648, 9.42941162, 11.6275545, 13.7959569, 15.9515905, 18.1114271, 20.2924385, 22.5115964, 24.7858725, 27.1322386, 29.5676665, 32.1091279, 34.7735946, 37.5780383, 40.5511987, 43.7117287, 47.0499821, 50.5854523, 54.3376328, 58.326017, 62.5700985, 67.0893705, 71.9033267, 77.0314605, 82.4932653, 88.3082345, 94.4958617, 101.194466, 108.261145, 115.679746, 123.434114, 131.891302, 140.843553, 149.98092, 159.172287, 168.286538, 177.192556, 185.759225, 196.565988, 205.946217, 213.238468, 217.781297, 218.913263, 220.603941, 216.403114, 204.491582, 183.050147, 150.259611, 110.459484, 48.1445475, -36.4226829, -148.479709, -304.493236, -515.888285, -794.089877, -1139.96526, -1577.89672, -2143.78413, -2822.4305, -3718.50609, -4854.41217, -6300.69119, -8209.45824, -10695.3853, -13682.9378, -17237.475, -22121.7794, -28704.1312, -34606.5305, -43366.0428, -56301.5839, -74732.0699, -99976.4165, -160726.175, -224084.468, -284527.051, -336529.682, -374568.119, -372705.547, -359438.674, -337991.502, -320001.461, -305362.235, -290461.918, -275369.475, -260153.874, -244884.081, -229629.062, -224370.708, -212155.598, -195205.698, -175742.978, -155989.404, -138166.944, -124497.565, -111606.666, -99947.4693},
        {-77.6286638, -76.6633192, -75.7160512, -74.7860167, -73.8723724, -72.9742749, -72.0908811, -71.2213476, -70.3648312, -69.5204885, -68.6874764, -67.8649514, -67.0520704, -66.2479901, -65.4518671, -64.6628583, -63.8801202, -63.1028098, -62.3300836, -61.5610983, -60.7950108, -60.0309777, -59.2681558, -58.5057017, -57.7427723, -56.9785241, -56.212114, -55.4426986, -54.6694347, -53.891479, -53.1079883, -52.3181192, -51.5210284, -50.7158727, -49.8465652, -48.9410463, -48.0279206, -47.1081375, -46.1826466, -45.2523974, -44.3183393, -43.3814218, -42.4425944, -41.5028065, -40.5630078, -39.6241475, -38.6871754, -37.7530407, -36.822693, -35.8970818, -34.9771566, -34.1128015, -33.2549706, -32.4035525, -31.5584358, -30.7195094, -29.8866617, -29.0597816, -28.2387577, -27.4234786, -26.6138331, -25.8097098, -25.0109974, -24.2175845, -23.42936, -22.6462123, -21.8680303, -21.0866041, -20.2939187, -19.5065649, -18.7249564, -17.9495068, -17.1806298, -16.418739, -15.6642482, -14.9175709, -14.1791208, -13.4493116, -12.728557, -12.0172705, -11.315866, -10.6247569, -9.944357, -9.27507995, -8.85973594, -8.57244186, -8.28440121, -7.9881614, -7.67626988, -7.34127406, -6.97572136, -6.57215922, -6.12313505, -5.62119629, -5.05889036, -4.42876468, -3.72336668, -2.93524379, -2.05694343, -1.08101302, 0, 2.23335256, 4.4188359, 6.57108541, 8.70473648, 10.8344245, 12.9747849, 15.1404531, 17.3460643, 19.6062542, 21.9356579, 24.348911, 26.8606489, 29.4855068, 32.2381203, 35.1331246, 38.1851553, 41.5098532, 45.1362829, 48.8506152, 52.6891295, 56.6881053, 60.8838219, 65.3125587, 70.0105953, 75.0142109, 80.3596851, 86.0832973, 92.2213267, 98.810053, 106.109282, 113.849807, 122.000849, 130.53163, 139.696833, 149.322948, 159.160344, 169.092602, 179.003306, 188.776039, 198.294384, 210.511526, 221.307219, 229.96447, 235.766289, 237.995684, 241.189307, 238.063112, 226.586695, 204.729655, 170.461588, 127.706773, 61.2021824, -28.1423632, -145.559275, -308.821357, -530.071034, -821.450728, -1183.89092, -1643.29415, -2237.79416, -2952.22953, -3895.69183, -5093.883, -6620.34636, -8629.68973, -11239.311, -14370.9086, -18091.1436, -23184.2415, -30028.5326, -36461.4013, -45793.0389, -59281.3341, -78184.1755, -103759.452, -163826.546, -226441.479, -286221.765, -337784.921, -375748.461, -374915.029, -362926.925, -342893.753, -326024.924, -312223.866, -298019.088, -283482.764, -268687.07, -253704.18, -238606.269, -233384.594, -221051.151, -203859.67, -184063.88, -163917.509, -145674.288, -131587.945, -117052.559, -105510.321},
        {-73.9283896, -73.0881671, -72.2567231, -71.433599, -70.6183365, -69.8104769, -69.0095619, -68.2151329, -67.4267315, -66.6438992, -65.8661774, -65.0931078, -64.3242318, -63.559091, -62.7972269, -62.0381809, -61.2814947, -60.5267097, -59.7733674, -59.0210094, -58.2691771, -57.5174122, -56.7652561, -56.0122503, -55.2579364, -54.5018558, -53.7435502, -52.9825609, -52.2184296, -51.4506977, -50.6789068, -49.9025983, -49.1213139, -48.3345949, -47.5086791, -46.6604604, -45.807233, -44.9496193, -44.0882415, -43.223722, -42.3566831, -41.4877469, -40.6175359, -39.7466723, -38.8757784, -38.0054766, -37.136389, -36.269138, -35.4043459, -34.5426351, -33.6846276, -32.8591442, -32.0379975, -31.2211986, -30.4087585, -29.6006882, -28.7969988, -27.9977013, -27.2028066, -26.4123258, -25.6262698, -24.8446498, -24.0674767, -23.2947616, -22.5265154, -21.7627491, -21.0034739, -20.2418329, -19.4711346, -18.7055538, -17.9455469, -17.19157, -16.4440794, -15.7035313, -14.9703819, -14.2450873, -13.5281039, -12.8198879, -12.1208954, -11.4315826, -10.7524059, -10.0838214, -9.42628523, -8.78025374, -8.40014235, -8.15408596, -7.90716708, -7.65160044, -7.37960081, -7.08338293, -6.75516155, -6.38715141, -5.97156727, -5.50062388, -4.96653599, -4.36151834, -3.67778569, -2.90755278, -2.04303436, -1.07644518, 0, 2.15164193, 4.27341621, 6.37986894, 8.48554621, 10.6049941, 12.7527588, 14.9433863, 17.1914228, 19.5114143, 21.9179069, 24.4254468, 27.04858, 29.8018526, 32.6998108, 35.7570006, 38.9879681, 42.5425098, 46.4544932, 50.4325411, 54.5201817, 58.7609435, 63.1983547, 67.8759435, 72.8372384, 78.1257677, 83.7850596, 89.8586424, 96.3900445, 103.422794, 111.241431, 119.560101, 128.35003, 137.582441, 147.609954, 158.213097, 169.070693, 180.039549, 190.976473, 201.738273, 212.181754, 225.049041, 236.433485, 245.627374, 251.922996, 254.61264, 259.195351, 257.204972, 246.382102, 224.467339, 189.201283, 143.679424, 73.7426029, -17.3013202, -134.266431, -298.549117, -524.100652, -824.872309, -1202.76863, -1686.52551, -2318.02074, -3088.96631, -4102.27217, -5373.62511, -6975.26682, -9075.26813, -11797.7127, -15077.717, -18990.3731, -24303.6533, -31393.5973, -38258.6396, -48018.2114, -61871.3701, -81017.173, -106654.678, -165550.576, -226944.382, -285643.244, -336454.313, -374184.735, -374654.325, -364315.791, -346113.711, -330624.063, -317770.663, -304297.579, -290287.615, -275823.577, -260988.268, -245864.493, -231393.035, -216193.224, -200524.371, -184645.787, -168816.785, -153296.675, -138344.769, -122551.547, -109762.863},
        {-70.4665418, -69.8390077, -69.1995366, -68.5485471, -67.8864579, -67.2136876, -66.5306548, -65.8377781, -65.135476, -64.4241673, -63.7042706, -62.9762044, -62.2403874, -61.4972382, -60.7471754, -59.9906176, -59.2279834, -58.4596915, -57.6861605, -56.9078089, -56.1250554, -55.3383187, -54.5480172, -53.7545698, -52.9583948, -52.1599111, -51.3595371, -50.5576915, -49.7547929, -48.95126, -48.1475113, -47.3439655, -46.5410412, -45.7391569, -44.9558175, -44.1825391, -43.4106325, -42.6399618, -41.8703913, -41.1017849, -40.3340069, -39.5669214, -38.8003925, -38.0342844, -37.2684612, -36.502787, -35.7371261, -34.9713424, -34.2053002, -33.4388635, -32.6718966, -31.8831775, -31.0941135, -30.3050259, -29.5162358, -28.7280646, -27.9408334, -27.1548636, -26.3704763, -25.5879928, -24.8077344, -24.0300222, -23.2551776, -22.4835217, -21.7153758, -20.9510612, -20.1908991, -19.4322445, -18.6725237, -17.9181755, -17.1697134, -16.427651, -15.6925017, -14.9647792, -14.2449969, -13.5336684, -12.8313072, -12.1384268, -11.4555407, -10.7831626, -10.1218058, -9.47198398, -8.83421062, -8.20899926, -7.8652159, -7.6635534, -7.46147923, -7.25079835, -7.02331574, -6.77083637, -6.48516522, -6.15810726, -5.78146747, -5.34705082, -4.84666228, -4.27210683, -3.61518945, -2.8677151, -2.02148875, -1.0683154, 0, 2.17818092, 4.32418555, 6.4539343, 8.5833476, 10.7283459, 12.9048495, 15.1287789, 17.4160546, 19.7825969, 22.2443262, 24.817163, 27.5170277, 30.3598407, 33.3615224, 36.5379933, 39.9051737, 43.6139746, 47.6996007, 51.8726368, 56.1779299, 60.660327, 65.3646748, 70.3358205, 75.618611, 81.2578932, 87.2985141, 93.7853206, 100.76316, 108.276878, 116.581446, 125.434542, 134.817977, 144.713561, 155.665285, 167.373869, 179.390118, 191.527189, 203.598239, 215.416424, 226.7949, 240.024006, 251.685791, 261.108745, 267.621359, 270.552122, 276.417337, 275.560729, 265.513837, 243.808198, 207.97535, 160.355461, 87.7942228, -3.83615897, -118.926965, -282.824093, -511.465246, -820.788127, -1213.26606, -1721.36121, -2392.22416, -3228.09977, -4320.43115, -5668.16389, -7340.18858, -9523.91383, -12352.3465, -15782.9058, -19903.0758, -25444.1625, -32773.2897, -40008.0293, -50111.1422, -64228.4433, -83505.7475, -109088.87, -166711.724, -226785.001, -284307.492, -334277.987, -371695.277, -373408.713, -364666.15, -348244.709, -334073.086, -322097.57, -309284.629, -295729.545, -281527.599, -266774.074, -251564.249, -225727.45, -207055.401, -193345.877, -182396.652, -172005.499, -159970.192, -144088.506, -127648.378, -113041.326},
        {-66.4848201, -65.9993522, -65.4909281, -64.960409, -64.4086564, -63.8365313, -63.244895, -62.6346089, -62.006534, -61.3615318, -60.7004633, -60.02419, -59.333573, -58.6294735, -57.9127529, -57.1842723, -56.444893, -55.6954764, -54.9368835, -54.1699757, -53.3956142, -52.6146602, -51.8279751, -51.03642, -50.2408562, -49.442145, -48.6411475, -47.8387252, -47.0357391, -46.2330505, -45.4315208, -44.6320111, -43.8353826, -43.0424967, -42.2936066, -41.569049, -40.848687, -40.1321037, -39.4188819, -38.7086045, -38.0008544, -37.2952145, -36.5912677, -35.888597, -35.1867852, -34.4854152, -33.7840699, -33.0823323, -32.3797852, -31.6760116, -30.9705943, -30.2275675, -29.4828336, -28.7367459, -27.9896581, -27.2419236, -26.493896, -25.7459288, -24.9983756, -24.2515898, -23.505925, -22.7617348, -22.0193726, -21.279192, -20.5415464, -19.8067896, -19.0752748, -18.3428811, -17.6055946, -16.8729976, -16.1457338, -15.4244467, -14.7097799, -14.002377, -13.3028817, -12.6119375, -11.9301881, -11.2582771, -10.596848, -9.94654441, -9.30801005, -8.68188845, -8.06882324, -7.46945802, -7.16996044, -7.02220648, -6.8746403, -6.71863972, -6.54558253, -6.34684654, -6.11380953, -5.83784931, -5.51034369, -5.12267046, -4.66620742, -4.13233237, -3.51242311, -2.79785745, -1.98001317, -1.05026809, 0, 2.17065396, 4.31059507, 6.43702394, 8.56714119, 10.7181474, 12.9072432, 15.1516293, 17.4685061, 19.8750744, 22.3885347, 25.0260876, 27.8049338, 30.7422739, 33.8553085, 37.1612381, 40.6772634, 44.5553389, 48.8319156, 53.2193939, 57.7638502, 62.5113611, 67.5080028, 72.7998518, 78.4329846, 84.4534775, 90.907407, 97.8408495, 105.299881, 113.330579, 122.174654, 131.610814, 141.626447, 152.208938, 163.973408, 176.593376, 189.574964, 202.717238, 215.819264, 228.680108, 241.098834, 255.251925, 267.837469, 278.189384, 285.641591, 289.528008, 296.968405, 297.564387, 288.703412, 267.772934, 232.160411, 184.689337, 110.667082, 14.1850309, -109.845837, -285.257861, -527.46801, -851.893257, -1255.3111, -1776.01423, -2472.62929, -3360.23104, -4518.87649, -5930.98288, -7662.92772, -9923.35453, -12855.6044, -16438.8197, -20773.7716, -26563.2928, -34169.6559, -41784.3975, -52274.4856, -66753.6016, -86335.4269, -112133.643, -169447.308, -229158.382, -286334.202, -336042.104, -373349.425, -375446.461, -367195.617, -351325.821, -337573.045, -325892.594, -313320.167, -299957.047, -285904.517, -271263.86, -256136.36, -235703.837, -218545.554, -203664.01, -190061.699, -176741.117, -162704.762, -146955.128, -131454.832, -116073.712},
        {-62.2183179, -61.8262068, -61.4070392, -60.9618109, -60.4915176, -59.9971551, -59.4797192, -58.9402056, -58.37961, -57.7989282, -57.1991559, -56.581289, -55.9463231, -55.295254, -54.6290774, -53.9487892, -53.255385, -52.5498607, -51.8332119, -51.1064344, -50.3705239, -49.6264763, -48.8752872, -48.1179525, -47.3554678, -46.588829, -45.8190317, -45.0470718, -44.2739449, -43.5006468, -42.7281733, -41.9575202, -41.1896831, -40.4256579, -39.7093161, -39.0193136, -38.3337913, -37.6523534, -36.9746043, -36.3001482, -35.6285897, -34.959533, -34.2925825, -33.6273425, -32.9634173, -32.3004114, -31.637929, -30.9755746, -30.3129524, -29.6496668, -28.9853222, -28.2902777, -27.5940167, -26.8967778, -26.1987991, -25.5003191, -24.801576, -24.1028082, -23.404254, -22.7061517, -22.0087397, -21.3122563, -20.6169398, -19.9230286, -19.230761, -18.5403752, -17.8521097, -17.1575872, -16.4486373, -15.7432672, -15.0422739, -14.3464542, -13.656605, -12.973523, -12.2980051, -11.6308482, -10.9728491, -10.3248046, -9.68751162, -9.06176696, -8.44836747, -7.84811, -7.26179138, -6.69020845, -6.4380967, -6.34789072, -6.25837197, -6.1604739, -6.04512997, -5.9032736, -5.72583827, -5.50375741, -5.22796447, -4.8893929, -4.47897615, -3.98764767, -3.40634091, -2.72598932, -1.93752633, -1.03188541, 0, 2.19515404, 4.35549668, 6.49990454, 8.64725425, 10.8164224, 13.0262857, 15.2957207, 17.6436041, 20.0888125, 22.6502225, 25.3467107, 28.1971537, 31.2204283, 34.435411, 37.8609784, 41.5160072, 45.5513125, 50.0046193, 54.5963719, 59.3737194, 64.383811, 69.6737959, 75.2908232, 81.2820421, 87.6946018, 94.5756514, 101.97234, 109.931817, 118.501231, 127.921476, 137.974917, 148.650581, 159.937493, 172.444013, 185.849501, 199.66816, 213.703216, 227.757893, 241.635419, 255.139018, 270.62755, 284.570806, 296.271996, 305.034332, 310.161025, 318.831887, 320.504377, 312.512552, 292.190474, 256.872202, 210.474952, 135.703547, 33.752724, -101.515623, -291.210823, -549.72043, -891.432, -1307.31821, -1840.50305, -2559.65404, -3488.80809, -4701.95511, -6171.21249, -7962.49222, -10300.636, -13337.5341, -17069.7314, -21608.8434, -27636.0861, -35510.4774, -43490.5597, -54358.9762, -69201.8323, -89105.2334, -115155.285, -172393.599, -231961.897, -288957.409, -338477.363, -375618.988, -377676.003, -369418.153, -353572.743, -339873.878, -328275.337, -315811.401, -302581.825, -288686.363, -274224.77, -259296.802, -248944.029, -234754.373, -217926.697, -199659.867, -181152.746, -163604.199, -148213.092, -134227.712, -118551.953},
        {-57.7832348, -57.4346396, -57.0621634, -56.6666176, -56.2488133, -55.8095618, -55.3496745, -54.8699626, -54.3712374, -53.8543102, -53.3199922, -52.7690947, -52.2024291, -51.6208065, -51.0250382, -50.4159356, -49.7943099, -49.1609723, -48.5167343, -47.8624069, -47.1988016, -46.5267296, -45.8470022, -45.1604305, -44.4678261, -43.77, -43.0677636, -42.3619281, -41.6533049, -40.9427052, -40.2309403, -39.5188215, -38.80716, -38.0967671, -37.4170162, -36.7538369, -36.0928152, -35.4338354, -34.7767821, -34.1215395, -33.4679921, -32.8160242, -32.1655202, -31.5163645, -30.8684414, -30.2216354, -29.5758308, -28.930912, -28.2867634, -27.6432694, -27.0003143, -26.3505534, -25.7012569, -25.0524659, -24.4042216, -23.7565649, -23.109537, -22.4631791, -21.8175321, -21.1726372, -20.5285354, -19.8852679, -19.2428758, -18.6014002, -17.9608821, -17.3213626, -16.6828829, -16.0317782, -15.354713, -14.6799956, -14.0085553, -13.3413219, -12.6792249, -12.0231939, -11.3741585, -10.7330483, -10.1007928, -9.47832165, -8.86656441, -8.26645066, -7.67890999, -7.10487198, -6.54526621, -6.00102227, -5.7950214, -5.76044554, -5.726607, -5.68398741, -5.6230684, -5.53433161, -5.40825868, -5.23533122, -5.00603088, -4.71083928, -4.34023805, -3.88470884, -3.33473326, -2.68079296, -1.91336956, -1.0229447, 0, 2.31129422, 4.56989503, 6.79716649, 9.0144727, 11.2431777, 13.5046457, 15.8202406, 18.2113266, 20.6992677, 23.3054281, 26.0511717, 28.9578628, 32.0468653, 35.3395434, 38.8572611, 42.6213825, 46.7807886, 51.3750602, 56.1371303, 61.1156879, 66.3594223, 71.9170225, 77.8371777, 84.1685771, 90.9599097, 98.2598647, 106.117131, 114.580398, 123.698355, 133.730248, 144.437005, 155.804147, 167.817198, 181.007896, 195.088655, 209.626229, 224.442277, 239.358455, 254.196419, 268.777826, 286.004803, 301.681003, 315.025383, 325.256898, 331.594505, 340.446895, 342.045855, 333.81291, 313.169583, 277.537396, 232.090429, 157.022301, 52.3460711, -88.8228978, -287.452036, -558.185887, -915.668993, -1350.83919, -1907.43561, -2655.05928, -3615.91003, -4869.72266, -6392.15666, -8252.43742, -10678.3326, -13824.8362, -17696.0352, -22407.7995, -28629.326, -36716.8186, -45027.9865, -56224.9891, -71355.8168, -91468.4603, -117610.91, -174385.236, -233396.829, -289805.161, -338769.701, -375449.921, -377378.777, -369093.263, -353307.07, -339736.115, -328322.375, -316090.231, -303127.538, -289522.153, -275361.929, -260734.724, -254767.367, -241998.486, -224506.703, -204370.639, -183668.916, -164480.155, -148882.977, -136111.463, -120334.284},
        {-53.2138641, -52.8419489, -52.4591179, -52.0655738, -51.661519, -51.2471563, -50.822688, -50.3883168, -49.9442453, -49.490676, -49.0278114, -48.5558542, -48.0750068, -47.5854719, -47.087452, -46.5811497, -46.0667675, -45.544508, -45.0145737, -44.4771673, -43.9324912, -43.3807481, -42.8221405, -42.2568709, -41.685142, -41.1071562, -40.5231162, -39.9332245, -39.3376836, -38.7366962, -38.1304648, -37.5191919, -36.9030801, -36.282332, -35.6489535, -35.007418, -34.3622971, -33.7140595, -33.0631737, -32.4101082, -31.7553316, -31.0993125, -30.4425194, -29.7854209, -29.1284855, -28.4721817, -27.8169783, -27.1633436, -26.5117462, -25.8626548, -25.2165378, -24.605514, -23.9977156, -23.3929252, -22.7909251, -22.1914978, -21.5944257, -20.9994913, -20.406477, -19.8151653, -19.2253386, -18.6367793, -18.04927, -17.4625929, -16.8765307, -16.2908656, -15.7053803, -15.1010393, -14.4592592, -13.8186326, -13.1801619, -12.5448492, -11.9136971, -11.2877077, -10.6678834, -10.0552265, -9.45073934, -8.85542424, -8.27028353, -7.69631951, -7.1345345, -6.58593083, -6.05151081, -5.53227676, -5.36707656, -5.38088362, -5.39461071, -5.39829642, -5.38197934, -5.33569804, -5.2494911, -5.11339712, -4.91745466, -4.65170232, -4.30617868, -3.87092231, -3.3359718, -2.69136573, -1.92714269, -1.03334125, 0, 2.5740786, 5.05636598, 7.47188818, 9.84567124, 12.2027412, 14.5681241, 16.966846, 19.4239328, 21.9644108, 24.6133058, 27.3956439, 30.3364512, 33.4607537, 36.7935774, 40.3599484, 44.1848927, 48.4183665, 53.1032428, 57.9819679, 63.1063385, 68.5281513, 74.2992032, 80.4712909, 87.0962111, 94.2257607, 101.911736, 110.205935, 119.160153, 128.826188, 139.514831, 150.923919, 163.02755, 175.799822, 189.610413, 204.23563, 219.346296, 234.797833, 250.445668, 266.145225, 281.751927, 301.209728, 319.04119, 334.301807, 346.047074, 353.332488, 360.568704, 360.117266, 349.694878, 327.018244, 289.80407, 244.324137, 169.283866, 67.7816793, -66.9374256, -261.745859, -534.483724, -902.991127, -1369.62035, -1971.21928, -2761.5316, -3742.67162, -5019.27867, -6593.09888, -8542.53829, -11075.899, -14341.8586, -18335.4165, -23165.4081, -29504.155, -37705.2394, -46297.3938, -57739.7955, -73017.705, -93116.3828, -119021.089, -174395.25, -231876.419, -286780.317, -334422.662, -370119.174, -372097.173, -364153.043, -348926.105, -335929.806, -325082.456, -313438.144, -301058.39, -288004.714, -274338.638, -260121.681, -243679.052, -228148.849, -213120.483, -198183.363, -182926.903, -166940.512, -149813.601, -137172.935, -121393.172},
        {-49.2914599, -48.9702988, -48.6404842, -48.3020675, -47.9551001, -47.5996333, -47.2357186, -46.8634073, -46.4827509, -46.0938008, -45.6966082, -45.2912247, -44.8777016, -44.4560903, -44.0264423, -43.5888088, -43.1432412, -42.6897911, -42.2285097, -41.7594485, -41.2826589, -40.7981921, -40.3060998, -39.8064331, -39.2992436, -38.7845826, -38.2625015, -37.7330517, -37.1962845, -36.6522515, -36.1010039, -35.5425933, -34.9770708, -34.404488, -33.8060872, -33.1917198, -32.5714635, -31.9459802, -31.3159316, -30.6819795, -30.0447857, -29.405012, -28.7633201, -28.1203719, -27.4768291, -26.8333535, -26.190607, -25.5492512, -24.909948, -24.2733591, -23.6401464, -23.0536382, -22.4709047, -21.891683, -21.3157097, -20.7427217, -20.172456, -19.6046494, -19.0390386, -18.4753607, -17.9133523, -17.3527505, -16.793292, -16.2347137, -15.6767525, -15.1191452, -14.5616286, -13.9832188, -13.3634291, -12.7447317, -12.1282068, -11.5149344, -10.9059948, -10.3024679, -9.70543392, -9.11597295, -8.53516511, -7.9640905, -7.40382923, -6.85546141, -6.32006715, -5.79872657, -5.29251977, -4.80252686, -4.68652986, -4.7597552, -4.83314198, -4.89619467, -4.9384177, -4.94931552, -4.91839259, -4.83515335, -4.68910224, -4.46974372, -4.16658223, -3.76912222, -3.26686814, -2.64932442, -1.90599553, -1.02638591, 0, 2.66739229, 5.23364454, 7.72551253, 10.1697521, 12.5931189, 15.022369, 17.4842579, 20.0055415, 22.6129757, 25.3333162, 28.1933188, 31.2197392, 34.4393334, 37.8788571, 41.565066, 45.524716, 49.9154438, 54.7827023, 59.8572074, 65.1937607, 70.8471639, 76.8722188, 83.3237269, 90.25649, 97.7253099, 105.784988, 114.490326, 123.896127, 134.05719, 145.287687, 157.287949, 170.034968, 183.505734, 198.095094, 213.571111, 229.590422, 246.004662, 262.665467, 279.424472, 296.133313, 317.106067, 336.373792, 352.915035, 365.708347, 373.732274, 381.298163, 380.718565, 369.63883, 345.704305, 306.560341, 258.579294, 180.013161, 75.3347257, -62.2236191, -262.378615, -544.543312, -928.130758, -1422.18926, -2058.85224, -2881.56008, -3876.53958, -5169.85345, -6783.31327, -8803.92105, -11429.1404, -14805.7312, -18928.7897, -23905.7556, -30400.1687, -38761.0071, -47680.1089, -59418.5369, -74898.8944, -95043.7852, -120775.813, -174836.871, -230875.45, -284359.331, -330756.294, -365534.121, -367604.819, -360029.124, -345370.149, -332913.898, -322562.341, -311416.238, -299517.618, -286908.509, -273630.938, -259726.934, -240373.674, -223955.214, -209468.83, -195911.797, -182281.392, -167574.889, -150789.566, -137496.514, -119061.119},
        {-46.0338544, -45.857255, -45.6606334, -45.4444765, -45.2092713, -44.9555048, -44.6836641, -44.3942361, -44.0877077, -43.7645661, -43.4252981, -43.0703909, -42.7003313, -42.3156064, -41.9167032, -41.5041086, -41.0783098, -40.6397935, -40.189047, -39.7265571, -39.2528108, -38.7682952, -38.2734972, -37.7689039, -37.2550021, -36.7322791, -36.2012216, -35.6623167, -35.1160515, -34.5629129, -34.0033879, -33.4379634, -32.8671266, -32.2913643, -31.7146997, -31.1357772, -30.5531991, -29.9673375, -29.3785647, -28.7872531, -28.1937747, -27.5985019, -27.0018068, -26.4040619, -25.8056392, -25.206911, -24.6082497, -24.0100273, -23.4126163, -22.8163887, -22.221717, -21.6502743, -21.0806701, -20.5128148, -19.9466191, -19.3819932, -18.8188478, -18.2570932, -17.6966401, -17.1373988, -16.5792798, -16.0221937, -15.4660509, -14.9107618, -14.356237, -13.802387, -13.2491221, -12.6768513, -12.0664516, -11.4580541, -10.8528336, -10.2519648, -9.6566224, -9.06798103, -8.48721541, -7.91550025, -7.35401023, -6.80392005, -6.26640441, -5.74263799, -5.23379551, -4.74105164, -4.26558109, -3.80855855, -3.74962432, -3.8927598, -4.03739747, -4.1724301, -4.28675044, -4.36925125, -4.40882529, -4.39436532, -4.3147641, -4.15891437, -3.91570891, -3.57404046, -3.12280179, -2.55088565, -1.8471848, -1.000592, 0, 2.56787248, 5.05816941, 7.49729564, 9.911656, 12.3276554, 14.7716986, 17.2701904, 19.8495358, 22.5361396, 25.3564066, 28.3367417, 31.5035497, 34.8832354, 38.5022038, 42.3868596, 46.5636077, 51.2012299, 56.3464706, 61.7036915, 67.329807, 73.2817311, 79.616378, 86.390662, 93.6614972, 101.485798, 109.920478, 119.022452, 128.848634, 139.455938, 151.107269, 163.578021, 176.863311, 190.958254, 206.480852, 223.114778, 240.377608, 258.077594, 276.022987, 294.02204, 311.883005, 333.508135, 353.365681, 370.462896, 383.807032, 392.40534, 402.600783, 404.230976, 394.469242, 370.488907, 329.463296, 276.928576, 191.473975, 76.275735, -75.7815881, -292.825632, -593.549988, -996.648249, -1510.0243, -2165.61364, -3007.12237, -4015.27788, -5327.12318, -6971.22627, -9040.23322, -11735.3668, -15208.0113, -19466.1224, -24623.8992, -31321.4457, -39901.3845, -49197.263, -61292.9442, -77053.4226, -97343.6929, -123028.75, -176030.682, -230893.117, -283216.775, -328602.377, -362650.644, -364889.16, -357707.208, -343593.997, -331571.241, -321542.364, -310688.917, -299050.418, -286666.389, -273576.347, -259819.815, -247558.755, -232864.476, -216561.042, -199472.518, -182422.966, -166236.453, -151737.041, -137260.894, -114248.554},
        {-42.358798, -42.3306763, -42.2705669, -42.1794288, -42.0582208, -41.9079019, -41.7294308, -41.5237664, -41.2918677, -41.0346935, -40.7532027, -40.448354, -40.1211065, -39.772419, -39.4032503, -39.0145594, -38.607305, -38.1824461, -37.7409416, -37.2837502, -36.811831, -36.3261427, -35.8276442, -35.3172945, -34.7960523, -34.2648765, -33.7247261, -33.1765598, -32.6213367, -32.0600154, -31.493555, -30.9229142, -30.349052, -29.7729272, -29.2215259, -28.682246, -28.1421724, -27.6014193, -27.060101, -26.5183318, -25.9762258, -25.4338973, -24.8914606, -24.3490298, -23.8067192, -23.2646431, -22.7229157, -22.1816512, -21.6409639, -21.100968, -20.5617777, -20.0319259, -19.5029257, -18.9747089, -18.4472072, -17.9203523, -17.394076, -16.86831, -16.342986, -15.8180358, -15.2933911, -14.7689838, -14.2447454, -13.7206077, -13.1965025, -12.6723616, -12.1481166, -11.6029559, -11.0165658, -10.4316599, -9.8495146, -9.27140628, -8.69861142, -8.13240647, -7.57406787, -7.02487209, -6.48609555, -5.95901472, -5.44490603, -4.94504595, -4.46071091, -3.99317737, -3.54372176, -3.11362055, -3.10117443, -3.30079685, -3.50212934, -3.69356414, -3.86349351, -4.00030969, -4.09240492, -4.12817146, -4.09600156, -3.98428746, -3.78142141, -3.47579565, -3.05580244, -2.50983402, -1.82628264, -0.993540554, 0, 2.55702215, 5.04433413, 7.48911779, 9.91855497, 12.3598275, 14.8401174, 17.3866063, 20.0264761, 22.7869088, 25.6950861, 28.7781899, 32.0634021, 35.5779045, 39.348879, 43.4035075, 47.7689717, 52.6098905, 57.9730796, 63.5748515, 69.4718389, 75.7206744, 82.3779907, 89.5004203, 97.1445959, 105.36715, 114.224716, 123.773925, 134.071411, 145.173806, 157.315542, 170.310259, 184.16125, 198.87181, 215.151912, 232.651509, 250.832105, 269.484987, 288.401445, 307.372766, 326.19024, 348.635884, 369.295688, 387.180144, 401.299748, 410.664991, 422.685348, 425.872587, 417.137457, 393.390709, 351.543089, 297.110692, 207.881279, 85.0556428, -79.0782607, -310.429509, -626.737081, -1045.73996, -1570.72774, -2237.17273, -3096.31022, -4134.87217, -5486.58191, -7175.31229, -9293.75128, -12051.1902, -15603.3974, -19969.0849, -25268.488, -32124.3826, -40875.2792, -50501.5736, -62908.8318, -78905.6813, -99300.7491, -124902.663, -176781.242, -230418.623, -281558.135, -325943.107, -359316.868, -361919.545, -355333.137, -341945.205, -330308.323, -320369.732, -309507.914, -297798.547, -285317.308, -272139.876, -258341.928, -249947.362, -236533.796, -219752.636, -201255.291, -182693.17, -165717.68, -151980.228, -136993.917, -114703.064},
        {-38.288519, -38.3820944, -38.4352129, -38.4491649, -38.4252408, -38.3647311, -38.2689261, -38.1391164, -37.9765924, -37.7826444, -37.558563, -37.3056385, -37.0251615, -36.7184222, -36.3867113, -36.0313191, -35.653536, -35.2546526, -34.8359591, -34.3987461, -33.944304, -33.4739232, -32.9888942, -32.4905074, -31.9800533, -31.4588222, -30.9281046, -30.389191, -29.8433718, -29.2919374, -28.7361782, -28.1773847, -27.6168474, -27.0558566, -26.5350861, -26.0353062, -25.5368141, -25.0396222, -24.5437428, -24.0491883, -23.555971, -23.0641034, -22.5735978, -22.0844666, -21.5967221, -21.1103768, -20.625443, -20.1419331, -19.6598595, -19.1792345, -18.7000705, -18.2330543, -17.7672925, -17.302566, -16.8386558, -16.3753429, -15.9124083, -15.4496329, -14.9867977, -14.5236836, -14.0600718, -13.595743, -13.1304784, -12.6640588, -12.1962653, -11.7268788, -11.2556803, -10.7577705, -10.2088426, -9.65957883, -9.11136012, -8.56556737, -8.02358146, -7.48678329, -6.95655376, -6.43427377, -5.92132422, -5.41908601, -4.92894002, -4.45226717, -3.99044834, -3.54486444, -3.11689636, -2.707925, -2.73200706, -2.97550545, -3.21982305, -3.45294863, -3.66287096, -3.83757883, -3.96506101, -4.03330627, -4.03030338, -3.94404112, -3.76250827, -3.47369359, -3.06558587, -2.52617387, -1.84344638, -1.00539216, 0, 2.68186148, 5.27799863, 7.81800602, 10.3314782, 12.8480098, 15.3971953, 18.0086293, 20.7119064, 23.5366212, 26.5123682, 29.6687419, 33.035337, 36.6417481, 40.5175696, 44.6923962, 49.1958224, 54.1699248, 59.6603669, 65.4423196, 71.5694807, 78.0955481, 85.0742197, 92.5591932, 100.604167, 109.262838, 118.588905, 128.636065, 139.458017, 151.108457, 163.817204, 177.397258, 191.849483, 207.174741, 224.039942, 242.112923, 260.883913, 280.153958, 299.724108, 319.395411, 338.968914, 362.461631, 384.175525, 403.086784, 418.171593, 428.406137, 441.149323, 444.898937, 436.535483, 412.939467, 370.991394, 316.921663, 226.833239, 100.166975, -71.4528167, -312.360598, -639.725665, -1070.71732, -1603.01429, -2277.70612, -3157.14475, -4241.56777, -5651.2515, -7398.50607, -9571.75657, -12388.4006, -16006.6353, -20450.5753, -25843.6558, -32800.947, -41659.4225, -51551.0433, -64203.1821, -80369.9325, -100805.388, -126263.641, -176940.102, -229287.219, -279198.757, -322568.481, -355290.154, -358388.341, -352538.807, -340007.832, -328743.085, -318761.961, -307714.61, -295736.791, -282964.261, -269532.779, -255578.102, -245927.785, -232735.631, -217081.658, -200045.886, -182708.332, -166149.016, -151447.954, -136678.19, -119765.531},
        {-34.5767865, -34.7355551, -34.8523292, -34.9284315, -34.965185, -34.9639126, -34.9259372, -34.8525817, -34.7451689, -34.6050218, -34.4334632, -34.2318161, -34.0014034, -33.7435479, -33.4595725, -33.1508001, -32.8185537, -32.4641561, -32.0889302, -31.694199, -31.2812852, -30.8515119, -30.4062018, -29.9466779, -29.4742632, -28.9902804, -28.4960524, -27.9929023, -27.4821528, -26.9651269, -26.4431474, -25.9175373, -25.3896194, -24.8607167, -24.3694427, -23.8976907, -23.4269056, -22.9572002, -22.4886871, -22.0214791, -21.555689, -21.0914294, -20.6288132, -20.1679531, -19.7089617, -19.2519519, -18.7970364, -18.3443279, -17.8939392, -17.445983, -17.000572, -16.5805889, -16.162883, -15.7470733, -15.332779, -14.9196191, -14.5072129, -14.0951795, -13.6831379, -13.2707073, -12.8575069, -12.4431557, -12.0272729, -11.6094777, -11.1893891, -10.7666264, -10.3408085, -9.88265227, -9.36356838, -8.84278497, -8.32179478, -7.80209058, -7.28516511, -6.77251113, -6.26562139, -5.76598865, -5.27510566, -4.79446518, -4.32555995, -3.86988274, -3.4289263, -3.00418338, -2.59714674, -2.20930912, -2.27287273, -2.56491621, -2.8573414, -3.1376637, -3.39339851, -3.61206126, -3.78116732, -3.88823213, -3.92077106, -3.86629954, -3.71233296, -3.44638674, -3.05597626, -2.52861695, -1.8518242, -1.01311341, 0, 2.90876032, 5.6976178, 8.39974378, 11.0483096, 13.6764865, 16.3174459, 19.0043591, 21.7703974, 24.6487321, 27.6725346, 30.8749762, 34.2892282, 37.9484619, 41.8858487, 46.1345598, 50.7277667, 55.7681495, 61.2988093, 67.1956951, 73.5068729, 80.2804088, 87.564369, 95.4068194, 103.855826, 112.959455, 122.765773, 133.322845, 144.678738, 156.881518, 170.131586, 184.268817, 199.295575, 215.214228, 232.720417, 251.46987, 270.933438, 290.905504, 311.180449, 331.552656, 351.816507, 376.034627, 398.434124, 417.974288, 433.614411, 444.313783, 457.905497, 462.256591, 454.107905, 430.200278, 387.27455, 331.615353, 238.755103, 107.779245, -69.9664473, -318.938036, -656.496883, -1100.00435, -1644.48872, -2334.72013, -3239.83206, -4369.34502, -5833.99274, -7639.68119, -9871.54204, -12748.9506, -16429.662, -20936.7945, -26393.1974, -33415.3895, -42340.6424, -52452.5297, -65304.128, -81597.4504, -102034.51, -127317.32, -176810.671, -227888.104, -276585.937, -318940.489, -350988.077, -354470.124, -349247.776, -337483.734, -326673.95, -316870.047, -305901.466, -293930.938, -281121.2, -267634.985, -253635.027, -240117.999, -226114.561, -211668.615, -196824.061, -181624.799, -166114.73, -150337.754, -135816.565, -122631.833},
        {-31.4352978, -31.5962291, -31.7212073, -31.8112598, -31.867414, -31.8906974, -31.8821373, -31.8427612, -31.7735965, -31.6756706, -31.5500109, -31.3976448, -31.2195997, -31.0169031, -30.7905823, -30.5416647, -30.2711779, -29.9801491, -29.6696058, -29.3405754, -28.9940853, -28.631163, -28.2528357, -27.8601311, -27.4540764, -27.035699, -26.6060265, -26.1660861, -25.7169053, -25.2595116, -24.7949322, -24.3241948, -23.8483265, -23.3683549, -22.9045653, -22.4479511, -21.9892741, -21.5289369, -21.0673419, -20.6048916, -20.1419884, -19.6790348, -19.2164332, -18.7545861, -18.2938959, -17.8347652, -17.3775963, -16.9227917, -16.4707538, -16.0218852, -15.5765883, -15.1761399, -14.7791821, -14.3852311, -13.9938034, -13.6044153, -13.2165833, -12.8298237, -12.4436528, -12.0575871, -11.6711429, -11.2838366, -10.8951846, -10.5047032, -10.1119089, -9.71631803, -9.31744693, -8.88266167, -8.38010022, -7.87558658, -7.37072133, -6.867105, -6.36633816, -5.87002136, -5.37975514, -4.89714006, -4.42377667, -3.96126553, -3.51120719, -3.0752022, -2.65485112, -2.25175448, -1.86751286, -1.5037268, -1.6140873, -1.96464058, -2.31599907, -2.65509212, -2.96884908, -3.24419929, -3.46807211, -3.62739687, -3.70910294, -3.70011965, -3.58737636, -3.35780241, -2.99832715, -2.49587994, -1.83739011, -1.00978701, 0, 3.2104886, 6.25355256, 9.16675387, 11.9876545, 14.7538164, 17.5028015, 20.2721719, 23.0994895, 26.0223162, 29.0782141, 32.3047451, 35.7394712, 39.4199543, 43.3837565, 47.6684396, 52.3115657, 57.3631705, 62.8625979, 68.8191955, 75.2731981, 82.2648408, 89.8343585, 98.0219859, 106.867958, 116.41251, 126.695876, 137.758292, 149.639993, 162.381212, 176.117414, 190.758687, 206.3167, 222.803119, 241.065489, 260.697537, 281.07009, 301.944051, 323.080323, 344.23981, 365.183414, 389.721681, 412.333371, 431.987068, 447.651353, 458.294808, 473.114212, 478.292902, 470.242409, 445.374265, 400.100005, 339.736416, 240.905573, 104.195549, -78.7952499, -334.822331, -682.400667, -1140.04523, -1704.71319, -2421.37683, -3358.52606, -4526.37838, -6035.14156, -7893.52118, -10186.2451, -13126.8567, -16869.771, -21431.5604, -26930.9031, -33991.8992, -42952.7535, -53237.492, -66239.1413, -82612.3527, -103011.778, -128092.068, -176475.763, -226357.654, -273900.419, -315266.737, -346619.288, -350291.006, -345494.144, -334318.414, -324047.147, -314714.248, -304189.613, -292615.349, -280133.559, -266886.35, -253015.826, -234137.748, -218432.007, -204792.44, -192112.883, -179287.173, -165209.147, -148772.64, -134332.734, -121226.051},
        {-28.5074737, -28.6560503, -28.7758124, -28.8674552, -28.931674, -28.9691639, -28.9806203, -28.9667385, -28.9282135, -28.8657407, -28.7800153, -28.6717326, -28.5415878, -28.3902761, -28.2184928, -28.0269332, -27.8162924, -27.5872657, -27.3405483, -27.0768356, -26.7968227, -26.5012049, -26.1906774, -25.8659355, -25.5276744, -25.1765893, -24.8133755, -24.4387283, -24.0533429, -23.6579144, -23.2531382, -22.8397096, -22.4183237, -21.9896757, -21.5556588, -21.116344, -20.6717882, -20.2226476, -19.7695788, -19.313238, -18.8542816, -18.393366, -17.9311476, -17.4682827, -17.0054278, -16.5432391, -16.0823732, -15.6234862, -15.1672347, -14.7142749, -14.2652633, -13.8740594, -13.486963, -13.1034772, -12.723105, -12.3453495, -11.9697137, -11.5957006, -11.2228133, -10.8505548, -10.4784281, -10.1059363, -9.73258237, -9.3578694, -8.98130039, -8.60237838, -8.22060641, -7.80215515, -7.31399558, -6.82437622, -6.33496089, -5.84741345, -5.36339771, -4.88457751, -4.41261669, -3.94917909, -3.49592853, -3.05452886, -2.62664391, -2.2139375, -1.81807349, -1.4407157, -1.08352796, -0.748174115, -0.908765789, -1.32080562, -1.73411778, -2.13503426, -2.50988706, -2.84500818, -3.12672962, -3.34138335, -3.47530139, -3.51481573, -3.44625836, -3.25596127, -2.93025647, -2.45547595, -1.8179517, -1.00401572, 0, 3.49798488, 6.78364084, 9.89873554, 12.8850366, 15.7843116, 18.6383283, 21.4888542, 24.377657, 27.3465044, 30.4371639, 33.6914032, 37.1509899, 40.8576917, 44.8532761, 49.1795109, 53.8781637, 58.9462133, 64.419029, 70.4396863, 77.0403553, 84.2532058, 92.1104081, 100.644132, 109.886548, 119.869826, 130.626135, 142.187647, 154.586531, 167.854957, 182.058359, 197.183447, 213.25241, 230.28744, 249.300373, 269.818576, 291.105511, 312.886477, 334.886769, 356.831685, 378.446522, 403.3226, 426.142576, 445.875353, 461.48983, 471.954908, 487.813431, 493.566872, 485.290647, 459.06017, 410.950858, 345.086664, 239.657362, 97.6297681, -88.9244963, -350.178164, -706.316707, -1177.5256, -1764.84706, -2511.96054, -3483.42906, -4687.04024, -6235.18036, -8141.0696, -10489.8024, -13488.2599, -17287.4214, -21899.1378, -27438.3178, -34530.007, -43510.689, -53903.1773, -66982.805, -83367.7104, -103676.032, -128525.907, -175933.661, -224769.7, -271302.614, -311800.996, -342533.438, -346322.994, -341847.484, -331137.874, -321325.491, -312422.406, -302315.539, -291123.219, -278963.775, -265955.536, -252216.831, -228541.805, -211264.317, -198273.44, -187458.248, -176707.816, -163911.219, -146957.53, -132675.506, -118653.508},
        {-25.3925761, -25.5830052, -25.7435067, -25.874831, -25.9777285, -26.0529496, -26.1012445, -26.1233636, -26.1200574, -26.0920761, -26.0401701, -25.9650897, -25.8675854, -25.7484074, -25.6083061, -25.4480319, -25.2683351, -25.0699661, -24.8536752, -24.6202128, -24.3703293, -24.1047749, -23.8243001, -23.5296552, -23.2215905, -22.9008565, -22.5682034, -22.2243816, -21.8701415, -21.5062334, -21.1334077, -20.7524148, -20.3640049, -19.9689284, -19.5765121, -19.1830376, -18.7846837, -18.3819226, -17.9752261, -17.5650664, -17.1519155, -16.7362454, -16.3185282, -15.8992359, -15.4788406, -15.0578142, -14.6366288, -14.2157565, -13.7956692, -13.3768391, -12.9597382, -12.577482, -12.1971914, -11.8186307, -11.4415645, -11.065757, -10.6909726, -10.3169758, -9.94353092, -9.57040236, -9.19735452, -8.82415178, -8.45055853, -8.07633916, -7.70125806, -7.32507962, -6.94756823, -6.54062216, -6.07680875, -5.61336415, -5.15185921, -4.69386476, -4.24095163, -3.79469066, -3.3566527, -2.92840857, -2.51152912, -2.10758517, -1.71814757, -1.34478716, -0.989074771, -0.652581237, -0.336877398, -0.0435340899, -0.258713154, -0.731497205, -1.20514445, -1.66549981, -2.09840822, -2.4897146, -2.82526387, -3.09090098, -3.27247083, -3.35581837, -3.32678851, -3.17122619, -2.87497632, -2.42388384, -1.80379368, -1.00055076, 0, 3.59449458, 6.96520846, 10.1560074, 13.2107571, 16.1733234, 19.0875719, 21.9973685, 24.9465788, 27.9790687, 31.1387038, 34.46935, 38.0148729, 41.8191383, 45.926012, 50.3793597, 55.2230471, 60.4164949, 65.9915053, 72.1829507, 79.0166016, 86.5182283, 94.7136012, 103.628491, 113.288667, 123.719901, 134.947962, 146.998621, 159.897649, 173.670815, 188.327203, 203.915388, 220.466149, 238.010262, 257.627247, 278.823508, 300.825787, 323.350238, 346.113015, 368.830272, 391.218163, 417.116526, 440.862798, 461.366324, 477.536451, 488.282525, 504.545956, 510.196011, 501.13402, 473.261314, 422.479223, 352.020301, 241.494272, 97.0987181, -88.4654295, -350.052909, -710.146527, -1191.22909, -1800.21883, -2578.70053, -3585.23849, -4825.19878, -6411.10712, -8353.10777, -10735.8379, -13770.159, -17608.9933, -22278.1898, -27897.726, -35053.5154, -44067.9117, -54424.3804, -67413.2232, -83647.1782, -103738.984, -128301.377, -175083.412, -223277.434, -269212.1, -309216.072, -339618.009, -343567.904, -339358.86, -328967.537, -319237.617, -310219.229, -299934.576, -288535.182, -276172.574, -262998.275, -249163.811, -226075.025, -209287.792, -196667.81, -186080.778, -175392.394, -162468.358, -145174.368, -131513.297, -117266.668},
        {-22.3891136, -22.6340567, -22.8453062, -23.0237972, -23.1704645, -23.286243, -23.3720677, -23.4288735, -23.4575953, -23.459168, -23.4345266, -23.3846059, -23.3103409, -23.2126666, -23.0925177, -22.9508294, -22.7885363, -22.6065736, -22.4058761, -22.1873787, -21.9520164, -21.700724, -21.4344365, -21.1540889, -20.8606159, -20.5549527, -20.2380339, -19.9107947, -19.5741699, -19.2290944, -18.8765032, -18.5173312, -18.1525132, -17.7829843, -17.4342498, -17.0944427, -16.7514004, -16.4052605, -16.0561606, -15.7042382, -15.3496308, -14.9924761, -14.6329115, -14.2710747, -13.9071032, -13.5411346, -13.1733064, -12.8037561, -12.4326214, -12.0600398, -11.6861488, -11.312232, -10.9372561, -10.5613338, -10.184578, -9.80710114, -9.42901611, -9.05043559, -8.67147229, -8.29223893, -7.91284823, -7.5334129, -7.15404566, -6.77485923, -6.39596632, -6.01747964, -5.63951192, -5.24185795, -4.80480032, -4.37035598, -3.93995475, -3.51502649, -3.09700103, -2.68730824, -2.28737794, -1.89863999, -1.52252423, -1.1604605, -0.813878653, -0.484208531, -0.17287998, 0.118677156, 0.389033032, 0.636757802, 0.366104979, -0.166799677, -0.699889096, -1.2185516, -1.7081755, -2.15414913, -2.5418608, -2.85669883, -3.08405154, -3.20930725, -3.21785427, -3.09508094, -2.82637556, -2.39712646, -1.79272195, -0.998550358, 0, 3.62066201, 7.01926823, 10.2408824, 13.3305681, 16.3333891, 19.2944091, 22.2586917, 25.2713007, 28.3772997, 31.6217525, 35.0497226, 38.7062738, 42.6364699, 46.8853743, 51.498051, 56.5195634, 61.8855312, 67.6253829, 74.0317349, 81.1261985, 88.9303851, 97.465906, 106.754373, 116.817396, 127.676588, 139.353559, 151.869922, 165.247286, 179.507265, 194.624085, 210.684115, 227.723182, 245.777112, 265.94862, 287.740089, 310.369396, 333.5523, 357.004562, 380.441942, 403.580198, 430.682785, 455.533688, 476.958898, 493.784409, 504.836213, 520.938778, 525.920004, 515.606267, 485.823942, 432.399403, 358.038604, 243.279967, 97.2907357, -86.6971845, -347.986531, -711.272649, -1201.25089, -1830.78146, -2639.08465, -3678.29177, -4950.64756, -6569.5334, -8541.77137, -10951.8537, -14015.8804, -17889.5412, -22616.1996, -28322.4238, -35546.6456, -44594.3233, -54870.8435, -67720.2092, -83759.4424, -103605.565, -127875.598, -174175.506, -221886.134, -267377.267, -307018.693, -337180.198, -341259.08, -337245.939, -327070.2, -317273.935, -307960.729, -297312.228, -285528.113, -272808.065, -259351.766, -245358.896, -224724.058, -209090.99, -196672.933, -185683.129, -174334.818, -160841.241, -143415.641, -130390.967, -116127.826},
        {-19.6174481, -19.9161147, -20.1764593, -20.3996467, -20.5868418, -20.7392094, -20.8579143, -20.9441214, -20.9989953, -21.023701, -21.0194033, -20.987267, -20.9284568, -20.8441376, -20.7354743, -20.6036316, -20.4497744, -20.2750674, -20.0806755, -19.8677635, -19.6374962, -19.3910384, -19.129555, -18.8542107, -18.5661704, -18.2665989, -17.956661, -17.6375214, -17.3103451, -16.9762969, -16.6365415, -16.2922438, -15.9445686, -15.5946806, -15.2868764, -14.9998478, -14.711768, -14.4224023, -14.1315157, -13.8388733, -13.5442403, -13.2473817, -12.9480627, -12.6460483, -12.3411038, -12.0329941, -11.7214845, -11.4063399, -11.0873256, -10.7642066, -10.4367481, -10.0728761, -9.704885, -9.33323006, -8.95836665, -8.58075006, -8.20083561, -7.81907864, -7.43593445, -7.05185836, -6.66730569, -6.28273176, -5.89859189, -5.51534139, -5.13343558, -4.75332979, -4.37547932, -3.9874162, -3.5769827, -3.1712875, -2.7716237, -2.37928436, -1.99556258, -1.62175144, -1.25914402, -0.90903341, -0.572712689, -0.251474941, 0.0533867493, 0.340579299, 0.608809623, 0.856784639, 1.08321126, 1.28679641, 0.962134413, 0.37059181, -0.220405816, -0.795792125, -1.34050078, -1.83946543, -2.27761976, -2.63989741, -2.91123205, -3.07655733, -3.12080693, -3.02891449, -2.78581369, -2.37643818, -1.78572162, -0.998597671, 0, 3.62136345, 7.02831841, 10.2666363, 13.3820886, 16.4204466, 19.4274818, 22.4489657, 25.5306696, 28.718365, 32.0578232, 35.5948157, 39.375114, 43.4444894, 47.8487134, 52.6335574, 57.8447928, 63.4107506, 69.3604251, 76.0052266, 83.3657609, 91.4626334, 100.31645, 109.947816, 120.377337, 131.625619, 143.713268, 156.660888, 170.489086, 185.218468, 200.820515, 217.382968, 234.941171, 253.530465, 274.243349, 296.586588, 319.785038, 343.556895, 367.620355, 391.693613, 415.494865, 443.778993, 469.694603, 491.974451, 509.351296, 520.557895, 535.871285, 539.593874, 527.572348, 495.653394, 439.683701, 362.388761, 244.329027, 96.5446092, -87.5059743, -350.259965, -718.038139, -1217.16128, -1865.11897, -2699.07027, -3765.80825, -5064.02749, -6709.52613, -8708.66029, -11146.5299, -14241.6617, -18151.3375, -22932.9263, -28717.8063, -36003.7027, -45081.1142, -55288.1192, -68014.1627, -83879.4273, -103504.096, -127508.35, -173370.895, -220638.762, -265717.501, -305012.666, -334929.808, -339080.416, -335193.482, -325157.726, -315240.32, -305596.848, -294558.544, -282372.211, -269284.652, -255542.671, -241393.071, -223687.797, -209322.975, -196988.373, -185373.763, -173168.914, -159063.597, -141747.581, -129139.624, -115137.506},
        {-17.4163321, -17.7396419, -18.0226609, -18.2666624, -18.4729195, -18.6427055, -18.7772935, -18.8779568, -18.9459685, -18.9826018, -18.98913, -18.9668262, -18.9169636, -18.8408154, -18.7396548, -18.614755, -18.4673893, -18.2988307, -18.1103525, -17.9032279, -17.6787301, -17.4381322, -17.1827076, -16.9137293, -16.6324705, -16.3402045, -16.0382045, -15.7277437, -15.4100951, -15.0865321, -14.7583279, -14.4267556, -14.0930884, -13.7585995, -13.4757131, -13.219051, -12.9626202, -12.706034, -12.4489056, -12.1908483, -11.9314753, -11.6703999, -11.4072354, -11.1415949, -10.8730918, -10.6013392, -10.3259505, -10.0465389, -9.76271762, -9.47409991, -9.18029904, -8.83851698, -8.49169766, -8.1403737, -7.78507776, -7.42634247, -7.06470048, -6.70068444, -6.33482699, -5.96766076, -5.5997184, -5.23153257, -4.86363589, -4.49656101, -4.13084058, -3.76700723, -3.40559362, -3.03500505, -2.643938, -2.25793682, -1.87832034, -1.50640736, -1.14351669, -0.790967141, -0.450077516, -0.122166627, 0.191446717, 0.489443707, 0.770505533, 1.03331339, 1.27654846, 1.49889194, 1.69902503, 1.8756289, 1.50580885, 0.860004677, 0.2149251, -0.413822546, -1.01063093, -1.55989271, -2.04600057, -2.45334716, -2.76632517, -2.96932725, -3.04674607, -2.9829743, -2.76240461, -2.36942967, -1.78844214, -1.0038347, 0, 3.70880767, 7.20003293, 10.5205147, 13.7170917, 16.836603, 19.9258873, 23.0317836, 26.2011306, 29.4807673, 32.9175326, 36.5582652, 40.4498041, 44.6389881, 49.1726561, 54.097647, 59.4607996, 65.2289813, 71.4376065, 78.3144019, 85.8890696, 94.1913117, 103.25083, 113.097328, 123.760506, 135.270067, 147.655713, 160.947147, 175.174069, 190.366184, 206.596086, 223.834856, 242.099327, 261.406333, 282.79331, 305.766791, 329.561147, 353.887034, 378.455103, 402.976008, 427.160403, 456.106326, 482.497406, 504.990244, 522.241439, 532.907591, 545.989126, 547.212861, 532.64944, 498.369505, 440.443701, 363.404818, 244.518727, 91.8378066, -102.250602, -377.341944, -758.672048, -1271.47674, -1930.83797, -2774.45314, -3851.09662, -5156.27858, -6813.74597, -8845.07705, -11338.7949, -14496.8569, -18469.1521, -23296.705, -29101.1804, -36403.4144, -45498.4492, -55842.1929, -68697.9195, -84651.1408, -104287.369, -128192.114, -173401.436, -219937.662, -264273.669, -302882.333, -332236.528, -336237.257, -332310.517, -322317.129, -312428.976, -302800.416, -291789.535, -279640.503, -266597.496, -252904.686, -238806.245, -220655.474, -206131.226, -193879.303, -182545.511, -170775.653, -157215.533, -140510.955, -127132.797, -114461.253},
        {-15.5373657, -15.8716708, -16.1646688, -16.4177006, -16.6321075, -16.8092303, -16.9504102, -17.0569883, -17.1303055, -17.1717029, -17.1825217, -17.1641027, -17.1177872, -17.044916, -16.9468304, -16.8248714, -16.6803799, -16.5146971, -16.3291639, -16.1251216, -15.903911, -15.6668734, -15.4153496, -15.1506808, -14.874208, -14.5872724, -14.2912148, -13.9873764, -13.6770983, -13.3617214, -13.0425869, -12.7210358, -12.3984092, -12.076048, -11.8109947, -11.5755681, -11.3414174, -11.1080761, -10.8750777, -10.6419555, -10.4082431, -10.1734739, -9.93718125, -9.69889868, -9.45815959, -9.21449744, -8.96744566, -8.71653769, -8.46130696, -8.20128692, -7.93601101, -7.61932304, -7.29743654, -6.97087541, -6.64016355, -6.30582486, -5.96838324, -5.6283626, -5.28628684, -4.94267986, -4.59806555, -4.25296783, -3.90791058, -3.56341772, -3.22001315, -2.87822076, -2.53856445, -2.18824639, -1.81478858, -1.44619002, -1.0838382, -0.729120625, -0.383424787, -0.0481381838, 0.27535169, 0.585657339, 0.881391267, 1.16116598, 1.42359398, 1.66728777, 1.89085985, 2.09292273, 2.27208892, 2.42697092, 2.0154081, 1.31777396, 0.620941685, -0.0589270867, -0.705670727, -1.3031276, -1.83513608, -2.28553454, -2.63816133, -2.87685482, -2.9854534, -2.94779541, -2.74771924, -2.36906324, -1.79566579, -1.01136526, 0, 3.80918371, 7.39682366, 10.8107479, 14.0987845, 17.3087615, 20.488507, 23.6858489, 26.9486154, 30.3246345, 33.8617342, 37.6077426, 41.6104877, 45.9177977, 50.5775004, 55.6374241, 61.1453967, 67.1168284, 73.5949161, 80.700814, 88.4754036, 96.9595661, 106.194183, 116.220135, 127.078305, 138.809573, 151.454821, 165.054929, 179.650781, 195.283256, 212.149913, 230.077508, 249.059921, 269.091028, 291.1432, 314.72107, 339.068338, 363.885337, 388.872396, 413.729846, 438.158019, 467.549386, 494.179745, 516.635749, 533.504048, 543.371294, 553.806974, 552.169193, 534.798893, 498.037014, 438.224498, 362.115665, 242.910615, 84.416259, -122.367801, -412.590216, -809.659603, -1336.98458, -2005.39575, -2853.73839, -3934.57924, -5240.33371, -6903.94909, -8965.47278, -11517.5178, -14741.767, -18779.0184, -23648.123, -29458.2631, -36765.7613, -45874.0125, -56393.0005, -69426.2098, -85517.0618, -105208.978, -129045.381, -173554.199, -219298.22, -262824.739, -300681.049, -329414.447, -333235.971, -329253.342, -319300.823, -309488.318, -299957.647, -289069.112, -277053.61, -264142.036, -250565.286, -236554.257, -217370.51, -202393.615, -190159.795, -179205.275, -168066.277, -155279.026, -139379.746, -124953.388, -113804.7},
        {-13.9407675, -14.2723178, -14.5626476, -14.8131183, -15.0250912, -15.1999277, -15.3389891, -15.4436366, -15.5152317, -15.5551356, -15.5647096, -15.5453151, -15.4983135, -15.4250659, -15.3269338, -15.2052784, -15.0614611, -14.8968432, -14.7127861, -14.510651, -14.2917993, -14.0575922, -13.8093912, -13.5485575, -13.2764525, -12.9944374, -12.7038737, -12.4061225, -12.1025453, -11.7945034, -11.4833581, -11.1704706, -10.8572024, -10.5449148, -10.291654, -10.0692466, -9.84883787, -9.62994952, -9.41210336, -9.19482117, -8.97762474, -8.76003585, -8.54157628, -8.32176783, -8.10013226, -7.87619138, -7.64946695, -7.41948078, -7.18575463, -6.9478103, -6.70516956, -6.41442489, -6.11895801, -5.81922132, -5.51566725, -5.20874818, -4.89891654, -4.58662473, -4.27232516, -3.95647023, -3.63951236, -3.32190395, -3.00409741, -2.68654514, -2.36969956, -2.05401307, -1.73993809, -1.41210142, -1.05550986, -0.703254905, -0.356814866, -0.0176680615, 0.312707191, 0.632832576, 0.941229777, 1.23642048, 1.51692636, 1.78126911, 2.02797041, 2.25555195, 2.4625354, 2.64744245, 2.80879479, 2.9451141, 2.49434378, 1.74708567, 1.00067932, 0.271838257, -0.422723998, -1.06629391, -1.64215797, -2.13360262, -2.52391436, -2.79637966, -2.93428498, -2.92091681, -2.7395616, -2.37350585, -1.80603601, -1.02043857, 0, 3.91477444, 7.60377597, 11.1157632, 14.4994947, 17.8037291, 21.0772249, 24.3687408, 27.7270353, 31.2008671, 34.8389948, 38.6901768, 42.8031719, 47.2267386, 52.0096355, 57.2006213, 62.8484544, 69.0144169, 75.7490566, 83.072805, 91.0371046, 99.6933973, 109.093125, 119.287731, 130.328657, 142.267344, 155.155236, 169.043774, 183.984401, 200.028558, 217.496523, 236.07233, 255.726771, 276.430639, 299.099638, 323.232103, 348.074393, 373.313825, 398.637718, 423.733388, 448.288155, 477.927602, 504.593482, 526.811278, 543.106471, 552.004545, 559.754207, 555.226909, 535.017327, 495.720139, 433.930021, 358.592698, 238.856072, 74.0092894, -146.372727, -452.548262, -865.870895, -1407.69421, -2084.06793, -2935.12763, -4017.704, -5321.3116, -6988.00083, -9076.07377, -11682.248, -14968.3339, -19066.2634, -23973.1569, -29785.9952, -37096.0667, -46215.0517, -56909.2191, -70119.5694, -86346.5789, -106090.724, -129852.48, -173629.599, -218550.963, -261242.73, -298331.058, -326442.104, -330120.667, -326128.503, -316269.497, -306584.892, -297201.284, -286479.7, -274635.015, -261882.101, -248435.831, -234511.08, -214135.851, -198596.67, -186317.061, -175720.55, -165230.661, -153270.919, -138264.848, -122754.603, -113036.89},
        {-12.5705434, -12.8833507, -13.1569747, -13.3927039, -13.5918267, -13.7556317, -13.8854074, -13.9824422, -14.0480247, -14.0834433, -14.0899864, -14.0689427, -14.0216006, -13.9492486, -13.8531751, -13.7346686, -13.5950177, -13.4355109, -13.2574365, -13.0620832, -12.8507394, -12.6246935, -12.3852341, -12.1336497, -11.8712287, -11.5992596, -11.319031, -11.0318313, -10.738949, -10.4416726, -10.1412905, -9.83909134, -9.53636351, -9.23439552, -8.98569251, -8.76485736, -8.54587738, -8.328379, -8.1119886, -7.89633262, -7.68103745, -7.46572951, -7.2500352, -7.03358094, -6.81599314, -6.5968982, -6.37592253, -6.15269256, -5.92683468, -5.6979753, -5.46574084, -5.19720356, -4.92524973, -4.65021147, -4.37242089, -4.09221011, -3.80991126, -3.52585646, -3.24037782, -2.95380746, -2.66647751, -2.37872008, -2.0908673, -1.80325128, -1.51620415, -1.23005802, -0.945145012, -0.643015259, -0.305669835, 0.0278226925, 0.355912642, 0.677050335, 0.989686091, 1.29227023, 1.58325307, 1.86108493, 2.12421614, 2.37109701, 2.60017786, 2.80990902, 2.9987408, 3.16512352, 3.3075075, 3.42434307, 2.93558434, 2.14109761, 1.34770228, 0.572621994, -0.166919628, -0.853698947, -1.47049233, -2.00007614, -2.42522674, -2.7287205, -2.89333379, -2.90184296, -2.7370244, -2.38165445, -1.81850949, -1.03036589, 0, 4.06959344, 7.89791424, 11.5355426, 15.0330587, 18.4410426, 21.8100747, 25.190735, 28.6336038, 32.1892612, 35.9082874, 39.8412626, 44.0387669, 48.5513806, 53.4296838, 58.7242567, 64.4856795, 70.7569698, 77.5876274, 85.0453382, 93.1790621, 102.037758, 111.670387, 122.125907, 133.453279, 145.701462, 158.919416, 173.1561, 188.460474, 204.881497, 222.744232, 241.720299, 261.775825, 282.876941, 305.987221, 330.574072, 355.838913, 381.448638, 407.070143, 432.370321, 457.016067, 486.7243, 513.27058, 535.138101, 550.810058, 558.769645, 564.918449, 558.466674, 536.042918, 494.275778, 429.793851, 352.34242, 229.401889, 58.42363, -171.645193, -489.22017, -914.921371, -1469.36886, -2157.72126, -3019.57406, -4111.38977, -5418.71468, -7090.33765, -9192.32151, -11824.313, -15142.8647, -19278.4086, -24230.6002, -30095.9666, -37450.5476, -46599.1435, -57340.1919, -70562.7577, -86746.9204, -106372.76, -129920.355, -172976.359, -217137.635, -259107.621, -295589.751, -323287.463, -327137.989, -323455.107, -313984.91, -304527.149, -295223.346, -284526.006, -272661.104, -259854.614, -246332.515, -232320.78, -211890.114, -196224.732, -183825.278, -173192.397, -162826.731, -151228.926, -136899.625, -120988.816, -111615.309},
        {-11.3905937, -11.6690525, -11.9123081, -12.1214848, -12.2977067, -12.4420981, -12.5557832, -12.6398862, -12.6955313, -12.7238427, -12.7259447, -12.7029615, -12.6560173, -12.5862363, -12.4947427, -12.3826607, -12.2511147, -12.1012287, -11.934127, -11.7509338, -11.5527734, -11.3407699, -11.1160476, -10.8797307, -10.6329434, -10.37681, -10.1124545, -9.84100139, -9.56357472, -9.28129876, -8.99529771, -8.7066958, -8.41661725, -8.12618627, -7.8763586, -7.64750491, -7.41951709, -7.19222676, -6.96546553, -6.739065, -6.51285679, -6.28667249, -6.06034373, -5.83370211, -5.60657924, -5.37880672, -5.15021617, -4.92063919, -4.6899074, -4.4578524, -4.22430579, -3.97263695, -3.71949659, -3.4650732, -3.20955525, -2.95313121, -2.69598958, -2.43831882, -2.18030742, -1.92214385, -1.66401659, -1.40611412, -1.14862491, -0.891737444, -0.635640202, -0.380521659, -0.126570294, 0.147575369, 0.462759399, 0.774348157, 1.08075617, 1.38039796, 1.67168806, 1.95304099, 2.22287128, 2.47959346, 2.72162204, 2.94737157, 3.15525656, 3.34369154, 3.51109104, 3.65586958, 3.77644169, 3.87122189, 3.34491288, 2.50527304, 1.66719668, 0.848365438, 0.06646092, -0.660835254, -1.31584147, -1.8808761, -2.33825754, -2.67030417, -2.85933437, -2.88766652, -2.73761901, -2.39151022, -1.83165854, -1.04038233, 0, 4.2541764, 8.24344603, 12.0209101, 15.6396699, 19.1528267, 22.6134816, 26.0747359, 29.5896908, 33.2114475, 36.9931074, 40.9877715, 45.2485412, 49.8285177, 54.7808022, 60.1584959, 66.0147001, 72.3084334, 79.0793566, 86.5933768, 94.8834349, 103.982471, 113.923427, 124.739242, 136.462858, 149.127215, 162.765254, 177.409915, 193.094139, 209.850867, 227.88627, 246.99654, 267.162649, 288.365568, 311.722031, 336.645128, 362.24508, 388.162132, 414.036526, 439.508505, 464.218314, 493.835234, 520.135544, 541.579894, 556.628934, 563.743311, 569.579599, 562.353542, 538.476808, 494.361062, 426.417974, 343.564506, 214.422781, 37.825192, -196.958169, -520.180243, -953.328528, -1517.89052, -2222.63256, -3104.65447, -4214.93373, -5534.39535, -7215.06185, -9318.3497, -11945.1291, -15263.1225, -19409.3661, -24412.6025, -30381.7918, -37822.9701, -47017.5647, -57658.9864, -70707.1434, -86647.0072, -105963.55, -129141.742, -171497.463, -214977.051, -256357.75, -292416.806, -319931.464, -324286.054, -321246.012, -312471.259, -303335.868, -294029.577, -283194.559, -271097.931, -258006.808, -244188.305, -229909.538, -210791.895, -195537.742, -182947.536, -171821.731, -160960.784, -149165.15, -135235.285, -119741.297, -109502.457},
        {-10.3900562, -10.6222681, -10.8241083, -10.9965067, -11.1403934, -11.2566984, -11.3463518, -11.4102836, -11.4494238, -11.4647024, -11.4570495, -11.427395, -11.3766691, -11.3058017, -11.2157228, -11.1073625, -10.9816508, -10.8395177, -10.6818933, -10.5097075, -10.3238905, -10.1253721, -9.91508255, -9.69395174, -9.46290973, -9.22288654, -8.97481219, -8.71961672, -8.45823014, -8.19158249, -7.92060377, -7.64622403, -7.36937328, -7.09098155, -6.83938347, -6.60123033, -6.36284407, -6.12426538, -5.88553494, -5.64669345, -5.40778161, -5.1688401, -4.92990962, -4.69103086, -4.45224451, -4.21359127, -3.97511182, -3.73684687, -3.4988371, -3.2611232, -3.02374587, -2.78578642, -2.54826573, -2.31124527, -2.07478654, -1.83895104, -1.60380024, -1.36939565, -1.13579875, -0.903071031, -0.671273985, -0.4404691, -0.210717868, 0.0179182222, 0.24537768, 0.471599014, 0.696520737, 0.943843681, 1.23669816, 1.52601467, 1.8101913, 2.08762616, 2.35671733, 2.61586291, 2.86346099, 3.09790968, 3.31760706, 3.52095124, 3.7063403, 3.87217234, 4.01684545, 4.13875774, 4.23630729, 4.3078922, 3.74436919, 2.86068587, 1.97909199, 1.11770156, 0.294628537, -0.472013083, -1.16410932, -1.7635462, -2.25220974, -2.61198595, -2.82476086, -2.87242049, -2.73685085, -2.39993796, -1.84356784, -1.04962652, 0, 4.3685315, 8.46146043, 12.3334692, 16.0392401, 19.6334556, 23.170798, 26.7059497, 30.2935931, 33.9884105, 37.8450843, 41.918297, 46.2627308, 50.9330681, 55.9839913, 61.4701829, 67.446325, 73.7960516, 80.5496094, 88.1485766, 96.6109822, 105.954855, 116.198225, 127.359121, 139.455571, 152.505605, 166.527252, 181.538541, 197.557502, 214.602163, 232.769861, 251.970238, 272.197531, 293.445975, 316.970492, 342.134976, 367.96714, 394.083015, 420.098633, 445.630025, 470.293223, 499.679361, 525.610858, 546.534705, 560.89789, 567.147405, 572.992887, 565.303016, 540.209118, 493.842522, 422.334554, 333.337983, 197.380937, 15.3727197, -222.851974, -550.069111, -988.902934, -1561.97768, -2281.9417, -3183.66238, -4312.88992, -5647.8153, -7340.47998, -9443.94704, -12058.9112, -15366.7763, -19511.8993, -24553.6286, -30614.2819, -38127.8758, -47352.1123, -57868.7656, -70715.4547, -86383.8566, -105365.648, -128152.507, -169799.019, -212593.224, -253386.074, -289028.518, -316371.506, -321247.694, -318871.857, -310817.07, -302032.599, -292755.113, -281814.535, -269517.871, -256172.127, -242084.31, -227561.424, -210010.258, -195242.373, -182410.547, -170667.558, -159166.184, -147059.203, -133499.393, -118631.607, -107301.487},
        {-9.54228205, -9.71797593, -9.86856741, -9.99478649, -10.0973632, -10.1770274, -10.2345091, -10.2705384, -10.2858453, -10.2811596, -10.2572114, -10.2147307, -10.1544474, -10.0770915, -9.9833931, -9.87408208, -9.74988844, -9.61154217, -9.45977326, -9.29531169, -9.11888744, -8.9312305, -8.73307085, -8.52513847, -8.30816335, -8.08287547, -7.85000482, -7.61028138, -7.36443513, -7.11319606, -6.85729415, -6.59745939, -6.33442176, -6.06891124, -5.81775408, -5.57329355, -5.32767951, -5.08111956, -4.83382135, -4.58599251, -4.33784066, -4.08957343, -3.84139845, -3.59352336, -3.34615577, -3.09950333, -2.85377366, -2.6091744, -2.36591316, -2.12419758, -1.88423529, -1.65663973, -1.43098714, -1.20725957, -0.985439081, -0.765507717, -0.547447533, -0.331240579, -0.116868909, 0.0956854256, 0.306440373, 0.515413881, 0.722623898, 0.928088371, 1.13182525, 1.33385248, 1.53418801, 1.75773044, 2.02878101, 2.29604317, 2.55792196, 2.81282242, 3.05914959, 3.29530849, 3.51970418, 3.73074167, 3.92682602, 4.10636226, 4.26775541, 4.40941053, 4.52973264, 4.62712679, 4.699998, 4.74675132, 4.14582394, 3.218645, 2.29409519, 1.3906987, 0.526979705, -0.278537607, -1.00732905, -1.64087044, -2.16063759, -2.54810632, -2.78475244, -2.85205176, -2.73148011, -2.40451329, -1.85262712, -1.05729742, 0, 4.36303973, 8.46257219, 12.3531873, 16.0894748, 19.7260248, 23.317427, 26.9182713, 30.5831477, 34.3666459, 38.3233559, 42.5078676, 46.9747708, 51.7786554, 56.9741113, 62.6157284, 68.7580966, 75.24492, 82.100662, 89.8569158, 98.5230816, 108.108559, 118.622749, 130.075051, 142.474866, 155.831592, 170.154631, 185.453383, 201.737247, 219.015624, 237.333562, 256.651741, 276.968868, 298.28365, 321.929663, 347.238178, 373.1845, 399.366872, 425.383539, 450.832745, 475.312735, 504.348357, 529.807503, 550.128864, 563.751132, 569.112998, 575.054509, 567.012663, 540.825811, 492.332307, 417.370503, 322.191811, 179.494452, -7.64716538, -248.761385, -579.107165, -1022.39094, -1602.31915, -2334.49548, -3252.57065, -4399.03475, -5752.4424, -7461.03475, -9567.87552, -12171.868, -15466.5324, -19601.9583, -24661.674, -30778.5124, -38329.364, -47553.0665, -57953.8665, -70618.2069, -86039.207, -104709.986, -127123.663, -168047.304, -210131.595, -250301.169, -285480.658, -312594.695, -317902.962, -316105.009, -308696.164, -300263.946, -291079.688, -280126.309, -267739.56, -254255.197, -240008.971, -225336.636, -209333.255, -195016.921, -181902.083, -169503.189, -157334.686, -144911.022, -131746.645, -117570.625, -105214.801},
        {-8.81868184, -8.92709312, -9.0156235, -9.08483244, -9.1352794, -9.16752382, -9.18212517, -9.17964289, -9.16063644, -9.12566528, -9.07528886, -9.01006663, -8.93055804, -8.83732256, -8.73091963, -8.61190871, -8.48084926, -8.33830072, -8.18482256, -8.02097422, -7.84731516, -7.66440483, -7.4728027, -7.2730682, -7.06576081, -6.85143996, -6.63066512, -6.40399574, -6.17199128, -5.93521118, -5.6942149, -5.4495619, -5.20181163, -4.95152354, -4.70895228, -4.46960577, -4.22887542, -3.98700608, -3.74424255, -3.50082968, -3.25701228, -3.01303518, -2.76914322, -2.52558121, -2.28259398, -2.04042636, -1.79932318, -1.55952927, -1.32128944, -1.08484854, -0.850451374, -0.628931185, -0.409714858, -0.192817686, 0.0217450408, 0.233958032, 0.443805995, 0.65127364, 0.856345675, 1.05900681, 1.25924175, 1.45703521, 1.6523719, 1.84523652, 2.03561378, 2.2234884, 2.40884507, 2.61533778, 2.8660522, 3.11215697, 3.35210241, 3.58433885, 3.80731662, 4.01948604, 4.21929744, 4.40520114, 4.57564748, 4.72908678, 4.86396937, 4.97874558, 5.07186573, 5.14178014, 5.18693915, 5.20579309, 4.56651508, 3.59595109, 2.62852252, 1.68313339, 0.778687752, -0.0659103734, -0.831756949, -1.49994794, -2.05157931, -2.46774704, -2.72954707, -2.81807539, -2.71442795, -2.39970073, -1.85498969, -1.06139079, 0, 4.09656554, 7.99168142, 11.7361976, 15.380964, 18.9768305, 22.5746472, 26.2252639, 29.9795305, 33.8882971, 38.0024135, 42.3727296, 47.0500955, 52.0853611, 57.5293762, 63.4329908, 69.847055, 76.670924, 83.9338057, 92.0292163, 100.975543, 110.791171, 121.494489, 133.103884, 145.637741, 159.114448, 173.552392, 188.969959, 205.385536, 222.817511, 241.414084, 261.016229, 281.603389, 303.155005, 326.932581, 352.274529, 378.177379, 404.235953, 430.045078, 455.199577, 479.294274, 507.847911, 532.728463, 552.371728, 565.213503, 569.689585, 575.622295, 567.214276, 540.054692, 489.732708, 411.837492, 311.42386, 163.046197, -28.7701626, -273.052419, -606.49754, -1053.37933, -1637.9716, -2375.97861, -3302.19133, -4460.74935, -5836.16181, -7567.29383, -9688.58173, -12294.8061, -15582.9972, -19703.2513, -24742.1096, -30833.1911, -38343.1484, -47506.817, -57862.7775, -70449.2365, -85742.041, -104217.038, -126350.074, -166525.911, -207834.253, -247273.717, -281842.922, -308540.484, -313989.734, -312481.438, -305460.767, -297336.251, -288380.066, -277629.651, -265418.919, -252081.779, -237952.143, -223363.922, -208407.31, -194296.33, -180850.356, -167888.761, -155230.918, -142696.201, -130103.982, -116461.169, -103609.727},
        {-8.19190423, -8.2231281, -8.23920509, -8.2405818, -8.22770483, -8.20102077, -8.16097621, -8.10801775, -8.04259199, -7.96514552, -7.87612494, -7.77597683, -7.6651478, -7.54408444, -7.41323335, -7.27304111, -7.12395433, -6.9664196, -6.80088352, -6.62779267, -6.44759366, -6.26073308, -6.06765752, -5.86881358, -5.66464786, -5.45560694, -5.24213743, -5.02468592, -4.803699, -4.57962327, -4.35290533, -4.12399176, -3.89332917, -3.66136415, -3.43949443, -3.22246069, -3.004872, -2.78681957, -2.56839459, -2.34968829, -2.13079187, -1.91179654, -1.69279352, -1.473874, -1.2551292, -1.03665033, -0.818528605, -0.600855223, -0.383721399, -0.167218341, 0.0485627401, 0.268756376, 0.487932333, 0.705886118, 0.922413238, 1.1373092, 1.35036951, 1.56138967, 1.7701652, 1.9764916, 2.18016437, 2.38097902, 2.57873106, 2.773216, 2.96422934, 3.15156659, 3.33502326, 3.53373218, 3.76636191, 3.99282617, 4.21166689, 4.42142603, 4.62064553, 4.80786733, 4.98163339, 5.14048565, 5.28296606, 5.40761656, 5.5129791, 5.59759562, 5.66000808, 5.69875841, 5.71238857, 5.69944049, 5.020256, 4.00590079, 2.99511122, 2.00714007, 1.0612401, 0.17666407, -0.627335239, -1.33150506, -1.91659264, -2.36334519, -2.65250995, -2.76483416, -2.68106505, -2.38194984, -1.84823578, -1.06067009, 0, 3.48636544, 6.89942236, 10.2814507, 13.6747305, 17.1215417, 20.6641641, 24.3448779, 28.205963, 32.2896993, 36.6383668, 41.2942454, 46.2996152, 51.696756, 57.5279479, 63.8354708, 70.6616047, 78.0958255, 86.1871554, 94.8711216, 104.200118, 114.226537, 125.002772, 136.581218, 149.014267, 162.354312, 176.653747, 191.964966, 208.340362, 225.832327, 244.913121, 265.061321, 286.203363, 308.265678, 332.225522, 357.483918, 383.161672, 408.869974, 434.220012, 458.822973, 482.290048, 510.238573, 534.445629, 553.347288, 565.379624, 568.978709, 574.578556, 565.617812, 537.533062, 485.760893, 405.737892, 301.841399, 149.639661, -46.2833889, -294.773016, -632.092434, -1082.19476, -1669.0331, -2404.09335, -3326.62931, -4489.49537, -5890.42073, -7652.29734, -9804.7135, -12435.5984, -15731.7366, -19834.5364, -24801.981, -30753.8499, -38115.8152, -47140.5188, -57566.86, -70240.2608, -85591.1972, -104050.145, -126047.581, -165443.581, -205881.621, -244434.775, -278176.12, -304178.731, -309336.603, -307687.945, -300668.298, -292772.961, -284226.319, -273977.852, -262315.688, -249527.955, -235902.78, -221728.293, -206969.212, -192670.814, -178849.188, -165520.427, -152700.623, -140405.867, -128652.251, -115211.197, -102748.363},
        {-7.64652465, -7.60455436, -7.55168892, -7.48827854, -7.41467343, -7.3312238, -7.23827986, -7.13619184, -7.02530993, -6.90598436, -6.77856533, -6.64340307, -6.50084777, -6.35124965, -6.19495893, -6.03232581, -5.86370051, -5.68943325, -5.50987423, -5.32537366, -5.13628177, -4.94294875, -4.74572483, -4.54496021, -4.34100512, -4.13420975, -3.92492432, -3.71349905, -3.50028415, -3.28562983, -3.0698863, -2.85340378, -2.63653247, -2.41962259, -2.21562842, -2.01833302, -1.82136798, -1.62467448, -1.42819371, -1.23186685, -1.0356351, -0.83943963, -0.643221638, -0.446922307, -0.250482823, -0.053844372, 0.14305186, 0.340264685, 0.537852919, 0.735875374, 0.934390864, 1.15381639, 1.37341126, 1.59279295, 1.81157896, 2.02938676, 2.24583386, 2.46053773, 2.67311587, 2.88318576, 3.09036489, 3.29427075, 3.49452082, 3.6907326, 3.88252358, 4.06951123, 4.25131306, 4.44292662, 4.65898021, 4.86737106, 5.06671963, 5.25564639, 5.43277179, 5.5967163, 5.74610036, 5.87954445, 5.99566902, 6.09309454, 6.17044146, 6.22633024, 6.25938135, 6.26821524, 6.25145237, 6.20771321, 5.48786587, 4.42803159, 3.37211707, 2.33976359, 1.35061241, 0.424304807, -0.419517955, -1.1612146, -1.78114387, -2.25966449, -2.57713519, -2.7139147, -2.65036175, -2.36683508, -1.84369341, -1.06129547, 0, 3.01227122, 6.05511894, 9.1638637, 12.373826, 15.7203264, 19.2386855, 22.9642237, 26.9322616, 31.1781197, 35.7371186, 40.6445788, 45.9358208, 51.6461652, 57.8109324, 64.4654431, 71.6450177, 79.5980034, 88.3901533, 97.5749418, 107.233338, 117.446311, 128.294829, 139.859863, 152.22238, 165.463351, 179.663744, 194.904529, 211.266674, 228.831149, 248.35842, 269.00081, 290.635439, 313.139425, 337.218861, 362.336381, 387.733559, 413.038823, 437.8806, 461.887318, 484.687406, 512.109488, 535.69478, 553.858626, 565.016372, 567.583365, 572.041446, 561.722841, 532.026272, 478.350461, 396.094128, 289.525016, 134.332914, -65.7134526, -319.541078, -661.99738, -1116.29203, -1705.63472, -2435.98003, -3351.71433, -4516.04769, -5940.98472, -7732.90623, -9916.85763, -12573.8552, -15879.7546, -19966.8788, -24864.4642, -30678.1077, -37891.3893, -46774.0802, -57257.8473, -70002.5825, -85395.0932, -103822.187, -125670.671, -164287.895, -203859.988, -241533.623, -274455.472, -299772.205, -304646.153, -302864.558, -295853.097, -288189.187, -280050.939, -270303.239, -259189.858, -246954.57, -233841.146, -220093.359, -205626.366, -191226.544, -177087.982, -163404.772, -150371.005, -138180.773, -127028.167, -113778.228, -101878.952},
        {-7.13641705, -7.04101452, -6.93791805, -6.82739194, -6.70970053, -6.58510812, -6.45387903, -6.31627758, -6.17256808, -6.02301485, -5.8678822, -5.70743446, -5.54193594, -5.37165095, -5.19684381, -5.01777884, -4.83472036, -4.64793267, -4.4576801, -4.26422697, -4.06783758, -3.86877626, -3.66730732, -3.46369508, -3.25820385, -3.05109796, -2.84264171, -2.63309943, -2.42273542, -2.21181402, -2.00059952, -1.78935625, -1.57834853, -1.36784067, -1.1688385, -0.976009635, -0.783892613, -0.592403164, -0.401457023, -0.210969923, -0.0208575979, 0.168964218, 0.358579791, 0.548073388, 0.737529274, 0.927031716, 1.11666498, 1.30651333, 1.49666104, 1.68719236, 1.87819158, 2.09363679, 2.30920046, 2.52444886, 2.73894831, 2.95226508, 3.16396547, 3.37361579, 3.58078231, 3.78503135, 3.98592918, 4.18304211, 4.37593643, 4.56417843, 4.74733442, 4.92497067, 5.09665349, 5.27625984, 5.47732308, 5.66989484, 5.8526137, 6.02411825, 6.1830471, 6.32803883, 6.45773203, 6.5707653, 6.66577723, 6.7414064, 6.79629142, 6.82907087, 6.83838334, 6.82286744, 6.78116174, 6.71190485, 5.95072499, 4.84171148, 3.73692248, 2.65651254, 1.62063623, 0.649448085, -0.236897334, -1.01824547, -1.67444179, -2.18533171, -2.5307607, -2.6905742, -2.64461766, -2.37273652, -1.85477623, -1.07058224, 0, 3.12297245, 6.27324294, 9.48661891, 12.7989078, 16.245917, 19.863454, 23.6873262, 27.753341, 32.0973058, 36.7550282, 41.7623155, 47.1549751, 52.9688145, 59.2396411, 66.0032624, 73.2954858, 81.3652747, 90.2788877, 99.5899931, 109.380075, 119.730616, 130.723102, 142.439015, 154.959839, 168.367058, 182.742156, 198.166616, 214.721923, 232.489559, 252.304587, 273.218601, 295.08701, 317.765226, 341.830329, 366.776894, 391.907051, 416.85971, 441.27378, 464.78817, 487.04179, 514.40527, 537.737013, 555.358854, 565.592627, 566.760166, 567.789436, 553.744606, 520.295976, 463.113846, 377.868517, 270.285226, 113.689969, -91.0047096, -353.504521, -705.194105, -1167.72471, -1762.74757, -2488.89434, -3396.06758, -4559.28215, -6004.24368, -7822.83973, -10034.3494, -12714.2888, -16033.4645, -20118.9432, -24989.015, -30742.5454, -37877.2472, -46665.2226, -57115.7285, -69804.1502, -85093.5749, -103347.09, -124927.782, -162768.347, -201519.862, -238403.011, -270638.48, -295446.954, -300323.102, -298697.635, -291959.531, -284588.985, -276751.367, -267324.041, -256536.778, -244619.349, -231801.525, -218313.078, -205018.233, -191010.155, -176703.538, -162513.081, -148853.48, -136139.431, -124785.631, -111914.45, -100609.597},
        {-6.67630935, -6.54020661, -6.39882098, -6.25234282, -6.10096249, -5.94487036, -5.78425676, -5.61931208, -5.45022665, -5.27719085, -5.10039504, -4.92002956, -4.73628478, -4.54935106, -4.35941876, -4.16667823, -3.97131984, -3.77353394, -3.57351089, -3.37144105, -3.16751478, -2.96192244, -2.75485439, -2.54650098, -2.33705258, -2.12669954, -1.91563222, -1.70404099, -1.49211619, -1.28004819, -1.06802734, -0.856244017, -0.644888564, -0.434151344, -0.231117459, -0.0323848746, 0.165531314, 0.362664494, 0.559048055, 0.754715387, 0.949699877, 1.14403492, 1.33775389, 1.53089019, 1.72347721, 1.91554832, 2.10713694, 2.29827643, 2.48900019, 2.67934161, 2.86933408, 3.07920814, 3.28836218, 3.49639177, 3.70289246, 3.90745981, 4.10968936, 4.30917668, 4.50551731, 4.69830682, 4.88714074, 5.07161465, 5.2513241, 5.42586463, 5.5948318, 5.75782117, 5.91442829, 6.07942228, 6.26720794, 6.44608649, 6.61466985, 6.77156993, 6.91539865, 7.04476792, 7.15828965, 7.25457576, 7.33223816, 7.38988877, 7.4261395, 7.43960226, 7.42888897, 7.39261155, 7.3293819, 7.23781194, 6.43455919, 5.27355676, 4.11693431, 2.98543449, 1.89979995, 0.880773332, -0.0509027195, -0.874485558, -1.56923254, -2.11440101, -2.48924834, -2.67303187, -2.64500896, -2.38443696, -1.87057323, -1.08267513, 0, 3.56542232, 7.09370925, 10.6253892, 14.2009906, 17.8610417, 21.6460711, 25.5966071, 29.7531782, 34.1563126, 38.8465389, 43.8643854, 49.2503804, 55.0450526, 61.2889301, 68.0225414, 75.286415, 83.2310511, 91.9126885, 101.123109, 110.926408, 121.386677, 132.568012, 144.534505, 157.350251, 171.079344, 185.785877, 201.533944, 218.387639, 236.411055, 256.370237, 277.369946, 299.263693, 321.904985, 345.857987, 370.620879, 395.502337, 420.142674, 444.182202, 467.261235, 489.020085, 516.441319, 539.588397, 556.665105, 565.875227, 565.422549, 562.13021, 543.427802, 505.364271, 443.988565, 355.34963, 247.168866, 89.556433, -120.049675, -392.363184, -754.740905, -1227.06724, -1829.22661, -2552.84225, -3452.48271, -4613.5469, -6072.62609, -7910.69962, -10141.7809, -12837.5295, -16167.2839, -20255.5463, -25121.2599, -30861.56, -37960.6308, -46683.7198, -57044.4621, -69598.427, -84697.9586, -102695.401, -123943.098, -161035.195, -199021.784, -235194.76, -266846.016, -291267.446, -296310.911, -295001.693, -288677.414, -281615.861, -273990.842, -264751.093, -254131.676, -242367.649, -229694.073, -216346.008, -204495.994, -191093.187, -176752.332, -162088.17, -147715.446, -134248.902, -122303.283, -109820.933, -99047.3579},
        {-6.28084603, -6.10990583, -5.93555457, -5.75792216, -5.57713849, -5.39333345, -5.20663695, -5.01717887, -4.82508912, -4.63049758, -4.43353416, -4.23432875, -4.03301124, -3.82971153, -3.62455951, -3.41768508, -3.20921814, -2.99928858, -2.7880263, -2.57556119, -2.36202314, -2.14754206, -1.93224783, -1.71627036, -1.49973954, -1.28278526, -1.06553742, -0.848125919, -0.630680645, -0.413331496, -0.196208369, 0.0205588424, 0.236840242, 0.452505934, 0.664863367, 0.875117875, 1.0845056, 1.2929798, 1.50049374, 1.70700071, 1.91245395, 2.11680675, 2.32001236, 2.52202406, 2.72279511, 2.92227879, 3.12042835, 3.31719708, 3.51253822, 3.70640506, 3.89875086, 4.10305479, 4.30545099, 4.50559953, 4.70316045, 4.89779382, 5.08915967, 5.27691807, 5.46072907, 5.64025272, 5.81514907, 5.98507818, 6.1497001, 6.30867488, 6.46166257, 6.60832323, 6.74831691, 6.8983436, 7.07469417, 7.24188507, 7.39847173, 7.54300958, 7.67405404, 7.79016053, 7.88988447, 7.97178129, 8.03440642, 8.07631526, 8.09606326, 8.09220582, 8.06329838, 8.00789636, 7.92455518, 7.81183026, 6.96568834, 5.7507572, 4.54010918, 3.35510006, 2.21708562, 1.14742163, 0.167463877, -0.701431871, -1.43790983, -2.02061423, -2.42818929, -2.63927924, -2.63252829, -2.38658066, -1.88008059, -1.0916723, 0, 4.08896906, 8.06073535, 11.9616062, 15.8378891, 19.7358912, 23.70192, 27.7822829, 32.0232871, 36.4712401, 41.1724493, 46.1732219, 51.5198653, 57.2586869, 63.4359941, 70.0980942, 77.2912946, 85.0263573, 93.3445117, 102.372465, 112.148909, 122.712533, 134.102028, 146.356083, 159.513391, 173.612641, 188.692524, 204.79173, 221.94895, 240.202875, 260.174319, 281.106402, 302.863178, 325.3087, 349.087431, 373.673219, 398.331802, 422.693109, 446.38707, 469.043614, 490.292671, 517.530954, 540.257805, 556.583567, 564.618581, 562.473188, 555.39066, 532.565177, 490.323849, 424.993784, 332.902094, 223.251982, 63.7570123, -150.782302, -431.838753, -803.693127, -1284.87568, -1893.91668, -2617.9299, -3514.00219, -4673.49288, -6138.66267, -7984.90451, -10223.2671, -12923.4923, -16254.7748, -20340.6587, -25206.3129, -30949.7769, -38029.7997, -46701.0928, -56949.6484, -69340.2384, -84220.5922, -101938.439, -122841.509, -159240.419, -196529.289, -232065.126, -263204.939, -287305.733, -292562.455, -291602.217, -285708.623, -278924.519, -271441.313, -262300.727, -251751.655, -240042.987, -227423.615, -214142.43, -203399.516, -190535.1, -176279.948, -161364.83, -146520.512, -132477.762, -119967.349, -107700.235, -97298.0662},
        {-5.97837511, -5.7718707, -5.56337842, -5.35298921, -5.14079399, -4.9268837, -4.71134928, -4.49428165, -4.27577175, -4.05591051, -3.83478886, -3.61249774, -3.38912809, -3.16477082, -2.93951689, -2.71345721, -2.48668273, -2.25928437, -2.03135307, -1.80297976, -1.57425537, -1.34527084, -1.11611711, -0.886885093, -0.657665736, -0.428549969, -0.199628724, 0.0290070662, 0.257266468, 0.485058548, 0.712292373, 0.938877012, 1.16472153, 1.38973499, 1.61439993, 1.83832661, 2.06111843, 2.28266583, 2.50285928, 2.72158923, 2.93874614, 3.15422047, 3.36790268, 3.57968321, 3.78945254, 3.99710111, 4.20251938, 4.40559782, 4.60622687, 4.804297, 4.99969866, 5.20031521, 5.39787093, 5.59208302, 5.78266866, 5.96934503, 6.15182933, 6.32983873, 6.50309044, 6.67130163, 6.83418949, 6.9914712, 7.14286396, 7.28808496, 7.42685137, 7.55888038, 7.68388919, 7.82071358, 7.98773033, 8.14522592, 8.29167816, 8.42556486, 8.54536381, 8.64955283, 8.73660971, 8.80501227, 8.8532383, 8.87976562, 8.88307203, 8.86163534, 8.81393334, 8.73844384, 8.63364466, 8.49801359, 7.60782876, 6.33785833, 5.0715507, 3.83082046, 2.63758217, 1.51375042, 0.481239764, -0.438035214, -1.22215994, -1.84921985, -2.29730036, -2.5444869, -2.5688649, -2.34851978, -1.86153697, -1.0860019, 0, 4.34492506, 8.53848901, 12.6296179, 16.6672377, 20.7002746, 24.7776544, 28.9483033, 33.2611473, 37.7651123, 42.5091245, 47.5421099, 52.9129945, 58.6707043, 64.8641654, 71.5423037, 78.7540454, 86.3821623, 94.451844, 103.387842, 113.203477, 123.912071, 135.526945, 148.061422, 161.528822, 175.942467, 191.31568, 207.66178, 224.994091, 243.325933, 263.085344, 283.718867, 305.115409, 327.163877, 350.748656, 375.260914, 399.826358, 424.035249, 447.477852, 469.744428, 490.425242, 516.43983, 537.818533, 552.717626, 559.293383, 555.702079, 548.851292, 525.958167, 482.991155, 415.918705, 320.709266, 204.878027, 38.734015, -181.160958, -465.615028, -840.483301, -1324.8651, -1937.85976, -2670.75726, -3576.78657, -4739.51409, -6192.39195, -8019.10974, -10235.9651, -12915.3362, -16224.6273, -20287.6241, -25135.6729, -30868.5938, -37925.1506, -46552.3815, -56727.6085, -69009.0804, -83737.1985, -101252.364, -121894.978, -157732.746, -194446.94, -229446.134, -260138.905, -283933.828, -289294.208, -288537.405, -282905.681, -276254.074, -268790.623, -259634.711, -249048.809, -237295.389, -224636.924, -211335.884, -200231.194, -187331.503, -173288.427, -158753.582, -144378.583, -130815.047, -118714.589, -105944.065, -95557.0232},
        {-5.73832998, -5.49574947, -5.25242248, -5.00840983, -4.76377231, -4.51857074, -4.27286592, -4.02671866, -3.78018977, -3.53334005, -3.28623031, -3.03892136, -2.79147399, -2.54394903, -2.29640728, -2.04890954, -1.80151662, -1.55428932, -1.30728846, -1.06057484, -0.81420927, -0.568252552, -0.322765495, -0.0778089069, 0.166556407, 0.410269639, 0.653269983, 0.895496632, 1.13688878, 1.37738562, 1.61692634, 1.85545014, 2.09289621, 2.32920375, 2.5674718, 2.80599317, 3.04302267, 3.27839696, 3.51195269, 3.74352651, 3.97295507, 4.20007501, 4.42472298, 4.64673564, 4.86594964, 5.08220162, 5.29532824, 5.50516614, 5.71155198, 5.9143224, 6.11331405, 6.3113944, 6.50530358, 6.69481254, 6.87969222, 7.05971358, 7.23464756, 7.40426511, 7.56833717, 7.7266347, 7.87892864, 8.02498994, 8.16458954, 8.29749841, 8.42348747, 8.54232768, 8.65378999, 8.77903086, 8.9386933, 9.08844223, 9.22666226, 9.35173801, 9.46205409, 9.5559951, 9.63194567, 9.68829039, 9.72341389, 9.73570077, 9.72353566, 9.68530314, 9.61938785, 9.52417439, 9.39804737, 9.23939141, 8.30446257, 6.97831941, 5.6549403, 4.35679336, 3.10634669, 1.9260684, 0.838426611, -0.134110579, -0.969075056, -1.64399871, -2.13641343, -2.42385111, -2.48384363, -2.29392289, -1.83162078, -1.07446919, 0, 4.49480523, 8.82342081, 13.0359656, 17.1825587, 21.3133188, 25.4783649, 29.7278159, 34.1117908, 38.6804085, 43.4837878, 48.5720478, 53.9953072, 59.8036851, 66.0473003, 72.7762718, 80.0407185, 87.6116664, 95.4993638, 104.385212, 114.259523, 125.112611, 136.934789, 149.716371, 163.447669, 178.118997, 193.720668, 210.242996, 227.676294, 246.010875, 265.489851, 285.758043, 306.729926, 328.319973, 351.678404, 376.101819, 400.557286, 424.588556, 447.739378, 469.553503, 489.574679, 514.020737, 533.730104, 546.940734, 551.890578, 546.817588, 540.927002, 518.747664, 475.775704, 407.507253, 309.43844, 186.765646, 13.4288896, -211.090823, -496.602893, -871.884642, -1357.19718, -1972.80163, -2716.88832, -3636.80522, -4803.97753, -6238.86159, -8036.9714, -10221.2991, -12867.8985, -16142.7336, -20170.4978, -24989.9552, -30703.5213, -37728.6877, -46304.721, -56392.769, -68553.0051, -83120.9679, -100432.196, -120822.229, -156159.246, -192362.985, -226885.823, -257180.139, -280698.312, -286132.877, -285535.949, -280114.125, -273558.746, -266092.865, -256904.22, -246267.629, -234457.908, -221749.874, -208418.343, -196284.327, -183136.073, -169415.674, -155565.222, -142026.81, -129242.529, -117654.472, -104261.187, -93601.0205},
        {-5.55046846, -5.26916742, -4.98838556, -4.70815687, -4.42851532, -4.1494949, -3.8711296, -3.5934534, -3.31650028, -3.04030422, -2.76489921, -2.49031923, -2.21659826, -1.94377029, -1.67186931, -1.40092928, -1.13098421, -0.862068064, -0.594214834, -0.327458503, -0.0618330523, 0.202627534, 0.465889272, 0.727918179, 0.988680271, 1.24814157, 1.50626808, 1.76302583, 2.01838083, 2.2722991, 2.52474665, 2.77568951, 3.02509368, 3.27292519, 3.5241899, 3.77622789, 4.02631868, 4.27426473, 4.51986851, 4.76293247, 5.00325909, 5.24065083, 5.47491014, 5.70583951, 5.93324138, 6.15691823, 6.37667251, 6.59230669, 6.80362325, 7.01042462, 7.2125133, 7.40937859, 7.60114288, 7.78761544, 7.96860551, 8.14392235, 8.31337521, 8.47677335, 8.633926, 8.78464243, 8.92873189, 9.06600363, 9.19626691, 9.31933097, 9.43500507, 9.54309845, 9.64342039, 9.75933206, 9.91362919, 10.0575469, 10.1893677, 10.3073741, 10.4098485, 10.4950733, 10.5613312, 10.6069045, 10.6300757, 10.6291274, 10.6023419, 10.5480018, 10.4643894, 10.3497875, 10.2024782, 10.0207443, 9.04055036, 7.65737214, 6.27584773, 4.91899828, 3.60984494, 2.37140885, 1.22671118, 0.19877306, -0.689384347, -1.41473989, -1.95427243, -2.2849608, -2.38378387, -2.22772047, -1.79374946, -1.05884969, 0, 4.52411442, 8.89036163, 13.1483086, 17.3475222, 21.5375693, 25.768017, 30.0884322, 34.5483818, 39.1974326, 44.0851517, 49.261106, 54.7748624, 60.6759879, 67.0140493, 73.8386137, 81.1992479, 88.7927132, 96.618176, 105.523216, 115.481798, 126.467888, 138.455453, 151.418457, 165.330868, 180.166649, 195.899767, 212.504188, 229.953878, 248.222802, 267.423177, 287.340025, 307.905838, 329.053105, 352.159119, 376.433968, 400.702463, 424.463655, 447.216594, 468.460331, 487.693915, 510.40644, 528.283857, 539.653252, 542.841711, 536.176318, 531.021269, 509.407262, 466.402106, 397.073609, 296.489578, 167.643819, -12.0091863, -239.805461, -524.530066, -898.34318, -1382.91003, -1999.89584, -2755.34481, -3690.09344, -4862.01934, -6277.94138, -8044.89626, -10191.6229, -12797.4897, -16027.6536, -20007.893, -24784.5338, -30462.7555, -37439.065, -45945.1778, -55918.8125, -67930.9121, -82314.9608, -99404.4424, -119532.841, -154409.048, -190147.287, -224237.192, -254168.394, -277430.524, -282908.625, -282433.313, -277181.902, -270718.551, -263275.782, -254091.596, -243448.242, -231627.965, -218913.013, -205585.634, -191889.405, -178340.524, -165045.292, -152110.007, -139640.972, -127744.485, -116526.848, -102595.967, -91351.2275},
        {-5.38083911, -5.05246188, -4.72621357, -4.40210567, -4.08014967, -3.76035704, -3.44273927, -3.12730784, -2.81407423, -2.50304992, -2.19424639, -1.88767513, -1.58334762, -1.28127535, -0.981469781, -0.683942411, -0.388704717, -0.0957681825, 0.194855711, 0.483155482, 0.769119647, 1.05273673, 1.33399524, 1.61288369, 1.88939062, 2.16350453, 2.43521394, 2.70450738, 2.97137335, 3.23580038, 3.49777698, 3.75729168, 4.01433299, 4.26888943, 4.52510129, 4.78079388, 5.0337426, 5.28380122, 5.53082355, 5.77466336, 6.01517443, 6.25221056, 6.48562552, 6.7152731, 6.94100709, 7.16268126, 7.38014941, 7.59326532, 7.80188278, 8.00585556, 8.20503746, 8.4025771, 8.594962, 8.78197451, 8.963397, 9.13901182, 9.30860134, 9.47194791, 9.62883389, 9.77904164, 9.92235352, 10.0585519, 10.1874191, 10.3087375, 10.4222895, 10.5278575, 10.6252237, 10.7391662, 10.8938629, 11.0375444, 11.1683728, 11.2845101, 11.3841183, 11.4653592, 11.5263951, 11.5653877, 11.5804991, 11.5698914, 11.5317265, 11.4641664, 11.365373, 11.2335085, 11.0667347, 10.8632136, 9.83569449, 8.39156575, 6.94765601, 5.52761544, 4.15509417, 2.85374236, 1.64721016, 0.559147729, -0.386794793, -1.16696725, -1.7577195, -2.13540137, -2.27636274, -2.15695343, -1.75352331, -1.04242221, 0, 4.28275028, 8.47283444, 12.6149594, 16.753832, 20.9341593, 25.200648, 29.5980052, 34.1709376, 38.9641522, 44.0223559, 49.3902557, 55.1125583, 61.2339707, 67.7991999, 74.8529526, 82.4399359, 90.3249684, 98.5127732, 107.681137, 117.814792, 128.898467, 140.916894, 153.854804, 167.696928, 182.427997, 198.032741, 214.495891, 231.802179, 249.936334, 269.141438, 289.049143, 309.566677, 330.601265, 353.453875, 377.334862, 401.082918, 424.187147, 446.136654, 466.420543, 484.527917, 505.84764, 522.173479, 531.840236, 533.182717, 524.535724, 516.640265, 492.323388, 446.818347, 375.358393, 273.176779, 143.705241, -36.0418212, -262.444596, -544.196353, -915.129567, -1397.78972, -2014.72231, -2776.33389, -3720.06771, -4896.60947, -6309.16784, -8063.34346, -10183.5555, -12747.2438, -15921.6884, -19832.9312, -24532.2241, -30126.2191, -36995.7979, -45366.7224, -55140.3593, -66916.7726, -81034.0149, -97830.1387, -117643.197, -152104.757, -187435.977, -221151.531, -250766.093, -273794.336, -279230.958, -278790.592, -273639.331, -267340.824, -260112.129, -251171.324, -240785.418, -229221.419, -216746.336, -203627.176, -187330.206, -173147.243, -160521.6, -148896.592, -137715.53, -126421.73, -114458.503, -100911.445, -88483.5469},
        {-5.2529074, -4.87205368, -4.49532701, -4.12270583, -3.75416862, -3.38969381, -3.02925987, -2.67284523, -2.32042837, -1.97198773, -1.62750176, -1.28694892, -0.950307661, -0.617556434, -0.288673693, 0.0363621082, 0.357572517, 0.674979081, 0.988603346, 1.29846686, 1.60459117, 1.90699782, 2.20570836, 2.50074433, 2.79212729, 3.07987878, 3.36402035, 3.64457354, 3.9215599, 4.19500097, 4.46491832, 4.73133347, 4.99426798, 5.2537434, 5.51282849, 5.7699571, 6.02352644, 6.27345916, 6.51967794, 6.76210543, 7.00066429, 7.23527717, 7.46586674, 7.69235565, 7.91466657, 8.13272214, 8.34644504, 8.55575792, 8.76058343, 8.96084424, 9.15646301, 9.35617008, 9.55088949, 9.74035297, 9.92429223, 10.102439, 10.274525, 10.440282, 10.5994417, 10.7517358, 10.896896, 11.0346541, 11.1647418, 11.2868909, 11.4008329, 11.5062997, 11.6030231, 11.7166367, 11.8721527, 12.0158815, 12.1458757, 12.260188, 12.3568708, 12.4339769, 12.4895588, 12.5216691, 12.5283604, 12.5076852, 12.4576963, 12.3764462, 12.2619874, 12.1123726, 11.9256544, 11.6998854, 10.6256051, 9.12160752, 7.61639251, 6.13421799, 4.69934191, 3.33602222, 2.06851685, 0.921083772, -0.082019082, -0.91653376, -1.55820232, -1.9827668, -2.16596927, -2.08355177, -1.71125635, -1.02482508, 0, 3.94901161, 7.88873984, 11.8575949, 15.8939869, 20.036326, 24.3230224, 28.7924863, 33.4831278, 38.4333572, 43.6815845, 49.26622, 55.2256738, 61.5983561, 68.422677, 75.7370467, 83.5798755, 91.8287384, 100.499069, 109.993072, 120.314691, 131.467873, 143.456563, 156.284707, 169.95625, 184.475137, 199.845315, 216.070728, 233.155323, 251.103045, 270.366091, 290.335796, 310.881629, 331.873061, 354.450946, 377.849062, 400.962149, 423.278261, 444.285452, 463.471775, 480.325285, 500.30944, 515.118288, 523.070781, 522.485872, 511.682514, 499.884881, 471.780399, 422.961716, 349.021477, 245.552331, 117.284753, -60.4948954, -284.062051, -562.335529, -930.047715, -1410.34123, -2026.35868, -2791.40913, -3740.70366, -4920.18089, -6332.8846, -8078.74143, -10174.9524, -12695.3359, -15809.6454, -19643.5984, -24252.0372, -29742.3043, -36481.6188, -44690.1988, -54228.7977, -65732.4008, -79545.6565, -96013.2131, -115479.719, -149514.418, -184431.215, -217768.607, -247065.096, -269859.179, -275243.755, -274833.326, -269785.936, -263693.248, -256749.998, -248136.911, -238099.422, -226882.965, -214732.975, -201894.888, -183003.633, -168217.111, -156284.047, -145953.168, -135973.197, -125092.862, -112060.887, -99125.4403, -85267.8462},
        {-5.17322923, -4.73490174, -4.30321322, -3.8780884, -3.45945199, -3.04722872, -2.64134331, -2.24172047, -1.84828493, -1.46096141, -1.07967462, -0.704349295, -0.334910146, 0.0287181049, 0.386610737, 0.73884303, 1.08549026, 1.42662772, 1.76233067, 2.09267441, 2.4177342, 2.73758533, 3.05230308, 3.36196273, 3.66663956, 3.96640885, 4.26134587, 4.55152592, 4.83702426, 5.11791617, 5.39427694, 5.66618185, 5.93370618, 6.1969252, 6.45883331, 6.71798484, 6.97289901, 7.22355639, 7.46993751, 7.71202293, 7.94979319, 8.18322885, 8.41231045, 8.63701854, 8.85733368, 9.0732364, 9.28470726, 9.49172681, 9.69427559, 9.89233416, 10.0858831, 10.2891384, 10.4875366, 10.6807495, 10.8684492, 11.0503076, 11.2259966, 11.3951882, 11.5575543, 11.7127669, 11.8604979, 12.0004193, 12.132203, 12.255521, 12.3700453, 12.4754477, 12.5714002, 12.6837938, 12.8378899, 12.9792856, 13.1059534, 13.2158656, 13.3069943, 13.3773119, 13.4247907, 13.4474029, 13.4431208, 13.4099168, 13.345763, 13.2486318, 13.1164954, 12.9473262, 12.7390964, 12.4897783, 11.3706989, 9.8105406, 8.24792118, 6.70763919, 5.21449313, 3.79328151, 2.46880284, 1.26585562, 0.209238369, -0.676250413, -1.36581221, -1.83464853, -2.05796085, -2.01095066, -1.66881946, -1.00676875, 0, 3.6044814, 7.28387623, 11.0701321, 14.9951965, 19.0910171, 23.3895415, 27.9227173, 32.722492, 37.8208132, 43.2496285, 49.0408856, 55.2265319, 61.8385151, 68.9087828, 76.4692825, 84.5519619, 93.1538549, 102.301921, 112.10708, 122.593798, 133.78654, 145.709772, 158.387962, 171.845574, 186.107076, 201.196932, 217.13961, 233.959574, 251.681293, 270.976053, 290.98433, 311.536545, 332.463115, 354.737997, 377.619842, 400.062358, 421.552903, 441.578835, 459.627512, 475.186293, 493.841052, 507.12456, 513.330989, 510.754508, 497.689287, 481.502842, 449.147661, 396.649579, 320.034431, 215.32805, 88.9083011, -86.2034943, -306.791894, -582.166756, -947.210425, -1425.32372, -2039.90747, -2805.83749, -3756.9771, -4936.48958, -6349.19759, -8087.54081, -10160.9752, -12639.0719, -15693.2639, -19448.2281, -23961.1593, -29339.2389, -35937.082, -43969.3412, -53256.3362, -64467.719, -77954.6423, -94068.2585, -113159.72, -146715.993, -181168.462, -214084.328, -243030.789, -265575.045, -270933.521, -270591.373, -265694.704, -259849.033, -253226.944, -244976.263, -235320.811, -224484.409, -212690.879, -200164.04, -179273.272, -164075.979, -152744.063, -143449.426, -134363.97, -123659.596, -109508.206, -97169.0915, -81799.0675},
        {-5.15647109, -4.65372543, -4.1613208, -3.67906782, -3.20677711, -2.74425927, -2.29132492, -1.84778466, -1.41344912, -0.988128905, -0.571634626, -0.163776896, 0.235633672, 0.626786465, 1.00987087, 1.38507628, 1.75259207, 2.11260764, 2.46531237, 2.81089565, 3.14954687, 3.48145542, 3.80681067, 4.12580203, 4.43861888, 4.74545059, 5.04648658, 5.34191621, 5.63192888, 5.91671397, 6.19646088, 6.47135899, 6.74159768, 7.00736635, 7.27616242, 7.54436754, 7.80827554, 8.06783863, 8.32300905, 8.57373902, 8.81998078, 9.06168654, 9.29880853, 9.531299, 9.75911015, 9.98219422, 10.2005034, 10.41399, 10.6226062, 10.8263042, 11.0250363, 11.2334133, 11.4364111, 11.6336641, 11.8248067, 12.0094734, 12.1872987, 12.3579171, 12.5209628, 12.6760706, 12.8228747, 12.9610097, 13.0901099, 13.20981, 13.3197442, 13.4195471, 13.5088532, 13.6129187, 13.7563848, 13.8860428, 13.9998661, 14.0958284, 14.171903, 14.2260635, 14.2562833, 14.2605359, 14.236795, 14.1830338, 14.097226, 13.977345, 13.8213643, 13.6272575, 13.3929979, 13.1165592, 11.9584156, 10.352906, 8.74443503, 7.15815476, 5.61921711, 4.15277404, 2.78397748, 1.5379794, 0.439931745, -0.485013534, -1.21170448, -1.71498915, -1.96971559, -1.95073185, -1.63288597, -0.991026002, 0, 3.41837533, 6.96055056, 10.6544556, 14.5280202, 18.6091744, 22.9258479, 27.5059707, 32.3774726, 37.5682834, 43.1063332, 49.0195516, 55.3358687, 62.0832142, 69.2895181, 76.9827101, 85.1907203, 93.9634987, 103.332121, 113.274709, 123.823912, 135.012377, 146.872754, 159.437691, 172.739836, 186.811838, 201.686346, 217.396008, 233.973472, 251.451387, 270.597926, 290.440521, 310.791163, 331.461843, 353.408327, 375.870718, 397.784114, 418.617373, 437.839354, 454.918914, 469.324912, 486.608134, 498.304489, 502.692896, 498.05227, 482.661527, 462.523854, 426.26283, 370.226304, 290.762126, 184.218144, 58.2509886, -115.922116, -336.321992, -612.676586, -978.615898, -1457.03839, -2070.84253, -2834.22264, -3780.75955, -4953.52947, -6359.36568, -8085.43482, -10138.0721, -12583.9081, -15591.7408, -19283.4719, -23716.3098, -28995.3838, -35463.0308, -43327.4096, -52376.5828, -63303.4356, -76461.5449, -92204.4874, -110885.84, -143809.055, -177640.365, -209995.877, -238491.7, -260743.939, -266211.413, -266105.714, -261538.933, -256011.962, -249691.325, -241748.206, -232398.894, -221859.676, -210346.841, -198076.676, -177508.866, -162582.029, -151425.028, -142166.727, -132935.989, -121861.677, -107072.655, -94860.0888, -78197.7522},
        {-5.20643077, -4.63261789, -4.07412598, -3.53058412, -3.00162136, -2.48686678, -1.98594944, -1.49849842, -1.02414279, -0.56251161, -0.113233956, 0.324061105, 0.749744503, 1.16418717, 1.56776003, 1.96083403, 2.34378009, 2.71696913, 3.0807721, 3.43555992, 3.78170353, 4.11957384, 4.44954181, 4.77197835, 5.08725439, 5.39574087, 5.69780872, 5.99382887, 6.28417225, 6.56920979, 6.84931241, 7.12485106, 7.39619666, 7.66372015, 7.94483134, 8.23102332, 8.51358441, 8.79233258, 9.06708582, 9.33766213, 9.60387948, 9.86555586, 10.1225093, 10.3745577, 10.621519, 10.8632114, 11.0994527, 11.330061, 11.5548542, 11.7736503, 11.9862674, 12.2012355, 12.4094717, 12.610605, 12.8042645, 12.9900794, 13.1676789, 13.3366919, 13.4967477, 13.6474754, 13.7885041, 13.9194629, 14.0399809, 14.1496873, 14.2482111, 14.3351816, 14.4102278, 14.4970318, 14.6186983, 14.7252489, 14.8147534, 14.8852817, 14.9349037, 14.9616892, 14.9637082, 14.9390304, 14.8857257, 14.8018641, 14.6855153, 14.5347492, 14.3476357, 14.1222446, 13.8566459, 13.5489094, 12.3584803, 10.7203801, 9.07971005, 7.46175489, 5.8917995, 4.39512872, 2.99702739, 1.72278036, 0.59767248, -0.353011415, -1.10398648, -1.62996786, -1.90567071, -1.9058102, -1.60510148, -0.97825969, 0, 3.45293977, 7.03010328, 10.7588079, 14.6663712, 18.7801103, 23.1273429, 27.7353862, 32.6315577, 37.8431747, 43.3975547, 49.3220151, 55.6438734, 62.3904467, 69.5890527, 77.2670087, 85.4516321, 94.1515501, 103.39141, 113.240805, 123.723047, 134.861448, 146.679322, 159.19998, 172.446734, 186.442898, 201.211783, 216.776701, 233.160966, 250.387889, 269.151577, 288.558591, 308.431004, 328.590889, 350.177399, 372.354067, 393.933211, 414.34178, 433.006725, 449.354999, 462.813551, 478.660077, 488.690123, 491.185929, 484.429735, 466.703782, 443.506624, 404.100109, 344.962397, 262.57165, 153.40603, 25.6921107, -150.19953, -374.12423, -656.158461, -1027.24684, -1508.96462, -2122.88705, -2880.26027, -3815.33615, -4973.63466, -6363.36633, -8070.15912, -10103.3996, -12528.7745, -15507.3686, -19156.4773, -23531.0394, -28732.1838, -35089.6807, -42803.9678, -51642.6477, -62305.5832, -75143.0396, -90505.282, -108742.576, -140846.569, -173865.798, -205490.182, -233409.642, -255314.097, -261051.29, -261381.434, -257355.382, -252220.303, -246157.674, -238433.839, -229276.029, -218911.479, -207567.42, -195471.087, -177983.064, -164129.632, -152638.945, -142239.161, -131658.433, -119624.919, -104866.772, -92140.9868, -74535.3447},
        {-5.30597629, -4.65906083, -4.03267842, -3.42626489, -2.83925608, -2.2710878, -1.72119589, -1.18901618, -0.673984489, -0.17553665, 0.306891509, 0.77386416, 1.22594548, 1.66369963, 2.08769079, 2.49848313, 2.89664082, 3.28272803, 3.65730894, 4.02094772, 4.37420853, 4.71765556, 5.05185297, 5.37736493, 5.69475563, 6.00458922, 6.30742988, 6.60384178, 6.89438909, 7.179636, 7.46014666, 7.73648525, 8.00921594, 8.2789029, 8.5738173, 8.88008703, 9.18350697, 9.48374216, 9.78045761, 10.0733184, 10.3619894, 10.6461358, 10.9254226, 11.1995148, 11.4680774, 11.7307755, 11.9872741, 12.2372382, 12.4803328, 12.7162229, 12.9445737, 13.1669272, 13.3810307, 13.5865085, 13.7829849, 13.9700842, 14.1474308, 14.3146491, 14.4713633, 14.6171978, 14.7517769, 14.8747249, 14.9856663, 15.0842252, 15.1700262, 15.2426934, 15.3018512, 15.3690915, 15.4654786, 15.5453305, 15.6068477, 15.6482302, 15.6676784, 15.6633926, 15.633573, 15.5764199, 15.4901336, 15.3729144, 15.2229624, 15.038478, 14.8176615, 14.5587132, 14.2598332, 13.9192219, 12.6990749, 11.0331791, 9.36536933, 7.72097587, 6.12532897, 4.60375883, 3.1815957, 1.8841698, 0.736811363, -0.235149385, -1.00638221, -1.55155689, -1.84534319, -1.86241088, -1.57742973, -0.965069513, 0, 3.56443062, 7.23613116, 11.0429942, 15.0129122, 19.1737777, 23.5534833, 28.1799216, 33.0809851, 38.2845663, 43.8185577, 49.710852, 55.9893416, 62.6819191, 69.8164771, 77.4209081, 85.5231046, 94.0561469, 103.034383, 112.700162, 123.061059, 134.124649, 145.89851, 158.390215, 171.607341, 185.557464, 200.248159, 215.687001, 231.881566, 248.839431, 267.121606, 285.979305, 305.254072, 324.787452, 345.958774, 367.840689, 389.095775, 409.104242, 427.246299, 442.902155, 455.452019, 469.822385, 478.159178, 478.757468, 469.912319, 449.918799, 424.102852, 381.970949, 320.060438, 234.908668, 123.052986, -7.03297588, -185.336943, -413.551801, -702.016913, -1078.72487, -1563.74233, -2177.13593, -2926.64059, -3847.45737, -4988.24017, -6358.70934, -8043.38248, -10055.4783, -12460.0334, -15409.8871, -19017.5756, -23335.7682, -28461.4801, -34711.2397, -42277.4061, -50914.1125, -61319.3639, -73837.2132, -88811.7134, -106586.918, -137791.73, -169914.139, -200726.986, -228003.111, -249515.354, -255547.756, -256354.49, -252917.883, -248193.588, -242382.822, -234864.584, -225881.203, -215675.006, -204488.324, -192563.484, -178855.846, -166294.303, -154365.074, -142554.375, -130348.427, -117233.446, -102695.652, -89194.7912, -70788.7446},
        {-5.44995917, -4.73004693, -4.03571524, -3.36621476, -2.72079616, -2.09871011, -1.49920729, -0.921538364, -0.364954004, 0.171295118, 0.687958332, 1.18578496, 1.66552435, 2.12792581, 2.57373867, 3.00371227, 3.41859593, 3.81913899, 4.20609077, 4.5802006, 4.9422178, 5.29289172, 5.63297167, 5.96320699, 6.284347, 6.59714103, 6.90233842, 7.20068848, 7.49294056, 7.77984398, 8.06214806, 8.34060214, 8.61595554, 8.88895759, 9.19825953, 9.5248625, 9.84931214, 10.1711278, 10.4898287, 10.8049344, 11.115964, 11.422437, 11.7238728, 12.0197905, 12.3097096, 12.5931494, 12.8696293, 13.1386686, 13.3997866, 13.6525027, 13.8963361, 14.1265212, 14.3469553, 14.5572506, 14.7570193, 14.9458737, 15.123426, 15.2892885, 15.4430734, 15.5843928, 15.7128591, 15.8280845, 15.9296812, 16.0172615, 16.0904375, 16.1488216, 16.1920259, 16.2391894, 16.3089823, 16.3607443, 16.3928218, 16.4035614, 16.3913092, 16.3544118, 16.2912156, 16.2000669, 16.0793121, 15.9272976, 15.7423698, 15.5228751, 15.26716, 14.9735707, 14.6404537, 14.2661554, 13.0176559, 11.3264714, 9.63411053, 7.96589905, 6.34716282, 4.80322767, 3.35941944, 2.04106396, 0.873487047, -0.11798546, -0.908027734, -1.47131394, -1.78251826, -1.81631485, -1.5473779, -0.950381554, 0, 3.72704523, 7.53278544, 11.4464779, 15.4973798, 19.7147484, 24.1278411, 28.7659149, 33.6582273, 38.8340354, 44.3225964, 50.1531677, 56.3550065, 62.95737, 69.9895155, 77.4807003, 85.4601815, 93.7610639, 102.384583, 111.803674, 122.005562, 132.977472, 144.706628, 157.180254, 170.385575, 184.309816, 198.940202, 214.263955, 230.268303, 246.940468, 264.669271, 282.893089, 301.478668, 320.292754, 340.980032, 362.518277, 383.41115, 402.992015, 420.594236, 435.551178, 447.196204, 460.048747, 466.677735, 465.401867, 454.539842, 432.410358, 404.483549, 360.138821, 295.907013, 208.318967, 93.9055218, -38.965338, -220.067536, -452.606318, -747.036123, -1128.61183, -1615.94322, -2227.64009, -2968.08066, -3873.34072, -4995.02136, -6344.31741, -8004.52174, -9992.04788, -12371.2429, -15287.6745, -18849.6288, -23108.5135, -28157.8683, -34299.7878, -41718.3961, -50161.3651, -60316.4178, -72518.7581, -87103.5898, -104406.117, -134659.131, -165825.85, -195767.083, -222343.635, -243416.314, -249726.28, -251000.418, -248150.809, -243833.843, -238271.149, -230957.563, -222151.056, -212109.596, -201091.153, -189353.698, -179496.817, -168265.517, -155918.302, -142713.679, -128910.153, -114766.231, -100540.418, -86066.8355, -66988.9838},
        {-5.62176446, -4.83732236, -4.08195083, -3.3547915, -2.65498596, -1.98167581, -1.33400266, -0.711108122, -0.11213379, 0.46377873, 1.01748783, 1.54985191, 2.06172937, 2.55397859, 3.02745798, 3.48302593, 3.92154084, 4.34386111, 4.75084512, 5.14335128, 5.52223797, 5.88836361, 6.24258657, 6.58576527, 6.91875808, 7.24242342, 7.55761967, 7.86520523, 8.1660385, 8.46097788, 8.75088175, 9.03660851, 9.31901657, 9.59896431, 9.92011877, 10.2610336, 10.599748, 10.935731, 11.2684518, 11.5973797, 11.9219838, 12.2417333, 12.5560972, 12.8645449, 13.1665455, 13.4615682, 13.7490821, 14.0285563, 14.2994602, 14.5612628, 14.8134333, 15.0512371, 15.2784383, 15.4945971, 15.699274, 15.8920292, 16.072423, 16.2400157, 16.3943677, 16.5350392, 16.6615906, 16.7735821, 16.870574, 16.9521267, 17.0178005, 17.0671556, 17.0997524, 17.1321358, 17.1804428, 17.2092044, 17.2168798, 17.2019282, 17.1628089, 17.0979812, 17.0059044, 16.8850377, 16.7338403, 16.5507716, 16.3342908, 16.0828571, 15.7949299, 15.4689684, 15.1034318, 14.6967794, 13.4178301, 11.6973138, 9.97593178, 8.27906687, 6.63210172, 5.06041901, 3.58940145, 2.24443171, 1.05089249, 0.0341664752, -0.780363646, -1.36731518, -1.70130545, -1.75695175, -1.5088714, -0.931681717, 0, 3.89219613, 7.83473149, 11.8581995, 15.9931937, 20.2703074, 24.7201342, 29.3732674, 34.2603005, 39.411827, 44.8584402, 50.6307337, 56.7593009, 63.2747353, 70.2076302, 77.5885791, 85.4481756, 93.5103881, 101.762007, 110.927181, 120.970799, 131.857748, 143.552918, 156.021196, 169.22747, 183.136628, 197.713559, 212.92315, 228.73029, 245.099867, 262.245524, 279.792185, 297.630109, 315.64956, 335.712065, 356.722254, 377.059081, 396.021429, 412.90818, 427.018213, 437.650411, 448.877119, 453.770961, 450.69688, 438.019819, 414.10472, 385.078754, 339.60408, 274.105082, 185.006147, 68.7316593, -67.1630715, -250.989044, -485.805859, -781.727899, -1163.27306, -1648.49235, -2255.43677, -2987.77961, -3881.21728, -4987.13101, -6317.87462, -7953.00248, -9905.48289, -12237.5977, -15095.1878, -18585.7391, -22765.1514, -27727.5362, -33756.1552, -41027.276, -49295.4135, -59223.4732, -71133.7141, -85348.3951, -102189.775, -131488.073, -161680.598, -190712.62, -216529.406, -237076.226, -243506.834, -245085.691, -242668.716, -238694.094, -233391.43, -226331.052, -217775.531, -207987.437, -197229.338, -185763.804, -177963.432, -167568.526, -155198.881, -141474.292, -127014.557, -112439.47, -98368.8276, -82792.5326, -63120.5492},
        {-5.81886307, -4.97907832, -4.17012754, -3.39112814, -2.6411975, -1.91945303, -1.2250121, -0.556992125, 0.0854895141, 0.703315421, 1.2973682, 1.86853047, 2.41768482, 2.94571387, 3.45350022, 3.94192647, 4.41187525, 4.86422915, 5.29987077, 5.71968274, 6.12454764, 6.5153481, 6.89296671, 7.25828609, 7.61218884, 7.95555756, 8.28927487, 8.61422337, 8.93128567, 9.24134437, 9.54528209, 9.84398142, 10.138325, 10.4291954, 10.7592697, 11.1076543, 11.4529529, 11.794692, 12.1323979, 12.4655967, 12.7938149, 13.1165786, 13.4334143, 13.7438481, 14.0474065, 14.3436156, 14.6320018, 14.9120914, 15.1834107, 15.4454859, 15.6978434, 15.9427758, 16.1769831, 16.3999317, 16.6110877, 16.8099176, 16.9958877, 17.1684643, 17.3271136, 17.4713022, 17.6004962, 17.714162, 17.8117659, 17.8927743, 17.9566534, 18.0028696, 18.0308893, 18.05466, 18.0877818, 18.0998543, 18.0894051, 18.0549617, 17.9950516, 17.9082025, 17.7929419, 17.6477973, 17.4712963, 17.2619664, 17.0183353, 16.7389304, 16.4222794, 16.0669097, 15.671349, 15.2341248, 13.9219425, 12.1667371, 10.4104412, 8.67857659, 6.99666513, 5.39022857, 3.8847887, 2.50586729, 1.27898613, 0.229667012, -0.616568291, -1.23419799, -1.5977003, -1.68155344, -1.46023563, -0.908225076, 0, 4.03928556, 8.10514061, 12.2291642, 16.4429554, 20.7781132, 25.2662366, 29.9389248, 34.8277767, 39.9643914, 45.3803679, 51.1073054, 57.1768027, 63.620459, 70.4698733, 77.7566447, 85.5123722, 93.3512729, 101.24532, 110.170191, 120.068044, 130.881042, 142.551343, 155.021108, 168.232496, 182.127669, 196.648785, 211.738005, 227.33749, 243.389398, 259.945447, 276.79807, 293.85656, 311.030209, 330.326874, 350.606566, 370.16968, 388.29594, 404.265072, 417.356799, 426.850845, 436.356163, 439.505658, 434.732031, 420.467983, 395.146215, 366.00031, 320.461867, 254.763367, 165.137291, 47.8161207, -91.120228, -277.322464, -511.947809, -804.309265, -1180.37122, -1658.62195, -2257.54969, -2983.08562, -3869.12898, -4963.27761, -6278.58659, -7888.23956, -9794.7485, -12056.8507, -14828.5539, -18220.1353, -22297.8943, -27160.7064, -33068.6983, -40190.784, -48300.2102, -58022.43, -69663.3189, -83528.7526, -99924.6068, -128287.505, -157509.889, -185614.945, -210625.863, -230565.832, -236938.986, -238633.603, -236466.751, -232756.092, -227724.71, -220970.44, -212747.628, -203310.619, -192913.76, -181811.396, -173925.041, -163772.974, -151844.855, -138630.344, -124619.101, -110300.786, -96165.0587, -79419.4822, -59225.1055},
        {-6.05520415, -5.1610458, -4.29902044, -3.46824556, -2.66783868, -1.89691728, -1.15459889, -0.440000993, 0.247758902, 0.909563294, 1.54629468, 2.15883556, 2.74806842, 3.31487577, 3.86014011, 4.38474392, 4.88956972, 5.37549999, 5.84341723, 6.29420395, 6.72874263, 7.14791578, 7.5526059, 7.94369547, 8.32206701, 8.688603, 9.04418594, 9.38969833, 9.72602267, 10.0540415, 10.3746372, 10.6886924, 10.9970895, 11.300711, 11.6395589, 11.994133, 12.3444632, 12.6901625, 13.0308439, 13.3661204, 13.695605, 14.0189107, 14.3356505, 14.6454374, 14.9478844, 15.2426044, 15.5292106, 15.8073158, 16.0765332, 16.3364756, 16.5867561, 16.8384969, 17.0795523, 17.3092858, 17.5270609, 17.7322411, 17.9241899, 18.1022708, 18.2658473, 18.4142828, 18.546941, 18.6631853, 18.7623791, 18.8438861, 18.9070697, 18.9512933, 18.9759206, 18.9923395, 19.0116493, 19.0084139, 18.9812172, 18.9286433, 18.8492762, 18.7416999, 18.6044984, 18.4362557, 18.2355557, 18.0009826, 17.7311203, 17.4245527, 17.079864, 16.695638, 16.2704589, 15.8029105, 14.4572744, 12.6668341, 10.8749013, 9.10713447, 7.38919248, 5.74673403, 4.20541783, 2.79090262, 1.52884713, 0.444910068, -0.435249824, -1.08597382, -1.4816032, -1.59647924, -1.4049432, -0.881336363, 0, 4.18045943, 8.36392861, 12.5829831, 16.8701986, 21.2581506, 25.7794146, 30.4665664, 35.3521814, 40.4688353, 45.8491037, 51.525562, 57.530786, 63.8973512, 70.6578332, 77.8448075, 85.4908498, 93.1004896, 100.630867, 109.308948, 119.054156, 129.785912, 141.423641, 153.886764, 167.094705, 180.966887, 195.422732, 210.381663, 225.763103, 241.486476, 257.447801, 273.598484, 289.86497, 306.173701, 324.646797, 344.112862, 362.814991, 380.020924, 394.998403, 407.015168, 415.338961, 423.13199, 424.582256, 418.19554, 402.477621, 375.93428, 346.987816, 301.74836, 236.242563, 146.497076, 28.5385496, -113.180553, -301.357611, -534.83568, -822.011151, -1190.78514, -1660.30154, -2249.70424, -2967.89915, -3846.59496, -4928.66905, -6227.43889, -7809.63523, -9666.51986, -11853.1442, -14532.6476, -17818.5553, -21788.2567, -26545.9086, -32328.1653, -39296.7328, -47245.0412, -56759.7305, -68129.9871, -81644.9972, -97593.947, -125030.867, -153288.888, -180465.986, -204660.136, -223969.313, -230223.922, -231967.89, -229985.294, -226503.133, -221734.327, -215290.58, -207414.873, -198350.188, -188339.505, -177625.808, -168936.661, -158840.126, -147527.227, -135188.989, -122016.439, -108200.6, -93932.499, -76009.5411, -55411.2063},
        {-6.3334436, -5.38378848, -4.46738217, -3.58335357, -2.73083157, -1.90894506, -1.11682293, -0.353594083, 0.381612601, 1.08966822, 1.7714439, 2.42781073, 3.05963982, 3.66780228, 4.25316923, 4.81661176, 5.35900099, 5.88120802, 6.38410396, 6.86855993, 7.33544702, 7.78563634, 8.21999901, 8.63940613, 9.04472881, 9.43683816, 9.81660528, 10.1849013, 10.5425973, 10.8905644, 11.2296737, 11.5607963, 11.8848033, 12.2025659, 12.550645, 12.9113162, 13.2664259, 13.615687, 13.9588124, 14.295515, 14.6255077, 14.9485035, 15.2642151, 15.5723556, 15.8726378, 16.1647745, 16.4484788, 16.7234636, 16.9894416, 17.2461258, 17.4932292, 17.7513216, 17.9988068, 18.2349455, 18.4589985, 18.6702266, 18.8678905, 19.0512509, 19.2195688, 19.3721047, 19.5081195, 19.626874, 19.7276288, 19.8096448, 19.8721827, 19.9145033, 19.9358673, 19.945248, 19.9513856, 19.9335094, 19.8902506, 19.8202402, 19.7221095, 19.5944896, 19.4360116, 19.2453066, 19.0210059, 18.7617404, 18.4661414, 18.13284, 17.7604673, 17.3476545, 16.8930328, 16.3952331, 15.0162583, 13.190651, 11.3630097, 9.55911552, 7.80474996, 6.12569432, 4.54772993, 3.09663814, 1.79820029, 0.678197715, -0.237588243, -0.923376244, -1.35338495, -1.50183302, -1.34293911, -0.850921882, 0, 4.30542299, 8.59212183, 12.8935074, 17.2429908, 21.6739827, 26.2198942, 30.9141362, 35.7901196, 40.8812553, 46.2209542, 51.8426274, 57.7796856, 64.0655399, 70.7336011, 77.8172802, 85.3499881, 92.7327004, 99.90848, 108.341246, 117.928889, 128.569296, 140.160358, 152.599963, 165.785999, 179.616357, 193.988926, 208.801593, 223.952249, 239.338783, 254.716666, 270.178425, 285.664672, 301.116023, 318.733811, 337.32829, 355.106856, 371.331907, 385.26584, 396.171053, 403.309946, 409.441252, 409.26693, 401.365312, 384.31473, 356.693514, 328.035986, 283.225491, 218.101366, 128.502946, 10.2695665, -133.734774, -323.281105, -554.842067, -835.886865, -1196.33602, -1656.02986, -2234.80874, -2944.77665, -3815.26852, -4884.05891, -6164.19721, -7716.60037, -9522.18812, -11632.5347, -14218.9346, -17397.7286, -21256.5505, -25904.2986, -31555.1308, -38363.6641, -46140.7095, -55437.5835, -66527.9131, -79685.3252, -95183.4469, -121715.933, -149031.375, -175298.393, -198685.609, -217361.641, -223458.919, -225206.382, -223359.587, -220074.781, -215553.083, -209411.996, -201881.821, -193192.861, -183575.417, -173259.792, -163260.979, -153092.162, -142523.723, -131326.043, -119269.503, -106124.485, -91661.371, -72614.7951, -51741.5636},
        {-6.65623729, -5.64787, -4.67396538, -3.73366197, -2.82609828, -1.95041283, -1.10574415, -0.291230751, 0.493988835, 1.25077609, 1.97999249, 2.68249952, 3.35915865, 4.01083137, 4.63837914, 5.24266346, 5.82454579, 6.38488762, 6.92455042, 7.44439567, 7.94528486, 8.42807945, 8.89364093, 9.34283077, 9.77651046, 10.1955415, 10.6007853, 10.9931034, 11.3733572, 11.7424083, 12.1011181, 12.4503481, 12.7909597, 13.1238145, 13.482187, 13.8500503, 14.2109881, 14.56481, 14.9113256, 15.2503445, 15.5816764, 15.9051307, 16.2205172, 16.5276455, 16.8263252, 17.1163658, 17.3975771, 17.6697686, 17.9327499, 18.1863306, 18.4303204, 18.694171, 18.9474073, 19.1891963, 19.4187051, 19.6351007, 19.83755, 20.0252203, 20.1972784, 20.3528914, 20.4912263, 20.6114502, 20.7127301, 20.794233, 20.8551259, 20.894576, 20.9117501, 20.9134593, 20.9063312, 20.8737675, 20.8144398, 20.7270196, 20.6101784, 20.4625877, 20.282919, 20.0698438, 19.8220336, 19.53816, 19.2168944, 18.8569082, 18.4568732, 18.0154606, 17.5313421, 17.003189, 15.5913269, 13.7312338, 11.8684639, 10.0288947, 8.23840374, 6.52286837, 4.90816607, 3.42017428, 2.08477045, 0.927832026, -0.0247635433, -0.747138812, -1.21341633, -1.39771866, -1.27416834, -0.816887939, 0, 4.40388149, 8.77074659, 13.1345882, 17.5293993, 21.9891726, 26.5479013, 31.2395781, 36.0981959, 41.1577478, 46.4522265, 52.0156251, 57.8819363, 64.0851532, 70.6592686, 77.6382755, 85.0561667, 92.2225673, 99.0679893, 107.264876, 116.692, 127.228131, 138.752041, 151.142501, 164.278282, 178.038156, 192.300894, 206.945267, 221.850046, 236.894003, 251.716127, 266.522891, 281.265001, 295.893161, 312.649894, 330.339996, 347.157115, 362.364413, 375.225049, 385.002185, 390.958983, 395.520603, 393.825858, 384.519257, 366.245308, 337.648519, 309.139533, 264.655192, 199.898471, 110.572346, -7.62020811, -153.173616, -343.279568, -572.339566, -846.989712, -1198.84507, -1648.30566, -2215.77151, -2916.27459, -3776.80298, -4830.20077, -6088.62722, -7608.54587, -9363.14446, -11401.0787, -13898.8806, -16974.3849, -20723.0876, -25257.0317, -30770.1693, -37410.1199, -44998.0183, -54058.1975, -64851.2912, -77637.9332, -92678.7572, -118340.478, -144751.128, -170144.82, -192755.668, -210817.786, -216741.254, -218466.907, -216724.873, -213610.601, -209313.779, -203455.21, -196253.026, -187925.359, -178690.34, -168766.102, -157160.682, -146851.266, -137112.072, -127217.32, -116441.227, -104058.013, -89341.8972, -69287.3299, -48278.8888},
        {-7.01295006, -5.92233413, -4.86997962, -3.85492393, -2.87620443, -1.93285852, -1.0239236, -0.148437051, 0.694563739, 1.50604138, 2.28695847, 3.03827764, 3.76096148, 4.45597261, 5.12427364, 5.76682718, 6.38459583, 6.97854221, 7.54962893, 8.0988186, 8.62707383, 9.13535722, 9.62463139, 10.095859, 10.5500025, 10.9880247, 11.4108881, 11.8195553, 12.2149889, 12.5981516, 12.970006, 13.3315146, 13.6836401, 14.0273451, 14.3999006, 14.7833514, 15.1593057, 15.5275478, 15.8878622, 16.240033, 16.5838447, 16.9190815, 17.2455279, 17.5629681, 17.8711864, 18.1699673, 18.4590949, 18.7383537, 19.007528, 19.2664021, 19.5147604, 19.781035, 20.0357415, 20.278043, 20.5071029, 20.7220844, 20.9221509, 21.1064657, 21.2741921, 21.4244934, 21.5565328, 21.6694737, 21.7624793, 21.8347131, 21.8853382, 21.913518, 21.9184157, 21.9067187, 21.8849336, 21.836706, 21.7607114, 21.6556256, 21.5201239, 21.3528819, 21.1525753, 20.9178795, 20.6474701, 20.3400227, 19.9942128, 19.6087159, 19.1822076, 18.7133636, 18.2008592, 17.64337, 16.1995586, 14.3057755, 12.4084677, 10.5335708, 8.7070204, 6.95475213, 5.30270157, 3.77680429, 2.4029959, 1.20721199, 0.21538816, -0.546540004, -1.0526369, -1.27696695, -1.19359454, -0.77658409, 0, 4.50519164, 8.95189676, 13.3748777, 17.8088967, 22.2887161, 26.8490983, 31.5248054, 36.3506, 41.3612442, 46.5915003, 52.0761308, 57.8498978, 63.9475637, 70.4038908, 77.2536415, 84.531578, 91.499262, 98.0809988, 106.060457, 115.306712, 125.688841, 137.075921, 149.337027, 162.341237, 175.957627, 190.055273, 204.503252, 219.17064, 233.926514, 248.38901, 262.770156, 277.014311, 291.065833, 307.132492, 324.026941, 339.958257, 354.191778, 365.992838, 374.626773, 379.35892, 382.491891, 379.329356, 368.542401, 348.802112, 318.779574, 288.019593, 241.601106, 175.476768, 85.5992344, -32.0788386, -175.243971, -362.078672, -588.351605, -860.959638, -1209.51973, -1653.44797, -2212.16046, -2899.57509, -3741.85577, -4771.16447, -5996.70034, -7478.62399, -9200.13229, -11210.6935, -13672.0687, -16694.4751, -20367.1182, -24794.7676, -30159.234, -36598.7494, -43870.8607, -52553.0516, -62915.3295, -75227.7017, -89760.1758, -114775.359, -140552.509, -165363.483, -187480.138, -205174.332, -211041.305, -212813.943, -211196.938, -208174.458, -203944.738, -198172.896, -191083.821, -182902.403, -173853.531, -164162.094, -150687.535, -140285.879, -131599.051, -123268.974, -113937.571, -102246.765, -86838.4797, -66258.8941, -45344.1289},
        {-7.42228891, -6.24580993, -5.11159355, -4.01856812, -2.96566197, -1.95180345, -0.97592092, -0.0369427114, 0.866202826, 1.73458735, 2.5692825, 3.37135995, 4.14189135, 4.88194835, 5.5926026, 6.27492577, 6.9299895, 7.55886544, 8.16262526, 8.74234061, 9.29908314, 9.83392451, 10.3479364, 10.8421904, 11.3177582, 11.7757114, 12.2171218, 12.6430609, 13.0546005, 13.4528121, 13.8387675, 14.2135382, 14.5781959, 14.9338124, 15.3223597, 15.7235989, 16.1167996, 16.5017061, 16.8780628, 17.2456141, 17.6041043, 17.9532778, 18.2928789, 18.622652, 18.9423414, 19.2516915, 19.5504467, 19.8383512, 20.1151496, 20.380586, 20.6344049, 20.902778, 21.1584494, 21.4005906, 21.6283729, 21.840968, 22.0375472, 22.2172819, 22.3793438, 22.5229041, 22.6471345, 22.7512063, 22.834291, 22.89556, 22.9341849, 22.9493372, 22.9401881, 22.91342, 22.8755347, 22.8102134, 22.7161406, 22.5920009, 22.4364789, 22.2482591, 22.0260261, 21.7684645, 21.4742589, 21.1420939, 20.7706539, 20.3586236, 19.9046875, 19.4075303, 18.8658365, 18.2782906, 16.8036222, 14.8781606, 12.948331, 11.0400761, 9.17933861, 7.39206123, 5.70418672, 4.14165781, 2.73041723, 1.4964077, 0.465571967, -0.336147251, -0.88280722, -1.14846521, -1.10717849, -0.73300433, 0, 4.55117311, 9.03077307, 13.4737391, 17.9150105, 22.3895265, 26.9322262, 31.578049, 36.3619341, 41.3188206, 46.483648, 51.8913553, 57.5768818, 63.5751668, 69.9211494, 76.649769, 83.7959648, 90.592384, 96.9593529, 104.746518, 113.8169, 124.033518, 135.259391, 147.357539, 160.190981, 173.622737, 187.515827, 201.73327, 216.138085, 230.593293, 244.725066, 258.720114, 272.51251, 286.036328, 301.427842, 317.520025, 332.558597, 345.816969, 356.568556, 364.086769, 367.645022, 369.522752, 365.072471, 352.979456, 331.928985, 300.606336, 267.416596, 218.89528, 151.297714, 60.8792236, -56.1048657, -196.28345, -379.214045, -602.464791, -873.310837, -1218.84582, -1657.26528, -2206.76475, -2879.57177, -3701.20626, -4703.69866, -5892.76447, -7333.80392, -9023.24841, -11012.051, -13443.5489, -16418.6732, -20017.0104, -24333.9581, -29543.178, -35774.0613, -42718.0017, -51009.7719, -60930.3774, -72760.824, -86782.1171, -111171.627, -136339.903, -160593.86, -182240.413, -199586.477, -205401.938, -207222.09, -205726.858, -202789.676, -198618.041, -192922.152, -185934.37, -177887.055, -169012.568, -159543.269, -144358.947, -133903.85, -126227.308, -119378.651, -111407.208, -100362.311, -84293.2893, -63364.6016, -42688.1347},
        {-7.89041817, -6.62718138, -5.41010979, -4.23802707, -3.10975692, -2.024123, -0.979948992, 0.0239414186, 0.988724556, 1.91557674, 2.8056743, 3.66019354, 4.4803108, 5.26720239, 6.02204463, 6.74601384, 7.44028636, 8.10603848, 8.74444655, 9.35668687, 9.94393578, 10.5073696, 11.0481646, 11.5674972, 12.0665436, 12.5464802, 13.0084834, 13.4537293, 13.8833944, 14.298655, 14.7006873, 15.0906678, 15.4697727, 15.8391784, 16.2455432, 16.6663459, 17.0785183, 17.4817608, 17.8757737, 18.2602574, 18.6349122, 18.9994384, 19.3535364, 19.6969065, 20.029249, 20.3502644, 20.6596529, 20.9571149, 21.2423507, 21.5150606, 21.774945, 22.0456419, 22.3023951, 22.5443859, 22.7707958, 22.9808062, 23.1735985, 23.3483541, 23.5042545, 23.640481, 23.7562151, 23.8506383, 23.9229318, 23.9722772, 23.9978558, 23.9988491, 23.9744385, 23.931084, 23.875071, 23.7905691, 23.6762881, 23.5309374, 23.3532266, 23.1418654, 22.8955632, 22.6130296, 22.2929743, 21.9341068, 21.5351366, 21.0947734, 20.6117266, 20.084706, 19.512421, 18.8935812, 17.3894446, 15.4350612, 13.4755466, 11.5367874, 9.64466987, 7.82508063, 6.10390615, 4.5070329, 3.06034739, 1.78973609, 0.721085489, -0.119717916, -0.706787643, -1.0142372, -1.01618011, -0.686729869, 0, 4.52106434, 8.96964213, 13.379999, 17.7864004, 22.2231121, 26.7243996, 31.3245284, 36.0577642, 40.9583725, 46.060619, 51.3987691, 57.0070886, 62.9198429, 69.1712976, 75.7957184, 82.8273708, 89.4886037, 95.6976946, 103.325387, 112.231965, 122.277712, 133.322911, 145.227845, 157.852798, 171.058054, 184.703896, 198.650606, 212.75847, 226.88777, 240.693989, 254.317305, 267.679441, 280.702118, 295.419616, 310.69738, 324.837509, 337.127958, 346.856684, 353.311645, 355.780796, 356.616174, 351.10911, 337.948068, 315.821511, 283.417901, 247.925411, 197.407867, 128.428806, 37.5517641, -78.6597234, -215.796851, -394.730014, -614.879917, -883.885877, -1226.16688, -1658.60818, -2198.09504, -2854.96036, -3654.13092, -4627.68152, -5777.58688, -7175.14863, -8831.41524, -10799.102, -13201.5264, -16129.5982, -19651.1212, -23851.0147, -28898.9208, -34916.5576, -41538.87, -49449.2375, -58937.8766, -70295.004, -83810.8361, -107560.558, -132103.176, -155782.664, -176942.996, -193928.148, -199689.956, -201559.291, -200191.041, -197351.833, -193256.495, -187656.632, -180790.233, -172895.29, -164209.79, -154971.723, -138493.284, -128067.943, -121266.902, -115661.362, -108822.524, -98321.5903, -81729.7614, -60609.1127, -40305.4516},
        {-8.42350218, -7.07533246, -5.77683095, -4.52673334, -3.32377531, -2.16669254, -1.05422071, 0.0149044915, 1.04194739, 2.02817231, 2.97484356, 3.88322547, 4.75458235, 5.59017852, 6.3912783, 7.15914602, 7.89504598, 8.60024251, 9.27599993, 9.92358256, 10.5442547, 11.1392807, 11.7099249, 12.2574515, 12.783125, 13.2882095, 13.7739695, 14.2416693, 14.6925732, 15.1279454, 15.5490504, 15.9571524, 16.3535157, 16.7394048, 17.1654304, 17.6071456, 18.0395104, 18.4621879, 18.8748408, 19.2771322, 19.668725, 20.049282, 20.4184663, 20.7759407, 21.1213682, 21.4544117, 21.7747341, 22.0819984, 22.3758675, 22.6560044, 22.9220718, 23.1958688, 23.4544424, 23.6969757, 23.9226518, 24.1306538, 24.3201647, 24.4903675, 24.6404454, 24.7695814, 24.8769586, 24.96176, 25.0231687, 25.0603678, 25.0725404, 25.0588694, 25.0185381, 24.9572314, 24.8804786, 24.7740525, 24.6367147, 24.4672269, 24.2643505, 24.0268471, 23.7534782, 23.4430055, 23.0941905, 22.7057948, 22.2765799, 21.8053073, 21.2907387, 20.7316357, 20.1267596, 19.4748722, 17.9429527, 15.9631491, 13.9776071, 12.0120811, 10.0923257, 8.24409532, 6.49314445, 4.86522763, 3.38609934, 2.08151407, 0.977226343, 0.0989906394, -0.527438537, -0.876306687, -0.921859314, -0.638341917, 0, 4.39410377, 8.73077055, 13.0424836, 17.3617262, 21.7209815, 26.1527328, 30.6894633, 35.3636564, 40.2077951, 45.2543629, 50.5358429, 56.0847184, 61.9334726, 68.1145887, 74.6605501, 81.6038399, 88.1745915, 94.2906668, 101.799388, 110.561306, 120.436973, 131.286938, 142.971755, 155.351972, 168.288143, 181.640817, 195.270546, 209.037882, 222.803374, 236.265473, 249.506269, 262.434943, 274.960677, 288.991487, 303.437136, 316.674368, 328.012714, 336.761706, 342.230874, 343.729751, 343.775144, 337.493181, 323.565883, 300.67527, 267.503363, 230.14091, 178.009024, 107.937543, 16.7563068, -98.7048453, -233.288972, -408.670906, -625.797778, -892.527326, -1230.82644, -1656.32728, -2184.66201, -2824.43658, -3599.90624, -4542.99125, -5651.93484, -7003.72109, -8623.55522, -10565.797, -12934.2067, -15809.8692, -19247.8073, -23322.3489, -28203.382, -34006.7401, -40332.8942, -47892.3274, -56979.2685, -67887.9458, -80912.588, -103973.431, -127832.195, -150876.605, -171494.389, -188073.272, -193772.164, -195693.489, -194465.896, -191756.506, -187782.908, -182329.989, -175636.973, -167943.08, -159487.533, -150509.552, -133408.908, -123140.923, -116987.894, -112232.121, -106155.903, -96041.5401, -79171.3312, -57997.0875, -38190.625},
        {-9.02770528, -7.59914714, -6.22305965, -4.89811945, -3.62300317, -2.39638747, -1.21694898, -0.0833643405, 1.0056898, 2.05153679, 3.05550001, 4.0189028, 4.94306853, 5.82932055, 6.67898222, 7.49337689, 8.27382793, 9.0216587, 9.73819255, 10.4247528, 11.0826629, 11.7132462, 12.3178259, 12.8977255, 13.4542684, 13.9887778, 14.5025772, 14.9969899, 15.4733392, 15.9329486, 16.3771414, 16.8072408, 17.2245704, 17.6304534, 18.0780001, 18.541551, 18.9948245, 19.4374633, 19.8691103, 20.2894081, 20.6979996, 21.0945275, 21.4786346, 21.8499637, 22.2081576, 22.5528591, 22.8837108, 23.2003556, 23.5024363, 23.7895957, 24.0614765, 24.3397007, 24.6014554, 24.8459069, 25.0722216, 25.2795657, 25.4671055, 25.6340074, 25.7794377, 25.9025626, 26.0025486, 26.0785618, 26.1297686, 26.1553353, 26.1544283, 26.1262138, 26.0698581, 25.9893831, 25.8886939, 25.7569429, 25.5929814, 25.3956612, 25.1638336, 24.8963504, 24.592063, 24.249823, 23.8684819, 23.4468912, 22.9839026, 22.4783676, 21.9291377, 21.3350645, 20.6949995, 20.0077942, 18.4500734, 16.4490963, 14.4420051, 12.4543341, 10.5116177, 8.63939026, 6.86318625, 5.20854005, 3.70098603, 2.3660586, 1.22929214, 0.316221054, -0.347620265, -0.736697425, -0.82547603, -0.588421687, 0, 4.14952984, 8.27642494, 12.4100194, 16.5796474, 20.8146429, 25.1443403, 29.5980734, 34.2051766, 38.9949839, 43.9968293, 49.2400472, 54.7539714, 60.5679363, 66.7112759, 73.2133243, 80.1034157, 86.6370178, 92.7329125, 100.170846, 108.814323, 118.52685, 129.171933, 140.613077, 152.713787, 165.33757, 178.347931, 191.608375, 204.982409, 218.333537, 231.409211, 244.231544, 256.698859, 268.709478, 282.027126, 295.617423, 307.94855, 318.35921, 326.188102, 330.77393, 331.455394, 331.00265, 324.278591, 309.950546, 286.685844, 253.151816, 214.657965, 161.568905, 90.8914215, -0.367697347, -115.201665, -248.264612, -421.081047, -635.419168, -899.077755, -1232.16802, -1649.27316, -2164.97634, -2786.69616, -3537.80871, -4449.50603, -5516.57561, -6820.58429, -8398.5908, -10306.0864, -12629.7953, -15442.1053, -18785.4257, -22724.3723, -27433.481, -33025.1105, -39099.5029, -46359.921, -55095.9945, -65597.3535, -78153.6278, -100441.524, -123516.826, -145822.396, -165801.092, -181895.777, -187515.366, -189492.628, -188427.829, -185899.274, -182120.087, -176895.879, -170460.15, -163046.401, -154888.135, -146218.852, -129424.186, -119485.554, -113660.344, -109205.94, -103379.729, -93439.0988, -76641.4339, -55533.1864, -36338.2001},
        {-9.72480593, -8.21120533, -6.75318798, -5.34936362, -3.99834198, -2.69873283, -1.44914588, -0.248190897, 0.90552239, 2.01338424, 3.0767849, 4.09711463, 5.0757637, 6.01412236, 6.91358087, 7.77552948, 8.60135846, 9.39245806, 10.1502185, 10.8760302, 11.5712832, 12.2373678, 12.8756744, 13.4875932, 14.0745143, 14.6378282, 15.178925, 15.699195, 16.2000284, 16.6828156, 17.1489468, 17.5998121, 18.036802, 18.4613066, 18.9282492, 19.4111275, 19.8827959, 20.3429074, 20.7911151, 21.2270719, 21.6504309, 22.060845, 22.4579673, 22.8414508, 23.2109485, 23.5661134, 23.9065985, 24.2320569, 24.5421415, 24.8365054, 25.1148015, 25.4030339, 25.6739334, 25.9265818, 26.1600607, 26.373452, 26.5658374, 26.7362987, 26.8839177, 27.0077762, 27.1069559, 27.1805386, 27.2276061, 27.2472401, 27.2385224, 27.2005348, 27.1323591, 27.0340643, 26.9056956, 26.7442988, 26.5488917, 26.3184919, 26.0521172, 25.7487856, 25.4075146, 25.0273221, 24.6072258, 24.1462436, 23.6433932, 23.0976923, 22.5081588, 21.8738104, 21.1936649, 20.46674, 18.887234, 16.8735026, 14.8535752, 12.8525873, 10.8956748, 9.00797329, 7.21461854, 5.54074625, 4.01149215, 2.65199195, 1.48738137, 0.542796129, -0.156628066, -0.585755493, -0.719450435, -0.532577177, 0, 3.61309535, 7.28716449, 11.0451106, 14.9098367, 18.9042461, 23.0512418, 27.3737269, 31.8946046, 36.6367781, 41.6231504, 46.8766247, 52.4201041, 58.2764918, 64.4686908, 71.0196044, 77.9521356, 84.5634485, 90.7727693, 98.2445536, 106.846189, 116.445064, 126.908564, 138.104079, 149.898996, 162.160702, 174.756585, 187.554032, 200.420432, 213.223171, 225.730962, 237.946049, 249.765423, 261.086074, 273.501373, 286.06012, 297.358749, 306.785334, 313.727953, 317.574681, 317.713596, 317.430317, 311.029168, 297.135663, 274.375315, 241.373638, 203.755345, 151.39695, 81.1741672, -10.0372901, -125.361708, -261.240593, -435.648853, -648.450815, -906.368711, -1230.12634, -1634.16109, -2132.91033, -2735.20563, -3462.36425, -4341.81884, -5364.42574, -6614.3731, -8134.66116, -9980.82628, -12229.6464, -14948.1493, -18166.8674, -21943.3666, -26469.0289, -31859.5423, -37819.0792, -44938.1534, -53474.9005, -63687.4564, -75833.9568, -97056.0619, -119007.502, -140225.531, -159247.405, -174610.378, -180129.846, -182212.828, -181420.23, -179253.923, -175902.35, -171161.886, -165240.579, -158346.476, -150687.623, -142472.067, -128483.197, -119332.913, -112962.396, -107312.828, -100325.388, -89941.2594, -74101.6219, -53046.3087, -34635.8882},
        {-10.4934981, -8.91388867, -7.39067643, -5.92249504, -4.50797818, -3.1457595, -1.8344727, -0.57275144, 0.640770601, 1.80745975, 2.92868233, 4.00580467, 5.0401931, 6.03321393, 6.98623351, 7.90061814, 8.77773416, 9.6189479, 10.4256257, 11.1991338, 11.9408386, 12.6521065, 13.3343037, 13.9887965, 14.6169514, 15.2201345, 15.7997124, 16.3570511, 16.8935172, 17.4104769, 17.9092966, 18.3913425, 18.857981, 19.3105785, 19.8018512, 20.3064104, 20.7982502, 21.2770704, 21.7425711, 22.194452, 22.6324132, 23.0561546, 23.465376, 23.8597774, 24.2390587, 24.6029199, 24.9510608, 25.2831814, 25.5989816, 25.8981612, 26.1804203, 26.4782784, 26.7579042, 27.0182863, 27.2584131, 27.477273, 27.6738546, 27.8471463, 27.9961365, 28.1198136, 28.2171663, 28.2871828, 28.3288516, 28.3411613, 28.3231002, 28.2736568, 28.1918197, 28.0729538, 27.9125116, 27.7169541, 27.4855047, 27.2173866, 26.9118232, 26.5680379, 26.1852539, 25.7626947, 25.2995835, 24.7951437, 24.2485986, 23.6591716, 23.026086, 22.348565, 21.6258322, 20.8571107, 19.2570742, 17.2337227, 15.2044863, 13.1940483, 11.2270923, 9.32830167, 7.52235998, 5.83395069, 4.28775729, 2.90846325, 1.72075207, 0.749307222, 0.0188121971, -0.446049522, -0.62059445, -0.480139105, 0, 3.03407025, 6.2208714, 9.57625079, 13.1160557, 16.8561336, 20.8123317, 25.0004973, 29.4364778, 34.1361206, 39.1152729, 44.3897821, 49.9754955, 55.8882604, 62.1439242, 68.7583342, 75.7473377, 82.4667767, 88.8382119, 96.3703602, 104.937639, 114.414466, 124.675259, 135.594435, 147.046413, 158.905609, 171.046441, 183.343327, 195.670684, 207.90293, 219.840698, 231.459244, 242.655122, 253.324886, 264.896835, 276.50165, 286.861545, 295.413547, 301.594686, 304.841992, 304.592494, 304.594145, 298.66103, 285.386734, 263.364845, 231.18895, 195.065392, 144.071813, 74.8986113, -15.7638159, -131.22507, -269.94123, -446.007507, -657.166098, -908.984169, -1223.00485, -1613.5984, -2095.13507, -2677.8606, -3381.20713, -4229.10627, -5210.0645, -6408.01558, -7867.25329, -9642.19152, -11804.3921, -14417.5455, -17503.2585, -21114.311, -25457.7909, -30653.5009, -36524.5593, -43533.0968, -51901.6143, -61852.6131, -73608.5941, -93742.8556, -114539.402, -134633.036, -152658.558, -167250.77, -172637.442, -174798.43, -174257.263, -172458.099, -169560.053, -165338.646, -159972.171, -153638.924, -146517.2, -138785.292, -128044.138, -119758.794, -112686.831, -105585.819, -97213.3297, -86326.934, -71684.2029, -50850.3353, -33266.1908},
        {-11.3273078, -9.71349963, -8.15319384, -6.64519281, -5.18829892, -3.78131451, -2.42304195, -1.11228359, 0.152158212, 1.37148109, 2.5468827, 3.67956067, 4.77071265, 5.82153628, 6.8332292, 7.80698905, 8.74401348, 9.64550013, 10.5126466, 11.3466507, 12.1487098, 12.9200217, 13.6617841, 14.3751946, 15.0614507, 15.7217502, 16.3572907, 16.9692699, 17.5588853, 18.1273346, 18.6758155, 19.2055257, 19.7176626, 20.2134241, 20.7372935, 21.2679151, 21.783413, 22.28358, 22.7682092, 23.2370933, 23.6900255, 24.1267986, 24.5472055, 24.9510392, 25.3380925, 25.7081586, 26.0610302, 26.3965003, 26.7143619, 27.0144078, 27.2964311, 27.6010593, 27.8863655, 28.1512573, 28.3946426, 28.615429, 28.8125242, 28.984836, 29.131272, 29.25074, 29.3421478, 29.4044029, 29.4364132, 29.4370864, 29.4053302, 29.3400522, 29.2401602, 29.0960641, 28.8983777, 28.663535, 28.3909946, 28.080215, 27.730655, 27.341773, 26.9130276, 26.4438776, 25.9337814, 25.3821978, 24.7885852, 24.1524022, 23.4731076, 22.7501599, 21.9830176, 21.1711395, 19.5507066, 17.5188709, 15.4819582, 13.4641967, 11.489815, 9.58304153, 7.76810461, 6.06923271, 4.51065422, 3.11659759, 1.91129122, 0.918963535, 0.163842951, -0.329842112, -0.537863238, -0.435992007, 0, 2.51675038, 5.26944617, 8.26733286, 11.519656, 15.035661, 18.8245934, 22.8956987, 27.2582224, 31.9214101, 36.8945071, 42.1867591, 47.8074116, 53.7657099, 60.0708996, 66.7322263, 73.7589355, 80.5834331, 87.132559, 94.7203147, 103.232337, 112.554264, 122.571731, 133.170376, 144.235836, 155.653748, 167.30975, 179.089477, 190.878568, 202.562659, 213.999772, 225.113285, 235.797119, 245.945196, 256.843882, 267.690876, 277.312559, 285.185119, 290.784745, 293.587624, 293.069945, 293.236553, 287.656692, 274.920509, 253.618152, 222.339767, 187.873515, 138.562025, 70.9459393, -18.4340987, -133.037446, -272.902261, -449.090565, -657.957339, -904.199943, -1209.54809, -1588.08432, -2053.89116, -2617.63532, -3297.59149, -4115.7607, -5061.02596, -6213.1802, -7612.52421, -9310.8856, -11379.6236, -13881.1321, -16832.9292, -20285.7744, -24454.9716, -29462.7491, -35231.6871, -42114.377, -50302.1562, -59986.3617, -71358.331, -90489.5429, -110221.158, -129274.477, -146370.802, -160231.436, -165441.895, -167616.456, -167248.899, -165750.131, -163250.905, -159499.836, -154646.782, -148841.602, -142234.153, -134974.295, -127123.113, -119629.198, -111976.835, -103650.309, -94133.9045, -82911.9068, -69468.6004, -49100.4618, -32316.3945},
        {-12.2111996, -10.6263643, -9.08491027, -7.58608935, -6.12915336, -4.71335412, -3.33794347, -2.00217324, -0.705295239, 0.553438688, 1.77477672, 2.95946702, 4.10825777, 5.22189713, 6.30113328, 7.3467144, 8.35938864, 9.33990419, 10.2890092, 11.2074519, 12.0959804, 12.9553428, 13.7862875, 14.5895625, 15.3659159, 16.1160961, 16.8408511, 17.5409291, 18.2170783, 18.8700469, 19.500583, 20.1094348, 20.6973505, 21.2650782, 21.8375488, 22.4029104, 22.9490207, 23.4758429, 23.9833405, 24.471477, 24.9402157, 25.3895199, 25.8193532, 26.2296789, 26.6204604, 26.9916611, 27.3432444, 27.6751737, 27.9874124, 28.279924, 28.5526717, 28.8559325, 29.1382657, 29.3985438, 29.6356397, 29.848426, 30.0357754, 30.1965606, 30.3296543, 30.4339292, 30.508258, 30.5515135, 30.5625683, 30.5402951, 30.4835666, 30.3912556, 30.2622347, 30.0819961, 29.8363532, 29.5517751, 29.2280019, 28.8647737, 28.4618305, 28.0189126, 27.5357599, 27.0121127, 26.447711, 25.8422949, 25.1956045, 24.5073799, 23.7773613, 23.0052888, 22.1909024, 21.3339422, 19.6908801, 17.648724, 15.6034152, 13.5788144, 11.5987822, 9.68717912, 7.86786583, 6.16470289, 4.60155089, 3.20227043, 1.99072209, 0.990766481, 0.226264185, -0.278924205, -0.500938094, -0.41591689, 0, 2.27382717, 4.8267865, 7.66383707, 10.789938, 14.2100484, 17.9291274, 21.952134, 26.2840274, 30.9297667, 35.8943109, 41.1826193, 46.7996508, 52.7503646, 59.0397198, 65.6726755, 72.6541909, 79.5390565, 86.2679218, 93.8639839, 102.235737, 111.291676, 120.940295, 131.090089, 141.649552, 152.527178, 163.631462, 174.870899, 186.153982, 197.389207, 208.583548, 219.510909, 230.050241, 240.080495, 250.796782, 261.419973, 270.819958, 278.48084, 283.886719, 286.521698, 285.869877, 285.581136, 279.705953, 266.913385, 245.872491, 215.252328, 182.114309, 134.637048, 69.3915137, -17.0513235, -128.120495, -263.69824, -435.095349, -640.004582, -883.319411, -1184.63475, -1556.94486, -2013.24402, -2561.72236, -3221.27012, -4016.01839, -4941.26902, -6064.39984, -7410.14815, -9024.75009, -10990.8704, -13374.4525, -16202.9714, -19532.3972, -23559.7508, -28402.2669, -33993.0856, -40661.8582, -48589.5443, -57957.1034, -68945.4952, -87394.3703, -106411.325, -124762.645, -141214.617, -154533.529, -159469.903, -161482.966, -161060.047, -159639.308, -157331.707, -153852.021, -149331.017, -143899.464, -137688.127, -130827.775, -124244.825, -117240.207, -109581.048, -101034.476, -91367.6193, -80347.6049, -67741.561, -48291.409, -32116.6107},
        {-13.164245, -11.6415822, -10.148368, -8.68445167, -7.24968232, -5.84390916, -4.46698137, -3.11874813, -1.79905861, -0.507761994, 0.755292537, 1.99025581, 3.19727863, 4.37651183, 5.52810624, 6.65221266, 7.74898192, 8.81856484, 9.86111225, 10.876775, 11.8657038, 12.8280496, 13.7639631, 14.6735952, 15.5570967, 16.4146185, 17.2463113, 18.052326, 18.8328133, 19.5879242, 20.3178094, 21.0226197, 21.7025061, 22.3576192, 22.9872135, 23.5919069, 24.1723281, 24.7286569, 25.2610733, 25.7697571, 26.2548883, 26.7166468, 27.1552125, 27.5707653, 27.9634852, 28.3335519, 28.6811455, 29.0064459, 29.3096329, 29.5908864, 29.8503865, 30.1492308, 30.4253609, 30.677636, 30.9049155, 31.1060587, 31.279925, 31.4253736, 31.5412639, 31.6264552, 31.6798069, 31.7001782, 31.6864286, 31.6374173, 31.5520037, 31.4290471, 31.2674068, 31.0479843, 30.752112, 30.4156863, 30.0387305, 29.6212682, 29.1633227, 28.6649177, 28.1260765, 27.5468226, 26.9271795, 26.2671706, 25.5668195, 24.8261496, 24.0451843, 23.2239472, 22.3624617, 21.4607512, 19.7951167, 17.7426655, 15.6892034, 13.6582401, 11.6732853, 9.75784852, 7.93543949, 6.22956788, 4.66374333, 3.2614755, 2.04627404, 1.04164862, 0.271108887, -0.241835498, -0.473674878, -0.400899598, 0, 2.11172671, 4.5322022, 7.26302541, 10.3057952, 13.6621106, 17.3335704, 21.3217736, 25.6283191, 30.2548058, 35.2028326, 40.4739984, 46.0699022, 51.9921429, 58.2423193, 64.8220305, 71.7328753, 78.6462251, 85.5165036, 93.0922527, 101.304308, 110.083506, 119.360683, 129.066673, 139.132314, 149.488441, 160.065889, 170.795495, 181.608095, 192.434524, 203.423015, 214.207295, 224.652982, 234.625692, 245.270866, 255.814209, 265.14014, 272.730329, 278.066446, 280.63016, 279.903141, 279.008783, 272.67868, 259.681992, 238.787877, 208.765494, 176.794124, 131.130272, 68.4405654, -14.6083672, -121.349898, -251.022062, -416.058981, -616.220423, -856.869558, -1154.80984, -1521.70828, -1969.23188, -2501.93415, -3140.21228, -3912.22362, -4822.09921, -5922.04363, -7218.89306, -8752.2898, -10617.4643, -12884.1312, -15591.1823, -18800.3413, -22686.3481, -27360.014, -32742.3336, -39163.1379, -46798.9443, -55826.2699, -66421.6322, -84266.2234, -102655.717, -120390.462, -136270.806, -149097.099, -153736.679, -155544.911, -155007.72, -153597.329, -151409.253, -148128.046, -143869.77, -138750.486, -132886.255, -126393.139, -120313.324, -113637.507, -106217.926, -97906.8199, -88556.4255, -78018.9809, -66146.7242, -47821.6842, -32223.8586},
        {-14.1934407, -12.7578921, -11.3350087, -9.92534694, -8.52946373, -7.14791572, -5.78125965, -4.43005223, -3.09485018, -1.77621022, -0.474689071, 0.809156547, 2.07476991, 3.32159431, 4.54907302, 5.75664932, 6.94376649, 8.10986781, 9.25439656, 10.376796, 11.4765095, 12.5529802, 13.6056515, 14.6339666, 15.6373688, 16.6153014, 17.5672077, 18.4925309, 19.3907144, 20.2612014, 21.1034352, 21.916859, 22.7009162, 23.4550501, 24.1481973, 24.7956961, 25.4132514, 26.0012964, 26.5602644, 27.0905887, 27.5927026, 28.0670394, 28.5140323, 28.9341145, 29.3277195, 29.6952804, 30.0372305, 30.3540032, 30.6460316, 30.9137491, 31.1575889, 31.4503845, 31.7185995, 31.9610976, 32.1767428, 32.3643986, 32.522929, 32.6511977, 32.7480684, 32.8124051, 32.8430713, 32.838931, 32.7988479, 32.7216858, 32.6063084, 32.4515796, 32.2563631, 31.9975857, 31.6527006, 31.2658154, 30.8372159, 30.3671878, 29.8560173, 29.30399, 28.7113918, 28.0785087, 27.4056266, 26.6930312, 25.9410085, 25.1498443, 24.3198245, 23.451235, 22.5443616, 21.5994903, 19.9119816, 17.8494111, 15.7878272, 13.750392, 11.7602676, 9.84061594, 8.01459913, 6.30537922, 4.73611828, 3.32997834, 2.11012148, 1.09970975, 0.321905208, -0.200130097, -0.443234104, -0.384244757, 0, 1.9807081, 4.29215526, 6.93326807, 9.90297311, 13.200197, 16.8238662, 20.7729074, 25.0462471, 29.642812, 34.5615286, 39.8013235, 45.3611233, 51.2398545, 57.4364437, 63.9498176, 70.7789026, 77.6823894, 84.624885, 92.1487071, 100.201303, 108.730121, 117.682607, 127.00621, 136.648376, 146.556553, 156.678189, 166.96073, 177.351625, 187.798321, 198.541893, 209.128496, 219.417491, 229.268237, 239.821956, 250.297076, 259.570193, 267.116107, 272.409619, 274.925528, 274.138633, 272.614879, 265.797139, 252.555424, 231.759746, 202.280116, 171.247944, 127.206492, 66.9604212, -12.6856075, -114.926933, -238.214166, -396.286485, -591.115252, -828.571499, -1122.48808, -1483.12316, -1920.73489, -2435.52823, -3050.09937, -3798.0116, -4693.46329, -5772.73741, -7026.17069, -8486.55731, -10258.6332, -12414.1118, -14999.8924, -18081.255, -21814.4657, -26304.3237, -31455.1558, -37603.0758, -44922.277, -53586.9531, -63771.2977, -81016.4321, -98787.8902, -115918.133, -131239.624, -143584.822, -147934.176, -149540.171, -148886.133, -147463.237, -145355.869, -142228.473, -138186.428, -133335.114, -127779.913, -121626.204, -115390.281, -108893.321, -101908.537, -94209.1409, -85568.3468, -75759.3668, -64555.4137, -47477.895, -32468.8604},
        {-15.3459475, -13.9983664, -12.6464259, -11.2913996, -9.93456123, -8.57718426, -7.2205423, -5.86590894, -4.51455778, -3.16776241, -1.82679641, -0.492933379, 0.832553098, 2.14838943, 3.45330203, 4.74601731, 6.02526167, 7.28976153, 8.53824329, 9.76943338, 10.9820582, 12.1748441, 13.3465176, 14.4958051, 15.6214329, 16.7221275, 17.7966153, 18.8436227, 19.8618761, 20.8501019, 21.8070266, 22.7313765, 23.621878, 24.4772576, 25.2356394, 25.9273257, 26.583347, 27.2043964, 27.7911671, 28.344352, 28.8646442, 29.3527369, 29.809323, 30.2350957, 30.6307481, 30.9969731, 31.3344638, 31.6439134, 31.926015, 32.1814614, 32.410946, 32.6972642, 32.9572268, 33.189747, 33.393738, 33.5681132, 33.7117857, 33.8236688, 33.9026757, 33.9477198, 33.9577142, 33.9315721, 33.868207, 33.7665319, 33.6254601, 33.4439049, 33.2207795, 32.9321426, 32.5546013, 32.134205, 31.6713485, 31.1664267, 30.6198344, 30.0319664, 29.4032175, 28.7339825, 28.0246563, 27.2756337, 26.4873095, 25.6600786, 24.7943357, 23.8904757, 22.9488935, 21.9699837, 20.2630028, 18.188452, 16.1151395, 14.0658889, 12.0635238, 10.1308678, 8.29074442, 6.56597736, 4.97939017, 3.55380646, 2.31204982, 1.27694386, 0.471312162, -0.08202167, -0.360234039, -0.340501348, 0, 1.673753, 3.71996097, 6.13437007, 8.91272645, 12.0507763, 15.5442657, 19.3889408, 23.5805478, 28.1148329, 32.9875421, 38.1944217, 43.7312178, 49.5936766, 55.7775441, 62.2785666, 69.0924902, 75.9346228, 82.7606479, 90.2006364, 98.1902404, 106.665112, 115.560904, 124.813268, 134.357857, 144.130323, 154.066317, 164.101493, 174.171502, 184.211997, 194.318431, 204.208061, 213.7686, 222.887759, 232.825151, 242.782538, 251.59584, 258.7412, 263.69476, 265.932661, 264.931044, 263.094403, 256.079292, 242.788913, 222.12647, 192.995167, 162.125729, 118.636958, 59.4751772, -18.4132904, -118.082121, -237.976284, -391.391029, -581.268, -813.379513, -1099.27717, -1447.91779, -1868.2582, -2354.89043, -2934.40082, -3649.04634, -4524.74298, -5583.35717, -6810.70094, -8233.32401, -9946.80385, -12015.033, -14469.3128, -17365.086, -20877.2976, -25113.2952, -30025.4567, -35899.9414, -42897.7407, -51179.8459, -60907.2484, -77320.464, -94231.0783, -110530.202, -125108.944, -136858.416, -141010.772, -142555.381, -141951.479, -140615.599, -138628.043, -135670.242, -131842.442, -127244.89, -121977.835, -116141.523, -109035.116, -102524.7, -96146.3108, -89435.9857, -81929.7614, -73163.6751, -62673.7634, -46699.3599, -32353.6393},
        {-16.5634187, -15.3221557, -14.0575448, -12.771651, -11.466539, -10.1442737, -8.80692005, -7.45654286, -6.09520701, -4.72497737, -3.34791879, -1.96609616, -0.581574331, 0.803581823, 2.18730744, 3.56753764, 4.94220757, 6.30925235, 7.66660713, 9.01220702, 10.3439872, 11.6598827, 12.9578288, 14.2357605, 15.491613, 16.7233214, 17.9288209, 19.1060465, 20.2529335, 21.3674169, 22.4474319, 23.4909136, 24.4957972, 25.4600177, 26.2875399, 27.0252608, 27.7212065, 28.3763618, 28.9917113, 29.5682397, 30.1069316, 30.6087717, 31.0747447, 31.5058351, 31.9030276, 32.2673069, 32.5996577, 32.9010646, 33.1725122, 33.4149852, 33.6294683, 33.9106443, 34.1637684, 34.3877942, 34.5816751, 34.7443645, 34.874816, 34.9719828, 35.0348187, 35.0622768, 35.0533109, 35.0068742, 34.9219203, 34.7974026, 34.6322745, 34.4254897, 34.1760014, 33.8588146, 33.4495093, 32.996431, 32.5000856, 31.960979, 31.3796172, 30.7565061, 30.0921516, 29.3870598, 28.6417366, 27.8566879, 27.0324196, 26.1694377, 25.2682481, 24.3293568, 23.3532697, 22.3404928, 20.6153537, 18.5306701, 16.4473867, 14.3879526, 12.374817, 10.4304293, 8.5772387, 6.83769433, 5.23424546, 3.78934133, 2.52543115, 1.46496416, 0.630389586, 0.0441566548, -0.271285405, -0.293487366, 0, 1.39699365, 3.19899869, 5.39939228, 7.99155156, 10.9688537, 14.3246759, 18.0523952, 22.1453889, 26.597034, 31.4007078, 36.5497874, 42.03765, 47.8576726, 54.0032326, 60.467707, 67.2444729, 73.978165, 80.6123399, 87.9291976, 95.8473845, 104.285547, 113.162333, 122.396387, 131.906357, 141.610889, 151.42863, 161.278226, 171.078325, 180.747572, 190.199564, 199.359849, 208.14859, 216.485947, 225.757186, 235.149916, 243.461054, 250.171231, 254.761076, 256.711221, 255.502297, 253.447943, 246.333561, 233.085498, 212.630099, 183.893712, 153.130701, 110.10739, 51.9181194, -24.342768, -121.580931, -238.419064, -387.280523, -571.554872, -796.739588, -1072.63581, -1407.11241, -1808.03827, -2265.20412, -2808.66565, -3488.20026, -4339.34215, -5372.65961, -6574.03631, -7964.95044, -9627.19314, -11614.8709, -13940.1454, -16646.2486, -19930.2101, -23902.0629, -28562.0738, -34146.9288, -40804.657, -48683.2872, -57930.8483, -73473.3718, -89486.3832, -104923.411, -118737.983, -129883.628, -133868.126, -135387.919, -134873.283, -133642.606, -131774.622, -128978.018, -125350.029, -120987.89, -115988.837, -110450.106, -102126.61, -95507.071, -89810.7205, -84256.7904, -78064.5127, -70453.1192, -60641.8418, -45837.6878, -32243.6757},
        {-17.8259259, -16.7116866, -15.5530015, -14.3528056, -13.1140343, -11.8396227, -10.532506, -9.19561952, -7.83189836, -6.44427777, -5.03569295, -3.6090791, -2.16737144, -0.713505179, 0.749584484, 2.21896234, 3.69169317, 5.16484179, 6.63547296, 8.1006515, 9.55744219, 11.0029098, 12.4341192, 13.8481351, 15.2420223, 16.6128456, 17.9576699, 19.2735598, 20.5575802, 21.8067959, 23.0182717, 24.1890723, 25.3162626, 26.3969074, 27.2971615, 28.0822982, 28.8191641, 29.5090725, 30.1533362, 30.7532685, 31.3101823, 31.8253908, 32.300207, 32.735944, 33.1339148, 33.4954325, 33.8218103, 34.1143611, 34.374398, 34.6032341, 34.8021826, 35.0811114, 35.3304253, 35.5490842, 35.7360479, 35.8902762, 36.0107288, 36.0963657, 36.1461466, 36.1590313, 36.1339796, 36.0699514, 35.9659064, 35.8208045, 35.6336054, 35.4032691, 35.1287552, 34.7826432, 34.3381458, 33.8486305, 33.3147673, 32.737226, 32.1166767, 31.4537893, 30.7492337, 30.0036799, 29.2177977, 28.3922573, 27.5277284, 26.6248812, 25.6843854, 24.706911, 23.693128, 22.6437064, 20.9016518, 18.8102994, 16.7208889, 14.6554328, 12.6359435, 10.6844334, 8.82291486, 7.07340039, 5.45790236, 3.99843319, 2.7170053, 1.6356311, 0.776323007, 0.161093434, -0.188045201, -0.249080484, 0, 1.20612183, 2.83224444, 4.87059102, 7.31338471, 10.1528487, 13.3812061, 16.9906801, 20.9734939, 25.3218707, 30.0280335, 35.0842056, 40.4826101, 46.2154702, 52.2750091, 58.6534499, 65.3430158, 71.9191958, 78.3146795, 85.4643032, 93.2709899, 101.637662, 110.467244, 119.662657, 129.126826, 138.762672, 148.473119, 158.161089, 167.729507, 177.081294, 185.971345, 194.504889, 202.629258, 210.291784, 218.965396, 227.834628, 235.677187, 241.982726, 246.240898, 247.941355, 246.573751, 244.380025, 237.259891, 224.164509, 204.045042, 175.852651, 145.397776, 103.05748, 46.0681064, -28.3340035, -122.912508, -236.607506, -380.679452, -558.395982, -774.842565, -1038.81607, -1357.44702, -1737.86594, -2166.99268, -2676.76178, -3321.15578, -4140.99357, -5140.51011, -6309.41526, -7665.963, -9276.77368, -11186.2667, -13389.3267, -15917.5308, -18984.0831, -22699.4493, -27095.5812, -32373.5371, -38670.9606, -46125.495, -54874.7838, -69544.5988, -84662.4229, -99243.8676, -112304.544, -122860.064, -126696.652, -128212.23, -127806.432, -126691.993, -124946.694, -122309.524, -118875.203, -114738.452, -109993.99, -104736.538, -95063.1245, -88284.8224, -83276.5385, -78913.179, -74069.6501, -67620.858, -58441.709, -44912.2953, -32190.0393},
        {-19.0363347, -18.0791501, -17.0536787, -15.9638286, -14.8135075, -13.6066232, -12.3470838, -11.0387969, -9.68567051, -8.29161244, -6.86053055, -5.39633271, -3.90292677, -2.38422058, -0.844122015, 0.713461078, 2.28462084, 3.86544941, 5.45203893, 7.04048154, 8.62686939, 10.2072946, 11.7778494, 13.3346258, 14.873716, 16.3912121, 17.8832064, 19.3457908, 20.7750577, 22.167099, 23.518007, 24.8238738, 26.0807916, 27.2848524, 28.2588146, 29.0890832, 29.8641463, 30.5857211, 31.2555247, 31.8752743, 32.4466871, 32.9714802, 33.4513708, 33.888076, 34.2833129, 34.6387989, 34.9562509, 35.2373861, 35.4839218, 35.697575, 35.8800629, 36.166596, 36.4225042, 36.6466109, 36.8377392, 36.9947126, 37.1163544, 37.2014876, 37.2489358, 37.2575221, 37.2260699, 37.1534023, 37.0383428, 36.8797146, 36.6763409, 36.4270452, 36.1306505, 35.7514197, 35.2544451, 34.7098288, 34.1186345, 33.4819257, 32.8007664, 32.0762201, 31.3093506, 30.5012215, 29.6528967, 28.7654397, 27.8399144, 26.8773844, 25.8789134, 24.8455651, 23.7784033, 22.6784916, 20.9194352, 18.8292248, 16.7433928, 14.6833651, 12.6705675, 10.7264262, 8.87236699, 7.12981585, 5.52019873, 4.06494157, 2.78547033, 1.70321094, 0.839589359, 0.216031527, -0.146036608, -0.225189099, 0, 1.30928964, 3.00369866, 5.0770937, 7.52334141, 10.3363084, 13.5098614, 17.037867, 20.9141918, 25.1327025, 29.6872658, 34.5717482, 39.7800164, 45.3059372, 51.143377, 57.2862026, 63.7282807, 70.0802146, 76.2811194, 83.1917661, 90.7238936, 98.7892408, 107.299547, 116.16655, 125.301989, 134.617603, 144.025131, 153.436312, 162.762885, 171.916588, 180.735652, 189.232278, 197.340255, 204.993375, 213.539964, 222.206546, 229.842434, 235.957059, 240.059851, 241.660241, 240.267659, 237.905447, 230.804021, 217.980958, 198.453836, 171.240234, 142.118855, 101.65587, 47.1785761, -23.9857301, -114.509752, -223.75707, -361.753993, -530.603449, -734.875794, -984.179051, -1285.85943, -1647.26316, -2058.92718, -2548.82725, -3163.27651, -3934.40022, -4874.07834, -5979.76309, -7271.67333, -8808.60478, -10631.6458, -12738.2126, -15158.0713, -18081.1183, -21606.6713, -25718.6937, -30649.6041, -36538.6502, -43525.0798, -51748.1408, -65639.9808, -79973.7229, -93815.39, -106231.005, -116286.591, -119995.669, -121511.765, -121209.117, -120226.342, -118635.112, -116188.605, -112974.416, -109080.14, -104593.373, -99601.7092, -88734.682, -81768.4939, -77289.2256, -73882.9578, -70135.7711, -64633.7462, -55962.9638, -44038.2379, -32490.6135},
        {-20.2695359, -19.4932309, -18.6225095, -17.6623363, -16.617676, -15.4934932, -14.2947525, -13.0264187, -11.6934562, -10.3008299, -8.8535042, -7.3564439, -5.81461358, -4.2329779, -2.61650149, -0.970148983, 0.701114968, 2.39232573, 4.09851865, 5.81472911, 7.53599245, 9.25734404, 10.9738192, 12.6804534, 14.3722819, 16.0443401, 17.6916633, 19.3092869, 20.8922463, 22.4355768, 23.9343138, 25.3834927, 26.7781487, 28.1133173, 29.1644752, 30.0410551, 30.8551288, 31.6088568, 32.3043995, 32.9439174, 33.529571, 34.0635207, 34.547927, 34.9849503, 35.3767512, 35.72549, 36.0333272, 36.3024234, 36.5349389, 36.7330342, 36.8988697, 37.1977906, 37.4652352, 37.6998263, 37.9001871, 38.0649403, 38.192709, 38.282116, 38.3317844, 38.3403371, 38.3063969, 38.2285869, 38.10553, 37.9358491, 37.7181672, 37.4511071, 37.133292, 36.717683, 36.1583381, 35.5480535, 34.8884123, 34.1809975, 33.4273921, 32.6291791, 31.7879414, 30.9052621, 29.9827241, 29.0219105, 28.0244043, 26.9917883, 25.9256457, 24.8275594, 23.6991124, 22.5418877, 20.7672161, 18.6842219, 16.6091237, 14.5626592, 12.5655663, 10.6385829, 8.80244663, 7.07789543, 5.48566707, 4.04649937, 2.78113014, 1.71029718, 0.854738321, 0.23519136, -0.127605888, -0.212915612, 0, 1.57494795, 3.47120326, 5.68602925, 8.21668924, 11.0604466, 14.2145645, 17.6763065, 21.4429357, 25.5117155, 29.8799093, 34.5447803, 39.5035919, 44.7536074, 50.2920902, 56.1163035, 62.2235106, 68.301733, 74.3040565, 80.9272204, 88.1022217, 95.7600577, 103.831726, 112.248223, 120.940546, 129.839693, 138.87666, 147.982445, 157.088045, 166.124457, 175.118797, 183.870699, 192.282328, 200.255843, 208.915916, 217.553454, 225.133364, 231.191054, 235.261933, 236.881412, 235.5849, 233.119554, 226.135896, 213.7366, 195.024343, 169.1018, 141.902616, 103.990758, 52.7611576, -14.3912491, -100.071529, -204.116737, -335.283073, -493.940986, -684.087554, -917.124219, -1201.12801, -1544.17597, -1943.80593, -2421.02938, -3008.35515, -3722.10336, -4588.82164, -5615.54672, -6827.04992, -8278.1528, -10009.8833, -12031.9991, -14377.4209, -17193.728, -20563.6423, -24385.6686, -28955.0872, -34418.3874, -40922.0586, -48612.5903, -61784.2664, -75398.8332, -88565.8315, -100394.802, -109995.286, -113576.211, -115081.472, -114863.73, -114006.707, -112573.686, -110323.338, -107334.175, -103684.711, -99453.4576, -94718.9264, -82935.8217, -75829.9805, -71768.3026, -69117.6879, -66245.0361, -61517.247, -53301.2203, -43117.4756, -32893.3493},
        {-21.5849875, -21.0085747, -20.309683, -19.4944037, -18.5688277, -17.5390462, -16.4111504, -15.1912312, -13.8853797, -12.4996871, -11.0402445, -9.51314291, -7.92447346, -6.28032724, -4.58679535, -2.84996886, -1.07593889, 0.729203488, 2.55936717, 4.40846107, 6.27039409, 8.13907514, 10.0084131, 11.872317, 13.7246955, 15.5594578, 17.3705126, 19.1517689, 20.8971355, 22.6005215, 24.2558356, 25.8569869, 27.3978841, 28.8724363, 30.0063284, 30.9332229, 31.7900064, 32.5792921, 33.3036934, 33.9658234, 34.5682956, 35.1137232, 35.6047194, 36.0438977, 36.4338712, 36.7772532, 37.0766571, 37.3346962, 37.5539836, 37.7371328, 37.886757, 38.1988601, 38.4784725, 38.724015, 38.9339088, 39.1065747, 39.2404337, 39.3339067, 39.3854148, 39.3933789, 39.3562199, 39.2723588, 39.1402166, 38.9582143, 38.7247727, 38.4383129, 38.0972558, 37.6425174, 37.0163942, 36.3359071, 35.6032047, 34.820436, 33.9897495, 33.1132943, 32.1932189, 31.2316722, 30.230803, 29.19276, 28.119692, 27.0137478, 25.8770762, 24.7118259, 23.5201457, 22.3041845, 20.5165058, 18.4456108, 16.3866062, 14.3594759, 12.3842036, 10.4807734, 8.66916908, 6.96937448, 5.40137348, 3.98514997, 2.74068781, 1.68797087, 0.846983029, 0.237708149, -0.119869896, -0.205767236, 0, 1.90202279, 4.04877391, 6.44185794, 9.08287944, 11.973443, 15.1151531, 18.5096144, 22.1584315, 26.0632088, 30.2255511, 34.6470627, 39.3293484, 44.2740127, 49.48266, 54.9568951, 60.6983225, 66.4675022, 72.231604, 78.537452, 85.3349984, 92.5741953, 100.204995, 108.17735, 116.441211, 124.946532, 133.643265, 142.481361, 151.410773, 160.381453, 169.610329, 178.682486, 187.467784, 195.836083, 204.676764, 213.349926, 220.943794, 227.02237, 231.149658, 232.889659, 231.806376, 229.379602, 222.674482, 210.880191, 193.185904, 168.780793, 143.768907, 108.695828, 61.0220115, -1.79208741, -82.286013, -180.766229, -304.734681, -452.651551, -627.926425, -844.073937, -1110.07176, -1434.89756, -1824.05697, -2290.78881, -2852.11959, -3506.83819, -4297.65956, -5241.1738, -6367.22379, -7728.10849, -9365.81916, -11304.7721, -13582.4019, -16300.5853, -19524.7467, -23063.1844, -27278.0072, -32323.6451, -38354.5279, -45525.0856, -58009.4788, -70937.4739, -83458.5681, -94722.2584, -103878.042, -107311.056, -108778.064, -108613.29, -107857.076, -106564.14, -104494.984, -101719.034, -98305.7202, -94324.4704, -89844.713, -77559.3603, -70434.3559, -66707.387, -64616.1409, -62398.3048, -58291.5657, -50533.6111, -42071.0853, -33195.5332},
        {-23.1223766, -22.7663403, -22.2562305, -21.5994224, -20.8032911, -19.8752119, -18.8225598, -17.6527102, -16.3730382, -14.9909189, -13.5137276, -11.9488394, -10.3036295, -8.58547308, -6.80174533, -4.95982143, -3.06707655, -1.13088588, 0.841375401, 2.84233212, 4.86460908, 6.90083112, 8.94362305, 10.9856097, 13.0194159, 15.0376664, 17.0329861, 18.9979997, 20.9253322, 22.8076083, 24.6374529, 26.4074907, 28.1103466, 29.7386454, 30.9603173, 31.9371638, 32.8348894, 33.6566557, 34.4056246, 35.0849578, 35.6978171, 36.2473643, 36.7367612, 37.1691696, 37.5477512, 37.8756679, 38.1560815, 38.3921536, 38.5870462, 38.7439209, 38.8659397, 39.1828793, 39.4663739, 39.7146726, 39.9260248, 40.0986797, 40.2308863, 40.320894, 40.366952, 40.3673095, 40.3202157, 40.2239198, 40.076671, 39.8767186, 39.6223117, 39.3116996, 38.9431315, 38.4463087, 37.7525782, 37.0015643, 36.1959598, 35.3384576, 34.4317507, 33.478532, 32.4814943, 31.4433305, 30.3667336, 29.2543964, 28.1090119, 26.9332729, 25.7298724, 24.5015032, 23.2508583, 21.9806304, 20.1864494, 18.1352051, 16.0998819, 14.099628, 12.1535913, 10.28092, 8.50076217, 6.83226583, 5.2945791, 3.90685003, 2.68822673, 1.65785725, 0.834889677, 0.23847209, -0.112247437, -0.198120826, 0, 2.18347441, 4.53725599, 7.06765406, 9.78097797, 12.6835371, 15.7816406, 19.0815981, 22.5897187, 26.3123119, 30.2556869, 34.4261531, 38.8300199, 43.4735965, 48.3631924, 53.5051168, 58.905679, 64.3119526, 69.6932131, 75.6421368, 82.1094826, 89.0460092, 96.4024754, 104.12964, 112.178262, 120.499099, 129.042912, 137.760458, 146.602496, 155.519785, 164.745247, 173.844017, 182.682207, 191.125925, 199.967604, 208.610192, 216.20962, 222.354103, 226.631854, 228.631087, 227.940016, 225.849966, 219.727701, 208.828218, 192.406513, 169.717583, 146.890306, 114.587325, 70.3451661, 11.700355, -63.8105829, -156.923955, -273.941384, -411.669657, -572.967233, -773.048072, -1021.57877, -1328.22591, -1704.44393, -2157.70525, -2693.53195, -3298.90245, -4025.97111, -4896.69438, -5943.34884, -7215.69792, -8755.47832, -10595.1656, -12774.0954, -15365.5608, -18423.6469, -21707.3448, -25611.232, -30291.5412, -35904.5052, -42606.3564, -54426.783, -66680.1984, -78554.4715, -89237.4712, -97917.0663, -101138.486, -102493.999, -102304.033, -101573.096, -100350.835, -98401.8643, -95790.0841, -92579.3932, -88833.6905, -84616.8754, -72835.1215, -66044.5914, -62553.5089, -60670.098, -58702.5824, -54959.1861, -47748.1328, -40786.5022, -33116.1582},
        {-24.9710939, -24.8553292, -24.5496485, -24.0629094, -23.4039693, -22.5816859, -21.6049167, -20.4825193, -19.2233512, -17.8362701, -16.3301334, -14.7137987, -12.9961237, -11.1859658, -9.29218264, -7.32363176, -5.28917073, -3.19765712, -1.05794849, 1.1210976, 3.33062359, 5.5617719, 7.80568499, 10.0535053, 12.2963752, 14.5254372, 16.7318337, 18.9067072, 21.0412001, 23.1264548, 25.1536137, 27.1138194, 28.9982141, 30.7979405, 32.1128544, 33.1379717, 34.0730112, 34.9218134, 35.6882186, 36.3760674, 36.9892001, 37.5314572, 38.0066791, 38.4187062, 38.7713791, 39.0685381, 39.3140236, 39.5116761, 39.665336, 39.7788439, 39.85604, 40.1636229, 40.4368768, 40.6739437, 40.872966, 41.0320857, 41.1494451, 41.2231863, 41.2514516, 41.232383, 41.1641229, 41.0448133, 40.8725966, 40.6456147, 40.36201, 40.0199247, 39.6175008, 39.0754746, 38.3164407, 37.4981854, 36.6238688, 35.6966511, 34.7196923, 33.6961526, 32.629192, 31.5219706, 30.3776486, 29.1993861, 27.9903431, 26.7536799, 25.4925563, 24.2101327, 22.9095691, 21.5940255, 19.8022565, 17.7793932, 15.7761483, 13.8107529, 11.9014379, 10.0664343, 8.32397299, 6.69228501, 5.18960128, 3.83415277, 2.64417045, 1.63788526, 0.833528166, 0.249330135, -0.096477877, -0.185664909, 0, 2.33717546, 4.78447581, 7.35275669, 10.0528737, 12.8956826, 15.8920389, 19.0527983, 22.3888165, 25.910949, 29.6300516, 33.5569798, 37.7025894, 42.0777359, 46.693275, 51.5600624, 56.6889536, 61.6736055, 66.465274, 72.0184786, 78.2546753, 85.0953199, 92.4618684, 100.275777, 108.458501, 116.931496, 125.616219, 134.434125, 143.30667, 152.155311, 160.937637, 169.52572, 177.830177, 185.761623, 194.254694, 202.707997, 210.247062, 216.475299, 220.996116, 223.412923, 223.329128, 221.978236, 216.837445, 207.191231, 192.324072, 171.520445, 150.661684, 120.786195, 79.5292421, 24.526088, -46.5880039, -134.896518, -245.611419, -374.494435, -523.935895, -709.859523, -942.112828, -1230.54331, -1588.32325, -2021.21976, -2531.37293, -3105.04303, -3791.22249, -4610.14999, -5591.65714, -6781.83026, -8219.29449, -9931.57361, -11954.4146, -14364.2695, -17214.5653, -20288.3817, -23951.0848, -28349.6833, -33631.1859, -39942.6013, -51113.9707, -62688.5121, -73892.4761, -83952.1131, -92093.6737, -95009.3938, -96148.8816, -95824.3938, -95009.6793, -93755.1543, -91835.6133, -89315.2534, -86258.2721, -82728.8665, -78791.2339, -68899.4043, -62947.9308, -59585.9057, -57462.4214, -55226.5703, -51527.4448, -45014.1373, -39181.3812, -32448.0447},
        {-26.8909579, -27.0212645, -26.926643, -26.6173624, -26.1036915, -25.3958993, -24.5042547, -23.4390267, -22.2104841, -20.8288959, -19.3045311, -17.6476584, -15.868547, -13.9774656, -11.9846832, -9.9004688, -7.73509124, -5.49881946, -3.20192239, -0.854668961, 1.53267191, 3.94983129, 6.38654027, 8.83252991, 11.2775313, 13.7112755, 16.1234936, 18.5039167, 20.8422758, 23.128302, 25.3517264, 27.5022801, 29.5696941, 31.5436996, 32.9636954, 34.0540508, 35.0445479, 35.9395316, 36.7433474, 37.4603404, 38.0948558, 38.6512389, 39.1338348, 39.5469888, 39.8950461, 40.1823519, 40.4132515, 40.59209, 40.7232127, 40.8109649, 40.8596916, 41.1594525, 41.4226848, 41.6475399, 41.8321693, 41.9747245, 42.073357, 42.1262183, 42.1314598, 42.0872331, 41.9916896, 41.8429809, 41.6392584, 41.3786736, 41.0593781, 40.6795232, 40.2372606, 39.6470062, 38.8251867, 37.9426514, 37.0029802, 36.0097526, 34.9665486, 33.8769477, 32.7445297, 31.5728742, 30.3655611, 29.12617, 27.8582806, 26.5654727, 25.251326, 23.9194202, 22.5733349, 21.21665, 19.428873, 17.4345395, 15.4632821, 13.5324424, 11.6593621, 9.86138283, 8.15584634, 6.56009426, 5.09146827, 3.76731004, 2.60496123, 1.62176353, 0.835058597, 0.262188107, -0.0795062706, -0.172682863, 0, 2.44423907, 4.94610649, 7.52071068, 10.1831601, 12.9485631, 15.8320281, 18.8486636, 22.0135779, 25.3418796, 28.848677, 32.5490784, 36.4581925, 40.5911275, 44.9629919, 49.5888941, 54.4839425, 59.0730494, 63.2870094, 68.4758986, 74.5283546, 81.3330155, 88.7785188, 96.7535025, 105.146604, 113.846462, 122.741714, 131.720997, 140.67295, 149.48621, 157.809314, 165.859039, 173.59605, 180.981017, 189.108742, 197.372826, 204.864433, 211.203988, 216.01192, 218.908656, 219.514622, 219.067768, 215.07871, 206.851402, 193.6898, 174.897859, 155.934512, 128.409993, 90.089513, 38.738282, -27.8784884, -111.015856, -215.08579, -335.219544, -473.443206, -645.859824, -862.25218, -1132.40305, -1469.83086, -1879.73374, -2363.23459, -2908.73964, -3559.03069, -4330.89532, -5250.60428, -6360.49704, -7695.53724, -9275.29025, -11130.7276, -13349.2812, -15986.4458, -18869.1032, -22316.5322, -26463.3629, -31444.2253, -37393.7491, -47930.1545, -58833.5839, -69367.7697, -78796.4445, -86383.3407, -88957.2927, -89840.1964, -89339.3118, -88410.0902, -87103.216, -85199.2943, -82762.5137, -79857.0628, -76547.1302, -72896.9043, -65275.6784, -60280.1475, -56984.3368, -54462.2717, -51787.9776, -48035.4798, -42278.8036, -37371.396, -31467.7701},
        {-28.8054375, -29.1802259, -29.2981061, -29.1705718, -28.8091171, -28.2252357, -27.4304214, -26.4361683, -25.2539701, -23.8953207, -22.371714, -20.6946439, -18.8756041, -16.9260887, -14.8575913, -12.681606, -10.4096266, -8.0531469, -5.62366083, -3.13266224, -0.591645, 1.98789702, 4.59446997, 7.21657996, 9.84273313, 12.4614356, 15.0611936, 17.6305131, 20.1579003, 22.6318614, 25.0409025, 27.3735296, 29.618249, 31.7635668, 33.3039021, 34.485197, 35.5593322, 36.5308477, 37.4042839, 38.184181, 38.8750791, 39.4815185, 40.0080394, 40.4591819, 40.8394863, 41.1534928, 41.4057417, 41.600773, 41.743127, 41.8373439, 41.887964, 42.1864521, 42.4442039, 42.6595394, 42.830779, 42.9562428, 43.0342512, 43.0631244, 43.0411826, 42.966746, 42.8381349, 42.6536696, 42.4116702, 42.110457, 41.7483503, 41.3236702, 40.8347371, 40.1931957, 39.3127717, 38.3705472, 37.3704611, 36.3164526, 35.2124606, 34.0624243, 32.8702827, 31.6399748, 30.3754398, 29.0806166, 27.7594444, 26.4158622, 25.0538091, 23.6772241, 22.2900462, 20.8962146, 19.1115136, 17.142124, 15.1988917, 13.2983523, 11.457041, 9.69149332, 8.01824472, 6.4538306, 5.01478638, 3.71764749, 2.57894934, 1.61522737, 0.843017005, 0.278853659, -0.0607272406, -0.159190268, 0, 2.50472077, 5.02495883, 7.57945466, 10.1869487, 12.8661815, 15.6358935, 18.5148251, 21.5217169, 24.6753093, 27.9943427, 31.4975577, 35.2036948, 39.1314943, 43.2996968, 47.7270428, 52.4322726, 56.6921901, 60.4195446, 65.3112671, 71.2271116, 78.026832, 85.5701824, 93.7169169, 102.326789, 111.259554, 120.374964, 129.532775, 138.59274, 147.414613, 155.375499, 162.994773, 170.276982, 177.226676, 185.055485, 193.164417, 200.63259, 207.102429, 212.216358, 215.616801, 216.946181, 217.500431, 214.765225, 208.059476, 196.702095, 180.011995, 162.836984, 137.584853, 102.19229, 54.5959812, -7.26738513, -84.5806422, -181.363723, -292.60896, -420.115765, -579.581345, -780.463081, -1032.21835, -1347.68115, -1732.16484, -2187.2992, -2705.02539, -3320.75861, -4047.35728, -4907.19707, -5938.78005, -7173.20602, -8620.92826, -10307.6938, -12334.9534, -14761.3357, -17466.348, -20715.2971, -24628.8148, -29327.5331, -34932.0838, -44843.0328, -55081.0947, -64946.918, -73741.1513, -80764.443, -82974.6706, -83577.101, -82876.4272, -81824.7551, -80470.5105, -78592.3636, -76252.2278, -73512.0163, -70433.6427, -67079.0204, -61791.7159, -57748.7811, -54474.511, -51493.2004, -48329.1444, -44506.6381, -39549.9763, -35373.8933, -30211.7007},
        {-30.6128386, -31.222611, -31.5491402, -31.6048194, -31.4020415, -30.9531996, -30.2706868, -29.3668961, -28.2542206, -26.9450532, -25.4517871, -23.7868153, -21.9625308, -19.9913267, -17.885596, -15.6577317, -13.320127, -10.8851748, -8.36526827, -5.77280035, -3.12016412, -0.419752628, 2.31604108, 5.07482396, 7.84420297, 10.6117851, 13.3651772, 16.0919863, 18.7798194, 21.4162834, 23.9889852, 26.4855318, 28.8935302, 31.2005873, 32.88134, 34.1906698, 35.389624, 36.4824564, 37.4734208, 38.3667711, 39.166761, 39.8776443, 40.503675, 41.0491068, 41.5181935, 41.915189, 42.2443471, 42.5099216, 42.7161664, 42.8673352, 42.9676819, 43.2772904, 43.5390386, 43.7516343, 43.9137855, 44.0242002, 44.0815861, 44.0846513, 44.0321037, 43.9226513, 43.7550018, 43.5278633, 43.2399437, 42.889951, 42.4765929, 41.9985776, 41.4546128, 40.7584223, 39.82577, 38.8306377, 37.7772425, 36.6698014, 35.5125317, 34.3096503, 33.0653745, 31.7839212, 30.4695078, 29.1263512, 27.7586685, 26.370677, 24.9665936, 23.5506355, 22.1270199, 20.6999638, 18.9139122, 16.9605356, 15.0358287, 13.1556839, 11.3359936, 9.59265031, 7.9415463, 6.39857405, 4.97962595, 3.70059442, 2.57737186, 1.62585069, 0.861923322, 0.301482157, -0.0395803917, -0.145371914, 0, 2.53637813, 5.05688471, 7.5830734, 10.1364978, 12.7387117, 15.4112686, 18.1757223, 21.0536264, 24.0665345, 27.2360003, 30.5835775, 34.1308197, 37.8992805, 41.9105137, 46.1860729, 50.7475118, 54.7871423, 58.2080551, 62.9074693, 68.7292441, 75.5172384, 83.1153112, 91.3673216, 100.117129, 109.208591, 118.485569, 127.79192, 136.971503, 145.868179, 153.715306, 161.191092, 168.322547, 175.136678, 182.835176, 190.857709, 198.330697, 204.928745, 210.326456, 214.198437, 216.219291, 217.800374, 216.350963, 211.205863, 201.69988, 187.167818, 171.615074, 148.538214, 116.104397, 72.4807807, 15.834522, -54.5918426, -143.020885, -244.933006, -362.092481, -509.093168, -694.768704, -927.952725, -1220.10976, -1576.84463, -2000.89207, -2487.41705, -3065.52302, -3745.12197, -4545.28172, -5500.58416, -6638.48993, -7961.37928, -9490.21054, -11337.9036, -13565.3598, -16101.0809, -19158.7367, -22845.059, -27266.7794, -32530.6298, -41821.5645, -51398.537, -60598.7235, -68759.3003, -75217.4434, -77053.8971, -77366.5574, -76459.6426, -75303.1498, -73937.76, -72126.7464, -69924.6603, -67386.0531, -64565.4761, -61517.4807, -58274.2053, -55040.1921, -51753.6212, -48352.6724, -44775.5258, -40960.3612, -36845.3586, -33205.9655, -28719.2451},
        {-31.968384, -32.7907152, -33.3157079, -33.5559579, -33.5240609, -33.2326127, -32.6942092, -31.9214459, -30.9269187, -29.7232233, -28.3229555, -26.738711, -24.9830856, -23.0686751, -21.0080752, -18.8138816, -16.4986901, -14.0750965, -11.5556965, -8.95308594, -6.27986047, -3.54861588, -0.771947934, 2.03754762, 4.86727502, 7.70463852, 10.5370424, 13.3518908, 16.136588, 18.8785383, 21.565146, 24.1838152, 26.7219502, 29.1669553, 31.0265735, 32.5381288, 33.94471, 35.2492874, 36.4548312, 37.5643115, 38.5806984, 39.5069622, 40.3460731, 41.1010011, 41.7747165, 42.3701894, 42.89039, 43.3382886, 43.7168551, 44.0290599, 44.2778731, 44.6248523, 44.9109424, 45.1356757, 45.2985844, 45.3992009, 45.4370575, 45.4116865, 45.3226202, 45.1693909, 44.9515309, 44.6685725, 44.320048, 43.9054898, 43.4244301, 42.8764011, 42.2609354, 41.5038306, 40.5326551, 39.4990128, 38.4072159, 37.2615767, 36.0664074, 34.8260202, 33.5447273, 32.2268409, 30.8766732, 29.4985363, 28.0967427, 26.6756043, 25.2394335, 23.7925424, 22.3392433, 20.8838483, 19.0787063, 17.1115142, 15.1741994, 13.2824959, 11.4521379, 9.69885945, 8.03839467, 6.48647766, 5.0588425, 3.7712233, 2.63935416, 1.67896917, 0.90580244, 0.335588057, -0.0159398767, -0.133047262, 0, 2.727979, 5.39692897, 8.0314464, 10.6561278, 13.2955696, 15.9743683, 18.7171205, 21.5484225, 24.492871, 27.5750623, 30.8195929, 34.2510594, 37.8940582, 41.7731859, 45.9130388, 50.3382135, 54.3309147, 57.8096829, 62.4783341, 68.2023809, 74.8473357, 82.2787112, 90.3620197, 98.962774, 107.946486, 117.17867, 126.524836, 135.850498, 145.021169, 153.525106, 161.743403, 169.654748, 177.237832, 185.260415, 193.30665, 200.750271, 207.333244, 212.797536, 216.885116, 219.337949, 221.86626, 221.644725, 218.030217, 210.379611, 198.049781, 183.652596, 162.476185, 133.063671, 93.9581789, 43.7028333, -17.1743784, -94.5411393, -185.693092, -292.800969, -427.988807, -598.911814, -813.225197, -1080.7271, -1406.44276, -1793.05967, -2234.78287, -2760.74211, -3382.33194, -4118.17178, -4999.12315, -6050.42827, -7272.90826, -8685.47394, -10396.58, -12464.0301, -14834.1209, -17693.3179, -21138.023, -25264.6385, -30169.5665, -38846.8852, -47770.9015, -56313.5986, -63846.9595, -69742.9674, -71186.2014, -71194.3176, -70076.7581, -68885.5603, -67636.2221, -66034.8261, -64112.3409, -61899.7349, -59427.9765, -56728.0341, -54539.3675, -51636.1384, -48243.3851, -44586.1459, -40889.4593, -37378.3635, -34277.8968, -30882.2458, -27059.2011},
        {-33.3787552, -34.3958722, -35.1064455, -35.5230169, -35.658128, -35.5243204, -35.134136, -34.5001163, -33.634803, -32.5507378, -31.2604624, -29.7765185, -28.1114477, -26.2777918, -24.2880924, -22.1548912, -19.8907298, -17.50815, -15.0196934, -12.4379018, -9.77531668, -7.04447988, -4.25793303, -1.42821782, 1.43212408, 4.31055099, 7.19452122, 10.0714931, 12.928925, 15.7542751, 18.5350018, 21.2585635, 23.9124184, 26.4840249, 28.531865, 30.266908, 31.9052795, 33.4483587, 34.8975252, 36.2541583, 37.5196374, 38.6953421, 39.7826517, 40.7829456, 41.6976033, 42.5280042, 43.2755277, 43.9415533, 44.5274603, 45.0346283, 45.4644365, 45.8608261, 46.1816921, 46.4274915, 46.5986809, 46.6957172, 46.7190572, 46.6691575, 46.546475, 46.3514664, 46.0845886, 45.7462982, 45.3370522, 44.8573072, 44.30752, 43.6881474, 42.9996462, 42.1815657, 41.1749175, 40.1057755, 38.978545, 37.7976309, 36.5674385, 35.292373, 33.9768395, 32.6252431, 31.241989, 29.8314825, 28.3981286, 26.9463324, 25.4804992, 24.0050342, 22.5243424, 21.042829, 19.2218792, 17.244623, 15.2983135, 13.3984875, 11.5606818, 9.80043327, 8.13327866, 6.57475482, 5.14039857, 3.84574672, 2.7063361, 1.73770352, 0.955385814, 0.374919787, 0.0118422624, -0.118309939, 0, 2.66928599, 5.28208518, 7.8629483, 10.4364261, 13.0270693, 15.6594287, 18.358055, 21.1474989, 24.0523112, 27.0970427, 30.306244, 33.7044659, 37.3162591, 41.1661744, 45.2787626, 49.6785743, 53.7854199, 57.5374587, 62.3271066, 68.0493271, 74.5990837, 81.8713399, 89.7610593, 98.1632052, 106.972741, 116.084631, 125.393838, 134.795326, 144.184058, 153.339179, 162.313938, 171.038045, 179.441207, 187.884907, 196.082967, 203.634072, 210.338399, 215.996126, 220.407431, 223.372492, 226.782189, 227.709696, 225.54614, 219.682649, 209.51035, 196.347805, 177.176847, 150.906744, 116.446764, 72.7061769, 21.4727867, -44.7408878, -125.094915, -222.171309, -345.538554, -501.628352, -696.872404, -939.344521, -1233.46109, -1581.8471, -1977.30114, -2449.83864, -3013.25743, -3686.12902, -4495.11375, -5463.15904, -6590.01918, -7892.62856, -9472.96313, -11384.8945, -13586.6867, -16242.5278, -19439.3845, -23264.2235, -27804.0114, -35857.4785, -44121.8868, -52006.262, -58919.6298, -64271.0159, -65366.807, -65120.4272, -63842.033, -62644.4614, -61520.1414, -60127.052, -58474.3449, -56571.1717, -54426.6839, -52050.0332, -50440.0884, -47735.0476, -44330.44, -40621.7948, -37004.6412, -33874.5085, -31626.9258, -28425.9808, -25197.4234},
        {-35.0428378, -36.2231421, -37.0922797, -37.6625303, -37.946174, -37.9554907, -37.7027603, -37.2002627, -36.4602779, -35.4950858, -34.3169663, -32.9381994, -31.3710649, -29.6278428, -27.7208131, -25.6622556, -23.4644503, -21.1396771, -18.700216, -16.1583468, -13.5263495, -10.816504, -8.04109033, -5.21238829, -2.34267785, 0.555761065, 3.47064852, 6.38970459, 9.30064934, 12.1912029, 15.0490852, 17.8620165, 20.6177167, 23.303906, 25.5417803, 27.510031, 29.391755, 31.1865668, 32.8940809, 34.5139117, 36.0456738, 37.4889816, 38.8434495, 40.1086922, 41.284324, 42.3699595, 43.3652132, 44.2696995, 45.0830329, 45.8048279, 46.434699, 46.8893469, 47.2530973, 47.527362, 47.713553, 47.8130822, 47.8273614, 47.7578025, 47.6058175, 47.3728182, 47.0602165, 46.6694243, 46.2018535, 45.6589161, 45.0420238, 44.3525886, 43.5920224, 42.7133922, 41.6709259, 40.5657428, 39.402389, 38.1854102, 36.9193526, 35.6087618, 34.2581838, 32.8721645, 31.4552498, 30.0119856, 28.5469178, 27.0645922, 25.5695548, 24.0663515, 22.5595281, 21.0536306, 19.2275149, 17.2554222, 15.3155085, 13.422889, 11.5926787, 9.83999297, 8.17994681, 6.62765545, 5.19823402, 3.90679771, 2.76846166, 1.79834104, 1.01155101, 0.42320673, 0.048423361, -0.097683936, 0, 2.12253614, 4.27470243, 6.47549046, 8.74389185, 11.0988982, 13.5595011, 16.1446923, 18.8734632, 21.7648055, 24.8377108, 28.1111708, 31.6041769, 35.335721, 39.3247944, 43.5903889, 48.1514961, 52.5989192, 56.8904799, 62.0136914, 67.8957905, 74.4640142, 81.6455996, 89.3677834, 97.5578027, 106.142894, 115.050295, 124.207243, 133.540973, 142.978724, 152.537417, 162.021719, 171.331962, 180.368479, 189.26532, 197.805958, 205.711538, 212.812277, 218.93839, 223.920093, 227.587602, 231.723465, 233.611376, 232.699574, 228.436296, 220.269778, 208.613893, 191.709836, 168.764436, 138.984522, 101.576923, 59.2355503, 3.4181614, -66.4496789, -153.203257, -264.293764, -405.099945, -581.000543, -798.774886, -1062.08282, -1373.05629, -1722.64285, -2142.34322, -2649.43045, -3262.62237, -4003.08487, -4890.86014, -5923.79199, -7116.16751, -8563.66769, -10315.9325, -12338.9626, -14778.905, -17714.3516, -21223.8947, -25386.1264, -32790.9213, -40377.7682, -47597.4386, -53900.7039, -58738.3355, -59577.9579, -59175.5874, -57831.9158, -56661.8745, -55643.8519, -54416.9696, -52976.5792, -51318.0323, -49436.6806, -47327.8756, -45679.2252, -43062.0347, -39857.3178, -36446.088, -33209.3588, -30528.1437, -28783.4563, -25848.908, -23140.3655},
        {-37.544799, -38.7951888, -39.7369347, -40.3818152, -40.7416085, -40.8280928, -40.6530464, -40.2282477, -39.5654749, -38.6765062, -37.57312, -36.2670946, -34.7702082, -33.094239, -31.2509655, -29.2521658, -27.1096183, -24.8351012, -22.4403928, -19.9372715, -17.3375153, -14.6529028, -11.8952121, -9.07622144, -6.20770922, -3.30145368, -0.369233102, 2.57717424, 5.52599006, 8.46543608, 11.383734, 14.2691056, 17.1097725, 19.8939565, 22.3148068, 24.5142878, 26.6379105, 28.6834722, 30.6487704, 32.5316025, 34.3297658, 36.0410577, 37.6632756, 39.194217, 40.6316791, 41.9734594, 43.2173553, 44.3611642, 45.4026834, 46.3397104, 47.1700425, 47.67991, 48.0832639, 48.3824878, 48.5799654, 48.6780806, 48.679217, 48.5857584, 48.4000885, 48.1245912, 47.7616501, 47.313649, 46.7829717, 46.1720019, 45.4831232, 44.7187196, 43.8811747, 42.9359056, 41.849217, 40.6997334, 39.4922353, 38.2315025, 36.9223155, 35.5694541, 34.1776987, 32.7518294, 31.2966263, 29.8168695, 28.3173393, 26.8028157, 25.2780789, 23.7479091, 22.2170863, 20.6903908, 18.8869056, 16.9602687, 15.0675518, 13.2228065, 11.4400845, 9.73343715, 8.11691618, 6.60457311, 5.21045952, 3.94862697, 2.83312703, 1.87801126, 1.09733124, 0.505138536, 0.115484716, -0.0575786525, 0, 0.376719205, 1.06954659, 2.0788564, 3.40502289, 5.04842031, 7.00942291, 9.28840494, 11.8857407, 14.8018043, 18.0369701, 21.5916124, 25.4661053, 29.6608232, 34.1761403, 39.0124308, 44.170069, 49.4441426, 54.8056993, 60.7194517, 67.1417843, 74.0290813, 81.337727, 89.0241056, 97.0446015, 105.355599, 113.913482, 122.674635, 131.595443, 140.632289, 149.815213, 158.999937, 168.12075, 177.111939, 186.412628, 195.704685, 204.53536, 212.687491, 219.943914, 226.087467, 230.900987, 235.296554, 237.584082, 237.325469, 234.082615, 227.417419, 217.901069, 203.833853, 184.525346, 159.285127, 127.422772, 91.4750844, 43.5800243, -16.6988907, -91.9362997, -189.072245, -313.086794, -468.96001, -663.921371, -900.190302, -1177.53232, -1483.52811, -1852.72987, -2309.21946, -2872.19676, -3552.5894, -4365.12212, -5300.77869, -6368.12879, -7662.79221, -9231.14827, -11046.9776, -13240.4272, -15883.0067, -19046.2259, -22801.5943, -29490.5746, -36347.2352, -42875.5975, -48579.6829, -52963.5128, -53750.7662, -53412.7018, -52216.3451, -51131.5953, -50147.2569, -48959.27, -47571.0825, -45986.1418, -44207.8957, -42239.7918, -39348.5611, -36683.3375, -34206.9573, -31882.2568, -29672.0723, -27539.2403, -25446.597, -23128.136, -20863.4382},
        {-39.8659112, -41.2301363, -42.2860015, -43.0448497, -43.5180238, -43.7168666, -43.6527209, -43.3369295, -42.7808355, -41.9957815, -40.9931106, -39.7841654, -38.3802888, -36.7928238, -35.0331132, -33.1124998, -31.0423264, -28.833936, -26.4986714, -24.0478754, -21.4928908, -18.8450606, -16.1157276, -13.3162347, -10.4579246, -7.55214031, -4.61022459, -1.64352033, 1.33662962, 4.31888242, 7.2918952, 10.2443251, 13.1648293, 16.042065, 18.635482, 21.0531622, 23.4059423, 25.6899176, 27.9011836, 30.0358355, 32.0899688, 34.0596788, 35.9410609, 37.7302104, 39.4232227, 41.0161931, 42.5052171, 43.8863899, 45.1558069, 46.3095636, 47.3437552, 47.9300917, 48.396086, 48.7448654, 48.9795573, 49.1032891, 49.1191883, 49.0303822, 48.8399982, 48.5511638, 48.1670063, 47.6906531, 47.1252316, 46.4738693, 45.7396936, 44.9258318, 44.0354113, 43.0400333, 41.9120561, 40.7196274, 39.4679183, 38.1620999, 36.8073434, 35.4088198, 33.9717003, 32.501156, 31.002358, 29.4804774, 27.9406854, 26.388153, 24.8280514, 23.2655517, 21.705825, 20.1540424, 18.3773872, 16.5050302, 14.6690025, 12.8821986, 11.1575126, 9.50783888, 7.94607167, 6.48510527, 5.13783395, 3.917152, 2.83595369, 1.9071333, 1.14358512, 0.558203416, 0.163882476, -0.0264834226, 0, -1.31031437, -2.0291916, -2.17417689, -1.76281545, -0.812652479, 0.658766815, 2.63389723, 5.09519356, 8.0251106, 11.4061032, 15.220626, 19.451134, 24.0800818, 29.0899244, 34.4631164, 40.1821128, 46.220279, 52.5587715, 59.2006039, 66.1262833, 73.3163168, 80.7512116, 88.4114747, 96.2776132, 104.330134, 112.549545, 120.916353, 129.411065, 138.014187, 146.715685, 155.483141, 164.294224, 173.126606, 182.768395, 192.792043, 202.55388, 211.788445, 220.230276, 227.61391, 233.673885, 238.474472, 241.320121, 241.880858, 239.826709, 234.827696, 227.746035, 216.761515, 201.246111, 180.571803, 154.110567, 124.056231, 83.5099949, 32.3191708, -31.763656, -115.262631, -222.781736, -358.924951, -530.990803, -740.276329, -985.139063, -1251.67118, -1574.90637, -1982.55115, -2493.31734, -3110.6758, -3845.09324, -4683.16266, -5629.16129, -6774.90331, -8162.68702, -9767.25574, -11708.8559, -14053.1702, -16865.8815, -20212.6724, -26191.7682, -32328.1735, -38179.4357, -43303.1017, -47256.7188, -48029.0676, -47792.3061, -46776.31, -45782.4036, -44814.2641, -43637.4651, -42267.5891, -40720.2184, -39010.9354, -37155.3227, -33082.1433, -30406.3341, -28674.3519, -27432.6535, -26227.6958, -24605.9357, -22113.8302, -20384.8411, -18471.2013},
        {-41.5002906, -43.096595, -44.3768233, -45.3521539, -46.0337653, -46.4328359, -46.5605441, -46.4280684, -46.0465871, -45.4272787, -44.5813217, -43.5198945, -42.2541754, -40.795343, -39.1545756, -37.3430517, -35.3719497, -33.2524481, -30.9957252, -28.6129595, -26.1153295, -23.5140135, -20.82019, -18.0450375, -15.1997342, -12.2954588, -9.34338957, -6.35470496, -3.34058343, -0.312203383, 2.71925674, 5.7426185, 8.74670349, 11.7203333, 14.4728027, 17.0864728, 19.6458624, 22.145619, 24.5803903, 26.9448236, 29.2335667, 31.4412671, 33.5625722, 35.5921298, 37.5245873, 39.3545923, 41.0767923, 42.685835, 44.1763678, 45.5430383, 46.7804942, 47.4756467, 48.0397185, 48.4761961, 48.788566, 48.9803147, 49.0549286, 49.0158942, 48.866698, 48.6108265, 48.2517661, 47.7930033, 47.2380246, 46.5903165, 45.8533655, 45.030658, 44.1256806, 43.1062647, 41.9410982, 40.707203, 39.4103768, 38.0564176, 36.6511232, 35.2002914, 33.70972, 32.1852069, 30.63255, 29.057547, 27.4659958, 25.8636943, 24.2564402, 22.6500315, 21.0502659, 19.4629413, 17.7068061, 15.8851171, 14.1033795, 12.3734613, 10.7072302, 9.11655445, 7.61330183, 6.20934034, 4.91653794, 3.74676257, 2.71188219, 1.82376474, 1.0942782, 0.535290495, 0.158669593, -0.0237165562, 0, -2.26524108, -3.78877533, -4.59797373, -4.72020728, -4.18284696, -3.01326376, -1.23882867, 1.11308732, 4.01511324, 7.43987808, 11.3600109, 15.7481406, 20.5768963, 25.818907, 31.4468017, 37.4332094, 43.855428, 50.7010391, 57.7059209, 64.8651313, 72.1737284, 79.6267703, 87.219315, 94.9464206, 102.803145, 110.784547, 118.885683, 127.101613, 135.427393, 143.812119, 152.313665, 160.94088, 169.702611, 179.554325, 190.03156, 200.417422, 210.436772, 219.814471, 228.275381, 235.544364, 241.302058, 245.331007, 247.364724, 247.136722, 244.380516, 239.887779, 232.069325, 220.394127, 204.331157, 183.34939, 159.533071, 126.539456, 83.9991555, 29.8493113, -41.4304352, -133.808096, -251.251681, -399.597988, -580.521519, -793.658107, -1028.02871, -1313.22348, -1672.51977, -2121.99975, -2665.55376, -3312.66837, -4052.74961, -4890.078, -5902.56279, -7126.46465, -8525.07899, -10218.8572, -12269.3171, -14737.9766, -17686.3536, -22999.1605, -28458.9219, -33671.3573, -38242.1857, -41777.1264, -42498.9557, -42318.2978, -41433.5619, -40505.81, -39550.3178, -38389.589, -37048.5376, -35552.0781, -33925.1246, -32192.5915, -27764.2335, -25252.9759, -24018.4611, -23420.3316, -22818.23, -21571.7988, -19040.6805, -17699.253, -16015.6087},
        {-41.8758817, -43.9059315, -45.5978014, -46.9629999, -48.0130351, -48.7594157, -49.2136499, -49.3872462, -49.2917129, -48.9385584, -48.3392912, -47.5054196, -46.4484521, -45.1798969, -43.7112626, -42.0540575, -40.21979, -38.2199685, -36.0661013, -33.7696969, -31.3422638, -28.7953101, -26.1403445, -23.3888752, -20.5524106, -17.6424592, -14.6705294, -11.6481294, -8.5867678, -5.4979529, -2.39319312, 0.716003155, 3.81812753, 6.90167161, 9.7981475, 12.5778384, 15.3129143, 17.9969611, 20.6235647, 23.1863109, 25.6787858, 28.0945751, 30.4272648, 32.6704407, 34.8176888, 36.8625949, 38.798745, 40.6197248, 42.3191204, 43.8905176, 45.3275024, 46.1753283, 46.8856324, 47.4617197, 47.9068949, 48.224463, 48.4177287, 48.489997, 48.4445727, 48.2847607, 48.0138658, 47.6351929, 47.1520468, 46.5677323, 45.8855545, 45.108818, 44.2408278, 43.2316763, 42.0327334, 40.7570508, 39.4113828, 38.0024838, 36.5371083, 35.0220106, 33.4639452, 31.8696665, 30.2459289, 28.5994867, 26.9370944, 25.2655064, 23.591477, 21.9217608, 20.2631121, 18.6222853, 16.8682011, 15.07943, 13.336199, 11.6497354, 10.0312667, 8.49202007, 7.04322303, 5.69610292, 4.46188711, 3.35180297, 2.37707788, 1.54893919, 0.878614297, 0.377330557, 0.0563153461, -0.0732039634, 0, -1.70309275, -2.77170539, -3.22636172, -3.08758556, -2.37590069, -1.11183092, 0.684099943, 2.9913681, 5.78944976, 9.05782111, 12.7759583, 16.9233377, 21.4794353, 26.4237274, 31.7356902, 37.3947998, 43.4518197, 49.8964097, 56.5468016, 63.3977473, 70.4439989, 77.6803084, 85.1014277, 92.7021088, 100.477104, 108.421165, 116.529043, 124.795492, 133.215262, 141.756101, 150.449668, 159.298816, 168.306399, 178.148987, 188.492574, 198.823499, 208.942499, 218.650312, 227.747678, 236.035335, 243.807499, 250.221242, 254.98075, 257.790214, 258.35382, 256.416627, 251.631736, 243.693116, 232.29474, 217.130576, 200.926659, 176.639041, 142.402804, 95.941931, 34.1410754, -45.7382668, -146.4346, -269.417078, -418.982045, -600.810591, -814.37646, -1073.76613, -1383.76561, -1754.23468, -2203.63037, -2746.46341, -3388.03773, -4140.47712, -5049.6637, -6142.38398, -7350.48711, -8810.30267, -10581.4801, -12723.6688, -15296.5181, -20030.0731, -24895.9888, -29536.3156, -33593.1042, -36708.405, -37262.206, -36999.9954, -36104.7079, -35183.4568, -34251.2724, -33146.4442, -31892.8396, -30514.3261, -29034.7712, -27478.0424, -24426.5813, -22419.629, -21127.8184, -20221.7822, -19372.153, -18249.5636, -16524.6467, -15160.9919, -13546.3997},
        {-41.7005097, -44.2792508, -46.4887723, -48.3412821, -49.8489878, -51.0240972, -51.8788179, -52.4253578, -52.6759245, -52.6427258, -52.3379693, -51.7738629, -50.9626141, -49.9164309, -48.6475208, -47.1680916, -45.490351, -43.6265068, -41.5887666, -39.3893382, -37.0404294, -34.5542477, -31.9430011, -29.2188971, -26.3941435, -23.4809481, -20.4915185, -17.4380625, -14.3327878, -11.1879021, -8.01561312, -4.82812864, -1.63765635, 1.54359602, 4.53698595, 7.41702386, 10.2602206, 13.0597695, 15.808864, 18.5006975, 21.1284635, 23.6853552, 26.1645663, 28.5592899, 30.8627196, 33.0680487, 35.1684707, 37.157179, 39.0273669, 40.7722279, 42.3849554, 43.4335254, 44.3455667, 45.1234903, 45.7697077, 46.2866302, 46.6766689, 46.9422352, 47.0857404, 47.1095958, 47.0162126, 46.8080022, 46.4873757, 46.0567446, 45.51852, 44.8751133, 44.1289358, 43.1985216, 42.0044189, 40.7220291, 39.3592009, 37.923783, 36.423624, 34.8665726, 33.2604776, 31.6131876, 29.9325514, 28.2264176, 26.5026349, 24.7690519, 23.0335175, 21.3038803, 19.587989, 17.8936922, 16.14266, 14.387643, 12.6824278, 11.0376597, 9.46398411, 7.97204633, 6.57249174, 5.27596569, 4.09311355, 3.03458068, 2.11101243, 1.33305415, 0.711351217, 0.256548976, -0.0207072122, -0.10977199, 0, -0.991403768, -1.48845113, -1.50236486, -1.04436775, -0.125682577, 1.24246789, 3.04886086, 5.28227355, 7.9314832, 10.985267, 14.4324022, 18.261666, 22.4618356, 27.0216883, 31.9300013, 37.1755517, 42.7453224, 48.6278344, 54.8159235, 61.2979824, 68.062404, 75.0975808, 82.3919056, 89.9337711, 97.71157, 105.713695, 113.928539, 122.344494, 130.949954, 139.697038, 148.623711, 157.729249, 167.012925, 176.917872, 187.221435, 197.582597, 207.867477, 217.94219, 227.672855, 236.925587, 246.363522, 254.813189, 261.984768, 267.588437, 271.334376, 272.653937, 271.605833, 267.969949, 261.526171, 252.054386, 242.268121, 225.428057, 198.673662, 159.752988, 107.656614, 40.8172475, -42.3324015, -138.949814, -256.014321, -405.315953, -596.47249, -828.919525, -1090.88204, -1386.34908, -1746.02105, -2188.3374, -2731.26879, -3393.13314, -4193.96723, -5152.83557, -6181.27859, -7422.427, -8932.56519, -10767.9775, -12984.9483, -17158.7223, -21446.8834, -25525.976, -29072.5445, -31763.133, -32118.5939, -31741.6249, -30804.0669, -29885.484, -28996.3268, -27977.1609, -26846.9311, -25624.5819, -24329.0582, -22979.3045, -21253.2517, -19734.5685, -18388.9963, -17182.2765, -16080.1506, -15048.36, -14052.6461, -12678.5815, -11157.6376},
        {-41.4908839, -44.7775377, -47.6472218, -50.113608, -52.1903682, -53.8911743, -55.2296982, -56.2196117, -56.8745866, -57.2082949, -57.2344084, -56.9665991, -56.4185386, -55.603899, -54.5363521, -53.2295698, -51.6972239, -49.9529862, -48.0105288, -45.8835233, -43.5856418, -41.1305561, -38.5319379, -35.8034593, -32.9587921, -30.0116081, -26.9755792, -23.8643772, -20.6916742, -17.4711418, -14.216452, -10.9412767, -7.65928763, -4.38415679, -1.37189644, 1.49008908, 4.31730816, 7.10395336, 9.84421722, 12.5322923, 15.1623711, 17.7286462, 20.2253102, 22.6465555, 24.9865748, 27.2395605, 29.3997053, 31.4612016, 33.4182421, 35.2650192, 36.9957255, 38.3268115, 39.5362327, 40.6242026, 41.5909345, 42.436642, 43.1615385, 43.7658375, 44.2497523, 44.6134965, 44.8572835, 44.9813267, 44.9858397, 44.8710357, 44.6371284, 44.2843311, 43.8128573, 43.0852898, 41.9675156, 40.7436012, 39.4226823, 38.0138943, 36.5263726, 34.9692528, 33.3516703, 31.6827607, 29.9716594, 28.2275019, 26.4594237, 24.6765603, 22.8880472, 21.1030199, 19.3306138, 17.5799645, 15.8363284, 14.1212829, 12.4566921, 10.8524665, 9.31851643, 7.86475237, 6.50108471, 5.23742386, 4.08368022, 3.04976421, 2.14558623, 1.38105669, 0.766086009, 0.310584589, 0.0244628415, -0.0823688246, 0, -0.889524519, -1.34073376, -1.36004087, -0.953859003, -0.128601305, 1.10931907, 2.75348898, 4.79749527, 7.2349248, 10.0593644, 13.2644009, 16.8436213, 20.7906122, 25.0989607, 29.7622535, 34.7740774, 40.0423341, 45.5483696, 51.4795828, 57.8111994, 64.5184451, 71.5765456, 78.9607266, 86.646214, 94.6082332, 102.82201, 111.26277, 119.90574, 128.726144, 137.520783, 146.508731, 155.718742, 165.179568, 175.549247, 186.541892, 197.703803, 208.874948, 219.895296, 230.604814, 240.843471, 250.221796, 258.879025, 266.700014, 273.56962, 279.372703, 286.035587, 290.891295, 293.314318, 292.679147, 288.360272, 280.680714, 266.90712, 245.242564, 214.369288, 173.947835, 123.199513, 61.3456283, -5.85210286, -87.1992413, -198.636342, -355.198989, -549.539661, -766.95522, -1008.53618, -1302.13452, -1663.31827, -2105.62124, -2644.08407, -3301.20619, -4095.26793, -4960.1165, -6011.45503, -7296.79067, -8863.63058, -10759.4819, -14371.2859, -18072.2572, -21575.0443, -24592.2959, -26836.6606, -26967.6822, -26453.1841, -25457.1457, -24555.1969, -23748.3721, -22866.1843, -21918.2437, -20914.1606, -19863.5453, -18776.0081, -17060.9436, -15753.4287, -14737.9303, -13898.9156, -13120.8517, -12288.2056, -11285.4443, -10209.3639, -8998.47221},
        {-40.3788505, -44.3174965, -47.8022578, -50.8478401, -53.4689489, -55.68029, -57.4965688, -58.932491, -60.0027622, -60.722088, -61.1051741, -61.1667259, -60.9214492, -60.3840495, -59.5692325, -58.4917037, -57.1661688, -55.6073333, -53.829903, -51.8485833, -49.6780798, -47.3330983, -44.8283443, -42.1785234, -39.3983413, -36.5025034, -33.5057155, -30.4226831, -27.2681119, -24.0567074, -20.8031753, -17.5222212, -14.2285507, -10.9368693, -7.9671766, -5.1751098, -2.41263672, 0.3154444, 3.00433534, 5.64923785, 8.24535371, 10.7878847, 13.2720325, 15.692999, 18.0459859, 20.326195, 22.528828, 24.6490867, 26.6821729, 28.6232883, 30.4676347, 32.0853229, 33.5993573, 35.0076515, 36.3081189, 37.4986731, 38.5772275, 39.5416957, 40.3899911, 41.1200272, 41.7297175, 42.2169756, 42.5797149, 42.8158489, 42.923291, 42.8999549, 42.743754, 42.2642795, 41.2756442, 40.1641621, 38.9399549, 37.6131442, 36.1938518, 34.6921992, 33.1183082, 31.4823004, 29.7942974, 28.0644209, 26.3027926, 24.5195342, 22.7247672, 20.9286134, 19.1411944, 17.3726319, 15.6577552, 13.9938125, 12.3777548, 10.8189021, 9.32657421, 7.91009098, 6.57877226, 5.34193789, 4.20890774, 3.18900164, 2.29153946, 1.52584103, 0.901226206, 0.427014843, 0.112526788, -0.0329181096, 0, -1.03226194, -1.63940511, -1.82605478, -1.59683621, -0.956374663, 0.0907045897, 1.53977629, 3.38621516, 5.62539596, 8.2526934, 11.2634822, 14.6531372, 18.417033, 22.5505444, 27.0490461, 31.9079129, 37.0067334, 42.3243415, 48.118009, 54.3582993, 61.0157759, 68.0610022, 75.4645416, 83.1969576, 91.2288136, 99.530673, 108.073099, 116.826656, 125.761906, 134.587418, 143.631815, 152.94426, 162.573916, 173.325306, 184.869913, 196.677787, 208.571482, 220.373553, 231.906553, 242.993035, 252.353206, 261.247465, 269.714043, 277.79117, 285.517077, 296.685476, 306.640247, 314.480748, 319.30634, 320.216383, 315.665534, 306.185826, 291.213525, 270.267472, 243.035096, 209.128134, 168.158321, 126.663412, 75.7550312, -1.0108532, -119.296675, -270.506679, -442.759518, -636.178326, -871.227156, -1159.03325, -1505.39358, -1920.05349, -2432.30254, -3060.37498, -3766.89323, -4634.25903, -5699.95686, -7001.47122, -8576.2866, -11599.1865, -14686.0317, -17589.9818, -20064.1967, -21861.836, -21806.0804, -21200.0548, -20195.4813, -19346.4663, -18645.4549, -17917.2322, -17162.6262, -16382.465, -15577.5768, -14748.7896, -12876.4331, -11733.5372, -11079.4761, -10673.6235, -10275.3534, -9644.03979, -8539.0565, -7815.16273, -6910.90896},
        {-36.7036308, -40.0990239, -43.1474893, -45.8590284, -48.2436426, -50.3113336, -52.0721027, -53.5359513, -54.7128811, -55.6128934, -56.2459898, -56.6221716, -56.7514404, -56.6437976, -56.3092448, -55.7577833, -54.9994146, -54.0441403, -52.9019618, -51.5828805, -50.096898, -48.4540157, -46.664235, -44.7375575, -42.6839847, -40.5135179, -38.2361586, -35.8619084, -33.4007687, -30.862741, -28.2578267, -25.5960274, -22.8873444, -20.1417793, -17.4726137, -14.8360368, -12.186996, -9.53214127, -6.87812239, -4.23158919, -1.59919151, 1.01242082, 3.59659799, 6.14669016, 8.6560475, 11.1180202, 13.5259584, 15.8732123, 18.153132, 20.3590678, 22.4843698, 24.3921585, 26.2088368, 27.9305782, 29.5535557, 31.0739429, 32.4879129, 33.791639, 34.9812946, 36.053053, 37.0030873, 37.8275711, 38.5226774, 39.0845797, 39.5094512, 39.7934653, 39.9327951, 39.6945733, 38.8514313, 37.8719222, 36.7670669, 35.5478865, 34.225402, 32.8106347, 31.3146055, 29.7483356, 28.122846, 26.4491577, 24.738292, 23.0012698, 21.2491122, 19.4928404, 17.7434754, 16.0120382, 14.3971827, 12.8642703, 11.3776092, 9.9453768, 8.57575021, 7.2769067, 6.0570235, 4.92427785, 3.88684697, 2.95290811, 2.13063849, 1.42821536, 0.853815928, 0.415617446, 0.121797141, -0.0194677516, 0, -0.194846855, -0.175436108, 0.068813246, 0.548482216, 1.27415181, 2.25640303, 3.5058169, 5.0329744, 6.84845656, 8.96284439, 11.3867189, 14.130661, 17.2052519, 20.6210724, 24.3887037, 28.5187266, 32.7517085, 37.059657, 42.0638984, 47.7171538, 53.9721439, 60.7815898, 68.0982123, 75.8747324, 84.063871, 92.618349, 101.490887, 110.634207, 120.001028, 129.399625, 138.98013, 148.738598, 158.671086, 168.854455, 179.244356, 189.774895, 200.417884, 211.145138, 221.928469, 232.73969, 243.862729, 254.862295, 265.649137, 276.134001, 286.227638, 295.250747, 303.851635, 312.088563, 320.019789, 327.703575, 337.292182, 344.190531, 346.268495, 342.10029, 331.698161, 314.428705, 289.658522, 259.983828, 220.807599, 164.009598, 83.0965692, -21.08103, -143.987635, -287.656855, -465.253399, -684.832616, -949.460561, -1265.89906, -1659.20404, -2144.08325, -2699.87183, -3382.09207, -4217.69608, -5233.63597, -6456.86384, -8810.27568, -11200.8938, -13431.6842, -15305.6132, -16625.6468, -16429.362, -15795.3737, -14854.6719, -14107.2362, -13537.9985, -12978.1449, -12420.2944, -11857.0661, -11281.0791, -10684.9524, -9203.86967, -8286.15713, -7754.87973, -7433.10244, -7143.89024, -6710.3081, -5955.42101, -5569.95265, -4667.95586}, 
        {-40.407689, -45.9605104, -50.9641119, -55.4359258, -59.3933845, -62.8539205, -65.8349661, -68.3539537, -70.4283157, -72.0754845, -73.3128924, -74.157972, -74.6281555, -74.7408754, -74.513564, -73.9636538, -73.1085771, -71.9657663, -70.5526538, -68.8866721, -66.9852535, -64.8658303, -62.5458351, -60.0427001, -57.3738578, -54.5567406, -51.6087808, -48.5474109, -45.3900632, -42.1541701, -38.857164, -35.5164774, -32.1495426, -28.7737919, -25.8449278, -23.1520292, -20.4789074, -17.8287722, -15.2048333, -12.6103003, -10.0483829, -7.52229088, -5.03523385, -2.59042153, -0.191063608, 2.15963023, 4.45845029, 6.70218688, 8.88763032, 11.0115709, 13.070799, 15.0994897, 17.0562382, 18.9370241, 20.7378275, 22.4546281, 24.0834059, 25.6201407, 27.0608124, 28.4014009, 29.637886, 30.7662477, 31.7824658, 32.6825202, 33.4623907, 34.1180573, 34.6454999, 34.8660398, 34.6051923, 34.2191569, 33.7152358, 33.1007312, 32.3829456, 31.569181, 30.6667399, 29.6829244, 28.6250368, 27.5003794, 26.3162545, 25.0799643, 23.798811, 22.480097, 21.1311246, 19.7591958, 18.1883795, 16.5214488, 14.863379, 13.2274189, 11.6268169, 10.0748216, 8.58468156, 7.16964529, 5.84296136, 4.61787832, 3.50764471, 2.52550909, 1.68472001, 0.998526007, 0.480175639, 0.142917454, 0, -5.04533406, -8.80453044, -11.3488094, -12.7493913, -13.0774962, -12.4043446, -10.8011566, -8.33915261, -5.0895528, -1.12357748, 3.4875531, 8.67261865, 14.3603989, 20.4796736, 26.9592225, 33.7278253, 42.0318743, 51.9883798, 60.5458071, 67.9152814, 74.3079281, 79.9348725, 85.0072398, 89.7361554, 94.3327447, 99.0081328, 103.973445, 109.439807, 115.618343, 123.664723, 132.499195, 142.049523, 152.243469, 163.219209, 174.799299, 186.750184, 198.936502, 211.222894, 223.473996, 235.554448, 245.902147, 256.242698, 266.719885, 277.477492, 288.659301, 302.37192, 316.305603, 330.113428, 343.448475, 355.96382, 367.363219, 377.187134, 386.398457, 394.765768, 399.619271, 399.383953, 392.4848, 376.052677, 350.100809, 316.054184, 277.150749, 225.938984, 159.72594, 74.3378929, -36.9079686, -179.542744, -358.643603, -579.623959, -849.561633, -1174.59296, -1512.49107, -1925.19601, -2435.58522, -3066.53617, -3840.92631, -5391.85238, -6979.41771, -8473.94495, -9745.75671, -10665.1756, -10620.2139, -10285.0449, -9736.69571, -9238.37693, -8793.41802, -8327.57191, -7848.04684, -7362.05105, -6876.79278, -6399.48027, -5223.19858, -4653.64555, -4470.72461, -4454.33919, -4384.39272, -4040.78864, -3203.43039, -3073.28089, -2486.75042}, 
        {-33.1017163, -39.8779159, -46.0391019, -51.6044334, -56.5930698, -61.0241704, -64.9168945, -68.2904012, -71.1638499, -73.5564, -75.4872105, -76.975441, -78.0402505, -78.7007984, -78.976244, -78.8857465, -78.4484652, -77.6835595, -76.6101885, -75.2475116, -73.6146881, -71.7308771, -69.6152381, -67.2869302, -64.7651128, -62.0689451, -59.2175864, -56.230196, -53.1259331, -49.9239571, -46.6434273, -43.3035028, -39.923343, -36.5221072, -33.5998286, -30.925115, -28.2607944, -25.6104208, -22.9775481, -20.3657304, -17.7785216, -15.2194758, -12.6921469, -10.2000889, -7.74685587, -5.33600173, -2.9710805, -0.655646153, 1.6067473, 3.81254588, 5.95819559, 8.04310715, 10.0606976, 12.0073486, 13.8794421, 15.6733596, 17.3854829, 19.0121938, 20.549874, 21.9949053, 23.3436693, 24.5925479, 25.7379228, 26.7761756, 27.7036882, 28.5168423, 29.2120196, 29.6417177, 29.6618888, 29.5656649, 29.3587552, 29.0468688, 28.6357149, 28.1310026, 27.5384412, 26.8637397, 26.1126074, 25.2907534, 24.4038869, 23.457717, 22.4579529, 21.4103038, 20.3204788, 19.1941871, 17.8175862, 16.310779, 14.7965074, 13.2876055, 11.7969071, 10.3372465, 8.92145751, 7.5623743, 6.27283088, 5.06566127, 3.95369952, 2.94977966, 2.06673573, 1.31740176, 0.714611794, 0.271199861, 0, -3.75444364, -6.54693917, -8.42515211, -9.43674799, -9.62939234, -9.05075068, -7.74848854, -5.77027146, -3.16376495, 0.0233654538, 3.74345422, 7.94883583, 12.5918448, 17.6248155, 23.0000824, 28.6699801, 35.414298, 43.3035783, 50.4185313, 56.8888031, 62.8440402, 68.4138886, 73.7279946, 78.9160046, 84.1075647, 89.4323211, 95.0199203, 101.000008, 107.502231, 115.077001, 123.278918, 132.111398, 141.577859, 152.006595, 163.238584, 175.028169, 187.281303, 199.903941, 212.802034, 225.881537, 238.082741, 250.571157, 263.441675, 276.78918, 290.708561, 307.276509, 324.110657, 340.810438, 356.97529, 372.204648, 385.236167, 397.584352, 410.959863, 425.707402, 439.382835, 450.794161, 458.749377, 460.631156, 456.996755, 449.958327, 442.820079, 429.642091, 407.921042, 374.6133, 325.75971, 257.821357, 167.249901, 50.5039835, -95.9231964, -275.557988, -456.431384, -683.025914, -971.15478, -1336.63119, -1795.26833, -2747.45513, -3728.28515, -4657.42767, -5454.55196, -6039.32731, -6040.7219, -5862.90682, -5550.13753, -5243.75717, -4949.59064, -4633.00967, -4301.86183, -3963.99468, -3627.25581, -3299.4928, -2822.42129, -2540.89081, -2374.94485, -2244.62694, -2069.98057, -1771.04926, -1267.87651, -1071.45202, -816.261377}, 
        {-31.0761186, -37.9526665, -44.3080895, -50.1566205, -55.5124922, -60.3899375, -64.8031891, -68.7664798, -72.2940424, -75.4001097, -78.0989144, -80.4046894, -82.3316674, -83.8940811, -85.1061635, -85.9821472, -86.5362651, -86.7827499, -86.7358345, -86.4097515, -85.8187338, -84.9770142, -83.8988254, -82.5984003, -81.0899715, -79.387772, -77.5060344, -75.4589916, -73.2608764, -70.9259215, -68.4683597, -65.9024238, -63.2423465, -60.5023608, -57.9287274, -55.4147849, -52.8510825, -50.2443233, -47.6012104, -44.9284468, -42.2327354, -39.5207794, -36.7992817, -34.0749454, -31.3544734, -28.6445689, -25.9519348, -23.2832741, -20.6452898, -18.0446851, -15.4881629, -13.1778984, -10.920885, -8.71958834, -6.57647386, -4.49400718, -2.47465385, -0.520879454, 1.36485046, 3.18007031, 4.92231455, 6.58911759, 8.17801388, 9.68653786, 11.1122239, 12.4526066, 13.7052202, 14.8225085, 15.7579979, 16.602219, 17.3556292, 18.0186861, 18.5918471, 19.0755697, 19.4703114, 19.7765296, 19.9946819, 20.1252257, 20.1686186, 20.1253179, 19.9957811, 19.7804659, 19.4798295, 19.0943296, 18.157605, 16.9133414, 15.6108353, 14.2656933, 12.893522, 11.5099281, 10.1305182, 8.77089891, 7.44667692, 6.17345886, 4.96685136, 3.84246106, 2.8158946, 1.90275863, 1.11865978, 0.479204691, 0, -0.184455236, -0.0921398179, 0.271999199, 0.903014759, 1.79595981, 2.94588729, 4.34785014, 5.99690132, 7.88809376, 10.0164804, 12.3771142, 14.9650481, 17.7753351, 20.803028, 24.0431799, 27.4908436, 31.3924113, 35.7788414, 40.0766818, 44.3348438, 48.6022387, 52.9277779, 57.3603727, 61.9489344, 66.7423743, 71.7896038, 77.1395341, 82.8410766, 88.9431427, 95.665327, 102.823274, 110.414689, 118.437278, 126.985746, 136.0093, 145.431282, 155.220298, 165.344957, 175.773864, 186.475628, 196.667608, 207.298298, 218.483291, 230.338175, 242.978541, 257.471725, 272.743635, 288.671927, 305.134253, 322.008269, 338.811733, 356.222067, 374.73871, 394.521675, 415.037984, 436.065796, 457.383274, 478.55606, 499.623132, 520.855303, 542.446458, 564.012548, 585.850716, 607.715591, 628.442546, 647.288905, 663.74577, 677.131078, 685.905582, 689.014907, 692.806716, 687.685821, 670.035601, 636.239436, 582.680708, 426.568934, 263.254822, 108.915219, -20.2730286, -108.133076, -80.1280218, -17.3477775, 71.3730515, 157.642359, 240.36705, 326.602988, 414.849631, 503.606434, 591.372856, 676.648353, 853.556775, 945.89716, 983.432795, 995.926967, 1013.14296, 1064.84407, 1180.79358, 1248.48544, 1248.76263}
    };
    
    /* YPtag: chiAmin tables above */
    
    /* Stuff for interpolating the NQC data */
    gsl_spline    *spline = NULL;
    gsl_interp_accel *acc = NULL;
    /* Interpolate the spin NQC data in 2-D parameter space -- spin and mass ratio */
    /* First, interpolate in spin dimension for all mass ratios */
    spline = gsl_spline_alloc( gsl_interp_cspline, adim );
    acc    = gsl_interp_accel_alloc();
    for (i = 0; i < qdim; i++)
    {
        gsl_spline_init( spline, alist, a3stab[i], adim );
        gsl_interp_accel_reset( acc );
        a3slist[i] = gsl_spline_eval( spline, chiA, acc );
        gsl_spline_init( spline, alist, a4tab[i], adim );
        gsl_interp_accel_reset( acc );
        a4list[i] = gsl_spline_eval( spline, chiA, acc );
        gsl_spline_init( spline, alist, a5tab[i], adim );
        gsl_interp_accel_reset( acc );
        a5list[i] = gsl_spline_eval( spline, chiA, acc );
        gsl_spline_init( spline, alist, b3tab[i], adim );
        gsl_interp_accel_reset( acc );
        b3list[i] = gsl_spline_eval( spline, chiA, acc );
        gsl_spline_init( spline, alist, b4tab[i], adim );
        gsl_interp_accel_reset( acc );
        b4list[i] = gsl_spline_eval( spline, chiA, acc );
        //  printf("%.15f\n",a3slist[i]);
    }
    //printf("%.15f\n",a);
    gsl_spline_free(spline);
    gsl_interp_accel_free(acc);
    /* Second, interpolate in mass ratio dimension */
    spline = gsl_spline_alloc( gsl_interp_cspline, qdim );
    acc    = gsl_interp_accel_alloc();
    gsl_spline_init( spline, etalist, a3slist, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->a3S = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, a4list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->a4 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, a5list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->a5 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, b3list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->b3 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, etalist, b4list, qdim );
    gsl_interp_accel_reset( acc );
    coeffs->b4 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_free(spline);
    gsl_interp_accel_free(acc);
    
    return XLAL_SUCCESS;
    
}

/**
 * Function to 3D-interpolate known amplitude NQC coeffcients of spin terms for SEOBNRv2,
 * namely a3s, a4 and a5.
 * The a3s, a4 and a5 values were calculated for
 * 11 mass ratios q=1,1.5,2,4,6,8,10,11,12,14 and 20,
 * 19 spin (\f$\chi\f$ defined in Taracchini et al. PRD 86, 024011 (2012)) values
 * chi = -1, -0.9, -0.8, ......, 0.3, 0.4, 0.5, 0.55, 0.6, 0.65, and
 * 3 chiA values, the max and min allowed chiA for given q and chi, as well as the mid value.
 * The calculation was done by Andrea Taracchini using a C++ code of the UMaryland group.
 * In principle, these numbers can be automatically calculated iteratively by the LAL code.
 * However, since such calcualtion increase the cost of each waveform generation by
 * about an order of magnitude, we prepare these numbers in advance reduce cost.
 * These number can be verified by confirming that
 * the peak amplitude and frequency agree well with the NR-fits predicted values,
 * and to get exact NR-fits predicted values, corrections on these numbers are ~1%.
 */
UNUSED static int XLALSimIMRGetEOBCalibratedSpinNQC3D(
                                                      EOBNonQCCoeffs * restrict coeffs,
                                                      INT4 UNUSED l,
                                                      INT4 UNUSED m,
                                                      REAL8 m1,
                                                      REAL8 m2,
                                                      REAL8 a,
                                                      REAL8 chiAin )
{
    
    REAL8 eta = m1*m2/((m1+m2)*(m1+m2));
    
#if OLDNSNQC
    const unsigned int nsqdim = 101;
    
    memset( coeffs, 0, sizeof( *coeffs ) );
    const double nsetalist[101] = {0.021, 0.0233, 0.0256, 0.0279, 0.0302, 0.0324, 0.0347, 0.037, 0.0393, 0.0416, 0.0439, 0.0462, 0.0485, 0.0508, 0.0531, 0.0553, 0.0576, 0.0599, 0.0622, 0.0645, 0.0668, 0.0691, 0.0714, 0.0737, 0.076, 0.0782, 0.0805, 0.0828, 0.0851, 0.0874, 0.0897, 0.092, 0.0943, 0.0966, 0.0989, 0.1012, 0.1034, 0.1057, 0.108, 0.1103, 0.1126, 0.1149, 0.1172, 0.1195, 0.1218, 0.124, 0.1263, 0.1286, 0.1309, 0.1332, 0.1355, 0.1378, 0.1401, 0.1424, 0.1447, 0.147, 0.1492, 0.1515, 0.1538, 0.1561, 0.1584, 0.1607, 0.163, 0.1653, 0.1676, 0.1698, 0.1721, 0.1744, 0.1767, 0.179, 0.1813, 0.1836, 0.1859, 0.1882, 0.1905, 0.1927, 0.195, 0.1973, 0.1996, 0.2019, 0.2042, 0.2065, 0.2088, 0.2111, 0.2134, 0.2156, 0.2179, 0.2202, 0.2225, 0.2248, 0.2271, 0.2294, 0.2317, 0.234, 0.2363, 0.2385, 0.2408, 0.2431, 0.2454, 0.2477, 0.25};
    
    const double a1list[101] = {-9.70365622284, -9.5487796595, -9.40705218072, -9.28242447497, -9.16287060127, -9.03811669547, -8.92394749379, -8.81661027414, -8.7095851328, -8.60723315388, -8.50836995352, -8.40985391052, -8.314500912, -8.22035943611, -8.12770592382, -8.03757606119, -7.94997595927, -7.86421160183, -7.77953168332, -7.69584399308, -7.61324344682, -7.53192249536, -7.45216063215, -7.37374612092, -7.29648777301, -7.22046897906, -7.14581213839, -7.07265232481, -7.00073270258, -6.92991647363, -6.85992514074, -6.79104613524, -6.72328769905, -6.65664446919, -6.59158891131, -6.52783781218, -6.4667874811, -6.41088676175, -6.35876275395, -6.3073217784, -6.25613493225, -6.20587481079, -6.15851113649, -6.1141742151, -6.07327317284, -6.03621329966, -6.0001893395, -5.96655531812, -5.93844433597, -5.91027864135, -5.88037825648, -5.85358373242, -5.83336708498, -5.82744285781, -5.82818819089, -5.8231589417, -5.81812839946, -5.81887431779, -5.82568678472, -5.83621354379, -5.84931204972, -5.86442954488, -5.88109070102, -5.89755497618, -5.91392008194, -5.93177097875, -5.95199881766, -5.97697408745, -6.00401909417, -6.032531535, -6.06148086596, -6.09023163793, -6.11920032326, -6.14820900838, -6.17641365359, -6.20357599269, -6.23098735047, -6.25903214508, -6.28832401322, -6.32033354968, -6.35344345473, -6.38745961465, -6.42231976572, -6.46800466293, -6.50219865357, -6.52990506622, -6.54965342702, -6.58271099648, -6.626558215, -6.67664802777, -6.72810147477, -6.77989789496, -6.83928403282, -6.91446932169, -6.99994425628, -7.08201636107, -7.21472596367, -7.36827339046, -7.55475356656, -7.81665470799, -8.24936799702};
    
    const double a2list[101] = {71.8355837876, 70.2888163877, 68.858392382, 67.578368761, 66.3483263783, 65.0816770892, 63.9139561883, 62.8042133759, 61.7074812267, 60.6519669001, 59.6324792699, 58.6222094978, 57.6434687679, 56.6791933949, 55.7311388293, 54.8071777364, 53.9062860992, 53.0218701557, 52.1495389357, 51.2895492368, 50.4445863927, 49.6138840037, 48.797074544, 47.9943062521, 47.2056828988, 46.4310888273, 45.6708900349, 44.9258152616, 44.1929201091, 43.4702452949, 42.7537119379, 42.0477065689, 41.3531483593, 40.6709264131, 40.0029858455, 39.3453826346, 38.7105183605, 38.119829408, 37.5632095631, 37.0166865581, 36.4716876499, 35.9345595307, 35.4276417334, 34.9454114602, 34.4897359229, 34.0666726398, 33.6585411373, 33.268152249, 32.9062444867, 32.5494651377, 32.1932422911, 31.8674681969, 31.598287958, 31.4450951286, 31.3504698588, 31.2224679676, 31.1007751184, 31.0231242897, 30.991643584, 30.991534397, 31.0161048282, 31.0616750366, 31.1249606145, 31.1939358163, 31.2696363303, 31.3636652679, 31.4828878696, 31.6456708233, 31.8311325739, 32.034527166, 32.2477681231, 32.4658509298, 32.6919766483, 32.9247097971, 33.1580140019, 33.3900369294, 33.6294910387, 33.8788248397, 34.1421252583, 34.4296046145, 34.7294034713, 35.0398350888, 35.3602204133, 35.7665471711, 36.0881399675, 36.3632831298, 36.5789495976, 36.9023316975, 37.3120873104, 37.7708043813, 38.2382659746, 38.7083475975, 39.2358324951, 39.8798584792, 40.5947669891, 41.2766466634, 42.3250573144, 43.5282809602, 44.9693438982, 46.9606371658, 50.0808316558};
    
    const double a3list[101] = {-75.3870348213, -73.5563054743, -71.8586610238, -70.3326010764, -68.8669937325, -67.3652081804, -65.9797869993, -64.6591310991, -63.3595036445, -62.1068621881, -60.8985027736, -59.7042086628, -58.5479728539, -57.4103808438, -56.2930281651, -55.2041741738, -54.1420295685, -53.0987898958, -52.0706872299, -51.0586173069, -50.0667002046, -49.0926497207, -48.1344431455, -47.1933388413, -46.2704158358, -45.3650449971, -44.4772042283, -43.6074103996, -42.7520134632, -41.9084025308, -41.0711187682, -40.2460776811, -39.4347613879, -38.6386667672, -37.8587722914, -37.0898368809, -36.3458764907, -35.6510020946, -34.9947343733, -34.3515708492, -33.7091654129, -33.0750000728, -32.4772646605, -31.9060849394, -31.3624600752, -30.8542478078, -30.364610911, -29.8930785724, -29.4449207659, -29.0043606667, -28.5703691139, -28.171109369, -27.8341554058, -27.6229720264, -27.4766951091, -27.2985750163, -27.1291617273, -27.0061494789, -26.9317625784, -26.8914625991, -26.8788122164, -26.8902006455, -26.9223881792, -26.9631792328, -27.013848345, -27.0865681299, -27.1887211003, -27.3400465839, -27.5181167553, -27.7178394005, -27.9305187037, -28.1507835284, -28.3821145422, -28.6229907203, -28.8672132002, -29.1128371338, -29.36865915, -29.6371071707, -29.922311561, -30.2347674515, -30.5620356831, -30.9022473284, -31.254609498, -31.7001465622, -32.0565827322, -32.3648712883, -32.6103616483, -32.9736989432, -33.4312219324, -33.9418551023, -34.4614297518, -34.9838283376, -35.5674613231, -36.2737268484, -37.0519888454, -37.7923554764, -38.9140097, -40.1981922876, -41.7294871636, -43.8332602049, -47.0600574352};
    
    const double b1list[101] = {-0.454231789035, -0.453740895324, -0.45114951071, -0.445455559729, -0.44030174389, -0.438246415293, -0.435617386371, -0.430628416274, -0.428155263233, -0.424563996869, -0.421496742404, -0.419542314684, -0.417786543725, -0.416452769889, -0.415181542102, -0.413649756645, -0.411562161412, -0.408719678496, -0.40593475615, -0.403713197659, -0.402947057134, -0.402692696086, -0.401882254635, -0.401361543256, -0.401857592043, -0.403135870926, -0.405023096371, -0.407445456424, -0.410163626113, -0.412943367674, -0.415232607237, -0.417748668538, -0.420844132902, -0.424905043844, -0.429142196414, -0.432959586521, -0.436617400517, -0.440401062979, -0.445048680123, -0.451651705843, -0.457760235328, -0.463733230566, -0.472390233324, -0.480892678638, -0.488468389259, -0.496054924285, -0.505287254412, -0.516712248163, -0.530702320051, -0.543402030632, -0.553376187444, -0.564523083803, -0.576716924164, -0.588673429838, -0.599807622576, -0.6083625871, -0.61629220663, -0.626333930393, -0.638492384934, -0.651370663496, -0.663578126675, -0.675399039036, -0.687337552218, -0.699811934302, -0.712809946825, -0.725971047683, -0.73913751946, -0.752013364118, -0.764774460887, -0.777402692471, -0.789886995534, -0.802209798764, -0.814336203683, -0.826241102491, -0.837954892052, -0.849457493421, -0.860601375162, -0.871314573433, -0.881462999336, -0.890680302818, -0.899437045192, -0.907818061982, -0.915897388877, -0.92305068548, -0.931064125926, -0.939248450894, -0.947470287824, -0.953643840407, -0.957725818058, -0.959847025364, -0.960115447844, -0.959051577361, -0.957073291984, -0.952866371063, -0.945294863841, -0.937774627549, -0.921959850725, -0.901024675728, -0.87472187553, -0.834748962688, -0.733155709571};
    
    const double b2list[101] = {1.24250927283, 1.23820882351, 1.22814744033, 1.20899501569, 1.19187482987, 1.18459068292, 1.17544439367, 1.16014645984, 1.15217431136, 1.14122006693, 1.13199616785, 1.12640121441, 1.121556275, 1.11826795578, 1.11552355075, 1.11238433978, 1.10807035787, 1.10209548232, 1.09668053255, 1.09318287767, 1.09397420715, 1.09653673342, 1.09802015935, 1.10068222122, 1.10643573897, 1.11470254998, 1.12510511039, 1.13755406941, 1.15120517214, 1.16532741744, 1.17824955526, 1.19205557182, 1.20770493667, 1.22624497606, 1.24563734056, 1.26435121321, 1.28310293005, 1.30265970794, 1.32505082997, 1.35327207812, 1.38087484583, 1.40864740792, 1.44367567593, 1.47847338398, 1.51107248182, 1.54412932876, 1.58210883511, 1.62865620041, 1.68826303807, 1.74233039828, 1.78373725125, 1.82868267503, 1.87681370736, 1.92312517747, 1.9656186804, 1.99785596132, 2.02740592217, 2.06495795508, 2.11055684374, 2.15878076242, 2.20440433139, 2.24836632819, 2.29241478185, 2.33795222667, 2.38484419984, 2.43171272272, 2.47786827853, 2.52194361567, 2.56488422632, 2.60668479559, 2.6474159293, 2.6870781889, 2.72548564593, 2.76255774767, 2.79845498187, 2.83311863463, 2.86593650322, 2.89662455284, 2.9246508029, 2.94855630091, 2.97022227107, 2.98999047671, 3.00814380105, 3.02093394891, 3.03817745817, 3.05662290422, 3.07635559396, 3.08574133623, 3.0854061584, 3.07702110074, 3.06226203175, 3.04321477098, 3.01960302068, 2.98621442345, 2.94157099234, 2.89879486315, 2.82337825218, 2.72933462053, 2.61436316871, 2.4455062038, 2.1171325295};
#endif
#if NEWNSNQC
    const unsigned int nsqdim = 126;
    
    memset( coeffs, 0, sizeof( *coeffs ) );
    const double nsetalist[126] = {0., 0.002, 0.004, 0.006, 0.008, 0.01, 0.012, 0.014, 0.016, 0.018, 0.02, 0.022, 0.024, 0.026, 0.028, 0.03, 0.032, 0.034, 0.036, 0.038, 0.04, 0.042, 0.044, 0.046, 0.048, 0.05, 0.052, 0.054, 0.056, 0.058, 0.06, 0.062, 0.064, 0.066, 0.068, 0.07, 0.072, 0.074, 0.076, 0.078, 0.08, 0.082, 0.084, 0.086, 0.088, 0.09, 0.092, 0.094, 0.096, 0.098, 0.1, 0.102, 0.104, 0.106, 0.108, 0.11, 0.112, 0.114, 0.116, 0.118, 0.12, 0.122, 0.124, 0.126, 0.128, 0.13, 0.132, 0.134, 0.136, 0.138, 0.14, 0.142, 0.144, 0.146, 0.148, 0.15, 0.152, 0.154, 0.156, 0.158, 0.16, 0.162, 0.164, 0.166, 0.168, 0.17, 0.172, 0.174, 0.176, 0.178, 0.18, 0.182, 0.184, 0.186, 0.188, 0.19, 0.192, 0.194, 0.196, 0.198, 0.2, 0.202, 0.204, 0.206, 0.208, 0.21, 0.212, 0.214, 0.216, 0.218, 0.22, 0.222, 0.224, 0.226, 0.228, 0.23, 0.232, 0.234, 0.236, 0.238, 0.24, 0.242, 0.244, 0.246, 0.248, 0.25};
    
    const double a1list[126] = {-12.9867168, -12.3922341, -11.8826575, -11.5044206, -11.2010798, -10.9291101, -10.6867579, -10.4603116, -10.2705936, -10.1096886, -9.94972042, -9.7993101, -9.67017397, -9.54568872, -9.42387643, -9.30374853, -9.19233175, -9.08443818, -8.97983162, -8.87983878, -8.78334655, -8.68874044, -8.5958085, -8.50609937, -8.42033859, -8.3368237, -8.25318119, -8.17059891, -8.0893624, -8.00915281, -7.92946769, -7.85089647, -7.77416021, -7.69874159, -7.6230207, -7.54778069, -7.47506574, -7.40420786, -7.33414965, -7.26472564, -7.19583939, -7.12778746, -7.06152986, -6.9966294, -6.93284618, -6.86298297, -6.78959864, -6.73156524, -6.67432199, -6.61818773, -6.56085433, -6.50488932, -6.45206856, -6.40421195, -6.35784803, -6.3138088, -6.26993274, -6.22465506, -6.18289146, -6.14263195, -6.10558528, -6.07015742, -6.03568352, -6.00473068, -5.97536374, -5.9488801, -5.92520739, -5.89986979, -5.87411946, -5.85174778, -5.83384542, -5.82778792, -5.82798078, -5.8256649, -5.82049035, -5.81770024, -5.81990472, -5.82666469, -5.83577969, -5.84678194, -5.8593275, -5.87380902, -5.88835704, -5.90227283, -5.91692664, -5.93375623, -5.95112447, -5.97214313, -5.99595282, -6.01998938, -6.0451582, -6.07037506, -6.09500227, -6.12087467, -6.1457118, -6.16991637, -6.19524357, -6.21896485, -6.24302387, -6.26778731, -6.29335024, -6.32209441, -6.35041468, -6.37940448, -6.40967178, -6.44669514, -6.4822263, -6.51034108, -6.53379526, -6.5510243, -6.5791139, -6.61538596, -6.65893348, -6.70365878, -6.7474139, -6.79400218, -6.8495302, -6.91475643, -6.98767056, -7.06220586, -7.16459696, -7.29175945, -7.43577701, -7.60975023, -7.86628294, -8.249368};
    
    const double a2list[126] = {103.047993, 97.488689, 92.7138991, 89.1259949, 86.2216369, 83.5996035, 81.2588265, 79.0565575, 77.1901259, 75.5962053, 74.0126613, 72.5193589, 71.2189287, 69.960102, 68.7278401, 67.5137535, 66.3813558, 65.2852124, 64.2251046, 63.2086791, 62.2258845, 61.2625581, 60.3162743, 59.4001895, 58.5221931, 57.666497, 56.8109787, 55.9670566, 55.1376064, 54.3195185, 53.5077914, 52.7084763, 51.9289985, 51.1641196, 50.3976552, 49.6369195, 48.9014988, 48.184488, 47.4747413, 46.7739132, 46.0794666, 45.3892075, 44.7162594, 44.0576325, 43.4108107, 42.7414687, 42.0328661, 41.4379676, 40.8519633, 40.2763931, 39.6864942, 39.1075193, 38.5558168, 38.0487152, 37.5534784, 37.0855881, 36.6187593, 36.1354218, 35.6891039, 35.2559943, 34.8506441, 34.4544071, 34.0606759, 33.7102444, 33.3728535, 33.0425553, 32.7368071, 32.424474, 32.1183409, 31.8446322, 31.6055987, 31.4663075, 31.3800944, 31.2792375, 31.1629634, 31.0686905, 31.0129002, 30.9902402, 30.9910664, 31.0100087, 31.0447735, 31.096186, 31.1545669, 31.2147854, 31.2846909, 31.3747464, 31.4774613, 31.6133923, 31.7749697, 31.9441796, 32.1267434, 32.3145467, 32.5026478, 32.7052324, 32.9044074, 33.103595, 33.3182241, 33.5238512, 33.7360635, 33.9573682, 34.1872127, 34.4454878, 34.7018803, 34.9661177, 35.2447179, 35.576539, 35.8979802, 36.1672599, 36.4055123, 36.592871, 36.8680094, 37.2087221, 37.6092315, 38.0164539, 38.4141907, 38.8351392, 39.3247457, 39.8822802, 40.4926366, 41.1147783, 41.9315173, 42.9304965, 44.0528309, 45.3955045, 47.3274085, 50.0808317};
    
    const double a3list[126] = {-112.067972, -105.543106, -99.9455353, -95.7264397, -92.3043493, -89.2092924, -86.4492519, -83.8470286, -81.6361633, -79.7483937, -77.874168, -76.1068934, -74.5632033, -73.0674934, -71.6039815, -70.1633203, -68.8189591, -67.519153, -66.2647734, -65.0620239, -63.8992811, -62.7604276, -61.6423525, -60.5597465, -59.5224349, -58.5119795, -57.5024912, -56.5074902, -55.53044, -54.5676408, -53.6130924, -52.6741387, -51.759812, -50.8637645, -49.9665188, -49.0765995, -48.2171201, -47.379661, -46.5505356, -45.7333476, -44.9242629, -44.118089, -43.3322908, -42.5639012, -41.8098449, -41.043258, -40.2287372, -39.5338046, -38.8499135, -38.1781038, -37.4888781, -36.8113298, -36.1641156, -35.5671876, -34.9832785, -34.4326665, -33.8826254, -33.3121583, -32.7856761, -32.2743216, -31.7933477, -31.3201055, -30.8470575, -30.4267354, -30.0203142, -29.6142881, -29.2352514, -28.8518367, -28.478952, -28.1429549, -27.8435553, -27.6538903, -27.5225161, -27.3767334, -27.2165867, -27.0811087, -26.986459, -26.9258851, -26.8924026, -26.8792128, -26.8840645, -26.9066881, -26.939216, -26.976347, -27.0248923, -27.0956124, -27.1838684, -27.3095255, -27.4637058, -27.6286706, -27.8094514, -27.9976633, -28.1882304, -28.3957603, -28.6018695, -28.8099936, -29.0365609, -29.2555204, -29.4831745, -29.7220365, -29.9712378, -30.2520714, -30.5319354, -30.8213449, -31.1276025, -31.4916256, -31.8453217, -32.1448723, -32.4128347, -32.6260766, -32.9352624, -33.3159744, -33.762117, -34.2149428, -34.6570779, -35.1244967, -35.6653064, -36.2763704, -36.9410027, -37.61746, -38.493775, -39.5607747, -40.7566336, -42.1829659, -44.2164331, -47.0600574};
    
    const double b1list[126] = {-0.0409516519, -0.109149589, -0.18280716, -0.243877638, -0.29427273, -0.337607837, -0.376514835, -0.408102051, -0.430340476, -0.447661929, -0.455696226, -0.455620099, -0.451829661, -0.448183565, -0.444485773, -0.441429709, -0.43796189, -0.435013771, -0.43317134, -0.431090718, -0.428804204, -0.426754544, -0.4248826, -0.422471012, -0.419498947, -0.416391867, -0.413725074, -0.411268992, -0.408996729, -0.407008137, -0.405444696, -0.404208778, -0.403206029, -0.402514117, -0.402371401, -0.402655408, -0.403094271, -0.403632835, -0.404112538, -0.405565386, -0.407338113, -0.407607441, -0.408032297, -0.409123955, -0.410936785, -0.415814407, -0.417806894, -0.420418936, -0.423769675, -0.427415314, -0.431003883, -0.434312853, -0.437542567, -0.440912596, -0.445155727, -0.450792563, -0.456134467, -0.461308333, -0.467704275, -0.475422557, -0.482471914, -0.489063765, -0.496174659, -0.503981188, -0.513338867, -0.525302557, -0.536971352, -0.546967233, -0.555720559, -0.565412016, -0.576313878, -0.586599947, -0.596427341, -0.604860719, -0.611970131, -0.619519476, -0.628821459, -0.639872749, -0.65090668, -0.661411175, -0.671571431, -0.682134421, -0.692764125, -0.703573421, -0.715112958, -0.727369423, -0.738621336, -0.749657644, -0.761066387, -0.77191596, -0.782873523, -0.793717364, -0.804225625, -0.815026186, -0.825216179, -0.835238792, -0.845945128, -0.855787288, -0.865303002, -0.874495405, -0.883003789, -0.891163975, -0.898663309, -0.905882454, -0.91320713, -0.919704848, -0.926061344, -0.933262667, -0.940928324, -0.947831369, -0.953153169, -0.95693697, -0.959330857, -0.960185331, -0.959843982, -0.958686944, -0.956619418, -0.952843922, -0.946447305, -0.939771599, -0.928224314, -0.911651306, -0.891719944, -0.867854188, -0.82502669, -0.73315571};
    
    const double b2list[126] = {-0.00166722842, 0.222250337, 0.448252485, 0.63206119, 0.781799547, 0.910224883, 1.02375301, 1.1157851, 1.17923771, 1.22659824, 1.24771432, 1.24548805, 1.2320439, 1.2192957, 1.2064752, 1.19588522, 1.18435194, 1.17453472, 1.16812456, 1.16113472, 1.15387353, 1.14747797, 1.14166252, 1.13468889, 1.12673472, 1.11873428, 1.11191795, 1.10593347, 1.10080563, 1.09667658, 1.09370713, 1.09190117, 1.09129942, 1.09194833, 1.0939851, 1.09725207, 1.1015065, 1.10645069, 1.11139159, 1.11928716, 1.12839242, 1.13375028, 1.14006609, 1.14868099, 1.15958421, 1.18126715, 1.19236439, 1.20564444, 1.22112152, 1.23765157, 1.25462197, 1.27123058, 1.28793848, 1.30524443, 1.32551845, 1.34960096, 1.37341634, 1.39766424, 1.4249208, 1.45597978, 1.48516877, 1.51368441, 1.5446322, 1.5767945, 1.61477856, 1.66514258, 1.71502079, 1.75728168, 1.79328757, 1.83222422, 1.87523717, 1.91514764, 1.95277569, 1.98471429, 2.01131332, 2.03946121, 2.07428419, 2.11572946, 2.15704477, 2.19631812, 2.23416378, 2.27326935, 2.3122869, 2.35158306, 2.39309063, 2.43665605, 2.47607718, 2.51394563, 2.55248145, 2.58860308, 2.62460385, 2.65980703, 2.6935071, 2.72765112, 2.75939003, 2.79018574, 2.82260365, 2.85186527, 2.87952241, 2.90554772, 2.92874698, 2.94977991, 2.96834961, 2.98551006, 3.0023529, 3.01499219, 3.02720026, 3.04304093, 3.06075921, 3.07707208, 3.08528399, 3.08627843, 3.08073147, 3.06990182, 3.05552041, 3.03783353, 3.0153212, 2.98607118, 2.94810984, 2.90941344, 2.85250477, 2.77660316, 2.68817051, 2.58115094, 2.4104722, 2.11713253};
#endif
    
    /* Stuff for interpolating the NQC data */
    gsl_spline    *spline = NULL;
    gsl_interp_accel *acc = NULL;
    /* Interpolate nonspin NQC data in the mass ratio dimension */
    spline = gsl_spline_alloc( gsl_interp_cspline, nsqdim );
    acc    = gsl_interp_accel_alloc();
    gsl_spline_init( spline, nsetalist, a1list, nsqdim );
    gsl_interp_accel_reset( acc );
    coeffs->a1 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, nsetalist, a2list, nsqdim );
    gsl_interp_accel_reset( acc );
    coeffs->a2 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, nsetalist, a3list, nsqdim );
    gsl_interp_accel_reset( acc );
    coeffs->a3 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, nsetalist, b1list, nsqdim );
    gsl_interp_accel_reset( acc );
    coeffs->b1 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_init( spline, nsetalist, b2list, nsqdim );
    gsl_interp_accel_reset( acc );
    coeffs->b2 = gsl_spline_eval( spline, eta, acc );
    gsl_spline_free(spline);
    gsl_interp_accel_free(acc);
    /* Andrea and I have different sign conventions, so I need to put a minus sign in front */
    coeffs->b1 = - coeffs->b1;
    coeffs->b2 = - coeffs->b2;
    
    REAL8 dM  = fabs(m1 - m2)/(m1+m2);
    REAL8 chiA = chiAin;
    if ( m2 > m1 )
    {
        chiA = -chiA;
    }
    REAL8 chi = a / ( 1.0 - 2.0 * eta );
    REAL8 chiAmax, chiAmed, chiAmin;
    REAL8 cmax, cmed, cmin;
    REAL8 nqcmax, nqcmed, nqcmin;
    EOBNonQCCoeffs chiAmaxCoeffs;
    EOBNonQCCoeffs chiAmedCoeffs;
    EOBNonQCCoeffs chiAminCoeffs;
    
    XLALSimIMRGetEOBCalibratedSpinNQCv2chiAmax( &chiAmaxCoeffs, l, m, eta, a );
    XLALSimIMRGetEOBCalibratedSpinNQCv2chiAmed( &chiAmedCoeffs, l, m, eta, a );
    XLALSimIMRGetEOBCalibratedSpinNQCv2chiAmin( &chiAminCoeffs, l, m, eta, a );
    
    if ( eta < 1.0e-15 ) {
        chiAmax = (chi + 1.0) / 2.0;
        chiAmin = (chi - 1.0) / 2.0;
    }
    else {
        if ( chi <= 0 ) {
            chiAmax = (1.+chi) * (1.-2.*eta) / (1.+dM-2.*eta);
            if ( (1.+dM-2.*eta+2.*chi*(1.-2.*eta))/(1.-dM-2.*eta) < 1. ) {
                chiAmin = -(1.+chi) * (1.-2.*eta) / (1.-dM-2.*eta);
            }
            else {
                chiAmin = -(1.-chi) * (1.-2.*eta) / (1.+dM-2.*eta);
            }
        }
        else {
            chiAmin = -(1.-chi) * (1.-2.*eta) / (1.+dM-2.*eta);
            if ( -(1.+dM-2.*eta-2.*chi*(1.-2.*eta))/(1.-dM-2.*eta) > -1. ) {
                chiAmax = (1.-chi) * (1.-2.*eta) / (1.-dM-2.*eta);
            }
            else {
                chiAmax = (1.+chi) * (1.-2.*eta) / (1.+dM-2.*eta);
            }
        }
    }
    chiAmed = (chiAmax + chiAmin) / 2.0;
    
    if ( chiAmax < 1.0e-15 )
    {
        cmax = 1.0;
        cmed = 0.0;
        cmin = 0.0;
    }
    else
    {
        cmax =  (chiA-chiAmed) * (chiA-chiAmin) / (chiAmax-chiAmed) / (chiAmax-chiAmin);
        cmed = -(chiA-chiAmax) * (chiA-chiAmin) / (chiAmax-chiAmed) / (chiAmed-chiAmin);
        cmin =  (chiA-chiAmax) * (chiA-chiAmed) / (chiAmax-chiAmin) / (chiAmed-chiAmin);
    }
    /*printf( "chi = %f\n",chi );
     printf( "cAmax = %f, cAmed = %f, cAmin = %f, cA = %f\n",  chiAmax, chiAmed, chiAmin, chiA );
     printf( "cmax = %f, cmed = %f, cmin = %f\n",  cmax, cmed, cmin );*/
    
    nqcmax = chiAmaxCoeffs.a3S;
    nqcmed = chiAmedCoeffs.a3S;
    nqcmin = chiAminCoeffs.a3S;
    coeffs->a3S = cmax*nqcmax + cmed*nqcmed + cmin*nqcmin;
    //printf( "a3max = %f, a3med = %f, a3min = %f, a3 = %f\n",
    //  nqcmax, nqcmed, nqcmin, coeffs->a3S );
    nqcmax = chiAmaxCoeffs.a4;
    nqcmed = chiAmedCoeffs.a4;
    nqcmin = chiAminCoeffs.a4;
    coeffs->a4  = cmax*nqcmax + cmed*nqcmed + cmin*nqcmin;
    //printf( "a4max = %f, a4med = %f, a4min = %f, a4 = %f\n",
    //  nqcmax, nqcmed, nqcmin, coeffs->a4 );
    nqcmax = chiAmaxCoeffs.a5;
    nqcmed = chiAmedCoeffs.a5;
    nqcmin = chiAminCoeffs.a5;
    coeffs->a5  = cmax*nqcmax + cmed*nqcmed + cmin*nqcmin;
    //printf( "a5max = %f, a5med = %f, a5min = %f, a5 = %f\n",
    //  nqcmax, nqcmed, nqcmin, coeffs->a5 );
    
    /*coeffs->a3S = 1775.87095769;
     coeffs->a4  = -5737.3899946;
     coeffs->a5  = 4641.54406645;
     coeffs->a3S = 5665.39993599;
     coeffs->a4  = -18576.8784995;
     coeffs->a5  = 15237.7821969;*/
    /*coeffs->a3S = 10322.2486021;
     coeffs->a4  = -32666.5282038;
     coeffs->a5  = 25854.6695187;*/
    /*coeffs->a3S = 10319.795279;
     coeffs->a4  = -32671.950978;
     coeffs->a5  = 25869.632474;*/
    /*coeffs->a3S = -145.193303874;
     coeffs->a4 = 322.844528702;
     coeffs->a5 = -157.092596165;*/
    /*coeffs->a3S = 2278.91726941;
     coeffs->a4 = -7128.39166422;
     coeffs->a5 = 5583.15077297;*/
    
    nqcmax = chiAmaxCoeffs.b3;
    nqcmed = chiAmedCoeffs.b3;
    nqcmin = chiAminCoeffs.b3;
    coeffs->b3  = cmax*nqcmax + cmed*nqcmed + cmin*nqcmin;
    //printf( "b3max = %f, b3med = %f, b3min = %f, b3 = %f\n",
    //  nqcmax, nqcmed, nqcmin, coeffs->b3 );
    nqcmax = chiAmaxCoeffs.b4;
    nqcmed = chiAmedCoeffs.b4;
    nqcmin = chiAminCoeffs.b4;
    coeffs->b4  = cmax*nqcmax + cmed*nqcmed + cmin*nqcmin;
    //printf( "b4max = %f, b4med = %f, b4min = %f, b4 = %f\n",
    //  nqcmax, nqcmed, nqcmin, coeffs->b4 );
    //coeffs->b3  = 2447.71718687;
    //coeffs->b4  = -3916.34664737;
    //  coeffs->b3  = 80437.4825363;
    //  coeffs->b4  = -124963.109912;
    //  coeffs->b3  = 80675.052985;
    //  coeffs->b4  = -125321.741652;
    /*coeffs->b3 = 422.973992013;
     coeffs->b4 = -655.437950088;*/
    return XLAL_SUCCESS;
}